<?php


class CNGeoIP {
    function __construct() {
	$this->is_valid = false;
	$this->error_message = "";
	$this->fp_binary_tree = fopen(dirname(__FILE__)."/".self::FILENAME_BINARY_TREE, 'r');
	if ( !$this->fp_binary_tree ) {
	    $this->error_message = "Can't open binary tree file \"".self::FILENAME_BINARY_TREE."\"";
	    return false;
	}
	$data = fread($this->fp_binary_tree, 4);
	$unp = unpack('Ndata', $data);
	if ( (($unp['data'] >> 8) & 0xffffff) != 0xcd301b ) {
	    $this->error_message = "Invalid datafile signature \"".(($unp['data'] >> 8) & 0xffffff)."\"";
	    return false;
	}
	if ( (($unp['data']) & 0xff) != 0x05 ) {
	    $this->error_message = "Incompatible datafile version \"".(($unp['data']) & 0xff)."\"";
	    return false;
	}
	$this->level_size = array();
	for(;;) {
	    $data = fread($this->fp_binary_tree, 1);
	    $unp = unpack('Cdata', $data);
	    $hi = ($unp['data']>>4) & 0x0f;
	    $lo = $unp['data'] & 0x0f;
	    $this->level_size[] = $hi;
	    if ( $hi == 0 )
		break;
	    $this->level_size[] = $lo;
	    if ( $lo == 0 )
		break;
	}
	$this->start_offset = ftell($this->fp_binary_tree);
	$this->is_valid = true;
    }


    function __destruct() {
	fclose($this->fp_binary_tree);
    }
    

    function get_place_by_ip($ip) {
	if ( !$this->is_valid )
	    return false;
	    
	$ipint = ip2long($ip);
	$shift = 32;
	$offset = $this->start_offset;
    
	for($l=0; $l<sizeof($this->level_size); $l++) {
	    $shift -= $this->level_size[$l];
	    $index = (($ipint>>$shift)) & ((1<<$this->level_size[$l])-1);
	    $tell = $offset+$index*4;
	    fseek($this->fp_binary_tree, $tell, 0);
	    $data = fread($this->fp_binary_tree, 4);
	    $unp = unpack('Ndata',$data);
	    $offset = $unp['data'] & 0xffffffff;
	    if ( $offset & 0x80000000 ) {
		$tell = $offset & 0x7fffffff;
		fseek($this->fp_binary_tree, $tell, 0);
		$data = fread($this->fp_binary_tree, 2);
		$unp = unpack('nlen', $data);
		$length = $unp['len'] & 0xffff;
		if ( !$length )
		    return "";
		$info = fread($this->fp_binary_tree, $length);
		return $info;
	    }
	}
	return false;
    }
    
    function get_description_by_ip($ip) {
	if ( !$this->is_valid )
	    return false;
        return $this->get_description_by_place($this->get_place_by_ip($ip));
    }

    function get_description_by_place($place) {
	if ( !$this->is_valid )
	    return false;
	if ( preg_match("/^\/.*$/", $place) ) {
	    $results = array();
	    while ( preg_match("/^(\/[^\s]+)*\/[^\s]+$/", $place, $matches) ) {
		$results[] = $GLOBALS['cngeo5dict'][$place];
		$place = $matches[1];
	    }
	}
	else {
	    $results = array('special'=>$place);
	}
	return $results;
    }
    
    const FILENAME_BINARY_TREE = "cngeoip.dat";
    
    private $is_valid;
    function is_valid() {
	return $this->is_valid;
    }
    
    private $error_message;
    function get_error() {
	return $this->error_message;
    }
    
    private $level_size;
    private $start_offset;
    
    private $fp_binary_tree;
    private $dbh;
    
}


$GLOBALS['cngeo5dict'] = array(
"" => array("type"=>"world"),
"/AB" => array("type"=>"country", "pop"=>180000, "name_ru"=>"Абхазия", "name_en"=>"Abkhazia", "lang"=>"ru", "lat"=>43, "lon"=>40.9833333333333),
"/AD" => array("type"=>"country", "pop"=>84484, "name_ru"=>"Андорра", "name_ca"=>"Andorra", "name_en"=>"Andorra", "lang"=>"ca", "lat"=>42.55, "lon"=>1.55),
"/AE" => array("type"=>"country", "pop"=>6122000, "name_ru"=>"Арабские Эмираты", "name_en"=>"UAE", "lang"=>"ar", "lat"=>27.78, "lon"=>54.00),
"/AF" => array("type"=>"country", "pop"=>32000987, "name_ru"=>"Афганистан", "name_ps"=>"افغانستان", "name_en"=>"Afghanistan", "lang"=>"ps", "lat"=>33.93, "lon"=>66.18),
"/AG" => array("type"=>"country", "pop"=>67448, "name_ru"=>"Антигуа и Барбуда", "name_en"=>"Antigua and Barbuda", "lang"=>"en", "lat"=>17.31, "lon"=>-62.0),
"/AL" => array("type"=>"country", "pop"=>3600523, "name_sq"=>"Shqipëria", "name_ru"=>"Албания", "name_en"=>"Albania", "lang"=>"sq", "lat"=>41.35, "lon"=>19.98),
"/AM" => array("type"=>"country", "pop"=>3240700, "name_ru"=>"Армения", "name_hy"=>"Հայաստան", "name_en"=>"Armenia", "lang"=>"hy", "lat"=>40.38, "lon"=>44.95),
"/AO" => array("type"=>"country", "pop"=>15941000, "name_ru"=>"Ангола", "name_pt"=>"Angola", "name_en"=>"Angola", "lang"=>"pt", "lat"=>-12.35, "lon"=>17.35),
"/AR" => array("type"=>"country", "pop"=>40482000, "name_ru"=>"Аргентина", "name_es"=>"Argentina", "name_en"=>"Argentina", "lang"=>"es", "lat"=>-37.18, "lon"=>-67.36),
"/AT" => array("type"=>"country", "pop"=>8151000, "name_de"=>"Österreich", "name_ru"=>"Австрия", "name_en"=>"Austria", "lang"=>"de", "lat"=>47.6, "lon"=>13.78),
"/AU" => array("type"=>"country", "pop"=>20001546, "name_ru"=>"Австралия", "name_en"=>"Australia", "lang"=>"en", "lat"=>-25.00, "lon"=>135.00),
"/AZ" => array("type"=>"country", "pop"=>8922000, "name_ru"=>"Азербайджан", "name_az"=>"Azərbaycan", "name_en"=>"Azerbaijan", "lang"=>"az", "lat"=>40.1, "lon"=>47.31),
"/BA" => array("type"=>"country", "pop"=>4613414, "name_ru"=>"Босния и Герцеговина", "name_en"=>"Bosnia and Herzegovina", "lang"=>"sr", "lat"=>44.1, "lon"=>17.96),
"/BB" => array("type"=>"country", "pop"=>279254, "name_ru"=>"Барбадос", "name_en"=>"Barbados", "lang"=>"en", "lat"=>13.16, "lon"=>-59.53),
"/BD" => array("type"=>"country", "pop"=>153546896, "name_bn"=>"গণপ্রজাতন্ত্রী বাংলাদেশ", "name_ru"=>"Бангладеш", "name_en"=>"Bangladesh", "lang"=>"bn", "lat"=>24.01, "lon"=>89.86),
"/BE" => array("type"=>"country", "pop"=>10665867, "name_ru"=>"Бельгия", "name_fr"=>"Belgique", "name_en"=>"Belgium", "lang"=>"fr", "lat"=>50.53, "lon"=>4.76),
"/BF" => array("type"=>"country", "pop"=>13440500, "name_ru"=>"Буркина Фасо", "name_en"=>"Burkina Faso", "lang"=>"fr", "lat"=>12.26, "lon"=>-2.06),
"/BG" => array("type"=>"country", "pop"=>7606551, "name_bg"=>"България", "name_ru"=>"Болгария", "name_en"=>"Bulgaria", "lang"=>"bg", "lat"=>42.65, "lon"=>25.4),
"/BH" => array("type"=>"country", "pop"=>791000, "name_ru"=>"Бахрейн", "name_ar"=>"مملكة البحرين‎‎", "name_en"=>"Bahrain", "lang"=>"ar", "lat"=>26.06, "lon"=>50.55),
"/BI" => array("type"=>"country", "pop"=>8391000, "name_ru"=>"Бурунди", "name_en"=>"Burundi", "lang"=>"rn", "lat"=>-3.66, "lon"=>29.81),
"/BJ" => array("type"=>"country", "pop"=>7460025, "name_ru"=>"Республика Бенин", "name_fr"=>"Bénin", "name_en"=>"Republic of Benin", "lang"=>"fr", "lat"=>8.83, "lon"=>2.18),
"/BN" => array("type"=>"country", "pop"=>372361, "name_ru"=>"Бруней", "name_en"=>"Brunei Darussalam", "lang"=>"ms", "lat"=>4.4, "lon"=>114.56),
"/BO" => array("type"=>"country", "pop"=>9627269, "name_ru"=>"Боливия", "name_en"=>"Bolivia", "lang"=>"es", "lat"=>-16.71, "lon"=>-64.66),
"/BR" => array("type"=>"country", "pop"=>190256710, "name_ru"=>"Бразилия", "name_en"=>"Brazil", "lang"=>"pt", "lat"=>-10.65, "lon"=>-52.95),
"/BS" => array("type"=>"country", "pop"=>301790, "name_ru"=>"Багамские Острова", "name_en"=>"Bahamas", "lang"=>"en", "lat"=>23.91, "lon"=>-77.66),
"/BT" => array("type"=>"country", "pop"=>672425, "name_ru"=>"Бутан", "name_en"=>"Bhutan", "lang"=>"dz", "lat"=>27.45, "lon"=>90.5),
"/BU" => array("type"=>"country", "pop"=>215446, "name_ru"=>"Вануату", "name_en"=>"Vanuatu", "lang"=>"en", "lat"=>-16.63, "lon"=>168.01),
"/BW" => array("type"=>"country", "pop"=>1573267, "name_ru"=>"Ботсвана", "name_en"=>"Botswana", "lang"=>"en", "lat"=>-21.76, "lon"=>24.03),
"/BY" => array("type"=>"country", "pop"=>9663500, "name_ru"=>"Белоруссия", "name_by"=>"Беларусь", "name_en"=>"Belarus", "lang"=>"by", "lat"=>53.51, "lon"=>28.76),
"/BY/BR" => array("type"=>"region", "pop"=>1440000, "name_ru"=>"Брестская область", "name_en"=>"Brest Voblast", "name_by"=>"Брэсцкая вобласць", "lang"=>"by"),
"/BY/BR/BRNV" => array("type"=>"city", "pop"=>169000, "name_ru"=>"Барановичи", "name_en"=>"Baranovichi", "name_by"=>"Баранавічы", "lang"=>"by", "lat"=>53.1166666666667, "lon"=>25.9833333333333),
"/BY/BR/BRST" => array("type"=>"city", "pop"=>318000, "name_ru"=>"Брест", "name_en"=>"Brest", "name_by"=>"Брэст", "lang"=>"by", "lat"=>52.1333333333333, "lon"=>23.6666666666667),
"/BY/BR/PNSK" => array("type"=>"city", "pop"=>131000, "name_ru"=>"Пинск", "name_en"=>"Pinsk", "name_by"=>"Пiнск", "lang"=>"by", "lat"=>52.1166666666667, "lon"=>26.0833333333333),
"/BY/HO" => array("type"=>"region", "pop"=>1464500, "name_ru"=>"Гомельская область", "name_en"=>"Homiel Voblast", "name_by"=>"Гомельская вобласць", "lang"=>"by"),
"/BY/HO/HOML" => array("type"=>"city", "pop"=>498700, "name_ru"=>"Гомель", "name_en"=>"Homiel", "name_by"=>"Гомель", "lang"=>"by", "lat"=>52.4416666666667, "lon"=>30.9833333333333),
"/BY/HO/MOZR" => array("type"=>"city", "pop"=>112500, "name_ru"=>"Мозырь", "name_en"=>"Mazyr", "name_by"=>"Мазыр", "lang"=>"by", "lat"=>52.05, "lon"=>29.2333333333333),
"/BY/HO/RCHC" => array("type"=>"city", "pop"=>65532, "name_ru"=>"Речица", "name_en"=>"Rechytsa", "name_by"=>"Рэчыца", "lang"=>"by", "lat"=>52.3638888888889, "lon"=>30.3947222222222),
"/BY/HO/SVTL" => array("type"=>"city", "pop"=>73400, "name_ru"=>"Светлогорск", "name_en"=>"Svetlahorsk", "name_by"=>"Светлагорск", "lang"=>"by", "lat"=>52.6333333333333, "lon"=>29.7333333333333),
"/BY/HO/ZHLB" => array("type"=>"city", "pop"=>73089, "name_ru"=>"Жлобин", "name_en"=>"Zhlobin", "name_by"=>"Жлобiн", "lang"=>"by", "lat"=>52.9, "lon"=>30.0333333333333),
"/BY/HR" => array("type"=>"region", "pop"=>1102800, "name_ru"=>"Гродненская область", "name_en"=>"Hrodna Voblast", "name_by"=>"Гродзенская вобласць", "lang"=>"by"),
"/BY/HR/HRDN" => array("type"=>"city", "pop"=>338200, "name_ru"=>"Гродно", "name_en"=>"Hrodna", "name_by"=>"Гро́дна", "lang"=>"by", "lat"=>53.6666666666667, "lon"=>23.8166666666667),
"/BY/HR/LIDA" => array("type"=>"city", "pop"=>95800, "name_ru"=>"Лида", "name_en"=>"Lida", "name_by"=>"Лiда", "lang"=>"by", "lat"=>53.8872222222222, "lon"=>25.3027777777778),
"/BY/HR/NVGR" => array("type"=>"city", "pop"=>30800, "name_ru"=>"Новогрудок", "name_en"=>"Novahrudak", "name_by"=>"Новагрудак", "lang"=>"by", "lat"=>53.5833333333333, "lon"=>25.8166666666667),
"/BY/HR/SLNM" => array("type"=>"city", "pop"=>51600, "name_ru"=>"Слоним", "name_en"=>"Slonim", "name_by"=>"Слонiм", "lang"=>"by", "lat"=>53.0833333333333, "lon"=>25.3166666666667),
"/BY/HR/SMRG" => array("type"=>"city", "pop"=>36900, "name_ru"=>"Сморгонь", "name_en"=>"Smarhon", "name_by"=>"Смаргонь", "lang"=>"by", "lat"=>54.4836111111111, "lon"=>26.4),
"/BY/MA" => array("type"=>"region", "pop"=>367700, "name_ru"=>"Могилевская область", "name_en"=>"Mahilyov Voblast", "name_by"=>"Магілёўская вобласць", "lang"=>"by"),
"/BY/MA/BBRS" => array("type"=>"city", "pop"=>219000, "name_ru"=>"Бобруйск", "name_en"=>"Babruysk", "name_by"=>"Бабруйск", "lang"=>"by", "lat"=>53.1, "lon"=>29.2166666666667),
"/BY/MA/KRCH" => array("type"=>"city", "pop"=>28200, "name_ru"=>"Кричев", "name_en"=>"Krychaw", "name_by"=>"Крычаў", "lang"=>"by", "lat"=>53.7194444444444, "lon"=>31.7138888888889),
"/BY/MA/MGLV" => array("type"=>"city", "pop"=>369200, "name_ru"=>"Могилев", "name_en"=>"Mahilyov", "name_by"=>"Магілёў", "lang"=>"by", "lat"=>53.9166666666667, "lon"=>30.35),
"/BY/MA/MSTS" => array("type"=>"city", "pop"=>11700, "name_ru"=>"Мстиславь", "name_en"=>"Mstsislaw", "name_by"=>"Мсці́слаў", "lang"=>"by", "lat"=>54.0166666666667, "lon"=>31.7166666666667),
"/BY/MI" => array("type"=>"region", "pop"=>1466000, "name_ru"=>"Минская область", "name_en"=>"Minsk Voblast", "name_by"=>"Мінская вобласць", "lang"=>"by"),
"/BY/MI/BRSV" => array("type"=>"city", "pop"=>154000, "name_ru"=>"Борисов", "name_en"=>"Barysaw", "name_by"=>"Барысаў", "lang"=>"by", "lat"=>54.2333333333333, "lon"=>28.5),
"/BY/MI/MLDC" => array("type"=>"city", "pop"=>98400, "name_ru"=>"Молодечно", "name_en"=>"Maladzyechna", "name_by"=>"Маладзечна", "lang"=>"by", "lat"=>54.3136111111111, "lon"=>26.8516666666667),
"/BY/MI/NSVZ" => array("type"=>"city", "pop"=>14300, "name_ru"=>"Несвиж", "name_en"=>"Nesvizh", "name_by"=>"Нясвіж", "lang"=>"by", "lat"=>53.2166666666667, "lon"=>26.6666666666667),
"/BY/MI/SLCK" => array("type"=>"city", "pop"=>62300, "name_ru"=>"Слуцк", "name_en"=>"Slutsk", "name_by"=>"Слуцк", "lang"=>"by", "lat"=>53.0166666666667, "lon"=>27.55),
"/BY/MNSK" => array("type"=>"city", "pop"=>1841800, "name_ru"=>"Минск", "name_en"=>"Minsk", "name_by"=>"Мінск", "lang"=>"by", "lat"=>53.9166666666667, "lon"=>27.55),
"/BY/VI" => array("type"=>"region", "pop"=>1265300, "name_ru"=>"Витебская область", "name_en"=>"Vitsebsk Voblast", "name_by"=>"Віцебская вобласць", "lang"=>"by"),
"/BY/VI/NVPL" => array("type"=>"city", "pop"=>101000, "name_ru"=>"Новополоцк", "name_en"=>"Novapolatsk", "name_by"=>"Новаполацк", "lang"=>"by", "lat"=>55.5333333333333, "lon"=>28.6666666666667),
"/BY/VI/ORSH" => array("type"=>"city", "pop"=>139000, "name_ru"=>"Орша", "name_en"=>"Orsha", "name_by"=>"Орша", "lang"=>"by", "lat"=>54.5, "lon"=>30.4),
"/BY/VI/PLCK" => array("type"=>"city", "pop"=>83700, "name_ru"=>"Полоцк", "name_en"=>"Polotsk", "name_by"=>"Полацк", "lang"=>"by", "lat"=>55.4833333333333, "lon"=>28.8),
"/BY/VI/VTBS" => array("type"=>"city", "pop"=>347500, "name_ru"=>"Витебск", "name_en"=>"Vitsebsk", "name_by"=>"Віцебск", "lang"=>"by", "lat"=>55.1833333333333, "lon"=>30.1666666666667),
"/BZ" => array("type"=>"country", "pop"=>301270, "name_ru"=>"Белиз", "name_en"=>"Belize", "lang"=>"en", "lat"=>17.25, "lon"=>-88.75),
"/CA" => array("type"=>"country", "pop"=>34000000, "name_ru"=>"Канада", "name_en"=>"Canada", "lang"=>"en", "lat"=>45.00, "lon"=>-75.00),
"/CD" => array("type"=>"country", "pop"=>63655000, "name_ru"=>"ДР Конго", "name_en"=>"Congo, the Democratic Republic of the", "lang"=>"fr", "lat"=>-2.33, "lon"=>22.8),
"/CF" => array("type"=>"country", "pop"=>3799897, "name_ru"=>"Центральноафриканская Республика", "name_en"=>"Central African Republic", "lang"=>"fr", "lat"=>6.96, "lon"=>20.61),
"/CG" => array("type"=>"country", "pop"=>3999000, "name_ru"=>"Конго", "name_en"=>"Congo", "lang"=>"fr", "lat"=>-0.75, "lon"=>15.38),
"/CH" => array("type"=>"country", "pop"=>7700200, "name_de"=>"Schweiz", "name_ru"=>"Швейцария", "name_en"=>"Switzerland", "lang"=>"de", "lat"=>46.8, "lon"=>8.23),
"/CI" => array("type"=>"country", "pop"=>21075000, "name_ru"=>"Кот-д'Ивуар", "name_fr"=>"Côte d’Ivoire", "name_en"=>"Côte d'Ivoire", "lang"=>"fr", "lat"=>8, "lon"=>-6),
"/CL" => array("type"=>"country", "pop"=>15980912, "name_ru"=>"Чили", "name_en"=>"Chile", "lang"=>"es", "lat"=>-36.95, "lon"=>-71.5),
"/CM" => array("type"=>"country", "pop"=>17795000, "name_ru"=>"Камерун", "name_en"=>"Cameroon", "lang"=>"fr", "lat"=>5.11, "lon"=>11.91),
"/CN" => array("type"=>"country", "pop"=>1322178190, "name_cn"=>"中國", "name_ru"=>"Китай", "name_en"=>"China", "lang"=>"cn", "lat"=>32.8, "lon"=>103.08),
"/CO" => array("type"=>"country", "pop"=>42954279, "name_ru"=>"Колумбия", "name_en"=>"Colombia", "lang"=>"es", "lat"=>3.8, "lon"=>-73.91),
"/CR" => array("type"=>"country", "pop"=>4327000, "name_ru"=>"Коста-Рика", "name_en"=>"Costa Rica", "lang"=>"es", "lat"=>9.5, "lon"=>-84.25),
"/CU" => array("type"=>"country", "pop"=>11461842, "name_ru"=>"Куба", "name_es"=>"Cuba", "name_en"=>"Cuba", "lang"=>"es", "lat"=>21.98, "lon"=>-79.03),
"/CV" => array("type"=>"country", "pop"=>542000, "name_ru"=>"Кабо-Верде", "name_en"=>"Cape Verde", "lang"=>"pt", "lat"=>15.91, "lon"=>-24.08),
"/CY" => array("type"=>"country", "pop"=>771657, "name_ru"=>"Кипр", "name_el"=>"Κύπρος", "name_en"=>"Cyprus", "lang"=>"el", "lat"=>35.16, "lon"=>33.35),
"/CZ" => array("type"=>"country", "pop"=>10403100, "name_cs"=>"Česko", "name_ru"=>"Чехия", "name_en"=>"Czech Republic", "lang"=>"cs", "lat"=>49.75, "lon"=>15.74),
"/DE" => array("type"=>"country", "pop"=>82060000, "name_de"=>"Deutschland", "name_ru"=>"Германия", "name_en"=>"Germany", "lang"=>"de", "lat"=>51.51, "lon"=>9.91),
"/DJ" => array("type"=>"country", "pop"=>476703, "name_ru"=>"Джибути", "name_en"=>"Djibouti", "lang"=>"ar", "lat"=>11.8, "lon"=>42.43),
"/DK" => array("type"=>"country", "pop"=>5511791, "name_ru"=>"Дания", "name_da"=>"Danmark", "name_en"=>"Denmark", "lang"=>"da", "lat"=>55.78, "lon"=>9.78),
"/DM" => array("type"=>"country", "pop"=>72512, "name_ru"=>"Доминика", "name_en"=>"Dominica", "lang"=>"en", "lat"=>15.41, "lon"=>-61.33),
"/DO" => array("type"=>"country", "pop"=>9183984, "name_ru"=>"Доминиканская Республика", "name_en"=>"Dominican Republic", "lang"=>"es", "lat"=>18.72, "lon"=>-70.18),
"/DZ" => array("type"=>"country", "pop"=>32531853, "name_ru"=>"Алжир", "name_ar"=>"الجزائر‎‎", "name_en"=>"Algeria", "lang"=>"ar", "lat"=>27.25, "lon"=>3.45),
"/EC" => array("type"=>"country", "pop"=>13850000, "name_ru"=>"Эквадор", "name_en"=>"Ecuador", "lang"=>"es", "lat"=>-1.8, "lon"=>-78.25),
"/EE" => array("type"=>"country", "pop"=>1340300, "name_et"=>"Eesti", "name_ru"=>"Эстония", "name_en"=>"Estonia", "lang"=>"et", "lat"=>58.7, "lon"=>25.51),
"/EG" => array("type"=>"country", "pop"=>77498000, "name_ru"=>"Египет", "name_ar"=>"مصر‎‎", "name_en"=>"Egypt", "lang"=>"ar", "lat"=>26.03, "lon"=>29.21),
"/ER" => array("type"=>"country", "pop"=>4401009, "name_ru"=>"Эритрея", "name_en"=>"Eritrea", "lang"=>"ar", "lat"=>15.48, "lon"=>38.25),
"/ES" => array("type"=>"country", "pop"=>40847371, "name_ru"=>"Испания", "name_es"=>"España", "name_en"=>"Spain", "lang"=>"es", "lat"=>39.92, "lon"=>-1.80),
"/ET" => array("type"=>"country", "pop"=>85237000, "name_ru"=>"Эфиопия", "name_en"=>"Ethiopia", "lang"=>"am", "lat"=>8.3, "lon"=>39.11),
"/EU" => array("type"=>"country", "pop"=>499794855, "name_ru"=>"Европейский Союз", "name_en"=>"European Union", "lat"=>50.1725, "lon"=>9.15),
"/FI" => array("type"=>"country", "pop"=>5340093, "name_fi"=>"Suomi", "name_ru"=>"Финляндия", "name_sv"=>"Finland", "name_en"=>"Finland", "lang"=>"fi", "lat"=>63.4, "lon"=>25.85),
"/FJ" => array("type"=>"country", "pop"=>905949, "name_ru"=>"Фиджи", "name_en"=>"Fiji", "lang"=>"fj", "lat"=>-17.00, "lon"=>179),
"/FM" => array("type"=>"country", "pop"=>107862, "name_ru"=>"Федеративные Штаты Микронезии", "name_en"=>"Federated States of Micronesia", "lang"=>"en", "lat"=>7.5, "lon"=>150.25),
"/FO" => array("type"=>"country", "pop"=>49006, "name_ru"=>"Фарерские острова", "name_da"=>"Færøerne", "name_en"=>"Faroe Islands", "lang"=>"da", "lat"=>62.0166666666667, "lon"=>-6.76666666666667),
"/FR" => array("type"=>"country", "pop"=>64473140, "name_ru"=>"Франция", "name_fr"=>"France", "name_en"=>"France", "lang"=>"fr", "lat"=>46.31, "lon"=>2.53),
"/GA" => array("type"=>"country", "pop"=>1497525, "name_ru"=>"Габон", "name_en"=>"Gabon", "lang"=>"fr", "lat"=>-0.68, "lon"=>11.5),
"/GB" => array("type"=>"country", "pop"=>60776238, "name_ru"=>"Великобритания", "name_en"=>"United Kingdom", "lang"=>"en", "lat"=>53.55, "lon"=>-2.43),
"/GD" => array("type"=>"country", "pop"=>89502, "name_ru"=>"Гренада", "name_en"=>"Grenada", "lang"=>"en", "lat"=>12.11, "lon"=>-61.66),
"/GE" => array("type"=>"country", "pop"=>4260000, "name_ka"=>"საქართველო", "name_ru"=>"Грузия", "name_en"=>"Georgia", "lang"=>"ka", "lat"=>42.01, "lon"=>43.73),
"/GH" => array("type"=>"country", "pop"=>21029853, "name_ru"=>"Гана", "name_en"=>"Ghana", "lang"=>"en", "lat"=>7.68, "lon"=>-0.98),
"/GM" => array("type"=>"country", "pop"=>1700000, "name_ru"=>"Гамбия", "name_en"=>"Gambia", "lang"=>"en", "lat"=>13.5, "lon"=>-15.5),
"/GN" => array("type"=>"country", "pop"=>9467866, "name_ru"=>"Гвинея", "name_en"=>"Guinea", "lang"=>"fr", "lat"=>10.96, "lon"=>-10.76),
"/GQ" => array("type"=>"country", "pop"=>551201, "name_ru"=>"Экваториальная Гвинея", "name_en"=>"Equatorial Guinea", "lang"=>"es", "lat"=>1.58, "lon"=>10.35),
"/GR" => array("type"=>"country", "pop"=>11216000, "name_ru"=>"Греция", "name_el"=>"Ελλάδα", "name_en"=>"Greece", "lang"=>"el", "lat"=>38.30, "lon"=>23.74),
"/GT" => array("type"=>"country", "pop"=>11237196, "name_ru"=>"Гватемала", "name_en"=>"Guatemala", "lang"=>"es", "lat"=>15.5, "lon"=>-90.25),
"/GU" => array("type"=>"country", "pop"=>173459, "name_ru"=>"Гуам", "name_en"=>"Guam", "lang"=>"en", "lat"=>13.4, "lon"=>144.716666666667),
"/GW" => array("type"=>"country", "pop"=>1416027, "name_ru"=>"Гвинея-Бисау", "name_en"=>"Guinea-Bissau", "lang"=>"pt", "lat"=>11.75, "lon"=>-15.15),
"/GY" => array("type"=>"country", "pop"=>765283, "name_ru"=>"Гайана", "name_en"=>"Guyana", "lang"=>"en", "lat"=>5.73, "lon"=>-59.31),
"/HK" => array("type"=>"country", "pop"=>7055071, "name_cn"=>"香港", "name_ru"=>"Гонконг", "name_en"=>"Hong Kong", "lang"=>"cn", "lat"=>22.3, "lon"=>114.166666666667),
"/HN" => array("type"=>"country", "pop"=>6975204, "name_ru"=>"Гондурас", "name_en"=>"Honduras", "lang"=>"es", "lat"=>14.63, "lon"=>-86.81),
"/HR" => array("type"=>"country", "pop"=>4555000, "name_ru"=>"Хорватия", "name_hr"=>"Hrvatska", "name_en"=>"Croatia", "lang"=>"hr", "lat"=>45.25, "lon"=>15.46),
"/HT" => array("type"=>"country", "pop"=>8121622, "name_ru"=>"Гаити", "name_en"=>"Haiti", "lang"=>"fr", "lat"=>19.1, "lon"=>-72.33),
"/HU" => array("type"=>"country", "pop"=>10020000, "name_ru"=>"Венгрия", "name_hu"=>"Magyarország", "name_en"=>"Hungary", "lang"=>"hu", "lat"=>47.49, "lon"=>19.04),
"/ID" => array("type"=>"country", "pop"=>237512352, "name_ru"=>"Индонезия", "name_en"=>"Indonesia", "lang"=>"id", "lat"=>-6.17, "lon"=>106.82),
"/IE" => array("type"=>"country", "pop"=>4339000, "name_ga"=>"Éire", "name_ru"=>"Ирландия", "name_en"=>"Ireland", "lang"=>"ga", "lat"=>53.13, "lon"=>-8.11),
"/IL" => array("type"=>"country", "pop"=>7411000, "name_ru"=>"Израиль", "name_he"=>"מדינת ישראל", "name_en"=>"Israel", "lang"=>"he", "lat"=>31.78, "lon"=>35.21),
"/IN" => array("type"=>"country", "pop"=>1131191071, "name_hi"=>"भारत गणराज्य", "name_ru"=>"Индия", "name_en"=>"India", "lang"=>"hi", "lat"=>21.12, "lon"=>78.31),
"/IQ" => array("type"=>"country", "pop"=>26783383, "name_ru"=>"Ирак", "name_en"=>"Iraq", "lang"=>"ar", "lat"=>33.33, "lon"=>44.38),
"/IR" => array("type"=>"country", "pop"=>71208000, "name_ru"=>"Иран", "name_en"=>"Iran", "lang"=>"fa", "lat"=>32.49, "lon"=>54.29),
"/IS" => array("type"=>"country", "pop"=>319756, "name_ru"=>"Исландия", "name_is"=>"Ísland", "name_en"=>"Iceland", "lang"=>"is", "lat"=>64.83, "lon"=>-17.98),
"/IT" => array("type"=>"country", "pop"=>60000000, "name_ru"=>"Италия", "name_it"=>"Italia", "name_en"=>"Italy", "lang"=>"it", "lat"=>43.00, "lon"=>12.00),
"/JM" => array("type"=>"country", "pop"=>2651000, "name_ru"=>"Ямайка", "name_en"=>"Jamaica", "lang"=>"en", "lat"=>18.15, "lon"=>-77.3),
"/JO" => array("type"=>"country", "pop"=>5759732, "name_ru"=>"Иордания", "name_en"=>"Jordan", "lang"=>"ar", "lat"=>31.01, "lon"=>36.61),
"/JP" => array("type"=>"country", "pop"=>127078679, "name_ru"=>"Япония", "name_ja"=>"日本", "name_en"=>"Japan", "lang"=>"ja", "lat"=>35.15, "lon"=>136.06),
"/KE" => array("type"=>"country", "pop"=>36913721, "name_ru"=>"Кения", "name_en"=>"Kenya", "lang"=>"en", "lat"=>-0.4, "lon"=>37.85),
"/KG" => array("type"=>"country", "pop"=>5146281, "name_ru"=>"Киргизия", "name_en"=>"Kyrgyzstan", "lang"=>"ru", "lat"=>41.26, "lon"=>74.98),
"/KH" => array("type"=>"country", "pop"=>14071000, "name_ru"=>"Камбоджа", "name_en"=>"Cambodia", "lang"=>"km", "lat"=>12.73, "lon"=>105.66),
"/KI" => array("type"=>"country", "pop"=>92533, "name_ru"=>"Кирибати", "name_en"=>"Kiribati", "lang"=>"en", "lat"=>1.32, "lon"=>172.98),
"/KM" => array("type"=>"country", "pop"=>798000, "name_ru"=>"Коморские Острова", "name_en"=>"Comoros", "lang"=>"fr", "lat"=>-12.18, "lon"=>44.23),
"/KN" => array("type"=>"country", "pop"=>42696, "name_ru"=>"Сент-Китс и Невис", "name_en"=>"Saint Kitts and Nevis", "lang"=>"en", "lat"=>17.13, "lon"=>-62.61),
"/KP" => array("type"=>"country", "pop"=>23906000, "name_ru"=>"Северная Корея", "name_ko"=>"조선 민주주의 인민 공화국", "name_en"=>"North Korea", "lang"=>"ko", "lat"=>39.74, "lon"=>126.40),
"/KR" => array("type"=>"country", "pop"=>49024737, "name_ru"=>"Южная Корея", "name_ko"=>"대한민국", "name_en"=>"South Korea", "lang"=>"ko", "lat"=>36.00, "lon"=>128.00),
"/KW" => array("type"=>"country", "pop"=>3100000, "name_ru"=>"Кувейт", "name_en"=>"Kuwait", "lang"=>"ar", "lat"=>29.16, "lon"=>47.6),
"/KY" => array("type"=>"country", "pop"=>49035, "name_ru"=>"Каймановы Острова", "name_en"=>"Cayman Islands", "lang"=>"en", "lat"=>19.1166666666667, "lon"=>-81.3666666666667),
"/KZ" => array("type"=>"country", "pop"=>15902000, "name_ru"=>"Казахстан", "name_en"=>"Kazakhstan", "lang"=>"ru", "lat"=>48.95, "lon"=>66.23),
"/KZ/ALMT" => array("type"=>"city", "pop"=>1344900, "name_ru"=>"Алма-Ата", "name_kz"=>"Алматы", "name_en"=>"Almaty", "lang"=>"ru", "lat"=>43.25, "lon"=>76.9),
"/KZ/ASTN" => array("type"=>"city", "pop"=>653800, "name_ru"=>"Астана", "name_kz"=>"Астана", "name_en"=>"Astana", "lang"=>"ru", "lat"=>51.1833333333333, "lon"=>71.4),
"/KZ/BKNR" => array("type"=>"city", "pop"=>59452, "name_ru"=>"Байконур", "name_kz"=>"Байқоңыр", "name_en"=>"Baikonur", "lang"=>"ru", "lat"=>45.6166666666667, "lon"=>63.3166666666667),
"/KZ/KAO" => array("type"=>"region", "pop"=>1344200, "name_ru"=>"Карагандинская область", "name_kz"=>"Қарағанды облысы", "name_en"=>"Karagandy Province", "lang"=>"ru"),
"/KZ/KAO/BLHS" => array("type"=>"city", "pop"=>66220, "name_ru"=>"Балхаш", "name_kz"=>"Балқаш", "name_en"=>"Balqash", "lang"=>"ru", "lat"=>46.85, "lon"=>75),
"/KZ/KAO/KRGN" => array("type"=>"city", "pop"=>465634, "name_ru"=>"Караганда", "name_kz"=>"Қарағанды", "name_en"=>"Karaganda", "lang"=>"ru", "lat"=>49.8333333333333, "lon"=>73.1666666666667),
"/KZ/KAO/SHHT" => array("type"=>"city", "pop"=>35916, "name_ru"=>"Шахтинск", "name_kz"=>"Шахтинск", "name_en"=>"Shakhtinsk", "lang"=>"ru", "lat"=>49.71, "lon"=>72.5872222222222),
"/KZ/KAO/STPV" => array("type"=>"city", "pop"=>62007, "name_ru"=>"Сатпаев", "name_kz"=>"Сәтбаев", "name_en"=>"Satpayev", "lang"=>"ru", "lat"=>47.9, "lon"=>67.5333333333333),
"/KZ/KAO/TMRT" => array("type"=>"city", "pop"=>164969, "name_ru"=>"Темиртау", "name_kz"=>"Теміртау", "name_en"=>"Temirtau", "lang"=>"ru", "lat"=>50.0666666666667, "lon"=>72.9666666666667),
"/KZ/KAO/ZHZK" => array("type"=>"city", "pop"=>92229, "name_ru"=>"Жезказган", "name_kz"=>"Жезқазған", "name_en"=>"Jezkazgan", "lang"=>"ru", "lat"=>47.7833333333333, "lon"=>67.7),
"/KZ/KOO" => array("type"=>"region", "pop"=>891800, "name_ru"=>"Костанайская область", "name_kz"=>"Қостанай облысы", "name_en"=>"Kostanay Province", "lang"=>"ru"),
"/KZ/KOO/ARKL" => array("type"=>"city", "pop"=>25108, "name_ru"=>"Аркалык", "name_kz"=>"Арқалық", "name_en"=>"Arkalyk", "lang"=>"ru", "lat"=>50.2486111111111, "lon"=>66.9113888888889),
"/KZ/KOO/KSTN" => array("type"=>"city", "pop"=>212378, "name_ru"=>"Костанай", "name_kz"=>"Қостанай", "name_en"=>"Kostanay", "lang"=>"ru", "lat"=>53.2, "lon"=>63.6333333333333),
"/KZ/MAO" => array("type"=>"region", "pop"=>416500, "name_ru"=>"Мангистауская область", "name_kz"=>"Маңғыстау облысы", "name_en"=>"Mangystau Province", "lang"=>"ru"),
"/KZ/MAO/AKT" => array("type"=>"city", "pop"=>152227, "name_ru"=>"Актау", "name_kz"=>"Ақтау", "name_en"=>"Aktau", "lang"=>"ru", "lat"=>43.65, "lon"=>51.15),
"/KZ/MAO/ZHN" => array("type"=>"city", "pop"=>71109, "name_ru"=>"Жанаозен", "name_kz"=>"Жаңаөзен", "name_en"=>"Zhanaozen", "lang"=>"ru", "lat"=>43.3, "lon"=>52.8),
"/KZ/SKO" => array("type"=>"region", "pop"=>651100, "name_ru"=>"Северо-Казахстанская область", "name_kz"=>"Солтүстік Қазақстан облысы", "name_en"=>"North Kazakhstan Province", "lang"=>"ru"),
"/KZ/SKO/PTPL" => array("type"=>"city", "pop"=>193268, "name_ru"=>"Петропавловск", "name_kz"=>"Петропавлов", "name_en"=>"Petropavlovsk", "lang"=>"ru", "lat"=>54.8833333333333, "lon"=>69.2166666666667),
"/KZ/VKO" => array("type"=>"region", "pop"=>1417600, "name_ru"=>"Восточно-Казахстанская область", "name_kz"=>"Шығыс Қазақстан облысы", "name_en"=>"East Kazakhstan Province", "lang"=>"ru"),
"/KZ/VKO/SMPL" => array("type"=>"city", "pop"=>330338, "name_ru"=>"Семипалатинск", "name_kz"=>"Семей", "name_en"=>"Semey", "lang"=>"ru", "lat"=>50.4111111111111, "lon"=>80.2275),
"/KZ/VKO/USTK" => array("type"=>"city", "pop"=>287308, "name_ru"=>"Усть-Каменогорск", "name_kz"=>"Өскемен", "name_en"=>"Oskemen", "lang"=>"ru", "lat"=>49.95, "lon"=>82.6166666666667),
"/KZ/YKO" => array("type"=>"region", "pop"=>2356500, "name_ru"=>"Южно-Казахстанская область", "name_kz"=>"Оңтүстік Қазақстан облысы", "name_en"=>"South Kazakhstan Province", "lang"=>"ru"),
"/KZ/YKO/KNT" => array("type"=>"city", "pop"=>58313, "name_ru"=>"Кентау", "name_kz"=>"Кентау", "name_en"=>"Kentau", "lang"=>"ru", "lat"=>43.5, "lon"=>68.5),
"/KZ/YKO/SHMK" => array("type"=>"city", "pop"=>566996, "name_ru"=>"Шымкент", "name_kz"=>"Шымкент", "name_en"=>"Shymkent", "lang"=>"ru", "lat"=>42.3, "lon"=>69.6),
"/KZ/YKO/TRKS" => array("type"=>"city", "pop"=>119168, "name_ru"=>"Туркестан", "name_kz"=>"Түркістан", "name_en"=>"Turkestan", "lang"=>"ru", "lat"=>43.3333333333333, "lon"=>68.25),
"/LA" => array("type"=>"country", "pop"=>5924000, "name_ru"=>"Лаос", "name_en"=>"Laos", "lang"=>"lo", "lat"=>19.36, "lon"=>102.56),
"/LB" => array("type"=>"country", "pop"=>3925502, "name_ru"=>"Ливан", "name_en"=>"Lebanon", "lang"=>"ar", "lat"=>33.83, "lon"=>35.76),
"/LC" => array("type"=>"country", "pop"=>160145, "name_ru"=>"Сент-Люсия", "name_en"=>"Saint Lucia", "lang"=>"en", "lat"=>13.88, "lon"=>-61.00),
"/LI" => array("type"=>"country", "pop"=>35360, "name_de"=>"Fürstentum Liechtenstein", "name_ru"=>"Лихтенштейн", "name_en"=>"Liechtenstein", "lang"=>"de", "lat"=>47.14, "lon"=>9.55),
"/LK" => array("type"=>"country", "pop"=>20064776, "name_ru"=>"Шри-Ланка", "name_en"=>"Sri Lanka", "lang"=>"si", "lat"=>7.65, "lon"=>80.75),
"/LR" => array("type"=>"country", "pop"=>3334587, "name_ru"=>"Либерия", "name_en"=>"Liberia", "lang"=>"en", "lat"=>6.53, "lon"=>-9.75),
"/LS" => array("type"=>"country", "pop"=>1861959, "name_ru"=>"Лесото", "name_en"=>"Lesotho", "lang"=>"en", "lat"=>-29.55, "lon"=>28.25),
"/LT" => array("type"=>"country", "pop"=>3555179, "name_ru"=>"Литва", "name_lt"=>"Lietuva", "name_en"=>"Lithuania", "lang"=>"lt", "lat"=>55.58, "lon"=>24.11),
"/LU" => array("type"=>"country", "pop"=>439539, "name_lb"=>"Lëtzebuerg", "name_ru"=>"Люксембург", "name_en"=>"Luxembourg", "lang"=>"lb", "lat"=>49.75, "lon"=>6.16),
"/LV" => array("type"=>"country", "pop"=>2231503, "name_ru"=>"Латвия", "name_lv"=>"Latvija", "name_en"=>"Latvia", "lang"=>"lv", "lat"=>56.86, "lon"=>24.84),
"/LV/CESI" => array("type"=>"city", "pop"=>18600, "name_ru"=>"Цесис", "name_lv"=>"Cēsis", "name_en"=>"Cesis", "lang"=>"lv", "lat"=>57.32, "lon"=>25.2775),
"/LV/DGVP" => array("type"=>"city", "pop"=>110400, "name_ru"=>"Даугавпилс", "name_lv"=>"Daugavpils", "name_en"=>"Daugavpils", "lang"=>"lv", "lat"=>55.8833333333333, "lon"=>26.5333333333333),
"/LV/JKBP" => array("type"=>"city", "pop"=>29000, "name_ru"=>"Екабпилс", "name_lv"=>"Jēkabpils", "name_en"=>"Jekabpils", "lang"=>"lv", "lat"=>56.5, "lon"=>25.85),
"/LV/JLGV" => array("type"=>"city", "pop"=>65400, "name_ru"=>"Елгава", "name_lv"=>"Jelgava", "name_en"=>"Jelgava", "lang"=>"lv", "lat"=>56.65, "lon"=>23.7294444444444),
"/LV/JRML" => array("type"=>"city", "pop"=>55600, "name_ru"=>"Юрмала", "name_lv"=>"Jūrmala", "name_en"=>"Jurmala", "lang"=>"lv", "lat"=>56.9666666666667, "lon"=>23.75),
"/LV/LIEP" => array("type"=>"city", "pop"=>85710, "name_ru"=>"Лиепая", "name_lv"=>"Liepāja", "name_en"=>"Liepaja", "lang"=>"lv", "lat"=>56.5166666666667, "lon"=>21),
"/LV/OGRE" => array("type"=>"city", "pop"=>26148, "name_ru"=>"Огре", "name_lv"=>"Ogre", "name_en"=>"Ogre", "lang"=>"lv", "lat"=>56.8152777777778, "lon"=>24.5991666666667),
"/LV/PRLI" => array("type"=>"city", "pop"=>8605, "name_ru"=>"Прейли", "name_lv"=>"Preili", "name_en"=>"Preili", "lang"=>"lv", "lat"=>56.2905555555556, "lon"=>26.7263888888889),
"/LV/RIGA" => array("type"=>"city", "pop"=>709145, "name_ru"=>"Рига", "name_lv"=>"Rīga", "name_en"=>"Riga", "lang"=>"lv", "lat"=>56.95, "lon"=>24.1),
"/LV/RZKN" => array("type"=>"city", "pop"=>35467, "name_ru"=>"Резекне", "name_lv"=>"Rēzekne", "name_en"=>"Rezekne", "lang"=>"lv", "lat"=>56.5063888888889, "lon"=>27.3305555555556),
"/LV/SLSP" => array("type"=>"city", "pop"=>17863, "name_ru"=>"Саласпилс", "name_lv"=>"Salaspils", "name_en"=>"Salaspils", "lang"=>"lv", "lat"=>56.8594444444444, "lon"=>24.3597222222222),
"/LV/TKMS" => array("type"=>"city", "pop"=>20134, "name_ru"=>"Тукумс", "name_lv"=>"Tukums", "name_en"=>"Tukums", "lang"=>"lv", "lat"=>56.9666666666667, "lon"=>23.15),
"/LV/VLMR" => array("type"=>"city", "pop"=>27600, "name_ru"=>"Валмиера", "name_lv"=>"Valmiera", "name_en"=>"Valmiera", "lang"=>"lv", "lat"=>57.55, "lon"=>25.4),
"/LV/VNTS" => array("type"=>"city", "pop"=>44000, "name_ru"=>"Вентспилс", "name_lv"=>"Ventspils", "name_en"=>"Ventspils", "lang"=>"lv", "lat"=>57.4166666666667, "lon"=>21.5333333333333),
"/LY" => array("type"=>"country", "pop"=>6365563, "name_ru"=>"Ливия", "name_en"=>"Libya", "lang"=>"ar", "lat"=>26.05, "lon"=>18.2),
"/MA" => array("type"=>"country", "pop"=>33241259, "name_ru"=>"Марокко", "name_en"=>"Morocco", "lang"=>"ar", "lat"=>31.5, "lon"=>-7.25),
"/MC" => array("type"=>"country", "pop"=>32965, "name_ru"=>"Монако", "name_en"=>"Monaco", "lang"=>"fr", "lat"=>43.73, "lon"=>7.41),
"/MD" => array("type"=>"country", "pop"=>3567500, "name_ru"=>"Молдавия", "name_ro"=>"Moldova", "name_en"=>"Moldova", "lang"=>"ro", "lat"=>47.25, "lon"=>28.51),
"/ME" => array("type"=>"country", "pop"=>684736, "name_ru"=>"Черногория", "name_sr"=>"Црна Гора", "name_en"=>"Montenegro", "lang"=>"sr", "lat"=>42.75, "lon"=>19.4),
"/MG" => array("type"=>"country", "pop"=>20042552, "name_ru"=>"Мадагаскар", "name_en"=>"Madagascar", "lang"=>"mg", "lat"=>-18.75, "lon"=>46.83),
"/MH" => array("type"=>"country", "pop"=>56429, "name_ru"=>"Маршалловы Острова", "name_en"=>"Marshall Islands", "lang"=>"mh", "lat"=>7.06, "lon"=>171.26),
"/MK" => array("type"=>"country", "pop"=>2045262, "name_ru"=>"Македония", "name_mk"=>"Македонија", "name_en"=>"Macedonia", "lang"=>"mk", "lat"=>41.6, "lon"=>21.75),
"/ML" => array("type"=>"country", "pop"=>13518000, "name_ru"=>"Мали", "name_en"=>"Mali", "lang"=>"fr", "lat"=>17.00, "lon"=>-4.36),
"/MM" => array("type"=>"country", "pop"=>50519000, "name_ru"=>"Мьянма", "name_en"=>"Myanmar", "lang"=>"my", "lat"=>21.1, "lon"=>95.45),
"/MN" => array("type"=>"country", "pop"=>2951786, "name_ru"=>"Монголия", "name_en"=>"Mongolia", "lang"=>"mn", "lat"=>47.6, "lon"=>103.23),
"/MO" => array("type"=>"country", "pop"=>546200, "name_cn"=>"澳門", "name_ru"=>"Аомынь", "name_pt"=>"Macau", "name_en"=>"Macau", "lang"=>"cn", "lat"=>22.1666666666667, "lon"=>113.55),
"/MR" => array("type"=>"country", "pop"=>3069000, "name_ru"=>"Мавритания", "name_en"=>"Mauritania", "lang"=>"ar", "lat"=>19.56, "lon"=>-10.65),
"/MT" => array("type"=>"country", "pop"=>402000, "name_ru"=>"Мальта", "name_en"=>"Malta", "lang"=>"mt", "lat"=>35.9, "lon"=>14.4),
"/MU" => array("type"=>"country", "pop"=>1240827, "name_ru"=>"Маврикий", "name_en"=>"Mauritius", "lang"=>"en", "lat"=>-20.28, "lon"=>57.55),
"/MV" => array("type"=>"country", "pop"=>298968, "name_ru"=>"Мальдивы", "name_en"=>"Maldives", "lang"=>"dv", "lat"=>3.78, "lon"=>72.96),
"/MW" => array("type"=>"country", "pop"=>13931000, "name_ru"=>"Малави", "name_en"=>"Malawi", "lang"=>"en", "lat"=>-13.2, "lon"=>34.4),
"/MX" => array("type"=>"country", "pop"=>107449525, "name_ru"=>"Мексика", "name_es"=>"Mexico", "name_en"=>"Mexico", "lang"=>"es", "lat"=>23.31, "lon"=>-102.36),
"/MY" => array("type"=>"country", "pop"=>27496000, "name_ru"=>"Малайзия", "name_en"=>"Malaysia", "lang"=>"ms", "lat"=>2.31, "lon"=>111.55),
"/MZ" => array("type"=>"country", "pop"=>21284700, "name_ru"=>"Мозамбик", "name_en"=>"Mozambique", "lang"=>"pt", "lat"=>-17.6, "lon"=>35.18),
"/NA" => array("type"=>"country", "pop"=>1820916, "name_ru"=>"Намибия", "name_en"=>"Namibia", "lang"=>"af", "lat"=>-22.56, "lon"=>17.15),
"/NC" => array("type"=>"country", "pop"=>249000, "name_ru"=>"Новая Каледония", "name_fr"=>"Nouvelle-Calédonie", "name_en"=>"New Caledonia", "lang"=>"fr", "lat"=>-22.2758333333333, "lon"=>166.458055555556),
"/NE" => array("type"=>"country", "pop"=>13272679, "name_ru"=>"Нигер", "name_en"=>"Niger", "lang"=>"fr", "lat"=>16.18, "lon"=>9.61),
"/NG" => array("type"=>"country", "pop"=>148000000, "name_ru"=>"Нигерия", "name_en"=>"Niger", "lang"=>"en", "lat"=>9.55, "lon"=>7.81),
"/NI" => array("type"=>"country", "pop"=>5697313, "name_ru"=>"Никарагуа", "name_en"=>"Nicaragua", "lang"=>"es", "lat"=>12.90, "lon"=>-84.92),
"/NL" => array("type"=>"country", "pop"=>16357373, "name_ru"=>"Нидерланды", "name_nl"=>"Nederland", "name_en"=>"Netherlands", "lang"=>"nl", "lat"=>51.91, "lon"=>5.56),
"/NO" => array("type"=>"country", "pop"=>4799252, "name_no"=>"Norge", "name_ru"=>"Норвегия", "name_en"=>"Norway", "lang"=>"no", "lat"=>62.76, "lon"=>9.45),
"/NP" => array("type"=>"country", "pop"=>27676547, "name_ru"=>"Непал", "name_en"=>"Nepal", "lang"=>"ne", "lat"=>28.25, "lon"=>83.91),
"/NR" => array("type"=>"country", "pop"=>13770, "name_ru"=>"Науру", "name_en"=>"Nauru", "lang"=>"en", "lat"=>-0.52, "lon"=>166.93),
"/NZ" => array("type"=>"country", "pop"=>4268000, "name_ru"=>"Новая Зеландия", "name_en"=>"New Zealand", "lang"=>"en", "lat"=>-41.31, "lon"=>172.76),
"/OM" => array("type"=>"country", "pop"=>3204897, "name_ru"=>"Оман", "name_en"=>"Oman", "lang"=>"ar", "lat"=>20.23, "lon"=>55.96),
"/PA" => array("type"=>"country", "pop"=>3360474, "name_ru"=>"Панама", "name_en"=>"Panama", "lang"=>"es", "lat"=>8.61, "lon"=>-80.36),
"/PE" => array("type"=>"country", "pop"=>29546963, "name_ru"=>"Перу", "name_en"=>"Peru", "lang"=>"es", "lat"=>-8.23, "lon"=>-76.01),
"/PF" => array("type"=>"country", "pop"=>264000, "name_ru"=>"Французская Полинезия", "name_fr"=>"Polynésie française", "name_en"=>"French Polynesia", "lang"=>"fr", "lat"=>-17.5666666666667, "lon"=>-149.6),
"/PG" => array("type"=>"country", "pop"=>6300000, "name_ru"=>"Папуа — Новая Гвинея", "name_en"=>"Papua New Guinea", "lang"=>"en", "lat"=>-6.36, "lon"=>146.1),
"/PH" => array("type"=>"country", "pop"=>87857473, "name_ru"=>"Филиппины", "name_en"=>"Philippines", "lang"=>"en", "lat"=>11.33, "lon"=>123.01),
"/PK" => array("type"=>"country", "pop"=>180808000, "name_ru"=>"Пакистан", "name_en"=>"Pakistan", "lang"=>"ur", "lat"=>29.4, "lon"=>69.18),
"/PL" => array("type"=>"country", "pop"=>38138000, "name_ru"=>"Польша", "name_pl"=>"Polska", "name_en"=>"Poland", "lang"=>"pl", "lat"=>52.14, "lon"=>19.37),
"/PR" => array("type"=>"country", "pop"=>3994259, "name_ru"=>"Пуэрто-Рико", "name_es"=>"Puerto Rico", "name_en"=>"Puerto Rico", "lang"=>"es", "lat"=>18.45, "lon"=>-66.0666666666667),
"/PS" => array("type"=>"country", "pop"=>4016416, "name_ru"=>"Палестина", "name_ar"=>"فلسطين", "name_en"=>"Palestine", "lang"=>"ar", "lat"=>31.8969444444444, "lon"=>35.2016666666667),
"/PT" => array("type"=>"country", "pop"=>10642836, "name_ru"=>"Португалия", "name_pt"=>"Portugal", "name_en"=>"Portugal", "lang"=>"pt", "lat"=>39.38, "lon"=>-8.2),
"/PW" => array("type"=>"country", "pop"=>20842, "name_ru"=>"Палау", "name_en"=>"Palau", "lang"=>"en", "lat"=>7.46, "lon"=>134.55),
"/PY" => array("type"=>"country", "pop"=>6347884, "name_ru"=>"Парагвай", "name_en"=>"Paraguay", "lang"=>"es", "lat"=>-23.26, "lon"=>-58.08),
"/QA" => array("type"=>"country", "pop"=>841000, "name_ru"=>"Катар", "name_en"=>"Qatar", "lang"=>"ar", "lat"=>25.26, "lon"=>51.21),
"/RO" => array("type"=>"country", "pop"=>22276056, "name_ru"=>"Румыния", "name_ro"=>"România", "name_en"=>"Romania", "lang"=>"ro", "lat"=>45.86, "lon"=>25.3),
"/RS" => array("type"=>"country", "pop"=>7365507, "name_ru"=>"Сербия", "name_sr"=>"Србија", "name_en"=>"Serbia", "lang"=>"sr", "lat"=>43.81, "lon"=>21.13),
"/RU" => array("type"=>"country", "pop"=>140000000, "name_ru"=>"Россия", "name_en"=>"Russia", "lang"=>"ru", "lat"=>65.00, "lon"=>55.00),
"/RU/CFO" => array("type"=>"federal region", "pop"=>37121812, "name_ru"=>"Центральный федеральный округ", "name_en"=>"Central Federal Region", "lang"=>"ru"),
"/RU/CFO/31" => array("type"=>"region", "pop"=>1519100, "name_ru"=>"Белгородская область", "name_en"=>"Belgorod Region", "lang"=>"ru"),
"/RU/CFO/31/BLG" => array("type"=>"city", "pop"=>353000, "name_ru"=>"Белгород", "name_en"=>"Belgorod", "lang"=>"ru", "lat"=>50.6, "lon"=>36.6),
"/RU/CFO/31/GBK" => array("type"=>"city", "pop"=>86400, "name_ru"=>"Губкин", "name_en"=>"Gubkin", "lang"=>"ru", "lat"=>51.2833333333333, "lon"=>57.55),
"/RU/CFO/31/SHBK" => array("type"=>"city", "pop"=>45600, "name_ru"=>"Шебекино", "name_en"=>"Shebekino", "lang"=>"ru", "lat"=>50.4166666666667, "lon"=>36.9),
"/RU/CFO/31/STOS" => array("type"=>"city", "pop"=>220200, "name_ru"=>"Старый Оскол", "name_en"=>"Staryy Oskol", "lang"=>"ru", "lat"=>51.3, "lon"=>37.8333333333333),
"/RU/CFO/32" => array("type"=>"region", "pop"=>1308500, "name_ru"=>"Брянская область", "name_en"=>"Brjansk Region", "lang"=>"ru"),
"/RU/CFO/32/BRJ" => array("type"=>"city", "pop"=>413900, "name_ru"=>"Брянск", "name_en"=>"Brjansk", "lang"=>"ru", "lat"=>53.25, "lon"=>34.3666666666667),
"/RU/CFO/32/DTK" => array("type"=>"city", "pop"=>32900, "name_ru"=>"Дятьково", "name_en"=>"Djatkovo", "lang"=>"ru", "lat"=>53.6, "lon"=>34.3333333333333),
"/RU/CFO/32/KLN" => array("type"=>"city", "pop"=>64600, "name_ru"=>"Клинцы", "name_en"=>"Klintsy", "lang"=>"ru", "lat"=>52.75, "lon"=>32.2333333333333),
"/RU/CFO/32/NVZ" => array("type"=>"city", "pop"=>42200, "name_ru"=>"Новозыбков", "name_en"=>"Novozybkov", "lang"=>"ru", "lat"=>52.5333333333333, "lon"=>31.9333333333333),
"/RU/CFO/33" => array("type"=>"region", "pop"=>1449500, "name_ru"=>"Владимирская область", "name_en"=>"Vladimir Region", "lang"=>"ru"),
"/RU/CFO/33/ALXN" => array("type"=>"city", "pop"=>63000, "name_ru"=>"Александров", "name_en"=>"Aleksandrov", "lang"=>"ru", "lat"=>56.4, "lon"=>38.7166666666667),
"/RU/CFO/33/GSHR" => array("type"=>"city", "pop"=>61900, "name_ru"=>"Гусь-Хрустальный", "name_en"=>"Gus-Khrustalnyy", "lang"=>"ru", "lat"=>55.6166666666667, "lon"=>40.65),
"/RU/CFO/33/KVR" => array("type"=>"city", "pop"=>150400, "name_ru"=>"Ковров", "name_en"=>"Kovrov", "lang"=>"ru", "lat"=>56.35, "lon"=>41.3166666666667),
"/RU/CFO/33/MRM" => array("type"=>"city", "pop"=>119400, "name_ru"=>"Муром", "name_en"=>"Murom", "lang"=>"ru", "lat"=>55.5666666666667, "lon"=>42.0333333333333),
"/RU/CFO/33/PKR" => array("type"=>"city", "pop"=>15700, "name_ru"=>"Покров", "name_en"=>"Pokrov", "lang"=>"ru", "lat"=>55.9166666666667, "lon"=>39.1833333333333),
"/RU/CFO/33/PTSH" => array("type"=>"city", "pop"=>14900, "name_ru"=>"Петушки", "name_en"=>"Petushki", "lang"=>"ru", "lat"=>55.9333333333333, "lon"=>39.4666666666667),
"/RU/CFO/33/SZD" => array("type"=>"city", "pop"=>11000, "name_ru"=>"Суздаль", "name_en"=>"Suzdal", "lang"=>"ru", "lat"=>56.4333333333333, "lon"=>40.4333333333333),
"/RU/CFO/33/VLD" => array("type"=>"city", "pop"=>339500, "name_ru"=>"Владимир", "name_en"=>"Vladimir", "lang"=>"ru", "lat"=>56.1333333333333, "lon"=>40.4166666666667),
"/RU/CFO/36" => array("type"=>"region", "pop"=>2280400, "name_ru"=>"Воронежская область", "name_en"=>"Voronezh Region", "lang"=>"ru"),
"/RU/CFO/36/BRG" => array("type"=>"city", "pop"=>65400, "name_ru"=>"Борисоглебск", "name_en"=>"Borisoglebsk", "lang"=>"ru", "lat"=>51.3666666666667, "lon"=>42.0833333333333),
"/RU/CFO/36/LSK" => array("type"=>"city", "pop"=>54800, "name_ru"=>"Лиски", "name_en"=>"Liski", "lang"=>"ru", "lat"=>50.9833333333333, "lon"=>39.5166666666667),
"/RU/CFO/36/RSH" => array("type"=>"city", "pop"=>62100, "name_ru"=>"Россошь", "name_en"=>"Rossosh'", "lang"=>"ru", "lat"=>50.2, "lon"=>39.5833333333333),
"/RU/CFO/36/VRN" => array("type"=>"city", "pop"=>839900, "name_ru"=>"Воронеж", "name_en"=>"Voronezh", "lang"=>"ru", "lat"=>51.6666666666667, "lon"=>39.1833333333333),
"/RU/CFO/37" => array("type"=>"region", "pop"=>1076900, "name_ru"=>"Ивановская область", "name_en"=>"Ivanovskaya Region", "lang"=>"ru"),
"/RU/CFO/37/IVN" => array("type"=>"city", "pop"=>406500, "name_ru"=>"Иваново", "name_en"=>"Ivanovo", "lang"=>"ru", "lat"=>57, "lon"=>40.9833333333333),
"/RU/CFO/37/KNSH" => array("type"=>"city", "pop"=>92400, "name_ru"=>"Кинешма", "name_en"=>"Kineshma", "lang"=>"ru", "lat"=>57.45, "lon"=>42.15),
"/RU/CFO/37/SHU" => array("type"=>"city", "pop"=>58900, "name_ru"=>"Шуя", "name_en"=>"Shuja", "lang"=>"ru", "lat"=>56.85, "lon"=>41.3666666666667),
"/RU/CFO/40" => array("type"=>"region", "pop"=>1005600, "name_ru"=>"Калужская область", "name_en"=>"Kaluzhskaya Region", "lang"=>"ru"),
"/RU/CFO/40/KLG" => array("type"=>"city", "pop"=>326900, "name_ru"=>"Калуга", "name_en"=>"Kaluga", "lang"=>"ru", "lat"=>54.5166666666667, "lon"=>36.2666666666667),
"/RU/CFO/40/KRV" => array("type"=>"city", "pop"=>38700, "name_ru"=>"Киров", "name_en"=>"Kirov", "lang"=>"ru", "lat"=>54.0833333333333, "lon"=>34.3),
"/RU/CFO/40/LDN" => array("type"=>"city", "pop"=>41500, "name_ru"=>"Людиново", "name_en"=>"Ljudinovo", "lang"=>"ru", "lat"=>53.8666666666667, "lon"=>34.45),
"/RU/CFO/40/OBN" => array("type"=>"city", "pop"=>105500, "name_ru"=>"Обнинск", "name_en"=>"Obninsk", "lang"=>"ru", "lat"=>55.1, "lon"=>36.6166666666667),
"/RU/CFO/44" => array("type"=>"region", "pop"=>697000, "name_ru"=>"Костромская область", "name_en"=>"Kostromskaya Region", "lang"=>"ru"),
"/RU/CFO/44/KST" => array("type"=>"city", "pop"=>271700, "name_ru"=>"Кострома", "name_en"=>"Kostroma", "lang"=>"ru", "lat"=>57.7666666666667, "lon"=>40.9333333333333),
"/RU/CFO/46" => array("type"=>"region", "pop"=>1162500, "name_ru"=>"Курская область", "name_en"=>"Kurskaya Region", "lang"=>"ru"),
"/RU/CFO/46/KRC" => array("type"=>"city", "pop"=>46900, "name_ru"=>"Курчатов", "name_en"=>"Kurchatov", "lang"=>"ru", "lat"=>51.6666666666667, "lon"=>35.65),
"/RU/CFO/46/KRS" => array("type"=>"city", "pop"=>408100, "name_ru"=>"Курск", "name_en"=>"Kursk", "lang"=>"ru", "lat"=>51.7166666666667, "lon"=>36.1833333333333),
"/RU/CFO/46/ZHG" => array("type"=>"city", "pop"=>97400, "name_ru"=>"Железногорск", "name_en"=>"Zheleznogorsk", "lang"=>"ru", "lat"=>52.3333333333333, "lon"=>35.3666666666667),
"/RU/CFO/48" => array("type"=>"region", "pop"=>1168800, "name_ru"=>"Липецкая область", "name_en"=>"Lipetskaya Region", "lang"=>"ru"),
"/RU/CFO/48/DNK" => array("type"=>"city", "pop"=>22300, "name_ru"=>"Данков", "name_en"=>"Dankov", "lang"=>"ru", "lat"=>53.25, "lon"=>39.15),
"/RU/CFO/48/ELC" => array("type"=>"city", "pop"=>112200, "name_ru"=>"Елец", "name_en"=>"Elets", "lang"=>"ru", "lat"=>52.6166666666667, "lon"=>38.5166666666667),
"/RU/CFO/48/GRZ" => array("type"=>"city", "pop"=>46500, "name_ru"=>"Грязи", "name_en"=>"Grjazi", "lang"=>"ru", "lat"=>52.5, "lon"=>39.9333333333333),
"/RU/CFO/48/LBD" => array("type"=>"city", "pop"=>20900, "name_ru"=>"Лебедянь", "name_en"=>"Lebedjan'", "lang"=>"ru", "lat"=>53.0166666666667, "lon"=>39.15),
"/RU/CFO/48/LPC" => array("type"=>"city", "pop"=>502500, "name_ru"=>"Липецк", "name_en"=>"Lipetsk", "lang"=>"ru", "lat"=>52.6166666666667, "lon"=>39.6),
"/RU/CFO/57" => array("type"=>"region", "pop"=>821900, "name_ru"=>"Орловская область", "name_en"=>"Orlovskaya Region", "lang"=>"ru"),
"/RU/CFO/57/LVN" => array("type"=>"city", "pop"=>51100, "name_ru"=>"Ливны", "name_en"=>"Livny", "lang"=>"ru", "lat"=>52.4166666666667, "lon"=>37.6),
"/RU/CFO/57/MCN" => array("type"=>"city", "pop"=>45700, "name_ru"=>"Мценск", "name_en"=>"Mtsensk", "lang"=>"ru", "lat"=>53.2833333333333, "lon"=>36.5666666666667),
"/RU/CFO/57/ORL" => array("type"=>"city", "pop"=>320800, "name_ru"=>"Орёл", "name_en"=>"Orel", "lang"=>"ru", "lat"=>52.9666666666667, "lon"=>36.0833333333333),
"/RU/CFO/62" => array("type"=>"region", "pop"=>1164500, "name_ru"=>"Рязанская область", "name_en"=>"Rjazanskaya Region", "lang"=>"ru"),
"/RU/CFO/62/KSM" => array("type"=>"city", "pop"=>33900, "name_ru"=>"Касимов", "name_en"=>"Kasimov", "lang"=>"ru", "lat"=>54.9333333333333, "lon"=>41.4),
"/RU/CFO/62/RZN" => array("type"=>"city", "pop"=>510800, "name_ru"=>"Рязань", "name_en"=>"Rjazan'", "lang"=>"ru", "lat"=>54.6166666666667, "lon"=>39.7166666666667),
"/RU/CFO/62/SKP" => array("type"=>"city", "pop"=>31300, "name_ru"=>"Скопин", "name_en"=>"Skopin", "lang"=>"ru", "lat"=>53.8166666666667, "lon"=>39.55),
"/RU/CFO/67" => array("type"=>"region", "pop"=>983200, "name_ru"=>"Смоленская область", "name_en"=>"Smolenskaya Region", "lang"=>"ru"),
"/RU/CFO/67/RSL" => array("type"=>"city", "pop"=>54700, "name_ru"=>"Рославль", "name_en"=>"Roslavl'", "lang"=>"ru", "lat"=>53.95, "lon"=>32.85),
"/RU/CFO/67/SML" => array("type"=>"city", "pop"=>316500, "name_ru"=>"Смоленск", "name_en"=>"Smolensk", "lang"=>"ru", "lat"=>54.7833333333333, "lon"=>32.05),
"/RU/CFO/67/VZM" => array("type"=>"city", "pop"=>55100, "name_ru"=>"Вязьма", "name_en"=>"Vjaz'ma", "lang"=>"ru", "lat"=>55.2, "lon"=>34.3),
"/RU/CFO/68" => array("type"=>"region", "pop"=>1106000, "name_ru"=>"Тамбовская область", "name_en"=>"Tambovskaya Region", "lang"=>"ru"),
"/RU/CFO/68/MCH" => array("type"=>"city", "pop"=>90200, "name_ru"=>"Мичуринск", "name_en"=>"Michurinsk", "lang"=>"ru", "lat"=>52.9, "lon"=>40.4833333333333),
"/RU/CFO/68/TMB" => array("type"=>"city", "pop"=>279800, "name_ru"=>"Тамбов", "name_en"=>"Tambov", "lang"=>"ru", "lat"=>52.7166666666667, "lon"=>41.4333333333333),
"/RU/CFO/69" => array("type"=>"region", "pop"=>1379500, "name_ru"=>"Тверская область", "name_en"=>"Tverskaya Region", "lang"=>"ru"),
"/RU/CFO/69/KMR" => array("type"=>"city", "pop"=>49400, "name_ru"=>"Кимры", "name_en"=>"Kimry", "lang"=>"ru", "lat"=>56.8666666666667, "lon"=>37.35),
"/RU/CFO/69/RZH" => array("type"=>"city", "pop"=>60700, "name_ru"=>"Ржев", "name_en"=>"Rzhev", "lang"=>"ru", "lat"=>56.25, "lon"=>34.3333333333333),
"/RU/CFO/69/TRZ" => array("type"=>"city", "pop"=>47600, "name_ru"=>"Торжок", "name_en"=>"Torzhok", "lang"=>"ru", "lat"=>57.0333333333333, "lon"=>34.9666666666667),
"/RU/CFO/69/TVR" => array("type"=>"city", "pop"=>407300, "name_ru"=>"Тверь", "name_en"=>"Tver'", "lang"=>"ru", "lat"=>56.8666666666667, "lon"=>35.9),
"/RU/CFO/69/VVL" => array("type"=>"city", "pop"=>52300, "name_ru"=>"Вышний Волочёк", "name_en"=>"Vyshniy Volochjok", "lang"=>"ru", "lat"=>57.5833333333333, "lon"=>34.5666666666667),
"/RU/CFO/71" => array("type"=>"region", "pop"=>1566300, "name_ru"=>"Тульская область", "name_en"=>"Tul'skaya Region", "lang"=>"ru"),
"/RU/CFO/71/ALX" => array("type"=>"city", "pop"=>65400, "name_ru"=>"Алексин", "name_en"=>"Aleksin", "lang"=>"ru", "lat"=>54.5, "lon"=>37.0666666666667),
"/RU/CFO/71/NVM" => array("type"=>"city", "pop"=>124100, "name_ru"=>"Новомосковск", "name_en"=>"Novomoskovsk", "lang"=>"ru", "lat"=>54.0333333333333, "lon"=>38.2666666666667),
"/RU/CFO/71/TUL" => array("type"=>"city", "pop"=>500000, "name_ru"=>"Тула", "name_en"=>"Tula", "lang"=>"ru", "lat"=>54.2, "lon"=>37.6166666666667),
"/RU/CFO/76" => array("type"=>"region", "pop"=>1315000, "name_ru"=>"Ярославская область", "name_en"=>"Yaroslavskaya Region", "lang"=>"ru"),
"/RU/CFO/76/RBN" => array("type"=>"city", "pop"=>211000, "name_ru"=>"Рыбинск", "name_en"=>"Rybinsk", "lang"=>"ru", "lat"=>58.05, "lon"=>38.8333333333333),
"/RU/CFO/76/YAR" => array("type"=>"city", "pop"=>605200, "name_ru"=>"Ярославль", "name_en"=>"Yaroslavl'", "lang"=>"ru", "lat"=>57.6333333333333, "lon"=>39.8666666666667),
"/RU/CFO/MO" => array("type"=>"region", "name_ru"=>"Москва и Московская область", "name_en"=>"Moscow and Moscow Region", "lang"=>"ru"),
"/RU/CFO/MO/50" => array("type"=>"region", "pop"=>6672800, "name_ru"=>"Московская область", "name_en"=>"Moscow Region", "lang"=>"ru"),
"/RU/CFO/MO/50/APRL" => array("type"=>"city", "pop"=>18300, "name_ru"=>"Апрелевка", "name_en"=>"Aprelevka", "lang"=>"ru", "lat"=>55.55, "lon"=>37.0666666666667),
"/RU/CFO/MO/50/BLSH" => array("type"=>"city", "pop"=>188700, "name_ru"=>"Балашиха", "name_en"=>"Balashiha", "lang"=>"ru", "lat"=>55.8, "lon"=>37.95),
"/RU/CFO/MO/50/BRNC" => array("type"=>"city", "pop"=>18600, "name_ru"=>"Бронницы", "name_en"=>"Bronnitsy", "lang"=>"ru", "lat"=>55.4333333333333, "lon"=>38.2666666666667),
"/RU/CFO/MO/50/CHG" => array("type"=>"city", "pop"=>20500, "name_ru"=>"Черноголовка", "name_en"=>"Chernogolovka", "lang"=>"ru", "lat"=>56, "lon"=>38.3666666666667),
"/RU/CFO/MO/50/CHHV" => array("type"=>"city", "pop"=>73600, "name_ru"=>"Чехов", "name_en"=>"Chehov", "lang"=>"ru", "lat"=>55.15, "lon"=>37.4833333333333),
"/RU/CFO/MO/50/DBN" => array("type"=>"city", "pop"=>62200, "name_ru"=>"Дубна", "name_en"=>"Dubna", "lang"=>"ru", "lat"=>56.75, "lon"=>37.15),
"/RU/CFO/MO/50/DDV" => array("type"=>"city", "pop"=>27800, "name_ru"=>"Дедовск", "name_en"=>"Dedovsk", "lang"=>"ru", "lat"=>55.8666666666667, "lon"=>37.1333333333333),
"/RU/CFO/MO/50/DLGP" => array("type"=>"city", "pop"=>80500, "name_ru"=>"Долгопрудный", "name_en"=>"Dolgoprudnyy", "lang"=>"ru", "lat"=>55.9333333333333, "lon"=>37.5166666666667),
"/RU/CFO/MO/50/DMDV" => array("type"=>"city", "pop"=>84900, "name_ru"=>"Домодедово", "name_en"=>"Domodedovo", "lang"=>"ru", "lat"=>55.4333333333333, "lon"=>37.75),
"/RU/CFO/MO/50/DMTR" => array("type"=>"city", "pop"=>62300, "name_ru"=>"Дмитров", "name_en"=>"Dmitrov", "lang"=>"ru", "lat"=>56.35, "lon"=>37.5333333333333),
"/RU/CFO/MO/50/DRZN" => array("type"=>"city", "pop"=>11600, "name_ru"=>"Дрезна", "name_en"=>"Drezna", "lang"=>"ru", "lat"=>55.75, "lon"=>38.8333333333333),
"/RU/CFO/MO/50/DZRZ" => array("type"=>"city", "pop"=>44300, "name_ru"=>"Дзержинский", "name_en"=>"Dzerzhinskiy", "lang"=>"ru", "lat"=>55.6333333333333, "lon"=>37.85),
"/RU/CFO/MO/50/EGRV" => array("type"=>"city", "pop"=>66800, "name_ru"=>"Егорьевск", "name_en"=>"Egorjevsk", "lang"=>"ru", "lat"=>55.3833333333333, "lon"=>39.0333333333333),
"/RU/CFO/MO/50/ELGR" => array("type"=>"city", "pop"=>20800, "name_ru"=>"Электрогорск", "name_en"=>"Elektrogorsk", "lang"=>"ru", "lat"=>55.8833333333333, "lon"=>38.7833333333333),
"/RU/CFO/MO/50/ELST" => array("type"=>"city", "pop"=>145800, "name_ru"=>"Электросталь", "name_en"=>"Elektrostal", "lang"=>"ru", "lat"=>55.8, "lon"=>38.45),
"/RU/CFO/MO/50/ELUG" => array("type"=>"city", "pop"=>20400, "name_ru"=>"Электроугли", "name_en"=>"Elektrougli", "lang"=>"ru", "lat"=>55.7166666666667, "lon"=>38.2166666666667),
"/RU/CFO/MO/50/FRZN" => array("type"=>"city", "pop"=>52600, "name_ru"=>"Фрязино", "name_en"=>"Frjazino", "lang"=>"ru", "lat"=>55.95, "lon"=>38.05),
"/RU/CFO/MO/50/GLCN" => array("type"=>"city", "pop"=>17000, "name_ru"=>"Голицыно", "name_en"=>"Golitsyno", "lang"=>"ru", "lat"=>55.6, "lon"=>36.9833333333333),
"/RU/CFO/MO/50/HMK" => array("type"=>"city", "pop"=>183600, "name_ru"=>"Химки", "name_en"=>"Himki", "lang"=>"ru", "lat"=>55.9, "lon"=>37.45),
"/RU/CFO/MO/50/HTKV" => array("type"=>"city", "pop"=>20200, "name_ru"=>"Хотьково", "name_en"=>"Hot'kovo", "lang"=>"ru", "lat"=>56.25, "lon"=>37.9833333333333),
"/RU/CFO/MO/50/ISTR" => array("type"=>"city", "pop"=>32300, "name_ru"=>"Истра", "name_en"=>"Istra", "lang"=>"ru", "lat"=>55.9166666666667, "lon"=>36.8666666666667),
"/RU/CFO/MO/50/IVNT" => array("type"=>"city", "pop"=>55700, "name_ru"=>"Ивантеевка", "name_en"=>"Ivanteevka", "lang"=>"ru", "lat"=>55.9833333333333, "lon"=>37.9333333333333),
"/RU/CFO/MO/50/JBLN" => array("type"=>"city", "pop"=>32300, "name_ru"=>"Юбилейный", "name_en"=>"Jubilejnyy", "lang"=>"ru", "lat"=>55.9333333333333, "lon"=>37.85),
"/RU/CFO/MO/50/JHRM" => array("type"=>"city", "pop"=>13100, "name_ru"=>"Яхрома", "name_en"=>"Jahroma", "lang"=>"ru", "lat"=>56.2833333333333, "lon"=>37.4833333333333),
"/RU/CFO/MO/50/KBNK" => array("type"=>"city", "pop"=>26300, "name_ru"=>"Кубинка", "name_en"=>"Kubinka", "lang"=>"ru", "lat"=>55.5833333333333, "lon"=>36.7),
"/RU/CFO/MO/50/KLIN" => array("type"=>"city", "pop"=>81500, "name_ru"=>"Клин", "name_en"=>"Klin", "lang"=>"ru", "lat"=>56.3333333333333, "lon"=>36.7),
"/RU/CFO/MO/50/KLMN" => array("type"=>"city", "pop"=>148000, "name_ru"=>"Коломна", "name_en"=>"Kolomna", "lang"=>"ru", "lat"=>55.0833333333333, "lon"=>38.7833333333333),
"/RU/CFO/MO/50/KLMV" => array("type"=>"city", "pop"=>55700, "name_ru"=>"Климовск", "name_en"=>"Klimovsck", "lang"=>"ru", "lat"=>55.3666666666667, "lon"=>37.5333333333333),
"/RU/CFO/MO/50/KRAR" => array("type"=>"city", "pop"=>26100, "name_ru"=>"Красноармейск", "name_en"=>"Krasnoarmejsk", "lang"=>"ru", "lat"=>56.1, "lon"=>38.1333333333333),
"/RU/CFO/MO/50/KRGR" => array("type"=>"city", "pop"=>99500, "name_ru"=>"Красногорск", "name_en"=>"Krasnogorsk", "lang"=>"ru", "lat"=>55.8166666666667, "lon"=>37.3333333333333),
"/RU/CFO/MO/50/KRLV" => array("type"=>"city", "pop"=>174600, "name_ru"=>"Королёв", "name_en"=>"Koroljov", "lang"=>"ru", "lat"=>55.9166666666667, "lon"=>37.8166666666667),
"/RU/CFO/MO/50/KRVS" => array("type"=>"city", "pop"=>18900, "name_ru"=>"Куровское", "name_en"=>"Kurovskoe", "lang"=>"ru", "lat"=>55.5833333333333, "lon"=>38.9166666666667),
"/RU/CFO/MO/50/KRZN" => array("type"=>"city", "pop"=>31500, "name_ru"=>"Краснознаменск", "name_en"=>"Krasnoznamensk", "lang"=>"ru", "lat"=>55.6, "lon"=>37.0333333333333),
"/RU/CFO/MO/50/KRZV" => array("type"=>"city", "pop"=>13700, "name_ru"=>"Краснозаводск", "name_en"=>"Krasnozavodsk", "lang"=>"ru", "lat"=>56.45, "lon"=>38.2166666666667),
"/RU/CFO/MO/50/KSHR" => array("type"=>"city", "pop"=>39500, "name_ru"=>"Кашира", "name_en"=>"Kashira", "lang"=>"ru", "lat"=>54.8333333333333, "lon"=>38.15),
"/RU/CFO/MO/50/KTLN" => array("type"=>"city", "pop"=>19800, "name_ru"=>"Котельники", "name_en"=>"Kotel'niki", "lang"=>"ru", "lat"=>55.65, "lon"=>37.8666666666667),
"/RU/CFO/MO/50/LBN" => array("type"=>"city", "pop"=>68400, "name_ru"=>"Лобня", "name_en"=>"Lobnja", "lang"=>"ru", "lat"=>56.0166666666667, "lon"=>37.4833333333333),
"/RU/CFO/MO/50/LBR" => array("type"=>"city", "pop"=>159000, "name_ru"=>"Люберцы", "name_en"=>"Lubertsy", "lang"=>"ru", "lat"=>55.6666666666667, "lon"=>37.9166666666667),
"/RU/CFO/MO/50/LHV" => array("type"=>"city", "pop"=>31900, "name_ru"=>"Луховицы", "name_en"=>"Luhovitsy", "lang"=>"ru", "lat"=>54.9833333333333, "lon"=>39.0333333333333),
"/RU/CFO/MO/50/LKDL" => array("type"=>"city", "pop"=>31000, "name_ru"=>"Ликино-Дулёво", "name_en"=>"Likino-Duljovo", "lang"=>"ru", "lat"=>55.7, "lon"=>38.95),
"/RU/CFO/MO/50/LSPT" => array("type"=>"city", "pop"=>22200, "name_ru"=>"Лосино-Петровский", "name_en"=>"Losino-Petrovskiy", "lang"=>"ru", "lat"=>55.8666666666667, "lon"=>38.2),
"/RU/CFO/MO/50/LTKR" => array("type"=>"city", "pop"=>51700, "name_ru"=>"Лыткарино", "name_en"=>"Lytkarino", "lang"=>"ru", "lat"=>55.5833333333333, "lon"=>37.9),
"/RU/CFO/MO/50/MSKY" => array("type"=>"city", "pop"=>15600, "name_ru"=>"Московский", "name_en"=>"Moskovskiy", "lang"=>"ru", "lat"=>55.6, "lon"=>37.35),
"/RU/CFO/MO/50/MTSC" => array("type"=>"city", "pop"=>163400, "name_ru"=>"Мытищи", "name_en"=>"Mytischi", "lang"=>"ru", "lat"=>55.9166666666667, "lon"=>37.7333333333333),
"/RU/CFO/MO/50/MZH" => array("type"=>"city", "pop"=>30900, "name_ru"=>"Можайск", "name_en"=>"Mozhajsk", "lang"=>"ru", "lat"=>55.5, "lon"=>36.0333333333333),
"/RU/CFO/MO/50/NGNS" => array("type"=>"city", "pop"=>115800, "name_ru"=>"Ногинск", "name_en"=>"Noginsk", "lang"=>"ru", "lat"=>55.85, "lon"=>38.4333333333333),
"/RU/CFO/MO/50/NRFM" => array("type"=>"city", "pop"=>71400, "name_ru"=>"Наро-Фоминск", "name_en"=>"Naro-Fominsk", "lang"=>"ru", "lat"=>55.3833333333333, "lon"=>36.7333333333333),
"/RU/CFO/MO/50/ODNC" => array("type"=>"city", "pop"=>129000, "name_ru"=>"Одинцово", "name_en"=>"Odintsovo", "lang"=>"ru", "lat"=>55.6666666666667, "lon"=>37.2833333333333),
"/RU/CFO/MO/50/ORZU" => array("type"=>"city", "pop"=>121700, "name_ru"=>"Орехово-Зуево", "name_en"=>"Orehovo-Zuevo", "lang"=>"ru", "lat"=>55.8, "lon"=>38.9666666666667),
"/RU/CFO/MO/50/OZHR" => array("type"=>"city", "pop"=>10600, "name_ru"=>"Ожерелье", "name_en"=>"Ozherelje", "lang"=>"ru", "lat"=>54.8, "lon"=>38.2666666666667),
"/RU/CFO/MO/50/OZR" => array("type"=>"city", "pop"=>26000, "name_ru"=>"Озёры", "name_en"=>"Ozjory", "lang"=>"ru", "lat"=>54.85, "lon"=>38.55),
"/RU/CFO/MO/50/PDL" => array("type"=>"city", "pop"=>180000, "name_ru"=>"Подольск", "name_en"=>"Podolsk", "lang"=>"ru", "lat"=>55.4166666666667, "lon"=>37.55),
"/RU/CFO/MO/50/PRSV" => array("type"=>"city", "pop"=>14600, "name_ru"=>"Пересвет", "name_en"=>"Peresvet", "lang"=>"ru", "lat"=>56.4166666666667, "lon"=>38.1833333333333),
"/RU/CFO/MO/50/PRTV" => array("type"=>"city", "pop"=>37100, "name_ru"=>"Протвино", "name_en"=>"Protvino", "lang"=>"ru", "lat"=>54.8666666666667, "lon"=>37.2166666666667),
"/RU/CFO/MO/50/PSCH" => array("type"=>"city", "pop"=>20100, "name_ru"=>"Пущино", "name_en"=>"Puschino", "lang"=>"ru", "lat"=>54.8333333333333, "lon"=>37.6166666666667),
"/RU/CFO/MO/50/PSHK" => array("type"=>"city", "pop"=>97700, "name_ru"=>"Пушкино", "name_en"=>"Pushkino", "lang"=>"ru", "lat"=>56.0166666666667, "lon"=>37.85),
"/RU/CFO/MO/50/PVPS" => array("type"=>"city", "pop"=>61400, "name_ru"=>"Павловский Посад", "name_en"=>"Pavlovskiy Posad", "lang"=>"ru", "lat"=>55.7833333333333, "lon"=>38.65),
"/RU/CFO/MO/50/REUT" => array("type"=>"city", "pop"=>81600, "name_ru"=>"Реутов", "name_en"=>"Reutov", "lang"=>"ru", "lat"=>55.7666666666667, "lon"=>37.8666666666667),
"/RU/CFO/MO/50/RMNS" => array("type"=>"city", "pop"=>82300, "name_ru"=>"Раменское", "name_en"=>"Ramenskoe", "lang"=>"ru", "lat"=>55.5666666666667, "lon"=>38.2166666666667),
"/RU/CFO/MO/50/RSHL" => array("type"=>"city", "pop"=>21400, "name_ru"=>"Рошаль", "name_en"=>"Roshal", "lang"=>"ru", "lat"=>55.6666666666667, "lon"=>39.8833333333333),
"/RU/CFO/MO/50/RUZA" => array("type"=>"city", "pop"=>13200, "name_ru"=>"Руза", "name_en"=>"Ruza", "lang"=>"ru", "lat"=>55.7, "lon"=>36.2),
"/RU/CFO/MO/50/SCHLK" => array("type"=>"city", "pop"=>112800, "name_ru"=>"Щёлково", "name_en"=>"Shchjolkovo", "lang"=>"ru", "lat"=>55.9166666666667, "lon"=>38),
"/RU/CFO/MO/50/SCHR" => array("type"=>"city", "pop"=>30900, "name_ru"=>"Щербинка", "name_en"=>"Scherbinka", "lang"=>"ru", "lat"=>55.5, "lon"=>37.5666666666667),
"/RU/CFO/MO/50/SHTR" => array("type"=>"city", "pop"=>31200, "name_ru"=>"Шатура", "name_en"=>"Shatura", "lang"=>"ru", "lat"=>55.5666666666667, "lon"=>39.5333333333333),
"/RU/CFO/MO/50/SLGR" => array("type"=>"city", "pop"=>56700, "name_ru"=>"Солнечногорск", "name_en"=>"Solnechnogorsk", "lang"=>"ru", "lat"=>56.1833333333333, "lon"=>36.9833333333333),
"/RU/CFO/MO/50/SRGP" => array("type"=>"city", "pop"=>109300, "name_ru"=>"Сергиев Посад", "name_en"=>"Sergiev Posad", "lang"=>"ru", "lat"=>56.3, "lon"=>38.1333333333333),
"/RU/CFO/MO/50/SRPH" => array("type"=>"city", "pop"=>123800, "name_ru"=>"Серпухов", "name_en"=>"Serpuhov", "lang"=>"ru", "lat"=>54.9166666666667, "lon"=>37.4),
"/RU/CFO/MO/50/STKP" => array("type"=>"city", "pop"=>21200, "name_ru"=>"Старая Купавна", "name_en"=>"Staraja Kupavna", "lang"=>"ru", "lat"=>55.8, "lon"=>38.1833333333333),
"/RU/CFO/MO/50/STPN" => array("type"=>"city", "pop"=>66500, "name_ru"=>"Ступино", "name_en"=>"Stupino", "lang"=>"ru", "lat"=>54.9, "lon"=>38.0666666666667),
"/RU/CFO/MO/50/TLDM" => array("type"=>"city", "pop"=>12600, "name_ru"=>"Талдом", "name_en"=>"Taldom", "lang"=>"ru", "lat"=>56.7333333333333, "lon"=>37.5333333333333),
"/RU/CFO/MO/50/TRCK" => array("type"=>"city", "pop"=>36100, "name_ru"=>"Троицк", "name_en"=>"Troitsk", "lang"=>"ru", "lat"=>55.4666666666667, "lon"=>37.3),
"/RU/CFO/MO/50/VDN" => array("type"=>"city", "pop"=>53700, "name_ru"=>"Видное", "name_en"=>"Vidnoe", "lang"=>"ru", "lat"=>55.55, "lon"=>37.7166666666667),
"/RU/CFO/MO/50/VLKL" => array("type"=>"city", "pop"=>24000, "name_ru"=>"Волоколамск", "name_en"=>"Volokolamsk", "lang"=>"ru", "lat"=>56.0333333333333, "lon"=>35.95),
"/RU/CFO/MO/50/VRJ" => array("type"=>"city", "pop"=>4800, "name_ru"=>"Верея", "name_en"=>"Vereja", "lang"=>"ru", "lat"=>55.35, "lon"=>36.1833333333333),
"/RU/CFO/MO/50/VSKR" => array("type"=>"city", "pop"=>90600, "name_ru"=>"Воскресенск", "name_en"=>"Voskresensk", "lang"=>"ru", "lat"=>55.3166666666667, "lon"=>38.6833333333333),
"/RU/CFO/MO/50/VSKV" => array("type"=>"city", "pop"=>10700, "name_ru"=>"Высоковск", "name_en"=>"Vysokovsk", "lang"=>"ru", "lat"=>56.3166666666667, "lon"=>36.55),
"/RU/CFO/MO/50/ZHDR" => array("type"=>"city", "pop"=>122600, "name_ru"=>"Железнодорожный", "name_en"=>"Zheleznodorozhnyy", "lang"=>"ru", "lat"=>55.75, "lon"=>38.0166666666667),
"/RU/CFO/MO/50/ZHKS" => array("type"=>"city", "pop"=>103100, "name_ru"=>"Жуковский", "name_en"=>"Zhukovskiy", "lang"=>"ru", "lat"=>55.6, "lon"=>38.1166666666667),
"/RU/CFO/MO/50/ZRSK" => array("type"=>"city", "pop"=>24100, "name_ru"=>"Зарайск", "name_en"=>"Zarajsk", "lang"=>"ru", "lat"=>54.7666666666667, "lon"=>38.8833333333333),
"/RU/CFO/MO/50/ZVGR" => array("type"=>"city", "pop"=>12900, "name_ru"=>"Звенигород", "name_en"=>"Zvenigorod", "lang"=>"ru", "lat"=>55.7333333333333, "lon"=>36.85),
"/RU/CFO/MO/MSK" => array("type"=>"city", "pop"=>10470300, "name_ru"=>"Москва", "name_en"=>"Moscow", "lang"=>"ru", "lat"=>55.75, "lon"=>37.6166666666667),
"/RU/CFO/MO/MSK/ZLN" => array("type"=>"city", "pop"=>217124, "name_ru"=>"Зеленоград", "name_en"=>"Zelenograd", "lang"=>"ru", "lat"=>55.997917, "lon"=>37.19041755),
"/RU/DVFO" => array("type"=>"federal region", "pop"=>6460094, "name_ru"=>"Дальневосточный федеральный округ", "name_en"=>"Dalnevostochnyy Federal Region", "lang"=>"ru"),
"/RU/DVFO/14" => array("type"=>"region", "pop"=>951400, "name_ru"=>"Республика Саха (Якутия)", "name_en"=>"Saha Republic", "lang"=>"ru"),
"/RU/DVFO/14/NER" => array("type"=>"city", "pop"=>64400, "name_ru"=>"Нерюнгри", "name_en"=>"Nerjungri", "lang"=>"ru", "lat"=>56.6666666666667, "lon"=>124.7),
"/RU/DVFO/14/YAK" => array("type"=>"city", "pop"=>255800, "name_ru"=>"Якутск", "name_en"=>"Yakutsk", "lang"=>"ru", "lat"=>62.0333333333333, "lon"=>129.733333333333),
"/RU/DVFO/25" => array("type"=>"region", "pop"=>1995800, "name_ru"=>"Приморский край", "name_en"=>"Primorskiy Kray", "lang"=>"ru"),
"/RU/DVFO/25/ART" => array("type"=>"city", "pop"=>102400, "name_ru"=>"Артём", "name_en"=>"Artjom", "lang"=>"ru", "lat"=>43.35, "lon"=>132.183333333333),
"/RU/DVFO/25/NHD" => array("type"=>"city", "pop"=>168500, "name_ru"=>"Находка", "name_en"=>"Nahodka", "lang"=>"ru", "lat"=>42.8166666666667, "lon"=>132.866666666667),
"/RU/DVFO/25/USS" => array("type"=>"city", "pop"=>153400, "name_ru"=>"Уссурийск", "name_en"=>"Ussurijsk", "lang"=>"ru", "lat"=>43.8, "lon"=>131.95),
"/RU/DVFO/25/VLA" => array("type"=>"city", "pop"=>578800, "name_ru"=>"Владивосток", "name_en"=>"Vladivostok", "lang"=>"ru", "lat"=>43.1166666666667, "lon"=>131.9),
"/RU/DVFO/27" => array("type"=>"region", "pop"=>1403700, "name_ru"=>"Хабаровский край", "name_en"=>"Habarovskiy Kray", "lang"=>"ru"),
"/RU/DVFO/27/AMR" => array("type"=>"city", "pop"=>46100, "name_ru"=>"Амурск", "name_en"=>"Amursk", "lang"=>"ru", "lat"=>50.2166666666667, "lon"=>136.9),
"/RU/DVFO/27/HAB" => array("type"=>"city", "pop"=>577300, "name_ru"=>"Хабаровск", "name_en"=>"Habarovsk", "lang"=>"ru", "lat"=>48.4833333333333, "lon"=>135.066666666667),
"/RU/DVFO/27/KNA" => array("type"=>"city", "pop"=>272400, "name_ru"=>"Комсомольск-на-Амуре", "name_en"=>"Komsomol'sk-on-Amur", "lang"=>"ru", "lat"=>50.55, "lon"=>137),
"/RU/DVFO/28" => array("type"=>"region", "pop"=>869600, "name_ru"=>"Амурская область", "name_en"=>"Amurskaya Region", "lang"=>"ru"),
"/RU/DVFO/28/BEL" => array("type"=>"city", "pop"=>67800, "name_ru"=>"Белогорск", "name_en"=>"Belogorsk", "lang"=>"ru", "lat"=>50.9166666666667, "lon"=>128.483333333333),
"/RU/DVFO/28/BLA" => array("type"=>"city", "pop"=>207300, "name_ru"=>"Благовещенск", "name_en"=>"Blagoveschensk", "lang"=>"ru", "lat"=>50.2833333333333, "lon"=>123.533333333333),
"/RU/DVFO/28/SVB" => array("type"=>"city", "pop"=>60000, "name_ru"=>"Свободный", "name_en"=>"Svobodnyy", "lang"=>"ru", "lat"=>51.3833333333333, "lon"=>128.133333333333),
"/RU/DVFO/41" => array("type"=>"region", "pop"=>345700, "name_ru"=>"Камчатский край", "name_en"=>"Kamchatskiy Kray", "lang"=>"ru"),
"/RU/DVFO/41/PPK" => array("type"=>"city", "pop"=>194900, "name_ru"=>"Петропавловск-Камчатский", "name_en"=>"Petropavlovsk-Kamchatskiy", "lang"=>"ru", "lat"=>53.0166666666667, "lon"=>158.65),
"/RU/DVFO/49" => array("type"=>"region", "pop"=>165800, "name_ru"=>"Магаданская область", "name_en"=>"Magadanskaya Region", "lang"=>"ru"),
"/RU/DVFO/49/MAG" => array("type"=>"city", "pop"=>99600, "name_ru"=>"Магадан", "name_en"=>"Magadan", "lang"=>"ru", "lat"=>59.5666666666667, "lon"=>150.8),
"/RU/DVFO/65" => array("type"=>"region", "pop"=>518500, "name_ru"=>"Сахалинская область", "name_en"=>"Sahalinskaya Region", "lang"=>"ru"),
"/RU/DVFO/65/YSH" => array("type"=>"city", "pop"=>173800, "name_ru"=>"Южно-Сахалинск", "name_en"=>"Yuzhno-Sahalinsk", "lang"=>"ru", "lat"=>46.95, "lon"=>142.733333333333),
"/RU/DVFO/79" => array("type"=>"region", "pop"=>185500, "name_ru"=>"Еврейская автономная область", "name_en"=>"Evreyskaya avtonomnaya oblast", "lang"=>"ru"),
"/RU/DVFO/79/BRB" => array("type"=>"city", "pop"=>75100, "name_ru"=>"Биробиджан", "name_en"=>"Birobidzhan", "lang"=>"ru", "lat"=>48.7833333333333, "lon"=>132.933333333333),
"/RU/DVFO/87" => array("type"=>"region", "pop"=>50300, "name_ru"=>"Чукотский Автономный округ", "name_en"=>"Chukotskiy Autonomous Okrug", "lang"=>"ru"),
"/RU/DVFO/87/ANA" => array("type"=>"city", "pop"=>11800, "name_ru"=>"Анадырь", "name_en"=>"Anadyr'", "lang"=>"ru", "lat"=>64.7333333333333, "lon"=>177.516666666667),
"/RU/PVFO" => array("type"=>"federal region", "pop"=>30157844, "name_ru"=>"Приволжский федеральный округ", "name_en"=>"Privolzhskiy Federal Region", "lang"=>"ru"),
"/RU/PVFO/02" => array("type"=>"region", "pop"=>4052700, "name_ru"=>"Республика Башкортостан", "name_en"=>"Republic of Bashkortostan", "lang"=>"ru"),
"/RU/PVFO/02/BLK" => array("type"=>"city", "pop"=>68800, "name_ru"=>"Белорецк", "name_en"=>"Beloretsk", "lang"=>"ru", "lat"=>53.9666666666667, "lon"=>58.4),
"/RU/PVFO/02/ISHB" => array("type"=>"city", "pop"=>68100, "name_ru"=>"Ишимбай", "name_en"=>"Ishimbay", "lang"=>"ru", "lat"=>53.45, "lon"=>56.0333333333333),
"/RU/PVFO/02/KMR" => array("type"=>"city", "pop"=>62400, "name_ru"=>"Кумертау", "name_en"=>"Kumertau", "lang"=>"ru", "lat"=>52.7666666666667, "lon"=>55.7833333333333),
"/RU/PVFO/02/MLZ" => array("type"=>"city", "pop"=>61800, "name_ru"=>"Мелеуз", "name_en"=>"Meleuz", "lang"=>"ru", "lat"=>52.95, "lon"=>55.9333333333333),
"/RU/PVFO/02/NFK" => array("type"=>"city", "pop"=>118000, "name_ru"=>"Нефтекамск", "name_en"=>"Neftekamsk", "lang"=>"ru", "lat"=>56.0833333333333, "lon"=>54.25),
"/RU/PVFO/02/OKT" => array("type"=>"city", "pop"=>108200, "name_ru"=>"Октябрьский", "name_en"=>"Oktjabr'skiy", "lang"=>"ru", "lat"=>54.4833333333333, "lon"=>53.4833333333333),
"/RU/PVFO/02/SBJ" => array("type"=>"city", "pop"=>65200, "name_ru"=>"СИбай", "name_en"=>"Sibay", "lang"=>"ru", "lat"=>52.7, "lon"=>58.65),
"/RU/PVFO/02/SLT" => array("type"=>"city", "pop"=>155900, "name_ru"=>"Салават", "name_en"=>"Salavat", "lang"=>"ru", "lat"=>53.3666666666667, "lon"=>55.9333333333333),
"/RU/PVFO/02/STK" => array("type"=>"city", "pop"=>268300, "name_ru"=>"Стерлитамак", "name_en"=>"Sterlitamak", "lang"=>"ru", "lat"=>53.6333333333333, "lon"=>55.95),
"/RU/PVFO/02/TMZ" => array("type"=>"city", "pop"=>65700, "name_ru"=>"Туймазы", "name_en"=>"Tujmazy", "lang"=>"ru", "lat"=>54.6, "lon"=>53.7),
"/RU/PVFO/02/UFA" => array("type"=>"city", "pop"=>1021500, "name_ru"=>"Уфа", "name_en"=>"Ufa", "lang"=>"ru", "lat"=>54.7333333333333, "lon"=>55.9666666666667),
"/RU/PVFO/12" => array("type"=>"region", "pop"=>703200, "name_ru"=>"Республика Марий Эл", "name_en"=>"Mari El Republic", "lang"=>"ru"),
"/RU/PVFO/12/VLG" => array("type"=>"city", "pop"=>56900, "name_ru"=>"Волжск", "name_en"=>"Volzhsk", "lang"=>"ru", "lat"=>55.8666666666667, "lon"=>48.35),
"/RU/PVFO/12/YSHR" => array("type"=>"city", "pop"=>248700, "name_ru"=>"Йошкар-Ола", "name_en"=>"Yoshkar-Ola", "lang"=>"ru", "lat"=>56.6333333333333, "lon"=>47.8666666666667),
"/RU/PVFO/13" => array("type"=>"region", "pop"=>840400, "name_ru"=>"Республика Мордовия", "name_en"=>"Repablik of Mordovia", "lang"=>"ru"),
"/RU/PVFO/13/SRN" => array("type"=>"city", "pop"=>295300, "name_ru"=>"Саранск", "name_en"=>"Saransk", "lang"=>"ru", "lat"=>54.1833333333333, "lon"=>45.1833333333333),
"/RU/PVFO/16" => array("type"=>"region", "pop"=>3762800, "name_ru"=>"Республика Татарстан", "name_en"=>"Tatarstan Republic", "lang"=>"ru"),
"/RU/PVFO/16/ALM" => array("type"=>"city", "pop"=>142000, "name_ru"=>"Альметьевск", "name_en"=>"Al'metjevsk", "lang"=>"ru", "lat"=>54.9, "lon"=>52.3),
"/RU/PVFO/16/BGL" => array("type"=>"city", "pop"=>90600, "name_ru"=>"Бугульма", "name_en"=>"Bugul'ma", "lang"=>"ru", "lat"=>54.5333333333333, "lon"=>52.7833333333333),
"/RU/PVFO/16/ELB" => array("type"=>"city", "pop"=>70000, "name_ru"=>"Елабуга", "name_en"=>"Elabuga", "lang"=>"ru", "lat"=>55.7666666666667, "lon"=>52.0333333333333),
"/RU/PVFO/16/KZN" => array("type"=>"city", "pop"=>1120200, "name_ru"=>"Казань", "name_en"=>"Kazan'", "lang"=>"ru", "lat"=>55.7833333333333, "lon"=>49.1166666666667),
"/RU/PVFO/16/NCH" => array("type"=>"city", "pop"=>506100, "name_ru"=>"Набережные Челны", "name_en"=>"Naberezhnye Chelny", "lang"=>"ru", "lat"=>55.7, "lon"=>52.3333333333333),
"/RU/PVFO/16/NZK" => array("type"=>"city", "pop"=>226400, "name_ru"=>"Нижнекамск", "name_en"=>"Nizhnekamsk", "lang"=>"ru", "lat"=>55.6333333333333, "lon"=>51.8166666666667),
"/RU/PVFO/16/ZLN" => array("type"=>"city", "pop"=>99100, "name_ru"=>"Зеленодольск", "name_en"=>"Zelenodol'sk", "lang"=>"ru", "lat"=>55.85, "lon"=>48.5166666666667),
"/RU/PVFO/18" => array("type"=>"region", "pop"=>1532700, "name_ru"=>"Удмуртская республика", "name_en"=>"Udmirtiya Republic", "lang"=>"ru"),
"/RU/PVFO/18/GLZ" => array("type"=>"city", "pop"=>98200, "name_ru"=>"Глазов", "name_en"=>"Glazov", "lang"=>"ru", "lat"=>58.1333333333333, "lon"=>52.65),
"/RU/PVFO/18/IZH" => array("type"=>"city", "pop"=>613300, "name_ru"=>"Ижевск", "name_en"=>"Izhevsk", "lang"=>"ru", "lat"=>56.85, "lon"=>53.2166666666667),
"/RU/PVFO/18/SRP" => array("type"=>"city", "pop"=>99400, "name_ru"=>"Сарапул", "name_en"=>"Sarapul", "lang"=>"ru", "lat"=>56.4666666666667, "lon"=>53.8),
"/RU/PVFO/18/VTK" => array("type"=>"city", "pop"=>97300, "name_ru"=>"Воткинск", "name_en"=>"Votkinsk", "lang"=>"ru", "lat"=>57.05, "lon"=>54),
"/RU/PVFO/21" => array("type"=>"region", "pop"=>1282600, "name_ru"=>"Чувашская республика", "name_en"=>"Chuvashskaya Republic", "lang"=>"ru"),
"/RU/PVFO/21/CHE" => array("type"=>"city", "pop"=>441600, "name_ru"=>"Чебоксары", "name_en"=>"Cheboksary", "lang"=>"ru", "lat"=>56.1333333333333, "lon"=>47.25),
"/RU/PVFO/21/KNS" => array("type"=>"city", "pop"=>47300, "name_ru"=>"Канаш", "name_en"=>"Kanash", "lang"=>"ru", "lat"=>55.5166666666667, "lon"=>47.5),
"/RU/PVFO/21/NVC" => array("type"=>"city", "pop"=>126300, "name_ru"=>"Новочебоксарск", "name_en"=>"Novocheboksarsk", "lang"=>"ru", "lat"=>56.1166666666667, "lon"=>47.5),
"/RU/PVFO/43" => array("type"=>"region", "pop"=>1413300, "name_ru"=>"Кировская область", "name_en"=>"Kirov Oblast", "lang"=>"ru"),
"/RU/PVFO/43/CHRP" => array("type"=>"city", "pop"=>84900, "name_ru"=>"Кирово-Чепецк", "name_en"=>"Kirovo-Chepetsk", "lang"=>"ru", "lat"=>58.55, "lon"=>50.0333333333333),
"/RU/PVFO/43/KRV" => array("type"=>"city", "pop"=>464600, "name_ru"=>"Киров", "name_en"=>"Kirov", "lang"=>"ru", "lat"=>58.6, "lon"=>49.65),
"/RU/PVFO/43/KTCH" => array("type"=>"city", "pop"=>26600, "name_ru"=>"Котельнич", "name_en"=>"Kotel'nich", "lang"=>"ru", "lat"=>58.3, "lon"=>48.3333333333333),
"/RU/PVFO/43/VPL" => array("type"=>"city", "pop"=>38800, "name_ru"=>"Вятские Поляны", "name_en"=>"Vjatskie Poljany", "lang"=>"ru", "lat"=>56.2166666666667, "lon"=>51.0666666666667),
"/RU/PVFO/52" => array("type"=>"region", "pop"=>3359800, "name_ru"=>"Нижегородская область", "name_en"=>"Nizhegorodskaya Region", "lang"=>"ru"),
"/RU/PVFO/52/ARZ" => array("type"=>"city", "pop"=>106100, "name_ru"=>"Арзамас", "name_en"=>"Arzamas", "lang"=>"ru", "lat"=>55.4, "lon"=>43.8166666666667),
"/RU/PVFO/52/BLH" => array("type"=>"city", "pop"=>54100, "name_ru"=>"Балахна", "name_en"=>"Balahna", "lang"=>"ru", "lat"=>56.5, "lon"=>43.6),
"/RU/PVFO/52/BOR" => array("type"=>"city", "pop"=>76200, "name_ru"=>"Бор", "name_en"=>"Bor", "lang"=>"ru", "lat"=>56.35, "lon"=>44.0833333333333),
"/RU/PVFO/52/DZR" => array("type"=>"city", "pop"=>247500, "name_ru"=>"Дзержинск", "name_en"=>"Dzershinsk", "lang"=>"ru", "lat"=>56.2333333333333, "lon"=>43.45),
"/RU/PVFO/52/KST" => array("type"=>"city", "pop"=>65600, "name_ru"=>"Кстово", "name_en"=>"Kstovo", "lang"=>"ru", "lat"=>56.15, "lon"=>44.2),
"/RU/PVFO/52/NNOV" => array("type"=>"city", "pop"=>1274700, "name_ru"=>"Нижний Новгород", "name_en"=>"Nizhniy Novgorod", "lang"=>"ru", "lat"=>56.3166666666667, "lon"=>44),
"/RU/PVFO/52/PVL" => array("type"=>"city", "pop"=>61100, "name_ru"=>"Павлово", "name_en"=>"Pavlovo", "lang"=>"ru", "lat"=>55.95, "lon"=>43.0666666666667),
"/RU/PVFO/52/SRV" => array("type"=>"city", "pop"=>87900, "name_ru"=>"Саров", "name_en"=>"Sarov", "lang"=>"ru", "lat"=>54.9333333333333, "lon"=>43.3166666666667),
"/RU/PVFO/52/VKS" => array("type"=>"city", "pop"=>57200, "name_ru"=>"Выкса", "name_en"=>"Vyksa", "lang"=>"ru", "lat"=>55.3166666666667, "lon"=>42.1833333333333),
"/RU/PVFO/56" => array("type"=>"region", "pop"=>2119000, "name_ru"=>"Оренбургская область", "name_en"=>"Orenburg Region", "lang"=>"ru"),
"/RU/PVFO/56/BGR" => array("type"=>"city", "pop"=>52500, "name_ru"=>"Бугуруслан", "name_en"=>"Buguruslan", "lang"=>"ru", "lat"=>53.6166666666667, "lon"=>52.4166666666667),
"/RU/PVFO/56/BZL" => array("type"=>"city", "pop"=>88900, "name_ru"=>"Бузулук", "name_en"=>"Buzuluk", "lang"=>"ru", "lat"=>52.7666666666667, "lon"=>52.2666666666667),
"/RU/PVFO/56/NVT" => array("type"=>"city", "pop"=>102200, "name_ru"=>"Новотроицк", "name_en"=>"Novotroitsk", "lang"=>"ru", "lat"=>51.2, "lon"=>58.3),
"/RU/PVFO/56/ORE" => array("type"=>"city", "pop"=>526400, "name_ru"=>"Оренбург", "name_en"=>"Orenburg", "lang"=>"ru", "lat"=>51.7666666666667, "lon"=>55.1),
"/RU/PVFO/56/ORS" => array("type"=>"city", "pop"=>245500, "name_ru"=>"Орск", "name_en"=>"Orsk", "lang"=>"ru", "lat"=>51.2, "lon"=>58.5666666666667),
"/RU/PVFO/58" => array("type"=>"region", "pop"=>1388000, "name_ru"=>"Пензенская область", "name_en"=>"Penza Region", "lang"=>"ru"),
"/RU/PVFO/58/KZN" => array("type"=>"city", "pop"=>88300, "name_ru"=>"Кузнецк", "name_en"=>"Kuznetsk", "lang"=>"ru", "lat"=>53.1166666666667, "lon"=>46.6),
"/RU/PVFO/58/PNZ" => array("type"=>"city", "pop"=>507800, "name_ru"=>"Пенза", "name_en"=>"Penza", "lang"=>"ru", "lat"=>53.2, "lon"=>45.0166666666667),
"/RU/PVFO/58/ZRC" => array("type"=>"city", "pop"=>62000, "name_ru"=>"Заречный", "name_en"=>"Zarechnyy", "lang"=>"ru", "lat"=>53.2, "lon"=>45.2),
"/RU/PVFO/59" => array("type"=>"region", "pop"=>2718200, "name_ru"=>"Пермский край", "name_en"=>"Perm Region", "lang"=>"ru"),
"/RU/PVFO/59/BRZ" => array("type"=>"city", "pop"=>166000, "name_ru"=>"Березники", "name_en"=>"Berezniki", "lang"=>"ru", "lat"=>59.4166666666667, "lon"=>56.7833333333333),
"/RU/PVFO/59/PRM" => array("type"=>"city", "pop"=>987200, "name_ru"=>"Пермь", "name_en"=>"Perm", "lang"=>"ru", "lat"=>58, "lon"=>56.25),
"/RU/PVFO/59/SLK" => array("type"=>"city", "pop"=>97300, "name_ru"=>"Соликамск", "name_en"=>"Solikamsk", "lang"=>"ru", "lat"=>59.6333333333333, "lon"=>56.7666666666667),
"/RU/PVFO/63" => array("type"=>"region", "pop"=>3172800, "name_ru"=>"Самарская область", "name_en"=>"Samara Region", "lang"=>"ru"),
"/RU/PVFO/63/NVK" => array("type"=>"city", "pop"=>111000, "name_ru"=>"Новокуйбышевск", "name_en"=>"Novokujbyshevsk", "lang"=>"ru", "lat"=>53.1, "lon"=>49.9166666666667),
"/RU/PVFO/63/SAM" => array("type"=>"city", "pop"=>1135400, "name_ru"=>"Самара", "name_en"=>"Samara", "lang"=>"ru", "lat"=>53.1833333333333, "lon"=>50.1166666666667),
"/RU/PVFO/63/SYZ" => array("type"=>"city", "pop"=>179100, "name_ru"=>"Сызрань", "name_en"=>"Syzran", "lang"=>"ru", "lat"=>53.1666666666667, "lon"=>48.4666666666667),
"/RU/PVFO/63/TLT" => array("type"=>"city", "pop"=>705500, "name_ru"=>"Тольятти", "name_en"=>"Togliatti", "lang"=>"ru", "lat"=>53.5166666666667, "lon"=>49.4166666666667),
"/RU/PVFO/64" => array("type"=>"region", "pop"=>2583800, "name_ru"=>"Саратовская область", "name_en"=>"Saratov Region", "lang"=>"ru"),
"/RU/PVFO/64/BLK" => array("type"=>"city", "pop"=>198700, "name_ru"=>"Балаково", "name_en"=>"Balakovo", "lang"=>"ru", "lat"=>52, "lon"=>47.8),
"/RU/PVFO/64/BLS" => array("type"=>"city", "pop"=>92500, "name_ru"=>"Балашов", "name_en"=>"Balashov", "lang"=>"ru", "lat"=>51.55, "lon"=>43.1666666666667),
"/RU/PVFO/64/ENG" => array("type"=>"city", "pop"=>204400, "name_ru"=>"Энгельс", "name_en"=>"Engels", "lang"=>"ru", "lat"=>51.4833333333333, "lon"=>46.1166666666667),
"/RU/PVFO/64/SRT" => array("type"=>"city", "pop"=>836100, "name_ru"=>"Саратов", "name_en"=>"Saratov", "lang"=>"ru", "lat"=>51.5333333333333, "lon"=>46),
"/RU/PVFO/64/VLS" => array("type"=>"city", "pop"=>68600, "name_ru"=>"Вольск", "name_en"=>"Volsk", "lang"=>"ru", "lat"=>52.05, "lon"=>47.3833333333333),
"/RU/PVFO/73" => array("type"=>"region", "pop"=>1312200, "name_ru"=>"Ульяновская область", "name_en"=>"Uljanovskaya Region", "lang"=>"ru"),
"/RU/PVFO/73/DMT" => array("type"=>"city", "pop"=>127600, "name_ru"=>"Димитровград", "name_en"=>"Dimitrovgrad", "lang"=>"ru", "lat"=>54.2166666666667, "lon"=>49.6),
"/RU/PVFO/73/ULJ" => array("type"=>"city", "pop"=>607000, "name_ru"=>"Ульяновск", "name_en"=>"Uljanovsk", "lang"=>"ru", "lat"=>54.3166666666667, "lon"=>48.3833333333333),
"/RU/SIFO" => array("type"=>"federal region", "pop"=>19545470, "name_ru"=>"Сибирский федеральный округ", "name_en"=>"Sibirskiy Federal Region", "lang"=>"ru"),
"/RU/SIFO/03" => array("type"=>"region", "pop"=>959900, "name_ru"=>"Республика Бурятия", "name_en"=>"Burjatiya Republic", "lang"=>"ru"),
"/RU/SIFO/03/ULU" => array("type"=>"city", "pop"=>340800, "name_ru"=>"Улан-Удэ", "name_en"=>"Ulan-Ude", "lang"=>"ru", "lat"=>51.8333333333333, "lon"=>107.616666666667),
"/RU/SIFO/04" => array("type"=>"region", "pop"=>207100, "name_ru"=>"Республика Алтай", "name_en"=>"Altaj Republic", "lang"=>"ru"),
"/RU/SIFO/04/GAL" => array("type"=>"city", "pop"=>54300, "name_ru"=>"Горно-Алтайск", "name_en"=>"Gorno-Altajsk", "lang"=>"ru", "lat"=>51.95, "lon"=>85.9666666666667),
"/RU/SIFO/17" => array("type"=>"region", "pop"=>311600, "name_ru"=>"Республика Тыва", "name_en"=>"Tyva Republic", "lang"=>"ru"),
"/RU/SIFO/17/KZL" => array("type"=>"city", "pop"=>108100, "name_ru"=>"Кызыл", "name_en"=>"Kyzyl", "lang"=>"ru", "lat"=>51.7166666666667, "lon"=>94.45),
"/RU/SIFO/19" => array("type"=>"region", "pop"=>537200, "name_ru"=>"Республика Хакасия", "name_en"=>"Khakassia Republic", "lang"=>"ru"),
"/RU/SIFO/19/ABK" => array("type"=>"city", "pop"=>163200, "name_ru"=>"Абакан", "name_en"=>"Abakan", "lang"=>"ru", "lat"=>53.7166666666667, "lon"=>91.4166666666667),
"/RU/SIFO/19/CHR" => array("type"=>"city", "pop"=>74200, "name_ru"=>"Черногорск", "name_en"=>"Chernogorsk", "lang"=>"ru", "lat"=>53.8166666666667, "lon"=>91.2833333333333),
"/RU/SIFO/19/SAY" => array("type"=>"city", "pop"=>48600, "name_ru"=>"Саяногорск", "name_en"=>"Sajanogorsk", "lang"=>"ru", "lat"=>53.1, "lon"=>91.4),
"/RU/SIFO/22" => array("type"=>"region", "pop"=>2508500, "name_ru"=>"Алтайский край", "name_en"=>"Altayskiy Kray", "lang"=>"ru"),
"/RU/SIFO/22/BAR" => array("type"=>"city", "pop"=>597200, "name_ru"=>"Барнаул", "name_en"=>"Barnaul", "lang"=>"ru", "lat"=>53.3666666666667, "lon"=>83.75),
"/RU/SIFO/22/BSK" => array("type"=>"city", "pop"=>221500, "name_ru"=>"Бийск", "name_en"=>"Bijsk", "lang"=>"ru", "lat"=>52.5166666666667, "lon"=>85.1666666666667),
"/RU/SIFO/22/NVA" => array("type"=>"city", "pop"=>73200, "name_ru"=>"Новоалтайск", "name_en"=>"Novoaltajsk", "lang"=>"ru", "lat"=>53.4, "lon"=>83.9333333333333),
"/RU/SIFO/22/RBC" => array("type"=>"city", "pop"=>156200, "name_ru"=>"Рубцовск", "name_en"=>"Rubtsovsk", "lang"=>"ru", "lat"=>51.5166666666667, "lon"=>81.2),
"/RU/SIFO/24" => array("type"=>"region", "pop"=>2890400, "name_ru"=>"Красноярский край", "name_en"=>"Krasnoyarskiy Kray", "lang"=>"ru"),
"/RU/SIFO/24/ACH" => array("type"=>"city", "pop"=>110800, "name_ru"=>"Ачинск", "name_en"=>"Achinsk", "lang"=>"ru", "lat"=>56.2666666666667, "lon"=>90.5),
"/RU/SIFO/24/KNS" => array("type"=>"city", "pop"=>99000, "name_ru"=>"Канск", "name_en"=>"Kansk", "lang"=>"ru", "lat"=>56.2, "lon"=>95.7),
"/RU/SIFO/24/KRY" => array("type"=>"city", "pop"=>936400, "name_ru"=>"Красноярск", "name_en"=>"Krasnoyarsk", "lang"=>"ru", "lat"=>56, "lon"=>92.9333333333333),
"/RU/SIFO/24/LSS" => array("type"=>"city", "pop"=>64200, "name_ru"=>"Лесосибирск", "name_en"=>"Lesosibirsk", "lang"=>"ru", "lat"=>58.2333333333333, "lon"=>92.4833333333333),
"/RU/SIFO/24/MNS" => array("type"=>"city", "pop"=>66800, "name_ru"=>"Минусинск", "name_en"=>"Minusinsk", "lang"=>"ru", "lat"=>53.7, "lon"=>91.6833333333333),
"/RU/SIFO/24/NRL" => array("type"=>"city", "pop"=>206400, "name_ru"=>"Норильск", "name_en"=>"Noril'sk", "lang"=>"ru", "lat"=>69.3333333333333, "lon"=>88.2166666666667),
"/RU/SIFO/24/NZR" => array("type"=>"city", "pop"=>53600, "name_ru"=>"Назарово", "name_en"=>"Nazarovo", "lang"=>"ru", "lat"=>56, "lon"=>90.3833333333333),
"/RU/SIFO/24/ZHG" => array("type"=>"city", "pop"=>93700, "name_ru"=>"Железногорск", "name_en"=>"Zheleznogorsk", "lang"=>"ru", "lat"=>56.25, "lon"=>93.5333333333333),
"/RU/SIFO/24/ZLN" => array("type"=>"city", "pop"=>68500, "name_ru"=>"Зеленогорск", "name_en"=>"Zelenogorsk", "lang"=>"ru", "lat"=>56.1, "lon"=>94.5833333333333),
"/RU/SIFO/38" => array("type"=>"region", "pop"=>2507700, "name_ru"=>"Иркутская область", "name_en"=>"Irkutskaya Region", "lang"=>"ru"),
"/RU/SIFO/38/ANG" => array("type"=>"city", "pop"=>242500, "name_ru"=>"Ангарск", "name_en"=>"Angarsk", "lang"=>"ru", "lat"=>52.5666666666667, "lon"=>103.916666666667),
"/RU/SIFO/38/BRT" => array("type"=>"city", "pop"=>252000, "name_ru"=>"Братск", "name_en"=>"Bratsk", "lang"=>"ru", "lat"=>56.1166666666667, "lon"=>101.6),
"/RU/SIFO/38/CHR" => array("type"=>"city", "pop"=>54300, "name_ru"=>"Черемхово", "name_en"=>"Cheremhovo", "lang"=>"ru", "lat"=>53.15, "lon"=>103.066666666667),
"/RU/SIFO/38/IRK" => array("type"=>"city", "pop"=>575800, "name_ru"=>"Иркутск", "name_en"=>"Irkutsk", "lang"=>"ru", "lat"=>52.2833333333333, "lon"=>104.3),
"/RU/SIFO/38/USO" => array("type"=>"city", "pop"=>85900, "name_ru"=>"Усолье-Сибирское", "name_en"=>"Usolje-Sibirskoe", "lang"=>"ru", "lat"=>52.75, "lon"=>103.65),
"/RU/SIFO/38/UST" => array("type"=>"city", "pop"=>98000, "name_ru"=>"Усть-Илимск", "name_en"=>"Ust'-Ilimsk", "lang"=>"ru", "lat"=>58, "lon"=>102.666666666667),
"/RU/SIFO/42" => array("type"=>"region", "pop"=>2823500, "name_ru"=>"Кемеровская область", "name_en"=>"Kemerovskaya Region", "lang"=>"ru"),
"/RU/SIFO/42/ANG" => array("type"=>"city", "pop"=>82900, "name_ru"=>"Анжеро-Судженск", "name_en"=>"Anzhero-Sudzhensk", "lang"=>"ru", "lat"=>56.0833333333333, "lon"=>86.0333333333333),
"/RU/SIFO/42/BLV" => array("type"=>"city", "pop"=>75800, "name_ru"=>"Белово", "name_en"=>"Belovo", "lang"=>"ru", "lat"=>54.4166666666667, "lon"=>86.3),
"/RU/SIFO/42/KEM" => array("type"=>"city", "pop"=>520000, "name_ru"=>"Кемерово", "name_en"=>"Kemerovo", "lang"=>"ru", "lat"=>55.3333333333333, "lon"=>86.0666666666667),
"/RU/SIFO/42/KSL" => array("type"=>"city", "pop"=>103800, "name_ru"=>"Киселёвск", "name_en"=>"Kiseljovsk", "lang"=>"ru", "lat"=>53.9833333333333, "lon"=>86.7),
"/RU/SIFO/42/LNK" => array("type"=>"city", "pop"=>105400, "name_ru"=>"Ленинск-Кузнецкий", "name_en"=>"Leninsk-Kuznetskiy", "lang"=>"ru", "lat"=>54.65, "lon"=>86.1666666666667),
"/RU/SIFO/42/MZD" => array("type"=>"city", "pop"=>103800, "name_ru"=>"Междуреченск", "name_en"=>"Mezhdurechensk", "lang"=>"ru", "lat"=>53.6833333333333, "lon"=>88.05),
"/RU/SIFO/42/NVK" => array("type"=>"city", "pop"=>562200, "name_ru"=>"Новокузнецк", "name_en"=>"Novokuznetsk", "lang"=>"ru", "lat"=>53.75, "lon"=>87.1166666666667),
"/RU/SIFO/42/PRK" => array("type"=>"city", "pop"=>213200, "name_ru"=>"Прокопьевск", "name_en"=>"Prokopjevsk", "lang"=>"ru", "lat"=>53.8833333333333, "lon"=>86.7166666666667),
"/RU/SIFO/42/URG" => array("type"=>"city", "pop"=>83800, "name_ru"=>"Юрга", "name_en"=>"Yurga", "lang"=>"ru", "lat"=>55.7333333333333, "lon"=>84.9),
"/RU/SIFO/54" => array("type"=>"region", "pop"=>2635600, "name_ru"=>"Новосибирская область", "name_en"=>"Novosibirskaya Region", "lang"=>"ru"),
"/RU/SIFO/54/BRD" => array("type"=>"city", "pop"=>94600, "name_ru"=>"Бердск", "name_en"=>"Berdsk", "lang"=>"ru", "lat"=>54.75, "lon"=>83.1),
"/RU/SIFO/54/ISK" => array("type"=>"city", "pop"=>64100, "name_ru"=>"Искитим", "name_en"=>"Iskitim", "lang"=>"ru", "lat"=>54.6333333333333, "lon"=>83.3),
"/RU/SIFO/54/KBS" => array("type"=>"city", "pop"=>48000, "name_ru"=>"Куйбышев", "name_en"=>"Kujbyshev", "lang"=>"ru", "lat"=>55.45, "lon"=>78.3),
"/RU/SIFO/54/NVS" => array("type"=>"city", "pop"=>1390500, "name_ru"=>"Новосибирск", "name_en"=>"Novosibirsk", "lang"=>"ru", "lat"=>55.0333333333333, "lon"=>82.9166666666667),
"/RU/SIFO/55" => array("type"=>"region", "pop"=>2018000, "name_ru"=>"Омская область", "name_en"=>"Omskaya Region", "lang"=>"ru"),
"/RU/SIFO/55/OMS" => array("type"=>"city", "pop"=>1131100, "name_ru"=>"Омск", "name_en"=>"Omsk", "lang"=>"ru", "lat"=>54.9666666666667, "lon"=>73.3833333333333),
"/RU/SIFO/70" => array("type"=>"region", "pop"=>1035000, "name_ru"=>"Томская область", "name_en"=>"Tomsk Region", "lang"=>"ru"),
"/RU/SIFO/70/SVR" => array("type"=>"city", "pop"=>107100, "name_ru"=>"Северск", "name_en"=>"Seversk", "lang"=>"ru", "lat"=>56.6, "lon"=>84.85),
"/RU/SIFO/70/TOM" => array("type"=>"city", "pop"=>496500, "name_ru"=>"Томск", "name_en"=>"Tomsk", "lang"=>"ru", "lat"=>56.5, "lon"=>84.9666666666667),
"/RU/SIFO/75" => array("type"=>"region", "pop"=>1118900, "name_ru"=>"Забайкальский край", "name_en"=>"Zabaykal'skiy Kray", "lang"=>"ru"),
"/RU/SIFO/75/CHT" => array("type"=>"city", "pop"=>306100, "name_ru"=>"Чита", "name_en"=>"Chita", "lang"=>"ru", "lat"=>52.0333333333333, "lon"=>113.5),
"/RU/SIFO/75/KRK" => array("type"=>"city", "pop"=>55200, "name_ru"=>"Краснокаменск", "name_en"=>"Krasnokamensk", "lang"=>"ru", "lat"=>50.1, "lon"=>118.033333333333),
"/RU/SZFO" => array("type"=>"federal region", "pop"=>13462259, "name_ru"=>"Северо-Западный федеральный округ", "name_en"=>"North-West Federal Region", "lang"=>"ru"),
"/RU/SZFO/10" => array("type"=>"region", "pop"=>690700, "name_ru"=>"Республика Карелия", "name_en"=>"Republic Karelia", "lang"=>"ru"),
"/RU/SZFO/10/KND" => array("type"=>"city", "pop"=>33700, "name_ru"=>"Кондопога", "name_en"=>"Kondopoga", "lang"=>"ru", "lat"=>64.2, "lon"=>34.2666666666667),
"/RU/SZFO/10/PTZ" => array("type"=>"city", "pop"=>268800, "name_ru"=>"Петрозаводск", "name_en"=>"Petrozavodsk", "lang"=>"ru", "lat"=>61.7833333333333, "lon"=>34.35),
"/RU/SZFO/11" => array("type"=>"region", "pop"=>968200, "name_ru"=>"Республика Коми", "name_en"=>"Komi Republic", "lang"=>"ru"),
"/RU/SZFO/11/SYK" => array("type"=>"city", "pop"=>231000, "name_ru"=>"Сыктывкар", "name_en"=>"Syktyvkar", "lang"=>"ru", "lat"=>61.6666666666667, "lon"=>50.8166666666667),
"/RU/SZFO/11/UHT" => array("type"=>"city", "pop"=>103400, "name_ru"=>"Ухта", "name_en"=>"Uhta", "lang"=>"ru", "lat"=>63.5666666666667, "lon"=>53.7),
"/RU/SZFO/11/VRK" => array("type"=>"city", "pop"=>74300, "name_ru"=>"Воркута", "name_en"=>"Vorkuta", "lang"=>"ru", "lat"=>67.5, "lon"=>64.0333333333333),
"/RU/SZFO/29" => array("type"=>"region", "pop"=>1271900, "name_ru"=>"Архангельская область", "name_en"=>"Arhangel'skaya Region", "lang"=>"ru"),
"/RU/SZFO/29/ARH" => array("type"=>"city", "pop"=>348700, "name_ru"=>"Архангельск", "name_en"=>"Arhangel'sk", "lang"=>"ru", "lat"=>64.55, "lon"=>40.5333333333333),
"/RU/SZFO/29/KTL" => array("type"=>"city", "pop"=>59200, "name_ru"=>"Котлас", "name_en"=>"Kotlas", "lang"=>"ru", "lat"=>61.25, "lon"=>46.65),
"/RU/SZFO/29/SVD" => array("type"=>"city", "pop"=>191400, "name_ru"=>"Северодвинск", "name_en"=>"Severodvinsk", "lang"=>"ru", "lat"=>64.5666666666667, "lon"=>39.8166666666667),
"/RU/SZFO/35" => array("type"=>"region", "pop"=>1222900, "name_ru"=>"Вологодская область", "name_en"=>"Vologodskaya Region", "lang"=>"ru"),
"/RU/SZFO/35/CHR" => array("type"=>"city", "pop"=>308000, "name_ru"=>"Череповец", "name_en"=>"Cherepovets", "lang"=>"ru", "lat"=>59.1333333333333, "lon"=>37.9166666666667),
"/RU/SZFO/35/SKL" => array("type"=>"city", "pop"=>40700, "name_ru"=>"Сокол", "name_en"=>"Sokol", "lang"=>"ru", "lat"=>59.4666666666667, "lon"=>40.1166666666667),
"/RU/SZFO/35/VLG" => array("type"=>"city", "pop"=>286100, "name_ru"=>"Вологда", "name_en"=>"Vologda", "lang"=>"ru", "lat"=>59.2166666666667, "lon"=>39.9),
"/RU/SZFO/35/VLU" => array("type"=>"city", "pop"=>32000, "name_ru"=>"Великий Устюг", "name_en"=>"Velikiy Ustjug", "lang"=>"ru", "lat"=>60.7666666666667, "lon"=>46.3),
"/RU/SZFO/39" => array("type"=>"region", "pop"=>937400, "name_ru"=>"Калининградская область", "name_en"=>"Kaliningradskaya Region", "lang"=>"ru"),
"/RU/SZFO/39/KGD" => array("type"=>"city", "pop"=>421700, "name_ru"=>"Калининград", "name_en"=>"Kaliningrad", "lang"=>"ru", "lat"=>54.7166666666667, "lon"=>20.5),
"/RU/SZFO/51" => array("type"=>"region", "pop"=>850900, "name_ru"=>"Мурманская область", "name_en"=>"Murmanskaya Region", "lang"=>"ru"),
"/RU/SZFO/51/APT" => array("type"=>"city", "pop"=>62200, "name_ru"=>"Апатиты", "name_en"=>"Apatity", "lang"=>"ru", "lat"=>67.5666666666667, "lon"=>33.4),
"/RU/SZFO/51/MUR" => array("type"=>"city", "pop"=>314700, "name_ru"=>"Мурманск", "name_en"=>"Murmansk", "lang"=>"ru", "lat"=>68.9666666666667, "lon"=>33.0833333333333),
"/RU/SZFO/51/SVM" => array("type"=>"city", "pop"=>53500, "name_ru"=>"Североморск", "name_en"=>"Severomorsk", "lang"=>"ru", "lat"=>69.0666666666667, "lon"=>33.4),
"/RU/SZFO/53" => array("type"=>"region", "pop"=>652400, "name_ru"=>"Новгородская область", "name_en"=>"Novgorod Region", "lang"=>"ru"),
"/RU/SZFO/53/BRV" => array("type"=>"city", "pop"=>56100, "name_ru"=>"Боровичи", "name_en"=>"Borovichi", "lang"=>"ru", "lat"=>58.4, "lon"=>33.9),
"/RU/SZFO/53/CHD" => array("type"=>"city", "pop"=>16400, "name_ru"=>"Чудово", "name_en"=>"Chudovo", "lang"=>"ru", "lat"=>59.1166666666667, "lon"=>31.6666666666667),
"/RU/SZFO/53/HLM" => array("type"=>"city", "pop"=>3900, "name_ru"=>"Холм", "name_en"=>"Holm", "lang"=>"ru", "lat"=>57.15, "lon"=>31.1833333333333),
"/RU/SZFO/53/MLV" => array("type"=>"city", "pop"=>12900, "name_ru"=>"Малая Вишера", "name_en"=>"Malaja Vishera", "lang"=>"ru", "lat"=>58.85, "lon"=>32.2166666666667),
"/RU/SZFO/53/NOV" => array("type"=>"city", "pop"=>216200, "name_ru"=>"Великий Новгород", "name_en"=>"Velikiy Novgorod", "lang"=>"ru", "lat"=>58.5166666666667, "lon"=>31.2833333333333),
"/RU/SZFO/53/OKL" => array("type"=>"city", "pop"=>12800, "name_ru"=>"Окуловка", "name_en"=>"Okulovka", "lang"=>"ru", "lat"=>58.3833333333333, "lon"=>33.3),
"/RU/SZFO/53/PST" => array("type"=>"city", "pop"=>15800, "name_ru"=>"Пестово", "name_en"=>"Pestovo", "lang"=>"ru", "lat"=>58.6, "lon"=>35.8166666666667),
"/RU/SZFO/53/SLC" => array("type"=>"city", "pop"=>10600, "name_ru"=>"Сольцы", "name_en"=>"Sol'tsy", "lang"=>"ru", "lat"=>58.1166666666667, "lon"=>30.3166666666667),
"/RU/SZFO/53/STRS" => array("type"=>"city", "pop"=>33300, "name_ru"=>"Старая Русса", "name_en"=>"Staraja Russa", "lang"=>"ru", "lat"=>57.9833333333333, "lon"=>31.35),
"/RU/SZFO/53/VLD" => array("type"=>"city", "pop"=>17300, "name_ru"=>"Валдай", "name_en"=>"Valdaj", "lang"=>"ru", "lat"=>57.9666666666667, "lon"=>33.25),
"/RU/SZFO/60" => array("type"=>"region", "pop"=>705300, "name_ru"=>"Псковская область", "name_en"=>"Pskovskaya Region", "lang"=>"ru"),
"/RU/SZFO/60/PSK" => array("type"=>"city", "pop"=>194200, "name_ru"=>"Псков", "name_en"=>"Pskov", "lang"=>"ru", "lat"=>57.8166666666667, "lon"=>28.3333333333333),
"/RU/SZFO/60/VLL" => array("type"=>"city", "pop"=>100300, "name_ru"=>"Великие Луки", "name_en"=>"Velikie Luki", "lang"=>"ru", "lat"=>56.3333333333333, "lon"=>30.5166666666667),
"/RU/SZFO/83" => array("type"=>"region", "pop"=>42000, "name_ru"=>"Ненецкий автономный округ", "name_en"=>"Nenets Autonomous Okrug", "lang"=>"ru"),
"/RU/SZFO/83/NMR" => array("type"=>"city", "pop"=>19300, "name_ru"=>"Нарьян-Мар", "name_en"=>"Naryan-Mar", "lang"=>"ru", "lat"=>67.6333333333333, "lon"=>53),
"/RU/SZFO/LE" => array("type"=>"region", "name_ru"=>"Санкт Петербург и Ленинградская область", "name_en"=>"Sankt-Petersburg and Leningrad Region", "lang"=>"ru"),
"/RU/SZFO/LE/47" => array("type"=>"region", "pop"=>1633400, "name_ru"=>"Ленинградская область", "name_en"=>"Leningrad Region", "lang"=>"ru"),
"/RU/SZFO/LE/47/BKSG" => array("type"=>"city", "pop"=>16800, "name_ru"=>"Бокситогорск", "name_en"=>"Boksitogorsk", "lang"=>"ru", "lat"=>59.4666666666667, "lon"=>33.85),
"/RU/SZFO/LE/47/GTCH" => array("type"=>"city", "pop"=>89900, "name_ru"=>"Гатчина", "name_en"=>"Gatchina", "lang"=>"ru", "lat"=>59.5666666666667, "lon"=>30.1333333333333),
"/RU/SZFO/LE/47/IVGR" => array("type"=>"city", "pop"=>10900, "name_ru"=>"Ивангород", "name_en"=>"Ivangorod", "lang"=>"ru", "lat"=>59.3666666666667, "lon"=>28.2166666666667),
"/RU/SZFO/LE/47/KMGR" => array("type"=>"city", "pop"=>6000, "name_ru"=>"Каменногорск", "name_en"=>"Kamennogorsk", "lang"=>"ru", "lat"=>60.95, "lon"=>29.1333333333333),
"/RU/SZFO/LE/47/KMMN" => array("type"=>"city", "pop"=>17700, "name_ru"=>"Коммунар", "name_en"=>"Kommunar", "lang"=>"ru", "lat"=>59.6166666666667, "lon"=>30.4),
"/RU/SZFO/LE/47/KNGS" => array("type"=>"city", "pop"=>50200, "name_ru"=>"Кингисепп", "name_en"=>"Kingisepp", "lang"=>"ru", "lat"=>59.3666666666667, "lon"=>28.6),
"/RU/SZFO/LE/47/KRSH" => array("type"=>"city", "pop"=>54800, "name_ru"=>"Кириши", "name_en"=>"Kirishi", "lang"=>"ru", "lat"=>59.45, "lon"=>32.0166666666667),
"/RU/SZFO/LE/47/KRVS" => array("type"=>"city", "pop"=>23400, "name_ru"=>"Кировск", "name_en"=>"Kirovsk", "lang"=>"ru", "lat"=>59.8666666666667, "lon"=>30.9833333333333),
"/RU/SZFO/LE/47/LBN" => array("type"=>"city", "pop"=>4500, "name_ru"=>"Любань", "name_en"=>"Ljuban'", "lang"=>"ru", "lat"=>59.35, "lon"=>31.25),
"/RU/SZFO/LE/47/LDPL" => array("type"=>"city", "pop"=>22100, "name_ru"=>"Лодейное Поле", "name_en"=>"Lodejnoe Pole", "lang"=>"ru", "lat"=>60.7333333333333, "lon"=>33.55),
"/RU/SZFO/LE/47/LUGA" => array("type"=>"city", "pop"=>38900, "name_ru"=>"Луга", "name_en"=>"Luga", "lang"=>"ru", "lat"=>58.7333333333333, "lon"=>29.85),
"/RU/SZFO/LE/47/NKLS" => array("type"=>"city", "pop"=>17200, "name_ru"=>"Никольское", "name_en"=>"Nikol'skoe", "lang"=>"ru", "lat"=>59.7, "lon"=>30.7833333333333),
"/RU/SZFO/LE/47/NVLD" => array("type"=>"city", "pop"=>9500, "name_ru"=>"Новая Ладога", "name_en"=>"Novaja Ladoga", "lang"=>"ru", "lat"=>60.1, "lon"=>32.3),
"/RU/SZFO/LE/47/OTRD" => array("type"=>"city", "pop"=>21500, "name_ru"=>"Отрадное", "name_en"=>"Otradnoe", "lang"=>"ru", "lat"=>59.7833333333333, "lon"=>30.8166666666667),
"/RU/SZFO/LE/47/PDPR" => array("type"=>"city", "pop"=>19000, "name_ru"=>"Подпорожье", "name_en"=>"Podporozhje", "lang"=>"ru", "lat"=>60.9166666666667, "lon"=>34.1666666666667),
"/RU/SZFO/LE/47/PKLV" => array("type"=>"city", "pop"=>22100, "name_ru"=>"Пикалёво", "name_en"=>"Pikaljovo", "lang"=>"ru", "lat"=>59.5166666666667, "lon"=>34.1666666666667),
"/RU/SZFO/LE/47/PRMR" => array("type"=>"city", "pop"=>5400, "name_ru"=>"Приморск", "name_en"=>"Primorsk", "lang"=>"ru", "lat"=>60.3666666666667, "lon"=>28.6166666666667),
"/RU/SZFO/LE/47/PRZR" => array("type"=>"city", "pop"=>17900, "name_ru"=>"Приозерск", "name_en"=>"Priozersk", "lang"=>"ru", "lat"=>61.0333333333333, "lon"=>30.1166666666667),
"/RU/SZFO/LE/47/SHLS" => array("type"=>"city", "pop"=>12600, "name_ru"=>"Шлиссельбург", "name_en"=>"Shlissel'burg", "lang"=>"ru", "lat"=>59.95, "lon"=>31.0333333333333),
"/RU/SZFO/LE/47/SLNC" => array("type"=>"city", "pop"=>35000, "name_ru"=>"Сланцы", "name_en"=>"Slantsy", "lang"=>"ru", "lat"=>59.1166666666667, "lon"=>28.0833333333333),
"/RU/SZFO/LE/47/SRTL" => array("type"=>"city", "pop"=>42300, "name_ru"=>"Сертолово", "name_en"=>"Sertolovo", "lang"=>"ru", "lat"=>60.15, "lon"=>30.2),
"/RU/SZFO/LE/47/SSBR" => array("type"=>"city", "pop"=>66800, "name_ru"=>"Сосновый Бор", "name_en"=>"Sosnovyy Bor", "lang"=>"ru", "lat"=>59.8833333333333, "lon"=>29.1),
"/RU/SZFO/LE/47/SSTR" => array("type"=>"city", "pop"=>13300, "name_ru"=>"Сясьстрой", "name_en"=>"Sjas'stroj", "lang"=>"ru", "lat"=>60.1333333333333, "lon"=>32.5666666666667),
"/RU/SZFO/LE/47/SVTG" => array("type"=>"city", "pop"=>15300, "name_ru"=>"Светогорск", "name_en"=>"Svetogorsk", "lang"=>"ru", "lat"=>61.1166666666667, "lon"=>28.85),
"/RU/SZFO/LE/47/THVN" => array("type"=>"city", "pop"=>60800, "name_ru"=>"Тихвин", "name_en"=>"Tihvin", "lang"=>"ru", "lat"=>59.6333333333333, "lon"=>33.5),
"/RU/SZFO/LE/47/TSN" => array("type"=>"city", "pop"=>36800, "name_ru"=>"Тосно", "name_en"=>"Tosno", "lang"=>"ru", "lat"=>59.55, "lon"=>30.9),
"/RU/SZFO/LE/47/VBRG" => array("type"=>"city", "pop"=>77600, "name_ru"=>"Выборг", "name_en"=>"Vyborg", "lang"=>"ru", "lat"=>60.7166666666667, "lon"=>28.75),
"/RU/SZFO/LE/47/VLHV" => array("type"=>"city", "pop"=>45600, "name_ru"=>"Волхов", "name_en"=>"Volhov", "lang"=>"ru", "lat"=>59.9166666666667, "lon"=>32.35),
"/RU/SZFO/LE/47/VLSV" => array("type"=>"city", "pop"=>11400, "name_ru"=>"Волосово", "name_en"=>"Volosovo", "lang"=>"ru", "lat"=>59.45, "lon"=>29.4833333333333),
"/RU/SZFO/LE/47/VSCK" => array("type"=>"city", "pop"=>1700, "name_ru"=>"Высоцк", "name_en"=>"Vysotsk", "lang"=>"ru", "lat"=>60.6166666666667, "lon"=>28.5666666666667),
"/RU/SZFO/LE/47/VSVL" => array("type"=>"city", "pop"=>46300, "name_ru"=>"Всеволожск", "name_en"=>"Vsevolozhsk", "lang"=>"ru", "lat"=>60.0166666666667, "lon"=>30.6833333333333),
"/RU/SZFO/LE/SPB" => array("type"=>"city", "pop"=>4568000, "name_ru"=>"Санкт-Петербург", "name_en"=>"Sankt-Peterburg", "lang"=>"ru", "lat"=>59.95, "lon"=>30.3166666666667),
"/RU/URFO" => array("type"=>"federal region", "pop"=>12254976, "name_ru"=>"Уральский федеральный округ", "name_en"=>"Uralskiy Federal Region", "lang"=>"ru"),
"/RU/URFO/45" => array("type"=>"region", "pop"=>960400, "name_ru"=>"Курганская область", "name_en"=>"Kurganskaya Region", "lang"=>"ru"),
"/RU/URFO/45/KUR" => array("type"=>"city", "pop"=>324100, "name_ru"=>"Курган", "name_en"=>"Kurgan", "lang"=>"ru", "lat"=>55.45, "lon"=>65.3333333333333),
"/RU/URFO/45/SHD" => array("type"=>"city", "pop"=>76600, "name_ru"=>"Шадринск", "name_en"=>"Shadrinsk", "lang"=>"ru", "lat"=>56.0833333333333, "lon"=>63.6333333333333),
"/RU/URFO/66" => array("type"=>"region", "pop"=>4395600, "name_ru"=>"Свердловская область", "name_en"=>"Sverdlovskaya Region", "lang"=>"ru"),
"/RU/URFO/66/EKA" => array("type"=>"city", "pop"=>1323000, "name_ru"=>"Екатеринбург", "name_en"=>"Ekaterinburg", "lang"=>"ru", "lat"=>56.8333333333333, "lon"=>60.5833333333333),
"/RU/URFO/66/KMU" => array("type"=>"city", "pop"=>180900, "name_ru"=>"Каменск-Уральский", "name_en"=>"Kamensk-Ural'skiy", "lang"=>"ru", "lat"=>56.4, "lon"=>61.9333333333333),
"/RU/URFO/66/NTG" => array("type"=>"city", "pop"=>375700, "name_ru"=>"Нижний Тагил", "name_en"=>"Nizhniy Tagil", "lang"=>"ru", "lat"=>57.9166666666667, "lon"=>59.9666666666667),
"/RU/URFO/66/NVU" => array("type"=>"city", "pop"=>92700, "name_ru"=>"Новоуральск", "name_en"=>"Novoural'sk", "lang"=>"ru", "lat"=>57.25, "lon"=>60.0833333333333),
"/RU/URFO/66/PRV" => array("type"=>"city", "pop"=>134000, "name_ru"=>"Первоуральск", "name_en"=>"Pervoural'sk", "lang"=>"ru", "lat"=>56.9166666666667, "lon"=>59.95),
"/RU/URFO/66/SRV" => array("type"=>"city", "pop"=>98500, "name_ru"=>"Серов", "name_en"=>"Serov", "lang"=>"ru", "lat"=>59.6, "lon"=>60.5666666666667),
"/RU/URFO/72" => array("type"=>"region", "pop"=>3373400, "name_ru"=>"Тюменская область", "name_en"=>"Tjumenskaya Region", "lang"=>"ru"),
"/RU/URFO/72/ISH" => array("type"=>"city", "pop"=>64000, "name_ru"=>"Ишим", "name_en"=>"Ishim", "lang"=>"ru", "lat"=>56.1166666666667, "lon"=>69.5),
"/RU/URFO/72/TMN" => array("type"=>"city", "pop"=>560000, "name_ru"=>"Тюмень", "name_en"=>"Tjumen'", "lang"=>"ru", "lat"=>57.15, "lon"=>65.5333333333333),
"/RU/URFO/72/TOB" => array("type"=>"city", "pop"=>99800, "name_ru"=>"Тобольск", "name_en"=>"Tobol'sk", "lang"=>"ru", "lat"=>58.2, "lon"=>68.25),
"/RU/URFO/74" => array("type"=>"region", "pop"=>3511000, "name_ru"=>"Челябинская область", "name_en"=>"Cheljabinsk Region", "lang"=>"ru"),
"/RU/URFO/74/CHE" => array("type"=>"city", "pop"=>1092500, "name_ru"=>"Челябинск", "name_en"=>"Cheljabinsk", "lang"=>"ru", "lat"=>55.15, "lon"=>61.4),
"/RU/URFO/74/KPS" => array("type"=>"city", "pop"=>137200, "name_ru"=>"Копейск", "name_en"=>"Kopejsk", "lang"=>"ru", "lat"=>55.1, "lon"=>61.6166666666667),
"/RU/URFO/74/MGN" => array("type"=>"city", "pop"=>409000, "name_ru"=>"Магнитогорск", "name_en"=>"Magnitogorsk", "lang"=>"ru", "lat"=>53.3833333333333, "lon"=>59.0333333333333),
"/RU/URFO/74/MSS" => array("type"=>"city", "pop"=>152600, "name_ru"=>"Миасс", "name_en"=>"Miass", "lang"=>"ru", "lat"=>55, "lon"=>60.1),
"/RU/URFO/74/OZR" => array("type"=>"city", "pop"=>86900, "name_ru"=>"Озёрск", "name_en"=>"Ozjorsk", "lang"=>"ru", "lat"=>55.75, "lon"=>60.7166666666667),
"/RU/URFO/74/SNZ" => array("type"=>"city", "pop"=>50100, "name_ru"=>"Снежинск", "name_en"=>"Snezhinsk", "lang"=>"ru", "lat"=>56.0833333333333, "lon"=>60.7333333333333),
"/RU/URFO/74/TRC" => array("type"=>"city", "pop"=>82400, "name_ru"=>"Троицк", "name_en"=>"Troitsk", "lang"=>"ru", "lat"=>54.0833333333333, "lon"=>61.5666666666667),
"/RU/URFO/74/ZLT" => array("type"=>"city", "pop"=>188800, "name_ru"=>"Златоуст", "name_en"=>"Zlatoust", "lang"=>"ru", "lat"=>55.1666666666667, "lon"=>59.6666666666667),
"/RU/URFO/86" => array("type"=>"region", "pop"=>1505200, "name_ru"=>"Ханты-Мансийский АО - Югра", "name_en"=>"Hanty-Mansiyskiy AO - Yugra", "lang"=>"ru"),
"/RU/URFO/86/HAN" => array("type"=>"city", "pop"=>67800, "name_ru"=>"Ханты-Мансийск", "name_en"=>"Hanty-Mansijsk", "lang"=>"ru", "lat"=>61, "lon"=>69.0166666666667),
"/RU/URFO/86/KGL" => array("type"=>"city", "pop"=>58900, "name_ru"=>"Когалым", "name_en"=>"Kogalym", "lang"=>"ru", "lat"=>62.3166666666667, "lon"=>74.4833333333333),
"/RU/URFO/86/NFT" => array("type"=>"city", "pop"=>115700, "name_ru"=>"Нефтеюганск", "name_en"=>"Neftejugansk", "lang"=>"ru", "lat"=>61.1, "lon"=>72.6),
"/RU/URFO/86/NGN" => array("type"=>"city", "pop"=>56100, "name_ru"=>"Нягань", "name_en"=>"Njagan'", "lang"=>"ru", "lat"=>62.1333333333333, "lon"=>65.3833333333333),
"/RU/URFO/86/NZH" => array("type"=>"city", "pop"=>243500, "name_ru"=>"Нижневартовск", "name_en"=>"Nizhnevartovsk", "lang"=>"ru", "lat"=>60.95, "lon"=>76.55),
"/RU/URFO/86/SRG" => array("type"=>"city", "pop"=>294200, "name_ru"=>"Сургут", "name_en"=>"Surgut", "lang"=>"ru", "lat"=>61.25, "lon"=>73.4333333333333),
"/RU/URFO/89" => array("type"=>"region", "pop"=>542700, "name_ru"=>"Ямало-Ненецкий автономный округ", "name_en"=>"Yamalo-Nenets Autonomous Okrug", "lang"=>"ru"),
"/RU/URFO/89/NBR" => array("type"=>"city", "pop"=>110400, "name_ru"=>"Ноябрьск", "name_en"=>"Nojabr'sk", "lang"=>"ru", "lat"=>63.2, "lon"=>75.45),
"/RU/URFO/89/NDM" => array("type"=>"city", "pop"=>48200, "name_ru"=>"Надым", "name_en"=>"Nadym", "lang"=>"ru", "lat"=>65.5333333333333, "lon"=>72.5166666666667),
"/RU/URFO/89/NUR" => array("type"=>"city", "pop"=>118400, "name_ru"=>"Новый Уренгой", "name_en"=>"Novyy Urengoy", "lang"=>"ru", "lat"=>66.0833333333333, "lon"=>76.6833333333333),
"/RU/URFO/89/SLH" => array("type"=>"city", "pop"=>41600, "name_ru"=>"Салехард", "name_en"=>"Salehard", "lang"=>"ru", "lat"=>66.5333333333333, "lon"=>66.6333333333333),
"/RU/URFO/89/TSL" => array("type"=>"city", "pop"=>20100, "name_ru"=>"Тарко-Сале", "name_en"=>"Tarko-Sale", "lang"=>"ru", "lat"=>64.9166666666667, "lon"=>77.7833333333333),
"/RU/YUFO" => array("type"=>"federal region", "pop"=>22901524, "name_ru"=>"Южный федеральный округ", "name_en"=>"Yuzhnyy Federal Region", "lang"=>"ru"),
"/RU/YUFO/01" => array("type"=>"region", "pop"=>441200, "name_ru"=>"Республика Адыгея", "name_en"=>"Adygeya Republic", "lang"=>"ru"),
"/RU/YUFO/01/MAY" => array("type"=>"city", "pop"=>153500, "name_ru"=>"Майкоп", "name_en"=>"Maykop", "lang"=>"ru", "lat"=>44.6, "lon"=>40.0833333333333),
"/RU/YUFO/05" => array("type"=>"region", "pop"=>2687800, "name_ru"=>"Республика Дагестан", "name_en"=>"Dagestan Republic", "lang"=>"ru"),
"/RU/YUFO/05/DER" => array("type"=>"city", "pop"=>109000, "name_ru"=>"Дербент", "name_en"=>"Derbent", "lang"=>"ru", "lat"=>42.0666666666667, "lon"=>48.3),
"/RU/YUFO/05/HAS" => array("type"=>"city", "pop"=>127100, "name_ru"=>"Хасавюрт", "name_en"=>"Hasavjurt", "lang"=>"ru", "lat"=>43.25, "lon"=>46.5833333333333),
"/RU/YUFO/05/KAS" => array("type"=>"city", "pop"=>82700, "name_ru"=>"Каспийск", "name_en"=>"Kaspijsk", "lang"=>"ru", "lat"=>42.8833333333333, "lon"=>47.6333333333333),
"/RU/YUFO/05/MAH" => array("type"=>"city", "pop"=>464200, "name_ru"=>"Махачкала", "name_en"=>"Mahachkala", "lang"=>"ru", "lat"=>42.9833333333333, "lon"=>47.4833333333333),
"/RU/YUFO/06" => array("type"=>"region", "pop"=>499500, "name_ru"=>"Республика Ингушетия", "name_en"=>"Ingushetiya Republic", "lang"=>"ru"),
"/RU/YUFO/06/KAR" => array("type"=>"city", "pop"=>35100, "name_ru"=>"Карабулак", "name_en"=>"Karabulak", "lang"=>"ru", "lat"=>43.3, "lon"=>44.9),
"/RU/YUFO/06/MAG" => array("type"=>"city", "pop"=>400, "name_ru"=>"Магас", "name_en"=>"Magas", "lang"=>"ru", "lat"=>43.1666666666667, "lon"=>44.8166666666667),
"/RU/YUFO/06/MAL" => array("type"=>"city", "pop"=>44700, "name_ru"=>"Малгобек", "name_en"=>"Malgobek", "lang"=>"ru", "lat"=>43.5166666666667, "lon"=>44.5833333333333),
"/RU/YUFO/06/NAZ" => array("type"=>"city", "pop"=>134300, "name_ru"=>"Назрань", "name_en"=>"Nazran'", "lang"=>"ru", "lat"=>43.2166666666667, "lon"=>44.7666666666667),
"/RU/YUFO/07" => array("type"=>"region", "pop"=>891300, "name_ru"=>"Кабардино-Балкарская республика", "name_en"=>"Kabardino-Balkarskaya Republic", "lang"=>"ru"),
"/RU/YUFO/07/BKS" => array("type"=>"city", "pop"=>57800, "name_ru"=>"Баксан", "name_en"=>"Baksan", "lang"=>"ru", "lat"=>43.6833333333333, "lon"=>43.5333333333333),
"/RU/YUFO/07/NLC" => array("type"=>"city", "pop"=>269600, "name_ru"=>"Нальчик", "name_en"=>"Nal'chik", "lang"=>"ru", "lat"=>43.4833333333333, "lon"=>43.6166666666667),
"/RU/YUFO/07/PRO" => array("type"=>"city", "pop"=>60300, "name_ru"=>"Прохладный", "name_en"=>"Prohladnyy", "lang"=>"ru", "lat"=>43.75, "lon"=>44.0333333333333),
"/RU/YUFO/08" => array("type"=>"region", "pop"=>285500, "name_ru"=>"Республика Калмыкия", "name_en"=>"Kalmykija Republic", "lang"=>"ru"),
"/RU/YUFO/08/ELS" => array("type"=>"city", "pop"=>102700, "name_ru"=>"Элиста", "name_en"=>"Elista", "lang"=>"ru", "lat"=>46.3166666666667, "lon"=>44.2666666666667),
"/RU/YUFO/09" => array("type"=>"region", "pop"=>427400, "name_ru"=>"Карачаево-Черкесская республика", "name_en"=>"Karachaevo-Cherkesskaya Republic", "lang"=>"ru"),
"/RU/YUFO/09/CHR" => array("type"=>"city", "pop"=>116300, "name_ru"=>"Черкесск", "name_en"=>"Cherkessk", "lang"=>"ru", "lat"=>44.2166666666667, "lon"=>42.05),
"/RU/YUFO/15" => array("type"=>"region", "pop"=>702500, "name_ru"=>"Республика Северная Осетия-Алания", "name_en"=>"Severnaya Osetiya Republic", "lang"=>"ru"),
"/RU/YUFO/15/VLK" => array("type"=>"city", "pop"=>312800, "name_ru"=>"Владикавказ", "name_en"=>"Vladikavkaz", "lang"=>"ru", "lat"=>43.0166666666667, "lon"=>44.6833333333333),
"/RU/YUFO/20" => array("type"=>"region", "pop"=>1209000, "name_ru"=>"Чеченская республика", "name_en"=>"Chechen Republic", "lang"=>"ru"),
"/RU/YUFO/20/GRZ" => array("type"=>"city", "pop"=>226100, "name_ru"=>"Грозный", "name_en"=>"Groznyy", "lang"=>"ru", "lat"=>43.3166666666667, "lon"=>45.7),
"/RU/YUFO/20/URM" => array("type"=>"city", "pop"=>48700, "name_ru"=>"Урус-Мартан", "name_en"=>"Urus-Martan", "lang"=>"ru", "lat"=>43.1333333333333, "lon"=>45.55),
"/RU/YUFO/23" => array("type"=>"region", "pop"=>5121800, "name_ru"=>"Краснодарский край", "name_en"=>"krasnodarskiy Kray", "lang"=>"ru"),
"/RU/YUFO/23/ARM" => array("type"=>"city", "pop"=>189100, "name_ru"=>"Армавир", "name_en"=>"Armavir", "lang"=>"ru", "lat"=>45, "lon"=>41.1166666666667),
"/RU/YUFO/23/ESK" => array("type"=>"city", "pop"=>87400, "name_ru"=>"Ейск", "name_en"=>"Ejsk", "lang"=>"ru", "lat"=>46.7, "lon"=>38.2833333333333),
"/RU/YUFO/23/KRS" => array("type"=>"city", "pop"=>709700, "name_ru"=>"Краснодар", "name_en"=>"Krasnodar", "lang"=>"ru", "lat"=>45.0333333333333, "lon"=>38.9833333333333),
"/RU/YUFO/23/NVR" => array("type"=>"city", "pop"=>227900, "name_ru"=>"Новороссийск", "name_en"=>"Novorossijsk", "lang"=>"ru", "lat"=>44.7166666666667, "lon"=>37.7666666666667),
"/RU/YUFO/23/SCH" => array("type"=>"city", "pop"=>334300, "name_ru"=>"Сочи", "name_en"=>"Sochi", "lang"=>"ru", "lat"=>43.6, "lon"=>39.7166666666667),
"/RU/YUFO/26" => array("type"=>"region", "pop"=>2705100, "name_ru"=>"Ставропольский край", "name_en"=>"Stavropolskiy Kray", "lang"=>"ru"),
"/RU/YUFO/26/ESS" => array("type"=>"city", "pop"=>81700, "name_ru"=>"Ессентуки", "name_en"=>"Essentuki", "lang"=>"ru", "lat"=>44.0333333333333, "lon"=>42.85),
"/RU/YUFO/26/KSL" => array("type"=>"city", "pop"=>129200, "name_ru"=>"Кисловодск", "name_en"=>"Kislovodsk", "lang"=>"ru", "lat"=>43.9166666666667, "lon"=>42.7166666666667),
"/RU/YUFO/26/MNV" => array("type"=>"city", "pop"=>76700, "name_ru"=>"Минеральные Воды", "name_en"=>"Mineral'nye Vody", "lang"=>"ru", "lat"=>44.2166666666667, "lon"=>43.1333333333333),
"/RU/YUFO/26/NVM" => array("type"=>"city", "pop"=>129100, "name_ru"=>"Невинномысск", "name_en"=>"Nevinnomyssk", "lang"=>"ru", "lat"=>44.6333333333333, "lon"=>41.9333333333333),
"/RU/YUFO/26/PTG" => array("type"=>"city", "pop"=>140100, "name_ru"=>"Пятигорск", "name_en"=>"Pjztigorsk", "lang"=>"ru", "lat"=>44.0333333333333, "lon"=>43.05),
"/RU/YUFO/26/STV" => array("type"=>"city", "pop"=>363700, "name_ru"=>"Ставрополь", "name_en"=>"Stavropol'", "lang"=>"ru", "lat"=>45.05, "lon"=>41.9833333333333),
"/RU/YUFO/30" => array("type"=>"region", "pop"=>1000900, "name_ru"=>"Астраханская область", "name_en"=>"Astrahanskaya Region", "lang"=>"ru"),
"/RU/YUFO/30/AST" => array("type"=>"city", "pop"=>503100, "name_ru"=>"Астрахань", "name_en"=>"Astrahan'", "lang"=>"ru", "lat"=>46.3333333333333, "lon"=>48.0166666666667),
"/RU/YUFO/34" => array("type"=>"region", "pop"=>2608800, "name_ru"=>"Волгоградская область", "name_en"=>"Volgogradskaya Region", "lang"=>"ru"),
"/RU/YUFO/34/KMS" => array("type"=>"city", "pop"=>118500, "name_ru"=>"Камышин", "name_en"=>"Kamyshin", "lang"=>"ru", "lat"=>50.0833333333333, "lon"=>45.4),
"/RU/YUFO/34/VLG" => array("type"=>"city", "pop"=>983900, "name_ru"=>"Волгоград", "name_en"=>"Volgograd", "lang"=>"ru", "lat"=>48.7, "lon"=>44.4833333333333),
"/RU/YUFO/34/VLZ" => array("type"=>"city", "pop"=>306400, "name_ru"=>"Волжский", "name_en"=>"Volzhskiy", "lang"=>"ru", "lat"=>48.7833333333333, "lon"=>44.7666666666667),
"/RU/YUFO/61" => array("type"=>"region", "pop"=>4254400, "name_ru"=>"Ростовская область", "name_en"=>"Rostovskaya Region", "lang"=>"ru"),
"/RU/YUFO/61/NVC" => array("type"=>"city", "pop"=>177000, "name_ru"=>"Новочеркасск", "name_en"=>"Novocherkassk", "lang"=>"ru", "lat"=>47.4333333333333, "lon"=>40.0833333333333),
"/RU/YUFO/61/RND" => array("type"=>"city", "pop"=>1048700, "name_ru"=>"Ростов-на-Дону", "name_en"=>"Rostov-on-Don", "lang"=>"ru", "lat"=>47.2333333333333, "lon"=>39.7166666666667),
"/RU/YUFO/61/SHH" => array("type"=>"city", "pop"=>244400, "name_ru"=>"Шахты", "name_en"=>"Shahty", "lang"=>"ru", "lat"=>47.7, "lon"=>40.2333333333333),
"/RU/YUFO/61/TGN" => array("type"=>"city", "pop"=>260700, "name_ru"=>"Таганрог", "name_en"=>"Taganrog", "lang"=>"ru", "lat"=>47.2333333333333, "lon"=>38.9),
"/RU/YUFO/61/VGD" => array("type"=>"city", "pop"=>169600, "name_ru"=>"Волгодонск", "name_en"=>"Volgodonsk", "lang"=>"ru", "lat"=>47.5166666666667, "lon"=>42.15),
"/RW" => array("type"=>"country", "pop"=>7600000, "name_ru"=>"Руанда", "name_en"=>"Rwanda", "lang"=>"fr", "lat"=>-1.98, "lon"=>29.71),
"/SA" => array("type"=>"country", "pop"=>27019731, "name_ru"=>"Саудовская Аравия", "name_en"=>"Saudi Arabia", "lang"=>"ar", "lat"=>23.71, "lon"=>44.11),
"/SB" => array("type"=>"country", "pop"=>478000, "name_ru"=>"Соломоновы Острова", "name_en"=>"Solomon Islands", "lang"=>"en", "lat"=>-8.00, "lon"=>159.18),
"/SC" => array("type"=>"country", "pop"=>80654, "name_ru"=>"Сейшельские Острова", "name_en"=>"Seychelles", "lang"=>"fr", "lat"=>-7.1, "lon"=>52.76),
"/SD" => array("type"=>"country", "pop"=>38114160, "name_ru"=>"Судан", "name_en"=>"Sudan", "lang"=>"ar", "lat"=>15.5, "lon"=>30.56),
"/SE" => array("type"=>"country", "pop"=>9263872, "name_ru"=>"Швеция", "name_sv"=>"Sverige", "name_en"=>"Sweden", "lang"=>"sv", "lat"=>61.31, "lon"=>14.83),
"/SG" => array("type"=>"country", "pop"=>4839400, "name_cn"=>"新加坡共和国", "name_ru"=>"Сингапур", "name_en"=>"Singapore", "lang"=>"cn", "lat"=>1.29, "lon"=>103.85),
"/SI" => array("type"=>"country", "pop"=>2009245, "name_ru"=>"Словения", "name_sl"=>"Slovenija", "name_en"=>"Slovenia", "lang"=>"sl", "lat"=>46.15, "lon"=>15.05),
"/SK" => array("type"=>"country", "pop"=>5431363, "name_ru"=>"Словакия", "name_sk"=>"Slovensko", "name_en"=>"Slovakia", "lang"=>"sk", "lat"=>48.65, "lon"=>19.55),
"/SL" => array("type"=>"country", "pop"=>5426618, "name_ru"=>"Сьерра-Леоне", "name_en"=>"Sierra Leone", "lang"=>"en", "lat"=>8.5, "lon"=>-11.91),
"/SM" => array("type"=>"country", "pop"=>28119, "name_ru"=>"Сан-Марино", "name_en"=>"San Marino", "lang"=>"it", "lat"=>43.93, "lon"=>12.44),
"/SN" => array("type"=>"country", "pop"=>10900000, "name_ru"=>"Сенегал", "name_en"=>"Senegal", "lang"=>"fr", "lat"=>14.36, "lon"=>-14.28),
"/SO" => array("type"=>"country", "pop"=>17700000, "name_ru"=>"Сомали", "name_en"=>"Somalia", "lang"=>"ar", "lat"=>8.31, "lon"=>47.3),
"/SR" => array("type"=>"country", "pop"=>438144, "name_ru"=>"Суринам", "name_en"=>"Suriname", "lang"=>"nl", "lat"=>4.3, "lon"=>-56),
"/ST" => array("type"=>"country", "pop"=>151912, "name_ru"=>"Сан-Томе и Принсипи", "name_en"=>"São Tomé and Príncipe", "lang"=>"pt", "lat"=>-0.31, "lon"=>6.6),
"/SV" => array("type"=>"country", "pop"=>6470379, "name_ru"=>"Сальвадор", "name_en"=>"El Salvador", "lang"=>"es", "lat"=>13.8, "lon"=>-88.9),
"/SY" => array("type"=>"country", "pop"=>19405000, "name_ru"=>"Сирия", "name_en"=>"Syrian", "lang"=>"ar", "lat"=>35.21, "lon"=>38.58),
"/SZ" => array("type"=>"country", "pop"=>1104343, "name_ru"=>"Свазиленд", "name_en"=>"Swaziland", "lang"=>"en", "lat"=>-26.48, "lon"=>31.43),
"/TC" => array("type"=>"country", "pop"=>21746, "name_ru"=>"Тёркс и Кайкос", "name_en"=>"Turks and Caicos Islands", "lang"=>"en", "lat"=>21.4602777777778, "lon"=>-71.1413888888889),
"/TD" => array("type"=>"country", "pop"=>9749000, "name_ru"=>"Чад", "name_en"=>"Chad", "lang"=>"ar", "lat"=>15.46, "lon"=>19.4),
"/TG" => array("type"=>"country", "pop"=>6300000, "name_ru"=>"Того", "name_en"=>"Togo", "lang"=>"fr", "lat"=>8.25, "lon"=>1.18),
"/TH" => array("type"=>"country", "pop"=>65444371, "name_ru"=>"Таиланд", "name_th"=>"ประเทศไทย", "name_en"=>"Thailand", "lang"=>"th", "lat"=>15.35, "lon"=>101.03),
"/TJ" => array("type"=>"country", "pop"=>7215000, "name_ru"=>"Таджикистан", "name_tg"=>"Тоҷикистон", "name_en"=>"Tajikistan", "lang"=>"tg", "lat"=>38.58, "lon"=>71.36),
"/TL" => array("type"=>"country", "pop"=>1040880, "name_ru"=>"Восточный Тимор", "name_en"=>"Timor-Leste", "lang"=>"pt", "lat"=>-8.76, "lon"=>126.09),
"/TM" => array("type"=>"country", "pop"=>4884887, "name_ru"=>"Туркмения", "name_tk"=>"Türkmenistan", "name_en"=>"Turkmenistan", "lang"=>"tk", "lat"=>39.33, "lon"=>59.65),
"/TN" => array("type"=>"country", "pop"=>10383577, "name_ru"=>"Тунис", "name_en"=>"Tunisia", "lang"=>"ar", "lat"=>33.31, "lon"=>9.35),
"/TO" => array("type"=>"country", "pop"=>116973, "name_ru"=>"Тонга", "name_en"=>"Tonga", "lang"=>"to", "lat"=>-20.58, "lon"=>-174.81),
"/TR" => array("type"=>"country", "pop"=>78110800, "name_ru"=>"Турция", "name_tr"=>"Türkiye", "name_en"=>"Turkey", "lang"=>"tr", "lat"=>39.16, "lon"=>35.66),
"/TT" => array("type"=>"country", "pop"=>1323000, "name_ru"=>"Тринидад и Тобаго", "name_en"=>"Trinidad and Tobago", "lang"=>"en", "lat"=>11.00, "lon"=>-61.00),
"/TV" => array("type"=>"country", "pop"=>12177, "name_ru"=>"Тувалу", "name_en"=>"Tuvalu", "lang"=>"en", "lat"=>-7.47, "lon"=>178.005),
"/TW" => array("type"=>"country", "pop"=>23046177, "name_cn"=>"台灣", "name_ru"=>"Тайвань", "name_en"=>"Taiwan", "lang"=>"cn", "lat"=>23.7666666666667, "lon"=>121),
"/TZ" => array("type"=>"country", "pop"=>37849133, "name_ru"=>"Танзания", "name_en"=>"Tanzania", "lang"=>"sw", "lat"=>-6.7, "lon"=>33.93),
"/UA" => array("type"=>"country", "pop"=>46022306, "name_uk"=>"Україна", "name_ru"=>"Украина", "name_en"=>"Ukraine", "lang"=>"uk", "lat"=>49.8, "lon"=>30.83),
"/UA/AB" => array("type"=>"region", "pop"=>1701000, "name_uk"=>"Вінницька область", "name_ru"=>"Винницкая область", "name_en"=>"Vinnytsia Oblast", "lang"=>"uk"),
"/UA/AB/MGPD" => array("type"=>"city", "pop"=>36000, "name_uk"=>"Могилів-Подільський", "name_ru"=>"Могилёв-Подольский", "name_en"=>"Mohyliv-Podilskyi", "lang"=>"uk", "lat"=>48.45, "lon"=>27.7833333333333),
"/UA/AB/VNNC" => array("type"=>"city", "pop"=>362500, "name_uk"=>"Вінниця", "name_ru"=>"Винница", "name_en"=>"Vinnytsia", "lang"=>"uk", "lat"=>49.2333333333333, "lon"=>28.4833333333333),
"/UA/AB/ZHMR" => array("type"=>"city", "pop"=>40300, "name_uk"=>"Жмеринка", "name_ru"=>"Жмеринка", "name_en"=>"Zhmerynka", "lang"=>"uk", "lat"=>49.0333333333333, "lon"=>28.1166666666667),
"/UA/AC" => array("type"=>"region", "pop"=>1040000, "name_uk"=>"Волинська область", "name_ru"=>"Волынская область", "name_en"=>"Volyn Oblast", "lang"=>"uk"),
"/UA/AC/KOVL" => array("type"=>"city", "pop"=>68000, "name_uk"=>"Ковель", "name_ru"=>"Ковель", "name_en"=>"Kovel", "lang"=>"uk", "lat"=>51.2166666666667, "lon"=>24.7166666666667),
"/UA/AC/LUCK" => array("type"=>"city", "pop"=>209800, "name_uk"=>"Луцьк", "name_ru"=>"Луцк", "name_en"=>"Lutsk", "lang"=>"uk", "lat"=>50.75, "lon"=>25.3333333333333),
"/UA/AC/NVVL" => array("type"=>"city", "pop"=>57800, "name_uk"=>"Нововолинськ", "name_ru"=>"Нововолынск", "name_en"=>"Novovolynsk", "lang"=>"uk", "lat"=>50.7333333333333, "lon"=>24.1666666666667),
"/UA/AC/VLVL" => array("type"=>"city", "pop"=>38400, "name_uk"=>"Володимир-Волинський", "name_ru"=>"Владимир-Волынский", "name_en"=>"Volodymyr-Volynskyi", "lang"=>"uk", "lat"=>50.8480555555556, "lon"=>24.3222222222222),
"/UA/AE" => array("type"=>"region", "pop"=>3366919, "name_uk"=>"Дніпропетровська область", "name_ru"=>"Днепропетровская область", "name_en"=>"Dnipropetrovsk Oblast", "lang"=>"uk"),
"/UA/AE/DNDZ" => array("type"=>"city", "pop"=>278000, "name_uk"=>"Дніпродзержинськ", "name_ru"=>"Днепродзержинск", "name_en"=>"Dniprodzerzhynsk", "lang"=>"uk", "lat"=>48.5, "lon"=>34.6166666666667),
"/UA/AE/DNPT" => array("type"=>"city", "pop"=>1013006, "name_uk"=>"Дніпропетровськ", "name_ru"=>"Днепропетровск", "name_en"=>"Dnipropetrovsk", "lang"=>"uk", "lat"=>48.4647222222222, "lon"=>35.0461111111111),
"/UA/AE/KRVR" => array("type"=>"city", "pop"=>684700, "name_uk"=>"Кривий Ріг", "name_ru"=>"Кривой Рог", "name_en"=>"Kryvyi Rih", "lang"=>"uk", "lat"=>47.9102777777778, "lon"=>33.3916666666667),
"/UA/AE/MRGN" => array("type"=>"city", "pop"=>50000, "name_uk"=>"Марганець", "name_ru"=>"Марганец", "name_en"=>"Marhanets", "lang"=>"uk", "lat"=>47.6397222222222, "lon"=>34.6363888888889),
"/UA/AE/NKPL" => array("type"=>"city", "pop"=>127336, "name_uk"=>"Нікополь", "name_ru"=>"Никополь", "name_en"=>"Nikopol", "lang"=>"uk", "lat"=>47.5680555555556, "lon"=>34.3911111111111),
"/UA/AE/NVMS" => array("type"=>"city", "pop"=>75000, "name_uk"=>"Новомосковськ", "name_ru"=>"Новомосковск", "name_en"=>"Novomoskovsk", "lang"=>"uk", "lat"=>48.6333333333333, "lon"=>35.2166666666667),
"/UA/AE/PVLG" => array("type"=>"city", "pop"=>112000, "name_uk"=>"Павлоград", "name_ru"=>"Павлоград", "name_en"=>"Pavlohrad", "lang"=>"uk", "lat"=>48.52, "lon"=>35.87),
"/UA/AH" => array("type"=>"region", "pop"=>4500000, "name_uk"=>"Донецька область", "name_ru"=>"Донецкая область", "name_en"=>"Donetsk Oblast", "lang"=>"uk"),
"/UA/AH/DNCK" => array("type"=>"city", "pop"=>985877, "name_uk"=>"Донецьк", "name_ru"=>"Донецк", "name_en"=>"Donetsk", "lang"=>"uk", "lat"=>48, "lon"=>37.8),
"/UA/AH/ENKV" => array("type"=>"city", "pop"=>104000, "name_uk"=>"Єнакієве", "name_ru"=>"Енакиево", "name_en"=>"Yenakiieve", "lang"=>"uk", "lat"=>48.2166666666667, "lon"=>38.2),
"/UA/AH/GRLV" => array("type"=>"city", "pop"=>292300, "name_uk"=>"Го́рлівка", "name_ru"=>"Горловка", "name_en"=>"Horlivka", "lang"=>"uk", "lat"=>48.3, "lon"=>38.05),
"/UA/AH/HRCZ" => array("type"=>"city", "pop"=>64200, "name_uk"=>"Харци́зьк", "name_ru"=>"Харцы́зск", "name_en"=>"Khartsyzk", "lang"=>"uk", "lat"=>48.0333333333333, "lon"=>38.15),
"/UA/AH/KRMT" => array("type"=>"city", "pop"=>169100, "name_uk"=>"Краматорськ", "name_ru"=>"Краматорск", "name_en"=>"Kramatorsk", "lang"=>"uk", "lat"=>48.7402777777778, "lon"=>37.5755555555556),
"/UA/AH/MKVK" => array("type"=>"city", "pop"=>389600, "name_uk"=>"Макі́ївка", "name_ru"=>"Макеевка", "name_en"=>"Makiivka", "lang"=>"uk", "lat"=>48.0416666666667, "lon"=>37.9638888888889),
"/UA/AH/MRPL" => array("type"=>"city", "pop"=>493245, "name_uk"=>"Марiуполь", "name_ru"=>"Мариуполь", "name_en"=>"Mariupol", "lang"=>"uk", "lat"=>47.1305555555556, "lon"=>37.5638888888889),
"/UA/AH/SLVN" => array("type"=>"city", "pop"=>124800, "name_uk"=>"Слов’я́нськ", "name_ru"=>"Славянск", "name_en"=>"Sloviansk", "lang"=>"uk", "lat"=>48.85, "lon"=>37.6),
"/UA/AI" => array("type"=>"region", "pop"=>1763000, "name_uk"=>"Київська область", "name_ru"=>"Киевская область", "name_en"=>"Kiev Oblast", "lang"=>"uk"),
"/UA/AI/BLCR" => array("type"=>"city", "pop"=>215000, "name_uk"=>"Біла Церква", "name_ru"=>"Белая Церковь", "name_en"=>"Bila Tserkva", "lang"=>"uk", "lat"=>49.7833333333333, "lon"=>30.1166666666667),
"/UA/AI/BRSP" => array("type"=>"city", "pop"=>55000, "name_uk"=>"Бориспiль", "name_ru"=>"Борисполь", "name_en"=>"Boryspil", "lang"=>"uk", "lat"=>50.35, "lon"=>30.95),
"/UA/AI/BRVR" => array("type"=>"city", "pop"=>86165, "name_uk"=>"Бровари", "name_ru"=>"Бровары", "name_en"=>"Brovary", "lang"=>"uk", "lat"=>50.5, "lon"=>30.7666666666667),
"/UA/AI/FSTV" => array("type"=>"city", "pop"=>50500, "name_uk"=>"Фастiв", "name_ru"=>"Фастов", "name_en"=>"Fastiv", "lang"=>"uk", "lat"=>50.0747222222222, "lon"=>29.9180555555556),
"/UA/AI/IRPN" => array("type"=>"city", "pop"=>41762, "name_uk"=>"Ірпінь", "name_ru"=>"Ирпень", "name_en"=>"Irpin", "lang"=>"uk", "lat"=>50.5166666666667, "lon"=>30.25),
"/UA/AI/SLVT" => array("type"=>"city", "pop"=>27833, "name_uk"=>"Славутич", "name_ru"=>"Славутич", "name_en"=>"Slavutych", "lang"=>"uk", "lat"=>51.5205555555556, "lon"=>30.7569444444444),
"/UA/AI/VSLK" => array("type"=>"city", "pop"=>39700, "name_uk"=>"Василькiв", "name_ru"=>"Васильков", "name_en"=>"Vasylkiv", "lang"=>"uk", "lat"=>50.1775, "lon"=>30.3216666666667),
"/UA/AK" => array("type"=>"region", "pop"=>1994300, "name_uk"=>"Автономна Республіка Крим", "name_ru"=>"Автономная республика Крым", "name_en"=>"Autonomous Republic of Crimea", "lang"=>"uk"),
"/UA/AK/EVPT" => array("type"=>"city", "pop"=>119600, "name_uk"=>"Євпаторія", "name_ru"=>"Евпатория", "name_en"=>"Yevpatoria", "lang"=>"uk", "lat"=>45.2036111111111, "lon"=>33.3613888888889),
"/UA/AK/FEOD" => array("type"=>"city", "pop"=>71700, "name_uk"=>"Феодосія", "name_ru"=>"Феодосия", "name_en"=>"Feodosiya", "lang"=>"uk", "lat"=>45.0488888888889, "lon"=>35.3791666666667),
"/UA/AK/KRCH" => array("type"=>"city", "pop"=>151300, "name_uk"=>"Керч", "name_ru"=>"Керчь", "name_en"=>"Kerch", "lang"=>"uk", "lat"=>45.3386111111111, "lon"=>36.4680555555556),
"/UA/AK/SMFR" => array("type"=>"city", "pop"=>360500, "name_uk"=>"Сімферополь", "name_ru"=>"Симферополь", "name_en"=>"Simferopol", "lang"=>"uk", "lat"=>44.9480555555556, "lon"=>34.1041666666667),
"/UA/AK/YALT" => array("type"=>"city", "pop"=>102321, "name_uk"=>"Ялта", "name_ru"=>"Ялта", "name_en"=>"Yalta", "lang"=>"uk", "lat"=>44.4994444444444, "lon"=>34.1552777777778),
"/UA/AM" => array("type"=>"region", "pop"=>1330000, "name_uk"=>"Житомирська область", "name_ru"=>"Житомирская область", "name_en"=>"Zhytomir Oblast", "lang"=>"uk"),
"/UA/AM/BRDC" => array("type"=>"city", "pop"=>86200, "name_uk"=>"Бердичiв", "name_ru"=>"Бердичев", "name_en"=>"Berdychiv", "lang"=>"uk", "lat"=>49.9, "lon"=>28.5666666666667),
"/UA/AM/KRST" => array("type"=>"city", "pop"=>67000, "name_uk"=>"Коростень", "name_ru"=>"Коростень", "name_en"=>"Korosten", "lang"=>"uk", "lat"=>50.95, "lon"=>28.65),
"/UA/AM/ZHTM" => array("type"=>"city", "pop"=>271348, "name_uk"=>"Житомир", "name_ru"=>"Житомир", "name_en"=>"Zhytomir", "lang"=>"uk", "lat"=>50.25, "lon"=>28.6666666666667),
"/UA/AO" => array("type"=>"region", "pop"=>1245000, "name_uk"=>"Закарпатська область", "name_ru"=>"Закарпатская область", "name_en"=>"Zakarpattia Oblast", "lang"=>"uk"),
"/UA/AO/HUST" => array("type"=>"city", "pop"=>31804, "name_uk"=>"Хуст", "name_ru"=>"Хуст", "name_en"=>"Khust", "lang"=>"uk", "lat"=>48.1813888888889, "lon"=>23.2977777777778),
"/UA/AO/MKCH" => array("type"=>"city", "pop"=>93738, "name_uk"=>"Мукачеве", "name_ru"=>"Мукачево", "name_en"=>"Mukachevo", "lang"=>"uk", "lat"=>48.4413888888889, "lon"=>22.7136111111111),
"/UA/AO/UZHG" => array("type"=>"city", "pop"=>117600, "name_uk"=>"Ужгород", "name_ru"=>"Ужгород", "name_en"=>"Uzhhorod", "lang"=>"uk", "lat"=>48.62, "lon"=>22.3),
"/UA/AO/VNGR" => array("type"=>"city", "pop"=>25760, "name_uk"=>"Виноградів", "name_ru"=>"Виноградов", "name_en"=>"Vynohradiv", "lang"=>"uk", "lat"=>48.15, "lon"=>22.7666666666667),
"/UA/AP" => array("type"=>"region", "pop"=>1861000, "name_uk"=>"Запорізька область", "name_ru"=>"Запорожская область", "name_en"=>"Zaporizhia Oblast", "lang"=>"uk"),
"/UA/AP/BRDN" => array("type"=>"city", "pop"=>112000, "name_uk"=>"Бердянськ", "name_ru"=>"Бердянск", "name_en"=>"Berdyansk", "lang"=>"uk", "lat"=>46.75, "lon"=>36.7833333333333),
"/UA/AP/MLTP" => array("type"=>"city", "pop"=>158400, "name_uk"=>"Мелітополь", "name_ru"=>"Мелитополь", "name_en"=>"Melitopol", "lang"=>"uk", "lat"=>46.8333333333333, "lon"=>35.3666666666667),
"/UA/AP/ZPRZ" => array("type"=>"city", "pop"=>781100, "name_uk"=>"Запоріжжя", "name_ru"=>"Запорожье", "name_en"=>"Zaporizhia", "lang"=>"uk", "lat"=>47.8333333333333, "lon"=>35.1666666666667),
"/UA/AT" => array("type"=>"region", "pop"=>1388000, "name_uk"=>"Івано-Франківська область", "name_ru"=>"Ивано-Франковская область", "name_en"=>"Ivano-Frankivsk Oblast", "lang"=>"uk"),
"/UA/AT/IVFR" => array("type"=>"city", "pop"=>238000, "name_uk"=>"Івано-Франківськ", "name_ru"=>"Ивано-Франковск", "name_en"=>"Ivano-Frankivsk", "lang"=>"uk", "lat"=>48.9166666666667, "lon"=>24.7166666666667),
"/UA/AT/KLMY" => array("type"=>"city", "pop"=>60000, "name_uk"=>"Коломия", "name_ru"=>"Коломыя", "name_en"=>"Kolomiya", "lang"=>"uk", "lat"=>48.5166666666667, "lon"=>25.0333333333333),
"/UA/AT/KLSH" => array("type"=>"city", "pop"=>67887, "name_uk"=>"Калуш", "name_ru"=>"Калуш", "name_en"=>"Kalush", "lang"=>"uk", "lat"=>49.0166666666667, "lon"=>24.3666666666667),
"/UA/AX" => array("type"=>"region", "pop"=>2829000, "name_uk"=>"Харківська область", "name_ru"=>"Харьковская область", "name_en"=>"Kharkiv Oblast", "lang"=>"uk"),
"/UA/AX/CHGV" => array("type"=>"city", "pop"=>37800, "name_uk"=>"Чугуїв", "name_ru"=>"Чугуев", "name_en"=>"Chuhuiv", "lang"=>"uk", "lat"=>49.8355555555556, "lon"=>36.6863888888889),
"/UA/AX/IZUM" => array("type"=>"city", "pop"=>53708, "name_uk"=>"Iзюм", "name_ru"=>"Изюм", "name_en"=>"Izium", "lang"=>"uk", "lat"=>49.2127777777778, "lon"=>37.2569444444444),
"/UA/AX/KHRK" => array("type"=>"city", "pop"=>1440000, "name_uk"=>"Харкiв", "name_ru"=>"Харьков", "name_en"=>"Kharkiv", "lang"=>"uk", "lat"=>50, "lon"=>36.25),
"/UA/AX/KPNS" => array("type"=>"city", "pop"=>30407, "name_uk"=>"Куп'янськ", "name_ru"=>"Купянск", "name_en"=>"Kupiansk", "lang"=>"uk", "lat"=>49.7166666666667, "lon"=>37.5833333333333),
"/UA/AX/LZV" => array("type"=>"city", "pop"=>71500, "name_uk"=>"Лозова", "name_ru"=>"Лозовая", "name_en"=>"Lozova", "lang"=>"uk", "lat"=>48.8833333333333, "lon"=>36.3833333333333),
"/UA/AX/PRVM" => array("type"=>"city", "pop"=>33319, "name_uk"=>"Первомайський", "name_ru"=>"Первомайский", "name_en"=>"Pervomaiskyi", "lang"=>"uk", "lat"=>49.3833333333333, "lon"=>36.2166666666667),
"/UA/BA" => array("type"=>"region", "pop"=>1067000, "name_uk"=>"Кiровоградська область", "name_ru"=>"Кировоградская область", "name_en"=>"Kirovohrad Oblast", "lang"=>"uk"),
"/UA/BA/ALXN" => array("type"=>"city", "pop"=>92638, "name_uk"=>"Олександрiя", "name_ru"=>"Александрия", "name_en"=>"Oleksandriia", "lang"=>"uk", "lat"=>48.6666666666667, "lon"=>33.1166666666667),
"/UA/BA/KRVG" => array("type"=>"city", "pop"=>258543, "name_uk"=>"Кiровоград", "name_ru"=>"Кировоград", "name_en"=>"Kirovohrad", "lang"=>"uk", "lat"=>48.5, "lon"=>32.2666666666667),
"/UA/BA/SVTL" => array("type"=>"city", "pop"=>49938, "name_uk"=>"Свiтроводськ", "name_ru"=>"Светловодск", "name_en"=>"Svitlovodsk", "lang"=>"uk", "lat"=>49.0833333333333, "lon"=>33.25),
"/UA/BB" => array("type"=>"region", "pop"=>2580600, "name_uk"=>"Луганська область", "name_ru"=>"Луганская область", "name_en"=>"Luhansk Oblast", "lang"=>"uk"),
"/UA/BB/ALCH" => array("type"=>"city", "pop"=>114700, "name_uk"=>"Алчевськ", "name_ru"=>"Алчевск", "name_en"=>"Alchevsk", "lang"=>"uk", "lat"=>48.4666666666667, "lon"=>38.8),
"/UA/BB/ANTR" => array("type"=>"city", "pop"=>81000, "name_uk"=>"Антрацит", "name_ru"=>"Антрацит", "name_en"=>"Antratsyt", "lang"=>"uk", "lat"=>48.1166666666667, "lon"=>39.0833333333333),
"/UA/BB/KRLC" => array("type"=>"city", "pop"=>96400, "name_uk"=>"Красний Луч", "name_ru"=>"Красный Луч", "name_en"=>"Krasnyi Luch", "lang"=>"uk", "lat"=>48.1213888888889, "lon"=>38.9452777777778),
"/UA/BB/LGNS" => array("type"=>"city", "pop"=>442000, "name_uk"=>"Луганськ", "name_ru"=>"Луганск", "name_en"=>"Luhansk", "lang"=>"uk", "lat"=>48.5833333333333, "lon"=>39.3333333333333),
"/UA/BB/LSCH" => array("type"=>"city", "pop"=>122000, "name_uk"=>"Лисичанськ", "name_ru"=>"Лисичанск", "name_en"=>"Lysychansk", "lang"=>"uk", "lat"=>49, "lon"=>38.3833333333333),
"/UA/BB/STHN" => array("type"=>"city", "pop"=>84325, "name_uk"=>"Стаханов", "name_ru"=>"Стаханов", "name_en"=>"Stakhanov", "lang"=>"uk", "lat"=>48.5472222222222, "lon"=>38.6361111111111),
"/UA/BB/SVRD" => array("type"=>"city", "pop"=>78462, "name_uk"=>"Свердловськ", "name_ru"=>"Свердловск", "name_en"=>"Sverdlovsk", "lang"=>"uk", "lat"=>48.85, "lon"=>39.65),
"/UA/BC" => array("type"=>"region", "pop"=>2577000, "name_uk"=>"Львівська область", "name_ru"=>"Львовская область", "name_en"=>"Lviv Oblast", "lang"=>"uk"),
"/UA/BC/DRGB" => array("type"=>"city", "pop"=>77049, "name_uk"=>"Дрогобич", "name_ru"=>"Дрогобыч", "name_en"=>"Drohobych", "lang"=>"uk", "lat"=>49.35, "lon"=>23.5),
"/UA/BC/LVIV" => array("type"=>"city", "pop"=>752300, "name_uk"=>"Львiв", "name_ru"=>"Львов", "name_en"=>"Lviv", "lang"=>"uk", "lat"=>49.8333333333333, "lon"=>24),
"/UA/BC/STRY" => array("type"=>"city", "pop"=>71000, "name_uk"=>"Стрий", "name_ru"=>"Стрый", "name_en"=>"Stryi", "lang"=>"uk", "lat"=>49.25, "lon"=>23.85),
"/UA/BC/TRSK" => array("type"=>"city", "pop"=>30000, "name_uk"=>"Трускавець", "name_ru"=>"Трускавец", "name_en"=>"Truskavets", "lang"=>"uk", "lat"=>49.2805555555556, "lon"=>23.505),
"/UA/BE" => array("type"=>"region", "pop"=>1219000, "name_uk"=>"Миколаївська область", "name_ru"=>"Николаевская область", "name_en"=>"Mykolaiv Oblast", "lang"=>"uk"),
"/UA/BE/MYKL" => array("type"=>"city", "pop"=>506381, "name_uk"=>"Миколаїв", "name_ru"=>"Николаев", "name_en"=>"Mykolaiv", "lang"=>"uk", "lat"=>46.9666666666667, "lon"=>32),
"/UA/BE/PRVM" => array("type"=>"city", "pop"=>70170, "name_uk"=>"Первомайськ", "name_ru"=>"Первомайск", "name_en"=>"Pervomaisk", "lang"=>"uk", "lat"=>48.05, "lon"=>30.85),
"/UA/BE/VZNS" => array("type"=>"city", "pop"=>42248, "name_uk"=>"Вознесенськ", "name_ru"=>"Вознесенск", "name_en"=>"Voznesensk", "lang"=>"uk", "lat"=>47.5725, "lon"=>31.3119444444444),
"/UA/BE/YZHN" => array("type"=>"city", "pop"=>39430, "name_uk"=>"Южноукраїнськ", "name_ru"=>"Южноукраинск", "name_en"=>"Yuzhnoukrainsk", "lang"=>"uk", "lat"=>47.8216666666667, "lon"=>31.175),
"/UA/BH" => array("type"=>"region", "pop"=>2402000, "name_uk"=>"Одеська область", "name_ru"=>"Одесская область", "name_en"=>"Odessa Oblast", "lang"=>"uk"),
"/UA/BH/BLDN" => array("type"=>"city", "pop"=>51000, "name_uk"=>"Білгород-Дністровський", "name_ru"=>"Белгород-Днестровский", "name_en"=>"Bilhorod-Dnistrovskyi", "lang"=>"uk", "lat"=>46.1833333333333, "lon"=>30.3333333333333),
"/UA/BH/ILCH" => array("type"=>"city", "pop"=>70000, "name_uk"=>"Іллічівськ", "name_ru"=>"Ильичёвск", "name_en"=>"Illichivsk", "lang"=>"uk", "lat"=>46.3, "lon"=>30.65),
"/UA/BH/IZML" => array("type"=>"city", "pop"=>77076, "name_uk"=>"Ізмаїл", "name_ru"=>"Измаил", "name_en"=>"Izmail", "lang"=>"uk", "lat"=>45.35, "lon"=>28.8333333333333),
"/UA/BH/KTVS" => array("type"=>"city", "pop"=>46664, "name_uk"=>"Котовськ", "name_ru"=>"Котовск", "name_en"=>"Kotovsk", "lang"=>"uk", "lat"=>47.7419444444444, "lon"=>29.535),
"/UA/BH/ODES" => array("type"=>"city", "pop"=>1008600, "name_uk"=>"Одеса", "name_ru"=>"Одесса", "name_en"=>"Odessa", "lang"=>"uk", "lat"=>46.4666666666667, "lon"=>30.7333333333333),
"/UA/BI" => array("type"=>"region", "pop"=>1554000, "name_uk"=>"Полтавська область", "name_ru"=>"Полтавская область", "name_en"=>"Poltava Oblast", "lang"=>"uk"),
"/UA/BI/KMSM" => array("type"=>"city", "pop"=>54414, "name_uk"=>"Комсомольськ", "name_ru"=>"Комсомольск", "name_en"=>"Komsomolsk", "lang"=>"uk", "lat"=>49.0166666666667, "lon"=>33.6666666666667),
"/UA/BI/KRMN" => array("type"=>"city", "pop"=>228566, "name_uk"=>"Кременчук", "name_ru"=>"Кременчуг", "name_en"=>"Kremenchuk", "lang"=>"uk", "lat"=>49.0666666666667, "lon"=>33.4166666666667),
"/UA/BI/LBNY" => array("type"=>"city", "pop"=>48725, "name_uk"=>"Лубни", "name_ru"=>"Лубны", "name_en"=>"Lubny", "lang"=>"uk", "lat"=>50, "lon"=>33.0166666666667),
"/UA/BI/MRGR" => array("type"=>"city", "pop"=>41400, "name_uk"=>"Миргород", "name_ru"=>"Миргород", "name_en"=>"Mirgorod", "lang"=>"uk", "lat"=>49.9666666666667, "lon"=>33.6),
"/UA/BI/PLTV" => array("type"=>"city", "pop"=>318000, "name_uk"=>"Полтава", "name_ru"=>"Полтава", "name_en"=>"Poltava", "lang"=>"uk", "lat"=>49.5833333333333, "lon"=>34.55),
"/UA/BK" => array("type"=>"region", "pop"=>1156000, "name_uk"=>"Рiвенська область", "name_ru"=>"Ровенская область", "name_en"=>"Rivne Oblast", "lang"=>"uk"),
"/UA/BK/DBN" => array("type"=>"city", "pop"=>38026, "name_uk"=>"Дубно", "name_ru"=>"Дубно", "name_en"=>"Dubno", "lang"=>"uk", "lat"=>50.4166666666667, "lon"=>25.75),
"/UA/BK/KSTP" => array("type"=>"city", "pop"=>30500, "name_uk"=>"Костопiль", "name_ru"=>"Костополь", "name_en"=>"Kostopil", "lang"=>"uk", "lat"=>50.8833333333333, "lon"=>26.45),
"/UA/BK/KZNC" => array("type"=>"city", "pop"=>40500, "name_uk"=>"Кузнецовськ", "name_ru"=>"Кузнецовск", "name_en"=>"Kuznetsovsk", "lang"=>"uk", "lat"=>51.3472222222222, "lon"=>25.8486111111111),
"/UA/BK/RVN" => array("type"=>"city", "pop"=>249000, "name_uk"=>"Рiвне", "name_ru"=>"Ровно", "name_en"=>"Rivne", "lang"=>"uk", "lat"=>50.6166666666667, "lon"=>26.25),
"/UA/BM" => array("type"=>"region", "pop"=>1226000, "name_uk"=>"Сумська область", "name_ru"=>"Сумская область", "name_en"=>"Sumy Oblast", "lang"=>"uk"),
"/UA/BM/AHTR" => array("type"=>"city", "pop"=>49721, "name_uk"=>"Охтирка", "name_ru"=>"Ахтырка", "name_en"=>"Okhtyrka", "lang"=>"uk", "lat"=>50.3152777777778, "lon"=>34.9005555555556),
"/UA/BM/KNTP" => array("type"=>"city", "pop"=>93000, "name_uk"=>"Конотоп", "name_ru"=>"Конотоп", "name_en"=>"Konotop", "lang"=>"uk", "lat"=>51.2166666666667, "lon"=>33.2),
"/UA/BM/LBDN" => array("type"=>"city", "pop"=>28814, "name_uk"=>"Лебедин", "name_ru"=>"Лебедин", "name_en"=>"Lebedyn", "lang"=>"uk", "lat"=>50.5833333333333, "lon"=>34.4833333333333),
"/UA/BM/RMNY" => array("type"=>"city", "pop"=>54000, "name_uk"=>"Ромни", "name_ru"=>"Ромны", "name_en"=>"Romny", "lang"=>"uk", "lat"=>50.75, "lon"=>33.4833333333333),
"/UA/BM/SHST" => array("type"=>"city", "pop"=>85800, "name_uk"=>"Шостка", "name_ru"=>"Шостка", "name_en"=>"Shostka", "lang"=>"uk", "lat"=>51.8666666666667, "lon"=>33.4833333333333),
"/UA/BM/SUMY" => array("type"=>"city", "pop"=>275426, "name_uk"=>"Суми", "name_ru"=>"Сумы", "name_en"=>"Sumy", "lang"=>"uk", "lat"=>50.9166666666667, "lon"=>34.75),
"/UA/BO" => array("type"=>"region", "pop"=>1112000, "name_uk"=>"Тернопiльська область", "name_ru"=>"Тернопольская область", "name_en"=>"Ternopil Oblast", "lang"=>"uk"),
"/UA/BO/CHRT" => array("type"=>"city", "pop"=>28855, "name_uk"=>"Чорткiв", "name_ru"=>"Чертков", "name_en"=>"Chortkiv", "lang"=>"uk", "lat"=>49.0166666666667, "lon"=>25.8),
"/UA/BO/KRMN" => array("type"=>"city", "pop"=>21880, "name_uk"=>"Кременець", "name_ru"=>"Кременец", "name_en"=>"Kremenets", "lang"=>"uk", "lat"=>50.1197222222222, "lon"=>25.7166666666667),
"/UA/BO/TRNP" => array("type"=>"city", "pop"=>245000, "name_uk"=>"Тернопiль", "name_ru"=>"Тернополь", "name_en"=>"Ternopil", "lang"=>"uk", "lat"=>49.5666666666667, "lon"=>25.6),
"/UA/BT" => array("type"=>"region", "pop"=>1142000, "name_uk"=>"Херсонська область", "name_ru"=>"Херсонская область", "name_en"=>"Kherson Oblast", "lang"=>"uk"),
"/UA/BT/CRPN" => array("type"=>"city", "pop"=>28800, "name_uk"=>"Цюрупинськ", "name_ru"=>"Цюрупинск", "name_en"=>"Tsiurupynsk", "lang"=>"uk", "lat"=>46.6166666666667, "lon"=>32.7166666666667),
"/UA/BT/KHRS" => array("type"=>"city", "pop"=>347000, "name_uk"=>"Херсон", "name_ru"=>"Херсон", "name_en"=>"Kherson", "lang"=>"uk", "lat"=>46.6333333333333, "lon"=>32.6),
"/UA/BT/KHVK" => array("type"=>"city", "pop"=>37892, "name_uk"=>"Каховка", "name_ru"=>"Каховка", "name_en"=>"Kakhovka", "lang"=>"uk", "lat"=>46.8, "lon"=>33.4666666666667),
"/UA/BT/NKHV" => array("type"=>"city", "pop"=>49518, "name_uk"=>"Нова Каховка", "name_ru"=>"Новая Каховка", "name_en"=>"Nova Kakhovka", "lang"=>"uk", "lat"=>46.7666666666667, "lon"=>33.3666666666667),
"/UA/BT/SKDV" => array("type"=>"city", "pop"=>19200, "name_uk"=>"Скадовськ", "name_ru"=>"Скадовск", "name_en"=>"Skadovsk", "lang"=>"uk", "lat"=>46.1166666666667, "lon"=>32.9166666666667),
"/UA/BT/ZHPR" => array("type"=>"city", "pop"=>1528, "name_uk"=>"Залізний Порт", "name_ru"=>"Железный Порт", "name_en"=>"Zaliznyi Port", "lang"=>"uk", "lat"=>46.1222222222222, "lon"=>32.2983333333333),
"/UA/BX" => array("type"=>"region", "pop"=>1373000, "name_uk"=>"Хмельницька область", "name_ru"=>"Хмельницкая область", "name_en"=>"Khmelnytskyi Oblast", "lang"=>"uk"),
"/UA/BX/DNVC" => array("type"=>"city", "pop"=>16448, "name_uk"=>"Дунаївці", "name_ru"=>"Дунаевцы", "name_en"=>"Dunaivtsi", "lang"=>"uk", "lat"=>48.9, "lon"=>26.8333333333333),
"/UA/BX/KHML" => array("type"=>"city", "pop"=>290100, "name_uk"=>"Хмельницький", "name_ru"=>"Хмельницкий", "name_en"=>"Khmelnytskyi", "lang"=>"uk", "lat"=>49.42, "lon"=>27),
"/UA/BX/KMPD" => array("type"=>"city", "pop"=>100025, "name_uk"=>"Кам'янець-Подільський", "name_ru"=>"Каменец-Подольский", "name_en"=>"Kamianets-Podilskyi", "lang"=>"uk", "lat"=>48.6805555555556, "lon"=>26.5805555555556),
"/UA/BX/SHPT" => array("type"=>"city", "pop"=>64844, "name_uk"=>"Шепетівка", "name_ru"=>"Шепетовка", "name_en"=>"Shepetivka", "lang"=>"uk", "lat"=>50.1833333333333, "lon"=>27.0666666666667),
"/UA/CA" => array("type"=>"region", "pop"=>1341000, "name_uk"=>"Черкаська область", "name_ru"=>"Черкасская область", "name_en"=>"Cherkasy Oblast", "lang"=>"uk"),
"/UA/CA/CHRK" => array("type"=>"city", "pop"=>290930, "name_uk"=>"Черкаси", "name_ru"=>"Черкассы", "name_en"=>"Cherkasy", "lang"=>"uk", "lat"=>49.4166666666667, "lon"=>32.1333333333333),
"/UA/CA/SMLA" => array("type"=>"city", "pop"=>68578, "name_uk"=>"Смiла", "name_ru"=>"Смела", "name_en"=>"Smila", "lang"=>"uk", "lat"=>49.2166666666667, "lon"=>31.8666666666667),
"/UA/CA/UMAN" => array("type"=>"city", "pop"=>87658, "name_uk"=>"Умань", "name_ru"=>"Умань", "name_en"=>"Uman", "lang"=>"uk", "lat"=>48.7666666666667, "lon"=>30.2333333333333),
"/UA/CA/ZLTN" => array("type"=>"city", "pop"=>28611, "name_uk"=>"Золотоноша", "name_ru"=>"Золотоноша", "name_en"=>"Zolotonosha", "lang"=>"uk", "lat"=>49.6833333333333, "lon"=>32.0333333333333),
"/UA/CB" => array("type"=>"region", "pop"=>1152000, "name_uk"=>"Чернігівська область", "name_ru"=>"Черниговская область", "name_en"=>"Chernihiv Oblast", "lang"=>"uk"),
"/UA/CB/CRNH" => array("type"=>"city", "pop"=>297816, "name_uk"=>"Чернігів", "name_ru"=>"Чернигов", "name_en"=>"Chernihiv", "lang"=>"uk", "lat"=>51.5, "lon"=>31.3),
"/UA/CB/NZHN" => array("type"=>"city", "pop"=>74727, "name_uk"=>"Нiжин", "name_ru"=>"Нежин", "name_en"=>"Nizhyn", "lang"=>"uk", "lat"=>51.05, "lon"=>31.9),
"/UA/CB/RPLK" => array("type"=>"city", "pop"=>62333, "name_uk"=>"Прилуки", "name_ru"=>"Прилуки", "name_en"=>"Pryluky", "lang"=>"uk", "lat"=>50.595, "lon"=>32.3827777777778),
"/UA/CE" => array("type"=>"region", "pop"=>908000, "name_uk"=>"Чернівецька область", "name_ru"=>"Черновицкая область", "name_en"=>"Chernivtsi Oblast", "lang"=>"uk"),
"/UA/CE/CHRN" => array("type"=>"city", "pop"=>244993, "name_uk"=>"Чернівці", "name_ru"=>"Черновцы", "name_en"=>"Chernivtsi", "lang"=>"uk", "lat"=>48.3, "lon"=>25.9333333333333),
"/UA/KIEV" => array("type"=>"city", "pop"=>2767230, "name_uk"=>"Київ", "name_ru"=>"Киев", "name_en"=>"Kiev", "lang"=>"uk", "lat"=>50.45, "lon"=>30.5),
"/UA/SVST" => array("type"=>"city", "pop"=>380200, "name_uk"=>"Севастополь", "name_ru"=>"Севастополь", "name_en"=>"Sevastopol", "lang"=>"uk", "lat"=>44.6, "lon"=>33.5333333333333),
"/UG" => array("type"=>"country", "pop"=>30900000, "name_ru"=>"Уганда", "name_en"=>"Uganda", "lang"=>"sw", "lat"=>1.06, "lon"=>31.88),
"/US" => array("type"=>"country", "pop"=>303500000, "name_ru"=>"США", "name_en"=>"USA", "lang"=>"en", "lat"=>40.00, "lon"=>-100.00),
"/US/US-AK" => array("type"=>"state", "pop"=>686293, "name_ru"=>"Аляска", "name_en"=>"Alaska", "lang"=>"en"),
"/US/US-AK/adak-station-474" => array("type"=>"cdp", "pop"=>4633, "name_ru"=>"Адак Стейшн", "name_en"=>"Adak Station", "lang"=>"en", "lat"=>51.842900, "lon"=>-176.640278),
"/US/US-AK/akhiok-475" => array("type"=>"city", "pop"=>72, "name_ru"=>"Акхайок", "name_en"=>"Akhiok", "lang"=>"en", "lat"=>56.977320, "lon"=>-154.217551),
"/US/US-AK/akiachak-476" => array("type"=>"city", "pop"=>481, "name_ru"=>"Акичак", "name_en"=>"Akiachak", "lang"=>"en", "lat"=>60.871740, "lon"=>-161.403575),
"/US/US-AK/akiak-477" => array("type"=>"city", "pop"=>309, "name_ru"=>"Акиак", "name_en"=>"Akiak", "lang"=>"en", "lat"=>60.885659, "lon"=>-161.192264),
"/US/US-AK/akutan-478" => array("type"=>"city", "pop"=>871, "name_ru"=>"Акутан", "name_en"=>"Akutan", "lang"=>"en", "lat"=>54.136299, "lon"=>-165.786036),
"/US/US-AK/alakanuk-479" => array("type"=>"city", "pop"=>711, "name_ru"=>"Алаканук", "name_en"=>"Alakanuk", "lang"=>"en", "lat"=>62.675611, "lon"=>-164.643602),
"/US/US-AK/alcan-480" => array("type"=>"cdp", "pop"=>27, "name_ru"=>"Алкан", "name_en"=>"Alcan", "lang"=>"en", "lat"=>62.722977, "lon"=>-141.188178),
"/US/US-AK/aleknagik-481" => array("type"=>"city", "pop"=>222, "name_ru"=>"Алекнеджик", "name_en"=>"Aleknagik", "lang"=>"en", "lat"=>59.285485, "lon"=>-158.628962),
"/US/US-AK/allakaket-482" => array("type"=>"city", "pop"=>85, "name_ru"=>"Аллакакет", "name_en"=>"Allakaket", "lang"=>"en", "lat"=>66.545170, "lon"=>-152.733117),
"/US/US-AK/ambler-483" => array("type"=>"city", "pop"=>322, "name_ru"=>"Амблер", "name_en"=>"Ambler", "lang"=>"en", "lat"=>67.076879, "lon"=>-157.918155),
"/US/US-AK/amchitka-484" => array("type"=>"cdp", "pop"=>25, "name_ru"=>"Амчитка", "name_en"=>"Amchitka", "lang"=>"en", "lat"=>51.567103, "lon"=>178.877380),
"/US/US-AK/anaktuvuk-pass-485" => array("type"=>"city", "pop"=>251, "name_ru"=>"Анкчувек Пасс", "name_en"=>"Anaktuvuk Pass", "lang"=>"en", "lat"=>68.115875, "lon"=>-151.679549),
"/US/US-AK/anchor-point-487" => array("type"=>"cdp", "pop"=>866, "name_ru"=>"Анкор Пойнт", "name_en"=>"Anchor Point", "lang"=>"en", "lat"=>59.770667, "lon"=>-151.813064),
"/US/US-AK/anchorage-486" => array("type"=>"city", "pop"=>279243, "name_ru"=>"Анкоридж", "name_en"=>"Anchorage", "lang"=>"en", "lat"=>61.178368, "lon"=>-149.186416),
"/US/US-AK/anderson-488" => array("type"=>"city", "pop"=>271, "name_ru"=>"Андерсон", "name_en"=>"Anderson", "lang"=>"en", "lat"=>64.306512, "lon"=>-149.159654),
"/US/US-AK/angoon-489" => array("type"=>"city", "pop"=>445, "name_ru"=>"Ангун", "name_en"=>"Angoon", "lang"=>"en", "lat"=>57.478409, "lon"=>-134.520029),
"/US/US-AK/aniak-490" => array("type"=>"city", "pop"=>572, "name_ru"=>"Аниак", "name_en"=>"Aniak", "lang"=>"en", "lat"=>61.580742, "lon"=>-159.539124),
"/US/US-AK/annette-491" => array("type"=>"cdp", "pop"=>43, "name_ru"=>"Аннет", "name_en"=>"Annette", "lang"=>"en", "lat"=>55.042123, "lon"=>-131.585763),
"/US/US-AK/anvik-492" => array("type"=>"city", "pop"=>91, "name_ru"=>"Анвик", "name_en"=>"Anvik", "lang"=>"en", "lat"=>62.655538, "lon"=>-160.208300),
"/US/US-AK/arctic-village-493" => array("type"=>"cdp", "pop"=>96, "name_ru"=>"Арктик Виллидж", "name_en"=>"Arctic Village", "lang"=>"en", "lat"=>68.125245, "lon"=>-145.526906),
"/US/US-AK/atka-494" => array("type"=>"city", "pop"=>71, "name_ru"=>"Атка", "name_en"=>"Atka", "lang"=>"en", "lat"=>52.242218, "lon"=>-174.205154),
"/US/US-AK/atmautluak-495" => array("type"=>"city", "pop"=>258, "name_ru"=>"Атмаутлуак", "name_en"=>"Atmautluak", "lang"=>"en", "lat"=>60.854280, "lon"=>-162.269012),
"/US/US-AK/atqasuk-496" => array("type"=>"city", "pop"=>203, "name_ru"=>"Аткэшек", "name_en"=>"Atqasuk", "lang"=>"en", "lat"=>70.465827, "lon"=>-157.408634),
"/US/US-AK/barrow-497" => array("type"=>"city", "pop"=>4010, "name_ru"=>"Барроу", "name_en"=>"Barrow", "lang"=>"en", "lat"=>71.268264, "lon"=>-156.806271),
"/US/US-AK/beaver-498" => array("type"=>"cdp", "pop"=>103, "name_ru"=>"Бивер", "name_en"=>"Beaver", "lang"=>"en", "lat"=>66.334203, "lon"=>-147.314663),
"/US/US-AK/bethel-499" => array("type"=>"city", "pop"=>6468, "name_ru"=>"Бетел", "name_en"=>"Bethel", "lang"=>"en", "lat"=>60.768211, "lon"=>-161.772218),
"/US/US-AK/bettles-500" => array("type"=>"city", "pop"=>38, "name_ru"=>"Бетлс", "name_en"=>"Bettles", "lang"=>"en", "lat"=>66.886581, "lon"=>-151.594416),
"/US/US-AK/big-delta-501" => array("type"=>"cdp", "pop"=>400, "name_ru"=>"Биг Дельта", "name_en"=>"Big Delta", "lang"=>"en", "lat"=>64.139919, "lon"=>-145.762114),
"/US/US-AK/big-lake-502" => array("type"=>"cdp", "pop"=>1477, "name_ru"=>"Биг Лейк", "name_en"=>"Big Lake", "lang"=>"en", "lat"=>61.521423, "lon"=>-149.954399),
"/US/US-AK/birch-creek-503" => array("type"=>"cdp", "pop"=>42, "name_ru"=>"Берч Крик", "name_en"=>"Birch Creek", "lang"=>"en", "lat"=>66.245449, "lon"=>-145.830269),
"/US/US-AK/brevig-mission-504" => array("type"=>"city", "pop"=>275, "name_ru"=>"Бревиг Мишен", "name_en"=>"Brevig Mission", "lang"=>"en", "lat"=>65.335095, "lon"=>-166.480947),
"/US/US-AK/buckland-505" => array("type"=>"city", "pop"=>422, "name_ru"=>"Бакленд", "name_en"=>"Buckland", "lang"=>"en", "lat"=>65.981052, "lon"=>-161.131676),
"/US/US-AK/butte-506" => array("type"=>"cdp", "pop"=>2039, "name_ru"=>"Бьютт", "name_en"=>"Butte", "lang"=>"en", "lat"=>61.542467, "lon"=>-149.051667),
"/US/US-AK/cantwell-507" => array("type"=>"cdp", "pop"=>147, "name_ru"=>"Кантуэлл", "name_en"=>"Cantwell", "lang"=>"en", "lat"=>63.371603, "lon"=>-148.920201),
"/US/US-AK/central-508" => array("type"=>"cdp", "pop"=>52, "name_ru"=>"Сентрал", "name_en"=>"Central", "lang"=>"en", "lat"=>65.555829, "lon"=>-144.857823),
"/US/US-AK/chalkyitsik-509" => array("type"=>"cdp", "pop"=>90, "name_ru"=>"Чалкиитзик", "name_en"=>"Chalkyitsik", "lang"=>"en", "lat"=>66.638020, "lon"=>-143.755047),
"/US/US-AK/chase-510" => array("type"=>"cdp", "pop"=>38, "name_ru"=>"Чейз", "name_en"=>"Chase", "lang"=>"en", "lat"=>62.449074, "lon"=>-150.101756),
"/US/US-AK/chefornak-511" => array("type"=>"city", "pop"=>394, "name_ru"=>"Чефорнак", "name_en"=>"Chefornak", "lang"=>"en", "lat"=>60.153746, "lon"=>-164.210294),
"/US/US-AK/chenega-512" => array("type"=>"cdp", "pop"=>94, "name_ru"=>"Ченега", "name_en"=>"Chenega", "lang"=>"en", "lat"=>60.104792, "lon"=>-147.943689),
"/US/US-AK/chevak-513" => array("type"=>"city", "pop"=>838, "name_ru"=>"Чевак", "name_en"=>"Chevak", "lang"=>"en", "lat"=>61.533490, "lon"=>-165.591558),
"/US/US-AK/chickaloon-514" => array("type"=>"cdp", "pop"=>145, "name_ru"=>"Чикалун", "name_en"=>"Chickaloon", "lang"=>"en", "lat"=>61.795050, "lon"=>-148.480527),
"/US/US-AK/chignik-515" => array("type"=>"city", "pop"=>63, "name_ru"=>"Чигник", "name_en"=>"Chignik", "lang"=>"en", "lat"=>56.301639, "lon"=>-158.415696),
"/US/US-AK/chignik-lagoon-516" => array("type"=>"cdp", "pop"=>53, "name_ru"=>"Чигник Лагун", "name_en"=>"Chignik Lagoon", "lang"=>"en", "lat"=>56.284127, "lon"=>-158.511127),
"/US/US-AK/chignik-lake-517" => array("type"=>"cdp", "pop"=>133, "name_ru"=>"Чигник Лейк", "name_en"=>"Chignik Lake", "lang"=>"en", "lat"=>56.256133, "lon"=>-158.772221),
"/US/US-AK/chiniak-518" => array("type"=>"cdp", "pop"=>69, "name_ru"=>"Чиниак", "name_en"=>"Chiniak", "lang"=>"en", "lat"=>57.631863, "lon"=>-152.182537),
"/US/US-AK/chistochina-519" => array("type"=>"cdp", "pop"=>60, "name_ru"=>"Чисточайна", "name_en"=>"Chistochina", "lang"=>"en", "lat"=>62.564429, "lon"=>-144.687193),
"/US/US-AK/chitina-520" => array("type"=>"cdp", "pop"=>49, "name_ru"=>"Читина", "name_en"=>"Chitina", "lang"=>"en", "lat"=>61.524841, "lon"=>-144.489121),
"/US/US-AK/chuathbaluk-521" => array("type"=>"city", "pop"=>119, "name_ru"=>"Чуатбалук", "name_en"=>"Chuathbaluk", "lang"=>"en", "lat"=>61.573195, "lon"=>-159.216410),
"/US/US-AK/circle-522" => array("type"=>"cdp", "pop"=>73, "name_ru"=>"Серкл", "name_en"=>"Circle", "lang"=>"en", "lat"=>65.824542, "lon"=>-144.082620),
"/US/US-AK/circle-hot-springs-station-523" => array("type"=>"cdp", "pop"=>29, "name_ru"=>"Серкл Хот Спрингс Стейшен", "name_en"=>"Circle Hot Springs Station", "lang"=>"en", "lat"=>65.470349, "lon"=>-144.689132),
"/US/US-AK/clam-gulch-524" => array("type"=>"cdp", "pop"=>79, "name_ru"=>"Клэм Галч", "name_en"=>"Clam Gulch", "lang"=>"en", "lat"=>60.224674, "lon"=>-151.402133),
"/US/US-AK/clarks-point-525" => array("type"=>"city", "pop"=>60, "name_ru"=>"Кларкс Пойнт", "name_en"=>"Clarks Point", "lang"=>"en", "lat"=>58.845722, "lon"=>-158.515717),
"/US/US-AK/coffman-cove-526" => array("type"=>"city", "pop"=>176, "name_ru"=>"Кофман Ков", "name_en"=>"Coffman Cove", "lang"=>"en", "lat"=>56.008167, "lon"=>-132.839060),
"/US/US-AK/cohoe-527" => array("type"=>"cdp", "pop"=>508, "name_ru"=>"Кохо", "name_en"=>"Cohoe", "lang"=>"en", "lat"=>60.316683, "lon"=>-151.564388),
"/US/US-AK/cold-bay-528" => array("type"=>"city", "pop"=>74, "name_ru"=>"Колд Бэй", "name_en"=>"Cold Bay", "lang"=>"en", "lat"=>55.225026, "lon"=>-162.738879),
"/US/US-AK/college-529" => array("type"=>"cdp", "pop"=>11249, "name_ru"=>"Колледж", "name_en"=>"College", "lang"=>"en", "lat"=>64.869541, "lon"=>-147.823404),
"/US/US-AK/cooper-landing-530" => array("type"=>"cdp", "pop"=>243, "name_ru"=>"Купер Ландинг", "name_en"=>"Cooper Landing", "lang"=>"en", "lat"=>60.491092, "lon"=>-149.792268),
"/US/US-AK/copper-center-531" => array("type"=>"cdp", "pop"=>449, "name_ru"=>"Коппер Сентер", "name_en"=>"Copper Center", "lang"=>"en", "lat"=>61.980756, "lon"=>-145.351126),
"/US/US-AK/copperville-532" => array("type"=>"cdp", "pop"=>163, "name_ru"=>"Коппервилл", "name_en"=>"Copperville", "lang"=>"en", "lat"=>62.043925, "lon"=>-145.420658),
"/US/US-AK/cordova-533" => array("type"=>"city", "pop"=>2242, "name_ru"=>"Кордова", "name_en"=>"Cordova", "lang"=>"en", "lat"=>60.547773, "lon"=>-145.747605),
"/US/US-AK/covenant-life-534" => array("type"=>"cdp", "pop"=>47, "name_ru"=>"Конвенант Лайф", "name_en"=>"Covenant Life", "lang"=>"en", "lat"=>59.417228, "lon"=>-136.025127),
"/US/US-AK/craig-535" => array("type"=>"city", "pop"=>1171, "name_ru"=>"Крейг", "name_en"=>"Craig", "lang"=>"en", "lat"=>55.482800, "lon"=>-133.123153),
"/US/US-AK/crooked-creek-536" => array("type"=>"cdp", "pop"=>106, "name_ru"=>"Крукед Крик", "name_en"=>"Crooked Creek", "lang"=>"en", "lat"=>61.876909, "lon"=>-158.077845),
"/US/US-AK/crown-point-537" => array("type"=>"cdp", "pop"=>62, "name_ru"=>"Краун Пойнт", "name_en"=>"Crown Point", "lang"=>"en", "lat"=>60.422530, "lon"=>-149.337543),
"/US/US-AK/cube-cove-538" => array("type"=>"cdp", "pop"=>156, "name_ru"=>"Каб Ков", "name_en"=>"Cube Cove", "lang"=>"en", "lat"=>57.935095, "lon"=>-134.711234),
"/US/US-AK/deadhorse-539" => array("type"=>"cdp", "pop"=>26, "name_ru"=>"Дэдхорс", "name_en"=>"Deadhorse", "lang"=>"en", "lat"=>70.205556, "lon"=>-148.511648),
"/US/US-AK/deering-540" => array("type"=>"city", "pop"=>141, "name_ru"=>"Диаринг", "name_en"=>"Deering", "lang"=>"en", "lat"=>66.073600, "lon"=>-162.731743),
"/US/US-AK/delta-junction-541" => array("type"=>"city", "pop"=>930, "name_ru"=>"Делта Джанкшен", "name_en"=>"Delta Junction", "lang"=>"en", "lat"=>64.063687, "lon"=>-145.708158),
"/US/US-AK/dillingham-542" => array("type"=>"city", "pop"=>2465, "name_ru"=>"Диллингем", "name_en"=>"Dillingham", "lang"=>"en", "lat"=>59.062164, "lon"=>-158.527978),
"/US/US-AK/diomede-543" => array("type"=>"city", "pop"=>146, "name_ru"=>"Диомед", "name_en"=>"Diomede", "lang"=>"en", "lat"=>65.774678, "lon"=>-168.915363),
"/US/US-AK/dora-bay-544" => array("type"=>"cdp", "pop"=>57, "name_ru"=>"Дора Бэй", "name_en"=>"Dora Bay", "lang"=>"en", "lat"=>55.188991, "lon"=>-132.295196),
"/US/US-AK/dot-lake-545" => array("type"=>"cdp", "pop"=>70, "name_ru"=>"Дот Лейк", "name_en"=>"Dot Lake", "lang"=>"en", "lat"=>63.629122, "lon"=>-144.095120),
"/US/US-AK/dry-creek-546" => array("type"=>"cdp", "pop"=>106, "name_ru"=>"Драй Крик", "name_en"=>"Dry Creek", "lang"=>"en", "lat"=>63.653404, "lon"=>-144.665296),
"/US/US-AK/eagle-547" => array("type"=>"city", "pop"=>145, "name_ru"=>"Игл", "name_en"=>"Eagle", "lang"=>"en", "lat"=>64.778147, "lon"=>-141.200633),
"/US/US-AK/eagle-village-548" => array("type"=>"cdp", "pop"=>35, "name_ru"=>"Игл Виллидж", "name_en"=>"Eagle Village", "lang"=>"en", "lat"=>64.790425, "lon"=>-141.109100),
"/US/US-AK/edna-bay-549" => array("type"=>"cdp", "pop"=>86, "name_ru"=>"Эдна Бэй", "name_en"=>"Edna Bay", "lang"=>"en", "lat"=>55.953930, "lon"=>-133.650728),
"/US/US-AK/eek-550" => array("type"=>"city", "pop"=>280, "name_ru"=>"Ик", "name_en"=>"Eek", "lang"=>"en", "lat"=>60.218890, "lon"=>-162.026425),
"/US/US-AK/egegik-551" => array("type"=>"cdp", "pop"=>93, "name_ru"=>"Эгиджик", "name_en"=>"Egegik", "lang"=>"en", "lat"=>58.228738, "lon"=>-157.392700),
"/US/US-AK/eielson-afb-552" => array("type"=>"cdp", "pop"=>5251, "name_ru"=>"Авиабаза Эйелсон ", "name_en"=>"Eielson AFB", "lang"=>"en", "lat"=>64.675713, "lon"=>-147.091316),
"/US/US-AK/ekwok-553" => array("type"=>"city", "pop"=>130, "name_ru"=>"Эквок", "name_en"=>"Ekwok", "lang"=>"en", "lat"=>59.360217, "lon"=>-157.479396),
"/US/US-AK/elfin-cove-554" => array("type"=>"cdp", "pop"=>57, "name_ru"=>"Элфин Ков", "name_en"=>"Elfin Cove", "lang"=>"en", "lat"=>58.186247, "lon"=>-136.336831),
"/US/US-AK/elim-555" => array("type"=>"city", "pop"=>313, "name_ru"=>"Элим", "name_en"=>"Elim", "lang"=>"en", "lat"=>64.621662, "lon"=>-162.260371),
"/US/US-AK/emmonak-556" => array("type"=>"city", "pop"=>841, "name_ru"=>"Эммонак", "name_en"=>"Emmonak", "lang"=>"en", "lat"=>62.778339, "lon"=>-164.537199),
"/US/US-AK/english-bay-557" => array("type"=>"cdp", "pop"=>158, "name_ru"=>"Инглиш Бэй", "name_en"=>"English Bay", "lang"=>"en", "lat"=>59.346854, "lon"=>-151.903489),
"/US/US-AK/ester-558" => array("type"=>"cdp", "pop"=>147, "name_ru"=>"Истер", "name_en"=>"Ester", "lang"=>"en", "lat"=>64.861264, "lon"=>-148.024538),
"/US/US-AK/evansville-559" => array("type"=>"cdp", "pop"=>33, "name_ru"=>"Эвансвилл", "name_en"=>"Evansville", "lang"=>"en", "lat"=>66.878557, "lon"=>-151.459670),
"/US/US-AK/eyak-560" => array("type"=>"cdp", "pop"=>172, "name_ru"=>"Эяк", "name_en"=>"Eyak", "lang"=>"en", "lat"=>60.528455, "lon"=>-145.592556),
"/US/US-AK/fairbanks-561" => array("type"=>"city", "pop"=>35132, "name_ru"=>"Фэрбенкс", "name_en"=>"Fairbanks", "lang"=>"en", "lat"=>64.837088, "lon"=>-147.649045),
"/US/US-AK/false-pass-562" => array("type"=>"cdp", "pop"=>54, "name_ru"=>"Фолс Пас", "name_en"=>"False Pass", "lang"=>"en", "lat"=>54.841028, "lon"=>-163.436845),
"/US/US-AK/ferry-563" => array("type"=>"cdp", "pop"=>56, "name_ru"=>"Ферри", "name_en"=>"Ferry", "lang"=>"en", "lat"=>63.927883, "lon"=>-149.130232),
"/US/US-AK/fort-greely-564" => array("type"=>"cdp", "pop"=>1147, "name_ru"=>"Форт Грили", "name_en"=>"Fort Greely", "lang"=>"en", "lat"=>63.958664, "lon"=>-145.756094),
"/US/US-AK/fort-yukon-565" => array("type"=>"city", "pop"=>520, "name_ru"=>"Форт Юкон", "name_en"=>"Fort Yukon", "lang"=>"en", "lat"=>66.564951, "lon"=>-145.261831),
"/US/US-AK/fox-566" => array("type"=>"cdp", "pop"=>275, "name_ru"=>"Фокс", "name_en"=>"Fox", "lang"=>"en", "lat"=>64.960744, "lon"=>-147.621258),
"/US/US-AK/fox-river-567" => array("type"=>"cdp", "pop"=>382, "name_ru"=>"Фокс Ривер", "name_en"=>"Fox River", "lang"=>"en", "lat"=>59.858323, "lon"=>-150.958233),
"/US/US-AK/freshwater-bay-568" => array("type"=>"cdp", "pop"=>68, "name_ru"=>"Фрэшуотер Бэй", "name_en"=>"Freshwater Bay", "lang"=>"en", "lat"=>57.891643, "lon"=>-135.079348),
"/US/US-AK/fritz-creek-569" => array("type"=>"cdp", "pop"=>1426, "name_ru"=>"Фриц Крик", "name_en"=>"Fritz Creek", "lang"=>"en", "lat"=>59.736089, "lon"=>-151.295346),
"/US/US-AK/gakona-570" => array("type"=>"cdp", "pop"=>25, "name_ru"=>"Гакона", "name_en"=>"Gakona", "lang"=>"en", "lat"=>62.292729, "lon"=>-145.309498),
"/US/US-AK/galena-571" => array("type"=>"city", "pop"=>599, "name_ru"=>"Галена", "name_en"=>"Galena", "lang"=>"en", "lat"=>64.740772, "lon"=>-156.819862),
"/US/US-AK/gambell-572" => array("type"=>"city", "pop"=>647, "name_ru"=>"Гамбелл", "name_en"=>"Gambell", "lang"=>"en", "lat"=>63.776555, "lon"=>-171.701685),
"/US/US-AK/game-creek-573" => array("type"=>"cdp", "pop"=>61, "name_ru"=>"Гейм Крик", "name_en"=>"Game Creek", "lang"=>"en", "lat"=>58.056687, "lon"=>-135.507890),
"/US/US-AK/glennallen-574" => array("type"=>"cdp", "pop"=>451, "name_ru"=>"Гленналлен", "name_en"=>"Glennallen", "lang"=>"en", "lat"=>62.084862, "lon"=>-145.603520),
"/US/US-AK/golovin-575" => array("type"=>"city", "pop"=>144, "name_ru"=>"Головин", "name_en"=>"Golovin", "lang"=>"en", "lat"=>64.578979, "lon"=>-162.990887),
"/US/US-AK/goodnews-bay-576" => array("type"=>"city", "pop"=>230, "name_ru"=>"Гуднбюс Бэй", "name_en"=>"Goodnews Bay", "lang"=>"en", "lat"=>59.109807, "lon"=>-161.565292),
"/US/US-AK/grayling-577" => array("type"=>"city", "pop"=>170, "name_ru"=>"Грйлинг", "name_en"=>"Grayling", "lang"=>"en", "lat"=>62.899503, "lon"=>-160.107398),
"/US/US-AK/gulkana-578" => array("type"=>"cdp", "pop"=>103, "name_ru"=>"Галкана", "name_en"=>"Gulkana", "lang"=>"en", "lat"=>62.252675, "lon"=>-145.399049),
"/US/US-AK/gustavus-579" => array("type"=>"cdp", "pop"=>355, "name_ru"=>"Гаставус", "name_en"=>"Gustavus", "lang"=>"en", "lat"=>58.406090, "lon"=>-135.801088),
"/US/US-AK/haines-580" => array("type"=>"city", "pop"=>2271, "name_ru"=>"Хайнс", "name_en"=>"Haines", "lang"=>"en", "lat"=>59.240250, "lon"=>-135.440405),
"/US/US-AK/halibut-cove-581" => array("type"=>"cdp", "pop"=>78, "name_ru"=>"Халибут Ков", "name_en"=>"Halibut Cove", "lang"=>"en", "lat"=>59.587217, "lon"=>-151.237659),
"/US/US-AK/happy-valley-582" => array("type"=>"cdp", "pop"=>309, "name_ru"=>"Хэппи Валли", "name_en"=>"Happy Valley", "lang"=>"en", "lat"=>59.947278, "lon"=>-151.732222),
"/US/US-AK/harding-lake-583" => array("type"=>"cdp", "pop"=>27, "name_ru"=>"Хардинг Лейк", "name_en"=>"Harding Lake", "lang"=>"en", "lat"=>64.421798, "lon"=>-146.850712),
"/US/US-AK/healy-584" => array("type"=>"cdp", "pop"=>487, "name_ru"=>"Хили", "name_en"=>"Healy", "lang"=>"en", "lat"=>63.811541, "lon"=>-149.002048),
"/US/US-AK/healy-lake-585" => array("type"=>"cdp", "pop"=>47, "name_ru"=>"Хили Лейк", "name_en"=>"Healy Lake", "lang"=>"en", "lat"=>63.977010, "lon"=>-144.724097),
"/US/US-AK/hobart-bay-586" => array("type"=>"cdp", "pop"=>187, "name_ru"=>"Хобарт Бэй", "name_en"=>"Hobart Bay", "lang"=>"en", "lat"=>57.465822, "lon"=>-133.353254),
"/US/US-AK/hollis-587" => array("type"=>"cdp", "pop"=>111, "name_ru"=>"Холлис", "name_en"=>"Hollis", "lang"=>"en", "lat"=>55.556697, "lon"=>-132.636312),
"/US/US-AK/holy-cross-588" => array("type"=>"city", "pop"=>199, "name_ru"=>"Холи Кросс", "name_en"=>"Holy Cross", "lang"=>"en", "lat"=>62.181376, "lon"=>-159.799406),
"/US/US-AK/homer-589" => array("type"=>"city", "pop"=>5802, "name_ru"=>"Гомер", "name_en"=>"Homer", "lang"=>"en", "lat"=>59.635463, "lon"=>-151.522165),
"/US/US-AK/hoonah-590" => array("type"=>"city", "pop"=>715, "name_ru"=>"Хуна", "name_en"=>"Hoonah", "lang"=>"en", "lat"=>58.111547, "lon"=>-135.419779),
"/US/US-AK/hooper-bay-591" => array("type"=>"city", "pop"=>1109, "name_ru"=>"Хупер Бэй", "name_en"=>"Hooper Bay", "lang"=>"en", "lat"=>61.539757, "lon"=>-166.103051),
"/US/US-AK/hope-592" => array("type"=>"cdp", "pop"=>161, "name_ru"=>"Хоуп", "name_en"=>"Hope", "lang"=>"en", "lat"=>60.897173, "lon"=>-149.631634),
"/US/US-AK/houston-593" => array("type"=>"city", "pop"=>2153, "name_ru"=>"Хьюстон", "name_en"=>"Houston", "lang"=>"en", "lat"=>61.624142, "lon"=>-149.781057),
"/US/US-AK/hughes-594" => array("type"=>"city", "pop"=>69, "name_ru"=>"Хьюз", "name_en"=>"Hughes", "lang"=>"en", "lat"=>66.037086, "lon"=>-154.267740),
"/US/US-AK/huslia-595" => array("type"=>"city", "pop"=>257, "name_ru"=>"Хуслиа", "name_en"=>"Huslia", "lang"=>"en", "lat"=>65.689918, "lon"=>-156.291976),
"/US/US-AK/hydaburg-596" => array("type"=>"city", "pop"=>339, "name_ru"=>"Хайдабург", "name_en"=>"Hydaburg", "lang"=>"en", "lat"=>55.208959, "lon"=>-132.817389),
"/US/US-AK/hyder-597" => array("type"=>"cdp", "pop"=>99, "name_ru"=>"Хайдер", "name_en"=>"Hyder", "lang"=>"en", "lat"=>55.980417, "lon"=>-130.058689),
"/US/US-AK/igiugig-598" => array("type"=>"cdp", "pop"=>33, "name_ru"=>"Игьюджиг", "name_en"=>"Igiugig", "lang"=>"en", "lat"=>59.335135, "lon"=>-155.907214),
"/US/US-AK/iliamna-599" => array("type"=>"cdp", "pop"=>94, "name_ru"=>"Иламна", "name_en"=>"Iliamna", "lang"=>"en", "lat"=>59.775077, "lon"=>-154.867589),
"/US/US-AK/ivanof-bay-600" => array("type"=>"cdp", "pop"=>35, "name_ru"=>"Иваноф Бэй", "name_en"=>"Ivanof Bay", "lang"=>"en", "lat"=>55.960836, "lon"=>-159.489478),
"/US/US-AK/jakolof-bay-601" => array("type"=>"cdp", "pop"=>28, "name_ru"=>"Джаколоф Бэй", "name_en"=>"Jakolof Bay", "lang"=>"en", "lat"=>59.446922, "lon"=>-151.510399),
"/US/US-AK/juneau-602" => array("type"=>"city", "pop"=>26751, "name_ru"=>"Джуно", "name_en"=>"Juneau", "lang"=>"en", "lat"=>58.388588, "lon"=>-134.133421),
"/US/US-AK/kachemak-603" => array("type"=>"city", "pop"=>442, "name_ru"=>"Качемак", "name_en"=>"Kachemak", "lang"=>"en", "lat"=>59.674003, "lon"=>-151.431041),
"/US/US-AK/kake-604" => array("type"=>"city", "pop"=>635, "name_ru"=>"Кейк", "name_en"=>"Kake", "lang"=>"en", "lat"=>56.968536, "lon"=>-133.924976),
"/US/US-AK/kaktovik-605" => array("type"=>"city", "pop"=>260, "name_ru"=>"Кактовик", "name_en"=>"Kaktovik", "lang"=>"en", "lat"=>70.049178, "lon"=>-143.637027),
"/US/US-AK/kalifonsky-606" => array("type"=>"cdp", "pop"=>285, "name_ru"=>"Калифорнски", "name_en"=>"Kalifonsky", "lang"=>"en", "lat"=>60.410079, "lon"=>-151.279770),
"/US/US-AK/kaltag-607" => array("type"=>"city", "pop"=>202, "name_ru"=>"Калтэг", "name_en"=>"Kaltag", "lang"=>"en", "lat"=>64.338617, "lon"=>-158.686168),
"/US/US-AK/karluk-608" => array("type"=>"cdp", "pop"=>71, "name_ru"=>"Карлек", "name_en"=>"Karluk", "lang"=>"en", "lat"=>57.586064, "lon"=>-154.395295),
"/US/US-AK/kasaan-609" => array("type"=>"city", "pop"=>34, "name_ru"=>"Касан", "name_en"=>"Kasaan", "lang"=>"en", "lat"=>55.546697, "lon"=>-132.404122),
"/US/US-AK/kasigluk-610" => array("type"=>"city", "pop"=>425, "name_ru"=>"Кейсиглек", "name_en"=>"Kasigluk", "lang"=>"en", "lat"=>60.888999, "lon"=>-162.545653),
"/US/US-AK/kasilof-611" => array("type"=>"cdp", "pop"=>383, "name_ru"=>"Кейсилоф", "name_en"=>"Kasilof", "lang"=>"en", "lat"=>60.332904, "lon"=>-151.229949),
"/US/US-AK/kenai-612" => array("type"=>"city", "pop"=>7742, "name_ru"=>"Кенай", "name_en"=>"Kenai", "lang"=>"en", "lat"=>60.553745, "lon"=>-151.207416),
"/US/US-AK/kenny-lake-613" => array("type"=>"cdp", "pop"=>423, "name_ru"=>"Кенни Лейк", "name_en"=>"Kenny Lake", "lang"=>"en", "lat"=>61.719264, "lon"=>-144.948272),
"/US/US-AK/ketchikan-614" => array("type"=>"city", "pop"=>7280, "name_ru"=>"Кетчикан", "name_en"=>"Ketchikan", "lang"=>"en", "lat"=>55.352515, "lon"=>-131.659582),
"/US/US-AK/kiana-615" => array("type"=>"city", "pop"=>403, "name_ru"=>"Киана", "name_en"=>"Kiana", "lang"=>"en", "lat"=>66.970152, "lon"=>-160.458160),
"/US/US-AK/king-cove-616" => array("type"=>"city", "pop"=>734, "name_ru"=>"Кинг Ков", "name_en"=>"King Cove", "lang"=>"en", "lat"=>55.058571, "lon"=>-162.299822),
"/US/US-AK/king-salmon-617" => array("type"=>"cdp", "pop"=>696, "name_ru"=>"Кинг Салмон", "name_en"=>"King Salmon", "lang"=>"en", "lat"=>58.742922, "lon"=>-156.544639),
"/US/US-AK/kipnuk-618" => array("type"=>"cdp", "pop"=>470, "name_ru"=>"Кипнек", "name_en"=>"Kipnuk", "lang"=>"en", "lat"=>59.923204, "lon"=>-164.101013),
"/US/US-AK/kivalina-619" => array("type"=>"city", "pop"=>391, "name_ru"=>"Кивалина", "name_en"=>"Kivalina", "lang"=>"en", "lat"=>67.722480, "lon"=>-164.533081),
"/US/US-AK/klawock-620" => array("type"=>"city", "pop"=>736, "name_ru"=>"Клоок", "name_en"=>"Klawock", "lang"=>"en", "lat"=>55.553792, "lon"=>-133.091015),
"/US/US-AK/klukwan-621" => array("type"=>"cdp", "pop"=>129, "name_ru"=>"Клукван", "name_en"=>"Klukwan", "lang"=>"en", "lat"=>59.403875, "lon"=>-135.884494),
"/US/US-AK/knik-622" => array("type"=>"cdp", "pop"=>272, "name_ru"=>"Найк", "name_en"=>"Knik", "lang"=>"en", "lat"=>61.453293, "lon"=>-149.741617),
"/US/US-AK/kobuk-623" => array("type"=>"city", "pop"=>113, "name_ru"=>"Кобек", "name_en"=>"Kobuk", "lang"=>"en", "lat"=>66.904607, "lon"=>-156.896358),
"/US/US-AK/kodiak-624" => array("type"=>"city", "pop"=>6228, "name_ru"=>"Кодиак", "name_en"=>"Kodiak", "lang"=>"en", "lat"=>57.798532, "lon"=>-152.402265),
"/US/US-AK/kodiak-station-625" => array("type"=>"cdp", "pop"=>2025, "name_ru"=>"Кодиак Стейшен", "name_en"=>"Kodiak Station", "lang"=>"en", "lat"=>57.762943, "lon"=>-152.526856),
"/US/US-AK/kokhanok-626" => array("type"=>"cdp", "pop"=>152, "name_ru"=>"Кокханок", "name_en"=>"Kokhanok", "lang"=>"en", "lat"=>59.407512, "lon"=>-154.744169),
"/US/US-AK/koliganek-627" => array("type"=>"cdp", "pop"=>181, "name_ru"=>"Колиганк", "name_en"=>"Koliganek", "lang"=>"en", "lat"=>59.736123, "lon"=>-157.266331),
"/US/US-AK/kongiganak-628" => array("type"=>"cdp", "pop"=>294, "name_ru"=>"Конгиганак", "name_en"=>"Kongiganak", "lang"=>"en", "lat"=>59.923976, "lon"=>-162.851942),
"/US/US-AK/kotlik-629" => array("type"=>"city", "pop"=>649, "name_ru"=>"Котлайк", "name_en"=>"Kotlik", "lang"=>"en", "lat"=>63.034125, "lon"=>-163.554705),
"/US/US-AK/kotzebue-630" => array("type"=>"city", "pop"=>3177, "name_ru"=>"Коцебу", "name_en"=>"Kotzebue", "lang"=>"en", "lat"=>66.888023, "lon"=>-162.535388),
"/US/US-AK/koyuk-631" => array("type"=>"city", "pop"=>296, "name_ru"=>"Коюк", "name_en"=>"Koyuk", "lang"=>"en", "lat"=>64.932850, "lon"=>-161.148348),
"/US/US-AK/koyukuk-632" => array("type"=>"city", "pop"=>89, "name_ru"=>"Коюкек", "name_en"=>"Koyukuk", "lang"=>"en", "lat"=>64.913529, "lon"=>-157.650652),
"/US/US-AK/kupreanof-633" => array("type"=>"city", "pop"=>21, "name_ru"=>"Купреаноф", "name_en"=>"Kupreanof", "lang"=>"en", "lat"=>56.808948, "lon"=>-132.985238),
"/US/US-AK/kwethluk-634" => array("type"=>"city", "pop"=>715, "name_ru"=>"Куэтлек", "name_en"=>"Kwethluk", "lang"=>"en", "lat"=>60.771814, "lon"=>-161.388490),
"/US/US-AK/kwigillingok-635" => array("type"=>"cdp", "pop"=>278, "name_ru"=>"Квигиллингок", "name_en"=>"Kwigillingok", "lang"=>"en", "lat"=>59.835882, "lon"=>-163.113734),
"/US/US-AK/labouchere-bay-636" => array("type"=>"cdp", "pop"=>149, "name_ru"=>"Лабушер Бэй", "name_en"=>"Labouchere Bay", "lang"=>"en", "lat"=>56.296341, "lon"=>-133.632253),
"/US/US-AK/lake-minchumina-637" => array("type"=>"cdp", "pop"=>32, "name_ru"=>"Лейк Минчумина", "name_en"=>"Lake Minchumina", "lang"=>"en", "lat"=>63.874443, "lon"=>-152.412786),
"/US/US-AK/larsen-bay-638" => array("type"=>"city", "pop"=>104, "name_ru"=>"Ларсен Бэй", "name_en"=>"Larsen Bay", "lang"=>"en", "lat"=>57.558681, "lon"=>-154.019734),
"/US/US-AK/lazy-mountain-639" => array("type"=>"cdp", "pop"=>838, "name_ru"=>"Лейзи Маунтин", "name_en"=>"Lazy Mountain", "lang"=>"en", "lat"=>61.626178, "lon"=>-148.945657),
"/US/US-AK/levelock-640" => array("type"=>"cdp", "pop"=>105, "name_ru"=>"Левелок", "name_en"=>"Levelock", "lang"=>"en", "lat"=>59.108594, "lon"=>-156.885009),
"/US/US-AK/lignite-641" => array("type"=>"cdp", "pop"=>99, "name_ru"=>"Лайнит", "name_en"=>"Lignite", "lang"=>"en", "lat"=>63.890511, "lon"=>-149.046410),
"/US/US-AK/lime-village-642" => array("type"=>"cdp", "pop"=>42, "name_ru"=>"Лайм Виллидж", "name_en"=>"Lime Village", "lang"=>"en", "lat"=>61.432815, "lon"=>-155.472526),
"/US/US-AK/long-island-643" => array("type"=>"cdp", "pop"=>198, "name_ru"=>"Лонг Айленд", "name_en"=>"Long Island", "lang"=>"en", "lat"=>54.917060, "lon"=>-132.679279),
"/US/US-AK/lower-kalskag-644" => array("type"=>"city", "pop"=>268, "name_ru"=>"Лоуэр Калскаг", "name_en"=>"Lower Kalskag", "lang"=>"en", "lat"=>61.518760, "lon"=>-160.349905),
"/US/US-AK/lutak-645" => array("type"=>"cdp", "pop"=>45, "name_ru"=>"Лутак", "name_en"=>"Lutak", "lang"=>"en", "lat"=>59.341066, "lon"=>-135.565509),
"/US/US-AK/manley-hot-springs-649" => array("type"=>"cdp", "pop"=>96, "name_ru"=>"Манли Хот Спрингс", "name_en"=>"Manley Hot Springs", "lang"=>"en", "lat"=>64.985763, "lon"=>-150.636587),
"/US/US-AK/manokotak-650" => array("type"=>"city", "pop"=>401, "name_ru"=>"Манокотак", "name_en"=>"Manokotak", "lang"=>"en", "lat"=>59.009559, "lon"=>-158.989699),
"/US/US-AK/marshall-651" => array("type"=>"city", "pop"=>382, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>61.875343, "lon"=>-162.060098),
"/US/US-AK/mccarthy-646" => array("type"=>"cdp", "pop"=>25, "name_ru"=>"Маккарти", "name_en"=>"McCarthy", "lang"=>"en", "lat"=>61.423206, "lon"=>-142.898032),
"/US/US-AK/mcgrath-647" => array("type"=>"city", "pop"=>351, "name_ru"=>"Мак-Грат", "name_en"=>"McGrath", "lang"=>"en", "lat"=>62.954315, "lon"=>-155.575753),
"/US/US-AK/mckinley-park-648" => array("type"=>"cdp", "pop"=>171, "name_ru"=>"Мак-Кинли Парк", "name_en"=>"McKinley Park", "lang"=>"en", "lat"=>63.696655, "lon"=>-148.954823),
"/US/US-AK/meadow-lakes-652" => array("type"=>"cdp", "pop"=>2374, "name_ru"=>"Медоу Лейкс", "name_en"=>"Meadow Lakes", "lang"=>"en", "lat"=>61.624594, "lon"=>-149.601230),
"/US/US-AK/mekoryuk-653" => array("type"=>"city", "pop"=>210, "name_ru"=>"Мекорюк", "name_en"=>"Mekoryuk", "lang"=>"en", "lat"=>60.365679, "lon"=>-166.283583),
"/US/US-AK/mendeltna-654" => array("type"=>"cdp", "pop"=>37, "name_ru"=>"Меделтна", "name_en"=>"Mendeltna", "lang"=>"en", "lat"=>62.066243, "lon"=>-146.449867),
"/US/US-AK/mentasta-lake-655" => array("type"=>"cdp", "pop"=>96, "name_ru"=>"Ментста Лейк", "name_en"=>"Mentasta Lake", "lang"=>"en", "lat"=>62.851791, "lon"=>-143.760878),
"/US/US-AK/metlakatla-656" => array("type"=>"cdp", "pop"=>1407, "name_ru"=>"Метлакатла", "name_en"=>"Metlakatla", "lang"=>"en", "lat"=>55.124552, "lon"=>-131.580632),
"/US/US-AK/meyers-chuck-657" => array("type"=>"cdp", "pop"=>37, "name_ru"=>"Мейерс Чак", "name_en"=>"Meyers Chuck", "lang"=>"en", "lat"=>55.714925, "lon"=>-132.222972),
"/US/US-AK/minto-658" => array("type"=>"cdp", "pop"=>218, "name_ru"=>"Минто", "name_en"=>"Minto", "lang"=>"en", "lat"=>65.026198, "lon"=>-149.529566),
"/US/US-AK/moose-creek-659" => array("type"=>"cdp", "pop"=>610, "name_ru"=>"Муз Крик", "name_en"=>"Moose Creek", "lang"=>"en", "lat"=>64.713723, "lon"=>-147.165180),
"/US/US-AK/moose-pass-660" => array("type"=>"cdp", "pop"=>81, "name_ru"=>"Муз Пас", "name_en"=>"Moose Pass", "lang"=>"en", "lat"=>60.467277, "lon"=>-149.388376),
"/US/US-AK/mosquito-lake-661" => array("type"=>"cdp", "pop"=>80, "name_ru"=>"Москито Лейк", "name_en"=>"Mosquito Lake", "lang"=>"en", "lat"=>59.529250, "lon"=>-136.090978),
"/US/US-AK/mountain-village-662" => array("type"=>"city", "pop"=>826, "name_ru"=>"Маунтин Виллидж", "name_en"=>"Mountain Village", "lang"=>"en", "lat"=>62.087808, "lon"=>-163.716689),
"/US/US-AK/naknek-663" => array("type"=>"cdp", "pop"=>575, "name_ru"=>"Нейкнек", "name_en"=>"Naknek", "lang"=>"en", "lat"=>58.760722, "lon"=>-156.901849),
"/US/US-AK/napakiak-664" => array("type"=>"city", "pop"=>353, "name_ru"=>"Напакиак", "name_en"=>"Napakiak", "lang"=>"en", "lat"=>60.676886, "lon"=>-161.978696),
"/US/US-AK/napaskiak-665" => array("type"=>"city", "pop"=>391, "name_ru"=>"Напаскиак", "name_en"=>"Napaskiak", "lang"=>"en", "lat"=>60.663758, "lon"=>-161.738144),
"/US/US-AK/naukati-bay-666" => array("type"=>"cdp", "pop"=>93, "name_ru"=>"Нокати Бэй", "name_en"=>"Naukati Bay", "lang"=>"en", "lat"=>55.873675, "lon"=>-133.184760),
"/US/US-AK/nelson-lagoon-667" => array("type"=>"cdp", "pop"=>83, "name_ru"=>"Нельсон Лагун", "name_en"=>"Nelson Lagoon", "lang"=>"en", "lat"=>55.920749, "lon"=>-161.189030),
"/US/US-AK/nenana-668" => array("type"=>"city", "pop"=>344, "name_ru"=>"Ненана", "name_en"=>"Nenana", "lang"=>"en", "lat"=>64.543954, "lon"=>-149.086694),
"/US/US-AK/new-stuyahok-670" => array("type"=>"city", "pop"=>473, "name_ru"=>"Нью Стуиахок", "name_en"=>"New Stuyahok", "lang"=>"en", "lat"=>59.473183, "lon"=>-157.262024),
"/US/US-AK/newhalen-669" => array("type"=>"city", "pop"=>128, "name_ru"=>"Ньюхален", "name_en"=>"Newhalen", "lang"=>"en", "lat"=>59.734437, "lon"=>-154.880083),
"/US/US-AK/newtok-671" => array("type"=>"city", "pop"=>207, "name_ru"=>"Ньюток", "name_en"=>"Newtok", "lang"=>"en", "lat"=>60.949358, "lon"=>-164.644051),
"/US/US-AK/nightmute-672" => array("type"=>"city", "pop"=>208, "name_ru"=>"Найтмут", "name_en"=>"Nightmute", "lang"=>"en", "lat"=>60.448283, "lon"=>-164.804066),
"/US/US-AK/nikiski-673" => array("type"=>"cdp", "pop"=>2743, "name_ru"=>"Никиски", "name_en"=>"Nikiski", "lang"=>"en", "lat"=>60.725224, "lon"=>-151.382747),
"/US/US-AK/nikolaevsk-674" => array("type"=>"cdp", "pop"=>371, "name_ru"=>"Николаевск", "name_en"=>"Nikolaevsk", "lang"=>"en", "lat"=>59.832340, "lon"=>-151.589226),
"/US/US-AK/nikolai-675" => array("type"=>"city", "pop"=>86, "name_ru"=>"Николай", "name_en"=>"Nikolai", "lang"=>"en", "lat"=>62.995238, "lon"=>-154.390664),
"/US/US-AK/nikolski-676" => array("type"=>"cdp", "pop"=>35, "name_ru"=>"Никольски", "name_en"=>"Nikolski", "lang"=>"en", "lat"=>52.961300, "lon"=>-168.822061),
"/US/US-AK/ninilchik-677" => array("type"=>"cdp", "pop"=>456, "name_ru"=>"Нинилчайк", "name_en"=>"Ninilchik", "lang"=>"en", "lat"=>60.073494, "lon"=>-151.631301),
"/US/US-AK/noatak-678" => array("type"=>"cdp", "pop"=>333, "name_ru"=>"Нотак", "name_en"=>"Noatak", "lang"=>"en", "lat"=>67.582831, "lon"=>-163.005221),
"/US/US-AK/nome-679" => array("type"=>"city", "pop"=>3576, "name_ru"=>"Ном", "name_en"=>"Nome", "lang"=>"en", "lat"=>64.509225, "lon"=>-165.415193),
"/US/US-AK/nondalton-680" => array("type"=>"city", "pop"=>177, "name_ru"=>"Нондалтон", "name_en"=>"Nondalton", "lang"=>"en", "lat"=>59.990396, "lon"=>-154.853101),
"/US/US-AK/noorvik-681" => array("type"=>"city", "pop"=>660, "name_ru"=>"Ноорвик", "name_en"=>"Noorvik", "lang"=>"en", "lat"=>66.828619, "lon"=>-161.042607),
"/US/US-AK/north-pole-682" => array("type"=>"city", "pop"=>2212, "name_ru"=>"Норт Пол", "name_en"=>"North Pole", "lang"=>"en", "lat"=>64.753195, "lon"=>-147.355951),
"/US/US-AK/northway-683" => array("type"=>"cdp", "pop"=>123, "name_ru"=>"Нортуэй", "name_en"=>"Northway", "lang"=>"en", "lat"=>62.932888, "lon"=>-141.874895),
"/US/US-AK/northway-junction-684" => array("type"=>"cdp", "pop"=>88, "name_ru"=>"Нортуэй Джанкшен", "name_en"=>"Northway Junction", "lang"=>"en", "lat"=>63.002092, "lon"=>-141.777345),
"/US/US-AK/northway-village-685" => array("type"=>"cdp", "pop"=>113, "name_ru"=>"Нортуэй Виллидж", "name_en"=>"Northway Village", "lang"=>"en", "lat"=>62.977401, "lon"=>-141.903572),
"/US/US-AK/nuiqsut-686" => array("type"=>"city", "pop"=>371, "name_ru"=>"Наиксут", "name_en"=>"Nuiqsut", "lang"=>"en", "lat"=>70.191531, "lon"=>-150.994600),
"/US/US-AK/nulato-687" => array("type"=>"city", "pop"=>295, "name_ru"=>"Нулато", "name_en"=>"Nulato", "lang"=>"en", "lat"=>64.707622, "lon"=>-158.223607),
"/US/US-AK/nunapitchuk-688" => array("type"=>"city", "pop"=>467, "name_ru"=>"Нанапитчек", "name_en"=>"Nunapitchuk", "lang"=>"en", "lat"=>60.879145, "lon"=>-162.452467),
"/US/US-AK/old-harbor-689" => array("type"=>"city", "pop"=>215, "name_ru"=>"Олд Харбор", "name_en"=>"Old Harbor", "lang"=>"en", "lat"=>57.245032, "lon"=>-153.373587),
"/US/US-AK/oscarville-690" => array("type"=>"cdp", "pop"=>57, "name_ru"=>"Оскарвилл", "name_en"=>"Oscarville", "lang"=>"en", "lat"=>60.696650, "lon"=>-161.759205),
"/US/US-AK/ouzinkie-691" => array("type"=>"city", "pop"=>204, "name_ru"=>"Озинки", "name_en"=>"Ouzinkie", "lang"=>"en", "lat"=>57.935175, "lon"=>-152.458165),
"/US/US-AK/palmer-692" => array("type"=>"city", "pop"=>8201, "name_ru"=>"Палмер", "name_en"=>"Palmer", "lang"=>"en", "lat"=>61.598923, "lon"=>-149.109874),
"/US/US-AK/paxson-693" => array("type"=>"cdp", "pop"=>30, "name_ru"=>"Пакссон", "name_en"=>"Paxson", "lang"=>"en", "lat"=>63.089858, "lon"=>-145.612966),
"/US/US-AK/pedro-bay-694" => array("type"=>"cdp", "pop"=>42, "name_ru"=>"Педро Бэй", "name_en"=>"Pedro Bay", "lang"=>"en", "lat"=>59.790507, "lon"=>-154.136625),
"/US/US-AK/pelican-695" => array("type"=>"city", "pop"=>127, "name_ru"=>"Пеликан", "name_en"=>"Pelican", "lang"=>"en", "lat"=>57.953164, "lon"=>-136.209853),
"/US/US-AK/perryville-696" => array("type"=>"cdp", "pop"=>108, "name_ru"=>"Перривилл", "name_en"=>"Perryville", "lang"=>"en", "lat"=>55.958639, "lon"=>-159.226086),
"/US/US-AK/petersburg-697" => array("type"=>"city", "pop"=>2825, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>56.773998, "lon"=>-132.861623),
"/US/US-AK/pilot-point-698" => array("type"=>"cdp", "pop"=>80, "name_ru"=>"Пайлот Пойнт", "name_en"=>"Pilot Point", "lang"=>"en", "lat"=>57.619308, "lon"=>-157.459012),
"/US/US-AK/pilot-station-699" => array("type"=>"city", "pop"=>604, "name_ru"=>"Пайлот Стейшен", "name_en"=>"Pilot Station", "lang"=>"en", "lat"=>61.941425, "lon"=>-162.875852),
"/US/US-AK/pitkas-point-700" => array("type"=>"cdp", "pop"=>135, "name_ru"=>"Питкас Пойнт", "name_en"=>"Pitkas Point", "lang"=>"en", "lat"=>62.031260, "lon"=>-163.269979),
"/US/US-AK/platinum-701" => array("type"=>"city", "pop"=>41, "name_ru"=>"Платинум", "name_en"=>"Platinum", "lang"=>"en", "lat"=>58.973944, "lon"=>-161.725226),
"/US/US-AK/pleasant-valley-702" => array("type"=>"cdp", "pop"=>401, "name_ru"=>"Плезант Валли", "name_en"=>"Pleasant Valley", "lang"=>"en", "lat"=>64.881333, "lon"=>-146.868546),
"/US/US-AK/point-baker-703" => array("type"=>"cdp", "pop"=>39, "name_ru"=>"Пойнт Бейкер", "name_en"=>"Point Baker", "lang"=>"en", "lat"=>56.354710, "lon"=>-133.623640),
"/US/US-AK/point-hope-704" => array("type"=>"city", "pop"=>674, "name_ru"=>"Пойнт Хоуп", "name_en"=>"Point Hope", "lang"=>"en", "lat"=>68.311254, "lon"=>-166.727131),
"/US/US-AK/point-lay-705" => array("type"=>"cdp", "pop"=>139, "name_ru"=>"Пойнт Лей", "name_en"=>"Point Lay", "lang"=>"en", "lat"=>69.711136, "lon"=>-163.008890),
"/US/US-AK/polk-inlet-706" => array("type"=>"cdp", "pop"=>135, "name_ru"=>"Полк Инлет", "name_en"=>"Polk Inlet", "lang"=>"en", "lat"=>55.349703, "lon"=>-132.545879),
"/US/US-AK/port-alexander-707" => array("type"=>"city", "pop"=>74, "name_ru"=>"Порт Александр", "name_en"=>"Port Alexander", "lang"=>"en", "lat"=>56.218086, "lon"=>-134.630990),
"/US/US-AK/port-alice-708" => array("type"=>"cdp", "pop"=>30, "name_ru"=>"Порт Элис", "name_en"=>"Port Alice", "lang"=>"en", "lat"=>55.826492, "lon"=>-133.602830),
"/US/US-AK/port-alsworth-709" => array("type"=>"cdp", "pop"=>55, "name_ru"=>"Порт Алсворт", "name_en"=>"Port Alsworth", "lang"=>"en", "lat"=>60.209500, "lon"=>-154.304037),
"/US/US-AK/port-clarence-710" => array("type"=>"cdp", "pop"=>26, "name_ru"=>"Порт Кларенс", "name_en"=>"Port Clarence", "lang"=>"en", "lat"=>65.066264, "lon"=>-166.824454),
"/US/US-AK/port-graham-711" => array("type"=>"cdp", "pop"=>166, "name_ru"=>"Порт Грейам", "name_en"=>"Port Graham", "lang"=>"en", "lat"=>59.348525, "lon"=>-151.837369),
"/US/US-AK/port-heiden-712" => array("type"=>"city", "pop"=>95, "name_ru"=>"Порт Гейден", "name_en"=>"Port Heiden", "lang"=>"en", "lat"=>56.955113, "lon"=>-158.587784),
"/US/US-AK/port-lions-713" => array("type"=>"city", "pop"=>232, "name_ru"=>"Порт Лайонс", "name_en"=>"Port Lions", "lang"=>"en", "lat"=>57.890760, "lon"=>-152.858909),
"/US/US-AK/port-protection-714" => array("type"=>"cdp", "pop"=>62, "name_ru"=>"Порт Протекшен", "name_en"=>"Port Protection", "lang"=>"en", "lat"=>56.317143, "lon"=>-133.598063),
"/US/US-AK/primrose-715" => array("type"=>"cdp", "pop"=>63, "name_ru"=>"Примройз", "name_en"=>"Primrose", "lang"=>"en", "lat"=>60.378301, "lon"=>-149.322824),
"/US/US-AK/prudhoe-bay-716" => array("type"=>"cdp", "pop"=>47, "name_ru"=>"Прудхо Бэй", "name_en"=>"Prudhoe Bay", "lang"=>"en", "lat"=>70.284047, "lon"=>-148.376162),
"/US/US-AK/quinhagak-717" => array("type"=>"city", "pop"=>554, "name_ru"=>"Квинхагак", "name_en"=>"Quinhagak", "lang"=>"en", "lat"=>59.739381, "lon"=>-161.872373),
"/US/US-AK/rampart-718" => array("type"=>"cdp", "pop"=>68, "name_ru"=>"Рампарт", "name_en"=>"Rampart", "lang"=>"en", "lat"=>65.383627, "lon"=>-150.011201),
"/US/US-AK/red-devil-719" => array("type"=>"cdp", "pop"=>53, "name_ru"=>"Ред Девил", "name_en"=>"Red Devil", "lang"=>"en", "lat"=>61.790327, "lon"=>-157.347864),
"/US/US-AK/ridgeway-720" => array("type"=>"cdp", "pop"=>2018, "name_ru"=>"Риджуэй", "name_en"=>"Ridgeway", "lang"=>"en", "lat"=>60.532030, "lon"=>-151.085226),
"/US/US-AK/rowan-bay-721" => array("type"=>"cdp", "pop"=>133, "name_ru"=>"Рован Бэй", "name_en"=>"Rowan Bay", "lang"=>"en", "lat"=>56.653931, "lon"=>-134.251244),
"/US/US-AK/ruby-722" => array("type"=>"city", "pop"=>165, "name_ru"=>"Раби", "name_en"=>"Ruby", "lang"=>"en", "lat"=>64.717570, "lon"=>-155.520662),
"/US/US-AK/russian-mission-723" => array("type"=>"city", "pop"=>324, "name_ru"=>"Рашен Мишен", "name_en"=>"Russian Mission", "lang"=>"en", "lat"=>61.794502, "lon"=>-161.353793),
"/US/US-AK/salamatof-729" => array("type"=>"cdp", "pop"=>999, "name_ru"=>"Саламатоф", "name_en"=>"Salamatof", "lang"=>"en", "lat"=>60.615924, "lon"=>-151.337406),
"/US/US-AK/salcha-730" => array("type"=>"cdp", "pop"=>354, "name_ru"=>"Салча", "name_en"=>"Salcha", "lang"=>"en", "lat"=>64.469257, "lon"=>-146.941490),
"/US/US-AK/sand-point-731" => array("type"=>"city", "pop"=>990, "name_ru"=>"Сэнд Пойнт", "name_en"=>"Sand Point", "lang"=>"en", "lat"=>55.319459, "lon"=>-160.491183),
"/US/US-AK/savoonga-732" => array("type"=>"city", "pop"=>642, "name_ru"=>"Савунга", "name_en"=>"Savoonga", "lang"=>"en", "lat"=>63.679737, "lon"=>-170.470908),
"/US/US-AK/saxman-733" => array("type"=>"city", "pop"=>370, "name_ru"=>"Саксман", "name_en"=>"Saxman", "lang"=>"en", "lat"=>55.324617, "lon"=>-131.592159),
"/US/US-AK/scammon-bay-734" => array("type"=>"city", "pop"=>511, "name_ru"=>"Скаммон Бэй", "name_en"=>"Scammon Bay", "lang"=>"en", "lat"=>61.845019, "lon"=>-165.581945),
"/US/US-AK/selawik-735" => array("type"=>"city", "pop"=>804, "name_ru"=>"Селавик", "name_en"=>"Selawik", "lang"=>"en", "lat"=>66.603302, "lon"=>-160.025040),
"/US/US-AK/seldovia-736" => array("type"=>"city", "pop"=>309, "name_ru"=>"Селдовиа", "name_en"=>"Seldovia", "lang"=>"en", "lat"=>59.439076, "lon"=>-151.709114),
"/US/US-AK/seward-737" => array("type"=>"city", "pop"=>3126, "name_ru"=>"Сьюард", "name_en"=>"Seward", "lang"=>"en", "lat"=>60.113554, "lon"=>-149.392990),
"/US/US-AK/shageluk-738" => array("type"=>"city", "pop"=>113, "name_ru"=>"Шаджелак", "name_en"=>"Shageluk", "lang"=>"en", "lat"=>62.668392, "lon"=>-159.565135),
"/US/US-AK/shaktoolik-739" => array("type"=>"city", "pop"=>230, "name_ru"=>"Шактулик", "name_en"=>"Shaktoolik", "lang"=>"en", "lat"=>64.364030, "lon"=>-161.204206),
"/US/US-AK/sheldon-point-740" => array("type"=>"city", "pop"=>109, "name_ru"=>"Шелдон Пойнт", "name_en"=>"Sheldon Point", "lang"=>"en", "lat"=>62.516887, "lon"=>-164.892782),
"/US/US-AK/shishmaref-741" => array("type"=>"city", "pop"=>560, "name_ru"=>"Шишмарф", "name_en"=>"Shishmaref", "lang"=>"en", "lat"=>66.250656, "lon"=>-166.119539),
"/US/US-AK/shungnak-742" => array("type"=>"city", "pop"=>267, "name_ru"=>"Шангнак", "name_en"=>"Shungnak", "lang"=>"en", "lat"=>66.875901, "lon"=>-157.155702),
"/US/US-AK/sitka-743" => array("type"=>"city", "pop"=>8588, "name_ru"=>"Ситка", "name_en"=>"Sitka", "lang"=>"en", "lat"=>57.214050, "lon"=>-135.446833),
"/US/US-AK/skagway-744" => array("type"=>"city", "pop"=>892, "name_ru"=>"Скагуэй", "name_en"=>"Skagway", "lang"=>"en", "lat"=>59.520367, "lon"=>-135.335701),
"/US/US-AK/skwentna-745" => array("type"=>"cdp", "pop"=>85, "name_ru"=>"Скуэнтна", "name_en"=>"Skwentna", "lang"=>"en", "lat"=>61.990238, "lon"=>-151.397818),
"/US/US-AK/slana-746" => array("type"=>"cdp", "pop"=>63, "name_ru"=>"Слана", "name_en"=>"Slana", "lang"=>"en", "lat"=>62.701524, "lon"=>-143.993976),
"/US/US-AK/sleetmute-747" => array("type"=>"cdp", "pop"=>106, "name_ru"=>"Слитмут", "name_en"=>"Sleetmute", "lang"=>"en", "lat"=>61.721141, "lon"=>-157.183563),
"/US/US-AK/soldotna-748" => array("type"=>"city", "pop"=>4386, "name_ru"=>"Солдотна", "name_en"=>"Soldotna", "lang"=>"en", "lat"=>60.487489, "lon"=>-151.064360),
"/US/US-AK/south-naknek-749" => array("type"=>"cdp", "pop"=>136, "name_ru"=>"Сауф Нейкнек", "name_en"=>"South Naknek", "lang"=>"en", "lat"=>58.685182, "lon"=>-156.980458),
"/US/US-AK/st.-george-724" => array("type"=>"city", "pop"=>118, "name_ru"=>"Сент-Джрдж", "name_en"=>"St. George", "lang"=>"en", "lat"=>56.614741, "lon"=>-169.594616),
"/US/US-AK/st.-john-harbor-725" => array("type"=>"cdp", "pop"=>69, "name_ru"=>"Сент-Джон Харбор", "name_en"=>"St. John Harbor", "lang"=>"en", "lat"=>56.439680, "lon"=>-132.963605),
"/US/US-AK/st.-mary's-726" => array("type"=>"city", "pop"=>548, "name_ru"=>"Сент-Мэриз", "name_en"=>"St. Mary's", "lang"=>"en", "lat"=>62.046064, "lon"=>-163.246256),
"/US/US-AK/st.-michael-727" => array("type"=>"city", "pop"=>366, "name_ru"=>"Сент-Майкл", "name_en"=>"St. Michael", "lang"=>"en", "lat"=>63.477590, "lon"=>-162.109141),
"/US/US-AK/st.-paul-728" => array("type"=>"city", "pop"=>415, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>57.210529, "lon"=>-170.218936),
"/US/US-AK/stebbins-750" => array("type"=>"city", "pop"=>547, "name_ru"=>"Стеббинс", "name_en"=>"Stebbins", "lang"=>"en", "lat"=>63.478468, "lon"=>-162.227355),
"/US/US-AK/sterling-751" => array("type"=>"cdp", "pop"=>3802, "name_ru"=>"Стерлинг", "name_en"=>"Sterling", "lang"=>"en", "lat"=>60.515179, "lon"=>-150.797051),
"/US/US-AK/stevens-village-752" => array("type"=>"cdp", "pop"=>102, "name_ru"=>"Стивенс Виллидж", "name_en"=>"Stevens Village", "lang"=>"en", "lat"=>65.999046, "lon"=>-149.054652),
"/US/US-AK/stony-river-753" => array("type"=>"cdp", "pop"=>51, "name_ru"=>"Стоуни Ривер", "name_en"=>"Stony River", "lang"=>"en", "lat"=>61.793128, "lon"=>-156.580762),
"/US/US-AK/sutton-754" => array("type"=>"cdp", "pop"=>308, "name_ru"=>"Саттон", "name_en"=>"Sutton", "lang"=>"en", "lat"=>61.717255, "lon"=>-148.881229),
"/US/US-AK/takotna-755" => array("type"=>"cdp", "pop"=>38, "name_ru"=>"Такотна", "name_en"=>"Takotna", "lang"=>"en", "lat"=>62.992833, "lon"=>-156.171777),
"/US/US-AK/talkeetna-756" => array("type"=>"cdp", "pop"=>250, "name_ru"=>"Талкитна", "name_en"=>"Talkeetna", "lang"=>"en", "lat"=>62.336940, "lon"=>-150.092575),
"/US/US-AK/tanacross-757" => array("type"=>"cdp", "pop"=>106, "name_ru"=>"Танакросс", "name_en"=>"Tanacross", "lang"=>"en", "lat"=>63.339100, "lon"=>-143.432723),
"/US/US-AK/tanana-758" => array("type"=>"city", "pop"=>268, "name_ru"=>"Танана", "name_en"=>"Tanana", "lang"=>"en", "lat"=>65.155992, "lon"=>-152.101918),
"/US/US-AK/tatitlek-759" => array("type"=>"cdp", "pop"=>119, "name_ru"=>"Татитлек", "name_en"=>"Tatitlek", "lang"=>"en", "lat"=>60.883869, "lon"=>-146.677745),
"/US/US-AK/teller-760" => array("type"=>"city", "pop"=>266, "name_ru"=>"Теллер", "name_en"=>"Teller", "lang"=>"en", "lat"=>65.256049, "lon"=>-166.363030),
"/US/US-AK/tenakee-springs-761" => array("type"=>"city", "pop"=>91, "name_ru"=>"Тенейки Спрингс", "name_en"=>"Tenakee Springs", "lang"=>"en", "lat"=>57.778393, "lon"=>-135.139441),
"/US/US-AK/tetlin-762" => array("type"=>"cdp", "pop"=>87, "name_ru"=>"Тетлин", "name_en"=>"Tetlin", "lang"=>"en", "lat"=>63.064643, "lon"=>-142.438924),
"/US/US-AK/thorne-bay-763" => array("type"=>"city", "pop"=>486, "name_ru"=>"Торн Бэй", "name_en"=>"Thorne Bay", "lang"=>"en", "lat"=>55.659194, "lon"=>-132.524696),
"/US/US-AK/togiak-764" => array("type"=>"city", "pop"=>811, "name_ru"=>"Тоджиак", "name_en"=>"Togiak", "lang"=>"en", "lat"=>58.936107, "lon"=>-160.576700),
"/US/US-AK/tok-765" => array("type"=>"cdp", "pop"=>935, "name_ru"=>"Ток", "name_en"=>"Tok", "lang"=>"en", "lat"=>63.300305, "lon"=>-143.038777),
"/US/US-AK/toksook-bay-766" => array("type"=>"city", "pop"=>534, "name_ru"=>"Токсук Бэй", "name_en"=>"Toksook Bay", "lang"=>"en", "lat"=>60.493074, "lon"=>-165.094243),
"/US/US-AK/tonsina-767" => array("type"=>"cdp", "pop"=>38, "name_ru"=>"Тонсина", "name_en"=>"Tonsina", "lang"=>"en", "lat"=>61.635818, "lon"=>-145.157477),
"/US/US-AK/trapper-creek-768" => array("type"=>"cdp", "pop"=>296, "name_ru"=>"Траппер Крик", "name_en"=>"Trapper Creek", "lang"=>"en", "lat"=>62.241445, "lon"=>-150.415685),
"/US/US-AK/tuluksak-769" => array("type"=>"city", "pop"=>358, "name_ru"=>"Талуксак", "name_en"=>"Tuluksak", "lang"=>"en", "lat"=>61.115757, "lon"=>-160.939181),
"/US/US-AK/tuntutuliak-770" => array("type"=>"cdp", "pop"=>300, "name_ru"=>"Тантутулиак", "name_en"=>"Tuntutuliak", "lang"=>"en", "lat"=>60.326313, "lon"=>-162.740699),
"/US/US-AK/tununak-771" => array("type"=>"city", "pop"=>316, "name_ru"=>"Танунак", "name_en"=>"Tununak", "lang"=>"en", "lat"=>60.578904, "lon"=>-165.244762),
"/US/US-AK/twin-hills-772" => array("type"=>"cdp", "pop"=>66, "name_ru"=>"Твин Хилс", "name_en"=>"Twin Hills", "lang"=>"en", "lat"=>59.079460, "lon"=>-160.210156),
"/US/US-AK/two-rivers-773" => array("type"=>"cdp", "pop"=>453, "name_ru"=>"Ту Риверс", "name_en"=>"Two Rivers", "lang"=>"en", "lat"=>64.870994, "lon"=>-147.090009),
"/US/US-AK/tyonek-774" => array("type"=>"cdp", "pop"=>154, "name_ru"=>"Тайонк", "name_en"=>"Tyonek", "lang"=>"en", "lat"=>61.066993, "lon"=>-151.215174),
"/US/US-AK/unalakleet-775" => array("type"=>"city", "pop"=>746, "name_ru"=>"Юналаклит", "name_en"=>"Unalakleet", "lang"=>"en", "lat"=>63.884662, "lon"=>-160.789767),
"/US/US-AK/unalaska-776" => array("type"=>"city", "pop"=>3532, "name_ru"=>"УКналащка", "name_en"=>"Unalaska", "lang"=>"en", "lat"=>53.934465, "lon"=>-166.510424),
"/US/US-AK/upper-kalskag-777" => array("type"=>"city", "pop"=>230, "name_ru"=>"Аппер Калскаг", "name_en"=>"Upper Kalskag", "lang"=>"en", "lat"=>61.539126, "lon"=>-160.328950),
"/US/US-AK/valdez-778" => array("type"=>"city", "pop"=>3787, "name_ru"=>"Вальдез", "name_en"=>"Valdez", "lang"=>"en", "lat"=>61.083369, "lon"=>-146.302374),
"/US/US-AK/venetie-779" => array("type"=>"cdp", "pop"=>182, "name_ru"=>"Венети", "name_en"=>"Venetie", "lang"=>"en", "lat"=>67.014588, "lon"=>-146.397946),
"/US/US-AK/wainwright-780" => array("type"=>"city", "pop"=>485, "name_ru"=>"Уэйнрайт", "name_en"=>"Wainwright", "lang"=>"en", "lat"=>70.599953, "lon"=>-160.071563),
"/US/US-AK/wales-781" => array("type"=>"city", "pop"=>152, "name_ru"=>"Уэльс", "name_en"=>"Wales", "lang"=>"en", "lat"=>65.609251, "lon"=>-168.079924),
"/US/US-AK/wasilla-782" => array("type"=>"city", "pop"=>10256, "name_ru"=>"Уэйсилла", "name_en"=>"Wasilla", "lang"=>"en", "lat"=>61.580211, "lon"=>-149.461598),
"/US/US-AK/whale-pass-783" => array("type"=>"cdp", "pop"=>75, "name_ru"=>"Уол Пасс", "name_en"=>"Whale Pass", "lang"=>"en", "lat"=>56.103172, "lon"=>-133.172668),
"/US/US-AK/white-mountain-784" => array("type"=>"city", "pop"=>203, "name_ru"=>"Уайт Маунтин", "name_en"=>"White Mountain", "lang"=>"en", "lat"=>64.683605, "lon"=>-163.419517),
"/US/US-AK/whitestone-logging-camp-785" => array("type"=>"cdp", "pop"=>164, "name_ru"=>"Уайтстон Логгинг Кэмп", "name_en"=>"Whitestone Logging Camp", "lang"=>"en", "lat"=>58.071247, "lon"=>-135.429846),
"/US/US-AK/whittier-786" => array("type"=>"city", "pop"=>159, "name_ru"=>"Уиттиер", "name_en"=>"Whittier", "lang"=>"en", "lat"=>60.793560, "lon"=>-148.651577),
"/US/US-AK/willow-787" => array("type"=>"cdp", "pop"=>285, "name_ru"=>"Уиллоу", "name_en"=>"Willow", "lang"=>"en", "lat"=>61.740786, "lon"=>-150.056317),
"/US/US-AK/womens-bay-788" => array("type"=>"cdp", "pop"=>620, "name_ru"=>"Уименс Бэй", "name_en"=>"Womens Bay", "lang"=>"en", "lat"=>57.682308, "lon"=>-152.668720),
"/US/US-AK/wrangell-789" => array("type"=>"city", "pop"=>1990, "name_ru"=>"Рангелл", "name_en"=>"Wrangell", "lang"=>"en", "lat"=>56.361602, "lon"=>-132.271579),
"/US/US-AK/yakutat-790" => array("type"=>"city", "pop"=>534, "name_ru"=>"Якутат", "name_en"=>"Yakutat", "lang"=>"en", "lat"=>59.557526, "lon"=>-139.762121),
"/US/US-AL" => array("type"=>"state", "pop"=>4661900, "name_ru"=>"Алабама", "name_en"=>"Alabama", "lang"=>"en"),
"/US/US-AL/abbeville-1" => array("type"=>"city", "pop"=>2937, "name_ru"=>"Аббевилл", "name_en"=>"Abbeville", "lang"=>"en", "lat"=>31.567949, "lon"=>-85.253681),
"/US/US-AL/adamsville-2" => array("type"=>"city", "pop"=>4739, "name_ru"=>"Адамсвилл", "name_en"=>"Adamsville", "lang"=>"en", "lat"=>33.595542, "lon"=>-86.957993),
"/US/US-AL/addison-3" => array("type"=>"town", "pop"=>711, "name_ru"=>"Эддиссон", "name_en"=>"Addison", "lang"=>"en", "lat"=>34.199940, "lon"=>-87.177921),
"/US/US-AL/akron-4" => array("type"=>"town", "pop"=>436, "name_ru"=>"Акрон", "name_en"=>"Akron", "lang"=>"en", "lat"=>32.876118, "lon"=>-87.738779),
"/US/US-AL/alabaster-5" => array("type"=>"city", "pop"=>29352, "name_ru"=>"Алебастр", "name_en"=>"Alabaster", "lang"=>"en", "lat"=>33.226547, "lon"=>-86.824620),
"/US/US-AL/albertville-6" => array("type"=>"city", "pop"=>19794, "name_ru"=>"Альбертивилл", "name_en"=>"Albertville", "lang"=>"en", "lat"=>34.262453, "lon"=>-86.209414),
"/US/US-AL/alexander-city-7" => array("type"=>"city", "pop"=>15080, "name_ru"=>"Александр Сити", "name_en"=>"Alexander City", "lang"=>"en", "lat"=>32.927015, "lon"=>-85.934009),
"/US/US-AL/aliceville-8" => array("type"=>"city", "pop"=>2466, "name_ru"=>"Алисевилл", "name_en"=>"Aliceville", "lang"=>"en", "lat"=>33.125078, "lon"=>-88.157999),
"/US/US-AL/allgood-9" => array("type"=>"town", "pop"=>724, "name_ru"=>"Алгуд", "name_en"=>"Allgood", "lang"=>"en", "lat"=>33.906430, "lon"=>-86.518044),
"/US/US-AL/altoona-10" => array("type"=>"town", "pop"=>965, "name_ru"=>"Альтоона", "name_en"=>"Altoona", "lang"=>"en", "lat"=>34.031381, "lon"=>-86.313552),
"/US/US-AL/andalusia-11" => array("type"=>"city", "pop"=>8978, "name_ru"=>"Андалусия", "name_en"=>"Andalusia", "lang"=>"en", "lat"=>31.307638, "lon"=>-86.481440),
"/US/US-AL/anderson-12" => array("type"=>"town", "pop"=>354, "name_ru"=>"Андерсон", "name_en"=>"Anderson", "lang"=>"en", "lat"=>34.913510, "lon"=>-87.270552),
"/US/US-AL/anniston-13" => array("type"=>"city", "pop"=>23662, "name_ru"=>"Аннистон", "name_en"=>"Anniston", "lang"=>"en", "lat"=>33.662650, "lon"=>-85.813937),
"/US/US-AL/arab-14" => array("type"=>"city", "pop"=>7924, "name_ru"=>"Араб", "name_en"=>"Arab", "lang"=>"en", "lat"=>34.329635, "lon"=>-86.500127),
"/US/US-AL/ardmore-15" => array("type"=>"town", "pop"=>1249, "name_ru"=>"Ардмор", "name_en"=>"Ardmore", "lang"=>"en", "lat"=>34.983360, "lon"=>-86.836627),
"/US/US-AL/argo-16" => array("type"=>"town", "pop"=>1908, "name_ru"=>"Арго", "name_en"=>"Argo", "lang"=>"en", "lat"=>33.686661, "lon"=>-86.510030),
"/US/US-AL/ariton-17" => array("type"=>"town", "pop"=>748, "name_ru"=>"Аритон", "name_en"=>"Ariton", "lang"=>"en", "lat"=>31.598373, "lon"=>-85.715372),
"/US/US-AL/arley-18" => array("type"=>"town", "pop"=>318, "name_ru"=>"Арли", "name_en"=>"Arley", "lang"=>"en", "lat"=>34.084983, "lon"=>-87.210631),
"/US/US-AL/ashford-19" => array("type"=>"city", "pop"=>2062, "name_ru"=>"Ашфорд", "name_en"=>"Ashford", "lang"=>"en", "lat"=>31.185801, "lon"=>-85.234603),
"/US/US-AL/ashland-20" => array("type"=>"town", "pop"=>1882, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>33.270967, "lon"=>-85.830700),
"/US/US-AL/ashville-21" => array("type"=>"town", "pop"=>2564, "name_ru"=>"Ашвилл", "name_en"=>"Ashville", "lang"=>"en", "lat"=>33.834691, "lon"=>-86.264011),
"/US/US-AL/athens-22" => array("type"=>"city", "pop"=>23715, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>34.786009, "lon"=>-86.959899),
"/US/US-AL/atmore-23" => array("type"=>"city", "pop"=>7428, "name_ru"=>"Атмор", "name_en"=>"Atmore", "lang"=>"en", "lat"=>31.021694, "lon"=>-87.495438),
"/US/US-AL/attalla-24" => array("type"=>"city", "pop"=>6540, "name_ru"=>"Аттолла", "name_en"=>"Attalla", "lang"=>"en", "lat"=>34.009983, "lon"=>-86.101863),
"/US/US-AL/auburn-25" => array("type"=>"city", "pop"=>56088, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>32.592900, "lon"=>-85.480322),
"/US/US-AL/autaugaville-26" => array("type"=>"town", "pop"=>870, "name_ru"=>"Отогавилл", "name_en"=>"Autaugaville", "lang"=>"en", "lat"=>32.430993, "lon"=>-86.655798),
"/US/US-AL/avon-27" => array("type"=>"town", "pop"=>475, "name_ru"=>"Эйвон", "name_en"=>"Avon", "lang"=>"en", "lat"=>31.180293, "lon"=>-85.276624),
"/US/US-AL/babbie-28" => array("type"=>"town", "pop"=>623, "name_ru"=>"Бабби", "name_en"=>"Babbie", "lang"=>"en", "lat"=>31.300292, "lon"=>-86.319340),
"/US/US-AL/baileyton-29" => array("type"=>"town", "pop"=>720, "name_ru"=>"Бейлейтон", "name_en"=>"Baileyton", "lang"=>"en", "lat"=>34.262079, "lon"=>-86.613498),
"/US/US-AL/banks-30" => array("type"=>"town", "pop"=>221, "name_ru"=>"Бэнкс", "name_en"=>"Banks", "lang"=>"en", "lat"=>31.814278, "lon"=>-85.838337),
"/US/US-AL/bay-minette-31" => array("type"=>"city", "pop"=>8043, "name_ru"=>"Бэй Минетт", "name_en"=>"Bay Minette", "lang"=>"en", "lat"=>30.880776, "lon"=>-87.776975),
"/US/US-AL/bayou-la-batre-32" => array("type"=>"city", "pop"=>2859, "name_ru"=>"Байу Ла Батр", "name_en"=>"Bayou La Batre", "lang"=>"en", "lat"=>30.403663, "lon"=>-88.250086),
"/US/US-AL/bear-creek-33" => array("type"=>"town", "pop"=>1000, "name_ru"=>"Бэар Крик", "name_en"=>"Bear Creek", "lang"=>"en", "lat"=>34.260539, "lon"=>-87.715574),
"/US/US-AL/beatrice-34" => array("type"=>"town", "pop"=>386, "name_ru"=>"Биатрис", "name_en"=>"Beatrice", "lang"=>"en", "lat"=>31.733754, "lon"=>-87.209042),
"/US/US-AL/beaverton-35" => array("type"=>"town", "pop"=>207, "name_ru"=>"Бивертон", "name_en"=>"Beaverton", "lang"=>"en", "lat"=>33.934330, "lon"=>-88.021408),
"/US/US-AL/belk-36" => array("type"=>"town", "pop"=>211, "name_ru"=>"Белк", "name_en"=>"Belk", "lang"=>"en", "lat"=>33.648083, "lon"=>-87.927732),
"/US/US-AL/benton-37" => array("type"=>"town", "pop"=>44, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>32.307439, "lon"=>-86.817579),
"/US/US-AL/berry-38" => array("type"=>"town", "pop"=>1189, "name_ru"=>"Берри", "name_en"=>"Berry", "lang"=>"en", "lat"=>33.663185, "lon"=>-87.607693),
"/US/US-AL/bessemer-39" => array("type"=>"city", "pop"=>28542, "name_ru"=>"Бессемер", "name_en"=>"Bessemer", "lang"=>"en", "lat"=>33.375236, "lon"=>-86.960601),
"/US/US-AL/billingsley-40" => array("type"=>"town", "pop"=>122, "name_ru"=>"Биллингсли", "name_en"=>"Billingsley", "lang"=>"en", "lat"=>32.660336, "lon"=>-86.708534),
"/US/US-AL/birmingham-41" => array("type"=>"city", "pop"=>228798, "name_ru"=>"Бирмингем", "name_en"=>"Birmingham", "lang"=>"en", "lat"=>33.527746, "lon"=>-86.799223),
"/US/US-AL/black-42" => array("type"=>"town", "pop"=>208, "name_ru"=>"Блэк", "name_en"=>"Black", "lang"=>"en", "lat"=>31.009386, "lon"=>-85.743210),
"/US/US-AL/blountsville-43" => array("type"=>"town", "pop"=>1985, "name_ru"=>"Блантсвилл", "name_en"=>"Blountsville", "lang"=>"en", "lat"=>34.080871, "lon"=>-86.588414),
"/US/US-AL/blue-mountain-44" => array("type"=>"town", "pop"=>221, "name_ru"=>"Блу Маунтин", "name_en"=>"Blue Mountain", "lang"=>"en", "lat"=>33.688517, "lon"=>-85.837612),
"/US/US-AL/blue-ridge-45" => array("type"=>"cdp", "pop"=>1151, "name_ru"=>"Блу Ридж", "name_en"=>"Blue Ridge", "lang"=>"en", "lat"=>32.492484, "lon"=>-86.190617),
"/US/US-AL/blue-springs-46" => array("type"=>"town", "pop"=>115, "name_ru"=>"Блу Спрингс", "name_en"=>"Blue Springs", "lang"=>"en", "lat"=>31.660501, "lon"=>-85.503092),
"/US/US-AL/boaz-47" => array("type"=>"city", "pop"=>8349, "name_ru"=>"Бос", "name_en"=>"Boaz", "lang"=>"en", "lat"=>34.202618, "lon"=>-86.155491),
"/US/US-AL/boligee-48" => array("type"=>"town", "pop"=>345, "name_ru"=>"Болиги", "name_en"=>"Boligee", "lang"=>"en", "lat"=>32.753000, "lon"=>-88.024254),
"/US/US-AL/bon-air-49" => array("type"=>"town", "pop"=>95, "name_ru"=>"Бон Эр", "name_en"=>"Bon Air", "lang"=>"en", "lat"=>33.261319, "lon"=>-86.338052),
"/US/US-AL/branchville-50" => array("type"=>"town", "pop"=>370, "name_ru"=>"Бранчвилл", "name_en"=>"Branchville", "lang"=>"en", "lat"=>33.661120, "lon"=>-86.427038),
"/US/US-AL/brantley-51" => array("type"=>"town", "pop"=>911, "name_ru"=>"Брэнтли", "name_en"=>"Brantley", "lang"=>"en", "lat"=>31.581672, "lon"=>-86.256677),
"/US/US-AL/brent-52" => array("type"=>"city", "pop"=>4219, "name_ru"=>"Брент", "name_en"=>"Brent", "lang"=>"en", "lat"=>32.942269, "lon"=>-87.176219),
"/US/US-AL/brewton-53" => array("type"=>"city", "pop"=>5279, "name_ru"=>"Брютон", "name_en"=>"Brewton", "lang"=>"en", "lat"=>31.114611, "lon"=>-87.072837),
"/US/US-AL/bridgeport-54" => array("type"=>"city", "pop"=>2634, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>34.944425, "lon"=>-85.719003),
"/US/US-AL/brighton-55" => array("type"=>"city", "pop"=>3263, "name_ru"=>"Брайтон", "name_en"=>"Brighton", "lang"=>"en", "lat"=>33.438850, "lon"=>-86.945474),
"/US/US-AL/brilliant-56" => array("type"=>"town", "pop"=>721, "name_ru"=>"Бриллиант", "name_en"=>"Brilliant", "lang"=>"en", "lat"=>34.016723, "lon"=>-87.775810),
"/US/US-AL/brookside-57" => array("type"=>"town", "pop"=>1326, "name_ru"=>"Бруксайд", "name_en"=>"Brookside", "lang"=>"en", "lat"=>33.631931, "lon"=>-86.914741),
"/US/US-AL/brookwood-58" => array("type"=>"town", "pop"=>1449, "name_ru"=>"Бруквуд", "name_en"=>"Brookwood", "lang"=>"en", "lat"=>33.247178, "lon"=>-87.323908),
"/US/US-AL/brundidge-59" => array("type"=>"city", "pop"=>2270, "name_ru"=>"Брундидж", "name_en"=>"Brundidge", "lang"=>"en", "lat"=>31.718862, "lon"=>-85.815274),
"/US/US-AL/butler-60" => array("type"=>"city", "pop"=>1711, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>32.093844, "lon"=>-88.216721),
"/US/US-AL/bynum-61" => array("type"=>"cdp", "pop"=>1917, "name_ru"=>"Бинум", "name_en"=>"Bynum", "lang"=>"en", "lat"=>33.605546, "lon"=>-85.957464),
"/US/US-AL/cahaba-heights-62" => array("type"=>"cdp", "pop"=>4778, "name_ru"=>"Кахаба Хайтс", "name_en"=>"Cahaba Heights", "lang"=>"en", "lat"=>33.459600, "lon"=>-86.728199),
"/US/US-AL/calera-63" => array("type"=>"city", "pop"=>10742, "name_ru"=>"Калера", "name_en"=>"Calera", "lang"=>"en", "lat"=>33.101269, "lon"=>-86.753675),
"/US/US-AL/camden-64" => array("type"=>"town", "pop"=>2227, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>32.002958, "lon"=>-87.298274),
"/US/US-AL/camp-hill-65" => array("type"=>"town", "pop"=>1207, "name_ru"=>"Кэмп Хилл", "name_en"=>"Camp Hill", "lang"=>"en", "lat"=>32.800119, "lon"=>-85.653335),
"/US/US-AL/carbon-hill-66" => array("type"=>"city", "pop"=>2036, "name_ru"=>"Карбон Хилл", "name_en"=>"Carbon Hill", "lang"=>"en", "lat"=>33.895817, "lon"=>-87.525037),
"/US/US-AL/cardiff-67" => array("type"=>"town", "pop"=>80, "name_ru"=>"Кардиф", "name_en"=>"Cardiff", "lang"=>"en", "lat"=>33.646128, "lon"=>-86.928960),
"/US/US-AL/carolina-68" => array("type"=>"town", "pop"=>245, "name_ru"=>"Каролина", "name_en"=>"Carolina", "lang"=>"en", "lat"=>31.231348, "lon"=>-86.520650),
"/US/US-AL/carrollton-69" => array("type"=>"town", "pop"=>929, "name_ru"=>"Карролтон", "name_en"=>"Carrollton", "lang"=>"en", "lat"=>33.261794, "lon"=>-88.095124),
"/US/US-AL/castleberry-70" => array("type"=>"town", "pop"=>556, "name_ru"=>"Каслберри", "name_en"=>"Castleberry", "lang"=>"en", "lat"=>31.299780, "lon"=>-87.028508),
"/US/US-AL/cedar-bluff-71" => array("type"=>"town", "pop"=>1652, "name_ru"=>"Сидар Блафф", "name_en"=>"Cedar Bluff", "lang"=>"en", "lat"=>34.220512, "lon"=>-85.596037),
"/US/US-AL/center-point-72" => array("type"=>"cdp", "pop"=>15408, "name_ru"=>"Сентер Поинт", "name_en"=>"Center Point", "lang"=>"en", "lat"=>33.643133, "lon"=>-86.678764),
"/US/US-AL/centre-73" => array("type"=>"city", "pop"=>3493, "name_ru"=>"Сентер", "name_en"=>"Centre", "lang"=>"en", "lat"=>34.153021, "lon"=>-85.671733),
"/US/US-AL/centreville-74" => array("type"=>"city", "pop"=>2554, "name_ru"=>"Сентревилл", "name_en"=>"Centreville", "lang"=>"en", "lat"=>32.958890, "lon"=>-87.132713),
"/US/US-AL/chatom-75" => array("type"=>"town", "pop"=>1164, "name_ru"=>"Чатом", "name_en"=>"Chatom", "lang"=>"en", "lat"=>31.469889, "lon"=>-88.253780),
"/US/US-AL/chelsea-76" => array("type"=>"cdp", "pop"=>4303, "name_ru"=>"Челси", "name_en"=>"Chelsea", "lang"=>"en", "lat"=>33.326283, "lon"=>-86.637432),
"/US/US-AL/cherokee-77" => array("type"=>"town", "pop"=>1167, "name_ru"=>"Чероки", "name_en"=>"Cherokee", "lang"=>"en", "lat"=>34.756365, "lon"=>-87.966692),
"/US/US-AL/chickasaw-78" => array("type"=>"city", "pop"=>5948, "name_ru"=>"Чикасо", "name_en"=>"Chickasaw", "lang"=>"en", "lat"=>30.772550, "lon"=>-88.084086),
"/US/US-AL/childersburg-79" => array("type"=>"city", "pop"=>4996, "name_ru"=>"Чилдерсберг", "name_en"=>"Childersburg", "lang"=>"en", "lat"=>33.279405, "lon"=>-86.353329),
"/US/US-AL/citronelle-80" => array("type"=>"city", "pop"=>3738, "name_ru"=>"Ситронелл", "name_en"=>"Citronelle", "lang"=>"en", "lat"=>31.096088, "lon"=>-88.249861),
"/US/US-AL/clanton-81" => array("type"=>"city", "pop"=>8836, "name_ru"=>"Клантон", "name_en"=>"Clanton", "lang"=>"en", "lat"=>32.840013, "lon"=>-86.624704),
"/US/US-AL/clayhatchee-82" => array("type"=>"town", "pop"=>493, "name_ru"=>"Клейхатчи", "name_en"=>"Clayhatchee", "lang"=>"en", "lat"=>31.237570, "lon"=>-85.712745),
"/US/US-AL/clayton-83" => array("type"=>"town", "pop"=>1377, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>31.880663, "lon"=>-85.451167),
"/US/US-AL/cleveland-84" => array("type"=>"town", "pop"=>1412, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>33.994936, "lon"=>-86.570521),
"/US/US-AL/clio-85" => array("type"=>"town", "pop"=>2210, "name_ru"=>"Клио", "name_en"=>"Clio", "lang"=>"en", "lat"=>31.709276, "lon"=>-85.610549),
"/US/US-AL/coffee-springs-86" => array("type"=>"town", "pop"=>256, "name_ru"=>"Коффи Спрингс", "name_en"=>"Coffee Springs", "lang"=>"en", "lat"=>31.166567, "lon"=>-85.909898),
"/US/US-AL/coffeeville-87" => array("type"=>"town", "pop"=>344, "name_ru"=>"Коффивилл", "name_en"=>"Coffeeville", "lang"=>"en", "lat"=>31.764206, "lon"=>-88.088292),
"/US/US-AL/collinsville-88" => array("type"=>"town", "pop"=>1693, "name_ru"=>"Коллинсвилл", "name_en"=>"Collinsville", "lang"=>"en", "lat"=>34.266551, "lon"=>-85.863044),
"/US/US-AL/colony-89" => array("type"=>"town", "pop"=>408, "name_ru"=>"Колони", "name_en"=>"Colony", "lang"=>"en", "lat"=>33.941382, "lon"=>-86.901185),
"/US/US-AL/columbia-90" => array("type"=>"town", "pop"=>841, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>31.297227, "lon"=>-85.109066),
"/US/US-AL/columbiana-91" => array("type"=>"city", "pop"=>3817, "name_ru"=>"Коламбиана", "name_en"=>"Columbiana", "lang"=>"en", "lat"=>33.177005, "lon"=>-86.600822),
"/US/US-AL/coosada-92" => array("type"=>"town", "pop"=>1649, "name_ru"=>"Кузада", "name_en"=>"Coosada", "lang"=>"en", "lat"=>32.495833, "lon"=>-86.330192),
"/US/US-AL/cordova-93" => array("type"=>"city", "pop"=>2284, "name_ru"=>"Кордова", "name_en"=>"Cordova", "lang"=>"en", "lat"=>33.760650, "lon"=>-87.192757),
"/US/US-AL/cottonwood-94" => array("type"=>"town", "pop"=>1202, "name_ru"=>"Коттонвуд", "name_en"=>"Cottonwood", "lang"=>"en", "lat"=>31.055584, "lon"=>-85.301372),
"/US/US-AL/county-line-95" => array("type"=>"town", "pop"=>264, "name_ru"=>"Каунти Лайн", "name_en"=>"County Line", "lang"=>"en", "lat"=>33.820967, "lon"=>-86.722833),
"/US/US-AL/county-line-96" => array("type"=>"town", "pop"=>127, "name_ru"=>"Каунти Лайн", "name_en"=>"County Line", "lang"=>"en", "lat"=>31.284426, "lon"=>-86.200746),
"/US/US-AL/courtland-97" => array("type"=>"town", "pop"=>764, "name_ru"=>"Кортленд", "name_en"=>"Courtland", "lang"=>"en", "lat"=>34.668861, "lon"=>-87.316645),
"/US/US-AL/cowarts-98" => array("type"=>"town", "pop"=>1630, "name_ru"=>"Ковартс", "name_en"=>"Cowarts", "lang"=>"en", "lat"=>31.199398, "lon"=>-85.306340),
"/US/US-AL/creola-99" => array("type"=>"city", "pop"=>2077, "name_ru"=>"Креола", "name_en"=>"Creola", "lang"=>"en", "lat"=>30.896900, "lon"=>-88.009802),
"/US/US-AL/crossville-100" => array("type"=>"town", "pop"=>1506, "name_ru"=>"Кроссвилл", "name_en"=>"Crossville", "lang"=>"en", "lat"=>34.286278, "lon"=>-85.993551),
"/US/US-AL/cuba-101" => array("type"=>"town", "pop"=>320, "name_ru"=>"Куба", "name_en"=>"Cuba", "lang"=>"en", "lat"=>32.440602, "lon"=>-88.374257),
"/US/US-AL/cullman-102" => array("type"=>"city", "pop"=>15295, "name_ru"=>"Калман", "name_en"=>"Cullman", "lang"=>"en", "lat"=>34.176059, "lon"=>-86.838934),
"/US/US-AL/dadeville-103" => array("type"=>"city", "pop"=>3228, "name_ru"=>"Дадевилл", "name_en"=>"Dadeville", "lang"=>"en", "lat"=>32.831922, "lon"=>-85.765204),
"/US/US-AL/daleville-104" => array("type"=>"city", "pop"=>4546, "name_ru"=>"Дейлвилл", "name_en"=>"Daleville", "lang"=>"en", "lat"=>31.300833, "lon"=>-85.716212),
"/US/US-AL/daphne-105" => array("type"=>"city", "pop"=>19093, "name_ru"=>"Дафни", "name_en"=>"Daphne", "lang"=>"en", "lat"=>30.626933, "lon"=>-87.897286),
"/US/US-AL/dauphin-island-106" => array("type"=>"town", "pop"=>1586, "name_ru"=>"Дауфин Айленд", "name_en"=>"Dauphin Island", "lang"=>"en", "lat"=>30.249243, "lon"=>-88.172996),
"/US/US-AL/daviston-107" => array("type"=>"town", "pop"=>258, "name_ru"=>"Дэвистон", "name_en"=>"Daviston", "lang"=>"en", "lat"=>33.055130, "lon"=>-85.639208),
"/US/US-AL/dayton-108" => array("type"=>"town", "pop"=>57, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>32.351726, "lon"=>-87.642865),
"/US/US-AL/decatur-109" => array("type"=>"city", "pop"=>56068, "name_ru"=>"Декатур", "name_en"=>"Decatur", "lang"=>"en", "lat"=>34.571833, "lon"=>-86.988166),
"/US/US-AL/demopolis-110" => array("type"=>"city", "pop"=>7350, "name_ru"=>"Демополис", "name_en"=>"Demopolis", "lang"=>"en", "lat"=>32.500546, "lon"=>-87.834170),
"/US/US-AL/detroit-111" => array("type"=>"town", "pop"=>226, "name_ru"=>"Детройт", "name_en"=>"Detroit", "lang"=>"en", "lat"=>34.026203, "lon"=>-88.168793),
"/US/US-AL/dora-112" => array("type"=>"city", "pop"=>2422, "name_ru"=>"Дора", "name_en"=>"Dora", "lang"=>"en", "lat"=>33.731426, "lon"=>-87.083205),
"/US/US-AL/dothan-113" => array("type"=>"city", "pop"=>66505, "name_ru"=>"Дотан", "name_en"=>"Dothan", "lang"=>"en", "lat"=>31.235950, "lon"=>-85.405299),
"/US/US-AL/double-springs-114" => array("type"=>"town", "pop"=>974, "name_ru"=>"Дабл Спрингс", "name_en"=>"Double Springs", "lang"=>"en", "lat"=>34.148362, "lon"=>-87.403137),
"/US/US-AL/douglas-115" => array("type"=>"town", "pop"=>597, "name_ru"=>"Дуглас", "name_en"=>"Douglas", "lang"=>"en", "lat"=>34.171471, "lon"=>-86.321247),
"/US/US-AL/dozier-116" => array("type"=>"town", "pop"=>396, "name_ru"=>"Дозьер", "name_en"=>"Dozier", "lang"=>"en", "lat"=>31.503627, "lon"=>-86.371660),
"/US/US-AL/dutton-117" => array("type"=>"town", "pop"=>306, "name_ru"=>"Даттон", "name_en"=>"Dutton", "lang"=>"en", "lat"=>34.607554, "lon"=>-85.915588),
"/US/US-AL/east-brewton-118" => array("type"=>"city", "pop"=>2483, "name_ru"=>"Ист Брютон", "name_en"=>"East Brewton", "lang"=>"en", "lat"=>31.090580, "lon"=>-87.055310),
"/US/US-AL/eclectic-119" => array("type"=>"town", "pop"=>1147, "name_ru"=>"Эклектик", "name_en"=>"Eclectic", "lang"=>"en", "lat"=>32.636763, "lon"=>-86.034740),
"/US/US-AL/edwardsville-120" => array("type"=>"town", "pop"=>195, "name_ru"=>"Эдвардсвилл", "name_en"=>"Edwardsville", "lang"=>"en", "lat"=>33.706678, "lon"=>-85.509838),
"/US/US-AL/elba-121" => array("type"=>"city", "pop"=>4164, "name_ru"=>"Эльба", "name_en"=>"Elba", "lang"=>"en", "lat"=>31.417151, "lon"=>-86.075176),
"/US/US-AL/elberta-122" => array("type"=>"town", "pop"=>1477, "name_ru"=>"Элберта", "name_en"=>"Elberta", "lang"=>"en", "lat"=>30.413442, "lon"=>-87.600790),
"/US/US-AL/eldridge-123" => array("type"=>"town", "pop"=>173, "name_ru"=>"Элдридж", "name_en"=>"Eldridge", "lang"=>"en", "lat"=>33.921959, "lon"=>-87.619773),
"/US/US-AL/elkmont-124" => array("type"=>"town", "pop"=>546, "name_ru"=>"Элкмонт", "name_en"=>"Elkmont", "lang"=>"en", "lat"=>34.928429, "lon"=>-86.975875),
"/US/US-AL/emelle-125" => array("type"=>"town", "pop"=>28, "name_ru"=>"Эмел", "name_en"=>"Emelle", "lang"=>"en", "lat"=>32.728537, "lon"=>-88.315248),
"/US/US-AL/enterprise-126" => array("type"=>"city", "pop"=>25351, "name_ru"=>"Энтерпрайз", "name_en"=>"Enterprise", "lang"=>"en", "lat"=>31.329809, "lon"=>-85.846454),
"/US/US-AL/epes-127" => array("type"=>"town", "pop"=>186, "name_ru"=>"Эпес", "name_en"=>"Epes", "lang"=>"en", "lat"=>32.691719, "lon"=>-88.125964),
"/US/US-AL/ethelsville-128" => array("type"=>"town", "pop"=>76, "name_ru"=>"Этелсвилл", "name_en"=>"Ethelsville", "lang"=>"en", "lat"=>33.414193, "lon"=>-88.216261),
"/US/US-AL/eufaula-129" => array("type"=>"city", "pop"=>14502, "name_ru"=>"Юфола", "name_en"=>"Eufaula", "lang"=>"en", "lat"=>31.909278, "lon"=>-85.150353),
"/US/US-AL/eunola-130" => array("type"=>"town", "pop"=>199, "name_ru"=>"Юнола", "name_en"=>"Eunola", "lang"=>"en", "lat"=>31.038587, "lon"=>-85.845481),
"/US/US-AL/eutaw-131" => array("type"=>"city", "pop"=>2976, "name_ru"=>"Юто", "name_en"=>"Eutaw", "lang"=>"en", "lat"=>32.840605, "lon"=>-87.891480),
"/US/US-AL/eva-132" => array("type"=>"town", "pop"=>587, "name_ru"=>"Эва", "name_en"=>"Eva", "lang"=>"en", "lat"=>34.330415, "lon"=>-86.755855),
"/US/US-AL/evergreen-133" => array("type"=>"city", "pop"=>3465, "name_ru"=>"Эвергрин", "name_en"=>"Evergreen", "lang"=>"en", "lat"=>31.437392, "lon"=>-86.955023),
"/US/US-AL/excel-134" => array("type"=>"town", "pop"=>602, "name_ru"=>"Эксель", "name_en"=>"Excel", "lang"=>"en", "lat"=>31.426167, "lon"=>-87.342419),
"/US/US-AL/fairfield-135" => array("type"=>"city", "pop"=>11316, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>33.473000, "lon"=>-86.917717),
"/US/US-AL/fairhope-136" => array("type"=>"city", "pop"=>17147, "name_ru"=>"Фэрхоп", "name_en"=>"Fairhope", "lang"=>"en", "lat"=>30.514468, "lon"=>-87.884052),
"/US/US-AL/fairview-137" => array("type"=>"town", "pop"=>546, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>34.246790, "lon"=>-86.687441),
"/US/US-AL/falkville-138" => array("type"=>"town", "pop"=>1166, "name_ru"=>"Фолквилл", "name_en"=>"Falkville", "lang"=>"en", "lat"=>34.376935, "lon"=>-86.909720),
"/US/US-AL/faunsdale-139" => array("type"=>"town", "pop"=>82, "name_ru"=>"Фаунсдейл", "name_en"=>"Faunsdale", "lang"=>"en", "lat"=>32.459085, "lon"=>-87.593247),
"/US/US-AL/fayette-140" => array("type"=>"city", "pop"=>4778, "name_ru"=>"Файет", "name_en"=>"Fayette", "lang"=>"en", "lat"=>33.694009, "lon"=>-87.831361),
"/US/US-AL/five-points-141" => array("type"=>"town", "pop"=>135, "name_ru"=>"Файв Поинтс", "name_en"=>"Five Points", "lang"=>"en", "lat"=>33.017066, "lon"=>-85.351261),
"/US/US-AL/flint-city-142" => array("type"=>"town", "pop"=>1033, "name_ru"=>"Флинт Сити", "name_en"=>"Flint City", "lang"=>"en", "lat"=>34.515500, "lon"=>-86.972787),
"/US/US-AL/flomaton-143" => array("type"=>"town", "pop"=>1537, "name_ru"=>"Фломатон", "name_en"=>"Flomaton", "lang"=>"en", "lat"=>31.016388, "lon"=>-87.254564),
"/US/US-AL/florala-144" => array("type"=>"city", "pop"=>1884, "name_ru"=>"Флорала", "name_en"=>"Florala", "lang"=>"en", "lat"=>31.021113, "lon"=>-86.308454),
"/US/US-AL/florence-145" => array("type"=>"city", "pop"=>37877, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>34.827850, "lon"=>-87.665057),
"/US/US-AL/foley-146" => array("type"=>"city", "pop"=>13807, "name_ru"=>"Фоли", "name_en"=>"Foley", "lang"=>"en", "lat"=>30.405662, "lon"=>-87.684233),
"/US/US-AL/forestdale-147" => array("type"=>"cdp", "pop"=>10395, "name_ru"=>"Форестдейл", "name_en"=>"Forestdale", "lang"=>"en", "lat"=>33.579250, "lon"=>-86.902874),
"/US/US-AL/forkland-148" => array("type"=>"town", "pop"=>592, "name_ru"=>"Форкленд", "name_en"=>"Forkland", "lang"=>"en", "lat"=>32.647569, "lon"=>-87.867223),
"/US/US-AL/fort-deposit-149" => array("type"=>"town", "pop"=>1193, "name_ru"=>"Форт Депозит", "name_en"=>"Fort Deposit", "lang"=>"en", "lat"=>31.985854, "lon"=>-86.576236),
"/US/US-AL/fort-mcclellan-150" => array("type"=>"cdp", "pop"=>4128, "name_ru"=>"Форт Мак-Клеллан", "name_en"=>"Fort McClellan", "lang"=>"en", "lat"=>33.718825, "lon"=>-85.790720),
"/US/US-AL/fort-payne-151" => array("type"=>"city", "pop"=>14061, "name_ru"=>"Форт Пейн", "name_en"=>"Fort Payne", "lang"=>"en", "lat"=>34.453792, "lon"=>-85.694440),
"/US/US-AL/fort-rucker-152" => array("type"=>"cdp", "pop"=>7593, "name_ru"=>"Форт Ракер", "name_en"=>"Fort Rucker", "lang"=>"en", "lat"=>31.343401, "lon"=>-85.715143),
"/US/US-AL/franklin-153" => array("type"=>"town", "pop"=>145, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>32.455250, "lon"=>-85.802790),
"/US/US-AL/frisco-city-154" => array("type"=>"town", "pop"=>1344, "name_ru"=>"Фриско Сити", "name_en"=>"Frisco City", "lang"=>"en", "lat"=>31.434326, "lon"=>-87.403563),
"/US/US-AL/fruithurst-155" => array("type"=>"city", "pop"=>284, "name_ru"=>"Фрутхерст", "name_en"=>"Fruithurst", "lang"=>"en", "lat"=>33.730255, "lon"=>-85.431860),
"/US/US-AL/fulton-156" => array("type"=>"town", "pop"=>300, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>31.793557, "lon"=>-87.741064),
"/US/US-AL/fultondale-157" => array("type"=>"city", "pop"=>6918, "name_ru"=>"Фултондейл", "name_en"=>"Fultondale", "lang"=>"en", "lat"=>33.617286, "lon"=>-86.800090),
"/US/US-AL/fyffe-158" => array("type"=>"town", "pop"=>1056, "name_ru"=>"Файф", "name_en"=>"Fyffe", "lang"=>"en", "lat"=>34.450980, "lon"=>-85.906355),
"/US/US-AL/gadsden-159" => array("type"=>"city", "pop"=>36816, "name_ru"=>"Гадсден", "name_en"=>"Gadsden", "lang"=>"en", "lat"=>34.010000, "lon"=>-86.014405),
"/US/US-AL/gainesville-160" => array("type"=>"town", "pop"=>199, "name_ru"=>"Гейнсвилл", "name_en"=>"Gainesville", "lang"=>"en", "lat"=>32.815274, "lon"=>-88.159585),
"/US/US-AL/gantt-161" => array("type"=>"town", "pop"=>237, "name_ru"=>"Гантт", "name_en"=>"Gantt", "lang"=>"en", "lat"=>31.408724, "lon"=>-86.484638),
"/US/US-AL/gantts-quarry-162" => array("type"=>"town", "pop"=>7, "name_ru"=>"Гантс Куорри", "name_en"=>"Gantts Quarry", "lang"=>"en", "lat"=>33.148196, "lon"=>-86.299365),
"/US/US-AL/garden-city-163" => array("type"=>"town", "pop"=>594, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>34.014782, "lon"=>-86.750077),
"/US/US-AL/gardendale-164" => array("type"=>"city", "pop"=>13665, "name_ru"=>"Гардендейл", "name_en"=>"Gardendale", "lang"=>"en", "lat"=>33.665221, "lon"=>-86.800561),
"/US/US-AL/gaylesville-165" => array("type"=>"town", "pop"=>142, "name_ru"=>"Гейлсвилл", "name_en"=>"Gaylesville", "lang"=>"en", "lat"=>34.267902, "lon"=>-85.558283),
"/US/US-AL/geiger-166" => array("type"=>"town", "pop"=>146, "name_ru"=>"Гейгер", "name_en"=>"Geiger", "lang"=>"en", "lat"=>32.873693, "lon"=>-88.313368),
"/US/US-AL/geneva-167" => array("type"=>"city", "pop"=>4445, "name_ru"=>"Дженива", "name_en"=>"Geneva", "lang"=>"en", "lat"=>31.036639, "lon"=>-85.876779),
"/US/US-AL/georgiana-168" => array("type"=>"town", "pop"=>1565, "name_ru"=>"Джорджиана", "name_en"=>"Georgiana", "lang"=>"en", "lat"=>31.636437, "lon"=>-86.742430),
"/US/US-AL/geraldine-169" => array("type"=>"town", "pop"=>839, "name_ru"=>"Джералдин", "name_en"=>"Geraldine", "lang"=>"en", "lat"=>34.360848, "lon"=>-86.002983),
"/US/US-AL/gilbertown-170" => array("type"=>"town", "pop"=>168, "name_ru"=>"Гилбертаун", "name_en"=>"Gilbertown", "lang"=>"en", "lat"=>31.875678, "lon"=>-88.318780),
"/US/US-AL/glen-allen-171" => array("type"=>"town", "pop"=>425, "name_ru"=>"Глен Аллен", "name_en"=>"Glen Allen", "lang"=>"en", "lat"=>33.884827, "lon"=>-87.744118),
"/US/US-AL/glencoe-172" => array("type"=>"city", "pop"=>5377, "name_ru"=>"Гленко", "name_en"=>"Glencoe", "lang"=>"en", "lat"=>33.946233, "lon"=>-85.931225),
"/US/US-AL/glenwood-173" => array("type"=>"town", "pop"=>194, "name_ru"=>"Гленвуд", "name_en"=>"Glenwood", "lang"=>"en", "lat"=>31.666077, "lon"=>-86.172433),
"/US/US-AL/goldville-174" => array("type"=>"town", "pop"=>36, "name_ru"=>"Голдвилл", "name_en"=>"Goldville", "lang"=>"en", "lat"=>33.083497, "lon"=>-85.784431),
"/US/US-AL/good-hope-175" => array("type"=>"town", "pop"=>2079, "name_ru"=>"Гуд Хоп", "name_en"=>"Good Hope", "lang"=>"en", "lat"=>34.110133, "lon"=>-86.864567),
"/US/US-AL/goodwater-176" => array("type"=>"city", "pop"=>1521, "name_ru"=>"Гудуотер", "name_en"=>"Goodwater", "lang"=>"en", "lat"=>33.060438, "lon"=>-86.051507),
"/US/US-AL/gordo-177" => array("type"=>"city", "pop"=>1557, "name_ru"=>"Гордо", "name_en"=>"Gordo", "lang"=>"en", "lat"=>33.321344, "lon"=>-87.903711),
"/US/US-AL/gordon-178" => array("type"=>"town", "pop"=>419, "name_ru"=>"Гордон", "name_en"=>"Gordon", "lang"=>"en", "lat"=>31.145641, "lon"=>-85.094350),
"/US/US-AL/goshen-179" => array("type"=>"town", "pop"=>296, "name_ru"=>"Гошен", "name_en"=>"Goshen", "lang"=>"en", "lat"=>31.719558, "lon"=>-86.125906),
"/US/US-AL/grand-bay-180" => array("type"=>"cdp", "pop"=>3383, "name_ru"=>"Гренд Бэй", "name_en"=>"Grand Bay", "lang"=>"en", "lat"=>30.471944, "lon"=>-88.344293),
"/US/US-AL/grant-181" => array("type"=>"town", "pop"=>709, "name_ru"=>"Грант", "name_en"=>"Grant", "lang"=>"en", "lat"=>34.517212, "lon"=>-86.252270),
"/US/US-AL/graysville-182" => array("type"=>"city", "pop"=>2374, "name_ru"=>"Грейсвилл", "name_en"=>"Graysville", "lang"=>"en", "lat"=>33.626400, "lon"=>-86.969199),
"/US/US-AL/greensboro-183" => array("type"=>"city", "pop"=>2564, "name_ru"=>"Гринсборо", "name_en"=>"Greensboro", "lang"=>"en", "lat"=>32.700618, "lon"=>-87.594062),
"/US/US-AL/greenville-184" => array("type"=>"city", "pop"=>7022, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>31.829719, "lon"=>-86.626699),
"/US/US-AL/grimes-185" => array("type"=>"town", "pop"=>453, "name_ru"=>"Граймс", "name_en"=>"Grimes", "lang"=>"en", "lat"=>31.302097, "lon"=>-85.451108),
"/US/US-AL/grove-hill-186" => array("type"=>"town", "pop"=>1342, "name_ru"=>"Гров Хилл", "name_en"=>"Grove Hill", "lang"=>"en", "lat"=>31.703793, "lon"=>-87.777428),
"/US/US-AL/gu-win-191" => array("type"=>"town", "pop"=>193, "name_ru"=>"Гу-Вин", "name_en"=>"Gu-Win", "lang"=>"en", "lat"=>33.958234, "lon"=>-87.871663),
"/US/US-AL/guin-187" => array("type"=>"town", "pop"=>2189, "name_ru"=>"Гин", "name_en"=>"Guin", "lang"=>"en", "lat"=>33.966090, "lon"=>-87.910549),
"/US/US-AL/gulf-shores-188" => array("type"=>"city", "pop"=>10248, "name_ru"=>"Галф Шорс", "name_en"=>"Gulf Shores", "lang"=>"en", "lat"=>30.270036, "lon"=>-87.701460),
"/US/US-AL/guntersville-189" => array("type"=>"city", "pop"=>8422, "name_ru"=>"Гантерсвилл", "name_en"=>"Guntersville", "lang"=>"en", "lat"=>34.356757, "lon"=>-86.270409),
"/US/US-AL/gurley-190" => array("type"=>"town", "pop"=>849, "name_ru"=>"Герли", "name_en"=>"Gurley", "lang"=>"en", "lat"=>34.703145, "lon"=>-86.373057),
"/US/US-AL/hackleburg-192" => array("type"=>"town", "pop"=>1448, "name_ru"=>"Хаклберг", "name_en"=>"Hackleburg", "lang"=>"en", "lat"=>34.274181, "lon"=>-87.833697),
"/US/US-AL/haleburg-193" => array("type"=>"town", "pop"=>109, "name_ru"=>"Халберг", "name_en"=>"Haleburg", "lang"=>"en", "lat"=>31.410858, "lon"=>-85.138674),
"/US/US-AL/haleyville-194" => array("type"=>"city", "pop"=>4047, "name_ru"=>"Халивилл", "name_en"=>"Haleyville", "lang"=>"en", "lat"=>34.238792, "lon"=>-87.618201),
"/US/US-AL/hamilton-195" => array("type"=>"city", "pop"=>6393, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>34.140314, "lon"=>-87.982764),
"/US/US-AL/hammondville-196" => array("type"=>"town", "pop"=>541, "name_ru"=>"Хаммондвилл", "name_en"=>"Hammondville", "lang"=>"en", "lat"=>34.569330, "lon"=>-85.638339),
"/US/US-AL/hanceville-197" => array("type"=>"city", "pop"=>3421, "name_ru"=>"Хансевилл", "name_en"=>"Hanceville", "lang"=>"en", "lat"=>34.064911, "lon"=>-86.763469),
"/US/US-AL/harpersville-198" => array("type"=>"town", "pop"=>1712, "name_ru"=>"Харперсвилл", "name_en"=>"Harpersville", "lang"=>"en", "lat"=>33.341799, "lon"=>-86.432691),
"/US/US-AL/hartford-199" => array("type"=>"city", "pop"=>2426, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>31.105115, "lon"=>-85.691343),
"/US/US-AL/hartselle-200" => array("type"=>"city", "pop"=>13888, "name_ru"=>"Хартсел", "name_en"=>"Hartselle", "lang"=>"en", "lat"=>34.436916, "lon"=>-86.939815),
"/US/US-AL/harvest-201" => array("type"=>"cdp", "pop"=>1922, "name_ru"=>"Харвест", "name_en"=>"Harvest", "lang"=>"en", "lat"=>34.855619, "lon"=>-86.749894),
"/US/US-AL/hayden-202" => array("type"=>"town", "pop"=>539, "name_ru"=>"Хейден", "name_en"=>"Hayden", "lang"=>"en", "lat"=>33.895328, "lon"=>-86.754358),
"/US/US-AL/hayneville-203" => array("type"=>"town", "pop"=>1122, "name_ru"=>"Хейнвилл", "name_en"=>"Hayneville", "lang"=>"en", "lat"=>32.182473, "lon"=>-86.578835),
"/US/US-AL/hazel-green-204" => array("type"=>"cdp", "pop"=>2208, "name_ru"=>"Хазел Грин", "name_en"=>"Hazel Green", "lang"=>"en", "lat"=>34.923624, "lon"=>-86.567213),
"/US/US-AL/headland-205" => array("type"=>"city", "pop"=>4019, "name_ru"=>"Хедленд", "name_en"=>"Headland", "lang"=>"en", "lat"=>31.351932, "lon"=>-85.343253),
"/US/US-AL/heath-206" => array("type"=>"town", "pop"=>247, "name_ru"=>"Хит", "name_en"=>"Heath", "lang"=>"en", "lat"=>31.355101, "lon"=>-86.466878),
"/US/US-AL/heflin-207" => array("type"=>"city", "pop"=>3551, "name_ru"=>"Хефлин", "name_en"=>"Heflin", "lang"=>"en", "lat"=>33.646285, "lon"=>-85.586732),
"/US/US-AL/helena-208" => array("type"=>"city", "pop"=>14639, "name_ru"=>"Хелена", "name_en"=>"Helena", "lang"=>"en", "lat"=>33.289736, "lon"=>-86.852423),
"/US/US-AL/henagar-209" => array("type"=>"town", "pop"=>2579, "name_ru"=>"Хенагар", "name_en"=>"Henagar", "lang"=>"en", "lat"=>34.633923, "lon"=>-85.737447),
"/US/US-AL/highland-lake-210" => array("type"=>"town", "pop"=>500, "name_ru"=>"Хайленд Лейк", "name_en"=>"Highland Lake", "lang"=>"en", "lat"=>33.889234, "lon"=>-86.423815),
"/US/US-AL/hillsboro-211" => array("type"=>"town", "pop"=>591, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>34.636721, "lon"=>-87.184584),
"/US/US-AL/hobson-city-212" => array("type"=>"town", "pop"=>866, "name_ru"=>"Хобсон Сити", "name_en"=>"Hobson City", "lang"=>"en", "lat"=>33.617033, "lon"=>-85.851727),
"/US/US-AL/hodges-213" => array("type"=>"town", "pop"=>261, "name_ru"=>"Ходжес", "name_en"=>"Hodges", "lang"=>"en", "lat"=>34.327387, "lon"=>-87.926532),
"/US/US-AL/hokes-bluff-214" => array("type"=>"city", "pop"=>4450, "name_ru"=>"Хокс Блафф", "name_en"=>"Hokes Bluff", "lang"=>"en", "lat"=>33.989650, "lon"=>-85.862970),
"/US/US-AL/holly-pond-215" => array("type"=>"town", "pop"=>728, "name_ru"=>"Холли Понд", "name_en"=>"Holly Pond", "lang"=>"en", "lat"=>34.174561, "lon"=>-86.617011),
"/US/US-AL/hollywood-216" => array("type"=>"town", "pop"=>924, "name_ru"=>"Холливуд", "name_en"=>"Hollywood", "lang"=>"en", "lat"=>34.716877, "lon"=>-85.965711),
"/US/US-AL/holt-217" => array("type"=>"cdp", "pop"=>4125, "name_ru"=>"Холт", "name_en"=>"Holt", "lang"=>"en", "lat"=>33.229550, "lon"=>-87.480949),
"/US/US-AL/homewood-218" => array("type"=>"city", "pop"=>23840, "name_ru"=>"Хомвуд", "name_en"=>"Homewood", "lang"=>"en", "lat"=>33.464800, "lon"=>-86.806999),
"/US/US-AL/hoover-219" => array("type"=>"city", "pop"=>71020, "name_ru"=>"Хувер", "name_en"=>"Hoover", "lang"=>"en", "lat"=>33.375332, "lon"=>-86.819037),
"/US/US-AL/hueytown-220" => array("type"=>"city", "pop"=>15751, "name_ru"=>"Хаейтаун", "name_en"=>"Hueytown", "lang"=>"en", "lat"=>33.437600, "lon"=>-86.997586),
"/US/US-AL/huguley-221" => array("type"=>"cdp", "pop"=>3161, "name_ru"=>"Хагали", "name_en"=>"Huguley", "lang"=>"en", "lat"=>32.843119, "lon"=>-85.240673),
"/US/US-AL/huntsville-222" => array("type"=>"city", "pop"=>176645, "name_ru"=>"Хантсвилл", "name_en"=>"Huntsville", "lang"=>"en", "lat"=>34.707048, "lon"=>-86.627723),
"/US/US-AL/hurtsboro-223" => array("type"=>"town", "pop"=>546, "name_ru"=>"Хертсборо", "name_en"=>"Hurtsboro", "lang"=>"en", "lat"=>32.239942, "lon"=>-85.414371),
"/US/US-AL/ider-224" => array("type"=>"town", "pop"=>715, "name_ru"=>"Айдер", "name_en"=>"Ider", "lang"=>"en", "lat"=>34.700676, "lon"=>-85.666937),
"/US/US-AL/inverness-225" => array("type"=>"cdp", "pop"=>2528, "name_ru"=>"Инвернесс", "name_en"=>"Inverness", "lang"=>"en", "lat"=>33.404139, "lon"=>-86.722944),
"/US/US-AL/irondale-226" => array("type"=>"city", "pop"=>9677, "name_ru"=>"Айрондейл", "name_en"=>"Irondale", "lang"=>"en", "lat"=>33.532317, "lon"=>-86.676303),
"/US/US-AL/jackson-227" => array("type"=>"city", "pop"=>5101, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>31.532540, "lon"=>-87.893012),
"/US/US-AL/jacksons'-gap-228" => array("type"=>"town", "pop"=>748, "name_ru"=>"Джэксонс Гэп", "name_en"=>"Jacksons' Gap", "lang"=>"en", "lat"=>32.884143, "lon"=>-85.816998),
"/US/US-AL/jacksonville-229" => array("type"=>"city", "pop"=>10077, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>33.810650, "lon"=>-85.760774),
"/US/US-AL/jasper-230" => array("type"=>"city", "pop"=>14131, "name_ru"=>"Джаспер", "name_en"=>"Jasper", "lang"=>"en", "lat"=>33.852042, "lon"=>-87.270067),
"/US/US-AL/jemison-231" => array("type"=>"town", "pop"=>2957, "name_ru"=>"Джемисон", "name_en"=>"Jemison", "lang"=>"en", "lat"=>32.964998, "lon"=>-86.744068),
"/US/US-AL/kansas-232" => array("type"=>"town", "pop"=>255, "name_ru"=>"Канзас", "name_en"=>"Kansas", "lang"=>"en", "lat"=>33.901778, "lon"=>-87.553343),
"/US/US-AL/kennedy-233" => array("type"=>"town", "pop"=>494, "name_ru"=>"Кеннеди", "name_en"=>"Kennedy", "lang"=>"en", "lat"=>33.581421, "lon"=>-87.985506),
"/US/US-AL/killen-234" => array("type"=>"town", "pop"=>1142, "name_ru"=>"Киллен", "name_en"=>"Killen", "lang"=>"en", "lat"=>34.861501, "lon"=>-87.529349),
"/US/US-AL/kimberly-235" => array("type"=>"town", "pop"=>2767, "name_ru"=>"Кимберли", "name_en"=>"Kimberly", "lang"=>"en", "lat"=>33.768213, "lon"=>-86.804802),
"/US/US-AL/kinsey-236" => array("type"=>"town", "pop"=>1965, "name_ru"=>"Кинси", "name_en"=>"Kinsey", "lang"=>"en", "lat"=>31.297409, "lon"=>-85.335936),
"/US/US-AL/kinston-237" => array("type"=>"town", "pop"=>614, "name_ru"=>"Кинстон", "name_en"=>"Kinston", "lang"=>"en", "lat"=>31.219426, "lon"=>-86.169941),
"/US/US-AL/ladonia-238" => array("type"=>"cdp", "pop"=>2905, "name_ru"=>"Ладония", "name_en"=>"Ladonia", "lang"=>"en", "lat"=>32.462379, "lon"=>-85.086805),
"/US/US-AL/lafayette-239" => array("type"=>"city", "pop"=>3151, "name_ru"=>"Ла Файет", "name_en"=>"Lafayette", "lang"=>"en", "lat"=>32.899542, "lon"=>-85.400846),
"/US/US-AL/lake-purdy-240" => array("type"=>"cdp", "pop"=>1840, "name_ru"=>"Лейк Перди", "name_en"=>"Lake Purdy", "lang"=>"en", "lat"=>33.429931, "lon"=>-86.680504),
"/US/US-AL/lakeview-241" => array("type"=>"town", "pop"=>174, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>34.392209, "lon"=>-85.973270),
"/US/US-AL/lanett-242" => array("type"=>"city", "pop"=>7121, "name_ru"=>"Ланетт", "name_en"=>"Lanett", "lang"=>"en", "lat"=>32.866205, "lon"=>-85.205117),
"/US/US-AL/langston-243" => array("type"=>"town", "pop"=>252, "name_ru"=>"Лангстон", "name_en"=>"Langston", "lang"=>"en", "lat"=>34.534187, "lon"=>-86.092361),
"/US/US-AL/leeds-244" => array("type"=>"city", "pop"=>11312, "name_ru"=>"Лидс", "name_en"=>"Leeds", "lang"=>"en", "lat"=>33.545976, "lon"=>-86.562745),
"/US/US-AL/leesburg-245" => array("type"=>"town", "pop"=>821, "name_ru"=>"Лисберг", "name_en"=>"Leesburg", "lang"=>"en", "lat"=>34.178976, "lon"=>-85.769101),
"/US/US-AL/leighton-246" => array("type"=>"town", "pop"=>830, "name_ru"=>"Лейтон", "name_en"=>"Leighton", "lang"=>"en", "lat"=>34.699553, "lon"=>-87.530673),
"/US/US-AL/lester-247" => array("type"=>"town", "pop"=>120, "name_ru"=>"Лестер", "name_en"=>"Lester", "lang"=>"en", "lat"=>34.984221, "lon"=>-87.147896),
"/US/US-AL/level-plains-248" => array("type"=>"town", "pop"=>1511, "name_ru"=>"Левел Плейнс", "name_en"=>"Level Plains", "lang"=>"en", "lat"=>31.306357, "lon"=>-85.766718),
"/US/US-AL/lexington-249" => array("type"=>"town", "pop"=>843, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>34.966032, "lon"=>-87.372872),
"/US/US-AL/libertyville-250" => array("type"=>"town", "pop"=>104, "name_ru"=>"Либертивилл", "name_en"=>"Libertyville", "lang"=>"en", "lat"=>31.243672, "lon"=>-86.460007),
"/US/US-AL/lincoln-251" => array("type"=>"city", "pop"=>5846, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>33.598010, "lon"=>-86.139391),
"/US/US-AL/linden-252" => array("type"=>"city", "pop"=>2255, "name_ru"=>"Линден", "name_en"=>"Linden", "lang"=>"en", "lat"=>32.301011, "lon"=>-87.792706),
"/US/US-AL/lineville-253" => array("type"=>"town", "pop"=>2374, "name_ru"=>"Лайнвилл", "name_en"=>"Lineville", "lang"=>"en", "lat"=>33.312421, "lon"=>-85.752616),
"/US/US-AL/lipscomb-254" => array("type"=>"city", "pop"=>2278, "name_ru"=>"Липскомб", "name_en"=>"Lipscomb", "lang"=>"en", "lat"=>33.426100, "lon"=>-86.925422),
"/US/US-AL/lisman-255" => array("type"=>"town", "pop"=>574, "name_ru"=>"Лисман", "name_en"=>"Lisman", "lang"=>"en", "lat"=>32.169306, "lon"=>-88.288072),
"/US/US-AL/littleville-256" => array("type"=>"town", "pop"=>1036, "name_ru"=>"Литлвилл", "name_en"=>"Littleville", "lang"=>"en", "lat"=>34.595005, "lon"=>-87.674038),
"/US/US-AL/livingston-257" => array("type"=>"city", "pop"=>3020, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>32.598810, "lon"=>-88.187144),
"/US/US-AL/loachapoka-258" => array("type"=>"town", "pop"=>162, "name_ru"=>"Лочапока", "name_en"=>"Loachapoka", "lang"=>"en", "lat"=>32.604706, "lon"=>-85.596938),
"/US/US-AL/lockhart-259" => array("type"=>"town", "pop"=>539, "name_ru"=>"Локхарт", "name_en"=>"Lockhart", "lang"=>"en", "lat"=>31.012731, "lon"=>-86.350243),
"/US/US-AL/locust-fork-260" => array("type"=>"town", "pop"=>1176, "name_ru"=>"Локуст Форк", "name_en"=>"Locust Fork", "lang"=>"en", "lat"=>33.902650, "lon"=>-86.621682),
"/US/US-AL/louisville-261" => array("type"=>"town", "pop"=>565, "name_ru"=>"Луисвилл", "name_en"=>"Louisville", "lang"=>"en", "lat"=>31.781088, "lon"=>-85.558042),
"/US/US-AL/lowndesboro-262" => array("type"=>"town", "pop"=>126, "name_ru"=>"Лоундсборо", "name_en"=>"Lowndesboro", "lang"=>"en", "lat"=>32.276737, "lon"=>-86.610257),
"/US/US-AL/loxley-263" => array("type"=>"town", "pop"=>1796, "name_ru"=>"Локсли", "name_en"=>"Loxley", "lang"=>"en", "lat"=>30.627129, "lon"=>-87.756379),
"/US/US-AL/luverne-264" => array("type"=>"city", "pop"=>2760, "name_ru"=>"Леверн", "name_en"=>"Luverne", "lang"=>"en", "lat"=>31.707854, "lon"=>-86.261088),
"/US/US-AL/lynn-265" => array("type"=>"town", "pop"=>712, "name_ru"=>"Линн", "name_en"=>"Lynn", "lang"=>"en", "lat"=>34.043769, "lon"=>-87.546695),
"/US/US-AL/madison-269" => array("type"=>"city", "pop"=>38714, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>34.702904, "lon"=>-86.749679),
"/US/US-AL/madrid-270" => array("type"=>"town", "pop"=>333, "name_ru"=>"Мадрид", "name_en"=>"Madrid", "lang"=>"en", "lat"=>31.035769, "lon"=>-85.397282),
"/US/US-AL/malvern-271" => array("type"=>"town", "pop"=>1222, "name_ru"=>"Малверн", "name_en"=>"Malvern", "lang"=>"en", "lat"=>31.135872, "lon"=>-85.519440),
"/US/US-AL/maplesville-272" => array("type"=>"town", "pop"=>690, "name_ru"=>"Мейплсвилл", "name_en"=>"Maplesville", "lang"=>"en", "lat"=>32.785716, "lon"=>-86.878395),
"/US/US-AL/margaret-273" => array("type"=>"town", "pop"=>1307, "name_ru"=>"Маргарет", "name_en"=>"Margaret", "lang"=>"en", "lat"=>33.687702, "lon"=>-86.474660),
"/US/US-AL/marion-274" => array("type"=>"city", "pop"=>3290, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>32.631553, "lon"=>-87.317598),
"/US/US-AL/maytown-275" => array("type"=>"town", "pop"=>429, "name_ru"=>"Мейтаун", "name_en"=>"Maytown", "lang"=>"en", "lat"=>33.551128, "lon"=>-86.995717),
"/US/US-AL/mcintosh-266" => array("type"=>"town", "pop"=>234, "name_ru"=>"Макинтош", "name_en"=>"McIntosh", "lang"=>"en", "lat"=>31.264812, "lon"=>-88.027882),
"/US/US-AL/mckenzie-267" => array("type"=>"town", "pop"=>606, "name_ru"=>"Маккензи", "name_en"=>"McKenzie", "lang"=>"en", "lat"=>31.543754, "lon"=>-86.716532),
"/US/US-AL/mcmullen-268" => array("type"=>"town", "pop"=>60, "name_ru"=>"Мак-Муллен", "name_en"=>"McMullen", "lang"=>"en", "lat"=>33.147105, "lon"=>-88.176123),
"/US/US-AL/meadowbrook-276" => array("type"=>"cdp", "pop"=>4621, "name_ru"=>"Медоубрук", "name_en"=>"Meadowbrook", "lang"=>"en", "lat"=>33.401928, "lon"=>-86.696801),
"/US/US-AL/memphis-277" => array("type"=>"town", "pop"=>31, "name_ru"=>"Мемфис", "name_en"=>"Memphis", "lang"=>"en", "lat"=>33.138332, "lon"=>-88.301314),
"/US/US-AL/mentone-278" => array("type"=>"town", "pop"=>482, "name_ru"=>"Ментон", "name_en"=>"Mentone", "lang"=>"en", "lat"=>34.568309, "lon"=>-85.582702),
"/US/US-AL/meridianville-279" => array("type"=>"cdp", "pop"=>2852, "name_ru"=>"Меридианвилл", "name_en"=>"Meridianville", "lang"=>"en", "lat"=>34.873760, "lon"=>-86.570935),
"/US/US-AL/midfield-280" => array("type"=>"city", "pop"=>5185, "name_ru"=>"Мидфилд", "name_en"=>"Midfield", "lang"=>"en", "lat"=>33.454100, "lon"=>-86.922808),
"/US/US-AL/midland-city-281" => array("type"=>"town", "pop"=>1883, "name_ru"=>"Мидленд Сити", "name_en"=>"Midland City", "lang"=>"en", "lat"=>31.316275, "lon"=>-85.493382),
"/US/US-AL/midway-282" => array("type"=>"town", "pop"=>419, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>32.074296, "lon"=>-85.520286),
"/US/US-AL/mignon-283" => array("type"=>"cdp", "pop"=>1548, "name_ru"=>"Мигнон", "name_en"=>"Mignon", "lang"=>"en", "lat"=>33.194043, "lon"=>-86.264008),
"/US/US-AL/millbrook-284" => array("type"=>"city", "pop"=>16640, "name_ru"=>"Милбрук", "name_en"=>"Millbrook", "lang"=>"en", "lat"=>32.494600, "lon"=>-86.367721),
"/US/US-AL/millport-285" => array("type"=>"town", "pop"=>1014, "name_ru"=>"Милпорт", "name_en"=>"Millport", "lang"=>"en", "lat"=>33.559467, "lon"=>-88.081194),
"/US/US-AL/millry-286" => array("type"=>"town", "pop"=>592, "name_ru"=>"Милри", "name_en"=>"Millry", "lang"=>"en", "lat"=>31.625865, "lon"=>-88.316669),
"/US/US-AL/minor-287" => array("type"=>"cdp", "pop"=>3313, "name_ru"=>"Майнор", "name_en"=>"Minor", "lang"=>"en", "lat"=>33.525300, "lon"=>-86.947612),
"/US/US-AL/mobile-288" => array("type"=>"city", "pop"=>191022, "name_ru"=>"Мобил", "name_en"=>"Mobile", "lang"=>"en", "lat"=>30.677450, "lon"=>-88.088959),
"/US/US-AL/monroeville-289" => array("type"=>"city", "pop"=>6372, "name_ru"=>"Монровилл", "name_en"=>"Monroeville", "lang"=>"en", "lat"=>31.516885, "lon"=>-87.327813),
"/US/US-AL/montevallo-290" => array("type"=>"city", "pop"=>6061, "name_ru"=>"Монтевалло", "name_en"=>"Montevallo", "lang"=>"en", "lat"=>33.109251, "lon"=>-86.851224),
"/US/US-AL/montgomery-291" => array("type"=>"city", "pop"=>202696, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>32.354400, "lon"=>-86.284287),
"/US/US-AL/moody-292" => array("type"=>"town", "pop"=>13400, "name_ru"=>"Муди", "name_en"=>"Moody", "lang"=>"en", "lat"=>33.596934, "lon"=>-86.494628),
"/US/US-AL/moores-mill-293" => array("type"=>"cdp", "pop"=>3362, "name_ru"=>"Мурс Милл", "name_en"=>"Moores Mill", "lang"=>"en", "lat"=>34.851888, "lon"=>-86.525755),
"/US/US-AL/mooresville-294" => array("type"=>"town", "pop"=>65, "name_ru"=>"Мурсвилл", "name_en"=>"Mooresville", "lang"=>"en", "lat"=>34.626832, "lon"=>-86.879986),
"/US/US-AL/morris-295" => array("type"=>"town", "pop"=>1888, "name_ru"=>"Моррис", "name_en"=>"Morris", "lang"=>"en", "lat"=>33.749733, "lon"=>-86.805340),
"/US/US-AL/mosses-296" => array("type"=>"town", "pop"=>1040, "name_ru"=>"Моссес", "name_en"=>"Mosses", "lang"=>"en", "lat"=>32.176151, "lon"=>-86.674589),
"/US/US-AL/moulton-297" => array("type"=>"city", "pop"=>3267, "name_ru"=>"Молтон", "name_en"=>"Moulton", "lang"=>"en", "lat"=>34.483899, "lon"=>-87.281139),
"/US/US-AL/moundville-298" => array("type"=>"town", "pop"=>2597, "name_ru"=>"Маундвилл", "name_en"=>"Moundville", "lang"=>"en", "lat"=>32.995882, "lon"=>-87.630660),
"/US/US-AL/mount-vernon-301" => array("type"=>"town", "pop"=>815, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>31.091175, "lon"=>-88.011747),
"/US/US-AL/mountain-brook-300" => array("type"=>"city", "pop"=>21061, "name_ru"=>"Маунтин Брук", "name_en"=>"Mountain Brook", "lang"=>"en", "lat"=>33.486867, "lon"=>-86.740477),
"/US/US-AL/mountainboro-299" => array("type"=>"town", "pop"=>331, "name_ru"=>"Маунтинборо", "name_en"=>"Mountainboro", "lang"=>"en", "lat"=>34.147710, "lon"=>-86.131208),
"/US/US-AL/mulga-302" => array("type"=>"town", "pop"=>889, "name_ru"=>"Малга", "name_en"=>"Mulga", "lang"=>"en", "lat"=>33.546587, "lon"=>-86.975349),
"/US/US-AL/muscle-shoals-303" => array("type"=>"city", "pop"=>13032, "name_ru"=>"Маскл Шолс", "name_en"=>"Muscle Shoals", "lang"=>"en", "lat"=>34.747117, "lon"=>-87.642039),
"/US/US-AL/myrtlewood-304" => array("type"=>"town", "pop"=>130, "name_ru"=>"Миртлвуд", "name_en"=>"Myrtlewood", "lang"=>"en", "lat"=>32.247111, "lon"=>-87.947127),
"/US/US-AL/napier-field-305" => array("type"=>"town", "pop"=>396, "name_ru"=>"Нейпир Филд", "name_en"=>"Napier Field", "lang"=>"en", "lat"=>31.315115, "lon"=>-85.455306),
"/US/US-AL/nauvoo-306" => array("type"=>"town", "pop"=>280, "name_ru"=>"Науву", "name_en"=>"Nauvoo", "lang"=>"en", "lat"=>33.988466, "lon"=>-87.487800),
"/US/US-AL/nectar-307" => array("type"=>"town", "pop"=>423, "name_ru"=>"Нектар", "name_en"=>"Nectar", "lang"=>"en", "lat"=>33.969026, "lon"=>-86.636264),
"/US/US-AL/needham-308" => array("type"=>"town", "pop"=>87, "name_ru"=>"Нидхэм", "name_en"=>"Needham", "lang"=>"en", "lat"=>31.986189, "lon"=>-88.338772),
"/US/US-AL/new-brockton-310" => array("type"=>"town", "pop"=>1234, "name_ru"=>"Нью Броктон", "name_en"=>"New Brockton", "lang"=>"en", "lat"=>31.380970, "lon"=>-85.924398),
"/US/US-AL/new-hope-311" => array("type"=>"town", "pop"=>2755, "name_ru"=>"Нью Хоуп", "name_en"=>"New Hope", "lang"=>"en", "lat"=>34.539140, "lon"=>-86.415896),
"/US/US-AL/new-market-312" => array("type"=>"cdp", "pop"=>1094, "name_ru"=>"Нью Маркет", "name_en"=>"New Market", "lang"=>"en", "lat"=>34.909465, "lon"=>-86.425117),
"/US/US-AL/new-site-313" => array("type"=>"town", "pop"=>831, "name_ru"=>"Нью Сайт", "name_en"=>"New Site", "lang"=>"en", "lat"=>33.035225, "lon"=>-85.775012),
"/US/US-AL/newbern-309" => array("type"=>"town", "pop"=>225, "name_ru"=>"Ньюберн", "name_en"=>"Newbern", "lang"=>"en", "lat"=>32.591251, "lon"=>-87.535972),
"/US/US-AL/newton-314" => array("type"=>"town", "pop"=>1653, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>31.339126, "lon"=>-85.588578),
"/US/US-AL/newville-315" => array("type"=>"town", "pop"=>543, "name_ru"=>"Ньювилл", "name_en"=>"Newville", "lang"=>"en", "lat"=>31.421670, "lon"=>-85.336500),
"/US/US-AL/north-courtland-316" => array("type"=>"town", "pop"=>798, "name_ru"=>"Норт Кортленд", "name_en"=>"North Courtland", "lang"=>"en", "lat"=>34.677186, "lon"=>-87.308365),
"/US/US-AL/north-johns-317" => array("type"=>"town", "pop"=>131, "name_ru"=>"Норт Джонс", "name_en"=>"North Johns", "lang"=>"en", "lat"=>33.368112, "lon"=>-87.101540),
"/US/US-AL/northport-318" => array("type"=>"city", "pop"=>23118, "name_ru"=>"Нортпорт", "name_en"=>"Northport", "lang"=>"en", "lat"=>33.244111, "lon"=>-87.592149),
"/US/US-AL/notasulga-319" => array("type"=>"town", "pop"=>845, "name_ru"=>"Нотасулга", "name_en"=>"Notasulga", "lang"=>"en", "lat"=>32.557554, "lon"=>-85.672449),
"/US/US-AL/oak-grove-320" => array("type"=>"town", "pop"=>465, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>33.186178, "lon"=>-86.302855),
"/US/US-AL/oak-hill-321" => array("type"=>"town", "pop"=>36, "name_ru"=>"Оук Хилл", "name_en"=>"Oak Hill", "lang"=>"en", "lat"=>31.922139, "lon"=>-87.088129),
"/US/US-AL/oakman-322" => array("type"=>"town", "pop"=>926, "name_ru"=>"Оукман", "name_en"=>"Oakman", "lang"=>"en", "lat"=>33.712982, "lon"=>-87.388470),
"/US/US-AL/odenville-323" => array("type"=>"town", "pop"=>2257, "name_ru"=>"Оденвилл", "name_en"=>"Odenville", "lang"=>"en", "lat"=>33.686021, "lon"=>-86.398370),
"/US/US-AL/ohatchee-324" => array("type"=>"town", "pop"=>1243, "name_ru"=>"Охатчи", "name_en"=>"Ohatchee", "lang"=>"en", "lat"=>33.784483, "lon"=>-86.018720),
"/US/US-AL/oneonta-325" => array("type"=>"city", "pop"=>7040, "name_ru"=>"Онеонта", "name_en"=>"Oneonta", "lang"=>"en", "lat"=>33.951165, "lon"=>-86.470595),
"/US/US-AL/onycha-326" => array("type"=>"town", "pop"=>207, "name_ru"=>"Онича", "name_en"=>"Onycha", "lang"=>"en", "lat"=>31.224328, "lon"=>-86.279091),
"/US/US-AL/opelika-327" => array("type"=>"city", "pop"=>26808, "name_ru"=>"Опелика", "name_en"=>"Opelika", "lang"=>"en", "lat"=>32.654200, "lon"=>-85.378805),
"/US/US-AL/opp-328" => array("type"=>"city", "pop"=>6582, "name_ru"=>"Опп", "name_en"=>"Opp", "lang"=>"en", "lat"=>31.291399, "lon"=>-86.255997),
"/US/US-AL/orange-beach-329" => array("type"=>"town", "pop"=>6231, "name_ru"=>"Ориндж Бич", "name_en"=>"Orange Beach", "lang"=>"en", "lat"=>30.284442, "lon"=>-87.590994),
"/US/US-AL/orrville-330" => array("type"=>"town", "pop"=>210, "name_ru"=>"Орвилл", "name_en"=>"Orrville", "lang"=>"en", "lat"=>32.308320, "lon"=>-87.245408),
"/US/US-AL/owens-crossroads-331" => array("type"=>"town", "pop"=>695, "name_ru"=>"Оуэнс Кросс Роудс", "name_en"=>"Owens Crossroads", "lang"=>"en", "lat"=>34.582017, "lon"=>-86.460163),
"/US/US-AL/oxford-332" => array("type"=>"city", "pop"=>20622, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>33.597030, "lon"=>-85.825836),
"/US/US-AL/ozark-333" => array("type"=>"city", "pop"=>14711, "name_ru"=>"Озарк", "name_en"=>"Ozark", "lang"=>"en", "lat"=>31.439127, "lon"=>-85.648755),
"/US/US-AL/paint-rock-334" => array("type"=>"town", "pop"=>182, "name_ru"=>"Пейнт Рок", "name_en"=>"Paint Rock", "lang"=>"en", "lat"=>34.660576, "lon"=>-86.328777),
"/US/US-AL/parrish-335" => array("type"=>"town", "pop"=>1246, "name_ru"=>"Парриш", "name_en"=>"Parrish", "lang"=>"en", "lat"=>33.733833, "lon"=>-87.280708),
"/US/US-AL/pelham-336" => array("type"=>"city", "pop"=>21266, "name_ru"=>"Пелхэм", "name_en"=>"Pelham", "lang"=>"en", "lat"=>33.303280, "lon"=>-86.794051),
"/US/US-AL/pell-city-337" => array("type"=>"city", "pop"=>12841, "name_ru"=>"Пелл Сити", "name_en"=>"Pell City", "lang"=>"en", "lat"=>33.560168, "lon"=>-86.272437),
"/US/US-AL/pennington-338" => array("type"=>"town", "pop"=>319, "name_ru"=>"Пеннингтон", "name_en"=>"Pennington", "lang"=>"en", "lat"=>32.196798, "lon"=>-88.052363),
"/US/US-AL/petrey-339" => array("type"=>"town", "pop"=>64, "name_ru"=>"Петри", "name_en"=>"Petrey", "lang"=>"en", "lat"=>31.849064, "lon"=>-86.207270),
"/US/US-AL/phenix-city-340" => array("type"=>"city", "pop"=>31125, "name_ru"=>"Феникс Сити", "name_en"=>"Phenix City", "lang"=>"en", "lat"=>32.465509, "lon"=>-85.014610),
"/US/US-AL/phil-campbell-341" => array("type"=>"town", "pop"=>1052, "name_ru"=>"Фил Кэмпбелл", "name_en"=>"Phil Campbell", "lang"=>"en", "lat"=>34.352028, "lon"=>-87.705810),
"/US/US-AL/pickensville-342" => array("type"=>"town", "pop"=>630, "name_ru"=>"Пиккенсвилл", "name_en"=>"Pickensville", "lang"=>"en", "lat"=>33.226646, "lon"=>-88.265574),
"/US/US-AL/piedmont-343" => array("type"=>"city", "pop"=>4964, "name_ru"=>"Пидмонт", "name_en"=>"Piedmont", "lang"=>"en", "lat"=>33.927382, "lon"=>-85.615204),
"/US/US-AL/pinckard-344" => array("type"=>"town", "pop"=>627, "name_ru"=>"Пинккард", "name_en"=>"Pinckard", "lang"=>"en", "lat"=>31.320588, "lon"=>-85.544246),
"/US/US-AL/pine-apple-345" => array("type"=>"town", "pop"=>305, "name_ru"=>"Пайн Эппл", "name_en"=>"Pine Apple", "lang"=>"en", "lat"=>31.873728, "lon"=>-86.987682),
"/US/US-AL/pine-hill-346" => array("type"=>"town", "pop"=>912, "name_ru"=>"Пайн Хилл", "name_en"=>"Pine Hill", "lang"=>"en", "lat"=>31.985636, "lon"=>-87.582518),
"/US/US-AL/pine-ridge-347" => array("type"=>"town", "pop"=>260, "name_ru"=>"Пайн Ридж", "name_en"=>"Pine Ridge", "lang"=>"en", "lat"=>34.444390, "lon"=>-85.779667),
"/US/US-AL/pinson-clay-chalkville-348" => array("type"=>"cdp", "pop"=>10987, "name_ru"=>"Пинсон-Клэй-Чолквилл", "name_en"=>"Pinson-Clay-Chalkville", "lang"=>"en", "lat"=>33.676696, "lon"=>-86.658635),
"/US/US-AL/pisgah-349" => array("type"=>"town", "pop"=>696, "name_ru"=>"Фасги", "name_en"=>"Pisgah", "lang"=>"en", "lat"=>34.682152, "lon"=>-85.847759),
"/US/US-AL/pleasant-grove-350" => array("type"=>"city", "pop"=>10254, "name_ru"=>"Плезант Гров", "name_en"=>"Pleasant Grove", "lang"=>"en", "lat"=>33.489168, "lon"=>-86.975939),
"/US/US-AL/point-clear-351" => array("type"=>"cdp", "pop"=>2125, "name_ru"=>"Пойнт Клир", "name_en"=>"Point Clear", "lang"=>"en", "lat"=>30.489634, "lon"=>-87.904391),
"/US/US-AL/pollard-352" => array("type"=>"town", "pop"=>118, "name_ru"=>"Поллард", "name_en"=>"Pollard", "lang"=>"en", "lat"=>31.022531, "lon"=>-87.172430),
"/US/US-AL/powell-353" => array("type"=>"town", "pop"=>977, "name_ru"=>"Пауэлл", "name_en"=>"Powell", "lang"=>"en", "lat"=>34.530771, "lon"=>-85.895102),
"/US/US-AL/prattville-354" => array("type"=>"city", "pop"=>32480, "name_ru"=>"Праттвилл", "name_en"=>"Prattville", "lang"=>"en", "lat"=>32.461700, "lon"=>-86.446691),
"/US/US-AL/priceville-355" => array("type"=>"town", "pop"=>2814, "name_ru"=>"Прайсвилл", "name_en"=>"Priceville", "lang"=>"en", "lat"=>34.524219, "lon"=>-86.885583),
"/US/US-AL/prichard-356" => array("type"=>"city", "pop"=>27662, "name_ru"=>"Причард", "name_en"=>"Prichard", "lang"=>"en", "lat"=>30.773100, "lon"=>-88.130673),
"/US/US-AL/providence-357" => array("type"=>"town", "pop"=>295, "name_ru"=>"Провиденс", "name_en"=>"Providence", "lang"=>"en", "lat"=>32.344452, "lon"=>-87.772103),
"/US/US-AL/ragland-358" => array("type"=>"town", "pop"=>2136, "name_ru"=>"Рагленд", "name_en"=>"Ragland", "lang"=>"en", "lat"=>33.748065, "lon"=>-86.136302),
"/US/US-AL/rainbow-city-359" => array("type"=>"city", "pop"=>9324, "name_ru"=>"Рейнбоу Сити", "name_en"=>"Rainbow City", "lang"=>"en", "lat"=>33.929750, "lon"=>-86.087374),
"/US/US-AL/rainsville-360" => array("type"=>"city", "pop"=>4990, "name_ru"=>"Рейнсвилл", "name_en"=>"Rainsville", "lang"=>"en", "lat"=>34.498776, "lon"=>-85.844933),
"/US/US-AL/ranburne-361" => array("type"=>"town", "pop"=>484, "name_ru"=>"Ранберн", "name_en"=>"Ranburne", "lang"=>"en", "lat"=>33.524874, "lon"=>-85.340155),
"/US/US-AL/red-bay-362" => array("type"=>"city", "pop"=>3296, "name_ru"=>"Ред Бэй", "name_en"=>"Red Bay", "lang"=>"en", "lat"=>34.436126, "lon"=>-88.136336),
"/US/US-AL/red-level-363" => array("type"=>"town", "pop"=>548, "name_ru"=>"Ред Левел", "name_en"=>"Red Level", "lang"=>"en", "lat"=>31.404179, "lon"=>-86.608199),
"/US/US-AL/redstone-arsenal-364" => array("type"=>"cdp", "pop"=>4909, "name_ru"=>"Редстоун Арсенал", "name_en"=>"Redstone Arsenal", "lang"=>"en", "lat"=>34.684075, "lon"=>-86.654036),
"/US/US-AL/reece-city-365" => array("type"=>"town", "pop"=>613, "name_ru"=>"Рис Сити", "name_en"=>"Reece City", "lang"=>"en", "lat"=>34.072433, "lon"=>-86.032368),
"/US/US-AL/reform-366" => array("type"=>"city", "pop"=>1790, "name_ru"=>"Реформ", "name_en"=>"Reform", "lang"=>"en", "lat"=>33.379561, "lon"=>-88.016697),
"/US/US-AL/repton-367" => array("type"=>"town", "pop"=>263, "name_ru"=>"Рептон", "name_en"=>"Repton", "lang"=>"en", "lat"=>31.408806, "lon"=>-87.239761),
"/US/US-AL/ridgeville-368" => array("type"=>"town", "pop"=>150, "name_ru"=>"Риджвилл", "name_en"=>"Ridgeville", "lang"=>"en", "lat"=>34.057014, "lon"=>-86.103258),
"/US/US-AL/river-falls-369" => array("type"=>"town", "pop"=>605, "name_ru"=>"Ривер Фолс", "name_en"=>"River Falls", "lang"=>"en", "lat"=>31.350363, "lon"=>-86.541895),
"/US/US-AL/riverside-370" => array("type"=>"town", "pop"=>2025, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>33.611281, "lon"=>-86.207701),
"/US/US-AL/riverview-371" => array("type"=>"town", "pop"=>98, "name_ru"=>"Ривервью", "name_en"=>"Riverview", "lang"=>"en", "lat"=>31.058462, "lon"=>-87.056720),
"/US/US-AL/roanoke-372" => array("type"=>"city", "pop"=>6688, "name_ru"=>"Роанок", "name_en"=>"Roanoke", "lang"=>"en", "lat"=>33.145335, "lon"=>-85.369345),
"/US/US-AL/robertsdale-373" => array("type"=>"city", "pop"=>4964, "name_ru"=>"Робертсдейл", "name_en"=>"Robertsdale", "lang"=>"en", "lat"=>30.552139, "lon"=>-87.707803),
"/US/US-AL/rockford-374" => array("type"=>"town", "pop"=>396, "name_ru"=>"Рокфорд", "name_en"=>"Rockford", "lang"=>"en", "lat"=>32.887588, "lon"=>-86.217227),
"/US/US-AL/rogersville-375" => array("type"=>"town", "pop"=>1204, "name_ru"=>"Роджерсвилл", "name_en"=>"Rogersville", "lang"=>"en", "lat"=>34.826241, "lon"=>-87.284514),
"/US/US-AL/rosa-376" => array("type"=>"town", "pop"=>356, "name_ru"=>"Роза", "name_en"=>"Rosa", "lang"=>"en", "lat"=>33.984540, "lon"=>-86.508688),
"/US/US-AL/russellville-377" => array("type"=>"city", "pop"=>8918, "name_ru"=>"Расселлвилл", "name_en"=>"Russellville", "lang"=>"en", "lat"=>34.505909, "lon"=>-87.729164),
"/US/US-AL/rutledge-378" => array("type"=>"town", "pop"=>509, "name_ru"=>"Рутледж", "name_en"=>"Rutledge", "lang"=>"en", "lat"=>31.729032, "lon"=>-86.301687),
"/US/US-AL/saks-380" => array("type"=>"cdp", "pop"=>11138, "name_ru"=>"Сакс", "name_en"=>"Saks", "lang"=>"en", "lat"=>33.711290, "lon"=>-85.853760),
"/US/US-AL/samson-381" => array("type"=>"city", "pop"=>2029, "name_ru"=>"Самсон", "name_en"=>"Samson", "lang"=>"en", "lat"=>31.112210, "lon"=>-86.046604),
"/US/US-AL/sand-rock-382" => array("type"=>"town", "pop"=>521, "name_ru"=>"Сэнд Рок", "name_en"=>"Sand Rock", "lang"=>"en", "lat"=>34.232126, "lon"=>-85.770626),
"/US/US-AL/sanford-383" => array("type"=>"town", "pop"=>268, "name_ru"=>"Санфорд", "name_en"=>"Sanford", "lang"=>"en", "lat"=>31.297556, "lon"=>-86.401696),
"/US/US-AL/saraland-384" => array("type"=>"city", "pop"=>12946, "name_ru"=>"Сараленд", "name_en"=>"Saraland", "lang"=>"en", "lat"=>30.820093, "lon"=>-88.098927),
"/US/US-AL/sardis-city-385" => array("type"=>"town", "pop"=>2121, "name_ru"=>"Сардис Сити", "name_en"=>"Sardis City", "lang"=>"en", "lat"=>34.176717, "lon"=>-86.113538),
"/US/US-AL/satsuma-386" => array("type"=>"city", "pop"=>6008, "name_ru"=>"Сатсума", "name_en"=>"Satsuma", "lang"=>"en", "lat"=>30.854100, "lon"=>-88.061099),
"/US/US-AL/scottsboro-387" => array("type"=>"city", "pop"=>14994, "name_ru"=>"Скотсборо", "name_en"=>"Scottsboro", "lang"=>"en", "lat"=>34.649346, "lon"=>-86.047376),
"/US/US-AL/section-388" => array("type"=>"town", "pop"=>759, "name_ru"=>"Секшен", "name_en"=>"Section", "lang"=>"en", "lat"=>34.581000, "lon"=>-85.983823),
"/US/US-AL/selma-389" => array("type"=>"city", "pop"=>18847, "name_ru"=>"Сельма", "name_en"=>"Selma", "lang"=>"en", "lat"=>32.416084, "lon"=>-87.032952),
"/US/US-AL/selmont-west-selmont-390" => array("type"=>"cdp", "pop"=>3823, "name_ru"=>"Сельмонт-Вэст Сельмонт", "name_en"=>"Selmont-West Selmont", "lang"=>"en", "lat"=>32.378357, "lon"=>-87.006673),
"/US/US-AL/sheffield-391" => array("type"=>"city", "pop"=>9127, "name_ru"=>"Шеффилд", "name_en"=>"Sheffield", "lang"=>"en", "lat"=>34.755250, "lon"=>-87.701261),
"/US/US-AL/shiloh-392" => array("type"=>"town", "pop"=>309, "name_ru"=>"Шилох", "name_en"=>"Shiloh", "lang"=>"en", "lat"=>34.465468, "lon"=>-85.877339),
"/US/US-AL/shorter-393" => array("type"=>"town", "pop"=>374, "name_ru"=>"Шортер", "name_en"=>"Shorter", "lang"=>"en", "lat"=>32.400754, "lon"=>-85.936607),
"/US/US-AL/silas-394" => array("type"=>"town", "pop"=>475, "name_ru"=>"Силас", "name_en"=>"Silas", "lang"=>"en", "lat"=>31.769209, "lon"=>-88.320906),
"/US/US-AL/silverhill-395" => array("type"=>"town", "pop"=>698, "name_ru"=>"Сильверхилл", "name_en"=>"Silverhill", "lang"=>"en", "lat"=>30.545174, "lon"=>-87.749023),
"/US/US-AL/sipsey-396" => array("type"=>"town", "pop"=>543, "name_ru"=>"Сипси", "name_en"=>"Sipsey", "lang"=>"en", "lat"=>33.822068, "lon"=>-87.083053),
"/US/US-AL/skyline-397" => array("type"=>"town", "pop"=>835, "name_ru"=>"Скайлайн", "name_en"=>"Skyline", "lang"=>"en", "lat"=>34.802861, "lon"=>-86.123512),
"/US/US-AL/slocomb-398" => array("type"=>"city", "pop"=>2047, "name_ru"=>"Слокомб", "name_en"=>"Slocomb", "lang"=>"en", "lat"=>31.110254, "lon"=>-85.594335),
"/US/US-AL/smiths-399" => array("type"=>"cdp", "pop"=>3456, "name_ru"=>"Смитс", "name_en"=>"Smiths", "lang"=>"en", "lat"=>32.539794, "lon"=>-85.087215),
"/US/US-AL/smoke-rise-400" => array("type"=>"cdp", "pop"=>1367, "name_ru"=>"Смоук Райз", "name_en"=>"Smoke Rise", "lang"=>"en", "lat"=>33.879865, "lon"=>-86.824469),
"/US/US-AL/snead-401" => array("type"=>"town", "pop"=>848, "name_ru"=>"Снид", "name_en"=>"Snead", "lang"=>"en", "lat"=>34.115874, "lon"=>-86.391528),
"/US/US-AL/somerville-402" => array("type"=>"town", "pop"=>505, "name_ru"=>"Сомервилл", "name_en"=>"Somerville", "lang"=>"en", "lat"=>34.474706, "lon"=>-86.801372),
"/US/US-AL/south-vinemont-404" => array("type"=>"town", "pop"=>515, "name_ru"=>"Саут Вайнмонт", "name_en"=>"South Vinemont", "lang"=>"en", "lat"=>34.236140, "lon"=>-86.862592),
"/US/US-AL/southside-403" => array("type"=>"city", "pop"=>8407, "name_ru"=>"Саутсайд", "name_en"=>"Southside", "lang"=>"en", "lat"=>33.903767, "lon"=>-86.021986),
"/US/US-AL/spanish-fort-405" => array("type"=>"cdp", "pop"=>5780, "name_ru"=>"Спэниш Форт", "name_en"=>"Spanish Fort", "lang"=>"en", "lat"=>30.675242, "lon"=>-87.884589),
"/US/US-AL/springville-406" => array("type"=>"town", "pop"=>3667, "name_ru"=>"Спрингвилл", "name_en"=>"Springville", "lang"=>"en", "lat"=>33.773623, "lon"=>-86.475147),
"/US/US-AL/st.-florian-379" => array("type"=>"town", "pop"=>474, "name_ru"=>"Св. Флориан", "name_en"=>"St. Florian", "lang"=>"en", "lat"=>34.865650, "lon"=>-87.623827),
"/US/US-AL/steele-407" => array("type"=>"town", "pop"=>1221, "name_ru"=>"Стил", "name_en"=>"Steele", "lang"=>"en", "lat"=>33.934000, "lon"=>-86.197259),
"/US/US-AL/stevenson-408" => array("type"=>"city", "pop"=>2022, "name_ru"=>"Стивенсон", "name_en"=>"Stevenson", "lang"=>"en", "lat"=>34.870608, "lon"=>-85.833098),
"/US/US-AL/sulligent-409" => array("type"=>"city", "pop"=>1968, "name_ru"=>"Салиджент", "name_en"=>"Sulligent", "lang"=>"en", "lat"=>33.894700, "lon"=>-88.131888),
"/US/US-AL/sumiton-410" => array("type"=>"city", "pop"=>2564, "name_ru"=>"Самитон", "name_en"=>"Sumiton", "lang"=>"en", "lat"=>33.753076, "lon"=>-87.048167),
"/US/US-AL/summerdale-411" => array("type"=>"town", "pop"=>745, "name_ru"=>"Саммердейл", "name_en"=>"Summerdale", "lang"=>"en", "lat"=>30.483286, "lon"=>-87.700361),
"/US/US-AL/susan-moore-412" => array("type"=>"town", "pop"=>822, "name_ru"=>"Сьюзан Мур", "name_en"=>"Susan Moore", "lang"=>"en", "lat"=>34.074059, "lon"=>-86.429739),
"/US/US-AL/sweet-water-413" => array("type"=>"town", "pop"=>220, "name_ru"=>"Свит Уотер", "name_en"=>"Sweet Water", "lang"=>"en", "lat"=>32.101690, "lon"=>-87.867328),
"/US/US-AL/sylacauga-414" => array("type"=>"city", "pop"=>12843, "name_ru"=>"Силакога", "name_en"=>"Sylacauga", "lang"=>"en", "lat"=>33.175958, "lon"=>-86.261318),
"/US/US-AL/sylvan-springs-416" => array("type"=>"town", "pop"=>1516, "name_ru"=>"Сильван Спрингс", "name_en"=>"Sylvan Springs", "lang"=>"en", "lat"=>33.512896, "lon"=>-87.030561),
"/US/US-AL/sylvania-415" => array("type"=>"town", "pop"=>1267, "name_ru"=>"Сильвания", "name_en"=>"Sylvania", "lang"=>"en", "lat"=>34.559790, "lon"=>-85.807570),
"/US/US-AL/talladega-417" => array("type"=>"city", "pop"=>16943, "name_ru"=>"Талладега", "name_en"=>"Talladega", "lang"=>"en", "lat"=>33.432048, "lon"=>-86.098001),
"/US/US-AL/talladega-springs-418" => array("type"=>"town", "pop"=>123, "name_ru"=>"Талладега Спрингс", "name_en"=>"Talladega Springs", "lang"=>"en", "lat"=>33.117523, "lon"=>-86.444543),
"/US/US-AL/tallassee-419" => array("type"=>"city", "pop"=>5172, "name_ru"=>"Талласси", "name_en"=>"Tallassee", "lang"=>"en", "lat"=>32.535160, "lon"=>-85.891439),
"/US/US-AL/tarrant-420" => array("type"=>"city", "pop"=>6499, "name_ru"=>"Таррант", "name_en"=>"Tarrant", "lang"=>"en", "lat"=>33.593700, "lon"=>-86.767888),
"/US/US-AL/taylor-421" => array("type"=>"town", "pop"=>1993, "name_ru"=>"Тейлор", "name_en"=>"Taylor", "lang"=>"en", "lat"=>31.164467, "lon"=>-85.467580),
"/US/US-AL/theodore-422" => array("type"=>"cdp", "pop"=>6509, "name_ru"=>"Теодор", "name_en"=>"Theodore", "lang"=>"en", "lat"=>30.545539, "lon"=>-88.176647),
"/US/US-AL/thomaston-423" => array("type"=>"town", "pop"=>363, "name_ru"=>"Томастон", "name_en"=>"Thomaston", "lang"=>"en", "lat"=>32.268081, "lon"=>-87.623610),
"/US/US-AL/thomasville-424" => array("type"=>"city", "pop"=>4478, "name_ru"=>"Томасвилл", "name_en"=>"Thomasville", "lang"=>"en", "lat"=>31.912389, "lon"=>-87.741895),
"/US/US-AL/thorsby-425" => array("type"=>"town", "pop"=>2069, "name_ru"=>"Торсби", "name_en"=>"Thorsby", "lang"=>"en", "lat"=>32.914998, "lon"=>-86.714133),
"/US/US-AL/tillmans-corner-426" => array("type"=>"cdp", "pop"=>17988, "name_ru"=>"Тилманс Корнер", "name_en"=>"Tillmans Corner", "lang"=>"en", "lat"=>30.585517, "lon"=>-88.191340),
"/US/US-AL/town-creek-427" => array("type"=>"town", "pop"=>1209, "name_ru"=>"Таун Крик", "name_en"=>"Town Creek", "lang"=>"en", "lat"=>34.672665, "lon"=>-87.408315),
"/US/US-AL/toxey-428" => array("type"=>"town", "pop"=>136, "name_ru"=>"Токсей", "name_en"=>"Toxey", "lang"=>"en", "lat"=>31.910129, "lon"=>-88.308050),
"/US/US-AL/trafford-429" => array("type"=>"town", "pop"=>495, "name_ru"=>"Траффорд", "name_en"=>"Trafford", "lang"=>"en", "lat"=>33.823647, "lon"=>-86.746489),
"/US/US-AL/triana-430" => array("type"=>"town", "pop"=>490, "name_ru"=>"Триана", "name_en"=>"Triana", "lang"=>"en", "lat"=>34.594573, "lon"=>-86.744271),
"/US/US-AL/trinity-431" => array("type"=>"town", "pop"=>1976, "name_ru"=>"Тринити", "name_en"=>"Trinity", "lang"=>"en", "lat"=>34.598597, "lon"=>-87.089825),
"/US/US-AL/troy-432" => array("type"=>"city", "pop"=>15148, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>31.804118, "lon"=>-85.968163),
"/US/US-AL/trussville-433" => array("type"=>"city", "pop"=>18985, "name_ru"=>"Трассвилл", "name_en"=>"Trussville", "lang"=>"en", "lat"=>33.635929, "lon"=>-86.593213),
"/US/US-AL/tuscaloosa-434" => array("type"=>"city", "pop"=>90221, "name_ru"=>"Тускалуча", "name_en"=>"Tuscaloosa", "lang"=>"en", "lat"=>33.237700, "lon"=>-87.540979),
"/US/US-AL/tuscumbia-435" => array("type"=>"city", "pop"=>8308, "name_ru"=>"Тускумбия", "name_en"=>"Tuscumbia", "lang"=>"en", "lat"=>34.728667, "lon"=>-87.705652),
"/US/US-AL/tuskegee-436" => array("type"=>"city", "pop"=>11357, "name_ru"=>"Таскджи", "name_en"=>"Tuskegee", "lang"=>"en", "lat"=>32.437748, "lon"=>-85.711143),
"/US/US-AL/underwood-petersville-437" => array("type"=>"cdp", "pop"=>3092, "name_ru"=>"Андервуд-Питерсвилл", "name_en"=>"Underwood-Petersville", "lang"=>"en", "lat"=>34.876969, "lon"=>-87.696010),
"/US/US-AL/union-438" => array("type"=>"town", "pop"=>213, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>32.993425, "lon"=>-87.905500),
"/US/US-AL/union-grove-439" => array("type"=>"town", "pop"=>99, "name_ru"=>"Юнион Гров", "name_en"=>"Union Grove", "lang"=>"en", "lat"=>34.399996, "lon"=>-86.446062),
"/US/US-AL/union-springs-440" => array("type"=>"city", "pop"=>4702, "name_ru"=>"Юнион Спрингс", "name_en"=>"Union Springs", "lang"=>"en", "lat"=>32.138811, "lon"=>-85.713660),
"/US/US-AL/uniontown-441" => array("type"=>"city", "pop"=>1418, "name_ru"=>"Юнионтаун", "name_en"=>"Uniontown", "lang"=>"en", "lat"=>32.447029, "lon"=>-87.508459),
"/US/US-AL/valley-442" => array("type"=>"city", "pop"=>9962, "name_ru"=>"Валли", "name_en"=>"Valley", "lang"=>"en", "lat"=>32.809037, "lon"=>-85.173016),
"/US/US-AL/valley-head-443" => array("type"=>"town", "pop"=>650, "name_ru"=>"Валли Хед", "name_en"=>"Valley Head", "lang"=>"en", "lat"=>34.550803, "lon"=>-85.625819),
"/US/US-AL/vance-444" => array("type"=>"town", "pop"=>945, "name_ru"=>"Ванс", "name_en"=>"Vance", "lang"=>"en", "lat"=>33.170926, "lon"=>-87.227909),
"/US/US-AL/vernon-445" => array("type"=>"city", "pop"=>1893, "name_ru"=>"Вернон", "name_en"=>"Vernon", "lang"=>"en", "lat"=>33.758330, "lon"=>-88.113754),
"/US/US-AL/vestavia-hills-446" => array("type"=>"city", "pop"=>30892, "name_ru"=>"Веставиа Хилс", "name_en"=>"Vestavia Hills", "lang"=>"en", "lat"=>33.433567, "lon"=>-86.787559),
"/US/US-AL/vina-447" => array("type"=>"town", "pop"=>402, "name_ru"=>"Вина", "name_en"=>"Vina", "lang"=>"en", "lat"=>34.374267, "lon"=>-88.056546),
"/US/US-AL/vincent-448" => array("type"=>"town", "pop"=>2006, "name_ru"=>"Винсент", "name_en"=>"Vincent", "lang"=>"en", "lat"=>33.385610, "lon"=>-86.409940),
"/US/US-AL/vredenburgh-449" => array("type"=>"town", "pop"=>307, "name_ru"=>"Вреденбург", "name_en"=>"Vredenburgh", "lang"=>"en", "lat"=>31.823284, "lon"=>-87.317533),
"/US/US-AL/wadley-450" => array("type"=>"town", "pop"=>647, "name_ru"=>"Уодли", "name_en"=>"Wadley", "lang"=>"en", "lat"=>33.123346, "lon"=>-85.568594),
"/US/US-AL/waldo-451" => array("type"=>"town", "pop"=>278, "name_ru"=>"Уолдо", "name_en"=>"Waldo", "lang"=>"en", "lat"=>33.390871, "lon"=>-86.035391),
"/US/US-AL/walnut-grove-452" => array("type"=>"town", "pop"=>697, "name_ru"=>"Уолнат Гров", "name_en"=>"Walnut Grove", "lang"=>"en", "lat"=>34.064583, "lon"=>-86.287287),
"/US/US-AL/warrior-453" => array("type"=>"city", "pop"=>3115, "name_ru"=>"Уорриор", "name_en"=>"Warrior", "lang"=>"en", "lat"=>33.813913, "lon"=>-86.808203),
"/US/US-AL/waterloo-454" => array("type"=>"town", "pop"=>210, "name_ru"=>"Ватрелоо", "name_en"=>"Waterloo", "lang"=>"en", "lat"=>34.916709, "lon"=>-88.064174),
"/US/US-AL/waverly-455" => array("type"=>"town", "pop"=>172, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>32.736134, "lon"=>-85.576150),
"/US/US-AL/weaver-456" => array("type"=>"city", "pop"=>2722, "name_ru"=>"Вивер", "name_en"=>"Weaver", "lang"=>"en", "lat"=>33.755600, "lon"=>-85.808576),
"/US/US-AL/webb-457" => array("type"=>"town", "pop"=>1367, "name_ru"=>"Уэбб", "name_en"=>"Webb", "lang"=>"en", "lat"=>31.256622, "lon"=>-85.285645),
"/US/US-AL/wedowee-458" => array("type"=>"town", "pop"=>824, "name_ru"=>"Ведоуи", "name_en"=>"Wedowee", "lang"=>"en", "lat"=>33.308498, "lon"=>-85.486532),
"/US/US-AL/west-blocton-459" => array("type"=>"town", "pop"=>1427, "name_ru"=>"Вэст Блоктон", "name_en"=>"West Blocton", "lang"=>"en", "lat"=>33.118109, "lon"=>-87.125824),
"/US/US-AL/west-end-cobb-town-460" => array("type"=>"cdp", "pop"=>4034, "name_ru"=>"Вэст Энд-Кобб Таун", "name_en"=>"West End-Cobb Town", "lang"=>"en", "lat"=>33.648156, "lon"=>-85.871219),
"/US/US-AL/west-jefferson-461" => array("type"=>"town", "pop"=>321, "name_ru"=>"Вэст Джефферсон", "name_en"=>"West Jefferson", "lang"=>"en", "lat"=>33.648736, "lon"=>-87.067931),
"/US/US-AL/west-point-463" => array("type"=>"town", "pop"=>605, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>34.238096, "lon"=>-86.948206),
"/US/US-AL/weston-462" => array("type"=>"town", "pop"=>384, "name_ru"=>"Вэстон", "name_en"=>"Weston", "lang"=>"en", "lat"=>34.162189, "lon"=>-88.030852),
"/US/US-AL/wetumpka-464" => array("type"=>"city", "pop"=>7719, "name_ru"=>"Уэтампка", "name_en"=>"Wetumpka", "lang"=>"en", "lat"=>32.544192, "lon"=>-86.201761),
"/US/US-AL/white-hall-465" => array("type"=>"town", "pop"=>965, "name_ru"=>"Уайт Холл", "name_en"=>"White Hall", "lang"=>"en", "lat"=>32.300783, "lon"=>-86.714537),
"/US/US-AL/wilmer-466" => array("type"=>"town", "pop"=>494, "name_ru"=>"Уилмер", "name_en"=>"Wilmer", "lang"=>"en", "lat"=>30.825683, "lon"=>-88.359191),
"/US/US-AL/wilsonville-467" => array("type"=>"town", "pop"=>1814, "name_ru"=>"Уилсонвилл", "name_en"=>"Wilsonville", "lang"=>"en", "lat"=>33.236035, "lon"=>-86.479006),
"/US/US-AL/wilton-468" => array("type"=>"town", "pop"=>659, "name_ru"=>"Уилтон", "name_en"=>"Wilton", "lang"=>"en", "lat"=>33.082152, "lon"=>-86.881057),
"/US/US-AL/winfield-469" => array("type"=>"city", "pop"=>4618, "name_ru"=>"Винфилд", "name_en"=>"Winfield", "lang"=>"en", "lat"=>33.929853, "lon"=>-87.796085),
"/US/US-AL/woodland-470" => array("type"=>"town", "pop"=>209, "name_ru"=>"Вудленд", "name_en"=>"Woodland", "lang"=>"en", "lat"=>33.374083, "lon"=>-85.399270),
"/US/US-AL/woodville-471" => array("type"=>"town", "pop"=>751, "name_ru"=>"Вудвилл", "name_en"=>"Woodville", "lang"=>"en", "lat"=>34.625248, "lon"=>-86.282166),
"/US/US-AL/yellow-bluff-472" => array("type"=>"town", "pop"=>180, "name_ru"=>"Йеллоу Блафф", "name_en"=>"Yellow Bluff", "lang"=>"en", "lat"=>31.959773, "lon"=>-87.482179),
"/US/US-AL/york-473" => array("type"=>"city", "pop"=>2515, "name_ru"=>"Йорк", "name_en"=>"York", "lang"=>"en", "lat"=>32.495962, "lon"=>-88.296126),
"/US/US-AR" => array("type"=>"state", "pop"=>2855390, "name_ru"=>"Арканзас", "name_en"=>"Arkansas", "lang"=>"en"),
"/US/US-AR/adona-970" => array("type"=>"town", "pop"=>187, "name_ru"=>"Адона", "name_en"=>"Adona", "lang"=>"en", "lat"=>35.039957, "lon"=>-92.898462),
"/US/US-AR/alexander-971" => array("type"=>"town", "pop"=>2291, "name_ru"=>"Александр", "name_en"=>"Alexander", "lang"=>"en", "lat"=>34.630272, "lon"=>-92.442774),
"/US/US-AR/alicia-972" => array("type"=>"town", "pop"=>138, "name_ru"=>"Алишиа", "name_en"=>"Alicia", "lang"=>"en", "lat"=>35.894420, "lon"=>-91.083028),
"/US/US-AR/allport-973" => array("type"=>"town", "pop"=>130, "name_ru"=>"Аллпорт", "name_en"=>"Allport", "lang"=>"en", "lat"=>34.539454, "lon"=>-91.784849),
"/US/US-AR/alma-974" => array("type"=>"city", "pop"=>5104, "name_ru"=>"Алма", "name_en"=>"Alma", "lang"=>"en", "lat"=>35.489943, "lon"=>-94.222852),
"/US/US-AR/almyra-975" => array("type"=>"town", "pop"=>303, "name_ru"=>"Альмира", "name_en"=>"Almyra", "lang"=>"en", "lat"=>34.405749, "lon"=>-91.410555),
"/US/US-AR/alpena-976" => array("type"=>"town", "pop"=>390, "name_ru"=>"Алпена", "name_en"=>"Alpena", "lang"=>"en", "lat"=>36.292970, "lon"=>-93.293346),
"/US/US-AR/altheimer-977" => array("type"=>"city", "pop"=>1110, "name_ru"=>"Алтеимер", "name_en"=>"Altheimer", "lang"=>"en", "lat"=>34.318947, "lon"=>-91.846228),
"/US/US-AR/altus-978" => array("type"=>"city", "pop"=>838, "name_ru"=>"Алтус", "name_en"=>"Altus", "lang"=>"en", "lat"=>35.442873, "lon"=>-93.762461),
"/US/US-AR/amagon-979" => array("type"=>"town", "pop"=>86, "name_ru"=>"Амагон", "name_en"=>"Amagon", "lang"=>"en", "lat"=>35.562441, "lon"=>-91.110573),
"/US/US-AR/amity-980" => array("type"=>"city", "pop"=>765, "name_ru"=>"Амити", "name_en"=>"Amity", "lang"=>"en", "lat"=>34.266149, "lon"=>-93.463909),
"/US/US-AR/antoine-981" => array("type"=>"town", "pop"=>145, "name_ru"=>"Антоайн", "name_en"=>"Antoine", "lang"=>"en", "lat"=>34.034016, "lon"=>-93.421059),
"/US/US-AR/arkadelphia-982" => array("type"=>"city", "pop"=>11130, "name_ru"=>"Аркадельфиа", "name_en"=>"Arkadelphia", "lang"=>"en", "lat"=>34.123759, "lon"=>-93.070631),
"/US/US-AR/arkansas-city-983" => array("type"=>"city", "pop"=>517, "name_ru"=>"Арканзас Сити", "name_en"=>"Arkansas City", "lang"=>"en", "lat"=>33.608824, "lon"=>-91.205659),
"/US/US-AR/ash-flat-985" => array("type"=>"city", "pop"=>1100, "name_ru"=>"Аш Флэт", "name_en"=>"Ash Flat", "lang"=>"en", "lat"=>36.225765, "lon"=>-91.606522),
"/US/US-AR/ashdown-984" => array("type"=>"city", "pop"=>4440, "name_ru"=>"Ашдаун", "name_en"=>"Ashdown", "lang"=>"en", "lat"=>33.674367, "lon"=>-94.125998),
"/US/US-AR/atkins-986" => array("type"=>"city", "pop"=>2958, "name_ru"=>"Аткинс", "name_en"=>"Atkins", "lang"=>"en", "lat"=>35.242555, "lon"=>-92.945561),
"/US/US-AR/aubrey-987" => array("type"=>"town", "pop"=>182, "name_ru"=>"Обри", "name_en"=>"Aubrey", "lang"=>"en", "lat"=>34.719780, "lon"=>-90.897847),
"/US/US-AR/augusta-988" => array("type"=>"city", "pop"=>2250, "name_ru"=>"Огаста", "name_en"=>"Augusta", "lang"=>"en", "lat"=>35.285466, "lon"=>-91.360486),
"/US/US-AR/austin-989" => array("type"=>"town", "pop"=>1797, "name_ru"=>"Остин", "name_en"=>"Austin", "lang"=>"en", "lat"=>34.998040, "lon"=>-91.981775),
"/US/US-AR/avoca-990" => array("type"=>"town", "pop"=>447, "name_ru"=>"Авока", "name_en"=>"Avoca", "lang"=>"en", "lat"=>36.399898, "lon"=>-94.073037),
"/US/US-AR/bald-knob-991" => array("type"=>"city", "pop"=>3420, "name_ru"=>"Болд Ноб", "name_en"=>"Bald Knob", "lang"=>"en", "lat"=>35.308177, "lon"=>-91.572525),
"/US/US-AR/banks-992" => array("type"=>"town", "pop"=>116, "name_ru"=>"Бэнкс", "name_en"=>"Banks", "lang"=>"en", "lat"=>33.575929, "lon"=>-92.268238),
"/US/US-AR/barling-993" => array("type"=>"city", "pop"=>4471, "name_ru"=>"Барлинг", "name_en"=>"Barling", "lang"=>"en", "lat"=>35.325153, "lon"=>-94.300068),
"/US/US-AR/bassett-994" => array("type"=>"town", "pop"=>154, "name_ru"=>"Бассет", "name_en"=>"Bassett", "lang"=>"en", "lat"=>35.536081, "lon"=>-90.130889),
"/US/US-AR/batesville-995" => array("type"=>"city", "pop"=>9511, "name_ru"=>"Батсвилл", "name_en"=>"Batesville", "lang"=>"en", "lat"=>35.768240, "lon"=>-91.622279),
"/US/US-AR/bauxite-996" => array("type"=>"town", "pop"=>548, "name_ru"=>"Бааксит", "name_en"=>"Bauxite", "lang"=>"en", "lat"=>34.556876, "lon"=>-92.526041),
"/US/US-AR/bay-997" => array("type"=>"city", "pop"=>2030, "name_ru"=>"Бэй", "name_en"=>"Bay", "lang"=>"en", "lat"=>35.740027, "lon"=>-90.559190),
"/US/US-AR/bearden-998" => array("type"=>"city", "pop"=>990, "name_ru"=>"Бирден", "name_en"=>"Bearden", "lang"=>"en", "lat"=>33.727515, "lon"=>-92.616261),
"/US/US-AR/beaver-999" => array("type"=>"town", "pop"=>117, "name_ru"=>"Бивер", "name_en"=>"Beaver", "lang"=>"en", "lat"=>36.476124, "lon"=>-93.772743),
"/US/US-AR/beebe-1000" => array("type"=>"city", "pop"=>6921, "name_ru"=>"Биб", "name_en"=>"Beebe", "lang"=>"en", "lat"=>35.072252, "lon"=>-91.893890),
"/US/US-AR/beedeville-1001" => array("type"=>"town", "pop"=>96, "name_ru"=>"Бидевилл", "name_en"=>"Beedeville", "lang"=>"en", "lat"=>35.428568, "lon"=>-91.108917),
"/US/US-AR/bella-vista-1002" => array("type"=>"cdp", "pop"=>16388, "name_ru"=>"Белла Виста", "name_en"=>"Bella Vista", "lang"=>"en", "lat"=>36.468136, "lon"=>-94.294387),
"/US/US-AR/bellefonte-1003" => array("type"=>"town", "pop"=>445, "name_ru"=>"Беллефонт", "name_en"=>"Bellefonte", "lang"=>"en", "lat"=>36.200434, "lon"=>-93.046363),
"/US/US-AR/belleville-1004" => array("type"=>"city", "pop"=>381, "name_ru"=>"Белелвилл", "name_en"=>"Belleville", "lang"=>"en", "lat"=>35.090666, "lon"=>-93.445305),
"/US/US-AR/ben-lomond-1005" => array("type"=>"town", "pop"=>132, "name_ru"=>"Бен Ломонд", "name_en"=>"Ben Lomond", "lang"=>"en", "lat"=>33.834431, "lon"=>-94.114988),
"/US/US-AR/benton-1006" => array("type"=>"city", "pop"=>29452, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>34.569243, "lon"=>-92.577250),
"/US/US-AR/bentonville-1007" => array("type"=>"city", "pop"=>35526, "name_ru"=>"Бентонвилль", "name_en"=>"Bentonville", "lang"=>"en", "lat"=>36.368850, "lon"=>-94.204849),
"/US/US-AR/bergman-1008" => array("type"=>"town", "pop"=>446, "name_ru"=>"Бергман", "name_en"=>"Bergman", "lang"=>"en", "lat"=>36.312940, "lon"=>-93.011061),
"/US/US-AR/berryville-1009" => array("type"=>"city", "pop"=>5247, "name_ru"=>"Берривилл", "name_en"=>"Berryville", "lang"=>"en", "lat"=>36.371001, "lon"=>-93.565389),
"/US/US-AR/bethel-heights-1010" => array("type"=>"town", "pop"=>1557, "name_ru"=>"Бетел Хайтс", "name_en"=>"Bethel Heights", "lang"=>"en", "lat"=>36.222767, "lon"=>-94.127457),
"/US/US-AR/big-flat-1012" => array("type"=>"town", "pop"=>110, "name_ru"=>"Биг Флэт", "name_en"=>"Big Flat", "lang"=>"en", "lat"=>36.005766, "lon"=>-92.404996),
"/US/US-AR/bigelow-1011" => array("type"=>"town", "pop"=>332, "name_ru"=>"Бигелоу", "name_en"=>"Bigelow", "lang"=>"en", "lat"=>34.998974, "lon"=>-92.631041),
"/US/US-AR/biggers-1013" => array("type"=>"town", "pop"=>344, "name_ru"=>"Биггерс", "name_en"=>"Biggers", "lang"=>"en", "lat"=>36.332002, "lon"=>-90.805928),
"/US/US-AR/birdsong-1014" => array("type"=>"town", "pop"=>37, "name_ru"=>"Бердсонг", "name_en"=>"Birdsong", "lang"=>"en", "lat"=>35.459018, "lon"=>-90.260501),
"/US/US-AR/black-oak-1015" => array("type"=>"town", "pop"=>300, "name_ru"=>"Блэк Оук", "name_en"=>"Black Oak", "lang"=>"en", "lat"=>35.835471, "lon"=>-90.367109),
"/US/US-AR/black-rock-1016" => array("type"=>"city", "pop"=>670, "name_ru"=>"Блэк Рок", "name_en"=>"Black Rock", "lang"=>"en", "lat"=>36.106382, "lon"=>-91.108629),
"/US/US-AR/black-springs-1017" => array("type"=>"town", "pop"=>110, "name_ru"=>"Блэк Спрингс", "name_en"=>"Black Springs", "lang"=>"en", "lat"=>34.461511, "lon"=>-93.713541),
"/US/US-AR/blevins-1018" => array("type"=>"city", "pop"=>353, "name_ru"=>"Блевинс", "name_en"=>"Blevins", "lang"=>"en", "lat"=>33.871131, "lon"=>-93.577248),
"/US/US-AR/blue-eye-1019" => array("type"=>"town", "pop"=>38, "name_ru"=>"Блу Ай", "name_en"=>"Blue Eye", "lang"=>"en", "lat"=>36.495973, "lon"=>-93.396921),
"/US/US-AR/blue-mountain-1020" => array("type"=>"town", "pop"=>135, "name_ru"=>"Блу Маунтин", "name_en"=>"Blue Mountain", "lang"=>"en", "lat"=>35.131714, "lon"=>-93.715311),
"/US/US-AR/bluff-city-1021" => array("type"=>"town", "pop"=>120, "name_ru"=>"Блафф Сити", "name_en"=>"Bluff City", "lang"=>"en", "lat"=>33.717233, "lon"=>-93.133062),
"/US/US-AR/blytheville-1022" => array("type"=>"city", "pop"=>16105, "name_ru"=>"Блитевилл", "name_en"=>"Blytheville", "lang"=>"en", "lat"=>35.934312, "lon"=>-89.924775),
"/US/US-AR/bodcaw-1023" => array("type"=>"town", "pop"=>117, "name_ru"=>"Бодко", "name_en"=>"Bodcaw", "lang"=>"en", "lat"=>33.556622, "lon"=>-93.401113),
"/US/US-AR/bonanza-1024" => array("type"=>"town", "pop"=>567, "name_ru"=>"Бонанза", "name_en"=>"Bonanza", "lang"=>"en", "lat"=>35.234127, "lon"=>-94.427445),
"/US/US-AR/bono-1025" => array("type"=>"town", "pop"=>1599, "name_ru"=>"Боно", "name_en"=>"Bono", "lang"=>"en", "lat"=>35.910539, "lon"=>-90.796673),
"/US/US-AR/booneville-1026" => array("type"=>"city", "pop"=>4086, "name_ru"=>"Бунвилл", "name_en"=>"Booneville", "lang"=>"en", "lat"=>35.139397, "lon"=>-93.918461),
"/US/US-AR/bradford-1027" => array("type"=>"city", "pop"=>853, "name_ru"=>"Брадфорд", "name_en"=>"Bradford", "lang"=>"en", "lat"=>35.423067, "lon"=>-91.456642),
"/US/US-AR/bradley-1028" => array("type"=>"city", "pop"=>521, "name_ru"=>"Брадли", "name_en"=>"Bradley", "lang"=>"en", "lat"=>33.099915, "lon"=>-93.656861),
"/US/US-AR/branch-1029" => array("type"=>"city", "pop"=>365, "name_ru"=>"Брэнч", "name_en"=>"Branch", "lang"=>"en", "lat"=>35.306056, "lon"=>-93.956258),
"/US/US-AR/brinkley-1030" => array("type"=>"city", "pop"=>3211, "name_ru"=>"Бринкли", "name_en"=>"Brinkley", "lang"=>"en", "lat"=>34.884330, "lon"=>-91.189038),
"/US/US-AR/brookland-1031" => array("type"=>"town", "pop"=>1670, "name_ru"=>"Брукленд", "name_en"=>"Brookland", "lang"=>"en", "lat"=>35.901410, "lon"=>-90.580836),
"/US/US-AR/bryant-1032" => array("type"=>"city", "pop"=>15040, "name_ru"=>"Брайант", "name_en"=>"Bryant", "lang"=>"en", "lat"=>34.612371, "lon"=>-92.492041),
"/US/US-AR/buckner-1033" => array("type"=>"city", "pop"=>355, "name_ru"=>"Бакнер", "name_en"=>"Buckner", "lang"=>"en", "lat"=>33.357167, "lon"=>-93.435380),
"/US/US-AR/bull-shoals-1034" => array("type"=>"city", "pop"=>2138, "name_ru"=>"Булл Шолс", "name_en"=>"Bull Shoals", "lang"=>"en", "lat"=>36.376039, "lon"=>-92.588413),
"/US/US-AR/burdette-1035" => array("type"=>"town", "pop"=>117, "name_ru"=>"Бердетт", "name_en"=>"Burdette", "lang"=>"en", "lat"=>35.814861, "lon"=>-89.945934),
"/US/US-AR/cabot-1036" => array("type"=>"city", "pop"=>23614, "name_ru"=>"Кабот", "name_en"=>"Cabot", "lang"=>"en", "lat"=>34.968176, "lon"=>-92.021989),
"/US/US-AR/caddo-valley-1037" => array("type"=>"town", "pop"=>640, "name_ru"=>"Каддо Валли", "name_en"=>"Caddo Valley", "lang"=>"en", "lat"=>34.187866, "lon"=>-93.070040),
"/US/US-AR/caldwell-1038" => array("type"=>"town", "pop"=>428, "name_ru"=>"Колдуэлл", "name_en"=>"Caldwell", "lang"=>"en", "lat"=>35.079327, "lon"=>-90.815078),
"/US/US-AR/cale-1039" => array("type"=>"town", "pop"=>56, "name_ru"=>"Кал", "name_en"=>"Cale", "lang"=>"en", "lat"=>33.627926, "lon"=>-93.233108),
"/US/US-AR/calico-rock-1040" => array("type"=>"city", "pop"=>969, "name_ru"=>"Калико Рок", "name_en"=>"Calico Rock", "lang"=>"en", "lat"=>36.128298, "lon"=>-92.131148),
"/US/US-AR/calion-1041" => array("type"=>"city", "pop"=>482, "name_ru"=>"Калион", "name_en"=>"Calion", "lang"=>"en", "lat"=>33.328838, "lon"=>-92.538397),
"/US/US-AR/camden-1042" => array("type"=>"city", "pop"=>11512, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>33.566689, "lon"=>-92.845657),
"/US/US-AR/cammack-village-1043" => array("type"=>"city", "pop"=>781, "name_ru"=>"Каммек Виллидж", "name_en"=>"Cammack Village", "lang"=>"en", "lat"=>34.779850, "lon"=>-92.346780),
"/US/US-AR/campbell-station-1044" => array("type"=>"town", "pop"=>205, "name_ru"=>"Кэмпбелл Стейшен", "name_en"=>"Campbell Station", "lang"=>"en", "lat"=>35.669481, "lon"=>-91.254380),
"/US/US-AR/caraway-1045" => array("type"=>"town", "pop"=>1393, "name_ru"=>"Карауэй", "name_en"=>"Caraway", "lang"=>"en", "lat"=>35.758035, "lon"=>-90.331987),
"/US/US-AR/carlisle-1046" => array("type"=>"city", "pop"=>2408, "name_ru"=>"Карлайл", "name_en"=>"Carlisle", "lang"=>"en", "lat"=>34.786086, "lon"=>-91.746513),
"/US/US-AR/carthage-1047" => array("type"=>"city", "pop"=>401, "name_ru"=>"Картаж", "name_en"=>"Carthage", "lang"=>"en", "lat"=>34.072252, "lon"=>-92.555548),
"/US/US-AR/casa-1048" => array("type"=>"town", "pop"=>213, "name_ru"=>"Каса", "name_en"=>"Casa", "lang"=>"en", "lat"=>35.025833, "lon"=>-93.044321),
"/US/US-AR/cash-1049" => array("type"=>"town", "pop"=>318, "name_ru"=>"Кэш", "name_en"=>"Cash", "lang"=>"en", "lat"=>35.799822, "lon"=>-90.932200),
"/US/US-AR/caulksville-1050" => array("type"=>"town", "pop"=>233, "name_ru"=>"Колксвилл", "name_en"=>"Caulksville", "lang"=>"en", "lat"=>35.301739, "lon"=>-93.868499),
"/US/US-AR/cave-city-1051" => array("type"=>"city", "pop"=>2046, "name_ru"=>"Кейв Сити", "name_en"=>"Cave City", "lang"=>"en", "lat"=>35.946680, "lon"=>-91.547830),
"/US/US-AR/cave-springs-1052" => array("type"=>"city", "pop"=>1695, "name_ru"=>"Кейв Спрингс", "name_en"=>"Cave Springs", "lang"=>"en", "lat"=>36.262694, "lon"=>-94.228719),
"/US/US-AR/centerton-1053" => array("type"=>"city", "pop"=>8593, "name_ru"=>"Сентертон", "name_en"=>"Centerton", "lang"=>"en", "lat"=>36.360312, "lon"=>-94.284966),
"/US/US-AR/central-city-1054" => array("type"=>"town", "pop"=>551, "name_ru"=>"Сентрал Сити", "name_en"=>"Central City", "lang"=>"en", "lat"=>35.327363, "lon"=>-94.236698),
"/US/US-AR/charleston-1055" => array("type"=>"city", "pop"=>3028, "name_ru"=>"Чарльстон", "name_en"=>"Charleston", "lang"=>"en", "lat"=>35.294511, "lon"=>-94.048435),
"/US/US-AR/cherokee-village-hidden-valley-1056" => array("type"=>"cdp", "pop"=>4416, "name_ru"=>"Чероки Виллидж - Хидн Валли", "name_en"=>"Cherokee Village-Hidden Valley", "lang"=>"en", "lat"=>36.301622, "lon"=>-91.569482),
"/US/US-AR/cherry-valley-1057" => array("type"=>"city", "pop"=>672, "name_ru"=>"Черри Валли", "name_en"=>"Cherry Valley", "lang"=>"en", "lat"=>35.402976, "lon"=>-90.753787),
"/US/US-AR/chester-1058" => array("type"=>"town", "pop"=>104, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>35.678047, "lon"=>-94.177474),
"/US/US-AR/chidester-1059" => array("type"=>"city", "pop"=>328, "name_ru"=>"Чидестер", "name_en"=>"Chidester", "lang"=>"en", "lat"=>33.699910, "lon"=>-93.018715),
"/US/US-AR/clarendon-1060" => array("type"=>"city", "pop"=>1666, "name_ru"=>"Кларендон", "name_en"=>"Clarendon", "lang"=>"en", "lat"=>34.695458, "lon"=>-91.305817),
"/US/US-AR/clarksville-1061" => array("type"=>"city", "pop"=>8579, "name_ru"=>"Кларксвилл", "name_en"=>"Clarksville", "lang"=>"en", "lat"=>35.457574, "lon"=>-93.472395),
"/US/US-AR/clinton-1062" => array("type"=>"city", "pop"=>2470, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>35.581075, "lon"=>-92.452359),
"/US/US-AR/coal-hill-1063" => array("type"=>"city", "pop"=>1071, "name_ru"=>"Кол Хилл", "name_en"=>"Coal Hill", "lang"=>"en", "lat"=>35.437023, "lon"=>-93.667554),
"/US/US-AR/college-city-1064" => array("type"=>"town", "pop"=>379, "name_ru"=>"Колледж Сити", "name_en"=>"College City", "lang"=>"en", "lat"=>36.125709, "lon"=>-90.939988),
"/US/US-AR/colt-1065" => array("type"=>"city", "pop"=>333, "name_ru"=>"Кольт", "name_en"=>"Colt", "lang"=>"en", "lat"=>35.132535, "lon"=>-90.812292),
"/US/US-AR/concord-1066" => array("type"=>"town", "pop"=>265, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>35.663429, "lon"=>-91.848729),
"/US/US-AR/conway-1067" => array("type"=>"city", "pop"=>57544, "name_ru"=>"Конвэй", "name_en"=>"Conway", "lang"=>"en", "lat"=>35.088809, "lon"=>-92.453098),
"/US/US-AR/corinth-1068" => array("type"=>"town", "pop"=>67, "name_ru"=>"Коринт", "name_en"=>"Corinth", "lang"=>"en", "lat"=>35.068243, "lon"=>-93.420544),
"/US/US-AR/corning-1069" => array("type"=>"city", "pop"=>3308, "name_ru"=>"Корнинг", "name_en"=>"Corning", "lang"=>"en", "lat"=>36.410818, "lon"=>-90.585851),
"/US/US-AR/cotter-1070" => array("type"=>"city", "pop"=>1079, "name_ru"=>"Коттер", "name_en"=>"Cotter", "lang"=>"en", "lat"=>36.282492, "lon"=>-92.520229),
"/US/US-AR/cotton-plant-1071" => array("type"=>"city", "pop"=>823, "name_ru"=>"Коттон Плант", "name_en"=>"Cotton Plant", "lang"=>"en", "lat"=>35.006938, "lon"=>-91.251472),
"/US/US-AR/cove-1072" => array("type"=>"town", "pop"=>380, "name_ru"=>"Ков", "name_en"=>"Cove", "lang"=>"en", "lat"=>34.435017, "lon"=>-94.409229),
"/US/US-AR/coy-1073" => array("type"=>"town", "pop"=>119, "name_ru"=>"Кой", "name_en"=>"Coy", "lang"=>"en", "lat"=>34.541788, "lon"=>-91.869600),
"/US/US-AR/crawfordsville-1074" => array("type"=>"town", "pop"=>489, "name_ru"=>"Крофордсвилл", "name_en"=>"Crawfordsville", "lang"=>"en", "lat"=>35.225653, "lon"=>-90.326220),
"/US/US-AR/crossett-1075" => array("type"=>"city", "pop"=>5518, "name_ru"=>"Кроссетт", "name_en"=>"Crossett", "lang"=>"en", "lat"=>33.127353, "lon"=>-91.966014),
"/US/US-AR/cushman-1076" => array("type"=>"town", "pop"=>468, "name_ru"=>"Кушман", "name_en"=>"Cushman", "lang"=>"en", "lat"=>35.863606, "lon"=>-91.778612),
"/US/US-AR/daisy-1077" => array("type"=>"town", "pop"=>112, "name_ru"=>"Дейзи", "name_en"=>"Daisy", "lang"=>"en", "lat"=>34.233088, "lon"=>-93.741895),
"/US/US-AR/damascus-1078" => array("type"=>"town", "pop"=>365, "name_ru"=>"Дамаскус", "name_en"=>"Damascus", "lang"=>"en", "lat"=>35.364640, "lon"=>-92.405811),
"/US/US-AR/danville-1079" => array("type"=>"city", "pop"=>2476, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>35.052081, "lon"=>-93.390282),
"/US/US-AR/dardanelle-1080" => array("type"=>"city", "pop"=>4435, "name_ru"=>"Дарданелл", "name_en"=>"Dardanelle", "lang"=>"en", "lat"=>35.226423, "lon"=>-93.165609),
"/US/US-AR/datto-1081" => array("type"=>"town", "pop"=>87, "name_ru"=>"Датто", "name_en"=>"Datto", "lang"=>"en", "lat"=>36.393166, "lon"=>-90.728081),
"/US/US-AR/de-queen-1087" => array("type"=>"city", "pop"=>5948, "name_ru"=>"Де Квин", "name_en"=>"De Queen", "lang"=>"en", "lat"=>34.042740, "lon"=>-94.341807),
"/US/US-AR/de-valls-bluff-1090" => array("type"=>"town", "pop"=>693, "name_ru"=>"Де Валлс Блафф", "name_en"=>"De Valls Bluff", "lang"=>"en", "lat"=>34.785728, "lon"=>-91.460788),
"/US/US-AR/de-witt-1091" => array("type"=>"city", "pop"=>3254, "name_ru"=>"Де Витт", "name_en"=>"De Witt", "lang"=>"en", "lat"=>34.289533, "lon"=>-91.338131),
"/US/US-AR/decatur-1082" => array("type"=>"city", "pop"=>1910, "name_ru"=>"Декатур", "name_en"=>"Decatur", "lang"=>"en", "lat"=>36.337579, "lon"=>-94.458484),
"/US/US-AR/delaplaine-1083" => array("type"=>"town", "pop"=>130, "name_ru"=>"Делаплейн", "name_en"=>"Delaplaine", "lang"=>"en", "lat"=>36.231320, "lon"=>-90.726015),
"/US/US-AR/delight-1084" => array("type"=>"city", "pop"=>288, "name_ru"=>"Делайт", "name_en"=>"Delight", "lang"=>"en", "lat"=>34.029446, "lon"=>-93.505771),
"/US/US-AR/dell-1085" => array("type"=>"town", "pop"=>226, "name_ru"=>"Делл", "name_en"=>"Dell", "lang"=>"en", "lat"=>35.855185, "lon"=>-90.033437),
"/US/US-AR/denning-1086" => array("type"=>"city", "pop"=>278, "name_ru"=>"Деннинг", "name_en"=>"Denning", "lang"=>"en", "lat"=>35.424602, "lon"=>-93.755219),
"/US/US-AR/dermott-1088" => array("type"=>"city", "pop"=>3225, "name_ru"=>"Дермотт", "name_en"=>"Dermott", "lang"=>"en", "lat"=>33.526602, "lon"=>-91.437725),
"/US/US-AR/des-arc-1089" => array("type"=>"city", "pop"=>1708, "name_ru"=>"Де Арк", "name_en"=>"Des Arc", "lang"=>"en", "lat"=>34.975676, "lon"=>-91.505167),
"/US/US-AR/diamond-city-1092" => array("type"=>"city", "pop"=>803, "name_ru"=>"Даймонд Сити", "name_en"=>"Diamond City", "lang"=>"en", "lat"=>36.455800, "lon"=>-92.915096),
"/US/US-AR/diaz-1093" => array("type"=>"city", "pop"=>1156, "name_ru"=>"Диас", "name_en"=>"Diaz", "lang"=>"en", "lat"=>35.647949, "lon"=>-91.259891),
"/US/US-AR/dierks-1094" => array("type"=>"city", "pop"=>1217, "name_ru"=>"Диркс", "name_en"=>"Dierks", "lang"=>"en", "lat"=>34.120579, "lon"=>-94.018769),
"/US/US-AR/dover-1095" => array("type"=>"city", "pop"=>1400, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>35.402166, "lon"=>-93.112419),
"/US/US-AR/dumas-1096" => array("type"=>"city", "pop"=>4622, "name_ru"=>"Дюма", "name_en"=>"Dumas", "lang"=>"en", "lat"=>33.883161, "lon"=>-91.485919),
"/US/US-AR/dyer-1097" => array("type"=>"town", "pop"=>615, "name_ru"=>"Дайер", "name_en"=>"Dyer", "lang"=>"en", "lat"=>35.494446, "lon"=>-94.138336),
"/US/US-AR/dyess-1098" => array("type"=>"town", "pop"=>474, "name_ru"=>"Дайесс", "name_en"=>"Dyess", "lang"=>"en", "lat"=>35.589359, "lon"=>-90.213131),
"/US/US-AR/earle-1099" => array("type"=>"city", "pop"=>2735, "name_ru"=>"Эрл", "name_en"=>"Earle", "lang"=>"en", "lat"=>35.273294, "lon"=>-90.464978),
"/US/US-AR/east-camden-1100" => array("type"=>"town", "pop"=>785, "name_ru"=>"Ист Камден", "name_en"=>"East Camden", "lang"=>"en", "lat"=>33.608507, "lon"=>-92.742409),
"/US/US-AR/edmondson-1101" => array("type"=>"town", "pop"=>474, "name_ru"=>"Эдмондсон", "name_en"=>"Edmondson", "lang"=>"en", "lat"=>35.103547, "lon"=>-90.309776),
"/US/US-AR/egypt-1102" => array("type"=>"town", "pop"=>106, "name_ru"=>"Эджипт", "name_en"=>"Egypt", "lang"=>"en", "lat"=>35.866649, "lon"=>-90.952883),
"/US/US-AR/el-dorado-1104" => array("type"=>"city", "pop"=>19905, "name_ru"=>"Эдь Дорадо", "name_en"=>"El Dorado", "lang"=>"en", "lat"=>33.219028, "lon"=>-92.662763),
"/US/US-AR/elaine-1103" => array("type"=>"city", "pop"=>698, "name_ru"=>"Элейн", "name_en"=>"Elaine", "lang"=>"en", "lat"=>34.308276, "lon"=>-90.854089),
"/US/US-AR/elkins-1105" => array("type"=>"city", "pop"=>2528, "name_ru"=>"Элкинс", "name_en"=>"Elkins", "lang"=>"en", "lat"=>36.019799, "lon"=>-94.028293),
"/US/US-AR/elm-springs-1106" => array("type"=>"town", "pop"=>1298, "name_ru"=>"Эльм Спрингс", "name_en"=>"Elm Springs", "lang"=>"en", "lat"=>36.201729, "lon"=>-94.215496),
"/US/US-AR/emerson-1107" => array("type"=>"town", "pop"=>341, "name_ru"=>"Эмерсон", "name_en"=>"Emerson", "lang"=>"en", "lat"=>33.097365, "lon"=>-93.196092),
"/US/US-AR/emmet-1108" => array("type"=>"city", "pop"=>420, "name_ru"=>"Эммет", "name_en"=>"Emmet", "lang"=>"en", "lat"=>33.722052, "lon"=>-93.466177),
"/US/US-AR/england-1109" => array("type"=>"city", "pop"=>3025, "name_ru"=>"Ингленд", "name_en"=>"England", "lang"=>"en", "lat"=>34.544911, "lon"=>-91.967932),
"/US/US-AR/enola-1110" => array("type"=>"town", "pop"=>205, "name_ru"=>"Энола", "name_en"=>"Enola", "lang"=>"en", "lat"=>35.200235, "lon"=>-92.201061),
"/US/US-AR/eudora-1111" => array("type"=>"city", "pop"=>2339, "name_ru"=>"Эудора", "name_en"=>"Eudora", "lang"=>"en", "lat"=>33.119766, "lon"=>-91.263117),
"/US/US-AR/eureka-springs-1112" => array("type"=>"city", "pop"=>2353, "name_ru"=>"Юрика Спрингс", "name_en"=>"Eureka Springs", "lang"=>"en", "lat"=>36.399212, "lon"=>-93.737378),
"/US/US-AR/evening-shade-1113" => array("type"=>"town", "pop"=>493, "name_ru"=>"Ивнинг Шейд", "name_en"=>"Evening Shade", "lang"=>"en", "lat"=>36.070437, "lon"=>-91.621271),
"/US/US-AR/everton-1114" => array("type"=>"town", "pop"=>182, "name_ru"=>"Эвертон", "name_en"=>"Everton", "lang"=>"en", "lat"=>36.155375, "lon"=>-92.908392),
"/US/US-AR/fairfield-bay-1115" => array("type"=>"cdp", "pop"=>2496, "name_ru"=>"Фэрфилд Бэй", "name_en"=>"Fairfield Bay", "lang"=>"en", "lat"=>35.605071, "lon"=>-92.269823),
"/US/US-AR/fargo-1116" => array("type"=>"town", "pop"=>101, "name_ru"=>"Фарго", "name_en"=>"Fargo", "lang"=>"en", "lat"=>34.955047, "lon"=>-91.177055),
"/US/US-AR/farmington-1117" => array("type"=>"city", "pop"=>4677, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>36.042343, "lon"=>-94.241974),
"/US/US-AR/fayetteville-1118" => array("type"=>"city", "pop"=>73372, "name_ru"=>"Фейетвилл", "name_en"=>"Fayetteville", "lang"=>"en", "lat"=>36.071800, "lon"=>-94.155684),
"/US/US-AR/felsenthal-1119" => array("type"=>"town", "pop"=>149, "name_ru"=>"Фелсентал", "name_en"=>"Felsenthal", "lang"=>"en", "lat"=>33.057875, "lon"=>-92.152623),
"/US/US-AR/fifty-six-1120" => array("type"=>"city", "pop"=>171, "name_ru"=>"Фифти-сикс", "name_en"=>"Fifty-Six", "lang"=>"en", "lat"=>35.960401, "lon"=>-92.229916),
"/US/US-AR/fisher-1121" => array("type"=>"town", "pop"=>255, "name_ru"=>"Фишер", "name_en"=>"Fisher", "lang"=>"en", "lat"=>35.491473, "lon"=>-90.972178),
"/US/US-AR/flippin-1122" => array("type"=>"city", "pop"=>1381, "name_ru"=>"Флиппин", "name_en"=>"Flippin", "lang"=>"en", "lat"=>36.277018, "lon"=>-92.592059),
"/US/US-AR/fordyce-1123" => array("type"=>"city", "pop"=>4230, "name_ru"=>"Фордис", "name_en"=>"Fordyce", "lang"=>"en", "lat"=>33.817686, "lon"=>-92.413813),
"/US/US-AR/foreman-1124" => array("type"=>"city", "pop"=>1020, "name_ru"=>"Форман", "name_en"=>"Foreman", "lang"=>"en", "lat"=>33.717334, "lon"=>-94.397549),
"/US/US-AR/forrest-city-1125" => array("type"=>"city", "pop"=>13281, "name_ru"=>"Форрест Сити", "name_en"=>"Forrest City", "lang"=>"en", "lat"=>35.014391, "lon"=>-90.785340),
"/US/US-AR/fort-smith-1126" => array("type"=>"city", "pop"=>84716, "name_ru"=>"Форт Смит", "name_en"=>"Fort Smith", "lang"=>"en", "lat"=>35.367850, "lon"=>-94.384737),
"/US/US-AR/fouke-1127" => array("type"=>"town", "pop"=>852, "name_ru"=>"Фоюк", "name_en"=>"Fouke", "lang"=>"en", "lat"=>33.260780, "lon"=>-93.886448),
"/US/US-AR/fountain-hill-1128" => array("type"=>"town", "pop"=>147, "name_ru"=>"Фонтейн Хилл", "name_en"=>"Fountain Hill", "lang"=>"en", "lat"=>33.356933, "lon"=>-91.850860),
"/US/US-AR/fourche-1129" => array("type"=>"town", "pop"=>60, "name_ru"=>"Форч", "name_en"=>"Fourche", "lang"=>"en", "lat"=>34.993105, "lon"=>-92.618812),
"/US/US-AR/franklin-1130" => array("type"=>"town", "pop"=>179, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>36.171398, "lon"=>-91.772334),
"/US/US-AR/fredonia-(biscoe)-1131" => array("type"=>"town", "pop"=>422, "name_ru"=>"Фредония (Биско)", "name_en"=>"Fredonia (Biscoe)", "lang"=>"en", "lat"=>34.818578, "lon"=>-91.409077),
"/US/US-AR/friendship-1132" => array("type"=>"town", "pop"=>213, "name_ru"=>"Френдшип", "name_en"=>"Friendship", "lang"=>"en", "lat"=>34.223753, "lon"=>-93.003008),
"/US/US-AR/fulton-1133" => array("type"=>"city", "pop"=>235, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>33.611864, "lon"=>-93.814287),
"/US/US-AR/garfield-1134" => array("type"=>"town", "pop"=>465, "name_ru"=>"Гарфилд", "name_en"=>"Garfield", "lang"=>"en", "lat"=>36.453719, "lon"=>-93.969783),
"/US/US-AR/garland-1135" => array("type"=>"town", "pop"=>354, "name_ru"=>"Гарленд", "name_en"=>"Garland", "lang"=>"en", "lat"=>33.359815, "lon"=>-93.712812),
"/US/US-AR/garner-1136" => array("type"=>"town", "pop"=>308, "name_ru"=>"Гарнер", "name_en"=>"Garner", "lang"=>"en", "lat"=>35.139480, "lon"=>-91.780130),
"/US/US-AR/gassville-1137" => array("type"=>"city", "pop"=>2148, "name_ru"=>"Гассвилл", "name_en"=>"Gassville", "lang"=>"en", "lat"=>36.282551, "lon"=>-92.494682),
"/US/US-AR/gateway-1138" => array("type"=>"town", "pop"=>522, "name_ru"=>"Гейтуэй", "name_en"=>"Gateway", "lang"=>"en", "lat"=>36.485566, "lon"=>-93.936335),
"/US/US-AR/gentry-1139" => array("type"=>"city", "pop"=>2978, "name_ru"=>"Джентри", "name_en"=>"Gentry", "lang"=>"en", "lat"=>36.266809, "lon"=>-94.482846),
"/US/US-AR/georgetown-1140" => array("type"=>"town", "pop"=>131, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>35.126261, "lon"=>-91.453233),
"/US/US-AR/gibson-1141" => array("type"=>"cdp", "pop"=>4288, "name_ru"=>"Гибсон", "name_en"=>"Gibson", "lang"=>"en", "lat"=>34.895368, "lon"=>-92.227056),
"/US/US-AR/gilbert-1142" => array("type"=>"town", "pop"=>32, "name_ru"=>"Гилберт", "name_en"=>"Gilbert", "lang"=>"en", "lat"=>35.990187, "lon"=>-92.718214),
"/US/US-AR/gillett-1143" => array("type"=>"city", "pop"=>757, "name_ru"=>"Гиллетт", "name_en"=>"Gillett", "lang"=>"en", "lat"=>34.119126, "lon"=>-91.378985),
"/US/US-AR/gillham-1144" => array("type"=>"town", "pop"=>199, "name_ru"=>"Гиллхэм", "name_en"=>"Gillham", "lang"=>"en", "lat"=>34.167868, "lon"=>-94.313507),
"/US/US-AR/gilmore-1145" => array("type"=>"town", "pop"=>276, "name_ru"=>"Гилмор", "name_en"=>"Gilmore", "lang"=>"en", "lat"=>35.411145, "lon"=>-90.277431),
"/US/US-AR/glenwood-1146" => array("type"=>"town", "pop"=>1999, "name_ru"=>"Гленвуд", "name_en"=>"Glenwood", "lang"=>"en", "lat"=>34.331128, "lon"=>-93.549694),
"/US/US-AR/goshen-1147" => array("type"=>"town", "pop"=>1142, "name_ru"=>"Гошен", "name_en"=>"Goshen", "lang"=>"en", "lat"=>36.108948, "lon"=>-94.002030),
"/US/US-AR/gosnell-1148" => array("type"=>"city", "pop"=>3582, "name_ru"=>"Госнелл", "name_en"=>"Gosnell", "lang"=>"en", "lat"=>35.961057, "lon"=>-89.972370),
"/US/US-AR/gould-1149" => array("type"=>"city", "pop"=>1143, "name_ru"=>"Гулд", "name_en"=>"Gould", "lang"=>"en", "lat"=>33.985411, "lon"=>-91.562706),
"/US/US-AR/grady-1150" => array("type"=>"city", "pop"=>456, "name_ru"=>"Грейди", "name_en"=>"Grady", "lang"=>"en", "lat"=>34.078716, "lon"=>-91.698318),
"/US/US-AR/grannis-1151" => array("type"=>"town", "pop"=>572, "name_ru"=>"Граннис", "name_en"=>"Grannis", "lang"=>"en", "lat"=>34.240052, "lon"=>-94.320320),
"/US/US-AR/gravel-ridge-1152" => array("type"=>"cdp", "pop"=>3846, "name_ru"=>"Гравел Ридж", "name_en"=>"Gravel Ridge", "lang"=>"en", "lat"=>34.871384, "lon"=>-92.187724),
"/US/US-AR/gravette-1153" => array("type"=>"city", "pop"=>2564, "name_ru"=>"Грейветт", "name_en"=>"Gravette", "lang"=>"en", "lat"=>36.422002, "lon"=>-94.454145),
"/US/US-AR/green-forest-1155" => array("type"=>"city", "pop"=>3049, "name_ru"=>"Грин Форест", "name_en"=>"Green Forest", "lang"=>"en", "lat"=>36.336126, "lon"=>-93.429060),
"/US/US-AR/greenbrier-1154" => array("type"=>"city", "pop"=>4452, "name_ru"=>"Гринбрир", "name_en"=>"Greenbrier", "lang"=>"en", "lat"=>35.229802, "lon"=>-92.383543),
"/US/US-AR/greenland-1156" => array("type"=>"town", "pop"=>1238, "name_ru"=>"Гринленд", "name_en"=>"Greenland", "lang"=>"en", "lat"=>36.000187, "lon"=>-94.183592),
"/US/US-AR/greenway-1157" => array("type"=>"town", "pop"=>222, "name_ru"=>"Гринуэй", "name_en"=>"Greenway", "lang"=>"en", "lat"=>36.341092, "lon"=>-90.222202),
"/US/US-AR/greenwood-1158" => array("type"=>"city", "pop"=>8635, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>35.211667, "lon"=>-94.235133),
"/US/US-AR/greers-ferry-1159" => array("type"=>"city", "pop"=>963, "name_ru"=>"Грирс Ферри", "name_en"=>"Greers Ferry", "lang"=>"en", "lat"=>35.567773, "lon"=>-92.169163),
"/US/US-AR/griffithville-1160" => array("type"=>"town", "pop"=>280, "name_ru"=>"Гриффитвилл", "name_en"=>"Griffithville", "lang"=>"en", "lat"=>35.124779, "lon"=>-91.645576),
"/US/US-AR/grubbs-1161" => array("type"=>"town", "pop"=>395, "name_ru"=>"Грубс", "name_en"=>"Grubbs", "lang"=>"en", "lat"=>35.652706, "lon"=>-91.074715),
"/US/US-AR/guion-1162" => array("type"=>"town", "pop"=>86, "name_ru"=>"Гаен", "name_en"=>"Guion", "lang"=>"en", "lat"=>35.929128, "lon"=>-91.940059),
"/US/US-AR/gum-springs-1163" => array("type"=>"town", "pop"=>194, "name_ru"=>"Гам Спрингс", "name_en"=>"Gum Springs", "lang"=>"en", "lat"=>34.064471, "lon"=>-93.095880),
"/US/US-AR/gurdon-1164" => array("type"=>"city", "pop"=>2285, "name_ru"=>"Гердон", "name_en"=>"Gurdon", "lang"=>"en", "lat"=>33.916435, "lon"=>-93.148141),
"/US/US-AR/guy-1165" => array("type"=>"town", "pop"=>560, "name_ru"=>"Гай", "name_en"=>"Guy", "lang"=>"en", "lat"=>35.324504, "lon"=>-92.334781),
"/US/US-AR/hackett-1166" => array("type"=>"city", "pop"=>761, "name_ru"=>"Хакетт", "name_en"=>"Hackett", "lang"=>"en", "lat"=>35.187533, "lon"=>-94.410317),
"/US/US-AR/hamburg-1167" => array("type"=>"city", "pop"=>2721, "name_ru"=>"Гамбург", "name_en"=>"Hamburg", "lang"=>"en", "lat"=>33.225806, "lon"=>-91.797045),
"/US/US-AR/hampton-1168" => array("type"=>"city", "pop"=>1468, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>33.536717, "lon"=>-92.465368),
"/US/US-AR/hardy-1169" => array("type"=>"city", "pop"=>809, "name_ru"=>"Харди", "name_en"=>"Hardy", "lang"=>"en", "lat"=>36.321492, "lon"=>-91.480417),
"/US/US-AR/harrell-1170" => array("type"=>"city", "pop"=>278, "name_ru"=>"Харрел", "name_en"=>"Harrell", "lang"=>"en", "lat"=>33.510001, "lon"=>-92.401651),
"/US/US-AR/harrisburg-1171" => array("type"=>"city", "pop"=>2117, "name_ru"=>"Харрисбург", "name_en"=>"Harrisburg", "lang"=>"en", "lat"=>35.560298, "lon"=>-90.721130),
"/US/US-AR/harrison-1172" => array("type"=>"city", "pop"=>13200, "name_ru"=>"Харрисон", "name_en"=>"Harrison", "lang"=>"en", "lat"=>36.241883, "lon"=>-93.119460),
"/US/US-AR/hartford-1173" => array("type"=>"city", "pop"=>788, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>35.023260, "lon"=>-94.378182),
"/US/US-AR/hartman-1174" => array("type"=>"city", "pop"=>639, "name_ru"=>"Хартман", "name_en"=>"Hartman", "lang"=>"en", "lat"=>35.434468, "lon"=>-93.618484),
"/US/US-AR/haskell-1175" => array("type"=>"city", "pop"=>3813, "name_ru"=>"Хаскел", "name_en"=>"Haskell", "lang"=>"en", "lat"=>34.505458, "lon"=>-92.636643),
"/US/US-AR/hatfield-1176" => array("type"=>"town", "pop"=>389, "name_ru"=>"Хэтфилд", "name_en"=>"Hatfield", "lang"=>"en", "lat"=>34.487103, "lon"=>-94.378659),
"/US/US-AR/havana-1177" => array("type"=>"city", "pop"=>404, "name_ru"=>"Гавана", "name_en"=>"Havana", "lang"=>"en", "lat"=>35.110232, "lon"=>-93.527921),
"/US/US-AR/haynes-1178" => array("type"=>"town", "pop"=>176, "name_ru"=>"Хайнс", "name_en"=>"Haynes", "lang"=>"en", "lat"=>34.889811, "lon"=>-90.792371),
"/US/US-AR/hazen-1179" => array("type"=>"city", "pop"=>1459, "name_ru"=>"Хазен", "name_en"=>"Hazen", "lang"=>"en", "lat"=>34.789298, "lon"=>-91.573738),
"/US/US-AR/heber-springs-1180" => array("type"=>"city", "pop"=>7203, "name_ru"=>"Хебер Спрингс", "name_en"=>"Heber Springs", "lang"=>"en", "lat"=>35.500186, "lon"=>-92.038592),
"/US/US-AR/hector-1181" => array("type"=>"town", "pop"=>531, "name_ru"=>"Гектор", "name_en"=>"Hector", "lang"=>"en", "lat"=>35.465699, "lon"=>-92.975251),
"/US/US-AR/helena-1182" => array("type"=>"city", "pop"=>7491, "name_ru"=>"Хелена", "name_en"=>"Helena", "lang"=>"en", "lat"=>34.526257, "lon"=>-90.604698),
"/US/US-AR/hermitage-1183" => array("type"=>"town", "pop"=>750, "name_ru"=>"Хермитедж", "name_en"=>"Hermitage", "lang"=>"en", "lat"=>33.447234, "lon"=>-92.170209),
"/US/US-AR/hickory-ridge-1184" => array("type"=>"city", "pop"=>351, "name_ru"=>"Хикори Ридж", "name_en"=>"Hickory Ridge", "lang"=>"en", "lat"=>35.403525, "lon"=>-90.994692),
"/US/US-AR/higden-1185" => array("type"=>"town", "pop"=>105, "name_ru"=>"Хайден", "name_en"=>"Higden", "lang"=>"en", "lat"=>35.566363, "lon"=>-92.205187),
"/US/US-AR/higginson-1186" => array("type"=>"town", "pop"=>408, "name_ru"=>"Хиггинсон", "name_en"=>"Higginson", "lang"=>"en", "lat"=>35.195789, "lon"=>-91.711479),
"/US/US-AR/highfill-1187" => array("type"=>"town", "pop"=>775, "name_ru"=>"Хайфилл", "name_en"=>"Highfill", "lang"=>"en", "lat"=>36.259237, "lon"=>-94.356268),
"/US/US-AR/hindsville-1188" => array("type"=>"town", "pop"=>83, "name_ru"=>"Хиндсвилл", "name_en"=>"Hindsville", "lang"=>"en", "lat"=>36.147569, "lon"=>-93.860076),
"/US/US-AR/holly-grove-1189" => array("type"=>"town", "pop"=>580, "name_ru"=>"Холли Гров", "name_en"=>"Holly Grove", "lang"=>"en", "lat"=>34.596165, "lon"=>-91.200667),
"/US/US-AR/hope-1190" => array("type"=>"city", "pop"=>10378, "name_ru"=>"Хоуп", "name_en"=>"Hope", "lang"=>"en", "lat"=>33.668223, "lon"=>-93.591234),
"/US/US-AR/horatio-1191" => array("type"=>"city", "pop"=>1025, "name_ru"=>"Хорейшо", "name_en"=>"Horatio", "lang"=>"en", "lat"=>33.939833, "lon"=>-94.355108),
"/US/US-AR/horseshoe-bend-1192" => array("type"=>"city", "pop"=>2220, "name_ru"=>"Хорсшу Бенд", "name_en"=>"Horseshoe Bend", "lang"=>"en", "lat"=>36.223312, "lon"=>-91.740954),
"/US/US-AR/horseshoe-lake-1193" => array("type"=>"town", "pop"=>290, "name_ru"=>"Хорсшу Лейк", "name_en"=>"Horseshoe Lake", "lang"=>"en", "lat"=>34.913338, "lon"=>-90.307035),
"/US/US-AR/hot-springs-1194" => array("type"=>"city", "pop"=>39467, "name_ru"=>"Хот Спрингс", "name_en"=>"Hot Springs", "lang"=>"en", "lat"=>34.493672, "lon"=>-93.052042),
"/US/US-AR/hot-springs-village-1195" => array("type"=>"cdp", "pop"=>6361, "name_ru"=>"Хот Спрингс Виллидж", "name_en"=>"Hot Springs Village", "lang"=>"en", "lat"=>34.665171, "lon"=>-93.000065),
"/US/US-AR/houston-1196" => array("type"=>"town", "pop"=>160, "name_ru"=>"Хьюстон", "name_en"=>"Houston", "lang"=>"en", "lat"=>35.034912, "lon"=>-92.694124),
"/US/US-AR/hoxie-1197" => array("type"=>"city", "pop"=>2647, "name_ru"=>"Хокси", "name_en"=>"Hoxie", "lang"=>"en", "lat"=>36.043709, "lon"=>-90.977491),
"/US/US-AR/hughes-1198" => array("type"=>"city", "pop"=>1685, "name_ru"=>"Хьюз", "name_en"=>"Hughes", "lang"=>"en", "lat"=>34.948477, "lon"=>-90.471270),
"/US/US-AR/humnoke-1199" => array("type"=>"city", "pop"=>289, "name_ru"=>"Хамнок", "name_en"=>"Humnoke", "lang"=>"en", "lat"=>34.542143, "lon"=>-91.758838),
"/US/US-AR/humphrey-1200" => array("type"=>"town", "pop"=>755, "name_ru"=>"Хамфри", "name_en"=>"Humphrey", "lang"=>"en", "lat"=>34.422617, "lon"=>-91.702285),
"/US/US-AR/hunter-1201" => array("type"=>"town", "pop"=>130, "name_ru"=>"Хантер", "name_en"=>"Hunter", "lang"=>"en", "lat"=>35.053363, "lon"=>-91.122745),
"/US/US-AR/huntington-1202" => array("type"=>"city", "pop"=>706, "name_ru"=>"Хантингтон", "name_en"=>"Huntington", "lang"=>"en", "lat"=>35.081796, "lon"=>-94.265984),
"/US/US-AR/huntsville-1203" => array("type"=>"city", "pop"=>2401, "name_ru"=>"Хантсвилл", "name_en"=>"Huntsville", "lang"=>"en", "lat"=>36.089081, "lon"=>-93.737230),
"/US/US-AR/huttig-1204" => array("type"=>"town", "pop"=>695, "name_ru"=>"Хаттиг", "name_en"=>"Huttig", "lang"=>"en", "lat"=>33.041275, "lon"=>-92.182249),
"/US/US-AR/imboden-1205" => array("type"=>"town", "pop"=>627, "name_ru"=>"Имбоден", "name_en"=>"Imboden", "lang"=>"en", "lat"=>36.201698, "lon"=>-91.179771),
"/US/US-AR/jacksonport-1206" => array("type"=>"town", "pop"=>206, "name_ru"=>"Джексонпорт", "name_en"=>"Jacksonport", "lang"=>"en", "lat"=>35.641209, "lon"=>-91.307773),
"/US/US-AR/jacksonville-1207" => array("type"=>"city", "pop"=>31351, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>34.869755, "lon"=>-92.127992),
"/US/US-AR/jasper-1208" => array("type"=>"city", "pop"=>482, "name_ru"=>"Джаспер", "name_en"=>"Jasper", "lang"=>"en", "lat"=>36.006054, "lon"=>-93.188294),
"/US/US-AR/jennette-1209" => array("type"=>"town", "pop"=>118, "name_ru"=>"Дженнетт", "name_en"=>"Jennette", "lang"=>"en", "lat"=>35.165561, "lon"=>-90.407515),
"/US/US-AR/jericho-1210" => array("type"=>"town", "pop"=>174, "name_ru"=>"Джерико", "name_en"=>"Jericho", "lang"=>"en", "lat"=>35.285907, "lon"=>-90.227305),
"/US/US-AR/jerome-1211" => array("type"=>"city", "pop"=>45, "name_ru"=>"Джером", "name_en"=>"Jerome", "lang"=>"en", "lat"=>33.398818, "lon"=>-91.470816),
"/US/US-AR/johnson-1212" => array("type"=>"city", "pop"=>3309, "name_ru"=>"Джонсон", "name_en"=>"Johnson", "lang"=>"en", "lat"=>36.133174, "lon"=>-94.172967),
"/US/US-AR/joiner-1213" => array("type"=>"city", "pop"=>499, "name_ru"=>"Джойнер", "name_en"=>"Joiner", "lang"=>"en", "lat"=>35.508311, "lon"=>-90.149739),
"/US/US-AR/jonesboro-1214" => array("type"=>"city", "pop"=>63960, "name_ru"=>"Джонсборо", "name_en"=>"Jonesboro", "lang"=>"en", "lat"=>35.821828, "lon"=>-90.685768),
"/US/US-AR/judsonia-1215" => array("type"=>"city", "pop"=>2180, "name_ru"=>"Джудзония", "name_en"=>"Judsonia", "lang"=>"en", "lat"=>35.275902, "lon"=>-91.640170),
"/US/US-AR/junction-city-1216" => array("type"=>"city", "pop"=>671, "name_ru"=>"Джанкшен Сити", "name_en"=>"Junction City", "lang"=>"en", "lat"=>33.024110, "lon"=>-92.723598),
"/US/US-AR/keiser-1217" => array("type"=>"town", "pop"=>801, "name_ru"=>"Кейзер", "name_en"=>"Keiser", "lang"=>"en", "lat"=>35.674485, "lon"=>-90.094762),
"/US/US-AR/kensett-1218" => array("type"=>"city", "pop"=>1846, "name_ru"=>"Кенсетт", "name_en"=>"Kensett", "lang"=>"en", "lat"=>35.235479, "lon"=>-91.670107),
"/US/US-AR/keo-1219" => array("type"=>"town", "pop"=>240, "name_ru"=>"Кео", "name_en"=>"Keo", "lang"=>"en", "lat"=>34.613057, "lon"=>-92.011133),
"/US/US-AR/kibler-1220" => array("type"=>"town", "pop"=>1031, "name_ru"=>"Киблер", "name_en"=>"Kibler", "lang"=>"en", "lat"=>35.425232, "lon"=>-94.236048),
"/US/US-AR/kingsland-1221" => array("type"=>"city", "pop"=>437, "name_ru"=>"Кингсленд", "name_en"=>"Kingsland", "lang"=>"en", "lat"=>33.861294, "lon"=>-92.294052),
"/US/US-AR/knobel-1222" => array("type"=>"town", "pop"=>326, "name_ru"=>"Нобел", "name_en"=>"Knobel", "lang"=>"en", "lat"=>36.319297, "lon"=>-90.601670),
"/US/US-AR/knoxville-1223" => array("type"=>"city", "pop"=>775, "name_ru"=>"Ноксвилл", "name_en"=>"Knoxville", "lang"=>"en", "lat"=>35.377831, "lon"=>-93.364787),
"/US/US-AR/lafe-1224" => array("type"=>"town", "pop"=>398, "name_ru"=>"Лафе", "name_en"=>"Lafe", "lang"=>"en", "lat"=>36.207081, "lon"=>-90.503120),
"/US/US-AR/lagrange-1225" => array("type"=>"town", "pop"=>112, "name_ru"=>"Лаграндж", "name_en"=>"LaGrange", "lang"=>"en", "lat"=>34.655514, "lon"=>-90.733726),
"/US/US-AR/lake-city-1226" => array("type"=>"town", "pop"=>2130, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>35.818970, "lon"=>-90.448580),
"/US/US-AR/lake-hamilton-1227" => array("type"=>"cdp", "pop"=>1331, "name_ru"=>"Лэйк Хамилтон", "name_en"=>"Lake Hamilton", "lang"=>"en", "lat"=>34.426023, "lon"=>-93.089268),
"/US/US-AR/lake-view-1229" => array("type"=>"town", "pop"=>450, "name_ru"=>"Лейк Вью", "name_en"=>"Lake View", "lang"=>"en", "lat"=>34.423149, "lon"=>-90.810284),
"/US/US-AR/lake-village-1230" => array("type"=>"city", "pop"=>2386, "name_ru"=>"Лейк Виллидж", "name_en"=>"Lake Village", "lang"=>"en", "lat"=>33.324203, "lon"=>-91.283532),
"/US/US-AR/lakeview-1228" => array("type"=>"town", "pop"=>847, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>36.375167, "lon"=>-92.539165),
"/US/US-AR/lamar-1231" => array("type"=>"city", "pop"=>1635, "name_ru"=>"Ламар", "name_en"=>"Lamar", "lang"=>"en", "lat"=>35.439201, "lon"=>-93.387796),
"/US/US-AR/lavaca-1232" => array("type"=>"city", "pop"=>2189, "name_ru"=>"Лавака", "name_en"=>"Lavaca", "lang"=>"en", "lat"=>35.334980, "lon"=>-94.181567),
"/US/US-AR/leachville-1233" => array("type"=>"city", "pop"=>1852, "name_ru"=>"Личвилл", "name_en"=>"Leachville", "lang"=>"en", "lat"=>35.921546, "lon"=>-90.256248),
"/US/US-AR/lead-hill-1234" => array("type"=>"town", "pop"=>307, "name_ru"=>"Лид Хилл", "name_en"=>"Lead Hill", "lang"=>"en", "lat"=>36.414723, "lon"=>-92.906736),
"/US/US-AR/leola-1235" => array("type"=>"town", "pop"=>541, "name_ru"=>"Леола", "name_en"=>"Leola", "lang"=>"en", "lat"=>34.169321, "lon"=>-92.590719),
"/US/US-AR/lepanto-1236" => array("type"=>"city", "pop"=>2019, "name_ru"=>"Лепанто", "name_en"=>"Lepanto", "lang"=>"en", "lat"=>35.607086, "lon"=>-90.333493),
"/US/US-AR/leslie-1237" => array("type"=>"city", "pop"=>467, "name_ru"=>"Лесли", "name_en"=>"Leslie", "lang"=>"en", "lat"=>35.829448, "lon"=>-92.555918),
"/US/US-AR/letona-1238" => array("type"=>"town", "pop"=>215, "name_ru"=>"Летона", "name_en"=>"Letona", "lang"=>"en", "lat"=>35.363082, "lon"=>-91.830150),
"/US/US-AR/lewisville-1239" => array("type"=>"city", "pop"=>1148, "name_ru"=>"Льюисвилл", "name_en"=>"Lewisville", "lang"=>"en", "lat"=>33.362966, "lon"=>-93.576618),
"/US/US-AR/lexa-1240" => array("type"=>"city", "pop"=>278, "name_ru"=>"Лекса", "name_en"=>"Lexa", "lang"=>"en", "lat"=>34.598203, "lon"=>-90.752614),
"/US/US-AR/lincoln-1241" => array("type"=>"city", "pop"=>2085, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>35.947224, "lon"=>-94.422081),
"/US/US-AR/little-flock-1242" => array("type"=>"city", "pop"=>3142, "name_ru"=>"Литл Флок", "name_en"=>"Little Flock", "lang"=>"en", "lat"=>36.387845, "lon"=>-94.136486),
"/US/US-AR/little-rock-1243" => array("type"=>"city", "pop"=>189515, "name_ru"=>"Литл Рок", "name_en"=>"Little Rock", "lang"=>"en", "lat"=>34.722400, "lon"=>-92.354076),
"/US/US-AR/lockesburg-1244" => array("type"=>"town", "pop"=>738, "name_ru"=>"Локсбург", "name_en"=>"Lockesburg", "lang"=>"en", "lat"=>33.970912, "lon"=>-94.173082),
"/US/US-AR/london-1245" => array("type"=>"city", "pop"=>1046, "name_ru"=>"Лондон", "name_en"=>"London", "lang"=>"en", "lat"=>35.328171, "lon"=>-93.240067),
"/US/US-AR/lonoke-1246" => array("type"=>"city", "pop"=>4600, "name_ru"=>"Лонок", "name_en"=>"Lonoke", "lang"=>"en", "lat"=>34.786011, "lon"=>-91.899551),
"/US/US-AR/lonsdale-1247" => array("type"=>"town", "pop"=>132, "name_ru"=>"Лонсдейл", "name_en"=>"Lonsdale", "lang"=>"en", "lat"=>34.544672, "lon"=>-92.811078),
"/US/US-AR/louann-1248" => array("type"=>"town", "pop"=>179, "name_ru"=>"Луэнн", "name_en"=>"Louann", "lang"=>"en", "lat"=>33.391590, "lon"=>-92.792748),
"/US/US-AR/lowell-1249" => array("type"=>"city", "pop"=>7173, "name_ru"=>"Лоуэлл", "name_en"=>"Lowell", "lang"=>"en", "lat"=>36.267300, "lon"=>-94.136673),
"/US/US-AR/luxora-1250" => array("type"=>"town", "pop"=>1219, "name_ru"=>"Лаксора", "name_en"=>"Luxora", "lang"=>"en", "lat"=>35.757668, "lon"=>-89.930233),
"/US/US-AR/lynn-1251" => array("type"=>"town", "pop"=>301, "name_ru"=>"Линн", "name_en"=>"Lynn", "lang"=>"en", "lat"=>36.006569, "lon"=>-91.251992),
"/US/US-AR/madison-1259" => array("type"=>"city", "pop"=>853, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>35.019419, "lon"=>-90.730510),
"/US/US-AR/magazine-1260" => array("type"=>"city", "pop"=>892, "name_ru"=>"Магазин", "name_en"=>"Magazine", "lang"=>"en", "lat"=>35.152692, "lon"=>-93.806329),
"/US/US-AR/magness-1261" => array("type"=>"town", "pop"=>193, "name_ru"=>"Магнесс", "name_en"=>"Magness", "lang"=>"en", "lat"=>35.701878, "lon"=>-91.482319),
"/US/US-AR/magnolia-1262" => array("type"=>"city", "pop"=>11081, "name_ru"=>"Магнолия", "name_en"=>"Magnolia", "lang"=>"en", "lat"=>33.274016, "lon"=>-93.235379),
"/US/US-AR/malvern-1263" => array("type"=>"city", "pop"=>8958, "name_ru"=>"Малверн", "name_en"=>"Malvern", "lang"=>"en", "lat"=>34.371124, "lon"=>-92.815565),
"/US/US-AR/mammoth-spring-1264" => array("type"=>"city", "pop"=>1112, "name_ru"=>"Маммот Спринг", "name_en"=>"Mammoth Spring", "lang"=>"en", "lat"=>36.494133, "lon"=>-91.544698),
"/US/US-AR/manila-1265" => array("type"=>"city", "pop"=>2828, "name_ru"=>"Манила", "name_en"=>"Manila", "lang"=>"en", "lat"=>35.883589, "lon"=>-90.161451),
"/US/US-AR/mansfield-1266" => array("type"=>"city", "pop"=>1120, "name_ru"=>"Мансфилд", "name_en"=>"Mansfield", "lang"=>"en", "lat"=>35.059073, "lon"=>-94.246744),
"/US/US-AR/marianna-1267" => array("type"=>"city", "pop"=>4396, "name_ru"=>"Марианна", "name_en"=>"Marianna", "lang"=>"en", "lat"=>34.774584, "lon"=>-90.766352),
"/US/US-AR/marie-1268" => array("type"=>"town", "pop"=>99, "name_ru"=>"Мари", "name_en"=>"Marie", "lang"=>"en", "lat"=>35.613285, "lon"=>-90.081687),
"/US/US-AR/marion-1269" => array("type"=>"city", "pop"=>12217, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>35.206516, "lon"=>-90.203462),
"/US/US-AR/marked-tree-1270" => array("type"=>"city", "pop"=>2634, "name_ru"=>"Маркед Три", "name_en"=>"Marked Tree", "lang"=>"en", "lat"=>35.529954, "lon"=>-90.417457),
"/US/US-AR/marmaduke-1271" => array("type"=>"city", "pop"=>1174, "name_ru"=>"Мармадьюк", "name_en"=>"Marmaduke", "lang"=>"en", "lat"=>36.188351, "lon"=>-90.387775),
"/US/US-AR/marshall-1272" => array("type"=>"city", "pop"=>1284, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>35.909875, "lon"=>-92.636207),
"/US/US-AR/marvell-1273" => array("type"=>"city", "pop"=>1128, "name_ru"=>"Марвелл", "name_en"=>"Marvell", "lang"=>"en", "lat"=>34.555991, "lon"=>-90.914514),
"/US/US-AR/maumelle-1274" => array("type"=>"city", "pop"=>16201, "name_ru"=>"Маумел", "name_en"=>"Maumelle", "lang"=>"en", "lat"=>34.852480, "lon"=>-92.410534),
"/US/US-AR/mayflower-1275" => array("type"=>"city", "pop"=>2231, "name_ru"=>"Мейфлауэр", "name_en"=>"Mayflower", "lang"=>"en", "lat"=>34.971815, "lon"=>-92.419024),
"/US/US-AR/maynard-1276" => array("type"=>"town", "pop"=>356, "name_ru"=>"Мейнард", "name_en"=>"Maynard", "lang"=>"en", "lat"=>36.421847, "lon"=>-90.901909),
"/US/US-AR/mccaskill-1252" => array("type"=>"city", "pop"=>81, "name_ru"=>"Маккаскилл", "name_en"=>"McCaskill", "lang"=>"en", "lat"=>33.917571, "lon"=>-93.641109),
"/US/US-AR/mccrory-1253" => array("type"=>"city", "pop"=>1549, "name_ru"=>"Мак-Крори", "name_en"=>"McCrory", "lang"=>"en", "lat"=>35.256855, "lon"=>-91.197745),
"/US/US-AR/mcdougal-1254" => array("type"=>"town", "pop"=>172, "name_ru"=>"Макдугал", "name_en"=>"McDougal", "lang"=>"en", "lat"=>36.436602, "lon"=>-90.390230),
"/US/US-AR/mcgehee-1255" => array("type"=>"city", "pop"=>3950, "name_ru"=>"Мак-Ги", "name_en"=>"McGehee", "lang"=>"en", "lat"=>33.625353, "lon"=>-91.392533),
"/US/US-AR/mcnab-1256" => array("type"=>"town", "pop"=>75, "name_ru"=>"Макнэб", "name_en"=>"McNab", "lang"=>"en", "lat"=>33.660716, "lon"=>-93.832225),
"/US/US-AR/mcneil-1257" => array("type"=>"city", "pop"=>625, "name_ru"=>"Макнейл", "name_en"=>"McNeil", "lang"=>"en", "lat"=>33.347876, "lon"=>-93.208805),
"/US/US-AR/mcrae-1258" => array("type"=>"city", "pop"=>707, "name_ru"=>"Мак-Рай", "name_en"=>"McRae", "lang"=>"en", "lat"=>35.114193, "lon"=>-91.822564),
"/US/US-AR/melbourne-1277" => array("type"=>"city", "pop"=>1685, "name_ru"=>"Мельбурн", "name_en"=>"Melbourne", "lang"=>"en", "lat"=>36.062337, "lon"=>-91.898652),
"/US/US-AR/mena-1278" => array("type"=>"city", "pop"=>5630, "name_ru"=>"Мена", "name_en"=>"Mena", "lang"=>"en", "lat"=>34.580450, "lon"=>-94.236696),
"/US/US-AR/menifee-1279" => array("type"=>"city", "pop"=>320, "name_ru"=>"Менифи", "name_en"=>"Menifee", "lang"=>"en", "lat"=>35.148312, "lon"=>-92.554820),
"/US/US-AR/midland-1280" => array("type"=>"town", "pop"=>258, "name_ru"=>"Мидленд", "name_en"=>"Midland", "lang"=>"en", "lat"=>35.093000, "lon"=>-94.352662),
"/US/US-AR/mineral-springs-1281" => array("type"=>"city", "pop"=>1269, "name_ru"=>"Минерал Спрингс", "name_en"=>"Mineral Springs", "lang"=>"en", "lat"=>33.877647, "lon"=>-93.919964),
"/US/US-AR/minturn-1282" => array("type"=>"town", "pop"=>110, "name_ru"=>"Минтум", "name_en"=>"Minturn", "lang"=>"en", "lat"=>35.975002, "lon"=>-91.027498),
"/US/US-AR/mitchellville-1283" => array("type"=>"city", "pop"=>478, "name_ru"=>"Митчелвилл", "name_en"=>"Mitchellville", "lang"=>"en", "lat"=>33.905872, "lon"=>-91.498487),
"/US/US-AR/monette-1284" => array("type"=>"city", "pop"=>1229, "name_ru"=>"Монетт", "name_en"=>"Monette", "lang"=>"en", "lat"=>35.893660, "lon"=>-90.344632),
"/US/US-AR/monticello-1285" => array("type"=>"city", "pop"=>9270, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>33.624763, "lon"=>-91.793784),
"/US/US-AR/montrose-1286" => array("type"=>"city", "pop"=>498, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>33.298565, "lon"=>-91.497030),
"/US/US-AR/moorefield-1287" => array("type"=>"town", "pop"=>160, "name_ru"=>"Мурфилд", "name_en"=>"Moorefield", "lang"=>"en", "lat"=>35.768123, "lon"=>-91.570961),
"/US/US-AR/moro-1288" => array("type"=>"town", "pop"=>196, "name_ru"=>"Моро", "name_en"=>"Moro", "lang"=>"en", "lat"=>34.792530, "lon"=>-90.989274),
"/US/US-AR/morrilton-1289" => array("type"=>"city", "pop"=>6581, "name_ru"=>"Моррилтон", "name_en"=>"Morrilton", "lang"=>"en", "lat"=>35.156241, "lon"=>-92.740892),
"/US/US-AR/morrison-bluff-1290" => array("type"=>"town", "pop"=>75, "name_ru"=>"Моррисон Блафф", "name_en"=>"Morrison Bluff", "lang"=>"en", "lat"=>35.382189, "lon"=>-93.527432),
"/US/US-AR/mount-ida-1295" => array("type"=>"city", "pop"=>945, "name_ru"=>"Маунт Ида", "name_en"=>"Mount Ida", "lang"=>"en", "lat"=>34.551784, "lon"=>-93.631119),
"/US/US-AR/mount-pleasant-1296" => array("type"=>"town", "pop"=>392, "name_ru"=>"Маунт Плезант", "name_en"=>"Mount Pleasant", "lang"=>"en", "lat"=>35.982389, "lon"=>-91.777680),
"/US/US-AR/mount-vernon-1297" => array("type"=>"town", "pop"=>155, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>35.224037, "lon"=>-92.130080),
"/US/US-AR/mountain-home-1292" => array("type"=>"city", "pop"=>12592, "name_ru"=>"Маунтин Хоум", "name_en"=>"Mountain Home", "lang"=>"en", "lat"=>36.335916, "lon"=>-92.379171),
"/US/US-AR/mountain-pine-1293" => array("type"=>"city", "pop"=>860, "name_ru"=>"Маунтин Пайн", "name_en"=>"Mountain Pine", "lang"=>"en", "lat"=>34.571426, "lon"=>-93.173951),
"/US/US-AR/mountain-view-1294" => array("type"=>"city", "pop"=>3112, "name_ru"=>"Маунтин Вью", "name_en"=>"Mountain View", "lang"=>"en", "lat"=>35.869951, "lon"=>-92.111107),
"/US/US-AR/mountainburg-1291" => array("type"=>"city", "pop"=>713, "name_ru"=>"Маунтинбург", "name_en"=>"Mountainburg", "lang"=>"en", "lat"=>35.634780, "lon"=>-94.165864),
"/US/US-AR/mulberry-1298" => array("type"=>"city", "pop"=>1719, "name_ru"=>"Малберри", "name_en"=>"Mulberry", "lang"=>"en", "lat"=>35.508409, "lon"=>-94.074215),
"/US/US-AR/murfreesboro-1299" => array("type"=>"city", "pop"=>1642, "name_ru"=>"Мерфрисборо", "name_en"=>"Murfreesboro", "lang"=>"en", "lat"=>34.063002, "lon"=>-93.691123),
"/US/US-AR/nashville-1300" => array("type"=>"city", "pop"=>4815, "name_ru"=>"Нэшвилл", "name_en"=>"Nashville", "lang"=>"en", "lat"=>33.940923, "lon"=>-93.851622),
"/US/US-AR/newark-1301" => array("type"=>"city", "pop"=>1227, "name_ru"=>"Ньюарк", "name_en"=>"Newark", "lang"=>"en", "lat"=>35.705340, "lon"=>-91.442858),
"/US/US-AR/newport-1302" => array("type"=>"city", "pop"=>7404, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>35.625546, "lon"=>-91.235159),
"/US/US-AR/nimmons-1303" => array("type"=>"town", "pop"=>92, "name_ru"=>"Ниммонс", "name_en"=>"Nimmons", "lang"=>"en", "lat"=>36.306965, "lon"=>-90.094498),
"/US/US-AR/norfork-1304" => array("type"=>"city", "pop"=>549, "name_ru"=>"Норфорк", "name_en"=>"Norfork", "lang"=>"en", "lat"=>36.208327, "lon"=>-92.281996),
"/US/US-AR/norman-1305" => array("type"=>"town", "pop"=>411, "name_ru"=>"Норман", "name_en"=>"Norman", "lang"=>"en", "lat"=>34.456677, "lon"=>-93.677786),
"/US/US-AR/norphlet-1306" => array("type"=>"city", "pop"=>788, "name_ru"=>"Норфлет", "name_en"=>"Norphlet", "lang"=>"en", "lat"=>33.319647, "lon"=>-92.663153),
"/US/US-AR/north-crossett-1307" => array("type"=>"cdp", "pop"=>3358, "name_ru"=>"Норт Кроссетт", "name_en"=>"North Crossett", "lang"=>"en", "lat"=>33.169800, "lon"=>-91.936218),
"/US/US-AR/north-little-rock-1308" => array("type"=>"city", "pop"=>59430, "name_ru"=>"Норт Литтл Рок", "name_en"=>"North Little Rock", "lang"=>"en", "lat"=>34.788400, "lon"=>-92.255437),
"/US/US-AR/o'kean-1315" => array("type"=>"town", "pop"=>196, "name_ru"=>"О'кен", "name_en"=>"O'Kean", "lang"=>"en", "lat"=>36.167799, "lon"=>-90.817207),
"/US/US-AR/oak-grove-1309" => array("type"=>"town", "pop"=>400, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>36.459686, "lon"=>-93.435898),
"/US/US-AR/oak-grove-heights-1310" => array("type"=>"town", "pop"=>820, "name_ru"=>"Оук Гров Хайтс", "name_en"=>"Oak Grove Heights", "lang"=>"en", "lat"=>36.127340, "lon"=>-90.515180),
"/US/US-AR/oakhaven-1311" => array("type"=>"city", "pop"=>51, "name_ru"=>"Оукхейвен", "name_en"=>"Oakhaven", "lang"=>"en", "lat"=>33.729130, "lon"=>-93.619952),
"/US/US-AR/oden-1312" => array("type"=>"town", "pop"=>212, "name_ru"=>"Оден", "name_en"=>"Oden", "lang"=>"en", "lat"=>34.620237, "lon"=>-93.785248),
"/US/US-AR/ogden-1313" => array("type"=>"town", "pop"=>203, "name_ru"=>"Огден", "name_en"=>"Ogden", "lang"=>"en", "lat"=>33.584847, "lon"=>-94.044513),
"/US/US-AR/oil-trough-1314" => array("type"=>"city", "pop"=>253, "name_ru"=>"Ойл Тро", "name_en"=>"Oil Trough", "lang"=>"en", "lat"=>35.629494, "lon"=>-91.461233),
"/US/US-AR/okolona-1316" => array("type"=>"town", "pop"=>158, "name_ru"=>"Околона", "name_en"=>"Okolona", "lang"=>"en", "lat"=>34.000486, "lon"=>-93.337187),
"/US/US-AR/ola-1317" => array("type"=>"city", "pop"=>1238, "name_ru"=>"Ола", "name_en"=>"Ola", "lang"=>"en", "lat"=>35.031002, "lon"=>-93.224087),
"/US/US-AR/omaha-1318" => array("type"=>"town", "pop"=>177, "name_ru"=>"Омаха", "name_en"=>"Omaha", "lang"=>"en", "lat"=>36.463215, "lon"=>-93.191876),
"/US/US-AR/oppelo-1319" => array("type"=>"city", "pop"=>748, "name_ru"=>"Оппело", "name_en"=>"Oppelo", "lang"=>"en", "lat"=>35.101014, "lon"=>-92.775129),
"/US/US-AR/osceola-1320" => array("type"=>"city", "pop"=>7914, "name_ru"=>"Осеола", "name_en"=>"Osceola", "lang"=>"en", "lat"=>35.701277, "lon"=>-89.994498),
"/US/US-AR/oxford-1321" => array("type"=>"city", "pop"=>629, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>36.207269, "lon"=>-91.919396),
"/US/US-AR/ozan-1322" => array("type"=>"city", "pop"=>77, "name_ru"=>"Озан", "name_en"=>"Ozan", "lang"=>"en", "lat"=>33.847339, "lon"=>-93.719426),
"/US/US-AR/ozark-1323" => array("type"=>"city", "pop"=>3582, "name_ru"=>"Озарк", "name_en"=>"Ozark", "lang"=>"en", "lat"=>35.496243, "lon"=>-93.840504),
"/US/US-AR/palestine-1324" => array("type"=>"city", "pop"=>686, "name_ru"=>"Палестин", "name_en"=>"Palestine", "lang"=>"en", "lat"=>34.970296, "lon"=>-90.905007),
"/US/US-AR/pangburn-1325" => array("type"=>"city", "pop"=>700, "name_ru"=>"Пангберн", "name_en"=>"Pangburn", "lang"=>"en", "lat"=>35.426167, "lon"=>-91.838669),
"/US/US-AR/paragould-1326" => array("type"=>"city", "pop"=>24800, "name_ru"=>"Парагулд", "name_en"=>"Paragould", "lang"=>"en", "lat"=>36.056123, "lon"=>-90.513515),
"/US/US-AR/paris-1327" => array("type"=>"city", "pop"=>3613, "name_ru"=>"Парис", "name_en"=>"Paris", "lang"=>"en", "lat"=>35.288875, "lon"=>-93.724606),
"/US/US-AR/parkdale-1328" => array("type"=>"city", "pop"=>348, "name_ru"=>"Паркдейл", "name_en"=>"Parkdale", "lang"=>"en", "lat"=>33.120779, "lon"=>-91.545842),
"/US/US-AR/parkers-iron-springs-1329" => array("type"=>"cdp", "pop"=>3611, "name_ru"=>"Паркерс-Айрон Спрингс", "name_en"=>"Parkers-Iron Springs", "lang"=>"en", "lat"=>34.605246, "lon"=>-92.322218),
"/US/US-AR/parkin-1330" => array("type"=>"city", "pop"=>1479, "name_ru"=>"Паркин", "name_en"=>"Parkin", "lang"=>"en", "lat"=>35.264471, "lon"=>-90.552454),
"/US/US-AR/patmos-1331" => array("type"=>"town", "pop"=>59, "name_ru"=>"Патмос", "name_en"=>"Patmos", "lang"=>"en", "lat"=>33.511943, "lon"=>-93.566663),
"/US/US-AR/patterson-1332" => array("type"=>"town", "pop"=>394, "name_ru"=>"Паттерсон", "name_en"=>"Patterson", "lang"=>"en", "lat"=>35.257567, "lon"=>-91.236613),
"/US/US-AR/pea-ridge-1334" => array("type"=>"city", "pop"=>4604, "name_ru"=>"Пи Ридж", "name_en"=>"Pea Ridge", "lang"=>"en", "lat"=>36.450521, "lon"=>-94.115902),
"/US/US-AR/peach-orchard-1333" => array("type"=>"town", "pop"=>174, "name_ru"=>"Пич Орчард", "name_en"=>"Peach Orchard", "lang"=>"en", "lat"=>36.280940, "lon"=>-90.661931),
"/US/US-AR/perla-1335" => array("type"=>"town", "pop"=>124, "name_ru"=>"Перла", "name_en"=>"Perla", "lang"=>"en", "lat"=>34.364442, "lon"=>-92.778604),
"/US/US-AR/perry-1336" => array("type"=>"town", "pop"=>308, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>35.044871, "lon"=>-92.793817),
"/US/US-AR/perrytown-1337" => array("type"=>"city", "pop"=>247, "name_ru"=>"Перритаун", "name_en"=>"Perrytown", "lang"=>"en", "lat"=>33.696509, "lon"=>-93.532593),
"/US/US-AR/perryville-1338" => array("type"=>"city", "pop"=>1442, "name_ru"=>"Перривилл", "name_en"=>"Perryville", "lang"=>"en", "lat"=>35.008437, "lon"=>-92.803460),
"/US/US-AR/piggott-1339" => array("type"=>"city", "pop"=>3482, "name_ru"=>"Пиггот", "name_en"=>"Piggott", "lang"=>"en", "lat"=>36.383881, "lon"=>-90.200814),
"/US/US-AR/pindall-1340" => array("type"=>"town", "pop"=>92, "name_ru"=>"Пиндалл", "name_en"=>"Pindall", "lang"=>"en", "lat"=>36.058345, "lon"=>-92.865268),
"/US/US-AR/pine-bluff-1341" => array("type"=>"city", "pop"=>50408, "name_ru"=>"Пайн Блафф", "name_en"=>"Pine Bluff", "lang"=>"en", "lat"=>34.213400, "lon"=>-92.018618),
"/US/US-AR/pineville-1342" => array("type"=>"town", "pop"=>243, "name_ru"=>"Пайнвилл", "name_en"=>"Pineville", "lang"=>"en", "lat"=>36.157697, "lon"=>-92.103924),
"/US/US-AR/piney-1343" => array("type"=>"cdp", "pop"=>2500, "name_ru"=>"Пайни", "name_en"=>"Piney", "lang"=>"en", "lat"=>34.503434, "lon"=>-93.142649),
"/US/US-AR/plainview-1344" => array("type"=>"city", "pop"=>776, "name_ru"=>"Плейнвью", "name_en"=>"Plainview", "lang"=>"en", "lat"=>34.989464, "lon"=>-93.297644),
"/US/US-AR/pleasant-plains-1345" => array("type"=>"town", "pop"=>270, "name_ru"=>"Плезант Плейнс", "name_en"=>"Pleasant Plains", "lang"=>"en", "lat"=>35.550638, "lon"=>-91.626953),
"/US/US-AR/plumerville-1346" => array("type"=>"city", "pop"=>864, "name_ru"=>"Пламервилл", "name_en"=>"Plumerville", "lang"=>"en", "lat"=>35.160603, "lon"=>-92.641221),
"/US/US-AR/pocahontas-1347" => array("type"=>"city", "pop"=>6713, "name_ru"=>"Покахонтас", "name_en"=>"Pocahontas", "lang"=>"en", "lat"=>36.263516, "lon"=>-90.968585),
"/US/US-AR/pollard-1348" => array("type"=>"town", "pop"=>218, "name_ru"=>"Поллард", "name_en"=>"Pollard", "lang"=>"en", "lat"=>36.430729, "lon"=>-90.266203),
"/US/US-AR/portia-1349" => array("type"=>"town", "pop"=>447, "name_ru"=>"Портиа", "name_en"=>"Portia", "lang"=>"en", "lat"=>36.085294, "lon"=>-91.071315),
"/US/US-AR/portland-1350" => array("type"=>"city", "pop"=>506, "name_ru"=>"Портленд", "name_en"=>"Portland", "lang"=>"en", "lat"=>33.238289, "lon"=>-91.511179),
"/US/US-AR/pottsville-1351" => array("type"=>"town", "pop"=>2805, "name_ru"=>"Поттсвилл", "name_en"=>"Pottsville", "lang"=>"en", "lat"=>35.238375, "lon"=>-93.058405),
"/US/US-AR/powhatan-1352" => array("type"=>"town", "pop"=>48, "name_ru"=>"Похатан", "name_en"=>"Powhatan", "lang"=>"en", "lat"=>36.081938, "lon"=>-91.118920),
"/US/US-AR/poyen-1353" => array("type"=>"town", "pop"=>282, "name_ru"=>"Пойен", "name_en"=>"Poyen", "lang"=>"en", "lat"=>34.322426, "lon"=>-92.642297),
"/US/US-AR/prairie-creek-1354" => array("type"=>"cdp", "pop"=>1268, "name_ru"=>"Прейри Крик", "name_en"=>"Prairie Creek", "lang"=>"en", "lat"=>36.338859, "lon"=>-94.061444),
"/US/US-AR/prairie-grove-1355" => array("type"=>"city", "pop"=>3741, "name_ru"=>"Прейри Гров", "name_en"=>"Prairie Grove", "lang"=>"en", "lat"=>35.978251, "lon"=>-94.317521),
"/US/US-AR/prattsville-1356" => array("type"=>"town", "pop"=>294, "name_ru"=>"Праттсвилл", "name_en"=>"Prattsville", "lang"=>"en", "lat"=>34.318115, "lon"=>-92.542959),
"/US/US-AR/prescott-1357" => array("type"=>"city", "pop"=>4353, "name_ru"=>"Прескотт", "name_en"=>"Prescott", "lang"=>"en", "lat"=>33.800696, "lon"=>-93.388000),
"/US/US-AR/pyatt-1358" => array("type"=>"town", "pop"=>253, "name_ru"=>"Пайат", "name_en"=>"Pyatt", "lang"=>"en", "lat"=>36.251260, "lon"=>-92.844418),
"/US/US-AR/quitman-1359" => array("type"=>"city", "pop"=>736, "name_ru"=>"Квитман", "name_en"=>"Quitman", "lang"=>"en", "lat"=>35.382269, "lon"=>-92.214278),
"/US/US-AR/ratcliff-1360" => array("type"=>"city", "pop"=>192, "name_ru"=>"Ратклиф", "name_en"=>"Ratcliff", "lang"=>"en", "lat"=>35.307962, "lon"=>-93.888233),
"/US/US-AR/ravenden-1361" => array("type"=>"town", "pop"=>492, "name_ru"=>"Равенден", "name_en"=>"Ravenden", "lang"=>"en", "lat"=>36.236199, "lon"=>-91.250444),
"/US/US-AR/ravenden-springs-1362" => array("type"=>"town", "pop"=>133, "name_ru"=>"Равенден Спрингс", "name_en"=>"Ravenden Springs", "lang"=>"en", "lat"=>36.314580, "lon"=>-91.224049),
"/US/US-AR/reader-1363" => array("type"=>"town", "pop"=>56, "name_ru"=>"Ридер", "name_en"=>"Reader", "lang"=>"en", "lat"=>33.751794, "lon"=>-93.101221),
"/US/US-AR/rector-1364" => array("type"=>"city", "pop"=>1758, "name_ru"=>"Ректор", "name_en"=>"Rector", "lang"=>"en", "lat"=>36.264242, "lon"=>-90.293344),
"/US/US-AR/redfield-1365" => array("type"=>"town", "pop"=>1173, "name_ru"=>"Редфилд", "name_en"=>"Redfield", "lang"=>"en", "lat"=>34.441891, "lon"=>-92.183852),
"/US/US-AR/reed-1366" => array("type"=>"town", "pop"=>248, "name_ru"=>"Рид", "name_en"=>"Reed", "lang"=>"en", "lat"=>33.701823, "lon"=>-91.443730),
"/US/US-AR/reyno-1367" => array("type"=>"town", "pop"=>470, "name_ru"=>"Рейно", "name_en"=>"Reyno", "lang"=>"en", "lat"=>36.360774, "lon"=>-90.757943),
"/US/US-AR/rison-1368" => array("type"=>"city", "pop"=>1281, "name_ru"=>"Рисон", "name_en"=>"Rison", "lang"=>"en", "lat"=>33.954456, "lon"=>-92.190689),
"/US/US-AR/rockport-1369" => array("type"=>"town", "pop"=>807, "name_ru"=>"Рокпорт", "name_en"=>"Rockport", "lang"=>"en", "lat"=>34.391874, "lon"=>-92.816267),
"/US/US-AR/rockwell-1370" => array("type"=>"cdp", "pop"=>2514, "name_ru"=>"Рокуэлл", "name_en"=>"Rockwell", "lang"=>"en", "lat"=>34.464259, "lon"=>-93.133634),
"/US/US-AR/roe-1371" => array("type"=>"town", "pop"=>104, "name_ru"=>"Ро", "name_en"=>"Roe", "lang"=>"en", "lat"=>34.630850, "lon"=>-91.385727),
"/US/US-AR/rogers-1372" => array("type"=>"city", "pop"=>56726, "name_ru"=>"Роджерс", "name_en"=>"Rogers", "lang"=>"en", "lat"=>36.331250, "lon"=>-94.133449),
"/US/US-AR/rondo-1373" => array("type"=>"town", "pop"=>196, "name_ru"=>"Рондо", "name_en"=>"Rondo", "lang"=>"en", "lat"=>34.657440, "lon"=>-90.820223),
"/US/US-AR/rose-bud-1374" => array("type"=>"town", "pop"=>457, "name_ru"=>"Роуз Бад", "name_en"=>"Rose Bud", "lang"=>"en", "lat"=>35.330884, "lon"=>-92.078734),
"/US/US-AR/rosston-1375" => array("type"=>"town", "pop"=>204, "name_ru"=>"Росстон", "name_en"=>"Rosston", "lang"=>"en", "lat"=>33.586914, "lon"=>-93.276039),
"/US/US-AR/rudy-1376" => array("type"=>"town", "pop"=>77, "name_ru"=>"Ради", "name_en"=>"Rudy", "lang"=>"en", "lat"=>35.527133, "lon"=>-94.270298),
"/US/US-AR/russell-1377" => array("type"=>"town", "pop"=>244, "name_ru"=>"Расселл", "name_en"=>"Russell", "lang"=>"en", "lat"=>35.362079, "lon"=>-91.510371),
"/US/US-AR/russellville-1378" => array("type"=>"city", "pop"=>27602, "name_ru"=>"Расселлвилл", "name_en"=>"Russellville", "lang"=>"en", "lat"=>35.271784, "lon"=>-93.140765),
"/US/US-AR/salem-1382" => array("type"=>"city", "pop"=>1614, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>36.371143, "lon"=>-91.823089),
"/US/US-AR/salem-1383" => array("type"=>"cdp", "pop"=>2950, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>34.629724, "lon"=>-92.555544),
"/US/US-AR/salesville-1384" => array("type"=>"town", "pop"=>483, "name_ru"=>"Сейлсвилл", "name_en"=>"Salesville", "lang"=>"en", "lat"=>36.244895, "lon"=>-92.270021),
"/US/US-AR/scranton-1385" => array("type"=>"city", "pop"=>259, "name_ru"=>"Скрантон", "name_en"=>"Scranton", "lang"=>"en", "lat"=>35.360260, "lon"=>-93.538476),
"/US/US-AR/searcy-1386" => array("type"=>"city", "pop"=>22299, "name_ru"=>"Сирси", "name_en"=>"Searcy", "lang"=>"en", "lat"=>35.242880, "lon"=>-91.730790),
"/US/US-AR/sedgwick-1387" => array("type"=>"town", "pop"=>207, "name_ru"=>"Седжвик", "name_en"=>"Sedgwick", "lang"=>"en", "lat"=>35.978516, "lon"=>-90.863194),
"/US/US-AR/shannon-hills-1388" => array("type"=>"city", "pop"=>3160, "name_ru"=>"Шэннон Хиллс", "name_en"=>"Shannon Hills", "lang"=>"en", "lat"=>34.620630, "lon"=>-92.408460),
"/US/US-AR/sheridan-1389" => array("type"=>"city", "pop"=>4572, "name_ru"=>"Шеридан", "name_en"=>"Sheridan", "lang"=>"en", "lat"=>34.309266, "lon"=>-92.405466),
"/US/US-AR/sherrill-1390" => array("type"=>"town", "pop"=>114, "name_ru"=>"Шеррилл", "name_en"=>"Sherrill", "lang"=>"en", "lat"=>34.384715, "lon"=>-91.950703),
"/US/US-AR/sherwood-1391" => array("type"=>"city", "pop"=>24542, "name_ru"=>"Шервуд", "name_en"=>"Sherwood", "lang"=>"en", "lat"=>34.832017, "lon"=>-92.207486),
"/US/US-AR/shirley-1392" => array("type"=>"town", "pop"=>337, "name_ru"=>"Ширли", "name_en"=>"Shirley", "lang"=>"en", "lat"=>35.643695, "lon"=>-92.313387),
"/US/US-AR/sidney-1393" => array("type"=>"town", "pop"=>287, "name_ru"=>"Сидни", "name_en"=>"Sidney", "lang"=>"en", "lat"=>36.004255, "lon"=>-91.658974),
"/US/US-AR/siloam-springs-1394" => array("type"=>"city", "pop"=>14825, "name_ru"=>"Силоэм Спрингс", "name_en"=>"Siloam Springs", "lang"=>"en", "lat"=>36.180173, "lon"=>-94.540455),
"/US/US-AR/smackover-1395" => array("type"=>"city", "pop"=>1881, "name_ru"=>"Смаковер", "name_en"=>"Smackover", "lang"=>"en", "lat"=>33.364593, "lon"=>-92.730822),
"/US/US-AR/smithville-1396" => array("type"=>"town", "pop"=>69, "name_ru"=>"Смитвилл", "name_en"=>"Smithville", "lang"=>"en", "lat"=>36.080042, "lon"=>-91.302844),
"/US/US-AR/south-lead-hill-1397" => array("type"=>"town", "pop"=>93, "name_ru"=>"Сауф Лид Хилл", "name_en"=>"South Lead Hill", "lang"=>"en", "lat"=>36.395010, "lon"=>-92.905424),
"/US/US-AR/sparkman-1398" => array("type"=>"town", "pop"=>511, "name_ru"=>"Спаркман", "name_en"=>"Sparkman", "lang"=>"en", "lat"=>33.917099, "lon"=>-92.848598),
"/US/US-AR/springdale-1399" => array("type"=>"city", "pop"=>68180, "name_ru"=>"Спрингдейл", "name_en"=>"Springdale", "lang"=>"en", "lat"=>36.181333, "lon"=>-94.145713),
"/US/US-AR/st.-charles-1379" => array("type"=>"town", "pop"=>243, "name_ru"=>"Сент-Чарльз", "name_en"=>"St. Charles", "lang"=>"en", "lat"=>34.375773, "lon"=>-91.132934),
"/US/US-AR/st.-francis-1380" => array("type"=>"city", "pop"=>228, "name_ru"=>"Сент-Фрэнсис", "name_en"=>"St. Francis", "lang"=>"en", "lat"=>36.453523, "lon"=>-90.143154),
"/US/US-AR/st.-paul-1381" => array("type"=>"town", "pop"=>176, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>35.823477, "lon"=>-93.764524),
"/US/US-AR/stamps-1400" => array("type"=>"city", "pop"=>1910, "name_ru"=>"Стампс", "name_en"=>"Stamps", "lang"=>"en", "lat"=>33.358658, "lon"=>-93.495704),
"/US/US-AR/star-city-1401" => array("type"=>"city", "pop"=>2229, "name_ru"=>"Стар Сити", "name_en"=>"Star City", "lang"=>"en", "lat"=>33.940972, "lon"=>-91.844364),
"/US/US-AR/stephens-1402" => array("type"=>"city", "pop"=>1017, "name_ru"=>"Стивенс", "name_en"=>"Stephens", "lang"=>"en", "lat"=>33.420069, "lon"=>-93.067427),
"/US/US-AR/strawberry-1403" => array("type"=>"town", "pop"=>271, "name_ru"=>"Строберри", "name_en"=>"Strawberry", "lang"=>"en", "lat"=>35.965570, "lon"=>-91.321442),
"/US/US-AR/strong-1404" => array("type"=>"city", "pop"=>626, "name_ru"=>"Стронг", "name_en"=>"Strong", "lang"=>"en", "lat"=>33.108043, "lon"=>-92.359374),
"/US/US-AR/stuttgart-1405" => array("type"=>"city", "pop"=>8971, "name_ru"=>"Статгарт", "name_en"=>"Stuttgart", "lang"=>"en", "lat"=>34.493522, "lon"=>-91.550915),
"/US/US-AR/subiaco-1406" => array("type"=>"town", "pop"=>442, "name_ru"=>"Сабиако", "name_en"=>"Subiaco", "lang"=>"en", "lat"=>35.297487, "lon"=>-93.639212),
"/US/US-AR/success-1407" => array("type"=>"town", "pop"=>157, "name_ru"=>"Саксесс", "name_en"=>"Success", "lang"=>"en", "lat"=>36.454518, "lon"=>-90.722854),
"/US/US-AR/sulphur-rock-1408" => array("type"=>"town", "pop"=>427, "name_ru"=>"Салфер Рок", "name_en"=>"Sulphur Rock", "lang"=>"en", "lat"=>35.750991, "lon"=>-91.499305),
"/US/US-AR/sulphur-springs-1409" => array("type"=>"city", "pop"=>691, "name_ru"=>"Салфер Спрингс", "name_en"=>"Sulphur Springs", "lang"=>"en", "lat"=>36.482788, "lon"=>-94.458607),
"/US/US-AR/summit-1410" => array("type"=>"city", "pop"=>582, "name_ru"=>"Саммит", "name_en"=>"Summit", "lang"=>"en", "lat"=>36.250142, "lon"=>-92.687427),
"/US/US-AR/sunset-1411" => array("type"=>"town", "pop"=>324, "name_ru"=>"Сансет", "name_en"=>"Sunset", "lang"=>"en", "lat"=>35.224658, "lon"=>-90.207601),
"/US/US-AR/swifton-1412" => array("type"=>"city", "pop"=>785, "name_ru"=>"Свифтон", "name_en"=>"Swifton", "lang"=>"en", "lat"=>35.824199, "lon"=>-91.128876),
"/US/US-AR/taylor-1413" => array("type"=>"city", "pop"=>532, "name_ru"=>"Тейлор", "name_en"=>"Taylor", "lang"=>"en", "lat"=>33.100329, "lon"=>-93.462106),
"/US/US-AR/texarkana-1414" => array("type"=>"city", "pop"=>30087, "name_ru"=>"Тексаркана", "name_en"=>"Texarkana", "lang"=>"en", "lat"=>33.450245, "lon"=>-94.015754),
"/US/US-AR/thornton-1415" => array("type"=>"city", "pop"=>491, "name_ru"=>"Торнтон", "name_en"=>"Thornton", "lang"=>"en", "lat"=>33.776218, "lon"=>-92.489189),
"/US/US-AR/tillar-1416" => array("type"=>"city", "pop"=>222, "name_ru"=>"Тиллар", "name_en"=>"Tillar", "lang"=>"en", "lat"=>33.710186, "lon"=>-91.451695),
"/US/US-AR/tinsman-1417" => array("type"=>"city", "pop"=>71, "name_ru"=>"Тинсман", "name_en"=>"Tinsman", "lang"=>"en", "lat"=>33.629013, "lon"=>-92.353521),
"/US/US-AR/tollette-1418" => array("type"=>"town", "pop"=>320, "name_ru"=>"Толлетт", "name_en"=>"Tollette", "lang"=>"en", "lat"=>33.818038, "lon"=>-93.896082),
"/US/US-AR/tontitown-1419" => array("type"=>"town", "pop"=>2019, "name_ru"=>"Тонтитаун", "name_en"=>"Tontitown", "lang"=>"en", "lat"=>36.174964, "lon"=>-94.236884),
"/US/US-AR/traskwood-1420" => array("type"=>"town", "pop"=>566, "name_ru"=>"Трасквуд", "name_en"=>"Traskwood", "lang"=>"en", "lat"=>34.451933, "lon"=>-92.666744),
"/US/US-AR/trumann-1421" => array("type"=>"city", "pop"=>6794, "name_ru"=>"Труман", "name_en"=>"Trumann", "lang"=>"en", "lat"=>35.674699, "lon"=>-90.519520),
"/US/US-AR/tuckerman-1422" => array("type"=>"city", "pop"=>1644, "name_ru"=>"Такерман", "name_en"=>"Tuckerman", "lang"=>"en", "lat"=>35.727875, "lon"=>-91.200588),
"/US/US-AR/tull-1423" => array("type"=>"town", "pop"=>374, "name_ru"=>"Талл", "name_en"=>"Tull", "lang"=>"en", "lat"=>34.442337, "lon"=>-92.573486),
"/US/US-AR/tupelo-1424" => array("type"=>"town", "pop"=>160, "name_ru"=>"Тупело", "name_en"=>"Tupelo", "lang"=>"en", "lat"=>35.391546, "lon"=>-91.229943),
"/US/US-AR/turrell-1425" => array("type"=>"city", "pop"=>866, "name_ru"=>"Террелл", "name_en"=>"Turrell", "lang"=>"en", "lat"=>35.377076, "lon"=>-90.257070),
"/US/US-AR/tyronza-1426" => array("type"=>"town", "pop"=>865, "name_ru"=>"Тайронза", "name_en"=>"Tyronza", "lang"=>"en", "lat"=>35.490332, "lon"=>-90.358904),
"/US/US-AR/ulm-1427" => array("type"=>"town", "pop"=>199, "name_ru"=>"Алм", "name_en"=>"Ulm", "lang"=>"en", "lat"=>34.576003, "lon"=>-91.461253),
"/US/US-AR/valley-springs-1428" => array("type"=>"town", "pop"=>178, "name_ru"=>"Валли Спрингс", "name_en"=>"Valley Springs", "lang"=>"en", "lat"=>36.152868, "lon"=>-92.991040),
"/US/US-AR/van-buren-1429" => array("type"=>"city", "pop"=>22543, "name_ru"=>"Ван Бурен", "name_en"=>"Van Buren", "lang"=>"en", "lat"=>35.446650, "lon"=>-94.352849),
"/US/US-AR/vandervoort-1430" => array("type"=>"town", "pop"=>119, "name_ru"=>"Вандервоорт", "name_en"=>"Vandervoort", "lang"=>"en", "lat"=>34.379628, "lon"=>-94.364453),
"/US/US-AR/victoria-1431" => array("type"=>"town", "pop"=>54, "name_ru"=>"Виктория", "name_en"=>"Victoria", "lang"=>"en", "lat"=>35.758100, "lon"=>-90.060683),
"/US/US-AR/vilonia-1432" => array("type"=>"town", "pop"=>3485, "name_ru"=>"Вилония", "name_en"=>"Vilonia", "lang"=>"en", "lat"=>35.084009, "lon"=>-92.211236),
"/US/US-AR/viola-1433" => array("type"=>"town", "pop"=>383, "name_ru"=>"Виола", "name_en"=>"Viola", "lang"=>"en", "lat"=>36.396156, "lon"=>-91.984945),
"/US/US-AR/wabbaseka-1434" => array("type"=>"city", "pop"=>296, "name_ru"=>"Ваббасика", "name_en"=>"Wabbaseka", "lang"=>"en", "lat"=>34.360292, "lon"=>-91.792724),
"/US/US-AR/waldenburg-1435" => array("type"=>"town", "pop"=>77, "name_ru"=>"Уолденбург", "name_en"=>"Waldenburg", "lang"=>"en", "lat"=>35.564863, "lon"=>-90.934830),
"/US/US-AR/waldo-1436" => array("type"=>"city", "pop"=>1437, "name_ru"=>"Уолдо", "name_en"=>"Waldo", "lang"=>"en", "lat"=>33.352467, "lon"=>-93.295324),
"/US/US-AR/waldron-1437" => array("type"=>"city", "pop"=>3569, "name_ru"=>"Уолдрон", "name_en"=>"Waldron", "lang"=>"en", "lat"=>34.899708, "lon"=>-94.093573),
"/US/US-AR/walnut-ridge-1438" => array("type"=>"city", "pop"=>4664, "name_ru"=>"Уолнат Ридж", "name_en"=>"Walnut Ridge", "lang"=>"en", "lat"=>36.069829, "lon"=>-90.958349),
"/US/US-AR/ward-1439" => array("type"=>"city", "pop"=>3908, "name_ru"=>"Уорд", "name_en"=>"Ward", "lang"=>"en", "lat"=>35.018859, "lon"=>-91.955031),
"/US/US-AR/warren-1440" => array("type"=>"city", "pop"=>6118, "name_ru"=>"Уоррен", "name_en"=>"Warren", "lang"=>"en", "lat"=>33.610137, "lon"=>-92.069666),
"/US/US-AR/washington-1441" => array("type"=>"city", "pop"=>135, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>33.773148, "lon"=>-93.682958),
"/US/US-AR/watson-1442" => array("type"=>"city", "pop"=>261, "name_ru"=>"Уотсон", "name_en"=>"Watson", "lang"=>"en", "lat"=>33.893580, "lon"=>-91.257366),
"/US/US-AR/weiner-1443" => array("type"=>"city", "pop"=>723, "name_ru"=>"Вейнер", "name_en"=>"Weiner", "lang"=>"en", "lat"=>35.619158, "lon"=>-90.902043),
"/US/US-AR/weldon-1444" => array("type"=>"town", "pop"=>91, "name_ru"=>"Уэлдон", "name_en"=>"Weldon", "lang"=>"en", "lat"=>35.447544, "lon"=>-91.231439),
"/US/US-AR/west-crossett-1445" => array("type"=>"cdp", "pop"=>2019, "name_ru"=>"Уэст Кроссетт", "name_en"=>"West Crossett", "lang"=>"en", "lat"=>33.144395, "lon"=>-92.019934),
"/US/US-AR/west-fork-1447" => array("type"=>"city", "pop"=>2326, "name_ru"=>"Вэст Форк", "name_en"=>"West Fork", "lang"=>"en", "lat"=>35.932478, "lon"=>-94.182318),
"/US/US-AR/west-helena-1448" => array("type"=>"city", "pop"=>9695, "name_ru"=>"Вэст Хелена", "name_en"=>"West Helena", "lang"=>"en", "lat"=>34.545959, "lon"=>-90.646490),
"/US/US-AR/west-memphis-1449" => array("type"=>"city", "pop"=>27070, "name_ru"=>"Вэст Мемфис", "name_en"=>"West Memphis", "lang"=>"en", "lat"=>35.149747, "lon"=>-90.182163),
"/US/US-AR/west-point-1450" => array("type"=>"town", "pop"=>222, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>35.204281, "lon"=>-91.610350),
"/US/US-AR/western-grove-1446" => array("type"=>"town", "pop"=>397, "name_ru"=>"Вестерн Гров", "name_en"=>"Western Grove", "lang"=>"en", "lat"=>36.101120, "lon"=>-92.953922),
"/US/US-AR/wheatley-1451" => array("type"=>"city", "pop"=>327, "name_ru"=>"Уитли", "name_en"=>"Wheatley", "lang"=>"en", "lat"=>34.918793, "lon"=>-91.107304),
"/US/US-AR/whelen-springs-1452" => array("type"=>"town", "pop"=>83, "name_ru"=>"Уэлен Спрингс", "name_en"=>"Whelen Springs", "lang"=>"en", "lat"=>33.831585, "lon"=>-93.125730),
"/US/US-AR/white-hall-1453" => array("type"=>"city", "pop"=>5121, "name_ru"=>"Уайт Холл", "name_en"=>"White Hall", "lang"=>"en", "lat"=>34.275567, "lon"=>-92.097805),
"/US/US-AR/wickes-1454" => array("type"=>"town", "pop"=>678, "name_ru"=>"Уикес", "name_en"=>"Wickes", "lang"=>"en", "lat"=>34.299734, "lon"=>-94.335223),
"/US/US-AR/widener-1455" => array("type"=>"town", "pop"=>300, "name_ru"=>"Уиденер", "name_en"=>"Widener", "lang"=>"en", "lat"=>35.021963, "lon"=>-90.682438),
"/US/US-AR/wiederkehr-village-1456" => array("type"=>"city", "pop"=>47, "name_ru"=>"Уиедеркер Виллидж", "name_en"=>"Wiederkehr Village", "lang"=>"en", "lat"=>35.482373, "lon"=>-93.757319),
"/US/US-AR/williford-1457" => array("type"=>"town", "pop"=>66, "name_ru"=>"Уиллифорд", "name_en"=>"Williford", "lang"=>"en", "lat"=>36.251381, "lon"=>-91.360705),
"/US/US-AR/willisville-1458" => array("type"=>"town", "pop"=>142, "name_ru"=>"Уиллисвилл", "name_en"=>"Willisville", "lang"=>"en", "lat"=>33.518456, "lon"=>-93.294778),
"/US/US-AR/wilmar-1459" => array("type"=>"city", "pop"=>560, "name_ru"=>"Уилмар", "name_en"=>"Wilmar", "lang"=>"en", "lat"=>33.625916, "lon"=>-91.928969),
"/US/US-AR/wilmot-1460" => array("type"=>"city", "pop"=>706, "name_ru"=>"Уилмот", "name_en"=>"Wilmot", "lang"=>"en", "lat"=>33.058282, "lon"=>-91.575019),
"/US/US-AR/wilson-1461" => array("type"=>"town", "pop"=>831, "name_ru"=>"Уилсон", "name_en"=>"Wilson", "lang"=>"en", "lat"=>35.566168, "lon"=>-90.041929),
"/US/US-AR/wilton-1462" => array("type"=>"town", "pop"=>420, "name_ru"=>"Уилтон", "name_en"=>"Wilton", "lang"=>"en", "lat"=>33.738798, "lon"=>-94.147459),
"/US/US-AR/winchester-1463" => array("type"=>"city", "pop"=>189, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>33.774788, "lon"=>-91.475009),
"/US/US-AR/winslow-1464" => array("type"=>"city", "pop"=>393, "name_ru"=>"Уинслоу", "name_en"=>"Winslow", "lang"=>"en", "lat"=>35.800937, "lon"=>-94.131694),
"/US/US-AR/winthrop-1465" => array("type"=>"city", "pop"=>177, "name_ru"=>"Уинтроп", "name_en"=>"Winthrop", "lang"=>"en", "lat"=>33.831222, "lon"=>-94.353426),
"/US/US-AR/wooster-1466" => array("type"=>"town", "pop"=>838, "name_ru"=>"Вустер", "name_en"=>"Wooster", "lang"=>"en", "lat"=>35.198880, "lon"=>-92.448299),
"/US/US-AR/wrightsville-1467" => array("type"=>"city", "pop"=>1718, "name_ru"=>"Райтсвилл", "name_en"=>"Wrightsville", "lang"=>"en", "lat"=>34.614014, "lon"=>-92.219905),
"/US/US-AR/wynne-1468" => array("type"=>"city", "pop"=>8437, "name_ru"=>"Винн", "name_en"=>"Wynne", "lang"=>"en", "lat"=>35.231267, "lon"=>-90.788414),
"/US/US-AR/yellville-1469" => array("type"=>"city", "pop"=>1378, "name_ru"=>"Йеллвилл", "name_en"=>"Yellville", "lang"=>"en", "lat"=>36.228739, "lon"=>-92.685939),
"/US/US-AR/zinc-1470" => array("type"=>"town", "pop"=>82, "name_ru"=>"Зинк", "name_en"=>"Zinc", "lang"=>"en", "lat"=>36.285314, "lon"=>-92.915255),
"/US/US-AS" => array("type"=>"outlying area", "name_ru"=>"Американское Самоа", "name_en"=>"American Samoa", "lang"=>"en"),
"/US/US-AS/aasu-23436" => array("type"=>"village", "pop"=>341, "name_ru"=>"Аасу", "name_en"=>"Aasu", "lang"=>"en", "lat"=>-14.309410, "lon"=>-170.756687),
"/US/US-AS/afao-23437" => array("type"=>"village", "pop"=>145, "name_ru"=>"Афао", "name_en"=>"Afao", "lang"=>"en", "lat"=>-14.338802, "lon"=>-170.800725),
"/US/US-AS/afono-23438" => array("type"=>"village", "pop"=>434, "name_ru"=>"Афоно", "name_en"=>"Afono", "lang"=>"en", "lat"=>-14.262155, "lon"=>-170.646079),
"/US/US-AS/agugulu-23439" => array("type"=>"village", "pop"=>42, "name_ru"=>"Агугулу", "name_en"=>"Agugulu", "lang"=>"en", "lat"=>-14.353292, "lon"=>-170.825669),
"/US/US-AS/alao-23440" => array("type"=>"village", "pop"=>463, "name_ru"=>"Алао", "name_en"=>"Alao", "lang"=>"en", "lat"=>-14.270032, "lon"=>-170.528230),
"/US/US-AS/alega-23441" => array("type"=>"village", "pop"=>49, "name_ru"=>"Алега", "name_en"=>"Alega", "lang"=>"en", "lat"=>-14.327011, "lon"=>-170.620617),
"/US/US-AS/alofau-23442" => array("type"=>"village", "pop"=>458, "name_ru"=>"Алофау", "name_en"=>"Alofau", "lang"=>"en", "lat"=>-14.301240, "lon"=>-170.604809),
"/US/US-AS/amaluia-23443" => array("type"=>"village", "pop"=>206, "name_ru"=>"Амалуиа", "name_en"=>"Amaluia", "lang"=>"en", "lat"=>-14.339470, "lon"=>-170.792781),
"/US/US-AS/amanave-23444" => array("type"=>"village", "pop"=>378, "name_ru"=>"Аманаве", "name_en"=>"Amanave", "lang"=>"en", "lat"=>-14.347735, "lon"=>-170.846635),
"/US/US-AS/amaua-23445" => array("type"=>"village", "pop"=>128, "name_ru"=>"Амауа", "name_en"=>"Amaua", "lang"=>"en", "lat"=>-14.276453, "lon"=>-170.623681),
"/US/US-AS/amouli-23446" => array("type"=>"village", "pop"=>463, "name_ru"=>"Амоули", "name_en"=>"Amouli", "lang"=>"en", "lat"=>-14.308231, "lon"=>-170.567327),
"/US/US-AS/anua-23447" => array("type"=>"village", "pop"=>65, "name_ru"=>"Ануа", "name_en"=>"Anua", "lang"=>"en", "lat"=>-14.276008, "lon"=>-170.688332),
"/US/US-AS/aoa-23448" => array("type"=>"village", "pop"=>491, "name_ru"=>"Аоа", "name_en"=>"Aoa", "lang"=>"en", "lat"=>-14.268208, "lon"=>-170.585237),
"/US/US-AS/aoloau-23449" => array("type"=>"village", "pop"=>544, "name_ru"=>"Аолоау", "name_en"=>"Aoloau", "lang"=>"en", "lat"=>-14.293258, "lon"=>-170.781376),
"/US/US-AS/asili-23450" => array("type"=>"village", "pop"=>203, "name_ru"=>"Асили", "name_en"=>"Asili", "lang"=>"en", "lat"=>-14.339648, "lon"=>-170.794708),
"/US/US-AS/atu'u-23451" => array("type"=>"village", "pop"=>408, "name_ru"=>"Атуу", "name_en"=>"Atu'u", "lang"=>"en", "lat"=>-14.274927, "lon"=>-170.683866),
"/US/US-AS/aua-23452" => array("type"=>"village", "pop"=>1896, "name_ru"=>"Ауа", "name_en"=>"Aua", "lang"=>"en", "lat"=>-14.280529, "lon"=>-170.663880),
"/US/US-AS/auasi-23453" => array("type"=>"village", "pop"=>122, "name_ru"=>"Ауаси", "name_en"=>"Auasi", "lang"=>"en", "lat"=>-14.275842, "lon"=>-170.576548),
"/US/US-AS/aumi-23454" => array("type"=>"village", "pop"=>228, "name_ru"=>"Ауми", "name_en"=>"Aumi", "lang"=>"en", "lat"=>-14.315075, "lon"=>-170.636562),
"/US/US-AS/aunu'u-23455" => array("type"=>"village", "pop"=>463, "name_ru"=>"Аунуу", "name_en"=>"Aunu'u", "lang"=>"en", "lat"=>-14.294780, "lon"=>-170.552652),
"/US/US-AS/auto-23456" => array("type"=>"village", "pop"=>255, "name_ru"=>"Ауто", "name_en"=>"Auto", "lang"=>"en", "lat"=>-14.281610, "lon"=>-170.626819),
"/US/US-AS/avaio-23457" => array("type"=>"village", "pop"=>50, "name_ru"=>"Аваио", "name_en"=>"Avaio", "lang"=>"en", "lat"=>-14.309324, "lon"=>-170.615941),
"/US/US-AS/faga'alu-23458" => array("type"=>"village", "pop"=>1006, "name_ru"=>"Фагаалу", "name_en"=>"Faga'alu", "lang"=>"en", "lat"=>-14.294253, "lon"=>-170.680564),
"/US/US-AS/faga'itua-23459" => array("type"=>"village", "pop"=>455, "name_ru"=>"Фагаитуа", "name_en"=>"Faga'itua", "lang"=>"en", "lat"=>-14.278934, "lon"=>-170.614286),
"/US/US-AS/fagali'i-23460" => array("type"=>"village", "pop"=>206, "name_ru"=>"Фагалии", "name_en"=>"Fagali'i", "lang"=>"en", "lat"=>-14.310065, "lon"=>-170.826830),
"/US/US-AS/fagamalo-23461" => array("type"=>"village", "pop"=>92, "name_ru"=>"Фагамало", "name_en"=>"Fagamalo", "lang"=>"en", "lat"=>-14.275365, "lon"=>-170.827994),
"/US/US-AS/faganeanea-23462" => array("type"=>"village", "pop"=>168, "name_ru"=>"Фаганеанеа", "name_en"=>"Faganeanea", "lang"=>"en", "lat"=>-14.305581, "lon"=>-170.693100),
"/US/US-AS/fagasa-23463" => array("type"=>"village", "pop"=>717, "name_ru"=>"Фпгаса", "name_en"=>"Fagasa", "lang"=>"en", "lat"=>-14.256071, "lon"=>-170.761794),
"/US/US-AS/fagatogo-23464" => array("type"=>"village", "pop"=>2323, "name_ru"=>"Фагатого", "name_en"=>"Fagatogo", "lang"=>"en", "lat"=>-14.287385, "lon"=>-170.690560),
"/US/US-AS/failolo-23465" => array("type"=>"village", "pop"=>81, "name_ru"=>"Фаилоло", "name_en"=>"Failolo", "lang"=>"en", "lat"=>-14.358363, "lon"=>-170.834110),
"/US/US-AS/faleasao-23466" => array("type"=>"village", "pop"=>246, "name_ru"=>"Фалеасао", "name_en"=>"Faleasao", "lang"=>"en", "lat"=>-14.195658, "lon"=>-169.496856),
"/US/US-AS/faleniu-23467" => array("type"=>"village", "pop"=>833, "name_ru"=>"Фалениу", "name_en"=>"Faleniu", "lang"=>"en", "lat"=>-14.331069, "lon"=>-170.743499),
"/US/US-AS/fatumafuti-23468" => array("type"=>"village", "pop"=>81, "name_ru"=>"Фатумафути", "name_en"=>"Fatumafuti", "lang"=>"en", "lat"=>-14.300277, "lon"=>-170.672771),
"/US/US-AS/futiga-23469" => array("type"=>"village", "pop"=>720, "name_ru"=>"Футига", "name_en"=>"Futiga", "lang"=>"en", "lat"=>-14.403069, "lon"=>-170.760835),
"/US/US-AS/ili'ili-23470" => array("type"=>"village", "pop"=>1790, "name_ru"=>"Илиили", "name_en"=>"Ili'ili", "lang"=>"en", "lat"=>-14.385522, "lon"=>-170.707020),
"/US/US-AS/lauli'i-23471" => array("type"=>"village", "pop"=>814, "name_ru"=>"Лаулии", "name_en"=>"Lauli'i", "lang"=>"en", "lat"=>-14.313939, "lon"=>-170.657627),
"/US/US-AS/leloaloa-23472" => array("type"=>"village", "pop"=>412, "name_ru"=>"Лелоалоа", "name_en"=>"Leloaloa", "lang"=>"en", "lat"=>-14.272670, "lon"=>-170.675780),
"/US/US-AS/leone-23473" => array("type"=>"village", "pop"=>3013, "name_ru"=>"Леоне", "name_en"=>"Leone", "lang"=>"en", "lat"=>-14.340273, "lon"=>-170.783669),
"/US/US-AS/leusoali'i-23474" => array("type"=>"village", "pop"=>247, "name_ru"=>"Леусоалии", "name_en"=>"Leusoali'i", "lang"=>"en", "lat"=>-14.269204, "lon"=>-169.410157),
"/US/US-AS/luma-23475" => array("type"=>"village", "pop"=>293, "name_ru"=>"Лума", "name_en"=>"Luma", "lang"=>"en", "lat"=>-14.218482, "lon"=>-169.544899),
"/US/US-AS/maia-23476" => array("type"=>"village", "pop"=>207, "name_ru"=>"Маиа", "name_en"=>"Maia", "lang"=>"en", "lat"=>-14.185938, "lon"=>-169.418664),
"/US/US-AS/malaeimi-23477" => array("type"=>"village", "pop"=>830, "name_ru"=>"Малаэими", "name_en"=>"Malaeimi", "lang"=>"en", "lat"=>-14.314659, "lon"=>-170.737161),
"/US/US-AS/malaeloa/aitulagi-23478" => array("type"=>"village", "pop"=>585, "name_ru"=>"Малаэлоа/Аитулаги", "name_en"=>"Malaeloa/Aitulagi", "lang"=>"en", "lat"=>-14.336979, "lon"=>-170.765642),
"/US/US-AS/malaeloa/ituau-23479" => array("type"=>"village", "pop"=>523, "name_ru"=>"Малаэлоа/Итуау", "name_en"=>"Malaeloa/Ituau", "lang"=>"en", "lat"=>-14.346685, "lon"=>-170.768915),
"/US/US-AS/maloata-23480" => array("type"=>"village", "pop"=>16, "name_ru"=>"Малоата", "name_en"=>"Maloata", "lang"=>"en", "lat"=>-14.311416, "lon"=>-170.816359),
"/US/US-AS/mapusagafou-23481" => array("type"=>"village", "pop"=>1133, "name_ru"=>"Мапусагафоу", "name_en"=>"Mapusagafou", "lang"=>"en", "lat"=>-14.324605, "lon"=>-170.750552),
"/US/US-AS/masausi-23482" => array("type"=>"village", "pop"=>157, "name_ru"=>"Масауси", "name_en"=>"Masausi", "lang"=>"en", "lat"=>-14.250285, "lon"=>-170.607304),
"/US/US-AS/masefau-23483" => array("type"=>"village", "pop"=>389, "name_ru"=>"Масефау", "name_en"=>"Masefau", "lang"=>"en", "lat"=>-14.253617, "lon"=>-170.623953),
"/US/US-AS/matu'u-23484" => array("type"=>"village", "pop"=>364, "name_ru"=>"Матуу", "name_en"=>"Matu'u", "lang"=>"en", "lat"=>-14.337338, "lon"=>-170.664080),
"/US/US-AS/mesepa-23485" => array("type"=>"village", "pop"=>483, "name_ru"=>"Месепа", "name_en"=>"Mesepa", "lang"=>"en", "lat"=>-14.325968, "lon"=>-170.741013),
"/US/US-AS/nu'uuli-23487" => array("type"=>"village", "pop"=>3893, "name_ru"=>"Нууули", "name_en"=>"Nu'uuli", "lang"=>"en", "lat"=>-14.334260, "lon"=>-170.701113),
"/US/US-AS/nua-23486" => array("type"=>"village", "pop"=>267, "name_ru"=>"Нуа", "name_en"=>"Nua", "lang"=>"en", "lat"=>-14.365469, "lon"=>-170.807996),
"/US/US-AS/ofu-23488" => array("type"=>"village", "pop"=>353, "name_ru"=>"Офу", "name_en"=>"Ofu", "lang"=>"en", "lat"=>-14.154052, "lon"=>-169.678669),
"/US/US-AS/olosega-23489" => array("type"=>"village", "pop"=>201, "name_ru"=>"Олосега", "name_en"=>"Olosega", "lang"=>"en", "lat"=>-14.201212, "lon"=>-169.599688),
"/US/US-AS/onenoa-23490" => array("type"=>"village", "pop"=>136, "name_ru"=>"Онеона", "name_en"=>"Onenoa", "lang"=>"en", "lat"=>-14.248976, "lon"=>-170.583532),
"/US/US-AS/pagai-23491" => array("type"=>"village", "pop"=>107, "name_ru"=>"Пагаи", "name_en"=>"Pagai", "lang"=>"en", "lat"=>-14.274003, "lon"=>-170.603383),
"/US/US-AS/pago-pago-23492" => array("type"=>"village", "pop"=>3519, "name_ru"=>"Паго-Паго", "name_en"=>"Pago Pago", "lang"=>"en", "lat"=>-14.241200, "lon"=>-170.719577),
"/US/US-AS/pava'ia'i-23493" => array("type"=>"village", "pop"=>1692, "name_ru"=>"Паваиаи", "name_en"=>"Pava'ia'i", "lang"=>"en", "lat"=>-14.338813, "lon"=>-170.753052),
"/US/US-AS/poloa-23494" => array("type"=>"village", "pop"=>176, "name_ru"=>"Полоа", "name_en"=>"Poloa", "lang"=>"en", "lat"=>-14.318268, "lon"=>-170.863482),
"/US/US-AS/sa'ilele-23495" => array("type"=>"village", "pop"=>130, "name_ru"=>"Саилеле", "name_en"=>"Sa'ilele", "lang"=>"en", "lat"=>-14.230761, "lon"=>-170.597286),
"/US/US-AS/se'etaga-23496" => array("type"=>"village", "pop"=>228, "name_ru"=>"Сеэтага", "name_en"=>"Se'etaga", "lang"=>"en", "lat"=>-14.363545, "lon"=>-170.816765),
"/US/US-AS/si'ufaga-23498" => array("type"=>"village", "pop"=>143, "name_ru"=>"Сиуфага", "name_en"=>"Si'ufaga", "lang"=>"en", "lat"=>-14.291047, "lon"=>-169.485796),
"/US/US-AS/sili-23497" => array("type"=>"village", "pop"=>24, "name_ru"=>"Сили", "name_en"=>"Sili", "lang"=>"en", "lat"=>-14.141575, "lon"=>-169.599164),
"/US/US-AS/swains-23499" => array("type"=>"village", "pop"=>16, "name_ru"=>"Суэйнс", "name_en"=>"Swains", "lang"=>"en", "lat"=>-11.103769, "lon"=>-171.048136),
"/US/US-AS/tafuna-23500" => array("type"=>"village", "pop"=>5174, "name_ru"=>"Тафуна", "name_en"=>"Tafuna", "lang"=>"en", "lat"=>-14.374595, "lon"=>-170.687313),
"/US/US-AS/taputimu-23501" => array("type"=>"village", "pop"=>520, "name_ru"=>"Тапутиму", "name_en"=>"Taputimu", "lang"=>"en", "lat"=>-14.357621, "lon"=>-170.770110),
"/US/US-AS/tula-23502" => array("type"=>"village", "pop"=>423, "name_ru"=>"Тула", "name_en"=>"Tula", "lang"=>"en", "lat"=>-14.230136, "lon"=>-170.546707),
"/US/US-AS/utulei-23503" => array("type"=>"village", "pop"=>930, "name_ru"=>"Утулеи", "name_en"=>"Utulei", "lang"=>"en", "lat"=>-14.288119, "lon"=>-170.678529),
"/US/US-AS/utumea-east-23504" => array("type"=>"village", "pop"=>57, "name_ru"=>"Утумеа-Ист", "name_en"=>"Utumea East", "lang"=>"en", "lat"=>-14.282883, "lon"=>-170.567480),
"/US/US-AS/utumea-west-23505" => array("type"=>"village", "pop"=>53, "name_ru"=>"Утумеа-Уэст", "name_en"=>"Utumea West", "lang"=>"en", "lat"=>-14.335483, "lon"=>-170.817207),
"/US/US-AS/vailoatai-23506" => array("type"=>"village", "pop"=>805, "name_ru"=>"Ваилоатаи", "name_en"=>"Vailoatai", "lang"=>"en", "lat"=>-14.385127, "lon"=>-170.782982),
"/US/US-AS/vaitogi-23507" => array("type"=>"village", "pop"=>1302, "name_ru"=>"Ваитоги", "name_en"=>"Vaitogi", "lang"=>"en", "lat"=>-14.390496, "lon"=>-170.731285),
"/US/US-AS/vatia-23508" => array("type"=>"village", "pop"=>608, "name_ru"=>"Ватиа", "name_en"=>"Vatia", "lang"=>"en", "lat"=>-14.201782, "lon"=>-170.652139),
"/US/US-AZ" => array("type"=>"state", "pop"=>6500180, "name_ru"=>"Аризона", "name_en"=>"Arizona", "lang"=>"en"),
"/US/US-AZ/ajo-791" => array("type"=>"cdp", "pop"=>2919, "name_ru"=>"Ахо", "name_en"=>"Ajo", "lang"=>"en", "lat"=>32.387246, "lon"=>-112.868364),
"/US/US-AZ/ak-chin-village-792" => array("type"=>"cdp", "pop"=>353, "name_ru"=>"Ак-Чин Виллидж", "name_en"=>"Ak-Chin Village", "lang"=>"en", "lat"=>33.028295, "lon"=>-112.086986),
"/US/US-AZ/apache-junction-793" => array("type"=>"city", "pop"=>32776, "name_ru"=>"Апач Джанкшен", "name_en"=>"Apache Junction", "lang"=>"en", "lat"=>33.405000, "lon"=>-111.546999),
"/US/US-AZ/arizona-city-794" => array("type"=>"cdp", "pop"=>1940, "name_ru"=>"Аризона Сити", "name_en"=>"Arizona City", "lang"=>"en", "lat"=>32.750803, "lon"=>-111.669818),
"/US/US-AZ/avondale-795" => array("type"=>"city", "pop"=>81299, "name_ru"=>"Авондейл", "name_en"=>"Avondale", "lang"=>"en", "lat"=>33.420072, "lon"=>-112.324044),
"/US/US-AZ/avra-valley-796" => array("type"=>"cdp", "pop"=>3403, "name_ru"=>"Авра Валли", "name_en"=>"Avra Valley", "lang"=>"en", "lat"=>32.437856, "lon"=>-111.314712),
"/US/US-AZ/bagdad-797" => array("type"=>"cdp", "pop"=>1858, "name_ru"=>"Багдад", "name_en"=>"Bagdad", "lang"=>"en", "lat"=>34.577922, "lon"=>-113.173730),
"/US/US-AZ/benson-798" => array("type"=>"city", "pop"=>4833, "name_ru"=>"Бенсон", "name_en"=>"Benson", "lang"=>"en", "lat"=>31.962193, "lon"=>-110.303107),
"/US/US-AZ/big-park-799" => array("type"=>"cdp", "pop"=>3024, "name_ru"=>"Биг Парк", "name_en"=>"Big Park", "lang"=>"en", "lat"=>34.780180, "lon"=>-111.761827),
"/US/US-AZ/bisbee-800" => array("type"=>"city", "pop"=>5991, "name_ru"=>"Бисби", "name_en"=>"Bisbee", "lang"=>"en", "lat"=>31.404563, "lon"=>-109.917479),
"/US/US-AZ/black-canyon-city-801" => array("type"=>"cdp", "pop"=>1811, "name_ru"=>"Блэк Каньон Сити", "name_en"=>"Black Canyon City", "lang"=>"en", "lat"=>34.062484, "lon"=>-112.109368),
"/US/US-AZ/blackwater-802" => array("type"=>"cdp", "pop"=>400, "name_ru"=>"Блэкуотер", "name_en"=>"Blackwater", "lang"=>"en", "lat"=>33.030663, "lon"=>-111.615270),
"/US/US-AZ/bluewater-803" => array("type"=>"cdp", "pop"=>505, "name_ru"=>"Блууотер", "name_en"=>"Bluewater", "lang"=>"en", "lat"=>34.168535, "lon"=>-114.264120),
"/US/US-AZ/buckeye-804" => array("type"=>"town", "pop"=>47261, "name_ru"=>"Бакай", "name_en"=>"Buckeye", "lang"=>"en", "lat"=>33.412639, "lon"=>-112.585529),
"/US/US-AZ/bullhead-city-805" => array("type"=>"city", "pop"=>40868, "name_ru"=>"Буллхед Сити", "name_en"=>"Bullhead City", "lang"=>"en", "lat"=>35.114187, "lon"=>-114.558590),
"/US/US-AZ/bylas-806" => array("type"=>"cdp", "pop"=>1219, "name_ru"=>"Байлас", "name_en"=>"Bylas", "lang"=>"en", "lat"=>33.130591, "lon"=>-110.110094),
"/US/US-AZ/cameron-807" => array("type"=>"cdp", "pop"=>493, "name_ru"=>"Кэмерон", "name_en"=>"Cameron", "lang"=>"en", "lat"=>35.848250, "lon"=>-111.431858),
"/US/US-AZ/camp-verde-808" => array("type"=>"town", "pop"=>10849, "name_ru"=>"Кэмп Верде", "name_en"=>"Camp Verde", "lang"=>"en", "lat"=>34.569655, "lon"=>-111.856192),
"/US/US-AZ/canyon-day-809" => array("type"=>"cdp", "pop"=>857, "name_ru"=>"Каньон Дэй", "name_en"=>"Canyon Day", "lang"=>"en", "lat"=>33.781420, "lon"=>-110.026176),
"/US/US-AZ/carefree-810" => array("type"=>"town", "pop"=>3862, "name_ru"=>"Кэрфри", "name_en"=>"Carefree", "lang"=>"en", "lat"=>33.822423, "lon"=>-111.918440),
"/US/US-AZ/casa-grande-811" => array("type"=>"city", "pop"=>41152, "name_ru"=>"Каса Грандэ", "name_en"=>"Casa Grande", "lang"=>"en", "lat"=>32.892313, "lon"=>-111.735927),
"/US/US-AZ/catalina-812" => array("type"=>"cdp", "pop"=>4864, "name_ru"=>"Каталина", "name_en"=>"Catalina", "lang"=>"en", "lat"=>32.476850, "lon"=>-110.900211),
"/US/US-AZ/cave-creek-813" => array("type"=>"town", "pop"=>5428, "name_ru"=>"Ков Крик", "name_en"=>"Cave Creek", "lang"=>"en", "lat"=>33.826165, "lon"=>-111.965206),
"/US/US-AZ/central-heights-midland-city-814" => array("type"=>"cdp", "pop"=>2969, "name_ru"=>"Сентрал Хайтс-Мидленд Сити", "name_en"=>"Central Heights-Midland City", "lang"=>"en", "lat"=>33.405183, "lon"=>-110.812166),
"/US/US-AZ/chandler-815" => array("type"=>"city", "pop"=>247140, "name_ru"=>"Чандлер", "name_en"=>"Chandler", "lang"=>"en", "lat"=>33.295000, "lon"=>-111.866856),
"/US/US-AZ/chinle-816" => array("type"=>"cdp", "pop"=>5059, "name_ru"=>"Чинл", "name_en"=>"Chinle", "lang"=>"en", "lat"=>36.151912, "lon"=>-109.580026),
"/US/US-AZ/chino-valley-817" => array("type"=>"town", "pop"=>11078, "name_ru"=>"Чино Валли", "name_en"=>"Chino Valley", "lang"=>"en", "lat"=>34.763216, "lon"=>-112.454696),
"/US/US-AZ/chuichu-818" => array("type"=>"cdp", "pop"=>330, "name_ru"=>"Чуичу", "name_en"=>"Chuichu", "lang"=>"en", "lat"=>32.744768, "lon"=>-111.797223),
"/US/US-AZ/cibecue-819" => array("type"=>"cdp", "pop"=>1254, "name_ru"=>"Сибекью", "name_en"=>"Cibecue", "lang"=>"en", "lat"=>34.029486, "lon"=>-110.482165),
"/US/US-AZ/clarkdale-820" => array("type"=>"town", "pop"=>4263, "name_ru"=>"Кларкдейл", "name_en"=>"Clarkdale", "lang"=>"en", "lat"=>34.759667, "lon"=>-112.056476),
"/US/US-AZ/claypool-821" => array("type"=>"cdp", "pop"=>1942, "name_ru"=>"Клэйпул", "name_en"=>"Claypool", "lang"=>"en", "lat"=>33.406687, "lon"=>-110.840115),
"/US/US-AZ/clifton-822" => array("type"=>"town", "pop"=>2451, "name_ru"=>"Слифтон", "name_en"=>"Clifton", "lang"=>"en", "lat"=>33.023511, "lon"=>-109.290494),
"/US/US-AZ/colorado-city-823" => array("type"=>"town", "pop"=>4750, "name_ru"=>"Колорадо Сити", "name_en"=>"Colorado City", "lang"=>"en", "lat"=>36.980841, "lon"=>-112.982502),
"/US/US-AZ/coolidge-824" => array("type"=>"city", "pop"=>10261, "name_ru"=>"Кулидж", "name_en"=>"Coolidge", "lang"=>"en", "lat"=>32.982439, "lon"=>-111.527600),
"/US/US-AZ/cornville-825" => array("type"=>"cdp", "pop"=>2089, "name_ru"=>"Корнвилл", "name_en"=>"Cornville", "lang"=>"en", "lat"=>34.721847, "lon"=>-111.907833),
"/US/US-AZ/cottonwood-826" => array("type"=>"city", "pop"=>11412, "name_ru"=>"Коттонвуд", "name_en"=>"Cottonwood", "lang"=>"en", "lat"=>34.734520, "lon"=>-112.024405),
"/US/US-AZ/cottonwood-verde-village-827" => array("type"=>"cdp", "pop"=>7037, "name_ru"=>"Коттонвуд-Верде Виллидж", "name_en"=>"Cottonwood-Verde Village", "lang"=>"en", "lat"=>34.711146, "lon"=>-111.994376),
"/US/US-AZ/dennehotso-828" => array("type"=>"cdp", "pop"=>616, "name_ru"=>"Дэннехотсо", "name_en"=>"Dennehotso", "lang"=>"en", "lat"=>36.821527, "lon"=>-109.883784),
"/US/US-AZ/desert-hills-829" => array("type"=>"cdp", "pop"=>1700, "name_ru"=>"Дезерт Хилс", "name_en"=>"Desert Hills", "lang"=>"en", "lat"=>34.553970, "lon"=>-114.371691),
"/US/US-AZ/dewey-humboldt-830" => array("type"=>"cdp", "pop"=>3822, "name_ru"=>"Дьюи-Гумбольдт", "name_en"=>"Dewey-Humboldt", "lang"=>"en", "lat"=>34.532554, "lon"=>-112.251800),
"/US/US-AZ/dolan-springs-831" => array("type"=>"cdp", "pop"=>1090, "name_ru"=>"Долан Спрингс", "name_en"=>"Dolan Springs", "lang"=>"en", "lat"=>35.604482, "lon"=>-114.266523),
"/US/US-AZ/douglas-832" => array("type"=>"city", "pop"=>17515, "name_ru"=>"Дуглас", "name_en"=>"Douglas", "lang"=>"en", "lat"=>31.342687, "lon"=>-109.528194),
"/US/US-AZ/dudleyville-833" => array("type"=>"cdp", "pop"=>1356, "name_ru"=>"Дадливилл", "name_en"=>"Dudleyville", "lang"=>"en", "lat"=>32.914267, "lon"=>-110.733779),
"/US/US-AZ/duncan-834" => array("type"=>"town", "pop"=>774, "name_ru"=>"Дункан", "name_en"=>"Duncan", "lang"=>"en", "lat"=>32.736254, "lon"=>-109.091780),
"/US/US-AZ/eagar-835" => array("type"=>"town", "pop"=>4488, "name_ru"=>"Игар", "name_en"=>"Eagar", "lang"=>"en", "lat"=>34.104928, "lon"=>-109.292112),
"/US/US-AZ/east-fork-836" => array("type"=>"cdp", "pop"=>752, "name_ru"=>"Ист Форк", "name_en"=>"East Fork", "lang"=>"en", "lat"=>33.806447, "lon"=>-109.931443),
"/US/US-AZ/ehrenberg-837" => array("type"=>"cdp", "pop"=>1226, "name_ru"=>"Эренберг", "name_en"=>"Ehrenberg", "lang"=>"en", "lat"=>33.622929, "lon"=>-114.493922),
"/US/US-AZ/el-mirage-838" => array("type"=>"town", "pop"=>24751, "name_ru"=>"Эль Мираж", "name_en"=>"El Mirage", "lang"=>"en", "lat"=>33.591210, "lon"=>-112.324018),
"/US/US-AZ/eloy-839" => array("type"=>"city", "pop"=>12750, "name_ru"=>"Элой", "name_en"=>"Eloy", "lang"=>"en", "lat"=>32.745043, "lon"=>-111.606508),
"/US/US-AZ/flagstaff-840" => array("type"=>"city", "pop"=>60222, "name_ru"=>"Флагстафф", "name_en"=>"Flagstaff", "lang"=>"en", "lat"=>35.189005, "lon"=>-111.619768),
"/US/US-AZ/florence-841" => array("type"=>"town", "pop"=>20781, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>33.037566, "lon"=>-111.373908),
"/US/US-AZ/flowing-wells-842" => array("type"=>"cdp", "pop"=>14013, "name_ru"=>"Флауинг Уэлс", "name_en"=>"Flowing Wells", "lang"=>"en", "lat"=>32.293812, "lon"=>-111.009162),
"/US/US-AZ/fort-defiance-843" => array("type"=>"cdp", "pop"=>4489, "name_ru"=>"Форт Дефианс", "name_en"=>"Fort Defiance", "lang"=>"en", "lat"=>35.743675, "lon"=>-109.066589),
"/US/US-AZ/fortuna-foothills-844" => array("type"=>"cdp", "pop"=>7737, "name_ru"=>"Фортуна Футхилс", "name_en"=>"Fortuna Foothills", "lang"=>"en", "lat"=>32.657665, "lon"=>-114.411030),
"/US/US-AZ/fountain-hills-845" => array("type"=>"town", "pop"=>25227, "name_ru"=>"Фонтейн Хилс", "name_en"=>"Fountain Hills", "lang"=>"en", "lat"=>33.605350, "lon"=>-111.741113),
"/US/US-AZ/fredonia-846" => array("type"=>"town", "pop"=>1125, "name_ru"=>"Фредониа", "name_en"=>"Fredonia", "lang"=>"en", "lat"=>36.964762, "lon"=>-112.518848),
"/US/US-AZ/ganado-847" => array("type"=>"cdp", "pop"=>1257, "name_ru"=>"Ганадо", "name_en"=>"Ganado", "lang"=>"en", "lat"=>35.707551, "lon"=>-109.552030),
"/US/US-AZ/gila-bend-848" => array("type"=>"town", "pop"=>1831, "name_ru"=>"Гила Бенд", "name_en"=>"Gila Bend", "lang"=>"en", "lat"=>32.951532, "lon"=>-112.684799),
"/US/US-AZ/gilbert-849" => array("type"=>"town", "pop"=>216449, "name_ru"=>"Гилберт", "name_en"=>"Gilbert", "lang"=>"en", "lat"=>33.332032, "lon"=>-111.764189),
"/US/US-AZ/glendale-850" => array("type"=>"city", "pop"=>251522, "name_ru"=>"Глендейл", "name_en"=>"Glendale", "lang"=>"en", "lat"=>33.584128, "lon"=>-112.201590),
"/US/US-AZ/globe-851" => array("type"=>"city", "pop"=>7197, "name_ru"=>"Глоуб", "name_en"=>"Globe", "lang"=>"en", "lat"=>33.382813, "lon"=>-110.754305),
"/US/US-AZ/golden-valley-852" => array("type"=>"cdp", "pop"=>2619, "name_ru"=>"Голден Валли", "name_en"=>"Golden Valley", "lang"=>"en", "lat"=>35.223239, "lon"=>-114.222090),
"/US/US-AZ/goodyear-853" => array("type"=>"city", "pop"=>59508, "name_ru"=>"Гудиир", "name_en"=>"Goodyear", "lang"=>"en", "lat"=>33.346954, "lon"=>-112.411365),
"/US/US-AZ/grand-canyon-village-854" => array("type"=>"cdp", "pop"=>1499, "name_ru"=>"Гранд Каньон Виллидж", "name_en"=>"Grand Canyon Village", "lang"=>"en", "lat"=>36.046441, "lon"=>-112.153457),
"/US/US-AZ/greasewood-855" => array("type"=>"cdp", "pop"=>196, "name_ru"=>"Грезвуд", "name_en"=>"Greasewood", "lang"=>"en", "lat"=>35.528024, "lon"=>-109.860090),
"/US/US-AZ/green-valley-856" => array("type"=>"cdp", "pop"=>13231, "name_ru"=>"Грин Валли", "name_en"=>"Green Valley", "lang"=>"en", "lat"=>31.848540, "lon"=>-111.009769),
"/US/US-AZ/guadalupe-857" => array("type"=>"town", "pop"=>5948, "name_ru"=>"Гвадалупе", "name_en"=>"Guadalupe", "lang"=>"en", "lat"=>33.365767, "lon"=>-111.961669),
"/US/US-AZ/hayden-858" => array("type"=>"town", "pop"=>814, "name_ru"=>"Хейден", "name_en"=>"Hayden", "lang"=>"en", "lat"=>32.999884, "lon"=>-110.780957),
"/US/US-AZ/heber-overgaard-859" => array("type"=>"cdp", "pop"=>1581, "name_ru"=>"Хебер-Овергаард", "name_en"=>"Heber-Overgaard", "lang"=>"en", "lat"=>34.414006, "lon"=>-110.568836),
"/US/US-AZ/holbrook-860" => array("type"=>"city", "pop"=>5081, "name_ru"=>"Холбрук", "name_en"=>"Holbrook", "lang"=>"en", "lat"=>34.912841, "lon"=>-110.155225),
"/US/US-AZ/hotevilla-861" => array("type"=>"cdp", "pop"=>869, "name_ru"=>"Хотевилла", "name_en"=>"Hotevilla", "lang"=>"en", "lat"=>35.923787, "lon"=>-110.657593),
"/US/US-AZ/huachuca-city-862" => array("type"=>"town", "pop"=>1963, "name_ru"=>"Хуачука Сити", "name_en"=>"Huachuca City", "lang"=>"en", "lat"=>31.632069, "lon"=>-110.340846),
"/US/US-AZ/jerome-863" => array("type"=>"town", "pop"=>353, "name_ru"=>"Джером", "name_en"=>"Jerome", "lang"=>"en", "lat"=>34.746705, "lon"=>-112.106123),
"/US/US-AZ/kachina-village-864" => array("type"=>"cdp", "pop"=>1711, "name_ru"=>"Качайна Виллидж", "name_en"=>"Kachina Village", "lang"=>"en", "lat"=>35.097003, "lon"=>-111.691933),
"/US/US-AZ/kaibito-865" => array("type"=>"cdp", "pop"=>641, "name_ru"=>"Кайбито", "name_en"=>"Kaibito", "lang"=>"en", "lat"=>36.589036, "lon"=>-111.110317),
"/US/US-AZ/kayenta-866" => array("type"=>"cdp", "pop"=>4372, "name_ru"=>"Кайента", "name_en"=>"Kayenta", "lang"=>"en", "lat"=>36.717960, "lon"=>-110.252566),
"/US/US-AZ/keams-canyon-867" => array("type"=>"cdp", "pop"=>393, "name_ru"=>"Кимс Каньон", "name_en"=>"Keams Canyon", "lang"=>"en", "lat"=>35.815911, "lon"=>-110.208894),
"/US/US-AZ/kearny-868" => array("type"=>"town", "pop"=>3297, "name_ru"=>"Кирни", "name_en"=>"Kearny", "lang"=>"en", "lat"=>33.056211, "lon"=>-110.905981),
"/US/US-AZ/kingman-869" => array("type"=>"city", "pop"=>27817, "name_ru"=>"Кингман", "name_en"=>"Kingman", "lang"=>"en", "lat"=>35.207020, "lon"=>-114.031970),
"/US/US-AZ/komatke-870" => array("type"=>"cdp", "pop"=>1116, "name_ru"=>"Коматк", "name_en"=>"Komatke", "lang"=>"en", "lat"=>33.306204, "lon"=>-112.178107),
"/US/US-AZ/kykotsmovi-village-871" => array("type"=>"cdp", "pop"=>773, "name_ru"=>"Кайкотсмови Виллидж", "name_en"=>"Kykotsmovi Village", "lang"=>"en", "lat"=>35.874240, "lon"=>-110.622749),
"/US/US-AZ/lake-havasu-city-872" => array("type"=>"city", "pop"=>56553, "name_ru"=>"Лейк Хавасу Виллидж", "name_en"=>"Lake Havasu City", "lang"=>"en", "lat"=>34.502866, "lon"=>-114.314272),
"/US/US-AZ/lake-montezuma-873" => array("type"=>"cdp", "pop"=>1841, "name_ru"=>"Лейк Монтесума", "name_en"=>"Lake Montezuma", "lang"=>"en", "lat"=>34.639323, "lon"=>-111.795516),
"/US/US-AZ/leupp-874" => array("type"=>"cdp", "pop"=>857, "name_ru"=>"Леапп", "name_en"=>"Leupp", "lang"=>"en", "lat"=>35.295387, "lon"=>-111.000574),
"/US/US-AZ/litchfield-park-875" => array("type"=>"city", "pop"=>5126, "name_ru"=>"Литчфилд Парк", "name_en"=>"Litchfield Park", "lang"=>"en", "lat"=>33.502526, "lon"=>-112.358817),
"/US/US-AZ/lukachukai-876" => array("type"=>"cdp", "pop"=>113, "name_ru"=>"Люкачукай", "name_en"=>"Lukachukai", "lang"=>"en", "lat"=>36.400071, "lon"=>-109.256877),
"/US/US-AZ/luke-afb-877" => array("type"=>"cdp", "pop"=>4371, "name_ru"=>"Военно-воздушная база Люк", "name_en"=>"Luke AFB", "lang"=>"en", "lat"=>33.533129, "lon"=>-112.370673),
"/US/US-AZ/mammoth-879" => array("type"=>"town", "pop"=>2573, "name_ru"=>"Маммот", "name_en"=>"Mammoth", "lang"=>"en", "lat"=>32.721506, "lon"=>-110.643497),
"/US/US-AZ/many-farms-880" => array("type"=>"cdp", "pop"=>1294, "name_ru"=>"Мани Фармс", "name_en"=>"Many Farms", "lang"=>"en", "lat"=>36.350271, "lon"=>-109.618463),
"/US/US-AZ/marana-881" => array("type"=>"town", "pop"=>34466, "name_ru"=>"Марана", "name_en"=>"Marana", "lang"=>"en", "lat"=>32.414432, "lon"=>-111.172754),
"/US/US-AZ/mcnary-878" => array("type"=>"cdp", "pop"=>355, "name_ru"=>"Макнари", "name_en"=>"McNary", "lang"=>"en", "lat"=>34.095785, "lon"=>-109.850548),
"/US/US-AZ/mesa-882" => array("type"=>"city", "pop"=>463552, "name_ru"=>"Меса", "name_en"=>"Mesa", "lang"=>"en", "lat"=>33.417736, "lon"=>-111.740337),
"/US/US-AZ/miami-883" => array("type"=>"town", "pop"=>1778, "name_ru"=>"Майами", "name_en"=>"Miami", "lang"=>"en", "lat"=>33.394327, "lon"=>-110.872055),
"/US/US-AZ/moenkopi-884" => array("type"=>"cdp", "pop"=>924, "name_ru"=>"Моенкопи", "name_en"=>"Moenkopi", "lang"=>"en", "lat"=>36.110912, "lon"=>-111.222177),
"/US/US-AZ/mohave-valley-885" => array("type"=>"cdp", "pop"=>6962, "name_ru"=>"Мохаве Валли", "name_en"=>"Mohave Valley", "lang"=>"en", "lat"=>34.933188, "lon"=>-114.588049),
"/US/US-AZ/morenci-886" => array("type"=>"cdp", "pop"=>1799, "name_ru"=>"Моренси", "name_en"=>"Morenci", "lang"=>"en", "lat"=>33.053047, "lon"=>-109.329545),
"/US/US-AZ/new-kingman-butler-887" => array("type"=>"cdp", "pop"=>11627, "name_ru"=>"Нью Кингман-Батлер", "name_en"=>"New Kingman-Butler", "lang"=>"en", "lat"=>35.260142, "lon"=>-114.029765),
"/US/US-AZ/nogales-888" => array("type"=>"city", "pop"=>19573, "name_ru"=>"Ногалес", "name_en"=>"Nogales", "lang"=>"en", "lat"=>31.363719, "lon"=>-110.932639),
"/US/US-AZ/oracle-889" => array("type"=>"cdp", "pop"=>3043, "name_ru"=>"Оракл", "name_en"=>"Oracle", "lang"=>"en", "lat"=>32.618765, "lon"=>-110.784194),
"/US/US-AZ/oro-valley-890" => array("type"=>"town", "pop"=>43465, "name_ru"=>"Оро Валли", "name_en"=>"Oro Valley", "lang"=>"en", "lat"=>32.424972, "lon"=>-110.964657),
"/US/US-AZ/page-891" => array("type"=>"city", "pop"=>6928, "name_ru"=>"Пейдж", "name_en"=>"Page", "lang"=>"en", "lat"=>36.904253, "lon"=>-111.457823),
"/US/US-AZ/paradise-valley-892" => array("type"=>"town", "pop"=>14990, "name_ru"=>"Парадайс Валли", "name_en"=>"Paradise Valley", "lang"=>"en", "lat"=>33.539267, "lon"=>-111.954622),
"/US/US-AZ/parker-893" => array("type"=>"town", "pop"=>3174, "name_ru"=>"Паркер", "name_en"=>"Parker", "lang"=>"en", "lat"=>34.021823, "lon"=>-114.231043),
"/US/US-AZ/parker-strip-894" => array("type"=>"cdp", "pop"=>1646, "name_ru"=>"Паркер Стрип", "name_en"=>"Parker Strip", "lang"=>"en", "lat"=>34.257749, "lon"=>-114.142052),
"/US/US-AZ/patagonia-895" => array("type"=>"town", "pop"=>780, "name_ru"=>"Патагония", "name_en"=>"Patagonia", "lang"=>"en", "lat"=>31.543280, "lon"=>-110.748950),
"/US/US-AZ/payson-896" => array("type"=>"town", "pop"=>15486, "name_ru"=>"Пейсон", "name_en"=>"Payson", "lang"=>"en", "lat"=>34.242604, "lon"=>-111.327002),
"/US/US-AZ/peach-springs-897" => array("type"=>"cdp", "pop"=>787, "name_ru"=>"Пич Стрингс", "name_en"=>"Peach Springs", "lang"=>"en", "lat"=>35.527421, "lon"=>-113.429819),
"/US/US-AZ/peoria-898" => array("type"=>"city", "pop"=>157960, "name_ru"=>"Пеория", "name_en"=>"Peoria", "lang"=>"en", "lat"=>33.688481, "lon"=>-112.244464),
"/US/US-AZ/peridot-899" => array("type"=>"cdp", "pop"=>957, "name_ru"=>"Перидот", "name_en"=>"Peridot", "lang"=>"en", "lat"=>33.302665, "lon"=>-110.454308),
"/US/US-AZ/phoenix-900" => array("type"=>"city", "pop"=>1567924, "name_ru"=>"Феникс", "name_en"=>"Phoenix", "lang"=>"en", "lat"=>33.542550, "lon"=>-112.071399),
"/US/US-AZ/picture-rocks-901" => array("type"=>"cdp", "pop"=>4026, "name_ru"=>"Пикчер Рокс", "name_en"=>"Picture Rocks", "lang"=>"en", "lat"=>32.345764, "lon"=>-111.245677),
"/US/US-AZ/pima-902" => array("type"=>"town", "pop"=>2165, "name_ru"=>"Пима", "name_en"=>"Pima", "lang"=>"en", "lat"=>32.888143, "lon"=>-109.833319),
"/US/US-AZ/pinetop-lakeside-903" => array("type"=>"town", "pop"=>4536, "name_ru"=>"Пайнтоп-Лейксайд", "name_en"=>"Pinetop-Lakeside", "lang"=>"en", "lat"=>34.148123, "lon"=>-109.966371),
"/US/US-AZ/pinon-904" => array("type"=>"cdp", "pop"=>468, "name_ru"=>"Пинон", "name_en"=>"Pinon", "lang"=>"en", "lat"=>36.100708, "lon"=>-110.221264),
"/US/US-AZ/pirtleville-905" => array("type"=>"cdp", "pop"=>1364, "name_ru"=>"Пиртлевилл", "name_en"=>"Pirtleville", "lang"=>"en", "lat"=>31.361653, "lon"=>-109.566932),
"/US/US-AZ/pisinemo-906" => array("type"=>"cdp", "pop"=>341, "name_ru"=>"Писайномо", "name_en"=>"Pisinemo", "lang"=>"en", "lat"=>32.037785, "lon"=>-112.320326),
"/US/US-AZ/polacca-907" => array("type"=>"cdp", "pop"=>1108, "name_ru"=>"Полакка", "name_en"=>"Polacca", "lang"=>"en", "lat"=>35.835638, "lon"=>-110.365594),
"/US/US-AZ/poston-908" => array("type"=>"cdp", "pop"=>480, "name_ru"=>"Постон", "name_en"=>"Poston", "lang"=>"en", "lat"=>33.959122, "lon"=>-114.415234),
"/US/US-AZ/prescott-909" => array("type"=>"city", "pop"=>42697, "name_ru"=>"Прескотт", "name_en"=>"Prescott", "lang"=>"en", "lat"=>34.576007, "lon"=>-112.449076),
"/US/US-AZ/prescott-valley-910" => array("type"=>"town", "pop"=>38535, "name_ru"=>"Прескотт Валли", "name_en"=>"Prescott Valley", "lang"=>"en", "lat"=>34.608234, "lon"=>-112.321417),
"/US/US-AZ/quartzsite-911" => array("type"=>"town", "pop"=>3481, "name_ru"=>"Квартцзит", "name_en"=>"Quartzsite", "lang"=>"en", "lat"=>33.666992, "lon"=>-114.214993),
"/US/US-AZ/queen-creek-912" => array("type"=>"town", "pop"=>24361, "name_ru"=>"Квин Крик", "name_en"=>"Queen Creek", "lang"=>"en", "lat"=>33.241430, "lon"=>-111.643236),
"/US/US-AZ/rio-rico-east-913" => array("type"=>"cdp", "pop"=>1407, "name_ru"=>"Рио Рико Ист", "name_en"=>"Rio Rico East", "lang"=>"en", "lat"=>31.479779, "lon"=>-110.975573),
"/US/US-AZ/rough-rock-914" => array("type"=>"cdp", "pop"=>523, "name_ru"=>"Раф Рок", "name_en"=>"Rough Rock", "lang"=>"en", "lat"=>36.409547, "lon"=>-109.867529),
"/US/US-AZ/sacaton-915" => array("type"=>"cdp", "pop"=>1452, "name_ru"=>"Сакатон", "name_en"=>"Sacaton", "lang"=>"en", "lat"=>33.082147, "lon"=>-111.746081),
"/US/US-AZ/safford-916" => array("type"=>"city", "pop"=>9823, "name_ru"=>"Саффорд", "name_en"=>"Safford", "lang"=>"en", "lat"=>32.831157, "lon"=>-109.698147),
"/US/US-AZ/san-carlos-920" => array("type"=>"cdp", "pop"=>2918, "name_ru"=>"Сан Карлос", "name_en"=>"San Carlos", "lang"=>"en", "lat"=>33.348635, "lon"=>-110.465051),
"/US/US-AZ/san-luis-921" => array("type"=>"city", "pop"=>24909, "name_ru"=>"Сан Луис", "name_en"=>"San Luis", "lang"=>"en", "lat"=>32.492453, "lon"=>-114.777861),
"/US/US-AZ/san-manuel-922" => array("type"=>"cdp", "pop"=>4009, "name_ru"=>"Сан Мануэль", "name_en"=>"San Manuel", "lang"=>"en", "lat"=>32.613999, "lon"=>-110.636202),
"/US/US-AZ/santa-rosa-924" => array("type"=>"cdp", "pop"=>493, "name_ru"=>"Санта Роза", "name_en"=>"Santa Rosa", "lang"=>"en", "lat"=>32.334847, "lon"=>-112.045401),
"/US/US-AZ/santan-923" => array("type"=>"cdp", "pop"=>330, "name_ru"=>"Сантан", "name_en"=>"Santan", "lang"=>"en", "lat"=>33.145476, "lon"=>-111.801546),
"/US/US-AZ/sawmill-925" => array("type"=>"cdp", "pop"=>507, "name_ru"=>"Сомилл", "name_en"=>"Sawmill", "lang"=>"en", "lat"=>35.891686, "lon"=>-109.153520),
"/US/US-AZ/scottsdale-926" => array("type"=>"city", "pop"=>235371, "name_ru"=>"Скотсдейл", "name_en"=>"Scottsdale", "lang"=>"en", "lat"=>33.685992, "lon"=>-111.870478),
"/US/US-AZ/second-mesa-927" => array("type"=>"cdp", "pop"=>929, "name_ru"=>"Секонд Меса", "name_en"=>"Second Mesa", "lang"=>"en", "lat"=>35.817608, "lon"=>-110.503592),
"/US/US-AZ/sedona-928" => array("type"=>"city", "pop"=>11599, "name_ru"=>"Седона", "name_en"=>"Sedona", "lang"=>"en", "lat"=>34.859275, "lon"=>-111.795122),
"/US/US-AZ/sells-929" => array("type"=>"cdp", "pop"=>2750, "name_ru"=>"Селс", "name_en"=>"Sells", "lang"=>"en", "lat"=>31.918458, "lon"=>-111.876443),
"/US/US-AZ/shongopovi-930" => array("type"=>"cdp", "pop"=>730, "name_ru"=>"Шонгопови", "name_en"=>"Shongopovi", "lang"=>"en", "lat"=>35.818099, "lon"=>-110.533999),
"/US/US-AZ/shonto-931" => array("type"=>"cdp", "pop"=>710, "name_ru"=>"Шонто", "name_en"=>"Shonto", "lang"=>"en", "lat"=>36.588018, "lon"=>-110.654182),
"/US/US-AZ/show-low-932" => array("type"=>"city", "pop"=>12324, "name_ru"=>"Шоу Лоу", "name_en"=>"Show Low", "lang"=>"en", "lat"=>34.237331, "lon"=>-110.044298),
"/US/US-AZ/sierra-vista-933" => array("type"=>"city", "pop"=>43320, "name_ru"=>"Сьерра Виста", "name_en"=>"Sierra Vista", "lang"=>"en", "lat"=>31.565113, "lon"=>-110.322461),
"/US/US-AZ/sierra-vista-southeast-934" => array("type"=>"cdp", "pop"=>9237, "name_ru"=>"Сьерра Виста Сауфист", "name_en"=>"Sierra Vista Southeast", "lang"=>"en", "lat"=>31.460592, "lon"=>-110.217428),
"/US/US-AZ/snowflake-935" => array("type"=>"town", "pop"=>5673, "name_ru"=>"Сноуфлейк", "name_en"=>"Snowflake", "lang"=>"en", "lat"=>34.521977, "lon"=>-110.084197),
"/US/US-AZ/somerton-936" => array("type"=>"city", "pop"=>12346, "name_ru"=>"Сомертон", "name_en"=>"Somerton", "lang"=>"en", "lat"=>32.597083, "lon"=>-114.709457),
"/US/US-AZ/south-tucson-937" => array("type"=>"city", "pop"=>6013, "name_ru"=>"Сауф Тусон", "name_en"=>"South Tucson", "lang"=>"en", "lat"=>32.196000, "lon"=>-110.968249),
"/US/US-AZ/springerville-938" => array("type"=>"town", "pop"=>1992, "name_ru"=>"Спрингервилл", "name_en"=>"Springerville", "lang"=>"en", "lat"=>34.156075, "lon"=>-109.296503),
"/US/US-AZ/st.-david-917" => array("type"=>"cdp", "pop"=>1468, "name_ru"=>"Сент-Дэвид", "name_en"=>"St. David", "lang"=>"en", "lat"=>31.900091, "lon"=>-110.229392),
"/US/US-AZ/st.-johns-918" => array("type"=>"city", "pop"=>3607, "name_ru"=>"Сент-Джонс", "name_en"=>"St. Johns", "lang"=>"en", "lat"=>34.509929, "lon"=>-109.376758),
"/US/US-AZ/st.-michaels-919" => array("type"=>"cdp", "pop"=>1119, "name_ru"=>"Сент-Майклс", "name_en"=>"St. Michaels", "lang"=>"en", "lat"=>35.661688, "lon"=>-109.093620),
"/US/US-AZ/sun-city-939" => array("type"=>"cdp", "pop"=>38126, "name_ru"=>"Сан Сити", "name_en"=>"Sun City", "lang"=>"en", "lat"=>33.616267, "lon"=>-112.281815),
"/US/US-AZ/sun-city-west-940" => array("type"=>"cdp", "pop"=>15997, "name_ru"=>"Сан Сити Уэст", "name_en"=>"Sun City West", "lang"=>"en", "lat"=>33.665416, "lon"=>-112.354123),
"/US/US-AZ/sun-lakes-941" => array("type"=>"cdp", "pop"=>6578, "name_ru"=>"Сан Лейкс", "name_en"=>"Sun Lakes", "lang"=>"en", "lat"=>33.216859, "lon"=>-111.870092),
"/US/US-AZ/supai-942" => array("type"=>"cdp", "pop"=>423, "name_ru"=>"Сапай", "name_en"=>"Supai", "lang"=>"en", "lat"=>36.224157, "lon"=>-112.693212),
"/US/US-AZ/superior-943" => array("type"=>"town", "pop"=>3335, "name_ru"=>"Супериор", "name_en"=>"Superior", "lang"=>"en", "lat"=>33.285854, "lon"=>-111.109484),
"/US/US-AZ/surprise-944" => array("type"=>"town", "pop"=>92897, "name_ru"=>"Серпрайз", "name_en"=>"Surprise", "lang"=>"en", "lat"=>33.664112, "lon"=>-112.468451),
"/US/US-AZ/swift-trail-junction-945" => array("type"=>"cdp", "pop"=>1203, "name_ru"=>"Свифт Трейл Джанкшен", "name_en"=>"Swift Trail Junction", "lang"=>"en", "lat"=>32.732465, "lon"=>-109.714932),
"/US/US-AZ/taylor-946" => array("type"=>"town", "pop"=>4139, "name_ru"=>"Тейлор", "name_en"=>"Taylor", "lang"=>"en", "lat"=>34.449926, "lon"=>-110.111550),
"/US/US-AZ/teec-nos-pos-947" => array("type"=>"cdp", "pop"=>317, "name_ru"=>"Теес Нос Пос", "name_en"=>"Teec Nos Pos", "lang"=>"en", "lat"=>36.929104, "lon"=>-109.082099),
"/US/US-AZ/tempe-948" => array("type"=>"city", "pop"=>175523, "name_ru"=>"Темп", "name_en"=>"Tempe", "lang"=>"en", "lat"=>33.388350, "lon"=>-111.930639),
"/US/US-AZ/thatcher-949" => array("type"=>"town", "pop"=>5002, "name_ru"=>"Тэтчер", "name_en"=>"Thatcher", "lang"=>"en", "lat"=>32.838355, "lon"=>-109.759036),
"/US/US-AZ/three-points-950" => array("type"=>"cdp", "pop"=>2175, "name_ru"=>"Три Пойнтс", "name_en"=>"Three Points", "lang"=>"en", "lat"=>32.067515, "lon"=>-111.278411),
"/US/US-AZ/tolleson-951" => array("type"=>"city", "pop"=>7199, "name_ru"=>"Толсон", "name_en"=>"Tolleson", "lang"=>"en", "lat"=>33.448800, "lon"=>-112.254824),
"/US/US-AZ/tombstone-952" => array("type"=>"city", "pop"=>1566, "name_ru"=>"Томбстон", "name_en"=>"Tombstone", "lang"=>"en", "lat"=>31.717305, "lon"=>-110.062216),
"/US/US-AZ/tsaile-953" => array("type"=>"cdp", "pop"=>1043, "name_ru"=>"Тсейл", "name_en"=>"Tsaile", "lang"=>"en", "lat"=>36.303712, "lon"=>-109.214705),
"/US/US-AZ/tuba-city-954" => array("type"=>"cdp", "pop"=>7323, "name_ru"=>"Таба Сити", "name_en"=>"Tuba City", "lang"=>"en", "lat"=>36.124981, "lon"=>-111.242465),
"/US/US-AZ/tucson-955" => array("type"=>"city", "pop"=>541811, "name_ru"=>"Тусон ", "name_en"=>"Tucson", "lang"=>"en", "lat"=>32.195816, "lon"=>-110.891717),
"/US/US-AZ/tucson-estates-956" => array("type"=>"cdp", "pop"=>2662, "name_ru"=>"Тусон Эстейтс", "name_en"=>"Tucson Estates", "lang"=>"en", "lat"=>32.189550, "lon"=>-111.084058),
"/US/US-AZ/valencia-west-957" => array("type"=>"cdp", "pop"=>3277, "name_ru"=>"Валенсия Уэст", "name_en"=>"Valencia West", "lang"=>"en", "lat"=>32.135139, "lon"=>-111.106913),
"/US/US-AZ/wellton-958" => array("type"=>"town", "pop"=>1911, "name_ru"=>"Уэллтон", "name_en"=>"Wellton", "lang"=>"en", "lat"=>32.672888, "lon"=>-114.139327),
"/US/US-AZ/whetstone-959" => array("type"=>"cdp", "pop"=>1289, "name_ru"=>"Уэтостон", "name_en"=>"Whetstone", "lang"=>"en", "lat"=>31.701705, "lon"=>-110.340746),
"/US/US-AZ/whiteriver-960" => array("type"=>"cdp", "pop"=>3775, "name_ru"=>"Уайтривер", "name_en"=>"Whiteriver", "lang"=>"en", "lat"=>33.838919, "lon"=>-109.961619),
"/US/US-AZ/wickenburg-961" => array("type"=>"town", "pop"=>6620, "name_ru"=>"Уикенбург", "name_en"=>"Wickenburg", "lang"=>"en", "lat"=>33.966615, "lon"=>-112.754409),
"/US/US-AZ/willcox-962" => array("type"=>"city", "pop"=>3799, "name_ru"=>"Уиллкокс", "name_en"=>"Willcox", "lang"=>"en", "lat"=>32.250670, "lon"=>-109.835163),
"/US/US-AZ/williams-963" => array("type"=>"city", "pop"=>3305, "name_ru"=>"Уилльямс", "name_en"=>"Williams", "lang"=>"en", "lat"=>35.264635, "lon"=>-112.172838),
"/US/US-AZ/willow-valley-964" => array("type"=>"cdp", "pop"=>355, "name_ru"=>"Уиллоу Валли", "name_en"=>"Willow Valley", "lang"=>"en", "lat"=>34.912035, "lon"=>-114.605802),
"/US/US-AZ/window-rock-965" => array("type"=>"cdp", "pop"=>3306, "name_ru"=>"Уиндоу Рок", "name_en"=>"Window Rock", "lang"=>"en", "lat"=>35.670895, "lon"=>-109.064370),
"/US/US-AZ/winkelman-966" => array("type"=>"town", "pop"=>429, "name_ru"=>"Уинкелман", "name_en"=>"Winkelman", "lang"=>"en", "lat"=>32.989003, "lon"=>-110.769577),
"/US/US-AZ/winslow-967" => array("type"=>"city", "pop"=>9867, "name_ru"=>"Уинслоу", "name_en"=>"Winslow", "lang"=>"en", "lat"=>35.027779, "lon"=>-110.706934),
"/US/US-AZ/youngtown-968" => array("type"=>"town", "pop"=>4896, "name_ru"=>"Янгтаун", "name_en"=>"Youngtown", "lang"=>"en", "lat"=>33.587150, "lon"=>-112.303049),
"/US/US-AZ/yuma-969" => array("type"=>"city", "pop"=>90041, "name_ru"=>"Юма", "name_en"=>"Yuma", "lang"=>"en", "lat"=>32.677450, "lon"=>-114.620157),
"/US/US-CA" => array("type"=>"state", "pop"=>36756666, "name_ru"=>"Калифорния", "name_en"=>"California", "lang"=>"en"),
"/US/US-CA/acton-1471" => array("type"=>"cdp", "pop"=>1471, "name_ru"=>"Актон", "name_en"=>"Acton", "lang"=>"en", "lat"=>34.482780, "lon"=>-118.182468),
"/US/US-CA/adelanto-1472" => array("type"=>"city", "pop"=>28553, "name_ru"=>"Аделанто", "name_en"=>"Adelanto", "lang"=>"en", "lat"=>34.569658, "lon"=>-117.436906),
"/US/US-CA/agoura-hills-1473" => array("type"=>"city", "pop"=>22283, "name_ru"=>"Агаура Хиллс", "name_en"=>"Agoura Hills", "lang"=>"en", "lat"=>34.151019, "lon"=>-118.759800),
"/US/US-CA/alameda-1474" => array("type"=>"city", "pop"=>70580, "name_ru"=>"Аламеда", "name_en"=>"Alameda", "lang"=>"en", "lat"=>37.752767, "lon"=>-122.275886),
"/US/US-CA/alamo-1475" => array("type"=>"cdp", "pop"=>12277, "name_ru"=>"Аламо", "name_en"=>"Alamo", "lang"=>"en", "lat"=>37.857067, "lon"=>-121.998635),
"/US/US-CA/albany-1476" => array("type"=>"city", "pop"=>16004, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>37.890817, "lon"=>-122.327153),
"/US/US-CA/alhambra-1477" => array("type"=>"city", "pop"=>85953, "name_ru"=>"Альгамбра", "name_en"=>"Alhambra", "lang"=>"en", "lat"=>34.084000, "lon"=>-118.134674),
"/US/US-CA/aliso-viejo-1478" => array("type"=>"cdp", "pop"=>41572, "name_ru"=>"Алисо Вьехо", "name_en"=>"Aliso Viejo", "lang"=>"en", "lat"=>33.567541, "lon"=>-117.725305),
"/US/US-CA/alondra-park-1479" => array("type"=>"cdp", "pop"=>12215, "name_ru"=>"Алондра Парк", "name_en"=>"Alondra Park", "lang"=>"en", "lat"=>33.889550, "lon"=>-118.329974),
"/US/US-CA/alpine-1480" => array("type"=>"cdp", "pop"=>1061, "name_ru"=>"Алпайн", "name_en"=>"Alpine", "lang"=>"en", "lat"=>32.842967, "lon"=>-116.761138),
"/US/US-CA/alta-sierra-1482" => array("type"=>"cdp", "pop"=>5709, "name_ru"=>"Алта Сьерра", "name_en"=>"Alta Sierra", "lang"=>"en", "lat"=>39.122861, "lon"=>-121.052610),
"/US/US-CA/altadena-1481" => array("type"=>"cdp", "pop"=>42658, "name_ru"=>"Алтадена", "name_en"=>"Altadena", "lang"=>"en", "lat"=>34.192250, "lon"=>-118.134716),
"/US/US-CA/alturas-1483" => array("type"=>"city", "pop"=>2788, "name_ru"=>"Алчерас", "name_en"=>"Alturas", "lang"=>"en", "lat"=>41.491507, "lon"=>-120.545442),
"/US/US-CA/amador-city-1484" => array("type"=>"city", "pop"=>211, "name_ru"=>"Амадор Сити", "name_en"=>"Amador City", "lang"=>"en", "lat"=>38.418460, "lon"=>-120.823284),
"/US/US-CA/american-canyon-1485" => array("type"=>"cdp", "pop"=>16315, "name_ru"=>"Американ Каньон", "name_en"=>"American Canyon", "lang"=>"en", "lat"=>38.173200, "lon"=>-122.261173),
"/US/US-CA/anaheim-1486" => array("type"=>"city", "pop"=>335288, "name_ru"=>"Анахайм", "name_en"=>"Anaheim", "lang"=>"en", "lat"=>33.838900, "lon"=>-117.872331),
"/US/US-CA/anderson-1487" => array("type"=>"city", "pop"=>10538, "name_ru"=>"Андерсон", "name_en"=>"Anderson", "lang"=>"en", "lat"=>40.449918, "lon"=>-122.294197),
"/US/US-CA/angels-1488" => array("type"=>"city", "pop"=>3720, "name_ru"=>"Анджелс", "name_en"=>"Angels", "lang"=>"en", "lat"=>38.075786, "lon"=>-120.547890),
"/US/US-CA/angwin-1489" => array("type"=>"cdp", "pop"=>3503, "name_ru"=>"Анджвин", "name_en"=>"Angwin", "lang"=>"en", "lat"=>38.577650, "lon"=>-122.447950),
"/US/US-CA/antioch-1490" => array("type"=>"city", "pop"=>100219, "name_ru"=>"Антиок", "name_en"=>"Antioch", "lang"=>"en", "lat"=>37.988733, "lon"=>-121.798891),
"/US/US-CA/apple-valley-1491" => array("type"=>"town", "pop"=>70200, "name_ru"=>"Апл Валли", "name_en"=>"Apple Valley", "lang"=>"en", "lat"=>34.529346, "lon"=>-117.214164),
"/US/US-CA/aptos-1492" => array("type"=>"cdp", "pop"=>9061, "name_ru"=>"Аптос", "name_en"=>"Aptos", "lang"=>"en", "lat"=>36.991873, "lon"=>-121.899608),
"/US/US-CA/aptos-hills-larkin-valley-1493" => array("type"=>"cdp", "pop"=>2205, "name_ru"=>"Аптос Хилс-Ларкин Валли", "name_en"=>"Aptos Hills-Larkin Valley", "lang"=>"en", "lat"=>36.960963, "lon"=>-121.830229),
"/US/US-CA/arbuckle-1494" => array("type"=>"cdp", "pop"=>1912, "name_ru"=>"Арбакл", "name_en"=>"Arbuckle", "lang"=>"en", "lat"=>39.016277, "lon"=>-122.060608),
"/US/US-CA/arcadia-1495" => array("type"=>"city", "pop"=>56248, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>34.132167, "lon"=>-118.035692),
"/US/US-CA/arcata-1496" => array("type"=>"city", "pop"=>17044, "name_ru"=>"Арката", "name_en"=>"Arcata", "lang"=>"en", "lat"=>40.867849, "lon"=>-124.080892),
"/US/US-CA/arden-arcade-1497" => array("type"=>"cdp", "pop"=>92040, "name_ru"=>"Арден-Аркейд", "name_en"=>"Arden-Arcade", "lang"=>"en", "lat"=>38.601098, "lon"=>-121.376691),
"/US/US-CA/armona-1498" => array("type"=>"cdp", "pop"=>3122, "name_ru"=>"Армона", "name_en"=>"Armona", "lang"=>"en", "lat"=>36.318018, "lon"=>-119.705896),
"/US/US-CA/arnold-1499" => array("type"=>"cdp", "pop"=>3788, "name_ru"=>"Арнолд", "name_en"=>"Arnold", "lang"=>"en", "lat"=>38.232194, "lon"=>-120.369553),
"/US/US-CA/aromas-1500" => array("type"=>"cdp", "pop"=>2275, "name_ru"=>"Аромас", "name_en"=>"Aromas", "lang"=>"en", "lat"=>36.877450, "lon"=>-121.640349),
"/US/US-CA/arroyo-grande-1501" => array("type"=>"city", "pop"=>17180, "name_ru"=>"Арройо Гранде", "name_en"=>"Arroyo Grande", "lang"=>"en", "lat"=>35.125330, "lon"=>-120.583599),
"/US/US-CA/artesia-1502" => array("type"=>"city", "pop"=>16244, "name_ru"=>"Артижа", "name_en"=>"Artesia", "lang"=>"en", "lat"=>33.867200, "lon"=>-118.079724),
"/US/US-CA/arvin-1503" => array("type"=>"city", "pop"=>15070, "name_ru"=>"Арвин", "name_en"=>"Arvin", "lang"=>"en", "lat"=>35.200284, "lon"=>-118.833705),
"/US/US-CA/ashland-1504" => array("type"=>"cdp", "pop"=>16590, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>37.694700, "lon"=>-122.115358),
"/US/US-CA/atascadero-1505" => array("type"=>"city", "pop"=>28452, "name_ru"=>"Атаскадеро", "name_en"=>"Atascadero", "lang"=>"en", "lat"=>35.485484, "lon"=>-120.687924),
"/US/US-CA/atherton-1506" => array("type"=>"town", "pop"=>7406, "name_ru"=>"Атертон", "name_en"=>"Atherton", "lang"=>"en", "lat"=>37.454650, "lon"=>-122.202434),
"/US/US-CA/atwater-1507" => array("type"=>"city", "pop"=>26565, "name_ru"=>"Атуотер", "name_en"=>"Atwater", "lang"=>"en", "lat"=>37.353278, "lon"=>-120.600006),
"/US/US-CA/auberry-1508" => array("type"=>"cdp", "pop"=>1866, "name_ru"=>"Оберри", "name_en"=>"Auberry", "lang"=>"en", "lat"=>37.077805, "lon"=>-119.493624),
"/US/US-CA/auburn-1509" => array("type"=>"city", "pop"=>13219, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>38.891993, "lon"=>-121.076060),
"/US/US-CA/august-1510" => array("type"=>"cdp", "pop"=>6376, "name_ru"=>"Огаст", "name_en"=>"August", "lang"=>"en", "lat"=>37.978950, "lon"=>-121.261002),
"/US/US-CA/avalon-1511" => array("type"=>"city", "pop"=>3101, "name_ru"=>"Авалон", "name_en"=>"Avalon", "lang"=>"en", "lat"=>33.342918, "lon"=>-118.327386),
"/US/US-CA/avenal-1512" => array("type"=>"city", "pop"=>17147, "name_ru"=>"Авенал", "name_en"=>"Avenal", "lang"=>"en", "lat"=>36.029588, "lon"=>-120.114713),
"/US/US-CA/avocado-heights-1513" => array("type"=>"cdp", "pop"=>14232, "name_ru"=>"Авакадо Хайтс", "name_en"=>"Avocado Heights", "lang"=>"en", "lat"=>34.038383, "lon"=>-118.001567),
"/US/US-CA/azusa-1514" => array("type"=>"city", "pop"=>46847, "name_ru"=>"Азуса", "name_en"=>"Azusa", "lang"=>"en", "lat"=>34.138100, "lon"=>-117.912256),
"/US/US-CA/bakersfield-1515" => array("type"=>"city", "pop"=>321078, "name_ru"=>"Бейкерсфилд", "name_en"=>"Bakersfield", "lang"=>"en", "lat"=>35.357700, "lon"=>-119.004553),
"/US/US-CA/baldwin-park-1516" => array("type"=>"city", "pop"=>77380, "name_ru"=>"Бодуин Парк", "name_en"=>"Baldwin Park", "lang"=>"en", "lat"=>34.082800, "lon"=>-117.970779),
"/US/US-CA/banning-1517" => array("type"=>"city", "pop"=>28917, "name_ru"=>"Баннинг", "name_en"=>"Banning", "lang"=>"en", "lat"=>33.937295, "lon"=>-116.892059),
"/US/US-CA/barstow-1518" => array("type"=>"city", "pop"=>24596, "name_ru"=>"Барстоу", "name_en"=>"Barstow", "lang"=>"en", "lat"=>34.877606, "lon"=>-117.065870),
"/US/US-CA/bayview-1519" => array("type"=>"cdp", "pop"=>1318, "name_ru"=>"Бэйвью", "name_en"=>"Bayview", "lang"=>"en", "lat"=>40.764984, "lon"=>-124.175941),
"/US/US-CA/bayview-montalvin-1520" => array("type"=>"cdp", "pop"=>3988, "name_ru"=>"Бэйвью-Монталвин", "name_en"=>"Bayview-Montalvin", "lang"=>"en", "lat"=>38.001633, "lon"=>-122.322699),
"/US/US-CA/baywood-los-osos-1521" => array("type"=>"cdp", "pop"=>14377, "name_ru"=>"Бэйвуд-Лос Осос", "name_en"=>"Baywood-Los Osos", "lang"=>"en", "lat"=>35.314604, "lon"=>-120.839374),
"/US/US-CA/beale-afb-1522" => array("type"=>"cdp", "pop"=>6912, "name_ru"=>"Военно-воздушная база Бил", "name_en"=>"Beale AFB", "lang"=>"en", "lat"=>39.113708, "lon"=>-121.361343),
"/US/US-CA/bear-valley-springs-1523" => array("type"=>"cdp", "pop"=>1593, "name_ru"=>"Бэар Валли Спрингс", "name_en"=>"Bear Valley Springs", "lang"=>"en", "lat"=>35.159261, "lon"=>-118.627518),
"/US/US-CA/beaumont-1524" => array("type"=>"city", "pop"=>32663, "name_ru"=>"Бомонт", "name_en"=>"Beaumont", "lang"=>"en", "lat"=>33.931794, "lon"=>-116.978523),
"/US/US-CA/bell-1525" => array("type"=>"city", "pop"=>36657, "name_ru"=>"Белл", "name_en"=>"Bell", "lang"=>"en", "lat"=>33.979650, "lon"=>-118.178255),
"/US/US-CA/bell-gardens-1527" => array("type"=>"city", "pop"=>44692, "name_ru"=>"Белл Гарденс", "name_en"=>"Bell Gardens", "lang"=>"en", "lat"=>33.967277, "lon"=>-118.152767),
"/US/US-CA/bellflower-1526" => array("type"=>"city", "pop"=>73033, "name_ru"=>"Беллфлауэр", "name_en"=>"Bellflower", "lang"=>"en", "lat"=>33.888133, "lon"=>-118.126632),
"/US/US-CA/belmont-1528" => array("type"=>"city", "pop"=>24776, "name_ru"=>"Белмонт", "name_en"=>"Belmont", "lang"=>"en", "lat"=>37.515250, "lon"=>-122.295138),
"/US/US-CA/belvedere-1529" => array("type"=>"city", "pop"=>2059, "name_ru"=>"Бельведер", "name_en"=>"Belvedere", "lang"=>"en", "lat"=>37.871650, "lon"=>-122.468708),
"/US/US-CA/ben-lomond-1531" => array("type"=>"cdp", "pop"=>7884, "name_ru"=>"Бен Ломонд", "name_en"=>"Ben Lomond", "lang"=>"en", "lat"=>37.103361, "lon"=>-122.084478),
"/US/US-CA/benicia-1530" => array("type"=>"city", "pop"=>26174, "name_ru"=>"Бенишия", "name_en"=>"Benicia", "lang"=>"en", "lat"=>38.071850, "lon"=>-122.154093),
"/US/US-CA/berkeley-1532" => array("type"=>"city", "pop"=>101371, "name_ru"=>"Беркли", "name_en"=>"Berkeley", "lang"=>"en", "lat"=>37.867250, "lon"=>-122.297290),
"/US/US-CA/bermuda-dunes-1533" => array("type"=>"cdp", "pop"=>4571, "name_ru"=>"Бермуда Дьюнс", "name_en"=>"Bermuda Dunes", "lang"=>"en", "lat"=>33.742668, "lon"=>-116.288365),
"/US/US-CA/bethel-island-1534" => array("type"=>"cdp", "pop"=>2115, "name_ru"=>"Бетел Айленд", "name_en"=>"Bethel Island", "lang"=>"en", "lat"=>38.032167, "lon"=>-121.646797),
"/US/US-CA/beverly-hills-1535" => array("type"=>"city", "pop"=>34445, "name_ru"=>"Беверли Хиллс", "name_en"=>"Beverly Hills", "lang"=>"en", "lat"=>34.079250, "lon"=>-118.401368),
"/US/US-CA/big-bear-city-1536" => array("type"=>"cdp", "pop"=>4920, "name_ru"=>"Биг Бэар Сити", "name_en"=>"Big Bear City", "lang"=>"en", "lat"=>34.266342, "lon"=>-116.845530),
"/US/US-CA/big-bear-lake-1537" => array("type"=>"city", "pop"=>6131, "name_ru"=>"Биг Бэар Лейк", "name_en"=>"Big Bear Lake", "lang"=>"en", "lat"=>34.243690, "lon"=>-116.894483),
"/US/US-CA/big-pine-1539" => array("type"=>"cdp", "pop"=>1158, "name_ru"=>"Биг Пайн", "name_en"=>"Big Pine", "lang"=>"en", "lat"=>37.165447, "lon"=>-118.295386),
"/US/US-CA/big-river-1540" => array("type"=>"cdp", "pop"=>705, "name_ru"=>"Биг Ривер", "name_en"=>"Big River", "lang"=>"en", "lat"=>34.140099, "lon"=>-114.360545),
"/US/US-CA/biggs-1538" => array("type"=>"city", "pop"=>1820, "name_ru"=>"Бигс", "name_en"=>"Biggs", "lang"=>"en", "lat"=>39.413933, "lon"=>-121.709232),
"/US/US-CA/bishop-1541" => array("type"=>"city", "pop"=>3410, "name_ru"=>"Бишоп", "name_en"=>"Bishop", "lang"=>"en", "lat"=>37.367696, "lon"=>-118.396804),
"/US/US-CA/blackhawk-1542" => array("type"=>"cdp", "pop"=>6199, "name_ru"=>"Блэкхок", "name_en"=>"Blackhawk", "lang"=>"en", "lat"=>37.820863, "lon"=>-121.906704),
"/US/US-CA/bloomington-1543" => array("type"=>"cdp", "pop"=>15116, "name_ru"=>"Блумингтон", "name_en"=>"Bloomington", "lang"=>"en", "lat"=>34.059450, "lon"=>-117.397849),
"/US/US-CA/blue-lake-1544" => array("type"=>"city", "pop"=>1104, "name_ru"=>"Блу Лейк", "name_en"=>"Blue Lake", "lang"=>"en", "lat"=>40.880236, "lon"=>-123.991588),
"/US/US-CA/bluewater-1545" => array("type"=>"cdp", "pop"=>261, "name_ru"=>"Блууотер", "name_en"=>"Bluewater", "lang"=>"en", "lat"=>34.174558, "lon"=>-114.270658),
"/US/US-CA/blythe-1546" => array("type"=>"city", "pop"=>21727, "name_ru"=>"Блайт", "name_en"=>"Blythe", "lang"=>"en", "lat"=>33.609383, "lon"=>-114.590083),
"/US/US-CA/bodega-bay-1547" => array("type"=>"cdp", "pop"=>1127, "name_ru"=>"Бодега Бэй", "name_en"=>"Bodega Bay", "lang"=>"en", "lat"=>38.322798, "lon"=>-123.029788),
"/US/US-CA/bodfish-1548" => array("type"=>"cdp", "pop"=>1283, "name_ru"=>"Бодфиш", "name_en"=>"Bodfish", "lang"=>"en", "lat"=>35.582538, "lon"=>-118.482780),
"/US/US-CA/bolinas-1549" => array("type"=>"cdp", "pop"=>1098, "name_ru"=>"Болинас", "name_en"=>"Bolinas", "lang"=>"en", "lat"=>37.905600, "lon"=>-122.696993),
"/US/US-CA/bonadelle-ranchos-madera-ranchos-1550" => array("type"=>"cdp", "pop"=>5705, "name_ru"=>"Бонадел Ранчос-Мадера Ранчос", "name_en"=>"Bonadelle Ranchos-Madera Ranchos", "lang"=>"en", "lat"=>36.959752, "lon"=>-119.895761),
"/US/US-CA/bonita-1551" => array("type"=>"cdp", "pop"=>12542, "name_ru"=>"Бонита", "name_en"=>"Bonita", "lang"=>"en", "lat"=>32.667483, "lon"=>-117.029885),
"/US/US-CA/bonsall-1552" => array("type"=>"cdp", "pop"=>1881, "name_ru"=>"Бонсолл", "name_en"=>"Bonsall", "lang"=>"en", "lat"=>33.282488, "lon"=>-117.220172),
"/US/US-CA/bootjack-1553" => array("type"=>"cdp", "pop"=>1295, "name_ru"=>"Бутджек", "name_en"=>"Bootjack", "lang"=>"en", "lat"=>37.469708, "lon"=>-119.883104),
"/US/US-CA/boron-1554" => array("type"=>"cdp", "pop"=>2101, "name_ru"=>"Борон", "name_en"=>"Boron", "lang"=>"en", "lat"=>35.004921, "lon"=>-117.650339),
"/US/US-CA/borrego-springs-1555" => array("type"=>"cdp", "pop"=>2244, "name_ru"=>"Боррего Спрингс", "name_en"=>"Borrego Springs", "lang"=>"en", "lat"=>33.238707, "lon"=>-116.354538),
"/US/US-CA/bostonia-1556" => array("type"=>"cdp", "pop"=>13670, "name_ru"=>"Бостония", "name_en"=>"Bostonia", "lang"=>"en", "lat"=>32.819757, "lon"=>-116.945731),
"/US/US-CA/boulder-creek-1557" => array("type"=>"cdp", "pop"=>6725, "name_ru"=>"Боулдер Крик", "name_en"=>"Boulder Creek", "lang"=>"en", "lat"=>37.137500, "lon"=>-122.127873),
"/US/US-CA/boyes-hot-springs-1558" => array("type"=>"cdp", "pop"=>5973, "name_ru"=>"Бойес Хот Спрингс", "name_en"=>"Boyes Hot Springs", "lang"=>"en", "lat"=>38.313050, "lon"=>-122.485011),
"/US/US-CA/bradbury-1559" => array("type"=>"city", "pop"=>1063, "name_ru"=>"Бредбери", "name_en"=>"Bradbury", "lang"=>"en", "lat"=>34.152490, "lon"=>-117.968406),
"/US/US-CA/brawley-1560" => array("type"=>"city", "pop"=>22725, "name_ru"=>"Броли", "name_en"=>"Brawley", "lang"=>"en", "lat"=>32.979830, "lon"=>-115.533373),
"/US/US-CA/brea-1561" => array("type"=>"city", "pop"=>38314, "name_ru"=>"Брея", "name_en"=>"Brea", "lang"=>"en", "lat"=>33.923567, "lon"=>-117.867211),
"/US/US-CA/brentwood-1562" => array("type"=>"city", "pop"=>49480, "name_ru"=>"Брентвуд", "name_en"=>"Brentwood", "lang"=>"en", "lat"=>37.933626, "lon"=>-121.708636),
"/US/US-CA/brisbane-1563" => array("type"=>"city", "pop"=>3677, "name_ru"=>"Брисбен", "name_en"=>"Brisbane", "lang"=>"en", "lat"=>37.689317, "lon"=>-122.399928),
"/US/US-CA/broadmoor-1564" => array("type"=>"cdp", "pop"=>3739, "name_ru"=>"Бродмур", "name_en"=>"Broadmoor", "lang"=>"en", "lat"=>37.691700, "lon"=>-122.479566),
"/US/US-CA/buellton-1565" => array("type"=>"cdp", "pop"=>4292, "name_ru"=>"Буэллтон", "name_en"=>"Buellton", "lang"=>"en", "lat"=>34.625929, "lon"=>-120.201215),
"/US/US-CA/buena-park-1566" => array("type"=>"city", "pop"=>79379, "name_ru"=>"Буэна Парк", "name_en"=>"Buena Park", "lang"=>"en", "lat"=>33.857700, "lon"=>-118.003274),
"/US/US-CA/burbank-1567" => array("type"=>"city", "pop"=>102968, "name_ru"=>"Бурбанк", "name_en"=>"Burbank", "lang"=>"en", "lat"=>34.191317, "lon"=>-118.325713),
"/US/US-CA/burbank-1568" => array("type"=>"cdp", "pop"=>4902, "name_ru"=>"Бурбанк", "name_en"=>"Burbank", "lang"=>"en", "lat"=>37.316403, "lon"=>-121.931747),
"/US/US-CA/burlingame-1569" => array("type"=>"city", "pop"=>27706, "name_ru"=>"Берлингейм", "name_en"=>"Burlingame", "lang"=>"en", "lat"=>37.590300, "lon"=>-122.362596),
"/US/US-CA/burney-1570" => array("type"=>"cdp", "pop"=>3423, "name_ru"=>"Берни", "name_en"=>"Burney", "lang"=>"en", "lat"=>40.884312, "lon"=>-121.668689),
"/US/US-CA/buttonwillow-1571" => array("type"=>"cdp", "pop"=>1301, "name_ru"=>"Баттонвиллоу", "name_en"=>"Buttonwillow", "lang"=>"en", "lat"=>35.401758, "lon"=>-119.469806),
"/US/US-CA/cabazon-1572" => array("type"=>"cdp", "pop"=>1588, "name_ru"=>"Кабазон", "name_en"=>"Cabazon", "lang"=>"en", "lat"=>33.911405, "lon"=>-116.771744),
"/US/US-CA/calexico-1573" => array("type"=>"city", "pop"=>38344, "name_ru"=>"Калексико", "name_en"=>"Calexico", "lang"=>"en", "lat"=>32.678473, "lon"=>-115.500754),
"/US/US-CA/california-city-1574" => array("type"=>"city", "pop"=>14556, "name_ru"=>"Калифорния Сити", "name_en"=>"California City", "lang"=>"en", "lat"=>35.189303, "lon"=>-117.811549),
"/US/US-CA/calimesa-1575" => array("type"=>"cdp", "pop"=>7454, "name_ru"=>"Калимеса", "name_en"=>"Calimesa", "lang"=>"en", "lat"=>33.995900, "lon"=>-117.045148),
"/US/US-CA/calipatria-1576" => array("type"=>"city", "pop"=>7594, "name_ru"=>"Калипатрия", "name_en"=>"Calipatria", "lang"=>"en", "lat"=>33.127239, "lon"=>-115.517630),
"/US/US-CA/calistoga-1577" => array("type"=>"city", "pop"=>5179, "name_ru"=>"Калистога", "name_en"=>"Calistoga", "lang"=>"en", "lat"=>38.581450, "lon"=>-122.581663),
"/US/US-CA/camarillo-1578" => array("type"=>"city", "pop"=>63324, "name_ru"=>"Камарилло", "name_en"=>"Camarillo", "lang"=>"en", "lat"=>34.222550, "lon"=>-119.029434),
"/US/US-CA/cambria-1579" => array("type"=>"cdp", "pop"=>5382, "name_ru"=>"Камбрия", "name_en"=>"Cambria", "lang"=>"en", "lat"=>35.546548, "lon"=>-121.079872),
"/US/US-CA/cambrian-park-1580" => array("type"=>"cdp", "pop"=>2998, "name_ru"=>"Камбриан Прак", "name_en"=>"Cambrian Park", "lang"=>"en", "lat"=>37.255650, "lon"=>-121.927849),
"/US/US-CA/cameron-park-1581" => array("type"=>"cdp", "pop"=>11897, "name_ru"=>"Камерон Парк", "name_en"=>"Cameron Park", "lang"=>"en", "lat"=>38.681727, "lon"=>-120.987207),
"/US/US-CA/camp-pendleton-north-1583" => array("type"=>"cdp", "pop"=>10373, "name_ru"=>"Кэип Пендлтон Норт", "name_en"=>"Camp Pendleton North", "lang"=>"en", "lat"=>33.313998, "lon"=>-117.314552),
"/US/US-CA/camp-pendleton-south-1584" => array("type"=>"cdp", "pop"=>11299, "name_ru"=>"Кэмп Пендлтон Сауф", "name_en"=>"Camp Pendleton South", "lang"=>"en", "lat"=>33.233315, "lon"=>-117.373547),
"/US/US-CA/campbell-1582" => array("type"=>"city", "pop"=>38617, "name_ru"=>"Кэмпбелл", "name_en"=>"Campbell", "lang"=>"en", "lat"=>37.278800, "lon"=>-121.953350),
"/US/US-CA/canyon-lake-1585" => array("type"=>"cdp", "pop"=>11243, "name_ru"=>"Каньон Лейк", "name_en"=>"Canyon Lake", "lang"=>"en", "lat"=>33.683493, "lon"=>-117.253358),
"/US/US-CA/capitola-1586" => array("type"=>"city", "pop"=>9612, "name_ru"=>"Капитола", "name_en"=>"Capitola", "lang"=>"en", "lat"=>36.976050, "lon"=>-121.952975),
"/US/US-CA/carlsbad-1587" => array("type"=>"city", "pop"=>96374, "name_ru"=>"Карлсбад", "name_en"=>"Carlsbad", "lang"=>"en", "lat"=>33.123826, "lon"=>-117.285865),
"/US/US-CA/carmel-by-the-sea-1588" => array("type"=>"city", "pop"=>3886, "name_ru"=>"Кармел-бай-зе-Си", "name_en"=>"Carmel-by-the-Sea", "lang"=>"en", "lat"=>36.553350, "lon"=>-121.920949),
"/US/US-CA/carmel-valley-village-1589" => array("type"=>"cdp", "pop"=>4407, "name_ru"=>"Кармел Валли Виллидж", "name_en"=>"Carmel Valley Village", "lang"=>"en", "lat"=>36.486245, "lon"=>-121.722892),
"/US/US-CA/carmichael-1590" => array("type"=>"cdp", "pop"=>48702, "name_ru"=>"Кармайкл", "name_en"=>"Carmichael", "lang"=>"en", "lat"=>38.633950, "lon"=>-121.318774),
"/US/US-CA/carpinteria-1591" => array("type"=>"city", "pop"=>13655, "name_ru"=>"Карпинтерия", "name_en"=>"Carpinteria", "lang"=>"en", "lat"=>34.387900, "lon"=>-119.513217),
"/US/US-CA/carson-1592" => array("type"=>"city", "pop"=>92366, "name_ru"=>"Карсон", "name_en"=>"Carson", "lang"=>"en", "lat"=>33.836183, "lon"=>-118.257889),
"/US/US-CA/caruthers-1593" => array("type"=>"cdp", "pop"=>1603, "name_ru"=>"Карутерс", "name_en"=>"Caruthers", "lang"=>"en", "lat"=>36.541800, "lon"=>-119.841688),
"/US/US-CA/casa-conejo-1594" => array("type"=>"cdp", "pop"=>3286, "name_ru"=>"Каса Коньехо", "name_en"=>"Casa Conejo", "lang"=>"en", "lat"=>34.183603, "lon"=>-118.942500),
"/US/US-CA/casa-de-oro-mount-helix-1595" => array("type"=>"cdp", "pop"=>30727, "name_ru"=>"Каса де Оро-Маунт Хеликс", "name_en"=>"Casa de Oro-Mount Helix", "lang"=>"en", "lat"=>32.765401, "lon"=>-116.938076),
"/US/US-CA/castro-valley-1596" => array("type"=>"cdp", "pop"=>48619, "name_ru"=>"Кастро Валли", "name_en"=>"Castro Valley", "lang"=>"en", "lat"=>37.710950, "lon"=>-122.061114),
"/US/US-CA/castroville-1597" => array("type"=>"cdp", "pop"=>5272, "name_ru"=>"Кастровилл", "name_en"=>"Castroville", "lang"=>"en", "lat"=>36.764050, "lon"=>-121.751849),
"/US/US-CA/cathedral-city-1598" => array("type"=>"city", "pop"=>52095, "name_ru"=>"Катедрал Сити", "name_en"=>"Cathedral City", "lang"=>"en", "lat"=>33.830833, "lon"=>-116.463409),
"/US/US-CA/cayucos-1599" => array("type"=>"cdp", "pop"=>2960, "name_ru"=>"Каюкос", "name_en"=>"Cayucos", "lang"=>"en", "lat"=>35.437951, "lon"=>-120.884804),
"/US/US-CA/central-valley-1600" => array("type"=>"cdp", "pop"=>4340, "name_ru"=>"Сентрал Валли", "name_en"=>"Central Valley", "lang"=>"en", "lat"=>40.679596, "lon"=>-122.374912),
"/US/US-CA/ceres-1601" => array("type"=>"city", "pop"=>42690, "name_ru"=>"Серс", "name_en"=>"Ceres", "lang"=>"en", "lat"=>37.599351, "lon"=>-120.956232),
"/US/US-CA/cerritos-1602" => array("type"=>"city", "pop"=>51326, "name_ru"=>"Серритос", "name_en"=>"Cerritos", "lang"=>"en", "lat"=>33.867603, "lon"=>-118.068402),
"/US/US-CA/challenge-brownsville-1603" => array("type"=>"cdp", "pop"=>1096, "name_ru"=>"Челлендж-Браунсвилл", "name_en"=>"Challenge-Brownsville", "lang"=>"en", "lat"=>39.461514, "lon"=>-121.258945),
"/US/US-CA/channel-islands-beach-1604" => array("type"=>"cdp", "pop"=>3317, "name_ru"=>"Чаннел Айлендс Бич", "name_en"=>"Channel Islands Beach", "lang"=>"en", "lat"=>34.158050, "lon"=>-119.222101),
"/US/US-CA/charter-oak-1605" => array("type"=>"cdp", "pop"=>8858, "name_ru"=>"Чартер Оук", "name_en"=>"Charter Oak", "lang"=>"en", "lat"=>34.102400, "lon"=>-117.853849),
"/US/US-CA/cherry-valley-1607" => array("type"=>"cdp", "pop"=>5945, "name_ru"=>"Черри Валли", "name_en"=>"Cherry Valley", "lang"=>"en", "lat"=>33.972450, "lon"=>-116.965307),
"/US/US-CA/cherryland-1606" => array("type"=>"cdp", "pop"=>11088, "name_ru"=>"Черриленд", "name_en"=>"Cherryland", "lang"=>"en", "lat"=>37.679350, "lon"=>-122.102268),
"/US/US-CA/chester-1608" => array("type"=>"cdp", "pop"=>2082, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>40.300372, "lon"=>-121.233191),
"/US/US-CA/chico-1609" => array("type"=>"city", "pop"=>83791, "name_ru"=>"Чико", "name_en"=>"Chico", "lang"=>"en", "lat"=>39.753803, "lon"=>-121.806931),
"/US/US-CA/chino-1610" => array("type"=>"city", "pop"=>83031, "name_ru"=>"Чино", "name_en"=>"Chino", "lang"=>"en", "lat"=>34.006900, "lon"=>-117.685998),
"/US/US-CA/chino-hills-1611" => array("type"=>"cdp", "pop"=>73879, "name_ru"=>"Чино Хиллс", "name_en"=>"Chino Hills", "lang"=>"en", "lat"=>33.974789, "lon"=>-117.748651),
"/US/US-CA/chowchilla-1612" => array("type"=>"city", "pop"=>19193, "name_ru"=>"Чоучилла", "name_en"=>"Chowchilla", "lang"=>"en", "lat"=>37.115422, "lon"=>-120.256583),
"/US/US-CA/chula-vista-1613" => array("type"=>"city", "pop"=>219318, "name_ru"=>"Чула Виста", "name_en"=>"Chula Vista", "lang"=>"en", "lat"=>32.628450, "lon"=>-117.044733),
"/US/US-CA/citrus-1614" => array("type"=>"cdp", "pop"=>9481, "name_ru"=>"Ситрес", "name_en"=>"Citrus", "lang"=>"en", "lat"=>34.114883, "lon"=>-117.890891),
"/US/US-CA/citrus-heights-1615" => array("type"=>"cdp", "pop"=>84432, "name_ru"=>"Ситрес Хайтс", "name_en"=>"Citrus Heights", "lang"=>"en", "lat"=>38.691650, "lon"=>-121.286669),
"/US/US-CA/claremont-1616" => array("type"=>"city", "pop"=>35470, "name_ru"=>"Клэрмонт", "name_en"=>"Claremont", "lang"=>"en", "lat"=>34.121850, "lon"=>-117.715734),
"/US/US-CA/clayton-1617" => array("type"=>"city", "pop"=>11278, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>37.940983, "lon"=>-121.928938),
"/US/US-CA/clearlake-1618" => array("type"=>"city", "pop"=>15040, "name_ru"=>"Клирлейк", "name_en"=>"Clearlake", "lang"=>"en", "lat"=>38.958202, "lon"=>-122.631587),
"/US/US-CA/clearlake-oaks-1619" => array("type"=>"cdp", "pop"=>2419, "name_ru"=>"Клирлейк Оукс", "name_en"=>"Clearlake Oaks", "lang"=>"en", "lat"=>39.022531, "lon"=>-122.665069),
"/US/US-CA/cloverdale-1620" => array("type"=>"city", "pop"=>8153, "name_ru"=>"Кловердейл", "name_en"=>"Cloverdale", "lang"=>"en", "lat"=>38.799681, "lon"=>-123.017422),
"/US/US-CA/clovis-1621" => array("type"=>"city", "pop"=>92318, "name_ru"=>"Кловис", "name_en"=>"Clovis", "lang"=>"en", "lat"=>36.819633, "lon"=>-119.696023),
"/US/US-CA/coachella-1622" => array("type"=>"city", "pop"=>39391, "name_ru"=>"Кочелла", "name_en"=>"Coachella", "lang"=>"en", "lat"=>33.680933, "lon"=>-116.152291),
"/US/US-CA/coalinga-1623" => array("type"=>"city", "pop"=>18974, "name_ru"=>"Колинга", "name_en"=>"Coalinga", "lang"=>"en", "lat"=>36.142446, "lon"=>-120.354600),
"/US/US-CA/cobb-1624" => array("type"=>"cdp", "pop"=>1477, "name_ru"=>"Кобб", "name_en"=>"Cobb", "lang"=>"en", "lat"=>38.837819, "lon"=>-122.724147),
"/US/US-CA/colfax-1625" => array("type"=>"city", "pop"=>1738, "name_ru"=>"Колфакс", "name_en"=>"Colfax", "lang"=>"en", "lat"=>39.095268, "lon"=>-120.953169),
"/US/US-CA/colma-1626" => array("type"=>"town", "pop"=>1422, "name_ru"=>"Колма", "name_en"=>"Colma", "lang"=>"en", "lat"=>37.673833, "lon"=>-122.453420),
"/US/US-CA/colton-1627" => array("type"=>"city", "pop"=>50517, "name_ru"=>"Колтон", "name_en"=>"Colton", "lang"=>"en", "lat"=>34.057900, "lon"=>-117.322002),
"/US/US-CA/columbia-1628" => array("type"=>"cdp", "pop"=>1799, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>38.026745, "lon"=>-120.401805),
"/US/US-CA/colusa-1629" => array("type"=>"city", "pop"=>5892, "name_ru"=>"Колуса", "name_en"=>"Colusa", "lang"=>"en", "lat"=>39.206097, "lon"=>-122.011330),
"/US/US-CA/commerce-1630" => array("type"=>"city", "pop"=>13435, "name_ru"=>"Коммерс", "name_en"=>"Commerce", "lang"=>"en", "lat"=>33.995367, "lon"=>-118.150241),
"/US/US-CA/compton-1631" => array("type"=>"city", "pop"=>93851, "name_ru"=>"Комптон", "name_en"=>"Compton", "lang"=>"en", "lat"=>33.892550, "lon"=>-118.226527),
"/US/US-CA/concord-1632" => array("type"=>"city", "pop"=>121160, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>37.973250, "lon"=>-121.999870),
"/US/US-CA/concow-1633" => array("type"=>"cdp", "pop"=>1392, "name_ru"=>"Конкоу", "name_en"=>"Concow", "lang"=>"en", "lat"=>39.745901, "lon"=>-121.493635),
"/US/US-CA/corcoran-1634" => array("type"=>"city", "pop"=>25139, "name_ru"=>"Коркоран", "name_en"=>"Corcoran", "lang"=>"en", "lat"=>36.081844, "lon"=>-119.559990),
"/US/US-CA/corning-1635" => array("type"=>"city", "pop"=>7207, "name_ru"=>"Корнинг", "name_en"=>"Corning", "lang"=>"en", "lat"=>39.926338, "lon"=>-122.182001),
"/US/US-CA/corona-1636" => array("type"=>"city", "pop"=>149923, "name_ru"=>"Корона", "name_en"=>"Corona", "lang"=>"en", "lat"=>33.867900, "lon"=>-117.569693),
"/US/US-CA/coronado-1637" => array("type"=>"city", "pop"=>22633, "name_ru"=>"Коронадо", "name_en"=>"Coronado", "lang"=>"en", "lat"=>32.630757, "lon"=>-117.173682),
"/US/US-CA/corralitos-1638" => array("type"=>"cdp", "pop"=>2513, "name_ru"=>"Корралитос", "name_en"=>"Corralitos", "lang"=>"en", "lat"=>36.996633, "lon"=>-121.798644),
"/US/US-CA/corte-madera-1639" => array("type"=>"town", "pop"=>9245, "name_ru"=>"Корте Мадера", "name_en"=>"Corte Madera", "lang"=>"en", "lat"=>37.923615, "lon"=>-122.507319),
"/US/US-CA/costa-mesa-1640" => array("type"=>"city", "pop"=>110080, "name_ru"=>"Коста Меса", "name_en"=>"Costa Mesa", "lang"=>"en", "lat"=>33.666850, "lon"=>-117.912649),
"/US/US-CA/cotati-1641" => array("type"=>"city", "pop"=>7169, "name_ru"=>"Котати", "name_en"=>"Cotati", "lang"=>"en", "lat"=>38.327793, "lon"=>-122.709471),
"/US/US-CA/coto-de-caza-1642" => array("type"=>"cdp", "pop"=>2853, "name_ru"=>"Кото де Каза", "name_en"=>"Coto De Caza", "lang"=>"en", "lat"=>33.604119, "lon"=>-117.586042),
"/US/US-CA/cottonwood-1643" => array("type"=>"cdp", "pop"=>1747, "name_ru"=>"Коттонвуд", "name_en"=>"Cottonwood", "lang"=>"en", "lat"=>40.390318, "lon"=>-122.279978),
"/US/US-CA/country-club-1644" => array("type"=>"cdp", "pop"=>9325, "name_ru"=>"Кантри Клаб", "name_en"=>"Country Club", "lang"=>"en", "lat"=>37.968973, "lon"=>-121.339679),
"/US/US-CA/covelo-1645" => array("type"=>"cdp", "pop"=>1057, "name_ru"=>"Ковело", "name_en"=>"Covelo", "lang"=>"en", "lat"=>39.800185, "lon"=>-123.247762),
"/US/US-CA/covina-1646" => array("type"=>"city", "pop"=>46944, "name_ru"=>"Ковина", "name_en"=>"Covina", "lang"=>"en", "lat"=>34.090150, "lon"=>-117.881072),
"/US/US-CA/crescent-city-1647" => array("type"=>"city", "pop"=>7852, "name_ru"=>"Кресцент Сити", "name_en"=>"Crescent City", "lang"=>"en", "lat"=>41.754127, "lon"=>-124.197830),
"/US/US-CA/crescent-city-north-1648" => array("type"=>"cdp", "pop"=>3853, "name_ru"=>"Кресценд Сити Норт", "name_en"=>"Crescent City North", "lang"=>"en", "lat"=>41.763674, "lon"=>-124.214092),
"/US/US-CA/crestline-1649" => array("type"=>"cdp", "pop"=>8594, "name_ru"=>"Крестлайн", "name_en"=>"Crestline", "lang"=>"en", "lat"=>34.246674, "lon"=>-117.295756),
"/US/US-CA/crockett-1650" => array("type"=>"cdp", "pop"=>3228, "name_ru"=>"Крокетт", "name_en"=>"Crockett", "lang"=>"en", "lat"=>38.051076, "lon"=>-122.219432),
"/US/US-CA/cudahy-1651" => array("type"=>"city", "pop"=>24334, "name_ru"=>"Кудаи", "name_en"=>"Cudahy", "lang"=>"en", "lat"=>33.964200, "lon"=>-118.181671),
"/US/US-CA/culver-city-1652" => array("type"=>"city", "pop"=>38580, "name_ru"=>"Калвер Сити", "name_en"=>"Culver City", "lang"=>"en", "lat"=>34.005800, "lon"=>-118.396116),
"/US/US-CA/cupertino-1653" => array("type"=>"city", "pop"=>53637, "name_ru"=>"Купертино", "name_en"=>"Cupertino", "lang"=>"en", "lat"=>37.314403, "lon"=>-122.047333),
"/US/US-CA/cutler-1654" => array("type"=>"cdp", "pop"=>4450, "name_ru"=>"Катлер", "name_en"=>"Cutler", "lang"=>"en", "lat"=>36.526687, "lon"=>-119.287222),
"/US/US-CA/cutten-1655" => array("type"=>"cdp", "pop"=>1516, "name_ru"=>"Каттен", "name_en"=>"Cutten", "lang"=>"en", "lat"=>40.767442, "lon"=>-124.141997),
"/US/US-CA/cypress-1656" => array("type"=>"city", "pop"=>47123, "name_ru"=>"Сайпресс", "name_en"=>"Cypress", "lang"=>"en", "lat"=>33.816350, "lon"=>-118.037911),
"/US/US-CA/daly-city-1657" => array("type"=>"city", "pop"=>101514, "name_ru"=>"Дейли Сити", "name_en"=>"Daly City", "lang"=>"en", "lat"=>37.687000, "lon"=>-122.467432),
"/US/US-CA/dana-point-1658" => array("type"=>"city", "pop"=>35700, "name_ru"=>"Дана Пойнт", "name_en"=>"Dana Point", "lang"=>"en", "lat"=>33.457000, "lon"=>-117.696337),
"/US/US-CA/danville-1659" => array("type"=>"city", "pop"=>41182, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>37.814400, "lon"=>-121.971381),
"/US/US-CA/davis-1660" => array("type"=>"city", "pop"=>62593, "name_ru"=>"Давис", "name_en"=>"Davis", "lang"=>"en", "lat"=>38.554950, "lon"=>-121.735946),
"/US/US-CA/day-valley-1661" => array("type"=>"cdp", "pop"=>2842, "name_ru"=>"Дей Валли", "name_en"=>"Day Valley", "lang"=>"en", "lat"=>37.035846, "lon"=>-121.861259),
"/US/US-CA/deer-park-1662" => array("type"=>"cdp", "pop"=>1825, "name_ru"=>"Дир Парк", "name_en"=>"Deer Park", "lang"=>"en", "lat"=>38.532767, "lon"=>-122.468616),
"/US/US-CA/del-aire-1663" => array("type"=>"cdp", "pop"=>8040, "name_ru"=>"Дель Эр", "name_en"=>"Del Aire", "lang"=>"en", "lat"=>33.916083, "lon"=>-118.368633),
"/US/US-CA/del-mar-1666" => array("type"=>"city", "pop"=>4447, "name_ru"=>"Дель Мар", "name_en"=>"Del Mar", "lang"=>"en", "lat"=>32.964900, "lon"=>-117.262449),
"/US/US-CA/del-monte-forest-1667" => array("type"=>"cdp", "pop"=>5069, "name_ru"=>"Дель Монте Форест", "name_en"=>"Del Monte Forest", "lang"=>"en", "lat"=>36.586251, "lon"=>-121.946294),
"/US/US-CA/del-rey-1668" => array("type"=>"cdp", "pop"=>1150, "name_ru"=>"Дель Рей", "name_en"=>"Del Rey", "lang"=>"en", "lat"=>36.657750, "lon"=>-119.595985),
"/US/US-CA/del-rey-oaks-1669" => array("type"=>"city", "pop"=>1529, "name_ru"=>"Дель Рей Оукс", "name_en"=>"Del Rey Oaks", "lang"=>"en", "lat"=>36.592791, "lon"=>-121.837299),
"/US/US-CA/delano-1664" => array("type"=>"city", "pop"=>53051, "name_ru"=>"Делано", "name_en"=>"Delano", "lang"=>"en", "lat"=>35.764844, "lon"=>-119.250428),
"/US/US-CA/delhi-1665" => array("type"=>"cdp", "pop"=>3280, "name_ru"=>"Дели", "name_en"=>"Delhi", "lang"=>"en", "lat"=>37.434130, "lon"=>-120.776917),
"/US/US-CA/denair-1670" => array("type"=>"cdp", "pop"=>3693, "name_ru"=>"Денейр", "name_en"=>"Denair", "lang"=>"en", "lat"=>37.529602, "lon"=>-120.800139),
"/US/US-CA/desert-hot-springs-1671" => array("type"=>"city", "pop"=>24489, "name_ru"=>"Дезерт Хот Спрингс", "name_en"=>"Desert Hot Springs", "lang"=>"en", "lat"=>33.959148, "lon"=>-116.500180),
"/US/US-CA/desert-view-highlands-1672" => array("type"=>"cdp", "pop"=>2154, "name_ru"=>"Дезерт Вью Хайлендс", "name_en"=>"Desert View Highlands", "lang"=>"en", "lat"=>34.589749, "lon"=>-118.152606),
"/US/US-CA/diamond-bar-1673" => array("type"=>"city", "pop"=>57235, "name_ru"=>"Даймонд Бар", "name_en"=>"Diamond Bar", "lang"=>"en", "lat"=>33.999403, "lon"=>-117.814277),
"/US/US-CA/diamond-springs-1674" => array("type"=>"cdp", "pop"=>2872, "name_ru"=>"Даймонд Спрингс", "name_en"=>"Diamond Springs", "lang"=>"en", "lat"=>38.685960, "lon"=>-120.821557),
"/US/US-CA/dinuba-1675" => array("type"=>"city", "pop"=>20335, "name_ru"=>"Динуба", "name_en"=>"Dinuba", "lang"=>"en", "lat"=>36.545611, "lon"=>-119.387310),
"/US/US-CA/discovery-bay-1676" => array("type"=>"cdp", "pop"=>5351, "name_ru"=>"Дискавери Бэй", "name_en"=>"Discovery Bay", "lang"=>"en", "lat"=>37.908783, "lon"=>-121.595220),
"/US/US-CA/dixon-1677" => array("type"=>"city", "pop"=>17412, "name_ru"=>"Диксон", "name_en"=>"Dixon", "lang"=>"en", "lat"=>38.444700, "lon"=>-121.823749),
"/US/US-CA/dixon-lane-meadow-creek-1678" => array("type"=>"cdp", "pop"=>2561, "name_ru"=>"Диксон Лейн-Медоу Крик", "name_en"=>"Dixon Lane-Meadow Creek", "lang"=>"en", "lat"=>37.387329, "lon"=>-118.412499),
"/US/US-CA/dollar-point-1679" => array("type"=>"cdp", "pop"=>1449, "name_ru"=>"Дллар Пойнт", "name_en"=>"Dollar Point", "lang"=>"en", "lat"=>39.189739, "lon"=>-120.108411),
"/US/US-CA/dorris-1680" => array("type"=>"city", "pop"=>831, "name_ru"=>"Доррис", "name_en"=>"Dorris", "lang"=>"en", "lat"=>41.965194, "lon"=>-121.919952),
"/US/US-CA/dos-palos-1681" => array("type"=>"city", "pop"=>4948, "name_ru"=>"Дос Палос", "name_en"=>"Dos Palos", "lang"=>"en", "lat"=>36.986524, "lon"=>-120.636821),
"/US/US-CA/downey-1682" => array("type"=>"city", "pop"=>107587, "name_ru"=>"Дауни", "name_en"=>"Downey", "lang"=>"en", "lat"=>33.938150, "lon"=>-118.129899),
"/US/US-CA/duarte-1683" => array("type"=>"city", "pop"=>21839, "name_ru"=>"Дуарте", "name_en"=>"Duarte", "lang"=>"en", "lat"=>34.157833, "lon"=>-117.953118),
"/US/US-CA/dublin-1684" => array("type"=>"city", "pop"=>44297, "name_ru"=>"Дублин", "name_en"=>"Dublin", "lang"=>"en", "lat"=>37.715911, "lon"=>-121.909908),
"/US/US-CA/dunsmuir-1685" => array("type"=>"city", "pop"=>1801, "name_ru"=>"Дансмьюир", "name_en"=>"Dunsmuir", "lang"=>"en", "lat"=>41.233544, "lon"=>-122.269889),
"/US/US-CA/durham-1686" => array("type"=>"cdp", "pop"=>4784, "name_ru"=>"Дарем", "name_en"=>"Durham", "lang"=>"en", "lat"=>39.599528, "lon"=>-121.827678),
"/US/US-CA/earlimart-1687" => array("type"=>"cdp", "pop"=>5881, "name_ru"=>"Эрлимарт", "name_en"=>"Earlimart", "lang"=>"en", "lat"=>35.880858, "lon"=>-119.270535),
"/US/US-CA/east-blythe-1688" => array("type"=>"cdp", "pop"=>1511, "name_ru"=>"Ист Блайт", "name_en"=>"East Blythe", "lang"=>"en", "lat"=>33.611706, "lon"=>-114.577309),
"/US/US-CA/east-compton-1689" => array("type"=>"cdp", "pop"=>7967, "name_ru"=>"Ист Комптон", "name_en"=>"East Compton", "lang"=>"en", "lat"=>33.898100, "lon"=>-118.194323),
"/US/US-CA/east-foothills-1690" => array("type"=>"cdp", "pop"=>14898, "name_ru"=>"Ист Футхиллс", "name_en"=>"East Foothills", "lang"=>"en", "lat"=>37.381133, "lon"=>-121.816431),
"/US/US-CA/east-hemet-1691" => array("type"=>"cdp", "pop"=>17611, "name_ru"=>"Ист Хемет", "name_en"=>"East Hemet", "lang"=>"en", "lat"=>33.739950, "lon"=>-116.937999),
"/US/US-CA/east-la-mirada-1692" => array("type"=>"cdp", "pop"=>9367, "name_ru"=>"Ист Ла Мирада", "name_en"=>"East La Mirada", "lang"=>"en", "lat"=>33.924390, "lon"=>-117.988079),
"/US/US-CA/east-los-angeles-1693" => array("type"=>"cdp", "pop"=>126379, "name_ru"=>"Ист Лос Анджелес", "name_en"=>"East Los Angeles", "lang"=>"en", "lat"=>34.032650, "lon"=>-118.168562),
"/US/US-CA/east-palo-alto-1695" => array("type"=>"city", "pop"=>33575, "name_ru"=>"Ист Пало Альто", "name_en"=>"East Palo Alto", "lang"=>"en", "lat"=>37.466050, "lon"=>-122.132199),
"/US/US-CA/east-pasadena-1696" => array("type"=>"cdp", "pop"=>5910, "name_ru"=>"Ист Пасадена", "name_en"=>"East Pasadena", "lang"=>"en", "lat"=>34.138250, "lon"=>-118.076824),
"/US/US-CA/east-porterville-1697" => array("type"=>"cdp", "pop"=>5790, "name_ru"=>"Ист Портервилл", "name_en"=>"East Porterville", "lang"=>"en", "lat"=>36.057400, "lon"=>-118.974699),
"/US/US-CA/east-richmond-heights-1698" => array("type"=>"cdp", "pop"=>3266, "name_ru"=>"Ист Ричмонд Хайтс", "name_en"=>"East Richmond Heights", "lang"=>"en", "lat"=>37.945100, "lon"=>-122.312599),
"/US/US-CA/east-san-gabriel-1699" => array("type"=>"cdp", "pop"=>12736, "name_ru"=>"Ист Сан Габриэль", "name_en"=>"East San Gabriel", "lang"=>"en", "lat"=>34.119850, "lon"=>-118.079696),
"/US/US-CA/east-sonora-1700" => array("type"=>"cdp", "pop"=>1675, "name_ru"=>"Ист Сонора", "name_en"=>"East Sonora", "lang"=>"en", "lat"=>37.980530, "lon"=>-120.339296),
"/US/US-CA/easton-1694" => array("type"=>"cdp", "pop"=>1877, "name_ru"=>"Истон", "name_en"=>"Easton", "lang"=>"en", "lat"=>36.651700, "lon"=>-119.788869),
"/US/US-CA/edwards-afb-1701" => array("type"=>"cdp", "pop"=>7423, "name_ru"=>"Военно-воздушная база Эдвардс", "name_en"=>"Edwards AFB", "lang"=>"en", "lat"=>34.909620, "lon"=>-117.934573),
"/US/US-CA/el-cajon-1702" => array("type"=>"city", "pop"=>92718, "name_ru"=>"Эль Каджон", "name_en"=>"El Cajon", "lang"=>"en", "lat"=>32.802450, "lon"=>-116.960849),
"/US/US-CA/el-centro-1703" => array("type"=>"city", "pop"=>40083, "name_ru"=>"Эль Сентро", "name_en"=>"El Centro", "lang"=>"en", "lat"=>32.787145, "lon"=>-115.559374),
"/US/US-CA/el-cerrito-1704" => array("type"=>"city", "pop"=>22222, "name_ru"=>"Эль Серрито", "name_en"=>"El Cerrito", "lang"=>"en", "lat"=>37.920500, "lon"=>-122.302070),
"/US/US-CA/el-cerrito-1705" => array("type"=>"cdp", "pop"=>4490, "name_ru"=>"Эль Серрито", "name_en"=>"El Cerrito", "lang"=>"en", "lat"=>33.840133, "lon"=>-117.521899),
"/US/US-CA/el-dorado-hills-1706" => array("type"=>"cdp", "pop"=>6395, "name_ru"=>"Элб Дорадо Хилс", "name_en"=>"El Dorado Hills", "lang"=>"en", "lat"=>38.685856, "lon"=>-121.080976),
"/US/US-CA/el-granada-1708" => array("type"=>"cdp", "pop"=>4426, "name_ru"=>"Эль Гранада", "name_en"=>"El Granada", "lang"=>"en", "lat"=>37.509667, "lon"=>-122.466581),
"/US/US-CA/el-monte-1711" => array("type"=>"city", "pop"=>121791, "name_ru"=>"Эль Монте", "name_en"=>"El Monte", "lang"=>"en", "lat"=>34.074800, "lon"=>-118.027797),
"/US/US-CA/el-paso-de-robles-(paso-robles)-1712" => array("type"=>"city", "pop"=>28715, "name_ru"=>"Эль Пасо де Роблес (Пасо Роблес)", "name_en"=>"El Paso de Robles (Paso Robles)", "lang"=>"en", "lat"=>35.632279, "lon"=>-120.664187),
"/US/US-CA/el-rio-1713" => array("type"=>"cdp", "pop"=>6419, "name_ru"=>"Эль Рио", "name_en"=>"El Rio", "lang"=>"en", "lat"=>34.242650, "lon"=>-119.158731),
"/US/US-CA/el-segundo-1714" => array("type"=>"city", "pop"=>16217, "name_ru"=>"Эль Секундо", "name_en"=>"El Segundo", "lang"=>"en", "lat"=>33.910563, "lon"=>-118.424817),
"/US/US-CA/el-sobrante-1715" => array("type"=>"cdp", "pop"=>9852, "name_ru"=>"Эль Собранте", "name_en"=>"El Sobrante", "lang"=>"en", "lat"=>37.976630, "lon"=>-122.292614),
"/US/US-CA/el-toro-1716" => array("type"=>"cdp", "pop"=>62685, "name_ru"=>"Эль Торо", "name_en"=>"El Toro", "lang"=>"en", "lat"=>33.647067, "lon"=>-117.682728),
"/US/US-CA/el-toro-station-1717" => array("type"=>"cdp", "pop"=>6869, "name_ru"=>"Эль Торо Стейшен", "name_en"=>"El Toro Station", "lang"=>"en", "lat"=>33.679176, "lon"=>-117.718429),
"/US/US-CA/el-verano-1718" => array("type"=>"cdp", "pop"=>3498, "name_ru"=>"Эль Верано", "name_en"=>"El Verano", "lang"=>"en", "lat"=>38.297740, "lon"=>-122.489391),
"/US/US-CA/eldridge-1707" => array("type"=>"cdp", "pop"=>1144, "name_ru"=>"Элдридж", "name_en"=>"Eldridge", "lang"=>"en", "lat"=>38.335182, "lon"=>-122.506871),
"/US/US-CA/elk-grove-1709" => array("type"=>"cdp", "pop"=>133003, "name_ru"=>"Элк Гров", "name_en"=>"Elk Grove", "lang"=>"en", "lat"=>38.400767, "lon"=>-121.367488),
"/US/US-CA/elkhorn-1710" => array("type"=>"cdp", "pop"=>1458, "name_ru"=>"Элкхом", "name_en"=>"Elkhorn", "lang"=>"en", "lat"=>36.810850, "lon"=>-121.717472),
"/US/US-CA/emerald-lake-hills-1719" => array("type"=>"cdp", "pop"=>3328, "name_ru"=>"Эмералд Лейк Хилс", "name_en"=>"Emerald Lake Hills", "lang"=>"en", "lat"=>37.464600, "lon"=>-122.266922),
"/US/US-CA/emeryville-1720" => array("type"=>"city", "pop"=>9583, "name_ru"=>"Эмеривилл", "name_en"=>"Emeryville", "lang"=>"en", "lat"=>37.839083, "lon"=>-122.299343),
"/US/US-CA/encinitas-1721" => array("type"=>"city", "pop"=>60372, "name_ru"=>"Энсинитас", "name_en"=>"Encinitas", "lang"=>"en", "lat"=>33.047655, "lon"=>-117.262552),
"/US/US-CA/escalon-1722" => array("type"=>"city", "pop"=>7271, "name_ru"=>"Эскалон", "name_en"=>"Escalon", "lang"=>"en", "lat"=>37.789488, "lon"=>-120.996215),
"/US/US-CA/escondido-1723" => array("type"=>"city", "pop"=>137103, "name_ru"=>"Эскондидо", "name_en"=>"Escondido", "lang"=>"en", "lat"=>33.136445, "lon"=>-117.071270),
"/US/US-CA/esparto-1724" => array("type"=>"cdp", "pop"=>1487, "name_ru"=>"Эспарто", "name_en"=>"Esparto", "lang"=>"en", "lat"=>38.693728, "lon"=>-122.021158),
"/US/US-CA/etna-1725" => array("type"=>"city", "pop"=>769, "name_ru"=>"Этна", "name_en"=>"Etna", "lang"=>"en", "lat"=>41.459278, "lon"=>-122.892843),
"/US/US-CA/eureka-1726" => array("type"=>"city", "pop"=>25300, "name_ru"=>"Юрика", "name_en"=>"Eureka", "lang"=>"en", "lat"=>40.795319, "lon"=>-124.155679),
"/US/US-CA/exeter-1727" => array("type"=>"city", "pop"=>9963, "name_ru"=>"Эксетер", "name_en"=>"Exeter", "lang"=>"en", "lat"=>36.294587, "lon"=>-119.143026),
"/US/US-CA/fair-oaks-1730" => array("type"=>"cdp", "pop"=>26867, "name_ru"=>"Фэр Оукс", "name_en"=>"Fair Oaks", "lang"=>"en", "lat"=>38.648500, "lon"=>-121.247178),
"/US/US-CA/fairfax-1728" => array("type"=>"town", "pop"=>7066, "name_ru"=>"Фэрфакс", "name_en"=>"Fairfax", "lang"=>"en", "lat"=>37.988633, "lon"=>-122.593828),
"/US/US-CA/fairfield-1729" => array("type"=>"city", "pop"=>103683, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>38.252500, "lon"=>-122.041655),
"/US/US-CA/fairview-1731" => array("type"=>"cdp", "pop"=>9045, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>37.678633, "lon"=>-122.044640),
"/US/US-CA/fallbrook-1732" => array("type"=>"cdp", "pop"=>22095, "name_ru"=>"Фолбрук", "name_en"=>"Fallbrook", "lang"=>"en", "lat"=>33.368600, "lon"=>-117.237752),
"/US/US-CA/farmersville-1733" => array("type"=>"city", "pop"=>10056, "name_ru"=>"Фармерсвилл", "name_en"=>"Farmersville", "lang"=>"en", "lat"=>36.304148, "lon"=>-119.206868),
"/US/US-CA/felton-1734" => array("type"=>"cdp", "pop"=>5350, "name_ru"=>"Фелтон", "name_en"=>"Felton", "lang"=>"en", "lat"=>37.042700, "lon"=>-122.072192),
"/US/US-CA/ferndale-1735" => array("type"=>"city", "pop"=>1384, "name_ru"=>"Ферндейл", "name_en"=>"Ferndale", "lang"=>"en", "lat"=>40.579486, "lon"=>-124.260625),
"/US/US-CA/fetters-hot-springs-agua-caliente-1736" => array("type"=>"cdp", "pop"=>2024, "name_ru"=>"Феттерс Хот Спрингс-Агуа Калиенте", "name_en"=>"Fetters Hot Springs-Agua Caliente", "lang"=>"en", "lat"=>38.322267, "lon"=>-122.479784),
"/US/US-CA/fillmore-1737" => array("type"=>"city", "pop"=>15053, "name_ru"=>"Фильмор", "name_en"=>"Fillmore", "lang"=>"en", "lat"=>34.400160, "lon"=>-118.918476),
"/US/US-CA/firebaugh-1738" => array("type"=>"city", "pop"=>6972, "name_ru"=>"Файербо", "name_en"=>"Firebaugh", "lang"=>"en", "lat"=>36.849218, "lon"=>-120.448936),
"/US/US-CA/florence-graham-1739" => array("type"=>"cdp", "pop"=>57147, "name_ru"=>"Флоренс-Грейам", "name_en"=>"Florence-Graham", "lang"=>"en", "lat"=>33.967050, "lon"=>-118.244374),
"/US/US-CA/florin-1740" => array("type"=>"cdp", "pop"=>24330, "name_ru"=>"Флорин", "name_en"=>"Florin", "lang"=>"en", "lat"=>38.490650, "lon"=>-121.404395),
"/US/US-CA/folsom-1741" => array("type"=>"city", "pop"=>67788, "name_ru"=>"Фолсом", "name_en"=>"Folsom", "lang"=>"en", "lat"=>38.676712, "lon"=>-121.146231),
"/US/US-CA/fontana-1742" => array("type"=>"city", "pop"=>184984, "name_ru"=>"Фонтана", "name_en"=>"Fontana", "lang"=>"en", "lat"=>34.097733, "lon"=>-117.457340),
"/US/US-CA/foothill-farms-1743" => array("type"=>"cdp", "pop"=>17135, "name_ru"=>"Футхил Фармс", "name_en"=>"Foothill Farms", "lang"=>"en", "lat"=>38.677589, "lon"=>-121.346716),
"/US/US-CA/ford-city-1744" => array("type"=>"cdp", "pop"=>3781, "name_ru"=>"Форд Сити", "name_en"=>"Ford City", "lang"=>"en", "lat"=>35.166489, "lon"=>-119.458412),
"/US/US-CA/foresthill-1745" => array("type"=>"cdp", "pop"=>1409, "name_ru"=>"Форестхилл", "name_en"=>"Foresthill", "lang"=>"en", "lat"=>38.999421, "lon"=>-120.832107),
"/US/US-CA/forestville-1746" => array("type"=>"cdp", "pop"=>2443, "name_ru"=>"Форествилл", "name_en"=>"Forestville", "lang"=>"en", "lat"=>38.481133, "lon"=>-122.888583),
"/US/US-CA/fort-bragg-1747" => array("type"=>"city", "pop"=>6604, "name_ru"=>"Форт Брагг", "name_en"=>"Fort Bragg", "lang"=>"en", "lat"=>39.442277, "lon"=>-123.802906),
"/US/US-CA/fort-jones-1748" => array("type"=>"city", "pop"=>647, "name_ru"=>"Форт Джонс", "name_en"=>"Fort Jones", "lang"=>"en", "lat"=>41.607051, "lon"=>-122.839824),
"/US/US-CA/fortuna-1749" => array("type"=>"city", "pop"=>11329, "name_ru"=>"Фортуна", "name_en"=>"Fortuna", "lang"=>"en", "lat"=>40.585811, "lon"=>-124.139668),
"/US/US-CA/foster-city-1750" => array("type"=>"city", "pop"=>29089, "name_ru"=>"Фостер Сити", "name_en"=>"Foster City", "lang"=>"en", "lat"=>37.563806, "lon"=>-122.244666),
"/US/US-CA/fountain-valley-1751" => array("type"=>"city", "pop"=>55516, "name_ru"=>"Фонтейн Валли", "name_en"=>"Fountain Valley", "lang"=>"en", "lat"=>33.710900, "lon"=>-117.950324),
"/US/US-CA/fowler-1752" => array("type"=>"city", "pop"=>5563, "name_ru"=>"Фоулер", "name_en"=>"Fowler", "lang"=>"en", "lat"=>36.624700, "lon"=>-119.672261),
"/US/US-CA/frazier-park-1753" => array("type"=>"cdp", "pop"=>2201, "name_ru"=>"Фразьер Парк", "name_en"=>"Frazier Park", "lang"=>"en", "lat"=>34.822120, "lon"=>-118.945313),
"/US/US-CA/freedom-1754" => array("type"=>"cdp", "pop"=>8361, "name_ru"=>"Фридом", "name_en"=>"Freedom", "lang"=>"en", "lat"=>36.940033, "lon"=>-121.789321),
"/US/US-CA/fremont-1755" => array("type"=>"city", "pop"=>202867, "name_ru"=>"Фремонт", "name_en"=>"Fremont", "lang"=>"en", "lat"=>37.528506, "lon"=>-121.997122),
"/US/US-CA/french-camp-1756" => array("type"=>"cdp", "pop"=>3018, "name_ru"=>"Френч Кэмп", "name_en"=>"French Camp", "lang"=>"en", "lat"=>37.882817, "lon"=>-121.278733),
"/US/US-CA/fresno-1757" => array("type"=>"city", "pop"=>476050, "name_ru"=>"Фресно", "name_en"=>"Fresno", "lang"=>"en", "lat"=>36.780600, "lon"=>-119.792874),
"/US/US-CA/fullerton-1758" => array("type"=>"city", "pop"=>131868, "name_ru"=>"Фуллертон", "name_en"=>"Fullerton", "lang"=>"en", "lat"=>33.884800, "lon"=>-117.928054),
"/US/US-CA/galt-1759" => array("type"=>"city", "pop"=>24026, "name_ru"=>"Галт", "name_en"=>"Galt", "lang"=>"en", "lat"=>38.268640, "lon"=>-121.301165),
"/US/US-CA/garden-acres-1761" => array("type"=>"cdp", "pop"=>8547, "name_ru"=>"Гарден Эйкерс", "name_en"=>"Garden Acres", "lang"=>"en", "lat"=>37.963900, "lon"=>-121.228317),
"/US/US-CA/garden-grove-1762" => array("type"=>"city", "pop"=>165796, "name_ru"=>"Гарден Гров", "name_en"=>"Garden Grove", "lang"=>"en", "lat"=>33.778750, "lon"=>-117.959375),
"/US/US-CA/gardena-1760" => array("type"=>"city", "pop"=>58554, "name_ru"=>"Гардена", "name_en"=>"Gardena", "lang"=>"en", "lat"=>33.893600, "lon"=>-118.304849),
"/US/US-CA/george-afb-1763" => array("type"=>"cdp", "pop"=>5085, "name_ru"=>"Военно-воздушная база Джордж", "name_en"=>"George AFB", "lang"=>"en", "lat"=>34.581726, "lon"=>-117.366487),
"/US/US-CA/gerber-las-flores-1764" => array("type"=>"cdp", "pop"=>1143, "name_ru"=>"Гербер-Лас Флорес", "name_en"=>"Gerber-Las Flores", "lang"=>"en", "lat"=>40.061481, "lon"=>-122.149355),
"/US/US-CA/gilroy-1765" => array("type"=>"city", "pop"=>49934, "name_ru"=>"Гилрой", "name_en"=>"Gilroy", "lang"=>"en", "lat"=>37.010100, "lon"=>-121.578090),
"/US/US-CA/glen-avon-1766" => array("type"=>"cdp", "pop"=>12663, "name_ru"=>"Глен Эйвон", "name_en"=>"Glen Avon", "lang"=>"en", "lat"=>34.017350, "lon"=>-117.487637),
"/US/US-CA/glen-ellen-1769" => array("type"=>"cdp", "pop"=>1191, "name_ru"=>"Глен Эллен", "name_en"=>"Glen Ellen", "lang"=>"en", "lat"=>38.355800, "lon"=>-122.538362),
"/US/US-CA/glendale-1767" => array("type"=>"city", "pop"=>197176, "name_ru"=>"Глендейл", "name_en"=>"Glendale", "lang"=>"en", "lat"=>34.176720, "lon"=>-118.253257),
"/US/US-CA/glendora-1768" => array("type"=>"city", "pop"=>49410, "name_ru"=>"Глендора", "name_en"=>"Glendora", "lang"=>"en", "lat"=>34.146050, "lon"=>-117.843941),
"/US/US-CA/glenshire-devonshire-1770" => array("type"=>"cdp", "pop"=>2133, "name_ru"=>"Гленшир-Девоншир", "name_en"=>"Glenshire-Devonshire", "lang"=>"en", "lat"=>39.350502, "lon"=>-120.087529),
"/US/US-CA/golden-hills-1771" => array("type"=>"cdp", "pop"=>5423, "name_ru"=>"Голден Хилс", "name_en"=>"Golden Hills", "lang"=>"en", "lat"=>35.142495, "lon"=>-118.489348),
"/US/US-CA/gonzales-1772" => array("type"=>"city", "pop"=>8537, "name_ru"=>"Гонзалес", "name_en"=>"Gonzales", "lang"=>"en", "lat"=>36.506530, "lon"=>-121.442551),
"/US/US-CA/grand-terrace-1773" => array("type"=>"city", "pop"=>12204, "name_ru"=>"Гранд Террас", "name_en"=>"Grand Terrace", "lang"=>"en", "lat"=>34.030567, "lon"=>-117.314481),
"/US/US-CA/granite-hills-1774" => array("type"=>"cdp", "pop"=>3157, "name_ru"=>"Гранит Хилс", "name_en"=>"Granite Hills", "lang"=>"en", "lat"=>32.803000, "lon"=>-116.903759),
"/US/US-CA/grass-valley-1775" => array("type"=>"city", "pop"=>12232, "name_ru"=>"Грас Валли", "name_en"=>"Grass Valley", "lang"=>"en", "lat"=>39.221460, "lon"=>-121.056670),
"/US/US-CA/graton-1776" => array("type"=>"cdp", "pop"=>1409, "name_ru"=>"Гратон", "name_en"=>"Graton", "lang"=>"en", "lat"=>38.437700, "lon"=>-122.865271),
"/US/US-CA/greenacres-1777" => array("type"=>"cdp", "pop"=>7379, "name_ru"=>"Гринакрс", "name_en"=>"Greenacres", "lang"=>"en", "lat"=>35.383459, "lon"=>-119.125744),
"/US/US-CA/greenfield-1778" => array("type"=>"city", "pop"=>15222, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>36.321675, "lon"=>-121.242393),
"/US/US-CA/greenville-1779" => array("type"=>"cdp", "pop"=>1396, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>40.135760, "lon"=>-120.937881),
"/US/US-CA/gridley-1780" => array("type"=>"city", "pop"=>6478, "name_ru"=>"Гридли", "name_en"=>"Gridley", "lang"=>"en", "lat"=>39.364373, "lon"=>-121.695158),
"/US/US-CA/groveland-big-oak-flat-1781" => array("type"=>"cdp", "pop"=>2753, "name_ru"=>"Гровленд-Биг Оук Флэт", "name_en"=>"Groveland-Big Oak Flat", "lang"=>"en", "lat"=>37.845543, "lon"=>-120.196422),
"/US/US-CA/grover-city-1782" => array("type"=>"city", "pop"=>11656, "name_ru"=>"Гровер Сити", "name_en"=>"Grover City", "lang"=>"en", "lat"=>35.120958, "lon"=>-120.618527),
"/US/US-CA/guadalupe-1783" => array("type"=>"city", "pop"=>6593, "name_ru"=>"Гвадалупе", "name_en"=>"Guadalupe", "lang"=>"en", "lat"=>34.963509, "lon"=>-120.574683),
"/US/US-CA/guerneville-1784" => array("type"=>"cdp", "pop"=>1966, "name_ru"=>"Герневилл", "name_en"=>"Guerneville", "lang"=>"en", "lat"=>38.508058, "lon"=>-122.984863),
"/US/US-CA/gustine-1785" => array("type"=>"city", "pop"=>5094, "name_ru"=>"Гастин", "name_en"=>"Gustine", "lang"=>"en", "lat"=>37.252508, "lon"=>-120.993198),
"/US/US-CA/hacienda-heights-1786" => array("type"=>"cdp", "pop"=>52354, "name_ru"=>"Хасиенда Хайтс", "name_en"=>"Hacienda Heights", "lang"=>"en", "lat"=>33.995750, "lon"=>-117.971372),
"/US/US-CA/half-moon-bay-1787" => array("type"=>"city", "pop"=>12449, "name_ru"=>"Халф Мун Бэй", "name_en"=>"Half Moon Bay", "lang"=>"en", "lat"=>37.470116, "lon"=>-122.436896),
"/US/US-CA/hamilton-city-1788" => array("type"=>"cdp", "pop"=>1811, "name_ru"=>"Гамильтон Сити", "name_en"=>"Hamilton City", "lang"=>"en", "lat"=>39.742271, "lon"=>-122.011657),
"/US/US-CA/hanford-1789" => array("type"=>"city", "pop"=>50103, "name_ru"=>"Ханфорд", "name_en"=>"Hanford", "lang"=>"en", "lat"=>36.325040, "lon"=>-119.646935),
"/US/US-CA/harbison-canyon-1790" => array("type"=>"cdp", "pop"=>2122, "name_ru"=>"Харбисон Каньон", "name_en"=>"Harbison Canyon", "lang"=>"en", "lat"=>32.819371, "lon"=>-116.829472),
"/US/US-CA/hawaiian-gardens-1791" => array("type"=>"city", "pop"=>15229, "name_ru"=>"Гавайиан Гарденс", "name_en"=>"Hawaiian Gardens", "lang"=>"en", "lat"=>33.830450, "lon"=>-118.071874),
"/US/US-CA/hawthorne-1792" => array("type"=>"city", "pop"=>84305, "name_ru"=>"Хоуторн", "name_en"=>"Hawthorne", "lang"=>"en", "lat"=>33.914650, "lon"=>-118.347049),
"/US/US-CA/hayfork-1793" => array("type"=>"cdp", "pop"=>2605, "name_ru"=>"Хейфорк", "name_en"=>"Hayfork", "lang"=>"en", "lat"=>40.573872, "lon"=>-123.125268),
"/US/US-CA/hayward-1794" => array("type"=>"city", "pop"=>142061, "name_ru"=>"Хейворд", "name_en"=>"Hayward", "lang"=>"en", "lat"=>37.627667, "lon"=>-122.104602),
"/US/US-CA/healdsburg-1795" => array("type"=>"city", "pop"=>10971, "name_ru"=>"Хелдсбург", "name_en"=>"Healdsburg", "lang"=>"en", "lat"=>38.618497, "lon"=>-122.862762),
"/US/US-CA/heber-1796" => array("type"=>"cdp", "pop"=>2566, "name_ru"=>"Хебер", "name_en"=>"Heber", "lang"=>"en", "lat"=>32.734137, "lon"=>-115.520123),
"/US/US-CA/hemet-1797" => array("type"=>"city", "pop"=>70991, "name_ru"=>"Хемет", "name_en"=>"Hemet", "lang"=>"en", "lat"=>33.731603, "lon"=>-116.997758),
"/US/US-CA/hercules-1798" => array("type"=>"city", "pop"=>24484, "name_ru"=>"Геркулес", "name_en"=>"Hercules", "lang"=>"en", "lat"=>38.020433, "lon"=>-122.296894),
"/US/US-CA/hermosa-beach-1799" => array("type"=>"city", "pop"=>19350, "name_ru"=>"Хермоса Бич", "name_en"=>"Hermosa Beach", "lang"=>"en", "lat"=>33.851670, "lon"=>-118.422709),
"/US/US-CA/hesperia-1800" => array("type"=>"city", "pop"=>85883, "name_ru"=>"Хесперия", "name_en"=>"Hesperia", "lang"=>"en", "lat"=>34.419075, "lon"=>-117.302401),
"/US/US-CA/hidden-hills-1801" => array("type"=>"city", "pop"=>2021, "name_ru"=>"Хидден Хилс", "name_en"=>"Hidden Hills", "lang"=>"en", "lat"=>34.167552, "lon"=>-118.659997),
"/US/US-CA/hidden-meadows-1802" => array("type"=>"cdp", "pop"=>2371, "name_ru"=>"Хидден Медоус", "name_en"=>"Hidden Meadows", "lang"=>"en", "lat"=>33.225153, "lon"=>-117.111703),
"/US/US-CA/hidden-valley-lake-1803" => array("type"=>"cdp", "pop"=>1961, "name_ru"=>"Хидден Валли Лейк", "name_en"=>"Hidden Valley Lake", "lang"=>"en", "lat"=>38.808482, "lon"=>-122.517282),
"/US/US-CA/highgrove-1804" => array("type"=>"cdp", "pop"=>3175, "name_ru"=>"Хайгров", "name_en"=>"Highgrove", "lang"=>"en", "lat"=>34.014548, "lon"=>-117.328035),
"/US/US-CA/highland-1805" => array("type"=>"city", "pop"=>51096, "name_ru"=>"Хайленд", "name_en"=>"Highland", "lang"=>"en", "lat"=>34.117300, "lon"=>-117.189141),
"/US/US-CA/highlands-1806" => array("type"=>"cdp", "pop"=>2644, "name_ru"=>"Хайлендс", "name_en"=>"Highlands", "lang"=>"en", "lat"=>37.519900, "lon"=>-122.342779),
"/US/US-CA/hillsborough-1807" => array("type"=>"town", "pop"=>10844, "name_ru"=>"Хилсборо", "name_en"=>"Hillsborough", "lang"=>"en", "lat"=>37.557550, "lon"=>-122.356241),
"/US/US-CA/hilmar-irwin-1808" => array("type"=>"cdp", "pop"=>3392, "name_ru"=>"Хилмар-Ирвинг", "name_en"=>"Hilmar-Irwin", "lang"=>"en", "lat"=>37.405598, "lon"=>-120.850608),
"/US/US-CA/hollister-1809" => array("type"=>"city", "pop"=>34877, "name_ru"=>"Холлистер", "name_en"=>"Hollister", "lang"=>"en", "lat"=>36.855501, "lon"=>-121.399055),
"/US/US-CA/holtville-1810" => array("type"=>"city", "pop"=>5412, "name_ru"=>"Хотвилл", "name_en"=>"Holtville", "lang"=>"en", "lat"=>32.813668, "lon"=>-115.377342),
"/US/US-CA/home-garden-1811" => array("type"=>"cdp", "pop"=>1549, "name_ru"=>"Хоум Гарден", "name_en"=>"Home Garden", "lang"=>"en", "lat"=>36.303319, "lon"=>-119.635254),
"/US/US-CA/home-gardens-1812" => array("type"=>"cdp", "pop"=>7780, "name_ru"=>"Хоум Гарденс", "name_en"=>"Home Gardens", "lang"=>"en", "lat"=>33.880900, "lon"=>-117.513154),
"/US/US-CA/homeland-1813" => array("type"=>"cdp", "pop"=>3312, "name_ru"=>"Хоумленд", "name_en"=>"Homeland", "lang"=>"en", "lat"=>33.748467, "lon"=>-117.108576),
"/US/US-CA/hughson-1814" => array("type"=>"city", "pop"=>6312, "name_ru"=>"Хафсон", "name_en"=>"Hughson", "lang"=>"en", "lat"=>37.601001, "lon"=>-120.866700),
"/US/US-CA/humboldt-hill-1815" => array("type"=>"cdp", "pop"=>2865, "name_ru"=>"Гумбольдт Хил", "name_en"=>"Humboldt Hill", "lang"=>"en", "lat"=>40.726199, "lon"=>-124.188560),
"/US/US-CA/huntington-beach-1816" => array("type"=>"city", "pop"=>192620, "name_ru"=>"Хантингтон Бич", "name_en"=>"Huntington Beach", "lang"=>"en", "lat"=>33.691755, "lon"=>-118.008115),
"/US/US-CA/huntington-park-1817" => array("type"=>"city", "pop"=>60898, "name_ru"=>"Хантингтон Парк", "name_en"=>"Huntington Park", "lang"=>"en", "lat"=>33.980300, "lon"=>-118.215149),
"/US/US-CA/huron-1818" => array("type"=>"city", "pop"=>7606, "name_ru"=>"Гурон", "name_en"=>"Huron", "lang"=>"en", "lat"=>36.203342, "lon"=>-120.095367),
"/US/US-CA/hydesville-1819" => array("type"=>"cdp", "pop"=>1131, "name_ru"=>"Хайдсвилл", "name_en"=>"Hydesville", "lang"=>"en", "lat"=>40.548669, "lon"=>-124.087099),
"/US/US-CA/idyllwild-pine-cove-1820" => array("type"=>"cdp", "pop"=>2853, "name_ru"=>"Айдилуайлд-Пайн Ков", "name_en"=>"Idyllwild-Pine Cove", "lang"=>"en", "lat"=>33.747262, "lon"=>-116.723161),
"/US/US-CA/imperial-1821" => array("type"=>"city", "pop"=>13648, "name_ru"=>"Империал", "name_en"=>"Imperial", "lang"=>"en", "lat"=>32.840509, "lon"=>-115.571025),
"/US/US-CA/imperial-beach-1822" => array("type"=>"city", "pop"=>26543, "name_ru"=>"Империал Бич", "name_en"=>"Imperial Beach", "lang"=>"en", "lat"=>32.570200, "lon"=>-117.118599),
"/US/US-CA/indian-wells-1823" => array("type"=>"city", "pop"=>5177, "name_ru"=>"Индиан Уэлс", "name_en"=>"Indian Wells", "lang"=>"en", "lat"=>33.705396, "lon"=>-116.340565),
"/US/US-CA/indio-1824" => array("type"=>"city", "pop"=>84443, "name_ru"=>"Индио", "name_en"=>"Indio", "lang"=>"en", "lat"=>33.719800, "lon"=>-116.228649),
"/US/US-CA/industry-1825" => array("type"=>"city", "pop"=>913, "name_ru"=>"Индастри", "name_en"=>"Industry", "lang"=>"en", "lat"=>34.013400, "lon"=>-117.921535),
"/US/US-CA/inglewood-1826" => array("type"=>"city", "pop"=>112714, "name_ru"=>"Инглвуд", "name_en"=>"Inglewood", "lang"=>"en", "lat"=>33.955850, "lon"=>-118.343649),
"/US/US-CA/interlaken-1827" => array("type"=>"cdp", "pop"=>6404, "name_ru"=>"Интерлакен", "name_en"=>"Interlaken", "lang"=>"en", "lat"=>36.951378, "lon"=>-121.732857),
"/US/US-CA/inverness-1828" => array("type"=>"cdp", "pop"=>1422, "name_ru"=>"Инвернесс", "name_en"=>"Inverness", "lang"=>"en", "lat"=>38.085070, "lon"=>-122.841089),
"/US/US-CA/ione-1829" => array("type"=>"city", "pop"=>7480, "name_ru"=>"Ен", "name_en"=>"Ione", "lang"=>"en", "lat"=>38.359575, "lon"=>-120.941374),
"/US/US-CA/irvine-1830" => array("type"=>"city", "pop"=>207500, "name_ru"=>"Ирвайн", "name_en"=>"Irvine", "lang"=>"en", "lat"=>33.662114, "lon"=>-117.795807),
"/US/US-CA/irwindale-1831" => array("type"=>"city", "pop"=>1439, "name_ru"=>"Ирвиндейл", "name_en"=>"Irwindale", "lang"=>"en", "lat"=>34.111933, "lon"=>-117.967557),
"/US/US-CA/isla-vista-1832" => array("type"=>"cdp", "pop"=>20395, "name_ru"=>"Айла Виста", "name_en"=>"Isla Vista", "lang"=>"en", "lat"=>34.416100, "lon"=>-119.856749),
"/US/US-CA/isleton-1833" => array("type"=>"city", "pop"=>838, "name_ru"=>"Айлтон", "name_en"=>"Isleton", "lang"=>"en", "lat"=>38.161394, "lon"=>-121.604257),
"/US/US-CA/ivanhoe-1834" => array("type"=>"cdp", "pop"=>3293, "name_ru"=>"Айванхо", "name_en"=>"Ivanhoe", "lang"=>"en", "lat"=>36.389309, "lon"=>-119.218685),
"/US/US-CA/jackson-1835" => array("type"=>"city", "pop"=>4320, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>38.350500, "lon"=>-120.773520),
"/US/US-CA/jamestown-1836" => array("type"=>"cdp", "pop"=>2178, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>37.957312, "lon"=>-120.413861),
"/US/US-CA/jamul-1837" => array("type"=>"cdp", "pop"=>2258, "name_ru"=>"Джеймул", "name_en"=>"Jamul", "lang"=>"en", "lat"=>32.720238, "lon"=>-116.888775),
"/US/US-CA/joshua-tree-1838" => array("type"=>"cdp", "pop"=>3898, "name_ru"=>"Джошуа Три", "name_en"=>"Joshua Tree", "lang"=>"en", "lat"=>34.126950, "lon"=>-116.317756),
"/US/US-CA/julian-1839" => array("type"=>"cdp", "pop"=>1284, "name_ru"=>"Джулиан", "name_en"=>"Julian", "lang"=>"en", "lat"=>33.073116, "lon"=>-116.587389),
"/US/US-CA/kelseyville-1840" => array("type"=>"cdp", "pop"=>2861, "name_ru"=>"Келсивилл", "name_en"=>"Kelseyville", "lang"=>"en", "lat"=>38.958092, "lon"=>-122.824467),
"/US/US-CA/kensington-1841" => array("type"=>"cdp", "pop"=>4974, "name_ru"=>"Кенсингтон", "name_en"=>"Kensington", "lang"=>"en", "lat"=>37.908233, "lon"=>-122.277319),
"/US/US-CA/kentfield-1842" => array("type"=>"cdp", "pop"=>6030, "name_ru"=>"Кентфилд", "name_en"=>"Kentfield", "lang"=>"en", "lat"=>37.950317, "lon"=>-122.547408),
"/US/US-CA/kerman-1843" => array("type"=>"city", "pop"=>12737, "name_ru"=>"Керман", "name_en"=>"Kerman", "lang"=>"en", "lat"=>36.724169, "lon"=>-120.061033),
"/US/US-CA/kernville-1844" => array("type"=>"cdp", "pop"=>1656, "name_ru"=>"Кернвилл", "name_en"=>"Kernville", "lang"=>"en", "lat"=>35.761605, "lon"=>-118.430066),
"/US/US-CA/kettleman-city-1845" => array("type"=>"cdp", "pop"=>1411, "name_ru"=>"Кетлман Сити", "name_en"=>"Kettleman City", "lang"=>"en", "lat"=>36.008697, "lon"=>-119.961919),
"/US/US-CA/keyes-1846" => array("type"=>"cdp", "pop"=>2878, "name_ru"=>"Кейес", "name_en"=>"Keyes", "lang"=>"en", "lat"=>37.564183, "lon"=>-120.917407),
"/US/US-CA/king-city-1847" => array("type"=>"city", "pop"=>11627, "name_ru"=>"Кинг Сити", "name_en"=>"King City", "lang"=>"en", "lat"=>36.212980, "lon"=>-121.128224),
"/US/US-CA/kings-beach-1848" => array("type"=>"cdp", "pop"=>2796, "name_ru"=>"Кингс Бич", "name_en"=>"Kings Beach", "lang"=>"en", "lat"=>39.248111, "lon"=>-120.017701),
"/US/US-CA/kingsburg-1849" => array("type"=>"city", "pop"=>11064, "name_ru"=>"Кингсбург", "name_en"=>"Kingsburg", "lang"=>"en", "lat"=>36.519500, "lon"=>-119.554794),
"/US/US-CA/klamath-1850" => array("type"=>"cdp", "pop"=>827, "name_ru"=>"Кламат", "name_en"=>"Klamath", "lang"=>"en", "lat"=>41.522520, "lon"=>-123.996561),
"/US/US-CA/la-canada-flintridge-1851" => array("type"=>"city", "pop"=>19378, "name_ru"=>"Ла Канада Флинтридж", "name_en"=>"La Canada Flintridge", "lang"=>"en", "lat"=>34.210550, "lon"=>-118.200441),
"/US/US-CA/la-crescenta-montrose-1852" => array("type"=>"cdp", "pop"=>16968, "name_ru"=>"Ла Кресента-Монтроз", "name_en"=>"La Crescenta-Montrose", "lang"=>"en", "lat"=>34.233850, "lon"=>-118.235207),
"/US/US-CA/la-habra-1860" => array("type"=>"city", "pop"=>59155, "name_ru"=>"Ла Хабра", "name_en"=>"La Habra", "lang"=>"en", "lat"=>33.927400, "lon"=>-117.950599),
"/US/US-CA/la-habra-heights-1861" => array("type"=>"city", "pop"=>5893, "name_ru"=>"Ла Хабра Хайтс", "name_en"=>"La Habra Heights", "lang"=>"en", "lat"=>33.962033, "lon"=>-117.950291),
"/US/US-CA/la-mesa-1874" => array("type"=>"city", "pop"=>54673, "name_ru"=>"Ла Меса", "name_en"=>"La Mesa", "lang"=>"en", "lat"=>32.769600, "lon"=>-117.018870),
"/US/US-CA/la-mirada-1875" => array("type"=>"city", "pop"=>49809, "name_ru"=>"Ла Мирада", "name_en"=>"La Mirada", "lang"=>"en", "lat"=>33.902117, "lon"=>-118.008766),
"/US/US-CA/la-palma-1878" => array("type"=>"city", "pop"=>15603, "name_ru"=>"Ла Палма", "name_en"=>"La Palma", "lang"=>"en", "lat"=>33.850850, "lon"=>-118.038699),
"/US/US-CA/la-puente-1879" => array("type"=>"city", "pop"=>40642, "name_ru"=>"Ла Пуэнт", "name_en"=>"La Puente", "lang"=>"en", "lat"=>34.032500, "lon"=>-117.952250),
"/US/US-CA/la-quinta-1880" => array("type"=>"city", "pop"=>43865, "name_ru"=>"Ла Квинта", "name_en"=>"La Quinta", "lang"=>"en", "lat"=>33.671360, "lon"=>-116.292241),
"/US/US-CA/la-riviera-1881" => array("type"=>"cdp", "pop"=>10986, "name_ru"=>"Ла Ривьера", "name_en"=>"La Riviera", "lang"=>"en", "lat"=>38.566867, "lon"=>-121.355710),
"/US/US-CA/la-verne-1887" => array("type"=>"city", "pop"=>33623, "name_ru"=>"Ла Верн", "name_en"=>"La Verne", "lang"=>"en", "lat"=>34.118233, "lon"=>-117.770949),
"/US/US-CA/ladera-heights-1853" => array("type"=>"cdp", "pop"=>6316, "name_ru"=>"Ладера Хайтс", "name_en"=>"Ladera Heights", "lang"=>"en", "lat"=>33.994050, "lon"=>-118.374499),
"/US/US-CA/lafayette-1854" => array("type"=>"city", "pop"=>25011, "name_ru"=>"Ла Файет", "name_en"=>"Lafayette", "lang"=>"en", "lat"=>37.893833, "lon"=>-122.117915),
"/US/US-CA/laguna-1855" => array("type"=>"cdp", "pop"=>9828, "name_ru"=>"Лагуна", "name_en"=>"Laguna", "lang"=>"en", "lat"=>38.421129, "lon"=>-121.422661),
"/US/US-CA/laguna-beach-1856" => array("type"=>"city", "pop"=>23995, "name_ru"=>"Лагуна Бич", "name_en"=>"Laguna Beach", "lang"=>"en", "lat"=>33.539900, "lon"=>-117.760721),
"/US/US-CA/laguna-hills-1857" => array("type"=>"cdp", "pop"=>31838, "name_ru"=>"Лагуна Хилс", "name_en"=>"Laguna Hills", "lang"=>"en", "lat"=>33.599767, "lon"=>-117.710878),
"/US/US-CA/laguna-niguel-1858" => array("type"=>"city", "pop"=>64469, "name_ru"=>"Лагуна Нигел", "name_en"=>"Laguna Niguel", "lang"=>"en", "lat"=>33.526533, "lon"=>-117.704318),
"/US/US-CA/lagunitas-forest-knolls-1859" => array("type"=>"cdp", "pop"=>1821, "name_ru"=>"Лагунитас-Форест Ноллс", "name_en"=>"Lagunitas-Forest Knolls", "lang"=>"en", "lat"=>38.015233, "lon"=>-122.687599),
"/US/US-CA/lake-arrowhead-1862" => array("type"=>"cdp", "pop"=>6539, "name_ru"=>"Лейк Арроухед", "name_en"=>"Lake Arrowhead", "lang"=>"en", "lat"=>34.262276, "lon"=>-117.188543),
"/US/US-CA/lake-elsinore-1863" => array("type"=>"city", "pop"=>50952, "name_ru"=>"Лейк Элзинор", "name_en"=>"Lake Elsinore", "lang"=>"en", "lat"=>33.668985, "lon"=>-117.321481),
"/US/US-CA/lake-isabella-1864" => array("type"=>"cdp", "pop"=>3323, "name_ru"=>"Лейк Изабелла", "name_en"=>"Lake Isabella", "lang"=>"en", "lat"=>35.615154, "lon"=>-118.465720),
"/US/US-CA/lake-los-angeles-1866" => array("type"=>"cdp", "pop"=>7977, "name_ru"=>"Лейк Лос Анджелес", "name_en"=>"Lake Los Angeles", "lang"=>"en", "lat"=>34.611896, "lon"=>-117.826891),
"/US/US-CA/lake-nacimiento-1867" => array("type"=>"cdp", "pop"=>1556, "name_ru"=>"Лейк Насимиенто", "name_en"=>"Lake Nacimiento", "lang"=>"en", "lat"=>35.728460, "lon"=>-120.878545),
"/US/US-CA/lake-of-the-pines-1868" => array("type"=>"cdp", "pop"=>3890, "name_ru"=>"Лейк оф зе Пайнс", "name_en"=>"Lake Of The Pines", "lang"=>"en", "lat"=>39.038675, "lon"=>-121.060297),
"/US/US-CA/lake-san-marcos-1870" => array("type"=>"cdp", "pop"=>3802, "name_ru"=>"Лейк Сан Маркос", "name_en"=>"Lake San Marcos", "lang"=>"en", "lat"=>33.119057, "lon"=>-117.209133),
"/US/US-CA/lakeland-village-1865" => array("type"=>"cdp", "pop"=>5159, "name_ru"=>"Лейкленд Виллидж", "name_en"=>"Lakeland Village", "lang"=>"en", "lat"=>33.638291, "lon"=>-117.344827),
"/US/US-CA/lakeport-1869" => array("type"=>"city", "pop"=>5142, "name_ru"=>"Лейкпорт", "name_en"=>"Lakeport", "lang"=>"en", "lat"=>39.042281, "lon"=>-122.919875),
"/US/US-CA/lakeside-1871" => array("type"=>"cdp", "pop"=>39412, "name_ru"=>"Лэйксайд", "name_en"=>"Lakeside", "lang"=>"en", "lat"=>32.849043, "lon"=>-116.909924),
"/US/US-CA/lakeview-1872" => array("type"=>"cdp", "pop"=>1448, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>33.828088, "lon"=>-117.123818),
"/US/US-CA/lakewood-1873" => array("type"=>"city", "pop"=>78444, "name_ru"=>"Лейквуд", "name_en"=>"Lakewood", "lang"=>"en", "lat"=>33.848800, "lon"=>-118.122049),
"/US/US-CA/lamont-1876" => array("type"=>"cdp", "pop"=>11517, "name_ru"=>"Ламонт", "name_en"=>"Lamont", "lang"=>"en", "lat"=>35.264088, "lon"=>-118.914015),
"/US/US-CA/lancaster-1877" => array("type"=>"city", "pop"=>145469, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>34.693064, "lon"=>-118.176425),
"/US/US-CA/larkfield-wikiup-1882" => array("type"=>"cdp", "pop"=>6779, "name_ru"=>"Ларкфилд-Уайкиап", "name_en"=>"Larkfield-Wikiup", "lang"=>"en", "lat"=>38.513386, "lon"=>-122.752478),
"/US/US-CA/larkspur-1883" => array("type"=>"city", "pop"=>11693, "name_ru"=>"Ларкспер", "name_en"=>"Larkspur", "lang"=>"en", "lat"=>37.941150, "lon"=>-122.529154),
"/US/US-CA/las-lomas-1884" => array("type"=>"cdp", "pop"=>2127, "name_ru"=>"Лас Ломас", "name_en"=>"Las Lomas", "lang"=>"en", "lat"=>36.868800, "lon"=>-121.730980),
"/US/US-CA/lathrop-1885" => array("type"=>"city", "pop"=>17197, "name_ru"=>"Латроп", "name_en"=>"Lathrop", "lang"=>"en", "lat"=>37.820900, "lon"=>-121.276624),
"/US/US-CA/laton-1886" => array("type"=>"cdp", "pop"=>1415, "name_ru"=>"Латон", "name_en"=>"Laton", "lang"=>"en", "lat"=>36.433950, "lon"=>-119.688880),
"/US/US-CA/lawndale-1888" => array("type"=>"city", "pop"=>31346, "name_ru"=>"Лондейл", "name_en"=>"Lawndale", "lang"=>"en", "lat"=>33.888350, "lon"=>-118.352649),
"/US/US-CA/laytonville-1889" => array("type"=>"cdp", "pop"=>1133, "name_ru"=>"Лейтонвилл", "name_en"=>"Laytonville", "lang"=>"en", "lat"=>39.674675, "lon"=>-123.495218),
"/US/US-CA/le-grand-1890" => array("type"=>"cdp", "pop"=>1205, "name_ru"=>"Ле Гранд", "name_en"=>"Le Grand", "lang"=>"en", "lat"=>37.219121, "lon"=>-120.260538),
"/US/US-CA/lemon-grove-1891" => array("type"=>"city", "pop"=>24089, "name_ru"=>"Лемон Гров", "name_en"=>"Lemon Grove", "lang"=>"en", "lat"=>32.733400, "lon"=>-117.032841),
"/US/US-CA/lemoore-1892" => array("type"=>"city", "pop"=>23873, "name_ru"=>"Лемур", "name_en"=>"Lemoore", "lang"=>"en", "lat"=>36.298625, "lon"=>-119.787481),
"/US/US-CA/lemoore-station-1893" => array("type"=>"cdp", "pop"=>0, "name_ru"=>"Лемур Стейшен", "name_en"=>"Lemoore Station", "lang"=>"en", "lat"=>36.264410, "lon"=>-119.903740),
"/US/US-CA/lennox-1894" => array("type"=>"cdp", "pop"=>22757, "name_ru"=>"Леннокс", "name_en"=>"Lennox", "lang"=>"en", "lat"=>33.939800, "lon"=>-118.357651),
"/US/US-CA/lenwood-1895" => array("type"=>"cdp", "pop"=>3190, "name_ru"=>"Ленвуд", "name_en"=>"Lenwood", "lang"=>"en", "lat"=>34.885777, "lon"=>-117.103606),
"/US/US-CA/lewiston-1896" => array("type"=>"cdp", "pop"=>1187, "name_ru"=>"Льюистон", "name_en"=>"Lewiston", "lang"=>"en", "lat"=>40.699233, "lon"=>-122.802019),
"/US/US-CA/lexington-hills-1897" => array("type"=>"cdp", "pop"=>2064, "name_ru"=>"Лексингтон Хилс", "name_en"=>"Lexington Hills", "lang"=>"en", "lat"=>37.164833, "lon"=>-121.971842),
"/US/US-CA/lincoln-1898" => array("type"=>"city", "pop"=>43602, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>38.899170, "lon"=>-121.316491),
"/US/US-CA/lincoln-village-1899" => array("type"=>"cdp", "pop"=>4236, "name_ru"=>"Линкольн Виллидж", "name_en"=>"Lincoln Village", "lang"=>"en", "lat"=>38.004200, "lon"=>-121.333485),
"/US/US-CA/linda-1900" => array("type"=>"cdp", "pop"=>13033, "name_ru"=>"Линда", "name_en"=>"Linda", "lang"=>"en", "lat"=>39.122847, "lon"=>-121.554720),
"/US/US-CA/linden-1901" => array("type"=>"cdp", "pop"=>1339, "name_ru"=>"Линден", "name_en"=>"Linden", "lang"=>"en", "lat"=>38.018867, "lon"=>-121.099938),
"/US/US-CA/lindsay-1902" => array("type"=>"city", "pop"=>10571, "name_ru"=>"Линдсей", "name_en"=>"Lindsay", "lang"=>"en", "lat"=>36.206261, "lon"=>-119.088848),
"/US/US-CA/littlerock-1903" => array("type"=>"cdp", "pop"=>1320, "name_ru"=>"Литлрок", "name_en"=>"Littlerock", "lang"=>"en", "lat"=>34.525080, "lon"=>-117.984751),
"/US/US-CA/live-oak-1904" => array("type"=>"city", "pop"=>8245, "name_ru"=>"Лайв Оук", "name_en"=>"Live Oak", "lang"=>"en", "lat"=>39.273598, "lon"=>-121.661287),
"/US/US-CA/live-oak-1905" => array("type"=>"cdp", "pop"=>15212, "name_ru"=>"Лайв Оук", "name_en"=>"Live Oak", "lang"=>"en", "lat"=>36.983700, "lon"=>-121.979536),
"/US/US-CA/livermore-1906" => array("type"=>"city", "pop"=>80188, "name_ru"=>"Ливермор", "name_en"=>"Livermore", "lang"=>"en", "lat"=>37.691926, "lon"=>-121.763663),
"/US/US-CA/livingston-1907" => array("type"=>"city", "pop"=>13400, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>37.388026, "lon"=>-120.721484),
"/US/US-CA/lockeford-1908" => array("type"=>"cdp", "pop"=>2722, "name_ru"=>"Локфорд", "name_en"=>"Lockeford", "lang"=>"en", "lat"=>38.151554, "lon"=>-121.155540),
"/US/US-CA/lodi-1909" => array("type"=>"city", "pop"=>61301, "name_ru"=>"Лоди", "name_en"=>"Lodi", "lang"=>"en", "lat"=>38.123250, "lon"=>-121.294424),
"/US/US-CA/loma-linda-1910" => array("type"=>"city", "pop"=>21601, "name_ru"=>"Лома Линда", "name_en"=>"Loma Linda", "lang"=>"en", "lat"=>34.044150, "lon"=>-117.249034),
"/US/US-CA/loma-rica-1911" => array("type"=>"cdp", "pop"=>1852, "name_ru"=>"Лома Рика", "name_en"=>"Loma Rica", "lang"=>"en", "lat"=>39.318711, "lon"=>-121.400357),
"/US/US-CA/lomita-1912" => array("type"=>"city", "pop"=>20156, "name_ru"=>"Ломита", "name_en"=>"Lomita", "lang"=>"en", "lat"=>33.793400, "lon"=>-118.317249),
"/US/US-CA/lompoc-1913" => array("type"=>"city", "pop"=>41099, "name_ru"=>"Ломпок", "name_en"=>"Lompoc", "lang"=>"en", "lat"=>34.660557, "lon"=>-120.469639),
"/US/US-CA/london-1914" => array("type"=>"cdp", "pop"=>1638, "name_ru"=>"Лондон", "name_en"=>"London", "lang"=>"en", "lat"=>36.480954, "lon"=>-119.443015),
"/US/US-CA/lone-pine-1915" => array("type"=>"cdp", "pop"=>1818, "name_ru"=>"Лоун Пайн", "name_en"=>"Lone Pine", "lang"=>"en", "lat"=>36.574949, "lon"=>-118.083910),
"/US/US-CA/long-beach-1916" => array("type"=>"city", "pop"=>463789, "name_ru"=>"Лонг Бич", "name_en"=>"Long Beach", "lang"=>"en", "lat"=>33.788900, "lon"=>-118.159824),
"/US/US-CA/loomis-1917" => array("type"=>"town", "pop"=>6795, "name_ru"=>"Лумис", "name_en"=>"Loomis", "lang"=>"en", "lat"=>38.809688, "lon"=>-121.194345),
"/US/US-CA/los-alamitos-1918" => array("type"=>"city", "pop"=>11656, "name_ru"=>"Лос Аламитос", "name_en"=>"Los Alamitos", "lang"=>"en", "lat"=>33.798567, "lon"=>-118.058087),
"/US/US-CA/los-altos-1919" => array("type"=>"city", "pop"=>28349, "name_ru"=>"Лос Алтос", "name_en"=>"Los Altos", "lang"=>"en", "lat"=>37.369167, "lon"=>-122.095482),
"/US/US-CA/los-altos-hills-1920" => array("type"=>"town", "pop"=>8592, "name_ru"=>"Лос Алтос Хилс", "name_en"=>"Los Altos Hills", "lang"=>"en", "lat"=>37.367950, "lon"=>-122.139327),
"/US/US-CA/los-angeles-1921" => array("type"=>"city", "pop"=>3833995, "name_ru"=>"Лос Анджелес", "name_en"=>"Los Angeles", "lang"=>"en", "lat"=>34.112101, "lon"=>-118.411201),
"/US/US-CA/los-banos-1922" => array("type"=>"city", "pop"=>34968, "name_ru"=>"Лос Банос", "name_en"=>"Los Banos", "lang"=>"en", "lat"=>37.064177, "lon"=>-120.835295),
"/US/US-CA/los-gatos-1923" => array("type"=>"town", "pop"=>29320, "name_ru"=>"Лос Гатос", "name_en"=>"Los Gatos", "lang"=>"en", "lat"=>37.230695, "lon"=>-121.959772),
"/US/US-CA/los-molinos-1924" => array("type"=>"cdp", "pop"=>1709, "name_ru"=>"Лос Молинос", "name_en"=>"Los Molinos", "lang"=>"en", "lat"=>40.027640, "lon"=>-122.097152),
"/US/US-CA/los-serranos-1925" => array("type"=>"cdp", "pop"=>7099, "name_ru"=>"Лос Серранос", "name_en"=>"Los Serranos", "lang"=>"en", "lat"=>33.973800, "lon"=>-117.703099),
"/US/US-CA/lost-hills-1926" => array("type"=>"cdp", "pop"=>1212, "name_ru"=>"Лост Хилс", "name_en"=>"Lost Hills", "lang"=>"en", "lat"=>35.623357, "lon"=>-119.693506),
"/US/US-CA/lower-lake-1927" => array("type"=>"cdp", "pop"=>1217, "name_ru"=>"Лоуер Лейк", "name_en"=>"Lower Lake", "lang"=>"en", "lat"=>38.911202, "lon"=>-122.610256),
"/US/US-CA/loyalton-1928" => array("type"=>"city", "pop"=>777, "name_ru"=>"Лойолтон", "name_en"=>"Loyalton", "lang"=>"en", "lat"=>39.676642, "lon"=>-120.243352),
"/US/US-CA/loyola-1929" => array("type"=>"cdp", "pop"=>3076, "name_ru"=>"Лойола", "name_en"=>"Loyola", "lang"=>"en", "lat"=>37.351450, "lon"=>-122.099453),
"/US/US-CA/lucas-valley-marinwood-1930" => array("type"=>"cdp", "pop"=>5982, "name_ru"=>"Лукас Валли-Маринвуд", "name_en"=>"Lucas Valley-Marinwood", "lang"=>"en", "lat"=>38.035200, "lon"=>-122.565489),
"/US/US-CA/lucerne-1931" => array("type"=>"cdp", "pop"=>2011, "name_ru"=>"Лукерне", "name_en"=>"Lucerne", "lang"=>"en", "lat"=>39.094232, "lon"=>-122.809886),
"/US/US-CA/lynwood-1932" => array("type"=>"city", "pop"=>70025, "name_ru"=>"Линвуд", "name_en"=>"Lynwood", "lang"=>"en", "lat"=>33.923700, "lon"=>-118.201393),
"/US/US-CA/madera-1936" => array("type"=>"city", "pop"=>56700, "name_ru"=>"Мадера", "name_en"=>"Madera", "lang"=>"en", "lat"=>36.967579, "lon"=>-120.076718),
"/US/US-CA/madera-acres-1937" => array("type"=>"cdp", "pop"=>5245, "name_ru"=>"Мадера Эйкерс", "name_en"=>"Madera Acres", "lang"=>"en", "lat"=>37.019221, "lon"=>-120.065872),
"/US/US-CA/magalia-1938" => array("type"=>"cdp", "pop"=>8987, "name_ru"=>"Магалия", "name_en"=>"Magalia", "lang"=>"en", "lat"=>39.827519, "lon"=>-121.610800),
"/US/US-CA/mammoth-lakes-1939" => array("type"=>"town", "pop"=>7402, "name_ru"=>"Маммот Лейкс", "name_en"=>"Mammoth Lakes", "lang"=>"en", "lat"=>37.627817, "lon"=>-118.990181),
"/US/US-CA/manhattan-beach-1940" => array("type"=>"city", "pop"=>36605, "name_ru"=>"Манхэттен Бич", "name_en"=>"Manhattan Beach", "lang"=>"en", "lat"=>33.895093, "lon"=>-118.421281),
"/US/US-CA/manteca-1941" => array("type"=>"city", "pop"=>65028, "name_ru"=>"Мантека", "name_en"=>"Manteca", "lang"=>"en", "lat"=>37.802328, "lon"=>-121.223474),
"/US/US-CA/march-afb-1942" => array("type"=>"cdp", "pop"=>5523, "name_ru"=>"Военно-воздушная база Марч", "name_en"=>"March AFB", "lang"=>"en", "lat"=>33.891333, "lon"=>-117.263874),
"/US/US-CA/maricopa-1943" => array("type"=>"city", "pop"=>1133, "name_ru"=>"Марикопа", "name_en"=>"Maricopa", "lang"=>"en", "lat"=>35.059689, "lon"=>-119.401433),
"/US/US-CA/marina-1944" => array("type"=>"city", "pop"=>17943, "name_ru"=>"Марина", "name_en"=>"Marina", "lang"=>"en", "lat"=>36.682750, "lon"=>-121.790833),
"/US/US-CA/marina-del-rey-1945" => array("type"=>"cdp", "pop"=>7431, "name_ru"=>"Марина Дель Рей", "name_en"=>"Marina del Rey", "lang"=>"en", "lat"=>33.980333, "lon"=>-118.451077),
"/US/US-CA/mariposa-1946" => array("type"=>"cdp", "pop"=>17976, "name_ru"=>"Марипоса", "name_en"=>"Mariposa", "lang"=>"en", "lat"=>37.488534, "lon"=>-119.966532),
"/US/US-CA/martinez-1947" => array("type"=>"city", "pop"=>35145, "name_ru"=>"Мартинез", "name_en"=>"Martinez", "lang"=>"en", "lat"=>38.000750, "lon"=>-122.114526),
"/US/US-CA/marysville-1948" => array("type"=>"city", "pop"=>11700, "name_ru"=>"Мэрисвилл", "name_en"=>"Marysville", "lang"=>"en", "lat"=>39.151624, "lon"=>-121.582255),
"/US/US-CA/mather-afb-1949" => array("type"=>"cdp", "pop"=>4885, "name_ru"=>"Военно-воздушная база Мазер", "name_en"=>"Mather AFB", "lang"=>"en", "lat"=>38.548433, "lon"=>-121.277199),
"/US/US-CA/mayflower-village-1950" => array("type"=>"cdp", "pop"=>4978, "name_ru"=>"Мейфлауэр Виллидж", "name_en"=>"Mayflower Village", "lang"=>"en", "lat"=>34.114950, "lon"=>-118.008949),
"/US/US-CA/maywood-1951" => array("type"=>"city", "pop"=>28224, "name_ru"=>"Мэйвуд", "name_en"=>"Maywood", "lang"=>"en", "lat"=>33.989100, "lon"=>-118.187726),
"/US/US-CA/mccloud-1933" => array("type"=>"cdp", "pop"=>1555, "name_ru"=>"Мак-Клауд", "name_en"=>"McCloud", "lang"=>"en", "lat"=>41.254804, "lon"=>-122.135206),
"/US/US-CA/mcfarland-1934" => array("type"=>"city", "pop"=>12098, "name_ru"=>"Мак-Фарленд", "name_en"=>"McFarland", "lang"=>"en", "lat"=>35.679071, "lon"=>-119.235579),
"/US/US-CA/mckinleyville-1935" => array("type"=>"cdp", "pop"=>10749, "name_ru"=>"Мак-Кинливилл", "name_en"=>"McKinleyville", "lang"=>"en", "lat"=>40.953314, "lon"=>-124.115608),
"/US/US-CA/meadow-vista-1952" => array("type"=>"cdp", "pop"=>3067, "name_ru"=>"Медоу Виста", "name_en"=>"Meadow Vista", "lang"=>"en", "lat"=>39.003967, "lon"=>-121.030723),
"/US/US-CA/mecca-1953" => array("type"=>"cdp", "pop"=>1966, "name_ru"=>"Мекка", "name_en"=>"Mecca", "lang"=>"en", "lat"=>33.575894, "lon"=>-116.071713),
"/US/US-CA/meiners-oaks-1954" => array("type"=>"cdp", "pop"=>3329, "name_ru"=>"Майнерс Оукс", "name_en"=>"Meiners Oaks", "lang"=>"en", "lat"=>34.450401, "lon"=>-119.272133),
"/US/US-CA/mendota-1955" => array("type"=>"city", "pop"=>10339, "name_ru"=>"Мендота", "name_en"=>"Mendota", "lang"=>"en", "lat"=>36.757599, "lon"=>-120.378905),
"/US/US-CA/menlo-park-1956" => array("type"=>"city", "pop"=>30087, "name_ru"=>"Менло Парк", "name_en"=>"Menlo Park", "lang"=>"en", "lat"=>37.480000, "lon"=>-122.144037),
"/US/US-CA/mentone-1957" => array("type"=>"cdp", "pop"=>5675, "name_ru"=>"Ментон", "name_en"=>"Mentone", "lang"=>"en", "lat"=>34.062950, "lon"=>-117.122774),
"/US/US-CA/merced-1958" => array("type"=>"city", "pop"=>77160, "name_ru"=>"Мерсед", "name_en"=>"Merced", "lang"=>"en", "lat"=>37.296274, "lon"=>-120.484357),
"/US/US-CA/mi-wuk-village-1966" => array("type"=>"cdp", "pop"=>1175, "name_ru"=>"Мивок Виллидж", "name_en"=>"Mi-Wuk Village", "lang"=>"en", "lat"=>38.057998, "lon"=>-120.176345),
"/US/US-CA/mill-valley-1960" => array("type"=>"city", "pop"=>13268, "name_ru"=>"Милл Валли", "name_en"=>"Mill Valley", "lang"=>"en", "lat"=>37.907967, "lon"=>-122.540980),
"/US/US-CA/millbrae-1959" => array("type"=>"city", "pop"=>20800, "name_ru"=>"Милбро", "name_en"=>"Millbrae", "lang"=>"en", "lat"=>37.599100, "lon"=>-122.400995),
"/US/US-CA/milpitas-1961" => array("type"=>"city", "pop"=>67503, "name_ru"=>"Милпитас", "name_en"=>"Milpitas", "lang"=>"en", "lat"=>37.434600, "lon"=>-121.891946),
"/US/US-CA/mira-loma-1962" => array("type"=>"cdp", "pop"=>15786, "name_ru"=>"Мира Лома", "name_en"=>"Mira Loma", "lang"=>"en", "lat"=>33.984600, "lon"=>-117.512733),
"/US/US-CA/mira-monte-1963" => array("type"=>"cdp", "pop"=>7744, "name_ru"=>"Мира Монте", "name_en"=>"Mira Monte", "lang"=>"en", "lat"=>34.427982, "lon"=>-119.282800),
"/US/US-CA/mission-hills-1964" => array("type"=>"cdp", "pop"=>3112, "name_ru"=>"Мишен Хилс", "name_en"=>"Mission Hills", "lang"=>"en", "lat"=>34.685999, "lon"=>-120.435840),
"/US/US-CA/mission-viejo-1965" => array("type"=>"city", "pop"=>94461, "name_ru"=>"Мишен Вьехо", "name_en"=>"Mission Viejo", "lang"=>"en", "lat"=>33.612417, "lon"=>-117.651626),
"/US/US-CA/modesto-1967" => array("type"=>"city", "pop"=>202967, "name_ru"=>"Модесто", "name_en"=>"Modesto", "lang"=>"en", "lat"=>37.659799, "lon"=>-120.993923),
"/US/US-CA/mojave-1968" => array("type"=>"cdp", "pop"=>3763, "name_ru"=>"Мохаве", "name_en"=>"Mojave", "lang"=>"en", "lat"=>35.052079, "lon"=>-118.179809),
"/US/US-CA/mono-vista-1969" => array("type"=>"cdp", "pop"=>2599, "name_ru"=>"Моно Виста", "name_en"=>"Mono Vista", "lang"=>"en", "lat"=>38.011478, "lon"=>-120.269652),
"/US/US-CA/monrovia-1970" => array("type"=>"city", "pop"=>37651, "name_ru"=>"Монровиа", "name_en"=>"Monrovia", "lang"=>"en", "lat"=>34.163385, "lon"=>-117.990416),
"/US/US-CA/montague-1971" => array("type"=>"city", "pop"=>1454, "name_ru"=>"Монтагю", "name_en"=>"Montague", "lang"=>"en", "lat"=>41.727457, "lon"=>-122.529580),
"/US/US-CA/montara-1972" => array("type"=>"cdp", "pop"=>2552, "name_ru"=>"Монтара", "name_en"=>"Montara", "lang"=>"en", "lat"=>37.545400, "lon"=>-122.497194),
"/US/US-CA/montclair-1973" => array("type"=>"city", "pop"=>36530, "name_ru"=>"Монтклэр", "name_en"=>"Montclair", "lang"=>"en", "lat"=>34.072784, "lon"=>-117.696560),
"/US/US-CA/monte-rio-1977" => array("type"=>"cdp", "pop"=>1058, "name_ru"=>"Монте Рио", "name_en"=>"Monte Rio", "lang"=>"en", "lat"=>38.465267, "lon"=>-123.012333),
"/US/US-CA/monte-sereno-1978" => array("type"=>"city", "pop"=>3602, "name_ru"=>"Монте Серено", "name_en"=>"Monte Sereno", "lang"=>"en", "lat"=>37.238250, "lon"=>-121.988574),
"/US/US-CA/montebello-1974" => array("type"=>"city", "pop"=>61906, "name_ru"=>"Монтебелло", "name_en"=>"Montebello", "lang"=>"en", "lat"=>34.015500, "lon"=>-118.109999),
"/US/US-CA/monterey-1975" => array("type"=>"city", "pop"=>27763, "name_ru"=>"Монтерей", "name_en"=>"Monterey", "lang"=>"en", "lat"=>36.601350, "lon"=>-121.881760),
"/US/US-CA/monterey-park-1976" => array("type"=>"city", "pop"=>61234, "name_ru"=>"Мотерей Парк", "name_en"=>"Monterey Park", "lang"=>"en", "lat"=>34.046833, "lon"=>-118.130565),
"/US/US-CA/moorpark-1979" => array("type"=>"city", "pop"=>36372, "name_ru"=>"Мурпарк", "name_en"=>"Moorpark", "lang"=>"en", "lat"=>34.285088, "lon"=>-118.877564),
"/US/US-CA/morada-1980" => array("type"=>"cdp", "pop"=>3570, "name_ru"=>"Морада", "name_en"=>"Morada", "lang"=>"en", "lat"=>38.038730, "lon"=>-121.244722),
"/US/US-CA/moraga-town-1981" => array("type"=>"city", "pop"=>15852, "name_ru"=>"Морага Таун", "name_en"=>"Moraga Town", "lang"=>"en", "lat"=>37.843850, "lon"=>-122.123425),
"/US/US-CA/moreno-valley-1982" => array("type"=>"city", "pop"=>190871, "name_ru"=>"Морено Валли", "name_en"=>"Moreno Valley", "lang"=>"en", "lat"=>33.926230, "lon"=>-117.210311),
"/US/US-CA/morgan-hill-1983" => array("type"=>"city", "pop"=>38096, "name_ru"=>"Морган Хил", "name_en"=>"Morgan Hill", "lang"=>"en", "lat"=>37.130550, "lon"=>-121.639638),
"/US/US-CA/morongo-valley-1984" => array("type"=>"cdp", "pop"=>1544, "name_ru"=>"Моронго Валли", "name_en"=>"Morongo Valley", "lang"=>"en", "lat"=>34.063888, "lon"=>-116.591763),
"/US/US-CA/morro-bay-1985" => array("type"=>"city", "pop"=>10333, "name_ru"=>"Морро Бэй", "name_en"=>"Morro Bay", "lang"=>"en", "lat"=>35.367406, "lon"=>-120.866294),
"/US/US-CA/moss-beach-1986" => array("type"=>"cdp", "pop"=>3002, "name_ru"=>"Мосс Бич", "name_en"=>"Moss Beach", "lang"=>"en", "lat"=>37.521800, "lon"=>-122.505604),
"/US/US-CA/mount-shasta-1990" => array("type"=>"city", "pop"=>3531, "name_ru"=>"Маунт Шаста", "name_en"=>"Mount Shasta", "lang"=>"en", "lat"=>41.322500, "lon"=>-122.314716),
"/US/US-CA/mountain-mesa-1987" => array("type"=>"cdp", "pop"=>1153, "name_ru"=>"Маунтин Меса", "name_en"=>"Mountain Mesa", "lang"=>"en", "lat"=>35.627489, "lon"=>-118.406054),
"/US/US-CA/mountain-view-1988" => array("type"=>"city", "pop"=>71348, "name_ru"=>"Маунтин Вью", "name_en"=>"Mountain View", "lang"=>"en", "lat"=>37.402000, "lon"=>-122.077641),
"/US/US-CA/mountain-view-acres-1989" => array("type"=>"cdp", "pop"=>2469, "name_ru"=>"Маунтин Вью Эйкрс", "name_en"=>"Mountain View Acres", "lang"=>"en", "lat"=>34.496782, "lon"=>-117.347950),
"/US/US-CA/murphys-1991" => array("type"=>"cdp", "pop"=>1517, "name_ru"=>"Мерфис", "name_en"=>"Murphys", "lang"=>"en", "lat"=>38.151953, "lon"=>-120.460894),
"/US/US-CA/murrieta-1992" => array("type"=>"cdp", "pop"=>97918, "name_ru"=>"Мерриета", "name_en"=>"Murrieta", "lang"=>"en", "lat"=>33.559673, "lon"=>-117.209601),
"/US/US-CA/murrieta-hot-springs-1993" => array("type"=>"cdp", "pop"=>1938, "name_ru"=>"Мерриета Хот Спрингс", "name_en"=>"Murrieta Hot Springs", "lang"=>"en", "lat"=>33.565088, "lon"=>-117.150758),
"/US/US-CA/muscoy-1994" => array("type"=>"cdp", "pop"=>7541, "name_ru"=>"Маской", "name_en"=>"Muscoy", "lang"=>"en", "lat"=>34.152350, "lon"=>-117.345674),
"/US/US-CA/myrtletown-1995" => array("type"=>"cdp", "pop"=>4413, "name_ru"=>"Миртлтаун", "name_en"=>"Myrtletown", "lang"=>"en", "lat"=>40.788961, "lon"=>-124.129083),
"/US/US-CA/napa-1996" => array("type"=>"city", "pop"=>74547, "name_ru"=>"Напа", "name_en"=>"Napa", "lang"=>"en", "lat"=>38.297933, "lon"=>-122.299907),
"/US/US-CA/national-city-1997" => array("type"=>"city", "pop"=>58680, "name_ru"=>"Нешнел Сити", "name_en"=>"National City", "lang"=>"en", "lat"=>32.666050, "lon"=>-117.098502),
"/US/US-CA/nebo-center-1998" => array("type"=>"cdp", "pop"=>1459, "name_ru"=>"Небо Сити", "name_en"=>"Nebo Center", "lang"=>"en", "lat"=>34.871567, "lon"=>-116.955593),
"/US/US-CA/needles-1999" => array("type"=>"city", "pop"=>5246, "name_ru"=>"Нидлс", "name_en"=>"Needles", "lang"=>"en", "lat"=>34.814126, "lon"=>-114.613217),
"/US/US-CA/nevada-city-2000" => array("type"=>"city", "pop"=>2929, "name_ru"=>"Невада Сити", "name_en"=>"Nevada City", "lang"=>"en", "lat"=>39.260358, "lon"=>-121.021246),
"/US/US-CA/newark-2001" => array("type"=>"city", "pop"=>41781, "name_ru"=>"Ньюарк", "name_en"=>"Newark", "lang"=>"en", "lat"=>37.519267, "lon"=>-122.030504),
"/US/US-CA/newman-2002" => array("type"=>"city", "pop"=>10226, "name_ru"=>"Ньюман", "name_en"=>"Newman", "lang"=>"en", "lat"=>37.315101, "lon"=>-121.021437),
"/US/US-CA/newport-beach-2003" => array("type"=>"city", "pop"=>79661, "name_ru"=>"Ньюпорт Бич", "name_en"=>"Newport Beach", "lang"=>"en", "lat"=>33.614050, "lon"=>-117.909409),
"/US/US-CA/nice-2004" => array("type"=>"cdp", "pop"=>2126, "name_ru"=>"Найс", "name_en"=>"Nice", "lang"=>"en", "lat"=>39.125284, "lon"=>-122.850510),
"/US/US-CA/niland-2005" => array("type"=>"cdp", "pop"=>1183, "name_ru"=>"Ниленд", "name_en"=>"Niland", "lang"=>"en", "lat"=>33.238540, "lon"=>-115.513108),
"/US/US-CA/nipomo-2006" => array("type"=>"cdp", "pop"=>7109, "name_ru"=>"Нипомо", "name_en"=>"Nipomo", "lang"=>"en", "lat"=>35.032363, "lon"=>-120.482977),
"/US/US-CA/norco-2007" => array("type"=>"city", "pop"=>26659, "name_ru"=>"Норко", "name_en"=>"Norco", "lang"=>"en", "lat"=>33.925460, "lon"=>-117.551199),
"/US/US-CA/north-auburn-2008" => array("type"=>"cdp", "pop"=>10301, "name_ru"=>"Норт Оберн", "name_en"=>"North Auburn", "lang"=>"en", "lat"=>38.931272, "lon"=>-121.080710),
"/US/US-CA/north-edwards-2009" => array("type"=>"cdp", "pop"=>1259, "name_ru"=>"Норт Эдвардс", "name_en"=>"North Edwards", "lang"=>"en", "lat"=>35.019101, "lon"=>-117.830274),
"/US/US-CA/north-el-monte-2010" => array("type"=>"cdp", "pop"=>3384, "name_ru"=>"Норт Эль Монте", "name_en"=>"North El Monte", "lang"=>"en", "lat"=>34.102900, "lon"=>-118.023351),
"/US/US-CA/north-fair-oaks-2011" => array("type"=>"cdp", "pop"=>13912, "name_ru"=>"Норт Фэр Оукс", "name_en"=>"North Fair Oaks", "lang"=>"en", "lat"=>37.474500, "lon"=>-122.201774),
"/US/US-CA/north-highlands-2012" => array("type"=>"cdp", "pop"=>42105, "name_ru"=>"Норт Халендс", "name_en"=>"North Highlands", "lang"=>"en", "lat"=>38.667650, "lon"=>-121.380409),
"/US/US-CA/norwalk-2013" => array("type"=>"city", "pop"=>102982, "name_ru"=>"Норуолк", "name_en"=>"Norwalk", "lang"=>"en", "lat"=>33.906900, "lon"=>-118.082499),
"/US/US-CA/novato-2014" => array("type"=>"city", "pop"=>52785, "name_ru"=>"Новато", "name_en"=>"Novato", "lang"=>"en", "lat"=>38.092100, "lon"=>-122.556710),
"/US/US-CA/nuevo-2015" => array("type"=>"cdp", "pop"=>3010, "name_ru"=>"Нуево", "name_en"=>"Nuevo", "lang"=>"en", "lat"=>33.806915, "lon"=>-117.136994),
"/US/US-CA/oak-park-2020" => array("type"=>"cdp", "pop"=>2412, "name_ru"=>"Оук Парк", "name_en"=>"Oak Park", "lang"=>"en", "lat"=>34.171488, "lon"=>-118.755479),
"/US/US-CA/oak-view-2021" => array("type"=>"cdp", "pop"=>3606, "name_ru"=>"Оук Вью", "name_en"=>"Oak View", "lang"=>"en", "lat"=>34.399669, "lon"=>-119.295454),
"/US/US-CA/oakdale-2016" => array("type"=>"city", "pop"=>20299, "name_ru"=>"Оукдейл", "name_en"=>"Oakdale", "lang"=>"en", "lat"=>37.768201, "lon"=>-120.851850),
"/US/US-CA/oakhurst-2017" => array("type"=>"cdp", "pop"=>2602, "name_ru"=>"Оукхерст", "name_en"=>"Oakhurst", "lang"=>"en", "lat"=>37.333328, "lon"=>-119.647006),
"/US/US-CA/oakland-2018" => array("type"=>"city", "pop"=>404155, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>37.771544, "lon"=>-122.224550),
"/US/US-CA/oakley-2019" => array("type"=>"cdp", "pop"=>32035, "name_ru"=>"Оукли", "name_en"=>"Oakley", "lang"=>"en", "lat"=>37.986485, "lon"=>-121.716026),
"/US/US-CA/occidental-2022" => array("type"=>"cdp", "pop"=>1300, "name_ru"=>"Оксидентал", "name_en"=>"Occidental", "lang"=>"en", "lat"=>38.395402, "lon"=>-122.937871),
"/US/US-CA/oceano-2023" => array("type"=>"cdp", "pop"=>6169, "name_ru"=>"Ошено", "name_en"=>"Oceano", "lang"=>"en", "lat"=>35.101778, "lon"=>-120.608751),
"/US/US-CA/oceanside-2024" => array("type"=>"city", "pop"=>169684, "name_ru"=>"Ошенсайд", "name_en"=>"Oceanside", "lang"=>"en", "lat"=>33.225074, "lon"=>-117.310286),
"/US/US-CA/oildale-2025" => array("type"=>"cdp", "pop"=>26553, "name_ru"=>"Ойлдейл", "name_en"=>"Oildale", "lang"=>"en", "lat"=>35.423500, "lon"=>-119.027899),
"/US/US-CA/ojai-2026" => array("type"=>"city", "pop"=>7775, "name_ru"=>"Охай", "name_en"=>"Ojai", "lang"=>"en", "lat"=>34.449085, "lon"=>-119.245717),
"/US/US-CA/olivehurst-2027" => array("type"=>"cdp", "pop"=>9738, "name_ru"=>"Оливерхерст", "name_en"=>"Olivehurst", "lang"=>"en", "lat"=>39.084320, "lon"=>-121.547506),
"/US/US-CA/ontario-2028" => array("type"=>"city", "pop"=>171691, "name_ru"=>"Онтарио", "name_en"=>"Ontario", "lang"=>"en", "lat"=>34.054412, "lon"=>-117.605810),
"/US/US-CA/opal-cliffs-2029" => array("type"=>"cdp", "pop"=>5940, "name_ru"=>"Опал Клифс", "name_en"=>"Opal Cliffs", "lang"=>"en", "lat"=>36.955200, "lon"=>-121.974739),
"/US/US-CA/orange-2030" => array("type"=>"city", "pop"=>136392, "name_ru"=>"Оранж", "name_en"=>"Orange", "lang"=>"en", "lat"=>33.805250, "lon"=>-117.823999),
"/US/US-CA/orange-cove-2031" => array("type"=>"city", "pop"=>10641, "name_ru"=>"Оранж Ков", "name_en"=>"Orange Cove", "lang"=>"en", "lat"=>36.622450, "lon"=>-119.317449),
"/US/US-CA/orangevale-2032" => array("type"=>"cdp", "pop"=>26266, "name_ru"=>"Оранжвейл", "name_en"=>"Orangevale", "lang"=>"en", "lat"=>38.688258, "lon"=>-121.217331),
"/US/US-CA/orinda-2033" => array("type"=>"city", "pop"=>18445, "name_ru"=>"Оринда", "name_en"=>"Orinda", "lang"=>"en", "lat"=>37.880917, "lon"=>-122.177338),
"/US/US-CA/orland-2034" => array("type"=>"city", "pop"=>7174, "name_ru"=>"Орленд", "name_en"=>"Orland", "lang"=>"en", "lat"=>39.745725, "lon"=>-122.187226),
"/US/US-CA/orosi-2035" => array("type"=>"cdp", "pop"=>5486, "name_ru"=>"Ороси", "name_en"=>"Orosi", "lang"=>"en", "lat"=>36.542499, "lon"=>-119.287859),
"/US/US-CA/oroville-2036" => array("type"=>"city", "pop"=>14684, "name_ru"=>"Оровилл", "name_en"=>"Oroville", "lang"=>"en", "lat"=>39.501618, "lon"=>-121.565935),
"/US/US-CA/oroville-east-2037" => array("type"=>"cdp", "pop"=>8462, "name_ru"=>"Оровилл Ист", "name_en"=>"Oroville East", "lang"=>"en", "lat"=>39.513534, "lon"=>-121.477427),
"/US/US-CA/oxnard-2038" => array("type"=>"city", "pop"=>185717, "name_ru"=>"Окснард", "name_en"=>"Oxnard", "lang"=>"en", "lat"=>34.197426, "lon"=>-119.213695),
"/US/US-CA/pacheco-2039" => array("type"=>"cdp", "pop"=>3325, "name_ru"=>"Пачеко", "name_en"=>"Pacheco", "lang"=>"en", "lat"=>37.987567, "lon"=>-122.062303),
"/US/US-CA/pacific-grove-2041" => array("type"=>"city", "pop"=>14601, "name_ru"=>"Пасифик Гров", "name_en"=>"Pacific Grove", "lang"=>"en", "lat"=>36.622800, "lon"=>-121.925065),
"/US/US-CA/pacifica-2040" => array("type"=>"city", "pop"=>37739, "name_ru"=>"Пасифика", "name_en"=>"Pacifica", "lang"=>"en", "lat"=>37.611250, "lon"=>-122.476320),
"/US/US-CA/pajaro-2042" => array("type"=>"cdp", "pop"=>3332, "name_ru"=>"Паджаро", "name_en"=>"Pajaro", "lang"=>"en", "lat"=>36.900100, "lon"=>-121.740859),
"/US/US-CA/palermo-2043" => array("type"=>"cdp", "pop"=>5260, "name_ru"=>"Палермо", "name_en"=>"Palermo", "lang"=>"en", "lat"=>39.436664, "lon"=>-121.530540),
"/US/US-CA/palm-desert-2046" => array("type"=>"city", "pop"=>50876, "name_ru"=>"Палм Дезерт", "name_en"=>"Palm Desert", "lang"=>"en", "lat"=>33.733242, "lon"=>-116.375392),
"/US/US-CA/palm-desert-country-2047" => array("type"=>"cdp", "pop"=>5626, "name_ru"=>"Палм Дезерт Каунти", "name_en"=>"Palm Desert Country", "lang"=>"en", "lat"=>33.740577, "lon"=>-116.314654),
"/US/US-CA/palm-springs-2048" => array("type"=>"city", "pop"=>47952, "name_ru"=>"Палм Спрингс", "name_en"=>"Palm Springs", "lang"=>"en", "lat"=>33.775750, "lon"=>-116.528788),
"/US/US-CA/palmdale-2044" => array("type"=>"city", "pop"=>143197, "name_ru"=>"Палмдейл", "name_en"=>"Palmdale", "lang"=>"en", "lat"=>34.605384, "lon"=>-118.086271),
"/US/US-CA/palmdale-east-2045" => array("type"=>"cdp", "pop"=>3052, "name_ru"=>"Палмдейл Ист", "name_en"=>"Palmdale East", "lang"=>"en", "lat"=>34.581928, "lon"=>-118.071831),
"/US/US-CA/palo-alto-2049" => array("type"=>"city", "pop"=>59395, "name_ru"=>"Пало-Альто", "name_en"=>"Palo Alto", "lang"=>"en", "lat"=>37.399200, "lon"=>-122.139354),
"/US/US-CA/palos-verdes-estates-2050" => array("type"=>"city", "pop"=>13585, "name_ru"=>"Палос Вердес Эстейтс", "name_en"=>"Palos Verdes Estates", "lang"=>"en", "lat"=>33.787767, "lon"=>-118.396203),
"/US/US-CA/paradise-2051" => array("type"=>"town", "pop"=>26469, "name_ru"=>"Парадайс", "name_en"=>"Paradise", "lang"=>"en", "lat"=>39.756412, "lon"=>-121.604334),
"/US/US-CA/paramount-2052" => array("type"=>"city", "pop"=>55236, "name_ru"=>"Парамаунт", "name_en"=>"Paramount", "lang"=>"en", "lat"=>33.895550, "lon"=>-118.164074),
"/US/US-CA/parksdale-2053" => array("type"=>"cdp", "pop"=>1911, "name_ru"=>"Парксдейл", "name_en"=>"Parksdale", "lang"=>"en", "lat"=>36.947219, "lon"=>-120.021905),
"/US/US-CA/parkway-south-sacramento-2054" => array("type"=>"cdp", "pop"=>31903, "name_ru"=>"Паркуэй-Сауф Сакраменто", "name_en"=>"Parkway-South Sacramento", "lang"=>"en", "lat"=>38.508550, "lon"=>-121.451949),
"/US/US-CA/parkwood-2055" => array("type"=>"cdp", "pop"=>1659, "name_ru"=>"Парквуд", "name_en"=>"Parkwood", "lang"=>"en", "lat"=>36.926818, "lon"=>-120.043582),
"/US/US-CA/parlier-2056" => array("type"=>"city", "pop"=>13273, "name_ru"=>"Парльер", "name_en"=>"Parlier", "lang"=>"en", "lat"=>36.608500, "lon"=>-119.538184),
"/US/US-CA/pasadena-2057" => array("type"=>"city", "pop"=>143080, "name_ru"=>"Пасадена", "name_en"=>"Pasadena", "lang"=>"en", "lat"=>34.160700, "lon"=>-118.138724),
"/US/US-CA/patterson-2058" => array("type"=>"city", "pop"=>19034, "name_ru"=>"Паттерсон", "name_en"=>"Patterson", "lang"=>"en", "lat"=>37.474251, "lon"=>-121.127600),
"/US/US-CA/pedley-2059" => array("type"=>"cdp", "pop"=>8869, "name_ru"=>"Педли", "name_en"=>"Pedley", "lang"=>"en", "lat"=>33.976733, "lon"=>-117.470649),
"/US/US-CA/penn-valley-2060" => array("type"=>"cdp", "pop"=>1242, "name_ru"=>"Пенн Валли", "name_en"=>"Penn Valley", "lang"=>"en", "lat"=>39.196169, "lon"=>-121.189948),
"/US/US-CA/perris-2061" => array("type"=>"city", "pop"=>55643, "name_ru"=>"Перрис", "name_en"=>"Perris", "lang"=>"en", "lat"=>33.793015, "lon"=>-117.223058),
"/US/US-CA/petaluma-2062" => array("type"=>"city", "pop"=>54666, "name_ru"=>"Петалума", "name_en"=>"Petaluma", "lang"=>"en", "lat"=>38.241150, "lon"=>-122.625455),
"/US/US-CA/phoenix-lake-cedar-ridge-2063" => array("type"=>"cdp", "pop"=>3569, "name_ru"=>"Финикс Лейк-Сидар Ридж", "name_en"=>"Phoenix Lake-Cedar Ridge", "lang"=>"en", "lat"=>38.023159, "lon"=>-120.300745),
"/US/US-CA/pico-rivera-2064" => array("type"=>"city", "pop"=>63138, "name_ru"=>"Пико Ривера", "name_en"=>"Pico Rivera", "lang"=>"en", "lat"=>33.989800, "lon"=>-118.088276),
"/US/US-CA/piedmont-2065" => array("type"=>"city", "pop"=>10481, "name_ru"=>"Пидмонт", "name_en"=>"Piedmont", "lang"=>"en", "lat"=>37.822850, "lon"=>-122.228903),
"/US/US-CA/pine-hills-2066" => array("type"=>"cdp", "pop"=>2947, "name_ru"=>"Пайн Хиллс", "name_en"=>"Pine Hills", "lang"=>"en", "lat"=>40.733455, "lon"=>-124.151114),
"/US/US-CA/pine-valley-2067" => array("type"=>"cdp", "pop"=>1297, "name_ru"=>"Пайн Валли", "name_en"=>"Pine Valley", "lang"=>"en", "lat"=>32.855420, "lon"=>-116.501194),
"/US/US-CA/pinole-2068" => array("type"=>"city", "pop"=>18808, "name_ru"=>"Пинол", "name_en"=>"Pinole", "lang"=>"en", "lat"=>38.012133, "lon"=>-122.315474),
"/US/US-CA/piru-2069" => array("type"=>"cdp", "pop"=>1157, "name_ru"=>"Пиру", "name_en"=>"Piru", "lang"=>"en", "lat"=>34.407301, "lon"=>-118.798748),
"/US/US-CA/pismo-beach-2070" => array("type"=>"city", "pop"=>8573, "name_ru"=>"Писмо Бич", "name_en"=>"Pismo Beach", "lang"=>"en", "lat"=>35.135958, "lon"=>-120.677005),
"/US/US-CA/pittsburg-2071" => array("type"=>"city", "pop"=>64148, "name_ru"=>"Питсбург", "name_en"=>"Pittsburg", "lang"=>"en", "lat"=>38.018150, "lon"=>-121.889470),
"/US/US-CA/pixley-2072" => array("type"=>"cdp", "pop"=>2457, "name_ru"=>"Пиксли", "name_en"=>"Pixley", "lang"=>"en", "lat"=>35.972878, "lon"=>-119.287173),
"/US/US-CA/placentia-2073" => array("type"=>"city", "pop"=>49692, "name_ru"=>"Пласентия", "name_en"=>"Placentia", "lang"=>"en", "lat"=>33.881567, "lon"=>-117.854773),
"/US/US-CA/placerville-2074" => array("type"=>"city", "pop"=>9994, "name_ru"=>"Пласервилл", "name_en"=>"Placerville", "lang"=>"en", "lat"=>38.730232, "lon"=>-120.795661),
"/US/US-CA/planada-2075" => array("type"=>"cdp", "pop"=>3531, "name_ru"=>"Планада", "name_en"=>"Planada", "lang"=>"en", "lat"=>37.288100, "lon"=>-120.323277),
"/US/US-CA/pleasant-hill-2076" => array("type"=>"city", "pop"=>32862, "name_ru"=>"Плезант Хилл", "name_en"=>"Pleasant Hill", "lang"=>"en", "lat"=>37.953650, "lon"=>-122.074186),
"/US/US-CA/pleasanton-2077" => array("type"=>"city", "pop"=>66828, "name_ru"=>"Плезантон", "name_en"=>"Pleasanton", "lang"=>"en", "lat"=>37.672450, "lon"=>-121.889069),
"/US/US-CA/plymouth-2078" => array("type"=>"city", "pop"=>1033, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>38.478405, "lon"=>-120.844794),
"/US/US-CA/point-arena-2079" => array("type"=>"city", "pop"=>460, "name_ru"=>"Пойнт Арена", "name_en"=>"Point Arena", "lang"=>"en", "lat"=>38.911447, "lon"=>-123.690721),
"/US/US-CA/point-dume-2080" => array("type"=>"cdp", "pop"=>2809, "name_ru"=>"Пойнт-Дам", "name_en"=>"Point Dume", "lang"=>"en", "lat"=>34.012315, "lon"=>-118.801587),
"/US/US-CA/pollock-pines-2081" => array("type"=>"cdp", "pop"=>4291, "name_ru"=>"Поллок Пайнс", "name_en"=>"Pollock Pines", "lang"=>"en", "lat"=>38.752135, "lon"=>-120.570574),
"/US/US-CA/pomona-2082" => array("type"=>"city", "pop"=>152699, "name_ru"=>"Помона", "name_en"=>"Pomona", "lang"=>"en", "lat"=>34.058950, "lon"=>-117.761399),
"/US/US-CA/poplar-cotton-center-2083" => array("type"=>"cdp", "pop"=>1901, "name_ru"=>"Поплар-Коттон Сентер", "name_en"=>"Poplar-Cotton Center", "lang"=>"en", "lat"=>36.056742, "lon"=>-119.146415),
"/US/US-CA/port-hueneme-2085" => array("type"=>"city", "pop"=>21478, "name_ru"=>"Порт Хаенем", "name_en"=>"Port Hueneme", "lang"=>"en", "lat"=>34.162088, "lon"=>-119.203429),
"/US/US-CA/porterville-2084" => array("type"=>"city", "pop"=>51830, "name_ru"=>"Портервилл", "name_en"=>"Porterville", "lang"=>"en", "lat"=>36.064567, "lon"=>-119.026952),
"/US/US-CA/portola-2086" => array("type"=>"city", "pop"=>2069, "name_ru"=>"Портола", "name_en"=>"Portola", "lang"=>"en", "lat"=>39.804874, "lon"=>-120.467640),
"/US/US-CA/portola-hills-2087" => array("type"=>"cdp", "pop"=>2677, "name_ru"=>"Портола Хилс", "name_en"=>"Portola Hills", "lang"=>"en", "lat"=>33.683652, "lon"=>-117.631991),
"/US/US-CA/portola-valley-2088" => array("type"=>"town", "pop"=>4532, "name_ru"=>"Портола Валли", "name_en"=>"Portola Valley", "lang"=>"en", "lat"=>37.371750, "lon"=>-122.218495),
"/US/US-CA/poway-2089" => array("type"=>"city", "pop"=>48858, "name_ru"=>"Пауэй", "name_en"=>"Poway", "lang"=>"en", "lat"=>32.985450, "lon"=>-117.017949),
"/US/US-CA/prunedale-2090" => array("type"=>"cdp", "pop"=>7393, "name_ru"=>"Прундейл", "name_en"=>"Prunedale", "lang"=>"en", "lat"=>36.799862, "lon"=>-121.656367),
"/US/US-CA/quail-valley-2091" => array("type"=>"cdp", "pop"=>1937, "name_ru"=>"Квейл Валли", "name_en"=>"Quail Valley", "lang"=>"en", "lat"=>33.705394, "lon"=>-117.252078),
"/US/US-CA/quartz-hill-2092" => array("type"=>"cdp", "pop"=>9626, "name_ru"=>"Квартц Хилл", "name_en"=>"Quartz Hill", "lang"=>"en", "lat"=>34.652216, "lon"=>-118.214898),
"/US/US-CA/quincy-east-quincy-2093" => array("type"=>"cdp", "pop"=>4271, "name_ru"=>"Куинси-Ист Куинси", "name_en"=>"Quincy-East Quincy", "lang"=>"en", "lat"=>39.923474, "lon"=>-120.923751),
"/US/US-CA/rainbow-2094" => array("type"=>"cdp", "pop"=>2006, "name_ru"=>"Рейнбоу", "name_en"=>"Rainbow", "lang"=>"en", "lat"=>33.405200, "lon"=>-117.147484),
"/US/US-CA/ramona-2095" => array("type"=>"cdp", "pop"=>13040, "name_ru"=>"Рамона", "name_en"=>"Ramona", "lang"=>"en", "lat"=>33.036150, "lon"=>-116.870051),
"/US/US-CA/rancho-cordova-2096" => array("type"=>"cdp", "pop"=>62265, "name_ru"=>"Ранчо Кордова", "name_en"=>"Rancho Cordova", "lang"=>"en", "lat"=>38.595400, "lon"=>-121.300248),
"/US/US-CA/rancho-cucamonga-2097" => array("type"=>"city", "pop"=>171176, "name_ru"=>"Ранчо Какамонга", "name_en"=>"Rancho Cucamonga", "lang"=>"en", "lat"=>34.123831, "lon"=>-117.570239),
"/US/US-CA/rancho-mirage-2098" => array("type"=>"city", "pop"=>16714, "name_ru"=>"Ранчо Мираж", "name_en"=>"Rancho Mirage", "lang"=>"en", "lat"=>33.757981, "lon"=>-116.425561),
"/US/US-CA/rancho-murieta-2099" => array("type"=>"cdp", "pop"=>2336, "name_ru"=>"Ранчо Мериета", "name_en"=>"Rancho Murieta", "lang"=>"en", "lat"=>38.520448, "lon"=>-121.060182),
"/US/US-CA/rancho-palos-verdes-2100" => array("type"=>"city", "pop"=>41106, "name_ru"=>"Ранчо Палос Вердес", "name_en"=>"Rancho Palos Verdes", "lang"=>"en", "lat"=>33.758200, "lon"=>-118.363006),
"/US/US-CA/rancho-rinconada-2101" => array("type"=>"cdp", "pop"=>4206, "name_ru"=>"Ранчо Ринконадо", "name_en"=>"Rancho Rinconada", "lang"=>"en", "lat"=>37.315000, "lon"=>-122.001835),
"/US/US-CA/rancho-san-diego-2102" => array("type"=>"cdp", "pop"=>6977, "name_ru"=>"Ранчо Сан Диего", "name_en"=>"Rancho San Diego", "lang"=>"en", "lat"=>32.747250, "lon"=>-116.934481),
"/US/US-CA/rancho-santa-margarita-2103" => array("type"=>"cdp", "pop"=>49591, "name_ru"=>"Ранчо Санта Мрагарита", "name_en"=>"Rancho Santa Margarita", "lang"=>"en", "lat"=>33.640695, "lon"=>-117.602178),
"/US/US-CA/red-bluff-2104" => array("type"=>"city", "pop"=>14025, "name_ru"=>"Ред Блафф", "name_en"=>"Red Bluff", "lang"=>"en", "lat"=>40.173752, "lon"=>-122.240366),
"/US/US-CA/redding-2105" => array("type"=>"city", "pop"=>90201, "name_ru"=>"Реддинг", "name_en"=>"Redding", "lang"=>"en", "lat"=>40.572400, "lon"=>-122.364918),
"/US/US-CA/redlands-2106" => array("type"=>"city", "pop"=>69689, "name_ru"=>"Редлендс", "name_en"=>"Redlands", "lang"=>"en", "lat"=>34.057900, "lon"=>-117.170852),
"/US/US-CA/redondo-beach-2107" => array("type"=>"city", "pop"=>66882, "name_ru"=>"Редондо Бич", "name_en"=>"Redondo Beach", "lang"=>"en", "lat"=>33.856500, "lon"=>-118.376174),
"/US/US-CA/redway-2108" => array("type"=>"cdp", "pop"=>1212, "name_ru"=>"Редуэй", "name_en"=>"Redway", "lang"=>"en", "lat"=>40.117712, "lon"=>-123.819356),
"/US/US-CA/redwood-city-2109" => array("type"=>"city", "pop"=>74060, "name_ru"=>"Редвуд Сити", "name_en"=>"Redwood City", "lang"=>"en", "lat"=>37.516428, "lon"=>-122.205943),
"/US/US-CA/reedley-2110" => array("type"=>"city", "pop"=>23439, "name_ru"=>"Ридли", "name_en"=>"Reedley", "lang"=>"en", "lat"=>36.600400, "lon"=>-119.448299),
"/US/US-CA/rialto-2111" => array("type"=>"city", "pop"=>98700, "name_ru"=>"Риалто", "name_en"=>"Rialto", "lang"=>"en", "lat"=>34.113567, "lon"=>-117.387023),
"/US/US-CA/richgrove-2112" => array("type"=>"cdp", "pop"=>1899, "name_ru"=>"Ричгров", "name_en"=>"Richgrove", "lang"=>"en", "lat"=>35.796724, "lon"=>-119.105718),
"/US/US-CA/richmond-2113" => array("type"=>"city", "pop"=>102285, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>37.951350, "lon"=>-122.360939),
"/US/US-CA/ridgecrest-2114" => array("type"=>"city", "pop"=>25638, "name_ru"=>"Риджкрест", "name_en"=>"Ridgecrest", "lang"=>"en", "lat"=>35.631004, "lon"=>-117.661100),
"/US/US-CA/rio-del-mar-2116" => array("type"=>"cdp", "pop"=>8919, "name_ru"=>"Рио дель Мар", "name_en"=>"Rio del Mar", "lang"=>"en", "lat"=>36.957844, "lon"=>-121.883838),
"/US/US-CA/rio-dell-2115" => array("type"=>"city", "pop"=>3182, "name_ru"=>"Рио Делл", "name_en"=>"Rio Dell", "lang"=>"en", "lat"=>40.500822, "lon"=>-124.106330),
"/US/US-CA/rio-linda-2117" => array("type"=>"cdp", "pop"=>9481, "name_ru"=>"Рио Линда", "name_en"=>"Rio Linda", "lang"=>"en", "lat"=>38.687600, "lon"=>-121.458181),
"/US/US-CA/rio-vista-2118" => array("type"=>"city", "pop"=>7804, "name_ru"=>"Рио Виста", "name_en"=>"Rio Vista", "lang"=>"en", "lat"=>38.165067, "lon"=>-121.694679),
"/US/US-CA/ripon-2119" => array("type"=>"city", "pop"=>14473, "name_ru"=>"Рипон", "name_en"=>"Ripon", "lang"=>"en", "lat"=>37.735050, "lon"=>-121.126452),
"/US/US-CA/riverbank-2120" => array("type"=>"city", "pop"=>20606, "name_ru"=>"Ривербэнк", "name_en"=>"Riverbank", "lang"=>"en", "lat"=>37.732810, "lon"=>-120.945099),
"/US/US-CA/riverdale-2121" => array("type"=>"cdp", "pop"=>1980, "name_ru"=>"Ривердейл", "name_en"=>"Riverdale", "lang"=>"en", "lat"=>36.430800, "lon"=>-119.865977),
"/US/US-CA/riverside-2122" => array("type"=>"city", "pop"=>295357, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>33.940437, "lon"=>-117.397363),
"/US/US-CA/rocklin-2123" => array("type"=>"city", "pop"=>52811, "name_ru"=>"Роклин", "name_en"=>"Rocklin", "lang"=>"en", "lat"=>38.802400, "lon"=>-121.240270),
"/US/US-CA/rodeo-2124" => array("type"=>"cdp", "pop"=>7589, "name_ru"=>"Родео", "name_en"=>"Rodeo", "lang"=>"en", "lat"=>38.045467, "lon"=>-122.237698),
"/US/US-CA/rohnert-park-2125" => array("type"=>"city", "pop"=>40496, "name_ru"=>"Ронерт Парк", "name_en"=>"Rohnert Park", "lang"=>"en", "lat"=>38.347767, "lon"=>-122.697672),
"/US/US-CA/rolling-hills-2126" => array("type"=>"city", "pop"=>1908, "name_ru"=>"Роллинг Хилс", "name_en"=>"Rolling Hills", "lang"=>"en", "lat"=>33.761750, "lon"=>-118.341798),
"/US/US-CA/rolling-hills-estates-2127" => array("type"=>"city", "pop"=>7860, "name_ru"=>"Роллинг Хилс Эстейтс", "name_en"=>"Rolling Hills Estates", "lang"=>"en", "lat"=>33.782589, "lon"=>-118.352060),
"/US/US-CA/romoland-2128" => array("type"=>"cdp", "pop"=>2319, "name_ru"=>"Ромоленд", "name_en"=>"Romoland", "lang"=>"en", "lat"=>33.744629, "lon"=>-117.171728),
"/US/US-CA/rosamond-2129" => array("type"=>"cdp", "pop"=>7430, "name_ru"=>"Розамонд", "name_en"=>"Rosamond", "lang"=>"en", "lat"=>34.860184, "lon"=>-118.199499),
"/US/US-CA/rosedale-2130" => array("type"=>"cdp", "pop"=>4673, "name_ru"=>"Роуздейл", "name_en"=>"Rosedale", "lang"=>"en", "lat"=>35.394464, "lon"=>-119.203533),
"/US/US-CA/roseland-2131" => array("type"=>"cdp", "pop"=>8779, "name_ru"=>"Роузлэнд", "name_en"=>"Roseland", "lang"=>"en", "lat"=>38.421500, "lon"=>-122.724872),
"/US/US-CA/rosemead-2132" => array("type"=>"city", "pop"=>54412, "name_ru"=>"Роузмид", "name_en"=>"Rosemead", "lang"=>"en", "lat"=>34.068850, "lon"=>-118.081874),
"/US/US-CA/rosemont-2133" => array("type"=>"cdp", "pop"=>22851, "name_ru"=>"Роузмонт", "name_en"=>"Rosemont", "lang"=>"en", "lat"=>38.547173, "lon"=>-121.350802),
"/US/US-CA/roseville-2134" => array("type"=>"city", "pop"=>112660, "name_ru"=>"Роузвилл", "name_en"=>"Roseville", "lang"=>"en", "lat"=>38.759812, "lon"=>-121.293695),
"/US/US-CA/ross-2135" => array("type"=>"town", "pop"=>2278, "name_ru"=>"Росс", "name_en"=>"Ross", "lang"=>"en", "lat"=>37.961850, "lon"=>-122.560620),
"/US/US-CA/rossmoor-2136" => array("type"=>"cdp", "pop"=>9893, "name_ru"=>"Россмур", "name_en"=>"Rossmoor", "lang"=>"en", "lat"=>33.789500, "lon"=>-118.078898),
"/US/US-CA/rowland-heights-2137" => array("type"=>"cdp", "pop"=>42647, "name_ru"=>"Роленд Хайтс", "name_en"=>"Rowland Heights", "lang"=>"en", "lat"=>33.979800, "lon"=>-117.891416),
"/US/US-CA/rubidoux-2138" => array("type"=>"cdp", "pop"=>24367, "name_ru"=>"Рабиду", "name_en"=>"Rubidoux", "lang"=>"en", "lat"=>33.995467, "lon"=>-117.421174),
"/US/US-CA/running-springs-2139" => array("type"=>"cdp", "pop"=>4195, "name_ru"=>"Раннинг Спрингс", "name_en"=>"Running Springs", "lang"=>"en", "lat"=>34.209738, "lon"=>-117.113496),
"/US/US-CA/sacramento-2140" => array("type"=>"city", "pop"=>463794, "name_ru"=>"Сакраменто", "name_en"=>"Sacramento", "lang"=>"en", "lat"=>38.566850, "lon"=>-121.467360),
"/US/US-CA/salida-2142" => array("type"=>"cdp", "pop"=>4499, "name_ru"=>"Салида", "name_en"=>"Salida", "lang"=>"en", "lat"=>37.708450, "lon"=>-121.084792),
"/US/US-CA/salinas-2143" => array("type"=>"city", "pop"=>143640, "name_ru"=>"Салинас", "name_en"=>"Salinas", "lang"=>"en", "lat"=>36.684833, "lon"=>-121.635081),
"/US/US-CA/san-andreas-2144" => array("type"=>"cdp", "pop"=>2115, "name_ru"=>"Сан Андреас", "name_en"=>"San Andreas", "lang"=>"en", "lat"=>38.189312, "lon"=>-120.674608),
"/US/US-CA/san-anselmo-2145" => array("type"=>"town", "pop"=>11986, "name_ru"=>"Сан Ансельмо", "name_en"=>"San Anselmo", "lang"=>"en", "lat"=>37.982600, "lon"=>-122.568887),
"/US/US-CA/san-antonio-heights-2146" => array("type"=>"cdp", "pop"=>2935, "name_ru"=>"Сан Антонио Хайтс", "name_en"=>"San Antonio Heights", "lang"=>"en", "lat"=>34.154333, "lon"=>-117.658434),
"/US/US-CA/san-bernardino-2147" => array("type"=>"city", "pop"=>198580, "name_ru"=>"Сан Бернардино", "name_en"=>"San Bernardino", "lang"=>"en", "lat"=>34.139750, "lon"=>-117.292299),
"/US/US-CA/san-bruno-2148" => array("type"=>"city", "pop"=>40315, "name_ru"=>"Сан Бруно", "name_en"=>"San Bruno", "lang"=>"en", "lat"=>37.624817, "lon"=>-122.428015),
"/US/US-CA/san-buenaventura-(ventura)-2149" => array("type"=>"city", "pop"=>103706, "name_ru"=>"Сан Буэновентура (Вентура)", "name_en"=>"San Buenaventura (Ventura)", "lang"=>"en", "lat"=>34.265900, "lon"=>-119.258366),
"/US/US-CA/san-carlos-2150" => array("type"=>"city", "pop"=>27238, "name_ru"=>"Сан Карлос", "name_en"=>"San Carlos", "lang"=>"en", "lat"=>37.497633, "lon"=>-122.267365),
"/US/US-CA/san-clemente-2151" => array("type"=>"city", "pop"=>61465, "name_ru"=>"Сан Клемент", "name_en"=>"San Clemente", "lang"=>"en", "lat"=>33.447211, "lon"=>-117.611068),
"/US/US-CA/san-diego-2153" => array("type"=>"city", "pop"=>1279329, "name_ru"=>"Сан Диего", "name_en"=>"San Diego", "lang"=>"en", "lat"=>32.814950, "lon"=>-117.135770),
"/US/US-CA/san-diego-country-estates-2154" => array("type"=>"cdp", "pop"=>6874, "name_ru"=>"Сан Диего Каунти Эстейтс", "name_en"=>"San Diego Country Estates", "lang"=>"en", "lat"=>33.006571, "lon"=>-116.782835),
"/US/US-CA/san-dimas-2155" => array("type"=>"city", "pop"=>35043, "name_ru"=>"Сан Димас", "name_en"=>"San Dimas", "lang"=>"en", "lat"=>34.109050, "lon"=>-117.807199),
"/US/US-CA/san-fernando-2156" => array("type"=>"city", "pop"=>23833, "name_ru"=>"Сан Фернандо", "name_en"=>"San Fernando", "lang"=>"en", "lat"=>34.289500, "lon"=>-118.434990),
"/US/US-CA/san-francisco-2157" => array("type"=>"city", "pop"=>808976, "name_ru"=>"Сан Франциско", "name_en"=>"San Francisco", "lang"=>"en", "lat"=>37.793250, "lon"=>-122.554783),
"/US/US-CA/san-gabriel-2158" => array("type"=>"city", "pop"=>40445, "name_ru"=>"Сан Габриэль", "name_en"=>"San Gabriel", "lang"=>"en", "lat"=>34.094700, "lon"=>-118.097599),
"/US/US-CA/san-jacinto-2160" => array("type"=>"city", "pop"=>37842, "name_ru"=>"Сан Джацинто", "name_en"=>"San Jacinto", "lang"=>"en", "lat"=>33.792655, "lon"=>-116.964224),
"/US/US-CA/san-joaquin-2161" => array("type"=>"city", "pop"=>4045, "name_ru"=>"Сан Джоакин", "name_en"=>"San Joaquin", "lang"=>"en", "lat"=>36.606936, "lon"=>-120.186935),
"/US/US-CA/san-jose-2162" => array("type"=>"city", "pop"=>948279, "name_ru"=>"Сан-Хосе", "name_en"=>"San Jose", "lang"=>"en", "lat"=>37.304000, "lon"=>-121.849783),
"/US/US-CA/san-juan-bautista-2163" => array("type"=>"city", "pop"=>1709, "name_ru"=>"Сан Хуан Баутиста", "name_en"=>"San Juan Bautista", "lang"=>"en", "lat"=>36.845960, "lon"=>-121.537056),
"/US/US-CA/san-juan-capistrano-2164" => array("type"=>"city", "pop"=>34793, "name_ru"=>"Сан Хуан Капистрано", "name_en"=>"San Juan Capistrano", "lang"=>"en", "lat"=>33.501640, "lon"=>-117.652632),
"/US/US-CA/san-leandro-2165" => array("type"=>"city", "pop"=>77880, "name_ru"=>"Сан Леандро", "name_en"=>"San Leandro", "lang"=>"en", "lat"=>37.704650, "lon"=>-122.160756),
"/US/US-CA/san-lorenzo-2166" => array("type"=>"cdp", "pop"=>19987, "name_ru"=>"Сан Лоренсо", "name_en"=>"San Lorenzo", "lang"=>"en", "lat"=>37.675050, "lon"=>-122.130114),
"/US/US-CA/san-luis-obispo-2167" => array("type"=>"city", "pop"=>43636, "name_ru"=>"Сан Луис Обиспо", "name_en"=>"San Luis Obispo", "lang"=>"en", "lat"=>35.272866, "lon"=>-120.664810),
"/US/US-CA/san-marcos-2168" => array("type"=>"city", "pop"=>79114, "name_ru"=>"Сан Маркос", "name_en"=>"San Marcos", "lang"=>"en", "lat"=>33.136633, "lon"=>-117.173109),
"/US/US-CA/san-marino-2169" => array("type"=>"city", "pop"=>12808, "name_ru"=>"Сан Марино", "name_en"=>"San Marino", "lang"=>"en", "lat"=>34.122650, "lon"=>-118.112061),
"/US/US-CA/san-martin-2170" => array("type"=>"cdp", "pop"=>1713, "name_ru"=>"Сан Мартин", "name_en"=>"San Martin", "lang"=>"en", "lat"=>37.084950, "lon"=>-121.606315),
"/US/US-CA/san-mateo-2171" => array("type"=>"city", "pop"=>92256, "name_ru"=>"Сан Матео", "name_en"=>"San Mateo", "lang"=>"en", "lat"=>37.561267, "lon"=>-122.312019),
"/US/US-CA/san-miguel-2172" => array("type"=>"cdp", "pop"=>1123, "name_ru"=>"Сан Мигель", "name_en"=>"San Miguel", "lang"=>"en", "lat"=>35.751687, "lon"=>-120.692239),
"/US/US-CA/san-pablo-2173" => array("type"=>"city", "pop"=>30729, "name_ru"=>"Сан Пабло", "name_en"=>"San Pablo", "lang"=>"en", "lat"=>37.962139, "lon"=>-122.342750),
"/US/US-CA/san-rafael-2174" => array("type"=>"city", "pop"=>55602, "name_ru"=>"Сан Рафаэль", "name_en"=>"San Rafael", "lang"=>"en", "lat"=>37.981100, "lon"=>-122.505944),
"/US/US-CA/san-ramon-2175" => array("type"=>"city", "pop"=>49161, "name_ru"=>"Сан Рамон", "name_en"=>"San Ramon", "lang"=>"en", "lat"=>37.757423, "lon"=>-121.951068),
"/US/US-CA/sand-city-2152" => array("type"=>"city", "pop"=>361, "name_ru"=>"Сэнд Сити", "name_en"=>"Sand City", "lang"=>"en", "lat"=>36.648633, "lon"=>-121.844715),
"/US/US-CA/sanger-2159" => array("type"=>"city", "pop"=>25447, "name_ru"=>"Сангер", "name_en"=>"Sanger", "lang"=>"en", "lat"=>36.700233, "lon"=>-119.553193),
"/US/US-CA/santa-ana-2176" => array("type"=>"city", "pop"=>339130, "name_ru"=>"Санта Ана", "name_en"=>"Santa Ana", "lang"=>"en", "lat"=>33.736400, "lon"=>-117.882114),
"/US/US-CA/santa-barbara-2177" => array("type"=>"city", "pop"=>86093, "name_ru"=>"Санта Брабара", "name_en"=>"Santa Barbara", "lang"=>"en", "lat"=>34.428500, "lon"=>-119.721904),
"/US/US-CA/santa-clara-2178" => array("type"=>"city", "pop"=>110200, "name_ru"=>"Санта Клара", "name_en"=>"Santa Clara", "lang"=>"en", "lat"=>37.365000, "lon"=>-121.966706),
"/US/US-CA/santa-clarita-2179" => array("type"=>"city", "pop"=>169500, "name_ru"=>"Санта Кларита", "name_en"=>"Santa Clarita", "lang"=>"en", "lat"=>34.413408, "lon"=>-118.510175),
"/US/US-CA/santa-cruz-2180" => array("type"=>"city", "pop"=>56124, "name_ru"=>"Санта Круз", "name_en"=>"Santa Cruz", "lang"=>"en", "lat"=>36.973250, "lon"=>-122.035821),
"/US/US-CA/santa-fe-springs-2181" => array("type"=>"city", "pop"=>17159, "name_ru"=>"Санта Фе Спрингс", "name_en"=>"Santa Fe Springs", "lang"=>"en", "lat"=>33.933496, "lon"=>-118.061569),
"/US/US-CA/santa-maria-2182" => array("type"=>"city", "pop"=>86356, "name_ru"=>"Санта Мария", "name_en"=>"Santa Maria", "lang"=>"en", "lat"=>34.937031, "lon"=>-120.436881),
"/US/US-CA/santa-monica-2183" => array("type"=>"city", "pop"=>87664, "name_ru"=>"Сантиа Моника", "name_en"=>"Santa Monica", "lang"=>"en", "lat"=>34.013400, "lon"=>-118.493229),
"/US/US-CA/santa-paula-2184" => array("type"=>"city", "pop"=>28618, "name_ru"=>"Санта Пола", "name_en"=>"Santa Paula", "lang"=>"en", "lat"=>34.352256, "lon"=>-119.069272),
"/US/US-CA/santa-rosa-2185" => array("type"=>"city", "pop"=>155796, "name_ru"=>"Санта Роза", "name_en"=>"Santa Rosa", "lang"=>"en", "lat"=>38.448600, "lon"=>-122.700799),
"/US/US-CA/santa-venetia-2186" => array("type"=>"cdp", "pop"=>3362, "name_ru"=>"Санта Венешиа", "name_en"=>"Santa Venetia", "lang"=>"en", "lat"=>38.009095, "lon"=>-122.503674),
"/US/US-CA/santa-ynez-2187" => array("type"=>"cdp", "pop"=>4200, "name_ru"=>"Санта Йнез", "name_en"=>"Santa Ynez", "lang"=>"en", "lat"=>34.608018, "lon"=>-120.102899),
"/US/US-CA/santee-2188" => array("type"=>"city", "pop"=>53860, "name_ru"=>"Санти", "name_en"=>"Santee", "lang"=>"en", "lat"=>32.855117, "lon"=>-116.982732),
"/US/US-CA/saratoga-2189" => array("type"=>"city", "pop"=>30445, "name_ru"=>"Саратога", "name_en"=>"Saratoga", "lang"=>"en", "lat"=>37.267800, "lon"=>-122.023026),
"/US/US-CA/sausalito-2190" => array("type"=>"city", "pop"=>7158, "name_ru"=>"Сосалито", "name_en"=>"Sausalito", "lang"=>"en", "lat"=>37.857667, "lon"=>-122.491510),
"/US/US-CA/scotts-valley-2191" => array("type"=>"city", "pop"=>11128, "name_ru"=>"Скотс Валли", "name_en"=>"Scotts Valley", "lang"=>"en", "lat"=>37.056276, "lon"=>-122.008013),
"/US/US-CA/seal-beach-2192" => array("type"=>"city", "pop"=>24127, "name_ru"=>"Сил Бич", "name_en"=>"Seal Beach", "lang"=>"en", "lat"=>33.758450, "lon"=>-118.075785),
"/US/US-CA/searles-valley-2193" => array("type"=>"cdp", "pop"=>2740, "name_ru"=>"Сирлс Валли", "name_en"=>"Searles Valley", "lang"=>"en", "lat"=>35.767637, "lon"=>-117.402927),
"/US/US-CA/seaside-2194" => array("type"=>"city", "pop"=>33797, "name_ru"=>"Сисайд", "name_en"=>"Seaside", "lang"=>"en", "lat"=>36.621400, "lon"=>-121.818584),
"/US/US-CA/sebastopol-2195" => array("type"=>"city", "pop"=>7518, "name_ru"=>"Себастопол", "name_en"=>"Sebastopol", "lang"=>"en", "lat"=>38.399250, "lon"=>-122.825932),
"/US/US-CA/sedco-hills-2196" => array("type"=>"cdp", "pop"=>3008, "name_ru"=>"Седко Хилс", "name_en"=>"Sedco Hills", "lang"=>"en", "lat"=>33.640021, "lon"=>-117.282508),
"/US/US-CA/seeley-2197" => array("type"=>"cdp", "pop"=>1228, "name_ru"=>"Сили", "name_en"=>"Seeley", "lang"=>"en", "lat"=>32.789028, "lon"=>-115.680191),
"/US/US-CA/selma-2198" => array("type"=>"city", "pop"=>22807, "name_ru"=>"Сельма", "name_en"=>"Selma", "lang"=>"en", "lat"=>36.574250, "lon"=>-119.615355),
"/US/US-CA/shafter-2199" => array("type"=>"city", "pop"=>15800, "name_ru"=>"Шафтер", "name_en"=>"Shafter", "lang"=>"en", "lat"=>35.501232, "lon"=>-119.272655),
"/US/US-CA/shingle-springs-2200" => array("type"=>"cdp", "pop"=>2049, "name_ru"=>"Шингл Спрингс", "name_en"=>"Shingle Springs", "lang"=>"en", "lat"=>38.670775, "lon"=>-120.939655),
"/US/US-CA/sierra-madre-2201" => array("type"=>"city", "pop"=>10834, "name_ru"=>"Сиерра Мадре", "name_en"=>"Sierra Madre", "lang"=>"en", "lat"=>34.168700, "lon"=>-118.049331),
"/US/US-CA/signal-hill-2202" => array("type"=>"city", "pop"=>10882, "name_ru"=>"Сигнал Хилл", "name_en"=>"Signal Hill", "lang"=>"en", "lat"=>33.803600, "lon"=>-118.167413),
"/US/US-CA/simi-valley-2203" => array("type"=>"city", "pop"=>120543, "name_ru"=>"Сими Валли", "name_en"=>"Simi Valley", "lang"=>"en", "lat"=>34.263050, "lon"=>-118.751299),
"/US/US-CA/solana-beach-2204" => array("type"=>"city", "pop"=>12825, "name_ru"=>"Солана Бич", "name_en"=>"Solana Beach", "lang"=>"en", "lat"=>32.995850, "lon"=>-117.255858),
"/US/US-CA/soledad-2205" => array("type"=>"city", "pop"=>28124, "name_ru"=>"Соледад", "name_en"=>"Soledad", "lang"=>"en", "lat"=>36.430099, "lon"=>-121.322549),
"/US/US-CA/solvang-2206" => array("type"=>"city", "pop"=>5148, "name_ru"=>"Солванг", "name_en"=>"Solvang", "lang"=>"en", "lat"=>34.592476, "lon"=>-120.139314),
"/US/US-CA/sonoma-2207" => array("type"=>"city", "pop"=>9893, "name_ru"=>"Сонома", "name_en"=>"Sonoma", "lang"=>"en", "lat"=>38.290300, "lon"=>-122.459265),
"/US/US-CA/sonora-2208" => array("type"=>"city", "pop"=>4559, "name_ru"=>"Сонора", "name_en"=>"Sonora", "lang"=>"en", "lat"=>37.983579, "lon"=>-120.381637),
"/US/US-CA/soquel-2209" => array("type"=>"cdp", "pop"=>9188, "name_ru"=>"Сокел", "name_en"=>"Soquel", "lang"=>"en", "lat"=>36.993800, "lon"=>-121.946728),
"/US/US-CA/soulsbyville-2210" => array("type"=>"cdp", "pop"=>1732, "name_ru"=>"Сайлсбивилл", "name_en"=>"Soulsbyville", "lang"=>"en", "lat"=>37.992577, "lon"=>-120.259868),
"/US/US-CA/south-dos-palos-2211" => array("type"=>"cdp", "pop"=>1214, "name_ru"=>"Саут Дос Палос", "name_en"=>"South Dos Palos", "lang"=>"en", "lat"=>36.971237, "lon"=>-120.648152),
"/US/US-CA/south-el-monte-2212" => array("type"=>"city", "pop"=>21397, "name_ru"=>"Саут Эль Монте", "name_en"=>"South El Monte", "lang"=>"en", "lat"=>34.048700, "lon"=>-118.047578),
"/US/US-CA/south-gate-2213" => array("type"=>"city", "pop"=>96640, "name_ru"=>"Саут Гейт", "name_en"=>"South Gate", "lang"=>"en", "lat"=>33.943950, "lon"=>-118.191930),
"/US/US-CA/south-lake-2214" => array("type"=>"cdp", "pop"=>1059, "name_ru"=>"Саут Лейк", "name_en"=>"South Lake", "lang"=>"en", "lat"=>35.633530, "lon"=>-118.355636),
"/US/US-CA/south-lake-tahoe-2215" => array("type"=>"city", "pop"=>23333, "name_ru"=>"Саут Лейк Тахо", "name_en"=>"South Lake Tahoe", "lang"=>"en", "lat"=>38.939269, "lon"=>-119.981136),
"/US/US-CA/south-oroville-2216" => array("type"=>"cdp", "pop"=>7463, "name_ru"=>"Саут Оровилл", "name_en"=>"South Oroville", "lang"=>"en", "lat"=>39.482512, "lon"=>-121.532762),
"/US/US-CA/south-pasadena-2217" => array("type"=>"city", "pop"=>24446, "name_ru"=>"Саут Пасадина", "name_en"=>"South Pasadena", "lang"=>"en", "lat"=>34.111350, "lon"=>-118.157024),
"/US/US-CA/south-san-francisco-2218" => array("type"=>"city", "pop"=>62502, "name_ru"=>"Саут Сан Франциско", "name_en"=>"South San Francisco", "lang"=>"en", "lat"=>37.658635, "lon"=>-122.308393),
"/US/US-CA/south-san-gabriel-2219" => array("type"=>"cdp", "pop"=>7700, "name_ru"=>"Саут Сан Габриэль", "name_en"=>"South San Gabriel", "lang"=>"en", "lat"=>34.048807, "lon"=>-118.096242),
"/US/US-CA/south-san-jose-hills-2220" => array("type"=>"cdp", "pop"=>17814, "name_ru"=>"Саут Сан Хосе Хилс", "name_en"=>"South San Jose Hills", "lang"=>"en", "lat"=>34.012833, "lon"=>-117.903753),
"/US/US-CA/south-santa-rosa-2221" => array("type"=>"cdp", "pop"=>4128, "name_ru"=>"Саут Санта Роза", "name_en"=>"South Santa Rosa", "lang"=>"en", "lat"=>38.404900, "lon"=>-122.726418),
"/US/US-CA/south-taft-2222" => array("type"=>"cdp", "pop"=>2170, "name_ru"=>"Саут Тафт", "name_en"=>"South Taft", "lang"=>"en", "lat"=>35.127023, "lon"=>-119.454830),
"/US/US-CA/south-whittier-2223" => array("type"=>"cdp", "pop"=>49514, "name_ru"=>"Саут Уиттиер", "name_en"=>"South Whittier", "lang"=>"en", "lat"=>33.933000, "lon"=>-118.029692),
"/US/US-CA/south-yuba-city-2224" => array("type"=>"cdp", "pop"=>8816, "name_ru"=>"Саут Иуба Сити", "name_en"=>"South Yuba City", "lang"=>"en", "lat"=>39.116533, "lon"=>-121.637973),
"/US/US-CA/spring-valley-2225" => array("type"=>"cdp", "pop"=>55331, "name_ru"=>"Спринг Валли", "name_en"=>"Spring Valley", "lang"=>"en", "lat"=>32.720850, "lon"=>-116.988860),
"/US/US-CA/squaw-valley-2226" => array("type"=>"cdp", "pop"=>2161, "name_ru"=>"Скво Валли", "name_en"=>"Squaw Valley", "lang"=>"en", "lat"=>36.696397, "lon"=>-119.194209),
"/US/US-CA/st.-helena-2141" => array("type"=>"city", "pop"=>5817, "name_ru"=>"Сент-Хэлена", "name_en"=>"St. Helena", "lang"=>"en", "lat"=>38.504700, "lon"=>-122.465723),
"/US/US-CA/stanford-2227" => array("type"=>"cdp", "pop"=>18097, "name_ru"=>"Стэнфорд", "name_en"=>"Stanford", "lang"=>"en", "lat"=>37.424050, "lon"=>-122.164887),
"/US/US-CA/stanton-2228" => array("type"=>"city", "pop"=>37589, "name_ru"=>"Стэнтон", "name_en"=>"Stanton", "lang"=>"en", "lat"=>33.800448, "lon"=>-117.990923),
"/US/US-CA/stockton-2229" => array("type"=>"city", "pop"=>287037, "name_ru"=>"Стоктон", "name_en"=>"Stockton", "lang"=>"en", "lat"=>37.969850, "lon"=>-121.306741),
"/US/US-CA/strathmore-2230" => array("type"=>"cdp", "pop"=>2353, "name_ru"=>"Стратмор", "name_en"=>"Strathmore", "lang"=>"en", "lat"=>36.144329, "lon"=>-119.059682),
"/US/US-CA/strawberry-2231" => array("type"=>"cdp", "pop"=>4377, "name_ru"=>"Строберри", "name_en"=>"Strawberry", "lang"=>"en", "lat"=>37.896967, "lon"=>-122.507657),
"/US/US-CA/suisun-city-2232" => array("type"=>"city", "pop"=>26977, "name_ru"=>"Саисан Сити", "name_en"=>"Suisun City", "lang"=>"en", "lat"=>38.245700, "lon"=>-122.009849),
"/US/US-CA/sun-city-2233" => array("type"=>"cdp", "pop"=>14930, "name_ru"=>"Сан Сити", "name_en"=>"Sun City", "lang"=>"en", "lat"=>33.712888, "lon"=>-117.202514),
"/US/US-CA/sunnyside-tahoe-city-2234" => array("type"=>"cdp", "pop"=>1643, "name_ru"=>"Саннисайд-Тахо Сити", "name_en"=>"Sunnyside-Tahoe City", "lang"=>"en", "lat"=>39.147519, "lon"=>-120.163737),
"/US/US-CA/sunnyslope-2235" => array("type"=>"cdp", "pop"=>3766, "name_ru"=>"Саннислоуп", "name_en"=>"Sunnyslope", "lang"=>"en", "lat"=>34.018940, "lon"=>-117.422732),
"/US/US-CA/sunnyvale-2236" => array("type"=>"city", "pop"=>132109, "name_ru"=>"Саннивейл", "name_en"=>"Sunnyvale", "lang"=>"en", "lat"=>37.385750, "lon"=>-122.025531),
"/US/US-CA/susanville-2237" => array("type"=>"city", "pop"=>17336, "name_ru"=>"Сьюзанвилл", "name_en"=>"Susanville", "lang"=>"en", "lat"=>40.411059, "lon"=>-120.650607),
"/US/US-CA/sutter-2238" => array("type"=>"cdp", "pop"=>92207, "name_ru"=>"Саттер", "name_en"=>"Sutter", "lang"=>"en", "lat"=>39.159256, "lon"=>-121.746207),
"/US/US-CA/sutter-creek-2239" => array("type"=>"city", "pop"=>2719, "name_ru"=>"Саттер Крик", "name_en"=>"Sutter Creek", "lang"=>"en", "lat"=>38.390792, "lon"=>-120.803471),
"/US/US-CA/taft-2240" => array("type"=>"city", "pop"=>9147, "name_ru"=>"Тафт", "name_en"=>"Taft", "lang"=>"en", "lat"=>35.142398, "lon"=>-119.454828),
"/US/US-CA/taft-heights-2241" => array("type"=>"cdp", "pop"=>2050, "name_ru"=>"Тафт Хайтс", "name_en"=>"Taft Heights", "lang"=>"en", "lat"=>35.130864, "lon"=>-119.473609),
"/US/US-CA/tahoe-vista-2242" => array("type"=>"cdp", "pop"=>1144, "name_ru"=>"Тахо Виста", "name_en"=>"Tahoe Vista", "lang"=>"en", "lat"=>39.247945, "lon"=>-120.053628),
"/US/US-CA/tamalpais-homestead-valley-2243" => array("type"=>"cdp", "pop"=>9601, "name_ru"=>"Тамалпаис-Хоумстед Валли", "name_en"=>"Tamalpais-Homestead Valley", "lang"=>"en", "lat"=>37.888350, "lon"=>-122.538999),
"/US/US-CA/tara-hills-2244" => array("type"=>"cdp", "pop"=>4998, "name_ru"=>"Тара Хилс", "name_en"=>"Tara Hills", "lang"=>"en", "lat"=>37.993600, "lon"=>-122.317479),
"/US/US-CA/tehachapi-2245" => array("type"=>"city", "pop"=>11736, "name_ru"=>"Техачапи", "name_en"=>"Tehachapi", "lang"=>"en", "lat"=>35.134790, "lon"=>-118.440465),
"/US/US-CA/tehama-2246" => array("type"=>"city", "pop"=>451, "name_ru"=>"Техама", "name_en"=>"Tehama", "lang"=>"en", "lat"=>40.023256, "lon"=>-122.123851),
"/US/US-CA/temecula-2247" => array("type"=>"city", "pop"=>97100, "name_ru"=>"Темекула", "name_en"=>"Temecula", "lang"=>"en", "lat"=>33.502683, "lon"=>-117.128798),
"/US/US-CA/temelec-2248" => array("type"=>"cdp", "pop"=>1594, "name_ru"=>"Темелек", "name_en"=>"Temelec", "lang"=>"en", "lat"=>38.256267, "lon"=>-122.503341),
"/US/US-CA/temple-city-2249" => array("type"=>"city", "pop"=>38341, "name_ru"=>"Темпл Сити", "name_en"=>"Temple City", "lang"=>"en", "lat"=>34.102633, "lon"=>-118.057078),
"/US/US-CA/templeton-2250" => array("type"=>"cdp", "pop"=>2887, "name_ru"=>"Темплтон", "name_en"=>"Templeton", "lang"=>"en", "lat"=>35.553885, "lon"=>-120.708444),
"/US/US-CA/terra-bella-2251" => array("type"=>"cdp", "pop"=>2740, "name_ru"=>"Терра Белла", "name_en"=>"Terra Bella", "lang"=>"en", "lat"=>35.959834, "lon"=>-119.038011),
"/US/US-CA/thermalito-2252" => array("type"=>"cdp", "pop"=>5646, "name_ru"=>"Термалито", "name_en"=>"Thermalito", "lang"=>"en", "lat"=>39.491123, "lon"=>-121.608019),
"/US/US-CA/thousand-oaks-2253" => array("type"=>"city", "pop"=>123091, "name_ru"=>"Таусанд Оукс", "name_en"=>"Thousand Oaks", "lang"=>"en", "lat"=>34.192700, "lon"=>-118.867557),
"/US/US-CA/thousand-palms-2254" => array("type"=>"cdp", "pop"=>4122, "name_ru"=>"Таусанд Палмс", "name_en"=>"Thousand Palms", "lang"=>"en", "lat"=>33.821800, "lon"=>-116.385825),
"/US/US-CA/tiburon-2255" => array("type"=>"town", "pop"=>8676, "name_ru"=>"Тибурон", "name_en"=>"Tiburon", "lang"=>"en", "lat"=>37.886792, "lon"=>-122.456764),
"/US/US-CA/tierra-buena-2256" => array("type"=>"cdp", "pop"=>2878, "name_ru"=>"Тьерра Буэна", "name_en"=>"Tierra Buena", "lang"=>"en", "lat"=>39.158647, "lon"=>-121.667729),
"/US/US-CA/tipton-2257" => array("type"=>"cdp", "pop"=>1383, "name_ru"=>"Типтон", "name_en"=>"Tipton", "lang"=>"en", "lat"=>36.058087, "lon"=>-119.311892),
"/US/US-CA/torrance-2258" => array("type"=>"city", "pop"=>140820, "name_ru"=>"Торранс", "name_en"=>"Torrance", "lang"=>"en", "lat"=>33.834800, "lon"=>-118.340424),
"/US/US-CA/trabuco-highlands-2259" => array("type"=>"cdp", "pop"=>3191, "name_ru"=>"Трабуко Хайлендс", "name_en"=>"Trabuco Highlands", "lang"=>"en", "lat"=>33.614859, "lon"=>-117.565460),
"/US/US-CA/tracy-2260" => array("type"=>"city", "pop"=>79196, "name_ru"=>"Трейси", "name_en"=>"Tracy", "lang"=>"en", "lat"=>37.736950, "lon"=>-121.431868),
"/US/US-CA/trinidad-2261" => array("type"=>"city", "pop"=>310, "name_ru"=>"Тринидад", "name_en"=>"Trinidad", "lang"=>"en", "lat"=>41.057727, "lon"=>-124.142084),
"/US/US-CA/truckee-2262" => array("type"=>"cdp", "pop"=>16185, "name_ru"=>"Траки", "name_en"=>"Truckee", "lang"=>"en", "lat"=>39.326365, "lon"=>-120.202296),
"/US/US-CA/tulare-2263" => array("type"=>"city", "pop"=>56654, "name_ru"=>"Тулар", "name_en"=>"Tulare", "lang"=>"en", "lat"=>36.195150, "lon"=>-119.343079),
"/US/US-CA/tulelake-2264" => array("type"=>"city", "pop"=>958, "name_ru"=>"Талелейк", "name_en"=>"Tulelake", "lang"=>"en", "lat"=>41.954152, "lon"=>-121.474804),
"/US/US-CA/tuolumne-city-2265" => array("type"=>"cdp", "pop"=>1686, "name_ru"=>"Таоламн Сити", "name_en"=>"Tuolumne City", "lang"=>"en", "lat"=>37.962821, "lon"=>-120.240278),
"/US/US-CA/turlock-2266" => array("type"=>"city", "pop"=>68404, "name_ru"=>"Терлок", "name_en"=>"Turlock", "lang"=>"en", "lat"=>37.504912, "lon"=>-120.850403),
"/US/US-CA/tustin-2267" => array("type"=>"city", "pop"=>71814, "name_ru"=>"Тастин", "name_en"=>"Tustin", "lang"=>"en", "lat"=>33.730005, "lon"=>-117.811099),
"/US/US-CA/tustin-foothills-2268" => array("type"=>"cdp", "pop"=>24358, "name_ru"=>"Тастин Футхилс", "name_en"=>"Tustin Foothills", "lang"=>"en", "lat"=>33.764550, "lon"=>-117.793022),
"/US/US-CA/twain-harte-2269" => array("type"=>"cdp", "pop"=>2170, "name_ru"=>"Твейн Харт", "name_en"=>"Twain Harte", "lang"=>"en", "lat"=>38.040468, "lon"=>-120.232649),
"/US/US-CA/twentynine-palms-2270" => array("type"=>"city", "pop"=>33480, "name_ru"=>"Твентинайн Палмс", "name_en"=>"Twentynine Palms", "lang"=>"en", "lat"=>34.139498, "lon"=>-116.065704),
"/US/US-CA/twentynine-palms-base-2271" => array("type"=>"cdp", "pop"=>10606, "name_ru"=>"Твентинайн Палмс Бейз", "name_en"=>"Twentynine Palms Base", "lang"=>"en", "lat"=>34.228116, "lon"=>-116.055732),
"/US/US-CA/twin-lakes-2272" => array("type"=>"cdp", "pop"=>5379, "name_ru"=>"Твин Лейкс", "name_en"=>"Twin Lakes", "lang"=>"en", "lat"=>36.961300, "lon"=>-121.991044),
"/US/US-CA/ukiah-2273" => array("type"=>"city", "pop"=>14956, "name_ru"=>"Юкайа", "name_en"=>"Ukiah", "lang"=>"en", "lat"=>39.145354, "lon"=>-123.208714),
"/US/US-CA/union-city-2274" => array("type"=>"city", "pop"=>72123, "name_ru"=>"Юнион Сити", "name_en"=>"Union City", "lang"=>"en", "lat"=>37.601650, "lon"=>-122.016072),
"/US/US-CA/upland-2275" => array("type"=>"city", "pop"=>72091, "name_ru"=>"Апленд", "name_en"=>"Upland", "lang"=>"en", "lat"=>34.118800, "lon"=>-117.657749),
"/US/US-CA/vacaville-2276" => array("type"=>"city", "pop"=>92219, "name_ru"=>"Вакавилл", "name_en"=>"Vacaville", "lang"=>"en", "lat"=>38.362267, "lon"=>-121.966053),
"/US/US-CA/val-verde-2281" => array("type"=>"cdp", "pop"=>1689, "name_ru"=>"Вал Верде", "name_en"=>"Val Verde", "lang"=>"en", "lat"=>34.445499, "lon"=>-118.655704),
"/US/US-CA/valinda-2277" => array("type"=>"cdp", "pop"=>18735, "name_ru"=>"Валинда", "name_en"=>"Valinda", "lang"=>"en", "lat"=>34.039350, "lon"=>-117.929279),
"/US/US-CA/valle-vista-2279" => array("type"=>"cdp", "pop"=>8751, "name_ru"=>"Валле Виста", "name_en"=>"Valle Vista", "lang"=>"en", "lat"=>33.750600, "lon"=>-116.896054),
"/US/US-CA/vallejo-2278" => array("type"=>"city", "pop"=>114729, "name_ru"=>"Вальехо", "name_en"=>"Vallejo", "lang"=>"en", "lat"=>38.107467, "lon"=>-122.264186),
"/US/US-CA/valley-center-2280" => array("type"=>"cdp", "pop"=>1711, "name_ru"=>"Валли Сентер", "name_en"=>"Valley Center", "lang"=>"en", "lat"=>33.217400, "lon"=>-117.010990),
"/US/US-CA/vandenberg-afb-2282" => array("type"=>"cdp", "pop"=>9846, "name_ru"=>"Военно-воздушная база Ванденберг", "name_en"=>"Vandenberg AFB", "lang"=>"en", "lat"=>34.749261, "lon"=>-120.517325),
"/US/US-CA/vandenberg-village-2283" => array("type"=>"cdp", "pop"=>5971, "name_ru"=>"Ванденберг Виллидж", "name_en"=>"Vandenberg Village", "lang"=>"en", "lat"=>34.711230, "lon"=>-120.462643),
"/US/US-CA/vernon-2284" => array("type"=>"city", "pop"=>90, "name_ru"=>"Вернон", "name_en"=>"Vernon", "lang"=>"en", "lat"=>34.001200, "lon"=>-118.210074),
"/US/US-CA/victorville-2285" => array("type"=>"city", "pop"=>110318, "name_ru"=>"Викторвилл", "name_en"=>"Victorville", "lang"=>"en", "lat"=>34.523049, "lon"=>-117.331596),
"/US/US-CA/view-park-windsor-hills-2286" => array("type"=>"cdp", "pop"=>11769, "name_ru"=>"Вью Парк-Виндзор Хилс", "name_en"=>"View Park-Windsor Hills", "lang"=>"en", "lat"=>33.996300, "lon"=>-118.348787),
"/US/US-CA/villa-park-2287" => array("type"=>"city", "pop"=>5967, "name_ru"=>"Вилла Парк", "name_en"=>"Villa Park", "lang"=>"en", "lat"=>33.818500, "lon"=>-117.810321),
"/US/US-CA/vincent-2288" => array("type"=>"cdp", "pop"=>13713, "name_ru"=>"Винсент", "name_en"=>"Vincent", "lang"=>"en", "lat"=>34.098547, "lon"=>-117.923499),
"/US/US-CA/vine-hill-2289" => array("type"=>"cdp", "pop"=>3214, "name_ru"=>"Вайн Хилл", "name_en"=>"Vine Hill", "lang"=>"en", "lat"=>38.016008, "lon"=>-122.091498),
"/US/US-CA/visalia-2290" => array("type"=>"city", "pop"=>121040, "name_ru"=>"Визалия", "name_en"=>"Visalia", "lang"=>"en", "lat"=>36.325400, "lon"=>-119.320574),
"/US/US-CA/vista-2291" => array("type"=>"city", "pop"=>91144, "name_ru"=>"Виста", "name_en"=>"Vista", "lang"=>"en", "lat"=>33.188783, "lon"=>-117.238216),
"/US/US-CA/walnut-2292" => array("type"=>"city", "pop"=>30744, "name_ru"=>"Уолнат", "name_en"=>"Walnut", "lang"=>"en", "lat"=>34.029016, "lon"=>-117.857378),
"/US/US-CA/walnut-creek-2293" => array("type"=>"city", "pop"=>63486, "name_ru"=>"Уолнат Крик", "name_en"=>"Walnut Creek", "lang"=>"en", "lat"=>37.902010, "lon"=>-122.040849),
"/US/US-CA/walnut-park-2294" => array("type"=>"cdp", "pop"=>14722, "name_ru"=>"Уолнат Парк", "name_en"=>"Walnut Park", "lang"=>"en", "lat"=>33.968450, "lon"=>-118.219404),
"/US/US-CA/wasco-2295" => array("type"=>"city", "pop"=>24628, "name_ru"=>"Васко", "name_en"=>"Wasco", "lang"=>"en", "lat"=>35.593643, "lon"=>-119.341049),
"/US/US-CA/waterford-2296" => array("type"=>"city", "pop"=>8875, "name_ru"=>"Уотерфорд", "name_en"=>"Waterford", "lang"=>"en", "lat"=>37.645551, "lon"=>-120.764600),
"/US/US-CA/watsonville-2297" => array("type"=>"city", "pop"=>50442, "name_ru"=>"Уотсонвилл", "name_en"=>"Watsonville", "lang"=>"en", "lat"=>36.919250, "lon"=>-121.768540),
"/US/US-CA/weaverville-2298" => array("type"=>"cdp", "pop"=>3370, "name_ru"=>"Уивервилл", "name_en"=>"Weaverville", "lang"=>"en", "lat"=>40.768754, "lon"=>-122.948859),
"/US/US-CA/weed-2299" => array("type"=>"city", "pop"=>3024, "name_ru"=>"Уид", "name_en"=>"Weed", "lang"=>"en", "lat"=>41.416513, "lon"=>-122.378377),
"/US/US-CA/weedpatch-2300" => array("type"=>"cdp", "pop"=>1892, "name_ru"=>"Уидпатч", "name_en"=>"Weedpatch", "lang"=>"en", "lat"=>35.236984, "lon"=>-118.912971),
"/US/US-CA/west-athens-2301" => array("type"=>"cdp", "pop"=>8859, "name_ru"=>"Уэст Афины", "name_en"=>"West Athens", "lang"=>"en", "lat"=>33.923350, "lon"=>-118.302374),
"/US/US-CA/west-bishop-2302" => array("type"=>"cdp", "pop"=>2908, "name_ru"=>"Уэст Бишоп", "name_en"=>"West Bishop", "lang"=>"en", "lat"=>37.361069, "lon"=>-118.454125),
"/US/US-CA/west-carson-2303" => array("type"=>"cdp", "pop"=>20143, "name_ru"=>"Уэст Карсон", "name_en"=>"West Carson", "lang"=>"en", "lat"=>33.821650, "lon"=>-118.291655),
"/US/US-CA/west-compton-2304" => array("type"=>"cdp", "pop"=>5451, "name_ru"=>"Уэст Комптон", "name_en"=>"West Compton", "lang"=>"en", "lat"=>33.894000, "lon"=>-118.269689),
"/US/US-CA/west-covina-2305" => array("type"=>"city", "pop"=>105790, "name_ru"=>"Уэст Ковина", "name_en"=>"West Covina", "lang"=>"en", "lat"=>34.054250, "lon"=>-117.909849),
"/US/US-CA/west-hollywood-2307" => array("type"=>"city", "pop"=>36005, "name_ru"=>"Уэст Голливуд", "name_en"=>"West Hollywood", "lang"=>"en", "lat"=>34.087900, "lon"=>-118.371249),
"/US/US-CA/west-menlo-park-2309" => array("type"=>"cdp", "pop"=>3959, "name_ru"=>"Уэст Менло Парк", "name_en"=>"West Menlo Park", "lang"=>"en", "lat"=>37.433650, "lon"=>-122.201899),
"/US/US-CA/west-pittsburg-2313" => array("type"=>"cdp", "pop"=>17453, "name_ru"=>"Уэст Питсбург", "name_en"=>"West Pittsburg", "lang"=>"en", "lat"=>38.038150, "lon"=>-121.974023),
"/US/US-CA/west-puente-valley-2314" => array("type"=>"cdp", "pop"=>20254, "name_ru"=>"Уэст Пуэнте Валли", "name_en"=>"West Puente Valley", "lang"=>"en", "lat"=>34.051733, "lon"=>-117.967385),
"/US/US-CA/west-sacramento-2315" => array("type"=>"city", "pop"=>47511, "name_ru"=>"Уэст Сакраменто", "name_en"=>"West Sacramento", "lang"=>"en", "lat"=>38.556749, "lon"=>-121.547839),
"/US/US-CA/west-whittier-los-nietos-2316" => array("type"=>"cdp", "pop"=>24164, "name_ru"=>"Уэст Уиттиер-Лос Найетос", "name_en"=>"West Whittier-Los Nietos", "lang"=>"en", "lat"=>33.977800, "lon"=>-118.067555),
"/US/US-CA/westhaven-moonstone-2306" => array("type"=>"cdp", "pop"=>1109, "name_ru"=>"Уэстхэвен-Мунстоун", "name_en"=>"Westhaven-Moonstone", "lang"=>"en", "lat"=>41.041972, "lon"=>-124.102836),
"/US/US-CA/westlake-village-2308" => array("type"=>"city", "pop"=>8469, "name_ru"=>"Уэстлейк Виллидж", "name_en"=>"Westlake Village", "lang"=>"en", "lat"=>34.136064, "lon"=>-118.821306),
"/US/US-CA/westminster-2310" => array("type"=>"city", "pop"=>88975, "name_ru"=>"Уэстминстер", "name_en"=>"Westminster", "lang"=>"en", "lat"=>33.751400, "lon"=>-117.993055),
"/US/US-CA/westmont-2311" => array("type"=>"cdp", "pop"=>31044, "name_ru"=>"Уэстмонт", "name_en"=>"Westmont", "lang"=>"en", "lat"=>33.941500, "lon"=>-118.301399),
"/US/US-CA/westmorland-2312" => array("type"=>"city", "pop"=>2197, "name_ru"=>"Уэстморленд", "name_en"=>"Westmorland", "lang"=>"en", "lat"=>33.036766, "lon"=>-115.621472),
"/US/US-CA/westwood-2317" => array("type"=>"cdp", "pop"=>2017, "name_ru"=>"Уэствуд", "name_en"=>"Westwood", "lang"=>"en", "lat"=>40.303167, "lon"=>-121.003924),
"/US/US-CA/wheatland-2318" => array("type"=>"city", "pop"=>3591, "name_ru"=>"Уитленд", "name_en"=>"Wheatland", "lang"=>"en", "lat"=>39.013025, "lon"=>-121.426749),
"/US/US-CA/whittier-2319" => array("type"=>"city", "pop"=>82267, "name_ru"=>"Уиттиер", "name_en"=>"Whittier", "lang"=>"en", "lat"=>33.966600, "lon"=>-118.021685),
"/US/US-CA/wildomar-2320" => array("type"=>"cdp", "pop"=>10411, "name_ru"=>"Уайлдомар", "name_en"=>"Wildomar", "lang"=>"en", "lat"=>33.609581, "lon"=>-117.251472),
"/US/US-CA/williams-2321" => array("type"=>"city", "pop"=>4760, "name_ru"=>"Уилльямс", "name_en"=>"Williams", "lang"=>"en", "lat"=>39.154581, "lon"=>-122.136795),
"/US/US-CA/willits-2322" => array("type"=>"city", "pop"=>4972, "name_ru"=>"Уиллитс", "name_en"=>"Willits", "lang"=>"en", "lat"=>39.405013, "lon"=>-123.347993),
"/US/US-CA/willow-creek-2324" => array("type"=>"cdp", "pop"=>1576, "name_ru"=>"Уиллоу Крик", "name_en"=>"Willow Creek", "lang"=>"en", "lat"=>40.882174, "lon"=>-123.665370),
"/US/US-CA/willowbrook-2323" => array("type"=>"cdp", "pop"=>32772, "name_ru"=>"Уиллоубрук", "name_en"=>"Willowbrook", "lang"=>"en", "lat"=>33.916950, "lon"=>-118.254149),
"/US/US-CA/willows-2325" => array("type"=>"city", "pop"=>6266, "name_ru"=>"Уиллоус", "name_en"=>"Willows", "lang"=>"en", "lat"=>39.515909, "lon"=>-122.198949),
"/US/US-CA/wilton-2326" => array("type"=>"cdp", "pop"=>3858, "name_ru"=>"Уилтон", "name_en"=>"Wilton", "lang"=>"en", "lat"=>38.419035, "lon"=>-121.207645),
"/US/US-CA/winchester-2327" => array("type"=>"cdp", "pop"=>1689, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>33.708900, "lon"=>-117.080649),
"/US/US-CA/windsor-2328" => array("type"=>"cdp", "pop"=>25362, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>38.545676, "lon"=>-122.797811),
"/US/US-CA/winters-2329" => array("type"=>"city", "pop"=>6977, "name_ru"=>"Уинтерс", "name_en"=>"Winters", "lang"=>"en", "lat"=>38.532088, "lon"=>-121.980245),
"/US/US-CA/winton-2330" => array("type"=>"cdp", "pop"=>7559, "name_ru"=>"Уинтон", "name_en"=>"Winton", "lang"=>"en", "lat"=>37.385554, "lon"=>-120.616341),
"/US/US-CA/wofford-heights-2331" => array("type"=>"cdp", "pop"=>2270, "name_ru"=>"Уоффорд Хайтс", "name_en"=>"Wofford Heights", "lang"=>"en", "lat"=>35.712640, "lon"=>-118.472769),
"/US/US-CA/woodacre-2332" => array("type"=>"cdp", "pop"=>1478, "name_ru"=>"Вудакр", "name_en"=>"Woodacre", "lang"=>"en", "lat"=>38.005750, "lon"=>-122.634683),
"/US/US-CA/woodbridge-2333" => array("type"=>"cdp", "pop"=>3456, "name_ru"=>"Вудбридж", "name_en"=>"Woodbridge", "lang"=>"en", "lat"=>38.163383, "lon"=>-121.317261),
"/US/US-CA/woodcrest-2334" => array("type"=>"cdp", "pop"=>7796, "name_ru"=>"Вудкрест", "name_en"=>"Woodcrest", "lang"=>"en", "lat"=>33.885850, "lon"=>-117.362024),
"/US/US-CA/woodlake-2335" => array("type"=>"city", "pop"=>7418, "name_ru"=>"Вудлейк", "name_en"=>"Woodlake", "lang"=>"en", "lat"=>36.412751, "lon"=>-119.097366),
"/US/US-CA/woodland-2336" => array("type"=>"city", "pop"=>54567, "name_ru"=>"Вудленд", "name_en"=>"Woodland", "lang"=>"en", "lat"=>38.676950, "lon"=>-121.766598),
"/US/US-CA/woodside-2337" => array("type"=>"town", "pop"=>5579, "name_ru"=>"Вудсайд", "name_en"=>"Woodside", "lang"=>"en", "lat"=>37.424783, "lon"=>-122.259273),
"/US/US-CA/woodville-2338" => array("type"=>"cdp", "pop"=>1557, "name_ru"=>"Вудвилл", "name_en"=>"Woodville", "lang"=>"en", "lat"=>36.090280, "lon"=>-119.203221),
"/US/US-CA/wrightwood-2339" => array("type"=>"cdp", "pop"=>3308, "name_ru"=>"Райтвуд", "name_en"=>"Wrightwood", "lang"=>"en", "lat"=>34.359400, "lon"=>-117.630309),
"/US/US-CA/yorba-linda-2340" => array("type"=>"city", "pop"=>65717, "name_ru"=>"Йорба Линда", "name_en"=>"Yorba Linda", "lang"=>"en", "lat"=>33.888661, "lon"=>-117.766326),
"/US/US-CA/yosemite-lakes-2341" => array("type"=>"cdp", "pop"=>2367, "name_ru"=>"Йосемити Лейкс", "name_en"=>"Yosemite Lakes", "lang"=>"en", "lat"=>37.191019, "lon"=>-119.771745),
"/US/US-CA/yountville-2342" => array("type"=>"town", "pop"=>3268, "name_ru"=>"Янтвилл", "name_en"=>"Yountville", "lang"=>"en", "lat"=>38.391353, "lon"=>-122.367805),
"/US/US-CA/yreka-2343" => array("type"=>"city", "pop"=>7368, "name_ru"=>"Ирика", "name_en"=>"Yreka", "lang"=>"en", "lat"=>41.729069, "lon"=>-122.631474),
"/US/US-CA/yuba-city-2344" => array("type"=>"city", "pop"=>61226, "name_ru"=>"Юба Сити", "name_en"=>"Yuba City", "lang"=>"en", "lat"=>39.135467, "lon"=>-121.622742),
"/US/US-CA/yucaipa-2345" => array("type"=>"city", "pop"=>49750, "name_ru"=>"Юкейпа", "name_en"=>"Yucaipa", "lang"=>"en", "lat"=>34.033050, "lon"=>-117.044449),
"/US/US-CA/yucca-valley-2346" => array("type"=>"cdp", "pop"=>20375, "name_ru"=>"Юкка Валли", "name_en"=>"Yucca Valley", "lang"=>"en", "lat"=>34.111240, "lon"=>-116.428552),
"/US/US-CO" => array("type"=>"state", "pop"=>4939456, "name_ru"=>"Колорадо", "name_en"=>"Colorado", "lang"=>"en"),
"/US/US-CO/aguilar-2347" => array("type"=>"town", "pop"=>570, "name_ru"=>"Агилар", "name_en"=>"Aguilar", "lang"=>"en", "lat"=>37.404138, "lon"=>-104.654367),
"/US/US-CO/air-force-academy-2348" => array("type"=>"cdp", "pop"=>9062, "name_ru"=>"Академия ВВС", "name_en"=>"Air Force Academy", "lang"=>"en", "lat"=>38.990448, "lon"=>-104.860139),
"/US/US-CO/akron-2349" => array("type"=>"town", "pop"=>1651, "name_ru"=>"Акрон", "name_en"=>"Akron", "lang"=>"en", "lat"=>40.162578, "lon"=>-103.211720),
"/US/US-CO/alamosa-2350" => array("type"=>"city", "pop"=>8745, "name_ru"=>"Аламоса", "name_en"=>"Alamosa", "lang"=>"en", "lat"=>37.468831, "lon"=>-105.873109),
"/US/US-CO/alamosa-east-2351" => array("type"=>"cdp", "pop"=>1389, "name_ru"=>"Аламоса Ист", "name_en"=>"Alamosa East", "lang"=>"en", "lat"=>37.477165, "lon"=>-105.842755),
"/US/US-CO/alma-2352" => array("type"=>"town", "pop"=>172, "name_ru"=>"Алма", "name_en"=>"Alma", "lang"=>"en", "lat"=>39.285770, "lon"=>-106.064709),
"/US/US-CO/antonito-2353" => array("type"=>"town", "pop"=>788, "name_ru"=>"Антонито", "name_en"=>"Antonito", "lang"=>"en", "lat"=>37.076460, "lon"=>-106.010155),
"/US/US-CO/applewood-2354" => array("type"=>"cdp", "pop"=>11069, "name_ru"=>"Эпплвуд", "name_en"=>"Applewood", "lang"=>"en", "lat"=>39.747242, "lon"=>-105.162296),
"/US/US-CO/arriba-2355" => array("type"=>"town", "pop"=>209, "name_ru"=>"Арриба", "name_en"=>"Arriba", "lang"=>"en", "lat"=>39.283804, "lon"=>-103.273512),
"/US/US-CO/arvada-2356" => array("type"=>"city", "pop"=>107361, "name_ru"=>"Арвада", "name_en"=>"Arvada", "lang"=>"en", "lat"=>39.821125, "lon"=>-105.106744),
"/US/US-CO/aspen-2357" => array("type"=>"city", "pop"=>5902, "name_ru"=>"Аспен", "name_en"=>"Aspen", "lang"=>"en", "lat"=>39.194552, "lon"=>-106.826851),
"/US/US-CO/ault-2358" => array("type"=>"town", "pop"=>1442, "name_ru"=>"Олт", "name_en"=>"Ault", "lang"=>"en", "lat"=>40.584019, "lon"=>-104.733126),
"/US/US-CO/aurora-2359" => array("type"=>"city", "pop"=>319057, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>39.712267, "lon"=>-104.729772),
"/US/US-CO/avon-2360" => array("type"=>"town", "pop"=>6578, "name_ru"=>"Эйвон", "name_en"=>"Avon", "lang"=>"en", "lat"=>39.652469, "lon"=>-106.533055),
"/US/US-CO/basalt-2361" => array("type"=>"town", "pop"=>3244, "name_ru"=>"Базальт", "name_en"=>"Basalt", "lang"=>"en", "lat"=>39.366732, "lon"=>-107.026970),
"/US/US-CO/battlement-mesa-2362" => array("type"=>"cdp", "pop"=>1477, "name_ru"=>"Бэтлмент Меса", "name_en"=>"Battlement Mesa", "lang"=>"en", "lat"=>39.441465, "lon"=>-108.024526),
"/US/US-CO/bayfield-2363" => array("type"=>"town", "pop"=>2024, "name_ru"=>"Бэйфилд", "name_en"=>"Bayfield", "lang"=>"en", "lat"=>37.234174, "lon"=>-107.594904),
"/US/US-CO/bennett-2364" => array("type"=>"town", "pop"=>2595, "name_ru"=>"Беннет", "name_en"=>"Bennett", "lang"=>"en", "lat"=>39.755144, "lon"=>-104.423413),
"/US/US-CO/berthoud-2365" => array("type"=>"town", "pop"=>5429, "name_ru"=>"Бертхауд", "name_en"=>"Berthoud", "lang"=>"en", "lat"=>40.305700, "lon"=>-105.081199),
"/US/US-CO/bethune-2366" => array("type"=>"town", "pop"=>209, "name_ru"=>"Бетун", "name_en"=>"Bethune", "lang"=>"en", "lat"=>39.303480, "lon"=>-102.422883),
"/US/US-CO/black-forest-2367" => array("type"=>"cdp", "pop"=>8143, "name_ru"=>"Блэк Форест", "name_en"=>"Black Forest", "lang"=>"en", "lat"=>39.046325, "lon"=>-104.668133),
"/US/US-CO/black-hawk-2368" => array("type"=>"town", "pop"=>105, "name_ru"=>"Блэк Хок", "name_en"=>"Black Hawk", "lang"=>"en", "lat"=>39.800302, "lon"=>-105.491285),
"/US/US-CO/blanca-2369" => array("type"=>"town", "pop"=>342, "name_ru"=>"Бланка", "name_en"=>"Blanca", "lang"=>"en", "lat"=>37.438857, "lon"=>-105.511948),
"/US/US-CO/blue-river-2370" => array("type"=>"town", "pop"=>954, "name_ru"=>"Блу Ривер", "name_en"=>"Blue River", "lang"=>"en", "lat"=>39.432545, "lon"=>-106.036578),
"/US/US-CO/bonanza-city-2371" => array("type"=>"town", "pop"=>16, "name_ru"=>"Бонанза Сити", "name_en"=>"Bonanza City", "lang"=>"en", "lat"=>38.294147, "lon"=>-106.141525),
"/US/US-CO/boone-2372" => array("type"=>"town", "pop"=>335, "name_ru"=>"Бун", "name_en"=>"Boone", "lang"=>"en", "lat"=>38.250072, "lon"=>-104.255768),
"/US/US-CO/boulder-2373" => array("type"=>"city", "pop"=>94171, "name_ru"=>"Боулдер", "name_en"=>"Boulder", "lang"=>"en", "lat"=>40.026881, "lon"=>-105.251025),
"/US/US-CO/bow-mar-2374" => array("type"=>"town", "pop"=>811, "name_ru"=>"Боу Мар", "name_en"=>"Bow Mar", "lang"=>"en", "lat"=>39.625800, "lon"=>-105.052054),
"/US/US-CO/branson-2375" => array("type"=>"town", "pop"=>77, "name_ru"=>"Брэнсон", "name_en"=>"Branson", "lang"=>"en", "lat"=>37.014581, "lon"=>-103.879112),
"/US/US-CO/breckenridge-2376" => array("type"=>"town", "pop"=>3420, "name_ru"=>"Брекенридж", "name_en"=>"Breckenridge", "lang"=>"en", "lat"=>39.507468, "lon"=>-106.047373),
"/US/US-CO/brighton-2377" => array("type"=>"city", "pop"=>31380, "name_ru"=>"Брайтон", "name_en"=>"Brighton", "lang"=>"en", "lat"=>39.957980, "lon"=>-104.794811),
"/US/US-CO/brookside-2378" => array("type"=>"town", "pop"=>212, "name_ru"=>"Бруксайд", "name_en"=>"Brookside", "lang"=>"en", "lat"=>38.413515, "lon"=>-105.191478),
"/US/US-CO/broomfield-2379" => array("type"=>"city", "pop"=>54858, "name_ru"=>"Брумфилд", "name_en"=>"Broomfield", "lang"=>"en", "lat"=>39.944000, "lon"=>-105.061654),
"/US/US-CO/brush-2380" => array("type"=>"city", "pop"=>5332, "name_ru"=>"Браш", "name_en"=>"Brush", "lang"=>"en", "lat"=>40.256755, "lon"=>-103.632300),
"/US/US-CO/buena-vista-2381" => array("type"=>"town", "pop"=>2134, "name_ru"=>"Буена Виста", "name_en"=>"Buena Vista", "lang"=>"en", "lat"=>38.832750, "lon"=>-106.141252),
"/US/US-CO/burlington-2382" => array("type"=>"city", "pop"=>3875, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>39.305441, "lon"=>-102.272190),
"/US/US-CO/byers-2383" => array("type"=>"cdp", "pop"=>1065, "name_ru"=>"Байерс", "name_en"=>"Byers", "lang"=>"en", "lat"=>39.710306, "lon"=>-104.224733),
"/US/US-CO/calhan-2384" => array("type"=>"town", "pop"=>868, "name_ru"=>"Калхэн", "name_en"=>"Calhan", "lang"=>"en", "lat"=>39.034421, "lon"=>-104.299323),
"/US/US-CO/campion-2385" => array("type"=>"cdp", "pop"=>1692, "name_ru"=>"Кэмпион", "name_en"=>"Campion", "lang"=>"en", "lat"=>40.344675, "lon"=>-105.094985),
"/US/US-CO/campo-2386" => array("type"=>"town", "pop"=>127, "name_ru"=>"Кампо", "name_en"=>"Campo", "lang"=>"en", "lat"=>37.104713, "lon"=>-102.577924),
"/US/US-CO/canon-city-2387" => array("type"=>"city", "pop"=>12687, "name_ru"=>"Кэнон Сити", "name_en"=>"Canon City", "lang"=>"en", "lat"=>38.441366, "lon"=>-105.234377),
"/US/US-CO/carbondale-2388" => array("type"=>"town", "pop"=>6266, "name_ru"=>"Карбондейл", "name_en"=>"Carbondale", "lang"=>"en", "lat"=>39.397260, "lon"=>-107.214044),
"/US/US-CO/cascade-chipita-park-2389" => array("type"=>"cdp", "pop"=>1479, "name_ru"=>"Каскейд-Чипита Парк", "name_en"=>"Cascade-Chipita Park", "lang"=>"en", "lat"=>38.951802, "lon"=>-105.002520),
"/US/US-CO/castle-rock-2390" => array("type"=>"city", "pop"=>44369, "name_ru"=>"Касл Рок", "name_en"=>"Castle Rock", "lang"=>"en", "lat"=>39.378872, "lon"=>-104.851305),
"/US/US-CO/castlewood-2391" => array("type"=>"cdp", "pop"=>24392, "name_ru"=>"Каслвуд", "name_en"=>"Castlewood", "lang"=>"en", "lat"=>39.584733, "lon"=>-104.900513),
"/US/US-CO/cedaredge-2392" => array("type"=>"town", "pop"=>2252, "name_ru"=>"Сидаредж", "name_en"=>"Cedaredge", "lang"=>"en", "lat"=>38.898784, "lon"=>-107.926338),
"/US/US-CO/center-2393" => array("type"=>"town", "pop"=>2389, "name_ru"=>"Сентер", "name_en"=>"Center", "lang"=>"en", "lat"=>37.751832, "lon"=>-106.109902),
"/US/US-CO/central-city-2394" => array("type"=>"city", "pop"=>566, "name_ru"=>"Сентрал Сити", "name_en"=>"Central City", "lang"=>"en", "lat"=>39.799812, "lon"=>-105.512389),
"/US/US-CO/cheraw-2395" => array("type"=>"town", "pop"=>201, "name_ru"=>"Черо", "name_en"=>"Cheraw", "lang"=>"en", "lat"=>38.107786, "lon"=>-103.510653),
"/US/US-CO/cherry-hills-village-2396" => array("type"=>"city", "pop"=>6355, "name_ru"=>"Черри Хилс Виллидж", "name_en"=>"Cherry Hills Village", "lang"=>"en", "lat"=>39.637450, "lon"=>-104.946916),
"/US/US-CO/cheyenne-wells-2397" => array("type"=>"town", "pop"=>786, "name_ru"=>"Шайенн Уэлс", "name_en"=>"Cheyenne Wells", "lang"=>"en", "lat"=>38.819351, "lon"=>-102.350803),
"/US/US-CO/cimarron-hills-2398" => array("type"=>"cdp", "pop"=>11160, "name_ru"=>"Симаррон Хилс", "name_en"=>"Cimarron Hills", "lang"=>"en", "lat"=>38.858598, "lon"=>-104.698423),
"/US/US-CO/clifton-2399" => array("type"=>"cdp", "pop"=>12671, "name_ru"=>"Слифтон", "name_en"=>"Clifton", "lang"=>"en", "lat"=>39.075250, "lon"=>-108.463374),
"/US/US-CO/coal-creek-2400" => array("type"=>"town", "pop"=>355, "name_ru"=>"Коал Крик", "name_en"=>"Coal Creek", "lang"=>"en", "lat"=>38.359508, "lon"=>-105.145133),
"/US/US-CO/cokedale-2401" => array("type"=>"town", "pop"=>134, "name_ru"=>"Коукдейл", "name_en"=>"Cokedale", "lang"=>"en", "lat"=>37.143358, "lon"=>-104.621507),
"/US/US-CO/collbran-2402" => array("type"=>"town", "pop"=>426, "name_ru"=>"Коллбран", "name_en"=>"Collbran", "lang"=>"en", "lat"=>39.239292, "lon"=>-107.958647),
"/US/US-CO/colorado-city-2403" => array("type"=>"cdp", "pop"=>1149, "name_ru"=>"Колорадо Сити", "name_en"=>"Colorado City", "lang"=>"en", "lat"=>37.944042, "lon"=>-104.846291),
"/US/US-CO/colorado-springs-2404" => array("type"=>"city", "pop"=>380307, "name_ru"=>"Колорадо Спрингс", "name_en"=>"Colorado Springs", "lang"=>"en", "lat"=>38.863200, "lon"=>-104.759899),
"/US/US-CO/columbine-2405" => array("type"=>"cdp", "pop"=>23969, "name_ru"=>"Колумбайн", "name_en"=>"Columbine", "lang"=>"en", "lat"=>39.587900, "lon"=>-105.068792),
"/US/US-CO/columbine-valley-2406" => array("type"=>"town", "pop"=>1320, "name_ru"=>"Колумбайн Валли", "name_en"=>"Columbine Valley", "lang"=>"en", "lat"=>39.598150, "lon"=>-105.033638),
"/US/US-CO/commerce-city-2407" => array("type"=>"city", "pop"=>42473, "name_ru"=>"Коммерс Сити", "name_en"=>"Commerce City", "lang"=>"en", "lat"=>39.867602, "lon"=>-104.868393),
"/US/US-CO/cortez-2408" => array("type"=>"city", "pop"=>8609, "name_ru"=>"Кортез", "name_en"=>"Cortez", "lang"=>"en", "lat"=>37.351531, "lon"=>-108.577303),
"/US/US-CO/craig-2409" => array("type"=>"city", "pop"=>9241, "name_ru"=>"Крейг", "name_en"=>"Craig", "lang"=>"en", "lat"=>40.516925, "lon"=>-107.553380),
"/US/US-CO/crawford-2410" => array("type"=>"town", "pop"=>392, "name_ru"=>"Крофорд", "name_en"=>"Crawford", "lang"=>"en", "lat"=>38.705087, "lon"=>-107.608638),
"/US/US-CO/creede-2411" => array("type"=>"town", "pop"=>421, "name_ru"=>"Крид", "name_en"=>"Creede", "lang"=>"en", "lat"=>37.850621, "lon"=>-106.926477),
"/US/US-CO/crested-butte-2412" => array("type"=>"town", "pop"=>1651, "name_ru"=>"Крестед Бьютт", "name_en"=>"Crested Butte", "lang"=>"en", "lat"=>38.870112, "lon"=>-106.983102),
"/US/US-CO/crestone-2413" => array("type"=>"town", "pop"=>107, "name_ru"=>"Крестоун", "name_en"=>"Crestone", "lang"=>"en", "lat"=>37.994985, "lon"=>-105.697286),
"/US/US-CO/cripple-creek-2414" => array("type"=>"city", "pop"=>1012, "name_ru"=>"Крипл Крик", "name_en"=>"Cripple Creek", "lang"=>"en", "lat"=>38.746299, "lon"=>-105.178709),
"/US/US-CO/crook-2415" => array("type"=>"town", "pop"=>124, "name_ru"=>"Крук", "name_en"=>"Crook", "lang"=>"en", "lat"=>40.858329, "lon"=>-102.801080),
"/US/US-CO/crowley-2416" => array("type"=>"town", "pop"=>165, "name_ru"=>"Кроули", "name_en"=>"Crowley", "lang"=>"en", "lat"=>38.193585, "lon"=>-103.858475),
"/US/US-CO/dacono-2417" => array("type"=>"city", "pop"=>4052, "name_ru"=>"Даконо", "name_en"=>"Dacono", "lang"=>"en", "lat"=>40.075775, "lon"=>-104.942909),
"/US/US-CO/de-beque-2418" => array("type"=>"town", "pop"=>522, "name_ru"=>"Де Бек", "name_en"=>"De Beque", "lang"=>"en", "lat"=>39.332505, "lon"=>-108.213659),
"/US/US-CO/deer-trail-2419" => array("type"=>"town", "pop"=>577, "name_ru"=>"Дир Трейл", "name_en"=>"Deer Trail", "lang"=>"en", "lat"=>39.616879, "lon"=>-104.041790),
"/US/US-CO/del-norte-2420" => array("type"=>"town", "pop"=>1603, "name_ru"=>"Дель Норте", "name_en"=>"Del Norte", "lang"=>"en", "lat"=>37.678262, "lon"=>-106.352143),
"/US/US-CO/delta-2421" => array("type"=>"city", "pop"=>9020, "name_ru"=>"Дельта", "name_en"=>"Delta", "lang"=>"en", "lat"=>38.744765, "lon"=>-108.073686),
"/US/US-CO/denver-2422" => array("type"=>"city", "pop"=>598707, "name_ru"=>"Денвер", "name_en"=>"Denver", "lang"=>"en", "lat"=>39.768035, "lon"=>-104.872655),
"/US/US-CO/derby-2423" => array("type"=>"cdp", "pop"=>6043, "name_ru"=>"Дерби", "name_en"=>"Derby", "lang"=>"en", "lat"=>39.839450, "lon"=>-104.918076),
"/US/US-CO/dillon-2424" => array("type"=>"town", "pop"=>808, "name_ru"=>"Диллон", "name_en"=>"Dillon", "lang"=>"en", "lat"=>39.624998, "lon"=>-106.039009),
"/US/US-CO/dinosaur-2425" => array("type"=>"town", "pop"=>335, "name_ru"=>"Дайносаер", "name_en"=>"Dinosaur", "lang"=>"en", "lat"=>40.241560, "lon"=>-109.008045),
"/US/US-CO/dolores-2426" => array("type"=>"town", "pop"=>920, "name_ru"=>"Долорес", "name_en"=>"Dolores", "lang"=>"en", "lat"=>37.474548, "lon"=>-108.497648),
"/US/US-CO/dove-creek-2427" => array("type"=>"town", "pop"=>745, "name_ru"=>"Дав Крик", "name_en"=>"Dove Creek", "lang"=>"en", "lat"=>37.765988, "lon"=>-108.905230),
"/US/US-CO/durango-2428" => array("type"=>"city", "pop"=>16416, "name_ru"=>"Дуранго", "name_en"=>"Durango", "lang"=>"en", "lat"=>37.285986, "lon"=>-107.869381),
"/US/US-CO/eads-2429" => array("type"=>"town", "pop"=>606, "name_ru"=>"Идс", "name_en"=>"Eads", "lang"=>"en", "lat"=>38.481474, "lon"=>-102.779450),
"/US/US-CO/eagle-2430" => array("type"=>"town", "pop"=>5897, "name_ru"=>"Игл", "name_en"=>"Eagle", "lang"=>"en", "lat"=>39.655644, "lon"=>-106.825441),
"/US/US-CO/eagle-vail-2431" => array("type"=>"cdp", "pop"=>1922, "name_ru"=>"Игл-Вейл", "name_en"=>"Eagle-Vail", "lang"=>"en", "lat"=>39.621987, "lon"=>-106.489146),
"/US/US-CO/eaton-2432" => array("type"=>"town", "pop"=>4213, "name_ru"=>"Итон", "name_en"=>"Eaton", "lang"=>"en", "lat"=>40.524877, "lon"=>-104.709823),
"/US/US-CO/eckley-2433" => array("type"=>"town", "pop"=>265, "name_ru"=>"Экли", "name_en"=>"Eckley", "lang"=>"en", "lat"=>40.111968, "lon"=>-102.488194),
"/US/US-CO/edgewater-2434" => array("type"=>"city", "pop"=>5136, "name_ru"=>"Эджвотер", "name_en"=>"Edgewater", "lang"=>"en", "lat"=>39.750350, "lon"=>-105.062649),
"/US/US-CO/el-jebel-2436" => array("type"=>"cdp", "pop"=>2605, "name_ru"=>"Эль Джебел", "name_en"=>"El Jebel", "lang"=>"en", "lat"=>39.399809, "lon"=>-107.089132),
"/US/US-CO/elizabeth-2435" => array("type"=>"town", "pop"=>1436, "name_ru"=>"Элизабет", "name_en"=>"Elizabeth", "lang"=>"en", "lat"=>39.360129, "lon"=>-104.599531),
"/US/US-CO/empire-2437" => array("type"=>"town", "pop"=>324, "name_ru"=>"Эмпаер", "name_en"=>"Empire", "lang"=>"en", "lat"=>39.759698, "lon"=>-105.682597),
"/US/US-CO/englewood-2438" => array("type"=>"city", "pop"=>32669, "name_ru"=>"Энглвуд", "name_en"=>"Englewood", "lang"=>"en", "lat"=>39.647417, "lon"=>-104.993841),
"/US/US-CO/erie-2439" => array("type"=>"town", "pop"=>16432, "name_ru"=>"Эри", "name_en"=>"Erie", "lang"=>"en", "lat"=>40.028855, "lon"=>-105.046033),
"/US/US-CO/estes-park-2440" => array("type"=>"town", "pop"=>6432, "name_ru"=>"Эстес Парк", "name_en"=>"Estes Park", "lang"=>"en", "lat"=>40.372060, "lon"=>-105.520972),
"/US/US-CO/evans-2441" => array("type"=>"city", "pop"=>18842, "name_ru"=>"Эванс", "name_en"=>"Evans", "lang"=>"en", "lat"=>40.377868, "lon"=>-104.696799),
"/US/US-CO/evergreen-2442" => array("type"=>"cdp", "pop"=>7582, "name_ru"=>"Эвергрин", "name_en"=>"Evergreen", "lang"=>"en", "lat"=>39.633684, "lon"=>-105.342303),
"/US/US-CO/fairplay-2443" => array("type"=>"town", "pop"=>658, "name_ru"=>"Фэрплей", "name_en"=>"Fairplay", "lang"=>"en", "lat"=>39.223809, "lon"=>-105.999055),
"/US/US-CO/federal-heights-2444" => array("type"=>"city", "pop"=>11732, "name_ru"=>"Федерал Хайтс", "name_en"=>"Federal Heights", "lang"=>"en", "lat"=>39.865150, "lon"=>-105.015871),
"/US/US-CO/firestone-2445" => array("type"=>"town", "pop"=>8117, "name_ru"=>"Файерстоун", "name_en"=>"Firestone", "lang"=>"en", "lat"=>40.120928, "lon"=>-104.934002),
"/US/US-CO/flagler-2446" => array("type"=>"town", "pop"=>567, "name_ru"=>"Флаглер", "name_en"=>"Flagler", "lang"=>"en", "lat"=>39.294041, "lon"=>-103.065341),
"/US/US-CO/fleming-2447" => array("type"=>"town", "pop"=>434, "name_ru"=>"Флеминг", "name_en"=>"Fleming", "lang"=>"en", "lat"=>40.681179, "lon"=>-102.839804),
"/US/US-CO/florence-2448" => array("type"=>"city", "pop"=>3625, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>38.385241, "lon"=>-105.116388),
"/US/US-CO/fort-carson-2449" => array("type"=>"cdp", "pop"=>11309, "name_ru"=>"Форт Карсон", "name_en"=>"Fort Carson", "lang"=>"en", "lat"=>38.741967, "lon"=>-104.782218),
"/US/US-CO/fort-collins-2450" => array("type"=>"city", "pop"=>136509, "name_ru"=>"Форт Коллинс", "name_en"=>"Fort Collins", "lang"=>"en", "lat"=>40.555550, "lon"=>-105.068481),
"/US/US-CO/fort-lupton-2451" => array("type"=>"city", "pop"=>7620, "name_ru"=>"Форт Луптон", "name_en"=>"Fort Lupton", "lang"=>"en", "lat"=>40.085241, "lon"=>-104.805071),
"/US/US-CO/fort-morgan-2452" => array("type"=>"city", "pop"=>10539, "name_ru"=>"Форт Морган", "name_en"=>"Fort Morgan", "lang"=>"en", "lat"=>40.265184, "lon"=>-103.794571),
"/US/US-CO/fountain-2453" => array("type"=>"city", "pop"=>19669, "name_ru"=>"Фонтейн", "name_en"=>"Fountain", "lang"=>"en", "lat"=>38.666888, "lon"=>-104.693286),
"/US/US-CO/fowler-2454" => array("type"=>"town", "pop"=>1094, "name_ru"=>"Фоулер", "name_en"=>"Fowler", "lang"=>"en", "lat"=>38.128338, "lon"=>-104.024492),
"/US/US-CO/fraser-2455" => array("type"=>"town", "pop"=>927, "name_ru"=>"Фрейзер", "name_en"=>"Fraser", "lang"=>"en", "lat"=>39.936999, "lon"=>-105.794446),
"/US/US-CO/frederick-2456" => array("type"=>"town", "pop"=>8127, "name_ru"=>"Фредерик", "name_en"=>"Frederick", "lang"=>"en", "lat"=>40.104972, "lon"=>-104.942424),
"/US/US-CO/frisco-2457" => array("type"=>"town", "pop"=>2696, "name_ru"=>"Фриско", "name_en"=>"Frisco", "lang"=>"en", "lat"=>39.579064, "lon"=>-106.096038),
"/US/US-CO/fruita-2458" => array("type"=>"city", "pop"=>7418, "name_ru"=>"Фрута", "name_en"=>"Fruita", "lang"=>"en", "lat"=>39.156510, "lon"=>-108.727298),
"/US/US-CO/fruitvale-2459" => array("type"=>"cdp", "pop"=>5222, "name_ru"=>"Фрутвейл", "name_en"=>"Fruitvale", "lang"=>"en", "lat"=>39.093200, "lon"=>-108.478371),
"/US/US-CO/garden-city-2460" => array("type"=>"town", "pop"=>337, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>40.393950, "lon"=>-104.688726),
"/US/US-CO/gateway-2461" => array("type"=>"cdp", "pop"=>7510, "name_ru"=>"Гейтуэй", "name_en"=>"Gateway", "lang"=>"en", "lat"=>39.549302, "lon"=>-104.905517),
"/US/US-CO/genesee-2462" => array("type"=>"cdp", "pop"=>2737, "name_ru"=>"Джинеси", "name_en"=>"Genesee", "lang"=>"en", "lat"=>39.685750, "lon"=>-105.272189),
"/US/US-CO/genoa-2463" => array("type"=>"town", "pop"=>180, "name_ru"=>"Дженоа", "name_en"=>"Genoa", "lang"=>"en", "lat"=>39.278439, "lon"=>-103.499412),
"/US/US-CO/georgetown-2464" => array("type"=>"town", "pop"=>1027, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>39.715212, "lon"=>-105.695987),
"/US/US-CO/gilcrest-2465" => array("type"=>"town", "pop"=>1151, "name_ru"=>"Джилкрест", "name_en"=>"Gilcrest", "lang"=>"en", "lat"=>40.282664, "lon"=>-104.781950),
"/US/US-CO/glendale-2466" => array("type"=>"city", "pop"=>4798, "name_ru"=>"Глендейл", "name_en"=>"Glendale", "lang"=>"en", "lat"=>39.704117, "lon"=>-104.934633),
"/US/US-CO/gleneagle-2467" => array("type"=>"cdp", "pop"=>1661, "name_ru"=>"Гленигл", "name_en"=>"Gleneagle", "lang"=>"en", "lat"=>39.045211, "lon"=>-104.823862),
"/US/US-CO/glenwood-springs-2468" => array("type"=>"city", "pop"=>9053, "name_ru"=>"Гленвуд Спрингс", "name_en"=>"Glenwood Springs", "lang"=>"en", "lat"=>39.544926, "lon"=>-107.330137),
"/US/US-CO/golden-2469" => array("type"=>"city", "pop"=>17321, "name_ru"=>"Гоулден", "name_en"=>"Golden", "lang"=>"en", "lat"=>39.738867, "lon"=>-105.215503),
"/US/US-CO/granada-2470" => array("type"=>"town", "pop"=>580, "name_ru"=>"Гранада", "name_en"=>"Granada", "lang"=>"en", "lat"=>38.063695, "lon"=>-102.310550),
"/US/US-CO/granby-2471" => array("type"=>"town", "pop"=>1662, "name_ru"=>"Грэнби", "name_en"=>"Granby", "lang"=>"en", "lat"=>40.088444, "lon"=>-105.937247),
"/US/US-CO/grand-junction-2472" => array("type"=>"city", "pop"=>49688, "name_ru"=>"Гранд Джанкшен", "name_en"=>"Grand Junction", "lang"=>"en", "lat"=>39.087300, "lon"=>-108.552924),
"/US/US-CO/grand-lake-2473" => array("type"=>"town", "pop"=>433, "name_ru"=>"Гранд Лейк", "name_en"=>"Grand Lake", "lang"=>"en", "lat"=>40.251171, "lon"=>-105.821384),
"/US/US-CO/greeley-2474" => array("type"=>"city", "pop"=>91492, "name_ru"=>"Грили", "name_en"=>"Greeley", "lang"=>"en", "lat"=>40.421324, "lon"=>-104.743871),
"/US/US-CO/green-mountain-falls-2475" => array("type"=>"town", "pop"=>803, "name_ru"=>"Грин Маунтин Фолс", "name_en"=>"Green Mountain Falls", "lang"=>"en", "lat"=>38.934150, "lon"=>-105.019668),
"/US/US-CO/greenwood-village-2476" => array("type"=>"city", "pop"=>14230, "name_ru"=>"Гринвуд Виллидж", "name_en"=>"Greenwood Village", "lang"=>"en", "lat"=>39.615533, "lon"=>-104.914957),
"/US/US-CO/grover-2477" => array("type"=>"town", "pop"=>154, "name_ru"=>"Гровер", "name_en"=>"Grover", "lang"=>"en", "lat"=>40.869176, "lon"=>-104.225602),
"/US/US-CO/gunbarrel-2478" => array("type"=>"cdp", "pop"=>9388, "name_ru"=>"Ганбарел", "name_en"=>"Gunbarrel", "lang"=>"en", "lat"=>40.065533, "lon"=>-105.171141),
"/US/US-CO/gunnison-2479" => array("type"=>"city", "pop"=>5461, "name_ru"=>"Ганнисон", "name_en"=>"Gunnison", "lang"=>"en", "lat"=>38.544702, "lon"=>-106.927045),
"/US/US-CO/gypsum-2480" => array("type"=>"town", "pop"=>6052, "name_ru"=>"Джипсум", "name_en"=>"Gypsum", "lang"=>"en", "lat"=>39.643423, "lon"=>-106.947734),
"/US/US-CO/hartman-2481" => array("type"=>"town", "pop"=>102, "name_ru"=>"Хартман", "name_en"=>"Hartman", "lang"=>"en", "lat"=>38.120894, "lon"=>-102.218059),
"/US/US-CO/haswell-2482" => array("type"=>"town", "pop"=>70, "name_ru"=>"Хасуэлл", "name_en"=>"Haswell", "lang"=>"en", "lat"=>38.452800, "lon"=>-103.163942),
"/US/US-CO/haxtun-2483" => array("type"=>"town", "pop"=>964, "name_ru"=>"Хакстун", "name_en"=>"Haxtun", "lang"=>"en", "lat"=>40.642719, "lon"=>-102.628787),
"/US/US-CO/hayden-2484" => array("type"=>"town", "pop"=>1579, "name_ru"=>"Хейден", "name_en"=>"Hayden", "lang"=>"en", "lat"=>40.488087, "lon"=>-107.260318),
"/US/US-CO/highlands-ranch-2485" => array("type"=>"cdp", "pop"=>10181, "name_ru"=>"Хайлендс Ранч", "name_en"=>"Highlands Ranch", "lang"=>"en", "lat"=>39.553962, "lon"=>-104.968901),
"/US/US-CO/hillrose-2486" => array("type"=>"town", "pop"=>270, "name_ru"=>"Хиллроуз", "name_en"=>"Hillrose", "lang"=>"en", "lat"=>40.324516, "lon"=>-103.522211),
"/US/US-CO/holly-2487" => array("type"=>"town", "pop"=>961, "name_ru"=>"Холли", "name_en"=>"Holly", "lang"=>"en", "lat"=>38.055625, "lon"=>-102.123988),
"/US/US-CO/holyoke-2488" => array("type"=>"city", "pop"=>2223, "name_ru"=>"Хольок", "name_en"=>"Holyoke", "lang"=>"en", "lat"=>40.584939, "lon"=>-102.303710),
"/US/US-CO/hooper-2489" => array("type"=>"town", "pop"=>124, "name_ru"=>"Хупер", "name_en"=>"Hooper", "lang"=>"en", "lat"=>37.745798, "lon"=>-105.877567),
"/US/US-CO/hot-sulphur-springs-2491" => array("type"=>"town", "pop"=>512, "name_ru"=>"Хот Салфер Спрингс", "name_en"=>"Hot Sulphur Springs", "lang"=>"en", "lat"=>40.073716, "lon"=>-106.099090),
"/US/US-CO/hotchkiss-2490" => array("type"=>"town", "pop"=>1084, "name_ru"=>"Хотчкисс", "name_en"=>"Hotchkiss", "lang"=>"en", "lat"=>38.801029, "lon"=>-107.717908),
"/US/US-CO/hudson-2492" => array("type"=>"town", "pop"=>1617, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>40.070976, "lon"=>-104.642904),
"/US/US-CO/hugo-2493" => array("type"=>"town", "pop"=>742, "name_ru"=>"Хьюго", "name_en"=>"Hugo", "lang"=>"en", "lat"=>39.134400, "lon"=>-103.467106),
"/US/US-CO/idaho-springs-2494" => array("type"=>"city", "pop"=>1736, "name_ru"=>"Айдахо Спрингс", "name_en"=>"Idaho Springs", "lang"=>"en", "lat"=>39.740963, "lon"=>-105.511887),
"/US/US-CO/ignacio-2495" => array("type"=>"town", "pop"=>767, "name_ru"=>"Игнасио", "name_en"=>"Ignacio", "lang"=>"en", "lat"=>37.116767, "lon"=>-107.634403),
"/US/US-CO/iliff-2496" => array("type"=>"town", "pop"=>216, "name_ru"=>"Илифф", "name_en"=>"Iliff", "lang"=>"en", "lat"=>40.757938, "lon"=>-103.065366),
"/US/US-CO/jamestown-2497" => array("type"=>"town", "pop"=>213, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>40.116133, "lon"=>-105.389630),
"/US/US-CO/johnstown-2498" => array("type"=>"town", "pop"=>9309, "name_ru"=>"Джонстаун", "name_en"=>"Johnstown", "lang"=>"en", "lat"=>40.335387, "lon"=>-104.911484),
"/US/US-CO/julesburg-2499" => array("type"=>"town", "pop"=>1252, "name_ru"=>"Джулсбург", "name_en"=>"Julesburg", "lang"=>"en", "lat"=>40.986738, "lon"=>-102.264581),
"/US/US-CO/keenesburg-2500" => array("type"=>"town", "pop"=>1177, "name_ru"=>"Кинсбург", "name_en"=>"Keenesburg", "lang"=>"en", "lat"=>40.108872, "lon"=>-104.518571),
"/US/US-CO/ken-caryl-2501" => array("type"=>"cdp", "pop"=>24391, "name_ru"=>"Кен Карил", "name_en"=>"Ken Caryl", "lang"=>"en", "lat"=>39.575901, "lon"=>-105.111688),
"/US/US-CO/keota-2502" => array("type"=>"town", "pop"=>5, "name_ru"=>"Кеота", "name_en"=>"Keota", "lang"=>"en", "lat"=>40.704059, "lon"=>-104.073720),
"/US/US-CO/kersey-2503" => array("type"=>"town", "pop"=>1460, "name_ru"=>"Керси", "name_en"=>"Kersey", "lang"=>"en", "lat"=>40.386088, "lon"=>-104.559763),
"/US/US-CO/kim-2504" => array("type"=>"town", "pop"=>65, "name_ru"=>"Ким", "name_en"=>"Kim", "lang"=>"en", "lat"=>37.246271, "lon"=>-103.351836),
"/US/US-CO/kiowa-2505" => array("type"=>"town", "pop"=>596, "name_ru"=>"Кайова", "name_en"=>"Kiowa", "lang"=>"en", "lat"=>39.345252, "lon"=>-104.460999),
"/US/US-CO/kit-carson-2506" => array("type"=>"town", "pop"=>202, "name_ru"=>"Кит Карсон", "name_en"=>"Kit Carson", "lang"=>"en", "lat"=>38.763094, "lon"=>-102.795696),
"/US/US-CO/kremmling-2507" => array("type"=>"town", "pop"=>1528, "name_ru"=>"Креммлинг", "name_en"=>"Kremmling", "lang"=>"en", "lat"=>40.056440, "lon"=>-106.376270),
"/US/US-CO/la-jara-2509" => array("type"=>"town", "pop"=>792, "name_ru"=>"Ла Джара", "name_en"=>"La Jara", "lang"=>"en", "lat"=>37.274397, "lon"=>-105.959494),
"/US/US-CO/la-junta-2510" => array("type"=>"city", "pop"=>7046, "name_ru"=>"Ла Джунта", "name_en"=>"La Junta", "lang"=>"en", "lat"=>37.980340, "lon"=>-103.544432),
"/US/US-CO/la-salle-2516" => array("type"=>"town", "pop"=>1986, "name_ru"=>"Ла Саль", "name_en"=>"La Salle", "lang"=>"en", "lat"=>40.347800, "lon"=>-104.704541),
"/US/US-CO/la-veta-2518" => array("type"=>"town", "pop"=>882, "name_ru"=>"Ла Вета", "name_en"=>"La Veta", "lang"=>"en", "lat"=>37.510010, "lon"=>-105.007282),
"/US/US-CO/lafayette-2508" => array("type"=>"city", "pop"=>25065, "name_ru"=>"Ла Файет", "name_en"=>"Lafayette", "lang"=>"en", "lat"=>39.994300, "lon"=>-105.097988),
"/US/US-CO/lake-city-2511" => array("type"=>"town", "pop"=>385, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>38.029111, "lon"=>-107.310322),
"/US/US-CO/lakeside-2512" => array("type"=>"town", "pop"=>19, "name_ru"=>"Лэйксайд", "name_en"=>"Lakeside", "lang"=>"en", "lat"=>39.778767, "lon"=>-105.057460),
"/US/US-CO/lakewood-2513" => array("type"=>"city", "pop"=>140989, "name_ru"=>"Лейквуд", "name_en"=>"Lakewood", "lang"=>"en", "lat"=>39.695200, "lon"=>-105.113556),
"/US/US-CO/lamar-2514" => array("type"=>"city", "pop"=>7953, "name_ru"=>"Ламар", "name_en"=>"Lamar", "lang"=>"en", "lat"=>38.080272, "lon"=>-102.617507),
"/US/US-CO/larkspur-2515" => array("type"=>"town", "pop"=>328, "name_ru"=>"Ларкспер", "name_en"=>"Larkspur", "lang"=>"en", "lat"=>39.181622, "lon"=>-104.895988),
"/US/US-CO/las-animas-2517" => array("type"=>"city", "pop"=>2345, "name_ru"=>"Лас Анимас", "name_en"=>"Las Animas", "lang"=>"en", "lat"=>38.066845, "lon"=>-103.225449),
"/US/US-CO/leadville-2519" => array("type"=>"city", "pop"=>2743, "name_ru"=>"Лидвилл", "name_en"=>"Leadville", "lang"=>"en", "lat"=>39.246776, "lon"=>-106.293776),
"/US/US-CO/leadville-north-2520" => array("type"=>"cdp", "pop"=>1757, "name_ru"=>"Лидвилл Норт", "name_en"=>"Leadville North", "lang"=>"en", "lat"=>39.260089, "lon"=>-106.310985),
"/US/US-CO/limon-2521" => array("type"=>"town", "pop"=>1750, "name_ru"=>"Лимон", "name_en"=>"Limon", "lang"=>"en", "lat"=>39.266080, "lon"=>-103.689495),
"/US/US-CO/lincoln-park-2522" => array("type"=>"cdp", "pop"=>3728, "name_ru"=>"Линкольн Парк", "name_en"=>"Lincoln Park", "lang"=>"en", "lat"=>38.425771, "lon"=>-105.213245),
"/US/US-CO/littleton-2523" => array("type"=>"city", "pop"=>40777, "name_ru"=>"Литтлтон", "name_en"=>"Littleton", "lang"=>"en", "lat"=>39.594567, "lon"=>-105.014396),
"/US/US-CO/lochbuie-2524" => array("type"=>"town", "pop"=>4800, "name_ru"=>"Локбуи", "name_en"=>"Lochbuie", "lang"=>"en", "lat"=>40.007700, "lon"=>-104.720098),
"/US/US-CO/log-lane-village-2525" => array("type"=>"town", "pop"=>973, "name_ru"=>"Лог Лейн Виллидж", "name_en"=>"Log Lane Village", "lang"=>"en", "lat"=>40.270019, "lon"=>-103.828392),
"/US/US-CO/longmont-2526" => array("type"=>"city", "pop"=>85928, "name_ru"=>"Лонгмонт", "name_en"=>"Longmont", "lang"=>"en", "lat"=>40.172807, "lon"=>-105.111699),
"/US/US-CO/louisville-2527" => array("type"=>"city", "pop"=>19133, "name_ru"=>"Луисвилл", "name_en"=>"Louisville", "lang"=>"en", "lat"=>39.968367, "lon"=>-105.139554),
"/US/US-CO/loveland-2528" => array("type"=>"city", "pop"=>65587, "name_ru"=>"Лавленд", "name_en"=>"Loveland", "lang"=>"en", "lat"=>40.418567, "lon"=>-105.071745),
"/US/US-CO/lyons-2529" => array("type"=>"town", "pop"=>2035, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>40.222850, "lon"=>-105.268499),
"/US/US-CO/manassa-2530" => array("type"=>"town", "pop"=>945, "name_ru"=>"Манасса", "name_en"=>"Manassa", "lang"=>"en", "lat"=>37.172935, "lon"=>-105.936032),
"/US/US-CO/mancos-2531" => array("type"=>"town", "pop"=>1261, "name_ru"=>"Манкос", "name_en"=>"Mancos", "lang"=>"en", "lat"=>37.346468, "lon"=>-108.293368),
"/US/US-CO/manitou-springs-2532" => array("type"=>"city", "pop"=>5161, "name_ru"=>"Манитоу Спрингс", "name_en"=>"Manitou Springs", "lang"=>"en", "lat"=>38.857054, "lon"=>-104.910925),
"/US/US-CO/manzanola-2533" => array("type"=>"town", "pop"=>472, "name_ru"=>"Манзанола", "name_en"=>"Manzanola", "lang"=>"en", "lat"=>38.108778, "lon"=>-103.866163),
"/US/US-CO/marble-2534" => array("type"=>"town", "pop"=>104, "name_ru"=>"Марбл", "name_en"=>"Marble", "lang"=>"en", "lat"=>39.071428, "lon"=>-107.188426),
"/US/US-CO/mead-2535" => array("type"=>"town", "pop"=>3408, "name_ru"=>"Мид", "name_en"=>"Mead", "lang"=>"en", "lat"=>40.234328, "lon"=>-104.987923),
"/US/US-CO/meeker-2536" => array("type"=>"town", "pop"=>2387, "name_ru"=>"Микер", "name_en"=>"Meeker", "lang"=>"en", "lat"=>40.054489, "lon"=>-107.889755),
"/US/US-CO/merino-2537" => array("type"=>"town", "pop"=>270, "name_ru"=>"Мерино", "name_en"=>"Merino", "lang"=>"en", "lat"=>40.484262, "lon"=>-103.353463),
"/US/US-CO/milliken-2538" => array("type"=>"town", "pop"=>6159, "name_ru"=>"Милликен", "name_en"=>"Milliken", "lang"=>"en", "lat"=>40.316600, "lon"=>-104.852745),
"/US/US-CO/minturn-2539" => array("type"=>"town", "pop"=>1185, "name_ru"=>"Минтум", "name_en"=>"Minturn", "lang"=>"en", "lat"=>39.584280, "lon"=>-106.426062),
"/US/US-CO/moffat-2540" => array("type"=>"town", "pop"=>126, "name_ru"=>"Моффат", "name_en"=>"Moffat", "lang"=>"en", "lat"=>38.001696, "lon"=>-105.904302),
"/US/US-CO/monte-vista-2541" => array("type"=>"city", "pop"=>4009, "name_ru"=>"Монте Виста", "name_en"=>"Monte Vista", "lang"=>"en", "lat"=>37.578603, "lon"=>-106.144234),
"/US/US-CO/montezuma-2542" => array("type"=>"town", "pop"=>43, "name_ru"=>"Мотезума", "name_en"=>"Montezuma", "lang"=>"en", "lat"=>39.581252, "lon"=>-105.868438),
"/US/US-CO/montrose-2543" => array("type"=>"city", "pop"=>17989, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>38.479417, "lon"=>-107.871351),
"/US/US-CO/monument-2544" => array("type"=>"town", "pop"=>2588, "name_ru"=>"Моньюмент", "name_en"=>"Monument", "lang"=>"en", "lat"=>39.070128, "lon"=>-104.857121),
"/US/US-CO/morrison-2545" => array("type"=>"town", "pop"=>415, "name_ru"=>"Моррисон", "name_en"=>"Morrison", "lang"=>"en", "lat"=>39.654505, "lon"=>-105.182381),
"/US/US-CO/mount-crested-butte-2547" => array("type"=>"town", "pop"=>848, "name_ru"=>"Маунт Крестид Бьютт", "name_en"=>"Mount Crested Butte", "lang"=>"en", "lat"=>38.907845, "lon"=>-106.966244),
"/US/US-CO/mountain-view-2546" => array("type"=>"town", "pop"=>520, "name_ru"=>"Маунтин Вью", "name_en"=>"Mountain View", "lang"=>"en", "lat"=>39.775700, "lon"=>-105.055749),
"/US/US-CO/naturita-2548" => array("type"=>"town", "pop"=>655, "name_ru"=>"Натурита", "name_en"=>"Naturita", "lang"=>"en", "lat"=>38.218988, "lon"=>-108.567990),
"/US/US-CO/nederland-2549" => array("type"=>"town", "pop"=>1369, "name_ru"=>"Недерленд", "name_en"=>"Nederland", "lang"=>"en", "lat"=>39.962600, "lon"=>-105.505302),
"/US/US-CO/new-castle-2550" => array("type"=>"town", "pop"=>3796, "name_ru"=>"Нью Касл", "name_en"=>"New Castle", "lang"=>"en", "lat"=>39.585419, "lon"=>-107.530591),
"/US/US-CO/niwot-2551" => array("type"=>"cdp", "pop"=>2666, "name_ru"=>"Нивот", "name_en"=>"Niwot", "lang"=>"en", "lat"=>40.095618, "lon"=>-105.155824),
"/US/US-CO/northglenn-2552" => array("type"=>"city", "pop"=>33697, "name_ru"=>"Нортгленн", "name_en"=>"Northglenn", "lang"=>"en", "lat"=>39.905700, "lon"=>-104.977916),
"/US/US-CO/norwood-2553" => array("type"=>"town", "pop"=>462, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>38.130288, "lon"=>-108.291613),
"/US/US-CO/nucla-2554" => array("type"=>"town", "pop"=>732, "name_ru"=>"Накла", "name_en"=>"Nucla", "lang"=>"en", "lat"=>38.266815, "lon"=>-108.549573),
"/US/US-CO/nunn-2555" => array("type"=>"town", "pop"=>541, "name_ru"=>"Нанн", "name_en"=>"Nunn", "lang"=>"en", "lat"=>40.703942, "lon"=>-104.780381),
"/US/US-CO/oak-creek-2556" => array("type"=>"town", "pop"=>813, "name_ru"=>"Оук Крик", "name_en"=>"Oak Creek", "lang"=>"en", "lat"=>40.274517, "lon"=>-106.957027),
"/US/US-CO/olathe-2557" => array("type"=>"town", "pop"=>1742, "name_ru"=>"Олэт", "name_en"=>"Olathe", "lang"=>"en", "lat"=>38.605729, "lon"=>-107.981697),
"/US/US-CO/olney-springs-2558" => array("type"=>"town", "pop"=>344, "name_ru"=>"Олни Спрингс", "name_en"=>"Olney Springs", "lang"=>"en", "lat"=>38.167339, "lon"=>-103.944096),
"/US/US-CO/ophir-2559" => array("type"=>"town", "pop"=>132, "name_ru"=>"Офир", "name_en"=>"Ophir", "lang"=>"en", "lat"=>37.856875, "lon"=>-107.828930),
"/US/US-CO/orchard-city-2560" => array("type"=>"town", "pop"=>3203, "name_ru"=>"Орчард Сити", "name_en"=>"Orchard City", "lang"=>"en", "lat"=>38.814758, "lon"=>-107.963614),
"/US/US-CO/orchard-mesa-2561" => array("type"=>"cdp", "pop"=>5977, "name_ru"=>"Орчард Меса", "name_en"=>"Orchard Mesa", "lang"=>"en", "lat"=>39.033845, "lon"=>-108.522449),
"/US/US-CO/ordway-2562" => array("type"=>"town", "pop"=>1098, "name_ru"=>"Ордуэй", "name_en"=>"Ordway", "lang"=>"en", "lat"=>38.220944, "lon"=>-103.756779),
"/US/US-CO/otis-2563" => array("type"=>"town", "pop"=>470, "name_ru"=>"Отис", "name_en"=>"Otis", "lang"=>"en", "lat"=>40.149689, "lon"=>-102.961586),
"/US/US-CO/ouray-2564" => array("type"=>"city", "pop"=>932, "name_ru"=>"Орей", "name_en"=>"Ouray", "lang"=>"en", "lat"=>38.026105, "lon"=>-107.671908),
"/US/US-CO/ovid-2565" => array("type"=>"town", "pop"=>291, "name_ru"=>"Овид", "name_en"=>"Ovid", "lang"=>"en", "lat"=>40.959928, "lon"=>-102.388073),
"/US/US-CO/pagosa-springs-2566" => array("type"=>"town", "pop"=>1744, "name_ru"=>"Пагоса Спрингс", "name_en"=>"Pagosa Springs", "lang"=>"en", "lat"=>37.270141, "lon"=>-107.018221),
"/US/US-CO/palisade-2567" => array("type"=>"town", "pop"=>2840, "name_ru"=>"Палисейд", "name_en"=>"Palisade", "lang"=>"en", "lat"=>39.107449, "lon"=>-108.357778),
"/US/US-CO/palmer-lake-2568" => array("type"=>"town", "pop"=>2326, "name_ru"=>"Палмер Лейк", "name_en"=>"Palmer Lake", "lang"=>"en", "lat"=>39.116774, "lon"=>-104.905903),
"/US/US-CO/paoli-2569" => array("type"=>"town", "pop"=>41, "name_ru"=>"Паоли", "name_en"=>"Paoli", "lang"=>"en", "lat"=>40.612129, "lon"=>-102.472401),
"/US/US-CO/paonia-2570" => array("type"=>"town", "pop"=>1633, "name_ru"=>"Паония", "name_en"=>"Paonia", "lang"=>"en", "lat"=>38.869454, "lon"=>-107.589546),
"/US/US-CO/parachute-2571" => array("type"=>"town", "pop"=>1290, "name_ru"=>"Парашут", "name_en"=>"Parachute", "lang"=>"en", "lat"=>39.445818, "lon"=>-108.055100),
"/US/US-CO/parker-2572" => array("type"=>"town", "pop"=>43767, "name_ru"=>"Паркер", "name_en"=>"Parker", "lang"=>"en", "lat"=>39.509599, "lon"=>-104.763798),
"/US/US-CO/peetz-2573" => array("type"=>"town", "pop"=>244, "name_ru"=>"Пеетц", "name_en"=>"Peetz", "lang"=>"en", "lat"=>40.961434, "lon"=>-103.114208),
"/US/US-CO/penrose-2574" => array("type"=>"cdp", "pop"=>2235, "name_ru"=>"Пенроуз", "name_en"=>"Penrose", "lang"=>"en", "lat"=>38.431138, "lon"=>-105.012507),
"/US/US-CO/pierce-2575" => array("type"=>"town", "pop"=>877, "name_ru"=>"Пирс", "name_en"=>"Pierce", "lang"=>"en", "lat"=>40.632856, "lon"=>-104.754637),
"/US/US-CO/pitkin-2576" => array("type"=>"town", "pop"=>117, "name_ru"=>"Питкин", "name_en"=>"Pitkin", "lang"=>"en", "lat"=>38.608060, "lon"=>-106.516131),
"/US/US-CO/platteville-2577" => array("type"=>"town", "pop"=>2634, "name_ru"=>"Платтевилл", "name_en"=>"Platteville", "lang"=>"en", "lat"=>40.216889, "lon"=>-104.822007),
"/US/US-CO/poncha-springs-2578" => array("type"=>"town", "pop"=>456, "name_ru"=>"Понча Спрингс", "name_en"=>"Poncha Springs", "lang"=>"en", "lat"=>38.514148, "lon"=>-106.074089),
"/US/US-CO/ponderosa-park-2579" => array("type"=>"cdp", "pop"=>1640, "name_ru"=>"Пондероза Парк", "name_en"=>"Ponderosa Park", "lang"=>"en", "lat"=>39.397293, "lon"=>-104.635487),
"/US/US-CO/pritchett-2580" => array("type"=>"town", "pop"=>116, "name_ru"=>"Питчетт", "name_en"=>"Pritchett", "lang"=>"en", "lat"=>37.370192, "lon"=>-102.858645),
"/US/US-CO/prospect-heights-2581" => array("type"=>"town", "pop"=>19, "name_ru"=>"Проспект Хайтс", "name_en"=>"Prospect Heights", "lang"=>"en", "lat"=>38.425919, "lon"=>-105.235753),
"/US/US-CO/pueblo-2582" => array("type"=>"city", "pop"=>104951, "name_ru"=>"Пуэбло", "name_en"=>"Pueblo", "lang"=>"en", "lat"=>38.273550, "lon"=>-104.620999),
"/US/US-CO/pueblo-west-2583" => array("type"=>"cdp", "pop"=>4386, "name_ru"=>"Пуэбло Уэст", "name_en"=>"Pueblo West", "lang"=>"en", "lat"=>38.349893, "lon"=>-104.722137),
"/US/US-CO/ramah-2584" => array("type"=>"town", "pop"=>120, "name_ru"=>"Рама", "name_en"=>"Ramah", "lang"=>"en", "lat"=>39.121851, "lon"=>-104.166593),
"/US/US-CO/rangely-2585" => array("type"=>"town", "pop"=>2138, "name_ru"=>"Рейнджели", "name_en"=>"Rangely", "lang"=>"en", "lat"=>40.087621, "lon"=>-108.778530),
"/US/US-CO/raymer-2586" => array("type"=>"town", "pop"=>91, "name_ru"=>"Реймер", "name_en"=>"Raymer", "lang"=>"en", "lat"=>40.607279, "lon"=>-103.843870),
"/US/US-CO/red-cliff-2587" => array("type"=>"town", "pop"=>323, "name_ru"=>"Ред Клиф", "name_en"=>"Red Cliff", "lang"=>"en", "lat"=>39.508386, "lon"=>-106.369504),
"/US/US-CO/redlands-2588" => array("type"=>"cdp", "pop"=>9355, "name_ru"=>"Редлендс", "name_en"=>"Redlands", "lang"=>"en", "lat"=>39.085033, "lon"=>-108.650762),
"/US/US-CO/rico-2589" => array("type"=>"town", "pop"=>258, "name_ru"=>"Рико", "name_en"=>"Rico", "lang"=>"en", "lat"=>37.687998, "lon"=>-108.030960),
"/US/US-CO/ridgway-2590" => array("type"=>"town", "pop"=>806, "name_ru"=>"Риджуэй", "name_en"=>"Ridgway", "lang"=>"en", "lat"=>38.155812, "lon"=>-107.752153),
"/US/US-CO/rifle-2591" => array("type"=>"city", "pop"=>9202, "name_ru"=>"Райфл", "name_en"=>"Rifle", "lang"=>"en", "lat"=>39.539337, "lon"=>-107.777751),
"/US/US-CO/rockvale-2592" => array("type"=>"town", "pop"=>436, "name_ru"=>"Роквейл", "name_en"=>"Rockvale", "lang"=>"en", "lat"=>38.369138, "lon"=>-105.163267),
"/US/US-CO/rocky-ford-2593" => array("type"=>"city", "pop"=>3950, "name_ru"=>"Роки Форд", "name_en"=>"Rocky Ford", "lang"=>"en", "lat"=>38.049916, "lon"=>-103.721897),
"/US/US-CO/romeo-2594" => array("type"=>"town", "pop"=>342, "name_ru"=>"Ромео", "name_en"=>"Romeo", "lang"=>"en", "lat"=>37.171573, "lon"=>-105.985903),
"/US/US-CO/rye-2595" => array("type"=>"town", "pop"=>201, "name_ru"=>"Рай", "name_en"=>"Rye", "lang"=>"en", "lat"=>37.921754, "lon"=>-104.930990),
"/US/US-CO/saguache-2596" => array("type"=>"town", "pop"=>581, "name_ru"=>"Сагуач", "name_en"=>"Saguache", "lang"=>"en", "lat"=>38.086131, "lon"=>-106.140940),
"/US/US-CO/salida-2597" => array("type"=>"city", "pop"=>5396, "name_ru"=>"Салида", "name_en"=>"Salida", "lang"=>"en", "lat"=>38.532665, "lon"=>-105.999794),
"/US/US-CO/san-luis-2599" => array("type"=>"town", "pop"=>642, "name_ru"=>"Сан Луис", "name_en"=>"San Luis", "lang"=>"en", "lat"=>37.200897, "lon"=>-105.423593),
"/US/US-CO/sanford-2598" => array("type"=>"town", "pop"=>740, "name_ru"=>"Санфорд", "name_en"=>"Sanford", "lang"=>"en", "lat"=>37.257428, "lon"=>-105.900014),
"/US/US-CO/sawpit-2600" => array("type"=>"town", "pop"=>24, "name_ru"=>"Сопит", "name_en"=>"Sawpit", "lang"=>"en", "lat"=>37.994813, "lon"=>-107.999947),
"/US/US-CO/security-widefield-2601" => array("type"=>"cdp", "pop"=>23822, "name_ru"=>"Секьюрити-Уайдфилд", "name_en"=>"Security-Widefield", "lang"=>"en", "lat"=>38.743968, "lon"=>-104.711494),
"/US/US-CO/sedgwick-2602" => array("type"=>"town", "pop"=>163, "name_ru"=>"Седжвик", "name_en"=>"Sedgwick", "lang"=>"en", "lat"=>40.934754, "lon"=>-102.524821),
"/US/US-CO/seibert-2603" => array("type"=>"town", "pop"=>165, "name_ru"=>"Сейберт", "name_en"=>"Seibert", "lang"=>"en", "lat"=>39.298126, "lon"=>-102.868355),
"/US/US-CO/severance-2604" => array("type"=>"town", "pop"=>3010, "name_ru"=>"Северанс", "name_en"=>"Severance", "lang"=>"en", "lat"=>40.522410, "lon"=>-104.849613),
"/US/US-CO/sheridan-2605" => array("type"=>"city", "pop"=>5412, "name_ru"=>"Шеридан", "name_en"=>"Sheridan", "lang"=>"en", "lat"=>39.648933, "lon"=>-105.017143),
"/US/US-CO/sheridan-lake-2606" => array("type"=>"town", "pop"=>56, "name_ru"=>"Шеридан Лейк", "name_en"=>"Sheridan Lake", "lang"=>"en", "lat"=>38.465971, "lon"=>-102.293521),
"/US/US-CO/sherrelwood-2607" => array("type"=>"cdp", "pop"=>16636, "name_ru"=>"Шеррелвуд", "name_en"=>"Sherrelwood", "lang"=>"en", "lat"=>39.837776, "lon"=>-105.000922),
"/US/US-CO/silt-2608" => array("type"=>"town", "pop"=>2667, "name_ru"=>"Силт", "name_en"=>"Silt", "lang"=>"en", "lat"=>39.549654, "lon"=>-107.653822),
"/US/US-CO/silver-cliff-2609" => array("type"=>"town", "pop"=>578, "name_ru"=>"Сильвер Клиф", "name_en"=>"Silver Cliff", "lang"=>"en", "lat"=>38.117254, "lon"=>-105.404494),
"/US/US-CO/silver-plume-2610" => array("type"=>"town", "pop"=>184, "name_ru"=>"Сильвер Плам", "name_en"=>"Silver Plume", "lang"=>"en", "lat"=>39.696084, "lon"=>-105.726703),
"/US/US-CO/silverthorne-2611" => array("type"=>"town", "pop"=>3980, "name_ru"=>"Смльверторн", "name_en"=>"Silverthorne", "lang"=>"en", "lat"=>39.651663, "lon"=>-106.081623),
"/US/US-CO/silverton-2612" => array("type"=>"town", "pop"=>524, "name_ru"=>"Сильвертон", "name_en"=>"Silverton", "lang"=>"en", "lat"=>37.810879, "lon"=>-107.663958),
"/US/US-CO/simla-2613" => array("type"=>"town", "pop"=>685, "name_ru"=>"Симла", "name_en"=>"Simla", "lang"=>"en", "lat"=>39.140781, "lon"=>-104.080890),
"/US/US-CO/snowmass-village-2614" => array("type"=>"town", "pop"=>1911, "name_ru"=>"Сноумасс Виллидж", "name_en"=>"Snowmass Village", "lang"=>"en", "lat"=>39.195468, "lon"=>-106.938652),
"/US/US-CO/southglenn-2615" => array("type"=>"cdp", "pop"=>43087, "name_ru"=>"Саутгленн", "name_en"=>"Southglenn", "lang"=>"en", "lat"=>39.587350, "lon"=>-104.952155),
"/US/US-CO/springfield-2616" => array("type"=>"town", "pop"=>1291, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>37.406614, "lon"=>-102.616771),
"/US/US-CO/starkville-2617" => array("type"=>"town", "pop"=>128, "name_ru"=>"Старквилл", "name_en"=>"Starkville", "lang"=>"en", "lat"=>37.117242, "lon"=>-104.523985),
"/US/US-CO/steamboat-springs-2618" => array("type"=>"city", "pop"=>9592, "name_ru"=>"Стимбоат Спрингс", "name_en"=>"Steamboat Springs", "lang"=>"en", "lat"=>40.477880, "lon"=>-106.821000),
"/US/US-CO/sterling-2619" => array("type"=>"city", "pop"=>12800, "name_ru"=>"Стерлинг", "name_en"=>"Sterling", "lang"=>"en", "lat"=>40.626486, "lon"=>-103.199144),
"/US/US-CO/stratmoor-2620" => array("type"=>"cdp", "pop"=>5854, "name_ru"=>"Стрэтмур", "name_en"=>"Stratmoor", "lang"=>"en", "lat"=>38.773950, "lon"=>-104.779128),
"/US/US-CO/stratton-2621" => array("type"=>"town", "pop"=>606, "name_ru"=>"Стрэттон", "name_en"=>"Stratton", "lang"=>"en", "lat"=>39.302982, "lon"=>-102.603794),
"/US/US-CO/sugar-city-2622" => array("type"=>"town", "pop"=>242, "name_ru"=>"Шугар Сити", "name_en"=>"Sugar City", "lang"=>"en", "lat"=>38.232879, "lon"=>-103.663081),
"/US/US-CO/superior-2623" => array("type"=>"town", "pop"=>10549, "name_ru"=>"Супериор", "name_en"=>"Superior", "lang"=>"en", "lat"=>39.929200, "lon"=>-105.154930),
"/US/US-CO/swink-2624" => array("type"=>"town", "pop"=>653, "name_ru"=>"Свинк", "name_en"=>"Swink", "lang"=>"en", "lat"=>38.013849, "lon"=>-103.627311),
"/US/US-CO/telluride-2625" => array("type"=>"town", "pop"=>2361, "name_ru"=>"Теллерайд", "name_en"=>"Telluride", "lang"=>"en", "lat"=>37.938672, "lon"=>-107.812360),
"/US/US-CO/the-pinery-2626" => array("type"=>"cdp", "pop"=>4885, "name_ru"=>"Зе Пайнри", "name_en"=>"The Pinery", "lang"=>"en", "lat"=>39.455234, "lon"=>-104.733935),
"/US/US-CO/thornton-2627" => array("type"=>"city", "pop"=>113429, "name_ru"=>"Торнтон", "name_en"=>"Thornton", "lang"=>"en", "lat"=>39.891950, "lon"=>-104.955228),
"/US/US-CO/timnath-2628" => array("type"=>"town", "pop"=>284, "name_ru"=>"Тимнат", "name_en"=>"Timnath", "lang"=>"en", "lat"=>40.529286, "lon"=>-104.983249),
"/US/US-CO/towaoc-2629" => array("type"=>"cdp", "pop"=>700, "name_ru"=>"Товаок", "name_en"=>"Towaoc", "lang"=>"en", "lat"=>37.209878, "lon"=>-108.727208),
"/US/US-CO/trinidad-2630" => array("type"=>"city", "pop"=>9125, "name_ru"=>"Тринидад", "name_en"=>"Trinidad", "lang"=>"en", "lat"=>37.168545, "lon"=>-104.505660),
"/US/US-CO/two-buttes-2631" => array("type"=>"town", "pop"=>57, "name_ru"=>"Ту Бьюттс", "name_en"=>"Two Buttes", "lang"=>"en", "lat"=>37.560549, "lon"=>-102.395909),
"/US/US-CO/vail-2632" => array("type"=>"town", "pop"=>4768, "name_ru"=>"Вейл", "name_en"=>"Vail", "lang"=>"en", "lat"=>39.639476, "lon"=>-106.357702),
"/US/US-CO/victor-2633" => array("type"=>"city", "pop"=>406, "name_ru"=>"Виктор", "name_en"=>"Victor", "lang"=>"en", "lat"=>38.708395, "lon"=>-105.140585),
"/US/US-CO/vilas-2634" => array("type"=>"town", "pop"=>94, "name_ru"=>"Вилас", "name_en"=>"Vilas", "lang"=>"en", "lat"=>37.374118, "lon"=>-102.446461),
"/US/US-CO/vona-2635" => array("type"=>"town", "pop"=>86, "name_ru"=>"Вона", "name_en"=>"Vona", "lang"=>"en", "lat"=>39.302702, "lon"=>-102.743628),
"/US/US-CO/walden-2636" => array("type"=>"town", "pop"=>586, "name_ru"=>"Уолден", "name_en"=>"Walden", "lang"=>"en", "lat"=>40.730629, "lon"=>-106.280937),
"/US/US-CO/walsenburg-2637" => array("type"=>"city", "pop"=>3830, "name_ru"=>"Уолсенбург", "name_en"=>"Walsenburg", "lang"=>"en", "lat"=>37.626730, "lon"=>-104.777060),
"/US/US-CO/walsh-2638" => array("type"=>"town", "pop"=>637, "name_ru"=>"Уолш", "name_en"=>"Walsh", "lang"=>"en", "lat"=>37.385518, "lon"=>-102.279593),
"/US/US-CO/ward-2639" => array("type"=>"town", "pop"=>168, "name_ru"=>"Уорд", "name_en"=>"Ward", "lang"=>"en", "lat"=>40.072500, "lon"=>-105.511284),
"/US/US-CO/welby-2640" => array("type"=>"cdp", "pop"=>10218, "name_ru"=>"Уэлби", "name_en"=>"Welby", "lang"=>"en", "lat"=>39.838250, "lon"=>-104.964658),
"/US/US-CO/wellington-2641" => array("type"=>"town", "pop"=>5052, "name_ru"=>"Веллингтон", "name_en"=>"Wellington", "lang"=>"en", "lat"=>40.702799, "lon"=>-105.001900),
"/US/US-CO/westcliffe-2642" => array("type"=>"town", "pop"=>455, "name_ru"=>"Уэстклифф", "name_en"=>"Westcliffe", "lang"=>"en", "lat"=>38.132869, "lon"=>-105.465067),
"/US/US-CO/westminster-2643" => array("type"=>"city", "pop"=>107056, "name_ru"=>"Уэстминстер", "name_en"=>"Westminster", "lang"=>"en", "lat"=>39.881500, "lon"=>-105.050854),
"/US/US-CO/westminster-east-2644" => array("type"=>"cdp", "pop"=>5197, "name_ru"=>"Уэстминтсер Ист", "name_en"=>"Westminster East", "lang"=>"en", "lat"=>39.822650, "lon"=>-105.003749),
"/US/US-CO/wheat-ridge-2645" => array("type"=>"city", "pop"=>30894, "name_ru"=>"Уит Ридж", "name_en"=>"Wheat Ridge", "lang"=>"en", "lat"=>39.773450, "lon"=>-105.098550),
"/US/US-CO/wiggins-2646" => array("type"=>"town", "pop"=>945, "name_ru"=>"Уиггинс", "name_en"=>"Wiggins", "lang"=>"en", "lat"=>40.230296, "lon"=>-104.073254),
"/US/US-CO/wiley-2647" => array("type"=>"town", "pop"=>446, "name_ru"=>"Уили", "name_en"=>"Wiley", "lang"=>"en", "lat"=>38.156068, "lon"=>-102.718665),
"/US/US-CO/williamsburg-2648" => array("type"=>"town", "pop"=>758, "name_ru"=>"Уилльямсбург", "name_en"=>"Williamsburg", "lang"=>"en", "lat"=>38.381331, "lon"=>-105.174880),
"/US/US-CO/windsor-2649" => array("type"=>"town", "pop"=>18058, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>40.485232, "lon"=>-104.906515),
"/US/US-CO/winter-park-2650" => array("type"=>"town", "pop"=>857, "name_ru"=>"Уинтер Парк", "name_en"=>"Winter Park", "lang"=>"en", "lat"=>39.886880, "lon"=>-105.779391),
"/US/US-CO/woodland-park-2651" => array("type"=>"city", "pop"=>6501, "name_ru"=>"Вудленд Парк", "name_en"=>"Woodland Park", "lang"=>"en", "lat"=>38.995032, "lon"=>-105.051732),
"/US/US-CO/woodmoor-2652" => array("type"=>"cdp", "pop"=>3858, "name_ru"=>"Вудмур", "name_en"=>"Woodmoor", "lang"=>"en", "lat"=>39.101480, "lon"=>-104.846914),
"/US/US-CO/wray-2653" => array("type"=>"city", "pop"=>2109, "name_ru"=>"Рей", "name_en"=>"Wray", "lang"=>"en", "lat"=>40.079443, "lon"=>-102.225968),
"/US/US-CO/yampa-2654" => array("type"=>"town", "pop"=>425, "name_ru"=>"Ямпа", "name_en"=>"Yampa", "lang"=>"en", "lat"=>40.153266, "lon"=>-106.907470),
"/US/US-CO/yuma-2655" => array("type"=>"city", "pop"=>3263, "name_ru"=>"Юма", "name_en"=>"Yuma", "lang"=>"en", "lat"=>40.123896, "lon"=>-102.724063),
"/US/US-CT" => array("type"=>"state", "pop"=>3501252, "name_ru"=>"Коннектикут", "name_en"=>"Connecticut", "lang"=>"en"),
"/US/US-CT/ansonia-2656" => array("type"=>"city", "pop"=>18503, "name_ru"=>"Ансония", "name_en"=>"Ansonia", "lang"=>"en", "lat"=>41.343390, "lon"=>-73.069041),
"/US/US-CT/bantam-2657" => array("type"=>"borough", "pop"=>795, "name_ru"=>"Бантэм", "name_en"=>"Bantam", "lang"=>"en", "lat"=>41.723486, "lon"=>-73.243819),
"/US/US-CT/bethel-2658" => array("type"=>"cdp", "pop"=>18438, "name_ru"=>"Бетел", "name_en"=>"Bethel", "lang"=>"en", "lat"=>41.371533, "lon"=>-73.411583),
"/US/US-CT/bethlehem-village-2659" => array("type"=>"cdp", "pop"=>1976, "name_ru"=>"Бетлхэм Виллидж", "name_en"=>"Bethlehem Village", "lang"=>"en", "lat"=>41.642640, "lon"=>-73.203497),
"/US/US-CT/blue-hills-2660" => array("type"=>"cdp", "pop"=>3206, "name_ru"=>"Блу Хилс", "name_en"=>"Blue Hills", "lang"=>"en", "lat"=>41.812896, "lon"=>-72.698158),
"/US/US-CT/branford-center-2661" => array("type"=>"cdp", "pop"=>5688, "name_ru"=>"Брэнфорд Сетер", "name_en"=>"Branford Center", "lang"=>"en", "lat"=>41.276183, "lon"=>-72.813943),
"/US/US-CT/bridgeport-2662" => array("type"=>"city", "pop"=>136405, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>41.186316, "lon"=>-73.196209),
"/US/US-CT/bristol-2663" => array("type"=>"city", "pop"=>60927, "name_ru"=>"Бристол", "name_en"=>"Bristol", "lang"=>"en", "lat"=>41.681201, "lon"=>-72.941017),
"/US/US-CT/broad-brook-2664" => array("type"=>"cdp", "pop"=>3585, "name_ru"=>"Броад Брук", "name_en"=>"Broad Brook", "lang"=>"en", "lat"=>41.909000, "lon"=>-72.544404),
"/US/US-CT/canaan-2665" => array("type"=>"cdp", "pop"=>1095, "name_ru"=>"Канаан", "name_en"=>"Canaan", "lang"=>"en", "lat"=>42.031616, "lon"=>-73.331443),
"/US/US-CT/canton-valley-2666" => array("type"=>"cdp", "pop"=>1563, "name_ru"=>"Кэнтон Валли", "name_en"=>"Canton Valley", "lang"=>"en", "lat"=>41.834150, "lon"=>-72.892201),
"/US/US-CT/central-manchester-2667" => array("type"=>"cdp", "pop"=>30934, "name_ru"=>"Сентрал Манчестер", "name_en"=>"Central Manchester", "lang"=>"en", "lat"=>41.779550, "lon"=>-72.520694),
"/US/US-CT/central-somers-2668" => array("type"=>"cdp", "pop"=>1644, "name_ru"=>"Сентрал Сомерс", "name_en"=>"Central Somers", "lang"=>"en", "lat"=>41.990683, "lon"=>-72.442303),
"/US/US-CT/central-waterford-2669" => array("type"=>"cdp", "pop"=>2939, "name_ru"=>"Сентрал Уотерфорд", "name_en"=>"Central Waterford", "lang"=>"en", "lat"=>41.345700, "lon"=>-72.130714),
"/US/US-CT/cheshire-village-2670" => array("type"=>"cdp", "pop"=>5759, "name_ru"=>"Чешир Виллидж", "name_en"=>"Cheshire Village", "lang"=>"en", "lat"=>41.504200, "lon"=>-72.900145),
"/US/US-CT/chester-center-2671" => array("type"=>"cdp", "pop"=>1563, "name_ru"=>"Честер Сентер", "name_en"=>"Chester Center", "lang"=>"en", "lat"=>41.400461, "lon"=>-72.453803),
"/US/US-CT/clinton-2672" => array("type"=>"cdp", "pop"=>13554, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>41.275833, "lon"=>-72.530078),
"/US/US-CT/colchester-2673" => array("type"=>"borough", "pop"=>15578, "name_ru"=>"Колчестер", "name_en"=>"Colchester", "lang"=>"en", "lat"=>41.573968, "lon"=>-72.331957),
"/US/US-CT/collinsville-2674" => array("type"=>"cdp", "pop"=>2591, "name_ru"=>"Коллинсвилл", "name_en"=>"Collinsville", "lang"=>"en", "lat"=>41.819218, "lon"=>-72.922999),
"/US/US-CT/conning-towers-nautilus-park-2675" => array("type"=>"cdp", "pop"=>10013, "name_ru"=>"Коннинг Тауэрс-Наутилус Парк", "name_en"=>"Conning Towers-Nautilus Park", "lang"=>"en", "lat"=>41.374450, "lon"=>-72.075095),
"/US/US-CT/coventry-lake-2676" => array("type"=>"cdp", "pop"=>2895, "name_ru"=>"Ковентри Лейк", "name_en"=>"Coventry Lake", "lang"=>"en", "lat"=>41.772305, "lon"=>-72.333013),
"/US/US-CT/crystal-lake-2677" => array("type"=>"cdp", "pop"=>1175, "name_ru"=>"Кристал Лэйк", "name_en"=>"Crystal Lake", "lang"=>"en", "lat"=>41.933450, "lon"=>-72.375049),
"/US/US-CT/danbury-2678" => array("type"=>"city", "pop"=>79256, "name_ru"=>"Данбери", "name_en"=>"Danbury", "lang"=>"en", "lat"=>41.402133, "lon"=>-73.471532),
"/US/US-CT/danielson-2679" => array("type"=>"borough", "pop"=>4310, "name_ru"=>"Даниелсон", "name_en"=>"Danielson", "lang"=>"en", "lat"=>41.809493, "lon"=>-71.887243),
"/US/US-CT/darien-2680" => array("type"=>"cdp", "pop"=>20177, "name_ru"=>"Дариен", "name_en"=>"Darien", "lang"=>"en", "lat"=>41.051137, "lon"=>-73.479716),
"/US/US-CT/deep-river-center-2681" => array("type"=>"cdp", "pop"=>2520, "name_ru"=>"Дип Ривер Сентер", "name_en"=>"Deep River Center", "lang"=>"en", "lat"=>41.381579, "lon"=>-72.440054),
"/US/US-CT/derby-2682" => array("type"=>"city", "pop"=>12393, "name_ru"=>"Дерби", "name_en"=>"Derby", "lang"=>"en", "lat"=>41.326667, "lon"=>-73.082789),
"/US/US-CT/durham-2683" => array("type"=>"cdp", "pop"=>7456, "name_ru"=>"Дарем", "name_en"=>"Durham", "lang"=>"en", "lat"=>41.473906, "lon"=>-72.681921),
"/US/US-CT/east-brooklyn-2684" => array("type"=>"cdp", "pop"=>1481, "name_ru"=>"Ист Бруклин", "name_en"=>"East Brooklyn", "lang"=>"en", "lat"=>41.789074, "lon"=>-71.897054),
"/US/US-CT/east-hampton-2685" => array("type"=>"cdp", "pop"=>12685, "name_ru"=>"Ист Хэмптон", "name_en"=>"East Hampton", "lang"=>"en", "lat"=>41.572921, "lon"=>-72.495990),
"/US/US-CT/east-hartford-2686" => array("type"=>"cdp", "pop"=>48571, "name_ru"=>"Ист Хартфорд", "name_en"=>"East Hartford", "lang"=>"en", "lat"=>41.761358, "lon"=>-72.615943),
"/US/US-CT/east-haven-2687" => array("type"=>"cdp", "pop"=>28590, "name_ru"=>"Ист Хейвен", "name_en"=>"East Haven", "lang"=>"en", "lat"=>41.295613, "lon"=>-72.863050),
"/US/US-CT/essex-village-2688" => array("type"=>"cdp", "pop"=>2500, "name_ru"=>"Эссекс Виллидж", "name_en"=>"Essex Village", "lang"=>"en", "lat"=>41.357041, "lon"=>-72.390705),
"/US/US-CT/fenwick-2689" => array("type"=>"borough", "pop"=>53, "name_ru"=>"Фенвик", "name_en"=>"Fenwick", "lang"=>"en", "lat"=>41.271150, "lon"=>-72.355393),
"/US/US-CT/georgetown-2690" => array("type"=>"cdp", "pop"=>1694, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>41.247283, "lon"=>-73.433430),
"/US/US-CT/glastonbury-center-2691" => array("type"=>"cdp", "pop"=>7082, "name_ru"=>"Гластонбери Сентер", "name_en"=>"Glastonbury Center", "lang"=>"en", "lat"=>41.700867, "lon"=>-72.599963),
"/US/US-CT/groton-2692" => array("type"=>"city", "pop"=>9316, "name_ru"=>"Гротон", "name_en"=>"Groton", "lang"=>"en", "lat"=>41.327333, "lon"=>-72.071670),
"/US/US-CT/guilford-center-2693" => array("type"=>"cdp", "pop"=>2588, "name_ru"=>"Гилфорд Сентер", "name_en"=>"Guilford Center", "lang"=>"en", "lat"=>41.280350, "lon"=>-72.677009),
"/US/US-CT/hartford-2694" => array("type"=>"city", "pop"=>124062, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>41.765700, "lon"=>-72.683866),
"/US/US-CT/hazardville-2695" => array("type"=>"cdp", "pop"=>5179, "name_ru"=>"Хазардвилл", "name_en"=>"Hazardville", "lang"=>"en", "lat"=>41.986800, "lon"=>-72.529468),
"/US/US-CT/heritage-village-2696" => array("type"=>"cdp", "pop"=>3623, "name_ru"=>"Херитейдж Виллидж", "name_en"=>"Heritage Village", "lang"=>"en", "lat"=>41.485576, "lon"=>-73.238307),
"/US/US-CT/higganum-2697" => array("type"=>"cdp", "pop"=>1692, "name_ru"=>"Хигганум", "name_en"=>"Higganum", "lang"=>"en", "lat"=>41.491249, "lon"=>-72.558013),
"/US/US-CT/jewett-city-2698" => array("type"=>"borough", "pop"=>2976, "name_ru"=>"Джеветт Сити", "name_en"=>"Jewett City", "lang"=>"en", "lat"=>41.605933, "lon"=>-71.981708),
"/US/US-CT/kensington-2699" => array("type"=>"cdp", "pop"=>8306, "name_ru"=>"Кенсингтон", "name_en"=>"Kensington", "lang"=>"en", "lat"=>41.628350, "lon"=>-72.771173),
"/US/US-CT/lake-pocotopaug-2700" => array("type"=>"cdp", "pop"=>3029, "name_ru"=>"Лейк Покотопог", "name_en"=>"Lake Pocotopaug", "lang"=>"en", "lat"=>41.592657, "lon"=>-72.511338),
"/US/US-CT/litchfield-2701" => array("type"=>"borough", "pop"=>1312, "name_ru"=>"Литчфилд", "name_en"=>"Litchfield", "lang"=>"en", "lat"=>41.744680, "lon"=>-73.190074),
"/US/US-CT/madison-center-2702" => array("type"=>"cdp", "pop"=>2139, "name_ru"=>"Мэдисон Сентер", "name_en"=>"Madison Center", "lang"=>"en", "lat"=>41.271650, "lon"=>-72.600585),
"/US/US-CT/meriden-2703" => array("type"=>"city", "pop"=>59186, "name_ru"=>"Мериден", "name_en"=>"Meriden", "lang"=>"en", "lat"=>41.536651, "lon"=>-72.795180),
"/US/US-CT/middletown-2704" => array("type"=>"city", "pop"=>48030, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>41.548371, "lon"=>-72.654264),
"/US/US-CT/milford-(remainder)-2705" => array("type"=>"(remainder)", "pop"=>48168, "name_ru"=>"Милфорд", "name_en"=>"Milford (remainder)", "lang"=>"en", "lat"=>41.226237, "lon"=>-73.062631),
"/US/US-CT/moodus-2706" => array("type"=>"cdp", "pop"=>1170, "name_ru"=>"Мудус", "name_en"=>"Moodus", "lang"=>"en", "lat"=>41.504004, "lon"=>-72.450117),
"/US/US-CT/moosup-2707" => array("type"=>"cdp", "pop"=>3289, "name_ru"=>"Мусап", "name_en"=>"Moosup", "lang"=>"en", "lat"=>41.717711, "lon"=>-71.873938),
"/US/US-CT/mystic-2708" => array("type"=>"cdp", "pop"=>2618, "name_ru"=>"Мистик", "name_en"=>"Mystic", "lang"=>"en", "lat"=>41.357300, "lon"=>-71.952454),
"/US/US-CT/naugatuck-2709" => array("type"=>"borough", "pop"=>31931, "name_ru"=>"Нагатак", "name_en"=>"Naugatuck", "lang"=>"en", "lat"=>41.489600, "lon"=>-73.052049),
"/US/US-CT/new-britain-2710" => array("type"=>"city", "pop"=>70486, "name_ru"=>"Нью Бритн", "name_en"=>"New Britain", "lang"=>"en", "lat"=>41.675100, "lon"=>-72.787805),
"/US/US-CT/new-hartford-center-2711" => array("type"=>"cdp", "pop"=>1269, "name_ru"=>"Нью Хартфорд", "name_en"=>"New Hartford Center", "lang"=>"en", "lat"=>41.879703, "lon"=>-72.973883),
"/US/US-CT/new-haven-2712" => array("type"=>"city", "pop"=>123669, "name_ru"=>"Нью Хейвен", "name_en"=>"New Haven", "lang"=>"en", "lat"=>41.310100, "lon"=>-72.924068),
"/US/US-CT/new-london-2714" => array("type"=>"city", "pop"=>25891, "name_ru"=>"Нью Лондон", "name_en"=>"New London", "lang"=>"en", "lat"=>41.329667, "lon"=>-72.095020),
"/US/US-CT/new-milford-2715" => array("type"=>"cdp", "pop"=>28338, "name_ru"=>"Нью Милфорд", "name_en"=>"New Milford", "lang"=>"en", "lat"=>41.588550, "lon"=>-73.406843),
"/US/US-CT/new-preston-2716" => array("type"=>"cdp", "pop"=>1217, "name_ru"=>"Нью Престон", "name_en"=>"New Preston", "lang"=>"en", "lat"=>41.681741, "lon"=>-73.354309),
"/US/US-CT/newington-2713" => array("type"=>"cdp", "pop"=>29699, "name_ru"=>"Ньюингтон", "name_en"=>"Newington", "lang"=>"en", "lat"=>41.687267, "lon"=>-72.730690),
"/US/US-CT/newtown-2717" => array("type"=>"borough", "pop"=>1807, "name_ru"=>"Ньютаун", "name_en"=>"Newtown", "lang"=>"en", "lat"=>41.413250, "lon"=>-73.316018),
"/US/US-CT/niantic-2718" => array("type"=>"cdp", "pop"=>3048, "name_ru"=>"Найантик", "name_en"=>"Niantic", "lang"=>"en", "lat"=>41.322833, "lon"=>-72.194263),
"/US/US-CT/north-granby-2719" => array("type"=>"cdp", "pop"=>1455, "name_ru"=>"Норт Гранби", "name_en"=>"North Granby", "lang"=>"en", "lat"=>42.017967, "lon"=>-72.843623),
"/US/US-CT/north-grosvenor-dale-2720" => array("type"=>"cdp", "pop"=>1705, "name_ru"=>"Норт Гросвенор Дейл", "name_en"=>"North Grosvenor Dale", "lang"=>"en", "lat"=>41.983613, "lon"=>-71.903771),
"/US/US-CT/north-haven-2721" => array("type"=>"cdp", "pop"=>23961, "name_ru"=>"Норт Хейвен", "name_en"=>"North Haven", "lang"=>"en", "lat"=>41.380906, "lon"=>-72.856365),
"/US/US-CT/northwest-harwinton-2722" => array("type"=>"cdp", "pop"=>3299, "name_ru"=>"Нортуэст Нарвинтон", "name_en"=>"Northwest Harwinton", "lang"=>"en", "lat"=>41.775992, "lon"=>-73.080385),
"/US/US-CT/norwalk-2723" => array("type"=>"city", "pop"=>83185, "name_ru"=>"Норуолк", "name_en"=>"Norwalk", "lang"=>"en", "lat"=>41.093950, "lon"=>-73.420177),
"/US/US-CT/norwich-2724" => array("type"=>"city", "pop"=>36388, "name_ru"=>"Норвич", "name_en"=>"Norwich", "lang"=>"en", "lat"=>41.550183, "lon"=>-72.087948),
"/US/US-CT/oakville-2725" => array("type"=>"cdp", "pop"=>8741, "name_ru"=>"Оуквилл", "name_en"=>"Oakville", "lang"=>"en", "lat"=>41.591950, "lon"=>-73.085878),
"/US/US-CT/old-saybrook-center-2726" => array("type"=>"cdp", "pop"=>1820, "name_ru"=>"Оулд Сейбрук Сентер", "name_en"=>"Old Saybrook Center", "lang"=>"en", "lat"=>41.291355, "lon"=>-72.368757),
"/US/US-CT/orange-2727" => array("type"=>"cdp", "pop"=>13781, "name_ru"=>"Оранж", "name_en"=>"Orange", "lang"=>"en", "lat"=>41.279450, "lon"=>-73.025867),
"/US/US-CT/pawcatuck-2728" => array("type"=>"cdp", "pop"=>5289, "name_ru"=>"Покатак", "name_en"=>"Pawcatuck", "lang"=>"en", "lat"=>41.375602, "lon"=>-71.852953),
"/US/US-CT/plainfield-village-2729" => array("type"=>"cdp", "pop"=>2856, "name_ru"=>"Плейнфилд Виллидж", "name_en"=>"Plainfield Village", "lang"=>"en", "lat"=>41.676501, "lon"=>-71.924849),
"/US/US-CT/poquonock-bridge-2730" => array("type"=>"cdp", "pop"=>2770, "name_ru"=>"Покнок Бридж", "name_en"=>"Poquonock Bridge", "lang"=>"en", "lat"=>41.349050, "lon"=>-72.028139),
"/US/US-CT/portland-2731" => array("type"=>"cdp", "pop"=>9551, "name_ru"=>"Портленд", "name_en"=>"Portland", "lang"=>"en", "lat"=>41.583995, "lon"=>-72.624783),
"/US/US-CT/putnam-district-2732" => array("type"=>"cdp", "pop"=>6835, "name_ru"=>"Путнам Дистрикт", "name_en"=>"Putnam District", "lang"=>"en", "lat"=>41.920850, "lon"=>-71.909275),
"/US/US-CT/quinebaug-2733" => array("type"=>"cdp", "pop"=>1031, "name_ru"=>"Квинбааг", "name_en"=>"Quinebaug", "lang"=>"en", "lat"=>42.009067, "lon"=>-71.937171),
"/US/US-CT/ridgefield-2734" => array("type"=>"cdp", "pop"=>24011, "name_ru"=>"Риджфилд", "name_en"=>"Ridgefield", "lang"=>"en", "lat"=>41.273832, "lon"=>-73.499880),
"/US/US-CT/salmon-brook-2735" => array("type"=>"cdp", "pop"=>2185, "name_ru"=>"Сэлмон Брук", "name_en"=>"Salmon Brook", "lang"=>"en", "lat"=>41.956441, "lon"=>-72.795949),
"/US/US-CT/saybrook-manor-2736" => array("type"=>"cdp", "pop"=>1073, "name_ru"=>"Сейбрук Манор", "name_en"=>"Saybrook Manor", "lang"=>"en", "lat"=>41.280250, "lon"=>-72.406226),
"/US/US-CT/shelton-2737" => array("type"=>"city", "pop"=>39991, "name_ru"=>"Шелтон", "name_en"=>"Shelton", "lang"=>"en", "lat"=>41.304050, "lon"=>-73.138553),
"/US/US-CT/sherwood-manor-2738" => array("type"=>"cdp", "pop"=>6357, "name_ru"=>"Шервуд Манор", "name_en"=>"Sherwood Manor", "lang"=>"en", "lat"=>42.013233, "lon"=>-72.564767),
"/US/US-CT/simsbury-center-2739" => array("type"=>"cdp", "pop"=>5577, "name_ru"=>"Симсбери Сентер", "name_en"=>"Simsbury Center", "lang"=>"en", "lat"=>41.882950, "lon"=>-72.811380),
"/US/US-CT/south-coventry-2740" => array("type"=>"cdp", "pop"=>1257, "name_ru"=>"Саут Совентри", "name_en"=>"South Coventry", "lang"=>"en", "lat"=>41.769866, "lon"=>-72.305516),
"/US/US-CT/south-windham-2741" => array("type"=>"cdp", "pop"=>1644, "name_ru"=>"Саут Уиндхэм", "name_en"=>"South Windham", "lang"=>"en", "lat"=>41.683455, "lon"=>-72.184655),
"/US/US-CT/south-woodstock-2743" => array("type"=>"cdp", "pop"=>1112, "name_ru"=>"Саут Вудсток", "name_en"=>"South Woodstock", "lang"=>"en", "lat"=>41.926964, "lon"=>-71.957750),
"/US/US-CT/southwood-acres-2742" => array("type"=>"cdp", "pop"=>8963, "name_ru"=>"Саутвуд Экрес", "name_en"=>"Southwood Acres", "lang"=>"en", "lat"=>41.962567, "lon"=>-72.571962),
"/US/US-CT/stafford-springs-2744" => array("type"=>"borough", "pop"=>4100, "name_ru"=>"Стэффорд Спрингс", "name_en"=>"Stafford Springs", "lang"=>"en", "lat"=>41.954108, "lon"=>-72.309460),
"/US/US-CT/stamford-2745" => array("type"=>"city", "pop"=>119303, "name_ru"=>"Стэмфорд", "name_en"=>"Stamford", "lang"=>"en", "lat"=>41.096733, "lon"=>-73.552688),
"/US/US-CT/stonington-2746" => array("type"=>"borough", "pop"=>1026, "name_ru"=>"Стонингтон", "name_en"=>"Stonington", "lang"=>"en", "lat"=>41.333793, "lon"=>-71.904579),
"/US/US-CT/storrs-2747" => array("type"=>"cdp", "pop"=>12198, "name_ru"=>"Сторрс", "name_en"=>"Storrs", "lang"=>"en", "lat"=>41.806054, "lon"=>-72.256675),
"/US/US-CT/stratford-2748" => array("type"=>"cdp", "pop"=>48853, "name_ru"=>"Стрэтфорд", "name_en"=>"Stratford", "lang"=>"en", "lat"=>41.204317, "lon"=>-73.130216),
"/US/US-CT/suffield-depot-2749" => array("type"=>"cdp", "pop"=>1353, "name_ru"=>"Саффилд Депо", "name_en"=>"Suffield Depot", "lang"=>"en", "lat"=>41.981200, "lon"=>-72.650349),
"/US/US-CT/tariffville-2750" => array("type"=>"cdp", "pop"=>1477, "name_ru"=>"Тариффвилл", "name_en"=>"Tariffville", "lang"=>"en", "lat"=>41.908200, "lon"=>-72.767233),
"/US/US-CT/terramuggus-2751" => array("type"=>"cdp", "pop"=>1044, "name_ru"=>"Террамаггус", "name_en"=>"Terramuggus", "lang"=>"en", "lat"=>41.634900, "lon"=>-72.470949),
"/US/US-CT/terryville-2752" => array("type"=>"cdp", "pop"=>5426, "name_ru"=>"Терривилл", "name_en"=>"Terryville", "lang"=>"en", "lat"=>41.679101, "lon"=>-73.007727),
"/US/US-CT/thompsonville-2753" => array("type"=>"cdp", "pop"=>8458, "name_ru"=>"Томпсонвилл", "name_en"=>"Thompsonville", "lang"=>"en", "lat"=>41.990800, "lon"=>-72.596374),
"/US/US-CT/torrington-2754" => array("type"=>"city", "pop"=>35312, "name_ru"=>"Торрингтон", "name_en"=>"Torrington", "lang"=>"en", "lat"=>41.834962, "lon"=>-73.129439),
"/US/US-CT/trumbull-2755" => array("type"=>"cdp", "pop"=>34688, "name_ru"=>"Трамбулл", "name_en"=>"Trumbull", "lang"=>"en", "lat"=>41.258166, "lon"=>-73.207474),
"/US/US-CT/uncasville-oxoboxo-valley-2756" => array("type"=>"cdp", "pop"=>2975, "name_ru"=>"Анкасвилл-Оксобоксо Валли", "name_en"=>"Uncasville-Oxoboxo Valley", "lang"=>"en", "lat"=>41.443615, "lon"=>-72.124446),
"/US/US-CT/wallingford-center-2757" => array("type"=>"cdp", "pop"=>17827, "name_ru"=>"Уоллингфорд Сентер", "name_en"=>"Wallingford Center", "lang"=>"en", "lat"=>41.449800, "lon"=>-72.819399),
"/US/US-CT/waterbury-2758" => array("type"=>"city", "pop"=>107037, "name_ru"=>"Уотербери", "name_en"=>"Waterbury", "lang"=>"en", "lat"=>41.558400, "lon"=>-73.037429),
"/US/US-CT/wauregan-2759" => array("type"=>"cdp", "pop"=>1079, "name_ru"=>"Уоереган", "name_en"=>"Wauregan", "lang"=>"en", "lat"=>41.752557, "lon"=>-71.911655),
"/US/US-CT/weatogue-2760" => array("type"=>"cdp", "pop"=>2521, "name_ru"=>"Уэтог", "name_en"=>"Weatogue", "lang"=>"en", "lat"=>41.843600, "lon"=>-72.828961),
"/US/US-CT/west-hartford-2762" => array("type"=>"cdp", "pop"=>60495, "name_ru"=>"Уэст Хартфорд", "name_en"=>"West Hartford", "lang"=>"en", "lat"=>41.767651, "lon"=>-72.754412),
"/US/US-CT/west-haven-2763" => array("type"=>"city", "pop"=>52420, "name_ru"=>"Уэст Хейвен", "name_en"=>"West Haven", "lang"=>"en", "lat"=>41.273950, "lon"=>-72.968305),
"/US/US-CT/west-mystic-2764" => array("type"=>"cdp", "pop"=>3595, "name_ru"=>"Уэст Мистик", "name_en"=>"West Mystic", "lang"=>"en", "lat"=>41.351800, "lon"=>-71.982149),
"/US/US-CT/west-simsbury-2766" => array("type"=>"cdp", "pop"=>2149, "name_ru"=>"Уэст Симсбери", "name_en"=>"West Simsbury", "lang"=>"en", "lat"=>41.873350, "lon"=>-72.844347),
"/US/US-CT/westbrook-center-2761" => array("type"=>"cdp", "pop"=>2060, "name_ru"=>"Уэстбрук Сентер", "name_en"=>"Westbrook Center", "lang"=>"en", "lat"=>41.280080, "lon"=>-72.443454),
"/US/US-CT/westport-2765" => array("type"=>"cdp", "pop"=>26592, "name_ru"=>"Уэстпорт", "name_en"=>"Westport", "lang"=>"en", "lat"=>41.123200, "lon"=>-73.347565),
"/US/US-CT/wethersfield-2767" => array("type"=>"cdp", "pop"=>25719, "name_ru"=>"Уэтерсфилд", "name_en"=>"Wethersfield", "lang"=>"en", "lat"=>41.701150, "lon"=>-72.670109),
"/US/US-CT/willimantic-2768" => array("type"=>"cdp", "pop"=>14746, "name_ru"=>"Уиллимантик", "name_en"=>"Willimantic", "lang"=>"en", "lat"=>41.715613, "lon"=>-72.217974),
"/US/US-CT/windsor-locks-2769" => array("type"=>"cdp", "pop"=>12495, "name_ru"=>"Виндзор Локс", "name_en"=>"Windsor Locks", "lang"=>"en", "lat"=>41.926050, "lon"=>-72.657999),
"/US/US-CT/winsted-2770" => array("type"=>"cdp", "pop"=>8254, "name_ru"=>"Уинстед", "name_en"=>"Winsted", "lang"=>"en", "lat"=>41.927043, "lon"=>-73.066908),
"/US/US-CT/woodbury-center-2771" => array("type"=>"cdp", "pop"=>1212, "name_ru"=>"Вудбери Сентер", "name_en"=>"Woodbury Center", "lang"=>"en", "lat"=>41.546635, "lon"=>-73.206983),
"/US/US-CT/woodmont-2772" => array("type"=>"borough", "pop"=>1771, "name_ru"=>"Вудмонт", "name_en"=>"Woodmont", "lang"=>"en", "lat"=>41.224200, "lon"=>-72.992598),
"/US/US-DC" => array("type"=>"district", "pop"=>591833, "name_ru"=>"Округ Колумбия", "name_en"=>"District of Columbia", "lang"=>"en"),
"/US/US-DC/washington-2845" => array("type"=>"city", "pop"=>591833, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>38.905050, "lon"=>-77.016167),
"/US/US-DE" => array("type"=>"state", "pop"=>873092, "name_ru"=>"Делавер", "name_en"=>"Delaware", "lang"=>"en"),
"/US/US-DE/arden-2773" => array("type"=>"village", "pop"=>484, "name_ru"=>"Арден", "name_en"=>"Arden", "lang"=>"en", "lat"=>39.810550, "lon"=>-75.489589),
"/US/US-DE/ardencroft-2774" => array("type"=>"village", "pop"=>275, "name_ru"=>"Арденкрофт", "name_en"=>"Ardencroft", "lang"=>"en", "lat"=>39.803987, "lon"=>-75.486825),
"/US/US-DE/ardentown-2775" => array("type"=>"village", "pop"=>308, "name_ru"=>"Ардентаун", "name_en"=>"Ardentown", "lang"=>"en", "lat"=>39.809038, "lon"=>-75.481888),
"/US/US-DE/bellefonte-2776" => array("type"=>"town", "pop"=>1290, "name_ru"=>"Беллефонт", "name_en"=>"Bellefonte", "lang"=>"en", "lat"=>39.766300, "lon"=>-75.498313),
"/US/US-DE/bethany-beach-2777" => array("type"=>"town", "pop"=>964, "name_ru"=>"Бетани Бич", "name_en"=>"Bethany Beach", "lang"=>"en", "lat"=>38.537989, "lon"=>-75.065363),
"/US/US-DE/bethel-2778" => array("type"=>"town", "pop"=>202, "name_ru"=>"Бетел", "name_en"=>"Bethel", "lang"=>"en", "lat"=>38.570057, "lon"=>-75.619792),
"/US/US-DE/blades-2779" => array("type"=>"town", "pop"=>1158, "name_ru"=>"Блэйдс", "name_en"=>"Blades", "lang"=>"en", "lat"=>38.634654, "lon"=>-75.609225),
"/US/US-DE/bowers-2780" => array("type"=>"town", "pop"=>346, "name_ru"=>"Бауэрс", "name_en"=>"Bowers", "lang"=>"en", "lat"=>39.060342, "lon"=>-75.400184),
"/US/US-DE/bridgeville-2781" => array("type"=>"town", "pop"=>1630, "name_ru"=>"Бриджвилл", "name_en"=>"Bridgeville", "lang"=>"en", "lat"=>38.743349, "lon"=>-75.602355),
"/US/US-DE/brookside-2782" => array("type"=>"cdp", "pop"=>15307, "name_ru"=>"Бруксайд", "name_en"=>"Brookside", "lang"=>"en", "lat"=>39.667833, "lon"=>-75.715923),
"/US/US-DE/camden-2783" => array("type"=>"town", "pop"=>2565, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>39.112756, "lon"=>-75.549488),
"/US/US-DE/cheswold-2784" => array("type"=>"town", "pop"=>473, "name_ru"=>"Чесволд", "name_en"=>"Cheswold", "lang"=>"en", "lat"=>39.216453, "lon"=>-75.584646),
"/US/US-DE/claymont-2785" => array("type"=>"cdp", "pop"=>9800, "name_ru"=>"Клэймонт", "name_en"=>"Claymont", "lang"=>"en", "lat"=>39.802950, "lon"=>-75.459426),
"/US/US-DE/clayton-2786" => array("type"=>"town", "pop"=>1481, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>39.292513, "lon"=>-75.633156),
"/US/US-DE/dagsboro-2787" => array("type"=>"town", "pop"=>578, "name_ru"=>"Дагсборо", "name_en"=>"Dagsboro", "lang"=>"en", "lat"=>38.547515, "lon"=>-75.247740),
"/US/US-DE/delaware-city-2788" => array("type"=>"city", "pop"=>1517, "name_ru"=>"Делавер Сити", "name_en"=>"Delaware City", "lang"=>"en", "lat"=>39.575520, "lon"=>-75.600378),
"/US/US-DE/delmar-2789" => array("type"=>"town", "pop"=>1516, "name_ru"=>"Делмар", "name_en"=>"Delmar", "lang"=>"en", "lat"=>38.458785, "lon"=>-75.570119),
"/US/US-DE/dewey-beach-2790" => array("type"=>"town", "pop"=>318, "name_ru"=>"Дьюи Бич", "name_en"=>"Dewey Beach", "lang"=>"en", "lat"=>38.695298, "lon"=>-75.076752),
"/US/US-DE/dover-2791" => array("type"=>"city", "pop"=>36107, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>39.158691, "lon"=>-75.517441),
"/US/US-DE/dover-base-housing-2792" => array("type"=>"cdp", "pop"=>4376, "name_ru"=>"Довер Бэйз Хаузин", "name_en"=>"Dover Base Housing", "lang"=>"en", "lat"=>39.118276, "lon"=>-75.482627),
"/US/US-DE/edgemoor-2793" => array("type"=>"cdp", "pop"=>5853, "name_ru"=>"Эджмур", "name_en"=>"Edgemoor", "lang"=>"en", "lat"=>39.755750, "lon"=>-75.507807),
"/US/US-DE/ellendale-2794" => array("type"=>"town", "pop"=>354, "name_ru"=>"Эллендэйл", "name_en"=>"Ellendale", "lang"=>"en", "lat"=>38.807108, "lon"=>-75.424251),
"/US/US-DE/elsmere-2795" => array("type"=>"town", "pop"=>5688, "name_ru"=>"Элсмер", "name_en"=>"Elsmere", "lang"=>"en", "lat"=>39.737850, "lon"=>-75.594924),
"/US/US-DE/farmington-2796" => array("type"=>"town", "pop"=>85, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>38.870729, "lon"=>-75.579564),
"/US/US-DE/felton-2797" => array("type"=>"town", "pop"=>905, "name_ru"=>"Фелтон", "name_en"=>"Felton", "lang"=>"en", "lat"=>39.009072, "lon"=>-75.577921),
"/US/US-DE/fenwick-island-2798" => array("type"=>"town", "pop"=>366, "name_ru"=>"Фенвик Айленд", "name_en"=>"Fenwick Island", "lang"=>"en", "lat"=>38.460739, "lon"=>-75.053832),
"/US/US-DE/frankford-2799" => array("type"=>"town", "pop"=>777, "name_ru"=>"Фрэнкфорд", "name_en"=>"Frankford", "lang"=>"en", "lat"=>38.522274, "lon"=>-75.232871),
"/US/US-DE/frederica-2800" => array("type"=>"town", "pop"=>751, "name_ru"=>"Фредерика", "name_en"=>"Frederica", "lang"=>"en", "lat"=>39.008850, "lon"=>-75.464773),
"/US/US-DE/georgetown-2801" => array("type"=>"town", "pop"=>5233, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>38.690798, "lon"=>-75.385889),
"/US/US-DE/greenwood-2802" => array("type"=>"town", "pop"=>907, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>38.807408, "lon"=>-75.592053),
"/US/US-DE/harrington-2803" => array("type"=>"city", "pop"=>3435, "name_ru"=>"Харрингтон", "name_en"=>"Harrington", "lang"=>"en", "lat"=>38.922743, "lon"=>-75.572318),
"/US/US-DE/hartly-2804" => array("type"=>"town", "pop"=>88, "name_ru"=>"Хартли", "name_en"=>"Hartly", "lang"=>"en", "lat"=>39.168374, "lon"=>-75.713060),
"/US/US-DE/henlopen-acres-2805" => array("type"=>"town", "pop"=>142, "name_ru"=>"Хенлопен Эйкерс", "name_en"=>"Henlopen Acres", "lang"=>"en", "lat"=>38.725826, "lon"=>-75.086069),
"/US/US-DE/highland-acres-2806" => array("type"=>"cdp", "pop"=>3151, "name_ru"=>"Хайленд Эйкерс", "name_en"=>"Highland Acres", "lang"=>"en", "lat"=>39.117248, "lon"=>-75.523096),
"/US/US-DE/houston-2807" => array("type"=>"town", "pop"=>483, "name_ru"=>"Хьюстон", "name_en"=>"Houston", "lang"=>"en", "lat"=>38.916814, "lon"=>-75.504505),
"/US/US-DE/kent-acres-2808" => array("type"=>"cdp", "pop"=>1807, "name_ru"=>"Кент Эйкерс", "name_en"=>"Kent Acres", "lang"=>"en", "lat"=>39.133104, "lon"=>-75.518816),
"/US/US-DE/kenton-2809" => array("type"=>"town", "pop"=>268, "name_ru"=>"Кентон", "name_en"=>"Kenton", "lang"=>"en", "lat"=>39.227202, "lon"=>-75.665840),
"/US/US-DE/laurel-2810" => array("type"=>"town", "pop"=>3982, "name_ru"=>"Лорел", "name_en"=>"Laurel", "lang"=>"en", "lat"=>38.554963, "lon"=>-75.573070),
"/US/US-DE/leipsic-2811" => array("type"=>"town", "pop"=>229, "name_ru"=>"Лейпсик", "name_en"=>"Leipsic", "lang"=>"en", "lat"=>39.243057, "lon"=>-75.517076),
"/US/US-DE/lewes-2812" => array("type"=>"city", "pop"=>3127, "name_ru"=>"Льюис", "name_en"=>"Lewes", "lang"=>"en", "lat"=>38.780287, "lon"=>-75.149782),
"/US/US-DE/little-creek-2813" => array("type"=>"town", "pop"=>217, "name_ru"=>"Литтл Крик", "name_en"=>"Little Creek", "lang"=>"en", "lat"=>39.165247, "lon"=>-75.448830),
"/US/US-DE/long-neck-2814" => array("type"=>"cdp", "pop"=>886, "name_ru"=>"Лонг Нек", "name_en"=>"Long Neck", "lang"=>"en", "lat"=>38.620123, "lon"=>-75.151043),
"/US/US-DE/magnolia-2815" => array("type"=>"town", "pop"=>250, "name_ru"=>"Магнолия", "name_en"=>"Magnolia", "lang"=>"en", "lat"=>39.069766, "lon"=>-75.476716),
"/US/US-DE/middletown-2816" => array("type"=>"town", "pop"=>12152, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>39.451120, "lon"=>-75.714658),
"/US/US-DE/milford-2817" => array("type"=>"city", "pop"=>8511, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>38.914055, "lon"=>-75.433108),
"/US/US-DE/millsboro-2818" => array("type"=>"town", "pop"=>2698, "name_ru"=>"Миллсборо", "name_en"=>"Millsboro", "lang"=>"en", "lat"=>38.585154, "lon"=>-75.291839),
"/US/US-DE/millville-2819" => array("type"=>"town", "pop"=>282, "name_ru"=>"Миллвилл", "name_en"=>"Millville", "lang"=>"en", "lat"=>38.545259, "lon"=>-75.112405),
"/US/US-DE/milton-2820" => array("type"=>"town", "pop"=>1835, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>38.778868, "lon"=>-75.313246),
"/US/US-DE/new-castle-2822" => array("type"=>"city", "pop"=>4977, "name_ru"=>"Нью Касл", "name_en"=>"New Castle", "lang"=>"en", "lat"=>39.666400, "lon"=>-75.570396),
"/US/US-DE/newark-2821" => array("type"=>"city", "pop"=>29886, "name_ru"=>"Ньюарк", "name_en"=>"Newark", "lang"=>"en", "lat"=>39.676600, "lon"=>-75.757699),
"/US/US-DE/newport-2823" => array("type"=>"town", "pop"=>1106, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>39.713250, "lon"=>-75.607026),
"/US/US-DE/ocean-view-2824" => array("type"=>"town", "pop"=>1138, "name_ru"=>"Оушн Вью", "name_en"=>"Ocean View", "lang"=>"en", "lat"=>38.545001, "lon"=>-75.092665),
"/US/US-DE/odessa-2825" => array("type"=>"town", "pop"=>337, "name_ru"=>"Одесса", "name_en"=>"Odessa", "lang"=>"en", "lat"=>39.456373, "lon"=>-75.660117),
"/US/US-DE/pike-creek-2826" => array("type"=>"cdp", "pop"=>10163, "name_ru"=>"Пайк Крик", "name_en"=>"Pike Creek", "lang"=>"en", "lat"=>39.730913, "lon"=>-75.704483),
"/US/US-DE/rehoboth-beach-2827" => array("type"=>"city", "pop"=>1587, "name_ru"=>"Рехобот Бич", "name_en"=>"Rehoboth Beach", "lang"=>"en", "lat"=>38.717342, "lon"=>-75.080910),
"/US/US-DE/rising-sun-lebanon-2828" => array("type"=>"cdp", "pop"=>2177, "name_ru"=>"Райзинг Сан-Лебанон", "name_en"=>"Rising Sun-Lebanon", "lang"=>"en", "lat"=>39.101117, "lon"=>-75.506645),
"/US/US-DE/riverview-2829" => array("type"=>"cdp", "pop"=>1138, "name_ru"=>"Ривервью", "name_en"=>"Riverview", "lang"=>"en", "lat"=>39.029070, "lon"=>-75.519805),
"/US/US-DE/rodney-village-2830" => array("type"=>"cdp", "pop"=>1745, "name_ru"=>"Родни Виллидж", "name_en"=>"Rodney Village", "lang"=>"en", "lat"=>39.126368, "lon"=>-75.536912),
"/US/US-DE/seaford-2831" => array("type"=>"city", "pop"=>7260, "name_ru"=>"Сифорд", "name_en"=>"Seaford", "lang"=>"en", "lat"=>38.648905, "lon"=>-75.616618),
"/US/US-DE/selbyville-2832" => array("type"=>"town", "pop"=>1853, "name_ru"=>"Селбивилл", "name_en"=>"Selbyville", "lang"=>"en", "lat"=>38.458056, "lon"=>-75.223961),
"/US/US-DE/slaughter-beach-2833" => array("type"=>"town", "pop"=>220, "name_ru"=>"Слаутер Бич", "name_en"=>"Slaughter Beach", "lang"=>"en", "lat"=>38.912275, "lon"=>-75.304348),
"/US/US-DE/smyrna-2834" => array("type"=>"town", "pop"=>8603, "name_ru"=>"Смирна", "name_en"=>"Smyrna", "lang"=>"en", "lat"=>39.295396, "lon"=>-75.607711),
"/US/US-DE/south-bethany-2835" => array("type"=>"town", "pop"=>526, "name_ru"=>"Саут Бетани", "name_en"=>"South Bethany", "lang"=>"en", "lat"=>38.516148, "lon"=>-75.058382),
"/US/US-DE/stanton-2836" => array("type"=>"cdp", "pop"=>5028, "name_ru"=>"Стэнтон", "name_en"=>"Stanton", "lang"=>"en", "lat"=>39.718067, "lon"=>-75.651623),
"/US/US-DE/talleyville-2837" => array("type"=>"cdp", "pop"=>6346, "name_ru"=>"Тэлливилл", "name_en"=>"Talleyville", "lang"=>"en", "lat"=>39.810200, "lon"=>-75.550341),
"/US/US-DE/townsend-2838" => array("type"=>"town", "pop"=>380, "name_ru"=>"Таунсенд", "name_en"=>"Townsend", "lang"=>"en", "lat"=>39.394089, "lon"=>-75.693018),
"/US/US-DE/viola-2839" => array("type"=>"town", "pop"=>174, "name_ru"=>"Виола", "name_en"=>"Viola", "lang"=>"en", "lat"=>39.042112, "lon"=>-75.571948),
"/US/US-DE/wilmington-2840" => array("type"=>"city", "pop"=>72592, "name_ru"=>"Уилмингтон", "name_en"=>"Wilmington", "lang"=>"en", "lat"=>39.735572, "lon"=>-75.529956),
"/US/US-DE/wilmington-manor-2841" => array("type"=>"cdp", "pop"=>8568, "name_ru"=>"Уилмингтон Мэнор", "name_en"=>"Wilmington Manor", "lang"=>"en", "lat"=>39.685500, "lon"=>-75.584769),
"/US/US-DE/woodside-2842" => array("type"=>"town", "pop"=>204, "name_ru"=>"Вудсайд", "name_en"=>"Woodside", "lang"=>"en", "lat"=>39.070306, "lon"=>-75.568536),
"/US/US-DE/woodside-east-2843" => array("type"=>"cdp", "pop"=>1655, "name_ru"=>"Вудсайд Ист", "name_en"=>"Woodside East", "lang"=>"en", "lat"=>39.067122, "lon"=>-75.538246),
"/US/US-DE/wyoming-2844" => array("type"=>"town", "pop"=>1410, "name_ru"=>"Вайоминг", "name_en"=>"Wyoming", "lang"=>"en", "lat"=>39.116819, "lon"=>-75.560592),
"/US/US-FL" => array("type"=>"state", "pop"=>18328340, "name_ru"=>"Флорида", "name_en"=>"Florida", "lang"=>"en"),
"/US/US-FL/aberdeen-2846" => array("type"=>"cdp", "pop"=>2572, "name_ru"=>"Абердин", "name_en"=>"Aberdeen", "lang"=>"en", "lat"=>26.550244, "lon"=>-80.148872),
"/US/US-FL/alachua-2847" => array("type"=>"city", "pop"=>9306, "name_ru"=>"Алачуа", "name_en"=>"Alachua", "lang"=>"en", "lat"=>29.779044, "lon"=>-82.480014),
"/US/US-FL/alford-2848" => array("type"=>"town", "pop"=>470, "name_ru"=>"Алфорд", "name_en"=>"Alford", "lang"=>"en", "lat"=>30.696135, "lon"=>-85.395834),
"/US/US-FL/altamonte-springs-2849" => array("type"=>"city", "pop"=>39947, "name_ru"=>"Алтамонте Спрингс", "name_en"=>"Altamonte Springs", "lang"=>"en", "lat"=>28.662080, "lon"=>-81.395249),
"/US/US-FL/altha-2850" => array("type"=>"town", "pop"=>520, "name_ru"=>"Алта", "name_en"=>"Altha", "lang"=>"en", "lat"=>30.571394, "lon"=>-85.126461),
"/US/US-FL/alva-2851" => array("type"=>"cdp", "pop"=>1036, "name_ru"=>"Алва", "name_en"=>"Alva", "lang"=>"en", "lat"=>26.715060, "lon"=>-81.611143),
"/US/US-FL/andover-2852" => array("type"=>"cdp", "pop"=>6251, "name_ru"=>"Андоувер", "name_en"=>"Andover", "lang"=>"en", "lat"=>25.967969, "lon"=>-80.212942),
"/US/US-FL/anna-maria-2853" => array("type"=>"city", "pop"=>1829, "name_ru"=>"Анна Мария", "name_en"=>"Anna Maria", "lang"=>"en", "lat"=>27.528900, "lon"=>-82.733023),
"/US/US-FL/apalachicola-2854" => array("type"=>"city", "pop"=>2210, "name_ru"=>"Апалачикола", "name_en"=>"Apalachicola", "lang"=>"en", "lat"=>29.726584, "lon"=>-84.994280),
"/US/US-FL/apollo-beach-2855" => array("type"=>"cdp", "pop"=>6025, "name_ru"=>"Аполло Бич", "name_en"=>"Apollo Beach", "lang"=>"en", "lat"=>27.772859, "lon"=>-82.407895),
"/US/US-FL/apopka-2856" => array("type"=>"city", "pop"=>37935, "name_ru"=>"Апопка", "name_en"=>"Apopka", "lang"=>"en", "lat"=>28.686750, "lon"=>-81.513276),
"/US/US-FL/arcadia-2857" => array("type"=>"city", "pop"=>6771, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>27.219614, "lon"=>-81.859672),
"/US/US-FL/archer-2858" => array("type"=>"city", "pop"=>1374, "name_ru"=>"Арчер", "name_en"=>"Archer", "lang"=>"en", "lat"=>29.534080, "lon"=>-82.518935),
"/US/US-FL/asbury-lake-2859" => array("type"=>"cdp", "pop"=>2072, "name_ru"=>"Асбери Лэйк", "name_en"=>"Asbury Lake", "lang"=>"en", "lat"=>30.048863, "lon"=>-81.821623),
"/US/US-FL/astatula-2860" => array("type"=>"town", "pop"=>1831, "name_ru"=>"Астатула", "name_en"=>"Astatula", "lang"=>"en", "lat"=>28.710550, "lon"=>-81.734899),
"/US/US-FL/astor-2861" => array("type"=>"cdp", "pop"=>1273, "name_ru"=>"Астор", "name_en"=>"Astor", "lang"=>"en", "lat"=>29.149722, "lon"=>-81.532212),
"/US/US-FL/atlantic-beach-2862" => array("type"=>"city", "pop"=>13147, "name_ru"=>"Атлантик Бич", "name_en"=>"Atlantic Beach", "lang"=>"en", "lat"=>30.325527, "lon"=>-81.390440),
"/US/US-FL/atlantis-2863" => array("type"=>"city", "pop"=>2068, "name_ru"=>"Атлантис", "name_en"=>"Atlantis", "lang"=>"en", "lat"=>26.596183, "lon"=>-80.101875),
"/US/US-FL/auburndale-2864" => array("type"=>"city", "pop"=>14353, "name_ru"=>"Ауберндэйл", "name_en"=>"Auburndale", "lang"=>"en", "lat"=>28.070173, "lon"=>-81.797825),
"/US/US-FL/aventura-2865" => array("type"=>"cdp", "pop"=>29254, "name_ru"=>"Авентура", "name_en"=>"Aventura", "lang"=>"en", "lat"=>25.959200, "lon"=>-80.133889),
"/US/US-FL/avon-park-2866" => array("type"=>"city", "pop"=>9033, "name_ru"=>"Эйвон Парк", "name_en"=>"Avon Park", "lang"=>"en", "lat"=>27.593227, "lon"=>-81.503658),
"/US/US-FL/azalea-park-2867" => array("type"=>"cdp", "pop"=>8926, "name_ru"=>"Азалеа Парк", "name_en"=>"Azalea Park", "lang"=>"en", "lat"=>28.545400, "lon"=>-81.295242),
"/US/US-FL/babson-park-2868" => array("type"=>"cdp", "pop"=>1125, "name_ru"=>"Бабсон Парк", "name_en"=>"Babson Park", "lang"=>"en", "lat"=>27.833567, "lon"=>-81.528056),
"/US/US-FL/bagdad-2869" => array("type"=>"cdp", "pop"=>1457, "name_ru"=>"Багдад", "name_en"=>"Bagdad", "lang"=>"en", "lat"=>30.588459, "lon"=>-87.034209),
"/US/US-FL/bal-harbour-2871" => array("type"=>"village", "pop"=>3118, "name_ru"=>"Бал Харбор", "name_en"=>"Bal Harbour", "lang"=>"en", "lat"=>25.891650, "lon"=>-80.124854),
"/US/US-FL/baldwin-2870" => array("type"=>"town", "pop"=>1546, "name_ru"=>"Болдуин", "name_en"=>"Baldwin", "lang"=>"en", "lat"=>30.304400, "lon"=>-81.975199),
"/US/US-FL/bartow-2872" => array("type"=>"city", "pop"=>16919, "name_ru"=>"Бартоу", "name_en"=>"Bartow", "lang"=>"en", "lat"=>27.895168, "lon"=>-81.847223),
"/US/US-FL/bascom-2873" => array("type"=>"town", "pop"=>107, "name_ru"=>"Баском", "name_en"=>"Bascom", "lang"=>"en", "lat"=>30.928582, "lon"=>-85.117138),
"/US/US-FL/baskin-2874" => array("type"=>"cdp", "pop"=>3834, "name_ru"=>"Баскин", "name_en"=>"Baskin", "lang"=>"en", "lat"=>27.893058, "lon"=>-82.809335),
"/US/US-FL/bassville-park-2875" => array("type"=>"cdp", "pop"=>2752, "name_ru"=>"Бассвилл Парк", "name_en"=>"Bassville Park", "lang"=>"en", "lat"=>28.838529, "lon"=>-81.767733),
"/US/US-FL/bay-harbor-islands-2876" => array("type"=>"town", "pop"=>4934, "name_ru"=>"Бэй Харбор Айлендс", "name_en"=>"Bay Harbor Islands", "lang"=>"en", "lat"=>25.886991, "lon"=>-80.134514),
"/US/US-FL/bay-hill-2877" => array("type"=>"cdp", "pop"=>5346, "name_ru"=>"Бэй Хилл", "name_en"=>"Bay Hill", "lang"=>"en", "lat"=>28.467650, "lon"=>-81.516283),
"/US/US-FL/bay-lake-2878" => array("type"=>"city", "pop"=>21, "name_ru"=>"Бэй Лэйк", "name_en"=>"Bay Lake", "lang"=>"en", "lat"=>28.388883, "lon"=>-81.576314),
"/US/US-FL/bay-pines-2880" => array("type"=>"cdp", "pop"=>4171, "name_ru"=>"Бэй Пайнс", "name_en"=>"Bay Pines", "lang"=>"en", "lat"=>27.815950, "lon"=>-82.775273),
"/US/US-FL/bayonet-point-2879" => array("type"=>"cdp", "pop"=>21860, "name_ru"=>"Байонет Поинт", "name_en"=>"Bayonet Point", "lang"=>"en", "lat"=>28.326257, "lon"=>-82.683478),
"/US/US-FL/bayshore-gardens-2881" => array("type"=>"cdp", "pop"=>17062, "name_ru"=>"Бэйшор Гарденс", "name_en"=>"Bayshore Gardens", "lang"=>"en", "lat"=>27.433475, "lon"=>-82.579967),
"/US/US-FL/beacon-square-2882" => array("type"=>"cdp", "pop"=>6265, "name_ru"=>"Бикон Сквер", "name_en"=>"Beacon Square", "lang"=>"en", "lat"=>28.211569, "lon"=>-82.749578),
"/US/US-FL/bee-ridge-2883" => array("type"=>"cdp", "pop"=>6406, "name_ru"=>"Би Ридж", "name_en"=>"Bee Ridge", "lang"=>"en", "lat"=>27.287350, "lon"=>-82.475577),
"/US/US-FL/bell-2884" => array("type"=>"town", "pop"=>416, "name_ru"=>"Белл", "name_en"=>"Bell", "lang"=>"en", "lat"=>29.755781, "lon"=>-82.861886),
"/US/US-FL/bellair-meadowbrook-terrace-2885" => array("type"=>"cdp", "pop"=>15606, "name_ru"=>"Беллэйр-Мидоубрук Террэйс", "name_en"=>"Bellair-Meadowbrook Terrace", "lang"=>"en", "lat"=>30.180150, "lon"=>-81.743599),
"/US/US-FL/belle-glade-2890" => array("type"=>"city", "pop"=>16681, "name_ru"=>"Белл Глэйд", "name_en"=>"Belle Glade", "lang"=>"en", "lat"=>26.685532, "lon"=>-80.671513),
"/US/US-FL/belle-glade-camp-2891" => array("type"=>"cdp", "pop"=>1616, "name_ru"=>"Белл Глэйд Кэмп", "name_en"=>"Belle Glade Camp", "lang"=>"en", "lat"=>26.657280, "lon"=>-80.682928),
"/US/US-FL/belle-isle-2892" => array("type"=>"city", "pop"=>6497, "name_ru"=>"Белл Айл", "name_en"=>"Belle Isle", "lang"=>"en", "lat"=>28.470133, "lon"=>-81.350336),
"/US/US-FL/belleair-2886" => array("type"=>"town", "pop"=>4073, "name_ru"=>"Беллэйр", "name_en"=>"Belleair", "lang"=>"en", "lat"=>27.936683, "lon"=>-82.812151),
"/US/US-FL/belleair-beach-2887" => array("type"=>"city", "pop"=>1586, "name_ru"=>"Беллэйр Бич", "name_en"=>"Belleair Beach", "lang"=>"en", "lat"=>27.925283, "lon"=>-82.837244),
"/US/US-FL/belleair-bluffs-2888" => array("type"=>"city", "pop"=>2164, "name_ru"=>"Беллэйр Блаффс", "name_en"=>"Belleair Bluffs", "lang"=>"en", "lat"=>27.919300, "lon"=>-82.819299),
"/US/US-FL/belleair-shore-2889" => array("type"=>"town", "pop"=>74, "name_ru"=>"Беллэйр Шор", "name_en"=>"Belleair Shore", "lang"=>"en", "lat"=>27.917422, "lon"=>-82.849452),
"/US/US-FL/belleview-2893" => array("type"=>"city", "pop"=>4467, "name_ru"=>"Беллевью", "name_en"=>"Belleview", "lang"=>"en", "lat"=>29.059844, "lon"=>-82.054828),
"/US/US-FL/bellview-2894" => array("type"=>"cdp", "pop"=>19386, "name_ru"=>"Беллвью", "name_en"=>"Bellview", "lang"=>"en", "lat"=>30.462460, "lon"=>-87.314559),
"/US/US-FL/beverly-beach-2895" => array("type"=>"town", "pop"=>585, "name_ru"=>"Беверли Бич", "name_en"=>"Beverly Beach", "lang"=>"en", "lat"=>29.516562, "lon"=>-81.148180),
"/US/US-FL/beverly-hills-2896" => array("type"=>"cdp", "pop"=>6163, "name_ru"=>"Беверли Хиллс", "name_en"=>"Beverly Hills", "lang"=>"en", "lat"=>28.912269, "lon"=>-82.464507),
"/US/US-FL/big-coppitt-key-2897" => array("type"=>"cdp", "pop"=>2388, "name_ru"=>"Биг Коппит Ки", "name_en"=>"Big Coppitt Key", "lang"=>"en", "lat"=>24.596233, "lon"=>-81.660317),
"/US/US-FL/big-pine-key-2898" => array("type"=>"cdp", "pop"=>4206, "name_ru"=>"Биг Пайн Ки", "name_en"=>"Big Pine Key", "lang"=>"en", "lat"=>24.686930, "lon"=>-81.366149),
"/US/US-FL/biscayne-park-2899" => array("type"=>"village", "pop"=>2962, "name_ru"=>"Бискэйн Парк", "name_en"=>"Biscayne Park", "lang"=>"en", "lat"=>25.881150, "lon"=>-80.181374),
"/US/US-FL/bithlo-2900" => array("type"=>"cdp", "pop"=>4834, "name_ru"=>"Битло", "name_en"=>"Bithlo", "lang"=>"en", "lat"=>28.569667, "lon"=>-81.109572),
"/US/US-FL/bloomingdale-2901" => array("type"=>"cdp", "pop"=>13912, "name_ru"=>"Блумингдэйл", "name_en"=>"Bloomingdale", "lang"=>"en", "lat"=>27.877231, "lon"=>-82.259925),
"/US/US-FL/blountstown-2902" => array("type"=>"city", "pop"=>2515, "name_ru"=>"Блантстаун", "name_en"=>"Blountstown", "lang"=>"en", "lat"=>30.443734, "lon"=>-85.045485),
"/US/US-FL/boca-del-mar-2903" => array("type"=>"cdp", "pop"=>17754, "name_ru"=>"Бока Дель Мар", "name_en"=>"Boca Del Mar", "lang"=>"en", "lat"=>26.344602, "lon"=>-80.146823),
"/US/US-FL/boca-pointe-2904" => array("type"=>"cdp", "pop"=>2147, "name_ru"=>"Бока Поинт", "name_en"=>"Boca Pointe", "lang"=>"en", "lat"=>26.332750, "lon"=>-80.159807),
"/US/US-FL/boca-raton-2905" => array("type"=>"city", "pop"=>85670, "name_ru"=>"Бока Ратон", "name_en"=>"Boca Raton", "lang"=>"en", "lat"=>26.372700, "lon"=>-80.106624),
"/US/US-FL/boca-west-2906" => array("type"=>"cdp", "pop"=>2847, "name_ru"=>"Бока Уэст", "name_en"=>"Boca West", "lang"=>"en", "lat"=>26.380342, "lon"=>-80.160062),
"/US/US-FL/bonifay-2907" => array("type"=>"city", "pop"=>2769, "name_ru"=>"Бонифэй", "name_en"=>"Bonifay", "lang"=>"en", "lat"=>30.784821, "lon"=>-85.683862),
"/US/US-FL/bonita-springs-2908" => array("type"=>"cdp", "pop"=>42730, "name_ru"=>"Бонита Спрингс", "name_en"=>"Bonita Springs", "lang"=>"en", "lat"=>26.340927, "lon"=>-81.785060),
"/US/US-FL/bowling-green-2909" => array("type"=>"city", "pop"=>2953, "name_ru"=>"Боулинг Грин", "name_en"=>"Bowling Green", "lang"=>"en", "lat"=>27.637288, "lon"=>-81.823810),
"/US/US-FL/boynton-beach-2910" => array("type"=>"city", "pop"=>68291, "name_ru"=>"Бойнтон Бич", "name_en"=>"Boynton Beach", "lang"=>"en", "lat"=>26.526830, "lon"=>-80.080687),
"/US/US-FL/bradenton-2911" => array("type"=>"city", "pop"=>53513, "name_ru"=>"Брэйдентон", "name_en"=>"Bradenton", "lang"=>"en", "lat"=>27.487900, "lon"=>-82.578974),
"/US/US-FL/bradenton-beach-2912" => array("type"=>"city", "pop"=>1553, "name_ru"=>"Брэйдентон Бич", "name_en"=>"Bradenton Beach", "lang"=>"en", "lat"=>27.463950, "lon"=>-82.697574),
"/US/US-FL/brandon-2913" => array("type"=>"cdp", "pop"=>57985, "name_ru"=>"Брэндон", "name_en"=>"Brandon", "lang"=>"en", "lat"=>27.929450, "lon"=>-82.291513),
"/US/US-FL/branford-2914" => array("type"=>"town", "pop"=>768, "name_ru"=>"Брэнфорд", "name_en"=>"Branford", "lang"=>"en", "lat"=>29.962147, "lon"=>-82.927607),
"/US/US-FL/brent-2915" => array("type"=>"cdp", "pop"=>21624, "name_ru"=>"Брент", "name_en"=>"Brent", "lang"=>"en", "lat"=>30.473767, "lon"=>-87.251739),
"/US/US-FL/briny-breezes-2916" => array("type"=>"town", "pop"=>392, "name_ru"=>"Брайни Бризес", "name_en"=>"Briny Breezes", "lang"=>"en", "lat"=>26.508700, "lon"=>-80.052749),
"/US/US-FL/bristol-2917" => array("type"=>"city", "pop"=>942, "name_ru"=>"Бристол", "name_en"=>"Bristol", "lang"=>"en", "lat"=>30.420258, "lon"=>-84.978937),
"/US/US-FL/broadview-park-2918" => array("type"=>"cdp", "pop"=>6109, "name_ru"=>"Броадвью Парк", "name_en"=>"Broadview Park", "lang"=>"en", "lat"=>26.098550, "lon"=>-80.209309),
"/US/US-FL/broadview-pompano-park-2919" => array("type"=>"cdp", "pop"=>5230, "name_ru"=>"Броадвью-Помпано Парк", "name_en"=>"Broadview-Pompano Park", "lang"=>"en", "lat"=>26.200733, "lon"=>-80.209458),
"/US/US-FL/bronson-2920" => array("type"=>"town", "pop"=>1059, "name_ru"=>"Бронсон", "name_en"=>"Bronson", "lang"=>"en", "lat"=>29.448760, "lon"=>-82.638672),
"/US/US-FL/brooker-2921" => array("type"=>"city", "pop"=>373, "name_ru"=>"Брукер", "name_en"=>"Brooker", "lang"=>"en", "lat"=>29.885991, "lon"=>-82.334044),
"/US/US-FL/brookridge-2922" => array("type"=>"cdp", "pop"=>2805, "name_ru"=>"Брукридж", "name_en"=>"Brookridge", "lang"=>"en", "lat"=>28.550711, "lon"=>-82.492134),
"/US/US-FL/brooksville-2923" => array("type"=>"city", "pop"=>8292, "name_ru"=>"Бруксвилл", "name_en"=>"Brooksville", "lang"=>"en", "lat"=>28.554423, "lon"=>-82.392955),
"/US/US-FL/browardale-2924" => array("type"=>"cdp", "pop"=>6257, "name_ru"=>"Бровардэйл", "name_en"=>"Browardale", "lang"=>"en", "lat"=>26.128300, "lon"=>-80.192810),
"/US/US-FL/brownsville-2925" => array("type"=>"cdp", "pop"=>15607, "name_ru"=>"Браунсвилл", "name_en"=>"Brownsville", "lang"=>"en", "lat"=>25.821296, "lon"=>-80.241500),
"/US/US-FL/buckhead-ridge-2926" => array("type"=>"cdp", "pop"=>1279, "name_ru"=>"Бакхед Ридж", "name_en"=>"Buckhead Ridge", "lang"=>"en", "lat"=>27.133495, "lon"=>-80.885739),
"/US/US-FL/buena-ventura-lakes-2927" => array("type"=>"cdp", "pop"=>14148, "name_ru"=>"Буена Виста Лэйкс", "name_en"=>"Buena Ventura Lakes", "lang"=>"en", "lat"=>28.335678, "lon"=>-81.353229),
"/US/US-FL/bunche-park-2928" => array("type"=>"cdp", "pop"=>4388, "name_ru"=>"Банч Парк", "name_en"=>"Bunche Park", "lang"=>"en", "lat"=>25.921050, "lon"=>-80.236958),
"/US/US-FL/bunnell-2929" => array("type"=>"city", "pop"=>3190, "name_ru"=>"Баннелл", "name_en"=>"Bunnell", "lang"=>"en", "lat"=>29.466791, "lon"=>-81.258873),
"/US/US-FL/bushnell-2930" => array("type"=>"city", "pop"=>2275, "name_ru"=>"Бушнелл", "name_en"=>"Bushnell", "lang"=>"en", "lat"=>28.664373, "lon"=>-82.114362),
"/US/US-FL/butler-beach-2931" => array("type"=>"cdp", "pop"=>3377, "name_ru"=>"Батлер Бич", "name_en"=>"Butler Beach", "lang"=>"en", "lat"=>29.798180, "lon"=>-81.267102),
"/US/US-FL/callahan-2932" => array("type"=>"town", "pop"=>1058, "name_ru"=>"Каллахан", "name_en"=>"Callahan", "lang"=>"en", "lat"=>30.561377, "lon"=>-81.832565),
"/US/US-FL/callaway-2933" => array("type"=>"city", "pop"=>13899, "name_ru"=>"Каллавэй", "name_en"=>"Callaway", "lang"=>"en", "lat"=>30.140900, "lon"=>-85.577824),
"/US/US-FL/campbell-2934" => array("type"=>"cdp", "pop"=>3884, "name_ru"=>"Кэмпбелл", "name_en"=>"Campbell", "lang"=>"en", "lat"=>28.257054, "lon"=>-81.452709),
"/US/US-FL/campbellton-2935" => array("type"=>"town", "pop"=>262, "name_ru"=>"Кэмпбеллтон", "name_en"=>"Campbellton", "lang"=>"en", "lat"=>30.951166, "lon"=>-85.397333),
"/US/US-FL/cape-canaveral-2936" => array("type"=>"city", "pop"=>10147, "name_ru"=>"Кэйп Канаверал", "name_en"=>"Cape Canaveral", "lang"=>"en", "lat"=>28.392484, "lon"=>-80.614619),
"/US/US-FL/cape-coral-2937" => array("type"=>"city", "pop"=>156835, "name_ru"=>"Кэйп Корал", "name_en"=>"Cape Coral", "lang"=>"en", "lat"=>26.637675, "lon"=>-81.997185),
"/US/US-FL/carol-city-2938" => array("type"=>"cdp", "pop"=>53331, "name_ru"=>"Кэрол Сити", "name_en"=>"Carol City", "lang"=>"en", "lat"=>25.944450, "lon"=>-80.270991),
"/US/US-FL/carrabelle-2939" => array("type"=>"city", "pop"=>1231, "name_ru"=>"Карабелл", "name_en"=>"Carrabelle", "lang"=>"en", "lat"=>29.847989, "lon"=>-84.670315),
"/US/US-FL/carrollwood-2940" => array("type"=>"cdp", "pop"=>7195, "name_ru"=>"Кэрроллвуд", "name_en"=>"Carrollwood", "lang"=>"en", "lat"=>28.049714, "lon"=>-82.493061),
"/US/US-FL/carrollwood-village-2941" => array("type"=>"cdp", "pop"=>15051, "name_ru"=>"Кэрроллвуд Виллидж", "name_en"=>"Carrollwood Village", "lang"=>"en", "lat"=>28.067325, "lon"=>-82.521159),
"/US/US-FL/caryville-2942" => array("type"=>"town", "pop"=>232, "name_ru"=>"Кэривилл", "name_en"=>"Caryville", "lang"=>"en", "lat"=>30.775101, "lon"=>-85.813132),
"/US/US-FL/casselberry-2943" => array("type"=>"city", "pop"=>24759, "name_ru"=>"Касселберри", "name_en"=>"Casselberry", "lang"=>"en", "lat"=>28.663417, "lon"=>-81.323518),
"/US/US-FL/cedar-grove-2944" => array("type"=>"town", "pop"=>5116, "name_ru"=>"Седар Гроув", "name_en"=>"Cedar Grove", "lang"=>"en", "lat"=>30.175833, "lon"=>-85.624477),
"/US/US-FL/cedar-key-2945" => array("type"=>"city", "pop"=>977, "name_ru"=>"Седар Ки", "name_en"=>"Cedar Key", "lang"=>"en", "lat"=>29.146273, "lon"=>-83.039567),
"/US/US-FL/center-hill-2946" => array("type"=>"city", "pop"=>1106, "name_ru"=>"Центр Хилл", "name_en"=>"Center Hill", "lang"=>"en", "lat"=>28.647449, "lon"=>-81.998114),
"/US/US-FL/century-2947" => array("type"=>"town", "pop"=>1863, "name_ru"=>"Сентури", "name_en"=>"Century", "lang"=>"en", "lat"=>30.977466, "lon"=>-87.261527),
"/US/US-FL/century-village-2948" => array("type"=>"cdp", "pop"=>8363, "name_ru"=>"Сентури Виллидж", "name_en"=>"Century Village", "lang"=>"en", "lat"=>26.713640, "lon"=>-80.129121),
"/US/US-FL/charlotte-harbor-2949" => array("type"=>"cdp", "pop"=>3327, "name_ru"=>"Шарлотт Харбор", "name_en"=>"Charlotte Harbor", "lang"=>"en", "lat"=>26.964737, "lon"=>-82.062494),
"/US/US-FL/charlotte-park-2950" => array("type"=>"cdp", "pop"=>2225, "name_ru"=>"Шарлотт Парк", "name_en"=>"Charlotte Park", "lang"=>"en", "lat"=>26.904046, "lon"=>-82.049196),
"/US/US-FL/chattahoochee-2951" => array("type"=>"city", "pop"=>3695, "name_ru"=>"Чатахучи", "name_en"=>"Chattahoochee", "lang"=>"en", "lat"=>30.697375, "lon"=>-84.836977),
"/US/US-FL/chiefland-2952" => array("type"=>"city", "pop"=>2133, "name_ru"=>"Чифлэнд", "name_en"=>"Chiefland", "lang"=>"en", "lat"=>29.478928, "lon"=>-82.861383),
"/US/US-FL/chipley-2953" => array("type"=>"city", "pop"=>3779, "name_ru"=>"Чипли", "name_en"=>"Chipley", "lang"=>"en", "lat"=>30.776124, "lon"=>-85.541475),
"/US/US-FL/chuluota-2954" => array("type"=>"cdp", "pop"=>1441, "name_ru"=>"Чулуота", "name_en"=>"Chuluota", "lang"=>"en", "lat"=>28.639544, "lon"=>-81.120348),
"/US/US-FL/cinco-bayou-2955" => array("type"=>"town", "pop"=>367, "name_ru"=>"Чинко Бэйоу", "name_en"=>"Cinco Bayou", "lang"=>"en", "lat"=>30.422300, "lon"=>-86.610278),
"/US/US-FL/citrus-springs-2956" => array("type"=>"cdp", "pop"=>2213, "name_ru"=>"Цитрус Спрингс", "name_en"=>"Citrus Springs", "lang"=>"en", "lat"=>28.996594, "lon"=>-82.474730),
"/US/US-FL/clearwater-2957" => array("type"=>"city", "pop"=>105774, "name_ru"=>"Клирвотер", "name_en"=>"Clearwater", "lang"=>"en", "lat"=>27.979400, "lon"=>-82.765549),
"/US/US-FL/clermont-2958" => array("type"=>"city", "pop"=>13100, "name_ru"=>"Клермонт", "name_en"=>"Clermont", "lang"=>"en", "lat"=>28.551190, "lon"=>-81.757656),
"/US/US-FL/cleveland-2959" => array("type"=>"cdp", "pop"=>2896, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>26.952071, "lon"=>-81.991207),
"/US/US-FL/clewiston-2960" => array("type"=>"city", "pop"=>7167, "name_ru"=>"Кльюистон", "name_en"=>"Clewiston", "lang"=>"en", "lat"=>26.753447, "lon"=>-80.939339),
"/US/US-FL/cloud-lake-2961" => array("type"=>"town", "pop"=>170, "name_ru"=>"Клауд Лэйк", "name_en"=>"Cloud Lake", "lang"=>"en", "lat"=>26.674700, "lon"=>-80.073587),
"/US/US-FL/cocoa-2962" => array("type"=>"city", "pop"=>16478, "name_ru"=>"Какао", "name_en"=>"Cocoa", "lang"=>"en", "lat"=>28.377001, "lon"=>-80.750173),
"/US/US-FL/cocoa-beach-2963" => array("type"=>"city", "pop"=>11920, "name_ru"=>"Какао Бич", "name_en"=>"Cocoa Beach", "lang"=>"en", "lat"=>28.335482, "lon"=>-80.626922),
"/US/US-FL/cocoa-west-2964" => array("type"=>"cdp", "pop"=>6160, "name_ru"=>"Какао Вест", "name_en"=>"Cocoa West", "lang"=>"en", "lat"=>28.357678, "lon"=>-80.769875),
"/US/US-FL/coconut-creek-2965" => array("type"=>"city", "pop"=>50436, "name_ru"=>"Коконат Крик", "name_en"=>"Coconut Creek", "lang"=>"en", "lat"=>26.279400, "lon"=>-80.184572),
"/US/US-FL/coleman-2966" => array("type"=>"city", "pop"=>763, "name_ru"=>"Коулман", "name_en"=>"Coleman", "lang"=>"en", "lat"=>28.801052, "lon"=>-82.069336),
"/US/US-FL/collier-manor-cresthaven-2967" => array("type"=>"cdp", "pop"=>7322, "name_ru"=>"Коллиер Мэнор-Крестхевен", "name_en"=>"Collier Manor-Cresthaven", "lang"=>"en", "lat"=>26.266000, "lon"=>-80.107408),
"/US/US-FL/combee-settlement-2968" => array("type"=>"cdp", "pop"=>5463, "name_ru"=>"Комби Сеттлмент", "name_en"=>"Combee Settlement", "lang"=>"en", "lat"=>28.057958, "lon"=>-81.905550),
"/US/US-FL/conway-2969" => array("type"=>"cdp", "pop"=>13159, "name_ru"=>"Конвэй", "name_en"=>"Conway", "lang"=>"en", "lat"=>28.495168, "lon"=>-81.332283),
"/US/US-FL/cooper-city-2970" => array("type"=>"city", "pop"=>29422, "name_ru"=>"Купер Сити", "name_en"=>"Cooper City", "lang"=>"en", "lat"=>26.044264, "lon"=>-80.289794),
"/US/US-FL/coral-gables-2971" => array("type"=>"city", "pop"=>42781, "name_ru"=>"Корал Гэйблс", "name_en"=>"Coral Gables", "lang"=>"en", "lat"=>25.699100, "lon"=>-80.264357),
"/US/US-FL/coral-springs-2972" => array("type"=>"city", "pop"=>125783, "name_ru"=>"Корал Спрингс", "name_en"=>"Coral Springs", "lang"=>"en", "lat"=>26.268917, "lon"=>-80.259040),
"/US/US-FL/coral-terrace-2973" => array("type"=>"cdp", "pop"=>23255, "name_ru"=>"Корал Террэйс", "name_en"=>"Coral Terrace", "lang"=>"en", "lat"=>25.745650, "lon"=>-80.304749),
"/US/US-FL/cortez-2974" => array("type"=>"cdp", "pop"=>4509, "name_ru"=>"Кортез", "name_en"=>"Cortez", "lang"=>"en", "lat"=>27.468450, "lon"=>-82.670635),
"/US/US-FL/cottondale-2975" => array("type"=>"town", "pop"=>878, "name_ru"=>"Коттондэйл", "name_en"=>"Cottondale", "lang"=>"en", "lat"=>30.798469, "lon"=>-85.376120),
"/US/US-FL/country-club-2976" => array("type"=>"cdp", "pop"=>3408, "name_ru"=>"Кантри Клаб", "name_en"=>"Country Club", "lang"=>"en", "lat"=>25.947900, "lon"=>-80.317299),
"/US/US-FL/country-club-trail-2977" => array("type"=>"cdp", "pop"=>4599, "name_ru"=>"Каунтри Клаб Трэйл", "name_en"=>"Country Club Trail", "lang"=>"en", "lat"=>26.527149, "lon"=>-80.120343),
"/US/US-FL/crescent-beach-2978" => array("type"=>"cdp", "pop"=>1081, "name_ru"=>"Кресцент Бич", "name_en"=>"Crescent Beach", "lang"=>"en", "lat"=>29.731059, "lon"=>-81.238802),
"/US/US-FL/crescent-city-2979" => array("type"=>"city", "pop"=>1814, "name_ru"=>"Кресцент Сити", "name_en"=>"Crescent City", "lang"=>"en", "lat"=>29.438255, "lon"=>-81.516572),
"/US/US-FL/crestview-2980" => array("type"=>"city", "pop"=>19444, "name_ru"=>"Крествью", "name_en"=>"Crestview", "lang"=>"en", "lat"=>30.752178, "lon"=>-86.574825),
"/US/US-FL/crooked-lake-park-2981" => array("type"=>"cdp", "pop"=>1575, "name_ru"=>"Крукед Лэйк Парк", "name_en"=>"Crooked Lake Park", "lang"=>"en", "lat"=>27.829904, "lon"=>-81.589740),
"/US/US-FL/cross-city-2982" => array("type"=>"town", "pop"=>1819, "name_ru"=>"Кросс Сити", "name_en"=>"Cross City", "lang"=>"en", "lat"=>29.638272, "lon"=>-83.124872),
"/US/US-FL/crystal-lake-2983" => array("type"=>"cdp", "pop"=>5300, "name_ru"=>"Кристал Лэйк", "name_en"=>"Crystal Lake", "lang"=>"en", "lat"=>28.035250, "lon"=>-81.908474),
"/US/US-FL/crystal-river-2984" => array("type"=>"city", "pop"=>3553, "name_ru"=>"Кристал Ривер", "name_en"=>"Crystal River", "lang"=>"en", "lat"=>28.901034, "lon"=>-82.597605),
"/US/US-FL/cudjoe-key-2985" => array("type"=>"cdp", "pop"=>1714, "name_ru"=>"Кадджо Ки", "name_en"=>"Cudjoe Key", "lang"=>"en", "lat"=>24.670991, "lon"=>-81.498486),
"/US/US-FL/cutler-2986" => array("type"=>"cdp", "pop"=>16201, "name_ru"=>"Катлер", "name_en"=>"Cutler", "lang"=>"en", "lat"=>25.627100, "lon"=>-80.316148),
"/US/US-FL/cutler-ridge-2987" => array("type"=>"cdp", "pop"=>21268, "name_ru"=>"Катлер Ридж", "name_en"=>"Cutler Ridge", "lang"=>"en", "lat"=>25.582300, "lon"=>-80.347024),
"/US/US-FL/cypress-gardens-2988" => array("type"=>"cdp", "pop"=>9188, "name_ru"=>"Сайпресс Гарденс", "name_en"=>"Cypress Gardens", "lang"=>"en", "lat"=>28.001157, "lon"=>-81.687042),
"/US/US-FL/cypress-lake-2989" => array("type"=>"cdp", "pop"=>10491, "name_ru"=>"Сайпресс Лэйк", "name_en"=>"Cypress Lake", "lang"=>"en", "lat"=>26.537700, "lon"=>-81.899574),
"/US/US-FL/cypress-lakes-2990" => array("type"=>"cdp", "pop"=>1260, "name_ru"=>"Сайпресс Лэйкс", "name_en"=>"Cypress Lakes", "lang"=>"en", "lat"=>26.727555, "lon"=>-80.122971),
"/US/US-FL/cypress-quarters-2991" => array("type"=>"cdp", "pop"=>1343, "name_ru"=>"Сайпресс Квотерс", "name_en"=>"Cypress Quarters", "lang"=>"en", "lat"=>27.247972, "lon"=>-80.813033),
"/US/US-FL/dade-city-2992" => array("type"=>"city", "pop"=>7223, "name_ru"=>"Дэйд Сити", "name_en"=>"Dade City", "lang"=>"en", "lat"=>28.361540, "lon"=>-82.195521),
"/US/US-FL/dade-city-north-2993" => array("type"=>"cdp", "pop"=>3058, "name_ru"=>"Дэйд Сити Норт", "name_en"=>"Dade City North", "lang"=>"en", "lat"=>28.383258, "lon"=>-82.193107),
"/US/US-FL/dania-2994" => array("type"=>"city", "pop"=>13024, "name_ru"=>"Даниа", "name_en"=>"Dania", "lang"=>"en", "lat"=>26.055817, "lon"=>-80.145074),
"/US/US-FL/davenport-2995" => array("type"=>"city", "pop"=>2560, "name_ru"=>"Дэйвенпорт", "name_en"=>"Davenport", "lang"=>"en", "lat"=>28.160496, "lon"=>-81.604943),
"/US/US-FL/davie-2996" => array("type"=>"town", "pop"=>90238, "name_ru"=>"Дэйви", "name_en"=>"Davie", "lang"=>"en", "lat"=>26.082150, "lon"=>-80.289246),
"/US/US-FL/daytona-beach-2997" => array("type"=>"city", "pop"=>64211, "name_ru"=>"Дайтона Бич", "name_en"=>"Daytona Beach", "lang"=>"en", "lat"=>29.192243, "lon"=>-81.052688),
"/US/US-FL/daytona-beach-shores-2998" => array("type"=>"city", "pop"=>5134, "name_ru"=>"Дайтона Бич Шорз", "name_en"=>"Daytona Beach Shores", "lang"=>"en", "lat"=>29.173515, "lon"=>-80.982395),
"/US/US-FL/de-bary-2999" => array("type"=>"cdp", "pop"=>7176, "name_ru"=>"Ди Бэйри", "name_en"=>"De Bary", "lang"=>"en", "lat"=>28.883424, "lon"=>-81.311756),
"/US/US-FL/de-funiak-springs-3001" => array("type"=>"city", "pop"=>5012, "name_ru"=>"Ди Фаниак Спрингс", "name_en"=>"De Funiak Springs", "lang"=>"en", "lat"=>30.719577, "lon"=>-86.123092),
"/US/US-FL/de-land-3002" => array("type"=>"city", "pop"=>27243, "name_ru"=>"Ди Лэнд", "name_en"=>"De Land", "lang"=>"en", "lat"=>29.036038, "lon"=>-81.297515),
"/US/US-FL/de-land-southwest-3003" => array("type"=>"cdp", "pop"=>1249, "name_ru"=>"Ди Лэнд Саутвест", "name_en"=>"De Land Southwest", "lang"=>"en", "lat"=>29.006895, "lon"=>-81.310957),
"/US/US-FL/de-leon-springs-3004" => array("type"=>"cdp", "pop"=>1481, "name_ru"=>"Ди Лайон Спрингс", "name_en"=>"De Leon Springs", "lang"=>"en", "lat"=>29.117440, "lon"=>-81.351785),
"/US/US-FL/deerfield-beach-3000" => array("type"=>"city", "pop"=>74584, "name_ru"=>"Дирфилд Бич", "name_en"=>"Deerfield Beach", "lang"=>"en", "lat"=>26.313133, "lon"=>-80.120741),
"/US/US-FL/del-rio-3006" => array("type"=>"cdp", "pop"=>8248, "name_ru"=>"Дель Рио", "name_en"=>"Del Rio", "lang"=>"en", "lat"=>28.008854, "lon"=>-82.399873),
"/US/US-FL/delray-beach-3005" => array("type"=>"city", "pop"=>64092, "name_ru"=>"Делрэй Бич", "name_en"=>"Delray Beach", "lang"=>"en", "lat"=>26.454667, "lon"=>-80.088682),
"/US/US-FL/deltona-3007" => array("type"=>"cdp", "pop"=>84408, "name_ru"=>"Делтона", "name_en"=>"Deltona", "lang"=>"en", "lat"=>28.909894, "lon"=>-81.208779),
"/US/US-FL/desoto-lakes-3008" => array("type"=>"cdp", "pop"=>2807, "name_ru"=>"Десото Лэйкс", "name_en"=>"Desoto Lakes", "lang"=>"en", "lat"=>27.371172, "lon"=>-82.489952),
"/US/US-FL/destin-3009" => array("type"=>"city", "pop"=>12563, "name_ru"=>"Дестин", "name_en"=>"Destin", "lang"=>"en", "lat"=>30.391100, "lon"=>-86.474409),
"/US/US-FL/doctor-phillips-3010" => array("type"=>"cdp", "pop"=>7963, "name_ru"=>"Доктор Филлипс", "name_en"=>"Doctor Phillips", "lang"=>"en", "lat"=>28.449233, "lon"=>-81.492486),
"/US/US-FL/doral-3011" => array("type"=>"cdp", "pop"=>23974, "name_ru"=>"Дорал", "name_en"=>"Doral", "lang"=>"en", "lat"=>25.819233, "lon"=>-80.355617),
"/US/US-FL/dover-3012" => array("type"=>"cdp", "pop"=>2606, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>27.992896, "lon"=>-82.220495),
"/US/US-FL/dundee-3013" => array("type"=>"town", "pop"=>3276, "name_ru"=>"Данди", "name_en"=>"Dundee", "lang"=>"en", "lat"=>28.020539, "lon"=>-81.620985),
"/US/US-FL/dunedin-3014" => array("type"=>"city", "pop"=>35988, "name_ru"=>"Дюндин", "name_en"=>"Dunedin", "lang"=>"en", "lat"=>28.044379, "lon"=>-82.808446),
"/US/US-FL/dunnellon-3015" => array("type"=>"city", "pop"=>2009, "name_ru"=>"Даннеллон", "name_en"=>"Dunnellon", "lang"=>"en", "lat"=>29.049537, "lon"=>-82.448450),
"/US/US-FL/eagle-lake-3016" => array("type"=>"city", "pop"=>2734, "name_ru"=>"Игл Лэйк", "name_en"=>"Eagle Lake", "lang"=>"en", "lat"=>27.978200, "lon"=>-81.757678),
"/US/US-FL/east-lake-orient-park-3017" => array("type"=>"cdp", "pop"=>6171, "name_ru"=>"Ист Лэйк-Ориент Парк", "name_en"=>"East Lake-Orient Park", "lang"=>"en", "lat"=>27.982550, "lon"=>-82.378474),
"/US/US-FL/east-naples-3018" => array("type"=>"cdp", "pop"=>22951, "name_ru"=>"Ист Нэплс", "name_en"=>"East Naples", "lang"=>"en", "lat"=>26.118952, "lon"=>-81.753926),
"/US/US-FL/east-palatka-3019" => array("type"=>"cdp", "pop"=>1989, "name_ru"=>"Ист Палатка", "name_en"=>"East Palatka", "lang"=>"en", "lat"=>29.651053, "lon"=>-81.599386),
"/US/US-FL/eastpoint-3020" => array("type"=>"cdp", "pop"=>1577, "name_ru"=>"Истпоинт", "name_en"=>"Eastpoint", "lang"=>"en", "lat"=>29.751316, "lon"=>-84.868075),
"/US/US-FL/eatonville-3021" => array("type"=>"town", "pop"=>2362, "name_ru"=>"Истонвилл", "name_en"=>"Eatonville", "lang"=>"en", "lat"=>28.616900, "lon"=>-81.390896),
"/US/US-FL/ebro-3022" => array("type"=>"town", "pop"=>267, "name_ru"=>"Эбро", "name_en"=>"Ebro", "lang"=>"en", "lat"=>30.442956, "lon"=>-85.880856),
"/US/US-FL/edgewater-3023" => array("type"=>"city", "pop"=>21415, "name_ru"=>"Эджвотер", "name_en"=>"Edgewater", "lang"=>"en", "lat"=>28.967650, "lon"=>-80.904841),
"/US/US-FL/edgewood-3024" => array("type"=>"city", "pop"=>2080, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>28.485600, "lon"=>-81.372099),
"/US/US-FL/eglin-afb-3025" => array("type"=>"cdp", "pop"=>8347, "name_ru"=>"Эглин АФБ", "name_en"=>"Eglin AFB", "lang"=>"en", "lat"=>30.460418, "lon"=>-86.549927),
"/US/US-FL/egypt-lake-3026" => array("type"=>"cdp", "pop"=>14580, "name_ru"=>"Египт Лэйк", "name_en"=>"Egypt Lake", "lang"=>"en", "lat"=>28.016733, "lon"=>-82.495589),
"/US/US-FL/el-portal-3029" => array("type"=>"village", "pop"=>2317, "name_ru"=>"Эль Портал", "name_en"=>"El Portal", "lang"=>"en", "lat"=>25.854800, "lon"=>-80.196415),
"/US/US-FL/elfers-3027" => array("type"=>"cdp", "pop"=>12356, "name_ru"=>"Элферс", "name_en"=>"Elfers", "lang"=>"en", "lat"=>28.213388, "lon"=>-82.722664),
"/US/US-FL/ellenton-3028" => array("type"=>"cdp", "pop"=>2573, "name_ru"=>"Эллентон", "name_en"=>"Ellenton", "lang"=>"en", "lat"=>27.523650, "lon"=>-82.527974),
"/US/US-FL/englewood-3030" => array("type"=>"cdp", "pop"=>15025, "name_ru"=>"Энглвуд", "name_en"=>"Englewood", "lang"=>"en", "lat"=>26.959500, "lon"=>-82.354277),
"/US/US-FL/ensley-3031" => array("type"=>"cdp", "pop"=>16362, "name_ru"=>"Энсли", "name_en"=>"Ensley", "lang"=>"en", "lat"=>30.526855, "lon"=>-87.276616),
"/US/US-FL/estero-3032" => array("type"=>"cdp", "pop"=>3177, "name_ru"=>"Эстеро", "name_en"=>"Estero", "lang"=>"en", "lat"=>26.441172, "lon"=>-81.810371),
"/US/US-FL/esto-3033" => array("type"=>"town", "pop"=>361, "name_ru"=>"Эсто", "name_en"=>"Esto", "lang"=>"en", "lat"=>30.984998, "lon"=>-85.642156),
"/US/US-FL/eustis-3034" => array("type"=>"city", "pop"=>19129, "name_ru"=>"Эустис", "name_en"=>"Eustis", "lang"=>"en", "lat"=>28.853850, "lon"=>-81.685249),
"/US/US-FL/everglades-3035" => array("type"=>"city", "pop"=>610, "name_ru"=>"Эверглэйдс", "name_en"=>"Everglades", "lang"=>"en", "lat"=>25.857163, "lon"=>-81.386760),
"/US/US-FL/fairview-shores-3036" => array("type"=>"cdp", "pop"=>13192, "name_ru"=>"Фэйрвью Шорс", "name_en"=>"Fairview Shores", "lang"=>"en", "lat"=>28.600250, "lon"=>-81.390874),
"/US/US-FL/fanning-springs-3037" => array("type"=>"city", "pop"=>829, "name_ru"=>"Фэннин Спрингс", "name_en"=>"Fanning Springs", "lang"=>"en", "lat"=>29.583715, "lon"=>-82.930537),
"/US/US-FL/feather-sound-3038" => array("type"=>"cdp", "pop"=>2690, "name_ru"=>"Фитер Саунд", "name_en"=>"Feather Sound", "lang"=>"en", "lat"=>27.909350, "lon"=>-82.657095),
"/US/US-FL/fellsmere-3039" => array("type"=>"city", "pop"=>4920, "name_ru"=>"Феллсмер", "name_en"=>"Fellsmere", "lang"=>"en", "lat"=>27.766454, "lon"=>-80.596867),
"/US/US-FL/fern-park-3041" => array("type"=>"cdp", "pop"=>8294, "name_ru"=>"Ферн Парк", "name_en"=>"Fern Park", "lang"=>"en", "lat"=>28.648200, "lon"=>-81.345810),
"/US/US-FL/fernandina-beach-3040" => array("type"=>"city", "pop"=>11610, "name_ru"=>"Фернандина Бич", "name_en"=>"Fernandina Beach", "lang"=>"en", "lat"=>30.657657, "lon"=>-81.449164),
"/US/US-FL/ferry-pass-3042" => array("type"=>"cdp", "pop"=>26301, "name_ru"=>"Ферри Пасс", "name_en"=>"Ferry Pass", "lang"=>"en", "lat"=>30.518183, "lon"=>-87.209235),
"/US/US-FL/five-points-3043" => array("type"=>"cdp", "pop"=>1136, "name_ru"=>"Файв Поинтс", "name_en"=>"Five Points", "lang"=>"en", "lat"=>30.222661, "lon"=>-82.644017),
"/US/US-FL/flagler-beach-3044" => array("type"=>"city", "pop"=>5859, "name_ru"=>"Флэглер Бич", "name_en"=>"Flagler Beach", "lang"=>"en", "lat"=>29.470251, "lon"=>-81.129959),
"/US/US-FL/floral-city-3045" => array("type"=>"cdp", "pop"=>2609, "name_ru"=>"Флорал Сити", "name_en"=>"Floral City", "lang"=>"en", "lat"=>28.741381, "lon"=>-82.291499),
"/US/US-FL/florida-city-3046" => array("type"=>"city", "pop"=>9935, "name_ru"=>"Флорида Сити", "name_en"=>"Florida City", "lang"=>"en", "lat"=>25.449850, "lon"=>-80.482299),
"/US/US-FL/florida-ridge-3047" => array("type"=>"cdp", "pop"=>12218, "name_ru"=>"Флорида Ридж", "name_en"=>"Florida Ridge", "lang"=>"en", "lat"=>27.580039, "lon"=>-80.386916),
"/US/US-FL/forest-city-3048" => array("type"=>"cdp", "pop"=>10638, "name_ru"=>"Форест Сити", "name_en"=>"Forest City", "lang"=>"en", "lat"=>28.661200, "lon"=>-81.445063),
"/US/US-FL/forest-island-park-3049" => array("type"=>"cdp", "pop"=>5988, "name_ru"=>"Форест Айленд Парк", "name_en"=>"Forest Island Park", "lang"=>"en", "lat"=>26.491065, "lon"=>-81.861081),
"/US/US-FL/fort-lauderdale-3050" => array("type"=>"city", "pop"=>183126, "name_ru"=>"Форт Лоудердэйл", "name_en"=>"Fort Lauderdale", "lang"=>"en", "lat"=>26.142550, "lon"=>-80.139499),
"/US/US-FL/fort-meade-3051" => array("type"=>"city", "pop"=>5723, "name_ru"=>"Форт Миад", "name_en"=>"Fort Meade", "lang"=>"en", "lat"=>27.752288, "lon"=>-81.795615),
"/US/US-FL/fort-myers-3052" => array("type"=>"city", "pop"=>65394, "name_ru"=>"Форт Майерс", "name_en"=>"Fort Myers", "lang"=>"en", "lat"=>26.630650, "lon"=>-81.850571),
"/US/US-FL/fort-myers-beach-3053" => array("type"=>"cdp", "pop"=>6410, "name_ru"=>"Форт Майерс Бич", "name_en"=>"Fort Myers Beach", "lang"=>"en", "lat"=>26.449971, "lon"=>-81.932957),
"/US/US-FL/fort-myers-shores-3054" => array("type"=>"cdp", "pop"=>5460, "name_ru"=>"Форт Майерс Шорс", "name_en"=>"Fort Myers Shores", "lang"=>"en", "lat"=>26.713509, "lon"=>-81.738695),
"/US/US-FL/fort-pierce-3055" => array("type"=>"city", "pop"=>40885, "name_ru"=>"Форт Пирс", "name_en"=>"Fort Pierce", "lang"=>"en", "lat"=>27.441250, "lon"=>-80.322034),
"/US/US-FL/fort-pierce-north-3056" => array("type"=>"cdp", "pop"=>5833, "name_ru"=>"Форт Пирс Норт", "name_en"=>"Fort Pierce North", "lang"=>"en", "lat"=>27.474783, "lon"=>-80.356184),
"/US/US-FL/fort-pierce-south-3057" => array("type"=>"cdp", "pop"=>5320, "name_ru"=>"Форт Пирс Саут", "name_en"=>"Fort Pierce South", "lang"=>"en", "lat"=>27.410750, "lon"=>-80.355137),
"/US/US-FL/fort-walton-beach-3058" => array("type"=>"city", "pop"=>18880, "name_ru"=>"Форт Волтон Бич", "name_en"=>"Fort Walton Beach", "lang"=>"en", "lat"=>30.422433, "lon"=>-86.618487),
"/US/US-FL/fort-white-3059" => array("type"=>"town", "pop"=>483, "name_ru"=>"Форт Уайт", "name_en"=>"Fort White", "lang"=>"en", "lat"=>29.921495, "lon"=>-82.714511),
"/US/US-FL/freeport-3060" => array("type"=>"city", "pop"=>1768, "name_ru"=>"Фрипорт", "name_en"=>"Freeport", "lang"=>"en", "lat"=>30.500088, "lon"=>-86.135245),
"/US/US-FL/frostproof-3061" => array("type"=>"city", "pop"=>2891, "name_ru"=>"Фростпруф", "name_en"=>"Frostproof", "lang"=>"en", "lat"=>27.746096, "lon"=>-81.531335),
"/US/US-FL/fruit-cove-3062" => array("type"=>"cdp", "pop"=>5904, "name_ru"=>"Фрут Коув", "name_en"=>"Fruit Cove", "lang"=>"en", "lat"=>30.100912, "lon"=>-81.618507),
"/US/US-FL/fruitland-park-3063" => array("type"=>"city", "pop"=>4293, "name_ru"=>"Фрутлэнд Парк", "name_en"=>"Fruitland Park", "lang"=>"en", "lat"=>28.857067, "lon"=>-81.909463),
"/US/US-FL/fruitville-3064" => array("type"=>"cdp", "pop"=>9808, "name_ru"=>"Фрутвилл", "name_en"=>"Fruitville", "lang"=>"en", "lat"=>27.332450, "lon"=>-82.463547),
"/US/US-FL/fussels-corner-3065" => array("type"=>"cdp", "pop"=>3840, "name_ru"=>"Фасселс Корнер", "name_en"=>"Fussels Corner", "lang"=>"en", "lat"=>28.053912, "lon"=>-81.860723),
"/US/US-FL/gainesville-3066" => array("type"=>"city", "pop"=>114916, "name_ru"=>"Гейнсвилл", "name_en"=>"Gainesville", "lang"=>"en", "lat"=>29.674150, "lon"=>-82.336277),
"/US/US-FL/gandy-3067" => array("type"=>"cdp", "pop"=>3164, "name_ru"=>"Гэнди", "name_en"=>"Gandy", "lang"=>"en", "lat"=>27.900000, "lon"=>-82.603712),
"/US/US-FL/gibsonia-3068" => array("type"=>"cdp", "pop"=>5168, "name_ru"=>"Гибсониа", "name_en"=>"Gibsonia", "lang"=>"en", "lat"=>28.117931, "lon"=>-81.968350),
"/US/US-FL/gibsonton-3069" => array("type"=>"cdp", "pop"=>7706, "name_ru"=>"Гибсонтон", "name_en"=>"Gibsonton", "lang"=>"en", "lat"=>27.840967, "lon"=>-82.370353),
"/US/US-FL/gifford-3070" => array("type"=>"cdp", "pop"=>6278, "name_ru"=>"Гиффорд", "name_en"=>"Gifford", "lang"=>"en", "lat"=>27.675324, "lon"=>-80.409641),
"/US/US-FL/gladeview-3071" => array("type"=>"cdp", "pop"=>15637, "name_ru"=>"Глэйдвью", "name_en"=>"Gladeview", "lang"=>"en", "lat"=>25.838950, "lon"=>-80.235899),
"/US/US-FL/glen-ridge-3073" => array("type"=>"town", "pop"=>281, "name_ru"=>"Глен Ридж", "name_en"=>"Glen Ridge", "lang"=>"en", "lat"=>26.671450, "lon"=>-80.076333),
"/US/US-FL/glen-st.-mary-3074" => array("type"=>"town", "pop"=>503, "name_ru"=>"Глен Ст. Мэри", "name_en"=>"Glen St. Mary", "lang"=>"en", "lat"=>30.273940, "lon"=>-82.161145),
"/US/US-FL/glencoe-3072" => array("type"=>"cdp", "pop"=>2282, "name_ru"=>"Гленко", "name_en"=>"Glencoe", "lang"=>"en", "lat"=>29.007067, "lon"=>-80.965866),
"/US/US-FL/glenvar-heights-3075" => array("type"=>"cdp", "pop"=>14823, "name_ru"=>"Гленвар Хайтс", "name_en"=>"Glenvar Heights", "lang"=>"en", "lat"=>25.708550, "lon"=>-80.316017),
"/US/US-FL/golden-beach-3076" => array("type"=>"town", "pop"=>875, "name_ru"=>"Голден Бич", "name_en"=>"Golden Beach", "lang"=>"en", "lat"=>25.964500, "lon"=>-80.122714),
"/US/US-FL/golden-gate-3077" => array("type"=>"cdp", "pop"=>14148, "name_ru"=>"Голден Гейт", "name_en"=>"Golden Gate", "lang"=>"en", "lat"=>26.183084, "lon"=>-81.703883),
"/US/US-FL/golden-glades-3078" => array("type"=>"cdp", "pop"=>25474, "name_ru"=>"Голден Глэйдс", "name_en"=>"Golden Glades", "lang"=>"en", "lat"=>25.911500, "lon"=>-80.200678),
"/US/US-FL/golden-lakes-3079" => array("type"=>"cdp", "pop"=>3867, "name_ru"=>"Голден Лэйкс", "name_en"=>"Golden Lakes", "lang"=>"en", "lat"=>26.700300, "lon"=>-80.164999),
"/US/US-FL/goldenrod-3080" => array("type"=>"cdp", "pop"=>12362, "name_ru"=>"Голденрод", "name_en"=>"Goldenrod", "lang"=>"en", "lat"=>28.611500, "lon"=>-81.291678),
"/US/US-FL/golf-3081" => array("type"=>"village", "pop"=>277, "name_ru"=>"Гольф", "name_en"=>"Golf", "lang"=>"en", "lat"=>26.505200, "lon"=>-80.110199),
"/US/US-FL/golfview-3082" => array("type"=>"town", "pop"=>153, "name_ru"=>"Гольфвью", "name_en"=>"Golfview", "lang"=>"en", "lat"=>26.689100, "lon"=>-80.111559),
"/US/US-FL/gonzalez-3083" => array("type"=>"cdp", "pop"=>7669, "name_ru"=>"Гонзалез", "name_en"=>"Gonzalez", "lang"=>"en", "lat"=>30.573152, "lon"=>-87.290427),
"/US/US-FL/goulding-3084" => array("type"=>"cdp", "pop"=>4159, "name_ru"=>"Гаулдинг", "name_en"=>"Goulding", "lang"=>"en", "lat"=>30.438800, "lon"=>-87.230449),
"/US/US-FL/goulds-3085" => array("type"=>"cdp", "pop"=>7284, "name_ru"=>"Гаулдс", "name_en"=>"Goulds", "lang"=>"en", "lat"=>25.561500, "lon"=>-80.387689),
"/US/US-FL/graceville-3086" => array("type"=>"city", "pop"=>2425, "name_ru"=>"Грэйсвилл", "name_en"=>"Graceville", "lang"=>"en", "lat"=>30.960492, "lon"=>-85.513373),
"/US/US-FL/grand-ridge-3087" => array("type"=>"town", "pop"=>802, "name_ru"=>"Гранд Ридж", "name_en"=>"Grand Ridge", "lang"=>"en", "lat"=>30.717025, "lon"=>-85.017338),
"/US/US-FL/greater-northdale-3088" => array("type"=>"cdp", "pop"=>16318, "name_ru"=>"Гритер Нортдэйл", "name_en"=>"Greater Northdale", "lang"=>"en", "lat"=>28.104178, "lon"=>-82.518366),
"/US/US-FL/green-cove-springs-3090" => array("type"=>"city", "pop"=>6551, "name_ru"=>"Грин Коув Спрингс", "name_en"=>"Green Cove Springs", "lang"=>"en", "lat"=>29.989731, "lon"=>-81.678855),
"/US/US-FL/greenacres-city-3089" => array("type"=>"city", "pop"=>18683, "name_ru"=>"Гринэкрс Сити", "name_en"=>"Greenacres City", "lang"=>"en", "lat"=>26.629733, "lon"=>-80.139658),
"/US/US-FL/greensboro-3091" => array("type"=>"town", "pop"=>600, "name_ru"=>"Гринсборо", "name_en"=>"Greensboro", "lang"=>"en", "lat"=>30.570232, "lon"=>-84.744375),
"/US/US-FL/greenville-3092" => array("type"=>"town", "pop"=>809, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>30.467020, "lon"=>-83.635231),
"/US/US-FL/greenwood-3093" => array("type"=>"town", "pop"=>748, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>30.870707, "lon"=>-85.160370),
"/US/US-FL/gretna-3094" => array("type"=>"city", "pop"=>1609, "name_ru"=>"Гретна", "name_en"=>"Gretna", "lang"=>"en", "lat"=>30.614112, "lon"=>-84.662163),
"/US/US-FL/grove-city-3095" => array("type"=>"cdp", "pop"=>2374, "name_ru"=>"Гроув Сити", "name_en"=>"Grove City", "lang"=>"en", "lat"=>26.907305, "lon"=>-82.322167),
"/US/US-FL/groveland-3096" => array("type"=>"city", "pop"=>7647, "name_ru"=>"Гроувлэнд", "name_en"=>"Groveland", "lang"=>"en", "lat"=>28.562971, "lon"=>-81.852131),
"/US/US-FL/gulf-breeze-3097" => array("type"=>"city", "pop"=>6580, "name_ru"=>"Галф Бриз", "name_en"=>"Gulf Breeze", "lang"=>"en", "lat"=>30.369001, "lon"=>-87.176161),
"/US/US-FL/gulf-gate-estates-3098" => array("type"=>"cdp", "pop"=>11622, "name_ru"=>"Галф Гейт Естэйтс", "name_en"=>"Gulf Gate Estates", "lang"=>"en", "lat"=>27.259100, "lon"=>-82.506849),
"/US/US-FL/gulf-stream-3100" => array("type"=>"town", "pop"=>735, "name_ru"=>"Галф Стрим", "name_en"=>"Gulf Stream", "lang"=>"en", "lat"=>26.485550, "lon"=>-80.058008),
"/US/US-FL/gulfport-3099" => array("type"=>"city", "pop"=>12327, "name_ru"=>"Галфпорт", "name_en"=>"Gulfport", "lang"=>"en", "lat"=>27.746350, "lon"=>-82.710199),
"/US/US-FL/haines-city-3101" => array("type"=>"city", "pop"=>19056, "name_ru"=>"Хайнес Сити", "name_en"=>"Haines City", "lang"=>"en", "lat"=>28.109016, "lon"=>-81.628924),
"/US/US-FL/hallandale-3102" => array("type"=>"city", "pop"=>30996, "name_ru"=>"Халландэйл", "name_en"=>"Hallandale", "lang"=>"en", "lat"=>25.985562, "lon"=>-80.141689),
"/US/US-FL/hammocks-3103" => array("type"=>"cdp", "pop"=>10897, "name_ru"=>"Хаммокс", "name_en"=>"Hammocks", "lang"=>"en", "lat"=>25.671018, "lon"=>-80.444711),
"/US/US-FL/hampton-3104" => array("type"=>"city", "pop"=>457, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>29.864018, "lon"=>-82.137178),
"/US/US-FL/hamptons-at-boca-raton-3105" => array("type"=>"cdp", "pop"=>11686, "name_ru"=>"Хэмптонс на Бока Рэйтон", "name_en"=>"Hamptons at Boca Raton", "lang"=>"en", "lat"=>26.382295, "lon"=>-80.185974),
"/US/US-FL/harbor-bluffs-3106" => array("type"=>"cdp", "pop"=>2659, "name_ru"=>"Харбор Блаффс", "name_en"=>"Harbor Bluffs", "lang"=>"en", "lat"=>27.909100, "lon"=>-82.827850),
"/US/US-FL/harbour-heights-3107" => array("type"=>"cdp", "pop"=>2523, "name_ru"=>"Харбор Хайтс", "name_en"=>"Harbour Heights", "lang"=>"en", "lat"=>26.992505, "lon"=>-82.007885),
"/US/US-FL/harlem-3108" => array("type"=>"cdp", "pop"=>2826, "name_ru"=>"Гарлем", "name_en"=>"Harlem", "lang"=>"en", "lat"=>26.735114, "lon"=>-80.951676),
"/US/US-FL/hastings-3109" => array("type"=>"town", "pop"=>740, "name_ru"=>"Гастингс", "name_en"=>"Hastings", "lang"=>"en", "lat"=>29.714909, "lon"=>-81.507893),
"/US/US-FL/havana-3110" => array("type"=>"town", "pop"=>1695, "name_ru"=>"Гавана", "name_en"=>"Havana", "lang"=>"en", "lat"=>30.624050, "lon"=>-84.415049),
"/US/US-FL/haverhill-3111" => array("type"=>"town", "pop"=>1562, "name_ru"=>"Хэйверхилл", "name_en"=>"Haverhill", "lang"=>"en", "lat"=>26.691074, "lon"=>-80.121677),
"/US/US-FL/hawthorne-3112" => array("type"=>"city", "pop"=>1553, "name_ru"=>"Хоуторн", "name_en"=>"Hawthorne", "lang"=>"en", "lat"=>29.591408, "lon"=>-82.085441),
"/US/US-FL/hawthorne-3113" => array("type"=>"cdp", "pop"=>1804, "name_ru"=>"Хоуторн", "name_en"=>"Hawthorne", "lang"=>"en", "lat"=>28.761224, "lon"=>-81.871034),
"/US/US-FL/hernando-3114" => array("type"=>"cdp", "pop"=>171689, "name_ru"=>"Гернандо", "name_en"=>"Hernando", "lang"=>"en", "lat"=>28.907251, "lon"=>-82.374562),
"/US/US-FL/hernando-beach-3115" => array("type"=>"cdp", "pop"=>1767, "name_ru"=>"Гернандо Бич", "name_en"=>"Hernando Beach", "lang"=>"en", "lat"=>28.469045, "lon"=>-82.659374),
"/US/US-FL/hialeah-3116" => array("type"=>"city", "pop"=>210542, "name_ru"=>"Хайалиа", "name_en"=>"Hialeah", "lang"=>"en", "lat"=>25.861750, "lon"=>-80.296772),
"/US/US-FL/hialeah-gardens-3117" => array("type"=>"city", "pop"=>19631, "name_ru"=>"Хайалиа Гарденс", "name_en"=>"Hialeah Gardens", "lang"=>"en", "lat"=>25.878167, "lon"=>-80.347413),
"/US/US-FL/high-point-3121" => array("type"=>"cdp", "pop"=>2814, "name_ru"=>"Хай Поинт", "name_en"=>"High Point", "lang"=>"en", "lat"=>28.546844, "lon"=>-82.525575),
"/US/US-FL/high-point-3122" => array("type"=>"cdp", "pop"=>2288, "name_ru"=>"Хай Поинт", "name_en"=>"High Point", "lang"=>"en", "lat"=>26.465475, "lon"=>-80.128277),
"/US/US-FL/high-springs-3124" => array("type"=>"city", "pop"=>4629, "name_ru"=>"Хай Спрингс", "name_en"=>"High Springs", "lang"=>"en", "lat"=>29.822088, "lon"=>-82.591915),
"/US/US-FL/highland-beach-3118" => array("type"=>"town", "pop"=>3965, "name_ru"=>"Хайленд Бич", "name_en"=>"Highland Beach", "lang"=>"en", "lat"=>26.411948, "lon"=>-80.066669),
"/US/US-FL/highland-city-3119" => array("type"=>"cdp", "pop"=>1919, "name_ru"=>"Хайленд Сити", "name_en"=>"Highland City", "lang"=>"en", "lat"=>27.960600, "lon"=>-81.879337),
"/US/US-FL/highland-park-3120" => array("type"=>"village", "pop"=>275, "name_ru"=>"Хайленд Парк", "name_en"=>"Highland Park", "lang"=>"en", "lat"=>27.865104, "lon"=>-81.563119),
"/US/US-FL/highpoint-3123" => array("type"=>"cdp", "pop"=>13818, "name_ru"=>"Хайпоинт", "name_en"=>"Highpoint", "lang"=>"en", "lat"=>27.916672, "lon"=>-82.713124),
"/US/US-FL/hiland-park-3125" => array("type"=>"cdp", "pop"=>3865, "name_ru"=>"Хайленд Парк", "name_en"=>"Hiland Park", "lang"=>"en", "lat"=>30.199300, "lon"=>-85.629899),
"/US/US-FL/hillcrest-heights-3126" => array("type"=>"town", "pop"=>295, "name_ru"=>"Хиллкрест Хайтс", "name_en"=>"Hillcrest Heights", "lang"=>"en", "lat"=>27.823784, "lon"=>-81.533315),
"/US/US-FL/hilliard-3127" => array("type"=>"town", "pop"=>2921, "name_ru"=>"Хиллиард", "name_en"=>"Hilliard", "lang"=>"en", "lat"=>30.694895, "lon"=>-81.922662),
"/US/US-FL/hillsboro-beach-3128" => array("type"=>"town", "pop"=>2254, "name_ru"=>"Хиллсборо Бич", "name_en"=>"Hillsboro Beach", "lang"=>"en", "lat"=>26.277650, "lon"=>-80.080526),
"/US/US-FL/hobe-sound-3129" => array("type"=>"cdp", "pop"=>11507, "name_ru"=>"Хоуб Саунд", "name_en"=>"Hobe Sound", "lang"=>"en", "lat"=>27.072445, "lon"=>-80.139562),
"/US/US-FL/holden-heights-3130" => array("type"=>"cdp", "pop"=>4387, "name_ru"=>"Холден Хайтс", "name_en"=>"Holden Heights", "lang"=>"en", "lat"=>28.496317, "lon"=>-81.388075),
"/US/US-FL/holiday-3131" => array("type"=>"cdp", "pop"=>19360, "name_ru"=>"Холидэй", "name_en"=>"Holiday", "lang"=>"en", "lat"=>28.185926, "lon"=>-82.742176),
"/US/US-FL/holly-hill-3132" => array("type"=>"city", "pop"=>13199, "name_ru"=>"Холли Хилл", "name_en"=>"Holly Hill", "lang"=>"en", "lat"=>29.242600, "lon"=>-81.044094),
"/US/US-FL/hollywood-3133" => array("type"=>"city", "pop"=>141740, "name_ru"=>"Холливуд", "name_en"=>"Hollywood", "lang"=>"en", "lat"=>26.029817, "lon"=>-80.164762),
"/US/US-FL/holmes-beach-3134" => array("type"=>"city", "pop"=>5017, "name_ru"=>"Холмс Бич", "name_en"=>"Holmes Beach", "lang"=>"en", "lat"=>27.510017, "lon"=>-82.715724),
"/US/US-FL/homestead-3135" => array("type"=>"city", "pop"=>57936, "name_ru"=>"Хоумстид", "name_en"=>"Homestead", "lang"=>"en", "lat"=>25.461893, "lon"=>-80.454208),
"/US/US-FL/homestead-afb-3136" => array("type"=>"cdp", "pop"=>5153, "name_ru"=>"Хоумстид АФБ", "name_en"=>"Homestead AFB", "lang"=>"en", "lat"=>25.495650, "lon"=>-80.387703),
"/US/US-FL/homosassa-3137" => array("type"=>"cdp", "pop"=>2113, "name_ru"=>"Хомосасса", "name_en"=>"Homosassa", "lang"=>"en", "lat"=>28.781225, "lon"=>-82.618365),
"/US/US-FL/homosassa-springs-3138" => array("type"=>"cdp", "pop"=>6271, "name_ru"=>"Хомосасса Спрингс", "name_en"=>"Homosassa Springs", "lang"=>"en", "lat"=>28.797169, "lon"=>-82.548293),
"/US/US-FL/horseshoe-beach-3139" => array("type"=>"town", "pop"=>294, "name_ru"=>"Хорсгу Бич", "name_en"=>"Horseshoe Beach", "lang"=>"en", "lat"=>29.440179, "lon"=>-83.288085),
"/US/US-FL/howey-in-the-hills-3140" => array("type"=>"town", "pop"=>1248, "name_ru"=>"Хауи-ин-зе-Хиллс", "name_en"=>"Howey-in-the-Hills", "lang"=>"en", "lat"=>28.715950, "lon"=>-81.774766),
"/US/US-FL/hudson-3141" => array("type"=>"cdp", "pop"=>7344, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>28.357673, "lon"=>-82.699510),
"/US/US-FL/hutchinson-island-south-3142" => array("type"=>"cdp", "pop"=>3893, "name_ru"=>"Хатчинсон Айленд Саут", "name_en"=>"Hutchinson Island South", "lang"=>"en", "lat"=>27.318729, "lon"=>-80.194808),
"/US/US-FL/hypoluxo-3143" => array("type"=>"town", "pop"=>2535, "name_ru"=>"Хайполуксо", "name_en"=>"Hypoluxo", "lang"=>"en", "lat"=>26.562550, "lon"=>-80.052576),
"/US/US-FL/immokalee-3144" => array("type"=>"cdp", "pop"=>14120, "name_ru"=>"Иммокали", "name_en"=>"Immokalee", "lang"=>"en", "lat"=>26.417078, "lon"=>-81.421562),
"/US/US-FL/indialantic-3145" => array("type"=>"town", "pop"=>2931, "name_ru"=>"Индиалантик", "name_en"=>"Indialantic", "lang"=>"en", "lat"=>28.087568, "lon"=>-80.569083),
"/US/US-FL/indian-creek-3146" => array("type"=>"village", "pop"=>37, "name_ru"=>"Индиан Крик", "name_en"=>"Indian Creek", "lang"=>"en", "lat"=>25.878061, "lon"=>-80.136678),
"/US/US-FL/indian-harbour-beach-3147" => array("type"=>"city", "pop"=>8311, "name_ru"=>"Индиан Харбор Бич", "name_en"=>"Indian Harbour Beach", "lang"=>"en", "lat"=>28.152951, "lon"=>-80.597825),
"/US/US-FL/indian-river-estates-3148" => array("type"=>"cdp", "pop"=>4858, "name_ru"=>"Индиан Ривер Эстэйтс", "name_en"=>"Indian River Estates", "lang"=>"en", "lat"=>27.364050, "lon"=>-80.310099),
"/US/US-FL/indian-river-shores-3149" => array("type"=>"town", "pop"=>3393, "name_ru"=>"Индиан Ривер Шорс", "name_en"=>"Indian River Shores", "lang"=>"en", "lat"=>27.705724, "lon"=>-80.382485),
"/US/US-FL/indian-rocks-beach-3150" => array("type"=>"city", "pop"=>5143, "name_ru"=>"Индиан Рокс Бич", "name_en"=>"Indian Rocks Beach", "lang"=>"en", "lat"=>27.897733, "lon"=>-82.844308),
"/US/US-FL/indian-shores-3151" => array("type"=>"town", "pop"=>1775, "name_ru"=>"Индиан Шорс", "name_en"=>"Indian Shores", "lang"=>"en", "lat"=>27.853767, "lon"=>-82.844432),
"/US/US-FL/indiantown-3152" => array("type"=>"cdp", "pop"=>4794, "name_ru"=>"Индиатаун", "name_en"=>"Indiantown", "lang"=>"en", "lat"=>27.036131, "lon"=>-80.469039),
"/US/US-FL/inglis-3153" => array("type"=>"town", "pop"=>1636, "name_ru"=>"Инглис", "name_en"=>"Inglis", "lang"=>"en", "lat"=>29.032626, "lon"=>-82.659361),
"/US/US-FL/interlachen-3154" => array("type"=>"town", "pop"=>1505, "name_ru"=>"Интерлэхен", "name_en"=>"Interlachen", "lang"=>"en", "lat"=>29.616885, "lon"=>-81.897907),
"/US/US-FL/inverness-3155" => array("type"=>"city", "pop"=>7276, "name_ru"=>"Инвернесс", "name_en"=>"Inverness", "lang"=>"en", "lat"=>28.839116, "lon"=>-82.343492),
"/US/US-FL/inwood-3156" => array("type"=>"cdp", "pop"=>6824, "name_ru"=>"Инвуд", "name_en"=>"Inwood", "lang"=>"en", "lat"=>28.038232, "lon"=>-81.767530),
"/US/US-FL/iona-3157" => array("type"=>"cdp", "pop"=>9565, "name_ru"=>"Айона", "name_en"=>"Iona", "lang"=>"en", "lat"=>26.513988, "lon"=>-81.960749),
"/US/US-FL/islamorada-3158" => array("type"=>"cdp", "pop"=>1220, "name_ru"=>"Исламорада", "name_en"=>"Islamorada", "lang"=>"en", "lat"=>24.918258, "lon"=>-80.635711),
"/US/US-FL/islandia-3159" => array("type"=>"city", "pop"=>13, "name_ru"=>"Исландия", "name_en"=>"Islandia", "lang"=>"en", "lat"=>25.383528, "lon"=>-80.216759),
"/US/US-FL/ives-estates-3160" => array("type"=>"cdp", "pop"=>13531, "name_ru"=>"Айвс Эстэйтс", "name_en"=>"Ives Estates", "lang"=>"en", "lat"=>25.962200, "lon"=>-80.183242),
"/US/US-FL/jacksonville-(remainder)-3162" => array("type"=>"(remainder)", "pop"=>635230, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville (remainder)", "lang"=>"en", "lat"=>30.334550, "lon"=>-81.657692),
"/US/US-FL/jacksonville-beach-3161" => array("type"=>"city", "pop"=>21673, "name_ru"=>"Джексонвилл Бич", "name_en"=>"Jacksonville Beach", "lang"=>"en", "lat"=>30.277163, "lon"=>-81.380727),
"/US/US-FL/jacob-city-3163" => array("type"=>"city", "pop"=>290, "name_ru"=>"Джейкоб Сити", "name_en"=>"Jacob City", "lang"=>"en", "lat"=>30.894147, "lon"=>-85.409871),
"/US/US-FL/jan-phyl-village-3164" => array("type"=>"cdp", "pop"=>5308, "name_ru"=>"Жан Фил Виллидж", "name_en"=>"Jan Phyl Village", "lang"=>"en", "lat"=>28.011608, "lon"=>-81.791197),
"/US/US-FL/jasmine-estates-3165" => array("type"=>"cdp", "pop"=>17136, "name_ru"=>"Жасмин Эстэйтс", "name_en"=>"Jasmine Estates", "lang"=>"en", "lat"=>28.292642, "lon"=>-82.690351),
"/US/US-FL/jasper-3166" => array("type"=>"city", "pop"=>2044, "name_ru"=>"Джаспер", "name_en"=>"Jasper", "lang"=>"en", "lat"=>30.517138, "lon"=>-82.956199),
"/US/US-FL/jay-3167" => array("type"=>"town", "pop"=>687, "name_ru"=>"Джэй", "name_en"=>"Jay", "lang"=>"en", "lat"=>30.951000, "lon"=>-87.152073),
"/US/US-FL/jennings-3168" => array("type"=>"town", "pop"=>843, "name_ru"=>"Дженнингс", "name_en"=>"Jennings", "lang"=>"en", "lat"=>30.601118, "lon"=>-83.106964),
"/US/US-FL/jensen-beach-3169" => array("type"=>"cdp", "pop"=>9884, "name_ru"=>"Дженсен Бич", "name_en"=>"Jensen Beach", "lang"=>"en", "lat"=>27.236089, "lon"=>-80.230891),
"/US/US-FL/june-park-3170" => array("type"=>"cdp", "pop"=>4080, "name_ru"=>"Джун Парк", "name_en"=>"June Park", "lang"=>"en", "lat"=>28.072465, "lon"=>-80.686990),
"/US/US-FL/juno-beach-3171" => array("type"=>"town", "pop"=>3332, "name_ru"=>"Джуно Бич", "name_en"=>"Juno Beach", "lang"=>"en", "lat"=>26.878373, "lon"=>-80.056420),
"/US/US-FL/jupiter-3172" => array("type"=>"town", "pop"=>48879, "name_ru"=>"Юпитер", "name_en"=>"Jupiter", "lang"=>"en", "lat"=>26.924100, "lon"=>-80.104591),
"/US/US-FL/jupiter-inlet-colony-3173" => array("type"=>"town", "pop"=>385, "name_ru"=>"Юпитер Инлет Колони", "name_en"=>"Jupiter Inlet Colony", "lang"=>"en", "lat"=>26.947850, "lon"=>-80.074999),
"/US/US-FL/jupiter-island-3174" => array("type"=>"town", "pop"=>667, "name_ru"=>"Юпитер Айленд", "name_en"=>"Jupiter Island", "lang"=>"en", "lat"=>27.042715, "lon"=>-80.109669),
"/US/US-FL/kathleen-3175" => array("type"=>"cdp", "pop"=>2743, "name_ru"=>"Катлин", "name_en"=>"Kathleen", "lang"=>"en", "lat"=>28.117900, "lon"=>-82.028110),
"/US/US-FL/kendale-lakes-3176" => array("type"=>"cdp", "pop"=>48524, "name_ru"=>"Кендэйл Лэйкс", "name_en"=>"Kendale Lakes", "lang"=>"en", "lat"=>25.707818, "lon"=>-80.407337),
"/US/US-FL/kendall-3177" => array("type"=>"cdp", "pop"=>87271, "name_ru"=>"Кендалл", "name_en"=>"Kendall", "lang"=>"en", "lat"=>25.667200, "lon"=>-80.339714),
"/US/US-FL/kendall-green-3178" => array("type"=>"cdp", "pop"=>3815, "name_ru"=>"Кендалл Грин", "name_en"=>"Kendall Green", "lang"=>"en", "lat"=>26.266850, "lon"=>-80.122724),
"/US/US-FL/kendall-lakes-west-3179" => array("type"=>"cdp", "pop"=>6038, "name_ru"=>"Кендалл Лэйкс Вест", "name_en"=>"Kendall Lakes West", "lang"=>"en", "lat"=>25.690950, "lon"=>-80.438574),
"/US/US-FL/kenneth-city-3180" => array("type"=>"town", "pop"=>4300, "name_ru"=>"Кеннет Сити", "name_en"=>"Kenneth City", "lang"=>"en", "lat"=>27.815733, "lon"=>-82.715518),
"/US/US-FL/kensington-park-3181" => array("type"=>"cdp", "pop"=>3026, "name_ru"=>"Кенсингтон Парк", "name_en"=>"Kensington Park", "lang"=>"en", "lat"=>27.359300, "lon"=>-82.496700),
"/US/US-FL/key-biscayne-3182" => array("type"=>"cdp", "pop"=>9735, "name_ru"=>"Ки Бискейн", "name_en"=>"Key Biscayne", "lang"=>"en", "lat"=>25.689950, "lon"=>-80.165349),
"/US/US-FL/key-colony-beach-3183" => array("type"=>"city", "pop"=>760, "name_ru"=>"Ки Колони Бич", "name_en"=>"Key Colony Beach", "lang"=>"en", "lat"=>24.722559, "lon"=>-81.023314),
"/US/US-FL/key-largo-3184" => array("type"=>"cdp", "pop"=>11336, "name_ru"=>"Ки Ларго", "name_en"=>"Key Largo", "lang"=>"en", "lat"=>25.106055, "lon"=>-80.441299),
"/US/US-FL/key-west-3186" => array("type"=>"city", "pop"=>22364, "name_ru"=>"Ки Вест", "name_en"=>"Key West", "lang"=>"en", "lat"=>24.562841, "lon"=>-81.775210),
"/US/US-FL/keystone-heights-3185" => array("type"=>"city", "pop"=>1494, "name_ru"=>"Кистон Хайтс", "name_en"=>"Keystone Heights", "lang"=>"en", "lat"=>29.780012, "lon"=>-82.033672),
"/US/US-FL/kings-point-3187" => array("type"=>"cdp", "pop"=>12422, "name_ru"=>"Кингс Поинт", "name_en"=>"Kings Point", "lang"=>"en", "lat"=>26.445012, "lon"=>-80.139978),
"/US/US-FL/kissimmee-3188" => array("type"=>"city", "pop"=>62291, "name_ru"=>"Киссимми", "name_en"=>"Kissimmee", "lang"=>"en", "lat"=>28.303533, "lon"=>-81.410505),
"/US/US-FL/la-belle-3189" => array("type"=>"city", "pop"=>2703, "name_ru"=>"Ла Бэлль", "name_en"=>"La Belle", "lang"=>"en", "lat"=>26.757169, "lon"=>-81.438040),
"/US/US-FL/la-crosse-3191" => array("type"=>"town", "pop"=>279, "name_ru"=>"Ла Кросс", "name_en"=>"La Crosse", "lang"=>"en", "lat"=>29.843922, "lon"=>-82.404452),
"/US/US-FL/lacoochee-3190" => array("type"=>"cdp", "pop"=>2072, "name_ru"=>"Лакучи", "name_en"=>"Lacoochee", "lang"=>"en", "lat"=>28.465046, "lon"=>-82.169237),
"/US/US-FL/lady-lake-3192" => array("type"=>"town", "pop"=>14940, "name_ru"=>"Леди Лэйк", "name_en"=>"Lady Lake", "lang"=>"en", "lat"=>28.927763, "lon"=>-81.928918),
"/US/US-FL/laguna-beach-3193" => array("type"=>"cdp", "pop"=>1876, "name_ru"=>"Лагуна Бич", "name_en"=>"Laguna Beach", "lang"=>"en", "lat"=>30.255082, "lon"=>-85.952773),
"/US/US-FL/lake-alfred-3194" => array("type"=>"city", "pop"=>4508, "name_ru"=>"Лэйк Алфред", "name_en"=>"Lake Alfred", "lang"=>"en", "lat"=>28.090656, "lon"=>-81.725771),
"/US/US-FL/lake-buena-vista-3195" => array("type"=>"city", "pop"=>15, "name_ru"=>"Лэйк Буена Виста", "name_en"=>"Lake Buena Vista", "lang"=>"en", "lat"=>28.387167, "lon"=>-81.525724),
"/US/US-FL/lake-butler-3196" => array("type"=>"city", "pop"=>1942, "name_ru"=>"Лэйк Батлер", "name_en"=>"Lake Butler", "lang"=>"en", "lat"=>30.020452, "lon"=>-82.339645),
"/US/US-FL/lake-city-3197" => array("type"=>"city", "pop"=>12427, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>30.192802, "lon"=>-82.642346),
"/US/US-FL/lake-clarke-shores-3198" => array("type"=>"town", "pop"=>3326, "name_ru"=>"Лэйк Кларк Шорс", "name_en"=>"Lake Clarke Shores", "lang"=>"en", "lat"=>26.646233, "lon"=>-80.075290),
"/US/US-FL/lake-hamilton-3199" => array("type"=>"town", "pop"=>1435, "name_ru"=>"Лэйк Хамилтон", "name_en"=>"Lake Hamilton", "lang"=>"en", "lat"=>28.047824, "lon"=>-81.625875),
"/US/US-FL/lake-helen-3200" => array("type"=>"city", "pop"=>2784, "name_ru"=>"Лэйк Хелен", "name_en"=>"Lake Helen", "lang"=>"en", "lat"=>28.983727, "lon"=>-81.231430),
"/US/US-FL/lake-lorraine-3203" => array("type"=>"cdp", "pop"=>6779, "name_ru"=>"Лэйк Лоппэйн", "name_en"=>"Lake Lorraine", "lang"=>"en", "lat"=>30.441350, "lon"=>-86.565274),
"/US/US-FL/lake-lucerne-3204" => array("type"=>"cdp", "pop"=>9478, "name_ru"=>"Лэйк Люсерн", "name_en"=>"Lake Lucerne", "lang"=>"en", "lat"=>25.966750, "lon"=>-80.259624),
"/US/US-FL/lake-magdalene-3205" => array("type"=>"cdp", "pop"=>15973, "name_ru"=>"Лэйк Магдален", "name_en"=>"Lake Magdalene", "lang"=>"en", "lat"=>28.073800, "lon"=>-82.472054),
"/US/US-FL/lake-mary-3206" => array("type"=>"city", "pop"=>15143, "name_ru"=>"Лэйк Мэри", "name_en"=>"Lake Mary", "lang"=>"en", "lat"=>28.758957, "lon"=>-81.334085),
"/US/US-FL/lake-panasoffkee-3207" => array("type"=>"cdp", "pop"=>2705, "name_ru"=>"Лэйк Панасоффки", "name_en"=>"Lake Panasoffkee", "lang"=>"en", "lat"=>28.790161, "lon"=>-82.130091),
"/US/US-FL/lake-park-3208" => array("type"=>"town", "pop"=>8668, "name_ru"=>"Лэйк Парк", "name_en"=>"Lake Park", "lang"=>"en", "lat"=>26.799800, "lon"=>-80.065977),
"/US/US-FL/lake-placid-3209" => array("type"=>"town", "pop"=>1878, "name_ru"=>"Лэйк Плэйсид", "name_en"=>"Lake Placid", "lang"=>"en", "lat"=>27.296885, "lon"=>-81.369662),
"/US/US-FL/lake-sarasota-3210" => array("type"=>"cdp", "pop"=>4117, "name_ru"=>"Лэйк Сарасота", "name_en"=>"Lake Sarasota", "lang"=>"en", "lat"=>27.292205, "lon"=>-82.437776),
"/US/US-FL/lake-wales-3214" => array("type"=>"city", "pop"=>14474, "name_ru"=>"Лэйк Уэйлс", "name_en"=>"Lake Wales", "lang"=>"en", "lat"=>27.901128, "lon"=>-81.587824),
"/US/US-FL/lake-worth-3216" => array("type"=>"city", "pop"=>35513, "name_ru"=>"Лэйк Ворт", "name_en"=>"Lake Worth", "lang"=>"en", "lat"=>26.619350, "lon"=>-80.058979),
"/US/US-FL/lakeland-3201" => array("type"=>"city", "pop"=>93333, "name_ru"=>"Лэйклэнд", "name_en"=>"Lakeland", "lang"=>"en", "lat"=>28.041167, "lon"=>-81.960006),
"/US/US-FL/lakeland-highlands-3202" => array("type"=>"cdp", "pop"=>9972, "name_ru"=>"Лэйклэнд Хайлендс", "name_en"=>"Lakeland Highlands", "lang"=>"en", "lat"=>27.959450, "lon"=>-81.950166),
"/US/US-FL/lakes-by-the-bay-3211" => array("type"=>"cdp", "pop"=>5615, "name_ru"=>"Лэйкс бай зе Бэй", "name_en"=>"Lakes by the Bay", "lang"=>"en", "lat"=>25.571850, "lon"=>-80.325527),
"/US/US-FL/lakeside-3212" => array("type"=>"cdp", "pop"=>29137, "name_ru"=>"Лэйксайд", "name_en"=>"Lakeside", "lang"=>"en", "lat"=>30.129650, "lon"=>-81.768400),
"/US/US-FL/lakeside-green-3213" => array("type"=>"cdp", "pop"=>2994, "name_ru"=>"Лэйксайд Грин", "name_en"=>"Lakeside Green", "lang"=>"en", "lat"=>26.738750, "lon"=>-80.114680),
"/US/US-FL/lakewood-park-3215" => array("type"=>"cdp", "pop"=>7211, "name_ru"=>"Лэйквуд Парк", "name_en"=>"Lakewood Park", "lang"=>"en", "lat"=>27.538006, "lon"=>-80.390289),
"/US/US-FL/land-o'-lakes-3217" => array("type"=>"cdp", "pop"=>7892, "name_ru"=>"Лэнд О-Лэйкс", "name_en"=>"Land O' Lakes", "lang"=>"en", "lat"=>28.221746, "lon"=>-82.453814),
"/US/US-FL/lantana-3218" => array("type"=>"town", "pop"=>10071, "name_ru"=>"Лантана", "name_en"=>"Lantana", "lang"=>"en", "lat"=>26.583250, "lon"=>-80.055724),
"/US/US-FL/largo-3219" => array("type"=>"city", "pop"=>72732, "name_ru"=>"Ларго", "name_en"=>"Largo", "lang"=>"en", "lat"=>27.908589, "lon"=>-82.781000),
"/US/US-FL/lauderdale-by-the-sea-3220" => array("type"=>"town", "pop"=>5831, "name_ru"=>"Лоудердэйл ба зе Си", "name_en"=>"Lauderdale-by-the-Sea", "lang"=>"en", "lat"=>26.188433, "lon"=>-80.095793),
"/US/US-FL/lauderdale-lakes-3221" => array("type"=>"city", "pop"=>31004, "name_ru"=>"Лоудердэйл Лэйкс", "name_en"=>"Lauderdale Lakes", "lang"=>"en", "lat"=>26.166650, "lon"=>-80.202090),
"/US/US-FL/lauderhill-3222" => array("type"=>"city", "pop"=>67073, "name_ru"=>"Лоудерхилл", "name_en"=>"Lauderhill", "lang"=>"en", "lat"=>26.165800, "lon"=>-80.232623),
"/US/US-FL/laurel-3223" => array("type"=>"cdp", "pop"=>8245, "name_ru"=>"Лорел", "name_en"=>"Laurel", "lang"=>"en", "lat"=>27.142550, "lon"=>-82.462413),
"/US/US-FL/laurel-hill-3224" => array("type"=>"city", "pop"=>558, "name_ru"=>"Лаурел Хилл", "name_en"=>"Laurel Hill", "lang"=>"en", "lat"=>30.965132, "lon"=>-86.456791),
"/US/US-FL/lawtey-3225" => array("type"=>"city", "pop"=>700, "name_ru"=>"Лоути", "name_en"=>"Lawtey", "lang"=>"en", "lat"=>30.045602, "lon"=>-82.070774),
"/US/US-FL/layton-3226" => array("type"=>"city", "pop"=>191, "name_ru"=>"Лэйтон", "name_en"=>"Layton", "lang"=>"en", "lat"=>24.822248, "lon"=>-80.811705),
"/US/US-FL/lazy-lake-3227" => array("type"=>"village", "pop"=>38, "name_ru"=>"Лэйзи Лэйк", "name_en"=>"Lazy Lake", "lang"=>"en", "lat"=>26.155750, "lon"=>-80.145349),
"/US/US-FL/lealman-3228" => array("type"=>"cdp", "pop"=>21748, "name_ru"=>"Лилман", "name_en"=>"Lealman", "lang"=>"en", "lat"=>27.822500, "lon"=>-82.688899),
"/US/US-FL/lecanto-3229" => array("type"=>"cdp", "pop"=>1243, "name_ru"=>"Леканто", "name_en"=>"Lecanto", "lang"=>"en", "lat"=>28.858050, "lon"=>-82.485780),
"/US/US-FL/lee-3230" => array("type"=>"town", "pop"=>358, "name_ru"=>"Ли", "name_en"=>"Lee", "lang"=>"en", "lat"=>30.418318, "lon"=>-83.301843),
"/US/US-FL/leesburg-3231" => array("type"=>"city", "pop"=>22113, "name_ru"=>"Лисберг", "name_en"=>"Leesburg", "lang"=>"en", "lat"=>28.808524, "lon"=>-81.882561),
"/US/US-FL/lehigh-acres-3232" => array("type"=>"cdp", "pop"=>13611, "name_ru"=>"Лихай Экрз", "name_en"=>"Lehigh Acres", "lang"=>"en", "lat"=>26.602937, "lon"=>-81.633092),
"/US/US-FL/leisure-city-3233" => array("type"=>"cdp", "pop"=>19379, "name_ru"=>"Лейже Сити", "name_en"=>"Leisure City", "lang"=>"en", "lat"=>25.493650, "lon"=>-80.436451),
"/US/US-FL/lely-3234" => array("type"=>"cdp", "pop"=>3014, "name_ru"=>"Лили", "name_en"=>"Lely", "lang"=>"en", "lat"=>26.100144, "lon"=>-81.728709),
"/US/US-FL/lighthouse-point-3235" => array("type"=>"city", "pop"=>11118, "name_ru"=>"Лайтхауз Поинт", "name_en"=>"Lighthouse Point", "lang"=>"en", "lat"=>26.277511, "lon"=>-80.089851),
"/US/US-FL/lindgren-acres-3236" => array("type"=>"cdp", "pop"=>22290, "name_ru"=>"Линдгрен Эйкрес", "name_en"=>"Lindgren Acres", "lang"=>"en", "lat"=>25.670300, "lon"=>-80.401437),
"/US/US-FL/live-oak-3237" => array("type"=>"city", "pop"=>7215, "name_ru"=>"Лайв Оук", "name_en"=>"Live Oak", "lang"=>"en", "lat"=>30.294347, "lon"=>-82.986078),
"/US/US-FL/lochmoor-waterway-estates-3238" => array("type"=>"cdp", "pop"=>4091, "name_ru"=>"Локмур Вотервэй Эстэйтс", "name_en"=>"Lochmoor Waterway Estates", "lang"=>"en", "lat"=>26.643750, "lon"=>-81.910138),
"/US/US-FL/lockhart-3239" => array("type"=>"cdp", "pop"=>11636, "name_ru"=>"Локхарт", "name_en"=>"Lockhart", "lang"=>"en", "lat"=>28.626100, "lon"=>-81.436625),
"/US/US-FL/longboat-key-3240" => array("type"=>"town", "pop"=>7309, "name_ru"=>"Лонгбоат Ки", "name_en"=>"Longboat Key", "lang"=>"en", "lat"=>27.387365, "lon"=>-82.636591),
"/US/US-FL/longwood-3241" => array("type"=>"city", "pop"=>13493, "name_ru"=>"Лонгвуд", "name_en"=>"Longwood", "lang"=>"en", "lat"=>28.701000, "lon"=>-81.348424),
"/US/US-FL/loughman-3242" => array("type"=>"cdp", "pop"=>1214, "name_ru"=>"Лауфмэн", "name_en"=>"Loughman", "lang"=>"en", "lat"=>28.236728, "lon"=>-81.571680),
"/US/US-FL/lower-grand-lagoon-3243" => array("type"=>"cdp", "pop"=>3329, "name_ru"=>"Лоуер Гранд Лагун", "name_en"=>"Lower Grand Lagoon", "lang"=>"en", "lat"=>30.143852, "lon"=>-85.750789),
"/US/US-FL/lutz-3244" => array("type"=>"cdp", "pop"=>10552, "name_ru"=>"Луц", "name_en"=>"Lutz", "lang"=>"en", "lat"=>28.138728, "lon"=>-82.463055),
"/US/US-FL/lynn-haven-3245" => array("type"=>"city", "pop"=>15254, "name_ru"=>"Линн Хэвен", "name_en"=>"Lynn Haven", "lang"=>"en", "lat"=>30.242850, "lon"=>-85.651649),
"/US/US-FL/macclenny-3246" => array("type"=>"city", "pop"=>6032, "name_ru"=>"Маккленни", "name_en"=>"Macclenny", "lang"=>"en", "lat"=>30.280214, "lon"=>-82.125912),
"/US/US-FL/madeira-beach-3249" => array("type"=>"city", "pop"=>4339, "name_ru"=>"Мадейра Бич", "name_en"=>"Madeira Beach", "lang"=>"en", "lat"=>27.795400, "lon"=>-82.791715),
"/US/US-FL/madison-3250" => array("type"=>"city", "pop"=>3039, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>30.468644, "lon"=>-83.415441),
"/US/US-FL/maitland-3251" => array("type"=>"city", "pop"=>14931, "name_ru"=>"Мэйтлэнд", "name_en"=>"Maitland", "lang"=>"en", "lat"=>28.627950, "lon"=>-81.367124),
"/US/US-FL/malabar-3252" => array("type"=>"town", "pop"=>2745, "name_ru"=>"Малабар", "name_en"=>"Malabar", "lang"=>"en", "lat"=>27.986723, "lon"=>-80.582315),
"/US/US-FL/malone-3253" => array("type"=>"town", "pop"=>2087, "name_ru"=>"Малон", "name_en"=>"Malone", "lang"=>"en", "lat"=>30.958396, "lon"=>-85.161197),
"/US/US-FL/manalapan-3254" => array("type"=>"town", "pop"=>339, "name_ru"=>"Маналапан", "name_en"=>"Manalapan", "lang"=>"en", "lat"=>26.558200, "lon"=>-80.042292),
"/US/US-FL/manasota-key-3255" => array("type"=>"cdp", "pop"=>1395, "name_ru"=>"Манасота Ки", "name_en"=>"Manasota Key", "lang"=>"en", "lat"=>26.924862, "lon"=>-82.352198),
"/US/US-FL/mango-3256" => array("type"=>"cdp", "pop"=>8700, "name_ru"=>"Манго", "name_en"=>"Mango", "lang"=>"en", "lat"=>27.990856, "lon"=>-82.307545),
"/US/US-FL/mangonia-park-3257" => array("type"=>"town", "pop"=>1230, "name_ru"=>"Мангониа Парк", "name_en"=>"Mangonia Park", "lang"=>"en", "lat"=>26.757767, "lon"=>-80.074104),
"/US/US-FL/marathon-3258" => array("type"=>"cdp", "pop"=>9526, "name_ru"=>"Марафон", "name_en"=>"Marathon", "lang"=>"en", "lat"=>24.711901, "lon"=>-81.080792),
"/US/US-FL/marco-3259" => array("type"=>"cdp", "pop"=>9493, "name_ru"=>"Марко", "name_en"=>"Marco", "lang"=>"en", "lat"=>25.942407, "lon"=>-81.717617),
"/US/US-FL/margate-3260" => array("type"=>"city", "pop"=>54086, "name_ru"=>"Маргэйт", "name_en"=>"Margate", "lang"=>"en", "lat"=>26.245400, "lon"=>-80.211224),
"/US/US-FL/marianna-3261" => array("type"=>"city", "pop"=>6284, "name_ru"=>"Марианна", "name_en"=>"Marianna", "lang"=>"en", "lat"=>30.774964, "lon"=>-85.239731),
"/US/US-FL/marineland-3262" => array("type"=>"town", "pop"=>14, "name_ru"=>"Маринлэнд", "name_en"=>"Marineland", "lang"=>"en", "lat"=>29.663992, "lon"=>-81.213317),
"/US/US-FL/mary-esther-3263" => array("type"=>"city", "pop"=>3918, "name_ru"=>"Мэри Эстер", "name_en"=>"Mary Esther", "lang"=>"en", "lat"=>30.411272, "lon"=>-86.657113),
"/US/US-FL/mascotte-3264" => array("type"=>"city", "pop"=>5591, "name_ru"=>"Маскотт", "name_en"=>"Mascotte", "lang"=>"en", "lat"=>28.576182, "lon"=>-81.888581),
"/US/US-FL/mayo-3265" => array("type"=>"town", "pop"=>1028, "name_ru"=>"Майо", "name_en"=>"Mayo", "lang"=>"en", "lat"=>30.051025, "lon"=>-83.176656),
"/US/US-FL/mcgregor-3247" => array("type"=>"cdp", "pop"=>6504, "name_ru"=>"Мак-Грегор", "name_en"=>"McGregor", "lang"=>"en", "lat"=>26.560433, "lon"=>-81.914701),
"/US/US-FL/mcintosh-3248" => array("type"=>"town", "pop"=>521, "name_ru"=>"Макинтош", "name_en"=>"McIntosh", "lang"=>"en", "lat"=>29.449074, "lon"=>-82.220848),
"/US/US-FL/meadow-wood-3266" => array("type"=>"cdp", "pop"=>4876, "name_ru"=>"Мидоу Вуд", "name_en"=>"Meadow Wood", "lang"=>"en", "lat"=>28.385300, "lon"=>-81.366609),
"/US/US-FL/medley-3267" => array("type"=>"town", "pop"=>1025, "name_ru"=>"Медли", "name_en"=>"Medley", "lang"=>"en", "lat"=>25.863045, "lon"=>-80.348135),
"/US/US-FL/medulla-3268" => array("type"=>"cdp", "pop"=>3977, "name_ru"=>"Медулла", "name_en"=>"Medulla", "lang"=>"en", "lat"=>27.959650, "lon"=>-81.988703),
"/US/US-FL/melbourne-3269" => array("type"=>"city", "pop"=>77351, "name_ru"=>"Мельбурн", "name_en"=>"Melbourne", "lang"=>"en", "lat"=>28.114987, "lon"=>-80.640115),
"/US/US-FL/melbourne-beach-3270" => array("type"=>"town", "pop"=>3150, "name_ru"=>"Мельбурн Бич", "name_en"=>"Melbourne Beach", "lang"=>"en", "lat"=>28.067401, "lon"=>-80.559550),
"/US/US-FL/melbourne-village-3271" => array("type"=>"town", "pop"=>671, "name_ru"=>"Мельбурн Виллидж", "name_en"=>"Melbourne Village", "lang"=>"en", "lat"=>28.088068, "lon"=>-80.667151),
"/US/US-FL/melrose-park-3272" => array("type"=>"cdp", "pop"=>6477, "name_ru"=>"Мелроуз Парк", "name_en"=>"Melrose Park", "lang"=>"en", "lat"=>26.113400, "lon"=>-80.193817),
"/US/US-FL/memphis-3273" => array("type"=>"cdp", "pop"=>6760, "name_ru"=>"Мемфис", "name_en"=>"Memphis", "lang"=>"en", "lat"=>27.539650, "lon"=>-82.558199),
"/US/US-FL/merritt-island-3274" => array("type"=>"cdp", "pop"=>32886, "name_ru"=>"Мерритт Айленд", "name_en"=>"Merritt Island", "lang"=>"en", "lat"=>28.310645, "lon"=>-80.664576),
"/US/US-FL/mexico-beach-3275" => array("type"=>"city", "pop"=>1264, "name_ru"=>"Мексико Бич", "name_en"=>"Mexico Beach", "lang"=>"en", "lat"=>29.942128, "lon"=>-85.405298),
"/US/US-FL/miami-3276" => array("type"=>"city", "pop"=>413201, "name_ru"=>"Майами", "name_en"=>"Miami", "lang"=>"en", "lat"=>25.775667, "lon"=>-80.210845),
"/US/US-FL/miami-beach-3277" => array("type"=>"city", "pop"=>84633, "name_ru"=>"Майами Бич", "name_en"=>"Miami Beach", "lang"=>"en", "lat"=>25.814700, "lon"=>-80.135932),
"/US/US-FL/miami-gardens-utopia-carver-3278" => array("type"=>"cdp", "pop"=>7448, "name_ru"=>"Майами Гарденс-Утопиа-Карвер", "name_en"=>"Miami Gardens-Utopia-Carver", "lang"=>"en", "lat"=>25.985850, "lon"=>-80.196924),
"/US/US-FL/miami-lakes-3279" => array("type"=>"cdp", "pop"=>21727, "name_ru"=>"Майами Лэйкс", "name_en"=>"Miami Lakes", "lang"=>"en", "lat"=>25.909950, "lon"=>-80.314469),
"/US/US-FL/miami-shores-3280" => array("type"=>"village", "pop"=>9551, "name_ru"=>"Майами Шорс", "name_en"=>"Miami Shores", "lang"=>"en", "lat"=>25.866400, "lon"=>-80.178627),
"/US/US-FL/miami-springs-3281" => array("type"=>"city", "pop"=>12534, "name_ru"=>"Майами Спрингс", "name_en"=>"Miami Springs", "lang"=>"en", "lat"=>25.819350, "lon"=>-80.289501),
"/US/US-FL/micanopy-3282" => array("type"=>"town", "pop"=>695, "name_ru"=>"Майкэнопи", "name_en"=>"Micanopy", "lang"=>"en", "lat"=>29.506107, "lon"=>-82.280777),
"/US/US-FL/micco-3283" => array("type"=>"cdp", "pop"=>8757, "name_ru"=>"Микко", "name_en"=>"Micco", "lang"=>"en", "lat"=>27.871070, "lon"=>-80.515542),
"/US/US-FL/mid-florida-lakes-3285" => array("type"=>"cdp", "pop"=>2776, "name_ru"=>"Мид Флорида Лэйкс", "name_en"=>"Mid Florida Lakes", "lang"=>"en", "lat"=>28.862770, "lon"=>-81.757009),
"/US/US-FL/middleburg-3284" => array("type"=>"cdp", "pop"=>6223, "name_ru"=>"Миддлберг", "name_en"=>"Middleburg", "lang"=>"en", "lat"=>30.050543, "lon"=>-81.902190),
"/US/US-FL/midway-3286" => array("type"=>"city", "pop"=>2752, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>30.491150, "lon"=>-84.462084),
"/US/US-FL/milton-3287" => array("type"=>"city", "pop"=>8656, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>30.631747, "lon"=>-87.049102),
"/US/US-FL/mims-3288" => array("type"=>"cdp", "pop"=>9412, "name_ru"=>"Мимс", "name_en"=>"Mims", "lang"=>"en", "lat"=>28.684572, "lon"=>-80.846254),
"/US/US-FL/minneola-3289" => array("type"=>"city", "pop"=>9006, "name_ru"=>"Миннеола", "name_en"=>"Minneola", "lang"=>"en", "lat"=>28.575050, "lon"=>-81.747074),
"/US/US-FL/miramar-3290" => array("type"=>"city", "pop"=>108484, "name_ru"=>"Мирамар", "name_en"=>"Miramar", "lang"=>"en", "lat"=>25.976929, "lon"=>-80.333446),
"/US/US-FL/miramar-beach-3291" => array("type"=>"cdp", "pop"=>1644, "name_ru"=>"Мирамар Бич", "name_en"=>"Miramar Beach", "lang"=>"en", "lat"=>30.379874, "lon"=>-86.360244),
"/US/US-FL/mission-bay-3292" => array("type"=>"cdp", "pop"=>1227, "name_ru"=>"Мишн Бэй", "name_en"=>"Mission Bay", "lang"=>"en", "lat"=>26.369606, "lon"=>-80.210105),
"/US/US-FL/molino-3293" => array("type"=>"cdp", "pop"=>1207, "name_ru"=>"Молино", "name_en"=>"Molino", "lang"=>"en", "lat"=>30.719378, "lon"=>-87.325689),
"/US/US-FL/monticello-3294" => array("type"=>"city", "pop"=>2541, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>30.541947, "lon"=>-83.870769),
"/US/US-FL/montverde-3295" => array("type"=>"town", "pop"=>1049, "name_ru"=>"Монтверде", "name_en"=>"Montverde", "lang"=>"en", "lat"=>28.596850, "lon"=>-81.675999),
"/US/US-FL/moore-haven-3296" => array("type"=>"city", "pop"=>1683, "name_ru"=>"Мур Хэвен", "name_en"=>"Moore Haven", "lang"=>"en", "lat"=>26.833781, "lon"=>-81.097686),
"/US/US-FL/morse-shores-3297" => array("type"=>"cdp", "pop"=>3771, "name_ru"=>"Морзе Шорс", "name_en"=>"Morse Shores", "lang"=>"en", "lat"=>26.680654, "lon"=>-81.797740),
"/US/US-FL/mount-dora-3298" => array("type"=>"city", "pop"=>12869, "name_ru"=>"Маунт Дора", "name_en"=>"Mount Dora", "lang"=>"en", "lat"=>28.806700, "lon"=>-81.641699),
"/US/US-FL/mount-plymouth-3299" => array("type"=>"cdp", "pop"=>1752, "name_ru"=>"Маунт Плимут", "name_en"=>"Mount Plymouth", "lang"=>"en", "lat"=>28.799684, "lon"=>-81.532259),
"/US/US-FL/mulberry-3300" => array("type"=>"city", "pop"=>3187, "name_ru"=>"Малберри", "name_en"=>"Mulberry", "lang"=>"en", "lat"=>27.901616, "lon"=>-81.973051),
"/US/US-FL/myrtle-grove-3301" => array("type"=>"cdp", "pop"=>17402, "name_ru"=>"Миртл Гроув", "name_en"=>"Myrtle Grove", "lang"=>"en", "lat"=>30.415936, "lon"=>-87.303775),
"/US/US-FL/naples-3302" => array("type"=>"city", "pop"=>21532, "name_ru"=>"Нэплс", "name_en"=>"Naples", "lang"=>"en", "lat"=>26.144700, "lon"=>-81.794685),
"/US/US-FL/naples-manor-3303" => array("type"=>"cdp", "pop"=>4574, "name_ru"=>"Нэплс Мэйнор", "name_en"=>"Naples Manor", "lang"=>"en", "lat"=>26.072214, "lon"=>-81.721479),
"/US/US-FL/naples-park-3304" => array("type"=>"cdp", "pop"=>8002, "name_ru"=>"Нэплс Парк", "name_en"=>"Naples Park", "lang"=>"en", "lat"=>26.263350, "lon"=>-81.815574),
"/US/US-FL/naranja-3305" => array("type"=>"cdp", "pop"=>5790, "name_ru"=>"Наранья", "name_en"=>"Naranja", "lang"=>"en", "lat"=>25.515650, "lon"=>-80.422199),
"/US/US-FL/nassau-village-ratliff-3306" => array("type"=>"cdp", "pop"=>4047, "name_ru"=>"Нассау Виллидж Рэтлиф", "name_en"=>"Nassau Village-Ratliff", "lang"=>"en", "lat"=>30.502092, "lon"=>-81.814696),
"/US/US-FL/neptune-beach-3307" => array("type"=>"city", "pop"=>6781, "name_ru"=>"Нептун Бич", "name_en"=>"Neptune Beach", "lang"=>"en", "lat"=>30.314741, "lon"=>-81.393738),
"/US/US-FL/new-port-richey-3309" => array("type"=>"city", "pop"=>17578, "name_ru"=>"Нью Порт Ричи", "name_en"=>"New Port Richey", "lang"=>"en", "lat"=>28.247714, "lon"=>-82.717824),
"/US/US-FL/new-port-richey-east-3310" => array("type"=>"cdp", "pop"=>9683, "name_ru"=>"Нью Порт Ричи Ист", "name_en"=>"New Port Richey East", "lang"=>"en", "lat"=>28.260829, "lon"=>-82.691873),
"/US/US-FL/new-smyrna-beach-3311" => array("type"=>"city", "pop"=>23325, "name_ru"=>"Нью Смирна Бич", "name_en"=>"New Smyrna Beach", "lang"=>"en", "lat"=>29.033650, "lon"=>-80.918708),
"/US/US-FL/newberry-3308" => array("type"=>"city", "pop"=>4247, "name_ru"=>"Ньюберри", "name_en"=>"Newberry", "lang"=>"en", "lat"=>29.644809, "lon"=>-82.612295),
"/US/US-FL/niceville-3312" => array("type"=>"city", "pop"=>12313, "name_ru"=>"Найсвилл", "name_en"=>"Niceville", "lang"=>"en", "lat"=>30.530000, "lon"=>-86.484170),
"/US/US-FL/nokomis-3313" => array("type"=>"cdp", "pop"=>3448, "name_ru"=>"Нокомис", "name_en"=>"Nokomis", "lang"=>"en", "lat"=>27.121900, "lon"=>-82.436699),
"/US/US-FL/noma-3314" => array("type"=>"town", "pop"=>215, "name_ru"=>"Нома", "name_en"=>"Noma", "lang"=>"en", "lat"=>30.981272, "lon"=>-85.617448),
"/US/US-FL/norland-3315" => array("type"=>"cdp", "pop"=>22109, "name_ru"=>"Норлэнд", "name_en"=>"Norland", "lang"=>"en", "lat"=>25.946800, "lon"=>-80.208699),
"/US/US-FL/north-andrews-gardens-3316" => array("type"=>"cdp", "pop"=>9002, "name_ru"=>"Норт Эндрюс Гарденс", "name_en"=>"North Andrews Gardens", "lang"=>"en", "lat"=>26.191100, "lon"=>-80.144574),
"/US/US-FL/north-bay-village-3317" => array("type"=>"city", "pop"=>7878, "name_ru"=>"Норт Бэй Виллидж", "name_en"=>"North Bay Village", "lang"=>"en", "lat"=>25.846600, "lon"=>-80.154971),
"/US/US-FL/north-brooksville-3318" => array("type"=>"cdp", "pop"=>1459, "name_ru"=>"Норт Бруксвилл", "name_en"=>"North Brooksville", "lang"=>"en", "lat"=>28.572900, "lon"=>-82.408277),
"/US/US-FL/north-de-land-3319" => array("type"=>"cdp", "pop"=>1493, "name_ru"=>"Норт Де Лэнд", "name_en"=>"North De Land", "lang"=>"en", "lat"=>29.049286, "lon"=>-81.298198),
"/US/US-FL/north-fort-myers-3320" => array("type"=>"cdp", "pop"=>30027, "name_ru"=>"Норт Форт Майерс", "name_en"=>"North Fort Myers", "lang"=>"en", "lat"=>26.705567, "lon"=>-81.892833),
"/US/US-FL/north-key-largo-3321" => array("type"=>"cdp", "pop"=>1490, "name_ru"=>"Норт Ки Ларго", "name_en"=>"North Key Largo", "lang"=>"en", "lat"=>25.266849, "lon"=>-80.323564),
"/US/US-FL/north-lauderdale-3322" => array("type"=>"city", "pop"=>41772, "name_ru"=>"Норт Лоудердэйл", "name_en"=>"North Lauderdale", "lang"=>"en", "lat"=>26.213150, "lon"=>-80.223771),
"/US/US-FL/north-miami-3323" => array("type"=>"city", "pop"=>55908, "name_ru"=>"Норт Майами", "name_en"=>"North Miami", "lang"=>"en", "lat"=>25.900300, "lon"=>-80.169299),
"/US/US-FL/north-miami-beach-3324" => array("type"=>"city", "pop"=>37997, "name_ru"=>"Норт Майами Бич", "name_en"=>"North Miami Beach", "lang"=>"en", "lat"=>25.928683, "lon"=>-80.166874),
"/US/US-FL/north-naples-3325" => array("type"=>"cdp", "pop"=>13422, "name_ru"=>"Норт Нэплс", "name_en"=>"North Naples", "lang"=>"en", "lat"=>26.197890, "lon"=>-81.785966),
"/US/US-FL/north-palm-beach-3326" => array("type"=>"village", "pop"=>12149, "name_ru"=>"Норт Палм Бич", "name_en"=>"North Palm Beach", "lang"=>"en", "lat"=>26.819200, "lon"=>-80.056669),
"/US/US-FL/north-port-3327" => array("type"=>"city", "pop"=>54907, "name_ru"=>"Норт Порт", "name_en"=>"North Port", "lang"=>"en", "lat"=>27.049559, "lon"=>-82.194888),
"/US/US-FL/north-redington-beach-3328" => array("type"=>"town", "pop"=>1476, "name_ru"=>"Норт Редингтон Бич", "name_en"=>"North Redington Beach", "lang"=>"en", "lat"=>27.817300, "lon"=>-82.821853),
"/US/US-FL/north-river-shores-3329" => array("type"=>"cdp", "pop"=>3250, "name_ru"=>"Норт Ривер Шорс", "name_en"=>"North River Shores", "lang"=>"en", "lat"=>27.222743, "lon"=>-80.275672),
"/US/US-FL/north-sarasota-3330" => array("type"=>"cdp", "pop"=>6702, "name_ru"=>"Норт Сарасота", "name_en"=>"North Sarasota", "lang"=>"en", "lat"=>27.373700, "lon"=>-82.518549),
"/US/US-FL/oak-hill-3331" => array("type"=>"city", "pop"=>1596, "name_ru"=>"Оук Хилл", "name_en"=>"Oak Hill", "lang"=>"en", "lat"=>28.878488, "lon"=>-80.830331),
"/US/US-FL/oak-ridge-3335" => array("type"=>"cdp", "pop"=>15388, "name_ru"=>"Оук Ридж", "name_en"=>"Oak Ridge", "lang"=>"en", "lat"=>28.470913, "lon"=>-81.424621),
"/US/US-FL/oakland-3332" => array("type"=>"city", "pop"=>1142, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>28.556100, "lon"=>-81.628913),
"/US/US-FL/oakland-park-3333" => array("type"=>"city", "pop"=>42293, "name_ru"=>"Оуклэнд Парк", "name_en"=>"Oakland Park", "lang"=>"en", "lat"=>26.174550, "lon"=>-80.153389),
"/US/US-FL/oakland-park-3334" => array("type"=>"cdp", "pop"=>1743, "name_ru"=>"Оуклэнд Парк", "name_en"=>"Oakland Park", "lang"=>"en", "lat"=>28.800000, "lon"=>-81.681378),
"/US/US-FL/ocala-3336" => array("type"=>"city", "pop"=>54801, "name_ru"=>"Окала", "name_en"=>"Ocala", "lang"=>"en", "lat"=>29.187450, "lon"=>-82.130799),
"/US/US-FL/ocean-breeze-park-3337" => array("type"=>"town", "pop"=>431, "name_ru"=>"Оушн Бриз Парк", "name_en"=>"Ocean Breeze Park", "lang"=>"en", "lat"=>27.240867, "lon"=>-80.226006),
"/US/US-FL/ocean-city-3338" => array("type"=>"cdp", "pop"=>5422, "name_ru"=>"Оушен Сити", "name_en"=>"Ocean City", "lang"=>"en", "lat"=>30.437600, "lon"=>-86.608378),
"/US/US-FL/ocean-ridge-3339" => array("type"=>"town", "pop"=>1639, "name_ru"=>"Оушн Ридж", "name_en"=>"Ocean Ridge", "lang"=>"en", "lat"=>26.534450, "lon"=>-80.047949),
"/US/US-FL/ocoee-3340" => array("type"=>"city", "pop"=>32290, "name_ru"=>"Осои", "name_en"=>"Ocoee", "lang"=>"en", "lat"=>28.572191, "lon"=>-81.532594),
"/US/US-FL/ojus-3341" => array("type"=>"cdp", "pop"=>15519, "name_ru"=>"Ойус", "name_en"=>"Ojus", "lang"=>"en", "lat"=>25.956367, "lon"=>-80.160199),
"/US/US-FL/okeechobee-3342" => array("type"=>"city", "pop"=>5987, "name_ru"=>"Окичоуби", "name_en"=>"Okeechobee", "lang"=>"en", "lat"=>27.241267, "lon"=>-80.829881),
"/US/US-FL/oldsmar-3343" => array("type"=>"city", "pop"=>13401, "name_ru"=>"Олдсмар", "name_en"=>"Oldsmar", "lang"=>"en", "lat"=>28.048481, "lon"=>-82.670667),
"/US/US-FL/olympia-heights-3344" => array("type"=>"cdp", "pop"=>37792, "name_ru"=>"Олимпиа Хайтс", "name_en"=>"Olympia Heights", "lang"=>"en", "lat"=>25.736900, "lon"=>-80.356532),
"/US/US-FL/opa-locka-3345" => array("type"=>"city", "pop"=>15287, "name_ru"=>"Опа-локка", "name_en"=>"Opa-locka", "lang"=>"en", "lat"=>25.899700, "lon"=>-80.255083),
"/US/US-FL/opa-locka-north-3346" => array("type"=>"cdp", "pop"=>6568, "name_ru"=>"Опа-локка Норт", "name_en"=>"Opa-locka North", "lang"=>"en", "lat"=>25.919000, "lon"=>-80.268101),
"/US/US-FL/orange-city-3347" => array("type"=>"city", "pop"=>9799, "name_ru"=>"Оранж Сити", "name_en"=>"Orange City", "lang"=>"en", "lat"=>28.936514, "lon"=>-81.292839),
"/US/US-FL/orange-park-3348" => array("type"=>"town", "pop"=>9043, "name_ru"=>"Оранж Парк", "name_en"=>"Orange Park", "lang"=>"en", "lat"=>30.168600, "lon"=>-81.704732),
"/US/US-FL/orchid-3349" => array("type"=>"town", "pop"=>281, "name_ru"=>"Орчид", "name_en"=>"Orchid", "lang"=>"en", "lat"=>27.777568, "lon"=>-80.421754),
"/US/US-FL/orlando-3350" => array("type"=>"city", "pop"=>230519, "name_ru"=>"Орландо", "name_en"=>"Orlando", "lang"=>"en", "lat"=>28.504748, "lon"=>-81.374248),
"/US/US-FL/orlovista-3351" => array("type"=>"cdp", "pop"=>5990, "name_ru"=>"Орловиста", "name_en"=>"Orlovista", "lang"=>"en", "lat"=>28.544050, "lon"=>-81.462999),
"/US/US-FL/ormond-beach-3352" => array("type"=>"city", "pop"=>37984, "name_ru"=>"Ормонд Бич", "name_en"=>"Ormond Beach", "lang"=>"en", "lat"=>29.294150, "lon"=>-81.094599),
"/US/US-FL/ormond-by-the-sea-3353" => array("type"=>"cdp", "pop"=>8157, "name_ru"=>"Ормонд-Бай-Зе-Си", "name_en"=>"Ormond-By-The-Sea", "lang"=>"en", "lat"=>29.343700, "lon"=>-81.068073),
"/US/US-FL/osprey-3354" => array("type"=>"cdp", "pop"=>2597, "name_ru"=>"Оспри", "name_en"=>"Osprey", "lang"=>"en", "lat"=>27.191500, "lon"=>-82.490763),
"/US/US-FL/otter-creek-3355" => array("type"=>"town", "pop"=>129, "name_ru"=>"Оттер Крик", "name_en"=>"Otter Creek", "lang"=>"en", "lat"=>29.323979, "lon"=>-82.773355),
"/US/US-FL/oviedo-3356" => array("type"=>"city", "pop"=>32606, "name_ru"=>"Овиедо", "name_en"=>"Oviedo", "lang"=>"en", "lat"=>28.657450, "lon"=>-81.181413),
"/US/US-FL/pace-3357" => array("type"=>"cdp", "pop"=>6277, "name_ru"=>"Пэйс", "name_en"=>"Pace", "lang"=>"en", "lat"=>30.599872, "lon"=>-87.159699),
"/US/US-FL/page-park-pine-manor-3358" => array("type"=>"cdp", "pop"=>5116, "name_ru"=>"Пэйдж Парк-Пайн Мэйнор", "name_en"=>"Page Park-Pine Manor", "lang"=>"en", "lat"=>26.573829, "lon"=>-81.870003),
"/US/US-FL/pahokee-3359" => array("type"=>"city", "pop"=>6598, "name_ru"=>"Пэхоуки", "name_en"=>"Pahokee", "lang"=>"en", "lat"=>26.818922, "lon"=>-80.661985),
"/US/US-FL/palatka-3360" => array("type"=>"city", "pop"=>10813, "name_ru"=>"Палатка", "name_en"=>"Palatka", "lang"=>"en", "lat"=>29.649206, "lon"=>-81.656566),
"/US/US-FL/palm-bay-3361" => array("type"=>"city", "pop"=>100786, "name_ru"=>"Палм Бэй", "name_en"=>"Palm Bay", "lang"=>"en", "lat"=>27.985661, "lon"=>-80.660192),
"/US/US-FL/palm-beach-3362" => array("type"=>"town", "pop"=>9535, "name_ru"=>"Палм Бич", "name_en"=>"Palm Beach", "lang"=>"en", "lat"=>26.692850, "lon"=>-80.038147),
"/US/US-FL/palm-beach-gardens-3363" => array("type"=>"city", "pop"=>49227, "name_ru"=>"Палм Бич Гарденс", "name_en"=>"Palm Beach Gardens", "lang"=>"en", "lat"=>26.837450, "lon"=>-80.120377),
"/US/US-FL/palm-beach-shores-3364" => array("type"=>"town", "pop"=>1538, "name_ru"=>"Палм Бич Шорс", "name_en"=>"Palm Beach Shores", "lang"=>"en", "lat"=>26.777050, "lon"=>-80.034699),
"/US/US-FL/palm-city-3365" => array("type"=>"cdp", "pop"=>3925, "name_ru"=>"Палм Сити", "name_en"=>"Palm City", "lang"=>"en", "lat"=>27.161273, "lon"=>-80.268880),
"/US/US-FL/palm-coast-3366" => array("type"=>"cdp", "pop"=>72316, "name_ru"=>"Палм Коаст", "name_en"=>"Palm Coast", "lang"=>"en", "lat"=>29.571147, "lon"=>-81.214104),
"/US/US-FL/palm-harbor-3369" => array("type"=>"cdp", "pop"=>50256, "name_ru"=>"Палм Харбор", "name_en"=>"Palm Harbor", "lang"=>"en", "lat"=>28.087639, "lon"=>-82.763236),
"/US/US-FL/palm-river-3370" => array("type"=>"cdp", "pop"=>3507, "name_ru"=>"Палм Ривер", "name_en"=>"Palm River", "lang"=>"en", "lat"=>26.276925, "lon"=>-81.776416),
"/US/US-FL/palm-river-clair-mel-3371" => array("type"=>"cdp", "pop"=>13691, "name_ru"=>"Палм Ривер-Клэйр Мел", "name_en"=>"Palm River-Clair Mel", "lang"=>"en", "lat"=>27.931550, "lon"=>-82.383249),
"/US/US-FL/palm-shores-3372" => array("type"=>"town", "pop"=>870, "name_ru"=>"Палм Шорс", "name_en"=>"Palm Shores", "lang"=>"en", "lat"=>28.186287, "lon"=>-80.657532),
"/US/US-FL/palm-springs-3373" => array("type"=>"village", "pop"=>16442, "name_ru"=>"Палм Спрингс", "name_en"=>"Palm Springs", "lang"=>"en", "lat"=>26.637767, "lon"=>-80.095318),
"/US/US-FL/palm-springs-north-3374" => array("type"=>"cdp", "pop"=>5300, "name_ru"=>"Палм Спрингс Норт", "name_en"=>"Palm Springs North", "lang"=>"en", "lat"=>25.932533, "lon"=>-80.333098),
"/US/US-FL/palm-valley-3375" => array("type"=>"cdp", "pop"=>9960, "name_ru"=>"Палм Вэлли", "name_en"=>"Palm Valley", "lang"=>"en", "lat"=>30.199661, "lon"=>-81.391146),
"/US/US-FL/palmetto-3367" => array("type"=>"city", "pop"=>14284, "name_ru"=>"Палметто", "name_en"=>"Palmetto", "lang"=>"en", "lat"=>27.521850, "lon"=>-82.577624),
"/US/US-FL/palmetto-estates-3368" => array("type"=>"cdp", "pop"=>12293, "name_ru"=>"Палметто Эстэйтс", "name_en"=>"Palmetto Estates", "lang"=>"en", "lat"=>25.621133, "lon"=>-80.362199),
"/US/US-FL/panama-city-3376" => array("type"=>"city", "pop"=>36644, "name_ru"=>"Панама Сити", "name_en"=>"Panama City", "lang"=>"en", "lat"=>30.173800, "lon"=>-85.673839),
"/US/US-FL/panama-city-beach-3377" => array("type"=>"city", "pop"=>14642, "name_ru"=>"Панама Сити Бич", "name_en"=>"Panama City Beach", "lang"=>"en", "lat"=>30.209769, "lon"=>-85.852074),
"/US/US-FL/parker-3378" => array("type"=>"city", "pop"=>4537, "name_ru"=>"Паркер", "name_en"=>"Parker", "lang"=>"en", "lat"=>30.128282, "lon"=>-85.600326),
"/US/US-FL/parkland-3379" => array("type"=>"city", "pop"=>24017, "name_ru"=>"Паркланд", "name_en"=>"Parkland", "lang"=>"en", "lat"=>26.314950, "lon"=>-80.237299),
"/US/US-FL/paxton-3380" => array("type"=>"town", "pop"=>809, "name_ru"=>"Пэкстон", "name_en"=>"Paxton", "lang"=>"en", "lat"=>30.975866, "lon"=>-86.312658),
"/US/US-FL/pembroke-park-3381" => array("type"=>"town", "pop"=>4744, "name_ru"=>"Пембрук Парк", "name_en"=>"Pembroke Park", "lang"=>"en", "lat"=>25.983867, "lon"=>-80.177532),
"/US/US-FL/pembroke-pines-3382" => array("type"=>"city", "pop"=>145661, "name_ru"=>"Пембруг Пайнс", "name_en"=>"Pembroke Pines", "lang"=>"en", "lat"=>26.009633, "lon"=>-80.340657),
"/US/US-FL/penney-farms-3383" => array("type"=>"town", "pop"=>623, "name_ru"=>"Пенни Фармс", "name_en"=>"Penney Farms", "lang"=>"en", "lat"=>29.980842, "lon"=>-81.811680),
"/US/US-FL/pensacola-3384" => array("type"=>"city", "pop"=>53820, "name_ru"=>"Пенсакола", "name_en"=>"Pensacola", "lang"=>"en", "lat"=>30.444741, "lon"=>-87.189537),
"/US/US-FL/perrine-3385" => array("type"=>"cdp", "pop"=>15576, "name_ru"=>"Перрин", "name_en"=>"Perrine", "lang"=>"en", "lat"=>25.605550, "lon"=>-80.352549),
"/US/US-FL/perry-3386" => array("type"=>"city", "pop"=>6785, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>30.108694, "lon"=>-83.581716),
"/US/US-FL/pierson-3387" => array("type"=>"town", "pop"=>2564, "name_ru"=>"Пирсон", "name_en"=>"Pierson", "lang"=>"en", "lat"=>29.237662, "lon"=>-81.460925),
"/US/US-FL/pine-castle-3388" => array("type"=>"cdp", "pop"=>8276, "name_ru"=>"Пайн Кастл", "name_en"=>"Pine Castle", "lang"=>"en", "lat"=>28.465550, "lon"=>-81.373049),
"/US/US-FL/pine-hills-3389" => array("type"=>"cdp", "pop"=>35322, "name_ru"=>"Пайн Хиллс", "name_en"=>"Pine Hills", "lang"=>"en", "lat"=>28.580000, "lon"=>-81.460152),
"/US/US-FL/pine-island-ridge-3390" => array("type"=>"cdp", "pop"=>5244, "name_ru"=>"Пайн Айленд Ридж", "name_en"=>"Pine Island Ridge", "lang"=>"en", "lat"=>26.094562, "lon"=>-80.274113),
"/US/US-FL/pinellas-park-3391" => array("type"=>"city", "pop"=>47173, "name_ru"=>"Пайнллас Парк", "name_en"=>"Pinellas Park", "lang"=>"en", "lat"=>27.857200, "lon"=>-82.706024),
"/US/US-FL/pinewood-3392" => array("type"=>"cdp", "pop"=>15518, "name_ru"=>"Пайнвуд", "name_en"=>"Pinewood", "lang"=>"en", "lat"=>25.868700, "lon"=>-80.217299),
"/US/US-FL/placid-lakes-3393" => array("type"=>"cdp", "pop"=>2045, "name_ru"=>"Плэйсид Лэйкс", "name_en"=>"Placid Lakes", "lang"=>"en", "lat"=>27.241226, "lon"=>-81.407113),
"/US/US-FL/plant-city-3397" => array("type"=>"city", "pop"=>32761, "name_ru"=>"Плант Сити", "name_en"=>"Plant City", "lang"=>"en", "lat"=>28.005792, "lon"=>-82.128473),
"/US/US-FL/plantation-3394" => array("type"=>"city", "pop"=>83628, "name_ru"=>"Плэнтейшен", "name_en"=>"Plantation", "lang"=>"en", "lat"=>26.125500, "lon"=>-80.261610),
"/US/US-FL/plantation-3395" => array("type"=>"cdp", "pop"=>1885, "name_ru"=>"Плэнтейшен", "name_en"=>"Plantation", "lang"=>"en", "lat"=>27.067879, "lon"=>-82.373151),
"/US/US-FL/plantation-key-3396" => array("type"=>"cdp", "pop"=>4405, "name_ru"=>"Плантэйшн Ки", "name_en"=>"Plantation Key", "lang"=>"en", "lat"=>24.984313, "lon"=>-80.561218),
"/US/US-FL/poinciana-place-3398" => array("type"=>"cdp", "pop"=>3618, "name_ru"=>"Пойинсиана Плэйс", "name_en"=>"Poinciana Place", "lang"=>"en", "lat"=>28.157250, "lon"=>-81.484911),
"/US/US-FL/polk-city-3399" => array("type"=>"town", "pop"=>1572, "name_ru"=>"Полк Сити", "name_en"=>"Polk City", "lang"=>"en", "lat"=>28.182064, "lon"=>-81.824744),
"/US/US-FL/pomona-park-3400" => array("type"=>"town", "pop"=>802, "name_ru"=>"Помона Парк", "name_en"=>"Pomona Park", "lang"=>"en", "lat"=>29.499203, "lon"=>-81.594302),
"/US/US-FL/pompano-beach-3401" => array("type"=>"city", "pop"=>101943, "name_ru"=>"Помпано Бич", "name_en"=>"Pompano Beach", "lang"=>"en", "lat"=>26.236867, "lon"=>-80.132862),
"/US/US-FL/pompano-beach-highlands-3402" => array("type"=>"cdp", "pop"=>17915, "name_ru"=>"Помпано Бич Хайлендс", "name_en"=>"Pompano Beach Highlands", "lang"=>"en", "lat"=>26.286050, "lon"=>-80.115099),
"/US/US-FL/ponce-de-leon-3403" => array("type"=>"town", "pop"=>465, "name_ru"=>"Понс де Лион", "name_en"=>"Ponce de Leon", "lang"=>"en", "lat"=>30.721953, "lon"=>-85.937037),
"/US/US-FL/ponce-inlet-3404" => array("type"=>"town", "pop"=>3140, "name_ru"=>"Понс Инлет", "name_en"=>"Ponce Inlet", "lang"=>"en", "lat"=>29.100010, "lon"=>-80.930444),
"/US/US-FL/port-charlotte-3405" => array("type"=>"cdp", "pop"=>41535, "name_ru"=>"Порт Шарлотт", "name_en"=>"Port Charlotte", "lang"=>"en", "lat"=>26.985499, "lon"=>-82.108584),
"/US/US-FL/port-la-belle-3406" => array("type"=>"cdp", "pop"=>1512, "name_ru"=>"Порт Ла Бэлль", "name_en"=>"Port La Belle", "lang"=>"en", "lat"=>26.756181, "lon"=>-81.405371),
"/US/US-FL/port-orange-3407" => array("type"=>"city", "pop"=>54616, "name_ru"=>"Порт Оранж", "name_en"=>"Port Orange", "lang"=>"en", "lat"=>29.109549, "lon"=>-81.005993),
"/US/US-FL/port-richey-3408" => array("type"=>"city", "pop"=>3412, "name_ru"=>"Порт Ричи", "name_en"=>"Port Richey", "lang"=>"en", "lat"=>28.275410, "lon"=>-82.728603),
"/US/US-FL/port-salerno-3413" => array("type"=>"cdp", "pop"=>7786, "name_ru"=>"Порт Салерно", "name_en"=>"Port Salerno", "lang"=>"en", "lat"=>27.146460, "lon"=>-80.189777),
"/US/US-FL/port-st.-joe-3409" => array("type"=>"city", "pop"=>3551, "name_ru"=>"Порт Ст. Джо", "name_en"=>"Port St. Joe", "lang"=>"en", "lat"=>29.808810, "lon"=>-85.296080),
"/US/US-FL/port-st.-john-3410" => array("type"=>"cdp", "pop"=>8933, "name_ru"=>"Порт Ст. Джон", "name_en"=>"Port St. John", "lang"=>"en", "lat"=>28.476768, "lon"=>-80.788800),
"/US/US-FL/port-st.-lucie-3411" => array("type"=>"city", "pop"=>154353, "name_ru"=>"Порт Ст. Люси", "name_en"=>"Port St. Lucie", "lang"=>"en", "lat"=>27.279800, "lon"=>-80.354527),
"/US/US-FL/port-st.-lucie-river-park-3412" => array("type"=>"cdp", "pop"=>4874, "name_ru"=>"Порт Ст. Люси-Ривер Парк", "name_en"=>"Port St. Lucie-River Park", "lang"=>"en", "lat"=>27.321286, "lon"=>-80.331805),
"/US/US-FL/pretty-bayou-3414" => array("type"=>"cdp", "pop"=>3839, "name_ru"=>"Притти Байоу", "name_en"=>"Pretty Bayou", "lang"=>"en", "lat"=>30.196450, "lon"=>-85.696799),
"/US/US-FL/princeton-3415" => array("type"=>"cdp", "pop"=>7073, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>25.539733, "lon"=>-80.398699),
"/US/US-FL/punta-gorda-3416" => array("type"=>"city", "pop"=>16343, "name_ru"=>"Пунта Горда", "name_en"=>"Punta Gorda", "lang"=>"en", "lat"=>26.897547, "lon"=>-82.057049),
"/US/US-FL/punta-rassa-3417" => array("type"=>"cdp", "pop"=>1493, "name_ru"=>"Пунта Расса", "name_en"=>"Punta Rassa", "lang"=>"en", "lat"=>26.500335, "lon"=>-82.000455),
"/US/US-FL/quincy-3418" => array("type"=>"city", "pop"=>6858, "name_ru"=>"Квинси", "name_en"=>"Quincy", "lang"=>"en", "lat"=>30.584276, "lon"=>-84.584130),
"/US/US-FL/raiford-3419" => array("type"=>"town", "pop"=>187, "name_ru"=>"Рэйфорд", "name_en"=>"Raiford", "lang"=>"en", "lat"=>30.064119, "lon"=>-82.238903),
"/US/US-FL/rainbow-lakes-3420" => array("type"=>"cdp", "pop"=>1496, "name_ru"=>"Рэйнбоу Лэйкс", "name_en"=>"Rainbow Lakes", "lang"=>"en", "lat"=>26.546914, "lon"=>-80.134986),
"/US/US-FL/reddick-3421" => array("type"=>"town", "pop"=>671, "name_ru"=>"Реддик", "name_en"=>"Reddick", "lang"=>"en", "lat"=>29.367630, "lon"=>-82.198295),
"/US/US-FL/redington-beach-3422" => array("type"=>"town", "pop"=>1482, "name_ru"=>"Редингтон Бич", "name_en"=>"Redington Beach", "lang"=>"en", "lat"=>27.809650, "lon"=>-82.811853),
"/US/US-FL/redington-shores-3423" => array("type"=>"town", "pop"=>2441, "name_ru"=>"Редингтон Шорс", "name_en"=>"Redington Shores", "lang"=>"en", "lat"=>27.829384, "lon"=>-82.832089),
"/US/US-FL/richmond-heights-3424" => array("type"=>"cdp", "pop"=>8583, "name_ru"=>"Ричмонд Хайтс", "name_en"=>"Richmond Heights", "lang"=>"en", "lat"=>25.633067, "lon"=>-80.371638),
"/US/US-FL/ridge-manor-3425" => array("type"=>"cdp", "pop"=>1947, "name_ru"=>"Ридж Мэйнор", "name_en"=>"Ridge Manor", "lang"=>"en", "lat"=>28.502995, "lon"=>-82.184073),
"/US/US-FL/ridge-wood-heights-3426" => array("type"=>"cdp", "pop"=>4851, "name_ru"=>"Ридж Вуд Хайтс", "name_en"=>"Ridge Wood Heights", "lang"=>"en", "lat"=>27.287022, "lon"=>-82.513470),
"/US/US-FL/rio-3427" => array("type"=>"cdp", "pop"=>1054, "name_ru"=>"Рио", "name_en"=>"Rio", "lang"=>"en", "lat"=>27.215457, "lon"=>-80.242711),
"/US/US-FL/riverland-3428" => array("type"=>"cdp", "pop"=>5376, "name_ru"=>"Риверлэнд", "name_en"=>"Riverland", "lang"=>"en", "lat"=>26.093600, "lon"=>-80.187082),
"/US/US-FL/riverview-3429" => array("type"=>"cdp", "pop"=>6478, "name_ru"=>"Ривервью", "name_en"=>"Riverview", "lang"=>"en", "lat"=>27.870512, "lon"=>-82.316708),
"/US/US-FL/riviera-beach-3430" => array("type"=>"city", "pop"=>36430, "name_ru"=>"Ривьера Бич", "name_en"=>"Riviera Beach", "lang"=>"en", "lat"=>26.782950, "lon"=>-80.069362),
"/US/US-FL/rockledge-3431" => array("type"=>"city", "pop"=>24747, "name_ru"=>"Рокледж", "name_en"=>"Rockledge", "lang"=>"en", "lat"=>28.319848, "lon"=>-80.729026),
"/US/US-FL/roseland-3432" => array("type"=>"cdp", "pop"=>1379, "name_ru"=>"Роузлэнд", "name_en"=>"Roseland", "lang"=>"en", "lat"=>27.839861, "lon"=>-80.489463),
"/US/US-FL/rotonda-3433" => array("type"=>"cdp", "pop"=>3576, "name_ru"=>"Ротонда", "name_en"=>"Rotonda", "lang"=>"en", "lat"=>26.881512, "lon"=>-82.276955),
"/US/US-FL/royal-palm-beach-3434" => array("type"=>"village", "pop"=>30312, "name_ru"=>"Роял Палм Бич", "name_en"=>"Royal Palm Beach", "lang"=>"en", "lat"=>26.705509, "lon"=>-80.226786),
"/US/US-FL/ruskin-3435" => array("type"=>"cdp", "pop"=>6046, "name_ru"=>"Раскин", "name_en"=>"Ruskin", "lang"=>"en", "lat"=>27.714360, "lon"=>-82.429076),
"/US/US-FL/safety-harbor-3436" => array("type"=>"city", "pop"=>17089, "name_ru"=>"Сэйфети Харбор", "name_en"=>"Safety Harbor", "lang"=>"en", "lat"=>28.009334, "lon"=>-82.696300),
"/US/US-FL/samoset-3448" => array("type"=>"cdp", "pop"=>3119, "name_ru"=>"Сэймоусет", "name_en"=>"Samoset", "lang"=>"en", "lat"=>27.477000, "lon"=>-82.541524),
"/US/US-FL/samsula-spruce-creek-3449" => array("type"=>"cdp", "pop"=>3404, "name_ru"=>"Самсула-Спрус Крик", "name_en"=>"Samsula-Spruce Creek", "lang"=>"en", "lat"=>29.048748, "lon"=>-81.055813),
"/US/US-FL/san-antonio-3450" => array("type"=>"city", "pop"=>1065, "name_ru"=>"Сан Антонио", "name_en"=>"San Antonio", "lang"=>"en", "lat"=>28.337350, "lon"=>-82.278966),
"/US/US-FL/san-carlos-park-3451" => array("type"=>"cdp", "pop"=>11785, "name_ru"=>"Сан Карлос Парк", "name_en"=>"San Carlos Park", "lang"=>"en", "lat"=>26.476042, "lon"=>-81.815175),
"/US/US-FL/sandalfoot-cove-3452" => array("type"=>"cdp", "pop"=>14214, "name_ru"=>"Сандалфут Ков", "name_en"=>"Sandalfoot Cove", "lang"=>"en", "lat"=>26.338891, "lon"=>-80.187784),
"/US/US-FL/sanford-3453" => array("type"=>"city", "pop"=>50634, "name_ru"=>"Санфорд", "name_en"=>"Sanford", "lang"=>"en", "lat"=>28.789250, "lon"=>-81.275698),
"/US/US-FL/sanibel-3454" => array("type"=>"city", "pop"=>5645, "name_ru"=>"Санибел", "name_en"=>"Sanibel", "lang"=>"en", "lat"=>26.443867, "lon"=>-82.100861),
"/US/US-FL/sarasota-3455" => array("type"=>"city", "pop"=>52340, "name_ru"=>"Сарасота", "name_en"=>"Sarasota", "lang"=>"en", "lat"=>27.336100, "lon"=>-82.546599),
"/US/US-FL/sarasota-springs-3456" => array("type"=>"cdp", "pop"=>16088, "name_ru"=>"Сарасота Спрингс", "name_en"=>"Sarasota Springs", "lang"=>"en", "lat"=>27.308700, "lon"=>-82.479799),
"/US/US-FL/satellite-beach-3457" => array("type"=>"city", "pop"=>11762, "name_ru"=>"Сателлайт Бич", "name_en"=>"Satellite Beach", "lang"=>"en", "lat"=>28.176001, "lon"=>-80.600200),
"/US/US-FL/sawgrass-3458" => array("type"=>"cdp", "pop"=>2999, "name_ru"=>"Соуграсс", "name_en"=>"Sawgrass", "lang"=>"en", "lat"=>30.192400, "lon"=>-81.370833),
"/US/US-FL/scott-lake-3459" => array("type"=>"cdp", "pop"=>14588, "name_ru"=>"Скотт Лэйк", "name_en"=>"Scott Lake", "lang"=>"en", "lat"=>25.941200, "lon"=>-80.232269),
"/US/US-FL/sea-ranch-lakes-3460" => array("type"=>"village", "pop"=>735, "name_ru"=>"Си Ранч Лэйкс", "name_en"=>"Sea Ranch Lakes", "lang"=>"en", "lat"=>26.200111, "lon"=>-80.096221),
"/US/US-FL/sebastian-3461" => array("type"=>"city", "pop"=>20265, "name_ru"=>"Себастьян", "name_en"=>"Sebastian", "lang"=>"en", "lat"=>27.787255, "lon"=>-80.480737),
"/US/US-FL/sebring-3462" => array("type"=>"city", "pop"=>10714, "name_ru"=>"Себринг", "name_en"=>"Sebring", "lang"=>"en", "lat"=>27.490746, "lon"=>-81.453472),
"/US/US-FL/seffner-3463" => array("type"=>"cdp", "pop"=>5371, "name_ru"=>"Сеффнер", "name_en"=>"Seffner", "lang"=>"en", "lat"=>27.997584, "lon"=>-82.274879),
"/US/US-FL/seminole-3464" => array("type"=>"city", "pop"=>19002, "name_ru"=>"Семиноул", "name_en"=>"Seminole", "lang"=>"en", "lat"=>27.841737, "lon"=>-82.786239),
"/US/US-FL/sewall's-point-3465" => array("type"=>"town", "pop"=>2004, "name_ru"=>"Сьюолс Поинт", "name_en"=>"Sewall's Point", "lang"=>"en", "lat"=>27.201516, "lon"=>-80.201535),
"/US/US-FL/shalimar-3466" => array("type"=>"town", "pop"=>701, "name_ru"=>"Шалимар", "name_en"=>"Shalimar", "lang"=>"en", "lat"=>30.444200, "lon"=>-86.581953),
"/US/US-FL/sharpes-3467" => array("type"=>"cdp", "pop"=>3348, "name_ru"=>"Шарпс", "name_en"=>"Sharpes", "lang"=>"en", "lat"=>28.437702, "lon"=>-80.758171),
"/US/US-FL/siesta-key-3468" => array("type"=>"cdp", "pop"=>7772, "name_ru"=>"Сиеста Ки", "name_en"=>"Siesta Key", "lang"=>"en", "lat"=>27.277067, "lon"=>-82.551526),
"/US/US-FL/silver-lake-3469" => array("type"=>"cdp", "pop"=>1573, "name_ru"=>"Сильвер Лэйк", "name_en"=>"Silver Lake", "lang"=>"en", "lat"=>28.841703, "lon"=>-81.798508),
"/US/US-FL/silver-springs-shores-3470" => array("type"=>"cdp", "pop"=>6421, "name_ru"=>"Сильвер Спрингс Шорс", "name_en"=>"Silver Springs Shores", "lang"=>"en", "lat"=>29.103622, "lon"=>-82.006348),
"/US/US-FL/sky-lake-3471" => array("type"=>"cdp", "pop"=>6202, "name_ru"=>"Скай Лэйк", "name_en"=>"Sky Lake", "lang"=>"en", "lat"=>28.460667, "lon"=>-81.391880),
"/US/US-FL/sneads-3472" => array("type"=>"town", "pop"=>1939, "name_ru"=>"Снидс", "name_en"=>"Sneads", "lang"=>"en", "lat"=>30.708752, "lon"=>-84.924551),
"/US/US-FL/solana-3473" => array("type"=>"cdp", "pop"=>1128, "name_ru"=>"Солана", "name_en"=>"Solana", "lang"=>"en", "lat"=>26.939144, "lon"=>-82.028440),
"/US/US-FL/sopchoppy-3474" => array("type"=>"city", "pop"=>494, "name_ru"=>"Сопчопии", "name_en"=>"Sopchoppy", "lang"=>"en", "lat"=>30.059438, "lon"=>-84.488788),
"/US/US-FL/south-apopka-3475" => array("type"=>"cdp", "pop"=>6360, "name_ru"=>"Саут Апопка", "name_en"=>"South Apopka", "lang"=>"en", "lat"=>28.652800, "lon"=>-81.507470),
"/US/US-FL/south-bay-3476" => array("type"=>"city", "pop"=>4503, "name_ru"=>"Саут Бэй", "name_en"=>"South Bay", "lang"=>"en", "lat"=>26.667349, "lon"=>-80.716536),
"/US/US-FL/south-beach-3477" => array("type"=>"cdp", "pop"=>2754, "name_ru"=>"Саут Бич", "name_en"=>"South Beach", "lang"=>"en", "lat"=>27.590826, "lon"=>-80.344323),
"/US/US-FL/south-bradenton-3478" => array("type"=>"cdp", "pop"=>20398, "name_ru"=>"Саут Брэйдентон", "name_en"=>"South Bradenton", "lang"=>"en", "lat"=>27.462669, "lon"=>-82.582013),
"/US/US-FL/south-brooksville-3479" => array("type"=>"cdp", "pop"=>1586, "name_ru"=>"Саут Бруксвилл", "name_en"=>"South Brooksville", "lang"=>"en", "lat"=>28.531208, "lon"=>-82.402404),
"/US/US-FL/south-daytona-3480" => array("type"=>"city", "pop"=>13662, "name_ru"=>"Саут Дайтона", "name_en"=>"South Daytona", "lang"=>"en", "lat"=>29.164833, "lon"=>-81.004809),
"/US/US-FL/south-gate-ridge-3483" => array("type"=>"cdp", "pop"=>5924, "name_ru"=>"Саут Гейт Ридж", "name_en"=>"South Gate Ridge", "lang"=>"en", "lat"=>27.286033, "lon"=>-82.496818),
"/US/US-FL/south-miami-3484" => array("type"=>"city", "pop"=>10707, "name_ru"=>"Саут Майами", "name_en"=>"South Miami", "lang"=>"en", "lat"=>25.708050, "lon"=>-80.295348),
"/US/US-FL/south-miami-heights-3485" => array("type"=>"cdp", "pop"=>30030, "name_ru"=>"Саут Майами Хайтс", "name_en"=>"South Miami Heights", "lang"=>"en", "lat"=>25.588400, "lon"=>-80.385373),
"/US/US-FL/south-palm-beach-3486" => array("type"=>"town", "pop"=>1468, "name_ru"=>"Саут Палм Бич", "name_en"=>"South Palm Beach", "lang"=>"en", "lat"=>26.590050, "lon"=>-80.034147),
"/US/US-FL/south-pasadena-3487" => array("type"=>"city", "pop"=>5548, "name_ru"=>"Саут Пасадина", "name_en"=>"South Pasadena", "lang"=>"en", "lat"=>27.751591, "lon"=>-82.740412),
"/US/US-FL/south-patrick-shores-3488" => array("type"=>"cdp", "pop"=>10249, "name_ru"=>"Саут Патрик Шорс", "name_en"=>"South Patrick Shores", "lang"=>"en", "lat"=>28.201839, "lon"=>-80.609753),
"/US/US-FL/south-sarasota-3489" => array("type"=>"cdp", "pop"=>5298, "name_ru"=>"Саут Сарасот", "name_en"=>"South Sarasota", "lang"=>"en", "lat"=>27.285897, "lon"=>-82.533187),
"/US/US-FL/south-venice-3490" => array("type"=>"cdp", "pop"=>11951, "name_ru"=>"Саут Венис", "name_en"=>"South Venice", "lang"=>"en", "lat"=>27.042600, "lon"=>-82.414266),
"/US/US-FL/southeast-arcadia-3481" => array("type"=>"cdp", "pop"=>4145, "name_ru"=>"Саутист Аркадия", "name_en"=>"Southeast Arcadia", "lang"=>"en", "lat"=>27.187085, "lon"=>-81.851688),
"/US/US-FL/southgate-3482" => array("type"=>"cdp", "pop"=>7324, "name_ru"=>"Саутгейт", "name_en"=>"Southgate", "lang"=>"en", "lat"=>27.307830, "lon"=>-82.510123),
"/US/US-FL/spring-hill-3492" => array("type"=>"cdp", "pop"=>31117, "name_ru"=>"Спринг Хилл", "name_en"=>"Spring Hill", "lang"=>"en", "lat"=>28.482112, "lon"=>-82.561903),
"/US/US-FL/springfield-3491" => array("type"=>"city", "pop"=>8803, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>30.166550, "lon"=>-85.608924),
"/US/US-FL/st.-augustine-3437" => array("type"=>"city", "pop"=>12404, "name_ru"=>"Сент Августин", "name_en"=>"St. Augustine", "lang"=>"en", "lat"=>29.891918, "lon"=>-81.310768),
"/US/US-FL/st.-augustine-beach-3438" => array("type"=>"city", "pop"=>6218, "name_ru"=>"Сент Августин Бич", "name_en"=>"St. Augustine Beach", "lang"=>"en", "lat"=>29.842092, "lon"=>-81.271067),
"/US/US-FL/st.-augustine-shores-3439" => array("type"=>"cdp", "pop"=>4411, "name_ru"=>"Сент Августин Шорс", "name_en"=>"St. Augustine Shores", "lang"=>"en", "lat"=>29.810521, "lon"=>-81.310657),
"/US/US-FL/st.-augustine-south-3440" => array("type"=>"cdp", "pop"=>4218, "name_ru"=>"Сент Августи Саут", "name_en"=>"St. Augustine South", "lang"=>"en", "lat"=>29.842487, "lon"=>-81.315741),
"/US/US-FL/st.-cloud-3441" => array("type"=>"city", "pop"=>28745, "name_ru"=>"Сент Клауд", "name_en"=>"St. Cloud", "lang"=>"en", "lat"=>28.241790, "lon"=>-81.289806),
"/US/US-FL/st.-james-city-3442" => array("type"=>"cdp", "pop"=>1904, "name_ru"=>"Сент Джэймс Сити", "name_en"=>"St. James City", "lang"=>"en", "lat"=>26.499704, "lon"=>-82.080898),
"/US/US-FL/st.-leo-3443" => array("type"=>"town", "pop"=>1384, "name_ru"=>"Сент Лео", "name_en"=>"St. Leo", "lang"=>"en", "lat"=>28.337047, "lon"=>-82.259460),
"/US/US-FL/st.-lucie-3444" => array("type"=>"village", "pop"=>265108, "name_ru"=>"Сент Люси", "name_en"=>"St. Lucie", "lang"=>"en", "lat"=>27.499040, "lon"=>-80.342875),
"/US/US-FL/st.-marks-3445" => array("type"=>"city", "pop"=>319, "name_ru"=>"Сент Маркс", "name_en"=>"St. Marks", "lang"=>"en", "lat"=>30.162766, "lon"=>-84.207287),
"/US/US-FL/st.-petersburg-3446" => array("type"=>"city", "pop"=>245314, "name_ru"=>"Сент Петерсберг", "name_en"=>"St. Petersburg", "lang"=>"en", "lat"=>27.758000, "lon"=>-82.642121),
"/US/US-FL/st.-petersburg-beach-3447" => array("type"=>"city", "pop"=>9200, "name_ru"=>"Сент Петерсберг Бич", "name_en"=>"St. Petersburg Beach", "lang"=>"en", "lat"=>27.710181, "lon"=>-82.778805),
"/US/US-FL/starke-3493" => array("type"=>"city", "pop"=>5972, "name_ru"=>"Старк", "name_en"=>"Starke", "lang"=>"en", "lat"=>29.946937, "lon"=>-82.112159),
"/US/US-FL/stock-island-3494" => array("type"=>"cdp", "pop"=>3613, "name_ru"=>"Сток Айленд", "name_en"=>"Stock Island", "lang"=>"en", "lat"=>24.566576, "lon"=>-81.738570),
"/US/US-FL/stuart-3495" => array("type"=>"city", "pop"=>15858, "name_ru"=>"Стюарт", "name_en"=>"Stuart", "lang"=>"en", "lat"=>27.194266, "lon"=>-80.243176),
"/US/US-FL/sugarmill-woods-3496" => array("type"=>"cdp", "pop"=>4073, "name_ru"=>"Шугармилл Вудс", "name_en"=>"Sugarmill Woods", "lang"=>"en", "lat"=>28.732070, "lon"=>-82.506342),
"/US/US-FL/sun-city-center-3497" => array("type"=>"cdp", "pop"=>8326, "name_ru"=>"Сна Сити Центр", "name_en"=>"Sun City Center", "lang"=>"en", "lat"=>27.721362, "lon"=>-82.352030),
"/US/US-FL/sun-valley-3503" => array("type"=>"cdp", "pop"=>2735, "name_ru"=>"Сан Вэлли", "name_en"=>"Sun Valley", "lang"=>"en", "lat"=>26.527700, "lon"=>-80.137733),
"/US/US-FL/suncoast-estates-3498" => array("type"=>"cdp", "pop"=>4483, "name_ru"=>"Санкоаст Эстэйтс", "name_en"=>"Suncoast Estates", "lang"=>"en", "lat"=>26.711250, "lon"=>-81.869274),
"/US/US-FL/sunny-isles-3499" => array("type"=>"cdp", "pop"=>11772, "name_ru"=>"Санни Айлс", "name_en"=>"Sunny Isles", "lang"=>"en", "lat"=>25.932100, "lon"=>-80.125841),
"/US/US-FL/sunnyside-3500" => array("type"=>"cdp", "pop"=>1008, "name_ru"=>"Саннисайд", "name_en"=>"Sunnyside", "lang"=>"en", "lat"=>28.801290, "lon"=>-81.837506),
"/US/US-FL/sunrise-3501" => array("type"=>"city", "pop"=>89026, "name_ru"=>"Санрайз", "name_en"=>"Sunrise", "lang"=>"en", "lat"=>26.151700, "lon"=>-80.297918),
"/US/US-FL/sunset-3502" => array("type"=>"cdp", "pop"=>15810, "name_ru"=>"Сансет", "name_en"=>"Sunset", "lang"=>"en", "lat"=>25.705450, "lon"=>-80.352561),
"/US/US-FL/surfside-3504" => array("type"=>"town", "pop"=>4520, "name_ru"=>"Серфсайд", "name_en"=>"Surfside", "lang"=>"en", "lat"=>25.878196, "lon"=>-80.123508),
"/US/US-FL/sweetwater-3505" => array("type"=>"city", "pop"=>13072, "name_ru"=>"Свитвотер", "name_en"=>"Sweetwater", "lang"=>"en", "lat"=>25.764700, "lon"=>-80.373824),
"/US/US-FL/sylvan-shores-3506" => array("type"=>"cdp", "pop"=>2155, "name_ru"=>"Силвиан Шорс", "name_en"=>"Sylvan Shores", "lang"=>"en", "lat"=>27.317714, "lon"=>-81.343822),
"/US/US-FL/tallahassee-3507" => array("type"=>"city", "pop"=>171922, "name_ru"=>"Таллахасси", "name_en"=>"Tallahassee", "lang"=>"en", "lat"=>30.457000, "lon"=>-84.281399),
"/US/US-FL/tamarac-3508" => array("type"=>"city", "pop"=>59340, "name_ru"=>"Тамарак", "name_en"=>"Tamarac", "lang"=>"en", "lat"=>26.204750, "lon"=>-80.257715),
"/US/US-FL/tamiami-3509" => array("type"=>"cdp", "pop"=>33845, "name_ru"=>"Тамайами", "name_en"=>"Tamiami", "lang"=>"en", "lat"=>25.758332, "lon"=>-80.398555),
"/US/US-FL/tampa-3510" => array("type"=>"city", "pop"=>340882, "name_ru"=>"Тампа", "name_en"=>"Tampa", "lang"=>"en", "lat"=>27.959000, "lon"=>-82.482120),
"/US/US-FL/tangelo-park-3511" => array("type"=>"cdp", "pop"=>2663, "name_ru"=>"Тангело Парк", "name_en"=>"Tangelo Park", "lang"=>"en", "lat"=>28.455750, "lon"=>-81.447449),
"/US/US-FL/tarpon-springs-3512" => array("type"=>"city", "pop"=>23369, "name_ru"=>"Тарпон Спрингс", "name_en"=>"Tarpon Springs", "lang"=>"en", "lat"=>28.146750, "lon"=>-82.781245),
"/US/US-FL/tavares-3513" => array("type"=>"city", "pop"=>13746, "name_ru"=>"Таварес", "name_en"=>"Tavares", "lang"=>"en", "lat"=>28.804767, "lon"=>-81.735685),
"/US/US-FL/tavernier-3514" => array("type"=>"cdp", "pop"=>2433, "name_ru"=>"Тавернир", "name_en"=>"Tavernier", "lang"=>"en", "lat"=>25.016422, "lon"=>-80.519153),
"/US/US-FL/taylor-creek-3515" => array("type"=>"cdp", "pop"=>4081, "name_ru"=>"Тэйлор Крик", "name_en"=>"Taylor Creek", "lang"=>"en", "lat"=>27.216298, "lon"=>-80.789829),
"/US/US-FL/temple-terrace-3516" => array("type"=>"city", "pop"=>22619, "name_ru"=>"Темпл Террэйс", "name_en"=>"Temple Terrace", "lang"=>"en", "lat"=>28.042450, "lon"=>-82.382202),
"/US/US-FL/tequesta-3517" => array("type"=>"village", "pop"=>5708, "name_ru"=>"Тикуеста", "name_en"=>"Tequesta", "lang"=>"en", "lat"=>26.960100, "lon"=>-80.101589),
"/US/US-FL/the-meadows-3518" => array("type"=>"cdp", "pop"=>3437, "name_ru"=>"Зе Мидоус", "name_en"=>"The Meadows", "lang"=>"en", "lat"=>27.361263, "lon"=>-82.469225),
"/US/US-FL/tice-3519" => array("type"=>"cdp", "pop"=>3971, "name_ru"=>"Тайс", "name_en"=>"Tice", "lang"=>"en", "lat"=>26.674467, "lon"=>-81.817759),
"/US/US-FL/timber-pines-3520" => array("type"=>"cdp", "pop"=>3182, "name_ru"=>"Тимбер Пайнс", "name_en"=>"Timber Pines", "lang"=>"en", "lat"=>28.469862, "lon"=>-82.603270),
"/US/US-FL/titusville-3521" => array("type"=>"city", "pop"=>44756, "name_ru"=>"Титусвилл", "name_en"=>"Titusville", "lang"=>"en", "lat"=>28.588801, "lon"=>-80.820192),
"/US/US-FL/town-'n'-country-3522" => array("type"=>"cdp", "pop"=>60946, "name_ru"=>"Таун-н-Кантри", "name_en"=>"Town 'n' Country", "lang"=>"en", "lat"=>28.010250, "lon"=>-82.577386),
"/US/US-FL/treasure-island-3523" => array("type"=>"city", "pop"=>7455, "name_ru"=>"Треже Айленд", "name_en"=>"Treasure Island", "lang"=>"en", "lat"=>27.767510, "lon"=>-82.768519),
"/US/US-FL/trenton-3524" => array("type"=>"city", "pop"=>1862, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>29.614202, "lon"=>-82.816761),
"/US/US-FL/tyndall-afb-3525" => array("type"=>"cdp", "pop"=>4318, "name_ru"=>"Военно-воздушная базаТиндолл", "name_en"=>"Tyndall AFB", "lang"=>"en", "lat"=>30.083718, "lon"=>-85.608648),
"/US/US-FL/umatilla-3526" => array("type"=>"city", "pop"=>3028, "name_ru"=>"Уматилла", "name_en"=>"Umatilla", "lang"=>"en", "lat"=>28.927521, "lon"=>-81.666858),
"/US/US-FL/union-park-3527" => array("type"=>"cdp", "pop"=>6890, "name_ru"=>"Юнион Парк", "name_en"=>"Union Park", "lang"=>"en", "lat"=>28.563850, "lon"=>-81.236017),
"/US/US-FL/university-west-3528" => array("type"=>"cdp", "pop"=>23760, "name_ru"=>"Юниверсити Вест", "name_en"=>"University West", "lang"=>"en", "lat"=>28.069300, "lon"=>-82.435199),
"/US/US-FL/upper-grand-lagoon-3529" => array("type"=>"cdp", "pop"=>7855, "name_ru"=>"Аппер Гранд Лагун", "name_en"=>"Upper Grand Lagoon", "lang"=>"en", "lat"=>30.162985, "lon"=>-85.740874),
"/US/US-FL/valparaiso-3530" => array("type"=>"city", "pop"=>6030, "name_ru"=>"Валпарайсо", "name_en"=>"Valparaiso", "lang"=>"en", "lat"=>30.493706, "lon"=>-86.503731),
"/US/US-FL/vamo-3531" => array("type"=>"cdp", "pop"=>3325, "name_ru"=>"Вамо", "name_en"=>"Vamo", "lang"=>"en", "lat"=>27.228600, "lon"=>-82.496494),
"/US/US-FL/venice-3532" => array("type"=>"city", "pop"=>20946, "name_ru"=>"Венис", "name_en"=>"Venice", "lang"=>"en", "lat"=>27.097900, "lon"=>-82.432749),
"/US/US-FL/venice-gardens-3533" => array("type"=>"cdp", "pop"=>7701, "name_ru"=>"Венис Гарденс", "name_en"=>"Venice Gardens", "lang"=>"en", "lat"=>27.069033, "lon"=>-82.405018),
"/US/US-FL/vernon-3534" => array("type"=>"city", "pop"=>782, "name_ru"=>"Вернон", "name_en"=>"Vernon", "lang"=>"en", "lat"=>30.618936, "lon"=>-85.705417),
"/US/US-FL/vero-beach-3535" => array("type"=>"city", "pop"=>16982, "name_ru"=>"Веро Бич", "name_en"=>"Vero Beach", "lang"=>"en", "lat"=>27.645877, "lon"=>-80.394210),
"/US/US-FL/vero-beach-south-3536" => array("type"=>"cdp", "pop"=>16973, "name_ru"=>"Веро Бич Саут", "name_en"=>"Vero Beach South", "lang"=>"en", "lat"=>27.615817, "lon"=>-80.413357),
"/US/US-FL/villages-of-oriole-3537" => array("type"=>"cdp", "pop"=>5698, "name_ru"=>"Вилиджис Оф Ориол", "name_en"=>"Villages of Oriole", "lang"=>"en", "lat"=>26.462016, "lon"=>-80.153164),
"/US/US-FL/villano-beach-3538" => array("type"=>"cdp", "pop"=>1867, "name_ru"=>"Виллано Бич", "name_en"=>"Villano Beach", "lang"=>"en", "lat"=>29.938254, "lon"=>-81.302093),
"/US/US-FL/villas-3539" => array("type"=>"cdp", "pop"=>9898, "name_ru"=>"Виллас", "name_en"=>"Villas", "lang"=>"en", "lat"=>26.550016, "lon"=>-81.869021),
"/US/US-FL/virginia-gardens-3540" => array("type"=>"village", "pop"=>2169, "name_ru"=>"Вирджиния Гарденс", "name_en"=>"Virginia Gardens", "lang"=>"en", "lat"=>25.809000, "lon"=>-80.298041),
"/US/US-FL/wabasso-3541" => array("type"=>"cdp", "pop"=>1145, "name_ru"=>"Вабассо", "name_en"=>"Wabasso", "lang"=>"en", "lat"=>27.750496, "lon"=>-80.437729),
"/US/US-FL/wahneta-3542" => array("type"=>"cdp", "pop"=>4024, "name_ru"=>"Ванета", "name_en"=>"Wahneta", "lang"=>"en", "lat"=>27.955904, "lon"=>-81.729271),
"/US/US-FL/waldo-3543" => array("type"=>"city", "pop"=>814, "name_ru"=>"Уолдо", "name_en"=>"Waldo", "lang"=>"en", "lat"=>29.790548, "lon"=>-82.171658),
"/US/US-FL/warm-mineral-springs-3544" => array("type"=>"cdp", "pop"=>4041, "name_ru"=>"Ворм Минерал Спрингс", "name_en"=>"Warm Mineral Springs", "lang"=>"en", "lat"=>27.048533, "lon"=>-82.270320),
"/US/US-FL/warrington-3545" => array("type"=>"cdp", "pop"=>16040, "name_ru"=>"Воррингтон", "name_en"=>"Warrington", "lang"=>"en", "lat"=>30.381200, "lon"=>-87.296496),
"/US/US-FL/washington-park-3546" => array("type"=>"cdp", "pop"=>6930, "name_ru"=>"Вашингтон Парк", "name_en"=>"Washington Park", "lang"=>"en", "lat"=>26.133083, "lon"=>-80.180439),
"/US/US-FL/watertown-3547" => array("type"=>"cdp", "pop"=>3340, "name_ru"=>"Уотертаун", "name_en"=>"Watertown", "lang"=>"en", "lat"=>30.187817, "lon"=>-82.599098),
"/US/US-FL/wauchula-3548" => array("type"=>"city", "pop"=>4511, "name_ru"=>"Ваучула", "name_en"=>"Wauchula", "lang"=>"en", "lat"=>27.545800, "lon"=>-81.809852),
"/US/US-FL/wausau-3549" => array("type"=>"town", "pop"=>419, "name_ru"=>"Ваусау", "name_en"=>"Wausau", "lang"=>"en", "lat"=>30.632477, "lon"=>-85.586662),
"/US/US-FL/waverly-3550" => array("type"=>"cdp", "pop"=>2071, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>27.967584, "lon"=>-81.618888),
"/US/US-FL/webster-3551" => array("type"=>"city", "pop"=>911, "name_ru"=>"Вебстер", "name_en"=>"Webster", "lang"=>"en", "lat"=>28.611769, "lon"=>-82.051440),
"/US/US-FL/weeki-wachee-3552" => array("type"=>"city", "pop"=>31, "name_ru"=>"Вики Вачи", "name_en"=>"Weeki Wachee", "lang"=>"en", "lat"=>28.515172, "lon"=>-82.578744),
"/US/US-FL/weeki-wachee-acres-3553" => array("type"=>"cdp", "pop"=>1394, "name_ru"=>"Вики Вачи Эйкрес", "name_en"=>"Weeki Wachee Acres", "lang"=>"en", "lat"=>28.483066, "lon"=>-82.597789),
"/US/US-FL/weeki-wachee-gardens-3554" => array("type"=>"cdp", "pop"=>1170, "name_ru"=>"Вики Вачи Гарденз", "name_en"=>"Weeki Wachee Gardens", "lang"=>"en", "lat"=>28.533464, "lon"=>-82.629853),
"/US/US-FL/wekiva-springs-3555" => array("type"=>"cdp", "pop"=>23026, "name_ru"=>"Векива Спрингс", "name_en"=>"Wekiva Springs", "lang"=>"en", "lat"=>28.698350, "lon"=>-81.425958),
"/US/US-FL/welaka-3556" => array("type"=>"town", "pop"=>802, "name_ru"=>"Вилака", "name_en"=>"Welaka", "lang"=>"en", "lat"=>29.480922, "lon"=>-81.663494),
"/US/US-FL/wellington-3557" => array("type"=>"cdp", "pop"=>54530, "name_ru"=>"Веллингтон", "name_en"=>"Wellington", "lang"=>"en", "lat"=>26.658422, "lon"=>-80.254826),
"/US/US-FL/west-bradenton-3558" => array("type"=>"cdp", "pop"=>4528, "name_ru"=>"Уэст Брэйдентон", "name_en"=>"West Bradenton", "lang"=>"en", "lat"=>27.502200, "lon"=>-82.614249),
"/US/US-FL/west-de-land-3560" => array("type"=>"cdp", "pop"=>3389, "name_ru"=>"Уэст Де Лэнд", "name_en"=>"West De Land", "lang"=>"en", "lat"=>29.015425, "lon"=>-81.333417),
"/US/US-FL/west-little-river-3562" => array("type"=>"cdp", "pop"=>33575, "name_ru"=>"Уэст Литтл Ривер", "name_en"=>"West Little River", "lang"=>"en", "lat"=>25.856550, "lon"=>-80.237249),
"/US/US-FL/west-melbourne-3563" => array("type"=>"city", "pop"=>15328, "name_ru"=>"Уэст Мелбурн", "name_en"=>"West Melbourne", "lang"=>"en", "lat"=>28.079801, "lon"=>-80.671588),
"/US/US-FL/west-miami-3564" => array("type"=>"city", "pop"=>5533, "name_ru"=>"Уэст Майами", "name_en"=>"West Miami", "lang"=>"en", "lat"=>25.757700, "lon"=>-80.297036),
"/US/US-FL/west-palm-beach-3565" => array("type"=>"city", "pop"=>98779, "name_ru"=>"Уэст Палм Бич", "name_en"=>"West Palm Beach", "lang"=>"en", "lat"=>26.747700, "lon"=>-80.126226),
"/US/US-FL/west-park-3566" => array("type"=>"cdp", "pop"=>14432, "name_ru"=>"Уэст Парк", "name_en"=>"West Park", "lang"=>"en", "lat"=>28.011500, "lon"=>-82.516323),
"/US/US-FL/west-pensacola-3567" => array("type"=>"cdp", "pop"=>22107, "name_ru"=>"Уэст Пенсакола", "name_en"=>"West Pensacola", "lang"=>"en", "lat"=>30.426400, "lon"=>-87.268618),
"/US/US-FL/west-samoset-3568" => array("type"=>"cdp", "pop"=>3819, "name_ru"=>"Уэст Самосет", "name_en"=>"West Samoset", "lang"=>"en", "lat"=>27.469250, "lon"=>-82.556899),
"/US/US-FL/westchester-3559" => array("type"=>"cdp", "pop"=>29883, "name_ru"=>"Уэстчестер", "name_en"=>"Westchester", "lang"=>"en", "lat"=>25.746600, "lon"=>-80.337099),
"/US/US-FL/westgate-belvedere-homes-3561" => array("type"=>"cdp", "pop"=>6880, "name_ru"=>"Уэстгейт-Белведер Хоумс", "name_en"=>"Westgate-Belvedere Homes", "lang"=>"en", "lat"=>26.698900, "lon"=>-80.094704),
"/US/US-FL/westview-3569" => array("type"=>"cdp", "pop"=>9668, "name_ru"=>"Уэствью", "name_en"=>"Westview", "lang"=>"en", "lat"=>25.881750, "lon"=>-80.242203),
"/US/US-FL/westville-3570" => array("type"=>"town", "pop"=>229, "name_ru"=>"Уэствилл", "name_en"=>"Westville", "lang"=>"en", "lat"=>30.769537, "lon"=>-85.845554),
"/US/US-FL/westwood-lakes-3571" => array("type"=>"cdp", "pop"=>11522, "name_ru"=>"Уэствуд Лэйкс", "name_en"=>"Westwood Lakes", "lang"=>"en", "lat"=>25.722450, "lon"=>-80.373599),
"/US/US-FL/wewahitchka-3572" => array("type"=>"city", "pop"=>1665, "name_ru"=>"Уивахитчка", "name_en"=>"Wewahitchka", "lang"=>"en", "lat"=>30.112677, "lon"=>-85.195990),
"/US/US-FL/whiskey-creek-3573" => array("type"=>"cdp", "pop"=>5061, "name_ru"=>"Уиски Крик", "name_en"=>"Whiskey Creek", "lang"=>"en", "lat"=>26.572150, "lon"=>-81.890399),
"/US/US-FL/whisper-walk-3574" => array("type"=>"cdp", "pop"=>3037, "name_ru"=>"Уиспер Уолк", "name_en"=>"Whisper Walk", "lang"=>"en", "lat"=>26.395908, "lon"=>-80.188394),
"/US/US-FL/white-city-3575" => array("type"=>"cdp", "pop"=>4645, "name_ru"=>"Уайт Сити", "name_en"=>"White City", "lang"=>"en", "lat"=>27.378967, "lon"=>-80.342252),
"/US/US-FL/white-springs-3576" => array("type"=>"town", "pop"=>829, "name_ru"=>"Уайт Спрингс", "name_en"=>"White Springs", "lang"=>"en", "lat"=>30.331512, "lon"=>-82.758041),
"/US/US-FL/whitfield-3577" => array("type"=>"cdp", "pop"=>3152, "name_ru"=>"Уайтфилд", "name_en"=>"Whitfield", "lang"=>"en", "lat"=>27.411441, "lon"=>-82.566230),
"/US/US-FL/wildwood-3578" => array("type"=>"city", "pop"=>3889, "name_ru"=>"Уайлдвуд", "name_en"=>"Wildwood", "lang"=>"en", "lat"=>28.854269, "lon"=>-82.040785),
"/US/US-FL/williamsburg-3579" => array("type"=>"cdp", "pop"=>3093, "name_ru"=>"Уилльямсбург", "name_en"=>"Williamsburg", "lang"=>"en", "lat"=>28.414293, "lon"=>-81.443026),
"/US/US-FL/williston-3580" => array("type"=>"city", "pop"=>2918, "name_ru"=>"Уиллистон", "name_en"=>"Williston", "lang"=>"en", "lat"=>29.370366, "lon"=>-82.456977),
"/US/US-FL/willow-oak-3581" => array("type"=>"cdp", "pop"=>4017, "name_ru"=>"Уиллоу Оук", "name_en"=>"Willow Oak", "lang"=>"en", "lat"=>27.918104, "lon"=>-82.021496),
"/US/US-FL/wilton-manors-3582" => array("type"=>"city", "pop"=>12616, "name_ru"=>"Уилтон Мэйнорс", "name_en"=>"Wilton Manors", "lang"=>"en", "lat"=>26.158350, "lon"=>-80.139749),
"/US/US-FL/wimauma-3583" => array("type"=>"cdp", "pop"=>2932, "name_ru"=>"Уимаума", "name_en"=>"Wimauma", "lang"=>"en", "lat"=>27.713078, "lon"=>-82.306646),
"/US/US-FL/windermere-3584" => array("type"=>"town", "pop"=>2567, "name_ru"=>"Уиндермер", "name_en"=>"Windermere", "lang"=>"en", "lat"=>28.495150, "lon"=>-81.533513),
"/US/US-FL/winston-3585" => array("type"=>"cdp", "pop"=>9118, "name_ru"=>"Уинстон", "name_en"=>"Winston", "lang"=>"en", "lat"=>28.042500, "lon"=>-82.011344),
"/US/US-FL/winter-garden-3586" => array("type"=>"city", "pop"=>29385, "name_ru"=>"Уинтер Гарден", "name_en"=>"Winter Garden", "lang"=>"en", "lat"=>28.562300, "lon"=>-81.583299),
"/US/US-FL/winter-haven-3587" => array("type"=>"city", "pop"=>33353, "name_ru"=>"Уинтер Хэвен", "name_en"=>"Winter Haven", "lang"=>"en", "lat"=>28.042261, "lon"=>-81.721820),
"/US/US-FL/winter-park-3588" => array("type"=>"city", "pop"=>27909, "name_ru"=>"Уинтер Парк", "name_en"=>"Winter Park", "lang"=>"en", "lat"=>28.595852, "lon"=>-81.343737),
"/US/US-FL/winter-springs-3589" => array("type"=>"city", "pop"=>32667, "name_ru"=>"Уинтер Спрингс", "name_en"=>"Winter Springs", "lang"=>"en", "lat"=>28.681450, "lon"=>-81.274719),
"/US/US-FL/woodville-3590" => array("type"=>"cdp", "pop"=>2760, "name_ru"=>"Вудвилл", "name_en"=>"Woodville", "lang"=>"en", "lat"=>30.311533, "lon"=>-84.263823),
"/US/US-FL/worthington-springs-3591" => array("type"=>"town", "pop"=>193, "name_ru"=>"Уортингтон Спрингс", "name_en"=>"Worthington Springs", "lang"=>"en", "lat"=>29.931223, "lon"=>-82.424303),
"/US/US-FL/wright-3592" => array("type"=>"cdp", "pop"=>18945, "name_ru"=>"Райт", "name_en"=>"Wright", "lang"=>"en", "lat"=>30.445650, "lon"=>-86.640298),
"/US/US-FL/yalaha-3593" => array("type"=>"cdp", "pop"=>1168, "name_ru"=>"Ялаха", "name_en"=>"Yalaha", "lang"=>"en", "lat"=>28.746503, "lon"=>-81.824237),
"/US/US-FL/yankeetown-3594" => array("type"=>"town", "pop"=>694, "name_ru"=>"Янкитаун", "name_en"=>"Yankeetown", "lang"=>"en", "lat"=>29.032918, "lon"=>-82.759133),
"/US/US-FL/yulee-3595" => array("type"=>"cdp", "pop"=>6915, "name_ru"=>"Юли", "name_en"=>"Yulee", "lang"=>"en", "lat"=>30.632681, "lon"=>-81.573886),
"/US/US-FL/zephyrhills-3596" => array("type"=>"city", "pop"=>13152, "name_ru"=>"Зефирхиллс", "name_en"=>"Zephyrhills", "lang"=>"en", "lat"=>28.235999, "lon"=>-82.176456),
"/US/US-FL/zephyrhills-north-3597" => array("type"=>"cdp", "pop"=>2320, "name_ru"=>"Зефирхиллс Норт", "name_en"=>"Zephyrhills North", "lang"=>"en", "lat"=>28.251235, "lon"=>-82.165749),
"/US/US-FL/zephyrhills-south-3598" => array("type"=>"cdp", "pop"=>2514, "name_ru"=>"Зефирхиллс Саут", "name_en"=>"Zephyrhills South", "lang"=>"en", "lat"=>28.215265, "lon"=>-82.187884),
"/US/US-FL/zephyrhills-west-3599" => array("type"=>"cdp", "pop"=>4249, "name_ru"=>"Зефирхиллс Вест", "name_en"=>"Zephyrhills West", "lang"=>"en", "lat"=>28.231095, "lon"=>-82.203731),
"/US/US-FL/zolfo-springs-3600" => array("type"=>"town", "pop"=>1705, "name_ru"=>"Золфо Спрингс", "name_en"=>"Zolfo Springs", "lang"=>"en", "lat"=>27.494205, "lon"=>-81.794457),
"/US/US-GA" => array("type"=>"state", "pop"=>9685744, "name_ru"=>"Джорджия", "name_en"=>"Georgia", "lang"=>"en"),
"/US/US-GA/abbeville-3601" => array("type"=>"city", "pop"=>2746, "name_ru"=>"Аббевилл", "name_en"=>"Abbeville", "lang"=>"en", "lat"=>31.992381, "lon"=>-83.306519),
"/US/US-GA/acworth-3602" => array("type"=>"city", "pop"=>19476, "name_ru"=>"Акворт", "name_en"=>"Acworth", "lang"=>"en", "lat"=>34.059900, "lon"=>-84.675483),
"/US/US-GA/adairsville-3603" => array("type"=>"city", "pop"=>3230, "name_ru"=>"Адэйрсвилл", "name_en"=>"Adairsville", "lang"=>"en", "lat"=>34.366302, "lon"=>-84.912247),
"/US/US-GA/adel-3604" => array("type"=>"city", "pop"=>5396, "name_ru"=>"Адел", "name_en"=>"Adel", "lang"=>"en", "lat"=>31.136233, "lon"=>-83.428600),
"/US/US-GA/adrian-3605" => array("type"=>"city", "pop"=>566, "name_ru"=>"Эдриан", "name_en"=>"Adrian", "lang"=>"en", "lat"=>32.529742, "lon"=>-82.591455),
"/US/US-GA/ailey-3606" => array("type"=>"city", "pop"=>535, "name_ru"=>"Эйли", "name_en"=>"Ailey", "lang"=>"en", "lat"=>32.187378, "lon"=>-82.570372),
"/US/US-GA/alamo-3607" => array("type"=>"city", "pop"=>2625, "name_ru"=>"Аламо", "name_en"=>"Alamo", "lang"=>"en", "lat"=>32.145969, "lon"=>-82.779901),
"/US/US-GA/alapaha-3608" => array("type"=>"town", "pop"=>692, "name_ru"=>"Алапаха", "name_en"=>"Alapaha", "lang"=>"en", "lat"=>31.383212, "lon"=>-83.223679),
"/US/US-GA/albany-3609" => array("type"=>"city", "pop"=>75831, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>31.577800, "lon"=>-84.175624),
"/US/US-GA/aldora-3610" => array("type"=>"town", "pop"=>91, "name_ru"=>"Алдора", "name_en"=>"Aldora", "lang"=>"en", "lat"=>33.051316, "lon"=>-84.174121),
"/US/US-GA/allenhurst-3611" => array("type"=>"town", "pop"=>709, "name_ru"=>"Алленхерст", "name_en"=>"Allenhurst", "lang"=>"en", "lat"=>31.781274, "lon"=>-81.606412),
"/US/US-GA/allentown-3612" => array("type"=>"town", "pop"=>283, "name_ru"=>"Аллентаун", "name_en"=>"Allentown", "lang"=>"en", "lat"=>32.593866, "lon"=>-83.225992),
"/US/US-GA/alma-3613" => array("type"=>"city", "pop"=>3490, "name_ru"=>"Алма", "name_en"=>"Alma", "lang"=>"en", "lat"=>31.543374, "lon"=>-82.476246),
"/US/US-GA/alpharetta-3614" => array("type"=>"city", "pop"=>49903, "name_ru"=>"Алфаретта", "name_en"=>"Alpharetta", "lang"=>"en", "lat"=>34.068883, "lon"=>-84.271995),
"/US/US-GA/alston-3615" => array("type"=>"town", "pop"=>165, "name_ru"=>"Алстон", "name_en"=>"Alston", "lang"=>"en", "lat"=>32.080515, "lon"=>-82.479623),
"/US/US-GA/alto-3616" => array("type"=>"town", "pop"=>913, "name_ru"=>"Алто", "name_en"=>"Alto", "lang"=>"en", "lat"=>34.467445, "lon"=>-83.574484),
"/US/US-GA/ambrose-3617" => array("type"=>"city", "pop"=>327, "name_ru"=>"Амброуз", "name_en"=>"Ambrose", "lang"=>"en", "lat"=>31.593421, "lon"=>-83.013676),
"/US/US-GA/americus-3618" => array("type"=>"city", "pop"=>16524, "name_ru"=>"Америкус", "name_en"=>"Americus", "lang"=>"en", "lat"=>32.072105, "lon"=>-84.226943),
"/US/US-GA/andersonville-3619" => array("type"=>"village", "pop"=>334, "name_ru"=>"Андерсонвилл", "name_en"=>"Andersonville", "lang"=>"en", "lat"=>32.197780, "lon"=>-84.146014),
"/US/US-GA/arabi-3620" => array("type"=>"town", "pop"=>452, "name_ru"=>"Араби", "name_en"=>"Arabi", "lang"=>"en", "lat"=>31.833296, "lon"=>-83.735019),
"/US/US-GA/aragon-3621" => array("type"=>"city", "pop"=>1079, "name_ru"=>"Арагон", "name_en"=>"Aragon", "lang"=>"en", "lat"=>34.046154, "lon"=>-85.057699),
"/US/US-GA/arcade-3622" => array("type"=>"city", "pop"=>1951, "name_ru"=>"Аркада", "name_en"=>"Arcade", "lang"=>"en", "lat"=>34.076868, "lon"=>-83.560982),
"/US/US-GA/argyle-3623" => array("type"=>"town", "pop"=>152, "name_ru"=>"Аргайл", "name_en"=>"Argyle", "lang"=>"en", "lat"=>31.073727, "lon"=>-82.649415),
"/US/US-GA/arlington-3624" => array("type"=>"city", "pop"=>1502, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>31.437790, "lon"=>-84.725442),
"/US/US-GA/arnoldsville-3625" => array("type"=>"city", "pop"=>329, "name_ru"=>"Арнолдсвилл", "name_en"=>"Arnoldsville", "lang"=>"en", "lat"=>33.910742, "lon"=>-83.218072),
"/US/US-GA/ashburn-3626" => array("type"=>"city", "pop"=>4236, "name_ru"=>"Эшберн", "name_en"=>"Ashburn", "lang"=>"en", "lat"=>31.708923, "lon"=>-83.652651),
"/US/US-GA/athens-3627" => array("type"=>"city", "pop"=>45734, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>33.955300, "lon"=>-83.393701),
"/US/US-GA/atlanta-3628" => array("type"=>"city", "pop"=>537958, "name_ru"=>"Атланта", "name_en"=>"Atlanta", "lang"=>"en", "lat"=>33.762900, "lon"=>-84.422592),
"/US/US-GA/attapulgus-3629" => array("type"=>"town", "pop"=>470, "name_ru"=>"Аттапалгас", "name_en"=>"Attapulgus", "lang"=>"en", "lat"=>30.749343, "lon"=>-84.484602),
"/US/US-GA/auburn-3630" => array("type"=>"city", "pop"=>7496, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>34.013282, "lon"=>-83.833184),
"/US/US-GA/augusta-3631" => array("type"=>"city", "pop"=>44639, "name_ru"=>"Огаста", "name_en"=>"Augusta", "lang"=>"en", "lat"=>33.457350, "lon"=>-81.994991),
"/US/US-GA/austell-3632" => array("type"=>"city", "pop"=>7062, "name_ru"=>"Остелл", "name_en"=>"Austell", "lang"=>"en", "lat"=>33.822000, "lon"=>-84.642932),
"/US/US-GA/avalon-3633" => array("type"=>"town", "pop"=>278, "name_ru"=>"Авалон", "name_en"=>"Avalon", "lang"=>"en", "lat"=>34.501508, "lon"=>-83.194444),
"/US/US-GA/avera-3634" => array("type"=>"city", "pop"=>202, "name_ru"=>"Авера", "name_en"=>"Avera", "lang"=>"en", "lat"=>33.193697, "lon"=>-82.529635),
"/US/US-GA/avondale-estates-3635" => array("type"=>"city", "pop"=>2830, "name_ru"=>"Эйвондэйл Эстэйтс", "name_en"=>"Avondale Estates", "lang"=>"en", "lat"=>33.768567, "lon"=>-84.264805),
"/US/US-GA/baconton-3636" => array("type"=>"city", "pop"=>985, "name_ru"=>"Бэйконтон", "name_en"=>"Baconton", "lang"=>"en", "lat"=>31.375824, "lon"=>-84.161563),
"/US/US-GA/bainbridge-3637" => array("type"=>"city", "pop"=>12337, "name_ru"=>"Бэйнбридж", "name_en"=>"Bainbridge", "lang"=>"en", "lat"=>30.906044, "lon"=>-84.577713),
"/US/US-GA/baldwin-3638" => array("type"=>"city", "pop"=>2967, "name_ru"=>"Болдуин", "name_en"=>"Baldwin", "lang"=>"en", "lat"=>34.492135, "lon"=>-83.550037),
"/US/US-GA/ball-ground-3639" => array("type"=>"city", "pop"=>953, "name_ru"=>"Болл Граунд", "name_en"=>"Ball Ground", "lang"=>"en", "lat"=>34.337221, "lon"=>-84.376942),
"/US/US-GA/barnesville-3640" => array("type"=>"city", "pop"=>5998, "name_ru"=>"Барнсвилл", "name_en"=>"Barnesville", "lang"=>"en", "lat"=>33.050812, "lon"=>-84.155138),
"/US/US-GA/bartow-3641" => array("type"=>"town", "pop"=>280, "name_ru"=>"Бартоу", "name_en"=>"Bartow", "lang"=>"en", "lat"=>32.882722, "lon"=>-82.472527),
"/US/US-GA/barwick-3642" => array("type"=>"town", "pop"=>438, "name_ru"=>"Барвик", "name_en"=>"Barwick", "lang"=>"en", "lat"=>30.892096, "lon"=>-83.739573),
"/US/US-GA/baxley-3643" => array("type"=>"city", "pop"=>4552, "name_ru"=>"Бэксли", "name_en"=>"Baxley", "lang"=>"en", "lat"=>31.764725, "lon"=>-82.351917),
"/US/US-GA/bellville-3644" => array("type"=>"city", "pop"=>142, "name_ru"=>"Беллвилл", "name_en"=>"Bellville", "lang"=>"en", "lat"=>32.152637, "lon"=>-81.973527),
"/US/US-GA/belvedere-park-3645" => array("type"=>"cdp", "pop"=>18089, "name_ru"=>"Белведер Парк", "name_en"=>"Belvedere Park", "lang"=>"en", "lat"=>33.748750, "lon"=>-84.259696),
"/US/US-GA/berkeley-lake-3646" => array("type"=>"city", "pop"=>2096, "name_ru"=>"Беркли Лэйк", "name_en"=>"Berkeley Lake", "lang"=>"en", "lat"=>33.982000, "lon"=>-84.185806),
"/US/US-GA/berlin-3647" => array("type"=>"town", "pop"=>616, "name_ru"=>"Берлин", "name_en"=>"Berlin", "lang"=>"en", "lat"=>31.067217, "lon"=>-83.623760),
"/US/US-GA/bethlehem-3648" => array("type"=>"town", "pop"=>1096, "name_ru"=>"Вифлеем", "name_en"=>"Bethlehem", "lang"=>"en", "lat"=>33.937608, "lon"=>-83.708437),
"/US/US-GA/between-3649" => array("type"=>"town", "pop"=>158, "name_ru"=>"Битвин", "name_en"=>"Between", "lang"=>"en", "lat"=>33.813045, "lon"=>-83.808915),
"/US/US-GA/bibb-city-3650" => array("type"=>"town", "pop"=>597, "name_ru"=>"Бибб Сити", "name_en"=>"Bibb City", "lang"=>"en", "lat"=>32.500600, "lon"=>-84.992474),
"/US/US-GA/bishop-3651" => array("type"=>"town", "pop"=>167, "name_ru"=>"Бишоп", "name_en"=>"Bishop", "lang"=>"en", "lat"=>33.816236, "lon"=>-83.436421),
"/US/US-GA/blackshear-3652" => array("type"=>"city", "pop"=>3498, "name_ru"=>"Блэкшир", "name_en"=>"Blackshear", "lang"=>"en", "lat"=>31.298946, "lon"=>-82.241661),
"/US/US-GA/blacksville-3653" => array("type"=>"cdp", "pop"=>1112, "name_ru"=>"Блэксвилл", "name_en"=>"Blacksville", "lang"=>"en", "lat"=>33.429529, "lon"=>-84.156330),
"/US/US-GA/blairsville-3654" => array("type"=>"city", "pop"=>720, "name_ru"=>"Блэйрсвилл", "name_en"=>"Blairsville", "lang"=>"en", "lat"=>34.873634, "lon"=>-83.951584),
"/US/US-GA/blakely-3655" => array("type"=>"city", "pop"=>5257, "name_ru"=>"Блэйкли", "name_en"=>"Blakely", "lang"=>"en", "lat"=>31.379125, "lon"=>-84.935271),
"/US/US-GA/bloomingdale-3656" => array("type"=>"city", "pop"=>2633, "name_ru"=>"Блумингдэйл", "name_en"=>"Bloomingdale", "lang"=>"en", "lat"=>32.118013, "lon"=>-81.308462),
"/US/US-GA/blue-ridge-3657" => array("type"=>"city", "pop"=>1066, "name_ru"=>"Блу Ридж", "name_en"=>"Blue Ridge", "lang"=>"en", "lat"=>34.864084, "lon"=>-84.327008),
"/US/US-GA/bluffton-3658" => array("type"=>"town", "pop"=>108, "name_ru"=>"Блаффтон", "name_en"=>"Bluffton", "lang"=>"en", "lat"=>31.520015, "lon"=>-84.866958),
"/US/US-GA/blythe-3659" => array("type"=>"town", "pop"=>816, "name_ru"=>"Блайт", "name_en"=>"Blythe", "lang"=>"en", "lat"=>33.292494, "lon"=>-82.199521),
"/US/US-GA/bogart-3660" => array("type"=>"town", "pop"=>152, "name_ru"=>"Богарт", "name_en"=>"Bogart", "lang"=>"en", "lat"=>33.947308, "lon"=>-83.532321),
"/US/US-GA/bonanza-3661" => array("type"=>"cdp", "pop"=>2010, "name_ru"=>"Бонанза", "name_en"=>"Bonanza", "lang"=>"en", "lat"=>33.455850, "lon"=>-84.336584),
"/US/US-GA/boston-3662" => array("type"=>"city", "pop"=>1441, "name_ru"=>"Бостон", "name_en"=>"Boston", "lang"=>"en", "lat"=>30.791702, "lon"=>-83.790843),
"/US/US-GA/bostwick-3663" => array("type"=>"town", "pop"=>359, "name_ru"=>"Боствик", "name_en"=>"Bostwick", "lang"=>"en", "lat"=>33.738094, "lon"=>-83.515285),
"/US/US-GA/bowdon-3664" => array("type"=>"city", "pop"=>2054, "name_ru"=>"Баудон", "name_en"=>"Bowdon", "lang"=>"en", "lat"=>33.537399, "lon"=>-85.254165),
"/US/US-GA/bowersville-3665" => array("type"=>"town", "pop"=>331, "name_ru"=>"Бауэрсвилл", "name_en"=>"Bowersville", "lang"=>"en", "lat"=>34.372247, "lon"=>-83.083337),
"/US/US-GA/bowman-3666" => array("type"=>"city", "pop"=>961, "name_ru"=>"Баумэн", "name_en"=>"Bowman", "lang"=>"en", "lat"=>34.203438, "lon"=>-83.028886),
"/US/US-GA/braselton-3667" => array("type"=>"town", "pop"=>5565, "name_ru"=>"Брэйзелтон", "name_en"=>"Braselton", "lang"=>"en", "lat"=>34.102260, "lon"=>-83.788163),
"/US/US-GA/braswell-3668" => array("type"=>"city", "pop"=>125, "name_ru"=>"Брэсвелл", "name_en"=>"Braswell", "lang"=>"en", "lat"=>33.983080, "lon"=>-84.966537),
"/US/US-GA/bremen-3669" => array("type"=>"city", "pop"=>5709, "name_ru"=>"Бремен", "name_en"=>"Bremen", "lang"=>"en", "lat"=>33.708369, "lon"=>-85.152252),
"/US/US-GA/brinson-3670" => array("type"=>"town", "pop"=>216, "name_ru"=>"Бринсон", "name_en"=>"Brinson", "lang"=>"en", "lat"=>30.978831, "lon"=>-84.737303),
"/US/US-GA/bronwood-3671" => array("type"=>"town", "pop"=>469, "name_ru"=>"Бронвуд", "name_en"=>"Bronwood", "lang"=>"en", "lat"=>31.830792, "lon"=>-84.364022),
"/US/US-GA/brooklet-3672" => array("type"=>"town", "pop"=>1335, "name_ru"=>"Бруклет", "name_en"=>"Brooklet", "lang"=>"en", "lat"=>32.388867, "lon"=>-81.669307),
"/US/US-GA/brooks-3673" => array("type"=>"town", "pop"=>661, "name_ru"=>"Брукс", "name_en"=>"Brooks", "lang"=>"en", "lat"=>33.294992, "lon"=>-84.459194),
"/US/US-GA/broxton-3674" => array("type"=>"city", "pop"=>1465, "name_ru"=>"Брокстон", "name_en"=>"Broxton", "lang"=>"en", "lat"=>31.623792, "lon"=>-82.886894),
"/US/US-GA/brunswick-3675" => array("type"=>"city", "pop"=>16269, "name_ru"=>"Брунсвик", "name_en"=>"Brunswick", "lang"=>"en", "lat"=>31.142877, "lon"=>-81.470978),
"/US/US-GA/buchanan-3676" => array("type"=>"city", "pop"=>1041, "name_ru"=>"Бачанан", "name_en"=>"Buchanan", "lang"=>"en", "lat"=>33.801232, "lon"=>-85.180571),
"/US/US-GA/buckhead-3677" => array("type"=>"town", "pop"=>223, "name_ru"=>"Бакхед", "name_en"=>"Buckhead", "lang"=>"en", "lat"=>33.568239, "lon"=>-83.363727),
"/US/US-GA/buena-vista-3678" => array("type"=>"city", "pop"=>1642, "name_ru"=>"Буена Виста", "name_en"=>"Buena Vista", "lang"=>"en", "lat"=>32.318357, "lon"=>-84.517594),
"/US/US-GA/buford-3679" => array("type"=>"city", "pop"=>11780, "name_ru"=>"Бьюфорд", "name_en"=>"Buford", "lang"=>"en", "lat"=>34.116980, "lon"=>-83.998629),
"/US/US-GA/butler-3680" => array("type"=>"city", "pop"=>1817, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>32.557059, "lon"=>-84.237504),
"/US/US-GA/byromville-3681" => array("type"=>"town", "pop"=>403, "name_ru"=>"Байромвилл", "name_en"=>"Byromville", "lang"=>"en", "lat"=>32.201334, "lon"=>-83.908126),
"/US/US-GA/byron-3682" => array("type"=>"city", "pop"=>4324, "name_ru"=>"Байрон", "name_en"=>"Byron", "lang"=>"en", "lat"=>32.651957, "lon"=>-83.757553),
"/US/US-GA/cadwell-3683" => array("type"=>"town", "pop"=>372, "name_ru"=>"Кэдвелл", "name_en"=>"Cadwell", "lang"=>"en", "lat"=>32.340183, "lon"=>-83.042593),
"/US/US-GA/cairo-3684" => array("type"=>"city", "pop"=>9809, "name_ru"=>"Каир", "name_en"=>"Cairo", "lang"=>"en", "lat"=>30.879305, "lon"=>-84.206033),
"/US/US-GA/calhoun-3685" => array("type"=>"city", "pop"=>14807, "name_ru"=>"Калхун", "name_en"=>"Calhoun", "lang"=>"en", "lat"=>34.494487, "lon"=>-84.943827),
"/US/US-GA/camak-3686" => array("type"=>"town", "pop"=>149, "name_ru"=>"Камак", "name_en"=>"Camak", "lang"=>"en", "lat"=>33.451321, "lon"=>-82.646691),
"/US/US-GA/camilla-3687" => array("type"=>"city", "pop"=>5660, "name_ru"=>"Камилла", "name_en"=>"Camilla", "lang"=>"en", "lat"=>31.229739, "lon"=>-84.210058),
"/US/US-GA/candler-mcafee-3688" => array("type"=>"cdp", "pop"=>29491, "name_ru"=>"Кандлер-Макафи", "name_en"=>"Candler-McAfee", "lang"=>"en", "lat"=>33.725600, "lon"=>-84.270298),
"/US/US-GA/canon-3689" => array("type"=>"city", "pop"=>812, "name_ru"=>"Кэнон", "name_en"=>"Canon", "lang"=>"en", "lat"=>34.346186, "lon"=>-83.110723),
"/US/US-GA/canton-3690" => array("type"=>"city", "pop"=>22724, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>34.249001, "lon"=>-84.490190),
"/US/US-GA/carl-3691" => array("type"=>"town", "pop"=>318, "name_ru"=>"Карл", "name_en"=>"Carl", "lang"=>"en", "lat"=>34.004960, "lon"=>-83.811341),
"/US/US-GA/carlton-3692" => array("type"=>"city", "pop"=>246, "name_ru"=>"Карлтон", "name_en"=>"Carlton", "lang"=>"en", "lat"=>34.042681, "lon"=>-83.038592),
"/US/US-GA/carnesville-3693" => array("type"=>"city", "pop"=>673, "name_ru"=>"Карнесвилл", "name_en"=>"Carnesville", "lang"=>"en", "lat"=>34.367204, "lon"=>-83.233317),
"/US/US-GA/carrollton-3694" => array("type"=>"city", "pop"=>23291, "name_ru"=>"Карролтон", "name_en"=>"Carrollton", "lang"=>"en", "lat"=>33.583836, "lon"=>-85.075877),
"/US/US-GA/cartersville-3695" => array("type"=>"city", "pop"=>19010, "name_ru"=>"Картерсвилл", "name_en"=>"Cartersville", "lang"=>"en", "lat"=>34.167365, "lon"=>-84.801394),
"/US/US-GA/cave-spring-3696" => array("type"=>"city", "pop"=>1051, "name_ru"=>"Кейв Спринг", "name_en"=>"Cave Spring", "lang"=>"en", "lat"=>34.109502, "lon"=>-85.333851),
"/US/US-GA/cecil-3697" => array("type"=>"town", "pop"=>318, "name_ru"=>"Сесил", "name_en"=>"Cecil", "lang"=>"en", "lat"=>31.045581, "lon"=>-83.393083),
"/US/US-GA/cedartown-3698" => array("type"=>"city", "pop"=>10123, "name_ru"=>"Сидартаун", "name_en"=>"Cedartown", "lang"=>"en", "lat"=>34.017272, "lon"=>-85.257241),
"/US/US-GA/centerville-3699" => array("type"=>"city", "pop"=>7240, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>32.631968, "lon"=>-83.676366),
"/US/US-GA/centralhatchee-3700" => array("type"=>"town", "pop"=>391, "name_ru"=>"Сентралхатчи", "name_en"=>"Centralhatchee", "lang"=>"en", "lat"=>33.368361, "lon"=>-85.104240),
"/US/US-GA/chamblee-3701" => array("type"=>"city", "pop"=>11202, "name_ru"=>"Чембли", "name_en"=>"Chamblee", "lang"=>"en", "lat"=>33.887150, "lon"=>-84.303012),
"/US/US-GA/chatsworth-3702" => array("type"=>"city", "pop"=>4130, "name_ru"=>"Чатсворт", "name_en"=>"Chatsworth", "lang"=>"en", "lat"=>34.778352, "lon"=>-84.776572),
"/US/US-GA/chattanooga-valley-3703" => array("type"=>"cdp", "pop"=>4088, "name_ru"=>"Чаттануга Валли", "name_en"=>"Chattanooga Valley", "lang"=>"en", "lat"=>34.922536, "lon"=>-85.348670),
"/US/US-GA/chauncey-3704" => array("type"=>"town", "pop"=>304, "name_ru"=>"Чаунси", "name_en"=>"Chauncey", "lang"=>"en", "lat"=>32.106517, "lon"=>-83.065235),
"/US/US-GA/chester-3705" => array("type"=>"town", "pop"=>1514, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>32.394352, "lon"=>-83.154510),
"/US/US-GA/chickamauga-3706" => array("type"=>"city", "pop"=>2589, "name_ru"=>"Чикамога", "name_en"=>"Chickamauga", "lang"=>"en", "lat"=>34.872394, "lon"=>-85.291487),
"/US/US-GA/clarkesville-3707" => array("type"=>"city", "pop"=>1703, "name_ru"=>"Кларксвилл", "name_en"=>"Clarkesville", "lang"=>"en", "lat"=>34.609473, "lon"=>-83.522820),
"/US/US-GA/clarkston-3708" => array("type"=>"city", "pop"=>7836, "name_ru"=>"Кларкстон", "name_en"=>"Clarkston", "lang"=>"en", "lat"=>33.810850, "lon"=>-84.240706),
"/US/US-GA/claxton-3709" => array("type"=>"city", "pop"=>2409, "name_ru"=>"Клэкстон", "name_en"=>"Claxton", "lang"=>"en", "lat"=>32.161414, "lon"=>-81.907931),
"/US/US-GA/clayton-3710" => array("type"=>"city", "pop"=>2185, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>34.877432, "lon"=>-83.400901),
"/US/US-GA/clermont-3711" => array("type"=>"town", "pop"=>782, "name_ru"=>"Клермонт", "name_en"=>"Clermont", "lang"=>"en", "lat"=>34.477074, "lon"=>-83.774141),
"/US/US-GA/cleveland-3712" => array("type"=>"city", "pop"=>2649, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>34.596939, "lon"=>-83.764152),
"/US/US-GA/climax-3713" => array("type"=>"city", "pop"=>291, "name_ru"=>"Клаймакс", "name_en"=>"Climax", "lang"=>"en", "lat"=>30.875624, "lon"=>-84.431252),
"/US/US-GA/cobbtown-3714" => array("type"=>"city", "pop"=>325, "name_ru"=>"Коббтаун", "name_en"=>"Cobbtown", "lang"=>"en", "lat"=>32.281091, "lon"=>-82.138926),
"/US/US-GA/cochran-3715" => array("type"=>"city", "pop"=>5211, "name_ru"=>"Кокран", "name_en"=>"Cochran", "lang"=>"en", "lat"=>32.387309, "lon"=>-83.354909),
"/US/US-GA/cohutta-3716" => array("type"=>"town", "pop"=>604, "name_ru"=>"Коатта", "name_en"=>"Cohutta", "lang"=>"en", "lat"=>34.954877, "lon"=>-84.954482),
"/US/US-GA/colbert-3717" => array("type"=>"city", "pop"=>527, "name_ru"=>"Колберт", "name_en"=>"Colbert", "lang"=>"en", "lat"=>34.036994, "lon"=>-83.212964),
"/US/US-GA/coleman-3718" => array("type"=>"city", "pop"=>137, "name_ru"=>"Коулман", "name_en"=>"Coleman", "lang"=>"en", "lat"=>31.672628, "lon"=>-84.890346),
"/US/US-GA/college-park-3719" => array("type"=>"city", "pop"=>19969, "name_ru"=>"Колледж Парк", "name_en"=>"College Park", "lang"=>"en", "lat"=>33.639067, "lon"=>-84.458587),
"/US/US-GA/collins-3720" => array("type"=>"city", "pop"=>552, "name_ru"=>"Коллинс", "name_en"=>"Collins", "lang"=>"en", "lat"=>32.179154, "lon"=>-82.110952),
"/US/US-GA/colquitt-3721" => array("type"=>"city", "pop"=>1913, "name_ru"=>"Колкитт", "name_en"=>"Colquitt", "lang"=>"en", "lat"=>31.171440, "lon"=>-84.734170),
"/US/US-GA/columbus-(remainder)-3722" => array("type"=>"(remainder)", "pop"=>178681, "name_ru"=>"Колумбус", "name_en"=>"Columbus (remainder)", "lang"=>"en", "lat"=>32.510710, "lon"=>-84.874972),
"/US/US-GA/comer-3723" => array("type"=>"city", "pop"=>1289, "name_ru"=>"Корнер", "name_en"=>"Comer", "lang"=>"en", "lat"=>34.061921, "lon"=>-83.125372),
"/US/US-GA/commerce-3724" => array("type"=>"city", "pop"=>6394, "name_ru"=>"Коммерс", "name_en"=>"Commerce", "lang"=>"en", "lat"=>34.205878, "lon"=>-83.461179),
"/US/US-GA/concord-3725" => array("type"=>"town", "pop"=>410, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>33.091499, "lon"=>-84.438140),
"/US/US-GA/conley-3726" => array("type"=>"cdp", "pop"=>5528, "name_ru"=>"Конли", "name_en"=>"Conley", "lang"=>"en", "lat"=>33.637950, "lon"=>-84.336605),
"/US/US-GA/conyers-3727" => array("type"=>"city", "pop"=>13545, "name_ru"=>"Конайерс", "name_en"=>"Conyers", "lang"=>"en", "lat"=>33.663967, "lon"=>-84.012364),
"/US/US-GA/coolidge-3728" => array("type"=>"city", "pop"=>558, "name_ru"=>"Кулидж", "name_en"=>"Coolidge", "lang"=>"en", "lat"=>31.010548, "lon"=>-83.866656),
"/US/US-GA/cordele-3729" => array("type"=>"city", "pop"=>11518, "name_ru"=>"Кордел", "name_en"=>"Cordele", "lang"=>"en", "lat"=>31.965170, "lon"=>-83.770026),
"/US/US-GA/corinth-3730" => array("type"=>"town", "pop"=>136, "name_ru"=>"Коринт", "name_en"=>"Corinth", "lang"=>"en", "lat"=>33.230333, "lon"=>-84.944008),
"/US/US-GA/cornelia-3731" => array("type"=>"city", "pop"=>3862, "name_ru"=>"Корнелия", "name_en"=>"Cornelia", "lang"=>"en", "lat"=>34.512437, "lon"=>-83.529422),
"/US/US-GA/country-club-estates-3732" => array("type"=>"cdp", "pop"=>7500, "name_ru"=>"Кантри Клаб Истейтс", "name_en"=>"Country Club Estates", "lang"=>"en", "lat"=>31.212452, "lon"=>-81.464353),
"/US/US-GA/covington-3733" => array("type"=>"city", "pop"=>14980, "name_ru"=>"Ковингтон", "name_en"=>"Covington", "lang"=>"en", "lat"=>33.603248, "lon"=>-83.854470),
"/US/US-GA/crawford-3734" => array("type"=>"city", "pop"=>815, "name_ru"=>"Крофорд", "name_en"=>"Crawford", "lang"=>"en", "lat"=>33.882949, "lon"=>-83.156345),
"/US/US-GA/crawfordville-3735" => array("type"=>"city", "pop"=>516, "name_ru"=>"Крофордвилл", "name_en"=>"Crawfordville", "lang"=>"en", "lat"=>33.554449, "lon"=>-82.896902),
"/US/US-GA/culloden-3736" => array("type"=>"city", "pop"=>227, "name_ru"=>"Каллоден", "name_en"=>"Culloden", "lang"=>"en", "lat"=>32.863021, "lon"=>-84.093709),
"/US/US-GA/cumming-3737" => array("type"=>"city", "pop"=>5719, "name_ru"=>"Камминг", "name_en"=>"Cumming", "lang"=>"en", "lat"=>34.208765, "lon"=>-84.135128),
"/US/US-GA/cusseta-3738" => array("type"=>"city", "pop"=>1107, "name_ru"=>"Кассета", "name_en"=>"Cusseta", "lang"=>"en", "lat"=>32.303957, "lon"=>-84.777558),
"/US/US-GA/cuthbert-3739" => array("type"=>"city", "pop"=>3418, "name_ru"=>"Катберт", "name_en"=>"Cuthbert", "lang"=>"en", "lat"=>31.770563, "lon"=>-84.793596),
"/US/US-GA/dacula-3740" => array("type"=>"city", "pop"=>4662, "name_ru"=>"Дакула", "name_en"=>"Dacula", "lang"=>"en", "lat"=>33.984850, "lon"=>-83.887798),
"/US/US-GA/dahlonega-3741" => array("type"=>"city", "pop"=>4947, "name_ru"=>"Далонега", "name_en"=>"Dahlonega", "lang"=>"en", "lat"=>34.536755, "lon"=>-83.984648),
"/US/US-GA/daisy-3742" => array("type"=>"city", "pop"=>138, "name_ru"=>"Дейзи", "name_en"=>"Daisy", "lang"=>"en", "lat"=>32.150852, "lon"=>-81.837125),
"/US/US-GA/dallas-3743" => array("type"=>"city", "pop"=>10584, "name_ru"=>"Даллас", "name_en"=>"Dallas", "lang"=>"en", "lat"=>33.919334, "lon"=>-84.831865),
"/US/US-GA/dalton-3744" => array("type"=>"city", "pop"=>33648, "name_ru"=>"Далтон", "name_en"=>"Dalton", "lang"=>"en", "lat"=>34.769008, "lon"=>-84.975906),
"/US/US-GA/damascus-3745" => array("type"=>"town", "pop"=>257, "name_ru"=>"Дамаскус", "name_en"=>"Damascus", "lang"=>"en", "lat"=>31.298401, "lon"=>-84.717511),
"/US/US-GA/danielsville-3746" => array("type"=>"city", "pop"=>453, "name_ru"=>"Дэниелсвилл", "name_en"=>"Danielsville", "lang"=>"en", "lat"=>34.123884, "lon"=>-83.220560),
"/US/US-GA/danville-3747" => array("type"=>"town", "pop"=>349, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>32.605526, "lon"=>-83.246101),
"/US/US-GA/darien-3748" => array("type"=>"city", "pop"=>1765, "name_ru"=>"Дариен", "name_en"=>"Darien", "lang"=>"en", "lat"=>31.371556, "lon"=>-81.429861),
"/US/US-GA/davisboro-3749" => array("type"=>"city", "pop"=>1762, "name_ru"=>"Дэвисборо", "name_en"=>"Davisboro", "lang"=>"en", "lat"=>32.983204, "lon"=>-82.606130),
"/US/US-GA/dawson-3750" => array("type"=>"city", "pop"=>4600, "name_ru"=>"Доусон", "name_en"=>"Dawson", "lang"=>"en", "lat"=>31.771673, "lon"=>-84.443979),
"/US/US-GA/dawsonville-3751" => array("type"=>"town", "pop"=>1571, "name_ru"=>"Доусонвилл", "name_en"=>"Dawsonville", "lang"=>"en", "lat"=>34.420672, "lon"=>-84.119359),
"/US/US-GA/de-soto-3758" => array("type"=>"village", "pop"=>204, "name_ru"=>"Де Сото", "name_en"=>"De Soto", "lang"=>"en", "lat"=>31.954331, "lon"=>-84.063741),
"/US/US-GA/dearing-3752" => array("type"=>"town", "pop"=>445, "name_ru"=>"Диэринг", "name_en"=>"Dearing", "lang"=>"en", "lat"=>33.415240, "lon"=>-82.384425),
"/US/US-GA/decatur-3753" => array("type"=>"city", "pop"=>18986, "name_ru"=>"Декатур", "name_en"=>"Decatur", "lang"=>"en", "lat"=>33.771250, "lon"=>-84.297809),
"/US/US-GA/deenwood-3754" => array("type"=>"cdp", "pop"=>2055, "name_ru"=>"Динвуд", "name_en"=>"Deenwood", "lang"=>"en", "lat"=>31.247188, "lon"=>-82.366023),
"/US/US-GA/deepstep-3755" => array("type"=>"town", "pop"=>130, "name_ru"=>"Дипстеп", "name_en"=>"Deepstep", "lang"=>"en", "lat"=>33.016053, "lon"=>-82.967588),
"/US/US-GA/demorest-3756" => array("type"=>"city", "pop"=>1797, "name_ru"=>"Деморест", "name_en"=>"Demorest", "lang"=>"en", "lat"=>34.565480, "lon"=>-83.545399),
"/US/US-GA/denton-3757" => array("type"=>"city", "pop"=>292, "name_ru"=>"Дентон", "name_en"=>"Denton", "lang"=>"en", "lat"=>31.722475, "lon"=>-82.696461),
"/US/US-GA/dexter-3759" => array("type"=>"town", "pop"=>547, "name_ru"=>"Декстер", "name_en"=>"Dexter", "lang"=>"en", "lat"=>32.433049, "lon"=>-83.059968),
"/US/US-GA/dillard-3760" => array("type"=>"town", "pop"=>234, "name_ru"=>"Диллард", "name_en"=>"Dillard", "lang"=>"en", "lat"=>34.972472, "lon"=>-83.385607),
"/US/US-GA/dock-junction-3761" => array("type"=>"cdp", "pop"=>7094, "name_ru"=>"Док Джанкшен", "name_en"=>"Dock Junction", "lang"=>"en", "lat"=>31.202152, "lon"=>-81.516921),
"/US/US-GA/doerun-3762" => array("type"=>"city", "pop"=>843, "name_ru"=>"Доеран", "name_en"=>"Doerun", "lang"=>"en", "lat"=>31.319148, "lon"=>-83.918033),
"/US/US-GA/donalsonville-3763" => array("type"=>"city", "pop"=>2694, "name_ru"=>"Дональсонвилл", "name_en"=>"Donalsonville", "lang"=>"en", "lat"=>31.040703, "lon"=>-84.878458),
"/US/US-GA/dooling-3764" => array("type"=>"town", "pop"=>163, "name_ru"=>"Дулинг", "name_en"=>"Dooling", "lang"=>"en", "lat"=>32.229381, "lon"=>-83.929061),
"/US/US-GA/doraville-3765" => array("type"=>"city", "pop"=>10268, "name_ru"=>"Доравилл", "name_en"=>"Doraville", "lang"=>"en", "lat"=>33.906967, "lon"=>-84.274308),
"/US/US-GA/douglas-3766" => array("type"=>"city", "pop"=>11413, "name_ru"=>"Дуглас", "name_en"=>"Douglas", "lang"=>"en", "lat"=>31.505176, "lon"=>-82.851520),
"/US/US-GA/douglasville-3767" => array("type"=>"city", "pop"=>31035, "name_ru"=>"Дугласвилл", "name_en"=>"Douglasville", "lang"=>"en", "lat"=>33.745193, "lon"=>-84.745120),
"/US/US-GA/druid-hills-3768" => array("type"=>"cdp", "pop"=>12174, "name_ru"=>"Друид Хиллс", "name_en"=>"Druid Hills", "lang"=>"en", "lat"=>33.787100, "lon"=>-84.326050),
"/US/US-GA/du-pont-3773" => array("type"=>"town", "pop"=>140, "name_ru"=>"Дю Понт", "name_en"=>"Du Pont", "lang"=>"en", "lat"=>30.989151, "lon"=>-82.869851),
"/US/US-GA/dublin-3769" => array("type"=>"city", "pop"=>17509, "name_ru"=>"Дублин", "name_en"=>"Dublin", "lang"=>"en", "lat"=>32.538258, "lon"=>-82.921587),
"/US/US-GA/dudley-3770" => array("type"=>"city", "pop"=>524, "name_ru"=>"Дадли", "name_en"=>"Dudley", "lang"=>"en", "lat"=>32.532816, "lon"=>-83.080683),
"/US/US-GA/duluth-3771" => array("type"=>"city", "pop"=>26125, "name_ru"=>"Далут", "name_en"=>"Duluth", "lang"=>"en", "lat"=>34.004689, "lon"=>-84.153215),
"/US/US-GA/dunwoody-3772" => array("type"=>"cdp", "pop"=>26302, "name_ru"=>"Данвуди", "name_en"=>"Dunwoody", "lang"=>"en", "lat"=>33.942267, "lon"=>-84.313968),
"/US/US-GA/east-boundary-3774" => array("type"=>"cdp", "pop"=>3271, "name_ru"=>"Ист Баундери", "name_en"=>"East Boundary", "lang"=>"en", "lat"=>33.459878, "lon"=>-81.941686),
"/US/US-GA/east-dublin-3775" => array("type"=>"town", "pop"=>2780, "name_ru"=>"Ист Дублин", "name_en"=>"East Dublin", "lang"=>"en", "lat"=>32.550688, "lon"=>-82.868798),
"/US/US-GA/east-ellijay-3776" => array("type"=>"city", "pop"=>698, "name_ru"=>"Ист Эллиджей", "name_en"=>"East Ellijay", "lang"=>"en", "lat"=>34.682099, "lon"=>-84.474186),
"/US/US-GA/east-griffin-3777" => array("type"=>"cdp", "pop"=>1746, "name_ru"=>"Ист Гриффин", "name_en"=>"East Griffin", "lang"=>"en", "lat"=>33.242990, "lon"=>-84.232335),
"/US/US-GA/east-newnan-3779" => array("type"=>"cdp", "pop"=>1173, "name_ru"=>"Ист Ньюнан", "name_en"=>"East Newnan", "lang"=>"en", "lat"=>33.346309, "lon"=>-84.777534),
"/US/US-GA/east-point-3780" => array("type"=>"city", "pop"=>43418, "name_ru"=>"Ист Пойнт", "name_en"=>"East Point", "lang"=>"en", "lat"=>33.670733, "lon"=>-84.469288),
"/US/US-GA/eastman-3778" => array("type"=>"city", "pop"=>5675, "name_ru"=>"Истман", "name_en"=>"Eastman", "lang"=>"en", "lat"=>32.196229, "lon"=>-83.179329),
"/US/US-GA/eatonton-3781" => array("type"=>"city", "pop"=>6373, "name_ru"=>"Итонтон", "name_en"=>"Eatonton", "lang"=>"en", "lat"=>33.322891, "lon"=>-83.387003),
"/US/US-GA/edge-hill-3782" => array("type"=>"city", "pop"=>33, "name_ru"=>"Эдж Хилл", "name_en"=>"Edge Hill", "lang"=>"en", "lat"=>33.152929, "lon"=>-82.626781),
"/US/US-GA/edison-3783" => array("type"=>"city", "pop"=>1250, "name_ru"=>"Эдисон", "name_en"=>"Edison", "lang"=>"en", "lat"=>31.560720, "lon"=>-84.738065),
"/US/US-GA/elberton-3784" => array("type"=>"city", "pop"=>4546, "name_ru"=>"Эльбертон", "name_en"=>"Elberton", "lang"=>"en", "lat"=>34.105522, "lon"=>-82.864037),
"/US/US-GA/ellaville-3785" => array("type"=>"city", "pop"=>1835, "name_ru"=>"Эллавилл", "name_en"=>"Ellaville", "lang"=>"en", "lat"=>32.237626, "lon"=>-84.309700),
"/US/US-GA/ellenton-3786" => array("type"=>"town", "pop"=>354, "name_ru"=>"Эллентон", "name_en"=>"Ellenton", "lang"=>"en", "lat"=>31.176376, "lon"=>-83.588127),
"/US/US-GA/ellijay-3787" => array("type"=>"city", "pop"=>1588, "name_ru"=>"Эллиджей", "name_en"=>"Ellijay", "lang"=>"en", "lat"=>34.692272, "lon"=>-84.483722),
"/US/US-GA/emerson-3788" => array("type"=>"city", "pop"=>1429, "name_ru"=>"Эмерсон", "name_en"=>"Emerson", "lang"=>"en", "lat"=>34.131791, "lon"=>-84.739428),
"/US/US-GA/enigma-3789" => array("type"=>"town", "pop"=>882, "name_ru"=>"Энигма", "name_en"=>"Enigma", "lang"=>"en", "lat"=>31.409055, "lon"=>-83.330164),
"/US/US-GA/ephesus-3790" => array("type"=>"town", "pop"=>386, "name_ru"=>"Эфесус", "name_en"=>"Ephesus", "lang"=>"en", "lat"=>33.408089, "lon"=>-85.261811),
"/US/US-GA/eton-3791" => array("type"=>"city", "pop"=>479, "name_ru"=>"Этон", "name_en"=>"Eton", "lang"=>"en", "lat"=>34.822955, "lon"=>-84.762858),
"/US/US-GA/euharlee-3792" => array("type"=>"town", "pop"=>4202, "name_ru"=>"Юхарли", "name_en"=>"Euharlee", "lang"=>"en", "lat"=>34.143403, "lon"=>-84.932815),
"/US/US-GA/evans-3793" => array("type"=>"cdp", "pop"=>11646, "name_ru"=>"Эванс", "name_en"=>"Evans", "lang"=>"en", "lat"=>33.513796, "lon"=>-82.145461),
"/US/US-GA/experiment-3794" => array("type"=>"cdp", "pop"=>3762, "name_ru"=>"Эксперимент", "name_en"=>"Experiment", "lang"=>"en", "lat"=>33.278223, "lon"=>-84.277346),
"/US/US-GA/fair-oaks-3797" => array("type"=>"cdp", "pop"=>6996, "name_ru"=>"Фэр Оукс", "name_en"=>"Fair Oaks", "lang"=>"en", "lat"=>33.919700, "lon"=>-84.544575),
"/US/US-GA/fairburn-3795" => array("type"=>"city", "pop"=>11024, "name_ru"=>"Фэрберн", "name_en"=>"Fairburn", "lang"=>"en", "lat"=>33.565450, "lon"=>-84.583545),
"/US/US-GA/fairmount-3796" => array("type"=>"city", "pop"=>809, "name_ru"=>"Фэрмаунт", "name_en"=>"Fairmount", "lang"=>"en", "lat"=>34.440778, "lon"=>-84.701344),
"/US/US-GA/fairview-3798" => array("type"=>"cdp", "pop"=>6444, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>34.930287, "lon"=>-85.292494),
"/US/US-GA/fayetteville-3799" => array("type"=>"city", "pop"=>15136, "name_ru"=>"Фейетвилл", "name_en"=>"Fayetteville", "lang"=>"en", "lat"=>33.445467, "lon"=>-84.461323),
"/US/US-GA/fitzgerald-3800" => array("type"=>"city", "pop"=>9185, "name_ru"=>"Фитцджеральд", "name_en"=>"Fitzgerald", "lang"=>"en", "lat"=>31.706159, "lon"=>-83.253622),
"/US/US-GA/flemington-3801" => array("type"=>"city", "pop"=>350, "name_ru"=>"Флемингтон", "name_en"=>"Flemington", "lang"=>"en", "lat"=>31.849489, "lon"=>-81.561795),
"/US/US-GA/flovilla-3802" => array("type"=>"city", "pop"=>696, "name_ru"=>"Фловилла", "name_en"=>"Flovilla", "lang"=>"en", "lat"=>33.252952, "lon"=>-83.899693),
"/US/US-GA/flowery-branch-3803" => array("type"=>"town", "pop"=>3991, "name_ru"=>"Флауэри Брэнч", "name_en"=>"Flowery Branch", "lang"=>"en", "lat"=>34.185600, "lon"=>-83.923150),
"/US/US-GA/folkston-3804" => array("type"=>"city", "pop"=>3231, "name_ru"=>"Фолкстон", "name_en"=>"Folkston", "lang"=>"en", "lat"=>30.836957, "lon"=>-82.009112),
"/US/US-GA/forest-park-3805" => array("type"=>"city", "pop"=>21726, "name_ru"=>"Форест Парк", "name_en"=>"Forest Park", "lang"=>"en", "lat"=>33.619033, "lon"=>-84.355363),
"/US/US-GA/forsyth-3806" => array("type"=>"city", "pop"=>5022, "name_ru"=>"Форсайт", "name_en"=>"Forsyth", "lang"=>"en", "lat"=>33.034291, "lon"=>-83.939984),
"/US/US-GA/fort-benning-south-3807" => array("type"=>"cdp", "pop"=>14617, "name_ru"=>"Форт Беннинг Саут", "name_en"=>"Fort Benning South", "lang"=>"en", "lat"=>32.362341, "lon"=>-84.949259),
"/US/US-GA/fort-gaines-3808" => array("type"=>"city", "pop"=>1002, "name_ru"=>"Форт Гейнс", "name_en"=>"Fort Gaines", "lang"=>"en", "lat"=>31.626379, "lon"=>-85.054812),
"/US/US-GA/fort-gordon-3809" => array("type"=>"cdp", "pop"=>9140, "name_ru"=>"Форт Гордон", "name_en"=>"Fort Gordon", "lang"=>"en", "lat"=>33.417109, "lon"=>-82.140946),
"/US/US-GA/fort-oglethorpe-3810" => array("type"=>"city", "pop"=>9663, "name_ru"=>"Форт Оглторп", "name_en"=>"Fort Oglethorpe", "lang"=>"en", "lat"=>34.922773, "lon"=>-85.253392),
"/US/US-GA/fort-stewart-3811" => array("type"=>"cdp", "pop"=>13774, "name_ru"=>"Форт Стьюарт", "name_en"=>"Fort Stewart", "lang"=>"en", "lat"=>31.872466, "lon"=>-81.610065),
"/US/US-GA/fort-valley-3812" => array("type"=>"city", "pop"=>8200, "name_ru"=>"Форт Валли", "name_en"=>"Fort Valley", "lang"=>"en", "lat"=>32.554816, "lon"=>-83.888662),
"/US/US-GA/franklin-3813" => array("type"=>"city", "pop"=>885, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>33.281410, "lon"=>-85.097673),
"/US/US-GA/franklin-springs-3814" => array("type"=>"city", "pop"=>804, "name_ru"=>"Франклин Спрингс", "name_en"=>"Franklin Springs", "lang"=>"en", "lat"=>34.283454, "lon"=>-83.146350),
"/US/US-GA/funston-3815" => array("type"=>"town", "pop"=>455, "name_ru"=>"Фунстон", "name_en"=>"Funston", "lang"=>"en", "lat"=>31.200219, "lon"=>-83.873257),
"/US/US-GA/gaines-school-3816" => array("type"=>"cdp", "pop"=>11354, "name_ru"=>"Гейнс Скул", "name_en"=>"Gaines School", "lang"=>"en", "lat"=>33.926600, "lon"=>-83.339449),
"/US/US-GA/gainesville-3817" => array("type"=>"city", "pop"=>35668, "name_ru"=>"Гейнсвилл", "name_en"=>"Gainesville", "lang"=>"en", "lat"=>34.294073, "lon"=>-83.834977),
"/US/US-GA/gainesville-mills-3818" => array("type"=>"cdp", "pop"=>1329, "name_ru"=>"Гейнсвилл Миллс", "name_en"=>"Gainesville Mills", "lang"=>"en", "lat"=>34.277357, "lon"=>-83.817911),
"/US/US-GA/garden-city-3819" => array("type"=>"city", "pop"=>9414, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>32.103483, "lon"=>-81.162865),
"/US/US-GA/garfield-3820" => array("type"=>"town", "pop"=>158, "name_ru"=>"Гарфилд", "name_en"=>"Garfield", "lang"=>"en", "lat"=>32.650349, "lon"=>-82.096478),
"/US/US-GA/gay-3821" => array("type"=>"town", "pop"=>132, "name_ru"=>"Гей", "name_en"=>"Gay", "lang"=>"en", "lat"=>33.093672, "lon"=>-84.573998),
"/US/US-GA/geneva-3822" => array("type"=>"town", "pop"=>108, "name_ru"=>"Дженива", "name_en"=>"Geneva", "lang"=>"en", "lat"=>32.579598, "lon"=>-84.552202),
"/US/US-GA/georgetown-3823" => array("type"=>"town", "pop"=>913, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>31.885383, "lon"=>-85.105906),
"/US/US-GA/georgetown-3824" => array("type"=>"cdp", "pop"=>5554, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>31.980952, "lon"=>-81.229807),
"/US/US-GA/gibson-3825" => array("type"=>"city", "pop"=>761, "name_ru"=>"Гибсон", "name_en"=>"Gibson", "lang"=>"en", "lat"=>33.232592, "lon"=>-82.595449),
"/US/US-GA/gillsville-3826" => array("type"=>"town", "pop"=>211, "name_ru"=>"Джилсвилл", "name_en"=>"Gillsville", "lang"=>"en", "lat"=>34.311517, "lon"=>-83.637487),
"/US/US-GA/girard-3827" => array("type"=>"town", "pop"=>223, "name_ru"=>"Жирар", "name_en"=>"Girard", "lang"=>"en", "lat"=>33.040840, "lon"=>-81.711274),
"/US/US-GA/glennville-3828" => array("type"=>"city", "pop"=>5397, "name_ru"=>"Гленсвилл", "name_en"=>"Glennville", "lang"=>"en", "lat"=>31.936232, "lon"=>-81.931620),
"/US/US-GA/glenwood-3829" => array("type"=>"city", "pop"=>895, "name_ru"=>"Гленвуд", "name_en"=>"Glenwood", "lang"=>"en", "lat"=>32.180598, "lon"=>-82.671673),
"/US/US-GA/good-hope-3830" => array("type"=>"town", "pop"=>280, "name_ru"=>"Гуд Хоп", "name_en"=>"Good Hope", "lang"=>"en", "lat"=>33.783912, "lon"=>-83.609770),
"/US/US-GA/gordon-3831" => array("type"=>"city", "pop"=>2108, "name_ru"=>"Гордон", "name_en"=>"Gordon", "lang"=>"en", "lat"=>32.885703, "lon"=>-83.335469),
"/US/US-GA/grantville-3832" => array("type"=>"city", "pop"=>2666, "name_ru"=>"Грантвилл", "name_en"=>"Grantville", "lang"=>"en", "lat"=>33.237972, "lon"=>-84.827867),
"/US/US-GA/gray-3833" => array("type"=>"city", "pop"=>2221, "name_ru"=>"Грей", "name_en"=>"Gray", "lang"=>"en", "lat"=>33.007390, "lon"=>-83.535748),
"/US/US-GA/grayson-3834" => array("type"=>"city", "pop"=>1507, "name_ru"=>"Грейсон", "name_en"=>"Grayson", "lang"=>"en", "lat"=>33.893200, "lon"=>-83.955514),
"/US/US-GA/greensboro-3835" => array("type"=>"city", "pop"=>3282, "name_ru"=>"Гринсборо", "name_en"=>"Greensboro", "lang"=>"en", "lat"=>33.574137, "lon"=>-83.183717),
"/US/US-GA/greenville-3836" => array("type"=>"city", "pop"=>924, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>33.026886, "lon"=>-84.715435),
"/US/US-GA/gresham-park-3837" => array("type"=>"cdp", "pop"=>9000, "name_ru"=>"Грешам Парк", "name_en"=>"Gresham Park", "lang"=>"en", "lat"=>33.703250, "lon"=>-84.314443),
"/US/US-GA/griffin-3838" => array("type"=>"city", "pop"=>23719, "name_ru"=>"Гриффин", "name_en"=>"Griffin", "lang"=>"en", "lat"=>33.241091, "lon"=>-84.272893),
"/US/US-GA/grovetown-3839" => array("type"=>"city", "pop"=>9224, "name_ru"=>"Гровтаун", "name_en"=>"Grovetown", "lang"=>"en", "lat"=>33.449392, "lon"=>-82.201828),
"/US/US-GA/gumbranch-3840" => array("type"=>"city", "pop"=>253, "name_ru"=>"Гамбранч", "name_en"=>"Gumbranch", "lang"=>"en", "lat"=>31.836753, "lon"=>-81.684743),
"/US/US-GA/gumlog-3841" => array("type"=>"cdp", "pop"=>1470, "name_ru"=>"Гамлог", "name_en"=>"Gumlog", "lang"=>"en", "lat"=>34.491764, "lon"=>-83.096661),
"/US/US-GA/guyton-3842" => array("type"=>"city", "pop"=>1926, "name_ru"=>"Гайтон", "name_en"=>"Guyton", "lang"=>"en", "lat"=>32.336145, "lon"=>-81.393931),
"/US/US-GA/hagan-3843" => array("type"=>"city", "pop"=>1055, "name_ru"=>"Хаган", "name_en"=>"Hagan", "lang"=>"en", "lat"=>32.154246, "lon"=>-81.929947),
"/US/US-GA/hahira-3844" => array("type"=>"city", "pop"=>2465, "name_ru"=>"Хаайра", "name_en"=>"Hahira", "lang"=>"en", "lat"=>30.990307, "lon"=>-83.372798),
"/US/US-GA/hamilton-3845" => array("type"=>"city", "pop"=>578, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>32.758852, "lon"=>-84.875568),
"/US/US-GA/hampton-3846" => array("type"=>"city", "pop"=>5295, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>33.384378, "lon"=>-84.285224),
"/US/US-GA/hapeville-3847" => array("type"=>"city", "pop"=>5978, "name_ru"=>"Хейпвилл", "name_en"=>"Hapeville", "lang"=>"en", "lat"=>33.661050, "lon"=>-84.409313),
"/US/US-GA/haralson-3848" => array("type"=>"town", "pop"=>159, "name_ru"=>"Харалсон", "name_en"=>"Haralson", "lang"=>"en", "lat"=>33.228937, "lon"=>-84.570273),
"/US/US-GA/harlem-3849" => array("type"=>"city", "pop"=>2057, "name_ru"=>"Гарлем", "name_en"=>"Harlem", "lang"=>"en", "lat"=>33.414036, "lon"=>-82.316978),
"/US/US-GA/harrison-3850" => array("type"=>"town", "pop"=>481, "name_ru"=>"Харрисон", "name_en"=>"Harrison", "lang"=>"en", "lat"=>32.825932, "lon"=>-82.725851),
"/US/US-GA/hartwell-3851" => array("type"=>"city", "pop"=>4298, "name_ru"=>"Хартвэлл", "name_en"=>"Hartwell", "lang"=>"en", "lat"=>34.353733, "lon"=>-82.932047),
"/US/US-GA/hawkinsville-3852" => array("type"=>"city", "pop"=>4253, "name_ru"=>"Хокинсвилл", "name_en"=>"Hawkinsville", "lang"=>"en", "lat"=>32.282556, "lon"=>-83.474150),
"/US/US-GA/hazlehurst-3853" => array("type"=>"city", "pop"=>3876, "name_ru"=>"Хейзлхерст", "name_en"=>"Hazlehurst", "lang"=>"en", "lat"=>31.865902, "lon"=>-82.599661),
"/US/US-GA/helen-3854" => array("type"=>"city", "pop"=>841, "name_ru"=>"Хэлен", "name_en"=>"Helen", "lang"=>"en", "lat"=>34.701995, "lon"=>-83.722248),
"/US/US-GA/helena-3855" => array("type"=>"town", "pop"=>2449, "name_ru"=>"Хелена", "name_en"=>"Helena", "lang"=>"en", "lat"=>32.079373, "lon"=>-82.910840),
"/US/US-GA/hephzibah-3856" => array("type"=>"city", "pop"=>4533, "name_ru"=>"Хефзиба", "name_en"=>"Hephzibah", "lang"=>"en", "lat"=>33.287275, "lon"=>-82.106050),
"/US/US-GA/hiawassee-3857" => array("type"=>"town", "pop"=>856, "name_ru"=>"Хайавасси", "name_en"=>"Hiawassee", "lang"=>"en", "lat"=>34.948861, "lon"=>-83.754261),
"/US/US-GA/higgston-3858" => array("type"=>"town", "pop"=>326, "name_ru"=>"Хайгстон", "name_en"=>"Higgston", "lang"=>"en", "lat"=>32.216791, "lon"=>-82.467177),
"/US/US-GA/hilltonia-3859" => array("type"=>"city", "pop"=>402, "name_ru"=>"Хилтония", "name_en"=>"Hilltonia", "lang"=>"en", "lat"=>32.883716, "lon"=>-81.661373),
"/US/US-GA/hinesville-3860" => array("type"=>"city", "pop"=>30152, "name_ru"=>"Хинесвилл", "name_en"=>"Hinesville", "lang"=>"en", "lat"=>31.832157, "lon"=>-81.598926),
"/US/US-GA/hiram-3861" => array("type"=>"city", "pop"=>2056, "name_ru"=>"Хирам", "name_en"=>"Hiram", "lang"=>"en", "lat"=>33.865471, "lon"=>-84.774656),
"/US/US-GA/hoboken-3862" => array("type"=>"city", "pop"=>521, "name_ru"=>"Хобокен", "name_en"=>"Hoboken", "lang"=>"en", "lat"=>31.181409, "lon"=>-82.131695),
"/US/US-GA/hogansville-3863" => array("type"=>"city", "pop"=>2912, "name_ru"=>"Хогансвилл", "name_en"=>"Hogansville", "lang"=>"en", "lat"=>33.168578, "lon"=>-84.902488),
"/US/US-GA/holly-springs-3864" => array("type"=>"city", "pop"=>8818, "name_ru"=>"Хоули Спрингс", "name_en"=>"Holly Springs", "lang"=>"en", "lat"=>34.172205, "lon"=>-84.497860),
"/US/US-GA/homeland-3865" => array("type"=>"city", "pop"=>792, "name_ru"=>"Хоумленд", "name_en"=>"Homeland", "lang"=>"en", "lat"=>30.854999, "lon"=>-82.024294),
"/US/US-GA/homer-3866" => array("type"=>"town", "pop"=>1092, "name_ru"=>"Гомер", "name_en"=>"Homer", "lang"=>"en", "lat"=>34.333825, "lon"=>-83.499673),
"/US/US-GA/homerville-3867" => array("type"=>"city", "pop"=>2842, "name_ru"=>"Гомерсвилл", "name_en"=>"Homerville", "lang"=>"en", "lat"=>31.036132, "lon"=>-82.753619),
"/US/US-GA/hoschton-3868" => array("type"=>"city", "pop"=>1612, "name_ru"=>"Хоштон", "name_en"=>"Hoschton", "lang"=>"en", "lat"=>34.092475, "lon"=>-83.761183),
"/US/US-GA/hull-3869" => array("type"=>"city", "pop"=>161, "name_ru"=>"Халл", "name_en"=>"Hull", "lang"=>"en", "lat"=>34.014750, "lon"=>-83.293489),
"/US/US-GA/ideal-3870" => array("type"=>"city", "pop"=>480, "name_ru"=>"Айдиал", "name_en"=>"Ideal", "lang"=>"en", "lat"=>32.372765, "lon"=>-84.188911),
"/US/US-GA/ila-3871" => array("type"=>"city", "pop"=>332, "name_ru"=>"Айла", "name_en"=>"Ila", "lang"=>"en", "lat"=>34.172278, "lon"=>-83.293543),
"/US/US-GA/indian-springs-3872" => array("type"=>"cdp", "pop"=>1273, "name_ru"=>"Индиан Спрингс", "name_en"=>"Indian Springs", "lang"=>"en", "lat"=>34.960800, "lon"=>-85.160017),
"/US/US-GA/iron-city-3873" => array("type"=>"town", "pop"=>302, "name_ru"=>"Айрон Сити", "name_en"=>"Iron City", "lang"=>"en", "lat"=>31.012955, "lon"=>-84.813760),
"/US/US-GA/irondale-3874" => array("type"=>"cdp", "pop"=>3352, "name_ru"=>"Айрондейл", "name_en"=>"Irondale", "lang"=>"en", "lat"=>33.480433, "lon"=>-84.358907),
"/US/US-GA/irwinton-3875" => array("type"=>"town", "pop"=>583, "name_ru"=>"Ирвингтон", "name_en"=>"Irwinton", "lang"=>"en", "lat"=>32.811413, "lon"=>-83.174671),
"/US/US-GA/isle-of-hope-dutch-island-3876" => array("type"=>"cdp", "pop"=>2637, "name_ru"=>"Айл оф Хоуп-Датч Айленд", "name_en"=>"Isle Of Hope-Dutch Island", "lang"=>"en", "lat"=>31.985500, "lon"=>-81.052864),
"/US/US-GA/ivey-3877" => array("type"=>"town", "pop"=>1068, "name_ru"=>"Айви", "name_en"=>"Ivey", "lang"=>"en", "lat"=>32.907855, "lon"=>-83.298695),
"/US/US-GA/jackson-3878" => array("type"=>"city", "pop"=>4496, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>33.292121, "lon"=>-83.962171),
"/US/US-GA/jacksonville-3879" => array("type"=>"town", "pop"=>116, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>31.815565, "lon"=>-82.978677),
"/US/US-GA/jakin-3880" => array("type"=>"town", "pop"=>150, "name_ru"=>"Джэкин", "name_en"=>"Jakin", "lang"=>"en", "lat"=>31.090392, "lon"=>-84.983252),
"/US/US-GA/jasper-3881" => array("type"=>"city", "pop"=>3166, "name_ru"=>"Джаспер", "name_en"=>"Jasper", "lang"=>"en", "lat"=>34.469781, "lon"=>-84.428287),
"/US/US-GA/jefferson-3882" => array("type"=>"city", "pop"=>7911, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>34.133390, "lon"=>-83.599698),
"/US/US-GA/jeffersonville-3883" => array("type"=>"city", "pop"=>1161, "name_ru"=>"Джефферсонвилл", "name_en"=>"Jeffersonville", "lang"=>"en", "lat"=>32.686935, "lon"=>-83.344229),
"/US/US-GA/jenkinsburg-3884" => array("type"=>"town", "pop"=>235, "name_ru"=>"Дженкинсбург", "name_en"=>"Jenkinsburg", "lang"=>"en", "lat"=>33.324929, "lon"=>-84.034314),
"/US/US-GA/jersey-3885" => array("type"=>"town", "pop"=>175, "name_ru"=>"Джерси", "name_en"=>"Jersey", "lang"=>"en", "lat"=>33.717487, "lon"=>-83.800340),
"/US/US-GA/jesup-3886" => array("type"=>"city", "pop"=>10459, "name_ru"=>"Джесап", "name_en"=>"Jesup", "lang"=>"en", "lat"=>31.598559, "lon"=>-81.889358),
"/US/US-GA/jonesboro-3887" => array("type"=>"city", "pop"=>4131, "name_ru"=>"Джонсборо", "name_en"=>"Jonesboro", "lang"=>"en", "lat"=>33.522050, "lon"=>-84.352924),
"/US/US-GA/junction-city-3888" => array("type"=>"town", "pop"=>170, "name_ru"=>"Джанкшен Сити", "name_en"=>"Junction City", "lang"=>"en", "lat"=>32.602943, "lon"=>-84.458270),
"/US/US-GA/kennesaw-3889" => array("type"=>"city", "pop"=>31628, "name_ru"=>"Кеннесо", "name_en"=>"Kennesaw", "lang"=>"en", "lat"=>34.023700, "lon"=>-84.616327),
"/US/US-GA/keysville-3890" => array("type"=>"town", "pop"=>237, "name_ru"=>"Кисвилл", "name_en"=>"Keysville", "lang"=>"en", "lat"=>33.236227, "lon"=>-82.229511),
"/US/US-GA/kings-bay-base-3891" => array("type"=>"cdp", "pop"=>3463, "name_ru"=>"Кингс Бэй Бейз", "name_en"=>"Kings Bay Base", "lang"=>"en", "lat"=>30.797829, "lon"=>-81.564416),
"/US/US-GA/kingsland-3892" => array("type"=>"city", "pop"=>13659, "name_ru"=>"Кингсленд", "name_en"=>"Kingsland", "lang"=>"en", "lat"=>30.791268, "lon"=>-81.663138),
"/US/US-GA/kingston-3893" => array("type"=>"city", "pop"=>669, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>34.235717, "lon"=>-84.946095),
"/US/US-GA/kite-3894" => array("type"=>"town", "pop"=>233, "name_ru"=>"Кайт", "name_en"=>"Kite", "lang"=>"en", "lat"=>32.691311, "lon"=>-82.515526),
"/US/US-GA/la-fayette-3895" => array("type"=>"city", "pop"=>6313, "name_ru"=>"Лафайетт", "name_en"=>"La Fayette", "lang"=>"en", "lat"=>34.707560, "lon"=>-85.281050),
"/US/US-GA/la-grange-3896" => array("type"=>"city", "pop"=>25597, "name_ru"=>"Ла Гранж", "name_en"=>"La Grange", "lang"=>"en", "lat"=>33.041430, "lon"=>-85.034615),
"/US/US-GA/lake-city-3897" => array("type"=>"city", "pop"=>2682, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>33.608000, "lon"=>-84.340557),
"/US/US-GA/lake-park-3899" => array("type"=>"city", "pop"=>609, "name_ru"=>"Лэйк Парк", "name_en"=>"Lake Park", "lang"=>"en", "lat"=>30.684606, "lon"=>-83.187768),
"/US/US-GA/lakeland-3898" => array("type"=>"city", "pop"=>2631, "name_ru"=>"Лэйклэнд", "name_en"=>"Lakeland", "lang"=>"en", "lat"=>31.040682, "lon"=>-83.074919),
"/US/US-GA/lakeview-3900" => array("type"=>"cdp", "pop"=>5237, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>34.977012, "lon"=>-85.253655),
"/US/US-GA/lakeview-estates-3901" => array("type"=>"cdp", "pop"=>1477, "name_ru"=>"Лейквью Истейтс", "name_en"=>"Lakeview Estates", "lang"=>"en", "lat"=>33.705550, "lon"=>-84.040003),
"/US/US-GA/lavonia-3902" => array("type"=>"city", "pop"=>2077, "name_ru"=>"Лавония", "name_en"=>"Lavonia", "lang"=>"en", "lat"=>34.435953, "lon"=>-83.106397),
"/US/US-GA/lawrenceville-3903" => array("type"=>"city", "pop"=>29258, "name_ru"=>"Лоренсвилл", "name_en"=>"Lawrenceville", "lang"=>"en", "lat"=>33.949512, "lon"=>-83.992180),
"/US/US-GA/leary-3904" => array("type"=>"city", "pop"=>624, "name_ru"=>"Лири", "name_en"=>"Leary", "lang"=>"en", "lat"=>31.484927, "lon"=>-84.512826),
"/US/US-GA/leesburg-3905" => array("type"=>"city", "pop"=>2920, "name_ru"=>"Лисберг", "name_en"=>"Leesburg", "lang"=>"en", "lat"=>31.733926, "lon"=>-84.170234),
"/US/US-GA/lenox-3906" => array("type"=>"town", "pop"=>858, "name_ru"=>"Ленокс", "name_en"=>"Lenox", "lang"=>"en", "lat"=>31.269877, "lon"=>-83.466357),
"/US/US-GA/leslie-3907" => array("type"=>"village", "pop"=>434, "name_ru"=>"Лесли", "name_en"=>"Leslie", "lang"=>"en", "lat"=>31.954731, "lon"=>-84.087001),
"/US/US-GA/lexington-3908" => array("type"=>"city", "pop"=>231, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>33.870233, "lon"=>-83.111044),
"/US/US-GA/lilburn-3909" => array("type"=>"city", "pop"=>11599, "name_ru"=>"Лилберн", "name_en"=>"Lilburn", "lang"=>"en", "lat"=>33.888233, "lon"=>-84.136866),
"/US/US-GA/lilly-3910" => array("type"=>"town", "pop"=>217, "name_ru"=>"Лилли", "name_en"=>"Lilly", "lang"=>"en", "lat"=>32.145885, "lon"=>-83.877770),
"/US/US-GA/lincolnton-3911" => array("type"=>"city", "pop"=>1534, "name_ru"=>"Линкольнтон", "name_en"=>"Lincolnton", "lang"=>"en", "lat"=>33.791802, "lon"=>-82.479314),
"/US/US-GA/lindale-3912" => array("type"=>"cdp", "pop"=>4187, "name_ru"=>"Линдейл", "name_en"=>"Lindale", "lang"=>"en", "lat"=>34.187550, "lon"=>-85.180659),
"/US/US-GA/linwood-3913" => array("type"=>"town", "pop"=>342, "name_ru"=>"Линвуд", "name_en"=>"Linwood", "lang"=>"en", "lat"=>34.711295, "lon"=>-85.290807),
"/US/US-GA/lithia-springs-3914" => array("type"=>"cdp", "pop"=>11403, "name_ru"=>"Лития Спрингс", "name_en"=>"Lithia Springs", "lang"=>"en", "lat"=>33.779000, "lon"=>-84.644197),
"/US/US-GA/lithonia-3915" => array("type"=>"city", "pop"=>2367, "name_ru"=>"Литония", "name_en"=>"Lithonia", "lang"=>"en", "lat"=>33.712550, "lon"=>-84.105981),
"/US/US-GA/locust-grove-3916" => array("type"=>"city", "pop"=>4811, "name_ru"=>"Локуст Гров", "name_en"=>"Locust Grove", "lang"=>"en", "lat"=>33.345347, "lon"=>-84.105033),
"/US/US-GA/loganville-3917" => array("type"=>"city", "pop"=>10562, "name_ru"=>"Логанвилл", "name_en"=>"Loganville", "lang"=>"en", "lat"=>33.835900, "lon"=>-83.895132),
"/US/US-GA/lone-oak-3918" => array("type"=>"town", "pop"=>97, "name_ru"=>"Лоун Оук", "name_en"=>"Lone Oak", "lang"=>"en", "lat"=>33.172029, "lon"=>-84.820106),
"/US/US-GA/lookout-mountain-3919" => array("type"=>"city", "pop"=>1524, "name_ru"=>"Лукаут Маунтин", "name_en"=>"Lookout Mountain", "lang"=>"en", "lat"=>34.961480, "lon"=>-85.360354),
"/US/US-GA/louisville-3920" => array("type"=>"city", "pop"=>2644, "name_ru"=>"Луисвилл", "name_en"=>"Louisville", "lang"=>"en", "lat"=>32.995834, "lon"=>-82.400352),
"/US/US-GA/lovejoy-3921" => array("type"=>"city", "pop"=>2418, "name_ru"=>"Лавджой", "name_en"=>"Lovejoy", "lang"=>"en", "lat"=>33.439550, "lon"=>-84.316211),
"/US/US-GA/ludowici-3922" => array("type"=>"city", "pop"=>1555, "name_ru"=>"Лудоуиси", "name_en"=>"Ludowici", "lang"=>"en", "lat"=>31.710665, "lon"=>-81.746387),
"/US/US-GA/lula-3923" => array("type"=>"city", "pop"=>2431, "name_ru"=>"Лула", "name_en"=>"Lula", "lang"=>"en", "lat"=>34.393016, "lon"=>-83.663072),
"/US/US-GA/lumber-city-3924" => array("type"=>"city", "pop"=>1182, "name_ru"=>"Ламбер Сити", "name_en"=>"Lumber City", "lang"=>"en", "lat"=>31.932588, "lon"=>-82.682703),
"/US/US-GA/lumpkin-3925" => array("type"=>"city", "pop"=>1220, "name_ru"=>"Лампкин", "name_en"=>"Lumpkin", "lang"=>"en", "lat"=>32.049657, "lon"=>-84.799552),
"/US/US-GA/luthersville-3926" => array("type"=>"town", "pop"=>821, "name_ru"=>"Лютерсвилл", "name_en"=>"Luthersville", "lang"=>"en", "lat"=>33.209851, "lon"=>-84.744442),
"/US/US-GA/lyerly-3927" => array("type"=>"town", "pop"=>519, "name_ru"=>"Лайерли", "name_en"=>"Lyerly", "lang"=>"en", "lat"=>34.403344, "lon"=>-85.404193),
"/US/US-GA/lyons-3928" => array("type"=>"city", "pop"=>4515, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>32.203864, "lon"=>-82.321071),
"/US/US-GA/mableton-3929" => array("type"=>"cdp", "pop"=>25725, "name_ru"=>"Мейблтон", "name_en"=>"Mableton", "lang"=>"en", "lat"=>33.812867, "lon"=>-84.562193),
"/US/US-GA/macon-3933" => array("type"=>"city", "pop"=>92775, "name_ru"=>"Мейкон", "name_en"=>"Macon", "lang"=>"en", "lat"=>32.836750, "lon"=>-83.656839),
"/US/US-GA/madison-3935" => array("type"=>"city", "pop"=>3925, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>33.577291, "lon"=>-83.476804),
"/US/US-GA/manassas-3936" => array("type"=>"city", "pop"=>106, "name_ru"=>"Манассас", "name_en"=>"Manassas", "lang"=>"en", "lat"=>32.163900, "lon"=>-82.021496),
"/US/US-GA/manchester-3937" => array("type"=>"city", "pop"=>3794, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>32.854598, "lon"=>-84.616526),
"/US/US-GA/mansfield-3938" => array("type"=>"town", "pop"=>583, "name_ru"=>"Мансфилд", "name_en"=>"Mansfield", "lang"=>"en", "lat"=>33.517985, "lon"=>-83.735183),
"/US/US-GA/marietta-3939" => array("type"=>"city", "pop"=>67562, "name_ru"=>"Мариетта", "name_en"=>"Marietta", "lang"=>"en", "lat"=>33.950950, "lon"=>-84.541149),
"/US/US-GA/marshallville-3940" => array("type"=>"city", "pop"=>1239, "name_ru"=>"Маршалвилл", "name_en"=>"Marshallville", "lang"=>"en", "lat"=>32.457180, "lon"=>-83.941089),
"/US/US-GA/martin-3941" => array("type"=>"town", "pop"=>309, "name_ru"=>"Мартин", "name_en"=>"Martin", "lang"=>"en", "lat"=>34.485118, "lon"=>-83.186364),
"/US/US-GA/martinez-3942" => array("type"=>"cdp", "pop"=>33731, "name_ru"=>"Мартинез", "name_en"=>"Martinez", "lang"=>"en", "lat"=>33.519403, "lon"=>-82.088293),
"/US/US-GA/maxeys-3943" => array("type"=>"town", "pop"=>215, "name_ru"=>"Максис", "name_en"=>"Maxeys", "lang"=>"en", "lat"=>33.754821, "lon"=>-83.173896),
"/US/US-GA/maysville-3944" => array("type"=>"town", "pop"=>1663, "name_ru"=>"Мейсвилл", "name_en"=>"Maysville", "lang"=>"en", "lat"=>34.253379, "lon"=>-83.555685),
"/US/US-GA/mccaysville-3930" => array("type"=>"city", "pop"=>986, "name_ru"=>"Мак-Кейсвилл", "name_en"=>"McCaysville", "lang"=>"en", "lat"=>34.981454, "lon"=>-84.370371),
"/US/US-GA/mcdonough-3931" => array("type"=>"city", "pop"=>19768, "name_ru"=>"Мак-Доно", "name_en"=>"McDonough", "lang"=>"en", "lat"=>33.450557, "lon"=>-84.144024),
"/US/US-GA/mcintyre-3932" => array("type"=>"town", "pop"=>713, "name_ru"=>"Макинтайр", "name_en"=>"McIntyre", "lang"=>"en", "lat"=>32.846947, "lon"=>-83.194806),
"/US/US-GA/mcrae-3934" => array("type"=>"city", "pop"=>4378, "name_ru"=>"Мак-Рай", "name_en"=>"McRae", "lang"=>"en", "lat"=>32.061583, "lon"=>-82.901043),
"/US/US-GA/meansville-3945" => array("type"=>"city", "pop"=>187, "name_ru"=>"Минсвилл", "name_en"=>"Meansville", "lang"=>"en", "lat"=>33.049899, "lon"=>-84.309786),
"/US/US-GA/meigs-3946" => array("type"=>"city", "pop"=>1077, "name_ru"=>"Мейс", "name_en"=>"Meigs", "lang"=>"en", "lat"=>31.072097, "lon"=>-84.091920),
"/US/US-GA/menlo-3947" => array("type"=>"city", "pop"=>529, "name_ru"=>"Менло", "name_en"=>"Menlo", "lang"=>"en", "lat"=>34.483271, "lon"=>-85.476850),
"/US/US-GA/metter-3948" => array("type"=>"city", "pop"=>4345, "name_ru"=>"Меттер", "name_en"=>"Metter", "lang"=>"en", "lat"=>32.395965, "lon"=>-82.061187),
"/US/US-GA/midville-3949" => array("type"=>"city", "pop"=>452, "name_ru"=>"Мидвилл", "name_en"=>"Midville", "lang"=>"en", "lat"=>32.821097, "lon"=>-82.235292),
"/US/US-GA/midway-3950" => array("type"=>"city", "pop"=>1040, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>31.802455, "lon"=>-81.426810),
"/US/US-GA/midway-hardwick-3951" => array("type"=>"cdp", "pop"=>4910, "name_ru"=>"Мидуэй-Хардвик", "name_en"=>"Midway-Hardwick", "lang"=>"en", "lat"=>33.051310, "lon"=>-83.244723),
"/US/US-GA/milan-3952" => array("type"=>"town", "pop"=>1027, "name_ru"=>"Милан", "name_en"=>"Milan", "lang"=>"en", "lat"=>32.019515, "lon"=>-83.063598),
"/US/US-GA/milledgeville-3953" => array("type"=>"city", "pop"=>20703, "name_ru"=>"Миледжвилл", "name_en"=>"Milledgeville", "lang"=>"en", "lat"=>33.085197, "lon"=>-83.239824),
"/US/US-GA/millen-3954" => array("type"=>"city", "pop"=>3450, "name_ru"=>"Миллен", "name_en"=>"Millen", "lang"=>"en", "lat"=>32.806697, "lon"=>-81.942307),
"/US/US-GA/milner-3955" => array("type"=>"city", "pop"=>669, "name_ru"=>"Милнер", "name_en"=>"Milner", "lang"=>"en", "lat"=>33.114105, "lon"=>-84.195116),
"/US/US-GA/mineral-bluff-3956" => array("type"=>"town", "pop"=>153, "name_ru"=>"Минерал Блафф", "name_en"=>"Mineral Bluff", "lang"=>"en", "lat"=>34.915766, "lon"=>-84.276932),
"/US/US-GA/mitchell-3957" => array("type"=>"town", "pop"=>192, "name_ru"=>"Митчелл", "name_en"=>"Mitchell", "lang"=>"en", "lat"=>33.219016, "lon"=>-82.703144),
"/US/US-GA/molena-3958" => array("type"=>"city", "pop"=>473, "name_ru"=>"Молена", "name_en"=>"Molena", "lang"=>"en", "lat"=>33.010729, "lon"=>-84.500956),
"/US/US-GA/monroe-3959" => array("type"=>"city", "pop"=>13381, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>33.794085, "lon"=>-83.711929),
"/US/US-GA/montezuma-3960" => array("type"=>"city", "pop"=>3853, "name_ru"=>"Мотезума", "name_en"=>"Montezuma", "lang"=>"en", "lat"=>32.299595, "lon"=>-84.026777),
"/US/US-GA/montgomery-3961" => array("type"=>"cdp", "pop"=>8930, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>31.944184, "lon"=>-81.109080),
"/US/US-GA/monticello-3962" => array("type"=>"city", "pop"=>2611, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>33.305556, "lon"=>-83.687921),
"/US/US-GA/montrose-3963" => array("type"=>"town", "pop"=>165, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>32.560350, "lon"=>-83.154282),
"/US/US-GA/moody-afb-3964" => array("type"=>"cdp", "pop"=>1288, "name_ru"=>"Военно-воздушная база Муди", "name_en"=>"Moody AFB", "lang"=>"en", "lat"=>30.982032, "lon"=>-83.214289),
"/US/US-GA/moreland-3965" => array("type"=>"town", "pop"=>429, "name_ru"=>"Морленд", "name_en"=>"Moreland", "lang"=>"en", "lat"=>33.286185, "lon"=>-84.770368),
"/US/US-GA/morgan-3966" => array("type"=>"city", "pop"=>1739, "name_ru"=>"Морган", "name_en"=>"Morgan", "lang"=>"en", "lat"=>31.537444, "lon"=>-84.599372),
"/US/US-GA/morganton-3967" => array("type"=>"town", "pop"=>280, "name_ru"=>"Моргантон", "name_en"=>"Morganton", "lang"=>"en", "lat"=>34.874943, "lon"=>-84.243978),
"/US/US-GA/morrow-3968" => array("type"=>"city", "pop"=>5539, "name_ru"=>"Морроу", "name_en"=>"Morrow", "lang"=>"en", "lat"=>33.578367, "lon"=>-84.340214),
"/US/US-GA/morven-3969" => array("type"=>"city", "pop"=>622, "name_ru"=>"Морвен", "name_en"=>"Morven", "lang"=>"en", "lat"=>30.942610, "lon"=>-83.500653),
"/US/US-GA/moultrie-3970" => array("type"=>"city", "pop"=>15409, "name_ru"=>"Молтри", "name_en"=>"Moultrie", "lang"=>"en", "lat"=>31.166820, "lon"=>-83.770165),
"/US/US-GA/mount-airy-3974" => array("type"=>"town", "pop"=>1144, "name_ru"=>"Маунт Эйри", "name_en"=>"Mount Airy", "lang"=>"en", "lat"=>34.519449, "lon"=>-83.503306),
"/US/US-GA/mount-vernon-3975" => array("type"=>"city", "pop"=>2135, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>32.181222, "lon"=>-82.593903),
"/US/US-GA/mount-zion-3976" => array("type"=>"city", "pop"=>1576, "name_ru"=>"Маунт Зион", "name_en"=>"Mount Zion", "lang"=>"en", "lat"=>33.633894, "lon"=>-85.180538),
"/US/US-GA/mountain-city-3971" => array("type"=>"town", "pop"=>746, "name_ru"=>"Маунтин Сити", "name_en"=>"Mountain City", "lang"=>"en", "lat"=>34.918796, "lon"=>-83.386016),
"/US/US-GA/mountain-park-3972" => array("type"=>"city", "pop"=>566, "name_ru"=>"Маунтин Парк", "name_en"=>"Mountain Park", "lang"=>"en", "lat"=>34.082583, "lon"=>-84.413551),
"/US/US-GA/mountain-park-3973" => array("type"=>"cdp", "pop"=>11025, "name_ru"=>"Маунтин Парк", "name_en"=>"Mountain Park", "lang"=>"en", "lat"=>33.844050, "lon"=>-84.129480),
"/US/US-GA/nahunta-3977" => array("type"=>"city", "pop"=>985, "name_ru"=>"Наанта", "name_en"=>"Nahunta", "lang"=>"en", "lat"=>31.208751, "lon"=>-81.981205),
"/US/US-GA/nashville-3978" => array("type"=>"city", "pop"=>4865, "name_ru"=>"Нэшвилл", "name_en"=>"Nashville", "lang"=>"en", "lat"=>31.204438, "lon"=>-83.249245),
"/US/US-GA/naylor-3979" => array("type"=>"town", "pop"=>111, "name_ru"=>"Нейлор", "name_en"=>"Naylor", "lang"=>"en", "lat"=>30.906748, "lon"=>-83.078967),
"/US/US-GA/nelson-3980" => array("type"=>"city", "pop"=>928, "name_ru"=>"Нельсон", "name_en"=>"Nelson", "lang"=>"en", "lat"=>34.381470, "lon"=>-84.371380),
"/US/US-GA/newborn-3981" => array("type"=>"town", "pop"=>792, "name_ru"=>"Ньюборн", "name_en"=>"Newborn", "lang"=>"en", "lat"=>33.515986, "lon"=>-83.695624),
"/US/US-GA/newington-3982" => array("type"=>"town", "pop"=>307, "name_ru"=>"Ньюингтон", "name_en"=>"Newington", "lang"=>"en", "lat"=>32.588900, "lon"=>-81.506181),
"/US/US-GA/newnan-3983" => array("type"=>"city", "pop"=>30474, "name_ru"=>"Ньюнан", "name_en"=>"Newnan", "lang"=>"en", "lat"=>33.379132, "lon"=>-84.788443),
"/US/US-GA/newton-3984" => array("type"=>"city", "pop"=>799, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>31.316373, "lon"=>-84.336688),
"/US/US-GA/nicholls-3985" => array("type"=>"city", "pop"=>2745, "name_ru"=>"Николс", "name_en"=>"Nicholls", "lang"=>"en", "lat"=>31.518242, "lon"=>-82.635783),
"/US/US-GA/nicholson-3986" => array("type"=>"city", "pop"=>2060, "name_ru"=>"Николсон", "name_en"=>"Nicholson", "lang"=>"en", "lat"=>34.113822, "lon"=>-83.432738),
"/US/US-GA/norcross-3987" => array("type"=>"city", "pop"=>10737, "name_ru"=>"Норкросс", "name_en"=>"Norcross", "lang"=>"en", "lat"=>33.940900, "lon"=>-84.207574),
"/US/US-GA/norman-park-3988" => array("type"=>"city", "pop"=>865, "name_ru"=>"Норман Парк", "name_en"=>"Norman Park", "lang"=>"en", "lat"=>31.268879, "lon"=>-83.684748),
"/US/US-GA/north-atlanta-3989" => array("type"=>"cdp", "pop"=>27812, "name_ru"=>"Норт Атланта", "name_en"=>"North Atlanta", "lang"=>"en", "lat"=>33.862550, "lon"=>-84.333528),
"/US/US-GA/north-decatur-3990" => array("type"=>"cdp", "pop"=>13936, "name_ru"=>"Норт Декатур", "name_en"=>"North Decatur", "lang"=>"en", "lat"=>33.807150, "lon"=>-84.289577),
"/US/US-GA/north-druid-hills-3991" => array("type"=>"cdp", "pop"=>14170, "name_ru"=>"Норт Друид Хилс", "name_en"=>"North Druid Hills", "lang"=>"en", "lat"=>33.821100, "lon"=>-84.327990),
"/US/US-GA/north-high-shoals-3992" => array("type"=>"town", "pop"=>576, "name_ru"=>"Норт Хай Шолс", "name_en"=>"North High Shoals", "lang"=>"en", "lat"=>33.828366, "lon"=>-83.501520),
"/US/US-GA/norwood-3993" => array("type"=>"town", "pop"=>284, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>33.462616, "lon"=>-82.706490),
"/US/US-GA/nunez-3994" => array("type"=>"town", "pop"=>137, "name_ru"=>"Нанез", "name_en"=>"Nunez", "lang"=>"en", "lat"=>32.491713, "lon"=>-82.346698),
"/US/US-GA/oak-park-3995" => array("type"=>"town", "pop"=>382, "name_ru"=>"Оук Парк", "name_en"=>"Oak Park", "lang"=>"en", "lat"=>32.357866, "lon"=>-82.314715),
"/US/US-GA/oakwood-3996" => array("type"=>"town", "pop"=>4392, "name_ru"=>"Оуквуд", "name_en"=>"Oakwood", "lang"=>"en", "lat"=>34.232149, "lon"=>-83.880627),
"/US/US-GA/ochlocknee-3997" => array("type"=>"town", "pop"=>614, "name_ru"=>"Ошлокни", "name_en"=>"Ochlocknee", "lang"=>"en", "lat"=>30.975181, "lon"=>-84.050843),
"/US/US-GA/ocilla-3998" => array("type"=>"city", "pop"=>3144, "name_ru"=>"Оцилла", "name_en"=>"Ocilla", "lang"=>"en", "lat"=>31.595651, "lon"=>-83.249230),
"/US/US-GA/oconee-3999" => array("type"=>"town", "pop"=>290, "name_ru"=>"Окони", "name_en"=>"Oconee", "lang"=>"en", "lat"=>32.852169, "lon"=>-82.957380),
"/US/US-GA/odum-4000" => array("type"=>"town", "pop"=>464, "name_ru"=>"Одум", "name_en"=>"Odum", "lang"=>"en", "lat"=>31.665261, "lon"=>-82.026412),
"/US/US-GA/oglethorpe-4001" => array("type"=>"city", "pop"=>1113, "name_ru"=>"Оглторп", "name_en"=>"Oglethorpe", "lang"=>"en", "lat"=>32.293171, "lon"=>-84.062708),
"/US/US-GA/oliver-4002" => array("type"=>"city", "pop"=>229, "name_ru"=>"Оливер", "name_en"=>"Oliver", "lang"=>"en", "lat"=>32.520661, "lon"=>-81.532794),
"/US/US-GA/omaha-4003" => array("type"=>"city", "pop"=>116, "name_ru"=>"Омаха", "name_en"=>"Omaha", "lang"=>"en", "lat"=>32.147832, "lon"=>-85.012792),
"/US/US-GA/omega-4004" => array("type"=>"city", "pop"=>1382, "name_ru"=>"Омега", "name_en"=>"Omega", "lang"=>"en", "lat"=>31.338325, "lon"=>-83.595604),
"/US/US-GA/orchard-hill-4005" => array("type"=>"town", "pop"=>239, "name_ru"=>"Орчард Хилл", "name_en"=>"Orchard Hill", "lang"=>"en", "lat"=>33.186735, "lon"=>-84.211459),
"/US/US-GA/oxford-4006" => array("type"=>"town", "pop"=>2546, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>33.623067, "lon"=>-83.870216),
"/US/US-GA/palmetto-4007" => array("type"=>"city", "pop"=>5053, "name_ru"=>"Палметто", "name_en"=>"Palmetto", "lang"=>"en", "lat"=>33.526128, "lon"=>-84.661658),
"/US/US-GA/panthersville-4008" => array("type"=>"cdp", "pop"=>9874, "name_ru"=>"Пантерсвилл", "name_en"=>"Panthersville", "lang"=>"en", "lat"=>33.706700, "lon"=>-84.277873),
"/US/US-GA/parrott-4009" => array("type"=>"town", "pop"=>133, "name_ru"=>"Парротт", "name_en"=>"Parrott", "lang"=>"en", "lat"=>31.893878, "lon"=>-84.511156),
"/US/US-GA/patterson-4010" => array("type"=>"city", "pop"=>681, "name_ru"=>"Паттерсон", "name_en"=>"Patterson", "lang"=>"en", "lat"=>31.384202, "lon"=>-82.138036),
"/US/US-GA/pavo-4011" => array("type"=>"city", "pop"=>701, "name_ru"=>"Паво", "name_en"=>"Pavo", "lang"=>"en", "lat"=>30.960118, "lon"=>-83.737557),
"/US/US-GA/payne-4012" => array("type"=>"city", "pop"=>164, "name_ru"=>"Пейн", "name_en"=>"Payne", "lang"=>"en", "lat"=>32.850400, "lon"=>-83.678809),
"/US/US-GA/peachtree-city-4013" => array("type"=>"city", "pop"=>34788, "name_ru"=>"Пичтри Сити", "name_en"=>"Peachtree City", "lang"=>"en", "lat"=>33.392612, "lon"=>-84.568876),
"/US/US-GA/pearson-4014" => array("type"=>"city", "pop"=>1949, "name_ru"=>"Пирсон", "name_en"=>"Pearson", "lang"=>"en", "lat"=>31.297206, "lon"=>-82.853178),
"/US/US-GA/pelham-4015" => array("type"=>"city", "pop"=>3882, "name_ru"=>"Пелхэм", "name_en"=>"Pelham", "lang"=>"en", "lat"=>31.126462, "lon"=>-84.152117),
"/US/US-GA/pembroke-4016" => array("type"=>"city", "pop"=>2544, "name_ru"=>"Пемброук", "name_en"=>"Pembroke", "lang"=>"en", "lat"=>32.135611, "lon"=>-81.622240),
"/US/US-GA/pendergrass-4017" => array("type"=>"city", "pop"=>642, "name_ru"=>"Пендерграсс", "name_en"=>"Pendergrass", "lang"=>"en", "lat"=>34.162570, "lon"=>-83.679185),
"/US/US-GA/perry-4018" => array("type"=>"city", "pop"=>12990, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>32.464094, "lon"=>-83.738013),
"/US/US-GA/phillipsburg-4019" => array("type"=>"cdp", "pop"=>1044, "name_ru"=>"Филлипсбург", "name_en"=>"Phillipsburg", "lang"=>"en", "lat"=>31.439068, "lon"=>-83.520538),
"/US/US-GA/pine-lake-4021" => array("type"=>"city", "pop"=>714, "name_ru"=>"Пайн Лейк", "name_en"=>"Pine Lake", "lang"=>"en", "lat"=>33.790350, "lon"=>-84.205974),
"/US/US-GA/pine-mountain-4022" => array("type"=>"town", "pop"=>1304, "name_ru"=>"Пайн Маунтин", "name_en"=>"Pine Mountain", "lang"=>"en", "lat"=>32.864280, "lon"=>-84.853128),
"/US/US-GA/pinehurst-4020" => array("type"=>"city", "pop"=>384, "name_ru"=>"Пайнхерст", "name_en"=>"Pinehurst", "lang"=>"en", "lat"=>32.195014, "lon"=>-83.760193),
"/US/US-GA/pineview-4023" => array("type"=>"town", "pop"=>518, "name_ru"=>"Пайнвью", "name_en"=>"Pineview", "lang"=>"en", "lat"=>32.108315, "lon"=>-83.502153),
"/US/US-GA/pitts-4024" => array("type"=>"city", "pop"=>297, "name_ru"=>"Питтс", "name_en"=>"Pitts", "lang"=>"en", "lat"=>31.946093, "lon"=>-83.541709),
"/US/US-GA/plains-4025" => array("type"=>"city", "pop"=>606, "name_ru"=>"Плейнс", "name_en"=>"Plains", "lang"=>"en", "lat"=>32.033088, "lon"=>-84.394659),
"/US/US-GA/plainville-4026" => array("type"=>"city", "pop"=>272, "name_ru"=>"Плейнвилл", "name_en"=>"Plainville", "lang"=>"en", "lat"=>34.404087, "lon"=>-85.036379),
"/US/US-GA/pooler-4027" => array("type"=>"city", "pop"=>14875, "name_ru"=>"Пулер", "name_en"=>"Pooler", "lang"=>"en", "lat"=>32.105276, "lon"=>-81.257493),
"/US/US-GA/port-wentworth-4030" => array("type"=>"city", "pop"=>4560, "name_ru"=>"Порт Уэнтворт", "name_en"=>"Port Wentworth", "lang"=>"en", "lat"=>32.192850, "lon"=>-81.194851),
"/US/US-GA/portal-4028" => array("type"=>"town", "pop"=>608, "name_ru"=>"Портал", "name_en"=>"Portal", "lang"=>"en", "lat"=>32.536270, "lon"=>-81.930583),
"/US/US-GA/porterdale-4029" => array("type"=>"town", "pop"=>1850, "name_ru"=>"Портердейл", "name_en"=>"Porterdale", "lang"=>"en", "lat"=>33.571880, "lon"=>-83.897036),
"/US/US-GA/poulan-4031" => array("type"=>"city", "pop"=>884, "name_ru"=>"Поулан", "name_en"=>"Poulan", "lang"=>"en", "lat"=>31.513077, "lon"=>-83.789903),
"/US/US-GA/powder-springs-4032" => array("type"=>"city", "pop"=>15614, "name_ru"=>"Паудер Спрингс", "name_en"=>"Powder Springs", "lang"=>"en", "lat"=>33.865333, "lon"=>-84.682607),
"/US/US-GA/preston-4033" => array("type"=>"city", "pop"=>418, "name_ru"=>"Престон", "name_en"=>"Preston", "lang"=>"en", "lat"=>32.060133, "lon"=>-84.540204),
"/US/US-GA/pulaski-4034" => array("type"=>"town", "pop"=>292, "name_ru"=>"Пуласки", "name_en"=>"Pulaski", "lang"=>"en", "lat"=>32.390804, "lon"=>-81.956328),
"/US/US-GA/putney-4035" => array("type"=>"cdp", "pop"=>3108, "name_ru"=>"Путни", "name_en"=>"Putney", "lang"=>"en", "lat"=>31.474750, "lon"=>-84.111384),
"/US/US-GA/quitman-4036" => array("type"=>"city", "pop"=>4621, "name_ru"=>"Квитман", "name_en"=>"Quitman", "lang"=>"en", "lat"=>30.785283, "lon"=>-83.560309),
"/US/US-GA/ranger-4037" => array("type"=>"town", "pop"=>93, "name_ru"=>"Рейнджер", "name_en"=>"Ranger", "lang"=>"en", "lat"=>34.500087, "lon"=>-84.711458),
"/US/US-GA/raoul-4038" => array("type"=>"cdp", "pop"=>1666, "name_ru"=>"Паул", "name_en"=>"Raoul", "lang"=>"en", "lat"=>34.457453, "lon"=>-83.601624),
"/US/US-GA/ray-city-4039" => array("type"=>"city", "pop"=>797, "name_ru"=>"Рей Сити", "name_en"=>"Ray City", "lang"=>"en", "lat"=>31.074835, "lon"=>-83.196746),
"/US/US-GA/rayle-4040" => array("type"=>"town", "pop"=>133, "name_ru"=>"Рейл", "name_en"=>"Rayle", "lang"=>"en", "lat"=>33.790830, "lon"=>-82.908847),
"/US/US-GA/rebecca-4041" => array("type"=>"town", "pop"=>236, "name_ru"=>"Ребекка", "name_en"=>"Rebecca", "lang"=>"en", "lat"=>31.807103, "lon"=>-83.487024),
"/US/US-GA/redan-4042" => array("type"=>"cdp", "pop"=>24376, "name_ru"=>"Редан", "name_en"=>"Redan", "lang"=>"en", "lat"=>33.736233, "lon"=>-84.159647),
"/US/US-GA/reed-creek-4043" => array("type"=>"cdp", "pop"=>1854, "name_ru"=>"Рид Крик", "name_en"=>"Reed Creek", "lang"=>"en", "lat"=>34.436630, "lon"=>-82.915082),
"/US/US-GA/register-4044" => array("type"=>"town", "pop"=>166, "name_ru"=>"Реджистер", "name_en"=>"Register", "lang"=>"en", "lat"=>32.365604, "lon"=>-81.884109),
"/US/US-GA/reidsville-4045" => array("type"=>"city", "pop"=>2463, "name_ru"=>"Райдсвилл", "name_en"=>"Reidsville", "lang"=>"en", "lat"=>32.084991, "lon"=>-82.118180),
"/US/US-GA/remerton-4046" => array("type"=>"city", "pop"=>1027, "name_ru"=>"Ремертон", "name_en"=>"Remerton", "lang"=>"en", "lat"=>30.843840, "lon"=>-83.309393),
"/US/US-GA/rentz-4047" => array("type"=>"town", "pop"=>326, "name_ru"=>"Рентс", "name_en"=>"Rentz", "lang"=>"en", "lat"=>32.383894, "lon"=>-82.991766),
"/US/US-GA/resaca-4048" => array("type"=>"city", "pop"=>778, "name_ru"=>"Ресака", "name_en"=>"Resaca", "lang"=>"en", "lat"=>34.581895, "lon"=>-84.944454),
"/US/US-GA/rest-haven-4049" => array("type"=>"town", "pop"=>146, "name_ru"=>"Рест Хейвен", "name_en"=>"Rest Haven", "lang"=>"en", "lat"=>34.133083, "lon"=>-83.977941),
"/US/US-GA/reynolds-4050" => array("type"=>"town", "pop"=>1058, "name_ru"=>"Рейнолдс", "name_en"=>"Reynolds", "lang"=>"en", "lat"=>32.558964, "lon"=>-84.096991),
"/US/US-GA/rhine-4051" => array("type"=>"town", "pop"=>433, "name_ru"=>"Райн", "name_en"=>"Rhine", "lang"=>"en", "lat"=>31.989026, "lon"=>-83.200752),
"/US/US-GA/riceboro-4052" => array("type"=>"city", "pop"=>706, "name_ru"=>"Райсборо", "name_en"=>"Riceboro", "lang"=>"en", "lat"=>31.727295, "lon"=>-81.440841),
"/US/US-GA/richland-4053" => array("type"=>"city", "pop"=>1600, "name_ru"=>"Ричленд", "name_en"=>"Richland", "lang"=>"en", "lat"=>32.084535, "lon"=>-84.675820),
"/US/US-GA/richmond-hill-4054" => array("type"=>"city", "pop"=>10571, "name_ru"=>"Ричмонд Хилл", "name_en"=>"Richmond Hill", "lang"=>"en", "lat"=>31.939658, "lon"=>-81.306432),
"/US/US-GA/riddleville-4055" => array("type"=>"town", "pop"=>122, "name_ru"=>"Риддлвилл", "name_en"=>"Riddleville", "lang"=>"en", "lat"=>32.907765, "lon"=>-82.667478),
"/US/US-GA/rincon-4056" => array("type"=>"town", "pop"=>7905, "name_ru"=>"Ринкон", "name_en"=>"Rincon", "lang"=>"en", "lat"=>32.291780, "lon"=>-81.234128),
"/US/US-GA/ringgold-4057" => array("type"=>"city", "pop"=>2803, "name_ru"=>"Ринггоулд", "name_en"=>"Ringgold", "lang"=>"en", "lat"=>34.917759, "lon"=>-85.109688),
"/US/US-GA/riverdale-4058" => array("type"=>"city", "pop"=>15279, "name_ru"=>"Ривердейл", "name_en"=>"Riverdale", "lang"=>"en", "lat"=>33.566100, "lon"=>-84.408168),
"/US/US-GA/riverside-4059" => array("type"=>"town", "pop"=>60, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>31.176531, "lon"=>-83.806135),
"/US/US-GA/roberta-4060" => array("type"=>"city", "pop"=>753, "name_ru"=>"Роберта", "name_en"=>"Roberta", "lang"=>"en", "lat"=>32.719675, "lon"=>-84.010883),
"/US/US-GA/robins-afb-4061" => array("type"=>"cdp", "pop"=>3092, "name_ru"=>"Военно-воздушная база Робинс", "name_en"=>"Robins AFB", "lang"=>"en", "lat"=>32.607801, "lon"=>-83.586783),
"/US/US-GA/rochelle-4062" => array("type"=>"city", "pop"=>1372, "name_ru"=>"Рошелл", "name_en"=>"Rochelle", "lang"=>"en", "lat"=>31.948826, "lon"=>-83.457213),
"/US/US-GA/rockmart-4063" => array("type"=>"city", "pop"=>4544, "name_ru"=>"Рокмарт", "name_en"=>"Rockmart", "lang"=>"en", "lat"=>34.009943, "lon"=>-85.049654),
"/US/US-GA/rocky-ford-4064" => array("type"=>"town", "pop"=>174, "name_ru"=>"Роки Форд", "name_en"=>"Rocky Ford", "lang"=>"en", "lat"=>32.663359, "lon"=>-81.829014),
"/US/US-GA/rome-4065" => array("type"=>"city", "pop"=>36041, "name_ru"=>"Роум", "name_en"=>"Rome", "lang"=>"en", "lat"=>34.262667, "lon"=>-85.186672),
"/US/US-GA/roopville-4066" => array("type"=>"town", "pop"=>205, "name_ru"=>"Рупвилл", "name_en"=>"Roopville", "lang"=>"en", "lat"=>33.456796, "lon"=>-85.132284),
"/US/US-GA/rossville-4067" => array("type"=>"city", "pop"=>3403, "name_ru"=>"Россвилл", "name_en"=>"Rossville", "lang"=>"en", "lat"=>34.974229, "lon"=>-85.289145),
"/US/US-GA/roswell-4068" => array("type"=>"city", "pop"=>87657, "name_ru"=>"Росуэл", "name_en"=>"Roswell", "lang"=>"en", "lat"=>34.037420, "lon"=>-84.356010),
"/US/US-GA/royston-4069" => array("type"=>"city", "pop"=>2735, "name_ru"=>"Ройстон", "name_en"=>"Royston", "lang"=>"en", "lat"=>34.285458, "lon"=>-83.109375),
"/US/US-GA/russell-4070" => array("type"=>"city", "pop"=>871, "name_ru"=>"Расселл", "name_en"=>"Russell", "lang"=>"en", "lat"=>33.975255, "lon"=>-83.695787),
"/US/US-GA/rutledge-4071" => array("type"=>"city", "pop"=>771, "name_ru"=>"Рутледж", "name_en"=>"Rutledge", "lang"=>"en", "lat"=>33.626539, "lon"=>-83.611620),
"/US/US-GA/sale-city-4074" => array("type"=>"town", "pop"=>317, "name_ru"=>"Сейл Сити", "name_en"=>"Sale City", "lang"=>"en", "lat"=>31.262788, "lon"=>-84.022282),
"/US/US-GA/sandersville-4075" => array("type"=>"city", "pop"=>6219, "name_ru"=>"Сэндерсвилл", "name_en"=>"Sandersville", "lang"=>"en", "lat"=>32.983281, "lon"=>-82.812698),
"/US/US-GA/sandy-springs-4076" => array("type"=>"cdp", "pop"=>82674, "name_ru"=>"Сэнди Спрингс", "name_en"=>"Sandy Springs", "lang"=>"en", "lat"=>33.937300, "lon"=>-84.368699),
"/US/US-GA/santa-claus-4077" => array("type"=>"city", "pop"=>250, "name_ru"=>"Санта Клос", "name_en"=>"Santa Claus", "lang"=>"en", "lat"=>32.171657, "lon"=>-82.329876),
"/US/US-GA/sardis-4078" => array("type"=>"town", "pop"=>1172, "name_ru"=>"Сардис", "name_en"=>"Sardis", "lang"=>"en", "lat"=>32.972972, "lon"=>-81.760354),
"/US/US-GA/sasser-4079" => array("type"=>"town", "pop"=>347, "name_ru"=>"Сассер", "name_en"=>"Sasser", "lang"=>"en", "lat"=>31.719117, "lon"=>-84.347864),
"/US/US-GA/savannah-4080" => array("type"=>"city", "pop"=>132410, "name_ru"=>"Саванна", "name_en"=>"Savannah", "lang"=>"en", "lat"=>32.024357, "lon"=>-81.131648),
"/US/US-GA/scotland-4081" => array("type"=>"city", "pop"=>287, "name_ru"=>"Скотленд", "name_en"=>"Scotland", "lang"=>"en", "lat"=>32.047407, "lon"=>-82.818723),
"/US/US-GA/scottdale-4082" => array("type"=>"cdp", "pop"=>8636, "name_ru"=>"Скотдейл", "name_en"=>"Scottdale", "lang"=>"en", "lat"=>33.794451, "lon"=>-84.263182),
"/US/US-GA/screven-4083" => array("type"=>"city", "pop"=>786, "name_ru"=>"Скревен", "name_en"=>"Screven", "lang"=>"en", "lat"=>31.483843, "lon"=>-82.017072),
"/US/US-GA/senoia-4084" => array("type"=>"city", "pop"=>3523, "name_ru"=>"Сеноия", "name_en"=>"Senoia", "lang"=>"en", "lat"=>33.311565, "lon"=>-84.552530),
"/US/US-GA/shady-dale-4085" => array("type"=>"town", "pop"=>244, "name_ru"=>"Шейди Дейл", "name_en"=>"Shady Dale", "lang"=>"en", "lat"=>33.400364, "lon"=>-83.589092),
"/US/US-GA/shannon-4086" => array("type"=>"cdp", "pop"=>1703, "name_ru"=>"Шэннон", "name_en"=>"Shannon", "lang"=>"en", "lat"=>34.339889, "lon"=>-85.086331),
"/US/US-GA/sharon-4087" => array("type"=>"city", "pop"=>95, "name_ru"=>"Шэрон", "name_en"=>"Sharon", "lang"=>"en", "lat"=>33.559110, "lon"=>-82.795328),
"/US/US-GA/sharpsburg-4088" => array("type"=>"town", "pop"=>330, "name_ru"=>"Шарпсбург", "name_en"=>"Sharpsburg", "lang"=>"en", "lat"=>33.341349, "lon"=>-84.647406),
"/US/US-GA/shellman-4089" => array("type"=>"city", "pop"=>1058, "name_ru"=>"Шеллман", "name_en"=>"Shellman", "lang"=>"en", "lat"=>31.756292, "lon"=>-84.615721),
"/US/US-GA/shiloh-4090" => array("type"=>"city", "pop"=>433, "name_ru"=>"Шилох", "name_en"=>"Shiloh", "lang"=>"en", "lat"=>32.812422, "lon"=>-84.697844),
"/US/US-GA/siloam-4091" => array("type"=>"town", "pop"=>331, "name_ru"=>"Силоэм", "name_en"=>"Siloam", "lang"=>"en", "lat"=>33.536500, "lon"=>-83.082374),
"/US/US-GA/skidaway-island-4092" => array("type"=>"cdp", "pop"=>4495, "name_ru"=>"Скидэвей Айленд", "name_en"=>"Skidaway Island", "lang"=>"en", "lat"=>31.934618, "lon"=>-81.047188),
"/US/US-GA/sky-valley-4093" => array("type"=>"city", "pop"=>217, "name_ru"=>"Скай Валли", "name_en"=>"Sky Valley", "lang"=>"en", "lat"=>34.983280, "lon"=>-83.332171),
"/US/US-GA/smithville-4094" => array("type"=>"city", "pop"=>964, "name_ru"=>"Смитвилл", "name_en"=>"Smithville", "lang"=>"en", "lat"=>31.901906, "lon"=>-84.255426),
"/US/US-GA/smyrna-4095" => array("type"=>"city", "pop"=>49854, "name_ru"=>"Смирна", "name_en"=>"Smyrna", "lang"=>"en", "lat"=>33.867481, "lon"=>-84.516894),
"/US/US-GA/snellville-4096" => array("type"=>"city", "pop"=>20112, "name_ru"=>"Снеллвилл", "name_en"=>"Snellville", "lang"=>"en", "lat"=>33.858333, "lon"=>-84.008618),
"/US/US-GA/social-circle-4097" => array("type"=>"city", "pop"=>4767, "name_ru"=>"Сошиал Серкл", "name_en"=>"Social Circle", "lang"=>"en", "lat"=>33.658541, "lon"=>-83.718023),
"/US/US-GA/soperton-4098" => array("type"=>"city", "pop"=>3295, "name_ru"=>"Сопретон", "name_en"=>"Soperton", "lang"=>"en", "lat"=>32.376587, "lon"=>-82.592211),
"/US/US-GA/south-augusta-4099" => array("type"=>"cdp", "pop"=>55998, "name_ru"=>"Саут Огаста", "name_en"=>"South Augusta", "lang"=>"en", "lat"=>33.417768, "lon"=>-82.048288),
"/US/US-GA/sparks-4100" => array("type"=>"town", "pop"=>1877, "name_ru"=>"Спаркс", "name_en"=>"Sparks", "lang"=>"en", "lat"=>31.167589, "lon"=>-83.439952),
"/US/US-GA/sparta-4101" => array("type"=>"city", "pop"=>1265, "name_ru"=>"Спарта", "name_en"=>"Sparta", "lang"=>"en", "lat"=>33.277125, "lon"=>-82.971598),
"/US/US-GA/springfield-4102" => array("type"=>"city", "pop"=>2126, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>32.366629, "lon"=>-81.308642),
"/US/US-GA/st.-marys-4072" => array("type"=>"city", "pop"=>16763, "name_ru"=>"Сент Мэрис", "name_en"=>"St. Marys", "lang"=>"en", "lat"=>30.759024, "lon"=>-81.577098),
"/US/US-GA/st.-simons-4073" => array("type"=>"cdp", "pop"=>12026, "name_ru"=>"Сент Саймонс", "name_en"=>"St. Simons", "lang"=>"en", "lat"=>31.176779, "lon"=>-81.384458),
"/US/US-GA/stapleton-4103" => array("type"=>"city", "pop"=>315, "name_ru"=>"Стейпелтон", "name_en"=>"Stapleton", "lang"=>"en", "lat"=>33.216202, "lon"=>-82.467013),
"/US/US-GA/statesboro-4104" => array("type"=>"city", "pop"=>27158, "name_ru"=>"Стейтсборо", "name_en"=>"Statesboro", "lang"=>"en", "lat"=>32.442438, "lon"=>-81.779167),
"/US/US-GA/statham-4105" => array("type"=>"city", "pop"=>2922, "name_ru"=>"Стэтхэм", "name_en"=>"Statham", "lang"=>"en", "lat"=>33.965384, "lon"=>-83.596823),
"/US/US-GA/stillmore-4106" => array("type"=>"town", "pop"=>763, "name_ru"=>"Стилмор", "name_en"=>"Stillmore", "lang"=>"en", "lat"=>32.441006, "lon"=>-82.214792),
"/US/US-GA/stockbridge-4107" => array("type"=>"city", "pop"=>14360, "name_ru"=>"Стокбридж", "name_en"=>"Stockbridge", "lang"=>"en", "lat"=>33.545800, "lon"=>-84.242949),
"/US/US-GA/stone-mountain-4108" => array("type"=>"city", "pop"=>7650, "name_ru"=>"Стоун Маунтин", "name_en"=>"Stone Mountain", "lang"=>"en", "lat"=>33.802833, "lon"=>-84.171666),
"/US/US-GA/sugar-hill-4109" => array("type"=>"city", "pop"=>17204, "name_ru"=>"Шугар Хилл", "name_en"=>"Sugar Hill", "lang"=>"en", "lat"=>34.099271, "lon"=>-84.040001),
"/US/US-GA/summertown-4110" => array("type"=>"town", "pop"=>146, "name_ru"=>"Саммертаун", "name_en"=>"Summertown", "lang"=>"en", "lat"=>32.746374, "lon"=>-82.276336),
"/US/US-GA/summerville-4111" => array("type"=>"city", "pop"=>5010, "name_ru"=>"Саммервилл", "name_en"=>"Summerville", "lang"=>"en", "lat"=>34.478458, "lon"=>-85.348997),
"/US/US-GA/sumner-4112" => array("type"=>"town", "pop"=>296, "name_ru"=>"Саммер", "name_en"=>"Sumner", "lang"=>"en", "lat"=>31.512160, "lon"=>-83.739991),
"/US/US-GA/sunny-side-4113" => array("type"=>"city", "pop"=>133, "name_ru"=>"Санни Сайд", "name_en"=>"Sunny Side", "lang"=>"en", "lat"=>33.341611, "lon"=>-84.290331),
"/US/US-GA/sunnyside-4114" => array("type"=>"cdp", "pop"=>1506, "name_ru"=>"Саннисайд", "name_en"=>"Sunnyside", "lang"=>"en", "lat"=>31.239229, "lon"=>-82.342115),
"/US/US-GA/surrency-4115" => array("type"=>"town", "pop"=>248, "name_ru"=>"Серренси", "name_en"=>"Surrency", "lang"=>"en", "lat"=>31.721782, "lon"=>-82.198275),
"/US/US-GA/suwanee-4116" => array("type"=>"city", "pop"=>16277, "name_ru"=>"Савани", "name_en"=>"Suwanee", "lang"=>"en", "lat"=>34.051153, "lon"=>-84.070734),
"/US/US-GA/swainsboro-4117" => array("type"=>"city", "pop"=>7667, "name_ru"=>"Свайнсборо", "name_en"=>"Swainsboro", "lang"=>"en", "lat"=>32.590121, "lon"=>-82.336143),
"/US/US-GA/sycamore-4118" => array("type"=>"city", "pop"=>526, "name_ru"=>"Сикамор", "name_en"=>"Sycamore", "lang"=>"en", "lat"=>31.670240, "lon"=>-83.635072),
"/US/US-GA/sylvania-4119" => array("type"=>"city", "pop"=>2511, "name_ru"=>"Сильвания", "name_en"=>"Sylvania", "lang"=>"en", "lat"=>32.750995, "lon"=>-81.637950),
"/US/US-GA/sylvester-4120" => array("type"=>"city", "pop"=>5833, "name_ru"=>"Силвестер", "name_en"=>"Sylvester", "lang"=>"en", "lat"=>31.528556, "lon"=>-83.836480),
"/US/US-GA/talbotton-4121" => array("type"=>"city", "pop"=>970, "name_ru"=>"Тэлботтон", "name_en"=>"Talbotton", "lang"=>"en", "lat"=>32.677289, "lon"=>-84.540867),
"/US/US-GA/talking-rock-4122" => array("type"=>"town", "pop"=>105, "name_ru"=>"Токинг Рок", "name_en"=>"Talking Rock", "lang"=>"en", "lat"=>34.509468, "lon"=>-84.505248),
"/US/US-GA/tallapoosa-4123" => array("type"=>"city", "pop"=>3123, "name_ru"=>"Таллапуза", "name_en"=>"Tallapoosa", "lang"=>"en", "lat"=>33.752476, "lon"=>-85.294907),
"/US/US-GA/tallulah-falls-4124" => array("type"=>"town", "pop"=>161, "name_ru"=>"Таллула Фолс", "name_en"=>"Tallulah Falls", "lang"=>"en", "lat"=>34.723946, "lon"=>-83.382488),
"/US/US-GA/talmo-4125" => array("type"=>"town", "pop"=>605, "name_ru"=>"Талмо", "name_en"=>"Talmo", "lang"=>"en", "lat"=>34.183945, "lon"=>-83.717964),
"/US/US-GA/tarrytown-4126" => array("type"=>"village", "pop"=>102, "name_ru"=>"Тэрритаун", "name_en"=>"Tarrytown", "lang"=>"en", "lat"=>32.318890, "lon"=>-82.559706),
"/US/US-GA/taylorsville-4127" => array("type"=>"town", "pop"=>226, "name_ru"=>"Тейлорсвилл", "name_en"=>"Taylorsville", "lang"=>"en", "lat"=>34.085785, "lon"=>-84.988191),
"/US/US-GA/temple-4128" => array("type"=>"city", "pop"=>4490, "name_ru"=>"Тэмпл", "name_en"=>"Temple", "lang"=>"en", "lat"=>33.734561, "lon"=>-85.031482),
"/US/US-GA/tennille-4129" => array("type"=>"city", "pop"=>1466, "name_ru"=>"Теннил", "name_en"=>"Tennille", "lang"=>"en", "lat"=>32.936199, "lon"=>-82.811280),
"/US/US-GA/the-rock-4130" => array("type"=>"town", "pop"=>88, "name_ru"=>"Зе Рок", "name_en"=>"The Rock", "lang"=>"en", "lat"=>32.963747, "lon"=>-84.242125),
"/US/US-GA/thomaston-4131" => array("type"=>"city", "pop"=>9191, "name_ru"=>"Томастон", "name_en"=>"Thomaston", "lang"=>"en", "lat"=>32.891644, "lon"=>-84.326983),
"/US/US-GA/thomasville-4132" => array("type"=>"city", "pop"=>19286, "name_ru"=>"Томасвилл", "name_en"=>"Thomasville", "lang"=>"en", "lat"=>30.839114, "lon"=>-83.978856),
"/US/US-GA/thomson-4133" => array("type"=>"city", "pop"=>6848, "name_ru"=>"Томсон", "name_en"=>"Thomson", "lang"=>"en", "lat"=>33.466098, "lon"=>-82.500003),
"/US/US-GA/thunderbolt-4134" => array("type"=>"town", "pop"=>2616, "name_ru"=>"Тандерболт", "name_en"=>"Thunderbolt", "lang"=>"en", "lat"=>32.034080, "lon"=>-81.049995),
"/US/US-GA/tifton-4135" => array("type"=>"city", "pop"=>17284, "name_ru"=>"Тифтон", "name_en"=>"Tifton", "lang"=>"en", "lat"=>31.465356, "lon"=>-83.510610),
"/US/US-GA/tiger-4136" => array("type"=>"town", "pop"=>315, "name_ru"=>"Тайгер", "name_en"=>"Tiger", "lang"=>"en", "lat"=>34.845465, "lon"=>-83.433403),
"/US/US-GA/tignall-4137" => array("type"=>"town", "pop"=>615, "name_ru"=>"Тигнолл", "name_en"=>"Tignall", "lang"=>"en", "lat"=>33.866861, "lon"=>-82.742424),
"/US/US-GA/toccoa-4138" => array("type"=>"city", "pop"=>9150, "name_ru"=>"Токкоа", "name_en"=>"Toccoa", "lang"=>"en", "lat"=>34.579903, "lon"=>-83.324713),
"/US/US-GA/toomsboro-4139" => array("type"=>"town", "pop"=>617, "name_ru"=>"Тумсборо", "name_en"=>"Toomsboro", "lang"=>"en", "lat"=>32.822141, "lon"=>-83.081051),
"/US/US-GA/trenton-4140" => array("type"=>"city", "pop"=>2375, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>34.874892, "lon"=>-85.509544),
"/US/US-GA/trion-4141" => array("type"=>"town", "pop"=>2058, "name_ru"=>"Трион", "name_en"=>"Trion", "lang"=>"en", "lat"=>34.549126, "lon"=>-85.310596),
"/US/US-GA/tucker-4142" => array("type"=>"cdp", "pop"=>25781, "name_ru"=>"Такер", "name_en"=>"Tucker", "lang"=>"en", "lat"=>33.854600, "lon"=>-84.223077),
"/US/US-GA/tunnel-hill-4143" => array("type"=>"city", "pop"=>1249, "name_ru"=>"Таннел Хилл", "name_en"=>"Tunnel Hill", "lang"=>"en", "lat"=>34.845840, "lon"=>-85.042990),
"/US/US-GA/turin-4144" => array("type"=>"town", "pop"=>409, "name_ru"=>"Турин", "name_en"=>"Turin", "lang"=>"en", "lat"=>33.326980, "lon"=>-84.636310),
"/US/US-GA/twin-city-4145" => array("type"=>"city", "pop"=>1894, "name_ru"=>"Твин Сити", "name_en"=>"Twin City", "lang"=>"en", "lat"=>32.584591, "lon"=>-82.157700),
"/US/US-GA/ty-ty-4148" => array("type"=>"town", "pop"=>773, "name_ru"=>"Тай Тай", "name_en"=>"Ty Ty", "lang"=>"en", "lat"=>31.472033, "lon"=>-83.650395),
"/US/US-GA/tybee-island-4146" => array("type"=>"city", "pop"=>3860, "name_ru"=>"Тайби Айленд", "name_en"=>"Tybee Island", "lang"=>"en", "lat"=>32.010127, "lon"=>-80.856661),
"/US/US-GA/tyrone-4147" => array("type"=>"town", "pop"=>6492, "name_ru"=>"Тайрон", "name_en"=>"Tyrone", "lang"=>"en", "lat"=>33.467950, "lon"=>-84.597016),
"/US/US-GA/unadilla-4149" => array("type"=>"city", "pop"=>2880, "name_ru"=>"Унадилла", "name_en"=>"Unadilla", "lang"=>"en", "lat"=>32.257447, "lon"=>-83.736115),
"/US/US-GA/union-city-4150" => array("type"=>"city", "pop"=>16961, "name_ru"=>"Юнион Сити", "name_en"=>"Union City", "lang"=>"en", "lat"=>33.577183, "lon"=>-84.544158),
"/US/US-GA/union-point-4151" => array("type"=>"city", "pop"=>1529, "name_ru"=>"Юнион Пойнт", "name_en"=>"Union Point", "lang"=>"en", "lat"=>33.617188, "lon"=>-83.075923),
"/US/US-GA/unionville-4152" => array("type"=>"cdp", "pop"=>2710, "name_ru"=>"Юнионвилл", "name_en"=>"Unionville", "lang"=>"en", "lat"=>31.436258, "lon"=>-83.509226),
"/US/US-GA/uvalda-4153" => array("type"=>"town", "pop"=>551, "name_ru"=>"Ювальда", "name_en"=>"Uvalda", "lang"=>"en", "lat"=>32.036479, "lon"=>-82.508582),
"/US/US-GA/valdosta-4154" => array("type"=>"city", "pop"=>48547, "name_ru"=>"Вальдоста", "name_en"=>"Valdosta", "lang"=>"en", "lat"=>30.849383, "lon"=>-83.278356),
"/US/US-GA/varnell-4155" => array("type"=>"city", "pop"=>1587, "name_ru"=>"Варнелл", "name_en"=>"Varnell", "lang"=>"en", "lat"=>34.901972, "lon"=>-84.970290),
"/US/US-GA/vernonburg-4156" => array("type"=>"town", "pop"=>133, "name_ru"=>"Вернонбург", "name_en"=>"Vernonburg", "lang"=>"en", "lat"=>31.966050, "lon"=>-81.120517),
"/US/US-GA/vidalia-4157" => array("type"=>"city", "pop"=>11355, "name_ru"=>"Видалия", "name_en"=>"Vidalia", "lang"=>"en", "lat"=>32.212266, "lon"=>-82.403361),
"/US/US-GA/vienna-4158" => array("type"=>"city", "pop"=>2851, "name_ru"=>"Вьенна", "name_en"=>"Vienna", "lang"=>"en", "lat"=>32.089773, "lon"=>-83.791566),
"/US/US-GA/villa-rica-4159" => array("type"=>"city", "pop"=>12838, "name_ru"=>"Вилла Рика", "name_en"=>"Villa Rica", "lang"=>"en", "lat"=>33.732925, "lon"=>-84.922869),
"/US/US-GA/vinings-4160" => array("type"=>"cdp", "pop"=>7417, "name_ru"=>"Винингс", "name_en"=>"Vinings", "lang"=>"en", "lat"=>33.861622, "lon"=>-84.467450),
"/US/US-GA/waco-4161" => array("type"=>"city", "pop"=>506, "name_ru"=>"Вако", "name_en"=>"Waco", "lang"=>"en", "lat"=>33.703049, "lon"=>-85.185066),
"/US/US-GA/wadley-4162" => array("type"=>"city", "pop"=>1940, "name_ru"=>"Уодли", "name_en"=>"Wadley", "lang"=>"en", "lat"=>32.864639, "lon"=>-82.402727),
"/US/US-GA/waleska-4163" => array("type"=>"city", "pop"=>889, "name_ru"=>"Уолска", "name_en"=>"Waleska", "lang"=>"en", "lat"=>34.315752, "lon"=>-84.550095),
"/US/US-GA/walnut-grove-4164" => array("type"=>"town", "pop"=>1261, "name_ru"=>"Уолнат Гров", "name_en"=>"Walnut Grove", "lang"=>"en", "lat"=>33.745110, "lon"=>-83.854309),
"/US/US-GA/walthourville-4165" => array("type"=>"city", "pop"=>3761, "name_ru"=>"Уолтоервилл", "name_en"=>"Walthourville", "lang"=>"en", "lat"=>31.775050, "lon"=>-81.628046),
"/US/US-GA/warm-springs-4166" => array("type"=>"city", "pop"=>477, "name_ru"=>"Уорм Спрингс", "name_en"=>"Warm Springs", "lang"=>"en", "lat"=>32.889128, "lon"=>-84.677602),
"/US/US-GA/warner-robins-4167" => array("type"=>"city", "pop"=>61336, "name_ru"=>"Уорнер Робинс", "name_en"=>"Warner Robins", "lang"=>"en", "lat"=>32.612695, "lon"=>-83.631339),
"/US/US-GA/warrenton-4168" => array("type"=>"city", "pop"=>1926, "name_ru"=>"Уоррентон", "name_en"=>"Warrenton", "lang"=>"en", "lat"=>33.406479, "lon"=>-82.665540),
"/US/US-GA/warwick-4169" => array("type"=>"city", "pop"=>401, "name_ru"=>"Уорвик", "name_en"=>"Warwick", "lang"=>"en", "lat"=>31.829847, "lon"=>-83.920365),
"/US/US-GA/washington-4170" => array("type"=>"city", "pop"=>4062, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>33.735506, "lon"=>-82.743969),
"/US/US-GA/watkinsville-4171" => array("type"=>"town", "pop"=>2880, "name_ru"=>"Уоткинсвилл", "name_en"=>"Watkinsville", "lang"=>"en", "lat"=>33.861887, "lon"=>-83.407595),
"/US/US-GA/waverly-hall-4172" => array("type"=>"town", "pop"=>781, "name_ru"=>"Уэверли Холл", "name_en"=>"Waverly Hall", "lang"=>"en", "lat"=>32.684022, "lon"=>-84.738164),
"/US/US-GA/waycross-4173" => array("type"=>"city", "pop"=>14774, "name_ru"=>"Уэйкросс", "name_en"=>"Waycross", "lang"=>"en", "lat"=>31.211945, "lon"=>-82.357737),
"/US/US-GA/waynesboro-4174" => array("type"=>"city", "pop"=>5874, "name_ru"=>"Уэйнсборо", "name_en"=>"Waynesboro", "lang"=>"en", "lat"=>33.090029, "lon"=>-82.013696),
"/US/US-GA/west-augusta-4175" => array("type"=>"cdp", "pop"=>27637, "name_ru"=>"Вэст Огаста", "name_en"=>"West Augusta", "lang"=>"en", "lat"=>33.486750, "lon"=>-82.058443),
"/US/US-GA/west-point-4177" => array("type"=>"city", "pop"=>3364, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>32.878614, "lon"=>-85.167863),
"/US/US-GA/weston-4176" => array("type"=>"town", "pop"=>69, "name_ru"=>"Вэстон", "name_en"=>"Weston", "lang"=>"en", "lat"=>31.976979, "lon"=>-84.612324),
"/US/US-GA/westside-4178" => array("type"=>"cdp", "pop"=>2180, "name_ru"=>"Вэстсайд", "name_en"=>"Westside", "lang"=>"en", "lat"=>34.271483, "lon"=>-83.845731),
"/US/US-GA/whigham-4179" => array("type"=>"city", "pop"=>605, "name_ru"=>"Уигам", "name_en"=>"Whigham", "lang"=>"en", "lat"=>30.883650, "lon"=>-84.324581),
"/US/US-GA/white-4180" => array("type"=>"town", "pop"=>707, "name_ru"=>"Уайт", "name_en"=>"White", "lang"=>"en", "lat"=>34.282447, "lon"=>-84.747660),
"/US/US-GA/white-plains-4182" => array("type"=>"town", "pop"=>303, "name_ru"=>"Уайт Плейнс", "name_en"=>"White Plains", "lang"=>"en", "lat"=>33.472598, "lon"=>-83.037615),
"/US/US-GA/whitemarsh-island-4181" => array("type"=>"cdp", "pop"=>2824, "name_ru"=>"Уайтмарш Айленд", "name_en"=>"Whitemarsh Island", "lang"=>"en", "lat"=>32.028651, "lon"=>-81.016965),
"/US/US-GA/whitesburg-4183" => array("type"=>"town", "pop"=>592, "name_ru"=>"Уайтсбург", "name_en"=>"Whitesburg", "lang"=>"en", "lat"=>33.493322, "lon"=>-84.913553),
"/US/US-GA/willacoochee-4184" => array("type"=>"town", "pop"=>1528, "name_ru"=>"Уиллакоочи", "name_en"=>"Willacoochee", "lang"=>"en", "lat"=>31.340153, "lon"=>-83.046170),
"/US/US-GA/williamson-4185" => array("type"=>"town", "pop"=>375, "name_ru"=>"Уильямсон", "name_en"=>"Williamson", "lang"=>"en", "lat"=>33.182138, "lon"=>-84.363144),
"/US/US-GA/wilmington-island-4186" => array("type"=>"cdp", "pop"=>11230, "name_ru"=>"Уилмингтон Айленд", "name_en"=>"Wilmington Island", "lang"=>"en", "lat"=>32.003451, "lon"=>-80.973872),
"/US/US-GA/winder-4187" => array("type"=>"city", "pop"=>14064, "name_ru"=>"Уиндер", "name_en"=>"Winder", "lang"=>"en", "lat"=>33.992013, "lon"=>-83.722172),
"/US/US-GA/winterville-4188" => array("type"=>"city", "pop"=>1187, "name_ru"=>"Уинтервилл", "name_en"=>"Winterville", "lang"=>"en", "lat"=>33.966606, "lon"=>-83.281792),
"/US/US-GA/woodbine-4189" => array("type"=>"city", "pop"=>1365, "name_ru"=>"Вудбайн", "name_en"=>"Woodbine", "lang"=>"en", "lat"=>30.959077, "lon"=>-81.721653),
"/US/US-GA/woodbury-4190" => array("type"=>"town", "pop"=>1073, "name_ru"=>"Вудбери", "name_en"=>"Woodbury", "lang"=>"en", "lat"=>32.983323, "lon"=>-84.582771),
"/US/US-GA/woodland-4191" => array("type"=>"city", "pop"=>392, "name_ru"=>"Вудленд", "name_en"=>"Woodland", "lang"=>"en", "lat"=>32.787461, "lon"=>-84.561096),
"/US/US-GA/woodstock-4192" => array("type"=>"city", "pop"=>23141, "name_ru"=>"Вудсток", "name_en"=>"Woodstock", "lang"=>"en", "lat"=>34.101467, "lon"=>-84.514440),
"/US/US-GA/woodville-4193" => array("type"=>"city", "pop"=>416, "name_ru"=>"Вудвилл", "name_en"=>"Woodville", "lang"=>"en", "lat"=>33.674096, "lon"=>-83.106664),
"/US/US-GA/woolsey-4194" => array("type"=>"town", "pop"=>200, "name_ru"=>"Вулси", "name_en"=>"Woolsey", "lang"=>"en", "lat"=>33.361322, "lon"=>-84.412031),
"/US/US-GA/wrens-4195" => array("type"=>"city", "pop"=>2217, "name_ru"=>"Ренс", "name_en"=>"Wrens", "lang"=>"en", "lat"=>33.206487, "lon"=>-82.387325),
"/US/US-GA/wrightsville-4196" => array("type"=>"city", "pop"=>3800, "name_ru"=>"Райтсвилл", "name_en"=>"Wrightsville", "lang"=>"en", "lat"=>32.724965, "lon"=>-82.720430),
"/US/US-GA/yatesville-4197" => array("type"=>"town", "pop"=>384, "name_ru"=>"Йетсвилл", "name_en"=>"Yatesville", "lang"=>"en", "lat"=>32.913164, "lon"=>-84.142653),
"/US/US-GA/young-harris-4198" => array("type"=>"city", "pop"=>709, "name_ru"=>"Янг Харрис", "name_en"=>"Young Harris", "lang"=>"en", "lat"=>34.933104, "lon"=>-83.848142),
"/US/US-GA/zebulon-4199" => array("type"=>"city", "pop"=>1250, "name_ru"=>"Зебулон", "name_en"=>"Zebulon", "lang"=>"en", "lat"=>33.100439, "lon"=>-84.342213),
"/US/US-GU" => array("type"=>"outlying area", "pop"=>173456, "name_ru"=>"Гуам", "name_en"=>"Guam", "lang"=>"en"),
"/US/US-GU/agana-23509" => array("type"=>"cdp", "pop"=>1139, "name_ru"=>"Агана", "name_en"=>"Agana", "lang"=>"en", "lat"=>13.475114, "lon"=>144.748441),
"/US/US-GU/agana-heights-23510" => array("type"=>"cdp", "pop"=>3347, "name_ru"=>"Агана Хайтс", "name_en"=>"Agana Heights", "lang"=>"en", "lat"=>13.462724, "lon"=>144.743889),
"/US/US-GU/agana-station-23511" => array("type"=>"cdp", "pop"=>2263, "name_ru"=>"Агана Стейшен", "name_en"=>"Agana Station", "lang"=>"en", "lat"=>13.479887, "lon"=>144.791370),
"/US/US-GU/agat-23512" => array("type"=>"cdp", "pop"=>3056, "name_ru"=>"Агат", "name_en"=>"Agat", "lang"=>"en", "lat"=>13.382383, "lon"=>144.657807),
"/US/US-GU/andersen-afb-23513" => array("type"=>"cdp", "pop"=>5531, "name_ru"=>"Военно-воздушная база Андерсон", "name_en"=>"Andersen AFB", "lang"=>"en", "lat"=>13.562699, "lon"=>144.927082),
"/US/US-GU/apra-harbor-23514" => array("type"=>"cdp", "pop"=>7303, "name_ru"=>"Апра Харбор", "name_en"=>"Apra Harbor", "lang"=>"en", "lat"=>13.426601, "lon"=>144.647249),
"/US/US-GU/asan-23515" => array("type"=>"cdp", "pop"=>740, "name_ru"=>"Асан", "name_en"=>"Asan", "lang"=>"en", "lat"=>13.467112, "lon"=>144.715269),
"/US/US-GU/barrigada-23516" => array("type"=>"cdp", "pop"=>3614, "name_ru"=>"Бэрригада", "name_en"=>"Barrigada", "lang"=>"en", "lat"=>13.461583, "lon"=>144.794688),
"/US/US-GU/barrigada-heights-23517" => array("type"=>"cdp", "pop"=>1063, "name_ru"=>"Бэрригада Хайтс", "name_en"=>"Barrigada Heights", "lang"=>"en", "lat"=>13.487702, "lon"=>144.823368),
"/US/US-GU/chalan-pago-23518" => array("type"=>"cdp", "pop"=>2736, "name_ru"=>"Чалан Паго", "name_en"=>"Chalan Pago", "lang"=>"en", "lat"=>13.436297, "lon"=>144.770396),
"/US/US-GU/dededo-23519" => array("type"=>"cdp", "pop"=>3166, "name_ru"=>"Дедедо", "name_en"=>"Dededo", "lang"=>"en", "lat"=>13.517008, "lon"=>144.836241),
"/US/US-GU/finegayan-station-23520" => array("type"=>"cdp", "pop"=>2895, "name_ru"=>"Файнгэян Стейшен", "name_en"=>"Finegayan Station", "lang"=>"en", "lat"=>13.568835, "lon"=>144.835327),
"/US/US-GU/inarajan-23521" => array("type"=>"cdp", "pop"=>1063, "name_ru"=>"Инарайан", "name_en"=>"Inarajan", "lang"=>"en", "lat"=>13.276068, "lon"=>144.737619),
"/US/US-GU/latte-heights-23522" => array("type"=>"cdp", "pop"=>1211, "name_ru"=>"Латте Хайтс", "name_en"=>"Latte Heights", "lang"=>"en", "lat"=>13.488334, "lon"=>144.840938),
"/US/US-GU/maina-23523" => array("type"=>"cdp", "pop"=>890, "name_ru"=>"Майна", "name_en"=>"Maina", "lang"=>"en", "lat"=>13.455331, "lon"=>144.729506),
"/US/US-GU/maite-23524" => array("type"=>"cdp", "pop"=>592, "name_ru"=>"Майте", "name_en"=>"Maite", "lang"=>"en", "lat"=>13.475734, "lon"=>144.762125),
"/US/US-GU/mangilao-23525" => array("type"=>"cdp", "pop"=>5608, "name_ru"=>"Мангилао", "name_en"=>"Mangilao", "lang"=>"en", "lat"=>13.441560, "lon"=>144.801665),
"/US/US-GU/marbo-annex-23526" => array("type"=>"cdp", "pop"=>1351, "name_ru"=>"Марбо Аннекс", "name_en"=>"Marbo Annex", "lang"=>"en", "lat"=>13.506643, "lon"=>144.865212),
"/US/US-GU/merizo-23527" => array("type"=>"cdp", "pop"=>1570, "name_ru"=>"Меризо", "name_en"=>"Merizo", "lang"=>"en", "lat"=>13.267846, "lon"=>144.670302),
"/US/US-GU/mongmong-23528" => array("type"=>"cdp", "pop"=>2346, "name_ru"=>"Монгмонг", "name_en"=>"Mongmong", "lang"=>"en", "lat"=>13.470568, "lon"=>144.765206),
"/US/US-GU/nimitz-hill-annex-23529" => array("type"=>"cdp", "pop"=>440, "name_ru"=>"Нимитс Хилл Аннекс", "name_en"=>"Nimitz Hill Annex", "lang"=>"en", "lat"=>13.459155, "lon"=>144.724799),
"/US/US-GU/ordot-23530" => array("type"=>"cdp", "pop"=>1715, "name_ru"=>"Ордот", "name_en"=>"Ordot", "lang"=>"en", "lat"=>13.446671, "lon"=>144.774397),
"/US/US-GU/piti-23531" => array("type"=>"cdp", "pop"=>723, "name_ru"=>"Пити", "name_en"=>"Piti", "lang"=>"en", "lat"=>13.464664, "lon"=>144.695410),
"/US/US-GU/santa-rita-23532" => array("type"=>"cdp", "pop"=>1243, "name_ru"=>"Санта Рита", "name_en"=>"Santa Rita", "lang"=>"en", "lat"=>13.387268, "lon"=>144.669309),
"/US/US-GU/santa-rosa-23533" => array("type"=>"cdp", "pop"=>926, "name_ru"=>"Санта Роза", "name_en"=>"Santa Rosa", "lang"=>"en", "lat"=>13.397062, "lon"=>144.671193),
"/US/US-GU/sinajana-23534" => array("type"=>"cdp", "pop"=>2024, "name_ru"=>"Синайана", "name_en"=>"Sinajana", "lang"=>"en", "lat"=>13.459616, "lon"=>144.752558),
"/US/US-GU/talofofo-23535" => array("type"=>"cdp", "pop"=>1594, "name_ru"=>"Талофофо", "name_en"=>"Talofofo", "lang"=>"en", "lat"=>13.355559, "lon"=>144.753380),
"/US/US-GU/tamuning-23536" => array("type"=>"cdp", "pop"=>9534, "name_ru"=>"Тамьюнинг", "name_en"=>"Tamuning", "lang"=>"en", "lat"=>13.493618, "lon"=>144.776709),
"/US/US-GU/toto-23537" => array("type"=>"cdp", "pop"=>2363, "name_ru"=>"Тото", "name_en"=>"Toto", "lang"=>"en", "lat"=>13.465512, "lon"=>144.777238),
"/US/US-GU/umatac-23538" => array("type"=>"cdp", "pop"=>395, "name_ru"=>"Юматак", "name_en"=>"Umatac", "lang"=>"en", "lat"=>13.296705, "lon"=>144.662035),
"/US/US-GU/yigo-23539" => array("type"=>"cdp", "pop"=>3358, "name_ru"=>"Йиго", "name_en"=>"Yigo", "lang"=>"en", "lat"=>13.538049, "lon"=>144.898527),
"/US/US-GU/yona-23540" => array("type"=>"cdp", "pop"=>2262, "name_ru"=>"Йона", "name_en"=>"Yona", "lang"=>"en", "lat"=>13.414336, "lon"=>144.773084),
"/US/US-HI" => array("type"=>"state", "pop"=>1288198, "name_ru"=>"Гавайи", "name_en"=>"Hawaii", "lang"=>"en"),
"/US/US-HI/ahuimanu-4200" => array("type"=>"cdp", "pop"=>8387, "name_ru"=>"Ауиману", "name_en"=>"Ahuimanu", "lang"=>"en", "lat"=>21.441237, "lon"=>-157.836518),
"/US/US-HI/aiea-4201" => array("type"=>"cdp", "pop"=>8906, "name_ru"=>"Эьеа", "name_en"=>"Aiea", "lang"=>"en", "lat"=>21.388450, "lon"=>-157.927831),
"/US/US-HI/aliamanu-4202" => array("type"=>"cdp", "pop"=>8835, "name_ru"=>"Алиаману", "name_en"=>"Aliamanu", "lang"=>"en", "lat"=>21.364854, "lon"=>-157.914388),
"/US/US-HI/anahola-4203" => array("type"=>"cdp", "pop"=>1181, "name_ru"=>"Анахола", "name_en"=>"Anahola", "lang"=>"en", "lat"=>22.147165, "lon"=>-159.317830),
"/US/US-HI/barbers-point-housing-4204" => array("type"=>"cdp", "pop"=>2218, "name_ru"=>"Барберс Пойнт Хаусинг", "name_en"=>"Barbers Point Housing", "lang"=>"en", "lat"=>21.324550, "lon"=>-158.083156),
"/US/US-HI/captain-cook-4205" => array("type"=>"cdp", "pop"=>2595, "name_ru"=>"Кэптейн Кук", "name_en"=>"Captain Cook", "lang"=>"en", "lat"=>19.501102, "lon"=>-155.903417),
"/US/US-HI/eleele-4206" => array("type"=>"cdp", "pop"=>1489, "name_ru"=>"Элил", "name_en"=>"Eleele", "lang"=>"en", "lat"=>21.910489, "lon"=>-159.584330),
"/US/US-HI/ewa-beach-4207" => array("type"=>"cdp", "pop"=>14315, "name_ru"=>"Ива Бич", "name_en"=>"Ewa Beach", "lang"=>"en", "lat"=>21.316150, "lon"=>-158.011805),
"/US/US-HI/ewa-gentry-4208" => array("type"=>"cdp", "pop"=>1992, "name_ru"=>"Ива Джентри", "name_en"=>"Ewa Gentry", "lang"=>"en", "lat"=>21.344150, "lon"=>-158.030775),
"/US/US-HI/ewa-villages-4209" => array("type"=>"cdp", "pop"=>3780, "name_ru"=>"Ива Виллиджис", "name_en"=>"Ewa Villages", "lang"=>"en", "lat"=>21.345450, "lon"=>-158.040879),
"/US/US-HI/fort-shafter-4210" => array("type"=>"cdp", "pop"=>2952, "name_ru"=>"Форт Шафтер", "name_en"=>"Fort Shafter", "lang"=>"en", "lat"=>21.353600, "lon"=>-157.874203),
"/US/US-HI/haiku-pauwela-4211" => array("type"=>"cdp", "pop"=>4509, "name_ru"=>"Хайку-Пауэла", "name_en"=>"Haiku-Pauwela", "lang"=>"en", "lat"=>20.921381, "lon"=>-156.304319),
"/US/US-HI/halaula-4212" => array("type"=>"cdp", "pop"=>496, "name_ru"=>"Халаула", "name_en"=>"Halaula", "lang"=>"en", "lat"=>20.231980, "lon"=>-155.783446),
"/US/US-HI/halawa-4213" => array("type"=>"cdp", "pop"=>13408, "name_ru"=>"Халава", "name_en"=>"Halawa", "lang"=>"en", "lat"=>21.379450, "lon"=>-157.921580),
"/US/US-HI/haleiwa-4214" => array("type"=>"cdp", "pop"=>2442, "name_ru"=>"Халейва", "name_en"=>"Haleiwa", "lang"=>"en", "lat"=>21.595476, "lon"=>-158.108837),
"/US/US-HI/haliimaile-4215" => array("type"=>"cdp", "pop"=>841, "name_ru"=>"Халиимейл", "name_en"=>"Haliimaile", "lang"=>"en", "lat"=>20.876254, "lon"=>-156.348277),
"/US/US-HI/hana-4216" => array("type"=>"cdp", "pop"=>683, "name_ru"=>"Хана", "name_en"=>"Hana", "lang"=>"en", "lat"=>20.770017, "lon"=>-155.994185),
"/US/US-HI/hanalei-4217" => array("type"=>"cdp", "pop"=>461, "name_ru"=>"Ханалей", "name_en"=>"Hanalei", "lang"=>"en", "lat"=>22.208978, "lon"=>-159.501701),
"/US/US-HI/hanamaulu-4218" => array("type"=>"cdp", "pop"=>3611, "name_ru"=>"Ханамаулу", "name_en"=>"Hanamaulu", "lang"=>"en", "lat"=>21.998151, "lon"=>-159.352431),
"/US/US-HI/hanapepe-4219" => array("type"=>"cdp", "pop"=>1395, "name_ru"=>"Ханапепе", "name_en"=>"Hanapepe", "lang"=>"en", "lat"=>21.916318, "lon"=>-159.590432),
"/US/US-HI/hauula-4220" => array("type"=>"cdp", "pop"=>3479, "name_ru"=>"Хауула", "name_en"=>"Hauula", "lang"=>"en", "lat"=>21.590639, "lon"=>-157.923293),
"/US/US-HI/hawaiian-beaches-4221" => array("type"=>"cdp", "pop"=>2846, "name_ru"=>"Гавайиан Бичис", "name_en"=>"Hawaiian Beaches", "lang"=>"en", "lat"=>19.543004, "lon"=>-154.915722),
"/US/US-HI/hawaiian-ocean-view-4222" => array("type"=>"cdp", "pop"=>969, "name_ru"=>"Гавайиан Оушен Вью", "name_en"=>"Hawaiian Ocean View", "lang"=>"en", "lat"=>19.068727, "lon"=>-155.764978),
"/US/US-HI/hawaiian-paradise-park-4223" => array("type"=>"cdp", "pop"=>3389, "name_ru"=>"Гавайиан Парадайз Парк", "name_en"=>"Hawaiian Paradise Park", "lang"=>"en", "lat"=>19.593256, "lon"=>-154.973125),
"/US/US-HI/hawi-4224" => array("type"=>"cdp", "pop"=>924, "name_ru"=>"Хои", "name_en"=>"Hawi", "lang"=>"en", "lat"=>20.242845, "lon"=>-155.834185),
"/US/US-HI/heeia-4225" => array("type"=>"cdp", "pop"=>5010, "name_ru"=>"Хеия", "name_en"=>"Heeia", "lang"=>"en", "lat"=>21.428000, "lon"=>-157.817183),
"/US/US-HI/hickam-housing-4226" => array("type"=>"cdp", "pop"=>6553, "name_ru"=>"Хикэм Хаузинг", "name_en"=>"Hickam Housing", "lang"=>"en", "lat"=>21.341267, "lon"=>-157.961371),
"/US/US-HI/hilo-4227" => array("type"=>"cdp", "pop"=>37808, "name_ru"=>"Хило", "name_en"=>"Hilo", "lang"=>"en", "lat"=>19.696151, "lon"=>-155.087501),
"/US/US-HI/holualoa-4228" => array("type"=>"cdp", "pop"=>3834, "name_ru"=>"Холуалоа", "name_en"=>"Holualoa", "lang"=>"en", "lat"=>19.635891, "lon"=>-155.929223),
"/US/US-HI/honalo-4229" => array("type"=>"cdp", "pop"=>1926, "name_ru"=>"Хонало", "name_en"=>"Honalo", "lang"=>"en", "lat"=>19.570685, "lon"=>-155.907667),
"/US/US-HI/honaunau-napoopoo-4230" => array("type"=>"cdp", "pop"=>2373, "name_ru"=>"Хонауно-Напупу", "name_en"=>"Honaunau-Napoopoo", "lang"=>"en", "lat"=>19.452742, "lon"=>-155.880959),
"/US/US-HI/honokaa-4231" => array("type"=>"cdp", "pop"=>2186, "name_ru"=>"Хонокаа", "name_en"=>"Honokaa", "lang"=>"en", "lat"=>20.077056, "lon"=>-155.467439),
"/US/US-HI/honolulu-4232" => array("type"=>"cdp", "pop"=>374676, "name_ru"=>"Гонолулу", "name_en"=>"Honolulu", "lang"=>"en", "lat"=>21.317250, "lon"=>-157.804233),
"/US/US-HI/honomu-4233" => array("type"=>"cdp", "pop"=>532, "name_ru"=>"Хоному", "name_en"=>"Honomu", "lang"=>"en", "lat"=>19.872718, "lon"=>-155.114899),
"/US/US-HI/iroquois-point-4234" => array("type"=>"cdp", "pop"=>4188, "name_ru"=>"Ирокез Пойнт", "name_en"=>"Iroquois Point", "lang"=>"en", "lat"=>21.325329, "lon"=>-157.981333),
"/US/US-HI/kaaawa-4235" => array("type"=>"cdp", "pop"=>1138, "name_ru"=>"Каава", "name_en"=>"Kaaawa", "lang"=>"en", "lat"=>21.557376, "lon"=>-157.857206),
"/US/US-HI/kaanapali-4236" => array("type"=>"cdp", "pop"=>579, "name_ru"=>"Каанапали", "name_en"=>"Kaanapali", "lang"=>"en", "lat"=>20.933598, "lon"=>-156.683469),
"/US/US-HI/kahaluu-4237" => array("type"=>"cdp", "pop"=>3068, "name_ru"=>"Кахалуу", "name_en"=>"Kahaluu", "lang"=>"en", "lat"=>21.461986, "lon"=>-157.834830),
"/US/US-HI/kahaluu-keauhou-4238" => array("type"=>"cdp", "pop"=>1990, "name_ru"=>"Кахальюу-Киоау", "name_en"=>"Kahaluu-Keauhou", "lang"=>"en", "lat"=>19.575554, "lon"=>-155.960519),
"/US/US-HI/kahuku-4239" => array("type"=>"cdp", "pop"=>2063, "name_ru"=>"Кахуку", "name_en"=>"Kahuku", "lang"=>"en", "lat"=>21.682228, "lon"=>-157.946296),
"/US/US-HI/kahului-4240" => array("type"=>"cdp", "pop"=>16889, "name_ru"=>"Каулуи", "name_en"=>"Kahului", "lang"=>"en", "lat"=>20.873837, "lon"=>-156.459344),
"/US/US-HI/kailua-4241" => array("type"=>"cdp", "pop"=>9126, "name_ru"=>"Кейлуа", "name_en"=>"Kailua", "lang"=>"en", "lat"=>19.664237, "lon"=>-155.964037),
"/US/US-HI/kailua-4242" => array("type"=>"cdp", "pop"=>36818, "name_ru"=>"Кейлуа", "name_en"=>"Kailua", "lang"=>"en", "lat"=>21.401600, "lon"=>-157.738455),
"/US/US-HI/kalaheo-4243" => array("type"=>"cdp", "pop"=>3592, "name_ru"=>"Калахео", "name_en"=>"Kalaheo", "lang"=>"en", "lat"=>21.920807, "lon"=>-159.526719),
"/US/US-HI/kalaoa-4244" => array("type"=>"cdp", "pop"=>4490, "name_ru"=>"Калаоа", "name_en"=>"Kalaoa", "lang"=>"en", "lat"=>19.725650, "lon"=>-156.018605),
"/US/US-HI/kalihiwai-4245" => array("type"=>"cdp", "pop"=>435, "name_ru"=>"Калиайвай", "name_en"=>"Kalihiwai", "lang"=>"en", "lat"=>22.219418, "lon"=>-159.421522),
"/US/US-HI/kaneohe-4246" => array("type"=>"cdp", "pop"=>35448, "name_ru"=>"Кано", "name_en"=>"Kaneohe", "lang"=>"en", "lat"=>21.411100, "lon"=>-157.792453),
"/US/US-HI/kaneohe-station-4247" => array("type"=>"cdp", "pop"=>11662, "name_ru"=>"Кано Стейшен", "name_en"=>"Kaneohe Station", "lang"=>"en", "lat"=>21.450050, "lon"=>-157.751644),
"/US/US-HI/kapaa-4248" => array("type"=>"cdp", "pop"=>8149, "name_ru"=>"Капаа", "name_en"=>"Kapaa", "lang"=>"en", "lat"=>22.097313, "lon"=>-159.353978),
"/US/US-HI/kapaau-4249" => array("type"=>"cdp", "pop"=>1083, "name_ru"=>"Капаау", "name_en"=>"Kapaau", "lang"=>"en", "lat"=>20.230666, "lon"=>-155.810858),
"/US/US-HI/kapalua-4250" => array("type"=>"cdp", "pop"=>408, "name_ru"=>"Капалуа", "name_en"=>"Kapalua", "lang"=>"en", "lat"=>20.999664, "lon"=>-156.658126),
"/US/US-HI/kaumakani-4251" => array("type"=>"cdp", "pop"=>803, "name_ru"=>"Каумакани", "name_en"=>"Kaumakani", "lang"=>"en", "lat"=>21.922135, "lon"=>-159.628085),
"/US/US-HI/kaunakakai-4252" => array("type"=>"cdp", "pop"=>2658, "name_ru"=>"Каунакакаи", "name_en"=>"Kaunakakai", "lang"=>"en", "lat"=>21.088593, "lon"=>-157.012690),
"/US/US-HI/kawela-bay-4253" => array("type"=>"cdp", "pop"=>366, "name_ru"=>"Кауэла Бэй", "name_en"=>"Kawela Bay", "lang"=>"en", "lat"=>21.703230, "lon"=>-158.009886),
"/US/US-HI/keaau-4254" => array("type"=>"cdp", "pop"=>1584, "name_ru"=>"Кио", "name_en"=>"Keaau", "lang"=>"en", "lat"=>19.619226, "lon"=>-155.039674),
"/US/US-HI/kealakekua-4255" => array("type"=>"cdp", "pop"=>1453, "name_ru"=>"Килейккуа", "name_en"=>"Kealakekua", "lang"=>"en", "lat"=>19.540023, "lon"=>-155.882395),
"/US/US-HI/kekaha-4256" => array("type"=>"cdp", "pop"=>3506, "name_ru"=>"Кекаха", "name_en"=>"Kekaha", "lang"=>"en", "lat"=>21.972283, "lon"=>-159.717815),
"/US/US-HI/kihei-4257" => array("type"=>"cdp", "pop"=>11107, "name_ru"=>"Кихей", "name_en"=>"Kihei", "lang"=>"en", "lat"=>20.763794, "lon"=>-156.453242),
"/US/US-HI/kilauea-4258" => array("type"=>"cdp", "pop"=>1685, "name_ru"=>"Килоея", "name_en"=>"Kilauea", "lang"=>"en", "lat"=>22.211103, "lon"=>-159.409609),
"/US/US-HI/koloa-4259" => array("type"=>"cdp", "pop"=>1791, "name_ru"=>"Колоа", "name_en"=>"Koloa", "lang"=>"en", "lat"=>21.906811, "lon"=>-159.464254),
"/US/US-HI/kualapuu-4260" => array("type"=>"cdp", "pop"=>1661, "name_ru"=>"Куалапуу", "name_en"=>"Kualapuu", "lang"=>"en", "lat"=>21.159996, "lon"=>-157.058374),
"/US/US-HI/kukuihaele-4261" => array("type"=>"cdp", "pop"=>316, "name_ru"=>"Кукил", "name_en"=>"Kukuihaele", "lang"=>"en", "lat"=>20.125490, "lon"=>-155.570697),
"/US/US-HI/kurtistown-4262" => array("type"=>"cdp", "pop"=>910, "name_ru"=>"Кертистаун", "name_en"=>"Kurtistown", "lang"=>"en", "lat"=>19.589112, "lon"=>-155.071063),
"/US/US-HI/lahaina-4263" => array("type"=>"cdp", "pop"=>9073, "name_ru"=>"Лахайна", "name_en"=>"Lahaina", "lang"=>"en", "lat"=>20.891278, "lon"=>-156.674264),
"/US/US-HI/laie-4264" => array("type"=>"cdp", "pop"=>5577, "name_ru"=>"Лаи", "name_en"=>"Laie", "lang"=>"en", "lat"=>21.649067, "lon"=>-157.925454),
"/US/US-HI/lanai-city-4265" => array("type"=>"cdp", "pop"=>2400, "name_ru"=>"Ланай Сити", "name_en"=>"Lanai City", "lang"=>"en", "lat"=>20.832701, "lon"=>-156.926533),
"/US/US-HI/laupahoehoe-4266" => array("type"=>"cdp", "pop"=>508, "name_ru"=>"Лопахохо", "name_en"=>"Laupahoehoe", "lang"=>"en", "lat"=>19.978529, "lon"=>-155.240079),
"/US/US-HI/lawai-4267" => array("type"=>"cdp", "pop"=>1787, "name_ru"=>"Лавай", "name_en"=>"Lawai", "lang"=>"en", "lat"=>21.924943, "lon"=>-159.504739),
"/US/US-HI/lihue-4268" => array("type"=>"cdp", "pop"=>5536, "name_ru"=>"Лию", "name_en"=>"Lihue", "lang"=>"en", "lat"=>21.975107, "lon"=>-159.354424),
"/US/US-HI/maalaea-4269" => array("type"=>"cdp", "pop"=>443, "name_ru"=>"Маллаея", "name_en"=>"Maalaea", "lang"=>"en", "lat"=>20.799968, "lon"=>-156.494404),
"/US/US-HI/maili-4270" => array("type"=>"cdp", "pop"=>6059, "name_ru"=>"Мейли", "name_en"=>"Maili", "lang"=>"en", "lat"=>21.418733, "lon"=>-158.180420),
"/US/US-HI/makaha-4271" => array("type"=>"cdp", "pop"=>7990, "name_ru"=>"Макаха", "name_en"=>"Makaha", "lang"=>"en", "lat"=>21.468274, "lon"=>-158.215062),
"/US/US-HI/makaha-valley-4272" => array("type"=>"cdp", "pop"=>1012, "name_ru"=>"Макаха Валли", "name_en"=>"Makaha Valley", "lang"=>"en", "lat"=>21.482350, "lon"=>-158.203782),
"/US/US-HI/makakilo-city-4273" => array("type"=>"cdp", "pop"=>9828, "name_ru"=>"Макакило Сити", "name_en"=>"Makakilo City", "lang"=>"en", "lat"=>21.358558, "lon"=>-158.088427),
"/US/US-HI/makawao-4274" => array("type"=>"cdp", "pop"=>5405, "name_ru"=>"Макавао", "name_en"=>"Makawao", "lang"=>"en", "lat"=>20.857539, "lon"=>-156.325011),
"/US/US-HI/maunaloa-4275" => array("type"=>"cdp", "pop"=>405, "name_ru"=>"Мауналоа", "name_en"=>"Maunaloa", "lang"=>"en", "lat"=>21.136455, "lon"=>-157.213884),
"/US/US-HI/maunawili-4276" => array("type"=>"cdp", "pop"=>4847, "name_ru"=>"Мауноили", "name_en"=>"Maunawili", "lang"=>"en", "lat"=>21.372400, "lon"=>-157.765766),
"/US/US-HI/mililani-town-4277" => array("type"=>"cdp", "pop"=>29359, "name_ru"=>"Милиани Таун", "name_en"=>"Mililani Town", "lang"=>"en", "lat"=>21.448972, "lon"=>-158.018162),
"/US/US-HI/mokuleia-4278" => array("type"=>"cdp", "pop"=>1776, "name_ru"=>"Мокулеия", "name_en"=>"Mokuleia", "lang"=>"en", "lat"=>21.582216, "lon"=>-158.178807),
"/US/US-HI/mountain-view-4279" => array("type"=>"cdp", "pop"=>3075, "name_ru"=>"Маунтин Вью", "name_en"=>"Mountain View", "lang"=>"en", "lat"=>19.520660, "lon"=>-155.149245),
"/US/US-HI/naalehu-4280" => array("type"=>"cdp", "pop"=>1027, "name_ru"=>"Наалеу", "name_en"=>"Naalehu", "lang"=>"en", "lat"=>19.077182, "lon"=>-155.575282),
"/US/US-HI/nanakuli-4281" => array("type"=>"cdp", "pop"=>9575, "name_ru"=>"Нанакули", "name_en"=>"Nanakuli", "lang"=>"en", "lat"=>21.391900, "lon"=>-158.156452),
"/US/US-HI/napili-honokowai-4282" => array("type"=>"cdp", "pop"=>4332, "name_ru"=>"Нейпили-Хоноковай", "name_en"=>"Napili-Honokowai", "lang"=>"en", "lat"=>20.971623, "lon"=>-156.669630),
"/US/US-HI/omao-4283" => array("type"=>"cdp", "pop"=>1142, "name_ru"=>"Омао", "name_en"=>"Omao", "lang"=>"en", "lat"=>21.925077, "lon"=>-159.486181),
"/US/US-HI/paauilo-4284" => array("type"=>"cdp", "pop"=>620, "name_ru"=>"Паоило", "name_en"=>"Paauilo", "lang"=>"en", "lat"=>20.043372, "lon"=>-155.371742),
"/US/US-HI/pahala-4285" => array("type"=>"cdp", "pop"=>1520, "name_ru"=>"Пахала", "name_en"=>"Pahala", "lang"=>"en", "lat"=>19.203250, "lon"=>-155.482409),
"/US/US-HI/pahoa-4286" => array("type"=>"cdp", "pop"=>1027, "name_ru"=>"Пахоа", "name_en"=>"Pahoa", "lang"=>"en", "lat"=>19.498664, "lon"=>-154.948882),
"/US/US-HI/paia-4287" => array("type"=>"cdp", "pop"=>2091, "name_ru"=>"Паия", "name_en"=>"Paia", "lang"=>"en", "lat"=>20.911731, "lon"=>-156.370982),
"/US/US-HI/pakala-village-4288" => array("type"=>"cdp", "pop"=>565, "name_ru"=>"Пакала Виллидж", "name_en"=>"Pakala Village", "lang"=>"en", "lat"=>21.945001, "lon"=>-159.648673),
"/US/US-HI/papaikou-4289" => array("type"=>"cdp", "pop"=>1634, "name_ru"=>"Папайкоу", "name_en"=>"Papaikou", "lang"=>"en", "lat"=>19.794022, "lon"=>-155.096531),
"/US/US-HI/paukaa-4290" => array("type"=>"cdp", "pop"=>495, "name_ru"=>"Паукаа", "name_en"=>"Paukaa", "lang"=>"en", "lat"=>19.764369, "lon"=>-155.098569),
"/US/US-HI/pearl-city-4291" => array("type"=>"cdp", "pop"=>30993, "name_ru"=>"Перл Сити", "name_en"=>"Pearl City", "lang"=>"en", "lat"=>21.404667, "lon"=>-157.969056),
"/US/US-HI/pepeekeo-4292" => array("type"=>"cdp", "pop"=>1813, "name_ru"=>"Пепеекео", "name_en"=>"Pepeekeo", "lang"=>"en", "lat"=>19.834682, "lon"=>-155.108823),
"/US/US-HI/poipu-4293" => array("type"=>"cdp", "pop"=>975, "name_ru"=>"Пойпу", "name_en"=>"Poipu", "lang"=>"en", "lat"=>21.881420, "lon"=>-159.462403),
"/US/US-HI/princeville-4294" => array("type"=>"cdp", "pop"=>1244, "name_ru"=>"Принсвилл", "name_en"=>"Princeville", "lang"=>"en", "lat"=>22.223513, "lon"=>-159.485392),
"/US/US-HI/puako-4295" => array("type"=>"cdp", "pop"=>397, "name_ru"=>"Пуако", "name_en"=>"Puako", "lang"=>"en", "lat"=>19.944758, "lon"=>-155.871074),
"/US/US-HI/puhi-4296" => array("type"=>"cdp", "pop"=>1210, "name_ru"=>"Пухи", "name_en"=>"Puhi", "lang"=>"en", "lat"=>21.968793, "lon"=>-159.399556),
"/US/US-HI/pukalani-4297" => array("type"=>"cdp", "pop"=>5879, "name_ru"=>"Пекалани", "name_en"=>"Pukalani", "lang"=>"en", "lat"=>20.836879, "lon"=>-156.341033),
"/US/US-HI/punaluu-4298" => array("type"=>"cdp", "pop"=>672, "name_ru"=>"Пунальюу", "name_en"=>"Punaluu", "lang"=>"en", "lat"=>21.592600, "lon"=>-157.896576),
"/US/US-HI/pupukea-4299" => array("type"=>"cdp", "pop"=>4111, "name_ru"=>"Пупекея", "name_en"=>"Pupukea", "lang"=>"en", "lat"=>21.664295, "lon"=>-158.053492),
"/US/US-HI/schofield-barracks-4300" => array("type"=>"cdp", "pop"=>19597, "name_ru"=>"Шофилд Баррекс", "name_en"=>"Schofield Barracks", "lang"=>"en", "lat"=>21.496344, "lon"=>-158.064555),
"/US/US-HI/village-park-4301" => array("type"=>"cdp", "pop"=>7407, "name_ru"=>"Виллидж Парк", "name_en"=>"Village Park", "lang"=>"en", "lat"=>21.398150, "lon"=>-158.030187),
"/US/US-HI/volcano-4302" => array("type"=>"cdp", "pop"=>1516, "name_ru"=>"Волкано", "name_en"=>"Volcano", "lang"=>"en", "lat"=>19.497664, "lon"=>-155.237292),
"/US/US-HI/wahiawa-4303" => array("type"=>"cdp", "pop"=>17386, "name_ru"=>"Вахайава", "name_en"=>"Wahiawa", "lang"=>"en", "lat"=>21.502900, "lon"=>-158.022129),
"/US/US-HI/waialua-4304" => array("type"=>"cdp", "pop"=>3943, "name_ru"=>"Уэйэлуа", "name_en"=>"Waialua", "lang"=>"en", "lat"=>21.577600, "lon"=>-158.128521),
"/US/US-HI/waianae-4305" => array("type"=>"cdp", "pop"=>8758, "name_ru"=>"Уоиана", "name_en"=>"Waianae", "lang"=>"en", "lat"=>21.449400, "lon"=>-158.184010),
"/US/US-HI/waihee-waiehue-4306" => array("type"=>"cdp", "pop"=>4004, "name_ru"=>"Уэйи-Уэйею", "name_en"=>"Waihee-Waiehue", "lang"=>"en", "lat"=>20.926241, "lon"=>-156.509572),
"/US/US-HI/waikane-4307" => array("type"=>"cdp", "pop"=>717, "name_ru"=>"Уэйкан", "name_en"=>"Waikane", "lang"=>"en", "lat"=>21.501379, "lon"=>-157.875226),
"/US/US-HI/waikapu-4308" => array("type"=>"cdp", "pop"=>729, "name_ru"=>"Уэйкапу", "name_en"=>"Waikapu", "lang"=>"en", "lat"=>20.839542, "lon"=>-156.533886),
"/US/US-HI/waikoloa-village-4309" => array("type"=>"cdp", "pop"=>2248, "name_ru"=>"Уэйколоа Виллидж", "name_en"=>"Waikoloa Village", "lang"=>"en", "lat"=>19.932380, "lon"=>-155.827678),
"/US/US-HI/wailea-makena-4310" => array("type"=>"cdp", "pop"=>3799, "name_ru"=>"Уэйлея-Макена", "name_en"=>"Wailea-Makena", "lang"=>"en", "lat"=>20.665356, "lon"=>-156.433127),
"/US/US-HI/wailua-4311" => array("type"=>"cdp", "pop"=>2018, "name_ru"=>"Уэйлуа", "name_en"=>"Wailua", "lang"=>"en", "lat"=>22.057933, "lon"=>-159.340675),
"/US/US-HI/wailua-homesteads-4312" => array("type"=>"cdp", "pop"=>3870, "name_ru"=>"Уэйлуа Хоумстэдс", "name_en"=>"Wailua Homesteads", "lang"=>"en", "lat"=>22.065682, "lon"=>-159.379603),
"/US/US-HI/wailuku-4313" => array("type"=>"cdp", "pop"=>10688, "name_ru"=>"Уэйлуку", "name_en"=>"Wailuku", "lang"=>"en", "lat"=>20.888525, "lon"=>-156.506775),
"/US/US-HI/waimalu-4314" => array("type"=>"cdp", "pop"=>29967, "name_ru"=>"Уэймалу", "name_en"=>"Waimalu", "lang"=>"en", "lat"=>21.404659, "lon"=>-157.943243),
"/US/US-HI/waimanalo-4315" => array("type"=>"cdp", "pop"=>3508, "name_ru"=>"Уэйманало", "name_en"=>"Waimanalo", "lang"=>"en", "lat"=>21.347617, "lon"=>-157.724982),
"/US/US-HI/waimanalo-beach-4316" => array("type"=>"cdp", "pop"=>4185, "name_ru"=>"Уэйманало Бич", "name_en"=>"Waimanalo Beach", "lang"=>"en", "lat"=>21.329967, "lon"=>-157.690232),
"/US/US-HI/waimea-4317" => array("type"=>"cdp", "pop"=>5972, "name_ru"=>"Уэймея", "name_en"=>"Waimea", "lang"=>"en", "lat"=>20.016551, "lon"=>-155.636608),
"/US/US-HI/waimea-4318" => array("type"=>"cdp", "pop"=>1840, "name_ru"=>"Уэймея", "name_en"=>"Waimea", "lang"=>"en", "lat"=>21.962076, "lon"=>-159.674993),
"/US/US-HI/wainaku-4319" => array("type"=>"cdp", "pop"=>1243, "name_ru"=>"Уэйнаку", "name_en"=>"Wainaku", "lang"=>"en", "lat"=>19.747258, "lon"=>-155.100954),
"/US/US-HI/waipahu-4320" => array("type"=>"cdp", "pop"=>31435, "name_ru"=>"Уэйпау", "name_en"=>"Waipahu", "lang"=>"en", "lat"=>21.390600, "lon"=>-158.012708),
"/US/US-HI/waipio-4321" => array("type"=>"cdp", "pop"=>11812, "name_ru"=>"Уэйпио", "name_en"=>"Waipio", "lang"=>"en", "lat"=>21.418307, "lon"=>-158.000602),
"/US/US-HI/waipio-acres-4322" => array("type"=>"cdp", "pop"=>5304, "name_ru"=>"Уэйпио Эйкрс", "name_en"=>"Waipio Acres", "lang"=>"en", "lat"=>21.474163, "lon"=>-158.020130),
"/US/US-HI/wheeler-afb-4323" => array("type"=>"cdp", "pop"=>2600, "name_ru"=>"Военно-воздушная база Уилер", "name_en"=>"Wheeler AFB", "lang"=>"en", "lat"=>21.481945, "lon"=>-158.041423),
"/US/US-HI/whitmore-village-4324" => array("type"=>"cdp", "pop"=>3373, "name_ru"=>"Уитмор Виллидж", "name_en"=>"Whitmore Village", "lang"=>"en", "lat"=>21.514185, "lon"=>-158.029166),
"/US/US-IA" => array("type"=>"state", "pop"=>3002555, "name_ru"=>"Айова", "name_en"=>"Iowa", "lang"=>"en"),
"/US/US-IA/ackley-6426" => array("type"=>"city", "pop"=>1665, "name_ru"=>"Акли", "name_en"=>"Ackley", "lang"=>"en", "lat"=>42.551525, "lon"=>-93.051942),
"/US/US-IA/ackworth-6427" => array("type"=>"city", "pop"=>116, "name_ru"=>"Акворт", "name_en"=>"Ackworth", "lang"=>"en", "lat"=>41.365123, "lon"=>-93.473022),
"/US/US-IA/adair-6428" => array("type"=>"city", "pop"=>713, "name_ru"=>"Адэр", "name_en"=>"Adair", "lang"=>"en", "lat"=>41.500107, "lon"=>-94.643587),
"/US/US-IA/adel-6429" => array("type"=>"city", "pop"=>4428, "name_ru"=>"Адел", "name_en"=>"Adel", "lang"=>"en", "lat"=>41.612554, "lon"=>-94.027058),
"/US/US-IA/afton-6430" => array("type"=>"city", "pop"=>905, "name_ru"=>"Афтон", "name_en"=>"Afton", "lang"=>"en", "lat"=>41.027924, "lon"=>-94.195625),
"/US/US-IA/agency-6431" => array("type"=>"city", "pop"=>644, "name_ru"=>"Эдженси", "name_en"=>"Agency", "lang"=>"en", "lat"=>40.996854, "lon"=>-92.307324),
"/US/US-IA/ainsworth-6432" => array("type"=>"city", "pop"=>537, "name_ru"=>"Эйнсворт", "name_en"=>"Ainsworth", "lang"=>"en", "lat"=>41.290155, "lon"=>-91.554122),
"/US/US-IA/akron-6433" => array("type"=>"city", "pop"=>1420, "name_ru"=>"Акрон", "name_en"=>"Akron", "lang"=>"en", "lat"=>42.827309, "lon"=>-96.556264),
"/US/US-IA/albert-city-6434" => array("type"=>"city", "pop"=>690, "name_ru"=>"Алберт Сити", "name_en"=>"Albert City", "lang"=>"en", "lat"=>42.781120, "lon"=>-94.948937),
"/US/US-IA/albia-6435" => array("type"=>"city", "pop"=>3559, "name_ru"=>"Албия", "name_en"=>"Albia", "lang"=>"en", "lat"=>41.026582, "lon"=>-92.802793),
"/US/US-IA/albion-6436" => array("type"=>"city", "pop"=>575, "name_ru"=>"Албион", "name_en"=>"Albion", "lang"=>"en", "lat"=>42.112766, "lon"=>-92.989184),
"/US/US-IA/alburnett-6437" => array("type"=>"city", "pop"=>551, "name_ru"=>"Олбернетт", "name_en"=>"Alburnett", "lang"=>"en", "lat"=>42.149531, "lon"=>-91.618161),
"/US/US-IA/alden-6438" => array("type"=>"city", "pop"=>818, "name_ru"=>"Олден", "name_en"=>"Alden", "lang"=>"en", "lat"=>42.512427, "lon"=>-93.379028),
"/US/US-IA/alexander-6439" => array("type"=>"city", "pop"=>157, "name_ru"=>"Александр", "name_en"=>"Alexander", "lang"=>"en", "lat"=>42.804790, "lon"=>-93.477190),
"/US/US-IA/algona-6440" => array("type"=>"city", "pop"=>5328, "name_ru"=>"Алгона", "name_en"=>"Algona", "lang"=>"en", "lat"=>43.073920, "lon"=>-94.230743),
"/US/US-IA/alleman-6441" => array("type"=>"city", "pop"=>435, "name_ru"=>"Аллеман", "name_en"=>"Alleman", "lang"=>"en", "lat"=>41.817124, "lon"=>-93.610348),
"/US/US-IA/allerton-6442" => array("type"=>"city", "pop"=>531, "name_ru"=>"Аллертон", "name_en"=>"Allerton", "lang"=>"en", "lat"=>40.708188, "lon"=>-93.367020),
"/US/US-IA/allison-6443" => array("type"=>"city", "pop"=>950, "name_ru"=>"Аллисон", "name_en"=>"Allison", "lang"=>"en", "lat"=>42.752472, "lon"=>-92.795645),
"/US/US-IA/alta-6444" => array("type"=>"city", "pop"=>1864, "name_ru"=>"Альта", "name_en"=>"Alta", "lang"=>"en", "lat"=>42.672049, "lon"=>-95.304481),
"/US/US-IA/alta-vista-6445" => array("type"=>"city", "pop"=>265, "name_ru"=>"Альта Виста", "name_en"=>"Alta Vista", "lang"=>"en", "lat"=>43.195675, "lon"=>-92.417126),
"/US/US-IA/alton-6446" => array("type"=>"city", "pop"=>1127, "name_ru"=>"Альтон", "name_en"=>"Alton", "lang"=>"en", "lat"=>42.986383, "lon"=>-96.007594),
"/US/US-IA/altoona-6447" => array("type"=>"city", "pop"=>13854, "name_ru"=>"Альтоона", "name_en"=>"Altoona", "lang"=>"en", "lat"=>41.650875, "lon"=>-93.476290),
"/US/US-IA/alvord-6448" => array("type"=>"city", "pop"=>177, "name_ru"=>"Алворд", "name_en"=>"Alvord", "lang"=>"en", "lat"=>43.342168, "lon"=>-96.302510),
"/US/US-IA/ames-6449" => array("type"=>"city", "pop"=>56510, "name_ru"=>"Эймс", "name_en"=>"Ames", "lang"=>"en", "lat"=>42.023350, "lon"=>-93.625622),
"/US/US-IA/anamosa-6450" => array("type"=>"city", "pop"=>5736, "name_ru"=>"Анамоса", "name_en"=>"Anamosa", "lang"=>"en", "lat"=>42.108951, "lon"=>-91.281306),
"/US/US-IA/andover-6451" => array("type"=>"city", "pop"=>84, "name_ru"=>"Андоувер", "name_en"=>"Andover", "lang"=>"en", "lat"=>41.978808, "lon"=>-90.251584),
"/US/US-IA/andrew-6452" => array("type"=>"city", "pop"=>437, "name_ru"=>"Эндрю", "name_en"=>"Andrew", "lang"=>"en", "lat"=>42.153101, "lon"=>-90.591667),
"/US/US-IA/anita-6453" => array("type"=>"city", "pop"=>1135, "name_ru"=>"Анита", "name_en"=>"Anita", "lang"=>"en", "lat"=>41.443563, "lon"=>-94.765348),
"/US/US-IA/ankeny-6454" => array("type"=>"city", "pop"=>42287, "name_ru"=>"Анкени", "name_en"=>"Ankeny", "lang"=>"en", "lat"=>41.724775, "lon"=>-93.605324),
"/US/US-IA/anthon-6455" => array("type"=>"city", "pop"=>631, "name_ru"=>"Энтон", "name_en"=>"Anthon", "lang"=>"en", "lat"=>42.387748, "lon"=>-95.866074),
"/US/US-IA/aplington-6456" => array("type"=>"city", "pop"=>999, "name_ru"=>"Аплингтон", "name_en"=>"Aplington", "lang"=>"en", "lat"=>42.581604, "lon"=>-92.880941),
"/US/US-IA/arcadia-6457" => array("type"=>"city", "pop"=>444, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>42.086973, "lon"=>-95.043506),
"/US/US-IA/archer-6458" => array("type"=>"city", "pop"=>115, "name_ru"=>"Арчер", "name_en"=>"Archer", "lang"=>"en", "lat"=>43.114467, "lon"=>-95.744050),
"/US/US-IA/aredale-6459" => array("type"=>"city", "pop"=>76, "name_ru"=>"Ардейл", "name_en"=>"Aredale", "lang"=>"en", "lat"=>42.833444, "lon"=>-93.004491),
"/US/US-IA/arion-6460" => array("type"=>"city", "pop"=>112, "name_ru"=>"Арион", "name_en"=>"Arion", "lang"=>"en", "lat"=>41.948576, "lon"=>-95.462481),
"/US/US-IA/arispe-6461" => array("type"=>"city", "pop"=>87, "name_ru"=>"Арисп", "name_en"=>"Arispe", "lang"=>"en", "lat"=>40.949020, "lon"=>-94.218744),
"/US/US-IA/arlington-6462" => array("type"=>"city", "pop"=>454, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>42.748748, "lon"=>-91.670352),
"/US/US-IA/armstrong-6463" => array("type"=>"city", "pop"=>917, "name_ru"=>"Армстронг", "name_en"=>"Armstrong", "lang"=>"en", "lat"=>43.393555, "lon"=>-94.479005),
"/US/US-IA/arnolds-park-6464" => array("type"=>"city", "pop"=>1249, "name_ru"=>"Арнольдс Парк", "name_en"=>"Arnolds Park", "lang"=>"en", "lat"=>43.362839, "lon"=>-95.131796),
"/US/US-IA/arthur-6465" => array("type"=>"city", "pop"=>209, "name_ru"=>"Артур", "name_en"=>"Arthur", "lang"=>"en", "lat"=>42.335240, "lon"=>-95.346384),
"/US/US-IA/asbury-6466" => array("type"=>"city", "pop"=>3848, "name_ru"=>"Асбери", "name_en"=>"Asbury", "lang"=>"en", "lat"=>42.514750, "lon"=>-90.758517),
"/US/US-IA/ashton-6467" => array("type"=>"city", "pop"=>423, "name_ru"=>"Аштон", "name_en"=>"Ashton", "lang"=>"en", "lat"=>43.309842, "lon"=>-95.792221),
"/US/US-IA/aspinwall-6468" => array("type"=>"city", "pop"=>64, "name_ru"=>"Аспинуолл", "name_en"=>"Aspinwall", "lang"=>"en", "lat"=>41.911447, "lon"=>-95.134765),
"/US/US-IA/atalissa-6469" => array("type"=>"city", "pop"=>303, "name_ru"=>"Аталисса", "name_en"=>"Atalissa", "lang"=>"en", "lat"=>41.571544, "lon"=>-91.166552),
"/US/US-IA/athelstan-6470" => array("type"=>"city", "pop"=>31, "name_ru"=>"Ательстан", "name_en"=>"Athelstan", "lang"=>"en", "lat"=>40.572675, "lon"=>-94.542237),
"/US/US-IA/atkins-6471" => array("type"=>"city", "pop"=>1560, "name_ru"=>"Аткинс", "name_en"=>"Atkins", "lang"=>"en", "lat"=>41.997250, "lon"=>-91.859293),
"/US/US-IA/atlantic-6472" => array("type"=>"city", "pop"=>6777, "name_ru"=>"Атлантик", "name_en"=>"Atlantic", "lang"=>"en", "lat"=>41.398249, "lon"=>-95.014841),
"/US/US-IA/auburn-6473" => array("type"=>"city", "pop"=>270, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>42.249944, "lon"=>-94.876943),
"/US/US-IA/audubon-6474" => array("type"=>"city", "pop"=>2110, "name_ru"=>"Одубон", "name_en"=>"Audubon", "lang"=>"en", "lat"=>41.718062, "lon"=>-94.929797),
"/US/US-IA/aurelia-6475" => array("type"=>"city", "pop"=>933, "name_ru"=>"Орелия", "name_en"=>"Aurelia", "lang"=>"en", "lat"=>42.712998, "lon"=>-95.436888),
"/US/US-IA/aurora-6476" => array("type"=>"city", "pop"=>186, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>42.619244, "lon"=>-91.729767),
"/US/US-IA/avoca-6477" => array("type"=>"city", "pop"=>1536, "name_ru"=>"Авока", "name_en"=>"Avoca", "lang"=>"en", "lat"=>41.480800, "lon"=>-95.335484),
"/US/US-IA/ayrshire-6478" => array("type"=>"city", "pop"=>177, "name_ru"=>"Эршир", "name_en"=>"Ayrshire", "lang"=>"en", "lat"=>43.040287, "lon"=>-94.833398),
"/US/US-IA/badger-6479" => array("type"=>"city", "pop"=>578, "name_ru"=>"Баджер", "name_en"=>"Badger", "lang"=>"en", "lat"=>42.612314, "lon"=>-94.142603),
"/US/US-IA/bagley-6480" => array("type"=>"city", "pop"=>315, "name_ru"=>"Багли", "name_en"=>"Bagley", "lang"=>"en", "lat"=>41.845719, "lon"=>-94.429929),
"/US/US-IA/baldwin-6481" => array("type"=>"city", "pop"=>134, "name_ru"=>"Болдуин", "name_en"=>"Baldwin", "lang"=>"en", "lat"=>42.073571, "lon"=>-90.839002),
"/US/US-IA/balltown-6482" => array("type"=>"city", "pop"=>72, "name_ru"=>"Боллтаун", "name_en"=>"Balltown", "lang"=>"en", "lat"=>42.637508, "lon"=>-90.868495),
"/US/US-IA/bancroft-6483" => array("type"=>"city", "pop"=>729, "name_ru"=>"Бэнкрофт", "name_en"=>"Bancroft", "lang"=>"en", "lat"=>43.292502, "lon"=>-94.216825),
"/US/US-IA/bankston-6484" => array("type"=>"city", "pop"=>26, "name_ru"=>"Бэнкстон", "name_en"=>"Bankston", "lang"=>"en", "lat"=>42.508660, "lon"=>-90.960199),
"/US/US-IA/barnes-city-6485" => array("type"=>"city", "pop"=>200, "name_ru"=>"Барнс Сити", "name_en"=>"Barnes City", "lang"=>"en", "lat"=>41.507444, "lon"=>-92.468786),
"/US/US-IA/barnum-6486" => array("type"=>"city", "pop"=>193, "name_ru"=>"Барнум", "name_en"=>"Barnum", "lang"=>"en", "lat"=>42.506687, "lon"=>-94.364329),
"/US/US-IA/bassett-6487" => array("type"=>"city", "pop"=>69, "name_ru"=>"Бассет", "name_en"=>"Bassett", "lang"=>"en", "lat"=>43.062381, "lon"=>-92.515724),
"/US/US-IA/batavia-6488" => array("type"=>"city", "pop"=>473, "name_ru"=>"Батавия", "name_en"=>"Batavia", "lang"=>"en", "lat"=>40.994795, "lon"=>-92.167598),
"/US/US-IA/battle-creek-6489" => array("type"=>"city", "pop"=>651, "name_ru"=>"Бэттл Крик", "name_en"=>"Battle Creek", "lang"=>"en", "lat"=>42.316854, "lon"=>-95.599303),
"/US/US-IA/baxter-6490" => array("type"=>"city", "pop"=>1043, "name_ru"=>"Бэкстер", "name_en"=>"Baxter", "lang"=>"en", "lat"=>41.825780, "lon"=>-93.150671),
"/US/US-IA/bayard-6491" => array("type"=>"city", "pop"=>474, "name_ru"=>"Байяр", "name_en"=>"Bayard", "lang"=>"en", "lat"=>41.852225, "lon"=>-94.558347),
"/US/US-IA/beacon-6492" => array("type"=>"city", "pop"=>527, "name_ru"=>"Бикон", "name_en"=>"Beacon", "lang"=>"en", "lat"=>41.274571, "lon"=>-92.681247),
"/US/US-IA/beaconsfield-6493" => array("type"=>"city", "pop"=>11, "name_ru"=>"Биконсфилд", "name_en"=>"Beaconsfield", "lang"=>"en", "lat"=>40.806979, "lon"=>-94.050417),
"/US/US-IA/beaman-6494" => array("type"=>"city", "pop"=>173, "name_ru"=>"Биман", "name_en"=>"Beaman", "lang"=>"en", "lat"=>42.220012, "lon"=>-92.822785),
"/US/US-IA/beaver-6495" => array("type"=>"city", "pop"=>51, "name_ru"=>"Бивер", "name_en"=>"Beaver", "lang"=>"en", "lat"=>42.037827, "lon"=>-94.140859),
"/US/US-IA/bedford-6496" => array("type"=>"city", "pop"=>1460, "name_ru"=>"Бедфорд", "name_en"=>"Bedford", "lang"=>"en", "lat"=>40.671062, "lon"=>-94.724136),
"/US/US-IA/belle-plaine-6497" => array("type"=>"city", "pop"=>2812, "name_ru"=>"Бель Плейн", "name_en"=>"Belle Plaine", "lang"=>"en", "lat"=>41.896665, "lon"=>-92.274073),
"/US/US-IA/bellevue-6498" => array("type"=>"city", "pop"=>2337, "name_ru"=>"Белвью", "name_en"=>"Bellevue", "lang"=>"en", "lat"=>42.258775, "lon"=>-90.428589),
"/US/US-IA/belmond-6499" => array("type"=>"city", "pop"=>2304, "name_ru"=>"Белмонд", "name_en"=>"Belmond", "lang"=>"en", "lat"=>42.849173, "lon"=>-93.609538),
"/US/US-IA/bennett-6500" => array("type"=>"city", "pop"=>365, "name_ru"=>"Беннет", "name_en"=>"Bennett", "lang"=>"en", "lat"=>41.739725, "lon"=>-90.974259),
"/US/US-IA/benton-6501" => array("type"=>"city", "pop"=>38, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>40.701891, "lon"=>-94.359411),
"/US/US-IA/berkley-6502" => array("type"=>"city", "pop"=>27, "name_ru"=>"Беркли", "name_en"=>"Berkley", "lang"=>"en", "lat"=>41.946140, "lon"=>-94.113584),
"/US/US-IA/bernard-6503" => array("type"=>"city", "pop"=>94, "name_ru"=>"Бернард", "name_en"=>"Bernard", "lang"=>"en", "lat"=>42.312711, "lon"=>-90.830969),
"/US/US-IA/bertram-6504" => array("type"=>"city", "pop"=>287, "name_ru"=>"Бертрам", "name_en"=>"Bertram", "lang"=>"en", "lat"=>41.954864, "lon"=>-91.536569),
"/US/US-IA/bettendorf-6505" => array("type"=>"city", "pop"=>32584, "name_ru"=>"Беттендорт", "name_en"=>"Bettendorf", "lang"=>"en", "lat"=>41.564233, "lon"=>-90.476182),
"/US/US-IA/bevington-6506" => array("type"=>"city", "pop"=>60, "name_ru"=>"Бевингтон", "name_en"=>"Bevington", "lang"=>"en", "lat"=>41.359623, "lon"=>-93.790176),
"/US/US-IA/birmingham-6507" => array("type"=>"city", "pop"=>420, "name_ru"=>"Бирмингем", "name_en"=>"Birmingham", "lang"=>"en", "lat"=>40.878362, "lon"=>-91.947741),
"/US/US-IA/blairsburg-6508" => array("type"=>"city", "pop"=>215, "name_ru"=>"Блэрсбург", "name_en"=>"Blairsburg", "lang"=>"en", "lat"=>42.479530, "lon"=>-93.642608),
"/US/US-IA/blairstown-6509" => array("type"=>"city", "pop"=>685, "name_ru"=>"Блэрстаун", "name_en"=>"Blairstown", "lang"=>"en", "lat"=>41.905844, "lon"=>-92.081764),
"/US/US-IA/blakesburg-6510" => array("type"=>"city", "pop"=>356, "name_ru"=>"Блейксбург", "name_en"=>"Blakesburg", "lang"=>"en", "lat"=>40.961751, "lon"=>-92.635359),
"/US/US-IA/blanchard-6511" => array("type"=>"city", "pop"=>62, "name_ru"=>"Бланчард", "name_en"=>"Blanchard", "lang"=>"en", "lat"=>40.580775, "lon"=>-95.220640),
"/US/US-IA/blencoe-6512" => array("type"=>"city", "pop"=>199, "name_ru"=>"Бленко", "name_en"=>"Blencoe", "lang"=>"en", "lat"=>41.929732, "lon"=>-96.081039),
"/US/US-IA/blockton-6513" => array("type"=>"city", "pop"=>160, "name_ru"=>"Блоктон", "name_en"=>"Blockton", "lang"=>"en", "lat"=>40.616754, "lon"=>-94.477700),
"/US/US-IA/bloomfield-6514" => array("type"=>"city", "pop"=>2582, "name_ru"=>"Блумфилд", "name_en"=>"Bloomfield", "lang"=>"en", "lat"=>40.748118, "lon"=>-92.418183),
"/US/US-IA/blue-grass-6515" => array("type"=>"city", "pop"=>1341, "name_ru"=>"Блу Грасс", "name_en"=>"Blue Grass", "lang"=>"en", "lat"=>41.508579, "lon"=>-90.763649),
"/US/US-IA/bode-6516" => array("type"=>"city", "pop"=>297, "name_ru"=>"Бод", "name_en"=>"Bode", "lang"=>"en", "lat"=>42.868102, "lon"=>-94.286424),
"/US/US-IA/bonaparte-6517" => array("type"=>"city", "pop"=>448, "name_ru"=>"Бонапарт", "name_en"=>"Bonaparte", "lang"=>"en", "lat"=>40.700901, "lon"=>-91.800504),
"/US/US-IA/bondurant-6518" => array("type"=>"city", "pop"=>2714, "name_ru"=>"Бондьюрант", "name_en"=>"Bondurant", "lang"=>"en", "lat"=>41.688368, "lon"=>-93.462180),
"/US/US-IA/boone-6519" => array("type"=>"city", "pop"=>12614, "name_ru"=>"Бун", "name_en"=>"Boone", "lang"=>"en", "lat"=>42.052483, "lon"=>-93.879237),
"/US/US-IA/bouton-6520" => array("type"=>"city", "pop"=>161, "name_ru"=>"Баутон", "name_en"=>"Bouton", "lang"=>"en", "lat"=>41.851715, "lon"=>-94.010756),
"/US/US-IA/boxholm-6521" => array("type"=>"city", "pop"=>200, "name_ru"=>"Боксхолм", "name_en"=>"Boxholm", "lang"=>"en", "lat"=>42.173265, "lon"=>-94.106334),
"/US/US-IA/boyden-6522" => array("type"=>"city", "pop"=>683, "name_ru"=>"Бойден", "name_en"=>"Boyden", "lang"=>"en", "lat"=>43.190230, "lon"=>-96.002019),
"/US/US-IA/braddyville-6523" => array("type"=>"city", "pop"=>163, "name_ru"=>"Брэддивилл", "name_en"=>"Braddyville", "lang"=>"en", "lat"=>40.577937, "lon"=>-95.044436),
"/US/US-IA/bradgate-6524" => array("type"=>"city", "pop"=>92, "name_ru"=>"Брэдгейт", "name_en"=>"Bradgate", "lang"=>"en", "lat"=>42.802596, "lon"=>-94.419303),
"/US/US-IA/brandon-6525" => array("type"=>"city", "pop"=>334, "name_ru"=>"Брэндон", "name_en"=>"Brandon", "lang"=>"en", "lat"=>42.314022, "lon"=>-92.002410),
"/US/US-IA/brayton-6526" => array("type"=>"city", "pop"=>132, "name_ru"=>"Брэйтон", "name_en"=>"Brayton", "lang"=>"en", "lat"=>41.542560, "lon"=>-94.925599),
"/US/US-IA/breda-6527" => array("type"=>"city", "pop"=>449, "name_ru"=>"Бреда", "name_en"=>"Breda", "lang"=>"en", "lat"=>42.183187, "lon"=>-94.977720),
"/US/US-IA/bridgewater-6528" => array("type"=>"city", "pop"=>161, "name_ru"=>"Бриджуотер", "name_en"=>"Bridgewater", "lang"=>"en", "lat"=>41.246876, "lon"=>-94.667446),
"/US/US-IA/brighton-6529" => array("type"=>"city", "pop"=>683, "name_ru"=>"Брайтон", "name_en"=>"Brighton", "lang"=>"en", "lat"=>41.174270, "lon"=>-91.820629),
"/US/US-IA/bristow-6530" => array("type"=>"city", "pop"=>190, "name_ru"=>"Бристоу", "name_en"=>"Bristow", "lang"=>"en", "lat"=>42.773536, "lon"=>-92.906731),
"/US/US-IA/britt-6531" => array("type"=>"city", "pop"=>1920, "name_ru"=>"Бритт", "name_en"=>"Britt", "lang"=>"en", "lat"=>43.097316, "lon"=>-93.803301),
"/US/US-IA/bronson-6532" => array("type"=>"city", "pop"=>283, "name_ru"=>"Бронсон", "name_en"=>"Bronson", "lang"=>"en", "lat"=>42.409458, "lon"=>-96.211263),
"/US/US-IA/brooklyn-6533" => array("type"=>"city", "pop"=>1355, "name_ru"=>"Бруклин", "name_en"=>"Brooklyn", "lang"=>"en", "lat"=>41.731213, "lon"=>-92.444583),
"/US/US-IA/brunsville-6534" => array("type"=>"city", "pop"=>139, "name_ru"=>"Брансвилл", "name_en"=>"Brunsville", "lang"=>"en", "lat"=>42.811308, "lon"=>-96.266385),
"/US/US-IA/buck-grove-6536" => array("type"=>"city", "pop"=>51, "name_ru"=>"Бак Гров", "name_en"=>"Buck Grove", "lang"=>"en", "lat"=>41.918658, "lon"=>-95.396989),
"/US/US-IA/buckeye-6535" => array("type"=>"city", "pop"=>96, "name_ru"=>"Бакай", "name_en"=>"Buckeye", "lang"=>"en", "lat"=>42.418393, "lon"=>-93.375935),
"/US/US-IA/buffalo-6537" => array("type"=>"city", "pop"=>1342, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>41.460250, "lon"=>-90.718190),
"/US/US-IA/buffalo-center-6538" => array("type"=>"city", "pop"=>856, "name_ru"=>"Буффало Сентер", "name_en"=>"Buffalo Center", "lang"=>"en", "lat"=>43.389788, "lon"=>-93.943072),
"/US/US-IA/burlington-6539" => array("type"=>"city", "pop"=>25327, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>40.808705, "lon"=>-91.122899),
"/US/US-IA/burt-6540" => array("type"=>"city", "pop"=>497, "name_ru"=>"Берт", "name_en"=>"Burt", "lang"=>"en", "lat"=>43.199175, "lon"=>-94.221368),
"/US/US-IA/bussey-6541" => array("type"=>"city", "pop"=>436, "name_ru"=>"Басси", "name_en"=>"Bussey", "lang"=>"en", "lat"=>41.205925, "lon"=>-92.884158),
"/US/US-IA/calamus-6542" => array("type"=>"city", "pop"=>395, "name_ru"=>"Каламус", "name_en"=>"Calamus", "lang"=>"en", "lat"=>41.826651, "lon"=>-90.759674),
"/US/US-IA/callender-6543" => array("type"=>"city", "pop"=>425, "name_ru"=>"Каллендер", "name_en"=>"Callender", "lang"=>"en", "lat"=>42.362183, "lon"=>-94.295575),
"/US/US-IA/calmar-6544" => array("type"=>"city", "pop"=>1017, "name_ru"=>"Калмар", "name_en"=>"Calmar", "lang"=>"en", "lat"=>43.179397, "lon"=>-91.866900),
"/US/US-IA/calumet-6545" => array("type"=>"city", "pop"=>152, "name_ru"=>"Калумет", "name_en"=>"Calumet", "lang"=>"en", "lat"=>42.944947, "lon"=>-95.550386),
"/US/US-IA/camanche-6546" => array("type"=>"city", "pop"=>4280, "name_ru"=>"Каманч", "name_en"=>"Camanche", "lang"=>"en", "lat"=>41.793908, "lon"=>-90.275012),
"/US/US-IA/cambridge-6547" => array("type"=>"city", "pop"=>809, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>41.899502, "lon"=>-93.530889),
"/US/US-IA/cantril-6548" => array("type"=>"city", "pop"=>255, "name_ru"=>"Кантрил", "name_en"=>"Cantril", "lang"=>"en", "lat"=>40.643059, "lon"=>-92.068830),
"/US/US-IA/carbon-6549" => array("type"=>"city", "pop"=>22, "name_ru"=>"Карбон", "name_en"=>"Carbon", "lang"=>"en", "lat"=>41.049260, "lon"=>-94.823511),
"/US/US-IA/carlisle-6550" => array("type"=>"city", "pop"=>3693, "name_ru"=>"Карлайл", "name_en"=>"Carlisle", "lang"=>"en", "lat"=>41.512537, "lon"=>-93.485192),
"/US/US-IA/carpenter-6551" => array("type"=>"city", "pop"=>128, "name_ru"=>"Карпентер", "name_en"=>"Carpenter", "lang"=>"en", "lat"=>43.414922, "lon"=>-93.017396),
"/US/US-IA/carroll-6552" => array("type"=>"city", "pop"=>9985, "name_ru"=>"Кэррол", "name_en"=>"Carroll", "lang"=>"en", "lat"=>42.069183, "lon"=>-94.864690),
"/US/US-IA/carson-6553" => array("type"=>"city", "pop"=>725, "name_ru"=>"Карсон", "name_en"=>"Carson", "lang"=>"en", "lat"=>41.236650, "lon"=>-95.416299),
"/US/US-IA/carter-lake-6554" => array("type"=>"city", "pop"=>3268, "name_ru"=>"Картер Лейк", "name_en"=>"Carter Lake", "lang"=>"en", "lat"=>41.288737, "lon"=>-95.916053),
"/US/US-IA/cascade-6555" => array("type"=>"city", "pop"=>2118, "name_ru"=>"Каскейд", "name_en"=>"Cascade", "lang"=>"en", "lat"=>42.299057, "lon"=>-91.009853),
"/US/US-IA/casey-6556" => array("type"=>"city", "pop"=>393, "name_ru"=>"Кейси", "name_en"=>"Casey", "lang"=>"en", "lat"=>41.507291, "lon"=>-94.521479),
"/US/US-IA/castalia-6557" => array("type"=>"city", "pop"=>165, "name_ru"=>"Касталия", "name_en"=>"Castalia", "lang"=>"en", "lat"=>43.111852, "lon"=>-91.676396),
"/US/US-IA/castana-6558" => array("type"=>"city", "pop"=>159, "name_ru"=>"Кастана", "name_en"=>"Castana", "lang"=>"en", "lat"=>42.073602, "lon"=>-95.909394),
"/US/US-IA/cedar-falls-6559" => array("type"=>"city", "pop"=>38059, "name_ru"=>"Сидар Фоллс", "name_en"=>"Cedar Falls", "lang"=>"en", "lat"=>42.520576, "lon"=>-92.452807),
"/US/US-IA/cedar-rapids-6560" => array("type"=>"city", "pop"=>128056, "name_ru"=>"Сидар Рэпидс", "name_en"=>"Cedar Rapids", "lang"=>"en", "lat"=>41.972967, "lon"=>-91.669578),
"/US/US-IA/center-junction-6561" => array("type"=>"city", "pop"=>114, "name_ru"=>"Сентер Джанкшен", "name_en"=>"Center Junction", "lang"=>"en", "lat"=>42.114372, "lon"=>-91.089337),
"/US/US-IA/center-point-6562" => array("type"=>"city", "pop"=>2342, "name_ru"=>"Сентер Поинт", "name_en"=>"Center Point", "lang"=>"en", "lat"=>42.184970, "lon"=>-91.776720),
"/US/US-IA/centerville-6563" => array("type"=>"city", "pop"=>5442, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>40.730829, "lon"=>-92.874342),
"/US/US-IA/central-city-6564" => array("type"=>"city", "pop"=>1183, "name_ru"=>"Сентрал Сити", "name_en"=>"Central City", "lang"=>"en", "lat"=>42.203452, "lon"=>-91.523609),
"/US/US-IA/centralia-6565" => array("type"=>"city", "pop"=>99, "name_ru"=>"Сентралия", "name_en"=>"Centralia", "lang"=>"en", "lat"=>42.472312, "lon"=>-90.836813),
"/US/US-IA/chariton-6566" => array("type"=>"city", "pop"=>4419, "name_ru"=>"Чаритон", "name_en"=>"Chariton", "lang"=>"en", "lat"=>41.017594, "lon"=>-93.309862),
"/US/US-IA/charles-city-6567" => array("type"=>"city", "pop"=>7470, "name_ru"=>"Чарльз Сити", "name_en"=>"Charles City", "lang"=>"en", "lat"=>43.067274, "lon"=>-92.675022),
"/US/US-IA/charlotte-6568" => array("type"=>"city", "pop"=>438, "name_ru"=>"Шарлотт", "name_en"=>"Charlotte", "lang"=>"en", "lat"=>41.961864, "lon"=>-90.467524),
"/US/US-IA/charter-oak-6569" => array("type"=>"city", "pop"=>500, "name_ru"=>"Чартер Оук", "name_en"=>"Charter Oak", "lang"=>"en", "lat"=>42.067909, "lon"=>-95.589110),
"/US/US-IA/chatsworth-6570" => array("type"=>"city", "pop"=>88, "name_ru"=>"Чатсворт", "name_en"=>"Chatsworth", "lang"=>"en", "lat"=>42.916392, "lon"=>-96.514712),
"/US/US-IA/chelsea-6571" => array("type"=>"city", "pop"=>272, "name_ru"=>"Челси", "name_en"=>"Chelsea", "lang"=>"en", "lat"=>41.920797, "lon"=>-92.395199),
"/US/US-IA/cherokee-6572" => array("type"=>"city", "pop"=>4697, "name_ru"=>"Чероки", "name_en"=>"Cherokee", "lang"=>"en", "lat"=>42.750593, "lon"=>-95.552486),
"/US/US-IA/chester-6573" => array("type"=>"city", "pop"=>137, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>43.490051, "lon"=>-92.363395),
"/US/US-IA/chillicothe-6574" => array("type"=>"city", "pop"=>85, "name_ru"=>"Чилликот", "name_en"=>"Chillicothe", "lang"=>"en", "lat"=>41.086033, "lon"=>-92.529456),
"/US/US-IA/churdan-6575" => array("type"=>"city", "pop"=>361, "name_ru"=>"Чердан", "name_en"=>"Churdan", "lang"=>"en", "lat"=>42.154688, "lon"=>-94.477727),
"/US/US-IA/cincinnati-6576" => array("type"=>"city", "pop"=>391, "name_ru"=>"Цинциннати", "name_en"=>"Cincinnati", "lang"=>"en", "lat"=>40.631751, "lon"=>-92.921791),
"/US/US-IA/clare-6577" => array("type"=>"city", "pop"=>174, "name_ru"=>"Клэр", "name_en"=>"Clare", "lang"=>"en", "lat"=>42.587695, "lon"=>-94.344625),
"/US/US-IA/clarence-6578" => array("type"=>"city", "pop"=>966, "name_ru"=>"Клэренс", "name_en"=>"Clarence", "lang"=>"en", "lat"=>41.885899, "lon"=>-91.058109),
"/US/US-IA/clarinda-6579" => array("type"=>"city", "pop"=>5493, "name_ru"=>"Кларинда", "name_en"=>"Clarinda", "lang"=>"en", "lat"=>40.738649, "lon"=>-95.035693),
"/US/US-IA/clarion-6580" => array("type"=>"city", "pop"=>2718, "name_ru"=>"Клэрион", "name_en"=>"Clarion", "lang"=>"en", "lat"=>42.732054, "lon"=>-93.732131),
"/US/US-IA/clarksville-6581" => array("type"=>"city", "pop"=>1336, "name_ru"=>"Кларксвилл", "name_en"=>"Clarksville", "lang"=>"en", "lat"=>42.780192, "lon"=>-92.668056),
"/US/US-IA/clayton-6582" => array("type"=>"city", "pop"=>51, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>42.902263, "lon"=>-91.150027),
"/US/US-IA/clear-lake-6584" => array("type"=>"city", "pop"=>7820, "name_ru"=>"Клир Лейк", "name_en"=>"Clear Lake", "lang"=>"en", "lat"=>43.135055, "lon"=>-93.373584),
"/US/US-IA/clearfield-6583" => array("type"=>"city", "pop"=>352, "name_ru"=>"Клирфилд", "name_en"=>"Clearfield", "lang"=>"en", "lat"=>40.802101, "lon"=>-94.476040),
"/US/US-IA/cleghorn-6585" => array("type"=>"city", "pop"=>226, "name_ru"=>"Клегхорн", "name_en"=>"Cleghorn", "lang"=>"en", "lat"=>42.811587, "lon"=>-95.713482),
"/US/US-IA/clemons-6586" => array("type"=>"city", "pop"=>147, "name_ru"=>"Клемонс", "name_en"=>"Clemons", "lang"=>"en", "lat"=>42.113243, "lon"=>-93.158212),
"/US/US-IA/clermont-6587" => array("type"=>"city", "pop"=>688, "name_ru"=>"Клермонт", "name_en"=>"Clermont", "lang"=>"en", "lat"=>43.000289, "lon"=>-91.653312),
"/US/US-IA/clinton-6588" => array("type"=>"city", "pop"=>26407, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>41.843216, "lon"=>-90.232994),
"/US/US-IA/clio-6589" => array("type"=>"city", "pop"=>71, "name_ru"=>"Клио", "name_en"=>"Clio", "lang"=>"en", "lat"=>40.634716, "lon"=>-93.451849),
"/US/US-IA/clive-6590" => array("type"=>"city", "pop"=>15234, "name_ru"=>"Клайв", "name_en"=>"Clive", "lang"=>"en", "lat"=>41.607664, "lon"=>-93.769483),
"/US/US-IA/clutier-6591" => array("type"=>"city", "pop"=>222, "name_ru"=>"Клатьер", "name_en"=>"Clutier", "lang"=>"en", "lat"=>42.078788, "lon"=>-92.403055),
"/US/US-IA/coburg-6592" => array("type"=>"city", "pop"=>29, "name_ru"=>"Коберг", "name_en"=>"Coburg", "lang"=>"en", "lat"=>40.917797, "lon"=>-95.265606),
"/US/US-IA/coggon-6593" => array("type"=>"city", "pop"=>714, "name_ru"=>"Коггон", "name_en"=>"Coggon", "lang"=>"en", "lat"=>42.279439, "lon"=>-91.530924),
"/US/US-IA/coin-6594" => array("type"=>"city", "pop"=>198, "name_ru"=>"Коин", "name_en"=>"Coin", "lang"=>"en", "lat"=>40.655885, "lon"=>-95.235226),
"/US/US-IA/colesburg-6595" => array("type"=>"city", "pop"=>376, "name_ru"=>"Коулсберг", "name_en"=>"Colesburg", "lang"=>"en", "lat"=>42.638542, "lon"=>-91.200884),
"/US/US-IA/colfax-6596" => array("type"=>"city", "pop"=>2212, "name_ru"=>"Колфакс", "name_en"=>"Colfax", "lang"=>"en", "lat"=>41.676158, "lon"=>-93.240186),
"/US/US-IA/college-springs-6597" => array("type"=>"city", "pop"=>227, "name_ru"=>"Коллидж Спрингс", "name_en"=>"College Springs", "lang"=>"en", "lat"=>40.620886, "lon"=>-95.121685),
"/US/US-IA/collins-6598" => array("type"=>"city", "pop"=>481, "name_ru"=>"Коллинс", "name_en"=>"Collins", "lang"=>"en", "lat"=>41.902458, "lon"=>-93.305433),
"/US/US-IA/colo-6599" => array("type"=>"city", "pop"=>876, "name_ru"=>"Коло", "name_en"=>"Colo", "lang"=>"en", "lat"=>42.015871, "lon"=>-93.318473),
"/US/US-IA/columbus-city-6600" => array("type"=>"city", "pop"=>349, "name_ru"=>"Коламбус Сити", "name_en"=>"Columbus City", "lang"=>"en", "lat"=>41.259309, "lon"=>-91.375026),
"/US/US-IA/columbus-junction-6601" => array("type"=>"city", "pop"=>1827, "name_ru"=>"Коламбус Джанкшен", "name_en"=>"Columbus Junction", "lang"=>"en", "lat"=>41.279400, "lon"=>-91.364650),
"/US/US-IA/colwell-6602" => array("type"=>"city", "pop"=>74, "name_ru"=>"Колуэлл", "name_en"=>"Colwell", "lang"=>"en", "lat"=>43.158088, "lon"=>-92.591457),
"/US/US-IA/conesville-6603" => array("type"=>"city", "pop"=>439, "name_ru"=>"Коунсвилл", "name_en"=>"Conesville", "lang"=>"en", "lat"=>41.380372, "lon"=>-91.348055),
"/US/US-IA/conrad-6604" => array("type"=>"city", "pop"=>992, "name_ru"=>"Конрад", "name_en"=>"Conrad", "lang"=>"en", "lat"=>42.223977, "lon"=>-92.873585),
"/US/US-IA/conway-6605" => array("type"=>"city", "pop"=>58, "name_ru"=>"Конвэй", "name_en"=>"Conway", "lang"=>"en", "lat"=>40.749869, "lon"=>-94.619446),
"/US/US-IA/coon-rapids-6606" => array("type"=>"city", "pop"=>1256, "name_ru"=>"Кун Рэпидс", "name_en"=>"Coon Rapids", "lang"=>"en", "lat"=>41.872477, "lon"=>-94.676531),
"/US/US-IA/coppock-6607" => array("type"=>"city", "pop"=>64, "name_ru"=>"Коппок", "name_en"=>"Coppock", "lang"=>"en", "lat"=>41.164169, "lon"=>-91.714680),
"/US/US-IA/coralville-6608" => array("type"=>"city", "pop"=>18478, "name_ru"=>"Коралвилл", "name_en"=>"Coralville", "lang"=>"en", "lat"=>41.690950, "lon"=>-91.600723),
"/US/US-IA/corning-6609" => array("type"=>"city", "pop"=>1600, "name_ru"=>"Корнинг", "name_en"=>"Corning", "lang"=>"en", "lat"=>40.993903, "lon"=>-94.739269),
"/US/US-IA/correctionville-6610" => array("type"=>"city", "pop"=>846, "name_ru"=>"Коррекшенвилл", "name_en"=>"Correctionville", "lang"=>"en", "lat"=>42.478211, "lon"=>-95.783399),
"/US/US-IA/corwith-6611" => array("type"=>"city", "pop"=>313, "name_ru"=>"Корвит", "name_en"=>"Corwith", "lang"=>"en", "lat"=>42.989300, "lon"=>-93.958609),
"/US/US-IA/corydon-6612" => array("type"=>"city", "pop"=>1459, "name_ru"=>"Коридон", "name_en"=>"Corydon", "lang"=>"en", "lat"=>40.759053, "lon"=>-93.317547),
"/US/US-IA/cotter-6613" => array("type"=>"city", "pop"=>47, "name_ru"=>"Коттер", "name_en"=>"Cotter", "lang"=>"en", "lat"=>41.292589, "lon"=>-91.470474),
"/US/US-IA/coulter-6614" => array("type"=>"city", "pop"=>250, "name_ru"=>"Колтер", "name_en"=>"Coulter", "lang"=>"en", "lat"=>42.733308, "lon"=>-93.370312),
"/US/US-IA/council-bluffs-6615" => array("type"=>"city", "pop"=>59536, "name_ru"=>"Каунсил Блаффс", "name_en"=>"Council Bluffs", "lang"=>"en", "lat"=>41.239950, "lon"=>-95.859499),
"/US/US-IA/craig-6616" => array("type"=>"city", "pop"=>96, "name_ru"=>"Крейг", "name_en"=>"Craig", "lang"=>"en", "lat"=>42.895546, "lon"=>-96.310123),
"/US/US-IA/crawfordsville-6617" => array("type"=>"city", "pop"=>301, "name_ru"=>"Крофордсвилл", "name_en"=>"Crawfordsville", "lang"=>"en", "lat"=>41.214439, "lon"=>-91.535923),
"/US/US-IA/crescent-6618" => array("type"=>"city", "pop"=>512, "name_ru"=>"Кресент", "name_en"=>"Crescent", "lang"=>"en", "lat"=>41.366300, "lon"=>-95.858799),
"/US/US-IA/cresco-6619" => array("type"=>"city", "pop"=>3730, "name_ru"=>"Креско", "name_en"=>"Cresco", "lang"=>"en", "lat"=>43.371768, "lon"=>-92.115655),
"/US/US-IA/creston-6620" => array("type"=>"city", "pop"=>7594, "name_ru"=>"Крестон", "name_en"=>"Creston", "lang"=>"en", "lat"=>41.059050, "lon"=>-94.364142),
"/US/US-IA/cromwell-6621" => array("type"=>"city", "pop"=>120, "name_ru"=>"Кромуэлл", "name_en"=>"Cromwell", "lang"=>"en", "lat"=>41.039969, "lon"=>-94.461925),
"/US/US-IA/crystal-lake-6622" => array("type"=>"city", "pop"=>259, "name_ru"=>"Кристал Лэйк", "name_en"=>"Crystal Lake", "lang"=>"en", "lat"=>43.223612, "lon"=>-93.792674),
"/US/US-IA/cumberland-6623" => array("type"=>"city", "pop"=>255, "name_ru"=>"Камберленд", "name_en"=>"Cumberland", "lang"=>"en", "lat"=>41.273685, "lon"=>-94.869662),
"/US/US-IA/cumming-6624" => array("type"=>"city", "pop"=>309, "name_ru"=>"Камминг", "name_en"=>"Cumming", "lang"=>"en", "lat"=>41.484693, "lon"=>-93.761756),
"/US/US-IA/curlew-6625" => array("type"=>"city", "pop"=>57, "name_ru"=>"Керлью", "name_en"=>"Curlew", "lang"=>"en", "lat"=>42.979468, "lon"=>-94.737206),
"/US/US-IA/cushing-6626" => array("type"=>"city", "pop"=>237, "name_ru"=>"Кушинг", "name_en"=>"Cushing", "lang"=>"en", "lat"=>42.465100, "lon"=>-95.676053),
"/US/US-IA/cylinder-6627" => array("type"=>"city", "pop"=>98, "name_ru"=>"Силиндер", "name_en"=>"Cylinder", "lang"=>"en", "lat"=>43.090414, "lon"=>-94.551259),
"/US/US-IA/dakota-city-6628" => array("type"=>"city", "pop"=>824, "name_ru"=>"Дакота Сити", "name_en"=>"Dakota City", "lang"=>"en", "lat"=>42.722221, "lon"=>-94.196903),
"/US/US-IA/dallas-center-6629" => array("type"=>"city", "pop"=>1734, "name_ru"=>"Даллас Сентер", "name_en"=>"Dallas Center", "lang"=>"en", "lat"=>41.684666, "lon"=>-93.984115),
"/US/US-IA/dana-6630" => array("type"=>"city", "pop"=>76, "name_ru"=>"Дана", "name_en"=>"Dana", "lang"=>"en", "lat"=>42.106674, "lon"=>-94.240134),
"/US/US-IA/danbury-6631" => array("type"=>"city", "pop"=>327, "name_ru"=>"Данбери", "name_en"=>"Danbury", "lang"=>"en", "lat"=>42.237168, "lon"=>-95.721458),
"/US/US-IA/danville-6632" => array("type"=>"city", "pop"=>854, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>40.861152, "lon"=>-91.314034),
"/US/US-IA/davenport-6633" => array("type"=>"city", "pop"=>100827, "name_ru"=>"Дэйвенпорт", "name_en"=>"Davenport", "lang"=>"en", "lat"=>41.556750, "lon"=>-90.604224),
"/US/US-IA/davis-city-6634" => array("type"=>"city", "pop"=>253, "name_ru"=>"Дэвис Сити", "name_en"=>"Davis City", "lang"=>"en", "lat"=>40.640222, "lon"=>-93.811928),
"/US/US-IA/dawson-6635" => array("type"=>"city", "pop"=>163, "name_ru"=>"Доусон", "name_en"=>"Dawson", "lang"=>"en", "lat"=>41.843119, "lon"=>-94.219922),
"/US/US-IA/dayton-6636" => array("type"=>"city", "pop"=>817, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>42.261748, "lon"=>-94.071497),
"/US/US-IA/de-soto-6652" => array("type"=>"city", "pop"=>1242, "name_ru"=>"Де Сото", "name_en"=>"De Soto", "lang"=>"en", "lat"=>41.534617, "lon"=>-94.008462),
"/US/US-IA/de-witt-6653" => array("type"=>"city", "pop"=>5260, "name_ru"=>"Де Витт", "name_en"=>"De Witt", "lang"=>"en", "lat"=>41.821639, "lon"=>-90.546833),
"/US/US-IA/decatur-city-6637" => array("type"=>"city", "pop"=>189, "name_ru"=>"Декатур Сити", "name_en"=>"Decatur City", "lang"=>"en", "lat"=>40.742269, "lon"=>-93.832512),
"/US/US-IA/decorah-6638" => array("type"=>"city", "pop"=>7906, "name_ru"=>"Декора", "name_en"=>"Decorah", "lang"=>"en", "lat"=>43.305258, "lon"=>-91.793297),
"/US/US-IA/dedham-6639" => array("type"=>"city", "pop"=>229, "name_ru"=>"Дедхэм", "name_en"=>"Dedham", "lang"=>"en", "lat"=>41.908665, "lon"=>-94.822571),
"/US/US-IA/deep-river-6640" => array("type"=>"city", "pop"=>272, "name_ru"=>"Дип Ривер", "name_en"=>"Deep River", "lang"=>"en", "lat"=>41.581479, "lon"=>-92.372432),
"/US/US-IA/defiance-6641" => array("type"=>"city", "pop"=>314, "name_ru"=>"Дефианс", "name_en"=>"Defiance", "lang"=>"en", "lat"=>41.825447, "lon"=>-95.339261),
"/US/US-IA/delaware-6642" => array("type"=>"city", "pop"=>181, "name_ru"=>"Делавэр", "name_en"=>"Delaware", "lang"=>"en", "lat"=>42.471864, "lon"=>-91.340136),
"/US/US-IA/delhi-6643" => array("type"=>"city", "pop"=>493, "name_ru"=>"Дели", "name_en"=>"Delhi", "lang"=>"en", "lat"=>42.430013, "lon"=>-91.331502),
"/US/US-IA/delmar-6644" => array("type"=>"city", "pop"=>488, "name_ru"=>"Делмар", "name_en"=>"Delmar", "lang"=>"en", "lat"=>41.999540, "lon"=>-90.607994),
"/US/US-IA/deloit-6645" => array("type"=>"city", "pop"=>280, "name_ru"=>"Делойт", "name_en"=>"Deloit", "lang"=>"en", "lat"=>42.097205, "lon"=>-95.317532),
"/US/US-IA/delphos-6646" => array("type"=>"city", "pop"=>25, "name_ru"=>"Делфос", "name_en"=>"Delphos", "lang"=>"en", "lat"=>40.663181, "lon"=>-94.339283),
"/US/US-IA/delta-6647" => array("type"=>"city", "pop"=>394, "name_ru"=>"Дельта", "name_en"=>"Delta", "lang"=>"en", "lat"=>41.323074, "lon"=>-92.329448),
"/US/US-IA/denison-6648" => array("type"=>"city", "pop"=>7184, "name_ru"=>"Денисон", "name_en"=>"Denison", "lang"=>"en", "lat"=>42.018298, "lon"=>-95.348941),
"/US/US-IA/denver-6649" => array("type"=>"city", "pop"=>1635, "name_ru"=>"Денвер", "name_en"=>"Denver", "lang"=>"en", "lat"=>42.671172, "lon"=>-92.334056),
"/US/US-IA/derby-6650" => array("type"=>"city", "pop"=>132, "name_ru"=>"Дерби", "name_en"=>"Derby", "lang"=>"en", "lat"=>40.929931, "lon"=>-93.456740),
"/US/US-IA/des-moines-6651" => array("type"=>"city", "pop"=>197052, "name_ru"=>"Дес Мойн", "name_en"=>"Des Moines", "lang"=>"en", "lat"=>41.576738, "lon"=>-93.617405),
"/US/US-IA/dexter-6654" => array("type"=>"city", "pop"=>887, "name_ru"=>"Декстер", "name_en"=>"Dexter", "lang"=>"en", "lat"=>41.515629, "lon"=>-94.226878),
"/US/US-IA/diagonal-6655" => array("type"=>"city", "pop"=>298, "name_ru"=>"Диагонал", "name_en"=>"Diagonal", "lang"=>"en", "lat"=>40.810774, "lon"=>-94.341239),
"/US/US-IA/dickens-6656" => array("type"=>"city", "pop"=>187, "name_ru"=>"Дикенс", "name_en"=>"Dickens", "lang"=>"en", "lat"=>43.133367, "lon"=>-95.022069),
"/US/US-IA/dike-6657" => array("type"=>"city", "pop"=>1142, "name_ru"=>"Дайк", "name_en"=>"Dike", "lang"=>"en", "lat"=>42.463100, "lon"=>-92.630140),
"/US/US-IA/dixon-6658" => array("type"=>"city", "pop"=>254, "name_ru"=>"Диксон", "name_en"=>"Dixon", "lang"=>"en", "lat"=>41.742774, "lon"=>-90.781439),
"/US/US-IA/dolliver-6659" => array("type"=>"city", "pop"=>61, "name_ru"=>"Долливер", "name_en"=>"Dolliver", "lang"=>"en", "lat"=>43.464865, "lon"=>-94.614620),
"/US/US-IA/donahue-6660" => array("type"=>"city", "pop"=>294, "name_ru"=>"Донахью", "name_en"=>"Donahue", "lang"=>"en", "lat"=>41.691883, "lon"=>-90.675050),
"/US/US-IA/donnan-6661" => array("type"=>"city", "pop"=>7, "name_ru"=>"Доннан", "name_en"=>"Donnan", "lang"=>"en", "lat"=>42.901018, "lon"=>-91.869312),
"/US/US-IA/donnellson-6662" => array("type"=>"city", "pop"=>874, "name_ru"=>"Доннелсон", "name_en"=>"Donnellson", "lang"=>"en", "lat"=>40.643362, "lon"=>-91.564923),
"/US/US-IA/doon-6663" => array("type"=>"city", "pop"=>522, "name_ru"=>"Дун", "name_en"=>"Doon", "lang"=>"en", "lat"=>43.278749, "lon"=>-96.231620),
"/US/US-IA/dougherty-6664" => array("type"=>"city", "pop"=>76, "name_ru"=>"Доэрти", "name_en"=>"Dougherty", "lang"=>"en", "lat"=>42.922570, "lon"=>-93.036294),
"/US/US-IA/dow-city-6665" => array("type"=>"city", "pop"=>456, "name_ru"=>"Доу Сити", "name_en"=>"Dow City", "lang"=>"en", "lat"=>41.928207, "lon"=>-95.494197),
"/US/US-IA/dows-6666" => array("type"=>"city", "pop"=>555, "name_ru"=>"Доус", "name_en"=>"Dows", "lang"=>"en", "lat"=>42.658736, "lon"=>-93.501991),
"/US/US-IA/drakesville-6667" => array("type"=>"city", "pop"=>217, "name_ru"=>"Дрейксвилл", "name_en"=>"Drakesville", "lang"=>"en", "lat"=>40.797991, "lon"=>-92.480916),
"/US/US-IA/dubuque-6668" => array("type"=>"city", "pop"=>57250, "name_ru"=>"Дабек", "name_en"=>"Dubuque", "lang"=>"en", "lat"=>42.504500, "lon"=>-90.690008),
"/US/US-IA/dumont-6669" => array("type"=>"city", "pop"=>633, "name_ru"=>"Думонт", "name_en"=>"Dumont", "lang"=>"en", "lat"=>42.751586, "lon"=>-92.973408),
"/US/US-IA/duncombe-6670" => array("type"=>"city", "pop"=>456, "name_ru"=>"Данкомб", "name_en"=>"Duncombe", "lang"=>"en", "lat"=>42.469925, "lon"=>-93.996030),
"/US/US-IA/dundee-6671" => array("type"=>"city", "pop"=>167, "name_ru"=>"Данди", "name_en"=>"Dundee", "lang"=>"en", "lat"=>42.579662, "lon"=>-91.546948),
"/US/US-IA/dunkerton-6672" => array("type"=>"city", "pop"=>812, "name_ru"=>"Данкертон", "name_en"=>"Dunkerton", "lang"=>"en", "lat"=>42.568344, "lon"=>-92.159896),
"/US/US-IA/dunlap-6673" => array("type"=>"city", "pop"=>1089, "name_ru"=>"Данлап", "name_en"=>"Dunlap", "lang"=>"en", "lat"=>41.852848, "lon"=>-95.599581),
"/US/US-IA/durango-6674" => array("type"=>"city", "pop"=>22, "name_ru"=>"Дуранго", "name_en"=>"Durango", "lang"=>"en", "lat"=>42.561555, "lon"=>-90.773458),
"/US/US-IA/durant-6675" => array("type"=>"city", "pop"=>1713, "name_ru"=>"Дурант", "name_en"=>"Durant", "lang"=>"en", "lat"=>41.600946, "lon"=>-90.909035),
"/US/US-IA/dyersville-6676" => array("type"=>"city", "pop"=>4204, "name_ru"=>"Дайерсвилл", "name_en"=>"Dyersville", "lang"=>"en", "lat"=>42.480441, "lon"=>-91.118251),
"/US/US-IA/dysart-6677" => array("type"=>"city", "pop"=>1277, "name_ru"=>"Дайсарт", "name_en"=>"Dysart", "lang"=>"en", "lat"=>42.172160, "lon"=>-92.309281),
"/US/US-IA/eagle-grove-6678" => array("type"=>"city", "pop"=>3310, "name_ru"=>"Игл Гров", "name_en"=>"Eagle Grove", "lang"=>"en", "lat"=>42.667339, "lon"=>-93.901318),
"/US/US-IA/earlham-6679" => array("type"=>"city", "pop"=>1347, "name_ru"=>"Эрлхэм", "name_en"=>"Earlham", "lang"=>"en", "lat"=>41.492189, "lon"=>-94.121966),
"/US/US-IA/earling-6680" => array("type"=>"city", "pop"=>431, "name_ru"=>"Эрлинг", "name_en"=>"Earling", "lang"=>"en", "lat"=>41.776102, "lon"=>-95.419658),
"/US/US-IA/earlville-6681" => array("type"=>"city", "pop"=>831, "name_ru"=>"Эрлвилл", "name_en"=>"Earlville", "lang"=>"en", "lat"=>42.483008, "lon"=>-91.269713),
"/US/US-IA/early-6682" => array("type"=>"city", "pop"=>520, "name_ru"=>"Эрли", "name_en"=>"Early", "lang"=>"en", "lat"=>42.460870, "lon"=>-95.152520),
"/US/US-IA/east-peru-6683" => array("type"=>"city", "pop"=>138, "name_ru"=>"Ист Перу", "name_en"=>"East Peru", "lang"=>"en", "lat"=>41.227475, "lon"=>-93.927599),
"/US/US-IA/eddyville-6684" => array("type"=>"city", "pop"=>1072, "name_ru"=>"Эддивилл", "name_en"=>"Eddyville", "lang"=>"en", "lat"=>41.158673, "lon"=>-92.629701),
"/US/US-IA/edgewood-6685" => array("type"=>"city", "pop"=>870, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>42.644859, "lon"=>-91.401652),
"/US/US-IA/elberon-6686" => array("type"=>"city", "pop"=>231, "name_ru"=>"Элберон", "name_en"=>"Elberon", "lang"=>"en", "lat"=>42.005418, "lon"=>-92.315827),
"/US/US-IA/eldon-6687" => array("type"=>"city", "pop"=>973, "name_ru"=>"Элдон", "name_en"=>"Eldon", "lang"=>"en", "lat"=>40.917710, "lon"=>-92.218179),
"/US/US-IA/eldora-6688" => array("type"=>"city", "pop"=>2734, "name_ru"=>"Элдора", "name_en"=>"Eldora", "lang"=>"en", "lat"=>42.360173, "lon"=>-93.101123),
"/US/US-IA/eldridge-6689" => array("type"=>"city", "pop"=>4880, "name_ru"=>"Элдридж", "name_en"=>"Eldridge", "lang"=>"en", "lat"=>41.635800, "lon"=>-90.573335),
"/US/US-IA/elgin-6690" => array("type"=>"city", "pop"=>619, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>42.955972, "lon"=>-91.634437),
"/US/US-IA/elk-horn-6693" => array("type"=>"city", "pop"=>609, "name_ru"=>"Элк Хорн", "name_en"=>"Elk Horn", "lang"=>"en", "lat"=>41.592718, "lon"=>-95.060398),
"/US/US-IA/elk-run-heights-6695" => array("type"=>"city", "pop"=>1079, "name_ru"=>"Элк Ран Хайтс", "name_en"=>"Elk Run Heights", "lang"=>"en", "lat"=>42.467056, "lon"=>-92.249300),
"/US/US-IA/elkader-6691" => array("type"=>"city", "pop"=>1346, "name_ru"=>"Элкадер", "name_en"=>"Elkader", "lang"=>"en", "lat"=>42.855986, "lon"=>-91.402832),
"/US/US-IA/elkhart-6692" => array("type"=>"city", "pop"=>473, "name_ru"=>"Элкхарт", "name_en"=>"Elkhart", "lang"=>"en", "lat"=>41.794093, "lon"=>-93.514548),
"/US/US-IA/elkport-6694" => array("type"=>"city", "pop"=>101, "name_ru"=>"Элкпорт", "name_en"=>"Elkport", "lang"=>"en", "lat"=>42.741105, "lon"=>-91.275374),
"/US/US-IA/elliott-6696" => array("type"=>"city", "pop"=>362, "name_ru"=>"Эллиотт", "name_en"=>"Elliott", "lang"=>"en", "lat"=>41.149845, "lon"=>-95.163035),
"/US/US-IA/ellston-6697" => array("type"=>"city", "pop"=>55, "name_ru"=>"Эллстон", "name_en"=>"Ellston", "lang"=>"en", "lat"=>40.840363, "lon"=>-94.108274),
"/US/US-IA/ellsworth-6698" => array("type"=>"city", "pop"=>514, "name_ru"=>"Эллсворт", "name_en"=>"Ellsworth", "lang"=>"en", "lat"=>42.310612, "lon"=>-93.581847),
"/US/US-IA/elma-6699" => array("type"=>"city", "pop"=>569, "name_ru"=>"Элма", "name_en"=>"Elma", "lang"=>"en", "lat"=>43.245657, "lon"=>-92.439954),
"/US/US-IA/ely-6700" => array("type"=>"city", "pop"=>1514, "name_ru"=>"Или", "name_en"=>"Ely", "lang"=>"en", "lat"=>41.875141, "lon"=>-91.587272),
"/US/US-IA/emerson-6701" => array("type"=>"city", "pop"=>490, "name_ru"=>"Эмерсон", "name_en"=>"Emerson", "lang"=>"en", "lat"=>41.018500, "lon"=>-95.402789),
"/US/US-IA/emmetsburg-6702" => array("type"=>"city", "pop"=>3616, "name_ru"=>"Эмметсбург", "name_en"=>"Emmetsburg", "lang"=>"en", "lat"=>43.113466, "lon"=>-94.680533),
"/US/US-IA/epworth-6703" => array("type"=>"city", "pop"=>1620, "name_ru"=>"Эпуорт", "name_en"=>"Epworth", "lang"=>"en", "lat"=>42.446801, "lon"=>-90.934160),
"/US/US-IA/essex-6704" => array("type"=>"city", "pop"=>795, "name_ru"=>"Эссекс", "name_en"=>"Essex", "lang"=>"en", "lat"=>40.833032, "lon"=>-95.303406),
"/US/US-IA/estherville-6705" => array("type"=>"city", "pop"=>6291, "name_ru"=>"Эстервилл", "name_en"=>"Estherville", "lang"=>"en", "lat"=>43.399352, "lon"=>-94.833186),
"/US/US-IA/evansdale-6706" => array("type"=>"city", "pop"=>5056, "name_ru"=>"Эвансдейл", "name_en"=>"Evansdale", "lang"=>"en", "lat"=>42.461784, "lon"=>-92.275646),
"/US/US-IA/everly-6707" => array("type"=>"city", "pop"=>636, "name_ru"=>"Эверли", "name_en"=>"Everly", "lang"=>"en", "lat"=>43.160516, "lon"=>-95.321613),
"/US/US-IA/exira-6708" => array("type"=>"city", "pop"=>722, "name_ru"=>"Эксира", "name_en"=>"Exira", "lang"=>"en", "lat"=>41.591621, "lon"=>-94.879292),
"/US/US-IA/exline-6709" => array("type"=>"city", "pop"=>164, "name_ru"=>"Экслайн", "name_en"=>"Exline", "lang"=>"en", "lat"=>40.648831, "lon"=>-92.841683),
"/US/US-IA/fairbank-6710" => array("type"=>"city", "pop"=>991, "name_ru"=>"Фэрбэнк", "name_en"=>"Fairbank", "lang"=>"en", "lat"=>42.638854, "lon"=>-92.050122),
"/US/US-IA/fairfax-6711" => array("type"=>"city", "pop"=>1875, "name_ru"=>"Фэрфакс", "name_en"=>"Fairfax", "lang"=>"en", "lat"=>41.921600, "lon"=>-91.780127),
"/US/US-IA/fairfield-6712" => array("type"=>"city", "pop"=>9218, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>41.007146, "lon"=>-91.967142),
"/US/US-IA/farley-6713" => array("type"=>"city", "pop"=>1391, "name_ru"=>"Фэрли", "name_en"=>"Farley", "lang"=>"en", "lat"=>42.443929, "lon"=>-91.006684),
"/US/US-IA/farmersburg-6714" => array("type"=>"city", "pop"=>273, "name_ru"=>"Фармерсбург", "name_en"=>"Farmersburg", "lang"=>"en", "lat"=>42.958532, "lon"=>-91.367419),
"/US/US-IA/farmington-6715" => array("type"=>"city", "pop"=>718, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>40.638807, "lon"=>-91.738843),
"/US/US-IA/farnhamville-6716" => array("type"=>"city", "pop"=>358, "name_ru"=>"Фарнхэмвилл", "name_en"=>"Farnhamville", "lang"=>"en", "lat"=>42.277462, "lon"=>-94.408009),
"/US/US-IA/farragut-6717" => array("type"=>"city", "pop"=>457, "name_ru"=>"Фаррагут", "name_en"=>"Farragut", "lang"=>"en", "lat"=>40.719801, "lon"=>-95.480050),
"/US/US-IA/fayette-6718" => array("type"=>"city", "pop"=>1258, "name_ru"=>"Файет", "name_en"=>"Fayette", "lang"=>"en", "lat"=>42.840385, "lon"=>-91.802615),
"/US/US-IA/fenton-6719" => array("type"=>"city", "pop"=>263, "name_ru"=>"Фэнтон", "name_en"=>"Fenton", "lang"=>"en", "lat"=>43.218264, "lon"=>-94.427682),
"/US/US-IA/ferguson-6720" => array("type"=>"city", "pop"=>123, "name_ru"=>"Фергюсон", "name_en"=>"Ferguson", "lang"=>"en", "lat"=>41.938398, "lon"=>-92.863422),
"/US/US-IA/fertile-6721" => array("type"=>"city", "pop"=>352, "name_ru"=>"Фертил", "name_en"=>"Fertile", "lang"=>"en", "lat"=>43.264874, "lon"=>-93.422579),
"/US/US-IA/floris-6722" => array("type"=>"city", "pop"=>152, "name_ru"=>"Флорис", "name_en"=>"Floris", "lang"=>"en", "lat"=>40.864078, "lon"=>-92.332096),
"/US/US-IA/floyd-6723" => array("type"=>"city", "pop"=>337, "name_ru"=>"Флойд", "name_en"=>"Floyd", "lang"=>"en", "lat"=>43.128368, "lon"=>-92.739429),
"/US/US-IA/fonda-6724" => array("type"=>"city", "pop"=>562, "name_ru"=>"Фонда", "name_en"=>"Fonda", "lang"=>"en", "lat"=>42.581767, "lon"=>-94.845516),
"/US/US-IA/fontanelle-6725" => array("type"=>"city", "pop"=>635, "name_ru"=>"Фонтанелл", "name_en"=>"Fontanelle", "lang"=>"en", "lat"=>41.289962, "lon"=>-94.560655),
"/US/US-IA/forest-city-6726" => array("type"=>"city", "pop"=>4098, "name_ru"=>"Форест Сити", "name_en"=>"Forest City", "lang"=>"en", "lat"=>43.257192, "lon"=>-93.637972),
"/US/US-IA/fort-atkinson-6727" => array("type"=>"city", "pop"=>361, "name_ru"=>"Форт Аткинсон", "name_en"=>"Fort Atkinson", "lang"=>"en", "lat"=>43.144061, "lon"=>-91.934676),
"/US/US-IA/fort-dodge-6728" => array("type"=>"city", "pop"=>25180, "name_ru"=>"Форт Додж", "name_en"=>"Fort Dodge", "lang"=>"en", "lat"=>42.507943, "lon"=>-94.176764),
"/US/US-IA/fort-madison-6729" => array("type"=>"city", "pop"=>10897, "name_ru"=>"Форт Мэдисон", "name_en"=>"Fort Madison", "lang"=>"en", "lat"=>40.619742, "lon"=>-91.352178),
"/US/US-IA/fostoria-6730" => array("type"=>"city", "pop"=>242, "name_ru"=>"Фостория", "name_en"=>"Fostoria", "lang"=>"en", "lat"=>43.241574, "lon"=>-95.154765),
"/US/US-IA/franklin-6731" => array("type"=>"city", "pop"=>124, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>40.666423, "lon"=>-91.511672),
"/US/US-IA/fraser-6732" => array("type"=>"city", "pop"=>117, "name_ru"=>"Фрейзер", "name_en"=>"Fraser", "lang"=>"en", "lat"=>42.126850, "lon"=>-93.973666),
"/US/US-IA/fredericksburg-6733" => array("type"=>"city", "pop"=>878, "name_ru"=>"Фредериксбург", "name_en"=>"Fredericksburg", "lang"=>"en", "lat"=>42.964609, "lon"=>-92.198290),
"/US/US-IA/frederika-6734" => array("type"=>"city", "pop"=>191, "name_ru"=>"Фредерика", "name_en"=>"Frederika", "lang"=>"en", "lat"=>42.883594, "lon"=>-92.305875),
"/US/US-IA/fredonia-6735" => array("type"=>"city", "pop"=>218, "name_ru"=>"Фредониа", "name_en"=>"Fredonia", "lang"=>"en", "lat"=>41.284703, "lon"=>-91.338951),
"/US/US-IA/fremont-6736" => array("type"=>"city", "pop"=>687, "name_ru"=>"Фремонт", "name_en"=>"Fremont", "lang"=>"en", "lat"=>41.212090, "lon"=>-92.434916),
"/US/US-IA/fruitland-6737" => array("type"=>"city", "pop"=>945, "name_ru"=>"Фрутленд", "name_en"=>"Fruitland", "lang"=>"en", "lat"=>41.346754, "lon"=>-91.125711),
"/US/US-IA/galt-6738" => array("type"=>"city", "pop"=>28, "name_ru"=>"Галт", "name_en"=>"Galt", "lang"=>"en", "lat"=>42.693902, "lon"=>-93.604840),
"/US/US-IA/galva-6739" => array("type"=>"city", "pop"=>296, "name_ru"=>"Галва", "name_en"=>"Galva", "lang"=>"en", "lat"=>42.506247, "lon"=>-95.416253),
"/US/US-IA/garber-6740" => array("type"=>"city", "pop"=>74, "name_ru"=>"Гарбер", "name_en"=>"Garber", "lang"=>"en", "lat"=>42.743904, "lon"=>-91.262872),
"/US/US-IA/garden-grove-6741" => array("type"=>"city", "pop"=>235, "name_ru"=>"Гарден Гров", "name_en"=>"Garden Grove", "lang"=>"en", "lat"=>40.826477, "lon"=>-93.607515),
"/US/US-IA/garnavillo-6742" => array("type"=>"city", "pop"=>686, "name_ru"=>"Гарнавилло", "name_en"=>"Garnavillo", "lang"=>"en", "lat"=>42.867814, "lon"=>-91.236414),
"/US/US-IA/garner-6743" => array("type"=>"city", "pop"=>2932, "name_ru"=>"Гарнер", "name_en"=>"Garner", "lang"=>"en", "lat"=>43.098218, "lon"=>-93.602483),
"/US/US-IA/garrison-6744" => array("type"=>"city", "pop"=>401, "name_ru"=>"Гаррисон", "name_en"=>"Garrison", "lang"=>"en", "lat"=>42.143610, "lon"=>-92.142832),
"/US/US-IA/garwin-6745" => array("type"=>"city", "pop"=>535, "name_ru"=>"Гарвин", "name_en"=>"Garwin", "lang"=>"en", "lat"=>42.093983, "lon"=>-92.678217),
"/US/US-IA/geneva-6746" => array("type"=>"city", "pop"=>171, "name_ru"=>"Дженива", "name_en"=>"Geneva", "lang"=>"en", "lat"=>42.675939, "lon"=>-93.129995),
"/US/US-IA/george-6747" => array("type"=>"city", "pop"=>970, "name_ru"=>"Джордж", "name_en"=>"George", "lang"=>"en", "lat"=>43.341687, "lon"=>-96.002703),
"/US/US-IA/gibson-6748" => array("type"=>"city", "pop"=>72, "name_ru"=>"Гибсон", "name_en"=>"Gibson", "lang"=>"en", "lat"=>41.480577, "lon"=>-92.391531),
"/US/US-IA/gilbert-6749" => array("type"=>"city", "pop"=>1050, "name_ru"=>"Гилберт", "name_en"=>"Gilbert", "lang"=>"en", "lat"=>42.107425, "lon"=>-93.647514),
"/US/US-IA/gilbertville-6750" => array("type"=>"city", "pop"=>798, "name_ru"=>"Гилбертвилл", "name_en"=>"Gilbertville", "lang"=>"en", "lat"=>42.418696, "lon"=>-92.213735),
"/US/US-IA/gillett-grove-6751" => array("type"=>"city", "pop"=>46, "name_ru"=>"Гиллет Гров", "name_en"=>"Gillett Grove", "lang"=>"en", "lat"=>43.015366, "lon"=>-95.035200),
"/US/US-IA/gilman-6752" => array("type"=>"city", "pop"=>570, "name_ru"=>"Гилман", "name_en"=>"Gilman", "lang"=>"en", "lat"=>41.878606, "lon"=>-92.787828),
"/US/US-IA/gilmore-city-6753" => array("type"=>"city", "pop"=>489, "name_ru"=>"Гилмор Сити", "name_en"=>"Gilmore City", "lang"=>"en", "lat"=>42.726580, "lon"=>-94.438756),
"/US/US-IA/gladbrook-6754" => array("type"=>"city", "pop"=>990, "name_ru"=>"Глэдбрук", "name_en"=>"Gladbrook", "lang"=>"en", "lat"=>42.186073, "lon"=>-92.714398),
"/US/US-IA/glenwood-6755" => array("type"=>"city", "pop"=>5646, "name_ru"=>"Гленвуд", "name_en"=>"Glenwood", "lang"=>"en", "lat"=>41.047789, "lon"=>-95.742174),
"/US/US-IA/glidden-6756" => array("type"=>"city", "pop"=>1227, "name_ru"=>"Глидден", "name_en"=>"Glidden", "lang"=>"en", "lat"=>42.057250, "lon"=>-94.725215),
"/US/US-IA/goldfield-6757" => array("type"=>"city", "pop"=>607, "name_ru"=>"Гоулдфилд", "name_en"=>"Goldfield", "lang"=>"en", "lat"=>42.735698, "lon"=>-93.921234),
"/US/US-IA/goodell-6758" => array("type"=>"city", "pop"=>143, "name_ru"=>"Гуделл", "name_en"=>"Goodell", "lang"=>"en", "lat"=>42.923644, "lon"=>-93.613851),
"/US/US-IA/goose-lake-6759" => array("type"=>"city", "pop"=>234, "name_ru"=>"Гус Лейк", "name_en"=>"Goose Lake", "lang"=>"en", "lat"=>41.968164, "lon"=>-90.382176),
"/US/US-IA/gowrie-6760" => array("type"=>"city", "pop"=>1012, "name_ru"=>"Гори", "name_en"=>"Gowrie", "lang"=>"en", "lat"=>42.276843, "lon"=>-94.288504),
"/US/US-IA/graettinger-6761" => array("type"=>"city", "pop"=>831, "name_ru"=>"Граеттингер", "name_en"=>"Graettinger", "lang"=>"en", "lat"=>43.236391, "lon"=>-94.750966),
"/US/US-IA/graf-6762" => array("type"=>"city", "pop"=>75, "name_ru"=>"Граф", "name_en"=>"Graf", "lang"=>"en", "lat"=>42.494709, "lon"=>-90.868891),
"/US/US-IA/grafton-6763" => array("type"=>"city", "pop"=>283, "name_ru"=>"Графтон", "name_en"=>"Grafton", "lang"=>"en", "lat"=>43.330148, "lon"=>-93.069914),
"/US/US-IA/grand-junction-6764" => array("type"=>"city", "pop"=>848, "name_ru"=>"Гранд Джанкшен", "name_en"=>"Grand Junction", "lang"=>"en", "lat"=>42.032188, "lon"=>-94.236501),
"/US/US-IA/grand-mound-6765" => array("type"=>"city", "pop"=>645, "name_ru"=>"Гранд Маунд", "name_en"=>"Grand Mound", "lang"=>"en", "lat"=>41.823463, "lon"=>-90.648942),
"/US/US-IA/grand-river-6766" => array("type"=>"city", "pop"=>191, "name_ru"=>"Гранд Ривер", "name_en"=>"Grand River", "lang"=>"en", "lat"=>40.818944, "lon"=>-93.962510),
"/US/US-IA/grandview-6767" => array("type"=>"city", "pop"=>597, "name_ru"=>"Грандвью", "name_en"=>"Grandview", "lang"=>"en", "lat"=>41.277211, "lon"=>-91.187432),
"/US/US-IA/granger-6768" => array("type"=>"city", "pop"=>1054, "name_ru"=>"Грейнджер", "name_en"=>"Granger", "lang"=>"en", "lat"=>41.760300, "lon"=>-93.823673),
"/US/US-IA/grant-6769" => array("type"=>"city", "pop"=>92, "name_ru"=>"Грант", "name_en"=>"Grant", "lang"=>"en", "lat"=>41.142218, "lon"=>-94.985105),
"/US/US-IA/granville-6770" => array("type"=>"city", "pop"=>321, "name_ru"=>"Гранвилл", "name_en"=>"Granville", "lang"=>"en", "lat"=>42.984742, "lon"=>-95.874833),
"/US/US-IA/gravity-6771" => array("type"=>"city", "pop"=>200, "name_ru"=>"Грэйвити", "name_en"=>"Gravity", "lang"=>"en", "lat"=>40.760128, "lon"=>-94.742785),
"/US/US-IA/gray-6772" => array("type"=>"city", "pop"=>74, "name_ru"=>"Грей", "name_en"=>"Gray", "lang"=>"en", "lat"=>41.841324, "lon"=>-94.984160),
"/US/US-IA/greeley-6773" => array("type"=>"city", "pop"=>257, "name_ru"=>"Грили", "name_en"=>"Greeley", "lang"=>"en", "lat"=>42.585126, "lon"=>-91.341940),
"/US/US-IA/green-island-6776" => array("type"=>"city", "pop"=>54, "name_ru"=>"Грин Айленд", "name_en"=>"Green Island", "lang"=>"en", "lat"=>42.154608, "lon"=>-90.322415),
"/US/US-IA/greene-6774" => array("type"=>"city", "pop"=>976, "name_ru"=>"Грин", "name_en"=>"Greene", "lang"=>"en", "lat"=>42.896846, "lon"=>-92.803925),
"/US/US-IA/greenfield-6775" => array("type"=>"city", "pop"=>1888, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>41.305504, "lon"=>-94.458483),
"/US/US-IA/greenville-6777" => array("type"=>"city", "pop"=>85, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>43.017204, "lon"=>-95.145209),
"/US/US-IA/grimes-6778" => array("type"=>"city", "pop"=>8419, "name_ru"=>"Граймс", "name_en"=>"Grimes", "lang"=>"en", "lat"=>41.660817, "lon"=>-93.789294),
"/US/US-IA/grinnell-6779" => array("type"=>"city", "pop"=>9167, "name_ru"=>"Гриннел", "name_en"=>"Grinnell", "lang"=>"en", "lat"=>41.740774, "lon"=>-92.722670),
"/US/US-IA/griswold-6780" => array("type"=>"city", "pop"=>951, "name_ru"=>"Грисволд", "name_en"=>"Griswold", "lang"=>"en", "lat"=>41.233943, "lon"=>-95.138872),
"/US/US-IA/grundy-center-6781" => array("type"=>"city", "pop"=>2519, "name_ru"=>"Гранди Сентер", "name_en"=>"Grundy Center", "lang"=>"en", "lat"=>42.363021, "lon"=>-92.774714),
"/US/US-IA/gruver-6782" => array("type"=>"city", "pop"=>97, "name_ru"=>"Грувер", "name_en"=>"Gruver", "lang"=>"en", "lat"=>43.393178, "lon"=>-94.703072),
"/US/US-IA/guernsey-6783" => array("type"=>"city", "pop"=>66, "name_ru"=>"Гернси", "name_en"=>"Guernsey", "lang"=>"en", "lat"=>41.649250, "lon"=>-92.343087),
"/US/US-IA/guthrie-center-6784" => array("type"=>"city", "pop"=>1517, "name_ru"=>"Гатри Сентер", "name_en"=>"Guthrie Center", "lang"=>"en", "lat"=>41.678648, "lon"=>-94.498282),
"/US/US-IA/guttenberg-6785" => array("type"=>"city", "pop"=>1913, "name_ru"=>"Гаттенберг", "name_en"=>"Guttenberg", "lang"=>"en", "lat"=>42.787898, "lon"=>-91.104097),
"/US/US-IA/halbur-6786" => array("type"=>"city", "pop"=>211, "name_ru"=>"Халбур", "name_en"=>"Halbur", "lang"=>"en", "lat"=>42.005497, "lon"=>-94.971432),
"/US/US-IA/hamburg-6787" => array("type"=>"city", "pop"=>1155, "name_ru"=>"Гамбург", "name_en"=>"Hamburg", "lang"=>"en", "lat"=>40.605815, "lon"=>-95.654752),
"/US/US-IA/hamilton-6788" => array("type"=>"city", "pop"=>137, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>41.169967, "lon"=>-92.904076),
"/US/US-IA/hampton-6789" => array("type"=>"city", "pop"=>4134, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>42.745045, "lon"=>-93.200911),
"/US/US-IA/hancock-6790" => array("type"=>"city", "pop"=>198, "name_ru"=>"Хэнкок", "name_en"=>"Hancock", "lang"=>"en", "lat"=>41.393306, "lon"=>-95.364990),
"/US/US-IA/hanlontown-6791" => array("type"=>"city", "pop"=>228, "name_ru"=>"Хэнлонтаун", "name_en"=>"Hanlontown", "lang"=>"en", "lat"=>43.280928, "lon"=>-93.379164),
"/US/US-IA/hansell-6792" => array("type"=>"city", "pop"=>96, "name_ru"=>"Ханселл", "name_en"=>"Hansell", "lang"=>"en", "lat"=>42.757246, "lon"=>-93.103440),
"/US/US-IA/harcourt-6793" => array("type"=>"city", "pop"=>317, "name_ru"=>"Харкорт", "name_en"=>"Harcourt", "lang"=>"en", "lat"=>42.261039, "lon"=>-94.174989),
"/US/US-IA/hardy-6794" => array("type"=>"city", "pop"=>88, "name_ru"=>"Харди", "name_en"=>"Hardy", "lang"=>"en", "lat"=>42.810237, "lon"=>-94.051481),
"/US/US-IA/harlan-6795" => array("type"=>"city", "pop"=>4986, "name_ru"=>"Харлан", "name_en"=>"Harlan", "lang"=>"en", "lat"=>41.649029, "lon"=>-95.326193),
"/US/US-IA/harper-6796" => array("type"=>"city", "pop"=>114, "name_ru"=>"Харпер", "name_en"=>"Harper", "lang"=>"en", "lat"=>41.363340, "lon"=>-92.051202),
"/US/US-IA/harpers-ferry-6797" => array("type"=>"city", "pop"=>312, "name_ru"=>"Харперс Ферри", "name_en"=>"Harpers Ferry", "lang"=>"en", "lat"=>43.200969, "lon"=>-91.151246),
"/US/US-IA/harris-6798" => array("type"=>"city", "pop"=>179, "name_ru"=>"Харрис", "name_en"=>"Harris", "lang"=>"en", "lat"=>43.447644, "lon"=>-95.432918),
"/US/US-IA/hartford-6799" => array("type"=>"city", "pop"=>779, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>41.457761, "lon"=>-93.403854),
"/US/US-IA/hartley-6800" => array("type"=>"city", "pop"=>1502, "name_ru"=>"Хартли", "name_en"=>"Hartley", "lang"=>"en", "lat"=>43.179348, "lon"=>-95.477538),
"/US/US-IA/hartwick-6801" => array("type"=>"city", "pop"=>78, "name_ru"=>"Хартвик", "name_en"=>"Hartwick", "lang"=>"en", "lat"=>41.783994, "lon"=>-92.346048),
"/US/US-IA/harvey-6802" => array("type"=>"city", "pop"=>264, "name_ru"=>"Харви", "name_en"=>"Harvey", "lang"=>"en", "lat"=>41.316623, "lon"=>-92.923405),
"/US/US-IA/hastings-6803" => array("type"=>"city", "pop"=>169, "name_ru"=>"Гастингс", "name_en"=>"Hastings", "lang"=>"en", "lat"=>41.023922, "lon"=>-95.495624),
"/US/US-IA/havelock-6804" => array("type"=>"city", "pop"=>169, "name_ru"=>"Хэвлок", "name_en"=>"Havelock", "lang"=>"en", "lat"=>42.833852, "lon"=>-94.700129),
"/US/US-IA/haverhill-6805" => array("type"=>"city", "pop"=>163, "name_ru"=>"Хэйверхилл", "name_en"=>"Haverhill", "lang"=>"en", "lat"=>41.944138, "lon"=>-92.961208),
"/US/US-IA/hawarden-6806" => array("type"=>"city", "pop"=>2403, "name_ru"=>"Гаварден", "name_en"=>"Hawarden", "lang"=>"en", "lat"=>43.002213, "lon"=>-96.482438),
"/US/US-IA/hawkeye-6807" => array("type"=>"city", "pop"=>440, "name_ru"=>"Хокай", "name_en"=>"Hawkeye", "lang"=>"en", "lat"=>42.938042, "lon"=>-91.950861),
"/US/US-IA/hayesville-6808" => array("type"=>"city", "pop"=>61, "name_ru"=>"Хайесвилл", "name_en"=>"Hayesville", "lang"=>"en", "lat"=>41.264468, "lon"=>-92.246970),
"/US/US-IA/hazleton-6809" => array("type"=>"city", "pop"=>901, "name_ru"=>"Хейзлтон", "name_en"=>"Hazleton", "lang"=>"en", "lat"=>42.618316, "lon"=>-91.905435),
"/US/US-IA/hedrick-6810" => array("type"=>"city", "pop"=>804, "name_ru"=>"Хэдрик", "name_en"=>"Hedrick", "lang"=>"en", "lat"=>41.170643, "lon"=>-92.307655),
"/US/US-IA/henderson-6811" => array("type"=>"city", "pop"=>170, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>41.139100, "lon"=>-95.431024),
"/US/US-IA/hepburn-6812" => array("type"=>"city", "pop"=>27, "name_ru"=>"Хэпберн", "name_en"=>"Hepburn", "lang"=>"en", "lat"=>40.848204, "lon"=>-95.016469),
"/US/US-IA/hiawatha-6813" => array("type"=>"city", "pop"=>6688, "name_ru"=>"Хайавата", "name_en"=>"Hiawatha", "lang"=>"en", "lat"=>42.046421, "lon"=>-91.683722),
"/US/US-IA/hills-6814" => array("type"=>"city", "pop"=>693, "name_ru"=>"Хиллс", "name_en"=>"Hills", "lang"=>"en", "lat"=>41.574653, "lon"=>-91.537073),
"/US/US-IA/hillsboro-6815" => array("type"=>"city", "pop"=>200, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>40.836758, "lon"=>-91.713493),
"/US/US-IA/hinton-6816" => array("type"=>"city", "pop"=>847, "name_ru"=>"Хинтон", "name_en"=>"Hinton", "lang"=>"en", "lat"=>42.624724, "lon"=>-96.293947),
"/US/US-IA/holland-6817" => array("type"=>"city", "pop"=>256, "name_ru"=>"Холланд", "name_en"=>"Holland", "lang"=>"en", "lat"=>42.399726, "lon"=>-92.799288),
"/US/US-IA/holstein-6818" => array("type"=>"city", "pop"=>1371, "name_ru"=>"Холстайн", "name_en"=>"Holstein", "lang"=>"en", "lat"=>42.486808, "lon"=>-95.542285),
"/US/US-IA/holy-cross-6819" => array("type"=>"city", "pop"=>354, "name_ru"=>"Холи Кросс", "name_en"=>"Holy Cross", "lang"=>"en", "lat"=>42.601492, "lon"=>-90.996271),
"/US/US-IA/hopkinton-6820" => array("type"=>"city", "pop"=>641, "name_ru"=>"Хопкинтон", "name_en"=>"Hopkinton", "lang"=>"en", "lat"=>42.343571, "lon"=>-91.248739),
"/US/US-IA/hornick-6821" => array("type"=>"city", "pop"=>240, "name_ru"=>"Хорник", "name_en"=>"Hornick", "lang"=>"en", "lat"=>42.230982, "lon"=>-96.096316),
"/US/US-IA/hospers-6822" => array("type"=>"city", "pop"=>666, "name_ru"=>"Хосперс", "name_en"=>"Hospers", "lang"=>"en", "lat"=>43.072010, "lon"=>-95.903609),
"/US/US-IA/houghton-6823" => array("type"=>"city", "pop"=>126, "name_ru"=>"Хоутон", "name_en"=>"Houghton", "lang"=>"en", "lat"=>40.782980, "lon"=>-91.612445),
"/US/US-IA/hubbard-6824" => array("type"=>"city", "pop"=>817, "name_ru"=>"Хаббард", "name_en"=>"Hubbard", "lang"=>"en", "lat"=>42.306123, "lon"=>-93.300481),
"/US/US-IA/hudson-6825" => array("type"=>"city", "pop"=>2186, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>42.423288, "lon"=>-92.454691),
"/US/US-IA/hull-6826" => array("type"=>"city", "pop"=>2028, "name_ru"=>"Халл", "name_en"=>"Hull", "lang"=>"en", "lat"=>43.190211, "lon"=>-96.132693),
"/US/US-IA/humboldt-6827" => array("type"=>"city", "pop"=>4198, "name_ru"=>"Гумбольдт", "name_en"=>"Humboldt", "lang"=>"en", "lat"=>42.721541, "lon"=>-94.223688),
"/US/US-IA/humeston-6828" => array("type"=>"city", "pop"=>522, "name_ru"=>"Хумстон", "name_en"=>"Humeston", "lang"=>"en", "lat"=>40.860324, "lon"=>-93.496816),
"/US/US-IA/huxley-6829" => array("type"=>"city", "pop"=>2830, "name_ru"=>"Хаксли", "name_en"=>"Huxley", "lang"=>"en", "lat"=>41.896000, "lon"=>-93.602470),
"/US/US-IA/ida-grove-6830" => array("type"=>"city", "pop"=>2056, "name_ru"=>"Ида Гров", "name_en"=>"Ida Grove", "lang"=>"en", "lat"=>42.344395, "lon"=>-95.472944),
"/US/US-IA/imogene-6831" => array("type"=>"city", "pop"=>57, "name_ru"=>"Имоджин", "name_en"=>"Imogene", "lang"=>"en", "lat"=>40.878718, "lon"=>-95.427418),
"/US/US-IA/independence-6832" => array("type"=>"city", "pop"=>6119, "name_ru"=>"Индепенденс", "name_en"=>"Independence", "lang"=>"en", "lat"=>42.468054, "lon"=>-91.890988),
"/US/US-IA/indianola-6833" => array("type"=>"city", "pop"=>14361, "name_ru"=>"Индианола", "name_en"=>"Indianola", "lang"=>"en", "lat"=>41.360712, "lon"=>-93.565484),
"/US/US-IA/inwood-6834" => array("type"=>"city", "pop"=>843, "name_ru"=>"Инвуд", "name_en"=>"Inwood", "lang"=>"en", "lat"=>43.286960, "lon"=>-96.440149),
"/US/US-IA/ionia-6835" => array("type"=>"city", "pop"=>260, "name_ru"=>"Иония", "name_en"=>"Ionia", "lang"=>"en", "lat"=>43.036246, "lon"=>-92.457636),
"/US/US-IA/iowa-city-6836" => array("type"=>"city", "pop"=>67831, "name_ru"=>"Айова Сити", "name_en"=>"Iowa City", "lang"=>"en", "lat"=>41.658250, "lon"=>-91.535124),
"/US/US-IA/iowa-falls-6837" => array("type"=>"city", "pop"=>4965, "name_ru"=>"Айова Фоллс", "name_en"=>"Iowa Falls", "lang"=>"en", "lat"=>42.520841, "lon"=>-93.265806),
"/US/US-IA/ireton-6838" => array("type"=>"city", "pop"=>584, "name_ru"=>"Айртон", "name_en"=>"Ireton", "lang"=>"en", "lat"=>42.975293, "lon"=>-96.321541),
"/US/US-IA/irwin-6839" => array("type"=>"city", "pop"=>353, "name_ru"=>"Ирвин", "name_en"=>"Irwin", "lang"=>"en", "lat"=>41.790150, "lon"=>-95.206704),
"/US/US-IA/jackson-junction-6840" => array("type"=>"city", "pop"=>63, "name_ru"=>"Джэксон Джанкшен", "name_en"=>"Jackson Junction", "lang"=>"en", "lat"=>43.103310, "lon"=>-92.047978),
"/US/US-IA/jamaica-6841" => array("type"=>"city", "pop"=>223, "name_ru"=>"Ямайка", "name_en"=>"Jamaica", "lang"=>"en", "lat"=>41.845502, "lon"=>-94.307192),
"/US/US-IA/janesville-6842" => array("type"=>"city", "pop"=>837, "name_ru"=>"Джейнсвилл", "name_en"=>"Janesville", "lang"=>"en", "lat"=>42.644374, "lon"=>-92.463878),
"/US/US-IA/jefferson-6843" => array("type"=>"city", "pop"=>4144, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>42.018767, "lon"=>-94.376498),
"/US/US-IA/jesup-6844" => array("type"=>"city", "pop"=>2315, "name_ru"=>"Джесап", "name_en"=>"Jesup", "lang"=>"en", "lat"=>42.475854, "lon"=>-92.066886),
"/US/US-IA/jewell-junction-6845" => array("type"=>"city", "pop"=>1144, "name_ru"=>"Джевелл Джанкшен", "name_en"=>"Jewell Junction", "lang"=>"en", "lat"=>42.311337, "lon"=>-93.638570),
"/US/US-IA/johnston-6846" => array("type"=>"city", "pop"=>15691, "name_ru"=>"Джонстон", "name_en"=>"Johnston", "lang"=>"en", "lat"=>41.685900, "lon"=>-93.702279),
"/US/US-IA/joice-6847" => array("type"=>"city", "pop"=>230, "name_ru"=>"Джойс", "name_en"=>"Joice", "lang"=>"en", "lat"=>43.364286, "lon"=>-93.457076),
"/US/US-IA/jolley-6848" => array("type"=>"city", "pop"=>31, "name_ru"=>"Джолли", "name_en"=>"Jolley", "lang"=>"en", "lat"=>42.479261, "lon"=>-94.717869),
"/US/US-IA/kalona-6849" => array("type"=>"city", "pop"=>2516, "name_ru"=>"Калона", "name_en"=>"Kalona", "lang"=>"en", "lat"=>41.487558, "lon"=>-91.700519),
"/US/US-IA/kamrar-6850" => array("type"=>"city", "pop"=>211, "name_ru"=>"Камар", "name_en"=>"Kamrar", "lang"=>"en", "lat"=>42.391749, "lon"=>-93.727436),
"/US/US-IA/kanawha-6851" => array("type"=>"city", "pop"=>657, "name_ru"=>"Канаха", "name_en"=>"Kanawha", "lang"=>"en", "lat"=>42.933965, "lon"=>-93.794052),
"/US/US-IA/kellerton-6852" => array("type"=>"city", "pop"=>330, "name_ru"=>"Келлертон", "name_en"=>"Kellerton", "lang"=>"en", "lat"=>40.710202, "lon"=>-94.048781),
"/US/US-IA/kelley-6853" => array("type"=>"city", "pop"=>340, "name_ru"=>"Келли", "name_en"=>"Kelley", "lang"=>"en", "lat"=>41.950950, "lon"=>-93.664685),
"/US/US-IA/kellogg-6854" => array("type"=>"city", "pop"=>601, "name_ru"=>"Келлогг", "name_en"=>"Kellogg", "lang"=>"en", "lat"=>41.716649, "lon"=>-92.907369),
"/US/US-IA/kensett-6855" => array("type"=>"city", "pop"=>267, "name_ru"=>"Кенсетт", "name_en"=>"Kensett", "lang"=>"en", "lat"=>43.353890, "lon"=>-93.210927),
"/US/US-IA/kent-6856" => array("type"=>"city", "pop"=>65, "name_ru"=>"Кент", "name_en"=>"Kent", "lang"=>"en", "lat"=>40.953118, "lon"=>-94.456327),
"/US/US-IA/keokuk-6857" => array("type"=>"city", "pop"=>10387, "name_ru"=>"Кеокек", "name_en"=>"Keokuk", "lang"=>"en", "lat"=>40.409588, "lon"=>-91.403515),
"/US/US-IA/keomah-village-6858" => array("type"=>"city", "pop"=>94, "name_ru"=>"Кеома Виллидж", "name_en"=>"Keomah Village", "lang"=>"en", "lat"=>41.288204, "lon"=>-92.536293),
"/US/US-IA/keosauqua-6859" => array("type"=>"city", "pop"=>1070, "name_ru"=>"Кеосоква", "name_en"=>"Keosauqua", "lang"=>"en", "lat"=>40.734637, "lon"=>-91.960279),
"/US/US-IA/keota-6860" => array("type"=>"city", "pop"=>890, "name_ru"=>"Кеота", "name_en"=>"Keota", "lang"=>"en", "lat"=>41.365993, "lon"=>-91.955498),
"/US/US-IA/keswick-6861" => array("type"=>"city", "pop"=>283, "name_ru"=>"Кесвик", "name_en"=>"Keswick", "lang"=>"en", "lat"=>41.455043, "lon"=>-92.237332),
"/US/US-IA/keystone-6862" => array("type"=>"city", "pop"=>668, "name_ru"=>"Кистоун", "name_en"=>"Keystone", "lang"=>"en", "lat"=>41.999759, "lon"=>-92.198579),
"/US/US-IA/kimballton-6863" => array("type"=>"city", "pop"=>296, "name_ru"=>"Кимболлтон", "name_en"=>"Kimballton", "lang"=>"en", "lat"=>41.628293, "lon"=>-95.074224),
"/US/US-IA/kingsley-6864" => array("type"=>"city", "pop"=>1241, "name_ru"=>"Кингсли", "name_en"=>"Kingsley", "lang"=>"en", "lat"=>42.587531, "lon"=>-95.968001),
"/US/US-IA/kinross-6865" => array("type"=>"city", "pop"=>76, "name_ru"=>"Кинросс", "name_en"=>"Kinross", "lang"=>"en", "lat"=>41.459412, "lon"=>-91.987212),
"/US/US-IA/kirkman-6866" => array("type"=>"city", "pop"=>71, "name_ru"=>"Киркман", "name_en"=>"Kirkman", "lang"=>"en", "lat"=>41.728993, "lon"=>-95.266284),
"/US/US-IA/kirkville-6867" => array("type"=>"city", "pop"=>212, "name_ru"=>"Кирквилл", "name_en"=>"Kirkville", "lang"=>"en", "lat"=>41.142850, "lon"=>-92.502293),
"/US/US-IA/kiron-6868" => array("type"=>"city", "pop"=>255, "name_ru"=>"Кирон", "name_en"=>"Kiron", "lang"=>"en", "lat"=>42.193816, "lon"=>-95.326706),
"/US/US-IA/klemme-6869" => array("type"=>"city", "pop"=>531, "name_ru"=>"Клемм", "name_en"=>"Klemme", "lang"=>"en", "lat"=>43.009149, "lon"=>-93.601096),
"/US/US-IA/knierim-6870" => array("type"=>"city", "pop"=>51, "name_ru"=>"Найрим", "name_en"=>"Knierim", "lang"=>"en", "lat"=>42.456075, "lon"=>-94.456254),
"/US/US-IA/knoxville-6871" => array("type"=>"city", "pop"=>7254, "name_ru"=>"Ноксвилл", "name_en"=>"Knoxville", "lang"=>"en", "lat"=>41.319100, "lon"=>-93.101306),
"/US/US-IA/la-motte-6883" => array("type"=>"city", "pop"=>271, "name_ru"=>"Ла Мотт", "name_en"=>"La Motte", "lang"=>"en", "lat"=>42.294921, "lon"=>-90.621260),
"/US/US-IA/la-porte-city-6886" => array("type"=>"city", "pop"=>2238, "name_ru"=>"Ла Порт Сити", "name_en"=>"La Porte City", "lang"=>"en", "lat"=>42.313077, "lon"=>-92.191201),
"/US/US-IA/lacona-6872" => array("type"=>"city", "pop"=>349, "name_ru"=>"Лакона", "name_en"=>"Lacona", "lang"=>"en", "lat"=>41.189919, "lon"=>-93.384185),
"/US/US-IA/ladora-6873" => array("type"=>"city", "pop"=>286, "name_ru"=>"Ладора", "name_en"=>"Ladora", "lang"=>"en", "lat"=>41.755806, "lon"=>-92.185401),
"/US/US-IA/lake-city-6874" => array("type"=>"city", "pop"=>1618, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>42.267750, "lon"=>-94.730559),
"/US/US-IA/lake-mills-6875" => array("type"=>"city", "pop"=>1977, "name_ru"=>"Лейк Миллс", "name_en"=>"Lake Mills", "lang"=>"en", "lat"=>43.416168, "lon"=>-93.532100),
"/US/US-IA/lake-park-6876" => array("type"=>"city", "pop"=>973, "name_ru"=>"Лэйк Парк", "name_en"=>"Lake Park", "lang"=>"en", "lat"=>43.448421, "lon"=>-95.324828),
"/US/US-IA/lake-view-6878" => array("type"=>"city", "pop"=>1196, "name_ru"=>"Лейк Вью", "name_en"=>"Lake View", "lang"=>"en", "lat"=>42.306171, "lon"=>-95.046625),
"/US/US-IA/lakeside-6877" => array("type"=>"city", "pop"=>509, "name_ru"=>"Лэйксайд", "name_en"=>"Lakeside", "lang"=>"en", "lat"=>42.619704, "lon"=>-95.173878),
"/US/US-IA/lakota-6879" => array("type"=>"city", "pop"=>240, "name_ru"=>"Лакота", "name_en"=>"Lakota", "lang"=>"en", "lat"=>43.377995, "lon"=>-94.094038),
"/US/US-IA/lambs-grove-6880" => array("type"=>"city", "pop"=>208, "name_ru"=>"Люмс Гров", "name_en"=>"Lambs Grove", "lang"=>"en", "lat"=>41.700339, "lon"=>-93.078292),
"/US/US-IA/lamoni-6881" => array("type"=>"city", "pop"=>2450, "name_ru"=>"Ламони", "name_en"=>"Lamoni", "lang"=>"en", "lat"=>40.620650, "lon"=>-93.937291),
"/US/US-IA/lamont-6882" => array("type"=>"city", "pop"=>485, "name_ru"=>"Ламонт", "name_en"=>"Lamont", "lang"=>"en", "lat"=>42.598697, "lon"=>-91.640822),
"/US/US-IA/lanesboro-6884" => array("type"=>"city", "pop"=>129, "name_ru"=>"Лейнсборо", "name_en"=>"Lanesboro", "lang"=>"en", "lat"=>42.183557, "lon"=>-94.691048),
"/US/US-IA/lansing-6885" => array("type"=>"city", "pop"=>959, "name_ru"=>"Ленсинг", "name_en"=>"Lansing", "lang"=>"en", "lat"=>43.361700, "lon"=>-91.225757),
"/US/US-IA/larchwood-6887" => array("type"=>"city", "pop"=>773, "name_ru"=>"Ларчвуд", "name_en"=>"Larchwood", "lang"=>"en", "lat"=>43.455267, "lon"=>-96.437381),
"/US/US-IA/larrabee-6888" => array("type"=>"city", "pop"=>127, "name_ru"=>"Ларраби", "name_en"=>"Larrabee", "lang"=>"en", "lat"=>42.861304, "lon"=>-95.544851),
"/US/US-IA/latimer-6889" => array("type"=>"city", "pop"=>537, "name_ru"=>"Латимер", "name_en"=>"Latimer", "lang"=>"en", "lat"=>42.763200, "lon"=>-93.364995),
"/US/US-IA/laurel-6890" => array("type"=>"city", "pop"=>266, "name_ru"=>"Лорел", "name_en"=>"Laurel", "lang"=>"en", "lat"=>41.883754, "lon"=>-92.921328),
"/US/US-IA/laurens-6891" => array("type"=>"city", "pop"=>1260, "name_ru"=>"Лоренс", "name_en"=>"Laurens", "lang"=>"en", "lat"=>42.847753, "lon"=>-94.848034),
"/US/US-IA/lawler-6892" => array("type"=>"city", "pop"=>418, "name_ru"=>"Лолер", "name_en"=>"Lawler", "lang"=>"en", "lat"=>43.071729, "lon"=>-92.153792),
"/US/US-IA/lawton-6893" => array("type"=>"city", "pop"=>768, "name_ru"=>"Лотон", "name_en"=>"Lawton", "lang"=>"en", "lat"=>42.478867, "lon"=>-96.184392),
"/US/US-IA/le-claire-6894" => array("type"=>"city", "pop"=>2994, "name_ru"=>"Ле Клэр", "name_en"=>"Le Claire", "lang"=>"en", "lat"=>41.598817, "lon"=>-90.364558),
"/US/US-IA/le-grand-6896" => array("type"=>"city", "pop"=>975, "name_ru"=>"Ле Гранд", "name_en"=>"Le Grand", "lang"=>"en", "lat"=>42.007172, "lon"=>-92.775078),
"/US/US-IA/le-mars-6900" => array("type"=>"city", "pop"=>9127, "name_ru"=>"Ле Марс", "name_en"=>"Le Mars", "lang"=>"en", "lat"=>42.787599, "lon"=>-96.169904),
"/US/US-IA/le-roy-6903" => array("type"=>"city", "pop"=>8, "name_ru"=>"Ле Рой", "name_en"=>"Le Roy", "lang"=>"en", "lat"=>40.878910, "lon"=>-93.592184),
"/US/US-IA/ledyard-6895" => array("type"=>"city", "pop"=>129, "name_ru"=>"Ледйард", "name_en"=>"Ledyard", "lang"=>"en", "lat"=>43.420259, "lon"=>-94.158721),
"/US/US-IA/lehigh-6897" => array("type"=>"city", "pop"=>451, "name_ru"=>"Лихай", "name_en"=>"Lehigh", "lang"=>"en", "lat"=>42.355863, "lon"=>-94.053074),
"/US/US-IA/leighton-6898" => array("type"=>"city", "pop"=>168, "name_ru"=>"Лейтон", "name_en"=>"Leighton", "lang"=>"en", "lat"=>41.338528, "lon"=>-92.786605),
"/US/US-IA/leland-6899" => array("type"=>"city", "pop"=>258, "name_ru"=>"Леленд", "name_en"=>"Leland", "lang"=>"en", "lat"=>43.333713, "lon"=>-93.637220),
"/US/US-IA/lenox-6901" => array("type"=>"city", "pop"=>1244, "name_ru"=>"Ленокс", "name_en"=>"Lenox", "lang"=>"en", "lat"=>40.883185, "lon"=>-94.558180),
"/US/US-IA/leon-6902" => array("type"=>"city", "pop"=>1884, "name_ru"=>"Леон", "name_en"=>"Leon", "lang"=>"en", "lat"=>40.741308, "lon"=>-93.753255),
"/US/US-IA/lester-6904" => array("type"=>"city", "pop"=>239, "name_ru"=>"Лестер", "name_en"=>"Lester", "lang"=>"en", "lat"=>43.439231, "lon"=>-96.330175),
"/US/US-IA/letts-6905" => array("type"=>"city", "pop"=>381, "name_ru"=>"Леттс", "name_en"=>"Letts", "lang"=>"en", "lat"=>41.330049, "lon"=>-91.234754),
"/US/US-IA/lewis-6906" => array("type"=>"city", "pop"=>401, "name_ru"=>"Льюис", "name_en"=>"Lewis", "lang"=>"en", "lat"=>41.306270, "lon"=>-95.084230),
"/US/US-IA/libertyville-6907" => array("type"=>"city", "pop"=>296, "name_ru"=>"Либертивилл", "name_en"=>"Libertyville", "lang"=>"en", "lat"=>40.958600, "lon"=>-92.048925),
"/US/US-IA/lidderdale-6908" => array("type"=>"city", "pop"=>175, "name_ru"=>"Лиддердейл", "name_en"=>"Lidderdale", "lang"=>"en", "lat"=>42.123930, "lon"=>-94.784208),
"/US/US-IA/lime-springs-6909" => array("type"=>"city", "pop"=>474, "name_ru"=>"Лайм Спрингс", "name_en"=>"Lime Springs", "lang"=>"en", "lat"=>43.450229, "lon"=>-92.281132),
"/US/US-IA/lincoln-6910" => array("type"=>"city", "pop"=>156, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>42.263233, "lon"=>-92.690987),
"/US/US-IA/linden-6911" => array("type"=>"city", "pop"=>275, "name_ru"=>"Линден", "name_en"=>"Linden", "lang"=>"en", "lat"=>41.643017, "lon"=>-94.270074),
"/US/US-IA/lineville-6912" => array("type"=>"city", "pop"=>225, "name_ru"=>"Лайнвилл", "name_en"=>"Lineville", "lang"=>"en", "lat"=>40.585965, "lon"=>-93.523615),
"/US/US-IA/linn-grove-6913" => array("type"=>"city", "pop"=>211, "name_ru"=>"Линн Гров", "name_en"=>"Linn Grove", "lang"=>"en", "lat"=>42.892560, "lon"=>-95.241718),
"/US/US-IA/lisbon-6914" => array("type"=>"city", "pop"=>2029, "name_ru"=>"Лисбон", "name_en"=>"Lisbon", "lang"=>"en", "lat"=>41.921238, "lon"=>-91.391137),
"/US/US-IA/liscomb-6915" => array("type"=>"city", "pop"=>271, "name_ru"=>"Лискомб", "name_en"=>"Liscomb", "lang"=>"en", "lat"=>42.190625, "lon"=>-93.005461),
"/US/US-IA/little-rock-6917" => array("type"=>"city", "pop"=>451, "name_ru"=>"Литл Рок", "name_en"=>"Little Rock", "lang"=>"en", "lat"=>43.444556, "lon"=>-95.880889),
"/US/US-IA/little-sioux-6918" => array("type"=>"city", "pop"=>196, "name_ru"=>"Литл Сиу", "name_en"=>"Little Sioux", "lang"=>"en", "lat"=>41.808269, "lon"=>-96.026743),
"/US/US-IA/littleport-6916" => array("type"=>"city", "pop"=>88, "name_ru"=>"Литлпорт", "name_en"=>"Littleport", "lang"=>"en", "lat"=>42.753117, "lon"=>-91.369046),
"/US/US-IA/livermore-6919" => array("type"=>"city", "pop"=>361, "name_ru"=>"Ливермор", "name_en"=>"Livermore", "lang"=>"en", "lat"=>42.867843, "lon"=>-94.183808),
"/US/US-IA/lockridge-6920" => array("type"=>"city", "pop"=>259, "name_ru"=>"Локридж", "name_en"=>"Lockridge", "lang"=>"en", "lat"=>40.993600, "lon"=>-91.747559),
"/US/US-IA/logan-6921" => array("type"=>"city", "pop"=>1411, "name_ru"=>"Логан", "name_en"=>"Logan", "lang"=>"en", "lat"=>41.644635, "lon"=>-95.791008),
"/US/US-IA/lohrville-6922" => array("type"=>"city", "pop"=>350, "name_ru"=>"Лорвилл", "name_en"=>"Lohrville", "lang"=>"en", "lat"=>42.268043, "lon"=>-94.550934),
"/US/US-IA/lone-rock-6923" => array("type"=>"city", "pop"=>146, "name_ru"=>"Лоун Рок", "name_en"=>"Lone Rock", "lang"=>"en", "lat"=>43.220402, "lon"=>-94.324859),
"/US/US-IA/lone-tree-6924" => array("type"=>"city", "pop"=>1184, "name_ru"=>"Лоун Три", "name_en"=>"Lone Tree", "lang"=>"en", "lat"=>41.485839, "lon"=>-91.424836),
"/US/US-IA/long-grove-6925" => array("type"=>"city", "pop"=>758, "name_ru"=>"Лонг Гров", "name_en"=>"Long Grove", "lang"=>"en", "lat"=>41.694500, "lon"=>-90.580444),
"/US/US-IA/lorimor-6926" => array("type"=>"city", "pop"=>399, "name_ru"=>"Лоримор", "name_en"=>"Lorimor", "lang"=>"en", "lat"=>41.127553, "lon"=>-94.056992),
"/US/US-IA/lost-nation-6927" => array("type"=>"city", "pop"=>461, "name_ru"=>"Лост Нейшен", "name_en"=>"Lost Nation", "lang"=>"en", "lat"=>41.966056, "lon"=>-90.817468),
"/US/US-IA/lovilia-6928" => array("type"=>"city", "pop"=>528, "name_ru"=>"Ловилия", "name_en"=>"Lovilia", "lang"=>"en", "lat"=>41.135437, "lon"=>-92.907941),
"/US/US-IA/low-moor-6930" => array("type"=>"city", "pop"=>222, "name_ru"=>"Лоу Мур", "name_en"=>"Low Moor", "lang"=>"en", "lat"=>41.802361, "lon"=>-90.354286),
"/US/US-IA/lowden-6929" => array("type"=>"city", "pop"=>804, "name_ru"=>"Лоуден", "name_en"=>"Lowden", "lang"=>"en", "lat"=>41.858552, "lon"=>-90.923051),
"/US/US-IA/lu-verne-6934" => array("type"=>"city", "pop"=>234, "name_ru"=>"Лу Верн", "name_en"=>"Lu Verne", "lang"=>"en", "lat"=>42.909600, "lon"=>-94.083398),
"/US/US-IA/luana-6931" => array("type"=>"city", "pop"=>240, "name_ru"=>"Луана", "name_en"=>"Luana", "lang"=>"en", "lat"=>43.060938, "lon"=>-91.454986),
"/US/US-IA/lucas-6932" => array("type"=>"city", "pop"=>245, "name_ru"=>"Лукас", "name_en"=>"Lucas", "lang"=>"en", "lat"=>41.031671, "lon"=>-93.460271),
"/US/US-IA/luther-6933" => array("type"=>"city", "pop"=>146, "name_ru"=>"Лютер", "name_en"=>"Luther", "lang"=>"en", "lat"=>41.967141, "lon"=>-93.819703),
"/US/US-IA/luxemburg-6935" => array("type"=>"city", "pop"=>240, "name_ru"=>"Люксембург", "name_en"=>"Luxemburg", "lang"=>"en", "lat"=>42.602955, "lon"=>-91.072988),
"/US/US-IA/luzerne-6936" => array("type"=>"city", "pop"=>99, "name_ru"=>"Лузерн", "name_en"=>"Luzerne", "lang"=>"en", "lat"=>41.906214, "lon"=>-92.180187),
"/US/US-IA/lynnville-6937" => array("type"=>"city", "pop"=>373, "name_ru"=>"Линнвилл", "name_en"=>"Lynnville", "lang"=>"en", "lat"=>41.576779, "lon"=>-92.786250),
"/US/US-IA/lytton-6938" => array("type"=>"city", "pop"=>261, "name_ru"=>"Литтон", "name_en"=>"Lytton", "lang"=>"en", "lat"=>42.423442, "lon"=>-94.860333),
"/US/US-IA/macedonia-6942" => array("type"=>"city", "pop"=>311, "name_ru"=>"Македония", "name_en"=>"Macedonia", "lang"=>"en", "lat"=>41.192050, "lon"=>-95.424827),
"/US/US-IA/macksburg-6945" => array("type"=>"city", "pop"=>129, "name_ru"=>"Максбург", "name_en"=>"Macksburg", "lang"=>"en", "lat"=>41.214261, "lon"=>-94.185386),
"/US/US-IA/madrid-6946" => array("type"=>"city", "pop"=>2391, "name_ru"=>"Мадрид", "name_en"=>"Madrid", "lang"=>"en", "lat"=>41.875450, "lon"=>-93.819706),
"/US/US-IA/magnolia-6947" => array("type"=>"city", "pop"=>187, "name_ru"=>"Магнолия", "name_en"=>"Magnolia", "lang"=>"en", "lat"=>41.687447, "lon"=>-95.874317),
"/US/US-IA/malcom-6948" => array("type"=>"city", "pop"=>300, "name_ru"=>"Малком", "name_en"=>"Malcom", "lang"=>"en", "lat"=>41.707524, "lon"=>-92.557842),
"/US/US-IA/mallard-6949" => array("type"=>"city", "pop"=>267, "name_ru"=>"Маллард", "name_en"=>"Mallard", "lang"=>"en", "lat"=>42.939736, "lon"=>-94.684109),
"/US/US-IA/maloy-6950" => array("type"=>"city", "pop"=>21, "name_ru"=>"Малой", "name_en"=>"Maloy", "lang"=>"en", "lat"=>40.674199, "lon"=>-94.411239),
"/US/US-IA/malvern-6951" => array("type"=>"city", "pop"=>1323, "name_ru"=>"Малверн", "name_en"=>"Malvern", "lang"=>"en", "lat"=>41.007520, "lon"=>-95.585408),
"/US/US-IA/manchester-6952" => array("type"=>"city", "pop"=>4876, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>42.485167, "lon"=>-91.455813),
"/US/US-IA/manilla-6953" => array("type"=>"city", "pop"=>808, "name_ru"=>"Манилла", "name_en"=>"Manilla", "lang"=>"en", "lat"=>41.889064, "lon"=>-95.235042),
"/US/US-IA/manly-6954" => array("type"=>"city", "pop"=>1338, "name_ru"=>"Манли", "name_en"=>"Manly", "lang"=>"en", "lat"=>43.288261, "lon"=>-93.201729),
"/US/US-IA/manning-6955" => array("type"=>"city", "pop"=>1490, "name_ru"=>"Мэннинг", "name_en"=>"Manning", "lang"=>"en", "lat"=>41.909548, "lon"=>-95.064288),
"/US/US-IA/manson-6956" => array("type"=>"city", "pop"=>1657, "name_ru"=>"Мэнсон", "name_en"=>"Manson", "lang"=>"en", "lat"=>42.528667, "lon"=>-94.539260),
"/US/US-IA/mapleton-6957" => array("type"=>"city", "pop"=>1149, "name_ru"=>"Мейплтон", "name_en"=>"Mapleton", "lang"=>"en", "lat"=>42.167717, "lon"=>-95.790779),
"/US/US-IA/maquoketa-6958" => array("type"=>"city", "pop"=>5908, "name_ru"=>"Маквокета", "name_en"=>"Maquoketa", "lang"=>"en", "lat"=>42.065327, "lon"=>-90.666103),
"/US/US-IA/marathon-6959" => array("type"=>"city", "pop"=>293, "name_ru"=>"Марафон", "name_en"=>"Marathon", "lang"=>"en", "lat"=>42.861042, "lon"=>-94.982935),
"/US/US-IA/marble-rock-6960" => array("type"=>"city", "pop"=>326, "name_ru"=>"Марбл Рок", "name_en"=>"Marble Rock", "lang"=>"en", "lat"=>42.964724, "lon"=>-92.867592),
"/US/US-IA/marcus-6961" => array("type"=>"city", "pop"=>1037, "name_ru"=>"Маркус", "name_en"=>"Marcus", "lang"=>"en", "lat"=>42.824072, "lon"=>-95.807196),
"/US/US-IA/marengo-6962" => array("type"=>"city", "pop"=>2522, "name_ru"=>"Маренго", "name_en"=>"Marengo", "lang"=>"en", "lat"=>41.797224, "lon"=>-92.069370),
"/US/US-IA/marion-6963" => array("type"=>"city", "pop"=>32821, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>42.034400, "lon"=>-91.589549),
"/US/US-IA/marne-6964" => array("type"=>"city", "pop"=>141, "name_ru"=>"Марн", "name_en"=>"Marne", "lang"=>"en", "lat"=>41.448461, "lon"=>-95.110290),
"/US/US-IA/marquette-6965" => array("type"=>"city", "pop"=>438, "name_ru"=>"Маркетт", "name_en"=>"Marquette", "lang"=>"en", "lat"=>43.042854, "lon"=>-91.184048),
"/US/US-IA/marshalltown-6966" => array("type"=>"city", "pop"=>25850, "name_ru"=>"Маршолтаун", "name_en"=>"Marshalltown", "lang"=>"en", "lat"=>42.035451, "lon"=>-92.911830),
"/US/US-IA/martelle-6967" => array("type"=>"city", "pop"=>256, "name_ru"=>"Мартелл", "name_en"=>"Martelle", "lang"=>"en", "lat"=>42.020417, "lon"=>-91.357629),
"/US/US-IA/martensdale-6968" => array("type"=>"city", "pop"=>460, "name_ru"=>"Мартенсдейл", "name_en"=>"Martensdale", "lang"=>"en", "lat"=>41.374099, "lon"=>-93.738220),
"/US/US-IA/martinsburg-6969" => array("type"=>"city", "pop"=>117, "name_ru"=>"Мартинсбург", "name_en"=>"Martinsburg", "lang"=>"en", "lat"=>41.178361, "lon"=>-92.252330),
"/US/US-IA/marysville-6970" => array("type"=>"city", "pop"=>51, "name_ru"=>"Мэрисвилл", "name_en"=>"Marysville", "lang"=>"en", "lat"=>41.182688, "lon"=>-92.953440),
"/US/US-IA/mason-city-6971" => array("type"=>"city", "pop"=>27337, "name_ru"=>"Мейсон Сити", "name_en"=>"Mason City", "lang"=>"en", "lat"=>43.151344, "lon"=>-93.198472),
"/US/US-IA/masonville-6972" => array("type"=>"city", "pop"=>101, "name_ru"=>"Мейсонвилл", "name_en"=>"Masonville", "lang"=>"en", "lat"=>42.479757, "lon"=>-91.591773),
"/US/US-IA/massena-6973" => array("type"=>"city", "pop"=>397, "name_ru"=>"Массена", "name_en"=>"Massena", "lang"=>"en", "lat"=>41.251253, "lon"=>-94.769251),
"/US/US-IA/matlock-6974" => array("type"=>"city", "pop"=>82, "name_ru"=>"Матлок", "name_en"=>"Matlock", "lang"=>"en", "lat"=>43.244532, "lon"=>-95.934227),
"/US/US-IA/maurice-6975" => array("type"=>"city", "pop"=>261, "name_ru"=>"Морис", "name_en"=>"Maurice", "lang"=>"en", "lat"=>42.966381, "lon"=>-96.182984),
"/US/US-IA/maxwell-6976" => array("type"=>"city", "pop"=>840, "name_ru"=>"Максуэлл", "name_en"=>"Maxwell", "lang"=>"en", "lat"=>41.891424, "lon"=>-93.398799),
"/US/US-IA/maynard-6977" => array("type"=>"city", "pop"=>456, "name_ru"=>"Мейнард", "name_en"=>"Maynard", "lang"=>"en", "lat"=>42.774199, "lon"=>-91.876911),
"/US/US-IA/maysville-6978" => array("type"=>"city", "pop"=>163, "name_ru"=>"Мейсвилл", "name_en"=>"Maysville", "lang"=>"en", "lat"=>41.649150, "lon"=>-90.718199),
"/US/US-IA/mccallsburg-6939" => array("type"=>"city", "pop"=>336, "name_ru"=>"Мак-Колсбург", "name_en"=>"McCallsburg", "lang"=>"en", "lat"=>42.165499, "lon"=>-93.390291),
"/US/US-IA/mccausland-6940" => array("type"=>"city", "pop"=>300, "name_ru"=>"Мак-Косленд", "name_en"=>"McCausland", "lang"=>"en", "lat"=>41.743413, "lon"=>-90.445219),
"/US/US-IA/mcclelland-6941" => array("type"=>"city", "pop"=>124, "name_ru"=>"Мак-Клелленд", "name_en"=>"McClelland", "lang"=>"en", "lat"=>41.327650, "lon"=>-95.683825),
"/US/US-IA/mcgregor-6943" => array("type"=>"city", "pop"=>831, "name_ru"=>"Мак-Грегор", "name_en"=>"McGregor", "lang"=>"en", "lat"=>43.022915, "lon"=>-91.180727),
"/US/US-IA/mcintire-6944" => array("type"=>"city", "pop"=>140, "name_ru"=>"Мак-Интайер", "name_en"=>"McIntire", "lang"=>"en", "lat"=>43.437179, "lon"=>-92.593086),
"/US/US-IA/mechanicsville-6979" => array("type"=>"city", "pop"=>1109, "name_ru"=>"Меканиксвилл", "name_en"=>"Mechanicsville", "lang"=>"en", "lat"=>41.904852, "lon"=>-91.253321),
"/US/US-IA/mediapolis-6980" => array("type"=>"city", "pop"=>1584, "name_ru"=>"Медиаполис", "name_en"=>"Mediapolis", "lang"=>"en", "lat"=>41.008199, "lon"=>-91.164525),
"/US/US-IA/melbourne-6981" => array("type"=>"city", "pop"=>808, "name_ru"=>"Мельбурн", "name_en"=>"Melbourne", "lang"=>"en", "lat"=>41.941800, "lon"=>-93.102202),
"/US/US-IA/melcher-dallas-6982" => array("type"=>"city", "pop"=>1237, "name_ru"=>"Мелчер-Даллас", "name_en"=>"Melcher-Dallas", "lang"=>"en", "lat"=>41.227831, "lon"=>-93.240848),
"/US/US-IA/melrose-6983" => array("type"=>"city", "pop"=>119, "name_ru"=>"Мелроуз", "name_en"=>"Melrose", "lang"=>"en", "lat"=>40.980841, "lon"=>-93.049328),
"/US/US-IA/melvin-6984" => array("type"=>"city", "pop"=>231, "name_ru"=>"Мэлвин", "name_en"=>"Melvin", "lang"=>"en", "lat"=>43.286710, "lon"=>-95.608549),
"/US/US-IA/menlo-6985" => array("type"=>"city", "pop"=>351, "name_ru"=>"Менло", "name_en"=>"Menlo", "lang"=>"en", "lat"=>41.522046, "lon"=>-94.403529),
"/US/US-IA/meriden-6986" => array("type"=>"city", "pop"=>168, "name_ru"=>"Мериден", "name_en"=>"Meriden", "lang"=>"en", "lat"=>42.794206, "lon"=>-95.632564),
"/US/US-IA/merrill-6987" => array("type"=>"city", "pop"=>739, "name_ru"=>"Меррил", "name_en"=>"Merrill", "lang"=>"en", "lat"=>42.720996, "lon"=>-96.249133),
"/US/US-IA/meservey-6988" => array("type"=>"city", "pop"=>223, "name_ru"=>"Мезерви", "name_en"=>"Meservey", "lang"=>"en", "lat"=>42.915070, "lon"=>-93.473190),
"/US/US-IA/middletown-6989" => array("type"=>"city", "pop"=>540, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>40.829173, "lon"=>-91.260817),
"/US/US-IA/miles-6990" => array("type"=>"city", "pop"=>441, "name_ru"=>"Майлс", "name_en"=>"Miles", "lang"=>"en", "lat"=>42.048285, "lon"=>-90.317493),
"/US/US-IA/milford-6991" => array("type"=>"city", "pop"=>2472, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>43.327042, "lon"=>-95.151798),
"/US/US-IA/millersburg-6992" => array("type"=>"city", "pop"=>187, "name_ru"=>"Миллерсберг", "name_en"=>"Millersburg", "lang"=>"en", "lat"=>41.572659, "lon"=>-92.159152),
"/US/US-IA/millerton-6993" => array("type"=>"city", "pop"=>41, "name_ru"=>"Миллертон", "name_en"=>"Millerton", "lang"=>"en", "lat"=>40.849318, "lon"=>-93.306934),
"/US/US-IA/millville-6994" => array("type"=>"city", "pop"=>21, "name_ru"=>"Миллвилл", "name_en"=>"Millville", "lang"=>"en", "lat"=>42.704840, "lon"=>-91.077171),
"/US/US-IA/milo-6995" => array("type"=>"city", "pop"=>821, "name_ru"=>"Майло", "name_en"=>"Milo", "lang"=>"en", "lat"=>41.289294, "lon"=>-93.438462),
"/US/US-IA/milton-6996" => array("type"=>"city", "pop"=>494, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>40.671284, "lon"=>-92.162518),
"/US/US-IA/minburn-6997" => array("type"=>"city", "pop"=>469, "name_ru"=>"Минберн", "name_en"=>"Minburn", "lang"=>"en", "lat"=>41.757930, "lon"=>-94.027762),
"/US/US-IA/minden-6998" => array("type"=>"city", "pop"=>553, "name_ru"=>"Минден", "name_en"=>"Minden", "lang"=>"en", "lat"=>41.467050, "lon"=>-95.541949),
"/US/US-IA/mingo-6999" => array("type"=>"city", "pop"=>263, "name_ru"=>"Минго", "name_en"=>"Mingo", "lang"=>"en", "lat"=>41.767400, "lon"=>-93.284326),
"/US/US-IA/missouri-valley-7000" => array("type"=>"city", "pop"=>2751, "name_ru"=>"Миссури Валли", "name_en"=>"Missouri Valley", "lang"=>"en", "lat"=>41.558921, "lon"=>-95.895137),
"/US/US-IA/mitchell-7001" => array("type"=>"city", "pop"=>147, "name_ru"=>"Митчелл", "name_en"=>"Mitchell", "lang"=>"en", "lat"=>43.321160, "lon"=>-92.870153),
"/US/US-IA/mitchellville-7002" => array("type"=>"city", "pop"=>2152, "name_ru"=>"Митчелвилл", "name_en"=>"Mitchellville", "lang"=>"en", "lat"=>41.661517, "lon"=>-93.360868),
"/US/US-IA/modale-7003" => array("type"=>"city", "pop"=>281, "name_ru"=>"Модейл", "name_en"=>"Modale", "lang"=>"en", "lat"=>41.619246, "lon"=>-96.012220),
"/US/US-IA/mondamin-7004" => array("type"=>"city", "pop"=>388, "name_ru"=>"Мондамин", "name_en"=>"Mondamin", "lang"=>"en", "lat"=>41.709843, "lon"=>-96.020779),
"/US/US-IA/moneta-7005" => array("type"=>"city", "pop"=>29, "name_ru"=>"Монета", "name_en"=>"Moneta", "lang"=>"en", "lat"=>43.132487, "lon"=>-95.392419),
"/US/US-IA/monmouth-7006" => array("type"=>"city", "pop"=>176, "name_ru"=>"Монмаут", "name_en"=>"Monmouth", "lang"=>"en", "lat"=>42.073687, "lon"=>-90.881751),
"/US/US-IA/monona-7007" => array("type"=>"city", "pop"=>1424, "name_ru"=>"Монона", "name_en"=>"Monona", "lang"=>"en", "lat"=>43.051276, "lon"=>-91.390392),
"/US/US-IA/monroe-7008" => array("type"=>"city", "pop"=>1826, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>41.520160, "lon"=>-93.104547),
"/US/US-IA/montezuma-7009" => array("type"=>"city", "pop"=>1362, "name_ru"=>"Мотезума", "name_en"=>"Montezuma", "lang"=>"en", "lat"=>41.582854, "lon"=>-92.527492),
"/US/US-IA/monticello-7010" => array("type"=>"city", "pop"=>3653, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>42.237655, "lon"=>-91.191010),
"/US/US-IA/montour-7011" => array("type"=>"city", "pop"=>277, "name_ru"=>"Монтор", "name_en"=>"Montour", "lang"=>"en", "lat"=>41.980516, "lon"=>-92.715422),
"/US/US-IA/montrose-7012" => array("type"=>"city", "pop"=>918, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>40.525386, "lon"=>-91.416238),
"/US/US-IA/moorhead-7013" => array("type"=>"city", "pop"=>196, "name_ru"=>"Мурхэд", "name_en"=>"Moorhead", "lang"=>"en", "lat"=>41.924284, "lon"=>-95.850407),
"/US/US-IA/moorland-7014" => array("type"=>"city", "pop"=>182, "name_ru"=>"Мурленд", "name_en"=>"Moorland", "lang"=>"en", "lat"=>42.441027, "lon"=>-94.294071),
"/US/US-IA/moravia-7015" => array("type"=>"city", "pop"=>684, "name_ru"=>"Моравия", "name_en"=>"Moravia", "lang"=>"en", "lat"=>40.892111, "lon"=>-92.818261),
"/US/US-IA/morley-7016" => array("type"=>"city", "pop"=>91, "name_ru"=>"Морли", "name_en"=>"Morley", "lang"=>"en", "lat"=>42.006548, "lon"=>-91.246558),
"/US/US-IA/morning-sun-7017" => array("type"=>"city", "pop"=>843, "name_ru"=>"Морнинг Сан", "name_en"=>"Morning Sun", "lang"=>"en", "lat"=>41.093999, "lon"=>-91.251550),
"/US/US-IA/morrison-7018" => array("type"=>"city", "pop"=>86, "name_ru"=>"Моррисон", "name_en"=>"Morrison", "lang"=>"en", "lat"=>42.343393, "lon"=>-92.672932),
"/US/US-IA/moulton-7019" => array("type"=>"city", "pop"=>597, "name_ru"=>"Молтон", "name_en"=>"Moulton", "lang"=>"en", "lat"=>40.685275, "lon"=>-92.679000),
"/US/US-IA/mount-auburn-7020" => array("type"=>"city", "pop"=>161, "name_ru"=>"Маунт Обурн", "name_en"=>"Mount Auburn", "lang"=>"en", "lat"=>42.256818, "lon"=>-92.093293),
"/US/US-IA/mount-ayr-7021" => array("type"=>"city", "pop"=>1711, "name_ru"=>"Маунт Эр", "name_en"=>"Mount Ayr", "lang"=>"en", "lat"=>40.714217, "lon"=>-94.237907),
"/US/US-IA/mount-pleasant-7022" => array("type"=>"city", "pop"=>8775, "name_ru"=>"Маунт Плезант", "name_en"=>"Mount Pleasant", "lang"=>"en", "lat"=>40.963150, "lon"=>-91.546474),
"/US/US-IA/mount-sterling-7023" => array("type"=>"city", "pop"=>37, "name_ru"=>"Маунт Стерлинг", "name_en"=>"Mount Sterling", "lang"=>"en", "lat"=>40.618981, "lon"=>-91.938905),
"/US/US-IA/mount-union-7024" => array("type"=>"city", "pop"=>125, "name_ru"=>"Маунт Юнион", "name_en"=>"Mount Union", "lang"=>"en", "lat"=>41.057440, "lon"=>-91.390612),
"/US/US-IA/mount-vernon-7025" => array("type"=>"city", "pop"=>4158, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>41.923129, "lon"=>-91.423760),
"/US/US-IA/moville-7026" => array("type"=>"city", "pop"=>1473, "name_ru"=>"Мовилл", "name_en"=>"Moville", "lang"=>"en", "lat"=>42.489590, "lon"=>-96.066824),
"/US/US-IA/murray-7027" => array("type"=>"city", "pop"=>796, "name_ru"=>"Мюррей", "name_en"=>"Murray", "lang"=>"en", "lat"=>41.041273, "lon"=>-93.948430),
"/US/US-IA/muscatine-7028" => array("type"=>"city", "pop"=>22504, "name_ru"=>"Маскатин", "name_en"=>"Muscatine", "lang"=>"en", "lat"=>41.417720, "lon"=>-91.071703),
"/US/US-IA/mystic-7029" => array("type"=>"city", "pop"=>579, "name_ru"=>"Мистик", "name_en"=>"Mystic", "lang"=>"en", "lat"=>40.779098, "lon"=>-92.943749),
"/US/US-IA/nashua-7030" => array("type"=>"city", "pop"=>1517, "name_ru"=>"Нашуа", "name_en"=>"Nashua", "lang"=>"en", "lat"=>42.950553, "lon"=>-92.539061),
"/US/US-IA/nemaha-7031" => array("type"=>"city", "pop"=>89, "name_ru"=>"Немаха", "name_en"=>"Nemaha", "lang"=>"en", "lat"=>42.515030, "lon"=>-95.087718),
"/US/US-IA/neola-7032" => array("type"=>"city", "pop"=>837, "name_ru"=>"Неола", "name_en"=>"Neola", "lang"=>"en", "lat"=>41.450950, "lon"=>-95.617249),
"/US/US-IA/nevada-7033" => array("type"=>"city", "pop"=>6688, "name_ru"=>"Невада", "name_en"=>"Nevada", "lang"=>"en", "lat"=>42.018182, "lon"=>-93.450168),
"/US/US-IA/new-albin-7034" => array("type"=>"city", "pop"=>510, "name_ru"=>"Нью Албин", "name_en"=>"New Albin", "lang"=>"en", "lat"=>43.497002, "lon"=>-91.288469),
"/US/US-IA/new-hampton-7037" => array("type"=>"city", "pop"=>3400, "name_ru"=>"Нью Хэмптон", "name_en"=>"New Hampton", "lang"=>"en", "lat"=>43.059727, "lon"=>-92.314527),
"/US/US-IA/new-hartford-7038" => array("type"=>"city", "pop"=>613, "name_ru"=>"Нью Хартфорд", "name_en"=>"New Hartford", "lang"=>"en", "lat"=>42.567208, "lon"=>-92.622447),
"/US/US-IA/new-liberty-7039" => array("type"=>"city", "pop"=>134, "name_ru"=>"Нью Либерти", "name_en"=>"New Liberty", "lang"=>"en", "lat"=>41.716614, "lon"=>-90.878562),
"/US/US-IA/new-london-7040" => array("type"=>"city", "pop"=>1841, "name_ru"=>"Нью Лондон", "name_en"=>"New London", "lang"=>"en", "lat"=>40.927501, "lon"=>-91.407308),
"/US/US-IA/new-market-7041" => array("type"=>"city", "pop"=>394, "name_ru"=>"Нью Маркет", "name_en"=>"New Market", "lang"=>"en", "lat"=>40.732427, "lon"=>-94.900034),
"/US/US-IA/new-providence-7042" => array("type"=>"city", "pop"=>213, "name_ru"=>"Нью Провиденс", "name_en"=>"New Providence", "lang"=>"en", "lat"=>42.281920, "lon"=>-93.172343),
"/US/US-IA/new-sharon-7043" => array("type"=>"city", "pop"=>1299, "name_ru"=>"Нью Шэрон", "name_en"=>"New Sharon", "lang"=>"en", "lat"=>41.470194, "lon"=>-92.650648),
"/US/US-IA/new-vienna-7045" => array("type"=>"city", "pop"=>402, "name_ru"=>"Нью Вьенна", "name_en"=>"New Vienna", "lang"=>"en", "lat"=>42.547837, "lon"=>-91.113435),
"/US/US-IA/new-virginia-7046" => array("type"=>"city", "pop"=>487, "name_ru"=>"Нью Вирджиния", "name_en"=>"New Virginia", "lang"=>"en", "lat"=>41.180942, "lon"=>-93.730470),
"/US/US-IA/newell-7035" => array("type"=>"city", "pop"=>869, "name_ru"=>"Ньюуэлл", "name_en"=>"Newell", "lang"=>"en", "lat"=>42.610625, "lon"=>-95.003800),
"/US/US-IA/newhall-7036" => array("type"=>"city", "pop"=>951, "name_ru"=>"Ньюхолл", "name_en"=>"Newhall", "lang"=>"en", "lat"=>41.993021, "lon"=>-91.966712),
"/US/US-IA/newton-7044" => array("type"=>"city", "pop"=>15042, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>41.695104, "lon"=>-93.045104),
"/US/US-IA/nichols-7047" => array("type"=>"city", "pop"=>357, "name_ru"=>"Николс", "name_en"=>"Nichols", "lang"=>"en", "lat"=>41.479090, "lon"=>-91.308134),
"/US/US-IA/nodaway-7048" => array("type"=>"city", "pop"=>121, "name_ru"=>"Нодэуэй", "name_en"=>"Nodaway", "lang"=>"en", "lat"=>40.936458, "lon"=>-94.895151),
"/US/US-IA/nora-springs-7049" => array("type"=>"city", "pop"=>1445, "name_ru"=>"Нора Спрингс", "name_en"=>"Nora Springs", "lang"=>"en", "lat"=>43.144042, "lon"=>-93.008729),
"/US/US-IA/north-buena-vista-7051" => array("type"=>"city", "pop"=>94, "name_ru"=>"Норт Буэна Виста", "name_en"=>"North Buena Vista", "lang"=>"en", "lat"=>42.682471, "lon"=>-90.956379),
"/US/US-IA/north-english-7052" => array("type"=>"city", "pop"=>989, "name_ru"=>"Норт Инглиш", "name_en"=>"North English", "lang"=>"en", "lat"=>41.515868, "lon"=>-92.078120),
"/US/US-IA/north-liberty-7053" => array("type"=>"city", "pop"=>11561, "name_ru"=>"Норт Либерти", "name_en"=>"North Liberty", "lang"=>"en", "lat"=>41.741514, "lon"=>-91.609326),
"/US/US-IA/north-washington-7054" => array("type"=>"city", "pop"=>103, "name_ru"=>"Норт Вошингтон", "name_en"=>"North Washington", "lang"=>"en", "lat"=>43.116280, "lon"=>-92.415000),
"/US/US-IA/northboro-7050" => array("type"=>"city", "pop"=>55, "name_ru"=>"Нортборо", "name_en"=>"Northboro", "lang"=>"en", "lat"=>40.608752, "lon"=>-95.293091),
"/US/US-IA/northwood-7055" => array("type"=>"city", "pop"=>2012, "name_ru"=>"Нортвуд", "name_en"=>"Northwood", "lang"=>"en", "lat"=>43.444615, "lon"=>-93.216361),
"/US/US-IA/norwalk-7056" => array("type"=>"city", "pop"=>8821, "name_ru"=>"Норуолк", "name_en"=>"Norwalk", "lang"=>"en", "lat"=>41.498053, "lon"=>-93.680877),
"/US/US-IA/norway-7057" => array("type"=>"city", "pop"=>597, "name_ru"=>"Норуэй", "name_en"=>"Norway", "lang"=>"en", "lat"=>41.902791, "lon"=>-91.921939),
"/US/US-IA/numa-7058" => array("type"=>"city", "pop"=>106, "name_ru"=>"Нума", "name_en"=>"Numa", "lang"=>"en", "lat"=>40.685427, "lon"=>-92.980551),
"/US/US-IA/oakland-7059" => array("type"=>"city", "pop"=>1417, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>41.308500, "lon"=>-95.396687),
"/US/US-IA/oakland-acres-7060" => array("type"=>"city", "pop"=>161, "name_ru"=>"Оукленд Эйкрес", "name_en"=>"Oakland Acres", "lang"=>"en", "lat"=>41.719327, "lon"=>-92.819980),
"/US/US-IA/oakville-7061" => array("type"=>"city", "pop"=>413, "name_ru"=>"Оуквилл", "name_en"=>"Oakville", "lang"=>"en", "lat"=>41.097820, "lon"=>-91.043480),
"/US/US-IA/ocheyedan-7062" => array("type"=>"city", "pop"=>488, "name_ru"=>"Ошайедан", "name_en"=>"Ocheyedan", "lang"=>"en", "lat"=>43.418230, "lon"=>-95.536537),
"/US/US-IA/odebolt-7063" => array("type"=>"city", "pop"=>1005, "name_ru"=>"Одеболт", "name_en"=>"Odebolt", "lang"=>"en", "lat"=>42.311521, "lon"=>-95.253789),
"/US/US-IA/oelwein-7064" => array("type"=>"city", "pop"=>6082, "name_ru"=>"Оелвейн", "name_en"=>"Oelwein", "lang"=>"en", "lat"=>42.676296, "lon"=>-91.913937),
"/US/US-IA/ogden-7065" => array("type"=>"city", "pop"=>1974, "name_ru"=>"Огден", "name_en"=>"Ogden", "lang"=>"en", "lat"=>42.039399, "lon"=>-94.028457),
"/US/US-IA/okoboji-7066" => array("type"=>"city", "pop"=>861, "name_ru"=>"Окободжи", "name_en"=>"Okoboji", "lang"=>"en", "lat"=>43.392120, "lon"=>-95.136181),
"/US/US-IA/olds-7067" => array("type"=>"city", "pop"=>238, "name_ru"=>"Олдс", "name_en"=>"Olds", "lang"=>"en", "lat"=>41.133322, "lon"=>-91.545435),
"/US/US-IA/olin-7068" => array("type"=>"city", "pop"=>703, "name_ru"=>"Олин", "name_en"=>"Olin", "lang"=>"en", "lat"=>41.997577, "lon"=>-91.141109),
"/US/US-IA/ollie-7069" => array("type"=>"city", "pop"=>162, "name_ru"=>"Олли", "name_en"=>"Ollie", "lang"=>"en", "lat"=>41.198829, "lon"=>-92.092786),
"/US/US-IA/onawa-7070" => array("type"=>"city", "pop"=>2746, "name_ru"=>"Онава", "name_en"=>"Onawa", "lang"=>"en", "lat"=>42.027413, "lon"=>-96.088057),
"/US/US-IA/oneida-7071" => array("type"=>"city", "pop"=>49, "name_ru"=>"Онида", "name_en"=>"Oneida", "lang"=>"en", "lat"=>42.541139, "lon"=>-91.352927),
"/US/US-IA/onslow-7072" => array("type"=>"city", "pop"=>213, "name_ru"=>"Онслоу", "name_en"=>"Onslow", "lang"=>"en", "lat"=>42.108039, "lon"=>-91.013682),
"/US/US-IA/orange-city-7073" => array("type"=>"city", "pop"=>5891, "name_ru"=>"Оранж Сити", "name_en"=>"Orange City", "lang"=>"en", "lat"=>43.003260, "lon"=>-96.058508),
"/US/US-IA/orchard-7074" => array("type"=>"city", "pop"=>86, "name_ru"=>"Орчард", "name_en"=>"Orchard", "lang"=>"en", "lat"=>43.227022, "lon"=>-92.774373),
"/US/US-IA/orient-7075" => array("type"=>"city", "pop"=>373, "name_ru"=>"Ориент", "name_en"=>"Orient", "lang"=>"en", "lat"=>41.202828, "lon"=>-94.417883),
"/US/US-IA/orleans-7076" => array("type"=>"city", "pop"=>602, "name_ru"=>"Орлеанс", "name_en"=>"Orleans", "lang"=>"en", "lat"=>43.445239, "lon"=>-95.098659),
"/US/US-IA/osage-7077" => array("type"=>"city", "pop"=>3427, "name_ru"=>"Осейдж", "name_en"=>"Osage", "lang"=>"en", "lat"=>43.283607, "lon"=>-92.811904),
"/US/US-IA/osceola-7078" => array("type"=>"city", "pop"=>4714, "name_ru"=>"Осеола", "name_en"=>"Osceola", "lang"=>"en", "lat"=>41.031279, "lon"=>-93.773837),
"/US/US-IA/oskaloosa-7079" => array("type"=>"city", "pop"=>11069, "name_ru"=>"Оскалуза", "name_en"=>"Oskaloosa", "lang"=>"en", "lat"=>41.292968, "lon"=>-92.639608),
"/US/US-IA/ossian-7080" => array("type"=>"city", "pop"=>811, "name_ru"=>"Оссиан", "name_en"=>"Ossian", "lang"=>"en", "lat"=>43.146826, "lon"=>-91.764518),
"/US/US-IA/osterdock-7081" => array("type"=>"city", "pop"=>49, "name_ru"=>"Оустердок", "name_en"=>"Osterdock", "lang"=>"en", "lat"=>42.732481, "lon"=>-91.158611),
"/US/US-IA/otho-7082" => array("type"=>"city", "pop"=>535, "name_ru"=>"Ото", "name_en"=>"Otho", "lang"=>"en", "lat"=>42.420936, "lon"=>-94.147627),
"/US/US-IA/oto-7083" => array("type"=>"city", "pop"=>144, "name_ru"=>"Ото", "name_en"=>"Oto", "lang"=>"en", "lat"=>42.281454, "lon"=>-95.893777),
"/US/US-IA/ottosen-7084" => array("type"=>"city", "pop"=>45, "name_ru"=>"Оттосен", "name_en"=>"Ottosen", "lang"=>"en", "lat"=>42.899605, "lon"=>-94.375327),
"/US/US-IA/ottumwa-7085" => array("type"=>"city", "pop"=>24342, "name_ru"=>"Оттумва", "name_en"=>"Ottumwa", "lang"=>"en", "lat"=>41.019928, "lon"=>-92.418227),
"/US/US-IA/owasa-7086" => array("type"=>"city", "pop"=>36, "name_ru"=>"Оваза", "name_en"=>"Owasa", "lang"=>"en", "lat"=>42.431145, "lon"=>-93.205112),
"/US/US-IA/oxford-7087" => array("type"=>"city", "pop"=>693, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>41.724653, "lon"=>-91.789682),
"/US/US-IA/oxford-junction-7088" => array("type"=>"city", "pop"=>544, "name_ru"=>"Оксфорд Джанкшен", "name_en"=>"Oxford Junction", "lang"=>"en", "lat"=>41.984604, "lon"=>-90.954347),
"/US/US-IA/oyens-7089" => array("type"=>"city", "pop"=>119, "name_ru"=>"Ойенс", "name_en"=>"Oyens", "lang"=>"en", "lat"=>42.819443, "lon"=>-96.057493),
"/US/US-IA/pacific-junction-7090" => array("type"=>"city", "pop"=>510, "name_ru"=>"Пасифик Джанкшен", "name_en"=>"Pacific Junction", "lang"=>"en", "lat"=>41.018320, "lon"=>-95.799599),
"/US/US-IA/packwood-7091" => array("type"=>"city", "pop"=>212, "name_ru"=>"Пэквуд", "name_en"=>"Packwood", "lang"=>"en", "lat"=>41.132457, "lon"=>-92.081797),
"/US/US-IA/palmer-7092" => array("type"=>"city", "pop"=>175, "name_ru"=>"Палмер", "name_en"=>"Palmer", "lang"=>"en", "lat"=>42.628805, "lon"=>-94.598312),
"/US/US-IA/palo-7093" => array("type"=>"city", "pop"=>875, "name_ru"=>"Пало", "name_en"=>"Palo", "lang"=>"en", "lat"=>42.064192, "lon"=>-91.795987),
"/US/US-IA/panama-7094" => array("type"=>"city", "pop"=>187, "name_ru"=>"Панама", "name_en"=>"Panama", "lang"=>"en", "lat"=>41.726724, "lon"=>-95.475273),
"/US/US-IA/panora-7095" => array("type"=>"city", "pop"=>1161, "name_ru"=>"Панора", "name_en"=>"Panora", "lang"=>"en", "lat"=>41.691038, "lon"=>-94.364942),
"/US/US-IA/panorama-park-7096" => array("type"=>"city", "pop"=>128, "name_ru"=>"Панорама Парк", "name_en"=>"Panorama Park", "lang"=>"en", "lat"=>41.555093, "lon"=>-90.454706),
"/US/US-IA/park-view-7098" => array("type"=>"cdp", "pop"=>2192, "name_ru"=>"Парк Вью", "name_en"=>"Park View", "lang"=>"en", "lat"=>41.689534, "lon"=>-90.540771),
"/US/US-IA/parkersburg-7097" => array("type"=>"city", "pop"=>1861, "name_ru"=>"Паркерсбург", "name_en"=>"Parkersburg", "lang"=>"en", "lat"=>42.573569, "lon"=>-92.782511),
"/US/US-IA/parnell-7099" => array("type"=>"city", "pop"=>214, "name_ru"=>"Парнелл", "name_en"=>"Parnell", "lang"=>"en", "lat"=>41.583469, "lon"=>-92.004753),
"/US/US-IA/paton-7100" => array("type"=>"city", "pop"=>240, "name_ru"=>"Патон", "name_en"=>"Paton", "lang"=>"en", "lat"=>42.163909, "lon"=>-94.254730),
"/US/US-IA/patterson-7101" => array("type"=>"city", "pop"=>132, "name_ru"=>"Паттерсон", "name_en"=>"Patterson", "lang"=>"en", "lat"=>41.348960, "lon"=>-93.879192),
"/US/US-IA/paullina-7102" => array("type"=>"city", "pop"=>989, "name_ru"=>"Поллина", "name_en"=>"Paullina", "lang"=>"en", "lat"=>42.979785, "lon"=>-95.686836),
"/US/US-IA/pella-7103" => array("type"=>"city", "pop"=>10208, "name_ru"=>"Пелла", "name_en"=>"Pella", "lang"=>"en", "lat"=>41.408653, "lon"=>-92.918999),
"/US/US-IA/peosta-7104" => array("type"=>"city", "pop"=>1187, "name_ru"=>"Пеоста", "name_en"=>"Peosta", "lang"=>"en", "lat"=>42.448848, "lon"=>-90.850738),
"/US/US-IA/perry-7105" => array("type"=>"city", "pop"=>9569, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>41.842123, "lon"=>-94.099789),
"/US/US-IA/persia-7106" => array("type"=>"city", "pop"=>329, "name_ru"=>"Персия", "name_en"=>"Persia", "lang"=>"en", "lat"=>41.579003, "lon"=>-95.570135),
"/US/US-IA/peterson-7107" => array("type"=>"city", "pop"=>342, "name_ru"=>"Петерсон", "name_en"=>"Peterson", "lang"=>"en", "lat"=>42.918513, "lon"=>-95.341224),
"/US/US-IA/pierson-7108" => array("type"=>"city", "pop"=>348, "name_ru"=>"Пирсон", "name_en"=>"Pierson", "lang"=>"en", "lat"=>42.543690, "lon"=>-95.866860),
"/US/US-IA/pilot-mound-7109" => array("type"=>"city", "pop"=>200, "name_ru"=>"Пайлот Маунд", "name_en"=>"Pilot Mound", "lang"=>"en", "lat"=>42.159078, "lon"=>-94.018674),
"/US/US-IA/pioneer-7110" => array("type"=>"city", "pop"=>19, "name_ru"=>"Пайониер", "name_en"=>"Pioneer", "lang"=>"en", "lat"=>42.653476, "lon"=>-94.391135),
"/US/US-IA/pisgah-7111" => array("type"=>"city", "pop"=>298, "name_ru"=>"Фасги", "name_en"=>"Pisgah", "lang"=>"en", "lat"=>41.830821, "lon"=>-95.926760),
"/US/US-IA/plainfield-7112" => array("type"=>"city", "pop"=>427, "name_ru"=>"Плейнфилд", "name_en"=>"Plainfield", "lang"=>"en", "lat"=>42.843933, "lon"=>-92.535458),
"/US/US-IA/plano-7113" => array("type"=>"city", "pop"=>56, "name_ru"=>"Плано", "name_en"=>"Plano", "lang"=>"en", "lat"=>40.755322, "lon"=>-93.046078),
"/US/US-IA/pleasant-hill-7114" => array("type"=>"city", "pop"=>8414, "name_ru"=>"Плезант Хилл", "name_en"=>"Pleasant Hill", "lang"=>"en", "lat"=>41.585650, "lon"=>-93.514449),
"/US/US-IA/pleasant-plain-7116" => array("type"=>"city", "pop"=>124, "name_ru"=>"Плезант Плейн", "name_en"=>"Pleasant Plain", "lang"=>"en", "lat"=>41.147372, "lon"=>-91.859697),
"/US/US-IA/pleasanton-7115" => array("type"=>"city", "pop"=>28, "name_ru"=>"Плезантон", "name_en"=>"Pleasanton", "lang"=>"en", "lat"=>40.582365, "lon"=>-93.742892),
"/US/US-IA/pleasantville-7117" => array("type"=>"city", "pop"=>1580, "name_ru"=>"Плезантвилл", "name_en"=>"Pleasantville", "lang"=>"en", "lat"=>41.389015, "lon"=>-93.267359),
"/US/US-IA/plover-7118" => array("type"=>"city", "pop"=>84, "name_ru"=>"Пловер", "name_en"=>"Plover", "lang"=>"en", "lat"=>42.877104, "lon"=>-94.622415),
"/US/US-IA/plymouth-7119" => array("type"=>"city", "pop"=>413, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>43.246901, "lon"=>-93.122398),
"/US/US-IA/pocahontas-7120" => array("type"=>"city", "pop"=>1799, "name_ru"=>"Покахонтас", "name_en"=>"Pocahontas", "lang"=>"en", "lat"=>42.736215, "lon"=>-94.666817),
"/US/US-IA/polk-city-7121" => array("type"=>"city", "pop"=>3201, "name_ru"=>"Полк Сити", "name_en"=>"Polk City", "lang"=>"en", "lat"=>41.773096, "lon"=>-93.716518),
"/US/US-IA/pomeroy-7122" => array("type"=>"city", "pop"=>613, "name_ru"=>"Померой", "name_en"=>"Pomeroy", "lang"=>"en", "lat"=>42.553500, "lon"=>-94.678531),
"/US/US-IA/popejoy-7123" => array("type"=>"city", "pop"=>78, "name_ru"=>"Поупджой", "name_en"=>"Popejoy", "lang"=>"en", "lat"=>42.594106, "lon"=>-93.426818),
"/US/US-IA/portsmouth-7124" => array("type"=>"city", "pop"=>200, "name_ru"=>"Портсмаут", "name_en"=>"Portsmouth", "lang"=>"en", "lat"=>41.650741, "lon"=>-95.519389),
"/US/US-IA/postville-7125" => array("type"=>"city", "pop"=>2306, "name_ru"=>"Поуствилл", "name_en"=>"Postville", "lang"=>"en", "lat"=>43.084083, "lon"=>-91.569347),
"/US/US-IA/prairie-city-7127" => array("type"=>"city", "pop"=>1436, "name_ru"=>"Прейри Сити", "name_en"=>"Prairie City", "lang"=>"en", "lat"=>41.598072, "lon"=>-93.235365),
"/US/US-IA/prairieburg-7126" => array("type"=>"city", "pop"=>170, "name_ru"=>"Прейрибург", "name_en"=>"Prairieburg", "lang"=>"en", "lat"=>42.237834, "lon"=>-91.425980),
"/US/US-IA/prescott-7128" => array("type"=>"city", "pop"=>245, "name_ru"=>"Прескотт", "name_en"=>"Prescott", "lang"=>"en", "lat"=>41.023433, "lon"=>-94.612397),
"/US/US-IA/preston-7129" => array("type"=>"city", "pop"=>920, "name_ru"=>"Престон", "name_en"=>"Preston", "lang"=>"en", "lat"=>42.048471, "lon"=>-90.399310),
"/US/US-IA/primghar-7130" => array("type"=>"city", "pop"=>809, "name_ru"=>"Примхар", "name_en"=>"Primghar", "lang"=>"en", "lat"=>43.087763, "lon"=>-95.626200),
"/US/US-IA/princeton-7131" => array("type"=>"city", "pop"=>953, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>41.671100, "lon"=>-90.359498),
"/US/US-IA/promise-city-7132" => array("type"=>"city", "pop"=>100, "name_ru"=>"Промис Сити", "name_en"=>"Promise City", "lang"=>"en", "lat"=>40.746944, "lon"=>-93.150911),
"/US/US-IA/protivin-7133" => array("type"=>"city", "pop"=>304, "name_ru"=>"Противин", "name_en"=>"Protivin", "lang"=>"en", "lat"=>43.217509, "lon"=>-92.088119),
"/US/US-IA/pulaski-7134" => array("type"=>"city", "pop"=>268, "name_ru"=>"Пуласки", "name_en"=>"Pulaski", "lang"=>"en", "lat"=>40.697498, "lon"=>-92.274064),
"/US/US-IA/quasqueton-7135" => array("type"=>"city", "pop"=>546, "name_ru"=>"Кваскетон", "name_en"=>"Quasqueton", "lang"=>"en", "lat"=>42.395451, "lon"=>-91.756433),
"/US/US-IA/quimby-7136" => array("type"=>"city", "pop"=>329, "name_ru"=>"Квимби", "name_en"=>"Quimby", "lang"=>"en", "lat"=>42.629177, "lon"=>-95.642679),
"/US/US-IA/radcliffe-7137" => array("type"=>"city", "pop"=>563, "name_ru"=>"Рэдклифф", "name_en"=>"Radcliffe", "lang"=>"en", "lat"=>42.318133, "lon"=>-93.434647),
"/US/US-IA/rake-7138" => array("type"=>"city", "pop"=>207, "name_ru"=>"Рейк", "name_en"=>"Rake", "lang"=>"en", "lat"=>43.481335, "lon"=>-93.921012),
"/US/US-IA/ralston-7139" => array("type"=>"city", "pop"=>79, "name_ru"=>"Ралстон", "name_en"=>"Ralston", "lang"=>"en", "lat"=>42.043504, "lon"=>-94.623905),
"/US/US-IA/randalia-7140" => array("type"=>"city", "pop"=>69, "name_ru"=>"Рандалия", "name_en"=>"Randalia", "lang"=>"en", "lat"=>42.863261, "lon"=>-91.886225),
"/US/US-IA/randall-7141" => array("type"=>"city", "pop"=>142, "name_ru"=>"Рэндолл", "name_en"=>"Randall", "lang"=>"en", "lat"=>42.237223, "lon"=>-93.602994),
"/US/US-IA/randolph-7142" => array("type"=>"city", "pop"=>168, "name_ru"=>"Рэндолф", "name_en"=>"Randolph", "lang"=>"en", "lat"=>40.872890, "lon"=>-95.564243),
"/US/US-IA/rathbun-7143" => array("type"=>"city", "pop"=>84, "name_ru"=>"Ратбун", "name_en"=>"Rathbun", "lang"=>"en", "lat"=>40.802084, "lon"=>-92.887894),
"/US/US-IA/raymond-7144" => array("type"=>"city", "pop"=>686, "name_ru"=>"Реймонд", "name_en"=>"Raymond", "lang"=>"en", "lat"=>42.466850, "lon"=>-92.228249),
"/US/US-IA/readlyn-7145" => array("type"=>"city", "pop"=>762, "name_ru"=>"Ридлин", "name_en"=>"Readlyn", "lang"=>"en", "lat"=>42.703363, "lon"=>-92.224752),
"/US/US-IA/reasnor-7146" => array("type"=>"city", "pop"=>188, "name_ru"=>"Риснор", "name_en"=>"Reasnor", "lang"=>"en", "lat"=>41.578503, "lon"=>-93.022616),
"/US/US-IA/red-oak-7149" => array("type"=>"city", "pop"=>5683, "name_ru"=>"Ред Оук", "name_en"=>"Red Oak", "lang"=>"en", "lat"=>41.012479, "lon"=>-95.224560),
"/US/US-IA/redding-7147" => array("type"=>"city", "pop"=>69, "name_ru"=>"Реддинг", "name_en"=>"Redding", "lang"=>"en", "lat"=>40.606046, "lon"=>-94.386794),
"/US/US-IA/redfield-7148" => array("type"=>"city", "pop"=>1041, "name_ru"=>"Редфилд", "name_en"=>"Redfield", "lang"=>"en", "lat"=>41.590105, "lon"=>-94.197623),
"/US/US-IA/reinbeck-7150" => array("type"=>"city", "pop"=>1652, "name_ru"=>"Райнбек", "name_en"=>"Reinbeck", "lang"=>"en", "lat"=>42.322233, "lon"=>-92.598502),
"/US/US-IA/rembrandt-7151" => array("type"=>"city", "pop"=>223, "name_ru"=>"Ребрандт", "name_en"=>"Rembrandt", "lang"=>"en", "lat"=>42.826440, "lon"=>-95.165655),
"/US/US-IA/remsen-7152" => array("type"=>"city", "pop"=>1654, "name_ru"=>"Ремсен", "name_en"=>"Remsen", "lang"=>"en", "lat"=>42.813116, "lon"=>-95.972750),
"/US/US-IA/renwick-7153" => array("type"=>"city", "pop"=>262, "name_ru"=>"Ренвик", "name_en"=>"Renwick", "lang"=>"en", "lat"=>42.827398, "lon"=>-93.980689),
"/US/US-IA/rhodes-7154" => array("type"=>"city", "pop"=>299, "name_ru"=>"Родс", "name_en"=>"Rhodes", "lang"=>"en", "lat"=>41.927428, "lon"=>-93.184007),
"/US/US-IA/riceville-7155" => array("type"=>"city", "pop"=>806, "name_ru"=>"Райсвилл", "name_en"=>"Riceville", "lang"=>"en", "lat"=>43.362515, "lon"=>-92.554913),
"/US/US-IA/richland-7156" => array("type"=>"city", "pop"=>563, "name_ru"=>"Ричленд", "name_en"=>"Richland", "lang"=>"en", "lat"=>41.185868, "lon"=>-91.994969),
"/US/US-IA/rickardsville-7157" => array("type"=>"city", "pop"=>181, "name_ru"=>"Рикардсвилл", "name_en"=>"Rickardsville", "lang"=>"en", "lat"=>42.583646, "lon"=>-90.875283),
"/US/US-IA/ricketts-7158" => array("type"=>"city", "pop"=>134, "name_ru"=>"Риккетс", "name_en"=>"Ricketts", "lang"=>"en", "lat"=>42.127464, "lon"=>-95.574231),
"/US/US-IA/ridgeway-7159" => array("type"=>"city", "pop"=>284, "name_ru"=>"Риджуэй", "name_en"=>"Ridgeway", "lang"=>"en", "lat"=>43.297170, "lon"=>-91.991094),
"/US/US-IA/rinard-7160" => array("type"=>"city", "pop"=>62, "name_ru"=>"Ринард", "name_en"=>"Rinard", "lang"=>"en", "lat"=>42.339480, "lon"=>-94.486467),
"/US/US-IA/ringsted-7161" => array("type"=>"city", "pop"=>418, "name_ru"=>"Рингстед", "name_en"=>"Ringsted", "lang"=>"en", "lat"=>43.296398, "lon"=>-94.506557),
"/US/US-IA/rippey-7162" => array("type"=>"city", "pop"=>276, "name_ru"=>"Риппи", "name_en"=>"Rippey", "lang"=>"en", "lat"=>41.934592, "lon"=>-94.199953),
"/US/US-IA/riverdale-7163" => array("type"=>"city", "pop"=>595, "name_ru"=>"Ривердейл", "name_en"=>"Riverdale", "lang"=>"en", "lat"=>41.535433, "lon"=>-90.467084),
"/US/US-IA/riverside-7164" => array("type"=>"city", "pop"=>982, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>41.482325, "lon"=>-91.573111),
"/US/US-IA/riverton-7165" => array("type"=>"city", "pop"=>275, "name_ru"=>"Ривертон", "name_en"=>"Riverton", "lang"=>"en", "lat"=>40.687368, "lon"=>-95.568770),
"/US/US-IA/robins-7166" => array("type"=>"city", "pop"=>2869, "name_ru"=>"Робинс", "name_en"=>"Robins", "lang"=>"en", "lat"=>42.078350, "lon"=>-91.669236),
"/US/US-IA/rock-falls-7167" => array("type"=>"city", "pop"=>154, "name_ru"=>"Рок Фоллс", "name_en"=>"Rock Falls", "lang"=>"en", "lat"=>43.207217, "lon"=>-93.087027),
"/US/US-IA/rock-rapids-7169" => array("type"=>"city", "pop"=>2437, "name_ru"=>"Рок Рэпидс", "name_en"=>"Rock Rapids", "lang"=>"en", "lat"=>43.425815, "lon"=>-96.166077),
"/US/US-IA/rock-valley-7170" => array("type"=>"city", "pop"=>2945, "name_ru"=>"Рок Валли", "name_en"=>"Rock Valley", "lang"=>"en", "lat"=>43.204626, "lon"=>-96.297157),
"/US/US-IA/rockford-7168" => array("type"=>"city", "pop"=>888, "name_ru"=>"Рокфорд", "name_en"=>"Rockford", "lang"=>"en", "lat"=>43.052841, "lon"=>-92.947496),
"/US/US-IA/rockwell-7171" => array("type"=>"city", "pop"=>992, "name_ru"=>"Рокуэлл", "name_en"=>"Rockwell", "lang"=>"en", "lat"=>42.979934, "lon"=>-93.186862),
"/US/US-IA/rockwell-city-7172" => array("type"=>"city", "pop"=>1988, "name_ru"=>"Рокуэлл Сити", "name_en"=>"Rockwell City", "lang"=>"en", "lat"=>42.398418, "lon"=>-94.631018),
"/US/US-IA/rodman-7173" => array("type"=>"city", "pop"=>51, "name_ru"=>"Родман", "name_en"=>"Rodman", "lang"=>"en", "lat"=>43.025979, "lon"=>-94.526295),
"/US/US-IA/rodney-7174" => array("type"=>"city", "pop"=>58, "name_ru"=>"Родни", "name_en"=>"Rodney", "lang"=>"en", "lat"=>42.204844, "lon"=>-95.951158),
"/US/US-IA/roland-7175" => array("type"=>"city", "pop"=>1317, "name_ru"=>"Роланд", "name_en"=>"Roland", "lang"=>"en", "lat"=>42.167027, "lon"=>-93.501239),
"/US/US-IA/rolfe-7176" => array("type"=>"city", "pop"=>581, "name_ru"=>"Ролф", "name_en"=>"Rolfe", "lang"=>"en", "lat"=>42.812820, "lon"=>-94.530314),
"/US/US-IA/rome-7177" => array("type"=>"city", "pop"=>113, "name_ru"=>"Роум", "name_en"=>"Rome", "lang"=>"en", "lat"=>40.983308, "lon"=>-91.680741),
"/US/US-IA/rose-hill-7178" => array("type"=>"city", "pop"=>205, "name_ru"=>"Роуз Хилл", "name_en"=>"Rose Hill", "lang"=>"en", "lat"=>41.320449, "lon"=>-92.462773),
"/US/US-IA/rossie-7179" => array("type"=>"city", "pop"=>56, "name_ru"=>"Росси", "name_en"=>"Rossie", "lang"=>"en", "lat"=>43.013234, "lon"=>-95.188832),
"/US/US-IA/rowan-7180" => array("type"=>"city", "pop"=>194, "name_ru"=>"Рован", "name_en"=>"Rowan", "lang"=>"en", "lat"=>42.740150, "lon"=>-93.549622),
"/US/US-IA/rowley-7181" => array("type"=>"city", "pop"=>286, "name_ru"=>"Роули", "name_en"=>"Rowley", "lang"=>"en", "lat"=>42.368392, "lon"=>-91.844842),
"/US/US-IA/royal-7182" => array("type"=>"city", "pop"=>424, "name_ru"=>"Ройал", "name_en"=>"Royal", "lang"=>"en", "lat"=>43.063954, "lon"=>-95.283150),
"/US/US-IA/rudd-7183" => array("type"=>"city", "pop"=>399, "name_ru"=>"Радд", "name_en"=>"Rudd", "lang"=>"en", "lat"=>43.126786, "lon"=>-92.904845),
"/US/US-IA/runnells-7184" => array("type"=>"city", "pop"=>392, "name_ru"=>"Раннелс", "name_en"=>"Runnells", "lang"=>"en", "lat"=>41.512323, "lon"=>-93.358055),
"/US/US-IA/russell-7185" => array("type"=>"city", "pop"=>564, "name_ru"=>"Расселл", "name_en"=>"Russell", "lang"=>"en", "lat"=>40.980396, "lon"=>-93.200776),
"/US/US-IA/ruthven-7186" => array("type"=>"city", "pop"=>687, "name_ru"=>"Рутвен", "name_en"=>"Ruthven", "lang"=>"en", "lat"=>43.129627, "lon"=>-94.899334),
"/US/US-IA/rutland-7187" => array("type"=>"city", "pop"=>123, "name_ru"=>"Рутленд", "name_en"=>"Rutland", "lang"=>"en", "lat"=>42.761205, "lon"=>-94.292564),
"/US/US-IA/ryan-7188" => array("type"=>"city", "pop"=>369, "name_ru"=>"Райан", "name_en"=>"Ryan", "lang"=>"en", "lat"=>42.352591, "lon"=>-91.482831),
"/US/US-IA/sabula-7189" => array("type"=>"city", "pop"=>649, "name_ru"=>"Сабула", "name_en"=>"Sabula", "lang"=>"en", "lat"=>42.067843, "lon"=>-90.174143),
"/US/US-IA/sac-city-7190" => array("type"=>"city", "pop"=>2140, "name_ru"=>"Сэк Сити", "name_en"=>"Sac City", "lang"=>"en", "lat"=>42.422133, "lon"=>-94.997313),
"/US/US-IA/sageville-7191" => array("type"=>"city", "pop"=>191, "name_ru"=>"Сейджвилл", "name_en"=>"Sageville", "lang"=>"en", "lat"=>42.548767, "lon"=>-90.703663),
"/US/US-IA/salem-7200" => array("type"=>"city", "pop"=>449, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>40.852014, "lon"=>-91.620676),
"/US/US-IA/salix-7201" => array("type"=>"city", "pop"=>383, "name_ru"=>"Саликс", "name_en"=>"Salix", "lang"=>"en", "lat"=>42.309517, "lon"=>-96.286630),
"/US/US-IA/sanborn-7202" => array("type"=>"city", "pop"=>1287, "name_ru"=>"Сэнборн", "name_en"=>"Sanborn", "lang"=>"en", "lat"=>43.181102, "lon"=>-95.656333),
"/US/US-IA/sandyville-7203" => array("type"=>"city", "pop"=>60, "name_ru"=>"Сэндивилл", "name_en"=>"Sandyville", "lang"=>"en", "lat"=>41.370231, "lon"=>-93.385465),
"/US/US-IA/saylorville-7204" => array("type"=>"cdp", "pop"=>2709, "name_ru"=>"Сейлорвилл", "name_en"=>"Saylorville", "lang"=>"en", "lat"=>41.684533, "lon"=>-93.625766),
"/US/US-IA/scarville-7205" => array("type"=>"city", "pop"=>89, "name_ru"=>"Скарвилл", "name_en"=>"Scarville", "lang"=>"en", "lat"=>43.470565, "lon"=>-93.616264),
"/US/US-IA/schaller-7206" => array("type"=>"city", "pop"=>701, "name_ru"=>"Шоллер", "name_en"=>"Schaller", "lang"=>"en", "lat"=>42.495743, "lon"=>-95.295389),
"/US/US-IA/schleswig-7207" => array("type"=>"city", "pop"=>816, "name_ru"=>"Шлесвиг", "name_en"=>"Schleswig", "lang"=>"en", "lat"=>42.161851, "lon"=>-95.434455),
"/US/US-IA/scranton-7208" => array("type"=>"city", "pop"=>522, "name_ru"=>"Скрантон", "name_en"=>"Scranton", "lang"=>"en", "lat"=>42.020419, "lon"=>-94.548185),
"/US/US-IA/searsboro-7209" => array("type"=>"city", "pop"=>151, "name_ru"=>"Сиэрсборо", "name_en"=>"Searsboro", "lang"=>"en", "lat"=>41.579885, "lon"=>-92.704772),
"/US/US-IA/sergeant-bluff-7210" => array("type"=>"city", "pop"=>4011, "name_ru"=>"Серджент Блафф", "name_en"=>"Sergeant Bluff", "lang"=>"en", "lat"=>42.400651, "lon"=>-96.355592),
"/US/US-IA/seymour-7211" => array("type"=>"city", "pop"=>738, "name_ru"=>"Сеймур", "name_en"=>"Seymour", "lang"=>"en", "lat"=>40.682845, "lon"=>-93.121832),
"/US/US-IA/shambaugh-7212" => array("type"=>"city", "pop"=>171, "name_ru"=>"Шамбо", "name_en"=>"Shambaugh", "lang"=>"en", "lat"=>40.657086, "lon"=>-95.036092),
"/US/US-IA/shannon-city-7213" => array("type"=>"city", "pop"=>73, "name_ru"=>"Шэннон Сити", "name_en"=>"Shannon City", "lang"=>"en", "lat"=>40.901005, "lon"=>-94.263638),
"/US/US-IA/sharpsburg-7214" => array("type"=>"city", "pop"=>90, "name_ru"=>"Шарпсбург", "name_en"=>"Sharpsburg", "lang"=>"en", "lat"=>40.802792, "lon"=>-94.641020),
"/US/US-IA/sheffield-7215" => array("type"=>"city", "pop"=>990, "name_ru"=>"Шеффилд", "name_en"=>"Sheffield", "lang"=>"en", "lat"=>42.894399, "lon"=>-93.210522),
"/US/US-IA/shelby-7216" => array("type"=>"city", "pop"=>644, "name_ru"=>"Шелби", "name_en"=>"Shelby", "lang"=>"en", "lat"=>41.508523, "lon"=>-95.451641),
"/US/US-IA/sheldahl-7217" => array("type"=>"city", "pop"=>319, "name_ru"=>"Шелдал", "name_en"=>"Sheldahl", "lang"=>"en", "lat"=>41.864128, "lon"=>-93.696234),
"/US/US-IA/sheldon-7218" => array("type"=>"city", "pop"=>4757, "name_ru"=>"Шелдон", "name_en"=>"Sheldon", "lang"=>"en", "lat"=>43.179305, "lon"=>-95.844911),
"/US/US-IA/shell-rock-7219" => array("type"=>"city", "pop"=>1235, "name_ru"=>"Шелл Рок", "name_en"=>"Shell Rock", "lang"=>"en", "lat"=>42.712648, "lon"=>-92.581550),
"/US/US-IA/shellsburg-7220" => array("type"=>"city", "pop"=>1004, "name_ru"=>"Шеллсбург", "name_en"=>"Shellsburg", "lang"=>"en", "lat"=>42.093167, "lon"=>-91.869416),
"/US/US-IA/shenandoah-7221" => array("type"=>"city", "pop"=>4944, "name_ru"=>"Шенандоа", "name_en"=>"Shenandoah", "lang"=>"en", "lat"=>40.759373, "lon"=>-95.369238),
"/US/US-IA/sherrill-7222" => array("type"=>"city", "pop"=>186, "name_ru"=>"Шеррилл", "name_en"=>"Sherrill", "lang"=>"en", "lat"=>42.602859, "lon"=>-90.782011),
"/US/US-IA/shueyville-7223" => array("type"=>"city", "pop"=>525, "name_ru"=>"Шуивилл", "name_en"=>"Shueyville", "lang"=>"en", "lat"=>41.848148, "lon"=>-91.648504),
"/US/US-IA/sibley-7224" => array("type"=>"city", "pop"=>2612, "name_ru"=>"Сибли", "name_en"=>"Sibley", "lang"=>"en", "lat"=>43.401661, "lon"=>-95.744927),
"/US/US-IA/sidney-7225" => array("type"=>"city", "pop"=>1175, "name_ru"=>"Сидни", "name_en"=>"Sidney", "lang"=>"en", "lat"=>40.745962, "lon"=>-95.644372),
"/US/US-IA/sigourney-7226" => array("type"=>"city", "pop"=>2090, "name_ru"=>"Сигурни", "name_en"=>"Sigourney", "lang"=>"en", "lat"=>41.334039, "lon"=>-92.206539),
"/US/US-IA/silver-city-7227" => array("type"=>"city", "pop"=>268, "name_ru"=>"Сильвер Сити", "name_en"=>"Silver City", "lang"=>"en", "lat"=>41.110987, "lon"=>-95.637783),
"/US/US-IA/sioux-center-7228" => array("type"=>"city", "pop"=>6581, "name_ru"=>"Сиу Сентер", "name_en"=>"Sioux Center", "lang"=>"en", "lat"=>43.076561, "lon"=>-96.171480),
"/US/US-IA/sioux-city-7229" => array("type"=>"city", "pop"=>82807, "name_ru"=>"Сиу Сити", "name_en"=>"Sioux City", "lang"=>"en", "lat"=>42.500151, "lon"=>-96.393950),
"/US/US-IA/sioux-rapids-7230" => array("type"=>"city", "pop"=>706, "name_ru"=>"Сиу Рэпидс", "name_en"=>"Sioux Rapids", "lang"=>"en", "lat"=>42.891884, "lon"=>-95.148061),
"/US/US-IA/slater-7231" => array("type"=>"city", "pop"=>1388, "name_ru"=>"Слейтер", "name_en"=>"Slater", "lang"=>"en", "lat"=>41.880279, "lon"=>-93.684062),
"/US/US-IA/sloan-7232" => array("type"=>"city", "pop"=>1019, "name_ru"=>"Слоан", "name_en"=>"Sloan", "lang"=>"en", "lat"=>42.234310, "lon"=>-96.224103),
"/US/US-IA/smithland-7233" => array("type"=>"city", "pop"=>213, "name_ru"=>"Смитленд", "name_en"=>"Smithland", "lang"=>"en", "lat"=>42.229088, "lon"=>-95.931588),
"/US/US-IA/soldier-7234" => array("type"=>"city", "pop"=>185, "name_ru"=>"Солджер", "name_en"=>"Soldier", "lang"=>"en", "lat"=>41.984610, "lon"=>-95.779631),
"/US/US-IA/solon-7235" => array("type"=>"city", "pop"=>1668, "name_ru"=>"Солон", "name_en"=>"Solon", "lang"=>"en", "lat"=>41.804968, "lon"=>-91.494550),
"/US/US-IA/somers-7236" => array("type"=>"city", "pop"=>136, "name_ru"=>"Сомерс", "name_en"=>"Somers", "lang"=>"en", "lat"=>42.377925, "lon"=>-94.430767),
"/US/US-IA/south-english-7237" => array("type"=>"city", "pop"=>205, "name_ru"=>"Саут Инглиш", "name_en"=>"South English", "lang"=>"en", "lat"=>41.452305, "lon"=>-92.090821),
"/US/US-IA/spencer-7238" => array("type"=>"city", "pop"=>10986, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>43.145866, "lon"=>-95.151124),
"/US/US-IA/spillville-7239" => array("type"=>"city", "pop"=>371, "name_ru"=>"Спиллвилл", "name_en"=>"Spillville", "lang"=>"en", "lat"=>43.204006, "lon"=>-91.951879),
"/US/US-IA/spirit-lake-7240" => array("type"=>"city", "pop"=>4692, "name_ru"=>"Спирит Лейк", "name_en"=>"Spirit Lake", "lang"=>"en", "lat"=>43.422766, "lon"=>-95.111409),
"/US/US-IA/spragueville-7241" => array("type"=>"city", "pop"=>84, "name_ru"=>"Спрагевилл", "name_en"=>"Spragueville", "lang"=>"en", "lat"=>42.070530, "lon"=>-90.430981),
"/US/US-IA/spring-hill-7243" => array("type"=>"city", "pop"=>91, "name_ru"=>"Спринг Хилл", "name_en"=>"Spring Hill", "lang"=>"en", "lat"=>41.412047, "lon"=>-93.648957),
"/US/US-IA/springbrook-7242" => array("type"=>"city", "pop"=>184, "name_ru"=>"Спрингбрук", "name_en"=>"Springbrook", "lang"=>"en", "lat"=>42.167481, "lon"=>-90.481896),
"/US/US-IA/springville-7244" => array("type"=>"city", "pop"=>1047, "name_ru"=>"Спрингвилл", "name_en"=>"Springville", "lang"=>"en", "lat"=>42.055781, "lon"=>-91.443975),
"/US/US-IA/st.-ansgar-7192" => array("type"=>"city", "pop"=>1056, "name_ru"=>"Сент Ансгар", "name_en"=>"St. Ansgar", "lang"=>"en", "lat"=>43.378061, "lon"=>-92.919611),
"/US/US-IA/st.-anthony-7193" => array("type"=>"city", "pop"=>187, "name_ru"=>"Сент Энтони", "name_en"=>"St. Anthony", "lang"=>"en", "lat"=>42.124383, "lon"=>-93.197803),
"/US/US-IA/st.-charles-7194" => array("type"=>"city", "pop"=>672, "name_ru"=>"Сент-Чарльз", "name_en"=>"St. Charles", "lang"=>"en", "lat"=>41.287888, "lon"=>-93.806296),
"/US/US-IA/st.-donatus-7195" => array("type"=>"city", "pop"=>155, "name_ru"=>"Сент Донатус", "name_en"=>"St. Donatus", "lang"=>"en", "lat"=>42.362057, "lon"=>-90.543140),
"/US/US-IA/st.-lucas-7196" => array("type"=>"city", "pop"=>163, "name_ru"=>"Сент Лукас", "name_en"=>"St. Lucas", "lang"=>"en", "lat"=>43.066608, "lon"=>-91.933894),
"/US/US-IA/st.-marys-7197" => array("type"=>"city", "pop"=>127, "name_ru"=>"Сент Мэрис", "name_en"=>"St. Marys", "lang"=>"en", "lat"=>41.308613, "lon"=>-93.733185),
"/US/US-IA/st.-olaf-7198" => array("type"=>"city", "pop"=>120, "name_ru"=>"Сент Олаф", "name_en"=>"St. Olaf", "lang"=>"en", "lat"=>42.927740, "lon"=>-91.387099),
"/US/US-IA/st.-paul-7199" => array("type"=>"city", "pop"=>116, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>40.766923, "lon"=>-91.516799),
"/US/US-IA/stacyville-7245" => array("type"=>"city", "pop"=>449, "name_ru"=>"Стейсивилл", "name_en"=>"Stacyville", "lang"=>"en", "lat"=>43.438439, "lon"=>-92.782832),
"/US/US-IA/stanhope-7246" => array("type"=>"city", "pop"=>437, "name_ru"=>"Стэнохоуп", "name_en"=>"Stanhope", "lang"=>"en", "lat"=>42.288589, "lon"=>-93.795036),
"/US/US-IA/stanley-7247" => array("type"=>"city", "pop"=>126, "name_ru"=>"Стэнли", "name_en"=>"Stanley", "lang"=>"en", "lat"=>42.641956, "lon"=>-91.811715),
"/US/US-IA/stanton-7248" => array("type"=>"city", "pop"=>683, "name_ru"=>"Стэнтон", "name_en"=>"Stanton", "lang"=>"en", "lat"=>40.981081, "lon"=>-95.103215),
"/US/US-IA/stanwood-7249" => array("type"=>"city", "pop"=>643, "name_ru"=>"Стэнвуд", "name_en"=>"Stanwood", "lang"=>"en", "lat"=>41.893632, "lon"=>-91.149958),
"/US/US-IA/state-center-7250" => array("type"=>"city", "pop"=>1341, "name_ru"=>"Стейт Сентер", "name_en"=>"State Center", "lang"=>"en", "lat"=>42.015297, "lon"=>-93.165390),
"/US/US-IA/steamboat-rock-7251" => array("type"=>"city", "pop"=>315, "name_ru"=>"Стимбоат Рок", "name_en"=>"Steamboat Rock", "lang"=>"en", "lat"=>42.407985, "lon"=>-93.066072),
"/US/US-IA/stockport-7252" => array("type"=>"city", "pop"=>282, "name_ru"=>"Стокпорт", "name_en"=>"Stockport", "lang"=>"en", "lat"=>40.856967, "lon"=>-91.833173),
"/US/US-IA/stockton-7253" => array("type"=>"city", "pop"=>181, "name_ru"=>"Стоктон", "name_en"=>"Stockton", "lang"=>"en", "lat"=>41.590423, "lon"=>-90.855234),
"/US/US-IA/storm-lake-7254" => array("type"=>"city", "pop"=>9591, "name_ru"=>"Сторм Лейк", "name_en"=>"Storm Lake", "lang"=>"en", "lat"=>42.645021, "lon"=>-95.199855),
"/US/US-IA/story-city-7255" => array("type"=>"city", "pop"=>3389, "name_ru"=>"Стори Сити", "name_en"=>"Story City", "lang"=>"en", "lat"=>42.186449, "lon"=>-93.591611),
"/US/US-IA/stout-7256" => array("type"=>"city", "pop"=>214, "name_ru"=>"Стаут", "name_en"=>"Stout", "lang"=>"en", "lat"=>42.527150, "lon"=>-92.711188),
"/US/US-IA/stratford-7257" => array("type"=>"city", "pop"=>686, "name_ru"=>"Стрэтфорд", "name_en"=>"Stratford", "lang"=>"en", "lat"=>42.270954, "lon"=>-93.927628),
"/US/US-IA/strawberry-point-7258" => array("type"=>"city", "pop"=>1247, "name_ru"=>"Строберри Поинт", "name_en"=>"Strawberry Point", "lang"=>"en", "lat"=>42.682963, "lon"=>-91.535525),
"/US/US-IA/struble-7259" => array("type"=>"city", "pop"=>80, "name_ru"=>"Струбл", "name_en"=>"Struble", "lang"=>"en", "lat"=>42.894844, "lon"=>-96.194068),
"/US/US-IA/stuart-7260" => array("type"=>"city", "pop"=>1703, "name_ru"=>"Стюарт", "name_en"=>"Stuart", "lang"=>"en", "lat"=>41.502357, "lon"=>-94.318254),
"/US/US-IA/sully-7261" => array("type"=>"city", "pop"=>856, "name_ru"=>"Салли", "name_en"=>"Sully", "lang"=>"en", "lat"=>41.578641, "lon"=>-92.846464),
"/US/US-IA/sumner-7262" => array("type"=>"city", "pop"=>1984, "name_ru"=>"Саммер", "name_en"=>"Sumner", "lang"=>"en", "lat"=>42.849915, "lon"=>-92.095976),
"/US/US-IA/superior-7263" => array("type"=>"city", "pop"=>126, "name_ru"=>"Супериор", "name_en"=>"Superior", "lang"=>"en", "lat"=>43.432670, "lon"=>-94.947077),
"/US/US-IA/sutherland-7264" => array("type"=>"city", "pop"=>632, "name_ru"=>"Сатерленд", "name_en"=>"Sutherland", "lang"=>"en", "lat"=>42.972377, "lon"=>-95.495251),
"/US/US-IA/swaledale-7265" => array("type"=>"city", "pop"=>167, "name_ru"=>"Суолдейл", "name_en"=>"Swaledale", "lang"=>"en", "lat"=>42.976892, "lon"=>-93.315329),
"/US/US-IA/swan-7266" => array("type"=>"city", "pop"=>84, "name_ru"=>"Свон", "name_en"=>"Swan", "lang"=>"en", "lat"=>41.465613, "lon"=>-93.309416),
"/US/US-IA/swea-city-7267" => array("type"=>"city", "pop"=>569, "name_ru"=>"Сви Сити", "name_en"=>"Swea City", "lang"=>"en", "lat"=>43.382838, "lon"=>-94.310178),
"/US/US-IA/swisher-7268" => array("type"=>"city", "pop"=>851, "name_ru"=>"Свишер", "name_en"=>"Swisher", "lang"=>"en", "lat"=>41.843369, "lon"=>-91.695787),
"/US/US-IA/tabor-7269" => array("type"=>"city", "pop"=>928, "name_ru"=>"Табор", "name_en"=>"Tabor", "lang"=>"en", "lat"=>40.893817, "lon"=>-95.672090),
"/US/US-IA/tama-7270" => array("type"=>"city", "pop"=>2557, "name_ru"=>"Тама", "name_en"=>"Tama", "lang"=>"en", "lat"=>41.963608, "lon"=>-92.574795),
"/US/US-IA/templeton-7271" => array("type"=>"city", "pop"=>334, "name_ru"=>"Темплтон", "name_en"=>"Templeton", "lang"=>"en", "lat"=>41.918088, "lon"=>-94.942275),
"/US/US-IA/tennant-7272" => array("type"=>"city", "pop"=>73, "name_ru"=>"Теннант", "name_en"=>"Tennant", "lang"=>"en", "lat"=>41.595526, "lon"=>-95.442950),
"/US/US-IA/terril-7273" => array("type"=>"city", "pop"=>364, "name_ru"=>"Террил", "name_en"=>"Terril", "lang"=>"en", "lat"=>43.308534, "lon"=>-94.968832),
"/US/US-IA/thayer-7274" => array("type"=>"city", "pop"=>54, "name_ru"=>"Тайер", "name_en"=>"Thayer", "lang"=>"en", "lat"=>41.029250, "lon"=>-94.049629),
"/US/US-IA/thompson-7275" => array("type"=>"city", "pop"=>546, "name_ru"=>"Томпсон", "name_en"=>"Thompson", "lang"=>"en", "lat"=>43.369608, "lon"=>-93.774333),
"/US/US-IA/thor-7276" => array("type"=>"city", "pop"=>158, "name_ru"=>"Тор", "name_en"=>"Thor", "lang"=>"en", "lat"=>42.688717, "lon"=>-94.049989),
"/US/US-IA/thornburg-7277" => array("type"=>"city", "pop"=>65, "name_ru"=>"Торнбург", "name_en"=>"Thornburg", "lang"=>"en", "lat"=>41.456253, "lon"=>-92.332585),
"/US/US-IA/thornton-7278" => array("type"=>"city", "pop"=>358, "name_ru"=>"Торнтон", "name_en"=>"Thornton", "lang"=>"en", "lat"=>42.944048, "lon"=>-93.387624),
"/US/US-IA/thurman-7279" => array("type"=>"city", "pop"=>209, "name_ru"=>"Турман", "name_en"=>"Thurman", "lang"=>"en", "lat"=>40.819439, "lon"=>-95.748368),
"/US/US-IA/tiffin-7280" => array("type"=>"city", "pop"=>1814, "name_ru"=>"Тиффин", "name_en"=>"Tiffin", "lang"=>"en", "lat"=>41.706398, "lon"=>-91.677278),
"/US/US-IA/tingley-7281" => array("type"=>"city", "pop"=>149, "name_ru"=>"Тингли", "name_en"=>"Tingley", "lang"=>"en", "lat"=>40.852761, "lon"=>-94.195537),
"/US/US-IA/tipton-7282" => array("type"=>"city", "pop"=>3012, "name_ru"=>"Типтон", "name_en"=>"Tipton", "lang"=>"en", "lat"=>41.770509, "lon"=>-91.128904),
"/US/US-IA/titonka-7283" => array("type"=>"city", "pop"=>515, "name_ru"=>"Титонка", "name_en"=>"Titonka", "lang"=>"en", "lat"=>43.236573, "lon"=>-94.041606),
"/US/US-IA/toledo-7284" => array("type"=>"city", "pop"=>2645, "name_ru"=>"Толедо", "name_en"=>"Toledo", "lang"=>"en", "lat"=>41.990461, "lon"=>-92.580521),
"/US/US-IA/toronto-7285" => array("type"=>"city", "pop"=>130, "name_ru"=>"Торонто", "name_en"=>"Toronto", "lang"=>"en", "lat"=>41.903319, "lon"=>-90.862866),
"/US/US-IA/traer-7286" => array("type"=>"city", "pop"=>1568, "name_ru"=>"Траер", "name_en"=>"Traer", "lang"=>"en", "lat"=>42.192468, "lon"=>-92.464734),
"/US/US-IA/treynor-7287" => array("type"=>"city", "pop"=>898, "name_ru"=>"Трейнор", "name_en"=>"Treynor", "lang"=>"en", "lat"=>41.231070, "lon"=>-95.606373),
"/US/US-IA/tripoli-7288" => array("type"=>"city", "pop"=>1258, "name_ru"=>"Триполи", "name_en"=>"Tripoli", "lang"=>"en", "lat"=>42.809756, "lon"=>-92.257580),
"/US/US-IA/truesdale-7289" => array("type"=>"city", "pop"=>87, "name_ru"=>"Трусдейл", "name_en"=>"Truesdale", "lang"=>"en", "lat"=>42.729475, "lon"=>-95.182731),
"/US/US-IA/truro-7290" => array("type"=>"city", "pop"=>485, "name_ru"=>"Труро", "name_en"=>"Truro", "lang"=>"en", "lat"=>41.210404, "lon"=>-93.845686),
"/US/US-IA/turin-7291" => array("type"=>"city", "pop"=>67, "name_ru"=>"Турин", "name_en"=>"Turin", "lang"=>"en", "lat"=>42.020358, "lon"=>-95.965795),
"/US/US-IA/udell-7292" => array("type"=>"city", "pop"=>47, "name_ru"=>"Уделл", "name_en"=>"Udell", "lang"=>"en", "lat"=>40.780172, "lon"=>-92.742524),
"/US/US-IA/underwood-7293" => array("type"=>"city", "pop"=>858, "name_ru"=>"Андервуд", "name_en"=>"Underwood", "lang"=>"en", "lat"=>41.385733, "lon"=>-95.679299),
"/US/US-IA/union-7294" => array("type"=>"city", "pop"=>388, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>42.242935, "lon"=>-93.062631),
"/US/US-IA/unionville-7295" => array("type"=>"city", "pop"=>122, "name_ru"=>"Юнионвилл", "name_en"=>"Unionville", "lang"=>"en", "lat"=>40.818969, "lon"=>-92.696827),
"/US/US-IA/university-heights-7296" => array("type"=>"city", "pop"=>1050, "name_ru"=>"Юниверсити Хайтс", "name_en"=>"University Heights", "lang"=>"en", "lat"=>41.654500, "lon"=>-91.558024),
"/US/US-IA/university-park-7297" => array("type"=>"city", "pop"=>542, "name_ru"=>"Юниверсити Парк", "name_en"=>"University Park", "lang"=>"en", "lat"=>41.286708, "lon"=>-92.615780),
"/US/US-IA/urbana-7298" => array("type"=>"city", "pop"=>1447, "name_ru"=>"Урбана", "name_en"=>"Urbana", "lang"=>"en", "lat"=>42.223287, "lon"=>-91.874975),
"/US/US-IA/urbandale-7299" => array("type"=>"city", "pop"=>38369, "name_ru"=>"Урбандейл", "name_en"=>"Urbandale", "lang"=>"en", "lat"=>41.636550, "lon"=>-93.740890),
"/US/US-IA/ute-7300" => array("type"=>"city", "pop"=>324, "name_ru"=>"Ют", "name_en"=>"Ute", "lang"=>"en", "lat"=>42.049723, "lon"=>-95.705596),
"/US/US-IA/vail-7301" => array("type"=>"city", "pop"=>435, "name_ru"=>"Вейл", "name_en"=>"Vail", "lang"=>"en", "lat"=>42.059761, "lon"=>-95.200605),
"/US/US-IA/valeria-7302" => array("type"=>"city", "pop"=>60, "name_ru"=>"Валерия", "name_en"=>"Valeria", "lang"=>"en", "lat"=>41.729487, "lon"=>-93.324547),
"/US/US-IA/van-horne-7303" => array("type"=>"city", "pop"=>722, "name_ru"=>"Ван Хорн", "name_en"=>"Van Horne", "lang"=>"en", "lat"=>42.010003, "lon"=>-92.090127),
"/US/US-IA/van-meter-7304" => array("type"=>"city", "pop"=>1208, "name_ru"=>"Ван Митер", "name_en"=>"Van Meter", "lang"=>"en", "lat"=>41.530461, "lon"=>-93.955683),
"/US/US-IA/van-wert-7305" => array("type"=>"city", "pop"=>214, "name_ru"=>"Ван Верт", "name_en"=>"Van Wert", "lang"=>"en", "lat"=>40.870770, "lon"=>-93.791880),
"/US/US-IA/varina-7306" => array("type"=>"city", "pop"=>78, "name_ru"=>"Варина", "name_en"=>"Varina", "lang"=>"en", "lat"=>42.658606, "lon"=>-94.898508),
"/US/US-IA/ventura-7307" => array("type"=>"city", "pop"=>718, "name_ru"=>"Вентура", "name_en"=>"Ventura", "lang"=>"en", "lat"=>43.126427, "lon"=>-93.459837),
"/US/US-IA/victor-7308" => array("type"=>"city", "pop"=>983, "name_ru"=>"Виктор", "name_en"=>"Victor", "lang"=>"en", "lat"=>41.730910, "lon"=>-92.294424),
"/US/US-IA/villisca-7309" => array("type"=>"city", "pop"=>1233, "name_ru"=>"Виллиска", "name_en"=>"Villisca", "lang"=>"en", "lat"=>40.931012, "lon"=>-94.977889),
"/US/US-IA/vincent-7310" => array("type"=>"city", "pop"=>151, "name_ru"=>"Винсент", "name_en"=>"Vincent", "lang"=>"en", "lat"=>42.591260, "lon"=>-94.017917),
"/US/US-IA/vining-7311" => array("type"=>"city", "pop"=>50, "name_ru"=>"Вининг", "name_en"=>"Vining", "lang"=>"en", "lat"=>41.990218, "lon"=>-92.385455),
"/US/US-IA/vinton-7312" => array("type"=>"city", "pop"=>5080, "name_ru"=>"Винтон", "name_en"=>"Vinton", "lang"=>"en", "lat"=>42.164154, "lon"=>-92.027014),
"/US/US-IA/volga-7313" => array("type"=>"city", "pop"=>222, "name_ru"=>"Волга", "name_en"=>"Volga", "lang"=>"en", "lat"=>42.802521, "lon"=>-91.542926),
"/US/US-IA/wadena-7314" => array("type"=>"city", "pop"=>218, "name_ru"=>"Вадена", "name_en"=>"Wadena", "lang"=>"en", "lat"=>42.839511, "lon"=>-91.657978),
"/US/US-IA/wahpeton-7315" => array("type"=>"city", "pop"=>438, "name_ru"=>"Уопетон", "name_en"=>"Wahpeton", "lang"=>"en", "lat"=>43.374476, "lon"=>-95.174545),
"/US/US-IA/walcott-7316" => array("type"=>"city", "pop"=>1628, "name_ru"=>"Уолкотт", "name_en"=>"Walcott", "lang"=>"en", "lat"=>41.594630, "lon"=>-90.775017),
"/US/US-IA/walford-7317" => array("type"=>"city", "pop"=>1425, "name_ru"=>"Уолфорд", "name_en"=>"Walford", "lang"=>"en", "lat"=>41.878325, "lon"=>-91.833245),
"/US/US-IA/walker-7318" => array("type"=>"city", "pop"=>775, "name_ru"=>"Уолкер", "name_en"=>"Walker", "lang"=>"en", "lat"=>42.286231, "lon"=>-91.780935),
"/US/US-IA/wall-lake-7320" => array("type"=>"city", "pop"=>747, "name_ru"=>"Уолл Лейк", "name_en"=>"Wall Lake", "lang"=>"en", "lat"=>42.269552, "lon"=>-95.092965),
"/US/US-IA/wallingford-7319" => array("type"=>"city", "pop"=>194, "name_ru"=>"Уоллингтон", "name_en"=>"Wallingford", "lang"=>"en", "lat"=>43.320165, "lon"=>-94.791994),
"/US/US-IA/walnut-7321" => array("type"=>"city", "pop"=>804, "name_ru"=>"Уолнат", "name_en"=>"Walnut", "lang"=>"en", "lat"=>41.479111, "lon"=>-95.222669),
"/US/US-IA/wapello-7322" => array("type"=>"city", "pop"=>2013, "name_ru"=>"Вапелло", "name_en"=>"Wapello", "lang"=>"en", "lat"=>41.178356, "lon"=>-91.188794),
"/US/US-IA/washington-7323" => array("type"=>"city", "pop"=>7254, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>41.298848, "lon"=>-91.690884),
"/US/US-IA/washta-7324" => array("type"=>"city", "pop"=>242, "name_ru"=>"Уошта", "name_en"=>"Washta", "lang"=>"en", "lat"=>42.575886, "lon"=>-95.719180),
"/US/US-IA/waterloo-7325" => array("type"=>"city", "pop"=>66662, "name_ru"=>"Ватрелоо", "name_en"=>"Waterloo", "lang"=>"en", "lat"=>42.491650, "lon"=>-92.351058),
"/US/US-IA/waterville-7326" => array("type"=>"city", "pop"=>133, "name_ru"=>"Уотервилл", "name_en"=>"Waterville", "lang"=>"en", "lat"=>43.208072, "lon"=>-91.294833),
"/US/US-IA/waucoma-7327" => array("type"=>"city", "pop"=>238, "name_ru"=>"Уоукома", "name_en"=>"Waucoma", "lang"=>"en", "lat"=>43.055343, "lon"=>-92.034598),
"/US/US-IA/waukee-7328" => array("type"=>"city", "pop"=>12367, "name_ru"=>"Уоуки", "name_en"=>"Waukee", "lang"=>"en", "lat"=>41.607608, "lon"=>-93.862387),
"/US/US-IA/waukon-7329" => array("type"=>"city", "pop"=>3908, "name_ru"=>"Уоукон", "name_en"=>"Waukon", "lang"=>"en", "lat"=>43.269817, "lon"=>-91.476350),
"/US/US-IA/waverly-7330" => array("type"=>"city", "pop"=>9298, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>42.725426, "lon"=>-92.469822),
"/US/US-IA/wayland-7331" => array("type"=>"city", "pop"=>926, "name_ru"=>"Уэйленд", "name_en"=>"Wayland", "lang"=>"en", "lat"=>41.147703, "lon"=>-91.657882),
"/US/US-IA/webb-7332" => array("type"=>"city", "pop"=>151, "name_ru"=>"Уэбб", "name_en"=>"Webb", "lang"=>"en", "lat"=>42.949249, "lon"=>-95.011967),
"/US/US-IA/webster-7333" => array("type"=>"city", "pop"=>107, "name_ru"=>"Вебстер", "name_en"=>"Webster", "lang"=>"en", "lat"=>41.437886, "lon"=>-92.170844),
"/US/US-IA/webster-city-7334" => array("type"=>"city", "pop"=>7720, "name_ru"=>"Вэбстер Сити", "name_en"=>"Webster City", "lang"=>"en", "lat"=>42.463422, "lon"=>-93.816453),
"/US/US-IA/weldon-7335" => array("type"=>"city", "pop"=>141, "name_ru"=>"Уэлдон", "name_en"=>"Weldon", "lang"=>"en", "lat"=>40.897947, "lon"=>-93.735103),
"/US/US-IA/wellman-7336" => array("type"=>"city", "pop"=>1427, "name_ru"=>"Уэллман", "name_en"=>"Wellman", "lang"=>"en", "lat"=>41.467872, "lon"=>-91.837020),
"/US/US-IA/wellsburg-7337" => array("type"=>"city", "pop"=>679, "name_ru"=>"Уэллсбург", "name_en"=>"Wellsburg", "lang"=>"en", "lat"=>42.433753, "lon"=>-92.924421),
"/US/US-IA/welton-7338" => array("type"=>"city", "pop"=>158, "name_ru"=>"Уэлтон", "name_en"=>"Welton", "lang"=>"en", "lat"=>41.911864, "lon"=>-90.597848),
"/US/US-IA/wesley-7339" => array("type"=>"city", "pop"=>414, "name_ru"=>"Уэсли", "name_en"=>"Wesley", "lang"=>"en", "lat"=>43.088899, "lon"=>-93.993002),
"/US/US-IA/west-bend-7340" => array("type"=>"city", "pop"=>794, "name_ru"=>"Вэст Бенд", "name_en"=>"West Bend", "lang"=>"en", "lat"=>42.960089, "lon"=>-94.446778),
"/US/US-IA/west-branch-7341" => array("type"=>"city", "pop"=>2296, "name_ru"=>"Вэст Брэнч", "name_en"=>"West Branch", "lang"=>"en", "lat"=>41.670083, "lon"=>-91.349199),
"/US/US-IA/west-burlington-7342" => array("type"=>"city", "pop"=>3283, "name_ru"=>"Вэст Бёрлингтон", "name_en"=>"West Burlington", "lang"=>"en", "lat"=>40.823522, "lon"=>-91.178564),
"/US/US-IA/west-chester-7343" => array("type"=>"city", "pop"=>159, "name_ru"=>"Вэст Честер", "name_en"=>"West Chester", "lang"=>"en", "lat"=>41.340050, "lon"=>-91.816172),
"/US/US-IA/west-des-moines-7344" => array("type"=>"city", "pop"=>55426, "name_ru"=>"Вэст Де Мойнс", "name_en"=>"West Des Moines", "lang"=>"en", "lat"=>41.570700, "lon"=>-93.752915),
"/US/US-IA/west-liberty-7347" => array("type"=>"city", "pop"=>3650, "name_ru"=>"Вэст Либерти", "name_en"=>"West Liberty", "lang"=>"en", "lat"=>41.572662, "lon"=>-91.263804),
"/US/US-IA/west-okoboji-7348" => array("type"=>"city", "pop"=>409, "name_ru"=>"Вэст", "name_en"=>"West Okoboji", "lang"=>"en", "lat"=>43.349416, "lon"=>-95.165323),
"/US/US-IA/west-point-7350" => array("type"=>"city", "pop"=>915, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>40.715817, "lon"=>-91.451524),
"/US/US-IA/west-union-7352" => array("type"=>"city", "pop"=>2429, "name_ru"=>"Вэст Юнион", "name_en"=>"West Union", "lang"=>"en", "lat"=>42.958695, "lon"=>-91.812620),
"/US/US-IA/westfield-7345" => array("type"=>"city", "pop"=>178, "name_ru"=>"Уэстфилд", "name_en"=>"Westfield", "lang"=>"en", "lat"=>42.756728, "lon"=>-96.605409),
"/US/US-IA/westgate-7346" => array("type"=>"city", "pop"=>216, "name_ru"=>"Вэстгейт", "name_en"=>"Westgate", "lang"=>"en", "lat"=>42.768485, "lon"=>-91.995092),
"/US/US-IA/westphalia-7349" => array("type"=>"city", "pop"=>142, "name_ru"=>"Вэстфалия", "name_en"=>"Westphalia", "lang"=>"en", "lat"=>41.719318, "lon"=>-95.393070),
"/US/US-IA/westside-7351" => array("type"=>"city", "pop"=>322, "name_ru"=>"Вэстсайд", "name_en"=>"Westside", "lang"=>"en", "lat"=>42.075243, "lon"=>-95.099952),
"/US/US-IA/westwood-7353" => array("type"=>"city", "pop"=>242, "name_ru"=>"Уэствуд", "name_en"=>"Westwood", "lang"=>"en", "lat"=>40.964805, "lon"=>-91.627208),
"/US/US-IA/what-cheer-7354" => array("type"=>"city", "pop"=>628, "name_ru"=>"Уот Чиер", "name_en"=>"What Cheer", "lang"=>"en", "lat"=>41.400137, "lon"=>-92.353882),
"/US/US-IA/wheatland-7355" => array("type"=>"city", "pop"=>745, "name_ru"=>"Уитленд", "name_en"=>"Wheatland", "lang"=>"en", "lat"=>41.832868, "lon"=>-90.837234),
"/US/US-IA/whiting-7356" => array("type"=>"city", "pop"=>703, "name_ru"=>"Уайтинг", "name_en"=>"Whiting", "lang"=>"en", "lat"=>42.126384, "lon"=>-96.150736),
"/US/US-IA/whittemore-7357" => array("type"=>"city", "pop"=>477, "name_ru"=>"Уиттмор", "name_en"=>"Whittemore", "lang"=>"en", "lat"=>43.063386, "lon"=>-94.423979),
"/US/US-IA/whitten-7358" => array("type"=>"city", "pop"=>134, "name_ru"=>"Уиттен", "name_en"=>"Whitten", "lang"=>"en", "lat"=>42.264798, "lon"=>-93.011372),
"/US/US-IA/willey-7359" => array("type"=>"city", "pop"=>99, "name_ru"=>"Уилли", "name_en"=>"Willey", "lang"=>"en", "lat"=>41.978396, "lon"=>-94.820719),
"/US/US-IA/williams-7360" => array("type"=>"city", "pop"=>374, "name_ru"=>"Уилльямс", "name_en"=>"Williams", "lang"=>"en", "lat"=>42.490148, "lon"=>-93.541557),
"/US/US-IA/williamsburg-7361" => array("type"=>"city", "pop"=>2812, "name_ru"=>"Уилльямсбург", "name_en"=>"Williamsburg", "lang"=>"en", "lat"=>41.668007, "lon"=>-92.008431),
"/US/US-IA/williamson-7362" => array("type"=>"city", "pop"=>158, "name_ru"=>"Уильямсон", "name_en"=>"Williamson", "lang"=>"en", "lat"=>41.087856, "lon"=>-93.256668),
"/US/US-IA/wilton-7363" => array("type"=>"city", "pop"=>2821, "name_ru"=>"Уилтон", "name_en"=>"Wilton", "lang"=>"en", "lat"=>41.589073, "lon"=>-91.024710),
"/US/US-IA/windsor-heights-7364" => array("type"=>"city", "pop"=>4643, "name_ru"=>"Виндзор Хайтс", "name_en"=>"Windsor Heights", "lang"=>"en", "lat"=>41.604800, "lon"=>-93.711899),
"/US/US-IA/winfield-7365" => array("type"=>"city", "pop"=>1082, "name_ru"=>"Винфилд", "name_en"=>"Winfield", "lang"=>"en", "lat"=>41.125825, "lon"=>-91.437214),
"/US/US-IA/winterset-7366" => array("type"=>"city", "pop"=>4820, "name_ru"=>"Уинтерсет", "name_en"=>"Winterset", "lang"=>"en", "lat"=>41.337977, "lon"=>-94.018532),
"/US/US-IA/winthrop-7367" => array("type"=>"city", "pop"=>734, "name_ru"=>"Уинтроп", "name_en"=>"Winthrop", "lang"=>"en", "lat"=>42.473258, "lon"=>-91.733670),
"/US/US-IA/wiota-7368" => array("type"=>"city", "pop"=>142, "name_ru"=>"Уиота", "name_en"=>"Wiota", "lang"=>"en", "lat"=>41.400590, "lon"=>-94.887469),
"/US/US-IA/woden-7369" => array("type"=>"city", "pop"=>219, "name_ru"=>"Воден", "name_en"=>"Woden", "lang"=>"en", "lat"=>43.230512, "lon"=>-93.910446),
"/US/US-IA/woodbine-7370" => array("type"=>"city", "pop"=>1527, "name_ru"=>"Вудбайн", "name_en"=>"Woodbine", "lang"=>"en", "lat"=>41.736491, "lon"=>-95.707075),
"/US/US-IA/woodburn-7371" => array("type"=>"city", "pop"=>236, "name_ru"=>"Вудберн", "name_en"=>"Woodburn", "lang"=>"en", "lat"=>41.010494, "lon"=>-93.596487),
"/US/US-IA/woodward-7372" => array("type"=>"city", "pop"=>1429, "name_ru"=>"Вудуорд", "name_en"=>"Woodward", "lang"=>"en", "lat"=>41.856267, "lon"=>-93.921275),
"/US/US-IA/woolstock-7373" => array("type"=>"city", "pop"=>188, "name_ru"=>"Вулсток", "name_en"=>"Woolstock", "lang"=>"en", "lat"=>42.565290, "lon"=>-93.843355),
"/US/US-IA/worthington-7374" => array("type"=>"city", "pop"=>366, "name_ru"=>"Уортингтон", "name_en"=>"Worthington", "lang"=>"en", "lat"=>42.396523, "lon"=>-91.120619),
"/US/US-IA/wyoming-7375" => array("type"=>"city", "pop"=>592, "name_ru"=>"Вайоминг", "name_en"=>"Wyoming", "lang"=>"en", "lat"=>42.060112, "lon"=>-91.004837),
"/US/US-IA/yale-7376" => array("type"=>"city", "pop"=>259, "name_ru"=>"Йель", "name_en"=>"Yale", "lang"=>"en", "lat"=>41.775349, "lon"=>-94.356783),
"/US/US-IA/yetter-7377" => array("type"=>"city", "pop"=>28, "name_ru"=>"Йеттер", "name_en"=>"Yetter", "lang"=>"en", "lat"=>42.316491, "lon"=>-94.843558),
"/US/US-IA/yorktown-7378" => array("type"=>"city", "pop"=>76, "name_ru"=>"Йорктаун", "name_en"=>"Yorktown", "lang"=>"en", "lat"=>40.736461, "lon"=>-95.154935),
"/US/US-IA/zearing-7379" => array("type"=>"city", "pop"=>589, "name_ru"=>"Зеринг", "name_en"=>"Zearing", "lang"=>"en", "lat"=>42.158880, "lon"=>-93.296533),
"/US/US-IA/zwingle-7380" => array("type"=>"city", "pop"=>109, "name_ru"=>"Звингл", "name_en"=>"Zwingle", "lang"=>"en", "lat"=>42.297826, "lon"=>-90.686908),
"/US/US-ID" => array("type"=>"state", "pop"=>1523816, "name_ru"=>"Айдахо", "name_en"=>"Idaho", "lang"=>"en"),
"/US/US-ID/aberdeen-4325" => array("type"=>"city", "pop"=>1749, "name_ru"=>"Абердин", "name_en"=>"Aberdeen", "lang"=>"en", "lat"=>42.944580, "lon"=>-112.836979),
"/US/US-ID/acequia-4326" => array("type"=>"city", "pop"=>132, "name_ru"=>"Асекия", "name_en"=>"Acequia", "lang"=>"en", "lat"=>42.668490, "lon"=>-113.595629),
"/US/US-ID/albion-4327" => array("type"=>"city", "pop"=>255, "name_ru"=>"Албион", "name_en"=>"Albion", "lang"=>"en", "lat"=>42.410972, "lon"=>-113.580061),
"/US/US-ID/american-falls-4328" => array("type"=>"city", "pop"=>4069, "name_ru"=>"Американ Фолс", "name_en"=>"American Falls", "lang"=>"en", "lat"=>42.781980, "lon"=>-112.853778),
"/US/US-ID/ammon-4329" => array("type"=>"city", "pop"=>13646, "name_ru"=>"Аммон", "name_en"=>"Ammon", "lang"=>"en", "lat"=>43.479650, "lon"=>-111.968295),
"/US/US-ID/arbon-valley-4330" => array("type"=>"cdp", "pop"=>613, "name_ru"=>"Арбон Валли", "name_en"=>"Arbon Valley", "lang"=>"en", "lat"=>42.883895, "lon"=>-112.597796),
"/US/US-ID/arco-4331" => array("type"=>"city", "pop"=>977, "name_ru"=>"Арко", "name_en"=>"Arco", "lang"=>"en", "lat"=>43.634125, "lon"=>-113.299764),
"/US/US-ID/arimo-4332" => array("type"=>"city", "pop"=>320, "name_ru"=>"Аримо", "name_en"=>"Arimo", "lang"=>"en", "lat"=>42.559543, "lon"=>-112.170601),
"/US/US-ID/ashton-4333" => array("type"=>"city", "pop"=>1073, "name_ru"=>"Аштон", "name_en"=>"Ashton", "lang"=>"en", "lat"=>44.073376, "lon"=>-111.448604),
"/US/US-ID/athol-4334" => array("type"=>"city", "pop"=>684, "name_ru"=>"Атол", "name_en"=>"Athol", "lang"=>"en", "lat"=>47.945068, "lon"=>-116.706480),
"/US/US-ID/atomic-city-4335" => array("type"=>"city", "pop"=>25, "name_ru"=>"Атомик Сити", "name_en"=>"Atomic City", "lang"=>"en", "lat"=>43.444285, "lon"=>-112.812550),
"/US/US-ID/bancroft-4336" => array("type"=>"city", "pop"=>344, "name_ru"=>"Бэнкрофт", "name_en"=>"Bancroft", "lang"=>"en", "lat"=>42.720210, "lon"=>-111.882327),
"/US/US-ID/basalt-4337" => array("type"=>"city", "pop"=>419, "name_ru"=>"Базальт", "name_en"=>"Basalt", "lang"=>"en", "lat"=>43.314356, "lon"=>-112.165153),
"/US/US-ID/bellevue-4338" => array("type"=>"city", "pop"=>2165, "name_ru"=>"Белвью", "name_en"=>"Bellevue", "lang"=>"en", "lat"=>43.465914, "lon"=>-114.257358),
"/US/US-ID/blackfoot-4339" => array("type"=>"city", "pop"=>11012, "name_ru"=>"Блэкфут", "name_en"=>"Blackfoot", "lang"=>"en", "lat"=>43.193985, "lon"=>-112.344922),
"/US/US-ID/bliss-4340" => array("type"=>"city", "pop"=>247, "name_ru"=>"Блисс", "name_en"=>"Bliss", "lang"=>"en", "lat"=>42.926222, "lon"=>-114.949110),
"/US/US-ID/bloomington-4341" => array("type"=>"city", "pop"=>225, "name_ru"=>"Блумингтон", "name_en"=>"Bloomington", "lang"=>"en", "lat"=>42.190887, "lon"=>-111.404838),
"/US/US-ID/boise-city-4342" => array("type"=>"city", "pop"=>205314, "name_ru"=>"Бойсе", "name_en"=>"Boise City", "lang"=>"en", "lat"=>43.606651, "lon"=>-116.226100),
"/US/US-ID/bonners-ferry-4343" => array("type"=>"city", "pop"=>2596, "name_ru"=>"Боннерс Ферри", "name_en"=>"Bonners Ferry", "lang"=>"en", "lat"=>48.691664, "lon"=>-116.315110),
"/US/US-ID/bovill-4344" => array("type"=>"city", "pop"=>287, "name_ru"=>"Бовилл", "name_en"=>"Bovill", "lang"=>"en", "lat"=>46.858257, "lon"=>-116.393307),
"/US/US-ID/buhl-4345" => array("type"=>"city", "pop"=>4077, "name_ru"=>"Бул", "name_en"=>"Buhl", "lang"=>"en", "lat"=>42.598563, "lon"=>-114.760714),
"/US/US-ID/burley-4346" => array("type"=>"city", "pop"=>9105, "name_ru"=>"Берли", "name_en"=>"Burley", "lang"=>"en", "lat"=>42.538855, "lon"=>-113.787891),
"/US/US-ID/butte-city-4347" => array("type"=>"city", "pop"=>72, "name_ru"=>"Бьютт Сити", "name_en"=>"Butte City", "lang"=>"en", "lat"=>43.606406, "lon"=>-113.238451),
"/US/US-ID/caldwell-4348" => array("type"=>"city", "pop"=>42331, "name_ru"=>"Колдуэлл", "name_en"=>"Caldwell", "lang"=>"en", "lat"=>43.658340, "lon"=>-116.674864),
"/US/US-ID/cambridge-4349" => array("type"=>"city", "pop"=>380, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>44.572589, "lon"=>-116.675909),
"/US/US-ID/cascade-4350" => array("type"=>"city", "pop"=>991, "name_ru"=>"Каскейд", "name_en"=>"Cascade", "lang"=>"en", "lat"=>44.513023, "lon"=>-116.040659),
"/US/US-ID/castleford-4351" => array("type"=>"city", "pop"=>271, "name_ru"=>"Каслфорд", "name_en"=>"Castleford", "lang"=>"en", "lat"=>42.521310, "lon"=>-114.871107),
"/US/US-ID/challis-4352" => array("type"=>"city", "pop"=>896, "name_ru"=>"Чаллис", "name_en"=>"Challis", "lang"=>"en", "lat"=>44.507739, "lon"=>-114.223933),
"/US/US-ID/chatcolet-4353" => array("type"=>"city", "pop"=>72, "name_ru"=>"Чатколет", "name_en"=>"Chatcolet", "lang"=>"en", "lat"=>47.354128, "lon"=>-116.752500),
"/US/US-ID/chubbuck-4354" => array("type"=>"city", "pop"=>11828, "name_ru"=>"Чуббак", "name_en"=>"Chubbuck", "lang"=>"en", "lat"=>42.922250, "lon"=>-112.464724),
"/US/US-ID/clark-fork-4355" => array("type"=>"city", "pop"=>578, "name_ru"=>"Кларк Форк", "name_en"=>"Clark Fork", "lang"=>"en", "lat"=>48.147926, "lon"=>-116.176681),
"/US/US-ID/clayton-4356" => array("type"=>"city", "pop"=>27, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>44.259058, "lon"=>-114.398496),
"/US/US-ID/clifton-4357" => array("type"=>"city", "pop"=>238, "name_ru"=>"Слифтон", "name_en"=>"Clifton", "lang"=>"en", "lat"=>42.187664, "lon"=>-112.005066),
"/US/US-ID/coeur-d'alene-4358" => array("type"=>"city", "pop"=>43360, "name_ru"=>"Коер Долин", "name_en"=>"Coeur d'Alene", "lang"=>"en", "lat"=>47.697788, "lon"=>-116.784816),
"/US/US-ID/cottonwood-4359" => array("type"=>"city", "pop"=>1041, "name_ru"=>"Коттонвуд", "name_en"=>"Cottonwood", "lang"=>"en", "lat"=>46.050476, "lon"=>-116.347867),
"/US/US-ID/council-4360" => array("type"=>"city", "pop"=>680, "name_ru"=>"Каунсил", "name_en"=>"Council", "lang"=>"en", "lat"=>44.728915, "lon"=>-116.435245),
"/US/US-ID/craigmont-4361" => array("type"=>"city", "pop"=>519, "name_ru"=>"Крейгмонт", "name_en"=>"Craigmont", "lang"=>"en", "lat"=>46.242843, "lon"=>-116.471306),
"/US/US-ID/crouch-4362" => array("type"=>"city", "pop"=>161, "name_ru"=>"Кроуч", "name_en"=>"Crouch", "lang"=>"en", "lat"=>44.115328, "lon"=>-115.971640),
"/US/US-ID/culdesac-4363" => array("type"=>"city", "pop"=>376, "name_ru"=>"Калдесак", "name_en"=>"Culdesac", "lang"=>"en", "lat"=>46.374995, "lon"=>-116.669352),
"/US/US-ID/dalton-gardens-4364" => array("type"=>"city", "pop"=>2389, "name_ru"=>"Долтон Гарденс", "name_en"=>"Dalton Gardens", "lang"=>"en", "lat"=>47.733467, "lon"=>-116.766867),
"/US/US-ID/dayton-4365" => array("type"=>"city", "pop"=>442, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>42.113012, "lon"=>-111.976392),
"/US/US-ID/deary-4366" => array("type"=>"city", "pop"=>541, "name_ru"=>"Диэри", "name_en"=>"Deary", "lang"=>"en", "lat"=>46.800633, "lon"=>-116.556477),
"/US/US-ID/declo-4367" => array("type"=>"city", "pop"=>329, "name_ru"=>"Дэксло", "name_en"=>"Declo", "lang"=>"en", "lat"=>42.519259, "lon"=>-113.628112),
"/US/US-ID/dietrich-4368" => array("type"=>"city", "pop"=>160, "name_ru"=>"Дитрих", "name_en"=>"Dietrich", "lang"=>"en", "lat"=>42.913698, "lon"=>-114.264033),
"/US/US-ID/donnelly-4369" => array("type"=>"city", "pop"=>142, "name_ru"=>"Доннелли", "name_en"=>"Donnelly", "lang"=>"en", "lat"=>44.730281, "lon"=>-116.075920),
"/US/US-ID/dover-4370" => array("type"=>"city", "pop"=>527, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>48.253618, "lon"=>-116.600323),
"/US/US-ID/downey-4371" => array("type"=>"city", "pop"=>562, "name_ru"=>"Дауни", "name_en"=>"Downey", "lang"=>"en", "lat"=>42.429403, "lon"=>-112.122813),
"/US/US-ID/driggs-4372" => array("type"=>"city", "pop"=>1362, "name_ru"=>"Дригс", "name_en"=>"Driggs", "lang"=>"en", "lat"=>43.725303, "lon"=>-111.107145),
"/US/US-ID/drummond-4373" => array("type"=>"city", "pop"=>14, "name_ru"=>"Драммонд", "name_en"=>"Drummond", "lang"=>"en", "lat"=>43.998887, "lon"=>-111.341649),
"/US/US-ID/dubois-4374" => array("type"=>"city", "pop"=>616, "name_ru"=>"Дюбуа", "name_en"=>"Dubois", "lang"=>"en", "lat"=>44.173538, "lon"=>-112.229441),
"/US/US-ID/eagle-4375" => array("type"=>"city", "pop"=>19471, "name_ru"=>"Игл", "name_en"=>"Eagle", "lang"=>"en", "lat"=>43.693529, "lon"=>-116.354783),
"/US/US-ID/east-hope-4376" => array("type"=>"city", "pop"=>218, "name_ru"=>"Ист Хоуп", "name_en"=>"East Hope", "lang"=>"en", "lat"=>48.241839, "lon"=>-116.291322),
"/US/US-ID/eden-4377" => array("type"=>"city", "pop"=>405, "name_ru"=>"Иден", "name_en"=>"Eden", "lang"=>"en", "lat"=>42.605150, "lon"=>-114.208724),
"/US/US-ID/elk-river-4378" => array("type"=>"city", "pop"=>136, "name_ru"=>"Элк Ривер", "name_en"=>"Elk River", "lang"=>"en", "lat"=>46.782472, "lon"=>-116.180394),
"/US/US-ID/emmett-4379" => array("type"=>"city", "pop"=>6357, "name_ru"=>"Эммет", "name_en"=>"Emmett", "lang"=>"en", "lat"=>43.873942, "lon"=>-116.493934),
"/US/US-ID/fairfield-4380" => array("type"=>"city", "pop"=>417, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>43.346066, "lon"=>-114.792050),
"/US/US-ID/ferdinand-4381" => array("type"=>"city", "pop"=>145, "name_ru"=>"Фердинанд", "name_en"=>"Ferdinand", "lang"=>"en", "lat"=>46.152817, "lon"=>-116.389461),
"/US/US-ID/fernan-lake-village-4382" => array("type"=>"city", "pop"=>183, "name_ru"=>"Фернан Лейк Виллидж", "name_en"=>"Fernan Lake Village", "lang"=>"en", "lat"=>47.673289, "lon"=>-116.746966),
"/US/US-ID/filer-4383" => array("type"=>"city", "pop"=>2141, "name_ru"=>"Файлер", "name_en"=>"Filer", "lang"=>"en", "lat"=>42.568899, "lon"=>-114.608896),
"/US/US-ID/firth-4384" => array("type"=>"city", "pop"=>486, "name_ru"=>"Ферт", "name_en"=>"Firth", "lang"=>"en", "lat"=>43.305604, "lon"=>-112.183132),
"/US/US-ID/fort-hall-4385" => array("type"=>"cdp", "pop"=>2681, "name_ru"=>"Форт Холл", "name_en"=>"Fort Hall", "lang"=>"en", "lat"=>43.006541, "lon"=>-112.447297),
"/US/US-ID/franklin-4386" => array("type"=>"city", "pop"=>665, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>42.016113, "lon"=>-111.802321),
"/US/US-ID/fruitland-4387" => array("type"=>"city", "pop"=>4732, "name_ru"=>"Фрутленд", "name_en"=>"Fruitland", "lang"=>"en", "lat"=>44.018037, "lon"=>-116.920087),
"/US/US-ID/garden-city-4388" => array("type"=>"city", "pop"=>11713, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>43.648467, "lon"=>-116.267014),
"/US/US-ID/genesee-4389" => array("type"=>"city", "pop"=>878, "name_ru"=>"Джинеси", "name_en"=>"Genesee", "lang"=>"en", "lat"=>46.551214, "lon"=>-116.927966),
"/US/US-ID/georgetown-4390" => array("type"=>"city", "pop"=>463, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>42.477765, "lon"=>-111.365048),
"/US/US-ID/glenns-ferry-4391" => array("type"=>"city", "pop"=>1388, "name_ru"=>"Гленнс Ферри", "name_en"=>"Glenns Ferry", "lang"=>"en", "lat"=>42.954024, "lon"=>-115.299930),
"/US/US-ID/gooding-4392" => array("type"=>"city", "pop"=>3200, "name_ru"=>"Гудинг", "name_en"=>"Gooding", "lang"=>"en", "lat"=>42.938121, "lon"=>-114.712604),
"/US/US-ID/grace-4393" => array("type"=>"city", "pop"=>939, "name_ru"=>"Грейс", "name_en"=>"Grace", "lang"=>"en", "lat"=>42.576337, "lon"=>-111.730031),
"/US/US-ID/grand-view-4394" => array("type"=>"city", "pop"=>462, "name_ru"=>"Гранд Вью", "name_en"=>"Grand View", "lang"=>"en", "lat"=>42.983514, "lon"=>-116.090005),
"/US/US-ID/grangeville-4395" => array("type"=>"city", "pop"=>3110, "name_ru"=>"Гранджевилл", "name_en"=>"Grangeville", "lang"=>"en", "lat"=>45.925913, "lon"=>-116.120807),
"/US/US-ID/greenleaf-4396" => array("type"=>"city", "pop"=>893, "name_ru"=>"Гринлиф", "name_en"=>"Greenleaf", "lang"=>"en", "lat"=>43.672274, "lon"=>-116.820313),
"/US/US-ID/hagerman-4397" => array("type"=>"city", "pop"=>791, "name_ru"=>"Хаджерман", "name_en"=>"Hagerman", "lang"=>"en", "lat"=>42.817002, "lon"=>-114.896655),
"/US/US-ID/hailey-4398" => array("type"=>"city", "pop"=>7883, "name_ru"=>"Хейли", "name_en"=>"Hailey", "lang"=>"en", "lat"=>43.511660, "lon"=>-114.299162),
"/US/US-ID/hamer-4399" => array("type"=>"city", "pop"=>12, "name_ru"=>"Хамер", "name_en"=>"Hamer", "lang"=>"en", "lat"=>43.925677, "lon"=>-112.203680),
"/US/US-ID/hansen-4400" => array("type"=>"city", "pop"=>1022, "name_ru"=>"Хансен", "name_en"=>"Hansen", "lang"=>"en", "lat"=>42.531333, "lon"=>-114.300686),
"/US/US-ID/harrison-4401" => array("type"=>"city", "pop"=>284, "name_ru"=>"Харрисон", "name_en"=>"Harrison", "lang"=>"en", "lat"=>47.452406, "lon"=>-116.780424),
"/US/US-ID/hauser-4402" => array("type"=>"city", "pop"=>791, "name_ru"=>"Хаузер", "name_en"=>"Hauser", "lang"=>"en", "lat"=>47.769663, "lon"=>-117.011826),
"/US/US-ID/hayden-4403" => array("type"=>"city", "pop"=>12873, "name_ru"=>"Хейден", "name_en"=>"Hayden", "lang"=>"en", "lat"=>47.760196, "lon"=>-116.788530),
"/US/US-ID/hayden-lake-4404" => array("type"=>"city", "pop"=>560, "name_ru"=>"Хэйден Лейк", "name_en"=>"Hayden Lake", "lang"=>"en", "lat"=>47.765774, "lon"=>-116.752820),
"/US/US-ID/hazelton-4405" => array("type"=>"city", "pop"=>742, "name_ru"=>"Хейзлтон", "name_en"=>"Hazelton", "lang"=>"en", "lat"=>42.594770, "lon"=>-114.134307),
"/US/US-ID/heyburn-4406" => array("type"=>"city", "pop"=>2696, "name_ru"=>"Хейберн", "name_en"=>"Heyburn", "lang"=>"en", "lat"=>42.555789, "lon"=>-113.760939),
"/US/US-ID/hollister-4407" => array("type"=>"city", "pop"=>233, "name_ru"=>"Холлистер", "name_en"=>"Hollister", "lang"=>"en", "lat"=>42.353844, "lon"=>-114.587242),
"/US/US-ID/homedale-4408" => array("type"=>"city", "pop"=>2471, "name_ru"=>"Хоумдейл", "name_en"=>"Homedale", "lang"=>"en", "lat"=>43.618071, "lon"=>-116.935870),
"/US/US-ID/hope-4409" => array("type"=>"city", "pop"=>86, "name_ru"=>"Хоуп", "name_en"=>"Hope", "lang"=>"en", "lat"=>48.247904, "lon"=>-116.306558),
"/US/US-ID/horseshoe-bend-4410" => array("type"=>"city", "pop"=>854, "name_ru"=>"Хорсшу Бенд", "name_en"=>"Horseshoe Bend", "lang"=>"en", "lat"=>43.913016, "lon"=>-116.194602),
"/US/US-ID/huetter-4411" => array("type"=>"city", "pop"=>97, "name_ru"=>"Хаеттер", "name_en"=>"Huetter", "lang"=>"en", "lat"=>47.703855, "lon"=>-116.847849),
"/US/US-ID/idaho-city-4412" => array("type"=>"city", "pop"=>476, "name_ru"=>"Айдахо Сити", "name_en"=>"Idaho City", "lang"=>"en", "lat"=>43.842468, "lon"=>-115.851068),
"/US/US-ID/idaho-falls-4413" => array("type"=>"city", "pop"=>54334, "name_ru"=>"Айдахо Фоллс", "name_en"=>"Idaho Falls", "lang"=>"en", "lat"=>43.492938, "lon"=>-112.034953),
"/US/US-ID/inkom-4414" => array("type"=>"city", "pop"=>725, "name_ru"=>"Инком", "name_en"=>"Inkom", "lang"=>"en", "lat"=>42.796915, "lon"=>-112.247848),
"/US/US-ID/iona-4415" => array("type"=>"city", "pop"=>1331, "name_ru"=>"Айона", "name_en"=>"Iona", "lang"=>"en", "lat"=>43.528544, "lon"=>-111.927669),
"/US/US-ID/irwin-4416" => array("type"=>"city", "pop"=>165, "name_ru"=>"Ирвин", "name_en"=>"Irwin", "lang"=>"en", "lat"=>43.397541, "lon"=>-111.265206),
"/US/US-ID/island-park-4417" => array("type"=>"city", "pop"=>287, "name_ru"=>"Айленд Парк", "name_en"=>"Island Park", "lang"=>"en", "lat"=>44.556888, "lon"=>-111.336972),
"/US/US-ID/jerome-4418" => array("type"=>"city", "pop"=>9157, "name_ru"=>"Джером", "name_en"=>"Jerome", "lang"=>"en", "lat"=>42.725664, "lon"=>-114.515744),
"/US/US-ID/juliaetta-4419" => array("type"=>"city", "pop"=>577, "name_ru"=>"Джулиаетта", "name_en"=>"Juliaetta", "lang"=>"en", "lat"=>46.574608, "lon"=>-116.707855),
"/US/US-ID/kamiah-4420" => array("type"=>"city", "pop"=>1088, "name_ru"=>"Камиа", "name_en"=>"Kamiah", "lang"=>"en", "lat"=>46.226920, "lon"=>-116.027686),
"/US/US-ID/kellogg-4421" => array("type"=>"city", "pop"=>2228, "name_ru"=>"Келлогг", "name_en"=>"Kellogg", "lang"=>"en", "lat"=>47.539262, "lon"=>-116.135254),
"/US/US-ID/kendrick-4422" => array("type"=>"city", "pop"=>347, "name_ru"=>"Кендрик", "name_en"=>"Kendrick", "lang"=>"en", "lat"=>46.613898, "lon"=>-116.652131),
"/US/US-ID/ketchum-4423" => array("type"=>"city", "pop"=>3272, "name_ru"=>"Кетчум", "name_en"=>"Ketchum", "lang"=>"en", "lat"=>43.688815, "lon"=>-114.375170),
"/US/US-ID/kimberly-4424" => array("type"=>"city", "pop"=>3148, "name_ru"=>"Кимберли", "name_en"=>"Kimberly", "lang"=>"en", "lat"=>42.534216, "lon"=>-114.364547),
"/US/US-ID/kooskia-4425" => array("type"=>"city", "pop"=>655, "name_ru"=>"Куския", "name_en"=>"Kooskia", "lang"=>"en", "lat"=>46.139277, "lon"=>-115.972000),
"/US/US-ID/kootenai-4426" => array("type"=>"city", "pop"=>498, "name_ru"=>"Кутенай", "name_en"=>"Kootenai", "lang"=>"en", "lat"=>48.310714, "lon"=>-116.517754),
"/US/US-ID/kuna-4427" => array("type"=>"city", "pop"=>13354, "name_ru"=>"Куна", "name_en"=>"Kuna", "lang"=>"en", "lat"=>43.490141, "lon"=>-116.417452),
"/US/US-ID/lapwai-4428" => array("type"=>"city", "pop"=>1120, "name_ru"=>"Лапвай", "name_en"=>"Lapwai", "lang"=>"en", "lat"=>46.404725, "lon"=>-116.803387),
"/US/US-ID/lava-hot-springs-4429" => array("type"=>"city", "pop"=>492, "name_ru"=>"Лава Хот Спрингс", "name_en"=>"Lava Hot Springs", "lang"=>"en", "lat"=>42.620256, "lon"=>-112.009584),
"/US/US-ID/leadore-4430" => array("type"=>"city", "pop"=>86, "name_ru"=>"Лидор", "name_en"=>"Leadore", "lang"=>"en", "lat"=>44.679748, "lon"=>-113.356652),
"/US/US-ID/lewiston-4431" => array("type"=>"city", "pop"=>31764, "name_ru"=>"Льюистон", "name_en"=>"Lewiston", "lang"=>"en", "lat"=>46.392909, "lon"=>-116.992215),
"/US/US-ID/lewisville-4432" => array("type"=>"city", "pop"=>525, "name_ru"=>"Льюисвилл", "name_en"=>"Lewisville", "lang"=>"en", "lat"=>43.694917, "lon"=>-112.013971),
"/US/US-ID/lost-river-4433" => array("type"=>"city", "pop"=>29, "name_ru"=>"Лост Ривер", "name_en"=>"Lost River", "lang"=>"en", "lat"=>43.721884, "lon"=>-113.544078),
"/US/US-ID/mackay-4436" => array("type"=>"city", "pop"=>548, "name_ru"=>"Маккей", "name_en"=>"Mackay", "lang"=>"en", "lat"=>43.911974, "lon"=>-113.610851),
"/US/US-ID/malad-city-4437" => array("type"=>"city", "pop"=>2084, "name_ru"=>"Мэлад Сити", "name_en"=>"Malad City", "lang"=>"en", "lat"=>42.189549, "lon"=>-112.249070),
"/US/US-ID/malta-4438" => array("type"=>"city", "pop"=>173, "name_ru"=>"Малта", "name_en"=>"Malta", "lang"=>"en", "lat"=>42.307597, "lon"=>-113.369693),
"/US/US-ID/marsing-4439" => array("type"=>"city", "pop"=>945, "name_ru"=>"Марсинг", "name_en"=>"Marsing", "lang"=>"en", "lat"=>43.545604, "lon"=>-116.806838),
"/US/US-ID/mccall-4434" => array("type"=>"city", "pop"=>2593, "name_ru"=>"Мак-Колл", "name_en"=>"McCall", "lang"=>"en", "lat"=>44.913402, "lon"=>-116.107510),
"/US/US-ID/mccammon-4435" => array("type"=>"city", "pop"=>821, "name_ru"=>"Мак-Кэммон", "name_en"=>"McCammon", "lang"=>"en", "lat"=>42.647022, "lon"=>-112.192315),
"/US/US-ID/melba-4440" => array("type"=>"city", "pop"=>560, "name_ru"=>"Мэлба", "name_en"=>"Melba", "lang"=>"en", "lat"=>43.373923, "lon"=>-116.529521),
"/US/US-ID/menan-4441" => array("type"=>"city", "pop"=>716, "name_ru"=>"Менан", "name_en"=>"Menan", "lang"=>"en", "lat"=>43.722049, "lon"=>-111.994114),
"/US/US-ID/meridian-4442" => array("type"=>"city", "pop"=>66916, "name_ru"=>"Меридиан", "name_en"=>"Meridian", "lang"=>"en", "lat"=>43.611129, "lon"=>-116.399676),
"/US/US-ID/middleton-4443" => array("type"=>"city", "pop"=>5594, "name_ru"=>"Мидлтон", "name_en"=>"Middleton", "lang"=>"en", "lat"=>43.707661, "lon"=>-116.623975),
"/US/US-ID/midvale-4444" => array("type"=>"city", "pop"=>209, "name_ru"=>"Мидвейл", "name_en"=>"Midvale", "lang"=>"en", "lat"=>44.470639, "lon"=>-116.733069),
"/US/US-ID/minidoka-4445" => array("type"=>"city", "pop"=>118, "name_ru"=>"Минидока", "name_en"=>"Minidoka", "lang"=>"en", "lat"=>42.753953, "lon"=>-113.489044),
"/US/US-ID/montpelier-4446" => array("type"=>"city", "pop"=>2356, "name_ru"=>"Монтпильер", "name_en"=>"Montpelier", "lang"=>"en", "lat"=>42.323086, "lon"=>-111.300940),
"/US/US-ID/moore-4447" => array("type"=>"city", "pop"=>185, "name_ru"=>"Мур", "name_en"=>"Moore", "lang"=>"en", "lat"=>43.734115, "lon"=>-113.366056),
"/US/US-ID/moscow-4448" => array("type"=>"city", "pop"=>22798, "name_ru"=>"Москва", "name_en"=>"Moscow", "lang"=>"en", "lat"=>46.729767, "lon"=>-116.996844),
"/US/US-ID/mountain-home-4449" => array("type"=>"city", "pop"=>12382, "name_ru"=>"Маунтин Хоум", "name_en"=>"Mountain Home", "lang"=>"en", "lat"=>43.138336, "lon"=>-115.688687),
"/US/US-ID/mountain-home-afb-4450" => array("type"=>"cdp", "pop"=>5936, "name_ru"=>"Военно-воздушная база Маунтин Хоум", "name_en"=>"Mountain Home AFB", "lang"=>"en", "lat"=>43.049618, "lon"=>-115.865528),
"/US/US-ID/moyie-springs-4451" => array("type"=>"city", "pop"=>727, "name_ru"=>"Мойи Спрингс", "name_en"=>"Moyie Springs", "lang"=>"en", "lat"=>48.724772, "lon"=>-116.192522),
"/US/US-ID/mud-lake-4452" => array("type"=>"city", "pop"=>286, "name_ru"=>"Мад Лейк", "name_en"=>"Mud Lake", "lang"=>"en", "lat"=>43.842647, "lon"=>-112.479066),
"/US/US-ID/mullan-4453" => array("type"=>"city", "pop"=>748, "name_ru"=>"Маллан", "name_en"=>"Mullan", "lang"=>"en", "lat"=>47.469146, "lon"=>-115.794037),
"/US/US-ID/murtaugh-4454" => array("type"=>"city", "pop"=>149, "name_ru"=>"Мерто", "name_en"=>"Murtaugh", "lang"=>"en", "lat"=>42.491908, "lon"=>-114.160213),
"/US/US-ID/nampa-4455" => array("type"=>"city", "pop"=>80362, "name_ru"=>"Нампа", "name_en"=>"Nampa", "lang"=>"en", "lat"=>43.579968, "lon"=>-116.563964),
"/US/US-ID/new-meadows-4457" => array("type"=>"city", "pop"=>491, "name_ru"=>"Нью Медоус", "name_en"=>"New Meadows", "lang"=>"en", "lat"=>44.970747, "lon"=>-116.284324),
"/US/US-ID/new-plymouth-4458" => array("type"=>"city", "pop"=>1448, "name_ru"=>"Нью Плимут", "name_en"=>"New Plymouth", "lang"=>"en", "lat"=>43.970792, "lon"=>-116.818032),
"/US/US-ID/newdale-4456" => array("type"=>"city", "pop"=>349, "name_ru"=>"Ньюдейл", "name_en"=>"Newdale", "lang"=>"en", "lat"=>43.886074, "lon"=>-111.604120),
"/US/US-ID/nezperce-4459" => array("type"=>"city", "pop"=>486, "name_ru"=>"Несперс", "name_en"=>"Nezperce", "lang"=>"en", "lat"=>46.233590, "lon"=>-116.238768),
"/US/US-ID/notus-4460" => array("type"=>"city", "pop"=>623, "name_ru"=>"Нотус", "name_en"=>"Notus", "lang"=>"en", "lat"=>43.725664, "lon"=>-116.799831),
"/US/US-ID/oakley-4461" => array("type"=>"city", "pop"=>707, "name_ru"=>"Оукли", "name_en"=>"Oakley", "lang"=>"en", "lat"=>42.242449, "lon"=>-113.880056),
"/US/US-ID/oldtown-4462" => array("type"=>"city", "pop"=>207, "name_ru"=>"Оулдтаун", "name_en"=>"Oldtown", "lang"=>"en", "lat"=>48.182360, "lon"=>-117.033956),
"/US/US-ID/onaway-4463" => array("type"=>"city", "pop"=>217, "name_ru"=>"Онуэй", "name_en"=>"Onaway", "lang"=>"en", "lat"=>46.928672, "lon"=>-116.888077),
"/US/US-ID/orofino-4464" => array("type"=>"city", "pop"=>3045, "name_ru"=>"Орофино", "name_en"=>"Orofino", "lang"=>"en", "lat"=>46.488489, "lon"=>-116.257808),
"/US/US-ID/osburn-4465" => array("type"=>"city", "pop"=>1389, "name_ru"=>"Осбурн", "name_en"=>"Osburn", "lang"=>"en", "lat"=>47.505870, "lon"=>-115.999613),
"/US/US-ID/oxford-4466" => array("type"=>"city", "pop"=>52, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>42.259818, "lon"=>-112.016565),
"/US/US-ID/paris-4467" => array("type"=>"city", "pop"=>483, "name_ru"=>"Парис", "name_en"=>"Paris", "lang"=>"en", "lat"=>42.227773, "lon"=>-111.401527),
"/US/US-ID/parker-4468" => array("type"=>"city", "pop"=>315, "name_ru"=>"Паркер", "name_en"=>"Parker", "lang"=>"en", "lat"=>43.958566, "lon"=>-111.758406),
"/US/US-ID/parma-4469" => array("type"=>"city", "pop"=>1870, "name_ru"=>"Парма", "name_en"=>"Parma", "lang"=>"en", "lat"=>43.785554, "lon"=>-116.940532),
"/US/US-ID/paul-4470" => array("type"=>"city", "pop"=>919, "name_ru"=>"Рол", "name_en"=>"Paul", "lang"=>"en", "lat"=>42.606440, "lon"=>-113.782385),
"/US/US-ID/payette-4471" => array("type"=>"city", "pop"=>7637, "name_ru"=>"Пейетт", "name_en"=>"Payette", "lang"=>"en", "lat"=>44.076958, "lon"=>-116.928950),
"/US/US-ID/peck-4472" => array("type"=>"city", "pop"=>184, "name_ru"=>"Пек", "name_en"=>"Peck", "lang"=>"en", "lat"=>46.473873, "lon"=>-116.424094),
"/US/US-ID/pierce-4473" => array("type"=>"city", "pop"=>527, "name_ru"=>"Пирс", "name_en"=>"Pierce", "lang"=>"en", "lat"=>46.492657, "lon"=>-115.798492),
"/US/US-ID/pinehurst-4474" => array("type"=>"city", "pop"=>1589, "name_ru"=>"Пайнхерст", "name_en"=>"Pinehurst", "lang"=>"en", "lat"=>47.535610, "lon"=>-116.233426),
"/US/US-ID/placerville-4475" => array("type"=>"city", "pop"=>57, "name_ru"=>"Пласервилл", "name_en"=>"Placerville", "lang"=>"en", "lat"=>43.942951, "lon"=>-115.944812),
"/US/US-ID/plummer-4476" => array("type"=>"city", "pop"=>997, "name_ru"=>"Пламмер", "name_en"=>"Plummer", "lang"=>"en", "lat"=>47.329698, "lon"=>-116.884071),
"/US/US-ID/pocatello-4477" => array("type"=>"city", "pop"=>54901, "name_ru"=>"Покателло", "name_en"=>"Pocatello", "lang"=>"en", "lat"=>42.876211, "lon"=>-112.468515),
"/US/US-ID/ponderay-4478" => array("type"=>"city", "pop"=>696, "name_ru"=>"Пондерей", "name_en"=>"Ponderay", "lang"=>"en", "lat"=>48.301218, "lon"=>-116.539497),
"/US/US-ID/post-falls-4479" => array("type"=>"city", "pop"=>26460, "name_ru"=>"Пост Фоллс", "name_en"=>"Post Falls", "lang"=>"en", "lat"=>47.712486, "lon"=>-116.944627),
"/US/US-ID/potlatch-4480" => array("type"=>"city", "pop"=>742, "name_ru"=>"Потлэтч", "name_en"=>"Potlatch", "lang"=>"en", "lat"=>46.923190, "lon"=>-116.897857),
"/US/US-ID/preston-4481" => array("type"=>"city", "pop"=>5059, "name_ru"=>"Престон", "name_en"=>"Preston", "lang"=>"en", "lat"=>42.097317, "lon"=>-111.874547),
"/US/US-ID/priest-river-4482" => array("type"=>"city", "pop"=>1913, "name_ru"=>"Прист Ривер", "name_en"=>"Priest River", "lang"=>"en", "lat"=>48.185137, "lon"=>-116.910110),
"/US/US-ID/rathdrum-4483" => array("type"=>"city", "pop"=>6821, "name_ru"=>"Рэтдрам", "name_en"=>"Rathdrum", "lang"=>"en", "lat"=>47.811661, "lon"=>-116.888723),
"/US/US-ID/reubens-4484" => array("type"=>"city", "pop"=>69, "name_ru"=>"Реюбенс", "name_en"=>"Reubens", "lang"=>"en", "lat"=>46.322219, "lon"=>-116.541110),
"/US/US-ID/rexburg-4485" => array("type"=>"city", "pop"=>28028, "name_ru"=>"Рексбург", "name_en"=>"Rexburg", "lang"=>"en", "lat"=>43.825173, "lon"=>-111.787419),
"/US/US-ID/richfield-4486" => array("type"=>"city", "pop"=>414, "name_ru"=>"Ричфилд", "name_en"=>"Richfield", "lang"=>"en", "lat"=>43.052283, "lon"=>-114.155095),
"/US/US-ID/rigby-4487" => array("type"=>"city", "pop"=>3394, "name_ru"=>"Ригбай", "name_en"=>"Rigby", "lang"=>"en", "lat"=>43.673931, "lon"=>-111.915509),
"/US/US-ID/riggins-4488" => array("type"=>"city", "pop"=>418, "name_ru"=>"Риггинс", "name_en"=>"Riggins", "lang"=>"en", "lat"=>45.421981, "lon"=>-116.314993),
"/US/US-ID/ririe-4489" => array("type"=>"city", "pop"=>534, "name_ru"=>"Рири", "name_en"=>"Ririe", "lang"=>"en", "lat"=>43.631344, "lon"=>-111.773434),
"/US/US-ID/roberts-4490" => array("type"=>"city", "pop"=>642, "name_ru"=>"Робертс", "name_en"=>"Roberts", "lang"=>"en", "lat"=>43.720717, "lon"=>-112.126837),
"/US/US-ID/rockland-4491" => array("type"=>"city", "pop"=>313, "name_ru"=>"Рокленд", "name_en"=>"Rockland", "lang"=>"en", "lat"=>42.573556, "lon"=>-112.874196),
"/US/US-ID/rupert-4492" => array("type"=>"city", "pop"=>5082, "name_ru"=>"Руперт", "name_en"=>"Rupert", "lang"=>"en", "lat"=>42.618059, "lon"=>-113.673698),
"/US/US-ID/salmon-4496" => array("type"=>"city", "pop"=>2986, "name_ru"=>"Сэлмон", "name_en"=>"Salmon", "lang"=>"en", "lat"=>45.177862, "lon"=>-113.898524),
"/US/US-ID/sandpoint-4497" => array("type"=>"city", "pop"=>8337, "name_ru"=>"Сэндпойнт", "name_en"=>"Sandpoint", "lang"=>"en", "lat"=>48.279862, "lon"=>-116.558117),
"/US/US-ID/shelley-4498" => array("type"=>"city", "pop"=>4257, "name_ru"=>"Шелли", "name_en"=>"Shelley", "lang"=>"en", "lat"=>43.380174, "lon"=>-112.123101),
"/US/US-ID/shoshone-4499" => array("type"=>"city", "pop"=>1564, "name_ru"=>"Шошони", "name_en"=>"Shoshone", "lang"=>"en", "lat"=>42.936174, "lon"=>-114.404628),
"/US/US-ID/smelterville-4500" => array("type"=>"city", "pop"=>598, "name_ru"=>"Смелтервилл", "name_en"=>"Smelterville", "lang"=>"en", "lat"=>47.542428, "lon"=>-116.176473),
"/US/US-ID/soda-springs-4501" => array("type"=>"city", "pop"=>3072, "name_ru"=>"Сода Спрингс", "name_en"=>"Soda Springs", "lang"=>"en", "lat"=>42.658587, "lon"=>-111.585514),
"/US/US-ID/spencer-4502" => array("type"=>"city", "pop"=>33, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>44.380830, "lon"=>-112.186378),
"/US/US-ID/spirit-lake-4503" => array("type"=>"city", "pop"=>1730, "name_ru"=>"Спирит Лейк", "name_en"=>"Spirit Lake", "lang"=>"en", "lat"=>47.967955, "lon"=>-116.869716),
"/US/US-ID/st.-anthony-4493" => array("type"=>"city", "pop"=>3401, "name_ru"=>"Сент Энтони", "name_en"=>"St. Anthony", "lang"=>"en", "lat"=>43.966168, "lon"=>-111.683685),
"/US/US-ID/st.-charles-4494" => array("type"=>"city", "pop"=>128, "name_ru"=>"Сент-Чарльз", "name_en"=>"St. Charles", "lang"=>"en", "lat"=>42.112561, "lon"=>-111.389213),
"/US/US-ID/st.-maries-4495" => array("type"=>"city", "pop"=>2644, "name_ru"=>"Сент Мэрис", "name_en"=>"St. Maries", "lang"=>"en", "lat"=>47.314917, "lon"=>-116.570640),
"/US/US-ID/stanley-4504" => array("type"=>"city", "pop"=>106, "name_ru"=>"Стэнли", "name_en"=>"Stanley", "lang"=>"en", "lat"=>44.215787, "lon"=>-114.936977),
"/US/US-ID/state-line-4505" => array("type"=>"city", "pop"=>60, "name_ru"=>"Стейт Лейн", "name_en"=>"State Line", "lang"=>"en", "lat"=>47.706154, "lon"=>-117.035122),
"/US/US-ID/stites-4506" => array("type"=>"city", "pop"=>225, "name_ru"=>"Стайтс", "name_en"=>"Stites", "lang"=>"en", "lat"=>46.092399, "lon"=>-115.975705),
"/US/US-ID/sugar-city-4507" => array("type"=>"city", "pop"=>1540, "name_ru"=>"Шугар Сити", "name_en"=>"Sugar City", "lang"=>"en", "lat"=>43.872406, "lon"=>-111.746534),
"/US/US-ID/sun-valley-4508" => array("type"=>"city", "pop"=>1466, "name_ru"=>"Сан Вэлли", "name_en"=>"Sun Valley", "lang"=>"en", "lat"=>43.681471, "lon"=>-114.328415),
"/US/US-ID/swan-valley-4509" => array("type"=>"city", "pop"=>244, "name_ru"=>"Суон Валли", "name_en"=>"Swan Valley", "lang"=>"en", "lat"=>43.439767, "lon"=>-111.315049),
"/US/US-ID/tensed-4510" => array("type"=>"city", "pop"=>125, "name_ru"=>"Тенст", "name_en"=>"Tensed", "lang"=>"en", "lat"=>47.159625, "lon"=>-116.923381),
"/US/US-ID/teton-4511" => array("type"=>"city", "pop"=>661, "name_ru"=>"Тетон", "name_en"=>"Teton", "lang"=>"en", "lat"=>43.887752, "lon"=>-111.668495),
"/US/US-ID/tetonia-4512" => array("type"=>"city", "pop"=>229, "name_ru"=>"Тетония", "name_en"=>"Tetonia", "lang"=>"en", "lat"=>43.815933, "lon"=>-111.158846),
"/US/US-ID/troy-4513" => array("type"=>"city", "pop"=>777, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>46.738694, "lon"=>-116.768048),
"/US/US-ID/twin-falls-4514" => array("type"=>"city", "pop"=>42197, "name_ru"=>"Твин Фоллс", "name_en"=>"Twin Falls", "lang"=>"en", "lat"=>42.561509, "lon"=>-114.462844),
"/US/US-ID/ucon-4515" => array("type"=>"city", "pop"=>1082, "name_ru"=>"Юкон", "name_en"=>"Ucon", "lang"=>"en", "lat"=>43.594155, "lon"=>-111.955432),
"/US/US-ID/victor-4516" => array("type"=>"city", "pop"=>1817, "name_ru"=>"Виктор", "name_en"=>"Victor", "lang"=>"en", "lat"=>43.603750, "lon"=>-111.113290),
"/US/US-ID/wallace-4517" => array("type"=>"city", "pop"=>861, "name_ru"=>"Уоллэс", "name_en"=>"Wallace", "lang"=>"en", "lat"=>47.472930, "lon"=>-115.922264),
"/US/US-ID/wardner-4518" => array("type"=>"city", "pop"=>197, "name_ru"=>"Уордер", "name_en"=>"Wardner", "lang"=>"en", "lat"=>47.520328, "lon"=>-116.133990),
"/US/US-ID/warm-river-4519" => array("type"=>"city", "pop"=>10, "name_ru"=>"Уорм Ривер", "name_en"=>"Warm River", "lang"=>"en", "lat"=>44.119608, "lon"=>-111.322711),
"/US/US-ID/weippe-4520" => array("type"=>"city", "pop"=>370, "name_ru"=>"Уэипп", "name_en"=>"Weippe", "lang"=>"en", "lat"=>46.378316, "lon"=>-115.938853),
"/US/US-ID/weiser-4521" => array("type"=>"city", "pop"=>5290, "name_ru"=>"Уайсер", "name_en"=>"Weiser", "lang"=>"en", "lat"=>44.250834, "lon"=>-116.966745),
"/US/US-ID/wendell-4522" => array("type"=>"city", "pop"=>2418, "name_ru"=>"Уэнделл", "name_en"=>"Wendell", "lang"=>"en", "lat"=>42.776071, "lon"=>-114.701816),
"/US/US-ID/weston-4523" => array("type"=>"city", "pop"=>437, "name_ru"=>"Вэстон", "name_en"=>"Weston", "lang"=>"en", "lat"=>42.038365, "lon"=>-111.977075),
"/US/US-ID/white-bird-4524" => array("type"=>"city", "pop"=>105, "name_ru"=>"Уайт Берд", "name_en"=>"White Bird", "lang"=>"en", "lat"=>45.762920, "lon"=>-116.298939),
"/US/US-ID/wilder-4525" => array("type"=>"city", "pop"=>1473, "name_ru"=>"Уайлдер", "name_en"=>"Wilder", "lang"=>"en", "lat"=>43.676570, "lon"=>-116.907833),
"/US/US-ID/winchester-4526" => array("type"=>"city", "pop"=>291, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>46.241051, "lon"=>-116.622418),
"/US/US-ID/worley-4527" => array("type"=>"city", "pop"=>219, "name_ru"=>"Уорли", "name_en"=>"Worley", "lang"=>"en", "lat"=>47.401335, "lon"=>-116.918404),
"/US/US-IL" => array("type"=>"state", "pop"=>12901563, "name_ru"=>"Иллинойс", "name_en"=>"Illinois", "lang"=>"en"),
"/US/US-IL/abingdon-4528" => array("type"=>"city", "pop"=>3269, "name_ru"=>"Абингдон", "name_en"=>"Abingdon", "lang"=>"en", "lat"=>40.802596, "lon"=>-90.400658),
"/US/US-IL/addieville-4529" => array("type"=>"village", "pop"=>252, "name_ru"=>"Аддивилл", "name_en"=>"Addieville", "lang"=>"en", "lat"=>38.391725, "lon"=>-89.487445),
"/US/US-IL/addison-4530" => array("type"=>"village", "pop"=>36917, "name_ru"=>"Эддиссон", "name_en"=>"Addison", "lang"=>"en", "lat"=>41.930567, "lon"=>-88.008623),
"/US/US-IL/adeline-4531" => array("type"=>"village", "pop"=>141, "name_ru"=>"Аделин", "name_en"=>"Adeline", "lang"=>"en", "lat"=>42.140817, "lon"=>-89.490471),
"/US/US-IL/albany-4532" => array("type"=>"village", "pop"=>913, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>41.785631, "lon"=>-90.219456),
"/US/US-IL/albers-4533" => array("type"=>"village", "pop"=>1085, "name_ru"=>"Алберс", "name_en"=>"Albers", "lang"=>"en", "lat"=>38.544809, "lon"=>-89.615089),
"/US/US-IL/albion-4534" => array("type"=>"city", "pop"=>1814, "name_ru"=>"Албион", "name_en"=>"Albion", "lang"=>"en", "lat"=>38.376182, "lon"=>-88.057199),
"/US/US-IL/aledo-4535" => array("type"=>"city", "pop"=>3565, "name_ru"=>"Аледо", "name_en"=>"Aledo", "lang"=>"en", "lat"=>41.198287, "lon"=>-90.747442),
"/US/US-IL/alexis-4536" => array("type"=>"village", "pop"=>797, "name_ru"=>"Алексис", "name_en"=>"Alexis", "lang"=>"en", "lat"=>41.063397, "lon"=>-90.554817),
"/US/US-IL/algonquin-4537" => array("type"=>"village", "pop"=>30352, "name_ru"=>"Алгонкин", "name_en"=>"Algonquin", "lang"=>"en", "lat"=>42.164233, "lon"=>-88.300464),
"/US/US-IL/alhambra-4538" => array("type"=>"village", "pop"=>642, "name_ru"=>"Альгамбра", "name_en"=>"Alhambra", "lang"=>"en", "lat"=>38.887553, "lon"=>-89.734998),
"/US/US-IL/allendale-4539" => array("type"=>"village", "pop"=>482, "name_ru"=>"Аллендейл", "name_en"=>"Allendale", "lang"=>"en", "lat"=>38.527573, "lon"=>-87.710291),
"/US/US-IL/allenville-4540" => array("type"=>"village", "pop"=>146, "name_ru"=>"Алленвилл", "name_en"=>"Allenville", "lang"=>"en", "lat"=>39.557945, "lon"=>-88.538346),
"/US/US-IL/allerton-4541" => array("type"=>"village", "pop"=>275, "name_ru"=>"Аллертон", "name_en"=>"Allerton", "lang"=>"en", "lat"=>39.915274, "lon"=>-87.935627),
"/US/US-IL/alma-4542" => array("type"=>"village", "pop"=>370, "name_ru"=>"Алма", "name_en"=>"Alma", "lang"=>"en", "lat"=>38.722948, "lon"=>-88.910130),
"/US/US-IL/alorton-4543" => array("type"=>"village", "pop"=>2526, "name_ru"=>"Алортон", "name_en"=>"Alorton", "lang"=>"en", "lat"=>38.585600, "lon"=>-90.115086),
"/US/US-IL/alpha-4544" => array("type"=>"village", "pop"=>683, "name_ru"=>"Альфа", "name_en"=>"Alpha", "lang"=>"en", "lat"=>41.192167, "lon"=>-90.380816),
"/US/US-IL/alsey-4545" => array("type"=>"village", "pop"=>227, "name_ru"=>"Алси", "name_en"=>"Alsey", "lang"=>"en", "lat"=>39.559349, "lon"=>-90.433526),
"/US/US-IL/alsip-4546" => array("type"=>"village", "pop"=>18692, "name_ru"=>"Алсип", "name_en"=>"Alsip", "lang"=>"en", "lat"=>41.670350, "lon"=>-87.735775),
"/US/US-IL/altamont-4547" => array("type"=>"city", "pop"=>2254, "name_ru"=>"Альтамонт", "name_en"=>"Altamont", "lang"=>"en", "lat"=>39.057980, "lon"=>-88.748511),
"/US/US-IL/alto-pass-4550" => array("type"=>"village", "pop"=>382, "name_ru"=>"Альто Пасс", "name_en"=>"Alto Pass", "lang"=>"en", "lat"=>37.572869, "lon"=>-89.319259),
"/US/US-IL/alton-4548" => array("type"=>"city", "pop"=>29393, "name_ru"=>"Альтон", "name_en"=>"Alton", "lang"=>"en", "lat"=>38.903967, "lon"=>-90.154017),
"/US/US-IL/altona-4549" => array("type"=>"village", "pop"=>525, "name_ru"=>"Альтона", "name_en"=>"Altona", "lang"=>"en", "lat"=>41.115206, "lon"=>-90.164686),
"/US/US-IL/alvin-4551" => array("type"=>"village", "pop"=>339, "name_ru"=>"Алвин", "name_en"=>"Alvin", "lang"=>"en", "lat"=>40.307685, "lon"=>-87.606812),
"/US/US-IL/amboy-4552" => array("type"=>"city", "pop"=>2574, "name_ru"=>"Амбой", "name_en"=>"Amboy", "lang"=>"en", "lat"=>41.716603, "lon"=>-89.333549),
"/US/US-IL/anchor-4553" => array("type"=>"village", "pop"=>171, "name_ru"=>"Анкор", "name_en"=>"Anchor", "lang"=>"en", "lat"=>40.568704, "lon"=>-88.538732),
"/US/US-IL/andalusia-4554" => array("type"=>"village", "pop"=>1067, "name_ru"=>"Андалусия", "name_en"=>"Andalusia", "lang"=>"en", "lat"=>41.438033, "lon"=>-90.725428),
"/US/US-IL/andover-4555" => array("type"=>"village", "pop"=>566, "name_ru"=>"Андоувер", "name_en"=>"Andover", "lang"=>"en", "lat"=>41.294976, "lon"=>-90.290535),
"/US/US-IL/anna-4556" => array("type"=>"city", "pop"=>5025, "name_ru"=>"Анна", "name_en"=>"Anna", "lang"=>"en", "lat"=>37.459972, "lon"=>-89.241147),
"/US/US-IL/annawan-4557" => array("type"=>"town", "pop"=>904, "name_ru"=>"Аннауан", "name_en"=>"Annawan", "lang"=>"en", "lat"=>41.398681, "lon"=>-89.907036),
"/US/US-IL/antioch-4558" => array("type"=>"village", "pop"=>13892, "name_ru"=>"Антиок", "name_en"=>"Antioch", "lang"=>"en", "lat"=>42.474731, "lon"=>-88.079536),
"/US/US-IL/apple-river-4559" => array("type"=>"village", "pop"=>359, "name_ru"=>"Эппл Ривер", "name_en"=>"Apple River", "lang"=>"en", "lat"=>42.501937, "lon"=>-90.094675),
"/US/US-IL/arcola-4560" => array("type"=>"city", "pop"=>2783, "name_ru"=>"Аркола", "name_en"=>"Arcola", "lang"=>"en", "lat"=>39.684612, "lon"=>-88.303771),
"/US/US-IL/arenzville-4561" => array("type"=>"village", "pop"=>401, "name_ru"=>"Арензвилл", "name_en"=>"Arenzville", "lang"=>"en", "lat"=>39.879078, "lon"=>-90.372024),
"/US/US-IL/argenta-4562" => array("type"=>"village", "pop"=>824, "name_ru"=>"Аргента", "name_en"=>"Argenta", "lang"=>"en", "lat"=>39.984984, "lon"=>-88.818681),
"/US/US-IL/arlington-4563" => array("type"=>"village", "pop"=>203, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>41.471174, "lon"=>-89.248228),
"/US/US-IL/arlington-heights-4564" => array("type"=>"village", "pop"=>73399, "name_ru"=>"Арлингтон Хайтс", "name_en"=>"Arlington Heights", "lang"=>"en", "lat"=>42.094950, "lon"=>-87.982024),
"/US/US-IL/armington-4565" => array("type"=>"village", "pop"=>349, "name_ru"=>"Армингтон", "name_en"=>"Armington", "lang"=>"en", "lat"=>40.339931, "lon"=>-89.313673),
"/US/US-IL/aroma-park-4566" => array("type"=>"village", "pop"=>1035, "name_ru"=>"Арома Парк", "name_en"=>"Aroma Park", "lang"=>"en", "lat"=>41.079767, "lon"=>-87.805449),
"/US/US-IL/arrowsmith-4567" => array("type"=>"village", "pop"=>288, "name_ru"=>"Арроусмит", "name_en"=>"Arrowsmith", "lang"=>"en", "lat"=>40.448863, "lon"=>-88.631478),
"/US/US-IL/arthur-4568" => array("type"=>"village", "pop"=>2129, "name_ru"=>"Артур", "name_en"=>"Arthur", "lang"=>"en", "lat"=>39.715291, "lon"=>-88.469038),
"/US/US-IL/ashkum-4569" => array("type"=>"village", "pop"=>693, "name_ru"=>"Ашкум", "name_en"=>"Ashkum", "lang"=>"en", "lat"=>40.878383, "lon"=>-87.953589),
"/US/US-IL/ashland-4570" => array("type"=>"village", "pop"=>1321, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>39.888694, "lon"=>-90.008661),
"/US/US-IL/ashley-4571" => array("type"=>"city", "pop"=>573, "name_ru"=>"Эшли", "name_en"=>"Ashley", "lang"=>"en", "lat"=>38.329052, "lon"=>-89.189394),
"/US/US-IL/ashmore-4572" => array("type"=>"village", "pop"=>755, "name_ru"=>"Ашмор", "name_en"=>"Ashmore", "lang"=>"en", "lat"=>39.531100, "lon"=>-88.020416),
"/US/US-IL/ashton-4573" => array("type"=>"village", "pop"=>1166, "name_ru"=>"Аштон", "name_en"=>"Ashton", "lang"=>"en", "lat"=>41.868127, "lon"=>-89.221499),
"/US/US-IL/assumption-4574" => array("type"=>"city", "pop"=>1196, "name_ru"=>"Ассумпшен", "name_en"=>"Assumption", "lang"=>"en", "lat"=>39.517961, "lon"=>-89.048435),
"/US/US-IL/astoria-4575" => array("type"=>"village", "pop"=>1124, "name_ru"=>"Астория", "name_en"=>"Astoria", "lang"=>"en", "lat"=>40.228322, "lon"=>-90.355894),
"/US/US-IL/athens-4576" => array("type"=>"city", "pop"=>1780, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>39.960906, "lon"=>-89.724849),
"/US/US-IL/atkinson-4577" => array("type"=>"town", "pop"=>955, "name_ru"=>"Аткинсон", "name_en"=>"Atkinson", "lang"=>"en", "lat"=>41.418736, "lon"=>-90.014514),
"/US/US-IL/atlanta-4578" => array("type"=>"city", "pop"=>1633, "name_ru"=>"Атланта", "name_en"=>"Atlanta", "lang"=>"en", "lat"=>40.262194, "lon"=>-89.232155),
"/US/US-IL/atwood-4579" => array("type"=>"village", "pop"=>1226, "name_ru"=>"Атвуд", "name_en"=>"Atwood", "lang"=>"en", "lat"=>39.799748, "lon"=>-88.463170),
"/US/US-IL/auburn-4580" => array("type"=>"city", "pop"=>4362, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>39.577236, "lon"=>-89.745815),
"/US/US-IL/augusta-4581" => array("type"=>"village", "pop"=>607, "name_ru"=>"Огаста", "name_en"=>"Augusta", "lang"=>"en", "lat"=>40.230878, "lon"=>-90.948936),
"/US/US-IL/aurora-4582" => array("type"=>"city", "pop"=>171782, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>41.767633, "lon"=>-88.293042),
"/US/US-IL/ava-4583" => array("type"=>"city", "pop"=>642, "name_ru"=>"Ава", "name_en"=>"Ava", "lang"=>"en", "lat"=>37.888414, "lon"=>-89.496258),
"/US/US-IL/aviston-4584" => array("type"=>"village", "pop"=>1723, "name_ru"=>"Авистон", "name_en"=>"Aviston", "lang"=>"en", "lat"=>38.607791, "lon"=>-89.606232),
"/US/US-IL/avon-4585" => array("type"=>"village", "pop"=>849, "name_ru"=>"Эйвон", "name_en"=>"Avon", "lang"=>"en", "lat"=>40.662215, "lon"=>-90.435357),
"/US/US-IL/baldwin-4586" => array("type"=>"village", "pop"=>404, "name_ru"=>"Болдуин", "name_en"=>"Baldwin", "lang"=>"en", "lat"=>38.183718, "lon"=>-89.844556),
"/US/US-IL/banner-4587" => array("type"=>"village", "pop"=>142, "name_ru"=>"Бэннер", "name_en"=>"Banner", "lang"=>"en", "lat"=>40.514691, "lon"=>-89.910425),
"/US/US-IL/bannockburn-4588" => array("type"=>"village", "pop"=>1879, "name_ru"=>"Баннокберн", "name_en"=>"Bannockburn", "lang"=>"en", "lat"=>42.192772, "lon"=>-87.868388),
"/US/US-IL/bardolph-4589" => array("type"=>"village", "pop"=>229, "name_ru"=>"Бардольф", "name_en"=>"Bardolph", "lang"=>"en", "lat"=>40.496777, "lon"=>-90.563214),
"/US/US-IL/barrington-4590" => array("type"=>"village", "pop"=>10374, "name_ru"=>"Бэррингтон", "name_en"=>"Barrington", "lang"=>"en", "lat"=>42.151837, "lon"=>-88.129236),
"/US/US-IL/barrington-hills-4591" => array("type"=>"village", "pop"=>4354, "name_ru"=>"Бэррингтон Хилс", "name_en"=>"Barrington Hills", "lang"=>"en", "lat"=>42.133750, "lon"=>-88.211186),
"/US/US-IL/barry-4592" => array("type"=>"city", "pop"=>1290, "name_ru"=>"Бэрри", "name_en"=>"Barry", "lang"=>"en", "lat"=>39.695602, "lon"=>-91.039734),
"/US/US-IL/bartelso-4593" => array("type"=>"village", "pop"=>583, "name_ru"=>"Бартелсо", "name_en"=>"Bartelso", "lang"=>"en", "lat"=>38.536523, "lon"=>-89.468387),
"/US/US-IL/bartlett-4594" => array("type"=>"village", "pop"=>42209, "name_ru"=>"Барлет", "name_en"=>"Bartlett", "lang"=>"en", "lat"=>41.979004, "lon"=>-88.205354),
"/US/US-IL/bartonville-4595" => array("type"=>"village", "pop"=>6160, "name_ru"=>"Бартонвилл", "name_en"=>"Bartonville", "lang"=>"en", "lat"=>40.637584, "lon"=>-89.664158),
"/US/US-IL/basco-4596" => array("type"=>"village", "pop"=>98, "name_ru"=>"Баско", "name_en"=>"Basco", "lang"=>"en", "lat"=>40.326835, "lon"=>-91.199069),
"/US/US-IL/batavia-4597" => array("type"=>"city", "pop"=>27502, "name_ru"=>"Батавия", "name_en"=>"Batavia", "lang"=>"en", "lat"=>41.850600, "lon"=>-88.300424),
"/US/US-IL/batchtown-4598" => array("type"=>"village", "pop"=>212, "name_ru"=>"Бэтчтаун", "name_en"=>"Batchtown", "lang"=>"en", "lat"=>39.032889, "lon"=>-90.653975),
"/US/US-IL/bath-4599" => array("type"=>"village", "pop"=>293, "name_ru"=>"Бат", "name_en"=>"Bath", "lang"=>"en", "lat"=>40.191131, "lon"=>-90.142378),
"/US/US-IL/bay-view-gardens-4601" => array("type"=>"village", "pop"=>356, "name_ru"=>"Бэй Вью Гарденс", "name_en"=>"Bay View Gardens", "lang"=>"en", "lat"=>40.809789, "lon"=>-89.518999),
"/US/US-IL/baylis-4600" => array("type"=>"village", "pop"=>244, "name_ru"=>"Бэйлис", "name_en"=>"Baylis", "lang"=>"en", "lat"=>39.729013, "lon"=>-90.909235),
"/US/US-IL/beach-park-4602" => array("type"=>"village", "pop"=>13119, "name_ru"=>"Бич Парк", "name_en"=>"Beach Park", "lang"=>"en", "lat"=>42.424950, "lon"=>-87.856229),
"/US/US-IL/beardstown-4603" => array("type"=>"city", "pop"=>5859, "name_ru"=>"Бирдстаун", "name_en"=>"Beardstown", "lang"=>"en", "lat"=>40.004295, "lon"=>-90.417872),
"/US/US-IL/beaverville-4604" => array("type"=>"village", "pop"=>379, "name_ru"=>"Бивервилл", "name_en"=>"Beaverville", "lang"=>"en", "lat"=>40.953525, "lon"=>-87.655042),
"/US/US-IL/beckemeyer-4605" => array("type"=>"village", "pop"=>1087, "name_ru"=>"Бекемайер", "name_en"=>"Beckemeyer", "lang"=>"en", "lat"=>38.605139, "lon"=>-89.433959),
"/US/US-IL/bedford-park-4606" => array("type"=>"village", "pop"=>533, "name_ru"=>"Бедфорд Парк", "name_en"=>"Bedford Park", "lang"=>"en", "lat"=>41.769767, "lon"=>-87.798046),
"/US/US-IL/beecher-4607" => array("type"=>"village", "pop"=>3029, "name_ru"=>"Бичер", "name_en"=>"Beecher", "lang"=>"en", "lat"=>41.347689, "lon"=>-87.614867),
"/US/US-IL/beecher-city-4608" => array("type"=>"village", "pop"=>487, "name_ru"=>"Бичер Сити", "name_en"=>"Beecher City", "lang"=>"en", "lat"=>39.186451, "lon"=>-88.787522),
"/US/US-IL/belgium-4609" => array("type"=>"village", "pop"=>461, "name_ru"=>"Белджиум", "name_en"=>"Belgium", "lang"=>"en", "lat"=>40.061500, "lon"=>-87.630680),
"/US/US-IL/belknap-4610" => array("type"=>"village", "pop"=>138, "name_ru"=>"Белнап", "name_en"=>"Belknap", "lang"=>"en", "lat"=>37.322303, "lon"=>-88.940140),
"/US/US-IL/belle-prairie-city-4611" => array("type"=>"town", "pop"=>58, "name_ru"=>"Бель Прейри Сити", "name_en"=>"Belle Prairie City", "lang"=>"en", "lat"=>38.223145, "lon"=>-88.551483),
"/US/US-IL/belle-rive-4612" => array("type"=>"village", "pop"=>372, "name_ru"=>"Бель Рив", "name_en"=>"Belle Rive", "lang"=>"en", "lat"=>38.232014, "lon"=>-88.739881),
"/US/US-IL/belleville-4613" => array("type"=>"city", "pop"=>41097, "name_ru"=>"Белелвилл", "name_en"=>"Belleville", "lang"=>"en", "lat"=>38.529967, "lon"=>-90.003230),
"/US/US-IL/bellevue-4614" => array("type"=>"village", "pop"=>1783, "name_ru"=>"Белвью", "name_en"=>"Bellevue", "lang"=>"en", "lat"=>40.686300, "lon"=>-89.673872),
"/US/US-IL/bellflower-4615" => array("type"=>"village", "pop"=>387, "name_ru"=>"Беллфлауэр", "name_en"=>"Bellflower", "lang"=>"en", "lat"=>40.340471, "lon"=>-88.526681),
"/US/US-IL/bellmont-4616" => array("type"=>"village", "pop"=>280, "name_ru"=>"Беллмонт", "name_en"=>"Bellmont", "lang"=>"en", "lat"=>38.382646, "lon"=>-87.909481),
"/US/US-IL/bellwood-4617" => array("type"=>"village", "pop"=>18973, "name_ru"=>"Беллвуд", "name_en"=>"Bellwood", "lang"=>"en", "lat"=>41.882900, "lon"=>-87.876321),
"/US/US-IL/belvidere-4618" => array("type"=>"city", "pop"=>26461, "name_ru"=>"Белвидер", "name_en"=>"Belvidere", "lang"=>"en", "lat"=>42.252592, "lon"=>-88.848322),
"/US/US-IL/bement-4619" => array("type"=>"village", "pop"=>1698, "name_ru"=>"Бермент", "name_en"=>"Bement", "lang"=>"en", "lat"=>39.922849, "lon"=>-88.571598),
"/US/US-IL/benld-4620" => array("type"=>"city", "pop"=>1463, "name_ru"=>"Бенлд", "name_en"=>"Benld", "lang"=>"en", "lat"=>39.093101, "lon"=>-89.801876),
"/US/US-IL/bensenville-4621" => array("type"=>"village", "pop"=>20138, "name_ru"=>"Бенсенвилл", "name_en"=>"Bensenville", "lang"=>"en", "lat"=>41.957333, "lon"=>-87.944515),
"/US/US-IL/benson-4622" => array("type"=>"village", "pop"=>407, "name_ru"=>"Бенсон", "name_en"=>"Benson", "lang"=>"en", "lat"=>40.850665, "lon"=>-89.120514),
"/US/US-IL/bentley-4623" => array("type"=>"town", "pop"=>40, "name_ru"=>"Бентли", "name_en"=>"Bentley", "lang"=>"en", "lat"=>40.344707, "lon"=>-91.111061),
"/US/US-IL/benton-4624" => array("type"=>"city", "pop"=>6957, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>38.007053, "lon"=>-88.918481),
"/US/US-IL/berkeley-4625" => array("type"=>"village", "pop"=>4913, "name_ru"=>"Беркли", "name_en"=>"Berkeley", "lang"=>"en", "lat"=>41.888450, "lon"=>-87.912299),
"/US/US-IL/berlin-4626" => array("type"=>"village", "pop"=>143, "name_ru"=>"Берлин", "name_en"=>"Berlin", "lang"=>"en", "lat"=>39.757662, "lon"=>-89.902507),
"/US/US-IL/berwyn-4627" => array("type"=>"city", "pop"=>49919, "name_ru"=>"Бервин", "name_en"=>"Berwyn", "lang"=>"en", "lat"=>41.842450, "lon"=>-87.791049),
"/US/US-IL/bethalto-4628" => array("type"=>"village", "pop"=>9933, "name_ru"=>"Бетальто", "name_en"=>"Bethalto", "lang"=>"en", "lat"=>38.900800, "lon"=>-90.046899),
"/US/US-IL/bethany-4629" => array("type"=>"village", "pop"=>1226, "name_ru"=>"Бетани", "name_en"=>"Bethany", "lang"=>"en", "lat"=>39.644168, "lon"=>-88.740938),
"/US/US-IL/biggsville-4630" => array("type"=>"village", "pop"=>307, "name_ru"=>"Биггсвилл", "name_en"=>"Biggsville", "lang"=>"en", "lat"=>40.853503, "lon"=>-90.861330),
"/US/US-IL/bingham-4631" => array("type"=>"village", "pop"=>116, "name_ru"=>"Бингхэм", "name_en"=>"Bingham", "lang"=>"en", "lat"=>39.112858, "lon"=>-89.211713),
"/US/US-IL/birds-4632" => array("type"=>"village", "pop"=>49, "name_ru"=>"Бердс", "name_en"=>"Birds", "lang"=>"en", "lat"=>38.837461, "lon"=>-87.670811),
"/US/US-IL/bishop-hill-4633" => array("type"=>"village", "pop"=>116, "name_ru"=>"Бишоп Хилл", "name_en"=>"Bishop Hill", "lang"=>"en", "lat"=>41.199646, "lon"=>-90.117243),
"/US/US-IL/blandinsville-4634" => array("type"=>"village", "pop"=>694, "name_ru"=>"Бландинсвилл", "name_en"=>"Blandinsville", "lang"=>"en", "lat"=>40.554727, "lon"=>-90.869371),
"/US/US-IL/bloomingdale-4635" => array("type"=>"village", "pop"=>21848, "name_ru"=>"Блумингдэйл", "name_en"=>"Bloomingdale", "lang"=>"en", "lat"=>41.948900, "lon"=>-88.086494),
"/US/US-IL/bloomington-4636" => array("type"=>"city", "pop"=>73026, "name_ru"=>"Блумингтон", "name_en"=>"Bloomington", "lang"=>"en", "lat"=>40.478950, "lon"=>-88.973549),
"/US/US-IL/blue-island-4637" => array("type"=>"city", "pop"=>22325, "name_ru"=>"Блу Айленд", "name_en"=>"Blue Island", "lang"=>"en", "lat"=>41.658400, "lon"=>-87.681424),
"/US/US-IL/blue-mound-4638" => array("type"=>"village", "pop"=>1020, "name_ru"=>"Блю Маунд", "name_en"=>"Blue Mound", "lang"=>"en", "lat"=>39.700667, "lon"=>-89.118879),
"/US/US-IL/bluffs-4639" => array("type"=>"village", "pop"=>701, "name_ru"=>"Блафс", "name_en"=>"Bluffs", "lang"=>"en", "lat"=>39.749954, "lon"=>-90.534444),
"/US/US-IL/bluford-4640" => array("type"=>"village", "pop"=>762, "name_ru"=>"Блафорд", "name_en"=>"Bluford", "lang"=>"en", "lat"=>38.326311, "lon"=>-88.735610),
"/US/US-IL/bolingbrook-4641" => array("type"=>"village", "pop"=>70823, "name_ru"=>"Болингбрук", "name_en"=>"Bolingbrook", "lang"=>"en", "lat"=>41.700400, "lon"=>-88.077165),
"/US/US-IL/bondville-4642" => array("type"=>"village", "pop"=>438, "name_ru"=>"Бондвилл", "name_en"=>"Bondville", "lang"=>"en", "lat"=>40.111424, "lon"=>-88.368867),
"/US/US-IL/bone-gap-4643" => array("type"=>"village", "pop"=>254, "name_ru"=>"Боун Гэп", "name_en"=>"Bone Gap", "lang"=>"en", "lat"=>38.444768, "lon"=>-87.997947),
"/US/US-IL/bonfield-4644" => array("type"=>"village", "pop"=>379, "name_ru"=>"Бонфилд", "name_en"=>"Bonfield", "lang"=>"en", "lat"=>41.147171, "lon"=>-88.056741),
"/US/US-IL/bonnie-4645" => array("type"=>"village", "pop"=>419, "name_ru"=>"Бонни", "name_en"=>"Bonnie", "lang"=>"en", "lat"=>38.201456, "lon"=>-88.907464),
"/US/US-IL/boulder-hill-4646" => array("type"=>"cdp", "pop"=>8894, "name_ru"=>"Боулдер Хилл", "name_en"=>"Boulder Hill", "lang"=>"en", "lat"=>41.712504, "lon"=>-88.336107),
"/US/US-IL/bourbonnais-4647" => array("type"=>"village", "pop"=>19024, "name_ru"=>"Бурбоннис", "name_en"=>"Bourbonnais", "lang"=>"en", "lat"=>41.163225, "lon"=>-87.880381),
"/US/US-IL/bowen-4648" => array("type"=>"village", "pop"=>493, "name_ru"=>"Боуэн", "name_en"=>"Bowen", "lang"=>"en", "lat"=>40.232024, "lon"=>-91.063580),
"/US/US-IL/braceville-4649" => array("type"=>"village", "pop"=>832, "name_ru"=>"Брасевилл", "name_en"=>"Braceville", "lang"=>"en", "lat"=>41.224192, "lon"=>-88.266376),
"/US/US-IL/bradford-4650" => array("type"=>"village", "pop"=>758, "name_ru"=>"Брадфорд", "name_en"=>"Bradford", "lang"=>"en", "lat"=>41.176877, "lon"=>-89.657163),
"/US/US-IL/bradley-4651" => array("type"=>"village", "pop"=>14964, "name_ru"=>"Брадли", "name_en"=>"Bradley", "lang"=>"en", "lat"=>41.147200, "lon"=>-87.855675),
"/US/US-IL/braidwood-4652" => array("type"=>"city", "pop"=>6664, "name_ru"=>"Брейдвуд", "name_en"=>"Braidwood", "lang"=>"en", "lat"=>41.268141, "lon"=>-88.223243),
"/US/US-IL/breese-4653" => array("type"=>"city", "pop"=>4357, "name_ru"=>"Бриз", "name_en"=>"Breese", "lang"=>"en", "lat"=>38.611045, "lon"=>-89.522238),
"/US/US-IL/bridgeport-4654" => array("type"=>"city", "pop"=>2085, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>38.709959, "lon"=>-87.758545),
"/US/US-IL/bridgeview-4655" => array("type"=>"village", "pop"=>14991, "name_ru"=>"Бриджвью", "name_en"=>"Bridgeview", "lang"=>"en", "lat"=>41.738750, "lon"=>-87.805887),
"/US/US-IL/brighton-4656" => array("type"=>"village", "pop"=>2376, "name_ru"=>"Брайтон", "name_en"=>"Brighton", "lang"=>"en", "lat"=>39.040255, "lon"=>-90.138968),
"/US/US-IL/brimfield-4657" => array("type"=>"village", "pop"=>902, "name_ru"=>"Бримфилд", "name_en"=>"Brimfield", "lang"=>"en", "lat"=>40.838173, "lon"=>-89.883891),
"/US/US-IL/broadlands-4658" => array("type"=>"village", "pop"=>305, "name_ru"=>"Броадлендс", "name_en"=>"Broadlands", "lang"=>"en", "lat"=>39.908516, "lon"=>-87.995424),
"/US/US-IL/broadview-4659" => array("type"=>"village", "pop"=>7625, "name_ru"=>"Броадвью", "name_en"=>"Broadview", "lang"=>"en", "lat"=>41.858050, "lon"=>-87.855939),
"/US/US-IL/broadwell-4660" => array("type"=>"village", "pop"=>167, "name_ru"=>"Броадуэлл", "name_en"=>"Broadwell", "lang"=>"en", "lat"=>40.067320, "lon"=>-89.442611),
"/US/US-IL/brocton-4661" => array("type"=>"village", "pop"=>295, "name_ru"=>"Броктон", "name_en"=>"Brocton", "lang"=>"en", "lat"=>39.716146, "lon"=>-87.933077),
"/US/US-IL/brookfield-4662" => array("type"=>"village", "pop"=>18077, "name_ru"=>"Брукфилд", "name_en"=>"Brookfield", "lang"=>"en", "lat"=>41.824300, "lon"=>-87.847758),
"/US/US-IL/brooklyn-4663" => array("type"=>"village", "pop"=>620, "name_ru"=>"Бруклин", "name_en"=>"Brooklyn", "lang"=>"en", "lat"=>38.651567, "lon"=>-90.167392),
"/US/US-IL/brookport-4664" => array("type"=>"city", "pop"=>1027, "name_ru"=>"Брукпорт", "name_en"=>"Brookport", "lang"=>"en", "lat"=>37.125001, "lon"=>-88.627170),
"/US/US-IL/broughton-4665" => array("type"=>"village", "pop"=>189, "name_ru"=>"Бротон", "name_en"=>"Broughton", "lang"=>"en", "lat"=>37.934080, "lon"=>-88.461361),
"/US/US-IL/browning-4666" => array("type"=>"village", "pop"=>124, "name_ru"=>"Браунинг", "name_en"=>"Browning", "lang"=>"en", "lat"=>40.128062, "lon"=>-90.372814),
"/US/US-IL/browns-4667" => array("type"=>"village", "pop"=>163, "name_ru"=>"Браунс", "name_en"=>"Browns", "lang"=>"en", "lat"=>38.377735, "lon"=>-87.983408),
"/US/US-IL/brownstown-4668" => array("type"=>"village", "pop"=>706, "name_ru"=>"Браунстаун", "name_en"=>"Brownstown", "lang"=>"en", "lat"=>38.996275, "lon"=>-88.953593),
"/US/US-IL/brussels-4669" => array("type"=>"village", "pop"=>142, "name_ru"=>"Брасселс", "name_en"=>"Brussels", "lang"=>"en", "lat"=>38.950996, "lon"=>-90.588827),
"/US/US-IL/bryant-4670" => array("type"=>"village", "pop"=>242, "name_ru"=>"Брайант", "name_en"=>"Bryant", "lang"=>"en", "lat"=>40.465490, "lon"=>-90.094501),
"/US/US-IL/buckingham-4671" => array("type"=>"village", "pop"=>241, "name_ru"=>"Бакингем", "name_en"=>"Buckingham", "lang"=>"en", "lat"=>41.046815, "lon"=>-88.176090),
"/US/US-IL/buckley-4672" => array("type"=>"village", "pop"=>564, "name_ru"=>"Бакли", "name_en"=>"Buckley", "lang"=>"en", "lat"=>40.597202, "lon"=>-88.036979),
"/US/US-IL/buckner-4673" => array("type"=>"village", "pop"=>488, "name_ru"=>"Бакнер", "name_en"=>"Buckner", "lang"=>"en", "lat"=>37.981081, "lon"=>-89.015621),
"/US/US-IL/buda-4674" => array("type"=>"village", "pop"=>567, "name_ru"=>"Буда", "name_en"=>"Buda", "lang"=>"en", "lat"=>41.327684, "lon"=>-89.678424),
"/US/US-IL/buffalo-4675" => array("type"=>"village", "pop"=>451, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>39.849490, "lon"=>-89.408953),
"/US/US-IL/buffalo-grove-4676" => array("type"=>"village", "pop"=>42972, "name_ru"=>"Буффало Гров", "name_en"=>"Buffalo Grove", "lang"=>"en", "lat"=>42.166307, "lon"=>-87.963323),
"/US/US-IL/bull-valley-4677" => array("type"=>"village", "pop"=>841, "name_ru"=>"Булл Валли", "name_en"=>"Bull Valley", "lang"=>"en", "lat"=>42.311377, "lon"=>-88.349915),
"/US/US-IL/bulpitt-4678" => array("type"=>"village", "pop"=>193, "name_ru"=>"Булпит", "name_en"=>"Bulpitt", "lang"=>"en", "lat"=>39.591556, "lon"=>-89.425061),
"/US/US-IL/buncombe-4679" => array("type"=>"village", "pop"=>193, "name_ru"=>"Банкомб", "name_en"=>"Buncombe", "lang"=>"en", "lat"=>37.470936, "lon"=>-88.974865),
"/US/US-IL/bunker-hill-4680" => array("type"=>"city", "pop"=>1741, "name_ru"=>"Банкер Хилл", "name_en"=>"Bunker Hill", "lang"=>"en", "lat"=>39.040468, "lon"=>-89.950759),
"/US/US-IL/burbank-4681" => array("type"=>"city", "pop"=>27567, "name_ru"=>"Бурбанк", "name_en"=>"Burbank", "lang"=>"en", "lat"=>41.744100, "lon"=>-87.768974),
"/US/US-IL/bureau-junction-4682" => array("type"=>"village", "pop"=>353, "name_ru"=>"Бюро Джанкшен", "name_en"=>"Bureau Junction", "lang"=>"en", "lat"=>41.287794, "lon"=>-89.364320),
"/US/US-IL/burlington-4683" => array("type"=>"village", "pop"=>482, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>42.050447, "lon"=>-88.547370),
"/US/US-IL/burnham-4684" => array("type"=>"village", "pop"=>3973, "name_ru"=>"Бернхэм", "name_en"=>"Burnham", "lang"=>"en", "lat"=>41.638200, "lon"=>-87.541174),
"/US/US-IL/burnt-prairie-4685" => array("type"=>"village", "pop"=>55, "name_ru"=>"Бернт Прейри", "name_en"=>"Burnt Prairie", "lang"=>"en", "lat"=>38.250490, "lon"=>-88.257709),
"/US/US-IL/burr-ridge-4686" => array("type"=>"village", "pop"=>11398, "name_ru"=>"Берр Ридж", "name_en"=>"Burr Ridge", "lang"=>"en", "lat"=>41.750900, "lon"=>-87.920599),
"/US/US-IL/bush-4687" => array("type"=>"village", "pop"=>261, "name_ru"=>"Буш", "name_en"=>"Bush", "lang"=>"en", "lat"=>37.841899, "lon"=>-89.129762),
"/US/US-IL/bushnell-4688" => array("type"=>"city", "pop"=>2921, "name_ru"=>"Бушнелл", "name_en"=>"Bushnell", "lang"=>"en", "lat"=>40.551546, "lon"=>-90.504549),
"/US/US-IL/butler-4689" => array("type"=>"village", "pop"=>190, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>39.197923, "lon"=>-89.534012),
"/US/US-IL/byron-4690" => array("type"=>"city", "pop"=>3850, "name_ru"=>"Байрон", "name_en"=>"Byron", "lang"=>"en", "lat"=>42.128400, "lon"=>-89.258263),
"/US/US-IL/cabery-4691" => array("type"=>"village", "pop"=>260, "name_ru"=>"Кабери", "name_en"=>"Cabery", "lang"=>"en", "lat"=>40.994696, "lon"=>-88.204756),
"/US/US-IL/cahokia-4692" => array("type"=>"village", "pop"=>15103, "name_ru"=>"Кахокия", "name_en"=>"Cahokia", "lang"=>"en", "lat"=>38.564758, "lon"=>-90.177589),
"/US/US-IL/cairo-4693" => array("type"=>"city", "pop"=>3105, "name_ru"=>"Каир", "name_en"=>"Cairo", "lang"=>"en", "lat"=>36.997623, "lon"=>-89.177608),
"/US/US-IL/calhoun-4694" => array("type"=>"village", "pop"=>211, "name_ru"=>"Калхун", "name_en"=>"Calhoun", "lang"=>"en", "lat"=>38.651435, "lon"=>-88.042813),
"/US/US-IL/calumet-city-4695" => array("type"=>"city", "pop"=>36800, "name_ru"=>"Калумет Сити", "name_en"=>"Calumet City", "lang"=>"en", "lat"=>41.611750, "lon"=>-87.549549),
"/US/US-IL/calumet-park-4696" => array("type"=>"village", "pop"=>7980, "name_ru"=>"Калумет парк", "name_en"=>"Calumet Park", "lang"=>"en", "lat"=>41.664067, "lon"=>-87.656435),
"/US/US-IL/camargo-4697" => array("type"=>"village", "pop"=>511, "name_ru"=>"Камарго", "name_en"=>"Camargo", "lang"=>"en", "lat"=>39.799484, "lon"=>-88.166699),
"/US/US-IL/cambria-4698" => array("type"=>"village", "pop"=>1336, "name_ru"=>"Камбрия", "name_en"=>"Cambria", "lang"=>"en", "lat"=>37.778647, "lon"=>-89.114321),
"/US/US-IL/cambridge-4699" => array("type"=>"village", "pop"=>2093, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>41.303291, "lon"=>-90.193292),
"/US/US-IL/camden-4700" => array("type"=>"village", "pop"=>93, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>40.153037, "lon"=>-90.772359),
"/US/US-IL/camp-point-4702" => array("type"=>"village", "pop"=>1193, "name_ru"=>"Кэмп Пойнт", "name_en"=>"Camp Point", "lang"=>"en", "lat"=>40.038050, "lon"=>-91.066632),
"/US/US-IL/campbell-hill-4701" => array("type"=>"village", "pop"=>308, "name_ru"=>"Кэмпбэлл Хилл", "name_en"=>"Campbell Hill", "lang"=>"en", "lat"=>37.930173, "lon"=>-89.551006),
"/US/US-IL/campus-4703" => array("type"=>"village", "pop"=>140, "name_ru"=>"Кампус", "name_en"=>"Campus", "lang"=>"en", "lat"=>41.024814, "lon"=>-88.307917),
"/US/US-IL/canton-4704" => array("type"=>"city", "pop"=>14545, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>40.562548, "lon"=>-90.040260),
"/US/US-IL/cantrall-4705" => array("type"=>"village", "pop"=>133, "name_ru"=>"Кантролл", "name_en"=>"Cantrall", "lang"=>"en", "lat"=>39.934360, "lon"=>-89.680145),
"/US/US-IL/capron-4706" => array("type"=>"village", "pop"=>1547, "name_ru"=>"Капрон", "name_en"=>"Capron", "lang"=>"en", "lat"=>42.398853, "lon"=>-88.742635),
"/US/US-IL/carbon-cliff-4707" => array("type"=>"village", "pop"=>1657, "name_ru"=>"Карбон Клифф", "name_en"=>"Carbon Cliff", "lang"=>"en", "lat"=>41.497633, "lon"=>-90.390349),
"/US/US-IL/carbon-hill-4709" => array("type"=>"village", "pop"=>419, "name_ru"=>"Карбон Хилл", "name_en"=>"Carbon Hill", "lang"=>"en", "lat"=>41.296285, "lon"=>-88.299471),
"/US/US-IL/carbondale-4708" => array("type"=>"city", "pop"=>26231, "name_ru"=>"Карбондейл", "name_en"=>"Carbondale", "lang"=>"en", "lat"=>37.724629, "lon"=>-89.217390),
"/US/US-IL/carlinville-4710" => array("type"=>"city", "pop"=>5962, "name_ru"=>"Карлинвилл", "name_en"=>"Carlinville", "lang"=>"en", "lat"=>39.280386, "lon"=>-89.880901),
"/US/US-IL/carlock-4711" => array("type"=>"village", "pop"=>478, "name_ru"=>"Карлок", "name_en"=>"Carlock", "lang"=>"en", "lat"=>40.582079, "lon"=>-89.132682),
"/US/US-IL/carlyle-4712" => array("type"=>"city", "pop"=>3378, "name_ru"=>"Карлайл", "name_en"=>"Carlyle", "lang"=>"en", "lat"=>38.619000, "lon"=>-89.372241),
"/US/US-IL/carmi-4713" => array("type"=>"city", "pop"=>5203, "name_ru"=>"Карми", "name_en"=>"Carmi", "lang"=>"en", "lat"=>38.086413, "lon"=>-88.171181),
"/US/US-IL/carol-stream-4714" => array("type"=>"village", "pop"=>40004, "name_ru"=>"Кэрол Стрим", "name_en"=>"Carol Stream", "lang"=>"en", "lat"=>41.916815, "lon"=>-88.131037),
"/US/US-IL/carpentersville-4715" => array("type"=>"village", "pop"=>37741, "name_ru"=>"Карпентерсвилл", "name_en"=>"Carpentersville", "lang"=>"en", "lat"=>42.121324, "lon"=>-88.271487),
"/US/US-IL/carriers-mills-4716" => array("type"=>"village", "pop"=>1991, "name_ru"=>"Карриэрс Миллс", "name_en"=>"Carriers Mills", "lang"=>"en", "lat"=>37.688577, "lon"=>-88.628900),
"/US/US-IL/carrollton-4717" => array("type"=>"city", "pop"=>2460, "name_ru"=>"Карролтон", "name_en"=>"Carrollton", "lang"=>"en", "lat"=>39.297815, "lon"=>-90.406825),
"/US/US-IL/carterville-4718" => array("type"=>"city", "pop"=>5441, "name_ru"=>"Картервилл", "name_en"=>"Carterville", "lang"=>"en", "lat"=>37.761710, "lon"=>-89.082743),
"/US/US-IL/carthage-4719" => array("type"=>"city", "pop"=>2504, "name_ru"=>"Картаж", "name_en"=>"Carthage", "lang"=>"en", "lat"=>40.414319, "lon"=>-91.135126),
"/US/US-IL/cary-4720" => array("type"=>"village", "pop"=>19559, "name_ru"=>"Кэри", "name_en"=>"Cary", "lang"=>"en", "lat"=>42.208950, "lon"=>-88.249749),
"/US/US-IL/casey-4721" => array("type"=>"city", "pop"=>2914, "name_ru"=>"Кейси", "name_en"=>"Casey", "lang"=>"en", "lat"=>39.302560, "lon"=>-87.989340),
"/US/US-IL/caseyville-4722" => array("type"=>"village", "pop"=>4280, "name_ru"=>"Кейсивилл", "name_en"=>"Caseyville", "lang"=>"en", "lat"=>38.633017, "lon"=>-90.032721),
"/US/US-IL/catlin-4723" => array("type"=>"village", "pop"=>2110, "name_ru"=>"Катлин", "name_en"=>"Catlin", "lang"=>"en", "lat"=>40.068394, "lon"=>-87.707853),
"/US/US-IL/cave-in-rock-4724" => array("type"=>"village", "pop"=>307, "name_ru"=>"Кейв-Ин-Рок", "name_en"=>"Cave-In-Rock", "lang"=>"en", "lat"=>37.470162, "lon"=>-88.165542),
"/US/US-IL/cedar-point-4725" => array("type"=>"village", "pop"=>261, "name_ru"=>"Сидар Пойнт", "name_en"=>"Cedar Point", "lang"=>"en", "lat"=>41.264677, "lon"=>-89.125165),
"/US/US-IL/cedarville-4726" => array("type"=>"village", "pop"=>683, "name_ru"=>"Сидарвилл", "name_en"=>"Cedarville", "lang"=>"en", "lat"=>42.375088, "lon"=>-89.636466),
"/US/US-IL/central-city-4727" => array("type"=>"village", "pop"=>1313, "name_ru"=>"Сентрал Сити", "name_en"=>"Central City", "lang"=>"en", "lat"=>38.548031, "lon"=>-89.128637),
"/US/US-IL/centralia-4728" => array("type"=>"city", "pop"=>13569, "name_ru"=>"Сентралия", "name_en"=>"Centralia", "lang"=>"en", "lat"=>38.522826, "lon"=>-89.126957),
"/US/US-IL/centreville-4729" => array("type"=>"city", "pop"=>5635, "name_ru"=>"Сентревилл", "name_en"=>"Centreville", "lang"=>"en", "lat"=>38.577217, "lon"=>-90.101576),
"/US/US-IL/cerro-gordo-4730" => array("type"=>"village", "pop"=>1343, "name_ru"=>"Серро Гордо", "name_en"=>"Cerro Gordo", "lang"=>"en", "lat"=>39.890095, "lon"=>-88.734551),
"/US/US-IL/chadwick-4731" => array("type"=>"village", "pop"=>471, "name_ru"=>"Чадвик", "name_en"=>"Chadwick", "lang"=>"en", "lat"=>42.014089, "lon"=>-89.888949),
"/US/US-IL/champaign-4732" => array("type"=>"city", "pop"=>79389, "name_ru"=>"Шампейн", "name_en"=>"Champaign", "lang"=>"en", "lat"=>40.113000, "lon"=>-88.264949),
"/US/US-IL/chandlerville-4733" => array("type"=>"village", "pop"=>693, "name_ru"=>"Чандлервилл", "name_en"=>"Chandlerville", "lang"=>"en", "lat"=>40.048194, "lon"=>-90.150718),
"/US/US-IL/channahon-4734" => array("type"=>"village", "pop"=>14030, "name_ru"=>"Чаннахон", "name_en"=>"Channahon", "lang"=>"en", "lat"=>41.434467, "lon"=>-88.218579),
"/US/US-IL/channel-lake-4735" => array("type"=>"cdp", "pop"=>1660, "name_ru"=>"Чаннел Лейк", "name_en"=>"Channel Lake", "lang"=>"en", "lat"=>42.484300, "lon"=>-88.150294),
"/US/US-IL/chapin-4736" => array("type"=>"village", "pop"=>568, "name_ru"=>"Чапин", "name_en"=>"Chapin", "lang"=>"en", "lat"=>39.767338, "lon"=>-90.403300),
"/US/US-IL/charleston-4737" => array("type"=>"city", "pop"=>21715, "name_ru"=>"Чарльстон", "name_en"=>"Charleston", "lang"=>"en", "lat"=>39.484830, "lon"=>-88.176072),
"/US/US-IL/chatham-4738" => array("type"=>"village", "pop"=>10676, "name_ru"=>"Чатем", "name_en"=>"Chatham", "lang"=>"en", "lat"=>39.672856, "lon"=>-89.697078),
"/US/US-IL/chatsworth-4739" => array("type"=>"town", "pop"=>1179, "name_ru"=>"Чатсворт", "name_en"=>"Chatsworth", "lang"=>"en", "lat"=>40.752990, "lon"=>-88.293781),
"/US/US-IL/chebanse-4740" => array("type"=>"village", "pop"=>1097, "name_ru"=>"Чебанс", "name_en"=>"Chebanse", "lang"=>"en", "lat"=>41.002200, "lon"=>-87.913849),
"/US/US-IL/chenoa-4741" => array("type"=>"city", "pop"=>1832, "name_ru"=>"Ченоа", "name_en"=>"Chenoa", "lang"=>"en", "lat"=>40.744201, "lon"=>-88.718135),
"/US/US-IL/cherry-4742" => array("type"=>"village", "pop"=>488, "name_ru"=>"Черри", "name_en"=>"Cherry", "lang"=>"en", "lat"=>41.429557, "lon"=>-89.213611),
"/US/US-IL/cherry-valley-4743" => array("type"=>"village", "pop"=>2267, "name_ru"=>"Черри Валли", "name_en"=>"Cherry Valley", "lang"=>"en", "lat"=>42.240050, "lon"=>-88.961824),
"/US/US-IL/chester-4744" => array("type"=>"city", "pop"=>7835, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>37.919682, "lon"=>-89.825053),
"/US/US-IL/chesterfield-4745" => array("type"=>"village", "pop"=>215, "name_ru"=>"Честерфилд", "name_en"=>"Chesterfield", "lang"=>"en", "lat"=>39.255707, "lon"=>-90.067681),
"/US/US-IL/chicago-4746" => array("type"=>"city", "pop"=>2853114, "name_ru"=>"Чикаго", "name_en"=>"Chicago", "lang"=>"en", "lat"=>41.837050, "lon"=>-87.684965),
"/US/US-IL/chicago-heights-4747" => array("type"=>"city", "pop"=>30586, "name_ru"=>"Чикаго Хайтс", "name_en"=>"Chicago Heights", "lang"=>"en", "lat"=>41.509900, "lon"=>-87.637567),
"/US/US-IL/chicago-ridge-4748" => array("type"=>"village", "pop"=>13366, "name_ru"=>"Чикаго Ридж", "name_en"=>"Chicago Ridge", "lang"=>"en", "lat"=>41.702450, "lon"=>-87.778638),
"/US/US-IL/chillicothe-4749" => array("type"=>"city", "pop"=>5901, "name_ru"=>"Чилликот", "name_en"=>"Chillicothe", "lang"=>"en", "lat"=>40.918303, "lon"=>-89.499265),
"/US/US-IL/chrisman-4750" => array("type"=>"city", "pop"=>1232, "name_ru"=>"Крисман", "name_en"=>"Chrisman", "lang"=>"en", "lat"=>39.804026, "lon"=>-87.674243),
"/US/US-IL/christopher-4751" => array("type"=>"city", "pop"=>2819, "name_ru"=>"Кристофер", "name_en"=>"Christopher", "lang"=>"en", "lat"=>37.972625, "lon"=>-89.051850),
"/US/US-IL/cicero-4752" => array("type"=>"town", "pop"=>80414, "name_ru"=>"Сисеро", "name_en"=>"Cicero", "lang"=>"en", "lat"=>41.844350, "lon"=>-87.759852),
"/US/US-IL/cisco-4753" => array("type"=>"village", "pop"=>274, "name_ru"=>"Сиско", "name_en"=>"Cisco", "lang"=>"en", "lat"=>40.013715, "lon"=>-88.723521),
"/US/US-IL/cisne-4754" => array("type"=>"village", "pop"=>649, "name_ru"=>"Сисн", "name_en"=>"Cisne", "lang"=>"en", "lat"=>38.513997, "lon"=>-88.437087),
"/US/US-IL/cissna-park-4755" => array("type"=>"village", "pop"=>788, "name_ru"=>"Сиссна Парк", "name_en"=>"Cissna Park", "lang"=>"en", "lat"=>40.566821, "lon"=>-87.892525),
"/US/US-IL/claremont-4756" => array("type"=>"village", "pop"=>201, "name_ru"=>"Клэрмонт", "name_en"=>"Claremont", "lang"=>"en", "lat"=>38.719086, "lon"=>-87.974480),
"/US/US-IL/clarendon-hills-4757" => array("type"=>"village", "pop"=>8582, "name_ru"=>"Карендон Хилс", "name_en"=>"Clarendon Hills", "lang"=>"en", "lat"=>41.798450, "lon"=>-87.957557),
"/US/US-IL/clay-city-4758" => array("type"=>"village", "pop"=>939, "name_ru"=>"Клей Сити", "name_en"=>"Clay City", "lang"=>"en", "lat"=>38.685496, "lon"=>-88.351193),
"/US/US-IL/clayton-4759" => array("type"=>"village", "pop"=>829, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>40.030328, "lon"=>-90.958001),
"/US/US-IL/clear-lake-4760" => array("type"=>"village", "pop"=>247, "name_ru"=>"Клир Лейк", "name_en"=>"Clear Lake", "lang"=>"en", "lat"=>39.814128, "lon"=>-89.566868),
"/US/US-IL/cleveland-4761" => array("type"=>"village", "pop"=>239, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>41.502850, "lon"=>-90.316440),
"/US/US-IL/clifton-4762" => array("type"=>"village", "pop"=>1257, "name_ru"=>"Слифтон", "name_en"=>"Clifton", "lang"=>"en", "lat"=>40.935549, "lon"=>-87.933446),
"/US/US-IL/clinton-4763" => array("type"=>"city", "pop"=>7154, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>40.150328, "lon"=>-88.961555),
"/US/US-IL/coal-city-4764" => array("type"=>"village", "pop"=>5729, "name_ru"=>"Коал Сити", "name_en"=>"Coal City", "lang"=>"en", "lat"=>41.288297, "lon"=>-88.278212),
"/US/US-IL/coal-valley-4766" => array("type"=>"village", "pop"=>4001, "name_ru"=>"Коал Валли", "name_en"=>"Coal Valley", "lang"=>"en", "lat"=>41.448950, "lon"=>-90.447751),
"/US/US-IL/coalton-4765" => array("type"=>"village", "pop"=>297, "name_ru"=>"Коалтон", "name_en"=>"Coalton", "lang"=>"en", "lat"=>39.284626, "lon"=>-89.304452),
"/US/US-IL/coatsburg-4767" => array("type"=>"village", "pop"=>222, "name_ru"=>"Коатсбург", "name_en"=>"Coatsburg", "lang"=>"en", "lat"=>40.032291, "lon"=>-91.159421),
"/US/US-IL/cobden-4768" => array("type"=>"village", "pop"=>1098, "name_ru"=>"Кобден", "name_en"=>"Cobden", "lang"=>"en", "lat"=>37.533439, "lon"=>-89.254224),
"/US/US-IL/coffeen-4769" => array("type"=>"city", "pop"=>687, "name_ru"=>"Коффин", "name_en"=>"Coffeen", "lang"=>"en", "lat"=>39.088863, "lon"=>-89.390527),
"/US/US-IL/colchester-4770" => array("type"=>"city", "pop"=>1355, "name_ru"=>"Колчестер", "name_en"=>"Colchester", "lang"=>"en", "lat"=>40.426753, "lon"=>-90.793005),
"/US/US-IL/coleta-4771" => array("type"=>"village", "pop"=>151, "name_ru"=>"Колета", "name_en"=>"Coleta", "lang"=>"en", "lat"=>41.905555, "lon"=>-89.799163),
"/US/US-IL/colfax-4772" => array("type"=>"village", "pop"=>1001, "name_ru"=>"Колфакс", "name_en"=>"Colfax", "lang"=>"en", "lat"=>40.566696, "lon"=>-88.616500),
"/US/US-IL/collinsville-4773" => array("type"=>"city", "pop"=>25960, "name_ru"=>"Коллинсвилл", "name_en"=>"Collinsville", "lang"=>"en", "lat"=>38.678400, "lon"=>-89.997271),
"/US/US-IL/colona-4774" => array("type"=>"village", "pop"=>5232, "name_ru"=>"Колона", "name_en"=>"Colona", "lang"=>"en", "lat"=>41.470650, "lon"=>-90.340470),
"/US/US-IL/colp-4775" => array("type"=>"village", "pop"=>227, "name_ru"=>"Колп", "name_en"=>"Colp", "lang"=>"en", "lat"=>37.807499, "lon"=>-89.077857),
"/US/US-IL/columbia-4776" => array("type"=>"city", "pop"=>9342, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>38.455243, "lon"=>-90.226249),
"/US/US-IL/columbus-4777" => array("type"=>"village", "pop"=>110, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>39.988094, "lon"=>-91.146454),
"/US/US-IL/compton-4778" => array("type"=>"village", "pop"=>333, "name_ru"=>"Комптон", "name_en"=>"Compton", "lang"=>"en", "lat"=>41.694192, "lon"=>-89.085917),
"/US/US-IL/concord-4779" => array("type"=>"village", "pop"=>169, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>39.816129, "lon"=>-90.371650),
"/US/US-IL/congerville-4780" => array("type"=>"village", "pop"=>500, "name_ru"=>"Конгервилл", "name_en"=>"Congerville", "lang"=>"en", "lat"=>40.616741, "lon"=>-89.208482),
"/US/US-IL/cooksville-4781" => array("type"=>"village", "pop"=>204, "name_ru"=>"Куксвилл", "name_en"=>"Cooksville", "lang"=>"en", "lat"=>40.542568, "lon"=>-88.715139),
"/US/US-IL/cordova-4782" => array("type"=>"village", "pop"=>682, "name_ru"=>"Кордова", "name_en"=>"Cordova", "lang"=>"en", "lat"=>41.678041, "lon"=>-90.322021),
"/US/US-IL/cornell-4783" => array("type"=>"village", "pop"=>488, "name_ru"=>"Корнелл", "name_en"=>"Cornell", "lang"=>"en", "lat"=>40.992329, "lon"=>-88.729737),
"/US/US-IL/cortland-4784" => array("type"=>"town", "pop"=>4164, "name_ru"=>"Кортленд", "name_en"=>"Cortland", "lang"=>"en", "lat"=>41.927279, "lon"=>-88.694448),
"/US/US-IL/coulterville-4785" => array("type"=>"village", "pop"=>1148, "name_ru"=>"Колтервилл", "name_en"=>"Coulterville", "lang"=>"en", "lat"=>38.184935, "lon"=>-89.604344),
"/US/US-IL/country-club-hills-4786" => array("type"=>"city", "pop"=>16723, "name_ru"=>"Кантри Клаб Хилс", "name_en"=>"Country Club Hills", "lang"=>"en", "lat"=>41.563150, "lon"=>-87.725255),
"/US/US-IL/countryside-4787" => array("type"=>"city", "pop"=>5795, "name_ru"=>"Кантрисайд", "name_en"=>"Countryside", "lang"=>"en", "lat"=>41.776550, "lon"=>-87.876561),
"/US/US-IL/cowden-4788" => array("type"=>"village", "pop"=>568, "name_ru"=>"Кауден", "name_en"=>"Cowden", "lang"=>"en", "lat"=>39.248911, "lon"=>-88.858792),
"/US/US-IL/crainville-4789" => array("type"=>"village", "pop"=>1353, "name_ru"=>"Крейнвилл", "name_en"=>"Crainville", "lang"=>"en", "lat"=>37.750287, "lon"=>-89.060122),
"/US/US-IL/creal-springs-4790" => array("type"=>"city", "pop"=>713, "name_ru"=>"Крил Спрингс", "name_en"=>"Creal Springs", "lang"=>"en", "lat"=>37.618577, "lon"=>-88.837608),
"/US/US-IL/crescent-city-4791" => array("type"=>"village", "pop"=>591, "name_ru"=>"Кресцент Сити", "name_en"=>"Crescent City", "lang"=>"en", "lat"=>40.771471, "lon"=>-87.856438),
"/US/US-IL/crest-hill-4792" => array("type"=>"city", "pop"=>20451, "name_ru"=>"Крест Хилл", "name_en"=>"Crest Hill", "lang"=>"en", "lat"=>41.566267, "lon"=>-88.105056),
"/US/US-IL/creston-4793" => array("type"=>"village", "pop"=>623, "name_ru"=>"Крестон", "name_en"=>"Creston", "lang"=>"en", "lat"=>41.930788, "lon"=>-88.965461),
"/US/US-IL/crestwood-4794" => array("type"=>"village", "pop"=>11051, "name_ru"=>"Крествуд", "name_en"=>"Crestwood", "lang"=>"en", "lat"=>41.645250, "lon"=>-87.740288),
"/US/US-IL/crete-4795" => array("type"=>"village", "pop"=>9041, "name_ru"=>"Крет", "name_en"=>"Crete", "lang"=>"en", "lat"=>41.453194, "lon"=>-87.615090),
"/US/US-IL/creve-coeur-4796" => array("type"=>"village", "pop"=>5206, "name_ru"=>"Крив Коер", "name_en"=>"Creve Coeur", "lang"=>"en", "lat"=>40.641251, "lon"=>-89.599693),
"/US/US-IL/crossville-4797" => array("type"=>"village", "pop"=>712, "name_ru"=>"Кроссвилл", "name_en"=>"Crossville", "lang"=>"en", "lat"=>38.162623, "lon"=>-88.064459),
"/US/US-IL/crystal-lake-4798" => array("type"=>"city", "pop"=>41797, "name_ru"=>"Кристал Лэйк", "name_en"=>"Crystal Lake", "lang"=>"en", "lat"=>42.231850, "lon"=>-88.330539),
"/US/US-IL/crystal-lawns-4799" => array("type"=>"cdp", "pop"=>3037, "name_ru"=>"Кристал Лонс", "name_en"=>"Crystal Lawns", "lang"=>"en", "lat"=>41.570267, "lon"=>-88.157931),
"/US/US-IL/cuba-4800" => array("type"=>"city", "pop"=>1333, "name_ru"=>"Куба", "name_en"=>"Cuba", "lang"=>"en", "lat"=>40.493382, "lon"=>-90.193186),
"/US/US-IL/cullom-4801" => array("type"=>"village", "pop"=>521, "name_ru"=>"Каллом", "name_en"=>"Cullom", "lang"=>"en", "lat"=>40.877864, "lon"=>-88.269593),
"/US/US-IL/cutler-4802" => array("type"=>"village", "pop"=>528, "name_ru"=>"Катлер", "name_en"=>"Cutler", "lang"=>"en", "lat"=>38.032745, "lon"=>-89.566585),
"/US/US-IL/cypress-4803" => array("type"=>"village", "pop"=>281, "name_ru"=>"Сайпресс", "name_en"=>"Cypress", "lang"=>"en", "lat"=>37.365504, "lon"=>-89.019042),
"/US/US-IL/dahlgren-4804" => array("type"=>"village", "pop"=>496, "name_ru"=>"Далгрен", "name_en"=>"Dahlgren", "lang"=>"en", "lat"=>38.198574, "lon"=>-88.684137),
"/US/US-IL/dakota-4805" => array("type"=>"town", "pop"=>471, "name_ru"=>"Дакота", "name_en"=>"Dakota", "lang"=>"en", "lat"=>42.387373, "lon"=>-89.527114),
"/US/US-IL/dallas-city-4806" => array("type"=>"city", "pop"=>974, "name_ru"=>"Даллас Сити", "name_en"=>"Dallas City", "lang"=>"en", "lat"=>40.634968, "lon"=>-91.164481),
"/US/US-IL/dalton-city-4807" => array("type"=>"village", "pop"=>546, "name_ru"=>"Далтон Сити", "name_en"=>"Dalton City", "lang"=>"en", "lat"=>39.714130, "lon"=>-88.807020),
"/US/US-IL/dalzell-4808" => array("type"=>"village", "pop"=>717, "name_ru"=>"Далцелл", "name_en"=>"Dalzell", "lang"=>"en", "lat"=>41.358133, "lon"=>-89.175595),
"/US/US-IL/damiansville-4809" => array("type"=>"village", "pop"=>408, "name_ru"=>"Дэмиансвилл", "name_en"=>"Damiansville", "lang"=>"en", "lat"=>38.509485, "lon"=>-89.624187),
"/US/US-IL/dana-4810" => array("type"=>"village", "pop"=>170, "name_ru"=>"Дана", "name_en"=>"Dana", "lang"=>"en", "lat"=>40.956284, "lon"=>-88.949787),
"/US/US-IL/danforth-4811" => array("type"=>"village", "pop"=>562, "name_ru"=>"Дэнфорт", "name_en"=>"Danforth", "lang"=>"en", "lat"=>40.821205, "lon"=>-87.977933),
"/US/US-IL/danvers-4812" => array("type"=>"village", "pop"=>1149, "name_ru"=>"Дэнверс", "name_en"=>"Danvers", "lang"=>"en", "lat"=>40.528892, "lon"=>-89.173267),
"/US/US-IL/danville-4813" => array("type"=>"city", "pop"=>32248, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>40.143898, "lon"=>-87.619436),
"/US/US-IL/darien-4814" => array("type"=>"city", "pop"=>22370, "name_ru"=>"Дариен", "name_en"=>"Darien", "lang"=>"en", "lat"=>41.749333, "lon"=>-87.976253),
"/US/US-IL/davis-4815" => array("type"=>"village", "pop"=>625, "name_ru"=>"Давис", "name_en"=>"Davis", "lang"=>"en", "lat"=>42.422185, "lon"=>-89.415712),
"/US/US-IL/davis-junction-4816" => array("type"=>"village", "pop"=>2270, "name_ru"=>"Дэвис Джанкшен", "name_en"=>"Davis Junction", "lang"=>"en", "lat"=>42.101917, "lon"=>-89.092025),
"/US/US-IL/dawson-4817" => array("type"=>"village", "pop"=>462, "name_ru"=>"Доусон", "name_en"=>"Dawson", "lang"=>"en", "lat"=>39.853834, "lon"=>-89.462711),
"/US/US-IL/de-kalb-4823" => array("type"=>"city", "pop"=>34925, "name_ru"=>"Де Калб", "name_en"=>"De Kalb", "lang"=>"en", "lat"=>41.930629, "lon"=>-88.751910),
"/US/US-IL/de-land-4824" => array("type"=>"village", "pop"=>458, "name_ru"=>"Ди Лэнд", "name_en"=>"De Land", "lang"=>"en", "lat"=>40.121585, "lon"=>-88.643827),
"/US/US-IL/de-pue-4826" => array("type"=>"village", "pop"=>1753, "name_ru"=>"Де Пу", "name_en"=>"De Pue", "lang"=>"en", "lat"=>41.329715, "lon"=>-89.295767),
"/US/US-IL/de-soto-4827" => array("type"=>"village", "pop"=>1559, "name_ru"=>"Де Сото", "name_en"=>"De Soto", "lang"=>"en", "lat"=>37.815898, "lon"=>-89.227202),
"/US/US-IL/de-witt-4830" => array("type"=>"village", "pop"=>178, "name_ru"=>"Де Витт", "name_en"=>"De Witt", "lang"=>"en", "lat"=>40.184437, "lon"=>-88.785907),
"/US/US-IL/decatur-4818" => array("type"=>"city", "pop"=>76256, "name_ru"=>"Декатур", "name_en"=>"Decatur", "lang"=>"en", "lat"=>39.852250, "lon"=>-88.933848),
"/US/US-IL/deer-creek-4819" => array("type"=>"village", "pop"=>741, "name_ru"=>"Дир Крик", "name_en"=>"Deer Creek", "lang"=>"en", "lat"=>40.629340, "lon"=>-89.333128),
"/US/US-IL/deer-grove-4821" => array("type"=>"village", "pop"=>47, "name_ru"=>"Дир Гров", "name_en"=>"Deer Grove", "lang"=>"en", "lat"=>41.610107, "lon"=>-89.684620),
"/US/US-IL/deer-park-4822" => array("type"=>"village", "pop"=>3349, "name_ru"=>"Дир Парк", "name_en"=>"Deer Park", "lang"=>"en", "lat"=>42.168134, "lon"=>-88.088195),
"/US/US-IL/deerfield-4820" => array("type"=>"village", "pop"=>19667, "name_ru"=>"Дирфилд", "name_en"=>"Deerfield", "lang"=>"en", "lat"=>42.165600, "lon"=>-87.849899),
"/US/US-IL/delavan-4825" => array("type"=>"city", "pop"=>1755, "name_ru"=>"Делаван", "name_en"=>"Delavan", "lang"=>"en", "lat"=>40.371440, "lon"=>-89.544856),
"/US/US-IL/des-plaines-4828" => array("type"=>"city", "pop"=>57062, "name_ru"=>"Дес Плейнс", "name_en"=>"Des Plaines", "lang"=>"en", "lat"=>42.034600, "lon"=>-87.900949),
"/US/US-IL/detroit-4829" => array("type"=>"village", "pop"=>86, "name_ru"=>"Детройт", "name_en"=>"Detroit", "lang"=>"en", "lat"=>39.619763, "lon"=>-90.676337),
"/US/US-IL/diamond-4831" => array("type"=>"village", "pop"=>2388, "name_ru"=>"Даймонд", "name_en"=>"Diamond", "lang"=>"en", "lat"=>41.288666, "lon"=>-88.255198),
"/US/US-IL/dieterich-4832" => array("type"=>"village", "pop"=>595, "name_ru"=>"Дитерих", "name_en"=>"Dieterich", "lang"=>"en", "lat"=>39.059632, "lon"=>-88.382119),
"/US/US-IL/divernon-4833" => array("type"=>"village", "pop"=>1129, "name_ru"=>"Дайвернон", "name_en"=>"Divernon", "lang"=>"en", "lat"=>39.569337, "lon"=>-89.653998),
"/US/US-IL/dix-4834" => array("type"=>"village", "pop"=>498, "name_ru"=>"Дикс", "name_en"=>"Dix", "lang"=>"en", "lat"=>38.440627, "lon"=>-88.942400),
"/US/US-IL/dixmoor-4835" => array("type"=>"village", "pop"=>3794, "name_ru"=>"Диксмур", "name_en"=>"Dixmoor", "lang"=>"en", "lat"=>41.631896, "lon"=>-87.668034),
"/US/US-IL/dixon-4836" => array("type"=>"city", "pop"=>15033, "name_ru"=>"Диксон", "name_en"=>"Dixon", "lang"=>"en", "lat"=>41.849499, "lon"=>-89.478777),
"/US/US-IL/dolton-4837" => array("type"=>"village", "pop"=>23889, "name_ru"=>"Долтон", "name_en"=>"Dolton", "lang"=>"en", "lat"=>41.629300, "lon"=>-87.598716),
"/US/US-IL/dongola-4838" => array("type"=>"village", "pop"=>791, "name_ru"=>"Донгола", "name_en"=>"Dongola", "lang"=>"en", "lat"=>37.360037, "lon"=>-89.163882),
"/US/US-IL/donnellson-4839" => array("type"=>"village", "pop"=>256, "name_ru"=>"Доннелсон", "name_en"=>"Donnellson", "lang"=>"en", "lat"=>39.030198, "lon"=>-89.474583),
"/US/US-IL/donovan-4840" => array("type"=>"village", "pop"=>325, "name_ru"=>"Донован", "name_en"=>"Donovan", "lang"=>"en", "lat"=>40.885405, "lon"=>-87.614721),
"/US/US-IL/dorchester-4841" => array("type"=>"village", "pop"=>166, "name_ru"=>"Дорчестер", "name_en"=>"Dorchester", "lang"=>"en", "lat"=>39.086168, "lon"=>-89.888164),
"/US/US-IL/dover-4842" => array("type"=>"village", "pop"=>165, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>41.434206, "lon"=>-89.395702),
"/US/US-IL/dowell-4843" => array("type"=>"village", "pop"=>422, "name_ru"=>"Дауэлл", "name_en"=>"Dowell", "lang"=>"en", "lat"=>37.939495, "lon"=>-89.239470),
"/US/US-IL/downers-grove-4844" => array("type"=>"village", "pop"=>49250, "name_ru"=>"Даунерс Гров", "name_en"=>"Downers Grove", "lang"=>"en", "lat"=>41.797200, "lon"=>-88.017399),
"/US/US-IL/downs-4845" => array("type"=>"village", "pop"=>761, "name_ru"=>"Даунс", "name_en"=>"Downs", "lang"=>"en", "lat"=>40.395889, "lon"=>-88.871729),
"/US/US-IL/du-bois-4846" => array("type"=>"village", "pop"=>207, "name_ru"=>"Дю Буа", "name_en"=>"Du Bois", "lang"=>"en", "lat"=>38.221905, "lon"=>-89.212810),
"/US/US-IL/du-quoin-4850" => array("type"=>"city", "pop"=>6326, "name_ru"=>"Дю Квойн", "name_en"=>"Du Quoin", "lang"=>"en", "lat"=>38.004201, "lon"=>-89.235385),
"/US/US-IL/dunfermline-4847" => array("type"=>"village", "pop"=>301, "name_ru"=>"Данфермлайн", "name_en"=>"Dunfermline", "lang"=>"en", "lat"=>40.490926, "lon"=>-90.032736),
"/US/US-IL/dunlap-4848" => array("type"=>"village", "pop"=>934, "name_ru"=>"Данлап", "name_en"=>"Dunlap", "lang"=>"en", "lat"=>40.860750, "lon"=>-89.676574),
"/US/US-IL/dupo-4849" => array("type"=>"village", "pop"=>4073, "name_ru"=>"Дюпо", "name_en"=>"Dupo", "lang"=>"en", "lat"=>38.517500, "lon"=>-90.209612),
"/US/US-IL/durand-4851" => array("type"=>"village", "pop"=>1084, "name_ru"=>"Дуранд", "name_en"=>"Durand", "lang"=>"en", "lat"=>42.435581, "lon"=>-89.328645),
"/US/US-IL/dwight-4852" => array("type"=>"village", "pop"=>4267, "name_ru"=>"Двайт", "name_en"=>"Dwight", "lang"=>"en", "lat"=>41.097004, "lon"=>-88.424813),
"/US/US-IL/eagarville-4853" => array("type"=>"village", "pop"=>124, "name_ru"=>"Игарвилл", "name_en"=>"Eagarville", "lang"=>"en", "lat"=>39.110375, "lon"=>-89.784195),
"/US/US-IL/earlville-4854" => array("type"=>"city", "pop"=>1822, "name_ru"=>"Эрлвилл", "name_en"=>"Earlville", "lang"=>"en", "lat"=>41.588510, "lon"=>-88.923875),
"/US/US-IL/east-alton-4855" => array("type"=>"village", "pop"=>6563, "name_ru"=>"Ист Альтон", "name_en"=>"East Alton", "lang"=>"en", "lat"=>38.883150, "lon"=>-90.107791),
"/US/US-IL/east-brooklyn-4856" => array("type"=>"village", "pop"=>127, "name_ru"=>"Ист Бруклин", "name_en"=>"East Brooklyn", "lang"=>"en", "lat"=>41.172432, "lon"=>-88.265670),
"/US/US-IL/east-cape-girardeau-4857" => array("type"=>"village", "pop"=>371, "name_ru"=>"Ист Кейп Джирардо", "name_en"=>"East Cape Girardeau", "lang"=>"en", "lat"=>37.290224, "lon"=>-89.482953),
"/US/US-IL/east-carondelet-4858" => array("type"=>"village", "pop"=>577, "name_ru"=>"Ист Каронделет", "name_en"=>"East Carondelet", "lang"=>"en", "lat"=>38.538583, "lon"=>-90.236993),
"/US/US-IL/east-dubuque-4859" => array("type"=>"city", "pop"=>1921, "name_ru"=>"Ист Дабек", "name_en"=>"East Dubuque", "lang"=>"en", "lat"=>42.491472, "lon"=>-90.638831),
"/US/US-IL/east-dundee-4860" => array("type"=>"village", "pop"=>3121, "name_ru"=>"Ист Данди", "name_en"=>"East Dundee", "lang"=>"en", "lat"=>42.095809, "lon"=>-88.255801),
"/US/US-IL/east-galesburg-4861" => array("type"=>"village", "pop"=>775, "name_ru"=>"Ист Гейлсбург", "name_en"=>"East Galesburg", "lang"=>"en", "lat"=>40.943031, "lon"=>-90.311218),
"/US/US-IL/east-gillespie-4862" => array("type"=>"village", "pop"=>238, "name_ru"=>"Ист Гиллеспи", "name_en"=>"East Gillespie", "lang"=>"en", "lat"=>39.138091, "lon"=>-89.812901),
"/US/US-IL/east-hazel-crest-4863" => array("type"=>"village", "pop"=>1536, "name_ru"=>"Ист Хейзл Крест", "name_en"=>"East Hazel Crest", "lang"=>"en", "lat"=>41.576350, "lon"=>-87.649699),
"/US/US-IL/east-moline-4864" => array("type"=>"city", "pop"=>20892, "name_ru"=>"Ист Молайн", "name_en"=>"East Moline", "lang"=>"en", "lat"=>41.514250, "lon"=>-90.427424),
"/US/US-IL/east-peoria-4866" => array("type"=>"city", "pop"=>22760, "name_ru"=>"Ист Пеория", "name_en"=>"East Peoria", "lang"=>"en", "lat"=>40.671500, "lon"=>-89.547825),
"/US/US-IL/east-st.-louis-4867" => array("type"=>"city", "pop"=>28773, "name_ru"=>"Ист Сент Луис", "name_en"=>"East St. Louis", "lang"=>"en", "lat"=>38.614650, "lon"=>-90.131029),
"/US/US-IL/easton-4865" => array("type"=>"village", "pop"=>353, "name_ru"=>"Истон", "name_en"=>"Easton", "lang"=>"en", "lat"=>40.232043, "lon"=>-89.842139),
"/US/US-IL/eddyville-4868" => array("type"=>"village", "pop"=>141, "name_ru"=>"Эддивилл", "name_en"=>"Eddyville", "lang"=>"en", "lat"=>37.500316, "lon"=>-88.585192),
"/US/US-IL/edgewood-4869" => array("type"=>"village", "pop"=>520, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>38.921839, "lon"=>-88.664098),
"/US/US-IL/edinburg-4870" => array("type"=>"village", "pop"=>1108, "name_ru"=>"Эдинбург", "name_en"=>"Edinburg", "lang"=>"en", "lat"=>39.657839, "lon"=>-89.389636),
"/US/US-IL/edwardsville-4871" => array("type"=>"city", "pop"=>24557, "name_ru"=>"Эдвардсвилл", "name_en"=>"Edwardsville", "lang"=>"en", "lat"=>38.796360, "lon"=>-89.964970),
"/US/US-IL/effingham-4872" => array("type"=>"city", "pop"=>12489, "name_ru"=>"Эффингхэм", "name_en"=>"Effingham", "lang"=>"en", "lat"=>39.119635, "lon"=>-88.550928),
"/US/US-IL/el-dara-4874" => array("type"=>"village", "pop"=>82, "name_ru"=>"Эль Дара", "name_en"=>"El Dara", "lang"=>"en", "lat"=>39.622150, "lon"=>-90.991136),
"/US/US-IL/el-paso-4890" => array("type"=>"city", "pop"=>2802, "name_ru"=>"Эль Пасо", "name_en"=>"El Paso", "lang"=>"en", "lat"=>40.739794, "lon"=>-89.014788),
"/US/US-IL/elburn-4873" => array("type"=>"village", "pop"=>5276, "name_ru"=>"Эльбурн", "name_en"=>"Elburn", "lang"=>"en", "lat"=>41.890200, "lon"=>-88.471749),
"/US/US-IL/eldorado-4875" => array("type"=>"city", "pop"=>4392, "name_ru"=>"Эльдорадо", "name_en"=>"Eldorado", "lang"=>"en", "lat"=>37.811842, "lon"=>-88.441672),
"/US/US-IL/eldred-4876" => array("type"=>"village", "pop"=>197, "name_ru"=>"Эльдред", "name_en"=>"Eldred", "lang"=>"en", "lat"=>39.286242, "lon"=>-90.553320),
"/US/US-IL/elgin-4877" => array("type"=>"city", "pop"=>106330, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>42.039563, "lon"=>-88.287952),
"/US/US-IL/elizabeth-4878" => array("type"=>"village", "pop"=>650, "name_ru"=>"Элизабет", "name_en"=>"Elizabeth", "lang"=>"en", "lat"=>42.316148, "lon"=>-90.224652),
"/US/US-IL/elizabethtown-4879" => array("type"=>"village", "pop"=>308, "name_ru"=>"Элизабеттаун", "name_en"=>"Elizabethtown", "lang"=>"en", "lat"=>37.450829, "lon"=>-88.304337),
"/US/US-IL/elk-grove-village-4880" => array("type"=>"village", "pop"=>33320, "name_ru"=>"Элк Гров Виллидж", "name_en"=>"Elk Grove Village", "lang"=>"en", "lat"=>42.005750, "lon"=>-87.992599),
"/US/US-IL/elkhart-4881" => array("type"=>"village", "pop"=>423, "name_ru"=>"Элкхарт", "name_en"=>"Elkhart", "lang"=>"en", "lat"=>40.019313, "lon"=>-89.481787),
"/US/US-IL/elkville-4882" => array("type"=>"village", "pop"=>920, "name_ru"=>"Элквилл", "name_en"=>"Elkville", "lang"=>"en", "lat"=>37.909539, "lon"=>-89.236475),
"/US/US-IL/elliott-4883" => array("type"=>"village", "pop"=>335, "name_ru"=>"Эллиотт", "name_en"=>"Elliott", "lang"=>"en", "lat"=>40.464155, "lon"=>-88.274837),
"/US/US-IL/ellis-grove-4884" => array("type"=>"village", "pop"=>360, "name_ru"=>"Эллис Гров", "name_en"=>"Ellis Grove", "lang"=>"en", "lat"=>38.009319, "lon"=>-89.908659),
"/US/US-IL/ellisville-4885" => array("type"=>"village", "pop"=>82, "name_ru"=>"Эллисвилл", "name_en"=>"Ellisville", "lang"=>"en", "lat"=>40.627008, "lon"=>-90.305726),
"/US/US-IL/ellsworth-4886" => array("type"=>"village", "pop"=>260, "name_ru"=>"Эллсворт", "name_en"=>"Ellsworth", "lang"=>"en", "lat"=>40.449455, "lon"=>-88.716576),
"/US/US-IL/elmhurst-4887" => array("type"=>"city", "pop"=>46013, "name_ru"=>"Элмхерст", "name_en"=>"Elmhurst", "lang"=>"en", "lat"=>41.897950, "lon"=>-87.943624),
"/US/US-IL/elmwood-4888" => array("type"=>"city", "pop"=>1871, "name_ru"=>"Элмвуд", "name_en"=>"Elmwood", "lang"=>"en", "lat"=>40.779130, "lon"=>-89.966892),
"/US/US-IL/elmwood-park-4889" => array("type"=>"village", "pop"=>24052, "name_ru"=>"Элмвуд Парк", "name_en"=>"Elmwood Park", "lang"=>"en", "lat"=>41.922450, "lon"=>-87.816499),
"/US/US-IL/elsah-4891" => array("type"=>"village", "pop"=>631, "name_ru"=>"Элса", "name_en"=>"Elsah", "lang"=>"en", "lat"=>38.953672, "lon"=>-90.354618),
"/US/US-IL/elvaston-4892" => array("type"=>"village", "pop"=>140, "name_ru"=>"Элвастон", "name_en"=>"Elvaston", "lang"=>"en", "lat"=>40.394343, "lon"=>-91.248433),
"/US/US-IL/elwood-4893" => array("type"=>"village", "pop"=>2341, "name_ru"=>"Элвуд", "name_en"=>"Elwood", "lang"=>"en", "lat"=>41.412067, "lon"=>-88.108977),
"/US/US-IL/emden-4894" => array("type"=>"village", "pop"=>477, "name_ru"=>"Эмден", "name_en"=>"Emden", "lang"=>"en", "lat"=>40.297730, "lon"=>-89.485166),
"/US/US-IL/emington-4895" => array("type"=>"village", "pop"=>110, "name_ru"=>"Эмингтон", "name_en"=>"Emington", "lang"=>"en", "lat"=>40.970109, "lon"=>-88.357389),
"/US/US-IL/energy-4896" => array("type"=>"village", "pop"=>1199, "name_ru"=>"Энерджи", "name_en"=>"Energy", "lang"=>"en", "lat"=>37.775369, "lon"=>-89.025754),
"/US/US-IL/enfield-4897" => array("type"=>"village", "pop"=>601, "name_ru"=>"Энфилд", "name_en"=>"Enfield", "lang"=>"en", "lat"=>38.101666, "lon"=>-88.337273),
"/US/US-IL/equality-4898" => array("type"=>"village", "pop"=>668, "name_ru"=>"Эквалити", "name_en"=>"Equality", "lang"=>"en", "lat"=>37.735917, "lon"=>-88.343078),
"/US/US-IL/erie-4899" => array("type"=>"village", "pop"=>1548, "name_ru"=>"Эри", "name_en"=>"Erie", "lang"=>"en", "lat"=>41.658676, "lon"=>-90.081243),
"/US/US-IL/essex-4900" => array("type"=>"village", "pop"=>705, "name_ru"=>"Эссекс", "name_en"=>"Essex", "lang"=>"en", "lat"=>41.178335, "lon"=>-88.194737),
"/US/US-IL/eureka-4901" => array("type"=>"city", "pop"=>5368, "name_ru"=>"Юрика", "name_en"=>"Eureka", "lang"=>"en", "lat"=>40.715584, "lon"=>-89.275135),
"/US/US-IL/evanston-4902" => array("type"=>"city", "pop"=>77693, "name_ru"=>"Эванстон", "name_en"=>"Evanston", "lang"=>"en", "lat"=>42.046350, "lon"=>-87.694549),
"/US/US-IL/evansville-4903" => array("type"=>"village", "pop"=>676, "name_ru"=>"Эвансвилл", "name_en"=>"Evansville", "lang"=>"en", "lat"=>38.088777, "lon"=>-89.932891),
"/US/US-IL/evergreen-park-4904" => array("type"=>"village", "pop"=>19359, "name_ru"=>"Эвергрин Парк", "name_en"=>"Evergreen Park", "lang"=>"en", "lat"=>41.721100, "lon"=>-87.701108),
"/US/US-IL/ewing-4905" => array("type"=>"village", "pop"=>319, "name_ru"=>"Юинг", "name_en"=>"Ewing", "lang"=>"en", "lat"=>38.089331, "lon"=>-88.852219),
"/US/US-IL/exeter-4906" => array("type"=>"village", "pop"=>65, "name_ru"=>"Эксетер", "name_en"=>"Exeter", "lang"=>"en", "lat"=>39.719046, "lon"=>-90.495903),
"/US/US-IL/fairbury-4907" => array("type"=>"city", "pop"=>3783, "name_ru"=>"Фэрбери", "name_en"=>"Fairbury", "lang"=>"en", "lat"=>40.746587, "lon"=>-88.512430),
"/US/US-IL/fairfield-4908" => array("type"=>"city", "pop"=>5165, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>38.380358, "lon"=>-88.368968),
"/US/US-IL/fairmont-4909" => array("type"=>"cdp", "pop"=>2894, "name_ru"=>"Фэрмонт", "name_en"=>"Fairmont", "lang"=>"en", "lat"=>41.561850, "lon"=>-88.059852),
"/US/US-IL/fairmont-city-4910" => array("type"=>"village", "pop"=>2244, "name_ru"=>"Фэрмонт Сити", "name_en"=>"Fairmont City", "lang"=>"en", "lat"=>38.651000, "lon"=>-90.101499),
"/US/US-IL/fairmount-4911" => array("type"=>"village", "pop"=>626, "name_ru"=>"Фэрмаунт", "name_en"=>"Fairmount", "lang"=>"en", "lat"=>40.046097, "lon"=>-87.828582),
"/US/US-IL/fairview-4912" => array("type"=>"village", "pop"=>480, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>40.650529, "lon"=>-90.191053),
"/US/US-IL/fairview-heights-4913" => array("type"=>"city", "pop"=>16643, "name_ru"=>"Фэрвью Хайтс", "name_en"=>"Fairview Heights", "lang"=>"en", "lat"=>38.596695, "lon"=>-90.004918),
"/US/US-IL/farina-4914" => array("type"=>"village", "pop"=>552, "name_ru"=>"Фарина", "name_en"=>"Farina", "lang"=>"en", "lat"=>38.831700, "lon"=>-88.777841),
"/US/US-IL/farmer-city-4915" => array("type"=>"city", "pop"=>1951, "name_ru"=>"Фармер Сити", "name_en"=>"Farmer City", "lang"=>"en", "lat"=>40.244443, "lon"=>-88.641249),
"/US/US-IL/farmersville-4916" => array("type"=>"village", "pop"=>747, "name_ru"=>"Фармерсвилл", "name_en"=>"Farmersville", "lang"=>"en", "lat"=>39.444156, "lon"=>-89.653032),
"/US/US-IL/farmington-4917" => array("type"=>"city", "pop"=>2433, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>40.697585, "lon"=>-90.006690),
"/US/US-IL/fayetteville-4918" => array("type"=>"village", "pop"=>371, "name_ru"=>"Фейетвилл", "name_en"=>"Fayetteville", "lang"=>"en", "lat"=>38.377882, "lon"=>-89.796115),
"/US/US-IL/ferris-4919" => array("type"=>"village", "pop"=>155, "name_ru"=>"Феррис", "name_en"=>"Ferris", "lang"=>"en", "lat"=>40.469486, "lon"=>-91.169392),
"/US/US-IL/fidelity-4920" => array("type"=>"village", "pop"=>103, "name_ru"=>"Фиделити", "name_en"=>"Fidelity", "lang"=>"en", "lat"=>39.154590, "lon"=>-90.163502),
"/US/US-IL/fieldon-4921" => array("type"=>"village", "pop"=>267, "name_ru"=>"Филдон", "name_en"=>"Fieldon", "lang"=>"en", "lat"=>39.108137, "lon"=>-90.499930),
"/US/US-IL/fillmore-4922" => array("type"=>"village", "pop"=>351, "name_ru"=>"Фильмор", "name_en"=>"Fillmore", "lang"=>"en", "lat"=>39.115750, "lon"=>-89.279282),
"/US/US-IL/findlay-4923" => array("type"=>"village", "pop"=>662, "name_ru"=>"Файндлей", "name_en"=>"Findlay", "lang"=>"en", "lat"=>39.521078, "lon"=>-88.754293),
"/US/US-IL/fisher-4924" => array("type"=>"village", "pop"=>1755, "name_ru"=>"Фишер", "name_en"=>"Fisher", "lang"=>"en", "lat"=>40.315784, "lon"=>-88.348547),
"/US/US-IL/fithian-4925" => array("type"=>"village", "pop"=>545, "name_ru"=>"Фитиан", "name_en"=>"Fithian", "lang"=>"en", "lat"=>40.113954, "lon"=>-87.874718),
"/US/US-IL/flanagan-4926" => array("type"=>"village", "pop"=>1024, "name_ru"=>"Флэнаган", "name_en"=>"Flanagan", "lang"=>"en", "lat"=>40.877334, "lon"=>-88.860557),
"/US/US-IL/flat-rock-4927" => array("type"=>"village", "pop"=>399, "name_ru"=>"Флэт Рок", "name_en"=>"Flat Rock", "lang"=>"en", "lat"=>38.904396, "lon"=>-87.672501),
"/US/US-IL/flora-4928" => array("type"=>"city", "pop"=>4772, "name_ru"=>"Флора", "name_en"=>"Flora", "lang"=>"en", "lat"=>38.670486, "lon"=>-88.478772),
"/US/US-IL/florence-4929" => array("type"=>"village", "pop"=>66, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>39.628201, "lon"=>-90.610202),
"/US/US-IL/flossmoor-4930" => array("type"=>"village", "pop"=>9309, "name_ru"=>"Флосмур", "name_en"=>"Flossmoor", "lang"=>"en", "lat"=>41.539350, "lon"=>-87.684199),
"/US/US-IL/foosland-4931" => array("type"=>"village", "pop"=>85, "name_ru"=>"Фусленд", "name_en"=>"Foosland", "lang"=>"en", "lat"=>40.360424, "lon"=>-88.428725),
"/US/US-IL/ford-heights-4932" => array("type"=>"village", "pop"=>3202, "name_ru"=>"Форд Хайтс", "name_en"=>"Ford Heights", "lang"=>"en", "lat"=>41.509630, "lon"=>-87.589049),
"/US/US-IL/forest-city-4933" => array("type"=>"village", "pop"=>267, "name_ru"=>"Форест Сити", "name_en"=>"Forest City", "lang"=>"en", "lat"=>40.370964, "lon"=>-89.831912),
"/US/US-IL/forest-lake-4934" => array("type"=>"cdp", "pop"=>1371, "name_ru"=>"Форест Лейк", "name_en"=>"Forest Lake", "lang"=>"en", "lat"=>42.205600, "lon"=>-88.052199),
"/US/US-IL/forest-park-4935" => array("type"=>"village", "pop"=>15224, "name_ru"=>"Форест Парк", "name_en"=>"Forest Park", "lang"=>"en", "lat"=>41.869050, "lon"=>-87.814847),
"/US/US-IL/forest-view-4936" => array("type"=>"village", "pop"=>718, "name_ru"=>"Форест Вью", "name_en"=>"Forest View", "lang"=>"en", "lat"=>41.806550, "lon"=>-87.780983),
"/US/US-IL/forrest-4937" => array("type"=>"village", "pop"=>1152, "name_ru"=>"Форрест", "name_en"=>"Forrest", "lang"=>"en", "lat"=>40.749984, "lon"=>-88.409939),
"/US/US-IL/forreston-4938" => array("type"=>"village", "pop"=>1490, "name_ru"=>"Форрестон", "name_en"=>"Forreston", "lang"=>"en", "lat"=>42.126126, "lon"=>-89.578701),
"/US/US-IL/forsyth-4939" => array("type"=>"village", "pop"=>3002, "name_ru"=>"Форсайт", "name_en"=>"Forsyth", "lang"=>"en", "lat"=>39.923608, "lon"=>-88.955887),
"/US/US-IL/fox-lake-4940" => array("type"=>"village", "pop"=>11087, "name_ru"=>"Фокс Лейк", "name_en"=>"Fox Lake", "lang"=>"en", "lat"=>42.418132, "lon"=>-88.183337),
"/US/US-IL/fox-lake-hills-4941" => array("type"=>"cdp", "pop"=>2681, "name_ru"=>"Фокс Лейк Хиллс", "name_en"=>"Fox Lake Hills", "lang"=>"en", "lat"=>42.407122, "lon"=>-88.123499),
"/US/US-IL/fox-river-grove-4942" => array("type"=>"village", "pop"=>5172, "name_ru"=>"Фокс Ривер Гров", "name_en"=>"Fox River Grove", "lang"=>"en", "lat"=>42.197000, "lon"=>-88.218958),
"/US/US-IL/fox-river-valley-gardens-4943" => array("type"=>"village", "pop"=>665, "name_ru"=>"Фокс Ривер Валли Гарденс", "name_en"=>"Fox River Valley Gardens", "lang"=>"en", "lat"=>42.244367, "lon"=>-88.194924),
"/US/US-IL/frankfort-4944" => array("type"=>"village", "pop"=>17981, "name_ru"=>"Франкфорт", "name_en"=>"Frankfort", "lang"=>"en", "lat"=>41.496986, "lon"=>-87.835935),
"/US/US-IL/frankfort-square-4945" => array("type"=>"cdp", "pop"=>6227, "name_ru"=>"Франкфорт Сквэр", "name_en"=>"Frankfort Square", "lang"=>"en", "lat"=>41.518900, "lon"=>-87.803142),
"/US/US-IL/franklin-4946" => array("type"=>"village", "pop"=>597, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>39.620362, "lon"=>-90.047204),
"/US/US-IL/franklin-grove-4947" => array("type"=>"village", "pop"=>1000, "name_ru"=>"Франклин Гров", "name_en"=>"Franklin Grove", "lang"=>"en", "lat"=>41.841020, "lon"=>-89.300136),
"/US/US-IL/franklin-park-4948" => array("type"=>"village", "pop"=>17965, "name_ru"=>"Франклин Парк", "name_en"=>"Franklin Park", "lang"=>"en", "lat"=>41.935400, "lon"=>-87.878176),
"/US/US-IL/freeburg-4949" => array("type"=>"village", "pop"=>4481, "name_ru"=>"Фрибург", "name_en"=>"Freeburg", "lang"=>"en", "lat"=>38.427850, "lon"=>-89.908124),
"/US/US-IL/freeman-spur-4950" => array("type"=>"village", "pop"=>278, "name_ru"=>"Фримэн Спер", "name_en"=>"Freeman Spur", "lang"=>"en", "lat"=>37.859922, "lon"=>-88.999806),
"/US/US-IL/freeport-4951" => array("type"=>"city", "pop"=>24615, "name_ru"=>"Фрипорт", "name_en"=>"Freeport", "lang"=>"en", "lat"=>42.290155, "lon"=>-89.634690),
"/US/US-IL/fulton-4952" => array("type"=>"city", "pop"=>3832, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>41.865027, "lon"=>-90.158725),
"/US/US-IL/fults-4953" => array("type"=>"village", "pop"=>29, "name_ru"=>"Фултс", "name_en"=>"Fults", "lang"=>"en", "lat"=>38.163743, "lon"=>-90.212572),
"/US/US-IL/gages-lake-4954" => array("type"=>"cdp", "pop"=>8349, "name_ru"=>"Гейджс Лейк", "name_en"=>"Gages Lake", "lang"=>"en", "lat"=>42.351600, "lon"=>-87.982441),
"/US/US-IL/galatia-4955" => array("type"=>"village", "pop"=>983, "name_ru"=>"Галатия", "name_en"=>"Galatia", "lang"=>"en", "lat"=>37.841103, "lon"=>-88.613879),
"/US/US-IL/galena-4956" => array("type"=>"city", "pop"=>3333, "name_ru"=>"Галена", "name_en"=>"Galena", "lang"=>"en", "lat"=>42.420750, "lon"=>-90.427345),
"/US/US-IL/galesburg-4957" => array("type"=>"city", "pop"=>31181, "name_ru"=>"Гейлсбург", "name_en"=>"Galesburg", "lang"=>"en", "lat"=>40.950043, "lon"=>-90.377932),
"/US/US-IL/galva-4958" => array("type"=>"city", "pop"=>2648, "name_ru"=>"Галва", "name_en"=>"Galva", "lang"=>"en", "lat"=>41.166594, "lon"=>-90.038910),
"/US/US-IL/gardner-4959" => array("type"=>"village", "pop"=>1489, "name_ru"=>"Гарднер", "name_en"=>"Gardner", "lang"=>"en", "lat"=>41.187734, "lon"=>-88.312156),
"/US/US-IL/garrett-4960" => array("type"=>"village", "pop"=>191, "name_ru"=>"Гарретт", "name_en"=>"Garrett", "lang"=>"en", "lat"=>39.796819, "lon"=>-88.425329),
"/US/US-IL/gays-4961" => array("type"=>"village", "pop"=>259, "name_ru"=>"Гейс", "name_en"=>"Gays", "lang"=>"en", "lat"=>39.457876, "lon"=>-88.495942),
"/US/US-IL/geneseo-4962" => array("type"=>"city", "pop"=>6437, "name_ru"=>"Дженесео", "name_en"=>"Geneseo", "lang"=>"en", "lat"=>41.447194, "lon"=>-90.154507),
"/US/US-IL/geneva-4963" => array("type"=>"city", "pop"=>24302, "name_ru"=>"Дженива", "name_en"=>"Geneva", "lang"=>"en", "lat"=>41.882327, "lon"=>-88.314486),
"/US/US-IL/genoa-4964" => array("type"=>"city", "pop"=>5101, "name_ru"=>"Дженоа", "name_en"=>"Genoa", "lang"=>"en", "lat"=>42.097648, "lon"=>-88.689890),
"/US/US-IL/georgetown-4965" => array("type"=>"city", "pop"=>3440, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>39.976455, "lon"=>-87.634985),
"/US/US-IL/german-valley-4968" => array("type"=>"village", "pop"=>450, "name_ru"=>"Джермэн Валли", "name_en"=>"German Valley", "lang"=>"en", "lat"=>42.215843, "lon"=>-89.483794),
"/US/US-IL/germantown-4966" => array("type"=>"village", "pop"=>1232, "name_ru"=>"Джермэнтаун", "name_en"=>"Germantown", "lang"=>"en", "lat"=>38.552266, "lon"=>-89.537583),
"/US/US-IL/germantown-hills-4967" => array("type"=>"village", "pop"=>3213, "name_ru"=>"Джермэнтаун Хилс", "name_en"=>"Germantown Hills", "lang"=>"en", "lat"=>40.768933, "lon"=>-89.467958),
"/US/US-IL/gibson-4969" => array("type"=>"city", "pop"=>3396, "name_ru"=>"Гибсон", "name_en"=>"Gibson", "lang"=>"en", "lat"=>40.465654, "lon"=>-88.378591),
"/US/US-IL/gifford-4970" => array("type"=>"village", "pop"=>993, "name_ru"=>"Гиффорд", "name_en"=>"Gifford", "lang"=>"en", "lat"=>40.306413, "lon"=>-88.021634),
"/US/US-IL/gilberts-4971" => array("type"=>"village", "pop"=>5996, "name_ru"=>"Гилбертс", "name_en"=>"Gilberts", "lang"=>"en", "lat"=>42.101650, "lon"=>-88.367199),
"/US/US-IL/gillespie-4972" => array("type"=>"city", "pop"=>3182, "name_ru"=>"Гиллеспи", "name_en"=>"Gillespie", "lang"=>"en", "lat"=>39.125357, "lon"=>-89.817173),
"/US/US-IL/gilman-4973" => array("type"=>"city", "pop"=>1714, "name_ru"=>"Гилман", "name_en"=>"Gilman", "lang"=>"en", "lat"=>40.765942, "lon"=>-87.996133),
"/US/US-IL/girard-4974" => array("type"=>"city", "pop"=>2166, "name_ru"=>"Жирар", "name_en"=>"Girard", "lang"=>"en", "lat"=>39.446834, "lon"=>-89.782360),
"/US/US-IL/gladstone-4975" => array("type"=>"village", "pop"=>255, "name_ru"=>"Глэдстоун", "name_en"=>"Gladstone", "lang"=>"en", "lat"=>40.863684, "lon"=>-90.957494),
"/US/US-IL/glasford-4976" => array("type"=>"village", "pop"=>1033, "name_ru"=>"Гласфорд", "name_en"=>"Glasford", "lang"=>"en", "lat"=>40.572316, "lon"=>-89.813174),
"/US/US-IL/glasgow-4977" => array("type"=>"village", "pop"=>159, "name_ru"=>"Гласго", "name_en"=>"Glasgow", "lang"=>"en", "lat"=>39.549678, "lon"=>-90.479572),
"/US/US-IL/glen-carbon-4979" => array("type"=>"village", "pop"=>12607, "name_ru"=>"Глен Карбон", "name_en"=>"Glen Carbon", "lang"=>"en", "lat"=>38.760476, "lon"=>-89.973014),
"/US/US-IL/glen-ellyn-4982" => array("type"=>"village", "pop"=>27142, "name_ru"=>"Глен Эллин", "name_en"=>"Glen Ellyn", "lang"=>"en", "lat"=>41.866550, "lon"=>-88.062599),
"/US/US-IL/glenbard-south-4978" => array("type"=>"cdp", "pop"=>3957, "name_ru"=>"Гленбард Саут", "name_en"=>"Glenbard South", "lang"=>"en", "lat"=>41.829700, "lon"=>-88.066355),
"/US/US-IL/glencoe-4980" => array("type"=>"village", "pop"=>9038, "name_ru"=>"Гленко", "name_en"=>"Glencoe", "lang"=>"en", "lat"=>42.134600, "lon"=>-87.763431),
"/US/US-IL/glendale-heights-4981" => array("type"=>"village", "pop"=>31847, "name_ru"=>"Глендейл Хайтс", "name_en"=>"Glendale Heights", "lang"=>"en", "lat"=>41.920200, "lon"=>-88.078849),
"/US/US-IL/glenview-4983" => array("type"=>"village", "pop"=>46180, "name_ru"=>"Гленвью", "name_en"=>"Glenview", "lang"=>"en", "lat"=>42.079648, "lon"=>-87.815191),
"/US/US-IL/glenwood-4984" => array("type"=>"village", "pop"=>8489, "name_ru"=>"Гленвуд", "name_en"=>"Glenwood", "lang"=>"en", "lat"=>41.546067, "lon"=>-87.617016),
"/US/US-IL/godfrey-4985" => array("type"=>"cdp", "pop"=>17524, "name_ru"=>"Годфри", "name_en"=>"Godfrey", "lang"=>"en", "lat"=>38.966700, "lon"=>-90.188329),
"/US/US-IL/godley-4986" => array("type"=>"village", "pop"=>703, "name_ru"=>"Годли", "name_en"=>"Godley", "lang"=>"en", "lat"=>41.237272, "lon"=>-88.244611),
"/US/US-IL/golconda-4987" => array("type"=>"city", "pop"=>653, "name_ru"=>"Голконда", "name_en"=>"Golconda", "lang"=>"en", "lat"=>37.362740, "lon"=>-88.486740),
"/US/US-IL/golden-4988" => array("type"=>"village", "pop"=>616, "name_ru"=>"Гоулден", "name_en"=>"Golden", "lang"=>"en", "lat"=>40.109726, "lon"=>-91.018413),
"/US/US-IL/golden-gate-4989" => array("type"=>"village", "pop"=>96, "name_ru"=>"Голден Гейт", "name_en"=>"Golden Gate", "lang"=>"en", "lat"=>38.358637, "lon"=>-88.205529),
"/US/US-IL/golf-4990" => array("type"=>"village", "pop"=>445, "name_ru"=>"Гольф", "name_en"=>"Golf", "lang"=>"en", "lat"=>42.058800, "lon"=>-87.786525),
"/US/US-IL/good-hope-4992" => array("type"=>"village", "pop"=>375, "name_ru"=>"Гуд Хоп", "name_en"=>"Good Hope", "lang"=>"en", "lat"=>40.557593, "lon"=>-90.674877),
"/US/US-IL/goodfield-4991" => array("type"=>"village", "pop"=>844, "name_ru"=>"Гудфилд", "name_en"=>"Goodfield", "lang"=>"en", "lat"=>40.629459, "lon"=>-89.271219),
"/US/US-IL/goodings-grove-4993" => array("type"=>"cdp", "pop"=>14054, "name_ru"=>"Гудингс Гров", "name_en"=>"Goodings Grove", "lang"=>"en", "lat"=>41.621600, "lon"=>-87.940594),
"/US/US-IL/goreville-4994" => array("type"=>"village", "pop"=>972, "name_ru"=>"Горвилл", "name_en"=>"Goreville", "lang"=>"en", "lat"=>37.555073, "lon"=>-88.973617),
"/US/US-IL/gorham-4995" => array("type"=>"village", "pop"=>245, "name_ru"=>"Горам", "name_en"=>"Gorham", "lang"=>"en", "lat"=>37.717806, "lon"=>-89.482577),
"/US/US-IL/grafton-4996" => array("type"=>"city", "pop"=>725, "name_ru"=>"Графтон", "name_en"=>"Grafton", "lang"=>"en", "lat"=>38.972508, "lon"=>-90.427000),
"/US/US-IL/grand-ridge-4997" => array("type"=>"village", "pop"=>520, "name_ru"=>"Гранд Ридж", "name_en"=>"Grand Ridge", "lang"=>"en", "lat"=>41.236168, "lon"=>-88.831464),
"/US/US-IL/grand-tower-4998" => array("type"=>"city", "pop"=>579, "name_ru"=>"Гранд Тауер", "name_en"=>"Grand Tower", "lang"=>"en", "lat"=>37.639600, "lon"=>-89.503313),
"/US/US-IL/grandview-4999" => array("type"=>"village", "pop"=>1410, "name_ru"=>"Грандвью", "name_en"=>"Grandview", "lang"=>"en", "lat"=>39.817800, "lon"=>-89.617024),
"/US/US-IL/grandwood-park-5000" => array("type"=>"cdp", "pop"=>2470, "name_ru"=>"Грандвуд Парк", "name_en"=>"Grandwood Park", "lang"=>"en", "lat"=>42.393153, "lon"=>-87.986688),
"/US/US-IL/granite-city-5001" => array("type"=>"city", "pop"=>30703, "name_ru"=>"Гранит Сити", "name_en"=>"Granite City", "lang"=>"en", "lat"=>38.714400, "lon"=>-90.129655),
"/US/US-IL/grant-park-5003" => array("type"=>"village", "pop"=>1718, "name_ru"=>"Грант Парк", "name_en"=>"Grant Park", "lang"=>"en", "lat"=>41.239250, "lon"=>-87.644646),
"/US/US-IL/grantfork-5002" => array("type"=>"village", "pop"=>327, "name_ru"=>"Грантфорк", "name_en"=>"Grantfork", "lang"=>"en", "lat"=>38.830873, "lon"=>-89.667787),
"/US/US-IL/granville-5004" => array("type"=>"village", "pop"=>1343, "name_ru"=>"Гранвилл", "name_en"=>"Granville", "lang"=>"en", "lat"=>41.264275, "lon"=>-89.229862),
"/US/US-IL/grayslake-5005" => array("type"=>"village", "pop"=>21735, "name_ru"=>"Грейслейк", "name_en"=>"Grayslake", "lang"=>"en", "lat"=>42.349800, "lon"=>-88.038536),
"/US/US-IL/grayville-5006" => array("type"=>"city", "pop"=>1585, "name_ru"=>"Грейвилл", "name_en"=>"Grayville", "lang"=>"en", "lat"=>38.259819, "lon"=>-87.996878),
"/US/US-IL/green-oaks-5008" => array("type"=>"village", "pop"=>3913, "name_ru"=>"Грин Оукс", "name_en"=>"Green Oaks", "lang"=>"en", "lat"=>42.301746, "lon"=>-87.915156),
"/US/US-IL/green-rock-5009" => array("type"=>"city", "pop"=>2615, "name_ru"=>"Грин Рок", "name_en"=>"Green Rock", "lang"=>"en", "lat"=>41.474860, "lon"=>-90.363616),
"/US/US-IL/green-valley-5011" => array("type"=>"village", "pop"=>694, "name_ru"=>"Грин Валли", "name_en"=>"Green Valley", "lang"=>"en", "lat"=>40.406429, "lon"=>-89.642452),
"/US/US-IL/greenfield-5007" => array("type"=>"city", "pop"=>1073, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>39.344323, "lon"=>-90.207606),
"/US/US-IL/greenup-5010" => array("type"=>"village", "pop"=>1477, "name_ru"=>"Гринэп", "name_en"=>"Greenup", "lang"=>"en", "lat"=>39.247818, "lon"=>-88.159890),
"/US/US-IL/greenview-5012" => array("type"=>"village", "pop"=>789, "name_ru"=>"Гринвью", "name_en"=>"Greenview", "lang"=>"en", "lat"=>40.084790, "lon"=>-89.740183),
"/US/US-IL/greenville-5013" => array("type"=>"city", "pop"=>7379, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>38.889298, "lon"=>-89.401640),
"/US/US-IL/gridley-5014" => array("type"=>"village", "pop"=>1429, "name_ru"=>"Гридли", "name_en"=>"Gridley", "lang"=>"en", "lat"=>40.743892, "lon"=>-88.880412),
"/US/US-IL/griggsville-5015" => array("type"=>"city", "pop"=>1172, "name_ru"=>"Григсвилл", "name_en"=>"Griggsville", "lang"=>"en", "lat"=>39.707864, "lon"=>-90.726954),
"/US/US-IL/gulf-port-5016" => array("type"=>"village", "pop"=>214, "name_ru"=>"Галф Порт", "name_en"=>"Gulf Port", "lang"=>"en", "lat"=>40.808950, "lon"=>-91.083589),
"/US/US-IL/gurnee-5017" => array("type"=>"village", "pop"=>30608, "name_ru"=>"Герни", "name_en"=>"Gurnee", "lang"=>"en", "lat"=>42.377092, "lon"=>-87.935973),
"/US/US-IL/hainesville-5018" => array("type"=>"village", "pop"=>3699, "name_ru"=>"Хейнсвилл", "name_en"=>"Hainesville", "lang"=>"en", "lat"=>42.339673, "lon"=>-88.064705),
"/US/US-IL/hamburg-5019" => array("type"=>"village", "pop"=>122, "name_ru"=>"Гамбург", "name_en"=>"Hamburg", "lang"=>"en", "lat"=>39.232785, "lon"=>-90.715599),
"/US/US-IL/hamel-5020" => array("type"=>"village", "pop"=>769, "name_ru"=>"Хамел", "name_en"=>"Hamel", "lang"=>"en", "lat"=>38.888587, "lon"=>-89.842670),
"/US/US-IL/hamilton-5021" => array("type"=>"city", "pop"=>2777, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>40.389121, "lon"=>-91.363221),
"/US/US-IL/hamletsburg-5022" => array("type"=>"village", "pop"=>85, "name_ru"=>"Хамлетсберг", "name_en"=>"Hamletsburg", "lang"=>"en", "lat"=>37.140809, "lon"=>-88.445957),
"/US/US-IL/hammond-5023" => array("type"=>"village", "pop"=>508, "name_ru"=>"Хаммонд", "name_en"=>"Hammond", "lang"=>"en", "lat"=>39.798522, "lon"=>-88.591537),
"/US/US-IL/hampshire-5024" => array("type"=>"village", "pop"=>5569, "name_ru"=>"Хэмпшир", "name_en"=>"Hampshire", "lang"=>"en", "lat"=>42.095792, "lon"=>-88.521725),
"/US/US-IL/hampton-5025" => array("type"=>"village", "pop"=>1774, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>41.553317, "lon"=>-90.405615),
"/US/US-IL/hanaford-5026" => array("type"=>"village", "pop"=>221, "name_ru"=>"Ханафорд", "name_en"=>"Hanaford", "lang"=>"en", "lat"=>37.958497, "lon"=>-88.835292),
"/US/US-IL/hanna-city-5027" => array("type"=>"village", "pop"=>977, "name_ru"=>"Ханна Сити", "name_en"=>"Hanna City", "lang"=>"en", "lat"=>40.692224, "lon"=>-89.793052),
"/US/US-IL/hanover-5028" => array("type"=>"village", "pop"=>793, "name_ru"=>"Хановер", "name_en"=>"Hanover", "lang"=>"en", "lat"=>42.254213, "lon"=>-90.278694),
"/US/US-IL/hanover-park-5029" => array("type"=>"village", "pop"=>36777, "name_ru"=>"Хановер Парк", "name_en"=>"Hanover Park", "lang"=>"en", "lat"=>41.984577, "lon"=>-88.144785),
"/US/US-IL/hardin-5030" => array("type"=>"village", "pop"=>935, "name_ru"=>"Хардин", "name_en"=>"Hardin", "lang"=>"en", "lat"=>39.157355, "lon"=>-90.623097),
"/US/US-IL/harmon-5031" => array("type"=>"village", "pop"=>144, "name_ru"=>"Хармон", "name_en"=>"Harmon", "lang"=>"en", "lat"=>41.719689, "lon"=>-89.556564),
"/US/US-IL/harrisburg-5032" => array("type"=>"city", "pop"=>9582, "name_ru"=>"Харрисбург", "name_en"=>"Harrisburg", "lang"=>"en", "lat"=>37.736476, "lon"=>-88.549251),
"/US/US-IL/harristown-5033" => array("type"=>"village", "pop"=>1244, "name_ru"=>"Харристаун", "name_en"=>"Harristown", "lang"=>"en", "lat"=>39.841608, "lon"=>-89.059525),
"/US/US-IL/hartford-5034" => array("type"=>"village", "pop"=>1477, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>38.828000, "lon"=>-90.091829),
"/US/US-IL/hartsburg-5035" => array("type"=>"village", "pop"=>330, "name_ru"=>"Хартсбург", "name_en"=>"Hartsburg", "lang"=>"en", "lat"=>40.250721, "lon"=>-89.440704),
"/US/US-IL/harvard-5036" => array("type"=>"city", "pop"=>9855, "name_ru"=>"Харвард", "name_en"=>"Harvard", "lang"=>"en", "lat"=>42.421163, "lon"=>-88.615282),
"/US/US-IL/harvel-5037" => array("type"=>"village", "pop"=>227, "name_ru"=>"Харвел", "name_en"=>"Harvel", "lang"=>"en", "lat"=>39.356963, "lon"=>-89.531073),
"/US/US-IL/harvey-5038" => array("type"=>"city", "pop"=>28056, "name_ru"=>"Харви", "name_en"=>"Harvey", "lang"=>"en", "lat"=>41.607250, "lon"=>-87.651896),
"/US/US-IL/harwood-heights-5039" => array("type"=>"village", "pop"=>8058, "name_ru"=>"Харвуд Хайтс", "name_en"=>"Harwood Heights", "lang"=>"en", "lat"=>41.966138, "lon"=>-87.805509),
"/US/US-IL/havana-5040" => array("type"=>"city", "pop"=>3328, "name_ru"=>"Гавана", "name_en"=>"Havana", "lang"=>"en", "lat"=>40.298431, "lon"=>-90.056834),
"/US/US-IL/hawthorn-woods-5041" => array("type"=>"village", "pop"=>8067, "name_ru"=>"Хоторн Вудс", "name_en"=>"Hawthorn Woods", "lang"=>"en", "lat"=>42.223600, "lon"=>-88.057323),
"/US/US-IL/hazel-crest-5042" => array("type"=>"village", "pop"=>14075, "name_ru"=>"Хэйзл Крест", "name_en"=>"Hazel Crest", "lang"=>"en", "lat"=>41.572300, "lon"=>-87.689399),
"/US/US-IL/hebron-5043" => array("type"=>"village", "pop"=>1401, "name_ru"=>"Хеброн", "name_en"=>"Hebron", "lang"=>"en", "lat"=>42.470950, "lon"=>-88.429865),
"/US/US-IL/hecker-5044" => array("type"=>"village", "pop"=>489, "name_ru"=>"Хэкер", "name_en"=>"Hecker", "lang"=>"en", "lat"=>38.304530, "lon"=>-89.993993),
"/US/US-IL/henderson-5045" => array("type"=>"village", "pop"=>295, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>41.024460, "lon"=>-90.352866),
"/US/US-IL/hennepin-5046" => array("type"=>"village", "pop"=>704, "name_ru"=>"Хеннепин", "name_en"=>"Hennepin", "lang"=>"en", "lat"=>41.255975, "lon"=>-89.319615),
"/US/US-IL/henning-5047" => array("type"=>"village", "pop"=>229, "name_ru"=>"Хеннинг", "name_en"=>"Henning", "lang"=>"en", "lat"=>40.306246, "lon"=>-87.700665),
"/US/US-IL/henry-5048" => array("type"=>"city", "pop"=>2415, "name_ru"=>"Хенри", "name_en"=>"Henry", "lang"=>"en", "lat"=>41.112328, "lon"=>-89.361795),
"/US/US-IL/herrick-5049" => array("type"=>"village", "pop"=>498, "name_ru"=>"Херрик", "name_en"=>"Herrick", "lang"=>"en", "lat"=>39.219161, "lon"=>-88.985296),
"/US/US-IL/herrin-5050" => array("type"=>"city", "pop"=>12337, "name_ru"=>"Херрин", "name_en"=>"Herrin", "lang"=>"en", "lat"=>37.801552, "lon"=>-89.031646),
"/US/US-IL/herscher-5051" => array("type"=>"village", "pop"=>1615, "name_ru"=>"Хершер", "name_en"=>"Herscher", "lang"=>"en", "lat"=>41.049121, "lon"=>-88.100596),
"/US/US-IL/hettick-5052" => array("type"=>"village", "pop"=>175, "name_ru"=>"Хеттик", "name_en"=>"Hettick", "lang"=>"en", "lat"=>39.355188, "lon"=>-90.037039),
"/US/US-IL/heyworth-5053" => array("type"=>"village", "pop"=>2602, "name_ru"=>"Хейворт", "name_en"=>"Heyworth", "lang"=>"en", "lat"=>40.313788, "lon"=>-88.978816),
"/US/US-IL/hickory-hills-5054" => array("type"=>"city", "pop"=>13326, "name_ru"=>"Хикори Хилс", "name_en"=>"Hickory Hills", "lang"=>"en", "lat"=>41.723750, "lon"=>-87.828200),
"/US/US-IL/hidalgo-5055" => array("type"=>"village", "pop"=>118, "name_ru"=>"Идальго", "name_en"=>"Hidalgo", "lang"=>"en", "lat"=>39.156589, "lon"=>-88.150659),
"/US/US-IL/highland-5056" => array("type"=>"city", "pop"=>9748, "name_ru"=>"Хайленд", "name_en"=>"Highland", "lang"=>"en", "lat"=>38.740999, "lon"=>-89.675819),
"/US/US-IL/highland-park-5057" => array("type"=>"city", "pop"=>31449, "name_ru"=>"Хайленд Парк", "name_en"=>"Highland Park", "lang"=>"en", "lat"=>42.182700, "lon"=>-87.809653),
"/US/US-IL/highwood-5058" => array("type"=>"city", "pop"=>5388, "name_ru"=>"Хайвуд", "name_en"=>"Highwood", "lang"=>"en", "lat"=>42.211604, "lon"=>-87.811705),
"/US/US-IL/hillcrest-5059" => array("type"=>"village", "pop"=>1288, "name_ru"=>"Хиллкрест", "name_en"=>"Hillcrest", "lang"=>"en", "lat"=>41.952484, "lon"=>-89.070372),
"/US/US-IL/hillsboro-5060" => array("type"=>"city", "pop"=>6164, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>39.163121, "lon"=>-89.484389),
"/US/US-IL/hillsdale-5061" => array("type"=>"village", "pop"=>561, "name_ru"=>"Хиллсдейл", "name_en"=>"Hillsdale", "lang"=>"en", "lat"=>41.611750, "lon"=>-90.175171),
"/US/US-IL/hillside-5062" => array("type"=>"village", "pop"=>8341, "name_ru"=>"Хиллсайд", "name_en"=>"Hillside", "lang"=>"en", "lat"=>41.870967, "lon"=>-87.898603),
"/US/US-IL/hillview-5063" => array("type"=>"village", "pop"=>166, "name_ru"=>"Хиллвью", "name_en"=>"Hillview", "lang"=>"en", "lat"=>39.449605, "lon"=>-90.537859),
"/US/US-IL/hinckley-5064" => array("type"=>"village", "pop"=>2081, "name_ru"=>"Хинкли", "name_en"=>"Hinckley", "lang"=>"en", "lat"=>41.769702, "lon"=>-88.641754),
"/US/US-IL/hindsboro-5065" => array("type"=>"village", "pop"=>334, "name_ru"=>"Хиндсборо", "name_en"=>"Hindsboro", "lang"=>"en", "lat"=>39.684302, "lon"=>-88.134208),
"/US/US-IL/hinsdale-5066" => array("type"=>"village", "pop"=>18439, "name_ru"=>"Хинсдейл", "name_en"=>"Hinsdale", "lang"=>"en", "lat"=>41.800550, "lon"=>-87.928275),
"/US/US-IL/hodgkins-5067" => array("type"=>"village", "pop"=>2014, "name_ru"=>"Ходжкинс", "name_en"=>"Hodgkins", "lang"=>"en", "lat"=>41.770800, "lon"=>-87.856258),
"/US/US-IL/hoffman-5068" => array("type"=>"village", "pop"=>449, "name_ru"=>"Хоффман", "name_en"=>"Hoffman", "lang"=>"en", "lat"=>38.538449, "lon"=>-89.262851),
"/US/US-IL/hoffman-estates-5069" => array("type"=>"village", "pop"=>53641, "name_ru"=>"Хоффман Истейтс", "name_en"=>"Hoffman Estates", "lang"=>"en", "lat"=>42.064550, "lon"=>-88.138303),
"/US/US-IL/holiday-hills-5070" => array("type"=>"village", "pop"=>772, "name_ru"=>"Холидей Хилс", "name_en"=>"Holiday Hills", "lang"=>"en", "lat"=>42.291300, "lon"=>-88.226885),
"/US/US-IL/hollowayville-5071" => array("type"=>"village", "pop"=>87, "name_ru"=>"Холлоуэйвилл", "name_en"=>"Hollowayville", "lang"=>"en", "lat"=>41.364917, "lon"=>-89.294538),
"/US/US-IL/homer-5072" => array("type"=>"village", "pop"=>1122, "name_ru"=>"Гомер", "name_en"=>"Homer", "lang"=>"en", "lat"=>40.031885, "lon"=>-87.958930),
"/US/US-IL/hometown-5073" => array("type"=>"city", "pop"=>4112, "name_ru"=>"Хоумтаун", "name_en"=>"Hometown", "lang"=>"en", "lat"=>41.732233, "lon"=>-87.731252),
"/US/US-IL/homewood-5074" => array("type"=>"village", "pop"=>18475, "name_ru"=>"Хомвуд", "name_en"=>"Homewood", "lang"=>"en", "lat"=>41.558400, "lon"=>-87.661124),
"/US/US-IL/hoopeston-5075" => array("type"=>"city", "pop"=>5656, "name_ru"=>"Хупестон", "name_en"=>"Hoopeston", "lang"=>"en", "lat"=>40.465183, "lon"=>-87.670882),
"/US/US-IL/hooppole-5076" => array("type"=>"village", "pop"=>152, "name_ru"=>"Хуппоул", "name_en"=>"Hooppole", "lang"=>"en", "lat"=>41.520883, "lon"=>-89.913298),
"/US/US-IL/hopedale-5077" => array("type"=>"village", "pop"=>924, "name_ru"=>"Хоупдейл", "name_en"=>"Hopedale", "lang"=>"en", "lat"=>40.421817, "lon"=>-89.416339),
"/US/US-IL/hopewell-5078" => array("type"=>"village", "pop"=>439, "name_ru"=>"Хоупуэлл", "name_en"=>"Hopewell", "lang"=>"en", "lat"=>40.983263, "lon"=>-89.456346),
"/US/US-IL/hopkins-park-5079" => array("type"=>"village", "pop"=>791, "name_ru"=>"Хопкинс Парк", "name_en"=>"Hopkins Park", "lang"=>"en", "lat"=>41.072249, "lon"=>-87.614745),
"/US/US-IL/hoyleton-5080" => array("type"=>"village", "pop"=>492, "name_ru"=>"Хойлтон", "name_en"=>"Hoyleton", "lang"=>"en", "lat"=>38.445279, "lon"=>-89.271552),
"/US/US-IL/hudson-5081" => array("type"=>"village", "pop"=>1794, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>40.607008, "lon"=>-88.988603),
"/US/US-IL/huey-5082" => array("type"=>"village", "pop"=>185, "name_ru"=>"Хаи", "name_en"=>"Huey", "lang"=>"en", "lat"=>38.605354, "lon"=>-89.291423),
"/US/US-IL/hull-5083" => array("type"=>"village", "pop"=>436, "name_ru"=>"Халл", "name_en"=>"Hull", "lang"=>"en", "lat"=>39.709115, "lon"=>-91.204976),
"/US/US-IL/humboldt-5084" => array("type"=>"village", "pop"=>466, "name_ru"=>"Гумбольдт", "name_en"=>"Humboldt", "lang"=>"en", "lat"=>39.605056, "lon"=>-88.319542),
"/US/US-IL/hume-5085" => array("type"=>"village", "pop"=>362, "name_ru"=>"Хум", "name_en"=>"Hume", "lang"=>"en", "lat"=>39.797627, "lon"=>-87.868827),
"/US/US-IL/huntley-5086" => array("type"=>"village", "pop"=>22923, "name_ru"=>"Хантли", "name_en"=>"Huntley", "lang"=>"en", "lat"=>42.168100, "lon"=>-88.424215),
"/US/US-IL/hurst-5087" => array("type"=>"city", "pop"=>787, "name_ru"=>"Херст", "name_en"=>"Hurst", "lang"=>"en", "lat"=>37.835634, "lon"=>-89.143384),
"/US/US-IL/hutsonville-5088" => array("type"=>"village", "pop"=>606, "name_ru"=>"Хатсонвилл", "name_en"=>"Hutsonville", "lang"=>"en", "lat"=>39.108695, "lon"=>-87.660631),
"/US/US-IL/illiopolis-5089" => array("type"=>"village", "pop"=>849, "name_ru"=>"Иллиополис", "name_en"=>"Illiopolis", "lang"=>"en", "lat"=>39.851309, "lon"=>-89.247047),
"/US/US-IL/ina-5090" => array("type"=>"village", "pop"=>2448, "name_ru"=>"Ина", "name_en"=>"Ina", "lang"=>"en", "lat"=>38.152175, "lon"=>-88.903721),
"/US/US-IL/indian-creek-5091" => array("type"=>"village", "pop"=>602, "name_ru"=>"Индиан Крик", "name_en"=>"Indian Creek", "lang"=>"en", "lat"=>42.225852, "lon"=>-87.976765),
"/US/US-IL/indian-head-park-5092" => array("type"=>"village", "pop"=>3614, "name_ru"=>"Индиан Хэд Парк", "name_en"=>"Indian Head Park", "lang"=>"en", "lat"=>41.767567, "lon"=>-87.897470),
"/US/US-IL/indianola-5093" => array("type"=>"village", "pop"=>220, "name_ru"=>"Индианола", "name_en"=>"Indianola", "lang"=>"en", "lat"=>39.927138, "lon"=>-87.740197),
"/US/US-IL/industry-5094" => array("type"=>"village", "pop"=>491, "name_ru"=>"Индастри", "name_en"=>"Industry", "lang"=>"en", "lat"=>40.327226, "lon"=>-90.607996),
"/US/US-IL/ingalls-park-5095" => array("type"=>"cdp", "pop"=>3173, "name_ru"=>"Инголс Парк", "name_en"=>"Ingalls Park", "lang"=>"en", "lat"=>41.521300, "lon"=>-88.033882),
"/US/US-IL/inverness-5096" => array("type"=>"village", "pop"=>7673, "name_ru"=>"Инвернесс", "name_en"=>"Inverness", "lang"=>"en", "lat"=>42.115100, "lon"=>-88.100773),
"/US/US-IL/iola-5097" => array("type"=>"village", "pop"=>164, "name_ru"=>"Айола", "name_en"=>"Iola", "lang"=>"en", "lat"=>38.834574, "lon"=>-88.627734),
"/US/US-IL/ipava-5098" => array("type"=>"village", "pop"=>470, "name_ru"=>"Ипава", "name_en"=>"Ipava", "lang"=>"en", "lat"=>40.352081, "lon"=>-90.323220),
"/US/US-IL/iroquois-5099" => array("type"=>"village", "pop"=>196, "name_ru"=>"Ирокез", "name_en"=>"Iroquois", "lang"=>"en", "lat"=>40.828738, "lon"=>-87.584737),
"/US/US-IL/irving-5100" => array("type"=>"village", "pop"=>479, "name_ru"=>"Ирвинг", "name_en"=>"Irving", "lang"=>"en", "lat"=>39.204949, "lon"=>-89.405989),
"/US/US-IL/irvington-5101" => array("type"=>"village", "pop"=>700, "name_ru"=>"Ирвингтон", "name_en"=>"Irvington", "lang"=>"en", "lat"=>38.437126, "lon"=>-89.162440),
"/US/US-IL/irwin-5102" => array("type"=>"village", "pop"=>89, "name_ru"=>"Ирвин", "name_en"=>"Irwin", "lang"=>"en", "lat"=>41.052533, "lon"=>-87.983962),
"/US/US-IL/island-lake-5103" => array("type"=>"village", "pop"=>8588, "name_ru"=>"Айленд Лейк", "name_en"=>"Island Lake", "lang"=>"en", "lat"=>42.276033, "lon"=>-88.201527),
"/US/US-IL/itasca-5104" => array("type"=>"village", "pop"=>8617, "name_ru"=>"Итаска", "name_en"=>"Itasca", "lang"=>"en", "lat"=>41.976700, "lon"=>-88.018486),
"/US/US-IL/iuka-5105" => array("type"=>"village", "pop"=>572, "name_ru"=>"Иука", "name_en"=>"Iuka", "lang"=>"en", "lat"=>38.614542, "lon"=>-88.789831),
"/US/US-IL/ivesdale-5106" => array("type"=>"village", "pop"=>272, "name_ru"=>"Айвсдейл", "name_en"=>"Ivesdale", "lang"=>"en", "lat"=>39.945125, "lon"=>-88.457009),
"/US/US-IL/jacksonville-5107" => array("type"=>"city", "pop"=>19415, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>39.729695, "lon"=>-90.234288),
"/US/US-IL/jeffersonville-5108" => array("type"=>"village", "pop"=>353, "name_ru"=>"Джефферсонвилл", "name_en"=>"Jeffersonville", "lang"=>"en", "lat"=>38.442971, "lon"=>-88.404666),
"/US/US-IL/jeiseyville-5109" => array("type"=>"village", "pop"=>126, "name_ru"=>"Джейзивилл", "name_en"=>"Jeiseyville", "lang"=>"en", "lat"=>39.577303, "lon"=>-89.407431),
"/US/US-IL/jerome-5110" => array("type"=>"village", "pop"=>1294, "name_ru"=>"Джером", "name_en"=>"Jerome", "lang"=>"en", "lat"=>39.766850, "lon"=>-89.681874),
"/US/US-IL/jerseyville-5111" => array("type"=>"city", "pop"=>8373, "name_ru"=>"Джерсивилл", "name_en"=>"Jerseyville", "lang"=>"en", "lat"=>39.119530, "lon"=>-90.327972),
"/US/US-IL/jewett-5112" => array("type"=>"village", "pop"=>223, "name_ru"=>"Джеветт", "name_en"=>"Jewett", "lang"=>"en", "lat"=>39.207670, "lon"=>-88.242779),
"/US/US-IL/johnsonville-5113" => array("type"=>"village", "pop"=>66, "name_ru"=>"Джонсонвилл", "name_en"=>"Johnsonville", "lang"=>"en", "lat"=>38.520610, "lon"=>-88.537154),
"/US/US-IL/johnston-city-5114" => array("type"=>"city", "pop"=>3477, "name_ru"=>"Джонстон Сити", "name_en"=>"Johnston City", "lang"=>"en", "lat"=>37.821637, "lon"=>-88.925526),
"/US/US-IL/joliet-5115" => array("type"=>"city", "pop"=>146125, "name_ru"=>"Джолиет", "name_en"=>"Joliet", "lang"=>"en", "lat"=>41.526717, "lon"=>-88.122436),
"/US/US-IL/jonesboro-5116" => array("type"=>"city", "pop"=>1818, "name_ru"=>"Джонсборо", "name_en"=>"Jonesboro", "lang"=>"en", "lat"=>37.454276, "lon"=>-89.269645),
"/US/US-IL/joppa-5117" => array("type"=>"village", "pop"=>406, "name_ru"=>"Джоппа", "name_en"=>"Joppa", "lang"=>"en", "lat"=>37.205777, "lon"=>-88.844258),
"/US/US-IL/joy-5118" => array("type"=>"village", "pop"=>348, "name_ru"=>"Джой", "name_en"=>"Joy", "lang"=>"en", "lat"=>41.196874, "lon"=>-90.879559),
"/US/US-IL/junction-5119" => array("type"=>"village", "pop"=>129, "name_ru"=>"Джанкшен", "name_en"=>"Junction", "lang"=>"en", "lat"=>37.722963, "lon"=>-88.237794),
"/US/US-IL/junction-city-5120" => array("type"=>"village", "pop"=>537, "name_ru"=>"Джанкшен Сити", "name_en"=>"Junction City", "lang"=>"en", "lat"=>38.577317, "lon"=>-89.125421),
"/US/US-IL/justice-5121" => array("type"=>"village", "pop"=>12452, "name_ru"=>"Джастис", "name_en"=>"Justice", "lang"=>"en", "lat"=>41.746400, "lon"=>-87.835524),
"/US/US-IL/kampsville-5122" => array("type"=>"village", "pop"=>293, "name_ru"=>"Кампсвилл", "name_en"=>"Kampsville", "lang"=>"en", "lat"=>39.309708, "lon"=>-90.625212),
"/US/US-IL/kane-5123" => array("type"=>"village", "pop"=>427, "name_ru"=>"Кейн", "name_en"=>"Kane", "lang"=>"en", "lat"=>39.190314, "lon"=>-90.351229),
"/US/US-IL/kangley-5124" => array("type"=>"village", "pop"=>271, "name_ru"=>"Кангли", "name_en"=>"Kangley", "lang"=>"en", "lat"=>41.147614, "lon"=>-88.872604),
"/US/US-IL/kankakee-5125" => array("type"=>"city", "pop"=>26710, "name_ru"=>"Канкаки", "name_en"=>"Kankakee", "lang"=>"en", "lat"=>41.115450, "lon"=>-87.861199),
"/US/US-IL/kansas-5126" => array("type"=>"village", "pop"=>804, "name_ru"=>"Канзас", "name_en"=>"Kansas", "lang"=>"en", "lat"=>39.553860, "lon"=>-87.939949),
"/US/US-IL/kappa-5127" => array("type"=>"village", "pop"=>215, "name_ru"=>"Каппа", "name_en"=>"Kappa", "lang"=>"en", "lat"=>40.674787, "lon"=>-89.008166),
"/US/US-IL/karnak-5128" => array("type"=>"village", "pop"=>540, "name_ru"=>"Карнак", "name_en"=>"Karnak", "lang"=>"en", "lat"=>37.293959, "lon"=>-88.975999),
"/US/US-IL/kaskaskia-5129" => array("type"=>"village", "pop"=>9, "name_ru"=>"Каскаския", "name_en"=>"Kaskaskia", "lang"=>"en", "lat"=>37.921343, "lon"=>-89.916359),
"/US/US-IL/keenes-5130" => array("type"=>"village", "pop"=>95, "name_ru"=>"Кинс", "name_en"=>"Keenes", "lang"=>"en", "lat"=>38.337901, "lon"=>-88.641192),
"/US/US-IL/keensburg-5131" => array("type"=>"village", "pop"=>239, "name_ru"=>"Кинсбург", "name_en"=>"Keensburg", "lang"=>"en", "lat"=>38.351983, "lon"=>-87.866989),
"/US/US-IL/keithsburg-5132" => array("type"=>"city", "pop"=>688, "name_ru"=>"Китсберг", "name_en"=>"Keithsburg", "lang"=>"en", "lat"=>41.100943, "lon"=>-90.935926),
"/US/US-IL/kell-5133" => array("type"=>"village", "pop"=>221, "name_ru"=>"Кэлл", "name_en"=>"Kell", "lang"=>"en", "lat"=>38.490798, "lon"=>-88.904158),
"/US/US-IL/kempton-5134" => array("type"=>"village", "pop"=>230, "name_ru"=>"Кэмптон", "name_en"=>"Kempton", "lang"=>"en", "lat"=>40.935805, "lon"=>-88.235794),
"/US/US-IL/kenilworth-5135" => array("type"=>"village", "pop"=>2397, "name_ru"=>"Кенилворт", "name_en"=>"Kenilworth", "lang"=>"en", "lat"=>42.088950, "lon"=>-87.714878),
"/US/US-IL/kenney-5136" => array("type"=>"village", "pop"=>355, "name_ru"=>"Кенни", "name_en"=>"Kenney", "lang"=>"en", "lat"=>40.098327, "lon"=>-89.086075),
"/US/US-IL/kewanee-5137" => array("type"=>"city", "pop"=>12302, "name_ru"=>"Кевани", "name_en"=>"Kewanee", "lang"=>"en", "lat"=>41.241143, "lon"=>-89.925572),
"/US/US-IL/keyesport-5138" => array("type"=>"village", "pop"=>466, "name_ru"=>"Кейеспорт", "name_en"=>"Keyesport", "lang"=>"en", "lat"=>38.740742, "lon"=>-89.276993),
"/US/US-IL/kilbourne-5139" => array("type"=>"village", "pop"=>350, "name_ru"=>"Килборн", "name_en"=>"Kilbourne", "lang"=>"en", "lat"=>40.151637, "lon"=>-90.009520),
"/US/US-IL/kildeer-5140" => array("type"=>"village", "pop"=>4126, "name_ru"=>"Килдир", "name_en"=>"Kildeer", "lang"=>"en", "lat"=>42.179300, "lon"=>-88.049121),
"/US/US-IL/kincaid-5141" => array("type"=>"village", "pop"=>1450, "name_ru"=>"Кинкейд", "name_en"=>"Kincaid", "lang"=>"en", "lat"=>39.588661, "lon"=>-89.414345),
"/US/US-IL/kinderhook-5142" => array("type"=>"village", "pop"=>231, "name_ru"=>"Киндерхук", "name_en"=>"Kinderhook", "lang"=>"en", "lat"=>39.704866, "lon"=>-91.154732),
"/US/US-IL/kingston-5143" => array("type"=>"village", "pop"=>1060, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>42.099193, "lon"=>-88.756899),
"/US/US-IL/kingston-mines-5144" => array("type"=>"village", "pop"=>252, "name_ru"=>"Кингстон Майнс", "name_en"=>"Kingston Mines", "lang"=>"en", "lat"=>40.558178, "lon"=>-89.770600),
"/US/US-IL/kinmundy-5145" => array("type"=>"city", "pop"=>854, "name_ru"=>"Кинманди", "name_en"=>"Kinmundy", "lang"=>"en", "lat"=>38.773987, "lon"=>-88.849583),
"/US/US-IL/kinsman-5146" => array("type"=>"village", "pop"=>110, "name_ru"=>"Кинсман", "name_en"=>"Kinsman", "lang"=>"en", "lat"=>41.190565, "lon"=>-88.569741),
"/US/US-IL/kirkland-5147" => array("type"=>"village", "pop"=>1697, "name_ru"=>"Киркленд", "name_en"=>"Kirkland", "lang"=>"en", "lat"=>42.092245, "lon"=>-88.854590),
"/US/US-IL/kirkwood-5148" => array("type"=>"village", "pop"=>707, "name_ru"=>"Кирквуд", "name_en"=>"Kirkwood", "lang"=>"en", "lat"=>40.867537, "lon"=>-90.748266),
"/US/US-IL/knoxville-5149" => array("type"=>"city", "pop"=>2917, "name_ru"=>"Ноксвилл", "name_en"=>"Knoxville", "lang"=>"en", "lat"=>40.906265, "lon"=>-90.285365),
"/US/US-IL/la-fayette-5152" => array("type"=>"village", "pop"=>219, "name_ru"=>"Лафайетт", "name_en"=>"La Fayette", "lang"=>"en", "lat"=>41.109788, "lon"=>-89.973522),
"/US/US-IL/la-grange-5153" => array("type"=>"village", "pop"=>15244, "name_ru"=>"Ла Гранж", "name_en"=>"La Grange", "lang"=>"en", "lat"=>41.807850, "lon"=>-87.873394),
"/US/US-IL/la-grange-park-5154" => array("type"=>"village", "pop"=>12404, "name_ru"=>"Ла Гранж Парк", "name_en"=>"La Grange Park", "lang"=>"en", "lat"=>41.831600, "lon"=>-87.871599),
"/US/US-IL/la-harpe-5155" => array("type"=>"city", "pop"=>1287, "name_ru"=>"Ла Харп", "name_en"=>"La Harpe", "lang"=>"en", "lat"=>40.584016, "lon"=>-90.970013),
"/US/US-IL/la-moille-5168" => array("type"=>"village", "pop"=>743, "name_ru"=>"Ла Моил", "name_en"=>"La Moille", "lang"=>"en", "lat"=>41.530142, "lon"=>-89.280930),
"/US/US-IL/la-prairie-5171" => array("type"=>"village", "pop"=>58, "name_ru"=>"Ла Прейри", "name_en"=>"La Prairie", "lang"=>"en", "lat"=>40.147112, "lon"=>-91.002704),
"/US/US-IL/la-rose-5172" => array("type"=>"village", "pop"=>144, "name_ru"=>"Ла Роуз", "name_en"=>"La Rose", "lang"=>"en", "lat"=>40.977386, "lon"=>-89.234532),
"/US/US-IL/la-salle-5173" => array("type"=>"city", "pop"=>9717, "name_ru"=>"Ла Саль", "name_en"=>"La Salle", "lang"=>"en", "lat"=>41.339952, "lon"=>-89.093484),
"/US/US-IL/lacon-5150" => array("type"=>"city", "pop"=>1832, "name_ru"=>"Лакон", "name_en"=>"Lacon", "lang"=>"en", "lat"=>41.021722, "lon"=>-89.406813),
"/US/US-IL/ladd-5151" => array("type"=>"village", "pop"=>1274, "name_ru"=>"Лэдд", "name_en"=>"Ladd", "lang"=>"en", "lat"=>41.382660, "lon"=>-89.214307),
"/US/US-IL/lake-barrington-5156" => array("type"=>"village", "pop"=>4975, "name_ru"=>"Лейк Бэррингтон", "name_en"=>"Lake Barrington", "lang"=>"en", "lat"=>42.210830, "lon"=>-88.165208),
"/US/US-IL/lake-bluff-5157" => array("type"=>"village", "pop"=>6222, "name_ru"=>"Лейк Блафф", "name_en"=>"Lake Bluff", "lang"=>"en", "lat"=>42.282550, "lon"=>-87.850299),
"/US/US-IL/lake-catherine-5158" => array("type"=>"cdp", "pop"=>1515, "name_ru"=>"Лейк Катэрин", "name_en"=>"Lake Catherine", "lang"=>"en", "lat"=>42.490280, "lon"=>-88.125565),
"/US/US-IL/lake-forest-5159" => array("type"=>"city", "pop"=>20990, "name_ru"=>"Лейк Форест", "name_en"=>"Lake Forest", "lang"=>"en", "lat"=>42.236300, "lon"=>-87.857554),
"/US/US-IL/lake-in-the-hills-5160" => array("type"=>"village", "pop"=>29698, "name_ru"=>"Лейк ин зе Хилс", "name_en"=>"Lake in the Hills", "lang"=>"en", "lat"=>42.192821, "lon"=>-88.321641),
"/US/US-IL/lake-of-the-woods-5162" => array("type"=>"cdp", "pop"=>2748, "name_ru"=>"Лейк оф зе Вудс", "name_en"=>"Lake of the Woods", "lang"=>"en", "lat"=>40.206456, "lon"=>-88.368557),
"/US/US-IL/lake-summerset-5163" => array("type"=>"cdp", "pop"=>1296, "name_ru"=>"Лейк Саммерсет", "name_en"=>"Lake Summerset", "lang"=>"en", "lat"=>42.449668, "lon"=>-89.397532),
"/US/US-IL/lake-villa-5164" => array("type"=>"village", "pop"=>8709, "name_ru"=>"Лейк Вилла", "name_en"=>"Lake Villa", "lang"=>"en", "lat"=>42.417526, "lon"=>-88.082278),
"/US/US-IL/lake-zurich-5167" => array("type"=>"village", "pop"=>20610, "name_ru"=>"Лейк Цурих", "name_en"=>"Lake Zurich", "lang"=>"en", "lat"=>42.192300, "lon"=>-88.088021),
"/US/US-IL/lakemoor-5161" => array("type"=>"village", "pop"=>5748, "name_ru"=>"Лейкмур", "name_en"=>"Lakemoor", "lang"=>"en", "lat"=>42.341491, "lon"=>-88.204869),
"/US/US-IL/lakewood-5165" => array("type"=>"village", "pop"=>3690, "name_ru"=>"Лейквуд", "name_en"=>"Lakewood", "lang"=>"en", "lat"=>42.220129, "lon"=>-88.378087),
"/US/US-IL/lakewood-shores-5166" => array("type"=>"cdp", "pop"=>1606, "name_ru"=>"Лейквуд Шорс", "name_en"=>"Lakewood Shores", "lang"=>"en", "lat"=>41.270814, "lon"=>-88.136305),
"/US/US-IL/lanark-5169" => array("type"=>"city", "pop"=>1457, "name_ru"=>"Ланарк", "name_en"=>"Lanark", "lang"=>"en", "lat"=>42.102782, "lon"=>-89.832056),
"/US/US-IL/lansing-5170" => array("type"=>"village", "pop"=>26672, "name_ru"=>"Ленсинг", "name_en"=>"Lansing", "lang"=>"en", "lat"=>41.567550, "lon"=>-87.545699),
"/US/US-IL/latham-5174" => array("type"=>"village", "pop"=>343, "name_ru"=>"Латам", "name_en"=>"Latham", "lang"=>"en", "lat"=>39.966888, "lon"=>-89.161882),
"/US/US-IL/lawrenceville-5175" => array("type"=>"city", "pop"=>4382, "name_ru"=>"Лоренсвилл", "name_en"=>"Lawrenceville", "lang"=>"en", "lat"=>38.726421, "lon"=>-87.687945),
"/US/US-IL/le-roy-5186" => array("type"=>"city", "pop"=>3527, "name_ru"=>"Ле Рой", "name_en"=>"Le Roy", "lang"=>"en", "lat"=>40.341583, "lon"=>-88.764677),
"/US/US-IL/leaf-river-5176" => array("type"=>"village", "pop"=>541, "name_ru"=>"Лиф Ривер", "name_en"=>"Leaf River", "lang"=>"en", "lat"=>42.123153, "lon"=>-89.402866),
"/US/US-IL/lebanon-5177" => array("type"=>"city", "pop"=>4477, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>38.603050, "lon"=>-89.814985),
"/US/US-IL/lee-5178" => array("type"=>"village", "pop"=>313, "name_ru"=>"Ли", "name_en"=>"Lee", "lang"=>"en", "lat"=>41.794022, "lon"=>-88.940154),
"/US/US-IL/leland-5179" => array("type"=>"village", "pop"=>962, "name_ru"=>"Леленд", "name_en"=>"Leland", "lang"=>"en", "lat"=>41.614694, "lon"=>-88.797976),
"/US/US-IL/leland-grove-5180" => array("type"=>"city", "pop"=>1443, "name_ru"=>"Леленд Гров", "name_en"=>"Leland Grove", "lang"=>"en", "lat"=>39.779129, "lon"=>-89.684613),
"/US/US-IL/lemont-5181" => array("type"=>"village", "pop"=>16176, "name_ru"=>"Лемонт", "name_en"=>"Lemont", "lang"=>"en", "lat"=>41.670050, "lon"=>-87.991475),
"/US/US-IL/lena-5182" => array("type"=>"village", "pop"=>2786, "name_ru"=>"Лена", "name_en"=>"Lena", "lang"=>"en", "lat"=>42.380936, "lon"=>-89.825817),
"/US/US-IL/lenzburg-5183" => array("type"=>"village", "pop"=>530, "name_ru"=>"Лензбург", "name_en"=>"Lenzburg", "lang"=>"en", "lat"=>38.285174, "lon"=>-89.818696),
"/US/US-IL/leonore-5184" => array("type"=>"village", "pop"=>108, "name_ru"=>"Леонор", "name_en"=>"Leonore", "lang"=>"en", "lat"=>41.189417, "lon"=>-88.982597),
"/US/US-IL/lerna-5185" => array("type"=>"village", "pop"=>312, "name_ru"=>"Лерна", "name_en"=>"Lerna", "lang"=>"en", "lat"=>39.417908, "lon"=>-88.288913),
"/US/US-IL/lewistown-5187" => array("type"=>"city", "pop"=>2374, "name_ru"=>"Льюистаун", "name_en"=>"Lewistown", "lang"=>"en", "lat"=>40.396876, "lon"=>-90.155349),
"/US/US-IL/lexington-5188" => array("type"=>"city", "pop"=>1899, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>40.643774, "lon"=>-88.782669),
"/US/US-IL/liberty-5189" => array("type"=>"village", "pop"=>510, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>39.880759, "lon"=>-91.108022),
"/US/US-IL/libertyville-5190" => array("type"=>"village", "pop"=>21843, "name_ru"=>"Либертивилл", "name_en"=>"Libertyville", "lang"=>"en", "lat"=>42.284100, "lon"=>-87.966026),
"/US/US-IL/lima-5191" => array("type"=>"village", "pop"=>156, "name_ru"=>"Лима", "name_en"=>"Lima", "lang"=>"en", "lat"=>40.176676, "lon"=>-91.375012),
"/US/US-IL/lincoln-5192" => array("type"=>"city", "pop"=>14541, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>40.150597, "lon"=>-89.366890),
"/US/US-IL/lincolnshire-5193" => array("type"=>"village", "pop"=>7960, "name_ru"=>"Линкольншир", "name_en"=>"Lincolnshire", "lang"=>"en", "lat"=>42.195600, "lon"=>-87.916066),
"/US/US-IL/lincolnwood-5194" => array("type"=>"village", "pop"=>11810, "name_ru"=>"Линкольнвуд", "name_en"=>"Lincolnwood", "lang"=>"en", "lat"=>42.005900, "lon"=>-87.732820),
"/US/US-IL/lindenhurst-5195" => array("type"=>"village", "pop"=>14695, "name_ru"=>"Линденхерст", "name_en"=>"Lindenhurst", "lang"=>"en", "lat"=>42.417402, "lon"=>-88.029812),
"/US/US-IL/lisbon-5196" => array("type"=>"village", "pop"=>335, "name_ru"=>"Лисбон", "name_en"=>"Lisbon", "lang"=>"en", "lat"=>41.480570, "lon"=>-88.482670),
"/US/US-IL/lisle-5197" => array("type"=>"village", "pop"=>23135, "name_ru"=>"Лайл", "name_en"=>"Lisle", "lang"=>"en", "lat"=>41.791800, "lon"=>-88.087777),
"/US/US-IL/litchfield-5198" => array("type"=>"city", "pop"=>6647, "name_ru"=>"Литчфилд", "name_en"=>"Litchfield", "lang"=>"en", "lat"=>39.175859, "lon"=>-89.656011),
"/US/US-IL/little-york-5200" => array("type"=>"village", "pop"=>247, "name_ru"=>"Литл Йорк", "name_en"=>"Little York", "lang"=>"en", "lat"=>41.010448, "lon"=>-90.746685),
"/US/US-IL/littleton-5199" => array("type"=>"village", "pop"=>189, "name_ru"=>"Литтлтон", "name_en"=>"Littleton", "lang"=>"en", "lat"=>40.234717, "lon"=>-90.623283),
"/US/US-IL/liverpool-5201" => array("type"=>"village", "pop"=>113, "name_ru"=>"Ливерпул", "name_en"=>"Liverpool", "lang"=>"en", "lat"=>40.390503, "lon"=>-90.001648),
"/US/US-IL/livingston-5202" => array("type"=>"village", "pop"=>787, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>38.967765, "lon"=>-89.763859),
"/US/US-IL/loami-5203" => array("type"=>"village", "pop"=>781, "name_ru"=>"Лоами", "name_en"=>"Loami", "lang"=>"en", "lat"=>39.674401, "lon"=>-89.847971),
"/US/US-IL/lockport-5204" => array("type"=>"city", "pop"=>24810, "name_ru"=>"Локпорт", "name_en"=>"Lockport", "lang"=>"en", "lat"=>41.589433, "lon"=>-88.047851),
"/US/US-IL/loda-5205" => array("type"=>"village", "pop"=>404, "name_ru"=>"Лода", "name_en"=>"Loda", "lang"=>"en", "lat"=>40.516709, "lon"=>-88.075349),
"/US/US-IL/lomax-5206" => array("type"=>"village", "pop"=>422, "name_ru"=>"Ломакс", "name_en"=>"Lomax", "lang"=>"en", "lat"=>40.677597, "lon"=>-91.076351),
"/US/US-IL/lombard-5207" => array("type"=>"village", "pop"=>42890, "name_ru"=>"Ломбард", "name_en"=>"Lombard", "lang"=>"en", "lat"=>41.874950, "lon"=>-88.014965),
"/US/US-IL/london-mills-5208" => array("type"=>"village", "pop"=>423, "name_ru"=>"Лондон Миллс", "name_en"=>"London Mills", "lang"=>"en", "lat"=>40.710694, "lon"=>-90.266686),
"/US/US-IL/long-creek-5209" => array("type"=>"village", "pop"=>1318, "name_ru"=>"Лонг Крик", "name_en"=>"Long Creek", "lang"=>"en", "lat"=>39.804333, "lon"=>-88.847303),
"/US/US-IL/long-grove-5210" => array("type"=>"village", "pop"=>8101, "name_ru"=>"Лонг Гров", "name_en"=>"Long Grove", "lang"=>"en", "lat"=>42.195388, "lon"=>-88.002841),
"/US/US-IL/long-lake-5211" => array("type"=>"cdp", "pop"=>2888, "name_ru"=>"Лонг Лейк", "name_en"=>"Long Lake", "lang"=>"en", "lat"=>42.376733, "lon"=>-88.126179),
"/US/US-IL/long-point-5212" => array("type"=>"village", "pop"=>232, "name_ru"=>"Лонг Пойнт", "name_en"=>"Long Point", "lang"=>"en", "lat"=>41.004653, "lon"=>-88.893616),
"/US/US-IL/longview-5213" => array("type"=>"village", "pop"=>144, "name_ru"=>"Лонгвью", "name_en"=>"Longview", "lang"=>"en", "lat"=>39.885526, "lon"=>-88.066464),
"/US/US-IL/loraine-5214" => array("type"=>"village", "pop"=>355, "name_ru"=>"Лорейн", "name_en"=>"Loraine", "lang"=>"en", "lat"=>40.153249, "lon"=>-91.222244),
"/US/US-IL/lostant-5215" => array("type"=>"village", "pop"=>472, "name_ru"=>"Лостант", "name_en"=>"Lostant", "lang"=>"en", "lat"=>41.140447, "lon"=>-89.061133),
"/US/US-IL/louisville-5216" => array("type"=>"village", "pop"=>1235, "name_ru"=>"Луисвилл", "name_en"=>"Louisville", "lang"=>"en", "lat"=>38.769536, "lon"=>-88.507330),
"/US/US-IL/loves-park-5217" => array("type"=>"city", "pop"=>24457, "name_ru"=>"Лавс Парк", "name_en"=>"Loves Park", "lang"=>"en", "lat"=>42.335848, "lon"=>-89.009975),
"/US/US-IL/lovington-5218" => array("type"=>"village", "pop"=>1170, "name_ru"=>"Лавингтон", "name_en"=>"Lovington", "lang"=>"en", "lat"=>39.714504, "lon"=>-88.630943),
"/US/US-IL/ludlow-5219" => array("type"=>"village", "pop"=>361, "name_ru"=>"Ладлоу", "name_en"=>"Ludlow", "lang"=>"en", "lat"=>40.386156, "lon"=>-88.126119),
"/US/US-IL/lyndon-5220" => array("type"=>"village", "pop"=>543, "name_ru"=>"Линдон", "name_en"=>"Lyndon", "lang"=>"en", "lat"=>41.717317, "lon"=>-89.925189),
"/US/US-IL/lynnville-5221" => array("type"=>"village", "pop"=>133, "name_ru"=>"Линнвилл", "name_en"=>"Lynnville", "lang"=>"en", "lat"=>39.685008, "lon"=>-90.346260),
"/US/US-IL/lynwood-5222" => array("type"=>"village", "pop"=>8519, "name_ru"=>"Линвуд", "name_en"=>"Lynwood", "lang"=>"en", "lat"=>41.526444, "lon"=>-87.541959),
"/US/US-IL/lyons-5223" => array("type"=>"village", "pop"=>10250, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>41.811733, "lon"=>-87.818644),
"/US/US-IL/macedonia-5226" => array("type"=>"village", "pop"=>50, "name_ru"=>"Македония", "name_en"=>"Macedonia", "lang"=>"en", "lat"=>38.053761, "lon"=>-88.703078),
"/US/US-IL/machesney-park-5228" => array("type"=>"village", "pop"=>22815, "name_ru"=>"Мачесни Парк", "name_en"=>"Machesney Park", "lang"=>"en", "lat"=>42.360433, "lon"=>-89.037983),
"/US/US-IL/mackinaw-5229" => array("type"=>"village", "pop"=>1715, "name_ru"=>"Маккино", "name_en"=>"Mackinaw", "lang"=>"en", "lat"=>40.535003, "lon"=>-89.358452),
"/US/US-IL/macomb-5233" => array("type"=>"city", "pop"=>19736, "name_ru"=>"Макомб", "name_en"=>"Macomb", "lang"=>"en", "lat"=>40.470393, "lon"=>-90.681440),
"/US/US-IL/macon-5234" => array("type"=>"city", "pop"=>1133, "name_ru"=>"Мейкон", "name_en"=>"Macon", "lang"=>"en", "lat"=>39.708952, "lon"=>-88.999366),
"/US/US-IL/madison-5235" => array("type"=>"city", "pop"=>4554, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>38.676700, "lon"=>-90.150327),
"/US/US-IL/maeystown-5236" => array("type"=>"village", "pop"=>151, "name_ru"=>"Маистаун", "name_en"=>"Maeystown", "lang"=>"en", "lat"=>38.226525, "lon"=>-90.231618),
"/US/US-IL/magnolia-5237" => array("type"=>"village", "pop"=>272, "name_ru"=>"Магнолия", "name_en"=>"Magnolia", "lang"=>"en", "lat"=>41.113966, "lon"=>-89.195182),
"/US/US-IL/mahomet-5238" => array("type"=>"village", "pop"=>6422, "name_ru"=>"Махомет", "name_en"=>"Mahomet", "lang"=>"en", "lat"=>40.193848, "lon"=>-88.401768),
"/US/US-IL/makanda-5239" => array("type"=>"village", "pop"=>408, "name_ru"=>"Маканда", "name_en"=>"Makanda", "lang"=>"en", "lat"=>37.619122, "lon"=>-89.237232),
"/US/US-IL/malden-5240" => array("type"=>"village", "pop"=>331, "name_ru"=>"Малден", "name_en"=>"Malden", "lang"=>"en", "lat"=>41.424878, "lon"=>-89.370167),
"/US/US-IL/malta-5241" => array("type"=>"village", "pop"=>1192, "name_ru"=>"Малта", "name_en"=>"Malta", "lang"=>"en", "lat"=>41.930180, "lon"=>-88.863090),
"/US/US-IL/manchester-5242" => array("type"=>"village", "pop"=>331, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>39.542217, "lon"=>-90.330368),
"/US/US-IL/manhattan-5243" => array("type"=>"village", "pop"=>7166, "name_ru"=>"Манхэттен", "name_en"=>"Manhattan", "lang"=>"en", "lat"=>41.426017, "lon"=>-87.988745),
"/US/US-IL/manito-5244" => array("type"=>"village", "pop"=>1631, "name_ru"=>"Манито", "name_en"=>"Manito", "lang"=>"en", "lat"=>40.421815, "lon"=>-89.780746),
"/US/US-IL/manlius-5245" => array("type"=>"village", "pop"=>340, "name_ru"=>"Манлиус", "name_en"=>"Manlius", "lang"=>"en", "lat"=>41.455705, "lon"=>-89.668033),
"/US/US-IL/mansfield-5246" => array("type"=>"village", "pop"=>910, "name_ru"=>"Мансфилд", "name_en"=>"Mansfield", "lang"=>"en", "lat"=>40.211999, "lon"=>-88.509030),
"/US/US-IL/manteno-5247" => array("type"=>"village", "pop"=>8562, "name_ru"=>"Мантено", "name_en"=>"Manteno", "lang"=>"en", "lat"=>41.251213, "lon"=>-87.836712),
"/US/US-IL/maple-park-5248" => array("type"=>"village", "pop"=>1290, "name_ru"=>"Мейпл Парк", "name_en"=>"Maple Park", "lang"=>"en", "lat"=>41.907211, "lon"=>-88.593732),
"/US/US-IL/mapleton-5249" => array("type"=>"village", "pop"=>236, "name_ru"=>"Мейплтон", "name_en"=>"Mapleton", "lang"=>"en", "lat"=>40.569733, "lon"=>-89.721407),
"/US/US-IL/maquon-5250" => array("type"=>"village", "pop"=>286, "name_ru"=>"Маквон", "name_en"=>"Maquon", "lang"=>"en", "lat"=>40.798256, "lon"=>-90.162679),
"/US/US-IL/marengo-5251" => array("type"=>"city", "pop"=>7520, "name_ru"=>"Маренго", "name_en"=>"Marengo", "lang"=>"en", "lat"=>42.247914, "lon"=>-88.600323),
"/US/US-IL/marietta-5252" => array("type"=>"village", "pop"=>142, "name_ru"=>"Мариетта", "name_en"=>"Marietta", "lang"=>"en", "lat"=>40.499844, "lon"=>-90.392349),
"/US/US-IL/marine-5253" => array("type"=>"village", "pop"=>935, "name_ru"=>"Марин", "name_en"=>"Marine", "lang"=>"en", "lat"=>38.785831, "lon"=>-89.778956),
"/US/US-IL/marion-5254" => array("type"=>"city", "pop"=>17388, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>37.732813, "lon"=>-88.941350),
"/US/US-IL/marissa-5255" => array("type"=>"village", "pop"=>1982, "name_ru"=>"Марисса", "name_en"=>"Marissa", "lang"=>"en", "lat"=>38.255010, "lon"=>-89.753695),
"/US/US-IL/mark-5256" => array("type"=>"village", "pop"=>514, "name_ru"=>"Марк", "name_en"=>"Mark", "lang"=>"en", "lat"=>41.264163, "lon"=>-89.248389),
"/US/US-IL/markham-5257" => array("type"=>"city", "pop"=>12248, "name_ru"=>"Маркхэм", "name_en"=>"Markham", "lang"=>"en", "lat"=>41.599200, "lon"=>-87.691228),
"/US/US-IL/maroa-5258" => array("type"=>"city", "pop"=>1538, "name_ru"=>"Мароа", "name_en"=>"Maroa", "lang"=>"en", "lat"=>40.038231, "lon"=>-88.954938),
"/US/US-IL/marquette-heights-5259" => array("type"=>"city", "pop"=>2831, "name_ru"=>"Маркетт Хайтс", "name_en"=>"Marquette Heights", "lang"=>"en", "lat"=>40.617182, "lon"=>-89.603685),
"/US/US-IL/marseilles-5260" => array("type"=>"city", "pop"=>4924, "name_ru"=>"Марсейллес", "name_en"=>"Marseilles", "lang"=>"en", "lat"=>41.327772, "lon"=>-88.687835),
"/US/US-IL/marshall-5261" => array("type"=>"city", "pop"=>3846, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>39.396183, "lon"=>-87.690247),
"/US/US-IL/martinsville-5262" => array("type"=>"city", "pop"=>1212, "name_ru"=>"Мартинсвилл", "name_en"=>"Martinsville", "lang"=>"en", "lat"=>39.336236, "lon"=>-87.880201),
"/US/US-IL/martinton-5263" => array("type"=>"village", "pop"=>383, "name_ru"=>"Мартинтон", "name_en"=>"Martinton", "lang"=>"en", "lat"=>40.915282, "lon"=>-87.726255),
"/US/US-IL/maryville-5264" => array("type"=>"village", "pop"=>7397, "name_ru"=>"Мэривилл", "name_en"=>"Maryville", "lang"=>"en", "lat"=>38.726367, "lon"=>-89.964749),
"/US/US-IL/mascoutah-5265" => array("type"=>"city", "pop"=>6701, "name_ru"=>"Маскаута", "name_en"=>"Mascoutah", "lang"=>"en", "lat"=>38.509115, "lon"=>-89.805431),
"/US/US-IL/mason-5266" => array("type"=>"town", "pop"=>397, "name_ru"=>"Мейсон", "name_en"=>"Mason", "lang"=>"en", "lat"=>38.953097, "lon"=>-88.625403),
"/US/US-IL/mason-city-5267" => array("type"=>"city", "pop"=>2365, "name_ru"=>"Мейсон Сити", "name_en"=>"Mason City", "lang"=>"en", "lat"=>40.201818, "lon"=>-89.696189),
"/US/US-IL/matherville-5268" => array("type"=>"village", "pop"=>749, "name_ru"=>"Матервилл", "name_en"=>"Matherville", "lang"=>"en", "lat"=>41.259134, "lon"=>-90.605731),
"/US/US-IL/matteson-5269" => array("type"=>"village", "pop"=>17420, "name_ru"=>"Мэттесон", "name_en"=>"Matteson", "lang"=>"en", "lat"=>41.511033, "lon"=>-87.737518),
"/US/US-IL/mattoon-5270" => array("type"=>"city", "pop"=>17177, "name_ru"=>"Мэттун", "name_en"=>"Mattoon", "lang"=>"en", "lat"=>39.477038, "lon"=>-88.368830),
"/US/US-IL/maunie-5271" => array("type"=>"village", "pop"=>170, "name_ru"=>"Мауни", "name_en"=>"Maunie", "lang"=>"en", "lat"=>38.035059, "lon"=>-88.045315),
"/US/US-IL/maywood-5272" => array("type"=>"village", "pop"=>25035, "name_ru"=>"Мэйвуд", "name_en"=>"Maywood", "lang"=>"en", "lat"=>41.880250, "lon"=>-87.844499),
"/US/US-IL/mazon-5273" => array("type"=>"village", "pop"=>1014, "name_ru"=>"Мазон", "name_en"=>"Mazon", "lang"=>"en", "lat"=>41.239583, "lon"=>-88.424856),
"/US/US-IL/mccook-5224" => array("type"=>"village", "pop"=>237, "name_ru"=>"Мак-Кук", "name_en"=>"McCook", "lang"=>"en", "lat"=>41.792833, "lon"=>-87.835825),
"/US/US-IL/mccullom-lake-5225" => array("type"=>"village", "pop"=>1100, "name_ru"=>"Маккалом Лейк", "name_en"=>"McCullom Lake", "lang"=>"en", "lat"=>42.368834, "lon"=>-88.293823),
"/US/US-IL/mchenry-5227" => array("type"=>"city", "pop"=>26862, "name_ru"=>"Мак-Генри", "name_en"=>"McHenry", "lang"=>"en", "lat"=>42.340665, "lon"=>-88.291283),
"/US/US-IL/mclean-5230" => array("type"=>"village", "pop"=>791, "name_ru"=>"Мак-Лин", "name_en"=>"McLean", "lang"=>"en", "lat"=>40.315559, "lon"=>-89.170437),
"/US/US-IL/mcleansboro-5231" => array("type"=>"city", "pop"=>2780, "name_ru"=>"Мак-Линсборо", "name_en"=>"McLeansboro", "lang"=>"en", "lat"=>38.092268, "lon"=>-88.532843),
"/US/US-IL/mcnabb-5232" => array("type"=>"village", "pop"=>282, "name_ru"=>"Мак-Набб", "name_en"=>"McNabb", "lang"=>"en", "lat"=>41.177280, "lon"=>-89.209830),
"/US/US-IL/mechanicsburg-5274" => array("type"=>"village", "pop"=>449, "name_ru"=>"Меканиксбург", "name_en"=>"Mechanicsburg", "lang"=>"en", "lat"=>39.810167, "lon"=>-89.398246),
"/US/US-IL/media-5275" => array("type"=>"village", "pop"=>114, "name_ru"=>"Медиа", "name_en"=>"Media", "lang"=>"en", "lat"=>40.772658, "lon"=>-90.833961),
"/US/US-IL/medinah-5276" => array("type"=>"cdp", "pop"=>2512, "name_ru"=>"Медина", "name_en"=>"Medinah", "lang"=>"en", "lat"=>41.975460, "lon"=>-88.055692),
"/US/US-IL/medora-5277" => array("type"=>"village", "pop"=>481, "name_ru"=>"Медора", "name_en"=>"Medora", "lang"=>"en", "lat"=>39.176017, "lon"=>-90.141692),
"/US/US-IL/melrose-park-5278" => array("type"=>"village", "pop"=>21853, "name_ru"=>"Мелроуз Парк", "name_en"=>"Melrose Park", "lang"=>"en", "lat"=>41.903050, "lon"=>-87.863448),
"/US/US-IL/melvin-5279" => array("type"=>"village", "pop"=>456, "name_ru"=>"Мэлвин", "name_en"=>"Melvin", "lang"=>"en", "lat"=>40.571367, "lon"=>-88.247363),
"/US/US-IL/mendon-5280" => array("type"=>"village", "pop"=>1009, "name_ru"=>"Мендон", "name_en"=>"Mendon", "lang"=>"en", "lat"=>40.089075, "lon"=>-91.284253),
"/US/US-IL/mendota-5281" => array("type"=>"city", "pop"=>6995, "name_ru"=>"Мендота", "name_en"=>"Mendota", "lang"=>"en", "lat"=>41.550884, "lon"=>-89.120071),
"/US/US-IL/menominee-5282" => array("type"=>"village", "pop"=>205, "name_ru"=>"Меномини", "name_en"=>"Menominee", "lang"=>"en", "lat"=>42.474648, "lon"=>-90.541986),
"/US/US-IL/meredosia-5283" => array("type"=>"village", "pop"=>966, "name_ru"=>"Мередозия", "name_en"=>"Meredosia", "lang"=>"en", "lat"=>39.831525, "lon"=>-90.558330),
"/US/US-IL/merrionette-park-5284" => array("type"=>"village", "pop"=>1982, "name_ru"=>"Меррионетт", "name_en"=>"Merrionette Park", "lang"=>"en", "lat"=>41.680300, "lon"=>-87.699749),
"/US/US-IL/metamora-5285" => array("type"=>"village", "pop"=>3398, "name_ru"=>"Метамора", "name_en"=>"Metamora", "lang"=>"en", "lat"=>40.792081, "lon"=>-89.363645),
"/US/US-IL/metcalf-5286" => array("type"=>"village", "pop"=>203, "name_ru"=>"Меткальф", "name_en"=>"Metcalf", "lang"=>"en", "lat"=>39.800273, "lon"=>-87.808035),
"/US/US-IL/metropolis-5287" => array("type"=>"city", "pop"=>6540, "name_ru"=>"Метрополис", "name_en"=>"Metropolis", "lang"=>"en", "lat"=>37.152187, "lon"=>-88.711955),
"/US/US-IL/mettawa-5288" => array("type"=>"village", "pop"=>514, "name_ru"=>"Меттава", "name_en"=>"Mettawa", "lang"=>"en", "lat"=>42.243733, "lon"=>-87.915394),
"/US/US-IL/middletown-5289" => array("type"=>"village", "pop"=>416, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>40.100639, "lon"=>-89.591344),
"/US/US-IL/midlothian-5290" => array("type"=>"village", "pop"=>13657, "name_ru"=>"Мидлотиан", "name_en"=>"Midlothian", "lang"=>"en", "lat"=>41.626300, "lon"=>-87.724274),
"/US/US-IL/milan-5291" => array("type"=>"village", "pop"=>5186, "name_ru"=>"Милан", "name_en"=>"Milan", "lang"=>"en", "lat"=>41.444150, "lon"=>-90.561254),
"/US/US-IL/milford-5292" => array("type"=>"village", "pop"=>1283, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>40.627692, "lon"=>-87.697040),
"/US/US-IL/mill-creek-5293" => array("type"=>"village", "pop"=>75, "name_ru"=>"Милл Крик", "name_en"=>"Mill Creek", "lang"=>"en", "lat"=>37.341499, "lon"=>-89.254202),
"/US/US-IL/mill-shoals-5296" => array("type"=>"village", "pop"=>225, "name_ru"=>"Милл Шоалс", "name_en"=>"Mill Shoals", "lang"=>"en", "lat"=>38.251426, "lon"=>-88.345118),
"/US/US-IL/milledgeville-5294" => array("type"=>"village", "pop"=>930, "name_ru"=>"Миледжвилл", "name_en"=>"Milledgeville", "lang"=>"en", "lat"=>41.963759, "lon"=>-89.775643),
"/US/US-IL/millington-5295" => array("type"=>"village", "pop"=>509, "name_ru"=>"Миллингтон", "name_en"=>"Millington", "lang"=>"en", "lat"=>41.560959, "lon"=>-88.599192),
"/US/US-IL/millstadt-5297" => array("type"=>"village", "pop"=>3348, "name_ru"=>"Миллштадт", "name_en"=>"Millstadt", "lang"=>"en", "lat"=>38.459400, "lon"=>-90.093074),
"/US/US-IL/milton-5298" => array("type"=>"village", "pop"=>254, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>39.564191, "lon"=>-90.649781),
"/US/US-IL/mineral-5299" => array("type"=>"village", "pop"=>261, "name_ru"=>"Минерал", "name_en"=>"Mineral", "lang"=>"en", "lat"=>41.381886, "lon"=>-89.836472),
"/US/US-IL/minier-5300" => array("type"=>"village", "pop"=>1247, "name_ru"=>"Миньер", "name_en"=>"Minier", "lang"=>"en", "lat"=>40.433782, "lon"=>-89.314015),
"/US/US-IL/minonk-5301" => array("type"=>"city", "pop"=>2154, "name_ru"=>"Минонк", "name_en"=>"Minonk", "lang"=>"en", "lat"=>40.903614, "lon"=>-89.037452),
"/US/US-IL/minooka-5302" => array("type"=>"village", "pop"=>10514, "name_ru"=>"Минуа", "name_en"=>"Minooka", "lang"=>"en", "lat"=>41.449100, "lon"=>-88.257713),
"/US/US-IL/modesto-5303" => array("type"=>"village", "pop"=>243, "name_ru"=>"Модесто", "name_en"=>"Modesto", "lang"=>"en", "lat"=>39.479072, "lon"=>-89.980192),
"/US/US-IL/mokena-5304" => array("type"=>"village", "pop"=>19236, "name_ru"=>"Мокена", "name_en"=>"Mokena", "lang"=>"en", "lat"=>41.532034, "lon"=>-87.875711),
"/US/US-IL/moline-5305" => array("type"=>"city", "pop"=>43088, "name_ru"=>"Молайн", "name_en"=>"Moline", "lang"=>"en", "lat"=>41.485100, "lon"=>-90.488799),
"/US/US-IL/momence-5306" => array("type"=>"city", "pop"=>3188, "name_ru"=>"Моменс", "name_en"=>"Momence", "lang"=>"en", "lat"=>41.163514, "lon"=>-87.663552),
"/US/US-IL/monee-5307" => array("type"=>"village", "pop"=>4993, "name_ru"=>"Мони", "name_en"=>"Monee", "lang"=>"en", "lat"=>41.422033, "lon"=>-87.754090),
"/US/US-IL/monmouth-5308" => array("type"=>"city", "pop"=>9321, "name_ru"=>"Монмаут", "name_en"=>"Monmouth", "lang"=>"en", "lat"=>40.913379, "lon"=>-90.642995),
"/US/US-IL/montgomery-5309" => array("type"=>"village", "pop"=>15229, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>41.729900, "lon"=>-88.341473),
"/US/US-IL/monticello-5310" => array("type"=>"city", "pop"=>5350, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>40.025822, "lon"=>-88.574677),
"/US/US-IL/montrose-5311" => array("type"=>"village", "pop"=>257, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>39.165728, "lon"=>-88.378342),
"/US/US-IL/morris-5312" => array("type"=>"city", "pop"=>13812, "name_ru"=>"Моррис", "name_en"=>"Morris", "lang"=>"en", "lat"=>41.368648, "lon"=>-88.427663),
"/US/US-IL/morrison-5313" => array("type"=>"city", "pop"=>4303, "name_ru"=>"Моррисон", "name_en"=>"Morrison", "lang"=>"en", "lat"=>41.808447, "lon"=>-89.965416),
"/US/US-IL/morrisonville-5314" => array("type"=>"village", "pop"=>1020, "name_ru"=>"Моррисонвилл", "name_en"=>"Morrisonville", "lang"=>"en", "lat"=>39.421220, "lon"=>-89.457795),
"/US/US-IL/morton-5315" => array("type"=>"village", "pop"=>16039, "name_ru"=>"Мортон", "name_en"=>"Morton", "lang"=>"en", "lat"=>40.611181, "lon"=>-89.467839),
"/US/US-IL/morton-grove-5316" => array("type"=>"village", "pop"=>22478, "name_ru"=>"Мортон Гров", "name_en"=>"Morton Grove", "lang"=>"en", "lat"=>42.042350, "lon"=>-87.788986),
"/US/US-IL/mound-city-5317" => array("type"=>"city", "pop"=>590, "name_ru"=>"Маунд Сити", "name_en"=>"Mound City", "lang"=>"en", "lat"=>37.085573, "lon"=>-89.163047),
"/US/US-IL/mound-station-5319" => array("type"=>"village", "pop"=>117, "name_ru"=>"Маунд Стейшен", "name_en"=>"Mound Station", "lang"=>"en", "lat"=>40.006667, "lon"=>-90.873102),
"/US/US-IL/mounds-5318" => array("type"=>"city", "pop"=>949, "name_ru"=>"Маундс", "name_en"=>"Mounds", "lang"=>"en", "lat"=>37.114844, "lon"=>-89.201246),
"/US/US-IL/mount-auburn-5320" => array("type"=>"village", "pop"=>542, "name_ru"=>"Маунт Обурн", "name_en"=>"Mount Auburn", "lang"=>"en", "lat"=>39.765864, "lon"=>-89.260987),
"/US/US-IL/mount-carmel-5321" => array("type"=>"city", "pop"=>7404, "name_ru"=>"Маунт Кармел", "name_en"=>"Mount Carmel", "lang"=>"en", "lat"=>38.417206, "lon"=>-87.770092),
"/US/US-IL/mount-carroll-5322" => array("type"=>"city", "pop"=>1659, "name_ru"=>"Маунт Кэррол", "name_en"=>"Mount Carroll", "lang"=>"en", "lat"=>42.095450, "lon"=>-89.976918),
"/US/US-IL/mount-clare-5323" => array("type"=>"village", "pop"=>421, "name_ru"=>"Маунт Клэр", "name_en"=>"Mount Clare", "lang"=>"en", "lat"=>39.098859, "lon"=>-89.824769),
"/US/US-IL/mount-erie-5324" => array("type"=>"village", "pop"=>101, "name_ru"=>"Маунт Эри", "name_en"=>"Mount Erie", "lang"=>"en", "lat"=>38.514626, "lon"=>-88.232680),
"/US/US-IL/mount-morris-5325" => array("type"=>"village", "pop"=>3078, "name_ru"=>"Маунт Моррис", "name_en"=>"Mount Morris", "lang"=>"en", "lat"=>42.048924, "lon"=>-89.431983),
"/US/US-IL/mount-olive-5326" => array("type"=>"city", "pop"=>2052, "name_ru"=>"Маунт Олив", "name_en"=>"Mount Olive", "lang"=>"en", "lat"=>39.072585, "lon"=>-89.727955),
"/US/US-IL/mount-prospect-5327" => array("type"=>"village", "pop"=>53338, "name_ru"=>"Маунт Проспект", "name_en"=>"Mount Prospect", "lang"=>"en", "lat"=>42.065400, "lon"=>-87.937549),
"/US/US-IL/mount-pulaski-5328" => array("type"=>"city", "pop"=>1580, "name_ru"=>"Маунт Пуласки", "name_en"=>"Mount Pulaski", "lang"=>"en", "lat"=>40.010051, "lon"=>-89.283078),
"/US/US-IL/mount-sterling-5329" => array("type"=>"city", "pop"=>1907, "name_ru"=>"Маунт Стерлинг", "name_en"=>"Mount Sterling", "lang"=>"en", "lat"=>39.984720, "lon"=>-90.763701),
"/US/US-IL/mount-vernon-5330" => array("type"=>"city", "pop"=>16298, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>38.317576, "lon"=>-88.910975),
"/US/US-IL/mount-zion-5331" => array("type"=>"village", "pop"=>5159, "name_ru"=>"Маунт Зион", "name_en"=>"Mount Zion", "lang"=>"en", "lat"=>39.777931, "lon"=>-88.878540),
"/US/US-IL/moweaqua-5332" => array("type"=>"village", "pop"=>1821, "name_ru"=>"Моуиква", "name_en"=>"Moweaqua", "lang"=>"en", "lat"=>39.625361, "lon"=>-89.017657),
"/US/US-IL/muddy-5333" => array("type"=>"village", "pop"=>73, "name_ru"=>"Мадди", "name_en"=>"Muddy", "lang"=>"en", "lat"=>37.764966, "lon"=>-88.516758),
"/US/US-IL/mulberry-grove-5334" => array("type"=>"village", "pop"=>658, "name_ru"=>"Малберри Гров", "name_en"=>"Mulberry Grove", "lang"=>"en", "lat"=>38.924388, "lon"=>-89.267338),
"/US/US-IL/muncie-5335" => array("type"=>"village", "pop"=>147, "name_ru"=>"Манси", "name_en"=>"Muncie", "lang"=>"en", "lat"=>40.115904, "lon"=>-87.842539),
"/US/US-IL/mundelein-5336" => array("type"=>"village", "pop"=>33717, "name_ru"=>"Манделей", "name_en"=>"Mundelein", "lang"=>"en", "lat"=>42.267050, "lon"=>-88.003149),
"/US/US-IL/murphysboro-5337" => array("type"=>"city", "pop"=>8184, "name_ru"=>"Мерфисборо", "name_en"=>"Murphysboro", "lang"=>"en", "lat"=>37.768065, "lon"=>-89.337230),
"/US/US-IL/murrayville-5338" => array("type"=>"village", "pop"=>619, "name_ru"=>"Мюррейвилл", "name_en"=>"Murrayville", "lang"=>"en", "lat"=>39.582328, "lon"=>-90.250931),
"/US/US-IL/naperville-5339" => array("type"=>"city", "pop"=>143117, "name_ru"=>"Непервилл", "name_en"=>"Naperville", "lang"=>"en", "lat"=>41.762600, "lon"=>-88.154299),
"/US/US-IL/naplate-5340" => array("type"=>"village", "pop"=>512, "name_ru"=>"Нейплейт", "name_en"=>"Naplate", "lang"=>"en", "lat"=>41.331271, "lon"=>-88.878228),
"/US/US-IL/naples-5341" => array("type"=>"town", "pop"=>123, "name_ru"=>"Нэплс", "name_en"=>"Naples", "lang"=>"en", "lat"=>39.753927, "lon"=>-90.607985),
"/US/US-IL/nashville-5342" => array("type"=>"city", "pop"=>3027, "name_ru"=>"Нэшвилл", "name_en"=>"Nashville", "lang"=>"en", "lat"=>38.348150, "lon"=>-89.377747),
"/US/US-IL/nason-5343" => array("type"=>"city", "pop"=>234, "name_ru"=>"Насон", "name_en"=>"Nason", "lang"=>"en", "lat"=>38.176292, "lon"=>-88.965948),
"/US/US-IL/national-city-5344" => array("type"=>"village", "pop"=>57, "name_ru"=>"Нешнел Сити", "name_en"=>"National City", "lang"=>"en", "lat"=>38.648400, "lon"=>-90.149924),
"/US/US-IL/nauvoo-5345" => array("type"=>"city", "pop"=>1155, "name_ru"=>"Науву", "name_en"=>"Nauvoo", "lang"=>"en", "lat"=>40.547541, "lon"=>-91.384777),
"/US/US-IL/nebo-5346" => array("type"=>"village", "pop"=>377, "name_ru"=>"Небо", "name_en"=>"Nebo", "lang"=>"en", "lat"=>39.441415, "lon"=>-90.790328),
"/US/US-IL/nelson-5347" => array("type"=>"village", "pop"=>157, "name_ru"=>"Нельсон", "name_en"=>"Nelson", "lang"=>"en", "lat"=>41.795550, "lon"=>-89.602565),
"/US/US-IL/neoga-5348" => array("type"=>"city", "pop"=>1753, "name_ru"=>"Неога", "name_en"=>"Neoga", "lang"=>"en", "lat"=>39.321640, "lon"=>-88.451538),
"/US/US-IL/neponset-5349" => array("type"=>"village", "pop"=>493, "name_ru"=>"Непонсет", "name_en"=>"Neponset", "lang"=>"en", "lat"=>41.297071, "lon"=>-89.790155),
"/US/US-IL/new-athens-5351" => array("type"=>"village", "pop"=>1999, "name_ru"=>"Нью Афины", "name_en"=>"New Athens", "lang"=>"en", "lat"=>38.318953, "lon"=>-89.877409),
"/US/US-IL/new-baden-5352" => array("type"=>"village", "pop"=>3248, "name_ru"=>"Нью Баден", "name_en"=>"New Baden", "lang"=>"en", "lat"=>38.536358, "lon"=>-89.697741),
"/US/US-IL/new-bedford-5353" => array("type"=>"village", "pop"=>91, "name_ru"=>"Нью Бедфорд", "name_en"=>"New Bedford", "lang"=>"en", "lat"=>41.511676, "lon"=>-89.717298),
"/US/US-IL/new-berlin-5354" => array("type"=>"village", "pop"=>1131, "name_ru"=>"Нью Берлин", "name_en"=>"New Berlin", "lang"=>"en", "lat"=>39.725157, "lon"=>-89.911885),
"/US/US-IL/new-boston-5355" => array("type"=>"city", "pop"=>616, "name_ru"=>"Нью Бостон", "name_en"=>"New Boston", "lang"=>"en", "lat"=>41.169372, "lon"=>-91.000535),
"/US/US-IL/new-burnside-5356" => array("type"=>"village", "pop"=>252, "name_ru"=>"Нью Бернсайд", "name_en"=>"New Burnside", "lang"=>"en", "lat"=>37.578378, "lon"=>-88.773058),
"/US/US-IL/new-canton-5357" => array("type"=>"town", "pop"=>385, "name_ru"=>"Нью Кантон", "name_en"=>"New Canton", "lang"=>"en", "lat"=>39.638184, "lon"=>-91.097791),
"/US/US-IL/new-douglas-5358" => array("type"=>"village", "pop"=>347, "name_ru"=>"Нью Дуглас", "name_en"=>"New Douglas", "lang"=>"en", "lat"=>38.970043, "lon"=>-89.665862),
"/US/US-IL/new-grand-chain-5359" => array("type"=>"village", "pop"=>203, "name_ru"=>"Нью Гранд Чейн", "name_en"=>"New Grand Chain", "lang"=>"en", "lat"=>37.253173, "lon"=>-89.017139),
"/US/US-IL/new-haven-5360" => array("type"=>"village", "pop"=>442, "name_ru"=>"Нью Хейвен", "name_en"=>"New Haven", "lang"=>"en", "lat"=>37.900150, "lon"=>-88.127793),
"/US/US-IL/new-holland-5361" => array("type"=>"village", "pop"=>304, "name_ru"=>"Нью Холланд", "name_en"=>"New Holland", "lang"=>"en", "lat"=>40.184325, "lon"=>-89.582484),
"/US/US-IL/new-lenox-5362" => array("type"=>"village", "pop"=>24137, "name_ru"=>"Нью Леннокс", "name_en"=>"New Lenox", "lang"=>"en", "lat"=>41.518715, "lon"=>-87.983963),
"/US/US-IL/new-millford-5364" => array("type"=>"village", "pop"=>463, "name_ru"=>"Нью Миллфорд", "name_en"=>"New Millford", "lang"=>"en", "lat"=>42.174632, "lon"=>-89.065408),
"/US/US-IL/new-minden-5365" => array("type"=>"village", "pop"=>193, "name_ru"=>"Нью Минден", "name_en"=>"New Minden", "lang"=>"en", "lat"=>38.437477, "lon"=>-89.370858),
"/US/US-IL/new-salem-5366" => array("type"=>"village", "pop"=>126, "name_ru"=>"Нью Салем", "name_en"=>"New Salem", "lang"=>"en", "lat"=>39.707999, "lon"=>-90.847358),
"/US/US-IL/newark-5350" => array("type"=>"village", "pop"=>1128, "name_ru"=>"Ньюарк", "name_en"=>"Newark", "lang"=>"en", "lat"=>41.537136, "lon"=>-88.580703),
"/US/US-IL/newman-5363" => array("type"=>"city", "pop"=>912, "name_ru"=>"Ньюман", "name_en"=>"Newman", "lang"=>"en", "lat"=>39.797253, "lon"=>-87.987515),
"/US/US-IL/newton-5367" => array("type"=>"city", "pop"=>2942, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>38.988069, "lon"=>-88.163337),
"/US/US-IL/niantic-5368" => array("type"=>"village", "pop"=>667, "name_ru"=>"Найантик", "name_en"=>"Niantic", "lang"=>"en", "lat"=>39.853864, "lon"=>-89.165493),
"/US/US-IL/niles-5369" => array("type"=>"village", "pop"=>28666, "name_ru"=>"Найлс", "name_en"=>"Niles", "lang"=>"en", "lat"=>42.026950, "lon"=>-87.809499),
"/US/US-IL/nilwood-5370" => array("type"=>"town", "pop"=>274, "name_ru"=>"Нилвуд", "name_en"=>"Nilwood", "lang"=>"en", "lat"=>39.399537, "lon"=>-89.807318),
"/US/US-IL/noble-5371" => array("type"=>"village", "pop"=>691, "name_ru"=>"Ноубл", "name_en"=>"Noble", "lang"=>"en", "lat"=>38.697195, "lon"=>-88.224570),
"/US/US-IL/nokomis-5372" => array("type"=>"city", "pop"=>2259, "name_ru"=>"Нокомис", "name_en"=>"Nokomis", "lang"=>"en", "lat"=>39.300390, "lon"=>-89.285329),
"/US/US-IL/nora-5373" => array("type"=>"village", "pop"=>113, "name_ru"=>"Нора", "name_en"=>"Nora", "lang"=>"en", "lat"=>42.457278, "lon"=>-89.944555),
"/US/US-IL/normal-5374" => array("type"=>"town", "pop"=>52056, "name_ru"=>"Нормал", "name_en"=>"Normal", "lang"=>"en", "lat"=>40.517370, "lon"=>-88.999406),
"/US/US-IL/norridge-5375" => array("type"=>"village", "pop"=>13954, "name_ru"=>"Норридж", "name_en"=>"Norridge", "lang"=>"en", "lat"=>41.965000, "lon"=>-87.822540),
"/US/US-IL/norris-5376" => array("type"=>"village", "pop"=>184, "name_ru"=>"Норрис", "name_en"=>"Norris", "lang"=>"en", "lat"=>40.625545, "lon"=>-90.031655),
"/US/US-IL/norris-city-5377" => array("type"=>"village", "pop"=>1015, "name_ru"=>"Норрис Сити", "name_en"=>"Norris City", "lang"=>"en", "lat"=>37.979467, "lon"=>-88.328241),
"/US/US-IL/north-aurora-5378" => array("type"=>"village", "pop"=>15812, "name_ru"=>"Норт Аврора", "name_en"=>"North Aurora", "lang"=>"en", "lat"=>41.803817, "lon"=>-88.330374),
"/US/US-IL/north-barrington-5379" => array("type"=>"village", "pop"=>3237, "name_ru"=>"Норт Бэррингтон", "name_en"=>"North Barrington", "lang"=>"en", "lat"=>42.207067, "lon"=>-88.131436),
"/US/US-IL/north-chicago-5381" => array("type"=>"city", "pop"=>32608, "name_ru"=>"Норт Чикаго", "name_en"=>"North Chicago", "lang"=>"en", "lat"=>42.318650, "lon"=>-87.859249),
"/US/US-IL/north-city-5382" => array("type"=>"village", "pop"=>653, "name_ru"=>"Норт Сити", "name_en"=>"North City", "lang"=>"en", "lat"=>37.991260, "lon"=>-89.061642),
"/US/US-IL/north-henderson-5384" => array("type"=>"village", "pop"=>181, "name_ru"=>"Норт Хендерсон", "name_en"=>"North Henderson", "lang"=>"en", "lat"=>41.090718, "lon"=>-90.475018),
"/US/US-IL/north-pekin-5386" => array("type"=>"village", "pop"=>1721, "name_ru"=>"Норт Пекин", "name_en"=>"North Pekin", "lang"=>"en", "lat"=>40.613498, "lon"=>-89.620494),
"/US/US-IL/north-riverside-5387" => array("type"=>"village", "pop"=>6240, "name_ru"=>"Норт Риверсайд", "name_en"=>"North Riverside", "lang"=>"en", "lat"=>41.847500, "lon"=>-87.824890),
"/US/US-IL/north-utica-5388" => array("type"=>"village", "pop"=>1039, "name_ru"=>"Норт Ютика", "name_en"=>"North Utica", "lang"=>"en", "lat"=>41.339835, "lon"=>-89.011292),
"/US/US-IL/northbrook-5380" => array("type"=>"village", "pop"=>33936, "name_ru"=>"Нортбрук", "name_en"=>"Northbrook", "lang"=>"en", "lat"=>42.129700, "lon"=>-87.832949),
"/US/US-IL/northfield-5383" => array("type"=>"village", "pop"=>5429, "name_ru"=>"Нортфилд", "name_en"=>"Northfield", "lang"=>"en", "lat"=>42.100600, "lon"=>-87.778724),
"/US/US-IL/northlake-5385" => array("type"=>"city", "pop"=>11462, "name_ru"=>"Нортлейк", "name_en"=>"Northlake", "lang"=>"en", "lat"=>41.913683, "lon"=>-87.903865),
"/US/US-IL/norwood-5389" => array("type"=>"village", "pop"=>456, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>40.706767, "lon"=>-89.699549),
"/US/US-IL/o'fallon-5405" => array("type"=>"city", "pop"=>27540, "name_ru"=>"О'фаллон", "name_en"=>"O'Fallon", "lang"=>"en", "lat"=>38.592000, "lon"=>-89.913299),
"/US/US-IL/oak-brook-5390" => array("type"=>"village", "pop"=>8803, "name_ru"=>"Оук Брук", "name_en"=>"Oak Brook", "lang"=>"en", "lat"=>41.836900, "lon"=>-87.952672),
"/US/US-IL/oak-forest-5394" => array("type"=>"city", "pop"=>27740, "name_ru"=>"Оук Форест", "name_en"=>"Oak Forest", "lang"=>"en", "lat"=>41.606500, "lon"=>-87.752599),
"/US/US-IL/oak-grove-5395" => array("type"=>"village", "pop"=>699, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>41.414015, "lon"=>-90.573207),
"/US/US-IL/oak-lawn-5397" => array("type"=>"village", "pop"=>53244, "name_ru"=>"Оук Лон", "name_en"=>"Oak Lawn", "lang"=>"en", "lat"=>41.715050, "lon"=>-87.752099),
"/US/US-IL/oak-park-5398" => array("type"=>"village", "pop"=>49557, "name_ru"=>"Оук Парк", "name_en"=>"Oak Park", "lang"=>"en", "lat"=>41.887556, "lon"=>-87.789463),
"/US/US-IL/oakbrook-terrace-5391" => array("type"=>"city", "pop"=>2221, "name_ru"=>"Оукбрук Террейс", "name_en"=>"Oakbrook Terrace", "lang"=>"en", "lat"=>41.853326, "lon"=>-87.968687),
"/US/US-IL/oakdale-5392" => array("type"=>"village", "pop"=>201, "name_ru"=>"Оукдейл", "name_en"=>"Oakdale", "lang"=>"en", "lat"=>38.263796, "lon"=>-89.504933),
"/US/US-IL/oakford-5393" => array("type"=>"village", "pop"=>292, "name_ru"=>"Оукфорд", "name_en"=>"Oakford", "lang"=>"en", "lat"=>40.101036, "lon"=>-89.965372),
"/US/US-IL/oakland-5396" => array("type"=>"city", "pop"=>934, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>39.657728, "lon"=>-88.026579),
"/US/US-IL/oakwood-5399" => array("type"=>"village", "pop"=>1426, "name_ru"=>"Оуквуд", "name_en"=>"Oakwood", "lang"=>"en", "lat"=>40.109007, "lon"=>-87.776541),
"/US/US-IL/oakwood-hills-5400" => array("type"=>"village", "pop"=>2376, "name_ru"=>"Оуквуд Хилс", "name_en"=>"Oakwood Hills", "lang"=>"en", "lat"=>42.247273, "lon"=>-88.244401),
"/US/US-IL/oblong-5401" => array("type"=>"village", "pop"=>1523, "name_ru"=>"Облонг", "name_en"=>"Oblong", "lang"=>"en", "lat"=>39.002217, "lon"=>-87.908446),
"/US/US-IL/oconee-5402" => array("type"=>"village", "pop"=>184, "name_ru"=>"Окони", "name_en"=>"Oconee", "lang"=>"en", "lat"=>39.286529, "lon"=>-89.106186),
"/US/US-IL/odell-5403" => array("type"=>"village", "pop"=>992, "name_ru"=>"Оделл", "name_en"=>"Odell", "lang"=>"en", "lat"=>41.002792, "lon"=>-88.522835),
"/US/US-IL/odin-5404" => array("type"=>"village", "pop"=>1061, "name_ru"=>"Один", "name_en"=>"Odin", "lang"=>"en", "lat"=>38.616379, "lon"=>-89.053918),
"/US/US-IL/ogden-5406" => array("type"=>"village", "pop"=>719, "name_ru"=>"Огден", "name_en"=>"Ogden", "lang"=>"en", "lat"=>40.115501, "lon"=>-87.955782),
"/US/US-IL/oglesby-5407" => array("type"=>"city", "pop"=>3671, "name_ru"=>"Оглсби", "name_en"=>"Oglesby", "lang"=>"en", "lat"=>41.294540, "lon"=>-89.064537),
"/US/US-IL/ohio-5408" => array("type"=>"village", "pop"=>518, "name_ru"=>"Огайо", "name_en"=>"Ohio", "lang"=>"en", "lat"=>41.556889, "lon"=>-89.459546),
"/US/US-IL/ohlman-5409" => array("type"=>"village", "pop"=>145, "name_ru"=>"Олман", "name_en"=>"Ohlman", "lang"=>"en", "lat"=>39.345529, "lon"=>-89.218285),
"/US/US-IL/okawville-5410" => array("type"=>"village", "pop"=>1337, "name_ru"=>"Оковилл", "name_en"=>"Okawville", "lang"=>"en", "lat"=>38.434176, "lon"=>-89.548260),
"/US/US-IL/old-mill-creek-5411" => array("type"=>"village", "pop"=>263, "name_ru"=>"Оулд Милл Крик", "name_en"=>"Old Mill Creek", "lang"=>"en", "lat"=>42.433100, "lon"=>-87.982107),
"/US/US-IL/old-ripley-5412" => array("type"=>"village", "pop"=>124, "name_ru"=>"Оулд Рипли", "name_en"=>"Old Ripley", "lang"=>"en", "lat"=>38.892349, "lon"=>-89.572954),
"/US/US-IL/old-shawneetown-5413" => array("type"=>"village", "pop"=>258, "name_ru"=>"Оулд Шонитаун", "name_en"=>"Old Shawneetown", "lang"=>"en", "lat"=>37.696927, "lon"=>-88.137842),
"/US/US-IL/olmsted-5414" => array("type"=>"village", "pop"=>261, "name_ru"=>"Олмстед", "name_en"=>"Olmsted", "lang"=>"en", "lat"=>37.181472, "lon"=>-89.083791),
"/US/US-IL/olney-5415" => array("type"=>"city", "pop"=>8407, "name_ru"=>"Олни", "name_en"=>"Olney", "lang"=>"en", "lat"=>38.731413, "lon"=>-88.085301),
"/US/US-IL/olympia-fields-5416" => array("type"=>"village", "pop"=>4735, "name_ru"=>"Олимпия Филдс", "name_en"=>"Olympia Fields", "lang"=>"en", "lat"=>41.517367, "lon"=>-87.692944),
"/US/US-IL/omaha-5417" => array("type"=>"village", "pop"=>244, "name_ru"=>"Омаха", "name_en"=>"Omaha", "lang"=>"en", "lat"=>37.889940, "lon"=>-88.304418),
"/US/US-IL/onarga-5418" => array("type"=>"village", "pop"=>1376, "name_ru"=>"Онарга", "name_en"=>"Onarga", "lang"=>"en", "lat"=>40.713123, "lon"=>-88.013249),
"/US/US-IL/oneida-5419" => array("type"=>"city", "pop"=>692, "name_ru"=>"Онида", "name_en"=>"Oneida", "lang"=>"en", "lat"=>41.072440, "lon"=>-90.225398),
"/US/US-IL/oquawka-5420" => array("type"=>"village", "pop"=>1378, "name_ru"=>"Оквока", "name_en"=>"Oquawka", "lang"=>"en", "lat"=>40.937574, "lon"=>-90.949548),
"/US/US-IL/orangeville-5421" => array("type"=>"village", "pop"=>759, "name_ru"=>"Оранжвилл", "name_en"=>"Orangeville", "lang"=>"en", "lat"=>42.468112, "lon"=>-89.644709),
"/US/US-IL/oreana-5422" => array("type"=>"village", "pop"=>924, "name_ru"=>"Ореана", "name_en"=>"Oreana", "lang"=>"en", "lat"=>39.937928, "lon"=>-88.869246),
"/US/US-IL/oregon-5423" => array("type"=>"city", "pop"=>4123, "name_ru"=>"Орегон", "name_en"=>"Oregon", "lang"=>"en", "lat"=>42.013054, "lon"=>-89.335378),
"/US/US-IL/orient-5424" => array("type"=>"city", "pop"=>312, "name_ru"=>"Ориент", "name_en"=>"Orient", "lang"=>"en", "lat"=>37.916410, "lon"=>-88.975365),
"/US/US-IL/orion-5425" => array("type"=>"village", "pop"=>1688, "name_ru"=>"Орион", "name_en"=>"Orion", "lang"=>"en", "lat"=>41.351192, "lon"=>-90.374075),
"/US/US-IL/orland-hills-5426" => array("type"=>"village", "pop"=>7195, "name_ru"=>"Орленд Хилс", "name_en"=>"Orland Hills", "lang"=>"en", "lat"=>41.592617, "lon"=>-87.842811),
"/US/US-IL/orland-park-5427" => array("type"=>"village", "pop"=>55514, "name_ru"=>"Орленд Парк", "name_en"=>"Orland Park", "lang"=>"en", "lat"=>41.610450, "lon"=>-87.850673),
"/US/US-IL/oswego-5428" => array("type"=>"village", "pop"=>31504, "name_ru"=>"Освего", "name_en"=>"Oswego", "lang"=>"en", "lat"=>41.697917, "lon"=>-88.332601),
"/US/US-IL/ottawa-5429" => array("type"=>"city", "pop"=>19419, "name_ru"=>"Оттава", "name_en"=>"Ottawa", "lang"=>"en", "lat"=>41.350243, "lon"=>-88.839395),
"/US/US-IL/otterville-5430" => array("type"=>"town", "pop"=>119, "name_ru"=>"Оттервилл", "name_en"=>"Otterville", "lang"=>"en", "lat"=>39.051026, "lon"=>-90.397486),
"/US/US-IL/owaneco-5431" => array("type"=>"village", "pop"=>244, "name_ru"=>"Ованеко", "name_en"=>"Owaneco", "lang"=>"en", "lat"=>39.481354, "lon"=>-89.194399),
"/US/US-IL/palatine-5432" => array("type"=>"village", "pop"=>67080, "name_ru"=>"Палатин", "name_en"=>"Palatine", "lang"=>"en", "lat"=>42.112500, "lon"=>-88.043024),
"/US/US-IL/palestine-5433" => array("type"=>"village", "pop"=>1324, "name_ru"=>"Палестин", "name_en"=>"Palestine", "lang"=>"en", "lat"=>39.002172, "lon"=>-87.612086),
"/US/US-IL/palmer-5434" => array("type"=>"village", "pop"=>232, "name_ru"=>"Палмер", "name_en"=>"Palmer", "lang"=>"en", "lat"=>39.458438, "lon"=>-89.407936),
"/US/US-IL/palmyra-5435" => array("type"=>"village", "pop"=>706, "name_ru"=>"Палмира", "name_en"=>"Palmyra", "lang"=>"en", "lat"=>39.434870, "lon"=>-89.996206),
"/US/US-IL/palos-heights-5436" => array("type"=>"city", "pop"=>12469, "name_ru"=>"Палос Хайтс", "name_en"=>"Palos Heights", "lang"=>"en", "lat"=>41.665300, "lon"=>-87.797149),
"/US/US-IL/palos-hills-5437" => array("type"=>"city", "pop"=>16924, "name_ru"=>"Палос Хилс", "name_en"=>"Palos Hills", "lang"=>"en", "lat"=>41.697600, "lon"=>-87.826124),
"/US/US-IL/palos-park-5438" => array("type"=>"village", "pop"=>4877, "name_ru"=>"Палос Парк", "name_en"=>"Palos Park", "lang"=>"en", "lat"=>41.664550, "lon"=>-87.844326),
"/US/US-IL/pana-5439" => array("type"=>"city", "pop"=>5718, "name_ru"=>"Пана", "name_en"=>"Pana", "lang"=>"en", "lat"=>39.386622, "lon"=>-89.078461),
"/US/US-IL/panama-5440" => array("type"=>"village", "pop"=>316, "name_ru"=>"Панама", "name_en"=>"Panama", "lang"=>"en", "lat"=>39.028710, "lon"=>-89.525171),
"/US/US-IL/panola-5441" => array("type"=>"village", "pop"=>33, "name_ru"=>"Панола", "name_en"=>"Panola", "lang"=>"en", "lat"=>40.785077, "lon"=>-89.020371),
"/US/US-IL/papineau-5442" => array("type"=>"village", "pop"=>199, "name_ru"=>"Папино", "name_en"=>"Papineau", "lang"=>"en", "lat"=>40.967704, "lon"=>-87.715949),
"/US/US-IL/paris-5443" => array("type"=>"city", "pop"=>8699, "name_ru"=>"Парис", "name_en"=>"Paris", "lang"=>"en", "lat"=>39.615536, "lon"=>-87.691507),
"/US/US-IL/park-city-5444" => array("type"=>"city", "pop"=>6609, "name_ru"=>"Парк Сити", "name_en"=>"Park City", "lang"=>"en", "lat"=>42.350750, "lon"=>-87.891018),
"/US/US-IL/park-forest-5446" => array("type"=>"village", "pop"=>22583, "name_ru"=>"Парк Форест", "name_en"=>"Park Forest", "lang"=>"en", "lat"=>41.482267, "lon"=>-87.686991),
"/US/US-IL/park-ridge-5447" => array("type"=>"city", "pop"=>36927, "name_ru"=>"Парк Ридж", "name_en"=>"Park Ridge", "lang"=>"en", "lat"=>42.011750, "lon"=>-87.843999),
"/US/US-IL/parkersburg-5445" => array("type"=>"village", "pop"=>222, "name_ru"=>"Паркерсбург", "name_en"=>"Parkersburg", "lang"=>"en", "lat"=>38.588701, "lon"=>-88.055684),
"/US/US-IL/patoka-5448" => array("type"=>"village", "pop"=>606, "name_ru"=>"Патока", "name_en"=>"Patoka", "lang"=>"en", "lat"=>38.753379, "lon"=>-89.096000),
"/US/US-IL/paw-paw-5450" => array("type"=>"village", "pop"=>876, "name_ru"=>"По По", "name_en"=>"Paw Paw", "lang"=>"en", "lat"=>41.687875, "lon"=>-88.980555),
"/US/US-IL/pawnee-5449" => array("type"=>"village", "pop"=>2544, "name_ru"=>"Пони", "name_en"=>"Pawnee", "lang"=>"en", "lat"=>39.592010, "lon"=>-89.580882),
"/US/US-IL/paxton-5451" => array("type"=>"city", "pop"=>4521, "name_ru"=>"Пэкстон", "name_en"=>"Paxton", "lang"=>"en", "lat"=>40.458367, "lon"=>-88.099462),
"/US/US-IL/payson-5452" => array("type"=>"village", "pop"=>1048, "name_ru"=>"Пейсон", "name_en"=>"Payson", "lang"=>"en", "lat"=>39.816783, "lon"=>-91.244105),
"/US/US-IL/pearl-5453" => array("type"=>"village", "pop"=>173, "name_ru"=>"Перл", "name_en"=>"Pearl", "lang"=>"en", "lat"=>39.458564, "lon"=>-90.624308),
"/US/US-IL/pearl-city-5454" => array("type"=>"village", "pop"=>771, "name_ru"=>"Перл Сити", "name_en"=>"Pearl City", "lang"=>"en", "lat"=>42.267673, "lon"=>-89.825540),
"/US/US-IL/pecatonica-5455" => array("type"=>"village", "pop"=>2209, "name_ru"=>"Пекатоника", "name_en"=>"Pecatonica", "lang"=>"en", "lat"=>42.309936, "lon"=>-89.358535),
"/US/US-IL/pekin-5456" => array("type"=>"city", "pop"=>33430, "name_ru"=>"Пекин", "name_en"=>"Pekin", "lang"=>"en", "lat"=>40.568233, "lon"=>-89.628095),
"/US/US-IL/peoria-5457" => array("type"=>"city", "pop"=>114114, "name_ru"=>"Пеория", "name_en"=>"Peoria", "lang"=>"en", "lat"=>40.744964, "lon"=>-89.609225),
"/US/US-IL/peoria-heights-5458" => array("type"=>"village", "pop"=>6230, "name_ru"=>"Пеория Хайтс", "name_en"=>"Peoria Heights", "lang"=>"en", "lat"=>40.756783, "lon"=>-89.557070),
"/US/US-IL/peotone-5459" => array("type"=>"village", "pop"=>4294, "name_ru"=>"Пеотоун", "name_en"=>"Peotone", "lang"=>"en", "lat"=>41.331642, "lon"=>-87.796277),
"/US/US-IL/percy-5460" => array("type"=>"village", "pop"=>910, "name_ru"=>"Перси", "name_en"=>"Percy", "lang"=>"en", "lat"=>38.015957, "lon"=>-89.616829),
"/US/US-IL/perry-5461" => array("type"=>"village", "pop"=>399, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>39.782433, "lon"=>-90.747241),
"/US/US-IL/peru-5462" => array("type"=>"city", "pop"=>9810, "name_ru"=>"Перу", "name_en"=>"Peru", "lang"=>"en", "lat"=>41.341745, "lon"=>-89.128290),
"/US/US-IL/pesotum-5463" => array("type"=>"village", "pop"=>499, "name_ru"=>"Песотум", "name_en"=>"Pesotum", "lang"=>"en", "lat"=>39.912903, "lon"=>-88.274051),
"/US/US-IL/petersburg-5464" => array("type"=>"city", "pop"=>2190, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>40.012359, "lon"=>-89.851412),
"/US/US-IL/phillipstown-5465" => array("type"=>"village", "pop"=>27, "name_ru"=>"Филлипстаун", "name_en"=>"Phillipstown", "lang"=>"en", "lat"=>38.140667, "lon"=>-88.019156),
"/US/US-IL/philo-5466" => array("type"=>"village", "pop"=>1608, "name_ru"=>"Фило", "name_en"=>"Philo", "lang"=>"en", "lat"=>40.003136, "lon"=>-88.157959),
"/US/US-IL/phoenix-5467" => array("type"=>"village", "pop"=>2041, "name_ru"=>"Феникс", "name_en"=>"Phoenix", "lang"=>"en", "lat"=>41.612350, "lon"=>-87.630473),
"/US/US-IL/pierron-5468" => array("type"=>"village", "pop"=>646, "name_ru"=>"Пиррон", "name_en"=>"Pierron", "lang"=>"en", "lat"=>38.778912, "lon"=>-89.566297),
"/US/US-IL/pinckneyville-5469" => array("type"=>"city", "pop"=>5384, "name_ru"=>"Пинкнивилл", "name_en"=>"Pinckneyville", "lang"=>"en", "lat"=>38.078145, "lon"=>-89.384212),
"/US/US-IL/pingree-grove-5470" => array("type"=>"village", "pop"=>4075, "name_ru"=>"Пингри Гров", "name_en"=>"Pingree Grove", "lang"=>"en", "lat"=>42.068600, "lon"=>-88.414706),
"/US/US-IL/piper-city-5471" => array("type"=>"village", "pop"=>736, "name_ru"=>"Пайпер Сити", "name_en"=>"Piper City", "lang"=>"en", "lat"=>40.755707, "lon"=>-88.187654),
"/US/US-IL/pistakee-highlands-5472" => array("type"=>"cdp", "pop"=>3848, "name_ru"=>"Пистейки Хайлендс", "name_en"=>"Pistakee Highlands", "lang"=>"en", "lat"=>42.403157, "lon"=>-88.212162),
"/US/US-IL/pittsburg-5473" => array("type"=>"village", "pop"=>581, "name_ru"=>"Питсбург", "name_en"=>"Pittsburg", "lang"=>"en", "lat"=>37.776669, "lon"=>-88.850276),
"/US/US-IL/pittsfield-5474" => array("type"=>"city", "pop"=>4453, "name_ru"=>"Питтсфилд", "name_en"=>"Pittsfield", "lang"=>"en", "lat"=>39.610708, "lon"=>-90.809288),
"/US/US-IL/plainfield-5475" => array("type"=>"village", "pop"=>36507, "name_ru"=>"Плейнфилд", "name_en"=>"Plainfield", "lang"=>"en", "lat"=>41.615915, "lon"=>-88.204069),
"/US/US-IL/plainville-5476" => array("type"=>"village", "pop"=>242, "name_ru"=>"Плейнвилл", "name_en"=>"Plainville", "lang"=>"en", "lat"=>39.784045, "lon"=>-91.181608),
"/US/US-IL/plano-5477" => array("type"=>"city", "pop"=>11886, "name_ru"=>"Плано", "name_en"=>"Plano", "lang"=>"en", "lat"=>41.663325, "lon"=>-88.534831),
"/US/US-IL/pleasant-hill-5478" => array("type"=>"village", "pop"=>970, "name_ru"=>"Плезант Хилл", "name_en"=>"Pleasant Hill", "lang"=>"en", "lat"=>39.444646, "lon"=>-90.872933),
"/US/US-IL/pleasant-plains-5479" => array("type"=>"village", "pop"=>720, "name_ru"=>"Плезант Плейнс", "name_en"=>"Pleasant Plains", "lang"=>"en", "lat"=>39.873209, "lon"=>-89.920016),
"/US/US-IL/plymouth-5480" => array("type"=>"village", "pop"=>520, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>40.291184, "lon"=>-90.917000),
"/US/US-IL/pocahontas-5481" => array("type"=>"village", "pop"=>717, "name_ru"=>"Покахонтас", "name_en"=>"Pocahontas", "lang"=>"en", "lat"=>38.824480, "lon"=>-89.538969),
"/US/US-IL/polo-5482" => array("type"=>"city", "pop"=>2477, "name_ru"=>"Поло", "name_en"=>"Polo", "lang"=>"en", "lat"=>41.984685, "lon"=>-89.578330),
"/US/US-IL/pontiac-5483" => array("type"=>"city", "pop"=>11258, "name_ru"=>"Понтиак", "name_en"=>"Pontiac", "lang"=>"en", "lat"=>40.876635, "lon"=>-88.641837),
"/US/US-IL/pontoon-beach-5484" => array("type"=>"village", "pop"=>6081, "name_ru"=>"Понтун Бич", "name_en"=>"Pontoon Beach", "lang"=>"en", "lat"=>38.726525, "lon"=>-90.051133),
"/US/US-IL/pontoosuc-5485" => array("type"=>"village", "pop"=>159, "name_ru"=>"Понтузук", "name_en"=>"Pontoosuc", "lang"=>"en", "lat"=>40.628563, "lon"=>-91.211607),
"/US/US-IL/poplar-grove-5486" => array("type"=>"village", "pop"=>4058, "name_ru"=>"Поплар Гров", "name_en"=>"Poplar Grove", "lang"=>"en", "lat"=>42.367300, "lon"=>-88.822887),
"/US/US-IL/port-byron-5487" => array("type"=>"village", "pop"=>1670, "name_ru"=>"Порт Бирон", "name_en"=>"Port Byron", "lang"=>"en", "lat"=>41.621252, "lon"=>-90.323718),
"/US/US-IL/posen-5488" => array("type"=>"village", "pop"=>4907, "name_ru"=>"Посен", "name_en"=>"Posen", "lang"=>"en", "lat"=>41.628200, "lon"=>-87.685674),
"/US/US-IL/potomac-5489" => array("type"=>"village", "pop"=>656, "name_ru"=>"Потомак", "name_en"=>"Potomac", "lang"=>"en", "lat"=>40.306395, "lon"=>-87.797248),
"/US/US-IL/prairie-city-5490" => array("type"=>"village", "pop"=>418, "name_ru"=>"Прейри Сити", "name_en"=>"Prairie City", "lang"=>"en", "lat"=>40.620604, "lon"=>-90.464500),
"/US/US-IL/prairie-du-rocher-5491" => array("type"=>"village", "pop"=>568, "name_ru"=>"Прейри дю Рочер", "name_en"=>"Prairie du Rocher", "lang"=>"en", "lat"=>38.081280, "lon"=>-90.097607),
"/US/US-IL/prairie-grove-5492" => array("type"=>"village", "pop"=>1995, "name_ru"=>"Прейри Гров", "name_en"=>"Prairie Grove", "lang"=>"en", "lat"=>42.280350, "lon"=>-88.266455),
"/US/US-IL/preston-heights-5493" => array("type"=>"cdp", "pop"=>2750, "name_ru"=>"Престон Хайтс", "name_en"=>"Preston Heights", "lang"=>"en", "lat"=>41.492800, "lon"=>-88.073101),
"/US/US-IL/princeton-5494" => array("type"=>"city", "pop"=>7496, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>41.378445, "lon"=>-89.466832),
"/US/US-IL/princeville-5495" => array("type"=>"village", "pop"=>1636, "name_ru"=>"Принсвилл", "name_en"=>"Princeville", "lang"=>"en", "lat"=>40.931537, "lon"=>-89.755599),
"/US/US-IL/prophetstown-5496" => array("type"=>"city", "pop"=>1925, "name_ru"=>"Профетстаун", "name_en"=>"Prophetstown", "lang"=>"en", "lat"=>41.669651, "lon"=>-89.934591),
"/US/US-IL/prospect-heights-5497" => array("type"=>"city", "pop"=>16047, "name_ru"=>"Проспект Хайтс", "name_en"=>"Prospect Heights", "lang"=>"en", "lat"=>42.102401, "lon"=>-87.924748),
"/US/US-IL/pulaski-5498" => array("type"=>"village", "pop"=>228, "name_ru"=>"Пуласки", "name_en"=>"Pulaski", "lang"=>"en", "lat"=>37.216013, "lon"=>-89.206953),
"/US/US-IL/quincy-5499" => array("type"=>"city", "pop"=>39983, "name_ru"=>"Квинси", "name_en"=>"Quincy", "lang"=>"en", "lat"=>39.933536, "lon"=>-91.388588),
"/US/US-IL/radom-5500" => array("type"=>"village", "pop"=>383, "name_ru"=>"Рэдом", "name_en"=>"Radom", "lang"=>"en", "lat"=>38.280596, "lon"=>-89.191838),
"/US/US-IL/raleigh-5501" => array("type"=>"village", "pop"=>319, "name_ru"=>"Роли", "name_en"=>"Raleigh", "lang"=>"en", "lat"=>37.825292, "lon"=>-88.531622),
"/US/US-IL/ramsey-5502" => array("type"=>"village", "pop"=>1047, "name_ru"=>"Рэмси", "name_en"=>"Ramsey", "lang"=>"en", "lat"=>39.143841, "lon"=>-89.109939),
"/US/US-IL/rankin-5503" => array("type"=>"village", "pop"=>584, "name_ru"=>"Ранкин", "name_en"=>"Rankin", "lang"=>"en", "lat"=>40.464741, "lon"=>-87.896053),
"/US/US-IL/ransom-5504" => array("type"=>"village", "pop"=>383, "name_ru"=>"Рэнсам", "name_en"=>"Ransom", "lang"=>"en", "lat"=>41.158072, "lon"=>-88.655272),
"/US/US-IL/rantoul-5505" => array("type"=>"village", "pop"=>12189, "name_ru"=>"Рантул", "name_en"=>"Rantoul", "lang"=>"en", "lat"=>40.303368, "lon"=>-88.155535),
"/US/US-IL/rapids-city-5506" => array("type"=>"village", "pop"=>990, "name_ru"=>"Рэпидс Сити", "name_en"=>"Rapids City", "lang"=>"en", "lat"=>41.581900, "lon"=>-90.341414),
"/US/US-IL/raritan-5507" => array("type"=>"village", "pop"=>122, "name_ru"=>"Раритан", "name_en"=>"Raritan", "lang"=>"en", "lat"=>40.695353, "lon"=>-90.824979),
"/US/US-IL/raymond-5508" => array("type"=>"village", "pop"=>899, "name_ru"=>"Реймонд", "name_en"=>"Raymond", "lang"=>"en", "lat"=>39.320216, "lon"=>-89.573738),
"/US/US-IL/red-bud-5509" => array("type"=>"city", "pop"=>3636, "name_ru"=>"Ред Бад", "name_en"=>"Red Bud", "lang"=>"en", "lat"=>38.210155, "lon"=>-89.996387),
"/US/US-IL/reddick-5510" => array("type"=>"village", "pop"=>225, "name_ru"=>"Реддик", "name_en"=>"Reddick", "lang"=>"en", "lat"=>41.097252, "lon"=>-88.248792),
"/US/US-IL/redmon-5511" => array("type"=>"village", "pop"=>189, "name_ru"=>"Редмон", "name_en"=>"Redmon", "lang"=>"en", "lat"=>39.644988, "lon"=>-87.861754),
"/US/US-IL/reynolds-5512" => array("type"=>"village", "pop"=>482, "name_ru"=>"Рейнолдс", "name_en"=>"Reynolds", "lang"=>"en", "lat"=>41.331315, "lon"=>-90.672797),
"/US/US-IL/richmond-5513" => array("type"=>"village", "pop"=>2316, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>42.481200, "lon"=>-88.307324),
"/US/US-IL/richton-park-5514" => array("type"=>"village", "pop"=>12842, "name_ru"=>"Рихтон Парк", "name_en"=>"Richton Park", "lang"=>"en", "lat"=>41.481950, "lon"=>-87.728328),
"/US/US-IL/richview-5515" => array("type"=>"village", "pop"=>286, "name_ru"=>"Ричвью", "name_en"=>"Richview", "lang"=>"en", "lat"=>38.375619, "lon"=>-89.183731),
"/US/US-IL/ridge-farm-5516" => array("type"=>"village", "pop"=>857, "name_ru"=>"Ридж Фарм", "name_en"=>"Ridge Farm", "lang"=>"en", "lat"=>39.895344, "lon"=>-87.652267),
"/US/US-IL/ridgway-5517" => array("type"=>"village", "pop"=>860, "name_ru"=>"Риджуэй", "name_en"=>"Ridgway", "lang"=>"en", "lat"=>37.797990, "lon"=>-88.260532),
"/US/US-IL/ridott-5518" => array("type"=>"village", "pop"=>147, "name_ru"=>"Ридотт", "name_en"=>"Ridott", "lang"=>"en", "lat"=>42.297522, "lon"=>-89.477324),
"/US/US-IL/rio-5519" => array("type"=>"village", "pop"=>221, "name_ru"=>"Рио", "name_en"=>"Rio", "lang"=>"en", "lat"=>41.108656, "lon"=>-90.398957),
"/US/US-IL/ripley-5520" => array("type"=>"village", "pop"=>93, "name_ru"=>"Рипли", "name_en"=>"Ripley", "lang"=>"en", "lat"=>40.024969, "lon"=>-90.637726),
"/US/US-IL/river-forest-5522" => array("type"=>"village", "pop"=>11176, "name_ru"=>"Ривер Форест", "name_en"=>"River Forest", "lang"=>"en", "lat"=>41.894850, "lon"=>-87.819046),
"/US/US-IL/river-grove-5523" => array("type"=>"village", "pop"=>9992, "name_ru"=>"Ривер Гров", "name_en"=>"River Grove", "lang"=>"en", "lat"=>41.924000, "lon"=>-87.837574),
"/US/US-IL/riverdale-5521" => array("type"=>"village", "pop"=>14163, "name_ru"=>"Ривердейл", "name_en"=>"Riverdale", "lang"=>"en", "lat"=>41.644600, "lon"=>-87.634662),
"/US/US-IL/riverside-5524" => array("type"=>"village", "pop"=>8272, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>41.830850, "lon"=>-87.815924),
"/US/US-IL/riverton-5525" => array("type"=>"village", "pop"=>2989, "name_ru"=>"Ривертон", "name_en"=>"Riverton", "lang"=>"en", "lat"=>39.849240, "lon"=>-89.540085),
"/US/US-IL/riverwoods-5526" => array("type"=>"village", "pop"=>4081, "name_ru"=>"Ривервудс", "name_en"=>"Riverwoods", "lang"=>"en", "lat"=>42.171050, "lon"=>-87.895874),
"/US/US-IL/roanoke-5527" => array("type"=>"village", "pop"=>2017, "name_ru"=>"Роанок", "name_en"=>"Roanoke", "lang"=>"en", "lat"=>40.796288, "lon"=>-89.201688),
"/US/US-IL/robbins-5528" => array("type"=>"village", "pop"=>6331, "name_ru"=>"Роббинс", "name_en"=>"Robbins", "lang"=>"en", "lat"=>41.642900, "lon"=>-87.708099),
"/US/US-IL/roberts-5529" => array("type"=>"village", "pop"=>380, "name_ru"=>"Робертс", "name_en"=>"Roberts", "lang"=>"en", "lat"=>40.614083, "lon"=>-88.183481),
"/US/US-IL/robinson-5530" => array("type"=>"city", "pop"=>6369, "name_ru"=>"Робинсон", "name_en"=>"Robinson", "lang"=>"en", "lat"=>39.007854, "lon"=>-87.740471),
"/US/US-IL/rochelle-5531" => array("type"=>"city", "pop"=>9847, "name_ru"=>"Рошелл", "name_en"=>"Rochelle", "lang"=>"en", "lat"=>41.918433, "lon"=>-89.062745),
"/US/US-IL/rochester-5532" => array("type"=>"village", "pop"=>3236, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>39.750576, "lon"=>-89.541493),
"/US/US-IL/rock-city-5534" => array("type"=>"village", "pop"=>313, "name_ru"=>"Рок Сити", "name_en"=>"Rock City", "lang"=>"en", "lat"=>42.412624, "lon"=>-89.470782),
"/US/US-IL/rock-falls-5536" => array("type"=>"city", "pop"=>9307, "name_ru"=>"Рок Фоллс", "name_en"=>"Rock Falls", "lang"=>"en", "lat"=>41.774174, "lon"=>-89.691113),
"/US/US-IL/rock-island-5538" => array("type"=>"city", "pop"=>38139, "name_ru"=>"Рок Айленд", "name_en"=>"Rock Island", "lang"=>"en", "lat"=>41.474600, "lon"=>-90.580299),
"/US/US-IL/rockbridge-5533" => array("type"=>"village", "pop"=>176, "name_ru"=>"Рокбридж", "name_en"=>"Rockbridge", "lang"=>"en", "lat"=>39.272935, "lon"=>-90.208917),
"/US/US-IL/rockdale-5535" => array("type"=>"village", "pop"=>1993, "name_ru"=>"Рокдейл", "name_en"=>"Rockdale", "lang"=>"en", "lat"=>41.505500, "lon"=>-88.118246),
"/US/US-IL/rockford-5537" => array("type"=>"city", "pop"=>157272, "name_ru"=>"Рокфорд", "name_en"=>"Rockford", "lang"=>"en", "lat"=>42.270300, "lon"=>-89.063149),
"/US/US-IL/rockton-5539" => array("type"=>"village", "pop"=>5512, "name_ru"=>"Роктон", "name_en"=>"Rockton", "lang"=>"en", "lat"=>42.449308, "lon"=>-89.069923),
"/US/US-IL/rockwood-5540" => array("type"=>"village", "pop"=>40, "name_ru"=>"Роквуд", "name_en"=>"Rockwood", "lang"=>"en", "lat"=>37.837867, "lon"=>-89.697439),
"/US/US-IL/rolling-meadows-5541" => array("type"=>"city", "pop"=>23463, "name_ru"=>"Роллинг Медоус", "name_en"=>"Rolling Meadows", "lang"=>"en", "lat"=>42.075033, "lon"=>-88.025016),
"/US/US-IL/rome-5542" => array("type"=>"cdp", "pop"=>1845, "name_ru"=>"Роум", "name_en"=>"Rome", "lang"=>"en", "lat"=>40.878383, "lon"=>-89.511709),
"/US/US-IL/romeoville-5543" => array("type"=>"village", "pop"=>38028, "name_ru"=>"Ромеовилл", "name_en"=>"Romeoville", "lang"=>"en", "lat"=>41.646050, "lon"=>-88.086299),
"/US/US-IL/roodhouse-5544" => array("type"=>"city", "pop"=>1937, "name_ru"=>"Рудхаус", "name_en"=>"Roodhouse", "lang"=>"en", "lat"=>39.484648, "lon"=>-90.373404),
"/US/US-IL/roscoe-5545" => array("type"=>"village", "pop"=>8807, "name_ru"=>"Роско", "name_en"=>"Roscoe", "lang"=>"en", "lat"=>42.416776, "lon"=>-89.006979),
"/US/US-IL/rose-hill-5546" => array("type"=>"village", "pop"=>76, "name_ru"=>"Роуз Хилл", "name_en"=>"Rose Hill", "lang"=>"en", "lat"=>39.104142, "lon"=>-88.151243),
"/US/US-IL/roselle-5547" => array("type"=>"village", "pop"=>23209, "name_ru"=>"Розл", "name_en"=>"Roselle", "lang"=>"en", "lat"=>41.980450, "lon"=>-88.083149),
"/US/US-IL/rosemont-5548" => array("type"=>"village", "pop"=>3929, "name_ru"=>"Роузмонт", "name_en"=>"Rosemont", "lang"=>"en", "lat"=>41.989100, "lon"=>-87.871474),
"/US/US-IL/roseville-5549" => array("type"=>"village", "pop"=>979, "name_ru"=>"Роузвилл", "name_en"=>"Roseville", "lang"=>"en", "lat"=>40.730809, "lon"=>-90.662709),
"/US/US-IL/rosewood-heights-5550" => array("type"=>"cdp", "pop"=>4821, "name_ru"=>"Роузвуд Хайтс", "name_en"=>"Rosewood Heights", "lang"=>"en", "lat"=>38.885750, "lon"=>-90.070269),
"/US/US-IL/rosiclare-5551" => array("type"=>"city", "pop"=>1085, "name_ru"=>"Розиклэр", "name_en"=>"Rosiclare", "lang"=>"en", "lat"=>37.422393, "lon"=>-88.350710),
"/US/US-IL/rossville-5552" => array("type"=>"village", "pop"=>1153, "name_ru"=>"Россвилл", "name_en"=>"Rossville", "lang"=>"en", "lat"=>40.381761, "lon"=>-87.669131),
"/US/US-IL/round-lake-5553" => array("type"=>"village", "pop"=>16748, "name_ru"=>"Раунд Лейк", "name_en"=>"Round Lake", "lang"=>"en", "lat"=>42.349950, "lon"=>-88.110336),
"/US/US-IL/round-lake-beach-5554" => array("type"=>"village", "pop"=>27855, "name_ru"=>"Раунд Лейк Бич", "name_en"=>"Round Lake Beach", "lang"=>"en", "lat"=>42.377600, "lon"=>-88.081697),
"/US/US-IL/round-lake-heights-5555" => array("type"=>"village", "pop"=>2802, "name_ru"=>"Раунд Лейк Хайтс", "name_en"=>"Round Lake Heights", "lang"=>"en", "lat"=>42.384000, "lon"=>-88.104102),
"/US/US-IL/round-lake-park-5556" => array("type"=>"village", "pop"=>6181, "name_ru"=>"Раунд Лейк Парк", "name_en"=>"Round Lake Park", "lang"=>"en", "lat"=>42.323221, "lon"=>-88.058559),
"/US/US-IL/roxana-5557" => array("type"=>"village", "pop"=>1552, "name_ru"=>"Роксана", "name_en"=>"Roxana", "lang"=>"en", "lat"=>38.836650, "lon"=>-90.052631),
"/US/US-IL/royal-5558" => array("type"=>"village", "pop"=>267, "name_ru"=>"Ройал", "name_en"=>"Royal", "lang"=>"en", "lat"=>40.191981, "lon"=>-87.971346),
"/US/US-IL/royal-lakes-5559" => array("type"=>"village", "pop"=>184, "name_ru"=>"Ройал Лейкс", "name_en"=>"Royal Lakes", "lang"=>"en", "lat"=>39.112486, "lon"=>-89.961790),
"/US/US-IL/royalton-5560" => array("type"=>"village", "pop"=>1159, "name_ru"=>"Ройалтон", "name_en"=>"Royalton", "lang"=>"en", "lat"=>37.877663, "lon"=>-89.113489),
"/US/US-IL/ruma-5561" => array("type"=>"village", "pop"=>285, "name_ru"=>"Рума", "name_en"=>"Ruma", "lang"=>"en", "lat"=>38.134433, "lon"=>-89.998396),
"/US/US-IL/rushville-5562" => array("type"=>"city", "pop"=>3096, "name_ru"=>"Рашвилл", "name_en"=>"Rushville", "lang"=>"en", "lat"=>40.119795, "lon"=>-90.565715),
"/US/US-IL/russellville-5563" => array("type"=>"village", "pop"=>114, "name_ru"=>"Расселлвилл", "name_en"=>"Russellville", "lang"=>"en", "lat"=>38.819951, "lon"=>-87.529254),
"/US/US-IL/rutland-5564" => array("type"=>"village", "pop"=>348, "name_ru"=>"Рутленд", "name_en"=>"Rutland", "lang"=>"en", "lat"=>40.984073, "lon"=>-89.039312),
"/US/US-IL/sadorus-5565" => array("type"=>"village", "pop"=>395, "name_ru"=>"Садорус", "name_en"=>"Sadorus", "lang"=>"en", "lat"=>39.966979, "lon"=>-88.345384),
"/US/US-IL/sailor-springs-5566" => array("type"=>"village", "pop"=>123, "name_ru"=>"Сейлор Спрингс", "name_en"=>"Sailor Springs", "lang"=>"en", "lat"=>38.764755, "lon"=>-88.358714),
"/US/US-IL/salem-5579" => array("type"=>"city", "pop"=>7446, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>38.628275, "lon"=>-88.951823),
"/US/US-IL/san-jose-5582" => array("type"=>"village", "pop"=>646, "name_ru"=>"Сан-Хосе", "name_en"=>"San Jose", "lang"=>"en", "lat"=>40.306002, "lon"=>-89.603363),
"/US/US-IL/sandoval-5580" => array("type"=>"village", "pop"=>1335, "name_ru"=>"Сандовал", "name_en"=>"Sandoval", "lang"=>"en", "lat"=>38.611273, "lon"=>-89.119349),
"/US/US-IL/sandwich-5581" => array("type"=>"city", "pop"=>7337, "name_ru"=>"Сэндвич", "name_en"=>"Sandwich", "lang"=>"en", "lat"=>41.646579, "lon"=>-88.623081),
"/US/US-IL/sauget-5583" => array("type"=>"village", "pop"=>237, "name_ru"=>"Согет", "name_en"=>"Sauget", "lang"=>"en", "lat"=>38.589367, "lon"=>-90.169658),
"/US/US-IL/sauk-village-5584" => array("type"=>"village", "pop"=>10254, "name_ru"=>"Сок Виллидж", "name_en"=>"Sauk Village", "lang"=>"en", "lat"=>41.488500, "lon"=>-87.565617),
"/US/US-IL/saunemin-5585" => array("type"=>"village", "pop"=>424, "name_ru"=>"Сонемин", "name_en"=>"Saunemin", "lang"=>"en", "lat"=>40.892856, "lon"=>-88.406477),
"/US/US-IL/savanna-5586" => array("type"=>"city", "pop"=>3211, "name_ru"=>"Саванна", "name_en"=>"Savanna", "lang"=>"en", "lat"=>42.090047, "lon"=>-90.139923),
"/US/US-IL/savoy-5587" => array("type"=>"village", "pop"=>5579, "name_ru"=>"Савой", "name_en"=>"Savoy", "lang"=>"en", "lat"=>40.065036, "lon"=>-88.252710),
"/US/US-IL/sawyerville-5588" => array("type"=>"village", "pop"=>284, "name_ru"=>"Сойервилл", "name_en"=>"Sawyerville", "lang"=>"en", "lat"=>39.076652, "lon"=>-89.802993),
"/US/US-IL/saybrook-5589" => array("type"=>"village", "pop"=>746, "name_ru"=>"Сейбрук", "name_en"=>"Saybrook", "lang"=>"en", "lat"=>40.428258, "lon"=>-88.526492),
"/US/US-IL/scales-mound-5590" => array("type"=>"village", "pop"=>381, "name_ru"=>"Скейлс Маунд", "name_en"=>"Scales Mound", "lang"=>"en", "lat"=>42.477461, "lon"=>-90.252011),
"/US/US-IL/schaumburg-5591" => array("type"=>"village", "pop"=>71716, "name_ru"=>"Шаамбург", "name_en"=>"Schaumburg", "lang"=>"en", "lat"=>42.033000, "lon"=>-88.083364),
"/US/US-IL/schiller-park-5592" => array("type"=>"village", "pop"=>11573, "name_ru"=>"Шиллер Парк", "name_en"=>"Schiller Park", "lang"=>"en", "lat"=>41.959827, "lon"=>-87.870024),
"/US/US-IL/schram-city-5593" => array("type"=>"village", "pop"=>614, "name_ru"=>"Шрам Сити", "name_en"=>"Schram City", "lang"=>"en", "lat"=>39.162589, "lon"=>-89.460576),
"/US/US-IL/sciota-5594" => array("type"=>"village", "pop"=>52, "name_ru"=>"Шиота", "name_en"=>"Sciota", "lang"=>"en", "lat"=>40.561946, "lon"=>-90.748986),
"/US/US-IL/scott-afb-5595" => array("type"=>"cdp", "pop"=>7245, "name_ru"=>"Военно-воздушная база Скотт", "name_en"=>"Scott AFB", "lang"=>"en", "lat"=>38.544249, "lon"=>-89.850445),
"/US/US-IL/scottville-5596" => array("type"=>"village", "pop"=>135, "name_ru"=>"Скоттвилл", "name_en"=>"Scottville", "lang"=>"en", "lat"=>39.477465, "lon"=>-90.103300),
"/US/US-IL/seaton-5597" => array("type"=>"village", "pop"=>226, "name_ru"=>"Ситон", "name_en"=>"Seaton", "lang"=>"en", "lat"=>41.101901, "lon"=>-90.799209),
"/US/US-IL/seatonville-5598" => array("type"=>"village", "pop"=>308, "name_ru"=>"Ситонвилл", "name_en"=>"Seatonville", "lang"=>"en", "lat"=>41.363787, "lon"=>-89.272887),
"/US/US-IL/secor-5599" => array("type"=>"village", "pop"=>372, "name_ru"=>"Секор", "name_en"=>"Secor", "lang"=>"en", "lat"=>40.740476, "lon"=>-89.135765),
"/US/US-IL/seneca-5600" => array("type"=>"village", "pop"=>2126, "name_ru"=>"Сенека", "name_en"=>"Seneca", "lang"=>"en", "lat"=>41.303370, "lon"=>-88.613521),
"/US/US-IL/sesser-5601" => array("type"=>"city", "pop"=>2121, "name_ru"=>"Сессер", "name_en"=>"Sesser", "lang"=>"en", "lat"=>38.090251, "lon"=>-89.050528),
"/US/US-IL/shabbona-5602" => array("type"=>"village", "pop"=>957, "name_ru"=>"Шаббона", "name_en"=>"Shabbona", "lang"=>"en", "lat"=>41.764856, "lon"=>-88.876403),
"/US/US-IL/shannon-5603" => array("type"=>"village", "pop"=>795, "name_ru"=>"Шэннон", "name_en"=>"Shannon", "lang"=>"en", "lat"=>42.151976, "lon"=>-89.740111),
"/US/US-IL/shawneetown-5604" => array("type"=>"city", "pop"=>1308, "name_ru"=>"Шонитаун", "name_en"=>"Shawneetown", "lang"=>"en", "lat"=>37.712801, "lon"=>-88.165498),
"/US/US-IL/sheffield-5605" => array("type"=>"village", "pop"=>935, "name_ru"=>"Шеффилд", "name_en"=>"Sheffield", "lang"=>"en", "lat"=>41.356999, "lon"=>-89.739583),
"/US/US-IL/shelbyville-5606" => array("type"=>"city", "pop"=>4605, "name_ru"=>"Шелбивилл", "name_en"=>"Shelbyville", "lang"=>"en", "lat"=>39.407051, "lon"=>-88.804026),
"/US/US-IL/sheldon-5607" => array("type"=>"village", "pop"=>1154, "name_ru"=>"Шелдон", "name_en"=>"Sheldon", "lang"=>"en", "lat"=>40.770869, "lon"=>-87.566401),
"/US/US-IL/sheridan-5608" => array("type"=>"village", "pop"=>1963, "name_ru"=>"Шеридан", "name_en"=>"Sheridan", "lang"=>"en", "lat"=>41.525927, "lon"=>-88.686669),
"/US/US-IL/sherman-5609" => array("type"=>"village", "pop"=>3827, "name_ru"=>"Шерман", "name_en"=>"Sherman", "lang"=>"en", "lat"=>39.886643, "lon"=>-89.608970),
"/US/US-IL/sherrard-5610" => array("type"=>"village", "pop"=>645, "name_ru"=>"Шеррад", "name_en"=>"Sherrard", "lang"=>"en", "lat"=>41.318435, "lon"=>-90.505264),
"/US/US-IL/shiloh-5611" => array("type"=>"village", "pop"=>11173, "name_ru"=>"Шилох", "name_en"=>"Shiloh", "lang"=>"en", "lat"=>38.553531, "lon"=>-89.905217),
"/US/US-IL/shipman-5612" => array("type"=>"town", "pop"=>634, "name_ru"=>"Шипман", "name_en"=>"Shipman", "lang"=>"en", "lat"=>39.120789, "lon"=>-90.045196),
"/US/US-IL/shorewood-5613" => array("type"=>"village", "pop"=>15521, "name_ru"=>"Шорвуд", "name_en"=>"Shorewood", "lang"=>"en", "lat"=>41.523200, "lon"=>-88.202374),
"/US/US-IL/shumway-5614" => array("type"=>"village", "pop"=>227, "name_ru"=>"Шамуэй", "name_en"=>"Shumway", "lang"=>"en", "lat"=>39.183211, "lon"=>-88.652974),
"/US/US-IL/sibley-5615" => array("type"=>"village", "pop"=>323, "name_ru"=>"Сибли", "name_en"=>"Sibley", "lang"=>"en", "lat"=>40.587553, "lon"=>-88.378078),
"/US/US-IL/sidell-5616" => array("type"=>"village", "pop"=>592, "name_ru"=>"Сиделл", "name_en"=>"Sidell", "lang"=>"en", "lat"=>39.910204, "lon"=>-87.821997),
"/US/US-IL/sidney-5617" => array("type"=>"village", "pop"=>1127, "name_ru"=>"Сидни", "name_en"=>"Sidney", "lang"=>"en", "lat"=>40.024376, "lon"=>-88.071838),
"/US/US-IL/sigel-5618" => array("type"=>"town", "pop"=>389, "name_ru"=>"Сигел", "name_en"=>"Sigel", "lang"=>"en", "lat"=>39.226175, "lon"=>-88.495113),
"/US/US-IL/silvis-5619" => array("type"=>"city", "pop"=>7770, "name_ru"=>"Силвис", "name_en"=>"Silvis", "lang"=>"en", "lat"=>41.506250, "lon"=>-90.412695),
"/US/US-IL/simpson-5620" => array("type"=>"village", "pop"=>56, "name_ru"=>"Симпсон", "name_en"=>"Simpson", "lang"=>"en", "lat"=>37.467591, "lon"=>-88.754964),
"/US/US-IL/sims-5621" => array("type"=>"village", "pop"=>262, "name_ru"=>"Симс", "name_en"=>"Sims", "lang"=>"en", "lat"=>38.361195, "lon"=>-88.534975),
"/US/US-IL/skokie-5622" => array("type"=>"village", "pop"=>66620, "name_ru"=>"Скоки", "name_en"=>"Skokie", "lang"=>"en", "lat"=>42.037050, "lon"=>-87.739946),
"/US/US-IL/sleepy-hollow-5623" => array("type"=>"village", "pop"=>3687, "name_ru"=>"Слипи Холлоу", "name_en"=>"Sleepy Hollow", "lang"=>"en", "lat"=>42.093450, "lon"=>-88.314157),
"/US/US-IL/smithboro-5624" => array("type"=>"village", "pop"=>232, "name_ru"=>"Смитборо", "name_en"=>"Smithboro", "lang"=>"en", "lat"=>38.894869, "lon"=>-89.341115),
"/US/US-IL/smithfield-5625" => array("type"=>"village", "pop"=>203, "name_ru"=>"Смитфилд", "name_en"=>"Smithfield", "lang"=>"en", "lat"=>40.474412, "lon"=>-90.295019),
"/US/US-IL/smithton-5626" => array("type"=>"village", "pop"=>3418, "name_ru"=>"Смиттон", "name_en"=>"Smithton", "lang"=>"en", "lat"=>38.408200, "lon"=>-89.990199),
"/US/US-IL/somonauk-5627" => array("type"=>"village", "pop"=>1651, "name_ru"=>"Сомонок", "name_en"=>"Somonauk", "lang"=>"en", "lat"=>41.632279, "lon"=>-88.683473),
"/US/US-IL/sorento-5628" => array("type"=>"village", "pop"=>602, "name_ru"=>"Соренто", "name_en"=>"Sorento", "lang"=>"en", "lat"=>38.999952, "lon"=>-89.572679),
"/US/US-IL/south-barrington-5629" => array("type"=>"village", "pop"=>4461, "name_ru"=>"Саут Бэррингтон", "name_en"=>"South Barrington", "lang"=>"en", "lat"=>42.087250, "lon"=>-88.153035),
"/US/US-IL/south-beloit-5630" => array("type"=>"city", "pop"=>5541, "name_ru"=>"Саут Белойт", "name_en"=>"South Beloit", "lang"=>"en", "lat"=>42.484718, "lon"=>-89.033360),
"/US/US-IL/south-chicago-heights-5631" => array("type"=>"village", "pop"=>3779, "name_ru"=>"Саут Чикаго Хайтс", "name_en"=>"South Chicago Heights", "lang"=>"en", "lat"=>41.483300, "lon"=>-87.637396),
"/US/US-IL/south-elgin-5632" => array("type"=>"village", "pop"=>21131, "name_ru"=>"Саут Элгин", "name_en"=>"South Elgin", "lang"=>"en", "lat"=>41.996300, "lon"=>-88.297999),
"/US/US-IL/south-holland-5634" => array("type"=>"village", "pop"=>21083, "name_ru"=>"Саут Холленд", "name_en"=>"South Holland", "lang"=>"en", "lat"=>41.597500, "lon"=>-87.601175),
"/US/US-IL/south-jacksonville-5635" => array("type"=>"village", "pop"=>3250, "name_ru"=>"Саут Джексонвилл", "name_en"=>"South Jacksonville", "lang"=>"en", "lat"=>39.707408, "lon"=>-90.230060),
"/US/US-IL/south-pekin-5636" => array("type"=>"village", "pop"=>1198, "name_ru"=>"Саут Пекин", "name_en"=>"South Pekin", "lang"=>"en", "lat"=>40.496886, "lon"=>-89.651776),
"/US/US-IL/south-roxana-5637" => array("type"=>"village", "pop"=>1807, "name_ru"=>"Саут Роксана", "name_en"=>"South Roxana", "lang"=>"en", "lat"=>38.822135, "lon"=>-90.059414),
"/US/US-IL/south-wilmington-5638" => array("type"=>"village", "pop"=>636, "name_ru"=>"Саут Уилмингтон", "name_en"=>"South Wilmington", "lang"=>"en", "lat"=>41.174220, "lon"=>-88.276258),
"/US/US-IL/southern-view-5633" => array("type"=>"village", "pop"=>1642, "name_ru"=>"Саутерн Вью", "name_en"=>"Southern View", "lang"=>"en", "lat"=>39.758100, "lon"=>-89.652074),
"/US/US-IL/sparland-5639" => array("type"=>"village", "pop"=>474, "name_ru"=>"Спарленд", "name_en"=>"Sparland", "lang"=>"en", "lat"=>41.030287, "lon"=>-89.440914),
"/US/US-IL/sparta-5640" => array("type"=>"city", "pop"=>4314, "name_ru"=>"Спарта", "name_en"=>"Sparta", "lang"=>"en", "lat"=>38.152784, "lon"=>-89.715544),
"/US/US-IL/spaulding-5641" => array("type"=>"village", "pop"=>821, "name_ru"=>"Сполдинг", "name_en"=>"Spaulding", "lang"=>"en", "lat"=>39.866704, "lon"=>-89.541963),
"/US/US-IL/spillertown-5642" => array("type"=>"village", "pop"=>223, "name_ru"=>"Спиллертаун", "name_en"=>"Spillertown", "lang"=>"en", "lat"=>37.765874, "lon"=>-88.921073),
"/US/US-IL/spring-bay-5643" => array("type"=>"village", "pop"=>522, "name_ru"=>"Спринг Бэй", "name_en"=>"Spring Bay", "lang"=>"en", "lat"=>40.823456, "lon"=>-89.525639),
"/US/US-IL/spring-grove-5646" => array("type"=>"village", "pop"=>5769, "name_ru"=>"Спринг Гров", "name_en"=>"Spring Grove", "lang"=>"en", "lat"=>42.444873, "lon"=>-88.243754),
"/US/US-IL/spring-valley-5647" => array("type"=>"city", "pop"=>5450, "name_ru"=>"Спринг Валли", "name_en"=>"Spring Valley", "lang"=>"en", "lat"=>41.326135, "lon"=>-89.198959),
"/US/US-IL/springerton-5644" => array("type"=>"village", "pop"=>128, "name_ru"=>"Спрингертон", "name_en"=>"Springerton", "lang"=>"en", "lat"=>38.178857, "lon"=>-88.354890),
"/US/US-IL/springfield-5645" => array("type"=>"city", "pop"=>117352, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>39.781433, "lon"=>-89.644654),
"/US/US-IL/st.-anne-5567" => array("type"=>"village", "pop"=>1247, "name_ru"=>"Сент Энн", "name_en"=>"St. Anne", "lang"=>"en", "lat"=>41.023331, "lon"=>-87.717014),
"/US/US-IL/st.-augustine-5568" => array("type"=>"village", "pop"=>137, "name_ru"=>"Сент Августин", "name_en"=>"St. Augustine", "lang"=>"en", "lat"=>40.719735, "lon"=>-90.408876),
"/US/US-IL/st.-charles-5569" => array("type"=>"city", "pop"=>32829, "name_ru"=>"Сент-Чарльз", "name_en"=>"St. Charles", "lang"=>"en", "lat"=>41.920333, "lon"=>-88.300830),
"/US/US-IL/st.-david-5570" => array("type"=>"village", "pop"=>547, "name_ru"=>"Сент-Дэвид", "name_en"=>"St. David", "lang"=>"en", "lat"=>40.492028, "lon"=>-90.051355),
"/US/US-IL/st.-elmo-5571" => array("type"=>"city", "pop"=>1420, "name_ru"=>"Сент Элмо", "name_en"=>"St. Elmo", "lang"=>"en", "lat"=>39.024697, "lon"=>-88.853242),
"/US/US-IL/st.-francisville-5573" => array("type"=>"city", "pop"=>728, "name_ru"=>"Сент Фрэнсисвилл", "name_en"=>"St. Francisville", "lang"=>"en", "lat"=>38.591924, "lon"=>-87.647610),
"/US/US-IL/st.-jacob-5574" => array("type"=>"village", "pop"=>1081, "name_ru"=>"Сент Якобс", "name_en"=>"St. Jacob", "lang"=>"en", "lat"=>38.716850, "lon"=>-89.768649),
"/US/US-IL/st.-johns-5575" => array("type"=>"village", "pop"=>213, "name_ru"=>"Сент-Джонс", "name_en"=>"St. Johns", "lang"=>"en", "lat"=>38.031241, "lon"=>-89.240378),
"/US/US-IL/st.-joseph-5576" => array("type"=>"village", "pop"=>4023, "name_ru"=>"Сент Джозеф", "name_en"=>"St. Joseph", "lang"=>"en", "lat"=>40.112672, "lon"=>-88.040643),
"/US/US-IL/st.-libory-5577" => array("type"=>"village", "pop"=>625, "name_ru"=>"Сент Либори", "name_en"=>"St. Libory", "lang"=>"en", "lat"=>38.364511, "lon"=>-89.714011),
"/US/US-IL/st.-peter-5578" => array("type"=>"village", "pop"=>381, "name_ru"=>"Сент Питер", "name_en"=>"St. Peter", "lang"=>"en", "lat"=>38.867394, "lon"=>-88.851802),
"/US/US-IL/standard-5648" => array("type"=>"village", "pop"=>250, "name_ru"=>"Стэндард", "name_en"=>"Standard", "lang"=>"en", "lat"=>41.256403, "lon"=>-89.180322),
"/US/US-IL/standard-city-5649" => array("type"=>"village", "pop"=>133, "name_ru"=>"Стэндард Сити", "name_en"=>"Standard City", "lang"=>"en", "lat"=>39.350372, "lon"=>-89.783831),
"/US/US-IL/stanford-5650" => array("type"=>"village", "pop"=>674, "name_ru"=>"Стэнфорд", "name_en"=>"Stanford", "lang"=>"en", "lat"=>40.433954, "lon"=>-89.220118),
"/US/US-IL/staunton-5651" => array("type"=>"city", "pop"=>5139, "name_ru"=>"Стонтон", "name_en"=>"Staunton", "lang"=>"en", "lat"=>39.012563, "lon"=>-89.788467),
"/US/US-IL/ste.-marie-5572" => array("type"=>"village", "pop"=>251, "name_ru"=>"Сент Мари", "name_en"=>"Ste. Marie", "lang"=>"en", "lat"=>38.932042, "lon"=>-88.026599),
"/US/US-IL/steeleville-5652" => array("type"=>"village", "pop"=>2014, "name_ru"=>"Стилвилл", "name_en"=>"Steeleville", "lang"=>"en", "lat"=>38.007528, "lon"=>-89.660743),
"/US/US-IL/steger-5653" => array("type"=>"village", "pop"=>10421, "name_ru"=>"Стиджер", "name_en"=>"Steger", "lang"=>"en", "lat"=>41.467800, "lon"=>-87.634999),
"/US/US-IL/sterling-5654" => array("type"=>"city", "pop"=>15124, "name_ru"=>"Стерлинг", "name_en"=>"Sterling", "lang"=>"en", "lat"=>41.798227, "lon"=>-89.691526),
"/US/US-IL/steward-5655" => array("type"=>"village", "pop"=>261, "name_ru"=>"Стьюард", "name_en"=>"Steward", "lang"=>"en", "lat"=>41.848456, "lon"=>-89.020578),
"/US/US-IL/stewardson-5656" => array("type"=>"village", "pop"=>712, "name_ru"=>"Стьюардсон", "name_en"=>"Stewardson", "lang"=>"en", "lat"=>39.264210, "lon"=>-88.629901),
"/US/US-IL/stickney-5657" => array("type"=>"village", "pop"=>5774, "name_ru"=>"Стикни", "name_en"=>"Stickney", "lang"=>"en", "lat"=>41.818000, "lon"=>-87.772817),
"/US/US-IL/stillman-valley-5658" => array("type"=>"village", "pop"=>1108, "name_ru"=>"Стилман Валли", "name_en"=>"Stillman Valley", "lang"=>"en", "lat"=>42.104010, "lon"=>-89.178073),
"/US/US-IL/stockton-5659" => array("type"=>"village", "pop"=>1797, "name_ru"=>"Стоктон", "name_en"=>"Stockton", "lang"=>"en", "lat"=>42.351601, "lon"=>-90.005995),
"/US/US-IL/stone-park-5661" => array("type"=>"village", "pop"=>4860, "name_ru"=>"Стоун Парк", "name_en"=>"Stone Park", "lang"=>"en", "lat"=>41.904300, "lon"=>-87.880425),
"/US/US-IL/stonefort-5660" => array("type"=>"village", "pop"=>291, "name_ru"=>"Стоун Форт", "name_en"=>"Stonefort", "lang"=>"en", "lat"=>37.617526, "lon"=>-88.707231),
"/US/US-IL/stonington-5662" => array("type"=>"village", "pop"=>900, "name_ru"=>"Стонингтон", "name_en"=>"Stonington", "lang"=>"en", "lat"=>39.639125, "lon"=>-89.192201),
"/US/US-IL/stoy-5663" => array("type"=>"village", "pop"=>115, "name_ru"=>"Стой", "name_en"=>"Stoy", "lang"=>"en", "lat"=>38.996837, "lon"=>-87.833427),
"/US/US-IL/strasburg-5664" => array("type"=>"village", "pop"=>563, "name_ru"=>"Страсбург", "name_en"=>"Strasburg", "lang"=>"en", "lat"=>39.350553, "lon"=>-88.623197),
"/US/US-IL/strawn-5665" => array("type"=>"village", "pop"=>97, "name_ru"=>"Строн", "name_en"=>"Strawn", "lang"=>"en", "lat"=>40.652689, "lon"=>-88.398950),
"/US/US-IL/streamwood-5666" => array("type"=>"village", "pop"=>37119, "name_ru"=>"Стримвуд", "name_en"=>"Streamwood", "lang"=>"en", "lat"=>42.020600, "lon"=>-88.173335),
"/US/US-IL/streator-5667" => array("type"=>"city", "pop"=>13780, "name_ru"=>"Стритор", "name_en"=>"Streator", "lang"=>"en", "lat"=>41.125271, "lon"=>-88.831477),
"/US/US-IL/stronghurst-5668" => array("type"=>"village", "pop"=>795, "name_ru"=>"Стронгхерст", "name_en"=>"Stronghurst", "lang"=>"en", "lat"=>40.746637, "lon"=>-90.909386),
"/US/US-IL/sublette-5669" => array("type"=>"village", "pop"=>442, "name_ru"=>"Саблетт", "name_en"=>"Sublette", "lang"=>"en", "lat"=>41.643459, "lon"=>-89.231673),
"/US/US-IL/sugar-grove-5670" => array("type"=>"village", "pop"=>9645, "name_ru"=>"Шугар Гров", "name_en"=>"Sugar Grove", "lang"=>"en", "lat"=>41.769600, "lon"=>-88.458848),
"/US/US-IL/sullivan-5671" => array("type"=>"city", "pop"=>4383, "name_ru"=>"Салливан", "name_en"=>"Sullivan", "lang"=>"en", "lat"=>39.597448, "lon"=>-88.607093),
"/US/US-IL/summerfield-5672" => array("type"=>"village", "pop"=>473, "name_ru"=>"Саммерфилд", "name_en"=>"Summerfield", "lang"=>"en", "lat"=>38.595692, "lon"=>-89.747263),
"/US/US-IL/summit-5673" => array("type"=>"village", "pop"=>10223, "name_ru"=>"Саммит", "name_en"=>"Summit", "lang"=>"en", "lat"=>41.784767, "lon"=>-87.816196),
"/US/US-IL/sumner-5674" => array("type"=>"city", "pop"=>2981, "name_ru"=>"Саммер", "name_en"=>"Sumner", "lang"=>"en", "lat"=>38.717078, "lon"=>-87.862331),
"/US/US-IL/sun-river-terrace-5676" => array("type"=>"village", "pop"=>526, "name_ru"=>"Сан Ривер Террейс", "name_en"=>"Sun River Terrace", "lang"=>"en", "lat"=>41.126476, "lon"=>-87.734600),
"/US/US-IL/sunnyside-5675" => array("type"=>"village", "pop"=>1529, "name_ru"=>"Саннисайд", "name_en"=>"Sunnyside", "lang"=>"en", "lat"=>42.395617, "lon"=>-88.233668),
"/US/US-IL/swansea-5677" => array("type"=>"village", "pop"=>12927, "name_ru"=>"Свонси", "name_en"=>"Swansea", "lang"=>"en", "lat"=>38.542100, "lon"=>-89.988577),
"/US/US-IL/sycamore-5678" => array("type"=>"city", "pop"=>17679, "name_ru"=>"Сикамор", "name_en"=>"Sycamore", "lang"=>"en", "lat"=>41.981704, "lon"=>-88.695770),
"/US/US-IL/symerton-5679" => array("type"=>"village", "pop"=>112, "name_ru"=>"Симертон", "name_en"=>"Symerton", "lang"=>"en", "lat"=>41.328102, "lon"=>-88.054497),
"/US/US-IL/table-grove-5680" => array("type"=>"village", "pop"=>375, "name_ru"=>"Тейбл Гров", "name_en"=>"Table Grove", "lang"=>"en", "lat"=>40.365575, "lon"=>-90.424925),
"/US/US-IL/tallula-5681" => array("type"=>"village", "pop"=>601, "name_ru"=>"Таллула", "name_en"=>"Tallula", "lang"=>"en", "lat"=>39.945806, "lon"=>-89.936340),
"/US/US-IL/tamaroa-5682" => array("type"=>"village", "pop"=>719, "name_ru"=>"Тамароа", "name_en"=>"Tamaroa", "lang"=>"en", "lat"=>38.135364, "lon"=>-89.228873),
"/US/US-IL/tamms-5683" => array("type"=>"village", "pop"=>916, "name_ru"=>"Таммс", "name_en"=>"Tamms", "lang"=>"en", "lat"=>37.239158, "lon"=>-89.267629),
"/US/US-IL/tampico-5684" => array("type"=>"village", "pop"=>728, "name_ru"=>"Тампико", "name_en"=>"Tampico", "lang"=>"en", "lat"=>41.630527, "lon"=>-89.785467),
"/US/US-IL/taylor-springs-5685" => array("type"=>"village", "pop"=>565, "name_ru"=>"Тейлор Спрингс", "name_en"=>"Taylor Springs", "lang"=>"en", "lat"=>39.130698, "lon"=>-89.495186),
"/US/US-IL/taylorville-5686" => array("type"=>"city", "pop"=>12238, "name_ru"=>"Тейлорвилл", "name_en"=>"Taylorville", "lang"=>"en", "lat"=>39.521466, "lon"=>-89.262165),
"/US/US-IL/tennessee-5687" => array("type"=>"village", "pop"=>130, "name_ru"=>"Теннесси", "name_en"=>"Tennessee", "lang"=>"en", "lat"=>40.411701, "lon"=>-90.835876),
"/US/US-IL/teutopolis-5688" => array("type"=>"village", "pop"=>1660, "name_ru"=>"Теутополис", "name_en"=>"Teutopolis", "lang"=>"en", "lat"=>39.132084, "lon"=>-88.478388),
"/US/US-IL/thawville-5689" => array("type"=>"village", "pop"=>296, "name_ru"=>"Товилл", "name_en"=>"Thawville", "lang"=>"en", "lat"=>40.673868, "lon"=>-88.113253),
"/US/US-IL/thayer-5690" => array("type"=>"village", "pop"=>688, "name_ru"=>"Тайер", "name_en"=>"Thayer", "lang"=>"en", "lat"=>39.539912, "lon"=>-89.758409),
"/US/US-IL/thebes-5691" => array("type"=>"village", "pop"=>406, "name_ru"=>"Тебес", "name_en"=>"Thebes", "lang"=>"en", "lat"=>37.210830, "lon"=>-89.453240),
"/US/US-IL/third-lake-5692" => array("type"=>"village", "pop"=>1382, "name_ru"=>"Терд Лейк", "name_en"=>"Third Lake", "lang"=>"en", "lat"=>42.367699, "lon"=>-88.008358),
"/US/US-IL/thomasboro-5693" => array("type"=>"village", "pop"=>1209, "name_ru"=>"Томасборо", "name_en"=>"Thomasboro", "lang"=>"en", "lat"=>40.243984, "lon"=>-88.186887),
"/US/US-IL/thompsonville-5694" => array("type"=>"village", "pop"=>588, "name_ru"=>"Томпсонвилл", "name_en"=>"Thompsonville", "lang"=>"en", "lat"=>37.914127, "lon"=>-88.761692),
"/US/US-IL/thomson-5695" => array("type"=>"village", "pop"=>524, "name_ru"=>"Томсон", "name_en"=>"Thomson", "lang"=>"en", "lat"=>41.957470, "lon"=>-90.103161),
"/US/US-IL/thornton-5696" => array("type"=>"village", "pop"=>2386, "name_ru"=>"Торнтон", "name_en"=>"Thornton", "lang"=>"en", "lat"=>41.574000, "lon"=>-87.618762),
"/US/US-IL/tilden-5697" => array("type"=>"village", "pop"=>870, "name_ru"=>"Тилден", "name_en"=>"Tilden", "lang"=>"en", "lat"=>38.211904, "lon"=>-89.683262),
"/US/US-IL/tilton-5698" => array("type"=>"village", "pop"=>2779, "name_ru"=>"Тилтон", "name_en"=>"Tilton", "lang"=>"en", "lat"=>40.094250, "lon"=>-87.638546),
"/US/US-IL/time-5699" => array("type"=>"village", "pop"=>27, "name_ru"=>"Тайм", "name_en"=>"Time", "lang"=>"en", "lat"=>39.561113, "lon"=>-90.722820),
"/US/US-IL/tinley-park-5700" => array("type"=>"village", "pop"=>59198, "name_ru"=>"Тинли Парк", "name_en"=>"Tinley Park", "lang"=>"en", "lat"=>41.574767, "lon"=>-87.800432),
"/US/US-IL/tiskilwa-5701" => array("type"=>"village", "pop"=>750, "name_ru"=>"Тискилва", "name_en"=>"Tiskilwa", "lang"=>"en", "lat"=>41.292736, "lon"=>-89.507841),
"/US/US-IL/toledo-5702" => array("type"=>"village", "pop"=>1129, "name_ru"=>"Толедо", "name_en"=>"Toledo", "lang"=>"en", "lat"=>39.274308, "lon"=>-88.242747),
"/US/US-IL/tolono-5703" => array("type"=>"village", "pop"=>2836, "name_ru"=>"Толоно", "name_en"=>"Tolono", "lang"=>"en", "lat"=>39.986319, "lon"=>-88.262956),
"/US/US-IL/toluca-5704" => array("type"=>"city", "pop"=>1250, "name_ru"=>"Толука", "name_en"=>"Toluca", "lang"=>"en", "lat"=>41.005337, "lon"=>-89.133169),
"/US/US-IL/tonica-5705" => array("type"=>"village", "pop"=>698, "name_ru"=>"Тоника", "name_en"=>"Tonica", "lang"=>"en", "lat"=>41.214060, "lon"=>-89.067794),
"/US/US-IL/topeka-5706" => array("type"=>"town", "pop"=>84, "name_ru"=>"Топека", "name_en"=>"Topeka", "lang"=>"en", "lat"=>40.329919, "lon"=>-89.930920),
"/US/US-IL/toulon-5707" => array("type"=>"city", "pop"=>1339, "name_ru"=>"Тоулон", "name_en"=>"Toulon", "lang"=>"en", "lat"=>41.094469, "lon"=>-89.863736),
"/US/US-IL/tovey-5708" => array("type"=>"village", "pop"=>486, "name_ru"=>"Тови", "name_en"=>"Tovey", "lang"=>"en", "lat"=>39.587575, "lon"=>-89.449369),
"/US/US-IL/towanda-5709" => array("type"=>"village", "pop"=>487, "name_ru"=>"Тованда", "name_en"=>"Towanda", "lang"=>"en", "lat"=>40.563211, "lon"=>-88.900438),
"/US/US-IL/tower-hill-5710" => array("type"=>"village", "pop"=>576, "name_ru"=>"Тауэр Хилл", "name_en"=>"Tower Hill", "lang"=>"en", "lat"=>39.386810, "lon"=>-88.959624),
"/US/US-IL/tower-lakes-5711" => array("type"=>"village", "pop"=>1302, "name_ru"=>"Тауэр Лейкс", "name_en"=>"Tower Lakes", "lang"=>"en", "lat"=>42.230217, "lon"=>-88.155793),
"/US/US-IL/tremont-5712" => array("type"=>"village", "pop"=>2070, "name_ru"=>"Тремонт", "name_en"=>"Tremont", "lang"=>"en", "lat"=>40.525840, "lon"=>-89.490343),
"/US/US-IL/trenton-5713" => array("type"=>"city", "pop"=>2661, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>38.606678, "lon"=>-89.681627),
"/US/US-IL/troy-5714" => array("type"=>"city", "pop"=>9840, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>38.731200, "lon"=>-89.893754),
"/US/US-IL/troy-grove-5715" => array("type"=>"village", "pop"=>301, "name_ru"=>"Трой Гров", "name_en"=>"Troy Grove", "lang"=>"en", "lat"=>41.465429, "lon"=>-89.081122),
"/US/US-IL/tuscola-5716" => array("type"=>"city", "pop"=>4514, "name_ru"=>"Таскола", "name_en"=>"Tuscola", "lang"=>"en", "lat"=>39.797651, "lon"=>-88.281495),
"/US/US-IL/ullin-5717" => array("type"=>"village", "pop"=>696, "name_ru"=>"Аллин", "name_en"=>"Ullin", "lang"=>"en", "lat"=>37.281641, "lon"=>-89.183788),
"/US/US-IL/union-5718" => array("type"=>"village", "pop"=>615, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>42.234279, "lon"=>-88.543511),
"/US/US-IL/union-hill-5719" => array("type"=>"village", "pop"=>63, "name_ru"=>"Юнион Хилл", "name_en"=>"Union Hill", "lang"=>"en", "lat"=>41.109296, "lon"=>-88.145458),
"/US/US-IL/university-park-5720" => array("type"=>"village", "pop"=>8171, "name_ru"=>"Юниверсити Парк", "name_en"=>"University Park", "lang"=>"en", "lat"=>41.442400, "lon"=>-87.709634),
"/US/US-IL/urbana-5721" => array("type"=>"city", "pop"=>39641, "name_ru"=>"Урбана", "name_en"=>"Urbana", "lang"=>"en", "lat"=>40.108536, "lon"=>-88.205151),
"/US/US-IL/ursa-5722" => array("type"=>"village", "pop"=>585, "name_ru"=>"Ерза", "name_en"=>"Ursa", "lang"=>"en", "lat"=>40.074607, "lon"=>-91.370627),
"/US/US-IL/valier-5723" => array("type"=>"village", "pop"=>664, "name_ru"=>"Вальер", "name_en"=>"Valier", "lang"=>"en", "lat"=>38.017766, "lon"=>-89.043475),
"/US/US-IL/valley-city-5724" => array("type"=>"village", "pop"=>13, "name_ru"=>"Валли Сити", "name_en"=>"Valley City", "lang"=>"en", "lat"=>39.708887, "lon"=>-90.649104),
"/US/US-IL/valmeyer-5725" => array("type"=>"village", "pop"=>1164, "name_ru"=>"Валмайер", "name_en"=>"Valmeyer", "lang"=>"en", "lat"=>38.304133, "lon"=>-90.313036),
"/US/US-IL/vandalia-5726" => array("type"=>"city", "pop"=>6321, "name_ru"=>"Вандалия", "name_en"=>"Vandalia", "lang"=>"en", "lat"=>38.980968, "lon"=>-89.101358),
"/US/US-IL/varna-5727" => array("type"=>"village", "pop"=>410, "name_ru"=>"Варна", "name_en"=>"Varna", "lang"=>"en", "lat"=>41.035637, "lon"=>-89.224513),
"/US/US-IL/venedy-5728" => array("type"=>"village", "pop"=>131, "name_ru"=>"Венеди", "name_en"=>"Venedy", "lang"=>"en", "lat"=>38.395887, "lon"=>-89.645866),
"/US/US-IL/venetian-village-5729" => array("type"=>"cdp", "pop"=>3133, "name_ru"=>"Венешиан Виллидж", "name_en"=>"Venetian Village", "lang"=>"en", "lat"=>42.399799, "lon"=>-88.046170),
"/US/US-IL/venice-5730" => array("type"=>"city", "pop"=>2421, "name_ru"=>"Венис", "name_en"=>"Venice", "lang"=>"en", "lat"=>38.672050, "lon"=>-90.168745),
"/US/US-IL/vergennes-5731" => array("type"=>"village", "pop"=>321, "name_ru"=>"Вердженес", "name_en"=>"Vergennes", "lang"=>"en", "lat"=>37.902024, "lon"=>-89.339735),
"/US/US-IL/vermilion-5732" => array("type"=>"village", "pop"=>219, "name_ru"=>"Вермилион", "name_en"=>"Vermilion", "lang"=>"en", "lat"=>39.579725, "lon"=>-87.587435),
"/US/US-IL/vermont-5733" => array("type"=>"village", "pop"=>751, "name_ru"=>"Вермонт", "name_en"=>"Vermont", "lang"=>"en", "lat"=>40.296004, "lon"=>-90.428122),
"/US/US-IL/vernon-5734" => array("type"=>"village", "pop"=>170, "name_ru"=>"Вернон", "name_en"=>"Vernon", "lang"=>"en", "lat"=>38.801965, "lon"=>-89.089136),
"/US/US-IL/vernon-hills-5735" => array("type"=>"village", "pop"=>24462, "name_ru"=>"Вермон Хиллс", "name_en"=>"Vernon Hills", "lang"=>"en", "lat"=>42.235950, "lon"=>-87.962719),
"/US/US-IL/verona-5736" => array("type"=>"village", "pop"=>258, "name_ru"=>"Верона", "name_en"=>"Verona", "lang"=>"en", "lat"=>41.215706, "lon"=>-88.505034),
"/US/US-IL/versailles-5737" => array("type"=>"village", "pop"=>527, "name_ru"=>"Версайллес", "name_en"=>"Versailles", "lang"=>"en", "lat"=>39.883849, "lon"=>-90.658868),
"/US/US-IL/victoria-5738" => array("type"=>"village", "pop"=>293, "name_ru"=>"Виктория", "name_en"=>"Victoria", "lang"=>"en", "lat"=>41.032984, "lon"=>-90.095339),
"/US/US-IL/vienna-5739" => array("type"=>"city", "pop"=>1282, "name_ru"=>"Вьенна", "name_en"=>"Vienna", "lang"=>"en", "lat"=>37.414176, "lon"=>-88.892714),
"/US/US-IL/villa-grove-5740" => array("type"=>"city", "pop"=>2449, "name_ru"=>"Вилла Гров", "name_en"=>"Villa Grove", "lang"=>"en", "lat"=>39.863635, "lon"=>-88.159536),
"/US/US-IL/villa-park-5741" => array("type"=>"village", "pop"=>22198, "name_ru"=>"Вилла Парк", "name_en"=>"Villa Park", "lang"=>"en", "lat"=>41.885500, "lon"=>-87.978216),
"/US/US-IL/viola-5742" => array("type"=>"village", "pop"=>928, "name_ru"=>"Виола", "name_en"=>"Viola", "lang"=>"en", "lat"=>41.205239, "lon"=>-90.587407),
"/US/US-IL/virden-5743" => array("type"=>"city", "pop"=>3364, "name_ru"=>"Вирден", "name_en"=>"Virden", "lang"=>"en", "lat"=>39.504886, "lon"=>-89.769493),
"/US/US-IL/virginia-5744" => array("type"=>"city", "pop"=>1664, "name_ru"=>"Вирджиния", "name_en"=>"Virginia", "lang"=>"en", "lat"=>39.950784, "lon"=>-90.211271),
"/US/US-IL/wadsworth-5745" => array("type"=>"village", "pop"=>3829, "name_ru"=>"Уодсворт", "name_en"=>"Wadsworth", "lang"=>"en", "lat"=>42.436912, "lon"=>-87.923369),
"/US/US-IL/waggoner-5746" => array("type"=>"village", "pop"=>238, "name_ru"=>"Ваггонер", "name_en"=>"Waggoner", "lang"=>"en", "lat"=>39.378224, "lon"=>-89.652332),
"/US/US-IL/walnut-5747" => array("type"=>"village", "pop"=>1400, "name_ru"=>"Уолнат", "name_en"=>"Walnut", "lang"=>"en", "lat"=>41.556821, "lon"=>-89.591052),
"/US/US-IL/walnut-hill-5748" => array("type"=>"village", "pop"=>104, "name_ru"=>"Уолнат Хилл", "name_en"=>"Walnut Hill", "lang"=>"en", "lat"=>38.477365, "lon"=>-89.044206),
"/US/US-IL/walshville-5749" => array("type"=>"village", "pop"=>86, "name_ru"=>"Уолшвилл", "name_en"=>"Walshville", "lang"=>"en", "lat"=>39.069415, "lon"=>-89.619500),
"/US/US-IL/waltonville-5750" => array("type"=>"village", "pop"=>428, "name_ru"=>"Уолтонвилл", "name_en"=>"Waltonville", "lang"=>"en", "lat"=>38.213235, "lon"=>-89.039129),
"/US/US-IL/wamac-5751" => array("type"=>"city", "pop"=>1270, "name_ru"=>"Вамак", "name_en"=>"Wamac", "lang"=>"en", "lat"=>38.500152, "lon"=>-89.145733),
"/US/US-IL/wapella-5752" => array("type"=>"village", "pop"=>638, "name_ru"=>"Вапелла", "name_en"=>"Wapella", "lang"=>"en", "lat"=>40.221326, "lon"=>-88.961715),
"/US/US-IL/warren-5753" => array("type"=>"village", "pop"=>1370, "name_ru"=>"Уоррен", "name_en"=>"Warren", "lang"=>"en", "lat"=>42.494729, "lon"=>-89.990179),
"/US/US-IL/warrensburg-5754" => array("type"=>"village", "pop"=>1172, "name_ru"=>"Уорренсбург", "name_en"=>"Warrensburg", "lang"=>"en", "lat"=>39.932336, "lon"=>-89.061223),
"/US/US-IL/warrenville-5755" => array("type"=>"city", "pop"=>13033, "name_ru"=>"Уорренвилл", "name_en"=>"Warrenville", "lang"=>"en", "lat"=>41.820050, "lon"=>-88.186203),
"/US/US-IL/warsaw-5756" => array("type"=>"city", "pop"=>1603, "name_ru"=>"Варшава", "name_en"=>"Warsaw", "lang"=>"en", "lat"=>40.352829, "lon"=>-91.426862),
"/US/US-IL/washburn-5757" => array("type"=>"village", "pop"=>1100, "name_ru"=>"Уошберн", "name_en"=>"Washburn", "lang"=>"en", "lat"=>40.920740, "lon"=>-89.291671),
"/US/US-IL/washington-5758" => array("type"=>"city", "pop"=>13942, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>40.698206, "lon"=>-89.435107),
"/US/US-IL/washington-park-5759" => array("type"=>"village", "pop"=>5558, "name_ru"=>"Вашингтон Парк", "name_en"=>"Washington Park", "lang"=>"en", "lat"=>38.629900, "lon"=>-90.092199),
"/US/US-IL/wataga-5760" => array("type"=>"village", "pop"=>787, "name_ru"=>"Уотага", "name_en"=>"Wataga", "lang"=>"en", "lat"=>41.025054, "lon"=>-90.275054),
"/US/US-IL/waterloo-5761" => array("type"=>"city", "pop"=>9813, "name_ru"=>"Ватрелоо", "name_en"=>"Waterloo", "lang"=>"en", "lat"=>38.331589, "lon"=>-90.155581),
"/US/US-IL/waterman-5762" => array("type"=>"village", "pop"=>1527, "name_ru"=>"Уотерман", "name_en"=>"Waterman", "lang"=>"en", "lat"=>41.769432, "lon"=>-88.773163),
"/US/US-IL/watseka-5763" => array("type"=>"city", "pop"=>5481, "name_ru"=>"Уотсика", "name_en"=>"Watseka", "lang"=>"en", "lat"=>40.775939, "lon"=>-87.732129),
"/US/US-IL/watson-5764" => array("type"=>"village", "pop"=>743, "name_ru"=>"Уотсон", "name_en"=>"Watson", "lang"=>"en", "lat"=>39.025570, "lon"=>-88.569539),
"/US/US-IL/wauconda-5765" => array("type"=>"village", "pop"=>12219, "name_ru"=>"Уоконда", "name_en"=>"Wauconda", "lang"=>"en", "lat"=>42.266621, "lon"=>-88.139510),
"/US/US-IL/waukegan-5766" => array("type"=>"city", "pop"=>90788, "name_ru"=>"Уокиган", "name_en"=>"Waukegan", "lang"=>"en", "lat"=>42.370900, "lon"=>-87.867999),
"/US/US-IL/waverly-5767" => array("type"=>"city", "pop"=>1235, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>39.592657, "lon"=>-89.952331),
"/US/US-IL/wayne-5768" => array("type"=>"village", "pop"=>2394, "name_ru"=>"Уэйн", "name_en"=>"Wayne", "lang"=>"en", "lat"=>41.952100, "lon"=>-88.261288),
"/US/US-IL/wayne-city-5769" => array("type"=>"village", "pop"=>1059, "name_ru"=>"Уэйн Сити", "name_en"=>"Wayne City", "lang"=>"en", "lat"=>38.348732, "lon"=>-88.587426),
"/US/US-IL/waynesville-5770" => array("type"=>"village", "pop"=>429, "name_ru"=>"Уэйнсвилл", "name_en"=>"Waynesville", "lang"=>"en", "lat"=>40.240896, "lon"=>-89.124967),
"/US/US-IL/weldon-5771" => array("type"=>"village", "pop"=>417, "name_ru"=>"Уэлдон", "name_en"=>"Weldon", "lang"=>"en", "lat"=>40.121517, "lon"=>-88.750550),
"/US/US-IL/wellington-5772" => array("type"=>"village", "pop"=>263, "name_ru"=>"Веллингтон", "name_en"=>"Wellington", "lang"=>"en", "lat"=>40.540470, "lon"=>-87.678293),
"/US/US-IL/wenona-5773" => array("type"=>"city", "pop"=>1043, "name_ru"=>"Уэнона", "name_en"=>"Wenona", "lang"=>"en", "lat"=>41.053047, "lon"=>-89.053335),
"/US/US-IL/wenonah-5774" => array("type"=>"village", "pop"=>43, "name_ru"=>"Уэнона", "name_en"=>"Wenonah", "lang"=>"en", "lat"=>39.320145, "lon"=>-89.289291),
"/US/US-IL/west-brooklyn-5775" => array("type"=>"village", "pop"=>167, "name_ru"=>"Уэст Бруклин", "name_en"=>"West Brooklyn", "lang"=>"en", "lat"=>41.693065, "lon"=>-89.147019),
"/US/US-IL/west-chicago-5777" => array("type"=>"city", "pop"=>26454, "name_ru"=>"Уэст Чикаго", "name_en"=>"West Chicago", "lang"=>"en", "lat"=>41.889894, "lon"=>-88.222825),
"/US/US-IL/west-city-5778" => array("type"=>"village", "pop"=>762, "name_ru"=>"Уэст Сити", "name_en"=>"West City", "lang"=>"en", "lat"=>37.995952, "lon"=>-88.948740),
"/US/US-IL/west-dundee-5779" => array("type"=>"village", "pop"=>7975, "name_ru"=>"Уэст Данди", "name_en"=>"West Dundee", "lang"=>"en", "lat"=>42.098067, "lon"=>-88.292022),
"/US/US-IL/west-frankfort-5782" => array("type"=>"city", "pop"=>8190, "name_ru"=>"Уэст Франкфорт", "name_en"=>"West Frankfort", "lang"=>"en", "lat"=>37.900648, "lon"=>-88.929418),
"/US/US-IL/west-peoria-5784" => array("type"=>"cdp", "pop"=>4456, "name_ru"=>"Уэст Пеория", "name_en"=>"West Peoria", "lang"=>"en", "lat"=>40.699150, "lon"=>-89.633549),
"/US/US-IL/west-point-5785" => array("type"=>"village", "pop"=>180, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>40.255620, "lon"=>-91.183099),
"/US/US-IL/west-salem-5786" => array("type"=>"village", "pop"=>937, "name_ru"=>"Уэст Салем", "name_en"=>"West Salem", "lang"=>"en", "lat"=>38.520012, "lon"=>-88.008813),
"/US/US-IL/westchester-5776" => array("type"=>"village", "pop"=>15710, "name_ru"=>"Уэстчестер", "name_en"=>"Westchester", "lang"=>"en", "lat"=>41.850900, "lon"=>-87.887125),
"/US/US-IL/western-springs-5780" => array("type"=>"village", "pop"=>12690, "name_ru"=>"Уэстерн Спрингс", "name_en"=>"Western Springs", "lang"=>"en", "lat"=>41.803400, "lon"=>-87.900750),
"/US/US-IL/westfield-5781" => array("type"=>"village", "pop"=>658, "name_ru"=>"Уэстфилд", "name_en"=>"Westfield", "lang"=>"en", "lat"=>39.455546, "lon"=>-87.996894),
"/US/US-IL/westmont-5783" => array("type"=>"village", "pop"=>24979, "name_ru"=>"Уэстмонт", "name_en"=>"Westmont", "lang"=>"en", "lat"=>41.796950, "lon"=>-87.975049),
"/US/US-IL/westville-5787" => array("type"=>"village", "pop"=>3001, "name_ru"=>"Уэствилл", "name_en"=>"Westville", "lang"=>"en", "lat"=>40.044850, "lon"=>-87.637999),
"/US/US-IL/wheaton-5788" => array("type"=>"city", "pop"=>54465, "name_ru"=>"Уитон", "name_en"=>"Wheaton", "lang"=>"en", "lat"=>41.856000, "lon"=>-88.107613),
"/US/US-IL/wheeler-5789" => array("type"=>"village", "pop"=>114, "name_ru"=>"Уилер", "name_en"=>"Wheeler", "lang"=>"en", "lat"=>39.042813, "lon"=>-88.318631),
"/US/US-IL/wheeling-5790" => array("type"=>"village", "pop"=>36063, "name_ru"=>"Уилинг", "name_en"=>"Wheeling", "lang"=>"en", "lat"=>42.130500, "lon"=>-87.923635),
"/US/US-IL/white-city-5792" => array("type"=>"village", "pop"=>222, "name_ru"=>"Уайт Сити", "name_en"=>"White City", "lang"=>"en", "lat"=>39.072382, "lon"=>-89.764035),
"/US/US-IL/white-hall-5793" => array("type"=>"city", "pop"=>2455, "name_ru"=>"Уайт Холл", "name_en"=>"White Hall", "lang"=>"en", "lat"=>39.437442, "lon"=>-90.403953),
"/US/US-IL/whiteash-5791" => array("type"=>"village", "pop"=>273, "name_ru"=>"Уайтэш", "name_en"=>"Whiteash", "lang"=>"en", "lat"=>37.785241, "lon"=>-88.930517),
"/US/US-IL/williamsfield-5794" => array("type"=>"village", "pop"=>573, "name_ru"=>"Уильямсфилд", "name_en"=>"Williamsfield", "lang"=>"en", "lat"=>40.926132, "lon"=>-90.017338),
"/US/US-IL/williamson-5795" => array("type"=>"village", "pop"=>244, "name_ru"=>"Уильямсон", "name_en"=>"Williamson", "lang"=>"en", "lat"=>38.987372, "lon"=>-89.763946),
"/US/US-IL/williamsville-5796" => array("type"=>"village", "pop"=>1388, "name_ru"=>"Уильямсвилл", "name_en"=>"Williamsville", "lang"=>"en", "lat"=>39.952788, "lon"=>-89.556280),
"/US/US-IL/willisville-5797" => array("type"=>"village", "pop"=>686, "name_ru"=>"Уиллисвилл", "name_en"=>"Willisville", "lang"=>"en", "lat"=>37.982882, "lon"=>-89.590565),
"/US/US-IL/willow-hill-5800" => array("type"=>"village", "pop"=>232, "name_ru"=>"Умллоу Хилл", "name_en"=>"Willow Hill", "lang"=>"en", "lat"=>38.996341, "lon"=>-88.021295),
"/US/US-IL/willow-springs-5801" => array("type"=>"village", "pop"=>5898, "name_ru"=>"Уиллоу Спрингс", "name_en"=>"Willow Springs", "lang"=>"en", "lat"=>41.733500, "lon"=>-87.882449),
"/US/US-IL/willowbrook-5798" => array("type"=>"village", "pop"=>8739, "name_ru"=>"Уиллоубрук", "name_en"=>"Willowbrook", "lang"=>"en", "lat"=>41.761880, "lon"=>-87.946293),
"/US/US-IL/willowbrook-5799" => array("type"=>"cdp", "pop"=>1808, "name_ru"=>"Уиллоубрук", "name_en"=>"Willowbrook", "lang"=>"en", "lat"=>41.457640, "lon"=>-87.536320),
"/US/US-IL/wilmette-5802" => array("type"=>"village", "pop"=>26418, "name_ru"=>"Уилметт", "name_en"=>"Wilmette", "lang"=>"en", "lat"=>42.077200, "lon"=>-87.728574),
"/US/US-IL/wilmington-5803" => array("type"=>"city", "pop"=>112, "name_ru"=>"Уилмингтон", "name_en"=>"Wilmington", "lang"=>"en", "lat"=>41.297665, "lon"=>-88.159634),
"/US/US-IL/wilmington-5804" => array("type"=>"village", "pop"=>129, "name_ru"=>"Уилмингтон", "name_en"=>"Wilmington", "lang"=>"en", "lat"=>39.482467, "lon"=>-90.489559),
"/US/US-IL/wilsonville-5805" => array("type"=>"village", "pop"=>580, "name_ru"=>"Уилсонвилл", "name_en"=>"Wilsonville", "lang"=>"en", "lat"=>39.068866, "lon"=>-89.855304),
"/US/US-IL/winchester-5806" => array("type"=>"city", "pop"=>1548, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>39.629418, "lon"=>-90.456565),
"/US/US-IL/windsor-5807" => array("type"=>"city", "pop"=>681, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>39.438141, "lon"=>-88.595754),
"/US/US-IL/windsor-5808" => array("type"=>"village", "pop"=>774, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>41.201607, "lon"=>-90.443867),
"/US/US-IL/winfield-5809" => array("type"=>"village", "pop"=>9984, "name_ru"=>"Винфилд", "name_en"=>"Winfield", "lang"=>"en", "lat"=>41.871381, "lon"=>-88.153634),
"/US/US-IL/winnebago-5810" => array("type"=>"village", "pop"=>3216, "name_ru"=>"Уиннебаго", "name_en"=>"Winnebago", "lang"=>"en", "lat"=>42.267952, "lon"=>-89.240455),
"/US/US-IL/winnetka-5811" => array("type"=>"village", "pop"=>12371, "name_ru"=>"Уиннетка", "name_en"=>"Winnetka", "lang"=>"en", "lat"=>42.106450, "lon"=>-87.742726),
"/US/US-IL/winslow-5812" => array("type"=>"village", "pop"=>326, "name_ru"=>"Уинслоу", "name_en"=>"Winslow", "lang"=>"en", "lat"=>42.492393, "lon"=>-89.796299),
"/US/US-IL/winthrop-harbor-5813" => array("type"=>"village", "pop"=>7210, "name_ru"=>"Уинтроп Харбор", "name_en"=>"Winthrop Harbor", "lang"=>"en", "lat"=>42.479750, "lon"=>-87.828174),
"/US/US-IL/witt-5814" => array("type"=>"city", "pop"=>959, "name_ru"=>"Витт", "name_en"=>"Witt", "lang"=>"en", "lat"=>39.255799, "lon"=>-89.348961),
"/US/US-IL/wonder-lake-5815" => array("type"=>"cdp", "pop"=>3878, "name_ru"=>"Уандер Лейк", "name_en"=>"Wonder Lake", "lang"=>"en", "lat"=>42.380044, "lon"=>-88.356723),
"/US/US-IL/wonder-lake-5816" => array("type"=>"village", "pop"=>1024, "name_ru"=>"Уандер Лейк", "name_en"=>"Wonder Lake", "lang"=>"en", "lat"=>42.384250, "lon"=>-88.367724),
"/US/US-IL/wood-dale-5817" => array("type"=>"city", "pop"=>13855, "name_ru"=>"Вуд Дейл", "name_en"=>"Wood Dale", "lang"=>"en", "lat"=>41.965850, "lon"=>-87.981071),
"/US/US-IL/wood-river-5822" => array("type"=>"city", "pop"=>10973, "name_ru"=>"Вуд Ривер", "name_en"=>"Wood River", "lang"=>"en", "lat"=>38.862150, "lon"=>-90.084774),
"/US/US-IL/woodhull-5818" => array("type"=>"village", "pop"=>790, "name_ru"=>"Вудхалл", "name_en"=>"Woodhull", "lang"=>"en", "lat"=>41.178426, "lon"=>-90.321742),
"/US/US-IL/woodland-5819" => array("type"=>"village", "pop"=>305, "name_ru"=>"Вудленд", "name_en"=>"Woodland", "lang"=>"en", "lat"=>40.712953, "lon"=>-87.730652),
"/US/US-IL/woodlawn-5820" => array("type"=>"village", "pop"=>722, "name_ru"=>"Вудлон", "name_en"=>"Woodlawn", "lang"=>"en", "lat"=>38.330523, "lon"=>-89.035274),
"/US/US-IL/woodridge-5821" => array("type"=>"village", "pop"=>34130, "name_ru"=>"Вудридж", "name_en"=>"Woodridge", "lang"=>"en", "lat"=>41.739078, "lon"=>-88.042528),
"/US/US-IL/woodson-5823" => array("type"=>"village", "pop"=>540, "name_ru"=>"Вудсон", "name_en"=>"Woodson", "lang"=>"en", "lat"=>39.627527, "lon"=>-90.222860),
"/US/US-IL/woodstock-5824" => array("type"=>"city", "pop"=>23930, "name_ru"=>"Вудсток", "name_en"=>"Woodstock", "lang"=>"en", "lat"=>42.310650, "lon"=>-88.436689),
"/US/US-IL/worden-5825" => array("type"=>"village", "pop"=>1037, "name_ru"=>"Уорден", "name_en"=>"Worden", "lang"=>"en", "lat"=>38.932195, "lon"=>-89.838539),
"/US/US-IL/worth-5826" => array("type"=>"village", "pop"=>10426, "name_ru"=>"Уорт", "name_en"=>"Worth", "lang"=>"en", "lat"=>41.687300, "lon"=>-87.792553),
"/US/US-IL/wyanet-5827" => array("type"=>"village", "pop"=>987, "name_ru"=>"Уайанет", "name_en"=>"Wyanet", "lang"=>"en", "lat"=>41.360123, "lon"=>-89.583777),
"/US/US-IL/wyoming-5828" => array("type"=>"city", "pop"=>1394, "name_ru"=>"Вайоминг", "name_en"=>"Wyoming", "lang"=>"en", "lat"=>41.064466, "lon"=>-89.772314),
"/US/US-IL/xenia-5829" => array("type"=>"village", "pop"=>394, "name_ru"=>"Ксениа", "name_en"=>"Xenia", "lang"=>"en", "lat"=>38.636771, "lon"=>-88.637068),
"/US/US-IL/yale-5830" => array("type"=>"village", "pop"=>92, "name_ru"=>"Йель", "name_en"=>"Yale", "lang"=>"en", "lat"=>39.120094, "lon"=>-88.024303),
"/US/US-IL/yates-city-5831" => array("type"=>"village", "pop"=>667, "name_ru"=>"Йетс Сити", "name_en"=>"Yates City", "lang"=>"en", "lat"=>40.778026, "lon"=>-90.013581),
"/US/US-IL/york-center-5832" => array("type"=>"cdp", "pop"=>4818, "name_ru"=>"Йорк Сентер", "name_en"=>"York Center", "lang"=>"en", "lat"=>41.854950, "lon"=>-87.994741),
"/US/US-IL/yorkville-5833" => array("type"=>"city", "pop"=>16505, "name_ru"=>"Йорквилл", "name_en"=>"Yorkville", "lang"=>"en", "lat"=>41.642250, "lon"=>-88.447549),
"/US/US-IL/zeigler-5834" => array("type"=>"city", "pop"=>1665, "name_ru"=>"Зейглер", "name_en"=>"Zeigler", "lang"=>"en", "lat"=>37.899949, "lon"=>-89.052509),
"/US/US-IL/zion-5835" => array("type"=>"city", "pop"=>25161, "name_ru"=>"Зион", "name_en"=>"Zion", "lang"=>"en", "lat"=>42.456650, "lon"=>-87.843067),
"/US/US-IN" => array("type"=>"state", "pop"=>6376792, "name_ru"=>"Индиана", "name_en"=>"Indiana", "lang"=>"en"),
"/US/US-IN/advance-5836" => array("type"=>"town", "pop"=>582, "name_ru"=>"Адванс", "name_en"=>"Advance", "lang"=>"en", "lat"=>39.995230, "lon"=>-86.620429),
"/US/US-IN/akron-5837" => array("type"=>"town", "pop"=>1011, "name_ru"=>"Акрон", "name_en"=>"Akron", "lang"=>"en", "lat"=>41.038807, "lon"=>-86.024790),
"/US/US-IN/alamo-5838" => array("type"=>"town", "pop"=>135, "name_ru"=>"Аламо", "name_en"=>"Alamo", "lang"=>"en", "lat"=>39.983666, "lon"=>-87.055069),
"/US/US-IN/albany-5839" => array("type"=>"town", "pop"=>2311, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>40.302100, "lon"=>-85.233625),
"/US/US-IN/albion-5840" => array("type"=>"town", "pop"=>2332, "name_ru"=>"Албион", "name_en"=>"Albion", "lang"=>"en", "lat"=>41.395197, "lon"=>-85.419695),
"/US/US-IN/alexandria-5841" => array("type"=>"city", "pop"=>5854, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>40.258178, "lon"=>-85.676177),
"/US/US-IN/alfordsville-5842" => array("type"=>"town", "pop"=>115, "name_ru"=>"Алфордсвилл", "name_en"=>"Alfordsville", "lang"=>"en", "lat"=>38.560124, "lon"=>-86.948694),
"/US/US-IN/alton-5843" => array("type"=>"town", "pop"=>52, "name_ru"=>"Альтон", "name_en"=>"Alton", "lang"=>"en", "lat"=>38.122341, "lon"=>-86.421240),
"/US/US-IN/altona-5844" => array("type"=>"town", "pop"=>192, "name_ru"=>"Альтона", "name_en"=>"Altona", "lang"=>"en", "lat"=>41.350892, "lon"=>-85.152849),
"/US/US-IN/ambia-5845" => array("type"=>"town", "pop"=>180, "name_ru"=>"Амбия", "name_en"=>"Ambia", "lang"=>"en", "lat"=>40.488823, "lon"=>-87.516182),
"/US/US-IN/amboy-5846" => array("type"=>"town", "pop"=>332, "name_ru"=>"Амбой", "name_en"=>"Amboy", "lang"=>"en", "lat"=>40.601629, "lon"=>-85.927384),
"/US/US-IN/amo-5847" => array("type"=>"town", "pop"=>458, "name_ru"=>"Амо", "name_en"=>"Amo", "lang"=>"en", "lat"=>39.689439, "lon"=>-86.614085),
"/US/US-IN/anderson-5848" => array("type"=>"city", "pop"=>57282, "name_ru"=>"Андерсон", "name_en"=>"Anderson", "lang"=>"en", "lat"=>40.092900, "lon"=>-85.688049),
"/US/US-IN/andrews-5849" => array("type"=>"town", "pop"=>1245, "name_ru"=>"Эндрюс", "name_en"=>"Andrews", "lang"=>"en", "lat"=>40.859465, "lon"=>-85.601001),
"/US/US-IN/angola-5850" => array("type"=>"city", "pop"=>7894, "name_ru"=>"Ангола", "name_en"=>"Angola", "lang"=>"en", "lat"=>41.642948, "lon"=>-85.000117),
"/US/US-IN/arcadia-5851" => array("type"=>"town", "pop"=>1834, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>40.174060, "lon"=>-86.021284),
"/US/US-IN/argos-5852" => array("type"=>"town", "pop"=>1799, "name_ru"=>"Аргос", "name_en"=>"Argos", "lang"=>"en", "lat"=>41.237734, "lon"=>-86.247670),
"/US/US-IN/ashley-5853" => array("type"=>"town", "pop"=>992, "name_ru"=>"Эшли", "name_en"=>"Ashley", "lang"=>"en", "lat"=>41.525602, "lon"=>-85.066084),
"/US/US-IN/atlanta-5854" => array("type"=>"town", "pop"=>874, "name_ru"=>"Атланта", "name_en"=>"Atlanta", "lang"=>"en", "lat"=>40.214836, "lon"=>-86.027128),
"/US/US-IN/attica-5855" => array("type"=>"city", "pop"=>3298, "name_ru"=>"Аттика", "name_en"=>"Attica", "lang"=>"en", "lat"=>40.287592, "lon"=>-87.246430),
"/US/US-IN/auburn-5856" => array("type"=>"city", "pop"=>13026, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>41.363316, "lon"=>-85.057625),
"/US/US-IN/aurora-5857" => array("type"=>"city", "pop"=>4063, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>39.066405, "lon"=>-84.903292),
"/US/US-IN/austin-5858" => array("type"=>"town", "pop"=>4603, "name_ru"=>"Остин", "name_en"=>"Austin", "lang"=>"en", "lat"=>38.743379, "lon"=>-85.810211),
"/US/US-IN/avilla-5859" => array("type"=>"town", "pop"=>2417, "name_ru"=>"Авилла", "name_en"=>"Avilla", "lang"=>"en", "lat"=>41.365002, "lon"=>-85.238055),
"/US/US-IN/bainbridge-5860" => array("type"=>"town", "pop"=>763, "name_ru"=>"Бэйнбридж", "name_en"=>"Bainbridge", "lang"=>"en", "lat"=>39.760554, "lon"=>-86.810980),
"/US/US-IN/bargersville-5861" => array("type"=>"town", "pop"=>2698, "name_ru"=>"Баргерсвилл", "name_en"=>"Bargersville", "lang"=>"en", "lat"=>39.519880, "lon"=>-86.165583),
"/US/US-IN/batesville-5862" => array("type"=>"city", "pop"=>6434, "name_ru"=>"Батсвилл", "name_en"=>"Batesville", "lang"=>"en", "lat"=>39.297938, "lon"=>-85.218294),
"/US/US-IN/battle-ground-5863" => array("type"=>"town", "pop"=>1425, "name_ru"=>"Бэттл Граунд", "name_en"=>"Battle Ground", "lang"=>"en", "lat"=>40.509988, "lon"=>-86.838484),
"/US/US-IN/bedford-5864" => array("type"=>"city", "pop"=>13486, "name_ru"=>"Бедфорд", "name_en"=>"Bedford", "lang"=>"en", "lat"=>38.860367, "lon"=>-86.490828),
"/US/US-IN/beech-grove-5865" => array("type"=>"city", "pop"=>14316, "name_ru"=>"Бич Гров", "name_en"=>"Beech Grove", "lang"=>"en", "lat"=>39.715000, "lon"=>-86.087370),
"/US/US-IN/berne-5866" => array("type"=>"city", "pop"=>4381, "name_ru"=>"Берн", "name_en"=>"Berne", "lang"=>"en", "lat"=>40.657128, "lon"=>-84.954289),
"/US/US-IN/bethany-5867" => array("type"=>"town", "pop"=>92, "name_ru"=>"Бетани", "name_en"=>"Bethany", "lang"=>"en", "lat"=>39.533745, "lon"=>-86.378854),
"/US/US-IN/beverly-shores-5868" => array("type"=>"town", "pop"=>729, "name_ru"=>"Беверли Шорс", "name_en"=>"Beverly Shores", "lang"=>"en", "lat"=>41.685472, "lon"=>-86.978845),
"/US/US-IN/bicknell-5869" => array("type"=>"city", "pop"=>3235, "name_ru"=>"Бикнелл", "name_en"=>"Bicknell", "lang"=>"en", "lat"=>38.775275, "lon"=>-87.308363),
"/US/US-IN/birdseye-5870" => array("type"=>"town", "pop"=>468, "name_ru"=>"Бердси", "name_en"=>"Birdseye", "lang"=>"en", "lat"=>38.313581, "lon"=>-86.695898),
"/US/US-IN/bloomfield-5871" => array("type"=>"town", "pop"=>2479, "name_ru"=>"Блумфилд", "name_en"=>"Bloomfield", "lang"=>"en", "lat"=>39.026387, "lon"=>-86.937007),
"/US/US-IN/bloomingdale-5872" => array("type"=>"town", "pop"=>307, "name_ru"=>"Блумингдэйл", "name_en"=>"Bloomingdale", "lang"=>"en", "lat"=>39.831521, "lon"=>-87.249508),
"/US/US-IN/bloomington-5873" => array("type"=>"city", "pop"=>71819, "name_ru"=>"Блумингтон", "name_en"=>"Bloomington", "lang"=>"en", "lat"=>39.166550, "lon"=>-86.521523),
"/US/US-IN/blountsville-5874" => array("type"=>"town", "pop"=>161, "name_ru"=>"Блантсвилл", "name_en"=>"Blountsville", "lang"=>"en", "lat"=>40.059566, "lon"=>-85.238874),
"/US/US-IN/bluffton-5875" => array("type"=>"city", "pop"=>9320, "name_ru"=>"Блаффтон", "name_en"=>"Bluffton", "lang"=>"en", "lat"=>40.741502, "lon"=>-85.170899),
"/US/US-IN/boonville-5876" => array("type"=>"city", "pop"=>6746, "name_ru"=>"Бунвилл", "name_en"=>"Boonville", "lang"=>"en", "lat"=>38.045123, "lon"=>-87.273722),
"/US/US-IN/boston-5877" => array("type"=>"town", "pop"=>168, "name_ru"=>"Бостон", "name_en"=>"Boston", "lang"=>"en", "lat"=>39.741523, "lon"=>-84.851140),
"/US/US-IN/boswell-5878" => array("type"=>"town", "pop"=>755, "name_ru"=>"Босуэлл", "name_en"=>"Boswell", "lang"=>"en", "lat"=>40.518989, "lon"=>-87.382617),
"/US/US-IN/bourbon-5879" => array("type"=>"town", "pop"=>1764, "name_ru"=>"Бурбон", "name_en"=>"Bourbon", "lang"=>"en", "lat"=>41.297771, "lon"=>-86.116921),
"/US/US-IN/brazil-5880" => array("type"=>"city", "pop"=>8272, "name_ru"=>"Бразил", "name_en"=>"Brazil", "lang"=>"en", "lat"=>39.523581, "lon"=>-87.123358),
"/US/US-IN/bremen-5881" => array("type"=>"town", "pop"=>4646, "name_ru"=>"Бремен", "name_en"=>"Bremen", "lang"=>"en", "lat"=>41.447018, "lon"=>-86.154357),
"/US/US-IN/bright-5882" => array("type"=>"cdp", "pop"=>3945, "name_ru"=>"Брайт", "name_en"=>"Bright", "lang"=>"en", "lat"=>39.202650, "lon"=>-84.857578),
"/US/US-IN/bristol-5883" => array("type"=>"town", "pop"=>1647, "name_ru"=>"Бристол", "name_en"=>"Bristol", "lang"=>"en", "lat"=>41.720955, "lon"=>-85.819411),
"/US/US-IN/brook-5884" => array("type"=>"town", "pop"=>980, "name_ru"=>"Брук", "name_en"=>"Brook", "lang"=>"en", "lat"=>40.865990, "lon"=>-87.365787),
"/US/US-IN/brooklyn-5885" => array("type"=>"town", "pop"=>1497, "name_ru"=>"Бруклин", "name_en"=>"Brooklyn", "lang"=>"en", "lat"=>39.542613, "lon"=>-86.371495),
"/US/US-IN/brooksburg-5886" => array("type"=>"town", "pop"=>72, "name_ru"=>"Бруксбург", "name_en"=>"Brooksburg", "lang"=>"en", "lat"=>38.735000, "lon"=>-85.243835),
"/US/US-IN/brookston-5887" => array("type"=>"town", "pop"=>1576, "name_ru"=>"Брукстон", "name_en"=>"Brookston", "lang"=>"en", "lat"=>40.600170, "lon"=>-86.865777),
"/US/US-IN/brookville-5888" => array("type"=>"town", "pop"=>2922, "name_ru"=>"Бруквилл", "name_en"=>"Brookville", "lang"=>"en", "lat"=>39.423592, "lon"=>-85.008363),
"/US/US-IN/brownsburg-5889" => array("type"=>"town", "pop"=>19994, "name_ru"=>"Браунсбург", "name_en"=>"Brownsburg", "lang"=>"en", "lat"=>39.841561, "lon"=>-86.394310),
"/US/US-IN/brownstown-5890" => array("type"=>"town", "pop"=>2981, "name_ru"=>"Браунстаун", "name_en"=>"Brownstown", "lang"=>"en", "lat"=>38.880334, "lon"=>-86.046127),
"/US/US-IN/bruceville-5891" => array("type"=>"town", "pop"=>450, "name_ru"=>"Брюсвилл", "name_en"=>"Bruceville", "lang"=>"en", "lat"=>38.758106, "lon"=>-87.414411),
"/US/US-IN/bryant-5892" => array("type"=>"town", "pop"=>258, "name_ru"=>"Брайант", "name_en"=>"Bryant", "lang"=>"en", "lat"=>40.534245, "lon"=>-84.963027),
"/US/US-IN/bunker-hill-5893" => array("type"=>"town", "pop"=>978, "name_ru"=>"Банкер Хилл", "name_en"=>"Bunker Hill", "lang"=>"en", "lat"=>40.660354, "lon"=>-86.101762),
"/US/US-IN/burket-5894" => array("type"=>"town", "pop"=>200, "name_ru"=>"Беркет", "name_en"=>"Burket", "lang"=>"en", "lat"=>41.154397, "lon"=>-85.969161),
"/US/US-IN/burlington-5895" => array("type"=>"town", "pop"=>431, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>40.481272, "lon"=>-86.394467),
"/US/US-IN/burnettsville-5896" => array("type"=>"town", "pop"=>341, "name_ru"=>"Бернетсвилл", "name_en"=>"Burnettsville", "lang"=>"en", "lat"=>40.761526, "lon"=>-86.594775),
"/US/US-IN/burns-harbor-5897" => array("type"=>"town", "pop"=>1100, "name_ru"=>"Бернс Харбор", "name_en"=>"Burns Harbor", "lang"=>"en", "lat"=>41.613988, "lon"=>-87.125472),
"/US/US-IN/butler-5898" => array("type"=>"city", "pop"=>2673, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>41.429168, "lon"=>-84.870607),
"/US/US-IN/cadiz-5899" => array("type"=>"town", "pop"=>175, "name_ru"=>"Кадиз", "name_en"=>"Cadiz", "lang"=>"en", "lat"=>39.950478, "lon"=>-85.486523),
"/US/US-IN/cambridge-city-5900" => array("type"=>"town", "pop"=>1966, "name_ru"=>"Кэмбридж Сити", "name_en"=>"Cambridge City", "lang"=>"en", "lat"=>39.812012, "lon"=>-85.170776),
"/US/US-IN/camden-5901" => array("type"=>"town", "pop"=>546, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>40.610167, "lon"=>-86.538682),
"/US/US-IN/campbellsburg-5902" => array("type"=>"town", "pop"=>580, "name_ru"=>"Кэмпбэллсбург", "name_en"=>"Campbellsburg", "lang"=>"en", "lat"=>38.651223, "lon"=>-86.260838),
"/US/US-IN/cannelburg-5903" => array("type"=>"town", "pop"=>157, "name_ru"=>"Каннелберг", "name_en"=>"Cannelburg", "lang"=>"en", "lat"=>38.669040, "lon"=>-86.997870),
"/US/US-IN/cannelton-5904" => array("type"=>"city", "pop"=>1140, "name_ru"=>"Каннелтон", "name_en"=>"Cannelton", "lang"=>"en", "lat"=>37.910149, "lon"=>-86.736774),
"/US/US-IN/carbon-5905" => array("type"=>"town", "pop"=>332, "name_ru"=>"Карбон", "name_en"=>"Carbon", "lang"=>"en", "lat"=>39.598933, "lon"=>-87.107496),
"/US/US-IN/carefree-5906" => array("type"=>"town", "pop"=>26, "name_ru"=>"Кэрфри", "name_en"=>"Carefree", "lang"=>"en", "lat"=>38.243272, "lon"=>-86.357890),
"/US/US-IN/carlisle-5907" => array("type"=>"town", "pop"=>677, "name_ru"=>"Карлайл", "name_en"=>"Carlisle", "lang"=>"en", "lat"=>38.962197, "lon"=>-87.400673),
"/US/US-IN/carmel-5908" => array("type"=>"city", "pop"=>66769, "name_ru"=>"Кармел", "name_en"=>"Carmel", "lang"=>"en", "lat"=>39.969319, "lon"=>-86.109507),
"/US/US-IN/carthage-5909" => array("type"=>"town", "pop"=>847, "name_ru"=>"Картаж", "name_en"=>"Carthage", "lang"=>"en", "lat"=>39.737217, "lon"=>-85.571579),
"/US/US-IN/castleton-5910" => array("type"=>"town", "pop"=>37, "name_ru"=>"Каслтон", "name_en"=>"Castleton", "lang"=>"en", "lat"=>39.904758, "lon"=>-86.050102),
"/US/US-IN/cayuga-5911" => array("type"=>"town", "pop"=>1087, "name_ru"=>"Каюга", "name_en"=>"Cayuga", "lang"=>"en", "lat"=>39.947091, "lon"=>-87.464838),
"/US/US-IN/cedar-grove-5912" => array("type"=>"town", "pop"=>183, "name_ru"=>"Седар Гроув", "name_en"=>"Cedar Grove", "lang"=>"en", "lat"=>39.355744, "lon"=>-84.937025),
"/US/US-IN/cedar-lake-5913" => array("type"=>"town", "pop"=>10981, "name_ru"=>"Сидар Лейк", "name_en"=>"Cedar Lake", "lang"=>"en", "lat"=>41.373272, "lon"=>-87.436177),
"/US/US-IN/center-point-5914" => array("type"=>"town", "pop"=>289, "name_ru"=>"Сентер Поинт", "name_en"=>"Center Point", "lang"=>"en", "lat"=>39.414788, "lon"=>-87.074537),
"/US/US-IN/centerville-5915" => array("type"=>"town", "pop"=>2588, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>39.817171, "lon"=>-84.995473),
"/US/US-IN/chalmers-5916" => array("type"=>"town", "pop"=>468, "name_ru"=>"Чалмерс", "name_en"=>"Chalmers", "lang"=>"en", "lat"=>40.662539, "lon"=>-86.867593),
"/US/US-IN/chandler-5917" => array("type"=>"town", "pop"=>3098, "name_ru"=>"Чандлер", "name_en"=>"Chandler", "lang"=>"en", "lat"=>38.042014, "lon"=>-87.368979),
"/US/US-IN/charlestown-5918" => array("type"=>"city", "pop"=>7324, "name_ru"=>"Чарльзтаун", "name_en"=>"Charlestown", "lang"=>"en", "lat"=>38.453806, "lon"=>-85.668056),
"/US/US-IN/chesterfield-5919" => array("type"=>"town", "pop"=>2755, "name_ru"=>"Честерфилд", "name_en"=>"Chesterfield", "lang"=>"en", "lat"=>40.111867, "lon"=>-85.592848),
"/US/US-IN/chesterton-5920" => array("type"=>"town", "pop"=>12705, "name_ru"=>"Честертон", "name_en"=>"Chesterton", "lang"=>"en", "lat"=>41.600920, "lon"=>-87.056959),
"/US/US-IN/chrisney-5921" => array("type"=>"town", "pop"=>530, "name_ru"=>"Кристни", "name_en"=>"Chrisney", "lang"=>"en", "lat"=>38.012263, "lon"=>-87.034057),
"/US/US-IN/churubusco-5922" => array("type"=>"town", "pop"=>1778, "name_ru"=>"Черабеско", "name_en"=>"Churubusco", "lang"=>"en", "lat"=>41.232432, "lon"=>-85.321411),
"/US/US-IN/cicero-5923" => array("type"=>"town", "pop"=>4451, "name_ru"=>"Сисеро", "name_en"=>"Cicero", "lang"=>"en", "lat"=>40.124701, "lon"=>-86.022567),
"/US/US-IN/clarks-hill-5924" => array("type"=>"town", "pop"=>688, "name_ru"=>"Кларкс Хилл", "name_en"=>"Clarks Hill", "lang"=>"en", "lat"=>40.246879, "lon"=>-86.724629),
"/US/US-IN/clarksville-5925" => array("type"=>"town", "pop"=>21873, "name_ru"=>"Кларксвилл", "name_en"=>"Clarksville", "lang"=>"en", "lat"=>38.316416, "lon"=>-85.765392),
"/US/US-IN/clay-city-5926" => array("type"=>"town", "pop"=>1010, "name_ru"=>"Клей Сити", "name_en"=>"Clay City", "lang"=>"en", "lat"=>39.277007, "lon"=>-87.112067),
"/US/US-IN/claypool-5927" => array("type"=>"town", "pop"=>300, "name_ru"=>"Клэйпул", "name_en"=>"Claypool", "lang"=>"en", "lat"=>41.127846, "lon"=>-85.879651),
"/US/US-IN/clayton-5928" => array("type"=>"town", "pop"=>825, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>39.689375, "lon"=>-86.523577),
"/US/US-IN/clear-lake-5929" => array("type"=>"town", "pop"=>273, "name_ru"=>"Клир Лейк", "name_en"=>"Clear Lake", "lang"=>"en", "lat"=>41.735823, "lon"=>-84.838338),
"/US/US-IN/clermont-5930" => array("type"=>"town", "pop"=>1485, "name_ru"=>"Клермонт", "name_en"=>"Clermont", "lang"=>"en", "lat"=>39.817362, "lon"=>-86.320442),
"/US/US-IN/clifford-5931" => array("type"=>"town", "pop"=>305, "name_ru"=>"Клиффорд", "name_en"=>"Clifford", "lang"=>"en", "lat"=>39.283044, "lon"=>-85.869638),
"/US/US-IN/clinton-5932" => array("type"=>"city", "pop"=>4799, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>39.660964, "lon"=>-87.404476),
"/US/US-IN/cloverdale-5933" => array("type"=>"town", "pop"=>2194, "name_ru"=>"Кловердейл", "name_en"=>"Cloverdale", "lang"=>"en", "lat"=>39.521273, "lon"=>-86.798724),
"/US/US-IN/coatesville-5934" => array("type"=>"town", "pop"=>569, "name_ru"=>"Коатсвилл", "name_en"=>"Coatesville", "lang"=>"en", "lat"=>39.687332, "lon"=>-86.669243),
"/US/US-IN/colfax-5935" => array("type"=>"town", "pop"=>772, "name_ru"=>"Колфакс", "name_en"=>"Colfax", "lang"=>"en", "lat"=>40.194311, "lon"=>-86.667250),
"/US/US-IN/collegeville-5936" => array("type"=>"cdp", "pop"=>993, "name_ru"=>"Коллиджвилл", "name_en"=>"Collegeville", "lang"=>"en", "lat"=>40.908135, "lon"=>-87.163922),
"/US/US-IN/columbia-city-5937" => array("type"=>"city", "pop"=>8283, "name_ru"=>"Колумбия Сити", "name_en"=>"Columbia City", "lang"=>"en", "lat"=>41.158558, "lon"=>-85.484816),
"/US/US-IN/columbus-5938" => array("type"=>"city", "pop"=>40001, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>39.210652, "lon"=>-85.913377),
"/US/US-IN/connersville-5939" => array("type"=>"city", "pop"=>13931, "name_ru"=>"Коннерсвилл", "name_en"=>"Connersville", "lang"=>"en", "lat"=>39.655704, "lon"=>-85.140720),
"/US/US-IN/converse-5940" => array("type"=>"town", "pop"=>1087, "name_ru"=>"Конверс", "name_en"=>"Converse", "lang"=>"en", "lat"=>40.579941, "lon"=>-85.877914),
"/US/US-IN/corunna-5941" => array("type"=>"town", "pop"=>247, "name_ru"=>"Коруна", "name_en"=>"Corunna", "lang"=>"en", "lat"=>41.435740, "lon"=>-85.144177),
"/US/US-IN/corydon-5942" => array("type"=>"town", "pop"=>2780, "name_ru"=>"Коридон", "name_en"=>"Corydon", "lang"=>"en", "lat"=>38.212657, "lon"=>-86.125347),
"/US/US-IN/country-club-heights-5943" => array("type"=>"town", "pop"=>86, "name_ru"=>"Сантри Клаб Хайтс", "name_en"=>"Country Club Heights", "lang"=>"en", "lat"=>40.125367, "lon"=>-85.687580),
"/US/US-IN/covington-5944" => array("type"=>"city", "pop"=>2432, "name_ru"=>"Ковингтон", "name_en"=>"Covington", "lang"=>"en", "lat"=>40.139851, "lon"=>-87.391241),
"/US/US-IN/crandall-5945" => array("type"=>"town", "pop"=>133, "name_ru"=>"Крэндолл", "name_en"=>"Crandall", "lang"=>"en", "lat"=>38.287841, "lon"=>-86.065830),
"/US/US-IN/crane-5946" => array("type"=>"town", "pop"=>193, "name_ru"=>"Крейн", "name_en"=>"Crane", "lang"=>"en", "lat"=>38.895119, "lon"=>-86.900374),
"/US/US-IN/crawfordsville-5947" => array("type"=>"city", "pop"=>15043, "name_ru"=>"Крофордсвилл", "name_en"=>"Crawfordsville", "lang"=>"en", "lat"=>40.044225, "lon"=>-86.896641),
"/US/US-IN/cromwell-5948" => array("type"=>"town", "pop"=>498, "name_ru"=>"Кромуэлл", "name_en"=>"Cromwell", "lang"=>"en", "lat"=>41.403489, "lon"=>-85.614435),
"/US/US-IN/crothersville-5949" => array("type"=>"town", "pop"=>1508, "name_ru"=>"Кротерсвилл", "name_en"=>"Crothersville", "lang"=>"en", "lat"=>38.796413, "lon"=>-85.839978),
"/US/US-IN/crown-point-5950" => array("type"=>"city", "pop"=>24376, "name_ru"=>"Краун Пойнт", "name_en"=>"Crown Point", "lang"=>"en", "lat"=>41.424248, "lon"=>-87.357715),
"/US/US-IN/crows-nest-5951" => array("type"=>"town", "pop"=>112, "name_ru"=>"Краус Нест", "name_en"=>"Crows Nest", "lang"=>"en", "lat"=>39.852167, "lon"=>-86.172823),
"/US/US-IN/culver-5952" => array("type"=>"town", "pop"=>1504, "name_ru"=>"Калвер", "name_en"=>"Culver", "lang"=>"en", "lat"=>41.217912, "lon"=>-86.422515),
"/US/US-IN/cumberland-5953" => array("type"=>"town", "pop"=>5580, "name_ru"=>"Камберленд", "name_en"=>"Cumberland", "lang"=>"en", "lat"=>39.782800, "lon"=>-85.952376),
"/US/US-IN/cynthiana-5954" => array("type"=>"town", "pop"=>643, "name_ru"=>"Синтиана", "name_en"=>"Cynthiana", "lang"=>"en", "lat"=>38.186920, "lon"=>-87.708794),
"/US/US-IN/dale-5955" => array("type"=>"town", "pop"=>1495, "name_ru"=>"Дейл", "name_en"=>"Dale", "lang"=>"en", "lat"=>38.167201, "lon"=>-86.986612),
"/US/US-IN/daleville-5956" => array("type"=>"town", "pop"=>1564, "name_ru"=>"Дейлвилл", "name_en"=>"Daleville", "lang"=>"en", "lat"=>40.118840, "lon"=>-85.556251),
"/US/US-IN/dana-5957" => array("type"=>"town", "pop"=>661, "name_ru"=>"Дана", "name_en"=>"Dana", "lang"=>"en", "lat"=>39.806678, "lon"=>-87.494238),
"/US/US-IN/danville-5958" => array("type"=>"town", "pop"=>8124, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>39.762218, "lon"=>-86.519790),
"/US/US-IN/darlington-5959" => array("type"=>"town", "pop"=>828, "name_ru"=>"Дарлингтон", "name_en"=>"Darlington", "lang"=>"en", "lat"=>40.108178, "lon"=>-86.776549),
"/US/US-IN/darmstadt-5960" => array("type"=>"town", "pop"=>1502, "name_ru"=>"Дармштадт", "name_en"=>"Darmstadt", "lang"=>"en", "lat"=>38.090998, "lon"=>-87.576189),
"/US/US-IN/dayton-5961" => array("type"=>"town", "pop"=>1208, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>40.377584, "lon"=>-86.776754),
"/US/US-IN/de-motte-5965" => array("type"=>"town", "pop"=>4352, "name_ru"=>"Де Мотте", "name_en"=>"De Motte", "lang"=>"en", "lat"=>41.199651, "lon"=>-87.199197),
"/US/US-IN/decatur-5962" => array("type"=>"city", "pop"=>9574, "name_ru"=>"Декатур", "name_en"=>"Decatur", "lang"=>"en", "lat"=>40.830015, "lon"=>-84.928236),
"/US/US-IN/decker-5963" => array("type"=>"town", "pop"=>272, "name_ru"=>"Декер", "name_en"=>"Decker", "lang"=>"en", "lat"=>38.518344, "lon"=>-87.523793),
"/US/US-IN/delphi-5964" => array("type"=>"city", "pop"=>2862, "name_ru"=>"Дэлфи", "name_en"=>"Delphi", "lang"=>"en", "lat"=>40.584611, "lon"=>-86.670341),
"/US/US-IN/denver-5966" => array("type"=>"town", "pop"=>499, "name_ru"=>"Денвер", "name_en"=>"Denver", "lang"=>"en", "lat"=>40.864247, "lon"=>-86.076391),
"/US/US-IN/dillsboro-5967" => array("type"=>"town", "pop"=>1456, "name_ru"=>"Диллсборо", "name_en"=>"Dillsboro", "lang"=>"en", "lat"=>39.018619, "lon"=>-85.061733),
"/US/US-IN/dublin-5968" => array("type"=>"town", "pop"=>648, "name_ru"=>"Дублин", "name_en"=>"Dublin", "lang"=>"en", "lat"=>39.812446, "lon"=>-85.205097),
"/US/US-IN/dugger-5969" => array("type"=>"town", "pop"=>934, "name_ru"=>"Даггер", "name_en"=>"Dugger", "lang"=>"en", "lat"=>39.068411, "lon"=>-87.260448),
"/US/US-IN/dune-acres-5970" => array("type"=>"town", "pop"=>231, "name_ru"=>"Дьюн Эйкрес", "name_en"=>"Dune Acres", "lang"=>"en", "lat"=>41.643656, "lon"=>-87.103317),
"/US/US-IN/dunkirk-5971" => array("type"=>"city", "pop"=>2590, "name_ru"=>"Дюнкерк", "name_en"=>"Dunkirk", "lang"=>"en", "lat"=>40.373519, "lon"=>-85.207933),
"/US/US-IN/dunlap-5972" => array("type"=>"cdp", "pop"=>5705, "name_ru"=>"Данлап", "name_en"=>"Dunlap", "lang"=>"en", "lat"=>41.635500, "lon"=>-85.919396),
"/US/US-IN/dunreith-5973" => array("type"=>"town", "pop"=>177, "name_ru"=>"Дюнрайт", "name_en"=>"Dunreith", "lang"=>"en", "lat"=>39.802895, "lon"=>-85.436878),
"/US/US-IN/dupont-5974" => array("type"=>"town", "pop"=>389, "name_ru"=>"Дюпон", "name_en"=>"Dupont", "lang"=>"en", "lat"=>38.891142, "lon"=>-85.516920),
"/US/US-IN/dyer-5975" => array("type"=>"town", "pop"=>15913, "name_ru"=>"Дайер", "name_en"=>"Dyer", "lang"=>"en", "lat"=>41.503400, "lon"=>-87.505200),
"/US/US-IN/earl-park-5976" => array("type"=>"town", "pop"=>443, "name_ru"=>"Ерл Парк", "name_en"=>"Earl Park", "lang"=>"en", "lat"=>40.685826, "lon"=>-87.420428),
"/US/US-IN/east-chicago-5977" => array("type"=>"city", "pop"=>29978, "name_ru"=>"Ист Чикаго", "name_en"=>"East Chicago", "lang"=>"en", "lat"=>41.648650, "lon"=>-87.451872),
"/US/US-IN/east-germantown-5978" => array("type"=>"town", "pop"=>224, "name_ru"=>"Ист Джермантаун", "name_en"=>"East Germantown", "lang"=>"en", "lat"=>39.812768, "lon"=>-85.136684),
"/US/US-IN/eaton-5979" => array("type"=>"town", "pop"=>1484, "name_ru"=>"Итон", "name_en"=>"Eaton", "lang"=>"en", "lat"=>40.340525, "lon"=>-85.354870),
"/US/US-IN/economy-5980" => array("type"=>"town", "pop"=>191, "name_ru"=>"Экономи", "name_en"=>"Economy", "lang"=>"en", "lat"=>39.977425, "lon"=>-85.087117),
"/US/US-IN/edgewood-5981" => array("type"=>"town", "pop"=>1867, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>40.102600, "lon"=>-85.737610),
"/US/US-IN/edinburgh-5982" => array("type"=>"town", "pop"=>4677, "name_ru"=>"Эдинбург", "name_en"=>"Edinburgh", "lang"=>"en", "lat"=>39.352678, "lon"=>-85.967475),
"/US/US-IN/edwardsport-5983" => array("type"=>"town", "pop"=>345, "name_ru"=>"Эдвардспорт", "name_en"=>"Edwardsport", "lang"=>"en", "lat"=>38.812387, "lon"=>-87.251592),
"/US/US-IN/elberfeld-5984" => array("type"=>"town", "pop"=>640, "name_ru"=>"Элбертфелд", "name_en"=>"Elberfeld", "lang"=>"en", "lat"=>38.161202, "lon"=>-87.448087),
"/US/US-IN/elizabeth-5985" => array("type"=>"town", "pop"=>145, "name_ru"=>"Элизабет", "name_en"=>"Elizabeth", "lang"=>"en", "lat"=>38.125328, "lon"=>-85.974766),
"/US/US-IN/elizabethtown-5986" => array("type"=>"town", "pop"=>406, "name_ru"=>"Элизабеттаун", "name_en"=>"Elizabethtown", "lang"=>"en", "lat"=>39.135345, "lon"=>-85.812700),
"/US/US-IN/elkhart-5987" => array("type"=>"city", "pop"=>52653, "name_ru"=>"Элкхарт", "name_en"=>"Elkhart", "lang"=>"en", "lat"=>41.686800, "lon"=>-85.968790),
"/US/US-IN/ellettsville-5988" => array("type"=>"town", "pop"=>6017, "name_ru"=>"Эллетсвилл", "name_en"=>"Ellettsville", "lang"=>"en", "lat"=>39.231400, "lon"=>-86.622149),
"/US/US-IN/elnora-5989" => array("type"=>"town", "pop"=>728, "name_ru"=>"Элнора", "name_en"=>"Elnora", "lang"=>"en", "lat"=>38.877223, "lon"=>-87.084914),
"/US/US-IN/elwood-5990" => array("type"=>"city", "pop"=>9027, "name_ru"=>"Элвуд", "name_en"=>"Elwood", "lang"=>"en", "lat"=>40.275726, "lon"=>-85.838933),
"/US/US-IN/english-5991" => array("type"=>"town", "pop"=>660, "name_ru"=>"Инглиш", "name_en"=>"English", "lang"=>"en", "lat"=>38.333024, "lon"=>-86.463277),
"/US/US-IN/etna-green-5992" => array("type"=>"town", "pop"=>626, "name_ru"=>"Этна Грин", "name_en"=>"Etna Green", "lang"=>"en", "lat"=>41.278108, "lon"=>-86.046250),
"/US/US-IN/evansville-5993" => array("type"=>"city", "pop"=>116309, "name_ru"=>"Эвансвилл", "name_en"=>"Evansville", "lang"=>"en", "lat"=>37.984150, "lon"=>-87.543299),
"/US/US-IN/fairland-5994" => array("type"=>"cdp", "pop"=>1348, "name_ru"=>"Фэрленд", "name_en"=>"Fairland", "lang"=>"en", "lat"=>39.598721, "lon"=>-85.858124),
"/US/US-IN/fairmount-5995" => array("type"=>"town", "pop"=>2734, "name_ru"=>"Фэрмаунт", "name_en"=>"Fairmount", "lang"=>"en", "lat"=>40.416906, "lon"=>-85.647852),
"/US/US-IN/fairview-park-5996" => array("type"=>"town", "pop"=>1465, "name_ru"=>"Фэрвью Парк", "name_en"=>"Fairview Park", "lang"=>"en", "lat"=>39.681759, "lon"=>-87.413552),
"/US/US-IN/farmersburg-5997" => array("type"=>"town", "pop"=>1178, "name_ru"=>"Фармерсбург", "name_en"=>"Farmersburg", "lang"=>"en", "lat"=>39.252427, "lon"=>-87.380350),
"/US/US-IN/farmland-5998" => array("type"=>"town", "pop"=>1341, "name_ru"=>"Фармленд", "name_en"=>"Farmland", "lang"=>"en", "lat"=>40.188956, "lon"=>-85.127135),
"/US/US-IN/ferdinand-5999" => array("type"=>"town", "pop"=>2336, "name_ru"=>"Фердинанд", "name_en"=>"Ferdinand", "lang"=>"en", "lat"=>38.228255, "lon"=>-86.862118),
"/US/US-IN/fillmore-6000" => array("type"=>"town", "pop"=>549, "name_ru"=>"Фильмор", "name_en"=>"Fillmore", "lang"=>"en", "lat"=>39.670966, "lon"=>-86.753606),
"/US/US-IN/fishers-6001" => array("type"=>"town", "pop"=>69011, "name_ru"=>"Фишерс", "name_en"=>"Fishers", "lang"=>"en", "lat"=>39.954133, "lon"=>-86.023576),
"/US/US-IN/flora-6002" => array("type"=>"town", "pop"=>2085, "name_ru"=>"Флора", "name_en"=>"Flora", "lang"=>"en", "lat"=>40.544936, "lon"=>-86.523363),
"/US/US-IN/fort-branch-6003" => array("type"=>"town", "pop"=>2550, "name_ru"=>"Форт Брэнч", "name_en"=>"Fort Branch", "lang"=>"en", "lat"=>38.246783, "lon"=>-87.574835),
"/US/US-IN/fort-wayne-6005" => array("type"=>"city", "pop"=>251591, "name_ru"=>"Форт Уэйн", "name_en"=>"Fort Wayne", "lang"=>"en", "lat"=>41.073750, "lon"=>-85.139024),
"/US/US-IN/fortville-6004" => array("type"=>"town", "pop"=>3732, "name_ru"=>"Фортвилл", "name_en"=>"Fortville", "lang"=>"en", "lat"=>39.935900, "lon"=>-85.846973),
"/US/US-IN/fountain-city-6006" => array("type"=>"town", "pop"=>705, "name_ru"=>"Фаунтейн Сити", "name_en"=>"Fountain City", "lang"=>"en", "lat"=>39.955873, "lon"=>-84.917851),
"/US/US-IN/fowler-6007" => array("type"=>"town", "pop"=>2199, "name_ru"=>"Фоулер", "name_en"=>"Fowler", "lang"=>"en", "lat"=>40.617206, "lon"=>-87.317011),
"/US/US-IN/fowlerton-6008" => array("type"=>"town", "pop"=>276, "name_ru"=>"Фаулертон", "name_en"=>"Fowlerton", "lang"=>"en", "lat"=>40.409638, "lon"=>-85.573035),
"/US/US-IN/francesville-6009" => array("type"=>"town", "pop"=>859, "name_ru"=>"Фрэнсисвилл", "name_en"=>"Francesville", "lang"=>"en", "lat"=>40.985819, "lon"=>-86.883851),
"/US/US-IN/francisco-6010" => array("type"=>"town", "pop"=>526, "name_ru"=>"Франциско", "name_en"=>"Francisco", "lang"=>"en", "lat"=>38.333962, "lon"=>-87.448628),
"/US/US-IN/frankfort-6011" => array("type"=>"city", "pop"=>16394, "name_ru"=>"Франкфорт", "name_en"=>"Frankfort", "lang"=>"en", "lat"=>40.281097, "lon"=>-86.510304),
"/US/US-IN/franklin-6012" => array("type"=>"city", "pop"=>23258, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>39.492213, "lon"=>-86.056539),
"/US/US-IN/frankton-6013" => array("type"=>"town", "pop"=>1860, "name_ru"=>"Франктон", "name_en"=>"Frankton", "lang"=>"en", "lat"=>40.220528, "lon"=>-85.772488),
"/US/US-IN/fredericksburg-6014" => array("type"=>"town", "pop"=>92, "name_ru"=>"Фредериксбург", "name_en"=>"Fredericksburg", "lang"=>"en", "lat"=>38.432950, "lon"=>-86.188459),
"/US/US-IN/fremont-6015" => array("type"=>"town", "pop"=>1609, "name_ru"=>"Фремонт", "name_en"=>"Fremont", "lang"=>"en", "lat"=>41.731131, "lon"=>-84.933742),
"/US/US-IN/french-lick-6016" => array("type"=>"town", "pop"=>1921, "name_ru"=>"Френч Лик", "name_en"=>"French Lick", "lang"=>"en", "lat"=>38.547232, "lon"=>-86.620173),
"/US/US-IN/fulton-6017" => array("type"=>"town", "pop"=>316, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>40.946533, "lon"=>-86.263991),
"/US/US-IN/galena-6018" => array("type"=>"cdp", "pop"=>1777, "name_ru"=>"Галена", "name_en"=>"Galena", "lang"=>"en", "lat"=>38.350336, "lon"=>-85.939396),
"/US/US-IN/galveston-6019" => array("type"=>"town", "pop"=>1406, "name_ru"=>"Галвестон", "name_en"=>"Galveston", "lang"=>"en", "lat"=>40.576440, "lon"=>-86.191700),
"/US/US-IN/garrett-6020" => array("type"=>"city", "pop"=>5681, "name_ru"=>"Гарретт", "name_en"=>"Garrett", "lang"=>"en", "lat"=>41.349898, "lon"=>-85.127733),
"/US/US-IN/gary-6021" => array("type"=>"city", "pop"=>95920, "name_ru"=>"Гэри", "name_en"=>"Gary", "lang"=>"en", "lat"=>41.595750, "lon"=>-87.344149),
"/US/US-IN/gas-city-6022" => array("type"=>"city", "pop"=>5683, "name_ru"=>"Гэс Сити", "name_en"=>"Gas City", "lang"=>"en", "lat"=>40.489150, "lon"=>-85.607749),
"/US/US-IN/gaston-6023" => array("type"=>"town", "pop"=>944, "name_ru"=>"Гэстон", "name_en"=>"Gaston", "lang"=>"en", "lat"=>40.313506, "lon"=>-85.501276),
"/US/US-IN/geneva-6024" => array("type"=>"town", "pop"=>1306, "name_ru"=>"Дженива", "name_en"=>"Geneva", "lang"=>"en", "lat"=>40.597194, "lon"=>-84.957128),
"/US/US-IN/gentryville-6025" => array("type"=>"town", "pop"=>241, "name_ru"=>"Джентривилл", "name_en"=>"Gentryville", "lang"=>"en", "lat"=>38.108813, "lon"=>-87.030393),
"/US/US-IN/georgetown-6026" => array("type"=>"town", "pop"=>3066, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>38.299240, "lon"=>-85.970696),
"/US/US-IN/georgetown-6027" => array("type"=>"cdp", "pop"=>3993, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>41.729397, "lon"=>-86.228314),
"/US/US-IN/glenwood-6028" => array("type"=>"town", "pop"=>291, "name_ru"=>"Гленвуд", "name_en"=>"Glenwood", "lang"=>"en", "lat"=>39.626465, "lon"=>-85.301286),
"/US/US-IN/goodland-6029" => array("type"=>"town", "pop"=>997, "name_ru"=>"Гудленд", "name_en"=>"Goodland", "lang"=>"en", "lat"=>40.764330, "lon"=>-87.294935),
"/US/US-IN/goshen-6030" => array("type"=>"city", "pop"=>32630, "name_ru"=>"Гошен", "name_en"=>"Goshen", "lang"=>"en", "lat"=>41.578500, "lon"=>-85.835070),
"/US/US-IN/gosport-6031" => array("type"=>"town", "pop"=>722, "name_ru"=>"Госпорт", "name_en"=>"Gosport", "lang"=>"en", "lat"=>39.350485, "lon"=>-86.665707),
"/US/US-IN/grabill-6032" => array("type"=>"town", "pop"=>1151, "name_ru"=>"Грабилл", "name_en"=>"Grabill", "lang"=>"en", "lat"=>41.210750, "lon"=>-84.967927),
"/US/US-IN/grandview-6033" => array("type"=>"town", "pop"=>661, "name_ru"=>"Грандвью", "name_en"=>"Grandview", "lang"=>"en", "lat"=>37.937966, "lon"=>-86.983068),
"/US/US-IN/granger-6034" => array("type"=>"cdp", "pop"=>20241, "name_ru"=>"Грейнджер", "name_en"=>"Granger", "lang"=>"en", "lat"=>41.735504, "lon"=>-86.139499),
"/US/US-IN/greencastle-6035" => array("type"=>"city", "pop"=>10145, "name_ru"=>"Гринкасл", "name_en"=>"Greencastle", "lang"=>"en", "lat"=>39.642000, "lon"=>-86.840600),
"/US/US-IN/greendale-6036" => array("type"=>"town", "pop"=>4364, "name_ru"=>"Гриндейл", "name_en"=>"Greendale", "lang"=>"en", "lat"=>39.117709, "lon"=>-84.860221),
"/US/US-IN/greenfield-6037" => array("type"=>"city", "pop"=>18787, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>39.792795, "lon"=>-85.770008),
"/US/US-IN/greens-fork-6040" => array("type"=>"town", "pop"=>339, "name_ru"=>"Гринсфорк", "name_en"=>"Greens Fork", "lang"=>"en", "lat"=>39.891405, "lon"=>-85.042278),
"/US/US-IN/greensboro-6038" => array("type"=>"town", "pop"=>138, "name_ru"=>"Гринсборо", "name_en"=>"Greensboro", "lang"=>"en", "lat"=>39.878274, "lon"=>-85.463031),
"/US/US-IN/greensburg-6039" => array("type"=>"city", "pop"=>10730, "name_ru"=>"Гринсбург", "name_en"=>"Greensburg", "lang"=>"en", "lat"=>39.341377, "lon"=>-85.480544),
"/US/US-IN/greentown-6041" => array("type"=>"town", "pop"=>2521, "name_ru"=>"Гринтаун", "name_en"=>"Greentown", "lang"=>"en", "lat"=>40.478518, "lon"=>-85.964697),
"/US/US-IN/greenville-6042" => array("type"=>"town", "pop"=>583, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>38.371088, "lon"=>-85.984630),
"/US/US-IN/greenwood-6043" => array("type"=>"city", "pop"=>47720, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>39.615544, "lon"=>-86.113759),
"/US/US-IN/griffin-6044" => array("type"=>"town", "pop"=>151, "name_ru"=>"Гриффин", "name_en"=>"Griffin", "lang"=>"en", "lat"=>38.204317, "lon"=>-87.915314),
"/US/US-IN/griffith-6045" => array("type"=>"town", "pop"=>16311, "name_ru"=>"Гриффит", "name_en"=>"Griffith", "lang"=>"en", "lat"=>41.522367, "lon"=>-87.423724),
"/US/US-IN/grissom-afb-6046" => array("type"=>"cdp", "pop"=>4271, "name_ru"=>"Военно-воздушная база Гриссом", "name_en"=>"Grissom AFB", "lang"=>"en", "lat"=>40.659386, "lon"=>-86.147603),
"/US/US-IN/gulivoire-park-6047" => array("type"=>"cdp", "pop"=>2788, "name_ru"=>"Галивуар Парк", "name_en"=>"Gulivoire Park", "lang"=>"en", "lat"=>41.613214, "lon"=>-86.245212),
"/US/US-IN/hagerstown-6048" => array("type"=>"town", "pop"=>1628, "name_ru"=>"Хагерстаун", "name_en"=>"Hagerstown", "lang"=>"en", "lat"=>39.911719, "lon"=>-85.156921),
"/US/US-IN/hamilton-6049" => array("type"=>"town", "pop"=>1484, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>41.532318, "lon"=>-84.917457),
"/US/US-IN/hamlet-6050" => array("type"=>"town", "pop"=>787, "name_ru"=>"Хамлет", "name_en"=>"Hamlet", "lang"=>"en", "lat"=>41.378486, "lon"=>-86.583171),
"/US/US-IN/hammond-6051" => array("type"=>"city", "pop"=>76732, "name_ru"=>"Хаммонд", "name_en"=>"Hammond", "lang"=>"en", "lat"=>41.621000, "lon"=>-87.490364),
"/US/US-IN/hanover-6052" => array("type"=>"town", "pop"=>3873, "name_ru"=>"Хановер", "name_en"=>"Hanover", "lang"=>"en", "lat"=>38.716062, "lon"=>-85.471950),
"/US/US-IN/hardinsburg-6053" => array("type"=>"town", "pop"=>244, "name_ru"=>"Хардинсбург", "name_en"=>"Hardinsburg", "lang"=>"en", "lat"=>38.460358, "lon"=>-86.272659),
"/US/US-IN/harmony-6054" => array("type"=>"town", "pop"=>607, "name_ru"=>"Хармони", "name_en"=>"Harmony", "lang"=>"en", "lat"=>39.534020, "lon"=>-87.073417),
"/US/US-IN/hartford-city-6055" => array("type"=>"city", "pop"=>6302, "name_ru"=>"Хартфорд Сити", "name_en"=>"Hartford City", "lang"=>"en", "lat"=>40.452511, "lon"=>-85.371786),
"/US/US-IN/hartsville-6056" => array("type"=>"town", "pop"=>383, "name_ru"=>"Хартсвилл", "name_en"=>"Hartsville", "lang"=>"en", "lat"=>39.266988, "lon"=>-85.698709),
"/US/US-IN/haubstadt-6057" => array("type"=>"town", "pop"=>1485, "name_ru"=>"Хаубштадт", "name_en"=>"Haubstadt", "lang"=>"en", "lat"=>38.203105, "lon"=>-87.574844),
"/US/US-IN/hazleton-6058" => array("type"=>"town", "pop"=>281, "name_ru"=>"Хейзлтон", "name_en"=>"Hazleton", "lang"=>"en", "lat"=>38.489004, "lon"=>-87.540397),
"/US/US-IN/hebron-6059" => array("type"=>"town", "pop"=>3698, "name_ru"=>"Хеброн", "name_en"=>"Hebron", "lang"=>"en", "lat"=>41.322373, "lon"=>-87.202840),
"/US/US-IN/hidden-valley-6060" => array("type"=>"cdp", "pop"=>2116, "name_ru"=>"Хидден Валли", "name_en"=>"Hidden Valley", "lang"=>"en", "lat"=>39.162246, "lon"=>-84.843070),
"/US/US-IN/highland-6061" => array("type"=>"town", "pop"=>22641, "name_ru"=>"Хайленд", "name_en"=>"Highland", "lang"=>"en", "lat"=>41.549100, "lon"=>-87.457656),
"/US/US-IN/highland-6062" => array("type"=>"cdp", "pop"=>3508, "name_ru"=>"Хайленд", "name_en"=>"Highland", "lang"=>"en", "lat"=>38.048147, "lon"=>-87.562182),
"/US/US-IN/hillsboro-6063" => array("type"=>"town", "pop"=>466, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>40.109496, "lon"=>-87.157077),
"/US/US-IN/hobart-6064" => array("type"=>"city", "pop"=>27891, "name_ru"=>"Хобарт", "name_en"=>"Hobart", "lang"=>"en", "lat"=>41.531600, "lon"=>-87.267321),
"/US/US-IN/holland-6065" => array("type"=>"town", "pop"=>694, "name_ru"=>"Холланд", "name_en"=>"Holland", "lang"=>"en", "lat"=>38.245890, "lon"=>-87.038356),
"/US/US-IN/holton-6066" => array("type"=>"town", "pop"=>396, "name_ru"=>"Холтон", "name_en"=>"Holton", "lang"=>"en", "lat"=>39.075586, "lon"=>-85.384593),
"/US/US-IN/homecroft-6067" => array("type"=>"town", "pop"=>742, "name_ru"=>"Хоумкрофт", "name_en"=>"Homecroft", "lang"=>"en", "lat"=>39.670050, "lon"=>-86.130114),
"/US/US-IN/hope-6068" => array("type"=>"town", "pop"=>2285, "name_ru"=>"Хоуп", "name_en"=>"Hope", "lang"=>"en", "lat"=>39.299909, "lon"=>-85.768129),
"/US/US-IN/hudson-6069" => array("type"=>"town", "pop"=>563, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>41.533560, "lon"=>-85.081634),
"/US/US-IN/huntertown-6070" => array("type"=>"town", "pop"=>2158, "name_ru"=>"Хантертаун", "name_en"=>"Huntertown", "lang"=>"en", "lat"=>41.231250, "lon"=>-85.169834),
"/US/US-IN/huntingburg-6071" => array("type"=>"city", "pop"=>6112, "name_ru"=>"Хантингбург", "name_en"=>"Huntingburg", "lang"=>"en", "lat"=>38.298150, "lon"=>-86.958733),
"/US/US-IN/huntington-6072" => array("type"=>"city", "pop"=>16521, "name_ru"=>"Хантингтон", "name_en"=>"Huntington", "lang"=>"en", "lat"=>40.880732, "lon"=>-85.505583),
"/US/US-IN/hymera-6073" => array("type"=>"town", "pop"=>800, "name_ru"=>"Хаймера", "name_en"=>"Hymera", "lang"=>"en", "lat"=>39.185700, "lon"=>-87.298865),
"/US/US-IN/indian-heights-6075" => array("type"=>"cdp", "pop"=>3669, "name_ru"=>"Индиан Хайтс", "name_en"=>"Indian Heights", "lang"=>"en", "lat"=>40.424950, "lon"=>-86.116464),
"/US/US-IN/indian-village-6076" => array("type"=>"town", "pop"=>139, "name_ru"=>"Индиан Виллидж", "name_en"=>"Indian Village", "lang"=>"en", "lat"=>41.714500, "lon"=>-86.231924),
"/US/US-IN/indianapolis-(remainder)-6074" => array("type"=>"(remainder)", "pop"=>731327, "name_ru"=>"Индианаполис", "name_en"=>"Indianapolis (remainder)", "lang"=>"en", "lat"=>39.776400, "lon"=>-86.146196),
"/US/US-IN/ingalls-6077" => array("type"=>"town", "pop"=>1776, "name_ru"=>"Инголс", "name_en"=>"Ingalls", "lang"=>"en", "lat"=>39.960250, "lon"=>-85.798774),
"/US/US-IN/jamestown-6078" => array("type"=>"town", "pop"=>993, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>39.927448, "lon"=>-86.628383),
"/US/US-IN/jasonville-6079" => array("type"=>"city", "pop"=>2439, "name_ru"=>"Джейсонвилл", "name_en"=>"Jasonville", "lang"=>"en", "lat"=>39.162545, "lon"=>-87.199559),
"/US/US-IN/jasper-6080" => array("type"=>"city", "pop"=>14116, "name_ru"=>"Джаспер", "name_en"=>"Jasper", "lang"=>"en", "lat"=>38.393153, "lon"=>-86.935230),
"/US/US-IN/jeffersonville-6081" => array("type"=>"city", "pop"=>30479, "name_ru"=>"Джефферсонвилл", "name_en"=>"Jeffersonville", "lang"=>"en", "lat"=>38.300900, "lon"=>-85.726568),
"/US/US-IN/jonesboro-6082" => array("type"=>"town", "pop"=>1742, "name_ru"=>"Джонсборо", "name_en"=>"Jonesboro", "lang"=>"en", "lat"=>40.479300, "lon"=>-85.629872),
"/US/US-IN/jonesville-6083" => array("type"=>"town", "pop"=>231, "name_ru"=>"Джонсвилл", "name_en"=>"Jonesville", "lang"=>"en", "lat"=>39.060046, "lon"=>-85.888816),
"/US/US-IN/judson-6084" => array("type"=>"town", "pop"=>61, "name_ru"=>"Джадсон", "name_en"=>"Judson", "lang"=>"en", "lat"=>39.813467, "lon"=>-87.135177),
"/US/US-IN/kempton-6085" => array("type"=>"town", "pop"=>707, "name_ru"=>"Кэмптон", "name_en"=>"Kempton", "lang"=>"en", "lat"=>40.287509, "lon"=>-86.230040),
"/US/US-IN/kendallville-6086" => array("type"=>"city", "pop"=>10452, "name_ru"=>"Кендоллвилл", "name_en"=>"Kendallville", "lang"=>"en", "lat"=>41.443428, "lon"=>-85.258087),
"/US/US-IN/kennard-6087" => array("type"=>"town", "pop"=>428, "name_ru"=>"Кеннард", "name_en"=>"Kennard", "lang"=>"en", "lat"=>39.904278, "lon"=>-85.519827),
"/US/US-IN/kentland-6088" => array("type"=>"town", "pop"=>1665, "name_ru"=>"Кентленд", "name_en"=>"Kentland", "lang"=>"en", "lat"=>40.772771, "lon"=>-87.446031),
"/US/US-IN/kewanna-6089" => array("type"=>"town", "pop"=>598, "name_ru"=>"Кеванна", "name_en"=>"Kewanna", "lang"=>"en", "lat"=>41.019075, "lon"=>-86.412705),
"/US/US-IN/kingman-6090" => array("type"=>"town", "pop"=>517, "name_ru"=>"Кингман", "name_en"=>"Kingman", "lang"=>"en", "lat"=>39.965821, "lon"=>-87.278267),
"/US/US-IN/kingsbury-6091" => array("type"=>"town", "pop"=>231, "name_ru"=>"Кингсбери", "name_en"=>"Kingsbury", "lang"=>"en", "lat"=>41.531170, "lon"=>-86.695569),
"/US/US-IN/kingsford-heights-6092" => array("type"=>"town", "pop"=>1404, "name_ru"=>"Кингсфорд Хайтс", "name_en"=>"Kingsford Heights", "lang"=>"en", "lat"=>41.479970, "lon"=>-86.691261),
"/US/US-IN/kirklin-6093" => array("type"=>"town", "pop"=>772, "name_ru"=>"Кирклин", "name_en"=>"Kirklin", "lang"=>"en", "lat"=>40.193045, "lon"=>-86.359671),
"/US/US-IN/knightstown-6094" => array("type"=>"town", "pop"=>2015, "name_ru"=>"Найтстаун", "name_en"=>"Knightstown", "lang"=>"en", "lat"=>39.795694, "lon"=>-85.528295),
"/US/US-IN/knightsville-6095" => array("type"=>"town", "pop"=>620, "name_ru"=>"Найтсвилл", "name_en"=>"Knightsville", "lang"=>"en", "lat"=>39.526146, "lon"=>-87.089762),
"/US/US-IN/knox-6096" => array("type"=>"city", "pop"=>3831, "name_ru"=>"Нокс", "name_en"=>"Knox", "lang"=>"en", "lat"=>41.291910, "lon"=>-86.621341),
"/US/US-IN/kokomo-6097" => array("type"=>"city", "pop"=>45694, "name_ru"=>"Кокомбо", "name_en"=>"Kokomo", "lang"=>"en", "lat"=>40.475550, "lon"=>-86.133099),
"/US/US-IN/koontz-lake-6098" => array("type"=>"cdp", "pop"=>1615, "name_ru"=>"Кунтс Лейк", "name_en"=>"Koontz Lake", "lang"=>"en", "lat"=>41.418244, "lon"=>-86.484078),
"/US/US-IN/kouts-6099" => array("type"=>"town", "pop"=>1849, "name_ru"=>"Каутс", "name_en"=>"Kouts", "lang"=>"en", "lat"=>41.316915, "lon"=>-87.026368),
"/US/US-IN/la-crosse-6101" => array("type"=>"town", "pop"=>546, "name_ru"=>"Ла Кросс", "name_en"=>"La Crosse", "lang"=>"en", "lat"=>41.317884, "lon"=>-86.889809),
"/US/US-IN/la-fontaine-6104" => array("type"=>"town", "pop"=>827, "name_ru"=>"Ла Фонтейн", "name_en"=>"La Fontaine", "lang"=>"en", "lat"=>40.673693, "lon"=>-85.721745),
"/US/US-IN/la-paz-6113" => array("type"=>"town", "pop"=>477, "name_ru"=>"Ла Паз", "name_en"=>"La Paz", "lang"=>"en", "lat"=>41.459389, "lon"=>-86.309986),
"/US/US-IN/la-porte-6115" => array("type"=>"city", "pop"=>21174, "name_ru"=>"Ла Порт", "name_en"=>"La Porte", "lang"=>"en", "lat"=>41.607482, "lon"=>-86.714766),
"/US/US-IN/laconia-6100" => array("type"=>"town", "pop"=>30, "name_ru"=>"Лакония", "name_en"=>"Laconia", "lang"=>"en", "lat"=>38.032103, "lon"=>-86.085304),
"/US/US-IN/ladoga-6102" => array("type"=>"town", "pop"=>1015, "name_ru"=>"Ладога", "name_en"=>"Ladoga", "lang"=>"en", "lat"=>39.916751, "lon"=>-86.798673),
"/US/US-IN/lafayette-6103" => array("type"=>"city", "pop"=>64049, "name_ru"=>"Ла Файет", "name_en"=>"Lafayette", "lang"=>"en", "lat"=>40.410850, "lon"=>-86.870685),
"/US/US-IN/lagrange-6105" => array("type"=>"town", "pop"=>2929, "name_ru"=>"Лаграндж", "name_en"=>"Lagrange", "lang"=>"en", "lat"=>41.643426, "lon"=>-85.417238),
"/US/US-IN/lagro-6106" => array("type"=>"town", "pop"=>417, "name_ru"=>"Лагро", "name_en"=>"Lagro", "lang"=>"en", "lat"=>40.838283, "lon"=>-85.728359),
"/US/US-IN/lake-dalecarlia-6107" => array("type"=>"cdp", "pop"=>1276, "name_ru"=>"Лейк Дейлкарлия", "name_en"=>"Lake Dalecarlia", "lang"=>"en", "lat"=>41.339087, "lon"=>-87.403323),
"/US/US-IN/lake-hart-6108" => array("type"=>"town", "pop"=>213, "name_ru"=>"Лейк Харт", "name_en"=>"Lake Hart", "lang"=>"en", "lat"=>39.568376, "lon"=>-86.430097),
"/US/US-IN/lake-station-6110" => array("type"=>"city", "pop"=>13246, "name_ru"=>"Лейк Стейшен", "name_en"=>"Lake Station", "lang"=>"en", "lat"=>41.572776, "lon"=>-87.261394),
"/US/US-IN/lakes-of-the-four-seasons-6109" => array("type"=>"cdp", "pop"=>6556, "name_ru"=>"Лейкс оф зе Фор Сизонс", "name_en"=>"Lakes of the Four Seasons", "lang"=>"en", "lat"=>41.408597, "lon"=>-87.222994),
"/US/US-IN/lakeville-6111" => array("type"=>"town", "pop"=>549, "name_ru"=>"Лейквилл", "name_en"=>"Lakeville", "lang"=>"en", "lat"=>41.525760, "lon"=>-86.274494),
"/US/US-IN/lanesville-6112" => array("type"=>"town", "pop"=>627, "name_ru"=>"Лейнсвилл", "name_en"=>"Lanesville", "lang"=>"en", "lat"=>38.238298, "lon"=>-85.984826),
"/US/US-IN/lapel-6114" => array("type"=>"town", "pop"=>1845, "name_ru"=>"Лапел", "name_en"=>"Lapel", "lang"=>"en", "lat"=>40.068850, "lon"=>-85.847549),
"/US/US-IN/larwill-6116" => array("type"=>"town", "pop"=>279, "name_ru"=>"Ларвилл", "name_en"=>"Larwill", "lang"=>"en", "lat"=>41.178887, "lon"=>-85.624374),
"/US/US-IN/laurel-6117" => array("type"=>"town", "pop"=>577, "name_ru"=>"Лорел", "name_en"=>"Laurel", "lang"=>"en", "lat"=>39.501586, "lon"=>-85.188355),
"/US/US-IN/lawrence-6118" => array("type"=>"city", "pop"=>43247, "name_ru"=>"Лоренс", "name_en"=>"Lawrence", "lang"=>"en", "lat"=>39.868556, "lon"=>-85.987189),
"/US/US-IN/lawrenceburg-6119" => array("type"=>"city", "pop"=>4791, "name_ru"=>"Лоренсбург", "name_en"=>"Lawrenceburg", "lang"=>"en", "lat"=>39.100361, "lon"=>-84.869679),
"/US/US-IN/leavenworth-6120" => array("type"=>"town", "pop"=>340, "name_ru"=>"Ливенуорт", "name_en"=>"Leavenworth", "lang"=>"en", "lat"=>38.200408, "lon"=>-86.346582),
"/US/US-IN/lebanon-6121" => array("type"=>"city", "pop"=>15400, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>40.051641, "lon"=>-86.473459),
"/US/US-IN/leesburg-6122" => array("type"=>"town", "pop"=>598, "name_ru"=>"Лисберг", "name_en"=>"Leesburg", "lang"=>"en", "lat"=>41.329533, "lon"=>-85.848559),
"/US/US-IN/lewisville-6123" => array("type"=>"town", "pop"=>339, "name_ru"=>"Льюисвилл", "name_en"=>"Lewisville", "lang"=>"en", "lat"=>39.805852, "lon"=>-85.353181),
"/US/US-IN/liberty-6124" => array("type"=>"town", "pop"=>1914, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>39.634466, "lon"=>-84.926969),
"/US/US-IN/ligonier-6125" => array("type"=>"city", "pop"=>4530, "name_ru"=>"Лигоньер", "name_en"=>"Ligonier", "lang"=>"en", "lat"=>41.464117, "lon"=>-85.594740),
"/US/US-IN/linden-6126" => array("type"=>"town", "pop"=>680, "name_ru"=>"Линден", "name_en"=>"Linden", "lang"=>"en", "lat"=>40.190677, "lon"=>-86.901964),
"/US/US-IN/linton-6127" => array("type"=>"city", "pop"=>5696, "name_ru"=>"Линтон", "name_en"=>"Linton", "lang"=>"en", "lat"=>39.035985, "lon"=>-87.159176),
"/US/US-IN/little-york-6128" => array("type"=>"town", "pop"=>186, "name_ru"=>"Литл Йорк", "name_en"=>"Little York", "lang"=>"en", "lat"=>38.699665, "lon"=>-85.903947),
"/US/US-IN/livonia-6129" => array("type"=>"town", "pop"=>110, "name_ru"=>"Ливония", "name_en"=>"Livonia", "lang"=>"en", "lat"=>38.557144, "lon"=>-86.278911),
"/US/US-IN/lizton-6130" => array("type"=>"town", "pop"=>360, "name_ru"=>"Лизтон", "name_en"=>"Lizton", "lang"=>"en", "lat"=>39.885657, "lon"=>-86.542506),
"/US/US-IN/logansport-6131" => array("type"=>"city", "pop"=>18663, "name_ru"=>"Логанспорт", "name_en"=>"Logansport", "lang"=>"en", "lat"=>40.753703, "lon"=>-86.357417),
"/US/US-IN/long-beach-6132" => array("type"=>"town", "pop"=>1547, "name_ru"=>"Лонг Бич", "name_en"=>"Long Beach", "lang"=>"en", "lat"=>41.746567, "lon"=>-86.852134),
"/US/US-IN/loogootee-6133" => array("type"=>"city", "pop"=>2584, "name_ru"=>"Лоогути", "name_en"=>"Loogootee", "lang"=>"en", "lat"=>38.675428, "lon"=>-86.914302),
"/US/US-IN/losantville-6134" => array("type"=>"town", "pop"=>258, "name_ru"=>"Лозантвилл", "name_en"=>"Losantville", "lang"=>"en", "lat"=>40.023075, "lon"=>-85.183082),
"/US/US-IN/lowell-6135" => array("type"=>"town", "pop"=>8405, "name_ru"=>"Лоуэлл", "name_en"=>"Lowell", "lang"=>"en", "lat"=>41.294726, "lon"=>-87.413239),
"/US/US-IN/lynn-6136" => array("type"=>"town", "pop"=>1051, "name_ru"=>"Линн", "name_en"=>"Lynn", "lang"=>"en", "lat"=>40.048676, "lon"=>-84.942260),
"/US/US-IN/lynnville-6137" => array("type"=>"town", "pop"=>810, "name_ru"=>"Линнвилл", "name_en"=>"Lynnville", "lang"=>"en", "lat"=>38.198929, "lon"=>-87.317378),
"/US/US-IN/lyons-6138" => array("type"=>"town", "pop"=>733, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>38.988069, "lon"=>-87.081627),
"/US/US-IN/mackey-6140" => array("type"=>"town", "pop"=>137, "name_ru"=>"Макей", "name_en"=>"Mackey", "lang"=>"en", "lat"=>38.249932, "lon"=>-87.390699),
"/US/US-IN/macy-6141" => array("type"=>"town", "pop"=>228, "name_ru"=>"Мейси", "name_en"=>"Macy", "lang"=>"en", "lat"=>40.957728, "lon"=>-86.128908),
"/US/US-IN/madison-6142" => array("type"=>"city", "pop"=>12716, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>38.755526, "lon"=>-85.399044),
"/US/US-IN/marengo-6143" => array("type"=>"town", "pop"=>816, "name_ru"=>"Маренго", "name_en"=>"Marengo", "lang"=>"en", "lat"=>38.371606, "lon"=>-86.343702),
"/US/US-IN/marion-6144" => array("type"=>"city", "pop"=>30212, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>40.549100, "lon"=>-85.666990),
"/US/US-IN/markle-6145" => array("type"=>"town", "pop"=>1087, "name_ru"=>"Маркл", "name_en"=>"Markle", "lang"=>"en", "lat"=>40.826431, "lon"=>-85.338104),
"/US/US-IN/markleville-6146" => array("type"=>"town", "pop"=>389, "name_ru"=>"Марклвилл", "name_en"=>"Markleville", "lang"=>"en", "lat"=>39.977228, "lon"=>-85.616120),
"/US/US-IN/marshall-6147" => array("type"=>"town", "pop"=>348, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>39.846960, "lon"=>-87.185785),
"/US/US-IN/martinsville-6148" => array("type"=>"city", "pop"=>11800, "name_ru"=>"Мартинсвилл", "name_en"=>"Martinsville", "lang"=>"en", "lat"=>39.422845, "lon"=>-86.421661),
"/US/US-IN/matthews-6149" => array("type"=>"town", "pop"=>535, "name_ru"=>"Мэттьюс", "name_en"=>"Matthews", "lang"=>"en", "lat"=>40.387306, "lon"=>-85.498169),
"/US/US-IN/mauckport-6150" => array("type"=>"town", "pop"=>81, "name_ru"=>"Маакпорт", "name_en"=>"Mauckport", "lang"=>"en", "lat"=>38.024313, "lon"=>-86.201638),
"/US/US-IN/mccordsville-6139" => array("type"=>"town", "pop"=>1306, "name_ru"=>"Мак-Кордсвилл", "name_en"=>"McCordsville", "lang"=>"en", "lat"=>39.908143, "lon"=>-85.931214),
"/US/US-IN/mecca-6151" => array("type"=>"town", "pop"=>347, "name_ru"=>"Мекка", "name_en"=>"Mecca", "lang"=>"en", "lat"=>39.727132, "lon"=>-87.331490),
"/US/US-IN/medaryville-6152" => array("type"=>"town", "pop"=>541, "name_ru"=>"Медаривилл", "name_en"=>"Medaryville", "lang"=>"en", "lat"=>41.080259, "lon"=>-86.890021),
"/US/US-IN/medora-6153" => array("type"=>"town", "pop"=>538, "name_ru"=>"Медора", "name_en"=>"Medora", "lang"=>"en", "lat"=>38.824612, "lon"=>-86.170697),
"/US/US-IN/mellott-6154" => array("type"=>"town", "pop"=>202, "name_ru"=>"Меллотт", "name_en"=>"Mellott", "lang"=>"en", "lat"=>40.163894, "lon"=>-87.149186),
"/US/US-IN/melody-hill-6155" => array("type"=>"cdp", "pop"=>2932, "name_ru"=>"Мелоди Хилл", "name_en"=>"Melody Hill", "lang"=>"en", "lat"=>38.024180, "lon"=>-87.513180),
"/US/US-IN/mentone-6156" => array("type"=>"town", "pop"=>937, "name_ru"=>"Ментон", "name_en"=>"Mentone", "lang"=>"en", "lat"=>41.173764, "lon"=>-86.036930),
"/US/US-IN/meridian-hills-6157" => array("type"=>"town", "pop"=>1743, "name_ru"=>"Меридиан Хилл", "name_en"=>"Meridian Hills", "lang"=>"en", "lat"=>39.886350, "lon"=>-86.156455),
"/US/US-IN/merom-6158" => array("type"=>"town", "pop"=>287, "name_ru"=>"Мером", "name_en"=>"Merom", "lang"=>"en", "lat"=>39.056875, "lon"=>-87.567849),
"/US/US-IN/merrillville-6159" => array("type"=>"town", "pop"=>33057, "name_ru"=>"Мерриллвилл", "name_en"=>"Merrillville", "lang"=>"en", "lat"=>41.471583, "lon"=>-87.325828),
"/US/US-IN/mexico-6160" => array("type"=>"cdp", "pop"=>1003, "name_ru"=>"Мексико", "name_en"=>"Mexico", "lang"=>"en", "lat"=>40.812920, "lon"=>-86.110866),
"/US/US-IN/michiana-shores-6161" => array("type"=>"town", "pop"=>333, "name_ru"=>"Мичиана Шорс", "name_en"=>"Michiana Shores", "lang"=>"en", "lat"=>41.756250, "lon"=>-86.818419),
"/US/US-IN/michigan-city-6162" => array("type"=>"city", "pop"=>32405, "name_ru"=>"Мичиган Сити", "name_en"=>"Michigan City", "lang"=>"en", "lat"=>41.712550, "lon"=>-86.876026),
"/US/US-IN/michigantown-6163" => array("type"=>"town", "pop"=>403, "name_ru"=>"Мичигантаун", "name_en"=>"Michigantown", "lang"=>"en", "lat"=>40.327794, "lon"=>-86.391013),
"/US/US-IN/middlebury-6164" => array("type"=>"town", "pop"=>3191, "name_ru"=>"Миддлбери", "name_en"=>"Middlebury", "lang"=>"en", "lat"=>41.673107, "lon"=>-85.709635),
"/US/US-IN/middletown-6165" => array("type"=>"town", "pop"=>2354, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>40.059500, "lon"=>-85.542749),
"/US/US-IN/milan-6166" => array("type"=>"town", "pop"=>1760, "name_ru"=>"Милан", "name_en"=>"Milan", "lang"=>"en", "lat"=>39.125257, "lon"=>-85.130705),
"/US/US-IN/milford-6167" => array("type"=>"town", "pop"=>1513, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>39.350126, "lon"=>-85.619454),
"/US/US-IN/milford-6168" => array("type"=>"town", "pop"=>1388, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>41.410380, "lon"=>-85.848349),
"/US/US-IN/millersburg-6169" => array("type"=>"town", "pop"=>911, "name_ru"=>"Миллерсберг", "name_en"=>"Millersburg", "lang"=>"en", "lat"=>41.526293, "lon"=>-85.695609),
"/US/US-IN/millhousen-6170" => array("type"=>"town", "pop"=>130, "name_ru"=>"Миллхаусен", "name_en"=>"Millhousen", "lang"=>"en", "lat"=>39.210675, "lon"=>-85.435305),
"/US/US-IN/milltown-6171" => array("type"=>"town", "pop"=>899, "name_ru"=>"Миллтаун", "name_en"=>"Milltown", "lang"=>"en", "lat"=>38.342761, "lon"=>-86.274460),
"/US/US-IN/milton-6172" => array("type"=>"town", "pop"=>589, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>39.786533, "lon"=>-85.156191),
"/US/US-IN/mishawaka-6173" => array("type"=>"city", "pop"=>50026, "name_ru"=>"Мишока", "name_en"=>"Mishawaka", "lang"=>"en", "lat"=>41.669200, "lon"=>-86.168474),
"/US/US-IN/mitchell-6174" => array("type"=>"city", "pop"=>4567, "name_ru"=>"Митчелл", "name_en"=>"Mitchell", "lang"=>"en", "lat"=>38.736570, "lon"=>-86.475495),
"/US/US-IN/modoc-6175" => array("type"=>"town", "pop"=>205, "name_ru"=>"Модок", "name_en"=>"Modoc", "lang"=>"en", "lat"=>40.044719, "lon"=>-85.126181),
"/US/US-IN/monon-6176" => array("type"=>"town", "pop"=>1596, "name_ru"=>"Монон", "name_en"=>"Monon", "lang"=>"en", "lat"=>40.863632, "lon"=>-86.878713),
"/US/US-IN/monroe-6177" => array("type"=>"town", "pop"=>771, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>40.744509, "lon"=>-84.940540),
"/US/US-IN/monroe-city-6178" => array("type"=>"town", "pop"=>524, "name_ru"=>"Монро Сити", "name_en"=>"Monroe City", "lang"=>"en", "lat"=>38.614151, "lon"=>-87.353466),
"/US/US-IN/monroeville-6179" => array("type"=>"town", "pop"=>1273, "name_ru"=>"Монровилл", "name_en"=>"Monroeville", "lang"=>"en", "lat"=>40.972225, "lon"=>-84.867629),
"/US/US-IN/monterey-6180" => array("type"=>"town", "pop"=>222, "name_ru"=>"Монтерей", "name_en"=>"Monterey", "lang"=>"en", "lat"=>41.156232, "lon"=>-86.481742),
"/US/US-IN/montezuma-6181" => array("type"=>"town", "pop"=>1097, "name_ru"=>"Мотезума", "name_en"=>"Montezuma", "lang"=>"en", "lat"=>39.791001, "lon"=>-87.369109),
"/US/US-IN/montgomery-6182" => array("type"=>"town", "pop"=>373, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>38.665216, "lon"=>-87.047121),
"/US/US-IN/monticello-6183" => array("type"=>"city", "pop"=>5293, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>40.746704, "lon"=>-86.763948),
"/US/US-IN/montpelier-6184" => array("type"=>"city", "pop"=>1759, "name_ru"=>"Монтпильер", "name_en"=>"Montpelier", "lang"=>"en", "lat"=>40.552858, "lon"=>-85.281747),
"/US/US-IN/mooreland-6185" => array("type"=>"town", "pop"=>388, "name_ru"=>"Мурленд", "name_en"=>"Mooreland", "lang"=>"en", "lat"=>39.997745, "lon"=>-85.251701),
"/US/US-IN/moores-hill-6186" => array("type"=>"town", "pop"=>649, "name_ru"=>"Мурс Хилл", "name_en"=>"Moores Hill", "lang"=>"en", "lat"=>39.113391, "lon"=>-85.089042),
"/US/US-IN/mooresville-6187" => array("type"=>"town", "pop"=>11763, "name_ru"=>"Мурсвилл", "name_en"=>"Mooresville", "lang"=>"en", "lat"=>39.615888, "lon"=>-86.370072),
"/US/US-IN/morgantown-6188" => array("type"=>"town", "pop"=>961, "name_ru"=>"Моргантаун", "name_en"=>"Morgantown", "lang"=>"en", "lat"=>39.373868, "lon"=>-86.259290),
"/US/US-IN/morocco-6189" => array("type"=>"town", "pop"=>1112, "name_ru"=>"Морокко", "name_en"=>"Morocco", "lang"=>"en", "lat"=>40.945605, "lon"=>-87.452692),
"/US/US-IN/morristown-6190" => array("type"=>"town", "pop"=>1320, "name_ru"=>"Морристаун", "name_en"=>"Morristown", "lang"=>"en", "lat"=>39.672322, "lon"=>-85.699787),
"/US/US-IN/mount-auburn-6191" => array("type"=>"town", "pop"=>72, "name_ru"=>"Маунт Обурн", "name_en"=>"Mount Auburn", "lang"=>"en", "lat"=>39.812904, "lon"=>-85.190300),
"/US/US-IN/mount-ayr-6192" => array("type"=>"town", "pop"=>130, "name_ru"=>"Маунт Эр", "name_en"=>"Mount Ayr", "lang"=>"en", "lat"=>40.951520, "lon"=>-87.298330),
"/US/US-IN/mount-carmel-6193" => array("type"=>"town", "pop"=>106, "name_ru"=>"Маунт Кармел", "name_en"=>"Mount Carmel", "lang"=>"en", "lat"=>39.407460, "lon"=>-84.875733),
"/US/US-IN/mount-etna-6194" => array("type"=>"town", "pop"=>104, "name_ru"=>"Маунт Этна", "name_en"=>"Mount Etna", "lang"=>"en", "lat"=>40.741750, "lon"=>-85.562513),
"/US/US-IN/mount-summit-6195" => array("type"=>"town", "pop"=>300, "name_ru"=>"Маунт Саммит", "name_en"=>"Mount Summit", "lang"=>"en", "lat"=>40.003245, "lon"=>-85.386732),
"/US/US-IN/mount-vernon-6196" => array("type"=>"city", "pop"=>6980, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>37.936011, "lon"=>-87.898709),
"/US/US-IN/mulberry-6197" => array("type"=>"town", "pop"=>1453, "name_ru"=>"Малберри", "name_en"=>"Mulberry", "lang"=>"en", "lat"=>40.345593, "lon"=>-86.668053),
"/US/US-IN/muncie-6198" => array("type"=>"city", "pop"=>64975, "name_ru"=>"Манси", "name_en"=>"Muncie", "lang"=>"en", "lat"=>40.197450, "lon"=>-85.390635),
"/US/US-IN/munster-6199" => array("type"=>"town", "pop"=>22105, "name_ru"=>"Манстер", "name_en"=>"Munster", "lang"=>"en", "lat"=>41.546700, "lon"=>-87.503864),
"/US/US-IN/napoleon-6200" => array("type"=>"town", "pop"=>229, "name_ru"=>"Наполеон", "name_en"=>"Napoleon", "lang"=>"en", "lat"=>39.205353, "lon"=>-85.327278),
"/US/US-IN/nappanee-6201" => array("type"=>"city", "pop"=>7153, "name_ru"=>"Наппании", "name_en"=>"Nappanee", "lang"=>"en", "lat"=>41.445973, "lon"=>-85.984578),
"/US/US-IN/nashville-6202" => array("type"=>"town", "pop"=>777, "name_ru"=>"Нэшвилл", "name_en"=>"Nashville", "lang"=>"en", "lat"=>39.203807, "lon"=>-86.236002),
"/US/US-IN/new-albany-6203" => array("type"=>"city", "pop"=>37296, "name_ru"=>"Нью Олбани", "name_en"=>"New Albany", "lang"=>"en", "lat"=>38.303987, "lon"=>-85.825814),
"/US/US-IN/new-amsterdam-6204" => array("type"=>"town", "pop"=>1, "name_ru"=>"Нью Амстердам", "name_en"=>"New Amsterdam", "lang"=>"en", "lat"=>38.102816, "lon"=>-86.276851),
"/US/US-IN/new-carlisle-6207" => array("type"=>"town", "pop"=>1740, "name_ru"=>"Нью Карлайл", "name_en"=>"New Carlisle", "lang"=>"en", "lat"=>41.704446, "lon"=>-86.510869),
"/US/US-IN/new-castle-6208" => array("type"=>"city", "pop"=>18339, "name_ru"=>"Нью Касл", "name_en"=>"New Castle", "lang"=>"en", "lat"=>39.920905, "lon"=>-85.364926),
"/US/US-IN/new-chicago-6209" => array("type"=>"town", "pop"=>1992, "name_ru"=>"Нью Чикаго", "name_en"=>"New Chicago", "lang"=>"en", "lat"=>41.558736, "lon"=>-87.272219),
"/US/US-IN/new-harmony-6210" => array("type"=>"town", "pop"=>865, "name_ru"=>"Нью Хармони", "name_en"=>"New Harmony", "lang"=>"en", "lat"=>38.128527, "lon"=>-87.932478),
"/US/US-IN/new-haven-6211" => array("type"=>"city", "pop"=>13685, "name_ru"=>"Нью Хейвен", "name_en"=>"New Haven", "lang"=>"en", "lat"=>41.069000, "lon"=>-85.025136),
"/US/US-IN/new-market-6212" => array("type"=>"town", "pop"=>642, "name_ru"=>"Нью Маркет", "name_en"=>"New Market", "lang"=>"en", "lat"=>39.951911, "lon"=>-86.922126),
"/US/US-IN/new-middletown-6213" => array("type"=>"town", "pop"=>78, "name_ru"=>"Нью Миддлтаун", "name_en"=>"New Middletown", "lang"=>"en", "lat"=>38.163424, "lon"=>-86.050147),
"/US/US-IN/new-palestine-6214" => array("type"=>"town", "pop"=>2091, "name_ru"=>"Нью Палестин", "name_en"=>"New Palestine", "lang"=>"en", "lat"=>39.721088, "lon"=>-85.891579),
"/US/US-IN/new-paris-6215" => array("type"=>"cdp", "pop"=>1007, "name_ru"=>"Нью Парис", "name_en"=>"New Paris", "lang"=>"en", "lat"=>41.504848, "lon"=>-85.826487),
"/US/US-IN/new-pekin-6216" => array("type"=>"town", "pop"=>1348, "name_ru"=>"Нью Пекин", "name_en"=>"New Pekin", "lang"=>"en", "lat"=>38.502532, "lon"=>-86.015393),
"/US/US-IN/new-providence-6219" => array("type"=>"town", "pop"=>270, "name_ru"=>"Нью Провиденс", "name_en"=>"New Providence", "lang"=>"en", "lat"=>38.457912, "lon"=>-85.920833),
"/US/US-IN/new-richmond-6220" => array("type"=>"town", "pop"=>337, "name_ru"=>"Нью Ричмонд", "name_en"=>"New Richmond", "lang"=>"en", "lat"=>40.193942, "lon"=>-86.978572),
"/US/US-IN/new-ross-6221" => array("type"=>"town", "pop"=>323, "name_ru"=>"Нью Росс", "name_en"=>"New Ross", "lang"=>"en", "lat"=>39.963863, "lon"=>-86.713665),
"/US/US-IN/new-whiteland-6223" => array("type"=>"town", "pop"=>5809, "name_ru"=>"Нью Уайтленд", "name_en"=>"New Whiteland", "lang"=>"en", "lat"=>39.562144, "lon"=>-86.096916),
"/US/US-IN/newberry-6205" => array("type"=>"town", "pop"=>201, "name_ru"=>"Ньюберри", "name_en"=>"Newberry", "lang"=>"en", "lat"=>38.923706, "lon"=>-87.018922),
"/US/US-IN/newburgh-6206" => array("type"=>"town", "pop"=>3312, "name_ru"=>"Ньюберг", "name_en"=>"Newburgh", "lang"=>"en", "lat"=>37.946536, "lon"=>-87.403727),
"/US/US-IN/newpoint-6217" => array("type"=>"town", "pop"=>296, "name_ru"=>"Ньюпойнт", "name_en"=>"Newpoint", "lang"=>"en", "lat"=>39.308897, "lon"=>-85.330099),
"/US/US-IN/newport-6218" => array("type"=>"town", "pop"=>542, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>39.884831, "lon"=>-87.406286),
"/US/US-IN/newtown-6222" => array("type"=>"town", "pop"=>155, "name_ru"=>"Ньютаун", "name_en"=>"Newtown", "lang"=>"en", "lat"=>40.204421, "lon"=>-87.147895),
"/US/US-IN/noblesville-6224" => array("type"=>"city", "pop"=>42155, "name_ru"=>"Ноублсвилл", "name_en"=>"Noblesville", "lang"=>"en", "lat"=>40.056634, "lon"=>-86.031692),
"/US/US-IN/north-crows-nest-6225" => array("type"=>"town", "pop"=>44, "name_ru"=>"Норт Краус Нест", "name_en"=>"North Crows Nest", "lang"=>"en", "lat"=>39.865096, "lon"=>-86.162035),
"/US/US-IN/north-judson-6226" => array("type"=>"town", "pop"=>1884, "name_ru"=>"Норт Джадсон", "name_en"=>"North Judson", "lang"=>"en", "lat"=>41.217038, "lon"=>-86.775914),
"/US/US-IN/north-liberty-6227" => array("type"=>"town", "pop"=>1404, "name_ru"=>"Норт Либерти", "name_en"=>"North Liberty", "lang"=>"en", "lat"=>41.532984, "lon"=>-86.430370),
"/US/US-IN/north-manchester-6228" => array("type"=>"town", "pop"=>5829, "name_ru"=>"Норт Манчестер", "name_en"=>"North Manchester", "lang"=>"en", "lat"=>41.004892, "lon"=>-85.774132),
"/US/US-IN/north-salem-6229" => array("type"=>"town", "pop"=>637, "name_ru"=>"Норт Салем", "name_en"=>"North Salem", "lang"=>"en", "lat"=>39.859356, "lon"=>-86.644493),
"/US/US-IN/north-terre-haute-6230" => array("type"=>"cdp", "pop"=>4331, "name_ru"=>"Норт Терр Хаут", "name_en"=>"North Terre Haute", "lang"=>"en", "lat"=>39.539933, "lon"=>-87.360166),
"/US/US-IN/north-vernon-6231" => array("type"=>"city", "pop"=>6296, "name_ru"=>"Норт Вернон", "name_en"=>"North Vernon", "lang"=>"en", "lat"=>39.007807, "lon"=>-85.631230),
"/US/US-IN/north-webster-6232" => array("type"=>"town", "pop"=>1035, "name_ru"=>"Норт Вэбстер", "name_en"=>"North Webster", "lang"=>"en", "lat"=>41.325305, "lon"=>-85.697179),
"/US/US-IN/oak-park-6234" => array("type"=>"cdp", "pop"=>5630, "name_ru"=>"Оук Парк", "name_en"=>"Oak Park", "lang"=>"en", "lat"=>38.304100, "lon"=>-85.690666),
"/US/US-IN/oakland-city-6233" => array("type"=>"city", "pop"=>2512, "name_ru"=>"Оукленд Сити", "name_en"=>"Oakland City", "lang"=>"en", "lat"=>38.340039, "lon"=>-87.348023),
"/US/US-IN/oaktown-6235" => array("type"=>"town", "pop"=>596, "name_ru"=>"Оуктаун", "name_en"=>"Oaktown", "lang"=>"en", "lat"=>38.871652, "lon"=>-87.441270),
"/US/US-IN/odon-6236" => array("type"=>"town", "pop"=>1393, "name_ru"=>"Одон", "name_en"=>"Odon", "lang"=>"en", "lat"=>38.842285, "lon"=>-86.988652),
"/US/US-IN/ogden-dunes-6237" => array("type"=>"town", "pop"=>1278, "name_ru"=>"Огден Дьюнс", "name_en"=>"Ogden Dunes", "lang"=>"en", "lat"=>41.622804, "lon"=>-87.193531),
"/US/US-IN/oldenburg-6238" => array("type"=>"town", "pop"=>644, "name_ru"=>"Олденбург", "name_en"=>"Oldenburg", "lang"=>"en", "lat"=>39.338463, "lon"=>-85.202998),
"/US/US-IN/onward-6239" => array("type"=>"town", "pop"=>79, "name_ru"=>"Онворд", "name_en"=>"Onward", "lang"=>"en", "lat"=>40.693968, "lon"=>-86.195390),
"/US/US-IN/oolitic-6240" => array("type"=>"town", "pop"=>1119, "name_ru"=>"Улитик", "name_en"=>"Oolitic", "lang"=>"en", "lat"=>38.891438, "lon"=>-86.526375),
"/US/US-IN/orestes-6241" => array("type"=>"town", "pop"=>321, "name_ru"=>"Орестес", "name_en"=>"Orestes", "lang"=>"en", "lat"=>40.270696, "lon"=>-85.725223),
"/US/US-IN/orland-6242" => array("type"=>"town", "pop"=>320, "name_ru"=>"Орленд", "name_en"=>"Orland", "lang"=>"en", "lat"=>41.731009, "lon"=>-85.171125),
"/US/US-IN/orleans-6243" => array("type"=>"town", "pop"=>2328, "name_ru"=>"Орлеанс", "name_en"=>"Orleans", "lang"=>"en", "lat"=>38.662020, "lon"=>-86.452886),
"/US/US-IN/osceola-6244" => array("type"=>"town", "pop"=>1784, "name_ru"=>"Осеола", "name_en"=>"Osceola", "lang"=>"en", "lat"=>41.665100, "lon"=>-86.078548),
"/US/US-IN/osgood-6245" => array("type"=>"town", "pop"=>1615, "name_ru"=>"Осгуд", "name_en"=>"Osgood", "lang"=>"en", "lat"=>39.128329, "lon"=>-85.291870),
"/US/US-IN/ossian-6246" => array("type"=>"town", "pop"=>2894, "name_ru"=>"Оссиан", "name_en"=>"Ossian", "lang"=>"en", "lat"=>40.878600, "lon"=>-85.168426),
"/US/US-IN/otterbein-6247" => array("type"=>"town", "pop"=>1228, "name_ru"=>"Оттербейн", "name_en"=>"Otterbein", "lang"=>"en", "lat"=>40.488280, "lon"=>-87.094135),
"/US/US-IN/owensville-6248" => array("type"=>"town", "pop"=>1282, "name_ru"=>"Оуэнсвилл", "name_en"=>"Owensville", "lang"=>"en", "lat"=>38.271724, "lon"=>-87.690627),
"/US/US-IN/oxford-6249" => array("type"=>"town", "pop"=>1166, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>40.521357, "lon"=>-87.248489),
"/US/US-IN/palmyra-6250" => array("type"=>"town", "pop"=>713, "name_ru"=>"Палмира", "name_en"=>"Palmyra", "lang"=>"en", "lat"=>38.408181, "lon"=>-86.114210),
"/US/US-IN/paoli-6251" => array("type"=>"town", "pop"=>3889, "name_ru"=>"Паоли", "name_en"=>"Paoli", "lang"=>"en", "lat"=>38.557626, "lon"=>-86.469222),
"/US/US-IN/paragon-6252" => array("type"=>"town", "pop"=>636, "name_ru"=>"Парагон", "name_en"=>"Paragon", "lang"=>"en", "lat"=>39.394855, "lon"=>-86.563052),
"/US/US-IN/parker-city-6253" => array("type"=>"town", "pop"=>1320, "name_ru"=>"Паркер Сити", "name_en"=>"Parker City", "lang"=>"en", "lat"=>40.189564, "lon"=>-85.203381),
"/US/US-IN/patoka-6254" => array("type"=>"town", "pop"=>760, "name_ru"=>"Патока", "name_en"=>"Patoka", "lang"=>"en", "lat"=>38.405756, "lon"=>-87.587502),
"/US/US-IN/patriot-6255" => array("type"=>"town", "pop"=>189, "name_ru"=>"Патриот", "name_en"=>"Patriot", "lang"=>"en", "lat"=>38.838153, "lon"=>-84.827816),
"/US/US-IN/pendleton-6256" => array("type"=>"town", "pop"=>4207, "name_ru"=>"Пендлтон", "name_en"=>"Pendleton", "lang"=>"en", "lat"=>40.000250, "lon"=>-85.742174),
"/US/US-IN/pennville-6257" => array("type"=>"town", "pop"=>713, "name_ru"=>"Пеннвилл", "name_en"=>"Pennville", "lang"=>"en", "lat"=>40.493279, "lon"=>-85.147444),
"/US/US-IN/perrysville-6258" => array("type"=>"town", "pop"=>490, "name_ru"=>"Перрисвилл", "name_en"=>"Perrysville", "lang"=>"en", "lat"=>40.053512, "lon"=>-87.435810),
"/US/US-IN/peru-6259" => array("type"=>"city", "pop"=>12301, "name_ru"=>"Перу", "name_en"=>"Peru", "lang"=>"en", "lat"=>40.753635, "lon"=>-86.068050),
"/US/US-IN/petersburg-6260" => array("type"=>"city", "pop"=>2447, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>38.491685, "lon"=>-87.281749),
"/US/US-IN/pierceton-6261" => array("type"=>"town", "pop"=>676, "name_ru"=>"Пирстон", "name_en"=>"Pierceton", "lang"=>"en", "lat"=>41.199560, "lon"=>-85.707625),
"/US/US-IN/pine-village-6262" => array("type"=>"town", "pop"=>242, "name_ru"=>"Пайн Виллидж", "name_en"=>"Pine Village", "lang"=>"en", "lat"=>40.449927, "lon"=>-87.252459),
"/US/US-IN/pittsboro-6263" => array("type"=>"town", "pop"=>2567, "name_ru"=>"Питтсборо", "name_en"=>"Pittsboro", "lang"=>"en", "lat"=>39.867295, "lon"=>-86.464767),
"/US/US-IN/plainfield-6264" => array("type"=>"town", "pop"=>27832, "name_ru"=>"Плейнфилд", "name_en"=>"Plainfield", "lang"=>"en", "lat"=>39.700437, "lon"=>-86.387942),
"/US/US-IN/plainville-6265" => array("type"=>"town", "pop"=>517, "name_ru"=>"Плейнвилл", "name_en"=>"Plainville", "lang"=>"en", "lat"=>38.804271, "lon"=>-87.151929),
"/US/US-IN/plymouth-6266" => array("type"=>"city", "pop"=>11038, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>41.347005, "lon"=>-86.314465),
"/US/US-IN/poneto-6267" => array("type"=>"town", "pop"=>232, "name_ru"=>"Понето", "name_en"=>"Poneto", "lang"=>"en", "lat"=>40.656127, "lon"=>-85.222222),
"/US/US-IN/portage-6268" => array("type"=>"city", "pop"=>36976, "name_ru"=>"Портейдж", "name_en"=>"Portage", "lang"=>"en", "lat"=>41.586699, "lon"=>-87.181522),
"/US/US-IN/porter-6269" => array("type"=>"town", "pop"=>5379, "name_ru"=>"Портер", "name_en"=>"Porter", "lang"=>"en", "lat"=>41.627296, "lon"=>-87.081908),
"/US/US-IN/portland-6270" => array("type"=>"city", "pop"=>6158, "name_ru"=>"Портленд", "name_en"=>"Portland", "lang"=>"en", "lat"=>40.433167, "lon"=>-84.981059),
"/US/US-IN/poseyville-6271" => array("type"=>"town", "pop"=>1100, "name_ru"=>"Посивилл", "name_en"=>"Poseyville", "lang"=>"en", "lat"=>38.170162, "lon"=>-87.782815),
"/US/US-IN/pottawattomie-park-6272" => array("type"=>"town", "pop"=>281, "name_ru"=>"Поттаваттами Парк", "name_en"=>"Pottawattomie Park", "lang"=>"en", "lat"=>41.722533, "lon"=>-86.867033),
"/US/US-IN/princes-lakes-6273" => array("type"=>"town", "pop"=>1612, "name_ru"=>"Принсес Лейкс", "name_en"=>"Princes Lakes", "lang"=>"en", "lat"=>39.350719, "lon"=>-86.109062),
"/US/US-IN/princeton-6274" => array("type"=>"city", "pop"=>8447, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>38.355678, "lon"=>-87.569208),
"/US/US-IN/redkey-6275" => array("type"=>"town", "pop"=>1396, "name_ru"=>"Редки", "name_en"=>"Redkey", "lang"=>"en", "lat"=>40.348003, "lon"=>-85.152670),
"/US/US-IN/remington-6276" => array("type"=>"town", "pop"=>1253, "name_ru"=>"Ремингтон", "name_en"=>"Remington", "lang"=>"en", "lat"=>40.762697, "lon"=>-87.151658),
"/US/US-IN/rensselaer-6277" => array("type"=>"city", "pop"=>6271, "name_ru"=>"Ренсселаер", "name_en"=>"Rensselaer", "lang"=>"en", "lat"=>40.935486, "lon"=>-87.149162),
"/US/US-IN/reynolds-6278" => array("type"=>"town", "pop"=>504, "name_ru"=>"Рейнолдс", "name_en"=>"Reynolds", "lang"=>"en", "lat"=>40.748718, "lon"=>-86.873657),
"/US/US-IN/richmond-6279" => array("type"=>"city", "pop"=>36733, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>39.829543, "lon"=>-84.885425),
"/US/US-IN/ridgeville-6280" => array("type"=>"town", "pop"=>779, "name_ru"=>"Риджвилл", "name_en"=>"Ridgeville", "lang"=>"en", "lat"=>40.289624, "lon"=>-85.029815),
"/US/US-IN/riley-6281" => array("type"=>"town", "pop"=>154, "name_ru"=>"Райли", "name_en"=>"Riley", "lang"=>"en", "lat"=>39.389800, "lon"=>-87.300403),
"/US/US-IN/rising-sun-6282" => array("type"=>"city", "pop"=>2342, "name_ru"=>"Райзинг Сайт", "name_en"=>"Rising Sun", "lang"=>"en", "lat"=>38.951626, "lon"=>-84.857110),
"/US/US-IN/river-forest-6283" => array("type"=>"town", "pop"=>27, "name_ru"=>"Ривер Форест", "name_en"=>"River Forest", "lang"=>"en", "lat"=>40.109467, "lon"=>-85.729383),
"/US/US-IN/roachdale-6284" => array("type"=>"town", "pop"=>947, "name_ru"=>"Роачдейл", "name_en"=>"Roachdale", "lang"=>"en", "lat"=>39.848779, "lon"=>-86.800408),
"/US/US-IN/roann-6285" => array("type"=>"town", "pop"=>384, "name_ru"=>"Роанн", "name_en"=>"Roann", "lang"=>"en", "lat"=>40.911170, "lon"=>-85.923888),
"/US/US-IN/roanoke-6286" => array("type"=>"town", "pop"=>1482, "name_ru"=>"Роанок", "name_en"=>"Roanoke", "lang"=>"en", "lat"=>40.964418, "lon"=>-85.375240),
"/US/US-IN/rochester-6287" => array("type"=>"city", "pop"=>6457, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>41.061459, "lon"=>-86.195480),
"/US/US-IN/rockport-6288" => array("type"=>"city", "pop"=>2045, "name_ru"=>"Рокпорт", "name_en"=>"Rockport", "lang"=>"en", "lat"=>37.883670, "lon"=>-87.050867),
"/US/US-IN/rockville-6289" => array("type"=>"town", "pop"=>2616, "name_ru"=>"Роквилл", "name_en"=>"Rockville", "lang"=>"en", "lat"=>39.766714, "lon"=>-87.230871),
"/US/US-IN/rocky-ripple-6290" => array("type"=>"town", "pop"=>707, "name_ru"=>"Роки Риппл", "name_en"=>"Rocky Ripple", "lang"=>"en", "lat"=>39.848900, "lon"=>-86.173199),
"/US/US-IN/rome-city-6291" => array("type"=>"town", "pop"=>1666, "name_ru"=>"Роум Сити", "name_en"=>"Rome City", "lang"=>"en", "lat"=>41.488348, "lon"=>-85.355812),
"/US/US-IN/rosedale-6292" => array("type"=>"town", "pop"=>724, "name_ru"=>"Роуздейл", "name_en"=>"Rosedale", "lang"=>"en", "lat"=>39.623150, "lon"=>-87.282099),
"/US/US-IN/roseland-6293" => array("type"=>"town", "pop"=>621, "name_ru"=>"Роузлэнд", "name_en"=>"Roseland", "lang"=>"en", "lat"=>41.716700, "lon"=>-86.251581),
"/US/US-IN/rossville-6294" => array("type"=>"town", "pop"=>1531, "name_ru"=>"Россвилл", "name_en"=>"Rossville", "lang"=>"en", "lat"=>40.421143, "lon"=>-86.595584),
"/US/US-IN/royal-center-6295" => array("type"=>"town", "pop"=>792, "name_ru"=>"Ройал Сити", "name_en"=>"Royal Center", "lang"=>"en", "lat"=>40.864624, "lon"=>-86.500512),
"/US/US-IN/rushville-6296" => array("type"=>"city", "pop"=>6068, "name_ru"=>"Рашвилл", "name_en"=>"Rushville", "lang"=>"en", "lat"=>39.614596, "lon"=>-85.447053),
"/US/US-IN/russellville-6297" => array("type"=>"town", "pop"=>337, "name_ru"=>"Расселлвилл", "name_en"=>"Russellville", "lang"=>"en", "lat"=>39.857147, "lon"=>-86.983445),
"/US/US-IN/russiaville-6298" => array("type"=>"town", "pop"=>1162, "name_ru"=>"Рашавилл", "name_en"=>"Russiaville", "lang"=>"en", "lat"=>40.418848, "lon"=>-86.270309),
"/US/US-IN/salamonia-6303" => array("type"=>"town", "pop"=>151, "name_ru"=>"Саламония", "name_en"=>"Salamonia", "lang"=>"en", "lat"=>40.382001, "lon"=>-84.866080),
"/US/US-IN/salem-6304" => array("type"=>"city", "pop"=>6502, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>38.604067, "lon"=>-86.098824),
"/US/US-IN/saltillo-6305" => array("type"=>"town", "pop"=>109, "name_ru"=>"Салтилло", "name_en"=>"Saltillo", "lang"=>"en", "lat"=>38.665718, "lon"=>-86.298328),
"/US/US-IN/sandborn-6306" => array("type"=>"town", "pop"=>429, "name_ru"=>"Сэндборн", "name_en"=>"Sandborn", "lang"=>"en", "lat"=>38.896819, "lon"=>-87.184600),
"/US/US-IN/santa-claus-6307" => array("type"=>"town", "pop"=>2314, "name_ru"=>"Санта Клос", "name_en"=>"Santa Claus", "lang"=>"en", "lat"=>38.115525, "lon"=>-86.925131),
"/US/US-IN/saratoga-6308" => array("type"=>"town", "pop"=>267, "name_ru"=>"Саратога", "name_en"=>"Saratoga", "lang"=>"en", "lat"=>40.236867, "lon"=>-84.915462),
"/US/US-IN/schererville-6309" => array("type"=>"town", "pop"=>28886, "name_ru"=>"Шерервилл", "name_en"=>"Schererville", "lang"=>"en", "lat"=>41.489200, "lon"=>-87.448183),
"/US/US-IN/schneider-6310" => array("type"=>"town", "pop"=>296, "name_ru"=>"Шнайдер", "name_en"=>"Schneider", "lang"=>"en", "lat"=>41.191948, "lon"=>-87.447656),
"/US/US-IN/scottsburg-6311" => array("type"=>"city", "pop"=>5912, "name_ru"=>"Скоттсбург", "name_en"=>"Scottsburg", "lang"=>"en", "lat"=>38.685861, "lon"=>-85.780920),
"/US/US-IN/seelyville-6312" => array("type"=>"town", "pop"=>1138, "name_ru"=>"Силивилл", "name_en"=>"Seelyville", "lang"=>"en", "lat"=>39.494000, "lon"=>-87.267049),
"/US/US-IN/sellersburg-6313" => array("type"=>"town", "pop"=>6346, "name_ru"=>"Селлерсбург", "name_en"=>"Sellersburg", "lang"=>"en", "lat"=>38.390872, "lon"=>-85.757161),
"/US/US-IN/selma-6314" => array("type"=>"town", "pop"=>835, "name_ru"=>"Сельма", "name_en"=>"Selma", "lang"=>"en", "lat"=>40.188395, "lon"=>-85.274315),
"/US/US-IN/seymour-6315" => array("type"=>"city", "pop"=>19223, "name_ru"=>"Сеймур", "name_en"=>"Seymour", "lang"=>"en", "lat"=>38.958469, "lon"=>-85.886388),
"/US/US-IN/shadeland-6316" => array("type"=>"town", "pop"=>1796, "name_ru"=>"Шейдленд", "name_en"=>"Shadeland", "lang"=>"en", "lat"=>40.344520, "lon"=>-86.963690),
"/US/US-IN/shamrock-lakes-6317" => array("type"=>"town", "pop"=>153, "name_ru"=>"Шэмрок Лейкс", "name_en"=>"Shamrock Lakes", "lang"=>"en", "lat"=>40.411008, "lon"=>-85.426224),
"/US/US-IN/sharpsville-6318" => array("type"=>"town", "pop"=>606, "name_ru"=>"Шарпсвилл", "name_en"=>"Sharpsville", "lang"=>"en", "lat"=>40.379949, "lon"=>-86.087290),
"/US/US-IN/shelburn-6319" => array("type"=>"town", "pop"=>1246, "name_ru"=>"Шелберн", "name_en"=>"Shelburn", "lang"=>"en", "lat"=>39.179465, "lon"=>-87.396881),
"/US/US-IN/shelbyville-6320" => array("type"=>"city", "pop"=>18516, "name_ru"=>"Шелбивилл", "name_en"=>"Shelbyville", "lang"=>"en", "lat"=>39.523264, "lon"=>-85.775817),
"/US/US-IN/sheridan-6321" => array("type"=>"town", "pop"=>3103, "name_ru"=>"Шеридан", "name_en"=>"Sheridan", "lang"=>"en", "lat"=>40.134699, "lon"=>-86.218863),
"/US/US-IN/shipshewana-6322" => array("type"=>"town", "pop"=>528, "name_ru"=>"Шипшевана", "name_en"=>"Shipshewana", "lang"=>"en", "lat"=>41.674866, "lon"=>-85.579326),
"/US/US-IN/shirley-6323" => array("type"=>"town", "pop"=>851, "name_ru"=>"Ширли", "name_en"=>"Shirley", "lang"=>"en", "lat"=>39.890472, "lon"=>-85.579848),
"/US/US-IN/shoals-6324" => array("type"=>"town", "pop"=>783, "name_ru"=>"Шоалс", "name_en"=>"Shoals", "lang"=>"en", "lat"=>38.667019, "lon"=>-86.793500),
"/US/US-IN/sidney-6325" => array("type"=>"town", "pop"=>155, "name_ru"=>"Сидни", "name_en"=>"Sidney", "lang"=>"en", "lat"=>41.105024, "lon"=>-85.742899),
"/US/US-IN/silver-lake-6326" => array("type"=>"town", "pop"=>968, "name_ru"=>"Сильвер Лэйк", "name_en"=>"Silver Lake", "lang"=>"en", "lat"=>41.072379, "lon"=>-85.891722),
"/US/US-IN/simonton-lake-6327" => array("type"=>"cdp", "pop"=>3554, "name_ru"=>"Симонтон Лейк", "name_en"=>"Simonton Lake", "lang"=>"en", "lat"=>41.747717, "lon"=>-85.968485),
"/US/US-IN/somerville-6328" => array("type"=>"town", "pop"=>303, "name_ru"=>"Сомервилл", "name_en"=>"Somerville", "lang"=>"en", "lat"=>38.277666, "lon"=>-87.376757),
"/US/US-IN/south-bend-6329" => array("type"=>"city", "pop"=>103807, "name_ru"=>"Саут Бенд", "name_en"=>"South Bend", "lang"=>"en", "lat"=>41.675300, "lon"=>-86.265699),
"/US/US-IN/south-haven-6330" => array("type"=>"cdp", "pop"=>6112, "name_ru"=>"Саут Хейвен", "name_en"=>"South Haven", "lang"=>"en", "lat"=>41.543584, "lon"=>-87.133907),
"/US/US-IN/south-whitley-6332" => array("type"=>"town", "pop"=>1849, "name_ru"=>"Саут Уайтли", "name_en"=>"South Whitley", "lang"=>"en", "lat"=>41.085664, "lon"=>-85.628639),
"/US/US-IN/southport-6331" => array("type"=>"city", "pop"=>1757, "name_ru"=>"Саутпорт", "name_en"=>"Southport", "lang"=>"en", "lat"=>39.659568, "lon"=>-86.116600),
"/US/US-IN/speedway-6333" => array("type"=>"town", "pop"=>12594, "name_ru"=>"Спидуэй", "name_en"=>"Speedway", "lang"=>"en", "lat"=>39.793450, "lon"=>-86.247424),
"/US/US-IN/spencer-6334" => array("type"=>"town", "pop"=>2466, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>39.287054, "lon"=>-86.769152),
"/US/US-IN/spiceland-6335" => array("type"=>"town", "pop"=>875, "name_ru"=>"Спайсленд", "name_en"=>"Spiceland", "lang"=>"en", "lat"=>39.838085, "lon"=>-85.438116),
"/US/US-IN/spring-grove-6336" => array("type"=>"town", "pop"=>369, "name_ru"=>"Спринг Гров", "name_en"=>"Spring Grove", "lang"=>"en", "lat"=>39.847234, "lon"=>-84.890473),
"/US/US-IN/spring-hill-6337" => array("type"=>"town", "pop"=>98, "name_ru"=>"Спринг Хилл", "name_en"=>"Spring Hill", "lang"=>"en", "lat"=>39.833435, "lon"=>-86.192661),
"/US/US-IN/spring-lake-6338" => array("type"=>"town", "pop"=>279, "name_ru"=>"Спринг Лейк", "name_en"=>"Spring Lake", "lang"=>"en", "lat"=>39.776781, "lon"=>-85.854368),
"/US/US-IN/springport-6339" => array("type"=>"town", "pop"=>158, "name_ru"=>"Спрингпорт", "name_en"=>"Springport", "lang"=>"en", "lat"=>40.046892, "lon"=>-85.392935),
"/US/US-IN/spurgeon-6340" => array("type"=>"town", "pop"=>213, "name_ru"=>"Сперджен", "name_en"=>"Spurgeon", "lang"=>"en", "lat"=>38.255607, "lon"=>-87.259359),
"/US/US-IN/st.-joe-6299" => array("type"=>"town", "pop"=>466, "name_ru"=>"Сент Джо", "name_en"=>"St. Joe", "lang"=>"en", "lat"=>41.314753, "lon"=>-84.903525),
"/US/US-IN/st.-john-6300" => array("type"=>"town", "pop"=>12999, "name_ru"=>"Сент Джон", "name_en"=>"St. John", "lang"=>"en", "lat"=>41.447024, "lon"=>-87.475848),
"/US/US-IN/st.-leon-6301" => array("type"=>"town", "pop"=>535, "name_ru"=>"Сент Леон", "name_en"=>"St. Leon", "lang"=>"en", "lat"=>39.287307, "lon"=>-84.967154),
"/US/US-IN/st.-paul-6302" => array("type"=>"town", "pop"=>990, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>39.427290, "lon"=>-85.629149),
"/US/US-IN/state-line-city-6341" => array("type"=>"town", "pop"=>135, "name_ru"=>"Стейт Лайн Сити", "name_en"=>"State Line City", "lang"=>"en", "lat"=>40.197197, "lon"=>-87.527246),
"/US/US-IN/staunton-6342" => array("type"=>"town", "pop"=>548, "name_ru"=>"Стонтон", "name_en"=>"Staunton", "lang"=>"en", "lat"=>39.486107, "lon"=>-87.189397),
"/US/US-IN/stilesville-6343" => array("type"=>"town", "pop"=>285, "name_ru"=>"Стилсвилл", "name_en"=>"Stilesville", "lang"=>"en", "lat"=>39.638263, "lon"=>-86.632263),
"/US/US-IN/stinesville-6344" => array("type"=>"town", "pop"=>197, "name_ru"=>"Стинсвилл", "name_en"=>"Stinesville", "lang"=>"en", "lat"=>39.300400, "lon"=>-86.649237),
"/US/US-IN/straughn-6345" => array("type"=>"town", "pop"=>233, "name_ru"=>"Строн", "name_en"=>"Straughn", "lang"=>"en", "lat"=>39.808294, "lon"=>-85.290429),
"/US/US-IN/sullivan-6346" => array("type"=>"city", "pop"=>4482, "name_ru"=>"Салливан", "name_en"=>"Sullivan", "lang"=>"en", "lat"=>39.096858, "lon"=>-87.407753),
"/US/US-IN/sulphur-springs-6347" => array("type"=>"town", "pop"=>322, "name_ru"=>"Салфер Спрингс", "name_en"=>"Sulphur Springs", "lang"=>"en", "lat"=>40.005671, "lon"=>-85.443416),
"/US/US-IN/summitville-6348" => array("type"=>"town", "pop"=>1043, "name_ru"=>"Саммитвилл", "name_en"=>"Summitville", "lang"=>"en", "lat"=>40.337321, "lon"=>-85.643010),
"/US/US-IN/sunman-6349" => array("type"=>"town", "pop"=>777, "name_ru"=>"Санмэн", "name_en"=>"Sunman", "lang"=>"en", "lat"=>39.233899, "lon"=>-85.096150),
"/US/US-IN/swayzee-6350" => array("type"=>"town", "pop"=>901, "name_ru"=>"Суэйзи", "name_en"=>"Swayzee", "lang"=>"en", "lat"=>40.506837, "lon"=>-85.824085),
"/US/US-IN/sweetser-6351" => array("type"=>"town", "pop"=>810, "name_ru"=>"Свитсер", "name_en"=>"Sweetser", "lang"=>"en", "lat"=>40.569692, "lon"=>-85.766992),
"/US/US-IN/switz-city-6352" => array("type"=>"town", "pop"=>303, "name_ru"=>"Свитс Сити", "name_en"=>"Switz City", "lang"=>"en", "lat"=>39.034714, "lon"=>-87.053609),
"/US/US-IN/syracuse-6353" => array("type"=>"town", "pop"=>2995, "name_ru"=>"Сиракьюс", "name_en"=>"Syracuse", "lang"=>"en", "lat"=>41.423213, "lon"=>-85.749049),
"/US/US-IN/taylorsville-6354" => array("type"=>"cdp", "pop"=>1044, "name_ru"=>"Тейлорсвилл", "name_en"=>"Taylorsville", "lang"=>"en", "lat"=>39.296233, "lon"=>-85.949227),
"/US/US-IN/tell-city-6355" => array("type"=>"city", "pop"=>7526, "name_ru"=>"Телл Сити", "name_en"=>"Tell City", "lang"=>"en", "lat"=>37.949946, "lon"=>-86.757419),
"/US/US-IN/tennyson-6356" => array("type"=>"town", "pop"=>297, "name_ru"=>"Теннисон", "name_en"=>"Tennyson", "lang"=>"en", "lat"=>38.081624, "lon"=>-87.118946),
"/US/US-IN/terre-haute-6357" => array("type"=>"city", "pop"=>60007, "name_ru"=>"Терр Хаут", "name_en"=>"Terre Haute", "lang"=>"en", "lat"=>39.470400, "lon"=>-87.383516),
"/US/US-IN/thorntown-6358" => array("type"=>"town", "pop"=>1590, "name_ru"=>"Торнтаун", "name_en"=>"Thorntown", "lang"=>"en", "lat"=>40.128728, "lon"=>-86.610078),
"/US/US-IN/tipton-6359" => array("type"=>"city", "pop"=>5006, "name_ru"=>"Типтон", "name_en"=>"Tipton", "lang"=>"en", "lat"=>40.282848, "lon"=>-86.040323),
"/US/US-IN/topeka-6360" => array("type"=>"town", "pop"=>1191, "name_ru"=>"Топека", "name_en"=>"Topeka", "lang"=>"en", "lat"=>41.538580, "lon"=>-85.542173),
"/US/US-IN/town-of-pines-6361" => array("type"=>"town", "pop"=>793, "name_ru"=>"Таун оф Пайнс", "name_en"=>"Town of Pines", "lang"=>"en", "lat"=>41.688551, "lon"=>-86.951660),
"/US/US-IN/trafalgar-6362" => array("type"=>"town", "pop"=>1125, "name_ru"=>"Трафалгар", "name_en"=>"Trafalgar", "lang"=>"en", "lat"=>39.414952, "lon"=>-86.149672),
"/US/US-IN/trail-creek-6363" => array("type"=>"town", "pop"=>2168, "name_ru"=>"Трейл Крик", "name_en"=>"Trail Creek", "lang"=>"en", "lat"=>41.696400, "lon"=>-86.856182),
"/US/US-IN/tri-lakes-6364" => array("type"=>"cdp", "pop"=>3299, "name_ru"=>"Три-Лейкс", "name_en"=>"Tri-Lakes", "lang"=>"en", "lat"=>41.221006, "lon"=>-85.484229),
"/US/US-IN/troy-6365" => array("type"=>"town", "pop"=>381, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>37.996066, "lon"=>-86.801540),
"/US/US-IN/ulen-6366" => array("type"=>"town", "pop"=>124, "name_ru"=>"Улен", "name_en"=>"Ulen", "lang"=>"en", "lat"=>40.064173, "lon"=>-86.463431),
"/US/US-IN/union-city-6367" => array("type"=>"city", "pop"=>3324, "name_ru"=>"Юнион Сити", "name_en"=>"Union City", "lang"=>"en", "lat"=>40.201608, "lon"=>-84.817480),
"/US/US-IN/uniondale-6368" => array("type"=>"town", "pop"=>268, "name_ru"=>"Юниондейл", "name_en"=>"Uniondale", "lang"=>"en", "lat"=>40.828450, "lon"=>-85.241358),
"/US/US-IN/universal-6369" => array("type"=>"town", "pop"=>409, "name_ru"=>"Юниверсал", "name_en"=>"Universal", "lang"=>"en", "lat"=>39.621600, "lon"=>-87.453024),
"/US/US-IN/upland-6370" => array("type"=>"town", "pop"=>3642, "name_ru"=>"Апленд", "name_en"=>"Upland", "lang"=>"en", "lat"=>40.462779, "lon"=>-85.513306),
"/US/US-IN/utica-6371" => array("type"=>"town", "pop"=>650, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>38.332792, "lon"=>-85.654865),
"/US/US-IN/valparaiso-6372" => array("type"=>"city", "pop"=>30429, "name_ru"=>"Валпарайсо", "name_en"=>"Valparaiso", "lang"=>"en", "lat"=>41.477312, "lon"=>-87.051004),
"/US/US-IN/van-buren-6373" => array("type"=>"town", "pop"=>835, "name_ru"=>"Ван Бурен", "name_en"=>"Van Buren", "lang"=>"en", "lat"=>40.616374, "lon"=>-85.505460),
"/US/US-IN/veedersburg-6374" => array("type"=>"town", "pop"=>2169, "name_ru"=>"Видерсбург", "name_en"=>"Veedersburg", "lang"=>"en", "lat"=>40.112726, "lon"=>-87.259194),
"/US/US-IN/vera-cruz-6375" => array("type"=>"town", "pop"=>53, "name_ru"=>"Вера Круз", "name_en"=>"Vera Cruz", "lang"=>"en", "lat"=>40.700855, "lon"=>-85.079582),
"/US/US-IN/vernon-6376" => array("type"=>"town", "pop"=>317, "name_ru"=>"Вернон", "name_en"=>"Vernon", "lang"=>"en", "lat"=>38.985132, "lon"=>-85.611090),
"/US/US-IN/versailles-6377" => array("type"=>"town", "pop"=>1713, "name_ru"=>"Версайллес", "name_en"=>"Versailles", "lang"=>"en", "lat"=>39.064344, "lon"=>-85.257538),
"/US/US-IN/vevay-6378" => array("type"=>"town", "pop"=>1604, "name_ru"=>"Вевэй", "name_en"=>"Vevay", "lang"=>"en", "lat"=>38.744357, "lon"=>-85.071159),
"/US/US-IN/vincennes-6379" => array("type"=>"city", "pop"=>17976, "name_ru"=>"Винсеннес", "name_en"=>"Vincennes", "lang"=>"en", "lat"=>38.676164, "lon"=>-87.511101),
"/US/US-IN/wabash-6380" => array("type"=>"city", "pop"=>10815, "name_ru"=>"Вабаш", "name_en"=>"Wabash", "lang"=>"en", "lat"=>40.799217, "lon"=>-85.825255),
"/US/US-IN/wakarusa-6381" => array("type"=>"town", "pop"=>1734, "name_ru"=>"Уэйкаруса", "name_en"=>"Wakarusa", "lang"=>"en", "lat"=>41.534544, "lon"=>-86.012750),
"/US/US-IN/walkerton-6382" => array("type"=>"town", "pop"=>2175, "name_ru"=>"Уолкертон", "name_en"=>"Walkerton", "lang"=>"en", "lat"=>41.464014, "lon"=>-86.482264),
"/US/US-IN/wallace-6383" => array("type"=>"town", "pop"=>95, "name_ru"=>"Уоллэс", "name_en"=>"Wallace", "lang"=>"en", "lat"=>39.987847, "lon"=>-87.147655),
"/US/US-IN/walton-6384" => array("type"=>"town", "pop"=>1007, "name_ru"=>"Уолтон", "name_en"=>"Walton", "lang"=>"en", "lat"=>40.662265, "lon"=>-86.244746),
"/US/US-IN/wanatah-6385" => array("type"=>"town", "pop"=>1059, "name_ru"=>"Ваната", "name_en"=>"Wanatah", "lang"=>"en", "lat"=>41.430752, "lon"=>-86.888618),
"/US/US-IN/warren-6386" => array("type"=>"town", "pop"=>1309, "name_ru"=>"Уоррен", "name_en"=>"Warren", "lang"=>"en", "lat"=>40.686674, "lon"=>-85.423367),
"/US/US-IN/warren-park-6387" => array("type"=>"town", "pop"=>1638, "name_ru"=>"Уоррен Парк", "name_en"=>"Warren Park", "lang"=>"en", "lat"=>39.782450, "lon"=>-86.052371),
"/US/US-IN/warsaw-6388" => array("type"=>"city", "pop"=>13627, "name_ru"=>"Варшава", "name_en"=>"Warsaw", "lang"=>"en", "lat"=>41.239594, "lon"=>-85.847571),
"/US/US-IN/washington-6389" => array("type"=>"city", "pop"=>11397, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>38.658171, "lon"=>-87.173005),
"/US/US-IN/waterloo-6390" => array("type"=>"town", "pop"=>2165, "name_ru"=>"Ватрелоо", "name_en"=>"Waterloo", "lang"=>"en", "lat"=>41.433369, "lon"=>-85.026601),
"/US/US-IN/waveland-6391" => array("type"=>"town", "pop"=>402, "name_ru"=>"Уейвленд", "name_en"=>"Waveland", "lang"=>"en", "lat"=>39.877029, "lon"=>-87.046299),
"/US/US-IN/waynetown-6392" => array("type"=>"town", "pop"=>879, "name_ru"=>"Уэйнтаун", "name_en"=>"Waynetown", "lang"=>"en", "lat"=>40.087518, "lon"=>-87.065554),
"/US/US-IN/west-baden-springs-6393" => array("type"=>"town", "pop"=>612, "name_ru"=>"Уэст Баден Спрингс", "name_en"=>"West Baden Springs", "lang"=>"en", "lat"=>38.567733, "lon"=>-86.612262),
"/US/US-IN/west-college-corner-6394" => array("type"=>"town", "pop"=>594, "name_ru"=>"Уэст Коллидж Корнер", "name_en"=>"West College Corner", "lang"=>"en", "lat"=>39.569600, "lon"=>-84.819299),
"/US/US-IN/west-harrison-6396" => array("type"=>"town", "pop"=>316, "name_ru"=>"Уэст Харрисон", "name_en"=>"West Harrison", "lang"=>"en", "lat"=>39.261229, "lon"=>-84.820918),
"/US/US-IN/west-lafayette-6397" => array("type"=>"city", "pop"=>30847, "name_ru"=>"Уэст Лафайет", "name_en"=>"West Lafayette", "lang"=>"en", "lat"=>40.444667, "lon"=>-86.911929),
"/US/US-IN/west-lebanon-6398" => array("type"=>"town", "pop"=>759, "name_ru"=>"Уэст Лебанон", "name_en"=>"West Lebanon", "lang"=>"en", "lat"=>40.271303, "lon"=>-87.383397),
"/US/US-IN/west-terre-haute-6400" => array("type"=>"town", "pop"=>2232, "name_ru"=>"Уэст Терр Хаут", "name_en"=>"West Terre Haute", "lang"=>"en", "lat"=>39.464300, "lon"=>-87.449049),
"/US/US-IN/westfield-6395" => array("type"=>"town", "pop"=>21171, "name_ru"=>"Уэстфилд", "name_en"=>"Westfield", "lang"=>"en", "lat"=>40.046985, "lon"=>-86.130913),
"/US/US-IN/westport-6399" => array("type"=>"town", "pop"=>1453, "name_ru"=>"Уэстпорт", "name_en"=>"Westport", "lang"=>"en", "lat"=>39.176063, "lon"=>-85.575266),
"/US/US-IN/westville-6401" => array("type"=>"town", "pop"=>5177, "name_ru"=>"Уэствилл", "name_en"=>"Westville", "lang"=>"en", "lat"=>41.539568, "lon"=>-86.906087),
"/US/US-IN/wheatfield-6402" => array("type"=>"town", "pop"=>920, "name_ru"=>"Уитфилд", "name_en"=>"Wheatfield", "lang"=>"en", "lat"=>41.191338, "lon"=>-87.052451),
"/US/US-IN/wheatland-6403" => array("type"=>"town", "pop"=>484, "name_ru"=>"Уитленд", "name_en"=>"Wheatland", "lang"=>"en", "lat"=>38.664134, "lon"=>-87.306664),
"/US/US-IN/whiteland-6404" => array("type"=>"town", "pop"=>4464, "name_ru"=>"Уайтленд", "name_en"=>"Whiteland", "lang"=>"en", "lat"=>39.550344, "lon"=>-86.088641),
"/US/US-IN/whitestown-6405" => array("type"=>"town", "pop"=>683, "name_ru"=>"Уайтстаун", "name_en"=>"Whitestown", "lang"=>"en", "lat"=>39.995573, "lon"=>-86.346354),
"/US/US-IN/whitewater-6406" => array("type"=>"town", "pop"=>75, "name_ru"=>"Уайтуотер", "name_en"=>"Whitewater", "lang"=>"en", "lat"=>39.944343, "lon"=>-84.830741),
"/US/US-IN/whiting-6407" => array("type"=>"city", "pop"=>4751, "name_ru"=>"Уайтинг", "name_en"=>"Whiting", "lang"=>"en", "lat"=>41.678167, "lon"=>-87.486699),
"/US/US-IN/wilkinson-6408" => array("type"=>"town", "pop"=>359, "name_ru"=>"Уилкинсон", "name_en"=>"Wilkinson", "lang"=>"en", "lat"=>39.884766, "lon"=>-85.608266),
"/US/US-IN/williams-creek-6409" => array("type"=>"town", "pop"=>418, "name_ru"=>"Уильямс Крик", "name_en"=>"Williams Creek", "lang"=>"en", "lat"=>39.900400, "lon"=>-86.149946),
"/US/US-IN/williamsport-6410" => array("type"=>"town", "pop"=>1851, "name_ru"=>"Уильямспорт", "name_en"=>"Williamsport", "lang"=>"en", "lat"=>40.287099, "lon"=>-87.293470),
"/US/US-IN/winamac-6411" => array("type"=>"town", "pop"=>2493, "name_ru"=>"Уинимак", "name_en"=>"Winamac", "lang"=>"en", "lat"=>41.052854, "lon"=>-86.603314),
"/US/US-IN/winchester-6412" => array("type"=>"city", "pop"=>4601, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>40.171909, "lon"=>-84.977114),
"/US/US-IN/windfall-city-6413" => array("type"=>"town", "pop"=>656, "name_ru"=>"Уиндфолл Сити", "name_en"=>"Windfall City", "lang"=>"en", "lat"=>40.362867, "lon"=>-85.957506),
"/US/US-IN/wingate-6414" => array("type"=>"town", "pop"=>291, "name_ru"=>"Уингейт", "name_en"=>"Wingate", "lang"=>"en", "lat"=>40.170515, "lon"=>-87.073537),
"/US/US-IN/winona-lake-6415" => array("type"=>"town", "pop"=>4292, "name_ru"=>"Уайнона Лейк", "name_en"=>"Winona Lake", "lang"=>"en", "lat"=>41.218637, "lon"=>-85.815286),
"/US/US-IN/winslow-6416" => array("type"=>"town", "pop"=>834, "name_ru"=>"Уинслоу", "name_en"=>"Winslow", "lang"=>"en", "lat"=>38.383753, "lon"=>-87.212662),
"/US/US-IN/wolcott-6417" => array("type"=>"town", "pop"=>905, "name_ru"=>"Уолкотт", "name_en"=>"Wolcott", "lang"=>"en", "lat"=>40.758213, "lon"=>-87.043334),
"/US/US-IN/wolcottville-6418" => array("type"=>"town", "pop"=>946, "name_ru"=>"Уолкоттвилл", "name_en"=>"Wolcottville", "lang"=>"en", "lat"=>41.525473, "lon"=>-85.366037),
"/US/US-IN/woodburn-6419" => array("type"=>"city", "pop"=>1633, "name_ru"=>"Вудберн", "name_en"=>"Woodburn", "lang"=>"en", "lat"=>41.126025, "lon"=>-84.852762),
"/US/US-IN/woodlawn-heights-6420" => array("type"=>"town", "pop"=>71, "name_ru"=>"Вудлон Хайтс", "name_en"=>"Woodlawn Heights", "lang"=>"en", "lat"=>40.117158, "lon"=>-85.696734),
"/US/US-IN/worthington-6421" => array("type"=>"town", "pop"=>1446, "name_ru"=>"Уортингтон", "name_en"=>"Worthington", "lang"=>"en", "lat"=>39.118519, "lon"=>-86.979844),
"/US/US-IN/wynnedale-6422" => array("type"=>"town", "pop"=>277, "name_ru"=>"Винндейл", "name_en"=>"Wynnedale", "lang"=>"en", "lat"=>39.832750, "lon"=>-86.198744),
"/US/US-IN/yeoman-6423" => array("type"=>"town", "pop"=>90, "name_ru"=>"Йеоман", "name_en"=>"Yeoman", "lang"=>"en", "lat"=>40.667116, "lon"=>-86.723427),
"/US/US-IN/yorktown-6424" => array("type"=>"town", "pop"=>8523, "name_ru"=>"Йорктаун", "name_en"=>"Yorktown", "lang"=>"en", "lat"=>40.173256, "lon"=>-85.478601),
"/US/US-IN/zionsville-6425" => array("type"=>"town", "pop"=>13655, "name_ru"=>"Зионсвилл", "name_en"=>"Zionsville", "lang"=>"en", "lat"=>39.952421, "lon"=>-86.272722),
"/US/US-KS" => array("type"=>"state", "pop"=>2802134, "name_ru"=>"Канзас", "name_en"=>"Kansas", "lang"=>"en"),
"/US/US-KS/abbyville-7381" => array("type"=>"city", "pop"=>124, "name_ru"=>"Аббивилл", "name_en"=>"Abbyville", "lang"=>"en", "lat"=>37.970606, "lon"=>-98.203424),
"/US/US-KS/abilene-7382" => array("type"=>"city", "pop"=>6400, "name_ru"=>"Абилин", "name_en"=>"Abilene", "lang"=>"en", "lat"=>38.919961, "lon"=>-97.221965),
"/US/US-KS/admire-7383" => array("type"=>"city", "pop"=>174, "name_ru"=>"Эдмаер", "name_en"=>"Admire", "lang"=>"en", "lat"=>38.641167, "lon"=>-96.102852),
"/US/US-KS/agenda-7384" => array("type"=>"city", "pop"=>67, "name_ru"=>"Агенда", "name_en"=>"Agenda", "lang"=>"en", "lat"=>39.707240, "lon"=>-97.431863),
"/US/US-KS/agra-7385" => array("type"=>"city", "pop"=>269, "name_ru"=>"Агра", "name_en"=>"Agra", "lang"=>"en", "lat"=>39.761032, "lon"=>-99.119750),
"/US/US-KS/albert-7386" => array("type"=>"city", "pop"=>175, "name_ru"=>"Альберт", "name_en"=>"Albert", "lang"=>"en", "lat"=>38.454514, "lon"=>-99.011724),
"/US/US-KS/alden-7387" => array("type"=>"city", "pop"=>154, "name_ru"=>"Олден", "name_en"=>"Alden", "lang"=>"en", "lat"=>38.244082, "lon"=>-98.311149),
"/US/US-KS/alexander-7388" => array("type"=>"city", "pop"=>66, "name_ru"=>"Александр", "name_en"=>"Alexander", "lang"=>"en", "lat"=>38.469357, "lon"=>-99.552024),
"/US/US-KS/allen-7389" => array("type"=>"city", "pop"=>212, "name_ru"=>"Аллен", "name_en"=>"Allen", "lang"=>"en", "lat"=>38.655458, "lon"=>-96.168512),
"/US/US-KS/alma-7390" => array("type"=>"city", "pop"=>755, "name_ru"=>"Алма", "name_en"=>"Alma", "lang"=>"en", "lat"=>39.014490, "lon"=>-96.288578),
"/US/US-KS/almena-7391" => array("type"=>"city", "pop"=>425, "name_ru"=>"Алмена", "name_en"=>"Almena", "lang"=>"en", "lat"=>39.891516, "lon"=>-99.709794),
"/US/US-KS/alta-vista-7393" => array("type"=>"city", "pop"=>420, "name_ru"=>"Альта Виста", "name_en"=>"Alta Vista", "lang"=>"en", "lat"=>38.864734, "lon"=>-96.488963),
"/US/US-KS/altamont-7392" => array("type"=>"city", "pop"=>1049, "name_ru"=>"Альтамонт", "name_en"=>"Altamont", "lang"=>"en", "lat"=>37.191544, "lon"=>-95.294719),
"/US/US-KS/alton-7394" => array("type"=>"city", "pop"=>102, "name_ru"=>"Альтон", "name_en"=>"Alton", "lang"=>"en", "lat"=>39.468197, "lon"=>-98.948247),
"/US/US-KS/altoona-7395" => array("type"=>"city", "pop"=>463, "name_ru"=>"Альтоона", "name_en"=>"Altoona", "lang"=>"en", "lat"=>37.525518, "lon"=>-95.661646),
"/US/US-KS/americus-7396" => array("type"=>"city", "pop"=>917, "name_ru"=>"Америкус", "name_en"=>"Americus", "lang"=>"en", "lat"=>38.506482, "lon"=>-96.260981),
"/US/US-KS/andale-7397" => array("type"=>"city", "pop"=>879, "name_ru"=>"Андейл", "name_en"=>"Andale", "lang"=>"en", "lat"=>37.792690, "lon"=>-97.629120),
"/US/US-KS/andover-7398" => array("type"=>"city", "pop"=>10351, "name_ru"=>"Андоувер", "name_en"=>"Andover", "lang"=>"en", "lat"=>37.691150, "lon"=>-97.139322),
"/US/US-KS/anthony-7399" => array("type"=>"city", "pop"=>2224, "name_ru"=>"Энтони", "name_en"=>"Anthony", "lang"=>"en", "lat"=>37.153880, "lon"=>-98.029057),
"/US/US-KS/arcadia-7400" => array("type"=>"city", "pop"=>386, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>37.641172, "lon"=>-94.624191),
"/US/US-KS/argonia-7401" => array("type"=>"city", "pop"=>472, "name_ru"=>"Аргония", "name_en"=>"Argonia", "lang"=>"en", "lat"=>37.265290, "lon"=>-97.764489),
"/US/US-KS/arkansas-city-7402" => array("type"=>"city", "pop"=>11070, "name_ru"=>"Арканзас Сити", "name_en"=>"Arkansas City", "lang"=>"en", "lat"=>37.068820, "lon"=>-97.040642),
"/US/US-KS/arlington-7403" => array("type"=>"city", "pop"=>434, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>37.896013, "lon"=>-98.177623),
"/US/US-KS/arma-7404" => array("type"=>"city", "pop"=>1513, "name_ru"=>"Арма", "name_en"=>"Arma", "lang"=>"en", "lat"=>37.542534, "lon"=>-94.700221),
"/US/US-KS/ashland-7405" => array("type"=>"city", "pop"=>867, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>37.186777, "lon"=>-99.768854),
"/US/US-KS/assaria-7406" => array("type"=>"city", "pop"=>449, "name_ru"=>"Ассария", "name_en"=>"Assaria", "lang"=>"en", "lat"=>38.680370, "lon"=>-97.603716),
"/US/US-KS/atchison-7407" => array("type"=>"city", "pop"=>10402, "name_ru"=>"Атчисон", "name_en"=>"Atchison", "lang"=>"en", "lat"=>39.561465, "lon"=>-95.133400),
"/US/US-KS/athol-7408" => array("type"=>"city", "pop"=>44, "name_ru"=>"Атол", "name_en"=>"Athol", "lang"=>"en", "lat"=>39.765713, "lon"=>-98.919862),
"/US/US-KS/atlanta-7409" => array("type"=>"city", "pop"=>239, "name_ru"=>"Атланта", "name_en"=>"Atlanta", "lang"=>"en", "lat"=>37.436361, "lon"=>-96.767101),
"/US/US-KS/attica-7410" => array("type"=>"city", "pop"=>586, "name_ru"=>"Аттика", "name_en"=>"Attica", "lang"=>"en", "lat"=>37.242482, "lon"=>-98.225900),
"/US/US-KS/atwood-7411" => array("type"=>"city", "pop"=>1069, "name_ru"=>"Атвуд", "name_en"=>"Atwood", "lang"=>"en", "lat"=>39.810446, "lon"=>-101.040967),
"/US/US-KS/auburn-7412" => array("type"=>"city", "pop"=>1151, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>38.905689, "lon"=>-95.816739),
"/US/US-KS/augusta-7413" => array("type"=>"city", "pop"=>8693, "name_ru"=>"Огаста", "name_en"=>"Augusta", "lang"=>"en", "lat"=>37.692298, "lon"=>-96.975258),
"/US/US-KS/aurora-7414" => array("type"=>"city", "pop"=>74, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>39.451643, "lon"=>-97.530084),
"/US/US-KS/axtell-7415" => array("type"=>"city", "pop"=>417, "name_ru"=>"Акстелл", "name_en"=>"Axtell", "lang"=>"en", "lat"=>39.871942, "lon"=>-96.257000),
"/US/US-KS/baldwin-city-7416" => array("type"=>"city", "pop"=>4308, "name_ru"=>"Болдуин Сити", "name_en"=>"Baldwin City", "lang"=>"en", "lat"=>38.776550, "lon"=>-95.185064),
"/US/US-KS/barnard-7417" => array("type"=>"city", "pop"=>114, "name_ru"=>"Барнард", "name_en"=>"Barnard", "lang"=>"en", "lat"=>39.189529, "lon"=>-98.044047),
"/US/US-KS/barnes-7418" => array("type"=>"city", "pop"=>136, "name_ru"=>"Барнс", "name_en"=>"Barnes", "lang"=>"en", "lat"=>39.711772, "lon"=>-96.872746),
"/US/US-KS/bartlett-7419" => array("type"=>"city", "pop"=>121, "name_ru"=>"Барлет", "name_en"=>"Bartlett", "lang"=>"en", "lat"=>37.054941, "lon"=>-95.211345),
"/US/US-KS/basehor-7420" => array("type"=>"city", "pop"=>4230, "name_ru"=>"Бейсхор", "name_en"=>"Basehor", "lang"=>"en", "lat"=>39.127152, "lon"=>-94.929055),
"/US/US-KS/bassett-7421" => array("type"=>"city", "pop"=>21, "name_ru"=>"Бассет", "name_en"=>"Bassett", "lang"=>"en", "lat"=>37.906704, "lon"=>-95.407161),
"/US/US-KS/baxter-springs-7422" => array("type"=>"city", "pop"=>4153, "name_ru"=>"Бэкстер Спрингс", "name_en"=>"Baxter Springs", "lang"=>"en", "lat"=>37.020959, "lon"=>-94.737246),
"/US/US-KS/bazine-7423" => array("type"=>"city", "pop"=>334, "name_ru"=>"Базин", "name_en"=>"Bazine", "lang"=>"en", "lat"=>38.445962, "lon"=>-99.692153),
"/US/US-KS/beattie-7424" => array("type"=>"city", "pop"=>259, "name_ru"=>"Битти", "name_en"=>"Beattie", "lang"=>"en", "lat"=>39.862730, "lon"=>-96.417560),
"/US/US-KS/bel-aire-7425" => array("type"=>"city", "pop"=>6797, "name_ru"=>"Бель Эр", "name_en"=>"Bel Aire", "lang"=>"en", "lat"=>37.764550, "lon"=>-97.268183),
"/US/US-KS/belle-plaine-7426" => array("type"=>"city", "pop"=>1528, "name_ru"=>"Бель Плейн", "name_en"=>"Belle Plaine", "lang"=>"en", "lat"=>37.394194, "lon"=>-97.278755),
"/US/US-KS/belleville-7427" => array("type"=>"city", "pop"=>1832, "name_ru"=>"Белелвилл", "name_en"=>"Belleville", "lang"=>"en", "lat"=>39.823075, "lon"=>-97.633543),
"/US/US-KS/beloit-7428" => array("type"=>"city", "pop"=>3640, "name_ru"=>"Белойт", "name_en"=>"Beloit", "lang"=>"en", "lat"=>39.465796, "lon"=>-98.107834),
"/US/US-KS/belpre-7429" => array("type"=>"city", "pop"=>94, "name_ru"=>"Белпре", "name_en"=>"Belpre", "lang"=>"en", "lat"=>37.950961, "lon"=>-99.099215),
"/US/US-KS/belvue-7430" => array("type"=>"city", "pop"=>220, "name_ru"=>"Белвью", "name_en"=>"Belvue", "lang"=>"en", "lat"=>39.216688, "lon"=>-96.178203),
"/US/US-KS/benedict-7431" => array("type"=>"city", "pop"=>98, "name_ru"=>"Бенедикт", "name_en"=>"Benedict", "lang"=>"en", "lat"=>37.624699, "lon"=>-95.741899),
"/US/US-KS/bennington-7432" => array("type"=>"city", "pop"=>609, "name_ru"=>"Беннингтон", "name_en"=>"Bennington", "lang"=>"en", "lat"=>39.031973, "lon"=>-97.593953),
"/US/US-KS/bentley-7433" => array("type"=>"city", "pop"=>519, "name_ru"=>"Бентли", "name_en"=>"Bentley", "lang"=>"en", "lat"=>37.886373, "lon"=>-97.515564),
"/US/US-KS/benton-7434" => array("type"=>"city", "pop"=>806, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>37.788839, "lon"=>-97.108085),
"/US/US-KS/bern-7435" => array("type"=>"city", "pop"=>192, "name_ru"=>"Берн", "name_en"=>"Bern", "lang"=>"en", "lat"=>39.962722, "lon"=>-95.972253),
"/US/US-KS/beverly-7436" => array("type"=>"city", "pop"=>184, "name_ru"=>"Беверли", "name_en"=>"Beverly", "lang"=>"en", "lat"=>39.013774, "lon"=>-97.975687),
"/US/US-KS/bird-city-7437" => array("type"=>"city", "pop"=>407, "name_ru"=>"Берд Сити", "name_en"=>"Bird City", "lang"=>"en", "lat"=>39.749939, "lon"=>-101.531664),
"/US/US-KS/bison-7438" => array("type"=>"city", "pop"=>208, "name_ru"=>"Бизон", "name_en"=>"Bison", "lang"=>"en", "lat"=>38.519802, "lon"=>-99.197158),
"/US/US-KS/blue-mound-7439" => array("type"=>"city", "pop"=>285, "name_ru"=>"Блю Маунд", "name_en"=>"Blue Mound", "lang"=>"en", "lat"=>38.089200, "lon"=>-95.009428),
"/US/US-KS/blue-rapids-7440" => array("type"=>"city", "pop"=>1018, "name_ru"=>"Блу Рэпидс", "name_en"=>"Blue Rapids", "lang"=>"en", "lat"=>39.678522, "lon"=>-96.659711),
"/US/US-KS/bluff-city-7441" => array("type"=>"city", "pop"=>73, "name_ru"=>"Блафф Сити", "name_en"=>"Bluff City", "lang"=>"en", "lat"=>37.075648, "lon"=>-97.875041),
"/US/US-KS/bogue-7442" => array("type"=>"city", "pop"=>160, "name_ru"=>"Бог", "name_en"=>"Bogue", "lang"=>"en", "lat"=>39.359641, "lon"=>-99.687890),
"/US/US-KS/bonner-springs-7443" => array("type"=>"city", "pop"=>7161, "name_ru"=>"Боннер Спрингс", "name_en"=>"Bonner Springs", "lang"=>"en", "lat"=>39.079304, "lon"=>-94.881545),
"/US/US-KS/brewster-7444" => array("type"=>"city", "pop"=>248, "name_ru"=>"Брюстер", "name_en"=>"Brewster", "lang"=>"en", "lat"=>39.362877, "lon"=>-101.377446),
"/US/US-KS/bronson-7445" => array("type"=>"city", "pop"=>336, "name_ru"=>"Бронсон", "name_en"=>"Bronson", "lang"=>"en", "lat"=>37.895885, "lon"=>-95.072479),
"/US/US-KS/brookville-7446" => array("type"=>"city", "pop"=>263, "name_ru"=>"Бруквилл", "name_en"=>"Brookville", "lang"=>"en", "lat"=>38.773700, "lon"=>-97.864189),
"/US/US-KS/brownell-7447" => array("type"=>"city", "pop"=>40, "name_ru"=>"Броунелл", "name_en"=>"Brownell", "lang"=>"en", "lat"=>38.640244, "lon"=>-99.743658),
"/US/US-KS/bucklin-7448" => array("type"=>"city", "pop"=>727, "name_ru"=>"Баклин", "name_en"=>"Bucklin", "lang"=>"en", "lat"=>37.548987, "lon"=>-99.634600),
"/US/US-KS/buffalo-7449" => array("type"=>"city", "pop"=>272, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>37.708610, "lon"=>-95.696686),
"/US/US-KS/buhler-7450" => array("type"=>"city", "pop"=>1332, "name_ru"=>"Булер", "name_en"=>"Buhler", "lang"=>"en", "lat"=>38.138148, "lon"=>-97.771312),
"/US/US-KS/bunker-hill-7451" => array("type"=>"city", "pop"=>92, "name_ru"=>"Банкер Хилл", "name_en"=>"Bunker Hill", "lang"=>"en", "lat"=>38.872590, "lon"=>-98.700770),
"/US/US-KS/burden-7452" => array("type"=>"city", "pop"=>529, "name_ru"=>"Берден", "name_en"=>"Burden", "lang"=>"en", "lat"=>37.313135, "lon"=>-96.755029),
"/US/US-KS/burdett-7453" => array("type"=>"city", "pop"=>222, "name_ru"=>"Бердетт", "name_en"=>"Burdett", "lang"=>"en", "lat"=>38.193395, "lon"=>-99.526061),
"/US/US-KS/burlingame-7454" => array("type"=>"city", "pop"=>958, "name_ru"=>"Берлингейм", "name_en"=>"Burlingame", "lang"=>"en", "lat"=>38.750841, "lon"=>-95.835454),
"/US/US-KS/burlington-7455" => array("type"=>"city", "pop"=>2679, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>38.194393, "lon"=>-95.743694),
"/US/US-KS/burns-7456" => array("type"=>"city", "pop"=>260, "name_ru"=>"Бернс", "name_en"=>"Burns", "lang"=>"en", "lat"=>38.089748, "lon"=>-96.887677),
"/US/US-KS/burr-oak-7457" => array("type"=>"city", "pop"=>212, "name_ru"=>"Берр Оук", "name_en"=>"Burr Oak", "lang"=>"en", "lat"=>39.870392, "lon"=>-98.303982),
"/US/US-KS/burrton-7458" => array("type"=>"city", "pop"=>896, "name_ru"=>"Берртон", "name_en"=>"Burrton", "lang"=>"en", "lat"=>38.022562, "lon"=>-97.671993),
"/US/US-KS/bushong-7459" => array("type"=>"city", "pop"=>50, "name_ru"=>"Бушонг", "name_en"=>"Bushong", "lang"=>"en", "lat"=>38.642691, "lon"=>-96.257478),
"/US/US-KS/bushton-7460" => array("type"=>"city", "pop"=>285, "name_ru"=>"Буштон", "name_en"=>"Bushton", "lang"=>"en", "lat"=>38.512836, "lon"=>-98.394732),
"/US/US-KS/byers-7461" => array("type"=>"city", "pop"=>49, "name_ru"=>"Байерс", "name_en"=>"Byers", "lang"=>"en", "lat"=>37.787814, "lon"=>-98.866730),
"/US/US-KS/caldwell-7462" => array("type"=>"city", "pop"=>1144, "name_ru"=>"Колдуэлл", "name_en"=>"Caldwell", "lang"=>"en", "lat"=>37.035152, "lon"=>-97.608679),
"/US/US-KS/cambridge-7463" => array("type"=>"city", "pop"=>96, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>37.317297, "lon"=>-96.666783),
"/US/US-KS/camp-forsyth-7464" => array("type"=>"cdp", "pop"=>1967, "name_ru"=>"Кэмп Форсайт", "name_en"=>"Camp Forsyth", "lang"=>"en", "lat"=>39.063363, "lon"=>-96.824380),
"/US/US-KS/caney-7465" => array("type"=>"city", "pop"=>1976, "name_ru"=>"Кэни", "name_en"=>"Caney", "lang"=>"en", "lat"=>37.013455, "lon"=>-95.931541),
"/US/US-KS/canton-7466" => array("type"=>"city", "pop"=>790, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>38.386125, "lon"=>-97.429868),
"/US/US-KS/carbondale-7467" => array("type"=>"city", "pop"=>1378, "name_ru"=>"Карбондейл", "name_en"=>"Carbondale", "lang"=>"en", "lat"=>38.820088, "lon"=>-95.692220),
"/US/US-KS/carlton-7468" => array("type"=>"city", "pop"=>38, "name_ru"=>"Карлтон", "name_en"=>"Carlton", "lang"=>"en", "lat"=>38.686783, "lon"=>-97.293712),
"/US/US-KS/cassoday-7469" => array("type"=>"city", "pop"=>128, "name_ru"=>"Кассодей", "name_en"=>"Cassoday", "lang"=>"en", "lat"=>38.038707, "lon"=>-96.636868),
"/US/US-KS/cawker-city-7470" => array("type"=>"city", "pop"=>463, "name_ru"=>"Кокер Сити", "name_en"=>"Cawker City", "lang"=>"en", "lat"=>39.509729, "lon"=>-98.432128),
"/US/US-KS/cedar-7471" => array("type"=>"city", "pop"=>22, "name_ru"=>"Сидар", "name_en"=>"Cedar", "lang"=>"en", "lat"=>39.656700, "lon"=>-98.940030),
"/US/US-KS/cedar-point-7472" => array("type"=>"city", "pop"=>50, "name_ru"=>"Сидар Пойнт", "name_en"=>"Cedar Point", "lang"=>"en", "lat"=>38.259906, "lon"=>-96.821711),
"/US/US-KS/cedar-vale-7473" => array("type"=>"city", "pop"=>617, "name_ru"=>"Сидар Вейл", "name_en"=>"Cedar Vale", "lang"=>"en", "lat"=>37.106366, "lon"=>-96.500577),
"/US/US-KS/centralia-7474" => array("type"=>"city", "pop"=>486, "name_ru"=>"Сентралия", "name_en"=>"Centralia", "lang"=>"en", "lat"=>39.724035, "lon"=>-96.130799),
"/US/US-KS/chanute-7475" => array("type"=>"city", "pop"=>8843, "name_ru"=>"Чанат", "name_en"=>"Chanute", "lang"=>"en", "lat"=>37.671039, "lon"=>-95.460266),
"/US/US-KS/chapman-7476" => array("type"=>"city", "pop"=>1332, "name_ru"=>"Чэпмен", "name_en"=>"Chapman", "lang"=>"en", "lat"=>38.971756, "lon"=>-97.020642),
"/US/US-KS/chase-7477" => array("type"=>"city", "pop"=>446, "name_ru"=>"Чейз", "name_en"=>"Chase", "lang"=>"en", "lat"=>38.355672, "lon"=>-98.348399),
"/US/US-KS/chautauqua-7478" => array("type"=>"city", "pop"=>97, "name_ru"=>"Чаутоква", "name_en"=>"Chautauqua", "lang"=>"en", "lat"=>37.023861, "lon"=>-96.176837),
"/US/US-KS/cheney-7479" => array("type"=>"city", "pop"=>2033, "name_ru"=>"Чини", "name_en"=>"Cheney", "lang"=>"en", "lat"=>37.627868, "lon"=>-97.780189),
"/US/US-KS/cherokee-7480" => array("type"=>"city", "pop"=>715, "name_ru"=>"Чероки", "name_en"=>"Cherokee", "lang"=>"en", "lat"=>37.344938, "lon"=>-94.821265),
"/US/US-KS/cherryvale-7481" => array("type"=>"city", "pop"=>2253, "name_ru"=>"Черривейл", "name_en"=>"Cherryvale", "lang"=>"en", "lat"=>37.268031, "lon"=>-95.551083),
"/US/US-KS/chetopa-7482" => array("type"=>"city", "pop"=>1234, "name_ru"=>"Четопа", "name_en"=>"Chetopa", "lang"=>"en", "lat"=>37.036306, "lon"=>-95.093273),
"/US/US-KS/cimarron-7483" => array("type"=>"city", "pop"=>1989, "name_ru"=>"Симаррон", "name_en"=>"Cimarron", "lang"=>"en", "lat"=>37.808185, "lon"=>-100.345697),
"/US/US-KS/circleville-7484" => array("type"=>"city", "pop"=>179, "name_ru"=>"Серклвилл", "name_en"=>"Circleville", "lang"=>"en", "lat"=>39.508945, "lon"=>-95.855329),
"/US/US-KS/claflin-7485" => array("type"=>"city", "pop"=>654, "name_ru"=>"Клафлин", "name_en"=>"Claflin", "lang"=>"en", "lat"=>38.523920, "lon"=>-98.536807),
"/US/US-KS/clay-center-7486" => array("type"=>"city", "pop"=>4442, "name_ru"=>"Клэй Сентер", "name_en"=>"Clay Center", "lang"=>"en", "lat"=>39.379924, "lon"=>-97.122870),
"/US/US-KS/clayton-7487" => array("type"=>"city", "pop"=>60, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>39.737243, "lon"=>-100.176406),
"/US/US-KS/clearwater-7488" => array("type"=>"city", "pop"=>2405, "name_ru"=>"Клирвотер", "name_en"=>"Clearwater", "lang"=>"en", "lat"=>37.505629, "lon"=>-97.501266),
"/US/US-KS/clifton-7489" => array("type"=>"city", "pop"=>502, "name_ru"=>"Слифтон", "name_en"=>"Clifton", "lang"=>"en", "lat"=>39.567703, "lon"=>-97.279679),
"/US/US-KS/climax-7490" => array("type"=>"city", "pop"=>58, "name_ru"=>"Клаймакс", "name_en"=>"Climax", "lang"=>"en", "lat"=>37.719278, "lon"=>-96.223556),
"/US/US-KS/clyde-7491" => array("type"=>"city", "pop"=>685, "name_ru"=>"Клайд", "name_en"=>"Clyde", "lang"=>"en", "lat"=>39.591707, "lon"=>-97.400612),
"/US/US-KS/coats-7492" => array("type"=>"city", "pop"=>109, "name_ru"=>"Коатс", "name_en"=>"Coats", "lang"=>"en", "lat"=>37.510901, "lon"=>-98.824380),
"/US/US-KS/coffeyville-7493" => array("type"=>"city", "pop"=>10312, "name_ru"=>"Коффивилл", "name_en"=>"Coffeyville", "lang"=>"en", "lat"=>37.040791, "lon"=>-95.632077),
"/US/US-KS/colby-7494" => array("type"=>"city", "pop"=>4803, "name_ru"=>"Колби", "name_en"=>"Colby", "lang"=>"en", "lat"=>39.387675, "lon"=>-101.045917),
"/US/US-KS/coldwater-7495" => array("type"=>"city", "pop"=>779, "name_ru"=>"Коулдуотер", "name_en"=>"Coldwater", "lang"=>"en", "lat"=>37.257856, "lon"=>-99.336841),
"/US/US-KS/collyer-7496" => array("type"=>"city", "pop"=>118, "name_ru"=>"Коллайер", "name_en"=>"Collyer", "lang"=>"en", "lat"=>39.037831, "lon"=>-100.118114),
"/US/US-KS/colony-7497" => array("type"=>"city", "pop"=>377, "name_ru"=>"Колони", "name_en"=>"Colony", "lang"=>"en", "lat"=>38.070639, "lon"=>-95.362163),
"/US/US-KS/columbus-7498" => array("type"=>"city", "pop"=>3185, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>37.170391, "lon"=>-94.844238),
"/US/US-KS/colwich-7499" => array("type"=>"city", "pop"=>1407, "name_ru"=>"Колуич", "name_en"=>"Colwich", "lang"=>"en", "lat"=>37.781700, "lon"=>-97.537550),
"/US/US-KS/concordia-7500" => array("type"=>"city", "pop"=>5208, "name_ru"=>"Конкордия", "name_en"=>"Concordia", "lang"=>"en", "lat"=>39.565075, "lon"=>-97.654624),
"/US/US-KS/conway-springs-7501" => array("type"=>"city", "pop"=>1193, "name_ru"=>"Конуэй Спрингс", "name_en"=>"Conway Springs", "lang"=>"en", "lat"=>37.389527, "lon"=>-97.644005),
"/US/US-KS/coolidge-7502" => array("type"=>"city", "pop"=>86, "name_ru"=>"Кулидж", "name_en"=>"Coolidge", "lang"=>"en", "lat"=>38.040864, "lon"=>-102.007406),
"/US/US-KS/copeland-7503" => array("type"=>"city", "pop"=>308, "name_ru"=>"Коупленд", "name_en"=>"Copeland", "lang"=>"en", "lat"=>37.540174, "lon"=>-100.628473),
"/US/US-KS/corning-7504" => array("type"=>"city", "pop"=>160, "name_ru"=>"Корнинг", "name_en"=>"Corning", "lang"=>"en", "lat"=>39.656422, "lon"=>-96.028671),
"/US/US-KS/cottonwood-falls-7505" => array("type"=>"city", "pop"=>878, "name_ru"=>"Коттонвуд Фоллс", "name_en"=>"Cottonwood Falls", "lang"=>"en", "lat"=>38.369060, "lon"=>-96.542619),
"/US/US-KS/council-grove-7506" => array("type"=>"city", "pop"=>2274, "name_ru"=>"Кансил Гров", "name_en"=>"Council Grove", "lang"=>"en", "lat"=>38.660671, "lon"=>-96.489392),
"/US/US-KS/countryside-7507" => array("type"=>"city", "pop"=>312, "name_ru"=>"Кантрисайд", "name_en"=>"Countryside", "lang"=>"en", "lat"=>39.015980, "lon"=>-94.655062),
"/US/US-KS/courtland-7508" => array("type"=>"city", "pop"=>298, "name_ru"=>"Кортленд", "name_en"=>"Courtland", "lang"=>"en", "lat"=>39.782794, "lon"=>-97.895310),
"/US/US-KS/coyville-7509" => array("type"=>"city", "pop"=>68, "name_ru"=>"Койвилл", "name_en"=>"Coyville", "lang"=>"en", "lat"=>37.687202, "lon"=>-95.895340),
"/US/US-KS/cuba-7510" => array("type"=>"city", "pop"=>191, "name_ru"=>"Куба", "name_en"=>"Cuba", "lang"=>"en", "lat"=>39.802219, "lon"=>-97.456812),
"/US/US-KS/cullison-7511" => array("type"=>"city", "pop"=>96, "name_ru"=>"Каллисон", "name_en"=>"Cullison", "lang"=>"en", "lat"=>37.629669, "lon"=>-98.905065),
"/US/US-KS/culver-7512" => array("type"=>"city", "pop"=>160, "name_ru"=>"Калвер", "name_en"=>"Culver", "lang"=>"en", "lat"=>38.968309, "lon"=>-97.759313),
"/US/US-KS/cunningham-7513" => array("type"=>"city", "pop"=>452, "name_ru"=>"Каннингхэм", "name_en"=>"Cunningham", "lang"=>"en", "lat"=>37.644947, "lon"=>-98.432014),
"/US/US-KS/damar-7514" => array("type"=>"city", "pop"=>142, "name_ru"=>"Дамар", "name_en"=>"Damar", "lang"=>"en", "lat"=>39.319137, "lon"=>-99.583768),
"/US/US-KS/danville-7515" => array("type"=>"city", "pop"=>54, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>37.285741, "lon"=>-97.891029),
"/US/US-KS/de-soto-7523" => array("type"=>"city", "pop"=>5420, "name_ru"=>"Де Сото", "name_en"=>"De Soto", "lang"=>"en", "lat"=>38.973608, "lon"=>-94.952031),
"/US/US-KS/dearing-7516" => array("type"=>"city", "pop"=>446, "name_ru"=>"Диэринг", "name_en"=>"Dearing", "lang"=>"en", "lat"=>37.059070, "lon"=>-95.712107),
"/US/US-KS/deerfield-7517" => array("type"=>"city", "pop"=>815, "name_ru"=>"Дирфилд", "name_en"=>"Deerfield", "lang"=>"en", "lat"=>37.981407, "lon"=>-101.132636),
"/US/US-KS/delia-7518" => array("type"=>"city", "pop"=>178, "name_ru"=>"Делия", "name_en"=>"Delia", "lang"=>"en", "lat"=>39.239325, "lon"=>-95.964275),
"/US/US-KS/delphos-7519" => array("type"=>"city", "pop"=>437, "name_ru"=>"Делфос", "name_en"=>"Delphos", "lang"=>"en", "lat"=>39.274770, "lon"=>-97.766062),
"/US/US-KS/denison-7520" => array("type"=>"city", "pop"=>223, "name_ru"=>"Денисон", "name_en"=>"Denison", "lang"=>"en", "lat"=>39.393210, "lon"=>-95.628152),
"/US/US-KS/denton-7521" => array("type"=>"city", "pop"=>176, "name_ru"=>"Дентон", "name_en"=>"Denton", "lang"=>"en", "lat"=>39.731569, "lon"=>-95.269773),
"/US/US-KS/derby-7522" => array("type"=>"city", "pop"=>22517, "name_ru"=>"Дерби", "name_en"=>"Derby", "lang"=>"en", "lat"=>37.549300, "lon"=>-97.258874),
"/US/US-KS/dexter-7524" => array("type"=>"city", "pop"=>336, "name_ru"=>"Декстер", "name_en"=>"Dexter", "lang"=>"en", "lat"=>37.179344, "lon"=>-96.715682),
"/US/US-KS/dighton-7525" => array("type"=>"city", "pop"=>1019, "name_ru"=>"Дайтон", "name_en"=>"Dighton", "lang"=>"en", "lat"=>38.481478, "lon"=>-100.465793),
"/US/US-KS/dodge-city-7526" => array("type"=>"city", "pop"=>25689, "name_ru"=>"Додж Сити", "name_en"=>"Dodge City", "lang"=>"en", "lat"=>37.758655, "lon"=>-100.015401),
"/US/US-KS/dorrance-7527" => array("type"=>"city", "pop"=>184, "name_ru"=>"Доранс", "name_en"=>"Dorrance", "lang"=>"en", "lat"=>38.846344, "lon"=>-98.590198),
"/US/US-KS/douglass-7528" => array("type"=>"city", "pop"=>1783, "name_ru"=>"Дугласс", "name_en"=>"Douglass", "lang"=>"en", "lat"=>37.517843, "lon"=>-97.011421),
"/US/US-KS/downs-7529" => array("type"=>"city", "pop"=>883, "name_ru"=>"Даунс", "name_en"=>"Downs", "lang"=>"en", "lat"=>39.503319, "lon"=>-98.546797),
"/US/US-KS/dresden-7530" => array("type"=>"city", "pop"=>43, "name_ru"=>"Дрезден", "name_en"=>"Dresden", "lang"=>"en", "lat"=>39.621421, "lon"=>-100.419655),
"/US/US-KS/dunlap-7531" => array("type"=>"city", "pop"=>81, "name_ru"=>"Данлап", "name_en"=>"Dunlap", "lang"=>"en", "lat"=>38.575809, "lon"=>-96.365839),
"/US/US-KS/durham-7532" => array("type"=>"city", "pop"=>101, "name_ru"=>"Дарем", "name_en"=>"Durham", "lang"=>"en", "lat"=>38.485077, "lon"=>-97.226970),
"/US/US-KS/dwight-7533" => array("type"=>"city", "pop"=>326, "name_ru"=>"Двайт", "name_en"=>"Dwight", "lang"=>"en", "lat"=>38.843878, "lon"=>-96.591899),
"/US/US-KS/earlton-7534" => array("type"=>"city", "pop"=>78, "name_ru"=>"Эрлтон", "name_en"=>"Earlton", "lang"=>"en", "lat"=>37.587315, "lon"=>-95.469214),
"/US/US-KS/eastborough-7535" => array("type"=>"city", "pop"=>806, "name_ru"=>"Истборо", "name_en"=>"Eastborough", "lang"=>"en", "lat"=>37.684700, "lon"=>-97.258439),
"/US/US-KS/easton-7536" => array("type"=>"city", "pop"=>344, "name_ru"=>"Истон", "name_en"=>"Easton", "lang"=>"en", "lat"=>39.345357, "lon"=>-95.116297),
"/US/US-KS/edgerton-7537" => array("type"=>"city", "pop"=>1817, "name_ru"=>"Эджертон", "name_en"=>"Edgerton", "lang"=>"en", "lat"=>38.763134, "lon"=>-95.012763),
"/US/US-KS/edmond-7538" => array("type"=>"city", "pop"=>42, "name_ru"=>"Эдмонд", "name_en"=>"Edmond", "lang"=>"en", "lat"=>39.626676, "lon"=>-99.820338),
"/US/US-KS/edna-7539" => array("type"=>"city", "pop"=>413, "name_ru"=>"Эдна", "name_en"=>"Edna", "lang"=>"en", "lat"=>37.058479, "lon"=>-95.358312),
"/US/US-KS/edwardsville-7540" => array("type"=>"city", "pop"=>4462, "name_ru"=>"Эдвардсвилл", "name_en"=>"Edwardsville", "lang"=>"en", "lat"=>39.078294, "lon"=>-94.818632),
"/US/US-KS/effingham-7541" => array("type"=>"city", "pop"=>575, "name_ru"=>"Эффингхэм", "name_en"=>"Effingham", "lang"=>"en", "lat"=>39.522319, "lon"=>-95.398193),
"/US/US-KS/el-dorado-7543" => array("type"=>"city", "pop"=>12591, "name_ru"=>"Эдь Дорадо", "name_en"=>"El Dorado", "lang"=>"en", "lat"=>37.821100, "lon"=>-96.857974),
"/US/US-KS/elbing-7542" => array("type"=>"city", "pop"=>209, "name_ru"=>"Элбинг", "name_en"=>"Elbing", "lang"=>"en", "lat"=>38.054101, "lon"=>-97.126809),
"/US/US-KS/elgin-7544" => array("type"=>"city", "pop"=>71, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>37.001557, "lon"=>-96.280431),
"/US/US-KS/elk-city-7545" => array("type"=>"city", "pop"=>296, "name_ru"=>"Элк Сити", "name_en"=>"Elk City", "lang"=>"en", "lat"=>37.289076, "lon"=>-95.909975),
"/US/US-KS/elk-falls-7546" => array("type"=>"city", "pop"=>104, "name_ru"=>"Элк Фоллс", "name_en"=>"Elk Falls", "lang"=>"en", "lat"=>37.373968, "lon"=>-96.192483),
"/US/US-KS/elkhart-7547" => array("type"=>"city", "pop"=>1901, "name_ru"=>"Элкхарт", "name_en"=>"Elkhart", "lang"=>"en", "lat"=>37.002173, "lon"=>-101.894527),
"/US/US-KS/ellinwood-7548" => array("type"=>"city", "pop"=>2035, "name_ru"=>"Эллинвуд", "name_en"=>"Ellinwood", "lang"=>"en", "lat"=>38.356215, "lon"=>-98.581997),
"/US/US-KS/ellis-7549" => array("type"=>"city", "pop"=>1952, "name_ru"=>"Эллис", "name_en"=>"Ellis", "lang"=>"en", "lat"=>38.936201, "lon"=>-99.558880),
"/US/US-KS/ellsworth-7550" => array("type"=>"city", "pop"=>2858, "name_ru"=>"Эллсворт", "name_en"=>"Ellsworth", "lang"=>"en", "lat"=>38.735624, "lon"=>-98.228060),
"/US/US-KS/elmdale-7551" => array("type"=>"city", "pop"=>47, "name_ru"=>"Элмдейл", "name_en"=>"Elmdale", "lang"=>"en", "lat"=>38.372980, "lon"=>-96.645689),
"/US/US-KS/elsmore-7552" => array("type"=>"city", "pop"=>66, "name_ru"=>"Элсмор", "name_en"=>"Elsmore", "lang"=>"en", "lat"=>37.794181, "lon"=>-95.149203),
"/US/US-KS/elwood-7553" => array("type"=>"city", "pop"=>1127, "name_ru"=>"Элвуд", "name_en"=>"Elwood", "lang"=>"en", "lat"=>39.753933, "lon"=>-94.878969),
"/US/US-KS/emmett-7554" => array("type"=>"city", "pop"=>262, "name_ru"=>"Эммет", "name_en"=>"Emmett", "lang"=>"en", "lat"=>39.307096, "lon"=>-96.056097),
"/US/US-KS/emporia-7555" => array("type"=>"city", "pop"=>26380, "name_ru"=>"Эмпория", "name_en"=>"Emporia", "lang"=>"en", "lat"=>38.411417, "lon"=>-96.193071),
"/US/US-KS/englewood-7556" => array("type"=>"city", "pop"=>95, "name_ru"=>"Энглвуд", "name_en"=>"Englewood", "lang"=>"en", "lat"=>37.039669, "lon"=>-99.985385),
"/US/US-KS/ensign-7557" => array("type"=>"city", "pop"=>191, "name_ru"=>"Энсайн", "name_en"=>"Ensign", "lang"=>"en", "lat"=>37.653583, "lon"=>-100.232242),
"/US/US-KS/enterprise-7558" => array("type"=>"city", "pop"=>816, "name_ru"=>"Энтерпрайз", "name_en"=>"Enterprise", "lang"=>"en", "lat"=>38.901832, "lon"=>-97.116284),
"/US/US-KS/erie-7559" => array("type"=>"city", "pop"=>1160, "name_ru"=>"Эри", "name_en"=>"Erie", "lang"=>"en", "lat"=>37.568486, "lon"=>-95.243673),
"/US/US-KS/esbon-7560" => array("type"=>"city", "pop"=>124, "name_ru"=>"Эсборн", "name_en"=>"Esbon", "lang"=>"en", "lat"=>39.821730, "lon"=>-98.433439),
"/US/US-KS/eskridge-7561" => array("type"=>"city", "pop"=>564, "name_ru"=>"Эскридж", "name_en"=>"Eskridge", "lang"=>"en", "lat"=>38.859569, "lon"=>-96.104461),
"/US/US-KS/eudora-7562" => array("type"=>"city", "pop"=>6226, "name_ru"=>"Эудора", "name_en"=>"Eudora", "lang"=>"en", "lat"=>38.938151, "lon"=>-95.099162),
"/US/US-KS/eureka-7563" => array("type"=>"city", "pop"=>2559, "name_ru"=>"Юрика", "name_en"=>"Eureka", "lang"=>"en", "lat"=>37.822710, "lon"=>-96.288439),
"/US/US-KS/everest-7564" => array("type"=>"city", "pop"=>295, "name_ru"=>"Эверест", "name_en"=>"Everest", "lang"=>"en", "lat"=>39.676742, "lon"=>-95.425014),
"/US/US-KS/fairview-7565" => array("type"=>"city", "pop"=>252, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>39.839885, "lon"=>-95.727106),
"/US/US-KS/fairway-7566" => array("type"=>"city", "pop"=>3843, "name_ru"=>"Фэруэй", "name_en"=>"Fairway", "lang"=>"en", "lat"=>39.025250, "lon"=>-94.628703),
"/US/US-KS/fall-river-7567" => array("type"=>"city", "pop"=>142, "name_ru"=>"Фолл Ривер", "name_en"=>"Fall River", "lang"=>"en", "lat"=>37.608410, "lon"=>-96.028483),
"/US/US-KS/florence-7568" => array("type"=>"city", "pop"=>596, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>38.242836, "lon"=>-96.929112),
"/US/US-KS/fontana-7569" => array("type"=>"city", "pop"=>217, "name_ru"=>"Фонтана", "name_en"=>"Fontana", "lang"=>"en", "lat"=>38.425862, "lon"=>-94.837850),
"/US/US-KS/ford-7570" => array("type"=>"city", "pop"=>324, "name_ru"=>"Форд", "name_en"=>"Ford", "lang"=>"en", "lat"=>37.636988, "lon"=>-99.754003),
"/US/US-KS/formoso-7571" => array("type"=>"city", "pop"=>108, "name_ru"=>"Формосо", "name_en"=>"Formoso", "lang"=>"en", "lat"=>39.779345, "lon"=>-97.993478),
"/US/US-KS/fort-riley-camp-whiteside-7572" => array("type"=>"cdp", "pop"=>112, "name_ru"=>"Форт Рили-Кэмп Уайтсайд", "name_en"=>"Fort Riley-Camp Whiteside", "lang"=>"en", "lat"=>39.084761, "lon"=>-96.771085),
"/US/US-KS/fort-riley-north-7573" => array("type"=>"cdp", "pop"=>12848, "name_ru"=>"Форт Рили Норт", "name_en"=>"Fort Riley North", "lang"=>"en", "lat"=>39.112072, "lon"=>-96.815965),
"/US/US-KS/fort-scott-7574" => array("type"=>"city", "pop"=>7941, "name_ru"=>"Форт Скотт", "name_en"=>"Fort Scott", "lang"=>"en", "lat"=>37.830158, "lon"=>-94.702603),
"/US/US-KS/fowler-7575" => array("type"=>"city", "pop"=>546, "name_ru"=>"Фоулер", "name_en"=>"Fowler", "lang"=>"en", "lat"=>37.383284, "lon"=>-100.195674),
"/US/US-KS/frankfort-7576" => array("type"=>"city", "pop"=>776, "name_ru"=>"Франкфорт", "name_en"=>"Frankfort", "lang"=>"en", "lat"=>39.703811, "lon"=>-96.417637),
"/US/US-KS/frederick-7577" => array("type"=>"city", "pop"=>11, "name_ru"=>"Фредерик", "name_en"=>"Frederick", "lang"=>"en", "lat"=>38.513099, "lon"=>-98.267417),
"/US/US-KS/fredonia-7578" => array("type"=>"city", "pop"=>2393, "name_ru"=>"Фредониа", "name_en"=>"Fredonia", "lang"=>"en", "lat"=>37.532912, "lon"=>-95.824031),
"/US/US-KS/freeport-7579" => array("type"=>"city", "pop"=>7, "name_ru"=>"Фрипорт", "name_en"=>"Freeport", "lang"=>"en", "lat"=>37.198656, "lon"=>-97.854166),
"/US/US-KS/frontenac-7580" => array("type"=>"city", "pop"=>3192, "name_ru"=>"Фронтенак", "name_en"=>"Frontenac", "lang"=>"en", "lat"=>37.456118, "lon"=>-94.696248),
"/US/US-KS/fulton-7581" => array("type"=>"city", "pop"=>179, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>38.009793, "lon"=>-94.719120),
"/US/US-KS/galatia-7582" => array("type"=>"city", "pop"=>59, "name_ru"=>"Галатия", "name_en"=>"Galatia", "lang"=>"en", "lat"=>38.641402, "lon"=>-98.957170),
"/US/US-KS/galena-7583" => array("type"=>"city", "pop"=>3126, "name_ru"=>"Галена", "name_en"=>"Galena", "lang"=>"en", "lat"=>37.075028, "lon"=>-94.635343),
"/US/US-KS/galesburg-7584" => array("type"=>"city", "pop"=>145, "name_ru"=>"Гейлсбург", "name_en"=>"Galesburg", "lang"=>"en", "lat"=>37.472359, "lon"=>-95.355637),
"/US/US-KS/galva-7585" => array("type"=>"city", "pop"=>804, "name_ru"=>"Галва", "name_en"=>"Galva", "lang"=>"en", "lat"=>38.382093, "lon"=>-97.537975),
"/US/US-KS/garden-city-7586" => array("type"=>"city", "pop"=>28557, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>37.977109, "lon"=>-100.862842),
"/US/US-KS/garden-plain-7587" => array("type"=>"city", "pop"=>854, "name_ru"=>"Гарден Плейн", "name_en"=>"Garden Plain", "lang"=>"en", "lat"=>37.661101, "lon"=>-97.681559),
"/US/US-KS/gardner-7588" => array("type"=>"city", "pop"=>17462, "name_ru"=>"Гарднер", "name_en"=>"Gardner", "lang"=>"en", "lat"=>38.812296, "lon"=>-94.925863),
"/US/US-KS/garfield-7589" => array("type"=>"city", "pop"=>171, "name_ru"=>"Гарфилд", "name_en"=>"Garfield", "lang"=>"en", "lat"=>38.077589, "lon"=>-99.244572),
"/US/US-KS/garnett-7590" => array("type"=>"city", "pop"=>3222, "name_ru"=>"Гарнетт", "name_en"=>"Garnett", "lang"=>"en", "lat"=>38.283639, "lon"=>-95.238552),
"/US/US-KS/gas-7591" => array("type"=>"city", "pop"=>537, "name_ru"=>"Гэс", "name_en"=>"Gas", "lang"=>"en", "lat"=>37.922707, "lon"=>-95.344704),
"/US/US-KS/gaylord-7592" => array("type"=>"city", "pop"=>120, "name_ru"=>"Гейлорд", "name_en"=>"Gaylord", "lang"=>"en", "lat"=>39.646121, "lon"=>-98.846677),
"/US/US-KS/gem-7593" => array("type"=>"city", "pop"=>87, "name_ru"=>"Джем", "name_en"=>"Gem", "lang"=>"en", "lat"=>39.425774, "lon"=>-100.896811),
"/US/US-KS/geneseo-7594" => array("type"=>"city", "pop"=>259, "name_ru"=>"Дженесео", "name_en"=>"Geneseo", "lang"=>"en", "lat"=>38.516790, "lon"=>-98.154049),
"/US/US-KS/geuda-springs-7595" => array("type"=>"city", "pop"=>191, "name_ru"=>"Джеуда Спрингс", "name_en"=>"Geuda Springs", "lang"=>"en", "lat"=>37.110760, "lon"=>-97.149359),
"/US/US-KS/girard-7596" => array("type"=>"city", "pop"=>2722, "name_ru"=>"Жирар", "name_en"=>"Girard", "lang"=>"en", "lat"=>37.510122, "lon"=>-94.842197),
"/US/US-KS/glade-7597" => array("type"=>"city", "pop"=>103, "name_ru"=>"Глейд", "name_en"=>"Glade", "lang"=>"en", "lat"=>39.682573, "lon"=>-99.310591),
"/US/US-KS/glasco-7598" => array("type"=>"city", "pop"=>488, "name_ru"=>"Гласко", "name_en"=>"Glasco", "lang"=>"en", "lat"=>39.360722, "lon"=>-97.836656),
"/US/US-KS/glen-elder-7599" => array("type"=>"city", "pop"=>395, "name_ru"=>"Глен Элдер", "name_en"=>"Glen Elder", "lang"=>"en", "lat"=>39.499704, "lon"=>-98.306176),
"/US/US-KS/goddard-7600" => array("type"=>"city", "pop"=>3869, "name_ru"=>"Годдард", "name_en"=>"Goddard", "lang"=>"en", "lat"=>37.659785, "lon"=>-97.576622),
"/US/US-KS/goessel-7601" => array("type"=>"city", "pop"=>512, "name_ru"=>"Гоессел", "name_en"=>"Goessel", "lang"=>"en", "lat"=>38.246473, "lon"=>-97.345449),
"/US/US-KS/goff-7602" => array("type"=>"city", "pop"=>164, "name_ru"=>"Гофф", "name_en"=>"Goff", "lang"=>"en", "lat"=>39.663584, "lon"=>-95.931291),
"/US/US-KS/goodland-7603" => array("type"=>"city", "pop"=>4388, "name_ru"=>"Гудленд", "name_en"=>"Goodland", "lang"=>"en", "lat"=>39.348573, "lon"=>-101.709277),
"/US/US-KS/gorham-7604" => array("type"=>"city", "pop"=>323, "name_ru"=>"Горам", "name_en"=>"Gorham", "lang"=>"en", "lat"=>38.880676, "lon"=>-99.023596),
"/US/US-KS/gove-city-7605" => array("type"=>"city", "pop"=>88, "name_ru"=>"Гоув Сити", "name_en"=>"Gove City", "lang"=>"en", "lat"=>38.959680, "lon"=>-100.489518),
"/US/US-KS/grainfield-7606" => array("type"=>"city", "pop"=>275, "name_ru"=>"Грейнфилд", "name_en"=>"Grainfield", "lang"=>"en", "lat"=>39.114117, "lon"=>-100.468644),
"/US/US-KS/grandview-plaza-7607" => array("type"=>"city", "pop"=>1382, "name_ru"=>"Грэндвью Плаза", "name_en"=>"Grandview Plaza", "lang"=>"en", "lat"=>39.031983, "lon"=>-96.794050),
"/US/US-KS/great-bend-7608" => array("type"=>"city", "pop"=>15564, "name_ru"=>"Грейт Бенд", "name_en"=>"Great Bend", "lang"=>"en", "lat"=>38.360826, "lon"=>-98.808830),
"/US/US-KS/greeley-7609" => array("type"=>"city", "pop"=>322, "name_ru"=>"Грили", "name_en"=>"Greeley", "lang"=>"en", "lat"=>38.367420, "lon"=>-95.127124),
"/US/US-KS/green-7610" => array("type"=>"city", "pop"=>139, "name_ru"=>"Грин", "name_en"=>"Green", "lang"=>"en", "lat"=>39.430270, "lon"=>-96.999970),
"/US/US-KS/greenleaf-7611" => array("type"=>"city", "pop"=>316, "name_ru"=>"Гринлиф", "name_en"=>"Greenleaf", "lang"=>"en", "lat"=>39.726757, "lon"=>-96.979731),
"/US/US-KS/greensburg-7612" => array("type"=>"city", "pop"=>1280, "name_ru"=>"Гринсбург", "name_en"=>"Greensburg", "lang"=>"en", "lat"=>37.605351, "lon"=>-99.291770),
"/US/US-KS/grenola-7613" => array("type"=>"city", "pop"=>215, "name_ru"=>"Гренола", "name_en"=>"Grenola", "lang"=>"en", "lat"=>37.350740, "lon"=>-96.448231),
"/US/US-KS/gridley-7614" => array("type"=>"city", "pop"=>350, "name_ru"=>"Гридли", "name_en"=>"Gridley", "lang"=>"en", "lat"=>38.097265, "lon"=>-95.883518),
"/US/US-KS/grinnell-7615" => array("type"=>"city", "pop"=>276, "name_ru"=>"Гриннел", "name_en"=>"Grinnell", "lang"=>"en", "lat"=>39.125937, "lon"=>-100.630654),
"/US/US-KS/gypsum-7616" => array("type"=>"city", "pop"=>404, "name_ru"=>"Джипсум", "name_en"=>"Gypsum", "lang"=>"en", "lat"=>38.705788, "lon"=>-97.426546),
"/US/US-KS/haddam-7617" => array("type"=>"city", "pop"=>151, "name_ru"=>"Хэддэм", "name_en"=>"Haddam", "lang"=>"en", "lat"=>39.855337, "lon"=>-97.303940),
"/US/US-KS/halstead-7618" => array("type"=>"city", "pop"=>1890, "name_ru"=>"Холстед", "name_en"=>"Halstead", "lang"=>"en", "lat"=>38.001031, "lon"=>-97.509308),
"/US/US-KS/hamilton-7619" => array("type"=>"city", "pop"=>303, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>37.981137, "lon"=>-96.163807),
"/US/US-KS/hamlin-7620" => array("type"=>"city", "pop"=>50, "name_ru"=>"Хэмлин", "name_en"=>"Hamlin", "lang"=>"en", "lat"=>39.915295, "lon"=>-95.627192),
"/US/US-KS/hanover-7621" => array("type"=>"city", "pop"=>573, "name_ru"=>"Хановер", "name_en"=>"Hanover", "lang"=>"en", "lat"=>39.892113, "lon"=>-96.876413),
"/US/US-KS/hanston-7622" => array("type"=>"city", "pop"=>241, "name_ru"=>"Хэнстон", "name_en"=>"Hanston", "lang"=>"en", "lat"=>38.122937, "lon"=>-99.712124),
"/US/US-KS/hardtner-7623" => array("type"=>"city", "pop"=>178, "name_ru"=>"Хардтнер", "name_en"=>"Hardtner", "lang"=>"en", "lat"=>37.014301, "lon"=>-98.648427),
"/US/US-KS/harper-7624" => array("type"=>"city", "pop"=>1416, "name_ru"=>"Харпер", "name_en"=>"Harper", "lang"=>"en", "lat"=>37.285233, "lon"=>-98.026277),
"/US/US-KS/harris-7625" => array("type"=>"city", "pop"=>39, "name_ru"=>"Харрис", "name_en"=>"Harris", "lang"=>"en", "lat"=>38.319380, "lon"=>-95.438479),
"/US/US-KS/hartford-7626" => array("type"=>"city", "pop"=>494, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>38.307871, "lon"=>-95.956889),
"/US/US-KS/harveyville-7627" => array("type"=>"city", "pop"=>253, "name_ru"=>"Харвивилл", "name_en"=>"Harveyville", "lang"=>"en", "lat"=>38.789161, "lon"=>-95.962131),
"/US/US-KS/havana-7628" => array("type"=>"city", "pop"=>84, "name_ru"=>"Гавана", "name_en"=>"Havana", "lang"=>"en", "lat"=>37.092223, "lon"=>-95.941313),
"/US/US-KS/haven-7629" => array("type"=>"city", "pop"=>1163, "name_ru"=>"Хейвен", "name_en"=>"Haven", "lang"=>"en", "lat"=>37.902295, "lon"=>-97.780630),
"/US/US-KS/havensville-7630" => array("type"=>"city", "pop"=>144, "name_ru"=>"Хейвенсвилл", "name_en"=>"Havensville", "lang"=>"en", "lat"=>39.511472, "lon"=>-96.075883),
"/US/US-KS/haviland-7631" => array("type"=>"city", "pop"=>469, "name_ru"=>"Хавиленд", "name_en"=>"Haviland", "lang"=>"en", "lat"=>37.617697, "lon"=>-99.105109),
"/US/US-KS/hays-7632" => array("type"=>"city", "pop"=>20368, "name_en"=>"Hays", "lang"=>"en", "lat"=>38.878244, "lon"=>-99.318778),
"/US/US-KS/haysville-7633" => array("type"=>"city", "pop"=>10364, "name_en"=>"Haysville", "lang"=>"en", "lat"=>37.567033, "lon"=>-97.349177),
"/US/US-KS/hazelton-7634" => array("type"=>"city", "pop"=>129, "name_ru"=>"Хейзлтон", "name_en"=>"Hazelton", "lang"=>"en", "lat"=>37.088235, "lon"=>-98.401255),
"/US/US-KS/hepler-7635" => array("type"=>"city", "pop"=>153, "name_en"=>"Hepler", "lang"=>"en", "lat"=>37.661901, "lon"=>-94.969891),
"/US/US-KS/herington-7636" => array("type"=>"city", "pop"=>2451, "name_en"=>"Herington", "lang"=>"en", "lat"=>38.673617, "lon"=>-96.945081),
"/US/US-KS/herndon-7637" => array("type"=>"city", "pop"=>128, "name_en"=>"Herndon", "lang"=>"en", "lat"=>39.908492, "lon"=>-100.785657),
"/US/US-KS/hesston-7638" => array("type"=>"city", "pop"=>3741, "name_en"=>"Hesston", "lang"=>"en", "lat"=>38.139887, "lon"=>-97.427647),
"/US/US-KS/hiawatha-7639" => array("type"=>"city", "pop"=>3169, "name_ru"=>"Хайавата", "name_en"=>"Hiawatha", "lang"=>"en", "lat"=>39.852531, "lon"=>-95.537901),
"/US/US-KS/highland-7640" => array("type"=>"city", "pop"=>945, "name_ru"=>"Хайленд", "name_en"=>"Highland", "lang"=>"en", "lat"=>39.860043, "lon"=>-95.265427),
"/US/US-KS/hill-city-7641" => array("type"=>"city", "pop"=>1392, "name_ru"=>"Хилл Сити", "name_en"=>"Hill City", "lang"=>"en", "lat"=>39.366250, "lon"=>-99.846501),
"/US/US-KS/hillsboro-7642" => array("type"=>"city", "pop"=>2638, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>38.351295, "lon"=>-97.202151),
"/US/US-KS/hoisington-7643" => array("type"=>"city", "pop"=>2883, "name_ru"=>"Хойсингтон", "name_en"=>"Hoisington", "lang"=>"en", "lat"=>38.518355, "lon"=>-98.778082),
"/US/US-KS/holcomb-7644" => array("type"=>"city", "pop"=>1991, "name_ru"=>"Холкомб", "name_en"=>"Holcomb", "lang"=>"en", "lat"=>37.990156, "lon"=>-100.988146),
"/US/US-KS/hollenberg-7645" => array("type"=>"city", "pop"=>28, "name_ru"=>"Холленберг", "name_en"=>"Hollenberg", "lang"=>"en", "lat"=>39.981213, "lon"=>-96.990926),
"/US/US-KS/holton-7646" => array("type"=>"city", "pop"=>3258, "name_ru"=>"Холтон", "name_en"=>"Holton", "lang"=>"en", "lat"=>39.470420, "lon"=>-95.730781),
"/US/US-KS/holyrood-7647" => array("type"=>"city", "pop"=>444, "name_ru"=>"Хоулируд", "name_en"=>"Holyrood", "lang"=>"en", "lat"=>38.587342, "lon"=>-98.411681),
"/US/US-KS/hope-7648" => array("type"=>"city", "pop"=>364, "name_ru"=>"Хоуп", "name_en"=>"Hope", "lang"=>"en", "lat"=>38.690949, "lon"=>-97.075075),
"/US/US-KS/horace-7649" => array("type"=>"city", "pop"=>121, "name_ru"=>"Хорейс", "name_en"=>"Horace", "lang"=>"en", "lat"=>38.476677, "lon"=>-101.790402),
"/US/US-KS/horton-7650" => array("type"=>"city", "pop"=>1799, "name_ru"=>"Хортон", "name_en"=>"Horton", "lang"=>"en", "lat"=>39.661724, "lon"=>-95.532438),
"/US/US-KS/howard-7651" => array("type"=>"city", "pop"=>757, "name_ru"=>"Ховарт", "name_en"=>"Howard", "lang"=>"en", "lat"=>37.469477, "lon"=>-96.262747),
"/US/US-KS/hoxie-7652" => array("type"=>"city", "pop"=>1115, "name_ru"=>"Хокси", "name_en"=>"Hoxie", "lang"=>"en", "lat"=>39.355332, "lon"=>-100.439549),
"/US/US-KS/hoyt-7653" => array("type"=>"city", "pop"=>569, "name_ru"=>"Хойт", "name_en"=>"Hoyt", "lang"=>"en", "lat"=>39.249619, "lon"=>-95.706490),
"/US/US-KS/hudson-7654" => array("type"=>"city", "pop"=>123, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>38.103896, "lon"=>-98.660124),
"/US/US-KS/hugoton-7655" => array("type"=>"city", "pop"=>3411, "name_ru"=>"Хаготон", "name_en"=>"Hugoton", "lang"=>"en", "lat"=>37.176354, "lon"=>-101.346822),
"/US/US-KS/humboldt-7656" => array("type"=>"city", "pop"=>1832, "name_ru"=>"Гумбольдт", "name_en"=>"Humboldt", "lang"=>"en", "lat"=>37.811624, "lon"=>-95.437181),
"/US/US-KS/hunnewell-7657" => array("type"=>"city", "pop"=>75, "name_ru"=>"Ханнуэлл", "name_en"=>"Hunnewell", "lang"=>"en", "lat"=>37.004471, "lon"=>-97.406911),
"/US/US-KS/hunter-7658" => array("type"=>"city", "pop"=>70, "name_ru"=>"Хантер", "name_en"=>"Hunter", "lang"=>"en", "lat"=>39.235596, "lon"=>-98.395613),
"/US/US-KS/huron-7659" => array("type"=>"city", "pop"=>86, "name_ru"=>"Гурон", "name_en"=>"Huron", "lang"=>"en", "lat"=>39.650545, "lon"=>-95.348403),
"/US/US-KS/hutchinson-7660" => array("type"=>"city", "pop"=>40889, "name_ru"=>"Хатчинсон", "name_en"=>"Hutchinson", "lang"=>"en", "lat"=>38.065490, "lon"=>-97.914214),
"/US/US-KS/independence-7661" => array("type"=>"city", "pop"=>9245, "name_ru"=>"Индепенденс", "name_en"=>"Independence", "lang"=>"en", "lat"=>37.231522, "lon"=>-95.713847),
"/US/US-KS/ingalls-7662" => array("type"=>"city", "pop"=>299, "name_ru"=>"Инголс", "name_en"=>"Ingalls", "lang"=>"en", "lat"=>37.827935, "lon"=>-100.452368),
"/US/US-KS/inman-7663" => array("type"=>"city", "pop"=>1187, "name_ru"=>"Инман", "name_en"=>"Inman", "lang"=>"en", "lat"=>38.229123, "lon"=>-97.773761),
"/US/US-KS/iola-7664" => array("type"=>"city", "pop"=>5782, "name_ru"=>"Айола", "name_en"=>"Iola", "lang"=>"en", "lat"=>37.928176, "lon"=>-95.400968),
"/US/US-KS/isabel-7665" => array("type"=>"city", "pop"=>97, "name_ru"=>"Изабел", "name_en"=>"Isabel", "lang"=>"en", "lat"=>37.467728, "lon"=>-98.551374),
"/US/US-KS/iuka-7666" => array("type"=>"city", "pop"=>181, "name_ru"=>"Иука", "name_en"=>"Iuka", "lang"=>"en", "lat"=>37.728994, "lon"=>-98.731441),
"/US/US-KS/jamestown-7667" => array("type"=>"city", "pop"=>377, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>39.600026, "lon"=>-97.861439),
"/US/US-KS/jennings-7668" => array("type"=>"city", "pop"=>118, "name_ru"=>"Дженнингс", "name_en"=>"Jennings", "lang"=>"en", "lat"=>39.680193, "lon"=>-100.293502),
"/US/US-KS/jetmore-7669" => array("type"=>"city", "pop"=>846, "name_ru"=>"Джетмор", "name_en"=>"Jetmore", "lang"=>"en", "lat"=>38.083088, "lon"=>-99.893002),
"/US/US-KS/jewell-7670" => array("type"=>"city", "pop"=>405, "name_ru"=>"Джевелл", "name_en"=>"Jewell", "lang"=>"en", "lat"=>39.671835, "lon"=>-98.151291),
"/US/US-KS/johnson-city-7671" => array("type"=>"city", "pop"=>1354, "name_ru"=>"Джонсон Сити", "name_en"=>"Johnson City", "lang"=>"en", "lat"=>37.569607, "lon"=>-101.752452),
"/US/US-KS/junction-city-7672" => array("type"=>"city", "pop"=>20671, "name_ru"=>"Джанкшен Сити", "name_en"=>"Junction City", "lang"=>"en", "lat"=>39.028561, "lon"=>-96.841979),
"/US/US-KS/kanopolis-7673" => array("type"=>"city", "pop"=>506, "name_ru"=>"Канополис", "name_en"=>"Kanopolis", "lang"=>"en", "lat"=>38.708524, "lon"=>-98.156704),
"/US/US-KS/kanorado-7674" => array("type"=>"city", "pop"=>222, "name_ru"=>"Канорадо", "name_en"=>"Kanorado", "lang"=>"en", "lat"=>39.337064, "lon"=>-102.036914),
"/US/US-KS/kansas-city-7675" => array("type"=>"city", "pop"=>142562, "name_ru"=>"Канзас Сити", "name_en"=>"Kansas City", "lang"=>"en", "lat"=>39.118400, "lon"=>-94.727037),
"/US/US-KS/kechi-7676" => array("type"=>"city", "pop"=>1753, "name_ru"=>"Кечи", "name_en"=>"Kechi", "lang"=>"en", "lat"=>37.795350, "lon"=>-97.278294),
"/US/US-KS/kensington-7677" => array("type"=>"city", "pop"=>464, "name_ru"=>"Кенсингтон", "name_en"=>"Kensington", "lang"=>"en", "lat"=>39.767393, "lon"=>-99.031961),
"/US/US-KS/kincaid-7678" => array("type"=>"city", "pop"=>175, "name_ru"=>"Кинкейд", "name_en"=>"Kincaid", "lang"=>"en", "lat"=>38.081535, "lon"=>-95.154688),
"/US/US-KS/kingman-7679" => array("type"=>"city", "pop"=>2996, "name_ru"=>"Кингман", "name_en"=>"Kingman", "lang"=>"en", "lat"=>37.647009, "lon"=>-98.113465),
"/US/US-KS/kinsley-7680" => array("type"=>"city", "pop"=>1449, "name_ru"=>"Кинсли", "name_en"=>"Kinsley", "lang"=>"en", "lat"=>37.922368, "lon"=>-99.411133),
"/US/US-KS/kiowa-7681" => array("type"=>"city", "pop"=>908, "name_ru"=>"Кайова", "name_en"=>"Kiowa", "lang"=>"en", "lat"=>37.017493, "lon"=>-98.484368),
"/US/US-KS/kirwin-7682" => array("type"=>"city", "pop"=>207, "name_ru"=>"Кирвин", "name_en"=>"Kirwin", "lang"=>"en", "lat"=>39.669821, "lon"=>-99.122777),
"/US/US-KS/kismet-7683" => array("type"=>"city", "pop"=>505, "name_ru"=>"Кисмет", "name_en"=>"Kismet", "lang"=>"en", "lat"=>37.204999, "lon"=>-100.700574),
"/US/US-KS/la-crosse-7685" => array("type"=>"city", "pop"=>1241, "name_ru"=>"Ла Кросс", "name_en"=>"La Crosse", "lang"=>"en", "lat"=>38.531727, "lon"=>-99.309792),
"/US/US-KS/la-cygne-7686" => array("type"=>"city", "pop"=>1108, "name_ru"=>"Ла Сигн", "name_en"=>"La Cygne", "lang"=>"en", "lat"=>38.346811, "lon"=>-94.761346),
"/US/US-KS/la-harpe-7687" => array("type"=>"city", "pop"=>654, "name_ru"=>"Ла Харп", "name_en"=>"La Harpe", "lang"=>"en", "lat"=>37.916499, "lon"=>-95.301944),
"/US/US-KS/labette-7684" => array("type"=>"city", "pop"=>67, "name_ru"=>"Лабетт", "name_en"=>"Labette", "lang"=>"en", "lat"=>37.230328, "lon"=>-95.183548),
"/US/US-KS/lake-quivira-7688" => array("type"=>"city", "pop"=>941, "name_ru"=>"Лейк Квивайра", "name_en"=>"Lake Quivira", "lang"=>"en", "lat"=>39.040450, "lon"=>-94.768039),
"/US/US-KS/lakin-7689" => array("type"=>"city", "pop"=>2123, "name_ru"=>"Лакин", "name_en"=>"Lakin", "lang"=>"en", "lat"=>37.940183, "lon"=>-101.258281),
"/US/US-KS/lancaster-7690" => array("type"=>"city", "pop"=>289, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>39.571782, "lon"=>-95.303402),
"/US/US-KS/lane-7691" => array("type"=>"city", "pop"=>253, "name_ru"=>"Лейн", "name_en"=>"Lane", "lang"=>"en", "lat"=>38.440258, "lon"=>-95.080842),
"/US/US-KS/langdon-7692" => array("type"=>"city", "pop"=>71, "name_ru"=>"Лэнгдон", "name_en"=>"Langdon", "lang"=>"en", "lat"=>37.853334, "lon"=>-98.324148),
"/US/US-KS/lansing-7693" => array("type"=>"city", "pop"=>10691, "name_ru"=>"Ленсинг", "name_en"=>"Lansing", "lang"=>"en", "lat"=>39.248224, "lon"=>-94.887465),
"/US/US-KS/larned-7694" => array("type"=>"city", "pop"=>3599, "name_ru"=>"Ларнд", "name_en"=>"Larned", "lang"=>"en", "lat"=>38.183161, "lon"=>-99.101326),
"/US/US-KS/latham-7695" => array("type"=>"city", "pop"=>162, "name_ru"=>"Латам", "name_en"=>"Latham", "lang"=>"en", "lat"=>37.535173, "lon"=>-96.641483),
"/US/US-KS/latimer-7696" => array("type"=>"city", "pop"=>21, "name_ru"=>"Латимер", "name_en"=>"Latimer", "lang"=>"en", "lat"=>38.739040, "lon"=>-96.845770),
"/US/US-KS/lawrence-7697" => array("type"=>"city", "pop"=>90520, "name_ru"=>"Лоренс", "name_en"=>"Lawrence", "lang"=>"en", "lat"=>38.962850, "lon"=>-95.255405),
"/US/US-KS/le-roy-7710" => array("type"=>"city", "pop"=>568, "name_ru"=>"Ле Рой", "name_en"=>"Le Roy", "lang"=>"en", "lat"=>38.086093, "lon"=>-95.632897),
"/US/US-KS/leavenworth-7698" => array("type"=>"city", "pop"=>34729, "name_ru"=>"Ливенуорт", "name_en"=>"Leavenworth", "lang"=>"en", "lat"=>39.324560, "lon"=>-94.923211),
"/US/US-KS/leawood-7699" => array("type"=>"city", "pop"=>31342, "name_ru"=>"Ливуд", "name_en"=>"Leawood", "lang"=>"en", "lat"=>38.907124, "lon"=>-94.625201),
"/US/US-KS/lebanon-7700" => array("type"=>"city", "pop"=>262, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>39.810491, "lon"=>-98.557047),
"/US/US-KS/lebo-7701" => array("type"=>"city", "pop"=>915, "name_ru"=>"Лебо", "name_en"=>"Lebo", "lang"=>"en", "lat"=>38.413839, "lon"=>-95.856370),
"/US/US-KS/lecompton-7702" => array("type"=>"city", "pop"=>666, "name_ru"=>"Лекомптон", "name_en"=>"Lecompton", "lang"=>"en", "lat"=>39.045169, "lon"=>-95.392003),
"/US/US-KS/lehigh-7703" => array("type"=>"city", "pop"=>190, "name_ru"=>"Лихай", "name_en"=>"Lehigh", "lang"=>"en", "lat"=>38.374891, "lon"=>-97.302415),
"/US/US-KS/lenexa-7704" => array("type"=>"city", "pop"=>46822, "name_ru"=>"Ленекса", "name_en"=>"Lenexa", "lang"=>"en", "lat"=>38.963536, "lon"=>-94.788457),
"/US/US-KS/lenora-7705" => array("type"=>"city", "pop"=>266, "name_ru"=>"Ленора", "name_en"=>"Lenora", "lang"=>"en", "lat"=>39.610971, "lon"=>-100.000987),
"/US/US-KS/leon-7706" => array("type"=>"city", "pop"=>637, "name_ru"=>"Леон", "name_en"=>"Leon", "lang"=>"en", "lat"=>37.688245, "lon"=>-96.783535),
"/US/US-KS/leona-7707" => array("type"=>"city", "pop"=>83, "name_ru"=>"Леона", "name_en"=>"Leona", "lang"=>"en", "lat"=>39.785798, "lon"=>-95.321904),
"/US/US-KS/leonardville-7708" => array("type"=>"city", "pop"=>475, "name_ru"=>"Леонардвилл", "name_en"=>"Leonardville", "lang"=>"en", "lat"=>39.363436, "lon"=>-96.857883),
"/US/US-KS/leoti-7709" => array("type"=>"city", "pop"=>1339, "name_ru"=>"Леоти", "name_en"=>"Leoti", "lang"=>"en", "lat"=>38.483026, "lon"=>-101.357187),
"/US/US-KS/lewis-7711" => array("type"=>"city", "pop"=>444, "name_ru"=>"Льюис", "name_en"=>"Lewis", "lang"=>"en", "lat"=>37.936999, "lon"=>-99.254340),
"/US/US-KS/liberal-7712" => array("type"=>"city", "pop"=>20074, "name_ru"=>"Либерал", "name_en"=>"Liberal", "lang"=>"en", "lat"=>37.044721, "lon"=>-100.935528),
"/US/US-KS/liberty-7713" => array("type"=>"city", "pop"=>92, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>37.156139, "lon"=>-95.597745),
"/US/US-KS/liebenthal-7714" => array("type"=>"city", "pop"=>102, "name_ru"=>"Либентал", "name_en"=>"Liebenthal", "lang"=>"en", "lat"=>38.654456, "lon"=>-99.320014),
"/US/US-KS/lincoln-center-7715" => array("type"=>"city", "pop"=>1213, "name_ru"=>"Линколн Сентер", "name_en"=>"Lincoln Center", "lang"=>"en", "lat"=>39.042655, "lon"=>-98.148173),
"/US/US-KS/lincolnville-7716" => array("type"=>"city", "pop"=>203, "name_ru"=>"Линколнвилл", "name_en"=>"Lincolnville", "lang"=>"en", "lat"=>38.493978, "lon"=>-96.961240),
"/US/US-KS/lindsborg-7717" => array("type"=>"city", "pop"=>3245, "name_ru"=>"Линдсборг", "name_en"=>"Lindsborg", "lang"=>"en", "lat"=>38.574268, "lon"=>-97.674868),
"/US/US-KS/linn-7718" => array("type"=>"city", "pop"=>375, "name_ru"=>"Линн", "name_en"=>"Linn", "lang"=>"en", "lat"=>39.679380, "lon"=>-97.086394),
"/US/US-KS/linwood-7719" => array("type"=>"city", "pop"=>388, "name_ru"=>"Линвуд", "name_en"=>"Linwood", "lang"=>"en", "lat"=>38.999957, "lon"=>-95.034671),
"/US/US-KS/little-river-7720" => array("type"=>"city", "pop"=>518, "name_ru"=>"Литтл Ривер", "name_en"=>"Little River", "lang"=>"en", "lat"=>38.397678, "lon"=>-98.012418),
"/US/US-KS/logan-7721" => array("type"=>"city", "pop"=>532, "name_ru"=>"Логан", "name_en"=>"Logan", "lang"=>"en", "lat"=>39.661267, "lon"=>-99.566504),
"/US/US-KS/lone-elm-7722" => array("type"=>"city", "pop"=>26, "name_ru"=>"Лоун Элм", "name_en"=>"Lone Elm", "lang"=>"en", "lat"=>38.079721, "lon"=>-95.242260),
"/US/US-KS/long-island-7724" => array("type"=>"city", "pop"=>140, "name_ru"=>"Лонг Айленд", "name_en"=>"Long Island", "lang"=>"en", "lat"=>39.946255, "lon"=>-99.532901),
"/US/US-KS/longford-7723" => array("type"=>"city", "pop"=>91, "name_ru"=>"Лонгфорд", "name_en"=>"Longford", "lang"=>"en", "lat"=>39.172574, "lon"=>-97.328135),
"/US/US-KS/longton-7725" => array("type"=>"city", "pop"=>367, "name_ru"=>"Лонгтон", "name_en"=>"Longton", "lang"=>"en", "lat"=>37.377453, "lon"=>-96.082566),
"/US/US-KS/lorraine-7726" => array("type"=>"city", "pop"=>130, "name_ru"=>"Лоррейн", "name_en"=>"Lorraine", "lang"=>"en", "lat"=>38.569613, "lon"=>-98.317127),
"/US/US-KS/lost-springs-7727" => array("type"=>"city", "pop"=>62, "name_ru"=>"Лост Спрингс", "name_en"=>"Lost Springs", "lang"=>"en", "lat"=>38.566484, "lon"=>-96.964932),
"/US/US-KS/louisburg-7728" => array("type"=>"city", "pop"=>3940, "name_ru"=>"Луисбург", "name_en"=>"Louisburg", "lang"=>"en", "lat"=>38.623768, "lon"=>-94.682900),
"/US/US-KS/louisville-7729" => array("type"=>"city", "pop"=>207, "name_ru"=>"Луисвилл", "name_en"=>"Louisville", "lang"=>"en", "lat"=>39.250877, "lon"=>-96.315108),
"/US/US-KS/lucas-7730" => array("type"=>"city", "pop"=>409, "name_ru"=>"Лукас", "name_en"=>"Lucas", "lang"=>"en", "lat"=>39.057502, "lon"=>-98.538479),
"/US/US-KS/luray-7731" => array("type"=>"city", "pop"=>180, "name_ru"=>"Лурей", "name_en"=>"Luray", "lang"=>"en", "lat"=>39.115012, "lon"=>-98.692391),
"/US/US-KS/lyndon-7732" => array("type"=>"city", "pop"=>1008, "name_ru"=>"Линдон", "name_en"=>"Lyndon", "lang"=>"en", "lat"=>38.610224, "lon"=>-95.685090),
"/US/US-KS/lyons-7733" => array("type"=>"city", "pop"=>3399, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>38.345959, "lon"=>-98.202162),
"/US/US-KS/macksville-7738" => array("type"=>"city", "pop"=>476, "name_ru"=>"Мэксвилл", "name_en"=>"Macksville", "lang"=>"en", "lat"=>37.957201, "lon"=>-98.968298),
"/US/US-KS/madison-7741" => array("type"=>"city", "pop"=>750, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>38.132993, "lon"=>-96.136979),
"/US/US-KS/mahaska-7742" => array("type"=>"city", "pop"=>96, "name_ru"=>"Махаска", "name_en"=>"Mahaska", "lang"=>"en", "lat"=>39.988235, "lon"=>-97.352732),
"/US/US-KS/maize-7743" => array("type"=>"city", "pop"=>3094, "name_ru"=>"Мейз", "name_en"=>"Maize", "lang"=>"en", "lat"=>37.773100, "lon"=>-97.464149),
"/US/US-KS/manchester-7744" => array("type"=>"city", "pop"=>102, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>39.091647, "lon"=>-97.320428),
"/US/US-KS/manhattan-7745" => array("type"=>"city", "pop"=>52284, "name_ru"=>"Манхэттен", "name_en"=>"Manhattan", "lang"=>"en", "lat"=>39.190100, "lon"=>-96.589981),
"/US/US-KS/mankato-7746" => array("type"=>"city", "pop"=>794, "name_ru"=>"Манкато", "name_en"=>"Mankato", "lang"=>"en", "lat"=>39.787433, "lon"=>-98.207633),
"/US/US-KS/manter-7747" => array("type"=>"city", "pop"=>161, "name_ru"=>"Мантер", "name_en"=>"Manter", "lang"=>"en", "lat"=>37.524199, "lon"=>-101.882696),
"/US/US-KS/maple-hill-7748" => array("type"=>"city", "pop"=>509, "name_ru"=>"Мейпл Хилл", "name_en"=>"Maple Hill", "lang"=>"en", "lat"=>39.086181, "lon"=>-96.026801),
"/US/US-KS/mapleton-7749" => array("type"=>"city", "pop"=>96, "name_ru"=>"Мейплтон", "name_en"=>"Mapleton", "lang"=>"en", "lat"=>38.015555, "lon"=>-94.882517),
"/US/US-KS/marion-7750" => array("type"=>"city", "pop"=>1880, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>38.348937, "lon"=>-97.015737),
"/US/US-KS/marquette-7751" => array("type"=>"city", "pop"=>583, "name_ru"=>"Маркетт", "name_en"=>"Marquette", "lang"=>"en", "lat"=>38.553806, "lon"=>-97.834039),
"/US/US-KS/marysville-7752" => array("type"=>"city", "pop"=>3126, "name_ru"=>"Мэрисвилл", "name_en"=>"Marysville", "lang"=>"en", "lat"=>39.846658, "lon"=>-96.642273),
"/US/US-KS/matfield-green-7753" => array("type"=>"city", "pop"=>57, "name_ru"=>"Мэтфилд Грин", "name_en"=>"Matfield Green", "lang"=>"en", "lat"=>38.158655, "lon"=>-96.563448),
"/US/US-KS/mayetta-7754" => array("type"=>"city", "pop"=>351, "name_ru"=>"Майетта", "name_en"=>"Mayetta", "lang"=>"en", "lat"=>39.338757, "lon"=>-95.721666),
"/US/US-KS/mayfield-7755" => array("type"=>"city", "pop"=>103, "name_ru"=>"Мейфилд", "name_en"=>"Mayfield", "lang"=>"en", "lat"=>37.262337, "lon"=>-97.545460),
"/US/US-KS/mccracken-7734" => array("type"=>"city", "pop"=>192, "name_ru"=>"Мак-Крэкен", "name_en"=>"McCracken", "lang"=>"en", "lat"=>38.582316, "lon"=>-99.568907),
"/US/US-KS/mccune-7735" => array("type"=>"city", "pop"=>423, "name_ru"=>"Маккан", "name_en"=>"McCune", "lang"=>"en", "lat"=>37.353855, "lon"=>-95.019090),
"/US/US-KS/mcdonald-7736" => array("type"=>"city", "pop"=>129, "name_ru"=>"Макдональд", "name_en"=>"McDonald", "lang"=>"en", "lat"=>39.785200, "lon"=>-101.370290),
"/US/US-KS/mcfarland-7737" => array("type"=>"city", "pop"=>270, "name_ru"=>"Мак-Фарленд", "name_en"=>"McFarland", "lang"=>"en", "lat"=>39.054536, "lon"=>-96.238027),
"/US/US-KS/mclouth-7739" => array("type"=>"city", "pop"=>835, "name_ru"=>"Мак-Лаут", "name_en"=>"McLouth", "lang"=>"en", "lat"=>39.195305, "lon"=>-95.208462),
"/US/US-KS/mcpherson-7740" => array("type"=>"city", "pop"=>13396, "name_ru"=>"Мак-Ферсон", "name_en"=>"McPherson", "lang"=>"en", "lat"=>38.371915, "lon"=>-97.661858),
"/US/US-KS/meade-7756" => array("type"=>"city", "pop"=>1538, "name_ru"=>"Мид", "name_en"=>"Meade", "lang"=>"en", "lat"=>37.285020, "lon"=>-100.337566),
"/US/US-KS/medicine-lodge-7757" => array("type"=>"city", "pop"=>1917, "name_ru"=>"Медисин Лодж", "name_en"=>"Medicine Lodge", "lang"=>"en", "lat"=>37.285987, "lon"=>-98.580304),
"/US/US-KS/melvern-7758" => array("type"=>"city", "pop"=>411, "name_ru"=>"Мэлверн", "name_en"=>"Melvern", "lang"=>"en", "lat"=>38.507433, "lon"=>-95.638071),
"/US/US-KS/menlo-7759" => array("type"=>"city", "pop"=>52, "name_ru"=>"Менло", "name_en"=>"Menlo", "lang"=>"en", "lat"=>39.356112, "lon"=>-100.723915),
"/US/US-KS/meriden-7760" => array("type"=>"city", "pop"=>755, "name_ru"=>"Мериден", "name_en"=>"Meriden", "lang"=>"en", "lat"=>39.190048, "lon"=>-95.566380),
"/US/US-KS/merriam-7761" => array("type"=>"city", "pop"=>10814, "name_ru"=>"Мерриам", "name_en"=>"Merriam", "lang"=>"en", "lat"=>39.019200, "lon"=>-94.691755),
"/US/US-KS/milan-7762" => array("type"=>"city", "pop"=>124, "name_ru"=>"Милан", "name_en"=>"Milan", "lang"=>"en", "lat"=>37.258066, "lon"=>-97.673046),
"/US/US-KS/mildred-7763" => array("type"=>"city", "pop"=>35, "name_ru"=>"Милдред", "name_en"=>"Mildred", "lang"=>"en", "lat"=>38.024774, "lon"=>-95.173800),
"/US/US-KS/milford-7764" => array("type"=>"city", "pop"=>426, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>39.174300, "lon"=>-96.912279),
"/US/US-KS/miltonvale-7765" => array("type"=>"city", "pop"=>470, "name_ru"=>"Милтонвейл", "name_en"=>"Miltonvale", "lang"=>"en", "lat"=>39.350662, "lon"=>-97.452882),
"/US/US-KS/minneapolis-7766" => array("type"=>"city", "pop"=>1978, "name_ru"=>"Миннеаполис", "name_en"=>"Minneapolis", "lang"=>"en", "lat"=>39.124731, "lon"=>-97.702056),
"/US/US-KS/minneola-7767" => array("type"=>"city", "pop"=>638, "name_ru"=>"Миннеола", "name_en"=>"Minneola", "lang"=>"en", "lat"=>37.442726, "lon"=>-100.012636),
"/US/US-KS/mission-7768" => array("type"=>"city", "pop"=>9765, "name_ru"=>"Мишен", "name_en"=>"Mission", "lang"=>"en", "lat"=>39.026700, "lon"=>-94.656699),
"/US/US-KS/mission-hills-7769" => array("type"=>"city", "pop"=>3568, "name_ru"=>"Мишен Хилс", "name_en"=>"Mission Hills", "lang"=>"en", "lat"=>39.013550, "lon"=>-94.616374),
"/US/US-KS/mission-woods-7770" => array("type"=>"city", "pop"=>160, "name_ru"=>"Мишен Вудс", "name_en"=>"Mission Woods", "lang"=>"en", "lat"=>39.033183, "lon"=>-94.612769),
"/US/US-KS/moline-7771" => array("type"=>"city", "pop"=>426, "name_ru"=>"Молайн", "name_en"=>"Moline", "lang"=>"en", "lat"=>37.363732, "lon"=>-96.301810),
"/US/US-KS/montezuma-7772" => array("type"=>"city", "pop"=>941, "name_ru"=>"Мотезума", "name_en"=>"Montezuma", "lang"=>"en", "lat"=>37.596452, "lon"=>-100.441750),
"/US/US-KS/moran-7773" => array("type"=>"city", "pop"=>522, "name_ru"=>"Моран", "name_en"=>"Moran", "lang"=>"en", "lat"=>37.916132, "lon"=>-95.170464),
"/US/US-KS/morganville-7774" => array("type"=>"city", "pop"=>198, "name_ru"=>"Морганвилл", "name_en"=>"Morganville", "lang"=>"en", "lat"=>39.466186, "lon"=>-97.203068),
"/US/US-KS/morland-7775" => array("type"=>"city", "pop"=>146, "name_ru"=>"Морленд", "name_en"=>"Morland", "lang"=>"en", "lat"=>39.348824, "lon"=>-100.074414),
"/US/US-KS/morrill-7776" => array("type"=>"city", "pop"=>247, "name_ru"=>"Моррил", "name_en"=>"Morrill", "lang"=>"en", "lat"=>39.929123, "lon"=>-95.694256),
"/US/US-KS/morrowville-7777" => array("type"=>"city", "pop"=>151, "name_ru"=>"Морроувилл", "name_en"=>"Morrowville", "lang"=>"en", "lat"=>39.844867, "lon"=>-97.172042),
"/US/US-KS/moscow-7778" => array("type"=>"city", "pop"=>232, "name_ru"=>"Москва", "name_en"=>"Moscow", "lang"=>"en", "lat"=>37.325468, "lon"=>-101.206340),
"/US/US-KS/mound-city-7779" => array("type"=>"city", "pop"=>791, "name_ru"=>"Маунд Сити", "name_en"=>"Mound City", "lang"=>"en", "lat"=>38.143757, "lon"=>-94.822412),
"/US/US-KS/mound-valley-7781" => array("type"=>"city", "pop"=>407, "name_ru"=>"Маунд Валли", "name_en"=>"Mound Valley", "lang"=>"en", "lat"=>37.207153, "lon"=>-95.403971),
"/US/US-KS/moundridge-7780" => array("type"=>"city", "pop"=>1636, "name_ru"=>"Маундридж", "name_en"=>"Moundridge", "lang"=>"en", "lat"=>38.201754, "lon"=>-97.515603),
"/US/US-KS/mount-hope-7782" => array("type"=>"city", "pop"=>857, "name_ru"=>"Маунт Хоуп", "name_en"=>"Mount Hope", "lang"=>"en", "lat"=>37.868354, "lon"=>-97.664233),
"/US/US-KS/mulberry-7783" => array("type"=>"city", "pop"=>567, "name_ru"=>"Малберри", "name_en"=>"Mulberry", "lang"=>"en", "lat"=>37.555909, "lon"=>-94.623171),
"/US/US-KS/mullinville-7784" => array("type"=>"city", "pop"=>202, "name_ru"=>"Муллинвилл", "name_en"=>"Mullinville", "lang"=>"en", "lat"=>37.588015, "lon"=>-99.474567),
"/US/US-KS/mulvane-7785" => array("type"=>"city", "pop"=>5866, "name_ru"=>"Малвейн", "name_en"=>"Mulvane", "lang"=>"en", "lat"=>37.480723, "lon"=>-97.240702),
"/US/US-KS/munden-7786" => array("type"=>"city", "pop"=>101, "name_ru"=>"Манден", "name_en"=>"Munden", "lang"=>"en", "lat"=>39.912473, "lon"=>-97.538383),
"/US/US-KS/muscotah-7787" => array("type"=>"city", "pop"=>197, "name_ru"=>"Маскота", "name_en"=>"Muscotah", "lang"=>"en", "lat"=>39.553326, "lon"=>-95.520475),
"/US/US-KS/narka-7788" => array("type"=>"city", "pop"=>77, "name_ru"=>"Нарка", "name_en"=>"Narka", "lang"=>"en", "lat"=>39.960225, "lon"=>-97.427034),
"/US/US-KS/nashville-7789" => array("type"=>"city", "pop"=>100, "name_ru"=>"Нэшвилл", "name_en"=>"Nashville", "lang"=>"en", "lat"=>37.438573, "lon"=>-98.422339),
"/US/US-KS/natoma-7790" => array("type"=>"city", "pop"=>308, "name_ru"=>"Натома", "name_en"=>"Natoma", "lang"=>"en", "lat"=>39.189044, "lon"=>-99.024465),
"/US/US-KS/neodesha-7791" => array("type"=>"city", "pop"=>2633, "name_ru"=>"Неодеша", "name_en"=>"Neodesha", "lang"=>"en", "lat"=>37.423453, "lon"=>-95.680927),
"/US/US-KS/neosho-falls-7792" => array("type"=>"city", "pop"=>159, "name_ru"=>"Неошо Фоллс", "name_en"=>"Neosho Falls", "lang"=>"en", "lat"=>38.005845, "lon"=>-95.555196),
"/US/US-KS/neosho-rapids-7793" => array("type"=>"city", "pop"=>271, "name_ru"=>"Неошо Рэпидс", "name_en"=>"Neosho Rapids", "lang"=>"en", "lat"=>38.368701, "lon"=>-95.991633),
"/US/US-KS/ness-city-7794" => array("type"=>"city", "pop"=>1261, "name_ru"=>"Несс Сити", "name_en"=>"Ness City", "lang"=>"en", "lat"=>38.453410, "lon"=>-99.905627),
"/US/US-KS/netawaka-7795" => array("type"=>"city", "pop"=>180, "name_ru"=>"Нетавака", "name_en"=>"Netawaka", "lang"=>"en", "lat"=>39.603242, "lon"=>-95.718492),
"/US/US-KS/new-albany-7796" => array("type"=>"city", "pop"=>70, "name_ru"=>"Нью Олбани", "name_en"=>"New Albany", "lang"=>"en", "lat"=>37.567686, "lon"=>-95.934488),
"/US/US-KS/new-cambria-7797" => array("type"=>"city", "pop"=>159, "name_ru"=>"Нью Камбрия", "name_en"=>"New Cambria", "lang"=>"en", "lat"=>38.879638, "lon"=>-97.505427),
"/US/US-KS/new-strawn-7798" => array("type"=>"city", "pop"=>393, "name_ru"=>"Нью Строн", "name_en"=>"New Strawn", "lang"=>"en", "lat"=>38.261340, "lon"=>-95.741307),
"/US/US-KS/newton-7799" => array("type"=>"city", "pop"=>18133, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>38.044387, "lon"=>-97.341909),
"/US/US-KS/nickerson-7800" => array("type"=>"city", "pop"=>1144, "name_ru"=>"Никерсон", "name_en"=>"Nickerson", "lang"=>"en", "lat"=>38.147715, "lon"=>-98.086729),
"/US/US-KS/niotaze-7801" => array("type"=>"city", "pop"=>105, "name_ru"=>"Ниотаз", "name_en"=>"Niotaze", "lang"=>"en", "lat"=>37.067275, "lon"=>-96.013969),
"/US/US-KS/norcatur-7802" => array("type"=>"city", "pop"=>144, "name_ru"=>"Норкатур", "name_en"=>"Norcatur", "lang"=>"en", "lat"=>39.834588, "lon"=>-100.187997),
"/US/US-KS/north-newton-7803" => array("type"=>"city", "pop"=>1592, "name_ru"=>"Норт Ньютон", "name_en"=>"North Newton", "lang"=>"en", "lat"=>38.073854, "lon"=>-97.346042),
"/US/US-KS/norton-7804" => array("type"=>"city", "pop"=>2651, "name_ru"=>"Нортон", "name_en"=>"Norton", "lang"=>"en", "lat"=>39.835542, "lon"=>-99.891833),
"/US/US-KS/nortonville-7805" => array("type"=>"city", "pop"=>585, "name_ru"=>"Нортонвилл", "name_en"=>"Nortonville", "lang"=>"en", "lat"=>39.415943, "lon"=>-95.331142),
"/US/US-KS/norwich-7806" => array("type"=>"city", "pop"=>499, "name_ru"=>"Норвич", "name_en"=>"Norwich", "lang"=>"en", "lat"=>37.456731, "lon"=>-97.845500),
"/US/US-KS/oak-hill-7807" => array("type"=>"city", "pop"=>35, "name_ru"=>"Оук Хилл", "name_en"=>"Oak Hill", "lang"=>"en", "lat"=>39.246471, "lon"=>-97.343337),
"/US/US-KS/oaklawn-sunview-7808" => array("type"=>"cdp", "pop"=>3240, "name_ru"=>"Оуклон-Санвью", "name_en"=>"Oaklawn-Sunview", "lang"=>"en", "lat"=>37.606750, "lon"=>-97.298791),
"/US/US-KS/oakley-7809" => array("type"=>"city", "pop"=>1845, "name_ru"=>"Оукли", "name_en"=>"Oakley", "lang"=>"en", "lat"=>39.125843, "lon"=>-100.854079),
"/US/US-KS/oberlin-7810" => array("type"=>"city", "pop"=>1656, "name_ru"=>"Оберлин", "name_en"=>"Oberlin", "lang"=>"en", "lat"=>39.823233, "lon"=>-100.530422),
"/US/US-KS/offerle-7811" => array("type"=>"city", "pop"=>201, "name_ru"=>"Офферл", "name_en"=>"Offerle", "lang"=>"en", "lat"=>37.890853, "lon"=>-99.560494),
"/US/US-KS/ogden-7812" => array("type"=>"city", "pop"=>2152, "name_ru"=>"Огден", "name_en"=>"Ogden", "lang"=>"en", "lat"=>39.111877, "lon"=>-96.707213),
"/US/US-KS/oketo-7813" => array("type"=>"city", "pop"=>81, "name_ru"=>"Окето", "name_en"=>"Oketo", "lang"=>"en", "lat"=>39.962880, "lon"=>-96.598598),
"/US/US-KS/olathe-7814" => array("type"=>"city", "pop"=>119993, "name_ru"=>"Олэт", "name_en"=>"Olathe", "lang"=>"en", "lat"=>38.886850, "lon"=>-94.814499),
"/US/US-KS/olivet-7815" => array("type"=>"city", "pop"=>61, "name_ru"=>"Оливет", "name_en"=>"Olivet", "lang"=>"en", "lat"=>38.481000, "lon"=>-95.752002),
"/US/US-KS/olmitz-7816" => array("type"=>"city", "pop"=>133, "name_ru"=>"Олмиц", "name_en"=>"Olmitz", "lang"=>"en", "lat"=>38.516448, "lon"=>-98.937018),
"/US/US-KS/olpe-7817" => array("type"=>"city", "pop"=>508, "name_ru"=>"Олп", "name_en"=>"Olpe", "lang"=>"en", "lat"=>38.262052, "lon"=>-96.167085),
"/US/US-KS/olsburg-7818" => array("type"=>"city", "pop"=>194, "name_ru"=>"Олсбург", "name_en"=>"Olsburg", "lang"=>"en", "lat"=>39.432051, "lon"=>-96.615433),
"/US/US-KS/onaga-7819" => array("type"=>"city", "pop"=>675, "name_ru"=>"Онага", "name_en"=>"Onaga", "lang"=>"en", "lat"=>39.489386, "lon"=>-96.170281),
"/US/US-KS/oneida-7820" => array("type"=>"city", "pop"=>67, "name_ru"=>"Онида", "name_en"=>"Oneida", "lang"=>"en", "lat"=>39.867995, "lon"=>-95.939796),
"/US/US-KS/osage-city-7821" => array("type"=>"city", "pop"=>2824, "name_ru"=>"Осейдж Сити", "name_en"=>"Osage City", "lang"=>"en", "lat"=>38.632883, "lon"=>-95.820880),
"/US/US-KS/osawatomie-7822" => array("type"=>"city", "pop"=>4488, "name_ru"=>"Осаватоми", "name_en"=>"Osawatomie", "lang"=>"en", "lat"=>38.502492, "lon"=>-94.951058),
"/US/US-KS/osborne-7823" => array("type"=>"city", "pop"=>1354, "name_ru"=>"Осборн", "name_en"=>"Osborne", "lang"=>"en", "lat"=>39.439439, "lon"=>-98.698289),
"/US/US-KS/oskaloosa-7824" => array("type"=>"city", "pop"=>1152, "name_ru"=>"Оскалуза", "name_en"=>"Oskaloosa", "lang"=>"en", "lat"=>39.215850, "lon"=>-95.313551),
"/US/US-KS/oswego-7825" => array("type"=>"city", "pop"=>1979, "name_ru"=>"Освего", "name_en"=>"Oswego", "lang"=>"en", "lat"=>37.167639, "lon"=>-95.111849),
"/US/US-KS/otis-7826" => array("type"=>"city", "pop"=>302, "name_ru"=>"Отис", "name_en"=>"Otis", "lang"=>"en", "lat"=>38.534921, "lon"=>-99.053064),
"/US/US-KS/ottawa-7827" => array("type"=>"city", "pop"=>12850, "name_ru"=>"Оттава", "name_en"=>"Ottawa", "lang"=>"en", "lat"=>38.609127, "lon"=>-95.266263),
"/US/US-KS/overbrook-7828" => array("type"=>"city", "pop"=>928, "name_ru"=>"Овербрук", "name_en"=>"Overbrook", "lang"=>"en", "lat"=>38.779964, "lon"=>-95.556944),
"/US/US-KS/overland-park-7829" => array("type"=>"city", "pop"=>171231, "name_ru"=>"Оверленд Парк", "name_en"=>"Overland Park", "lang"=>"en", "lat"=>38.914592, "lon"=>-94.684916),
"/US/US-KS/oxford-7830" => array("type"=>"city", "pop"=>1068, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>37.274817, "lon"=>-97.169507),
"/US/US-KS/ozawkie-7831" => array("type"=>"city", "pop"=>560, "name_ru"=>"Озавки", "name_en"=>"Ozawkie", "lang"=>"en", "lat"=>39.235848, "lon"=>-95.467276),
"/US/US-KS/palco-7832" => array("type"=>"city", "pop"=>217, "name_ru"=>"Палко", "name_en"=>"Palco", "lang"=>"en", "lat"=>39.253948, "lon"=>-99.564161),
"/US/US-KS/palmer-7833" => array("type"=>"city", "pop"=>97, "name_ru"=>"Палмер", "name_en"=>"Palmer", "lang"=>"en", "lat"=>39.633734, "lon"=>-97.139308),
"/US/US-KS/paola-7834" => array("type"=>"city", "pop"=>5351, "name_ru"=>"Паола", "name_en"=>"Paola", "lang"=>"en", "lat"=>38.576986, "lon"=>-94.866243),
"/US/US-KS/paradise-7835" => array("type"=>"city", "pop"=>59, "name_ru"=>"Парадайс", "name_en"=>"Paradise", "lang"=>"en", "lat"=>39.114553, "lon"=>-98.917564),
"/US/US-KS/park-7836" => array("type"=>"city", "pop"=>127, "name_ru"=>"Парк", "name_en"=>"Park", "lang"=>"en", "lat"=>39.112292, "lon"=>-100.361658),
"/US/US-KS/park-city-7837" => array("type"=>"city", "pop"=>7787, "name_ru"=>"Парк Сити", "name_en"=>"Park City", "lang"=>"en", "lat"=>37.795964, "lon"=>-97.321908),
"/US/US-KS/parker-7838" => array("type"=>"city", "pop"=>310, "name_ru"=>"Паркер", "name_en"=>"Parker", "lang"=>"en", "lat"=>38.328854, "lon"=>-94.990550),
"/US/US-KS/parkerville-7839" => array("type"=>"city", "pop"=>72, "name_ru"=>"Паркервилл", "name_en"=>"Parkerville", "lang"=>"en", "lat"=>38.763862, "lon"=>-96.662445),
"/US/US-KS/parsons-7840" => array("type"=>"city", "pop"=>11065, "name_ru"=>"Парсонс", "name_en"=>"Parsons", "lang"=>"en", "lat"=>37.339018, "lon"=>-95.267879),
"/US/US-KS/partridge-7841" => array("type"=>"city", "pop"=>250, "name_ru"=>"Партридж", "name_en"=>"Partridge", "lang"=>"en", "lat"=>37.967364, "lon"=>-98.093870),
"/US/US-KS/pawnee-rock-7842" => array("type"=>"city", "pop"=>329, "name_ru"=>"Пони Рок", "name_en"=>"Pawnee Rock", "lang"=>"en", "lat"=>38.265270, "lon"=>-98.982241),
"/US/US-KS/paxico-7843" => array("type"=>"city", "pop"=>209, "name_ru"=>"Паксико", "name_en"=>"Paxico", "lang"=>"en", "lat"=>39.069410, "lon"=>-96.167109),
"/US/US-KS/peabody-7844" => array("type"=>"city", "pop"=>1203, "name_ru"=>"Пибоди", "name_en"=>"Peabody", "lang"=>"en", "lat"=>38.168778, "lon"=>-97.106864),
"/US/US-KS/penalosa-7845" => array("type"=>"city", "pop"=>24, "name_ru"=>"Пеналоса", "name_en"=>"Penalosa", "lang"=>"en", "lat"=>37.715830, "lon"=>-98.319735),
"/US/US-KS/perry-7846" => array("type"=>"city", "pop"=>845, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>39.073179, "lon"=>-95.387383),
"/US/US-KS/peru-7847" => array("type"=>"city", "pop"=>158, "name_ru"=>"Перу", "name_en"=>"Peru", "lang"=>"en", "lat"=>37.081239, "lon"=>-96.096013),
"/US/US-KS/phillipsburg-7848" => array("type"=>"city", "pop"=>2367, "name_ru"=>"Филлипсбург", "name_en"=>"Phillipsburg", "lang"=>"en", "lat"=>39.751409, "lon"=>-99.321935),
"/US/US-KS/pittsburg-7849" => array("type"=>"city", "pop"=>19649, "name_ru"=>"Питсбург", "name_en"=>"Pittsburg", "lang"=>"en", "lat"=>37.413516, "lon"=>-94.699564),
"/US/US-KS/plains-city-7850" => array("type"=>"city", "pop"=>957, "name_ru"=>"Плейнс Сити", "name_en"=>"Plains City", "lang"=>"en", "lat"=>37.263642, "lon"=>-100.589897),
"/US/US-KS/plainville-7851" => array("type"=>"city", "pop"=>1821, "name_ru"=>"Плейнвилл", "name_en"=>"Plainville", "lang"=>"en", "lat"=>39.234151, "lon"=>-99.301345),
"/US/US-KS/pleasanton-7852" => array("type"=>"city", "pop"=>1316, "name_ru"=>"Плезантон", "name_en"=>"Pleasanton", "lang"=>"en", "lat"=>38.175212, "lon"=>-94.713011),
"/US/US-KS/plevna-7853" => array("type"=>"city", "pop"=>97, "name_ru"=>"Плевна", "name_en"=>"Plevna", "lang"=>"en", "lat"=>37.971123, "lon"=>-98.309115),
"/US/US-KS/pomona-7854" => array("type"=>"city", "pop"=>945, "name_ru"=>"Помона", "name_en"=>"Pomona", "lang"=>"en", "lat"=>38.610983, "lon"=>-95.451796),
"/US/US-KS/portis-7855" => array("type"=>"city", "pop"=>108, "name_ru"=>"Портис", "name_en"=>"Portis", "lang"=>"en", "lat"=>39.563249, "lon"=>-98.691036),
"/US/US-KS/potwin-7856" => array("type"=>"city", "pop"=>432, "name_ru"=>"Потвин", "name_en"=>"Potwin", "lang"=>"en", "lat"=>37.938881, "lon"=>-97.018723),
"/US/US-KS/powhattan-7857" => array("type"=>"city", "pop"=>85, "name_ru"=>"Похаттан", "name_en"=>"Powhattan", "lang"=>"en", "lat"=>39.762086, "lon"=>-95.633392),
"/US/US-KS/prairie-view-7858" => array("type"=>"city", "pop"=>127, "name_ru"=>"Прейри Вью", "name_en"=>"Prairie View", "lang"=>"en", "lat"=>39.831729, "lon"=>-99.572939),
"/US/US-KS/prairie-village-7859" => array("type"=>"city", "pop"=>21479, "name_ru"=>"Прейри Виллидж", "name_en"=>"Prairie Village", "lang"=>"en", "lat"=>38.986700, "lon"=>-94.635899),
"/US/US-KS/pratt-7860" => array("type"=>"city", "pop"=>6397, "name_ru"=>"Прэтт", "name_en"=>"Pratt", "lang"=>"en", "lat"=>37.646663, "lon"=>-98.735913),
"/US/US-KS/prescott-7861" => array("type"=>"city", "pop"=>271, "name_ru"=>"Прескотт", "name_en"=>"Prescott", "lang"=>"en", "lat"=>38.063031, "lon"=>-94.696103),
"/US/US-KS/preston-7862" => array("type"=>"city", "pop"=>160, "name_ru"=>"Престон", "name_en"=>"Preston", "lang"=>"en", "lat"=>37.757972, "lon"=>-98.555492),
"/US/US-KS/pretty-prairie-7863" => array("type"=>"city", "pop"=>598, "name_ru"=>"Притти Прейри", "name_en"=>"Pretty Prairie", "lang"=>"en", "lat"=>37.780231, "lon"=>-98.019735),
"/US/US-KS/princeton-7864" => array("type"=>"city", "pop"=>334, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>38.488614, "lon"=>-95.272498),
"/US/US-KS/protection-7865" => array("type"=>"city", "pop"=>547, "name_ru"=>"Протекшен", "name_en"=>"Protection", "lang"=>"en", "lat"=>37.200159, "lon"=>-99.480345),
"/US/US-KS/quenemo-7866" => array("type"=>"city", "pop"=>424, "name_ru"=>"Квинмо", "name_en"=>"Quenemo", "lang"=>"en", "lat"=>38.580402, "lon"=>-95.526304),
"/US/US-KS/quinter-7867" => array("type"=>"city", "pop"=>782, "name_ru"=>"Квинтер", "name_en"=>"Quinter", "lang"=>"en", "lat"=>39.066569, "lon"=>-100.233676),
"/US/US-KS/radium-7868" => array("type"=>"city", "pop"=>37, "name_ru"=>"Радиум", "name_en"=>"Radium", "lang"=>"en", "lat"=>38.173686, "lon"=>-98.893860),
"/US/US-KS/ramona-7869" => array("type"=>"city", "pop"=>82, "name_ru"=>"Рамона", "name_en"=>"Ramona", "lang"=>"en", "lat"=>38.597459, "lon"=>-97.063605),
"/US/US-KS/randall-7870" => array("type"=>"city", "pop"=>70, "name_ru"=>"Рэндолл", "name_en"=>"Randall", "lang"=>"en", "lat"=>39.641643, "lon"=>-98.045110),
"/US/US-KS/randolph-7871" => array("type"=>"city", "pop"=>198, "name_ru"=>"Рэндолф", "name_en"=>"Randolph", "lang"=>"en", "lat"=>39.430057, "lon"=>-96.759270),
"/US/US-KS/ransom-7872" => array("type"=>"city", "pop"=>279, "name_ru"=>"Рэнсам", "name_en"=>"Ransom", "lang"=>"en", "lat"=>38.636349, "lon"=>-99.931990),
"/US/US-KS/rantoul-7873" => array("type"=>"city", "pop"=>239, "name_ru"=>"Рантул", "name_en"=>"Rantoul", "lang"=>"en", "lat"=>38.548529, "lon"=>-95.100175),
"/US/US-KS/raymond-7874" => array("type"=>"city", "pop"=>90, "name_ru"=>"Реймонд", "name_en"=>"Raymond", "lang"=>"en", "lat"=>38.278453, "lon"=>-98.414530),
"/US/US-KS/reading-7875" => array("type"=>"city", "pop"=>244, "name_ru"=>"Ридинг", "name_en"=>"Reading", "lang"=>"en", "lat"=>38.519091, "lon"=>-95.957497),
"/US/US-KS/redfield-7876" => array("type"=>"city", "pop"=>136, "name_ru"=>"Редфилд", "name_en"=>"Redfield", "lang"=>"en", "lat"=>37.836261, "lon"=>-94.881953),
"/US/US-KS/republic-7877" => array("type"=>"city", "pop"=>133, "name_ru"=>"Репаблик", "name_en"=>"Republic", "lang"=>"en", "lat"=>39.923649, "lon"=>-97.824387),
"/US/US-KS/reserve-7878" => array("type"=>"city", "pop"=>94, "name_ru"=>"Резерв", "name_en"=>"Reserve", "lang"=>"en", "lat"=>39.977291, "lon"=>-95.565429),
"/US/US-KS/rexford-7879" => array("type"=>"city", "pop"=>142, "name_ru"=>"Рексфорд", "name_en"=>"Rexford", "lang"=>"en", "lat"=>39.470508, "lon"=>-100.743550),
"/US/US-KS/richfield-7880" => array("type"=>"city", "pop"=>41, "name_ru"=>"Ричфилд", "name_en"=>"Richfield", "lang"=>"en", "lat"=>37.265715, "lon"=>-101.783235),
"/US/US-KS/richmond-7881" => array("type"=>"city", "pop"=>503, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>38.401214, "lon"=>-95.253957),
"/US/US-KS/riley-7882" => array("type"=>"city", "pop"=>1026, "name_ru"=>"Райли", "name_en"=>"Riley", "lang"=>"en", "lat"=>39.298246, "lon"=>-96.828367),
"/US/US-KS/robinson-7883" => array("type"=>"city", "pop"=>193, "name_ru"=>"Робинсон", "name_en"=>"Robinson", "lang"=>"en", "lat"=>39.814775, "lon"=>-95.410958),
"/US/US-KS/roeland-park-7884" => array("type"=>"city", "pop"=>6960, "name_ru"=>"Роленд Парк", "name_en"=>"Roeland Park", "lang"=>"en", "lat"=>39.035850, "lon"=>-94.637488),
"/US/US-KS/rolla-7885" => array("type"=>"city", "pop"=>413, "name_ru"=>"Ролла", "name_en"=>"Rolla", "lang"=>"en", "lat"=>37.119342, "lon"=>-101.630602),
"/US/US-KS/rose-hill-7886" => array("type"=>"city", "pop"=>4034, "name_ru"=>"Роуз Хилл", "name_en"=>"Rose Hill", "lang"=>"en", "lat"=>37.566208, "lon"=>-97.132071),
"/US/US-KS/roseland-7887" => array("type"=>"city", "pop"=>94, "name_ru"=>"Роузлэнд", "name_en"=>"Roseland", "lang"=>"en", "lat"=>37.279367, "lon"=>-94.851257),
"/US/US-KS/rossville-7888" => array("type"=>"city", "pop"=>1114, "name_ru"=>"Россвилл", "name_en"=>"Rossville", "lang"=>"en", "lat"=>39.135978, "lon"=>-95.949707),
"/US/US-KS/rozel-7889" => array("type"=>"city", "pop"=>156, "name_ru"=>"Розел", "name_en"=>"Rozel", "lang"=>"en", "lat"=>38.195856, "lon"=>-99.402053),
"/US/US-KS/rush-center-7890" => array("type"=>"city", "pop"=>164, "name_ru"=>"Раш Сентер", "name_en"=>"Rush Center", "lang"=>"en", "lat"=>38.464868, "lon"=>-99.310520),
"/US/US-KS/russell-7891" => array("type"=>"city", "pop"=>4217, "name_ru"=>"Расселл", "name_en"=>"Russell", "lang"=>"en", "lat"=>38.887920, "lon"=>-98.853485),
"/US/US-KS/russell-springs-7892" => array("type"=>"city", "pop"=>28, "name_ru"=>"Рассел Спрингс", "name_en"=>"Russell Springs", "lang"=>"en", "lat"=>38.912296, "lon"=>-101.175495),
"/US/US-KS/sabetha-7893" => array("type"=>"city", "pop"=>2482, "name_ru"=>"Сабета", "name_en"=>"Sabetha", "lang"=>"en", "lat"=>39.903097, "lon"=>-95.794503),
"/US/US-KS/salina-7899" => array("type"=>"city", "pop"=>46483, "name_ru"=>"Салина", "name_en"=>"Salina", "lang"=>"en", "lat"=>38.815788, "lon"=>-97.617762),
"/US/US-KS/satanta-7900" => array("type"=>"city", "pop"=>1090, "name_ru"=>"Сатанта", "name_en"=>"Satanta", "lang"=>"en", "lat"=>37.435868, "lon"=>-100.987942),
"/US/US-KS/savonburg-7901" => array("type"=>"city", "pop"=>86, "name_ru"=>"Сейвонбург", "name_en"=>"Savonburg", "lang"=>"en", "lat"=>37.749130, "lon"=>-95.141682),
"/US/US-KS/sawyer-7902" => array("type"=>"city", "pop"=>121, "name_ru"=>"Сойер", "name_en"=>"Sawyer", "lang"=>"en", "lat"=>37.497728, "lon"=>-98.682685),
"/US/US-KS/scammon-7903" => array("type"=>"city", "pop"=>463, "name_ru"=>"Скаммон", "name_en"=>"Scammon", "lang"=>"en", "lat"=>37.278263, "lon"=>-94.822050),
"/US/US-KS/scandia-7904" => array("type"=>"city", "pop"=>343, "name_ru"=>"Скандия", "name_en"=>"Scandia", "lang"=>"en", "lat"=>39.797204, "lon"=>-97.783452),
"/US/US-KS/schoenchen-7905" => array("type"=>"city", "pop"=>217, "name_ru"=>"Скинчен", "name_en"=>"Schoenchen", "lang"=>"en", "lat"=>38.713232, "lon"=>-99.330379),
"/US/US-KS/scott-city-7906" => array("type"=>"city", "pop"=>3502, "name_ru"=>"Скотт Сити", "name_en"=>"Scott City", "lang"=>"en", "lat"=>38.479791, "lon"=>-100.908398),
"/US/US-KS/scottsville-7907" => array("type"=>"city", "pop"=>19, "name_ru"=>"Скоттсвилл", "name_en"=>"Scottsville", "lang"=>"en", "lat"=>39.543052, "lon"=>-97.952034),
"/US/US-KS/scranton-7908" => array("type"=>"city", "pop"=>681, "name_ru"=>"Скрантон", "name_en"=>"Scranton", "lang"=>"en", "lat"=>38.777929, "lon"=>-95.740785),
"/US/US-KS/sedan-7909" => array("type"=>"city", "pop"=>1168, "name_ru"=>"Седан", "name_en"=>"Sedan", "lang"=>"en", "lat"=>37.127368, "lon"=>-96.184694),
"/US/US-KS/sedgwick-7910" => array("type"=>"city", "pop"=>1667, "name_ru"=>"Седжвик", "name_en"=>"Sedgwick", "lang"=>"en", "lat"=>37.917763, "lon"=>-97.426145),
"/US/US-KS/selden-7911" => array("type"=>"city", "pop"=>178, "name_ru"=>"Селден", "name_en"=>"Selden", "lang"=>"en", "lat"=>39.541191, "lon"=>-100.566643),
"/US/US-KS/seneca-7912" => array("type"=>"city", "pop"=>2013, "name_ru"=>"Сенека", "name_en"=>"Seneca", "lang"=>"en", "lat"=>39.835695, "lon"=>-96.065776),
"/US/US-KS/severance-7913" => array("type"=>"city", "pop"=>102, "name_ru"=>"Северанс", "name_en"=>"Severance", "lang"=>"en", "lat"=>39.766862, "lon"=>-95.248849),
"/US/US-KS/severy-7914" => array("type"=>"city", "pop"=>328, "name_ru"=>"Севери", "name_en"=>"Severy", "lang"=>"en", "lat"=>37.621559, "lon"=>-96.227338),
"/US/US-KS/seward-7915" => array("type"=>"city", "pop"=>58, "name_ru"=>"Сьюард", "name_en"=>"Seward", "lang"=>"en", "lat"=>38.177580, "lon"=>-98.793787),
"/US/US-KS/sharon-7916" => array("type"=>"city", "pop"=>189, "name_ru"=>"Шэрон", "name_en"=>"Sharon", "lang"=>"en", "lat"=>37.249970, "lon"=>-98.418260),
"/US/US-KS/sharon-springs-7917" => array("type"=>"city", "pop"=>657, "name_ru"=>"Шэрон Спрингс", "name_en"=>"Sharon Springs", "lang"=>"en", "lat"=>38.894247, "lon"=>-101.752140),
"/US/US-KS/shawnee-7918" => array("type"=>"city", "pop"=>60954, "name_ru"=>"Шони", "name_en"=>"Shawnee", "lang"=>"en", "lat"=>39.015236, "lon"=>-94.804453),
"/US/US-KS/silver-lake-7919" => array("type"=>"city", "pop"=>1380, "name_ru"=>"Сильвер Лэйк", "name_en"=>"Silver Lake", "lang"=>"en", "lat"=>39.098640, "lon"=>-95.856016),
"/US/US-KS/simpson-7920" => array("type"=>"city", "pop"=>100, "name_ru"=>"Симпсон", "name_en"=>"Simpson", "lang"=>"en", "lat"=>39.386296, "lon"=>-97.933720),
"/US/US-KS/smith-center-7921" => array("type"=>"city", "pop"=>1644, "name_ru"=>"Смит Сентер", "name_en"=>"Smith Center", "lang"=>"en", "lat"=>39.779717, "lon"=>-98.783135),
"/US/US-KS/smolan-7922" => array("type"=>"city", "pop"=>213, "name_ru"=>"Смолан", "name_en"=>"Smolan", "lang"=>"en", "lat"=>38.738005, "lon"=>-97.683625),
"/US/US-KS/soldier-7923" => array("type"=>"city", "pop"=>119, "name_ru"=>"Солджер", "name_en"=>"Soldier", "lang"=>"en", "lat"=>39.537837, "lon"=>-95.964602),
"/US/US-KS/solomon-7924" => array("type"=>"city", "pop"=>1078, "name_ru"=>"Соломон", "name_en"=>"Solomon", "lang"=>"en", "lat"=>38.919712, "lon"=>-97.371420),
"/US/US-KS/south-haven-7925" => array("type"=>"city", "pop"=>350, "name_ru"=>"Саут Хейвен", "name_en"=>"South Haven", "lang"=>"en", "lat"=>37.049130, "lon"=>-97.400318),
"/US/US-KS/south-hutchinson-7926" => array("type"=>"city", "pop"=>2547, "name_ru"=>"Саут Хатчинсон", "name_en"=>"South Hutchinson", "lang"=>"en", "lat"=>38.028422, "lon"=>-97.941044),
"/US/US-KS/spearville-7927" => array("type"=>"city", "pop"=>861, "name_ru"=>"Спирвилл", "name_en"=>"Spearville", "lang"=>"en", "lat"=>37.847975, "lon"=>-99.754257),
"/US/US-KS/speed-7928" => array("type"=>"city", "pop"=>40, "name_ru"=>"Спид", "name_en"=>"Speed", "lang"=>"en", "lat"=>39.676331, "lon"=>-99.420253),
"/US/US-KS/spivey-7929" => array("type"=>"city", "pop"=>72, "name_ru"=>"Спайви", "name_en"=>"Spivey", "lang"=>"en", "lat"=>37.448229, "lon"=>-98.164962),
"/US/US-KS/spring-hill-7930" => array("type"=>"city", "pop"=>5227, "name_ru"=>"Спринг Хилл", "name_en"=>"Spring Hill", "lang"=>"en", "lat"=>38.751723, "lon"=>-94.827412),
"/US/US-KS/st.-francis-7894" => array("type"=>"city", "pop"=>1282, "name_ru"=>"Сент-Фрэнсис", "name_en"=>"St. Francis", "lang"=>"en", "lat"=>39.771397, "lon"=>-101.800265),
"/US/US-KS/st.-george-7895" => array("type"=>"city", "pop"=>562, "name_ru"=>"Сент-Джрдж", "name_en"=>"St. George", "lang"=>"en", "lat"=>39.190088, "lon"=>-96.417255),
"/US/US-KS/st.-john-7896" => array("type"=>"city", "pop"=>1173, "name_ru"=>"Сент Джон", "name_en"=>"St. John", "lang"=>"en", "lat"=>38.001015, "lon"=>-98.760679),
"/US/US-KS/st.-marys-7897" => array("type"=>"city", "pop"=>2277, "name_ru"=>"Сент Мэрис", "name_en"=>"St. Marys", "lang"=>"en", "lat"=>39.193978, "lon"=>-96.065308),
"/US/US-KS/st.-paul-7898" => array("type"=>"city", "pop"=>663, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>37.517342, "lon"=>-95.174842),
"/US/US-KS/stafford-7931" => array("type"=>"city", "pop"=>1028, "name_ru"=>"Стэффорд", "name_en"=>"Stafford", "lang"=>"en", "lat"=>37.962774, "lon"=>-98.599300),
"/US/US-KS/stark-7932" => array("type"=>"city", "pop"=>103, "name_ru"=>"Старк", "name_en"=>"Stark", "lang"=>"en", "lat"=>37.689547, "lon"=>-95.143319),
"/US/US-KS/sterling-7933" => array("type"=>"city", "pop"=>2535, "name_ru"=>"Стерлинг", "name_en"=>"Sterling", "lang"=>"en", "lat"=>38.209159, "lon"=>-98.205783),
"/US/US-KS/stockton-7934" => array("type"=>"city", "pop"=>1402, "name_ru"=>"Стоктон", "name_en"=>"Stockton", "lang"=>"en", "lat"=>39.436319, "lon"=>-99.271267),
"/US/US-KS/strong-city-7935" => array("type"=>"city", "pop"=>527, "name_ru"=>"Стронг Сити", "name_en"=>"Strong City", "lang"=>"en", "lat"=>38.395645, "lon"=>-96.536462),
"/US/US-KS/sublette-7936" => array("type"=>"city", "pop"=>1468, "name_ru"=>"Саблетт", "name_en"=>"Sublette", "lang"=>"en", "lat"=>37.479625, "lon"=>-100.846525),
"/US/US-KS/summerfield-7937" => array("type"=>"city", "pop"=>198, "name_ru"=>"Саммерфилд", "name_en"=>"Summerfield", "lang"=>"en", "lat"=>39.996439, "lon"=>-96.348958),
"/US/US-KS/sun-city-7938" => array("type"=>"city", "pop"=>72, "name_ru"=>"Сан Сити", "name_en"=>"Sun City", "lang"=>"en", "lat"=>37.378510, "lon"=>-98.916440),
"/US/US-KS/susank-7939" => array("type"=>"city", "pop"=>55, "name_ru"=>"Сасанк", "name_en"=>"Susank", "lang"=>"en", "lat"=>38.640390, "lon"=>-98.774566),
"/US/US-KS/sylvan-grove-7940" => array("type"=>"city", "pop"=>286, "name_ru"=>"Сильван Гров", "name_en"=>"Sylvan Grove", "lang"=>"en", "lat"=>39.012794, "lon"=>-98.393516),
"/US/US-KS/sylvia-7941" => array("type"=>"city", "pop"=>292, "name_ru"=>"Сильвия", "name_en"=>"Sylvia", "lang"=>"en", "lat"=>37.959897, "lon"=>-98.407804),
"/US/US-KS/syracuse-7942" => array("type"=>"city", "pop"=>1768, "name_ru"=>"Сиракьюс", "name_en"=>"Syracuse", "lang"=>"en", "lat"=>37.981118, "lon"=>-101.751443),
"/US/US-KS/tampa-7943" => array("type"=>"city", "pop"=>129, "name_ru"=>"Тампа", "name_en"=>"Tampa", "lang"=>"en", "lat"=>38.546984, "lon"=>-97.153976),
"/US/US-KS/tescott-7944" => array("type"=>"city", "pop"=>326, "name_ru"=>"Тескотт", "name_en"=>"Tescott", "lang"=>"en", "lat"=>39.011792, "lon"=>-97.877357),
"/US/US-KS/thayer-7945" => array("type"=>"city", "pop"=>487, "name_ru"=>"Тайер", "name_en"=>"Thayer", "lang"=>"en", "lat"=>37.487921, "lon"=>-95.473987),
"/US/US-KS/timken-7946" => array("type"=>"city", "pop"=>77, "name_ru"=>"Тимкен", "name_en"=>"Timken", "lang"=>"en", "lat"=>38.472877, "lon"=>-99.176998),
"/US/US-KS/tipton-7947" => array("type"=>"city", "pop"=>226, "name_ru"=>"Типтон", "name_en"=>"Tipton", "lang"=>"en", "lat"=>39.339371, "lon"=>-98.471068),
"/US/US-KS/tonganoxie-7948" => array("type"=>"city", "pop"=>4305, "name_ru"=>"Тонганокси", "name_en"=>"Tonganoxie", "lang"=>"en", "lat"=>39.110745, "lon"=>-95.084159),
"/US/US-KS/topeka-7949" => array("type"=>"city", "pop"=>123446, "name_ru"=>"Топека", "name_en"=>"Topeka", "lang"=>"en", "lat"=>39.037900, "lon"=>-95.691999),
"/US/US-KS/toronto-7950" => array("type"=>"city", "pop"=>265, "name_ru"=>"Торонто", "name_en"=>"Toronto", "lang"=>"en", "lat"=>37.798562, "lon"=>-95.949288),
"/US/US-KS/towanda-7951" => array("type"=>"city", "pop"=>1359, "name_ru"=>"Тованда", "name_en"=>"Towanda", "lang"=>"en", "lat"=>37.796852, "lon"=>-96.997017),
"/US/US-KS/treece-7952" => array("type"=>"city", "pop"=>139, "name_ru"=>"Трис", "name_en"=>"Treece", "lang"=>"en", "lat"=>37.000195, "lon"=>-94.843494),
"/US/US-KS/tribune-7953" => array("type"=>"city", "pop"=>678, "name_ru"=>"Трибьюн", "name_en"=>"Tribune", "lang"=>"en", "lat"=>38.471325, "lon"=>-101.754047),
"/US/US-KS/troy-7954" => array("type"=>"city", "pop"=>1015, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>39.786319, "lon"=>-95.088584),
"/US/US-KS/turon-7955" => array("type"=>"city", "pop"=>428, "name_ru"=>"Турон", "name_en"=>"Turon", "lang"=>"en", "lat"=>37.807400, "lon"=>-98.428050),
"/US/US-KS/tyro-7956" => array("type"=>"city", "pop"=>220, "name_ru"=>"Тайро", "name_en"=>"Tyro", "lang"=>"en", "lat"=>37.036697, "lon"=>-95.821415),
"/US/US-KS/udall-7957" => array("type"=>"city", "pop"=>743, "name_ru"=>"Удолл", "name_en"=>"Udall", "lang"=>"en", "lat"=>37.386969, "lon"=>-97.117491),
"/US/US-KS/ulysses-7958" => array("type"=>"city", "pop"=>5557, "name_ru"=>"Улисс", "name_en"=>"Ulysses", "lang"=>"en", "lat"=>37.579018, "lon"=>-101.355783),
"/US/US-KS/uniontown-7959" => array("type"=>"city", "pop"=>276, "name_ru"=>"Юнионтаун", "name_en"=>"Uniontown", "lang"=>"en", "lat"=>37.847238, "lon"=>-94.976042),
"/US/US-KS/utica-7960" => array("type"=>"city", "pop"=>188, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>38.642908, "lon"=>-100.169895),
"/US/US-KS/valley-center-7961" => array("type"=>"city", "pop"=>6521, "name_ru"=>"Валли Сентер", "name_en"=>"Valley Center", "lang"=>"en", "lat"=>37.835200, "lon"=>-97.373472),
"/US/US-KS/valley-falls-7962" => array("type"=>"city", "pop"=>1156, "name_ru"=>"Валли Фоллс", "name_en"=>"Valley Falls", "lang"=>"en", "lat"=>39.341371, "lon"=>-95.460316),
"/US/US-KS/vermillion-7963" => array("type"=>"city", "pop"=>95, "name_ru"=>"Вермиллион", "name_en"=>"Vermillion", "lang"=>"en", "lat"=>39.717844, "lon"=>-96.264930),
"/US/US-KS/victoria-7964" => array("type"=>"city", "pop"=>1204, "name_ru"=>"Виктория", "name_en"=>"Victoria", "lang"=>"en", "lat"=>38.853597, "lon"=>-99.147082),
"/US/US-KS/vining-7965" => array("type"=>"city", "pop"=>56, "name_ru"=>"Вининг", "name_en"=>"Vining", "lang"=>"en", "lat"=>39.567462, "lon"=>-97.292801),
"/US/US-KS/viola-7966" => array("type"=>"city", "pop"=>205, "name_ru"=>"Виола", "name_en"=>"Viola", "lang"=>"en", "lat"=>37.482455, "lon"=>-97.644563),
"/US/US-KS/virgil-7967" => array("type"=>"city", "pop"=>102, "name_ru"=>"Верджил", "name_en"=>"Virgil", "lang"=>"en", "lat"=>37.980575, "lon"=>-96.010820),
"/US/US-KS/wakeeney-7968" => array("type"=>"city", "pop"=>1716, "name_ru"=>"Уэйкини", "name_en"=>"Wakeeney", "lang"=>"en", "lat"=>39.023425, "lon"=>-99.881339),
"/US/US-KS/wakefield-7969" => array("type"=>"city", "pop"=>870, "name_ru"=>"Уэйкфилд", "name_en"=>"Wakefield", "lang"=>"en", "lat"=>39.216713, "lon"=>-97.017837),
"/US/US-KS/waldo-7970" => array("type"=>"city", "pop"=>44, "name_ru"=>"Уолдо", "name_en"=>"Waldo", "lang"=>"en", "lat"=>39.120150, "lon"=>-98.797499),
"/US/US-KS/waldron-7971" => array("type"=>"city", "pop"=>15, "name_ru"=>"Уолдрон", "name_en"=>"Waldron", "lang"=>"en", "lat"=>37.000757, "lon"=>-98.182805),
"/US/US-KS/wallace-7972" => array("type"=>"city", "pop"=>55, "name_ru"=>"Уоллэс", "name_en"=>"Wallace", "lang"=>"en", "lat"=>38.913687, "lon"=>-101.592435),
"/US/US-KS/walnut-7973" => array("type"=>"city", "pop"=>219, "name_ru"=>"Уолнат", "name_en"=>"Walnut", "lang"=>"en", "lat"=>37.600311, "lon"=>-95.074494),
"/US/US-KS/walton-7974" => array("type"=>"city", "pop"=>286, "name_ru"=>"Уолтон", "name_en"=>"Walton", "lang"=>"en", "lat"=>38.118276, "lon"=>-97.256082),
"/US/US-KS/wamego-7975" => array("type"=>"city", "pop"=>4312, "name_ru"=>"Вамего", "name_en"=>"Wamego", "lang"=>"en", "lat"=>39.204078, "lon"=>-96.308053),
"/US/US-KS/washington-7976" => array("type"=>"city", "pop"=>1105, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>39.816874, "lon"=>-97.052198),
"/US/US-KS/waterville-7977" => array("type"=>"city", "pop"=>616, "name_ru"=>"Уотервилл", "name_en"=>"Waterville", "lang"=>"en", "lat"=>39.691921, "lon"=>-96.748117),
"/US/US-KS/wathena-7978" => array("type"=>"city", "pop"=>1296, "name_ru"=>"Уотена", "name_en"=>"Wathena", "lang"=>"en", "lat"=>39.764296, "lon"=>-94.949091),
"/US/US-KS/waverly-7979" => array("type"=>"city", "pop"=>548, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>38.397279, "lon"=>-95.603310),
"/US/US-KS/webber-7980" => array("type"=>"city", "pop"=>31, "name_ru"=>"Вэббер", "name_en"=>"Webber", "lang"=>"en", "lat"=>39.934402, "lon"=>-98.035843),
"/US/US-KS/weir-7981" => array("type"=>"city", "pop"=>703, "name_ru"=>"Вейр", "name_en"=>"Weir", "lang"=>"en", "lat"=>37.309571, "lon"=>-94.775678),
"/US/US-KS/wellington-7982" => array("type"=>"city", "pop"=>7735, "name_ru"=>"Веллингтон", "name_en"=>"Wellington", "lang"=>"en", "lat"=>37.271284, "lon"=>-97.397201),
"/US/US-KS/wellsville-7983" => array("type"=>"city", "pop"=>1737, "name_ru"=>"Уэллсвилл", "name_en"=>"Wellsville", "lang"=>"en", "lat"=>38.717478, "lon"=>-95.081104),
"/US/US-KS/west-mineral-7984" => array("type"=>"city", "pop"=>228, "name_ru"=>"Вэст Минерал", "name_en"=>"West Mineral", "lang"=>"en", "lat"=>37.283346, "lon"=>-94.927007),
"/US/US-KS/westmoreland-7985" => array("type"=>"city", "pop"=>746, "name_ru"=>"Вэстморленд", "name_en"=>"Westmoreland", "lang"=>"en", "lat"=>39.394774, "lon"=>-96.413874),
"/US/US-KS/westphalia-7986" => array("type"=>"city", "pop"=>162, "name_ru"=>"Вэстфалия", "name_en"=>"Westphalia", "lang"=>"en", "lat"=>38.182333, "lon"=>-95.490439),
"/US/US-KS/westwood-7987" => array("type"=>"city", "pop"=>1481, "name_ru"=>"Уэствуд", "name_en"=>"Westwood", "lang"=>"en", "lat"=>39.039900, "lon"=>-94.615024),
"/US/US-KS/westwood-hills-7988" => array("type"=>"city", "pop"=>364, "name_ru"=>"Уэствуд Хиллс", "name_en"=>"Westwood Hills", "lang"=>"en", "lat"=>39.039050, "lon"=>-94.610574),
"/US/US-KS/wetmore-7989" => array("type"=>"city", "pop"=>341, "name_ru"=>"Уэстмур", "name_en"=>"Wetmore", "lang"=>"en", "lat"=>39.634892, "lon"=>-95.810937),
"/US/US-KS/wheaton-7990" => array("type"=>"city", "pop"=>91, "name_ru"=>"Уитон", "name_en"=>"Wheaton", "lang"=>"en", "lat"=>39.502572, "lon"=>-96.318964),
"/US/US-KS/white-city-7991" => array("type"=>"city", "pop"=>531, "name_ru"=>"Уайт Сити", "name_en"=>"White City", "lang"=>"en", "lat"=>38.794319, "lon"=>-96.734895),
"/US/US-KS/white-cloud-7992" => array("type"=>"city", "pop"=>227, "name_ru"=>"Уайт Клауд", "name_en"=>"White Cloud", "lang"=>"en", "lat"=>39.974195, "lon"=>-95.297244),
"/US/US-KS/whitewater-7993" => array("type"=>"city", "pop"=>633, "name_ru"=>"Уайтуотер", "name_en"=>"Whitewater", "lang"=>"en", "lat"=>37.963164, "lon"=>-97.146785),
"/US/US-KS/whiting-7994" => array("type"=>"city", "pop"=>206, "name_ru"=>"Уайтинг", "name_en"=>"Whiting", "lang"=>"en", "lat"=>39.588654, "lon"=>-95.611409),
"/US/US-KS/wichita-7995" => array("type"=>"city", "pop"=>366046, "name_ru"=>"Уичита", "name_en"=>"Wichita", "lang"=>"en", "lat"=>37.687350, "lon"=>-97.342674),
"/US/US-KS/willard-7996" => array("type"=>"city", "pop"=>84, "name_ru"=>"Уиллард", "name_en"=>"Willard", "lang"=>"en", "lat"=>39.093918, "lon"=>-95.943068),
"/US/US-KS/williamsburg-7997" => array("type"=>"city", "pop"=>357, "name_ru"=>"Уилльямсбург", "name_en"=>"Williamsburg", "lang"=>"en", "lat"=>38.480101, "lon"=>-95.465759),
"/US/US-KS/willis-7998" => array("type"=>"city", "pop"=>65, "name_ru"=>"Уиллс", "name_en"=>"Willis", "lang"=>"en", "lat"=>39.722657, "lon"=>-95.505179),
"/US/US-KS/willowbrook-7999" => array("type"=>"city", "pop"=>87, "name_ru"=>"Уиллоубрук", "name_en"=>"Willowbrook", "lang"=>"en", "lat"=>38.101625, "lon"=>-97.991064),
"/US/US-KS/wilmore-8000" => array("type"=>"city", "pop"=>58, "name_ru"=>"Уилмор", "name_en"=>"Wilmore", "lang"=>"en", "lat"=>37.335902, "lon"=>-99.209305),
"/US/US-KS/wilsey-8001" => array("type"=>"city", "pop"=>188, "name_ru"=>"Уилси", "name_en"=>"Wilsey", "lang"=>"en", "lat"=>38.635937, "lon"=>-96.676796),
"/US/US-KS/wilson-8002" => array("type"=>"city", "pop"=>758, "name_ru"=>"Уилсон", "name_en"=>"Wilson", "lang"=>"en", "lat"=>38.825609, "lon"=>-98.474405),
"/US/US-KS/winchester-8003" => array("type"=>"city", "pop"=>570, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>39.321660, "lon"=>-95.267894),
"/US/US-KS/windom-8004" => array("type"=>"city", "pop"=>131, "name_ru"=>"Уиндом", "name_en"=>"Windom", "lang"=>"en", "lat"=>38.384264, "lon"=>-97.910226),
"/US/US-KS/winfield-8005" => array("type"=>"city", "pop"=>11497, "name_ru"=>"Винфилд", "name_en"=>"Winfield", "lang"=>"en", "lat"=>37.268551, "lon"=>-96.968532),
"/US/US-KS/winona-8006" => array("type"=>"city", "pop"=>189, "name_ru"=>"Уайнона", "name_en"=>"Winona", "lang"=>"en", "lat"=>39.061632, "lon"=>-101.244559),
"/US/US-KS/woodbine-8007" => array("type"=>"city", "pop"=>210, "name_ru"=>"Вудбайн", "name_en"=>"Woodbine", "lang"=>"en", "lat"=>38.795439, "lon"=>-96.959711),
"/US/US-KS/woodston-8008" => array("type"=>"city", "pop"=>106, "name_ru"=>"Вудстон", "name_en"=>"Woodston", "lang"=>"en", "lat"=>39.454290, "lon"=>-99.098747),
"/US/US-KS/yates-center-8009" => array("type"=>"city", "pop"=>1376, "name_ru"=>"Йетс Сентер", "name_en"=>"Yates Center", "lang"=>"en", "lat"=>37.872372, "lon"=>-95.742766),
"/US/US-KS/zenda-8010" => array("type"=>"city", "pop"=>111, "name_ru"=>"Зенда", "name_en"=>"Zenda", "lang"=>"en", "lat"=>37.444976, "lon"=>-98.282218),
"/US/US-KS/zurich-8011" => array("type"=>"city", "pop"=>115, "name_ru"=>"Цурих", "name_en"=>"Zurich", "lang"=>"en", "lat"=>39.231645, "lon"=>-99.434448),
"/US/US-KY" => array("type"=>"state", "pop"=>4269245, "name_ru"=>"Кентукки", "name_en"=>"Kentucky", "lang"=>"en"),
"/US/US-KY/adairville-8012" => array("type"=>"city", "pop"=>921, "name_ru"=>"Адэрвилл", "name_en"=>"Adairville", "lang"=>"en", "lat"=>36.667828, "lon"=>-86.854005),
"/US/US-KY/albany-8013" => array("type"=>"city", "pop"=>2311, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>36.690548, "lon"=>-85.135420),
"/US/US-KY/alexandria-8014" => array("type"=>"city", "pop"=>8620, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>38.960538, "lon"=>-84.384225),
"/US/US-KY/allen-8015" => array("type"=>"city", "pop"=>148, "name_ru"=>"Аллен", "name_en"=>"Allen", "lang"=>"en", "lat"=>37.611227, "lon"=>-82.728398),
"/US/US-KY/allensville-8016" => array("type"=>"city", "pop"=>218, "name_ru"=>"Алленсвилл", "name_en"=>"Allensville", "lang"=>"en", "lat"=>36.717069, "lon"=>-87.068448),
"/US/US-KY/anchorage-8017" => array("type"=>"city", "pop"=>3288, "name_ru"=>"Анкоридж", "name_en"=>"Anchorage", "lang"=>"en", "lat"=>38.270220, "lon"=>-85.538103),
"/US/US-KY/annville-8018" => array("type"=>"city", "pop"=>470, "name_ru"=>"Эннвилл", "name_en"=>"Annville", "lang"=>"en", "lat"=>37.323003, "lon"=>-83.960735),
"/US/US-KY/arlington-8019" => array("type"=>"city", "pop"=>379, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>36.790088, "lon"=>-89.012291),
"/US/US-KY/ashland-8020" => array("type"=>"city", "pop"=>21346, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>38.459192, "lon"=>-82.644921),
"/US/US-KY/auburn-8021" => array("type"=>"city", "pop"=>1500, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>36.864102, "lon"=>-86.712860),
"/US/US-KY/audubon-park-8022" => array("type"=>"city", "pop"=>1662, "name_ru"=>"Одубон Парк", "name_en"=>"Audubon Park", "lang"=>"en", "lat"=>38.204900, "lon"=>-85.727637),
"/US/US-KY/augusta-8023" => array("type"=>"city", "pop"=>1253, "name_ru"=>"Огаста", "name_en"=>"Augusta", "lang"=>"en", "lat"=>38.772486, "lon"=>-84.001601),
"/US/US-KY/bancroft-8024" => array("type"=>"city", "pop"=>603, "name_ru"=>"Бэнкрофт", "name_en"=>"Bancroft", "lang"=>"en", "lat"=>38.282973, "lon"=>-85.611646),
"/US/US-KY/barbourmeade-8025" => array("type"=>"city", "pop"=>1408, "name_ru"=>"Барбормид", "name_en"=>"Barbourmeade", "lang"=>"en", "lat"=>38.300150, "lon"=>-85.599735),
"/US/US-KY/barbourville-8026" => array("type"=>"city", "pop"=>3611, "name_ru"=>"Барборвилл", "name_en"=>"Barbourville", "lang"=>"en", "lat"=>36.866368, "lon"=>-83.884016),
"/US/US-KY/bardstown-8027" => array("type"=>"city", "pop"=>11242, "name_ru"=>"Бардстаун", "name_en"=>"Bardstown", "lang"=>"en", "lat"=>37.813579, "lon"=>-85.463193),
"/US/US-KY/bardwell-8028" => array("type"=>"city", "pop"=>775, "name_ru"=>"Бардуэлл", "name_en"=>"Bardwell", "lang"=>"en", "lat"=>36.872635, "lon"=>-89.009962),
"/US/US-KY/barlow-8029" => array("type"=>"city", "pop"=>715, "name_ru"=>"Барлоу", "name_en"=>"Barlow", "lang"=>"en", "lat"=>37.050706, "lon"=>-89.045062),
"/US/US-KY/beattyville-8030" => array("type"=>"city", "pop"=>1122, "name_ru"=>"Биттивилл", "name_en"=>"Beattyville", "lang"=>"en", "lat"=>37.580219, "lon"=>-83.707238),
"/US/US-KY/beaver-dam-8031" => array("type"=>"city", "pop"=>3133, "name_ru"=>"Бивер Дэм", "name_en"=>"Beaver Dam", "lang"=>"en", "lat"=>37.408540, "lon"=>-86.877959),
"/US/US-KY/bedford-8032" => array("type"=>"city", "pop"=>744, "name_ru"=>"Бедфорд", "name_en"=>"Bedford", "lang"=>"en", "lat"=>38.593045, "lon"=>-85.316892),
"/US/US-KY/beechwood-village-8033" => array("type"=>"city", "pop"=>1267, "name_ru"=>"Бичвуд Виллидж", "name_en"=>"Beechwood Village", "lang"=>"en", "lat"=>38.256700, "lon"=>-85.629811),
"/US/US-KY/bellefonte-8034" => array("type"=>"city", "pop"=>843, "name_ru"=>"Беллефонт", "name_en"=>"Bellefonte", "lang"=>"en", "lat"=>38.498400, "lon"=>-82.688429),
"/US/US-KY/bellemeade-8035" => array("type"=>"city", "pop"=>955, "name_ru"=>"Беллемид", "name_en"=>"Bellemeade", "lang"=>"en", "lat"=>38.249400, "lon"=>-85.591079),
"/US/US-KY/bellevue-8036" => array("type"=>"city", "pop"=>5852, "name_ru"=>"Белвью", "name_en"=>"Bellevue", "lang"=>"en", "lat"=>39.101201, "lon"=>-84.477500),
"/US/US-KY/bellewood-8037" => array("type"=>"city", "pop"=>331, "name_ru"=>"Белвуд", "name_en"=>"Bellewood", "lang"=>"en", "lat"=>38.260385, "lon"=>-85.659131),
"/US/US-KY/benham-8038" => array("type"=>"city", "pop"=>531, "name_ru"=>"Бенхэм", "name_en"=>"Benham", "lang"=>"en", "lat"=>36.964640, "lon"=>-82.951945),
"/US/US-KY/benton-8039" => array("type"=>"city", "pop"=>4367, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>36.852851, "lon"=>-88.358069),
"/US/US-KY/berea-8040" => array("type"=>"city", "pop"=>14431, "name_ru"=>"Берея", "name_en"=>"Berea", "lang"=>"en", "lat"=>37.576718, "lon"=>-84.293601),
"/US/US-KY/berry-8041" => array("type"=>"city", "pop"=>307, "name_ru"=>"Берри", "name_en"=>"Berry", "lang"=>"en", "lat"=>38.520387, "lon"=>-84.384209),
"/US/US-KY/blaine-8042" => array("type"=>"city", "pop"=>257, "name_ru"=>"Блейн", "name_en"=>"Blaine", "lang"=>"en", "lat"=>38.026234, "lon"=>-82.854099),
"/US/US-KY/blandville-8043" => array("type"=>"city", "pop"=>95, "name_ru"=>"Блэндвилл", "name_en"=>"Blandville", "lang"=>"en", "lat"=>36.943464, "lon"=>-88.964600),
"/US/US-KY/bloomfield-8044" => array("type"=>"city", "pop"=>891, "name_ru"=>"Блумфилд", "name_en"=>"Bloomfield", "lang"=>"en", "lat"=>37.912063, "lon"=>-85.315868),
"/US/US-KY/blue-ridge-manor-8045" => array("type"=>"city", "pop"=>682, "name_ru"=>"Блу Ридж Мэнор", "name_en"=>"Blue Ridge Manor", "lang"=>"en", "lat"=>38.244867, "lon"=>-85.564177),
"/US/US-KY/bonnieville-8046" => array("type"=>"city", "pop"=>367, "name_ru"=>"Боннивилл", "name_en"=>"Bonnieville", "lang"=>"en", "lat"=>37.376936, "lon"=>-85.901481),
"/US/US-KY/booneville-8047" => array("type"=>"city", "pop"=>148, "name_ru"=>"Бунвилл", "name_en"=>"Booneville", "lang"=>"en", "lat"=>37.476386, "lon"=>-83.679198),
"/US/US-KY/bowling-green-8048" => array("type"=>"city", "pop"=>55097, "name_ru"=>"Боулинг Грин", "name_en"=>"Bowling Green", "lang"=>"en", "lat"=>36.973703, "lon"=>-86.441242),
"/US/US-KY/bradfordsville-8049" => array("type"=>"city", "pop"=>322, "name_ru"=>"Брэдфордсвилл", "name_en"=>"Bradfordsville", "lang"=>"en", "lat"=>37.495438, "lon"=>-85.148237),
"/US/US-KY/brandenburg-8050" => array("type"=>"city", "pop"=>2183, "name_ru"=>"Брэнденбург", "name_en"=>"Brandenburg", "lang"=>"en", "lat"=>37.996847, "lon"=>-86.180410),
"/US/US-KY/breckinridge-center-8051" => array("type"=>"cdp", "pop"=>2375, "name_ru"=>"Брекинридж Сентер", "name_en"=>"Breckinridge Center", "lang"=>"en", "lat"=>37.682654, "lon"=>-87.863082),
"/US/US-KY/bremen-8052" => array("type"=>"city", "pop"=>359, "name_ru"=>"Бремен", "name_en"=>"Bremen", "lang"=>"en", "lat"=>37.361147, "lon"=>-87.218986),
"/US/US-KY/briarwood-8053" => array("type"=>"city", "pop"=>594, "name_ru"=>"Брайрвуд", "name_en"=>"Briarwood", "lang"=>"en", "lat"=>38.277708, "lon"=>-85.590214),
"/US/US-KY/broad-fields-8054" => array("type"=>"city", "pop"=>273, "name_ru"=>"Броад Филдс", "name_en"=>"Broad Fields", "lang"=>"en", "lat"=>38.241050, "lon"=>-85.651848),
"/US/US-KY/brodhead-8055" => array("type"=>"city", "pop"=>1199, "name_ru"=>"Бродхэд", "name_en"=>"Brodhead", "lang"=>"en", "lat"=>37.402930, "lon"=>-84.416948),
"/US/US-KY/broeck-pointe-8056" => array("type"=>"city", "pop"=>327, "name_ru"=>"Броек Пуэнт", "name_en"=>"Broeck Pointe", "lang"=>"en", "lat"=>38.295903, "lon"=>-85.585567),
"/US/US-KY/bromley-8057" => array("type"=>"city", "pop"=>769, "name_ru"=>"Бромли", "name_en"=>"Bromley", "lang"=>"en", "lat"=>39.080613, "lon"=>-84.562401),
"/US/US-KY/brooks-8058" => array("type"=>"cdp", "pop"=>2464, "name_ru"=>"Брукс", "name_en"=>"Brooks", "lang"=>"en", "lat"=>38.068184, "lon"=>-85.716881),
"/US/US-KY/brooksville-8059" => array("type"=>"city", "pop"=>571, "name_ru"=>"Бруксвилл", "name_en"=>"Brooksville", "lang"=>"en", "lat"=>38.682542, "lon"=>-84.066686),
"/US/US-KY/brownsboro-farm-8060" => array("type"=>"city", "pop"=>745, "name_ru"=>"Браунсборо Фарм", "name_en"=>"Brownsboro Farm", "lang"=>"en", "lat"=>38.302767, "lon"=>-85.592194),
"/US/US-KY/brownsboro-village-8061" => array("type"=>"city", "pop"=>340, "name_ru"=>"Браунсборо Виллидж", "name_en"=>"Brownsboro Village", "lang"=>"en", "lat"=>38.263550, "lon"=>-85.666144),
"/US/US-KY/brownsville-8062" => array("type"=>"city", "pop"=>1046, "name_ru"=>"Браунсвилл", "name_en"=>"Brownsville", "lang"=>"en", "lat"=>37.188238, "lon"=>-86.259325),
"/US/US-KY/buechel-8063" => array("type"=>"cdp", "pop"=>7081, "name_ru"=>"Буэшел", "name_en"=>"Buechel", "lang"=>"en", "lat"=>38.188800, "lon"=>-85.640395),
"/US/US-KY/burgin-8064" => array("type"=>"city", "pop"=>884, "name_ru"=>"Бургин", "name_en"=>"Burgin", "lang"=>"en", "lat"=>37.754040, "lon"=>-84.765352),
"/US/US-KY/burkesville-8065" => array("type"=>"city", "pop"=>1680, "name_ru"=>"Берксвилл", "name_en"=>"Burkesville", "lang"=>"en", "lat"=>36.790675, "lon"=>-85.368794),
"/US/US-KY/burlington-8066" => array("type"=>"cdp", "pop"=>6070, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>39.024446, "lon"=>-84.722956),
"/US/US-KY/burnside-8067" => array("type"=>"city", "pop"=>691, "name_ru"=>"Бернсайд", "name_en"=>"Burnside", "lang"=>"en", "lat"=>36.987296, "lon"=>-84.600918),
"/US/US-KY/butler-8068" => array("type"=>"city", "pop"=>627, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>38.787924, "lon"=>-84.370715),
"/US/US-KY/cadiz-8069" => array("type"=>"city", "pop"=>2563, "name_ru"=>"Кадиз", "name_en"=>"Cadiz", "lang"=>"en", "lat"=>36.869683, "lon"=>-87.823269),
"/US/US-KY/calhoun-8070" => array("type"=>"city", "pop"=>788, "name_ru"=>"Калхун", "name_en"=>"Calhoun", "lang"=>"en", "lat"=>37.541224, "lon"=>-87.259249),
"/US/US-KY/california-8071" => array("type"=>"city", "pop"=>82, "name_ru"=>"Калифорния", "name_en"=>"California", "lang"=>"en", "lat"=>38.919652, "lon"=>-84.262616),
"/US/US-KY/calvert-city-8072" => array("type"=>"city", "pop"=>2763, "name_ru"=>"Калверт Сити", "name_en"=>"Calvert City", "lang"=>"en", "lat"=>37.033273, "lon"=>-88.352934),
"/US/US-KY/camargo-8073" => array("type"=>"city", "pop"=>1034, "name_ru"=>"Камарго", "name_en"=>"Camargo", "lang"=>"en", "lat"=>37.996763, "lon"=>-83.889613),
"/US/US-KY/cambridge-8074" => array("type"=>"city", "pop"=>207, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>38.221500, "lon"=>-85.615931),
"/US/US-KY/campbellsburg-8075" => array("type"=>"city", "pop"=>697, "name_ru"=>"Кэмпбэллсбург", "name_en"=>"Campbellsburg", "lang"=>"en", "lat"=>38.526221, "lon"=>-85.206808),
"/US/US-KY/campbellsville-8076" => array("type"=>"city", "pop"=>11010, "name_ru"=>"Кэмпбелсвилл", "name_en"=>"Campbellsville", "lang"=>"en", "lat"=>37.347499, "lon"=>-85.349088),
"/US/US-KY/campton-8077" => array("type"=>"city", "pop"=>412, "name_ru"=>"Кэмптон", "name_en"=>"Campton", "lang"=>"en", "lat"=>37.735153, "lon"=>-83.547325),
"/US/US-KY/caneyville-8078" => array("type"=>"city", "pop"=>660, "name_ru"=>"Канивилл", "name_en"=>"Caneyville", "lang"=>"en", "lat"=>37.423744, "lon"=>-86.487983),
"/US/US-KY/carlisle-8079" => array("type"=>"city", "pop"=>2074, "name_ru"=>"Карлайл", "name_en"=>"Carlisle", "lang"=>"en", "lat"=>38.313613, "lon"=>-84.037572),
"/US/US-KY/carrollton-8080" => array("type"=>"city", "pop"=>3924, "name_ru"=>"Карролтон", "name_en"=>"Carrollton", "lang"=>"en", "lat"=>38.678639, "lon"=>-85.164587),
"/US/US-KY/carrsville-8081" => array("type"=>"city", "pop"=>71, "name_ru"=>"Каррсвилл", "name_en"=>"Carrsville", "lang"=>"en", "lat"=>37.397669, "lon"=>-88.373223),
"/US/US-KY/catlettsburg-8082" => array("type"=>"city", "pop"=>1938, "name_ru"=>"Катлеттсбург", "name_en"=>"Catlettsburg", "lang"=>"en", "lat"=>38.417632, "lon"=>-82.601246),
"/US/US-KY/cave-city-8083" => array("type"=>"city", "pop"=>2002, "name_ru"=>"Кейв Сити", "name_en"=>"Cave City", "lang"=>"en", "lat"=>37.138520, "lon"=>-85.964291),
"/US/US-KY/cedarville-8084" => array("type"=>"city", "pop"=>65, "name_ru"=>"Сидарвилл", "name_en"=>"Cedarville", "lang"=>"en", "lat"=>37.314144, "lon"=>-82.358305),
"/US/US-KY/centertown-8085" => array("type"=>"city", "pop"=>428, "name_ru"=>"Сентертаун", "name_en"=>"Centertown", "lang"=>"en", "lat"=>37.417995, "lon"=>-86.995459),
"/US/US-KY/central-city-8086" => array("type"=>"city", "pop"=>5698, "name_ru"=>"Сентрал Сити", "name_en"=>"Central City", "lang"=>"en", "lat"=>37.295489, "lon"=>-87.128706),
"/US/US-KY/cherrywood-village-8087" => array("type"=>"city", "pop"=>340, "name_ru"=>"Черривуд", "name_en"=>"Cherrywood Village", "lang"=>"en", "lat"=>38.261400, "lon"=>-85.654753),
"/US/US-KY/clarkson-8088" => array("type"=>"city", "pop"=>838, "name_ru"=>"Кларксон", "name_en"=>"Clarkson", "lang"=>"en", "lat"=>37.494301, "lon"=>-86.220505),
"/US/US-KY/claryville-8089" => array("type"=>"cdp", "pop"=>2038, "name_ru"=>"Кларивилл", "name_en"=>"Claryville", "lang"=>"en", "lat"=>38.924645, "lon"=>-84.409251),
"/US/US-KY/clay-8090" => array("type"=>"city", "pop"=>1142, "name_ru"=>"Клей", "name_en"=>"Clay", "lang"=>"en", "lat"=>37.476881, "lon"=>-87.820423),
"/US/US-KY/clay-city-8091" => array("type"=>"city", "pop"=>1358, "name_ru"=>"Клей Сити", "name_en"=>"Clay City", "lang"=>"en", "lat"=>37.863122, "lon"=>-83.928360),
"/US/US-KY/clinton-8092" => array("type"=>"city", "pop"=>1330, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>36.666078, "lon"=>-88.994030),
"/US/US-KY/cloverport-8093" => array("type"=>"city", "pop"=>1234, "name_ru"=>"Клаверпорт", "name_en"=>"Cloverport", "lang"=>"en", "lat"=>37.830949, "lon"=>-86.629931),
"/US/US-KY/coal-run-8094" => array("type"=>"city", "pop"=>262, "name_ru"=>"Коул Ран", "name_en"=>"Coal Run", "lang"=>"en", "lat"=>37.512946, "lon"=>-82.558689),
"/US/US-KY/cold-spring-8095" => array("type"=>"city", "pop"=>5809, "name_ru"=>"Коулд Спринг", "name_en"=>"Cold Spring", "lang"=>"en", "lat"=>39.023960, "lon"=>-84.437384),
"/US/US-KY/coldstream-8096" => array("type"=>"city", "pop"=>1061, "name_ru"=>"Коулдстрим", "name_en"=>"Coldstream", "lang"=>"en", "lat"=>38.315908, "lon"=>-85.524888),
"/US/US-KY/columbia-8097" => array("type"=>"city", "pop"=>4288, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>37.098008, "lon"=>-85.308116),
"/US/US-KY/columbus-8098" => array("type"=>"city", "pop"=>214, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>36.759523, "lon"=>-89.101807),
"/US/US-KY/concord-8099" => array("type"=>"city", "pop"=>27, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>38.688052, "lon"=>-83.491930),
"/US/US-KY/concord-8100" => array("type"=>"cdp", "pop"=>1560, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>37.072135, "lon"=>-88.700105),
"/US/US-KY/corbin-8101" => array("type"=>"city", "pop"=>8324, "name_ru"=>"Корбин", "name_en"=>"Corbin", "lang"=>"en", "lat"=>36.937388, "lon"=>-84.107432),
"/US/US-KY/corinth-8102" => array("type"=>"city", "pop"=>194, "name_ru"=>"Коринт", "name_en"=>"Corinth", "lang"=>"en", "lat"=>38.495907, "lon"=>-84.564066),
"/US/US-KY/corydon-8103" => array("type"=>"city", "pop"=>777, "name_ru"=>"Коридон", "name_en"=>"Corydon", "lang"=>"en", "lat"=>37.739666, "lon"=>-87.706539),
"/US/US-KY/covington-8104" => array("type"=>"city", "pop"=>43235, "name_ru"=>"Ковингтон", "name_en"=>"Covington", "lang"=>"en", "lat"=>39.036831, "lon"=>-84.514346),
"/US/US-KY/crab-orchard-8105" => array("type"=>"city", "pop"=>853, "name_ru"=>"Крэб Орчард", "name_en"=>"Crab Orchard", "lang"=>"en", "lat"=>37.463247, "lon"=>-84.506068),
"/US/US-KY/creekside-8106" => array("type"=>"city", "pop"=>387, "name_ru"=>"Криксайд", "name_en"=>"Creekside", "lang"=>"en", "lat"=>38.292110, "lon"=>-85.568161),
"/US/US-KY/crescent-park-8107" => array("type"=>"city", "pop"=>364, "name_ru"=>"Кресент Парк", "name_en"=>"Crescent Park", "lang"=>"en", "lat"=>39.043212, "lon"=>-84.576262),
"/US/US-KY/crescent-springs-8108" => array("type"=>"city", "pop"=>4043, "name_ru"=>"Кресент Спрингс", "name_en"=>"Crescent Springs", "lang"=>"en", "lat"=>39.052451, "lon"=>-84.580245),
"/US/US-KY/crestview-8109" => array("type"=>"city", "pop"=>454, "name_ru"=>"Крествью", "name_en"=>"Crestview", "lang"=>"en", "lat"=>39.022444, "lon"=>-84.416088),
"/US/US-KY/crestview-hills-8110" => array("type"=>"city", "pop"=>3709, "name_ru"=>"Крествью Хиллс", "name_en"=>"Crestview Hills", "lang"=>"en", "lat"=>39.026299, "lon"=>-84.571437),
"/US/US-KY/crestwood-8111" => array("type"=>"city", "pop"=>2951, "name_ru"=>"Крествуд", "name_en"=>"Crestwood", "lang"=>"en", "lat"=>38.337865, "lon"=>-85.480673),
"/US/US-KY/crittenden-8112" => array("type"=>"city", "pop"=>2636, "name_ru"=>"Криттенден", "name_en"=>"Crittenden", "lang"=>"en", "lat"=>38.781110, "lon"=>-84.605541),
"/US/US-KY/crofton-8113" => array("type"=>"city", "pop"=>962, "name_ru"=>"Крофтон", "name_en"=>"Crofton", "lang"=>"en", "lat"=>37.048144, "lon"=>-87.483422),
"/US/US-KY/crossgate-8114" => array("type"=>"city", "pop"=>274, "name_ru"=>"Кроссгейт", "name_en"=>"Crossgate", "lang"=>"en", "lat"=>38.279998, "lon"=>-85.629229),
"/US/US-KY/cumberland-8115" => array("type"=>"city", "pop"=>2352, "name_ru"=>"Камберленд", "name_en"=>"Cumberland", "lang"=>"en", "lat"=>36.984161, "lon"=>-82.986039),
"/US/US-KY/cynthiana-8116" => array("type"=>"city", "pop"=>6283, "name_ru"=>"Синтиана", "name_en"=>"Cynthiana", "lang"=>"en", "lat"=>38.386358, "lon"=>-84.295311),
"/US/US-KY/danville-8117" => array("type"=>"city", "pop"=>15524, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>37.642119, "lon"=>-84.772076),
"/US/US-KY/dawson-springs-8118" => array("type"=>"city", "pop"=>2903, "name_ru"=>"Досон Спрингс", "name_en"=>"Dawson Springs", "lang"=>"en", "lat"=>37.172665, "lon"=>-87.688628),
"/US/US-KY/dayton-8119" => array("type"=>"city", "pop"=>5453, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>39.112501, "lon"=>-84.462350),
"/US/US-KY/dixon-8120" => array("type"=>"city", "pop"=>611, "name_ru"=>"Диксон", "name_en"=>"Dixon", "lang"=>"en", "lat"=>37.517209, "lon"=>-87.690669),
"/US/US-KY/douglass-hills-8121" => array("type"=>"city", "pop"=>6002, "name_ru"=>"Дугласс Хиллс", "name_en"=>"Douglass Hills", "lang"=>"en", "lat"=>38.236000, "lon"=>-85.547682),
"/US/US-KY/dover-8122" => array("type"=>"city", "pop"=>328, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>38.755782, "lon"=>-83.883401),
"/US/US-KY/drakesboro-8123" => array("type"=>"city", "pop"=>615, "name_ru"=>"Дрейксборо", "name_en"=>"Drakesboro", "lang"=>"en", "lat"=>37.216503, "lon"=>-87.049750),
"/US/US-KY/druid-hills-8124" => array("type"=>"city", "pop"=>346, "name_ru"=>"Друид Хиллс", "name_en"=>"Druid Hills", "lang"=>"en", "lat"=>38.264000, "lon"=>-85.662128),
"/US/US-KY/dry-ridge-8125" => array("type"=>"city", "pop"=>2228, "name_ru"=>"Драй Ридж", "name_en"=>"Dry Ridge", "lang"=>"en", "lat"=>38.682175, "lon"=>-84.596428),
"/US/US-KY/dycusburg-8126" => array("type"=>"city", "pop"=>47, "name_ru"=>"Дикусбург", "name_en"=>"Dycusburg", "lang"=>"en", "lat"=>37.159964, "lon"=>-88.184922),
"/US/US-KY/earlington-8127" => array("type"=>"city", "pop"=>1571, "name_ru"=>"Эрлингтон", "name_en"=>"Earlington", "lang"=>"en", "lat"=>37.273447, "lon"=>-87.515546),
"/US/US-KY/eddyville-8128" => array("type"=>"city", "pop"=>2416, "name_ru"=>"Эддивилл", "name_en"=>"Eddyville", "lang"=>"en", "lat"=>37.075261, "lon"=>-88.079251),
"/US/US-KY/edgewood-8129" => array("type"=>"city", "pop"=>8910, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>39.006134, "lon"=>-84.562253),
"/US/US-KY/edmonton-8130" => array("type"=>"city", "pop"=>1646, "name_ru"=>"Эдмонтон", "name_en"=>"Edmonton", "lang"=>"en", "lat"=>36.982888, "lon"=>-85.620429),
"/US/US-KY/ekron-8131" => array("type"=>"city", "pop"=>161, "name_ru"=>"Экрон", "name_en"=>"Ekron", "lang"=>"en", "lat"=>37.930698, "lon"=>-86.177294),
"/US/US-KY/elizabethtown-8132" => array("type"=>"city", "pop"=>24144, "name_ru"=>"Элизабеттаун", "name_en"=>"Elizabethtown", "lang"=>"en", "lat"=>37.703087, "lon"=>-85.871655),
"/US/US-KY/elkhorn-city-8133" => array("type"=>"city", "pop"=>1002, "name_ru"=>"Элкхорн Сити", "name_en"=>"Elkhorn City", "lang"=>"en", "lat"=>37.302932, "lon"=>-82.348010),
"/US/US-KY/elkton-8134" => array("type"=>"city", "pop"=>1981, "name_ru"=>"Элктон", "name_en"=>"Elkton", "lang"=>"en", "lat"=>36.810121, "lon"=>-87.158023),
"/US/US-KY/elsmere-8135" => array("type"=>"city", "pop"=>7932, "name_ru"=>"Элсмер", "name_en"=>"Elsmere", "lang"=>"en", "lat"=>38.994251, "lon"=>-84.600446),
"/US/US-KY/eminence-8136" => array("type"=>"city", "pop"=>2216, "name_ru"=>"Эминенс", "name_en"=>"Eminence", "lang"=>"en", "lat"=>38.365379, "lon"=>-85.174322),
"/US/US-KY/erlanger-8137" => array("type"=>"city", "pop"=>17188, "name_ru"=>"Эрлангер", "name_en"=>"Erlanger", "lang"=>"en", "lat"=>39.009851, "lon"=>-84.588081),
"/US/US-KY/eubank-8138" => array("type"=>"city", "pop"=>376, "name_ru"=>"Ойбанк", "name_en"=>"Eubank", "lang"=>"en", "lat"=>37.279321, "lon"=>-84.654913),
"/US/US-KY/evarts-8139" => array("type"=>"city", "pop"=>1027, "name_ru"=>"Эвартс", "name_en"=>"Evarts", "lang"=>"en", "lat"=>36.863692, "lon"=>-83.197653),
"/US/US-KY/ewing-8140" => array("type"=>"city", "pop"=>292, "name_ru"=>"Юинг", "name_en"=>"Ewing", "lang"=>"en", "lat"=>38.427314, "lon"=>-83.862306),
"/US/US-KY/fairdale-8141" => array("type"=>"cdp", "pop"=>6563, "name_ru"=>"Фэрдейл", "name_en"=>"Fairdale", "lang"=>"en", "lat"=>38.111806, "lon"=>-85.757523),
"/US/US-KY/fairfield-8142" => array("type"=>"city", "pop"=>77, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>37.933246, "lon"=>-85.384024),
"/US/US-KY/fairmeade-8143" => array("type"=>"city", "pop"=>280, "name_ru"=>"Фэрмид", "name_en"=>"Fairmeade", "lang"=>"en", "lat"=>38.249400, "lon"=>-85.635299),
"/US/US-KY/fairview-8144" => array("type"=>"city", "pop"=>142, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>38.998571, "lon"=>-84.482838),
"/US/US-KY/falmouth-8145" => array("type"=>"city", "pop"=>2059, "name_ru"=>"Фолмут", "name_en"=>"Falmouth", "lang"=>"en", "lat"=>38.673790, "lon"=>-84.334276),
"/US/US-KY/ferguson-8146" => array("type"=>"city", "pop"=>937, "name_ru"=>"Фергюсон", "name_en"=>"Ferguson", "lang"=>"en", "lat"=>37.066653, "lon"=>-84.594532),
"/US/US-KY/fern-creek-8147" => array("type"=>"cdp", "pop"=>16406, "name_ru"=>"Ферн Крик", "name_en"=>"Fern Creek", "lang"=>"en", "lat"=>38.158300, "lon"=>-85.596670),
"/US/US-KY/fincastle-8148" => array("type"=>"city", "pop"=>919, "name_ru"=>"Финкасл", "name_en"=>"Fincastle", "lang"=>"en", "lat"=>38.307650, "lon"=>-85.542299),
"/US/US-KY/flatwoods-8149" => array("type"=>"city", "pop"=>7655, "name_ru"=>"Флэтвудс", "name_en"=>"Flatwoods", "lang"=>"en", "lat"=>38.519304, "lon"=>-82.720046),
"/US/US-KY/fleming-neon-8150" => array("type"=>"city", "pop"=>789, "name_ru"=>"Флеминг-Неон", "name_en"=>"Fleming-Neon", "lang"=>"en", "lat"=>37.191800, "lon"=>-82.703780),
"/US/US-KY/flemingsburg-8151" => array("type"=>"city", "pop"=>2673, "name_ru"=>"Флемингсбург", "name_en"=>"Flemingsburg", "lang"=>"en", "lat"=>38.423506, "lon"=>-83.736919),
"/US/US-KY/florence-8152" => array("type"=>"city", "pop"=>27745, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>38.991501, "lon"=>-84.642912),
"/US/US-KY/fordsville-8153" => array("type"=>"city", "pop"=>548, "name_ru"=>"Фордсвилл", "name_en"=>"Fordsville", "lang"=>"en", "lat"=>37.636184, "lon"=>-86.715963),
"/US/US-KY/forest-hills-8154" => array("type"=>"city", "pop"=>586, "name_ru"=>"Форест Хиллс", "name_en"=>"Forest Hills", "lang"=>"en", "lat"=>38.215570, "lon"=>-85.583746),
"/US/US-KY/fort-campbell-north-8155" => array("type"=>"cdp", "pop"=>18861, "name_ru"=>"Форт Кэмпбэлл Норт", "name_en"=>"Fort Campbell North", "lang"=>"en", "lat"=>36.653584, "lon"=>-87.459706),
"/US/US-KY/fort-knox-8156" => array("type"=>"cdp", "pop"=>21495, "name_ru"=>"Форт Нокс", "name_en"=>"Fort Knox", "lang"=>"en", "lat"=>37.891042, "lon"=>-85.963200),
"/US/US-KY/fort-mitchell-8157" => array("type"=>"city", "pop"=>7562, "name_ru"=>"Форт Митчелл", "name_en"=>"Fort Mitchell", "lang"=>"en", "lat"=>39.045151, "lon"=>-84.555541),
"/US/US-KY/fort-thomas-8158" => array("type"=>"city", "pop"=>15222, "name_ru"=>"Форт Томас", "name_en"=>"Fort Thomas", "lang"=>"en", "lat"=>39.082401, "lon"=>-84.450317),
"/US/US-KY/fort-wright-8159" => array("type"=>"city", "pop"=>5450, "name_ru"=>"Форт Райт", "name_en"=>"Fort Wright", "lang"=>"en", "lat"=>39.047120, "lon"=>-84.536883),
"/US/US-KY/foster-8160" => array("type"=>"city", "pop"=>65, "name_ru"=>"Фостер", "name_en"=>"Foster", "lang"=>"en", "lat"=>38.799768, "lon"=>-84.213868),
"/US/US-KY/fountain-run-8161" => array("type"=>"city", "pop"=>239, "name_ru"=>"Фонтейн Ран", "name_en"=>"Fountain Run", "lang"=>"en", "lat"=>36.722147, "lon"=>-85.960454),
"/US/US-KY/fox-chase-8162" => array("type"=>"city", "pop"=>515, "name_ru"=>"Фокс Чейз", "name_en"=>"Fox Chase", "lang"=>"en", "lat"=>38.045718, "lon"=>-85.689727),
"/US/US-KY/frankfort-8163" => array("type"=>"city", "pop"=>27322, "name_ru"=>"Франкфорт", "name_en"=>"Frankfort", "lang"=>"en", "lat"=>38.190770, "lon"=>-84.865203),
"/US/US-KY/franklin-8164" => array("type"=>"city", "pop"=>7997, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>36.728023, "lon"=>-86.573935),
"/US/US-KY/fredonia-8165" => array("type"=>"city", "pop"=>414, "name_ru"=>"Фредониа", "name_en"=>"Fredonia", "lang"=>"en", "lat"=>37.209060, "lon"=>-88.058829),
"/US/US-KY/frenchburg-8166" => array("type"=>"city", "pop"=>547, "name_ru"=>"Френчбург", "name_en"=>"Frenchburg", "lang"=>"en", "lat"=>37.954806, "lon"=>-83.622449),
"/US/US-KY/fulton-8167" => array("type"=>"city", "pop"=>2438, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>36.512451, "lon"=>-88.881213),
"/US/US-KY/gamaliel-8168" => array("type"=>"city", "pop"=>427, "name_ru"=>"Гамалиел", "name_en"=>"Gamaliel", "lang"=>"en", "lat"=>36.639900, "lon"=>-85.793396),
"/US/US-KY/georgetown-8169" => array("type"=>"city", "pop"=>21589, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>38.205168, "lon"=>-84.552190),
"/US/US-KY/germantown-8170" => array("type"=>"city", "pop"=>141, "name_ru"=>"Джермэнтаун", "name_en"=>"Germantown", "lang"=>"en", "lat"=>38.651960, "lon"=>-83.967160),
"/US/US-KY/ghent-8171" => array("type"=>"city", "pop"=>389, "name_ru"=>"Гент", "name_en"=>"Ghent", "lang"=>"en", "lat"=>38.735732, "lon"=>-85.060954),
"/US/US-KY/glasgow-8172" => array("type"=>"city", "pop"=>14302, "name_ru"=>"Гласго", "name_en"=>"Glasgow", "lang"=>"en", "lat"=>36.999128, "lon"=>-85.921544),
"/US/US-KY/glencoe-8173" => array("type"=>"city", "pop"=>379, "name_ru"=>"Гленко", "name_en"=>"Glencoe", "lang"=>"en", "lat"=>38.711117, "lon"=>-84.821827),
"/US/US-KY/glenview-8174" => array("type"=>"city", "pop"=>894, "name_ru"=>"Гленвью", "name_en"=>"Glenview", "lang"=>"en", "lat"=>38.301071, "lon"=>-85.650717),
"/US/US-KY/glenview-hills-8175" => array("type"=>"city", "pop"=>377, "name_ru"=>"Гленвью Хиллс", "name_en"=>"Glenview Hills", "lang"=>"en", "lat"=>38.294100, "lon"=>-85.637993),
"/US/US-KY/glenview-manor-8176" => array("type"=>"city", "pop"=>220, "name_ru"=>"Гленвью Мэнор", "name_en"=>"Glenview Manor", "lang"=>"en", "lat"=>38.290700, "lon"=>-85.632624),
"/US/US-KY/goose-creek-8177" => array("type"=>"city", "pop"=>298, "name_ru"=>"Гуз Крик", "name_en"=>"Goose Creek", "lang"=>"en", "lat"=>38.292878, "lon"=>-85.589953),
"/US/US-KY/goshen-8178" => array("type"=>"cdp", "pop"=>992, "name_ru"=>"Гошен", "name_en"=>"Goshen", "lang"=>"en", "lat"=>38.408632, "lon"=>-85.591620),
"/US/US-KY/grand-rivers-8179" => array("type"=>"city", "pop"=>333, "name_ru"=>"Грэнд Риверс", "name_en"=>"Grand Rivers", "lang"=>"en", "lat"=>37.005663, "lon"=>-88.235878),
"/US/US-KY/gratz-8180" => array("type"=>"city", "pop"=>95, "name_ru"=>"Гратц", "name_en"=>"Gratz", "lang"=>"en", "lat"=>38.473007, "lon"=>-84.946387),
"/US/US-KY/graymoor-devondale-8181" => array("type"=>"city", "pop"=>3185, "name_ru"=>"Греймур-Девондейл", "name_en"=>"Graymoor-Devondale", "lang"=>"en", "lat"=>38.273300, "lon"=>-85.617298),
"/US/US-KY/grayson-8182" => array("type"=>"city", "pop"=>3985, "name_ru"=>"Грейсон", "name_en"=>"Grayson", "lang"=>"en", "lat"=>38.333556, "lon"=>-82.938289),
"/US/US-KY/green-spring-8184" => array("type"=>"city", "pop"=>868, "name_ru"=>"Грин Спринг", "name_en"=>"Green Spring", "lang"=>"en", "lat"=>38.316894, "lon"=>-85.614459),
"/US/US-KY/greensburg-8183" => array("type"=>"city", "pop"=>2415, "name_ru"=>"Гринсбург", "name_en"=>"Greensburg", "lang"=>"en", "lat"=>37.255951, "lon"=>-85.495281),
"/US/US-KY/greenup-8185" => array("type"=>"city", "pop"=>1185, "name_ru"=>"Гринэп", "name_en"=>"Greenup", "lang"=>"en", "lat"=>38.573000, "lon"=>-82.829066),
"/US/US-KY/greenville-8186" => array("type"=>"city", "pop"=>4235, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>37.206738, "lon"=>-87.178453),
"/US/US-KY/guthrie-8187" => array("type"=>"city", "pop"=>1454, "name_ru"=>"Гатри", "name_en"=>"Guthrie", "lang"=>"en", "lat"=>36.647352, "lon"=>-87.170723),
"/US/US-KY/hanson-8188" => array("type"=>"city", "pop"=>583, "name_ru"=>"Хэнсон", "name_en"=>"Hanson", "lang"=>"en", "lat"=>37.417360, "lon"=>-87.472121),
"/US/US-KY/hardin-8189" => array("type"=>"city", "pop"=>572, "name_ru"=>"Хардин", "name_en"=>"Hardin", "lang"=>"en", "lat"=>36.764269, "lon"=>-88.301496),
"/US/US-KY/hardinsburg-8190" => array("type"=>"city", "pop"=>2433, "name_ru"=>"Хардинсбург", "name_en"=>"Hardinsburg", "lang"=>"en", "lat"=>37.776058, "lon"=>-86.454965),
"/US/US-KY/harlan-8191" => array("type"=>"city", "pop"=>1864, "name_ru"=>"Харлан", "name_en"=>"Harlan", "lang"=>"en", "lat"=>36.841399, "lon"=>-83.320163),
"/US/US-KY/harrodsburg-8192" => array("type"=>"city", "pop"=>8192, "name_ru"=>"Нэрродсбург", "name_en"=>"Harrodsburg", "lang"=>"en", "lat"=>37.766572, "lon"=>-84.847241),
"/US/US-KY/hartford-8193" => array("type"=>"city", "pop"=>2666, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>37.451051, "lon"=>-86.893820),
"/US/US-KY/hawesville-8194" => array("type"=>"city", "pop"=>982, "name_ru"=>"Хосвилл", "name_en"=>"Hawesville", "lang"=>"en", "lat"=>37.898810, "lon"=>-86.755582),
"/US/US-KY/hazard-8195" => array("type"=>"city", "pop"=>4796, "name_ru"=>"Хазард", "name_en"=>"Hazard", "lang"=>"en", "lat"=>37.252960, "lon"=>-83.195280),
"/US/US-KY/hazel-8196" => array("type"=>"city", "pop"=>454, "name_ru"=>"Хейзел", "name_en"=>"Hazel", "lang"=>"en", "lat"=>36.505365, "lon"=>-88.325987),
"/US/US-KY/hebron-estates-8197" => array("type"=>"city", "pop"=>1181, "name_ru"=>"Хеброн Эстейтс", "name_en"=>"Hebron Estates", "lang"=>"en", "lat"=>38.050648, "lon"=>-85.669405),
"/US/US-KY/henderson-8198" => array("type"=>"city", "pop"=>27933, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>37.839918, "lon"=>-87.582002),
"/US/US-KY/hendron-8199" => array("type"=>"cdp", "pop"=>3712, "name_ru"=>"Хендрон", "name_en"=>"Hendron", "lang"=>"en", "lat"=>37.036901, "lon"=>-88.644296),
"/US/US-KY/hickman-8200" => array("type"=>"city", "pop"=>2239, "name_ru"=>"Хикман", "name_en"=>"Hickman", "lang"=>"en", "lat"=>36.562878, "lon"=>-89.187811),
"/US/US-KY/hickory-hill-8201" => array("type"=>"city", "pop"=>158, "name_ru"=>"Хикори Хилл", "name_en"=>"Hickory Hill", "lang"=>"en", "lat"=>38.287486, "lon"=>-85.568295),
"/US/US-KY/highland-heights-8202" => array("type"=>"city", "pop"=>5720, "name_ru"=>"Хайленд Хайтс", "name_en"=>"Highland Heights", "lang"=>"en", "lat"=>39.036901, "lon"=>-84.454220),
"/US/US-KY/highview-8203" => array("type"=>"cdp", "pop"=>14814, "name_ru"=>"Хайвью", "name_en"=>"Highview", "lang"=>"en", "lat"=>38.140200, "lon"=>-85.635014),
"/US/US-KY/hills-and-dales-8204" => array("type"=>"city", "pop"=>183, "name_ru"=>"Хиллс энд Дейлс", "name_en"=>"Hills and Dales", "lang"=>"en", "lat"=>38.301250, "lon"=>-85.624249),
"/US/US-KY/hillview-8205" => array("type"=>"city", "pop"=>7572, "name_ru"=>"Хиллвью", "name_en"=>"Hillview", "lang"=>"en", "lat"=>38.069720, "lon"=>-85.685661),
"/US/US-KY/hindman-8206" => array("type"=>"city", "pop"=>767, "name_ru"=>"Хиндман", "name_en"=>"Hindman", "lang"=>"en", "lat"=>37.333733, "lon"=>-82.983182),
"/US/US-KY/hiseville-8207" => array("type"=>"city", "pop"=>220, "name_ru"=>"Хизвилл", "name_en"=>"Hiseville", "lang"=>"en", "lat"=>37.100078, "lon"=>-85.815135),
"/US/US-KY/hodgenville-8208" => array("type"=>"city", "pop"=>2775, "name_ru"=>"Ходженвилл", "name_en"=>"Hodgenville", "lang"=>"en", "lat"=>37.568128, "lon"=>-85.735723),
"/US/US-KY/hollow-creek-8209" => array("type"=>"city", "pop"=>906, "name_ru"=>"Холлоу Крик", "name_en"=>"Hollow Creek", "lang"=>"en", "lat"=>38.151504, "lon"=>-85.625267),
"/US/US-KY/hollyvilla-8210" => array("type"=>"city", "pop"=>576, "name_ru"=>"Холливилла", "name_en"=>"Hollyvilla", "lang"=>"en", "lat"=>38.093833, "lon"=>-85.745241),
"/US/US-KY/hopkinsville-8211" => array("type"=>"city", "pop"=>32076, "name_ru"=>"Хопкинсвилл", "name_en"=>"Hopkinsville", "lang"=>"en", "lat"=>36.854422, "lon"=>-87.490928),
"/US/US-KY/horse-cave-8212" => array("type"=>"city", "pop"=>2326, "name_ru"=>"Хорс Кейв", "name_en"=>"Horse Cave", "lang"=>"en", "lat"=>37.172430, "lon"=>-85.911335),
"/US/US-KY/houston-acres-8213" => array("type"=>"city", "pop"=>542, "name_ru"=>"Хьюстон Эйкрес", "name_en"=>"Houston Acres", "lang"=>"en", "lat"=>38.214050, "lon"=>-85.613649),
"/US/US-KY/hunters-hollow-8214" => array("type"=>"city", "pop"=>395, "name_ru"=>"Хантерс Холлоу", "name_en"=>"Hunters Hollow", "lang"=>"en", "lat"=>38.077184, "lon"=>-85.691384),
"/US/US-KY/hurstbourne-8215" => array("type"=>"city", "pop"=>4423, "name_ru"=>"Херстборн", "name_en"=>"Hurstbourne", "lang"=>"en", "lat"=>38.238150, "lon"=>-85.586623),
"/US/US-KY/hurstbourne-acres-8216" => array("type"=>"city", "pop"=>1615, "name_ru"=>"Херстборн Эйкрес", "name_en"=>"Hurstbourne Acres", "lang"=>"en", "lat"=>38.220200, "lon"=>-85.590530),
"/US/US-KY/hustonville-8217" => array("type"=>"city", "pop"=>350, "name_ru"=>"Хьюстонвилл", "name_en"=>"Hustonville", "lang"=>"en", "lat"=>37.472725, "lon"=>-84.817720),
"/US/US-KY/hyden-8218" => array("type"=>"city", "pop"=>190, "name_ru"=>"Хайден", "name_en"=>"Hyden", "lang"=>"en", "lat"=>37.161700, "lon"=>-83.373608),
"/US/US-KY/independence-8219" => array("type"=>"city", "pop"=>21694, "name_ru"=>"Индепенденс", "name_en"=>"Independence", "lang"=>"en", "lat"=>38.954417, "lon"=>-84.547352),
"/US/US-KY/indian-hills-8220" => array("type"=>"city", "pop"=>3480, "name_ru"=>"Индиан Хиллс", "name_en"=>"Indian Hills", "lang"=>"en", "lat"=>38.279072, "lon"=>-85.667636),
"/US/US-KY/indian-hills-cherokee-section-8221" => array("type"=>"city", "pop"=>1005, "name_ru"=>"Индиан Хиллс Чероки Секшен", "name_en"=>"Indian Hills Cherokee Section", "lang"=>"en", "lat"=>38.277907, "lon"=>-85.651570),
"/US/US-KY/inez-8222" => array("type"=>"city", "pop"=>435, "name_ru"=>"Инез", "name_en"=>"Inez", "lang"=>"en", "lat"=>37.866143, "lon"=>-82.537532),
"/US/US-KY/irvine-8223" => array("type"=>"city", "pop"=>2666, "name_ru"=>"Ирвайн", "name_en"=>"Irvine", "lang"=>"en", "lat"=>37.696366, "lon"=>-83.968348),
"/US/US-KY/irvington-8224" => array("type"=>"city", "pop"=>1400, "name_ru"=>"Ирвингтон", "name_en"=>"Irvington", "lang"=>"en", "lat"=>37.878901, "lon"=>-86.284664),
"/US/US-KY/island-8225" => array("type"=>"city", "pop"=>425, "name_ru"=>"Айленд", "name_en"=>"Island", "lang"=>"en", "lat"=>37.442742, "lon"=>-87.147648),
"/US/US-KY/jackson-8226" => array("type"=>"city", "pop"=>2400, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>37.557654, "lon"=>-83.379731),
"/US/US-KY/jamestown-8227" => array("type"=>"city", "pop"=>1738, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>36.993056, "lon"=>-85.070578),
"/US/US-KY/jeffersontown-8228" => array("type"=>"city", "pop"=>26226, "name_ru"=>"Джефферсонтаун", "name_en"=>"Jeffersontown", "lang"=>"en", "lat"=>38.205600, "lon"=>-85.570920),
"/US/US-KY/jeffersonville-8229" => array("type"=>"city", "pop"=>1638, "name_ru"=>"Джефферсонвилл", "name_en"=>"Jeffersonville", "lang"=>"en", "lat"=>37.965041, "lon"=>-83.828135),
"/US/US-KY/jenkins-8230" => array("type"=>"city", "pop"=>2244, "name_ru"=>"Дженкинс", "name_en"=>"Jenkins", "lang"=>"en", "lat"=>37.178124, "lon"=>-82.626625),
"/US/US-KY/junction-city-8231" => array("type"=>"city", "pop"=>2211, "name_ru"=>"Джанкшен Сити", "name_en"=>"Junction City", "lang"=>"en", "lat"=>37.585527, "lon"=>-84.789043),
"/US/US-KY/keeneland-8232" => array("type"=>"city", "pop"=>393, "name_ru"=>"Кинленд", "name_en"=>"Keeneland", "lang"=>"en", "lat"=>38.275950, "lon"=>-85.566488),
"/US/US-KY/kenton-vale-8233" => array("type"=>"city", "pop"=>144, "name_ru"=>"Кентонвейл", "name_en"=>"Kenton Vale", "lang"=>"en", "lat"=>39.051373, "lon"=>-84.519909),
"/US/US-KY/kevil-8234" => array("type"=>"city", "pop"=>576, "name_ru"=>"Кевил", "name_en"=>"Kevil", "lang"=>"en", "lat"=>37.084852, "lon"=>-88.885091),
"/US/US-KY/kingsley-8235" => array("type"=>"city", "pop"=>426, "name_ru"=>"Кингсли", "name_en"=>"Kingsley", "lang"=>"en", "lat"=>38.221500, "lon"=>-85.672590),
"/US/US-KY/kuttawa-8236" => array("type"=>"city", "pop"=>647, "name_ru"=>"Каттава", "name_en"=>"Kuttawa", "lang"=>"en", "lat"=>37.061927, "lon"=>-88.109645),
"/US/US-KY/la-center-8237" => array("type"=>"city", "pop"=>1039, "name_ru"=>"Ла Сентер", "name_en"=>"La Center", "lang"=>"en", "lat"=>37.074824, "lon"=>-88.974103),
"/US/US-KY/la-grange-8239" => array("type"=>"city", "pop"=>6286, "name_ru"=>"Ла Гранж", "name_en"=>"La Grange", "lang"=>"en", "lat"=>38.402794, "lon"=>-85.378551),
"/US/US-KY/lafayette-8238" => array("type"=>"city", "pop"=>215, "name_ru"=>"Ла Файет", "name_en"=>"LaFayette", "lang"=>"en", "lat"=>36.660116, "lon"=>-87.658155),
"/US/US-KY/lakeside-park-8240" => array("type"=>"city", "pop"=>2681, "name_ru"=>"Лейксайд Парк", "name_en"=>"Lakeside Park", "lang"=>"en", "lat"=>39.034251, "lon"=>-84.566471),
"/US/US-KY/lakeview-heights-8241" => array("type"=>"city", "pop"=>201, "name_ru"=>"Лейквью Хайтс", "name_en"=>"Lakeview Heights", "lang"=>"en", "lat"=>38.152133, "lon"=>-83.504486),
"/US/US-KY/lancaster-8242" => array("type"=>"city", "pop"=>4403, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>37.618545, "lon"=>-84.579493),
"/US/US-KY/langdon-place-8243" => array("type"=>"city", "pop"=>1076, "name_ru"=>"Лэнгдон Плейс", "name_en"=>"Langdon Place", "lang"=>"en", "lat"=>38.286848, "lon"=>-85.584904),
"/US/US-KY/latonia-lakes-8244" => array("type"=>"city", "pop"=>410, "name_ru"=>"Латония Лейкс", "name_en"=>"Latonia Lakes", "lang"=>"en", "lat"=>38.967453, "lon"=>-84.495688),
"/US/US-KY/lawrenceburg-8245" => array("type"=>"city", "pop"=>9978, "name_ru"=>"Лоренсбург", "name_en"=>"Lawrenceburg", "lang"=>"en", "lat"=>38.034807, "lon"=>-84.894548),
"/US/US-KY/lebanon-8246" => array("type"=>"city", "pop"=>5973, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>37.567033, "lon"=>-85.254444),
"/US/US-KY/lebanon-junction-8247" => array("type"=>"city", "pop"=>2012, "name_ru"=>"Лебанон Джанкшен", "name_en"=>"Lebanon Junction", "lang"=>"en", "lat"=>37.833759, "lon"=>-85.725419),
"/US/US-KY/ledbetter-8248" => array("type"=>"cdp", "pop"=>1694, "name_ru"=>"Ледбеттер", "name_en"=>"Ledbetter", "lang"=>"en", "lat"=>37.048054, "lon"=>-88.497176),
"/US/US-KY/leitchfield-8249" => array("type"=>"city", "pop"=>6537, "name_ru"=>"Леитчфилд", "name_en"=>"Leitchfield", "lang"=>"en", "lat"=>37.480659, "lon"=>-86.293349),
"/US/US-KY/lewisburg-8250" => array("type"=>"city", "pop"=>912, "name_ru"=>"Льюисбург", "name_en"=>"Lewisburg", "lang"=>"en", "lat"=>36.986562, "lon"=>-86.950241),
"/US/US-KY/lewisport-8251" => array("type"=>"city", "pop"=>1658, "name_ru"=>"Льюиспорт", "name_en"=>"Lewisport", "lang"=>"en", "lat"=>37.932202, "lon"=>-86.902373),
"/US/US-KY/lexington-fayette-8252" => array("type"=>"lexington-fayette", "pop"=>225366, "name_ru"=>"Лексингтон-Файет", "name_en"=>"Lexington-Fayette", "lang"=>"en", "lat"=>38.042746, "lon"=>-84.459460),
"/US/US-KY/liberty-8253" => array("type"=>"city", "pop"=>1893, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>37.317849, "lon"=>-84.930604),
"/US/US-KY/lincolnshire-8254" => array("type"=>"city", "pop"=>169, "name_ru"=>"Линкольншир", "name_en"=>"Lincolnshire", "lang"=>"en", "lat"=>38.223828, "lon"=>-85.621188),
"/US/US-KY/livermore-8255" => array("type"=>"city", "pop"=>1430, "name_ru"=>"Ливермор", "name_en"=>"Livermore", "lang"=>"en", "lat"=>37.491572, "lon"=>-87.134061),
"/US/US-KY/livingston-8256" => array("type"=>"city", "pop"=>228, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>37.298466, "lon"=>-84.216728),
"/US/US-KY/london-8257" => array("type"=>"city", "pop"=>7962, "name_ru"=>"Лондон", "name_en"=>"London", "lang"=>"en", "lat"=>37.127965, "lon"=>-84.082201),
"/US/US-KY/lone-oak-8258" => array("type"=>"city", "pop"=>438, "name_ru"=>"Лоун Оук", "name_en"=>"Lone Oak", "lang"=>"en", "lat"=>37.036110, "lon"=>-88.666081),
"/US/US-KY/loretto-8259" => array("type"=>"city", "pop"=>654, "name_ru"=>"Лоретто", "name_en"=>"Loretto", "lang"=>"en", "lat"=>37.635485, "lon"=>-85.402097),
"/US/US-KY/louisa-8260" => array("type"=>"city", "pop"=>2088, "name_ru"=>"Луиза", "name_en"=>"Louisa", "lang"=>"en", "lat"=>38.109433, "lon"=>-82.603437),
"/US/US-KY/louisville-8261" => array("type"=>"city", "pop"=>269063, "name_ru"=>"Луисвилл", "name_en"=>"Louisville", "lang"=>"en", "lat"=>38.224750, "lon"=>-85.741156),
"/US/US-KY/loyall-8262" => array("type"=>"city", "pop"=>703, "name_ru"=>"Лойолл", "name_en"=>"Loyall", "lang"=>"en", "lat"=>36.851800, "lon"=>-83.350828),
"/US/US-KY/ludlow-8263" => array("type"=>"city", "pop"=>4808, "name_ru"=>"Ладлоу", "name_en"=>"Ludlow", "lang"=>"en", "lat"=>39.089351, "lon"=>-84.550517),
"/US/US-KY/lynch-8264" => array("type"=>"city", "pop"=>820, "name_ru"=>"Линч", "name_en"=>"Lynch", "lang"=>"en", "lat"=>36.964537, "lon"=>-82.913632),
"/US/US-KY/lyndon-8265" => array("type"=>"city", "pop"=>11242, "name_ru"=>"Линдон", "name_en"=>"Lyndon", "lang"=>"en", "lat"=>38.264414, "lon"=>-85.589349),
"/US/US-KY/lynnview-8266" => array("type"=>"city", "pop"=>1036, "name_ru"=>"Линнвилл", "name_en"=>"Lynnview", "lang"=>"en", "lat"=>38.178850, "lon"=>-85.710298),
"/US/US-KY/mackville-8269" => array("type"=>"city", "pop"=>218, "name_ru"=>"Маквилл", "name_en"=>"Mackville", "lang"=>"en", "lat"=>37.733588, "lon"=>-85.068434),
"/US/US-KY/madisonville-8271" => array("type"=>"city", "pop"=>19106, "name_ru"=>"Мэдисонвилл", "name_en"=>"Madisonville", "lang"=>"en", "lat"=>37.336612, "lon"=>-87.505401),
"/US/US-KY/manchester-8272" => array("type"=>"city", "pop"=>1951, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>37.152504, "lon"=>-83.768055),
"/US/US-KY/manor-creek-8273" => array("type"=>"city", "pop"=>247, "name_ru"=>"Мэнор Крик", "name_en"=>"Manor Creek", "lang"=>"en", "lat"=>38.297951, "lon"=>-85.588437),
"/US/US-KY/marion-8274" => array("type"=>"city", "pop"=>3102, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>37.331460, "lon"=>-88.079178),
"/US/US-KY/martin-8275" => array("type"=>"city", "pop"=>633, "name_ru"=>"Мартин", "name_en"=>"Martin", "lang"=>"en", "lat"=>37.571834, "lon"=>-82.756017),
"/US/US-KY/maryhill-estates-8276" => array("type"=>"city", "pop"=>194, "name_ru"=>"Мэрихилл Эстейтс", "name_en"=>"Maryhill Estates", "lang"=>"en", "lat"=>38.266300, "lon"=>-85.652765),
"/US/US-KY/masonville-8277" => array("type"=>"cdp", "pop"=>1119, "name_ru"=>"Мейсонвилл", "name_en"=>"Masonville", "lang"=>"en", "lat"=>37.673096, "lon"=>-87.034524),
"/US/US-KY/massac-8278" => array("type"=>"cdp", "pop"=>3733, "name_ru"=>"Массак", "name_en"=>"Massac", "lang"=>"en", "lat"=>37.033796, "lon"=>-88.685724),
"/US/US-KY/mayfield-8279" => array("type"=>"city", "pop"=>10186, "name_ru"=>"Мейфилд", "name_en"=>"Mayfield", "lang"=>"en", "lat"=>36.736326, "lon"=>-88.645742),
"/US/US-KY/maysville-8280" => array("type"=>"city", "pop"=>9246, "name_ru"=>"Мейсвилл", "name_en"=>"Maysville", "lang"=>"en", "lat"=>38.629536, "lon"=>-83.780051),
"/US/US-KY/mchenry-8267" => array("type"=>"city", "pop"=>435, "name_ru"=>"Мак-Генри", "name_en"=>"McHenry", "lang"=>"en", "lat"=>37.379933, "lon"=>-86.921297),
"/US/US-KY/mckee-8268" => array("type"=>"city", "pop"=>858, "name_ru"=>"Мак-Ки", "name_en"=>"McKee", "lang"=>"en", "lat"=>37.430323, "lon"=>-83.987093),
"/US/US-KY/mcroberts-8270" => array("type"=>"cdp", "pop"=>1101, "name_ru"=>"Мак-Робертс", "name_en"=>"McRoberts", "lang"=>"en", "lat"=>37.211796, "lon"=>-82.667974),
"/US/US-KY/meadow-vale-8282" => array("type"=>"city", "pop"=>849, "name_ru"=>"Медоу Вейл", "name_en"=>"Meadow Vale", "lang"=>"en", "lat"=>38.283773, "lon"=>-85.573224),
"/US/US-KY/meadowbrook-farm-8281" => array("type"=>"city", "pop"=>155, "name_ru"=>"Медоубрук Фарм", "name_en"=>"Meadowbrook Farm", "lang"=>"en", "lat"=>38.278650, "lon"=>-85.575540),
"/US/US-KY/meadowview-estates-8283" => array("type"=>"city", "pop"=>451, "name_ru"=>"Медоувью Эстейтс", "name_en"=>"Meadowview Estates", "lang"=>"en", "lat"=>38.223100, "lon"=>-85.636156),
"/US/US-KY/melbourne-8284" => array("type"=>"city", "pop"=>448, "name_ru"=>"Мельбурн", "name_en"=>"Melbourne", "lang"=>"en", "lat"=>39.031716, "lon"=>-84.370968),
"/US/US-KY/mentor-8285" => array("type"=>"city", "pop"=>166, "name_ru"=>"Ментор", "name_en"=>"Mentor", "lang"=>"en", "lat"=>38.890628, "lon"=>-84.243686),
"/US/US-KY/middlesborough-8286" => array("type"=>"city", "pop"=>9929, "name_ru"=>"Миддлсборо", "name_en"=>"Middlesborough", "lang"=>"en", "lat"=>36.613198, "lon"=>-83.722370),
"/US/US-KY/middletown-8287" => array("type"=>"city", "pop"=>7147, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>38.242467, "lon"=>-85.519626),
"/US/US-KY/midway-8288" => array("type"=>"city", "pop"=>1604, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>38.149740, "lon"=>-84.683498),
"/US/US-KY/millersburg-8289" => array("type"=>"city", "pop"=>865, "name_ru"=>"Миллерсберг", "name_en"=>"Millersburg", "lang"=>"en", "lat"=>38.303393, "lon"=>-84.146154),
"/US/US-KY/milton-8290" => array("type"=>"city", "pop"=>594, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>38.712121, "lon"=>-85.374281),
"/US/US-KY/minor-lane-heights-8291" => array("type"=>"city", "pop"=>1675, "name_ru"=>"Майнор Лейн Хайтс", "name_en"=>"Minor Lane Heights", "lang"=>"en", "lat"=>38.123256, "lon"=>-85.724787),
"/US/US-KY/mockingbird-valley-8292" => array("type"=>"city", "pop"=>243, "name_ru"=>"Мокингбёрд Валли", "name_en"=>"Mockingbird Valley", "lang"=>"en", "lat"=>38.268450, "lon"=>-85.680725),
"/US/US-KY/monterey-8293" => array("type"=>"city", "pop"=>177, "name_ru"=>"Монтерей", "name_en"=>"Monterey", "lang"=>"en", "lat"=>38.424961, "lon"=>-84.871290),
"/US/US-KY/monticello-8294" => array("type"=>"city", "pop"=>6162, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>36.840481, "lon"=>-84.847596),
"/US/US-KY/moorland-8295" => array("type"=>"city", "pop"=>501, "name_ru"=>"Мурленд", "name_en"=>"Moorland", "lang"=>"en", "lat"=>38.272550, "lon"=>-85.580299),
"/US/US-KY/morehead-8296" => array("type"=>"city", "pop"=>7707, "name_ru"=>"Морхэд", "name_en"=>"Morehead", "lang"=>"en", "lat"=>38.190146, "lon"=>-83.442203),
"/US/US-KY/morganfield-8297" => array("type"=>"city", "pop"=>3294, "name_ru"=>"Морганфилд", "name_en"=>"Morganfield", "lang"=>"en", "lat"=>37.684239, "lon"=>-87.906882),
"/US/US-KY/morgantown-8298" => array("type"=>"city", "pop"=>2564, "name_ru"=>"Моргантаун", "name_en"=>"Morgantown", "lang"=>"en", "lat"=>37.218016, "lon"=>-86.701849),
"/US/US-KY/mortons-gap-8299" => array("type"=>"city", "pop"=>937, "name_ru"=>"Мортонс Гэп", "name_en"=>"Mortons Gap", "lang"=>"en", "lat"=>37.237081, "lon"=>-87.471846),
"/US/US-KY/mount-olivet-8300" => array("type"=>"city", "pop"=>278, "name_ru"=>"Маунт Оливет", "name_en"=>"Mount Olivet", "lang"=>"en", "lat"=>38.531753, "lon"=>-84.037240),
"/US/US-KY/mount-sterling-8301" => array("type"=>"city", "pop"=>6921, "name_ru"=>"Маунт Стерлинг", "name_en"=>"Mount Sterling", "lang"=>"en", "lat"=>38.060366, "lon"=>-83.945261),
"/US/US-KY/mount-vernon-8302" => array("type"=>"city", "pop"=>2611, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>37.363073, "lon"=>-84.344211),
"/US/US-KY/mount-washington-8303" => array("type"=>"city", "pop"=>11980, "name_ru"=>"Маунт Вашингтон", "name_en"=>"Mount Washington", "lang"=>"en", "lat"=>38.047552, "lon"=>-85.550146),
"/US/US-KY/muldraugh-8304" => array("type"=>"city", "pop"=>1223, "name_ru"=>"Малдро", "name_en"=>"Muldraugh", "lang"=>"en", "lat"=>37.936904, "lon"=>-85.991450),
"/US/US-KY/munfordville-8305" => array("type"=>"city", "pop"=>1611, "name_ru"=>"Манфордвилл", "name_en"=>"Munfordville", "lang"=>"en", "lat"=>37.281010, "lon"=>-85.899298),
"/US/US-KY/murray-8306" => array("type"=>"city", "pop"=>16557, "name_ru"=>"Мюррей", "name_en"=>"Murray", "lang"=>"en", "lat"=>36.611477, "lon"=>-88.315674),
"/US/US-KY/murray-hill-8307" => array("type"=>"city", "pop"=>667, "name_ru"=>"Мюррей Хилл", "name_en"=>"Murray Hill", "lang"=>"en", "lat"=>38.289833, "lon"=>-85.585907),
"/US/US-KY/nebo-8308" => array("type"=>"city", "pop"=>216, "name_ru"=>"Небо", "name_en"=>"Nebo", "lang"=>"en", "lat"=>37.384381, "lon"=>-87.641423),
"/US/US-KY/new-castle-8310" => array("type"=>"city", "pop"=>909, "name_ru"=>"Нью Касл", "name_en"=>"New Castle", "lang"=>"en", "lat"=>38.433680, "lon"=>-85.171066),
"/US/US-KY/new-haven-8311" => array("type"=>"city", "pop"=>882, "name_ru"=>"Нью Хейвен", "name_en"=>"New Haven", "lang"=>"en", "lat"=>37.659187, "lon"=>-85.589287),
"/US/US-KY/newburg-8309" => array("type"=>"cdp", "pop"=>21647, "name_ru"=>"Ньюбург", "name_en"=>"Newburg", "lang"=>"en", "lat"=>38.171783, "lon"=>-85.686630),
"/US/US-KY/newport-8312" => array("type"=>"city", "pop"=>15766, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>39.085001, "lon"=>-84.486619),
"/US/US-KY/nicholasville-8313" => array("type"=>"city", "pop"=>26444, "name_ru"=>"Николасвилл", "name_en"=>"Nicholasville", "lang"=>"en", "lat"=>37.878256, "lon"=>-84.580040),
"/US/US-KY/norbourne-estates-8314" => array("type"=>"city", "pop"=>492, "name_ru"=>"Норборн Эстейтс", "name_en"=>"Norbourne Estates", "lang"=>"en", "lat"=>38.246900, "lon"=>-85.647125),
"/US/US-KY/north-corbin-8315" => array("type"=>"cdp", "pop"=>1601, "name_ru"=>"Норт Корбин", "name_en"=>"North Corbin", "lang"=>"en", "lat"=>36.966459, "lon"=>-84.094720),
"/US/US-KY/north-middletown-8317" => array("type"=>"city", "pop"=>556, "name_ru"=>"Норт Миддлтаун", "name_en"=>"North Middletown", "lang"=>"en", "lat"=>38.140969, "lon"=>-84.110569),
"/US/US-KY/northfield-8316" => array("type"=>"city", "pop"=>1092, "name_ru"=>"Нортфилд", "name_en"=>"Northfield", "lang"=>"en", "lat"=>38.286250, "lon"=>-85.636415),
"/US/US-KY/nortonville-8318" => array("type"=>"city", "pop"=>1234, "name_ru"=>"Нортонвилл", "name_en"=>"Nortonville", "lang"=>"en", "lat"=>37.186343, "lon"=>-87.457364),
"/US/US-KY/norwood-8319" => array("type"=>"city", "pop"=>435, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>38.253000, "lon"=>-85.610599),
"/US/US-KY/oak-grove-8321" => array("type"=>"city", "pop"=>9205, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>36.668000, "lon"=>-87.427261),
"/US/US-KY/oakbrook-8320" => array("type"=>"cdp", "pop"=>4113, "name_ru"=>"Оукбрук", "name_en"=>"Oakbrook", "lang"=>"en", "lat"=>38.999594, "lon"=>-84.685271),
"/US/US-KY/oakland-8322" => array("type"=>"city", "pop"=>262, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>37.040933, "lon"=>-86.252333),
"/US/US-KY/okolona-8323" => array("type"=>"cdp", "pop"=>18902, "name_ru"=>"Околона", "name_en"=>"Okolona", "lang"=>"en", "lat"=>38.139000, "lon"=>-85.686046),
"/US/US-KY/old-brownsboro-place-8324" => array("type"=>"city", "pop"=>431, "name_ru"=>"Оулд Браунсборо Плейс", "name_en"=>"Old Brownsboro Place", "lang"=>"en", "lat"=>38.289450, "lon"=>-85.613690),
"/US/US-KY/olive-hill-8325" => array("type"=>"city", "pop"=>1819, "name_ru"=>"Олив Хилл", "name_en"=>"Olive Hill", "lang"=>"en", "lat"=>38.303237, "lon"=>-83.169383),
"/US/US-KY/orchard-grass-hills-8326" => array("type"=>"city", "pop"=>1444, "name_ru"=>"Орчард Грасс Хиллс", "name_en"=>"Orchard Grass Hills", "lang"=>"en", "lat"=>38.324688, "lon"=>-85.523381),
"/US/US-KY/owensboro-8327" => array("type"=>"city", "pop"=>55516, "name_ru"=>"Оуэнсборо", "name_en"=>"Owensboro", "lang"=>"en", "lat"=>37.759350, "lon"=>-87.119319),
"/US/US-KY/owenton-8328" => array("type"=>"city", "pop"=>1476, "name_ru"=>"Оуэнтон", "name_en"=>"Owenton", "lang"=>"en", "lat"=>38.538573, "lon"=>-84.839658),
"/US/US-KY/owingsville-8329" => array("type"=>"city", "pop"=>1672, "name_ru"=>"Оуингсвилл", "name_en"=>"Owingsville", "lang"=>"en", "lat"=>38.136940, "lon"=>-83.759484),
"/US/US-KY/paducah-8330" => array("type"=>"city", "pop"=>25521, "name_ru"=>"Падака", "name_en"=>"Paducah", "lang"=>"en", "lat"=>37.072586, "lon"=>-88.635996),
"/US/US-KY/paintsville-8331" => array("type"=>"city", "pop"=>4216, "name_ru"=>"Пейнтсвилл", "name_en"=>"Paintsville", "lang"=>"en", "lat"=>37.817095, "lon"=>-82.807143),
"/US/US-KY/paris-8332" => array("type"=>"city", "pop"=>9286, "name_ru"=>"Парис", "name_en"=>"Paris", "lang"=>"en", "lat"=>38.206760, "lon"=>-84.260993),
"/US/US-KY/park-city-8333" => array("type"=>"city", "pop"=>545, "name_ru"=>"Парк Сити", "name_en"=>"Park City", "lang"=>"en", "lat"=>37.093870, "lon"=>-86.049481),
"/US/US-KY/park-hills-8334" => array("type"=>"city", "pop"=>2782, "name_ru"=>"Парк Хиллс", "name_en"=>"Park Hills", "lang"=>"en", "lat"=>39.070201, "lon"=>-84.530912),
"/US/US-KY/park-lake-8335" => array("type"=>"city", "pop"=>263, "name_ru"=>"Парк Лейк", "name_en"=>"Park Lake", "lang"=>"en", "lat"=>38.345286, "lon"=>-85.474541),
"/US/US-KY/parkway-village-8336" => array("type"=>"city", "pop"=>752, "name_ru"=>"Паркуэй Виллидж", "name_en"=>"Parkway Village", "lang"=>"en", "lat"=>38.211330, "lon"=>-85.738440),
"/US/US-KY/pembroke-8337" => array("type"=>"city", "pop"=>952, "name_ru"=>"Пемброук", "name_en"=>"Pembroke", "lang"=>"en", "lat"=>36.773989, "lon"=>-87.358755),
"/US/US-KY/perryville-8338" => array("type"=>"city", "pop"=>759, "name_ru"=>"Перривилл", "name_en"=>"Perryville", "lang"=>"en", "lat"=>37.649852, "lon"=>-84.951371),
"/US/US-KY/pewee-valley-8339" => array("type"=>"city", "pop"=>1606, "name_ru"=>"Пьюи Валли", "name_en"=>"Pewee Valley", "lang"=>"en", "lat"=>38.311960, "lon"=>-85.489300),
"/US/US-KY/phelps-8340" => array("type"=>"cdp", "pop"=>1298, "name_ru"=>"Фэлпс", "name_en"=>"Phelps", "lang"=>"en", "lat"=>37.509089, "lon"=>-82.160147),
"/US/US-KY/pikeville-8341" => array("type"=>"city", "pop"=>6413, "name_ru"=>"Пайквилл", "name_en"=>"Pikeville", "lang"=>"en", "lat"=>37.477164, "lon"=>-82.513846),
"/US/US-KY/pine-knot-8342" => array("type"=>"cdp", "pop"=>1549, "name_ru"=>"Пайн Нот", "name_en"=>"Pine Knot", "lang"=>"en", "lat"=>36.663935, "lon"=>-84.438877),
"/US/US-KY/pineville-8343" => array("type"=>"city", "pop"=>1990, "name_ru"=>"Пайнвилл", "name_en"=>"Pineville", "lang"=>"en", "lat"=>36.762839, "lon"=>-83.701272),
"/US/US-KY/pioneer-village-8344" => array("type"=>"city", "pop"=>2712, "name_ru"=>"Пайонир Виллидж", "name_en"=>"Pioneer Village", "lang"=>"en", "lat"=>38.059221, "lon"=>-85.677939),
"/US/US-KY/pippa-passes-8345" => array("type"=>"city", "pop"=>447, "name_ru"=>"Пиппа Пассис", "name_en"=>"Pippa Passes", "lang"=>"en", "lat"=>37.334061, "lon"=>-82.874351),
"/US/US-KY/plantation-8346" => array("type"=>"city", "pop"=>971, "name_ru"=>"Плэнтейшен", "name_en"=>"Plantation", "lang"=>"en", "lat"=>38.283100, "lon"=>-85.594011),
"/US/US-KY/pleasure-ridge-park-8347" => array("type"=>"cdp", "pop"=>25131, "name_ru"=>"Плежер Ридж Парк", "name_en"=>"Pleasure Ridge Park", "lang"=>"en", "lat"=>38.153700, "lon"=>-85.852874),
"/US/US-KY/pleasureville-8348" => array("type"=>"city", "pop"=>874, "name_ru"=>"Плежервилл", "name_en"=>"Pleasureville", "lang"=>"en", "lat"=>38.351913, "lon"=>-85.111648),
"/US/US-KY/plum-springs-8349" => array("type"=>"city", "pop"=>479, "name_ru"=>"Плам Спрингс", "name_en"=>"Plum Springs", "lang"=>"en", "lat"=>37.021377, "lon"=>-86.382758),
"/US/US-KY/plymouth-village-8350" => array("type"=>"city", "pop"=>162, "name_ru"=>"Плимут Виллидж", "name_en"=>"Plymouth Village", "lang"=>"en", "lat"=>38.240400, "lon"=>-85.649239),
"/US/US-KY/poplar-hills-8351" => array("type"=>"city", "pop"=>417, "name_ru"=>"Поплар Хиллс", "name_en"=>"Poplar Hills", "lang"=>"en", "lat"=>38.176421, "lon"=>-85.693532),
"/US/US-KY/powderly-8352" => array("type"=>"city", "pop"=>879, "name_ru"=>"Паудерили", "name_en"=>"Powderly", "lang"=>"en", "lat"=>37.235424, "lon"=>-87.162310),
"/US/US-KY/prestonsburg-8353" => array("type"=>"city", "pop"=>3862, "name_ru"=>"Престонсбург", "name_en"=>"Prestonsburg", "lang"=>"en", "lat"=>37.673650, "lon"=>-82.766648),
"/US/US-KY/prestonville-8354" => array("type"=>"city", "pop"=>176, "name_ru"=>"Престонвилл", "name_en"=>"Prestonville", "lang"=>"en", "lat"=>38.679479, "lon"=>-85.195156),
"/US/US-KY/princeton-8355" => array("type"=>"city", "pop"=>6398, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>37.107021, "lon"=>-87.884689),
"/US/US-KY/prospect-8356" => array("type"=>"city", "pop"=>5802, "name_ru"=>"Проспект", "name_en"=>"Prospect", "lang"=>"en", "lat"=>38.347771, "lon"=>-85.609314),
"/US/US-KY/providence-8357" => array("type"=>"city", "pop"=>3419, "name_ru"=>"Провиденс", "name_en"=>"Providence", "lang"=>"en", "lat"=>37.399198, "lon"=>-87.750210),
"/US/US-KY/raceland-8358" => array("type"=>"city", "pop"=>2601, "name_ru"=>"Рейсленд", "name_en"=>"Raceland", "lang"=>"en", "lat"=>38.536201, "lon"=>-82.733693),
"/US/US-KY/radcliff-8359" => array("type"=>"city", "pop"=>22013, "name_ru"=>"Рэдклифф", "name_en"=>"Radcliff", "lang"=>"en", "lat"=>37.824751, "lon"=>-85.940066),
"/US/US-KY/ravenna-8360" => array("type"=>"city", "pop"=>670, "name_ru"=>"Равенна", "name_en"=>"Ravenna", "lang"=>"en", "lat"=>37.687185, "lon"=>-83.947824),
"/US/US-KY/raywick-8361" => array("type"=>"city", "pop"=>151, "name_ru"=>"Рейвик", "name_en"=>"Raywick", "lang"=>"en", "lat"=>37.559746, "lon"=>-85.431182),
"/US/US-KY/reidland-8362" => array("type"=>"cdp", "pop"=>4054, "name_ru"=>"Райдленд", "name_en"=>"Reidland", "lang"=>"en", "lat"=>37.007602, "lon"=>-88.525554),
"/US/US-KY/richlawn-8363" => array("type"=>"city", "pop"=>490, "name_ru"=>"Ричлон", "name_en"=>"Richlawn", "lang"=>"en", "lat"=>38.255250, "lon"=>-85.641278),
"/US/US-KY/richmond-8364" => array("type"=>"city", "pop"=>32895, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>37.742752, "lon"=>-84.293637),
"/US/US-KY/river-bluff-8365" => array("type"=>"city", "pop"=>442, "name_ru"=>"Ривер Блафф", "name_en"=>"River Bluff", "lang"=>"en", "lat"=>38.369327, "lon"=>-85.602820),
"/US/US-KY/riverwood-8366" => array("type"=>"city", "pop"=>531, "name_ru"=>"Ривервуд", "name_en"=>"Riverwood", "lang"=>"en", "lat"=>38.283450, "lon"=>-85.662006),
"/US/US-KY/robinswood-8367" => array("type"=>"city", "pop"=>250, "name_ru"=>"Робинсвуд", "name_en"=>"Robinswood", "lang"=>"en", "lat"=>38.284717, "lon"=>-85.650891),
"/US/US-KY/rochester-8368" => array("type"=>"city", "pop"=>187, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>37.208572, "lon"=>-86.891063),
"/US/US-KY/rockport-8369" => array("type"=>"city", "pop"=>346, "name_ru"=>"Рокпорт", "name_en"=>"Rockport", "lang"=>"en", "lat"=>37.336384, "lon"=>-86.992783),
"/US/US-KY/rolling-fields-8370" => array("type"=>"city", "pop"=>724, "name_ru"=>"Роллинг Филдс", "name_en"=>"Rolling Fields", "lang"=>"en", "lat"=>38.269050, "lon"=>-85.670156),
"/US/US-KY/rolling-hills-8371" => array("type"=>"city", "pop"=>984, "name_ru"=>"Роллинг Хилс", "name_en"=>"Rolling Hills", "lang"=>"en", "lat"=>38.282600, "lon"=>-85.577810),
"/US/US-KY/russell-8372" => array("type"=>"city", "pop"=>3589, "name_ru"=>"Расселл", "name_en"=>"Russell", "lang"=>"en", "lat"=>38.509696, "lon"=>-82.701190),
"/US/US-KY/russell-springs-8373" => array("type"=>"city", "pop"=>2327, "name_ru"=>"Рассел Спрингс", "name_en"=>"Russell Springs", "lang"=>"en", "lat"=>37.050441, "lon"=>-85.078075),
"/US/US-KY/russellville-8374" => array("type"=>"city", "pop"=>7268, "name_ru"=>"Расселлвилл", "name_en"=>"Russellville", "lang"=>"en", "lat"=>36.839482, "lon"=>-86.893639),
"/US/US-KY/ryland-heights-8375" => array("type"=>"city", "pop"=>827, "name_ru"=>"Райленд Хайтс", "name_en"=>"Ryland Heights", "lang"=>"en", "lat"=>38.957960, "lon"=>-84.469318),
"/US/US-KY/sacramento-8376" => array("type"=>"city", "pop"=>502, "name_ru"=>"Сакраменто", "name_en"=>"Sacramento", "lang"=>"en", "lat"=>37.416733, "lon"=>-87.267148),
"/US/US-KY/sadieville-8377" => array("type"=>"city", "pop"=>319, "name_ru"=>"Сэдивилл", "name_en"=>"Sadieville", "lang"=>"en", "lat"=>38.391284, "lon"=>-84.536465),
"/US/US-KY/salem-8382" => array("type"=>"city", "pop"=>753, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>37.265502, "lon"=>-88.239486),
"/US/US-KY/salt-lick-8383" => array("type"=>"city", "pop"=>360, "name_ru"=>"Солт Лик", "name_en"=>"Salt Lick", "lang"=>"en", "lat"=>38.119475, "lon"=>-83.616091),
"/US/US-KY/salyersville-8384" => array("type"=>"city", "pop"=>1569, "name_ru"=>"Сейлерсвилл", "name_en"=>"Salyersville", "lang"=>"en", "lat"=>37.746807, "lon"=>-83.063720),
"/US/US-KY/sanders-8385" => array("type"=>"city", "pop"=>258, "name_ru"=>"Сэндерс", "name_en"=>"Sanders", "lang"=>"en", "lat"=>38.654648, "lon"=>-84.946757),
"/US/US-KY/sandy-hook-8386" => array("type"=>"city", "pop"=>723, "name_ru"=>"Сэнди Хук", "name_en"=>"Sandy Hook", "lang"=>"en", "lat"=>38.092791, "lon"=>-83.123306),
"/US/US-KY/sardis-8387" => array("type"=>"city", "pop"=>185, "name_ru"=>"Сардис", "name_en"=>"Sardis", "lang"=>"en", "lat"=>38.534187, "lon"=>-83.957529),
"/US/US-KY/science-hill-8388" => array("type"=>"city", "pop"=>670, "name_ru"=>"Сайнс Хилл", "name_en"=>"Science Hill", "lang"=>"en", "lat"=>37.176021, "lon"=>-84.636405),
"/US/US-KY/scottsville-8389" => array("type"=>"city", "pop"=>4581, "name_ru"=>"Скоттсвилл", "name_en"=>"Scottsville", "lang"=>"en", "lat"=>36.749561, "lon"=>-86.196415),
"/US/US-KY/sebree-8390" => array("type"=>"city", "pop"=>1507, "name_ru"=>"Себри", "name_en"=>"Sebree", "lang"=>"en", "lat"=>37.604031, "lon"=>-87.528616),
"/US/US-KY/seneca-gardens-8391" => array("type"=>"city", "pop"=>753, "name_ru"=>"Сенека Гарденс", "name_en"=>"Seneca Gardens", "lang"=>"en", "lat"=>38.227800, "lon"=>-85.676634),
"/US/US-KY/sharpsburg-8392" => array("type"=>"city", "pop"=>345, "name_ru"=>"Шарпсбург", "name_en"=>"Sharpsburg", "lang"=>"en", "lat"=>38.202111, "lon"=>-83.928393),
"/US/US-KY/shelbyville-8393" => array("type"=>"city", "pop"=>11294, "name_ru"=>"Шелбивилл", "name_en"=>"Shelbyville", "lang"=>"en", "lat"=>38.217584, "lon"=>-85.231209),
"/US/US-KY/shepherdsville-8394" => array("type"=>"city", "pop"=>9203, "name_ru"=>"Шефердсвилл", "name_en"=>"Shepherdsville", "lang"=>"en", "lat"=>37.987414, "lon"=>-85.713166),
"/US/US-KY/shively-8395" => array("type"=>"city", "pop"=>16641, "name_ru"=>"Шивели", "name_en"=>"Shively", "lang"=>"en", "lat"=>38.195900, "lon"=>-85.813199),
"/US/US-KY/silver-grove-8396" => array("type"=>"city", "pop"=>1153, "name_ru"=>"Силвер Гров", "name_en"=>"Silver Grove", "lang"=>"en", "lat"=>39.036733, "lon"=>-84.396867),
"/US/US-KY/simpsonville-8397" => array("type"=>"city", "pop"=>1436, "name_ru"=>"Симпсонвилл", "name_en"=>"Simpsonville", "lang"=>"en", "lat"=>38.218304, "lon"=>-85.353101),
"/US/US-KY/slaughters-8398" => array("type"=>"city", "pop"=>232, "name_ru"=>"Слотерс", "name_en"=>"Slaughters", "lang"=>"en", "lat"=>37.491304, "lon"=>-87.501225),
"/US/US-KY/smithfield-8399" => array("type"=>"city", "pop"=>101, "name_ru"=>"Смитфилд", "name_en"=>"Smithfield", "lang"=>"en", "lat"=>38.387855, "lon"=>-85.257311),
"/US/US-KY/smithland-8400" => array("type"=>"city", "pop"=>388, "name_ru"=>"Смитленд", "name_en"=>"Smithland", "lang"=>"en", "lat"=>37.137790, "lon"=>-88.403337),
"/US/US-KY/smiths-grove-8401" => array("type"=>"city", "pop"=>781, "name_ru"=>"Смитс Гров", "name_en"=>"Smiths Grove", "lang"=>"en", "lat"=>37.050289, "lon"=>-86.208117),
"/US/US-KY/somerset-8402" => array("type"=>"city", "pop"=>12449, "name_ru"=>"Сомерсет", "name_en"=>"Somerset", "lang"=>"en", "lat"=>37.082338, "lon"=>-84.609803),
"/US/US-KY/sonora-8403" => array("type"=>"city", "pop"=>331, "name_ru"=>"Сонора", "name_en"=>"Sonora", "lang"=>"en", "lat"=>37.523537, "lon"=>-85.891454),
"/US/US-KY/south-carrollton-8404" => array("type"=>"city", "pop"=>181, "name_ru"=>"Саут Кэрроллтон", "name_en"=>"South Carrollton", "lang"=>"en", "lat"=>37.336380, "lon"=>-87.140390),
"/US/US-KY/south-park-view-8406" => array("type"=>"city", "pop"=>233, "name_ru"=>"Саут Парк Вью", "name_en"=>"South Park View", "lang"=>"en", "lat"=>38.119201, "lon"=>-85.718207),
"/US/US-KY/south-shore-8407" => array("type"=>"city", "pop"=>1252, "name_ru"=>"Саут Шор", "name_en"=>"South Shore", "lang"=>"en", "lat"=>38.722008, "lon"=>-82.964288),
"/US/US-KY/south-wallins-8408" => array("type"=>"cdp", "pop"=>1022, "name_ru"=>"Саут Уоллинс", "name_en"=>"South Wallins", "lang"=>"en", "lat"=>36.807446, "lon"=>-83.406303),
"/US/US-KY/southgate-8405" => array("type"=>"city", "pop"=>3285, "name_ru"=>"Саутгейт", "name_en"=>"Southgate", "lang"=>"en", "lat"=>39.061285, "lon"=>-84.470771),
"/US/US-KY/sparta-8409" => array("type"=>"city", "pop"=>207, "name_ru"=>"Спарта", "name_en"=>"Sparta", "lang"=>"en", "lat"=>38.676327, "lon"=>-84.903842),
"/US/US-KY/spring-mill-8412" => array("type"=>"city", "pop"=>419, "name_ru"=>"Спринг Милл", "name_en"=>"Spring Mill", "lang"=>"en", "lat"=>38.143718, "lon"=>-85.629870),
"/US/US-KY/spring-valley-8413" => array("type"=>"city", "pop"=>754, "name_ru"=>"Спринг Валли", "name_en"=>"Spring Valley", "lang"=>"en", "lat"=>38.298992, "lon"=>-85.609959),
"/US/US-KY/springfield-8410" => array("type"=>"city", "pop"=>2877, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>37.687661, "lon"=>-85.222838),
"/US/US-KY/springlee-8411" => array("type"=>"city", "pop"=>451, "name_ru"=>"Спрингли", "name_en"=>"Springlee", "lang"=>"en", "lat"=>38.241150, "lon"=>-85.642285),
"/US/US-KY/st.-charles-8378" => array("type"=>"city", "pop"=>307, "name_ru"=>"Сент-Чарльз", "name_en"=>"St. Charles", "lang"=>"en", "lat"=>37.188009, "lon"=>-87.556870),
"/US/US-KY/st.-dennis-8379" => array("type"=>"cdp", "pop"=>10326, "name_ru"=>"Сент Деннис", "name_en"=>"St. Dennis", "lang"=>"en", "lat"=>38.188392, "lon"=>-85.845933),
"/US/US-KY/st.-matthews-8380" => array("type"=>"city", "pop"=>18709, "name_ru"=>"Сент Мэтьюс", "name_en"=>"St. Matthews", "lang"=>"en", "lat"=>38.249150, "lon"=>-85.637801),
"/US/US-KY/st.-regis-park-8381" => array("type"=>"city", "pop"=>1665, "name_ru"=>"Сент Регис Парк", "name_en"=>"St. Regis Park", "lang"=>"en", "lat"=>38.229750, "lon"=>-85.614504),
"/US/US-KY/stamping-ground-8414" => array("type"=>"city", "pop"=>678, "name_ru"=>"Стэмпинг Граунд", "name_en"=>"Stamping Ground", "lang"=>"en", "lat"=>38.270035, "lon"=>-84.686092),
"/US/US-KY/stanford-8415" => array("type"=>"city", "pop"=>3399, "name_ru"=>"Стэнфорд", "name_en"=>"Stanford", "lang"=>"en", "lat"=>37.531861, "lon"=>-84.662689),
"/US/US-KY/stanton-8416" => array("type"=>"city", "pop"=>3149, "name_ru"=>"Стэнтон", "name_en"=>"Stanton", "lang"=>"en", "lat"=>37.846913, "lon"=>-83.857242),
"/US/US-KY/stearns-8417" => array("type"=>"cdp", "pop"=>1550, "name_ru"=>"Стернс", "name_en"=>"Stearns", "lang"=>"en", "lat"=>36.694975, "lon"=>-84.478032),
"/US/US-KY/strathmoor-gardens-8418" => array("type"=>"city", "pop"=>300, "name_ru"=>"Стрэтмур Гарденс", "name_en"=>"Strathmoor Gardens", "lang"=>"en", "lat"=>38.218592, "lon"=>-85.676859),
"/US/US-KY/strathmoor-manor-8419" => array("type"=>"city", "pop"=>354, "name_ru"=>"Стрэтмур Мэнор", "name_en"=>"Strathmoor Manor", "lang"=>"en", "lat"=>38.219250, "lon"=>-85.683551),
"/US/US-KY/strathmoor-village-8420" => array("type"=>"city", "pop"=>693, "name_ru"=>"Стрэтмур Виллидж", "name_en"=>"Strathmoor Village", "lang"=>"en", "lat"=>38.221350, "lon"=>-85.678857),
"/US/US-KY/sturgis-8421" => array("type"=>"city", "pop"=>1925, "name_ru"=>"Стерджис", "name_en"=>"Sturgis", "lang"=>"en", "lat"=>37.545892, "lon"=>-87.987338),
"/US/US-KY/sycamore-8422" => array("type"=>"city", "pop"=>284, "name_ru"=>"Сикамор", "name_en"=>"Sycamore", "lang"=>"en", "lat"=>38.246640, "lon"=>-85.560722),
"/US/US-KY/taylor-mill-8423" => array("type"=>"city", "pop"=>6746, "name_ru"=>"Тейлор Милл", "name_en"=>"Taylor Mill", "lang"=>"en", "lat"=>39.011761, "lon"=>-84.497210),
"/US/US-KY/taylorsville-8424" => array("type"=>"city", "pop"=>1232, "name_ru"=>"Тейлорсвилл", "name_en"=>"Taylorsville", "lang"=>"en", "lat"=>38.034774, "lon"=>-85.343844),
"/US/US-KY/ten-broeck-8425" => array("type"=>"city", "pop"=>195, "name_ru"=>"Тен Броек", "name_en"=>"Ten Broeck", "lang"=>"en", "lat"=>38.295500, "lon"=>-85.579305),
"/US/US-KY/thornhill-8426" => array("type"=>"city", "pop"=>192, "name_ru"=>"Торнхилл", "name_en"=>"Thornhill", "lang"=>"en", "lat"=>38.288100, "lon"=>-85.625231),
"/US/US-KY/tompkinsville-8427" => array("type"=>"city", "pop"=>2609, "name_ru"=>"Томпкинсвилл", "name_en"=>"Tompkinsville", "lang"=>"en", "lat"=>36.700704, "lon"=>-85.692157),
"/US/US-KY/trenton-8428" => array("type"=>"city", "pop"=>427, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>36.722770, "lon"=>-87.263245),
"/US/US-KY/union-8429" => array("type"=>"city", "pop"=>3612, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>38.952911, "lon"=>-84.669166),
"/US/US-KY/uniontown-8430" => array("type"=>"city", "pop"=>1020, "name_ru"=>"Юнионтаун", "name_en"=>"Uniontown", "lang"=>"en", "lat"=>37.773279, "lon"=>-87.932244),
"/US/US-KY/upton-8431" => array("type"=>"city", "pop"=>616, "name_ru"=>"Аптон", "name_en"=>"Upton", "lang"=>"en", "lat"=>37.464584, "lon"=>-85.898625),
"/US/US-KY/valley-station-8432" => array("type"=>"cdp", "pop"=>22840, "name_ru"=>"Валли Стейшен", "name_en"=>"Valley Station", "lang"=>"en", "lat"=>38.103450, "lon"=>-85.858579),
"/US/US-KY/van-lear-8434" => array("type"=>"cdp", "pop"=>1050, "name_ru"=>"Ван Лир", "name_en"=>"Van Lear", "lang"=>"en", "lat"=>37.769941, "lon"=>-82.756393),
"/US/US-KY/vanceburg-8433" => array("type"=>"city", "pop"=>1698, "name_ru"=>"Вансбург", "name_en"=>"Vanceburg", "lang"=>"en", "lat"=>38.594228, "lon"=>-83.320959),
"/US/US-KY/versailles-8435" => array("type"=>"city", "pop"=>7828, "name_ru"=>"Версайллес", "name_en"=>"Versailles", "lang"=>"en", "lat"=>38.054440, "lon"=>-84.729358),
"/US/US-KY/vicco-8436" => array("type"=>"city", "pop"=>314, "name_ru"=>"Викко", "name_en"=>"Vicco", "lang"=>"en", "lat"=>37.216090, "lon"=>-83.061201),
"/US/US-KY/villa-hills-8437" => array("type"=>"city", "pop"=>7727, "name_ru"=>"Вилла Хиллс", "name_en"=>"Villa Hills", "lang"=>"en", "lat"=>39.063933, "lon"=>-84.596330),
"/US/US-KY/vine-grove-8438" => array("type"=>"city", "pop"=>4362, "name_ru"=>"Вайн Гров", "name_en"=>"Vine Grove", "lang"=>"en", "lat"=>37.811222, "lon"=>-85.976192),
"/US/US-KY/visalia-8439" => array("type"=>"city", "pop"=>190, "name_ru"=>"Визалия", "name_en"=>"Visalia", "lang"=>"en", "lat"=>38.915006, "lon"=>-84.451254),
"/US/US-KY/wallins-creek-8440" => array("type"=>"city", "pop"=>233, "name_ru"=>"Уоллинс Крик", "name_en"=>"Wallins Creek", "lang"=>"en", "lat"=>36.827949, "lon"=>-83.414541),
"/US/US-KY/walton-8441" => array("type"=>"city", "pop"=>3038, "name_ru"=>"Уолтон", "name_en"=>"Walton", "lang"=>"en", "lat"=>38.866167, "lon"=>-84.615580),
"/US/US-KY/warfield-8442" => array("type"=>"city", "pop"=>260, "name_ru"=>"Ворфилд", "name_en"=>"Warfield", "lang"=>"en", "lat"=>37.838673, "lon"=>-82.423169),
"/US/US-KY/warsaw-8443" => array("type"=>"city", "pop"=>1794, "name_ru"=>"Варшава", "name_en"=>"Warsaw", "lang"=>"en", "lat"=>38.785303, "lon"=>-84.894804),
"/US/US-KY/washington-8444" => array("type"=>"city", "pop"=>11595, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>38.615892, "lon"=>-83.808053),
"/US/US-KY/water-valley-8445" => array("type"=>"city", "pop"=>321, "name_ru"=>"Уотер Валли", "name_en"=>"Water Valley", "lang"=>"en", "lat"=>36.568331, "lon"=>-88.810794),
"/US/US-KY/watterson-park-8446" => array("type"=>"city", "pop"=>1243, "name_ru"=>"Уоттерсон Парк", "name_en"=>"Watterson Park", "lang"=>"en", "lat"=>38.190643, "lon"=>-85.688538),
"/US/US-KY/waverly-8447" => array("type"=>"city", "pop"=>280, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>37.709695, "lon"=>-87.815258),
"/US/US-KY/wayland-8448" => array("type"=>"city", "pop"=>291, "name_ru"=>"Уэйленд", "name_en"=>"Wayland", "lang"=>"en", "lat"=>37.448190, "lon"=>-82.800473),
"/US/US-KY/wellington-8449" => array("type"=>"city", "pop"=>596, "name_ru"=>"Веллингтон", "name_en"=>"Wellington", "lang"=>"en", "lat"=>38.216700, "lon"=>-85.670078),
"/US/US-KY/west-buechel-8450" => array("type"=>"city", "pop"=>1509, "name_ru"=>"Вэст Буэшел", "name_en"=>"West Buechel", "lang"=>"en", "lat"=>38.195367, "lon"=>-85.670282),
"/US/US-KY/west-liberty-8451" => array("type"=>"city", "pop"=>3285, "name_ru"=>"Вэст Либерти", "name_en"=>"West Liberty", "lang"=>"en", "lat"=>37.913405, "lon"=>-83.265661),
"/US/US-KY/west-point-8452" => array("type"=>"city", "pop"=>974, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>37.985020, "lon"=>-85.954114),
"/US/US-KY/westwood-8453" => array("type"=>"city", "pop"=>668, "name_ru"=>"Уэствуд", "name_en"=>"Westwood", "lang"=>"en", "lat"=>38.279800, "lon"=>-85.585393),
"/US/US-KY/westwood-8454" => array("type"=>"cdp", "pop"=>5300, "name_ru"=>"Уэствуд", "name_en"=>"Westwood", "lang"=>"en", "lat"=>38.481168, "lon"=>-82.677953),
"/US/US-KY/wheatcroft-8455" => array("type"=>"city", "pop"=>168, "name_ru"=>"Уиткрофт", "name_en"=>"Wheatcroft", "lang"=>"en", "lat"=>37.489303, "lon"=>-87.861182),
"/US/US-KY/wheelwright-8456" => array("type"=>"city", "pop"=>1029, "name_ru"=>"Уилрайт", "name_en"=>"Wheelwright", "lang"=>"en", "lat"=>37.338942, "lon"=>-82.720273),
"/US/US-KY/whipps-millgate-8457" => array("type"=>"city", "pop"=>454, "name_ru"=>"Уипс Милгейт", "name_en"=>"Whipps Millgate", "lang"=>"en", "lat"=>38.273600, "lon"=>-85.576734),
"/US/US-KY/white-plains-8458" => array("type"=>"city", "pop"=>790, "name_ru"=>"Уайт Плейнс", "name_en"=>"White Plains", "lang"=>"en", "lat"=>37.187689, "lon"=>-87.386429),
"/US/US-KY/whitesburg-8459" => array("type"=>"city", "pop"=>1478, "name_ru"=>"Уайтсбург", "name_en"=>"Whitesburg", "lang"=>"en", "lat"=>37.118114, "lon"=>-82.825677),
"/US/US-KY/whitesville-8460" => array("type"=>"city", "pop"=>595, "name_ru"=>"Уайтсвилл", "name_en"=>"Whitesville", "lang"=>"en", "lat"=>37.683868, "lon"=>-86.869783),
"/US/US-KY/whitley-city-8461" => array("type"=>"cdp", "pop"=>1133, "name_ru"=>"Уати Сити", "name_en"=>"Whitley City", "lang"=>"en", "lat"=>36.724652, "lon"=>-84.475416),
"/US/US-KY/wickliffe-8462" => array("type"=>"city", "pop"=>794, "name_ru"=>"Уиклайфф", "name_en"=>"Wickliffe", "lang"=>"en", "lat"=>36.966278, "lon"=>-89.081482),
"/US/US-KY/wilder-8463" => array("type"=>"city", "pop"=>2922, "name_ru"=>"Уайлдер", "name_en"=>"Wilder", "lang"=>"en", "lat"=>39.035550, "lon"=>-84.478572),
"/US/US-KY/wildwood-8464" => array("type"=>"city", "pop"=>272, "name_ru"=>"Уайлдвуд", "name_en"=>"Wildwood", "lang"=>"en", "lat"=>38.250400, "lon"=>-85.574345),
"/US/US-KY/williamsburg-8465" => array("type"=>"city", "pop"=>5204, "name_ru"=>"Уилльямсбург", "name_en"=>"Williamsburg", "lang"=>"en", "lat"=>36.738323, "lon"=>-84.167462),
"/US/US-KY/williamstown-8466" => array("type"=>"city", "pop"=>3533, "name_ru"=>"Уильямстаун", "name_en"=>"Williamstown", "lang"=>"en", "lat"=>38.637528, "lon"=>-84.560940),
"/US/US-KY/willisburg-8467" => array("type"=>"city", "pop"=>368, "name_ru"=>"Уиллисбург", "name_en"=>"Willisburg", "lang"=>"en", "lat"=>37.810070, "lon"=>-85.121961),
"/US/US-KY/wilmore-8468" => array("type"=>"city", "pop"=>6004, "name_ru"=>"Уилмор", "name_en"=>"Wilmore", "lang"=>"en", "lat"=>37.866594, "lon"=>-84.658951),
"/US/US-KY/winchester-8469" => array("type"=>"city", "pop"=>16598, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>37.997293, "lon"=>-84.188284),
"/US/US-KY/winding-falls-8470" => array("type"=>"city", "pop"=>657, "name_ru"=>"Уиндинг Фоллс", "name_en"=>"Winding Falls", "lang"=>"en", "lat"=>38.282950, "lon"=>-85.643762),
"/US/US-KY/windy-hills-8471" => array("type"=>"city", "pop"=>2778, "name_ru"=>"Уинди Хилс", "name_en"=>"Windy Hills", "lang"=>"en", "lat"=>38.270167, "lon"=>-85.634707),
"/US/US-KY/wingo-8472" => array("type"=>"city", "pop"=>594, "name_ru"=>"Уинго", "name_en"=>"Wingo", "lang"=>"en", "lat"=>36.641905, "lon"=>-88.737889),
"/US/US-KY/woodburn-8473" => array("type"=>"city", "pop"=>349, "name_ru"=>"Вудберн", "name_en"=>"Woodburn", "lang"=>"en", "lat"=>36.841076, "lon"=>-86.529290),
"/US/US-KY/woodbury-8474" => array("type"=>"city", "pop"=>87, "name_ru"=>"Вудбери", "name_en"=>"Woodbury", "lang"=>"en", "lat"=>37.183040, "lon"=>-86.635047),
"/US/US-KY/woodland-hills-8475" => array("type"=>"city", "pop"=>735, "name_ru"=>"Вудленд Хилс", "name_en"=>"Woodland Hills", "lang"=>"en", "lat"=>38.239200, "lon"=>-85.529489),
"/US/US-KY/woodlawn-8476" => array("type"=>"city", "pop"=>249, "name_ru"=>"Вудлон", "name_en"=>"Woodlawn", "lang"=>"en", "lat"=>39.090751, "lon"=>-84.472032),
"/US/US-KY/woodlawn-oakdale-8477" => array("type"=>"cdp", "pop"=>4954, "name_ru"=>"Вудлон-Оукдейл", "name_en"=>"Woodlawn-Oakdale", "lang"=>"en", "lat"=>37.038228, "lon"=>-88.573260),
"/US/US-KY/woodlawn-park-8478" => array("type"=>"city", "pop"=>1128, "name_ru"=>"Вудлон Парк", "name_en"=>"Woodlawn Park", "lang"=>"en", "lat"=>38.261750, "lon"=>-85.630205),
"/US/US-KY/worthington-8479" => array("type"=>"city", "pop"=>1676, "name_ru"=>"Уортингтон", "name_en"=>"Worthington", "lang"=>"en", "lat"=>38.551712, "lon"=>-82.736294),
"/US/US-KY/worthington-hills-8480" => array("type"=>"city", "pop"=>1740, "name_ru"=>"Уортингтон Хилс", "name_en"=>"Worthington Hills", "lang"=>"en", "lat"=>38.310650, "lon"=>-85.527078),
"/US/US-KY/worthville-8481" => array("type"=>"city", "pop"=>227, "name_ru"=>"Уортвилл", "name_en"=>"Worthville", "lang"=>"en", "lat"=>38.609399, "lon"=>-85.068489),
"/US/US-KY/wurtland-8482" => array("type"=>"city", "pop"=>1047, "name_ru"=>"Уертленд", "name_en"=>"Wurtland", "lang"=>"en", "lat"=>38.550355, "lon"=>-82.773220),
"/US/US-LA" => array("type"=>"state", "pop"=>4410796, "name_ru"=>"Луизиана", "name_en"=>"Louisiana", "lang"=>"en"),
"/US/US-LA/abbeville-8483" => array("type"=>"city", "pop"=>12103, "name_ru"=>"Аббевилл", "name_en"=>"Abbeville", "lang"=>"en", "lat"=>29.974206, "lon"=>-92.123607),
"/US/US-LA/abita-springs-8484" => array("type"=>"town", "pop"=>2409, "name_ru"=>"Абита Спрингс", "name_en"=>"Abita Springs", "lang"=>"en", "lat"=>30.477606, "lon"=>-90.031319),
"/US/US-LA/addis-8485" => array("type"=>"town", "pop"=>3569, "name_ru"=>"Аддис", "name_en"=>"Addis", "lang"=>"en", "lat"=>30.354465, "lon"=>-91.269446),
"/US/US-LA/albany-8486" => array("type"=>"village", "pop"=>1131, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>30.503500, "lon"=>-90.584517),
"/US/US-LA/alexandria-8487" => array("type"=>"city", "pop"=>48639, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>31.291950, "lon"=>-92.463306),
"/US/US-LA/amelia-8488" => array("type"=>"cdp", "pop"=>2447, "name_ru"=>"Амелия", "name_en"=>"Amelia", "lang"=>"en", "lat"=>29.663690, "lon"=>-91.106814),
"/US/US-LA/amite-city-8489" => array("type"=>"town", "pop"=>4318, "name_ru"=>"Амит Сити", "name_en"=>"Amite City", "lang"=>"en", "lat"=>30.733392, "lon"=>-90.509992),
"/US/US-LA/anacoco-8490" => array("type"=>"village", "pop"=>729, "name_ru"=>"Анакоко", "name_en"=>"Anacoco", "lang"=>"en", "lat"=>31.255743, "lon"=>-93.343918),
"/US/US-LA/angie-8491" => array("type"=>"village", "pop"=>229, "name_ru"=>"Энджи", "name_en"=>"Angie", "lang"=>"en", "lat"=>30.963949, "lon"=>-89.815808),
"/US/US-LA/arabi-8492" => array("type"=>"cdp", "pop"=>8787, "name_ru"=>"Араби", "name_en"=>"Arabi", "lang"=>"en", "lat"=>29.954984, "lon"=>-89.998512),
"/US/US-LA/arcadia-8493" => array("type"=>"town", "pop"=>2731, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>32.550966, "lon"=>-92.924554),
"/US/US-LA/arnaudville-8494" => array("type"=>"town", "pop"=>1406, "name_ru"=>"Арнодвилл", "name_en"=>"Arnaudville", "lang"=>"en", "lat"=>30.401578, "lon"=>-91.932266),
"/US/US-LA/ashland-8495" => array("type"=>"village", "pop"=>288, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>32.116270, "lon"=>-93.114587),
"/US/US-LA/athens-8496" => array("type"=>"village", "pop"=>244, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>32.651131, "lon"=>-93.025904),
"/US/US-LA/atlanta-8497" => array("type"=>"village", "pop"=>137, "name_ru"=>"Атланта", "name_en"=>"Atlanta", "lang"=>"en", "lat"=>31.805585, "lon"=>-92.736858),
"/US/US-LA/avondale-8498" => array("type"=>"cdp", "pop"=>5813, "name_ru"=>"Авондейл", "name_en"=>"Avondale", "lang"=>"en", "lat"=>29.902121, "lon"=>-90.191570),
"/US/US-LA/baker-8499" => array("type"=>"city", "pop"=>13360, "name_ru"=>"Бейкер", "name_en"=>"Baker", "lang"=>"en", "lat"=>30.585450, "lon"=>-91.156989),
"/US/US-LA/baldwin-8500" => array("type"=>"town", "pop"=>2613, "name_ru"=>"Болдуин", "name_en"=>"Baldwin", "lang"=>"en", "lat"=>29.833446, "lon"=>-91.553740),
"/US/US-LA/ball-8501" => array("type"=>"town", "pop"=>3751, "name_ru"=>"Болл", "name_en"=>"Ball", "lang"=>"en", "lat"=>31.418250, "lon"=>-92.410051),
"/US/US-LA/barataria-8502" => array("type"=>"cdp", "pop"=>1160, "name_ru"=>"Баратария", "name_en"=>"Barataria", "lang"=>"en", "lat"=>29.715585, "lon"=>-90.124123),
"/US/US-LA/basile-8503" => array("type"=>"town", "pop"=>2377, "name_ru"=>"Бейзил", "name_en"=>"Basile", "lang"=>"en", "lat"=>30.485667, "lon"=>-92.601929),
"/US/US-LA/baskin-8504" => array("type"=>"village", "pop"=>177, "name_ru"=>"Баскин", "name_en"=>"Baskin", "lang"=>"en", "lat"=>32.259326, "lon"=>-91.747796),
"/US/US-LA/bastrop-8505" => array("type"=>"city", "pop"=>11814, "name_ru"=>"Бастроп", "name_en"=>"Bastrop", "lang"=>"en", "lat"=>32.774481, "lon"=>-91.909335),
"/US/US-LA/baton-rouge-8506" => array("type"=>"city", "pop"=>223689, "name_ru"=>"Батон-Руж", "name_en"=>"Baton Rouge", "lang"=>"en", "lat"=>30.448967, "lon"=>-91.126043),
"/US/US-LA/bayou-cane-8507" => array("type"=>"cdp", "pop"=>15876, "name_ru"=>"Байу-Кейн", "name_en"=>"Bayou Cane", "lang"=>"en", "lat"=>29.623867, "lon"=>-90.751082),
"/US/US-LA/bayou-vista-8508" => array("type"=>"cdp", "pop"=>4733, "name_ru"=>"Байу Виста", "name_en"=>"Bayou Vista", "lang"=>"en", "lat"=>29.692250, "lon"=>-91.267078),
"/US/US-LA/belcher-8509" => array("type"=>"village", "pop"=>280, "name_ru"=>"Белчер", "name_en"=>"Belcher", "lang"=>"en", "lat"=>32.753276, "lon"=>-93.835111),
"/US/US-LA/belle-chasse-8510" => array("type"=>"cdp", "pop"=>8512, "name_ru"=>"Бель Час", "name_en"=>"Belle Chasse", "lang"=>"en", "lat"=>29.845881, "lon"=>-90.003942),
"/US/US-LA/benton-8511" => array("type"=>"town", "pop"=>2845, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>32.691850, "lon"=>-93.740102),
"/US/US-LA/bernice-8512" => array("type"=>"town", "pop"=>1643, "name_ru"=>"Бернис", "name_en"=>"Bernice", "lang"=>"en", "lat"=>32.827303, "lon"=>-92.659157),
"/US/US-LA/berwick-8513" => array("type"=>"town", "pop"=>4281, "name_ru"=>"Бервик", "name_en"=>"Berwick", "lang"=>"en", "lat"=>29.700294, "lon"=>-91.235778),
"/US/US-LA/bienville-8514" => array("type"=>"village", "pop"=>246, "name_ru"=>"Бинвилл", "name_en"=>"Bienville", "lang"=>"en", "lat"=>32.361296, "lon"=>-92.976342),
"/US/US-LA/blanchard-8515" => array("type"=>"town", "pop"=>2608, "name_ru"=>"Бланчард", "name_en"=>"Blanchard", "lang"=>"en", "lat"=>32.591963, "lon"=>-93.886144),
"/US/US-LA/bogalusa-8516" => array("type"=>"city", "pop"=>12607, "name_ru"=>"Богалуса", "name_en"=>"Bogalusa", "lang"=>"en", "lat"=>30.782723, "lon"=>-89.864092),
"/US/US-LA/bonita-8517" => array("type"=>"village", "pop"=>305, "name_ru"=>"Бонита", "name_en"=>"Bonita", "lang"=>"en", "lat"=>32.919944, "lon"=>-91.674933),
"/US/US-LA/boothville-venice-8518" => array("type"=>"cdp", "pop"=>2743, "name_ru"=>"Бутвилл-Венис", "name_en"=>"Boothville-Venice", "lang"=>"en", "lat"=>29.323286, "lon"=>-89.395795),
"/US/US-LA/bossier-city-8519" => array("type"=>"city", "pop"=>62384, "name_ru"=>"Бошьер Сити", "name_en"=>"Bossier City", "lang"=>"en", "lat"=>32.518643, "lon"=>-93.663767),
"/US/US-LA/boutte-8520" => array("type"=>"cdp", "pop"=>2702, "name_ru"=>"Баутт", "name_en"=>"Boutte", "lang"=>"en", "lat"=>29.896162, "lon"=>-90.397537),
"/US/US-LA/boyce-8521" => array("type"=>"town", "pop"=>1183, "name_ru"=>"Бойс", "name_en"=>"Boyce", "lang"=>"en", "lat"=>31.390000, "lon"=>-92.669357),
"/US/US-LA/breaux-bridge-8522" => array("type"=>"city", "pop"=>8047, "name_ru"=>"Бро Бридж", "name_en"=>"Breaux Bridge", "lang"=>"en", "lat"=>30.278200, "lon"=>-91.901010),
"/US/US-LA/bridge-city-8523" => array("type"=>"cdp", "pop"=>8327, "name_ru"=>"Бридж Сити", "name_en"=>"Bridge City", "lang"=>"en", "lat"=>29.933768, "lon"=>-90.161419),
"/US/US-LA/broadmoor-8524" => array("type"=>"cdp", "pop"=>3218, "name_ru"=>"Бродмур", "name_en"=>"Broadmoor", "lang"=>"en", "lat"=>30.178000, "lon"=>-92.062153),
"/US/US-LA/broussard-8525" => array("type"=>"town", "pop"=>7767, "name_ru"=>"Брассард", "name_en"=>"Broussard", "lang"=>"en", "lat"=>30.150050, "lon"=>-91.970050),
"/US/US-LA/brownfields-8526" => array("type"=>"cdp", "pop"=>5229, "name_ru"=>"Браунфилдс", "name_en"=>"Brownfields", "lang"=>"en", "lat"=>30.546450, "lon"=>-91.120649),
"/US/US-LA/brownsville-bawcomville-8527" => array("type"=>"cdp", "pop"=>7397, "name_ru"=>"Браунсвилл-Баукомвилл", "name_en"=>"Brownsville-Bawcomville", "lang"=>"en", "lat"=>32.483750, "lon"=>-92.163860),
"/US/US-LA/brusly-8528" => array("type"=>"town", "pop"=>2159, "name_ru"=>"Брашли", "name_en"=>"Brusly", "lang"=>"en", "lat"=>30.394660, "lon"=>-91.251663),
"/US/US-LA/bryceland-8529" => array("type"=>"village", "pop"=>108, "name_ru"=>"Брайсленд", "name_en"=>"Bryceland", "lang"=>"en", "lat"=>32.443754, "lon"=>-92.989647),
"/US/US-LA/bunkie-8530" => array("type"=>"city", "pop"=>4465, "name_ru"=>"Банки", "name_en"=>"Bunkie", "lang"=>"en", "lat"=>30.953558, "lon"=>-92.186702),
"/US/US-LA/buras-triumph-8531" => array("type"=>"cdp", "pop"=>3702, "name_ru"=>"Бурас-Триумф", "name_en"=>"Buras-Triumph", "lang"=>"en", "lat"=>29.342110, "lon"=>-89.496457),
"/US/US-LA/calvin-8532" => array("type"=>"village", "pop"=>218, "name_ru"=>"Калвин", "name_en"=>"Calvin", "lang"=>"en", "lat"=>31.959748, "lon"=>-92.778301),
"/US/US-LA/cameron-8533" => array("type"=>"cdp", "pop"=>7238, "name_ru"=>"Кэмерон", "name_en"=>"Cameron", "lang"=>"en", "lat"=>29.797616, "lon"=>-93.299571),
"/US/US-LA/campti-8534" => array("type"=>"town", "pop"=>1069, "name_ru"=>"Кэмпти", "name_en"=>"Campti", "lang"=>"en", "lat"=>31.897893, "lon"=>-93.115014),
"/US/US-LA/cankton-8535" => array("type"=>"village", "pop"=>418, "name_ru"=>"Кэнктон", "name_en"=>"Cankton", "lang"=>"en", "lat"=>30.348950, "lon"=>-92.111449),
"/US/US-LA/carencro-8536" => array("type"=>"city", "pop"=>6653, "name_ru"=>"Каренкро", "name_en"=>"Carencro", "lang"=>"en", "lat"=>30.310500, "lon"=>-92.036579),
"/US/US-LA/carlyss-8537" => array("type"=>"cdp", "pop"=>3305, "name_ru"=>"Карлисс", "name_en"=>"Carlyss", "lang"=>"en", "lat"=>30.175800, "lon"=>-93.369599),
"/US/US-LA/carville-8538" => array("type"=>"cdp", "pop"=>1108, "name_ru"=>"Карвилл", "name_en"=>"Carville", "lang"=>"en", "lat"=>30.223935, "lon"=>-91.104521),
"/US/US-LA/castor-8539" => array("type"=>"village", "pop"=>205, "name_ru"=>"Кастор", "name_en"=>"Castor", "lang"=>"en", "lat"=>32.254142, "lon"=>-93.166959),
"/US/US-LA/cecilia-8540" => array("type"=>"cdp", "pop"=>1374, "name_ru"=>"Сесилия", "name_en"=>"Cecilia", "lang"=>"en", "lat"=>30.336148, "lon"=>-91.847758),
"/US/US-LA/chackbay-8541" => array("type"=>"cdp", "pop"=>2276, "name_ru"=>"Чакбей", "name_en"=>"Chackbay", "lang"=>"en", "lat"=>29.887506, "lon"=>-90.784328),
"/US/US-LA/chalmette-8542" => array("type"=>"cdp", "pop"=>31860, "name_ru"=>"Чалметт", "name_en"=>"Chalmette", "lang"=>"en", "lat"=>29.944151, "lon"=>-89.967296),
"/US/US-LA/charenton-8543" => array("type"=>"cdp", "pop"=>1584, "name_ru"=>"Чарентон", "name_en"=>"Charenton", "lang"=>"en", "lat"=>29.868545, "lon"=>-91.536370),
"/US/US-LA/chataignier-8544" => array("type"=>"village", "pop"=>379, "name_ru"=>"Чатейньер", "name_en"=>"Chataignier", "lang"=>"en", "lat"=>30.567473, "lon"=>-92.315392),
"/US/US-LA/chatham-8545" => array("type"=>"town", "pop"=>617, "name_ru"=>"Чатем", "name_en"=>"Chatham", "lang"=>"en", "lat"=>32.308883, "lon"=>-92.451215),
"/US/US-LA/chauvin-8546" => array("type"=>"cdp", "pop"=>3375, "name_ru"=>"Човин", "name_en"=>"Chauvin", "lang"=>"en", "lat"=>29.445492, "lon"=>-90.594431),
"/US/US-LA/cheneyville-8547" => array("type"=>"town", "pop"=>881, "name_ru"=>"Ченивилл", "name_en"=>"Cheneyville", "lang"=>"en", "lat"=>31.010850, "lon"=>-92.289888),
"/US/US-LA/choudrant-8548" => array("type"=>"village", "pop"=>719, "name_ru"=>"Чаудрант", "name_en"=>"Choudrant", "lang"=>"en", "lat"=>32.530336, "lon"=>-92.517665),
"/US/US-LA/church-point-8549" => array("type"=>"town", "pop"=>4661, "name_ru"=>"Черч Пойнт", "name_en"=>"Church Point", "lang"=>"en", "lat"=>30.403854, "lon"=>-92.214462),
"/US/US-LA/claiborne-8550" => array("type"=>"cdp", "pop"=>16142, "name_ru"=>"Клэрборн", "name_en"=>"Claiborne", "lang"=>"en", "lat"=>32.542200, "lon"=>-92.196734),
"/US/US-LA/clarence-8551" => array("type"=>"village", "pop"=>499, "name_ru"=>"Клэренс", "name_en"=>"Clarence", "lang"=>"en", "lat"=>31.819980, "lon"=>-93.028612),
"/US/US-LA/clarks-8552" => array("type"=>"village", "pop"=>1049, "name_ru"=>"Кларкс", "name_en"=>"Clarks", "lang"=>"en", "lat"=>32.028616, "lon"=>-92.139663),
"/US/US-LA/clayton-8553" => array("type"=>"town", "pop"=>779, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>31.720952, "lon"=>-91.539728),
"/US/US-LA/clinton-8554" => array("type"=>"town", "pop"=>1891, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>30.861068, "lon"=>-91.015020),
"/US/US-LA/colfax-8555" => array("type"=>"town", "pop"=>1667, "name_ru"=>"Колфакс", "name_en"=>"Colfax", "lang"=>"en", "lat"=>31.518867, "lon"=>-92.705693),
"/US/US-LA/collinston-8556" => array("type"=>"village", "pop"=>302, "name_ru"=>"Коллинстон", "name_en"=>"Collinston", "lang"=>"en", "lat"=>32.688820, "lon"=>-91.871497),
"/US/US-LA/columbia-8557" => array("type"=>"town", "pop"=>450, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>32.101453, "lon"=>-92.077223),
"/US/US-LA/converse-8558" => array("type"=>"village", "pop"=>405, "name_ru"=>"Конверс", "name_en"=>"Converse", "lang"=>"en", "lat"=>31.780364, "lon"=>-93.699795),
"/US/US-LA/cotton-valley-8560" => array("type"=>"town", "pop"=>1140, "name_ru"=>"Коттон Валли", "name_en"=>"Cotton Valley", "lang"=>"en", "lat"=>32.812721, "lon"=>-93.422998),
"/US/US-LA/cottonport-8559" => array("type"=>"town", "pop"=>2246, "name_ru"=>"Коттонпорт", "name_en"=>"Cottonport", "lang"=>"en", "lat"=>30.988276, "lon"=>-92.052927),
"/US/US-LA/coushatta-8561" => array("type"=>"town", "pop"=>2088, "name_ru"=>"Кошатта", "name_en"=>"Coushatta", "lang"=>"en", "lat"=>32.024823, "lon"=>-93.340516),
"/US/US-LA/covington-8562" => array("type"=>"city", "pop"=>9155, "name_ru"=>"Ковингтон", "name_en"=>"Covington", "lang"=>"en", "lat"=>30.478602, "lon"=>-90.112087),
"/US/US-LA/crowley-8563" => array("type"=>"city", "pop"=>13922, "name_ru"=>"Кроули", "name_en"=>"Crowley", "lang"=>"en", "lat"=>30.212611, "lon"=>-92.375933),
"/US/US-LA/cullen-8564" => array("type"=>"town", "pop"=>1357, "name_ru"=>"Каллен", "name_en"=>"Cullen", "lang"=>"en", "lat"=>32.970056, "lon"=>-93.446346),
"/US/US-LA/cut-off-8565" => array("type"=>"cdp", "pop"=>5325, "name_ru"=>"Кат Оф", "name_en"=>"Cut Off", "lang"=>"en", "lat"=>29.529406, "lon"=>-90.334637),
"/US/US-LA/de-quincy-8570" => array("type"=>"city", "pop"=>3474, "name_ru"=>"Де Квинси", "name_en"=>"De Quincy", "lang"=>"en", "lat"=>30.448288, "lon"=>-93.443480),
"/US/US-LA/de-ridder-8571" => array("type"=>"city", "pop"=>9868, "name_ru"=>"Де Риддер", "name_en"=>"De Ridder", "lang"=>"en", "lat"=>30.845656, "lon"=>-93.290977),
"/US/US-LA/delcambre-8566" => array("type"=>"town", "pop"=>2161, "name_ru"=>"Делкамбр", "name_en"=>"Delcambre", "lang"=>"en", "lat"=>29.950090, "lon"=>-91.989608),
"/US/US-LA/delhi-8567" => array("type"=>"town", "pop"=>2988, "name_ru"=>"Дели", "name_en"=>"Delhi", "lang"=>"en", "lat"=>32.456405, "lon"=>-91.489221),
"/US/US-LA/delta-8568" => array("type"=>"village", "pop"=>221, "name_ru"=>"Дельта", "name_en"=>"Delta", "lang"=>"en", "lat"=>32.323844, "lon"=>-90.923231),
"/US/US-LA/denham-springs-8569" => array("type"=>"city", "pop"=>10308, "name_ru"=>"Денхэм Спрингс", "name_en"=>"Denham Springs", "lang"=>"en", "lat"=>30.479700, "lon"=>-90.959308),
"/US/US-LA/des-allemands-8572" => array("type"=>"cdp", "pop"=>2504, "name_ru"=>"Дес Аллемандс", "name_en"=>"Des Allemands", "lang"=>"en", "lat"=>29.801732, "lon"=>-90.481997),
"/US/US-LA/destrehan-8573" => array("type"=>"cdp", "pop"=>8031, "name_ru"=>"Дестрхен", "name_en"=>"Destrehan", "lang"=>"en", "lat"=>29.961795, "lon"=>-90.369046),
"/US/US-LA/deville-8574" => array("type"=>"cdp", "pop"=>1113, "name_ru"=>"Девилл", "name_en"=>"Deville", "lang"=>"en", "lat"=>31.346591, "lon"=>-92.157906),
"/US/US-LA/dixie-inn-8575" => array("type"=>"village", "pop"=>345, "name_ru"=>"Дикси Инн", "name_en"=>"Dixie Inn", "lang"=>"en", "lat"=>32.593977, "lon"=>-93.335032),
"/US/US-LA/dodson-8576" => array("type"=>"village", "pop"=>329, "name_ru"=>"Додсон", "name_en"=>"Dodson", "lang"=>"en", "lat"=>32.079172, "lon"=>-92.658163),
"/US/US-LA/donaldsonville-8577" => array("type"=>"city", "pop"=>7527, "name_ru"=>"Дональдсонвилл", "name_en"=>"Donaldsonville", "lang"=>"en", "lat"=>30.098052, "lon"=>-90.996683),
"/US/US-LA/downsville-8578" => array("type"=>"village", "pop"=>110, "name_ru"=>"Даунсвилл", "name_en"=>"Downsville", "lang"=>"en", "lat"=>32.626968, "lon"=>-92.414341),
"/US/US-LA/doyline-8579" => array("type"=>"village", "pop"=>816, "name_ru"=>"Дойлайн", "name_en"=>"Doyline", "lang"=>"en", "lat"=>32.518880, "lon"=>-93.415911),
"/US/US-LA/dry-prong-8580" => array("type"=>"village", "pop"=>441, "name_ru"=>"Драй Пронг", "name_en"=>"Dry Prong", "lang"=>"en", "lat"=>31.578645, "lon"=>-92.529345),
"/US/US-LA/dubach-8581" => array("type"=>"town", "pop"=>782, "name_ru"=>"Дабах", "name_en"=>"Dubach", "lang"=>"en", "lat"=>32.697998, "lon"=>-92.655952),
"/US/US-LA/dubberly-8582" => array("type"=>"village", "pop"=>280, "name_ru"=>"Дабберли", "name_en"=>"Dubberly", "lang"=>"en", "lat"=>32.539482, "lon"=>-93.237586),
"/US/US-LA/dulac-8583" => array("type"=>"cdp", "pop"=>3273, "name_ru"=>"Дулак", "name_en"=>"Dulac", "lang"=>"en", "lat"=>29.369785, "lon"=>-90.700356),
"/US/US-LA/duson-8584" => array("type"=>"town", "pop"=>1652, "name_ru"=>"Дасон", "name_en"=>"Duson", "lang"=>"en", "lat"=>30.232179, "lon"=>-92.191855),
"/US/US-LA/east-hodge-8585" => array("type"=>"village", "pop"=>356, "name_ru"=>"Ист Ходж", "name_en"=>"East Hodge", "lang"=>"en", "lat"=>32.276869, "lon"=>-92.713963),
"/US/US-LA/eastwood-8586" => array("type"=>"cdp", "pop"=>2987, "name_ru"=>"Иствуд", "name_en"=>"Eastwood", "lang"=>"en", "lat"=>32.561704, "lon"=>-93.566195),
"/US/US-LA/eden-isle-8587" => array("type"=>"cdp", "pop"=>3768, "name_ru"=>"Иден Айл", "name_en"=>"Eden Isle", "lang"=>"en", "lat"=>30.228292, "lon"=>-89.798707),
"/US/US-LA/edgard-8588" => array("type"=>"cdp", "pop"=>2753, "name_ru"=>"Эдгар", "name_en"=>"Edgard", "lang"=>"en", "lat"=>30.025064, "lon"=>-90.556745),
"/US/US-LA/edgefield-8589" => array("type"=>"village", "pop"=>181, "name_ru"=>"Эджфилд", "name_en"=>"Edgefield", "lang"=>"en", "lat"=>32.048787, "lon"=>-93.335794),
"/US/US-LA/elizabeth-8590" => array("type"=>"town", "pop"=>581, "name_ru"=>"Элизабет", "name_en"=>"Elizabeth", "lang"=>"en", "lat"=>30.868371, "lon"=>-92.797683),
"/US/US-LA/elton-8591" => array("type"=>"town", "pop"=>1237, "name_ru"=>"Элтон", "name_en"=>"Elton", "lang"=>"en", "lat"=>30.481577, "lon"=>-92.697065),
"/US/US-LA/empire-8592" => array("type"=>"cdp", "pop"=>2654, "name_ru"=>"Эмпаер", "name_en"=>"Empire", "lang"=>"en", "lat"=>29.411894, "lon"=>-89.610610),
"/US/US-LA/epps-8593" => array("type"=>"village", "pop"=>1101, "name_ru"=>"Эппс", "name_en"=>"Epps", "lang"=>"en", "lat"=>32.605015, "lon"=>-91.479526),
"/US/US-LA/erath-8594" => array("type"=>"town", "pop"=>2175, "name_ru"=>"Эрат", "name_en"=>"Erath", "lang"=>"en", "lat"=>29.958231, "lon"=>-92.037549),
"/US/US-LA/eros-8595" => array("type"=>"town", "pop"=>206, "name_ru"=>"Эрос", "name_en"=>"Eros", "lang"=>"en", "lat"=>32.392215, "lon"=>-92.424310),
"/US/US-LA/estelle-8596" => array("type"=>"cdp", "pop"=>14091, "name_ru"=>"Эстелл", "name_en"=>"Estelle", "lang"=>"en", "lat"=>29.843602, "lon"=>-90.099343),
"/US/US-LA/estherwood-8597" => array("type"=>"village", "pop"=>854, "name_ru"=>"Эстервуд", "name_en"=>"Estherwood", "lang"=>"en", "lat"=>30.181071, "lon"=>-92.463201),
"/US/US-LA/eunice-8598" => array("type"=>"city", "pop"=>11500, "name_ru"=>"Юнис", "name_en"=>"Eunice", "lang"=>"en", "lat"=>30.491080, "lon"=>-92.419072),
"/US/US-LA/evergreen-8599" => array("type"=>"town", "pop"=>310, "name_ru"=>"Эвергрин", "name_en"=>"Evergreen", "lang"=>"en", "lat"=>30.954021, "lon"=>-92.106022),
"/US/US-LA/farmerville-8600" => array("type"=>"town", "pop"=>3633, "name_ru"=>"Фармервилл", "name_en"=>"Farmerville", "lang"=>"en", "lat"=>32.773902, "lon"=>-92.400460),
"/US/US-LA/fenton-8601" => array("type"=>"village", "pop"=>374, "name_ru"=>"Фэнтон", "name_en"=>"Fenton", "lang"=>"en", "lat"=>30.363469, "lon"=>-92.916758),
"/US/US-LA/ferriday-8602" => array("type"=>"town", "pop"=>3571, "name_ru"=>"Ферридей", "name_en"=>"Ferriday", "lang"=>"en", "lat"=>31.634519, "lon"=>-91.555303),
"/US/US-LA/fisher-8603" => array("type"=>"village", "pop"=>266, "name_ru"=>"Фишер", "name_en"=>"Fisher", "lang"=>"en", "lat"=>31.494411, "lon"=>-93.460259),
"/US/US-LA/florien-8604" => array("type"=>"village", "pop"=>693, "name_ru"=>"Флориен", "name_en"=>"Florien", "lang"=>"en", "lat"=>31.449439, "lon"=>-93.459485),
"/US/US-LA/folsom-8605" => array("type"=>"village", "pop"=>667, "name_ru"=>"Фолсом", "name_en"=>"Folsom", "lang"=>"en", "lat"=>30.631905, "lon"=>-90.198048),
"/US/US-LA/fordoche-8606" => array("type"=>"village", "pop"=>941, "name_ru"=>"Фордош", "name_en"=>"Fordoche", "lang"=>"en", "lat"=>30.594712, "lon"=>-91.616725),
"/US/US-LA/forest-8607" => array("type"=>"village", "pop"=>256, "name_ru"=>"Форест", "name_en"=>"Forest", "lang"=>"en", "lat"=>32.792423, "lon"=>-91.412160),
"/US/US-LA/forest-hill-8608" => array("type"=>"village", "pop"=>460, "name_ru"=>"Форест Хилл", "name_en"=>"Forest Hill", "lang"=>"en", "lat"=>31.050050, "lon"=>-92.524607),
"/US/US-LA/fort-polk-north-8609" => array("type"=>"cdp", "pop"=>3819, "name_ru"=>"Форт Полк Норт", "name_en"=>"Fort Polk North", "lang"=>"en", "lat"=>31.112595, "lon"=>-93.171064),
"/US/US-LA/fort-polk-south-8610" => array("type"=>"cdp", "pop"=>10911, "name_ru"=>"Форт Полк Саут", "name_en"=>"Fort Polk South", "lang"=>"en", "lat"=>31.049528, "lon"=>-93.217192),
"/US/US-LA/franklin-8611" => array("type"=>"city", "pop"=>7680, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>29.800838, "lon"=>-91.514460),
"/US/US-LA/franklinton-8612" => array("type"=>"town", "pop"=>3748, "name_ru"=>"Франклинтон", "name_en"=>"Franklinton", "lang"=>"en", "lat"=>30.848779, "lon"=>-90.143638),
"/US/US-LA/french-settlement-8613" => array("type"=>"village", "pop"=>1057, "name_ru"=>"Фрэнч Сеттлмент", "name_en"=>"French Settlement", "lang"=>"en", "lat"=>30.311464, "lon"=>-90.803296),
"/US/US-LA/galliano-8614" => array("type"=>"cdp", "pop"=>4294, "name_ru"=>"Галлиано", "name_en"=>"Galliano", "lang"=>"en", "lat"=>29.430977, "lon"=>-90.302740),
"/US/US-LA/gardere-8615" => array("type"=>"cdp", "pop"=>7209, "name_ru"=>"Гардер", "name_en"=>"Gardere", "lang"=>"en", "lat"=>30.357200, "lon"=>-91.134581),
"/US/US-LA/garyville-8616" => array("type"=>"cdp", "pop"=>3181, "name_ru"=>"Гаривилл", "name_en"=>"Garyville", "lang"=>"en", "lat"=>30.077293, "lon"=>-90.624545),
"/US/US-LA/georgetown-8617" => array("type"=>"village", "pop"=>316, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>31.759299, "lon"=>-92.380966),
"/US/US-LA/gibsland-8618" => array("type"=>"town", "pop"=>1047, "name_ru"=>"Гибсленд", "name_en"=>"Gibsland", "lang"=>"en", "lat"=>32.538982, "lon"=>-93.053329),
"/US/US-LA/gilbert-8619" => array("type"=>"village", "pop"=>512, "name_ru"=>"Гилберт", "name_en"=>"Gilbert", "lang"=>"en", "lat"=>32.050615, "lon"=>-91.655371),
"/US/US-LA/gilliam-8620" => array("type"=>"village", "pop"=>180, "name_ru"=>"Гиллиам", "name_en"=>"Gilliam", "lang"=>"en", "lat"=>32.827255, "lon"=>-93.842724),
"/US/US-LA/glenmora-8621" => array("type"=>"town", "pop"=>1554, "name_ru"=>"Гленмора", "name_en"=>"Glenmora", "lang"=>"en", "lat"=>30.973200, "lon"=>-92.582774),
"/US/US-LA/golden-meadow-8622" => array("type"=>"town", "pop"=>2110, "name_ru"=>"Голден Медоу", "name_en"=>"Golden Meadow", "lang"=>"en", "lat"=>29.387666, "lon"=>-90.273402),
"/US/US-LA/goldonna-8623" => array("type"=>"village", "pop"=>457, "name_ru"=>"Голдонна", "name_en"=>"Goldonna", "lang"=>"en", "lat"=>32.020417, "lon"=>-92.911970),
"/US/US-LA/gonzales-8624" => array("type"=>"city", "pop"=>9390, "name_ru"=>"Гонзалес", "name_en"=>"Gonzales", "lang"=>"en", "lat"=>30.216807, "lon"=>-90.923629),
"/US/US-LA/grambling-8625" => array("type"=>"town", "pop"=>4516, "name_ru"=>"Грэмблинг", "name_en"=>"Grambling", "lang"=>"en", "lat"=>32.527019, "lon"=>-92.711646),
"/US/US-LA/gramercy-8626" => array("type"=>"town", "pop"=>3254, "name_ru"=>"Грэмерси", "name_en"=>"Gramercy", "lang"=>"en", "lat"=>30.059365, "lon"=>-90.695526),
"/US/US-LA/grand-cane-8627" => array("type"=>"village", "pop"=>197, "name_ru"=>"Гранд Кейн", "name_en"=>"Grand Cane", "lang"=>"en", "lat"=>32.084513, "lon"=>-93.809428),
"/US/US-LA/grand-coteau-8628" => array("type"=>"town", "pop"=>1047, "name_ru"=>"Гранд Кото", "name_en"=>"Grand Coteau", "lang"=>"en", "lat"=>30.418158, "lon"=>-92.043835),
"/US/US-LA/grand-isle-8629" => array("type"=>"town", "pop"=>1619, "name_ru"=>"Гранд Айл", "name_en"=>"Grand Isle", "lang"=>"en", "lat"=>29.209852, "lon"=>-90.032785),
"/US/US-LA/gray-8630" => array("type"=>"cdp", "pop"=>4260, "name_ru"=>"Грей", "name_en"=>"Gray", "lang"=>"en", "lat"=>29.676895, "lon"=>-90.781601),
"/US/US-LA/grayson-8631" => array("type"=>"village", "pop"=>510, "name_ru"=>"Грейсон", "name_en"=>"Grayson", "lang"=>"en", "lat"=>32.048482, "lon"=>-92.111531),
"/US/US-LA/greensburg-8632" => array("type"=>"town", "pop"=>618, "name_ru"=>"Гринсбург", "name_en"=>"Greensburg", "lang"=>"en", "lat"=>30.829552, "lon"=>-90.670044),
"/US/US-LA/greenwood-8633" => array("type"=>"town", "pop"=>2759, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>32.441448, "lon"=>-93.973819),
"/US/US-LA/gretna-8634" => array("type"=>"city", "pop"=>16399, "name_ru"=>"Гретна", "name_en"=>"Gretna", "lang"=>"en", "lat"=>29.913968, "lon"=>-90.052502),
"/US/US-LA/grosse-tete-8635" => array("type"=>"village", "pop"=>632, "name_ru"=>"Гросс Тет", "name_en"=>"Grosse Tete", "lang"=>"en", "lat"=>30.414393, "lon"=>-91.435156),
"/US/US-LA/gueydan-8636" => array("type"=>"town", "pop"=>1602, "name_ru"=>"Гейдан", "name_en"=>"Gueydan", "lang"=>"en", "lat"=>30.028878, "lon"=>-92.508505),
"/US/US-LA/hackberry-8637" => array("type"=>"cdp", "pop"=>1664, "name_ru"=>"Хэкберри", "name_en"=>"Hackberry", "lang"=>"en", "lat"=>29.968284, "lon"=>-93.433840),
"/US/US-LA/hahnville-8638" => array("type"=>"cdp", "pop"=>2599, "name_ru"=>"Ханвилл", "name_en"=>"Hahnville", "lang"=>"en", "lat"=>29.972950, "lon"=>-90.421030),
"/US/US-LA/hall-summit-8639" => array("type"=>"village", "pop"=>255, "name_ru"=>"Холл Саммит", "name_en"=>"Hall Summit", "lang"=>"en", "lat"=>32.176860, "lon"=>-93.304948),
"/US/US-LA/hammond-8640" => array("type"=>"city", "pop"=>19803, "name_ru"=>"Хаммонд", "name_en"=>"Hammond", "lang"=>"en", "lat"=>30.507653, "lon"=>-90.460110),
"/US/US-LA/harahan-8641" => array("type"=>"city", "pop"=>9306, "name_ru"=>"Харахан", "name_en"=>"Harahan", "lang"=>"en", "lat"=>29.936951, "lon"=>-90.203655),
"/US/US-LA/harrisonburg-8642" => array("type"=>"village", "pop"=>730, "name_ru"=>"Харрисонбург", "name_en"=>"Harrisonburg", "lang"=>"en", "lat"=>31.767023, "lon"=>-91.823779),
"/US/US-LA/harvey-8643" => array("type"=>"cdp", "pop"=>21222, "name_ru"=>"Харви", "name_en"=>"Harvey", "lang"=>"en", "lat"=>29.886034, "lon"=>-90.067275),
"/US/US-LA/haughton-8644" => array("type"=>"town", "pop"=>3012, "name_ru"=>"Хотон", "name_en"=>"Haughton", "lang"=>"en", "lat"=>32.530072, "lon"=>-93.506256),
"/US/US-LA/haynesville-8645" => array("type"=>"town", "pop"=>2429, "name_ru"=>"Хейнсвилл", "name_en"=>"Haynesville", "lang"=>"en", "lat"=>32.966880, "lon"=>-93.137829),
"/US/US-LA/heflin-8646" => array("type"=>"village", "pop"=>239, "name_ru"=>"Хефлин", "name_en"=>"Heflin", "lang"=>"en", "lat"=>32.457867, "lon"=>-93.267421),
"/US/US-LA/henderson-8647" => array("type"=>"town", "pop"=>1585, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>30.314256, "lon"=>-91.799265),
"/US/US-LA/hessmer-8648" => array("type"=>"village", "pop"=>667, "name_ru"=>"Хессмер", "name_en"=>"Hessmer", "lang"=>"en", "lat"=>31.053623, "lon"=>-92.121137),
"/US/US-LA/hodge-8649" => array("type"=>"village", "pop"=>475, "name_ru"=>"Ходж", "name_en"=>"Hodge", "lang"=>"en", "lat"=>32.271651, "lon"=>-92.729109),
"/US/US-LA/homer-8650" => array("type"=>"town", "pop"=>3385, "name_ru"=>"Гомер", "name_en"=>"Homer", "lang"=>"en", "lat"=>32.793567, "lon"=>-93.058679),
"/US/US-LA/hornbeck-8651" => array("type"=>"town", "pop"=>376, "name_ru"=>"Хорнбек", "name_en"=>"Hornbeck", "lang"=>"en", "lat"=>31.326695, "lon"=>-93.396229),
"/US/US-LA/hosston-8652" => array("type"=>"village", "pop"=>389, "name_ru"=>"Хосстон", "name_en"=>"Hosston", "lang"=>"en", "lat"=>32.883714, "lon"=>-93.882800),
"/US/US-LA/houma-8653" => array("type"=>"city", "pop"=>32512, "name_ru"=>"Хаума", "name_en"=>"Houma", "lang"=>"en", "lat"=>29.578733, "lon"=>-90.706916),
"/US/US-LA/ida-8654" => array("type"=>"village", "pop"=>262, "name_ru"=>"Ида", "name_en"=>"Ida", "lang"=>"en", "lat"=>33.002103, "lon"=>-93.894968),
"/US/US-LA/independence-8655" => array("type"=>"town", "pop"=>1795, "name_ru"=>"Индепенденс", "name_en"=>"Independence", "lang"=>"en", "lat"=>30.635474, "lon"=>-90.505060),
"/US/US-LA/inniswold-8656" => array("type"=>"cdp", "pop"=>3474, "name_ru"=>"Иннисволд", "name_en"=>"Inniswold", "lang"=>"en", "lat"=>30.398600, "lon"=>-91.073220),
"/US/US-LA/iota-8657" => array("type"=>"town", "pop"=>1403, "name_ru"=>"Иота", "name_en"=>"Iota", "lang"=>"en", "lat"=>30.326496, "lon"=>-92.494800),
"/US/US-LA/iowa-8658" => array("type"=>"town", "pop"=>2620, "name_ru"=>"Айова", "name_en"=>"Iowa", "lang"=>"en", "lat"=>30.237224, "lon"=>-93.012678),
"/US/US-LA/jackson-8659" => array("type"=>"town", "pop"=>3933, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>30.834196, "lon"=>-91.210204),
"/US/US-LA/jamestown-8660" => array("type"=>"village", "pop"=>139, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>32.337608, "lon"=>-93.205315),
"/US/US-LA/jean-lafitte-8662" => array("type"=>"town", "pop"=>2265, "name_ru"=>"Джин Лафитт", "name_en"=>"Jean Lafitte", "lang"=>"en", "lat"=>29.749475, "lon"=>-90.106125),
"/US/US-LA/jeanerette-8661" => array("type"=>"city", "pop"=>5889, "name_ru"=>"Жанеретт", "name_en"=>"Jeanerette", "lang"=>"en", "lat"=>29.916320, "lon"=>-91.675349),
"/US/US-LA/jefferson-8663" => array("type"=>"cdp", "pop"=>436181, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>29.958190, "lon"=>-90.172608),
"/US/US-LA/jena-8664" => array("type"=>"town", "pop"=>2868, "name_ru"=>"Джена", "name_en"=>"Jena", "lang"=>"en", "lat"=>31.687557, "lon"=>-92.125753),
"/US/US-LA/jennings-8665" => array("type"=>"city", "pop"=>10529, "name_ru"=>"Дженнингс", "name_en"=>"Jennings", "lang"=>"en", "lat"=>30.222915, "lon"=>-92.658353),
"/US/US-LA/jonesboro-8666" => array("type"=>"town", "pop"=>3724, "name_ru"=>"Джонсборо", "name_en"=>"Jonesboro", "lang"=>"en", "lat"=>32.236795, "lon"=>-92.709974),
"/US/US-LA/jonesville-8667" => array("type"=>"town", "pop"=>2317, "name_ru"=>"Джонсвилл", "name_en"=>"Jonesville", "lang"=>"en", "lat"=>31.622767, "lon"=>-91.830484),
"/US/US-LA/junction-city-8668" => array("type"=>"village", "pop"=>603, "name_ru"=>"Джанкшен Сити", "name_en"=>"Junction City", "lang"=>"en", "lat"=>33.009046, "lon"=>-92.722767),
"/US/US-LA/kaplan-8669" => array("type"=>"city", "pop"=>5053, "name_ru"=>"Каплан", "name_en"=>"Kaplan", "lang"=>"en", "lat"=>30.003407, "lon"=>-92.284318),
"/US/US-LA/keachi-8670" => array("type"=>"town", "pop"=>335, "name_ru"=>"Кичи", "name_en"=>"Keachi", "lang"=>"en", "lat"=>32.175712, "lon"=>-93.916249),
"/US/US-LA/kenner-8671" => array("type"=>"city", "pop"=>66942, "name_ru"=>"Кеннер", "name_en"=>"Kenner", "lang"=>"en", "lat"=>30.128464, "lon"=>-90.227872),
"/US/US-LA/kentwood-8672" => array("type"=>"town", "pop"=>2274, "name_ru"=>"Кентвуд", "name_en"=>"Kentwood", "lang"=>"en", "lat"=>30.934665, "lon"=>-90.515259),
"/US/US-LA/kilbourne-8673" => array("type"=>"village", "pop"=>406, "name_ru"=>"Килборн", "name_en"=>"Kilbourne", "lang"=>"en", "lat"=>32.997769, "lon"=>-91.313191),
"/US/US-LA/killian-8674" => array("type"=>"village", "pop"=>1375, "name_ru"=>"Киллиан", "name_en"=>"Killian", "lang"=>"en", "lat"=>30.350191, "lon"=>-90.578254),
"/US/US-LA/kinder-8675" => array("type"=>"town", "pop"=>2403, "name_ru"=>"Киндер", "name_en"=>"Kinder", "lang"=>"en", "lat"=>30.486467, "lon"=>-92.848663),
"/US/US-LA/krotz-springs-8676" => array("type"=>"town", "pop"=>1271, "name_ru"=>"Кротц Спрингс", "name_en"=>"Krotz Springs", "lang"=>"en", "lat"=>30.534324, "lon"=>-91.753536),
"/US/US-LA/labadieville-8677" => array("type"=>"cdp", "pop"=>1821, "name_ru"=>"Лабадьевилл", "name_en"=>"Labadieville", "lang"=>"en", "lat"=>29.823635, "lon"=>-90.956422),
"/US/US-LA/lacombe-8678" => array("type"=>"cdp", "pop"=>6523, "name_ru"=>"Лакомб", "name_en"=>"Lacombe", "lang"=>"en", "lat"=>30.307692, "lon"=>-89.927980),
"/US/US-LA/lafayette-8679" => array("type"=>"city", "pop"=>113656, "name_ru"=>"Ла Файет", "name_en"=>"Lafayette", "lang"=>"en", "lat"=>30.215250, "lon"=>-92.029499),
"/US/US-LA/lafitte-8680" => array("type"=>"cdp", "pop"=>1507, "name_ru"=>"Лафитт", "name_en"=>"Lafitte", "lang"=>"en", "lat"=>29.688635, "lon"=>-90.097137),
"/US/US-LA/lake-arthur-8681" => array("type"=>"town", "pop"=>2879, "name_ru"=>"Лейк Артур", "name_en"=>"Lake Arthur", "lang"=>"en", "lat"=>30.079831, "lon"=>-92.677428),
"/US/US-LA/lake-charles-8682" => array("type"=>"city", "pop"=>71061, "name_ru"=>"Лейк Чарльз", "name_en"=>"Lake Charles", "lang"=>"en", "lat"=>30.210850, "lon"=>-93.202910),
"/US/US-LA/lake-providence-8683" => array("type"=>"town", "pop"=>4235, "name_ru"=>"Лейк Провиденс", "name_en"=>"Lake Providence", "lang"=>"en", "lat"=>32.810552, "lon"=>-91.184352),
"/US/US-LA/laplace-8684" => array("type"=>"cdp", "pop"=>24194, "name_ru"=>"Лаплас", "name_en"=>"Laplace", "lang"=>"en", "lat"=>30.072348, "lon"=>-90.472775),
"/US/US-LA/larose-8685" => array("type"=>"cdp", "pop"=>5772, "name_ru"=>"Лароуз", "name_en"=>"Larose", "lang"=>"en", "lat"=>29.570768, "lon"=>-90.374542),
"/US/US-LA/lecompte-8686" => array("type"=>"town", "pop"=>1317, "name_ru"=>"Лекомпт", "name_en"=>"Lecompte", "lang"=>"en", "lat"=>31.088867, "lon"=>-92.398583),
"/US/US-LA/leesville-8687" => array("type"=>"city", "pop"=>5704, "name_ru"=>"Лисвилл", "name_en"=>"Leesville", "lang"=>"en", "lat"=>31.137383, "lon"=>-93.274845),
"/US/US-LA/leonville-8688" => array("type"=>"town", "pop"=>1029, "name_ru"=>"Леонвилл", "name_en"=>"Leonville", "lang"=>"en", "lat"=>30.464406, "lon"=>-91.980139),
"/US/US-LA/lillie-8689" => array("type"=>"village", "pop"=>130, "name_ru"=>"Лилли", "name_en"=>"Lillie", "lang"=>"en", "lat"=>32.920179, "lon"=>-92.661410),
"/US/US-LA/lisbon-8690" => array("type"=>"village", "pop"=>149, "name_ru"=>"Лисбон", "name_en"=>"Lisbon", "lang"=>"en", "lat"=>32.788107, "lon"=>-92.865507),
"/US/US-LA/livingston-8691" => array("type"=>"town", "pop"=>1741, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>30.502175, "lon"=>-90.745216),
"/US/US-LA/livonia-8692" => array("type"=>"town", "pop"=>1350, "name_ru"=>"Ливония", "name_en"=>"Livonia", "lang"=>"en", "lat"=>30.564852, "lon"=>-91.550077),
"/US/US-LA/lockport-8693" => array("type"=>"town", "pop"=>2635, "name_ru"=>"Локпорт", "name_en"=>"Lockport", "lang"=>"en", "lat"=>29.640412, "lon"=>-90.538043),
"/US/US-LA/logansport-8694" => array("type"=>"town", "pop"=>1673, "name_ru"=>"Логанспорт", "name_en"=>"Logansport", "lang"=>"en", "lat"=>31.974047, "lon"=>-93.994065),
"/US/US-LA/lone-star-8695" => array("type"=>"cdp", "pop"=>1367, "name_ru"=>"Лоун Стар", "name_en"=>"Lone Star", "lang"=>"en", "lat"=>29.920052, "lon"=>-90.340519),
"/US/US-LA/longstreet-8696" => array("type"=>"village", "pop"=>170, "name_ru"=>"Лонгстрит", "name_en"=>"Longstreet", "lang"=>"en", "lat"=>32.096771, "lon"=>-93.950119),
"/US/US-LA/loreauville-8697" => array("type"=>"village", "pop"=>946, "name_ru"=>"Лоровилл", "name_en"=>"Loreauville", "lang"=>"en", "lat"=>30.059129, "lon"=>-91.738598),
"/US/US-LA/lucky-8698" => array("type"=>"village", "pop"=>337, "name_ru"=>"Лаки", "name_en"=>"Lucky", "lang"=>"en", "lat"=>32.243530, "lon"=>-93.012787),
"/US/US-LA/luling-8699" => array("type"=>"cdp", "pop"=>2803, "name_ru"=>"Лулинг", "name_en"=>"Luling", "lang"=>"en", "lat"=>29.923839, "lon"=>-90.366909),
"/US/US-LA/lutcher-8700" => array("type"=>"town", "pop"=>3464, "name_ru"=>"Латчер", "name_en"=>"Lutcher", "lang"=>"en", "lat"=>30.063039, "lon"=>-90.711684),
"/US/US-LA/lydia-8701" => array("type"=>"cdp", "pop"=>1136, "name_ru"=>"Лидия", "name_en"=>"Lydia", "lang"=>"en", "lat"=>29.924494, "lon"=>-91.782375),
"/US/US-LA/madisonville-8703" => array("type"=>"town", "pop"=>752, "name_ru"=>"Мэдисонвилл", "name_en"=>"Madisonville", "lang"=>"en", "lat"=>30.395163, "lon"=>-90.164207),
"/US/US-LA/mamou-8704" => array("type"=>"town", "pop"=>3414, "name_ru"=>"Маму", "name_en"=>"Mamou", "lang"=>"en", "lat"=>30.634843, "lon"=>-92.417595),
"/US/US-LA/mandeville-8705" => array("type"=>"city", "pop"=>12421, "name_ru"=>"Мандевилл", "name_en"=>"Mandeville", "lang"=>"en", "lat"=>30.370119, "lon"=>-90.082159),
"/US/US-LA/mangham-8706" => array("type"=>"town", "pop"=>544, "name_ru"=>"Мэнгхэм", "name_en"=>"Mangham", "lang"=>"en", "lat"=>32.306050, "lon"=>-91.780948),
"/US/US-LA/mansfield-8707" => array("type"=>"city", "pop"=>5402, "name_ru"=>"Мансфилд", "name_en"=>"Mansfield", "lang"=>"en", "lat"=>32.033808, "lon"=>-93.702364),
"/US/US-LA/mansura-8708" => array("type"=>"town", "pop"=>1547, "name_ru"=>"Мансура", "name_en"=>"Mansura", "lang"=>"en", "lat"=>31.058034, "lon"=>-92.050818),
"/US/US-LA/many-8709" => array("type"=>"town", "pop"=>2755, "name_ru"=>"Мэни", "name_en"=>"Many", "lang"=>"en", "lat"=>31.565357, "lon"=>-93.474864),
"/US/US-LA/maringouin-8710" => array("type"=>"town", "pop"=>1193, "name_ru"=>"Марингуин", "name_en"=>"Maringouin", "lang"=>"en", "lat"=>30.491096, "lon"=>-91.518272),
"/US/US-LA/marion-8711" => array("type"=>"village", "pop"=>750, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>32.899813, "lon"=>-92.239463),
"/US/US-LA/marksville-8712" => array("type"=>"city", "pop"=>5684, "name_ru"=>"Марксвилл", "name_en"=>"Marksville", "lang"=>"en", "lat"=>31.126750, "lon"=>-92.063066),
"/US/US-LA/marrero-8713" => array("type"=>"cdp", "pop"=>36671, "name_ru"=>"Марреро", "name_en"=>"Marrero", "lang"=>"en", "lat"=>29.885025, "lon"=>-90.114319),
"/US/US-LA/martin-8714" => array("type"=>"village", "pop"=>602, "name_ru"=>"Мартин", "name_en"=>"Martin", "lang"=>"en", "lat"=>32.099131, "lon"=>-93.218327),
"/US/US-LA/mathews-8715" => array("type"=>"cdp", "pop"=>3009, "name_ru"=>"Мэтьюс", "name_en"=>"Mathews", "lang"=>"en", "lat"=>29.687240, "lon"=>-90.561424),
"/US/US-LA/maurice-8716" => array("type"=>"village", "pop"=>853, "name_ru"=>"Морис", "name_en"=>"Maurice", "lang"=>"en", "lat"=>30.106189, "lon"=>-92.123852),
"/US/US-LA/mcnary-8702" => array("type"=>"village", "pop"=>200, "name_ru"=>"Макнари", "name_en"=>"McNary", "lang"=>"en", "lat"=>30.986743, "lon"=>-92.577551),
"/US/US-LA/melville-8717" => array("type"=>"town", "pop"=>1384, "name_ru"=>"Меовилл", "name_en"=>"Melville", "lang"=>"en", "lat"=>30.691772, "lon"=>-91.747332),
"/US/US-LA/mer-rouge-8720" => array("type"=>"village", "pop"=>659, "name_ru"=>"Мер Руж", "name_en"=>"Mer Rouge", "lang"=>"en", "lat"=>32.776913, "lon"=>-91.794255),
"/US/US-LA/meraux-8718" => array("type"=>"cdp", "pop"=>8849, "name_ru"=>"Меро", "name_en"=>"Meraux", "lang"=>"en", "lat"=>29.927735, "lon"=>-89.918872),
"/US/US-LA/mermentau-8719" => array("type"=>"village", "pop"=>691, "name_ru"=>"Мерменто", "name_en"=>"Mermentau", "lang"=>"en", "lat"=>30.184978, "lon"=>-92.584979),
"/US/US-LA/merrydale-8721" => array("type"=>"cdp", "pop"=>10395, "name_ru"=>"Меридейл", "name_en"=>"Merrydale", "lang"=>"en", "lat"=>30.501000, "lon"=>-91.108427),
"/US/US-LA/merryville-8722" => array("type"=>"town", "pop"=>1171, "name_ru"=>"Меривилл", "name_en"=>"Merryville", "lang"=>"en", "lat"=>30.753539, "lon"=>-93.532568),
"/US/US-LA/metairie-8723" => array("type"=>"cdp", "pop"=>149428, "name_ru"=>"Метейри", "name_en"=>"Metairie", "lang"=>"en", "lat"=>29.997601, "lon"=>-90.177396),
"/US/US-LA/midway-8724" => array("type"=>"cdp", "pop"=>1586, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>31.685122, "lon"=>-92.151166),
"/US/US-LA/mimosa-park-8725" => array("type"=>"cdp", "pop"=>4516, "name_ru"=>"Мимоза Парк", "name_en"=>"Mimosa Park", "lang"=>"en", "lat"=>29.898223, "lon"=>-90.347319),
"/US/US-LA/minden-8726" => array("type"=>"city", "pop"=>12962, "name_ru"=>"Минден", "name_en"=>"Minden", "lang"=>"en", "lat"=>32.616646, "lon"=>-93.280812),
"/US/US-LA/monroe-8727" => array("type"=>"city", "pop"=>51215, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>32.511650, "lon"=>-92.084921),
"/US/US-LA/montegut-8728" => array("type"=>"cdp", "pop"=>1784, "name_ru"=>"Монтегат", "name_en"=>"Montegut", "lang"=>"en", "lat"=>29.469838, "lon"=>-90.559880),
"/US/US-LA/montgomery-8729" => array("type"=>"town", "pop"=>839, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>31.664490, "lon"=>-92.889165),
"/US/US-LA/monticello-8730" => array("type"=>"cdp", "pop"=>4710, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>30.486700, "lon"=>-91.044099),
"/US/US-LA/montpelier-8731" => array("type"=>"village", "pop"=>210, "name_ru"=>"Монтпильер", "name_en"=>"Montpelier", "lang"=>"en", "lat"=>30.679511, "lon"=>-90.654490),
"/US/US-LA/mooringsport-8732" => array("type"=>"town", "pop"=>804, "name_ru"=>"Мурингспорт", "name_en"=>"Mooringsport", "lang"=>"en", "lat"=>32.682847, "lon"=>-93.961562),
"/US/US-LA/moreauville-8733" => array("type"=>"village", "pop"=>929, "name_ru"=>"Моровилл", "name_en"=>"Moreauville", "lang"=>"en", "lat"=>31.033276, "lon"=>-91.980420),
"/US/US-LA/morgan-city-8734" => array("type"=>"city", "pop"=>11604, "name_ru"=>"Морган Сити", "name_en"=>"Morgan City", "lang"=>"en", "lat"=>29.701700, "lon"=>-91.191876),
"/US/US-LA/morganza-8735" => array("type"=>"village", "pop"=>614, "name_ru"=>"Морганза", "name_en"=>"Morganza", "lang"=>"en", "lat"=>30.736621, "lon"=>-91.590691),
"/US/US-LA/morse-8736" => array("type"=>"village", "pop"=>747, "name_ru"=>"Морс", "name_en"=>"Morse", "lang"=>"en", "lat"=>30.120897, "lon"=>-92.497955),
"/US/US-LA/moss-bluff-8737" => array("type"=>"cdp", "pop"=>8039, "name_ru"=>"Мосс Блафф", "name_en"=>"Moss Bluff", "lang"=>"en", "lat"=>30.304144, "lon"=>-93.203772),
"/US/US-LA/mound-8738" => array("type"=>"village", "pop"=>13, "name_ru"=>"Маунд", "name_en"=>"Mound", "lang"=>"en", "lat"=>32.336792, "lon"=>-91.021466),
"/US/US-LA/mount-lebanon-8739" => array("type"=>"town", "pop"=>68, "name_ru"=>"Маунт Лебанон", "name_en"=>"Mount Lebanon", "lang"=>"en", "lat"=>32.502920, "lon"=>-93.049278),
"/US/US-LA/napoleonville-8740" => array("type"=>"village", "pop"=>658, "name_ru"=>"Наполеонвилл", "name_en"=>"Napoleonville", "lang"=>"en", "lat"=>29.937576, "lon"=>-91.026651),
"/US/US-LA/natalbany-8741" => array("type"=>"cdp", "pop"=>1289, "name_ru"=>"Наталбани", "name_en"=>"Natalbany", "lang"=>"en", "lat"=>30.547666, "lon"=>-90.483676),
"/US/US-LA/natchez-8742" => array("type"=>"village", "pop"=>587, "name_ru"=>"Натчез", "name_en"=>"Natchez", "lang"=>"en", "lat"=>31.673964, "lon"=>-93.045335),
"/US/US-LA/natchitoches-8743" => array("type"=>"city", "pop"=>18241, "name_ru"=>"Натчиточес", "name_en"=>"Natchitoches", "lang"=>"en", "lat"=>31.755477, "lon"=>-93.097665),
"/US/US-LA/new-iberia-8745" => array("type"=>"city", "pop"=>32959, "name_ru"=>"Нью Иберия", "name_en"=>"New Iberia", "lang"=>"en", "lat"=>30.005107, "lon"=>-91.817162),
"/US/US-LA/new-orleans-8747" => array("type"=>"city", "pop"=>311853, "name_ru"=>"Нью Орлеанс", "name_en"=>"New Orleans", "lang"=>"en", "lat"=>30.065846, "lon"=>-89.931355),
"/US/US-LA/new-roads-8748" => array("type"=>"city", "pop"=>4746, "name_ru"=>"Нью Роадс", "name_en"=>"New Roads", "lang"=>"en", "lat"=>30.694134, "lon"=>-91.454137),
"/US/US-LA/new-sarpy-8749" => array("type"=>"cdp", "pop"=>2946, "name_ru"=>"Нью Сарпи", "name_en"=>"New Sarpy", "lang"=>"en", "lat"=>29.977370, "lon"=>-90.386204),
"/US/US-LA/newellton-8744" => array("type"=>"town", "pop"=>1250, "name_ru"=>"Ньюэллтон", "name_en"=>"Newellton", "lang"=>"en", "lat"=>32.072586, "lon"=>-91.239118),
"/US/US-LA/newllano-8746" => array("type"=>"town", "pop"=>2660, "name_ru"=>"Ньюллано", "name_en"=>"Newllano", "lang"=>"en", "lat"=>31.112829, "lon"=>-93.279098),
"/US/US-LA/noble-8750" => array("type"=>"village", "pop"=>267, "name_ru"=>"Ноубл", "name_en"=>"Noble", "lang"=>"en", "lat"=>31.690082, "lon"=>-93.683017),
"/US/US-LA/norco-8751" => array("type"=>"cdp", "pop"=>3385, "name_ru"=>"Норко", "name_en"=>"Norco", "lang"=>"en", "lat"=>30.001719, "lon"=>-90.409565),
"/US/US-LA/north-hodge-8752" => array("type"=>"village", "pop"=>434, "name_ru"=>"Норт Ходж", "name_en"=>"North Hodge", "lang"=>"en", "lat"=>32.284094, "lon"=>-92.716162),
"/US/US-LA/north-vacherie-8753" => array("type"=>"cdp", "pop"=>2354, "name_ru"=>"Норт Вачери", "name_en"=>"North Vacherie", "lang"=>"en", "lat"=>29.996757, "lon"=>-90.705614),
"/US/US-LA/norwood-8754" => array("type"=>"village", "pop"=>333, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>30.964542, "lon"=>-91.106519),
"/US/US-LA/oak-grove-8756" => array("type"=>"town", "pop"=>1969, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>32.862168, "lon"=>-91.391198),
"/US/US-LA/oak-hills-place-8757" => array("type"=>"cdp", "pop"=>5479, "name_ru"=>"Оук Хиллс Плейс", "name_en"=>"Oak Hills Place", "lang"=>"en", "lat"=>30.366667, "lon"=>-91.086914),
"/US/US-LA/oak-ridge-8758" => array("type"=>"village", "pop"=>129, "name_ru"=>"Оук Ридж", "name_en"=>"Oak Ridge", "lang"=>"en", "lat"=>32.624718, "lon"=>-91.773274),
"/US/US-LA/oakdale-8755" => array("type"=>"city", "pop"=>8140, "name_ru"=>"Оукдейл", "name_en"=>"Oakdale", "lang"=>"en", "lat"=>30.818414, "lon"=>-92.656475),
"/US/US-LA/oberlin-8759" => array("type"=>"town", "pop"=>1918, "name_ru"=>"Оберлин", "name_en"=>"Oberlin", "lang"=>"en", "lat"=>30.618277, "lon"=>-92.764263),
"/US/US-LA/oil-city-8760" => array("type"=>"town", "pop"=>1192, "name_ru"=>"Ойл Сити", "name_en"=>"Oil City", "lang"=>"en", "lat"=>32.744650, "lon"=>-93.974859),
"/US/US-LA/old-jefferson-8761" => array("type"=>"cdp", "pop"=>4531, "name_ru"=>"Оулд Джефферсон", "name_en"=>"Old Jefferson", "lang"=>"en", "lat"=>30.379100, "lon"=>-91.003391),
"/US/US-LA/olla-8762" => array("type"=>"town", "pop"=>1354, "name_ru"=>"Олла", "name_en"=>"Olla", "lang"=>"en", "lat"=>31.899845, "lon"=>-92.237596),
"/US/US-LA/opelousas-8763" => array("type"=>"city", "pop"=>23220, "name_ru"=>"Опелусас", "name_en"=>"Opelousas", "lang"=>"en", "lat"=>30.524697, "lon"=>-92.083513),
"/US/US-LA/paincourtville-8764" => array("type"=>"cdp", "pop"=>1550, "name_ru"=>"Пейнкортвилл", "name_en"=>"Paincourtville", "lang"=>"en", "lat"=>29.991407, "lon"=>-91.053897),
"/US/US-LA/palmetto-8765" => array("type"=>"village", "pop"=>184, "name_ru"=>"Палметто", "name_en"=>"Palmetto", "lang"=>"en", "lat"=>30.717795, "lon"=>-91.907964),
"/US/US-LA/parks-8766" => array("type"=>"village", "pop"=>543, "name_ru"=>"Паркс", "name_en"=>"Parks", "lang"=>"en", "lat"=>30.215292, "lon"=>-91.829406),
"/US/US-LA/patterson-8767" => array("type"=>"city", "pop"=>5271, "name_ru"=>"Паттерсон", "name_en"=>"Patterson", "lang"=>"en", "lat"=>29.691724, "lon"=>-91.308695),
"/US/US-LA/pearl-river-8768" => array("type"=>"town", "pop"=>2193, "name_ru"=>"Перл Ривер", "name_en"=>"Pearl River", "lang"=>"en", "lat"=>30.372195, "lon"=>-89.747964),
"/US/US-LA/pierre-part-8769" => array("type"=>"cdp", "pop"=>3053, "name_ru"=>"Пирр Парт", "name_en"=>"Pierre Part", "lang"=>"en", "lat"=>29.957554, "lon"=>-91.208863),
"/US/US-LA/pine-prairie-8770" => array("type"=>"village", "pop"=>1538, "name_ru"=>"Пайн Прейри", "name_en"=>"Pine Prairie", "lang"=>"en", "lat"=>30.782902, "lon"=>-92.422894),
"/US/US-LA/pineville-8771" => array("type"=>"city", "pop"=>14776, "name_ru"=>"Пайнвилл", "name_en"=>"Pineville", "lang"=>"en", "lat"=>31.331967, "lon"=>-92.414989),
"/US/US-LA/pioneer-8772" => array("type"=>"village", "pop"=>153, "name_ru"=>"Пайониер", "name_en"=>"Pioneer", "lang"=>"en", "lat"=>32.738050, "lon"=>-91.435896),
"/US/US-LA/plain-dealing-8773" => array("type"=>"town", "pop"=>1034, "name_ru"=>"Плейн Дилинг", "name_en"=>"Plain Dealing", "lang"=>"en", "lat"=>32.906779, "lon"=>-93.698919),
"/US/US-LA/plaquemine-8774" => array("type"=>"city", "pop"=>6751, "name_ru"=>"Плакемин", "name_en"=>"Plaquemine", "lang"=>"en", "lat"=>30.283200, "lon"=>-91.240548),
"/US/US-LA/plaucheville-8775" => array("type"=>"village", "pop"=>276, "name_ru"=>"Плошевилл", "name_en"=>"Plaucheville", "lang"=>"en", "lat"=>30.964631, "lon"=>-91.982949),
"/US/US-LA/pleasant-hill-8776" => array("type"=>"village", "pop"=>716, "name_ru"=>"Плезант Хилл", "name_en"=>"Pleasant Hill", "lang"=>"en", "lat"=>31.816116, "lon"=>-93.515160),
"/US/US-LA/pollock-8777" => array("type"=>"town", "pop"=>383, "name_ru"=>"Поллок", "name_en"=>"Pollock", "lang"=>"en", "lat"=>31.524561, "lon"=>-92.407611),
"/US/US-LA/ponchatoula-8778" => array("type"=>"city", "pop"=>6320, "name_ru"=>"Пончатаула", "name_en"=>"Ponchatoula", "lang"=>"en", "lat"=>30.437309, "lon"=>-90.443493),
"/US/US-LA/port-allen-8779" => array("type"=>"city", "pop"=>4984, "name_ru"=>"Порт Аллен", "name_en"=>"Port Allen", "lang"=>"en", "lat"=>30.450750, "lon"=>-91.208774),
"/US/US-LA/port-barre-8780" => array("type"=>"town", "pop"=>2401, "name_ru"=>"Порт Барр", "name_en"=>"Port Barre", "lang"=>"en", "lat"=>30.556784, "lon"=>-91.956353),
"/US/US-LA/port-sulphur-8781" => array("type"=>"cdp", "pop"=>3523, "name_ru"=>"Порт Салфер", "name_en"=>"Port Sulphur", "lang"=>"en", "lat"=>29.494430, "lon"=>-89.708470),
"/US/US-LA/port-vincent-8782" => array("type"=>"village", "pop"=>527, "name_ru"=>"Порт Винсент", "name_en"=>"Port Vincent", "lang"=>"en", "lat"=>30.337057, "lon"=>-90.841830),
"/US/US-LA/powhatan-8783" => array("type"=>"village", "pop"=>138, "name_ru"=>"Похатан", "name_en"=>"Powhatan", "lang"=>"en", "lat"=>31.874126, "lon"=>-93.203325),
"/US/US-LA/poydras-8784" => array("type"=>"cdp", "pop"=>4029, "name_ru"=>"Пойдрас", "name_en"=>"Poydras", "lang"=>"en", "lat"=>29.859180, "lon"=>-89.886857),
"/US/US-LA/prien-8785" => array("type"=>"cdp", "pop"=>6448, "name_ru"=>"Приен", "name_en"=>"Prien", "lang"=>"en", "lat"=>30.166350, "lon"=>-93.267600),
"/US/US-LA/provencal-8786" => array("type"=>"village", "pop"=>719, "name_ru"=>"Провенкал", "name_en"=>"Provencal", "lang"=>"en", "lat"=>31.652669, "lon"=>-93.200332),
"/US/US-LA/quitman-8787" => array("type"=>"village", "pop"=>165, "name_ru"=>"Квитман", "name_en"=>"Quitman", "lang"=>"en", "lat"=>32.349908, "lon"=>-92.721525),
"/US/US-LA/raceland-8788" => array("type"=>"cdp", "pop"=>5564, "name_ru"=>"Рейсленд", "name_en"=>"Raceland", "lang"=>"en", "lat"=>29.718095, "lon"=>-90.596831),
"/US/US-LA/rayne-8789" => array("type"=>"city", "pop"=>8601, "name_ru"=>"Рейн", "name_en"=>"Rayne", "lang"=>"en", "lat"=>30.238243, "lon"=>-92.267066),
"/US/US-LA/rayville-8790" => array("type"=>"town", "pop"=>4016, "name_ru"=>"Рейвилл", "name_en"=>"Rayville", "lang"=>"en", "lat"=>32.471577, "lon"=>-91.757249),
"/US/US-LA/red-chute-8791" => array("type"=>"cdp", "pop"=>5431, "name_ru"=>"Ред Шут", "name_en"=>"Red Chute", "lang"=>"en", "lat"=>32.573541, "lon"=>-93.611561),
"/US/US-LA/reeves-8792" => array("type"=>"village", "pop"=>213, "name_ru"=>"Ривс", "name_en"=>"Reeves", "lang"=>"en", "lat"=>30.519868, "lon"=>-93.043884),
"/US/US-LA/reserve-8793" => array("type"=>"cdp", "pop"=>8847, "name_ru"=>"Резерв", "name_en"=>"Reserve", "lang"=>"en", "lat"=>30.075440, "lon"=>-90.560559),
"/US/US-LA/richmond-8794" => array("type"=>"village", "pop"=>401, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>32.387775, "lon"=>-91.182373),
"/US/US-LA/richwood-8795" => array("type"=>"town", "pop"=>2090, "name_ru"=>"Ричвуд", "name_en"=>"Richwood", "lang"=>"en", "lat"=>32.446576, "lon"=>-92.072531),
"/US/US-LA/ridgecrest-8796" => array("type"=>"town", "pop"=>707, "name_ru"=>"Риджкрест", "name_en"=>"Ridgecrest", "lang"=>"en", "lat"=>31.601874, "lon"=>-91.529712),
"/US/US-LA/ringgold-8797" => array("type"=>"town", "pop"=>1520, "name_ru"=>"Ринггоулд", "name_en"=>"Ringgold", "lang"=>"en", "lat"=>32.326044, "lon"=>-93.284390),
"/US/US-LA/river-ridge-8798" => array("type"=>"cdp", "pop"=>14800, "name_ru"=>"Ривер Ридж", "name_en"=>"River Ridge", "lang"=>"en", "lat"=>29.958851, "lon"=>-90.217809),
"/US/US-LA/robeline-8799" => array("type"=>"village", "pop"=>184, "name_ru"=>"Роублайн", "name_en"=>"Robeline", "lang"=>"en", "lat"=>31.691763, "lon"=>-93.303322),
"/US/US-LA/rodessa-8800" => array("type"=>"village", "pop"=>315, "name_ru"=>"Родесса", "name_en"=>"Rodessa", "lang"=>"en", "lat"=>32.971669, "lon"=>-93.991979),
"/US/US-LA/rosedale-8801" => array("type"=>"village", "pop"=>726, "name_ru"=>"Роуздейл", "name_en"=>"Rosedale", "lang"=>"en", "lat"=>30.444939, "lon"=>-91.457101),
"/US/US-LA/roseland-8802" => array("type"=>"town", "pop"=>1287, "name_ru"=>"Роузлэнд", "name_en"=>"Roseland", "lang"=>"en", "lat"=>30.764672, "lon"=>-90.512821),
"/US/US-LA/rosepine-8803" => array("type"=>"town", "pop"=>1342, "name_ru"=>"Роузпайн", "name_en"=>"Rosepine", "lang"=>"en", "lat"=>30.919382, "lon"=>-93.284335),
"/US/US-LA/ruston-8804" => array("type"=>"city", "pop"=>21123, "name_ru"=>"Растон", "name_en"=>"Ruston", "lang"=>"en", "lat"=>32.531802, "lon"=>-92.639625),
"/US/US-LA/saline-8809" => array("type"=>"village", "pop"=>279, "name_ru"=>"Сейлайн", "name_en"=>"Saline", "lang"=>"en", "lat"=>32.164012, "lon"=>-92.976968),
"/US/US-LA/sarepta-8810" => array("type"=>"village", "pop"=>899, "name_ru"=>"Сарепта", "name_en"=>"Sarepta", "lang"=>"en", "lat"=>32.895234, "lon"=>-93.451371),
"/US/US-LA/schriever-8811" => array("type"=>"cdp", "pop"=>4958, "name_ru"=>"Шривер", "name_en"=>"Schriever", "lang"=>"en", "lat"=>29.736087, "lon"=>-90.826802),
"/US/US-LA/scott-8812" => array("type"=>"town", "pop"=>9010, "name_ru"=>"Скотт", "name_en"=>"Scott", "lang"=>"en", "lat"=>30.237690, "lon"=>-92.090499),
"/US/US-LA/shenandoah-8813" => array("type"=>"cdp", "pop"=>13429, "name_ru"=>"Шенандоа", "name_en"=>"Shenandoah", "lang"=>"en", "lat"=>30.401050, "lon"=>-91.000947),
"/US/US-LA/shongaloo-8814" => array("type"=>"village", "pop"=>157, "name_ru"=>"Шонгалу", "name_en"=>"Shongaloo", "lang"=>"en", "lat"=>32.940365, "lon"=>-93.293799),
"/US/US-LA/shreveport-8815" => array("type"=>"city", "pop"=>199729, "name_ru"=>"Шривпорт", "name_en"=>"Shreveport", "lang"=>"en", "lat"=>32.471550, "lon"=>-93.797808),
"/US/US-LA/sibley-8816" => array("type"=>"town", "pop"=>1067, "name_ru"=>"Сибли", "name_en"=>"Sibley", "lang"=>"en", "lat"=>32.543656, "lon"=>-93.294117),
"/US/US-LA/sicily-island-8817" => array("type"=>"village", "pop"=>446, "name_ru"=>"Сисили Айленд", "name_en"=>"Sicily Island", "lang"=>"en", "lat"=>31.849703, "lon"=>-91.659224),
"/US/US-LA/sikes-8818" => array("type"=>"village", "pop"=>110, "name_ru"=>"Сайкс", "name_en"=>"Sikes", "lang"=>"en", "lat"=>32.079286, "lon"=>-92.485873),
"/US/US-LA/simmesport-8819" => array("type"=>"town", "pop"=>2207, "name_ru"=>"Симмеспорт", "name_en"=>"Simmesport", "lang"=>"en", "lat"=>30.981858, "lon"=>-91.814378),
"/US/US-LA/simpson-8820" => array("type"=>"village", "pop"=>487, "name_ru"=>"Симпсон", "name_en"=>"Simpson", "lang"=>"en", "lat"=>31.255830, "lon"=>-93.018360),
"/US/US-LA/simsboro-8821" => array("type"=>"village", "pop"=>675, "name_ru"=>"Симсборо", "name_en"=>"Simsboro", "lang"=>"en", "lat"=>32.533183, "lon"=>-92.783037),
"/US/US-LA/slaughter-8822" => array("type"=>"town", "pop"=>985, "name_ru"=>"Слотер", "name_en"=>"Slaughter", "lang"=>"en", "lat"=>30.718900, "lon"=>-91.141604),
"/US/US-LA/slidell-8823" => array("type"=>"city", "pop"=>27183, "name_ru"=>"Слиделл", "name_en"=>"Slidell", "lang"=>"en", "lat"=>30.281033, "lon"=>-89.777674),
"/US/US-LA/sorrento-8824" => array("type"=>"town", "pop"=>1430, "name_ru"=>"Сорренто", "name_en"=>"Sorrento", "lang"=>"en", "lat"=>30.188343, "lon"=>-90.865808),
"/US/US-LA/south-mansfield-8825" => array("type"=>"village", "pop"=>335, "name_ru"=>"Саут Мэнсфилд", "name_en"=>"South Mansfield", "lang"=>"en", "lat"=>32.018542, "lon"=>-93.722875),
"/US/US-LA/south-vacherie-8826" => array("type"=>"cdp", "pop"=>3462, "name_ru"=>"Саут Вачери", "name_en"=>"South Vacherie", "lang"=>"en", "lat"=>29.927206, "lon"=>-90.699604),
"/US/US-LA/spearsville-8827" => array("type"=>"village", "pop"=>142, "name_ru"=>"Спирсвилл", "name_en"=>"Spearsville", "lang"=>"en", "lat"=>32.935434, "lon"=>-92.603927),
"/US/US-LA/springfield-8828" => array("type"=>"town", "pop"=>405, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>30.426515, "lon"=>-90.542562),
"/US/US-LA/springhill-8829" => array("type"=>"city", "pop"=>5115, "name_ru"=>"Спрингхилл", "name_en"=>"Springhill", "lang"=>"en", "lat"=>33.002123, "lon"=>-93.461277),
"/US/US-LA/st.-francisville-8805" => array("type"=>"town", "pop"=>1500, "name_ru"=>"Сент Фрэнсисвилл", "name_en"=>"St. Francisville", "lang"=>"en", "lat"=>30.783604, "lon"=>-91.379115),
"/US/US-LA/st.-joseph-8806" => array("type"=>"town", "pop"=>1078, "name_ru"=>"Сент Джозеф", "name_en"=>"St. Joseph", "lang"=>"en", "lat"=>31.920135, "lon"=>-91.239320),
"/US/US-LA/st.-martinville-8807" => array("type"=>"city", "pop"=>7024, "name_ru"=>"Сент Мартинвилл", "name_en"=>"St. Martinville", "lang"=>"en", "lat"=>30.124958, "lon"=>-91.830527),
"/US/US-LA/st.-rose-8808" => array("type"=>"cdp", "pop"=>6259, "name_ru"=>"Сент Роуз", "name_en"=>"St. Rose", "lang"=>"en", "lat"=>29.958489, "lon"=>-90.314535),
"/US/US-LA/stanley-8830" => array("type"=>"village", "pop"=>146, "name_ru"=>"Стэнли", "name_en"=>"Stanley", "lang"=>"en", "lat"=>31.963471, "lon"=>-93.903469),
"/US/US-LA/sterlington-8831" => array("type"=>"town", "pop"=>1400, "name_ru"=>"Стерлингтон", "name_en"=>"Sterlington", "lang"=>"en", "lat"=>32.690622, "lon"=>-92.064792),
"/US/US-LA/stonewall-8832" => array("type"=>"town", "pop"=>1912, "name_ru"=>"Стоунуолл", "name_en"=>"Stonewall", "lang"=>"en", "lat"=>32.264024, "lon"=>-93.816384),
"/US/US-LA/sulphur-8833" => array("type"=>"city", "pop"=>19450, "name_ru"=>"Салфер", "name_en"=>"Sulphur", "lang"=>"en", "lat"=>30.229500, "lon"=>-93.356449),
"/US/US-LA/sun-8834" => array("type"=>"village", "pop"=>472, "name_ru"=>"Сан", "name_en"=>"Sun", "lang"=>"en", "lat"=>30.649563, "lon"=>-89.904542),
"/US/US-LA/sunset-8835" => array("type"=>"town", "pop"=>2625, "name_ru"=>"Сансет", "name_en"=>"Sunset", "lang"=>"en", "lat"=>30.408258, "lon"=>-92.066012),
"/US/US-LA/supreme-8836" => array("type"=>"cdp", "pop"=>1020, "name_ru"=>"Суприм", "name_en"=>"Supreme", "lang"=>"en", "lat"=>29.859592, "lon"=>-90.992624),
"/US/US-LA/swartz-8837" => array("type"=>"cdp", "pop"=>3698, "name_ru"=>"Суорц", "name_en"=>"Swartz", "lang"=>"en", "lat"=>32.569296, "lon"=>-91.986125),
"/US/US-LA/tallulah-8838" => array("type"=>"city", "pop"=>7658, "name_ru"=>"Таллла", "name_en"=>"Tallulah", "lang"=>"en", "lat"=>32.407365, "lon"=>-91.189234),
"/US/US-LA/tangipahoa-8839" => array("type"=>"village", "pop"=>772, "name_ru"=>"Танджипахоа", "name_en"=>"Tangipahoa", "lang"=>"en", "lat"=>30.875272, "lon"=>-90.513875),
"/US/US-LA/terrytown-8840" => array("type"=>"cdp", "pop"=>23787, "name_ru"=>"Территаун", "name_en"=>"Terrytown", "lang"=>"en", "lat"=>29.900001, "lon"=>-90.028286),
"/US/US-LA/thibodaux-8841" => array("type"=>"city", "pop"=>14052, "name_ru"=>"Тибодо", "name_en"=>"Thibodaux", "lang"=>"en", "lat"=>29.789993, "lon"=>-90.820861),
"/US/US-LA/tickfaw-8842" => array("type"=>"village", "pop"=>686, "name_ru"=>"Тикфо", "name_en"=>"Tickfaw", "lang"=>"en", "lat"=>30.576217, "lon"=>-90.485704),
"/US/US-LA/timberlane-8843" => array("type"=>"cdp", "pop"=>12614, "name_ru"=>"Тимберлейн", "name_en"=>"Timberlane", "lang"=>"en", "lat"=>29.877350, "lon"=>-90.032042),
"/US/US-LA/tullos-8844" => array("type"=>"town", "pop"=>416, "name_ru"=>"Таллос", "name_en"=>"Tullos", "lang"=>"en", "lat"=>31.820249, "lon"=>-92.329111),
"/US/US-LA/turkey-creek-8845" => array("type"=>"village", "pop"=>361, "name_ru"=>"Терки Крик", "name_en"=>"Turkey Creek", "lang"=>"en", "lat"=>30.874031, "lon"=>-92.415956),
"/US/US-LA/urania-8846" => array("type"=>"town", "pop"=>688, "name_ru"=>"Урания", "name_en"=>"Urania", "lang"=>"en", "lat"=>31.863875, "lon"=>-92.290517),
"/US/US-LA/varnado-8847" => array("type"=>"village", "pop"=>339, "name_ru"=>"Варнадо", "name_en"=>"Varnado", "lang"=>"en", "lat"=>30.894726, "lon"=>-89.831560),
"/US/US-LA/vidalia-8848" => array("type"=>"town", "pop"=>4139, "name_ru"=>"Видалия", "name_en"=>"Vidalia", "lang"=>"en", "lat"=>31.568320, "lon"=>-91.440467),
"/US/US-LA/vienna-8849" => array("type"=>"town", "pop"=>414, "name_ru"=>"Вьенна", "name_en"=>"Vienna", "lang"=>"en", "lat"=>32.608624, "lon"=>-92.651797),
"/US/US-LA/village-st.-george-8850" => array("type"=>"cdp", "pop"=>6242, "name_ru"=>"Виллидж Сент Джордж", "name_en"=>"Village St. George", "lang"=>"en", "lat"=>30.362050, "lon"=>-91.067115),
"/US/US-LA/ville-platte-8851" => array("type"=>"town", "pop"=>8190, "name_ru"=>"Вилл Платт", "name_en"=>"Ville Platte", "lang"=>"en", "lat"=>30.690044, "lon"=>-92.280254),
"/US/US-LA/vinton-8852" => array("type"=>"town", "pop"=>3255, "name_ru"=>"Винтон", "name_en"=>"Vinton", "lang"=>"en", "lat"=>30.197176, "lon"=>-93.580462),
"/US/US-LA/violet-8853" => array("type"=>"cdp", "pop"=>8574, "name_ru"=>"Вайолет", "name_en"=>"Violet", "lang"=>"en", "lat"=>29.896191, "lon"=>-89.892536),
"/US/US-LA/vivian-8854" => array("type"=>"town", "pop"=>3874, "name_ru"=>"Вивиан", "name_en"=>"Vivian", "lang"=>"en", "lat"=>32.873093, "lon"=>-93.986731),
"/US/US-LA/waggaman-8855" => array("type"=>"cdp", "pop"=>9405, "name_ru"=>"Ваггаман", "name_en"=>"Waggaman", "lang"=>"en", "lat"=>29.937601, "lon"=>-90.233653),
"/US/US-LA/walker-8856" => array("type"=>"town", "pop"=>6249, "name_ru"=>"Уолкер", "name_en"=>"Walker", "lang"=>"en", "lat"=>30.484387, "lon"=>-90.865992),
"/US/US-LA/washington-8857" => array("type"=>"town", "pop"=>1053, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>30.615412, "lon"=>-92.059665),
"/US/US-LA/waterproof-8858" => array("type"=>"town", "pop"=>706, "name_ru"=>"Уотерпруф", "name_en"=>"Waterproof", "lang"=>"en", "lat"=>31.806776, "lon"=>-91.385118),
"/US/US-LA/welsh-8859" => array("type"=>"town", "pop"=>3280, "name_ru"=>"Уэлш", "name_en"=>"Welsh", "lang"=>"en", "lat"=>30.235480, "lon"=>-92.812123),
"/US/US-LA/west-ferriday-8860" => array("type"=>"cdp", "pop"=>1632, "name_ru"=>"Вэст Ферридей", "name_en"=>"West Ferriday", "lang"=>"en", "lat"=>31.640156, "lon"=>-91.573179),
"/US/US-LA/west-monroe-8863" => array("type"=>"city", "pop"=>12899, "name_ru"=>"Вэст Монро", "name_en"=>"West Monroe", "lang"=>"en", "lat"=>32.514800, "lon"=>-92.148460),
"/US/US-LA/westlake-8861" => array("type"=>"city", "pop"=>4577, "name_ru"=>"Вэстлейк", "name_en"=>"Westlake", "lang"=>"en", "lat"=>30.251100, "lon"=>-93.258924),
"/US/US-LA/westminster-8862" => array("type"=>"cdp", "pop"=>2582, "name_ru"=>"Уэстминстер", "name_en"=>"Westminster", "lang"=>"en", "lat"=>30.407050, "lon"=>-91.089768),
"/US/US-LA/westwego-8864" => array("type"=>"city", "pop"=>10098, "name_ru"=>"Вэствего", "name_en"=>"Westwego", "lang"=>"en", "lat"=>29.904851, "lon"=>-90.142767),
"/US/US-LA/white-castle-8865" => array("type"=>"town", "pop"=>1838, "name_ru"=>"Уайт Касл", "name_en"=>"White Castle", "lang"=>"en", "lat"=>30.160677, "lon"=>-91.149278),
"/US/US-LA/wilson-8866" => array("type"=>"village", "pop"=>653, "name_ru"=>"Уилсон", "name_en"=>"Wilson", "lang"=>"en", "lat"=>30.919736, "lon"=>-91.111398),
"/US/US-LA/winnfield-8867" => array("type"=>"city", "pop"=>5095, "name_ru"=>"Уиннфилд", "name_en"=>"Winnfield", "lang"=>"en", "lat"=>31.922092, "lon"=>-92.641234),
"/US/US-LA/winnsboro-8868" => array("type"=>"town", "pop"=>4854, "name_ru"=>"Уиннсборо", "name_en"=>"Winnsboro", "lang"=>"en", "lat"=>32.160079, "lon"=>-91.720468),
"/US/US-LA/wisner-8869" => array("type"=>"town", "pop"=>1030, "name_ru"=>"Уиснер", "name_en"=>"Wisner", "lang"=>"en", "lat"=>31.980872, "lon"=>-91.655124),
"/US/US-LA/woodworth-8870" => array("type"=>"village", "pop"=>1130, "name_ru"=>"Вудворт", "name_en"=>"Woodworth", "lang"=>"en", "lat"=>31.150033, "lon"=>-92.501551),
"/US/US-LA/youngsville-8871" => array("type"=>"town", "pop"=>6980, "name_ru"=>"Янгсвилл", "name_en"=>"Youngsville", "lang"=>"en", "lat"=>30.099909, "lon"=>-91.992456),
"/US/US-LA/zachary-8872" => array("type"=>"city", "pop"=>14252, "name_ru"=>"Захари", "name_en"=>"Zachary", "lang"=>"en", "lat"=>30.661098, "lon"=>-91.152130),
"/US/US-LA/zwolle-8873" => array("type"=>"town", "pop"=>1754, "name_ru"=>"Зволл", "name_en"=>"Zwolle", "lang"=>"en", "lat"=>31.636301, "lon"=>-93.643472),
"/US/US-MA" => array("type"=>"state", "pop"=>6497967, "name_ru"=>"Массачусетс", "name_en"=>"Massachusetts", "lang"=>"en"),
"/US/US-MA/abington-9309" => array("type"=>"cdp", "pop"=>16689, "name_ru"=>"Абингтон", "name_en"=>"Abington", "lang"=>"en", "lat"=>42.117400, "lon"=>-70.959072),
"/US/US-MA/acushnet-center-9310" => array("type"=>"cdp", "pop"=>3170, "name_ru"=>"Акушнет", "name_en"=>"Acushnet Center", "lang"=>"en", "lat"=>41.686115, "lon"=>-70.906118),
"/US/US-MA/adams-9311" => array("type"=>"cdp", "pop"=>8295, "name_ru"=>"Адамс", "name_en"=>"Adams", "lang"=>"en", "lat"=>42.627526, "lon"=>-73.119281),
"/US/US-MA/amesbury-9312" => array("type"=>"cdp", "pop"=>16584, "name_ru"=>"Амсбери", "name_en"=>"Amesbury", "lang"=>"en", "lat"=>42.856371, "lon"=>-70.928545),
"/US/US-MA/amherst-9313" => array("type"=>"cdp", "pop"=>35565, "name_ru"=>"Амхерст", "name_en"=>"Amherst", "lang"=>"en", "lat"=>42.375121, "lon"=>-72.519989),
"/US/US-MA/andover-9314" => array("type"=>"cdp", "pop"=>33418, "name_ru"=>"Андоувер", "name_en"=>"Andover", "lang"=>"en", "lat"=>42.655250, "lon"=>-71.143145),
"/US/US-MA/arlington-9315" => array("type"=>"cdp", "pop"=>40993, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>42.418600, "lon"=>-71.164350),
"/US/US-MA/athol-9316" => array("type"=>"cdp", "pop"=>11570, "name_ru"=>"Атол", "name_en"=>"Athol", "lang"=>"en", "lat"=>42.591792, "lon"=>-72.232032),
"/US/US-MA/attleboro-9317" => array("type"=>"city", "pop"=>42833, "name_ru"=>"Аттлборо", "name_en"=>"Attleboro", "lang"=>"en", "lat"=>41.930669, "lon"=>-71.295741),
"/US/US-MA/ayer-9318" => array("type"=>"cdp", "pop"=>7399, "name_ru"=>"Айер", "name_en"=>"Ayer", "lang"=>"en", "lat"=>42.562206, "lon"=>-71.585282),
"/US/US-MA/baldwinville-9319" => array("type"=>"cdp", "pop"=>1795, "name_ru"=>"Болдвинвилл", "name_en"=>"Baldwinville", "lang"=>"en", "lat"=>42.606335, "lon"=>-72.075942),
"/US/US-MA/barnstable-village-9320" => array("type"=>"cdp", "pop"=>2790, "name_ru"=>"Барнстейбл Виллидж", "name_en"=>"Barnstable Village", "lang"=>"en", "lat"=>41.701418, "lon"=>-70.300061),
"/US/US-MA/barre-9321" => array("type"=>"cdp", "pop"=>5431, "name_ru"=>"Барр", "name_en"=>"Barre", "lang"=>"en", "lat"=>42.422461, "lon"=>-72.107656),
"/US/US-MA/belchertown-9322" => array("type"=>"cdp", "pop"=>14233, "name_ru"=>"Белчертаун", "name_en"=>"Belchertown", "lang"=>"en", "lat"=>42.273195, "lon"=>-72.403189),
"/US/US-MA/bellingham-9323" => array("type"=>"cdp", "pop"=>15900, "name_ru"=>"Беллингхэм", "name_en"=>"Bellingham", "lang"=>"en", "lat"=>42.094143, "lon"=>-71.474552),
"/US/US-MA/belmont-9324" => array("type"=>"cdp", "pop"=>23291, "name_ru"=>"Белмонт", "name_en"=>"Belmont", "lang"=>"en", "lat"=>42.396450, "lon"=>-71.180132),
"/US/US-MA/beverly-9325" => array("type"=>"city", "pop"=>39343, "name_ru"=>"Беверли", "name_en"=>"Beverly", "lang"=>"en", "lat"=>42.557750, "lon"=>-70.840461),
"/US/US-MA/bliss-corner-9326" => array("type"=>"cdp", "pop"=>4908, "name_ru"=>"Блисс Корнер", "name_en"=>"Bliss Corner", "lang"=>"en", "lat"=>41.605272, "lon"=>-70.943053),
"/US/US-MA/bondsville-9327" => array("type"=>"cdp", "pop"=>1992, "name_ru"=>"Блондсвилл", "name_en"=>"Bondsville", "lang"=>"en", "lat"=>42.206984, "lon"=>-72.342643),
"/US/US-MA/boston-9328" => array("type"=>"city", "pop"=>609023, "name_ru"=>"Бостон", "name_en"=>"Boston", "lang"=>"en", "lat"=>42.336029, "lon"=>-71.017892),
"/US/US-MA/bourne-9329" => array("type"=>"cdp", "pop"=>19392, "name_ru"=>"Борн", "name_en"=>"Bourne", "lang"=>"en", "lat"=>41.732691, "lon"=>-70.613968),
"/US/US-MA/boxford-9330" => array("type"=>"cdp", "pop"=>8131, "name_ru"=>"Боксфорд", "name_en"=>"Boxford", "lang"=>"en", "lat"=>42.673527, "lon"=>-70.986579),
"/US/US-MA/braintree-9331" => array("type"=>"cdp", "pop"=>35294, "name_ru"=>"Брейнтри", "name_en"=>"Braintree", "lang"=>"en", "lat"=>42.202550, "lon"=>-71.003381),
"/US/US-MA/brewster-9332" => array("type"=>"cdp", "pop"=>9936, "name_ru"=>"Брюстер", "name_en"=>"Brewster", "lang"=>"en", "lat"=>41.761559, "lon"=>-70.082688),
"/US/US-MA/bridgewater-9333" => array("type"=>"cdp", "pop"=>25774, "name_ru"=>"Бриджуотер", "name_en"=>"Bridgewater", "lang"=>"en", "lat"=>41.988633, "lon"=>-70.971326),
"/US/US-MA/brockton-9334" => array("type"=>"city", "pop"=>93007, "name_ru"=>"Броктон", "name_en"=>"Brockton", "lang"=>"en", "lat"=>42.081133, "lon"=>-71.025492),
"/US/US-MA/brookline-9335" => array("type"=>"cdp", "pop"=>54896, "name_ru"=>"Бруклин", "name_en"=>"Brookline", "lang"=>"en", "lat"=>42.323133, "lon"=>-71.142436),
"/US/US-MA/burlington-9336" => array("type"=>"cdp", "pop"=>24985, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>42.503109, "lon"=>-71.202055),
"/US/US-MA/buzzards-bay-9337" => array("type"=>"cdp", "pop"=>3250, "name_ru"=>"Баззардс Бэй", "name_en"=>"Buzzards Bay", "lang"=>"en", "lat"=>41.755013, "lon"=>-70.614947),
"/US/US-MA/cambridge-9338" => array("type"=>"city", "pop"=>105596, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>42.375800, "lon"=>-71.118700),
"/US/US-MA/centerville-9339" => array("type"=>"cdp", "pop"=>9190, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>41.653383, "lon"=>-70.344440),
"/US/US-MA/chatham-9340" => array("type"=>"cdp", "pop"=>6701, "name_ru"=>"Чатем", "name_en"=>"Chatham", "lang"=>"en", "lat"=>41.676373, "lon"=>-69.962643),
"/US/US-MA/chelmsford-9341" => array("type"=>"cdp", "pop"=>34409, "name_ru"=>"Челмсфорд", "name_en"=>"Chelmsford", "lang"=>"en", "lat"=>42.599857, "lon"=>-71.367670),
"/US/US-MA/chelsea-9342" => array("type"=>"city", "pop"=>41577, "name_ru"=>"Челси", "name_en"=>"Chelsea", "lang"=>"en", "lat"=>42.395900, "lon"=>-71.033490),
"/US/US-MA/chicopee-9343" => array("type"=>"city", "pop"=>54941, "name_ru"=>"Чикопи", "name_en"=>"Chicopee", "lang"=>"en", "lat"=>42.175900, "lon"=>-72.573123),
"/US/US-MA/clinton-9344" => array("type"=>"cdp", "pop"=>13965, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>42.418744, "lon"=>-71.685307),
"/US/US-MA/cochituate-9345" => array("type"=>"cdp", "pop"=>6046, "name_ru"=>"Кочайчуат", "name_en"=>"Cochituate", "lang"=>"en", "lat"=>42.328994, "lon"=>-71.360633),
"/US/US-MA/cordaville-9346" => array("type"=>"cdp", "pop"=>1530, "name_ru"=>"Кордавилл", "name_en"=>"Cordaville", "lang"=>"en", "lat"=>42.271844, "lon"=>-71.521700),
"/US/US-MA/cotuit-9347" => array("type"=>"cdp", "pop"=>2364, "name_ru"=>"Котут", "name_en"=>"Cotuit", "lang"=>"en", "lat"=>41.620307, "lon"=>-70.440007),
"/US/US-MA/danvers-9348" => array("type"=>"cdp", "pop"=>26762, "name_ru"=>"Дэнверс", "name_en"=>"Danvers", "lang"=>"en", "lat"=>42.571867, "lon"=>-70.951113),
"/US/US-MA/dedham-9349" => array("type"=>"cdp", "pop"=>24630, "name_ru"=>"Дедхэм", "name_en"=>"Dedham", "lang"=>"en", "lat"=>42.246601, "lon"=>-71.179223),
"/US/US-MA/dennis-9350" => array("type"=>"cdp", "pop"=>15349, "name_ru"=>"Деннис", "name_en"=>"Dennis", "lang"=>"en", "lat"=>41.731646, "lon"=>-70.201945),
"/US/US-MA/dennis-port-9351" => array("type"=>"cdp", "pop"=>2775, "name_ru"=>"Деннис Порт", "name_en"=>"Dennis Port", "lang"=>"en", "lat"=>41.669735, "lon"=>-70.139251),
"/US/US-MA/dover-9352" => array("type"=>"cdp", "pop"=>5644, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>42.247530, "lon"=>-71.275848),
"/US/US-MA/duxbury-9353" => array("type"=>"cdp", "pop"=>14496, "name_ru"=>"Даксбери", "name_en"=>"Duxbury", "lang"=>"en", "lat"=>42.042789, "lon"=>-70.673447),
"/US/US-MA/east-brookfield-9354" => array("type"=>"cdp", "pop"=>2057, "name_ru"=>"Ист Брукфилд", "name_en"=>"East Brookfield", "lang"=>"en", "lat"=>42.228671, "lon"=>-72.048160),
"/US/US-MA/east-dennis-9355" => array("type"=>"cdp", "pop"=>2584, "name_ru"=>"Ист Деннис", "name_en"=>"East Dennis", "lang"=>"en", "lat"=>41.740117, "lon"=>-70.157641),
"/US/US-MA/east-douglas-9356" => array("type"=>"cdp", "pop"=>1945, "name_ru"=>"Ист Дуглас", "name_en"=>"East Douglas", "lang"=>"en", "lat"=>42.075136, "lon"=>-71.712584),
"/US/US-MA/east-falmouth-9357" => array("type"=>"cdp", "pop"=>5577, "name_ru"=>"Ист Фолмаут", "name_en"=>"East Falmouth", "lang"=>"en", "lat"=>41.564332, "lon"=>-70.558215),
"/US/US-MA/east-harwich-9358" => array("type"=>"cdp", "pop"=>3828, "name_ru"=>"Ист Харвич", "name_en"=>"East Harwich", "lang"=>"en", "lat"=>41.707770, "lon"=>-70.030705),
"/US/US-MA/east-pepperell-9359" => array("type"=>"cdp", "pop"=>2296, "name_ru"=>"Ист Пепперелл", "name_en"=>"East Pepperell", "lang"=>"en", "lat"=>42.666867, "lon"=>-71.564551),
"/US/US-MA/east-sandwich-9360" => array("type"=>"cdp", "pop"=>3171, "name_ru"=>"Ист  Сэндвич", "name_en"=>"East Sandwich", "lang"=>"en", "lat"=>41.738346, "lon"=>-70.434565),
"/US/US-MA/essex-9361" => array("type"=>"cdp", "pop"=>736457, "name_ru"=>"Эссекс", "name_en"=>"Essex", "lang"=>"en", "lat"=>42.630780, "lon"=>-70.772386),
"/US/US-MA/everett-9362" => array("type"=>"city", "pop"=>37353, "name_ru"=>"Эверетт", "name_en"=>"Everett", "lang"=>"en", "lat"=>42.406550, "lon"=>-71.054647),
"/US/US-MA/fall-river-9363" => array("type"=>"city", "pop"=>90931, "name_ru"=>"Фолл Ривер", "name_en"=>"Fall River", "lang"=>"en", "lat"=>41.714600, "lon"=>-71.105075),
"/US/US-MA/falmouth-9364" => array("type"=>"cdp", "pop"=>33123, "name_ru"=>"Фолмут", "name_en"=>"Falmouth", "lang"=>"en", "lat"=>41.546187, "lon"=>-70.607847),
"/US/US-MA/fiskdale-9365" => array("type"=>"cdp", "pop"=>2189, "name_ru"=>"Фискдейл", "name_en"=>"Fiskdale", "lang"=>"en", "lat"=>42.124382, "lon"=>-72.110762),
"/US/US-MA/fitchburg-9366" => array("type"=>"city", "pop"=>40239, "name_ru"=>"Фитчбург", "name_en"=>"Fitchburg", "lang"=>"en", "lat"=>42.589250, "lon"=>-71.815550),
"/US/US-MA/forestdale-9367" => array("type"=>"cdp", "pop"=>2833, "name_ru"=>"Форестдейл", "name_en"=>"Forestdale", "lang"=>"en", "lat"=>41.688151, "lon"=>-70.508575),
"/US/US-MA/fort-devens-9368" => array("type"=>"cdp", "pop"=>8973, "name_ru"=>"Форт Девенс", "name_en"=>"Fort Devens", "lang"=>"en", "lat"=>42.536062, "lon"=>-71.617668),
"/US/US-MA/foxborough-9369" => array("type"=>"cdp", "pop"=>16347, "name_ru"=>"Фоксборо", "name_en"=>"Foxborough", "lang"=>"en", "lat"=>42.065150, "lon"=>-71.248359),
"/US/US-MA/framingham-9370" => array("type"=>"cdp", "pop"=>64885, "name_ru"=>"Фрэмингхэм", "name_en"=>"Framingham", "lang"=>"en", "lat"=>42.305901, "lon"=>-71.437922),
"/US/US-MA/franklin-9371" => array("type"=>"cdp", "pop"=>32148, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>42.086892, "lon"=>-71.387314),
"/US/US-MA/gardner-9372" => array("type"=>"city", "pop"=>20682, "name_ru"=>"Гарднер", "name_en"=>"Gardner", "lang"=>"en", "lat"=>42.584133, "lon"=>-71.986223),
"/US/US-MA/gloucester-9373" => array("type"=>"city", "pop"=>30243, "name_ru"=>"Глостер", "name_en"=>"Gloucester", "lang"=>"en", "lat"=>42.631798, "lon"=>-70.683429),
"/US/US-MA/granby-9374" => array("type"=>"cdp", "pop"=>6281, "name_ru"=>"Грэнби", "name_en"=>"Granby", "lang"=>"en", "lat"=>42.262742, "lon"=>-72.527444),
"/US/US-MA/great-barrington-9375" => array("type"=>"cdp", "pop"=>7379, "name_ru"=>"Грейт Бэррингтон", "name_en"=>"Great Barrington", "lang"=>"en", "lat"=>42.194439, "lon"=>-73.356367),
"/US/US-MA/green-harbor-cedar-crest-9377" => array("type"=>"cdp", "pop"=>2205, "name_ru"=>"Грин Харбор-Сидар Крест", "name_en"=>"Green Harbor-Cedar Crest", "lang"=>"en", "lat"=>42.074742, "lon"=>-70.659787),
"/US/US-MA/greenfield-9376" => array("type"=>"cdp", "pop"=>17828, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>42.589344, "lon"=>-72.598905),
"/US/US-MA/groton-9378" => array("type"=>"cdp", "pop"=>10632, "name_ru"=>"Гротон", "name_en"=>"Groton", "lang"=>"en", "lat"=>42.606410, "lon"=>-71.571504),
"/US/US-MA/hanson-9379" => array("type"=>"cdp", "pop"=>10019, "name_ru"=>"Хэнсон", "name_en"=>"Hanson", "lang"=>"en", "lat"=>42.064400, "lon"=>-70.850355),
"/US/US-MA/harwich-center-9380" => array("type"=>"cdp", "pop"=>1668, "name_ru"=>"Харвич Сентер", "name_en"=>"Harwich Center", "lang"=>"en", "lat"=>41.691805, "lon"=>-70.072815),
"/US/US-MA/harwich-port-9381" => array("type"=>"cdp", "pop"=>1742, "name_ru"=>"Харвич Порт", "name_en"=>"Harwich Port", "lang"=>"en", "lat"=>41.669764, "lon"=>-70.064758),
"/US/US-MA/hatfield-9382" => array("type"=>"cdp", "pop"=>3227, "name_ru"=>"Хэтфилд", "name_en"=>"Hatfield", "lang"=>"en", "lat"=>42.371877, "lon"=>-72.608825),
"/US/US-MA/haverhill-9383" => array("type"=>"city", "pop"=>61275, "name_ru"=>"Хэйверхилл", "name_en"=>"Haverhill", "lang"=>"en", "lat"=>42.784300, "lon"=>-71.086212),
"/US/US-MA/hingham-9384" => array("type"=>"cdp", "pop"=>22561, "name_ru"=>"Хингхэм", "name_en"=>"Hingham", "lang"=>"en", "lat"=>42.237199, "lon"=>-70.887663),
"/US/US-MA/holbrook-9385" => array("type"=>"cdp", "pop"=>10644, "name_ru"=>"Холбрук", "name_en"=>"Holbrook", "lang"=>"en", "lat"=>42.147778, "lon"=>-71.007085),
"/US/US-MA/holland-9386" => array("type"=>"cdp", "pop"=>2529, "name_ru"=>"Холланд", "name_en"=>"Holland", "lang"=>"en", "lat"=>42.052844, "lon"=>-72.152104),
"/US/US-MA/holyoke-9387" => array("type"=>"city", "pop"=>39947, "name_ru"=>"Хольок", "name_en"=>"Holyoke", "lang"=>"en", "lat"=>42.213467, "lon"=>-72.642426),
"/US/US-MA/hopedale-9388" => array("type"=>"cdp", "pop"=>6142, "name_ru"=>"Хоупдейл", "name_en"=>"Hopedale", "lang"=>"en", "lat"=>42.126773, "lon"=>-71.542179),
"/US/US-MA/hopkinton-9389" => array("type"=>"cdp", "pop"=>14338, "name_ru"=>"Хопкинтон", "name_en"=>"Hopkinton", "lang"=>"en", "lat"=>42.222848, "lon"=>-71.521444),
"/US/US-MA/housatonic-9390" => array("type"=>"cdp", "pop"=>1184, "name_ru"=>"Хаузатоник", "name_en"=>"Housatonic", "lang"=>"en", "lat"=>42.250377, "lon"=>-73.362717),
"/US/US-MA/hudson-9391" => array("type"=>"cdp", "pop"=>19597, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>42.392320, "lon"=>-71.564026),
"/US/US-MA/hull-9392" => array("type"=>"cdp", "pop"=>11041, "name_ru"=>"Халл", "name_en"=>"Hull", "lang"=>"en", "lat"=>42.302970, "lon"=>-70.883293),
"/US/US-MA/hyannis-9393" => array("type"=>"cdp", "pop"=>14120, "name_ru"=>"Хейннис", "name_en"=>"Hyannis", "lang"=>"en", "lat"=>41.652871, "lon"=>-70.295796),
"/US/US-MA/ipswich-9394" => array("type"=>"cdp", "pop"=>13219, "name_ru"=>"Ипсвич", "name_en"=>"Ipswich", "lang"=>"en", "lat"=>42.678087, "lon"=>-70.831499),
"/US/US-MA/kingston-9395" => array("type"=>"cdp", "pop"=>12328, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>41.997200, "lon"=>-70.714492),
"/US/US-MA/lawrence-9396" => array("type"=>"city", "pop"=>70014, "name_ru"=>"Лоренс", "name_en"=>"Lawrence", "lang"=>"en", "lat"=>42.699967, "lon"=>-71.162016),
"/US/US-MA/lee-9397" => array("type"=>"cdp", "pop"=>5763, "name_ru"=>"Ли", "name_en"=>"Lee", "lang"=>"en", "lat"=>42.307160, "lon"=>-73.251533),
"/US/US-MA/lenox-9398" => array("type"=>"cdp", "pop"=>5095, "name_ru"=>"Ленокс", "name_en"=>"Lenox", "lang"=>"en", "lat"=>42.358141, "lon"=>-73.286736),
"/US/US-MA/leominster-9399" => array("type"=>"city", "pop"=>41055, "name_ru"=>"Леоминстер", "name_en"=>"Leominster", "lang"=>"en", "lat"=>42.519075, "lon"=>-71.772163),
"/US/US-MA/lexington-9400" => array("type"=>"cdp", "pop"=>30272, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>42.445650, "lon"=>-71.231424),
"/US/US-MA/littleton-common-9401" => array("type"=>"cdp", "pop"=>2867, "name_ru"=>"Литтлтон Коммон", "name_en"=>"Littleton Common", "lang"=>"en", "lat"=>42.536515, "lon"=>-71.472476),
"/US/US-MA/longmeadow-9402" => array("type"=>"cdp", "pop"=>15329, "name_ru"=>"Лонгмэдоу", "name_en"=>"Longmeadow", "lang"=>"en", "lat"=>42.047752, "lon"=>-72.570116),
"/US/US-MA/lowell-9403" => array("type"=>"city", "pop"=>103615, "name_ru"=>"Лоуэлл", "name_en"=>"Lowell", "lang"=>"en", "lat"=>42.638711, "lon"=>-71.322141),
"/US/US-MA/lunenburg-9404" => array("type"=>"cdp", "pop"=>9946, "name_ru"=>"Луненбург", "name_en"=>"Lunenburg", "lang"=>"en", "lat"=>42.594967, "lon"=>-71.725138),
"/US/US-MA/lynn-9405" => array("type"=>"city", "pop"=>86957, "name_ru"=>"Линн", "name_en"=>"Lynn", "lang"=>"en", "lat"=>42.474167, "lon"=>-70.962548),
"/US/US-MA/lynnfield-9406" => array("type"=>"cdp", "pop"=>11412, "name_ru"=>"Линфилд", "name_en"=>"Lynnfield", "lang"=>"en", "lat"=>42.535650, "lon"=>-71.040322),
"/US/US-MA/malden-9407" => array("type"=>"city", "pop"=>55597, "name_ru"=>"Малден", "name_en"=>"Malden", "lang"=>"en", "lat"=>42.430367, "lon"=>-71.058435),
"/US/US-MA/mansfield-center-9408" => array("type"=>"cdp", "pop"=>7170, "name_ru"=>"Мэнсфилд Сентер", "name_en"=>"Mansfield Center", "lang"=>"en", "lat"=>42.023700, "lon"=>-71.218663),
"/US/US-MA/marblehead-9409" => array("type"=>"cdp", "pop"=>19951, "name_ru"=>"Марблхед", "name_en"=>"Marblehead", "lang"=>"en", "lat"=>42.494016, "lon"=>-70.836144),
"/US/US-MA/marion-center-9410" => array("type"=>"cdp", "pop"=>1426, "name_ru"=>"Мэрион Сентер", "name_en"=>"Marion Center", "lang"=>"en", "lat"=>41.703144, "lon"=>-70.762382),
"/US/US-MA/marlborough-9411" => array("type"=>"city", "pop"=>37932, "name_ru"=>"Марлборо", "name_en"=>"Marlborough", "lang"=>"en", "lat"=>42.349904, "lon"=>-71.547693),
"/US/US-MA/marshfield-9412" => array("type"=>"cdp", "pop"=>24735, "name_ru"=>"Маршфилд", "name_en"=>"Marshfield", "lang"=>"en", "lat"=>42.091408, "lon"=>-70.706017),
"/US/US-MA/marshfield-hills-9413" => array("type"=>"cdp", "pop"=>2201, "name_ru"=>"Маршфилд Хилс", "name_en"=>"Marshfield Hills", "lang"=>"en", "lat"=>42.150329, "lon"=>-70.729987),
"/US/US-MA/marstons-mills-9414" => array("type"=>"cdp", "pop"=>8017, "name_ru"=>"Марстонс Милс", "name_en"=>"Marstons Mills", "lang"=>"en", "lat"=>41.673537, "lon"=>-70.408331),
"/US/US-MA/mattapoisett-center-9415" => array("type"=>"cdp", "pop"=>2949, "name_ru"=>"Мэттапойсетт Сентер", "name_en"=>"Mattapoisett Center", "lang"=>"en", "lat"=>41.662568, "lon"=>-70.808586),
"/US/US-MA/maynard-9416" => array("type"=>"cdp", "pop"=>10182, "name_ru"=>"Мейнард", "name_en"=>"Maynard", "lang"=>"en", "lat"=>42.425651, "lon"=>-71.456950),
"/US/US-MA/medfield-9417" => array("type"=>"cdp", "pop"=>12275, "name_ru"=>"Медфилд", "name_en"=>"Medfield", "lang"=>"en", "lat"=>42.187854, "lon"=>-71.304664),
"/US/US-MA/medford-9418" => array("type"=>"city", "pop"=>55573, "name_ru"=>"Медфорд", "name_en"=>"Medford", "lang"=>"en", "lat"=>42.422450, "lon"=>-71.108742),
"/US/US-MA/melrose-9419" => array("type"=>"city", "pop"=>26708, "name_ru"=>"Мелроуз", "name_en"=>"Melrose", "lang"=>"en", "lat"=>42.454900, "lon"=>-71.059107),
"/US/US-MA/middleborough-center-9420" => array("type"=>"cdp", "pop"=>6837, "name_ru"=>"Миддлборо Сентер", "name_en"=>"Middleborough Center", "lang"=>"en", "lat"=>41.896232, "lon"=>-70.927886),
"/US/US-MA/milford-9421" => array("type"=>"cdp", "pop"=>27246, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>42.148634, "lon"=>-71.515529),
"/US/US-MA/millers-falls-9422" => array("type"=>"cdp", "pop"=>1084, "name_ru"=>"Миллерс Фоллс", "name_en"=>"Millers Falls", "lang"=>"en", "lat"=>42.579927, "lon"=>-72.492897),
"/US/US-MA/millis-clicquot-9423" => array("type"=>"cdp", "pop"=>4081, "name_ru"=>"Миллис-Клайквот", "name_en"=>"Millis-Clicquot", "lang"=>"en", "lat"=>42.168948, "lon"=>-71.355657),
"/US/US-MA/milton-9424" => array("type"=>"cdp", "pop"=>26187, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>42.243309, "lon"=>-71.083171),
"/US/US-MA/monson-center-9425" => array("type"=>"cdp", "pop"=>2101, "name_ru"=>"Монсон Сентер", "name_en"=>"Monson Center", "lang"=>"en", "lat"=>42.097071, "lon"=>-72.310407),
"/US/US-MA/monument-beach-9426" => array("type"=>"cdp", "pop"=>1842, "name_ru"=>"Моньюмент Бич", "name_en"=>"Monument Beach", "lang"=>"en", "lat"=>41.719358, "lon"=>-70.609789),
"/US/US-MA/nahant-9427" => array("type"=>"cdp", "pop"=>3498, "name_ru"=>"Нахант", "name_en"=>"Nahant", "lang"=>"en", "lat"=>42.419150, "lon"=>-70.913494),
"/US/US-MA/nantucket-9428" => array("type"=>"cdp", "pop"=>11215, "name_ru"=>"Нантукет", "name_en"=>"Nantucket", "lang"=>"en", "lat"=>41.285598, "lon"=>-70.102717),
"/US/US-MA/needham-9429" => array("type"=>"cdp", "pop"=>28560, "name_ru"=>"Нидхэм", "name_en"=>"Needham", "lang"=>"en", "lat"=>42.280700, "lon"=>-71.240989),
"/US/US-MA/new-bedford-9430" => array("type"=>"city", "pop"=>91365, "name_ru"=>"Нью Бедфорд", "name_en"=>"New Bedford", "lang"=>"en", "lat"=>41.661300, "lon"=>-70.938117),
"/US/US-MA/newburyport-9431" => array("type"=>"city", "pop"=>17542, "name_ru"=>"Ньюберипорт", "name_en"=>"Newburyport", "lang"=>"en", "lat"=>42.814828, "lon"=>-70.889458),
"/US/US-MA/newton-9432" => array("type"=>"city", "pop"=>82139, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>42.331567, "lon"=>-71.207082),
"/US/US-MA/north-adams-9433" => array("type"=>"city", "pop"=>13711, "name_ru"=>"Норт Адамс", "name_en"=>"North Adams", "lang"=>"en", "lat"=>42.686771, "lon"=>-73.113822),
"/US/US-MA/north-amherst-9434" => array("type"=>"cdp", "pop"=>6239, "name_ru"=>"Норт Амхерст", "name_en"=>"North Amherst", "lang"=>"en", "lat"=>42.409195, "lon"=>-72.521768),
"/US/US-MA/north-attleborough-center-9436" => array("type"=>"cdp", "pop"=>16178, "name_ru"=>"Норт Аттлборо Сентер", "name_en"=>"North Attleborough Center", "lang"=>"en", "lat"=>41.973150, "lon"=>-71.325176),
"/US/US-MA/north-brookfield-9438" => array("type"=>"cdp", "pop"=>4833, "name_ru"=>"Норт Брукфилд", "name_en"=>"North Brookfield", "lang"=>"en", "lat"=>42.271653, "lon"=>-72.085020),
"/US/US-MA/north-eastham-9439" => array("type"=>"cdp", "pop"=>1570, "name_ru"=>"Норт Истхэм", "name_en"=>"North Eastham", "lang"=>"en", "lat"=>41.841949, "lon"=>-70.023398),
"/US/US-MA/north-falmouth-9440" => array("type"=>"cdp", "pop"=>2625, "name_ru"=>"Норт Фолмаут", "name_en"=>"North Falmouth", "lang"=>"en", "lat"=>41.641112, "lon"=>-70.632203),
"/US/US-MA/north-lakeville-9442" => array("type"=>"cdp", "pop"=>2048, "name_ru"=>"Норт Лейквилл", "name_en"=>"North Lakeville", "lang"=>"en", "lat"=>41.857624, "lon"=>-70.942651),
"/US/US-MA/north-pembroke-9443" => array("type"=>"cdp", "pop"=>2485, "name_ru"=>"Норт Пемброук", "name_en"=>"North Pembroke", "lang"=>"en", "lat"=>42.096233, "lon"=>-70.783047),
"/US/US-MA/north-plymouth-9444" => array("type"=>"cdp", "pop"=>3450, "name_ru"=>"Норт Плимут", "name_en"=>"North Plymouth", "lang"=>"en", "lat"=>41.978040, "lon"=>-70.682747),
"/US/US-MA/north-scituate-9445" => array("type"=>"cdp", "pop"=>4891, "name_ru"=>"Норт Сайчуат", "name_en"=>"North Scituate", "lang"=>"en", "lat"=>42.212207, "lon"=>-70.766992),
"/US/US-MA/north-seekonk-9446" => array("type"=>"cdp", "pop"=>2635, "name_ru"=>"Норт Сиконк", "name_en"=>"North Seekonk", "lang"=>"en", "lat"=>41.889222, "lon"=>-71.330594),
"/US/US-MA/north-westport-9448" => array("type"=>"cdp", "pop"=>4697, "name_ru"=>"Норт Вэстпорт", "name_en"=>"North Westport", "lang"=>"en", "lat"=>41.660393, "lon"=>-71.088920),
"/US/US-MA/northampton-9435" => array("type"=>"city", "pop"=>28379, "name_ru"=>"Нортгемптон", "name_en"=>"Northampton", "lang"=>"en", "lat"=>42.327100, "lon"=>-72.675140),
"/US/US-MA/northborough-9437" => array("type"=>"cdp", "pop"=>14646, "name_ru"=>"Нортборо", "name_en"=>"Northborough", "lang"=>"en", "lat"=>42.311832, "lon"=>-71.648902),
"/US/US-MA/northfield-9441" => array("type"=>"cdp", "pop"=>3026, "name_ru"=>"Нортфилд", "name_en"=>"Northfield", "lang"=>"en", "lat"=>42.709397, "lon"=>-72.437253),
"/US/US-MA/northwest-harwich-9447" => array("type"=>"cdp", "pop"=>3037, "name_ru"=>"Нортвэст Харвич", "name_en"=>"Northwest Harwich", "lang"=>"en", "lat"=>41.690061, "lon"=>-70.105675),
"/US/US-MA/norton-center-9449" => array("type"=>"cdp", "pop"=>1899, "name_ru"=>"Нортон Сентер", "name_en"=>"Norton Center", "lang"=>"en", "lat"=>41.969752, "lon"=>-71.184164),
"/US/US-MA/norwood-9450" => array("type"=>"cdp", "pop"=>28211, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>42.185267, "lon"=>-71.195276),
"/US/US-MA/ocean-bluff-brant-rock-9451" => array("type"=>"cdp", "pop"=>4541, "name_ru"=>"Оушен Блафф-Брэнт Рок", "name_en"=>"Ocean Bluff-Brant Rock", "lang"=>"en", "lat"=>42.102671, "lon"=>-70.658848),
"/US/US-MA/ocean-grove-9452" => array("type"=>"cdp", "pop"=>3169, "name_ru"=>"Оушен Гров", "name_en"=>"Ocean Grove", "lang"=>"en", "lat"=>41.728650, "lon"=>-71.209874),
"/US/US-MA/onset-9453" => array("type"=>"cdp", "pop"=>1461, "name_ru"=>"Онсет", "name_en"=>"Onset", "lang"=>"en", "lat"=>41.747584, "lon"=>-70.667579),
"/US/US-MA/orange-9454" => array("type"=>"cdp", "pop"=>7688, "name_ru"=>"Оранж", "name_en"=>"Orange", "lang"=>"en", "lat"=>42.594236, "lon"=>-72.304258),
"/US/US-MA/orleans-9455" => array("type"=>"cdp", "pop"=>6269, "name_ru"=>"Орлеанс", "name_en"=>"Orleans", "lang"=>"en", "lat"=>41.790982, "lon"=>-70.000277),
"/US/US-MA/osterville-9456" => array("type"=>"cdp", "pop"=>2911, "name_ru"=>"Остервилл", "name_en"=>"Osterville", "lang"=>"en", "lat"=>41.625023, "lon"=>-70.387801),
"/US/US-MA/oxford-9457" => array("type"=>"cdp", "pop"=>13615, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>42.115680, "lon"=>-71.870864),
"/US/US-MA/palmer-9458" => array("type"=>"cdp", "pop"=>12933, "name_ru"=>"Палмер", "name_en"=>"Palmer", "lang"=>"en", "lat"=>42.160907, "lon"=>-72.322318),
"/US/US-MA/peabody-9459" => array("type"=>"city", "pop"=>51331, "name_ru"=>"Пибоди", "name_en"=>"Peabody", "lang"=>"en", "lat"=>42.532936, "lon"=>-70.970279),
"/US/US-MA/pepperell-9460" => array("type"=>"cdp", "pop"=>11382, "name_ru"=>"Пепперелл", "name_en"=>"Pepperell", "lang"=>"en", "lat"=>42.667797, "lon"=>-71.588189),
"/US/US-MA/pinehurst-9461" => array("type"=>"cdp", "pop"=>6614, "name_ru"=>"Пайнхерст", "name_en"=>"Pinehurst", "lang"=>"en", "lat"=>42.534307, "lon"=>-71.236167),
"/US/US-MA/pittsfield-9462" => array("type"=>"city", "pop"=>42652, "name_ru"=>"Питтсфилд", "name_en"=>"Pittsfield", "lang"=>"en", "lat"=>42.451800, "lon"=>-73.260899),
"/US/US-MA/plymouth-9463" => array("type"=>"cdp", "pop"=>492066, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>41.958122, "lon"=>-70.664297),
"/US/US-MA/pocasset-9464" => array("type"=>"cdp", "pop"=>2756, "name_ru"=>"Покассет", "name_en"=>"Pocasset", "lang"=>"en", "lat"=>41.691240, "lon"=>-70.637309),
"/US/US-MA/provincetown-9465" => array("type"=>"cdp", "pop"=>3376, "name_ru"=>"Провинстаун", "name_en"=>"Provincetown", "lang"=>"en", "lat"=>42.051089, "lon"=>-70.179391),
"/US/US-MA/quincy-9466" => array("type"=>"city", "pop"=>92339, "name_ru"=>"Квинси", "name_en"=>"Quincy", "lang"=>"en", "lat"=>42.261315, "lon"=>-71.007022),
"/US/US-MA/randolph-9467" => array("type"=>"cdp", "pop"=>30082, "name_ru"=>"Рэндолф", "name_en"=>"Randolph", "lang"=>"en", "lat"=>42.175152, "lon"=>-71.054343),
"/US/US-MA/raynham-center-9468" => array("type"=>"cdp", "pop"=>3709, "name_ru"=>"Рейнхэм Сентер", "name_en"=>"Raynham Center", "lang"=>"en", "lat"=>41.931754, "lon"=>-71.043663),
"/US/US-MA/reading-9469" => array("type"=>"cdp", "pop"=>23052, "name_ru"=>"Ридинг", "name_en"=>"Reading", "lang"=>"en", "lat"=>42.535774, "lon"=>-71.107000),
"/US/US-MA/revere-9470" => array("type"=>"city", "pop"=>60204, "name_ru"=>"Ревир", "name_en"=>"Revere", "lang"=>"en", "lat"=>42.418250, "lon"=>-70.992491),
"/US/US-MA/rockport-9471" => array("type"=>"cdp", "pop"=>7612, "name_ru"=>"Рокпорт", "name_en"=>"Rockport", "lang"=>"en", "lat"=>42.646021, "lon"=>-70.619433),
"/US/US-MA/rowley-9472" => array("type"=>"cdp", "pop"=>5738, "name_ru"=>"Роули", "name_en"=>"Rowley", "lang"=>"en", "lat"=>42.718691, "lon"=>-70.874647),
"/US/US-MA/rutland-9473" => array("type"=>"cdp", "pop"=>7899, "name_ru"=>"Рутленд", "name_en"=>"Rutland", "lang"=>"en", "lat"=>42.362813, "lon"=>-71.951948),
"/US/US-MA/sagamore-9474" => array("type"=>"cdp", "pop"=>2589, "name_ru"=>"Сагамор", "name_en"=>"Sagamore", "lang"=>"en", "lat"=>41.783233, "lon"=>-70.534189),
"/US/US-MA/salem-9475" => array("type"=>"city", "pop"=>41256, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>42.526468, "lon"=>-70.869521),
"/US/US-MA/salisbury-9476" => array("type"=>"cdp", "pop"=>8417, "name_ru"=>"Солсбери", "name_en"=>"Salisbury", "lang"=>"en", "lat"=>42.832625, "lon"=>-70.840270),
"/US/US-MA/sandwich-9477" => array("type"=>"cdp", "pop"=>20129, "name_ru"=>"Сэндвич", "name_en"=>"Sandwich", "lang"=>"en", "lat"=>41.757831, "lon"=>-70.499815),
"/US/US-MA/saugus-9478" => array("type"=>"cdp", "pop"=>27478, "name_ru"=>"Согус", "name_en"=>"Saugus", "lang"=>"en", "lat"=>42.467433, "lon"=>-71.015299),
"/US/US-MA/scituate-9479" => array("type"=>"cdp", "pop"=>17926, "name_ru"=>"Сайчуат", "name_en"=>"Scituate", "lang"=>"en", "lat"=>42.189130, "lon"=>-70.734347),
"/US/US-MA/sharon-9480" => array("type"=>"cdp", "pop"=>17373, "name_ru"=>"Шэрон", "name_en"=>"Sharon", "lang"=>"en", "lat"=>42.117518, "lon"=>-71.182167),
"/US/US-MA/shelburne-falls-9481" => array("type"=>"cdp", "pop"=>1996, "name_ru"=>"Шелбурн Фолс", "name_en"=>"Shelburne Falls", "lang"=>"en", "lat"=>42.603537, "lon"=>-72.742788),
"/US/US-MA/shirley-9482" => array("type"=>"cdp", "pop"=>7904, "name_ru"=>"Ширли", "name_en"=>"Shirley", "lang"=>"en", "lat"=>42.542000, "lon"=>-71.652926),
"/US/US-MA/smith-mills-9483" => array("type"=>"cdp", "pop"=>4593, "name_ru"=>"Смит Милс", "name_en"=>"Smith Mills", "lang"=>"en", "lat"=>41.639627, "lon"=>-70.994481),
"/US/US-MA/somerset-9484" => array("type"=>"cdp", "pop"=>18055, "name_ru"=>"Сомерсет", "name_en"=>"Somerset", "lang"=>"en", "lat"=>41.739400, "lon"=>-71.159042),
"/US/US-MA/somerville-9485" => array("type"=>"city", "pop"=>75662, "name_ru"=>"Сомервилл", "name_en"=>"Somerville", "lang"=>"en", "lat"=>42.390319, "lon"=>-71.101828),
"/US/US-MA/south-amherst-9486" => array("type"=>"cdp", "pop"=>5053, "name_ru"=>"Саут Амхерст", "name_en"=>"South Amherst", "lang"=>"en", "lat"=>42.338800, "lon"=>-72.516776),
"/US/US-MA/south-ashburnham-9487" => array("type"=>"cdp", "pop"=>1110, "name_ru"=>"Саут Эшбурнхэм", "name_en"=>"South Ashburnham", "lang"=>"en", "lat"=>42.610200, "lon"=>-71.939655),
"/US/US-MA/south-deerfield-9489" => array("type"=>"cdp", "pop"=>1906, "name_ru"=>"Саут Дирфилд", "name_en"=>"South Deerfield", "lang"=>"en", "lat"=>42.479756, "lon"=>-72.590094),
"/US/US-MA/south-dennis-9490" => array("type"=>"cdp", "pop"=>3559, "name_ru"=>"Саут Деннис", "name_en"=>"South Dennis", "lang"=>"en", "lat"=>41.704316, "lon"=>-70.155928),
"/US/US-MA/south-duxbury-9491" => array("type"=>"cdp", "pop"=>3017, "name_ru"=>"Саут Даксбери", "name_en"=>"South Duxbury", "lang"=>"en", "lat"=>42.016935, "lon"=>-70.690552),
"/US/US-MA/south-lancaster-9492" => array("type"=>"cdp", "pop"=>1772, "name_ru"=>"Саут Ланкастер", "name_en"=>"South Lancaster", "lang"=>"en", "lat"=>42.435872, "lon"=>-71.693546),
"/US/US-MA/south-yarmouth-9493" => array("type"=>"cdp", "pop"=>10358, "name_ru"=>"Саут Ярмут", "name_en"=>"South Yarmouth", "lang"=>"en", "lat"=>41.670139, "lon"=>-70.200738),
"/US/US-MA/southbridge-9488" => array("type"=>"cdp", "pop"=>16852, "name_ru"=>"Саутбридж", "name_en"=>"Southbridge", "lang"=>"en", "lat"=>42.083854, "lon"=>-72.034797),
"/US/US-MA/spencer-9494" => array("type"=>"cdp", "pop"=>11922, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>42.243872, "lon"=>-71.991970),
"/US/US-MA/springfield-9495" => array("type"=>"city", "pop"=>150640, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>42.115033, "lon"=>-72.539045),
"/US/US-MA/stoneham-9496" => array("type"=>"cdp", "pop"=>21471, "name_ru"=>"Стоунхэм", "name_en"=>"Stoneham", "lang"=>"en", "lat"=>42.474101, "lon"=>-71.098199),
"/US/US-MA/sturbridge-9497" => array("type"=>"cdp", "pop"=>9103, "name_ru"=>"Стербридж", "name_en"=>"Sturbridge", "lang"=>"en", "lat"=>42.099928, "lon"=>-72.064343),
"/US/US-MA/swampscott-9498" => array("type"=>"cdp", "pop"=>13944, "name_ru"=>"Свомпскотт", "name_en"=>"Swampscott", "lang"=>"en", "lat"=>42.466650, "lon"=>-70.892337),
"/US/US-MA/taunton-9499" => array("type"=>"city", "pop"=>55702, "name_ru"=>"Тонтон", "name_en"=>"Taunton", "lang"=>"en", "lat"=>41.903049, "lon"=>-71.093252),
"/US/US-MA/teaticket-9500" => array("type"=>"cdp", "pop"=>1856, "name_ru"=>"Титикет", "name_en"=>"Teaticket", "lang"=>"en", "lat"=>41.558362, "lon"=>-70.587589),
"/US/US-MA/three-rivers-9501" => array("type"=>"cdp", "pop"=>3006, "name_ru"=>"Три Риверс", "name_en"=>"Three Rivers", "lang"=>"en", "lat"=>42.180363, "lon"=>-72.371660),
"/US/US-MA/topsfield-9502" => array("type"=>"cdp", "pop"=>6051, "name_ru"=>"Топсфилд", "name_en"=>"Topsfield", "lang"=>"en", "lat"=>42.639160, "lon"=>-70.955871),
"/US/US-MA/townsend-9503" => array("type"=>"cdp", "pop"=>9400, "name_ru"=>"Таунсенд", "name_en"=>"Townsend", "lang"=>"en", "lat"=>42.669850, "lon"=>-71.703272),
"/US/US-MA/turners-falls-9504" => array("type"=>"cdp", "pop"=>4731, "name_ru"=>"Тёрнерс Фолс", "name_en"=>"Turners Falls", "lang"=>"en", "lat"=>42.596058, "lon"=>-72.557987),
"/US/US-MA/upton-west-upton-9505" => array("type"=>"cdp", "pop"=>2347, "name_ru"=>"Аптон-Вэст Аптон", "name_en"=>"Upton-West Upton", "lang"=>"en", "lat"=>42.173725, "lon"=>-71.610200),
"/US/US-MA/vineyard-haven-9506" => array("type"=>"cdp", "pop"=>1762, "name_ru"=>"Вайнярд Хейвен", "name_en"=>"Vineyard Haven", "lang"=>"en", "lat"=>41.455991, "lon"=>-70.605598),
"/US/US-MA/wakefield-9507" => array("type"=>"cdp", "pop"=>24717, "name_ru"=>"Уэйкфилд", "name_en"=>"Wakefield", "lang"=>"en", "lat"=>42.504568, "lon"=>-71.066591),
"/US/US-MA/walpole-9508" => array("type"=>"cdp", "pop"=>23133, "name_ru"=>"Уолпол", "name_en"=>"Walpole", "lang"=>"en", "lat"=>42.137763, "lon"=>-71.244749),
"/US/US-MA/waltham-9509" => array("type"=>"city", "pop"=>60236, "name_ru"=>"Уолтэм", "name_en"=>"Waltham", "lang"=>"en", "lat"=>42.387845, "lon"=>-71.242214),
"/US/US-MA/ware-9510" => array("type"=>"cdp", "pop"=>9824, "name_ru"=>"Вэр", "name_en"=>"Ware", "lang"=>"en", "lat"=>42.253190, "lon"=>-72.245471),
"/US/US-MA/wareham-center-9511" => array("type"=>"cdp", "pop"=>2607, "name_ru"=>"Вэрхэм Сентер", "name_en"=>"Wareham Center", "lang"=>"en", "lat"=>41.752267, "lon"=>-70.721449),
"/US/US-MA/warren-9512" => array("type"=>"cdp", "pop"=>5068, "name_ru"=>"Уоррен", "name_en"=>"Warren", "lang"=>"en", "lat"=>42.215217, "lon"=>-72.190669),
"/US/US-MA/watertown-9513" => array("type"=>"cdp", "pop"=>32365, "name_ru"=>"Уотертаун", "name_en"=>"Watertown", "lang"=>"en", "lat"=>42.369500, "lon"=>-71.177577),
"/US/US-MA/webster-9514" => array("type"=>"cdp", "pop"=>16655, "name_ru"=>"Вебстер", "name_en"=>"Webster", "lang"=>"en", "lat"=>42.046056, "lon"=>-71.874695),
"/US/US-MA/wellesley-9515" => array("type"=>"cdp", "pop"=>27244, "name_ru"=>"Уэлсли", "name_en"=>"Wellesley", "lang"=>"en", "lat"=>42.304450, "lon"=>-71.286803),
"/US/US-MA/west-barnstable-9516" => array("type"=>"cdp", "pop"=>1508, "name_ru"=>"Вэст Барнстейбл", "name_en"=>"West Barnstable", "lang"=>"en", "lat"=>41.706526, "lon"=>-70.373287),
"/US/US-MA/west-brookfield-9518" => array("type"=>"cdp", "pop"=>3806, "name_ru"=>"Вэст Брукфилд", "name_en"=>"West Brookfield", "lang"=>"en", "lat"=>42.238484, "lon"=>-72.148094),
"/US/US-MA/west-chatham-9519" => array("type"=>"cdp", "pop"=>1504, "name_ru"=>"Вэст Чатем", "name_en"=>"West Chatham", "lang"=>"en", "lat"=>41.680743, "lon"=>-69.993164),
"/US/US-MA/west-concord-9520" => array("type"=>"cdp", "pop"=>5761, "name_ru"=>"Вэст Конкорд", "name_en"=>"West Concord", "lang"=>"en", "lat"=>42.451408, "lon"=>-71.403337),
"/US/US-MA/west-dennis-9521" => array("type"=>"cdp", "pop"=>2307, "name_ru"=>"Вэст Деннис", "name_en"=>"West Dennis", "lang"=>"en", "lat"=>41.666858, "lon"=>-70.167440),
"/US/US-MA/west-falmouth-9522" => array("type"=>"cdp", "pop"=>1752, "name_ru"=>"Вэст Фолмаут", "name_en"=>"West Falmouth", "lang"=>"en", "lat"=>41.601256, "lon"=>-70.640029),
"/US/US-MA/west-springfield-9524" => array("type"=>"cdp", "pop"=>27459, "name_ru"=>"Вэст Спрингфилд", "name_en"=>"West Springfield", "lang"=>"en", "lat"=>42.123217, "lon"=>-72.653054),
"/US/US-MA/west-wareham-9525" => array("type"=>"cdp", "pop"=>2059, "name_ru"=>"Вэст Вэрхэм", "name_en"=>"West Wareham", "lang"=>"en", "lat"=>41.789851, "lon"=>-70.756560),
"/US/US-MA/west-yarmouth-9526" => array("type"=>"cdp", "pop"=>5409, "name_ru"=>"Вэст Ярмут", "name_en"=>"West Yarmouth", "lang"=>"en", "lat"=>41.647080, "lon"=>-70.248132),
"/US/US-MA/westborough-9517" => array("type"=>"cdp", "pop"=>18467, "name_ru"=>"Вэстборо", "name_en"=>"Westborough", "lang"=>"en", "lat"=>42.265904, "lon"=>-71.617070),
"/US/US-MA/westfield-9523" => array("type"=>"city", "pop"=>40608, "name_ru"=>"Уэстфилд", "name_en"=>"Westfield", "lang"=>"en", "lat"=>42.138900, "lon"=>-72.756049),
"/US/US-MA/weweantic-9527" => array("type"=>"cdp", "pop"=>1812, "name_ru"=>"Уэуэантик", "name_en"=>"Weweantic", "lang"=>"en", "lat"=>41.735152, "lon"=>-70.732467),
"/US/US-MA/weymouth-9528" => array("type"=>"cdp", "pop"=>53261, "name_ru"=>"Уэймут", "name_en"=>"Weymouth", "lang"=>"en", "lat"=>42.206250, "lon"=>-70.945100),
"/US/US-MA/white-island-shores-9529" => array("type"=>"cdp", "pop"=>1827, "name_ru"=>"Уайт Айленд Шорс", "name_en"=>"White Island Shores", "lang"=>"en", "lat"=>41.793824, "lon"=>-70.639054),
"/US/US-MA/whitinsville-9530" => array("type"=>"cdp", "pop"=>5639, "name_ru"=>"Уайтинсвилл", "name_en"=>"Whitinsville", "lang"=>"en", "lat"=>42.112424, "lon"=>-71.671354),
"/US/US-MA/wilbraham-9531" => array("type"=>"cdp", "pop"=>13970, "name_ru"=>"Уилбрэхэм", "name_en"=>"Wilbraham", "lang"=>"en", "lat"=>42.135792, "lon"=>-72.436281),
"/US/US-MA/williamstown-9532" => array("type"=>"cdp", "pop"=>7968, "name_ru"=>"Уильямстаун", "name_en"=>"Williamstown", "lang"=>"en", "lat"=>42.707149, "lon"=>-73.201875),
"/US/US-MA/wilmington-9533" => array("type"=>"cdp", "pop"=>21649, "name_ru"=>"Уилмингтон", "name_en"=>"Wilmington", "lang"=>"en", "lat"=>42.559366, "lon"=>-71.166580),
"/US/US-MA/winchendon-9534" => array("type"=>"cdp", "pop"=>10164, "name_ru"=>"Уинчендон", "name_en"=>"Winchendon", "lang"=>"en", "lat"=>42.681961, "lon"=>-72.044746),
"/US/US-MA/winchester-9535" => array("type"=>"cdp", "pop"=>21090, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>42.451850, "lon"=>-71.146924),
"/US/US-MA/winthrop-9536" => array("type"=>"cdp", "pop"=>21880, "name_ru"=>"Уинтроп", "name_en"=>"Winthrop", "lang"=>"en", "lat"=>42.380871, "lon"=>-70.968707),
"/US/US-MA/woburn-9537" => array("type"=>"city", "pop"=>36871, "name_ru"=>"Уоберн", "name_en"=>"Woburn", "lang"=>"en", "lat"=>42.485533, "lon"=>-71.155243),
"/US/US-MA/worcester-9538" => array("type"=>"city", "pop"=>175011, "name_ru"=>"Вустер", "name_en"=>"Worcester", "lang"=>"en", "lat"=>42.269633, "lon"=>-71.808920),
"/US/US-MA/yarmouth-port-9539" => array("type"=>"cdp", "pop"=>4271, "name_ru"=>"Ярмут Порт", "name_en"=>"Yarmouth Port", "lang"=>"en", "lat"=>41.709445, "lon"=>-70.228882),
"/US/US-MD" => array("type"=>"state", "pop"=>5633597, "name_ru"=>"Мэриленд", "name_en"=>"Maryland", "lang"=>"en"),
"/US/US-MD/aberdeen-8980" => array("type"=>"town", "pop"=>13993, "name_ru"=>"Абердин", "name_en"=>"Aberdeen", "lang"=>"en", "lat"=>39.511708, "lon"=>-76.172506),
"/US/US-MD/aberdeen-proving-ground-8981" => array("type"=>"cdp", "pop"=>5267, "name_ru"=>"Абердин Прувинг Граунд", "name_en"=>"Aberdeen Proving Ground", "lang"=>"en", "lat"=>39.472467, "lon"=>-76.129966),
"/US/US-MD/accident-8982" => array("type"=>"town", "pop"=>330, "name_ru"=>"Эксидент", "name_en"=>"Accident", "lang"=>"en", "lat"=>39.626945, "lon"=>-79.320884),
"/US/US-MD/accokeek-8983" => array("type"=>"cdp", "pop"=>4477, "name_ru"=>"Аккокик", "name_en"=>"Accokeek", "lang"=>"en", "lat"=>38.675967, "lon"=>-77.016258),
"/US/US-MD/adelphi-8984" => array("type"=>"cdp", "pop"=>13524, "name_ru"=>"Аделфи", "name_en"=>"Adelphi", "lang"=>"en", "lat"=>39.001900, "lon"=>-76.965460),
"/US/US-MD/andrews-afb-8985" => array("type"=>"cdp", "pop"=>10228, "name_ru"=>"Военно-воздушная база Эндрюс", "name_en"=>"Andrews AFB", "lang"=>"en", "lat"=>38.803376, "lon"=>-76.871814),
"/US/US-MD/annapolis-8986" => array("type"=>"city", "pop"=>36524, "name_ru"=>"Аннаполис", "name_en"=>"Annapolis", "lang"=>"en", "lat"=>38.971650, "lon"=>-76.503033),
"/US/US-MD/arbutus-8987" => array("type"=>"cdp", "pop"=>19750, "name_ru"=>"Арбутус", "name_en"=>"Arbutus", "lang"=>"en", "lat"=>39.242600, "lon"=>-76.692294),
"/US/US-MD/arden-on-the-severn-8988" => array("type"=>"cdp", "pop"=>2427, "name_ru"=>"Арден-он-зе-Севен", "name_en"=>"Arden-on-the-Severn", "lang"=>"en", "lat"=>39.069950, "lon"=>-76.594920),
"/US/US-MD/arnold-8989" => array("type"=>"cdp", "pop"=>20261, "name_ru"=>"Арнолд", "name_en"=>"Arnold", "lang"=>"en", "lat"=>39.042729, "lon"=>-76.499428),
"/US/US-MD/ashton-sandy-springs-8990" => array("type"=>"cdp", "pop"=>3092, "name_ru"=>"Эштон-Сэнди Спрингс", "name_en"=>"Ashton-Sandy Springs", "lang"=>"en", "lat"=>39.149467, "lon"=>-77.012680),
"/US/US-MD/aspen-hill-8991" => array("type"=>"cdp", "pop"=>45494, "name_ru"=>"Аспен Хилл", "name_en"=>"Aspen Hill", "lang"=>"en", "lat"=>39.092100, "lon"=>-77.078981),
"/US/US-MD/ballenger-creek-8992" => array("type"=>"cdp", "pop"=>5546, "name_ru"=>"Бэлленджер Крик", "name_en"=>"Ballenger Creek", "lang"=>"en", "lat"=>39.372384, "lon"=>-77.435622),
"/US/US-MD/baltimore-8993" => array("type"=>"city", "pop"=>636919, "name_ru"=>"Балтимор", "name_en"=>"Baltimore", "lang"=>"en", "lat"=>39.300800, "lon"=>-76.610616),
"/US/US-MD/barclay-8994" => array("type"=>"town", "pop"=>146, "name_ru"=>"Барклей", "name_en"=>"Barclay", "lang"=>"en", "lat"=>39.143813, "lon"=>-75.864230),
"/US/US-MD/barnesville-8995" => array("type"=>"town", "pop"=>197, "name_ru"=>"Барнсвилл", "name_en"=>"Barnesville", "lang"=>"en", "lat"=>39.219863, "lon"=>-77.381149),
"/US/US-MD/barton-8996" => array("type"=>"town", "pop"=>452, "name_ru"=>"Бартон", "name_en"=>"Barton", "lang"=>"en", "lat"=>39.532537, "lon"=>-79.017017),
"/US/US-MD/bel-air-8997" => array("type"=>"town", "pop"=>9826, "name_ru"=>"Бель Эр", "name_en"=>"Bel Air", "lang"=>"en", "lat"=>39.534600, "lon"=>-76.346914),
"/US/US-MD/bel-air-north-8998" => array("type"=>"cdp", "pop"=>14880, "name_ru"=>"Бель Эр Норт", "name_en"=>"Bel Air North", "lang"=>"en", "lat"=>39.553533, "lon"=>-76.373653),
"/US/US-MD/bel-air-south-8999" => array("type"=>"cdp", "pop"=>26421, "name_ru"=>"Бель Эр Саут", "name_en"=>"Bel Air South", "lang"=>"en", "lat"=>39.504033, "lon"=>-76.318100),
"/US/US-MD/beltsville-9000" => array("type"=>"cdp", "pop"=>14476, "name_ru"=>"Белтсвилл", "name_en"=>"Beltsville", "lang"=>"en", "lat"=>39.036787, "lon"=>-76.923534),
"/US/US-MD/berlin-9001" => array("type"=>"town", "pop"=>4056, "name_ru"=>"Берлин", "name_en"=>"Berlin", "lang"=>"en", "lat"=>38.325164, "lon"=>-75.218057),
"/US/US-MD/berwyn-heights-9002" => array("type"=>"town", "pop"=>2932, "name_ru"=>"Бервин Хайтс", "name_en"=>"Berwyn Heights", "lang"=>"en", "lat"=>38.992100, "lon"=>-76.911984),
"/US/US-MD/bethesda-9003" => array("type"=>"cdp", "pop"=>62936, "name_ru"=>"Бетесда", "name_en"=>"Bethesda", "lang"=>"en", "lat"=>38.989750, "lon"=>-77.120299),
"/US/US-MD/betterton-9004" => array("type"=>"town", "pop"=>354, "name_ru"=>"Беттертон", "name_en"=>"Betterton", "lang"=>"en", "lat"=>39.366967, "lon"=>-76.072507),
"/US/US-MD/bladensburg-9005" => array("type"=>"town", "pop"=>7568, "name_ru"=>"Блейденсбург", "name_en"=>"Bladensburg", "lang"=>"en", "lat"=>38.943650, "lon"=>-76.926162),
"/US/US-MD/boonsboro-9006" => array("type"=>"town", "pop"=>3399, "name_ru"=>"Бунсборо", "name_en"=>"Boonsboro", "lang"=>"en", "lat"=>39.509454, "lon"=>-77.655614),
"/US/US-MD/bowie-9007" => array("type"=>"city", "pop"=>52544, "name_ru"=>"Боуи", "name_en"=>"Bowie", "lang"=>"en", "lat"=>38.959467, "lon"=>-76.737739),
"/US/US-MD/bowleys-quarters-9008" => array("type"=>"cdp", "pop"=>5595, "name_ru"=>"Боулис Куортерс", "name_en"=>"Bowleys Quarters", "lang"=>"en", "lat"=>39.312267, "lon"=>-76.382310),
"/US/US-MD/braddock-heights-9009" => array("type"=>"cdp", "pop"=>4778, "name_ru"=>"Брэддок Хайтс", "name_en"=>"Braddock Heights", "lang"=>"en", "lat"=>39.414288, "lon"=>-77.507335),
"/US/US-MD/brandywine-9010" => array("type"=>"cdp", "pop"=>1406, "name_ru"=>"Брэндиуайн", "name_en"=>"Brandywine", "lang"=>"en", "lat"=>38.698525, "lon"=>-76.854286),
"/US/US-MD/brentwood-9011" => array("type"=>"town", "pop"=>2795, "name_ru"=>"Брентвуд", "name_en"=>"Brentwood", "lang"=>"en", "lat"=>38.943583, "lon"=>-76.957115),
"/US/US-MD/bridgeport-9012" => array("type"=>"cdp", "pop"=>2702, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>39.640896, "lon"=>-77.673239),
"/US/US-MD/brookeville-9013" => array("type"=>"town", "pop"=>131, "name_ru"=>"Бруквилл", "name_en"=>"Brookeville", "lang"=>"en", "lat"=>39.176647, "lon"=>-77.057357),
"/US/US-MD/brooklyn-park-9014" => array("type"=>"cdp", "pop"=>10987, "name_ru"=>"Бруклин Парк", "name_en"=>"Brooklyn Park", "lang"=>"en", "lat"=>39.217000, "lon"=>-76.610499),
"/US/US-MD/brookview-9015" => array("type"=>"town", "pop"=>62, "name_ru"=>"Бруквью", "name_en"=>"Brookview", "lang"=>"en", "lat"=>38.574426, "lon"=>-75.795107),
"/US/US-MD/brunswick-9016" => array("type"=>"town", "pop"=>5228, "name_ru"=>"Брунсвик", "name_en"=>"Brunswick", "lang"=>"en", "lat"=>39.316371, "lon"=>-77.623220),
"/US/US-MD/bryans-road-9017" => array("type"=>"cdp", "pop"=>3809, "name_ru"=>"Брайанс Роуд", "name_en"=>"Bryans Road", "lang"=>"en", "lat"=>38.631850, "lon"=>-77.079629),
"/US/US-MD/burkittsville-9018" => array("type"=>"town", "pop"=>186, "name_ru"=>"Беркитсвилл", "name_en"=>"Burkittsville", "lang"=>"en", "lat"=>39.391359, "lon"=>-77.627387),
"/US/US-MD/burtonsville-9019" => array("type"=>"cdp", "pop"=>5853, "name_ru"=>"Бертонсвилл", "name_en"=>"Burtonsville", "lang"=>"en", "lat"=>39.111200, "lon"=>-76.932216),
"/US/US-MD/cabin-john-brookmont-9020" => array("type"=>"cdp", "pop"=>5341, "name_ru"=>"Кэбин Джон-Брукмонт", "name_en"=>"Cabin John-Brookmont", "lang"=>"en", "lat"=>38.967367, "lon"=>-77.147626),
"/US/US-MD/california-9021" => array("type"=>"cdp", "pop"=>7626, "name_ru"=>"Калифорния", "name_en"=>"California", "lang"=>"en", "lat"=>38.296148, "lon"=>-76.496161),
"/US/US-MD/calvert-beach-long-beach-9022" => array("type"=>"cdp", "pop"=>1728, "name_ru"=>"Калверт Бич-Лонг Бич", "name_en"=>"Calvert Beach-Long Beach", "lang"=>"en", "lat"=>38.460869, "lon"=>-76.478114),
"/US/US-MD/calverton-9023" => array("type"=>"cdp", "pop"=>12046, "name_ru"=>"Калвертон", "name_en"=>"Calverton", "lang"=>"en", "lat"=>39.056350, "lon"=>-76.948224),
"/US/US-MD/cambridge-9024" => array("type"=>"city", "pop"=>11752, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>38.561840, "lon"=>-76.077827),
"/US/US-MD/camp-springs-9025" => array("type"=>"cdp", "pop"=>16392, "name_ru"=>"Кэмп Спрингс", "name_en"=>"Camp Springs", "lang"=>"en", "lat"=>38.803550, "lon"=>-76.920756),
"/US/US-MD/cape-st.-claire-9026" => array("type"=>"cdp", "pop"=>7878, "name_ru"=>"Кейп Сент Клэр", "name_en"=>"Cape St. Claire", "lang"=>"en", "lat"=>39.044300, "lon"=>-76.445924),
"/US/US-MD/capitol-heights-9027" => array("type"=>"town", "pop"=>4096, "name_ru"=>"Капитол Хайтс", "name_en"=>"Capitol Heights", "lang"=>"en", "lat"=>38.877850, "lon"=>-76.908071),
"/US/US-MD/carmody-hills-pepper-mill-village-9028" => array("type"=>"cdp", "pop"=>4815, "name_ru"=>"Кармоди Хиллс-Пеппер Милл Виллидж", "name_en"=>"Carmody Hills-Pepper Mill Village", "lang"=>"en", "lat"=>38.893550, "lon"=>-76.888482),
"/US/US-MD/carney-9029" => array("type"=>"cdp", "pop"=>25578, "name_ru"=>"Карни", "name_en"=>"Carney", "lang"=>"en", "lat"=>39.404433, "lon"=>-76.522748),
"/US/US-MD/catonsville-9030" => array("type"=>"cdp", "pop"=>35233, "name_ru"=>"Катонсвилл", "name_en"=>"Catonsville", "lang"=>"en", "lat"=>39.264500, "lon"=>-76.741662),
"/US/US-MD/cecilton-9031" => array("type"=>"town", "pop"=>493, "name_ru"=>"Сесилтон", "name_en"=>"Cecilton", "lang"=>"en", "lat"=>39.402807, "lon"=>-75.869769),
"/US/US-MD/centreville-9032" => array("type"=>"town", "pop"=>3458, "name_ru"=>"Сентревилл", "name_en"=>"Centreville", "lang"=>"en", "lat"=>39.043521, "lon"=>-76.065290),
"/US/US-MD/charlestown-9033" => array("type"=>"town", "pop"=>1103, "name_ru"=>"Чарльзтаун", "name_en"=>"Charlestown", "lang"=>"en", "lat"=>39.575241, "lon"=>-75.979251),
"/US/US-MD/charlotte-hall-9034" => array("type"=>"cdp", "pop"=>1992, "name_ru"=>"Шарлот Холл", "name_en"=>"Charlotte Hall", "lang"=>"en", "lat"=>38.478243, "lon"=>-76.789501),
"/US/US-MD/chesapeake-beach-9035" => array("type"=>"town", "pop"=>3399, "name_ru"=>"Чесапик Бич", "name_en"=>"Chesapeake Beach", "lang"=>"en", "lat"=>38.687286, "lon"=>-76.535865),
"/US/US-MD/chesapeake-city-9036" => array("type"=>"town", "pop"=>830, "name_ru"=>"Чесапик Сити", "name_en"=>"Chesapeake City", "lang"=>"en", "lat"=>39.529664, "lon"=>-75.810670),
"/US/US-MD/chesapeake-ranch-estates-9037" => array("type"=>"cdp", "pop"=>5423, "name_ru"=>"Чесапик Рэнч Эстетйс", "name_en"=>"Chesapeake Ranch Estates", "lang"=>"en", "lat"=>38.355647, "lon"=>-76.415853),
"/US/US-MD/chestertown-9038" => array("type"=>"town", "pop"=>4899, "name_ru"=>"Честертаун", "name_en"=>"Chestertown", "lang"=>"en", "lat"=>39.219211, "lon"=>-76.070548),
"/US/US-MD/cheverly-9039" => array("type"=>"town", "pop"=>6372, "name_ru"=>"Чеверли", "name_en"=>"Cheverly", "lang"=>"en", "lat"=>38.925750, "lon"=>-76.913860),
"/US/US-MD/chevy-chase-9040" => array("type"=>"town", "pop"=>2803, "name_ru"=>"Чеви Чейз", "name_en"=>"Chevy Chase", "lang"=>"en", "lat"=>38.980500, "lon"=>-77.084122),
"/US/US-MD/chevy-chase-9041" => array("type"=>"cdp", "pop"=>8559, "name_ru"=>"Чеви Чейз", "name_en"=>"Chevy Chase", "lang"=>"en", "lat"=>38.994950, "lon"=>-77.073132),
"/US/US-MD/chevy-chase-section-five-9042" => array("type"=>"village", "pop"=>659, "name_ru"=>"Чеви Чейз Секшен Файв", "name_en"=>"Chevy Chase Section Five", "lang"=>"en", "lat"=>38.982950, "lon"=>-77.075087),
"/US/US-MD/chevy-chase-section-three-9043" => array("type"=>"village", "pop"=>793, "name_ru"=>"Чеви Чейз Секшен Три", "name_en"=>"Chevy Chase Section Three", "lang"=>"en", "lat"=>38.969450, "lon"=>-77.079267),
"/US/US-MD/chevy-chase-village-9044" => array("type"=>"town", "pop"=>2109, "name_ru"=>"Чеви Чейз Виллидж", "name_en"=>"Chevy Chase Village", "lang"=>"en", "lat"=>38.979100, "lon"=>-77.074474),
"/US/US-MD/chillum-9045" => array("type"=>"cdp", "pop"=>31309, "name_ru"=>"Чиллум", "name_en"=>"Chillum", "lang"=>"en", "lat"=>38.968200, "lon"=>-76.977843),
"/US/US-MD/church-creek-9046" => array("type"=>"town", "pop"=>78, "name_ru"=>"Черч Крик", "name_en"=>"Church Creek", "lang"=>"en", "lat"=>38.505306, "lon"=>-76.154733),
"/US/US-MD/church-hill-9047" => array("type"=>"town", "pop"=>670, "name_ru"=>"Черч Хилл", "name_en"=>"Church Hill", "lang"=>"en", "lat"=>39.143569, "lon"=>-75.980708),
"/US/US-MD/clear-spring-9048" => array("type"=>"town", "pop"=>461, "name_ru"=>"Клир Спринг", "name_en"=>"Clear Spring", "lang"=>"en", "lat"=>39.655916, "lon"=>-77.931108),
"/US/US-MD/clinton-9049" => array("type"=>"cdp", "pop"=>19987, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>38.763750, "lon"=>-76.895582),
"/US/US-MD/clover-hill-9050" => array("type"=>"cdp", "pop"=>2823, "name_ru"=>"Клавер Хилл", "name_en"=>"Clover Hill", "lang"=>"en", "lat"=>39.457464, "lon"=>-77.429928),
"/US/US-MD/cloverly-9051" => array("type"=>"cdp", "pop"=>7904, "name_ru"=>"Клаверли", "name_en"=>"Cloverly", "lang"=>"en", "lat"=>39.102850, "lon"=>-76.976577),
"/US/US-MD/cockeysville-9052" => array("type"=>"cdp", "pop"=>18668, "name_ru"=>"Кокисвилл", "name_en"=>"Cockeysville", "lang"=>"en", "lat"=>39.477850, "lon"=>-76.631824),
"/US/US-MD/colesville-9053" => array("type"=>"cdp", "pop"=>18819, "name_ru"=>"Коулсвилл", "name_en"=>"Colesville", "lang"=>"en", "lat"=>39.088300, "lon"=>-77.007138),
"/US/US-MD/college-park-9054" => array("type"=>"city", "pop"=>26925, "name_ru"=>"Колледж Парк", "name_en"=>"College Park", "lang"=>"en", "lat"=>38.996062, "lon"=>-76.934785),
"/US/US-MD/colmar-manor-9055" => array("type"=>"town", "pop"=>1256, "name_ru"=>"Колмар Мэнор", "name_en"=>"Colmar Manor", "lang"=>"en", "lat"=>38.930500, "lon"=>-76.943417),
"/US/US-MD/columbia-9056" => array("type"=>"cdp", "pop"=>75883, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>39.203000, "lon"=>-76.857981),
"/US/US-MD/coral-hills-9057" => array("type"=>"cdp", "pop"=>11032, "name_ru"=>"Корал Хиллс", "name_en"=>"Coral Hills", "lang"=>"en", "lat"=>38.871650, "lon"=>-76.923516),
"/US/US-MD/cottage-city-9058" => array("type"=>"town", "pop"=>1122, "name_ru"=>"Коттедж Сити", "name_en"=>"Cottage City", "lang"=>"en", "lat"=>38.936550, "lon"=>-76.949729),
"/US/US-MD/cresaptown-bel-air-9059" => array("type"=>"cdp", "pop"=>4586, "name_ru"=>"Кресаптаун", "name_en"=>"Cresaptown-Bel Air", "lang"=>"en", "lat"=>39.588188, "lon"=>-78.858863),
"/US/US-MD/crisfield-9060" => array("type"=>"city", "pop"=>2749, "name_ru"=>"Крисфилд", "name_en"=>"Crisfield", "lang"=>"en", "lat"=>37.977064, "lon"=>-75.859340),
"/US/US-MD/crofton-9061" => array("type"=>"cdp", "pop"=>12781, "name_ru"=>"Крофтон", "name_en"=>"Crofton", "lang"=>"en", "lat"=>39.011000, "lon"=>-76.687261),
"/US/US-MD/crownsville-9062" => array("type"=>"cdp", "pop"=>1514, "name_ru"=>"Краунсвилл", "name_en"=>"Crownsville", "lang"=>"en", "lat"=>39.016750, "lon"=>-76.595520),
"/US/US-MD/cumberland-9063" => array("type"=>"city", "pop"=>20495, "name_ru"=>"Камберленд", "name_en"=>"Cumberland", "lang"=>"en", "lat"=>39.651067, "lon"=>-78.763805),
"/US/US-MD/damascus-9064" => array("type"=>"cdp", "pop"=>9817, "name_ru"=>"Дамаскус", "name_en"=>"Damascus", "lang"=>"en", "lat"=>39.277733, "lon"=>-77.204975),
"/US/US-MD/deale-9065" => array("type"=>"cdp", "pop"=>4151, "name_ru"=>"Дил", "name_en"=>"Deale", "lang"=>"en", "lat"=>38.793200, "lon"=>-76.546581),
"/US/US-MD/deer-park-9066" => array("type"=>"town", "pop"=>382, "name_ru"=>"Дир Парк", "name_en"=>"Deer Park", "lang"=>"en", "lat"=>39.423484, "lon"=>-79.326701),
"/US/US-MD/delmar-9067" => array("type"=>"town", "pop"=>3356, "name_ru"=>"Делмар", "name_en"=>"Delmar", "lang"=>"en", "lat"=>38.452601, "lon"=>-75.574754),
"/US/US-MD/denton-9068" => array("type"=>"town", "pop"=>4022, "name_ru"=>"Дентон", "name_en"=>"Denton", "lang"=>"en", "lat"=>38.882323, "lon"=>-75.822302),
"/US/US-MD/discovery-spring-garden-9069" => array("type"=>"cdp", "pop"=>2443, "name_ru"=>"Дискавери-Спринг Гарден", "name_en"=>"Discovery-Spring Garden", "lang"=>"en", "lat"=>39.460335, "lon"=>-77.360793),
"/US/US-MD/district-heights-9070" => array("type"=>"city", "pop"=>6034, "name_ru"=>"Дистрикт Хайтс", "name_en"=>"District Heights", "lang"=>"en", "lat"=>38.859117, "lon"=>-76.887214),
"/US/US-MD/dodge-park-9071" => array("type"=>"cdp", "pop"=>4842, "name_ru"=>"Додж Парк", "name_en"=>"Dodge Park", "lang"=>"en", "lat"=>38.931150, "lon"=>-76.881436),
"/US/US-MD/dundalk-9072" => array("type"=>"cdp", "pop"=>65800, "name_ru"=>"Дандолк", "name_en"=>"Dundalk", "lang"=>"en", "lat"=>39.268567, "lon"=>-76.497368),
"/US/US-MD/eagle-harbor-9073" => array("type"=>"town", "pop"=>56, "name_ru"=>"Игл Харбор", "name_en"=>"Eagle Harbor", "lang"=>"en", "lat"=>38.566373, "lon"=>-76.687049),
"/US/US-MD/east-new-market-9074" => array("type"=>"town", "pop"=>266, "name_ru"=>"Ист Нью Маркет", "name_en"=>"East New Market", "lang"=>"en", "lat"=>38.599102, "lon"=>-75.923875),
"/US/US-MD/east-riverdale-9076" => array("type"=>"cdp", "pop"=>14187, "name_ru"=>"Ист Ривердейл", "name_en"=>"East Riverdale", "lang"=>"en", "lat"=>38.960567, "lon"=>-76.910454),
"/US/US-MD/easton-9075" => array("type"=>"town", "pop"=>14677, "name_ru"=>"Истон", "name_en"=>"Easton", "lang"=>"en", "lat"=>38.774637, "lon"=>-76.069815),
"/US/US-MD/edgemere-9077" => array("type"=>"cdp", "pop"=>9226, "name_ru"=>"Эджмир", "name_en"=>"Edgemere", "lang"=>"en", "lat"=>39.222933, "lon"=>-76.459688),
"/US/US-MD/edgewood-9078" => array("type"=>"cdp", "pop"=>23903, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>39.420450, "lon"=>-76.299249),
"/US/US-MD/edmonston-9079" => array("type"=>"town", "pop"=>1329, "name_ru"=>"Эдмонстон", "name_en"=>"Edmonston", "lang"=>"en", "lat"=>38.950050, "lon"=>-76.933543),
"/US/US-MD/eldersburg-9080" => array("type"=>"cdp", "pop"=>9720, "name_ru"=>"Эльдерсбург", "name_en"=>"Eldersburg", "lang"=>"en", "lat"=>39.408400, "lon"=>-76.943179),
"/US/US-MD/eldorado-9081" => array("type"=>"town", "pop"=>57, "name_ru"=>"Эльдорадо", "name_en"=>"Eldorado", "lang"=>"en", "lat"=>38.582494, "lon"=>-75.790752),
"/US/US-MD/elkridge-9082" => array("type"=>"cdp", "pop"=>12953, "name_ru"=>"Элкридж", "name_en"=>"Elkridge", "lang"=>"en", "lat"=>39.201943, "lon"=>-76.750501),
"/US/US-MD/elkton-9083" => array("type"=>"town", "pop"=>14842, "name_ru"=>"Элктон", "name_en"=>"Elkton", "lang"=>"en", "lat"=>39.603504, "lon"=>-75.820762),
"/US/US-MD/ellicott-city-9084" => array("type"=>"cdp", "pop"=>41396, "name_ru"=>"Элликотт Сити", "name_en"=>"Ellicott City", "lang"=>"en", "lat"=>39.272050, "lon"=>-76.834574),
"/US/US-MD/emmitsburg-9085" => array("type"=>"town", "pop"=>2365, "name_ru"=>"Эммитсбург", "name_en"=>"Emmitsburg", "lang"=>"en", "lat"=>39.704600, "lon"=>-77.325586),
"/US/US-MD/essex-9086" => array("type"=>"cdp", "pop"=>40872, "name_ru"=>"Эссекс", "name_en"=>"Essex", "lang"=>"en", "lat"=>39.302700, "lon"=>-76.446009),
"/US/US-MD/fairland-9087" => array("type"=>"cdp", "pop"=>19828, "name_ru"=>"Фэрленд", "name_en"=>"Fairland", "lang"=>"en", "lat"=>39.082175, "lon"=>-76.951554),
"/US/US-MD/fairmount-heights-9088" => array("type"=>"town", "pop"=>1495, "name_ru"=>"Фэрмаунт Хайтс", "name_en"=>"Fairmount Heights", "lang"=>"en", "lat"=>38.901650, "lon"=>-76.914809),
"/US/US-MD/fallston-9089" => array("type"=>"cdp", "pop"=>5730, "name_ru"=>"Фоллстон", "name_en"=>"Fallston", "lang"=>"en", "lat"=>39.521033, "lon"=>-76.426489),
"/US/US-MD/federalsburg-9090" => array("type"=>"town", "pop"=>2625, "name_ru"=>"Федералсбург", "name_en"=>"Federalsburg", "lang"=>"en", "lat"=>38.691012, "lon"=>-75.773641),
"/US/US-MD/ferndale-9091" => array("type"=>"cdp", "pop"=>16355, "name_ru"=>"Ферндейл", "name_en"=>"Ferndale", "lang"=>"en", "lat"=>39.186500, "lon"=>-76.634380),
"/US/US-MD/forest-heights-9092" => array("type"=>"town", "pop"=>2554, "name_ru"=>"Форест Хайтс", "name_en"=>"Forest Heights", "lang"=>"en", "lat"=>38.810150, "lon"=>-76.998817),
"/US/US-MD/forestville-9093" => array("type"=>"cdp", "pop"=>16731, "name_ru"=>"Форествилл", "name_en"=>"Forestville", "lang"=>"en", "lat"=>38.849700, "lon"=>-76.874808),
"/US/US-MD/fort-meade-9094" => array("type"=>"cdp", "pop"=>12509, "name_ru"=>"Форт Миад", "name_en"=>"Fort Meade", "lang"=>"en", "lat"=>39.108100, "lon"=>-76.741969),
"/US/US-MD/fort-ritchie-9095" => array("type"=>"cdp", "pop"=>1249, "name_ru"=>"Форт Ритчи", "name_en"=>"Fort Ritchie", "lang"=>"en", "lat"=>39.700507, "lon"=>-77.503781),
"/US/US-MD/fort-washington-9096" => array("type"=>"cdp", "pop"=>24032, "name_ru"=>"Форт Вашингтон", "name_en"=>"Fort Washington", "lang"=>"en", "lat"=>38.735300, "lon"=>-77.003206),
"/US/US-MD/frederick-9097" => array("type"=>"city", "pop"=>59213, "name_ru"=>"Фредерик", "name_en"=>"Frederick", "lang"=>"en", "lat"=>39.428187, "lon"=>-77.416905),
"/US/US-MD/friendly-9098" => array("type"=>"cdp", "pop"=>9028, "name_ru"=>"Френдли", "name_en"=>"Friendly", "lang"=>"en", "lat"=>38.756400, "lon"=>-76.965782),
"/US/US-MD/friendsville-9099" => array("type"=>"town", "pop"=>503, "name_ru"=>"Френдсвилл", "name_en"=>"Friendsville", "lang"=>"en", "lat"=>39.662983, "lon"=>-79.405180),
"/US/US-MD/frostburg-9100" => array("type"=>"city", "pop"=>7719, "name_ru"=>"Фростбург", "name_en"=>"Frostburg", "lang"=>"en", "lat"=>39.650000, "lon"=>-78.928982),
"/US/US-MD/fruitland-9101" => array("type"=>"city", "pop"=>4431, "name_ru"=>"Фрутленд", "name_en"=>"Fruitland", "lang"=>"en", "lat"=>38.321944, "lon"=>-75.626272),
"/US/US-MD/funkstown-9102" => array("type"=>"town", "pop"=>937, "name_ru"=>"Фанкстаун", "name_en"=>"Funkstown", "lang"=>"en", "lat"=>39.608300, "lon"=>-77.708040),
"/US/US-MD/gaithersburg-9103" => array("type"=>"city", "pop"=>58744, "name_ru"=>"Гейтерсбург", "name_en"=>"Gaithersburg", "lang"=>"en", "lat"=>39.141250, "lon"=>-77.216641),
"/US/US-MD/galena-9104" => array("type"=>"town", "pop"=>504, "name_ru"=>"Галена", "name_en"=>"Galena", "lang"=>"en", "lat"=>39.343308, "lon"=>-75.879160),
"/US/US-MD/galestown-9105" => array("type"=>"town", "pop"=>97, "name_ru"=>"Гейлстаун", "name_en"=>"Galestown", "lang"=>"en", "lat"=>38.561982, "lon"=>-75.716927),
"/US/US-MD/garrett-park-9106" => array("type"=>"town", "pop"=>955, "name_ru"=>"Гарретт парк", "name_en"=>"Garrett Park", "lang"=>"en", "lat"=>39.036150, "lon"=>-77.094646),
"/US/US-MD/garrison-9107" => array("type"=>"cdp", "pop"=>5045, "name_ru"=>"Гаррисон", "name_en"=>"Garrison", "lang"=>"en", "lat"=>39.401350, "lon"=>-76.749328),
"/US/US-MD/germantown-9108" => array("type"=>"cdp", "pop"=>41145, "name_ru"=>"Джермэнтаун", "name_en"=>"Germantown", "lang"=>"en", "lat"=>39.178192, "lon"=>-77.260652),
"/US/US-MD/glen-burnie-9110" => array("type"=>"cdp", "pop"=>37305, "name_ru"=>"Глен Берни", "name_en"=>"Glen Burnie", "lang"=>"en", "lat"=>39.166350, "lon"=>-76.602399),
"/US/US-MD/glen-echo-9111" => array("type"=>"town", "pop"=>255, "name_ru"=>"Глен Эко", "name_en"=>"Glen Echo", "lang"=>"en", "lat"=>38.967350, "lon"=>-77.140766),
"/US/US-MD/glenarden-9109" => array("type"=>"town", "pop"=>6306, "name_ru"=>"Гленарден", "name_en"=>"Glenarden", "lang"=>"en", "lat"=>38.930750, "lon"=>-76.862903),
"/US/US-MD/glenn-dale-9112" => array("type"=>"cdp", "pop"=>9689, "name_ru"=>"Гленн Дейл", "name_en"=>"Glenn Dale", "lang"=>"en", "lat"=>38.986864, "lon"=>-76.800947),
"/US/US-MD/goddard-9113" => array("type"=>"cdp", "pop"=>4576, "name_ru"=>"Годдард", "name_en"=>"Goddard", "lang"=>"en", "lat"=>38.989500, "lon"=>-76.853651),
"/US/US-MD/golden-beach-9114" => array("type"=>"cdp", "pop"=>2944, "name_ru"=>"Голден Бич", "name_en"=>"Golden Beach", "lang"=>"en", "lat"=>38.487269, "lon"=>-76.693242),
"/US/US-MD/goldsboro-9115" => array("type"=>"town", "pop"=>216, "name_ru"=>"Гоулсборо", "name_en"=>"Goldsboro", "lang"=>"en", "lat"=>39.035370, "lon"=>-75.787430),
"/US/US-MD/grantsville-9116" => array("type"=>"town", "pop"=>649, "name_ru"=>"Грантсвилл", "name_en"=>"Grantsville", "lang"=>"en", "lat"=>39.696539, "lon"=>-79.157117),
"/US/US-MD/grasonville-9117" => array("type"=>"cdp", "pop"=>2439, "name_ru"=>"Грейсонвилл", "name_en"=>"Grasonville", "lang"=>"en", "lat"=>38.957977, "lon"=>-76.194228),
"/US/US-MD/greater-upper-marlboro-9118" => array("type"=>"cdp", "pop"=>11528, "name_ru"=>"Грейтер Аппер Марлборо", "name_en"=>"Greater Upper Marlboro", "lang"=>"en", "lat"=>38.830296, "lon"=>-76.752666),
"/US/US-MD/green-haven-9120" => array("type"=>"cdp", "pop"=>14416, "name_ru"=>"Грин Хейвен", "name_en"=>"Green Haven", "lang"=>"en", "lat"=>39.136700, "lon"=>-76.540106),
"/US/US-MD/green-valley-9122" => array("type"=>"cdp", "pop"=>9424, "name_ru"=>"Грин Валли", "name_en"=>"Green Valley", "lang"=>"en", "lat"=>39.343600, "lon"=>-77.245170),
"/US/US-MD/greenbelt-9119" => array("type"=>"city", "pop"=>21098, "name_ru"=>"Гринбелт", "name_en"=>"Greenbelt", "lang"=>"en", "lat"=>38.994833, "lon"=>-76.885732),
"/US/US-MD/greensboro-9121" => array("type"=>"town", "pop"=>1998, "name_ru"=>"Гринсборо", "name_en"=>"Greensboro", "lang"=>"en", "lat"=>38.976141, "lon"=>-75.806812),
"/US/US-MD/hagerstown-9123" => array("type"=>"city", "pop"=>39728, "name_ru"=>"Хагерстаун", "name_en"=>"Hagerstown", "lang"=>"en", "lat"=>39.637950, "lon"=>-77.719277),
"/US/US-MD/halfway-9124" => array("type"=>"cdp", "pop"=>8873, "name_ru"=>"Хафуэй", "name_en"=>"Halfway", "lang"=>"en", "lat"=>39.613167, "lon"=>-77.770650),
"/US/US-MD/hampstead-9125" => array("type"=>"town", "pop"=>5467, "name_ru"=>"Хемпстед", "name_en"=>"Hampstead", "lang"=>"en", "lat"=>39.618433, "lon"=>-76.855755),
"/US/US-MD/hampton-9126" => array("type"=>"cdp", "pop"=>4926, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>39.418250, "lon"=>-76.572776),
"/US/US-MD/hancock-9127" => array("type"=>"town", "pop"=>1743, "name_ru"=>"Хэнкок", "name_en"=>"Hancock", "lang"=>"en", "lat"=>39.702891, "lon"=>-78.173663),
"/US/US-MD/havre-de-grace-9128" => array("type"=>"city", "pop"=>13085, "name_ru"=>"Хавр де Грейс", "name_en"=>"Havre de Grace", "lang"=>"en", "lat"=>39.547167, "lon"=>-76.098992),
"/US/US-MD/hebron-9129" => array("type"=>"town", "pop"=>1044, "name_ru"=>"Хеброн", "name_en"=>"Hebron", "lang"=>"en", "lat"=>38.417595, "lon"=>-75.688022),
"/US/US-MD/henderson-9130" => array("type"=>"town", "pop"=>122, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>39.074902, "lon"=>-75.767124),
"/US/US-MD/herald-harbor-9131" => array("type"=>"cdp", "pop"=>1707, "name_ru"=>"Херальд Харбор", "name_en"=>"Herald Harbor", "lang"=>"en", "lat"=>39.050367, "lon"=>-76.574225),
"/US/US-MD/highland-beach-9132" => array("type"=>"town", "pop"=>110, "name_ru"=>"Хайленд Бич", "name_en"=>"Highland Beach", "lang"=>"en", "lat"=>38.931767, "lon"=>-76.466780),
"/US/US-MD/hillandale-9133" => array("type"=>"cdp", "pop"=>10318, "name_ru"=>"Хилландейл", "name_en"=>"Hillandale", "lang"=>"en", "lat"=>39.019600, "lon"=>-76.974740),
"/US/US-MD/hillcrest-heights-9134" => array("type"=>"cdp", "pop"=>17136, "name_ru"=>"Хиллкрест Хайтс", "name_en"=>"Hillcrest Heights", "lang"=>"en", "lat"=>38.836550, "lon"=>-76.964044),
"/US/US-MD/hillsboro-9135" => array("type"=>"town", "pop"=>157, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>38.916171, "lon"=>-75.941694),
"/US/US-MD/hillsmere-shores-9136" => array("type"=>"cdp", "pop"=>3321, "name_ru"=>"Хиллсмер Шорс", "name_en"=>"Hillsmere Shores", "lang"=>"en", "lat"=>38.931900, "lon"=>-76.497074),
"/US/US-MD/hughesville-9137" => array("type"=>"cdp", "pop"=>1319, "name_ru"=>"Хьюсвилл", "name_en"=>"Hughesville", "lang"=>"en", "lat"=>38.538725, "lon"=>-76.783719),
"/US/US-MD/hurlock-9138" => array("type"=>"town", "pop"=>2048, "name_ru"=>"Херлок", "name_en"=>"Hurlock", "lang"=>"en", "lat"=>38.629139, "lon"=>-75.871027),
"/US/US-MD/hyattsville-9139" => array("type"=>"city", "pop"=>15376, "name_ru"=>"Хайатсвилл", "name_en"=>"Hyattsville", "lang"=>"en", "lat"=>38.959233, "lon"=>-76.953446),
"/US/US-MD/indian-head-9140" => array("type"=>"town", "pop"=>3758, "name_ru"=>"Индиан Хед", "name_en"=>"Indian Head", "lang"=>"en", "lat"=>38.598477, "lon"=>-77.157869),
"/US/US-MD/jarrettsville-9141" => array("type"=>"cdp", "pop"=>2148, "name_ru"=>"Джарреттсвилл", "name_en"=>"Jarrettsville", "lang"=>"en", "lat"=>39.603667, "lon"=>-76.472672),
"/US/US-MD/jessup-9142" => array("type"=>"cdp", "pop"=>6537, "name_ru"=>"Джессап", "name_en"=>"Jessup", "lang"=>"en", "lat"=>39.146133, "lon"=>-76.772990),
"/US/US-MD/joppatowne-9143" => array("type"=>"cdp", "pop"=>11084, "name_ru"=>"Джоппатаун", "name_en"=>"Joppatowne", "lang"=>"en", "lat"=>39.415450, "lon"=>-76.354196),
"/US/US-MD/keedysville-9144" => array("type"=>"town", "pop"=>862, "name_ru"=>"Кидисвилл", "name_en"=>"Keedysville", "lang"=>"en", "lat"=>39.486965, "lon"=>-77.699236),
"/US/US-MD/kensington-9145" => array("type"=>"town", "pop"=>1946, "name_ru"=>"Кенсингтон", "name_en"=>"Kensington", "lang"=>"en", "lat"=>39.025900, "lon"=>-77.073191),
"/US/US-MD/kentland-9146" => array("type"=>"cdp", "pop"=>7967, "name_ru"=>"Кентленд", "name_en"=>"Kentland", "lang"=>"en", "lat"=>38.918083, "lon"=>-76.894011),
"/US/US-MD/kettering-9147" => array("type"=>"cdp", "pop"=>9901, "name_ru"=>"Кеттеринг", "name_en"=>"Kettering", "lang"=>"en", "lat"=>38.890067, "lon"=>-76.785621),
"/US/US-MD/kingstown-9148" => array("type"=>"cdp", "pop"=>1660, "name_ru"=>"Кингстаун", "name_en"=>"Kingstown", "lang"=>"en", "lat"=>39.206169, "lon"=>-76.048123),
"/US/US-MD/kingsville-9149" => array("type"=>"cdp", "pop"=>3550, "name_ru"=>"Кингсвилл", "name_en"=>"Kingsville", "lang"=>"en", "lat"=>39.449050, "lon"=>-76.418345),
"/US/US-MD/kitzmiller-9150" => array("type"=>"town", "pop"=>278, "name_ru"=>"Кицмиллер", "name_en"=>"Kitzmiller", "lang"=>"en", "lat"=>39.389212, "lon"=>-79.183667),
"/US/US-MD/la-plata-9157" => array("type"=>"town", "pop"=>8879, "name_ru"=>"Ла Плата", "name_en"=>"La Plata", "lang"=>"en", "lat"=>38.534112, "lon"=>-76.970488),
"/US/US-MD/la-vale-9160" => array("type"=>"cdp", "pop"=>4694, "name_ru"=>"Ла Вейл", "name_en"=>"La Vale", "lang"=>"en", "lat"=>39.663400, "lon"=>-78.818612),
"/US/US-MD/lake-shore-9151" => array("type"=>"cdp", "pop"=>13269, "name_ru"=>"Лейк Шор", "name_en"=>"Lake Shore", "lang"=>"en", "lat"=>39.097603, "lon"=>-76.485638),
"/US/US-MD/landover-9152" => array("type"=>"cdp", "pop"=>5052, "name_ru"=>"Лэндовер", "name_en"=>"Landover", "lang"=>"en", "lat"=>38.936150, "lon"=>-76.901485),
"/US/US-MD/landover-hills-9153" => array("type"=>"town", "pop"=>1514, "name_ru"=>"Лэндовер Хилс", "name_en"=>"Landover Hills", "lang"=>"en", "lat"=>38.943133, "lon"=>-76.891116),
"/US/US-MD/langley-park-9154" => array("type"=>"cdp", "pop"=>17474, "name_ru"=>"Лэнгли Парк", "name_en"=>"Langley Park", "lang"=>"en", "lat"=>38.993950, "lon"=>-76.982061),
"/US/US-MD/lanham-seabrook-9155" => array("type"=>"cdp", "pop"=>16792, "name_ru"=>"Лэнхэм-Сибрук", "name_en"=>"Lanham-Seabrook", "lang"=>"en", "lat"=>38.966800, "lon"=>-76.844047),
"/US/US-MD/lansdowne-baltimore-highlands-9156" => array("type"=>"cdp", "pop"=>15509, "name_ru"=>"Лансдаун-Балтимор Хайлендс", "name_en"=>"Lansdowne-Baltimore Highlands", "lang"=>"en", "lat"=>39.236450, "lon"=>-76.654650),
"/US/US-MD/largo-9158" => array("type"=>"cdp", "pop"=>9475, "name_ru"=>"Ларго", "name_en"=>"Largo", "lang"=>"en", "lat"=>38.879500, "lon"=>-76.830178),
"/US/US-MD/laurel-9159" => array("type"=>"city", "pop"=>22329, "name_ru"=>"Лорел", "name_en"=>"Laurel", "lang"=>"en", "lat"=>39.097750, "lon"=>-76.858554),
"/US/US-MD/lawsonia-9161" => array("type"=>"cdp", "pop"=>1326, "name_ru"=>"Лосония", "name_en"=>"Lawsonia", "lang"=>"en", "lat"=>37.957497, "lon"=>-75.842039),
"/US/US-MD/laytonsville-9162" => array("type"=>"town", "pop"=>353, "name_ru"=>"Лейтонсвилл", "name_en"=>"Laytonsville", "lang"=>"en", "lat"=>39.209633, "lon"=>-77.141557),
"/US/US-MD/leonardtown-9163" => array("type"=>"town", "pop"=>2214, "name_ru"=>"Леонардтаун", "name_en"=>"Leonardtown", "lang"=>"en", "lat"=>38.299620, "lon"=>-76.641688),
"/US/US-MD/lexington-park-9164" => array("type"=>"cdp", "pop"=>9943, "name_ru"=>"Лексингтон Парк", "name_en"=>"Lexington Park", "lang"=>"en", "lat"=>38.260413, "lon"=>-76.445858),
"/US/US-MD/linganore-bartonsville-9165" => array("type"=>"cdp", "pop"=>4079, "name_ru"=>"Линганор-Бартонсвилл", "name_en"=>"Linganore-Bartonsville", "lang"=>"en", "lat"=>39.408565, "lon"=>-77.324090),
"/US/US-MD/linthicum-9166" => array("type"=>"cdp", "pop"=>7547, "name_ru"=>"Линтикум", "name_en"=>"Linthicum", "lang"=>"en", "lat"=>39.205400, "lon"=>-76.666030),
"/US/US-MD/loch-lynn-heights-9168" => array("type"=>"town", "pop"=>435, "name_ru"=>"Лох Линн Хайтс", "name_en"=>"Loch Lynn Heights", "lang"=>"en", "lat"=>39.391295, "lon"=>-79.373123),
"/US/US-MD/lochearn-9167" => array("type"=>"cdp", "pop"=>25240, "name_ru"=>"Лошерн", "name_en"=>"Lochearn", "lang"=>"en", "lat"=>39.347333, "lon"=>-76.730832),
"/US/US-MD/lonaconing-9169" => array("type"=>"town", "pop"=>1125, "name_ru"=>"Лонакоунинг", "name_en"=>"Lonaconing", "lang"=>"en", "lat"=>39.565234, "lon"=>-78.979871),
"/US/US-MD/londontowne-9170" => array("type"=>"cdp", "pop"=>6992, "name_ru"=>"Лондонтаун", "name_en"=>"Londontowne", "lang"=>"en", "lat"=>38.935750, "lon"=>-76.558016),
"/US/US-MD/long-meadow-9171" => array("type"=>"cdp", "pop"=>5594, "name_ru"=>"Лонг Медоу", "name_en"=>"Long Meadow", "lang"=>"en", "lat"=>39.686167, "lon"=>-77.712449),
"/US/US-MD/luke-9172" => array("type"=>"town", "pop"=>73, "name_ru"=>"Льюк", "name_en"=>"Luke", "lang"=>"en", "lat"=>39.475564, "lon"=>-79.060367),
"/US/US-MD/lutherville-timonium-9173" => array("type"=>"cdp", "pop"=>16442, "name_ru"=>"Лютервилл-Тимониум", "name_en"=>"Lutherville-Timonium", "lang"=>"en", "lat"=>39.440700, "lon"=>-76.612274),
"/US/US-MD/manchester-9174" => array("type"=>"town", "pop"=>3546, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>39.658486, "lon"=>-76.890748),
"/US/US-MD/marbury-9175" => array("type"=>"cdp", "pop"=>1244, "name_ru"=>"Марбери", "name_en"=>"Marbury", "lang"=>"en", "lat"=>38.556055, "lon"=>-77.158485),
"/US/US-MD/mardela-springs-9176" => array("type"=>"town", "pop"=>357, "name_ru"=>"Мардела Спрингс", "name_en"=>"Mardela Springs", "lang"=>"en", "lat"=>38.458179, "lon"=>-75.756568),
"/US/US-MD/marlow-heights-9177" => array("type"=>"cdp", "pop"=>5885, "name_ru"=>"Марлоу Хайтс", "name_en"=>"Marlow Heights", "lang"=>"en", "lat"=>38.824600, "lon"=>-76.949574),
"/US/US-MD/marlton-9178" => array("type"=>"cdp", "pop"=>5523, "name_ru"=>"Марлтон", "name_en"=>"Marlton", "lang"=>"en", "lat"=>38.758133, "lon"=>-76.789890),
"/US/US-MD/martin's-additions-9179" => array("type"=>"village", "pop"=>900, "name_ru"=>"Мартинс Аддишинс", "name_en"=>"Martin's Additions", "lang"=>"en", "lat"=>38.978439, "lon"=>-77.069689),
"/US/US-MD/marydel-9180" => array("type"=>"town", "pop"=>143, "name_ru"=>"Мэридел", "name_en"=>"Marydel", "lang"=>"en", "lat"=>39.113915, "lon"=>-75.748479),
"/US/US-MD/maryland-city-9181" => array("type"=>"cdp", "pop"=>6813, "name_ru"=>"Мэриленд Сити", "name_en"=>"Maryland City", "lang"=>"en", "lat"=>39.094057, "lon"=>-76.824351),
"/US/US-MD/mayo-9182" => array("type"=>"cdp", "pop"=>2537, "name_ru"=>"Майо", "name_en"=>"Mayo", "lang"=>"en", "lat"=>38.891683, "lon"=>-76.500401),
"/US/US-MD/mays-chapel-9183" => array("type"=>"cdp", "pop"=>10132, "name_ru"=>"Мейс Чэпэл", "name_en"=>"Mays Chapel", "lang"=>"en", "lat"=>39.432965, "lon"=>-76.649674),
"/US/US-MD/middle-river-9184" => array("type"=>"cdp", "pop"=>24616, "name_ru"=>"Миддл Ривер", "name_en"=>"Middle River", "lang"=>"en", "lat"=>39.334050, "lon"=>-76.434030),
"/US/US-MD/middletown-9185" => array("type"=>"town", "pop"=>2859, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>39.442242, "lon"=>-77.544192),
"/US/US-MD/midland-9186" => array("type"=>"town", "pop"=>440, "name_ru"=>"Мидленд", "name_en"=>"Midland", "lang"=>"en", "lat"=>39.589754, "lon"=>-78.949025),
"/US/US-MD/milford-mill-9187" => array("type"=>"cdp", "pop"=>22547, "name_ru"=>"Милфорд Милл", "name_en"=>"Milford Mill", "lang"=>"en", "lat"=>39.347650, "lon"=>-76.770276),
"/US/US-MD/millington-9188" => array("type"=>"town", "pop"=>445, "name_ru"=>"Миллингтон", "name_en"=>"Millington", "lang"=>"en", "lat"=>39.258201, "lon"=>-75.841656),
"/US/US-MD/mitchellville-9189" => array("type"=>"cdp", "pop"=>12593, "name_ru"=>"Митчелвилл", "name_en"=>"Mitchellville", "lang"=>"en", "lat"=>38.929167, "lon"=>-76.823262),
"/US/US-MD/montgomery-village-9190" => array("type"=>"cdp", "pop"=>32315, "name_ru"=>"Монтгомери Виллидж", "name_en"=>"Montgomery Village", "lang"=>"en", "lat"=>39.181157, "lon"=>-77.193613),
"/US/US-MD/morningside-9191" => array("type"=>"town", "pop"=>1305, "name_ru"=>"Морнингсайд", "name_en"=>"Morningside", "lang"=>"en", "lat"=>38.829400, "lon"=>-76.890709),
"/US/US-MD/mount-aetna-9192" => array("type"=>"cdp", "pop"=>3608, "name_ru"=>"Маунт Аэтна", "name_en"=>"Mount Aetna", "lang"=>"en", "lat"=>39.611050, "lon"=>-77.671098),
"/US/US-MD/mount-airy-9194" => array("type"=>"town", "pop"=>8761, "name_ru"=>"Маунт Эйри", "name_en"=>"Mount Airy", "lang"=>"en", "lat"=>39.374167, "lon"=>-77.159269),
"/US/US-MD/mount-rainier-9195" => array("type"=>"city", "pop"=>8322, "name_ru"=>"Маунт Рейнир", "name_en"=>"Mount Rainier", "lang"=>"en", "lat"=>38.941483, "lon"=>-76.963999),
"/US/US-MD/mountain-lake-park-9193" => array("type"=>"town", "pop"=>2106, "name_ru"=>"Маунтин Лейк Парк", "name_en"=>"Mountain Lake Park", "lang"=>"en", "lat"=>39.403762, "lon"=>-79.379401),
"/US/US-MD/myersville-9196" => array("type"=>"town", "pop"=>1511, "name_ru"=>"Майерсвилл", "name_en"=>"Myersville", "lang"=>"en", "lat"=>39.506153, "lon"=>-77.565080),
"/US/US-MD/naval-academy-9197" => array("type"=>"cdp", "pop"=>5420, "name_ru"=>"Нейвал Академи", "name_en"=>"Naval Academy", "lang"=>"en", "lat"=>38.985782, "lon"=>-76.489075),
"/US/US-MD/new-carrollton-9198" => array("type"=>"city", "pop"=>12460, "name_ru"=>"Нью Кэрролтон", "name_en"=>"New Carrollton", "lang"=>"en", "lat"=>38.966250, "lon"=>-76.876949),
"/US/US-MD/new-market-9199" => array("type"=>"town", "pop"=>465, "name_ru"=>"Нью Маркет", "name_en"=>"New Market", "lang"=>"en", "lat"=>39.384073, "lon"=>-77.274394),
"/US/US-MD/new-windsor-9200" => array("type"=>"town", "pop"=>1352, "name_ru"=>"Нью Виндзор", "name_en"=>"New Windsor", "lang"=>"en", "lat"=>39.543967, "lon"=>-77.101713),
"/US/US-MD/north-beach-9201" => array("type"=>"town", "pop"=>1869, "name_ru"=>"Норт Бич", "name_en"=>"North Beach", "lang"=>"en", "lat"=>38.708660, "lon"=>-76.534665),
"/US/US-MD/north-bethesda-9202" => array("type"=>"cdp", "pop"=>29656, "name_ru"=>"Норт Бетесда", "name_en"=>"North Bethesda", "lang"=>"en", "lat"=>39.044450, "lon"=>-77.119122),
"/US/US-MD/north-brentwood-9203" => array("type"=>"town", "pop"=>464, "name_ru"=>"Норт Брентвуд", "name_en"=>"North Brentwood", "lang"=>"en", "lat"=>38.945050, "lon"=>-76.951053),
"/US/US-MD/north-east-9204" => array("type"=>"town", "pop"=>2862, "name_ru"=>"Норт Ист", "name_en"=>"North East", "lang"=>"en", "lat"=>39.600243, "lon"=>-75.942471),
"/US/US-MD/north-kensington-9205" => array("type"=>"cdp", "pop"=>8607, "name_ru"=>"Норт Кенсингтон", "name_en"=>"North Kensington", "lang"=>"en", "lat"=>39.039700, "lon"=>-77.071671),
"/US/US-MD/north-laurel-9206" => array("type"=>"cdp", "pop"=>15008, "name_ru"=>"Норт Лорел", "name_en"=>"North Laurel", "lang"=>"en", "lat"=>39.138950, "lon"=>-76.870929),
"/US/US-MD/north-potomac-9207" => array("type"=>"cdp", "pop"=>18456, "name_ru"=>"Норт Потомак", "name_en"=>"North Potomac", "lang"=>"en", "lat"=>39.096432, "lon"=>-77.244955),
"/US/US-MD/oakland-9208" => array("type"=>"town", "pop"=>1844, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>39.413886, "lon"=>-79.407798),
"/US/US-MD/oakland-9209" => array("type"=>"cdp", "pop"=>2078, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>39.412350, "lon"=>-76.901930),
"/US/US-MD/ocean-city-9210" => array("type"=>"town", "pop"=>7042, "name_ru"=>"Оушен Сити", "name_en"=>"Ocean City", "lang"=>"en", "lat"=>38.379449, "lon"=>-75.036439),
"/US/US-MD/ocean-pines-9211" => array("type"=>"cdp", "pop"=>4251, "name_ru"=>"Оушен Пайнс", "name_en"=>"Ocean Pines", "lang"=>"en", "lat"=>38.383086, "lon"=>-75.149603),
"/US/US-MD/odenton-9212" => array("type"=>"cdp", "pop"=>12833, "name_ru"=>"Одентон", "name_en"=>"Odenton", "lang"=>"en", "lat"=>39.061234, "lon"=>-76.695762),
"/US/US-MD/olney-9213" => array("type"=>"cdp", "pop"=>23019, "name_ru"=>"Олни", "name_en"=>"Olney", "lang"=>"en", "lat"=>39.149173, "lon"=>-77.082231),
"/US/US-MD/overlea-9214" => array("type"=>"cdp", "pop"=>12137, "name_ru"=>"Оверли", "name_en"=>"Overlea", "lang"=>"en", "lat"=>39.363400, "lon"=>-76.517778),
"/US/US-MD/owings-mills-9215" => array("type"=>"cdp", "pop"=>9474, "name_ru"=>"Оуингс Миллс", "name_en"=>"Owings Mills", "lang"=>"en", "lat"=>39.411315, "lon"=>-76.793193),
"/US/US-MD/oxford-9216" => array("type"=>"town", "pop"=>706, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>38.689317, "lon"=>-76.170820),
"/US/US-MD/oxon-hill-glassmanor-9217" => array("type"=>"cdp", "pop"=>35794, "name_ru"=>"Оксон Хилл-Глассмэнор", "name_en"=>"Oxon Hill-Glassmanor", "lang"=>"en", "lat"=>38.796600, "lon"=>-76.975241),
"/US/US-MD/palmer-park-9218" => array("type"=>"cdp", "pop"=>7019, "name_ru"=>"Палмер Парк", "name_en"=>"Palmer Park", "lang"=>"en", "lat"=>38.918100, "lon"=>-76.872282),
"/US/US-MD/parkville-9219" => array("type"=>"cdp", "pop"=>31617, "name_ru"=>"Парквилл", "name_en"=>"Parkville", "lang"=>"en", "lat"=>39.384025, "lon"=>-76.551815),
"/US/US-MD/parole-9220" => array("type"=>"cdp", "pop"=>10054, "name_ru"=>"Парол", "name_en"=>"Parole", "lang"=>"en", "lat"=>38.984150, "lon"=>-76.553084),
"/US/US-MD/pasadena-9221" => array("type"=>"cdp", "pop"=>10012, "name_ru"=>"Пасадена", "name_en"=>"Pasadena", "lang"=>"en", "lat"=>39.112100, "lon"=>-76.552447),
"/US/US-MD/perry-hall-9222" => array("type"=>"cdp", "pop"=>22723, "name_ru"=>"Перри Хилл", "name_en"=>"Perry Hall", "lang"=>"en", "lat"=>39.406600, "lon"=>-76.478620),
"/US/US-MD/perryman-9223" => array("type"=>"cdp", "pop"=>2160, "name_ru"=>"Перримэн", "name_en"=>"Perryman", "lang"=>"en", "lat"=>39.469869, "lon"=>-76.205746),
"/US/US-MD/perryville-9224" => array("type"=>"town", "pop"=>3802, "name_ru"=>"Перривилл", "name_en"=>"Perryville", "lang"=>"en", "lat"=>39.570967, "lon"=>-76.068775),
"/US/US-MD/pikesville-9225" => array("type"=>"cdp", "pop"=>24815, "name_ru"=>"Пайксвилл", "name_en"=>"Pikesville", "lang"=>"en", "lat"=>39.391387, "lon"=>-76.700714),
"/US/US-MD/pittsville-9226" => array("type"=>"town", "pop"=>1182, "name_ru"=>"Питтсвилл", "name_en"=>"Pittsville", "lang"=>"en", "lat"=>38.393334, "lon"=>-75.413218),
"/US/US-MD/pleasant-hills-9227" => array("type"=>"cdp", "pop"=>2591, "name_ru"=>"Плезант Хилс", "name_en"=>"Pleasant Hills", "lang"=>"en", "lat"=>39.487050, "lon"=>-76.395234),
"/US/US-MD/pocomoke-city-9228" => array("type"=>"city", "pop"=>3879, "name_ru"=>"Покомоук Сити", "name_en"=>"Pocomoke City", "lang"=>"en", "lat"=>38.067106, "lon"=>-75.561944),
"/US/US-MD/poolesville-9229" => array("type"=>"town", "pop"=>5674, "name_ru"=>"Пулсвилл", "name_en"=>"Poolesville", "lang"=>"en", "lat"=>39.144750, "lon"=>-77.410014),
"/US/US-MD/port-deposit-9230" => array("type"=>"town", "pop"=>701, "name_ru"=>"Порт Депозит", "name_en"=>"Port Deposit", "lang"=>"en", "lat"=>39.603590, "lon"=>-76.115805),
"/US/US-MD/port-tobacco-village-9231" => array("type"=>"town", "pop"=>19, "name_ru"=>"Порт Тобакко Виллидж", "name_en"=>"Port Tobacco Village", "lang"=>"en", "lat"=>38.511379, "lon"=>-77.020404),
"/US/US-MD/potomac-9232" => array("type"=>"cdp", "pop"=>45634, "name_ru"=>"Потомак", "name_en"=>"Potomac", "lang"=>"en", "lat"=>39.025751, "lon"=>-77.197731),
"/US/US-MD/potomac-heights-9233" => array("type"=>"cdp", "pop"=>1524, "name_ru"=>"Потомак Хайтс", "name_en"=>"Potomac Heights", "lang"=>"en", "lat"=>38.607567, "lon"=>-77.122760),
"/US/US-MD/preston-9234" => array("type"=>"town", "pop"=>672, "name_ru"=>"Престон", "name_en"=>"Preston", "lang"=>"en", "lat"=>38.710768, "lon"=>-75.908639),
"/US/US-MD/prince-frederick-9235" => array("type"=>"cdp", "pop"=>1885, "name_ru"=>"Принс Фредерик", "name_en"=>"Prince Frederick", "lang"=>"en", "lat"=>38.548598, "lon"=>-76.589069),
"/US/US-MD/princess-anne-9236" => array("type"=>"town", "pop"=>3015, "name_ru"=>"Прнцесс Энн", "name_en"=>"Princess Anne", "lang"=>"en", "lat"=>38.204318, "lon"=>-75.695375),
"/US/US-MD/pumphrey-9237" => array("type"=>"cdp", "pop"=>5483, "name_ru"=>"Памфри", "name_en"=>"Pumphrey", "lang"=>"en", "lat"=>39.218000, "lon"=>-76.642924),
"/US/US-MD/queen-anne-9238" => array("type"=>"town", "pop"=>166, "name_ru"=>"Квин Энн", "name_en"=>"Queen Anne", "lang"=>"en", "lat"=>38.918870, "lon"=>-75.954036),
"/US/US-MD/queenstown-9239" => array("type"=>"town", "pop"=>635, "name_ru"=>"Квинстаун", "name_en"=>"Queenstown", "lang"=>"en", "lat"=>38.989912, "lon"=>-76.155911),
"/US/US-MD/randallstown-9240" => array("type"=>"cdp", "pop"=>26277, "name_ru"=>"Рэндаллстаун", "name_en"=>"Randallstown", "lang"=>"en", "lat"=>39.371700, "lon"=>-76.802171),
"/US/US-MD/redland-9241" => array("type"=>"cdp", "pop"=>16145, "name_ru"=>"Редленд", "name_en"=>"Redland", "lang"=>"en", "lat"=>39.131598, "lon"=>-77.150117),
"/US/US-MD/reisterstown-9242" => array("type"=>"cdp", "pop"=>19314, "name_ru"=>"Райстерстаун", "name_en"=>"Reisterstown", "lang"=>"en", "lat"=>39.456850, "lon"=>-76.814912),
"/US/US-MD/ridgely-9243" => array("type"=>"town", "pop"=>1517, "name_ru"=>"Риджли", "name_en"=>"Ridgely", "lang"=>"en", "lat"=>38.947526, "lon"=>-75.885302),
"/US/US-MD/rising-sun-9244" => array("type"=>"town", "pop"=>1809, "name_ru"=>"Райзинг Сайт", "name_en"=>"Rising Sun", "lang"=>"en", "lat"=>39.698297, "lon"=>-76.060579),
"/US/US-MD/riva-9245" => array("type"=>"cdp", "pop"=>3438, "name_ru"=>"Рива", "name_en"=>"Riva", "lang"=>"en", "lat"=>38.946038, "lon"=>-76.589623),
"/US/US-MD/riverdale-9246" => array("type"=>"town", "pop"=>5185, "name_ru"=>"Ривердейл", "name_en"=>"Riverdale", "lang"=>"en", "lat"=>38.964300, "lon"=>-76.927598),
"/US/US-MD/riviera-beach-9247" => array("type"=>"cdp", "pop"=>11376, "name_ru"=>"Ривьера Бич", "name_en"=>"Riviera Beach", "lang"=>"en", "lat"=>39.160650, "lon"=>-76.523349),
"/US/US-MD/rock-hall-9248" => array("type"=>"town", "pop"=>1486, "name_ru"=>"Рок Холл", "name_en"=>"Rock Hall", "lang"=>"en", "lat"=>39.140474, "lon"=>-76.240924),
"/US/US-MD/rockville-9249" => array("type"=>"city", "pop"=>60734, "name_ru"=>"Роквилл", "name_en"=>"Rockville", "lang"=>"en", "lat"=>39.081150, "lon"=>-77.153919),
"/US/US-MD/rosaryville-9250" => array("type"=>"cdp", "pop"=>8976, "name_ru"=>"Розаривилл", "name_en"=>"Rosaryville", "lang"=>"en", "lat"=>38.776100, "lon"=>-76.834491),
"/US/US-MD/rosedale-9251" => array("type"=>"cdp", "pop"=>18703, "name_ru"=>"Роуздейл", "name_en"=>"Rosedale", "lang"=>"en", "lat"=>39.327608, "lon"=>-76.508422),
"/US/US-MD/rosemont-9252" => array("type"=>"village", "pop"=>310, "name_ru"=>"Роузмонт", "name_en"=>"Rosemont", "lang"=>"en", "lat"=>39.334766, "lon"=>-77.621381),
"/US/US-MD/rossmoor-9253" => array("type"=>"cdp", "pop"=>6182, "name_ru"=>"Россмур", "name_en"=>"Rossmoor", "lang"=>"en", "lat"=>39.103717, "lon"=>-77.071282),
"/US/US-MD/rossville-9254" => array("type"=>"cdp", "pop"=>9492, "name_ru"=>"Россвилл", "name_en"=>"Rossville", "lang"=>"en", "lat"=>39.354800, "lon"=>-76.480842),
"/US/US-MD/salisbury-9257" => array("type"=>"city", "pop"=>28200, "name_ru"=>"Солсбери", "name_en"=>"Salisbury", "lang"=>"en", "lat"=>38.371916, "lon"=>-75.589027),
"/US/US-MD/savage-guilford-9258" => array("type"=>"cdp", "pop"=>9669, "name_ru"=>"Сэвадж-Гилфорд", "name_en"=>"Savage-Guilford", "lang"=>"en", "lat"=>39.149150, "lon"=>-76.827803),
"/US/US-MD/seat-pleasant-9259" => array("type"=>"city", "pop"=>4829, "name_ru"=>"Сит Плезант", "name_en"=>"Seat Pleasant", "lang"=>"en", "lat"=>38.895250, "lon"=>-76.902510),
"/US/US-MD/secretary-9260" => array("type"=>"town", "pop"=>487, "name_ru"=>"Секретари", "name_en"=>"Secretary", "lang"=>"en", "lat"=>38.607388, "lon"=>-75.947456),
"/US/US-MD/selby-on-the-bay-9261" => array("type"=>"cdp", "pop"=>3101, "name_ru"=>"Селби-он-зе-Бэй", "name_en"=>"Selby-on-the-Bay", "lang"=>"en", "lat"=>38.916000, "lon"=>-76.522718),
"/US/US-MD/severn-9262" => array("type"=>"cdp", "pop"=>24499, "name_ru"=>"Северн", "name_en"=>"Severn", "lang"=>"en", "lat"=>39.135250, "lon"=>-76.691576),
"/US/US-MD/severna-park-9263" => array("type"=>"cdp", "pop"=>25879, "name_ru"=>"Северна Парк", "name_en"=>"Severna Park", "lang"=>"en", "lat"=>39.081950, "lon"=>-76.572528),
"/US/US-MD/shady-side-9264" => array("type"=>"cdp", "pop"=>4107, "name_ru"=>"Шейди Сайд", "name_en"=>"Shady Side", "lang"=>"en", "lat"=>38.830134, "lon"=>-76.524511),
"/US/US-MD/sharpsburg-9265" => array("type"=>"town", "pop"=>661, "name_ru"=>"Шарпсбург", "name_en"=>"Sharpsburg", "lang"=>"en", "lat"=>39.457566, "lon"=>-77.749797),
"/US/US-MD/sharptown-9266" => array("type"=>"town", "pop"=>620, "name_ru"=>"Шарптаун", "name_en"=>"Sharptown", "lang"=>"en", "lat"=>38.539199, "lon"=>-75.720263),
"/US/US-MD/silver-spring-9267" => array("type"=>"cdp", "pop"=>76046, "name_ru"=>"Силвер Спринг", "name_en"=>"Silver Spring", "lang"=>"en", "lat"=>39.013750, "lon"=>-77.024158),
"/US/US-MD/smithsburg-9268" => array("type"=>"town", "pop"=>2908, "name_ru"=>"Смитсбург", "name_en"=>"Smithsburg", "lang"=>"en", "lat"=>39.658778, "lon"=>-77.577950),
"/US/US-MD/snow-hill-9269" => array("type"=>"town", "pop"=>2321, "name_ru"=>"Сноу Хилл", "name_en"=>"Snow Hill", "lang"=>"en", "lat"=>38.173212, "lon"=>-75.391622),
"/US/US-MD/somerset-9270" => array("type"=>"town", "pop"=>1169, "name_ru"=>"Сомерсет", "name_en"=>"Somerset", "lang"=>"en", "lat"=>38.965833, "lon"=>-77.095380),
"/US/US-MD/south-gate-9271" => array("type"=>"cdp", "pop"=>27564, "name_ru"=>"Саут Гейт", "name_en"=>"South Gate", "lang"=>"en", "lat"=>39.128896, "lon"=>-76.625983),
"/US/US-MD/south-kensington-9272" => array("type"=>"cdp", "pop"=>8777, "name_ru"=>"Саут Кенсингтон", "name_en"=>"South Kensington", "lang"=>"en", "lat"=>39.018750, "lon"=>-77.080303),
"/US/US-MD/south-laurel-9273" => array("type"=>"cdp", "pop"=>18591, "name_ru"=>"Саут Лорел", "name_en"=>"South Laurel", "lang"=>"en", "lat"=>39.069600, "lon"=>-76.850499),
"/US/US-MD/st.-charles-9255" => array("type"=>"cdp", "pop"=>28717, "name_ru"=>"Сент-Чарльз", "name_en"=>"St. Charles", "lang"=>"en", "lat"=>38.605706, "lon"=>-76.923226),
"/US/US-MD/st.-michaels-9256" => array("type"=>"town", "pop"=>1068, "name_ru"=>"Сент-Майклс", "name_en"=>"St. Michaels", "lang"=>"en", "lat"=>38.788405, "lon"=>-76.222227),
"/US/US-MD/stevensville-9274" => array("type"=>"cdp", "pop"=>1862, "name_ru"=>"Стивенсвилл", "name_en"=>"Stevensville", "lang"=>"en", "lat"=>38.993919, "lon"=>-76.308934),
"/US/US-MD/stevensville-south-9275" => array("type"=>"cdp", "pop"=>1751, "name_ru"=>"Стивенсвилл Саут", "name_en"=>"Stevensville South", "lang"=>"en", "lat"=>38.960179, "lon"=>-76.329345),
"/US/US-MD/sudlersville-9276" => array("type"=>"town", "pop"=>392, "name_ru"=>"Садлерсвилл", "name_en"=>"Sudlersville", "lang"=>"en", "lat"=>39.186238, "lon"=>-75.858480),
"/US/US-MD/suitland-silver-hill-9277" => array("type"=>"cdp", "pop"=>35111, "name_ru"=>"Сьютленд-Силвер Хилл", "name_en"=>"Suitland-Silver Hill", "lang"=>"en", "lat"=>38.845550, "lon"=>-76.922779),
"/US/US-MD/sykesville-9278" => array("type"=>"town", "pop"=>4419, "name_ru"=>"Сайксвилл", "name_en"=>"Sykesville", "lang"=>"en", "lat"=>39.370050, "lon"=>-76.975436),
"/US/US-MD/takoma-park-9279" => array("type"=>"city", "pop"=>17701, "name_ru"=>"Такома Парк", "name_en"=>"Takoma Park", "lang"=>"en", "lat"=>38.982100, "lon"=>-77.002789),
"/US/US-MD/taneytown-9280" => array("type"=>"city", "pop"=>5422, "name_ru"=>"Тенитаун", "name_en"=>"Taneytown", "lang"=>"en", "lat"=>39.655150, "lon"=>-77.168697),
"/US/US-MD/temple-hills-9281" => array("type"=>"cdp", "pop"=>6865, "name_ru"=>"Темпл Хилс", "name_en"=>"Temple Hills", "lang"=>"en", "lat"=>38.809808, "lon"=>-76.948758),
"/US/US-MD/templeville-9282" => array("type"=>"town", "pop"=>82, "name_ru"=>"Темплвилл", "name_en"=>"Templeville", "lang"=>"en", "lat"=>39.134759, "lon"=>-75.768038),
"/US/US-MD/thurmont-9283" => array("type"=>"town", "pop"=>6037, "name_ru"=>"Термонт", "name_en"=>"Thurmont", "lang"=>"en", "lat"=>39.623835, "lon"=>-77.409715),
"/US/US-MD/towson-9284" => array("type"=>"cdp", "pop"=>49445, "name_ru"=>"Таусон", "name_en"=>"Towson", "lang"=>"en", "lat"=>39.395033, "lon"=>-76.621258),
"/US/US-MD/trappe-9285" => array("type"=>"town", "pop"=>1143, "name_ru"=>"Трапп", "name_en"=>"Trappe", "lang"=>"en", "lat"=>38.659753, "lon"=>-76.058391),
"/US/US-MD/union-bridge-9286" => array("type"=>"town", "pop"=>1077, "name_ru"=>"Юнион Бридж", "name_en"=>"Union Bridge", "lang"=>"en", "lat"=>39.568200, "lon"=>-77.176185),
"/US/US-MD/university-park-9287" => array("type"=>"town", "pop"=>2282, "name_ru"=>"Юниверсити Парк", "name_en"=>"University Park", "lang"=>"en", "lat"=>38.971271, "lon"=>-76.943703),
"/US/US-MD/upper-marlboro-9288" => array("type"=>"town", "pop"=>657, "name_ru"=>"Аппер Марлборо", "name_en"=>"Upper Marlboro", "lang"=>"en", "lat"=>38.817552, "lon"=>-76.755370),
"/US/US-MD/vienna-9289" => array("type"=>"town", "pop"=>319, "name_ru"=>"Вьенна", "name_en"=>"Vienna", "lang"=>"en", "lat"=>38.485035, "lon"=>-75.827364),
"/US/US-MD/waldorf-9290" => array("type"=>"cdp", "pop"=>15058, "name_ru"=>"Уолдорф", "name_en"=>"Waldorf", "lang"=>"en", "lat"=>38.641900, "lon"=>-76.904529),
"/US/US-MD/walker-mill-9291" => array("type"=>"cdp", "pop"=>10920, "name_ru"=>"Уолкер Милл", "name_en"=>"Walker Mill", "lang"=>"en", "lat"=>38.875303, "lon"=>-76.888660),
"/US/US-MD/walkersville-9292" => array("type"=>"town", "pop"=>5606, "name_ru"=>"Уолкерсвилл", "name_en"=>"Walkersville", "lang"=>"en", "lat"=>39.484450, "lon"=>-77.353323),
"/US/US-MD/washington-grove-9293" => array("type"=>"town", "pop"=>566, "name_ru"=>"Вашингтон Гров", "name_en"=>"Washington Grove", "lang"=>"en", "lat"=>39.141150, "lon"=>-77.174934),
"/US/US-MD/west-laurel-9295" => array("type"=>"cdp", "pop"=>4151, "name_ru"=>"Вэст Лорел", "name_en"=>"West Laurel", "lang"=>"en", "lat"=>39.101033, "lon"=>-76.899965),
"/US/US-MD/west-ocean-city-9298" => array("type"=>"cdp", "pop"=>1928, "name_ru"=>"Вэст Ошен Сити", "name_en"=>"West Ocean City", "lang"=>"en", "lat"=>38.345317, "lon"=>-75.109650),
"/US/US-MD/westernport-9294" => array("type"=>"town", "pop"=>1945, "name_ru"=>"Вэстернпорт", "name_en"=>"Westernport", "lang"=>"en", "lat"=>39.488887, "lon"=>-79.042489),
"/US/US-MD/westminster-9296" => array("type"=>"city", "pop"=>17689, "name_ru"=>"Уэстминстер", "name_en"=>"Westminster", "lang"=>"en", "lat"=>39.578774, "lon"=>-77.007657),
"/US/US-MD/westminster-south-9297" => array("type"=>"cdp", "pop"=>4284, "name_ru"=>"Вэстминстер Саут", "name_en"=>"Westminster South", "lang"=>"en", "lat"=>39.548072, "lon"=>-76.982334),
"/US/US-MD/wheaton-glenmont-9299" => array("type"=>"cdp", "pop"=>53720, "name_ru"=>"Уитон-Гленмонт", "name_en"=>"Wheaton-Glenmont", "lang"=>"en", "lat"=>39.056967, "lon"=>-77.052456),
"/US/US-MD/white-marsh-9300" => array("type"=>"cdp", "pop"=>8183, "name_ru"=>"Уайт Марш", "name_en"=>"White Marsh", "lang"=>"en", "lat"=>39.383514, "lon"=>-76.458633),
"/US/US-MD/white-oak-9301" => array("type"=>"cdp", "pop"=>18671, "name_ru"=>"Уайт Оук", "name_en"=>"White Oak", "lang"=>"en", "lat"=>39.042000, "lon"=>-76.988575),
"/US/US-MD/white-plains-9302" => array("type"=>"cdp", "pop"=>3560, "name_ru"=>"Уайт Плейнс", "name_en"=>"White Plains", "lang"=>"en", "lat"=>38.594357, "lon"=>-76.976424),
"/US/US-MD/willards-9303" => array("type"=>"town", "pop"=>967, "name_ru"=>"Уиллардс", "name_en"=>"Willards", "lang"=>"en", "lat"=>38.391866, "lon"=>-75.349862),
"/US/US-MD/williamsport-9304" => array("type"=>"town", "pop"=>2278, "name_ru"=>"Уильямспорт", "name_en"=>"Williamsport", "lang"=>"en", "lat"=>39.596300, "lon"=>-77.818616),
"/US/US-MD/woodlawn-9305" => array("type"=>"cdp", "pop"=>32907, "name_ru"=>"Вудлон", "name_en"=>"Woodlawn", "lang"=>"en", "lat"=>39.306667, "lon"=>-76.747820),
"/US/US-MD/woodlawn-9306" => array("type"=>"cdp", "pop"=>5329, "name_ru"=>"Вудлон", "name_en"=>"Woodlawn", "lang"=>"en", "lat"=>38.949250, "lon"=>-76.898321),
"/US/US-MD/woodmore-9307" => array("type"=>"cdp", "pop"=>2874, "name_ru"=>"Вудмор", "name_en"=>"Woodmore", "lang"=>"en", "lat"=>38.934750, "lon"=>-76.775039),
"/US/US-MD/woodsboro-9308" => array("type"=>"town", "pop"=>914, "name_ru"=>"Вудсборо", "name_en"=>"Woodsboro", "lang"=>"en", "lat"=>39.532117, "lon"=>-77.310950),
"/US/US-ME" => array("type"=>"state", "pop"=>1316456, "name_ru"=>"Мэн", "name_en"=>"Maine", "lang"=>"en"),
"/US/US-ME/auburn-8874" => array("type"=>"city", "pop"=>23177, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>44.084200, "lon"=>-70.249540),
"/US/US-ME/augusta-8875" => array("type"=>"city", "pop"=>18282, "name_ru"=>"Огаста", "name_en"=>"Augusta", "lang"=>"en", "lat"=>44.330647, "lon"=>-69.729714),
"/US/US-ME/bangor-8876" => array("type"=>"city", "pop"=>31756, "name_ru"=>"Бангор", "name_en"=>"Bangor", "lang"=>"en", "lat"=>44.832203, "lon"=>-68.787550),
"/US/US-ME/bar-harbor-8877" => array("type"=>"cdp", "pop"=>5129, "name_ru"=>"Бар Харбор", "name_en"=>"Bar Harbor", "lang"=>"en", "lat"=>44.382417, "lon"=>-68.211869),
"/US/US-ME/bath-8878" => array("type"=>"city", "pop"=>8885, "name_ru"=>"Бат", "name_en"=>"Bath", "lang"=>"en", "lat"=>43.938075, "lon"=>-69.837416),
"/US/US-ME/belfast-8879" => array("type"=>"city", "pop"=>6721, "name_ru"=>"Бэлфаст", "name_en"=>"Belfast", "lang"=>"en", "lat"=>44.425803, "lon"=>-69.026902),
"/US/US-ME/biddeford-8880" => array("type"=>"city", "pop"=>21435, "name_ru"=>"Биддефорд", "name_en"=>"Biddeford", "lang"=>"en", "lat"=>43.465484, "lon"=>-70.443436),
"/US/US-ME/bingham-8881" => array("type"=>"cdp", "pop"=>971, "name_ru"=>"Бингхэм", "name_en"=>"Bingham", "lang"=>"en", "lat"=>45.058646, "lon"=>-69.878750),
"/US/US-ME/boothbay-harbor-8882" => array("type"=>"cdp", "pop"=>2251, "name_ru"=>"Бутбэй Харбор", "name_en"=>"Boothbay Harbor", "lang"=>"en", "lat"=>43.854580, "lon"=>-69.624853),
"/US/US-ME/brewer-8883" => array("type"=>"city", "pop"=>9035, "name_ru"=>"Брюэр", "name_en"=>"Brewer", "lang"=>"en", "lat"=>44.778803, "lon"=>-68.732014),
"/US/US-ME/bridgton-8884" => array("type"=>"cdp", "pop"=>5442, "name_ru"=>"Бриджтон", "name_en"=>"Bridgton", "lang"=>"en", "lat"=>44.067122, "lon"=>-70.721212),
"/US/US-ME/brunswick-8885" => array("type"=>"cdp", "pop"=>21720, "name_ru"=>"Брунсвик", "name_en"=>"Brunswick", "lang"=>"en", "lat"=>43.910657, "lon"=>-69.938445),
"/US/US-ME/brunswick-station-8886" => array("type"=>"cdp", "pop"=>1829, "name_ru"=>"Брансвик Стейшен", "name_en"=>"Brunswick Station", "lang"=>"en", "lat"=>43.896240, "lon"=>-69.931446),
"/US/US-ME/bucksport-8887" => array("type"=>"cdp", "pop"=>4893, "name_ru"=>"Бакспорт", "name_en"=>"Bucksport", "lang"=>"en", "lat"=>44.607913, "lon"=>-68.792491),
"/US/US-ME/calais-8888" => array("type"=>"city", "pop"=>3184, "name_ru"=>"Кале", "name_en"=>"Calais", "lang"=>"en", "lat"=>45.133728, "lon"=>-67.209105),
"/US/US-ME/camden-8889" => array("type"=>"cdp", "pop"=>5220, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>44.213667, "lon"=>-69.068777),
"/US/US-ME/cape-neddick-8890" => array("type"=>"cdp", "pop"=>2193, "name_ru"=>"Кейп Неддик", "name_en"=>"Cape Neddick", "lang"=>"en", "lat"=>43.170024, "lon"=>-70.619281),
"/US/US-ME/caribou-8891" => array("type"=>"city", "pop"=>8093, "name_ru"=>"Капибу", "name_en"=>"Caribou", "lang"=>"en", "lat"=>46.866791, "lon"=>-67.990562),
"/US/US-ME/chisholm-8892" => array("type"=>"cdp", "pop"=>1653, "name_ru"=>"Чизем", "name_en"=>"Chisholm", "lang"=>"en", "lat"=>44.492286, "lon"=>-70.193092),
"/US/US-ME/clinton-8893" => array("type"=>"cdp", "pop"=>3315, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>44.644884, "lon"=>-69.487277),
"/US/US-ME/cumberland-center-8894" => array("type"=>"cdp", "pop"=>1890, "name_ru"=>"Камберленд Сентер", "name_en"=>"Cumberland Center", "lang"=>"en", "lat"=>43.800756, "lon"=>-70.252697),
"/US/US-ME/damariscotta-newcastle-8895" => array("type"=>"cdp", "pop"=>1567, "name_ru"=>"Дамарискотта-Ньюкасл", "name_en"=>"Damariscotta-Newcastle", "lang"=>"en", "lat"=>44.029331, "lon"=>-69.530127),
"/US/US-ME/dexter-8896" => array("type"=>"cdp", "pop"=>3687, "name_ru"=>"Декстер", "name_en"=>"Dexter", "lang"=>"en", "lat"=>45.015525, "lon"=>-69.294604),
"/US/US-ME/dixfield-8897" => array("type"=>"cdp", "pop"=>2530, "name_ru"=>"Диксфилд", "name_en"=>"Dixfield", "lang"=>"en", "lat"=>44.538406, "lon"=>-70.452916),
"/US/US-ME/dover-foxcroft-8898" => array("type"=>"cdp", "pop"=>4220, "name_ru"=>"Довер-Фокскрофт", "name_en"=>"Dover-Foxcroft", "lang"=>"en", "lat"=>45.189279, "lon"=>-69.219420),
"/US/US-ME/east-millinocket-8899" => array("type"=>"cdp", "pop"=>1723, "name_ru"=>"Ист Миллинокет", "name_en"=>"East Millinocket", "lang"=>"en", "lat"=>45.625497, "lon"=>-68.575059),
"/US/US-ME/eastport-8900" => array("type"=>"city", "pop"=>1536, "name_ru"=>"Истпорт", "name_en"=>"Eastport", "lang"=>"en", "lat"=>44.918892, "lon"=>-67.012040),
"/US/US-ME/ellsworth-8901" => array("type"=>"city", "pop"=>7103, "name_ru"=>"Эллсворт", "name_en"=>"Ellsworth", "lang"=>"en", "lat"=>44.585720, "lon"=>-68.496141),
"/US/US-ME/fairfield-8902" => array("type"=>"cdp", "pop"=>6703, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>44.590371, "lon"=>-69.605662),
"/US/US-ME/falmouth-foreside-8903" => array("type"=>"cdp", "pop"=>1708, "name_ru"=>"Фолмаут Форсайд", "name_en"=>"Falmouth Foreside", "lang"=>"en", "lat"=>43.732274, "lon"=>-70.215043),
"/US/US-ME/farmingdale-8904" => array("type"=>"cdp", "pop"=>2884, "name_ru"=>"Фармингдейл", "name_en"=>"Farmingdale", "lang"=>"en", "lat"=>44.251423, "lon"=>-69.785455),
"/US/US-ME/farmington-8905" => array("type"=>"cdp", "pop"=>7545, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>44.668487, "lon"=>-70.145884),
"/US/US-ME/fort-fairfield-8906" => array("type"=>"cdp", "pop"=>3450, "name_ru"=>"Форт Фэрфилд", "name_en"=>"Fort Fairfield", "lang"=>"en", "lat"=>46.768038, "lon"=>-67.836475),
"/US/US-ME/fort-kent-8907" => array("type"=>"cdp", "pop"=>4182, "name_ru"=>"Форт Кент", "name_en"=>"Fort Kent", "lang"=>"en", "lat"=>47.263242, "lon"=>-68.577372),
"/US/US-ME/freeport-8908" => array("type"=>"cdp", "pop"=>8195, "name_ru"=>"Фрипорт", "name_en"=>"Freeport", "lang"=>"en", "lat"=>43.856863, "lon"=>-70.100249),
"/US/US-ME/fryeburg-8909" => array("type"=>"cdp", "pop"=>3335, "name_ru"=>"Фрайербург", "name_en"=>"Fryeburg", "lang"=>"en", "lat"=>44.018924, "lon"=>-70.974460),
"/US/US-ME/gardiner-8910" => array("type"=>"city", "pop"=>6100, "name_ru"=>"Гардинер", "name_en"=>"Gardiner", "lang"=>"en", "lat"=>44.190672, "lon"=>-69.789227),
"/US/US-ME/gorham-8911" => array("type"=>"cdp", "pop"=>15563, "name_ru"=>"Горам", "name_en"=>"Gorham", "lang"=>"en", "lat"=>43.680762, "lon"=>-70.446465),
"/US/US-ME/greenville-8912" => array("type"=>"cdp", "pop"=>1712, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>45.466699, "lon"=>-69.592804),
"/US/US-ME/guilford-8913" => array("type"=>"cdp", "pop"=>1438, "name_ru"=>"Гилфорд", "name_en"=>"Guilford", "lang"=>"en", "lat"=>45.171320, "lon"=>-69.391813),
"/US/US-ME/hallowell-8914" => array("type"=>"city", "pop"=>2437, "name_ru"=>"Халлоуэлл", "name_en"=>"Hallowell", "lang"=>"en", "lat"=>44.293968, "lon"=>-69.814664),
"/US/US-ME/hampden-8915" => array("type"=>"cdp", "pop"=>6914, "name_ru"=>"Хэмпдэн", "name_en"=>"Hampden", "lang"=>"en", "lat"=>44.749038, "lon"=>-68.836646),
"/US/US-ME/hartland-8916" => array("type"=>"cdp", "pop"=>1892, "name_ru"=>"Хартленд", "name_en"=>"Hartland", "lang"=>"en", "lat"=>44.882521, "lon"=>-69.455229),
"/US/US-ME/houlton-8917" => array("type"=>"cdp", "pop"=>6144, "name_ru"=>"Хьюлтон", "name_en"=>"Houlton", "lang"=>"en", "lat"=>46.117330, "lon"=>-67.834207),
"/US/US-ME/howland-8918" => array("type"=>"cdp", "pop"=>1326, "name_ru"=>"Хауленд", "name_en"=>"Howland", "lang"=>"en", "lat"=>45.250377, "lon"=>-68.667753),
"/US/US-ME/kennebunk-8919" => array("type"=>"cdp", "pop"=>11448, "name_ru"=>"Кеннебанк", "name_en"=>"Kennebunk", "lang"=>"en", "lat"=>43.386027, "lon"=>-70.545231),
"/US/US-ME/kennebunkport-8920" => array("type"=>"cdp", "pop"=>3984, "name_ru"=>"Кеннебанкпорт", "name_en"=>"Kennebunkport", "lang"=>"en", "lat"=>43.354442, "lon"=>-70.461509),
"/US/US-ME/kittery-8921" => array("type"=>"cdp", "pop"=>10427, "name_ru"=>"Киттери", "name_en"=>"Kittery", "lang"=>"en", "lat"=>43.092217, "lon"=>-70.738137),
"/US/US-ME/kittery-point-8922" => array("type"=>"cdp", "pop"=>1093, "name_ru"=>"Киттери Пойнт", "name_en"=>"Kittery Point", "lang"=>"en", "lat"=>43.084683, "lon"=>-70.699528),
"/US/US-ME/lewiston-8923" => array("type"=>"city", "pop"=>35131, "name_ru"=>"Льюистон", "name_en"=>"Lewiston", "lang"=>"en", "lat"=>44.089250, "lon"=>-70.172478),
"/US/US-ME/limestone-8924" => array("type"=>"cdp", "pop"=>2262, "name_ru"=>"Лаймстоун", "name_en"=>"Limestone", "lang"=>"en", "lat"=>46.911838, "lon"=>-67.830890),
"/US/US-ME/lincoln-8925" => array("type"=>"cdp", "pop"=>34628, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>45.363886, "lon"=>-68.500633),
"/US/US-ME/lisbon-falls-8926" => array("type"=>"cdp", "pop"=>4674, "name_ru"=>"Лисбон Фоллс", "name_en"=>"Lisbon Falls", "lang"=>"en", "lat"=>44.008150, "lon"=>-70.057972),
"/US/US-ME/little-falls-south-windham-8927" => array("type"=>"cdp", "pop"=>1715, "name_ru"=>"Литтл Фоллс-Саут Уиндхэм", "name_en"=>"Little Falls-South Windham", "lang"=>"en", "lat"=>43.735400, "lon"=>-70.426074),
"/US/US-ME/livermore-falls-8928" => array("type"=>"cdp", "pop"=>3134, "name_ru"=>"Ливермор Фоллс", "name_en"=>"Livermore Falls", "lang"=>"en", "lat"=>44.472512, "lon"=>-70.181076),
"/US/US-ME/loring-afb-8929" => array("type"=>"cdp", "pop"=>7829, "name_ru"=>"Военно-воздушная база Лоринг", "name_en"=>"Loring AFB", "lang"=>"en", "lat"=>46.940446, "lon"=>-67.894849),
"/US/US-ME/machias-8930" => array("type"=>"cdp", "pop"=>2134, "name_ru"=>"Макиас", "name_en"=>"Machias", "lang"=>"en", "lat"=>44.707972, "lon"=>-67.478123),
"/US/US-ME/madawaska-8931" => array("type"=>"cdp", "pop"=>4336, "name_ru"=>"Мадаваска", "name_en"=>"Madawaska", "lang"=>"en", "lat"=>47.343032, "lon"=>-68.333878),
"/US/US-ME/madison-8932" => array("type"=>"cdp", "pop"=>4570, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>44.799332, "lon"=>-69.869967),
"/US/US-ME/mars-hill-blaine-8933" => array("type"=>"cdp", "pop"=>1717, "name_ru"=>"Марс Хилл-Блейн", "name_en"=>"Mars Hill-Blaine", "lang"=>"en", "lat"=>46.508638, "lon"=>-67.870026),
"/US/US-ME/mechanic-falls-8934" => array("type"=>"cdp", "pop"=>3231, "name_ru"=>"Меканик Фоллс", "name_en"=>"Mechanic Falls", "lang"=>"en", "lat"=>44.111684, "lon"=>-70.394497),
"/US/US-ME/mexico-8935" => array("type"=>"cdp", "pop"=>2866, "name_ru"=>"Мексико", "name_en"=>"Mexico", "lang"=>"en", "lat"=>44.554496, "lon"=>-70.536006),
"/US/US-ME/milford-8936" => array("type"=>"cdp", "pop"=>2980, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>44.950785, "lon"=>-68.633968),
"/US/US-ME/millinocket-8937" => array("type"=>"cdp", "pop"=>4902, "name_ru"=>"Миллинокет", "name_en"=>"Millinocket", "lang"=>"en", "lat"=>45.659144, "lon"=>-68.698905),
"/US/US-ME/milo-8938" => array("type"=>"cdp", "pop"=>2320, "name_ru"=>"Майло", "name_en"=>"Milo", "lang"=>"en", "lat"=>45.241736, "lon"=>-68.976049),
"/US/US-ME/newport-8939" => array("type"=>"cdp", "pop"=>3104, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>44.835602, "lon"=>-69.266093),
"/US/US-ME/norridgewock-8940" => array("type"=>"cdp", "pop"=>3248, "name_ru"=>"Норриджвок", "name_en"=>"Norridgewock", "lang"=>"en", "lat"=>44.711951, "lon"=>-69.790027),
"/US/US-ME/north-berwick-8941" => array("type"=>"cdp", "pop"=>4832, "name_ru"=>"Норт Бервик", "name_en"=>"North Berwick", "lang"=>"en", "lat"=>43.303061, "lon"=>-70.734446),
"/US/US-ME/north-windham-8942" => array("type"=>"cdp", "pop"=>4077, "name_ru"=>"Норт Уиндхэм", "name_en"=>"North Windham", "lang"=>"en", "lat"=>43.821437, "lon"=>-70.429731),
"/US/US-ME/norway-8943" => array("type"=>"cdp", "pop"=>4762, "name_ru"=>"Норуэй", "name_en"=>"Norway", "lang"=>"en", "lat"=>44.213498, "lon"=>-70.550178),
"/US/US-ME/oakland-8944" => array("type"=>"cdp", "pop"=>6184, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>44.551794, "lon"=>-69.711463),
"/US/US-ME/old-orchard-beach-8945" => array("type"=>"cdp", "pop"=>9395, "name_ru"=>"Оулд Орчард Бич", "name_en"=>"Old Orchard Beach", "lang"=>"en", "lat"=>43.524284, "lon"=>-70.384218),
"/US/US-ME/old-town-8946" => array("type"=>"city", "pop"=>7709, "name_ru"=>"Оулд Таун", "name_en"=>"Old Town", "lang"=>"en", "lat"=>44.953440, "lon"=>-68.736587),
"/US/US-ME/orono-8947" => array("type"=>"cdp", "pop"=>9670, "name_ru"=>"Ороно", "name_en"=>"Orono", "lang"=>"en", "lat"=>44.883750, "lon"=>-68.682174),
"/US/US-ME/oxford-8948" => array("type"=>"cdp", "pop"=>56741, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>44.140352, "lon"=>-70.507917),
"/US/US-ME/pittsfield-8949" => array("type"=>"cdp", "pop"=>4217, "name_ru"=>"Питтсфилд", "name_en"=>"Pittsfield", "lang"=>"en", "lat"=>44.774783, "lon"=>-69.382104),
"/US/US-ME/portland-8950" => array("type"=>"city", "pop"=>62561, "name_ru"=>"Портленд", "name_en"=>"Portland", "lang"=>"en", "lat"=>43.667134, "lon"=>-70.207166),
"/US/US-ME/presque-isle-8951" => array("type"=>"city", "pop"=>9045, "name_ru"=>"Преск Айл", "name_en"=>"Presque Isle", "lang"=>"en", "lat"=>46.686040, "lon"=>-67.987296),
"/US/US-ME/randolph-8952" => array("type"=>"cdp", "pop"=>1855, "name_ru"=>"Рэндолф", "name_en"=>"Randolph", "lang"=>"en", "lat"=>44.236036, "lon"=>-69.752234),
"/US/US-ME/richmond-8953" => array("type"=>"cdp", "pop"=>3415, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>44.096341, "lon"=>-69.807915),
"/US/US-ME/rockland-8954" => array("type"=>"city", "pop"=>7436, "name_ru"=>"Рокленд", "name_en"=>"Rockland", "lang"=>"en", "lat"=>44.123358, "lon"=>-69.131344),
"/US/US-ME/rumford-8955" => array("type"=>"cdp", "pop"=>6310, "name_ru"=>"Рамфорд", "name_en"=>"Rumford", "lang"=>"en", "lat"=>44.549553, "lon"=>-70.564239),
"/US/US-ME/saco-8956" => array("type"=>"city", "pop"=>18125, "name_ru"=>"Сако", "name_en"=>"Saco", "lang"=>"en", "lat"=>43.537139, "lon"=>-70.454685),
"/US/US-ME/sanford-8957" => array("type"=>"cdp", "pop"=>21156, "name_ru"=>"Санфорд", "name_en"=>"Sanford", "lang"=>"en", "lat"=>43.441004, "lon"=>-70.781343),
"/US/US-ME/scarborough-8958" => array("type"=>"cdp", "pop"=>19054, "name_ru"=>"Скарборо", "name_en"=>"Scarborough", "lang"=>"en", "lat"=>43.597774, "lon"=>-70.331846),
"/US/US-ME/searsport-8959" => array("type"=>"cdp", "pop"=>2584, "name_ru"=>"Сиэрспорт", "name_en"=>"Searsport", "lang"=>"en", "lat"=>44.462244, "lon"=>-68.918640),
"/US/US-ME/skowhegan-8960" => array("type"=>"cdp", "pop"=>8679, "name_ru"=>"Скоухеган", "name_en"=>"Skowhegan", "lang"=>"en", "lat"=>44.775657, "lon"=>-69.713439),
"/US/US-ME/south-eliot-8961" => array("type"=>"cdp", "pop"=>3112, "name_ru"=>"Саут Элиот", "name_en"=>"South Eliot", "lang"=>"en", "lat"=>43.127394, "lon"=>-70.798200),
"/US/US-ME/south-paris-8962" => array("type"=>"cdp", "pop"=>2320, "name_ru"=>"Саут Пэрис", "name_en"=>"South Paris", "lang"=>"en", "lat"=>44.219060, "lon"=>-70.513720),
"/US/US-ME/south-portland-8963" => array("type"=>"city", "pop"=>23803, "name_ru"=>"Саут Портленд", "name_en"=>"South Portland", "lang"=>"en", "lat"=>43.632850, "lon"=>-70.285546),
"/US/US-ME/south-sanford-8964" => array("type"=>"cdp", "pop"=>3929, "name_ru"=>"Саут Санфорд", "name_en"=>"South Sanford", "lang"=>"en", "lat"=>43.401923, "lon"=>-70.715112),
"/US/US-ME/springvale-8965" => array("type"=>"cdp", "pop"=>3542, "name_ru"=>"Спрингвейл", "name_en"=>"Springvale", "lang"=>"en", "lat"=>43.466956, "lon"=>-70.803247),
"/US/US-ME/thomaston-8966" => array("type"=>"cdp", "pop"=>3656, "name_ru"=>"Томастон", "name_en"=>"Thomaston", "lang"=>"en", "lat"=>44.080016, "lon"=>-69.180154),
"/US/US-ME/topsham-8967" => array("type"=>"cdp", "pop"=>9827, "name_ru"=>"Топшэм", "name_en"=>"Topsham", "lang"=>"en", "lat"=>43.934774, "lon"=>-69.945606),
"/US/US-ME/van-buren-8968" => array("type"=>"cdp", "pop"=>2479, "name_ru"=>"Ван Бурен", "name_en"=>"Van Buren", "lang"=>"en", "lat"=>47.168239, "lon"=>-67.951115),
"/US/US-ME/waldoboro-8969" => array("type"=>"cdp", "pop"=>5004, "name_ru"=>"Уолдоборо", "name_en"=>"Waldoboro", "lang"=>"en", "lat"=>44.096099, "lon"=>-69.381520),
"/US/US-ME/waterville-8970" => array("type"=>"city", "pop"=>16016, "name_ru"=>"Уотервилл", "name_en"=>"Waterville", "lang"=>"en", "lat"=>44.545731, "lon"=>-69.661259),
"/US/US-ME/westbrook-8971" => array("type"=>"city", "pop"=>16534, "name_ru"=>"Вэстбрук", "name_en"=>"Westbrook", "lang"=>"en", "lat"=>43.693933, "lon"=>-70.354040),
"/US/US-ME/wilton-8972" => array("type"=>"cdp", "pop"=>4171, "name_ru"=>"Уилтон", "name_en"=>"Wilton", "lang"=>"en", "lat"=>44.595428, "lon"=>-70.238117),
"/US/US-ME/winslow-8973" => array("type"=>"cdp", "pop"=>7845, "name_ru"=>"Уинслоу", "name_en"=>"Winslow", "lang"=>"en", "lat"=>44.553345, "lon"=>-69.608957),
"/US/US-ME/winterport-8974" => array("type"=>"cdp", "pop"=>3527, "name_ru"=>"Уинтерпорт", "name_en"=>"Winterport", "lang"=>"en", "lat"=>44.652054, "lon"=>-68.856208),
"/US/US-ME/winthrop-8975" => array("type"=>"cdp", "pop"=>6429, "name_ru"=>"Уинтроп", "name_en"=>"Winthrop", "lang"=>"en", "lat"=>44.313566, "lon"=>-69.968177),
"/US/US-ME/wiscasset-8976" => array("type"=>"cdp", "pop"=>3752, "name_ru"=>"Вискассет", "name_en"=>"Wiscasset", "lang"=>"en", "lat"=>44.012362, "lon"=>-69.670377),
"/US/US-ME/woodland-8977" => array("type"=>"cdp", "pop"=>1447, "name_ru"=>"Вудленд", "name_en"=>"Woodland", "lang"=>"en", "lat"=>45.159426, "lon"=>-67.411748),
"/US/US-ME/yarmouth-8978" => array("type"=>"cdp", "pop"=>8097, "name_ru"=>"Ярмут", "name_en"=>"Yarmouth", "lang"=>"en", "lat"=>43.796183, "lon"=>-70.200541),
"/US/US-ME/york-harbor-8979" => array("type"=>"cdp", "pop"=>2555, "name_ru"=>"Йорк Харбор", "name_en"=>"York Harbor", "lang"=>"en", "lat"=>43.142485, "lon"=>-70.647610),
"/US/US-MI" => array("type"=>"state", "pop"=>10003422, "name_ru"=>"Мичиган", "name_en"=>"Michigan", "lang"=>"en"),
"/US/US-MI/addison-9540" => array("type"=>"village", "pop"=>586, "name_ru"=>"Эддиссон", "name_en"=>"Addison", "lang"=>"en", "lat"=>41.985892, "lon"=>-84.349122),
"/US/US-MI/adrian-9541" => array("type"=>"city", "pop"=>21391, "name_ru"=>"Эдриан", "name_en"=>"Adrian", "lang"=>"en", "lat"=>41.898263, "lon"=>-84.043470),
"/US/US-MI/ahmeek-9542" => array("type"=>"village", "pop"=>145, "name_ru"=>"Амик", "name_en"=>"Ahmeek", "lang"=>"en", "lat"=>47.297930, "lon"=>-88.397029),
"/US/US-MI/akron-9543" => array("type"=>"village", "pop"=>432, "name_ru"=>"Акрон", "name_en"=>"Akron", "lang"=>"en", "lat"=>43.566999, "lon"=>-83.514294),
"/US/US-MI/alanson-9544" => array("type"=>"village", "pop"=>783, "name_ru"=>"Алансон", "name_en"=>"Alanson", "lang"=>"en", "lat"=>45.441125, "lon"=>-84.787351),
"/US/US-MI/albion-9545" => array("type"=>"city", "pop"=>9103, "name_ru"=>"Албион", "name_en"=>"Albion", "lang"=>"en", "lat"=>42.246359, "lon"=>-84.756871),
"/US/US-MI/algonac-9546" => array("type"=>"city", "pop"=>4540, "name_ru"=>"Алгонак", "name_en"=>"Algonac", "lang"=>"en", "lat"=>42.622103, "lon"=>-82.534640),
"/US/US-MI/allegan-9547" => array("type"=>"city", "pop"=>4816, "name_ru"=>"Аллеган", "name_en"=>"Allegan", "lang"=>"en", "lat"=>42.528326, "lon"=>-85.845492),
"/US/US-MI/allen-9548" => array("type"=>"village", "pop"=>212, "name_ru"=>"Аллен", "name_en"=>"Allen", "lang"=>"en", "lat"=>41.957899, "lon"=>-84.767422),
"/US/US-MI/allen-park-9550" => array("type"=>"city", "pop"=>26014, "name_ru"=>"Аллен Парк", "name_en"=>"Allen Park", "lang"=>"en", "lat"=>42.259600, "lon"=>-83.210436),
"/US/US-MI/allendale-9549" => array("type"=>"cdp", "pop"=>6950, "name_ru"=>"Аллендейл", "name_en"=>"Allendale", "lang"=>"en", "lat"=>42.985740, "lon"=>-85.951750),
"/US/US-MI/alma-9551" => array("type"=>"city", "pop"=>9223, "name_ru"=>"Алма", "name_en"=>"Alma", "lang"=>"en", "lat"=>43.379406, "lon"=>-84.654849),
"/US/US-MI/almont-9552" => array("type"=>"village", "pop"=>2744, "name_ru"=>"Алмонт", "name_en"=>"Almont", "lang"=>"en", "lat"=>42.921368, "lon"=>-83.044454),
"/US/US-MI/alpena-9553" => array("type"=>"city", "pop"=>10465, "name_ru"=>"Алпена", "name_en"=>"Alpena", "lang"=>"en", "lat"=>45.072780, "lon"=>-83.437549),
"/US/US-MI/alpha-9554" => array("type"=>"village", "pop"=>175, "name_ru"=>"Альфа", "name_en"=>"Alpha", "lang"=>"en", "lat"=>46.043942, "lon"=>-88.378261),
"/US/US-MI/anchorville-9555" => array("type"=>"cdp", "pop"=>3202, "name_ru"=>"Анкорвилл", "name_en"=>"Anchorville", "lang"=>"en", "lat"=>42.692680, "lon"=>-82.696268),
"/US/US-MI/ann-arbor-9556" => array("type"=>"city", "pop"=>114386, "name_ru"=>"Энн Арбор", "name_en"=>"Ann Arbor", "lang"=>"en", "lat"=>42.275350, "lon"=>-83.730841),
"/US/US-MI/applegate-9557" => array("type"=>"village", "pop"=>268, "name_ru"=>"Эпплгейт", "name_en"=>"Applegate", "lang"=>"en", "lat"=>43.354902, "lon"=>-82.636471),
"/US/US-MI/argentine-9558" => array("type"=>"cdp", "pop"=>7042, "name_ru"=>"Аргентин", "name_en"=>"Argentine", "lang"=>"en", "lat"=>42.790181, "lon"=>-83.831974),
"/US/US-MI/armada-9559" => array("type"=>"village", "pop"=>1657, "name_ru"=>"Армада", "name_en"=>"Armada", "lang"=>"en", "lat"=>42.841002, "lon"=>-82.883464),
"/US/US-MI/ashley-9560" => array("type"=>"village", "pop"=>507, "name_ru"=>"Эшли", "name_en"=>"Ashley", "lang"=>"en", "lat"=>43.187725, "lon"=>-84.476341),
"/US/US-MI/athens-9561" => array("type"=>"village", "pop"=>1034, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>42.085847, "lon"=>-85.236177),
"/US/US-MI/au-gres-9564" => array("type"=>"city", "pop"=>926, "name_ru"=>"О Грс", "name_en"=>"Au Gres", "lang"=>"en", "lat"=>44.044209, "lon"=>-83.692805),
"/US/US-MI/au-sable-9566" => array("type"=>"cdp", "pop"=>547, "name_ru"=>"О Сейбл", "name_en"=>"Au Sable", "lang"=>"en", "lat"=>44.410754, "lon"=>-83.339756),
"/US/US-MI/auburn-9562" => array("type"=>"city", "pop"=>2031, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>43.601928, "lon"=>-84.077075),
"/US/US-MI/auburn-hills-9563" => array("type"=>"city", "pop"=>20931, "name_ru"=>"Обурн Хилс", "name_en"=>"Auburn Hills", "lang"=>"en", "lat"=>42.674500, "lon"=>-83.243530),
"/US/US-MI/augusta-9565" => array("type"=>"village", "pop"=>857, "name_ru"=>"Огаста", "name_en"=>"Augusta", "lang"=>"en", "lat"=>42.337550, "lon"=>-85.351374),
"/US/US-MI/bad-axe-9567" => array("type"=>"city", "pop"=>3069, "name_ru"=>"Бэд Экс", "name_en"=>"Bad Axe", "lang"=>"en", "lat"=>43.802807, "lon"=>-82.996890),
"/US/US-MI/baldwin-9568" => array("type"=>"village", "pop"=>1171, "name_ru"=>"Болдуин", "name_en"=>"Baldwin", "lang"=>"en", "lat"=>43.897576, "lon"=>-85.852362),
"/US/US-MI/bancroft-9569" => array("type"=>"village", "pop"=>587, "name_ru"=>"Бэнкрофт", "name_en"=>"Bancroft", "lang"=>"en", "lat"=>42.876814, "lon"=>-84.065519),
"/US/US-MI/bangor-9570" => array("type"=>"city", "pop"=>1840, "name_ru"=>"Бангор", "name_en"=>"Bangor", "lang"=>"en", "lat"=>42.311465, "lon"=>-86.114519),
"/US/US-MI/baraga-9571" => array("type"=>"village", "pop"=>1185, "name_ru"=>"Барага", "name_en"=>"Baraga", "lang"=>"en", "lat"=>46.777265, "lon"=>-88.493535),
"/US/US-MI/barnes-lake-millers-lake-9572" => array("type"=>"cdp", "pop"=>1304, "name_ru"=>"Барнс Лейк-Миллерс Лейк", "name_en"=>"Barnes Lake-Millers Lake", "lang"=>"en", "lat"=>43.178102, "lon"=>-83.310387),
"/US/US-MI/baroda-9573" => array("type"=>"village", "pop"=>868, "name_ru"=>"Барода", "name_en"=>"Baroda", "lang"=>"en", "lat"=>41.954572, "lon"=>-86.487629),
"/US/US-MI/barryton-9574" => array("type"=>"village", "pop"=>370, "name_ru"=>"Бэрритон", "name_en"=>"Barryton", "lang"=>"en", "lat"=>43.749581, "lon"=>-85.142652),
"/US/US-MI/barton-hills-9575" => array("type"=>"village", "pop"=>325, "name_ru"=>"Бартон Хилс", "name_en"=>"Barton Hills", "lang"=>"en", "lat"=>42.317700, "lon"=>-83.755396),
"/US/US-MI/battle-creek-9576" => array("type"=>"city", "pop"=>52053, "name_ru"=>"Бэттл Крик", "name_en"=>"Battle Creek", "lang"=>"en", "lat"=>42.298850, "lon"=>-85.229015),
"/US/US-MI/bay-city-9577" => array("type"=>"city", "pop"=>33874, "name_ru"=>"Бэй Сити", "name_en"=>"Bay City", "lang"=>"en", "lat"=>43.590050, "lon"=>-83.888649),
"/US/US-MI/beal-city-9578" => array("type"=>"cdp", "pop"=>345, "name_ru"=>"Бил Сити", "name_en"=>"Beal City", "lang"=>"en", "lat"=>43.669852, "lon"=>-84.909769),
"/US/US-MI/bear-lake-9579" => array("type"=>"village", "pop"=>288, "name_ru"=>"Бэр Лейк", "name_en"=>"Bear Lake", "lang"=>"en", "lat"=>44.419994, "lon"=>-86.145574),
"/US/US-MI/beaverton-9580" => array("type"=>"city", "pop"=>1060, "name_ru"=>"Бивертон", "name_en"=>"Beaverton", "lang"=>"en", "lat"=>43.881698, "lon"=>-84.487558),
"/US/US-MI/beecher-9581" => array("type"=>"cdp", "pop"=>14465, "name_ru"=>"Бичер", "name_en"=>"Beecher", "lang"=>"en", "lat"=>43.090800, "lon"=>-83.704879),
"/US/US-MI/beechwood-9582" => array("type"=>"cdp", "pop"=>2676, "name_ru"=>"Бичвуд", "name_en"=>"Beechwood", "lang"=>"en", "lat"=>42.797008, "lon"=>-86.125881),
"/US/US-MI/belding-9583" => array("type"=>"city", "pop"=>5691, "name_ru"=>"Белдинг", "name_en"=>"Belding", "lang"=>"en", "lat"=>43.096512, "lon"=>-85.232983),
"/US/US-MI/bellaire-9584" => array("type"=>"village", "pop"=>1123, "name_ru"=>"Беллэр", "name_en"=>"Bellaire", "lang"=>"en", "lat"=>44.975887, "lon"=>-85.206711),
"/US/US-MI/belleville-9585" => array("type"=>"city", "pop"=>3556, "name_ru"=>"Белелвилл", "name_en"=>"Belleville", "lang"=>"en", "lat"=>42.203400, "lon"=>-83.482295),
"/US/US-MI/bellevue-9586" => array("type"=>"village", "pop"=>1357, "name_ru"=>"Белвью", "name_en"=>"Bellevue", "lang"=>"en", "lat"=>42.444150, "lon"=>-85.018649),
"/US/US-MI/benton-harbor-9587" => array("type"=>"city", "pop"=>10814, "name_ru"=>"Бентон Харбор", "name_en"=>"Benton Harbor", "lang"=>"en", "lat"=>42.116400, "lon"=>-86.446166),
"/US/US-MI/benton-heights-9588" => array("type"=>"cdp", "pop"=>5465, "name_ru"=>"Бентон Хайтс", "name_en"=>"Benton Heights", "lang"=>"en", "lat"=>42.121100, "lon"=>-86.415009),
"/US/US-MI/benzonia-9589" => array("type"=>"village", "pop"=>473, "name_ru"=>"Бензония", "name_en"=>"Benzonia", "lang"=>"en", "lat"=>44.616520, "lon"=>-86.097927),
"/US/US-MI/berkley-9590" => array("type"=>"city", "pop"=>14787, "name_ru"=>"Беркли", "name_en"=>"Berkley", "lang"=>"en", "lat"=>42.497000, "lon"=>-83.185399),
"/US/US-MI/berrien-springs-9591" => array("type"=>"village", "pop"=>1913, "name_ru"=>"Беррин Спрингс", "name_en"=>"Berrien Springs", "lang"=>"en", "lat"=>41.947316, "lon"=>-86.340049),
"/US/US-MI/bessemer-9592" => array("type"=>"city", "pop"=>1846, "name_ru"=>"Бессемер", "name_en"=>"Bessemer", "lang"=>"en", "lat"=>46.477588, "lon"=>-90.049631),
"/US/US-MI/beulah-9593" => array("type"=>"village", "pop"=>385, "name_ru"=>"Бойла", "name_en"=>"Beulah", "lang"=>"en", "lat"=>44.629616, "lon"=>-86.095488),
"/US/US-MI/beverly-hills-9594" => array("type"=>"village", "pop"=>9865, "name_ru"=>"Беверли Хиллс", "name_en"=>"Beverly Hills", "lang"=>"en", "lat"=>42.520000, "lon"=>-83.241534),
"/US/US-MI/big-rapids-9595" => array("type"=>"city", "pop"=>10257, "name_ru"=>"Биг Рэпидс", "name_en"=>"Big Rapids", "lang"=>"en", "lat"=>43.703879, "lon"=>-85.486257),
"/US/US-MI/bingham-farms-9596" => array("type"=>"village", "pop"=>989, "name_ru"=>"Бингхэм Фармс", "name_en"=>"Bingham Farms", "lang"=>"en", "lat"=>42.514948, "lon"=>-83.278017),
"/US/US-MI/birch-run-9597" => array("type"=>"village", "pop"=>1640, "name_ru"=>"Берч Ран", "name_en"=>"Birch Run", "lang"=>"en", "lat"=>43.250056, "lon"=>-83.790099),
"/US/US-MI/birmingham-9598" => array("type"=>"city", "pop"=>18932, "name_ru"=>"Бирмингем", "name_en"=>"Birmingham", "lang"=>"en", "lat"=>42.544750, "lon"=>-83.216474),
"/US/US-MI/blissfield-9599" => array("type"=>"village", "pop"=>3171, "name_ru"=>"Блиссфилд", "name_en"=>"Blissfield", "lang"=>"en", "lat"=>41.832406, "lon"=>-83.864375),
"/US/US-MI/bloomfield-hills-9600" => array("type"=>"city", "pop"=>3772, "name_ru"=>"Блумфилд Хилс", "name_en"=>"Bloomfield Hills", "lang"=>"en", "lat"=>42.578850, "lon"=>-83.247381),
"/US/US-MI/bloomfield-township-9601" => array("type"=>"cdp", "pop"=>42137, "name_ru"=>"Блумфилд Тауншип", "name_en"=>"Bloomfield Township", "lang"=>"en", "lat"=>42.575567, "lon"=>-83.272411),
"/US/US-MI/bloomingdale-9602" => array("type"=>"village", "pop"=>494, "name_ru"=>"Блумингдэйл", "name_en"=>"Bloomingdale", "lang"=>"en", "lat"=>42.383442, "lon"=>-85.957755),
"/US/US-MI/boyne-city-9603" => array("type"=>"city", "pop"=>3142, "name_ru"=>"Бойн Сити", "name_en"=>"Boyne City", "lang"=>"en", "lat"=>45.213241, "lon"=>-85.013854),
"/US/US-MI/boyne-falls-9604" => array("type"=>"village", "pop"=>330, "name_ru"=>"Бойн Фолс", "name_en"=>"Boyne Falls", "lang"=>"en", "lat"=>45.166966, "lon"=>-84.913560),
"/US/US-MI/breckenridge-9605" => array("type"=>"village", "pop"=>1284, "name_ru"=>"Брекенридж", "name_en"=>"Breckenridge", "lang"=>"en", "lat"=>43.407713, "lon"=>-84.478425),
"/US/US-MI/breedsville-9606" => array("type"=>"village", "pop"=>219, "name_ru"=>"Бридсвилл", "name_en"=>"Breedsville", "lang"=>"en", "lat"=>42.347057, "lon"=>-86.070874),
"/US/US-MI/bridgeport-9607" => array("type"=>"cdp", "pop"=>8569, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>43.375850, "lon"=>-83.883242),
"/US/US-MI/bridgman-9608" => array("type"=>"city", "pop"=>2400, "name_ru"=>"Бриджмэн", "name_en"=>"Bridgman", "lang"=>"en", "lat"=>41.940710, "lon"=>-86.563567),
"/US/US-MI/brighton-9609" => array("type"=>"city", "pop"=>7204, "name_ru"=>"Брайтон", "name_en"=>"Brighton", "lang"=>"en", "lat"=>42.530113, "lon"=>-83.784806),
"/US/US-MI/britton-9610" => array("type"=>"village", "pop"=>650, "name_ru"=>"Бриттон", "name_en"=>"Britton", "lang"=>"en", "lat"=>41.986268, "lon"=>-83.831868),
"/US/US-MI/bronson-9611" => array("type"=>"city", "pop"=>2274, "name_ru"=>"Бронсон", "name_en"=>"Bronson", "lang"=>"en", "lat"=>41.873653, "lon"=>-85.191519),
"/US/US-MI/brooklyn-9612" => array("type"=>"village", "pop"=>1340, "name_ru"=>"Бруклин", "name_en"=>"Brooklyn", "lang"=>"en", "lat"=>42.105500, "lon"=>-84.248724),
"/US/US-MI/brown-city-9613" => array("type"=>"city", "pop"=>1252, "name_ru"=>"Браун Сити", "name_en"=>"Brown City", "lang"=>"en", "lat"=>43.211639, "lon"=>-82.988679),
"/US/US-MI/brownlee-park-9614" => array("type"=>"cdp", "pop"=>2536, "name_ru"=>"Браунли Парк", "name_en"=>"Brownlee Park", "lang"=>"en", "lat"=>42.324200, "lon"=>-85.133674),
"/US/US-MI/buchanan-9615" => array("type"=>"city", "pop"=>4354, "name_ru"=>"Бачанан", "name_en"=>"Buchanan", "lang"=>"en", "lat"=>41.827896, "lon"=>-86.366811),
"/US/US-MI/buckley-9616" => array("type"=>"village", "pop"=>550, "name_ru"=>"Бакли", "name_en"=>"Buckley", "lang"=>"en", "lat"=>44.503904, "lon"=>-85.671742),
"/US/US-MI/buena-vista-9617" => array("type"=>"cdp", "pop"=>8196, "name_ru"=>"Буена Виста", "name_en"=>"Buena Vista", "lang"=>"en", "lat"=>43.420250, "lon"=>-83.898749),
"/US/US-MI/burlington-9618" => array("type"=>"village", "pop"=>382, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>42.104768, "lon"=>-85.079074),
"/US/US-MI/burr-oak-9619" => array("type"=>"village", "pop"=>750, "name_ru"=>"Берр Оук", "name_en"=>"Burr Oak", "lang"=>"en", "lat"=>41.848209, "lon"=>-85.320902),
"/US/US-MI/burt-9620" => array("type"=>"cdp", "pop"=>490, "name_ru"=>"Берт", "name_en"=>"Burt", "lang"=>"en", "lat"=>43.236790, "lon"=>-83.906267),
"/US/US-MI/burton-9621" => array("type"=>"city", "pop"=>29845, "name_ru"=>"Бертон", "name_en"=>"Burton", "lang"=>"en", "lat"=>42.996700, "lon"=>-83.619591),
"/US/US-MI/byron-9622" => array("type"=>"village", "pop"=>562, "name_ru"=>"Байрон", "name_en"=>"Byron", "lang"=>"en", "lat"=>42.825493, "lon"=>-83.948894),
"/US/US-MI/cadillac-9623" => array("type"=>"city", "pop"=>10245, "name_ru"=>"Кадиллак", "name_en"=>"Cadillac", "lang"=>"en", "lat"=>44.248268, "lon"=>-85.416103),
"/US/US-MI/caledonia-9624" => array("type"=>"village", "pop"=>1354, "name_ru"=>"Каледония", "name_en"=>"Caledonia", "lang"=>"en", "lat"=>42.793861, "lon"=>-85.513242),
"/US/US-MI/calumet-9625" => array("type"=>"village", "pop"=>800, "name_ru"=>"Калумет", "name_en"=>"Calumet", "lang"=>"en", "lat"=>47.247436, "lon"=>-88.453224),
"/US/US-MI/camden-9626" => array("type"=>"village", "pop"=>524, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>41.756387, "lon"=>-84.757041),
"/US/US-MI/canton-9627" => array("type"=>"cdp", "pop"=>57047, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>42.308512, "lon"=>-83.482185),
"/US/US-MI/capac-9628" => array("type"=>"village", "pop"=>2063, "name_ru"=>"Капак", "name_en"=>"Capac", "lang"=>"en", "lat"=>43.013286, "lon"=>-82.929350),
"/US/US-MI/carleton-9629" => array("type"=>"village", "pop"=>2572, "name_ru"=>"Карлтон", "name_en"=>"Carleton", "lang"=>"en", "lat"=>42.057538, "lon"=>-83.389816),
"/US/US-MI/carney-9630" => array("type"=>"village", "pop"=>212, "name_ru"=>"Карни", "name_en"=>"Carney", "lang"=>"en", "lat"=>45.589060, "lon"=>-87.553984),
"/US/US-MI/caro-9631" => array("type"=>"village", "pop"=>4020, "name_ru"=>"Каро", "name_en"=>"Caro", "lang"=>"en", "lat"=>43.488012, "lon"=>-83.402060),
"/US/US-MI/carrollton-9632" => array("type"=>"cdp", "pop"=>5968, "name_ru"=>"Карролтон", "name_en"=>"Carrollton", "lang"=>"en", "lat"=>43.459450, "lon"=>-83.939557),
"/US/US-MI/carson-city-9633" => array("type"=>"city", "pop"=>1177, "name_ru"=>"Карсон Сити", "name_en"=>"Carson City", "lang"=>"en", "lat"=>43.178424, "lon"=>-84.846903),
"/US/US-MI/carsonville-9634" => array("type"=>"village", "pop"=>470, "name_ru"=>"Карсонвилл", "name_en"=>"Carsonville", "lang"=>"en", "lat"=>43.425694, "lon"=>-82.672612),
"/US/US-MI/caseville-9635" => array("type"=>"village", "pop"=>835, "name_ru"=>"Кейсвилл", "name_en"=>"Caseville", "lang"=>"en", "lat"=>43.942185, "lon"=>-83.274571),
"/US/US-MI/casnovia-9636" => array("type"=>"village", "pop"=>311, "name_ru"=>"Касновия", "name_en"=>"Casnovia", "lang"=>"en", "lat"=>43.234809, "lon"=>-85.791290),
"/US/US-MI/caspian-9637" => array("type"=>"city", "pop"=>882, "name_ru"=>"Каспиан", "name_en"=>"Caspian", "lang"=>"en", "lat"=>46.065189, "lon"=>-88.625747),
"/US/US-MI/cass-city-9638" => array("type"=>"village", "pop"=>2493, "name_ru"=>"Касс Сити", "name_en"=>"Cass City", "lang"=>"en", "lat"=>43.602469, "lon"=>-83.175170),
"/US/US-MI/cassopolis-9639" => array("type"=>"village", "pop"=>1920, "name_ru"=>"Кассополис", "name_en"=>"Cassopolis", "lang"=>"en", "lat"=>41.912463, "lon"=>-86.007880),
"/US/US-MI/cedar-springs-9640" => array("type"=>"city", "pop"=>3250, "name_ru"=>"Сидар Спрингс", "name_en"=>"Cedar Springs", "lang"=>"en", "lat"=>43.221419, "lon"=>-85.553153),
"/US/US-MI/cement-city-9641" => array("type"=>"village", "pop"=>419, "name_ru"=>"Семент Сити", "name_en"=>"Cement City", "lang"=>"en", "lat"=>42.068401, "lon"=>-84.327317),
"/US/US-MI/center-line-9642" => array("type"=>"city", "pop"=>8135, "name_ru"=>"Сентер Лайн", "name_en"=>"Center Line", "lang"=>"en", "lat"=>42.481151, "lon"=>-83.025199),
"/US/US-MI/central-lake-9643" => array("type"=>"village", "pop"=>965, "name_ru"=>"Сентрал Лейк", "name_en"=>"Central Lake", "lang"=>"en", "lat"=>45.069793, "lon"=>-85.263081),
"/US/US-MI/centreville-9644" => array("type"=>"village", "pop"=>1521, "name_ru"=>"Сентревилл", "name_en"=>"Centreville", "lang"=>"en", "lat"=>41.921831, "lon"=>-85.527418),
"/US/US-MI/charlevoix-9645" => array("type"=>"city", "pop"=>2656, "name_ru"=>"Чарлвуа", "name_en"=>"Charlevoix", "lang"=>"en", "lat"=>45.314147, "lon"=>-85.254073),
"/US/US-MI/charlotte-9646" => array("type"=>"city", "pop"=>8984, "name_ru"=>"Шарлотт", "name_en"=>"Charlotte", "lang"=>"en", "lat"=>42.564368, "lon"=>-84.832479),
"/US/US-MI/chatham-9647" => array("type"=>"village", "pop"=>211, "name_ru"=>"Чатем", "name_en"=>"Chatham", "lang"=>"en", "lat"=>46.344309, "lon"=>-86.930774),
"/US/US-MI/cheboygan-9648" => array("type"=>"city", "pop"=>4942, "name_ru"=>"Чебой ган", "name_en"=>"Cheboygan", "lang"=>"en", "lat"=>45.641902, "lon"=>-84.468807),
"/US/US-MI/chelsea-9649" => array("type"=>"village", "pop"=>5005, "name_ru"=>"Челси", "name_en"=>"Chelsea", "lang"=>"en", "lat"=>42.314435, "lon"=>-84.021965),
"/US/US-MI/chesaning-9650" => array("type"=>"village", "pop"=>2345, "name_ru"=>"Чесанинг", "name_en"=>"Chesaning", "lang"=>"en", "lat"=>43.184590, "lon"=>-84.119938),
"/US/US-MI/clare-9651" => array("type"=>"city", "pop"=>3093, "name_ru"=>"Клэр", "name_en"=>"Clare", "lang"=>"en", "lat"=>43.824199, "lon"=>-84.765017),
"/US/US-MI/clarkston-9652" => array("type"=>"village", "pop"=>1005, "name_ru"=>"Кларкстон", "name_en"=>"Clarkston", "lang"=>"en", "lat"=>42.735384, "lon"=>-83.421104),
"/US/US-MI/clarksville-9653" => array("type"=>"village", "pop"=>310, "name_ru"=>"Кларксвилл", "name_en"=>"Clarksville", "lang"=>"en", "lat"=>42.841502, "lon"=>-85.242033),
"/US/US-MI/clawson-9654" => array("type"=>"city", "pop"=>12157, "name_ru"=>"Клосон", "name_en"=>"Clawson", "lang"=>"en", "lat"=>42.536950, "lon"=>-83.151400),
"/US/US-MI/clayton-9655" => array("type"=>"village", "pop"=>303, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>41.864442, "lon"=>-84.235856),
"/US/US-MI/clifford-9656" => array("type"=>"village", "pop"=>310, "name_ru"=>"Клиффорд", "name_en"=>"Clifford", "lang"=>"en", "lat"=>43.314255, "lon"=>-83.178819),
"/US/US-MI/climax-9657" => array("type"=>"village", "pop"=>750, "name_ru"=>"Клаймакс", "name_en"=>"Climax", "lang"=>"en", "lat"=>42.239450, "lon"=>-85.337027),
"/US/US-MI/clinton-9658" => array("type"=>"village", "pop"=>2391, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>42.068868, "lon"=>-83.972408),
"/US/US-MI/clinton-9659" => array("type"=>"cdp", "pop"=>85866, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>42.586850, "lon"=>-82.920074),
"/US/US-MI/clio-9660" => array("type"=>"city", "pop"=>2489, "name_ru"=>"Клио", "name_en"=>"Clio", "lang"=>"en", "lat"=>43.177100, "lon"=>-83.736149),
"/US/US-MI/coldwater-9661" => array("type"=>"city", "pop"=>10514, "name_ru"=>"Коулдуотер", "name_en"=>"Coldwater", "lang"=>"en", "lat"=>41.940646, "lon"=>-85.004549),
"/US/US-MI/coleman-9662" => array("type"=>"city", "pop"=>1229, "name_ru"=>"Коулман", "name_en"=>"Coleman", "lang"=>"en", "lat"=>43.756749, "lon"=>-84.586692),
"/US/US-MI/coloma-9663" => array("type"=>"city", "pop"=>1459, "name_ru"=>"Колома", "name_en"=>"Coloma", "lang"=>"en", "lat"=>42.186579, "lon"=>-86.307679),
"/US/US-MI/colon-9664" => array("type"=>"village", "pop"=>1166, "name_ru"=>"Колон", "name_en"=>"Colon", "lang"=>"en", "lat"=>41.959023, "lon"=>-85.323467),
"/US/US-MI/columbiaville-9665" => array("type"=>"village", "pop"=>786, "name_ru"=>"Колумбиявилл", "name_en"=>"Columbiaville", "lang"=>"en", "lat"=>43.157976, "lon"=>-83.407801),
"/US/US-MI/comstock-northwest-9666" => array("type"=>"cdp", "pop"=>3402, "name_ru"=>"Комсток Нортвэст", "name_en"=>"Comstock Northwest", "lang"=>"en", "lat"=>42.322300, "lon"=>-85.517060),
"/US/US-MI/comstock-park-9667" => array("type"=>"cdp", "pop"=>6530, "name_ru"=>"Комсток Парк", "name_en"=>"Comstock Park", "lang"=>"en", "lat"=>43.044950, "lon"=>-85.672326),
"/US/US-MI/concord-9668" => array("type"=>"village", "pop"=>1072, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>42.175240, "lon"=>-84.643897),
"/US/US-MI/constantine-9669" => array("type"=>"village", "pop"=>2103, "name_ru"=>"Константин", "name_en"=>"Constantine", "lang"=>"en", "lat"=>41.838008, "lon"=>-85.665298),
"/US/US-MI/coopersville-9670" => array("type"=>"city", "pop"=>4138, "name_ru"=>"Куперсвилл", "name_en"=>"Coopersville", "lang"=>"en", "lat"=>43.065450, "lon"=>-85.933849),
"/US/US-MI/copemish-9671" => array("type"=>"village", "pop"=>214, "name_ru"=>"Коупмиш", "name_en"=>"Copemish", "lang"=>"en", "lat"=>44.480956, "lon"=>-85.923894),
"/US/US-MI/copper-city-9672" => array("type"=>"village", "pop"=>193, "name_ru"=>"Коппер Сити", "name_en"=>"Copper City", "lang"=>"en", "lat"=>47.284500, "lon"=>-88.386931),
"/US/US-MI/corunna-9673" => array("type"=>"city", "pop"=>3285, "name_ru"=>"Коруна", "name_en"=>"Corunna", "lang"=>"en", "lat"=>42.984028, "lon"=>-84.116731),
"/US/US-MI/croswell-9674" => array("type"=>"city", "pop"=>2457, "name_ru"=>"Кросуэлл", "name_en"=>"Croswell", "lang"=>"en", "lat"=>43.273723, "lon"=>-82.618686),
"/US/US-MI/crystal-falls-9675" => array("type"=>"city", "pop"=>1594, "name_ru"=>"Кристал Фоллс", "name_en"=>"Crystal Falls", "lang"=>"en", "lat"=>46.097305, "lon"=>-88.327727),
"/US/US-MI/custer-9676" => array("type"=>"village", "pop"=>315, "name_ru"=>"Кастер", "name_en"=>"Custer", "lang"=>"en", "lat"=>43.950256, "lon"=>-86.219441),
"/US/US-MI/cutlerville-9677" => array("type"=>"cdp", "pop"=>11228, "name_ru"=>"Катлервилл", "name_en"=>"Cutlerville", "lang"=>"en", "lat"=>42.841550, "lon"=>-85.661574),
"/US/US-MI/daggett-9678" => array("type"=>"village", "pop"=>253, "name_ru"=>"Даггетт", "name_en"=>"Daggett", "lang"=>"en", "lat"=>45.461950, "lon"=>-87.604466),
"/US/US-MI/dansville-9679" => array("type"=>"village", "pop"=>436, "name_ru"=>"Дэнсвилл", "name_en"=>"Dansville", "lang"=>"en", "lat"=>42.555350, "lon"=>-84.302499),
"/US/US-MI/davison-9680" => array("type"=>"city", "pop"=>5136, "name_ru"=>"Дэвисон", "name_en"=>"Davison", "lang"=>"en", "lat"=>43.032100, "lon"=>-83.518299),
"/US/US-MI/de-tour-village-9686" => array("type"=>"village", "pop"=>405, "name_ru"=>"Де Тор Виллидж", "name_en"=>"De Tour Village", "lang"=>"en", "lat"=>45.985957, "lon"=>-83.902284),
"/US/US-MI/de-witt-9689" => array("type"=>"city", "pop"=>3964, "name_ru"=>"Де Витт", "name_en"=>"De Witt", "lang"=>"en", "lat"=>42.835950, "lon"=>-84.575649),
"/US/US-MI/dearborn-9681" => array("type"=>"city", "pop"=>86477, "name_ru"=>"Дирборн", "name_en"=>"Dearborn", "lang"=>"en", "lat"=>42.311334, "lon"=>-83.213482),
"/US/US-MI/dearborn-heights-9682" => array("type"=>"city", "pop"=>51972, "name_ru"=>"Дирборн Хайтс", "name_en"=>"Dearborn Heights", "lang"=>"en", "lat"=>42.327201, "lon"=>-83.272975),
"/US/US-MI/decatur-9683" => array("type"=>"village", "pop"=>1775, "name_ru"=>"Декатур", "name_en"=>"Decatur", "lang"=>"en", "lat"=>42.109715, "lon"=>-85.973431),
"/US/US-MI/deckerville-9684" => array("type"=>"village", "pop"=>890, "name_ru"=>"Декервилл", "name_en"=>"Deckerville", "lang"=>"en", "lat"=>43.525632, "lon"=>-82.742118),
"/US/US-MI/deerfield-9685" => array("type"=>"village", "pop"=>930, "name_ru"=>"Дирфилд", "name_en"=>"Deerfield", "lang"=>"en", "lat"=>41.890142, "lon"=>-83.778592),
"/US/US-MI/detroit-9687" => array("type"=>"city", "pop"=>912062, "name_ru"=>"Детройт", "name_en"=>"Detroit", "lang"=>"en", "lat"=>42.383100, "lon"=>-83.102198),
"/US/US-MI/detroit-beach-9688" => array("type"=>"cdp", "pop"=>2113, "name_ru"=>"Детройт Бич", "name_en"=>"Detroit Beach", "lang"=>"en", "lat"=>41.931639, "lon"=>-83.328656),
"/US/US-MI/dexter-9690" => array("type"=>"village", "pop"=>3272, "name_ru"=>"Декстер", "name_en"=>"Dexter", "lang"=>"en", "lat"=>42.331465, "lon"=>-83.879810),
"/US/US-MI/dimondale-9691" => array("type"=>"village", "pop"=>1319, "name_ru"=>"Даймондейл", "name_en"=>"Dimondale", "lang"=>"en", "lat"=>42.648800, "lon"=>-84.649142),
"/US/US-MI/douglas-9692" => array("type"=>"village", "pop"=>1174, "name_ru"=>"Дуглас", "name_en"=>"Douglas", "lang"=>"en", "lat"=>42.641807, "lon"=>-86.210364),
"/US/US-MI/dowagiac-9693" => array("type"=>"city", "pop"=>5703, "name_ru"=>"Доваджиак", "name_en"=>"Dowagiac", "lang"=>"en", "lat"=>41.983831, "lon"=>-86.110638),
"/US/US-MI/dryden-9694" => array("type"=>"village", "pop"=>777, "name_ru"=>"Драйден", "name_en"=>"Dryden", "lang"=>"en", "lat"=>42.945403, "lon"=>-83.124527),
"/US/US-MI/dundee-9695" => array("type"=>"village", "pop"=>4253, "name_ru"=>"Данди", "name_en"=>"Dundee", "lang"=>"en", "lat"=>41.957005, "lon"=>-83.660428),
"/US/US-MI/durand-9696" => array("type"=>"city", "pop"=>3740, "name_ru"=>"Дуранд", "name_en"=>"Durand", "lang"=>"en", "lat"=>42.912973, "lon"=>-83.988283),
"/US/US-MI/eagle-9697" => array("type"=>"village", "pop"=>120, "name_ru"=>"Игл", "name_en"=>"Eagle", "lang"=>"en", "lat"=>42.809784, "lon"=>-84.790520),
"/US/US-MI/east-detroit-9698" => array("type"=>"city", "pop"=>35283, "name_ru"=>"Ист Детройт", "name_en"=>"East Detroit", "lang"=>"en", "lat"=>42.466050, "lon"=>-82.946499),
"/US/US-MI/east-grand-rapids-9699" => array("type"=>"city", "pop"=>10467, "name_ru"=>"Ист Гранд Рэпидс", "name_en"=>"East Grand Rapids", "lang"=>"en", "lat"=>42.946433, "lon"=>-85.607372),
"/US/US-MI/east-jordan-9700" => array("type"=>"city", "pop"=>2234, "name_ru"=>"Ист Джордан", "name_en"=>"East Jordan", "lang"=>"en", "lat"=>45.157840, "lon"=>-85.130250),
"/US/US-MI/east-lansing-9702" => array("type"=>"city", "pop"=>45857, "name_ru"=>"Ист Лэнсинг", "name_en"=>"East Lansing", "lang"=>"en", "lat"=>42.735950, "lon"=>-84.484320),
"/US/US-MI/east-tawas-9703" => array("type"=>"city", "pop"=>2723, "name_ru"=>"Ист Тавас", "name_en"=>"East Tawas", "lang"=>"en", "lat"=>44.286187, "lon"=>-83.484098),
"/US/US-MI/eastlake-9701" => array("type"=>"village", "pop"=>563, "name_ru"=>"Истлейк", "name_en"=>"Eastlake", "lang"=>"en", "lat"=>44.245496, "lon"=>-86.293858),
"/US/US-MI/eastwood-9704" => array("type"=>"cdp", "pop"=>6340, "name_ru"=>"Иствуд", "name_en"=>"Eastwood", "lang"=>"en", "lat"=>42.302000, "lon"=>-85.544099),
"/US/US-MI/eaton-rapids-9705" => array("type"=>"city", "pop"=>5246, "name_ru"=>"Итон Рэпидс", "name_en"=>"Eaton Rapids", "lang"=>"en", "lat"=>42.510037, "lon"=>-84.653889),
"/US/US-MI/eau-claire-9706" => array("type"=>"village", "pop"=>623, "name_ru"=>"О Клэр", "name_en"=>"Eau Claire", "lang"=>"en", "lat"=>41.983910, "lon"=>-86.303807),
"/US/US-MI/ecorse-9707" => array("type"=>"city", "pop"=>9858, "name_ru"=>"Экорс", "name_en"=>"Ecorse", "lang"=>"en", "lat"=>42.249750, "lon"=>-83.140468),
"/US/US-MI/edgemont-park-9708" => array("type"=>"cdp", "pop"=>2532, "name_ru"=>"Эджмонт Парк", "name_en"=>"Edgemont Park", "lang"=>"en", "lat"=>42.746750, "lon"=>-84.593652),
"/US/US-MI/edmore-9709" => array("type"=>"village", "pop"=>1222, "name_ru"=>"Эдмор", "name_en"=>"Edmore", "lang"=>"en", "lat"=>43.407584, "lon"=>-85.037666),
"/US/US-MI/edwardsburg-9710" => array("type"=>"village", "pop"=>1108, "name_ru"=>"Эдвардсбург", "name_en"=>"Edwardsburg", "lang"=>"en", "lat"=>41.796747, "lon"=>-86.083494),
"/US/US-MI/elberta-9711" => array("type"=>"village", "pop"=>437, "name_ru"=>"Элберта", "name_en"=>"Elberta", "lang"=>"en", "lat"=>44.621424, "lon"=>-86.229707),
"/US/US-MI/elk-rapids-9712" => array("type"=>"village", "pop"=>1672, "name_ru"=>"Элк Рэпидс", "name_en"=>"Elk Rapids", "lang"=>"en", "lat"=>44.894852, "lon"=>-85.404233),
"/US/US-MI/elkton-9713" => array("type"=>"village", "pop"=>758, "name_ru"=>"Элктон", "name_en"=>"Elkton", "lang"=>"en", "lat"=>43.818861, "lon"=>-83.180663),
"/US/US-MI/ellsworth-9714" => array("type"=>"village", "pop"=>455, "name_ru"=>"Эллсворт", "name_en"=>"Ellsworth", "lang"=>"en", "lat"=>45.166159, "lon"=>-85.244417),
"/US/US-MI/elsie-9715" => array("type"=>"village", "pop"=>979, "name_ru"=>"Элси", "name_en"=>"Elsie", "lang"=>"en", "lat"=>43.089001, "lon"=>-84.390573),
"/US/US-MI/emmett-9716" => array("type"=>"village", "pop"=>243, "name_ru"=>"Эммет", "name_en"=>"Emmett", "lang"=>"en", "lat"=>42.991117, "lon"=>-82.765915),
"/US/US-MI/empire-9717" => array("type"=>"village", "pop"=>367, "name_ru"=>"Эмпаер", "name_en"=>"Empire", "lang"=>"en", "lat"=>44.813168, "lon"=>-86.058938),
"/US/US-MI/escanaba-9718" => array("type"=>"city", "pop"=>12214, "name_ru"=>"Эсканаба", "name_en"=>"Escanaba", "lang"=>"en", "lat"=>45.746226, "lon"=>-87.080801),
"/US/US-MI/essexville-9719" => array("type"=>"city", "pop"=>3493, "name_ru"=>"Эссексвилл", "name_en"=>"Essexville", "lang"=>"en", "lat"=>43.611233, "lon"=>-83.843532),
"/US/US-MI/estral-beach-9720" => array("type"=>"village", "pop"=>464, "name_ru"=>"Эстрал Бич", "name_en"=>"Estral Beach", "lang"=>"en", "lat"=>41.985859, "lon"=>-83.237157),
"/US/US-MI/evart-9721" => array("type"=>"city", "pop"=>1667, "name_ru"=>"Эварт", "name_en"=>"Evart", "lang"=>"en", "lat"=>43.902403, "lon"=>-85.273752),
"/US/US-MI/fair-haven-9723" => array("type"=>"cdp", "pop"=>1505, "name_ru"=>"Фэр Хейвен", "name_en"=>"Fair Haven", "lang"=>"en", "lat"=>42.677400, "lon"=>-82.648124),
"/US/US-MI/fair-plain-9724" => array("type"=>"cdp", "pop"=>8051, "name_ru"=>"Фэр Плейн", "name_en"=>"Fair Plain", "lang"=>"en", "lat"=>42.081750, "lon"=>-86.453531),
"/US/US-MI/fairgrove-9722" => array("type"=>"village", "pop"=>591, "name_ru"=>"Фэргров", "name_en"=>"Fairgrove", "lang"=>"en", "lat"=>43.524021, "lon"=>-83.541654),
"/US/US-MI/farmington-9725" => array("type"=>"city", "pop"=>9839, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>42.463200, "lon"=>-83.377274),
"/US/US-MI/farmington-hills-9726" => array("type"=>"city", "pop"=>78522, "name_ru"=>"Фармингтон Хилс", "name_en"=>"Farmington Hills", "lang"=>"en", "lat"=>42.486086, "lon"=>-83.375583),
"/US/US-MI/farwell-9727" => array("type"=>"village", "pop"=>807, "name_ru"=>"Фаруэлл", "name_en"=>"Farwell", "lang"=>"en", "lat"=>43.835664, "lon"=>-84.867711),
"/US/US-MI/fennville-9728" => array("type"=>"city", "pop"=>1393, "name_ru"=>"Феннвилл", "name_en"=>"Fennville", "lang"=>"en", "lat"=>42.594795, "lon"=>-86.105157),
"/US/US-MI/fenton-9729" => array("type"=>"city", "pop"=>11678, "name_ru"=>"Фэнтон", "name_en"=>"Fenton", "lang"=>"en", "lat"=>42.800023, "lon"=>-83.714245),
"/US/US-MI/ferndale-9730" => array("type"=>"city", "pop"=>21112, "name_ru"=>"Ферндейл", "name_en"=>"Ferndale", "lang"=>"en", "lat"=>42.459300, "lon"=>-83.130978),
"/US/US-MI/ferrysburg-9731" => array("type"=>"city", "pop"=>3053, "name_ru"=>"Феррисбург", "name_en"=>"Ferrysburg", "lang"=>"en", "lat"=>43.086168, "lon"=>-86.219319),
"/US/US-MI/fife-lake-9732" => array("type"=>"village", "pop"=>461, "name_ru"=>"Файф Лейк", "name_en"=>"Fife Lake", "lang"=>"en", "lat"=>44.573690, "lon"=>-85.353299),
"/US/US-MI/flat-rock-9733" => array("type"=>"city", "pop"=>8771, "name_ru"=>"Флэт Рок", "name_en"=>"Flat Rock", "lang"=>"en", "lat"=>42.101800, "lon"=>-83.272928),
"/US/US-MI/flint-9734" => array("type"=>"city", "pop"=>112900, "name_ru"=>"Флинт", "name_en"=>"Flint", "lang"=>"en", "lat"=>43.022850, "lon"=>-83.692799),
"/US/US-MI/flushing-9735" => array("type"=>"city", "pop"=>7815, "name_ru"=>"Флашинг", "name_en"=>"Flushing", "lang"=>"en", "lat"=>43.064867, "lon"=>-83.841639),
"/US/US-MI/forest-hills-9736" => array("type"=>"cdp", "pop"=>16690, "name_ru"=>"Форест Хиллс", "name_en"=>"Forest Hills", "lang"=>"en", "lat"=>42.959518, "lon"=>-85.489589),
"/US/US-MI/forestville-9737" => array("type"=>"village", "pop"=>119, "name_ru"=>"Форествилл", "name_en"=>"Forestville", "lang"=>"en", "lat"=>43.660722, "lon"=>-82.613083),
"/US/US-MI/fountain-9738" => array("type"=>"village", "pop"=>174, "name_ru"=>"Фонтейн", "name_en"=>"Fountain", "lang"=>"en", "lat"=>44.047779, "lon"=>-86.179487),
"/US/US-MI/fowler-9739" => array("type"=>"village", "pop"=>1054, "name_ru"=>"Фоулер", "name_en"=>"Fowler", "lang"=>"en", "lat"=>43.004032, "lon"=>-84.740262),
"/US/US-MI/fowlerville-9740" => array("type"=>"village", "pop"=>3101, "name_ru"=>"Фоулервилл", "name_en"=>"Fowlerville", "lang"=>"en", "lat"=>42.660076, "lon"=>-84.072272),
"/US/US-MI/frankenmuth-9741" => array("type"=>"city", "pop"=>4652, "name_ru"=>"Франкенмут", "name_en"=>"Frankenmuth", "lang"=>"en", "lat"=>43.330463, "lon"=>-83.740220),
"/US/US-MI/frankfort-9742" => array("type"=>"city", "pop"=>1451, "name_ru"=>"Франкфорт", "name_en"=>"Frankfort", "lang"=>"en", "lat"=>44.636861, "lon"=>-86.233984),
"/US/US-MI/franklin-9743" => array("type"=>"village", "pop"=>2914, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>42.517350, "lon"=>-83.302051),
"/US/US-MI/fraser-9744" => array("type"=>"city", "pop"=>14902, "name_ru"=>"Фрейзер", "name_en"=>"Fraser", "lang"=>"en", "lat"=>42.539750, "lon"=>-82.948599),
"/US/US-MI/free-soil-9747" => array("type"=>"village", "pop"=>176, "name_ru"=>"Фри Сойл", "name_en"=>"Free Soil", "lang"=>"en", "lat"=>44.107697, "lon"=>-86.213785),
"/US/US-MI/freeland-9745" => array("type"=>"cdp", "pop"=>1421, "name_ru"=>"Фриленд", "name_en"=>"Freeland", "lang"=>"en", "lat"=>43.521056, "lon"=>-84.117734),
"/US/US-MI/freeport-9746" => array("type"=>"village", "pop"=>436, "name_ru"=>"Фрипорт", "name_en"=>"Freeport", "lang"=>"en", "lat"=>42.763533, "lon"=>-85.314455),
"/US/US-MI/fremont-9748" => array("type"=>"city", "pop"=>4195, "name_ru"=>"Фремонт", "name_en"=>"Fremont", "lang"=>"en", "lat"=>43.461490, "lon"=>-85.952173),
"/US/US-MI/fruitport-9749" => array("type"=>"village", "pop"=>1080, "name_ru"=>"Фрутпорт", "name_en"=>"Fruitport", "lang"=>"en", "lat"=>43.125696, "lon"=>-86.155867),
"/US/US-MI/gaastra-9750" => array("type"=>"city", "pop"=>300, "name_ru"=>"Гаастра", "name_en"=>"Gaastra", "lang"=>"en", "lat"=>46.058840, "lon"=>-88.605416),
"/US/US-MI/gagetown-9751" => array("type"=>"village", "pop"=>366, "name_ru"=>"Гейджтаун", "name_en"=>"Gagetown", "lang"=>"en", "lat"=>43.656768, "lon"=>-83.245198),
"/US/US-MI/gaines-9752" => array("type"=>"village", "pop"=>349, "name_ru"=>"Гейнс", "name_en"=>"Gaines", "lang"=>"en", "lat"=>42.872281, "lon"=>-83.911751),
"/US/US-MI/galesburg-9753" => array("type"=>"city", "pop"=>1933, "name_ru"=>"Гейлсбург", "name_en"=>"Galesburg", "lang"=>"en", "lat"=>42.290650, "lon"=>-85.417552),
"/US/US-MI/galien-9754" => array("type"=>"village", "pop"=>554, "name_ru"=>"Галиен", "name_en"=>"Galien", "lang"=>"en", "lat"=>41.801643, "lon"=>-86.499715),
"/US/US-MI/garden-9755" => array("type"=>"village", "pop"=>223, "name_ru"=>"Гарден", "name_en"=>"Garden", "lang"=>"en", "lat"=>45.775303, "lon"=>-86.552452),
"/US/US-MI/garden-city-9756" => array("type"=>"city", "pop"=>26743, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>42.323800, "lon"=>-83.342387),
"/US/US-MI/gaylord-9757" => array("type"=>"city", "pop"=>3593, "name_ru"=>"Гейлорд", "name_en"=>"Gaylord", "lang"=>"en", "lat"=>45.023538, "lon"=>-84.678075),
"/US/US-MI/gibraltar-9758" => array("type"=>"city", "pop"=>4826, "name_ru"=>"Гибралтар", "name_en"=>"Gibraltar", "lang"=>"en", "lat"=>42.094817, "lon"=>-83.202700),
"/US/US-MI/gladstone-9759" => array("type"=>"city", "pop"=>5067, "name_ru"=>"Глэдстоун", "name_en"=>"Gladstone", "lang"=>"en", "lat"=>45.851966, "lon"=>-87.024726),
"/US/US-MI/gladwin-9760" => array("type"=>"city", "pop"=>2906, "name_ru"=>"Глэдвин", "name_en"=>"Gladwin", "lang"=>"en", "lat"=>43.983942, "lon"=>-84.491557),
"/US/US-MI/gobles-9761" => array("type"=>"city", "pop"=>782, "name_ru"=>"Гоблс", "name_en"=>"Gobles", "lang"=>"en", "lat"=>42.361907, "lon"=>-85.876711),
"/US/US-MI/goodrich-9762" => array("type"=>"village", "pop"=>1619, "name_ru"=>"Гудрич", "name_en"=>"Goodrich", "lang"=>"en", "lat"=>42.916576, "lon"=>-83.508271),
"/US/US-MI/grand-beach-9763" => array("type"=>"village", "pop"=>240, "name_ru"=>"Гранд Бич", "name_en"=>"Grand Beach", "lang"=>"en", "lat"=>41.774650, "lon"=>-86.788302),
"/US/US-MI/grand-blanc-9764" => array("type"=>"city", "pop"=>7551, "name_ru"=>"Гранд Бланк", "name_en"=>"Grand Blanc", "lang"=>"en", "lat"=>42.925800, "lon"=>-83.618739),
"/US/US-MI/grand-haven-9765" => array("type"=>"city", "pop"=>10608, "name_ru"=>"Гранд Хейвен", "name_en"=>"Grand Haven", "lang"=>"en", "lat"=>43.055256, "lon"=>-86.224190),
"/US/US-MI/grand-ledge-9766" => array("type"=>"city", "pop"=>7669, "name_ru"=>"Гранд Ледж", "name_en"=>"Grand Ledge", "lang"=>"en", "lat"=>42.752560, "lon"=>-84.748572),
"/US/US-MI/grand-rapids-9767" => array("type"=>"city", "pop"=>193396, "name_ru"=>"Гранд Рэпидс", "name_en"=>"Grand Rapids", "lang"=>"en", "lat"=>42.961250, "lon"=>-85.655727),
"/US/US-MI/grandville-9768" => array("type"=>"city", "pop"=>16718, "name_ru"=>"Грандвилл", "name_en"=>"Grandville", "lang"=>"en", "lat"=>42.903800, "lon"=>-85.756148),
"/US/US-MI/grant-9769" => array("type"=>"city", "pop"=>856, "name_ru"=>"Грант", "name_en"=>"Grant", "lang"=>"en", "lat"=>43.334060, "lon"=>-85.810058),
"/US/US-MI/grass-lake-9770" => array("type"=>"village", "pop"=>1170, "name_ru"=>"Грасс Лейк", "name_en"=>"Grass Lake", "lang"=>"en", "lat"=>42.250867, "lon"=>-84.205952),
"/US/US-MI/grayling-9771" => array("type"=>"city", "pop"=>1835, "name_ru"=>"Грйлинг", "name_en"=>"Grayling", "lang"=>"en", "lat"=>44.657275, "lon"=>-84.709520),
"/US/US-MI/greater-galesburg-9772" => array("type"=>"cdp", "pop"=>1260, "name_ru"=>"Грейтер Гейлсбург", "name_en"=>"Greater Galesburg", "lang"=>"en", "lat"=>42.281147, "lon"=>-85.420038),
"/US/US-MI/greenville-9773" => array("type"=>"city", "pop"=>8154, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>43.178034, "lon"=>-85.253870),
"/US/US-MI/greilickville-9774" => array("type"=>"cdp", "pop"=>1165, "name_ru"=>"Грайликвилл", "name_en"=>"Greilickville", "lang"=>"en", "lat"=>44.799128, "lon"=>-85.645795),
"/US/US-MI/grosse-ile-9775" => array("type"=>"cdp", "pop"=>9884, "name_ru"=>"Гросс Айл", "name_en"=>"Grosse Ile", "lang"=>"en", "lat"=>42.128401, "lon"=>-83.149386),
"/US/US-MI/grosse-pointe-9776" => array("type"=>"city", "pop"=>4996, "name_ru"=>"Гросс Пуэнт", "name_en"=>"Grosse Pointe", "lang"=>"en", "lat"=>42.383250, "lon"=>-82.906053),
"/US/US-MI/grosse-pointe-farms-9777" => array("type"=>"city", "pop"=>8614, "name_ru"=>"Гросс Пуэнт Фармс", "name_en"=>"Grosse Pointe Farms", "lang"=>"en", "lat"=>42.383300, "lon"=>-82.858904),
"/US/US-MI/grosse-pointe-park-9778" => array("type"=>"city", "pop"=>10956, "name_ru"=>"Гросс Пуэнт Парк", "name_en"=>"Grosse Pointe Park", "lang"=>"en", "lat"=>42.373850, "lon"=>-82.923431),
"/US/US-MI/grosse-pointe-shores-9779" => array("type"=>"village", "pop"=>2509, "name_ru"=>"Гросс Пуэнт Шорс", "name_en"=>"Grosse Pointe Shores", "lang"=>"en", "lat"=>42.445233, "lon"=>-82.874135),
"/US/US-MI/grosse-pointe-woods-9780" => array("type"=>"city", "pop"=>15154, "name_ru"=>"Гросс Пуэнт Вудс", "name_en"=>"Grosse Pointe Woods", "lang"=>"en", "lat"=>42.435700, "lon"=>-82.898366),
"/US/US-MI/gwinn-9781" => array("type"=>"cdp", "pop"=>2370, "name_ru"=>"Гвинн", "name_en"=>"Gwinn", "lang"=>"en", "lat"=>46.290925, "lon"=>-87.439314),
"/US/US-MI/hamtramck-9782" => array("type"=>"city", "pop"=>20512, "name_ru"=>"Хэмтрэмк", "name_en"=>"Hamtramck", "lang"=>"en", "lat"=>42.395900, "lon"=>-83.055499),
"/US/US-MI/hancock-9783" => array("type"=>"city", "pop"=>4158, "name_ru"=>"Хэнкок", "name_en"=>"Hancock", "lang"=>"en", "lat"=>47.132266, "lon"=>-88.600701),
"/US/US-MI/hanover-9784" => array("type"=>"village", "pop"=>413, "name_ru"=>"Хановер", "name_en"=>"Hanover", "lang"=>"en", "lat"=>42.100565, "lon"=>-84.552688),
"/US/US-MI/harbor-beach-9785" => array("type"=>"city", "pop"=>1617, "name_ru"=>"Харбор Бич", "name_en"=>"Harbor Beach", "lang"=>"en", "lat"=>43.846593, "lon"=>-82.654888),
"/US/US-MI/harbor-springs-9786" => array("type"=>"city", "pop"=>1547, "name_ru"=>"Харбор Спрингс", "name_en"=>"Harbor Springs", "lang"=>"en", "lat"=>45.431987, "lon"=>-84.989638),
"/US/US-MI/harper-woods-9787" => array("type"=>"city", "pop"=>12584, "name_ru"=>"Харпер Вудс", "name_en"=>"Harper Woods", "lang"=>"en", "lat"=>42.438150, "lon"=>-82.928393),
"/US/US-MI/harrietta-9788" => array("type"=>"village", "pop"=>169, "name_ru"=>"Харриетта", "name_en"=>"Harrietta", "lang"=>"en", "lat"=>44.309502, "lon"=>-85.700490),
"/US/US-MI/harrison-9789" => array("type"=>"city", "pop"=>1993, "name_ru"=>"Харрисон", "name_en"=>"Harrison", "lang"=>"en", "lat"=>44.017368, "lon"=>-84.807265),
"/US/US-MI/harrison-9790" => array("type"=>"cdp", "pop"=>24685, "name_ru"=>"Харрисон", "name_en"=>"Harrison", "lang"=>"en", "lat"=>42.586251, "lon"=>-82.820197),
"/US/US-MI/harrisville-9791" => array("type"=>"city", "pop"=>491, "name_ru"=>"Харрисвилл", "name_en"=>"Harrisville", "lang"=>"en", "lat"=>44.656658, "lon"=>-83.294323),
"/US/US-MI/hart-9792" => array("type"=>"city", "pop"=>1915, "name_ru"=>"Харт", "name_en"=>"Hart", "lang"=>"en", "lat"=>43.697055, "lon"=>-86.363087),
"/US/US-MI/hartford-9793" => array("type"=>"city", "pop"=>2453, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>42.207519, "lon"=>-86.166162),
"/US/US-MI/harvey-9794" => array("type"=>"cdp", "pop"=>1377, "name_ru"=>"Харви", "name_en"=>"Harvey", "lang"=>"en", "lat"=>46.492829, "lon"=>-87.350898),
"/US/US-MI/haslett-9795" => array("type"=>"cdp", "pop"=>10230, "name_ru"=>"Хаслетт", "name_en"=>"Haslett", "lang"=>"en", "lat"=>42.752750, "lon"=>-84.405799),
"/US/US-MI/hastings-9796" => array("type"=>"city", "pop"=>6858, "name_ru"=>"Гастингс", "name_en"=>"Hastings", "lang"=>"en", "lat"=>42.650169, "lon"=>-85.288492),
"/US/US-MI/hazel-park-9797" => array("type"=>"city", "pop"=>17955, "name_ru"=>"Хейзел Парк", "name_en"=>"Hazel Park", "lang"=>"en", "lat"=>42.462784, "lon"=>-83.098103),
"/US/US-MI/hemlock-9798" => array("type"=>"cdp", "pop"=>1601, "name_ru"=>"Хемлок", "name_en"=>"Hemlock", "lang"=>"en", "lat"=>43.415750, "lon"=>-84.231593),
"/US/US-MI/hersey-9799" => array("type"=>"village", "pop"=>363, "name_ru"=>"Херси", "name_en"=>"Hersey", "lang"=>"en", "lat"=>43.849713, "lon"=>-85.442130),
"/US/US-MI/hesperia-9800" => array("type"=>"village", "pop"=>949, "name_ru"=>"Хесперия", "name_en"=>"Hesperia", "lang"=>"en", "lat"=>43.568218, "lon"=>-86.040242),
"/US/US-MI/highland-park-9801" => array("type"=>"city", "pop"=>14153, "name_ru"=>"Хайленд Парк", "name_en"=>"Highland Park", "lang"=>"en", "lat"=>42.404850, "lon"=>-83.098767),
"/US/US-MI/hillman-9802" => array("type"=>"village", "pop"=>673, "name_ru"=>"Хиллман", "name_en"=>"Hillman", "lang"=>"en", "lat"=>45.065917, "lon"=>-83.900606),
"/US/US-MI/hillsdale-9803" => array("type"=>"city", "pop"=>7762, "name_ru"=>"Хиллсдейл", "name_en"=>"Hillsdale", "lang"=>"en", "lat"=>41.925789, "lon"=>-84.635734),
"/US/US-MI/holland-9804" => array("type"=>"city", "pop"=>34076, "name_ru"=>"Холланд", "name_en"=>"Holland", "lang"=>"en", "lat"=>42.771706, "lon"=>-86.102231),
"/US/US-MI/holly-9805" => array("type"=>"village", "pop"=>6273, "name_ru"=>"Холли", "name_en"=>"Holly", "lang"=>"en", "lat"=>42.799016, "lon"=>-83.622581),
"/US/US-MI/holt-9806" => array("type"=>"cdp", "pop"=>11744, "name_ru"=>"Холт", "name_en"=>"Holt", "lang"=>"en", "lat"=>42.637250, "lon"=>-84.525599),
"/US/US-MI/homer-9807" => array("type"=>"village", "pop"=>1747, "name_ru"=>"Гомер", "name_en"=>"Homer", "lang"=>"en", "lat"=>42.145475, "lon"=>-84.809983),
"/US/US-MI/honor-9808" => array("type"=>"village", "pop"=>287, "name_ru"=>"Онор", "name_en"=>"Honor", "lang"=>"en", "lat"=>44.666779, "lon"=>-86.020194),
"/US/US-MI/hopkins-9809" => array("type"=>"village", "pop"=>552, "name_ru"=>"Хопкинс", "name_en"=>"Hopkins", "lang"=>"en", "lat"=>42.625333, "lon"=>-85.763843),
"/US/US-MI/houghton-9810" => array("type"=>"city", "pop"=>6878, "name_ru"=>"Хоутон", "name_en"=>"Houghton", "lang"=>"en", "lat"=>47.113581, "lon"=>-88.561793),
"/US/US-MI/houghton-lake-9811" => array("type"=>"cdp", "pop"=>3353, "name_ru"=>"Хоутон Лейк", "name_en"=>"Houghton Lake", "lang"=>"en", "lat"=>44.313436, "lon"=>-84.764140),
"/US/US-MI/howard-city-9812" => array("type"=>"village", "pop"=>1577, "name_ru"=>"Ховард Сити", "name_en"=>"Howard City", "lang"=>"en", "lat"=>43.395153, "lon"=>-85.466172),
"/US/US-MI/howell-9813" => array("type"=>"city", "pop"=>9694, "name_ru"=>"Хауэлл", "name_en"=>"Howell", "lang"=>"en", "lat"=>42.609354, "lon"=>-83.935717),
"/US/US-MI/hubbardston-9814" => array("type"=>"village", "pop"=>387, "name_ru"=>"Хаббардстон", "name_en"=>"Hubbardston", "lang"=>"en", "lat"=>43.094159, "lon"=>-84.841352),
"/US/US-MI/hubbell-9815" => array("type"=>"cdp", "pop"=>1174, "name_ru"=>"Хаббел", "name_en"=>"Hubbell", "lang"=>"en", "lat"=>47.179352, "lon"=>-88.436492),
"/US/US-MI/hudson-9816" => array("type"=>"city", "pop"=>2313, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>41.856299, "lon"=>-84.345581),
"/US/US-MI/hudsonville-9817" => array("type"=>"city", "pop"=>6997, "name_ru"=>"Хадсонвилл", "name_en"=>"Hudsonville", "lang"=>"en", "lat"=>42.865000, "lon"=>-85.863074),
"/US/US-MI/huntington-woods-9818" => array("type"=>"city", "pop"=>5811, "name_ru"=>"Хантингтон Вудс", "name_en"=>"Huntington Woods", "lang"=>"en", "lat"=>42.481600, "lon"=>-83.168199),
"/US/US-MI/imlay-city-9819" => array("type"=>"city", "pop"=>3687, "name_ru"=>"Имлей Сити", "name_en"=>"Imlay City", "lang"=>"en", "lat"=>43.016374, "lon"=>-83.077974),
"/US/US-MI/inkster-9820" => array("type"=>"city", "pop"=>26621, "name_ru"=>"Инкстер", "name_en"=>"Inkster", "lang"=>"en", "lat"=>42.293350, "lon"=>-83.316522),
"/US/US-MI/ionia-9821" => array("type"=>"city", "pop"=>12589, "name_ru"=>"Иония", "name_en"=>"Ionia", "lang"=>"en", "lat"=>42.985102, "lon"=>-85.057966),
"/US/US-MI/iron-mountain-9822" => array("type"=>"city", "pop"=>7763, "name_ru"=>"Айрон Маунтин", "name_en"=>"Iron Mountain", "lang"=>"en", "lat"=>45.827651, "lon"=>-88.061473),
"/US/US-MI/iron-river-9823" => array("type"=>"city", "pop"=>3003, "name_ru"=>"Айрон Ривер", "name_en"=>"Iron River", "lang"=>"en", "lat"=>46.098837, "lon"=>-88.638203),
"/US/US-MI/ironwood-9824" => array("type"=>"city", "pop"=>5368, "name_ru"=>"Айронвуд", "name_en"=>"Ironwood", "lang"=>"en", "lat"=>46.454459, "lon"=>-90.152560),
"/US/US-MI/ishpeming-9825" => array("type"=>"city", "pop"=>6480, "name_ru"=>"Ищпеминг", "name_en"=>"Ishpeming", "lang"=>"en", "lat"=>46.487498, "lon"=>-87.663632),
"/US/US-MI/ithaca-9826" => array("type"=>"city", "pop"=>3017, "name_ru"=>"Итака", "name_en"=>"Ithaca", "lang"=>"en", "lat"=>43.292290, "lon"=>-84.604765),
"/US/US-MI/jackson-9827" => array("type"=>"city", "pop"=>33518, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>42.243200, "lon"=>-84.404774),
"/US/US-MI/jenison-9828" => array("type"=>"cdp", "pop"=>17882, "name_ru"=>"Дженисон", "name_en"=>"Jenison", "lang"=>"en", "lat"=>42.908679, "lon"=>-85.829827),
"/US/US-MI/jonesville-9829" => array("type"=>"village", "pop"=>2228, "name_ru"=>"Джонсвилл", "name_en"=>"Jonesville", "lang"=>"en", "lat"=>41.981040, "lon"=>-84.666470),
"/US/US-MI/k.-i.-sawyer-afb-9840" => array("type"=>"cdp", "pop"=>6577, "name_ru"=>"Военно-воздушная база К.И. Сойер", "name_en"=>"K. I. Sawyer AFB", "lang"=>"en", "lat"=>46.344678, "lon"=>-87.382495),
"/US/US-MI/kalamazoo-9830" => array("type"=>"city", "pop"=>72179, "name_ru"=>"Каламазу", "name_en"=>"Kalamazoo", "lang"=>"en", "lat"=>42.274700, "lon"=>-85.588287),
"/US/US-MI/kaleva-9831" => array("type"=>"village", "pop"=>464, "name_ru"=>"Калева", "name_en"=>"Kaleva", "lang"=>"en", "lat"=>44.372327, "lon"=>-86.013029),
"/US/US-MI/kalkaska-9832" => array("type"=>"village", "pop"=>2172, "name_ru"=>"Калкаска", "name_en"=>"Kalkaska", "lang"=>"en", "lat"=>44.734563, "lon"=>-85.179891),
"/US/US-MI/keego-harbor-9833" => array("type"=>"city", "pop"=>2838, "name_ru"=>"Киго Харбор", "name_en"=>"Keego Harbor", "lang"=>"en", "lat"=>42.606937, "lon"=>-83.344200),
"/US/US-MI/kent-city-9834" => array("type"=>"village", "pop"=>1088, "name_ru"=>"Кент Сити", "name_en"=>"Kent City", "lang"=>"en", "lat"=>43.219068, "lon"=>-85.752057),
"/US/US-MI/kentwood-9835" => array("type"=>"city", "pop"=>47419, "name_ru"=>"Кентвуд", "name_en"=>"Kentwood", "lang"=>"en", "lat"=>42.882377, "lon"=>-85.592193),
"/US/US-MI/kinde-9836" => array("type"=>"village", "pop"=>472, "name_ru"=>"Кинд", "name_en"=>"Kinde", "lang"=>"en", "lat"=>43.939852, "lon"=>-82.994877),
"/US/US-MI/kingsford-9837" => array("type"=>"city", "pop"=>5297, "name_ru"=>"Кингсфорд", "name_en"=>"Kingsford", "lang"=>"en", "lat"=>45.807976, "lon"=>-88.099035),
"/US/US-MI/kingsley-9838" => array("type"=>"village", "pop"=>1499, "name_ru"=>"Кингсли", "name_en"=>"Kingsley", "lang"=>"en", "lat"=>44.584291, "lon"=>-85.535372),
"/US/US-MI/kingston-9839" => array("type"=>"village", "pop"=>423, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>43.413568, "lon"=>-83.186667),
"/US/US-MI/l'anse-9853" => array("type"=>"village", "pop"=>1882, "name_ru"=>"Л'Анс", "name_en"=>"L'Anse", "lang"=>"en", "lat"=>46.753574, "lon"=>-88.449376),
"/US/US-MI/laingsburg-9841" => array("type"=>"city", "pop"=>1248, "name_ru"=>"Лейнгсбург", "name_en"=>"Laingsburg", "lang"=>"en", "lat"=>42.889951, "lon"=>-84.348843),
"/US/US-MI/lake-angelus-9842" => array("type"=>"city", "pop"=>313, "name_ru"=>"Лейк Ангелус", "name_en"=>"Lake Angelus", "lang"=>"en", "lat"=>42.691900, "lon"=>-83.325491),
"/US/US-MI/lake-ann-9843" => array("type"=>"village", "pop"=>266, "name_ru"=>"Лейк Энн", "name_en"=>"Lake Ann", "lang"=>"en", "lat"=>44.725741, "lon"=>-85.844474),
"/US/US-MI/lake-city-9844" => array("type"=>"city", "pop"=>919, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>44.330286, "lon"=>-85.208499),
"/US/US-MI/lake-fenton-9845" => array("type"=>"cdp", "pop"=>4091, "name_ru"=>"Лейк Фентон", "name_en"=>"Lake Fenton", "lang"=>"en", "lat"=>42.846096, "lon"=>-83.707908),
"/US/US-MI/lake-linden-9846" => array("type"=>"village", "pop"=>1050, "name_ru"=>"Лейк Линден", "name_en"=>"Lake Linden", "lang"=>"en", "lat"=>47.195121, "lon"=>-88.403458),
"/US/US-MI/lake-michigan-beach-9847" => array("type"=>"cdp", "pop"=>1694, "name_ru"=>"Лейк Мичиган Бич", "name_en"=>"Lake Michigan Beach", "lang"=>"en", "lat"=>42.214325, "lon"=>-86.382433),
"/US/US-MI/lake-odessa-9848" => array("type"=>"village", "pop"=>2215, "name_ru"=>"Лейк Одесса", "name_en"=>"Lake Odessa", "lang"=>"en", "lat"=>42.782789, "lon"=>-85.137411),
"/US/US-MI/lake-orion-9849" => array("type"=>"village", "pop"=>2713, "name_ru"=>"Лейк Орион", "name_en"=>"Lake Orion", "lang"=>"en", "lat"=>42.783661, "lon"=>-83.244258),
"/US/US-MI/lakeview-9850" => array("type"=>"village", "pop"=>1091, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>43.446490, "lon"=>-85.275460),
"/US/US-MI/lakewood-club-9851" => array("type"=>"village", "pop"=>1353, "name_ru"=>"Лейквуд Клаб", "name_en"=>"Lakewood Club", "lang"=>"en", "lat"=>43.376598, "lon"=>-86.255091),
"/US/US-MI/lambertville-9852" => array("type"=>"cdp", "pop"=>7860, "name_ru"=>"Ламбертвилл", "name_en"=>"Lambertville", "lang"=>"en", "lat"=>41.747297, "lon"=>-83.622907),
"/US/US-MI/lansing-9854" => array("type"=>"city", "pop"=>113968, "name_ru"=>"Ленсинг", "name_en"=>"Lansing", "lang"=>"en", "lat"=>42.709100, "lon"=>-84.553996),
"/US/US-MI/lapeer-9855" => array("type"=>"city", "pop"=>9018, "name_ru"=>"Лапир", "name_en"=>"Lapeer", "lang"=>"en", "lat"=>43.047039, "lon"=>-83.324877),
"/US/US-MI/lathrup-village-9856" => array("type"=>"city", "pop"=>4060, "name_ru"=>"Латрап Виллидж", "name_en"=>"Lathrup Village", "lang"=>"en", "lat"=>42.492250, "lon"=>-83.227290),
"/US/US-MI/laurium-9857" => array("type"=>"village", "pop"=>2003, "name_ru"=>"Лориум", "name_en"=>"Laurium", "lang"=>"en", "lat"=>47.236479, "lon"=>-88.437595),
"/US/US-MI/lawrence-9858" => array("type"=>"village", "pop"=>994, "name_ru"=>"Лоренс", "name_en"=>"Lawrence", "lang"=>"en", "lat"=>42.218660, "lon"=>-86.052821),
"/US/US-MI/lawton-9859" => array("type"=>"village", "pop"=>1800, "name_ru"=>"Лотон", "name_en"=>"Lawton", "lang"=>"en", "lat"=>42.167178, "lon"=>-85.846436),
"/US/US-MI/le-roy-9862" => array("type"=>"village", "pop"=>256, "name_ru"=>"Ле Рой", "name_en"=>"Le Roy", "lang"=>"en", "lat"=>44.038509, "lon"=>-85.452685),
"/US/US-MI/lennon-9860" => array("type"=>"village", "pop"=>491, "name_ru"=>"Леннон", "name_en"=>"Lennon", "lang"=>"en", "lat"=>42.985312, "lon"=>-83.931372),
"/US/US-MI/leonard-9861" => array("type"=>"village", "pop"=>331, "name_ru"=>"Леонард", "name_en"=>"Leonard", "lang"=>"en", "lat"=>42.865851, "lon"=>-83.143202),
"/US/US-MI/leslie-9863" => array("type"=>"city", "pop"=>2272, "name_ru"=>"Лесли", "name_en"=>"Leslie", "lang"=>"en", "lat"=>42.450033, "lon"=>-84.430808),
"/US/US-MI/level-park-oak-park-9864" => array("type"=>"cdp", "pop"=>3502, "name_ru"=>"Левел Парк-Оук Парк", "name_en"=>"Level Park-Oak Park", "lang"=>"en", "lat"=>42.363485, "lon"=>-85.269361),
"/US/US-MI/lexington-9865" => array("type"=>"village", "pop"=>1046, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>43.268540, "lon"=>-82.533813),
"/US/US-MI/lincoln-9866" => array("type"=>"village", "pop"=>346, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>44.685207, "lon"=>-83.412645),
"/US/US-MI/lincoln-park-9867" => array("type"=>"city", "pop"=>35317, "name_ru"=>"Линкольн Парк", "name_en"=>"Lincoln Park", "lang"=>"en", "lat"=>42.243450, "lon"=>-83.181045),
"/US/US-MI/linden-9868" => array("type"=>"city", "pop"=>3358, "name_ru"=>"Линден", "name_en"=>"Linden", "lang"=>"en", "lat"=>42.819212, "lon"=>-83.781539),
"/US/US-MI/litchfield-9869" => array("type"=>"city", "pop"=>1392, "name_ru"=>"Литчфилд", "name_en"=>"Litchfield", "lang"=>"en", "lat"=>42.043077, "lon"=>-84.757528),
"/US/US-MI/livonia-9870" => array("type"=>"city", "pop"=>91220, "name_ru"=>"Ливония", "name_en"=>"Livonia", "lang"=>"en", "lat"=>42.398150, "lon"=>-83.372524),
"/US/US-MI/lowell-9871" => array("type"=>"city", "pop"=>4153, "name_ru"=>"Лоуэлл", "name_en"=>"Lowell", "lang"=>"en", "lat"=>42.933247, "lon"=>-85.345696),
"/US/US-MI/ludington-9872" => array("type"=>"city", "pop"=>8324, "name_ru"=>"Лудингтон", "name_en"=>"Ludington", "lang"=>"en", "lat"=>43.957070, "lon"=>-86.442992),
"/US/US-MI/luna-pier-9873" => array("type"=>"city", "pop"=>1520, "name_ru"=>"Луна Пир", "name_en"=>"Luna Pier", "lang"=>"en", "lat"=>41.805050, "lon"=>-83.442453),
"/US/US-MI/luther-9874" => array("type"=>"village", "pop"=>322, "name_ru"=>"Лютер", "name_en"=>"Luther", "lang"=>"en", "lat"=>44.038668, "lon"=>-85.682920),
"/US/US-MI/lyons-9875" => array("type"=>"village", "pop"=>713, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>42.983202, "lon"=>-84.945356),
"/US/US-MI/mackinac-island-9878" => array("type"=>"city", "pop"=>462, "name_ru"=>"Макинак Айленд", "name_en"=>"Mackinac Island", "lang"=>"en", "lat"=>45.855991, "lon"=>-84.621667),
"/US/US-MI/mackinaw-city-9879" => array("type"=>"village", "pop"=>840, "name_ru"=>"Макино Сити", "name_en"=>"Mackinaw City", "lang"=>"en", "lat"=>45.779004, "lon"=>-84.749772),
"/US/US-MI/madison-heights-9880" => array("type"=>"city", "pop"=>29531, "name_ru"=>"Мэдисон Хайтс", "name_en"=>"Madison Heights", "lang"=>"en", "lat"=>42.503700, "lon"=>-83.102650),
"/US/US-MI/mancelona-9881" => array("type"=>"village", "pop"=>1355, "name_ru"=>"Манселона", "name_en"=>"Mancelona", "lang"=>"en", "lat"=>44.902124, "lon"=>-85.061040),
"/US/US-MI/manchester-9882" => array("type"=>"village", "pop"=>2222, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>42.148950, "lon"=>-84.035491),
"/US/US-MI/manistee-9883" => array("type"=>"city", "pop"=>6110, "name_ru"=>"Манисти", "name_en"=>"Manistee", "lang"=>"en", "lat"=>44.244617, "lon"=>-86.326440),
"/US/US-MI/manistique-9884" => array("type"=>"city", "pop"=>3085, "name_ru"=>"Манистик", "name_en"=>"Manistique", "lang"=>"en", "lat"=>45.958617, "lon"=>-86.250091),
"/US/US-MI/manitou-beach-devils-lake-9885" => array("type"=>"cdp", "pop"=>2061, "name_ru"=>"Маниту Бич-Девилс Лейк", "name_en"=>"Manitou Beach-Devils Lake", "lang"=>"en", "lat"=>41.974954, "lon"=>-84.279777),
"/US/US-MI/manton-9886" => array("type"=>"city", "pop"=>1167, "name_ru"=>"Мантон", "name_en"=>"Manton", "lang"=>"en", "lat"=>44.411436, "lon"=>-85.400405),
"/US/US-MI/maple-rapids-9887" => array("type"=>"village", "pop"=>598, "name_ru"=>"Мейпл Рэпидс", "name_en"=>"Maple Rapids", "lang"=>"en", "lat"=>43.106613, "lon"=>-84.690081),
"/US/US-MI/marcellus-9888" => array("type"=>"village", "pop"=>1040, "name_ru"=>"Марселлус", "name_en"=>"Marcellus", "lang"=>"en", "lat"=>42.025977, "lon"=>-85.813250),
"/US/US-MI/marine-city-9889" => array("type"=>"city", "pop"=>4377, "name_ru"=>"Марин Сити", "name_en"=>"Marine City", "lang"=>"en", "lat"=>42.714237, "lon"=>-82.501142),
"/US/US-MI/marion-9890" => array("type"=>"village", "pop"=>802, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>44.102755, "lon"=>-85.147033),
"/US/US-MI/marlette-9891" => array("type"=>"city", "pop"=>1984, "name_ru"=>"Марлетт", "name_en"=>"Marlette", "lang"=>"en", "lat"=>43.326451, "lon"=>-83.080694),
"/US/US-MI/marquette-9892" => array("type"=>"city", "pop"=>20916, "name_ru"=>"Маркетт", "name_en"=>"Marquette", "lang"=>"en", "lat"=>46.550806, "lon"=>-87.395726),
"/US/US-MI/marshall-9893" => array("type"=>"city", "pop"=>7121, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>42.261692, "lon"=>-84.958353),
"/US/US-MI/martin-9894" => array("type"=>"village", "pop"=>407, "name_ru"=>"Мартин", "name_en"=>"Martin", "lang"=>"en", "lat"=>42.537129, "lon"=>-85.637479),
"/US/US-MI/marysville-9895" => array("type"=>"city", "pop"=>9943, "name_ru"=>"Мэрисвилл", "name_en"=>"Marysville", "lang"=>"en", "lat"=>42.910168, "lon"=>-82.479817),
"/US/US-MI/mason-9896" => array("type"=>"city", "pop"=>8099, "name_ru"=>"Мейсон", "name_en"=>"Mason", "lang"=>"en", "lat"=>42.581700, "lon"=>-84.443899),
"/US/US-MI/mattawan-9897" => array("type"=>"village", "pop"=>2795, "name_ru"=>"Мэттаван", "name_en"=>"Mattawan", "lang"=>"en", "lat"=>42.214025, "lon"=>-85.787622),
"/US/US-MI/maybee-9898" => array("type"=>"village", "pop"=>520, "name_ru"=>"Мейби", "name_en"=>"Maybee", "lang"=>"en", "lat"=>42.005765, "lon"=>-83.516083),
"/US/US-MI/mayville-9899" => array("type"=>"village", "pop"=>988, "name_ru"=>"Мейвилл", "name_en"=>"Mayville", "lang"=>"en", "lat"=>43.336045, "lon"=>-83.353261),
"/US/US-MI/mcbain-9876" => array("type"=>"city", "pop"=>583, "name_ru"=>"МакБейн", "name_en"=>"McBain", "lang"=>"en", "lat"=>44.194441, "lon"=>-85.214414),
"/US/US-MI/mcbride-9877" => array("type"=>"village", "pop"=>227, "name_ru"=>"МакБрайд", "name_en"=>"McBride", "lang"=>"en", "lat"=>43.352841, "lon"=>-85.043106),
"/US/US-MI/mecosta-9900" => array("type"=>"village", "pop"=>433, "name_ru"=>"Мекоста", "name_en"=>"Mecosta", "lang"=>"en", "lat"=>43.618918, "lon"=>-85.230505),
"/US/US-MI/melvin-9901" => array("type"=>"village", "pop"=>150, "name_ru"=>"Мэлвин", "name_en"=>"Melvin", "lang"=>"en", "lat"=>43.184602, "lon"=>-82.863451),
"/US/US-MI/melvindale-9902" => array("type"=>"city", "pop"=>9812, "name_ru"=>"Мелвиндейл", "name_en"=>"Melvindale", "lang"=>"en", "lat"=>42.281119, "lon"=>-83.179087),
"/US/US-MI/memphis-9903" => array("type"=>"city", "pop"=>1110, "name_ru"=>"Мемфис", "name_en"=>"Memphis", "lang"=>"en", "lat"=>42.895331, "lon"=>-82.769524),
"/US/US-MI/mendon-9904" => array("type"=>"village", "pop"=>909, "name_ru"=>"Мендон", "name_en"=>"Mendon", "lang"=>"en", "lat"=>42.008412, "lon"=>-85.454098),
"/US/US-MI/menominee-9905" => array("type"=>"city", "pop"=>8324, "name_ru"=>"Меномини", "name_en"=>"Menominee", "lang"=>"en", "lat"=>45.121146, "lon"=>-87.623526),
"/US/US-MI/merrill-9906" => array("type"=>"village", "pop"=>719, "name_ru"=>"Меррил", "name_en"=>"Merrill", "lang"=>"en", "lat"=>43.409274, "lon"=>-84.335297),
"/US/US-MI/mesick-9907" => array("type"=>"village", "pop"=>451, "name_ru"=>"Месик", "name_en"=>"Mesick", "lang"=>"en", "lat"=>44.403559, "lon"=>-85.719382),
"/US/US-MI/metamora-9908" => array("type"=>"village", "pop"=>483, "name_ru"=>"Метамора", "name_en"=>"Metamora", "lang"=>"en", "lat"=>42.941206, "lon"=>-83.290574),
"/US/US-MI/michiana-9909" => array("type"=>"village", "pop"=>193, "name_ru"=>"Мичиана", "name_en"=>"Michiana", "lang"=>"en", "lat"=>41.763400, "lon"=>-86.811193),
"/US/US-MI/michigan-center-9910" => array("type"=>"cdp", "pop"=>4863, "name_ru"=>"Мичиган Сентер", "name_en"=>"Michigan Center", "lang"=>"en", "lat"=>42.226860, "lon"=>-84.322535),
"/US/US-MI/middletown-9911" => array("type"=>"cdp", "pop"=>1010, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>42.985708, "lon"=>-84.144908),
"/US/US-MI/middleville-9912" => array("type"=>"village", "pop"=>2697, "name_ru"=>"Миддлвилл", "name_en"=>"Middleville", "lang"=>"en", "lat"=>42.713413, "lon"=>-85.465535),
"/US/US-MI/midland-9913" => array("type"=>"city", "pop"=>40917, "name_ru"=>"Мидленд", "name_en"=>"Midland", "lang"=>"en", "lat"=>43.623618, "lon"=>-84.229365),
"/US/US-MI/milan-9914" => array("type"=>"city", "pop"=>5657, "name_ru"=>"Милан", "name_en"=>"Milan", "lang"=>"en", "lat"=>42.083993, "lon"=>-83.683394),
"/US/US-MI/milford-9915" => array("type"=>"village", "pop"=>6482, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>42.585650, "lon"=>-83.601124),
"/US/US-MI/millersburg-9916" => array("type"=>"village", "pop"=>242, "name_ru"=>"Миллерсберг", "name_en"=>"Millersburg", "lang"=>"en", "lat"=>45.333835, "lon"=>-84.060604),
"/US/US-MI/millington-9917" => array("type"=>"village", "pop"=>1070, "name_ru"=>"Миллингтон", "name_en"=>"Millington", "lang"=>"en", "lat"=>43.280318, "lon"=>-83.528819),
"/US/US-MI/minden-city-9918" => array("type"=>"village", "pop"=>226, "name_ru"=>"Минден Сити", "name_en"=>"Minden City", "lang"=>"en", "lat"=>43.670969, "lon"=>-82.775205),
"/US/US-MI/mineral-hills-9919" => array("type"=>"village", "pop"=>200, "name_ru"=>"Минерал Хилс", "name_en"=>"Mineral Hills", "lang"=>"en", "lat"=>46.112255, "lon"=>-88.644668),
"/US/US-MI/mio-9920" => array("type"=>"cdp", "pop"=>1886, "name_ru"=>"Мио", "name_en"=>"Mio", "lang"=>"en", "lat"=>44.663030, "lon"=>-84.144591),
"/US/US-MI/monroe-9921" => array("type"=>"city", "pop"=>21374, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>41.916698, "lon"=>-83.386026),
"/US/US-MI/montague-9922" => array("type"=>"city", "pop"=>2294, "name_ru"=>"Монтагю", "name_en"=>"Montague", "lang"=>"en", "lat"=>43.412338, "lon"=>-86.362913),
"/US/US-MI/montgomery-9923" => array("type"=>"village", "pop"=>367, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>41.776933, "lon"=>-84.805994),
"/US/US-MI/montrose-9924" => array("type"=>"city", "pop"=>1480, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>43.175624, "lon"=>-83.892606),
"/US/US-MI/morenci-9925" => array("type"=>"city", "pop"=>2259, "name_ru"=>"Моренси", "name_en"=>"Morenci", "lang"=>"en", "lat"=>41.721626, "lon"=>-84.216417),
"/US/US-MI/morley-9926" => array("type"=>"village", "pop"=>481, "name_ru"=>"Морли", "name_en"=>"Morley", "lang"=>"en", "lat"=>43.491239, "lon"=>-85.446039),
"/US/US-MI/morrice-9927" => array("type"=>"village", "pop"=>872, "name_ru"=>"Моррис", "name_en"=>"Morrice", "lang"=>"en", "lat"=>42.838032, "lon"=>-84.179535),
"/US/US-MI/mount-clemens-9928" => array("type"=>"city", "pop"=>16781, "name_ru"=>"Маунт Клеменс", "name_en"=>"Mount Clemens", "lang"=>"en", "lat"=>42.598050, "lon"=>-82.881931),
"/US/US-MI/mount-morris-9929" => array("type"=>"city", "pop"=>3168, "name_ru"=>"Маунт Моррис", "name_en"=>"Mount Morris", "lang"=>"en", "lat"=>43.116350, "lon"=>-83.698949),
"/US/US-MI/mount-pleasant-9930" => array("type"=>"city", "pop"=>26675, "name_ru"=>"Маунт Плезант", "name_en"=>"Mount Pleasant", "lang"=>"en", "lat"=>43.596552, "lon"=>-84.778251),
"/US/US-MI/muir-9931" => array("type"=>"village", "pop"=>623, "name_ru"=>"Мьюир", "name_en"=>"Muir", "lang"=>"en", "lat"=>42.997390, "lon"=>-84.935483),
"/US/US-MI/mulliken-9932" => array("type"=>"village", "pop"=>550, "name_ru"=>"Мулликен", "name_en"=>"Mulliken", "lang"=>"en", "lat"=>42.762978, "lon"=>-84.895069),
"/US/US-MI/munising-9933" => array("type"=>"city", "pop"=>2311, "name_ru"=>"Мунизинг", "name_en"=>"Munising", "lang"=>"en", "lat"=>46.417026, "lon"=>-86.641215),
"/US/US-MI/muskegon-9934" => array("type"=>"city", "pop"=>39401, "name_ru"=>"Маскегон", "name_en"=>"Muskegon", "lang"=>"en", "lat"=>43.227550, "lon"=>-86.255534),
"/US/US-MI/muskegon-heights-9935" => array("type"=>"city", "pop"=>11623, "name_ru"=>"Маскегон Хайтс", "name_en"=>"Muskegon Heights", "lang"=>"en", "lat"=>43.201600, "lon"=>-86.241049),
"/US/US-MI/napoleon-9936" => array("type"=>"cdp", "pop"=>6961, "name_ru"=>"Наполеон", "name_en"=>"Napoleon", "lang"=>"en", "lat"=>42.168350, "lon"=>-84.246274),
"/US/US-MI/nashville-9937" => array("type"=>"village", "pop"=>1655, "name_ru"=>"Нэшвилл", "name_en"=>"Nashville", "lang"=>"en", "lat"=>42.602910, "lon"=>-85.093877),
"/US/US-MI/negaunee-9938" => array("type"=>"city", "pop"=>4456, "name_ru"=>"Негауни", "name_en"=>"Negaunee", "lang"=>"en", "lat"=>46.499736, "lon"=>-87.596870),
"/US/US-MI/new-baltimore-9940" => array("type"=>"city", "pop"=>11346, "name_ru"=>"Нью Балтимор", "name_en"=>"New Baltimore", "lang"=>"en", "lat"=>42.683150, "lon"=>-82.738144),
"/US/US-MI/new-buffalo-9942" => array("type"=>"city", "pop"=>2423, "name_ru"=>"Нью Буффало", "name_en"=>"New Buffalo", "lang"=>"en", "lat"=>41.790369, "lon"=>-86.745272),
"/US/US-MI/new-era-9943" => array("type"=>"village", "pop"=>456, "name_ru"=>"Нью Эра", "name_en"=>"New Era", "lang"=>"en", "lat"=>43.559520, "lon"=>-86.346908),
"/US/US-MI/new-haven-9944" => array("type"=>"village", "pop"=>5132, "name_ru"=>"Нью Хейвен", "name_en"=>"New Haven", "lang"=>"en", "lat"=>42.732600, "lon"=>-82.794106),
"/US/US-MI/new-lothrop-9945" => array("type"=>"village", "pop"=>578, "name_ru"=>"Нью Лотроп", "name_en"=>"New Lothrop", "lang"=>"en", "lat"=>43.116096, "lon"=>-83.968927),
"/US/US-MI/newaygo-9939" => array("type"=>"city", "pop"=>1626, "name_ru"=>"Ньюейго", "name_en"=>"Newaygo", "lang"=>"en", "lat"=>43.420045, "lon"=>-85.800517),
"/US/US-MI/newberry-9941" => array("type"=>"village", "pop"=>1517, "name_ru"=>"Ньюберри", "name_en"=>"Newberry", "lang"=>"en", "lat"=>46.353302, "lon"=>-85.508968),
"/US/US-MI/niles-9946" => array("type"=>"city", "pop"=>11258, "name_ru"=>"Найлс", "name_en"=>"Niles", "lang"=>"en", "lat"=>41.834472, "lon"=>-86.247981),
"/US/US-MI/north-adams-9947" => array("type"=>"village", "pop"=>485, "name_ru"=>"Норт Адамс", "name_en"=>"North Adams", "lang"=>"en", "lat"=>41.971888, "lon"=>-84.524225),
"/US/US-MI/north-branch-9948" => array("type"=>"village", "pop"=>977, "name_ru"=>"Норт Брэнч", "name_en"=>"North Branch", "lang"=>"en", "lat"=>43.229814, "lon"=>-83.191706),
"/US/US-MI/north-muskegon-9949" => array("type"=>"city", "pop"=>3915, "name_ru"=>"Норт Маскегон", "name_en"=>"North Muskegon", "lang"=>"en", "lat"=>43.250440, "lon"=>-86.272456),
"/US/US-MI/northport-9950" => array("type"=>"village", "pop"=>631, "name_ru"=>"Нортпорт", "name_en"=>"Northport", "lang"=>"en", "lat"=>45.130567, "lon"=>-85.616919),
"/US/US-MI/northview-9951" => array("type"=>"cdp", "pop"=>13712, "name_ru"=>"Нортвью", "name_en"=>"Northview", "lang"=>"en", "lat"=>43.045650, "lon"=>-85.600610),
"/US/US-MI/northville-9952" => array("type"=>"city", "pop"=>5996, "name_ru"=>"Нортвилл", "name_en"=>"Northville", "lang"=>"en", "lat"=>42.436508, "lon"=>-83.488384),
"/US/US-MI/norton-shores-9953" => array("type"=>"city", "pop"=>23307, "name_ru"=>"Нортон Шорс", "name_en"=>"Norton Shores", "lang"=>"en", "lat"=>43.160050, "lon"=>-86.252524),
"/US/US-MI/norway-9954" => array("type"=>"city", "pop"=>2821, "name_ru"=>"Норуэй", "name_en"=>"Norway", "lang"=>"en", "lat"=>45.802597, "lon"=>-87.914647),
"/US/US-MI/novi-9955" => array("type"=>"city", "pop"=>54054, "name_ru"=>"Нови", "name_en"=>"Novi", "lang"=>"en", "lat"=>42.475493, "lon"=>-83.490466),
"/US/US-MI/oak-park-9957" => array("type"=>"city", "pop"=>30547, "name_ru"=>"Оук Парк", "name_en"=>"Oak Park", "lang"=>"en", "lat"=>42.463133, "lon"=>-83.180499),
"/US/US-MI/oakley-9956" => array("type"=>"village", "pop"=>315, "name_ru"=>"Оукли", "name_en"=>"Oakley", "lang"=>"en", "lat"=>43.143106, "lon"=>-84.168568),
"/US/US-MI/okemos-9958" => array("type"=>"cdp", "pop"=>20216, "name_ru"=>"Окмос", "name_en"=>"Okemos", "lang"=>"en", "lat"=>42.708250, "lon"=>-84.414449),
"/US/US-MI/olivet-9959" => array("type"=>"city", "pop"=>1813, "name_ru"=>"Оливет", "name_en"=>"Olivet", "lang"=>"en", "lat"=>42.446079, "lon"=>-84.923556),
"/US/US-MI/omer-9960" => array("type"=>"city", "pop"=>297, "name_ru"=>"Омер", "name_en"=>"Omer", "lang"=>"en", "lat"=>44.049523, "lon"=>-83.857210),
"/US/US-MI/onaway-9961" => array("type"=>"city", "pop"=>902, "name_ru"=>"Онуэй", "name_en"=>"Onaway", "lang"=>"en", "lat"=>45.358475, "lon"=>-84.228302),
"/US/US-MI/onekama-9962" => array("type"=>"village", "pop"=>597, "name_ru"=>"Уанкама", "name_en"=>"Onekama", "lang"=>"en", "lat"=>44.366836, "lon"=>-86.203089),
"/US/US-MI/onsted-9963" => array("type"=>"village", "pop"=>993, "name_ru"=>"Онстед", "name_en"=>"Onsted", "lang"=>"en", "lat"=>42.007600, "lon"=>-84.190199),
"/US/US-MI/ontonagon-9964" => array("type"=>"village", "pop"=>1509, "name_ru"=>"Онтонагон", "name_en"=>"Ontonagon", "lang"=>"en", "lat"=>46.867129, "lon"=>-89.314271),
"/US/US-MI/orchard-lake-village-9965" => array("type"=>"city", "pop"=>2209, "name_ru"=>"Орчард Лейк Виллидж", "name_en"=>"Orchard Lake Village", "lang"=>"en", "lat"=>42.582401, "lon"=>-83.377467),
"/US/US-MI/ortonville-9966" => array("type"=>"village", "pop"=>1482, "name_ru"=>"Ортонвилл", "name_en"=>"Ortonville", "lang"=>"en", "lat"=>42.851440, "lon"=>-83.444212),
"/US/US-MI/oscoda-9967" => array("type"=>"cdp", "pop"=>8836, "name_ru"=>"Оскода", "name_en"=>"Oscoda", "lang"=>"en", "lat"=>44.422184, "lon"=>-83.334773),
"/US/US-MI/ossineke-9968" => array("type"=>"cdp", "pop"=>1656, "name_ru"=>"Оссинк", "name_en"=>"Ossineke", "lang"=>"en", "lat"=>44.907188, "lon"=>-83.431972),
"/US/US-MI/otisville-9969" => array("type"=>"village", "pop"=>803, "name_ru"=>"Отисвилл", "name_en"=>"Otisville", "lang"=>"en", "lat"=>43.165272, "lon"=>-83.524336),
"/US/US-MI/otsego-9970" => array("type"=>"city", "pop"=>3821, "name_ru"=>"Отсего", "name_en"=>"Otsego", "lang"=>"en", "lat"=>42.457655, "lon"=>-85.697143),
"/US/US-MI/otter-lake-9971" => array("type"=>"village", "pop"=>411, "name_ru"=>"Оттер Лейк", "name_en"=>"Otter Lake", "lang"=>"en", "lat"=>43.213144, "lon"=>-83.460491),
"/US/US-MI/ovid-9972" => array("type"=>"village", "pop"=>1399, "name_ru"=>"Овид", "name_en"=>"Ovid", "lang"=>"en", "lat"=>43.002506, "lon"=>-84.374545),
"/US/US-MI/owendale-9973" => array("type"=>"village", "pop"=>265, "name_ru"=>"Оуэндейл", "name_en"=>"Owendale", "lang"=>"en", "lat"=>43.727025, "lon"=>-83.267746),
"/US/US-MI/owosso-9974" => array("type"=>"city", "pop"=>14963, "name_ru"=>"Оуоссо", "name_en"=>"Owosso", "lang"=>"en", "lat"=>42.995223, "lon"=>-84.175771),
"/US/US-MI/oxford-9975" => array("type"=>"village", "pop"=>3526, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>42.821039, "lon"=>-83.253996),
"/US/US-MI/parchment-9976" => array("type"=>"city", "pop"=>1808, "name_ru"=>"Парчмент", "name_en"=>"Parchment", "lang"=>"en", "lat"=>42.326950, "lon"=>-85.565487),
"/US/US-MI/parma-9977" => array("type"=>"village", "pop"=>846, "name_ru"=>"Парма", "name_en"=>"Parma", "lang"=>"en", "lat"=>42.257250, "lon"=>-84.598349),
"/US/US-MI/paw-paw-9978" => array("type"=>"village", "pop"=>3222, "name_ru"=>"По По", "name_en"=>"Paw Paw", "lang"=>"en", "lat"=>42.217998, "lon"=>-85.889994),
"/US/US-MI/paw-paw-lake-9979" => array("type"=>"cdp", "pop"=>3782, "name_ru"=>"По По Лейк", "name_en"=>"Paw Paw Lake", "lang"=>"en", "lat"=>42.211244, "lon"=>-86.275590),
"/US/US-MI/pearl-beach-9980" => array("type"=>"cdp", "pop"=>3394, "name_ru"=>"Перл Бич", "name_en"=>"Pearl Beach", "lang"=>"en", "lat"=>42.624170, "lon"=>-82.590966),
"/US/US-MI/peck-9981" => array("type"=>"village", "pop"=>562, "name_ru"=>"Пек", "name_en"=>"Peck", "lang"=>"en", "lat"=>43.259068, "lon"=>-82.816509),
"/US/US-MI/pellston-9982" => array("type"=>"village", "pop"=>772, "name_ru"=>"Пеллстон", "name_en"=>"Pellston", "lang"=>"en", "lat"=>45.551509, "lon"=>-84.783290),
"/US/US-MI/pentwater-9983" => array("type"=>"village", "pop"=>942, "name_ru"=>"Пентуотер", "name_en"=>"Pentwater", "lang"=>"en", "lat"=>43.779425, "lon"=>-86.430942),
"/US/US-MI/perrinton-9984" => array("type"=>"village", "pop"=>422, "name_ru"=>"Перринтон", "name_en"=>"Perrinton", "lang"=>"en", "lat"=>43.181687, "lon"=>-84.678604),
"/US/US-MI/perry-9985" => array("type"=>"city", "pop"=>2025, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>42.820885, "lon"=>-84.228250),
"/US/US-MI/petersburg-9986" => array("type"=>"city", "pop"=>1120, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>41.899572, "lon"=>-83.711817),
"/US/US-MI/petoskey-9987" => array("type"=>"city", "pop"=>6028, "name_ru"=>"Петоски", "name_en"=>"Petoskey", "lang"=>"en", "lat"=>45.366115, "lon"=>-84.955072),
"/US/US-MI/pewamo-9988" => array("type"=>"village", "pop"=>549, "name_ru"=>"Пьюамо", "name_en"=>"Pewamo", "lang"=>"en", "lat"=>43.001264, "lon"=>-84.846455),
"/US/US-MI/pierson-9989" => array("type"=>"village", "pop"=>181, "name_ru"=>"Пирсон", "name_en"=>"Pierson", "lang"=>"en", "lat"=>43.319043, "lon"=>-85.497514),
"/US/US-MI/pigeon-9990" => array("type"=>"village", "pop"=>1066, "name_ru"=>"Пиджен", "name_en"=>"Pigeon", "lang"=>"en", "lat"=>43.829503, "lon"=>-83.269995),
"/US/US-MI/pinckney-9991" => array("type"=>"village", "pop"=>2423, "name_ru"=>"Пинкни", "name_en"=>"Pinckney", "lang"=>"en", "lat"=>42.454456, "lon"=>-83.945546),
"/US/US-MI/pinconning-9992" => array("type"=>"city", "pop"=>1314, "name_ru"=>"Пинконнинг", "name_en"=>"Pinconning", "lang"=>"en", "lat"=>43.857922, "lon"=>-83.964618),
"/US/US-MI/plainwell-9993" => array("type"=>"city", "pop"=>3856, "name_ru"=>"Плейнуэлл", "name_en"=>"Plainwell", "lang"=>"en", "lat"=>42.445143, "lon"=>-85.644083),
"/US/US-MI/pleasant-ridge-9994" => array("type"=>"city", "pop"=>2469, "name_ru"=>"Плезант Ридж", "name_en"=>"Pleasant Ridge", "lang"=>"en", "lat"=>42.470900, "lon"=>-83.144913),
"/US/US-MI/plymouth-9995" => array("type"=>"city", "pop"=>8487, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>42.371150, "lon"=>-83.467418),
"/US/US-MI/plymouth-township-9996" => array("type"=>"cdp", "pop"=>23646, "name_ru"=>"Плимут Тауншип", "name_en"=>"Plymouth Township", "lang"=>"en", "lat"=>42.371255, "lon"=>-83.494223),
"/US/US-MI/pontiac-9997" => array("type"=>"city", "pop"=>66095, "name_ru"=>"Понтиак", "name_en"=>"Pontiac", "lang"=>"en", "lat"=>42.649100, "lon"=>-83.287192),
"/US/US-MI/port-austin-9999" => array("type"=>"village", "pop"=>652, "name_ru"=>"Порт Остин", "name_en"=>"Port Austin", "lang"=>"en", "lat"=>44.042681, "lon"=>-82.995338),
"/US/US-MI/port-hope-10000" => array("type"=>"village", "pop"=>273, "name_ru"=>"Порт Хоуп", "name_en"=>"Port Hope", "lang"=>"en", "lat"=>43.939643, "lon"=>-82.715139),
"/US/US-MI/port-huron-10001" => array("type"=>"city", "pop"=>30869, "name_ru"=>"Порт Хурон", "name_en"=>"Port Huron", "lang"=>"en", "lat"=>42.993719, "lon"=>-82.433628),
"/US/US-MI/port-sanilac-10003" => array("type"=>"village", "pop"=>620, "name_ru"=>"Порт Санилак", "name_en"=>"Port Sanilac", "lang"=>"en", "lat"=>43.430088, "lon"=>-82.546057),
"/US/US-MI/portage-9998" => array("type"=>"city", "pop"=>46133, "name_ru"=>"Портейдж", "name_en"=>"Portage", "lang"=>"en", "lat"=>42.201383, "lon"=>-85.590555),
"/US/US-MI/portland-10002" => array("type"=>"city", "pop"=>3687, "name_ru"=>"Портленд", "name_en"=>"Portland", "lang"=>"en", "lat"=>42.871205, "lon"=>-84.901098),
"/US/US-MI/posen-10004" => array("type"=>"village", "pop"=>267, "name_ru"=>"Посен", "name_en"=>"Posen", "lang"=>"en", "lat"=>45.262113, "lon"=>-83.699409),
"/US/US-MI/potterville-10005" => array("type"=>"city", "pop"=>2147, "name_ru"=>"Поттервилл", "name_en"=>"Potterville", "lang"=>"en", "lat"=>42.628034, "lon"=>-84.745795),
"/US/US-MI/powers-10006" => array("type"=>"village", "pop"=>415, "name_ru"=>"Пауэрс", "name_en"=>"Powers", "lang"=>"en", "lat"=>45.687504, "lon"=>-87.526371),
"/US/US-MI/prescott-10007" => array("type"=>"village", "pop"=>270, "name_ru"=>"Прескотт", "name_en"=>"Prescott", "lang"=>"en", "lat"=>44.191391, "lon"=>-83.931860),
"/US/US-MI/prudenville-10008" => array("type"=>"cdp", "pop"=>1513, "name_ru"=>"Пруденвилл", "name_en"=>"Prudenville", "lang"=>"en", "lat"=>44.301361, "lon"=>-84.664645),
"/US/US-MI/quincy-10009" => array("type"=>"village", "pop"=>1594, "name_ru"=>"Квинси", "name_en"=>"Quincy", "lang"=>"en", "lat"=>41.943929, "lon"=>-84.882951),
"/US/US-MI/quinnesec-10010" => array("type"=>"cdp", "pop"=>1254, "name_ru"=>"Квиннесек", "name_en"=>"Quinnesec", "lang"=>"en", "lat"=>45.801436, "lon"=>-87.998456),
"/US/US-MI/ravenna-10011" => array("type"=>"village", "pop"=>1211, "name_ru"=>"Равенна", "name_en"=>"Ravenna", "lang"=>"en", "lat"=>43.189575, "lon"=>-85.940273),
"/US/US-MI/reading-10012" => array("type"=>"city", "pop"=>1071, "name_ru"=>"Ридинг", "name_en"=>"Reading", "lang"=>"en", "lat"=>41.839201, "lon"=>-84.747325),
"/US/US-MI/redford-10013" => array("type"=>"cdp", "pop"=>54387, "name_ru"=>"Редфорд", "name_en"=>"Redford", "lang"=>"en", "lat"=>42.394800, "lon"=>-83.294216),
"/US/US-MI/reed-city-10014" => array("type"=>"city", "pop"=>2330, "name_ru"=>"Рид Сити", "name_en"=>"Reed City", "lang"=>"en", "lat"=>43.872548, "lon"=>-85.509184),
"/US/US-MI/reese-10015" => array("type"=>"village", "pop"=>1346, "name_ru"=>"Рис", "name_en"=>"Reese", "lang"=>"en", "lat"=>43.451769, "lon"=>-83.688570),
"/US/US-MI/richland-10016" => array("type"=>"village", "pop"=>752, "name_ru"=>"Ричленд", "name_en"=>"Richland", "lang"=>"en", "lat"=>42.375633, "lon"=>-85.456132),
"/US/US-MI/richmond-10017" => array("type"=>"city", "pop"=>5606, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>42.808963, "lon"=>-82.754602),
"/US/US-MI/river-rouge-10018" => array("type"=>"city", "pop"=>8449, "name_ru"=>"Ривер Руж", "name_en"=>"River Rouge", "lang"=>"en", "lat"=>42.274350, "lon"=>-83.123543),
"/US/US-MI/riverview-10019" => array("type"=>"city", "pop"=>11817, "name_ru"=>"Ривервью", "name_en"=>"Riverview", "lang"=>"en", "lat"=>42.171351, "lon"=>-83.192119),
"/US/US-MI/robin-glen-indiantown-10020" => array("type"=>"cdp", "pop"=>1395, "name_ru"=>"Робин Глен-Индиантаун", "name_en"=>"Robin Glen-Indiantown", "lang"=>"en", "lat"=>43.460656, "lon"=>-83.836832),
"/US/US-MI/rochester-10021" => array("type"=>"city", "pop"=>11002, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>42.686668, "lon"=>-83.123785),
"/US/US-MI/rochester-hills-10022" => array("type"=>"city", "pop"=>69014, "name_ru"=>"Рочестер Хилс", "name_en"=>"Rochester Hills", "lang"=>"en", "lat"=>42.665674, "lon"=>-83.158205),
"/US/US-MI/rockford-10023" => array("type"=>"city", "pop"=>5411, "name_ru"=>"Рокфорд", "name_en"=>"Rockford", "lang"=>"en", "lat"=>43.126209, "lon"=>-85.556629),
"/US/US-MI/rockwood-10024" => array("type"=>"city", "pop"=>3160, "name_ru"=>"Роквуд", "name_en"=>"Rockwood", "lang"=>"en", "lat"=>42.070300, "lon"=>-83.244377),
"/US/US-MI/rogers-city-10025" => array("type"=>"city", "pop"=>3031, "name_ru"=>"Роджерс Сити", "name_en"=>"Rogers City", "lang"=>"en", "lat"=>45.418522, "lon"=>-83.806452),
"/US/US-MI/romeo-10026" => array("type"=>"village", "pop"=>3737, "name_ru"=>"Ромео", "name_en"=>"Romeo", "lang"=>"en", "lat"=>42.804175, "lon"=>-83.005303),
"/US/US-MI/romulus-10027" => array("type"=>"city", "pop"=>22963, "name_ru"=>"Ромулус", "name_en"=>"Romulus", "lang"=>"en", "lat"=>42.223750, "lon"=>-83.366220),
"/US/US-MI/roosevelt-park-10028" => array("type"=>"city", "pop"=>3769, "name_ru"=>"Рузвельт Парк", "name_en"=>"Roosevelt Park", "lang"=>"en", "lat"=>43.197550, "lon"=>-86.273525),
"/US/US-MI/roscommon-10029" => array("type"=>"village", "pop"=>1059, "name_ru"=>"Роскоммон", "name_en"=>"Roscommon", "lang"=>"en", "lat"=>44.490787, "lon"=>-84.588894),
"/US/US-MI/rose-city-10031" => array("type"=>"city", "pop"=>685, "name_ru"=>"Роуз Сити", "name_en"=>"Rose City", "lang"=>"en", "lat"=>44.421030, "lon"=>-84.115391),
"/US/US-MI/rosebush-10030" => array("type"=>"village", "pop"=>341, "name_ru"=>"Роузбуш", "name_en"=>"Rosebush", "lang"=>"en", "lat"=>43.699612, "lon"=>-84.765802),
"/US/US-MI/roseville-10032" => array("type"=>"city", "pop"=>46782, "name_ru"=>"Роузвилл", "name_en"=>"Roseville", "lang"=>"en", "lat"=>42.507450, "lon"=>-82.936608),
"/US/US-MI/rothbury-10033" => array("type"=>"village", "pop"=>433, "name_ru"=>"Ротбери", "name_en"=>"Rothbury", "lang"=>"en", "lat"=>43.507353, "lon"=>-86.348429),
"/US/US-MI/royal-oak-10034" => array("type"=>"city", "pop"=>57110, "name_ru"=>"Ройал Оук", "name_en"=>"Royal Oak", "lang"=>"en", "lat"=>42.508600, "lon"=>-83.154622),
"/US/US-MI/saginaw-10035" => array("type"=>"city", "pop"=>55620, "name_ru"=>"Сагино", "name_en"=>"Saginaw", "lang"=>"en", "lat"=>43.419550, "lon"=>-83.949521),
"/US/US-MI/saginaw-township-north-10036" => array("type"=>"cdp", "pop"=>23018, "name_ru"=>"Сагино Тауншип Норт", "name_en"=>"Saginaw Township North", "lang"=>"en", "lat"=>43.458935, "lon"=>-84.007741),
"/US/US-MI/saginaw-township-south-10037" => array("type"=>"cdp", "pop"=>13987, "name_ru"=>"Сагино Тауншип Саут", "name_en"=>"Saginaw Township South", "lang"=>"en", "lat"=>43.418950, "lon"=>-84.018991),
"/US/US-MI/saline-10046" => array("type"=>"city", "pop"=>8882, "name_ru"=>"Сейлайн", "name_en"=>"Saline", "lang"=>"en", "lat"=>42.175814, "lon"=>-83.779976),
"/US/US-MI/sand-lake-10047" => array("type"=>"village", "pop"=>515, "name_ru"=>"Сэнд Лейк", "name_en"=>"Sand Lake", "lang"=>"en", "lat"=>43.290864, "lon"=>-85.518209),
"/US/US-MI/sandusky-10048" => array("type"=>"city", "pop"=>2609, "name_ru"=>"Сэндаски", "name_en"=>"Sandusky", "lang"=>"en", "lat"=>43.422938, "lon"=>-82.831813),
"/US/US-MI/sanford-10049" => array("type"=>"village", "pop"=>910, "name_ru"=>"Санфорд", "name_en"=>"Sanford", "lang"=>"en", "lat"=>43.675577, "lon"=>-84.380649),
"/US/US-MI/saranac-10050" => array("type"=>"village", "pop"=>1296, "name_ru"=>"Саранак", "name_en"=>"Saranac", "lang"=>"en", "lat"=>42.929797, "lon"=>-85.209376),
"/US/US-MI/saugatuck-10051" => array("type"=>"city", "pop"=>1009, "name_ru"=>"Сотак", "name_en"=>"Saugatuck", "lang"=>"en", "lat"=>42.656111, "lon"=>-86.205805),
"/US/US-MI/sault-ste.-marie-10052" => array("type"=>"city", "pop"=>14087, "name_ru"=>"Солт Сент Мари", "name_en"=>"Sault Ste. Marie", "lang"=>"en", "lat"=>46.484562, "lon"=>-84.365456),
"/US/US-MI/schoolcraft-10053" => array("type"=>"village", "pop"=>1512, "name_ru"=>"Скулкрафт", "name_en"=>"Schoolcraft", "lang"=>"en", "lat"=>42.115154, "lon"=>-85.633543),
"/US/US-MI/scottville-10054" => array("type"=>"city", "pop"=>1250, "name_ru"=>"Скоттвилл", "name_en"=>"Scottville", "lang"=>"en", "lat"=>43.950399, "lon"=>-86.280008),
"/US/US-MI/sebewaing-10055" => array("type"=>"village", "pop"=>1750, "name_ru"=>"Себуэйнг", "name_en"=>"Sebewaing", "lang"=>"en", "lat"=>43.731772, "lon"=>-83.451018),
"/US/US-MI/shelby-10056" => array("type"=>"cdp", "pop"=>1885, "name_ru"=>"Шелби", "name_en"=>"Shelby", "lang"=>"en", "lat"=>42.670801, "lon"=>-83.033075),
"/US/US-MI/shelby-10057" => array("type"=>"village", "pop"=>1871, "name_ru"=>"Шелби", "name_en"=>"Shelby", "lang"=>"en", "lat"=>43.611089, "lon"=>-86.365229),
"/US/US-MI/shepherd-10058" => array("type"=>"village", "pop"=>1353, "name_ru"=>"Шеперд", "name_en"=>"Shepherd", "lang"=>"en", "lat"=>43.524579, "lon"=>-84.693847),
"/US/US-MI/sheridan-10059" => array("type"=>"village", "pop"=>694, "name_ru"=>"Шеридан", "name_en"=>"Sheridan", "lang"=>"en", "lat"=>43.210439, "lon"=>-85.072649),
"/US/US-MI/sherwood-10060" => array("type"=>"village", "pop"=>308, "name_ru"=>"Шервуд", "name_en"=>"Sherwood", "lang"=>"en", "lat"=>41.999990, "lon"=>-85.239985),
"/US/US-MI/shields-10061" => array("type"=>"cdp", "pop"=>6634, "name_ru"=>"Шилдс", "name_en"=>"Shields", "lang"=>"en", "lat"=>43.417656, "lon"=>-84.073195),
"/US/US-MI/shoreham-10062" => array("type"=>"village", "pop"=>823, "name_ru"=>"Шорхэм", "name_en"=>"Shoreham", "lang"=>"en", "lat"=>42.057700, "lon"=>-86.511647),
"/US/US-MI/shorewood-tower-hills-harbert-10063" => array("type"=>"cdp", "pop"=>1636, "name_ru"=>"Шорвуд-Тауэр Хилс-Харберт", "name_en"=>"Shorewood-Tower Hills-Harbert", "lang"=>"en", "lat"=>41.884532, "lon"=>-86.620411),
"/US/US-MI/skidway-lake-10064" => array("type"=>"cdp", "pop"=>2569, "name_ru"=>"Скидуэй Лейк", "name_en"=>"Skidway Lake", "lang"=>"en", "lat"=>44.193995, "lon"=>-84.044639),
"/US/US-MI/south-gull-lake-10067" => array("type"=>"cdp", "pop"=>1453, "name_ru"=>"Саут Галл Лейк", "name_en"=>"South Gull Lake", "lang"=>"en", "lat"=>42.387532, "lon"=>-85.396686),
"/US/US-MI/south-haven-10068" => array("type"=>"city", "pop"=>5138, "name_ru"=>"Саут Хейвен", "name_en"=>"South Haven", "lang"=>"en", "lat"=>42.400374, "lon"=>-86.272547),
"/US/US-MI/south-lyon-10069" => array("type"=>"city", "pop"=>10959, "name_ru"=>"Саут Лайон", "name_en"=>"South Lyon", "lang"=>"en", "lat"=>42.460734, "lon"=>-83.651291),
"/US/US-MI/south-monroe-10070" => array("type"=>"cdp", "pop"=>5266, "name_ru"=>"Саут Монро", "name_en"=>"South Monroe", "lang"=>"en", "lat"=>41.895798, "lon"=>-83.417707),
"/US/US-MI/south-range-10071" => array("type"=>"village", "pop"=>688, "name_ru"=>"Саут Рэндж", "name_en"=>"South Range", "lang"=>"en", "lat"=>47.070348, "lon"=>-88.644612),
"/US/US-MI/south-rockwood-10072" => array("type"=>"village", "pop"=>1654, "name_ru"=>"Саут Роквуд", "name_en"=>"South Rockwood", "lang"=>"en", "lat"=>42.061972, "lon"=>-83.260596),
"/US/US-MI/southfield-10065" => array("type"=>"city", "pop"=>75392, "name_ru"=>"Саут филд", "name_en"=>"Southfield", "lang"=>"en", "lat"=>42.475600, "lon"=>-83.259196),
"/US/US-MI/southgate-10066" => array("type"=>"city", "pop"=>27739, "name_ru"=>"Саутгейт", "name_en"=>"Southgate", "lang"=>"en", "lat"=>42.203575, "lon"=>-83.206206),
"/US/US-MI/sparlingville-10073" => array("type"=>"cdp", "pop"=>1974, "name_ru"=>"Спарлингвилл", "name_en"=>"Sparlingville", "lang"=>"en", "lat"=>42.960358, "lon"=>-82.525870),
"/US/US-MI/sparta-10074" => array("type"=>"village", "pop"=>4054, "name_ru"=>"Спарта", "name_en"=>"Sparta", "lang"=>"en", "lat"=>43.157910, "lon"=>-85.709358),
"/US/US-MI/spring-arbor-10075" => array("type"=>"cdp", "pop"=>8324, "name_ru"=>"Спринг Арбор", "name_en"=>"Spring Arbor", "lang"=>"en", "lat"=>42.206550, "lon"=>-84.555918),
"/US/US-MI/spring-lake-10077" => array("type"=>"village", "pop"=>2420, "name_ru"=>"Спринг Лейк", "name_en"=>"Spring Lake", "lang"=>"en", "lat"=>43.074792, "lon"=>-86.190485),
"/US/US-MI/springfield-10076" => array("type"=>"city", "pop"=>5037, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>42.323984, "lon"=>-85.237628),
"/US/US-MI/springport-10078" => array("type"=>"village", "pop"=>653, "name_ru"=>"Спрингпорт", "name_en"=>"Springport", "lang"=>"en", "lat"=>42.377804, "lon"=>-84.697330),
"/US/US-MI/st.-charles-10038" => array("type"=>"village", "pop"=>2016, "name_ru"=>"Сент-Чарльз", "name_en"=>"St. Charles", "lang"=>"en", "lat"=>43.299401, "lon"=>-84.149661),
"/US/US-MI/st.-clair-10039" => array("type"=>"city", "pop"=>5763, "name_ru"=>"Сент Клэр", "name_en"=>"St. Clair", "lang"=>"en", "lat"=>42.824738, "lon"=>-82.492915),
"/US/US-MI/st.-clair-shores-10040" => array("type"=>"city", "pop"=>60364, "name_ru"=>"Сент Клэр Шорс", "name_en"=>"St. Clair Shores", "lang"=>"en", "lat"=>42.493100, "lon"=>-82.891134),
"/US/US-MI/st.-helen-10041" => array("type"=>"cdp", "pop"=>2390, "name_ru"=>"Сент Хэлен", "name_en"=>"St. Helen", "lang"=>"en", "lat"=>44.362844, "lon"=>-84.418600),
"/US/US-MI/st.-ignace-10042" => array("type"=>"city", "pop"=>2335, "name_ru"=>"Сент Игнас", "name_en"=>"St. Ignace", "lang"=>"en", "lat"=>45.868396, "lon"=>-84.725655),
"/US/US-MI/st.-johns-10043" => array("type"=>"city", "pop"=>7278, "name_ru"=>"Сент-Джонс", "name_en"=>"St. Johns", "lang"=>"en", "lat"=>43.000157, "lon"=>-84.557514),
"/US/US-MI/st.-joseph-10044" => array("type"=>"city", "pop"=>8458, "name_ru"=>"Сент Джозеф", "name_en"=>"St. Joseph", "lang"=>"en", "lat"=>42.096850, "lon"=>-86.491273),
"/US/US-MI/st.-louis-10045" => array("type"=>"city", "pop"=>6971, "name_ru"=>"Сент Луис", "name_en"=>"St. Louis", "lang"=>"en", "lat"=>43.409271, "lon"=>-84.615744),
"/US/US-MI/stambaugh-10079" => array("type"=>"city", "pop"=>1146, "name_ru"=>"Стэмбо", "name_en"=>"Stambaugh", "lang"=>"en", "lat"=>46.079203, "lon"=>-88.633405),
"/US/US-MI/standish-10080" => array("type"=>"city", "pop"=>1981, "name_ru"=>"Стэндиш", "name_en"=>"Standish", "lang"=>"en", "lat"=>43.980581, "lon"=>-83.963366),
"/US/US-MI/stanton-10081" => array("type"=>"city", "pop"=>1483, "name_ru"=>"Стэнтон", "name_en"=>"Stanton", "lang"=>"en", "lat"=>43.293058, "lon"=>-85.079309),
"/US/US-MI/stanwood-10082" => array("type"=>"village", "pop"=>198, "name_ru"=>"Стэнвуд", "name_en"=>"Stanwood", "lang"=>"en", "lat"=>43.580124, "lon"=>-85.447867),
"/US/US-MI/stephenson-10083" => array("type"=>"city", "pop"=>803, "name_ru"=>"Стивенсон", "name_en"=>"Stephenson", "lang"=>"en", "lat"=>45.413679, "lon"=>-87.608931),
"/US/US-MI/sterling-10084" => array("type"=>"village", "pop"=>485, "name_ru"=>"Стерлинг", "name_en"=>"Sterling", "lang"=>"en", "lat"=>44.032125, "lon"=>-84.020315),
"/US/US-MI/sterling-heights-10085" => array("type"=>"city", "pop"=>127160, "name_ru"=>"Стерлинг Хайтс", "name_en"=>"Sterling Heights", "lang"=>"en", "lat"=>42.580035, "lon"=>-83.030534),
"/US/US-MI/stevensville-10086" => array("type"=>"village", "pop"=>1131, "name_ru"=>"Стивенсвилл", "name_en"=>"Stevensville", "lang"=>"en", "lat"=>42.013850, "lon"=>-86.525224),
"/US/US-MI/stockbridge-10087" => array("type"=>"village", "pop"=>1274, "name_ru"=>"Стокбридж", "name_en"=>"Stockbridge", "lang"=>"en", "lat"=>42.450750, "lon"=>-84.178299),
"/US/US-MI/stony-point-10088" => array("type"=>"cdp", "pop"=>1598, "name_ru"=>"Стоуни Пойнт", "name_en"=>"Stony Point", "lang"=>"en", "lat"=>41.947151, "lon"=>-83.271417),
"/US/US-MI/sturgis-10089" => array("type"=>"city", "pop"=>10916, "name_ru"=>"Стерджис", "name_en"=>"Sturgis", "lang"=>"en", "lat"=>41.801662, "lon"=>-85.421199),
"/US/US-MI/sunfield-10090" => array("type"=>"village", "pop"=>583, "name_ru"=>"Санфилд", "name_en"=>"Sunfield", "lang"=>"en", "lat"=>42.761730, "lon"=>-84.991736),
"/US/US-MI/suttons-bay-10091" => array("type"=>"village", "pop"=>578, "name_ru"=>"Саттонс Бэй", "name_en"=>"Suttons Bay", "lang"=>"en", "lat"=>44.978980, "lon"=>-85.651460),
"/US/US-MI/swartz-creek-10092" => array("type"=>"city", "pop"=>5223, "name_ru"=>"Свортс Крик", "name_en"=>"Swartz Creek", "lang"=>"en", "lat"=>42.962100, "lon"=>-83.826461),
"/US/US-MI/sylvan-lake-10093" => array("type"=>"city", "pop"=>1639, "name_ru"=>"Сильван Лейк", "name_en"=>"Sylvan Lake", "lang"=>"en", "lat"=>42.617067, "lon"=>-83.332945),
"/US/US-MI/tawas-city-10094" => array("type"=>"city", "pop"=>1873, "name_ru"=>"Тавас Сити", "name_en"=>"Tawas City", "lang"=>"en", "lat"=>44.267358, "lon"=>-83.521996),
"/US/US-MI/taylor-10095" => array("type"=>"city", "pop"=>60619, "name_ru"=>"Тейлор", "name_en"=>"Taylor", "lang"=>"en", "lat"=>42.225251, "lon"=>-83.268448),
"/US/US-MI/tecumseh-10096" => array("type"=>"city", "pop"=>8610, "name_ru"=>"Текамсе", "name_en"=>"Tecumseh", "lang"=>"en", "lat"=>42.006076, "lon"=>-83.944478),
"/US/US-MI/tekonsha-10097" => array("type"=>"village", "pop"=>679, "name_ru"=>"Теконша", "name_en"=>"Tekonsha", "lang"=>"en", "lat"=>42.094718, "lon"=>-84.989280),
"/US/US-MI/temperance-10098" => array("type"=>"cdp", "pop"=>6542, "name_ru"=>"Темперанс", "name_en"=>"Temperance", "lang"=>"en", "lat"=>41.767600, "lon"=>-83.572300),
"/US/US-MI/thompsonville-10099" => array("type"=>"village", "pop"=>440, "name_ru"=>"Томпсонвилл", "name_en"=>"Thompsonville", "lang"=>"en", "lat"=>44.520749, "lon"=>-85.938742),
"/US/US-MI/three-oaks-10100" => array("type"=>"village", "pop"=>1695, "name_ru"=>"Три Оукс", "name_en"=>"Three Oaks", "lang"=>"en", "lat"=>41.799329, "lon"=>-86.612736),
"/US/US-MI/three-rivers-10101" => array("type"=>"city", "pop"=>7179, "name_ru"=>"Три Риверс", "name_en"=>"Three Rivers", "lang"=>"en", "lat"=>41.944075, "lon"=>-85.632125),
"/US/US-MI/traverse-city-10102" => array("type"=>"city", "pop"=>14398, "name_ru"=>"Трэверс Сити", "name_en"=>"Traverse City", "lang"=>"en", "lat"=>44.754428, "lon"=>-85.602895),
"/US/US-MI/trenton-10103" => array("type"=>"city", "pop"=>17965, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>42.139950, "lon"=>-83.192952),
"/US/US-MI/trowbridge-park-10104" => array("type"=>"cdp", "pop"=>1831, "name_ru"=>"Троубридж Парк", "name_en"=>"Trowbridge Park", "lang"=>"en", "lat"=>46.555717, "lon"=>-87.441086),
"/US/US-MI/troy-10105" => array("type"=>"city", "pop"=>80264, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>42.580161, "lon"=>-83.143050),
"/US/US-MI/turner-10106" => array("type"=>"village", "pop"=>125, "name_ru"=>"Тернер", "name_en"=>"Turner", "lang"=>"en", "lat"=>44.141769, "lon"=>-83.788197),
"/US/US-MI/tustin-10107" => array("type"=>"village", "pop"=>226, "name_ru"=>"Тастин", "name_en"=>"Tustin", "lang"=>"en", "lat"=>44.101991, "lon"=>-85.458103),
"/US/US-MI/twin-lake-10109" => array("type"=>"cdp", "pop"=>1328, "name_ru"=>"Твин Лейк", "name_en"=>"Twin Lake", "lang"=>"en", "lat"=>43.369039, "lon"=>-86.179904),
"/US/US-MI/twining-10108" => array("type"=>"village", "pop"=>172, "name_ru"=>"Твайнинг", "name_en"=>"Twining", "lang"=>"en", "lat"=>44.112637, "lon"=>-83.807823),
"/US/US-MI/ubly-10110" => array("type"=>"village", "pop"=>771, "name_ru"=>"Абли", "name_en"=>"Ubly", "lang"=>"en", "lat"=>43.710872, "lon"=>-82.934609),
"/US/US-MI/union-city-10111" => array("type"=>"village", "pop"=>1717, "name_ru"=>"Юнион Сити", "name_en"=>"Union City", "lang"=>"en", "lat"=>42.066303, "lon"=>-85.140537),
"/US/US-MI/unionville-10112" => array("type"=>"village", "pop"=>567, "name_ru"=>"Юнионвилл", "name_en"=>"Unionville", "lang"=>"en", "lat"=>43.654115, "lon"=>-83.466936),
"/US/US-MI/utica-10113" => array("type"=>"city", "pop"=>4924, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>42.628400, "lon"=>-83.020899),
"/US/US-MI/vandalia-10114" => array("type"=>"village", "pop"=>392, "name_ru"=>"Вандалия", "name_en"=>"Vandalia", "lang"=>"en", "lat"=>41.918402, "lon"=>-85.914957),
"/US/US-MI/vanderbilt-10115" => array("type"=>"village", "pop"=>571, "name_ru"=>"Вандербилт", "name_en"=>"Vanderbilt", "lang"=>"en", "lat"=>45.143507, "lon"=>-84.663499),
"/US/US-MI/vandercook-lake-10116" => array("type"=>"cdp", "pop"=>4642, "name_ru"=>"Вандеркук Лейк", "name_en"=>"Vandercook Lake", "lang"=>"en", "lat"=>42.191600, "lon"=>-84.385425),
"/US/US-MI/vassar-10117" => array("type"=>"city", "pop"=>2665, "name_ru"=>"Вассар", "name_en"=>"Vassar", "lang"=>"en", "lat"=>43.371530, "lon"=>-83.578680),
"/US/US-MI/vermontville-10118" => array("type"=>"village", "pop"=>785, "name_ru"=>"Вермонтвилл", "name_en"=>"Vermontville", "lang"=>"en", "lat"=>42.626021, "lon"=>-85.026110),
"/US/US-MI/vernon-10119" => array("type"=>"village", "pop"=>790, "name_ru"=>"Вернон", "name_en"=>"Vernon", "lang"=>"en", "lat"=>42.939672, "lon"=>-84.029936),
"/US/US-MI/vicksburg-10120" => array("type"=>"village", "pop"=>2194, "name_ru"=>"Виксбург", "name_en"=>"Vicksburg", "lang"=>"en", "lat"=>42.120436, "lon"=>-85.534191),
"/US/US-MI/wakefield-10121" => array("type"=>"city", "pop"=>1846, "name_ru"=>"Уэйкфилд", "name_en"=>"Wakefield", "lang"=>"en", "lat"=>46.476750, "lon"=>-89.933433),
"/US/US-MI/waldron-10122" => array("type"=>"village", "pop"=>560, "name_ru"=>"Уолдрон", "name_en"=>"Waldron", "lang"=>"en", "lat"=>41.724557, "lon"=>-84.418201),
"/US/US-MI/walker-10123" => array("type"=>"city", "pop"=>23722, "name_ru"=>"Уолкер", "name_en"=>"Walker", "lang"=>"en", "lat"=>42.984700, "lon"=>-85.747674),
"/US/US-MI/walkerville-10124" => array("type"=>"village", "pop"=>253, "name_ru"=>"Уолкервилл", "name_en"=>"Walkerville", "lang"=>"en", "lat"=>43.714738, "lon"=>-86.125725),
"/US/US-MI/walled-lake-10125" => array("type"=>"city", "pop"=>6876, "name_ru"=>"Уоллд Лейк", "name_en"=>"Walled Lake", "lang"=>"en", "lat"=>42.533924, "lon"=>-83.478582),
"/US/US-MI/warren-10126" => array("type"=>"city", "pop"=>133939, "name_ru"=>"Уоррен", "name_en"=>"Warren", "lang"=>"en", "lat"=>42.493000, "lon"=>-83.028197),
"/US/US-MI/waterford-10127" => array("type"=>"cdp", "pop"=>66692, "name_ru"=>"Уотерфорд", "name_en"=>"Waterford", "lang"=>"en", "lat"=>42.662934, "lon"=>-83.387965),
"/US/US-MI/watervliet-10128" => array("type"=>"city", "pop"=>1737, "name_ru"=>"Уотервлит", "name_en"=>"Watervliet", "lang"=>"en", "lat"=>42.186641, "lon"=>-86.257159),
"/US/US-MI/waverly-10129" => array("type"=>"cdp", "pop"=>507, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>42.739200, "lon"=>-84.620756),
"/US/US-MI/wayland-10130" => array("type"=>"city", "pop"=>3809, "name_ru"=>"Уэйленд", "name_en"=>"Wayland", "lang"=>"en", "lat"=>42.671932, "lon"=>-85.641398),
"/US/US-MI/wayne-10131" => array("type"=>"city", "pop"=>17186, "name_ru"=>"Уэйн", "name_en"=>"Wayne", "lang"=>"en", "lat"=>42.276750, "lon"=>-83.388204),
"/US/US-MI/webberville-10132" => array("type"=>"village", "pop"=>1480, "name_ru"=>"Уэббервилл", "name_en"=>"Webberville", "lang"=>"en", "lat"=>42.667440, "lon"=>-84.180198),
"/US/US-MI/weidman-10133" => array("type"=>"cdp", "pop"=>696, "name_ru"=>"Уидман", "name_en"=>"Weidman", "lang"=>"en", "lat"=>43.694034, "lon"=>-84.972941),
"/US/US-MI/west-bloomfield-township-10134" => array("type"=>"cdp", "pop"=>54843, "name_ru"=>"Вэст Блумфилд Тауншип", "name_en"=>"West Bloomfield Township", "lang"=>"en", "lat"=>42.569214, "lon"=>-83.382149),
"/US/US-MI/west-branch-10135" => array("type"=>"city", "pop"=>1816, "name_ru"=>"Вэст Брэнч", "name_en"=>"West Branch", "lang"=>"en", "lat"=>44.275656, "lon"=>-84.235311),
"/US/US-MI/west-monroe-10137" => array("type"=>"cdp", "pop"=>3919, "name_ru"=>"Вэст Монро", "name_en"=>"West Monroe", "lang"=>"en", "lat"=>41.913760, "lon"=>-83.431656),
"/US/US-MI/westland-10136" => array("type"=>"city", "pop"=>78961, "name_ru"=>"Вэстленд", "name_en"=>"Westland", "lang"=>"en", "lat"=>42.318950, "lon"=>-83.380923),
"/US/US-MI/westphalia-10138" => array("type"=>"village", "pop"=>816, "name_ru"=>"Вэстфалия", "name_en"=>"Westphalia", "lang"=>"en", "lat"=>42.929696, "lon"=>-84.797843),
"/US/US-MI/westwood-10139" => array("type"=>"cdp", "pop"=>8957, "name_ru"=>"Уэствуд", "name_en"=>"Westwood", "lang"=>"en", "lat"=>42.303450, "lon"=>-85.630099),
"/US/US-MI/white-cloud-10140" => array("type"=>"city", "pop"=>1390, "name_ru"=>"Уайт Клауд", "name_en"=>"White Cloud", "lang"=>"en", "lat"=>43.552923, "lon"=>-85.772648),
"/US/US-MI/white-pigeon-10142" => array("type"=>"village", "pop"=>1559, "name_ru"=>"Уайт Пиджен", "name_en"=>"White Pigeon", "lang"=>"en", "lat"=>41.798368, "lon"=>-85.642793),
"/US/US-MI/whitehall-10141" => array("type"=>"city", "pop"=>2803, "name_ru"=>"Уайтхолл", "name_en"=>"Whitehall", "lang"=>"en", "lat"=>43.398308, "lon"=>-86.341262),
"/US/US-MI/whitmore-lake-10143" => array("type"=>"cdp", "pop"=>3251, "name_ru"=>"Уитмор Лейк", "name_en"=>"Whitmore Lake", "lang"=>"en", "lat"=>42.429114, "lon"=>-83.745407),
"/US/US-MI/whittemore-10144" => array("type"=>"city", "pop"=>440, "name_ru"=>"Уиттмор", "name_en"=>"Whittemore", "lang"=>"en", "lat"=>44.233111, "lon"=>-83.802800),
"/US/US-MI/williamston-10145" => array("type"=>"city", "pop"=>3765, "name_ru"=>"Уилльямстон", "name_en"=>"Williamston", "lang"=>"en", "lat"=>42.689824, "lon"=>-84.282191),
"/US/US-MI/wixom-10146" => array("type"=>"city", "pop"=>13425, "name_ru"=>"Уиксом", "name_en"=>"Wixom", "lang"=>"en", "lat"=>42.524105, "lon"=>-83.535500),
"/US/US-MI/wolf-lake-10147" => array("type"=>"cdp", "pop"=>4110, "name_ru"=>"Вулф Лейк", "name_en"=>"Wolf Lake", "lang"=>"en", "lat"=>43.245472, "lon"=>-86.105879),
"/US/US-MI/wolverine-10148" => array("type"=>"village", "pop"=>335, "name_ru"=>"Вулверин", "name_en"=>"Wolverine", "lang"=>"en", "lat"=>45.273487, "lon"=>-84.605484),
"/US/US-MI/wolverine-lake-10149" => array("type"=>"village", "pop"=>4303, "name_ru"=>"Вулверин Лейк", "name_en"=>"Wolverine Lake", "lang"=>"en", "lat"=>42.556717, "lon"=>-83.484431),
"/US/US-MI/woodhaven-10150" => array("type"=>"city", "pop"=>12634, "name_ru"=>"Вудхейвен", "name_en"=>"Woodhaven", "lang"=>"en", "lat"=>42.132350, "lon"=>-83.236719),
"/US/US-MI/woodland-10151" => array("type"=>"village", "pop"=>488, "name_ru"=>"Вудленд", "name_en"=>"Woodland", "lang"=>"en", "lat"=>42.726493, "lon"=>-85.134864),
"/US/US-MI/woodland-beach-10152" => array("type"=>"cdp", "pop"=>2309, "name_ru"=>"Вудленд Бич", "name_en"=>"Woodland Beach", "lang"=>"en", "lat"=>41.941713, "lon"=>-83.314301),
"/US/US-MI/wurtsmith-afb-10153" => array("type"=>"cdp", "pop"=>5080, "name_ru"=>"Военно-воздушная база Уертсмит", "name_en"=>"Wurtsmith AFB", "lang"=>"en", "lat"=>44.461210, "lon"=>-83.384458),
"/US/US-MI/wyandotte-10154" => array("type"=>"city", "pop"=>24739, "name_ru"=>"Уайандотт", "name_en"=>"Wyandotte", "lang"=>"en", "lat"=>42.211433, "lon"=>-83.156628),
"/US/US-MI/wyoming-10155" => array("type"=>"city", "pop"=>70462, "name_ru"=>"Вайоминг", "name_en"=>"Wyoming", "lang"=>"en", "lat"=>42.892350, "lon"=>-85.702524),
"/US/US-MI/yale-10156" => array("type"=>"city", "pop"=>1956, "name_ru"=>"Йель", "name_en"=>"Yale", "lang"=>"en", "lat"=>43.128040, "lon"=>-82.796357),
"/US/US-MI/ypsilanti-10157" => array("type"=>"city", "pop"=>21464, "name_ru"=>"Ипсиланти", "name_en"=>"Ypsilanti", "lang"=>"en", "lat"=>42.244200, "lon"=>-83.621106),
"/US/US-MI/zeeland-10158" => array("type"=>"city", "pop"=>5448, "name_ru"=>"Зиленд", "name_en"=>"Zeeland", "lang"=>"en", "lat"=>42.813124, "lon"=>-86.013121),
"/US/US-MI/zilwaukee-10159" => array("type"=>"city", "pop"=>1651, "name_ru"=>"Зилуоки", "name_en"=>"Zilwaukee", "lang"=>"en", "lat"=>43.480050, "lon"=>-83.921849),
"/US/US-MN" => array("type"=>"state", "pop"=>5220393, "name_ru"=>"Миннесота", "name_en"=>"Minnesota", "lang"=>"en"),
"/US/US-MN/ada-10160" => array("type"=>"city", "pop"=>1462, "name_ru"=>"Ада", "name_en"=>"Ada", "lang"=>"en", "lat"=>47.299518, "lon"=>-96.513929),
"/US/US-MN/adams-10161" => array("type"=>"city", "pop"=>755, "name_ru"=>"Адамс", "name_en"=>"Adams", "lang"=>"en", "lat"=>43.564995, "lon"=>-92.718455),
"/US/US-MN/adrian-10162" => array("type"=>"city", "pop"=>1216, "name_ru"=>"Эдриан", "name_en"=>"Adrian", "lang"=>"en", "lat"=>43.634563, "lon"=>-95.932407),
"/US/US-MN/afton-10163" => array("type"=>"city", "pop"=>2839, "name_ru"=>"Афтон", "name_en"=>"Afton", "lang"=>"en", "lat"=>44.903166, "lon"=>-92.818844),
"/US/US-MN/aitkin-10164" => array("type"=>"city", "pop"=>1948, "name_ru"=>"Айткин", "name_en"=>"Aitkin", "lang"=>"en", "lat"=>46.530125, "lon"=>-93.710415),
"/US/US-MN/akeley-10165" => array("type"=>"city", "pop"=>412, "name_ru"=>"Акели", "name_en"=>"Akeley", "lang"=>"en", "lat"=>47.000362, "lon"=>-94.727579),
"/US/US-MN/albany-10166" => array("type"=>"city", "pop"=>2113, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>45.630920, "lon"=>-94.568651),
"/US/US-MN/albert-lea-10168" => array("type"=>"city", "pop"=>17389, "name_ru"=>"Алберт Ли", "name_en"=>"Albert Lea", "lang"=>"en", "lat"=>43.655050, "lon"=>-93.366569),
"/US/US-MN/alberta-10167" => array("type"=>"city", "pop"=>136, "name_ru"=>"Алберта", "name_en"=>"Alberta", "lang"=>"en", "lat"=>45.574763, "lon"=>-96.050315),
"/US/US-MN/albertville-10169" => array("type"=>"city", "pop"=>6360, "name_ru"=>"Альбертивилл", "name_en"=>"Albertville", "lang"=>"en", "lat"=>45.239200, "lon"=>-93.660018),
"/US/US-MN/alden-10170" => array("type"=>"city", "pop"=>612, "name_ru"=>"Олден", "name_en"=>"Alden", "lang"=>"en", "lat"=>43.670138, "lon"=>-93.573239),
"/US/US-MN/aldrich-10171" => array("type"=>"city", "pop"=>47, "name_ru"=>"Олдрич", "name_en"=>"Aldrich", "lang"=>"en", "lat"=>46.374743, "lon"=>-94.939431),
"/US/US-MN/alexandria-10172" => array("type"=>"city", "pop"=>11237, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>45.874047, "lon"=>-95.375140),
"/US/US-MN/alpha-10173" => array("type"=>"city", "pop"=>116, "name_ru"=>"Альфа", "name_en"=>"Alpha", "lang"=>"en", "lat"=>43.639189, "lon"=>-94.871384),
"/US/US-MN/altura-10174" => array("type"=>"city", "pop"=>421, "name_ru"=>"Олтура", "name_en"=>"Altura", "lang"=>"en", "lat"=>44.069823, "lon"=>-91.938942),
"/US/US-MN/alvarado-10175" => array("type"=>"city", "pop"=>341, "name_ru"=>"Альварадо", "name_en"=>"Alvarado", "lang"=>"en", "lat"=>48.192551, "lon"=>-96.998257),
"/US/US-MN/amboy-10176" => array("type"=>"city", "pop"=>537, "name_ru"=>"Амбой", "name_en"=>"Amboy", "lang"=>"en", "lat"=>43.888464, "lon"=>-94.157303),
"/US/US-MN/andover-10177" => array("type"=>"city", "pop"=>30442, "name_ru"=>"Андоувер", "name_en"=>"Andover", "lang"=>"en", "lat"=>45.257533, "lon"=>-93.331116),
"/US/US-MN/annandale-10178" => array("type"=>"city", "pop"=>3100, "name_ru"=>"Аннандейл", "name_en"=>"Annandale", "lang"=>"en", "lat"=>45.262219, "lon"=>-94.118894),
"/US/US-MN/anoka-10179" => array("type"=>"city", "pop"=>17289, "name_ru"=>"Анока", "name_en"=>"Anoka", "lang"=>"en", "lat"=>45.210400, "lon"=>-93.390070),
"/US/US-MN/apple-valley-10181" => array("type"=>"city", "pop"=>50004, "name_ru"=>"Апл Валли", "name_en"=>"Apple Valley", "lang"=>"en", "lat"=>44.749150, "lon"=>-93.197454),
"/US/US-MN/appleton-10180" => array("type"=>"city", "pop"=>2950, "name_ru"=>"Эпплтон", "name_en"=>"Appleton", "lang"=>"en", "lat"=>45.201447, "lon"=>-96.022933),
"/US/US-MN/arco-10182" => array("type"=>"city", "pop"=>85, "name_ru"=>"Арко", "name_en"=>"Arco", "lang"=>"en", "lat"=>44.383492, "lon"=>-96.180814),
"/US/US-MN/arden-hills-10183" => array("type"=>"city", "pop"=>9608, "name_ru"=>"Арден Хилс", "name_en"=>"Arden Hills", "lang"=>"en", "lat"=>45.071442, "lon"=>-93.165405),
"/US/US-MN/argyle-10184" => array("type"=>"city", "pop"=>599, "name_ru"=>"Аргайл", "name_en"=>"Argyle", "lang"=>"en", "lat"=>48.337383, "lon"=>-96.815705),
"/US/US-MN/arlington-10185" => array("type"=>"city", "pop"=>2028, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>44.607147, "lon"=>-94.077936),
"/US/US-MN/arnold-10186" => array("type"=>"cdp", "pop"=>2891, "name_ru"=>"Арнолд", "name_en"=>"Arnold", "lang"=>"en", "lat"=>46.875067, "lon"=>-92.107844),
"/US/US-MN/ashby-10187" => array("type"=>"city", "pop"=>478, "name_ru"=>"Эшби", "name_en"=>"Ashby", "lang"=>"en", "lat"=>46.093236, "lon"=>-95.816082),
"/US/US-MN/askov-10188" => array("type"=>"city", "pop"=>356, "name_ru"=>"Асков", "name_en"=>"Askov", "lang"=>"en", "lat"=>46.188706, "lon"=>-92.781511),
"/US/US-MN/atwater-10189" => array("type"=>"city", "pop"=>1027, "name_ru"=>"Атуотер", "name_en"=>"Atwater", "lang"=>"en", "lat"=>45.135474, "lon"=>-94.776862),
"/US/US-MN/audubon-10190" => array("type"=>"city", "pop"=>445, "name_ru"=>"Одубон", "name_en"=>"Audubon", "lang"=>"en", "lat"=>46.864168, "lon"=>-95.980888),
"/US/US-MN/aurora-10191" => array("type"=>"city", "pop"=>1746, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>47.531973, "lon"=>-92.240044),
"/US/US-MN/austin-10192" => array("type"=>"city", "pop"=>22800, "name_ru"=>"Остин", "name_en"=>"Austin", "lang"=>"en", "lat"=>43.671518, "lon"=>-92.970636),
"/US/US-MN/avoca-10193" => array("type"=>"city", "pop"=>130, "name_ru"=>"Авока", "name_en"=>"Avoca", "lang"=>"en", "lat"=>43.949488, "lon"=>-95.645347),
"/US/US-MN/avon-10194" => array("type"=>"city", "pop"=>1315, "name_ru"=>"Эйвон", "name_en"=>"Avon", "lang"=>"en", "lat"=>45.608798, "lon"=>-94.451538),
"/US/US-MN/babbitt-10195" => array("type"=>"city", "pop"=>1587, "name_ru"=>"Бэббит", "name_en"=>"Babbitt", "lang"=>"en", "lat"=>47.645110, "lon"=>-91.940056),
"/US/US-MN/backus-10196" => array("type"=>"city", "pop"=>297, "name_ru"=>"Бэкус", "name_en"=>"Backus", "lang"=>"en", "lat"=>46.820918, "lon"=>-94.514141),
"/US/US-MN/badger-10197" => array("type"=>"city", "pop"=>459, "name_ru"=>"Баджер", "name_en"=>"Badger", "lang"=>"en", "lat"=>48.776088, "lon"=>-96.020713),
"/US/US-MN/bagley-10198" => array("type"=>"city", "pop"=>1288, "name_ru"=>"Багли", "name_en"=>"Bagley", "lang"=>"en", "lat"=>47.523436, "lon"=>-95.403619),
"/US/US-MN/balaton-10199" => array("type"=>"city", "pop"=>632, "name_ru"=>"Балатон", "name_en"=>"Balaton", "lang"=>"en", "lat"=>44.233288, "lon"=>-95.871075),
"/US/US-MN/barnesville-10200" => array("type"=>"city", "pop"=>2308, "name_ru"=>"Барнсвилл", "name_en"=>"Barnesville", "lang"=>"en", "lat"=>46.648857, "lon"=>-96.417554),
"/US/US-MN/barnum-10201" => array("type"=>"city", "pop"=>595, "name_ru"=>"Барнум", "name_en"=>"Barnum", "lang"=>"en", "lat"=>46.504776, "lon"=>-92.689303),
"/US/US-MN/barrett-10202" => array("type"=>"city", "pop"=>344, "name_ru"=>"Барретт", "name_en"=>"Barrett", "lang"=>"en", "lat"=>45.910743, "lon"=>-95.891352),
"/US/US-MN/barry-10203" => array("type"=>"city", "pop"=>22, "name_ru"=>"Бэрри", "name_en"=>"Barry", "lang"=>"en", "lat"=>45.558469, "lon"=>-96.559727),
"/US/US-MN/battle-lake-10204" => array("type"=>"city", "pop"=>772, "name_ru"=>"Бэттл Лейк", "name_en"=>"Battle Lake", "lang"=>"en", "lat"=>46.288832, "lon"=>-95.715122),
"/US/US-MN/baudette-10205" => array("type"=>"city", "pop"=>925, "name_ru"=>"Бодетт", "name_en"=>"Baudette", "lang"=>"en", "lat"=>48.711171, "lon"=>-94.587017),
"/US/US-MN/baxter-10206" => array("type"=>"city", "pop"=>8271, "name_ru"=>"Бэкстер", "name_en"=>"Baxter", "lang"=>"en", "lat"=>46.341221, "lon"=>-94.282414),
"/US/US-MN/bayport-10207" => array("type"=>"city", "pop"=>3291, "name_ru"=>"Бэйпорт", "name_en"=>"Bayport", "lang"=>"en", "lat"=>45.017800, "lon"=>-92.780189),
"/US/US-MN/beardsley-10208" => array("type"=>"city", "pop"=>230, "name_ru"=>"Бирдсли", "name_en"=>"Beardsley", "lang"=>"en", "lat"=>45.557806, "lon"=>-96.714191),
"/US/US-MN/beaver-bay-10209" => array("type"=>"city", "pop"=>170, "name_ru"=>"Бивер Бэй", "name_en"=>"Beaver Bay", "lang"=>"en", "lat"=>47.256919, "lon"=>-91.298619),
"/US/US-MN/beaver-creek-10210" => array("type"=>"city", "pop"=>260, "name_ru"=>"Бивер Крик", "name_en"=>"Beaver Creek", "lang"=>"en", "lat"=>43.612602, "lon"=>-96.361775),
"/US/US-MN/becker-10211" => array("type"=>"city", "pop"=>4176, "name_ru"=>"Бекер", "name_en"=>"Becker", "lang"=>"en", "lat"=>45.371341, "lon"=>-93.873398),
"/US/US-MN/bejou-10212" => array("type"=>"city", "pop"=>93, "name_ru"=>"Бежу", "name_en"=>"Bejou", "lang"=>"en", "lat"=>47.441760, "lon"=>-95.971973),
"/US/US-MN/belgrade-10213" => array("type"=>"city", "pop"=>714, "name_ru"=>"Белгрейд", "name_en"=>"Belgrade", "lang"=>"en", "lat"=>45.449971, "lon"=>-94.999034),
"/US/US-MN/belle-plaine-10215" => array("type"=>"city", "pop"=>6648, "name_ru"=>"Бель Плейн", "name_en"=>"Belle Plaine", "lang"=>"en", "lat"=>44.623600, "lon"=>-93.764199),
"/US/US-MN/bellechester-10214" => array("type"=>"city", "pop"=>163, "name_ru"=>"Беллечестер", "name_en"=>"Bellechester", "lang"=>"en", "lat"=>44.370862, "lon"=>-92.511746),
"/US/US-MN/bellingham-10216" => array("type"=>"city", "pop"=>179, "name_ru"=>"Беллингхэм", "name_en"=>"Bellingham", "lang"=>"en", "lat"=>45.135616, "lon"=>-96.282785),
"/US/US-MN/beltrami-10217" => array("type"=>"city", "pop"=>92, "name_ru"=>"Белтрами", "name_en"=>"Beltrami", "lang"=>"en", "lat"=>47.543161, "lon"=>-96.528011),
"/US/US-MN/belview-10218" => array("type"=>"city", "pop"=>376, "name_ru"=>"Белвью", "name_en"=>"Belview", "lang"=>"en", "lat"=>44.605887, "lon"=>-95.328762),
"/US/US-MN/bemidji-10219" => array("type"=>"city", "pop"=>13541, "name_ru"=>"Бемидджи", "name_en"=>"Bemidji", "lang"=>"en", "lat"=>47.482323, "lon"=>-94.878559),
"/US/US-MN/bena-10220" => array("type"=>"city", "pop"=>100, "name_ru"=>"Бена", "name_en"=>"Bena", "lang"=>"en", "lat"=>47.341083, "lon"=>-94.206176),
"/US/US-MN/benson-10221" => array("type"=>"city", "pop"=>3049, "name_ru"=>"Бенсон", "name_en"=>"Benson", "lang"=>"en", "lat"=>45.317146, "lon"=>-95.607116),
"/US/US-MN/bertha-10222" => array("type"=>"city", "pop"=>426, "name_ru"=>"Берта", "name_en"=>"Bertha", "lang"=>"en", "lat"=>46.266012, "lon"=>-95.063128),
"/US/US-MN/bethel-10223" => array("type"=>"city", "pop"=>539, "name_ru"=>"Бетел", "name_en"=>"Bethel", "lang"=>"en", "lat"=>45.400586, "lon"=>-93.269367),
"/US/US-MN/big-falls-10225" => array("type"=>"city", "pop"=>236, "name_ru"=>"Биг Фолс", "name_en"=>"Big Falls", "lang"=>"en", "lat"=>48.185065, "lon"=>-93.808196),
"/US/US-MN/big-lake-10227" => array("type"=>"city", "pop"=>9795, "name_ru"=>"Биг Лейк", "name_en"=>"Big Lake", "lang"=>"en", "lat"=>45.336605, "lon"=>-93.746560),
"/US/US-MN/bigelow-10224" => array("type"=>"city", "pop"=>221, "name_ru"=>"Бигелоу", "name_en"=>"Bigelow", "lang"=>"en", "lat"=>43.505361, "lon"=>-95.689089),
"/US/US-MN/bigfork-10226" => array("type"=>"city", "pop"=>443, "name_ru"=>"Бигфорк", "name_en"=>"Bigfork", "lang"=>"en", "lat"=>47.748396, "lon"=>-93.653415),
"/US/US-MN/bingham-lake-10228" => array("type"=>"city", "pop"=>169, "name_ru"=>"Бингхэм Лейк", "name_en"=>"Bingham Lake", "lang"=>"en", "lat"=>43.909404, "lon"=>-95.045418),
"/US/US-MN/birchwood-village-10229" => array("type"=>"city", "pop"=>901, "name_ru"=>"Берчвуд Виллидж", "name_en"=>"Birchwood Village", "lang"=>"en", "lat"=>45.059970, "lon"=>-92.977714),
"/US/US-MN/bird-island-10230" => array("type"=>"city", "pop"=>1111, "name_ru"=>"Берд Айленд", "name_en"=>"Bird Island", "lang"=>"en", "lat"=>44.764801, "lon"=>-94.894029),
"/US/US-MN/biscay-10231" => array("type"=>"city", "pop"=>108, "name_ru"=>"Бискей", "name_en"=>"Biscay", "lang"=>"en", "lat"=>44.826458, "lon"=>-94.274071),
"/US/US-MN/biwabik-10232" => array("type"=>"city", "pop"=>974, "name_ru"=>"Бивабик", "name_en"=>"Biwabik", "lang"=>"en", "lat"=>47.534241, "lon"=>-92.349983),
"/US/US-MN/blackduck-10233" => array("type"=>"city", "pop"=>761, "name_ru"=>"Блэкдак", "name_en"=>"Blackduck", "lang"=>"en", "lat"=>47.726434, "lon"=>-94.548232),
"/US/US-MN/blaine-10234" => array("type"=>"city", "pop"=>55042, "name_ru"=>"Блейн", "name_en"=>"Blaine", "lang"=>"en", "lat"=>45.166133, "lon"=>-93.208382),
"/US/US-MN/blomkest-10235" => array("type"=>"city", "pop"=>171, "name_ru"=>"Бломкест", "name_en"=>"Blomkest", "lang"=>"en", "lat"=>44.943215, "lon"=>-95.022387),
"/US/US-MN/blooming-prairie-10236" => array("type"=>"city", "pop"=>1967, "name_ru"=>"Блуминг Прейри", "name_en"=>"Blooming Prairie", "lang"=>"en", "lat"=>43.868468, "lon"=>-93.054462),
"/US/US-MN/bloomington-10237" => array("type"=>"city", "pop"=>81280, "name_ru"=>"Блумингтон", "name_en"=>"Bloomington", "lang"=>"en", "lat"=>44.829450, "lon"=>-93.315086),
"/US/US-MN/blue-earth-10238" => array("type"=>"city", "pop"=>3247, "name_ru"=>"Блу Эрт", "name_en"=>"Blue Earth", "lang"=>"en", "lat"=>43.639087, "lon"=>-94.099818),
"/US/US-MN/bluffton-10239" => array("type"=>"city", "pop"=>209, "name_ru"=>"Блаффтон", "name_en"=>"Bluffton", "lang"=>"en", "lat"=>46.469350, "lon"=>-95.233535),
"/US/US-MN/bock-10240" => array("type"=>"city", "pop"=>110, "name_ru"=>"Бок", "name_en"=>"Bock", "lang"=>"en", "lat"=>45.784980, "lon"=>-93.553232),
"/US/US-MN/borup-10241" => array("type"=>"city", "pop"=>83, "name_ru"=>"Бореп", "name_en"=>"Borup", "lang"=>"en", "lat"=>47.180797, "lon"=>-96.505846),
"/US/US-MN/bovey-10242" => array("type"=>"city", "pop"=>665, "name_ru"=>"Бови", "name_en"=>"Bovey", "lang"=>"en", "lat"=>47.294207, "lon"=>-93.402280),
"/US/US-MN/bowlus-10243" => array("type"=>"city", "pop"=>231, "name_ru"=>"Боулус", "name_en"=>"Bowlus", "lang"=>"en", "lat"=>45.820064, "lon"=>-94.407461),
"/US/US-MN/boy-river-10245" => array("type"=>"city", "pop"=>35, "name_ru"=>"Бой Ривер", "name_en"=>"Boy River", "lang"=>"en", "lat"=>47.166898, "lon"=>-94.122461),
"/US/US-MN/boyd-10244" => array("type"=>"city", "pop"=>188, "name_ru"=>"Бойд", "name_en"=>"Boyd", "lang"=>"en", "lat"=>44.850959, "lon"=>-95.900770),
"/US/US-MN/braham-10246" => array("type"=>"city", "pop"=>1646, "name_ru"=>"Брэхэм", "name_en"=>"Braham", "lang"=>"en", "lat"=>45.723742, "lon"=>-93.173363),
"/US/US-MN/brainerd-10247" => array("type"=>"city", "pop"=>13691, "name_ru"=>"Брейнерд", "name_en"=>"Brainerd", "lang"=>"en", "lat"=>46.354369, "lon"=>-94.194098),
"/US/US-MN/branch-10248" => array("type"=>"city", "pop"=>2400, "name_ru"=>"Брэнч", "name_en"=>"Branch", "lang"=>"en", "lat"=>45.514587, "lon"=>-92.957565),
"/US/US-MN/brandon-10249" => array("type"=>"city", "pop"=>439, "name_ru"=>"Брэндон", "name_en"=>"Brandon", "lang"=>"en", "lat"=>45.965422, "lon"=>-95.596947),
"/US/US-MN/breckenridge-10250" => array("type"=>"city", "pop"=>3196, "name_ru"=>"Брекенридж", "name_en"=>"Breckenridge", "lang"=>"en", "lat"=>46.265330, "lon"=>-96.585547),
"/US/US-MN/breezy-point-10251" => array("type"=>"city", "pop"=>1614, "name_ru"=>"Бризи Пойнт", "name_en"=>"Breezy Point", "lang"=>"en", "lat"=>46.601407, "lon"=>-94.219254),
"/US/US-MN/brewster-10252" => array("type"=>"city", "pop"=>483, "name_ru"=>"Брюстер", "name_en"=>"Brewster", "lang"=>"en", "lat"=>43.696200, "lon"=>-95.464761),
"/US/US-MN/bricelyn-10253" => array("type"=>"city", "pop"=>331, "name_ru"=>"Брайслин", "name_en"=>"Bricelyn", "lang"=>"en", "lat"=>43.560710, "lon"=>-93.813030),
"/US/US-MN/brook-park-10256" => array("type"=>"city", "pop"=>157, "name_ru"=>"Брук Парк", "name_en"=>"Brook Park", "lang"=>"en", "lat"=>45.948927, "lon"=>-93.073144),
"/US/US-MN/brooklyn-center-10254" => array("type"=>"city", "pop"=>27402, "name_ru"=>"Бруклин Сентер", "name_en"=>"Brooklyn Center", "lang"=>"en", "lat"=>45.068850, "lon"=>-93.317849),
"/US/US-MN/brooklyn-park-10255" => array("type"=>"city", "pop"=>71308, "name_ru"=>"Бруклин Парк", "name_en"=>"Brooklyn Park", "lang"=>"en", "lat"=>45.109567, "lon"=>-93.352082),
"/US/US-MN/brooks-10257" => array("type"=>"city", "pop"=>133, "name_ru"=>"Брукс", "name_en"=>"Brooks", "lang"=>"en", "lat"=>47.817052, "lon"=>-96.001128),
"/US/US-MN/brookston-10258" => array("type"=>"city", "pop"=>90, "name_ru"=>"Брукстон", "name_en"=>"Brookston", "lang"=>"en", "lat"=>46.865996, "lon"=>-92.603238),
"/US/US-MN/brooten-10259" => array("type"=>"city", "pop"=>616, "name_ru"=>"Брутен", "name_en"=>"Brooten", "lang"=>"en", "lat"=>45.500329, "lon"=>-95.119038),
"/US/US-MN/browerville-10260" => array("type"=>"city", "pop"=>691, "name_ru"=>"Брауэрвилл", "name_en"=>"Browerville", "lang"=>"en", "lat"=>46.085735, "lon"=>-94.868056),
"/US/US-MN/browns-valley-10262" => array("type"=>"city", "pop"=>603, "name_ru"=>"Браунс Валли", "name_en"=>"Browns Valley", "lang"=>"en", "lat"=>45.594809, "lon"=>-96.832811),
"/US/US-MN/brownsdale-10261" => array("type"=>"city", "pop"=>688, "name_ru"=>"Браунсдейл", "name_en"=>"Brownsdale", "lang"=>"en", "lat"=>43.740335, "lon"=>-92.867383),
"/US/US-MN/brownsville-10263" => array("type"=>"city", "pop"=>518, "name_ru"=>"Браунсвилл", "name_en"=>"Brownsville", "lang"=>"en", "lat"=>43.698629, "lon"=>-91.280524),
"/US/US-MN/brownton-10264" => array("type"=>"city", "pop"=>784, "name_ru"=>"Браунтон", "name_en"=>"Brownton", "lang"=>"en", "lat"=>44.732596, "lon"=>-94.351204),
"/US/US-MN/bruno-10265" => array("type"=>"city", "pop"=>106, "name_ru"=>"Бруно", "name_en"=>"Bruno", "lang"=>"en", "lat"=>46.280542, "lon"=>-92.667532),
"/US/US-MN/buckman-10266" => array("type"=>"city", "pop"=>231, "name_ru"=>"Бакмэн", "name_en"=>"Buckman", "lang"=>"en", "lat"=>45.897977, "lon"=>-94.092348),
"/US/US-MN/buffalo-10267" => array("type"=>"city", "pop"=>14222, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>45.176982, "lon"=>-93.868544),
"/US/US-MN/buffalo-lake-10268" => array("type"=>"city", "pop"=>731, "name_ru"=>"Буффало Лейк", "name_en"=>"Buffalo Lake", "lang"=>"en", "lat"=>44.736701, "lon"=>-94.617659),
"/US/US-MN/buhl-10269" => array("type"=>"city", "pop"=>983, "name_ru"=>"Бул", "name_en"=>"Buhl", "lang"=>"en", "lat"=>47.498508, "lon"=>-92.766890),
"/US/US-MN/burnsville-10270" => array("type"=>"city", "pop"=>59139, "name_ru"=>"Бернсвилл", "name_en"=>"Burnsville", "lang"=>"en", "lat"=>44.764565, "lon"=>-93.279860),
"/US/US-MN/burtrum-10271" => array("type"=>"city", "pop"=>132, "name_ru"=>"Бертрам", "name_en"=>"Burtrum", "lang"=>"en", "lat"=>45.865865, "lon"=>-94.687207),
"/US/US-MN/butterfield-10272" => array("type"=>"city", "pop"=>503, "name_ru"=>"Баттерфилд", "name_en"=>"Butterfield", "lang"=>"en", "lat"=>43.958245, "lon"=>-94.793001),
"/US/US-MN/byron-10273" => array("type"=>"city", "pop"=>4550, "name_ru"=>"Байрон", "name_en"=>"Byron", "lang"=>"en", "lat"=>44.033416, "lon"=>-92.646091),
"/US/US-MN/caledonia-10274" => array("type"=>"city", "pop"=>2821, "name_ru"=>"Каледония", "name_en"=>"Caledonia", "lang"=>"en", "lat"=>43.634135, "lon"=>-91.500260),
"/US/US-MN/callaway-10275" => array("type"=>"city", "pop"=>210, "name_ru"=>"Каллавэй", "name_en"=>"Callaway", "lang"=>"en", "lat"=>46.983003, "lon"=>-95.911272),
"/US/US-MN/calumet-10276" => array("type"=>"city", "pop"=>355, "name_ru"=>"Калумет", "name_en"=>"Calumet", "lang"=>"en", "lat"=>47.319156, "lon"=>-93.269615),
"/US/US-MN/cambridge-10277" => array("type"=>"city", "pop"=>7692, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>45.564339, "lon"=>-93.226930),
"/US/US-MN/campbell-10278" => array("type"=>"city", "pop"=>209, "name_ru"=>"Кэмпбелл", "name_en"=>"Campbell", "lang"=>"en", "lat"=>46.097905, "lon"=>-96.405387),
"/US/US-MN/canby-10279" => array("type"=>"city", "pop"=>1710, "name_ru"=>"Кэнби", "name_en"=>"Canby", "lang"=>"en", "lat"=>44.714588, "lon"=>-96.269023),
"/US/US-MN/cannon-falls-10280" => array("type"=>"city", "pop"=>4034, "name_ru"=>"Кэннон Фоллс", "name_en"=>"Cannon Falls", "lang"=>"en", "lat"=>44.516500, "lon"=>-92.902621),
"/US/US-MN/canton-10281" => array("type"=>"city", "pop"=>323, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>43.530081, "lon"=>-91.929443),
"/US/US-MN/carlos-10282" => array("type"=>"city", "pop"=>421, "name_ru"=>"Карлос", "name_en"=>"Carlos", "lang"=>"en", "lat"=>45.973921, "lon"=>-95.292810),
"/US/US-MN/carlton-10283" => array("type"=>"city", "pop"=>770, "name_ru"=>"Карлтон", "name_en"=>"Carlton", "lang"=>"en", "lat"=>46.660259, "lon"=>-92.423155),
"/US/US-MN/carver-10284" => array("type"=>"city", "pop"=>2915, "name_ru"=>"Карвер", "name_en"=>"Carver", "lang"=>"en", "lat"=>44.760367, "lon"=>-93.627473),
"/US/US-MN/cass-lake-10285" => array("type"=>"city", "pop"=>828, "name_ru"=>"Касс Лейк", "name_en"=>"Cass Lake", "lang"=>"en", "lat"=>47.378771, "lon"=>-94.600580),
"/US/US-MN/cedar-mills-10286" => array("type"=>"city", "pop"=>53, "name_ru"=>"Сидар Миллс", "name_en"=>"Cedar Mills", "lang"=>"en", "lat"=>44.947244, "lon"=>-94.517286),
"/US/US-MN/center-city-10287" => array("type"=>"city", "pop"=>608, "name_ru"=>"Сентер Сити", "name_en"=>"Center City", "lang"=>"en", "lat"=>45.390862, "lon"=>-92.818134),
"/US/US-MN/centerville-10288" => array("type"=>"city", "pop"=>3737, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>45.164952, "lon"=>-93.053023),
"/US/US-MN/ceylon-10289" => array("type"=>"city", "pop"=>372, "name_ru"=>"Цейлон", "name_en"=>"Ceylon", "lang"=>"en", "lat"=>43.531864, "lon"=>-94.631997),
"/US/US-MN/champlin-10290" => array("type"=>"city", "pop"=>23503, "name_ru"=>"Чэмплин", "name_en"=>"Champlin", "lang"=>"en", "lat"=>45.165460, "lon"=>-93.389228),
"/US/US-MN/chandler-10291" => array("type"=>"city", "pop"=>239, "name_ru"=>"Чандлер", "name_en"=>"Chandler", "lang"=>"en", "lat"=>43.929426, "lon"=>-95.949128),
"/US/US-MN/chanhassen-10292" => array("type"=>"city", "pop"=>23480, "name_ru"=>"Чэнхассен", "name_en"=>"Chanhassen", "lang"=>"en", "lat"=>44.857207, "lon"=>-93.557830),
"/US/US-MN/chaska-10293" => array("type"=>"city", "pop"=>24116, "name_ru"=>"Часка", "name_en"=>"Chaska", "lang"=>"en", "lat"=>44.821100, "lon"=>-93.605138),
"/US/US-MN/chatfield-10294" => array("type"=>"city", "pop"=>2518, "name_ru"=>"Чэтфилд", "name_en"=>"Chatfield", "lang"=>"en", "lat"=>43.843927, "lon"=>-92.185937),
"/US/US-MN/chickamaw-beach-10295" => array("type"=>"city", "pop"=>145, "name_ru"=>"Чикамо Бич", "name_en"=>"Chickamaw Beach", "lang"=>"en", "lat"=>46.748745, "lon"=>-94.389776),
"/US/US-MN/chisago-city-10296" => array("type"=>"city", "pop"=>4202, "name_ru"=>"Чисаго Сити", "name_en"=>"Chisago City", "lang"=>"en", "lat"=>45.370350, "lon"=>-92.888046),
"/US/US-MN/chisholm-10297" => array("type"=>"city", "pop"=>4594, "name_ru"=>"Чизем", "name_en"=>"Chisholm", "lang"=>"en", "lat"=>47.487520, "lon"=>-92.879158),
"/US/US-MN/chokio-10298" => array("type"=>"city", "pop"=>424, "name_ru"=>"Чокио", "name_en"=>"Chokio", "lang"=>"en", "lat"=>45.573689, "lon"=>-96.174119),
"/US/US-MN/circle-pines-10299" => array("type"=>"city", "pop"=>5395, "name_ru"=>"Серкл Пайнс", "name_en"=>"Circle Pines", "lang"=>"en", "lat"=>45.138500, "lon"=>-93.152445),
"/US/US-MN/clara-city-10300" => array("type"=>"city", "pop"=>1310, "name_ru"=>"Клара Сити", "name_en"=>"Clara City", "lang"=>"en", "lat"=>44.957994, "lon"=>-95.366030),
"/US/US-MN/claremont-10301" => array("type"=>"city", "pop"=>594, "name_ru"=>"Клэрмонт", "name_en"=>"Claremont", "lang"=>"en", "lat"=>44.044687, "lon"=>-92.996160),
"/US/US-MN/clarissa-10302" => array("type"=>"city", "pop"=>592, "name_ru"=>"Кларисса", "name_en"=>"Clarissa", "lang"=>"en", "lat"=>46.129191, "lon"=>-94.949391),
"/US/US-MN/clarkfield-10303" => array("type"=>"city", "pop"=>846, "name_ru"=>"Кларкфилд", "name_en"=>"Clarkfield", "lang"=>"en", "lat"=>44.791219, "lon"=>-95.807339),
"/US/US-MN/clarks-grove-10304" => array("type"=>"city", "pop"=>704, "name_ru"=>"Кларкс Гров", "name_en"=>"Clarks Grove", "lang"=>"en", "lat"=>43.764421, "lon"=>-93.327803),
"/US/US-MN/clear-lake-10306" => array("type"=>"city", "pop"=>478, "name_ru"=>"Клир Лейк", "name_en"=>"Clear Lake", "lang"=>"en", "lat"=>45.446087, "lon"=>-93.997964),
"/US/US-MN/clearbrook-10305" => array("type"=>"city", "pop"=>534, "name_ru"=>"Клирбрук", "name_en"=>"Clearbrook", "lang"=>"en", "lat"=>47.694714, "lon"=>-95.428346),
"/US/US-MN/clearwater-10307" => array("type"=>"city", "pop"=>1493, "name_ru"=>"Клирвотер", "name_en"=>"Clearwater", "lang"=>"en", "lat"=>45.410163, "lon"=>-94.042694),
"/US/US-MN/clements-10308" => array("type"=>"city", "pop"=>169, "name_ru"=>"Клементс", "name_en"=>"Clements", "lang"=>"en", "lat"=>44.379947, "lon"=>-95.054772),
"/US/US-MN/cleveland-10309" => array("type"=>"city", "pop"=>697, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>44.323795, "lon"=>-93.835069),
"/US/US-MN/climax-10310" => array("type"=>"city", "pop"=>235, "name_ru"=>"Клаймакс", "name_en"=>"Climax", "lang"=>"en", "lat"=>47.606788, "lon"=>-96.812274),
"/US/US-MN/clinton-10311" => array("type"=>"city", "pop"=>401, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>45.462186, "lon"=>-96.443274),
"/US/US-MN/clitherall-10312" => array("type"=>"city", "pop"=>113, "name_ru"=>"Клиттеролл", "name_en"=>"Clitherall", "lang"=>"en", "lat"=>46.274338, "lon"=>-95.629678),
"/US/US-MN/clontarf-10313" => array("type"=>"city", "pop"=>153, "name_ru"=>"Клонтарф", "name_en"=>"Clontarf", "lang"=>"en", "lat"=>45.376334, "lon"=>-95.677167),
"/US/US-MN/cloquet-10314" => array("type"=>"city", "pop"=>11387, "name_ru"=>"Клокет", "name_en"=>"Cloquet", "lang"=>"en", "lat"=>46.729040, "lon"=>-92.494426),
"/US/US-MN/coates-10315" => array("type"=>"city", "pop"=>161, "name_ru"=>"Коатс", "name_en"=>"Coates", "lang"=>"en", "lat"=>44.715700, "lon"=>-93.035048),
"/US/US-MN/cobden-10316" => array("type"=>"city", "pop"=>59, "name_ru"=>"Кобден", "name_en"=>"Cobden", "lang"=>"en", "lat"=>44.283854, "lon"=>-94.847722),
"/US/US-MN/cokato-10317" => array("type"=>"city", "pop"=>2711, "name_ru"=>"Кокато", "name_en"=>"Cokato", "lang"=>"en", "lat"=>45.076759, "lon"=>-94.189375),
"/US/US-MN/cold-spring-10318" => array("type"=>"city", "pop"=>3751, "name_ru"=>"Коулд Спринг", "name_en"=>"Cold Spring", "lang"=>"en", "lat"=>45.457498, "lon"=>-94.431480),
"/US/US-MN/coleraine-10319" => array("type"=>"city", "pop"=>1008, "name_ru"=>"Колерейн", "name_en"=>"Coleraine", "lang"=>"en", "lat"=>47.306279, "lon"=>-93.423125),
"/US/US-MN/cologne-10320" => array("type"=>"city", "pop"=>1416, "name_ru"=>"Кологн", "name_en"=>"Cologne", "lang"=>"en", "lat"=>44.771600, "lon"=>-93.788015),
"/US/US-MN/columbia-heights-10321" => array("type"=>"city", "pop"=>18210, "name_ru"=>"Колумбия Хайтс", "name_en"=>"Columbia Heights", "lang"=>"en", "lat"=>45.049100, "lon"=>-93.246649),
"/US/US-MN/comfrey-10322" => array("type"=>"city", "pop"=>344, "name_ru"=>"Комфри", "name_en"=>"Comfrey", "lang"=>"en", "lat"=>44.110900, "lon"=>-94.902683),
"/US/US-MN/comstock-10323" => array("type"=>"city", "pop"=>117, "name_ru"=>"Комсток", "name_en"=>"Comstock", "lang"=>"en", "lat"=>46.659970, "lon"=>-96.746575),
"/US/US-MN/conger-10324" => array("type"=>"city", "pop"=>124, "name_ru"=>"Конгер", "name_en"=>"Conger", "lang"=>"en", "lat"=>43.615347, "lon"=>-93.528637),
"/US/US-MN/cook-10325" => array("type"=>"city", "pop"=>618, "name_ru"=>"Кук", "name_en"=>"Cook", "lang"=>"en", "lat"=>47.853082, "lon"=>-92.688050),
"/US/US-MN/coon-rapids-10326" => array("type"=>"city", "pop"=>61832, "name_ru"=>"Кун Рэпидс", "name_en"=>"Coon Rapids", "lang"=>"en", "lat"=>45.174050, "lon"=>-93.311091),
"/US/US-MN/corcoran-10327" => array("type"=>"city", "pop"=>5630, "name_ru"=>"Коркоран", "name_en"=>"Corcoran", "lang"=>"en", "lat"=>45.117050, "lon"=>-93.586274),
"/US/US-MN/correll-10328" => array("type"=>"city", "pop"=>44, "name_ru"=>"Коррел", "name_en"=>"Correll", "lang"=>"en", "lat"=>45.231820, "lon"=>-96.162698),
"/US/US-MN/cosmos-10329" => array("type"=>"city", "pop"=>555, "name_ru"=>"Космос", "name_en"=>"Cosmos", "lang"=>"en", "lat"=>44.935870, "lon"=>-94.697372),
"/US/US-MN/cottage-grove-10330" => array("type"=>"city", "pop"=>33623, "name_ru"=>"Коттэдж Гров", "name_en"=>"Cottage Grove", "lang"=>"en", "lat"=>44.817499, "lon"=>-92.926990),
"/US/US-MN/cottonwood-10331" => array("type"=>"city", "pop"=>1118, "name_ru"=>"Коттонвуд", "name_en"=>"Cottonwood", "lang"=>"en", "lat"=>44.609905, "lon"=>-95.670826),
"/US/US-MN/courtland-10332" => array("type"=>"city", "pop"=>617, "name_ru"=>"Кортленд", "name_en"=>"Courtland", "lang"=>"en", "lat"=>44.266483, "lon"=>-94.344042),
"/US/US-MN/cromwell-10333" => array("type"=>"city", "pop"=>203, "name_ru"=>"Кромуэлл", "name_en"=>"Cromwell", "lang"=>"en", "lat"=>46.679505, "lon"=>-92.870657),
"/US/US-MN/crookston-10334" => array("type"=>"city", "pop"=>7790, "name_ru"=>"Крукстон", "name_en"=>"Crookston", "lang"=>"en", "lat"=>47.774211, "lon"=>-96.605647),
"/US/US-MN/crosby-10335" => array("type"=>"city", "pop"=>2243, "name_ru"=>"Кросби", "name_en"=>"Crosby", "lang"=>"en", "lat"=>46.489951, "lon"=>-93.956643),
"/US/US-MN/crosslake-10336" => array("type"=>"city", "pop"=>2108, "name_ru"=>"Кросслейк", "name_en"=>"Crosslake", "lang"=>"en", "lat"=>46.682454, "lon"=>-94.094837),
"/US/US-MN/crystal-10337" => array("type"=>"city", "pop"=>21707, "name_ru"=>"Кристал", "name_en"=>"Crystal", "lang"=>"en", "lat"=>45.037300, "lon"=>-93.359349),
"/US/US-MN/currie-10338" => array("type"=>"city", "pop"=>198, "name_ru"=>"Карри", "name_en"=>"Currie", "lang"=>"en", "lat"=>44.070972, "lon"=>-95.666553),
"/US/US-MN/cuyuna-10339" => array("type"=>"city", "pop"=>273, "name_ru"=>"Каиуна", "name_en"=>"Cuyuna", "lang"=>"en", "lat"=>46.511770, "lon"=>-93.928965),
"/US/US-MN/cyrus-10340" => array("type"=>"city", "pop"=>267, "name_ru"=>"Сирус", "name_en"=>"Cyrus", "lang"=>"en", "lat"=>45.615140, "lon"=>-95.736983),
"/US/US-MN/dakota-10341" => array("type"=>"city", "pop"=>320, "name_ru"=>"Дакота", "name_en"=>"Dakota", "lang"=>"en", "lat"=>43.911776, "lon"=>-91.358561),
"/US/US-MN/dalton-10342" => array("type"=>"city", "pop"=>241, "name_ru"=>"Далтон", "name_en"=>"Dalton", "lang"=>"en", "lat"=>46.173443, "lon"=>-95.916328),
"/US/US-MN/danube-10343" => array("type"=>"city", "pop"=>478, "name_ru"=>"Дануб", "name_en"=>"Danube", "lang"=>"en", "lat"=>44.791308, "lon"=>-95.102542),
"/US/US-MN/danvers-10344" => array("type"=>"city", "pop"=>91, "name_ru"=>"Дэнверс", "name_en"=>"Danvers", "lang"=>"en", "lat"=>45.282623, "lon"=>-95.752738),
"/US/US-MN/darfur-10345" => array("type"=>"city", "pop"=>121, "name_ru"=>"Дарфур", "name_en"=>"Darfur", "lang"=>"en", "lat"=>44.053605, "lon"=>-94.837318),
"/US/US-MN/darwin-10346" => array("type"=>"city", "pop"=>304, "name_ru"=>"Дарвин", "name_en"=>"Darwin", "lang"=>"en", "lat"=>45.097012, "lon"=>-94.404057),
"/US/US-MN/dassel-10347" => array("type"=>"city", "pop"=>1320, "name_ru"=>"Дассел", "name_en"=>"Dassel", "lang"=>"en", "lat"=>45.081778, "lon"=>-94.307372),
"/US/US-MN/dawson-10348" => array("type"=>"city", "pop"=>1357, "name_ru"=>"Доусон", "name_en"=>"Dawson", "lang"=>"en", "lat"=>44.928855, "lon"=>-96.049024),
"/US/US-MN/dayton-10349" => array("type"=>"city", "pop"=>4643, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>45.189900, "lon"=>-93.471732),
"/US/US-MN/de-graff-10354" => array("type"=>"city", "pop"=>115, "name_ru"=>"Де Графф", "name_en"=>"De Graff", "lang"=>"en", "lat"=>45.259240, "lon"=>-95.468899),
"/US/US-MN/deephaven-10350" => array("type"=>"city", "pop"=>3775, "name_ru"=>"Дипхейвен", "name_en"=>"Deephaven", "lang"=>"en", "lat"=>44.929350, "lon"=>-93.522994),
"/US/US-MN/deer-creek-10351" => array("type"=>"city", "pop"=>325, "name_ru"=>"Дир Крик", "name_en"=>"Deer Creek", "lang"=>"en", "lat"=>46.391268, "lon"=>-95.322024),
"/US/US-MN/deer-river-10352" => array("type"=>"city", "pop"=>918, "name_ru"=>"Дир Ривер", "name_en"=>"Deer River", "lang"=>"en", "lat"=>47.335953, "lon"=>-93.793989),
"/US/US-MN/deerwood-10353" => array("type"=>"city", "pop"=>573, "name_ru"=>"Дирвуд", "name_en"=>"Deerwood", "lang"=>"en", "lat"=>46.473237, "lon"=>-93.899033),
"/US/US-MN/delano-10355" => array("type"=>"city", "pop"=>5309, "name_ru"=>"Делано", "name_en"=>"Delano", "lang"=>"en", "lat"=>45.041262, "lon"=>-93.785483),
"/US/US-MN/delavan-10356" => array("type"=>"city", "pop"=>205, "name_ru"=>"Делаван", "name_en"=>"Delavan", "lang"=>"en", "lat"=>43.767316, "lon"=>-94.018460),
"/US/US-MN/delhi-10357" => array("type"=>"city", "pop"=>63, "name_ru"=>"Дели", "name_en"=>"Delhi", "lang"=>"en", "lat"=>44.598184, "lon"=>-95.213029),
"/US/US-MN/dellwood-10358" => array("type"=>"city", "pop"=>1044, "name_ru"=>"Деллвуд", "name_en"=>"Dellwood", "lang"=>"en", "lat"=>45.102083, "lon"=>-92.966442),
"/US/US-MN/denham-10359" => array("type"=>"city", "pop"=>40, "name_ru"=>"Денхэм", "name_en"=>"Denham", "lang"=>"en", "lat"=>46.362328, "lon"=>-92.943620),
"/US/US-MN/dennison-10360" => array("type"=>"city", "pop"=>164, "name_ru"=>"Деннисон", "name_en"=>"Dennison", "lang"=>"en", "lat"=>44.403398, "lon"=>-93.028696),
"/US/US-MN/dent-10361" => array("type"=>"city", "pop"=>184, "name_ru"=>"Дент", "name_en"=>"Dent", "lang"=>"en", "lat"=>46.552118, "lon"=>-95.717164),
"/US/US-MN/detroit-lakes-10362" => array("type"=>"city", "pop"=>8141, "name_ru"=>"Детройт Лейк", "name_en"=>"Detroit Lakes", "lang"=>"en", "lat"=>46.813129, "lon"=>-95.842203),
"/US/US-MN/dexter-10363" => array("type"=>"city", "pop"=>314, "name_ru"=>"Декстер", "name_en"=>"Dexter", "lang"=>"en", "lat"=>43.718083, "lon"=>-92.702479),
"/US/US-MN/dilworth-10364" => array("type"=>"city", "pop"=>3677, "name_ru"=>"Дилворт", "name_en"=>"Dilworth", "lang"=>"en", "lat"=>46.875400, "lon"=>-96.698724),
"/US/US-MN/dodge-center-10365" => array("type"=>"city", "pop"=>2593, "name_ru"=>"Додж Сентер", "name_en"=>"Dodge Center", "lang"=>"en", "lat"=>44.028596, "lon"=>-92.849044),
"/US/US-MN/donaldson-10366" => array("type"=>"city", "pop"=>35, "name_ru"=>"Дональдсон", "name_en"=>"Donaldson", "lang"=>"en", "lat"=>48.574738, "lon"=>-96.897280),
"/US/US-MN/donnelly-10367" => array("type"=>"city", "pop"=>243, "name_ru"=>"Доннелли", "name_en"=>"Donnelly", "lang"=>"en", "lat"=>45.689984, "lon"=>-96.014109),
"/US/US-MN/doran-10368" => array("type"=>"city", "pop"=>51, "name_ru"=>"Доран", "name_en"=>"Doran", "lang"=>"en", "lat"=>46.185236, "lon"=>-96.485627),
"/US/US-MN/dover-10369" => array("type"=>"city", "pop"=>637, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>43.969850, "lon"=>-92.133696),
"/US/US-MN/dovray-10370" => array("type"=>"city", "pop"=>57, "name_ru"=>"Доврей", "name_en"=>"Dovray", "lang"=>"en", "lat"=>44.054791, "lon"=>-95.546097),
"/US/US-MN/duluth-10371" => array("type"=>"city", "pop"=>84284, "name_ru"=>"Далут", "name_en"=>"Duluth", "lang"=>"en", "lat"=>46.781017, "lon"=>-92.117979),
"/US/US-MN/dumont-10372" => array("type"=>"city", "pop"=>105, "name_ru"=>"Думонт", "name_en"=>"Dumont", "lang"=>"en", "lat"=>45.716795, "lon"=>-96.421922),
"/US/US-MN/dundas-10373" => array("type"=>"city", "pop"=>1080, "name_ru"=>"Дандас", "name_en"=>"Dundas", "lang"=>"en", "lat"=>44.429719, "lon"=>-93.204856),
"/US/US-MN/dundee-10374" => array("type"=>"city", "pop"=>98, "name_ru"=>"Данди", "name_en"=>"Dundee", "lang"=>"en", "lat"=>43.844332, "lon"=>-95.466092),
"/US/US-MN/dunnell-10375" => array("type"=>"city", "pop"=>175, "name_ru"=>"Даннел", "name_en"=>"Dunnell", "lang"=>"en", "lat"=>43.560476, "lon"=>-94.773651),
"/US/US-MN/eagan-10376" => array("type"=>"city", "pop"=>63985, "name_ru"=>"Иган", "name_en"=>"Eagan", "lang"=>"en", "lat"=>44.819400, "lon"=>-93.163004),
"/US/US-MN/eagle-bend-10377" => array("type"=>"city", "pop"=>567, "name_ru"=>"Игл Бенд", "name_en"=>"Eagle Bend", "lang"=>"en", "lat"=>46.165003, "lon"=>-95.034013),
"/US/US-MN/eagle-lake-10378" => array("type"=>"city", "pop"=>2285, "name_ru"=>"Игл Лэйк", "name_en"=>"Eagle Lake", "lang"=>"en", "lat"=>44.159412, "lon"=>-93.882549),
"/US/US-MN/east-bethel-10379" => array("type"=>"city", "pop"=>11895, "name_ru"=>"Ист Бетел", "name_en"=>"East Bethel", "lang"=>"en", "lat"=>45.344159, "lon"=>-93.201447),
"/US/US-MN/east-grand-forks-10380" => array("type"=>"city", "pop"=>7812, "name_ru"=>"Ист Гранд Форкс", "name_en"=>"East Grand Forks", "lang"=>"en", "lat"=>47.930000, "lon"=>-97.015904),
"/US/US-MN/east-gull-lake-10381" => array("type"=>"city", "pop"=>1027, "name_ru"=>"Ист Галл Лейк", "name_en"=>"East Gull Lake", "lang"=>"en", "lat"=>46.410756, "lon"=>-94.352185),
"/US/US-MN/easton-10382" => array("type"=>"city", "pop"=>188, "name_ru"=>"Истон", "name_en"=>"Easton", "lang"=>"en", "lat"=>43.765871, "lon"=>-93.900264),
"/US/US-MN/echo-10383" => array("type"=>"city", "pop"=>242, "name_ru"=>"Эхо", "name_en"=>"Echo", "lang"=>"en", "lat"=>44.622749, "lon"=>-95.411259),
"/US/US-MN/eden-prairie-10384" => array("type"=>"city", "pop"=>61191, "name_ru"=>"Иден Прейри", "name_en"=>"Eden Prairie", "lang"=>"en", "lat"=>44.849250, "lon"=>-93.458137),
"/US/US-MN/eden-valley-10385" => array("type"=>"city", "pop"=>954, "name_ru"=>"Иден Валли", "name_en"=>"Eden Valley", "lang"=>"en", "lat"=>45.325683, "lon"=>-94.545664),
"/US/US-MN/edgerton-10386" => array("type"=>"city", "pop"=>988, "name_ru"=>"Эджертон", "name_en"=>"Edgerton", "lang"=>"en", "lat"=>43.874196, "lon"=>-96.130556),
"/US/US-MN/edina-10387" => array("type"=>"city", "pop"=>45608, "name_ru"=>"Эдина", "name_en"=>"Edina", "lang"=>"en", "lat"=>44.891050, "lon"=>-93.359624),
"/US/US-MN/effie-10388" => array("type"=>"city", "pop"=>85, "name_ru"=>"Эффи", "name_en"=>"Effie", "lang"=>"en", "lat"=>47.840684, "lon"=>-93.637775),
"/US/US-MN/eitzen-10389" => array("type"=>"city", "pop"=>228, "name_ru"=>"Айтзен", "name_en"=>"Eitzen", "lang"=>"en", "lat"=>43.508301, "lon"=>-91.464700),
"/US/US-MN/elba-10390" => array("type"=>"city", "pop"=>210, "name_ru"=>"Эльба", "name_en"=>"Elba", "lang"=>"en", "lat"=>44.087072, "lon"=>-92.017298),
"/US/US-MN/elbow-lake-10391" => array("type"=>"city", "pop"=>1184, "name_ru"=>"Элбоу Лейк", "name_en"=>"Elbow Lake", "lang"=>"en", "lat"=>45.991780, "lon"=>-95.977350),
"/US/US-MN/elgin-10392" => array("type"=>"city", "pop"=>938, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>44.130912, "lon"=>-92.253922),
"/US/US-MN/elizabeth-10393" => array("type"=>"city", "pop"=>183, "name_ru"=>"Элизабет", "name_en"=>"Elizabeth", "lang"=>"en", "lat"=>46.379313, "lon"=>-96.130404),
"/US/US-MN/elk-river-10395" => array("type"=>"city", "pop"=>23194, "name_ru"=>"Элк Ривер", "name_en"=>"Elk River", "lang"=>"en", "lat"=>45.332591, "lon"=>-93.565768),
"/US/US-MN/elko-10394" => array("type"=>"city", "pop"=>223, "name_ru"=>"Элко", "name_en"=>"Elko", "lang"=>"en", "lat"=>44.565287, "lon"=>-93.324481),
"/US/US-MN/elkton-10396" => array("type"=>"city", "pop"=>147, "name_ru"=>"Элктон", "name_en"=>"Elkton", "lang"=>"en", "lat"=>43.659837, "lon"=>-92.708710),
"/US/US-MN/ellendale-10397" => array("type"=>"city", "pop"=>673, "name_ru"=>"Эллендэйл", "name_en"=>"Ellendale", "lang"=>"en", "lat"=>43.873579, "lon"=>-93.301214),
"/US/US-MN/ellsworth-10398" => array("type"=>"city", "pop"=>513, "name_ru"=>"Эллсворт", "name_en"=>"Ellsworth", "lang"=>"en", "lat"=>43.521138, "lon"=>-96.018471),
"/US/US-MN/elmdale-10399" => array("type"=>"city", "pop"=>98, "name_ru"=>"Элмдейл", "name_en"=>"Elmdale", "lang"=>"en", "lat"=>45.837255, "lon"=>-94.490231),
"/US/US-MN/elmore-10400" => array("type"=>"city", "pop"=>656, "name_ru"=>"Элмор", "name_en"=>"Elmore", "lang"=>"en", "lat"=>43.505852, "lon"=>-94.088455),
"/US/US-MN/elrosa-10401" => array("type"=>"city", "pop"=>178, "name_ru"=>"Элроза", "name_en"=>"Elrosa", "lang"=>"en", "lat"=>45.562494, "lon"=>-94.947593),
"/US/US-MN/ely-10402" => array("type"=>"city", "pop"=>3473, "name_ru"=>"Или", "name_en"=>"Ely", "lang"=>"en", "lat"=>47.905657, "lon"=>-91.850604),
"/US/US-MN/elysian-10403" => array("type"=>"city", "pop"=>616, "name_ru"=>"Элизиан", "name_en"=>"Elysian", "lang"=>"en", "lat"=>44.199968, "lon"=>-93.677074),
"/US/US-MN/emily-10404" => array("type"=>"city", "pop"=>991, "name_ru"=>"Эмили", "name_en"=>"Emily", "lang"=>"en", "lat"=>46.756883, "lon"=>-93.961521),
"/US/US-MN/emmons-10405" => array("type"=>"city", "pop"=>405, "name_ru"=>"Эммонс", "name_en"=>"Emmons", "lang"=>"en", "lat"=>43.507373, "lon"=>-93.483779),
"/US/US-MN/erhard-10406" => array("type"=>"city", "pop"=>149, "name_ru"=>"Эрхард", "name_en"=>"Erhard", "lang"=>"en", "lat"=>46.483992, "lon"=>-96.097135),
"/US/US-MN/erskine-10407" => array("type"=>"city", "pop"=>403, "name_ru"=>"Эрскин", "name_en"=>"Erskine", "lang"=>"en", "lat"=>47.663606, "lon"=>-96.013744),
"/US/US-MN/evan-10408" => array("type"=>"city", "pop"=>92, "name_ru"=>"Эван", "name_en"=>"Evan", "lang"=>"en", "lat"=>44.352508, "lon"=>-94.836885),
"/US/US-MN/evansville-10409" => array("type"=>"city", "pop"=>568, "name_ru"=>"Эвансвилл", "name_en"=>"Evansville", "lang"=>"en", "lat"=>46.007246, "lon"=>-95.683726),
"/US/US-MN/eveleth-10410" => array("type"=>"city", "pop"=>3582, "name_ru"=>"Эвелет", "name_en"=>"Eveleth", "lang"=>"en", "lat"=>47.463676, "lon"=>-92.542861),
"/US/US-MN/excelsior-10411" => array("type"=>"city", "pop"=>2291, "name_ru"=>"Экселсиор", "name_en"=>"Excelsior", "lang"=>"en", "lat"=>44.900653, "lon"=>-93.566506),
"/US/US-MN/eyota-10412" => array("type"=>"city", "pop"=>1765, "name_ru"=>"Эйота", "name_en"=>"Eyota", "lang"=>"en", "lat"=>43.987640, "lon"=>-92.228225),
"/US/US-MN/fairfax-10413" => array("type"=>"city", "pop"=>1208, "name_ru"=>"Фэрфакс", "name_en"=>"Fairfax", "lang"=>"en", "lat"=>44.528218, "lon"=>-94.722894),
"/US/US-MN/fairmont-10414" => array("type"=>"city", "pop"=>10221, "name_ru"=>"Фэрмонт", "name_en"=>"Fairmont", "lang"=>"en", "lat"=>43.642388, "lon"=>-94.457775),
"/US/US-MN/falcon-heights-10415" => array("type"=>"city", "pop"=>5643, "name_ru"=>"Фэлкон Хайтс", "name_en"=>"Falcon Heights", "lang"=>"en", "lat"=>44.987900, "lon"=>-93.179427),
"/US/US-MN/faribault-10416" => array("type"=>"city", "pop"=>22009, "name_ru"=>"Фариболт", "name_en"=>"Faribault", "lang"=>"en", "lat"=>44.297441, "lon"=>-93.274921),
"/US/US-MN/farmington-10417" => array("type"=>"city", "pop"=>19057, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>44.657150, "lon"=>-93.178338),
"/US/US-MN/farwell-10418" => array("type"=>"city", "pop"=>49, "name_ru"=>"Фаруэлл", "name_en"=>"Farwell", "lang"=>"en", "lat"=>45.751869, "lon"=>-95.618113),
"/US/US-MN/federal-dam-10419" => array("type"=>"city", "pop"=>93, "name_ru"=>"Федерал Дэм", "name_en"=>"Federal Dam", "lang"=>"en", "lat"=>47.238477, "lon"=>-94.218836),
"/US/US-MN/felton-10420" => array("type"=>"city", "pop"=>201, "name_ru"=>"Фелтон", "name_en"=>"Felton", "lang"=>"en", "lat"=>47.076437, "lon"=>-96.504848),
"/US/US-MN/fergus-falls-10421" => array("type"=>"city", "pop"=>13636, "name_ru"=>"Фергус Фоллс", "name_en"=>"Fergus Falls", "lang"=>"en", "lat"=>46.282125, "lon"=>-96.067652),
"/US/US-MN/fertile-10422" => array("type"=>"city", "pop"=>839, "name_ru"=>"Фертил", "name_en"=>"Fertile", "lang"=>"en", "lat"=>47.532071, "lon"=>-96.291338),
"/US/US-MN/fifty-lakes-10423" => array("type"=>"city", "pop"=>421, "name_ru"=>"Фифти Лейкс", "name_en"=>"Fifty Lakes", "lang"=>"en", "lat"=>46.748755, "lon"=>-94.095963),
"/US/US-MN/finlayson-10424" => array("type"=>"city", "pop"=>323, "name_ru"=>"Финлейсон", "name_en"=>"Finlayson", "lang"=>"en", "lat"=>46.204952, "lon"=>-92.927570),
"/US/US-MN/fisher-10425" => array("type"=>"city", "pop"=>403, "name_ru"=>"Фишер", "name_en"=>"Fisher", "lang"=>"en", "lat"=>47.799277, "lon"=>-96.799617),
"/US/US-MN/flensburg-10426" => array("type"=>"city", "pop"=>216, "name_ru"=>"Фленсбург", "name_en"=>"Flensburg", "lang"=>"en", "lat"=>45.947678, "lon"=>-94.530077),
"/US/US-MN/floodwood-10427" => array("type"=>"city", "pop"=>497, "name_ru"=>"Фладвуд", "name_en"=>"Floodwood", "lang"=>"en", "lat"=>46.926918, "lon"=>-92.915527),
"/US/US-MN/florence-10428" => array("type"=>"city", "pop"=>56, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>44.236867, "lon"=>-96.051474),
"/US/US-MN/foley-10429" => array("type"=>"city", "pop"=>2381, "name_ru"=>"Фоли", "name_en"=>"Foley", "lang"=>"en", "lat"=>45.663225, "lon"=>-93.909405),
"/US/US-MN/forada-10430" => array("type"=>"city", "pop"=>193, "name_ru"=>"Форада", "name_en"=>"Forada", "lang"=>"en", "lat"=>45.789511, "lon"=>-95.355397),
"/US/US-MN/forest-lake-10431" => array("type"=>"city", "pop"=>17373, "name_ru"=>"Форест Лейк", "name_en"=>"Forest Lake", "lang"=>"en", "lat"=>45.275800, "lon"=>-92.985766),
"/US/US-MN/foreston-10432" => array("type"=>"city", "pop"=>539, "name_ru"=>"Форстон", "name_en"=>"Foreston", "lang"=>"en", "lat"=>45.733192, "lon"=>-93.709365),
"/US/US-MN/fort-ripley-10433" => array("type"=>"city", "pop"=>73, "name_ru"=>"Форт Рипли", "name_en"=>"Fort Ripley", "lang"=>"en", "lat"=>46.166390, "lon"=>-94.361659),
"/US/US-MN/fosston-10434" => array("type"=>"city", "pop"=>1470, "name_ru"=>"Фосстон", "name_en"=>"Fosston", "lang"=>"en", "lat"=>47.580888, "lon"=>-95.747661),
"/US/US-MN/fountain-10435" => array("type"=>"city", "pop"=>340, "name_ru"=>"Фонтейн", "name_en"=>"Fountain", "lang"=>"en", "lat"=>43.740730, "lon"=>-92.134000),
"/US/US-MN/foxhome-10436" => array("type"=>"city", "pop"=>126, "name_ru"=>"Фоксхоум", "name_en"=>"Foxhome", "lang"=>"en", "lat"=>46.276738, "lon"=>-96.310228),
"/US/US-MN/franklin-10437" => array("type"=>"city", "pop"=>465, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>44.530854, "lon"=>-94.886123),
"/US/US-MN/franklin-10438" => array("type"=>"city", "pop"=>22, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>47.523644, "lon"=>-92.519954),
"/US/US-MN/frazee-10439" => array("type"=>"city", "pop"=>1375, "name_ru"=>"Фрази", "name_en"=>"Frazee", "lang"=>"en", "lat"=>46.727005, "lon"=>-95.699569),
"/US/US-MN/freeborn-10440" => array("type"=>"city", "pop"=>286, "name_ru"=>"Фриборн", "name_en"=>"Freeborn", "lang"=>"en", "lat"=>43.766389, "lon"=>-93.563177),
"/US/US-MN/freeport-10441" => array("type"=>"city", "pop"=>469, "name_ru"=>"Фрипорт", "name_en"=>"Freeport", "lang"=>"en", "lat"=>45.662979, "lon"=>-94.687158),
"/US/US-MN/fridley-10442" => array("type"=>"city", "pop"=>25868, "name_ru"=>"Фридли", "name_en"=>"Fridley", "lang"=>"en", "lat"=>45.084767, "lon"=>-93.258919),
"/US/US-MN/frost-10443" => array("type"=>"city", "pop"=>221, "name_ru"=>"Фрост", "name_en"=>"Frost", "lang"=>"en", "lat"=>43.583724, "lon"=>-93.925371),
"/US/US-MN/fulda-10444" => array("type"=>"city", "pop"=>1196, "name_ru"=>"Фулда", "name_en"=>"Fulda", "lang"=>"en", "lat"=>43.869996, "lon"=>-95.602296),
"/US/US-MN/funkley-10445" => array("type"=>"city", "pop"=>16, "name_ru"=>"Фанкли", "name_en"=>"Funkley", "lang"=>"en", "lat"=>47.786394, "lon"=>-94.427479),
"/US/US-MN/garfield-10446" => array("type"=>"city", "pop"=>242, "name_ru"=>"Гарфилд", "name_en"=>"Garfield", "lang"=>"en", "lat"=>45.940351, "lon"=>-95.494442),
"/US/US-MN/garrison-10447" => array("type"=>"city", "pop"=>215, "name_ru"=>"Гаррисон", "name_en"=>"Garrison", "lang"=>"en", "lat"=>46.302644, "lon"=>-93.823463),
"/US/US-MN/garvin-10448" => array("type"=>"city", "pop"=>141, "name_ru"=>"Гарвин", "name_en"=>"Garvin", "lang"=>"en", "lat"=>44.214750, "lon"=>-95.760944),
"/US/US-MN/gary-10449" => array("type"=>"city", "pop"=>194, "name_ru"=>"Гэри", "name_en"=>"Gary", "lang"=>"en", "lat"=>47.372828, "lon"=>-96.265055),
"/US/US-MN/gaylord-10450" => array("type"=>"city", "pop"=>2123, "name_ru"=>"Гейлорд", "name_en"=>"Gaylord", "lang"=>"en", "lat"=>44.554521, "lon"=>-94.218228),
"/US/US-MN/gem-lake-10451" => array("type"=>"city", "pop"=>442, "name_ru"=>"Джем Лейк", "name_en"=>"Gem Lake", "lang"=>"en", "lat"=>45.058550, "lon"=>-93.041598),
"/US/US-MN/geneva-10452" => array("type"=>"city", "pop"=>445, "name_ru"=>"Дженива", "name_en"=>"Geneva", "lang"=>"en", "lat"=>43.823921, "lon"=>-93.268270),
"/US/US-MN/genola-10453" => array("type"=>"city", "pop"=>80, "name_ru"=>"Дженола", "name_en"=>"Genola", "lang"=>"en", "lat"=>45.968750, "lon"=>-94.111037),
"/US/US-MN/georgetown-10454" => array("type"=>"city", "pop"=>128, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>47.078139, "lon"=>-96.794847),
"/US/US-MN/ghent-10455" => array("type"=>"city", "pop"=>367, "name_ru"=>"Гент", "name_en"=>"Ghent", "lang"=>"en", "lat"=>44.512803, "lon"=>-95.892357),
"/US/US-MN/gibbon-10456" => array("type"=>"city", "pop"=>738, "name_ru"=>"Гиббон", "name_en"=>"Gibbon", "lang"=>"en", "lat"=>44.535133, "lon"=>-94.522999),
"/US/US-MN/gilbert-10457" => array("type"=>"city", "pop"=>1755, "name_ru"=>"Гилберт", "name_en"=>"Gilbert", "lang"=>"en", "lat"=>47.491865, "lon"=>-92.460521),
"/US/US-MN/gilman-10458" => array("type"=>"city", "pop"=>202, "name_ru"=>"Гилман", "name_en"=>"Gilman", "lang"=>"en", "lat"=>45.735265, "lon"=>-93.950116),
"/US/US-MN/glencoe-10459" => array("type"=>"city", "pop"=>5550, "name_ru"=>"Гленко", "name_en"=>"Glencoe", "lang"=>"en", "lat"=>44.769543, "lon"=>-94.149478),
"/US/US-MN/glenville-10460" => array("type"=>"city", "pop"=>678, "name_ru"=>"Гленвилл", "name_en"=>"Glenville", "lang"=>"en", "lat"=>43.568968, "lon"=>-93.275618),
"/US/US-MN/glenwood-10461" => array("type"=>"city", "pop"=>2525, "name_ru"=>"Гленвуд", "name_en"=>"Glenwood", "lang"=>"en", "lat"=>45.653295, "lon"=>-95.382812),
"/US/US-MN/glyndon-10462" => array("type"=>"city", "pop"=>1155, "name_ru"=>"Глиндон", "name_en"=>"Glyndon", "lang"=>"en", "lat"=>46.872709, "lon"=>-96.579378),
"/US/US-MN/golden-valley-10463" => array("type"=>"city", "pop"=>20344, "name_ru"=>"Голден Валли", "name_en"=>"Golden Valley", "lang"=>"en", "lat"=>44.989867, "lon"=>-93.358633),
"/US/US-MN/gonvick-10464" => array("type"=>"city", "pop"=>287, "name_ru"=>"Гонвик", "name_en"=>"Gonvick", "lang"=>"en", "lat"=>47.738256, "lon"=>-95.512505),
"/US/US-MN/good-thunder-10467" => array("type"=>"city", "pop"=>569, "name_ru"=>"Гуд Тандер", "name_en"=>"Good Thunder", "lang"=>"en", "lat"=>44.006676, "lon"=>-94.070141),
"/US/US-MN/goodhue-10465" => array("type"=>"city", "pop"=>989, "name_ru"=>"Гудхью", "name_en"=>"Goodhue", "lang"=>"en", "lat"=>44.400564, "lon"=>-92.621580),
"/US/US-MN/goodridge-10466" => array("type"=>"city", "pop"=>100, "name_ru"=>"Гудридж", "name_en"=>"Goodridge", "lang"=>"en", "lat"=>48.144703, "lon"=>-95.803965),
"/US/US-MN/goodview-10468" => array("type"=>"city", "pop"=>3565, "name_ru"=>"Гудвью", "name_en"=>"Goodview", "lang"=>"en", "lat"=>44.067232, "lon"=>-91.710733),
"/US/US-MN/graceville-10469" => array("type"=>"city", "pop"=>557, "name_ru"=>"Грэйсвилл", "name_en"=>"Graceville", "lang"=>"en", "lat"=>45.568482, "lon"=>-96.437491),
"/US/US-MN/granada-10470" => array("type"=>"city", "pop"=>294, "name_ru"=>"Гранада", "name_en"=>"Granada", "lang"=>"en", "lat"=>43.693616, "lon"=>-94.349283),
"/US/US-MN/grand-marais-10471" => array("type"=>"city", "pop"=>1418, "name_ru"=>"Гранд Мараис", "name_en"=>"Grand Marais", "lang"=>"en", "lat"=>47.755209, "lon"=>-90.344837),
"/US/US-MN/grand-meadow-10472" => array("type"=>"city", "pop"=>916, "name_ru"=>"Гранд Медоу", "name_en"=>"Grand Meadow", "lang"=>"en", "lat"=>43.705207, "lon"=>-92.571843),
"/US/US-MN/grand-rapids-10473" => array("type"=>"city", "pop"=>8743, "name_ru"=>"Гранд Рэпидс", "name_en"=>"Grand Rapids", "lang"=>"en", "lat"=>47.228766, "lon"=>-93.524536),
"/US/US-MN/granite-falls-10474" => array("type"=>"city", "pop"=>2908, "name_ru"=>"Гранти Фоллс", "name_en"=>"Granite Falls", "lang"=>"en", "lat"=>44.811064, "lon"=>-95.538695),
"/US/US-MN/grasston-10475" => array("type"=>"city", "pop"=>114, "name_ru"=>"Грасстон", "name_en"=>"Grasston", "lang"=>"en", "lat"=>45.794485, "lon"=>-93.153882),
"/US/US-MN/green-isle-10478" => array("type"=>"city", "pop"=>435, "name_ru"=>"Грин Айл", "name_en"=>"Green Isle", "lang"=>"en", "lat"=>44.678146, "lon"=>-94.006185),
"/US/US-MN/greenbush-10476" => array("type"=>"city", "pop"=>739, "name_ru"=>"Гринбуш", "name_en"=>"Greenbush", "lang"=>"en", "lat"=>48.698239, "lon"=>-96.183912),
"/US/US-MN/greenfield-10477" => array("type"=>"city", "pop"=>3110, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>45.104900, "lon"=>-93.696359),
"/US/US-MN/greenwald-10479" => array("type"=>"city", "pop"=>177, "name_ru"=>"Гринуолд", "name_en"=>"Greenwald", "lang"=>"en", "lat"=>45.602209, "lon"=>-94.859010),
"/US/US-MN/greenwood-10480" => array("type"=>"city", "pop"=>812, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>44.908310, "lon"=>-93.550531),
"/US/US-MN/grey-eagle-10481" => array("type"=>"city", "pop"=>326, "name_ru"=>"Грей Игл", "name_en"=>"Grey Eagle", "lang"=>"en", "lat"=>45.825048, "lon"=>-94.748648),
"/US/US-MN/grove-city-10482" => array("type"=>"city", "pop"=>581, "name_ru"=>"Гроув Сити", "name_en"=>"Grove City", "lang"=>"en", "lat"=>45.151891, "lon"=>-94.681368),
"/US/US-MN/grygla-10483" => array("type"=>"city", "pop"=>224, "name_ru"=>"Грайгла", "name_en"=>"Grygla", "lang"=>"en", "lat"=>48.299676, "lon"=>-95.619657),
"/US/US-MN/gully-10484" => array("type"=>"city", "pop"=>102, "name_ru"=>"Галли", "name_en"=>"Gully", "lang"=>"en", "lat"=>47.769030, "lon"=>-95.623470),
"/US/US-MN/hackensack-10485" => array("type"=>"city", "pop"=>283, "name_ru"=>"Хэкенсэк", "name_en"=>"Hackensack", "lang"=>"en", "lat"=>46.927861, "lon"=>-94.518580),
"/US/US-MN/hadley-10486" => array("type"=>"city", "pop"=>70, "name_ru"=>"Хэдли", "name_en"=>"Hadley", "lang"=>"en", "lat"=>43.997437, "lon"=>-95.858131),
"/US/US-MN/hallock-10487" => array("type"=>"city", "pop"=>1020, "name_ru"=>"Холлок", "name_en"=>"Hallock", "lang"=>"en", "lat"=>48.766411, "lon"=>-96.943588),
"/US/US-MN/halma-10488" => array("type"=>"city", "pop"=>66, "name_ru"=>"Халма", "name_en"=>"Halma", "lang"=>"en", "lat"=>48.660796, "lon"=>-96.597633),
"/US/US-MN/halstad-10489" => array("type"=>"city", "pop"=>541, "name_ru"=>"Халстэд", "name_en"=>"Halstad", "lang"=>"en", "lat"=>47.351164, "lon"=>-96.823993),
"/US/US-MN/ham-lake-10491" => array("type"=>"city", "pop"=>15020, "name_ru"=>"Хэм Лейк", "name_en"=>"Ham Lake", "lang"=>"en", "lat"=>45.260000, "lon"=>-93.200800),
"/US/US-MN/hamburg-10490" => array("type"=>"city", "pop"=>546, "name_ru"=>"Гамбург", "name_en"=>"Hamburg", "lang"=>"en", "lat"=>44.731600, "lon"=>-93.963180),
"/US/US-MN/hammond-10492" => array("type"=>"city", "pop"=>181, "name_ru"=>"Хаммонд", "name_en"=>"Hammond", "lang"=>"en", "lat"=>44.221744, "lon"=>-92.373747),
"/US/US-MN/hampton-10493" => array("type"=>"city", "pop"=>657, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>44.609324, "lon"=>-93.002298),
"/US/US-MN/hancock-10494" => array("type"=>"city", "pop"=>685, "name_ru"=>"Хэнкок", "name_en"=>"Hancock", "lang"=>"en", "lat"=>45.498285, "lon"=>-95.793876),
"/US/US-MN/hanley-falls-10495" => array("type"=>"city", "pop"=>285, "name_ru"=>"Хэнли Фоллс", "name_en"=>"Hanley Falls", "lang"=>"en", "lat"=>44.691918, "lon"=>-95.619111),
"/US/US-MN/hanover-10496" => array("type"=>"city", "pop"=>2423, "name_ru"=>"Хановер", "name_en"=>"Hanover", "lang"=>"en", "lat"=>45.156130, "lon"=>-93.658607),
"/US/US-MN/hanska-10497" => array("type"=>"city", "pop"=>409, "name_ru"=>"Ханска", "name_en"=>"Hanska", "lang"=>"en", "lat"=>44.148095, "lon"=>-94.494552),
"/US/US-MN/harding-10498" => array("type"=>"city", "pop"=>106, "name_ru"=>"Хардинг", "name_en"=>"Harding", "lang"=>"en", "lat"=>46.120805, "lon"=>-94.043807),
"/US/US-MN/hardwick-10499" => array("type"=>"city", "pop"=>205, "name_ru"=>"Хардвик", "name_en"=>"Hardwick", "lang"=>"en", "lat"=>43.775205, "lon"=>-96.196735),
"/US/US-MN/harmony-10500" => array("type"=>"city", "pop"=>1076, "name_ru"=>"Хармони", "name_en"=>"Harmony", "lang"=>"en", "lat"=>43.555304, "lon"=>-92.009383),
"/US/US-MN/harris-10501" => array("type"=>"city", "pop"=>1252, "name_ru"=>"Харрис", "name_en"=>"Harris", "lang"=>"en", "lat"=>45.602868, "lon"=>-92.989142),
"/US/US-MN/hartland-10502" => array("type"=>"city", "pop"=>270, "name_ru"=>"Хартленд", "name_en"=>"Hartland", "lang"=>"en", "lat"=>43.804264, "lon"=>-93.485624),
"/US/US-MN/hastings-10503" => array("type"=>"city", "pop"=>22126, "name_ru"=>"Гастингс", "name_en"=>"Hastings", "lang"=>"en", "lat"=>44.732850, "lon"=>-92.852502),
"/US/US-MN/hatfield-10504" => array("type"=>"city", "pop"=>43, "name_ru"=>"Хэтфилд", "name_en"=>"Hatfield", "lang"=>"en", "lat"=>43.951638, "lon"=>-96.189094),
"/US/US-MN/hawley-10505" => array("type"=>"city", "pop"=>1874, "name_ru"=>"Холи", "name_en"=>"Hawley", "lang"=>"en", "lat"=>46.878187, "lon"=>-96.316208),
"/US/US-MN/hayfield-10506" => array("type"=>"city", "pop"=>1337, "name_ru"=>"Хэйфилд", "name_en"=>"Hayfield", "lang"=>"en", "lat"=>43.890115, "lon"=>-92.847030),
"/US/US-MN/hayward-10507" => array("type"=>"city", "pop"=>234, "name_ru"=>"Хейворд", "name_en"=>"Hayward", "lang"=>"en", "lat"=>43.649354, "lon"=>-93.246704),
"/US/US-MN/hazel-run-10508" => array("type"=>"city", "pop"=>55, "name_ru"=>"Хейзел Ран", "name_en"=>"Hazel Run", "lang"=>"en", "lat"=>44.749780, "lon"=>-95.716136),
"/US/US-MN/hector-10509" => array("type"=>"city", "pop"=>1072, "name_ru"=>"Гектор", "name_en"=>"Hector", "lang"=>"en", "lat"=>44.741432, "lon"=>-94.712689),
"/US/US-MN/heidelberg-10510" => array("type"=>"city", "pop"=>129, "name_ru"=>"Хайделберг", "name_en"=>"Heidelberg", "lang"=>"en", "lat"=>44.493378, "lon"=>-93.623605),
"/US/US-MN/henderson-10511" => array("type"=>"city", "pop"=>918, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>44.530250, "lon"=>-93.907949),
"/US/US-MN/hendricks-10512" => array("type"=>"city", "pop"=>648, "name_ru"=>"Хендрикс", "name_en"=>"Hendricks", "lang"=>"en", "lat"=>44.508624, "lon"=>-96.426307),
"/US/US-MN/hendrum-10513" => array("type"=>"city", "pop"=>285, "name_ru"=>"Хендрам", "name_en"=>"Hendrum", "lang"=>"en", "lat"=>47.263839, "lon"=>-96.810241),
"/US/US-MN/henning-10514" => array("type"=>"city", "pop"=>800, "name_ru"=>"Хеннинг", "name_en"=>"Henning", "lang"=>"en", "lat"=>46.322377, "lon"=>-95.442898),
"/US/US-MN/henriette-10515" => array("type"=>"city", "pop"=>106, "name_ru"=>"Хенриетт", "name_en"=>"Henriette", "lang"=>"en", "lat"=>45.871338, "lon"=>-93.120023),
"/US/US-MN/herman-10516" => array("type"=>"city", "pop"=>422, "name_ru"=>"Херман", "name_en"=>"Herman", "lang"=>"en", "lat"=>45.810610, "lon"=>-96.139928),
"/US/US-MN/hermantown-10517" => array("type"=>"city", "pop"=>9361, "name_ru"=>"Хермантаун", "name_en"=>"Hermantown", "lang"=>"en", "lat"=>46.806205, "lon"=>-92.243804),
"/US/US-MN/heron-lake-10518" => array("type"=>"city", "pop"=>733, "name_ru"=>"Херон Лейк", "name_en"=>"Heron Lake", "lang"=>"en", "lat"=>43.795336, "lon"=>-95.323641),
"/US/US-MN/hewitt-10519" => array("type"=>"city", "pop"=>259, "name_ru"=>"Хьюитт", "name_en"=>"Hewitt", "lang"=>"en", "lat"=>46.322936, "lon"=>-95.087899),
"/US/US-MN/hibbing-10520" => array("type"=>"city", "pop"=>16209, "name_ru"=>"Хиббинг", "name_en"=>"Hibbing", "lang"=>"en", "lat"=>47.399983, "lon"=>-92.948143),
"/US/US-MN/hill-city-10521" => array("type"=>"city", "pop"=>440, "name_ru"=>"Хилл Сити", "name_en"=>"Hill City", "lang"=>"en", "lat"=>46.986054, "lon"=>-93.596413),
"/US/US-MN/hillman-10522" => array("type"=>"city", "pop"=>32, "name_ru"=>"Хиллман", "name_en"=>"Hillman", "lang"=>"en", "lat"=>46.008981, "lon"=>-93.888171),
"/US/US-MN/hills-10523" => array("type"=>"city", "pop"=>594, "name_ru"=>"Хиллс", "name_en"=>"Hills", "lang"=>"en", "lat"=>43.526472, "lon"=>-96.360752),
"/US/US-MN/hilltop-10524" => array("type"=>"city", "pop"=>683, "name_ru"=>"Хиллтоп", "name_en"=>"Hilltop", "lang"=>"en", "lat"=>45.053348, "lon"=>-93.248851),
"/US/US-MN/hinckley-10525" => array("type"=>"city", "pop"=>1409, "name_ru"=>"Хинкли", "name_en"=>"Hinckley", "lang"=>"en", "lat"=>46.014510, "lon"=>-92.938386),
"/US/US-MN/hitterdal-10526" => array("type"=>"city", "pop"=>199, "name_ru"=>"Хиттердал", "name_en"=>"Hitterdal", "lang"=>"en", "lat"=>46.978464, "lon"=>-96.258164),
"/US/US-MN/hoffman-10527" => array("type"=>"city", "pop"=>640, "name_ru"=>"Хоффман", "name_en"=>"Hoffman", "lang"=>"en", "lat"=>45.831924, "lon"=>-95.785895),
"/US/US-MN/hokah-10528" => array("type"=>"city", "pop"=>561, "name_ru"=>"Хока", "name_en"=>"Hokah", "lang"=>"en", "lat"=>43.760100, "lon"=>-91.350016),
"/US/US-MN/holdingford-10529" => array("type"=>"city", "pop"=>723, "name_ru"=>"Холдингфорд", "name_en"=>"Holdingford", "lang"=>"en", "lat"=>45.730840, "lon"=>-94.471151),
"/US/US-MN/holland-10530" => array("type"=>"city", "pop"=>198, "name_ru"=>"Холланд", "name_en"=>"Holland", "lang"=>"en", "lat"=>44.090859, "lon"=>-96.191739),
"/US/US-MN/hollandale-10531" => array("type"=>"city", "pop"=>274, "name_ru"=>"Холландейл", "name_en"=>"Hollandale", "lang"=>"en", "lat"=>43.759834, "lon"=>-93.205389),
"/US/US-MN/holloway-10532" => array("type"=>"city", "pop"=>96, "name_ru"=>"Холлоуэй", "name_en"=>"Holloway", "lang"=>"en", "lat"=>45.247588, "lon"=>-95.910531),
"/US/US-MN/holt-10533" => array("type"=>"city", "pop"=>84, "name_ru"=>"Холт", "name_en"=>"Holt", "lang"=>"en", "lat"=>48.292124, "lon"=>-96.193941),
"/US/US-MN/hopkins-10534" => array("type"=>"city", "pop"=>16961, "name_ru"=>"Хопкинс", "name_en"=>"Hopkins", "lang"=>"en", "lat"=>44.926700, "lon"=>-93.405282),
"/US/US-MN/houston-10535" => array("type"=>"city", "pop"=>956, "name_ru"=>"Хьюстон", "name_en"=>"Houston", "lang"=>"en", "lat"=>43.760691, "lon"=>-91.572312),
"/US/US-MN/howard-lake-10536" => array("type"=>"city", "pop"=>1987, "name_ru"=>"Ховард Лейк", "name_en"=>"Howard Lake", "lang"=>"en", "lat"=>45.059579, "lon"=>-94.065774),
"/US/US-MN/hoyt-lakes-10537" => array("type"=>"city", "pop"=>1956, "name_ru"=>"Хойт Лейкс", "name_en"=>"Hoyt Lakes", "lang"=>"en", "lat"=>47.559406, "lon"=>-92.119667),
"/US/US-MN/hugo-10538" => array("type"=>"city", "pop"=>13139, "name_ru"=>"Хьюго", "name_en"=>"Hugo", "lang"=>"en", "lat"=>45.159091, "lon"=>-92.963030),
"/US/US-MN/humboldt-10539" => array("type"=>"city", "pop"=>52, "name_ru"=>"Гумбольдт", "name_en"=>"Humboldt", "lang"=>"en", "lat"=>48.921454, "lon"=>-97.094340),
"/US/US-MN/hutchinson-10540" => array("type"=>"city", "pop"=>13913, "name_ru"=>"Хатчинсон", "name_en"=>"Hutchinson", "lang"=>"en", "lat"=>44.888827, "lon"=>-94.374744),
"/US/US-MN/ihlen-10541" => array("type"=>"city", "pop"=>103, "name_ru"=>"Илен", "name_en"=>"Ihlen", "lang"=>"en", "lat"=>43.910010, "lon"=>-96.371934),
"/US/US-MN/independence-10542" => array("type"=>"city", "pop"=>3546, "name_ru"=>"Индепенденс", "name_en"=>"Independence", "lang"=>"en", "lat"=>45.022050, "lon"=>-93.705186),
"/US/US-MN/international-falls-10543" => array("type"=>"city", "pop"=>5906, "name_ru"=>"Интернешнл Фоллс", "name_en"=>"International Falls", "lang"=>"en", "lat"=>48.587622, "lon"=>-93.406181),
"/US/US-MN/inver-grove-heights-10544" => array("type"=>"city", "pop"=>33383, "name_ru"=>"Инвер Гров Хайтс", "name_en"=>"Inver Grove Heights", "lang"=>"en", "lat"=>44.825233, "lon"=>-93.057742),
"/US/US-MN/iona-10545" => array("type"=>"city", "pop"=>155, "name_ru"=>"Айона", "name_en"=>"Iona", "lang"=>"en", "lat"=>43.914934, "lon"=>-95.785024),
"/US/US-MN/iron-junction-10546" => array("type"=>"city", "pop"=>85, "name_ru"=>"Айрон Джанкшен", "name_en"=>"Iron Junction", "lang"=>"en", "lat"=>47.416427, "lon"=>-92.606650),
"/US/US-MN/ironton-10547" => array("type"=>"city", "pop"=>528, "name_ru"=>"Айронтон", "name_en"=>"Ironton", "lang"=>"en", "lat"=>46.480930, "lon"=>-93.996650),
"/US/US-MN/isanti-10548" => array("type"=>"city", "pop"=>6123, "name_ru"=>"Исанти", "name_en"=>"Isanti", "lang"=>"en", "lat"=>45.489858, "lon"=>-93.252619),
"/US/US-MN/island-view-10549" => array("type"=>"city", "pop"=>150, "name_ru"=>"Айленд Вью", "name_en"=>"Island View", "lang"=>"en", "lat"=>48.605963, "lon"=>-93.190687),
"/US/US-MN/isle-10550" => array("type"=>"city", "pop"=>909, "name_ru"=>"Айл", "name_en"=>"Isle", "lang"=>"en", "lat"=>46.142212, "lon"=>-93.460424),
"/US/US-MN/ivanhoe-10551" => array("type"=>"city", "pop"=>589, "name_ru"=>"Айванхо", "name_en"=>"Ivanhoe", "lang"=>"en", "lat"=>44.464102, "lon"=>-96.250410),
"/US/US-MN/jackson-10552" => array("type"=>"city", "pop"=>3305, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>43.624364, "lon"=>-94.988967),
"/US/US-MN/janesville-10553" => array("type"=>"city", "pop"=>2242, "name_ru"=>"Джейнсвилл", "name_en"=>"Janesville", "lang"=>"en", "lat"=>44.116655, "lon"=>-93.707807),
"/US/US-MN/jasper-10554" => array("type"=>"city", "pop"=>564, "name_ru"=>"Джаспер", "name_en"=>"Jasper", "lang"=>"en", "lat"=>43.849642, "lon"=>-96.399508),
"/US/US-MN/jeffers-10555" => array("type"=>"city", "pop"=>362, "name_ru"=>"Джефферс", "name_en"=>"Jeffers", "lang"=>"en", "lat"=>44.055510, "lon"=>-95.195454),
"/US/US-MN/jenkins-10556" => array("type"=>"city", "pop"=>403, "name_ru"=>"Дженкинс", "name_en"=>"Jenkins", "lang"=>"en", "lat"=>46.652315, "lon"=>-94.328960),
"/US/US-MN/johnson-10557" => array("type"=>"city", "pop"=>28, "name_ru"=>"Джонсон", "name_en"=>"Johnson", "lang"=>"en", "lat"=>45.571884, "lon"=>-96.293777),
"/US/US-MN/jordan-10558" => array("type"=>"city", "pop"=>5445, "name_ru"=>"Джордан", "name_en"=>"Jordan", "lang"=>"en", "lat"=>44.668497, "lon"=>-93.632249),
"/US/US-MN/kandiyohi-10559" => array("type"=>"city", "pop"=>530, "name_ru"=>"Кандийоай", "name_en"=>"Kandiyohi", "lang"=>"en", "lat"=>45.131104, "lon"=>-94.932315),
"/US/US-MN/karlstad-10560" => array("type"=>"city", "pop"=>674, "name_ru"=>"Карлстад", "name_en"=>"Karlstad", "lang"=>"en", "lat"=>48.578096, "lon"=>-96.515583),
"/US/US-MN/kasota-10561" => array("type"=>"city", "pop"=>690, "name_ru"=>"Касота", "name_en"=>"Kasota", "lang"=>"en", "lat"=>44.290247, "lon"=>-93.968192),
"/US/US-MN/kasson-10562" => array("type"=>"city", "pop"=>5564, "name_ru"=>"Кассон", "name_en"=>"Kasson", "lang"=>"en", "lat"=>44.029248, "lon"=>-92.752884),
"/US/US-MN/keewatin-10563" => array("type"=>"city", "pop"=>1070, "name_ru"=>"Киуотин", "name_en"=>"Keewatin", "lang"=>"en", "lat"=>47.404108, "lon"=>-93.085511),
"/US/US-MN/kelliher-10564" => array("type"=>"city", "pop"=>306, "name_ru"=>"Келлиер", "name_en"=>"Kelliher", "lang"=>"en", "lat"=>47.943000, "lon"=>-94.445999),
"/US/US-MN/kellogg-10565" => array("type"=>"city", "pop"=>472, "name_ru"=>"Келлогг", "name_en"=>"Kellogg", "lang"=>"en", "lat"=>44.307048, "lon"=>-91.998804),
"/US/US-MN/kennedy-10566" => array("type"=>"city", "pop"=>211, "name_ru"=>"Кеннеди", "name_en"=>"Kennedy", "lang"=>"en", "lat"=>48.643456, "lon"=>-96.910108),
"/US/US-MN/kenneth-10567" => array("type"=>"city", "pop"=>59, "name_ru"=>"Кеннет", "name_en"=>"Kenneth", "lang"=>"en", "lat"=>43.753267, "lon"=>-96.071818),
"/US/US-MN/kensington-10568" => array("type"=>"city", "pop"=>267, "name_ru"=>"Кенсингтон", "name_en"=>"Kensington", "lang"=>"en", "lat"=>45.777914, "lon"=>-95.695324),
"/US/US-MN/kent-10569" => array("type"=>"city", "pop"=>100, "name_ru"=>"Кент", "name_en"=>"Kent", "lang"=>"en", "lat"=>46.439614, "lon"=>-96.681103),
"/US/US-MN/kenyon-10570" => array("type"=>"city", "pop"=>1680, "name_ru"=>"Кеньон", "name_en"=>"Kenyon", "lang"=>"en", "lat"=>44.272420, "lon"=>-92.985834),
"/US/US-MN/kerkhoven-10571" => array("type"=>"city", "pop"=>674, "name_ru"=>"Керкховен", "name_en"=>"Kerkhoven", "lang"=>"en", "lat"=>45.192934, "lon"=>-95.319958),
"/US/US-MN/kerrick-10572" => array("type"=>"city", "pop"=>71, "name_ru"=>"Керрик", "name_en"=>"Kerrick", "lang"=>"en", "lat"=>46.339710, "lon"=>-92.585580),
"/US/US-MN/kettle-river-10573" => array("type"=>"city", "pop"=>171, "name_ru"=>"Кеттл Ривер", "name_en"=>"Kettle River", "lang"=>"en", "lat"=>46.486409, "lon"=>-92.878017),
"/US/US-MN/kiester-10574" => array("type"=>"city", "pop"=>477, "name_ru"=>"Кистер", "name_en"=>"Kiester", "lang"=>"en", "lat"=>43.536637, "lon"=>-93.711413),
"/US/US-MN/kilkenny-10575" => array("type"=>"city", "pop"=>154, "name_ru"=>"Килкенни", "name_en"=>"Kilkenny", "lang"=>"en", "lat"=>44.313199, "lon"=>-93.574346),
"/US/US-MN/kimball-prairie-10576" => array("type"=>"city", "pop"=>690, "name_ru"=>"Кимбалл", "name_en"=>"Kimball Prairie", "lang"=>"en", "lat"=>45.313317, "lon"=>-94.300803),
"/US/US-MN/kinbrae-10577" => array("type"=>"city", "pop"=>21, "name_ru"=>"Кинбро", "name_en"=>"Kinbrae", "lang"=>"en", "lat"=>43.826329, "lon"=>-95.480170),
"/US/US-MN/kingston-10578" => array("type"=>"city", "pop"=>118, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>45.196651, "lon"=>-94.309741),
"/US/US-MN/kinney-10579" => array("type"=>"city", "pop"=>190, "name_ru"=>"Кинни", "name_en"=>"Kinney", "lang"=>"en", "lat"=>47.522344, "lon"=>-92.725270),
"/US/US-MN/la-crescent-10580" => array("type"=>"city", "pop"=>4912, "name_ru"=>"Ла Кресент", "name_en"=>"La Crescent", "lang"=>"en", "lat"=>43.829400, "lon"=>-91.300843),
"/US/US-MN/la-prairie-10602" => array("type"=>"city", "pop"=>593, "name_ru"=>"Ла Прейри", "name_en"=>"La Prairie", "lang"=>"en", "lat"=>47.227320, "lon"=>-93.490079),
"/US/US-MN/la-salle-10603" => array("type"=>"city", "pop"=>83, "name_ru"=>"Ла Саль", "name_en"=>"La Salle", "lang"=>"en", "lat"=>44.072086, "lon"=>-94.569983),
"/US/US-MN/lafayette-10581" => array("type"=>"city", "pop"=>515, "name_ru"=>"Ла Файет", "name_en"=>"Lafayette", "lang"=>"en", "lat"=>44.448281, "lon"=>-94.393294),
"/US/US-MN/lake-benton-10582" => array("type"=>"city", "pop"=>633, "name_ru"=>"Лейк Бентон", "name_en"=>"Lake Benton", "lang"=>"en", "lat"=>44.262551, "lon"=>-96.292547),
"/US/US-MN/lake-bronson-10583" => array("type"=>"city", "pop"=>196, "name_ru"=>"Лейк Бронсон", "name_en"=>"Lake Bronson", "lang"=>"en", "lat"=>48.732592, "lon"=>-96.662562),
"/US/US-MN/lake-city-10584" => array("type"=>"city", "pop"=>5285, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>44.446533, "lon"=>-92.278112),
"/US/US-MN/lake-crystal-10585" => array("type"=>"city", "pop"=>2598, "name_ru"=>"Лейк Кристал", "name_en"=>"Lake Crystal", "lang"=>"en", "lat"=>44.103409, "lon"=>-94.217851),
"/US/US-MN/lake-elmo-10586" => array("type"=>"city", "pop"=>7484, "name_ru"=>"Лейк Элмо", "name_en"=>"Lake Elmo", "lang"=>"en", "lat"=>44.995350, "lon"=>-92.907744),
"/US/US-MN/lake-henry-10588" => array("type"=>"city", "pop"=>86, "name_ru"=>"Лейк Хенри", "name_en"=>"Lake Henry", "lang"=>"en", "lat"=>45.457808, "lon"=>-94.797178),
"/US/US-MN/lake-lillian-10591" => array("type"=>"city", "pop"=>244, "name_ru"=>"Лейк Лиллиан", "name_en"=>"Lake Lillian", "lang"=>"en", "lat"=>44.945745, "lon"=>-94.880520),
"/US/US-MN/lake-park-10592" => array("type"=>"city", "pop"=>812, "name_ru"=>"Лэйк Парк", "name_en"=>"Lake Park", "lang"=>"en", "lat"=>46.883579, "lon"=>-96.096480),
"/US/US-MN/lake-shore-10594" => array("type"=>"city", "pop"=>1068, "name_ru"=>"Лейк Шор", "name_en"=>"Lake Shore", "lang"=>"en", "lat"=>46.495343, "lon"=>-94.360961),
"/US/US-MN/lake-st.-croix-beach-10593" => array("type"=>"city", "pop"=>1094, "name_ru"=>"Лейк Сент Круа Бич", "name_en"=>"Lake St. Croix Beach", "lang"=>"en", "lat"=>44.924400, "lon"=>-92.765528),
"/US/US-MN/lake-wilson-10596" => array("type"=>"city", "pop"=>232, "name_ru"=>"Лейк Уилсон", "name_en"=>"Lake Wilson", "lang"=>"en", "lat"=>43.994810, "lon"=>-95.953011),
"/US/US-MN/lakefield-10587" => array("type"=>"city", "pop"=>1614, "name_ru"=>"Лейкфилд", "name_en"=>"Lakefield", "lang"=>"en", "lat"=>43.677914, "lon"=>-95.171063),
"/US/US-MN/lakeland-10589" => array("type"=>"city", "pop"=>1791, "name_ru"=>"Лэйклэнд", "name_en"=>"Lakeland", "lang"=>"en", "lat"=>44.950550, "lon"=>-92.770599),
"/US/US-MN/lakeland-shores-10590" => array("type"=>"city", "pop"=>342, "name_ru"=>"Лейкленд Шорс", "name_en"=>"Lakeland Shores", "lang"=>"en", "lat"=>44.949167, "lon"=>-92.763205),
"/US/US-MN/lakeville-10595" => array("type"=>"city", "pop"=>54542, "name_ru"=>"Лейквилл", "name_en"=>"Lakeville", "lang"=>"en", "lat"=>44.679367, "lon"=>-93.244813),
"/US/US-MN/lamberton-10597" => array("type"=>"city", "pop"=>771, "name_ru"=>"Лэмбертон", "name_en"=>"Lamberton", "lang"=>"en", "lat"=>44.229729, "lon"=>-95.264132),
"/US/US-MN/lancaster-10598" => array("type"=>"city", "pop"=>305, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>48.857915, "lon"=>-96.802812),
"/US/US-MN/landfall-10599" => array("type"=>"city", "pop"=>634, "name_ru"=>"Лендфолл", "name_en"=>"Landfall", "lang"=>"en", "lat"=>44.950100, "lon"=>-92.977449),
"/US/US-MN/lanesboro-10600" => array("type"=>"city", "pop"=>728, "name_ru"=>"Лейнсборо", "name_en"=>"Lanesboro", "lang"=>"en", "lat"=>43.719529, "lon"=>-91.973542),
"/US/US-MN/laporte-10601" => array("type"=>"city", "pop"=>148, "name_ru"=>"Лапорт", "name_en"=>"Laporte", "lang"=>"en", "lat"=>47.214286, "lon"=>-94.757249),
"/US/US-MN/lastrup-10604" => array("type"=>"city", "pop"=>87, "name_ru"=>"Ластреп", "name_en"=>"Lastrup", "lang"=>"en", "lat"=>46.039818, "lon"=>-94.061275),
"/US/US-MN/lauderdale-10605" => array("type"=>"city", "pop"=>2212, "name_ru"=>"Лодердейл", "name_en"=>"Lauderdale", "lang"=>"en", "lat"=>44.994400, "lon"=>-93.202874),
"/US/US-MN/le-center-10606" => array("type"=>"city", "pop"=>2286, "name_ru"=>"Ле Сентер", "name_en"=>"Le Center", "lang"=>"en", "lat"=>44.387579, "lon"=>-93.732948),
"/US/US-MN/le-roy-10610" => array("type"=>"city", "pop"=>906, "name_ru"=>"Ле Рой", "name_en"=>"Le Roy", "lang"=>"en", "lat"=>43.513607, "lon"=>-92.507778),
"/US/US-MN/le-sueur-10612" => array("type"=>"city", "pop"=>4252, "name_ru"=>"Ле Саер", "name_en"=>"Le Sueur", "lang"=>"en", "lat"=>44.463195, "lon"=>-93.906144),
"/US/US-MN/lengby-10607" => array("type"=>"city", "pop"=>76, "name_ru"=>"Ленгби", "name_en"=>"Lengby", "lang"=>"en", "lat"=>47.514888, "lon"=>-95.633503),
"/US/US-MN/leonard-10608" => array("type"=>"city", "pop"=>28, "name_ru"=>"Леонард", "name_en"=>"Leonard", "lang"=>"en", "lat"=>47.652385, "lon"=>-95.266366),
"/US/US-MN/leonidas-10609" => array("type"=>"city", "pop"=>55, "name_ru"=>"Леонидас", "name_en"=>"Leonidas", "lang"=>"en", "lat"=>47.467347, "lon"=>-92.567912),
"/US/US-MN/lester-prairie-10611" => array("type"=>"city", "pop"=>1747, "name_ru"=>"Лестер Прейри", "name_en"=>"Lester Prairie", "lang"=>"en", "lat"=>44.883724, "lon"=>-94.039744),
"/US/US-MN/lewiston-10613" => array("type"=>"city", "pop"=>1480, "name_ru"=>"Льюистон", "name_en"=>"Lewiston", "lang"=>"en", "lat"=>43.984931, "lon"=>-91.867904),
"/US/US-MN/lewisville-10614" => array("type"=>"city", "pop"=>253, "name_ru"=>"Льюисвилл", "name_en"=>"Lewisville", "lang"=>"en", "lat"=>43.924026, "lon"=>-94.435086),
"/US/US-MN/lexington-10615" => array("type"=>"city", "pop"=>1997, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>45.138750, "lon"=>-93.171091),
"/US/US-MN/lilydale-10616" => array("type"=>"city", "pop"=>615, "name_ru"=>"Лилидейл", "name_en"=>"Lilydale", "lang"=>"en", "lat"=>44.914133, "lon"=>-93.130009),
"/US/US-MN/lindstrom-10617" => array("type"=>"city", "pop"=>3995, "name_ru"=>"Линдстром", "name_en"=>"Lindstrom", "lang"=>"en", "lat"=>45.385887, "lon"=>-92.840380),
"/US/US-MN/lino-lakes-10618" => array("type"=>"city", "pop"=>20174, "name_ru"=>"Лино Лейкс", "name_en"=>"Lino Lakes", "lang"=>"en", "lat"=>45.161850, "lon"=>-93.082507),
"/US/US-MN/lismore-10619" => array("type"=>"city", "pop"=>224, "name_ru"=>"Лисмор", "name_en"=>"Lismore", "lang"=>"en", "lat"=>43.749361, "lon"=>-95.948183),
"/US/US-MN/litchfield-10620" => array("type"=>"city", "pop"=>6560, "name_ru"=>"Литчфилд", "name_en"=>"Litchfield", "lang"=>"en", "lat"=>45.119504, "lon"=>-94.527050),
"/US/US-MN/little-canada-10621" => array("type"=>"city", "pop"=>9570, "name_ru"=>"Литтл Канада", "name_en"=>"Little Canada", "lang"=>"en", "lat"=>45.023700, "lon"=>-93.083861),
"/US/US-MN/little-falls-10622" => array("type"=>"city", "pop"=>8121, "name_ru"=>"Литтл Фоллс", "name_en"=>"Little Falls", "lang"=>"en", "lat"=>45.977931, "lon"=>-94.361085),
"/US/US-MN/little-rock-10624" => array("type"=>"cdp", "pop"=>257, "name_ru"=>"Литл Рок", "name_en"=>"Little Rock", "lang"=>"en", "lat"=>47.867978, "lon"=>-95.110411),
"/US/US-MN/littlefork-10623" => array("type"=>"city", "pop"=>680, "name_ru"=>"Литтлфорк", "name_en"=>"Littlefork", "lang"=>"en", "lat"=>48.396455, "lon"=>-93.556807),
"/US/US-MN/long-beach-10625" => array("type"=>"city", "pop"=>300, "name_ru"=>"Лонг Бич", "name_en"=>"Long Beach", "lang"=>"en", "lat"=>45.648146, "lon"=>-95.429820),
"/US/US-MN/long-lake-10626" => array("type"=>"city", "pop"=>1795, "name_ru"=>"Лонг Лейк", "name_en"=>"Long Lake", "lang"=>"en", "lat"=>44.985350, "lon"=>-93.568760),
"/US/US-MN/long-prairie-10627" => array("type"=>"city", "pop"=>2809, "name_ru"=>"Лонг Прейри", "name_en"=>"Long Prairie", "lang"=>"en", "lat"=>45.979600, "lon"=>-94.863013),
"/US/US-MN/longville-10628" => array("type"=>"city", "pop"=>178, "name_ru"=>"Лонгвилл", "name_en"=>"Longville", "lang"=>"en", "lat"=>46.986719, "lon"=>-94.215181),
"/US/US-MN/lonsdale-10629" => array("type"=>"city", "pop"=>3041, "name_ru"=>"Лонсдейл", "name_en"=>"Lonsdale", "lang"=>"en", "lat"=>44.477648, "lon"=>-93.427759),
"/US/US-MN/loretto-10630" => array("type"=>"city", "pop"=>602, "name_ru"=>"Лоретто", "name_en"=>"Loretto", "lang"=>"en", "lat"=>45.055367, "lon"=>-93.635211),
"/US/US-MN/louisburg-10631" => array("type"=>"city", "pop"=>23, "name_ru"=>"Луисбург", "name_en"=>"Louisburg", "lang"=>"en", "lat"=>45.166151, "lon"=>-96.170107),
"/US/US-MN/lowry-10632" => array("type"=>"city", "pop"=>257, "name_ru"=>"Лоури", "name_en"=>"Lowry", "lang"=>"en", "lat"=>45.705517, "lon"=>-95.517180),
"/US/US-MN/lucan-10633" => array("type"=>"city", "pop"=>200, "name_ru"=>"Лукан", "name_en"=>"Lucan", "lang"=>"en", "lat"=>44.409275, "lon"=>-95.412690),
"/US/US-MN/luverne-10634" => array("type"=>"city", "pop"=>4425, "name_ru"=>"Леверн", "name_en"=>"Luverne", "lang"=>"en", "lat"=>43.654820, "lon"=>-96.209945),
"/US/US-MN/lyle-10635" => array("type"=>"city", "pop"=>558, "name_ru"=>"Лайл", "name_en"=>"Lyle", "lang"=>"en", "lat"=>43.505351, "lon"=>-92.941984),
"/US/US-MN/lynd-10636" => array("type"=>"city", "pop"=>404, "name_ru"=>"Линд", "name_en"=>"Lynd", "lang"=>"en", "lat"=>44.385099, "lon"=>-95.895578),
"/US/US-MN/mabel-10637" => array("type"=>"city", "pop"=>709, "name_ru"=>"Мейбел", "name_en"=>"Mabel", "lang"=>"en", "lat"=>43.520305, "lon"=>-91.768285),
"/US/US-MN/madelia-10642" => array("type"=>"city", "pop"=>2144, "name_ru"=>"Маделия", "name_en"=>"Madelia", "lang"=>"en", "lat"=>44.049265, "lon"=>-94.417123),
"/US/US-MN/madison-10643" => array("type"=>"city", "pop"=>1563, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>45.013764, "lon"=>-96.189096),
"/US/US-MN/madison-lake-10644" => array("type"=>"city", "pop"=>954, "name_ru"=>"Мэдисон Лейк", "name_en"=>"Madison Lake", "lang"=>"en", "lat"=>44.202198, "lon"=>-93.814387),
"/US/US-MN/magnolia-10645" => array("type"=>"city", "pop"=>208, "name_ru"=>"Магнолия", "name_en"=>"Magnolia", "lang"=>"en", "lat"=>43.644881, "lon"=>-96.074516),
"/US/US-MN/mahnomen-10646" => array("type"=>"city", "pop"=>1185, "name_ru"=>"Маномен", "name_en"=>"Mahnomen", "lang"=>"en", "lat"=>47.314796, "lon"=>-95.969136),
"/US/US-MN/mahtomedi-10647" => array("type"=>"city", "pop"=>8025, "name_ru"=>"Матомеди", "name_en"=>"Mahtomedi", "lang"=>"en", "lat"=>45.056100, "lon"=>-92.960815),
"/US/US-MN/manchester-10648" => array("type"=>"city", "pop"=>75, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>43.724946, "lon"=>-93.450641),
"/US/US-MN/manhattan-beach-10649" => array("type"=>"city", "pop"=>52, "name_ru"=>"Манхэттен Бич", "name_en"=>"Manhattan Beach", "lang"=>"en", "lat"=>46.734387, "lon"=>-94.144431),
"/US/US-MN/mankato-10650" => array("type"=>"city", "pop"=>36245, "name_ru"=>"Манкато", "name_en"=>"Mankato", "lang"=>"en", "lat"=>44.166661, "lon"=>-93.991068),
"/US/US-MN/mantorville-10651" => array("type"=>"city", "pop"=>1160, "name_ru"=>"Манторвилл", "name_en"=>"Mantorville", "lang"=>"en", "lat"=>44.065882, "lon"=>-92.753436),
"/US/US-MN/maple-grove-10652" => array("type"=>"city", "pop"=>61542, "name_ru"=>"Мейпл Гров", "name_en"=>"Maple Grove", "lang"=>"en", "lat"=>45.106018, "lon"=>-93.459963),
"/US/US-MN/maple-lake-10653" => array("type"=>"city", "pop"=>2035, "name_ru"=>"Мейпл Лейк", "name_en"=>"Maple Lake", "lang"=>"en", "lat"=>45.231364, "lon"=>-94.002952),
"/US/US-MN/maple-plain-10654" => array("type"=>"city", "pop"=>1983, "name_ru"=>"Мейпл Плейн", "name_en"=>"Maple Plain", "lang"=>"en", "lat"=>45.008900, "lon"=>-93.661499),
"/US/US-MN/mapleton-10655" => array("type"=>"city", "pop"=>1658, "name_ru"=>"Мейплтон", "name_en"=>"Mapleton", "lang"=>"en", "lat"=>43.925921, "lon"=>-93.953943),
"/US/US-MN/mapleview-10656" => array("type"=>"city", "pop"=>175, "name_ru"=>"Мейплвью", "name_en"=>"Mapleview", "lang"=>"en", "lat"=>43.688959, "lon"=>-92.974853),
"/US/US-MN/maplewood-10657" => array("type"=>"city", "pop"=>36087, "name_ru"=>"Мейплвуд", "name_en"=>"Maplewood", "lang"=>"en", "lat"=>44.992933, "lon"=>-93.023699),
"/US/US-MN/marble-10658" => array("type"=>"city", "pop"=>672, "name_ru"=>"Марбл", "name_en"=>"Marble", "lang"=>"en", "lat"=>47.328583, "lon"=>-93.293408),
"/US/US-MN/marietta-10659" => array("type"=>"city", "pop"=>151, "name_ru"=>"Мариетта", "name_en"=>"Marietta", "lang"=>"en", "lat"=>45.010340, "lon"=>-96.418011),
"/US/US-MN/marine-on-st.-croix-10660" => array("type"=>"city", "pop"=>630, "name_ru"=>"Марин он Сент Круа", "name_en"=>"Marine on St. Croix", "lang"=>"en", "lat"=>45.198038, "lon"=>-92.773156),
"/US/US-MN/marshall-10661" => array("type"=>"city", "pop"=>12647, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>44.450690, "lon"=>-95.789516),
"/US/US-MN/mayer-10662" => array("type"=>"city", "pop"=>1560, "name_ru"=>"Майер", "name_en"=>"Mayer", "lang"=>"en", "lat"=>44.884167, "lon"=>-93.888347),
"/US/US-MN/maynard-10663" => array("type"=>"city", "pop"=>354, "name_ru"=>"Мейнард", "name_en"=>"Maynard", "lang"=>"en", "lat"=>44.904935, "lon"=>-95.467150),
"/US/US-MN/mazeppa-10664" => array("type"=>"city", "pop"=>771, "name_ru"=>"Мазеппа", "name_en"=>"Mazeppa", "lang"=>"en", "lat"=>44.273560, "lon"=>-92.540802),
"/US/US-MN/mcgrath-10638" => array("type"=>"city", "pop"=>63, "name_ru"=>"Мак-Грат", "name_en"=>"McGrath", "lang"=>"en", "lat"=>46.242900, "lon"=>-93.274379),
"/US/US-MN/mcgregor-10639" => array("type"=>"city", "pop"=>378, "name_ru"=>"Мак-Грегор", "name_en"=>"McGregor", "lang"=>"en", "lat"=>46.608273, "lon"=>-93.305955),
"/US/US-MN/mcintosh-10640" => array("type"=>"city", "pop"=>604, "name_ru"=>"Макинтош", "name_en"=>"McIntosh", "lang"=>"en", "lat"=>47.637473, "lon"=>-95.885209),
"/US/US-MN/mckinley-10641" => array("type"=>"city", "pop"=>83, "name_ru"=>"Мак-Кинли", "name_en"=>"McKinley", "lang"=>"en", "lat"=>47.514530, "lon"=>-92.407499),
"/US/US-MN/meadowlands-10665" => array("type"=>"city", "pop"=>106, "name_ru"=>"Медоулендс", "name_en"=>"Meadowlands", "lang"=>"en", "lat"=>47.072783, "lon"=>-92.731392),
"/US/US-MN/medford-10666" => array("type"=>"city", "pop"=>1222, "name_ru"=>"Медфорд", "name_en"=>"Medford", "lang"=>"en", "lat"=>44.171745, "lon"=>-93.244467),
"/US/US-MN/medicine-lake-10667" => array("type"=>"city", "pop"=>340, "name_ru"=>"Медисин Лейк", "name_en"=>"Medicine Lake", "lang"=>"en", "lat"=>45.002500, "lon"=>-93.422943),
"/US/US-MN/medina-10668" => array("type"=>"city", "pop"=>5076, "name_ru"=>"Медина", "name_en"=>"Medina", "lang"=>"en", "lat"=>45.031167, "lon"=>-93.595328),
"/US/US-MN/meire-grove-10669" => array("type"=>"city", "pop"=>144, "name_ru"=>"Майр Гров", "name_en"=>"Meire Grove", "lang"=>"en", "lat"=>45.627081, "lon"=>-94.868899),
"/US/US-MN/melrose-10670" => array("type"=>"city", "pop"=>3145, "name_ru"=>"Мелроуз", "name_en"=>"Melrose", "lang"=>"en", "lat"=>45.675357, "lon"=>-94.812051),
"/US/US-MN/menahga-10671" => array("type"=>"city", "pop"=>1196, "name_ru"=>"Менага", "name_en"=>"Menahga", "lang"=>"en", "lat"=>46.748797, "lon"=>-95.102078),
"/US/US-MN/mendota-10672" => array("type"=>"city", "pop"=>202, "name_ru"=>"Мендота", "name_en"=>"Mendota", "lang"=>"en", "lat"=>44.886900, "lon"=>-93.159303),
"/US/US-MN/mendota-heights-10673" => array("type"=>"city", "pop"=>11594, "name_ru"=>"Мендота Хайтс", "name_en"=>"Mendota Heights", "lang"=>"en", "lat"=>44.882150, "lon"=>-93.140299),
"/US/US-MN/mentor-10674" => array("type"=>"city", "pop"=>139, "name_ru"=>"Ментор", "name_en"=>"Mentor", "lang"=>"en", "lat"=>47.696629, "lon"=>-96.144673),
"/US/US-MN/middle-river-10675" => array("type"=>"city", "pop"=>298, "name_ru"=>"Миддл Ривер", "name_en"=>"Middle River", "lang"=>"en", "lat"=>48.435637, "lon"=>-96.161914),
"/US/US-MN/miesville-10676" => array("type"=>"city", "pop"=>137, "name_ru"=>"Мисвилл", "name_en"=>"Miesville", "lang"=>"en", "lat"=>44.600227, "lon"=>-92.810013),
"/US/US-MN/milaca-10677" => array("type"=>"city", "pop"=>3003, "name_ru"=>"Милака", "name_en"=>"Milaca", "lang"=>"en", "lat"=>45.757874, "lon"=>-93.652903),
"/US/US-MN/milan-10678" => array("type"=>"city", "pop"=>310, "name_ru"=>"Милан", "name_en"=>"Milan", "lang"=>"en", "lat"=>45.114065, "lon"=>-95.913533),
"/US/US-MN/millerville-10679" => array("type"=>"city", "pop"=>111, "name_ru"=>"Миллервилл", "name_en"=>"Millerville", "lang"=>"en", "lat"=>46.071086, "lon"=>-95.552586),
"/US/US-MN/millville-10680" => array("type"=>"city", "pop"=>170, "name_ru"=>"Миллвилл", "name_en"=>"Millville", "lang"=>"en", "lat"=>44.245072, "lon"=>-92.296175),
"/US/US-MN/milroy-10681" => array("type"=>"city", "pop"=>246, "name_ru"=>"Милрой", "name_en"=>"Milroy", "lang"=>"en", "lat"=>44.417381, "lon"=>-95.554031),
"/US/US-MN/miltona-10682" => array("type"=>"city", "pop"=>370, "name_ru"=>"Милтона", "name_en"=>"Miltona", "lang"=>"en", "lat"=>46.044840, "lon"=>-95.291972),
"/US/US-MN/minneapolis-10683" => array("type"=>"city", "pop"=>382605, "name_ru"=>"Миннеаполис", "name_en"=>"Minneapolis", "lang"=>"en", "lat"=>44.961850, "lon"=>-93.266849),
"/US/US-MN/minneiska-10684" => array("type"=>"city", "pop"=>108, "name_ru"=>"Миннайска", "name_en"=>"Minneiska", "lang"=>"en", "lat"=>44.194798, "lon"=>-91.871545),
"/US/US-MN/minneota-10685" => array("type"=>"city", "pop"=>1366, "name_ru"=>"Миннеота", "name_en"=>"Minneota", "lang"=>"en", "lat"=>44.563661, "lon"=>-95.982005),
"/US/US-MN/minnesota-city-10686" => array("type"=>"city", "pop"=>217, "name_ru"=>"Миннесота Сити", "name_en"=>"Minnesota City", "lang"=>"en", "lat"=>44.092008, "lon"=>-91.750079),
"/US/US-MN/minnesota-lake-10687" => array("type"=>"city", "pop"=>638, "name_ru"=>"Миннесота Лейк", "name_en"=>"Minnesota Lake", "lang"=>"en", "lat"=>43.840461, "lon"=>-93.828743),
"/US/US-MN/minnetonka-10688" => array("type"=>"city", "pop"=>50081, "name_ru"=>"Миннетонка", "name_en"=>"Minnetonka", "lang"=>"en", "lat"=>44.932054, "lon"=>-93.461364),
"/US/US-MN/minnetonka-beach-10689" => array("type"=>"city", "pop"=>593, "name_ru"=>"Миннетонка Бич", "name_en"=>"Minnetonka Beach", "lang"=>"en", "lat"=>44.939851, "lon"=>-93.588295),
"/US/US-MN/minnetrista-10690" => array("type"=>"city", "pop"=>5960, "name_ru"=>"Миннетриста", "name_en"=>"Minnetrista", "lang"=>"en", "lat"=>44.942550, "lon"=>-93.703212),
"/US/US-MN/mizpah-10691" => array("type"=>"city", "pop"=>71, "name_ru"=>"Мизпа", "name_en"=>"Mizpah", "lang"=>"en", "lat"=>47.922547, "lon"=>-94.215437),
"/US/US-MN/montevideo-10692" => array("type"=>"city", "pop"=>5233, "name_ru"=>"Монтевидео", "name_en"=>"Montevideo", "lang"=>"en", "lat"=>44.951132, "lon"=>-95.718546),
"/US/US-MN/montgomery-10693" => array("type"=>"city", "pop"=>3266, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>44.439189, "lon"=>-93.580531),
"/US/US-MN/monticello-10694" => array("type"=>"city", "pop"=>11780, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>45.304738, "lon"=>-93.802205),
"/US/US-MN/montrose-10695" => array("type"=>"city", "pop"=>2789, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>45.067224, "lon"=>-93.912294),
"/US/US-MN/moorhead-10696" => array("type"=>"city", "pop"=>36012, "name_ru"=>"Мурхэд", "name_en"=>"Moorhead", "lang"=>"en", "lat"=>46.864584, "lon"=>-96.755710),
"/US/US-MN/moose-lake-10697" => array("type"=>"city", "pop"=>2386, "name_ru"=>"Муз Лейк", "name_en"=>"Moose Lake", "lang"=>"en", "lat"=>46.446040, "lon"=>-92.768440),
"/US/US-MN/mora-10698" => array("type"=>"city", "pop"=>3416, "name_ru"=>"Мора", "name_en"=>"Mora", "lang"=>"en", "lat"=>45.877392, "lon"=>-93.293263),
"/US/US-MN/morgan-10699" => array("type"=>"city", "pop"=>811, "name_ru"=>"Морган", "name_en"=>"Morgan", "lang"=>"en", "lat"=>44.415742, "lon"=>-94.924631),
"/US/US-MN/morris-10700" => array("type"=>"city", "pop"=>4996, "name_ru"=>"Моррис", "name_en"=>"Morris", "lang"=>"en", "lat"=>45.585887, "lon"=>-95.905292),
"/US/US-MN/morristown-10701" => array("type"=>"city", "pop"=>1039, "name_ru"=>"Морристаун", "name_en"=>"Morristown", "lang"=>"en", "lat"=>44.224666, "lon"=>-93.444829),
"/US/US-MN/morton-10702" => array("type"=>"city", "pop"=>402, "name_ru"=>"Мортон", "name_en"=>"Morton", "lang"=>"en", "lat"=>44.552977, "lon"=>-94.984855),
"/US/US-MN/motley-10703" => array("type"=>"city", "pop"=>639, "name_ru"=>"Мотли", "name_en"=>"Motley", "lang"=>"en", "lat"=>46.335661, "lon"=>-94.643247),
"/US/US-MN/mound-10704" => array("type"=>"city", "pop"=>9546, "name_ru"=>"Маунд", "name_en"=>"Mound", "lang"=>"en", "lat"=>44.933650, "lon"=>-93.660153),
"/US/US-MN/mounds-view-10705" => array("type"=>"city", "pop"=>12005, "name_ru"=>"Маундс Вью", "name_en"=>"Mounds View", "lang"=>"en", "lat"=>45.106033, "lon"=>-93.206113),
"/US/US-MN/mountain-iron-10706" => array("type"=>"city", "pop"=>2921, "name_ru"=>"Маунтин Айрон", "name_en"=>"Mountain Iron", "lang"=>"en", "lat"=>47.526414, "lon"=>-92.626850),
"/US/US-MN/mountain-lake-10707" => array("type"=>"city", "pop"=>1928, "name_ru"=>"Маунтин Лейк", "name_en"=>"Mountain Lake", "lang"=>"en", "lat"=>43.940329, "lon"=>-94.926195),
"/US/US-MN/murdock-10708" => array("type"=>"city", "pop"=>272, "name_ru"=>"Мердок", "name_en"=>"Murdock", "lang"=>"en", "lat"=>45.223655, "lon"=>-95.394118),
"/US/US-MN/myrtle-10709" => array("type"=>"city", "pop"=>59, "name_ru"=>"Миртл", "name_en"=>"Myrtle", "lang"=>"en", "lat"=>43.563178, "lon"=>-93.162878),
"/US/US-MN/nashua-10710" => array("type"=>"city", "pop"=>60, "name_ru"=>"Нашуа", "name_en"=>"Nashua", "lang"=>"en", "lat"=>46.036026, "lon"=>-96.304678),
"/US/US-MN/nashwauk-10711" => array("type"=>"city", "pop"=>915, "name_ru"=>"Нашуок", "name_en"=>"Nashwauk", "lang"=>"en", "lat"=>47.384182, "lon"=>-93.167741),
"/US/US-MN/nassau-10712" => array("type"=>"city", "pop"=>74, "name_ru"=>"Нассо", "name_en"=>"Nassau", "lang"=>"en", "lat"=>45.067967, "lon"=>-96.441124),
"/US/US-MN/naytahwaush-10713" => array("type"=>"cdp", "pop"=>378, "name_ru"=>"Нейтавауш", "name_en"=>"Naytahwaush", "lang"=>"en", "lat"=>47.285616, "lon"=>-95.630845),
"/US/US-MN/nelson-10714" => array("type"=>"city", "pop"=>164, "name_ru"=>"Нельсон", "name_en"=>"Nelson", "lang"=>"en", "lat"=>45.885004, "lon"=>-95.265705),
"/US/US-MN/nerstrand-10715" => array("type"=>"city", "pop"=>226, "name_ru"=>"Нерстренд", "name_en"=>"Nerstrand", "lang"=>"en", "lat"=>44.340754, "lon"=>-93.062467),
"/US/US-MN/nevis-10716" => array("type"=>"city", "pop"=>389, "name_ru"=>"Невис", "name_en"=>"Nevis", "lang"=>"en", "lat"=>46.964553, "lon"=>-94.841702),
"/US/US-MN/new-auburn-10717" => array("type"=>"city", "pop"=>502, "name_ru"=>"Нью Обурн", "name_en"=>"New Auburn", "lang"=>"en", "lat"=>44.672843, "lon"=>-94.228911),
"/US/US-MN/new-brighton-10718" => array("type"=>"city", "pop"=>21563, "name_ru"=>"Нью Брайтон", "name_en"=>"New Brighton", "lang"=>"en", "lat"=>45.065550, "lon"=>-93.204326),
"/US/US-MN/new-germany-10720" => array("type"=>"city", "pop"=>371, "name_ru"=>"Нью Джермани", "name_en"=>"New Germany", "lang"=>"en", "lat"=>44.883700, "lon"=>-93.970899),
"/US/US-MN/new-hope-10721" => array("type"=>"city", "pop"=>20428, "name_ru"=>"Нью Хоуп", "name_en"=>"New Hope", "lang"=>"en", "lat"=>45.036967, "lon"=>-93.386953),
"/US/US-MN/new-london-10722" => array("type"=>"city", "pop"=>1319, "name_ru"=>"Нью Лондон", "name_en"=>"New London", "lang"=>"en", "lat"=>45.299294, "lon"=>-94.945627),
"/US/US-MN/new-market-10723" => array("type"=>"city", "pop"=>3668, "name_ru"=>"Нью Маркет", "name_en"=>"New Market", "lang"=>"en", "lat"=>44.572103, "lon"=>-93.349657),
"/US/US-MN/new-munich-10724" => array("type"=>"city", "pop"=>344, "name_ru"=>"Нью Муник", "name_en"=>"New Munich", "lang"=>"en", "lat"=>45.630588, "lon"=>-94.752814),
"/US/US-MN/new-prague-10726" => array("type"=>"city", "pop"=>6712, "name_ru"=>"Нью Прага", "name_en"=>"New Prague", "lang"=>"en", "lat"=>44.546550, "lon"=>-93.573999),
"/US/US-MN/new-richland-10727" => array("type"=>"city", "pop"=>1156, "name_ru"=>"Нью Ричленд", "name_en"=>"New Richland", "lang"=>"en", "lat"=>43.893412, "lon"=>-93.492977),
"/US/US-MN/new-trier-10728" => array("type"=>"city", "pop"=>114, "name_ru"=>"Нью Триер", "name_en"=>"New Trier", "lang"=>"en", "lat"=>44.603914, "lon"=>-92.933168),
"/US/US-MN/new-ulm-10729" => array("type"=>"city", "pop"=>13048, "name_ru"=>"Нью Ульм", "name_en"=>"New Ulm", "lang"=>"en", "lat"=>44.313096, "lon"=>-94.463721),
"/US/US-MN/new-york-mills-10730" => array("type"=>"city", "pop"=>1203, "name_ru"=>"Нью Йорк Миллс", "name_en"=>"New York Mills", "lang"=>"en", "lat"=>46.519714, "lon"=>-95.374175),
"/US/US-MN/newfolden-10719" => array("type"=>"city", "pop"=>340, "name_ru"=>"Ньюфолден", "name_en"=>"Newfolden", "lang"=>"en", "lat"=>48.355182, "lon"=>-96.329725),
"/US/US-MN/newport-10725" => array("type"=>"city", "pop"=>3484, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>44.874500, "lon"=>-92.998556),
"/US/US-MN/nicollet-10731" => array("type"=>"city", "pop"=>1074, "name_ru"=>"Николлет", "name_en"=>"Nicollet", "lang"=>"en", "lat"=>44.275567, "lon"=>-94.188126),
"/US/US-MN/nielsville-10732" => array("type"=>"city", "pop"=>83, "name_ru"=>"Нилсвилл", "name_en"=>"Nielsville", "lang"=>"en", "lat"=>47.528773, "lon"=>-96.816266),
"/US/US-MN/nimrod-10733" => array("type"=>"city", "pop"=>72, "name_ru"=>"Нимрод", "name_en"=>"Nimrod", "lang"=>"en", "lat"=>46.637780, "lon"=>-94.880994),
"/US/US-MN/nisswa-10734" => array("type"=>"city", "pop"=>2110, "name_ru"=>"Ниссва", "name_en"=>"Nisswa", "lang"=>"en", "lat"=>46.500027, "lon"=>-94.297279),
"/US/US-MN/norcross-10735" => array("type"=>"city", "pop"=>56, "name_ru"=>"Норкросс", "name_en"=>"Norcross", "lang"=>"en", "lat"=>45.868575, "lon"=>-96.195878),
"/US/US-MN/north-branch-10736" => array("type"=>"city", "pop"=>10417, "name_ru"=>"Норт Брэнч", "name_en"=>"North Branch", "lang"=>"en", "lat"=>45.514287, "lon"=>-92.978367),
"/US/US-MN/north-mankato-10738" => array("type"=>"city", "pop"=>12530, "name_ru"=>"Норт Манкато", "name_en"=>"North Mankato", "lang"=>"en", "lat"=>44.176204, "lon"=>-94.030850),
"/US/US-MN/north-oaks-10739" => array("type"=>"city", "pop"=>4689, "name_ru"=>"Норт Оукс", "name_en"=>"North Oaks", "lang"=>"en", "lat"=>45.106800, "lon"=>-93.087204),
"/US/US-MN/north-redwood-10741" => array("type"=>"city", "pop"=>203, "name_ru"=>"Норт Редвуд", "name_en"=>"North Redwood", "lang"=>"en", "lat"=>44.563553, "lon"=>-95.093522),
"/US/US-MN/north-st.-paul-10743" => array("type"=>"city", "pop"=>11247, "name_ru"=>"Норт Сент Пол", "name_en"=>"North St. Paul", "lang"=>"en", "lat"=>45.014133, "lon"=>-92.998321),
"/US/US-MN/northfield-10737" => array("type"=>"city", "pop"=>19633, "name_ru"=>"Нортфилд", "name_en"=>"Northfield", "lang"=>"en", "lat"=>44.454379, "lon"=>-93.165305),
"/US/US-MN/northome-10740" => array("type"=>"city", "pop"=>207, "name_ru"=>"Нортхоум", "name_en"=>"Northome", "lang"=>"en", "lat"=>47.874635, "lon"=>-94.265764),
"/US/US-MN/northrop-10742" => array("type"=>"city", "pop"=>237, "name_ru"=>"Нортроп", "name_en"=>"Northrop", "lang"=>"en", "lat"=>43.735640, "lon"=>-94.435601),
"/US/US-MN/norwood-10744" => array("type"=>"city", "pop"=>1351, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>44.768500, "lon"=>-93.926044),
"/US/US-MN/oak-park-heights-10746" => array("type"=>"city", "pop"=>4680, "name_ru"=>"Оук Парк Хайтс", "name_en"=>"Oak Park Heights", "lang"=>"en", "lat"=>45.030663, "lon"=>-92.805499),
"/US/US-MN/oakdale-10745" => array("type"=>"city", "pop"=>27034, "name_ru"=>"Оукдейл", "name_en"=>"Oakdale", "lang"=>"en", "lat"=>44.988700, "lon"=>-92.966300),
"/US/US-MN/oakport-10747" => array("type"=>"cdp", "pop"=>1638, "name_ru"=>"Оукпорт", "name_en"=>"Oakport", "lang"=>"en", "lat"=>46.931880, "lon"=>-96.778571),
"/US/US-MN/odessa-10748" => array("type"=>"city", "pop"=>103, "name_ru"=>"Одесса", "name_en"=>"Odessa", "lang"=>"en", "lat"=>45.261385, "lon"=>-96.331186),
"/US/US-MN/odin-10749" => array("type"=>"city", "pop"=>115, "name_ru"=>"Один", "name_en"=>"Odin", "lang"=>"en", "lat"=>43.867575, "lon"=>-94.741716),
"/US/US-MN/ogema-10750" => array("type"=>"city", "pop"=>142, "name_ru"=>"Огема", "name_en"=>"Ogema", "lang"=>"en", "lat"=>47.101608, "lon"=>-95.922461),
"/US/US-MN/ogilvie-10751" => array("type"=>"city", "pop"=>467, "name_ru"=>"Огилви", "name_en"=>"Ogilvie", "lang"=>"en", "lat"=>45.830684, "lon"=>-93.419918),
"/US/US-MN/okabena-10752" => array("type"=>"city", "pop"=>169, "name_ru"=>"Окабена", "name_en"=>"Okabena", "lang"=>"en", "lat"=>43.738576, "lon"=>-95.315844),
"/US/US-MN/oklee-10753" => array("type"=>"city", "pop"=>358, "name_ru"=>"Окли", "name_en"=>"Oklee", "lang"=>"en", "lat"=>47.838834, "lon"=>-95.850097),
"/US/US-MN/olivia-10754" => array("type"=>"city", "pop"=>2362, "name_ru"=>"Оливия", "name_en"=>"Olivia", "lang"=>"en", "lat"=>44.775810, "lon"=>-94.997087),
"/US/US-MN/onamia-10755" => array("type"=>"city", "pop"=>907, "name_ru"=>"Онамия", "name_en"=>"Onamia", "lang"=>"en", "lat"=>46.069261, "lon"=>-93.665158),
"/US/US-MN/ormsby-10756" => array("type"=>"city", "pop"=>136, "name_ru"=>"Ормсби", "name_en"=>"Ormsby", "lang"=>"en", "lat"=>43.850958, "lon"=>-94.698579),
"/US/US-MN/orono-10757" => array("type"=>"city", "pop"=>7732, "name_ru"=>"Ороно", "name_en"=>"Orono", "lang"=>"en", "lat"=>44.965200, "lon"=>-93.586730),
"/US/US-MN/oronoco-10758" => array("type"=>"city", "pop"=>949, "name_ru"=>"Ороноко", "name_en"=>"Oronoco", "lang"=>"en", "lat"=>44.163520, "lon"=>-92.540912),
"/US/US-MN/orr-10759" => array("type"=>"city", "pop"=>242, "name_ru"=>"Орр", "name_en"=>"Orr", "lang"=>"en", "lat"=>48.063953, "lon"=>-92.820369),
"/US/US-MN/ortonville-10760" => array("type"=>"city", "pop"=>1974, "name_ru"=>"Ортонвилл", "name_en"=>"Ortonville", "lang"=>"en", "lat"=>45.303639, "lon"=>-96.440883),
"/US/US-MN/osakis-10761" => array("type"=>"city", "pop"=>1572, "name_ru"=>"Осакис", "name_en"=>"Osakis", "lang"=>"en", "lat"=>45.865362, "lon"=>-95.152754),
"/US/US-MN/oslo-10762" => array("type"=>"city", "pop"=>319, "name_ru"=>"Осло", "name_en"=>"Oslo", "lang"=>"en", "lat"=>48.194926, "lon"=>-97.127863),
"/US/US-MN/osseo-10763" => array("type"=>"city", "pop"=>2551, "name_ru"=>"Оссео", "name_en"=>"Osseo", "lang"=>"en", "lat"=>45.117900, "lon"=>-93.399149),
"/US/US-MN/ostrander-10764" => array("type"=>"city", "pop"=>199, "name_ru"=>"Острэндер", "name_en"=>"Ostrander", "lang"=>"en", "lat"=>43.614031, "lon"=>-92.425629),
"/US/US-MN/ottertail-10765" => array("type"=>"city", "pop"=>533, "name_ru"=>"Оттертейл", "name_en"=>"Ottertail", "lang"=>"en", "lat"=>46.428472, "lon"=>-95.558689),
"/US/US-MN/owatonna-10766" => array("type"=>"city", "pop"=>24855, "name_ru"=>"Оуотонна", "name_en"=>"Owatonna", "lang"=>"en", "lat"=>44.085186, "lon"=>-93.224317),
"/US/US-MN/palisade-10767" => array("type"=>"city", "pop"=>108, "name_ru"=>"Палисейд", "name_en"=>"Palisade", "lang"=>"en", "lat"=>46.713132, "lon"=>-93.490725),
"/US/US-MN/park-rapids-10769" => array("type"=>"city", "pop"=>3648, "name_ru"=>"Парк Рэпидс", "name_en"=>"Park Rapids", "lang"=>"en", "lat"=>46.917477, "lon"=>-95.060139),
"/US/US-MN/parkers-prairie-10768" => array("type"=>"city", "pop"=>1008, "name_ru"=>"Паркерс Прейри", "name_en"=>"Parkers Prairie", "lang"=>"en", "lat"=>46.153798, "lon"=>-95.329307),
"/US/US-MN/paynesville-10770" => array("type"=>"city", "pop"=>2262, "name_ru"=>"Пейнсвилл", "name_en"=>"Paynesville", "lang"=>"en", "lat"=>45.378526, "lon"=>-94.719729),
"/US/US-MN/pease-10771" => array("type"=>"city", "pop"=>198, "name_ru"=>"Пис", "name_en"=>"Pease", "lang"=>"en", "lat"=>45.698108, "lon"=>-93.650601),
"/US/US-MN/pelican-rapids-10772" => array("type"=>"city", "pop"=>2311, "name_ru"=>"Пеликан Рэпидс", "name_en"=>"Pelican Rapids", "lang"=>"en", "lat"=>46.570815, "lon"=>-96.085626),
"/US/US-MN/pemberton-10773" => array("type"=>"city", "pop"=>257, "name_ru"=>"Пембертон", "name_en"=>"Pemberton", "lang"=>"en", "lat"=>44.008945, "lon"=>-93.783856),
"/US/US-MN/pennock-10774" => array("type"=>"city", "pop"=>491, "name_ru"=>"Пеннок", "name_en"=>"Pennock", "lang"=>"en", "lat"=>45.146318, "lon"=>-95.175336),
"/US/US-MN/pequot-lakes-10775" => array("type"=>"city", "pop"=>1925, "name_ru"=>"Пеквот Лейкс", "name_en"=>"Pequot Lakes", "lang"=>"en", "lat"=>46.601948, "lon"=>-94.320593),
"/US/US-MN/perham-10776" => array("type"=>"city", "pop"=>2730, "name_ru"=>"Перхэм", "name_en"=>"Perham", "lang"=>"en", "lat"=>46.599606, "lon"=>-95.573368),
"/US/US-MN/perley-10777" => array("type"=>"city", "pop"=>107, "name_ru"=>"Перли", "name_en"=>"Perley", "lang"=>"en", "lat"=>47.176594, "lon"=>-96.800471),
"/US/US-MN/peterson-10778" => array("type"=>"city", "pop"=>254, "name_ru"=>"Петерсон", "name_en"=>"Peterson", "lang"=>"en", "lat"=>43.785990, "lon"=>-91.833174),
"/US/US-MN/pierz-10779" => array("type"=>"city", "pop"=>1371, "name_ru"=>"Пирз", "name_en"=>"Pierz", "lang"=>"en", "lat"=>45.978505, "lon"=>-94.100291),
"/US/US-MN/pillager-10780" => array("type"=>"city", "pop"=>492, "name_ru"=>"Пиллагер", "name_en"=>"Pillager", "lang"=>"en", "lat"=>46.326909, "lon"=>-94.476446),
"/US/US-MN/pine-city-10781" => array("type"=>"city", "pop"=>3285, "name_ru"=>"Пайн Сити", "name_en"=>"Pine City", "lang"=>"en", "lat"=>45.830860, "lon"=>-92.969947),
"/US/US-MN/pine-island-10782" => array("type"=>"city", "pop"=>3326, "name_ru"=>"Пайн Айленд", "name_en"=>"Pine Island", "lang"=>"en", "lat"=>44.204880, "lon"=>-92.649583),
"/US/US-MN/pine-river-10783" => array("type"=>"city", "pop"=>877, "name_ru"=>"Пайн Ривер", "name_en"=>"Pine River", "lang"=>"en", "lat"=>46.722554, "lon"=>-94.398310),
"/US/US-MN/pine-springs-10784" => array("type"=>"city", "pop"=>392, "name_ru"=>"Пайн Спрингс", "name_en"=>"Pine Springs", "lang"=>"en", "lat"=>45.036483, "lon"=>-92.952116),
"/US/US-MN/pipestone-10785" => array("type"=>"city", "pop"=>4131, "name_ru"=>"Пайпстоун", "name_en"=>"Pipestone", "lang"=>"en", "lat"=>44.000020, "lon"=>-96.314082),
"/US/US-MN/plainview-10786" => array("type"=>"city", "pop"=>3225, "name_ru"=>"Плейнвью", "name_en"=>"Plainview", "lang"=>"en", "lat"=>44.165551, "lon"=>-92.168932),
"/US/US-MN/plato-10787" => array("type"=>"city", "pop"=>317, "name_ru"=>"Плато", "name_en"=>"Plato", "lang"=>"en", "lat"=>44.772430, "lon"=>-94.038958),
"/US/US-MN/pleasant-lake-10788" => array("type"=>"city", "pop"=>79, "name_ru"=>"Плезент Лейк", "name_en"=>"Pleasant Lake", "lang"=>"en", "lat"=>45.499042, "lon"=>-94.284961),
"/US/US-MN/plummer-10789" => array("type"=>"city", "pop"=>255, "name_ru"=>"Пламмер", "name_en"=>"Plummer", "lang"=>"en", "lat"=>47.915656, "lon"=>-96.039448),
"/US/US-MN/plymouth-10790" => array("type"=>"city", "pop"=>71486, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>45.024281, "lon"=>-93.460215),
"/US/US-MN/ponemah-10791" => array("type"=>"cdp", "pop"=>704, "name_ru"=>"Поунма", "name_en"=>"Ponemah", "lang"=>"en", "lat"=>48.042477, "lon"=>-94.907827),
"/US/US-MN/porter-10792" => array("type"=>"city", "pop"=>163, "name_ru"=>"Портер", "name_en"=>"Porter", "lang"=>"en", "lat"=>44.640748, "lon"=>-96.169607),
"/US/US-MN/preston-10793" => array("type"=>"city", "pop"=>1324, "name_ru"=>"Престон", "name_en"=>"Preston", "lang"=>"en", "lat"=>43.671565, "lon"=>-92.082529),
"/US/US-MN/princeton-10794" => array("type"=>"city", "pop"=>4763, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>45.570560, "lon"=>-93.590842),
"/US/US-MN/prinsburg-10795" => array("type"=>"city", "pop"=>449, "name_ru"=>"Принсбург", "name_en"=>"Prinsburg", "lang"=>"en", "lat"=>44.935586, "lon"=>-95.185833),
"/US/US-MN/prior-lake-10796" => array("type"=>"city", "pop"=>23836, "name_ru"=>"Прайор Лейк", "name_en"=>"Prior Lake", "lang"=>"en", "lat"=>44.728550, "lon"=>-93.434020),
"/US/US-MN/proctor-10797" => array("type"=>"city", "pop"=>2839, "name_ru"=>"Проктор", "name_en"=>"Proctor", "lang"=>"en", "lat"=>46.743100, "lon"=>-92.227386),
"/US/US-MN/quamba-10798" => array("type"=>"city", "pop"=>107, "name_ru"=>"Квамба", "name_en"=>"Quamba", "lang"=>"en", "lat"=>45.915469, "lon"=>-93.174714),
"/US/US-MN/racine-10799" => array("type"=>"city", "pop"=>510, "name_ru"=>"Расин", "name_en"=>"Racine", "lang"=>"en", "lat"=>43.776489, "lon"=>-92.481962),
"/US/US-MN/ramsey-10800" => array("type"=>"city", "pop"=>23734, "name_ru"=>"Рэмси", "name_en"=>"Ramsey", "lang"=>"en", "lat"=>45.263400, "lon"=>-93.447670),
"/US/US-MN/randall-10801" => array("type"=>"city", "pop"=>633, "name_ru"=>"Рэндолл", "name_en"=>"Randall", "lang"=>"en", "lat"=>46.089363, "lon"=>-94.499700),
"/US/US-MN/randolph-10802" => array("type"=>"city", "pop"=>351, "name_ru"=>"Рэндолф", "name_en"=>"Randolph", "lang"=>"en", "lat"=>44.526350, "lon"=>-93.018562),
"/US/US-MN/ranier-10803" => array("type"=>"city", "pop"=>168, "name_ru"=>"Раньер", "name_en"=>"Ranier", "lang"=>"en", "lat"=>48.613222, "lon"=>-93.346587),
"/US/US-MN/raymond-10804" => array("type"=>"city", "pop"=>763, "name_ru"=>"Реймонд", "name_en"=>"Raymond", "lang"=>"en", "lat"=>45.016016, "lon"=>-95.237185),
"/US/US-MN/red-lake-10806" => array("type"=>"cdp", "pop"=>4069, "name_ru"=>"Ред Лейк", "name_en"=>"Red Lake", "lang"=>"en", "lat"=>47.859683, "lon"=>-95.001623),
"/US/US-MN/red-lake-falls-10807" => array("type"=>"city", "pop"=>1503, "name_ru"=>"Ред Лейк Фоллс", "name_en"=>"Red Lake Falls", "lang"=>"en", "lat"=>47.885186, "lon"=>-96.273906),
"/US/US-MN/red-wing-10808" => array("type"=>"city", "pop"=>15687, "name_ru"=>"Ред Уинг", "name_en"=>"Red Wing", "lang"=>"en", "lat"=>44.581929, "lon"=>-92.602057),
"/US/US-MN/redby-10805" => array("type"=>"cdp", "pop"=>787, "name_ru"=>"Редби", "name_en"=>"Redby", "lang"=>"en", "lat"=>47.844192, "lon"=>-94.930397),
"/US/US-MN/redwood-falls-10809" => array("type"=>"city", "pop"=>5077, "name_ru"=>"Редвуд Фоллс", "name_en"=>"Redwood Falls", "lang"=>"en", "lat"=>44.541646, "lon"=>-95.113700),
"/US/US-MN/regal-10810" => array("type"=>"city", "pop"=>38, "name_ru"=>"Регал", "name_en"=>"Regal", "lang"=>"en", "lat"=>45.405000, "lon"=>-94.840155),
"/US/US-MN/remer-10811" => array("type"=>"city", "pop"=>357, "name_ru"=>"Ремер", "name_en"=>"Remer", "lang"=>"en", "lat"=>47.058046, "lon"=>-93.913671),
"/US/US-MN/renville-10812" => array("type"=>"city", "pop"=>1207, "name_ru"=>"Ренвилл", "name_en"=>"Renville", "lang"=>"en", "lat"=>44.794221, "lon"=>-95.211105),
"/US/US-MN/revere-10813" => array("type"=>"city", "pop"=>90, "name_ru"=>"Ревир", "name_en"=>"Revere", "lang"=>"en", "lat"=>44.218927, "lon"=>-95.363350),
"/US/US-MN/rice-10814" => array("type"=>"city", "pop"=>1454, "name_ru"=>"Райс", "name_en"=>"Rice", "lang"=>"en", "lat"=>45.748625, "lon"=>-94.228686),
"/US/US-MN/richfield-10815" => array("type"=>"city", "pop"=>32728, "name_ru"=>"Ричфилд", "name_en"=>"Richfield", "lang"=>"en", "lat"=>44.877200, "lon"=>-93.282124),
"/US/US-MN/richmond-10816" => array("type"=>"city", "pop"=>1268, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>45.452642, "lon"=>-94.517730),
"/US/US-MN/richville-10817" => array("type"=>"city", "pop"=>123, "name_ru"=>"Ричвилл", "name_en"=>"Richville", "lang"=>"en", "lat"=>46.506659, "lon"=>-95.626068),
"/US/US-MN/riverton-10818" => array("type"=>"city", "pop"=>106, "name_ru"=>"Ривертон", "name_en"=>"Riverton", "lang"=>"en", "lat"=>46.460951, "lon"=>-94.048726),
"/US/US-MN/robbinsdale-10819" => array("type"=>"city", "pop"=>13518, "name_ru"=>"Роббинсдейл", "name_en"=>"Robbinsdale", "lang"=>"en", "lat"=>45.026700, "lon"=>-93.333181),
"/US/US-MN/rochester-10820" => array("type"=>"city", "pop"=>100413, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>44.013650, "lon"=>-92.475699),
"/US/US-MN/rock-creek-10821" => array("type"=>"city", "pop"=>1502, "name_ru"=>"Рок Крик", "name_en"=>"Rock Creek", "lang"=>"en", "lat"=>45.759306, "lon"=>-92.907751),
"/US/US-MN/rockford-10822" => array("type"=>"city", "pop"=>4051, "name_ru"=>"Рокфорд", "name_en"=>"Rockford", "lang"=>"en", "lat"=>45.089100, "lon"=>-93.735560),
"/US/US-MN/rockville-10823" => array("type"=>"city", "pop"=>2530, "name_ru"=>"Роквилл", "name_en"=>"Rockville", "lang"=>"en", "lat"=>45.472428, "lon"=>-94.339637),
"/US/US-MN/rogers-10824" => array("type"=>"city", "pop"=>7536, "name_ru"=>"Роджерс", "name_en"=>"Rogers", "lang"=>"en", "lat"=>45.193900, "lon"=>-93.551700),
"/US/US-MN/rollingstone-10825" => array("type"=>"city", "pop"=>641, "name_ru"=>"Роллингстоун", "name_en"=>"Rollingstone", "lang"=>"en", "lat"=>44.099896, "lon"=>-91.819110),
"/US/US-MN/ronneby-10826" => array("type"=>"city", "pop"=>32, "name_ru"=>"Роннеби", "name_en"=>"Ronneby", "lang"=>"en", "lat"=>45.682234, "lon"=>-93.865733),
"/US/US-MN/roosevelt-10827" => array("type"=>"city", "pop"=>161, "name_ru"=>"Рузвельт", "name_en"=>"Roosevelt", "lang"=>"en", "lat"=>48.806734, "lon"=>-95.100530),
"/US/US-MN/roscoe-10828" => array("type"=>"city", "pop"=>108, "name_ru"=>"Роско", "name_en"=>"Roscoe", "lang"=>"en", "lat"=>45.432352, "lon"=>-94.635243),
"/US/US-MN/rose-creek-10830" => array("type"=>"city", "pop"=>342, "name_ru"=>"Роуз Крик", "name_en"=>"Rose Creek", "lang"=>"en", "lat"=>43.604565, "lon"=>-92.829646),
"/US/US-MN/roseau-10829" => array("type"=>"city", "pop"=>2758, "name_ru"=>"Розо", "name_en"=>"Roseau", "lang"=>"en", "lat"=>48.845759, "lon"=>-95.761853),
"/US/US-MN/rosemount-10831" => array("type"=>"city", "pop"=>21197, "name_ru"=>"Роузмаунт", "name_en"=>"Rosemount", "lang"=>"en", "lat"=>44.747570, "lon"=>-93.070769),
"/US/US-MN/roseville-10832" => array("type"=>"city", "pop"=>32815, "name_ru"=>"Роузвилл", "name_en"=>"Roseville", "lang"=>"en", "lat"=>45.015200, "lon"=>-93.152547),
"/US/US-MN/rothsay-10833" => array("type"=>"city", "pop"=>478, "name_ru"=>"Ротсей", "name_en"=>"Rothsay", "lang"=>"en", "lat"=>46.474732, "lon"=>-96.286669),
"/US/US-MN/round-lake-10834" => array("type"=>"city", "pop"=>403, "name_ru"=>"Раунд Лейк", "name_en"=>"Round Lake", "lang"=>"en", "lat"=>43.538760, "lon"=>-95.469290),
"/US/US-MN/royalton-10835" => array("type"=>"city", "pop"=>976, "name_ru"=>"Ройалтон", "name_en"=>"Royalton", "lang"=>"en", "lat"=>45.831774, "lon"=>-94.289438),
"/US/US-MN/rush-city-10836" => array("type"=>"city", "pop"=>2981, "name_ru"=>"Раш Сити", "name_en"=>"Rush City", "lang"=>"en", "lat"=>45.681677, "lon"=>-92.965832),
"/US/US-MN/rushford-10837" => array("type"=>"city", "pop"=>1727, "name_ru"=>"Рашфорд", "name_en"=>"Rushford", "lang"=>"en", "lat"=>43.812340, "lon"=>-91.752408),
"/US/US-MN/rushford-village-10838" => array("type"=>"city", "pop"=>771, "name_ru"=>"Рашфорд Виллидж", "name_en"=>"Rushford Village", "lang"=>"en", "lat"=>43.804398, "lon"=>-91.787296),
"/US/US-MN/rushmore-10839" => array("type"=>"city", "pop"=>367, "name_ru"=>"Рашмор", "name_en"=>"Rushmore", "lang"=>"en", "lat"=>43.620149, "lon"=>-95.798826),
"/US/US-MN/russell-10840" => array("type"=>"city", "pop"=>335, "name_ru"=>"Расселл", "name_en"=>"Russell", "lang"=>"en", "lat"=>44.318931, "lon"=>-95.947961),
"/US/US-MN/ruthton-10841" => array("type"=>"city", "pop"=>259, "name_ru"=>"Руттон", "name_en"=>"Ruthton", "lang"=>"en", "lat"=>44.177824, "lon"=>-96.103912),
"/US/US-MN/rutledge-10842" => array("type"=>"city", "pop"=>195, "name_ru"=>"Рутледж", "name_en"=>"Rutledge", "lang"=>"en", "lat"=>46.260029, "lon"=>-92.869185),
"/US/US-MN/sabin-10843" => array("type"=>"city", "pop"=>441, "name_ru"=>"Сабин", "name_en"=>"Sabin", "lang"=>"en", "lat"=>46.780465, "lon"=>-96.652909),
"/US/US-MN/sacred-heart-10844" => array("type"=>"city", "pop"=>492, "name_ru"=>"Сейкред Харт", "name_en"=>"Sacred Heart", "lang"=>"en", "lat"=>44.783110, "lon"=>-95.350337),
"/US/US-MN/sanborn-10866" => array("type"=>"city", "pop"=>385, "name_ru"=>"Сэнборн", "name_en"=>"Sanborn", "lang"=>"en", "lat"=>44.210166, "lon"=>-95.128953),
"/US/US-MN/sandstone-10867" => array("type"=>"city", "pop"=>2792, "name_ru"=>"Сэндстоун", "name_en"=>"Sandstone", "lang"=>"en", "lat"=>46.125170, "lon"=>-92.864587),
"/US/US-MN/sargeant-10868" => array("type"=>"city", "pop"=>71, "name_ru"=>"Сраджент", "name_en"=>"Sargeant", "lang"=>"en", "lat"=>43.805587, "lon"=>-92.800905),
"/US/US-MN/sartell-10869" => array("type"=>"city", "pop"=>13915, "name_ru"=>"Сартелл", "name_en"=>"Sartell", "lang"=>"en", "lat"=>45.620501, "lon"=>-94.206267),
"/US/US-MN/sauk-centre-10870" => array("type"=>"city", "pop"=>3945, "name_ru"=>"Сок Сентер", "name_en"=>"Sauk Centre", "lang"=>"en", "lat"=>45.737775, "lon"=>-94.957256),
"/US/US-MN/sauk-rapids-10871" => array("type"=>"city", "pop"=>12013, "name_ru"=>"Сок Рэпидс", "name_en"=>"Sauk Rapids", "lang"=>"en", "lat"=>45.594401, "lon"=>-94.166069),
"/US/US-MN/savage-10872" => array("type"=>"city", "pop"=>27692, "name_ru"=>"Сэвидж", "name_en"=>"Savage", "lang"=>"en", "lat"=>44.755381, "lon"=>-93.357236),
"/US/US-MN/scanlon-10873" => array("type"=>"city", "pop"=>843, "name_ru"=>"Скэнлон", "name_en"=>"Scanlon", "lang"=>"en", "lat"=>46.707391, "lon"=>-92.428919),
"/US/US-MN/seaforth-10874" => array("type"=>"city", "pop"=>71, "name_ru"=>"Сифорт", "name_en"=>"Seaforth", "lang"=>"en", "lat"=>44.476922, "lon"=>-95.329252),
"/US/US-MN/sebeka-10875" => array("type"=>"city", "pop"=>650, "name_ru"=>"Себека", "name_en"=>"Sebeka", "lang"=>"en", "lat"=>46.629082, "lon"=>-95.088174),
"/US/US-MN/sedan-10876" => array("type"=>"city", "pop"=>60, "name_ru"=>"Седан", "name_en"=>"Sedan", "lang"=>"en", "lat"=>45.576991, "lon"=>-95.245501),
"/US/US-MN/shafer-10877" => array("type"=>"city", "pop"=>378, "name_ru"=>"Шафтер", "name_en"=>"Shafer", "lang"=>"en", "lat"=>45.387725, "lon"=>-92.745837),
"/US/US-MN/shakopee-10878" => array("type"=>"city", "pop"=>33455, "name_ru"=>"Шакопи", "name_en"=>"Shakopee", "lang"=>"en", "lat"=>44.770600, "lon"=>-93.475582),
"/US/US-MN/shelly-10879" => array("type"=>"city", "pop"=>239, "name_ru"=>"Шелли", "name_en"=>"Shelly", "lang"=>"en", "lat"=>47.457525, "lon"=>-96.820044),
"/US/US-MN/sherburn-10880" => array("type"=>"city", "pop"=>1019, "name_ru"=>"Шерберн", "name_en"=>"Sherburn", "lang"=>"en", "lat"=>43.653761, "lon"=>-94.726488),
"/US/US-MN/shevlin-10881" => array("type"=>"city", "pop"=>155, "name_ru"=>"Шевлин", "name_en"=>"Shevlin", "lang"=>"en", "lat"=>47.530714, "lon"=>-95.259245),
"/US/US-MN/shoreview-10882" => array("type"=>"city", "pop"=>25111, "name_ru"=>"Шорвью", "name_en"=>"Shoreview", "lang"=>"en", "lat"=>45.083900, "lon"=>-93.135949),
"/US/US-MN/shorewood-10883" => array("type"=>"city", "pop"=>7426, "name_ru"=>"Шорвуд", "name_en"=>"Shorewood", "lang"=>"en", "lat"=>44.919835, "lon"=>-93.565347),
"/US/US-MN/silver-bay-10884" => array("type"=>"city", "pop"=>1840, "name_ru"=>"Силвер Бэй", "name_en"=>"Silver Bay", "lang"=>"en", "lat"=>47.291888, "lon"=>-91.278600),
"/US/US-MN/silver-lake-10885" => array("type"=>"city", "pop"=>799, "name_ru"=>"Сильвер Лэйк", "name_en"=>"Silver Lake", "lang"=>"en", "lat"=>44.903819, "lon"=>-94.197517),
"/US/US-MN/skyline-10886" => array("type"=>"city", "pop"=>295, "name_ru"=>"Скайлайн", "name_en"=>"Skyline", "lang"=>"en", "lat"=>44.140196, "lon"=>-94.035228),
"/US/US-MN/slayton-10887" => array("type"=>"city", "pop"=>1835, "name_ru"=>"Слейтон", "name_en"=>"Slayton", "lang"=>"en", "lat"=>43.990539, "lon"=>-95.754868),
"/US/US-MN/sleepy-eye-10888" => array("type"=>"city", "pop"=>3423, "name_ru"=>"Слипи Ай", "name_en"=>"Sleepy Eye", "lang"=>"en", "lat"=>44.299925, "lon"=>-94.724831),
"/US/US-MN/sobieski-10889" => array("type"=>"city", "pop"=>181, "name_ru"=>"Собиски", "name_en"=>"Sobieski", "lang"=>"en", "lat"=>45.924316, "lon"=>-94.482118),
"/US/US-MN/solway-10890" => array("type"=>"city", "pop"=>73, "name_ru"=>"Солуэй", "name_en"=>"Solway", "lang"=>"en", "lat"=>47.520658, "lon"=>-95.129749),
"/US/US-MN/south-haven-10891" => array("type"=>"city", "pop"=>195, "name_ru"=>"Саут Хейвен", "name_en"=>"South Haven", "lang"=>"en", "lat"=>45.291468, "lon"=>-94.215855),
"/US/US-MN/south-st.-paul-10892" => array("type"=>"city", "pop"=>19552, "name_ru"=>"Саут Сент Пол", "name_en"=>"South St. Paul", "lang"=>"en", "lat"=>44.887950, "lon"=>-93.040125),
"/US/US-MN/spicer-10893" => array("type"=>"city", "pop"=>1087, "name_ru"=>"Спайсер", "name_en"=>"Spicer", "lang"=>"en", "lat"=>45.232639, "lon"=>-94.938648),
"/US/US-MN/spring-grove-10895" => array("type"=>"city", "pop"=>1239, "name_ru"=>"Спринг Гров", "name_en"=>"Spring Grove", "lang"=>"en", "lat"=>43.560975, "lon"=>-91.639475),
"/US/US-MN/spring-hill-10896" => array("type"=>"city", "pop"=>49, "name_ru"=>"Спринг Хилл", "name_en"=>"Spring Hill", "lang"=>"en", "lat"=>45.523389, "lon"=>-94.834397),
"/US/US-MN/spring-lake-park-10897" => array("type"=>"city", "pop"=>6458, "name_ru"=>"Спринг Лейк Парк", "name_en"=>"Spring Lake Park", "lang"=>"en", "lat"=>45.116050, "lon"=>-93.244849),
"/US/US-MN/spring-park-10898" => array("type"=>"city", "pop"=>1852, "name_ru"=>"Спринг Парк", "name_en"=>"Spring Park", "lang"=>"en", "lat"=>44.935600, "lon"=>-93.633131),
"/US/US-MN/spring-valley-10899" => array("type"=>"city", "pop"=>2355, "name_ru"=>"Спринг Валли", "name_en"=>"Spring Valley", "lang"=>"en", "lat"=>43.686015, "lon"=>-92.389911),
"/US/US-MN/springfield-10894" => array("type"=>"city", "pop"=>2148, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>44.237943, "lon"=>-94.976801),
"/US/US-MN/squaw-lake-10900" => array("type"=>"city", "pop"=>94, "name_ru"=>"Скво Лейк", "name_en"=>"Squaw Lake", "lang"=>"en", "lat"=>47.626939, "lon"=>-94.140982),
"/US/US-MN/st.-anthony-10845" => array("type"=>"city", "pop"=>8341, "name_ru"=>"Сент Энтони", "name_en"=>"St. Anthony", "lang"=>"en", "lat"=>45.027350, "lon"=>-93.215299),
"/US/US-MN/st.-anthony-10846" => array("type"=>"city", "pop"=>81, "name_ru"=>"Сент Энтони", "name_en"=>"St. Anthony", "lang"=>"en", "lat"=>45.689368, "lon"=>-94.609127),
"/US/US-MN/st.-bonifacius-10847" => array("type"=>"city", "pop"=>2364, "name_ru"=>"Сент Бонивациус", "name_en"=>"St. Bonifacius", "lang"=>"en", "lat"=>44.904950, "lon"=>-93.746999),
"/US/US-MN/st.-charles-10848" => array("type"=>"city", "pop"=>3577, "name_ru"=>"Сент-Чарльз", "name_en"=>"St. Charles", "lang"=>"en", "lat"=>43.969640, "lon"=>-92.064080),
"/US/US-MN/st.-clair-10849" => array("type"=>"city", "pop"=>780, "name_ru"=>"Сент Клэр", "name_en"=>"St. Clair", "lang"=>"en", "lat"=>44.081088, "lon"=>-93.857536),
"/US/US-MN/st.-cloud-10850" => array("type"=>"city", "pop"=>66948, "name_ru"=>"Сент Клауд", "name_en"=>"St. Cloud", "lang"=>"en", "lat"=>45.553951, "lon"=>-94.170359),
"/US/US-MN/st.-francis-10851" => array("type"=>"city", "pop"=>7261, "name_ru"=>"Сент-Фрэнсис", "name_en"=>"St. Francis", "lang"=>"en", "lat"=>45.399760, "lon"=>-93.389566),
"/US/US-MN/st.-hilaire-10852" => array("type"=>"city", "pop"=>279, "name_ru"=>"Сент Хилари", "name_en"=>"St. Hilaire", "lang"=>"en", "lat"=>48.013225, "lon"=>-96.213764),
"/US/US-MN/st.-james-10853" => array("type"=>"city", "pop"=>4281, "name_ru"=>"Сент Джеймс", "name_en"=>"St. James", "lang"=>"en", "lat"=>43.984766, "lon"=>-94.625320),
"/US/US-MN/st.-joseph-10854" => array("type"=>"city", "pop"=>6114, "name_ru"=>"Сент Джозеф", "name_en"=>"St. Joseph", "lang"=>"en", "lat"=>45.566518, "lon"=>-94.317123),
"/US/US-MN/st.-leo-10855" => array("type"=>"city", "pop"=>91, "name_ru"=>"Сент Лео", "name_en"=>"St. Leo", "lang"=>"en", "lat"=>44.716142, "lon"=>-96.054675),
"/US/US-MN/st.-louis-park-10856" => array("type"=>"city", "pop"=>45465, "name_ru"=>"Сент Льюис Парк", "name_en"=>"St. Louis Park", "lang"=>"en", "lat"=>44.948600, "lon"=>-93.364874),
"/US/US-MN/st.-martin-10857" => array("type"=>"city", "pop"=>283, "name_ru"=>"Сент Мартин", "name_en"=>"St. Martin", "lang"=>"en", "lat"=>45.504101, "lon"=>-94.662696),
"/US/US-MN/st.-marys-point-10858" => array("type"=>"city", "pop"=>362, "name_ru"=>"Сент Мэрис Пойнт", "name_en"=>"St. Marys Point", "lang"=>"en", "lat"=>44.913550, "lon"=>-92.769149),
"/US/US-MN/st.-michael-10859" => array("type"=>"city", "pop"=>15278, "name_ru"=>"Сент-Майкл", "name_en"=>"St. Michael", "lang"=>"en", "lat"=>45.210900, "lon"=>-93.663674),
"/US/US-MN/st.-paul-10860" => array("type"=>"city", "pop"=>279590, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>44.947744, "lon"=>-93.103686),
"/US/US-MN/st.-paul-park-10861" => array("type"=>"city", "pop"=>5239, "name_ru"=>"Сент Пол Парк", "name_en"=>"St. Paul Park", "lang"=>"en", "lat"=>44.839450, "lon"=>-92.994968),
"/US/US-MN/st.-peter-10862" => array("type"=>"city", "pop"=>10971, "name_ru"=>"Сент Питер", "name_en"=>"St. Peter", "lang"=>"en", "lat"=>44.329019, "lon"=>-93.963852),
"/US/US-MN/st.-rosa-10863" => array("type"=>"city", "pop"=>46, "name_ru"=>"Сент Роза", "name_en"=>"St. Rosa", "lang"=>"en", "lat"=>45.731313, "lon"=>-94.715018),
"/US/US-MN/st.-stephen-10864" => array("type"=>"city", "pop"=>803, "name_ru"=>"Сент Стифен", "name_en"=>"St. Stephen", "lang"=>"en", "lat"=>45.702090, "lon"=>-94.276458),
"/US/US-MN/st.-vincent-10865" => array("type"=>"city", "pop"=>97, "name_ru"=>"Сент Винсент", "name_en"=>"St. Vincent", "lang"=>"en", "lat"=>48.971397, "lon"=>-97.225288),
"/US/US-MN/stacy-10901" => array("type"=>"city", "pop"=>1383, "name_ru"=>"Стейси", "name_en"=>"Stacy", "lang"=>"en", "lat"=>45.397876, "lon"=>-92.987774),
"/US/US-MN/staples-10902" => array("type"=>"city", "pop"=>3019, "name_ru"=>"Стейплс", "name_en"=>"Staples", "lang"=>"en", "lat"=>46.363978, "lon"=>-94.803428),
"/US/US-MN/starbuck-10903" => array("type"=>"city", "pop"=>1240, "name_ru"=>"Старбак", "name_en"=>"Starbuck", "lang"=>"en", "lat"=>45.612068, "lon"=>-95.532998),
"/US/US-MN/steen-10904" => array("type"=>"city", "pop"=>175, "name_ru"=>"Стин", "name_en"=>"Steen", "lang"=>"en", "lat"=>43.513236, "lon"=>-96.262222),
"/US/US-MN/stephen-10905" => array("type"=>"city", "pop"=>639, "name_ru"=>"Стифен", "name_en"=>"Stephen", "lang"=>"en", "lat"=>48.451481, "lon"=>-96.877165),
"/US/US-MN/stewart-10906" => array("type"=>"city", "pop"=>536, "name_ru"=>"Стьюарт", "name_en"=>"Stewart", "lang"=>"en", "lat"=>44.725091, "lon"=>-94.487871),
"/US/US-MN/stewartville-10907" => array("type"=>"city", "pop"=>5553, "name_ru"=>"Стьюартвилл", "name_en"=>"Stewartville", "lang"=>"en", "lat"=>43.857250, "lon"=>-92.487032),
"/US/US-MN/stillwater-10908" => array("type"=>"city", "pop"=>17896, "name_ru"=>"Стиллуотер", "name_en"=>"Stillwater", "lang"=>"en", "lat"=>45.056500, "lon"=>-92.822174),
"/US/US-MN/stockton-10909" => array("type"=>"city", "pop"=>819, "name_ru"=>"Стоктон", "name_en"=>"Stockton", "lang"=>"en", "lat"=>44.027960, "lon"=>-91.772304),
"/US/US-MN/storden-10910" => array("type"=>"city", "pop"=>248, "name_ru"=>"Сторден", "name_en"=>"Storden", "lang"=>"en", "lat"=>44.039589, "lon"=>-95.319928),
"/US/US-MN/strandquist-10911" => array("type"=>"city", "pop"=>83, "name_ru"=>"Стрэндквист", "name_en"=>"Strandquist", "lang"=>"en", "lat"=>48.489960, "lon"=>-96.448088),
"/US/US-MN/strathcona-10912" => array("type"=>"city", "pop"=>28, "name_ru"=>"Стрэткона", "name_en"=>"Strathcona", "lang"=>"en", "lat"=>48.552248, "lon"=>-96.167849),
"/US/US-MN/sturgeon-lake-10913" => array("type"=>"city", "pop"=>372, "name_ru"=>"Стерджион Лейк", "name_en"=>"Sturgeon Lake", "lang"=>"en", "lat"=>46.383276, "lon"=>-92.827052),
"/US/US-MN/sunburg-10914" => array("type"=>"city", "pop"=>101, "name_ru"=>"Санбург", "name_en"=>"Sunburg", "lang"=>"en", "lat"=>45.347944, "lon"=>-95.239418),
"/US/US-MN/sunfish-lake-10915" => array("type"=>"city", "pop"=>523, "name_ru"=>"Санфиш Лейк", "name_en"=>"Sunfish Lake", "lang"=>"en", "lat"=>44.874900, "lon"=>-93.096838),
"/US/US-MN/swanville-10916" => array("type"=>"city", "pop"=>338, "name_ru"=>"Свонвилл", "name_en"=>"Swanville", "lang"=>"en", "lat"=>45.916206, "lon"=>-94.639631),
"/US/US-MN/taconite-10917" => array("type"=>"city", "pop"=>293, "name_ru"=>"Таконит", "name_en"=>"Taconite", "lang"=>"en", "lat"=>47.323157, "lon"=>-93.360161),
"/US/US-MN/tamarack-10918" => array("type"=>"city", "pop"=>54, "name_ru"=>"Тамарак", "name_en"=>"Tamarack", "lang"=>"en", "lat"=>46.651941, "lon"=>-93.118953),
"/US/US-MN/taopi-10919" => array("type"=>"city", "pop"=>94, "name_ru"=>"Таопи", "name_en"=>"Taopi", "lang"=>"en", "lat"=>43.558436, "lon"=>-92.640416),
"/US/US-MN/taunton-10920" => array("type"=>"city", "pop"=>195, "name_ru"=>"Тонтон", "name_en"=>"Taunton", "lang"=>"en", "lat"=>44.593070, "lon"=>-96.063154),
"/US/US-MN/taylors-falls-10921" => array("type"=>"city", "pop"=>1031, "name_ru"=>"Тейлорс Фоллс", "name_en"=>"Taylors Falls", "lang"=>"en", "lat"=>45.411443, "lon"=>-92.661845),
"/US/US-MN/tenney-10922" => array("type"=>"city", "pop"=>5, "name_ru"=>"Тенни", "name_en"=>"Tenney", "lang"=>"en", "lat"=>46.044305, "lon"=>-96.453958),
"/US/US-MN/tenstrike-10923" => array("type"=>"city", "pop"=>207, "name_ru"=>"Тенстрайк", "name_en"=>"Tenstrike", "lang"=>"en", "lat"=>47.660751, "lon"=>-94.684612),
"/US/US-MN/thief-river-falls-10924" => array("type"=>"city", "pop"=>8472, "name_ru"=>"Тиф Ривер Фоллс", "name_en"=>"Thief River Falls", "lang"=>"en", "lat"=>48.113497, "lon"=>-96.175416),
"/US/US-MN/thomson-10925" => array("type"=>"city", "pop"=>163, "name_ru"=>"Томсон", "name_en"=>"Thomson", "lang"=>"en", "lat"=>46.664140, "lon"=>-92.389757),
"/US/US-MN/tintah-10926" => array("type"=>"city", "pop"=>68, "name_ru"=>"Тинта", "name_en"=>"Tintah", "lang"=>"en", "lat"=>46.011027, "lon"=>-96.320069),
"/US/US-MN/tonka-bay-10927" => array("type"=>"city", "pop"=>1545, "name_ru"=>"Тонка Бэй", "name_en"=>"Tonka Bay", "lang"=>"en", "lat"=>44.915667, "lon"=>-93.590424),
"/US/US-MN/tower-10928" => array("type"=>"city", "pop"=>477, "name_ru"=>"Тауэр", "name_en"=>"Tower", "lang"=>"en", "lat"=>47.810071, "lon"=>-92.290468),
"/US/US-MN/tracy-10929" => array("type"=>"city", "pop"=>2065, "name_ru"=>"Трейси", "name_en"=>"Tracy", "lang"=>"en", "lat"=>44.237272, "lon"=>-95.616916),
"/US/US-MN/trail-10930" => array("type"=>"city", "pop"=>60, "name_ru"=>"Трейл", "name_en"=>"Trail", "lang"=>"en", "lat"=>47.783720, "lon"=>-95.699059),
"/US/US-MN/trimont-10931" => array("type"=>"city", "pop"=>686, "name_ru"=>"Тримонт", "name_en"=>"Trimont", "lang"=>"en", "lat"=>43.761433, "lon"=>-94.715424),
"/US/US-MN/trommald-10932" => array("type"=>"city", "pop"=>129, "name_ru"=>"Троммалд", "name_en"=>"Trommald", "lang"=>"en", "lat"=>46.504507, "lon"=>-94.016677),
"/US/US-MN/trosky-10933" => array("type"=>"city", "pop"=>110, "name_ru"=>"Троски", "name_en"=>"Trosky", "lang"=>"en", "lat"=>43.888350, "lon"=>-96.255681),
"/US/US-MN/truman-10934" => array("type"=>"city", "pop"=>1172, "name_ru"=>"Труман", "name_en"=>"Truman", "lang"=>"en", "lat"=>43.828043, "lon"=>-94.436157),
"/US/US-MN/turtle-river-10935" => array("type"=>"city", "pop"=>80, "name_ru"=>"Тертл Ривер", "name_en"=>"Turtle River", "lang"=>"en", "lat"=>47.588454, "lon"=>-94.760193),
"/US/US-MN/twin-lakes-10936" => array("type"=>"city", "pop"=>161, "name_ru"=>"Твин Лейкс", "name_en"=>"Twin Lakes", "lang"=>"en", "lat"=>43.560892, "lon"=>-93.423337),
"/US/US-MN/twin-valley-10937" => array("type"=>"city", "pop"=>783, "name_ru"=>"Твин Валли", "name_en"=>"Twin Valley", "lang"=>"en", "lat"=>47.257764, "lon"=>-96.258444),
"/US/US-MN/two-harbors-10938" => array("type"=>"city", "pop"=>3287, "name_ru"=>"Ту Харборс", "name_en"=>"Two Harbors", "lang"=>"en", "lat"=>47.030338, "lon"=>-91.675668),
"/US/US-MN/tyler-10939" => array("type"=>"city", "pop"=>1122, "name_ru"=>"Тайлер", "name_en"=>"Tyler", "lang"=>"en", "lat"=>44.279253, "lon"=>-96.135121),
"/US/US-MN/ulen-10940" => array("type"=>"city", "pop"=>560, "name_ru"=>"Улен", "name_en"=>"Ulen", "lang"=>"en", "lat"=>47.079294, "lon"=>-96.259537),
"/US/US-MN/underwood-10941" => array("type"=>"city", "pop"=>337, "name_ru"=>"Андервуд", "name_en"=>"Underwood", "lang"=>"en", "lat"=>46.285238, "lon"=>-95.870004),
"/US/US-MN/upsala-10942" => array("type"=>"city", "pop"=>396, "name_ru"=>"Упсала", "name_en"=>"Upsala", "lang"=>"en", "lat"=>45.808094, "lon"=>-94.569019),
"/US/US-MN/urbank-10943" => array("type"=>"city", "pop"=>58, "name_ru"=>"Урбанк", "name_en"=>"Urbank", "lang"=>"en", "lat"=>46.125635, "lon"=>-95.510558),
"/US/US-MN/utica-10944" => array("type"=>"city", "pop"=>251, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>43.976449, "lon"=>-91.950681),
"/US/US-MN/vadnais-heights-10945" => array("type"=>"city", "pop"=>12506, "name_ru"=>"Ваднейс Хайтс", "name_en"=>"Vadnais Heights", "lang"=>"en", "lat"=>45.056424, "lon"=>-93.072971),
"/US/US-MN/vergas-10946" => array("type"=>"city", "pop"=>293, "name_ru"=>"Вергас", "name_en"=>"Vergas", "lang"=>"en", "lat"=>46.655339, "lon"=>-95.803807),
"/US/US-MN/vermillion-10947" => array("type"=>"city", "pop"=>425, "name_ru"=>"Вермиллион", "name_en"=>"Vermillion", "lang"=>"en", "lat"=>44.674550, "lon"=>-92.964078),
"/US/US-MN/verndale-10948" => array("type"=>"city", "pop"=>529, "name_ru"=>"Верндейл", "name_en"=>"Verndale", "lang"=>"en", "lat"=>46.397301, "lon"=>-95.011513),
"/US/US-MN/vernon-center-10949" => array("type"=>"city", "pop"=>330, "name_ru"=>"Вернон Сентер", "name_en"=>"Vernon Center", "lang"=>"en", "lat"=>43.963401, "lon"=>-94.166685),
"/US/US-MN/vesta-10950" => array("type"=>"city", "pop"=>309, "name_ru"=>"Веста", "name_en"=>"Vesta", "lang"=>"en", "lat"=>44.506452, "lon"=>-95.413710),
"/US/US-MN/victoria-10951" => array("type"=>"city", "pop"=>6643, "name_ru"=>"Виктория", "name_en"=>"Victoria", "lang"=>"en", "lat"=>44.866077, "lon"=>-93.658408),
"/US/US-MN/viking-10952" => array("type"=>"city", "pop"=>87, "name_ru"=>"Вайкинг", "name_en"=>"Viking", "lang"=>"en", "lat"=>48.217945, "lon"=>-96.406416),
"/US/US-MN/villard-10953" => array("type"=>"city", "pop"=>220, "name_ru"=>"Виллард", "name_en"=>"Villard", "lang"=>"en", "lat"=>45.713472, "lon"=>-95.269644),
"/US/US-MN/vineland-10954" => array("type"=>"cdp", "pop"=>137, "name_ru"=>"Вайнленд", "name_en"=>"Vineland", "lang"=>"en", "lat"=>46.182741, "lon"=>-93.773614),
"/US/US-MN/vining-10955" => array("type"=>"city", "pop"=>66, "name_ru"=>"Вининг", "name_en"=>"Vining", "lang"=>"en", "lat"=>46.263168, "lon"=>-95.533832),
"/US/US-MN/virginia-10956" => array("type"=>"city", "pop"=>8480, "name_ru"=>"Вирджиния", "name_en"=>"Virginia", "lang"=>"en", "lat"=>47.515789, "lon"=>-92.507070),
"/US/US-MN/wabasha-10957" => array("type"=>"city", "pop"=>2520, "name_ru"=>"Вабаша", "name_en"=>"Wabasha", "lang"=>"en", "lat"=>44.371480, "lon"=>-92.045050),
"/US/US-MN/wabasso-10958" => array("type"=>"city", "pop"=>639, "name_ru"=>"Вабассо", "name_en"=>"Wabasso", "lang"=>"en", "lat"=>44.402633, "lon"=>-95.255028),
"/US/US-MN/waconia-10959" => array("type"=>"city", "pop"=>9934, "name_ru"=>"Вакония", "name_en"=>"Waconia", "lang"=>"en", "lat"=>44.847900, "lon"=>-93.794149),
"/US/US-MN/wadena-10960" => array("type"=>"city", "pop"=>3980, "name_ru"=>"Вадена", "name_en"=>"Wadena", "lang"=>"en", "lat"=>46.439940, "lon"=>-95.134812),
"/US/US-MN/wahkon-10961" => array("type"=>"city", "pop"=>333, "name_ru"=>"Вакон", "name_en"=>"Wahkon", "lang"=>"en", "lat"=>46.118622, "lon"=>-93.523663),
"/US/US-MN/waite-park-10962" => array("type"=>"city", "pop"=>6796, "name_ru"=>"Уэйт Парк", "name_en"=>"Waite Park", "lang"=>"en", "lat"=>45.552501, "lon"=>-94.223564),
"/US/US-MN/waldorf-10963" => array("type"=>"city", "pop"=>232, "name_ru"=>"Уолдорф", "name_en"=>"Waldorf", "lang"=>"en", "lat"=>43.932588, "lon"=>-93.696680),
"/US/US-MN/walker-10964" => array("type"=>"city", "pop"=>1289, "name_ru"=>"Уолкер", "name_en"=>"Walker", "lang"=>"en", "lat"=>47.089800, "lon"=>-94.578927),
"/US/US-MN/walnut-grove-10965" => array("type"=>"city", "pop"=>536, "name_ru"=>"Уолнат Гров", "name_en"=>"Walnut Grove", "lang"=>"en", "lat"=>44.225639, "lon"=>-95.468603),
"/US/US-MN/walters-10966" => array("type"=>"city", "pop"=>77, "name_ru"=>"Уолтерс", "name_en"=>"Walters", "lang"=>"en", "lat"=>43.605203, "lon"=>-93.673512),
"/US/US-MN/waltham-10967" => array("type"=>"city", "pop"=>183, "name_ru"=>"Уолтэм", "name_en"=>"Waltham", "lang"=>"en", "lat"=>43.820748, "lon"=>-92.875204),
"/US/US-MN/wanamingo-10968" => array("type"=>"city", "pop"=>1044, "name_ru"=>"Ванаминго", "name_en"=>"Wanamingo", "lang"=>"en", "lat"=>44.305268, "lon"=>-92.791526),
"/US/US-MN/wanda-10969" => array("type"=>"city", "pop"=>95, "name_ru"=>"Ванда", "name_en"=>"Wanda", "lang"=>"en", "lat"=>44.315966, "lon"=>-95.212141),
"/US/US-MN/warba-10970" => array("type"=>"city", "pop"=>176, "name_ru"=>"Варба", "name_en"=>"Warba", "lang"=>"en", "lat"=>47.135766, "lon"=>-93.270903),
"/US/US-MN/warren-10971" => array("type"=>"city", "pop"=>1556, "name_ru"=>"Уоррен", "name_en"=>"Warren", "lang"=>"en", "lat"=>48.196064, "lon"=>-96.769842),
"/US/US-MN/warroad-10972" => array("type"=>"city", "pop"=>1656, "name_ru"=>"Уорроуд", "name_en"=>"Warroad", "lang"=>"en", "lat"=>48.908543, "lon"=>-95.320706),
"/US/US-MN/waseca-10973" => array("type"=>"city", "pop"=>9438, "name_ru"=>"Васека", "name_en"=>"Waseca", "lang"=>"en", "lat"=>44.081124, "lon"=>-93.501928),
"/US/US-MN/watertown-10974" => array("type"=>"city", "pop"=>4118, "name_ru"=>"Уотертаун", "name_en"=>"Watertown", "lang"=>"en", "lat"=>44.963147, "lon"=>-93.845730),
"/US/US-MN/waterville-10975" => array("type"=>"city", "pop"=>1893, "name_ru"=>"Уотервилл", "name_en"=>"Waterville", "lang"=>"en", "lat"=>44.222666, "lon"=>-93.574261),
"/US/US-MN/watkins-10976" => array("type"=>"city", "pop"=>927, "name_ru"=>"Уоткинс", "name_en"=>"Watkins", "lang"=>"en", "lat"=>45.314811, "lon"=>-94.407942),
"/US/US-MN/watson-10977" => array("type"=>"city", "pop"=>192, "name_ru"=>"Уотсон", "name_en"=>"Watson", "lang"=>"en", "lat"=>45.010278, "lon"=>-95.799886),
"/US/US-MN/waubun-10978" => array("type"=>"city", "pop"=>394, "name_ru"=>"Уобан", "name_en"=>"Waubun", "lang"=>"en", "lat"=>47.184079, "lon"=>-95.939491),
"/US/US-MN/waverly-10979" => array("type"=>"city", "pop"=>1076, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>45.068935, "lon"=>-93.969003),
"/US/US-MN/wayzata-10980" => array("type"=>"city", "pop"=>3884, "name_ru"=>"Уэйзата", "name_en"=>"Wayzata", "lang"=>"en", "lat"=>44.970300, "lon"=>-93.513819),
"/US/US-MN/welcome-10981" => array("type"=>"city", "pop"=>678, "name_ru"=>"Вэлкам", "name_en"=>"Welcome", "lang"=>"en", "lat"=>43.668387, "lon"=>-94.618527),
"/US/US-MN/wells-10982" => array("type"=>"city", "pop"=>2346, "name_ru"=>"Уэллс", "name_en"=>"Wells", "lang"=>"en", "lat"=>43.743689, "lon"=>-93.726157),
"/US/US-MN/wendell-10983" => array("type"=>"city", "pop"=>170, "name_ru"=>"Уэнделл", "name_en"=>"Wendell", "lang"=>"en", "lat"=>46.035427, "lon"=>-96.100969),
"/US/US-MN/west-concord-10985" => array("type"=>"city", "pop"=>806, "name_ru"=>"Вэст Конкорд", "name_en"=>"West Concord", "lang"=>"en", "lat"=>44.153026, "lon"=>-92.899718),
"/US/US-MN/west-st.-paul-10987" => array("type"=>"city", "pop"=>18935, "name_ru"=>"Вэст Сент Пол", "name_en"=>"West St. Paul", "lang"=>"en", "lat"=>44.902600, "lon"=>-93.085599),
"/US/US-MN/west-union-10988" => array("type"=>"city", "pop"=>89, "name_ru"=>"Вэст Юнион", "name_en"=>"West Union", "lang"=>"en", "lat"=>45.801418, "lon"=>-95.083326),
"/US/US-MN/westbrook-10984" => array("type"=>"city", "pop"=>678, "name_ru"=>"Вэстбрук", "name_en"=>"Westbrook", "lang"=>"en", "lat"=>44.042607, "lon"=>-95.434947),
"/US/US-MN/westport-10986" => array("type"=>"city", "pop"=>66, "name_ru"=>"Уэстпорт", "name_en"=>"Westport", "lang"=>"en", "lat"=>45.714575, "lon"=>-95.167291),
"/US/US-MN/whalan-10989" => array("type"=>"city", "pop"=>60, "name_ru"=>"Уолан", "name_en"=>"Whalan", "lang"=>"en", "lat"=>43.732796, "lon"=>-91.927708),
"/US/US-MN/wheaton-10990" => array("type"=>"city", "pop"=>1422, "name_ru"=>"Уитон", "name_en"=>"Wheaton", "lang"=>"en", "lat"=>45.805762, "lon"=>-96.497250),
"/US/US-MN/white-bear-lake-10991" => array("type"=>"city", "pop"=>24095, "name_ru"=>"Уайт Бэр Лейк", "name_en"=>"White Bear Lake", "lang"=>"en", "lat"=>45.067650, "lon"=>-93.012671),
"/US/US-MN/white-earth-10992" => array("type"=>"cdp", "pop"=>851, "name_ru"=>"Уайт Эрт", "name_en"=>"White Earth", "lang"=>"en", "lat"=>47.098278, "lon"=>-95.855370),
"/US/US-MN/wilder-10993" => array("type"=>"city", "pop"=>66, "name_ru"=>"Уайлдер", "name_en"=>"Wilder", "lang"=>"en", "lat"=>43.825531, "lon"=>-95.206304),
"/US/US-MN/willernie-10994" => array("type"=>"city", "pop"=>519, "name_ru"=>"Уиллерни", "name_en"=>"Willernie", "lang"=>"en", "lat"=>45.053900, "lon"=>-92.956332),
"/US/US-MN/williams-10995" => array("type"=>"city", "pop"=>178, "name_ru"=>"Уилльямс", "name_en"=>"Williams", "lang"=>"en", "lat"=>48.768484, "lon"=>-94.952694),
"/US/US-MN/willmar-10996" => array("type"=>"city", "pop"=>17774, "name_ru"=>"Уиллмар", "name_en"=>"Willmar", "lang"=>"en", "lat"=>45.124601, "lon"=>-95.049567),
"/US/US-MN/willow-river-10997" => array("type"=>"city", "pop"=>391, "name_ru"=>"Уиллоу Ривер", "name_en"=>"Willow River", "lang"=>"en", "lat"=>46.319991, "lon"=>-92.836685),
"/US/US-MN/wilmont-10998" => array("type"=>"city", "pop"=>319, "name_ru"=>"Уилмонт", "name_en"=>"Wilmont", "lang"=>"en", "lat"=>43.763971, "lon"=>-95.828306),
"/US/US-MN/wilton-10999" => array("type"=>"city", "pop"=>195, "name_ru"=>"Уилтон", "name_en"=>"Wilton", "lang"=>"en", "lat"=>47.505211, "lon"=>-94.991175),
"/US/US-MN/windom-11000" => array("type"=>"city", "pop"=>4197, "name_ru"=>"Уиндом", "name_en"=>"Windom", "lang"=>"en", "lat"=>43.869463, "lon"=>-95.118300),
"/US/US-MN/winger-11001" => array("type"=>"city", "pop"=>196, "name_ru"=>"Уингер", "name_en"=>"Winger", "lang"=>"en", "lat"=>47.536970, "lon"=>-95.985068),
"/US/US-MN/winnebago-11002" => array("type"=>"city", "pop"=>1335, "name_ru"=>"Уиннебаго", "name_en"=>"Winnebago", "lang"=>"en", "lat"=>43.766038, "lon"=>-94.168267),
"/US/US-MN/winona-11003" => array("type"=>"city", "pop"=>26785, "name_ru"=>"Уайнона", "name_en"=>"Winona", "lang"=>"en", "lat"=>44.047776, "lon"=>-91.657138),
"/US/US-MN/winsted-11004" => array("type"=>"city", "pop"=>2418, "name_ru"=>"Уинстед", "name_en"=>"Winsted", "lang"=>"en", "lat"=>44.965810, "lon"=>-94.048757),
"/US/US-MN/winthrop-11005" => array("type"=>"city", "pop"=>1256, "name_ru"=>"Уинтроп", "name_en"=>"Winthrop", "lang"=>"en", "lat"=>44.543722, "lon"=>-94.364646),
"/US/US-MN/winton-11006" => array("type"=>"city", "pop"=>187, "name_ru"=>"Уинтон", "name_en"=>"Winton", "lang"=>"en", "lat"=>47.929005, "lon"=>-91.801368),
"/US/US-MN/wolf-lake-11007" => array("type"=>"city", "pop"=>38, "name_ru"=>"Вулф Лейк", "name_en"=>"Wolf Lake", "lang"=>"en", "lat"=>46.802881, "lon"=>-95.355118),
"/US/US-MN/wolverton-11008" => array("type"=>"city", "pop"=>103, "name_ru"=>"Вулвертон", "name_en"=>"Wolverton", "lang"=>"en", "lat"=>46.564714, "lon"=>-96.735502),
"/US/US-MN/wood-lake-11010" => array("type"=>"city", "pop"=>381, "name_ru"=>"Вуд Лейк", "name_en"=>"Wood Lake", "lang"=>"en", "lat"=>44.650779, "lon"=>-95.534939),
"/US/US-MN/woodbury-11009" => array("type"=>"city", "pop"=>56128, "name_ru"=>"Вудбери", "name_en"=>"Woodbury", "lang"=>"en", "lat"=>44.905483, "lon"=>-92.920907),
"/US/US-MN/woodland-11011" => array("type"=>"city", "pop"=>469, "name_ru"=>"Вудленд", "name_en"=>"Woodland", "lang"=>"en", "lat"=>44.950732, "lon"=>-93.514772),
"/US/US-MN/woodstock-11012" => array("type"=>"city", "pop"=>124, "name_ru"=>"Вудсток", "name_en"=>"Woodstock", "lang"=>"en", "lat"=>44.010390, "lon"=>-96.096519),
"/US/US-MN/worthington-11013" => array("type"=>"city", "pop"=>11036, "name_ru"=>"Уортингтон", "name_en"=>"Worthington", "lang"=>"en", "lat"=>43.627158, "lon"=>-95.598436),
"/US/US-MN/wrenshall-11014" => array("type"=>"city", "pop"=>386, "name_ru"=>"Ренсхолл", "name_en"=>"Wrenshall", "lang"=>"en", "lat"=>46.618288, "lon"=>-92.383553),
"/US/US-MN/wright-11015" => array("type"=>"city", "pop"=>91, "name_ru"=>"Райт", "name_en"=>"Wright", "lang"=>"en", "lat"=>46.672086, "lon"=>-93.006655),
"/US/US-MN/wykoff-11016" => array("type"=>"city", "pop"=>433, "name_ru"=>"Уайкофф", "name_en"=>"Wykoff", "lang"=>"en", "lat"=>43.708161, "lon"=>-92.267421),
"/US/US-MN/wyoming-11017" => array("type"=>"city", "pop"=>3803, "name_ru"=>"Вайоминг", "name_en"=>"Wyoming", "lang"=>"en", "lat"=>45.333600, "lon"=>-92.995999),
"/US/US-MN/young-america-11018" => array("type"=>"city", "pop"=>829, "name_ru"=>"Янг Америка", "name_en"=>"Young America", "lang"=>"en", "lat"=>44.780750, "lon"=>-93.915324),
"/US/US-MN/zemple-11019" => array("type"=>"city", "pop"=>74, "name_ru"=>"Земпл", "name_en"=>"Zemple", "lang"=>"en", "lat"=>47.323393, "lon"=>-93.786176),
"/US/US-MN/zimmerman-11020" => array("type"=>"city", "pop"=>5313, "name_ru"=>"Циммерман", "name_en"=>"Zimmerman", "lang"=>"en", "lat"=>45.443971, "lon"=>-93.592354),
"/US/US-MN/zumbro-falls-11021" => array("type"=>"city", "pop"=>172, "name_ru"=>"Зумбро Фоллс", "name_en"=>"Zumbro Falls", "lang"=>"en", "lat"=>44.286852, "lon"=>-92.426334),
"/US/US-MN/zumbrota-11022" => array("type"=>"city", "pop"=>3074, "name_ru"=>"Зумброта", "name_en"=>"Zumbrota", "lang"=>"en", "lat"=>44.294462, "lon"=>-92.672607),
"/US/US-MO" => array("type"=>"state", "pop"=>5911605, "name_ru"=>"Миссури", "name_en"=>"Missouri", "lang"=>"en"),
"/US/US-MO/adrian-11347" => array("type"=>"city", "pop"=>1899, "name_ru"=>"Эдриан", "name_en"=>"Adrian", "lang"=>"en", "lat"=>38.397857, "lon"=>-94.350932),
"/US/US-MO/advance-11348" => array("type"=>"city", "pop"=>1215, "name_ru"=>"Адванс", "name_en"=>"Advance", "lang"=>"en", "lat"=>37.103553, "lon"=>-89.914533),
"/US/US-MO/affton-11349" => array("type"=>"cdp", "pop"=>21106, "name_ru"=>"Аффтон", "name_en"=>"Affton", "lang"=>"en", "lat"=>38.549450, "lon"=>-90.328911),
"/US/US-MO/agency-11350" => array("type"=>"town", "pop"=>630, "name_ru"=>"Эдженси", "name_en"=>"Agency", "lang"=>"en", "lat"=>39.655688, "lon"=>-94.748538),
"/US/US-MO/airport-drive-11351" => array("type"=>"village", "pop"=>730, "name_ru"=>"Эрпорт Драйв", "name_en"=>"Airport Drive", "lang"=>"en", "lat"=>37.138977, "lon"=>-94.515768),
"/US/US-MO/alba-11352" => array("type"=>"city", "pop"=>587, "name_ru"=>"Альба", "name_en"=>"Alba", "lang"=>"en", "lat"=>37.237146, "lon"=>-94.416954),
"/US/US-MO/albany-11353" => array("type"=>"city", "pop"=>1736, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>40.248160, "lon"=>-94.333880),
"/US/US-MO/aldrich-11354" => array("type"=>"village", "pop"=>81, "name_ru"=>"Олдрич", "name_en"=>"Aldrich", "lang"=>"en", "lat"=>37.549089, "lon"=>-93.552305),
"/US/US-MO/alexandria-11355" => array("type"=>"city", "pop"=>164, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>40.359732, "lon"=>-91.459222),
"/US/US-MO/allendale-11356" => array("type"=>"town", "pop"=>47, "name_ru"=>"Аллендейл", "name_en"=>"Allendale", "lang"=>"en", "lat"=>40.486215, "lon"=>-94.288700),
"/US/US-MO/allenville-11357" => array("type"=>"village", "pop"=>108, "name_ru"=>"Алленвилл", "name_en"=>"Allenville", "lang"=>"en", "lat"=>37.221563, "lon"=>-89.754943),
"/US/US-MO/alma-11358" => array("type"=>"city", "pop"=>368, "name_ru"=>"Алма", "name_en"=>"Alma", "lang"=>"en", "lat"=>39.095831, "lon"=>-93.548012),
"/US/US-MO/altamont-11359" => array("type"=>"town", "pop"=>217, "name_ru"=>"Альтамонт", "name_en"=>"Altamont", "lang"=>"en", "lat"=>39.888852, "lon"=>-94.086745),
"/US/US-MO/altenburg-11360" => array("type"=>"city", "pop"=>312, "name_ru"=>"Альтенбург", "name_en"=>"Altenburg", "lang"=>"en", "lat"=>37.631251, "lon"=>-89.581415),
"/US/US-MO/alton-11361" => array("type"=>"city", "pop"=>638, "name_ru"=>"Альтон", "name_en"=>"Alton", "lang"=>"en", "lat"=>36.693862, "lon"=>-91.398939),
"/US/US-MO/amazonia-11362" => array("type"=>"town", "pop"=>283, "name_ru"=>"Амазония", "name_en"=>"Amazonia", "lang"=>"en", "lat"=>39.888343, "lon"=>-94.891896),
"/US/US-MO/amity-11363" => array("type"=>"town", "pop"=>66, "name_ru"=>"Амити", "name_en"=>"Amity", "lang"=>"en", "lat"=>39.868577, "lon"=>-94.434502),
"/US/US-MO/amoret-11364" => array("type"=>"city", "pop"=>216, "name_ru"=>"Аморет", "name_en"=>"Amoret", "lang"=>"en", "lat"=>38.255652, "lon"=>-94.587089),
"/US/US-MO/amsterdam-11365" => array("type"=>"city", "pop"=>288, "name_ru"=>"Амсердам", "name_en"=>"Amsterdam", "lang"=>"en", "lat"=>38.349338, "lon"=>-94.588866),
"/US/US-MO/anderson-11366" => array("type"=>"city", "pop"=>2055, "name_ru"=>"Андерсон", "name_en"=>"Anderson", "lang"=>"en", "lat"=>36.654834, "lon"=>-94.448585),
"/US/US-MO/annada-11367" => array("type"=>"town", "pop"=>49, "name_ru"=>"Аннада", "name_en"=>"Annada", "lang"=>"en", "lat"=>39.262161, "lon"=>-90.828539),
"/US/US-MO/annapolis-11368" => array("type"=>"city", "pop"=>288, "name_ru"=>"Аннаполис", "name_en"=>"Annapolis", "lang"=>"en", "lat"=>37.359944, "lon"=>-90.698281),
"/US/US-MO/anniston-11369" => array("type"=>"town", "pop"=>271, "name_ru"=>"Аннистон", "name_en"=>"Anniston", "lang"=>"en", "lat"=>36.824519, "lon"=>-89.326199),
"/US/US-MO/appleton-city-11370" => array("type"=>"city", "pop"=>1263, "name_ru"=>"Эпплстон Сити", "name_en"=>"Appleton City", "lang"=>"en", "lat"=>38.189975, "lon"=>-94.030303),
"/US/US-MO/arbela-11371" => array("type"=>"town", "pop"=>39, "name_ru"=>"Арбела", "name_en"=>"Arbela", "lang"=>"en", "lat"=>40.462952, "lon"=>-92.015962),
"/US/US-MO/arbyrd-11372" => array("type"=>"city", "pop"=>489, "name_ru"=>"Арбид", "name_en"=>"Arbyrd", "lang"=>"en", "lat"=>36.052644, "lon"=>-90.239999),
"/US/US-MO/arcadia-11373" => array("type"=>"city", "pop"=>508, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>37.585383, "lon"=>-90.628499),
"/US/US-MO/archie-11374" => array("type"=>"city", "pop"=>996, "name_ru"=>"Арчи", "name_en"=>"Archie", "lang"=>"en", "lat"=>38.482853, "lon"=>-94.349042),
"/US/US-MO/arcola-11375" => array("type"=>"village", "pop"=>43, "name_ru"=>"Аркола", "name_en"=>"Arcola", "lang"=>"en", "lat"=>37.549932, "lon"=>-93.875667),
"/US/US-MO/argyle-11376" => array("type"=>"town", "pop"=>169, "name_ru"=>"Аргайл", "name_en"=>"Argyle", "lang"=>"en", "lat"=>38.295011, "lon"=>-92.025457),
"/US/US-MO/arkoe-11377" => array("type"=>"town", "pop"=>57, "name_ru"=>"Арко", "name_en"=>"Arkoe", "lang"=>"en", "lat"=>40.258788, "lon"=>-94.829581),
"/US/US-MO/armstrong-11378" => array("type"=>"city", "pop"=>281, "name_ru"=>"Армстронг", "name_en"=>"Armstrong", "lang"=>"en", "lat"=>39.269349, "lon"=>-92.705009),
"/US/US-MO/arnold-11379" => array("type"=>"city", "pop"=>20603, "name_ru"=>"Арнолд", "name_en"=>"Arnold", "lang"=>"en", "lat"=>38.428433, "lon"=>-90.369299),
"/US/US-MO/arrow-rock-11380" => array("type"=>"town", "pop"=>75, "name_ru"=>"Эрроу Рок", "name_en"=>"Arrow Rock", "lang"=>"en", "lat"=>39.069966, "lon"=>-92.948002),
"/US/US-MO/asbury-11381" => array("type"=>"city", "pop"=>232, "name_ru"=>"Асбери", "name_en"=>"Asbury", "lang"=>"en", "lat"=>37.273778, "lon"=>-94.602936),
"/US/US-MO/ash-grove-11383" => array("type"=>"city", "pop"=>1529, "name_ru"=>"Аш Гров", "name_en"=>"Ash Grove", "lang"=>"en", "lat"=>37.317440, "lon"=>-93.580526),
"/US/US-MO/ashburn-11382" => array("type"=>"town", "pop"=>51, "name_ru"=>"Эшберн", "name_en"=>"Ashburn", "lang"=>"en", "lat"=>39.546259, "lon"=>-91.170989),
"/US/US-MO/ashland-11384" => array("type"=>"city", "pop"=>2190, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>38.773867, "lon"=>-92.256677),
"/US/US-MO/atlanta-11385" => array("type"=>"city", "pop"=>450, "name_ru"=>"Атланта", "name_en"=>"Atlanta", "lang"=>"en", "lat"=>39.898021, "lon"=>-92.479815),
"/US/US-MO/augusta-11386" => array("type"=>"city", "pop"=>215, "name_ru"=>"Огаста", "name_en"=>"Augusta", "lang"=>"en", "lat"=>38.572857, "lon"=>-90.882795),
"/US/US-MO/aullville-11387" => array("type"=>"village", "pop"=>84, "name_ru"=>"Олвилл", "name_en"=>"Aullville", "lang"=>"en", "lat"=>39.018239, "lon"=>-93.677468),
"/US/US-MO/aurora-11388" => array("type"=>"city", "pop"=>7478, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>36.969899, "lon"=>-93.720373),
"/US/US-MO/auxvasse-11389" => array("type"=>"city", "pop"=>994, "name_ru"=>"Оксвассе", "name_en"=>"Auxvasse", "lang"=>"en", "lat"=>39.017365, "lon"=>-91.894500),
"/US/US-MO/ava-11390" => array("type"=>"city", "pop"=>3149, "name_ru"=>"Ава", "name_en"=>"Ava", "lang"=>"en", "lat"=>36.953096, "lon"=>-92.665107),
"/US/US-MO/avilla-11391" => array("type"=>"town", "pop"=>146, "name_ru"=>"Авилла", "name_en"=>"Avilla", "lang"=>"en", "lat"=>37.194257, "lon"=>-94.130210),
"/US/US-MO/avondale-11392" => array("type"=>"city", "pop"=>543, "name_ru"=>"Авондейл", "name_en"=>"Avondale", "lang"=>"en", "lat"=>39.153505, "lon"=>-94.544969),
"/US/US-MO/bagnell-11393" => array("type"=>"town", "pop"=>91, "name_ru"=>"Бэгнелл", "name_en"=>"Bagnell", "lang"=>"en", "lat"=>38.228862, "lon"=>-92.605088),
"/US/US-MO/baker-11394" => array("type"=>"village", "pop"=>224, "name_ru"=>"Бейкер", "name_en"=>"Baker", "lang"=>"en", "lat"=>36.773560, "lon"=>-89.761430),
"/US/US-MO/bakersfield-11395" => array("type"=>"village", "pop"=>277, "name_ru"=>"Бейкерсфилд", "name_en"=>"Bakersfield", "lang"=>"en", "lat"=>36.524084, "lon"=>-92.144227),
"/US/US-MO/baldwin-park-11396" => array("type"=>"village", "pop"=>119, "name_ru"=>"Бодуин Парк", "name_en"=>"Baldwin Park", "lang"=>"en", "lat"=>38.795312, "lon"=>-94.245417),
"/US/US-MO/ballwin-11397" => array("type"=>"city", "pop"=>30033, "name_ru"=>"Боллуин", "name_en"=>"Ballwin", "lang"=>"en", "lat"=>38.594100, "lon"=>-90.552374),
"/US/US-MO/baring-11398" => array("type"=>"city", "pop"=>147, "name_ru"=>"Баринг", "name_en"=>"Baring", "lang"=>"en", "lat"=>40.244678, "lon"=>-92.205683),
"/US/US-MO/barnard-11399" => array("type"=>"city", "pop"=>242, "name_ru"=>"Барнард", "name_en"=>"Barnard", "lang"=>"en", "lat"=>40.175700, "lon"=>-94.822578),
"/US/US-MO/barnett-11400" => array("type"=>"city", "pop"=>229, "name_ru"=>"Барнетт", "name_en"=>"Barnett", "lang"=>"en", "lat"=>38.377020, "lon"=>-92.674004),
"/US/US-MO/barnhart-11401" => array("type"=>"cdp", "pop"=>4911, "name_ru"=>"Барнхарт", "name_en"=>"Barnhart", "lang"=>"en", "lat"=>38.334394, "lon"=>-90.404100),
"/US/US-MO/bates-city-11402" => array("type"=>"village", "pop"=>266, "name_ru"=>"Бейтс Сити", "name_en"=>"Bates City", "lang"=>"en", "lat"=>39.006192, "lon"=>-94.070240),
"/US/US-MO/battlefield-11403" => array("type"=>"town", "pop"=>4249, "name_ru"=>"Бэттлфилд", "name_en"=>"Battlefield", "lang"=>"en", "lat"=>37.118100, "lon"=>-93.367849),
"/US/US-MO/bel-nor-11410" => array("type"=>"village", "pop"=>1481, "name_ru"=>"Бел-Нор", "name_en"=>"Bel-Nor", "lang"=>"en", "lat"=>38.703200, "lon"=>-90.316372),
"/US/US-MO/bel-ridge-11411" => array("type"=>"village", "pop"=>2898, "name_ru"=>"Бел-Ридж", "name_en"=>"Bel-Ridge", "lang"=>"en", "lat"=>38.714050, "lon"=>-90.327904),
"/US/US-MO/bell-city-11405" => array("type"=>"city", "pop"=>450, "name_ru"=>"Белл-Сити", "name_en"=>"Bell City", "lang"=>"en", "lat"=>37.023770, "lon"=>-89.819370),
"/US/US-MO/bella-villa-11404" => array("type"=>"city", "pop"=>636, "name_ru"=>"Белла Вилла", "name_en"=>"Bella Villa", "lang"=>"en", "lat"=>38.543450, "lon"=>-90.286183),
"/US/US-MO/belle-11406" => array("type"=>"city", "pop"=>1381, "name_ru"=>"Бель", "name_en"=>"Belle", "lang"=>"en", "lat"=>38.285772, "lon"=>-91.722289),
"/US/US-MO/bellefontaine-neighbors-11407" => array("type"=>"city", "pop"=>10241, "name_ru"=>"Бельфонтейн Нейборс", "name_en"=>"Bellefontaine Neighbors", "lang"=>"en", "lat"=>38.752033, "lon"=>-90.226895),
"/US/US-MO/bellerive-11408" => array("type"=>"village", "pop"=>252, "name_ru"=>"Беллерайв", "name_en"=>"Bellerive", "lang"=>"en", "lat"=>38.713050, "lon"=>-90.309848),
"/US/US-MO/bellflower-11409" => array("type"=>"city", "pop"=>389, "name_ru"=>"Беллфлауэр", "name_en"=>"Bellflower", "lang"=>"en", "lat"=>39.003849, "lon"=>-91.352721),
"/US/US-MO/belton-11412" => array("type"=>"city", "pop"=>24591, "name_ru"=>"Белтон", "name_en"=>"Belton", "lang"=>"en", "lat"=>38.819488, "lon"=>-94.531163),
"/US/US-MO/benton-11413" => array("type"=>"city", "pop"=>745, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>37.098041, "lon"=>-89.562077),
"/US/US-MO/benton-city-11414" => array("type"=>"town", "pop"=>123, "name_ru"=>"Бентон Сити", "name_en"=>"Benton City", "lang"=>"en", "lat"=>39.134759, "lon"=>-91.764277),
"/US/US-MO/berger-11415" => array("type"=>"city", "pop"=>201, "name_ru"=>"Бергер", "name_en"=>"Berger", "lang"=>"en", "lat"=>38.673956, "lon"=>-91.335245),
"/US/US-MO/berkeley-11416" => array("type"=>"city", "pop"=>9387, "name_ru"=>"Беркли", "name_en"=>"Berkeley", "lang"=>"en", "lat"=>38.743301, "lon"=>-90.335607),
"/US/US-MO/bernie-11417" => array("type"=>"city", "pop"=>1789, "name_ru"=>"Берни", "name_en"=>"Bernie", "lang"=>"en", "lat"=>36.671970, "lon"=>-89.970467),
"/US/US-MO/bertrand-11418" => array("type"=>"city", "pop"=>710, "name_ru"=>"Бертранд", "name_en"=>"Bertrand", "lang"=>"en", "lat"=>36.908977, "lon"=>-89.449667),
"/US/US-MO/bethany-11419" => array("type"=>"city", "pop"=>3069, "name_ru"=>"Бетани", "name_en"=>"Bethany", "lang"=>"en", "lat"=>40.267755, "lon"=>-94.028257),
"/US/US-MO/bethel-11420" => array("type"=>"town", "pop"=>115, "name_ru"=>"Бетел", "name_en"=>"Bethel", "lang"=>"en", "lat"=>39.878229, "lon"=>-92.022721),
"/US/US-MO/beverly-hills-11421" => array("type"=>"city", "pop"=>558, "name_ru"=>"Беверли Хиллс", "name_en"=>"Beverly Hills", "lang"=>"en", "lat"=>38.697750, "lon"=>-90.290371),
"/US/US-MO/bevier-11422" => array("type"=>"city", "pop"=>732, "name_ru"=>"Бевир", "name_en"=>"Bevier", "lang"=>"en", "lat"=>39.749094, "lon"=>-92.563714),
"/US/US-MO/big-lake-11424" => array("type"=>"village", "pop"=>120, "name_ru"=>"Биг Лейк", "name_en"=>"Big Lake", "lang"=>"en", "lat"=>40.074140, "lon"=>-95.351467),
"/US/US-MO/bigelow-11423" => array("type"=>"village", "pop"=>35, "name_ru"=>"Бигелоу", "name_en"=>"Bigelow", "lang"=>"en", "lat"=>40.109418, "lon"=>-95.288481),
"/US/US-MO/billings-11425" => array("type"=>"city", "pop"=>1118, "name_ru"=>"Биллингс", "name_en"=>"Billings", "lang"=>"en", "lat"=>37.063486, "lon"=>-93.555674),
"/US/US-MO/birch-tree-11426" => array("type"=>"city", "pop"=>624, "name_ru"=>"Берч Три", "name_en"=>"Birch Tree", "lang"=>"en", "lat"=>36.991619, "lon"=>-91.492409),
"/US/US-MO/birmingham-11427" => array("type"=>"village", "pop"=>222, "name_ru"=>"Бирмингем", "name_en"=>"Birmingham", "lang"=>"en", "lat"=>39.166650, "lon"=>-94.450449),
"/US/US-MO/bismarck-11428" => array("type"=>"city", "pop"=>1538, "name_ru"=>"Бисмарк", "name_en"=>"Bismarck", "lang"=>"en", "lat"=>37.767369, "lon"=>-90.622712),
"/US/US-MO/black-jack-11430" => array("type"=>"city", "pop"=>6818, "name_ru"=>"Блэк Джек", "name_en"=>"Black Jack", "lang"=>"en", "lat"=>38.798150, "lon"=>-90.263598),
"/US/US-MO/blackburn-11429" => array("type"=>"city", "pop"=>270, "name_ru"=>"Блэкберн", "name_en"=>"Blackburn", "lang"=>"en", "lat"=>39.104915, "lon"=>-93.486146),
"/US/US-MO/blackwater-11431" => array("type"=>"city", "pop"=>198, "name_ru"=>"Блэкуотер", "name_en"=>"Blackwater", "lang"=>"en", "lat"=>38.979382, "lon"=>-92.990677),
"/US/US-MO/blairstown-11432" => array("type"=>"city", "pop"=>141, "name_ru"=>"Блэрстаун", "name_en"=>"Blairstown", "lang"=>"en", "lat"=>38.557391, "lon"=>-93.957430),
"/US/US-MO/bland-11433" => array("type"=>"city", "pop"=>558, "name_ru"=>"Блэнд", "name_en"=>"Bland", "lang"=>"en", "lat"=>38.300579, "lon"=>-91.632546),
"/US/US-MO/blodgett-11434" => array("type"=>"town", "pop"=>266, "name_ru"=>"Блодгетт", "name_en"=>"Blodgett", "lang"=>"en", "lat"=>37.004252, "lon"=>-89.526157),
"/US/US-MO/bloomfield-11435" => array("type"=>"city", "pop"=>1869, "name_ru"=>"Блумфилд", "name_en"=>"Bloomfield", "lang"=>"en", "lat"=>36.888637, "lon"=>-89.931594),
"/US/US-MO/bloomsdale-11436" => array("type"=>"city", "pop"=>422, "name_ru"=>"Блумсдейл", "name_en"=>"Bloomsdale", "lang"=>"en", "lat"=>38.011423, "lon"=>-90.219273),
"/US/US-MO/blue-eye-11437" => array("type"=>"town", "pop"=>97, "name_ru"=>"Блу Ай", "name_en"=>"Blue Eye", "lang"=>"en", "lat"=>36.499860, "lon"=>-93.396414),
"/US/US-MO/blue-springs-11438" => array("type"=>"city", "pop"=>55698, "name_ru"=>"Блу Спрингс", "name_en"=>"Blue Springs", "lang"=>"en", "lat"=>39.022000, "lon"=>-94.273649),
"/US/US-MO/blythedale-11439" => array("type"=>"village", "pop"=>235, "name_ru"=>"Блитедейл", "name_en"=>"Blythedale", "lang"=>"en", "lat"=>40.475203, "lon"=>-93.927170),
"/US/US-MO/bogard-11440" => array("type"=>"city", "pop"=>223, "name_ru"=>"Богард", "name_en"=>"Bogard", "lang"=>"en", "lat"=>39.457949, "lon"=>-93.524091),
"/US/US-MO/bolckow-11441" => array("type"=>"city", "pop"=>230, "name_ru"=>"Болкоу", "name_en"=>"Bolckow", "lang"=>"en", "lat"=>40.115560, "lon"=>-94.820554),
"/US/US-MO/bolivar-11442" => array("type"=>"city", "pop"=>11013, "name_ru"=>"Боливар", "name_en"=>"Bolivar", "lang"=>"en", "lat"=>37.609635, "lon"=>-93.414272),
"/US/US-MO/bonne-terre-11443" => array("type"=>"city", "pop"=>6854, "name_ru"=>"Бонн Терр", "name_en"=>"Bonne Terre", "lang"=>"en", "lat"=>37.922655, "lon"=>-90.549815),
"/US/US-MO/boonville-11444" => array("type"=>"city", "pop"=>8770, "name_ru"=>"Бунвилл", "name_en"=>"Boonville", "lang"=>"en", "lat"=>38.964213, "lon"=>-92.747821),
"/US/US-MO/bosworth-11445" => array("type"=>"city", "pop"=>366, "name_ru"=>"Босуорт", "name_en"=>"Bosworth", "lang"=>"en", "lat"=>39.469751, "lon"=>-93.335575),
"/US/US-MO/bourbon-11446" => array("type"=>"city", "pop"=>1528, "name_ru"=>"Бурбон", "name_en"=>"Bourbon", "lang"=>"en", "lat"=>38.151592, "lon"=>-91.247242),
"/US/US-MO/bowling-green-11447" => array("type"=>"city", "pop"=>5285, "name_ru"=>"Боулинг Грин", "name_en"=>"Bowling Green", "lang"=>"en", "lat"=>39.338957, "lon"=>-91.195565),
"/US/US-MO/bragg-city-11448" => array("type"=>"town", "pop"=>178, "name_ru"=>"Брагг Сити", "name_en"=>"Bragg City", "lang"=>"en", "lat"=>36.268424, "lon"=>-89.911504),
"/US/US-MO/brandsville-11449" => array("type"=>"city", "pop"=>179, "name_ru"=>"Брэндсвилл", "name_en"=>"Brandsville", "lang"=>"en", "lat"=>36.650450, "lon"=>-91.696670),
"/US/US-MO/branson-11450" => array("type"=>"city", "pop"=>7651, "name_ru"=>"Брэнсон", "name_en"=>"Branson", "lang"=>"en", "lat"=>36.645804, "lon"=>-93.245211),
"/US/US-MO/brashear-11451" => array("type"=>"city", "pop"=>274, "name_ru"=>"Брашир", "name_en"=>"Brashear", "lang"=>"en", "lat"=>40.147889, "lon"=>-92.378369),
"/US/US-MO/braymer-11452" => array("type"=>"city", "pop"=>946, "name_ru"=>"Бреймер", "name_en"=>"Braymer", "lang"=>"en", "lat"=>39.590174, "lon"=>-93.795640),
"/US/US-MO/breckenridge-11453" => array("type"=>"city", "pop"=>454, "name_ru"=>"Брекенридж", "name_en"=>"Breckenridge", "lang"=>"en", "lat"=>39.761461, "lon"=>-93.804841),
"/US/US-MO/breckenridge-hills-11454" => array("type"=>"village", "pop"=>4479, "name_ru"=>"Брекенридж Хиллс", "name_en"=>"Breckenridge Hills", "lang"=>"en", "lat"=>38.715450, "lon"=>-90.367477),
"/US/US-MO/brentwood-11455" => array("type"=>"city", "pop"=>7182, "name_ru"=>"Брентвуд", "name_en"=>"Brentwood", "lang"=>"en", "lat"=>38.619050, "lon"=>-90.348599),
"/US/US-MO/bridgeton-11456" => array("type"=>"city", "pop"=>15028, "name_ru"=>"Бриджтон", "name_en"=>"Bridgeton", "lang"=>"en", "lat"=>38.766300, "lon"=>-90.426548),
"/US/US-MO/brimson-11457" => array("type"=>"village", "pop"=>61, "name_ru"=>"Бримосн", "name_en"=>"Brimson", "lang"=>"en", "lat"=>40.144221, "lon"=>-93.737755),
"/US/US-MO/bronaugh-11458" => array("type"=>"city", "pop"=>241, "name_ru"=>"Брнаф", "name_en"=>"Bronaugh", "lang"=>"en", "lat"=>37.694485, "lon"=>-94.467867),
"/US/US-MO/brookfield-11459" => array("type"=>"city", "pop"=>4321, "name_ru"=>"Брукфилд", "name_en"=>"Brookfield", "lang"=>"en", "lat"=>39.785750, "lon"=>-93.076317),
"/US/US-MO/brookline-11460" => array("type"=>"village", "pop"=>283, "name_ru"=>"Бруклин", "name_en"=>"Brookline", "lang"=>"en", "lat"=>37.171017, "lon"=>-93.410150),
"/US/US-MO/brooklyn-heights-11461" => array("type"=>"village", "pop"=>134, "name_ru"=>"Бруклин Хайтс", "name_en"=>"Brooklyn Heights", "lang"=>"en", "lat"=>37.170391, "lon"=>-94.386138),
"/US/US-MO/browning-11462" => array("type"=>"city", "pop"=>293, "name_ru"=>"Браунинг", "name_en"=>"Browning", "lang"=>"en", "lat"=>40.034914, "lon"=>-93.160237),
"/US/US-MO/brownington-11463" => array("type"=>"town", "pop"=>120, "name_ru"=>"Браунингтон", "name_en"=>"Brownington", "lang"=>"en", "lat"=>38.245701, "lon"=>-93.722482),
"/US/US-MO/brumley-11464" => array("type"=>"town", "pop"=>108, "name_ru"=>"Брамли", "name_en"=>"Brumley", "lang"=>"en", "lat"=>38.087322, "lon"=>-92.486128),
"/US/US-MO/brunswick-11465" => array("type"=>"city", "pop"=>853, "name_ru"=>"Брунсвик", "name_en"=>"Brunswick", "lang"=>"en", "lat"=>39.424008, "lon"=>-93.126011),
"/US/US-MO/bucklin-11466" => array("type"=>"city", "pop"=>474, "name_ru"=>"Баклин", "name_en"=>"Bucklin", "lang"=>"en", "lat"=>39.782373, "lon"=>-92.889254),
"/US/US-MO/buckner-11467" => array("type"=>"city", "pop"=>2788, "name_ru"=>"Бакнер", "name_en"=>"Buckner", "lang"=>"en", "lat"=>39.134800, "lon"=>-94.197999),
"/US/US-MO/buffalo-11468" => array("type"=>"city", "pop"=>3303, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>37.644276, "lon"=>-93.094335),
"/US/US-MO/bunceton-11469" => array("type"=>"city", "pop"=>359, "name_ru"=>"Банстон", "name_en"=>"Bunceton", "lang"=>"en", "lat"=>38.789660, "lon"=>-92.798526),
"/US/US-MO/bunker-11470" => array("type"=>"city", "pop"=>430, "name_ru"=>"Банкер", "name_en"=>"Bunker", "lang"=>"en", "lat"=>37.457453, "lon"=>-91.211527),
"/US/US-MO/burgess-11471" => array("type"=>"town", "pop"=>70, "name_ru"=>"Бергесс", "name_en"=>"Burgess", "lang"=>"en", "lat"=>37.556165, "lon"=>-94.614851),
"/US/US-MO/burlington-junction-11472" => array("type"=>"city", "pop"=>614, "name_ru"=>"Берлингтон Джанкшен", "name_en"=>"Burlington Junction", "lang"=>"en", "lat"=>40.447238, "lon"=>-95.067409),
"/US/US-MO/butler-11473" => array("type"=>"city", "pop"=>4306, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>38.259321, "lon"=>-94.339395),
"/US/US-MO/butterfield-11474" => array("type"=>"village", "pop"=>419, "name_ru"=>"Баттерфилд", "name_en"=>"Butterfield", "lang"=>"en", "lat"=>36.747576, "lon"=>-93.904438),
"/US/US-MO/byrnes-mill-11475" => array("type"=>"city", "pop"=>2910, "name_ru"=>"Бернс Милл", "name_en"=>"Byrnes Mill", "lang"=>"en", "lat"=>38.439250, "lon"=>-90.573649),
"/US/US-MO/cabool-11476" => array("type"=>"city", "pop"=>2141, "name_ru"=>"Кабул", "name_en"=>"Cabool", "lang"=>"en", "lat"=>37.126832, "lon"=>-92.103919),
"/US/US-MO/cainsville-11477" => array("type"=>"city", "pop"=>371, "name_ru"=>"Кейнсвилл", "name_en"=>"Cainsville", "lang"=>"en", "lat"=>40.440073, "lon"=>-93.774466),
"/US/US-MO/cairo-11478" => array("type"=>"village", "pop"=>306, "name_ru"=>"Каир", "name_en"=>"Cairo", "lang"=>"en", "lat"=>39.511993, "lon"=>-92.440705),
"/US/US-MO/caledonia-11479" => array("type"=>"village", "pop"=>166, "name_ru"=>"Каледония", "name_en"=>"Caledonia", "lang"=>"en", "lat"=>37.763565, "lon"=>-90.770982),
"/US/US-MO/calhoun-11480" => array("type"=>"city", "pop"=>493, "name_ru"=>"Калхун", "name_en"=>"Calhoun", "lang"=>"en", "lat"=>38.467948, "lon"=>-93.624456),
"/US/US-MO/california-11481" => array("type"=>"city", "pop"=>4154, "name_ru"=>"Калифорния", "name_en"=>"California", "lang"=>"en", "lat"=>38.631573, "lon"=>-92.566782),
"/US/US-MO/callao-11482" => array("type"=>"city", "pop"=>277, "name_ru"=>"Каллао", "name_en"=>"Callao", "lang"=>"en", "lat"=>39.762198, "lon"=>-92.623218),
"/US/US-MO/calverton-park-11483" => array("type"=>"village", "pop"=>1271, "name_ru"=>"Калвертон Парк", "name_en"=>"Calverton Park", "lang"=>"en", "lat"=>38.766550, "lon"=>-90.308857),
"/US/US-MO/camden-11484" => array("type"=>"city", "pop"=>199, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>39.200208, "lon"=>-94.018869),
"/US/US-MO/camden-point-11485" => array("type"=>"city", "pop"=>539, "name_ru"=>"Камден Пойнт", "name_en"=>"Camden Point", "lang"=>"en", "lat"=>39.452757, "lon"=>-94.747121),
"/US/US-MO/camdenton-11486" => array("type"=>"city", "pop"=>3498, "name_ru"=>"Камдентон", "name_en"=>"Camdenton", "lang"=>"en", "lat"=>38.008856, "lon"=>-92.746300),
"/US/US-MO/cameron-11487" => array("type"=>"city", "pop"=>9279, "name_ru"=>"Кэмерон", "name_en"=>"Cameron", "lang"=>"en", "lat"=>39.742872, "lon"=>-94.236295),
"/US/US-MO/campbell-11488" => array("type"=>"city", "pop"=>1841, "name_ru"=>"Кэмпбелл", "name_en"=>"Campbell", "lang"=>"en", "lat"=>36.492104, "lon"=>-90.074335),
"/US/US-MO/canalou-11489" => array("type"=>"city", "pop"=>299, "name_ru"=>"Каналоу", "name_en"=>"Canalou", "lang"=>"en", "lat"=>36.754339, "lon"=>-89.687494),
"/US/US-MO/canton-11490" => array("type"=>"city", "pop"=>2469, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>40.129668, "lon"=>-91.523848),
"/US/US-MO/cape-girardeau-11491" => array("type"=>"city", "pop"=>37370, "name_ru"=>"Кейп Жирардо", "name_en"=>"Cape Girardeau", "lang"=>"en", "lat"=>37.308309, "lon"=>-89.550300),
"/US/US-MO/cardwell-11492" => array("type"=>"city", "pop"=>723, "name_ru"=>"Кардуэлл", "name_en"=>"Cardwell", "lang"=>"en", "lat"=>36.047239, "lon"=>-90.291162),
"/US/US-MO/carl-junction-11493" => array("type"=>"city", "pop"=>7327, "name_ru"=>"Карл Джанкшен", "name_en"=>"Carl Junction", "lang"=>"en", "lat"=>37.163731, "lon"=>-94.545703),
"/US/US-MO/carrollton-11494" => array("type"=>"city", "pop"=>3857, "name_ru"=>"Карролтон", "name_en"=>"Carrollton", "lang"=>"en", "lat"=>39.363101, "lon"=>-93.496345),
"/US/US-MO/carterville-11495" => array("type"=>"city", "pop"=>1972, "name_ru"=>"Картервилл", "name_en"=>"Carterville", "lang"=>"en", "lat"=>37.150850, "lon"=>-94.438374),
"/US/US-MO/carthage-11496" => array("type"=>"city", "pop"=>13916, "name_ru"=>"Картаж", "name_en"=>"Carthage", "lang"=>"en", "lat"=>37.167703, "lon"=>-94.313535),
"/US/US-MO/caruthersville-11497" => array("type"=>"city", "pop"=>6145, "name_ru"=>"Карутерсвилл", "name_en"=>"Caruthersville", "lang"=>"en", "lat"=>36.179139, "lon"=>-89.665059),
"/US/US-MO/carytown-11498" => array("type"=>"city", "pop"=>232, "name_ru"=>"Кэритаун", "name_en"=>"Carytown", "lang"=>"en", "lat"=>37.249087, "lon"=>-94.334260),
"/US/US-MO/cassville-11499" => array("type"=>"city", "pop"=>3309, "name_ru"=>"Кассвилл", "name_en"=>"Cassville", "lang"=>"en", "lat"=>36.678803, "lon"=>-93.867638),
"/US/US-MO/castle-point-11500" => array("type"=>"cdp", "pop"=>4975, "name_ru"=>"Касл Пойнт", "name_en"=>"Castle Point", "lang"=>"en", "lat"=>38.757950, "lon"=>-90.248074),
"/US/US-MO/catron-11501" => array("type"=>"town", "pop"=>56, "name_ru"=>"Катрон", "name_en"=>"Catron", "lang"=>"en", "lat"=>36.612375, "lon"=>-89.706254),
"/US/US-MO/cave-11502" => array("type"=>"town", "pop"=>9, "name_ru"=>"Кейв", "name_en"=>"Cave", "lang"=>"en", "lat"=>39.023759, "lon"=>-91.045204),
"/US/US-MO/cedar-hill-11503" => array("type"=>"cdp", "pop"=>1966, "name_ru"=>"Сидар Хилл", "name_en"=>"Cedar Hill", "lang"=>"en", "lat"=>38.356140, "lon"=>-90.641554),
"/US/US-MO/cedar-hill-lakes-11504" => array("type"=>"village", "pop"=>214, "name_ru"=>"Сидар Хилл Лейкс", "name_en"=>"Cedar Hill Lakes", "lang"=>"en", "lat"=>38.330479, "lon"=>-90.657282),
"/US/US-MO/center-11505" => array("type"=>"city", "pop"=>641, "name_ru"=>"Сентер", "name_en"=>"Center", "lang"=>"en", "lat"=>39.509847, "lon"=>-91.528510),
"/US/US-MO/centertown-11506" => array("type"=>"town", "pop"=>251, "name_ru"=>"Сентертаун", "name_en"=>"Centertown", "lang"=>"en", "lat"=>38.618972, "lon"=>-92.409253),
"/US/US-MO/centerview-11507" => array("type"=>"town", "pop"=>255, "name_ru"=>"Сентервью", "name_en"=>"Centerview", "lang"=>"en", "lat"=>38.745559, "lon"=>-93.845562),
"/US/US-MO/centerville-11508" => array("type"=>"city", "pop"=>169, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>37.436327, "lon"=>-90.960333),
"/US/US-MO/centralia-11509" => array("type"=>"city", "pop"=>3669, "name_ru"=>"Сентралия", "name_en"=>"Centralia", "lang"=>"en", "lat"=>39.211524, "lon"=>-92.131575),
"/US/US-MO/chaffee-11510" => array("type"=>"city", "pop"=>2950, "name_ru"=>"Чаффи", "name_en"=>"Chaffee", "lang"=>"en", "lat"=>37.180376, "lon"=>-89.661338),
"/US/US-MO/chain-o-lakes-11511" => array("type"=>"village", "pop"=>131, "name_ru"=>"Чейн-О-Лейкс", "name_en"=>"Chain-O-Lakes", "lang"=>"en", "lat"=>36.533312, "lon"=>-93.724782),
"/US/US-MO/chamois-11512" => array("type"=>"city", "pop"=>480, "name_ru"=>"Чамуа", "name_en"=>"Chamois", "lang"=>"en", "lat"=>38.677153, "lon"=>-91.770006),
"/US/US-MO/champ-11513" => array("type"=>"village", "pop"=>17, "name_ru"=>"Чэмп", "name_en"=>"Champ", "lang"=>"en", "lat"=>38.743733, "lon"=>-90.449379),
"/US/US-MO/charlack-11514" => array("type"=>"city", "pop"=>1345, "name_ru"=>"Чарлэк", "name_en"=>"Charlack", "lang"=>"en", "lat"=>38.703600, "lon"=>-90.342504),
"/US/US-MO/charleston-11515" => array("type"=>"city", "pop"=>5111, "name_ru"=>"Чарльстон", "name_en"=>"Charleston", "lang"=>"en", "lat"=>36.918935, "lon"=>-89.337906),
"/US/US-MO/chesterfield-11516" => array("type"=>"city", "pop"=>46064, "name_ru"=>"Честерфилд", "name_en"=>"Chesterfield", "lang"=>"en", "lat"=>38.654300, "lon"=>-90.581223),
"/US/US-MO/chilhowee-11517" => array("type"=>"town", "pop"=>337, "name_ru"=>"Чил хауи", "name_en"=>"Chilhowee", "lang"=>"en", "lat"=>38.588948, "lon"=>-93.856327),
"/US/US-MO/chillicothe-11518" => array("type"=>"city", "pop"=>8706, "name_ru"=>"Чилликот", "name_en"=>"Chillicothe", "lang"=>"en", "lat"=>39.791614, "lon"=>-93.554060),
"/US/US-MO/chula-11519" => array("type"=>"city", "pop"=>196, "name_ru"=>"Чула", "name_en"=>"Chula", "lang"=>"en", "lat"=>39.921628, "lon"=>-93.476345),
"/US/US-MO/clarence-11520" => array("type"=>"city", "pop"=>868, "name_ru"=>"Клэренс", "name_en"=>"Clarence", "lang"=>"en", "lat"=>39.743014, "lon"=>-92.260266),
"/US/US-MO/clark-11521" => array("type"=>"city", "pop"=>291, "name_ru"=>"Кларк", "name_en"=>"Clark", "lang"=>"en", "lat"=>39.280604, "lon"=>-92.341858),
"/US/US-MO/clarksburg-11522" => array("type"=>"city", "pop"=>385, "name_ru"=>"Кларксбург", "name_en"=>"Clarksburg", "lang"=>"en", "lat"=>38.661362, "lon"=>-92.666126),
"/US/US-MO/clarksdale-11523" => array("type"=>"city", "pop"=>331, "name_ru"=>"Кларксдейл", "name_en"=>"Clarksdale", "lang"=>"en", "lat"=>39.813455, "lon"=>-94.550245),
"/US/US-MO/clarkson-valley-11524" => array("type"=>"city", "pop"=>2569, "name_ru"=>"Кларксон Валли", "name_en"=>"Clarkson Valley", "lang"=>"en", "lat"=>38.625033, "lon"=>-90.594121),
"/US/US-MO/clarksville-11525" => array("type"=>"city", "pop"=>497, "name_ru"=>"Кларксвилл", "name_en"=>"Clarksville", "lang"=>"en", "lat"=>39.371793, "lon"=>-90.905639),
"/US/US-MO/clarkton-11526" => array("type"=>"city", "pop"=>1238, "name_ru"=>"Кларктон", "name_en"=>"Clarkton", "lang"=>"en", "lat"=>36.452361, "lon"=>-89.965251),
"/US/US-MO/claycomo-11527" => array("type"=>"village", "pop"=>1311, "name_ru"=>"Клейкомо", "name_en"=>"Claycomo", "lang"=>"en", "lat"=>39.199500, "lon"=>-94.476773),
"/US/US-MO/clayton-11528" => array("type"=>"city", "pop"=>16088, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>38.644300, "lon"=>-90.328568),
"/US/US-MO/clearmont-11529" => array("type"=>"city", "pop"=>185, "name_ru"=>"Клирмонт", "name_en"=>"Clearmont", "lang"=>"en", "lat"=>40.507442, "lon"=>-95.032879),
"/US/US-MO/cleveland-11530" => array("type"=>"town", "pop"=>679, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>38.677799, "lon"=>-94.597049),
"/US/US-MO/clever-11531" => array("type"=>"city", "pop"=>1634, "name_ru"=>"Клевер", "name_en"=>"Clever", "lang"=>"en", "lat"=>37.029800, "lon"=>-93.471124),
"/US/US-MO/cliff-village-11532" => array("type"=>"village", "pop"=>34, "name_ru"=>"Клифф Виллидж", "name_en"=>"Cliff Village", "lang"=>"en", "lat"=>37.025322, "lon"=>-94.516902),
"/US/US-MO/clifton-hill-11533" => array("type"=>"city", "pop"=>130, "name_ru"=>"Клифтон Хилл", "name_en"=>"Clifton Hill", "lang"=>"en", "lat"=>39.439230, "lon"=>-92.667052),
"/US/US-MO/climax-springs-11534" => array("type"=>"village", "pop"=>87, "name_ru"=>"Клаймакс Спрингс", "name_en"=>"Climax Springs", "lang"=>"en", "lat"=>38.101591, "lon"=>-93.051658),
"/US/US-MO/clinton-11535" => array("type"=>"city", "pop"=>9381, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>38.371292, "lon"=>-93.768081),
"/US/US-MO/clyde-11536" => array("type"=>"village", "pop"=>73, "name_ru"=>"Клайд", "name_en"=>"Clyde", "lang"=>"en", "lat"=>40.266216, "lon"=>-94.670153),
"/US/US-MO/cobalt-city-11537" => array("type"=>"village", "pop"=>254, "name_ru"=>"Кобальт Сити", "name_en"=>"Cobalt City", "lang"=>"en", "lat"=>37.544885, "lon"=>-90.286853),
"/US/US-MO/coffey-11538" => array("type"=>"town", "pop"=>139, "name_ru"=>"Коффи", "name_en"=>"Coffey", "lang"=>"en", "lat"=>40.105089, "lon"=>-94.006174),
"/US/US-MO/cole-camp-11539" => array("type"=>"city", "pop"=>1138, "name_ru"=>"Коул Кэмп", "name_en"=>"Cole Camp", "lang"=>"en", "lat"=>38.459479, "lon"=>-93.205468),
"/US/US-MO/collins-11540" => array("type"=>"village", "pop"=>173, "name_ru"=>"Коллинс", "name_en"=>"Collins", "lang"=>"en", "lat"=>37.890113, "lon"=>-93.620774),
"/US/US-MO/columbia-11541" => array("type"=>"city", "pop"=>100733, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>38.954100, "lon"=>-92.326696),
"/US/US-MO/commerce-11542" => array("type"=>"town", "pop"=>109, "name_ru"=>"Коммерс", "name_en"=>"Commerce", "lang"=>"en", "lat"=>37.158088, "lon"=>-89.446561),
"/US/US-MO/conception-junction-11543" => array("type"=>"town", "pop"=>197, "name_ru"=>"Консепшен Джанкшен", "name_en"=>"Conception Junction", "lang"=>"en", "lat"=>40.268281, "lon"=>-94.690723),
"/US/US-MO/concord-11544" => array("type"=>"cdp", "pop"=>19859, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>38.514333, "lon"=>-90.353549),
"/US/US-MO/concordia-11545" => array("type"=>"city", "pop"=>2384, "name_ru"=>"Конкордия", "name_en"=>"Concordia", "lang"=>"en", "lat"=>38.987551, "lon"=>-93.568405),
"/US/US-MO/conway-11546" => array("type"=>"city", "pop"=>780, "name_ru"=>"Конвэй", "name_en"=>"Conway", "lang"=>"en", "lat"=>37.500768, "lon"=>-92.824176),
"/US/US-MO/cool-valley-11547" => array("type"=>"city", "pop"=>1004, "name_ru"=>"Кул Валли", "name_en"=>"Cool Valley", "lang"=>"en", "lat"=>38.724767, "lon"=>-90.304757),
"/US/US-MO/cooter-11548" => array("type"=>"city", "pop"=>414, "name_ru"=>"Кутер", "name_en"=>"Cooter", "lang"=>"en", "lat"=>36.046971, "lon"=>-89.809713),
"/US/US-MO/corder-11549" => array("type"=>"city", "pop"=>414, "name_ru"=>"Кордер", "name_en"=>"Corder", "lang"=>"en", "lat"=>39.099483, "lon"=>-93.638557),
"/US/US-MO/corning-11550" => array("type"=>"town", "pop"=>19, "name_ru"=>"Корнинг", "name_en"=>"Corning", "lang"=>"en", "lat"=>40.248834, "lon"=>-95.454742),
"/US/US-MO/cosby-11551" => array("type"=>"town", "pop"=>144, "name_ru"=>"Косби", "name_en"=>"Cosby", "lang"=>"en", "lat"=>39.863528, "lon"=>-94.679499),
"/US/US-MO/cottleville-11552" => array("type"=>"town", "pop"=>3166, "name_ru"=>"Коттлвилл", "name_en"=>"Cottleville", "lang"=>"en", "lat"=>38.754050, "lon"=>-90.653751),
"/US/US-MO/country-club-11553" => array("type"=>"village", "pop"=>2443, "name_ru"=>"Кантри Клаб", "name_en"=>"Country Club", "lang"=>"en", "lat"=>39.827650, "lon"=>-94.826868),
"/US/US-MO/country-club-hills-11554" => array("type"=>"city", "pop"=>1280, "name_ru"=>"Кантри Клаб Хилс", "name_en"=>"Country Club Hills", "lang"=>"en", "lat"=>38.720700, "lon"=>-90.274759),
"/US/US-MO/country-life-acres-11555" => array("type"=>"village", "pop"=>79, "name_ru"=>"Кантри Лайф Эйкрес", "name_en"=>"Country Life Acres", "lang"=>"en", "lat"=>38.624750, "lon"=>-90.455932),
"/US/US-MO/cowgill-11556" => array("type"=>"city", "pop"=>253, "name_ru"=>"Коуджил", "name_en"=>"Cowgill", "lang"=>"en", "lat"=>39.560560, "lon"=>-93.925791),
"/US/US-MO/craig-11557" => array("type"=>"city", "pop"=>286, "name_ru"=>"Крейг", "name_en"=>"Craig", "lang"=>"en", "lat"=>40.192223, "lon"=>-95.373360),
"/US/US-MO/crane-11558" => array("type"=>"city", "pop"=>1375, "name_ru"=>"Крейн", "name_en"=>"Crane", "lang"=>"en", "lat"=>36.903146, "lon"=>-93.571210),
"/US/US-MO/creighton-11559" => array("type"=>"city", "pop"=>346, "name_ru"=>"Крейтон", "name_en"=>"Creighton", "lang"=>"en", "lat"=>38.496654, "lon"=>-94.071764),
"/US/US-MO/crestwood-11560" => array("type"=>"city", "pop"=>11408, "name_ru"=>"Крествуд", "name_en"=>"Crestwood", "lang"=>"en", "lat"=>38.555600, "lon"=>-90.381590),
"/US/US-MO/creve-coeur-11561" => array("type"=>"city", "pop"=>16868, "name_ru"=>"Крив Коер", "name_en"=>"Creve Coeur", "lang"=>"en", "lat"=>38.660467, "lon"=>-90.435470),
"/US/US-MO/crocker-11562" => array("type"=>"city", "pop"=>1000, "name_ru"=>"Крокер", "name_en"=>"Crocker", "lang"=>"en", "lat"=>37.949150, "lon"=>-92.267260),
"/US/US-MO/cross-timbers-11563" => array("type"=>"city", "pop"=>187, "name_ru"=>"Кросс Тимберс", "name_en"=>"Cross Timbers", "lang"=>"en", "lat"=>38.024202, "lon"=>-93.228771),
"/US/US-MO/crystal-city-11564" => array("type"=>"city", "pop"=>4538, "name_ru"=>"Кристал Сити", "name_en"=>"Crystal City", "lang"=>"en", "lat"=>38.225902, "lon"=>-90.383030),
"/US/US-MO/crystal-lake-park-11565" => array("type"=>"city", "pop"=>440, "name_ru"=>"Кристал Лейк Парк", "name_en"=>"Crystal Lake Park", "lang"=>"en", "lat"=>38.620500, "lon"=>-90.430892),
"/US/US-MO/crystal-lakes-11566" => array("type"=>"village", "pop"=>404, "name_ru"=>"Кристал Лейкс", "name_en"=>"Crystal Lakes", "lang"=>"en", "lat"=>39.358168, "lon"=>-94.189958),
"/US/US-MO/cuba-11567" => array("type"=>"city", "pop"=>3555, "name_ru"=>"Куба", "name_en"=>"Cuba", "lang"=>"en", "lat"=>38.062968, "lon"=>-91.399883),
"/US/US-MO/curryville-11568" => array("type"=>"city", "pop"=>246, "name_ru"=>"Карривилл", "name_en"=>"Curryville", "lang"=>"en", "lat"=>39.345921, "lon"=>-91.342545),
"/US/US-MO/dadeville-11569" => array("type"=>"village", "pop"=>212, "name_ru"=>"Дадевилл", "name_en"=>"Dadeville", "lang"=>"en", "lat"=>37.479813, "lon"=>-93.674760),
"/US/US-MO/dalton-11570" => array("type"=>"town", "pop"=>25, "name_ru"=>"Далтон", "name_en"=>"Dalton", "lang"=>"en", "lat"=>39.398152, "lon"=>-92.991999),
"/US/US-MO/dardenne-prairie-11571" => array("type"=>"town", "pop"=>8197, "name_ru"=>"Дарденн Прейри", "name_en"=>"Dardenne Prairie", "lang"=>"en", "lat"=>38.754327, "lon"=>-90.732060),
"/US/US-MO/darlington-11572" => array("type"=>"village", "pop"=>103, "name_ru"=>"Дарлингтон", "name_en"=>"Darlington", "lang"=>"en", "lat"=>40.198773, "lon"=>-94.400168),
"/US/US-MO/de-kalb-11576" => array("type"=>"town", "pop"=>272, "name_ru"=>"Де Калб", "name_en"=>"De Kalb", "lang"=>"en", "lat"=>39.587169, "lon"=>-94.923861),
"/US/US-MO/de-soto-11583" => array("type"=>"city", "pop"=>6477, "name_ru"=>"Де Сото", "name_en"=>"De Soto", "lang"=>"en", "lat"=>38.138991, "lon"=>-90.560464),
"/US/US-MO/de-witt-11585" => array("type"=>"city", "pop"=>115, "name_ru"=>"Де Витт", "name_en"=>"De Witt", "lang"=>"en", "lat"=>39.384916, "lon"=>-93.220147),
"/US/US-MO/dearborn-11573" => array("type"=>"city", "pop"=>535, "name_ru"=>"Дирборн", "name_en"=>"Dearborn", "lang"=>"en", "lat"=>39.524935, "lon"=>-94.772946),
"/US/US-MO/deepwater-11574" => array("type"=>"city", "pop"=>490, "name_ru"=>"Дипуотер", "name_en"=>"Deepwater", "lang"=>"en", "lat"=>38.259175, "lon"=>-93.773801),
"/US/US-MO/deerfield-11575" => array("type"=>"village", "pop"=>74, "name_ru"=>"Дирфилд", "name_en"=>"Deerfield", "lang"=>"en", "lat"=>37.838964, "lon"=>-94.506587),
"/US/US-MO/dellwood-11577" => array("type"=>"city", "pop"=>4901, "name_ru"=>"Деллвуд", "name_en"=>"Dellwood", "lang"=>"en", "lat"=>38.755200, "lon"=>-90.277679),
"/US/US-MO/delta-11578" => array("type"=>"city", "pop"=>540, "name_ru"=>"Дельта", "name_en"=>"Delta", "lang"=>"en", "lat"=>37.198159, "lon"=>-89.737025),
"/US/US-MO/dennis-acres-11579" => array("type"=>"village", "pop"=>74, "name_ru"=>"Деннис Эйкрес", "name_en"=>"Dennis Acres", "lang"=>"en", "lat"=>37.045805, "lon"=>-94.504096),
"/US/US-MO/denver-11580" => array("type"=>"village", "pop"=>35, "name_ru"=>"Денвер", "name_en"=>"Denver", "lang"=>"en", "lat"=>40.398994, "lon"=>-94.323713),
"/US/US-MO/des-arc-11581" => array("type"=>"village", "pop"=>186, "name_ru"=>"Де Арк", "name_en"=>"Des Arc", "lang"=>"en", "lat"=>37.284185, "lon"=>-90.634945),
"/US/US-MO/des-peres-11584" => array("type"=>"city", "pop"=>8603, "name_ru"=>"Дес Перес", "name_en"=>"Des Peres", "lang"=>"en", "lat"=>38.595970, "lon"=>-90.447180),
"/US/US-MO/desloge-11582" => array("type"=>"city", "pop"=>5190, "name_ru"=>"Деслодж", "name_en"=>"Desloge", "lang"=>"en", "lat"=>37.873771, "lon"=>-90.520677),
"/US/US-MO/dexter-11586" => array("type"=>"city", "pop"=>7661, "name_ru"=>"Декстер", "name_en"=>"Dexter", "lang"=>"en", "lat"=>36.790367, "lon"=>-89.960320),
"/US/US-MO/diamond-11587" => array("type"=>"town", "pop"=>895, "name_ru"=>"Даймонд", "name_en"=>"Diamond", "lang"=>"en", "lat"=>36.995483, "lon"=>-94.315068),
"/US/US-MO/diehlstadt-11588" => array("type"=>"town", "pop"=>164, "name_ru"=>"Дилстадт", "name_en"=>"Diehlstadt", "lang"=>"en", "lat"=>36.959371, "lon"=>-89.432372),
"/US/US-MO/diggins-11589" => array("type"=>"village", "pop"=>339, "name_ru"=>"Диггинс", "name_en"=>"Diggins", "lang"=>"en", "lat"=>37.174035, "lon"=>-92.852274),
"/US/US-MO/dixon-11590" => array("type"=>"city", "pop"=>1536, "name_ru"=>"Диксон", "name_en"=>"Dixon", "lang"=>"en", "lat"=>37.995234, "lon"=>-92.095812),
"/US/US-MO/doniphan-11591" => array("type"=>"city", "pop"=>1887, "name_ru"=>"Донифан", "name_en"=>"Doniphan", "lang"=>"en", "lat"=>36.622906, "lon"=>-90.822043),
"/US/US-MO/doolittle-11592" => array("type"=>"city", "pop"=>653, "name_ru"=>"Дулиттл", "name_en"=>"Doolittle", "lang"=>"en", "lat"=>37.941488, "lon"=>-91.887797),
"/US/US-MO/dover-11593" => array("type"=>"town", "pop"=>105, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>39.194481, "lon"=>-93.689503),
"/US/US-MO/downing-11594" => array("type"=>"city", "pop"=>376, "name_ru"=>"Даунинг", "name_en"=>"Downing", "lang"=>"en", "lat"=>40.486443, "lon"=>-92.368503),
"/US/US-MO/drexel-11595" => array("type"=>"city", "pop"=>1097, "name_ru"=>"Дрексел", "name_en"=>"Drexel", "lang"=>"en", "lat"=>38.482810, "lon"=>-94.607367),
"/US/US-MO/dudley-11596" => array("type"=>"city", "pop"=>289, "name_ru"=>"Дадли", "name_en"=>"Dudley", "lang"=>"en", "lat"=>36.789139, "lon"=>-90.091422),
"/US/US-MO/duenweg-11597" => array("type"=>"city", "pop"=>1249, "name_ru"=>"Даенвег", "name_en"=>"Duenweg", "lang"=>"en", "lat"=>37.078976, "lon"=>-94.410510),
"/US/US-MO/duquesne-11598" => array("type"=>"village", "pop"=>1746, "name_ru"=>"Даквесн", "name_en"=>"Duquesne", "lang"=>"en", "lat"=>37.071250, "lon"=>-94.458667),
"/US/US-MO/eagleville-11599" => array("type"=>"town", "pop"=>324, "name_ru"=>"Иглвилл", "name_en"=>"Eagleville", "lang"=>"en", "lat"=>40.468716, "lon"=>-93.988123),
"/US/US-MO/east-lynne-11600" => array("type"=>"town", "pop"=>303, "name_ru"=>"Ист Линн", "name_en"=>"East Lynne", "lang"=>"en", "lat"=>38.668211, "lon"=>-94.229422),
"/US/US-MO/east-prairie-11602" => array("type"=>"city", "pop"=>3066, "name_ru"=>"Ист Прейри", "name_en"=>"East Prairie", "lang"=>"en", "lat"=>36.779048, "lon"=>-89.384117),
"/US/US-MO/easton-11601" => array("type"=>"city", "pop"=>269, "name_ru"=>"Истон", "name_en"=>"Easton", "lang"=>"en", "lat"=>39.721664, "lon"=>-94.639551),
"/US/US-MO/edgar-springs-11603" => array("type"=>"city", "pop"=>183, "name_ru"=>"Эдгар Спрингс", "name_en"=>"Edgar Springs", "lang"=>"en", "lat"=>37.701710, "lon"=>-91.865826),
"/US/US-MO/edgerton-11604" => array("type"=>"city", "pop"=>555, "name_ru"=>"Эджертон", "name_en"=>"Edgerton", "lang"=>"en", "lat"=>39.505013, "lon"=>-94.630648),
"/US/US-MO/edina-11605" => array("type"=>"city", "pop"=>1122, "name_ru"=>"Эдина", "name_en"=>"Edina", "lang"=>"en", "lat"=>40.167951, "lon"=>-92.173086),
"/US/US-MO/edmundson-11606" => array("type"=>"village", "pop"=>781, "name_ru"=>"Эдмундсон", "name_en"=>"Edmundson", "lang"=>"en", "lat"=>38.734300, "lon"=>-90.365871),
"/US/US-MO/el-dorado-springs-11608" => array("type"=>"city", "pop"=>3678, "name_ru"=>"Эль Дорадо Спрингс", "name_en"=>"El Dorado Springs", "lang"=>"en", "lat"=>37.869704, "lon"=>-94.019617),
"/US/US-MO/eldon-11607" => array("type"=>"city", "pop"=>4986, "name_ru"=>"Элдон", "name_en"=>"Eldon", "lang"=>"en", "lat"=>38.350906, "lon"=>-92.576408),
"/US/US-MO/ellington-11609" => array("type"=>"city", "pop"=>984, "name_ru"=>"Эллингтон", "name_en"=>"Ellington", "lang"=>"en", "lat"=>37.235416, "lon"=>-90.971978),
"/US/US-MO/ellisville-11610" => array("type"=>"city", "pop"=>9212, "name_ru"=>"Эллисвилл", "name_en"=>"Ellisville", "lang"=>"en", "lat"=>38.589150, "lon"=>-90.586002),
"/US/US-MO/ellsinore-11611" => array("type"=>"town", "pop"=>362, "name_ru"=>"Эллсинор", "name_en"=>"Ellsinore", "lang"=>"en", "lat"=>36.933636, "lon"=>-90.746047),
"/US/US-MO/elmer-11612" => array("type"=>"city", "pop"=>98, "name_ru"=>"Элмер", "name_en"=>"Elmer", "lang"=>"en", "lat"=>39.957645, "lon"=>-92.649811),
"/US/US-MO/elmira-11613" => array("type"=>"village", "pop"=>63, "name_ru"=>"Элмира", "name_en"=>"Elmira", "lang"=>"en", "lat"=>39.508349, "lon"=>-94.153051),
"/US/US-MO/elmo-11614" => array("type"=>"city", "pop"=>161, "name_ru"=>"Элмо", "name_en"=>"Elmo", "lang"=>"en", "lat"=>40.518780, "lon"=>-95.117152),
"/US/US-MO/elsberry-11615" => array("type"=>"city", "pop"=>2634, "name_ru"=>"Элсберри", "name_en"=>"Elsberry", "lang"=>"en", "lat"=>39.170039, "lon"=>-90.789891),
"/US/US-MO/elvins-11616" => array("type"=>"city", "pop"=>1391, "name_ru"=>"Элвинс", "name_en"=>"Elvins", "lang"=>"en", "lat"=>37.835291, "lon"=>-90.534998),
"/US/US-MO/emerald-beach-11617" => array("type"=>"village", "pop"=>284, "name_ru"=>"Эмеральд Бич", "name_en"=>"Emerald Beach", "lang"=>"en", "lat"=>36.574136, "lon"=>-93.670752),
"/US/US-MO/eminence-11618" => array("type"=>"city", "pop"=>554, "name_ru"=>"Эминенс", "name_en"=>"Eminence", "lang"=>"en", "lat"=>37.146271, "lon"=>-91.358528),
"/US/US-MO/emma-11619" => array("type"=>"city", "pop"=>233, "name_ru"=>"Эмма", "name_en"=>"Emma", "lang"=>"en", "lat"=>38.976470, "lon"=>-93.495505),
"/US/US-MO/eolia-11620" => array("type"=>"village", "pop"=>442, "name_ru"=>"Эолия", "name_en"=>"Eolia", "lang"=>"en", "lat"=>39.239865, "lon"=>-91.011154),
"/US/US-MO/essex-11621" => array("type"=>"city", "pop"=>521, "name_ru"=>"Эссекс", "name_en"=>"Essex", "lang"=>"en", "lat"=>36.811987, "lon"=>-89.862427),
"/US/US-MO/esther-11622" => array("type"=>"city", "pop"=>1071, "name_ru"=>"Эстер", "name_en"=>"Esther", "lang"=>"en", "lat"=>37.852064, "lon"=>-90.494665),
"/US/US-MO/ethel-11623" => array("type"=>"town", "pop"=>100, "name_ru"=>"Этел", "name_en"=>"Ethel", "lang"=>"en", "lat"=>39.893528, "lon"=>-92.740446),
"/US/US-MO/eugene-11624" => array("type"=>"town", "pop"=>179, "name_ru"=>"Юджин", "name_en"=>"Eugene", "lang"=>"en", "lat"=>38.353727, "lon"=>-92.403408),
"/US/US-MO/eureka-11625" => array("type"=>"city", "pop"=>9316, "name_ru"=>"Юрика", "name_en"=>"Eureka", "lang"=>"en", "lat"=>38.502900, "lon"=>-90.642680),
"/US/US-MO/everton-11626" => array("type"=>"city", "pop"=>305, "name_ru"=>"Эвертон", "name_en"=>"Everton", "lang"=>"en", "lat"=>37.343527, "lon"=>-93.702406),
"/US/US-MO/ewing-11627" => array("type"=>"city", "pop"=>441, "name_ru"=>"Юинг", "name_en"=>"Ewing", "lang"=>"en", "lat"=>40.008433, "lon"=>-91.714362),
"/US/US-MO/excelsior-estates-11628" => array("type"=>"village", "pop"=>256, "name_ru"=>"Экселсиор Истейтс", "name_en"=>"Excelsior Estates", "lang"=>"en", "lat"=>39.390189, "lon"=>-94.207618),
"/US/US-MO/excelsior-springs-11629" => array("type"=>"city", "pop"=>12002, "name_ru"=>"Экселсиор Спрингс", "name_en"=>"Excelsior Springs", "lang"=>"en", "lat"=>39.337896, "lon"=>-94.235459),
"/US/US-MO/exeter-11630" => array("type"=>"city", "pop"=>749, "name_ru"=>"Эксетер", "name_en"=>"Exeter", "lang"=>"en", "lat"=>36.670722, "lon"=>-93.939533),
"/US/US-MO/fair-grove-11632" => array("type"=>"city", "pop"=>1408, "name_ru"=>"Фэр Гров", "name_en"=>"Fair Grove", "lang"=>"en", "lat"=>37.382589, "lon"=>-93.151384),
"/US/US-MO/fair-play-11633" => array("type"=>"city", "pop"=>455, "name_ru"=>"Фэр Плей", "name_en"=>"Fair Play", "lang"=>"en", "lat"=>37.633887, "lon"=>-93.575429),
"/US/US-MO/fairfax-11631" => array("type"=>"city", "pop"=>603, "name_ru"=>"Фэрфакс", "name_en"=>"Fairfax", "lang"=>"en", "lat"=>40.339533, "lon"=>-95.391703),
"/US/US-MO/fairview-11634" => array("type"=>"town", "pop"=>415, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>36.816868, "lon"=>-94.084954),
"/US/US-MO/farber-11635" => array("type"=>"city", "pop"=>394, "name_ru"=>"Фарбер", "name_en"=>"Farber", "lang"=>"en", "lat"=>39.274071, "lon"=>-91.576402),
"/US/US-MO/farley-11636" => array("type"=>"village", "pop"=>256, "name_ru"=>"Фэрли", "name_en"=>"Farley", "lang"=>"en", "lat"=>39.282315, "lon"=>-94.828413),
"/US/US-MO/farmington-11637" => array("type"=>"city", "pop"=>16097, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>37.780003, "lon"=>-90.425863),
"/US/US-MO/fayette-11638" => array("type"=>"city", "pop"=>2695, "name_ru"=>"Файет", "name_en"=>"Fayette", "lang"=>"en", "lat"=>39.147508, "lon"=>-92.685566),
"/US/US-MO/fenton-11639" => array("type"=>"city", "pop"=>4285, "name_ru"=>"Фэнтон", "name_en"=>"Fenton", "lang"=>"en", "lat"=>38.536100, "lon"=>-90.451364),
"/US/US-MO/ferguson-11640" => array("type"=>"city", "pop"=>20964, "name_ru"=>"Фергюсон", "name_en"=>"Ferguson", "lang"=>"en", "lat"=>38.748800, "lon"=>-90.295774),
"/US/US-MO/ferrelview-11641" => array("type"=>"village", "pop"=>578, "name_ru"=>"Феррелвью", "name_en"=>"Ferrelview", "lang"=>"en", "lat"=>39.313659, "lon"=>-94.665302),
"/US/US-MO/festus-11642" => array("type"=>"city", "pop"=>11246, "name_ru"=>"Фестус", "name_en"=>"Festus", "lang"=>"en", "lat"=>38.222847, "lon"=>-90.405399),
"/US/US-MO/fidelity-11643" => array("type"=>"town", "pop"=>271, "name_ru"=>"Фиделити", "name_en"=>"Fidelity", "lang"=>"en", "lat"=>37.082006, "lon"=>-94.309470),
"/US/US-MO/fillmore-11644" => array("type"=>"city", "pop"=>213, "name_ru"=>"Фильмор", "name_en"=>"Fillmore", "lang"=>"en", "lat"=>40.025484, "lon"=>-94.972960),
"/US/US-MO/fisk-11645" => array("type"=>"city", "pop"=>368, "name_ru"=>"Фиск", "name_en"=>"Fisk", "lang"=>"en", "lat"=>36.782222, "lon"=>-90.207411),
"/US/US-MO/flat-river-11646" => array("type"=>"city", "pop"=>4823, "name_ru"=>"Флэт Ривер", "name_en"=>"Flat River", "lang"=>"en", "lat"=>37.853497, "lon"=>-90.516660),
"/US/US-MO/fleming-11647" => array("type"=>"city", "pop"=>117, "name_ru"=>"Флеминг", "name_en"=>"Fleming", "lang"=>"en", "lat"=>39.192885, "lon"=>-94.051221),
"/US/US-MO/flemington-11648" => array("type"=>"village", "pop"=>134, "name_ru"=>"Флемингтон", "name_en"=>"Flemington", "lang"=>"en", "lat"=>37.803680, "lon"=>-93.499921),
"/US/US-MO/flint-hill-11649" => array("type"=>"village", "pop"=>452, "name_ru"=>"Флинт Хилл", "name_en"=>"Flint Hill", "lang"=>"en", "lat"=>38.854800, "lon"=>-90.855363),
"/US/US-MO/flordell-hills-11650" => array("type"=>"city", "pop"=>861, "name_ru"=>"Флорделл Хиллс", "name_en"=>"Flordell Hills", "lang"=>"en", "lat"=>38.717567, "lon"=>-90.265271),
"/US/US-MO/florida-11651" => array("type"=>"village", "pop"=>2, "name_ru"=>"Флорида", "name_en"=>"Florida", "lang"=>"en", "lat"=>39.493264, "lon"=>-91.790119),
"/US/US-MO/florissant-11652" => array("type"=>"city", "pop"=>50561, "name_ru"=>"Флориссант", "name_en"=>"Florissant", "lang"=>"en", "lat"=>38.797350, "lon"=>-90.324303),
"/US/US-MO/foley-11653" => array("type"=>"city", "pop"=>211, "name_ru"=>"Фоли", "name_en"=>"Foley", "lang"=>"en", "lat"=>39.045677, "lon"=>-90.741341),
"/US/US-MO/ford-city-11654" => array("type"=>"town", "pop"=>29, "name_ru"=>"Форд Сити", "name_en"=>"Ford City", "lang"=>"en", "lat"=>40.112244, "lon"=>-94.464789),
"/US/US-MO/fordland-11655" => array("type"=>"city", "pop"=>762, "name_ru"=>"Фордленд", "name_en"=>"Fordland", "lang"=>"en", "lat"=>37.156930, "lon"=>-92.942730),
"/US/US-MO/forest-city-11656" => array("type"=>"city", "pop"=>313, "name_ru"=>"Форест Сити", "name_en"=>"Forest City", "lang"=>"en", "lat"=>39.982689, "lon"=>-95.187902),
"/US/US-MO/foristell-11657" => array("type"=>"city", "pop"=>331, "name_ru"=>"Фористелл", "name_en"=>"Foristell", "lang"=>"en", "lat"=>38.818546, "lon"=>-90.966381),
"/US/US-MO/forsyth-11658" => array("type"=>"city", "pop"=>1725, "name_ru"=>"Форсайт", "name_en"=>"Forsyth", "lang"=>"en", "lat"=>36.687271, "lon"=>-93.108177),
"/US/US-MO/fort-leonard-wood-11660" => array("type"=>"cdp", "pop"=>15863, "name_ru"=>"Форт Леонард Вуд", "name_en"=>"Fort Leonard Wood", "lang"=>"en", "lat"=>37.705209, "lon"=>-92.157961),
"/US/US-MO/fortescue-11659" => array("type"=>"town", "pop"=>47, "name_ru"=>"Фортескью", "name_en"=>"Fortescue", "lang"=>"en", "lat"=>40.051972, "lon"=>-95.317475),
"/US/US-MO/foster-11661" => array("type"=>"town", "pop"=>133, "name_ru"=>"Фостер", "name_en"=>"Foster", "lang"=>"en", "lat"=>38.166701, "lon"=>-94.507984),
"/US/US-MO/frankford-11662" => array("type"=>"city", "pop"=>355, "name_ru"=>"Фрэнкфорд", "name_en"=>"Frankford", "lang"=>"en", "lat"=>39.493211, "lon"=>-91.320684),
"/US/US-MO/franklin-11663" => array("type"=>"city", "pop"=>111, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>39.011334, "lon"=>-92.754720),
"/US/US-MO/fredericktown-11664" => array("type"=>"city", "pop"=>4151, "name_ru"=>"Фредериктаун", "name_en"=>"Fredericktown", "lang"=>"en", "lat"=>37.564601, "lon"=>-90.296455),
"/US/US-MO/freeburg-11665" => array("type"=>"village", "pop"=>438, "name_ru"=>"Фрибург", "name_en"=>"Freeburg", "lang"=>"en", "lat"=>38.315468, "lon"=>-91.922473),
"/US/US-MO/freeman-11666" => array("type"=>"city", "pop"=>599, "name_ru"=>"Фримэн", "name_en"=>"Freeman", "lang"=>"en", "lat"=>38.618632, "lon"=>-94.505765),
"/US/US-MO/freistatt-11667" => array("type"=>"town", "pop"=>194, "name_ru"=>"Фрайстадт", "name_en"=>"Freistatt", "lang"=>"en", "lat"=>37.021264, "lon"=>-93.896242),
"/US/US-MO/fremont-hills-11668" => array("type"=>"city", "pop"=>612, "name_ru"=>"Фремонт Хилс", "name_en"=>"Fremont Hills", "lang"=>"en", "lat"=>37.065766, "lon"=>-93.252635),
"/US/US-MO/frohna-11669" => array("type"=>"city", "pop"=>202, "name_ru"=>"Фрона", "name_en"=>"Frohna", "lang"=>"en", "lat"=>37.636500, "lon"=>-89.620472),
"/US/US-MO/frontenac-11670" => array("type"=>"city", "pop"=>3546, "name_ru"=>"Фронтенак", "name_en"=>"Frontenac", "lang"=>"en", "lat"=>38.630800, "lon"=>-90.415977),
"/US/US-MO/fulton-11671" => array("type"=>"city", "pop"=>12707, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>38.853137, "lon"=>-91.948843),
"/US/US-MO/gainesville-11672" => array("type"=>"city", "pop"=>594, "name_ru"=>"Гейнсвилл", "name_en"=>"Gainesville", "lang"=>"en", "lat"=>36.605686, "lon"=>-92.423751),
"/US/US-MO/galena-11673" => array("type"=>"city", "pop"=>500, "name_ru"=>"Галена", "name_en"=>"Galena", "lang"=>"en", "lat"=>36.804588, "lon"=>-93.470131),
"/US/US-MO/gallatin-11674" => array("type"=>"city", "pop"=>1730, "name_ru"=>"Галлатин", "name_en"=>"Gallatin", "lang"=>"en", "lat"=>39.910293, "lon"=>-93.963803),
"/US/US-MO/galt-11675" => array("type"=>"city", "pop"=>267, "name_ru"=>"Галт", "name_en"=>"Galt", "lang"=>"en", "lat"=>40.127195, "lon"=>-93.387588),
"/US/US-MO/garden-city-11676" => array("type"=>"city", "pop"=>1652, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>38.560823, "lon"=>-94.188368),
"/US/US-MO/gasconade-11677" => array("type"=>"city", "pop"=>255, "name_ru"=>"Гаскондейл", "name_en"=>"Gasconade", "lang"=>"en", "lat"=>38.669357, "lon"=>-91.560289),
"/US/US-MO/gentry-11678" => array("type"=>"village", "pop"=>92, "name_ru"=>"Джентри", "name_en"=>"Gentry", "lang"=>"en", "lat"=>40.333086, "lon"=>-94.423538),
"/US/US-MO/gerald-11679" => array("type"=>"city", "pop"=>1248, "name_ru"=>"Джеральд", "name_en"=>"Gerald", "lang"=>"en", "lat"=>38.399388, "lon"=>-91.331133),
"/US/US-MO/gerster-11680" => array("type"=>"town", "pop"=>33, "name_ru"=>"Джерстер", "name_en"=>"Gerster", "lang"=>"en", "lat"=>37.954033, "lon"=>-93.576508),
"/US/US-MO/gibbs-11681" => array("type"=>"town", "pop"=>99, "name_ru"=>"Гиббс", "name_en"=>"Gibbs", "lang"=>"en", "lat"=>40.097962, "lon"=>-92.416295),
"/US/US-MO/gideon-11682" => array("type"=>"city", "pop"=>960, "name_ru"=>"Гидеон", "name_en"=>"Gideon", "lang"=>"en", "lat"=>36.450819, "lon"=>-89.910639),
"/US/US-MO/gilliam-11683" => array("type"=>"town", "pop"=>218, "name_ru"=>"Гиллиам", "name_en"=>"Gilliam", "lang"=>"en", "lat"=>39.232767, "lon"=>-93.003659),
"/US/US-MO/gilman-city-11684" => array("type"=>"city", "pop"=>381, "name_ru"=>"Гилман Сити", "name_en"=>"Gilman City", "lang"=>"en", "lat"=>40.138325, "lon"=>-93.872372),
"/US/US-MO/gladstone-11685" => array("type"=>"city", "pop"=>28243, "name_ru"=>"Глэдстоун", "name_en"=>"Gladstone", "lang"=>"en", "lat"=>39.212800, "lon"=>-94.558908),
"/US/US-MO/glasgow-11686" => array("type"=>"city", "pop"=>1199, "name_ru"=>"Гласго", "name_en"=>"Glasgow", "lang"=>"en", "lat"=>39.226985, "lon"=>-92.837922),
"/US/US-MO/glasgow-village-11687" => array("type"=>"cdp", "pop"=>5199, "name_ru"=>"Гласгоу Филлидж", "name_en"=>"Glasgow Village", "lang"=>"en", "lat"=>38.758213, "lon"=>-90.196928),
"/US/US-MO/glen-echo-park-11691" => array("type"=>"village", "pop"=>158, "name_ru"=>"Глен Эко Парк", "name_en"=>"Glen Echo Park", "lang"=>"en", "lat"=>38.700989, "lon"=>-90.295484),
"/US/US-MO/glenaire-11688" => array("type"=>"village", "pop"=>603, "name_ru"=>"Гленэр", "name_en"=>"Glenaire", "lang"=>"en", "lat"=>39.220033, "lon"=>-94.449986),
"/US/US-MO/glenallen-11689" => array("type"=>"town", "pop"=>96, "name_ru"=>"Гленаллен", "name_en"=>"Glenallen", "lang"=>"en", "lat"=>37.316819, "lon"=>-90.028145),
"/US/US-MO/glendale-11690" => array("type"=>"city", "pop"=>5490, "name_ru"=>"Глендейл", "name_en"=>"Glendale", "lang"=>"en", "lat"=>38.593000, "lon"=>-90.381174),
"/US/US-MO/glenwood-11692" => array("type"=>"village", "pop"=>193, "name_ru"=>"Гленвуд", "name_en"=>"Glenwood", "lang"=>"en", "lat"=>40.523581, "lon"=>-92.576343),
"/US/US-MO/golden-city-11693" => array("type"=>"city", "pop"=>877, "name_ru"=>"Голден Сити", "name_en"=>"Golden City", "lang"=>"en", "lat"=>37.393558, "lon"=>-94.093212),
"/US/US-MO/goodman-11694" => array("type"=>"town", "pop"=>1255, "name_ru"=>"Гудман", "name_en"=>"Goodman", "lang"=>"en", "lat"=>36.737872, "lon"=>-94.409275),
"/US/US-MO/gordonville-11695" => array("type"=>"town", "pop"=>444, "name_ru"=>"Гордонвилл", "name_en"=>"Gordonville", "lang"=>"en", "lat"=>37.310375, "lon"=>-89.673041),
"/US/US-MO/gower-11696" => array("type"=>"city", "pop"=>1431, "name_ru"=>"Гауэр", "name_en"=>"Gower", "lang"=>"en", "lat"=>39.612325, "lon"=>-94.594698),
"/US/US-MO/graham-11697" => array("type"=>"town", "pop"=>186, "name_ru"=>"Грэхэм", "name_en"=>"Graham", "lang"=>"en", "lat"=>40.201493, "lon"=>-95.039767),
"/US/US-MO/grain-valley-11698" => array("type"=>"city", "pop"=>10352, "name_ru"=>"Грейн Валли", "name_en"=>"Grain Valley", "lang"=>"en", "lat"=>39.007218, "lon"=>-94.206824),
"/US/US-MO/granby-11699" => array("type"=>"city", "pop"=>2231, "name_ru"=>"Грэнби", "name_en"=>"Granby", "lang"=>"en", "lat"=>36.916970, "lon"=>-94.260071),
"/US/US-MO/grand-pass-11701" => array("type"=>"town", "pop"=>50, "name_ru"=>"Гранд Пасс", "name_en"=>"Grand Pass", "lang"=>"en", "lat"=>39.205239, "lon"=>-93.443171),
"/US/US-MO/grandin-11700" => array("type"=>"city", "pop"=>235, "name_ru"=>"Грандин", "name_en"=>"Grandin", "lang"=>"en", "lat"=>36.829898, "lon"=>-90.821764),
"/US/US-MO/grandview-11702" => array("type"=>"city", "pop"=>24033, "name_ru"=>"Грандвью", "name_en"=>"Grandview", "lang"=>"en", "lat"=>38.881350, "lon"=>-94.522500),
"/US/US-MO/granger-11703" => array("type"=>"town", "pop"=>43, "name_ru"=>"Грейнджер", "name_en"=>"Granger", "lang"=>"en", "lat"=>40.467509, "lon"=>-91.973104),
"/US/US-MO/grant-city-11704" => array("type"=>"town", "pop"=>776, "name_ru"=>"Грант Сити", "name_en"=>"Grant City", "lang"=>"en", "lat"=>40.485962, "lon"=>-94.413853),
"/US/US-MO/grantwood-village-11705" => array("type"=>"town", "pop"=>866, "name_ru"=>"Грантвуд Виллидж", "name_en"=>"Grantwood Village", "lang"=>"en", "lat"=>38.550383, "lon"=>-90.354337),
"/US/US-MO/gravois-mills-11706" => array("type"=>"village", "pop"=>228, "name_ru"=>"Грейвуа Миллс", "name_en"=>"Gravois Mills", "lang"=>"en", "lat"=>38.309598, "lon"=>-92.824520),
"/US/US-MO/gray-summit-11707" => array("type"=>"cdp", "pop"=>2505, "name_ru"=>"Грей Саммит", "name_en"=>"Gray Summit", "lang"=>"en", "lat"=>38.494850, "lon"=>-90.815180),
"/US/US-MO/green-city-11709" => array("type"=>"city", "pop"=>633, "name_ru"=>"Грин Сити", "name_en"=>"Green City", "lang"=>"en", "lat"=>40.265683, "lon"=>-92.957055),
"/US/US-MO/green-ridge-11712" => array("type"=>"town", "pop"=>468, "name_ru"=>"Грин Ридж", "name_en"=>"Green Ridge", "lang"=>"en", "lat"=>38.620187, "lon"=>-93.409798),
"/US/US-MO/greencastle-11708" => array("type"=>"city", "pop"=>286, "name_ru"=>"Гринкасл", "name_en"=>"Greencastle", "lang"=>"en", "lat"=>40.261539, "lon"=>-92.878342),
"/US/US-MO/greendale-11710" => array("type"=>"city", "pop"=>690, "name_ru"=>"Гриндейл", "name_en"=>"Greendale", "lang"=>"en", "lat"=>38.691467, "lon"=>-90.312228),
"/US/US-MO/greenfield-11711" => array("type"=>"city", "pop"=>1233, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>37.416325, "lon"=>-93.842657),
"/US/US-MO/greentop-11713" => array("type"=>"city", "pop"=>412, "name_ru"=>"Гринтоп", "name_en"=>"Greentop", "lang"=>"en", "lat"=>40.351100, "lon"=>-92.565875),
"/US/US-MO/greenville-11714" => array("type"=>"city", "pop"=>432, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>37.126879, "lon"=>-90.446001),
"/US/US-MO/greenwood-11715" => array("type"=>"city", "pop"=>4649, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>38.851150, "lon"=>-94.341999),
"/US/US-MO/guilford-11716" => array("type"=>"town", "pop"=>84, "name_ru"=>"Гилфорд", "name_en"=>"Guilford", "lang"=>"en", "lat"=>40.168940, "lon"=>-94.736085),
"/US/US-MO/gunn-city-11717" => array("type"=>"village", "pop"=>86, "name_ru"=>"Ганн Сити", "name_en"=>"Gunn City", "lang"=>"en", "lat"=>38.666213, "lon"=>-94.163882),
"/US/US-MO/hale-11718" => array("type"=>"city", "pop"=>452, "name_ru"=>"Хейл", "name_en"=>"Hale", "lang"=>"en", "lat"=>39.605186, "lon"=>-93.343239),
"/US/US-MO/halfway-11719" => array("type"=>"village", "pop"=>191, "name_ru"=>"Хафуэй", "name_en"=>"Halfway", "lang"=>"en", "lat"=>37.616048, "lon"=>-93.235598),
"/US/US-MO/hallsville-11720" => array("type"=>"city", "pop"=>958, "name_ru"=>"Холлсвилл", "name_en"=>"Hallsville", "lang"=>"en", "lat"=>39.115784, "lon"=>-92.219663),
"/US/US-MO/halltown-11721" => array("type"=>"town", "pop"=>200, "name_ru"=>"Холлтаун", "name_en"=>"Halltown", "lang"=>"en", "lat"=>37.194322, "lon"=>-93.628852),
"/US/US-MO/hamilton-11722" => array("type"=>"city", "pop"=>1784, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>39.743312, "lon"=>-94.002102),
"/US/US-MO/hanley-hills-11723" => array("type"=>"village", "pop"=>2026, "name_ru"=>"Хэнли Хиллс", "name_en"=>"Hanley Hills", "lang"=>"en", "lat"=>38.684867, "lon"=>-90.324360),
"/US/US-MO/hannibal-11724" => array("type"=>"city", "pop"=>17432, "name_ru"=>"Ханнибал", "name_en"=>"Hannibal", "lang"=>"en", "lat"=>39.707547, "lon"=>-91.383638),
"/US/US-MO/hardin-11725" => array("type"=>"city", "pop"=>556, "name_ru"=>"Хардин", "name_en"=>"Hardin", "lang"=>"en", "lat"=>39.267612, "lon"=>-93.830314),
"/US/US-MO/harris-11726" => array("type"=>"town", "pop"=>97, "name_ru"=>"Харрис", "name_en"=>"Harris", "lang"=>"en", "lat"=>40.306389, "lon"=>-93.350093),
"/US/US-MO/harrisburg-11727" => array("type"=>"town", "pop"=>182, "name_ru"=>"Харрисбург", "name_en"=>"Harrisburg", "lang"=>"en", "lat"=>39.142215, "lon"=>-92.459708),
"/US/US-MO/harrisonville-11728" => array("type"=>"city", "pop"=>9749, "name_ru"=>"Харрисонвилл", "name_en"=>"Harrisonville", "lang"=>"en", "lat"=>38.655411, "lon"=>-94.349244),
"/US/US-MO/hartsburg-11729" => array("type"=>"town", "pop"=>104, "name_ru"=>"Хартсбург", "name_en"=>"Hartsburg", "lang"=>"en", "lat"=>38.695153, "lon"=>-92.309845),
"/US/US-MO/hartville-11730" => array("type"=>"city", "pop"=>604, "name_ru"=>"Хартвилл", "name_en"=>"Hartville", "lang"=>"en", "lat"=>37.250120, "lon"=>-92.511750),
"/US/US-MO/hartwell-11731" => array("type"=>"village", "pop"=>14, "name_ru"=>"Хартвэлл", "name_en"=>"Hartwell", "lang"=>"en", "lat"=>38.434907, "lon"=>-93.933909),
"/US/US-MO/harwood-11732" => array("type"=>"town", "pop"=>88, "name_ru"=>"Харвуд", "name_en"=>"Harwood", "lang"=>"en", "lat"=>37.956830, "lon"=>-94.153741),
"/US/US-MO/hawk-point-11733" => array("type"=>"city", "pop"=>578, "name_ru"=>"Хок Пойнт", "name_en"=>"Hawk Point", "lang"=>"en", "lat"=>38.971163, "lon"=>-91.133741),
"/US/US-MO/hay-wood-city-11737" => array("type"=>"village", "pop"=>263, "name_ru"=>"Хэй-Вуд Сити", "name_en"=>"Hay-Wood City", "lang"=>"en", "lat"=>37.011785, "lon"=>-89.600124),
"/US/US-MO/hayti-11734" => array("type"=>"city", "pop"=>2959, "name_ru"=>"Хейти", "name_en"=>"Hayti", "lang"=>"en", "lat"=>36.232476, "lon"=>-89.748474),
"/US/US-MO/hayti-heights-11735" => array("type"=>"city", "pop"=>744, "name_ru"=>"Хейти Хайтс", "name_en"=>"Hayti Heights", "lang"=>"en", "lat"=>36.232223, "lon"=>-89.768513),
"/US/US-MO/hayward-11736" => array("type"=>"town", "pop"=>103, "name_ru"=>"Хейворд", "name_en"=>"Hayward", "lang"=>"en", "lat"=>36.396120, "lon"=>-89.666034),
"/US/US-MO/hazelwood-11738" => array("type"=>"city", "pop"=>25345, "name_ru"=>"Хейзелвуд", "name_en"=>"Hazelwood", "lang"=>"en", "lat"=>38.777650, "lon"=>-90.359629),
"/US/US-MO/henrietta-11739" => array("type"=>"city", "pop"=>432, "name_ru"=>"Хенриетта", "name_en"=>"Henrietta", "lang"=>"en", "lat"=>39.237353, "lon"=>-93.937951),
"/US/US-MO/herculaneum-11740" => array("type"=>"city", "pop"=>3582, "name_ru"=>"Геркуланеум", "name_en"=>"Herculaneum", "lang"=>"en", "lat"=>38.259193, "lon"=>-90.387670),
"/US/US-MO/hermann-11741" => array("type"=>"city", "pop"=>2717, "name_ru"=>"Херманн", "name_en"=>"Hermann", "lang"=>"en", "lat"=>38.699075, "lon"=>-91.435026),
"/US/US-MO/hermitage-11742" => array("type"=>"city", "pop"=>501, "name_ru"=>"Хермитедж", "name_en"=>"Hermitage", "lang"=>"en", "lat"=>37.944980, "lon"=>-93.323249),
"/US/US-MO/higbee-11743" => array("type"=>"city", "pop"=>655, "name_ru"=>"Хигби", "name_en"=>"Higbee", "lang"=>"en", "lat"=>39.306395, "lon"=>-92.512153),
"/US/US-MO/higginsville-11744" => array("type"=>"city", "pop"=>4571, "name_ru"=>"Хиггинсвилл", "name_en"=>"Higginsville", "lang"=>"en", "lat"=>39.064764, "lon"=>-93.726546),
"/US/US-MO/high-hill-11745" => array("type"=>"city", "pop"=>212, "name_ru"=>"Хай Хилл", "name_en"=>"High Hill", "lang"=>"en", "lat"=>38.875396, "lon"=>-91.375599),
"/US/US-MO/high-ridge-11746" => array("type"=>"cdp", "pop"=>4423, "name_ru"=>"Хай Ридж", "name_en"=>"High Ridge", "lang"=>"en", "lat"=>38.462850, "lon"=>-90.534149),
"/US/US-MO/hillsboro-11747" => array("type"=>"city", "pop"=>2062, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>38.232349, "lon"=>-90.566767),
"/US/US-MO/hillsdale-11748" => array("type"=>"village", "pop"=>1408, "name_ru"=>"Хиллсдейл", "name_en"=>"Hillsdale", "lang"=>"en", "lat"=>38.685100, "lon"=>-90.288486),
"/US/US-MO/hoberg-11749" => array("type"=>"village", "pop"=>63, "name_ru"=>"Хоберг", "name_en"=>"Hoberg", "lang"=>"en", "lat"=>37.068349, "lon"=>-93.849270),
"/US/US-MO/holcomb-11750" => array("type"=>"city", "pop"=>671, "name_ru"=>"Холкомб", "name_en"=>"Holcomb", "lang"=>"en", "lat"=>36.400223, "lon"=>-90.023053),
"/US/US-MO/holden-11751" => array("type"=>"city", "pop"=>2543, "name_ru"=>"Холден", "name_en"=>"Holden", "lang"=>"en", "lat"=>38.713437, "lon"=>-93.989118),
"/US/US-MO/holland-11752" => array("type"=>"town", "pop"=>232, "name_ru"=>"Холланд", "name_en"=>"Holland", "lang"=>"en", "lat"=>36.057329, "lon"=>-89.870382),
"/US/US-MO/holliday-11753" => array("type"=>"village", "pop"=>122, "name_ru"=>"Холлидей", "name_en"=>"Holliday", "lang"=>"en", "lat"=>39.493559, "lon"=>-92.131861),
"/US/US-MO/hollister-11754" => array("type"=>"city", "pop"=>4032, "name_ru"=>"Холлистер", "name_en"=>"Hollister", "lang"=>"en", "lat"=>36.619134, "lon"=>-93.220313),
"/US/US-MO/holt-11755" => array("type"=>"city", "pop"=>472, "name_ru"=>"Холт", "name_en"=>"Holt", "lang"=>"en", "lat"=>39.454240, "lon"=>-94.344142),
"/US/US-MO/holts-summit-11756" => array("type"=>"city", "pop"=>3735, "name_ru"=>"Холтс Саммит", "name_en"=>"Holts Summit", "lang"=>"en", "lat"=>38.644661, "lon"=>-92.115963),
"/US/US-MO/homestead-11757" => array("type"=>"village", "pop"=>169, "name_ru"=>"Хоумстид", "name_en"=>"Homestead", "lang"=>"en", "lat"=>39.359712, "lon"=>-94.201891),
"/US/US-MO/homestown-11758" => array("type"=>"city", "pop"=>169, "name_ru"=>"Хоумстаун", "name_en"=>"Homestown", "lang"=>"en", "lat"=>36.331702, "lon"=>-89.824601),
"/US/US-MO/hopkins-11759" => array("type"=>"city", "pop"=>561, "name_ru"=>"Хопкинс", "name_en"=>"Hopkins", "lang"=>"en", "lat"=>40.551113, "lon"=>-94.816187),
"/US/US-MO/horine-11760" => array("type"=>"cdp", "pop"=>1043, "name_ru"=>"Хорайн", "name_en"=>"Horine", "lang"=>"en", "lat"=>38.266086, "lon"=>-90.430169),
"/US/US-MO/hornersville-11761" => array("type"=>"city", "pop"=>656, "name_ru"=>"Хорнерсвилл", "name_en"=>"Hornersville", "lang"=>"en", "lat"=>36.040266, "lon"=>-90.114795),
"/US/US-MO/houston-11762" => array("type"=>"city", "pop"=>2038, "name_ru"=>"Хьюстон", "name_en"=>"Houston", "lang"=>"en", "lat"=>37.322645, "lon"=>-91.959635),
"/US/US-MO/houston-lake-11764" => array("type"=>"city", "pop"=>262, "name_ru"=>"Хьюстон Лейк", "name_en"=>"Houston Lake", "lang"=>"en", "lat"=>39.191717, "lon"=>-94.623504),
"/US/US-MO/houstonia-11763" => array("type"=>"city", "pop"=>290, "name_ru"=>"Хьюстония", "name_en"=>"Houstonia", "lang"=>"en", "lat"=>38.899431, "lon"=>-93.359803),
"/US/US-MO/howardville-11765" => array("type"=>"city", "pop"=>314, "name_ru"=>"Ховардвилл", "name_en"=>"Howardville", "lang"=>"en", "lat"=>36.568339, "lon"=>-89.597328),
"/US/US-MO/hughesville-11766" => array("type"=>"village", "pop"=>183, "name_ru"=>"Хьюсвилл", "name_en"=>"Hughesville", "lang"=>"en", "lat"=>38.837877, "lon"=>-93.295926),
"/US/US-MO/humansville-11767" => array("type"=>"city", "pop"=>1017, "name_ru"=>"Хьюмансвилл", "name_en"=>"Humansville", "lang"=>"en", "lat"=>37.795607, "lon"=>-93.575854),
"/US/US-MO/hume-11768" => array("type"=>"town", "pop"=>344, "name_ru"=>"Хум", "name_en"=>"Hume", "lang"=>"en", "lat"=>38.089534, "lon"=>-94.581555),
"/US/US-MO/humphreys-11769" => array("type"=>"town", "pop"=>150, "name_ru"=>"Хамфрис", "name_en"=>"Humphreys", "lang"=>"en", "lat"=>40.125100, "lon"=>-93.318416),
"/US/US-MO/hunnewell-11770" => array("type"=>"city", "pop"=>219, "name_ru"=>"Ханнуэлл", "name_en"=>"Hunnewell", "lang"=>"en", "lat"=>39.668263, "lon"=>-91.858853),
"/US/US-MO/huntleigh-11771" => array("type"=>"city", "pop"=>328, "name_ru"=>"Хантли", "name_en"=>"Huntleigh", "lang"=>"en", "lat"=>38.613577, "lon"=>-90.407613),
"/US/US-MO/huntsville-11772" => array("type"=>"city", "pop"=>1636, "name_ru"=>"Хантсвилл", "name_en"=>"Huntsville", "lang"=>"en", "lat"=>39.437361, "lon"=>-92.543633),
"/US/US-MO/hurdland-11773" => array("type"=>"city", "pop"=>223, "name_ru"=>"Хердленд", "name_en"=>"Hurdland", "lang"=>"en", "lat"=>40.150731, "lon"=>-92.302239),
"/US/US-MO/hurley-11774" => array("type"=>"city", "pop"=>150, "name_ru"=>"Херли", "name_en"=>"Hurley", "lang"=>"en", "lat"=>36.930857, "lon"=>-93.496655),
"/US/US-MO/iatan-11775" => array("type"=>"village", "pop"=>66, "name_ru"=>"Айатан", "name_en"=>"Iatan", "lang"=>"en", "lat"=>39.475169, "lon"=>-94.981626),
"/US/US-MO/iberia-11776" => array("type"=>"city", "pop"=>679, "name_ru"=>"Иберия", "name_en"=>"Iberia", "lang"=>"en", "lat"=>38.088818, "lon"=>-92.294448),
"/US/US-MO/imperial-11777" => array("type"=>"cdp", "pop"=>4156, "name_ru"=>"Империал", "name_en"=>"Imperial", "lang"=>"en", "lat"=>38.367733, "lon"=>-90.372219),
"/US/US-MO/independence-11778" => array("type"=>"city", "pop"=>110440, "name_ru"=>"Индепенденс", "name_en"=>"Independence", "lang"=>"en", "lat"=>39.089344, "lon"=>-94.353413),
"/US/US-MO/indian-point-11779" => array("type"=>"village", "pop"=>707, "name_ru"=>"Индиан Пойнт", "name_en"=>"Indian Point", "lang"=>"en", "lat"=>36.640735, "lon"=>-93.346944),
"/US/US-MO/ionia-11780" => array("type"=>"town", "pop"=>111, "name_ru"=>"Иония", "name_en"=>"Ionia", "lang"=>"en", "lat"=>38.503847, "lon"=>-93.322796),
"/US/US-MO/iron-gates-11782" => array("type"=>"village", "pop"=>309, "name_ru"=>"Айрон Гейтс", "name_en"=>"Iron Gates", "lang"=>"en", "lat"=>37.063500, "lon"=>-94.556949),
"/US/US-MO/iron-mountain-lake-11783" => array("type"=>"city", "pop"=>699, "name_ru"=>"Айрон Маунтин Лейк", "name_en"=>"Iron Mountain Lake", "lang"=>"en", "lat"=>37.685081, "lon"=>-90.614669),
"/US/US-MO/irondale-11781" => array("type"=>"city", "pop"=>455, "name_ru"=>"Айрондейл", "name_en"=>"Irondale", "lang"=>"en", "lat"=>37.834238, "lon"=>-90.670881),
"/US/US-MO/ironton-11784" => array("type"=>"city", "pop"=>1324, "name_ru"=>"Айронтон", "name_en"=>"Ironton", "lang"=>"en", "lat"=>37.596732, "lon"=>-90.635830),
"/US/US-MO/jackson-11785" => array("type"=>"city", "pop"=>13744, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>37.380320, "lon"=>-89.654461),
"/US/US-MO/jacksonville-11786" => array("type"=>"village", "pop"=>172, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>39.587481, "lon"=>-92.472844),
"/US/US-MO/jameson-11787" => array("type"=>"town", "pop"=>119, "name_ru"=>"Джеймсон", "name_en"=>"Jameson", "lang"=>"en", "lat"=>40.005320, "lon"=>-93.987576),
"/US/US-MO/jamesport-11788" => array("type"=>"city", "pop"=>501, "name_ru"=>"Джеймспорт", "name_en"=>"Jamesport", "lang"=>"en", "lat"=>39.974715, "lon"=>-93.802104),
"/US/US-MO/jamestown-11789" => array("type"=>"town", "pop"=>393, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>38.765723, "lon"=>-92.477771),
"/US/US-MO/jasper-11790" => array("type"=>"city", "pop"=>1067, "name_ru"=>"Джаспер", "name_en"=>"Jasper", "lang"=>"en", "lat"=>37.334596, "lon"=>-94.302581),
"/US/US-MO/jefferson-city-11791" => array("type"=>"city", "pop"=>40771, "name_ru"=>"Джефферсон Сити", "name_en"=>"Jefferson City", "lang"=>"en", "lat"=>38.571902, "lon"=>-92.190459),
"/US/US-MO/jennings-11792" => array("type"=>"city", "pop"=>14625, "name_ru"=>"Дженнингс", "name_en"=>"Jennings", "lang"=>"en", "lat"=>38.723578, "lon"=>-90.264457),
"/US/US-MO/jerico-springs-11793" => array("type"=>"village", "pop"=>257, "name_ru"=>"Джерико Спрингс", "name_en"=>"Jerico Springs", "lang"=>"en", "lat"=>37.618886, "lon"=>-94.010536),
"/US/US-MO/jonesburg-11794" => array("type"=>"city", "pop"=>726, "name_ru"=>"Джоунсбург", "name_en"=>"Jonesburg", "lang"=>"en", "lat"=>38.853582, "lon"=>-91.306467),
"/US/US-MO/joplin-11795" => array("type"=>"city", "pop"=>49775, "name_ru"=>"Джоплин", "name_en"=>"Joplin", "lang"=>"en", "lat"=>37.081800, "lon"=>-94.502974),
"/US/US-MO/josephville-11796" => array("type"=>"village", "pop"=>276, "name_ru"=>"Джоузефвилл", "name_en"=>"Josephville", "lang"=>"en", "lat"=>38.830995, "lon"=>-90.792882),
"/US/US-MO/junction-city-11797" => array("type"=>"village", "pop"=>333, "name_ru"=>"Джанкшен Сити", "name_en"=>"Junction City", "lang"=>"en", "lat"=>37.573930, "lon"=>-90.293901),
"/US/US-MO/kahoka-11798" => array("type"=>"city", "pop"=>2182, "name_ru"=>"Кахока", "name_en"=>"Kahoka", "lang"=>"en", "lat"=>40.423831, "lon"=>-91.718387),
"/US/US-MO/kansas-city-11799" => array("type"=>"city", "pop"=>451572, "name_ru"=>"Канзас Сити", "name_en"=>"Kansas City", "lang"=>"en", "lat"=>39.122312, "lon"=>-94.552009),
"/US/US-MO/kearney-11800" => array("type"=>"city", "pop"=>8599, "name_ru"=>"Кирни", "name_en"=>"Kearney", "lang"=>"en", "lat"=>39.364360, "lon"=>-94.361465),
"/US/US-MO/kelso-11801" => array("type"=>"town", "pop"=>519, "name_ru"=>"Келсо", "name_en"=>"Kelso", "lang"=>"en", "lat"=>37.190060, "lon"=>-89.550434),
"/US/US-MO/kennett-11802" => array("type"=>"city", "pop"=>10707, "name_ru"=>"Кеннетт", "name_en"=>"Kennett", "lang"=>"en", "lat"=>36.239991, "lon"=>-90.048708),
"/US/US-MO/keytesville-11803" => array("type"=>"city", "pop"=>491, "name_ru"=>"Китесвилл", "name_en"=>"Keytesville", "lang"=>"en", "lat"=>39.432734, "lon"=>-92.936814),
"/US/US-MO/kidder-11804" => array("type"=>"city", "pop"=>278, "name_ru"=>"Киддер", "name_en"=>"Kidder", "lang"=>"en", "lat"=>39.781903, "lon"=>-94.102381),
"/US/US-MO/kimberling-city-11805" => array("type"=>"city", "pop"=>2444, "name_ru"=>"Кимберлинг Сити", "name_en"=>"Kimberling City", "lang"=>"en", "lat"=>36.641881, "lon"=>-93.423908),
"/US/US-MO/kimmswick-11806" => array("type"=>"city", "pop"=>111, "name_ru"=>"Киммсвик", "name_en"=>"Kimmswick", "lang"=>"en", "lat"=>38.366700, "lon"=>-90.365149),
"/US/US-MO/king-city-11807" => array("type"=>"city", "pop"=>891, "name_ru"=>"Кинг Сити", "name_en"=>"King City", "lang"=>"en", "lat"=>40.050636, "lon"=>-94.524713),
"/US/US-MO/kingdom-city-11808" => array("type"=>"village", "pop"=>136, "name_ru"=>"Кингдом Сити", "name_en"=>"Kingdom City", "lang"=>"en", "lat"=>38.946310, "lon"=>-91.937225),
"/US/US-MO/kingston-11809" => array("type"=>"city", "pop"=>293, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>39.642238, "lon"=>-94.038183),
"/US/US-MO/kingsville-11810" => array("type"=>"city", "pop"=>257, "name_ru"=>"Кингсвилл", "name_en"=>"Kingsville", "lang"=>"en", "lat"=>38.742883, "lon"=>-94.068705),
"/US/US-MO/kinloch-11811" => array("type"=>"city", "pop"=>425, "name_ru"=>"Кинлок", "name_en"=>"Kinloch", "lang"=>"en", "lat"=>38.738150, "lon"=>-90.325319),
"/US/US-MO/kirksville-11812" => array("type"=>"city", "pop"=>17368, "name_ru"=>"Кирксвилл", "name_en"=>"Kirksville", "lang"=>"en", "lat"=>40.195113, "lon"=>-92.578426),
"/US/US-MO/kirkwood-11813" => array("type"=>"city", "pop"=>26760, "name_ru"=>"Кирквуд", "name_en"=>"Kirkwood", "lang"=>"en", "lat"=>38.580400, "lon"=>-90.420055),
"/US/US-MO/knob-noster-11814" => array("type"=>"city", "pop"=>3282, "name_ru"=>"Ноб Ностер", "name_en"=>"Knob Noster", "lang"=>"en", "lat"=>38.768044, "lon"=>-93.565842),
"/US/US-MO/knox-city-11815" => array("type"=>"city", "pop"=>207, "name_ru"=>"Нокс Сити", "name_en"=>"Knox City", "lang"=>"en", "lat"=>40.143639, "lon"=>-92.010271),
"/US/US-MO/koshkonong-11816" => array("type"=>"town", "pop"=>205, "name_ru"=>"Кошкононг", "name_en"=>"Koshkonong", "lang"=>"en", "lat"=>36.594437, "lon"=>-91.645549),
"/US/US-MO/la-belle-11817" => array("type"=>"city", "pop"=>648, "name_ru"=>"Ла Бэлль", "name_en"=>"La Belle", "lang"=>"en", "lat"=>40.116429, "lon"=>-91.913851),
"/US/US-MO/la-due-11820" => array("type"=>"village", "pop"=>20, "name_ru"=>"Ла Дью", "name_en"=>"La Due", "lang"=>"en", "lat"=>38.312754, "lon"=>-93.877758),
"/US/US-MO/la-grange-11822" => array("type"=>"city", "pop"=>926, "name_ru"=>"Ла Гранж", "name_en"=>"La Grange", "lang"=>"en", "lat"=>40.047970, "lon"=>-91.500483),
"/US/US-MO/la-monte-11839" => array("type"=>"city", "pop"=>1097, "name_ru"=>"Ла Монте", "name_en"=>"La Monte", "lang"=>"en", "lat"=>38.771143, "lon"=>-93.422515),
"/US/US-MO/la-plata-11842" => array("type"=>"city", "pop"=>1476, "name_ru"=>"Ла Плата", "name_en"=>"La Plata", "lang"=>"en", "lat"=>40.025043, "lon"=>-92.490854),
"/US/US-MO/la-russell-11844" => array("type"=>"city", "pop"=>148, "name_ru"=>"Ла Рассел", "name_en"=>"La Russell", "lang"=>"en", "lat"=>37.140755, "lon"=>-94.060840),
"/US/US-MO/la-tour-11846" => array("type"=>"town", "pop"=>66, "name_ru"=>"Ла Тур", "name_en"=>"La Tour", "lang"=>"en", "lat"=>38.633974, "lon"=>-94.102408),
"/US/US-MO/laclede-11818" => array("type"=>"city", "pop"=>388, "name_ru"=>"Лаклид", "name_en"=>"Laclede", "lang"=>"en", "lat"=>39.788222, "lon"=>-93.169515),
"/US/US-MO/laddonia-11819" => array("type"=>"city", "pop"=>598, "name_ru"=>"Ладдония", "name_en"=>"Laddonia", "lang"=>"en", "lat"=>39.241492, "lon"=>-91.642935),
"/US/US-MO/ladue-11821" => array("type"=>"city", "pop"=>8193, "name_ru"=>"Ладью", "name_en"=>"Ladue", "lang"=>"en", "lat"=>38.636837, "lon"=>-90.381605),
"/US/US-MO/lake-annette-11823" => array("type"=>"village", "pop"=>164, "name_ru"=>"Лейк Аннет", "name_en"=>"Lake Annette", "lang"=>"en", "lat"=>38.655296, "lon"=>-94.507820),
"/US/US-MO/lake-lotawana-11825" => array("type"=>"city", "pop"=>1954, "name_ru"=>"Лейк Лотавана", "name_en"=>"Lake Lotawana", "lang"=>"en", "lat"=>38.929500, "lon"=>-94.257855),
"/US/US-MO/lake-mykee-town-11826" => array("type"=>"village", "pop"=>344, "name_ru"=>"Лейк Майки Таун", "name_en"=>"Lake Mykee Town", "lang"=>"en", "lat"=>38.676468, "lon"=>-92.100803),
"/US/US-MO/lake-ozark-11827" => array("type"=>"city", "pop"=>2040, "name_ru"=>"Лейк Озарк", "name_en"=>"Lake Ozark", "lang"=>"en", "lat"=>38.188100, "lon"=>-92.629343),
"/US/US-MO/lake-st.-louis-11828" => array("type"=>"city", "pop"=>14160, "name_ru"=>"Лейк Сент Луис", "name_en"=>"Lake St. Louis", "lang"=>"en", "lat"=>38.789005, "lon"=>-90.782767),
"/US/US-MO/lake-tapawingo-11831" => array("type"=>"city", "pop"=>786, "name_ru"=>"Лейк Тапавинго", "name_en"=>"Lake Tapawingo", "lang"=>"en", "lat"=>39.018200, "lon"=>-94.310387),
"/US/US-MO/lake-waukomis-11834" => array("type"=>"city", "pop"=>899, "name_ru"=>"Лейк Уокомис", "name_en"=>"Lake Waukomis", "lang"=>"en", "lat"=>39.230300, "lon"=>-94.636464),
"/US/US-MO/lake-winnebago-11835" => array("type"=>"city", "pop"=>1139, "name_ru"=>"Лейк Уиннебаго", "name_en"=>"Lake Winnebago", "lang"=>"en", "lat"=>38.821392, "lon"=>-94.357169),
"/US/US-MO/lakeland-11824" => array("type"=>"town", "pop"=>351, "name_ru"=>"Лэйклэнд", "name_en"=>"Lakeland", "lang"=>"en", "lat"=>38.219550, "lon"=>-92.630923),
"/US/US-MO/lakeshire-11829" => array("type"=>"city", "pop"=>1285, "name_ru"=>"Лейкшир", "name_en"=>"Lakeshire", "lang"=>"en", "lat"=>38.541300, "lon"=>-90.336990),
"/US/US-MO/lakeside-11830" => array("type"=>"city", "pop"=>38, "name_ru"=>"Лэйксайд", "name_en"=>"Lakeside", "lang"=>"en", "lat"=>38.204030, "lon"=>-92.622673),
"/US/US-MO/lakeview-11832" => array("type"=>"village", "pop"=>110, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>38.213828, "lon"=>-92.627086),
"/US/US-MO/lakeview-11833" => array("type"=>"city", "pop"=>37, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>36.697712, "lon"=>-93.366309),
"/US/US-MO/lamar-11836" => array("type"=>"city", "pop"=>4474, "name_ru"=>"Ламар", "name_en"=>"Lamar", "lang"=>"en", "lat"=>37.493675, "lon"=>-94.273193),
"/US/US-MO/lamar-heights-11837" => array("type"=>"village", "pop"=>218, "name_ru"=>"Ламар Хайтс", "name_en"=>"Lamar Heights", "lang"=>"en", "lat"=>37.493563, "lon"=>-94.297801),
"/US/US-MO/lambert-11838" => array("type"=>"village", "pop"=>36, "name_ru"=>"Ламберт", "name_en"=>"Lambert", "lang"=>"en", "lat"=>37.093626, "lon"=>-89.555719),
"/US/US-MO/lanagan-11840" => array("type"=>"town", "pop"=>430, "name_ru"=>"Ланаган", "name_en"=>"Lanagan", "lang"=>"en", "lat"=>36.606714, "lon"=>-94.450813),
"/US/US-MO/lancaster-11841" => array("type"=>"city", "pop"=>702, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>40.522673, "lon"=>-92.530385),
"/US/US-MO/laredo-11843" => array("type"=>"city", "pop"=>241, "name_ru"=>"Ларедо", "name_en"=>"Laredo", "lang"=>"en", "lat"=>40.026325, "lon"=>-93.447314),
"/US/US-MO/lathrop-11845" => array("type"=>"city", "pop"=>2329, "name_ru"=>"Латроп", "name_en"=>"Lathrop", "lang"=>"en", "lat"=>39.549151, "lon"=>-94.329524),
"/US/US-MO/laurie-11847" => array("type"=>"village", "pop"=>734, "name_ru"=>"Лори", "name_en"=>"Laurie", "lang"=>"en", "lat"=>38.203638, "lon"=>-92.829105),
"/US/US-MO/lawson-11848" => array("type"=>"city", "pop"=>2341, "name_ru"=>"Лосон", "name_en"=>"Lawson", "lang"=>"en", "lat"=>39.436283, "lon"=>-94.213592),
"/US/US-MO/leadington-11849" => array("type"=>"city", "pop"=>418, "name_ru"=>"Лидингтон", "name_en"=>"Leadington", "lang"=>"en", "lat"=>37.834310, "lon"=>-90.481574),
"/US/US-MO/leadwood-11850" => array("type"=>"city", "pop"=>1159, "name_ru"=>"Лидвуд", "name_en"=>"Leadwood", "lang"=>"en", "lat"=>37.862247, "lon"=>-90.589351),
"/US/US-MO/leasburg-11851" => array("type"=>"village", "pop"=>337, "name_ru"=>"Лисбург", "name_en"=>"Leasburg", "lang"=>"en", "lat"=>38.094168, "lon"=>-91.295045),
"/US/US-MO/leawood-11852" => array("type"=>"village", "pop"=>959, "name_ru"=>"Ливуд", "name_en"=>"Leawood", "lang"=>"en", "lat"=>37.033400, "lon"=>-94.490499),
"/US/US-MO/lebanon-11853" => array("type"=>"city", "pop"=>14336, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>37.670549, "lon"=>-92.660859),
"/US/US-MO/lee's-summit-11854" => array("type"=>"city", "pop"=>84208, "name_ru"=>"Ли'с Саммит", "name_en"=>"Lee's Summit", "lang"=>"en", "lat"=>38.916343, "lon"=>-94.383516),
"/US/US-MO/leeton-11855" => array("type"=>"city", "pop"=>625, "name_ru"=>"Литон", "name_en"=>"Leeton", "lang"=>"en", "lat"=>38.583644, "lon"=>-93.695096),
"/US/US-MO/lemay-11856" => array("type"=>"cdp", "pop"=>18005, "name_ru"=>"Лемэй", "name_en"=>"Lemay", "lang"=>"en", "lat"=>38.531800, "lon"=>-90.283936),
"/US/US-MO/leonard-11857" => array("type"=>"town", "pop"=>62, "name_ru"=>"Леонард", "name_en"=>"Leonard", "lang"=>"en", "lat"=>39.895226, "lon"=>-92.180741),
"/US/US-MO/leslie-11858" => array("type"=>"village", "pop"=>86, "name_ru"=>"Лесли", "name_en"=>"Leslie", "lang"=>"en", "lat"=>38.417623, "lon"=>-91.231228),
"/US/US-MO/levasy-11859" => array("type"=>"city", "pop"=>97, "name_ru"=>"Леваси", "name_en"=>"Levasy", "lang"=>"en", "lat"=>39.133100, "lon"=>-94.132274),
"/US/US-MO/lewis-and-clark-village-11860" => array("type"=>"town", "pop"=>158, "name_ru"=>"Льюис энд Кларк Виллидж", "name_en"=>"Lewis and Clark Village", "lang"=>"en", "lat"=>39.539890, "lon"=>-95.050767),
"/US/US-MO/lewistown-11861" => array("type"=>"town", "pop"=>564, "name_ru"=>"Льюистаун", "name_en"=>"Lewistown", "lang"=>"en", "lat"=>40.084159, "lon"=>-91.812668),
"/US/US-MO/lexington-11862" => array("type"=>"city", "pop"=>4551, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>39.182099, "lon"=>-93.874859),
"/US/US-MO/liberal-11863" => array("type"=>"city", "pop"=>774, "name_ru"=>"Либерал", "name_en"=>"Liberal", "lang"=>"en", "lat"=>37.558777, "lon"=>-94.515979),
"/US/US-MO/liberty-11864" => array("type"=>"city", "pop"=>30568, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>39.242050, "lon"=>-94.419112),
"/US/US-MO/licking-11865" => array("type"=>"city", "pop"=>2772, "name_ru"=>"Ликинг", "name_en"=>"Licking", "lang"=>"en", "lat"=>37.496124, "lon"=>-91.860728),
"/US/US-MO/lilbourn-11866" => array("type"=>"city", "pop"=>1251, "name_ru"=>"Лилборн", "name_en"=>"Lilbourn", "lang"=>"en", "lat"=>36.590771, "lon"=>-89.613709),
"/US/US-MO/lincoln-11867" => array("type"=>"city", "pop"=>1081, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>38.394100, "lon"=>-93.330602),
"/US/US-MO/linn-11868" => array("type"=>"city", "pop"=>1432, "name_ru"=>"Линн", "name_en"=>"Linn", "lang"=>"en", "lat"=>38.480756, "lon"=>-91.844397),
"/US/US-MO/linn-creek-11869" => array("type"=>"city", "pop"=>310, "name_ru"=>"Линн Крик", "name_en"=>"Linn Creek", "lang"=>"en", "lat"=>38.044115, "lon"=>-92.701864),
"/US/US-MO/linneus-11870" => array("type"=>"city", "pop"=>342, "name_ru"=>"Линнеус", "name_en"=>"Linneus", "lang"=>"en", "lat"=>39.876922, "lon"=>-93.187488),
"/US/US-MO/lithium-11871" => array("type"=>"village", "pop"=>7, "name_ru"=>"Литиум", "name_en"=>"Lithium", "lang"=>"en", "lat"=>37.832205, "lon"=>-89.883776),
"/US/US-MO/livonia-11872" => array("type"=>"village", "pop"=>107, "name_ru"=>"Ливония", "name_en"=>"Livonia", "lang"=>"en", "lat"=>40.491624, "lon"=>-92.699667),
"/US/US-MO/lock-springs-11873" => array("type"=>"town", "pop"=>68, "name_ru"=>"Лок Спрингс", "name_en"=>"Lock Springs", "lang"=>"en", "lat"=>39.849093, "lon"=>-93.775871),
"/US/US-MO/lockwood-11874" => array("type"=>"city", "pop"=>918, "name_ru"=>"Локвуд", "name_en"=>"Lockwood", "lang"=>"en", "lat"=>37.387171, "lon"=>-93.958698),
"/US/US-MO/lohman-11875" => array("type"=>"town", "pop"=>164, "name_ru"=>"Ломан", "name_en"=>"Lohman", "lang"=>"en", "lat"=>38.542477, "lon"=>-92.363473),
"/US/US-MO/lone-jack-11876" => array("type"=>"village", "pop"=>917, "name_ru"=>"Лоун Джэк", "name_en"=>"Lone Jack", "lang"=>"en", "lat"=>38.870372, "lon"=>-94.166417),
"/US/US-MO/longtown-11877" => array("type"=>"town", "pop"=>77, "name_ru"=>"Лонгтаун", "name_en"=>"Longtown", "lang"=>"en", "lat"=>37.669850, "lon"=>-89.773513),
"/US/US-MO/louisburg-11878" => array("type"=>"village", "pop"=>156, "name_ru"=>"Луисбург", "name_en"=>"Louisburg", "lang"=>"en", "lat"=>37.754894, "lon"=>-93.140139),
"/US/US-MO/louisiana-11879" => array("type"=>"city", "pop"=>3768, "name_ru"=>"Луизианна", "name_en"=>"Louisiana", "lang"=>"en", "lat"=>39.443522, "lon"=>-91.059239),
"/US/US-MO/lowry-city-11880" => array("type"=>"city", "pop"=>724, "name_ru"=>"Лоури Сити", "name_en"=>"Lowry City", "lang"=>"en", "lat"=>38.141014, "lon"=>-93.726749),
"/US/US-MO/lucerne-11881" => array("type"=>"village", "pop"=>85, "name_ru"=>"Лукерне", "name_en"=>"Lucerne", "lang"=>"en", "lat"=>40.463717, "lon"=>-93.291573),
"/US/US-MO/ludlow-11882" => array("type"=>"town", "pop"=>212, "name_ru"=>"Ладлоу", "name_en"=>"Ludlow", "lang"=>"en", "lat"=>39.653542, "lon"=>-93.703093),
"/US/US-MO/lupus-11883" => array("type"=>"town", "pop"=>29, "name_ru"=>"Лупус", "name_en"=>"Lupus", "lang"=>"en", "lat"=>38.845717, "lon"=>-92.454216),
"/US/US-MO/luray-11884" => array("type"=>"town", "pop"=>99, "name_ru"=>"Лурей", "name_en"=>"Luray", "lang"=>"en", "lat"=>40.452545, "lon"=>-91.884301),
"/US/US-MO/mac-kenzie-11886" => array("type"=>"village", "pop"=>148, "name_ru"=>"Мак Кензи", "name_en"=>"Mac Kenzie", "lang"=>"en", "lat"=>38.580300, "lon"=>-90.316475),
"/US/US-MO/macks-creek-11888" => array("type"=>"city", "pop"=>288, "name_ru"=>"Мэкс Крик", "name_en"=>"Macks Creek", "lang"=>"en", "lat"=>37.965994, "lon"=>-92.971976),
"/US/US-MO/macon-11889" => array("type"=>"city", "pop"=>5489, "name_ru"=>"Мейкон", "name_en"=>"Macon", "lang"=>"en", "lat"=>39.738743, "lon"=>-92.470528),
"/US/US-MO/madison-11890" => array("type"=>"city", "pop"=>536, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>39.473160, "lon"=>-92.211919),
"/US/US-MO/maitland-11891" => array("type"=>"city", "pop"=>303, "name_ru"=>"Мэйтлэнд", "name_en"=>"Maitland", "lang"=>"en", "lat"=>40.203071, "lon"=>-95.078713),
"/US/US-MO/malden-11892" => array("type"=>"city", "pop"=>4481, "name_ru"=>"Малден", "name_en"=>"Malden", "lang"=>"en", "lat"=>36.584867, "lon"=>-89.975562),
"/US/US-MO/malta-bend-11893" => array("type"=>"town", "pop"=>237, "name_ru"=>"Мальта Бенд", "name_en"=>"Malta Bend", "lang"=>"en", "lat"=>39.194254, "lon"=>-93.364139),
"/US/US-MO/manchester-11894" => array("type"=>"city", "pop"=>18569, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>38.592550, "lon"=>-90.514820),
"/US/US-MO/mansfield-11895" => array("type"=>"city", "pop"=>1361, "name_ru"=>"Мансфилд", "name_en"=>"Mansfield", "lang"=>"en", "lat"=>37.109739, "lon"=>-92.580899),
"/US/US-MO/maplewood-11896" => array("type"=>"city", "pop"=>8626, "name_ru"=>"Мейплвуд", "name_en"=>"Maplewood", "lang"=>"en", "lat"=>38.612150, "lon"=>-90.324320),
"/US/US-MO/marble-hill-11897" => array("type"=>"city", "pop"=>1482, "name_ru"=>"Марбл Хилл", "name_en"=>"Marble Hill", "lang"=>"en", "lat"=>37.302286, "lon"=>-89.980434),
"/US/US-MO/marceline-11898" => array("type"=>"city", "pop"=>2306, "name_ru"=>"Марслайн", "name_en"=>"Marceline", "lang"=>"en", "lat"=>39.716957, "lon"=>-92.947428),
"/US/US-MO/marionville-11899" => array("type"=>"city", "pop"=>2176, "name_ru"=>"Мэрионвилл", "name_en"=>"Marionville", "lang"=>"en", "lat"=>37.000281, "lon"=>-93.636692),
"/US/US-MO/marlborough-11900" => array("type"=>"village", "pop"=>2140, "name_ru"=>"Марлборо", "name_en"=>"Marlborough", "lang"=>"en", "lat"=>38.567833, "lon"=>-90.338899),
"/US/US-MO/marquand-11901" => array("type"=>"city", "pop"=>266, "name_ru"=>"Маркванд", "name_en"=>"Marquand", "lang"=>"en", "lat"=>37.428691, "lon"=>-90.167065),
"/US/US-MO/marshall-11902" => array("type"=>"city", "pop"=>12118, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>39.114954, "lon"=>-93.202167),
"/US/US-MO/marshfield-11903" => array("type"=>"city", "pop"=>7274, "name_ru"=>"Маршфилд", "name_en"=>"Marshfield", "lang"=>"en", "lat"=>37.339437, "lon"=>-92.910354),
"/US/US-MO/marston-11904" => array("type"=>"city", "pop"=>529, "name_ru"=>"Марстон", "name_en"=>"Marston", "lang"=>"en", "lat"=>36.517574, "lon"=>-89.608739),
"/US/US-MO/marthasville-11905" => array("type"=>"city", "pop"=>863, "name_ru"=>"Мартасвилл", "name_en"=>"Marthasville", "lang"=>"en", "lat"=>38.628410, "lon"=>-91.053329),
"/US/US-MO/martinsburg-11906" => array("type"=>"town", "pop"=>327, "name_ru"=>"Мартинсбург", "name_en"=>"Martinsburg", "lang"=>"en", "lat"=>39.101143, "lon"=>-91.647191),
"/US/US-MO/maryland-heights-11907" => array("type"=>"city", "pop"=>25949, "name_ru"=>"Мэриленд Хайтс", "name_en"=>"Maryland Heights", "lang"=>"en", "lat"=>38.722474, "lon"=>-90.477403),
"/US/US-MO/maryville-11908" => array("type"=>"city", "pop"=>10800, "name_ru"=>"Мэривилл", "name_en"=>"Maryville", "lang"=>"en", "lat"=>40.343340, "lon"=>-94.871001),
"/US/US-MO/matthews-11909" => array("type"=>"city", "pop"=>534, "name_ru"=>"Мэттьюс", "name_en"=>"Matthews", "lang"=>"en", "lat"=>36.759300, "lon"=>-89.583625),
"/US/US-MO/maysville-11910" => array("type"=>"city", "pop"=>1139, "name_ru"=>"Мейсвилл", "name_en"=>"Maysville", "lang"=>"en", "lat"=>39.887040, "lon"=>-94.361341),
"/US/US-MO/mayview-11911" => array("type"=>"city", "pop"=>286, "name_ru"=>"Мэйвью", "name_en"=>"Mayview", "lang"=>"en", "lat"=>39.052581, "lon"=>-93.833864),
"/US/US-MO/mcfall-11885" => array("type"=>"city", "pop"=>123, "name_ru"=>"Мак-Фолл", "name_en"=>"McFall", "lang"=>"en", "lat"=>40.111743, "lon"=>-94.222546),
"/US/US-MO/mckittrick-11887" => array("type"=>"town", "pop"=>66, "name_ru"=>"Мак-Киттрик", "name_en"=>"McKittrick", "lang"=>"en", "lat"=>38.735197, "lon"=>-91.442689),
"/US/US-MO/meadville-11912" => array("type"=>"city", "pop"=>425, "name_ru"=>"Мидвилл", "name_en"=>"Meadville", "lang"=>"en", "lat"=>39.787765, "lon"=>-93.302680),
"/US/US-MO/mehlville-11913" => array("type"=>"cdp", "pop"=>27557, "name_ru"=>"Мелвилл", "name_en"=>"Mehlville", "lang"=>"en", "lat"=>38.501133, "lon"=>-90.315674),
"/US/US-MO/memphis-11914" => array("type"=>"city", "pop"=>1959, "name_ru"=>"Мемфис", "name_en"=>"Memphis", "lang"=>"en", "lat"=>40.460855, "lon"=>-92.169727),
"/US/US-MO/mendon-11915" => array("type"=>"town", "pop"=>191, "name_ru"=>"Мендон", "name_en"=>"Mendon", "lang"=>"en", "lat"=>39.591102, "lon"=>-93.133580),
"/US/US-MO/mercer-11916" => array("type"=>"town", "pop"=>326, "name_ru"=>"Мерсер", "name_en"=>"Mercer", "lang"=>"en", "lat"=>40.511038, "lon"=>-93.529349),
"/US/US-MO/merriam-woods-11917" => array("type"=>"village", "pop"=>1446, "name_ru"=>"Мерриам Вудс", "name_en"=>"Merriam Woods", "lang"=>"en", "lat"=>36.716777, "lon"=>-93.170714),
"/US/US-MO/merwin-11918" => array("type"=>"town", "pop"=>85, "name_ru"=>"Мервин", "name_en"=>"Merwin", "lang"=>"en", "lat"=>38.404195, "lon"=>-94.589763),
"/US/US-MO/meta-11919" => array("type"=>"city", "pop"=>257, "name_ru"=>"Мета", "name_en"=>"Meta", "lang"=>"en", "lat"=>38.312400, "lon"=>-92.166574),
"/US/US-MO/metz-11920" => array("type"=>"town", "pop"=>66, "name_ru"=>"Метс", "name_en"=>"Metz", "lang"=>"en", "lat"=>37.996205, "lon"=>-94.442108),
"/US/US-MO/mexico-11921" => array("type"=>"city", "pop"=>11074, "name_ru"=>"Мексико", "name_en"=>"Mexico", "lang"=>"en", "lat"=>39.166478, "lon"=>-91.870638),
"/US/US-MO/miami-11922" => array("type"=>"city", "pop"=>153, "name_ru"=>"Майами", "name_en"=>"Miami", "lang"=>"en", "lat"=>39.321734, "lon"=>-93.225488),
"/US/US-MO/middletown-11923" => array("type"=>"town", "pop"=>181, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>39.128532, "lon"=>-91.414512),
"/US/US-MO/milan-11924" => array("type"=>"city", "pop"=>1774, "name_ru"=>"Милан", "name_en"=>"Milan", "lang"=>"en", "lat"=>40.203223, "lon"=>-93.123926),
"/US/US-MO/milford-11925" => array("type"=>"village", "pop"=>52, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>37.586347, "lon"=>-94.156276),
"/US/US-MO/mill-spring-11928" => array("type"=>"village", "pop"=>208, "name_ru"=>"Милл Спрингс", "name_en"=>"Mill Spring", "lang"=>"en", "lat"=>37.064698, "lon"=>-90.681424),
"/US/US-MO/millard-11926" => array("type"=>"village", "pop"=>75, "name_ru"=>"Миллард", "name_en"=>"Millard", "lang"=>"en", "lat"=>40.107135, "lon"=>-92.545328),
"/US/US-MO/miller-11927" => array("type"=>"city", "pop"=>796, "name_ru"=>"Миллер", "name_en"=>"Miller", "lang"=>"en", "lat"=>37.215307, "lon"=>-93.840318),
"/US/US-MO/milo-11929" => array("type"=>"town", "pop"=>82, "name_ru"=>"Майло", "name_en"=>"Milo", "lang"=>"en", "lat"=>37.755146, "lon"=>-94.305250),
"/US/US-MO/mindenmines-11930" => array("type"=>"city", "pop"=>408, "name_ru"=>"Минденминс", "name_en"=>"Mindenmines", "lang"=>"en", "lat"=>37.473278, "lon"=>-94.589658),
"/US/US-MO/miner-11931" => array("type"=>"city", "pop"=>1313, "name_ru"=>"Майнер", "name_en"=>"Miner", "lang"=>"en", "lat"=>36.893830, "lon"=>-89.534602),
"/US/US-MO/mineral-point-11932" => array("type"=>"town", "pop"=>385, "name_ru"=>"Минерал Пойнт", "name_en"=>"Mineral Point", "lang"=>"en", "lat"=>37.945132, "lon"=>-90.724364),
"/US/US-MO/missouri-city-11933" => array("type"=>"city", "pop"=>331, "name_ru"=>"Миссури Сити", "name_en"=>"Missouri City", "lang"=>"en", "lat"=>39.238144, "lon"=>-94.300903),
"/US/US-MO/moberly-11934" => array("type"=>"city", "pop"=>14227, "name_ru"=>"Моберли", "name_en"=>"Moberly", "lang"=>"en", "lat"=>39.419748, "lon"=>-92.436222),
"/US/US-MO/mokane-11935" => array("type"=>"city", "pop"=>195, "name_ru"=>"Мокейн", "name_en"=>"Mokane", "lang"=>"en", "lat"=>38.674357, "lon"=>-91.872748),
"/US/US-MO/moline-acres-11936" => array("type"=>"city", "pop"=>2517, "name_ru"=>"Молайн Эйкерс", "name_en"=>"Moline Acres", "lang"=>"en", "lat"=>38.745100, "lon"=>-90.242261),
"/US/US-MO/monett-11937" => array("type"=>"city", "pop"=>8939, "name_ru"=>"Монетт", "name_en"=>"Monett", "lang"=>"en", "lat"=>36.921123, "lon"=>-93.923706),
"/US/US-MO/monroe-city-11938" => array("type"=>"city", "pop"=>2473, "name_ru"=>"Монро Сити", "name_en"=>"Monroe City", "lang"=>"en", "lat"=>39.654088, "lon"=>-91.732167),
"/US/US-MO/montgomery-city-11939" => array("type"=>"city", "pop"=>2485, "name_ru"=>"Монтгомери Сити", "name_en"=>"Montgomery City", "lang"=>"en", "lat"=>38.975622, "lon"=>-91.505121),
"/US/US-MO/monticello-11940" => array("type"=>"town", "pop"=>119, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>40.119112, "lon"=>-91.712614),
"/US/US-MO/montrose-11941" => array("type"=>"city", "pop"=>419, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>38.259402, "lon"=>-93.981985),
"/US/US-MO/mooresville-11942" => array("type"=>"village", "pop"=>84, "name_ru"=>"Мурсвилл", "name_en"=>"Mooresville", "lang"=>"en", "lat"=>39.746380, "lon"=>-93.720208),
"/US/US-MO/morehouse-11943" => array("type"=>"city", "pop"=>922, "name_ru"=>"Морхаус", "name_en"=>"Morehouse", "lang"=>"en", "lat"=>36.846021, "lon"=>-89.690938),
"/US/US-MO/morley-11944" => array("type"=>"city", "pop"=>796, "name_ru"=>"Морли", "name_en"=>"Morley", "lang"=>"en", "lat"=>37.043329, "lon"=>-89.612562),
"/US/US-MO/morrison-11945" => array("type"=>"city", "pop"=>122, "name_ru"=>"Моррисон", "name_en"=>"Morrison", "lang"=>"en", "lat"=>38.670653, "lon"=>-91.632823),
"/US/US-MO/morrisville-11946" => array("type"=>"town", "pop"=>375, "name_ru"=>"Моррисвилл", "name_en"=>"Morrisville", "lang"=>"en", "lat"=>37.480662, "lon"=>-93.428575),
"/US/US-MO/mosby-11947" => array("type"=>"city", "pop"=>249, "name_ru"=>"Мосби", "name_en"=>"Mosby", "lang"=>"en", "lat"=>39.319584, "lon"=>-94.300825),
"/US/US-MO/moscow-mills-11948" => array("type"=>"city", "pop"=>2460, "name_ru"=>"Москау Миллс", "name_en"=>"Moscow Mills", "lang"=>"en", "lat"=>38.945889, "lon"=>-90.919181),
"/US/US-MO/mound-city-11949" => array("type"=>"city", "pop"=>1074, "name_ru"=>"Маунд Сити", "name_en"=>"Mound City", "lang"=>"en", "lat"=>40.135971, "lon"=>-95.234249),
"/US/US-MO/moundville-11950" => array("type"=>"town", "pop"=>101, "name_ru"=>"Маундвилл", "name_en"=>"Moundville", "lang"=>"en", "lat"=>37.765935, "lon"=>-94.451239),
"/US/US-MO/mount-leonard-11953" => array("type"=>"town", "pop"=>119, "name_ru"=>"Маунт Леонард", "name_en"=>"Mount Leonard", "lang"=>"en", "lat"=>39.124982, "lon"=>-93.394692),
"/US/US-MO/mount-moriah-11954" => array("type"=>"town", "pop"=>144, "name_ru"=>"Маунт Морайа", "name_en"=>"Mount Moriah", "lang"=>"en", "lat"=>40.329326, "lon"=>-93.797329),
"/US/US-MO/mount-vernon-11955" => array("type"=>"city", "pop"=>4600, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>37.101622, "lon"=>-93.817924),
"/US/US-MO/mountain-grove-11951" => array("type"=>"city", "pop"=>4679, "name_ru"=>"Маунтейн Гров", "name_en"=>"Mountain Grove", "lang"=>"en", "lat"=>37.132215, "lon"=>-92.264875),
"/US/US-MO/mountain-view-11952" => array("type"=>"city", "pop"=>2603, "name_ru"=>"Маунтин Вью", "name_en"=>"Mountain View", "lang"=>"en", "lat"=>36.994755, "lon"=>-91.698687),
"/US/US-MO/murphy-11956" => array("type"=>"cdp", "pop"=>9342, "name_ru"=>"Мерфи", "name_en"=>"Murphy", "lang"=>"en", "lat"=>38.492900, "lon"=>-90.486423),
"/US/US-MO/napoleon-11957" => array("type"=>"city", "pop"=>196, "name_ru"=>"Наполеон", "name_en"=>"Napoleon", "lang"=>"en", "lat"=>39.132304, "lon"=>-94.088352),
"/US/US-MO/naylor-11958" => array("type"=>"city", "pop"=>596, "name_ru"=>"Нейлор", "name_en"=>"Naylor", "lang"=>"en", "lat"=>36.573965, "lon"=>-90.605344),
"/US/US-MO/neck-city-11959" => array("type"=>"city", "pop"=>127, "name_ru"=>"Нек Сити", "name_en"=>"Neck City", "lang"=>"en", "lat"=>37.256272, "lon"=>-94.443598),
"/US/US-MO/neelyville-11960" => array("type"=>"city", "pop"=>500, "name_ru"=>"Нилвилл", "name_en"=>"Neelyville", "lang"=>"en", "lat"=>36.559233, "lon"=>-90.511012),
"/US/US-MO/nelson-11961" => array("type"=>"city", "pop"=>202, "name_ru"=>"Нельсон", "name_en"=>"Nelson", "lang"=>"en", "lat"=>38.993546, "lon"=>-93.030342),
"/US/US-MO/neosho-11962" => array("type"=>"city", "pop"=>11317, "name_ru"=>"Неошо", "name_en"=>"Neosho", "lang"=>"en", "lat"=>36.844680, "lon"=>-94.375221),
"/US/US-MO/nevada-11963" => array("type"=>"city", "pop"=>8327, "name_ru"=>"Невада", "name_en"=>"Nevada", "lang"=>"en", "lat"=>37.841168, "lon"=>-94.352951),
"/US/US-MO/new-bloomfield-11965" => array("type"=>"city", "pop"=>739, "name_ru"=>"Нью Блумфилд", "name_en"=>"New Bloomfield", "lang"=>"en", "lat"=>38.719421, "lon"=>-92.089667),
"/US/US-MO/new-cambria-11967" => array("type"=>"town", "pop"=>222, "name_ru"=>"Нью Камбрия", "name_en"=>"New Cambria", "lang"=>"en", "lat"=>39.775253, "lon"=>-92.751283),
"/US/US-MO/new-florence-11968" => array("type"=>"city", "pop"=>753, "name_ru"=>"Нью Флоренс", "name_en"=>"New Florence", "lang"=>"en", "lat"=>38.908224, "lon"=>-91.449896),
"/US/US-MO/new-franklin-11969" => array("type"=>"city", "pop"=>1111, "name_ru"=>"Нью Франклин", "name_en"=>"New Franklin", "lang"=>"en", "lat"=>39.018627, "lon"=>-92.740597),
"/US/US-MO/new-hampton-11970" => array("type"=>"city", "pop"=>328, "name_ru"=>"Нью Хэмптон", "name_en"=>"New Hampton", "lang"=>"en", "lat"=>40.265111, "lon"=>-94.195752),
"/US/US-MO/new-haven-11971" => array("type"=>"city", "pop"=>2029, "name_ru"=>"Нью Хейвен", "name_en"=>"New Haven", "lang"=>"en", "lat"=>38.607467, "lon"=>-91.218133),
"/US/US-MO/new-london-11972" => array("type"=>"city", "pop"=>1009, "name_ru"=>"Нью Лондон", "name_en"=>"New London", "lang"=>"en", "lat"=>39.584179, "lon"=>-91.398815),
"/US/US-MO/new-madrid-11973" => array("type"=>"city", "pop"=>3012, "name_ru"=>"Нью Мадрид", "name_en"=>"New Madrid", "lang"=>"en", "lat"=>36.589283, "lon"=>-89.548616),
"/US/US-MO/new-melle-11974" => array("type"=>"village", "pop"=>281, "name_ru"=>"Нью Мелл", "name_en"=>"New Melle", "lang"=>"en", "lat"=>38.710165, "lon"=>-90.880258),
"/US/US-MO/newark-11964" => array("type"=>"town", "pop"=>93, "name_ru"=>"Ньюарк", "name_en"=>"Newark", "lang"=>"en", "lat"=>39.994435, "lon"=>-91.973079),
"/US/US-MO/newburg-11966" => array("type"=>"city", "pop"=>467, "name_ru"=>"Ньюбург", "name_en"=>"Newburg", "lang"=>"en", "lat"=>37.916669, "lon"=>-91.901362),
"/US/US-MO/newtonia-11975" => array("type"=>"town", "pop"=>242, "name_ru"=>"Ньютония", "name_en"=>"Newtonia", "lang"=>"en", "lat"=>36.879518, "lon"=>-94.184065),
"/US/US-MO/newtown-11976" => array("type"=>"town", "pop"=>188, "name_ru"=>"Ньютаун", "name_en"=>"Newtown", "lang"=>"en", "lat"=>40.376577, "lon"=>-93.332824),
"/US/US-MO/niangua-11977" => array("type"=>"city", "pop"=>494, "name_ru"=>"Найангуа", "name_en"=>"Niangua", "lang"=>"en", "lat"=>37.387649, "lon"=>-92.829729),
"/US/US-MO/nixa-11978" => array("type"=>"city", "pop"=>18829, "name_ru"=>"Никса", "name_en"=>"Nixa", "lang"=>"en", "lat"=>37.047050, "lon"=>-93.297924),
"/US/US-MO/noel-11979" => array("type"=>"city", "pop"=>1608, "name_ru"=>"Ноел", "name_en"=>"Noel", "lang"=>"en", "lat"=>36.545028, "lon"=>-94.487179),
"/US/US-MO/norborne-11980" => array("type"=>"city", "pop"=>753, "name_ru"=>"Норборн", "name_en"=>"Norborne", "lang"=>"en", "lat"=>39.302252, "lon"=>-93.675825),
"/US/US-MO/normandy-11981" => array("type"=>"city", "pop"=>4902, "name_ru"=>"Норманди", "name_en"=>"Normandy", "lang"=>"en", "lat"=>38.707700, "lon"=>-90.300412),
"/US/US-MO/north-kansas-city-11982" => array("type"=>"city", "pop"=>5746, "name_ru"=>"Норт Канзас Сити", "name_en"=>"North Kansas City", "lang"=>"en", "lat"=>39.139150, "lon"=>-94.563963),
"/US/US-MO/north-lilbourn-11983" => array("type"=>"village", "pop"=>85, "name_ru"=>"Норт Лилборн", "name_en"=>"North Lilbourn", "lang"=>"en", "lat"=>36.601706, "lon"=>-89.621941),
"/US/US-MO/north-wardell-11985" => array("type"=>"village", "pop"=>135, "name_ru"=>"Норт Уорделл", "name_en"=>"North Wardell", "lang"=>"en", "lat"=>36.357773, "lon"=>-89.815463),
"/US/US-MO/northmoor-11984" => array("type"=>"city", "pop"=>399, "name_ru"=>"Нортмур", "name_en"=>"Northmoor", "lang"=>"en", "lat"=>39.184333, "lon"=>-94.606646),
"/US/US-MO/northwoods-11986" => array("type"=>"city", "pop"=>4311, "name_ru"=>"Нортвудс", "name_en"=>"Northwoods", "lang"=>"en", "lat"=>38.703200, "lon"=>-90.282224),
"/US/US-MO/norwood-11987" => array("type"=>"city", "pop"=>578, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>37.106795, "lon"=>-92.416677),
"/US/US-MO/norwood-court-11988" => array("type"=>"town", "pop"=>1019, "name_ru"=>"Норвуд Корт", "name_en"=>"Norwood Court", "lang"=>"en", "lat"=>38.715700, "lon"=>-90.290352),
"/US/US-MO/novelty-11989" => array("type"=>"town", "pop"=>110, "name_ru"=>"Новелти", "name_en"=>"Novelty", "lang"=>"en", "lat"=>40.012571, "lon"=>-92.207434),
"/US/US-MO/novinger-11990" => array("type"=>"city", "pop"=>526, "name_ru"=>"Новингер", "name_en"=>"Novinger", "lang"=>"en", "lat"=>40.234097, "lon"=>-92.705893),
"/US/US-MO/o'fallon-12002" => array("type"=>"city", "pop"=>76819, "name_ru"=>"О'фаллон", "name_en"=>"O'Fallon", "lang"=>"en", "lat"=>38.784306, "lon"=>-90.707878),
"/US/US-MO/oak-grove-11991" => array("type"=>"city", "pop"=>6852, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>39.003500, "lon"=>-94.125727),
"/US/US-MO/oak-grove-11992" => array("type"=>"village", "pop"=>402, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>38.226886, "lon"=>-91.151246),
"/US/US-MO/oak-ridge-11995" => array("type"=>"town", "pop"=>247, "name_ru"=>"Оук Ридж", "name_en"=>"Oak Ridge", "lang"=>"en", "lat"=>37.499766, "lon"=>-89.729591),
"/US/US-MO/oakland-11993" => array("type"=>"city", "pop"=>1555, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>38.575817, "lon"=>-90.384422),
"/US/US-MO/oakland-park-11994" => array("type"=>"village", "pop"=>89, "name_ru"=>"Оуклэнд Парк", "name_en"=>"Oakland Park", "lang"=>"en", "lat"=>37.110044, "lon"=>-94.472644),
"/US/US-MO/oaks-11996" => array("type"=>"village", "pop"=>139, "name_ru"=>"Оукс", "name_en"=>"Oaks", "lang"=>"en", "lat"=>39.196950, "lon"=>-94.571875),
"/US/US-MO/oakview-11997" => array("type"=>"village", "pop"=>396, "name_ru"=>"Оуквью", "name_en"=>"Oakview", "lang"=>"en", "lat"=>39.208000, "lon"=>-94.570799),
"/US/US-MO/oakville-11998" => array("type"=>"cdp", "pop"=>31750, "name_ru"=>"Оуквилл", "name_en"=>"Oakville", "lang"=>"en", "lat"=>38.444138, "lon"=>-90.316902),
"/US/US-MO/oakwood-11999" => array("type"=>"village", "pop"=>205, "name_ru"=>"Оуквуд", "name_en"=>"Oakwood", "lang"=>"en", "lat"=>39.200300, "lon"=>-94.572149),
"/US/US-MO/oakwood-park-12000" => array("type"=>"village", "pop"=>182, "name_ru"=>"Оуквуд Парк", "name_en"=>"Oakwood Park", "lang"=>"en", "lat"=>39.205717, "lon"=>-94.572815),
"/US/US-MO/odessa-12001" => array("type"=>"city", "pop"=>4737, "name_ru"=>"Одесса", "name_en"=>"Odessa", "lang"=>"en", "lat"=>38.996954, "lon"=>-93.954753),
"/US/US-MO/old-appleton-12003" => array("type"=>"town", "pop"=>85, "name_ru"=>"Оулд Эпплтон", "name_en"=>"Old Appleton", "lang"=>"en", "lat"=>37.593205, "lon"=>-89.709394),
"/US/US-MO/old-monroe-12004" => array("type"=>"city", "pop"=>306, "name_ru"=>"Оулд Монро", "name_en"=>"Old Monroe", "lang"=>"en", "lat"=>38.931280, "lon"=>-90.747024),
"/US/US-MO/olean-12005" => array("type"=>"town", "pop"=>163, "name_ru"=>"Олин", "name_en"=>"Olean", "lang"=>"en", "lat"=>38.410352, "lon"=>-92.530127),
"/US/US-MO/olivette-12006" => array("type"=>"city", "pop"=>7449, "name_ru"=>"Оливетт", "name_en"=>"Olivette", "lang"=>"en", "lat"=>38.672300, "lon"=>-90.377225),
"/US/US-MO/olympian-village-12007" => array("type"=>"city", "pop"=>692, "name_ru"=>"Олимпиан Виллидж", "name_en"=>"Olympian Village", "lang"=>"en", "lat"=>38.134643, "lon"=>-90.457239),
"/US/US-MO/oran-12008" => array("type"=>"city", "pop"=>1248, "name_ru"=>"Оран", "name_en"=>"Oran", "lang"=>"en", "lat"=>37.085028, "lon"=>-89.652894),
"/US/US-MO/oregon-12009" => array("type"=>"city", "pop"=>872, "name_ru"=>"Орегон", "name_en"=>"Oregon", "lang"=>"en", "lat"=>39.984498, "lon"=>-95.143632),
"/US/US-MO/oronogo-12010" => array("type"=>"city", "pop"=>2174, "name_ru"=>"Ороного", "name_en"=>"Oronogo", "lang"=>"en", "lat"=>37.187869, "lon"=>-94.469185),
"/US/US-MO/orrick-12011" => array("type"=>"city", "pop"=>822, "name_ru"=>"Оррик", "name_en"=>"Orrick", "lang"=>"en", "lat"=>39.213600, "lon"=>-94.125155),
"/US/US-MO/osage-beach-12012" => array("type"=>"city", "pop"=>4720, "name_ru"=>"Осейдж Бич", "name_en"=>"Osage Beach", "lang"=>"en", "lat"=>38.135190, "lon"=>-92.652280),
"/US/US-MO/osborn-12013" => array("type"=>"city", "pop"=>434, "name_ru"=>"Осборн", "name_en"=>"Osborn", "lang"=>"en", "lat"=>39.750020, "lon"=>-94.357112),
"/US/US-MO/osceola-12014" => array("type"=>"city", "pop"=>785, "name_ru"=>"Осеола", "name_en"=>"Osceola", "lang"=>"en", "lat"=>38.046446, "lon"=>-93.698017),
"/US/US-MO/osgood-12015" => array("type"=>"town", "pop"=>47, "name_ru"=>"Осгуд", "name_en"=>"Osgood", "lang"=>"en", "lat"=>40.198095, "lon"=>-93.350577),
"/US/US-MO/otterville-12016" => array("type"=>"city", "pop"=>490, "name_ru"=>"Оттервилл", "name_en"=>"Otterville", "lang"=>"en", "lat"=>38.702277, "lon"=>-93.002815),
"/US/US-MO/overland-12017" => array("type"=>"city", "pop"=>15632, "name_ru"=>"Оверленд", "name_en"=>"Overland", "lang"=>"en", "lat"=>38.696000, "lon"=>-90.368176),
"/US/US-MO/owensville-12018" => array("type"=>"city", "pop"=>2588, "name_ru"=>"Оуэнсвилл", "name_en"=>"Owensville", "lang"=>"en", "lat"=>38.347584, "lon"=>-91.497677),
"/US/US-MO/ozark-12019" => array("type"=>"city", "pop"=>18095, "name_ru"=>"Озарк", "name_en"=>"Ozark", "lang"=>"en", "lat"=>37.019700, "lon"=>-93.208001),
"/US/US-MO/pacific-12020" => array("type"=>"city", "pop"=>7268, "name_ru"=>"Пасифик", "name_en"=>"Pacific", "lang"=>"en", "lat"=>38.478333, "lon"=>-90.752919),
"/US/US-MO/pagedale-12021" => array("type"=>"city", "pop"=>3395, "name_ru"=>"Пейдждейл", "name_en"=>"Pagedale", "lang"=>"en", "lat"=>38.680887, "lon"=>-90.308335),
"/US/US-MO/palmyra-12022" => array("type"=>"city", "pop"=>3416, "name_ru"=>"Палмира", "name_en"=>"Palmyra", "lang"=>"en", "lat"=>39.796424, "lon"=>-91.524703),
"/US/US-MO/paris-12023" => array("type"=>"city", "pop"=>1399, "name_ru"=>"Парис", "name_en"=>"Paris", "lang"=>"en", "lat"=>39.477148, "lon"=>-92.003931),
"/US/US-MO/parkdale-12024" => array("type"=>"town", "pop"=>201, "name_ru"=>"Паркдейл", "name_en"=>"Parkdale", "lang"=>"en", "lat"=>38.478950, "lon"=>-90.526568),
"/US/US-MO/parkville-12025" => array("type"=>"city", "pop"=>5279, "name_ru"=>"Парквилл", "name_en"=>"Parkville", "lang"=>"en", "lat"=>39.195100, "lon"=>-94.674138),
"/US/US-MO/parkway-12026" => array("type"=>"village", "pop"=>276, "name_ru"=>"Паркуэй", "name_en"=>"Parkway", "lang"=>"en", "lat"=>38.338538, "lon"=>-90.973738),
"/US/US-MO/parma-12027" => array("type"=>"city", "pop"=>740, "name_ru"=>"Парма", "name_en"=>"Parma", "lang"=>"en", "lat"=>36.610923, "lon"=>-89.818238),
"/US/US-MO/parnell-12028" => array("type"=>"city", "pop"=>191, "name_ru"=>"Парнелл", "name_en"=>"Parnell", "lang"=>"en", "lat"=>40.440063, "lon"=>-94.621755),
"/US/US-MO/pasadena-hills-12029" => array("type"=>"city", "pop"=>1089, "name_ru"=>"Пасадена Хиллс", "name_en"=>"Pasadena Hills", "lang"=>"en", "lat"=>38.708567, "lon"=>-90.291365),
"/US/US-MO/pasadena-park-12030" => array("type"=>"village", "pop"=>458, "name_ru"=>"Пасадена Парк", "name_en"=>"Pasadena Park", "lang"=>"en", "lat"=>38.710200, "lon"=>-90.296927),
"/US/US-MO/pascola-12031" => array("type"=>"town", "pop"=>131, "name_ru"=>"Паскола", "name_en"=>"Pascola", "lang"=>"en", "lat"=>36.266066, "lon"=>-89.826508),
"/US/US-MO/passaic-12032" => array("type"=>"town", "pop"=>41, "name_ru"=>"Пассаик", "name_en"=>"Passaic", "lang"=>"en", "lat"=>38.321552, "lon"=>-94.347885),
"/US/US-MO/pattonsburg-12033" => array("type"=>"city", "pop"=>255, "name_ru"=>"Паттонсбург", "name_en"=>"Pattonsburg", "lang"=>"en", "lat"=>40.047962, "lon"=>-94.137564),
"/US/US-MO/paynesville-12034" => array("type"=>"village", "pop"=>93, "name_ru"=>"Пейнсвилл", "name_en"=>"Paynesville", "lang"=>"en", "lat"=>39.262177, "lon"=>-90.899648),
"/US/US-MO/peculiar-12035" => array("type"=>"city", "pop"=>4649, "name_ru"=>"Пекьюлиар", "name_en"=>"Peculiar", "lang"=>"en", "lat"=>38.723094, "lon"=>-94.457858),
"/US/US-MO/peerless-park-12036" => array("type"=>"village", "pop"=>33, "name_ru"=>"Пирлесс Парк", "name_en"=>"Peerless Park", "lang"=>"en", "lat"=>38.542400, "lon"=>-90.501053),
"/US/US-MO/penermon-12037" => array("type"=>"village", "pop"=>75, "name_ru"=>"Пенермон", "name_en"=>"Penermon", "lang"=>"en", "lat"=>36.793349, "lon"=>-89.829566),
"/US/US-MO/perry-12038" => array("type"=>"city", "pop"=>663, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>39.429952, "lon"=>-91.667929),
"/US/US-MO/perryville-12039" => array("type"=>"city", "pop"=>8172, "name_ru"=>"Перривилл", "name_en"=>"Perryville", "lang"=>"en", "lat"=>37.726092, "lon"=>-89.872177),
"/US/US-MO/pevely-12040" => array("type"=>"city", "pop"=>5808, "name_ru"=>"Пивли", "name_en"=>"Pevely", "lang"=>"en", "lat"=>38.289141, "lon"=>-90.398001),
"/US/US-MO/phelps-city-12041" => array("type"=>"town", "pop"=>32, "name_ru"=>"Фелпс Сити", "name_en"=>"Phelps City", "lang"=>"en", "lat"=>40.402675, "lon"=>-95.598122),
"/US/US-MO/phillipsburg-12042" => array("type"=>"town", "pop"=>226, "name_ru"=>"Филлипсбург", "name_en"=>"Phillipsburg", "lang"=>"en", "lat"=>37.553870, "lon"=>-92.785161),
"/US/US-MO/pickering-12043" => array("type"=>"town", "pop"=>149, "name_ru"=>"Пикеринг", "name_en"=>"Pickering", "lang"=>"en", "lat"=>40.450552, "lon"=>-94.841718),
"/US/US-MO/piedmont-12044" => array("type"=>"city", "pop"=>1914, "name_ru"=>"Пидмонт", "name_en"=>"Piedmont", "lang"=>"en", "lat"=>37.147823, "lon"=>-90.698403),
"/US/US-MO/pierce-city-12045" => array("type"=>"city", "pop"=>1465, "name_ru"=>"Пирс Сити", "name_en"=>"Pierce City", "lang"=>"en", "lat"=>36.946935, "lon"=>-94.002221),
"/US/US-MO/pilot-grove-12046" => array("type"=>"city", "pop"=>744, "name_ru"=>"Пайлот Гров", "name_en"=>"Pilot Grove", "lang"=>"en", "lat"=>38.873785, "lon"=>-92.912018),
"/US/US-MO/pilot-knob-12047" => array("type"=>"city", "pop"=>671, "name_ru"=>"Пайлот Ноб", "name_en"=>"Pilot Knob", "lang"=>"en", "lat"=>37.623524, "lon"=>-90.643573),
"/US/US-MO/pine-lawn-12048" => array("type"=>"city", "pop"=>3987, "name_ru"=>"Пайн Лон", "name_en"=>"Pine Lawn", "lang"=>"en", "lat"=>38.695100, "lon"=>-90.275395),
"/US/US-MO/pineville-12049" => array("type"=>"town", "pop"=>861, "name_ru"=>"Пайнвилл", "name_en"=>"Pineville", "lang"=>"en", "lat"=>36.593060, "lon"=>-94.383853),
"/US/US-MO/platte-city-12050" => array("type"=>"city", "pop"=>4806, "name_ru"=>"Платт Сити", "name_en"=>"Platte City", "lang"=>"en", "lat"=>39.361464, "lon"=>-94.777759),
"/US/US-MO/platte-woods-12051" => array("type"=>"city", "pop"=>457, "name_ru"=>"Платт Вудс", "name_en"=>"Platte Woods", "lang"=>"en", "lat"=>39.229300, "lon"=>-94.651962),
"/US/US-MO/plattsburg-12052" => array("type"=>"city", "pop"=>2417, "name_ru"=>"Платтсбург", "name_en"=>"Plattsburg", "lang"=>"en", "lat"=>39.565049, "lon"=>-94.462349),
"/US/US-MO/pleasant-hill-12053" => array("type"=>"city", "pop"=>7178, "name_ru"=>"Плезант Хилл", "name_en"=>"Pleasant Hill", "lang"=>"en", "lat"=>38.796552, "lon"=>-94.265686),
"/US/US-MO/pleasant-hope-12054" => array("type"=>"city", "pop"=>598, "name_ru"=>"Плезант Хоуп", "name_en"=>"Pleasant Hope", "lang"=>"en", "lat"=>37.461780, "lon"=>-93.273580),
"/US/US-MO/pleasant-valley-12055" => array("type"=>"city", "pop"=>3524, "name_ru"=>"Плезант Валли", "name_en"=>"Pleasant Valley", "lang"=>"en", "lat"=>39.218400, "lon"=>-94.481301),
"/US/US-MO/pocahontas-12056" => array("type"=>"town", "pop"=>128, "name_ru"=>"Покахонтас", "name_en"=>"Pocahontas", "lang"=>"en", "lat"=>37.501111, "lon"=>-89.639739),
"/US/US-MO/pollock-12057" => array("type"=>"village", "pop"=>121, "name_ru"=>"Поллок", "name_en"=>"Pollock", "lang"=>"en", "lat"=>40.359007, "lon"=>-93.084095),
"/US/US-MO/polo-12058" => array("type"=>"city", "pop"=>601, "name_ru"=>"Поло", "name_en"=>"Polo", "lang"=>"en", "lat"=>39.552194, "lon"=>-94.040045),
"/US/US-MO/poplar-bluff-12059" => array("type"=>"city", "pop"=>17096, "name_ru"=>"Поплар Блаф", "name_en"=>"Poplar Bluff", "lang"=>"en", "lat"=>36.763931, "lon"=>-90.407212),
"/US/US-MO/portage-des-sioux-12060" => array("type"=>"city", "pop"=>344, "name_ru"=>"Портедж Де Сиу", "name_en"=>"Portage Des Sioux", "lang"=>"en", "lat"=>38.926100, "lon"=>-90.341949),
"/US/US-MO/portageville-12061" => array("type"=>"city", "pop"=>2913, "name_ru"=>"Портеджвилл", "name_en"=>"Portageville", "lang"=>"en", "lat"=>36.428707, "lon"=>-89.699783),
"/US/US-MO/potosi-12062" => array("type"=>"city", "pop"=>2698, "name_ru"=>"Потоси", "name_en"=>"Potosi", "lang"=>"en", "lat"=>37.935997, "lon"=>-90.782128),
"/US/US-MO/powersville-12063" => array("type"=>"village", "pop"=>80, "name_ru"=>"Пауэрсвилл", "name_en"=>"Powersville", "lang"=>"en", "lat"=>40.549908, "lon"=>-93.300763),
"/US/US-MO/prairie-home-12064" => array("type"=>"city", "pop"=>226, "name_ru"=>"Прейри Ноум", "name_en"=>"Prairie Home", "lang"=>"en", "lat"=>38.814940, "lon"=>-92.589568),
"/US/US-MO/prathersville-12065" => array("type"=>"village", "pop"=>136, "name_ru"=>"Пратерсвилл", "name_en"=>"Prathersville", "lang"=>"en", "lat"=>39.314136, "lon"=>-94.276275),
"/US/US-MO/preston-12066" => array("type"=>"town", "pop"=>220, "name_ru"=>"Престон", "name_en"=>"Preston", "lang"=>"en", "lat"=>37.940076, "lon"=>-93.213544),
"/US/US-MO/princeton-12067" => array("type"=>"city", "pop"=>951, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>40.397137, "lon"=>-93.588098),
"/US/US-MO/purcell-12068" => array("type"=>"city", "pop"=>391, "name_ru"=>"Перселл", "name_en"=>"Purcell", "lang"=>"en", "lat"=>37.243292, "lon"=>-94.438818),
"/US/US-MO/purdin-12069" => array("type"=>"city", "pop"=>206, "name_ru"=>"Пурдин", "name_en"=>"Purdin", "lang"=>"en", "lat"=>39.950249, "lon"=>-93.166986),
"/US/US-MO/purdy-12070" => array("type"=>"city", "pop"=>1158, "name_ru"=>"Перди", "name_en"=>"Purdy", "lang"=>"en", "lat"=>36.818456, "lon"=>-93.920684),
"/US/US-MO/puxico-12071" => array("type"=>"city", "pop"=>1139, "name_ru"=>"Паксико", "name_en"=>"Puxico", "lang"=>"en", "lat"=>36.950663, "lon"=>-90.158995),
"/US/US-MO/queen-city-12072" => array("type"=>"city", "pop"=>611, "name_ru"=>"Квин Сити", "name_en"=>"Queen City", "lang"=>"en", "lat"=>40.413673, "lon"=>-92.566570),
"/US/US-MO/quitman-12073" => array("type"=>"town", "pop"=>45, "name_ru"=>"Квитман", "name_en"=>"Quitman", "lang"=>"en", "lat"=>40.373113, "lon"=>-95.076663),
"/US/US-MO/qulin-12074" => array("type"=>"city", "pop"=>479, "name_ru"=>"Кулин", "name_en"=>"Qulin", "lang"=>"en", "lat"=>36.597889, "lon"=>-90.249129),
"/US/US-MO/randolph-12075" => array("type"=>"village", "pop"=>51, "name_ru"=>"Рэндолф", "name_en"=>"Randolph", "lang"=>"en", "lat"=>39.156376, "lon"=>-94.490624),
"/US/US-MO/ravenwood-12076" => array("type"=>"town", "pop"=>444, "name_ru"=>"Рейвенвуд", "name_en"=>"Ravenwood", "lang"=>"en", "lat"=>40.352587, "lon"=>-94.671226),
"/US/US-MO/raymondville-12077" => array("type"=>"town", "pop"=>438, "name_ru"=>"Реймондвилл", "name_en"=>"Raymondville", "lang"=>"en", "lat"=>37.339746, "lon"=>-91.835815),
"/US/US-MO/raymore-12078" => array("type"=>"city", "pop"=>17703, "name_ru"=>"Реймор", "name_en"=>"Raymore", "lang"=>"en", "lat"=>38.811192, "lon"=>-94.457256),
"/US/US-MO/raytown-12079" => array("type"=>"city", "pop"=>28194, "name_ru"=>"Рейтаун", "name_en"=>"Raytown", "lang"=>"en", "lat"=>38.994867, "lon"=>-94.465214),
"/US/US-MO/rayville-12080" => array("type"=>"village", "pop"=>193, "name_ru"=>"Рейвилл", "name_en"=>"Rayville", "lang"=>"en", "lat"=>39.347978, "lon"=>-94.063590),
"/US/US-MO/rea-12081" => array("type"=>"town", "pop"=>57, "name_ru"=>"Ри", "name_en"=>"Rea", "lang"=>"en", "lat"=>40.061101, "lon"=>-94.764109),
"/US/US-MO/redings-mill-12082" => array("type"=>"village", "pop"=>167, "name_ru"=>"Редингс Милл", "name_en"=>"Redings Mill", "lang"=>"en", "lat"=>37.019668, "lon"=>-94.516542),
"/US/US-MO/reeds-12083" => array("type"=>"town", "pop"=>110, "name_ru"=>"Ридс", "name_en"=>"Reeds", "lang"=>"en", "lat"=>37.116488, "lon"=>-94.167454),
"/US/US-MO/reeds-spring-12084" => array("type"=>"city", "pop"=>756, "name_ru"=>"Ридс Спринг", "name_en"=>"Reeds Spring", "lang"=>"en", "lat"=>36.750065, "lon"=>-93.380846),
"/US/US-MO/renick-12085" => array("type"=>"village", "pop"=>236, "name_ru"=>"Реник", "name_en"=>"Renick", "lang"=>"en", "lat"=>39.342050, "lon"=>-92.410683),
"/US/US-MO/republic-12086" => array("type"=>"city", "pop"=>13715, "name_ru"=>"Репаблик", "name_en"=>"Republic", "lang"=>"en", "lat"=>37.117300, "lon"=>-93.475121),
"/US/US-MO/revere-12087" => array("type"=>"town", "pop"=>113, "name_ru"=>"Ревир", "name_en"=>"Revere", "lang"=>"en", "lat"=>40.494178, "lon"=>-91.676044),
"/US/US-MO/rhineland-12088" => array("type"=>"town", "pop"=>163, "name_ru"=>"Райнленд", "name_en"=>"Rhineland", "lang"=>"en", "lat"=>38.718605, "lon"=>-91.517290),
"/US/US-MO/rich-hill-12090" => array("type"=>"city", "pop"=>1494, "name_ru"=>"Рич Хилл", "name_en"=>"Rich Hill", "lang"=>"en", "lat"=>38.095751, "lon"=>-94.362986),
"/US/US-MO/richards-12089" => array("type"=>"town", "pop"=>93, "name_ru"=>"Ричардс", "name_en"=>"Richards", "lang"=>"en", "lat"=>37.909826, "lon"=>-94.557178),
"/US/US-MO/richland-12091" => array("type"=>"city", "pop"=>1783, "name_ru"=>"Ричленд", "name_en"=>"Richland", "lang"=>"en", "lat"=>37.861147, "lon"=>-92.399486),
"/US/US-MO/richmond-12092" => array("type"=>"city", "pop"=>5890, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>39.278651, "lon"=>-93.973342),
"/US/US-MO/richmond-heights-12093" => array("type"=>"city", "pop"=>9093, "name_ru"=>"Ричмонд Хайтс", "name_en"=>"Richmond Heights", "lang"=>"en", "lat"=>38.630167, "lon"=>-90.331730),
"/US/US-MO/ridgely-12094" => array("type"=>"town", "pop"=>97, "name_ru"=>"Риджли", "name_en"=>"Ridgely", "lang"=>"en", "lat"=>39.452162, "lon"=>-94.639532),
"/US/US-MO/ridgeway-12095" => array("type"=>"city", "pop"=>533, "name_ru"=>"Риджуэй", "name_en"=>"Ridgeway", "lang"=>"en", "lat"=>40.377784, "lon"=>-93.937715),
"/US/US-MO/risco-12096" => array("type"=>"city", "pop"=>351, "name_ru"=>"Риско", "name_en"=>"Risco", "lang"=>"en", "lat"=>36.551477, "lon"=>-89.818664),
"/US/US-MO/ritchey-12097" => array("type"=>"town", "pop"=>80, "name_ru"=>"Ритчи", "name_en"=>"Ritchey", "lang"=>"en", "lat"=>36.943744, "lon"=>-94.185674),
"/US/US-MO/rivermines-12098" => array("type"=>"village", "pop"=>459, "name_ru"=>"Ривермайнс", "name_en"=>"Rivermines", "lang"=>"en", "lat"=>37.844146, "lon"=>-90.529876),
"/US/US-MO/riverside-12099" => array("type"=>"city", "pop"=>2986, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>39.167300, "lon"=>-94.633653),
"/US/US-MO/riverview-12100" => array("type"=>"village", "pop"=>2918, "name_ru"=>"Ривервью", "name_en"=>"Riverview", "lang"=>"en", "lat"=>38.743700, "lon"=>-90.211728),
"/US/US-MO/rocheport-12101" => array("type"=>"city", "pop"=>202, "name_ru"=>"Рочепорт", "name_en"=>"Rocheport", "lang"=>"en", "lat"=>38.979473, "lon"=>-92.563313),
"/US/US-MO/rock-hill-12103" => array("type"=>"city", "pop"=>4580, "name_ru"=>"Рок Хилл", "name_en"=>"Rock Hill", "lang"=>"en", "lat"=>38.609000, "lon"=>-90.366930),
"/US/US-MO/rock-port-12104" => array("type"=>"city", "pop"=>1295, "name_ru"=>"Рок Порт", "name_en"=>"Rock Port", "lang"=>"en", "lat"=>40.411379, "lon"=>-95.516385),
"/US/US-MO/rockaway-beach-12102" => array("type"=>"city", "pop"=>594, "name_ru"=>"Рокуэй Бич", "name_en"=>"Rockaway Beach", "lang"=>"en", "lat"=>36.701882, "lon"=>-93.159020),
"/US/US-MO/rockville-12105" => array("type"=>"city", "pop"=>167, "name_ru"=>"Роквилл", "name_en"=>"Rockville", "lang"=>"en", "lat"=>38.070406, "lon"=>-94.080183),
"/US/US-MO/rocky-ridge-12106" => array("type"=>"village", "pop"=>362, "name_ru"=>"Роки Ридж", "name_en"=>"Rocky Ridge", "lang"=>"en", "lat"=>37.933441, "lon"=>-90.242678),
"/US/US-MO/rogersville-12107" => array("type"=>"town", "pop"=>3047, "name_ru"=>"Роджерсвилл", "name_en"=>"Rogersville", "lang"=>"en", "lat"=>37.115904, "lon"=>-93.059080),
"/US/US-MO/rolla-12108" => array("type"=>"city", "pop"=>18438, "name_ru"=>"Ролла", "name_en"=>"Rolla", "lang"=>"en", "lat"=>37.948735, "lon"=>-91.765760),
"/US/US-MO/roscoe-12109" => array("type"=>"village", "pop"=>107, "name_ru"=>"Роско", "name_en"=>"Roscoe", "lang"=>"en", "lat"=>37.976777, "lon"=>-93.812436),
"/US/US-MO/rosebud-12110" => array("type"=>"city", "pop"=>382, "name_ru"=>"Роузбад", "name_en"=>"Rosebud", "lang"=>"en", "lat"=>38.386554, "lon"=>-91.402972),
"/US/US-MO/rosendale-12111" => array("type"=>"city", "pop"=>182, "name_ru"=>"Розендейл", "name_en"=>"Rosendale", "lang"=>"en", "lat"=>40.042623, "lon"=>-94.824899),
"/US/US-MO/rothville-12112" => array("type"=>"village", "pop"=>85, "name_ru"=>"Ротвилл", "name_en"=>"Rothville", "lang"=>"en", "lat"=>39.654691, "lon"=>-93.060887),
"/US/US-MO/rush-hill-12113" => array("type"=>"town", "pop"=>130, "name_ru"=>"Раш Хилл", "name_en"=>"Rush Hill", "lang"=>"en", "lat"=>39.209904, "lon"=>-91.725121),
"/US/US-MO/rushville-12114" => array("type"=>"town", "pop"=>283, "name_ru"=>"Рашвилл", "name_en"=>"Rushville", "lang"=>"en", "lat"=>39.587288, "lon"=>-95.023304),
"/US/US-MO/russellville-12115" => array("type"=>"city", "pop"=>736, "name_ru"=>"Расселлвилл", "name_en"=>"Russellville", "lang"=>"en", "lat"=>38.512662, "lon"=>-92.437864),
"/US/US-MO/rutledge-12116" => array("type"=>"town", "pop"=>100, "name_ru"=>"Рутледж", "name_en"=>"Rutledge", "lang"=>"en", "lat"=>40.313037, "lon"=>-92.087387),
"/US/US-MO/saginaw-12117" => array("type"=>"village", "pop"=>309, "name_ru"=>"Сагино", "name_en"=>"Saginaw", "lang"=>"en", "lat"=>37.026423, "lon"=>-94.469537),
"/US/US-MO/salem-12135" => array("type"=>"city", "pop"=>4870, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>37.639916, "lon"=>-91.533329),
"/US/US-MO/salisbury-12136" => array("type"=>"city", "pop"=>1557, "name_ru"=>"Солсбери", "name_en"=>"Salisbury", "lang"=>"en", "lat"=>39.423360, "lon"=>-92.802348),
"/US/US-MO/sappington-12137" => array("type"=>"cdp", "pop"=>10917, "name_ru"=>"Сэппингтон", "name_en"=>"Sappington", "lang"=>"en", "lat"=>38.519900, "lon"=>-90.385613),
"/US/US-MO/sarcoxie-12138" => array("type"=>"city", "pop"=>1377, "name_ru"=>"Саркокси", "name_en"=>"Sarcoxie", "lang"=>"en", "lat"=>37.066782, "lon"=>-94.122305),
"/US/US-MO/savannah-12139" => array("type"=>"city", "pop"=>5075, "name_ru"=>"Саванна", "name_en"=>"Savannah", "lang"=>"en", "lat"=>39.938762, "lon"=>-94.827804),
"/US/US-MO/schell-city-12140" => array("type"=>"city", "pop"=>282, "name_ru"=>"Шелл Сити", "name_en"=>"Schell City", "lang"=>"en", "lat"=>38.019137, "lon"=>-94.116395),
"/US/US-MO/scotsdale-12141" => array("type"=>"town", "pop"=>209, "name_ru"=>"Скотсдейл", "name_en"=>"Scotsdale", "lang"=>"en", "lat"=>38.395950, "lon"=>-90.589671),
"/US/US-MO/scott-city-12142" => array("type"=>"city", "pop"=>4538, "name_ru"=>"Скотт Сити", "name_en"=>"Scott City", "lang"=>"en", "lat"=>37.222526, "lon"=>-89.532052),
"/US/US-MO/sedalia-12143" => array("type"=>"city", "pop"=>20990, "name_ru"=>"Седалия", "name_en"=>"Sedalia", "lang"=>"en", "lat"=>38.702981, "lon"=>-93.233272),
"/US/US-MO/sedgewickville-12144" => array("type"=>"village", "pop"=>192, "name_ru"=>"Седжвиквилл", "name_en"=>"Sedgewickville", "lang"=>"en", "lat"=>37.515078, "lon"=>-89.905943),
"/US/US-MO/seligman-12145" => array("type"=>"city", "pop"=>944, "name_ru"=>"Селигман", "name_en"=>"Seligman", "lang"=>"en", "lat"=>36.522699, "lon"=>-93.938028),
"/US/US-MO/senath-12146" => array("type"=>"city", "pop"=>1586, "name_ru"=>"Сенат", "name_en"=>"Senath", "lang"=>"en", "lat"=>36.133499, "lon"=>-90.161098),
"/US/US-MO/seneca-12147" => array("type"=>"city", "pop"=>2267, "name_ru"=>"Сенека", "name_en"=>"Seneca", "lang"=>"en", "lat"=>36.837883, "lon"=>-94.608492),
"/US/US-MO/seymour-12148" => array("type"=>"city", "pop"=>2073, "name_ru"=>"Сеймур", "name_en"=>"Seymour", "lang"=>"en", "lat"=>37.148526, "lon"=>-92.767107),
"/US/US-MO/shelbina-12149" => array("type"=>"city", "pop"=>1793, "name_ru"=>"Шелбина", "name_en"=>"Shelbina", "lang"=>"en", "lat"=>39.692274, "lon"=>-92.039684),
"/US/US-MO/shelbyville-12150" => array("type"=>"city", "pop"=>654, "name_ru"=>"Шелбивилл", "name_en"=>"Shelbyville", "lang"=>"en", "lat"=>39.807136, "lon"=>-92.039744),
"/US/US-MO/sheldon-12151" => array("type"=>"city", "pop"=>520, "name_ru"=>"Шелдон", "name_en"=>"Sheldon", "lang"=>"en", "lat"=>37.658604, "lon"=>-94.295675),
"/US/US-MO/sheridan-12152" => array("type"=>"town", "pop"=>161, "name_ru"=>"Шеридан", "name_en"=>"Sheridan", "lang"=>"en", "lat"=>40.517366, "lon"=>-94.614649),
"/US/US-MO/shoal-creek-drive-12153" => array("type"=>"village", "pop"=>358, "name_ru"=>"Шоал Крик Драйв", "name_en"=>"Shoal Creek Drive", "lang"=>"en", "lat"=>37.037510, "lon"=>-94.523239),
"/US/US-MO/shoal-creek-estates-12154" => array("type"=>"town", "pop"=>54, "name_ru"=>"Шоал Крик Истейтс", "name_en"=>"Shoal Creek Estates", "lang"=>"en", "lat"=>37.018750, "lon"=>-94.494107),
"/US/US-MO/shrewsbury-12155" => array("type"=>"city", "pop"=>6235, "name_ru"=>"Шрюсберри", "name_en"=>"Shrewsbury", "lang"=>"en", "lat"=>38.587050, "lon"=>-90.327920),
"/US/US-MO/sibley-12156" => array("type"=>"village", "pop"=>327, "name_ru"=>"Сибли", "name_en"=>"Sibley", "lang"=>"en", "lat"=>39.179200, "lon"=>-94.196916),
"/US/US-MO/sikeston-12157" => array("type"=>"city", "pop"=>17053, "name_ru"=>"Сайкстон", "name_en"=>"Sikeston", "lang"=>"en", "lat"=>36.880171, "lon"=>-89.580707),
"/US/US-MO/silex-12158" => array("type"=>"village", "pop"=>244, "name_ru"=>"Силекс", "name_en"=>"Silex", "lang"=>"en", "lat"=>39.126275, "lon"=>-91.057467),
"/US/US-MO/silver-creek-12159" => array("type"=>"village", "pop"=>638, "name_ru"=>"Силвер Крик", "name_en"=>"Silver Creek", "lang"=>"en", "lat"=>37.038750, "lon"=>-94.471899),
"/US/US-MO/skidmore-12160" => array("type"=>"city", "pop"=>331, "name_ru"=>"Скидмор", "name_en"=>"Skidmore", "lang"=>"en", "lat"=>40.287490, "lon"=>-95.078671),
"/US/US-MO/slater-12161" => array("type"=>"city", "pop"=>1900, "name_ru"=>"Слейтер", "name_en"=>"Slater", "lang"=>"en", "lat"=>39.222766, "lon"=>-93.064691),
"/US/US-MO/smithton-12162" => array("type"=>"city", "pop"=>518, "name_ru"=>"Смиттон", "name_en"=>"Smithton", "lang"=>"en", "lat"=>38.681619, "lon"=>-93.092234),
"/US/US-MO/smithville-12163" => array("type"=>"city", "pop"=>8079, "name_ru"=>"Смитвилл", "name_en"=>"Smithville", "lang"=>"en", "lat"=>39.381505, "lon"=>-94.574127),
"/US/US-MO/south-gifford-12164" => array("type"=>"village", "pop"=>72, "name_ru"=>"Саут Гиффорд", "name_en"=>"South Gifford", "lang"=>"en", "lat"=>40.025919, "lon"=>-92.682295),
"/US/US-MO/south-gorin-12165" => array("type"=>"town", "pop"=>139, "name_ru"=>"Саут Горин", "name_en"=>"South Gorin", "lang"=>"en", "lat"=>40.360065, "lon"=>-92.024102),
"/US/US-MO/south-greenfield-12166" => array("type"=>"village", "pop"=>128, "name_ru"=>"Саут Гринфилд", "name_en"=>"South Greenfield", "lang"=>"en", "lat"=>37.375861, "lon"=>-93.842263),
"/US/US-MO/south-lineville-12167" => array("type"=>"town", "pop"=>35, "name_ru"=>"Саут Лайнвилл", "name_en"=>"South Lineville", "lang"=>"en", "lat"=>40.577897, "lon"=>-93.523636),
"/US/US-MO/south-west-city-12168" => array("type"=>"town", "pop"=>600, "name_ru"=>"Саут Вэст Сити", "name_en"=>"South West City", "lang"=>"en", "lat"=>36.523694, "lon"=>-94.608627),
"/US/US-MO/spanish-lake-12169" => array("type"=>"cdp", "pop"=>20322, "name_ru"=>"Спэниш Лейк", "name_en"=>"Spanish Lake", "lang"=>"en", "lat"=>38.786900, "lon"=>-90.207699),
"/US/US-MO/sparta-12170" => array("type"=>"city", "pop"=>1199, "name_ru"=>"Спарта", "name_en"=>"Sparta", "lang"=>"en", "lat"=>37.001677, "lon"=>-93.083196),
"/US/US-MO/spickard-12171" => array("type"=>"city", "pop"=>304, "name_en"=>"Spickard", "lang"=>"en", "lat"=>40.242628, "lon"=>-93.592173),
"/US/US-MO/springfield-12172" => array("type"=>"city", "pop"=>156206, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>37.196200, "lon"=>-93.286099),
"/US/US-MO/st.-ann-12118" => array("type"=>"city", "pop"=>12766, "name_en"=>"St. Ann", "lang"=>"en", "lat"=>38.725850, "lon"=>-90.387779),
"/US/US-MO/st.-charles-12119" => array("type"=>"city", "pop"=>64386, "name_ru"=>"Сент-Чарльз", "name_en"=>"St. Charles", "lang"=>"en", "lat"=>38.790550, "lon"=>-90.516394),
"/US/US-MO/st.-clair-12120" => array("type"=>"city", "pop"=>4472, "name_ru"=>"Сент Клэр", "name_en"=>"St. Clair", "lang"=>"en", "lat"=>38.351092, "lon"=>-90.980854),
"/US/US-MO/st.-cloud-12121" => array("type"=>"village", "pop"=>59, "name_ru"=>"Сент Клауд", "name_en"=>"St. Cloud", "lang"=>"en", "lat"=>38.172521, "lon"=>-91.212560),
"/US/US-MO/st.-elizabeth-12123" => array("type"=>"town", "pop"=>312, "name_en"=>"St. Elizabeth", "lang"=>"en", "lat"=>38.256207, "lon"=>-92.266041),
"/US/US-MO/st.-george-12124" => array("type"=>"city", "pop"=>1212, "name_ru"=>"Сент-Джрдж", "name_en"=>"St. George", "lang"=>"en", "lat"=>38.537800, "lon"=>-90.313142),
"/US/US-MO/st.-james-12125" => array("type"=>"city", "pop"=>4013, "name_ru"=>"Сент Джеймс", "name_en"=>"St. James", "lang"=>"en", "lat"=>38.000998, "lon"=>-91.613989),
"/US/US-MO/st.-john-12126" => array("type"=>"city", "pop"=>6381, "name_ru"=>"Сент Джон", "name_en"=>"St. John", "lang"=>"en", "lat"=>38.714500, "lon"=>-90.346097),
"/US/US-MO/st.-joseph-12127" => array("type"=>"city", "pop"=>76197, "name_ru"=>"Сент Джозеф", "name_en"=>"St. Joseph", "lang"=>"en", "lat"=>39.759200, "lon"=>-94.821149),
"/US/US-MO/st.-louis-12128" => array("type"=>"city", "pop"=>354361, "name_ru"=>"Сент Луис", "name_en"=>"St. Louis", "lang"=>"en", "lat"=>38.636050, "lon"=>-90.244299),
"/US/US-MO/st.-martins-12129" => array("type"=>"city", "pop"=>1122, "name_en"=>"St. Martins", "lang"=>"en", "lat"=>38.595722, "lon"=>-92.333338),
"/US/US-MO/st.-mary-12130" => array("type"=>"city", "pop"=>380, "name_en"=>"St. Mary", "lang"=>"en", "lat"=>37.876100, "lon"=>-89.948700),
"/US/US-MO/st.-paul-12131" => array("type"=>"village", "pop"=>1662, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>38.849900, "lon"=>-90.736647),
"/US/US-MO/st.-peters-12132" => array("type"=>"city", "pop"=>55500, "name_en"=>"St. Peters", "lang"=>"en", "lat"=>38.778475, "lon"=>-90.605281),
"/US/US-MO/st.-robert-12133" => array("type"=>"city", "pop"=>3456, "name_en"=>"St. Robert", "lang"=>"en", "lat"=>37.821311, "lon"=>-92.155588),
"/US/US-MO/st.-thomas-12134" => array("type"=>"town", "pop"=>285, "name_en"=>"St. Thomas", "lang"=>"en", "lat"=>38.368813, "lon"=>-92.215627),
"/US/US-MO/stanberry-12173" => array("type"=>"city", "pop"=>1124, "name_en"=>"Stanberry", "lang"=>"en", "lat"=>40.216686, "lon"=>-94.537975),
"/US/US-MO/stark-city-12174" => array("type"=>"town", "pop"=>164, "name_en"=>"Stark City", "lang"=>"en", "lat"=>36.860218, "lon"=>-94.188612),
"/US/US-MO/ste.-genevieve-12122" => array("type"=>"city", "pop"=>4418, "name_en"=>"Ste. Genevieve", "lang"=>"en", "lat"=>37.974925, "lon"=>-90.049497),
"/US/US-MO/steele-12175" => array("type"=>"city", "pop"=>2093, "name_ru"=>"Стил", "name_en"=>"Steele", "lang"=>"en", "lat"=>36.084997, "lon"=>-89.846744),
"/US/US-MO/steelville-12176" => array("type"=>"city", "pop"=>1493, "name_en"=>"Steelville", "lang"=>"en", "lat"=>37.969437, "lon"=>-91.355418),
"/US/US-MO/stella-12177" => array("type"=>"town", "pop"=>187, "name_en"=>"Stella", "lang"=>"en", "lat"=>36.762301, "lon"=>-94.190293),
"/US/US-MO/stewartsville-12178" => array("type"=>"city", "pop"=>742, "name_en"=>"Stewartsville", "lang"=>"en", "lat"=>39.754392, "lon"=>-94.498700),
"/US/US-MO/stockton-12179" => array("type"=>"city", "pop"=>2016, "name_ru"=>"Стоктон", "name_en"=>"Stockton", "lang"=>"en", "lat"=>37.696777, "lon"=>-93.795661),
"/US/US-MO/stotesbury-12180" => array("type"=>"town", "pop"=>42, "name_en"=>"Stotesbury", "lang"=>"en", "lat"=>37.974435, "lon"=>-94.564748),
"/US/US-MO/stotts-city-12181" => array("type"=>"city", "pop"=>265, "name_en"=>"Stotts City", "lang"=>"en", "lat"=>37.101941, "lon"=>-93.947597),
"/US/US-MO/stoutland-12182" => array("type"=>"village", "pop"=>195, "name_en"=>"Stoutland", "lang"=>"en", "lat"=>37.812981, "lon"=>-92.514288),
"/US/US-MO/stoutsville-12183" => array("type"=>"village", "pop"=>42, "name_en"=>"Stoutsville", "lang"=>"en", "lat"=>39.551684, "lon"=>-91.852930),
"/US/US-MO/stover-12184" => array("type"=>"city", "pop"=>1043, "name_en"=>"Stover", "lang"=>"en", "lat"=>38.442175, "lon"=>-92.989741),
"/US/US-MO/strafford-12185" => array("type"=>"city", "pop"=>2178, "name_en"=>"Strafford", "lang"=>"en", "lat"=>37.269776, "lon"=>-93.117539),
"/US/US-MO/strasburg-12186" => array("type"=>"town", "pop"=>136, "name_ru"=>"Страсбург", "name_en"=>"Strasburg", "lang"=>"en", "lat"=>38.759240, "lon"=>-94.164699),
"/US/US-MO/sturgeon-12187" => array("type"=>"city", "pop"=>915, "name_en"=>"Sturgeon", "lang"=>"en", "lat"=>39.233841, "lon"=>-92.281301),
"/US/US-MO/sugar-creek-12188" => array("type"=>"city", "pop"=>3533, "name_en"=>"Sugar Creek", "lang"=>"en", "lat"=>39.138132, "lon"=>-94.411166),
"/US/US-MO/sullivan-12189" => array("type"=>"city", "pop"=>6726, "name_ru"=>"Салливан", "name_en"=>"Sullivan", "lang"=>"en", "lat"=>38.211040, "lon"=>-91.163238),
"/US/US-MO/summersville-12190" => array("type"=>"city", "pop"=>554, "name_en"=>"Summersville", "lang"=>"en", "lat"=>37.178092, "lon"=>-91.658377),
"/US/US-MO/sumner-12191" => array("type"=>"town", "pop"=>131, "name_ru"=>"Саммер", "name_en"=>"Sumner", "lang"=>"en", "lat"=>39.655657, "lon"=>-93.243235),
"/US/US-MO/sundown-12192" => array("type"=>"village", "pop"=>35, "name_en"=>"Sundown", "lang"=>"en", "lat"=>36.565528, "lon"=>-92.637305),
"/US/US-MO/sunrise-beach-12193" => array("type"=>"village", "pop"=>426, "name_en"=>"Sunrise Beach", "lang"=>"en", "lat"=>38.179680, "lon"=>-92.772375),
"/US/US-MO/sunset-hills-12194" => array("type"=>"city", "pop"=>8190, "name_en"=>"Sunset Hills", "lang"=>"en", "lat"=>38.537333, "lon"=>-90.416342),
"/US/US-MO/sweet-springs-12195" => array("type"=>"city", "pop"=>1517, "name_en"=>"Sweet Springs", "lang"=>"en", "lat"=>38.964238, "lon"=>-93.416398),
"/US/US-MO/sycamore-hills-12196" => array("type"=>"village", "pop"=>685, "name_en"=>"Sycamore Hills", "lang"=>"en", "lat"=>38.701000, "lon"=>-90.348801),
"/US/US-MO/syracuse-12197" => array("type"=>"city", "pop"=>187, "name_ru"=>"Сиракьюс", "name_en"=>"Syracuse", "lang"=>"en", "lat"=>38.669188, "lon"=>-92.875736),
"/US/US-MO/table-rock-12198" => array("type"=>"village", "pop"=>100, "name_en"=>"Table Rock", "lang"=>"en", "lat"=>36.604273, "lon"=>-93.300281),
"/US/US-MO/tallapoosa-12199" => array("type"=>"city", "pop"=>182, "name_ru"=>"Таллапуза", "name_en"=>"Tallapoosa", "lang"=>"en", "lat"=>36.507499, "lon"=>-89.818690),
"/US/US-MO/taneyville-12200" => array("type"=>"village", "pop"=>367, "name_en"=>"Taneyville", "lang"=>"en", "lat"=>36.737466, "lon"=>-93.034744),
"/US/US-MO/taos-12201" => array("type"=>"city", "pop"=>863, "name_en"=>"Taos", "lang"=>"en", "lat"=>38.497818, "lon"=>-92.080785),
"/US/US-MO/tarkio-12202" => array("type"=>"city", "pop"=>1806, "name_en"=>"Tarkio", "lang"=>"en", "lat"=>40.443063, "lon"=>-95.383766),
"/US/US-MO/tarrants-12203" => array("type"=>"village", "pop"=>43, "name_en"=>"Tarrants", "lang"=>"en", "lat"=>39.358067, "lon"=>-91.183427),
"/US/US-MO/thayer-12204" => array("type"=>"city", "pop"=>2151, "name_ru"=>"Тайер", "name_en"=>"Thayer", "lang"=>"en", "lat"=>36.524636, "lon"=>-91.541148),
"/US/US-MO/the-landing-12205" => array("type"=>"village", "pop"=>8, "name_en"=>"The Landing", "lang"=>"en", "lat"=>39.559407, "lon"=>-91.658977),
"/US/US-MO/theodosia-12206" => array("type"=>"village", "pop"=>249, "name_en"=>"Theodosia", "lang"=>"en", "lat"=>36.579884, "lon"=>-92.665127),
"/US/US-MO/tightwad-12207" => array("type"=>"village", "pop"=>64, "name_en"=>"Tightwad", "lang"=>"en", "lat"=>38.311141, "lon"=>-93.541663),
"/US/US-MO/tina-12208" => array("type"=>"town", "pop"=>185, "name_en"=>"Tina", "lang"=>"en", "lat"=>39.537193, "lon"=>-93.441260),
"/US/US-MO/tindall-12209" => array("type"=>"town", "pop"=>63, "name_en"=>"Tindall", "lang"=>"en", "lat"=>40.160419, "lon"=>-93.609403),
"/US/US-MO/tipton-12210" => array("type"=>"city", "pop"=>3281, "name_ru"=>"Типтон", "name_en"=>"Tipton", "lang"=>"en", "lat"=>38.653085, "lon"=>-92.780004),
"/US/US-MO/town-and-country-12211" => array("type"=>"city", "pop"=>10710, "name_en"=>"Town and Country", "lang"=>"en", "lat"=>38.624750, "lon"=>-90.476329),
"/US/US-MO/tracy-12212" => array("type"=>"city", "pop"=>209, "name_ru"=>"Трейси", "name_en"=>"Tracy", "lang"=>"en", "lat"=>39.378572, "lon"=>-94.792995),
"/US/US-MO/trenton-12213" => array("type"=>"city", "pop"=>6047, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>40.081148, "lon"=>-93.604496),
"/US/US-MO/trimble-12214" => array("type"=>"city", "pop"=>475, "name_en"=>"Trimble", "lang"=>"en", "lat"=>39.474618, "lon"=>-94.561028),
"/US/US-MO/triplett-12215" => array("type"=>"city", "pop"=>59, "name_en"=>"Triplett", "lang"=>"en", "lat"=>39.497681, "lon"=>-93.193521),
"/US/US-MO/troy-12216" => array("type"=>"city", "pop"=>12270, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>38.976901, "lon"=>-90.975354),
"/US/US-MO/truesdale-12217" => array("type"=>"city", "pop"=>661, "name_ru"=>"Трусдейл", "name_en"=>"Truesdale", "lang"=>"en", "lat"=>38.811010, "lon"=>-91.125069),
"/US/US-MO/truxton-12218" => array("type"=>"village", "pop"=>119, "name_en"=>"Truxton", "lang"=>"en", "lat"=>39.002199, "lon"=>-91.240612),
"/US/US-MO/turney-12219" => array("type"=>"village", "pop"=>157, "name_en"=>"Turney", "lang"=>"en", "lat"=>39.636911, "lon"=>-94.320589),
"/US/US-MO/tuscumbia-12220" => array("type"=>"town", "pop"=>224, "name_ru"=>"Тускумбия", "name_en"=>"Tuscumbia", "lang"=>"en", "lat"=>38.235407, "lon"=>-92.459524),
"/US/US-MO/twin-bridges-12221" => array("type"=>"town", "pop"=>44, "name_en"=>"Twin Bridges", "lang"=>"en", "lat"=>37.552166, "lon"=>-92.597051),
"/US/US-MO/twin-oaks-12222" => array("type"=>"village", "pop"=>346, "name_en"=>"Twin Oaks", "lang"=>"en", "lat"=>38.566483, "lon"=>-90.500487),
"/US/US-MO/umber-view-heights-12223" => array("type"=>"village", "pop"=>57, "name_en"=>"Umber View Heights", "lang"=>"en", "lat"=>37.624195, "lon"=>-93.803846),
"/US/US-MO/union-12224" => array("type"=>"city", "pop"=>9684, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>38.448260, "lon"=>-91.012618),
"/US/US-MO/union-star-12225" => array("type"=>"town", "pop"=>397, "name_en"=>"Union Star", "lang"=>"en", "lat"=>39.979439, "lon"=>-94.598047),
"/US/US-MO/unionville-12226" => array("type"=>"city", "pop"=>1865, "name_ru"=>"Юнионвилл", "name_en"=>"Unionville", "lang"=>"en", "lat"=>40.475683, "lon"=>-93.003693),
"/US/US-MO/unity-village-12227" => array("type"=>"village", "pop"=>114, "name_en"=>"Unity Village", "lang"=>"en", "lat"=>38.950550, "lon"=>-94.401487),
"/US/US-MO/university-city-12228" => array("type"=>"city", "pop"=>36289, "name_en"=>"University City", "lang"=>"en", "lat"=>38.666233, "lon"=>-90.331081),
"/US/US-MO/uplands-park-12229" => array("type"=>"village", "pop"=>437, "name_en"=>"Uplands Park", "lang"=>"en", "lat"=>38.692700, "lon"=>-90.282465),
"/US/US-MO/urbana-12230" => array("type"=>"city", "pop"=>433, "name_ru"=>"Урбана", "name_en"=>"Urbana", "lang"=>"en", "lat"=>37.843990, "lon"=>-93.166124),
"/US/US-MO/urich-12231" => array("type"=>"city", "pop"=>501, "name_en"=>"Urich", "lang"=>"en", "lat"=>38.459931, "lon"=>-93.998990),
"/US/US-MO/utica-12232" => array("type"=>"village", "pop"=>262, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>39.745593, "lon"=>-93.629276),
"/US/US-MO/valley-park-12233" => array("type"=>"city", "pop"=>6274, "name_en"=>"Valley Park", "lang"=>"en", "lat"=>38.554550, "lon"=>-90.482384),
"/US/US-MO/van-buren-12234" => array("type"=>"town", "pop"=>815, "name_ru"=>"Ван Бурен", "name_en"=>"Van Buren", "lang"=>"en", "lat"=>37.007324, "lon"=>-91.012218),
"/US/US-MO/vandalia-12235" => array("type"=>"city", "pop"=>4342, "name_ru"=>"Вандалия", "name_en"=>"Vandalia", "lang"=>"en", "lat"=>39.309279, "lon"=>-91.490290),
"/US/US-MO/vandiver-12236" => array("type"=>"village", "pop"=>81, "name_en"=>"Vandiver", "lang"=>"en", "lat"=>39.161279, "lon"=>-91.848762),
"/US/US-MO/vanduser-12237" => array("type"=>"village", "pop"=>209, "name_en"=>"Vanduser", "lang"=>"en", "lat"=>36.991231, "lon"=>-89.686462),
"/US/US-MO/velda-village-12238" => array("type"=>"city", "pop"=>1597, "name_en"=>"Velda Village", "lang"=>"en", "lat"=>38.694000, "lon"=>-90.294050),
"/US/US-MO/velda-village-hills-12239" => array("type"=>"village", "pop"=>1035, "name_en"=>"Velda Village Hills", "lang"=>"en", "lat"=>38.692300, "lon"=>-90.286874),
"/US/US-MO/verona-12240" => array("type"=>"town", "pop"=>724, "name_ru"=>"Верона", "name_en"=>"Verona", "lang"=>"en", "lat"=>36.962953, "lon"=>-93.794140),
"/US/US-MO/versailles-12241" => array("type"=>"city", "pop"=>2715, "name_ru"=>"Версайллес", "name_en"=>"Versailles", "lang"=>"en", "lat"=>38.433224, "lon"=>-92.846046),
"/US/US-MO/viburnum-12242" => array("type"=>"town", "pop"=>783, "name_en"=>"Viburnum", "lang"=>"en", "lat"=>37.715613, "lon"=>-91.129153),
"/US/US-MO/vienna-12243" => array("type"=>"city", "pop"=>638, "name_ru"=>"Вьенна", "name_en"=>"Vienna", "lang"=>"en", "lat"=>38.187528, "lon"=>-91.947968),
"/US/US-MO/villa-ridge-12245" => array("type"=>"cdp", "pop"=>1865, "name_en"=>"Villa Ridge", "lang"=>"en", "lat"=>38.471173, "lon"=>-90.883525),
"/US/US-MO/village-of-four-seasons-12244" => array("type"=>"town", "pop"=>1610, "name_en"=>"Village of Four Seasons", "lang"=>"en", "lat"=>38.198178, "lon"=>-92.715451),
"/US/US-MO/vinita-park-12246" => array("type"=>"city", "pop"=>1784, "name_en"=>"Vinita Park", "lang"=>"en", "lat"=>38.690021, "lon"=>-90.339656),
"/US/US-MO/vinita-terrace-12247" => array("type"=>"village", "pop"=>278, "name_en"=>"Vinita Terrace", "lang"=>"en", "lat"=>38.684950, "lon"=>-90.329674),
"/US/US-MO/vista-12248" => array("type"=>"village", "pop"=>52, "name_ru"=>"Виста", "name_en"=>"Vista", "lang"=>"en", "lat"=>37.988793, "lon"=>-93.663829),
"/US/US-MO/waco-12249" => array("type"=>"town", "pop"=>92, "name_ru"=>"Вако", "name_en"=>"Waco", "lang"=>"en", "lat"=>37.246271, "lon"=>-94.599775),
"/US/US-MO/wakenda-12250" => array("type"=>"town", "pop"=>448, "name_en"=>"Wakenda", "lang"=>"en", "lat"=>39.315512, "lon"=>-93.377895),
"/US/US-MO/walker-12251" => array("type"=>"city", "pop"=>272, "name_ru"=>"Уолкер", "name_en"=>"Walker", "lang"=>"en", "lat"=>37.899484, "lon"=>-94.230373),
"/US/US-MO/walnut-grove-12252" => array("type"=>"city", "pop"=>667, "name_ru"=>"Уолнат Гров", "name_en"=>"Walnut Grove", "lang"=>"en", "lat"=>37.411862, "lon"=>-93.548207),
"/US/US-MO/wardell-12253" => array("type"=>"town", "pop"=>400, "name_en"=>"Wardell", "lang"=>"en", "lat"=>36.350698, "lon"=>-89.816886),
"/US/US-MO/wardsville-12254" => array("type"=>"town", "pop"=>996, "name_en"=>"Wardsville", "lang"=>"en", "lat"=>38.487970, "lon"=>-92.173402),
"/US/US-MO/warrensburg-12255" => array("type"=>"city", "pop"=>19034, "name_ru"=>"Уорренсбург", "name_en"=>"Warrensburg", "lang"=>"en", "lat"=>38.759819, "lon"=>-93.728640),
"/US/US-MO/warrenton-12256" => array("type"=>"city", "pop"=>7398, "name_ru"=>"Уоррентон", "name_en"=>"Warrenton", "lang"=>"en", "lat"=>38.821258, "lon"=>-91.140098),
"/US/US-MO/warsaw-12257" => array("type"=>"city", "pop"=>2222, "name_ru"=>"Варшава", "name_en"=>"Warsaw", "lang"=>"en", "lat"=>38.242981, "lon"=>-93.371317),
"/US/US-MO/warson-woods-12258" => array("type"=>"city", "pop"=>1865, "name_en"=>"Warson Woods", "lang"=>"en", "lat"=>38.606867, "lon"=>-90.391522),
"/US/US-MO/washburn-12259" => array("type"=>"city", "pop"=>482, "name_ru"=>"Уошберн", "name_en"=>"Washburn", "lang"=>"en", "lat"=>36.589501, "lon"=>-93.966324),
"/US/US-MO/washington-12260" => array("type"=>"city", "pop"=>14392, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>38.553148, "lon"=>-91.012640),
"/US/US-MO/watson-12261" => array("type"=>"village", "pop"=>114, "name_ru"=>"Уотсон", "name_en"=>"Watson", "lang"=>"en", "lat"=>40.480143, "lon"=>-95.623146),
"/US/US-MO/waverly-12262" => array("type"=>"city", "pop"=>787, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>39.206632, "lon"=>-93.519927),
"/US/US-MO/wayland-12263" => array("type"=>"city", "pop"=>397, "name_ru"=>"Уэйленд", "name_en"=>"Wayland", "lang"=>"en", "lat"=>40.395154, "lon"=>-91.581718),
"/US/US-MO/waynesville-12264" => array("type"=>"city", "pop"=>3960, "name_ru"=>"Уэйнсвилл", "name_en"=>"Waynesville", "lang"=>"en", "lat"=>37.827589, "lon"=>-92.216933),
"/US/US-MO/weatherby-12265" => array("type"=>"town", "pop"=>116, "name_en"=>"Weatherby", "lang"=>"en", "lat"=>39.909232, "lon"=>-94.241307),
"/US/US-MO/weatherby-lake-12266" => array("type"=>"city", "pop"=>1863, "name_en"=>"Weatherby Lake", "lang"=>"en", "lat"=>39.238600, "lon"=>-94.697642),
"/US/US-MO/weaubleau-12267" => array("type"=>"city", "pop"=>524, "name_en"=>"Weaubleau", "lang"=>"en", "lat"=>37.891614, "lon"=>-93.540784),
"/US/US-MO/webb-city-12268" => array("type"=>"city", "pop"=>11532, "name_en"=>"Webb City", "lang"=>"en", "lat"=>37.141250, "lon"=>-94.471249),
"/US/US-MO/webster-groves-12269" => array("type"=>"city", "pop"=>22335, "name_en"=>"Webster Groves", "lang"=>"en", "lat"=>38.585600, "lon"=>-90.354588),
"/US/US-MO/weldon-spring-12270" => array("type"=>"town", "pop"=>5295, "name_en"=>"Weldon Spring", "lang"=>"en", "lat"=>38.716500, "lon"=>-90.648981),
"/US/US-MO/weldon-spring-heights-12271" => array("type"=>"town", "pop"=>73, "name_en"=>"Weldon Spring Heights", "lang"=>"en", "lat"=>38.704131, "lon"=>-90.684555),
"/US/US-MO/wellington-12272" => array("type"=>"city", "pop"=>763, "name_ru"=>"Веллингтон", "name_en"=>"Wellington", "lang"=>"en", "lat"=>39.136200, "lon"=>-93.987706),
"/US/US-MO/wellston-12273" => array("type"=>"city", "pop"=>2311, "name_en"=>"Wellston", "lang"=>"en", "lat"=>38.674900, "lon"=>-90.293503),
"/US/US-MO/wellsville-12274" => array("type"=>"city", "pop"=>1324, "name_ru"=>"Уэллсвилл", "name_en"=>"Wellsville", "lang"=>"en", "lat"=>39.073742, "lon"=>-91.567912),
"/US/US-MO/wentworth-12275" => array("type"=>"village", "pop"=>148, "name_en"=>"Wentworth", "lang"=>"en", "lat"=>36.993319, "lon"=>-94.074939),
"/US/US-MO/wentzville-12276" => array("type"=>"city", "pop"=>23768, "name_en"=>"Wentzville", "lang"=>"en", "lat"=>38.813367, "lon"=>-90.859549),
"/US/US-MO/west-line-12278" => array("type"=>"village", "pop"=>131, "name_en"=>"West Line", "lang"=>"en", "lat"=>38.632623, "lon"=>-94.585129),
"/US/US-MO/west-plains-12281" => array("type"=>"city", "pop"=>12045, "name_en"=>"West Plains", "lang"=>"en", "lat"=>36.741352, "lon"=>-91.862995),
"/US/US-MO/westboro-12277" => array("type"=>"city", "pop"=>156, "name_en"=>"Westboro", "lang"=>"en", "lat"=>40.535779, "lon"=>-95.321866),
"/US/US-MO/weston-12279" => array("type"=>"city", "pop"=>1663, "name_ru"=>"Вэстон", "name_en"=>"Weston", "lang"=>"en", "lat"=>39.414596, "lon"=>-94.896518),
"/US/US-MO/westphalia-12280" => array("type"=>"city", "pop"=>328, "name_ru"=>"Вэстфалия", "name_en"=>"Westphalia", "lang"=>"en", "lat"=>38.441218, "lon"=>-92.000488),
"/US/US-MO/westwood-12282" => array("type"=>"village", "pop"=>289, "name_ru"=>"Уэствуд", "name_en"=>"Westwood", "lang"=>"en", "lat"=>38.646900, "lon"=>-90.434367),
"/US/US-MO/wheatland-12283" => array("type"=>"city", "pop"=>392, "name_ru"=>"Уитленд", "name_en"=>"Wheatland", "lang"=>"en", "lat"=>37.943464, "lon"=>-93.403909),
"/US/US-MO/wheaton-12284" => array("type"=>"city", "pop"=>749, "name_ru"=>"Уитон", "name_en"=>"Wheaton", "lang"=>"en", "lat"=>36.761310, "lon"=>-94.056195),
"/US/US-MO/wheeling-12285" => array("type"=>"city", "pop"=>255, "name_ru"=>"Уилинг", "name_en"=>"Wheeling", "lang"=>"en", "lat"=>39.786588, "lon"=>-93.386427),
"/US/US-MO/whiteman-afb-12286" => array("type"=>"cdp", "pop"=>4174, "name_en"=>"Whiteman AFB", "lang"=>"en", "lat"=>38.732745, "lon"=>-93.554640),
"/US/US-MO/whiteside-12287" => array("type"=>"village", "pop"=>84, "name_en"=>"Whiteside", "lang"=>"en", "lat"=>39.183400, "lon"=>-91.016535),
"/US/US-MO/whitewater-12288" => array("type"=>"town", "pop"=>117, "name_ru"=>"Уайтуотер", "name_en"=>"Whitewater", "lang"=>"en", "lat"=>37.236620, "lon"=>-89.797684),
"/US/US-MO/wilbur-park-12289" => array("type"=>"village", "pop"=>454, "name_en"=>"Wilbur Park", "lang"=>"en", "lat"=>38.552726, "lon"=>-90.307311),
"/US/US-MO/willard-12290" => array("type"=>"city", "pop"=>3336, "name_ru"=>"Уиллард", "name_en"=>"Willard", "lang"=>"en", "lat"=>37.293550, "lon"=>-93.410946),
"/US/US-MO/williamsville-12291" => array("type"=>"city", "pop"=>361, "name_ru"=>"Уильямсвилл", "name_en"=>"Williamsville", "lang"=>"en", "lat"=>36.973595, "lon"=>-90.548054),
"/US/US-MO/willow-springs-12292" => array("type"=>"city", "pop"=>2159, "name_ru"=>"Уиллоу Спрингс", "name_en"=>"Willow Springs", "lang"=>"en", "lat"=>36.994292, "lon"=>-91.969943),
"/US/US-MO/wilson-city-12293" => array("type"=>"town", "pop"=>162, "name_en"=>"Wilson City", "lang"=>"en", "lat"=>36.922989, "lon"=>-89.222828),
"/US/US-MO/winchester-12294" => array("type"=>"city", "pop"=>1545, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>38.589950, "lon"=>-90.525634),
"/US/US-MO/windsor-12295" => array("type"=>"city", "pop"=>3186, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>38.533359, "lon"=>-93.522358),
"/US/US-MO/winfield-12296" => array("type"=>"city", "pop"=>1138, "name_ru"=>"Винфилд", "name_en"=>"Winfield", "lang"=>"en", "lat"=>38.995448, "lon"=>-90.739752),
"/US/US-MO/winona-12297" => array("type"=>"city", "pop"=>1326, "name_ru"=>"Уайнона", "name_en"=>"Winona", "lang"=>"en", "lat"=>37.006341, "lon"=>-91.332055),
"/US/US-MO/winston-12298" => array("type"=>"town", "pop"=>245, "name_ru"=>"Уинстон", "name_en"=>"Winston", "lang"=>"en", "lat"=>39.869911, "lon"=>-94.141509),
"/US/US-MO/woods-heights-12299" => array("type"=>"city", "pop"=>708, "name_en"=>"Woods Heights", "lang"=>"en", "lat"=>39.340005, "lon"=>-94.162246),
"/US/US-MO/woodson-terrace-12300" => array("type"=>"city", "pop"=>4006, "name_en"=>"Woodson Terrace", "lang"=>"en", "lat"=>38.728400, "lon"=>-90.360138),
"/US/US-MO/wooldridge-12301" => array("type"=>"town", "pop"=>43, "name_en"=>"Wooldridge", "lang"=>"en", "lat"=>38.906460, "lon"=>-92.521793),
"/US/US-MO/worth-12302" => array("type"=>"town", "pop"=>82, "name_ru"=>"Уорт", "name_en"=>"Worth", "lang"=>"en", "lat"=>40.405497, "lon"=>-94.446843),
"/US/US-MO/worthington-12303" => array("type"=>"village", "pop"=>83, "name_ru"=>"Уортингтон", "name_en"=>"Worthington", "lang"=>"en", "lat"=>40.408050, "lon"=>-92.688775),
"/US/US-MO/wright-city-12304" => array("type"=>"city", "pop"=>2967, "name_en"=>"Wright City", "lang"=>"en", "lat"=>38.827612, "lon"=>-91.024317),
"/US/US-MO/wyaconda-12305" => array("type"=>"city", "pop"=>291, "name_en"=>"Wyaconda", "lang"=>"en", "lat"=>40.393022, "lon"=>-91.926328),
"/US/US-MO/wyatt-12306" => array("type"=>"city", "pop"=>341, "name_en"=>"Wyatt", "lang"=>"en", "lat"=>36.911957, "lon"=>-89.223604),
"/US/US-MO/zalma-12307" => array("type"=>"village", "pop"=>94, "name_en"=>"Zalma", "lang"=>"en", "lat"=>37.143420, "lon"=>-90.080297),
"/US/US-MP" => array("type"=>"outlying area", "pop"=>84546, "name_ru"=>"Северные Марианские О-ва", "name_en"=>"Northern Mariana Islands", "lang"=>"en"),
"/US/US-MP/capital-hill-23541" => array("type"=>"cdp", "pop"=>1234, "name_en"=>"Capital Hill", "lang"=>"en", "lat"=>15.200100, "lon"=>145.751828),
"/US/US-MP/chalan-kanoa-23542" => array("type"=>"cdp", "pop"=>2549, "name_en"=>"Chalan Kanoa", "lang"=>"en", "lat"=>15.146235, "lon"=>145.704129),
"/US/US-MP/dandan-23543" => array("type"=>"cdp", "pop"=>901, "name_en"=>"Dandan", "lang"=>"en", "lat"=>15.134480, "lon"=>145.737087),
"/US/US-MP/garapan-23544" => array("type"=>"cdp", "pop"=>3904, "name_en"=>"Garapan", "lang"=>"en", "lat"=>15.206059, "lon"=>145.717948),
"/US/US-MP/gualo-rai-23545" => array("type"=>"cdp", "pop"=>1746, "name_en"=>"Gualo Rai", "lang"=>"en", "lat"=>15.190903, "lon"=>145.722614),
"/US/US-MP/kagman-23546" => array("type"=>"cdp", "pop"=>390, "name_en"=>"Kagman", "lang"=>"en", "lat"=>15.166715, "lon"=>145.778611),
"/US/US-MP/koblerville-23547" => array("type"=>"cdp", "pop"=>2811, "name_en"=>"Koblerville", "lang"=>"en", "lat"=>15.120451, "lon"=>145.706447),
"/US/US-MP/navy-hill-23548" => array("type"=>"cdp", "pop"=>419, "name_en"=>"Navy Hill", "lang"=>"en", "lat"=>15.209373, "lon"=>145.733735),
"/US/US-MP/san-antonio-23549" => array("type"=>"cdp", "pop"=>2887, "name_ru"=>"Сан Антонио", "name_en"=>"San Antonio", "lang"=>"en", "lat"=>15.133282, "lon"=>145.699118),
"/US/US-MP/san-jose-(saipan)-23550" => array("type"=>"cdp", "pop"=>839, "name_en"=>"San Jose (Saipan)", "lang"=>"en", "lat"=>15.166863, "lon"=>145.708628),
"/US/US-MP/san-jose-(tinian)-23551" => array("type"=>"cdp", "pop"=>1358, "name_en"=>"San Jose (Tinian)", "lang"=>"en", "lat"=>14.968683, "lon"=>145.619687),
"/US/US-MP/san-roque-23552" => array("type"=>"cdp", "pop"=>911, "name_en"=>"San Roque", "lang"=>"en", "lat"=>15.247416, "lon"=>145.772041),
"/US/US-MP/san-vicente-23553" => array("type"=>"cdp", "pop"=>1669, "name_en"=>"San Vicente", "lang"=>"en", "lat"=>15.151564, "lon"=>145.737763),
"/US/US-MP/songsong-23554" => array("type"=>"cdp", "pop"=>1313, "name_en"=>"Songsong", "lang"=>"en", "lat"=>14.132934, "lon"=>145.145342),
"/US/US-MP/susupe-23555" => array("type"=>"cdp", "pop"=>1776, "name_en"=>"Susupe", "lang"=>"en", "lat"=>15.156364, "lon"=>145.705708),
"/US/US-MP/tanapag-23556" => array("type"=>"cdp", "pop"=>1602, "name_en"=>"Tanapag", "lang"=>"en", "lat"=>15.238895, "lon"=>145.753791),
"/US/US-MS" => array("type"=>"state", "pop"=>2938618, "name_ru"=>"Миссисипи", "name_en"=>"Mississippi", "lang"=>"en"),
"/US/US-MS/abbeville-11023" => array("type"=>"town", "pop"=>449, "name_ru"=>"Аббевилл", "name_en"=>"Abbeville", "lang"=>"en", "lat"=>34.502676, "lon"=>-89.502230),
"/US/US-MS/aberdeen-11024" => array("type"=>"city", "pop"=>6089, "name_ru"=>"Абердин", "name_en"=>"Aberdeen", "lang"=>"en", "lat"=>33.828537, "lon"=>-88.551589),
"/US/US-MS/ackerman-11025" => array("type"=>"town", "pop"=>1524, "name_ru"=>"Акерман", "name_en"=>"Ackerman", "lang"=>"en", "lat"=>33.310706, "lon"=>-89.171865),
"/US/US-MS/algoma-11026" => array("type"=>"town", "pop"=>549, "name_ru"=>"Алгома", "name_en"=>"Algoma", "lang"=>"en", "lat"=>34.178828, "lon"=>-89.035956),
"/US/US-MS/alligator-11027" => array("type"=>"town", "pop"=>196, "name_ru"=>"Аллигатор", "name_en"=>"Alligator", "lang"=>"en", "lat"=>34.089808, "lon"=>-90.720204),
"/US/US-MS/amory-11028" => array("type"=>"city", "pop"=>7262, "name_ru"=>"Амори", "name_en"=>"Amory", "lang"=>"en", "lat"=>33.987883, "lon"=>-88.482896),
"/US/US-MS/anguilla-11029" => array("type"=>"town", "pop"=>763, "name_ru"=>"Ангилья", "name_en"=>"Anguilla", "lang"=>"en", "lat"=>32.972144, "lon"=>-90.828241),
"/US/US-MS/arcola-11030" => array("type"=>"town", "pop"=>484, "name_ru"=>"Аркола", "name_en"=>"Arcola", "lang"=>"en", "lat"=>33.271097, "lon"=>-90.880267),
"/US/US-MS/artesia-11031" => array("type"=>"town", "pop"=>473, "name_ru"=>"Артижа", "name_en"=>"Artesia", "lang"=>"en", "lat"=>33.417047, "lon"=>-88.642892),
"/US/US-MS/ashland-11032" => array("type"=>"town", "pop"=>564, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>34.832755, "lon"=>-89.176623),
"/US/US-MS/baldwyn-11033" => array("type"=>"city", "pop"=>3362, "name_ru"=>"Болдуин", "name_en"=>"Baldwyn", "lang"=>"en", "lat"=>34.502966, "lon"=>-88.633681),
"/US/US-MS/bassfield-11034" => array("type"=>"town", "pop"=>274, "name_ru"=>"Бассфилд", "name_en"=>"Bassfield", "lang"=>"en", "lat"=>31.497325, "lon"=>-89.742467),
"/US/US-MS/batesville-11035" => array("type"=>"city", "pop"=>7880, "name_ru"=>"Батсвилл", "name_en"=>"Batesville", "lang"=>"en", "lat"=>34.319458, "lon"=>-89.936843),
"/US/US-MS/bay-springs-11037" => array("type"=>"town", "pop"=>2198, "name_ru"=>"Бэй Спрингс", "name_en"=>"Bay Springs", "lang"=>"en", "lat"=>31.980159, "lon"=>-89.284145),
"/US/US-MS/bay-st.-louis-11036" => array("type"=>"city", "pop"=>8052, "name_ru"=>"Бэй Сент Луис", "name_en"=>"Bay St. Louis", "lang"=>"en", "lat"=>30.310150, "lon"=>-89.330021),
"/US/US-MS/beaumont-11038" => array("type"=>"town", "pop"=>973, "name_ru"=>"Бомонт", "name_en"=>"Beaumont", "lang"=>"en", "lat"=>31.172550, "lon"=>-88.924812),
"/US/US-MS/beauregard-11039" => array("type"=>"village", "pop"=>278, "name_ru"=>"Борегард", "name_en"=>"Beauregard", "lang"=>"en", "lat"=>31.721384, "lon"=>-90.388004),
"/US/US-MS/belmont-11040" => array("type"=>"town", "pop"=>1930, "name_ru"=>"Белмонт", "name_en"=>"Belmont", "lang"=>"en", "lat"=>34.507034, "lon"=>-88.208027),
"/US/US-MS/belzoni-11041" => array("type"=>"city", "pop"=>2432, "name_ru"=>"Белзони", "name_en"=>"Belzoni", "lang"=>"en", "lat"=>33.179311, "lon"=>-90.485548),
"/US/US-MS/benoit-11042" => array("type"=>"town", "pop"=>556, "name_ru"=>"Беноит", "name_en"=>"Benoit", "lang"=>"en", "lat"=>33.651542, "lon"=>-91.008528),
"/US/US-MS/bentonia-11043" => array("type"=>"town", "pop"=>482, "name_ru"=>"Бентония", "name_en"=>"Bentonia", "lang"=>"en", "lat"=>32.647313, "lon"=>-90.370168),
"/US/US-MS/beulah-11044" => array("type"=>"town", "pop"=>435, "name_ru"=>"Бойла", "name_en"=>"Beulah", "lang"=>"en", "lat"=>33.790377, "lon"=>-90.980396),
"/US/US-MS/big-creek-11045" => array("type"=>"village", "pop"=>121, "name_ru"=>"Биг Крик", "name_en"=>"Big Creek", "lang"=>"en", "lat"=>33.843340, "lon"=>-89.415728),
"/US/US-MS/biloxi-11046" => array("type"=>"city", "pop"=>45670, "name_ru"=>"Билокси", "name_en"=>"Biloxi", "lang"=>"en", "lat"=>30.415767, "lon"=>-88.931918),
"/US/US-MS/blue-mountain-11047" => array("type"=>"town", "pop"=>737, "name_ru"=>"Блу Маунтин", "name_en"=>"Blue Mountain", "lang"=>"en", "lat"=>34.671045, "lon"=>-89.027161),
"/US/US-MS/blue-springs-11048" => array("type"=>"village", "pop"=>155, "name_ru"=>"Блу Спрингс", "name_en"=>"Blue Springs", "lang"=>"en", "lat"=>34.402876, "lon"=>-88.872831),
"/US/US-MS/bogue-chitto-11049" => array("type"=>"cdp", "pop"=>689, "name_ru"=>"Бог Читто", "name_en"=>"Bogue Chitto", "lang"=>"en", "lat"=>32.836257, "lon"=>-88.918449),
"/US/US-MS/bolton-11050" => array("type"=>"town", "pop"=>602, "name_ru"=>"Болтон", "name_en"=>"Bolton", "lang"=>"en", "lat"=>32.352576, "lon"=>-90.457399),
"/US/US-MS/booneville-11051" => array("type"=>"city", "pop"=>8749, "name_ru"=>"Бунвилл", "name_en"=>"Booneville", "lang"=>"en", "lat"=>34.662613, "lon"=>-88.567728),
"/US/US-MS/boyle-11052" => array("type"=>"town", "pop"=>700, "name_ru"=>"Бойл", "name_en"=>"Boyle", "lang"=>"en", "lat"=>33.703812, "lon"=>-90.725988),
"/US/US-MS/brandon-11053" => array("type"=>"city", "pop"=>22160, "name_ru"=>"Брэндон", "name_en"=>"Brandon", "lang"=>"en", "lat"=>32.280472, "lon"=>-90.003836),
"/US/US-MS/braxton-11054" => array("type"=>"village", "pop"=>184, "name_ru"=>"Брэкстон", "name_en"=>"Braxton", "lang"=>"en", "lat"=>32.024654, "lon"=>-89.972233),
"/US/US-MS/brookhaven-11055" => array("type"=>"city", "pop"=>13296, "name_ru"=>"Брукхейвен", "name_en"=>"Brookhaven", "lang"=>"en", "lat"=>31.583413, "lon"=>-90.445291),
"/US/US-MS/brooksville-11056" => array("type"=>"town", "pop"=>1121, "name_ru"=>"Бруксвилл", "name_en"=>"Brooksville", "lang"=>"en", "lat"=>33.232551, "lon"=>-88.582558),
"/US/US-MS/bruce-11057" => array("type"=>"town", "pop"=>2004, "name_ru"=>"Брюс", "name_en"=>"Bruce", "lang"=>"en", "lat"=>33.991796, "lon"=>-89.344900),
"/US/US-MS/bude-11058" => array("type"=>"town", "pop"=>1016, "name_ru"=>"Бьюд", "name_en"=>"Bude", "lang"=>"en", "lat"=>31.462635, "lon"=>-90.851233),
"/US/US-MS/burnsville-11059" => array("type"=>"town", "pop"=>1026, "name_ru"=>"Бернсвилл", "name_en"=>"Burnsville", "lang"=>"en", "lat"=>34.841982, "lon"=>-88.315965),
"/US/US-MS/byhalia-11060" => array("type"=>"town", "pop"=>1310, "name_ru"=>"Байхалия", "name_en"=>"Byhalia", "lang"=>"en", "lat"=>34.864514, "lon"=>-89.687371),
"/US/US-MS/caledonia-11061" => array("type"=>"town", "pop"=>948, "name_ru"=>"Каледония", "name_en"=>"Caledonia", "lang"=>"en", "lat"=>33.683449, "lon"=>-88.325952),
"/US/US-MS/calhoun-city-11062" => array("type"=>"town", "pop"=>1786, "name_ru"=>"Калхун Сити", "name_en"=>"Calhoun City", "lang"=>"en", "lat"=>33.859316, "lon"=>-89.314559),
"/US/US-MS/canton-11063" => array("type"=>"city", "pop"=>12520, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>32.610652, "lon"=>-90.031075),
"/US/US-MS/carrollton-11064" => array("type"=>"town", "pop"=>394, "name_ru"=>"Карролтон", "name_en"=>"Carrollton", "lang"=>"en", "lat"=>33.504936, "lon"=>-89.922573),
"/US/US-MS/carthage-11065" => array("type"=>"city", "pop"=>4815, "name_ru"=>"Картаж", "name_en"=>"Carthage", "lang"=>"en", "lat"=>32.744036, "lon"=>-89.533445),
"/US/US-MS/cary-11066" => array("type"=>"town", "pop"=>360, "name_ru"=>"Кэри", "name_en"=>"Cary", "lang"=>"en", "lat"=>32.804695, "lon"=>-90.925900),
"/US/US-MS/centreville-11067" => array("type"=>"town", "pop"=>1566, "name_ru"=>"Сентревилл", "name_en"=>"Centreville", "lang"=>"en", "lat"=>31.086376, "lon"=>-91.065772),
"/US/US-MS/charleston-11068" => array("type"=>"city", "pop"=>1882, "name_ru"=>"Чарльстон", "name_en"=>"Charleston", "lang"=>"en", "lat"=>34.007424, "lon"=>-90.055123),
"/US/US-MS/chunky-11069" => array("type"=>"town", "pop"=>355, "name_ru"=>"Чанки", "name_en"=>"Chunky", "lang"=>"en", "lat"=>32.327058, "lon"=>-88.929665),
"/US/US-MS/clarksdale-11070" => array("type"=>"city", "pop"=>18006, "name_ru"=>"Кларксдейл", "name_en"=>"Clarksdale", "lang"=>"en", "lat"=>34.200385, "lon"=>-90.576670),
"/US/US-MS/cleveland-11071" => array("type"=>"city", "pop"=>12218, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>33.743243, "lon"=>-90.728105),
"/US/US-MS/clinton-11072" => array("type"=>"city", "pop"=>26313, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>32.339700, "lon"=>-90.334928),
"/US/US-MS/coahoma-11073" => array("type"=>"town", "pop"=>343, "name_ru"=>"Коахома", "name_en"=>"Coahoma", "lang"=>"en", "lat"=>34.365896, "lon"=>-90.522211),
"/US/US-MS/coffeeville-11074" => array("type"=>"town", "pop"=>949, "name_ru"=>"Коффивилл", "name_en"=>"Coffeeville", "lang"=>"en", "lat"=>33.976741, "lon"=>-89.677499),
"/US/US-MS/coldwater-11075" => array("type"=>"town", "pop"=>1619, "name_ru"=>"Коулдуотер", "name_en"=>"Coldwater", "lang"=>"en", "lat"=>34.690244, "lon"=>-89.975442),
"/US/US-MS/collins-11076" => array("type"=>"city", "pop"=>2765, "name_ru"=>"Коллинс", "name_en"=>"Collins", "lang"=>"en", "lat"=>31.637752, "lon"=>-89.557846),
"/US/US-MS/collinsville-11077" => array("type"=>"cdp", "pop"=>1364, "name_ru"=>"Коллинсвилл", "name_en"=>"Collinsville", "lang"=>"en", "lat"=>32.494410, "lon"=>-88.842497),
"/US/US-MS/columbia-11078" => array("type"=>"city", "pop"=>6567, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>31.256604, "lon"=>-89.828700),
"/US/US-MS/columbus-11079" => array("type"=>"city", "pop"=>23798, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>33.504977, "lon"=>-88.404172),
"/US/US-MS/columbus-afb-11080" => array("type"=>"cdp", "pop"=>2890, "name_ru"=>"Военно-воздушная база Колумбус", "name_en"=>"Columbus AFB", "lang"=>"en", "lat"=>33.634487, "lon"=>-88.449848),
"/US/US-MS/como-11081" => array("type"=>"town", "pop"=>1315, "name_ru"=>"Комо", "name_en"=>"Como", "lang"=>"en", "lat"=>34.514867, "lon"=>-89.939962),
"/US/US-MS/conehatta-11082" => array("type"=>"cdp", "pop"=>925, "name_ru"=>"Конхэтта", "name_en"=>"Conehatta", "lang"=>"en", "lat"=>32.457401, "lon"=>-89.271089),
"/US/US-MS/corinth-11083" => array("type"=>"city", "pop"=>14253, "name_ru"=>"Коринт", "name_en"=>"Corinth", "lang"=>"en", "lat"=>34.937680, "lon"=>-88.516842),
"/US/US-MS/courtland-11084" => array("type"=>"town", "pop"=>472, "name_ru"=>"Кортленд", "name_en"=>"Courtland", "lang"=>"en", "lat"=>34.241099, "lon"=>-89.942507),
"/US/US-MS/crawford-11085" => array("type"=>"town", "pop"=>624, "name_ru"=>"Крофорд", "name_en"=>"Crawford", "lang"=>"en", "lat"=>33.302698, "lon"=>-88.623442),
"/US/US-MS/crenshaw-11086" => array("type"=>"town", "pop"=>910, "name_ru"=>"Креншо", "name_en"=>"Crenshaw", "lang"=>"en", "lat"=>34.503888, "lon"=>-90.194689),
"/US/US-MS/crosby-11087" => array("type"=>"town", "pop"=>351, "name_ru"=>"Кросби", "name_en"=>"Crosby", "lang"=>"en", "lat"=>31.278645, "lon"=>-91.061384),
"/US/US-MS/crowder-11088" => array("type"=>"town", "pop"=>729, "name_ru"=>"Краудер", "name_en"=>"Crowder", "lang"=>"en", "lat"=>34.172970, "lon"=>-90.137576),
"/US/US-MS/cruger-11089" => array("type"=>"town", "pop"=>428, "name_ru"=>"Крюгер", "name_en"=>"Cruger", "lang"=>"en", "lat"=>33.323761, "lon"=>-90.234855),
"/US/US-MS/crystal-springs-11090" => array("type"=>"city", "pop"=>5950, "name_ru"=>"Кристал Спрингс", "name_en"=>"Crystal Springs", "lang"=>"en", "lat"=>31.989708, "lon"=>-90.355013),
"/US/US-MS/d'iberville-11095" => array("type"=>"city", "pop"=>7928, "name_ru"=>"Дибервилл", "name_en"=>"D'Iberville", "lang"=>"en", "lat"=>30.443733, "lon"=>-88.897634),
"/US/US-MS/d'lo-11096" => array("type"=>"town", "pop"=>391, "name_ru"=>"Дло", "name_en"=>"D'Lo", "lang"=>"en", "lat"=>31.986510, "lon"=>-89.900952),
"/US/US-MS/de-kalb-11092" => array("type"=>"town", "pop"=>876, "name_ru"=>"Де Калб", "name_en"=>"De Kalb", "lang"=>"en", "lat"=>32.768494, "lon"=>-88.655469),
"/US/US-MS/decatur-11091" => array("type"=>"town", "pop"=>1943, "name_ru"=>"Декатур", "name_en"=>"Decatur", "lang"=>"en", "lat"=>32.438177, "lon"=>-89.110864),
"/US/US-MS/derma-11093" => array("type"=>"town", "pop"=>1113, "name_ru"=>"Дерма", "name_en"=>"Derma", "lang"=>"en", "lat"=>33.854541, "lon"=>-89.286280),
"/US/US-MS/diamondhead-11094" => array("type"=>"cdp", "pop"=>2661, "name_ru"=>"Даймондхед", "name_en"=>"Diamondhead", "lang"=>"en", "lat"=>30.383600, "lon"=>-89.368584),
"/US/US-MS/doddsville-11097" => array("type"=>"town", "pop"=>98, "name_ru"=>"Доддсвилл", "name_en"=>"Doddsville", "lang"=>"en", "lat"=>33.658489, "lon"=>-90.524425),
"/US/US-MS/drew-11098" => array("type"=>"city", "pop"=>2172, "name_ru"=>"Дрю", "name_en"=>"Drew", "lang"=>"en", "lat"=>33.809810, "lon"=>-90.530423),
"/US/US-MS/duck-hill-11099" => array("type"=>"town", "pop"=>1690, "name_ru"=>"Дак Хилл", "name_en"=>"Duck Hill", "lang"=>"en", "lat"=>33.631357, "lon"=>-89.712527),
"/US/US-MS/dumas-11100" => array("type"=>"town", "pop"=>457, "name_ru"=>"Дюма", "name_en"=>"Dumas", "lang"=>"en", "lat"=>34.635337, "lon"=>-88.836972),
"/US/US-MS/duncan-11101" => array("type"=>"town", "pop"=>552, "name_ru"=>"Дункан", "name_en"=>"Duncan", "lang"=>"en", "lat"=>34.043546, "lon"=>-90.744606),
"/US/US-MS/durant-11102" => array("type"=>"city", "pop"=>2720, "name_ru"=>"Дурант", "name_en"=>"Durant", "lang"=>"en", "lat"=>33.080185, "lon"=>-89.856479),
"/US/US-MS/ecru-11103" => array("type"=>"town", "pop"=>1051, "name_ru"=>"Экру", "name_en"=>"Ecru", "lang"=>"en", "lat"=>34.356215, "lon"=>-89.025420),
"/US/US-MS/eden-11104" => array("type"=>"village", "pop"=>122, "name_ru"=>"Иден", "name_en"=>"Eden", "lang"=>"en", "lat"=>32.985710, "lon"=>-90.323186),
"/US/US-MS/edwards-11105" => array("type"=>"town", "pop"=>1292, "name_ru"=>"Эдвардс", "name_en"=>"Edwards", "lang"=>"en", "lat"=>32.332308, "lon"=>-90.604719),
"/US/US-MS/ellisville-11106" => array("type"=>"city", "pop"=>4392, "name_ru"=>"Эллисвилл", "name_en"=>"Ellisville", "lang"=>"en", "lat"=>31.598750, "lon"=>-89.207089),
"/US/US-MS/enterprise-11107" => array("type"=>"town", "pop"=>454, "name_ru"=>"Энтерпрайз", "name_en"=>"Enterprise", "lang"=>"en", "lat"=>32.174912, "lon"=>-88.813683),
"/US/US-MS/escatawpa-11108" => array("type"=>"cdp", "pop"=>3902, "name_ru"=>"Эскатопа", "name_en"=>"Escatawpa", "lang"=>"en", "lat"=>30.489002, "lon"=>-88.551151),
"/US/US-MS/ethel-11109" => array("type"=>"town", "pop"=>458, "name_ru"=>"Этел", "name_en"=>"Ethel", "lang"=>"en", "lat"=>33.121947, "lon"=>-89.464834),
"/US/US-MS/eupora-11110" => array("type"=>"town", "pop"=>2213, "name_ru"=>"Юпора", "name_en"=>"Eupora", "lang"=>"en", "lat"=>33.543862, "lon"=>-89.277771),
"/US/US-MS/falcon-11111" => array("type"=>"town", "pop"=>284, "name_ru"=>"Фэлкон", "name_en"=>"Falcon", "lang"=>"en", "lat"=>34.392646, "lon"=>-90.255475),
"/US/US-MS/falkner-11112" => array("type"=>"town", "pop"=>204, "name_ru"=>"Фолкнер", "name_en"=>"Falkner", "lang"=>"en", "lat"=>34.847467, "lon"=>-88.931896),
"/US/US-MS/fayette-11113" => array("type"=>"city", "pop"=>1986, "name_ru"=>"Файет", "name_en"=>"Fayette", "lang"=>"en", "lat"=>31.711683, "lon"=>-91.061743),
"/US/US-MS/flora-11114" => array("type"=>"town", "pop"=>1489, "name_ru"=>"Флора", "name_en"=>"Flora", "lang"=>"en", "lat"=>32.547420, "lon"=>-90.311974),
"/US/US-MS/florence-11115" => array("type"=>"town", "pop"=>3474, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>32.154848, "lon"=>-90.128419),
"/US/US-MS/flowood-11116" => array("type"=>"town", "pop"=>7016, "name_ru"=>"Фловуд", "name_en"=>"Flowood", "lang"=>"en", "lat"=>32.323200, "lon"=>-90.105709),
"/US/US-MS/forest-11117" => array("type"=>"city", "pop"=>5972, "name_ru"=>"Форест", "name_en"=>"Forest", "lang"=>"en", "lat"=>32.356889, "lon"=>-89.473975),
"/US/US-MS/french-camp-11118" => array("type"=>"village", "pop"=>380, "name_ru"=>"Френч Кэмп", "name_en"=>"French Camp", "lang"=>"en", "lat"=>33.292087, "lon"=>-89.398478),
"/US/US-MS/friars-point-11119" => array("type"=>"town", "pop"=>1313, "name_ru"=>"Фрайарс Пойнт", "name_en"=>"Friars Point", "lang"=>"en", "lat"=>34.367289, "lon"=>-90.638126),
"/US/US-MS/fulton-11120" => array("type"=>"city", "pop"=>4073, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>34.262308, "lon"=>-88.401722),
"/US/US-MS/gattman-11121" => array("type"=>"village", "pop"=>114, "name_ru"=>"Гаттман", "name_en"=>"Gattman", "lang"=>"en", "lat"=>33.885485, "lon"=>-88.236091),
"/US/US-MS/gautier-11122" => array("type"=>"city", "pop"=>16306, "name_ru"=>"Готье", "name_en"=>"Gautier", "lang"=>"en", "lat"=>30.388850, "lon"=>-88.639869),
"/US/US-MS/georgetown-11123" => array("type"=>"town", "pop"=>352, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>31.869783, "lon"=>-90.164788),
"/US/US-MS/glendora-11124" => array("type"=>"village", "pop"=>250, "name_ru"=>"Глендора", "name_en"=>"Glendora", "lang"=>"en", "lat"=>33.828470, "lon"=>-90.292058),
"/US/US-MS/gloster-11125" => array("type"=>"town", "pop"=>1026, "name_ru"=>"Глостер", "name_en"=>"Gloster", "lang"=>"en", "lat"=>31.194551, "lon"=>-91.018002),
"/US/US-MS/golden-11126" => array("type"=>"town", "pop"=>199, "name_ru"=>"Гоулден", "name_en"=>"Golden", "lang"=>"en", "lat"=>34.485194, "lon"=>-88.186118),
"/US/US-MS/goodman-11127" => array("type"=>"town", "pop"=>1226, "name_ru"=>"Гудман", "name_en"=>"Goodman", "lang"=>"en", "lat"=>32.967927, "lon"=>-89.912542),
"/US/US-MS/greenville-11128" => array("type"=>"city", "pop"=>35764, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>33.385402, "lon"=>-91.051887),
"/US/US-MS/greenwood-11129" => array("type"=>"city", "pop"=>16084, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>33.517785, "lon"=>-90.191606),
"/US/US-MS/grenada-11130" => array("type"=>"city", "pop"=>14664, "name_ru"=>"Гренада", "name_en"=>"Grenada", "lang"=>"en", "lat"=>33.778249, "lon"=>-89.810400),
"/US/US-MS/gulf-hills-11131" => array("type"=>"cdp", "pop"=>5004, "name_ru"=>"Галф Хиллс", "name_en"=>"Gulf Hills", "lang"=>"en", "lat"=>30.437479, "lon"=>-88.815369),
"/US/US-MS/gulf-park-estates-11132" => array("type"=>"cdp", "pop"=>2314, "name_ru"=>"Галф Парк Эстейтс", "name_en"=>"Gulf Park Estates", "lang"=>"en", "lat"=>30.379188, "lon"=>-88.758496),
"/US/US-MS/gulfport-11133" => array("type"=>"city", "pop"=>70055, "name_ru"=>"Галфпорт", "name_en"=>"Gulfport", "lang"=>"en", "lat"=>30.386100, "lon"=>-89.067682),
"/US/US-MS/gunnison-11134" => array("type"=>"town", "pop"=>585, "name_ru"=>"Ганнисон", "name_en"=>"Gunnison", "lang"=>"en", "lat"=>33.944202, "lon"=>-90.946813),
"/US/US-MS/guntown-11135" => array("type"=>"town", "pop"=>1723, "name_ru"=>"Гантаун", "name_en"=>"Guntown", "lang"=>"en", "lat"=>34.444865, "lon"=>-88.663034),
"/US/US-MS/hatley-11136" => array("type"=>"town", "pop"=>466, "name_ru"=>"Хэтли", "name_en"=>"Hatley", "lang"=>"en", "lat"=>33.977230, "lon"=>-88.418053),
"/US/US-MS/hattiesburg-11137" => array("type"=>"city", "pop"=>51993, "name_ru"=>"Хаттисбург", "name_en"=>"Hattiesburg", "lang"=>"en", "lat"=>31.312750, "lon"=>-89.306919),
"/US/US-MS/hazlehurst-11138" => array("type"=>"city", "pop"=>4387, "name_ru"=>"Хейзлхерст", "name_en"=>"Hazlehurst", "lang"=>"en", "lat"=>31.864750, "lon"=>-90.391261),
"/US/US-MS/heidelberg-11139" => array("type"=>"town", "pop"=>812, "name_ru"=>"Хайделберг", "name_en"=>"Heidelberg", "lang"=>"en", "lat"=>31.890084, "lon"=>-88.987961),
"/US/US-MS/hernando-11140" => array("type"=>"city", "pop"=>12318, "name_ru"=>"Гернандо", "name_en"=>"Hernando", "lang"=>"en", "lat"=>34.827613, "lon"=>-89.986922),
"/US/US-MS/hickory-11141" => array("type"=>"town", "pop"=>509, "name_ru"=>"Хикори", "name_en"=>"Hickory", "lang"=>"en", "lat"=>32.316549, "lon"=>-89.021077),
"/US/US-MS/hickory-flat-11142" => array("type"=>"town", "pop"=>556, "name_ru"=>"Хикори Флэт", "name_en"=>"Hickory Flat", "lang"=>"en", "lat"=>34.615152, "lon"=>-89.186953),
"/US/US-MS/hollandale-11143" => array("type"=>"city", "pop"=>2962, "name_ru"=>"Холландейл", "name_en"=>"Hollandale", "lang"=>"en", "lat"=>33.174418, "lon"=>-90.853223),
"/US/US-MS/holly-springs-11144" => array("type"=>"city", "pop"=>7994, "name_ru"=>"Хоули Спрингс", "name_en"=>"Holly Springs", "lang"=>"en", "lat"=>34.772684, "lon"=>-89.443457),
"/US/US-MS/horn-lake-11145" => array("type"=>"city", "pop"=>24669, "name_ru"=>"Хорн Лейк", "name_en"=>"Horn Lake", "lang"=>"en", "lat"=>34.956076, "lon"=>-90.035284),
"/US/US-MS/houston-11146" => array("type"=>"city", "pop"=>3877, "name_ru"=>"Хьюстон", "name_en"=>"Houston", "lang"=>"en", "lat"=>33.896169, "lon"=>-89.001593),
"/US/US-MS/indianola-11147" => array("type"=>"city", "pop"=>10805, "name_ru"=>"Индианола", "name_en"=>"Indianola", "lang"=>"en", "lat"=>33.448619, "lon"=>-90.644781),
"/US/US-MS/inverness-11148" => array("type"=>"town", "pop"=>1001, "name_ru"=>"Инвернесс", "name_en"=>"Inverness", "lang"=>"en", "lat"=>33.353991, "lon"=>-90.590748),
"/US/US-MS/isola-11149" => array("type"=>"town", "pop"=>684, "name_ru"=>"Айсола", "name_en"=>"Isola", "lang"=>"en", "lat"=>33.262549, "lon"=>-90.591665),
"/US/US-MS/itta-bena-11150" => array("type"=>"city", "pop"=>1875, "name_ru"=>"Итта Бена", "name_en"=>"Itta Bena", "lang"=>"en", "lat"=>33.499720, "lon"=>-90.325028),
"/US/US-MS/iuka-11151" => array("type"=>"city", "pop"=>2938, "name_ru"=>"Иука", "name_en"=>"Iuka", "lang"=>"en", "lat"=>34.808033, "lon"=>-88.197296),
"/US/US-MS/jackson-11152" => array("type"=>"city", "pop"=>173861, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>32.320500, "lon"=>-90.207591),
"/US/US-MS/jonestown-11153" => array("type"=>"town", "pop"=>1556, "name_ru"=>"Джоунстаун", "name_en"=>"Jonestown", "lang"=>"en", "lat"=>34.321154, "lon"=>-90.454271),
"/US/US-MS/jumpertown-11154" => array("type"=>"town", "pop"=>407, "name_ru"=>"Джампертаун", "name_en"=>"Jumpertown", "lang"=>"en", "lat"=>34.706414, "lon"=>-88.663833),
"/US/US-MS/kilmichael-11155" => array("type"=>"town", "pop"=>676, "name_ru"=>"Килмайкл", "name_en"=>"Kilmichael", "lang"=>"en", "lat"=>33.439027, "lon"=>-89.566951),
"/US/US-MS/kiln-11156" => array("type"=>"cdp", "pop"=>1262, "name_ru"=>"Килн", "name_en"=>"Kiln", "lang"=>"en", "lat"=>30.416539, "lon"=>-89.434231),
"/US/US-MS/kosciusko-11157" => array("type"=>"city", "pop"=>7371, "name_ru"=>"Костюшко", "name_en"=>"Kosciusko", "lang"=>"en", "lat"=>33.057979, "lon"=>-89.588225),
"/US/US-MS/kossuth-11158" => array("type"=>"village", "pop"=>174, "name_ru"=>"Кошут", "name_en"=>"Kossuth", "lang"=>"en", "lat"=>34.871398, "lon"=>-88.644064),
"/US/US-MS/lake-11159" => array("type"=>"town", "pop"=>406, "name_ru"=>"Лейк", "name_en"=>"Lake", "lang"=>"en", "lat"=>32.343086, "lon"=>-89.327450),
"/US/US-MS/lambert-11160" => array("type"=>"town", "pop"=>1654, "name_ru"=>"Ламберт", "name_en"=>"Lambert", "lang"=>"en", "lat"=>34.202013, "lon"=>-90.283003),
"/US/US-MS/latimer-11161" => array("type"=>"cdp", "pop"=>3222, "name_ru"=>"Латимер", "name_en"=>"Latimer", "lang"=>"en", "lat"=>30.495029, "lon"=>-88.859595),
"/US/US-MS/laurel-11162" => array("type"=>"city", "pop"=>18693, "name_ru"=>"Лорел", "name_en"=>"Laurel", "lang"=>"en", "lat"=>31.693694, "lon"=>-89.144659),
"/US/US-MS/leakesville-11163" => array("type"=>"town", "pop"=>1025, "name_ru"=>"Ликсвилл", "name_en"=>"Leakesville", "lang"=>"en", "lat"=>31.153983, "lon"=>-88.558289),
"/US/US-MS/learned-11164" => array("type"=>"town", "pop"=>47, "name_ru"=>"Лернд", "name_en"=>"Learned", "lang"=>"en", "lat"=>32.197659, "lon"=>-90.548724),
"/US/US-MS/leland-11165" => array("type"=>"city", "pop"=>4790, "name_ru"=>"Леленд", "name_en"=>"Leland", "lang"=>"en", "lat"=>33.404708, "lon"=>-90.899929),
"/US/US-MS/lena-11166" => array("type"=>"town", "pop"=>164, "name_ru"=>"Лена", "name_en"=>"Lena", "lang"=>"en", "lat"=>32.594565, "lon"=>-89.593524),
"/US/US-MS/lexington-11167" => array("type"=>"city", "pop"=>1880, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>33.116781, "lon"=>-90.049246),
"/US/US-MS/liberty-11168" => array("type"=>"town", "pop"=>670, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>31.160155, "lon"=>-90.801209),
"/US/US-MS/long-beach-11169" => array("type"=>"city", "pop"=>12234, "name_ru"=>"Лонг Бич", "name_en"=>"Long Beach", "lang"=>"en", "lat"=>30.362000, "lon"=>-89.167343),
"/US/US-MS/louin-11170" => array("type"=>"town", "pop"=>336, "name_ru"=>"Луин", "name_en"=>"Louin", "lang"=>"en", "lat"=>32.072488, "lon"=>-89.261696),
"/US/US-MS/louise-11171" => array("type"=>"town", "pop"=>285, "name_ru"=>"Луис", "name_en"=>"Louise", "lang"=>"en", "lat"=>32.982623, "lon"=>-90.591237),
"/US/US-MS/louisville-11172" => array("type"=>"city", "pop"=>6604, "name_ru"=>"Луисвилл", "name_en"=>"Louisville", "lang"=>"en", "lat"=>33.122828, "lon"=>-89.054596),
"/US/US-MS/lucedale-11173" => array("type"=>"city", "pop"=>3077, "name_ru"=>"Льюсдейл", "name_en"=>"Lucedale", "lang"=>"en", "lat"=>30.921643, "lon"=>-88.593554),
"/US/US-MS/lula-11174" => array("type"=>"town", "pop"=>324, "name_ru"=>"Лула", "name_en"=>"Lula", "lang"=>"en", "lat"=>34.454834, "lon"=>-90.477551),
"/US/US-MS/lumberton-11175" => array("type"=>"city", "pop"=>2698, "name_ru"=>"Ламбертон", "name_en"=>"Lumberton", "lang"=>"en", "lat"=>31.004154, "lon"=>-89.451738),
"/US/US-MS/lyman-11176" => array("type"=>"cdp", "pop"=>1117, "name_ru"=>"Лиман", "name_en"=>"Lyman", "lang"=>"en", "lat"=>30.501206, "lon"=>-89.130772),
"/US/US-MS/lynchburg-11177" => array("type"=>"cdp", "pop"=>2071, "name_ru"=>"Линчбург", "name_en"=>"Lynchburg", "lang"=>"en", "lat"=>34.967100, "lon"=>-90.094451),
"/US/US-MS/lyon-11178" => array("type"=>"town", "pop"=>368, "name_ru"=>"Лион", "name_en"=>"Lyon", "lang"=>"en", "lat"=>34.217062, "lon"=>-90.542179),
"/US/US-MS/maben-11179" => array("type"=>"town", "pop"=>764, "name_ru"=>"Мабен", "name_en"=>"Maben", "lang"=>"en", "lat"=>33.554048, "lon"=>-89.081007),
"/US/US-MS/macon-11183" => array("type"=>"city", "pop"=>2744, "name_ru"=>"Мейкон", "name_en"=>"Macon", "lang"=>"en", "lat"=>33.114923, "lon"=>-88.561201),
"/US/US-MS/madison-11184" => array("type"=>"city", "pop"=>17681, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>32.453416, "lon"=>-90.108450),
"/US/US-MS/magee-11185" => array("type"=>"city", "pop"=>4332, "name_ru"=>"Маги", "name_en"=>"Magee", "lang"=>"en", "lat"=>31.871225, "lon"=>-89.731234),
"/US/US-MS/magnolia-11186" => array("type"=>"city", "pop"=>2090, "name_ru"=>"Магнолия", "name_en"=>"Magnolia", "lang"=>"en", "lat"=>31.150240, "lon"=>-90.460716),
"/US/US-MS/mantachie-11187" => array("type"=>"town", "pop"=>1119, "name_ru"=>"Мантачи", "name_en"=>"Mantachie", "lang"=>"en", "lat"=>34.323980, "lon"=>-88.493029),
"/US/US-MS/mantee-11188" => array("type"=>"village", "pop"=>162, "name_ru"=>"Манти", "name_en"=>"Mantee", "lang"=>"en", "lat"=>33.730498, "lon"=>-89.060012),
"/US/US-MS/marietta-11189" => array("type"=>"town", "pop"=>245, "name_ru"=>"Мариетта", "name_en"=>"Marietta", "lang"=>"en", "lat"=>34.500892, "lon"=>-88.471097),
"/US/US-MS/marion-11190" => array("type"=>"town", "pop"=>1497, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>32.427622, "lon"=>-88.648350),
"/US/US-MS/marks-11191" => array("type"=>"city", "pop"=>1721, "name_ru"=>"Маркс", "name_en"=>"Marks", "lang"=>"en", "lat"=>34.253834, "lon"=>-90.272875),
"/US/US-MS/martin-bluff-11192" => array("type"=>"cdp", "pop"=>1928, "name_ru"=>"Мартин Блафф", "name_en"=>"Martin Bluff", "lang"=>"en", "lat"=>30.456622, "lon"=>-88.639358),
"/US/US-MS/mathiston-11193" => array("type"=>"town", "pop"=>696, "name_ru"=>"Матисон", "name_en"=>"Mathiston", "lang"=>"en", "lat"=>33.539659, "lon"=>-89.128273),
"/US/US-MS/mayersville-11194" => array("type"=>"town", "pop"=>645, "name_ru"=>"Мейерсвилл", "name_en"=>"Mayersville", "lang"=>"en", "lat"=>32.897430, "lon"=>-91.044918),
"/US/US-MS/mccomb-11180" => array("type"=>"city", "pop"=>13684, "name_ru"=>"Мак-Комб", "name_en"=>"McComb", "lang"=>"en", "lat"=>31.244953, "lon"=>-90.465200),
"/US/US-MS/mccool-11181" => array("type"=>"town", "pop"=>190, "name_ru"=>"Мак-Кул", "name_en"=>"McCool", "lang"=>"en", "lat"=>33.198327, "lon"=>-89.343022),
"/US/US-MS/mclain-11182" => array("type"=>"town", "pop"=>595, "name_ru"=>"Мак-Лейн", "name_en"=>"McLain", "lang"=>"en", "lat"=>31.103708, "lon"=>-88.824203),
"/US/US-MS/meadville-11195" => array("type"=>"town", "pop"=>494, "name_ru"=>"Мидвилл", "name_en"=>"Meadville", "lang"=>"en", "lat"=>31.471809, "lon"=>-90.893013),
"/US/US-MS/memphis-11196" => array("type"=>"village", "pop"=>70, "name_ru"=>"Мемфис", "name_en"=>"Memphis", "lang"=>"en", "lat"=>34.925726, "lon"=>-90.140916),
"/US/US-MS/mendenhall-11197" => array("type"=>"city", "pop"=>2553, "name_ru"=>"Менденхолл", "name_en"=>"Mendenhall", "lang"=>"en", "lat"=>31.960705, "lon"=>-89.869440),
"/US/US-MS/meridian-11198" => array("type"=>"city", "pop"=>38232, "name_ru"=>"Меридиан", "name_en"=>"Meridian", "lang"=>"en", "lat"=>32.380874, "lon"=>-88.712501),
"/US/US-MS/meridian-station-11199" => array("type"=>"cdp", "pop"=>2503, "name_ru"=>"Меридиан Стейшен", "name_en"=>"Meridian Station", "lang"=>"en", "lat"=>32.546151, "lon"=>-88.617615),
"/US/US-MS/merigold-11200" => array("type"=>"town", "pop"=>599, "name_ru"=>"Меригоулд", "name_en"=>"Merigold", "lang"=>"en", "lat"=>33.839071, "lon"=>-90.726912),
"/US/US-MS/metcalfe-11201" => array("type"=>"town", "pop"=>1259, "name_ru"=>"Меткалф", "name_en"=>"Metcalfe", "lang"=>"en", "lat"=>33.450096, "lon"=>-91.002282),
"/US/US-MS/mize-11202" => array("type"=>"town", "pop"=>277, "name_ru"=>"Майз", "name_en"=>"Mize", "lang"=>"en", "lat"=>31.867173, "lon"=>-89.554262),
"/US/US-MS/monticello-11203" => array("type"=>"town", "pop"=>1699, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>31.550346, "lon"=>-90.112650),
"/US/US-MS/montrose-11204" => array("type"=>"town", "pop"=>128, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>32.124383, "lon"=>-89.235436),
"/US/US-MS/moorhead-11205" => array("type"=>"city", "pop"=>2342, "name_ru"=>"Мурхэд", "name_en"=>"Moorhead", "lang"=>"en", "lat"=>33.449222, "lon"=>-90.506760),
"/US/US-MS/morgan-city-11206" => array("type"=>"town", "pop"=>283, "name_ru"=>"Морган Сити", "name_en"=>"Morgan City", "lang"=>"en", "lat"=>33.378854, "lon"=>-90.347333),
"/US/US-MS/morton-11207" => array("type"=>"city", "pop"=>3437, "name_ru"=>"Мортон", "name_en"=>"Morton", "lang"=>"en", "lat"=>32.355537, "lon"=>-89.656393),
"/US/US-MS/moss-point-11208" => array("type"=>"city", "pop"=>13951, "name_ru"=>"Мосс Пойнт", "name_en"=>"Moss Point", "lang"=>"en", "lat"=>30.425500, "lon"=>-88.529034),
"/US/US-MS/mound-bayou-11209" => array("type"=>"city", "pop"=>1936, "name_ru"=>"Маунд Байу", "name_en"=>"Mound Bayou", "lang"=>"en", "lat"=>33.880521, "lon"=>-90.727868),
"/US/US-MS/mount-olive-11210" => array("type"=>"town", "pop"=>965, "name_ru"=>"Маунт Олив", "name_en"=>"Mount Olive", "lang"=>"en", "lat"=>31.759398, "lon"=>-89.656237),
"/US/US-MS/myrtle-11211" => array("type"=>"town", "pop"=>564, "name_ru"=>"Миртл", "name_en"=>"Myrtle", "lang"=>"en", "lat"=>34.559276, "lon"=>-89.115806),
"/US/US-MS/natchez-11212" => array("type"=>"city", "pop"=>16413, "name_ru"=>"Натчез", "name_en"=>"Natchez", "lang"=>"en", "lat"=>31.547860, "lon"=>-91.388593),
"/US/US-MS/nellieburg-11213" => array("type"=>"cdp", "pop"=>1208, "name_ru"=>"Неллибург", "name_en"=>"Nellieburg", "lang"=>"en", "lat"=>32.387177, "lon"=>-88.784188),
"/US/US-MS/nettleton-11214" => array("type"=>"town", "pop"=>2008, "name_ru"=>"Неттлтон", "name_en"=>"Nettleton", "lang"=>"en", "lat"=>34.091284, "lon"=>-88.616540),
"/US/US-MS/new-albany-11215" => array("type"=>"city", "pop"=>8178, "name_ru"=>"Нью Олбани", "name_en"=>"New Albany", "lang"=>"en", "lat"=>34.490617, "lon"=>-89.020701),
"/US/US-MS/new-augusta-11216" => array("type"=>"town", "pop"=>693, "name_ru"=>"Нью Огаста", "name_en"=>"New Augusta", "lang"=>"en", "lat"=>31.203269, "lon"=>-89.028584),
"/US/US-MS/new-hebron-11217" => array("type"=>"village", "pop"=>436, "name_ru"=>"Нью Хеброн", "name_en"=>"New Hebron", "lang"=>"en", "lat"=>31.732929, "lon"=>-89.983209),
"/US/US-MS/new-hope-11218" => array("type"=>"cdp", "pop"=>1663, "name_ru"=>"Нью Хоуп", "name_en"=>"New Hope", "lang"=>"en", "lat"=>33.446506, "lon"=>-88.324931),
"/US/US-MS/new-houlka-11219" => array("type"=>"town", "pop"=>692, "name_ru"=>"Нью Хаулка", "name_en"=>"New Houlka", "lang"=>"en", "lat"=>34.036313, "lon"=>-89.020039),
"/US/US-MS/newport-11220" => array("type"=>"village", "pop"=>35, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>34.893220, "lon"=>-90.228506),
"/US/US-MS/newton-11221" => array("type"=>"city", "pop"=>3674, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>32.319731, "lon"=>-89.158901),
"/US/US-MS/north-carrollton-11222" => array("type"=>"town", "pop"=>464, "name_ru"=>"Норт Кэрролтон", "name_en"=>"North Carrollton", "lang"=>"en", "lat"=>33.518244, "lon"=>-89.919862),
"/US/US-MS/north-gulfport-11223" => array("type"=>"cdp", "pop"=>4966, "name_ru"=>"Норт Галфпорт", "name_en"=>"North Gulfport", "lang"=>"en", "lat"=>30.414033, "lon"=>-89.094282),
"/US/US-MS/north-tunica-11224" => array("type"=>"cdp", "pop"=>1314, "name_ru"=>"Норт Тунис", "name_en"=>"North Tunica", "lang"=>"en", "lat"=>34.700890, "lon"=>-90.378121),
"/US/US-MS/noxapater-11225" => array("type"=>"town", "pop"=>410, "name_ru"=>"Ноксапейтер", "name_en"=>"Noxapater", "lang"=>"en", "lat"=>32.992825, "lon"=>-89.063456),
"/US/US-MS/oakland-11226" => array("type"=>"town", "pop"=>601, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>34.053062, "lon"=>-89.914640),
"/US/US-MS/ocean-springs-11227" => array("type"=>"city", "pop"=>17149, "name_ru"=>"Ошен Спрингс", "name_en"=>"Ocean Springs", "lang"=>"en", "lat"=>30.406067, "lon"=>-88.801925),
"/US/US-MS/okolona-11228" => array("type"=>"city", "pop"=>2872, "name_ru"=>"Околона", "name_en"=>"Okolona", "lang"=>"en", "lat"=>34.005624, "lon"=>-88.750533),
"/US/US-MS/olive-branch-11229" => array("type"=>"city", "pop"=>31830, "name_ru"=>"Олив Брэнч", "name_en"=>"Olive Branch", "lang"=>"en", "lat"=>34.954779, "lon"=>-89.829036),
"/US/US-MS/orange-grove-11230" => array("type"=>"cdp", "pop"=>15676, "name_ru"=>"Оранж Гров", "name_en"=>"Orange Grove", "lang"=>"en", "lat"=>30.453100, "lon"=>-89.087649),
"/US/US-MS/osyka-11231" => array("type"=>"town", "pop"=>496, "name_ru"=>"Осыка", "name_en"=>"Osyka", "lang"=>"en", "lat"=>31.006983, "lon"=>-90.471091),
"/US/US-MS/oxford-11232" => array("type"=>"city", "pop"=>17265, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>34.359751, "lon"=>-89.526155),
"/US/US-MS/pace-11233" => array("type"=>"town", "pop"=>357, "name_ru"=>"Пэйс", "name_en"=>"Pace", "lang"=>"en", "lat"=>33.792058, "lon"=>-90.859080),
"/US/US-MS/pachuta-11234" => array("type"=>"town", "pop"=>233, "name_ru"=>"Пачута", "name_en"=>"Pachuta", "lang"=>"en", "lat"=>32.042816, "lon"=>-88.884364),
"/US/US-MS/paden-11235" => array("type"=>"village", "pop"=>105, "name_ru"=>"Пейден", "name_en"=>"Paden", "lang"=>"en", "lat"=>34.664568, "lon"=>-88.264633),
"/US/US-MS/pascagoula-11236" => array("type"=>"city", "pop"=>23609, "name_ru"=>"Паскагоула", "name_en"=>"Pascagoula", "lang"=>"en", "lat"=>30.366133, "lon"=>-88.547305),
"/US/US-MS/pass-christian-11237" => array("type"=>"city", "pop"=>3993, "name_ru"=>"Пасс Кристиан", "name_en"=>"Pass Christian", "lang"=>"en", "lat"=>30.318800, "lon"=>-89.241721),
"/US/US-MS/pearl-11238" => array("type"=>"city", "pop"=>24400, "name_ru"=>"Перл", "name_en"=>"Pearl", "lang"=>"en", "lat"=>32.271833, "lon"=>-90.105181),
"/US/US-MS/pearl-river-11240" => array("type"=>"cdp", "pop"=>57466, "name_ru"=>"Перл Ривер", "name_en"=>"Pearl River", "lang"=>"en", "lat"=>32.787602, "lon"=>-89.242769),
"/US/US-MS/pearlington-11239" => array("type"=>"cdp", "pop"=>1603, "name_ru"=>"Перлингтон", "name_en"=>"Pearlington", "lang"=>"en", "lat"=>30.248370, "lon"=>-89.601168),
"/US/US-MS/pelahatchie-11241" => array("type"=>"city", "pop"=>1474, "name_ru"=>"Пелахатчи", "name_en"=>"Pelahatchie", "lang"=>"en", "lat"=>32.315366, "lon"=>-89.801138),
"/US/US-MS/petal-11242" => array("type"=>"city", "pop"=>10575, "name_ru"=>"Петал", "name_en"=>"Petal", "lang"=>"en", "lat"=>31.344150, "lon"=>-89.251442),
"/US/US-MS/philadelphia-11243" => array("type"=>"city", "pop"=>7977, "name_ru"=>"Филадельфия", "name_en"=>"Philadelphia", "lang"=>"en", "lat"=>32.773977, "lon"=>-89.109214),
"/US/US-MS/picayune-11244" => array("type"=>"city", "pop"=>11787, "name_ru"=>"Пикаюн", "name_en"=>"Picayune", "lang"=>"en", "lat"=>30.528745, "lon"=>-89.678334),
"/US/US-MS/pickens-11245" => array("type"=>"town", "pop"=>1200, "name_ru"=>"Пикенс", "name_en"=>"Pickens", "lang"=>"en", "lat"=>32.890893, "lon"=>-89.968420),
"/US/US-MS/pittsboro-11246" => array("type"=>"village", "pop"=>205, "name_ru"=>"Питтсборо", "name_en"=>"Pittsboro", "lang"=>"en", "lat"=>33.941251, "lon"=>-89.336752),
"/US/US-MS/plantersville-11247" => array("type"=>"town", "pop"=>1338, "name_ru"=>"Плантерсвилл", "name_en"=>"Plantersville", "lang"=>"en", "lat"=>34.211527, "lon"=>-88.665750),
"/US/US-MS/polkville-11248" => array("type"=>"village", "pop"=>643, "name_ru"=>"Полквилл", "name_en"=>"Polkville", "lang"=>"en", "lat"=>32.191467, "lon"=>-89.692470),
"/US/US-MS/pontotoc-11249" => array("type"=>"city", "pop"=>5924, "name_ru"=>"Понтоток", "name_en"=>"Pontotoc", "lang"=>"en", "lat"=>34.248506, "lon"=>-89.006667),
"/US/US-MS/pope-11250" => array("type"=>"village", "pop"=>244, "name_ru"=>"Поуп", "name_en"=>"Pope", "lang"=>"en", "lat"=>34.213658, "lon"=>-89.947632),
"/US/US-MS/poplarville-11251" => array("type"=>"city", "pop"=>3003, "name_ru"=>"Попларвилл", "name_en"=>"Poplarville", "lang"=>"en", "lat"=>30.838025, "lon"=>-89.534128),
"/US/US-MS/port-gibson-11252" => array("type"=>"city", "pop"=>1668, "name_ru"=>"Порт Гибсон", "name_en"=>"Port Gibson", "lang"=>"en", "lat"=>31.955112, "lon"=>-90.983439),
"/US/US-MS/potts-camp-11253" => array("type"=>"town", "pop"=>508, "name_ru"=>"Поттс Кэмп", "name_en"=>"Potts Camp", "lang"=>"en", "lat"=>34.648254, "lon"=>-89.306898),
"/US/US-MS/prentiss-11254" => array("type"=>"town", "pop"=>1000, "name_ru"=>"Прентисс", "name_en"=>"Prentiss", "lang"=>"en", "lat"=>31.596317, "lon"=>-89.872018),
"/US/US-MS/puckett-11255" => array("type"=>"village", "pop"=>356, "name_ru"=>"Пакетт", "name_en"=>"Puckett", "lang"=>"en", "lat"=>32.084432, "lon"=>-89.778370),
"/US/US-MS/purvis-11256" => array("type"=>"city", "pop"=>2647, "name_ru"=>"Первис", "name_en"=>"Purvis", "lang"=>"en", "lat"=>31.142460, "lon"=>-89.408766),
"/US/US-MS/quitman-11257" => array("type"=>"city", "pop"=>2320, "name_ru"=>"Квитман", "name_en"=>"Quitman", "lang"=>"en", "lat"=>32.041343, "lon"=>-88.720424),
"/US/US-MS/raleigh-11258" => array("type"=>"town", "pop"=>1230, "name_ru"=>"Роли", "name_en"=>"Raleigh", "lang"=>"en", "lat"=>32.030354, "lon"=>-89.523626),
"/US/US-MS/raymond-11259" => array("type"=>"town", "pop"=>1663, "name_ru"=>"Реймонд", "name_en"=>"Raymond", "lang"=>"en", "lat"=>32.260304, "lon"=>-90.422715),
"/US/US-MS/redwater-11260" => array("type"=>"cdp", "pop"=>289, "name_ru"=>"Редуотер", "name_en"=>"Redwater", "lang"=>"en", "lat"=>32.782640, "lon"=>-89.550554),
"/US/US-MS/renova-11261" => array("type"=>"town", "pop"=>577, "name_ru"=>"Ренова", "name_en"=>"Renova", "lang"=>"en", "lat"=>33.780926, "lon"=>-90.722647),
"/US/US-MS/richland-11262" => array("type"=>"city", "pop"=>7258, "name_ru"=>"Ричленд", "name_en"=>"Richland", "lang"=>"en", "lat"=>32.233184, "lon"=>-90.164963),
"/US/US-MS/richton-11263" => array("type"=>"town", "pop"=>1005, "name_ru"=>"Ричтон", "name_en"=>"Richton", "lang"=>"en", "lat"=>31.349365, "lon"=>-88.940651),
"/US/US-MS/ridgeland-11264" => array("type"=>"city", "pop"=>21509, "name_ru"=>"Риджленд", "name_en"=>"Ridgeland", "lang"=>"en", "lat"=>32.421840, "lon"=>-90.131248),
"/US/US-MS/rienzi-11265" => array("type"=>"town", "pop"=>325, "name_ru"=>"Ринзи", "name_en"=>"Rienzi", "lang"=>"en", "lat"=>34.763471, "lon"=>-88.531267),
"/US/US-MS/ripley-11266" => array("type"=>"city", "pop"=>5646, "name_ru"=>"Рипли", "name_en"=>"Ripley", "lang"=>"en", "lat"=>34.729596, "lon"=>-88.944441),
"/US/US-MS/rolling-fork-11267" => array("type"=>"city", "pop"=>2082, "name_ru"=>"Роллинг Форк", "name_en"=>"Rolling Fork", "lang"=>"en", "lat"=>32.906916, "lon"=>-90.876971),
"/US/US-MS/rosedale-11268" => array("type"=>"city", "pop"=>2330, "name_ru"=>"Роуздейл", "name_en"=>"Rosedale", "lang"=>"en", "lat"=>33.853480, "lon"=>-91.032180),
"/US/US-MS/roxie-11269" => array("type"=>"town", "pop"=>557, "name_ru"=>"Рокси", "name_en"=>"Roxie", "lang"=>"en", "lat"=>31.505077, "lon"=>-91.067216),
"/US/US-MS/ruleville-11270" => array("type"=>"city", "pop"=>2831, "name_ru"=>"Рулвилл", "name_en"=>"Ruleville", "lang"=>"en", "lat"=>33.724972, "lon"=>-90.550097),
"/US/US-MS/sallis-11272" => array("type"=>"town", "pop"=>108, "name_ru"=>"Саллис", "name_en"=>"Sallis", "lang"=>"en", "lat"=>33.021655, "lon"=>-89.764021),
"/US/US-MS/saltillo-11273" => array("type"=>"town", "pop"=>4279, "name_ru"=>"Салтилло", "name_en"=>"Saltillo", "lang"=>"en", "lat"=>34.383264, "lon"=>-88.689077),
"/US/US-MS/sandersville-11274" => array("type"=>"town", "pop"=>807, "name_ru"=>"Сэндерсвилл", "name_en"=>"Sandersville", "lang"=>"en", "lat"=>31.787624, "lon"=>-89.036132),
"/US/US-MS/sardis-11275" => array("type"=>"town", "pop"=>2009, "name_ru"=>"Сардис", "name_en"=>"Sardis", "lang"=>"en", "lat"=>34.434699, "lon"=>-89.912652),
"/US/US-MS/satartia-11276" => array("type"=>"village", "pop"=>66, "name_ru"=>"Сатартия", "name_en"=>"Satartia", "lang"=>"en", "lat"=>32.670954, "lon"=>-90.543880),
"/US/US-MS/schlater-11277" => array("type"=>"town", "pop"=>352, "name_ru"=>"Шлейтр", "name_en"=>"Schlater", "lang"=>"en", "lat"=>33.639061, "lon"=>-90.349374),
"/US/US-MS/scooba-11278" => array("type"=>"town", "pop"=>582, "name_ru"=>"Скуба", "name_en"=>"Scooba", "lang"=>"en", "lat"=>32.831442, "lon"=>-88.477668),
"/US/US-MS/sebastopol-11279" => array("type"=>"town", "pop"=>234, "name_ru"=>"Себастопол", "name_en"=>"Sebastopol", "lang"=>"en", "lat"=>32.569098, "lon"=>-89.333820),
"/US/US-MS/seminary-11280" => array("type"=>"town", "pop"=>352, "name_ru"=>"Семинари", "name_en"=>"Seminary", "lang"=>"en", "lat"=>31.559202, "lon"=>-89.495310),
"/US/US-MS/senatobia-11281" => array("type"=>"city", "pop"=>7398, "name_ru"=>"Сенатобия", "name_en"=>"Senatobia", "lang"=>"en", "lat"=>34.617512, "lon"=>-89.972147),
"/US/US-MS/shannon-11282" => array("type"=>"town", "pop"=>1761, "name_ru"=>"Шэннон", "name_en"=>"Shannon", "lang"=>"en", "lat"=>34.113740, "lon"=>-88.694183),
"/US/US-MS/shaw-11283" => array("type"=>"city", "pop"=>2135, "name_ru"=>"Шо", "name_en"=>"Shaw", "lang"=>"en", "lat"=>33.600301, "lon"=>-90.771459),
"/US/US-MS/shelby-11284" => array("type"=>"city", "pop"=>2600, "name_ru"=>"Шелби", "name_en"=>"Shelby", "lang"=>"en", "lat"=>33.951576, "lon"=>-90.765718),
"/US/US-MS/sherman-11285" => array("type"=>"town", "pop"=>623, "name_ru"=>"Шерман", "name_en"=>"Sherman", "lang"=>"en", "lat"=>34.359353, "lon"=>-88.838676),
"/US/US-MS/shoreline-park-11286" => array("type"=>"cdp", "pop"=>2775, "name_ru"=>"Шорлайн Парк", "name_en"=>"Shoreline Park", "lang"=>"en", "lat"=>30.319412, "lon"=>-89.409285),
"/US/US-MS/shubuta-11287" => array("type"=>"town", "pop"=>636, "name_ru"=>"Шубута", "name_en"=>"Shubuta", "lang"=>"en", "lat"=>31.860753, "lon"=>-88.702361),
"/US/US-MS/shuqualak-11288" => array("type"=>"town", "pop"=>524, "name_ru"=>"Шуквалак", "name_en"=>"Shuqualak", "lang"=>"en", "lat"=>32.979449, "lon"=>-88.568749),
"/US/US-MS/sidon-11289" => array("type"=>"town", "pop"=>566, "name_ru"=>"Сидон", "name_en"=>"Sidon", "lang"=>"en", "lat"=>33.407420, "lon"=>-90.208227),
"/US/US-MS/silver-city-11290" => array("type"=>"town", "pop"=>298, "name_ru"=>"Сильвер Сити", "name_en"=>"Silver City", "lang"=>"en", "lat"=>33.096309, "lon"=>-90.497249),
"/US/US-MS/silver-creek-11291" => array("type"=>"town", "pop"=>235, "name_ru"=>"Силвер Крик", "name_en"=>"Silver Creek", "lang"=>"en", "lat"=>31.603299, "lon"=>-90.004643),
"/US/US-MS/slate-spring-11292" => array("type"=>"village", "pop"=>118, "name_ru"=>"Слейт Спринг", "name_en"=>"Slate Spring", "lang"=>"en", "lat"=>33.742629, "lon"=>-89.373116),
"/US/US-MS/sledge-11293" => array("type"=>"town", "pop"=>453, "name_ru"=>"Следж", "name_en"=>"Sledge", "lang"=>"en", "lat"=>34.432500, "lon"=>-90.221339),
"/US/US-MS/smithville-11294" => array("type"=>"town", "pop"=>868, "name_ru"=>"Смитвилл", "name_en"=>"Smithville", "lang"=>"en", "lat"=>34.067039, "lon"=>-88.398132),
"/US/US-MS/soso-11295" => array("type"=>"town", "pop"=>387, "name_ru"=>"Сосо", "name_en"=>"Soso", "lang"=>"en", "lat"=>31.754575, "lon"=>-89.276357),
"/US/US-MS/southaven-11296" => array("type"=>"city", "pop"=>44076, "name_ru"=>"Саутхейвен", "name_en"=>"Southaven", "lang"=>"en", "lat"=>34.974351, "lon"=>-89.999007),
"/US/US-MS/st.-martin-11271" => array("type"=>"cdp", "pop"=>6349, "name_ru"=>"Сент Мартин", "name_en"=>"St. Martin", "lang"=>"en", "lat"=>30.437700, "lon"=>-88.868174),
"/US/US-MS/standing-pine-11297" => array("type"=>"cdp", "pop"=>346, "name_ru"=>"Стэндинг Пайн", "name_en"=>"Standing Pine", "lang"=>"en", "lat"=>32.673534, "lon"=>-89.451536),
"/US/US-MS/starkville-11298" => array("type"=>"city", "pop"=>24187, "name_ru"=>"Старквилл", "name_en"=>"Starkville", "lang"=>"en", "lat"=>33.456723, "lon"=>-88.822160),
"/US/US-MS/state-line-11299" => array("type"=>"town", "pop"=>546, "name_ru"=>"Стейт Лейн", "name_en"=>"State Line", "lang"=>"en", "lat"=>31.435727, "lon"=>-88.474618),
"/US/US-MS/stonewall-11300" => array("type"=>"town", "pop"=>1075, "name_ru"=>"Стоунуолл", "name_en"=>"Stonewall", "lang"=>"en", "lat"=>32.137541, "lon"=>-88.795909),
"/US/US-MS/sturgis-11301" => array("type"=>"town", "pop"=>188, "name_ru"=>"Стерджис", "name_en"=>"Sturgis", "lang"=>"en", "lat"=>33.343904, "lon"=>-89.046556),
"/US/US-MS/summit-11302" => array("type"=>"town", "pop"=>1624, "name_ru"=>"Саммит", "name_en"=>"Summit", "lang"=>"en", "lat"=>31.284662, "lon"=>-90.466152),
"/US/US-MS/sumner-11303" => array("type"=>"town", "pop"=>349, "name_ru"=>"Саммер", "name_en"=>"Sumner", "lang"=>"en", "lat"=>33.969754, "lon"=>-90.369545),
"/US/US-MS/sumrall-11304" => array("type"=>"town", "pop"=>1352, "name_ru"=>"Самролл", "name_en"=>"Sumrall", "lang"=>"en", "lat"=>31.418507, "lon"=>-89.546491),
"/US/US-MS/sunflower-11305" => array("type"=>"town", "pop"=>1166, "name_ru"=>"Санфлауэр", "name_en"=>"Sunflower", "lang"=>"en", "lat"=>33.546623, "lon"=>-90.538386),
"/US/US-MS/sylvarena-11306" => array("type"=>"village", "pop"=>117, "name_ru"=>"Сильварена", "name_en"=>"Sylvarena", "lang"=>"en", "lat"=>32.012968, "lon"=>-89.380376),
"/US/US-MS/taylor-11307" => array("type"=>"village", "pop"=>309, "name_ru"=>"Тейлор", "name_en"=>"Taylor", "lang"=>"en", "lat"=>34.272116, "lon"=>-89.582358),
"/US/US-MS/taylorsville-11308" => array("type"=>"town", "pop"=>1265, "name_ru"=>"Тейлорсвилл", "name_en"=>"Taylorsville", "lang"=>"en", "lat"=>31.831021, "lon"=>-89.432027),
"/US/US-MS/tchula-11309" => array("type"=>"town", "pop"=>2215, "name_ru"=>"Чула", "name_en"=>"Tchula", "lang"=>"en", "lat"=>33.183071, "lon"=>-90.222646),
"/US/US-MS/terry-11310" => array("type"=>"town", "pop"=>825, "name_ru"=>"Терри", "name_en"=>"Terry", "lang"=>"en", "lat"=>32.102803, "lon"=>-90.298841),
"/US/US-MS/thaxton-11311" => array("type"=>"town", "pop"=>601, "name_ru"=>"Тэкстон", "name_en"=>"Thaxton", "lang"=>"en", "lat"=>34.308154, "lon"=>-89.174924),
"/US/US-MS/tillatoba-11312" => array("type"=>"village", "pop"=>124, "name_ru"=>"Тиллатоба", "name_en"=>"Tillatoba", "lang"=>"en", "lat"=>33.984949, "lon"=>-89.895733),
"/US/US-MS/tishomingo-11313" => array("type"=>"town", "pop"=>311, "name_ru"=>"Тишоминго", "name_en"=>"Tishomingo", "lang"=>"en", "lat"=>34.636515, "lon"=>-88.229899),
"/US/US-MS/toccopola-11314" => array("type"=>"town", "pop"=>283, "name_ru"=>"Токкопола", "name_en"=>"Toccopola", "lang"=>"en", "lat"=>34.255668, "lon"=>-89.234400),
"/US/US-MS/tremont-11315" => array("type"=>"town", "pop"=>393, "name_ru"=>"Тремонт", "name_en"=>"Tremont", "lang"=>"en", "lat"=>34.233514, "lon"=>-88.249026),
"/US/US-MS/tucker-11316" => array("type"=>"cdp", "pop"=>459, "name_ru"=>"Такер", "name_en"=>"Tucker", "lang"=>"en", "lat"=>32.707472, "lon"=>-89.048896),
"/US/US-MS/tunica-11317" => array("type"=>"town", "pop"=>1070, "name_ru"=>"Тунис", "name_en"=>"Tunica", "lang"=>"en", "lat"=>34.688234, "lon"=>-90.381519),
"/US/US-MS/tupelo-11318" => array("type"=>"city", "pop"=>36233, "name_ru"=>"Тупело", "name_en"=>"Tupelo", "lang"=>"en", "lat"=>34.263488, "lon"=>-88.731995),
"/US/US-MS/tutwiler-11319" => array("type"=>"town", "pop"=>1247, "name_ru"=>"Татуилер", "name_en"=>"Tutwiler", "lang"=>"en", "lat"=>34.011814, "lon"=>-90.429886),
"/US/US-MS/tylertown-11320" => array("type"=>"town", "pop"=>1918, "name_ru"=>"Тайлертаун", "name_en"=>"Tylertown", "lang"=>"en", "lat"=>31.116946, "lon"=>-90.142487),
"/US/US-MS/union-11321" => array("type"=>"town", "pop"=>2093, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>32.571457, "lon"=>-89.116306),
"/US/US-MS/utica-11322" => array("type"=>"town", "pop"=>909, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>32.103273, "lon"=>-90.622116),
"/US/US-MS/vaiden-11323" => array("type"=>"town", "pop"=>880, "name_ru"=>"Вейден", "name_en"=>"Vaiden", "lang"=>"en", "lat"=>33.332190, "lon"=>-89.752526),
"/US/US-MS/vancleave-11324" => array("type"=>"cdp", "pop"=>3214, "name_ru"=>"Ванклив", "name_en"=>"Vancleave", "lang"=>"en", "lat"=>30.541737, "lon"=>-88.667095),
"/US/US-MS/vardaman-11325" => array("type"=>"town", "pop"=>1010, "name_ru"=>"Вардаман", "name_en"=>"Vardaman", "lang"=>"en", "lat"=>33.881738, "lon"=>-89.178577),
"/US/US-MS/verona-11326" => array("type"=>"town", "pop"=>3394, "name_ru"=>"Верона", "name_en"=>"Verona", "lang"=>"en", "lat"=>34.190190, "lon"=>-88.719052),
"/US/US-MS/vicksburg-11327" => array("type"=>"city", "pop"=>24974, "name_ru"=>"Виксбург", "name_en"=>"Vicksburg", "lang"=>"en", "lat"=>32.329087, "lon"=>-90.870795),
"/US/US-MS/walnut-11328" => array("type"=>"town", "pop"=>752, "name_ru"=>"Уолнат", "name_en"=>"Walnut", "lang"=>"en", "lat"=>34.948526, "lon"=>-88.900103),
"/US/US-MS/walnut-grove-11329" => array("type"=>"town", "pop"=>1476, "name_ru"=>"Уолнат Гров", "name_en"=>"Walnut Grove", "lang"=>"en", "lat"=>32.597820, "lon"=>-89.458320),
"/US/US-MS/walthall-11330" => array("type"=>"village", "pop"=>162, "name_ru"=>"Уолтолл", "name_en"=>"Walthall", "lang"=>"en", "lat"=>33.608044, "lon"=>-89.275989),
"/US/US-MS/water-valley-11331" => array("type"=>"city", "pop"=>3951, "name_ru"=>"Уотер Валли", "name_en"=>"Water Valley", "lang"=>"en", "lat"=>34.161637, "lon"=>-89.630273),
"/US/US-MS/waveland-11332" => array("type"=>"city", "pop"=>5249, "name_ru"=>"Уейвленд", "name_en"=>"Waveland", "lang"=>"en", "lat"=>30.286300, "lon"=>-89.384474),
"/US/US-MS/waynesboro-11333" => array("type"=>"city", "pop"=>5647, "name_ru"=>"Уэйнсборо", "name_en"=>"Waynesboro", "lang"=>"en", "lat"=>31.670279, "lon"=>-88.641046),
"/US/US-MS/webb-11334" => array("type"=>"town", "pop"=>492, "name_ru"=>"Уэбб", "name_en"=>"Webb", "lang"=>"en", "lat"=>33.948038, "lon"=>-90.346184),
"/US/US-MS/weir-11335" => array("type"=>"town", "pop"=>504, "name_ru"=>"Вейр", "name_en"=>"Weir", "lang"=>"en", "lat"=>33.263298, "lon"=>-89.289371),
"/US/US-MS/wesson-11336" => array("type"=>"town", "pop"=>1673, "name_ru"=>"Вессон", "name_en"=>"Wesson", "lang"=>"en", "lat"=>31.700454, "lon"=>-90.396738),
"/US/US-MS/west-11337" => array("type"=>"town", "pop"=>193, "name_ru"=>"Вэст", "name_en"=>"West", "lang"=>"en", "lat"=>33.198124, "lon"=>-89.781419),
"/US/US-MS/west-hattiesburg-11338" => array("type"=>"cdp", "pop"=>5450, "name_ru"=>"Вэст Хаттисбург", "name_en"=>"West Hattiesburg", "lang"=>"en", "lat"=>31.318813, "lon"=>-89.375024),
"/US/US-MS/west-point-11339" => array("type"=>"city", "pop"=>11292, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>33.606507, "lon"=>-88.644616),
"/US/US-MS/wiggins-11340" => array("type"=>"city", "pop"=>4901, "name_ru"=>"Уиггинс", "name_en"=>"Wiggins", "lang"=>"en", "lat"=>30.860268, "lon"=>-89.138275),
"/US/US-MS/winona-11341" => array("type"=>"city", "pop"=>4524, "name_ru"=>"Уайнона", "name_en"=>"Winona", "lang"=>"en", "lat"=>33.489447, "lon"=>-89.726052),
"/US/US-MS/winstonville-11342" => array("type"=>"village", "pop"=>291, "name_ru"=>"Уинстонвилл", "name_en"=>"Winstonville", "lang"=>"en", "lat"=>33.911421, "lon"=>-90.752730),
"/US/US-MS/woodland-11343" => array("type"=>"village", "pop"=>154, "name_ru"=>"Вудленд", "name_en"=>"Woodland", "lang"=>"en", "lat"=>33.779976, "lon"=>-89.050106),
"/US/US-MS/woodville-11344" => array("type"=>"town", "pop"=>1161, "name_ru"=>"Вудвилл", "name_en"=>"Woodville", "lang"=>"en", "lat"=>31.103343, "lon"=>-91.298771),
"/US/US-MS/wool-market-11345" => array("type"=>"cdp", "pop"=>1166, "name_ru"=>"Вул Маркет", "name_en"=>"Wool Market", "lang"=>"en", "lat"=>30.467529, "lon"=>-88.995700),
"/US/US-MS/yazoo-city-11346" => array("type"=>"city", "pop"=>11425, "name_ru"=>"Язу Сити", "name_en"=>"Yazoo City", "lang"=>"en", "lat"=>32.861367, "lon"=>-90.407695),
"/US/US-MT" => array("type"=>"state", "pop"=>967440, "name_ru"=>"Монтана", "name_en"=>"Montana", "lang"=>"en"),
"/US/US-MT/absarokee-12308" => array("type"=>"cdp", "pop"=>1067, "name_en"=>"Absarokee", "lang"=>"en", "lat"=>45.524902, "lon"=>-109.444492),
"/US/US-MT/alberton-12309" => array("type"=>"town", "pop"=>420, "name_en"=>"Alberton", "lang"=>"en", "lat"=>47.002104, "lon"=>-114.476236),
"/US/US-MT/anaconda-deer-lodge-county-12310" => array("type"=>"county", "pop"=>10278, "name_en"=>"Anaconda-Deer Lodge County", "lang"=>"en", "lat"=>46.065829, "lon"=>-113.083591),
"/US/US-MT/arlee-12311" => array("type"=>"cdp", "pop"=>489, "name_en"=>"Arlee", "lang"=>"en", "lat"=>47.169227, "lon"=>-114.088957),
"/US/US-MT/ashland-12312" => array("type"=>"cdp", "pop"=>484, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>45.623531, "lon"=>-106.314347),
"/US/US-MT/bainville-12313" => array("type"=>"town", "pop"=>146, "name_en"=>"Bainville", "lang"=>"en", "lat"=>48.140462, "lon"=>-104.218362),
"/US/US-MT/baker-12314" => array("type"=>"city", "pop"=>1634, "name_ru"=>"Бейкер", "name_en"=>"Baker", "lang"=>"en", "lat"=>46.363097, "lon"=>-104.273753),
"/US/US-MT/bearcreek-12315" => array("type"=>"town", "pop"=>83, "name_en"=>"Bearcreek", "lang"=>"en", "lat"=>45.160596, "lon"=>-109.156994),
"/US/US-MT/belgrade-12316" => array("type"=>"city", "pop"=>8185, "name_ru"=>"Белгрейд", "name_en"=>"Belgrade", "lang"=>"en", "lat"=>45.777858, "lon"=>-111.177394),
"/US/US-MT/belt-12317" => array("type"=>"city", "pop"=>592, "name_en"=>"Belt", "lang"=>"en", "lat"=>47.385970, "lon"=>-110.925797),
"/US/US-MT/big-sandy-12318" => array("type"=>"town", "pop"=>610, "name_en"=>"Big Sandy", "lang"=>"en", "lat"=>48.178900, "lon"=>-110.112877),
"/US/US-MT/big-timber-12319" => array("type"=>"city", "pop"=>1796, "name_en"=>"Big Timber", "lang"=>"en", "lat"=>45.834096, "lon"=>-109.950609),
"/US/US-MT/billings-12320" => array("type"=>"city", "pop"=>103994, "name_ru"=>"Биллингс", "name_en"=>"Billings", "lang"=>"en", "lat"=>45.791401, "lon"=>-108.538543),
"/US/US-MT/bonner-west-riverside-12321" => array("type"=>"cdp", "pop"=>1669, "name_en"=>"Bonner-West Riverside", "lang"=>"en", "lat"=>46.877793, "lon"=>-113.887978),
"/US/US-MT/boulder-12322" => array("type"=>"town", "pop"=>1442, "name_ru"=>"Боулдер", "name_en"=>"Boulder", "lang"=>"en", "lat"=>46.236053, "lon"=>-112.119457),
"/US/US-MT/bozeman-12323" => array("type"=>"city", "pop"=>39442, "name_en"=>"Bozeman", "lang"=>"en", "lat"=>45.676908, "lon"=>-111.042474),
"/US/US-MT/bridger-12324" => array("type"=>"town", "pop"=>725, "name_en"=>"Bridger", "lang"=>"en", "lat"=>45.293429, "lon"=>-108.914774),
"/US/US-MT/broadus-12325" => array("type"=>"town", "pop"=>456, "name_en"=>"Broadus", "lang"=>"en", "lat"=>45.442909, "lon"=>-105.407041),
"/US/US-MT/broadview-12326" => array("type"=>"town", "pop"=>151, "name_ru"=>"Броадвью", "name_en"=>"Broadview", "lang"=>"en", "lat"=>46.097756, "lon"=>-108.877236),
"/US/US-MT/brockton-12327" => array("type"=>"town", "pop"=>238, "name_ru"=>"Броктон", "name_en"=>"Brockton", "lang"=>"en", "lat"=>48.149101, "lon"=>-104.913600),
"/US/US-MT/browning-12328" => array("type"=>"town", "pop"=>1057, "name_ru"=>"Браунинг", "name_en"=>"Browning", "lang"=>"en", "lat"=>48.556705, "lon"=>-113.013659),
"/US/US-MT/busby-12329" => array("type"=>"cdp", "pop"=>409, "name_en"=>"Busby", "lang"=>"en", "lat"=>45.531213, "lon"=>-106.955747),
"/US/US-MT/butte-silver-bow-(remainder)-12330" => array("type"=>"(remainder)", "pop"=>33336, "name_en"=>"Butte-Silver Bow (remainder)", "lang"=>"en", "lat"=>45.895956, "lon"=>-112.657607),
"/US/US-MT/cascade-12331" => array("type"=>"town", "pop"=>773, "name_ru"=>"Каскейд", "name_en"=>"Cascade", "lang"=>"en", "lat"=>47.270702, "lon"=>-111.702470),
"/US/US-MT/charlo-12332" => array("type"=>"cdp", "pop"=>358, "name_en"=>"Charlo", "lang"=>"en", "lat"=>47.442385, "lon"=>-114.170678),
"/US/US-MT/chester-12333" => array("type"=>"town", "pop"=>704, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>48.511134, "lon"=>-110.965896),
"/US/US-MT/chinook-12334" => array("type"=>"city", "pop"=>1274, "name_en"=>"Chinook", "lang"=>"en", "lat"=>48.590404, "lon"=>-109.231233),
"/US/US-MT/choteau-12335" => array("type"=>"city", "pop"=>1691, "name_en"=>"Choteau", "lang"=>"en", "lat"=>47.812629, "lon"=>-112.179602),
"/US/US-MT/circle-12336" => array("type"=>"town", "pop"=>542, "name_ru"=>"Серкл", "name_en"=>"Circle", "lang"=>"en", "lat"=>47.417494, "lon"=>-105.586142),
"/US/US-MT/clyde-park-12337" => array("type"=>"town", "pop"=>347, "name_en"=>"Clyde Park", "lang"=>"en", "lat"=>45.883655, "lon"=>-110.605444),
"/US/US-MT/colstrip-12338" => array("type"=>"cdp", "pop"=>2345, "name_en"=>"Colstrip", "lang"=>"en", "lat"=>45.894749, "lon"=>-106.634063),
"/US/US-MT/columbia-falls-12339" => array("type"=>"city", "pop"=>5261, "name_en"=>"Columbia Falls", "lang"=>"en", "lat"=>48.371248, "lon"=>-114.187289),
"/US/US-MT/columbus-12340" => array("type"=>"town", "pop"=>1959, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>45.637179, "lon"=>-109.250614),
"/US/US-MT/conrad-12341" => array("type"=>"city", "pop"=>2501, "name_ru"=>"Конрад", "name_en"=>"Conrad", "lang"=>"en", "lat"=>48.172835, "lon"=>-111.946279),
"/US/US-MT/crow-agency-12342" => array("type"=>"cdp", "pop"=>1446, "name_en"=>"Crow Agency", "lang"=>"en", "lat"=>45.600279, "lon"=>-107.460739),
"/US/US-MT/culbertson-12343" => array("type"=>"town", "pop"=>689, "name_en"=>"Culbertson", "lang"=>"en", "lat"=>48.147277, "lon"=>-104.515404),
"/US/US-MT/cut-bank-12344" => array("type"=>"city", "pop"=>3105, "name_en"=>"Cut Bank", "lang"=>"en", "lat"=>48.634818, "lon"=>-112.330206),
"/US/US-MT/darby-12345" => array("type"=>"town", "pop"=>863, "name_en"=>"Darby", "lang"=>"en", "lat"=>46.022122, "lon"=>-114.178368),
"/US/US-MT/deer-lodge-12346" => array("type"=>"city", "pop"=>3462, "name_en"=>"Deer Lodge", "lang"=>"en", "lat"=>46.397295, "lon"=>-112.731983),
"/US/US-MT/denton-12347" => array("type"=>"town", "pop"=>278, "name_ru"=>"Дентон", "name_en"=>"Denton", "lang"=>"en", "lat"=>47.323147, "lon"=>-109.948862),
"/US/US-MT/dillon-12348" => array("type"=>"city", "pop"=>4193, "name_ru"=>"Диллон", "name_en"=>"Dillon", "lang"=>"en", "lat"=>45.216126, "lon"=>-112.634532),
"/US/US-MT/dodson-12349" => array("type"=>"town", "pop"=>102, "name_ru"=>"Додсон", "name_en"=>"Dodson", "lang"=>"en", "lat"=>48.395406, "lon"=>-108.245201),
"/US/US-MT/drummond-12350" => array("type"=>"town", "pop"=>315, "name_ru"=>"Драммонд", "name_en"=>"Drummond", "lang"=>"en", "lat"=>46.666283, "lon"=>-113.145390),
"/US/US-MT/dutton-12351" => array("type"=>"town", "pop"=>354, "name_ru"=>"Даттон", "name_en"=>"Dutton", "lang"=>"en", "lat"=>47.847604, "lon"=>-111.713580),
"/US/US-MT/east-glacier-park-village-12352" => array("type"=>"cdp", "pop"=>326, "name_en"=>"East Glacier Park Village", "lang"=>"en", "lat"=>48.447060, "lon"=>-113.222330),
"/US/US-MT/east-helena-12353" => array("type"=>"town", "pop"=>2114, "name_en"=>"East Helena", "lang"=>"en", "lat"=>46.588345, "lon"=>-111.917963),
"/US/US-MT/ekalaka-12354" => array("type"=>"town", "pop"=>370, "name_en"=>"Ekalaka", "lang"=>"en", "lat"=>45.889234, "lon"=>-104.549208),
"/US/US-MT/ennis-12355" => array("type"=>"town", "pop"=>1057, "name_en"=>"Ennis", "lang"=>"en", "lat"=>45.345343, "lon"=>-111.729705),
"/US/US-MT/eureka-12356" => array("type"=>"town", "pop"=>1013, "name_ru"=>"Юрика", "name_en"=>"Eureka", "lang"=>"en", "lat"=>48.875929, "lon"=>-115.048011),
"/US/US-MT/evergreen-12357" => array("type"=>"cdp", "pop"=>4109, "name_ru"=>"Эвергрин", "name_en"=>"Evergreen", "lang"=>"en", "lat"=>48.225676, "lon"=>-114.274992),
"/US/US-MT/fairfield-12358" => array("type"=>"town", "pop"=>616, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>47.615299, "lon"=>-111.980025),
"/US/US-MT/fairview-12359" => array("type"=>"city", "pop"=>666, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>47.850840, "lon"=>-104.051024),
"/US/US-MT/finley-point-12360" => array("type"=>"cdp", "pop"=>395, "name_en"=>"Finley Point", "lang"=>"en", "lat"=>47.738192, "lon"=>-114.062413),
"/US/US-MT/flaxville-12361" => array("type"=>"town", "pop"=>70, "name_en"=>"Flaxville", "lang"=>"en", "lat"=>48.803708, "lon"=>-105.173166),
"/US/US-MT/forsyth-12362" => array("type"=>"city", "pop"=>1857, "name_ru"=>"Форсайт", "name_en"=>"Forsyth", "lang"=>"en", "lat"=>46.266681, "lon"=>-106.677094),
"/US/US-MT/fort-belknap-12363" => array("type"=>"cdp", "pop"=>422, "name_en"=>"Fort Belknap", "lang"=>"en", "lat"=>48.483068, "lon"=>-108.766341),
"/US/US-MT/fort-benton-12364" => array("type"=>"city", "pop"=>1459, "name_en"=>"Fort Benton", "lang"=>"en", "lat"=>47.828640, "lon"=>-110.655139),
"/US/US-MT/fort-peck-12365" => array("type"=>"town", "pop"=>220, "name_en"=>"Fort Peck", "lang"=>"en", "lat"=>48.006580, "lon"=>-106.454748),
"/US/US-MT/frazer-12366" => array("type"=>"cdp", "pop"=>403, "name_en"=>"Frazer", "lang"=>"en", "lat"=>48.052478, "lon"=>-106.049872),
"/US/US-MT/froid-12367" => array("type"=>"town", "pop"=>192, "name_en"=>"Froid", "lang"=>"en", "lat"=>48.335471, "lon"=>-104.491010),
"/US/US-MT/fromberg-12368" => array("type"=>"town", "pop"=>467, "name_en"=>"Fromberg", "lang"=>"en", "lat"=>45.391129, "lon"=>-108.905668),
"/US/US-MT/geraldine-12369" => array("type"=>"town", "pop"=>243, "name_ru"=>"Джералдин", "name_en"=>"Geraldine", "lang"=>"en", "lat"=>47.603461, "lon"=>-110.265923),
"/US/US-MT/glasgow-12370" => array("type"=>"city", "pop"=>2921, "name_ru"=>"Гласго", "name_en"=>"Glasgow", "lang"=>"en", "lat"=>48.198505, "lon"=>-106.631865),
"/US/US-MT/glendive-12371" => array("type"=>"city", "pop"=>4585, "name_en"=>"Glendive", "lang"=>"en", "lat"=>47.108557, "lon"=>-104.707533),
"/US/US-MT/grass-range-12372" => array("type"=>"town", "pop"=>141, "name_en"=>"Grass Range", "lang"=>"en", "lat"=>47.026799, "lon"=>-108.802793),
"/US/US-MT/great-falls-12373" => array("type"=>"city", "pop"=>59251, "name_en"=>"Great Falls", "lang"=>"en", "lat"=>47.504800, "lon"=>-111.290624),
"/US/US-MT/hamilton-12374" => array("type"=>"city", "pop"=>4817, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>46.251388, "lon"=>-114.162076),
"/US/US-MT/hardin-12375" => array("type"=>"city", "pop"=>3487, "name_ru"=>"Хардин", "name_en"=>"Hardin", "lang"=>"en", "lat"=>45.731892, "lon"=>-107.613369),
"/US/US-MT/harlem-12376" => array("type"=>"city", "pop"=>787, "name_ru"=>"Гарлем", "name_en"=>"Harlem", "lang"=>"en", "lat"=>48.531755, "lon"=>-108.783695),
"/US/US-MT/harlowton-12377" => array("type"=>"city", "pop"=>931, "name_en"=>"Harlowton", "lang"=>"en", "lat"=>46.437218, "lon"=>-109.833991),
"/US/US-MT/havre-12378" => array("type"=>"city", "pop"=>9575, "name_en"=>"Havre", "lang"=>"en", "lat"=>48.543815, "lon"=>-109.679098),
"/US/US-MT/havre-north-12379" => array("type"=>"cdp", "pop"=>1110, "name_en"=>"Havre North", "lang"=>"en", "lat"=>48.561267, "lon"=>-109.668846),
"/US/US-MT/hays-12380" => array("type"=>"cdp", "pop"=>333, "name_en"=>"Hays", "lang"=>"en", "lat"=>47.990494, "lon"=>-108.649395),
"/US/US-MT/heart-butte-12381" => array("type"=>"cdp", "pop"=>499, "name_en"=>"Heart Butte", "lang"=>"en", "lat"=>48.290479, "lon"=>-112.832945),
"/US/US-MT/helena-12382" => array("type"=>"city", "pop"=>29351, "name_ru"=>"Хелена", "name_en"=>"Helena", "lang"=>"en", "lat"=>46.596522, "lon"=>-112.020381),
"/US/US-MT/helena-valley-northeast-12383" => array("type"=>"cdp", "pop"=>1585, "name_en"=>"Helena Valley Northeast", "lang"=>"en", "lat"=>46.703537, "lon"=>-111.958315),
"/US/US-MT/helena-valley-northwest-12384" => array("type"=>"cdp", "pop"=>1215, "name_en"=>"Helena Valley Northwest", "lang"=>"en", "lat"=>46.727308, "lon"=>-112.056176),
"/US/US-MT/helena-valley-southeast-12385" => array("type"=>"cdp", "pop"=>4601, "name_en"=>"Helena Valley Southeast", "lang"=>"en", "lat"=>46.616356, "lon"=>-111.921353),
"/US/US-MT/helena-valley-west-central-12386" => array("type"=>"cdp", "pop"=>6327, "name_en"=>"Helena Valley West Central", "lang"=>"en", "lat"=>46.660869, "lon"=>-112.058378),
"/US/US-MT/helena-west-side-12387" => array("type"=>"cdp", "pop"=>1847, "name_en"=>"Helena West Side", "lang"=>"en", "lat"=>46.596667, "lon"=>-112.107080),
"/US/US-MT/hingham-12388" => array("type"=>"town", "pop"=>151, "name_ru"=>"Хингхэм", "name_en"=>"Hingham", "lang"=>"en", "lat"=>48.555649, "lon"=>-110.420745),
"/US/US-MT/hobson-12389" => array("type"=>"town", "pop"=>211, "name_en"=>"Hobson", "lang"=>"en", "lat"=>46.998904, "lon"=>-109.873163),
"/US/US-MT/hot-springs-12390" => array("type"=>"town", "pop"=>564, "name_ru"=>"Хот Спрингс", "name_en"=>"Hot Springs", "lang"=>"en", "lat"=>47.608895, "lon"=>-114.671100),
"/US/US-MT/hysham-12391" => array("type"=>"town", "pop"=>243, "name_en"=>"Hysham", "lang"=>"en", "lat"=>46.290548, "lon"=>-107.229311),
"/US/US-MT/ismay-12392" => array("type"=>"town", "pop"=>25, "name_en"=>"Ismay", "lang"=>"en", "lat"=>46.500208, "lon"=>-104.793196),
"/US/US-MT/joliet-12393" => array("type"=>"town", "pop"=>625, "name_ru"=>"Джолиет", "name_en"=>"Joliet", "lang"=>"en", "lat"=>45.484574, "lon"=>-108.970553),
"/US/US-MT/jordan-12394" => array("type"=>"town", "pop"=>336, "name_ru"=>"Джордан", "name_en"=>"Jordan", "lang"=>"en", "lat"=>47.321151, "lon"=>-106.910021),
"/US/US-MT/judith-gap-12395" => array("type"=>"city", "pop"=>143, "name_en"=>"Judith Gap", "lang"=>"en", "lat"=>46.678930, "lon"=>-109.754006),
"/US/US-MT/kalispell-12396" => array("type"=>"city", "pop"=>21182, "name_en"=>"Kalispell", "lang"=>"en", "lat"=>48.199360, "lon"=>-114.317006),
"/US/US-MT/kevin-12397" => array("type"=>"town", "pop"=>141, "name_en"=>"Kevin", "lang"=>"en", "lat"=>48.745800, "lon"=>-111.965471),
"/US/US-MT/kicking-horse-12398" => array("type"=>"cdp", "pop"=>281, "name_en"=>"Kicking Horse", "lang"=>"en", "lat"=>47.460580, "lon"=>-114.074169),
"/US/US-MT/lame-deer-12399" => array("type"=>"cdp", "pop"=>1918, "name_en"=>"Lame Deer", "lang"=>"en", "lat"=>45.617609, "lon"=>-106.609754),
"/US/US-MT/laurel-12400" => array("type"=>"city", "pop"=>6605, "name_ru"=>"Лорел", "name_en"=>"Laurel", "lang"=>"en", "lat"=>45.674025, "lon"=>-108.771711),
"/US/US-MT/lavina-12401" => array("type"=>"town", "pop"=>218, "name_en"=>"Lavina", "lang"=>"en", "lat"=>46.296351, "lon"=>-108.939133),
"/US/US-MT/lewistown-12402" => array("type"=>"city", "pop"=>5954, "name_ru"=>"Льюистаун", "name_en"=>"Lewistown", "lang"=>"en", "lat"=>47.063562, "lon"=>-109.427667),
"/US/US-MT/libby-12403" => array("type"=>"city", "pop"=>2921, "name_en"=>"Libby", "lang"=>"en", "lat"=>48.389033, "lon"=>-115.557065),
"/US/US-MT/lima-12404" => array("type"=>"town", "pop"=>228, "name_ru"=>"Лима", "name_en"=>"Lima", "lang"=>"en", "lat"=>44.638428, "lon"=>-112.591449),
"/US/US-MT/livingston-12405" => array("type"=>"city", "pop"=>7500, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>45.662975, "lon"=>-110.563729),
"/US/US-MT/lockwood-12406" => array("type"=>"cdp", "pop"=>3967, "name_ru"=>"Локвуд", "name_en"=>"Lockwood", "lang"=>"en", "lat"=>45.819301, "lon"=>-108.414105),
"/US/US-MT/lodge-grass-12407" => array("type"=>"town", "pop"=>507, "name_en"=>"Lodge Grass", "lang"=>"en", "lat"=>45.314400, "lon"=>-107.366594),
"/US/US-MT/lolo-12408" => array("type"=>"cdp", "pop"=>2746, "name_en"=>"Lolo", "lang"=>"en", "lat"=>46.767700, "lon"=>-114.105038),
"/US/US-MT/malmstrom-afb-12409" => array("type"=>"cdp", "pop"=>5938, "name_en"=>"Malmstrom AFB", "lang"=>"en", "lat"=>47.509383, "lon"=>-111.190691),
"/US/US-MT/malta-12410" => array("type"=>"city", "pop"=>1801, "name_ru"=>"Малта", "name_en"=>"Malta", "lang"=>"en", "lat"=>48.354776, "lon"=>-107.869957),
"/US/US-MT/manhattan-12411" => array("type"=>"town", "pop"=>1622, "name_ru"=>"Манхэттен", "name_en"=>"Manhattan", "lang"=>"en", "lat"=>45.857439, "lon"=>-111.330212),
"/US/US-MT/medicine-lake-12412" => array("type"=>"town", "pop"=>213, "name_ru"=>"Медисин Лейк", "name_en"=>"Medicine Lake", "lang"=>"en", "lat"=>48.503375, "lon"=>-104.500211),
"/US/US-MT/melstone-12413" => array("type"=>"town", "pop"=>136, "name_en"=>"Melstone", "lang"=>"en", "lat"=>46.599111, "lon"=>-107.867773),
"/US/US-MT/miles-city-12414" => array("type"=>"city", "pop"=>8102, "name_en"=>"Miles City", "lang"=>"en", "lat"=>46.407318, "lon"=>-105.839096),
"/US/US-MT/missoula-12415" => array("type"=>"city", "pop"=>68202, "name_en"=>"Missoula", "lang"=>"en", "lat"=>46.868752, "lon"=>-114.007576),
"/US/US-MT/moore-12416" => array("type"=>"town", "pop"=>180, "name_ru"=>"Мур", "name_en"=>"Moore", "lang"=>"en", "lat"=>46.974929, "lon"=>-109.694877),
"/US/US-MT/muddy-12417" => array("type"=>"cdp", "pop"=>387, "name_ru"=>"Мадди", "name_en"=>"Muddy", "lang"=>"en", "lat"=>45.590044, "lon"=>-106.794740),
"/US/US-MT/nashua-12418" => array("type"=>"town", "pop"=>291, "name_ru"=>"Нашуа", "name_en"=>"Nashua", "lang"=>"en", "lat"=>48.134292, "lon"=>-106.356239),
"/US/US-MT/neihart-12419" => array("type"=>"town", "pop"=>86, "name_en"=>"Neihart", "lang"=>"en", "lat"=>46.933534, "lon"=>-110.735661),
"/US/US-MT/north-browning-12420" => array("type"=>"cdp", "pop"=>1630, "name_en"=>"North Browning", "lang"=>"en", "lat"=>48.570235, "lon"=>-113.008488),
"/US/US-MT/opheim-12421" => array("type"=>"town", "pop"=>99, "name_en"=>"Opheim", "lang"=>"en", "lat"=>48.856939, "lon"=>-106.407673),
"/US/US-MT/orchard-homes-12422" => array("type"=>"cdp", "pop"=>10317, "name_en"=>"Orchard Homes", "lang"=>"en", "lat"=>46.859900, "lon"=>-114.066407),
"/US/US-MT/outlook-12423" => array("type"=>"town", "pop"=>65, "name_en"=>"Outlook", "lang"=>"en", "lat"=>48.888799, "lon"=>-104.781798),
"/US/US-MT/pablo-12424" => array("type"=>"cdp", "pop"=>1298, "name_en"=>"Pablo", "lang"=>"en", "lat"=>47.604287, "lon"=>-114.104740),
"/US/US-MT/philipsburg-12425" => array("type"=>"town", "pop"=>911, "name_en"=>"Philipsburg", "lang"=>"en", "lat"=>46.332282, "lon"=>-113.294974),
"/US/US-MT/pinesdale-12426" => array("type"=>"town", "pop"=>822, "name_en"=>"Pinesdale", "lang"=>"en", "lat"=>46.336514, "lon"=>-114.219417),
"/US/US-MT/plains-12427" => array("type"=>"town", "pop"=>1244, "name_ru"=>"Плейнс", "name_en"=>"Plains", "lang"=>"en", "lat"=>47.460397, "lon"=>-114.883470),
"/US/US-MT/plentywood-12428" => array("type"=>"city", "pop"=>1658, "name_en"=>"Plentywood", "lang"=>"en", "lat"=>48.774016, "lon"=>-104.554651),
"/US/US-MT/plevna-12429" => array("type"=>"town", "pop"=>131, "name_ru"=>"Плевна", "name_en"=>"Plevna", "lang"=>"en", "lat"=>46.416318, "lon"=>-104.517337),
"/US/US-MT/polson-12430" => array("type"=>"city", "pop"=>5228, "name_en"=>"Polson", "lang"=>"en", "lat"=>47.689785, "lon"=>-114.157875),
"/US/US-MT/poplar-12431" => array("type"=>"city", "pop"=>866, "name_en"=>"Poplar", "lang"=>"en", "lat"=>48.110290, "lon"=>-105.195885),
"/US/US-MT/pryor-12432" => array("type"=>"cdp", "pop"=>654, "name_en"=>"Pryor", "lang"=>"en", "lat"=>45.419135, "lon"=>-108.531890),
"/US/US-MT/red-lodge-12433" => array("type"=>"city", "pop"=>2450, "name_en"=>"Red Lodge", "lang"=>"en", "lat"=>45.191514, "lon"=>-109.247086),
"/US/US-MT/rexford-12434" => array("type"=>"town", "pop"=>149, "name_ru"=>"Рексфорд", "name_en"=>"Rexford", "lang"=>"en", "lat"=>48.901010, "lon"=>-115.170344),
"/US/US-MT/richey-12435" => array("type"=>"town", "pop"=>174, "name_en"=>"Richey", "lang"=>"en", "lat"=>47.644151, "lon"=>-105.068590),
"/US/US-MT/ronan-12436" => array("type"=>"city", "pop"=>2008, "name_en"=>"Ronan", "lang"=>"en", "lat"=>47.528398, "lon"=>-114.099799),
"/US/US-MT/roundup-12437" => array("type"=>"city", "pop"=>1922, "name_en"=>"Roundup", "lang"=>"en", "lat"=>46.448482, "lon"=>-108.538839),
"/US/US-MT/ryegate-12438" => array("type"=>"town", "pop"=>280, "name_en"=>"Ryegate", "lang"=>"en", "lat"=>46.298766, "lon"=>-109.253387),
"/US/US-MT/saco-12439" => array("type"=>"town", "pop"=>189, "name_ru"=>"Сако", "name_en"=>"Saco", "lang"=>"en", "lat"=>48.457055, "lon"=>-107.340382),
"/US/US-MT/scobey-12441" => array("type"=>"city", "pop"=>891, "name_en"=>"Scobey", "lang"=>"en", "lat"=>48.790718, "lon"=>-105.420575),
"/US/US-MT/shelby-12442" => array("type"=>"city", "pop"=>3541, "name_ru"=>"Шелби", "name_en"=>"Shelby", "lang"=>"en", "lat"=>48.509520, "lon"=>-111.856678),
"/US/US-MT/sheridan-12443" => array("type"=>"town", "pop"=>704, "name_ru"=>"Шеридан", "name_en"=>"Sheridan", "lang"=>"en", "lat"=>45.457147, "lon"=>-112.191461),
"/US/US-MT/sidney-12444" => array("type"=>"city", "pop"=>4782, "name_ru"=>"Сидни", "name_en"=>"Sidney", "lang"=>"en", "lat"=>47.712379, "lon"=>-104.163718),
"/US/US-MT/south-browning-12445" => array("type"=>"cdp", "pop"=>1748, "name_en"=>"South Browning", "lang"=>"en", "lat"=>48.546199, "lon"=>-113.013455),
"/US/US-MT/st.-ignatius-12440" => array("type"=>"town", "pop"=>813, "name_en"=>"St. Ignatius", "lang"=>"en", "lat"=>47.319241, "lon"=>-114.095182),
"/US/US-MT/stanford-12446" => array("type"=>"town", "pop"=>392, "name_ru"=>"Стэнфорд", "name_en"=>"Stanford", "lang"=>"en", "lat"=>47.152007, "lon"=>-110.218439),
"/US/US-MT/starr-school-12447" => array("type"=>"cdp", "pop"=>260, "name_en"=>"Starr School", "lang"=>"en", "lat"=>48.602008, "lon"=>-113.149438),
"/US/US-MT/stevensville-12448" => array("type"=>"town", "pop"=>1984, "name_ru"=>"Стивенсвилл", "name_en"=>"Stevensville", "lang"=>"en", "lat"=>46.507813, "lon"=>-114.090930),
"/US/US-MT/sun-prairie-12450" => array("type"=>"cdp", "pop"=>1424, "name_en"=>"Sun Prairie", "lang"=>"en", "lat"=>47.536923, "lon"=>-111.480552),
"/US/US-MT/sunburst-12449" => array("type"=>"town", "pop"=>327, "name_en"=>"Sunburst", "lang"=>"en", "lat"=>48.876528, "lon"=>-111.906220),
"/US/US-MT/superior-12451" => array("type"=>"town", "pop"=>873, "name_ru"=>"Супериор", "name_en"=>"Superior", "lang"=>"en", "lat"=>47.194363, "lon"=>-114.894234),
"/US/US-MT/terry-12452" => array("type"=>"town", "pop"=>544, "name_ru"=>"Терри", "name_en"=>"Terry", "lang"=>"en", "lat"=>46.792185, "lon"=>-105.311543),
"/US/US-MT/thompson-falls-12453" => array("type"=>"city", "pop"=>1423, "name_en"=>"Thompson Falls", "lang"=>"en", "lat"=>47.599577, "lon"=>-115.333665),
"/US/US-MT/three-forks-12454" => array("type"=>"town", "pop"=>1928, "name_en"=>"Three Forks", "lang"=>"en", "lat"=>45.889372, "lon"=>-111.553677),
"/US/US-MT/townsend-12455" => array("type"=>"city", "pop"=>2030, "name_ru"=>"Таунсенд", "name_en"=>"Townsend", "lang"=>"en", "lat"=>46.318898, "lon"=>-111.519155),
"/US/US-MT/troy-12456" => array("type"=>"city", "pop"=>985, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>48.460002, "lon"=>-115.889969),
"/US/US-MT/twin-bridges-12457" => array("type"=>"town", "pop"=>428, "name_en"=>"Twin Bridges", "lang"=>"en", "lat"=>45.543359, "lon"=>-112.332646),
"/US/US-MT/valier-12458" => array("type"=>"town", "pop"=>454, "name_ru"=>"Вальер", "name_en"=>"Valier", "lang"=>"en", "lat"=>48.301852, "lon"=>-112.251209),
"/US/US-MT/virginia-city-12459" => array("type"=>"town", "pop"=>144, "name_en"=>"Virginia City", "lang"=>"en", "lat"=>45.298620, "lon"=>-111.935464),
"/US/US-MT/walkerville-12460" => array("type"=>"city", "pop"=>684, "name_ru"=>"Уолкервилл", "name_en"=>"Walkerville", "lang"=>"en", "lat"=>46.036511, "lon"=>-112.540280),
"/US/US-MT/west-yellowstone-12462" => array("type"=>"town", "pop"=>1511, "name_en"=>"West Yellowstone", "lang"=>"en", "lat"=>44.664590, "lon"=>-111.106722),
"/US/US-MT/westby-12461" => array("type"=>"town", "pop"=>139, "name_en"=>"Westby", "lang"=>"en", "lat"=>48.870342, "lon"=>-104.056641),
"/US/US-MT/white-sulphur-springs-12465" => array("type"=>"city", "pop"=>951, "name_en"=>"White Sulphur Springs", "lang"=>"en", "lat"=>46.547158, "lon"=>-110.903855),
"/US/US-MT/whitefish-12463" => array("type"=>"city", "pop"=>8281, "name_en"=>"Whitefish", "lang"=>"en", "lat"=>48.415564, "lon"=>-114.344725),
"/US/US-MT/whitehall-12464" => array("type"=>"town", "pop"=>1167, "name_ru"=>"Уайтхолл", "name_en"=>"Whitehall", "lang"=>"en", "lat"=>45.871398, "lon"=>-112.095889),
"/US/US-MT/wibaux-12466" => array("type"=>"town", "pop"=>465, "name_en"=>"Wibaux", "lang"=>"en", "lat"=>46.986978, "lon"=>-104.189357),
"/US/US-MT/winifred-12467" => array("type"=>"town", "pop"=>146, "name_en"=>"Winifred", "lang"=>"en", "lat"=>47.561548, "lon"=>-109.375343),
"/US/US-MT/winnett-12468" => array("type"=>"town", "pop"=>163, "name_en"=>"Winnett", "lang"=>"en", "lat"=>47.004714, "lon"=>-108.346183),
"/US/US-MT/wolf-point-12469" => array("type"=>"city", "pop"=>2509, "name_en"=>"Wolf Point", "lang"=>"en", "lat"=>48.092539, "lon"=>-105.640538),
"/US/US-NC" => array("type"=>"state", "pop"=>9222414, "name_ru"=>"Северная Каролина", "name_en"=>"North Carolina", "lang"=>"en"),
"/US/US-NC/aberdeen-14767" => array("type"=>"town", "pop"=>5627, "name_ru"=>"Абердин", "name_en"=>"Aberdeen", "lang"=>"en", "lat"=>35.139499, "lon"=>-79.421376),
"/US/US-NC/ahoskie-14768" => array("type"=>"town", "pop"=>4259, "name_en"=>"Ahoskie", "lang"=>"en", "lat"=>36.286799, "lon"=>-76.985859),
"/US/US-NC/alamance-14769" => array("type"=>"village", "pop"=>357, "name_en"=>"Alamance", "lang"=>"en", "lat"=>36.033350, "lon"=>-79.485187),
"/US/US-NC/albemarle-14770" => array("type"=>"city", "pop"=>15522, "name_en"=>"Albemarle", "lang"=>"en", "lat"=>35.363359, "lon"=>-80.192167),
"/US/US-NC/alexander-mills-14771" => array("type"=>"town", "pop"=>662, "name_en"=>"Alexander Mills", "lang"=>"en", "lat"=>35.309832, "lon"=>-81.857023),
"/US/US-NC/alliance-14772" => array("type"=>"town", "pop"=>725, "name_en"=>"Alliance", "lang"=>"en", "lat"=>35.144482, "lon"=>-76.807886),
"/US/US-NC/altamahaw-ossipee-14773" => array("type"=>"cdp", "pop"=>1076, "name_en"=>"Altamahaw-Ossipee", "lang"=>"en", "lat"=>36.179273, "lon"=>-79.506835),
"/US/US-NC/andrews-14774" => array("type"=>"town", "pop"=>1703, "name_ru"=>"Эндрюс", "name_en"=>"Andrews", "lang"=>"en", "lat"=>35.198531, "lon"=>-83.822859),
"/US/US-NC/angier-14775" => array("type"=>"town", "pop"=>4344, "name_en"=>"Angier", "lang"=>"en", "lat"=>35.510739, "lon"=>-78.738823),
"/US/US-NC/ansonville-14776" => array("type"=>"town", "pop"=>595, "name_en"=>"Ansonville", "lang"=>"en", "lat"=>35.104503, "lon"=>-80.109877),
"/US/US-NC/apex-14777" => array("type"=>"town", "pop"=>33075, "name_en"=>"Apex", "lang"=>"en", "lat"=>35.725791, "lon"=>-78.845034),
"/US/US-NC/arapahoe-14778" => array("type"=>"town", "pop"=>405, "name_en"=>"Arapahoe", "lang"=>"en", "lat"=>35.019102, "lon"=>-76.827318),
"/US/US-NC/archdale-14779" => array("type"=>"city", "pop"=>9332, "name_en"=>"Archdale", "lang"=>"en", "lat"=>35.903808, "lon"=>-79.966315),
"/US/US-NC/arlington-14780" => array("type"=>"town", "pop"=>795, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>36.229709, "lon"=>-80.832727),
"/US/US-NC/asheboro-14781" => array("type"=>"city", "pop"=>24609, "name_en"=>"Asheboro", "lang"=>"en", "lat"=>35.719826, "lon"=>-79.811695),
"/US/US-NC/asheville-14782" => array("type"=>"city", "pop"=>74543, "name_en"=>"Asheville", "lang"=>"en", "lat"=>35.574123, "lon"=>-82.548826),
"/US/US-NC/askewville-14783" => array("type"=>"town", "pop"=>165, "name_en"=>"Askewville", "lang"=>"en", "lat"=>36.109714, "lon"=>-76.941240),
"/US/US-NC/atkinson-14784" => array("type"=>"town", "pop"=>270, "name_ru"=>"Аткинсон", "name_en"=>"Atkinson", "lang"=>"en", "lat"=>34.527476, "lon"=>-78.170327),
"/US/US-NC/atlantic-beach-14785" => array("type"=>"town", "pop"=>1815, "name_ru"=>"Атлантик Бич", "name_en"=>"Atlantic Beach", "lang"=>"en", "lat"=>34.700247, "lon"=>-76.739476),
"/US/US-NC/aulander-14786" => array("type"=>"town", "pop"=>843, "name_en"=>"Aulander", "lang"=>"en", "lat"=>36.228866, "lon"=>-77.113641),
"/US/US-NC/aurora-14787" => array("type"=>"town", "pop"=>597, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>35.303528, "lon"=>-76.788438),
"/US/US-NC/autryville-14788" => array("type"=>"town", "pop"=>204, "name_en"=>"Autryville", "lang"=>"en", "lat"=>34.996471, "lon"=>-78.640612),
"/US/US-NC/avery-creek-14789" => array("type"=>"cdp", "pop"=>1144, "name_en"=>"Avery Creek", "lang"=>"en", "lat"=>35.462784, "lon"=>-82.574497),
"/US/US-NC/ayden-14790" => array("type"=>"town", "pop"=>5005, "name_en"=>"Ayden", "lang"=>"en", "lat"=>35.470811, "lon"=>-77.421053),
"/US/US-NC/badin-14791" => array("type"=>"cdp", "pop"=>1318, "name_en"=>"Badin", "lang"=>"en", "lat"=>35.407094, "lon"=>-80.111267),
"/US/US-NC/bailey-14792" => array("type"=>"town", "pop"=>684, "name_en"=>"Bailey", "lang"=>"en", "lat"=>35.780105, "lon"=>-78.112912),
"/US/US-NC/bakersville-14793" => array("type"=>"town", "pop"=>350, "name_en"=>"Bakersville", "lang"=>"en", "lat"=>36.014758, "lon"=>-82.158429),
"/US/US-NC/bald-head-island-14794" => array("type"=>"village", "pop"=>292, "name_en"=>"Bald Head Island", "lang"=>"en", "lat"=>33.863137, "lon"=>-77.988326),
"/US/US-NC/balfour-14795" => array("type"=>"cdp", "pop"=>1118, "name_en"=>"Balfour", "lang"=>"en", "lat"=>35.348608, "lon"=>-82.478919),
"/US/US-NC/banner-elk-14796" => array("type"=>"town", "pop"=>900, "name_en"=>"Banner Elk", "lang"=>"en", "lat"=>36.162426, "lon"=>-81.872327),
"/US/US-NC/barker-heights-14797" => array("type"=>"cdp", "pop"=>1137, "name_en"=>"Barker Heights", "lang"=>"en", "lat"=>35.309849, "lon"=>-82.441110),
"/US/US-NC/barker-ten-mile-14798" => array("type"=>"cdp", "pop"=>1087, "name_en"=>"Barker Ten Mile", "lang"=>"en", "lat"=>34.679778, "lon"=>-78.988787),
"/US/US-NC/bath-14799" => array("type"=>"town", "pop"=>268, "name_ru"=>"Бат", "name_en"=>"Bath", "lang"=>"en", "lat"=>35.465409, "lon"=>-76.815063),
"/US/US-NC/battleboro-14800" => array("type"=>"town", "pop"=>447, "name_en"=>"Battleboro", "lang"=>"en", "lat"=>36.048612, "lon"=>-77.750773),
"/US/US-NC/bayboro-14801" => array("type"=>"town", "pop"=>689, "name_en"=>"Bayboro", "lang"=>"en", "lat"=>35.143895, "lon"=>-76.768706),
"/US/US-NC/bayshore-14802" => array("type"=>"cdp", "pop"=>1661, "name_en"=>"Bayshore", "lang"=>"en", "lat"=>34.289451, "lon"=>-77.787736),
"/US/US-NC/beargrass-14803" => array("type"=>"town", "pop"=>77, "name_en"=>"Beargrass", "lang"=>"en", "lat"=>35.765829, "lon"=>-77.129773),
"/US/US-NC/beaufort-14804" => array("type"=>"town", "pop"=>4189, "name_en"=>"Beaufort", "lang"=>"en", "lat"=>34.721472, "lon"=>-76.652526),
"/US/US-NC/beech-mountain-14805" => array("type"=>"town", "pop"=>338, "name_en"=>"Beech Mountain", "lang"=>"en", "lat"=>36.210764, "lon"=>-81.889032),
"/US/US-NC/belhaven-14806" => array("type"=>"town", "pop"=>2035, "name_en"=>"Belhaven", "lang"=>"en", "lat"=>35.543701, "lon"=>-76.624726),
"/US/US-NC/belmont-14807" => array("type"=>"city", "pop"=>9374, "name_ru"=>"Белмонт", "name_en"=>"Belmont", "lang"=>"en", "lat"=>35.249936, "lon"=>-81.040242),
"/US/US-NC/belville-14808" => array("type"=>"town", "pop"=>1263, "name_en"=>"Belville", "lang"=>"en", "lat"=>34.230016, "lon"=>-77.986452),
"/US/US-NC/belwood-14809" => array("type"=>"town", "pop"=>1018, "name_en"=>"Belwood", "lang"=>"en", "lat"=>35.477169, "lon"=>-81.521208),
"/US/US-NC/benson-14810" => array("type"=>"town", "pop"=>3495, "name_ru"=>"Бенсон", "name_en"=>"Benson", "lang"=>"en", "lat"=>35.379902, "lon"=>-78.546996),
"/US/US-NC/bent-creek-14811" => array("type"=>"cdp", "pop"=>1487, "name_en"=>"Bent Creek", "lang"=>"en", "lat"=>35.510944, "lon"=>-82.614315),
"/US/US-NC/bessemer-city-14812" => array("type"=>"city", "pop"=>5562, "name_en"=>"Bessemer City", "lang"=>"en", "lat"=>35.284021, "lon"=>-81.285177),
"/US/US-NC/bethel-14813" => array("type"=>"town", "pop"=>1743, "name_ru"=>"Бетел", "name_en"=>"Bethel", "lang"=>"en", "lat"=>35.807589, "lon"=>-77.376081),
"/US/US-NC/bethlehem-14814" => array("type"=>"cdp", "pop"=>3186, "name_ru"=>"Вифлеем", "name_en"=>"Bethlehem", "lang"=>"en", "lat"=>35.813445, "lon"=>-81.292371),
"/US/US-NC/beulaville-14815" => array("type"=>"town", "pop"=>1140, "name_en"=>"Beulaville", "lang"=>"en", "lat"=>34.922311, "lon"=>-77.772515),
"/US/US-NC/biltmore-forest-14816" => array("type"=>"town", "pop"=>1559, "name_en"=>"Biltmore Forest", "lang"=>"en", "lat"=>35.533469, "lon"=>-82.541003),
"/US/US-NC/biscoe-14817" => array("type"=>"town", "pop"=>1672, "name_en"=>"Biscoe", "lang"=>"en", "lat"=>35.360310, "lon"=>-79.780517),
"/US/US-NC/black-creek-14818" => array("type"=>"town", "pop"=>697, "name_en"=>"Black Creek", "lang"=>"en", "lat"=>35.635818, "lon"=>-77.933546),
"/US/US-NC/black-mountain-14819" => array("type"=>"town", "pop"=>7877, "name_en"=>"Black Mountain", "lang"=>"en", "lat"=>35.614255, "lon"=>-82.330313),
"/US/US-NC/bladenboro-14820" => array("type"=>"town", "pop"=>1672, "name_en"=>"Bladenboro", "lang"=>"en", "lat"=>34.539846, "lon"=>-78.795330),
"/US/US-NC/blowing-rock-14821" => array("type"=>"town", "pop"=>1490, "name_en"=>"Blowing Rock", "lang"=>"en", "lat"=>36.129148, "lon"=>-81.670642),
"/US/US-NC/boger-city-14822" => array("type"=>"cdp", "pop"=>1373, "name_en"=>"Boger City", "lang"=>"en", "lat"=>35.479363, "lon"=>-81.202206),
"/US/US-NC/boiling-spring-lakes-14823" => array("type"=>"city", "pop"=>4686, "name_en"=>"Boiling Spring Lakes", "lang"=>"en", "lat"=>34.031892, "lon"=>-78.066972),
"/US/US-NC/boiling-springs-14824" => array("type"=>"town", "pop"=>3881, "name_en"=>"Boiling Springs", "lang"=>"en", "lat"=>35.248928, "lon"=>-81.667211),
"/US/US-NC/bolivia-14825" => array("type"=>"town", "pop"=>188, "name_en"=>"Bolivia", "lang"=>"en", "lat"=>34.070687, "lon"=>-78.146833),
"/US/US-NC/bolton-14826" => array("type"=>"town", "pop"=>491, "name_ru"=>"Болтон", "name_en"=>"Bolton", "lang"=>"en", "lat"=>34.318650, "lon"=>-78.402764),
"/US/US-NC/bonnie-doone-14827" => array("type"=>"cdp", "pop"=>3893, "name_en"=>"Bonnie Doone", "lang"=>"en", "lat"=>35.098888, "lon"=>-78.959847),
"/US/US-NC/boone-14828" => array("type"=>"town", "pop"=>13945, "name_ru"=>"Бун", "name_en"=>"Boone", "lang"=>"en", "lat"=>36.209499, "lon"=>-81.669862),
"/US/US-NC/boonville-14829" => array("type"=>"town", "pop"=>1116, "name_ru"=>"Бунвилл", "name_en"=>"Boonville", "lang"=>"en", "lat"=>36.233202, "lon"=>-80.708066),
"/US/US-NC/bostic-14830" => array("type"=>"town", "pop"=>321, "name_en"=>"Bostic", "lang"=>"en", "lat"=>35.361467, "lon"=>-81.835273),
"/US/US-NC/brevard-14831" => array("type"=>"city", "pop"=>6716, "name_en"=>"Brevard", "lang"=>"en", "lat"=>35.237733, "lon"=>-82.730195),
"/US/US-NC/bridgeton-14832" => array("type"=>"town", "pop"=>321, "name_ru"=>"Бриджтон", "name_en"=>"Bridgeton", "lang"=>"en", "lat"=>35.122404, "lon"=>-77.021656),
"/US/US-NC/broadway-14833" => array("type"=>"town", "pop"=>1154, "name_en"=>"Broadway", "lang"=>"en", "lat"=>35.457145, "lon"=>-79.053562),
"/US/US-NC/brogden-14834" => array("type"=>"cdp", "pop"=>3246, "name_en"=>"Brogden", "lang"=>"en", "lat"=>35.296644, "lon"=>-78.025308),
"/US/US-NC/brookford-14835" => array("type"=>"town", "pop"=>427, "name_en"=>"Brookford", "lang"=>"en", "lat"=>35.702450, "lon"=>-81.346856),
"/US/US-NC/brunswick-14836" => array("type"=>"town", "pop"=>430, "name_ru"=>"Брунсвик", "name_en"=>"Brunswick", "lang"=>"en", "lat"=>34.292653, "lon"=>-78.707192),
"/US/US-NC/bryson-city-14837" => array("type"=>"town", "pop"=>1374, "name_en"=>"Bryson City", "lang"=>"en", "lat"=>35.426527, "lon"=>-83.447333),
"/US/US-NC/buies-creek-14838" => array("type"=>"cdp", "pop"=>2085, "name_en"=>"Buies Creek", "lang"=>"en", "lat"=>35.407485, "lon"=>-78.740210),
"/US/US-NC/bunn-14839" => array("type"=>"town", "pop"=>409, "name_en"=>"Bunn", "lang"=>"en", "lat"=>35.958658, "lon"=>-78.252642),
"/US/US-NC/burgaw-14840" => array("type"=>"town", "pop"=>4275, "name_en"=>"Burgaw", "lang"=>"en", "lat"=>34.550340, "lon"=>-77.926366),
"/US/US-NC/burlington-14841" => array("type"=>"city", "pop"=>50857, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>36.084400, "lon"=>-79.447234),
"/US/US-NC/burnsville-14842" => array("type"=>"town", "pop"=>1637, "name_ru"=>"Бернсвилл", "name_en"=>"Burnsville", "lang"=>"en", "lat"=>35.916525, "lon"=>-82.297882),
"/US/US-NC/butner-14843" => array("type"=>"cdp", "pop"=>6169, "name_en"=>"Butner", "lang"=>"en", "lat"=>36.135195, "lon"=>-78.770414),
"/US/US-NC/cajah's-mountain-14844" => array("type"=>"town", "pop"=>2688, "name_en"=>"Cajah's Mountain", "lang"=>"en", "lat"=>35.849927, "lon"=>-81.538901),
"/US/US-NC/calabash-14845" => array("type"=>"town", "pop"=>1419, "name_en"=>"Calabash", "lang"=>"en", "lat"=>33.898543, "lon"=>-78.576996),
"/US/US-NC/calypso-14846" => array("type"=>"town", "pop"=>479, "name_en"=>"Calypso", "lang"=>"en", "lat"=>35.154589, "lon"=>-78.105007),
"/US/US-NC/cameron-14847" => array("type"=>"town", "pop"=>303, "name_ru"=>"Кэмерон", "name_en"=>"Cameron", "lang"=>"en", "lat"=>35.327050, "lon"=>-79.251575),
"/US/US-NC/camp-lejeune-central-14848" => array("type"=>"cdp", "pop"=>36716, "name_en"=>"Camp Lejeune Central", "lang"=>"en", "lat"=>34.695804, "lon"=>-77.366475),
"/US/US-NC/candor-14849" => array("type"=>"town", "pop"=>840, "name_en"=>"Candor", "lang"=>"en", "lat"=>35.291234, "lon"=>-79.743649),
"/US/US-NC/canton-14850" => array("type"=>"town", "pop"=>3867, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>35.546143, "lon"=>-82.843009),
"/US/US-NC/cape-carteret-14851" => array("type"=>"town", "pop"=>1389, "name_en"=>"Cape Carteret", "lang"=>"en", "lat"=>34.694316, "lon"=>-77.059460),
"/US/US-NC/carolina-beach-14852" => array("type"=>"town", "pop"=>5864, "name_en"=>"Carolina Beach", "lang"=>"en", "lat"=>34.040246, "lon"=>-77.895161),
"/US/US-NC/carrboro-14853" => array("type"=>"town", "pop"=>18162, "name_en"=>"Carrboro", "lang"=>"en", "lat"=>35.915333, "lon"=>-79.083790),
"/US/US-NC/carthage-14854" => array("type"=>"town", "pop"=>2033, "name_ru"=>"Картаж", "name_en"=>"Carthage", "lang"=>"en", "lat"=>35.347354, "lon"=>-79.418261),
"/US/US-NC/cary-14855" => array("type"=>"town", "pop"=>129545, "name_ru"=>"Кэри", "name_en"=>"Cary", "lang"=>"en", "lat"=>35.783411, "lon"=>-78.798920),
"/US/US-NC/casar-14856" => array("type"=>"town", "pop"=>311, "name_en"=>"Casar", "lang"=>"en", "lat"=>35.512909, "lon"=>-81.617716),
"/US/US-NC/castalia-14857" => array("type"=>"town", "pop"=>353, "name_ru"=>"Касталия", "name_en"=>"Castalia", "lang"=>"en", "lat"=>36.082274, "lon"=>-78.057335),
"/US/US-NC/castle-hayne-14858" => array("type"=>"cdp", "pop"=>1182, "name_en"=>"Castle Hayne", "lang"=>"en", "lat"=>34.355003, "lon"=>-77.903160),
"/US/US-NC/caswell-beach-14859" => array("type"=>"town", "pop"=>471, "name_en"=>"Caswell Beach", "lang"=>"en", "lat"=>33.904822, "lon"=>-78.048537),
"/US/US-NC/catawba-14860" => array("type"=>"town", "pop"=>808, "name_en"=>"Catawba", "lang"=>"en", "lat"=>35.706187, "lon"=>-81.075775),
"/US/US-NC/cedar-point-14861" => array("type"=>"town", "pop"=>966, "name_ru"=>"Сидар Пойнт", "name_en"=>"Cedar Point", "lang"=>"en", "lat"=>34.681423, "lon"=>-77.081468),
"/US/US-NC/centerville-14862" => array("type"=>"town", "pop"=>114, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>36.186518, "lon"=>-78.110253),
"/US/US-NC/cerro-gordo-14863" => array("type"=>"town", "pop"=>237, "name_ru"=>"Серро Гордо", "name_en"=>"Cerro Gordo", "lang"=>"en", "lat"=>34.322011, "lon"=>-78.928352),
"/US/US-NC/chadbourn-14864" => array("type"=>"town", "pop"=>2059, "name_en"=>"Chadbourn", "lang"=>"en", "lat"=>34.321875, "lon"=>-78.827249),
"/US/US-NC/chapel-hill-14865" => array("type"=>"town", "pop"=>52542, "name_en"=>"Chapel Hill", "lang"=>"en", "lat"=>35.927613, "lon"=>-79.040627),
"/US/US-NC/charlotte-14866" => array("type"=>"city", "pop"=>687456, "name_ru"=>"Шарлотт", "name_en"=>"Charlotte", "lang"=>"en", "lat"=>35.197550, "lon"=>-80.834514),
"/US/US-NC/cherryville-14867" => array("type"=>"city", "pop"=>5627, "name_en"=>"Cherryville", "lang"=>"en", "lat"=>35.381583, "lon"=>-81.375634),
"/US/US-NC/china-grove-14868" => array("type"=>"town", "pop"=>3748, "name_en"=>"China Grove", "lang"=>"en", "lat"=>35.571250, "lon"=>-80.580322),
"/US/US-NC/chocowinity-14869" => array("type"=>"town", "pop"=>720, "name_en"=>"Chocowinity", "lang"=>"en", "lat"=>35.513885, "lon"=>-77.099689),
"/US/US-NC/claremont-14870" => array("type"=>"city", "pop"=>1144, "name_ru"=>"Клэрмонт", "name_en"=>"Claremont", "lang"=>"en", "lat"=>35.713619, "lon"=>-81.150859),
"/US/US-NC/clarkton-14871" => array("type"=>"town", "pop"=>687, "name_ru"=>"Кларктон", "name_en"=>"Clarkton", "lang"=>"en", "lat"=>34.489052, "lon"=>-78.654722),
"/US/US-NC/clayton-14872" => array("type"=>"town", "pop"=>15841, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>35.647010, "lon"=>-78.458345),
"/US/US-NC/clemmons-14873" => array("type"=>"village", "pop"=>17234, "name_en"=>"Clemmons", "lang"=>"en", "lat"=>36.021243, "lon"=>-80.387704),
"/US/US-NC/cleveland-14874" => array("type"=>"town", "pop"=>838, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>35.730882, "lon"=>-80.679174),
"/US/US-NC/clinton-14875" => array("type"=>"city", "pop"=>8868, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>35.000182, "lon"=>-78.331777),
"/US/US-NC/clyde-14876" => array("type"=>"town", "pop"=>1321, "name_ru"=>"Клайд", "name_en"=>"Clyde", "lang"=>"en", "lat"=>35.533193, "lon"=>-82.911807),
"/US/US-NC/coats-14877" => array("type"=>"town", "pop"=>2134, "name_ru"=>"Коатс", "name_en"=>"Coats", "lang"=>"en", "lat"=>35.406223, "lon"=>-78.668761),
"/US/US-NC/cofield-14878" => array("type"=>"village", "pop"=>324, "name_en"=>"Cofield", "lang"=>"en", "lat"=>36.355561, "lon"=>-76.909733),
"/US/US-NC/colerain-14879" => array("type"=>"town", "pop"=>201, "name_en"=>"Colerain", "lang"=>"en", "lat"=>36.200982, "lon"=>-76.768378),
"/US/US-NC/columbia-14880" => array("type"=>"town", "pop"=>803, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>35.919710, "lon"=>-76.251213),
"/US/US-NC/columbus-14881" => array("type"=>"town", "pop"=>975, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>35.246769, "lon"=>-82.204890),
"/US/US-NC/como-14882" => array("type"=>"town", "pop"=>74, "name_ru"=>"Комо", "name_en"=>"Como", "lang"=>"en", "lat"=>36.499721, "lon"=>-77.014767),
"/US/US-NC/concord-14883" => array("type"=>"city", "pop"=>66311, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>35.407333, "lon"=>-80.594311),
"/US/US-NC/conetoe-14884" => array("type"=>"town", "pop"=>328, "name_en"=>"Conetoe", "lang"=>"en", "lat"=>35.817192, "lon"=>-77.457421),
"/US/US-NC/connelly-springs-14885" => array("type"=>"town", "pop"=>1878, "name_en"=>"Connelly Springs", "lang"=>"en", "lat"=>35.748659, "lon"=>-81.500045),
"/US/US-NC/conover-14886" => array("type"=>"city", "pop"=>7286, "name_en"=>"Conover", "lang"=>"en", "lat"=>35.709533, "lon"=>-81.217549),
"/US/US-NC/conway-14887" => array("type"=>"town", "pop"=>665, "name_ru"=>"Конвэй", "name_en"=>"Conway", "lang"=>"en", "lat"=>36.436703, "lon"=>-77.229711),
"/US/US-NC/cooleemee-14888" => array("type"=>"town", "pop"=>978, "name_en"=>"Cooleemee", "lang"=>"en", "lat"=>35.812497, "lon"=>-80.557464),
"/US/US-NC/cornelius-14889" => array("type"=>"town", "pop"=>24487, "name_en"=>"Cornelius", "lang"=>"en", "lat"=>35.480783, "lon"=>-80.868563),
"/US/US-NC/cove-city-14890" => array("type"=>"town", "pop"=>412, "name_en"=>"Cove City", "lang"=>"en", "lat"=>35.188420, "lon"=>-77.321865),
"/US/US-NC/cramerton-14891" => array("type"=>"town", "pop"=>3144, "name_en"=>"Cramerton", "lang"=>"en", "lat"=>35.234576, "lon"=>-81.074179),
"/US/US-NC/creedmoor-14892" => array("type"=>"city", "pop"=>3684, "name_en"=>"Creedmoor", "lang"=>"en", "lat"=>36.120889, "lon"=>-78.687182),
"/US/US-NC/creswell-14893" => array("type"=>"town", "pop"=>259, "name_en"=>"Creswell", "lang"=>"en", "lat"=>35.870875, "lon"=>-76.393178),
"/US/US-NC/cricket-14894" => array("type"=>"cdp", "pop"=>2015, "name_en"=>"Cricket", "lang"=>"en", "lat"=>36.166257, "lon"=>-81.191412),
"/US/US-NC/crossnore-14895" => array("type"=>"town", "pop"=>227, "name_en"=>"Crossnore", "lang"=>"en", "lat"=>36.022406, "lon"=>-81.931375),
"/US/US-NC/cullowhee-14896" => array("type"=>"cdp", "pop"=>4029, "name_en"=>"Cullowhee", "lang"=>"en", "lat"=>35.311612, "lon"=>-83.178352),
"/US/US-NC/dallas-14897" => array("type"=>"town", "pop"=>3801, "name_ru"=>"Даллас", "name_en"=>"Dallas", "lang"=>"en", "lat"=>35.314738, "lon"=>-81.176832),
"/US/US-NC/danbury-14898" => array("type"=>"town", "pop"=>112, "name_ru"=>"Данбери", "name_en"=>"Danbury", "lang"=>"en", "lat"=>36.408773, "lon"=>-80.211684),
"/US/US-NC/davidson-14899" => array("type"=>"town", "pop"=>10305, "name_en"=>"Davidson", "lang"=>"en", "lat"=>35.498480, "lon"=>-80.846507),
"/US/US-NC/dellview-14900" => array("type"=>"town", "pop"=>10, "name_en"=>"Dellview", "lang"=>"en", "lat"=>35.387133, "lon"=>-81.410136),
"/US/US-NC/denton-14901" => array("type"=>"town", "pop"=>1486, "name_ru"=>"Дентон", "name_en"=>"Denton", "lang"=>"en", "lat"=>35.636564, "lon"=>-80.111947),
"/US/US-NC/dillsboro-14902" => array("type"=>"town", "pop"=>208, "name_ru"=>"Диллсборо", "name_en"=>"Dillsboro", "lang"=>"en", "lat"=>35.368983, "lon"=>-83.252090),
"/US/US-NC/dobbins-heights-14903" => array("type"=>"town", "pop"=>876, "name_en"=>"Dobbins Heights", "lang"=>"en", "lat"=>34.906553, "lon"=>-79.693471),
"/US/US-NC/dobson-14904" => array("type"=>"town", "pop"=>1458, "name_en"=>"Dobson", "lang"=>"en", "lat"=>36.391565, "lon"=>-80.721337),
"/US/US-NC/dortches-14905" => array("type"=>"town", "pop"=>871, "name_en"=>"Dortches", "lang"=>"en", "lat"=>36.011134, "lon"=>-77.858399),
"/US/US-NC/dover-14906" => array("type"=>"town", "pop"=>418, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>35.216097, "lon"=>-77.433496),
"/US/US-NC/drexel-14907" => array("type"=>"town", "pop"=>1882, "name_ru"=>"Дрексел", "name_en"=>"Drexel", "lang"=>"en", "lat"=>35.757256, "lon"=>-81.608158),
"/US/US-NC/dublin-14908" => array("type"=>"town", "pop"=>243, "name_ru"=>"Дублин", "name_en"=>"Dublin", "lang"=>"en", "lat"=>34.655874, "lon"=>-78.723427),
"/US/US-NC/dunn-14909" => array("type"=>"city", "pop"=>10029, "name_en"=>"Dunn", "lang"=>"en", "lat"=>35.311299, "lon"=>-78.615260),
"/US/US-NC/durham-14910" => array("type"=>"city", "pop"=>223284, "name_ru"=>"Дарем", "name_en"=>"Durham", "lang"=>"en", "lat"=>35.980433, "lon"=>-78.914969),
"/US/US-NC/earl-14911" => array("type"=>"town", "pop"=>234, "name_en"=>"Earl", "lang"=>"en", "lat"=>35.194934, "lon"=>-81.534287),
"/US/US-NC/east-arcadia-14912" => array("type"=>"town", "pop"=>515, "name_en"=>"East Arcadia", "lang"=>"en", "lat"=>34.381355, "lon"=>-78.321888),
"/US/US-NC/east-bend-14913" => array("type"=>"town", "pop"=>665, "name_en"=>"East Bend", "lang"=>"en", "lat"=>36.217332, "lon"=>-80.509284),
"/US/US-NC/east-flat-rock-14914" => array("type"=>"cdp", "pop"=>3218, "name_en"=>"East Flat Rock", "lang"=>"en", "lat"=>35.284466, "lon"=>-82.419173),
"/US/US-NC/east-laurinburg-14915" => array("type"=>"town", "pop"=>279, "name_en"=>"East Laurinburg", "lang"=>"en", "lat"=>34.770187, "lon"=>-79.445275),
"/US/US-NC/east-rockingham-14917" => array("type"=>"cdp", "pop"=>4158, "name_en"=>"East Rockingham", "lang"=>"en", "lat"=>34.911272, "lon"=>-79.765490),
"/US/US-NC/east-spencer-14918" => array("type"=>"town", "pop"=>1791, "name_en"=>"East Spencer", "lang"=>"en", "lat"=>35.678642, "lon"=>-80.430777),
"/US/US-NC/eastover-14916" => array("type"=>"cdp", "pop"=>3443, "name_en"=>"Eastover", "lang"=>"en", "lat"=>35.100683, "lon"=>-78.782457),
"/US/US-NC/eden-14919" => array("type"=>"city", "pop"=>15469, "name_ru"=>"Иден", "name_en"=>"Eden", "lang"=>"en", "lat"=>36.505713, "lon"=>-79.742160),
"/US/US-NC/edenton-14920" => array("type"=>"town", "pop"=>4966, "name_en"=>"Edenton", "lang"=>"en", "lat"=>36.055772, "lon"=>-76.601869),
"/US/US-NC/elizabeth-city-14921" => array("type"=>"city", "pop"=>20025, "name_en"=>"Elizabeth City", "lang"=>"en", "lat"=>36.296049, "lon"=>-76.220583),
"/US/US-NC/elizabethtown-14922" => array("type"=>"town", "pop"=>3714, "name_ru"=>"Элизабеттаун", "name_en"=>"Elizabethtown", "lang"=>"en", "lat"=>34.624453, "lon"=>-78.610347),
"/US/US-NC/elk-park-14924" => array("type"=>"town", "pop"=>429, "name_en"=>"Elk Park", "lang"=>"en", "lat"=>36.157967, "lon"=>-81.980759),
"/US/US-NC/elkin-14923" => array("type"=>"town", "pop"=>4108, "name_en"=>"Elkin", "lang"=>"en", "lat"=>36.257504, "lon"=>-80.849151),
"/US/US-NC/ellenboro-14925" => array("type"=>"town", "pop"=>466, "name_en"=>"Ellenboro", "lang"=>"en", "lat"=>35.329566, "lon"=>-81.759134),
"/US/US-NC/ellerbe-14926" => array("type"=>"town", "pop"=>982, "name_en"=>"Ellerbe", "lang"=>"en", "lat"=>35.070737, "lon"=>-79.760813),
"/US/US-NC/elm-city-14927" => array("type"=>"town", "pop"=>1373, "name_en"=>"Elm City", "lang"=>"en", "lat"=>35.808923, "lon"=>-77.862877),
"/US/US-NC/elon-college-14928" => array("type"=>"town", "pop"=>4394, "name_en"=>"Elon College", "lang"=>"en", "lat"=>36.094150, "lon"=>-79.511673),
"/US/US-NC/elroy-14929" => array("type"=>"cdp", "pop"=>4028, "name_en"=>"Elroy", "lang"=>"en", "lat"=>35.333959, "lon"=>-77.920607),
"/US/US-NC/emerald-isle-14930" => array("type"=>"town", "pop"=>3641, "name_en"=>"Emerald Isle", "lang"=>"en", "lat"=>34.666425, "lon"=>-77.029184),
"/US/US-NC/enfield-14931" => array("type"=>"town", "pop"=>2348, "name_ru"=>"Энфилд", "name_en"=>"Enfield", "lang"=>"en", "lat"=>36.179439, "lon"=>-77.668150),
"/US/US-NC/enochville-14932" => array("type"=>"cdp", "pop"=>2901, "name_en"=>"Enochville", "lang"=>"en", "lat"=>35.519950, "lon"=>-80.665990),
"/US/US-NC/erwin-14933" => array("type"=>"town", "pop"=>4829, "name_en"=>"Erwin", "lang"=>"en", "lat"=>35.325231, "lon"=>-78.672277),
"/US/US-NC/etowah-14934" => array("type"=>"cdp", "pop"=>1997, "name_en"=>"Etowah", "lang"=>"en", "lat"=>35.315303, "lon"=>-82.595120),
"/US/US-NC/eureka-14935" => array("type"=>"town", "pop"=>239, "name_ru"=>"Юрика", "name_en"=>"Eureka", "lang"=>"en", "lat"=>35.540893, "lon"=>-77.877462),
"/US/US-NC/everetts-14936" => array("type"=>"town", "pop"=>149, "name_en"=>"Everetts", "lang"=>"en", "lat"=>35.833323, "lon"=>-77.171799),
"/US/US-NC/fair-bluff-14937" => array("type"=>"town", "pop"=>1141, "name_en"=>"Fair Bluff", "lang"=>"en", "lat"=>34.311736, "lon"=>-79.034612),
"/US/US-NC/fairmont-14938" => array("type"=>"town", "pop"=>2730, "name_ru"=>"Фэрмонт", "name_en"=>"Fairmont", "lang"=>"en", "lat"=>34.495134, "lon"=>-79.115384),
"/US/US-NC/fairplains-14939" => array("type"=>"cdp", "pop"=>2339, "name_en"=>"Fairplains", "lang"=>"en", "lat"=>36.196278, "lon"=>-81.152737),
"/US/US-NC/fairview-14940" => array("type"=>"cdp", "pop"=>5610, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>35.522837, "lon"=>-82.404547),
"/US/US-NC/faison-14941" => array("type"=>"town", "pop"=>789, "name_en"=>"Faison", "lang"=>"en", "lat"=>35.116600, "lon"=>-78.136596),
"/US/US-NC/faith-14942" => array("type"=>"town", "pop"=>714, "name_en"=>"Faith", "lang"=>"en", "lat"=>35.588062, "lon"=>-80.461231),
"/US/US-NC/falcon-14943" => array("type"=>"town", "pop"=>343, "name_ru"=>"Фэлкон", "name_en"=>"Falcon", "lang"=>"en", "lat"=>35.194399, "lon"=>-78.648583),
"/US/US-NC/falkland-14944" => array("type"=>"town", "pop"=>113, "name_en"=>"Falkland", "lang"=>"en", "lat"=>35.698552, "lon"=>-77.514779),
"/US/US-NC/fallston-14945" => array("type"=>"town", "pop"=>608, "name_ru"=>"Фоллстон", "name_en"=>"Fallston", "lang"=>"en", "lat"=>35.429199, "lon"=>-81.502407),
"/US/US-NC/farmville-14946" => array("type"=>"town", "pop"=>4615, "name_en"=>"Farmville", "lang"=>"en", "lat"=>35.594932, "lon"=>-77.590970),
"/US/US-NC/fayetteville-14947" => array("type"=>"city", "pop"=>174091, "name_ru"=>"Фейетвилл", "name_en"=>"Fayetteville", "lang"=>"en", "lat"=>35.071650, "lon"=>-78.899240),
"/US/US-NC/fearrington-14948" => array("type"=>"cdp", "pop"=>1101, "name_en"=>"Fearrington", "lang"=>"en", "lat"=>35.806335, "lon"=>-79.073511),
"/US/US-NC/flat-rock-14949" => array("type"=>"cdp", "pop"=>2910, "name_ru"=>"Флэт Рок", "name_en"=>"Flat Rock", "lang"=>"en", "lat"=>36.506963, "lon"=>-80.583212),
"/US/US-NC/fletcher-14950" => array("type"=>"town", "pop"=>4714, "name_en"=>"Fletcher", "lang"=>"en", "lat"=>35.434282, "lon"=>-82.502866),
"/US/US-NC/forest-city-14951" => array("type"=>"town", "pop"=>7172, "name_ru"=>"Форест Сити", "name_en"=>"Forest City", "lang"=>"en", "lat"=>35.336022, "lon"=>-81.869061),
"/US/US-NC/forest-oaks-14952" => array("type"=>"cdp", "pop"=>3054, "name_en"=>"Forest Oaks", "lang"=>"en", "lat"=>35.988008, "lon"=>-79.706425),
"/US/US-NC/fort-bragg-14953" => array("type"=>"cdp", "pop"=>34744, "name_ru"=>"Форт Брагг", "name_en"=>"Fort Bragg", "lang"=>"en", "lat"=>35.138533, "lon"=>-79.001022),
"/US/US-NC/fountain-14954" => array("type"=>"town", "pop"=>531, "name_ru"=>"Фонтейн", "name_en"=>"Fountain", "lang"=>"en", "lat"=>35.674142, "lon"=>-77.637077),
"/US/US-NC/four-oaks-14955" => array("type"=>"town", "pop"=>2053, "name_en"=>"Four Oaks", "lang"=>"en", "lat"=>35.447712, "lon"=>-78.425047),
"/US/US-NC/foxfire-14956" => array("type"=>"village", "pop"=>488, "name_en"=>"Foxfire", "lang"=>"en", "lat"=>35.173554, "lon"=>-79.568303),
"/US/US-NC/franklin-14957" => array("type"=>"town", "pop"=>3967, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>35.181050, "lon"=>-83.381809),
"/US/US-NC/franklinton-14958" => array("type"=>"town", "pop"=>1985, "name_ru"=>"Франклинтон", "name_en"=>"Franklinton", "lang"=>"en", "lat"=>36.103642, "lon"=>-78.456609),
"/US/US-NC/franklinville-14959" => array("type"=>"town", "pop"=>1301, "name_en"=>"Franklinville", "lang"=>"en", "lat"=>35.744991, "lon"=>-79.691766),
"/US/US-NC/fremont-14960" => array("type"=>"town", "pop"=>1430, "name_ru"=>"Фремонт", "name_en"=>"Fremont", "lang"=>"en", "lat"=>35.543251, "lon"=>-77.976035),
"/US/US-NC/fuquay-varina-14961" => array("type"=>"town", "pop"=>17002, "name_en"=>"Fuquay-Varina", "lang"=>"en", "lat"=>35.590728, "lon"=>-78.793953),
"/US/US-NC/gamewell-14962" => array("type"=>"town", "pop"=>3872, "name_en"=>"Gamewell", "lang"=>"en", "lat"=>35.867105, "lon"=>-81.600792),
"/US/US-NC/garland-14963" => array("type"=>"town", "pop"=>835, "name_ru"=>"Гарленд", "name_en"=>"Garland", "lang"=>"en", "lat"=>34.785953, "lon"=>-78.394986),
"/US/US-NC/garner-14964" => array("type"=>"town", "pop"=>27138, "name_ru"=>"Гарнер", "name_en"=>"Garner", "lang"=>"en", "lat"=>35.694519, "lon"=>-78.626913),
"/US/US-NC/garysburg-14965" => array("type"=>"town", "pop"=>1133, "name_en"=>"Garysburg", "lang"=>"en", "lat"=>36.446789, "lon"=>-77.559740),
"/US/US-NC/gaston-14966" => array("type"=>"town", "pop"=>881, "name_ru"=>"Гэстон", "name_en"=>"Gaston", "lang"=>"en", "lat"=>36.496160, "lon"=>-77.644253),
"/US/US-NC/gastonia-14967" => array("type"=>"city", "pop"=>72505, "name_en"=>"Gastonia", "lang"=>"en", "lat"=>35.252750, "lon"=>-81.171507),
"/US/US-NC/gatesville-14968" => array("type"=>"town", "pop"=>312, "name_en"=>"Gatesville", "lang"=>"en", "lat"=>36.407183, "lon"=>-76.755440),
"/US/US-NC/gibson-14969" => array("type"=>"town", "pop"=>552, "name_ru"=>"Гибсон", "name_en"=>"Gibson", "lang"=>"en", "lat"=>34.758714, "lon"=>-79.606764),
"/US/US-NC/gibsonville-14970" => array("type"=>"town", "pop"=>4738, "name_en"=>"Gibsonville", "lang"=>"en", "lat"=>36.103456, "lon"=>-79.540830),
"/US/US-NC/glen-alpine-14971" => array("type"=>"town", "pop"=>1084, "name_en"=>"Glen Alpine", "lang"=>"en", "lat"=>35.729329, "lon"=>-81.778627),
"/US/US-NC/glen-raven-14972" => array("type"=>"cdp", "pop"=>2616, "name_en"=>"Glen Raven", "lang"=>"en", "lat"=>36.120950, "lon"=>-79.470028),
"/US/US-NC/godwin-14973" => array("type"=>"town", "pop"=>115, "name_en"=>"Godwin", "lang"=>"en", "lat"=>35.217367, "lon"=>-78.682007),
"/US/US-NC/goldsboro-14974" => array("type"=>"city", "pop"=>37597, "name_ru"=>"Гоулсборо", "name_en"=>"Goldsboro", "lang"=>"en", "lat"=>35.373800, "lon"=>-77.976109),
"/US/US-NC/goldston-14975" => array("type"=>"town", "pop"=>379, "name_en"=>"Goldston", "lang"=>"en", "lat"=>35.593433, "lon"=>-79.328523),
"/US/US-NC/gorman-14976" => array("type"=>"cdp", "pop"=>1090, "name_en"=>"Gorman", "lang"=>"en", "lat"=>36.038000, "lon"=>-78.813063),
"/US/US-NC/graham-14977" => array("type"=>"city", "pop"=>14533, "name_ru"=>"Грэхэм", "name_en"=>"Graham", "lang"=>"en", "lat"=>36.063100, "lon"=>-79.390502),
"/US/US-NC/grandfather-14978" => array("type"=>"village", "pop"=>69, "name_en"=>"Grandfather", "lang"=>"en", "lat"=>36.098498, "lon"=>-81.847526),
"/US/US-NC/granite-falls-14979" => array("type"=>"town", "pop"=>4610, "name_ru"=>"Гранти Фоллс", "name_en"=>"Granite Falls", "lang"=>"en", "lat"=>35.801669, "lon"=>-81.427049),
"/US/US-NC/granite-quarry-14980" => array("type"=>"town", "pop"=>2272, "name_en"=>"Granite Quarry", "lang"=>"en", "lat"=>35.613187, "lon"=>-80.445710),
"/US/US-NC/greenevers-14981" => array("type"=>"town", "pop"=>595, "name_en"=>"Greenevers", "lang"=>"en", "lat"=>34.827017, "lon"=>-77.924354),
"/US/US-NC/greensboro-14982" => array("type"=>"city", "pop"=>250642, "name_ru"=>"Гринсборо", "name_en"=>"Greensboro", "lang"=>"en", "lat"=>36.078900, "lon"=>-79.826888),
"/US/US-NC/greenville-14983" => array("type"=>"city", "pop"=>79629, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>35.599826, "lon"=>-77.374398),
"/US/US-NC/grifton-14984" => array("type"=>"town", "pop"=>2240, "name_en"=>"Grifton", "lang"=>"en", "lat"=>35.373804, "lon"=>-77.434612),
"/US/US-NC/grimesland-14985" => array("type"=>"town", "pop"=>472, "name_en"=>"Grimesland", "lang"=>"en", "lat"=>35.562310, "lon"=>-77.191684),
"/US/US-NC/grover-14986" => array("type"=>"town", "pop"=>697, "name_ru"=>"Гровер", "name_en"=>"Grover", "lang"=>"en", "lat"=>35.170910, "lon"=>-81.451735),
"/US/US-NC/half-moon-14987" => array("type"=>"cdp", "pop"=>6306, "name_en"=>"Half Moon", "lang"=>"en", "lat"=>34.827012, "lon"=>-77.466490),
"/US/US-NC/halifax-14988" => array("type"=>"town", "pop"=>328, "name_en"=>"Halifax", "lang"=>"en", "lat"=>36.325242, "lon"=>-77.589997),
"/US/US-NC/hamilton-14989" => array("type"=>"town", "pop"=>467, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>35.943958, "lon"=>-77.208088),
"/US/US-NC/hamlet-14990" => array("type"=>"city", "pop"=>5807, "name_ru"=>"Хамлет", "name_en"=>"Hamlet", "lang"=>"en", "lat"=>34.889594, "lon"=>-79.709324),
"/US/US-NC/harkers-island-14991" => array("type"=>"cdp", "pop"=>1759, "name_en"=>"Harkers Island", "lang"=>"en", "lat"=>34.698355, "lon"=>-76.558709),
"/US/US-NC/harmony-14992" => array("type"=>"town", "pop"=>636, "name_ru"=>"Хармони", "name_en"=>"Harmony", "lang"=>"en", "lat"=>35.958397, "lon"=>-80.772797),
"/US/US-NC/harrells-14993" => array("type"=>"town", "pop"=>213, "name_en"=>"Harrells", "lang"=>"en", "lat"=>34.729604, "lon"=>-78.200981),
"/US/US-NC/harrellsville-14994" => array("type"=>"town", "pop"=>94, "name_en"=>"Harrellsville", "lang"=>"en", "lat"=>36.301328, "lon"=>-76.792724),
"/US/US-NC/harrisburg-14995" => array("type"=>"town", "pop"=>5819, "name_ru"=>"Харрисбург", "name_en"=>"Harrisburg", "lang"=>"en", "lat"=>35.321385, "lon"=>-80.665075),
"/US/US-NC/hassell-14996" => array("type"=>"town", "pop"=>68, "name_en"=>"Hassell", "lang"=>"en", "lat"=>35.908214, "lon"=>-77.277833),
"/US/US-NC/havelock-14997" => array("type"=>"city", "pop"=>21764, "name_ru"=>"Хэвлок", "name_en"=>"Havelock", "lang"=>"en", "lat"=>34.907358, "lon"=>-76.894906),
"/US/US-NC/haw-river-14998" => array("type"=>"town", "pop"=>2012, "name_en"=>"Haw River", "lang"=>"en", "lat"=>36.092419, "lon"=>-79.360083),
"/US/US-NC/hayesville-14999" => array("type"=>"town", "pop"=>463, "name_ru"=>"Хайесвилл", "name_en"=>"Hayesville", "lang"=>"en", "lat"=>35.046290, "lon"=>-83.819255),
"/US/US-NC/hays-15000" => array("type"=>"cdp", "pop"=>1522, "name_en"=>"Hays", "lang"=>"en", "lat"=>36.249125, "lon"=>-81.116121),
"/US/US-NC/hazelwood-15001" => array("type"=>"town", "pop"=>1678, "name_ru"=>"Хейзелвуд", "name_en"=>"Hazelwood", "lang"=>"en", "lat"=>35.477320, "lon"=>-83.004845),
"/US/US-NC/hemby-bridge-15002" => array("type"=>"cdp", "pop"=>1931, "name_en"=>"Hemby Bridge", "lang"=>"en", "lat"=>35.118556, "lon"=>-80.630443),
"/US/US-NC/henderson-15003" => array("type"=>"city", "pop"=>15850, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>36.324836, "lon"=>-78.411972),
"/US/US-NC/hendersonville-15004" => array("type"=>"city", "pop"=>12005, "name_en"=>"Hendersonville", "lang"=>"en", "lat"=>35.321501, "lon"=>-82.462154),
"/US/US-NC/hertford-15005" => array("type"=>"town", "pop"=>2185, "name_en"=>"Hertford", "lang"=>"en", "lat"=>36.183138, "lon"=>-76.474553),
"/US/US-NC/hickory-15006" => array("type"=>"city", "pop"=>41305, "name_ru"=>"Хикори", "name_en"=>"Hickory", "lang"=>"en", "lat"=>35.739200, "lon"=>-81.326216),
"/US/US-NC/high-point-15008" => array("type"=>"city", "pop"=>101835, "name_ru"=>"Хай Поинт", "name_en"=>"High Point", "lang"=>"en", "lat"=>35.981300, "lon"=>-79.998156),
"/US/US-NC/high-shoals-15009" => array("type"=>"city", "pop"=>802, "name_en"=>"High Shoals", "lang"=>"en", "lat"=>35.398224, "lon"=>-81.201501),
"/US/US-NC/highlands-15007" => array("type"=>"town", "pop"=>951, "name_ru"=>"Хайлендс", "name_en"=>"Highlands", "lang"=>"en", "lat"=>35.053950, "lon"=>-83.198100),
"/US/US-NC/hildebran-15010" => array("type"=>"town", "pop"=>1841, "name_en"=>"Hildebran", "lang"=>"en", "lat"=>35.715232, "lon"=>-81.421473),
"/US/US-NC/hillsborough-15011" => array("type"=>"town", "pop"=>5653, "name_ru"=>"Хилсборо", "name_en"=>"Hillsborough", "lang"=>"en", "lat"=>36.071158, "lon"=>-79.102676),
"/US/US-NC/hobgood-15012" => array("type"=>"town", "pop"=>378, "name_en"=>"Hobgood", "lang"=>"en", "lat"=>36.027446, "lon"=>-77.395226),
"/US/US-NC/hoffman-15013" => array("type"=>"town", "pop"=>599, "name_ru"=>"Хоффман", "name_en"=>"Hoffman", "lang"=>"en", "lat"=>35.032496, "lon"=>-79.549683),
"/US/US-NC/holden-beach-15014" => array("type"=>"town", "pop"=>850, "name_en"=>"Holden Beach", "lang"=>"en", "lat"=>33.909120, "lon"=>-78.312884),
"/US/US-NC/holly-ridge-15015" => array("type"=>"town", "pop"=>908, "name_en"=>"Holly Ridge", "lang"=>"en", "lat"=>34.492593, "lon"=>-77.555846),
"/US/US-NC/holly-springs-15016" => array("type"=>"town", "pop"=>20870, "name_ru"=>"Хоули Спрингс", "name_en"=>"Holly Springs", "lang"=>"en", "lat"=>35.655242, "lon"=>-78.835670),
"/US/US-NC/hookerton-15017" => array("type"=>"town", "pop"=>493, "name_en"=>"Hookerton", "lang"=>"en", "lat"=>35.422899, "lon"=>-77.589242),
"/US/US-NC/hope-mills-15018" => array("type"=>"town", "pop"=>13262, "name_en"=>"Hope Mills", "lang"=>"en", "lat"=>34.973253, "lon"=>-78.953440),
"/US/US-NC/hot-springs-15019" => array("type"=>"town", "pop"=>638, "name_ru"=>"Хот Спрингс", "name_en"=>"Hot Springs", "lang"=>"en", "lat"=>35.891647, "lon"=>-82.828259),
"/US/US-NC/hudson-15020" => array("type"=>"town", "pop"=>3045, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>35.848450, "lon"=>-81.485887),
"/US/US-NC/huntersville-15021" => array("type"=>"town", "pop"=>44054, "name_en"=>"Huntersville", "lang"=>"en", "lat"=>35.415803, "lon"=>-80.845085),
"/US/US-NC/icard-15022" => array("type"=>"cdp", "pop"=>2553, "name_en"=>"Icard", "lang"=>"en", "lat"=>35.725676, "lon"=>-81.458569),
"/US/US-NC/indian-beach-15023" => array("type"=>"town", "pop"=>92, "name_en"=>"Indian Beach", "lang"=>"en", "lat"=>34.685872, "lon"=>-76.894403),
"/US/US-NC/indian-trail-15024" => array("type"=>"town", "pop"=>18976, "name_en"=>"Indian Trail", "lang"=>"en", "lat"=>35.072384, "lon"=>-80.679204),
"/US/US-NC/jackson-15025" => array("type"=>"town", "pop"=>643, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>36.390336, "lon"=>-77.419793),
"/US/US-NC/jacksonville-15026" => array("type"=>"city", "pop"=>76233, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>34.764350, "lon"=>-77.404146),
"/US/US-NC/james-city-15027" => array("type"=>"cdp", "pop"=>4279, "name_en"=>"James City", "lang"=>"en", "lat"=>35.065451, "lon"=>-77.018123),
"/US/US-NC/jamestown-15028" => array("type"=>"town", "pop"=>3212, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>35.995608, "lon"=>-79.934467),
"/US/US-NC/jamesville-15029" => array("type"=>"town", "pop"=>453, "name_en"=>"Jamesville", "lang"=>"en", "lat"=>35.811630, "lon"=>-76.899743),
"/US/US-NC/jefferson-15030" => array("type"=>"town", "pop"=>1348, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>36.420246, "lon"=>-81.469860),
"/US/US-NC/jonesville-15031" => array("type"=>"town", "pop"=>2270, "name_ru"=>"Джонсвилл", "name_en"=>"Jonesville", "lang"=>"en", "lat"=>36.234756, "lon"=>-80.840351),
"/US/US-NC/kannapolis-15032" => array("type"=>"city", "pop"=>42521, "name_en"=>"Kannapolis", "lang"=>"en", "lat"=>35.491150, "lon"=>-80.620364),
"/US/US-NC/kelford-15033" => array("type"=>"town", "pop"=>224, "name_en"=>"Kelford", "lang"=>"en", "lat"=>36.180771, "lon"=>-77.224229),
"/US/US-NC/kenansville-15034" => array("type"=>"town", "pop"=>1182, "name_en"=>"Kenansville", "lang"=>"en", "lat"=>34.961953, "lon"=>-77.965952),
"/US/US-NC/kenly-15035" => array("type"=>"town", "pop"=>1945, "name_en"=>"Kenly", "lang"=>"en", "lat"=>35.593418, "lon"=>-78.125272),
"/US/US-NC/kernersville-15036" => array("type"=>"town", "pop"=>22392, "name_en"=>"Kernersville", "lang"=>"en", "lat"=>36.118164, "lon"=>-80.076350),
"/US/US-NC/kill-devil-hills-15037" => array("type"=>"town", "pop"=>6642, "name_en"=>"Kill Devil Hills", "lang"=>"en", "lat"=>36.016020, "lon"=>-75.667526),
"/US/US-NC/king-15038" => array("type"=>"city", "pop"=>6906, "name_en"=>"King", "lang"=>"en", "lat"=>36.277976, "lon"=>-80.357614),
"/US/US-NC/kings-mountain-15039" => array("type"=>"city", "pop"=>11175, "name_en"=>"Kings Mountain", "lang"=>"en", "lat"=>35.241507, "lon"=>-81.347970),
"/US/US-NC/kingstown-15040" => array("type"=>"town", "pop"=>847, "name_ru"=>"Кингстаун", "name_en"=>"Kingstown", "lang"=>"en", "lat"=>35.362503, "lon"=>-81.623129),
"/US/US-NC/kinston-15041" => array("type"=>"city", "pop"=>22360, "name_ru"=>"Кинстон", "name_en"=>"Kinston", "lang"=>"en", "lat"=>35.275433, "lon"=>-77.590314),
"/US/US-NC/kittrell-15042" => array("type"=>"town", "pop"=>145, "name_en"=>"Kittrell", "lang"=>"en", "lat"=>36.221902, "lon"=>-78.441464),
"/US/US-NC/kitty-hawk-15043" => array("type"=>"town", "pop"=>3260, "name_en"=>"Kitty Hawk", "lang"=>"en", "lat"=>36.070215, "lon"=>-75.719466),
"/US/US-NC/knightdale-15044" => array("type"=>"town", "pop"=>8174, "name_en"=>"Knightdale", "lang"=>"en", "lat"=>35.792525, "lon"=>-78.486377),
"/US/US-NC/kure-beach-15045" => array("type"=>"town", "pop"=>2519, "name_en"=>"Kure Beach", "lang"=>"en", "lat"=>33.996040, "lon"=>-77.910678),
"/US/US-NC/la-grange-15046" => array("type"=>"town", "pop"=>2757, "name_ru"=>"Ла Гранж", "name_en"=>"La Grange", "lang"=>"en", "lat"=>35.305853, "lon"=>-77.788362),
"/US/US-NC/lake-junaluska-15047" => array("type"=>"cdp", "pop"=>2482, "name_en"=>"Lake Junaluska", "lang"=>"en", "lat"=>35.527726, "lon"=>-82.976636),
"/US/US-NC/lake-lure-15048" => array("type"=>"town", "pop"=>1012, "name_en"=>"Lake Lure", "lang"=>"en", "lat"=>35.440574, "lon"=>-82.206649),
"/US/US-NC/lake-waccamaw-15049" => array("type"=>"town", "pop"=>1445, "name_en"=>"Lake Waccamaw", "lang"=>"en", "lat"=>34.319635, "lon"=>-78.513374),
"/US/US-NC/landis-15050" => array("type"=>"town", "pop"=>3127, "name_en"=>"Landis", "lang"=>"en", "lat"=>35.546300, "lon"=>-80.611890),
"/US/US-NC/lansing-15051" => array("type"=>"town", "pop"=>147, "name_ru"=>"Ленсинг", "name_en"=>"Lansing", "lang"=>"en", "lat"=>36.499817, "lon"=>-81.509817),
"/US/US-NC/lasker-15052" => array("type"=>"town", "pop"=>90, "name_en"=>"Lasker", "lang"=>"en", "lat"=>36.349732, "lon"=>-77.305595),
"/US/US-NC/lattimore-15053" => array("type"=>"town", "pop"=>419, "name_en"=>"Lattimore", "lang"=>"en", "lat"=>35.317277, "lon"=>-81.662018),
"/US/US-NC/laurel-park-15054" => array("type"=>"town", "pop"=>2161, "name_en"=>"Laurel Park", "lang"=>"en", "lat"=>35.311883, "lon"=>-82.500958),
"/US/US-NC/laurinburg-15055" => array("type"=>"city", "pop"=>15548, "name_en"=>"Laurinburg", "lang"=>"en", "lat"=>34.766060, "lon"=>-79.470058),
"/US/US-NC/lawndale-15056" => array("type"=>"town", "pop"=>633, "name_ru"=>"Лондейл", "name_en"=>"Lawndale", "lang"=>"en", "lat"=>35.414061, "lon"=>-81.561603),
"/US/US-NC/leggett-15057" => array("type"=>"town", "pop"=>69, "name_en"=>"Leggett", "lang"=>"en", "lat"=>35.991106, "lon"=>-77.579614),
"/US/US-NC/leland-15058" => array("type"=>"town", "pop"=>5094, "name_ru"=>"Леленд", "name_en"=>"Leland", "lang"=>"en", "lat"=>34.240576, "lon"=>-78.005656),
"/US/US-NC/lenoir-15059" => array("type"=>"city", "pop"=>17890, "name_en"=>"Lenoir", "lang"=>"en", "lat"=>35.906117, "lon"=>-81.534562),
"/US/US-NC/lewiston-woodville-15060" => array("type"=>"town", "pop"=>563, "name_en"=>"Lewiston Woodville", "lang"=>"en", "lat"=>36.114322, "lon"=>-77.179900),
"/US/US-NC/lewisville-15061" => array("type"=>"cdp", "pop"=>12878, "name_ru"=>"Льюисвилл", "name_en"=>"Lewisville", "lang"=>"en", "lat"=>36.093129, "lon"=>-80.402726),
"/US/US-NC/lexington-15062" => array("type"=>"city", "pop"=>20407, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>35.809046, "lon"=>-80.257005),
"/US/US-NC/liberty-15063" => array("type"=>"town", "pop"=>2731, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>35.854160, "lon"=>-79.571362),
"/US/US-NC/light-oak-15064" => array("type"=>"cdp", "pop"=>1339, "name_en"=>"Light Oak", "lang"=>"en", "lat"=>35.284625, "lon"=>-81.476924),
"/US/US-NC/lilesville-15065" => array("type"=>"town", "pop"=>424, "name_en"=>"Lilesville", "lang"=>"en", "lat"=>34.967774, "lon"=>-79.985241),
"/US/US-NC/lillington-15066" => array("type"=>"town", "pop"=>3283, "name_en"=>"Lillington", "lang"=>"en", "lat"=>35.397867, "lon"=>-78.816088),
"/US/US-NC/lincolnton-15067" => array("type"=>"city", "pop"=>10859, "name_ru"=>"Линкольнтон", "name_en"=>"Lincolnton", "lang"=>"en", "lat"=>35.475762, "lon"=>-81.245525),
"/US/US-NC/linden-15068" => array("type"=>"town", "pop"=>134, "name_ru"=>"Линден", "name_en"=>"Linden", "lang"=>"en", "lat"=>35.254365, "lon"=>-78.748474),
"/US/US-NC/littleton-15069" => array("type"=>"town", "pop"=>646, "name_ru"=>"Литтлтон", "name_en"=>"Littleton", "lang"=>"en", "lat"=>36.434220, "lon"=>-77.910593),
"/US/US-NC/locust-15070" => array("type"=>"city", "pop"=>2637, "name_en"=>"Locust", "lang"=>"en", "lat"=>35.258144, "lon"=>-80.430599),
"/US/US-NC/long-beach-15071" => array("type"=>"town", "pop"=>3816, "name_ru"=>"Лонг Бич", "name_en"=>"Long Beach", "lang"=>"en", "lat"=>33.918282, "lon"=>-78.149443),
"/US/US-NC/long-view-15072" => array("type"=>"town", "pop"=>4944, "name_en"=>"Long View", "lang"=>"en", "lat"=>35.726000, "lon"=>-81.384002),
"/US/US-NC/louisburg-15073" => array("type"=>"town", "pop"=>3785, "name_ru"=>"Луисбург", "name_en"=>"Louisburg", "lang"=>"en", "lat"=>36.100263, "lon"=>-78.299292),
"/US/US-NC/love-valley-15074" => array("type"=>"town", "pop"=>58, "name_en"=>"Love Valley", "lang"=>"en", "lat"=>35.990163, "lon"=>-80.985156),
"/US/US-NC/lowell-15075" => array("type"=>"city", "pop"=>2794, "name_ru"=>"Лоуэлл", "name_en"=>"Lowell", "lang"=>"en", "lat"=>35.268576, "lon"=>-81.101370),
"/US/US-NC/lowesville-15076" => array("type"=>"cdp", "pop"=>1092, "name_en"=>"Lowesville", "lang"=>"en", "lat"=>35.421190, "lon"=>-81.002614),
"/US/US-NC/lucama-15077" => array("type"=>"town", "pop"=>854, "name_en"=>"Lucama", "lang"=>"en", "lat"=>35.643525, "lon"=>-78.008623),
"/US/US-NC/lumber-bridge-15078" => array("type"=>"town", "pop"=>120, "name_en"=>"Lumber Bridge", "lang"=>"en", "lat"=>34.889423, "lon"=>-79.072972),
"/US/US-NC/lumberton-15079" => array("type"=>"city", "pop"=>22006, "name_ru"=>"Ламбертон", "name_en"=>"Lumberton", "lang"=>"en", "lat"=>34.632313, "lon"=>-79.014947),
"/US/US-NC/macclesfield-15081" => array("type"=>"town", "pop"=>408, "name_en"=>"Macclesfield", "lang"=>"en", "lat"=>35.752244, "lon"=>-77.670564),
"/US/US-NC/macon-15085" => array("type"=>"town", "pop"=>106, "name_ru"=>"Мейкон", "name_en"=>"Macon", "lang"=>"en", "lat"=>36.439014, "lon"=>-78.084665),
"/US/US-NC/madison-15086" => array("type"=>"town", "pop"=>2255, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>36.386230, "lon"=>-79.975532),
"/US/US-NC/maggie-valley-15087" => array("type"=>"town", "pop"=>815, "name_en"=>"Maggie Valley", "lang"=>"en", "lat"=>35.517172, "lon"=>-83.091549),
"/US/US-NC/magnolia-15088" => array("type"=>"town", "pop"=>989, "name_ru"=>"Магнолия", "name_en"=>"Magnolia", "lang"=>"en", "lat"=>34.895494, "lon"=>-78.055214),
"/US/US-NC/maiden-15089" => array("type"=>"town", "pop"=>3501, "name_en"=>"Maiden", "lang"=>"en", "lat"=>35.575776, "lon"=>-81.206146),
"/US/US-NC/manteo-15090" => array("type"=>"town", "pop"=>1301, "name_en"=>"Manteo", "lang"=>"en", "lat"=>35.898990, "lon"=>-75.666383),
"/US/US-NC/mar-mac-15093" => array("type"=>"cdp", "pop"=>3282, "name_en"=>"Mar-Mac", "lang"=>"en", "lat"=>35.334851, "lon"=>-78.056042),
"/US/US-NC/marietta-15091" => array("type"=>"town", "pop"=>169, "name_ru"=>"Мариетта", "name_en"=>"Marietta", "lang"=>"en", "lat"=>34.368909, "lon"=>-79.127437),
"/US/US-NC/marion-15092" => array("type"=>"city", "pop"=>5144, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>35.682550, "lon"=>-82.008591),
"/US/US-NC/mars-hill-15095" => array("type"=>"town", "pop"=>1772, "name_en"=>"Mars Hill", "lang"=>"en", "lat"=>35.828150, "lon"=>-82.550552),
"/US/US-NC/marshall-15094" => array("type"=>"town", "pop"=>832, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>35.796738, "lon"=>-82.684031),
"/US/US-NC/marshville-15096" => array("type"=>"town", "pop"=>3208, "name_en"=>"Marshville", "lang"=>"en", "lat"=>34.987093, "lon"=>-80.368236),
"/US/US-NC/masonboro-15097" => array("type"=>"cdp", "pop"=>7010, "name_en"=>"Masonboro", "lang"=>"en", "lat"=>34.169149, "lon"=>-77.866246),
"/US/US-NC/matthews-15098" => array("type"=>"town", "pop"=>26901, "name_ru"=>"Мэттьюс", "name_en"=>"Matthews", "lang"=>"en", "lat"=>35.122010, "lon"=>-80.713190),
"/US/US-NC/maxton-15099" => array("type"=>"town", "pop"=>2696, "name_en"=>"Maxton", "lang"=>"en", "lat"=>34.737191, "lon"=>-79.351699),
"/US/US-NC/mayodan-15100" => array("type"=>"town", "pop"=>2581, "name_en"=>"Mayodan", "lang"=>"en", "lat"=>36.413512, "lon"=>-79.972880),
"/US/US-NC/maysville-15101" => array("type"=>"town", "pop"=>959, "name_ru"=>"Мейсвилл", "name_en"=>"Maysville", "lang"=>"en", "lat"=>34.903738, "lon"=>-77.230946),
"/US/US-NC/mcadenville-15080" => array("type"=>"town", "pop"=>664, "name_en"=>"McAdenville", "lang"=>"en", "lat"=>35.264720, "lon"=>-81.080016),
"/US/US-NC/mcdonald-15082" => array("type"=>"town", "pop"=>106, "name_ru"=>"Макдональд", "name_en"=>"McDonald", "lang"=>"en", "lat"=>34.553081, "lon"=>-79.176298),
"/US/US-NC/mcfarlan-15083" => array("type"=>"town", "pop"=>83, "name_en"=>"McFarlan", "lang"=>"en", "lat"=>34.813293, "lon"=>-79.976171),
"/US/US-NC/mcleansville-15084" => array("type"=>"cdp", "pop"=>1154, "name_en"=>"McLeansville", "lang"=>"en", "lat"=>36.103616, "lon"=>-79.660543),
"/US/US-NC/mebane-15102" => array("type"=>"city", "pop"=>10624, "name_en"=>"Mebane", "lang"=>"en", "lat"=>36.091189, "lon"=>-79.275306),
"/US/US-NC/mesic-15103" => array("type"=>"town", "pop"=>238, "name_en"=>"Mesic", "lang"=>"en", "lat"=>35.204519, "lon"=>-76.648289),
"/US/US-NC/micro-15104" => array("type"=>"town", "pop"=>537, "name_en"=>"Micro", "lang"=>"en", "lat"=>35.562643, "lon"=>-78.204164),
"/US/US-NC/middleburg-15105" => array("type"=>"town", "pop"=>160, "name_ru"=>"Миддлберг", "name_en"=>"Middleburg", "lang"=>"en", "lat"=>36.398400, "lon"=>-78.324798),
"/US/US-NC/middlesex-15106" => array("type"=>"town", "pop"=>858, "name_en"=>"Middlesex", "lang"=>"en", "lat"=>35.787675, "lon"=>-78.204598),
"/US/US-NC/millers-creek-15107" => array("type"=>"cdp", "pop"=>1787, "name_en"=>"Millers Creek", "lang"=>"en", "lat"=>36.188934, "lon"=>-81.237554),
"/US/US-NC/milton-15108" => array("type"=>"town", "pop"=>124, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>36.536540, "lon"=>-79.208427),
"/US/US-NC/minnesott-beach-15109" => array("type"=>"town", "pop"=>359, "name_en"=>"Minnesott Beach", "lang"=>"en", "lat"=>34.980307, "lon"=>-76.818749),
"/US/US-NC/mint-hill-15110" => array("type"=>"town", "pop"=>20369, "name_en"=>"Mint Hill", "lang"=>"en", "lat"=>35.176880, "lon"=>-80.661749),
"/US/US-NC/mocksville-15111" => array("type"=>"town", "pop"=>4641, "name_en"=>"Mocksville", "lang"=>"en", "lat"=>35.897531, "lon"=>-80.555993),
"/US/US-NC/monroe-15112" => array("type"=>"city", "pop"=>32543, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>34.987064, "lon"=>-80.541826),
"/US/US-NC/montreat-15113" => array("type"=>"town", "pop"=>696, "name_en"=>"Montreat", "lang"=>"en", "lat"=>35.646526, "lon"=>-82.300637),
"/US/US-NC/mooresboro-15114" => array("type"=>"city", "pop"=>331, "name_en"=>"Mooresboro", "lang"=>"en", "lat"=>35.299060, "lon"=>-81.699563),
"/US/US-NC/mooresville-15115" => array("type"=>"town", "pop"=>21954, "name_ru"=>"Мурсвилл", "name_en"=>"Mooresville", "lang"=>"en", "lat"=>35.579391, "lon"=>-80.812154),
"/US/US-NC/moravian-falls-15116" => array("type"=>"cdp", "pop"=>1736, "name_en"=>"Moravian Falls", "lang"=>"en", "lat"=>36.107178, "lon"=>-81.180673),
"/US/US-NC/morehead-city-15117" => array("type"=>"town", "pop"=>9513, "name_en"=>"Morehead City", "lang"=>"en", "lat"=>34.724684, "lon"=>-76.731995),
"/US/US-NC/morganton-15118" => array("type"=>"city", "pop"=>17194, "name_ru"=>"Моргантон", "name_en"=>"Morganton", "lang"=>"en", "lat"=>35.741598, "lon"=>-81.698417),
"/US/US-NC/morrisville-15119" => array("type"=>"town", "pop"=>13699, "name_ru"=>"Моррисвилл", "name_en"=>"Morrisville", "lang"=>"en", "lat"=>35.833057, "lon"=>-78.833062),
"/US/US-NC/morven-15120" => array("type"=>"town", "pop"=>541, "name_ru"=>"Морвен", "name_en"=>"Morven", "lang"=>"en", "lat"=>34.863265, "lon"=>-80.000413),
"/US/US-NC/mount-airy-15123" => array("type"=>"city", "pop"=>9501, "name_ru"=>"Маунт Эйри", "name_en"=>"Mount Airy", "lang"=>"en", "lat"=>36.503840, "lon"=>-80.613141),
"/US/US-NC/mount-gilead-15124" => array("type"=>"town", "pop"=>1396, "name_en"=>"Mount Gilead", "lang"=>"en", "lat"=>35.215063, "lon"=>-80.003035),
"/US/US-NC/mount-holly-15125" => array("type"=>"city", "pop"=>10081, "name_en"=>"Mount Holly", "lang"=>"en", "lat"=>35.305185, "lon"=>-81.031384),
"/US/US-NC/mount-olive-15126" => array("type"=>"town", "pop"=>4389, "name_ru"=>"Маунт Олив", "name_en"=>"Mount Olive", "lang"=>"en", "lat"=>35.196407, "lon"=>-78.067675),
"/US/US-NC/mount-pleasant-15127" => array("type"=>"town", "pop"=>1632, "name_ru"=>"Маунт Плезант", "name_en"=>"Mount Pleasant", "lang"=>"en", "lat"=>35.402487, "lon"=>-80.436100),
"/US/US-NC/mountain-home-15121" => array("type"=>"cdp", "pop"=>1898, "name_ru"=>"Маунтин Хоум", "name_en"=>"Mountain Home", "lang"=>"en", "lat"=>35.369921, "lon"=>-82.504433),
"/US/US-NC/mountain-view-15122" => array("type"=>"cdp", "pop"=>3697, "name_ru"=>"Маунтин Вью", "name_en"=>"Mountain View", "lang"=>"en", "lat"=>35.683003, "lon"=>-81.369177),
"/US/US-NC/mulberry-15128" => array("type"=>"cdp", "pop"=>2339, "name_ru"=>"Малберри", "name_en"=>"Mulberry", "lang"=>"en", "lat"=>36.227713, "lon"=>-81.166137),
"/US/US-NC/murfreesboro-15129" => array("type"=>"town", "pop"=>2380, "name_ru"=>"Мерфрисборо", "name_en"=>"Murfreesboro", "lang"=>"en", "lat"=>36.440638, "lon"=>-77.099127),
"/US/US-NC/murphy-15130" => array("type"=>"town", "pop"=>1559, "name_ru"=>"Мерфи", "name_en"=>"Murphy", "lang"=>"en", "lat"=>35.093327, "lon"=>-84.027441),
"/US/US-NC/myrtle-grove-15131" => array("type"=>"cdp", "pop"=>4275, "name_ru"=>"Миртл Гроув", "name_en"=>"Myrtle Grove", "lang"=>"en", "lat"=>34.124384, "lon"=>-77.883749),
"/US/US-NC/nags-head-15132" => array("type"=>"town", "pop"=>3016, "name_en"=>"Nags Head", "lang"=>"en", "lat"=>35.944576, "lon"=>-75.626527),
"/US/US-NC/nashville-15133" => array("type"=>"town", "pop"=>4538, "name_ru"=>"Нэшвилл", "name_en"=>"Nashville", "lang"=>"en", "lat"=>35.967272, "lon"=>-77.956053),
"/US/US-NC/navassa-15134" => array("type"=>"town", "pop"=>1847, "name_en"=>"Navassa", "lang"=>"en", "lat"=>34.257187, "lon"=>-78.002710),
"/US/US-NC/neuse-forest-15135" => array("type"=>"cdp", "pop"=>1110, "name_en"=>"Neuse Forest", "lang"=>"en", "lat"=>34.963598, "lon"=>-76.945077),
"/US/US-NC/new-bern-15136" => array("type"=>"city", "pop"=>28586, "name_en"=>"New Bern", "lang"=>"en", "lat"=>35.119914, "lon"=>-77.078835),
"/US/US-NC/new-hope-15137" => array("type"=>"cdp", "pop"=>5694, "name_ru"=>"Нью Хоуп", "name_en"=>"New Hope", "lang"=>"en", "lat"=>35.807100, "lon"=>-78.555231),
"/US/US-NC/new-hope-15138" => array("type"=>"cdp", "pop"=>4491, "name_ru"=>"Нью Хоуп", "name_en"=>"New Hope", "lang"=>"en", "lat"=>35.384937, "lon"=>-77.904355),
"/US/US-NC/new-london-15140" => array("type"=>"town", "pop"=>612, "name_ru"=>"Нью Лондон", "name_en"=>"New London", "lang"=>"en", "lat"=>35.442553, "lon"=>-80.219846),
"/US/US-NC/new-river-station-15142" => array("type"=>"cdp", "pop"=>9732, "name_en"=>"New River Station", "lang"=>"en", "lat"=>34.709705, "lon"=>-77.442526),
"/US/US-NC/newland-15139" => array("type"=>"town", "pop"=>655, "name_en"=>"Newland", "lang"=>"en", "lat"=>36.086563, "lon"=>-81.926362),
"/US/US-NC/newport-15141" => array("type"=>"town", "pop"=>4298, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>34.787931, "lon"=>-76.860969),
"/US/US-NC/newton-15143" => array("type"=>"city", "pop"=>13372, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>35.665450, "lon"=>-81.218696),
"/US/US-NC/newton-grove-15144" => array("type"=>"town", "pop"=>630, "name_en"=>"Newton Grove", "lang"=>"en", "lat"=>35.252223, "lon"=>-78.355935),
"/US/US-NC/norlina-15145" => array("type"=>"town", "pop"=>1019, "name_en"=>"Norlina", "lang"=>"en", "lat"=>36.445663, "lon"=>-78.195502),
"/US/US-NC/norman-15146" => array("type"=>"town", "pop"=>82, "name_ru"=>"Норман", "name_en"=>"Norman", "lang"=>"en", "lat"=>35.170164, "lon"=>-79.722550),
"/US/US-NC/north-hickory-15147" => array("type"=>"cdp", "pop"=>4299, "name_en"=>"North Hickory", "lang"=>"en", "lat"=>35.777932, "lon"=>-81.323790),
"/US/US-NC/north-wilkesboro-15149" => array("type"=>"town", "pop"=>4158, "name_en"=>"North Wilkesboro", "lang"=>"en", "lat"=>36.169673, "lon"=>-81.142171),
"/US/US-NC/northlakes-15148" => array("type"=>"cdp", "pop"=>1219, "name_en"=>"Northlakes", "lang"=>"en", "lat"=>35.781606, "lon"=>-81.375229),
"/US/US-NC/norwood-15150" => array("type"=>"town", "pop"=>2415, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>35.224681, "lon"=>-80.120285),
"/US/US-NC/oak-city-15152" => array("type"=>"town", "pop"=>337, "name_en"=>"Oak City", "lang"=>"en", "lat"=>35.962063, "lon"=>-77.304916),
"/US/US-NC/oakboro-15151" => array("type"=>"town", "pop"=>1168, "name_en"=>"Oakboro", "lang"=>"en", "lat"=>35.223050, "lon"=>-80.325760),
"/US/US-NC/ocean-isle-beach-15153" => array("type"=>"town", "pop"=>528, "name_en"=>"Ocean Isle Beach", "lang"=>"en", "lat"=>33.892542, "lon"=>-78.432233),
"/US/US-NC/ogden-15154" => array("type"=>"cdp", "pop"=>3228, "name_ru"=>"Огден", "name_en"=>"Ogden", "lang"=>"en", "lat"=>34.265000, "lon"=>-77.799484),
"/US/US-NC/old-fort-15155" => array("type"=>"town", "pop"=>959, "name_en"=>"Old Fort", "lang"=>"en", "lat"=>35.632100, "lon"=>-82.175842),
"/US/US-NC/oriental-15156" => array("type"=>"town", "pop"=>815, "name_en"=>"Oriental", "lang"=>"en", "lat"=>35.034714, "lon"=>-76.684584),
"/US/US-NC/orrum-15157" => array("type"=>"town", "pop"=>102, "name_en"=>"Orrum", "lang"=>"en", "lat"=>34.465832, "lon"=>-79.010062),
"/US/US-NC/oxford-15158" => array("type"=>"city", "pop"=>8641, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>36.310937, "lon"=>-78.589576),
"/US/US-NC/pantego-15159" => array("type"=>"town", "pop"=>167, "name_en"=>"Pantego", "lang"=>"en", "lat"=>35.587375, "lon"=>-76.658540),
"/US/US-NC/parkton-15160" => array("type"=>"town", "pop"=>433, "name_en"=>"Parkton", "lang"=>"en", "lat"=>34.902233, "lon"=>-79.009954),
"/US/US-NC/parkwood-15161" => array("type"=>"cdp", "pop"=>4123, "name_ru"=>"Парквуд", "name_en"=>"Parkwood", "lang"=>"en", "lat"=>35.888533, "lon"=>-78.908842),
"/US/US-NC/parmele-15162" => array("type"=>"town", "pop"=>262, "name_en"=>"Parmele", "lang"=>"en", "lat"=>35.819690, "lon"=>-77.312433),
"/US/US-NC/patterson-springs-15163" => array("type"=>"town", "pop"=>613, "name_en"=>"Patterson Springs", "lang"=>"en", "lat"=>35.232385, "lon"=>-81.517494),
"/US/US-NC/peachland-15164" => array("type"=>"town", "pop"=>529, "name_en"=>"Peachland", "lang"=>"en", "lat"=>34.993710, "lon"=>-80.265787),
"/US/US-NC/pembroke-15165" => array("type"=>"town", "pop"=>2757, "name_ru"=>"Пемброук", "name_en"=>"Pembroke", "lang"=>"en", "lat"=>34.680773, "lon"=>-79.198689),
"/US/US-NC/pikeville-15166" => array("type"=>"town", "pop"=>704, "name_ru"=>"Пайквилл", "name_en"=>"Pikeville", "lang"=>"en", "lat"=>35.497529, "lon"=>-77.984322),
"/US/US-NC/pilot-mountain-15167" => array("type"=>"town", "pop"=>1272, "name_en"=>"Pilot Mountain", "lang"=>"en", "lat"=>36.384329, "lon"=>-80.471886),
"/US/US-NC/pine-knoll-shores-15170" => array("type"=>"town", "pop"=>1547, "name_en"=>"Pine Knoll Shores", "lang"=>"en", "lat"=>34.695679, "lon"=>-76.818216),
"/US/US-NC/pine-level-15171" => array("type"=>"town", "pop"=>1774, "name_en"=>"Pine Level", "lang"=>"en", "lat"=>35.510016, "lon"=>-78.246165),
"/US/US-NC/pinebluff-15168" => array("type"=>"town", "pop"=>1391, "name_en"=>"Pinebluff", "lang"=>"en", "lat"=>35.107120, "lon"=>-79.472309),
"/US/US-NC/pinehurst-15169" => array("type"=>"town", "pop"=>12331, "name_ru"=>"Пайнхерст", "name_en"=>"Pinehurst", "lang"=>"en", "lat"=>35.180343, "lon"=>-79.461416),
"/US/US-NC/pinetops-15172" => array("type"=>"town", "pop"=>1266, "name_en"=>"Pinetops", "lang"=>"en", "lat"=>35.791144, "lon"=>-77.637350),
"/US/US-NC/pineville-15173" => array("type"=>"town", "pop"=>6791, "name_ru"=>"Пайнвилл", "name_en"=>"Pineville", "lang"=>"en", "lat"=>35.084568, "lon"=>-80.885719),
"/US/US-NC/piney-green-15174" => array("type"=>"cdp", "pop"=>8999, "name_en"=>"Piney Green", "lang"=>"en", "lat"=>34.754433, "lon"=>-77.321620),
"/US/US-NC/pink-hill-15175" => array("type"=>"town", "pop"=>529, "name_en"=>"Pink Hill", "lang"=>"en", "lat"=>35.056413, "lon"=>-77.744839),
"/US/US-NC/pittsboro-15176" => array("type"=>"town", "pop"=>2624, "name_ru"=>"Питтсборо", "name_en"=>"Pittsboro", "lang"=>"en", "lat"=>35.719407, "lon"=>-79.174513),
"/US/US-NC/pleasant-garden-15177" => array("type"=>"cdp", "pop"=>5229, "name_en"=>"Pleasant Garden", "lang"=>"en", "lat"=>35.962171, "lon"=>-79.774094),
"/US/US-NC/pleasant-hill-15178" => array("type"=>"cdp", "pop"=>1114, "name_ru"=>"Плезант Хилл", "name_en"=>"Pleasant Hill", "lang"=>"en", "lat"=>36.253041, "lon"=>-80.885307),
"/US/US-NC/plymouth-15179" => array("type"=>"town", "pop"=>3863, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>35.860853, "lon"=>-76.748400),
"/US/US-NC/polkton-15180" => array("type"=>"town", "pop"=>2871, "name_en"=>"Polkton", "lang"=>"en", "lat"=>35.008212, "lon"=>-80.201931),
"/US/US-NC/polkville-15181" => array("type"=>"city", "pop"=>535, "name_ru"=>"Полквилл", "name_en"=>"Polkville", "lang"=>"en", "lat"=>35.416897, "lon"=>-81.644762),
"/US/US-NC/pollocksville-15182" => array("type"=>"town", "pop"=>256, "name_en"=>"Pollocksville", "lang"=>"en", "lat"=>35.006972, "lon"=>-77.221348),
"/US/US-NC/pope-afb-15183" => array("type"=>"cdp", "pop"=>2857, "name_en"=>"Pope AFB", "lang"=>"en", "lat"=>35.168657, "lon"=>-79.010556),
"/US/US-NC/poplar-tent-15184" => array("type"=>"cdp", "pop"=>3872, "name_en"=>"Poplar Tent", "lang"=>"en", "lat"=>35.409911, "lon"=>-80.635875),
"/US/US-NC/powellsville-15185" => array("type"=>"town", "pop"=>236, "name_en"=>"Powellsville", "lang"=>"en", "lat"=>36.224796, "lon"=>-76.932000),
"/US/US-NC/princeton-15186" => array("type"=>"town", "pop"=>1287, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>35.465510, "lon"=>-78.160354),
"/US/US-NC/princeville-15187" => array("type"=>"town", "pop"=>1668, "name_ru"=>"Принсвилл", "name_en"=>"Princeville", "lang"=>"en", "lat"=>35.888271, "lon"=>-77.525953),
"/US/US-NC/proctorville-15188" => array("type"=>"town", "pop"=>131, "name_en"=>"Proctorville", "lang"=>"en", "lat"=>34.475423, "lon"=>-79.037873),
"/US/US-NC/pumpkin-center-15189" => array("type"=>"cdp", "pop"=>2857, "name_en"=>"Pumpkin Center", "lang"=>"en", "lat"=>34.780391, "lon"=>-77.366559),
"/US/US-NC/raeford-15190" => array("type"=>"city", "pop"=>3517, "name_en"=>"Raeford", "lang"=>"en", "lat"=>34.979003, "lon"=>-79.224593),
"/US/US-NC/raleigh-15191" => array("type"=>"city", "pop"=>392552, "name_ru"=>"Роли", "name_en"=>"Raleigh", "lang"=>"en", "lat"=>35.821950, "lon"=>-78.658753),
"/US/US-NC/ramseur-15192" => array("type"=>"town", "pop"=>1755, "name_en"=>"Ramseur", "lang"=>"en", "lat"=>35.733398, "lon"=>-79.650637),
"/US/US-NC/randleman-15193" => array("type"=>"city", "pop"=>3686, "name_en"=>"Randleman", "lang"=>"en", "lat"=>35.817032, "lon"=>-79.802373),
"/US/US-NC/ranlo-15194" => array("type"=>"town", "pop"=>2344, "name_en"=>"Ranlo", "lang"=>"en", "lat"=>35.287048, "lon"=>-81.131266),
"/US/US-NC/raynham-15195" => array("type"=>"town", "pop"=>74, "name_en"=>"Raynham", "lang"=>"en", "lat"=>34.574864, "lon"=>-79.190473),
"/US/US-NC/red-oak-15196" => array("type"=>"town", "pop"=>2923, "name_ru"=>"Ред Оук", "name_en"=>"Red Oak", "lang"=>"en", "lat"=>36.038457, "lon"=>-77.906783),
"/US/US-NC/red-springs-15197" => array("type"=>"town", "pop"=>3525, "name_en"=>"Red Springs", "lang"=>"en", "lat"=>34.814200, "lon"=>-79.184524),
"/US/US-NC/reidsville-15198" => array("type"=>"city", "pop"=>14871, "name_ru"=>"Райдсвилл", "name_en"=>"Reidsville", "lang"=>"en", "lat"=>36.348313, "lon"=>-79.672387),
"/US/US-NC/rennert-15199" => array("type"=>"town", "pop"=>318, "name_en"=>"Rennert", "lang"=>"en", "lat"=>34.814708, "lon"=>-79.079427),
"/US/US-NC/rhodhiss-15200" => array("type"=>"town", "pop"=>902, "name_en"=>"Rhodhiss", "lang"=>"en", "lat"=>35.770538, "lon"=>-81.430956),
"/US/US-NC/rich-square-15203" => array("type"=>"town", "pop"=>871, "name_en"=>"Rich Square", "lang"=>"en", "lat"=>36.273690, "lon"=>-77.284361),
"/US/US-NC/richfield-15201" => array("type"=>"town", "pop"=>514, "name_ru"=>"Ричфилд", "name_en"=>"Richfield", "lang"=>"en", "lat"=>35.471524, "lon"=>-80.256651),
"/US/US-NC/richlands-15202" => array("type"=>"town", "pop"=>938, "name_en"=>"Richlands", "lang"=>"en", "lat"=>34.899793, "lon"=>-77.547828),
"/US/US-NC/river-bend-15204" => array("type"=>"town", "pop"=>3125, "name_en"=>"River Bend", "lang"=>"en", "lat"=>35.073108, "lon"=>-77.152238),
"/US/US-NC/river-road-15205" => array("type"=>"cdp", "pop"=>3892, "name_en"=>"River Road", "lang"=>"en", "lat"=>35.506644, "lon"=>-76.991035),
"/US/US-NC/roanoke-rapids-15206" => array("type"=>"city", "pop"=>16403, "name_en"=>"Roanoke Rapids", "lang"=>"en", "lat"=>36.453236, "lon"=>-77.651664),
"/US/US-NC/robbins-15207" => array("type"=>"city", "pop"=>1225, "name_ru"=>"Роббинс", "name_en"=>"Robbins", "lang"=>"en", "lat"=>35.432419, "lon"=>-79.584213),
"/US/US-NC/robbinsville-15208" => array("type"=>"town", "pop"=>695, "name_en"=>"Robbinsville", "lang"=>"en", "lat"=>35.322837, "lon"=>-83.806917),
"/US/US-NC/robersonville-15209" => array("type"=>"town", "pop"=>1553, "name_en"=>"Robersonville", "lang"=>"en", "lat"=>35.824865, "lon"=>-77.253025),
"/US/US-NC/rockingham-15210" => array("type"=>"city", "pop"=>8880, "name_en"=>"Rockingham", "lang"=>"en", "lat"=>34.939266, "lon"=>-79.762317),
"/US/US-NC/rockwell-15211" => array("type"=>"town", "pop"=>2012, "name_ru"=>"Рокуэлл", "name_en"=>"Rockwell", "lang"=>"en", "lat"=>35.551840, "lon"=>-80.406925),
"/US/US-NC/rocky-mount-15212" => array("type"=>"city", "pop"=>57010, "name_en"=>"Rocky Mount", "lang"=>"en", "lat"=>35.957311, "lon"=>-77.810201),
"/US/US-NC/rolesville-15213" => array("type"=>"town", "pop"=>2844, "name_en"=>"Rolesville", "lang"=>"en", "lat"=>35.927555, "lon"=>-78.457053),
"/US/US-NC/ronda-15214" => array("type"=>"town", "pop"=>456, "name_en"=>"Ronda", "lang"=>"en", "lat"=>36.221584, "lon"=>-80.943341),
"/US/US-NC/roper-15215" => array("type"=>"town", "pop"=>571, "name_en"=>"Roper", "lang"=>"en", "lat"=>35.877418, "lon"=>-76.617092),
"/US/US-NC/rose-hill-15217" => array("type"=>"town", "pop"=>1409, "name_ru"=>"Роуз Хилл", "name_en"=>"Rose Hill", "lang"=>"en", "lat"=>34.825040, "lon"=>-78.028231),
"/US/US-NC/roseboro-15216" => array("type"=>"town", "pop"=>1309, "name_en"=>"Roseboro", "lang"=>"en", "lat"=>34.953882, "lon"=>-78.512105),
"/US/US-NC/rosman-15218" => array("type"=>"town", "pop"=>558, "name_en"=>"Rosman", "lang"=>"en", "lat"=>35.143771, "lon"=>-82.821964),
"/US/US-NC/rowland-15219" => array("type"=>"town", "pop"=>1158, "name_en"=>"Rowland", "lang"=>"en", "lat"=>34.536626, "lon"=>-79.292672),
"/US/US-NC/roxboro-15220" => array("type"=>"city", "pop"=>8667, "name_en"=>"Roxboro", "lang"=>"en", "lat"=>36.401595, "lon"=>-78.980086),
"/US/US-NC/roxobel-15221" => array("type"=>"town", "pop"=>241, "name_en"=>"Roxobel", "lang"=>"en", "lat"=>36.201767, "lon"=>-77.238215),
"/US/US-NC/royal-pines-15222" => array("type"=>"cdp", "pop"=>4418, "name_en"=>"Royal Pines", "lang"=>"en", "lat"=>35.476936, "lon"=>-82.505321),
"/US/US-NC/rural-hall-15223" => array("type"=>"town", "pop"=>2706, "name_en"=>"Rural Hall", "lang"=>"en", "lat"=>36.234700, "lon"=>-80.293915),
"/US/US-NC/ruth-15224" => array("type"=>"town", "pop"=>321, "name_en"=>"Ruth", "lang"=>"en", "lat"=>35.383726, "lon"=>-81.945411),
"/US/US-NC/rutherford-college-15225" => array("type"=>"town", "pop"=>1275, "name_en"=>"Rutherford College", "lang"=>"en", "lat"=>35.753174, "lon"=>-81.527975),
"/US/US-NC/rutherfordton-15226" => array("type"=>"town", "pop"=>4040, "name_en"=>"Rutherfordton", "lang"=>"en", "lat"=>35.364809, "lon"=>-81.960852),
"/US/US-NC/salem-15230" => array("type"=>"cdp", "pop"=>2271, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>35.697352, "lon"=>-81.697486),
"/US/US-NC/salemburg-15231" => array("type"=>"town", "pop"=>485, "name_en"=>"Salemburg", "lang"=>"en", "lat"=>35.014190, "lon"=>-78.503074),
"/US/US-NC/salisbury-15232" => array("type"=>"city", "pop"=>28972, "name_ru"=>"Солсбери", "name_en"=>"Salisbury", "lang"=>"en", "lat"=>35.666500, "lon"=>-80.482902),
"/US/US-NC/saluda-15233" => array("type"=>"city", "pop"=>581, "name_en"=>"Saluda", "lang"=>"en", "lat"=>35.237908, "lon"=>-82.348146),
"/US/US-NC/sandy-creek-15234" => array("type"=>"town", "pop"=>308, "name_en"=>"Sandy Creek", "lang"=>"en", "lat"=>34.284963, "lon"=>-78.157036),
"/US/US-NC/sanford-15235" => array("type"=>"city", "pop"=>29284, "name_ru"=>"Санфорд", "name_en"=>"Sanford", "lang"=>"en", "lat"=>35.477045, "lon"=>-79.181317),
"/US/US-NC/santeetlah-15236" => array("type"=>"town", "pop"=>47, "name_en"=>"Santeetlah", "lang"=>"en", "lat"=>35.363757, "lon"=>-83.866497),
"/US/US-NC/saratoga-15237" => array("type"=>"town", "pop"=>369, "name_ru"=>"Саратога", "name_en"=>"Saratoga", "lang"=>"en", "lat"=>35.653472, "lon"=>-77.776793),
"/US/US-NC/sawmills-15238" => array("type"=>"town", "pop"=>4931, "name_en"=>"Sawmills", "lang"=>"en", "lat"=>35.818711, "lon"=>-81.481676),
"/US/US-NC/saxapahaw-15239" => array("type"=>"cdp", "pop"=>1178, "name_en"=>"Saxapahaw", "lang"=>"en", "lat"=>35.947911, "lon"=>-79.321062),
"/US/US-NC/scotland-neck-15240" => array("type"=>"town", "pop"=>2170, "name_en"=>"Scotland Neck", "lang"=>"en", "lat"=>36.130286, "lon"=>-77.421272),
"/US/US-NC/seaboard-15241" => array("type"=>"town", "pop"=>620, "name_en"=>"Seaboard", "lang"=>"en", "lat"=>36.490303, "lon"=>-77.441128),
"/US/US-NC/seagate-15242" => array("type"=>"cdp", "pop"=>5444, "name_en"=>"Seagate", "lang"=>"en", "lat"=>34.203514, "lon"=>-77.842929),
"/US/US-NC/seagrove-15243" => array("type"=>"town", "pop"=>259, "name_en"=>"Seagrove", "lang"=>"en", "lat"=>35.541485, "lon"=>-79.776871),
"/US/US-NC/sedge-garden-15244" => array("type"=>"cdp", "pop"=>2784, "name_en"=>"Sedge Garden", "lang"=>"en", "lat"=>36.088500, "lon"=>-80.139045),
"/US/US-NC/selma-15245" => array("type"=>"town", "pop"=>6948, "name_ru"=>"Сельма", "name_en"=>"Selma", "lang"=>"en", "lat"=>35.536776, "lon"=>-78.288744),
"/US/US-NC/seven-devils-15246" => array("type"=>"town", "pop"=>169, "name_en"=>"Seven Devils", "lang"=>"en", "lat"=>36.149722, "lon"=>-81.813875),
"/US/US-NC/seven-lakes-15247" => array("type"=>"cdp", "pop"=>2049, "name_en"=>"Seven Lakes", "lang"=>"en", "lat"=>35.264223, "lon"=>-79.585022),
"/US/US-NC/seven-springs-15248" => array("type"=>"town", "pop"=>85, "name_en"=>"Seven Springs", "lang"=>"en", "lat"=>35.225362, "lon"=>-77.846037),
"/US/US-NC/severn-15249" => array("type"=>"town", "pop"=>238, "name_ru"=>"Северн", "name_en"=>"Severn", "lang"=>"en", "lat"=>36.513742, "lon"=>-77.189775),
"/US/US-NC/shallotte-15250" => array("type"=>"town", "pop"=>2169, "name_en"=>"Shallotte", "lang"=>"en", "lat"=>33.975871, "lon"=>-78.385879),
"/US/US-NC/sharpsburg-15251" => array("type"=>"town", "pop"=>2405, "name_ru"=>"Шарпсбург", "name_en"=>"Sharpsburg", "lang"=>"en", "lat"=>35.867148, "lon"=>-77.832390),
"/US/US-NC/shelby-15252" => array("type"=>"city", "pop"=>21449, "name_ru"=>"Шелби", "name_en"=>"Shelby", "lang"=>"en", "lat"=>35.288907, "lon"=>-81.542040),
"/US/US-NC/sherrills-ford-15253" => array("type"=>"cdp", "pop"=>3185, "name_en"=>"Sherrills Ford", "lang"=>"en", "lat"=>35.576758, "lon"=>-80.990735),
"/US/US-NC/siler-city-15254" => array("type"=>"town", "pop"=>8564, "name_en"=>"Siler City", "lang"=>"en", "lat"=>35.724859, "lon"=>-79.462602),
"/US/US-NC/silver-city-15255" => array("type"=>"cdp", "pop"=>1343, "name_ru"=>"Сильвер Сити", "name_en"=>"Silver City", "lang"=>"en", "lat"=>34.996009, "lon"=>-79.227770),
"/US/US-NC/silver-lake-15256" => array("type"=>"cdp", "pop"=>4071, "name_ru"=>"Сильвер Лэйк", "name_en"=>"Silver Lake", "lang"=>"en", "lat"=>34.148608, "lon"=>-77.918031),
"/US/US-NC/simpson-15257" => array("type"=>"village", "pop"=>463, "name_ru"=>"Симпсон", "name_en"=>"Simpson", "lang"=>"en", "lat"=>35.575044, "lon"=>-77.278831),
"/US/US-NC/sims-15258" => array("type"=>"town", "pop"=>124, "name_ru"=>"Симс", "name_en"=>"Sims", "lang"=>"en", "lat"=>35.760380, "lon"=>-78.058855),
"/US/US-NC/smith-creek-15259" => array("type"=>"cdp", "pop"=>7461, "name_en"=>"Smith Creek", "lang"=>"en", "lat"=>34.262738, "lon"=>-77.863754),
"/US/US-NC/smithfield-15260" => array("type"=>"town", "pop"=>12965, "name_ru"=>"Смитфилд", "name_en"=>"Smithfield", "lang"=>"en", "lat"=>35.507322, "lon"=>-78.344108),
"/US/US-NC/sneads-ferry-15261" => array("type"=>"cdp", "pop"=>2031, "name_en"=>"Sneads Ferry", "lang"=>"en", "lat"=>34.553706, "lon"=>-77.378741),
"/US/US-NC/snow-hill-15262" => array("type"=>"town", "pop"=>1564, "name_ru"=>"Сноу Хилл", "name_en"=>"Snow Hill", "lang"=>"en", "lat"=>35.451438, "lon"=>-77.674517),
"/US/US-NC/south-gastonia-15265" => array("type"=>"cdp", "pop"=>5487, "name_en"=>"South Gastonia", "lang"=>"en", "lat"=>35.210184, "lon"=>-81.208726),
"/US/US-NC/south-henderson-15266" => array("type"=>"cdp", "pop"=>1374, "name_en"=>"South Henderson", "lang"=>"en", "lat"=>36.302453, "lon"=>-78.405583),
"/US/US-NC/south-rosemary-15268" => array("type"=>"cdp", "pop"=>1955, "name_en"=>"South Rosemary", "lang"=>"en", "lat"=>36.451326, "lon"=>-77.697540),
"/US/US-NC/south-weldon-15269" => array("type"=>"cdp", "pop"=>1640, "name_en"=>"South Weldon", "lang"=>"en", "lat"=>36.403173, "lon"=>-77.619042),
"/US/US-NC/southern-pines-15263" => array("type"=>"town", "pop"=>12675, "name_en"=>"Southern Pines", "lang"=>"en", "lat"=>35.183549, "lon"=>-79.402215),
"/US/US-NC/southern-shores-15264" => array("type"=>"town", "pop"=>2587, "name_en"=>"Southern Shores", "lang"=>"en", "lat"=>36.118707, "lon"=>-75.733183),
"/US/US-NC/southport-15267" => array("type"=>"city", "pop"=>3080, "name_ru"=>"Саутпорт", "name_en"=>"Southport", "lang"=>"en", "lat"=>33.924073, "lon"=>-78.022003),
"/US/US-NC/sparta-15270" => array("type"=>"town", "pop"=>1775, "name_ru"=>"Спарта", "name_en"=>"Sparta", "lang"=>"en", "lat"=>36.503753, "lon"=>-81.121296),
"/US/US-NC/speed-15271" => array("type"=>"town", "pop"=>66, "name_ru"=>"Спид", "name_en"=>"Speed", "lang"=>"en", "lat"=>35.968307, "lon"=>-77.444259),
"/US/US-NC/spencer-15272" => array("type"=>"town", "pop"=>3390, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>35.694350, "lon"=>-80.433076),
"/US/US-NC/spencer-mountain-15273" => array("type"=>"town", "pop"=>68, "name_en"=>"Spencer Mountain", "lang"=>"en", "lat"=>35.308856, "lon"=>-81.111354),
"/US/US-NC/spindale-15274" => array("type"=>"town", "pop"=>3878, "name_en"=>"Spindale", "lang"=>"en", "lat"=>35.360287, "lon"=>-81.924699),
"/US/US-NC/spring-hope-15275" => array("type"=>"town", "pop"=>1288, "name_en"=>"Spring Hope", "lang"=>"en", "lat"=>35.944535, "lon"=>-78.109445),
"/US/US-NC/spring-lake-15276" => array("type"=>"town", "pop"=>8133, "name_ru"=>"Спринг Лейк", "name_en"=>"Spring Lake", "lang"=>"en", "lat"=>35.177602, "lon"=>-78.982411),
"/US/US-NC/spruce-pine-15277" => array("type"=>"town", "pop"=>1981, "name_en"=>"Spruce Pine", "lang"=>"en", "lat"=>35.916208, "lon"=>-82.069943),
"/US/US-NC/st.-helena-15227" => array("type"=>"village", "pop"=>512, "name_ru"=>"Сент-Хэлена", "name_en"=>"St. Helena", "lang"=>"en", "lat"=>34.513757, "lon"=>-77.915384),
"/US/US-NC/st.-pauls-15228" => array("type"=>"town", "pop"=>2041, "name_en"=>"St. Pauls", "lang"=>"en", "lat"=>34.807166, "lon"=>-78.972995),
"/US/US-NC/st.-stephens-15229" => array("type"=>"cdp", "pop"=>8734, "name_en"=>"St. Stephens", "lang"=>"en", "lat"=>35.764542, "lon"=>-81.273403),
"/US/US-NC/staley-15278" => array("type"=>"town", "pop"=>366, "name_en"=>"Staley", "lang"=>"en", "lat"=>35.794620, "lon"=>-79.551368),
"/US/US-NC/stallings-15279" => array("type"=>"town", "pop"=>8970, "name_en"=>"Stallings", "lang"=>"en", "lat"=>35.084656, "lon"=>-80.689535),
"/US/US-NC/stanfield-15280" => array("type"=>"town", "pop"=>1112, "name_en"=>"Stanfield", "lang"=>"en", "lat"=>35.235268, "lon"=>-80.427752),
"/US/US-NC/stanley-15281" => array("type"=>"town", "pop"=>3186, "name_ru"=>"Стэнли", "name_en"=>"Stanley", "lang"=>"en", "lat"=>35.356497, "lon"=>-81.097541),
"/US/US-NC/stanleyville-15282" => array("type"=>"cdp", "pop"=>4779, "name_en"=>"Stanleyville", "lang"=>"en", "lat"=>36.198800, "lon"=>-80.266935),
"/US/US-NC/stantonsburg-15283" => array("type"=>"town", "pop"=>707, "name_en"=>"Stantonsburg", "lang"=>"en", "lat"=>35.604484, "lon"=>-77.820182),
"/US/US-NC/star-15284" => array("type"=>"town", "pop"=>792, "name_en"=>"Star", "lang"=>"en", "lat"=>35.400460, "lon"=>-79.784223),
"/US/US-NC/statesville-15285" => array("type"=>"city", "pop"=>26414, "name_en"=>"Statesville", "lang"=>"en", "lat"=>35.786717, "lon"=>-80.886326),
"/US/US-NC/stedman-15286" => array("type"=>"town", "pop"=>676, "name_en"=>"Stedman", "lang"=>"en", "lat"=>35.010477, "lon"=>-78.698278),
"/US/US-NC/stem-15287" => array("type"=>"town", "pop"=>278, "name_en"=>"Stem", "lang"=>"en", "lat"=>36.199814, "lon"=>-78.723233),
"/US/US-NC/stokesdale-15288" => array("type"=>"town", "pop"=>3783, "name_en"=>"Stokesdale", "lang"=>"en", "lat"=>36.235678, "lon"=>-79.984177),
"/US/US-NC/stoneville-15289" => array("type"=>"town", "pop"=>974, "name_en"=>"Stoneville", "lang"=>"en", "lat"=>36.465698, "lon"=>-79.906580),
"/US/US-NC/stonewall-15290" => array("type"=>"town", "pop"=>265, "name_ru"=>"Стоунуолл", "name_en"=>"Stonewall", "lang"=>"en", "lat"=>35.137487, "lon"=>-76.741523),
"/US/US-NC/stony-point-15291" => array("type"=>"cdp", "pop"=>1286, "name_ru"=>"Стоуни Пойнт", "name_en"=>"Stony Point", "lang"=>"en", "lat"=>35.866712, "lon"=>-81.049100),
"/US/US-NC/stovall-15292" => array("type"=>"town", "pop"=>393, "name_en"=>"Stovall", "lang"=>"en", "lat"=>36.447709, "lon"=>-78.569449),
"/US/US-NC/sugar-mountain-15293" => array("type"=>"village", "pop"=>212, "name_en"=>"Sugar Mountain", "lang"=>"en", "lat"=>36.124225, "lon"=>-81.863315),
"/US/US-NC/summerfield-15294" => array("type"=>"cdp", "pop"=>7693, "name_ru"=>"Саммерфилд", "name_en"=>"Summerfield", "lang"=>"en", "lat"=>36.202555, "lon"=>-79.894244),
"/US/US-NC/sunset-beach-15295" => array("type"=>"town", "pop"=>2602, "name_en"=>"Sunset Beach", "lang"=>"en", "lat"=>33.874691, "lon"=>-78.509975),
"/US/US-NC/surf-city-15296" => array("type"=>"town", "pop"=>2057, "name_en"=>"Surf City", "lang"=>"en", "lat"=>34.423912, "lon"=>-77.558444),
"/US/US-NC/swannanoa-15297" => array("type"=>"cdp", "pop"=>3538, "name_en"=>"Swannanoa", "lang"=>"en", "lat"=>35.601640, "lon"=>-82.392084),
"/US/US-NC/swansboro-15298" => array("type"=>"town", "pop"=>1962, "name_en"=>"Swansboro", "lang"=>"en", "lat"=>34.690513, "lon"=>-77.128884),
"/US/US-NC/swepsonville-15299" => array("type"=>"cdp", "pop"=>1053, "name_en"=>"Swepsonville", "lang"=>"en", "lat"=>36.028692, "lon"=>-79.359771),
"/US/US-NC/sylva-15300" => array("type"=>"town", "pop"=>2437, "name_en"=>"Sylva", "lang"=>"en", "lat"=>35.374360, "lon"=>-83.222775),
"/US/US-NC/tabor-city-15301" => array("type"=>"town", "pop"=>2740, "name_en"=>"Tabor City", "lang"=>"en", "lat"=>34.149436, "lon"=>-78.874213),
"/US/US-NC/tar-heel-15303" => array("type"=>"town", "pop"=>70, "name_en"=>"Tar Heel", "lang"=>"en", "lat"=>34.731519, "lon"=>-78.792437),
"/US/US-NC/tarboro-15302" => array("type"=>"town", "pop"=>10257, "name_en"=>"Tarboro", "lang"=>"en", "lat"=>35.907719, "lon"=>-77.556588),
"/US/US-NC/taylorsville-15304" => array("type"=>"town", "pop"=>1854, "name_ru"=>"Тейлорсвилл", "name_en"=>"Taylorsville", "lang"=>"en", "lat"=>35.921136, "lon"=>-81.179257),
"/US/US-NC/taylortown-15305" => array("type"=>"town", "pop"=>883, "name_en"=>"Taylortown", "lang"=>"en", "lat"=>35.215900, "lon"=>-79.491954),
"/US/US-NC/teachey-15306" => array("type"=>"town", "pop"=>261, "name_en"=>"Teachey", "lang"=>"en", "lat"=>34.767381, "lon"=>-78.008768),
"/US/US-NC/thomasville-15307" => array("type"=>"city", "pop"=>26526, "name_ru"=>"Томасвилл", "name_en"=>"Thomasville", "lang"=>"en", "lat"=>35.888949, "lon"=>-80.076577),
"/US/US-NC/toast-15308" => array("type"=>"cdp", "pop"=>2125, "name_en"=>"Toast", "lang"=>"en", "lat"=>36.496870, "lon"=>-80.632925),
"/US/US-NC/topsail-beach-15309" => array("type"=>"town", "pop"=>581, "name_en"=>"Topsail Beach", "lang"=>"en", "lat"=>34.375243, "lon"=>-77.633981),
"/US/US-NC/trent-woods-15311" => array("type"=>"town", "pop"=>3986, "name_en"=>"Trent Woods", "lang"=>"en", "lat"=>35.077913, "lon"=>-77.096271),
"/US/US-NC/trenton-15310" => array("type"=>"town", "pop"=>246, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>35.062954, "lon"=>-77.354950),
"/US/US-NC/trinity-15312" => array("type"=>"cdp", "pop"=>6977, "name_ru"=>"Тринити", "name_en"=>"Trinity", "lang"=>"en", "lat"=>35.887856, "lon"=>-80.014531),
"/US/US-NC/troutman-15313" => array("type"=>"town", "pop"=>1815, "name_en"=>"Troutman", "lang"=>"en", "lat"=>35.704684, "lon"=>-80.890548),
"/US/US-NC/troy-15314" => array("type"=>"town", "pop"=>3415, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>35.364059, "lon"=>-79.893160),
"/US/US-NC/tryon-15315" => array("type"=>"town", "pop"=>1708, "name_en"=>"Tryon", "lang"=>"en", "lat"=>35.208888, "lon"=>-82.238710),
"/US/US-NC/turkey-15316" => array("type"=>"town", "pop"=>272, "name_en"=>"Turkey", "lang"=>"en", "lat"=>34.993187, "lon"=>-78.185367),
"/US/US-NC/valdese-15317" => array("type"=>"town", "pop"=>4557, "name_en"=>"Valdese", "lang"=>"en", "lat"=>35.746532, "lon"=>-81.557684),
"/US/US-NC/valley-hill-15318" => array("type"=>"cdp", "pop"=>1802, "name_en"=>"Valley Hill", "lang"=>"en", "lat"=>35.295966, "lon"=>-82.490601),
"/US/US-NC/vanceboro-15319" => array("type"=>"town", "pop"=>849, "name_en"=>"Vanceboro", "lang"=>"en", "lat"=>35.304375, "lon"=>-77.157588),
"/US/US-NC/vandemere-15320" => array("type"=>"town", "pop"=>268, "name_en"=>"Vandemere", "lang"=>"en", "lat"=>35.189083, "lon"=>-76.664895),
"/US/US-NC/vander-15321" => array("type"=>"cdp", "pop"=>1179, "name_en"=>"Vander", "lang"=>"en", "lat"=>35.036784, "lon"=>-78.786969),
"/US/US-NC/varnamtown-15322" => array("type"=>"town", "pop"=>593, "name_en"=>"Varnamtown", "lang"=>"en", "lat"=>33.941877, "lon"=>-78.239722),
"/US/US-NC/vass-15323" => array("type"=>"town", "pop"=>800, "name_en"=>"Vass", "lang"=>"en", "lat"=>35.255760, "lon"=>-79.282240),
"/US/US-NC/waco-15324" => array("type"=>"town", "pop"=>327, "name_ru"=>"Вако", "name_en"=>"Waco", "lang"=>"en", "lat"=>35.361894, "lon"=>-81.429304),
"/US/US-NC/wade-15325" => array("type"=>"town", "pop"=>535, "name_en"=>"Wade", "lang"=>"en", "lat"=>35.164279, "lon"=>-78.735350),
"/US/US-NC/wadesboro-15326" => array("type"=>"town", "pop"=>5066, "name_en"=>"Wadesboro", "lang"=>"en", "lat"=>34.966580, "lon"=>-80.068493),
"/US/US-NC/wagram-15327" => array("type"=>"town", "pop"=>775, "name_en"=>"Wagram", "lang"=>"en", "lat"=>34.889945, "lon"=>-79.365660),
"/US/US-NC/wake-forest-15328" => array("type"=>"town", "pop"=>27068, "name_en"=>"Wake Forest", "lang"=>"en", "lat"=>35.970567, "lon"=>-78.520881),
"/US/US-NC/walkertown-15329" => array("type"=>"town", "pop"=>4547, "name_en"=>"Walkertown", "lang"=>"en", "lat"=>36.171483, "lon"=>-80.152958),
"/US/US-NC/wallace-15330" => array("type"=>"town", "pop"=>3591, "name_ru"=>"Уоллэс", "name_en"=>"Wallace", "lang"=>"en", "lat"=>34.735008, "lon"=>-77.997879),
"/US/US-NC/walnut-cove-15331" => array("type"=>"town", "pop"=>1596, "name_en"=>"Walnut Cove", "lang"=>"en", "lat"=>36.294816, "lon"=>-80.139344),
"/US/US-NC/walnut-creek-15332" => array("type"=>"village", "pop"=>855, "name_ru"=>"Уолнат Крик", "name_en"=>"Walnut Creek", "lang"=>"en", "lat"=>35.308556, "lon"=>-77.868885),
"/US/US-NC/walstonburg-15333" => array("type"=>"town", "pop"=>227, "name_en"=>"Walstonburg", "lang"=>"en", "lat"=>35.594676, "lon"=>-77.699159),
"/US/US-NC/wanchese-15334" => array("type"=>"cdp", "pop"=>1380, "name_en"=>"Wanchese", "lang"=>"en", "lat"=>35.837470, "lon"=>-75.640249),
"/US/US-NC/warrenton-15335" => array("type"=>"town", "pop"=>840, "name_ru"=>"Уоррентон", "name_en"=>"Warrenton", "lang"=>"en", "lat"=>36.400113, "lon"=>-78.157096),
"/US/US-NC/warsaw-15336" => array("type"=>"town", "pop"=>3173, "name_ru"=>"Варшава", "name_en"=>"Warsaw", "lang"=>"en", "lat"=>34.998919, "lon"=>-78.087955),
"/US/US-NC/washington-15337" => array("type"=>"city", "pop"=>10112, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>35.556277, "lon"=>-77.050504),
"/US/US-NC/washington-park-15338" => array("type"=>"town", "pop"=>453, "name_ru"=>"Вашингтон Парк", "name_en"=>"Washington Park", "lang"=>"en", "lat"=>35.532465, "lon"=>-77.031218),
"/US/US-NC/watha-15339" => array("type"=>"town", "pop"=>230, "name_en"=>"Watha", "lang"=>"en", "lat"=>34.642293, "lon"=>-77.962853),
"/US/US-NC/waxhaw-15340" => array("type"=>"town", "pop"=>3672, "name_en"=>"Waxhaw", "lang"=>"en", "lat"=>34.924334, "lon"=>-80.742145),
"/US/US-NC/waynesville-15341" => array("type"=>"town", "pop"=>9877, "name_ru"=>"Уэйнсвилл", "name_en"=>"Waynesville", "lang"=>"en", "lat"=>35.484525, "lon"=>-82.997517),
"/US/US-NC/weaverville-15342" => array("type"=>"town", "pop"=>3084, "name_ru"=>"Уивервилл", "name_en"=>"Weaverville", "lang"=>"en", "lat"=>35.697018, "lon"=>-82.557867),
"/US/US-NC/webster-15343" => array("type"=>"town", "pop"=>451, "name_ru"=>"Вебстер", "name_en"=>"Webster", "lang"=>"en", "lat"=>35.352886, "lon"=>-83.214963),
"/US/US-NC/weddington-15344" => array("type"=>"town", "pop"=>9882, "name_en"=>"Weddington", "lang"=>"en", "lat"=>35.031588, "lon"=>-80.738072),
"/US/US-NC/welcome-15345" => array("type"=>"cdp", "pop"=>3377, "name_ru"=>"Вэлкам", "name_en"=>"Welcome", "lang"=>"en", "lat"=>35.910849, "lon"=>-80.254200),
"/US/US-NC/weldon-15346" => array("type"=>"town", "pop"=>1669, "name_ru"=>"Уэлдон", "name_en"=>"Weldon", "lang"=>"en", "lat"=>36.424246, "lon"=>-77.605613),
"/US/US-NC/wendell-15347" => array("type"=>"town", "pop"=>5300, "name_ru"=>"Уэнделл", "name_en"=>"Wendell", "lang"=>"en", "lat"=>35.781971, "lon"=>-78.368474),
"/US/US-NC/west-canton-15348" => array("type"=>"cdp", "pop"=>1119, "name_en"=>"West Canton", "lang"=>"en", "lat"=>35.538554, "lon"=>-82.867222),
"/US/US-NC/west-jefferson-15349" => array("type"=>"town", "pop"=>1121, "name_ru"=>"Вэст Джефферсон", "name_en"=>"West Jefferson", "lang"=>"en", "lat"=>36.397064, "lon"=>-81.491056),
"/US/US-NC/west-marion-15350" => array("type"=>"cdp", "pop"=>1291, "name_en"=>"West Marion", "lang"=>"en", "lat"=>35.653445, "lon"=>-82.020140),
"/US/US-NC/west-smithfield-15352" => array("type"=>"cdp", "pop"=>2411, "name_en"=>"West Smithfield", "lang"=>"en", "lat"=>35.530582, "lon"=>-78.363806),
"/US/US-NC/westport-15351" => array("type"=>"cdp", "pop"=>1280, "name_ru"=>"Уэстпорт", "name_en"=>"Westport", "lang"=>"en", "lat"=>35.507726, "lon"=>-80.977951),
"/US/US-NC/whispering-pines-15353" => array("type"=>"village", "pop"=>2149, "name_en"=>"Whispering Pines", "lang"=>"en", "lat"=>35.254882, "lon"=>-79.379490),
"/US/US-NC/whitakers-15354" => array("type"=>"town", "pop"=>773, "name_en"=>"Whitakers", "lang"=>"en", "lat"=>36.105945, "lon"=>-77.712692),
"/US/US-NC/white-lake-15355" => array("type"=>"town", "pop"=>516, "name_en"=>"White Lake", "lang"=>"en", "lat"=>34.638817, "lon"=>-78.500350),
"/US/US-NC/white-plains-15356" => array("type"=>"cdp", "pop"=>1027, "name_ru"=>"Уайт Плейнс", "name_en"=>"White Plains", "lang"=>"en", "lat"=>36.449711, "lon"=>-80.638075),
"/US/US-NC/whiteville-15357" => array("type"=>"city", "pop"=>5233, "name_en"=>"Whiteville", "lang"=>"en", "lat"=>34.329927, "lon"=>-78.704795),
"/US/US-NC/wilkesboro-15358" => array("type"=>"town", "pop"=>3140, "name_en"=>"Wilkesboro", "lang"=>"en", "lat"=>36.140969, "lon"=>-81.164837),
"/US/US-NC/williamston-15359" => array("type"=>"town", "pop"=>5382, "name_ru"=>"Уилльямстон", "name_en"=>"Williamston", "lang"=>"en", "lat"=>35.851357, "lon"=>-77.058562),
"/US/US-NC/wilmington-15360" => array("type"=>"city", "pop"=>100192, "name_ru"=>"Уилмингтон", "name_en"=>"Wilmington", "lang"=>"en", "lat"=>34.212450, "lon"=>-77.911299),
"/US/US-NC/wilson-15361" => array("type"=>"city", "pop"=>48433, "name_ru"=>"Уилсон", "name_en"=>"Wilson", "lang"=>"en", "lat"=>35.731769, "lon"=>-77.924457),
"/US/US-NC/windemere-15362" => array("type"=>"cdp", "pop"=>4604, "name_en"=>"Windemere", "lang"=>"en", "lat"=>34.243750, "lon"=>-77.845026),
"/US/US-NC/windsor-15363" => array("type"=>"town", "pop"=>2987, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>36.000682, "lon"=>-76.943879),
"/US/US-NC/winfall-15364" => array("type"=>"town", "pop"=>589, "name_en"=>"Winfall", "lang"=>"en", "lat"=>36.210623, "lon"=>-76.455220),
"/US/US-NC/wingate-15365" => array("type"=>"town", "pop"=>3879, "name_ru"=>"Уингейт", "name_en"=>"Wingate", "lang"=>"en", "lat"=>34.983312, "lon"=>-80.448455),
"/US/US-NC/winston-salem-15366" => array("type"=>"city", "pop"=>217600, "name_en"=>"Winston-Salem", "lang"=>"en", "lat"=>36.102100, "lon"=>-80.262910),
"/US/US-NC/winterville-15367" => array("type"=>"town", "pop"=>4813, "name_ru"=>"Уинтервилл", "name_en"=>"Winterville", "lang"=>"en", "lat"=>35.528737, "lon"=>-77.397670),
"/US/US-NC/winton-15368" => array("type"=>"town", "pop"=>896, "name_ru"=>"Уинтон", "name_en"=>"Winton", "lang"=>"en", "lat"=>36.389869, "lon"=>-76.936099),
"/US/US-NC/woodfin-15369" => array("type"=>"town", "pop"=>5731, "name_en"=>"Woodfin", "lang"=>"en", "lat"=>35.636550, "lon"=>-82.585515),
"/US/US-NC/woodland-15370" => array("type"=>"town", "pop"=>768, "name_ru"=>"Вудленд", "name_en"=>"Woodland", "lang"=>"en", "lat"=>36.332152, "lon"=>-77.210928),
"/US/US-NC/wrightsboro-15371" => array("type"=>"cdp", "pop"=>4752, "name_en"=>"Wrightsboro", "lang"=>"en", "lat"=>34.282467, "lon"=>-77.924189),
"/US/US-NC/wrightsville-beach-15372" => array("type"=>"town", "pop"=>2648, "name_en"=>"Wrightsville Beach", "lang"=>"en", "lat"=>34.211067, "lon"=>-77.795725),
"/US/US-NC/yadkinville-15373" => array("type"=>"town", "pop"=>2863, "name_en"=>"Yadkinville", "lang"=>"en", "lat"=>36.131361, "lon"=>-80.659581),
"/US/US-NC/yanceyville-15374" => array("type"=>"town", "pop"=>2186, "name_en"=>"Yanceyville", "lang"=>"en", "lat"=>36.407603, "lon"=>-79.343110),
"/US/US-NC/yaupon-beach-15375" => array("type"=>"town", "pop"=>734, "name_en"=>"Yaupon Beach", "lang"=>"en", "lat"=>33.907521, "lon"=>-78.076377),
"/US/US-NC/youngsville-15376" => array("type"=>"town", "pop"=>775, "name_ru"=>"Янгсвилл", "name_en"=>"Youngsville", "lang"=>"en", "lat"=>36.025864, "lon"=>-78.477078),
"/US/US-NC/zebulon-15377" => array("type"=>"town", "pop"=>4732, "name_ru"=>"Зебулон", "name_en"=>"Zebulon", "lang"=>"en", "lat"=>35.824440, "lon"=>-78.318488),
"/US/US-ND" => array("type"=>"state", "pop"=>641481, "name_ru"=>"Северная Дакота", "name_en"=>"North Dakota", "lang"=>"en"),
"/US/US-ND/abercrombie-15378" => array("type"=>"city", "pop"=>271, "name_en"=>"Abercrombie", "lang"=>"en", "lat"=>46.447223, "lon"=>-96.726126),
"/US/US-ND/adams-15379" => array("type"=>"city", "pop"=>175, "name_ru"=>"Адамс", "name_en"=>"Adams", "lang"=>"en", "lat"=>48.419694, "lon"=>-98.073753),
"/US/US-ND/alamo-15380" => array("type"=>"city", "pop"=>49, "name_ru"=>"Аламо", "name_en"=>"Alamo", "lang"=>"en", "lat"=>48.582025, "lon"=>-103.467037),
"/US/US-ND/alexander-15381" => array("type"=>"city", "pop"=>215, "name_ru"=>"Александр", "name_en"=>"Alexander", "lang"=>"en", "lat"=>47.841038, "lon"=>-103.642611),
"/US/US-ND/alice-15382" => array("type"=>"city", "pop"=>50, "name_en"=>"Alice", "lang"=>"en", "lat"=>46.760452, "lon"=>-97.556214),
"/US/US-ND/almont-15383" => array("type"=>"city", "pop"=>84, "name_ru"=>"Алмонт", "name_en"=>"Almont", "lang"=>"en", "lat"=>46.727815, "lon"=>-101.502758),
"/US/US-ND/alsen-15384" => array("type"=>"city", "pop"=>54, "name_en"=>"Alsen", "lang"=>"en", "lat"=>48.629511, "lon"=>-98.705407),
"/US/US-ND/ambrose-15385" => array("type"=>"city", "pop"=>20, "name_ru"=>"Амброуз", "name_en"=>"Ambrose", "lang"=>"en", "lat"=>48.955170, "lon"=>-103.480765),
"/US/US-ND/amenia-15386" => array("type"=>"city", "pop"=>89, "name_en"=>"Amenia", "lang"=>"en", "lat"=>47.008233, "lon"=>-97.223907),
"/US/US-ND/amidon-15387" => array("type"=>"city", "pop"=>23, "name_en"=>"Amidon", "lang"=>"en", "lat"=>46.482099, "lon"=>-103.318779),
"/US/US-ND/anamoose-15388" => array("type"=>"city", "pop"=>246, "name_en"=>"Anamoose", "lang"=>"en", "lat"=>47.883153, "lon"=>-100.241468),
"/US/US-ND/aneta-15389" => array("type"=>"city", "pop"=>246, "name_en"=>"Aneta", "lang"=>"en", "lat"=>47.678891, "lon"=>-97.988831),
"/US/US-ND/antler-15390" => array("type"=>"city", "pop"=>42, "name_en"=>"Antler", "lang"=>"en", "lat"=>48.971296, "lon"=>-101.283023),
"/US/US-ND/ardoch-15391" => array("type"=>"city", "pop"=>53, "name_en"=>"Ardoch", "lang"=>"en", "lat"=>48.205857, "lon"=>-97.340927),
"/US/US-ND/argusville-15392" => array("type"=>"city", "pop"=>134, "name_en"=>"Argusville", "lang"=>"en", "lat"=>47.050605, "lon"=>-96.940981),
"/US/US-ND/arnegard-15393" => array("type"=>"city", "pop"=>104, "name_en"=>"Arnegard", "lang"=>"en", "lat"=>47.808127, "lon"=>-103.438150),
"/US/US-ND/arthur-15394" => array("type"=>"city", "pop"=>378, "name_ru"=>"Артур", "name_en"=>"Arthur", "lang"=>"en", "lat"=>47.103454, "lon"=>-97.218255),
"/US/US-ND/ashley-15395" => array("type"=>"city", "pop"=>687, "name_ru"=>"Эшли", "name_en"=>"Ashley", "lang"=>"en", "lat"=>46.034920, "lon"=>-99.373343),
"/US/US-ND/ayr-15396" => array("type"=>"city", "pop"=>22, "name_en"=>"Ayr", "lang"=>"en", "lat"=>47.040681, "lon"=>-97.491220),
"/US/US-ND/balfour-15397" => array("type"=>"city", "pop"=>17, "name_en"=>"Balfour", "lang"=>"en", "lat"=>47.952344, "lon"=>-100.534659),
"/US/US-ND/balta-15398" => array("type"=>"city", "pop"=>62, "name_en"=>"Balta", "lang"=>"en", "lat"=>48.166160, "lon"=>-100.037208),
"/US/US-ND/bantry-15399" => array("type"=>"city", "pop"=>17, "name_en"=>"Bantry", "lang"=>"en", "lat"=>48.497823, "lon"=>-100.609234),
"/US/US-ND/barney-15400" => array("type"=>"city", "pop"=>61, "name_en"=>"Barney", "lang"=>"en", "lat"=>46.266509, "lon"=>-96.999192),
"/US/US-ND/barton-15401" => array("type"=>"city", "pop"=>24, "name_ru"=>"Бартон", "name_en"=>"Barton", "lang"=>"en", "lat"=>48.505937, "lon"=>-100.178091),
"/US/US-ND/bathgate-15402" => array("type"=>"city", "pop"=>56, "name_en"=>"Bathgate", "lang"=>"en", "lat"=>48.880554, "lon"=>-97.472860),
"/US/US-ND/beach-15403" => array("type"=>"city", "pop"=>939, "name_en"=>"Beach", "lang"=>"en", "lat"=>46.913893, "lon"=>-104.007726),
"/US/US-ND/belcourt-15404" => array("type"=>"cdp", "pop"=>2458, "name_en"=>"Belcourt", "lang"=>"en", "lat"=>48.841599, "lon"=>-99.746315),
"/US/US-ND/belfield-15405" => array("type"=>"city", "pop"=>811, "name_en"=>"Belfield", "lang"=>"en", "lat"=>46.886946, "lon"=>-103.195416),
"/US/US-ND/benedict-15406" => array("type"=>"city", "pop"=>47, "name_ru"=>"Бенедикт", "name_en"=>"Benedict", "lang"=>"en", "lat"=>47.830541, "lon"=>-101.083728),
"/US/US-ND/bergen-15407" => array("type"=>"city", "pop"=>10, "name_en"=>"Bergen", "lang"=>"en", "lat"=>48.004669, "lon"=>-100.718272),
"/US/US-ND/berlin-15408" => array("type"=>"city", "pop"=>30, "name_ru"=>"Берлин", "name_en"=>"Berlin", "lang"=>"en", "lat"=>46.378439, "lon"=>-98.487486),
"/US/US-ND/berthold-15409" => array("type"=>"city", "pop"=>439, "name_en"=>"Berthold", "lang"=>"en", "lat"=>48.315019, "lon"=>-101.735334),
"/US/US-ND/beulah-15410" => array("type"=>"city", "pop"=>2863, "name_ru"=>"Бойла", "name_en"=>"Beulah", "lang"=>"en", "lat"=>47.266887, "lon"=>-101.773358),
"/US/US-ND/binford-15411" => array("type"=>"city", "pop"=>169, "name_en"=>"Binford", "lang"=>"en", "lat"=>47.559812, "lon"=>-98.345286),
"/US/US-ND/bisbee-15412" => array("type"=>"city", "pop"=>130, "name_ru"=>"Бисби", "name_en"=>"Bisbee", "lang"=>"en", "lat"=>48.626606, "lon"=>-99.378270),
"/US/US-ND/bismarck-15413" => array("type"=>"city", "pop"=>60389, "name_ru"=>"Бисмарк", "name_en"=>"Bismarck", "lang"=>"en", "lat"=>46.805467, "lon"=>-100.767298),
"/US/US-ND/bottineau-15414" => array("type"=>"city", "pop"=>2052, "name_en"=>"Bottineau", "lang"=>"en", "lat"=>48.825448, "lon"=>-100.443288),
"/US/US-ND/bowbells-15415" => array("type"=>"city", "pop"=>322, "name_en"=>"Bowbells", "lang"=>"en", "lat"=>48.803013, "lon"=>-102.246775),
"/US/US-ND/bowdon-15416" => array("type"=>"city", "pop"=>112, "name_ru"=>"Баудон", "name_en"=>"Bowdon", "lang"=>"en", "lat"=>47.468929, "lon"=>-99.708084),
"/US/US-ND/bowman-15417" => array("type"=>"city", "pop"=>1501, "name_ru"=>"Баумэн", "name_en"=>"Bowman", "lang"=>"en", "lat"=>46.183808, "lon"=>-103.399782),
"/US/US-ND/braddock-15418" => array("type"=>"city", "pop"=>34, "name_en"=>"Braddock", "lang"=>"en", "lat"=>46.563451, "lon"=>-100.089129),
"/US/US-ND/briarwood-15419" => array("type"=>"city", "pop"=>82, "name_ru"=>"Брайрвуд", "name_en"=>"Briarwood", "lang"=>"en", "lat"=>46.786228, "lon"=>-96.793762),
"/US/US-ND/brinsmade-15420" => array("type"=>"city", "pop"=>29, "name_en"=>"Brinsmade", "lang"=>"en", "lat"=>48.183311, "lon"=>-99.324418),
"/US/US-ND/brocket-15421" => array("type"=>"city", "pop"=>58, "name_en"=>"Brocket", "lang"=>"en", "lat"=>48.211188, "lon"=>-98.356241),
"/US/US-ND/buchanan-15422" => array("type"=>"city", "pop"=>70, "name_ru"=>"Бачанан", "name_en"=>"Buchanan", "lang"=>"en", "lat"=>47.062962, "lon"=>-98.829198),
"/US/US-ND/bucyrus-15423" => array("type"=>"city", "pop"=>22, "name_en"=>"Bucyrus", "lang"=>"en", "lat"=>46.063577, "lon"=>-102.788308),
"/US/US-ND/buffalo-15424" => array("type"=>"city", "pop"=>190, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>46.921427, "lon"=>-97.550291),
"/US/US-ND/burlington-15425" => array("type"=>"city", "pop"=>1000, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>48.275555, "lon"=>-101.425654),
"/US/US-ND/butte-15426" => array("type"=>"city", "pop"=>82, "name_ru"=>"Бьютт", "name_en"=>"Butte", "lang"=>"en", "lat"=>47.837410, "lon"=>-100.665924),
"/US/US-ND/buxton-15427" => array("type"=>"city", "pop"=>325, "name_en"=>"Buxton", "lang"=>"en", "lat"=>47.602166, "lon"=>-97.099188),
"/US/US-ND/calio-15428" => array("type"=>"city", "pop"=>19, "name_en"=>"Calio", "lang"=>"en", "lat"=>48.623603, "lon"=>-98.937883),
"/US/US-ND/calvin-15429" => array("type"=>"city", "pop"=>21, "name_ru"=>"Калвин", "name_en"=>"Calvin", "lang"=>"en", "lat"=>48.851156, "lon"=>-98.936990),
"/US/US-ND/cando-15430" => array("type"=>"city", "pop"=>1019, "name_en"=>"Cando", "lang"=>"en", "lat"=>48.487640, "lon"=>-99.202632),
"/US/US-ND/cannon-ball-15431" => array("type"=>"cdp", "pop"=>51, "name_en"=>"Cannon Ball", "lang"=>"en", "lat"=>46.312470, "lon"=>-100.632061),
"/US/US-ND/canton-city-15432" => array("type"=>"city", "pop"=>37, "name_en"=>"Canton City", "lang"=>"en", "lat"=>48.687698, "lon"=>-97.666944),
"/US/US-ND/carpio-15433" => array("type"=>"city", "pop"=>139, "name_en"=>"Carpio", "lang"=>"en", "lat"=>48.442372, "lon"=>-101.715382),
"/US/US-ND/carrington-15434" => array("type"=>"city", "pop"=>2071, "name_en"=>"Carrington", "lang"=>"en", "lat"=>47.450083, "lon"=>-99.123491),
"/US/US-ND/carson-15435" => array("type"=>"city", "pop"=>257, "name_ru"=>"Карсон", "name_en"=>"Carson", "lang"=>"en", "lat"=>46.421896, "lon"=>-101.569768),
"/US/US-ND/casselton-15436" => array("type"=>"city", "pop"=>2004, "name_en"=>"Casselton", "lang"=>"en", "lat"=>46.900649, "lon"=>-97.211870),
"/US/US-ND/cathay-15437" => array("type"=>"city", "pop"=>46, "name_en"=>"Cathay", "lang"=>"en", "lat"=>47.553932, "lon"=>-99.411007),
"/US/US-ND/cavalier-15438" => array("type"=>"city", "pop"=>1322, "name_en"=>"Cavalier", "lang"=>"en", "lat"=>48.795310, "lon"=>-97.624296),
"/US/US-ND/cayuga-15439" => array("type"=>"city", "pop"=>56, "name_ru"=>"Каюга", "name_en"=>"Cayuga", "lang"=>"en", "lat"=>46.075834, "lon"=>-97.383414),
"/US/US-ND/center-15440" => array("type"=>"city", "pop"=>554, "name_ru"=>"Сентер", "name_en"=>"Center", "lang"=>"en", "lat"=>47.114953, "lon"=>-101.296922),
"/US/US-ND/christine-15441" => array("type"=>"city", "pop"=>173, "name_en"=>"Christine", "lang"=>"en", "lat"=>46.575313, "lon"=>-96.806691),
"/US/US-ND/churchs-ferry-15442" => array("type"=>"city", "pop"=>71, "name_en"=>"Churchs Ferry", "lang"=>"en", "lat"=>48.268574, "lon"=>-99.194174),
"/US/US-ND/cleveland-15443" => array("type"=>"city", "pop"=>98, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>46.890874, "lon"=>-99.119957),
"/US/US-ND/clifford-15444" => array("type"=>"city", "pop"=>47, "name_ru"=>"Клиффорд", "name_en"=>"Clifford", "lang"=>"en", "lat"=>47.348268, "lon"=>-97.409289),
"/US/US-ND/cogswell-15445" => array("type"=>"city", "pop"=>146, "name_en"=>"Cogswell", "lang"=>"en", "lat"=>46.107110, "lon"=>-97.784032),
"/US/US-ND/coleharbor-15446" => array("type"=>"city", "pop"=>94, "name_en"=>"Coleharbor", "lang"=>"en", "lat"=>47.542395, "lon"=>-101.221171),
"/US/US-ND/colfax-15447" => array("type"=>"city", "pop"=>86, "name_ru"=>"Колфакс", "name_en"=>"Colfax", "lang"=>"en", "lat"=>46.470157, "lon"=>-96.873986),
"/US/US-ND/columbus-15448" => array("type"=>"city", "pop"=>119, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>48.904483, "lon"=>-102.781049),
"/US/US-ND/conway-15449" => array("type"=>"city", "pop"=>20, "name_ru"=>"Конвэй", "name_en"=>"Conway", "lang"=>"en", "lat"=>48.234004, "lon"=>-97.674392),
"/US/US-ND/cooperstown-15450" => array("type"=>"city", "pop"=>888, "name_en"=>"Cooperstown", "lang"=>"en", "lat"=>47.444117, "lon"=>-98.124094),
"/US/US-ND/courtenay-15451" => array("type"=>"city", "pop"=>47, "name_en"=>"Courtenay", "lang"=>"en", "lat"=>47.224499, "lon"=>-98.568733),
"/US/US-ND/crary-15452" => array("type"=>"city", "pop"=>137, "name_en"=>"Crary", "lang"=>"en", "lat"=>48.071471, "lon"=>-98.640036),
"/US/US-ND/crosby-15453" => array("type"=>"city", "pop"=>952, "name_ru"=>"Кросби", "name_en"=>"Crosby", "lang"=>"en", "lat"=>48.911401, "lon"=>-103.294733),
"/US/US-ND/crystal-15454" => array("type"=>"city", "pop"=>147, "name_ru"=>"Кристал", "name_en"=>"Crystal", "lang"=>"en", "lat"=>48.598451, "lon"=>-97.668365),
"/US/US-ND/davenport-15455" => array("type"=>"city", "pop"=>251, "name_ru"=>"Дэйвенпорт", "name_en"=>"Davenport", "lang"=>"en", "lat"=>46.713752, "lon"=>-97.068807),
"/US/US-ND/dawson-15456" => array("type"=>"city", "pop"=>62, "name_ru"=>"Доусон", "name_en"=>"Dawson", "lang"=>"en", "lat"=>46.868147, "lon"=>-99.752122),
"/US/US-ND/dazey-15457" => array("type"=>"city", "pop"=>82, "name_en"=>"Dazey", "lang"=>"en", "lat"=>47.188179, "lon"=>-98.200541),
"/US/US-ND/deering-15458" => array("type"=>"city", "pop"=>99, "name_ru"=>"Диаринг", "name_en"=>"Deering", "lang"=>"en", "lat"=>48.395253, "lon"=>-101.049411),
"/US/US-ND/des-lacs-15459" => array("type"=>"city", "pop"=>194, "name_en"=>"Des Lacs", "lang"=>"en", "lat"=>48.256852, "lon"=>-101.563157),
"/US/US-ND/devils-lake-15460" => array("type"=>"city", "pop"=>6708, "name_en"=>"Devils Lake", "lang"=>"en", "lat"=>48.112753, "lon"=>-98.874074),
"/US/US-ND/dickey-15461" => array("type"=>"city", "pop"=>48, "name_en"=>"Dickey", "lang"=>"en", "lat"=>46.536496, "lon"=>-98.467421),
"/US/US-ND/dickinson-15462" => array("type"=>"city", "pop"=>16035, "name_en"=>"Dickinson", "lang"=>"en", "lat"=>46.886113, "lon"=>-102.782719),
"/US/US-ND/dodge-15463" => array("type"=>"city", "pop"=>118, "name_en"=>"Dodge", "lang"=>"en", "lat"=>47.305310, "lon"=>-102.201955),
"/US/US-ND/donnybrook-15464" => array("type"=>"city", "pop"=>83, "name_en"=>"Donnybrook", "lang"=>"en", "lat"=>48.507881, "lon"=>-101.885642),
"/US/US-ND/douglas-15465" => array("type"=>"city", "pop"=>60, "name_ru"=>"Дуглас", "name_en"=>"Douglas", "lang"=>"en", "lat"=>47.858592, "lon"=>-101.500799),
"/US/US-ND/drake-15466" => array("type"=>"city", "pop"=>268, "name_en"=>"Drake", "lang"=>"en", "lat"=>47.921941, "lon"=>-100.375864),
"/US/US-ND/drayton-15467" => array("type"=>"city", "pop"=>782, "name_en"=>"Drayton", "lang"=>"en", "lat"=>48.564558, "lon"=>-97.179555),
"/US/US-ND/dunn-center-15468" => array("type"=>"city", "pop"=>111, "name_en"=>"Dunn Center", "lang"=>"en", "lat"=>47.353295, "lon"=>-102.623758),
"/US/US-ND/dunseith-15469" => array("type"=>"city", "pop"=>733, "name_en"=>"Dunseith", "lang"=>"en", "lat"=>48.812563, "lon"=>-100.061288),
"/US/US-ND/dwight-15470" => array("type"=>"city", "pop"=>69, "name_ru"=>"Двайт", "name_en"=>"Dwight", "lang"=>"en", "lat"=>46.302904, "lon"=>-96.738623),
"/US/US-ND/east-dunseith-15471" => array("type"=>"cdp", "pop"=>260, "name_en"=>"East Dunseith", "lang"=>"en", "lat"=>48.865516, "lon"=>-100.016369),
"/US/US-ND/edgeley-15472" => array("type"=>"city", "pop"=>526, "name_en"=>"Edgeley", "lang"=>"en", "lat"=>46.362554, "lon"=>-98.711775),
"/US/US-ND/edinburg-15473" => array("type"=>"city", "pop"=>220, "name_ru"=>"Эдинбург", "name_en"=>"Edinburg", "lang"=>"en", "lat"=>48.495567, "lon"=>-97.862170),
"/US/US-ND/edmore-15474" => array("type"=>"city", "pop"=>234, "name_ru"=>"Эдмор", "name_en"=>"Edmore", "lang"=>"en", "lat"=>48.412677, "lon"=>-98.453227),
"/US/US-ND/egeland-15475" => array("type"=>"city", "pop"=>37, "name_en"=>"Egeland", "lang"=>"en", "lat"=>48.628014, "lon"=>-99.097628),
"/US/US-ND/elgin-15476" => array("type"=>"city", "pop"=>543, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>46.401656, "lon"=>-101.846554),
"/US/US-ND/ellendale-15477" => array("type"=>"city", "pop"=>1456, "name_ru"=>"Эллендэйл", "name_en"=>"Ellendale", "lang"=>"en", "lat"=>46.003942, "lon"=>-98.524783),
"/US/US-ND/elliott-15478" => array("type"=>"city", "pop"=>40, "name_ru"=>"Эллиотт", "name_en"=>"Elliott", "lang"=>"en", "lat"=>46.402409, "lon"=>-97.813880),
"/US/US-ND/emerado-15479" => array("type"=>"city", "pop"=>477, "name_en"=>"Emerado", "lang"=>"en", "lat"=>47.921530, "lon"=>-97.363919),
"/US/US-ND/enderlin-15480" => array("type"=>"city", "pop"=>1019, "name_en"=>"Enderlin", "lang"=>"en", "lat"=>46.623041, "lon"=>-97.598852),
"/US/US-ND/epping-15481" => array("type"=>"city", "pop"=>76, "name_en"=>"Epping", "lang"=>"en", "lat"=>48.280740, "lon"=>-103.357674),
"/US/US-ND/esmond-15482" => array("type"=>"city", "pop"=>152, "name_en"=>"Esmond", "lang"=>"en", "lat"=>48.032857, "lon"=>-99.764826),
"/US/US-ND/fairdale-15483" => array("type"=>"city", "pop"=>44, "name_ru"=>"Фэрдейл", "name_en"=>"Fairdale", "lang"=>"en", "lat"=>48.489481, "lon"=>-98.231549),
"/US/US-ND/fairmount-15484" => array("type"=>"city", "pop"=>363, "name_ru"=>"Фэрмаунт", "name_en"=>"Fairmount", "lang"=>"en", "lat"=>46.054467, "lon"=>-96.602705),
"/US/US-ND/fargo-15485" => array("type"=>"city", "pop"=>93531, "name_ru"=>"Фарго", "name_en"=>"Fargo", "lang"=>"en", "lat"=>46.876633, "lon"=>-96.817112),
"/US/US-ND/fessenden-15486" => array("type"=>"city", "pop"=>500, "name_en"=>"Fessenden", "lang"=>"en", "lat"=>47.649666, "lon"=>-99.625742),
"/US/US-ND/fingal-15487" => array("type"=>"city", "pop"=>118, "name_en"=>"Fingal", "lang"=>"en", "lat"=>46.762051, "lon"=>-97.792429),
"/US/US-ND/finley-15488" => array("type"=>"city", "pop"=>402, "name_en"=>"Finley", "lang"=>"en", "lat"=>47.511417, "lon"=>-97.837071),
"/US/US-ND/flasher-15489" => array("type"=>"city", "pop"=>265, "name_en"=>"Flasher", "lang"=>"en", "lat"=>46.452083, "lon"=>-101.232019),
"/US/US-ND/flaxton-15490" => array("type"=>"city", "pop"=>58, "name_en"=>"Flaxton", "lang"=>"en", "lat"=>48.897279, "lon"=>-102.392677),
"/US/US-ND/forbes-15491" => array("type"=>"city", "pop"=>58, "name_en"=>"Forbes", "lang"=>"en", "lat"=>45.942153, "lon"=>-98.782750),
"/US/US-ND/fordville-15492" => array("type"=>"city", "pop"=>230, "name_en"=>"Fordville", "lang"=>"en", "lat"=>48.216745, "lon"=>-97.794480),
"/US/US-ND/forest-river-15493" => array("type"=>"city", "pop"=>133, "name_en"=>"Forest River", "lang"=>"en", "lat"=>48.217118, "lon"=>-97.470006),
"/US/US-ND/forman-15494" => array("type"=>"city", "pop"=>470, "name_en"=>"Forman", "lang"=>"en", "lat"=>46.105203, "lon"=>-97.636852),
"/US/US-ND/fort-ransom-15495" => array("type"=>"city", "pop"=>97, "name_en"=>"Fort Ransom", "lang"=>"en", "lat"=>46.524665, "lon"=>-97.930353),
"/US/US-ND/fort-totten-15496" => array("type"=>"cdp", "pop"=>1639, "name_en"=>"Fort Totten", "lang"=>"en", "lat"=>47.976136, "lon"=>-98.994825),
"/US/US-ND/fort-yates-15498" => array("type"=>"city", "pop"=>284, "name_en"=>"Fort Yates", "lang"=>"en", "lat"=>46.087016, "lon"=>-100.629763),
"/US/US-ND/fortuna-15497" => array("type"=>"city", "pop"=>27, "name_ru"=>"Фортуна", "name_en"=>"Fortuna", "lang"=>"en", "lat"=>48.911062, "lon"=>-103.776897),
"/US/US-ND/four-bears-village-15499" => array("type"=>"cdp", "pop"=>309, "name_en"=>"Four Bears Village", "lang"=>"en", "lat"=>47.989357, "lon"=>-102.597078),
"/US/US-ND/fredonia-15500" => array("type"=>"city", "pop"=>43, "name_ru"=>"Фредониа", "name_en"=>"Fredonia", "lang"=>"en", "lat"=>46.328594, "lon"=>-99.094874),
"/US/US-ND/frontier-15501" => array("type"=>"city", "pop"=>261, "name_en"=>"Frontier", "lang"=>"en", "lat"=>46.800378, "lon"=>-96.833009),
"/US/US-ND/fullerton-15502" => array("type"=>"city", "pop"=>75, "name_ru"=>"Фуллертон", "name_en"=>"Fullerton", "lang"=>"en", "lat"=>46.163639, "lon"=>-98.425828),
"/US/US-ND/gackle-15503" => array("type"=>"city", "pop"=>281, "name_en"=>"Gackle", "lang"=>"en", "lat"=>46.626362, "lon"=>-99.141422),
"/US/US-ND/galesburg-15504" => array("type"=>"city", "pop"=>148, "name_ru"=>"Гейлсбург", "name_en"=>"Galesburg", "lang"=>"en", "lat"=>47.270591, "lon"=>-97.408458),
"/US/US-ND/gardena-15505" => array("type"=>"city", "pop"=>34, "name_ru"=>"Гардена", "name_en"=>"Gardena", "lang"=>"en", "lat"=>48.700980, "lon"=>-100.498080),
"/US/US-ND/gardner-15506" => array("type"=>"city", "pop"=>85, "name_ru"=>"Гарднер", "name_en"=>"Gardner", "lang"=>"en", "lat"=>47.143398, "lon"=>-96.968833),
"/US/US-ND/garrison-15507" => array("type"=>"city", "pop"=>1173, "name_ru"=>"Гаррисон", "name_en"=>"Garrison", "lang"=>"en", "lat"=>47.653748, "lon"=>-101.423373),
"/US/US-ND/gascoyne-15508" => array("type"=>"city", "pop"=>21, "name_en"=>"Gascoyne", "lang"=>"en", "lat"=>46.117493, "lon"=>-103.079229),
"/US/US-ND/gilby-15509" => array("type"=>"city", "pop"=>226, "name_en"=>"Gilby", "lang"=>"en", "lat"=>48.083752, "lon"=>-97.468205),
"/US/US-ND/gladstone-15510" => array("type"=>"city", "pop"=>232, "name_ru"=>"Глэдстоун", "name_en"=>"Gladstone", "lang"=>"en", "lat"=>46.859996, "lon"=>-102.567413),
"/US/US-ND/glen-ullin-15513" => array("type"=>"city", "pop"=>803, "name_en"=>"Glen Ullin", "lang"=>"en", "lat"=>46.812375, "lon"=>-101.832099),
"/US/US-ND/glenburn-15511" => array("type"=>"city", "pop"=>314, "name_en"=>"Glenburn", "lang"=>"en", "lat"=>48.513136, "lon"=>-101.220495),
"/US/US-ND/glenfield-15512" => array("type"=>"city", "pop"=>127, "name_en"=>"Glenfield", "lang"=>"en", "lat"=>47.454932, "lon"=>-98.566321),
"/US/US-ND/golden-valley-15514" => array("type"=>"city", "pop"=>163, "name_ru"=>"Голден Валли", "name_en"=>"Golden Valley", "lang"=>"en", "lat"=>47.290429, "lon"=>-102.064717),
"/US/US-ND/golva-15515" => array("type"=>"city", "pop"=>90, "name_en"=>"Golva", "lang"=>"en", "lat"=>46.734604, "lon"=>-103.982577),
"/US/US-ND/goodrich-15516" => array("type"=>"city", "pop"=>116, "name_ru"=>"Гудрич", "name_en"=>"Goodrich", "lang"=>"en", "lat"=>47.476117, "lon"=>-100.124686),
"/US/US-ND/grace-city-15517" => array("type"=>"city", "pop"=>66, "name_en"=>"Grace City", "lang"=>"en", "lat"=>47.550461, "lon"=>-98.803979),
"/US/US-ND/grafton-15518" => array("type"=>"city", "pop"=>3978, "name_ru"=>"Графтон", "name_en"=>"Grafton", "lang"=>"en", "lat"=>48.415110, "lon"=>-97.403069),
"/US/US-ND/grand-forks-15519" => array("type"=>"city", "pop"=>51313, "name_en"=>"Grand Forks", "lang"=>"en", "lat"=>47.916900, "lon"=>-97.072006),
"/US/US-ND/grand-forks-afb-15520" => array("type"=>"cdp", "pop"=>9343, "name_en"=>"Grand Forks AFB", "lang"=>"en", "lat"=>47.962155, "lon"=>-97.383914),
"/US/US-ND/grandin-15521" => array("type"=>"city", "pop"=>166, "name_ru"=>"Грандин", "name_en"=>"Grandin", "lang"=>"en", "lat"=>47.236055, "lon"=>-97.002356),
"/US/US-ND/grano-15522" => array("type"=>"city", "pop"=>8, "name_en"=>"Grano", "lang"=>"en", "lat"=>48.616998, "lon"=>-101.586461),
"/US/US-ND/granville-15523" => array("type"=>"city", "pop"=>238, "name_ru"=>"Гранвилл", "name_en"=>"Granville", "lang"=>"en", "lat"=>48.267459, "lon"=>-100.842938),
"/US/US-ND/great-bend-15524" => array("type"=>"city", "pop"=>109, "name_ru"=>"Грейт Бенд", "name_en"=>"Great Bend", "lang"=>"en", "lat"=>46.154529, "lon"=>-96.801050),
"/US/US-ND/grenora-15525" => array("type"=>"city", "pop"=>194, "name_en"=>"Grenora", "lang"=>"en", "lat"=>48.619194, "lon"=>-103.936101),
"/US/US-ND/gwinner-15526" => array("type"=>"city", "pop"=>712, "name_en"=>"Gwinner", "lang"=>"en", "lat"=>46.224179, "lon"=>-97.658478),
"/US/US-ND/hague-15527" => array("type"=>"city", "pop"=>71, "name_en"=>"Hague", "lang"=>"en", "lat"=>46.029389, "lon"=>-99.998238),
"/US/US-ND/halliday-15528" => array("type"=>"city", "pop"=>206, "name_en"=>"Halliday", "lang"=>"en", "lat"=>47.352723, "lon"=>-102.337484),
"/US/US-ND/hamberg-15529" => array("type"=>"city", "pop"=>23, "name_en"=>"Hamberg", "lang"=>"en", "lat"=>47.762641, "lon"=>-99.514818),
"/US/US-ND/hamilton-15530" => array("type"=>"city", "pop"=>62, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>48.808142, "lon"=>-97.451638),
"/US/US-ND/hampden-15531" => array("type"=>"city", "pop"=>55, "name_ru"=>"Хэмпдэн", "name_en"=>"Hampden", "lang"=>"en", "lat"=>48.539547, "lon"=>-98.654492),
"/US/US-ND/hankinson-15532" => array("type"=>"city", "pop"=>981, "name_en"=>"Hankinson", "lang"=>"en", "lat"=>46.071506, "lon"=>-96.893756),
"/US/US-ND/hanks-15533" => array("type"=>"city", "pop"=>11, "name_en"=>"Hanks", "lang"=>"en", "lat"=>48.602536, "lon"=>-103.802025),
"/US/US-ND/hannaford-15534" => array("type"=>"city", "pop"=>151, "name_en"=>"Hannaford", "lang"=>"en", "lat"=>47.313865, "lon"=>-98.187852),
"/US/US-ND/hannah-15535" => array("type"=>"city", "pop"=>16, "name_en"=>"Hannah", "lang"=>"en", "lat"=>48.973115, "lon"=>-98.690240),
"/US/US-ND/hansboro-15536" => array("type"=>"city", "pop"=>6, "name_en"=>"Hansboro", "lang"=>"en", "lat"=>48.952241, "lon"=>-99.380829),
"/US/US-ND/harvey-15537" => array("type"=>"city", "pop"=>1619, "name_ru"=>"Харви", "name_en"=>"Harvey", "lang"=>"en", "lat"=>47.770045, "lon"=>-99.930718),
"/US/US-ND/harwood-15538" => array("type"=>"city", "pop"=>701, "name_ru"=>"Харвуд", "name_en"=>"Harwood", "lang"=>"en", "lat"=>46.980277, "lon"=>-96.879857),
"/US/US-ND/hatton-15539" => array("type"=>"city", "pop"=>640, "name_en"=>"Hatton", "lang"=>"en", "lat"=>47.636022, "lon"=>-97.457642),
"/US/US-ND/havana-15540" => array("type"=>"city", "pop"=>84, "name_ru"=>"Гавана", "name_en"=>"Havana", "lang"=>"en", "lat"=>45.950652, "lon"=>-97.617655),
"/US/US-ND/haynes-15541" => array("type"=>"city", "pop"=>16, "name_ru"=>"Хайнс", "name_en"=>"Haynes", "lang"=>"en", "lat"=>45.973719, "lon"=>-102.471555),
"/US/US-ND/hazelton-15542" => array("type"=>"city", "pop"=>187, "name_ru"=>"Хейзлтон", "name_en"=>"Hazelton", "lang"=>"en", "lat"=>46.484288, "lon"=>-100.278421),
"/US/US-ND/hazen-15543" => array("type"=>"city", "pop"=>2206, "name_ru"=>"Хазен", "name_en"=>"Hazen", "lang"=>"en", "lat"=>47.299545, "lon"=>-101.624781),
"/US/US-ND/hebron-15544" => array("type"=>"city", "pop"=>727, "name_ru"=>"Хеброн", "name_en"=>"Hebron", "lang"=>"en", "lat"=>46.902828, "lon"=>-102.044084),
"/US/US-ND/hettinger-15545" => array("type"=>"city", "pop"=>1138, "name_en"=>"Hettinger", "lang"=>"en", "lat"=>46.003627, "lon"=>-102.634396),
"/US/US-ND/hillsboro-15546" => array("type"=>"city", "pop"=>1469, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>47.402820, "lon"=>-97.061380),
"/US/US-ND/hoople-15547" => array("type"=>"city", "pop"=>252, "name_en"=>"Hoople", "lang"=>"en", "lat"=>48.535501, "lon"=>-97.637484),
"/US/US-ND/hope-15548" => array("type"=>"city", "pop"=>256, "name_ru"=>"Хоуп", "name_en"=>"Hope", "lang"=>"en", "lat"=>47.324820, "lon"=>-97.719629),
"/US/US-ND/horace-15549" => array("type"=>"city", "pop"=>1757, "name_ru"=>"Хорейс", "name_en"=>"Horace", "lang"=>"en", "lat"=>46.757430, "lon"=>-96.904152),
"/US/US-ND/hove-mobile-park-15550" => array("type"=>"city", "pop"=>0, "name_en"=>"Hove Mobile Park", "lang"=>"en", "lat"=>48.582492, "lon"=>-98.229094),
"/US/US-ND/hunter-15551" => array("type"=>"city", "pop"=>304, "name_ru"=>"Хантер", "name_en"=>"Hunter", "lang"=>"en", "lat"=>47.192107, "lon"=>-97.216270),
"/US/US-ND/hurdsfield-15552" => array("type"=>"city", "pop"=>76, "name_en"=>"Hurdsfield", "lang"=>"en", "lat"=>47.447824, "lon"=>-99.930015),
"/US/US-ND/inkster-15553" => array("type"=>"city", "pop"=>95, "name_ru"=>"Инкстер", "name_en"=>"Inkster", "lang"=>"en", "lat"=>48.152046, "lon"=>-97.644316),
"/US/US-ND/jamestown-15554" => array("type"=>"city", "pop"=>14630, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>46.910122, "lon"=>-98.698223),
"/US/US-ND/jud-15555" => array("type"=>"city", "pop"=>65, "name_en"=>"Jud", "lang"=>"en", "lat"=>46.525010, "lon"=>-98.897809),
"/US/US-ND/karlsruhe-15556" => array("type"=>"city", "pop"=>104, "name_en"=>"Karlsruhe", "lang"=>"en", "lat"=>48.090778, "lon"=>-100.617069),
"/US/US-ND/kathryn-15557" => array("type"=>"city", "pop"=>54, "name_en"=>"Kathryn", "lang"=>"en", "lat"=>46.679790, "lon"=>-97.967604),
"/US/US-ND/kenmare-15558" => array("type"=>"city", "pop"=>1075, "name_en"=>"Kenmare", "lang"=>"en", "lat"=>48.673974, "lon"=>-102.079828),
"/US/US-ND/kensal-15559" => array("type"=>"city", "pop"=>147, "name_en"=>"Kensal", "lang"=>"en", "lat"=>47.300315, "lon"=>-98.732540),
"/US/US-ND/kief-15560" => array("type"=>"city", "pop"=>11, "name_en"=>"Kief", "lang"=>"en", "lat"=>47.860535, "lon"=>-100.512554),
"/US/US-ND/killdeer-15561" => array("type"=>"city", "pop"=>673, "name_en"=>"Killdeer", "lang"=>"en", "lat"=>47.365435, "lon"=>-102.752696),
"/US/US-ND/kindred-15562" => array("type"=>"city", "pop"=>619, "name_en"=>"Kindred", "lang"=>"en", "lat"=>46.650926, "lon"=>-97.021058),
"/US/US-ND/knox-15563" => array("type"=>"city", "pop"=>59, "name_ru"=>"Нокс", "name_en"=>"Knox", "lang"=>"en", "lat"=>48.342686, "lon"=>-99.689291),
"/US/US-ND/kramer-15564" => array("type"=>"city", "pop"=>39, "name_en"=>"Kramer", "lang"=>"en", "lat"=>48.691193, "lon"=>-100.707072),
"/US/US-ND/kulm-15565" => array("type"=>"city", "pop"=>356, "name_en"=>"Kulm", "lang"=>"en", "lat"=>46.301875, "lon"=>-98.947751),
"/US/US-ND/la-moure-15567" => array("type"=>"city", "pop"=>113, "name_en"=>"La Moure", "lang"=>"en", "lat"=>46.357765, "lon"=>-98.296620),
"/US/US-ND/lakota-15566" => array("type"=>"city", "pop"=>696, "name_ru"=>"Лакота", "name_en"=>"Lakota", "lang"=>"en", "lat"=>48.043053, "lon"=>-98.347137),
"/US/US-ND/landa-15568" => array("type"=>"city", "pop"=>25, "name_en"=>"Landa", "lang"=>"en", "lat"=>48.895436, "lon"=>-100.910968),
"/US/US-ND/langdon-15569" => array("type"=>"city", "pop"=>1654, "name_ru"=>"Лэнгдон", "name_en"=>"Langdon", "lang"=>"en", "lat"=>48.762645, "lon"=>-98.372873),
"/US/US-ND/lankin-15570" => array("type"=>"city", "pop"=>113, "name_en"=>"Lankin", "lang"=>"en", "lat"=>48.314824, "lon"=>-97.920782),
"/US/US-ND/lansford-15571" => array("type"=>"city", "pop"=>226, "name_en"=>"Lansford", "lang"=>"en", "lat"=>48.626641, "lon"=>-101.376361),
"/US/US-ND/larimore-15572" => array("type"=>"city", "pop"=>1312, "name_en"=>"Larimore", "lang"=>"en", "lat"=>47.908389, "lon"=>-97.627338),
"/US/US-ND/larson-15573" => array("type"=>"city", "pop"=>26, "name_en"=>"Larson", "lang"=>"en", "lat"=>48.891116, "lon"=>-102.865588),
"/US/US-ND/lawton-15574" => array("type"=>"city", "pop"=>38, "name_ru"=>"Лотон", "name_en"=>"Lawton", "lang"=>"en", "lat"=>48.302895, "lon"=>-98.367374),
"/US/US-ND/leal-15575" => array("type"=>"city", "pop"=>32, "name_en"=>"Leal", "lang"=>"en", "lat"=>47.105118, "lon"=>-98.314002),
"/US/US-ND/leeds-15576" => array("type"=>"city", "pop"=>450, "name_ru"=>"Лидс", "name_en"=>"Leeds", "lang"=>"en", "lat"=>48.289357, "lon"=>-99.438500),
"/US/US-ND/lehr-15577" => array("type"=>"city", "pop"=>92, "name_en"=>"Lehr", "lang"=>"en", "lat"=>46.282793, "lon"=>-99.351733),
"/US/US-ND/leith-15578" => array("type"=>"city", "pop"=>23, "name_en"=>"Leith", "lang"=>"en", "lat"=>46.359906, "lon"=>-101.637285),
"/US/US-ND/leonard-15579" => array("type"=>"city", "pop"=>249, "name_ru"=>"Леонард", "name_en"=>"Leonard", "lang"=>"en", "lat"=>46.651329, "lon"=>-97.245507),
"/US/US-ND/lidgerwood-15580" => array("type"=>"city", "pop"=>706, "name_en"=>"Lidgerwood", "lang"=>"en", "lat"=>46.074862, "lon"=>-97.146303),
"/US/US-ND/lignite-15581" => array("type"=>"city", "pop"=>141, "name_ru"=>"Лайнит", "name_en"=>"Lignite", "lang"=>"en", "lat"=>48.876876, "lon"=>-102.563704),
"/US/US-ND/lincoln-15582" => array("type"=>"city", "pop"=>2733, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>46.765140, "lon"=>-100.700862),
"/US/US-ND/linton-15583" => array("type"=>"city", "pop"=>1015, "name_ru"=>"Линтон", "name_en"=>"Linton", "lang"=>"en", "lat"=>46.268382, "lon"=>-100.233211),
"/US/US-ND/lisbon-15584" => array("type"=>"city", "pop"=>2182, "name_ru"=>"Лисбон", "name_en"=>"Lisbon", "lang"=>"en", "lat"=>46.438294, "lon"=>-97.683152),
"/US/US-ND/litchville-15585" => array("type"=>"city", "pop"=>165, "name_en"=>"Litchville", "lang"=>"en", "lat"=>46.657391, "lon"=>-98.191326),
"/US/US-ND/loma-15586" => array("type"=>"city", "pop"=>17, "name_en"=>"Loma", "lang"=>"en", "lat"=>48.637891, "lon"=>-98.525007),
"/US/US-ND/loraine-15587" => array("type"=>"city", "pop"=>17, "name_ru"=>"Лорейн", "name_en"=>"Loraine", "lang"=>"en", "lat"=>48.867124, "lon"=>-101.567322),
"/US/US-ND/ludden-15588" => array("type"=>"city", "pop"=>27, "name_en"=>"Ludden", "lang"=>"en", "lat"=>46.007928, "lon"=>-98.124996),
"/US/US-ND/luverne-15589" => array("type"=>"city", "pop"=>35, "name_ru"=>"Леверн", "name_en"=>"Luverne", "lang"=>"en", "lat"=>47.250311, "lon"=>-97.934541),
"/US/US-ND/maddock-15593" => array("type"=>"city", "pop"=>479, "name_en"=>"Maddock", "lang"=>"en", "lat"=>47.962170, "lon"=>-99.529055),
"/US/US-ND/makoti-15594" => array("type"=>"city", "pop"=>136, "name_en"=>"Makoti", "lang"=>"en", "lat"=>47.960533, "lon"=>-101.804716),
"/US/US-ND/mandan-15595" => array("type"=>"city", "pop"=>18091, "name_en"=>"Mandan", "lang"=>"en", "lat"=>46.829350, "lon"=>-100.887849),
"/US/US-ND/mandaree-15596" => array("type"=>"cdp", "pop"=>367, "name_en"=>"Mandaree", "lang"=>"en", "lat"=>47.738391, "lon"=>-102.676696),
"/US/US-ND/mantador-15597" => array("type"=>"city", "pop"=>66, "name_en"=>"Mantador", "lang"=>"en", "lat"=>46.165891, "lon"=>-96.977774),
"/US/US-ND/manvel-15598" => array("type"=>"city", "pop"=>333, "name_en"=>"Manvel", "lang"=>"en", "lat"=>48.073209, "lon"=>-97.175670),
"/US/US-ND/mapleton-15599" => array("type"=>"city", "pop"=>725, "name_ru"=>"Мейплтон", "name_en"=>"Mapleton", "lang"=>"en", "lat"=>46.890778, "lon"=>-97.053440),
"/US/US-ND/marion-15600" => array("type"=>"city", "pop"=>123, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>46.609534, "lon"=>-98.334780),
"/US/US-ND/marmarth-15601" => array("type"=>"city", "pop"=>119, "name_en"=>"Marmarth", "lang"=>"en", "lat"=>46.295855, "lon"=>-103.933918),
"/US/US-ND/martin-15602" => array("type"=>"city", "pop"=>71, "name_ru"=>"Мартин", "name_en"=>"Martin", "lang"=>"en", "lat"=>47.826684, "lon"=>-100.114243),
"/US/US-ND/max-15603" => array("type"=>"city", "pop"=>249, "name_en"=>"Max", "lang"=>"en", "lat"=>47.819736, "lon"=>-101.290907),
"/US/US-ND/maxbass-15604" => array("type"=>"city", "pop"=>73, "name_en"=>"Maxbass", "lang"=>"en", "lat"=>48.722238, "lon"=>-101.141705),
"/US/US-ND/mayville-15605" => array("type"=>"city", "pop"=>1769, "name_ru"=>"Мейвилл", "name_en"=>"Mayville", "lang"=>"en", "lat"=>47.500094, "lon"=>-97.328248),
"/US/US-ND/maza-15606" => array("type"=>"city", "pop"=>20, "name_en"=>"Maza", "lang"=>"en", "lat"=>48.394901, "lon"=>-99.199532),
"/US/US-ND/mcclusky-15590" => array("type"=>"city", "pop"=>301, "name_en"=>"McClusky", "lang"=>"en", "lat"=>47.484283, "lon"=>-100.441697),
"/US/US-ND/mchenry-15591" => array("type"=>"city", "pop"=>67, "name_ru"=>"Мак-Генри", "name_en"=>"McHenry", "lang"=>"en", "lat"=>47.575838, "lon"=>-98.590759),
"/US/US-ND/mcville-15592" => array("type"=>"city", "pop"=>407, "name_en"=>"McVille", "lang"=>"en", "lat"=>47.765051, "lon"=>-98.176145),
"/US/US-ND/medina-15607" => array("type"=>"city", "pop"=>298, "name_ru"=>"Медина", "name_en"=>"Medina", "lang"=>"en", "lat"=>46.895093, "lon"=>-99.299986),
"/US/US-ND/medora-15608" => array("type"=>"city", "pop"=>93, "name_ru"=>"Медора", "name_en"=>"Medora", "lang"=>"en", "lat"=>46.913704, "lon"=>-103.524353),
"/US/US-ND/mercer-15609" => array("type"=>"city", "pop"=>75, "name_ru"=>"Мерсер", "name_en"=>"Mercer", "lang"=>"en", "lat"=>47.490965, "lon"=>-100.710211),
"/US/US-ND/merricourt-15610" => array("type"=>"city", "pop"=>9, "name_en"=>"Merricourt", "lang"=>"en", "lat"=>46.207927, "lon"=>-98.761200),
"/US/US-ND/michigan-city-15611" => array("type"=>"city", "pop"=>285, "name_ru"=>"Мичиган Сити", "name_en"=>"Michigan City", "lang"=>"en", "lat"=>48.024489, "lon"=>-98.120382),
"/US/US-ND/milnor-15612" => array("type"=>"city", "pop"=>677, "name_en"=>"Milnor", "lang"=>"en", "lat"=>46.260831, "lon"=>-97.456649),
"/US/US-ND/milton-15613" => array("type"=>"city", "pop"=>68, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>48.627216, "lon"=>-98.043728),
"/US/US-ND/minnewaukan-15614" => array("type"=>"city", "pop"=>295, "name_en"=>"Minnewaukan", "lang"=>"en", "lat"=>48.069847, "lon"=>-99.250100),
"/US/US-ND/minot-15615" => array("type"=>"city", "pop"=>35419, "name_en"=>"Minot", "lang"=>"en", "lat"=>48.233706, "lon"=>-101.296102),
"/US/US-ND/minot-afb-15616" => array("type"=>"cdp", "pop"=>9095, "name_en"=>"Minot AFB", "lang"=>"en", "lat"=>48.419890, "lon"=>-101.336341),
"/US/US-ND/minto-15617" => array("type"=>"city", "pop"=>589, "name_ru"=>"Минто", "name_en"=>"Minto", "lang"=>"en", "lat"=>48.292226, "lon"=>-97.372055),
"/US/US-ND/mohall-15618" => array("type"=>"city", "pop"=>696, "name_en"=>"Mohall", "lang"=>"en", "lat"=>48.766004, "lon"=>-101.510630),
"/US/US-ND/monango-15619" => array("type"=>"city", "pop"=>25, "name_en"=>"Monango", "lang"=>"en", "lat"=>46.172768, "lon"=>-98.595113),
"/US/US-ND/montpelier-15620" => array("type"=>"city", "pop"=>89, "name_ru"=>"Монтпильер", "name_en"=>"Montpelier", "lang"=>"en", "lat"=>46.698796, "lon"=>-98.587009),
"/US/US-ND/mooreton-15621" => array("type"=>"city", "pop"=>184, "name_en"=>"Mooreton", "lang"=>"en", "lat"=>46.269051, "lon"=>-96.876296),
"/US/US-ND/mott-15622" => array("type"=>"city", "pop"=>675, "name_en"=>"Mott", "lang"=>"en", "lat"=>46.374748, "lon"=>-102.318086),
"/US/US-ND/mountain-15623" => array("type"=>"city", "pop"=>122, "name_en"=>"Mountain", "lang"=>"en", "lat"=>48.683451, "lon"=>-97.863404),
"/US/US-ND/munich-15624" => array("type"=>"city", "pop"=>208, "name_en"=>"Munich", "lang"=>"en", "lat"=>48.669076, "lon"=>-98.832244),
"/US/US-ND/mylo-15625" => array("type"=>"city", "pop"=>19, "name_en"=>"Mylo", "lang"=>"en", "lat"=>48.635874, "lon"=>-99.617855),
"/US/US-ND/napoleon-15626" => array("type"=>"city", "pop"=>719, "name_ru"=>"Наполеон", "name_en"=>"Napoleon", "lang"=>"en", "lat"=>46.503592, "lon"=>-99.767966),
"/US/US-ND/neche-15627" => array("type"=>"city", "pop"=>393, "name_en"=>"Neche", "lang"=>"en", "lat"=>48.982019, "lon"=>-97.550883),
"/US/US-ND/nekoma-15628" => array("type"=>"city", "pop"=>41, "name_en"=>"Nekoma", "lang"=>"en", "lat"=>48.576053, "lon"=>-98.376117),
"/US/US-ND/new-england-15630" => array("type"=>"city", "pop"=>557, "name_en"=>"New England", "lang"=>"en", "lat"=>46.540945, "lon"=>-102.866675),
"/US/US-ND/new-leipzig-15631" => array("type"=>"city", "pop"=>224, "name_en"=>"New Leipzig", "lang"=>"en", "lat"=>46.376157, "lon"=>-101.949677),
"/US/US-ND/new-rockford-15632" => array("type"=>"city", "pop"=>1250, "name_en"=>"New Rockford", "lang"=>"en", "lat"=>47.680581, "lon"=>-99.135782),
"/US/US-ND/new-salem-15633" => array("type"=>"city", "pop"=>862, "name_ru"=>"Нью Салем", "name_en"=>"New Salem", "lang"=>"en", "lat"=>46.842778, "lon"=>-101.417423),
"/US/US-ND/new-town-15634" => array("type"=>"city", "pop"=>1712, "name_en"=>"New Town", "lang"=>"en", "lat"=>47.980802, "lon"=>-102.489660),
"/US/US-ND/newburg-15629" => array("type"=>"city", "pop"=>79, "name_ru"=>"Ньюбург", "name_en"=>"Newburg", "lang"=>"en", "lat"=>48.714715, "lon"=>-100.912298),
"/US/US-ND/niagara-15635" => array("type"=>"city", "pop"=>53, "name_en"=>"Niagara", "lang"=>"en", "lat"=>47.997361, "lon"=>-97.870463),
"/US/US-ND/nome-15636" => array("type"=>"city", "pop"=>63, "name_ru"=>"Ном", "name_en"=>"Nome", "lang"=>"en", "lat"=>46.675567, "lon"=>-97.815558),
"/US/US-ND/noonan-15637" => array("type"=>"city", "pop"=>139, "name_en"=>"Noonan", "lang"=>"en", "lat"=>48.890321, "lon"=>-103.009888),
"/US/US-ND/north-river-15638" => array("type"=>"city", "pop"=>62, "name_en"=>"North River", "lang"=>"en", "lat"=>46.949289, "lon"=>-96.802105),
"/US/US-ND/northwood-15639" => array("type"=>"city", "pop"=>927, "name_ru"=>"Нортвуд", "name_en"=>"Northwood", "lang"=>"en", "lat"=>47.735778, "lon"=>-97.568662),
"/US/US-ND/oakes-15640" => array("type"=>"city", "pop"=>1763, "name_en"=>"Oakes", "lang"=>"en", "lat"=>46.139420, "lon"=>-98.086225),
"/US/US-ND/oberon-15641" => array("type"=>"city", "pop"=>81, "name_en"=>"Oberon", "lang"=>"en", "lat"=>47.923870, "lon"=>-99.205248),
"/US/US-ND/oriska-15642" => array("type"=>"city", "pop"=>112, "name_en"=>"Oriska", "lang"=>"en", "lat"=>46.931369, "lon"=>-97.788725),
"/US/US-ND/osnabrock-15643" => array("type"=>"city", "pop"=>148, "name_en"=>"Osnabrock", "lang"=>"en", "lat"=>48.670213, "lon"=>-98.148933),
"/US/US-ND/overly-15644" => array("type"=>"city", "pop"=>17, "name_en"=>"Overly", "lang"=>"en", "lat"=>48.681228, "lon"=>-100.150731),
"/US/US-ND/oxbow-15645" => array("type"=>"city", "pop"=>238, "name_en"=>"Oxbow", "lang"=>"en", "lat"=>46.670659, "lon"=>-96.801595),
"/US/US-ND/page-15646" => array("type"=>"city", "pop"=>205, "name_ru"=>"Пейдж", "name_en"=>"Page", "lang"=>"en", "lat"=>47.158003, "lon"=>-97.570600),
"/US/US-ND/palermo-15647" => array("type"=>"city", "pop"=>69, "name_ru"=>"Палермо", "name_en"=>"Palermo", "lang"=>"en", "lat"=>48.338275, "lon"=>-102.228523),
"/US/US-ND/park-river-15648" => array("type"=>"city", "pop"=>1370, "name_en"=>"Park River", "lang"=>"en", "lat"=>48.393394, "lon"=>-97.744809),
"/US/US-ND/parshall-15649" => array("type"=>"city", "pop"=>1055, "name_en"=>"Parshall", "lang"=>"en", "lat"=>47.954793, "lon"=>-102.133684),
"/US/US-ND/pekin-15650" => array("type"=>"city", "pop"=>69, "name_ru"=>"Пекин", "name_en"=>"Pekin", "lang"=>"en", "lat"=>47.791023, "lon"=>-98.326644),
"/US/US-ND/pembina-15651" => array("type"=>"city", "pop"=>550, "name_en"=>"Pembina", "lang"=>"en", "lat"=>48.968001, "lon"=>-97.245656),
"/US/US-ND/perth-15652" => array("type"=>"city", "pop"=>10, "name_en"=>"Perth", "lang"=>"en", "lat"=>48.714638, "lon"=>-99.457207),
"/US/US-ND/petersburg-15653" => array("type"=>"city", "pop"=>165, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>48.011930, "lon"=>-97.999675),
"/US/US-ND/pettibone-15654" => array("type"=>"city", "pop"=>71, "name_en"=>"Pettibone", "lang"=>"en", "lat"=>47.117787, "lon"=>-99.520099),
"/US/US-ND/pick-city-15655" => array("type"=>"city", "pop"=>157, "name_en"=>"Pick City", "lang"=>"en", "lat"=>47.511902, "lon"=>-101.456093),
"/US/US-ND/pillsbury-15656" => array("type"=>"city", "pop"=>22, "name_en"=>"Pillsbury", "lang"=>"en", "lat"=>47.206820, "lon"=>-97.796055),
"/US/US-ND/pingree-15657" => array("type"=>"city", "pop"=>58, "name_en"=>"Pingree", "lang"=>"en", "lat"=>47.163298, "lon"=>-98.909006),
"/US/US-ND/pisek-15658" => array("type"=>"city", "pop"=>83, "name_en"=>"Pisek", "lang"=>"en", "lat"=>48.311098, "lon"=>-97.710304),
"/US/US-ND/plaza-15659" => array("type"=>"city", "pop"=>155, "name_en"=>"Plaza", "lang"=>"en", "lat"=>48.025641, "lon"=>-101.959354),
"/US/US-ND/portal-15660" => array("type"=>"city", "pop"=>110, "name_ru"=>"Портал", "name_en"=>"Portal", "lang"=>"en", "lat"=>48.995610, "lon"=>-102.547865),
"/US/US-ND/portland-15661" => array("type"=>"city", "pop"=>551, "name_ru"=>"Портленд", "name_en"=>"Portland", "lang"=>"en", "lat"=>47.499065, "lon"=>-97.368863),
"/US/US-ND/powers-lake-15662" => array("type"=>"city", "pop"=>244, "name_en"=>"Powers Lake", "lang"=>"en", "lat"=>48.563616, "lon"=>-102.643499),
"/US/US-ND/prairie-rose-15663" => array("type"=>"city", "pop"=>74, "name_en"=>"Prairie Rose", "lang"=>"en", "lat"=>46.817298, "lon"=>-96.834368),
"/US/US-ND/rawson-15664" => array("type"=>"city", "pop"=>9, "name_en"=>"Rawson", "lang"=>"en", "lat"=>47.819994, "lon"=>-103.540669),
"/US/US-ND/ray-15665" => array("type"=>"city", "pop"=>532, "name_en"=>"Ray", "lang"=>"en", "lat"=>48.341308, "lon"=>-103.162675),
"/US/US-ND/reeder-15666" => array("type"=>"city", "pop"=>154, "name_en"=>"Reeder", "lang"=>"en", "lat"=>46.107059, "lon"=>-102.942195),
"/US/US-ND/regan-15667" => array("type"=>"city", "pop"=>41, "name_en"=>"Regan", "lang"=>"en", "lat"=>47.157325, "lon"=>-100.527514),
"/US/US-ND/regent-15668" => array("type"=>"city", "pop"=>172, "name_en"=>"Regent", "lang"=>"en", "lat"=>46.422710, "lon"=>-102.557594),
"/US/US-ND/reile's-acres-15669" => array("type"=>"city", "pop"=>285, "name_en"=>"Reile's Acres", "lang"=>"en", "lat"=>46.922520, "lon"=>-96.862391),
"/US/US-ND/reynolds-15670" => array("type"=>"city", "pop"=>325, "name_ru"=>"Рейнолдс", "name_en"=>"Reynolds", "lang"=>"en", "lat"=>47.671078, "lon"=>-97.108427),
"/US/US-ND/rhame-15671" => array("type"=>"city", "pop"=>173, "name_en"=>"Rhame", "lang"=>"en", "lat"=>46.234038, "lon"=>-103.654237),
"/US/US-ND/richardton-15672" => array("type"=>"city", "pop"=>574, "name_en"=>"Richardton", "lang"=>"en", "lat"=>46.884624, "lon"=>-102.314998),
"/US/US-ND/riverdale-15673" => array("type"=>"city", "pop"=>264, "name_ru"=>"Ривердейл", "name_en"=>"Riverdale", "lang"=>"en", "lat"=>47.496642, "lon"=>-101.365531),
"/US/US-ND/robinson-15674" => array("type"=>"city", "pop"=>59, "name_ru"=>"Робинсон", "name_en"=>"Robinson", "lang"=>"en", "lat"=>47.143094, "lon"=>-99.780173),
"/US/US-ND/rocklake-15675" => array("type"=>"city", "pop"=>143, "name_en"=>"Rocklake", "lang"=>"en", "lat"=>48.790658, "lon"=>-99.244981),
"/US/US-ND/rogers-15676" => array("type"=>"city", "pop"=>54, "name_ru"=>"Роджерс", "name_en"=>"Rogers", "lang"=>"en", "lat"=>47.073121, "lon"=>-98.202089),
"/US/US-ND/rolette-15677" => array("type"=>"city", "pop"=>529, "name_en"=>"Rolette", "lang"=>"en", "lat"=>48.661244, "lon"=>-99.840553),
"/US/US-ND/rolla-15678" => array("type"=>"city", "pop"=>1420, "name_ru"=>"Ролла", "name_en"=>"Rolla", "lang"=>"en", "lat"=>48.858702, "lon"=>-99.617829),
"/US/US-ND/ross-15679" => array("type"=>"city", "pop"=>45, "name_ru"=>"Росс", "name_en"=>"Ross", "lang"=>"en", "lat"=>48.313086, "lon"=>-102.542806),
"/US/US-ND/rugby-15680" => array("type"=>"city", "pop"=>2574, "name_en"=>"Rugby", "lang"=>"en", "lat"=>48.367116, "lon"=>-99.990960),
"/US/US-ND/ruso-15681" => array("type"=>"city", "pop"=>5, "name_en"=>"Ruso", "lang"=>"en", "lat"=>47.836510, "lon"=>-100.933535),
"/US/US-ND/russell-15682" => array("type"=>"city", "pop"=>44, "name_ru"=>"Расселл", "name_en"=>"Russell", "lang"=>"en", "lat"=>48.673502, "lon"=>-100.901988),
"/US/US-ND/rutland-15683" => array("type"=>"city", "pop"=>195, "name_ru"=>"Рутленд", "name_en"=>"Rutland", "lang"=>"en", "lat"=>46.053771, "lon"=>-97.507908),
"/US/US-ND/ryder-15684" => array("type"=>"city", "pop"=>82, "name_en"=>"Ryder", "lang"=>"en", "lat"=>47.917302, "lon"=>-101.673115),
"/US/US-ND/sanborn-15687" => array("type"=>"city", "pop"=>181, "name_ru"=>"Сэнборн", "name_en"=>"Sanborn", "lang"=>"en", "lat"=>46.942689, "lon"=>-98.223332),
"/US/US-ND/sarles-15688" => array("type"=>"city", "pop"=>20, "name_en"=>"Sarles", "lang"=>"en", "lat"=>48.945225, "lon"=>-98.996459),
"/US/US-ND/sawyer-15689" => array("type"=>"city", "pop"=>340, "name_ru"=>"Сойер", "name_en"=>"Sawyer", "lang"=>"en", "lat"=>48.089326, "lon"=>-101.053992),
"/US/US-ND/scranton-15690" => array("type"=>"city", "pop"=>282, "name_ru"=>"Скрантон", "name_en"=>"Scranton", "lang"=>"en", "lat"=>46.147786, "lon"=>-103.142412),
"/US/US-ND/selfridge-15691" => array("type"=>"city", "pop"=>219, "name_en"=>"Selfridge", "lang"=>"en", "lat"=>46.042028, "lon"=>-100.924266),
"/US/US-ND/sentinel-butte-15692" => array("type"=>"city", "pop"=>52, "name_en"=>"Sentinel Butte", "lang"=>"en", "lat"=>46.919413, "lon"=>-103.839976),
"/US/US-ND/sharon-15693" => array("type"=>"city", "pop"=>87, "name_ru"=>"Шэрон", "name_en"=>"Sharon", "lang"=>"en", "lat"=>47.597132, "lon"=>-97.898553),
"/US/US-ND/sheldon-15694" => array("type"=>"city", "pop"=>120, "name_ru"=>"Шелдон", "name_en"=>"Sheldon", "lang"=>"en", "lat"=>46.586995, "lon"=>-97.492178),
"/US/US-ND/shell-valley-15695" => array("type"=>"cdp", "pop"=>470, "name_en"=>"Shell Valley", "lang"=>"en", "lat"=>48.797917, "lon"=>-99.864330),
"/US/US-ND/sherwood-15696" => array("type"=>"city", "pop"=>213, "name_ru"=>"Шервуд", "name_en"=>"Sherwood", "lang"=>"en", "lat"=>48.961451, "lon"=>-101.633126),
"/US/US-ND/sheyenne-15697" => array("type"=>"city", "pop"=>280, "name_en"=>"Sheyenne", "lang"=>"en", "lat"=>47.826823, "lon"=>-99.116619),
"/US/US-ND/sibley-15698" => array("type"=>"city", "pop"=>42, "name_ru"=>"Сибли", "name_en"=>"Sibley", "lang"=>"en", "lat"=>47.217288, "lon"=>-97.965274),
"/US/US-ND/solen-15699" => array("type"=>"city", "pop"=>88, "name_en"=>"Solen", "lang"=>"en", "lat"=>46.388735, "lon"=>-100.795431),
"/US/US-ND/souris-15700" => array("type"=>"city", "pop"=>74, "name_en"=>"Souris", "lang"=>"en", "lat"=>48.910297, "lon"=>-100.680471),
"/US/US-ND/south-heart-15701" => array("type"=>"city", "pop"=>298, "name_en"=>"South Heart", "lang"=>"en", "lat"=>46.863874, "lon"=>-102.991808),
"/US/US-ND/spiritwood-lake-15702" => array("type"=>"city", "pop"=>77, "name_en"=>"Spiritwood Lake", "lang"=>"en", "lat"=>47.074717, "lon"=>-98.587218),
"/US/US-ND/spring-brook-15703" => array("type"=>"city", "pop"=>29, "name_en"=>"Spring Brook", "lang"=>"en", "lat"=>48.251899, "lon"=>-103.462083),
"/US/US-ND/st.-john-15685" => array("type"=>"city", "pop"=>354, "name_ru"=>"Сент Джон", "name_en"=>"St. John", "lang"=>"en", "lat"=>48.943626, "lon"=>-99.710680),
"/US/US-ND/st.-thomas-15686" => array("type"=>"city", "pop"=>381, "name_en"=>"St. Thomas", "lang"=>"en", "lat"=>48.620171, "lon"=>-97.446899),
"/US/US-ND/stanley-15704" => array("type"=>"city", "pop"=>1218, "name_ru"=>"Стэнли", "name_en"=>"Stanley", "lang"=>"en", "lat"=>48.316833, "lon"=>-102.386459),
"/US/US-ND/stanton-15705" => array("type"=>"city", "pop"=>304, "name_ru"=>"Стэнтон", "name_en"=>"Stanton", "lang"=>"en", "lat"=>47.318490, "lon"=>-101.381991),
"/US/US-ND/starkweather-15706" => array("type"=>"city", "pop"=>144, "name_en"=>"Starkweather", "lang"=>"en", "lat"=>48.452332, "lon"=>-98.877746),
"/US/US-ND/steele-15707" => array("type"=>"city", "pop"=>650, "name_ru"=>"Стил", "name_en"=>"Steele", "lang"=>"en", "lat"=>46.856093, "lon"=>-99.916223),
"/US/US-ND/strasburg-15708" => array("type"=>"city", "pop"=>440, "name_ru"=>"Страсбург", "name_en"=>"Strasburg", "lang"=>"en", "lat"=>46.133157, "lon"=>-100.160415),
"/US/US-ND/streeter-15709" => array("type"=>"city", "pop"=>152, "name_en"=>"Streeter", "lang"=>"en", "lat"=>46.656921, "lon"=>-99.357111),
"/US/US-ND/surrey-15710" => array("type"=>"city", "pop"=>877, "name_en"=>"Surrey", "lang"=>"en", "lat"=>48.236040, "lon"=>-101.132585),
"/US/US-ND/sykeston-15711" => array("type"=>"city", "pop"=>128, "name_en"=>"Sykeston", "lang"=>"en", "lat"=>47.465713, "lon"=>-99.398533),
"/US/US-ND/tappen-15712" => array("type"=>"city", "pop"=>175, "name_en"=>"Tappen", "lang"=>"en", "lat"=>46.872924, "lon"=>-99.623363),
"/US/US-ND/taylor-15713" => array("type"=>"city", "pop"=>144, "name_ru"=>"Тейлор", "name_en"=>"Taylor", "lang"=>"en", "lat"=>46.902103, "lon"=>-102.422376),
"/US/US-ND/thompson-15714" => array("type"=>"city", "pop"=>959, "name_ru"=>"Томпсон", "name_en"=>"Thompson", "lang"=>"en", "lat"=>47.775027, "lon"=>-97.105774),
"/US/US-ND/tioga-15715" => array("type"=>"city", "pop"=>1096, "name_en"=>"Tioga", "lang"=>"en", "lat"=>48.393495, "lon"=>-102.938506),
"/US/US-ND/tolley-15716" => array("type"=>"city", "pop"=>55, "name_en"=>"Tolley", "lang"=>"en", "lat"=>48.730463, "lon"=>-101.826525),
"/US/US-ND/tolna-15717" => array("type"=>"city", "pop"=>176, "name_en"=>"Tolna", "lang"=>"en", "lat"=>47.825053, "lon"=>-98.437383),
"/US/US-ND/tower-city-15718" => array("type"=>"city", "pop"=>241, "name_en"=>"Tower City", "lang"=>"en", "lat"=>46.925312, "lon"=>-97.676059),
"/US/US-ND/towner-15719" => array("type"=>"city", "pop"=>483, "name_en"=>"Towner", "lang"=>"en", "lat"=>48.347610, "lon"=>-100.407194),
"/US/US-ND/turtle-lake-15720" => array("type"=>"city", "pop"=>502, "name_en"=>"Turtle Lake", "lang"=>"en", "lat"=>47.521494, "lon"=>-100.890316),
"/US/US-ND/tuttle-15721" => array("type"=>"city", "pop"=>88, "name_en"=>"Tuttle", "lang"=>"en", "lat"=>47.144056, "lon"=>-99.993589),
"/US/US-ND/underwood-15722" => array("type"=>"city", "pop"=>710, "name_ru"=>"Андервуд", "name_en"=>"Underwood", "lang"=>"en", "lat"=>47.454348, "lon"=>-101.141166),
"/US/US-ND/upham-15723" => array("type"=>"city", "pop"=>135, "name_en"=>"Upham", "lang"=>"en", "lat"=>48.580364, "lon"=>-100.728145),
"/US/US-ND/valley-city-15724" => array("type"=>"city", "pop"=>6230, "name_ru"=>"Валли Сити", "name_en"=>"Valley City", "lang"=>"en", "lat"=>46.923142, "lon"=>-98.005644),
"/US/US-ND/velva-15725" => array("type"=>"city", "pop"=>906, "name_en"=>"Velva", "lang"=>"en", "lat"=>48.058301, "lon"=>-100.931233),
"/US/US-ND/venturia-15726" => array("type"=>"city", "pop"=>18, "name_en"=>"Venturia", "lang"=>"en", "lat"=>45.997366, "lon"=>-99.549265),
"/US/US-ND/verona-15727" => array("type"=>"city", "pop"=>92, "name_ru"=>"Верона", "name_en"=>"Verona", "lang"=>"en", "lat"=>46.365262, "lon"=>-98.071045),
"/US/US-ND/voltaire-15728" => array("type"=>"city", "pop"=>45, "name_en"=>"Voltaire", "lang"=>"en", "lat"=>48.019273, "lon"=>-100.843899),
"/US/US-ND/wahpeton-15729" => array("type"=>"city", "pop"=>7585, "name_ru"=>"Уопетон", "name_en"=>"Wahpeton", "lang"=>"en", "lat"=>46.271012, "lon"=>-96.610088),
"/US/US-ND/walcott-15730" => array("type"=>"city", "pop"=>176, "name_ru"=>"Уолкотт", "name_en"=>"Walcott", "lang"=>"en", "lat"=>46.550440, "lon"=>-96.937006),
"/US/US-ND/wales-15731" => array("type"=>"city", "pop"=>24, "name_ru"=>"Уэльс", "name_en"=>"Wales", "lang"=>"en", "lat"=>48.894097, "lon"=>-98.601380),
"/US/US-ND/walhalla-15732" => array("type"=>"city", "pop"=>899, "name_en"=>"Walhalla", "lang"=>"en", "lat"=>48.921225, "lon"=>-97.916739),
"/US/US-ND/warwick-15733" => array("type"=>"city", "pop"=>75, "name_ru"=>"Уорвик", "name_en"=>"Warwick", "lang"=>"en", "lat"=>47.855686, "lon"=>-98.707065),
"/US/US-ND/washburn-15734" => array("type"=>"city", "pop"=>1239, "name_ru"=>"Уошберн", "name_en"=>"Washburn", "lang"=>"en", "lat"=>47.291320, "lon"=>-101.027224),
"/US/US-ND/watford-city-15735" => array("type"=>"city", "pop"=>1386, "name_en"=>"Watford City", "lang"=>"en", "lat"=>47.802323, "lon"=>-103.271413),
"/US/US-ND/west-fargo-15736" => array("type"=>"city", "pop"=>23708, "name_en"=>"West Fargo", "lang"=>"en", "lat"=>46.872850, "lon"=>-96.897044),
"/US/US-ND/westhope-15737" => array("type"=>"city", "pop"=>460, "name_en"=>"Westhope", "lang"=>"en", "lat"=>48.911833, "lon"=>-101.016605),
"/US/US-ND/wheelock-15738" => array("type"=>"city", "pop"=>56, "name_en"=>"Wheelock", "lang"=>"en", "lat"=>48.294988, "lon"=>-103.252488),
"/US/US-ND/white-earth-15739" => array("type"=>"city", "pop"=>59, "name_ru"=>"Уайт Эрт", "name_en"=>"White Earth", "lang"=>"en", "lat"=>48.380091, "lon"=>-102.771785),
"/US/US-ND/white-shield-15740" => array("type"=>"cdp", "pop"=>274, "name_en"=>"White Shield", "lang"=>"en", "lat"=>47.661348, "lon"=>-101.844803),
"/US/US-ND/wildrose-15741" => array("type"=>"city", "pop"=>126, "name_en"=>"Wildrose", "lang"=>"en", "lat"=>48.629216, "lon"=>-103.183081),
"/US/US-ND/williston-15742" => array("type"=>"city", "pop"=>12641, "name_ru"=>"Уиллистон", "name_en"=>"Williston", "lang"=>"en", "lat"=>48.162292, "lon"=>-103.630917),
"/US/US-ND/willow-city-15743" => array("type"=>"city", "pop"=>197, "name_en"=>"Willow City", "lang"=>"en", "lat"=>48.604665, "lon"=>-100.292705),
"/US/US-ND/wilton-15744" => array("type"=>"city", "pop"=>720, "name_ru"=>"Уилтон", "name_en"=>"Wilton", "lang"=>"en", "lat"=>47.158653, "lon"=>-100.784011),
"/US/US-ND/wimbledon-15745" => array("type"=>"city", "pop"=>205, "name_en"=>"Wimbledon", "lang"=>"en", "lat"=>47.170670, "lon"=>-98.459568),
"/US/US-ND/wing-15746" => array("type"=>"city", "pop"=>118, "name_en"=>"Wing", "lang"=>"en", "lat"=>47.142583, "lon"=>-100.282196),
"/US/US-ND/wishek-15747" => array("type"=>"city", "pop"=>876, "name_en"=>"Wishek", "lang"=>"en", "lat"=>46.255654, "lon"=>-99.555200),
"/US/US-ND/wolford-15748" => array("type"=>"city", "pop"=>44, "name_en"=>"Wolford", "lang"=>"en", "lat"=>48.497782, "lon"=>-99.704533),
"/US/US-ND/woodworth-15749" => array("type"=>"city", "pop"=>71, "name_ru"=>"Вудворт", "name_en"=>"Woodworth", "lang"=>"en", "lat"=>47.142227, "lon"=>-99.303873),
"/US/US-ND/wyndmere-15750" => array("type"=>"city", "pop"=>472, "name_en"=>"Wyndmere", "lang"=>"en", "lat"=>46.264222, "lon"=>-97.131048),
"/US/US-ND/york-15751" => array("type"=>"city", "pop"=>26, "name_ru"=>"Йорк", "name_en"=>"York", "lang"=>"en", "lat"=>48.313476, "lon"=>-99.573416),
"/US/US-ND/zap-15752" => array("type"=>"city", "pop"=>209, "name_en"=>"Zap", "lang"=>"en", "lat"=>47.284757, "lon"=>-101.921894),
"/US/US-ND/zeeland-15753" => array("type"=>"city", "pop"=>111, "name_ru"=>"Зиленд", "name_en"=>"Zeeland", "lang"=>"en", "lat"=>45.973062, "lon"=>-99.832250),
"/US/US-NE" => array("type"=>"state", "pop"=>1783432, "name_ru"=>"Небраска", "name_en"=>"Nebraska", "lang"=>"en"),
"/US/US-NE/abie-12470" => array("type"=>"village", "pop"=>103, "name_en"=>"Abie", "lang"=>"en", "lat"=>41.333708, "lon"=>-96.949340),
"/US/US-NE/adams-12471" => array("type"=>"village", "pop"=>499, "name_ru"=>"Адамс", "name_en"=>"Adams", "lang"=>"en", "lat"=>40.456224, "lon"=>-96.511204),
"/US/US-NE/ainsworth-12472" => array("type"=>"city", "pop"=>1628, "name_ru"=>"Эйнсворт", "name_en"=>"Ainsworth", "lang"=>"en", "lat"=>42.548978, "lon"=>-99.857184),
"/US/US-NE/albion-12473" => array("type"=>"city", "pop"=>1601, "name_ru"=>"Албион", "name_en"=>"Albion", "lang"=>"en", "lat"=>41.689331, "lon"=>-98.002789),
"/US/US-NE/alda-12474" => array("type"=>"village", "pop"=>640, "name_en"=>"Alda", "lang"=>"en", "lat"=>40.864702, "lon"=>-98.468808),
"/US/US-NE/alexandria-12475" => array("type"=>"village", "pop"=>177, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>40.247414, "lon"=>-97.386837),
"/US/US-NE/allen-12476" => array("type"=>"village", "pop"=>412, "name_ru"=>"Аллен", "name_en"=>"Allen", "lang"=>"en", "lat"=>42.414555, "lon"=>-96.842811),
"/US/US-NE/alliance-12477" => array("type"=>"city", "pop"=>8109, "name_en"=>"Alliance", "lang"=>"en", "lat"=>42.100372, "lon"=>-102.873929),
"/US/US-NE/alma-12478" => array("type"=>"city", "pop"=>1074, "name_ru"=>"Алма", "name_en"=>"Alma", "lang"=>"en", "lat"=>40.101696, "lon"=>-99.362235),
"/US/US-NE/alvo-12479" => array("type"=>"village", "pop"=>139, "name_en"=>"Alvo", "lang"=>"en", "lat"=>40.871493, "lon"=>-96.386645),
"/US/US-NE/amherst-12480" => array("type"=>"village", "pop"=>272, "name_ru"=>"Амхерст", "name_en"=>"Amherst", "lang"=>"en", "lat"=>40.838766, "lon"=>-99.269146),
"/US/US-NE/anoka-12481" => array("type"=>"village", "pop"=>9, "name_ru"=>"Анока", "name_en"=>"Anoka", "lang"=>"en", "lat"=>42.946917, "lon"=>-98.829878),
"/US/US-NE/anselmo-12482" => array("type"=>"village", "pop"=>143, "name_en"=>"Anselmo", "lang"=>"en", "lat"=>41.618499, "lon"=>-99.864376),
"/US/US-NE/ansley-12483" => array("type"=>"village", "pop"=>472, "name_en"=>"Ansley", "lang"=>"en", "lat"=>41.287596, "lon"=>-99.382172),
"/US/US-NE/arapahoe-12484" => array("type"=>"city", "pop"=>886, "name_en"=>"Arapahoe", "lang"=>"en", "lat"=>40.305028, "lon"=>-99.898085),
"/US/US-NE/arcadia-12485" => array("type"=>"village", "pop"=>317, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>41.424689, "lon"=>-99.125625),
"/US/US-NE/arlington-12486" => array("type"=>"village", "pop"=>1184, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>41.454248, "lon"=>-96.355542),
"/US/US-NE/arnold-12487" => array("type"=>"village", "pop"=>586, "name_ru"=>"Арнолд", "name_en"=>"Arnold", "lang"=>"en", "lat"=>41.423859, "lon"=>-100.193820),
"/US/US-NE/arthur-12488" => array("type"=>"village", "pop"=>110, "name_ru"=>"Артур", "name_en"=>"Arthur", "lang"=>"en", "lat"=>41.572292, "lon"=>-101.691890),
"/US/US-NE/ashland-12489" => array("type"=>"city", "pop"=>2563, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>41.040500, "lon"=>-96.370899),
"/US/US-NE/ashton-12490" => array("type"=>"village", "pop"=>210, "name_ru"=>"Аштон", "name_en"=>"Ashton", "lang"=>"en", "lat"=>41.247588, "lon"=>-98.794461),
"/US/US-NE/atkinson-12491" => array("type"=>"city", "pop"=>1095, "name_ru"=>"Аткинсон", "name_en"=>"Atkinson", "lang"=>"en", "lat"=>42.531670, "lon"=>-98.976440),
"/US/US-NE/atlanta-12492" => array("type"=>"village", "pop"=>123, "name_ru"=>"Атланта", "name_en"=>"Atlanta", "lang"=>"en", "lat"=>40.368256, "lon"=>-99.472360),
"/US/US-NE/auburn-12493" => array("type"=>"city", "pop"=>3358, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>40.388267, "lon"=>-95.841968),
"/US/US-NE/aurora-12494" => array("type"=>"city", "pop"=>4198, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>40.865743, "lon"=>-98.003120),
"/US/US-NE/avoca-12495" => array("type"=>"village", "pop"=>264, "name_ru"=>"Авока", "name_en"=>"Avoca", "lang"=>"en", "lat"=>40.796374, "lon"=>-96.117829),
"/US/US-NE/axtell-12496" => array("type"=>"village", "pop"=>693, "name_ru"=>"Акстелл", "name_en"=>"Axtell", "lang"=>"en", "lat"=>40.478954, "lon"=>-99.128250),
"/US/US-NE/ayr-12497" => array("type"=>"village", "pop"=>102, "name_en"=>"Ayr", "lang"=>"en", "lat"=>40.437876, "lon"=>-98.440226),
"/US/US-NE/bancroft-12498" => array("type"=>"village", "pop"=>470, "name_ru"=>"Бэнкрофт", "name_en"=>"Bancroft", "lang"=>"en", "lat"=>42.010360, "lon"=>-96.573024),
"/US/US-NE/barada-12499" => array("type"=>"village", "pop"=>25, "name_en"=>"Barada", "lang"=>"en", "lat"=>40.218265, "lon"=>-95.578618),
"/US/US-NE/barneston-12500" => array("type"=>"village", "pop"=>120, "name_en"=>"Barneston", "lang"=>"en", "lat"=>40.048656, "lon"=>-96.576738),
"/US/US-NE/bartlett-12501" => array("type"=>"village", "pop"=>113, "name_ru"=>"Барлет", "name_en"=>"Bartlett", "lang"=>"en", "lat"=>41.884144, "lon"=>-98.551379),
"/US/US-NE/bartley-12502" => array("type"=>"village", "pop"=>336, "name_en"=>"Bartley", "lang"=>"en", "lat"=>40.252501, "lon"=>-100.309250),
"/US/US-NE/bassett-12503" => array("type"=>"city", "pop"=>636, "name_ru"=>"Бассет", "name_en"=>"Bassett", "lang"=>"en", "lat"=>42.582587, "lon"=>-99.535671),
"/US/US-NE/battle-creek-12504" => array("type"=>"city", "pop"=>1139, "name_ru"=>"Бэттл Крик", "name_en"=>"Battle Creek", "lang"=>"en", "lat"=>41.999042, "lon"=>-97.598672),
"/US/US-NE/bayard-12505" => array("type"=>"city", "pop"=>1117, "name_ru"=>"Байяр", "name_en"=>"Bayard", "lang"=>"en", "lat"=>41.757062, "lon"=>-103.322955),
"/US/US-NE/bazile-mills-12506" => array("type"=>"village", "pop"=>24, "name_en"=>"Bazile Mills", "lang"=>"en", "lat"=>42.513774, "lon"=>-97.910020),
"/US/US-NE/beatrice-12507" => array("type"=>"city", "pop"=>12769, "name_ru"=>"Биатрис", "name_en"=>"Beatrice", "lang"=>"en", "lat"=>40.269843, "lon"=>-96.745630),
"/US/US-NE/beaver-city-12508" => array("type"=>"city", "pop"=>552, "name_en"=>"Beaver City", "lang"=>"en", "lat"=>40.137734, "lon"=>-99.827346),
"/US/US-NE/beaver-crossing-12509" => array("type"=>"village", "pop"=>434, "name_en"=>"Beaver Crossing", "lang"=>"en", "lat"=>40.777775, "lon"=>-97.281873),
"/US/US-NE/bee-12510" => array("type"=>"village", "pop"=>215, "name_en"=>"Bee", "lang"=>"en", "lat"=>41.006923, "lon"=>-97.057677),
"/US/US-NE/beemer-12511" => array("type"=>"village", "pop"=>691, "name_en"=>"Beemer", "lang"=>"en", "lat"=>41.930439, "lon"=>-96.809549),
"/US/US-NE/belden-12512" => array("type"=>"village", "pop"=>114, "name_en"=>"Belden", "lang"=>"en", "lat"=>42.411685, "lon"=>-97.207560),
"/US/US-NE/belgrade-12513" => array("type"=>"village", "pop"=>116, "name_ru"=>"Белгрейд", "name_en"=>"Belgrade", "lang"=>"en", "lat"=>41.471319, "lon"=>-98.067068),
"/US/US-NE/bellevue-12514" => array("type"=>"city", "pop"=>49699, "name_ru"=>"Белвью", "name_en"=>"Bellevue", "lang"=>"en", "lat"=>41.156450, "lon"=>-95.922642),
"/US/US-NE/bellwood-12515" => array("type"=>"village", "pop"=>413, "name_ru"=>"Беллвуд", "name_en"=>"Bellwood", "lang"=>"en", "lat"=>41.342139, "lon"=>-97.240163),
"/US/US-NE/belvidere-12516" => array("type"=>"village", "pop"=>83, "name_ru"=>"Белвидер", "name_en"=>"Belvidere", "lang"=>"en", "lat"=>40.254413, "lon"=>-97.556857),
"/US/US-NE/benedict-12517" => array("type"=>"village", "pop"=>274, "name_ru"=>"Бенедикт", "name_en"=>"Benedict", "lang"=>"en", "lat"=>41.006617, "lon"=>-97.606784),
"/US/US-NE/benkelman-12518" => array("type"=>"city", "pop"=>859, "name_en"=>"Benkelman", "lang"=>"en", "lat"=>40.051627, "lon"=>-101.534567),
"/US/US-NE/bennet-12519" => array("type"=>"village", "pop"=>732, "name_en"=>"Bennet", "lang"=>"en", "lat"=>40.681250, "lon"=>-96.505880),
"/US/US-NE/bennington-12520" => array("type"=>"village", "pop"=>1014, "name_ru"=>"Беннингтон", "name_en"=>"Bennington", "lang"=>"en", "lat"=>41.367600, "lon"=>-96.156932),
"/US/US-NE/bertrand-12521" => array("type"=>"village", "pop"=>763, "name_ru"=>"Бертранд", "name_en"=>"Bertrand", "lang"=>"en", "lat"=>40.526670, "lon"=>-99.632692),
"/US/US-NE/berwyn-12522" => array("type"=>"village", "pop"=>124, "name_ru"=>"Бервин", "name_en"=>"Berwyn", "lang"=>"en", "lat"=>41.351281, "lon"=>-99.500159),
"/US/US-NE/big-springs-12523" => array("type"=>"village", "pop"=>379, "name_en"=>"Big Springs", "lang"=>"en", "lat"=>41.063408, "lon"=>-102.074651),
"/US/US-NE/bladen-12524" => array("type"=>"village", "pop"=>253, "name_en"=>"Bladen", "lang"=>"en", "lat"=>40.323565, "lon"=>-98.595131),
"/US/US-NE/blair-12525" => array("type"=>"city", "pop"=>7789, "name_en"=>"Blair", "lang"=>"en", "lat"=>41.544184, "lon"=>-96.136700),
"/US/US-NE/bloomfield-12526" => array("type"=>"city", "pop"=>986, "name_ru"=>"Блумфилд", "name_en"=>"Bloomfield", "lang"=>"en", "lat"=>42.598818, "lon"=>-97.648840),
"/US/US-NE/bloomington-12527" => array("type"=>"village", "pop"=>106, "name_ru"=>"Блумингтон", "name_en"=>"Bloomington", "lang"=>"en", "lat"=>40.093603, "lon"=>-99.037783),
"/US/US-NE/blue-hill-12528" => array("type"=>"city", "pop"=>747, "name_en"=>"Blue Hill", "lang"=>"en", "lat"=>40.333032, "lon"=>-98.447635),
"/US/US-NE/blue-springs-12529" => array("type"=>"city", "pop"=>368, "name_ru"=>"Блу Спрингс", "name_en"=>"Blue Springs", "lang"=>"en", "lat"=>40.136983, "lon"=>-96.662591),
"/US/US-NE/boys-town-12530" => array("type"=>"village", "pop"=>977, "name_en"=>"Boys Town", "lang"=>"en", "lat"=>41.259450, "lon"=>-96.131156),
"/US/US-NE/bradshaw-12531" => array("type"=>"village", "pop"=>314, "name_en"=>"Bradshaw", "lang"=>"en", "lat"=>40.883692, "lon"=>-97.746242),
"/US/US-NE/brady-12532" => array("type"=>"village", "pop"=>371, "name_en"=>"Brady", "lang"=>"en", "lat"=>41.022788, "lon"=>-100.367418),
"/US/US-NE/brainard-12533" => array("type"=>"village", "pop"=>340, "name_en"=>"Brainard", "lang"=>"en", "lat"=>41.182708, "lon"=>-97.002173),
"/US/US-NE/brewster-12534" => array("type"=>"village", "pop"=>21, "name_ru"=>"Брюстер", "name_en"=>"Brewster", "lang"=>"en", "lat"=>41.938373, "lon"=>-99.865089),
"/US/US-NE/bridgeport-12535" => array("type"=>"city", "pop"=>1448, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>41.666665, "lon"=>-103.098300),
"/US/US-NE/bristow-12536" => array("type"=>"village", "pop"=>76, "name_ru"=>"Бристоу", "name_en"=>"Bristow", "lang"=>"en", "lat"=>42.839903, "lon"=>-98.583653),
"/US/US-NE/broadwater-12537" => array("type"=>"village", "pop"=>130, "name_en"=>"Broadwater", "lang"=>"en", "lat"=>41.597594, "lon"=>-102.851509),
"/US/US-NE/brock-12538" => array("type"=>"village", "pop"=>146, "name_en"=>"Brock", "lang"=>"en", "lat"=>40.481143, "lon"=>-95.959660),
"/US/US-NE/broken-bow-12539" => array("type"=>"city", "pop"=>3154, "name_en"=>"Broken Bow", "lang"=>"en", "lat"=>41.405371, "lon"=>-99.639055),
"/US/US-NE/brownville-12540" => array("type"=>"village", "pop"=>136, "name_en"=>"Brownville", "lang"=>"en", "lat"=>40.397059, "lon"=>-95.661445),
"/US/US-NE/brule-12541" => array("type"=>"village", "pop"=>311, "name_en"=>"Brule", "lang"=>"en", "lat"=>41.095890, "lon"=>-101.888434),
"/US/US-NE/bruning-12542" => array("type"=>"village", "pop"=>245, "name_en"=>"Bruning", "lang"=>"en", "lat"=>40.335672, "lon"=>-97.564562),
"/US/US-NE/bruno-12543" => array("type"=>"village", "pop"=>98, "name_ru"=>"Бруно", "name_en"=>"Bruno", "lang"=>"en", "lat"=>41.283292, "lon"=>-96.960502),
"/US/US-NE/brunswick-12544" => array("type"=>"village", "pop"=>157, "name_ru"=>"Брунсвик", "name_en"=>"Brunswick", "lang"=>"en", "lat"=>42.338234, "lon"=>-97.971226),
"/US/US-NE/burchard-12545" => array("type"=>"village", "pop"=>87, "name_en"=>"Burchard", "lang"=>"en", "lat"=>40.149754, "lon"=>-96.347839),
"/US/US-NE/burr-12546" => array("type"=>"village", "pop"=>64, "name_en"=>"Burr", "lang"=>"en", "lat"=>40.535931, "lon"=>-96.298871),
"/US/US-NE/burton-12547" => array("type"=>"village", "pop"=>9, "name_ru"=>"Бертон", "name_en"=>"Burton", "lang"=>"en", "lat"=>42.911356, "lon"=>-99.590602),
"/US/US-NE/burwell-12548" => array("type"=>"city", "pop"=>1005, "name_en"=>"Burwell", "lang"=>"en", "lat"=>41.780007, "lon"=>-99.133867),
"/US/US-NE/bushnell-12549" => array("type"=>"village", "pop"=>136, "name_ru"=>"Бушнелл", "name_en"=>"Bushnell", "lang"=>"en", "lat"=>41.231959, "lon"=>-103.890380),
"/US/US-NE/butte-12550" => array("type"=>"village", "pop"=>321, "name_ru"=>"Бьютт", "name_en"=>"Butte", "lang"=>"en", "lat"=>42.912626, "lon"=>-98.847950),
"/US/US-NE/byron-12551" => array("type"=>"village", "pop"=>123, "name_ru"=>"Байрон", "name_en"=>"Byron", "lang"=>"en", "lat"=>40.005235, "lon"=>-97.768099),
"/US/US-NE/cairo-12552" => array("type"=>"village", "pop"=>782, "name_ru"=>"Каир", "name_en"=>"Cairo", "lang"=>"en", "lat"=>41.001623, "lon"=>-98.607726),
"/US/US-NE/callaway-12553" => array("type"=>"village", "pop"=>601, "name_ru"=>"Каллавэй", "name_en"=>"Callaway", "lang"=>"en", "lat"=>41.291330, "lon"=>-99.920042),
"/US/US-NE/cambridge-12554" => array("type"=>"city", "pop"=>915, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>40.283353, "lon"=>-100.170484),
"/US/US-NE/campbell-12555" => array("type"=>"village", "pop"=>338, "name_ru"=>"Кэмпбелл", "name_en"=>"Campbell", "lang"=>"en", "lat"=>40.296765, "lon"=>-98.730893),
"/US/US-NE/carleton-12556" => array("type"=>"village", "pop"=>116, "name_ru"=>"Карлтон", "name_en"=>"Carleton", "lang"=>"en", "lat"=>40.301750, "lon"=>-97.673527),
"/US/US-NE/carroll-12557" => array("type"=>"village", "pop"=>214, "name_ru"=>"Кэррол", "name_en"=>"Carroll", "lang"=>"en", "lat"=>42.276349, "lon"=>-97.191057),
"/US/US-NE/cedar-bluffs-12558" => array("type"=>"village", "pop"=>585, "name_en"=>"Cedar Bluffs", "lang"=>"en", "lat"=>41.397837, "lon"=>-96.609220),
"/US/US-NE/cedar-creek-12559" => array("type"=>"village", "pop"=>398, "name_en"=>"Cedar Creek", "lang"=>"en", "lat"=>41.045388, "lon"=>-96.097905),
"/US/US-NE/cedar-rapids-12560" => array("type"=>"village", "pop"=>348, "name_ru"=>"Сидар Рэпидс", "name_en"=>"Cedar Rapids", "lang"=>"en", "lat"=>41.558804, "lon"=>-98.149284),
"/US/US-NE/center-12561" => array("type"=>"village", "pop"=>79, "name_ru"=>"Сентер", "name_en"=>"Center", "lang"=>"en", "lat"=>42.608908, "lon"=>-97.875879),
"/US/US-NE/central-city-12562" => array("type"=>"city", "pop"=>2828, "name_ru"=>"Сентрал Сити", "name_en"=>"Central City", "lang"=>"en", "lat"=>41.114425, "lon"=>-98.002771),
"/US/US-NE/ceresco-12563" => array("type"=>"village", "pop"=>857, "name_en"=>"Ceresco", "lang"=>"en", "lat"=>41.057729, "lon"=>-96.645330),
"/US/US-NE/chadron-12564" => array("type"=>"city", "pop"=>5429, "name_en"=>"Chadron", "lang"=>"en", "lat"=>42.827062, "lon"=>-103.000093),
"/US/US-NE/chalco-12565" => array("type"=>"cdp", "pop"=>7337, "name_en"=>"Chalco", "lang"=>"en", "lat"=>41.183413, "lon"=>-96.133891),
"/US/US-NE/chambers-12566" => array("type"=>"village", "pop"=>294, "name_en"=>"Chambers", "lang"=>"en", "lat"=>42.205128, "lon"=>-98.747804),
"/US/US-NE/chapman-12567" => array("type"=>"village", "pop"=>306, "name_ru"=>"Чэпмен", "name_en"=>"Chapman", "lang"=>"en", "lat"=>41.022863, "lon"=>-98.159654),
"/US/US-NE/chappell-12568" => array("type"=>"city", "pop"=>878, "name_en"=>"Chappell", "lang"=>"en", "lat"=>41.094095, "lon"=>-102.467044),
"/US/US-NE/chester-12569" => array("type"=>"village", "pop"=>239, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>40.009581, "lon"=>-97.617665),
"/US/US-NE/clarks-12570" => array("type"=>"village", "pop"=>318, "name_ru"=>"Кларкс", "name_en"=>"Clarks", "lang"=>"en", "lat"=>41.216283, "lon"=>-97.839419),
"/US/US-NE/clarkson-12571" => array("type"=>"city", "pop"=>654, "name_ru"=>"Кларксон", "name_en"=>"Clarkson", "lang"=>"en", "lat"=>41.724658, "lon"=>-97.121257),
"/US/US-NE/clatonia-12572" => array("type"=>"village", "pop"=>262, "name_en"=>"Clatonia", "lang"=>"en", "lat"=>40.464926, "lon"=>-96.851161),
"/US/US-NE/clay-center-12573" => array("type"=>"city", "pop"=>751, "name_ru"=>"Клэй Сентер", "name_en"=>"Clay Center", "lang"=>"en", "lat"=>40.523321, "lon"=>-98.054742),
"/US/US-NE/clearwater-12574" => array("type"=>"village", "pop"=>338, "name_ru"=>"Клирвотер", "name_en"=>"Clearwater", "lang"=>"en", "lat"=>42.170635, "lon"=>-98.188572),
"/US/US-NE/clinton-12575" => array("type"=>"village", "pop"=>27, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>42.759385, "lon"=>-102.348223),
"/US/US-NE/cody-12576" => array("type"=>"village", "pop"=>134, "name_en"=>"Cody", "lang"=>"en", "lat"=>42.937077, "lon"=>-101.247736),
"/US/US-NE/coleridge-12577" => array("type"=>"village", "pop"=>468, "name_en"=>"Coleridge", "lang"=>"en", "lat"=>42.506349, "lon"=>-97.202338),
"/US/US-NE/colon-12578" => array("type"=>"village", "pop"=>131, "name_ru"=>"Колон", "name_en"=>"Colon", "lang"=>"en", "lat"=>41.297759, "lon"=>-96.606453),
"/US/US-NE/columbus-12579" => array("type"=>"city", "pop"=>21595, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>41.434707, "lon"=>-97.355620),
"/US/US-NE/comstock-12580" => array("type"=>"village", "pop"=>97, "name_ru"=>"Комсток", "name_en"=>"Comstock", "lang"=>"en", "lat"=>41.557871, "lon"=>-99.242539),
"/US/US-NE/concord-12581" => array("type"=>"village", "pop"=>159, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>42.384224, "lon"=>-96.988750),
"/US/US-NE/cook-12582" => array("type"=>"village", "pop"=>285, "name_ru"=>"Кук", "name_en"=>"Cook", "lang"=>"en", "lat"=>40.510309, "lon"=>-96.161227),
"/US/US-NE/cordova-12583" => array("type"=>"village", "pop"=>124, "name_ru"=>"Кордова", "name_en"=>"Cordova", "lang"=>"en", "lat"=>40.716394, "lon"=>-97.350853),
"/US/US-NE/cornlea-12584" => array("type"=>"village", "pop"=>41, "name_en"=>"Cornlea", "lang"=>"en", "lat"=>41.680723, "lon"=>-97.567089),
"/US/US-NE/cortland-12585" => array("type"=>"village", "pop"=>493, "name_ru"=>"Кортленд", "name_en"=>"Cortland", "lang"=>"en", "lat"=>40.506118, "lon"=>-96.705562),
"/US/US-NE/cotesfield-12586" => array("type"=>"village", "pop"=>62, "name_en"=>"Cotesfield", "lang"=>"en", "lat"=>41.357692, "lon"=>-98.633024),
"/US/US-NE/cowles-12587" => array("type"=>"village", "pop"=>42, "name_en"=>"Cowles", "lang"=>"en", "lat"=>40.170983, "lon"=>-98.447969),
"/US/US-NE/cozad-12588" => array("type"=>"city", "pop"=>4218, "name_en"=>"Cozad", "lang"=>"en", "lat"=>40.862922, "lon"=>-99.986486),
"/US/US-NE/crab-orchard-12589" => array("type"=>"village", "pop"=>45, "name_ru"=>"Крэб Орчард", "name_en"=>"Crab Orchard", "lang"=>"en", "lat"=>40.334402, "lon"=>-96.421705),
"/US/US-NE/craig-12590" => array("type"=>"village", "pop"=>220, "name_ru"=>"Крейг", "name_en"=>"Craig", "lang"=>"en", "lat"=>41.784945, "lon"=>-96.361591),
"/US/US-NE/crawford-12591" => array("type"=>"city", "pop"=>1028, "name_ru"=>"Крофорд", "name_en"=>"Crawford", "lang"=>"en", "lat"=>42.685456, "lon"=>-103.414236),
"/US/US-NE/creighton-12592" => array("type"=>"city", "pop"=>1128, "name_ru"=>"Крейтон", "name_en"=>"Creighton", "lang"=>"en", "lat"=>42.465109, "lon"=>-97.906846),
"/US/US-NE/creston-12593" => array("type"=>"village", "pop"=>215, "name_ru"=>"Крестон", "name_en"=>"Creston", "lang"=>"en", "lat"=>41.707065, "lon"=>-97.361518),
"/US/US-NE/crete-12594" => array("type"=>"city", "pop"=>6241, "name_ru"=>"Крет", "name_en"=>"Crete", "lang"=>"en", "lat"=>40.625640, "lon"=>-96.958305),
"/US/US-NE/crofton-12595" => array("type"=>"city", "pop"=>684, "name_ru"=>"Крофтон", "name_en"=>"Crofton", "lang"=>"en", "lat"=>42.731790, "lon"=>-97.497663),
"/US/US-NE/crookston-12596" => array("type"=>"village", "pop"=>85, "name_ru"=>"Крукстон", "name_en"=>"Crookston", "lang"=>"en", "lat"=>42.925346, "lon"=>-100.752864),
"/US/US-NE/culbertson-12597" => array("type"=>"village", "pop"=>530, "name_en"=>"Culbertson", "lang"=>"en", "lat"=>40.228464, "lon"=>-100.835648),
"/US/US-NE/curtis-12598" => array("type"=>"city", "pop"=>681, "name_en"=>"Curtis", "lang"=>"en", "lat"=>40.634522, "lon"=>-100.514248),
"/US/US-NE/cushing-12599" => array("type"=>"village", "pop"=>29, "name_ru"=>"Кушинг", "name_en"=>"Cushing", "lang"=>"en", "lat"=>41.294589, "lon"=>-98.369002),
"/US/US-NE/dakota-city-12600" => array("type"=>"city", "pop"=>1872, "name_ru"=>"Дакота Сити", "name_en"=>"Dakota City", "lang"=>"en", "lat"=>42.415951, "lon"=>-96.417447),
"/US/US-NE/dalton-12601" => array("type"=>"village", "pop"=>310, "name_ru"=>"Далтон", "name_en"=>"Dalton", "lang"=>"en", "lat"=>41.407666, "lon"=>-102.970362),
"/US/US-NE/danbury-12602" => array("type"=>"village", "pop"=>120, "name_ru"=>"Данбери", "name_en"=>"Danbury", "lang"=>"en", "lat"=>40.037755, "lon"=>-100.404420),
"/US/US-NE/dannebrog-12603" => array("type"=>"village", "pop"=>330, "name_en"=>"Dannebrog", "lang"=>"en", "lat"=>41.118596, "lon"=>-98.545246),
"/US/US-NE/davenport-12604" => array("type"=>"village", "pop"=>277, "name_ru"=>"Дэйвенпорт", "name_en"=>"Davenport", "lang"=>"en", "lat"=>40.311647, "lon"=>-97.810110),
"/US/US-NE/davey-12605" => array("type"=>"village", "pop"=>164, "name_en"=>"Davey", "lang"=>"en", "lat"=>40.982570, "lon"=>-96.668552),
"/US/US-NE/david-city-12606" => array("type"=>"city", "pop"=>2474, "name_en"=>"David City", "lang"=>"en", "lat"=>41.254531, "lon"=>-97.126143),
"/US/US-NE/dawson-12607" => array("type"=>"village", "pop"=>186, "name_ru"=>"Доусон", "name_en"=>"Dawson", "lang"=>"en", "lat"=>40.130916, "lon"=>-95.829768),
"/US/US-NE/daykin-12608" => array("type"=>"village", "pop"=>153, "name_en"=>"Daykin", "lang"=>"en", "lat"=>40.321804, "lon"=>-97.297443),
"/US/US-NE/de-witt-12613" => array("type"=>"village", "pop"=>545, "name_ru"=>"Де Витт", "name_en"=>"De Witt", "lang"=>"en", "lat"=>40.394980, "lon"=>-96.921771),
"/US/US-NE/decatur-12609" => array("type"=>"village", "pop"=>544, "name_ru"=>"Декатур", "name_en"=>"Decatur", "lang"=>"en", "lat"=>42.008100, "lon"=>-96.250721),
"/US/US-NE/denton-12610" => array("type"=>"village", "pop"=>221, "name_ru"=>"Дентон", "name_en"=>"Denton", "lang"=>"en", "lat"=>40.739527, "lon"=>-96.844614),
"/US/US-NE/deshler-12611" => array("type"=>"city", "pop"=>742, "name_en"=>"Deshler", "lang"=>"en", "lat"=>40.139002, "lon"=>-97.723229),
"/US/US-NE/deweese-12612" => array("type"=>"village", "pop"=>72, "name_en"=>"Deweese", "lang"=>"en", "lat"=>40.354290, "lon"=>-98.139151),
"/US/US-NE/dickens-12614" => array("type"=>"village", "pop"=>16, "name_ru"=>"Дикенс", "name_en"=>"Dickens", "lang"=>"en", "lat"=>40.825826, "lon"=>-100.993447),
"/US/US-NE/diller-12615" => array("type"=>"village", "pop"=>257, "name_en"=>"Diller", "lang"=>"en", "lat"=>40.108715, "lon"=>-96.937534),
"/US/US-NE/dix-12616" => array("type"=>"village", "pop"=>232, "name_ru"=>"Дикс", "name_en"=>"Dix", "lang"=>"en", "lat"=>41.234712, "lon"=>-103.486305),
"/US/US-NE/dixon-12617" => array("type"=>"village", "pop"=>107, "name_ru"=>"Диксон", "name_en"=>"Dixon", "lang"=>"en", "lat"=>42.415620, "lon"=>-96.994499),
"/US/US-NE/dodge-12618" => array("type"=>"village", "pop"=>673, "name_en"=>"Dodge", "lang"=>"en", "lat"=>41.721635, "lon"=>-96.879371),
"/US/US-NE/doniphan-12619" => array("type"=>"village", "pop"=>754, "name_ru"=>"Донифан", "name_en"=>"Doniphan", "lang"=>"en", "lat"=>40.773536, "lon"=>-98.371173),
"/US/US-NE/dorchester-12620" => array("type"=>"village", "pop"=>602, "name_ru"=>"Дорчестер", "name_en"=>"Dorchester", "lang"=>"en", "lat"=>40.647589, "lon"=>-97.114845),
"/US/US-NE/douglas-12621" => array("type"=>"village", "pop"=>224, "name_ru"=>"Дуглас", "name_en"=>"Douglas", "lang"=>"en", "lat"=>40.592304, "lon"=>-96.388487),
"/US/US-NE/du-bois-12622" => array("type"=>"village", "pop"=>138, "name_ru"=>"Дю Буа", "name_en"=>"Du Bois", "lang"=>"en", "lat"=>40.033965, "lon"=>-96.045850),
"/US/US-NE/dunbar-12623" => array("type"=>"village", "pop"=>230, "name_en"=>"Dunbar", "lang"=>"en", "lat"=>40.668672, "lon"=>-96.030311),
"/US/US-NE/duncan-12624" => array("type"=>"village", "pop"=>348, "name_ru"=>"Дункан", "name_en"=>"Duncan", "lang"=>"en", "lat"=>41.389981, "lon"=>-97.493501),
"/US/US-NE/dunning-12625" => array("type"=>"village", "pop"=>80, "name_en"=>"Dunning", "lang"=>"en", "lat"=>41.827721, "lon"=>-100.103680),
"/US/US-NE/dwight-12626" => array("type"=>"village", "pop"=>242, "name_ru"=>"Двайт", "name_en"=>"Dwight", "lang"=>"en", "lat"=>41.083002, "lon"=>-97.019142),
"/US/US-NE/eagle-12627" => array("type"=>"village", "pop"=>1132, "name_ru"=>"Игл", "name_en"=>"Eagle", "lang"=>"en", "lat"=>40.816061, "lon"=>-96.432058),
"/US/US-NE/eddyville-12628" => array("type"=>"village", "pop"=>99, "name_ru"=>"Эддивилл", "name_en"=>"Eddyville", "lang"=>"en", "lat"=>41.012712, "lon"=>-99.623256),
"/US/US-NE/edgar-12629" => array("type"=>"city", "pop"=>471, "name_en"=>"Edgar", "lang"=>"en", "lat"=>40.368390, "lon"=>-97.970483),
"/US/US-NE/edison-12630" => array("type"=>"village", "pop"=>137, "name_ru"=>"Эдисон", "name_en"=>"Edison", "lang"=>"en", "lat"=>40.277592, "lon"=>-99.775692),
"/US/US-NE/elba-12631" => array("type"=>"village", "pop"=>227, "name_ru"=>"Эльба", "name_en"=>"Elba", "lang"=>"en", "lat"=>41.284421, "lon"=>-98.569057),
"/US/US-NE/elgin-12632" => array("type"=>"city", "pop"=>649, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>41.983713, "lon"=>-98.083247),
"/US/US-NE/elk-creek-12633" => array("type"=>"village", "pop"=>102, "name_en"=>"Elk Creek", "lang"=>"en", "lat"=>40.287494, "lon"=>-96.127988),
"/US/US-NE/elkhorn-12634" => array("type"=>"city", "pop"=>267, "name_ru"=>"Элкхом", "name_en"=>"Elkhorn", "lang"=>"en", "lat"=>41.282700, "lon"=>-96.236199),
"/US/US-NE/elm-creek-12635" => array("type"=>"village", "pop"=>878, "name_en"=>"Elm Creek", "lang"=>"en", "lat"=>40.719765, "lon"=>-99.374984),
"/US/US-NE/elmwood-12636" => array("type"=>"village", "pop"=>693, "name_ru"=>"Элмвуд", "name_en"=>"Elmwood", "lang"=>"en", "lat"=>40.842197, "lon"=>-96.294251),
"/US/US-NE/elsie-12637" => array("type"=>"village", "pop"=>128, "name_ru"=>"Элси", "name_en"=>"Elsie", "lang"=>"en", "lat"=>40.846848, "lon"=>-101.388902),
"/US/US-NE/elwood-12638" => array("type"=>"village", "pop"=>679, "name_ru"=>"Элвуд", "name_en"=>"Elwood", "lang"=>"en", "lat"=>40.589632, "lon"=>-99.861625),
"/US/US-NE/elyria-12639" => array("type"=>"village", "pop"=>48, "name_en"=>"Elyria", "lang"=>"en", "lat"=>41.679687, "lon"=>-99.005259),
"/US/US-NE/emerson-12640" => array("type"=>"village", "pop"=>819, "name_ru"=>"Эмерсон", "name_en"=>"Emerson", "lang"=>"en", "lat"=>42.278249, "lon"=>-96.725649),
"/US/US-NE/emmet-12641" => array("type"=>"village", "pop"=>67, "name_ru"=>"Эммет", "name_en"=>"Emmet", "lang"=>"en", "lat"=>42.476388, "lon"=>-98.809857),
"/US/US-NE/endicott-12642" => array("type"=>"village", "pop"=>124, "name_en"=>"Endicott", "lang"=>"en", "lat"=>40.080872, "lon"=>-97.095296),
"/US/US-NE/ericson-12643" => array("type"=>"village", "pop"=>95, "name_en"=>"Ericson", "lang"=>"en", "lat"=>41.780537, "lon"=>-98.677612),
"/US/US-NE/eustis-12644" => array("type"=>"village", "pop"=>378, "name_ru"=>"Эустис", "name_en"=>"Eustis", "lang"=>"en", "lat"=>40.664271, "lon"=>-100.028617),
"/US/US-NE/ewing-12645" => array("type"=>"village", "pop"=>405, "name_ru"=>"Юинг", "name_en"=>"Ewing", "lang"=>"en", "lat"=>42.258481, "lon"=>-98.342600),
"/US/US-NE/exeter-12646" => array("type"=>"village", "pop"=>641, "name_ru"=>"Эксетер", "name_en"=>"Exeter", "lang"=>"en", "lat"=>40.644678, "lon"=>-97.449228),
"/US/US-NE/fairbury-12647" => array("type"=>"city", "pop"=>3755, "name_ru"=>"Фэрбери", "name_en"=>"Fairbury", "lang"=>"en", "lat"=>40.140915, "lon"=>-97.177232),
"/US/US-NE/fairfield-12648" => array("type"=>"city", "pop"=>406, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>40.432017, "lon"=>-98.103244),
"/US/US-NE/fairmont-12649" => array("type"=>"village", "pop"=>622, "name_ru"=>"Фэрмонт", "name_en"=>"Fairmont", "lang"=>"en", "lat"=>40.635512, "lon"=>-97.584242),
"/US/US-NE/falls-city-12650" => array("type"=>"city", "pop"=>4017, "name_en"=>"Falls City", "lang"=>"en", "lat"=>40.065312, "lon"=>-95.599552),
"/US/US-NE/farnam-12651" => array("type"=>"village", "pop"=>230, "name_en"=>"Farnam", "lang"=>"en", "lat"=>40.706286, "lon"=>-100.215001),
"/US/US-NE/farwell-12652" => array("type"=>"village", "pop"=>139, "name_ru"=>"Фаруэлл", "name_en"=>"Farwell", "lang"=>"en", "lat"=>41.215838, "lon"=>-98.627407),
"/US/US-NE/filley-12653" => array("type"=>"village", "pop"=>172, "name_en"=>"Filley", "lang"=>"en", "lat"=>40.285421, "lon"=>-96.533793),
"/US/US-NE/firth-12654" => array("type"=>"village", "pop"=>691, "name_ru"=>"Ферт", "name_en"=>"Firth", "lang"=>"en", "lat"=>40.534345, "lon"=>-96.603990),
"/US/US-NE/fordyce-12655" => array("type"=>"village", "pop"=>159, "name_ru"=>"Фордис", "name_en"=>"Fordyce", "lang"=>"en", "lat"=>42.698061, "lon"=>-97.362694),
"/US/US-NE/fort-calhoun-12656" => array("type"=>"city", "pop"=>908, "name_en"=>"Fort Calhoun", "lang"=>"en", "lat"=>41.455091, "lon"=>-96.025631),
"/US/US-NE/foster-12657" => array("type"=>"village", "pop"=>56, "name_ru"=>"Фостер", "name_en"=>"Foster", "lang"=>"en", "lat"=>42.273677, "lon"=>-97.665018),
"/US/US-NE/franklin-12658" => array("type"=>"city", "pop"=>889, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>40.096603, "lon"=>-98.951019),
"/US/US-NE/fremont-12659" => array("type"=>"city", "pop"=>25224, "name_ru"=>"Фремонт", "name_en"=>"Fremont", "lang"=>"en", "lat"=>41.440821, "lon"=>-96.492059),
"/US/US-NE/friend-12660" => array("type"=>"city", "pop"=>1151, "name_en"=>"Friend", "lang"=>"en", "lat"=>40.651070, "lon"=>-97.284049),
"/US/US-NE/fullerton-12661" => array("type"=>"city", "pop"=>1221, "name_ru"=>"Фуллертон", "name_en"=>"Fullerton", "lang"=>"en", "lat"=>41.364113, "lon"=>-97.972540),
"/US/US-NE/funk-12662" => array("type"=>"village", "pop"=>185, "name_en"=>"Funk", "lang"=>"en", "lat"=>40.462935, "lon"=>-99.249530),
"/US/US-NE/gandy-12663" => array("type"=>"village", "pop"=>28, "name_ru"=>"Гэнди", "name_en"=>"Gandy", "lang"=>"en", "lat"=>41.470124, "lon"=>-100.457427),
"/US/US-NE/garland-12664" => array("type"=>"village", "pop"=>242, "name_ru"=>"Гарленд", "name_en"=>"Garland", "lang"=>"en", "lat"=>40.944187, "lon"=>-96.985069),
"/US/US-NE/garrison-12665" => array("type"=>"village", "pop"=>63, "name_ru"=>"Гаррисон", "name_en"=>"Garrison", "lang"=>"en", "lat"=>41.175494, "lon"=>-97.163046),
"/US/US-NE/geneva-12666" => array("type"=>"city", "pop"=>2037, "name_ru"=>"Дженива", "name_en"=>"Geneva", "lang"=>"en", "lat"=>40.527631, "lon"=>-97.601552),
"/US/US-NE/genoa-12667" => array("type"=>"city", "pop"=>856, "name_ru"=>"Дженоа", "name_en"=>"Genoa", "lang"=>"en", "lat"=>41.445909, "lon"=>-97.732847),
"/US/US-NE/gering-12668" => array("type"=>"city", "pop"=>7679, "name_en"=>"Gering", "lang"=>"en", "lat"=>41.828469, "lon"=>-103.665189),
"/US/US-NE/gibbon-12669" => array("type"=>"city", "pop"=>1847, "name_ru"=>"Гиббон", "name_en"=>"Gibbon", "lang"=>"en", "lat"=>40.746503, "lon"=>-98.844744),
"/US/US-NE/gilead-12670" => array("type"=>"village", "pop"=>34, "name_en"=>"Gilead", "lang"=>"en", "lat"=>40.146501, "lon"=>-97.414950),
"/US/US-NE/giltner-12671" => array("type"=>"village", "pop"=>401, "name_en"=>"Giltner", "lang"=>"en", "lat"=>40.774335, "lon"=>-98.153949),
"/US/US-NE/glenvil-12672" => array("type"=>"village", "pop"=>286, "name_en"=>"Glenvil", "lang"=>"en", "lat"=>40.502585, "lon"=>-98.254773),
"/US/US-NE/goehner-12673" => array("type"=>"village", "pop"=>172, "name_en"=>"Goehner", "lang"=>"en", "lat"=>40.832655, "lon"=>-97.219873),
"/US/US-NE/gordon-12674" => array("type"=>"city", "pop"=>1499, "name_ru"=>"Гордон", "name_en"=>"Gordon", "lang"=>"en", "lat"=>42.806331, "lon"=>-102.203634),
"/US/US-NE/gothenburg-12675" => array("type"=>"city", "pop"=>3693, "name_en"=>"Gothenburg", "lang"=>"en", "lat"=>40.929404, "lon"=>-100.161172),
"/US/US-NE/grafton-12676" => array("type"=>"village", "pop"=>136, "name_ru"=>"Графтон", "name_en"=>"Grafton", "lang"=>"en", "lat"=>40.629625, "lon"=>-97.715062),
"/US/US-NE/grainton-12677" => array("type"=>"village", "pop"=>16, "name_en"=>"Grainton", "lang"=>"en", "lat"=>40.822373, "lon"=>-101.286525),
"/US/US-NE/grand-island-12678" => array("type"=>"city", "pop"=>45801, "name_en"=>"Grand Island", "lang"=>"en", "lat"=>40.921832, "lon"=>-98.365852),
"/US/US-NE/grant-12679" => array("type"=>"city", "pop"=>1083, "name_ru"=>"Грант", "name_en"=>"Grant", "lang"=>"en", "lat"=>40.844409, "lon"=>-101.725660),
"/US/US-NE/greeley-center-12680" => array("type"=>"village", "pop"=>436, "name_en"=>"Greeley Center", "lang"=>"en", "lat"=>41.548214, "lon"=>-98.529982),
"/US/US-NE/greenwood-12681" => array("type"=>"village", "pop"=>581, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>40.961596, "lon"=>-96.442697),
"/US/US-NE/gresham-12682" => array("type"=>"village", "pop"=>252, "name_en"=>"Gresham", "lang"=>"en", "lat"=>41.027704, "lon"=>-97.401139),
"/US/US-NE/gretna-12683" => array("type"=>"city", "pop"=>6572, "name_ru"=>"Гретна", "name_en"=>"Gretna", "lang"=>"en", "lat"=>41.140708, "lon"=>-96.243633),
"/US/US-NE/gross-12684" => array("type"=>"village", "pop"=>4, "name_en"=>"Gross", "lang"=>"en", "lat"=>42.946760, "lon"=>-98.568872),
"/US/US-NE/guide-rock-12685" => array("type"=>"village", "pop"=>203, "name_en"=>"Guide Rock", "lang"=>"en", "lat"=>40.073019, "lon"=>-98.329201),
"/US/US-NE/gurley-12686" => array("type"=>"village", "pop"=>222, "name_ru"=>"Герли", "name_en"=>"Gurley", "lang"=>"en", "lat"=>41.320555, "lon"=>-102.973879),
"/US/US-NE/hadar-12687" => array("type"=>"village", "pop"=>317, "name_en"=>"Hadar", "lang"=>"en", "lat"=>42.105822, "lon"=>-97.451474),
"/US/US-NE/haigler-12688" => array("type"=>"village", "pop"=>186, "name_en"=>"Haigler", "lang"=>"en", "lat"=>40.012414, "lon"=>-101.938522),
"/US/US-NE/hallam-12689" => array("type"=>"village", "pop"=>615, "name_en"=>"Hallam", "lang"=>"en", "lat"=>40.536736, "lon"=>-96.785876),
"/US/US-NE/halsey-12690" => array("type"=>"village", "pop"=>47, "name_en"=>"Halsey", "lang"=>"en", "lat"=>41.902752, "lon"=>-100.269074),
"/US/US-NE/hamlet-12691" => array("type"=>"village", "pop"=>51, "name_ru"=>"Хамлет", "name_en"=>"Hamlet", "lang"=>"en", "lat"=>40.384428, "lon"=>-101.234900),
"/US/US-NE/hampton-12692" => array("type"=>"village", "pop"=>418, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>40.881040, "lon"=>-97.887868),
"/US/US-NE/harbine-12693" => array("type"=>"village", "pop"=>53, "name_en"=>"Harbine", "lang"=>"en", "lat"=>40.191527, "lon"=>-96.973465),
"/US/US-NE/hardy-12694" => array("type"=>"village", "pop"=>160, "name_ru"=>"Харди", "name_en"=>"Hardy", "lang"=>"en", "lat"=>40.011625, "lon"=>-97.923306),
"/US/US-NE/harrison-12695" => array("type"=>"village", "pop"=>242, "name_ru"=>"Харрисон", "name_en"=>"Harrison", "lang"=>"en", "lat"=>42.688018, "lon"=>-103.881626),
"/US/US-NE/hartington-12696" => array("type"=>"city", "pop"=>1480, "name_en"=>"Hartington", "lang"=>"en", "lat"=>42.620022, "lon"=>-97.264998),
"/US/US-NE/harvard-12697" => array("type"=>"city", "pop"=>873, "name_ru"=>"Харвард", "name_en"=>"Harvard", "lang"=>"en", "lat"=>40.619316, "lon"=>-98.096225),
"/US/US-NE/hastings-12698" => array("type"=>"city", "pop"=>25394, "name_ru"=>"Гастингс", "name_en"=>"Hastings", "lang"=>"en", "lat"=>40.589256, "lon"=>-98.390054),
"/US/US-NE/hay-springs-12700" => array("type"=>"village", "pop"=>549, "name_en"=>"Hay Springs", "lang"=>"en", "lat"=>42.683127, "lon"=>-102.689013),
"/US/US-NE/hayes-center-12699" => array("type"=>"village", "pop"=>221, "name_en"=>"Hayes Center", "lang"=>"en", "lat"=>40.511247, "lon"=>-101.019978),
"/US/US-NE/hazard-12701" => array("type"=>"village", "pop"=>58, "name_ru"=>"Хазард", "name_en"=>"Hazard", "lang"=>"en", "lat"=>41.091152, "lon"=>-99.078059),
"/US/US-NE/heartwell-12702" => array("type"=>"village", "pop"=>77, "name_en"=>"Heartwell", "lang"=>"en", "lat"=>40.569734, "lon"=>-98.788451),
"/US/US-NE/hebron-12703" => array("type"=>"city", "pop"=>1328, "name_ru"=>"Хеброн", "name_en"=>"Hebron", "lang"=>"en", "lat"=>40.168352, "lon"=>-97.587738),
"/US/US-NE/hemingford-12704" => array("type"=>"village", "pop"=>882, "name_en"=>"Hemingford", "lang"=>"en", "lat"=>42.321150, "lon"=>-103.075742),
"/US/US-NE/henderson-12705" => array("type"=>"city", "pop"=>977, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>40.779033, "lon"=>-97.811586),
"/US/US-NE/hendley-12706" => array("type"=>"village", "pop"=>34, "name_en"=>"Hendley", "lang"=>"en", "lat"=>40.130877, "lon"=>-99.970432),
"/US/US-NE/henry-12707" => array("type"=>"village", "pop"=>160, "name_ru"=>"Хенри", "name_en"=>"Henry", "lang"=>"en", "lat"=>41.998285, "lon"=>-104.046297),
"/US/US-NE/herman-12708" => array("type"=>"village", "pop"=>298, "name_ru"=>"Херман", "name_en"=>"Herman", "lang"=>"en", "lat"=>41.673087, "lon"=>-96.216617),
"/US/US-NE/hershey-12709" => array("type"=>"village", "pop"=>699, "name_en"=>"Hershey", "lang"=>"en", "lat"=>41.158122, "lon"=>-101.001292),
"/US/US-NE/hickman-12710" => array("type"=>"village", "pop"=>1627, "name_ru"=>"Хикман", "name_en"=>"Hickman", "lang"=>"en", "lat"=>40.621488, "lon"=>-96.631485),
"/US/US-NE/hildreth-12711" => array("type"=>"village", "pop"=>320, "name_en"=>"Hildreth", "lang"=>"en", "lat"=>40.337505, "lon"=>-99.045992),
"/US/US-NE/holbrook-12712" => array("type"=>"village", "pop"=>200, "name_ru"=>"Холбрук", "name_en"=>"Holbrook", "lang"=>"en", "lat"=>40.303856, "lon"=>-100.010437),
"/US/US-NE/holdrege-12713" => array("type"=>"city", "pop"=>5172, "name_en"=>"Holdrege", "lang"=>"en", "lat"=>40.440585, "lon"=>-99.375516),
"/US/US-NE/holstein-12714" => array("type"=>"village", "pop"=>240, "name_ru"=>"Холстайн", "name_en"=>"Holstein", "lang"=>"en", "lat"=>40.464811, "lon"=>-98.651067),
"/US/US-NE/homer-12715" => array("type"=>"village", "pop"=>604, "name_ru"=>"Гомер", "name_en"=>"Homer", "lang"=>"en", "lat"=>42.321971, "lon"=>-96.491098),
"/US/US-NE/hooper-12716" => array("type"=>"city", "pop"=>788, "name_ru"=>"Хупер", "name_en"=>"Hooper", "lang"=>"en", "lat"=>41.612229, "lon"=>-96.548039),
"/US/US-NE/hordville-12717" => array("type"=>"village", "pop"=>143, "name_en"=>"Hordville", "lang"=>"en", "lat"=>41.079849, "lon"=>-97.890341),
"/US/US-NE/hoskins-12718" => array("type"=>"village", "pop"=>258, "name_en"=>"Hoskins", "lang"=>"en", "lat"=>42.113669, "lon"=>-97.304593),
"/US/US-NE/howard-city-12719" => array("type"=>"village", "pop"=>207, "name_ru"=>"Ховард Сити", "name_en"=>"Howard City", "lang"=>"en", "lat"=>41.075270, "lon"=>-98.715008),
"/US/US-NE/howells-12720" => array("type"=>"village", "pop"=>609, "name_en"=>"Howells", "lang"=>"en", "lat"=>41.723431, "lon"=>-97.004680),
"/US/US-NE/hubbard-12721" => array("type"=>"village", "pop"=>240, "name_ru"=>"Хаббард", "name_en"=>"Hubbard", "lang"=>"en", "lat"=>42.385473, "lon"=>-96.590391),
"/US/US-NE/hubbell-12722" => array("type"=>"village", "pop"=>62, "name_ru"=>"Хаббел", "name_en"=>"Hubbell", "lang"=>"en", "lat"=>40.007612, "lon"=>-97.496701),
"/US/US-NE/humboldt-12723" => array("type"=>"city", "pop"=>810, "name_ru"=>"Гумбольдт", "name_en"=>"Humboldt", "lang"=>"en", "lat"=>40.166160, "lon"=>-95.944187),
"/US/US-NE/humphrey-12724" => array("type"=>"city", "pop"=>795, "name_ru"=>"Хамфри", "name_en"=>"Humphrey", "lang"=>"en", "lat"=>41.688239, "lon"=>-97.487368),
"/US/US-NE/huntley-12725" => array("type"=>"village", "pop"=>59, "name_ru"=>"Хантли", "name_en"=>"Huntley", "lang"=>"en", "lat"=>40.210797, "lon"=>-99.291039),
"/US/US-NE/hyannis-12726" => array("type"=>"village", "pop"=>231, "name_ru"=>"Хейннис", "name_en"=>"Hyannis", "lang"=>"en", "lat"=>42.000428, "lon"=>-101.760627),
"/US/US-NE/imperial-12727" => array("type"=>"city", "pop"=>1762, "name_ru"=>"Империал", "name_en"=>"Imperial", "lang"=>"en", "lat"=>40.516226, "lon"=>-101.637609),
"/US/US-NE/indianola-12728" => array("type"=>"city", "pop"=>592, "name_ru"=>"Индианола", "name_en"=>"Indianola", "lang"=>"en", "lat"=>40.234867, "lon"=>-100.418902),
"/US/US-NE/inglewood-12729" => array("type"=>"village", "pop"=>386, "name_ru"=>"Инглвуд", "name_en"=>"Inglewood", "lang"=>"en", "lat"=>41.416725, "lon"=>-96.501613),
"/US/US-NE/inman-12730" => array("type"=>"village", "pop"=>124, "name_ru"=>"Инман", "name_en"=>"Inman", "lang"=>"en", "lat"=>42.381308, "lon"=>-98.528939),
"/US/US-NE/ithaca-12731" => array("type"=>"village", "pop"=>159, "name_ru"=>"Итака", "name_en"=>"Ithaca", "lang"=>"en", "lat"=>41.160392, "lon"=>-96.539564),
"/US/US-NE/jackson-12732" => array("type"=>"village", "pop"=>213, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>42.448794, "lon"=>-96.565896),
"/US/US-NE/jansen-12733" => array("type"=>"village", "pop"=>128, "name_en"=>"Jansen", "lang"=>"en", "lat"=>40.185709, "lon"=>-97.082394),
"/US/US-NE/johnson-12734" => array("type"=>"village", "pop"=>253, "name_ru"=>"Джонсон", "name_en"=>"Johnson", "lang"=>"en", "lat"=>40.411250, "lon"=>-95.998384),
"/US/US-NE/johnstown-12735" => array("type"=>"village", "pop"=>48, "name_ru"=>"Джонстаун", "name_en"=>"Johnstown", "lang"=>"en", "lat"=>42.572467, "lon"=>-100.055227),
"/US/US-NE/julian-12736" => array("type"=>"village", "pop"=>73, "name_ru"=>"Джулиан", "name_en"=>"Julian", "lang"=>"en", "lat"=>40.520326, "lon"=>-95.867149),
"/US/US-NE/juniata-12737" => array("type"=>"village", "pop"=>770, "name_en"=>"Juniata", "lang"=>"en", "lat"=>40.589481, "lon"=>-98.505403),
"/US/US-NE/kearney-12738" => array("type"=>"city", "pop"=>30417, "name_ru"=>"Кирни", "name_en"=>"Kearney", "lang"=>"en", "lat"=>40.700764, "lon"=>-99.084612),
"/US/US-NE/kenesaw-12739" => array("type"=>"village", "pop"=>912, "name_en"=>"Kenesaw", "lang"=>"en", "lat"=>40.619617, "lon"=>-98.658091),
"/US/US-NE/kennard-12740" => array("type"=>"village", "pop"=>390, "name_ru"=>"Кеннард", "name_en"=>"Kennard", "lang"=>"en", "lat"=>41.474259, "lon"=>-96.203411),
"/US/US-NE/kilgore-12741" => array("type"=>"village", "pop"=>90, "name_en"=>"Kilgore", "lang"=>"en", "lat"=>42.938521, "lon"=>-100.956700),
"/US/US-NE/kimball-12742" => array("type"=>"city", "pop"=>2191, "name_en"=>"Kimball", "lang"=>"en", "lat"=>41.232535, "lon"=>-103.658998),
"/US/US-NE/la-vista-12745" => array("type"=>"city", "pop"=>16643, "name_en"=>"La Vista", "lang"=>"en", "lat"=>41.184333, "lon"=>-96.039139),
"/US/US-NE/lamar-12743" => array("type"=>"village", "pop"=>17, "name_ru"=>"Ламар", "name_en"=>"Lamar", "lang"=>"en", "lat"=>40.572443, "lon"=>-101.979465),
"/US/US-NE/laurel-12744" => array("type"=>"city", "pop"=>874, "name_ru"=>"Лорел", "name_en"=>"Laurel", "lang"=>"en", "lat"=>42.427975, "lon"=>-97.094038),
"/US/US-NE/lawrence-12746" => array("type"=>"village", "pop"=>281, "name_ru"=>"Лоренс", "name_en"=>"Lawrence", "lang"=>"en", "lat"=>40.290160, "lon"=>-98.259011),
"/US/US-NE/lebanon-12747" => array("type"=>"village", "pop"=>66, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>40.049206, "lon"=>-100.275930),
"/US/US-NE/leigh-12748" => array("type"=>"village", "pop"=>411, "name_en"=>"Leigh", "lang"=>"en", "lat"=>41.704086, "lon"=>-97.236835),
"/US/US-NE/leshara-12749" => array("type"=>"village", "pop"=>104, "name_en"=>"Leshara", "lang"=>"en", "lat"=>41.330124, "lon"=>-96.428694),
"/US/US-NE/lewellen-12750" => array("type"=>"village", "pop"=>216, "name_en"=>"Lewellen", "lang"=>"en", "lat"=>41.330695, "lon"=>-102.143093),
"/US/US-NE/lewiston-12751" => array("type"=>"village", "pop"=>73, "name_ru"=>"Льюистон", "name_en"=>"Lewiston", "lang"=>"en", "lat"=>40.243363, "lon"=>-96.407450),
"/US/US-NE/lexington-12752" => array("type"=>"city", "pop"=>10126, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>40.778366, "lon"=>-99.742106),
"/US/US-NE/liberty-12753" => array("type"=>"village", "pop"=>84, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>40.085027, "lon"=>-96.483040),
"/US/US-NE/lincoln-12754" => array("type"=>"city", "pop"=>251624, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>40.816400, "lon"=>-96.688171),
"/US/US-NE/lindsay-12755" => array("type"=>"village", "pop"=>277, "name_ru"=>"Линдсей", "name_en"=>"Lindsay", "lang"=>"en", "lat"=>41.700613, "lon"=>-97.694275),
"/US/US-NE/linwood-12756" => array("type"=>"village", "pop"=>110, "name_ru"=>"Линвуд", "name_en"=>"Linwood", "lang"=>"en", "lat"=>41.411797, "lon"=>-96.932394),
"/US/US-NE/litchfield-12757" => array("type"=>"village", "pop"=>248, "name_ru"=>"Литчфилд", "name_en"=>"Litchfield", "lang"=>"en", "lat"=>41.156073, "lon"=>-99.151899),
"/US/US-NE/lodgepole-12758" => array("type"=>"village", "pop"=>349, "name_en"=>"Lodgepole", "lang"=>"en", "lat"=>41.148485, "lon"=>-102.638320),
"/US/US-NE/long-pine-12759" => array("type"=>"city", "pop"=>321, "name_en"=>"Long Pine", "lang"=>"en", "lat"=>42.534781, "lon"=>-99.702391),
"/US/US-NE/loomis-12760" => array("type"=>"village", "pop"=>363, "name_ru"=>"Лумис", "name_en"=>"Loomis", "lang"=>"en", "lat"=>40.478076, "lon"=>-99.507402),
"/US/US-NE/lorton-village-12761" => array("type"=>"village", "pop"=>61, "name_en"=>"Lorton Village", "lang"=>"en", "lat"=>40.597510, "lon"=>-96.023488),
"/US/US-NE/louisville-12762" => array("type"=>"village", "pop"=>1039, "name_ru"=>"Луисвилл", "name_en"=>"Louisville", "lang"=>"en", "lat"=>40.998793, "lon"=>-96.160303),
"/US/US-NE/loup-city-12763" => array("type"=>"city", "pop"=>883, "name_en"=>"Loup City", "lang"=>"en", "lat"=>41.276339, "lon"=>-98.967720),
"/US/US-NE/lushton-12764" => array("type"=>"village", "pop"=>32, "name_en"=>"Lushton", "lang"=>"en", "lat"=>40.723797, "lon"=>-97.724150),
"/US/US-NE/lyman-12765" => array("type"=>"village", "pop"=>400, "name_ru"=>"Лиман", "name_en"=>"Lyman", "lang"=>"en", "lat"=>41.918070, "lon"=>-104.036489),
"/US/US-NE/lynch-12766" => array("type"=>"village", "pop"=>220, "name_ru"=>"Линч", "name_en"=>"Lynch", "lang"=>"en", "lat"=>42.830904, "lon"=>-98.466850),
"/US/US-NE/lyons-12767" => array("type"=>"city", "pop"=>858, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>41.936118, "lon"=>-96.471949),
"/US/US-NE/macy-12772" => array("type"=>"cdp", "pop"=>836, "name_ru"=>"Мейси", "name_en"=>"Macy", "lang"=>"en", "lat"=>42.112647, "lon"=>-96.363028),
"/US/US-NE/madison-12773" => array("type"=>"city", "pop"=>2186, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>41.828402, "lon"=>-97.456551),
"/US/US-NE/madrid-12774" => array("type"=>"village", "pop"=>242, "name_ru"=>"Мадрид", "name_en"=>"Madrid", "lang"=>"en", "lat"=>40.849598, "lon"=>-101.542662),
"/US/US-NE/magnet-12775" => array("type"=>"village", "pop"=>69, "name_en"=>"Magnet", "lang"=>"en", "lat"=>42.457410, "lon"=>-97.469156),
"/US/US-NE/malcolm-12776" => array("type"=>"village", "pop"=>483, "name_en"=>"Malcolm", "lang"=>"en", "lat"=>40.908450, "lon"=>-96.866029),
"/US/US-NE/malmo-12777" => array("type"=>"village", "pop"=>100, "name_en"=>"Malmo", "lang"=>"en", "lat"=>41.266270, "lon"=>-96.721046),
"/US/US-NE/manley-12778" => array("type"=>"village", "pop"=>190, "name_en"=>"Manley", "lang"=>"en", "lat"=>40.918750, "lon"=>-96.165187),
"/US/US-NE/marquette-12779" => array("type"=>"village", "pop"=>266, "name_ru"=>"Маркетт", "name_en"=>"Marquette", "lang"=>"en", "lat"=>41.006034, "lon"=>-98.009485),
"/US/US-NE/marsland-12780" => array("type"=>"village", "pop"=>10, "name_en"=>"Marsland", "lang"=>"en", "lat"=>42.444572, "lon"=>-103.300905),
"/US/US-NE/martinsburg-12781" => array("type"=>"village", "pop"=>102, "name_ru"=>"Мартинсбург", "name_en"=>"Martinsburg", "lang"=>"en", "lat"=>42.507991, "lon"=>-96.831647),
"/US/US-NE/maskell-12782" => array("type"=>"village", "pop"=>67, "name_en"=>"Maskell", "lang"=>"en", "lat"=>42.690393, "lon"=>-96.981520),
"/US/US-NE/mason-city-12783" => array("type"=>"village", "pop"=>165, "name_ru"=>"Мейсон Сити", "name_en"=>"Mason City", "lang"=>"en", "lat"=>41.222477, "lon"=>-99.297977),
"/US/US-NE/maxwell-12784" => array("type"=>"village", "pop"=>319, "name_ru"=>"Максуэлл", "name_en"=>"Maxwell", "lang"=>"en", "lat"=>41.078587, "lon"=>-100.525804),
"/US/US-NE/maywood-12785" => array("type"=>"village", "pop"=>268, "name_ru"=>"Мэйвуд", "name_en"=>"Maywood", "lang"=>"en", "lat"=>40.658057, "lon"=>-100.622322),
"/US/US-NE/mccook-12768" => array("type"=>"city", "pop"=>7436, "name_ru"=>"Мак-Кук", "name_en"=>"McCook", "lang"=>"en", "lat"=>40.204112, "lon"=>-100.620341),
"/US/US-NE/mccool-junction-12769" => array("type"=>"village", "pop"=>398, "name_en"=>"McCool Junction", "lang"=>"en", "lat"=>40.744330, "lon"=>-97.595163),
"/US/US-NE/mcgrew-12770" => array("type"=>"village", "pop"=>99, "name_en"=>"McGrew", "lang"=>"en", "lat"=>41.747249, "lon"=>-103.417179),
"/US/US-NE/mclean-12771" => array("type"=>"village", "pop"=>34, "name_ru"=>"Мак-Лин", "name_en"=>"McLean", "lang"=>"en", "lat"=>42.386179, "lon"=>-97.467665),
"/US/US-NE/mead-12786" => array("type"=>"village", "pop"=>598, "name_ru"=>"Мид", "name_en"=>"Mead", "lang"=>"en", "lat"=>41.229103, "lon"=>-96.489390),
"/US/US-NE/meadow-grove-12787" => array("type"=>"village", "pop"=>283, "name_en"=>"Meadow Grove", "lang"=>"en", "lat"=>42.029157, "lon"=>-97.736155),
"/US/US-NE/melbeta-12788" => array("type"=>"village", "pop"=>138, "name_en"=>"Melbeta", "lang"=>"en", "lat"=>41.782016, "lon"=>-103.516791),
"/US/US-NE/memphis-12789" => array("type"=>"village", "pop"=>109, "name_ru"=>"Мемфис", "name_en"=>"Memphis", "lang"=>"en", "lat"=>41.094646, "lon"=>-96.432760),
"/US/US-NE/merna-12790" => array("type"=>"village", "pop"=>363, "name_en"=>"Merna", "lang"=>"en", "lat"=>41.484612, "lon"=>-99.760583),
"/US/US-NE/merriman-12791" => array("type"=>"village", "pop"=>106, "name_en"=>"Merriman", "lang"=>"en", "lat"=>42.921087, "lon"=>-101.699849),
"/US/US-NE/milford-12792" => array("type"=>"city", "pop"=>2041, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>40.772760, "lon"=>-97.052015),
"/US/US-NE/miller-12793" => array("type"=>"village", "pop"=>156, "name_ru"=>"Миллер", "name_en"=>"Miller", "lang"=>"en", "lat"=>40.926994, "lon"=>-99.390544),
"/US/US-NE/milligan-12794" => array("type"=>"village", "pop"=>281, "name_en"=>"Milligan", "lang"=>"en", "lat"=>40.499935, "lon"=>-97.388049),
"/US/US-NE/minatare-12795" => array("type"=>"city", "pop"=>779, "name_en"=>"Minatare", "lang"=>"en", "lat"=>41.811024, "lon"=>-103.502233),
"/US/US-NE/minden-12796" => array("type"=>"city", "pop"=>2802, "name_ru"=>"Минден", "name_en"=>"Minden", "lang"=>"en", "lat"=>40.500811, "lon"=>-98.950813),
"/US/US-NE/mitchell-12797" => array("type"=>"city", "pop"=>1785, "name_ru"=>"Митчелл", "name_en"=>"Mitchell", "lang"=>"en", "lat"=>41.942080, "lon"=>-103.808976),
"/US/US-NE/monowi-12798" => array("type"=>"village", "pop"=>2, "name_en"=>"Monowi", "lang"=>"en", "lat"=>42.830059, "lon"=>-98.328791),
"/US/US-NE/monroe-12799" => array("type"=>"village", "pop"=>309, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>41.473724, "lon"=>-97.600106),
"/US/US-NE/moorefield-12800" => array("type"=>"village", "pop"=>43, "name_ru"=>"Мурфилд", "name_en"=>"Moorefield", "lang"=>"en", "lat"=>40.689742, "lon"=>-100.399901),
"/US/US-NE/morrill-12801" => array("type"=>"village", "pop"=>935, "name_ru"=>"Моррил", "name_en"=>"Morrill", "lang"=>"en", "lat"=>41.964051, "lon"=>-103.923837),
"/US/US-NE/morse-bluff-12802" => array("type"=>"village", "pop"=>127, "name_en"=>"Morse Bluff", "lang"=>"en", "lat"=>41.431114, "lon"=>-96.765690),
"/US/US-NE/mullen-12803" => array("type"=>"village", "pop"=>485, "name_en"=>"Mullen", "lang"=>"en", "lat"=>42.042028, "lon"=>-101.043326),
"/US/US-NE/murdock-12804" => array("type"=>"village", "pop"=>264, "name_ru"=>"Мердок", "name_en"=>"Murdock", "lang"=>"en", "lat"=>40.926021, "lon"=>-96.280287),
"/US/US-NE/murray-12805" => array("type"=>"village", "pop"=>509, "name_ru"=>"Мюррей", "name_en"=>"Murray", "lang"=>"en", "lat"=>40.916631, "lon"=>-95.927883),
"/US/US-NE/naper-12806" => array("type"=>"village", "pop"=>90, "name_en"=>"Naper", "lang"=>"en", "lat"=>42.964170, "lon"=>-99.097139),
"/US/US-NE/naponee-12807" => array("type"=>"village", "pop"=>113, "name_en"=>"Naponee", "lang"=>"en", "lat"=>40.074005, "lon"=>-99.138077),
"/US/US-NE/nebraska-city-12808" => array("type"=>"city", "pop"=>7032, "name_en"=>"Nebraska City", "lang"=>"en", "lat"=>40.676244, "lon"=>-95.860696),
"/US/US-NE/nehawka-12809" => array("type"=>"village", "pop"=>219, "name_en"=>"Nehawka", "lang"=>"en", "lat"=>40.829812, "lon"=>-95.989287),
"/US/US-NE/neligh-12810" => array("type"=>"city", "pop"=>1460, "name_en"=>"Neligh", "lang"=>"en", "lat"=>42.130005, "lon"=>-98.028759),
"/US/US-NE/nelson-12811" => array("type"=>"city", "pop"=>509, "name_ru"=>"Нельсон", "name_en"=>"Nelson", "lang"=>"en", "lat"=>40.200926, "lon"=>-98.069041),
"/US/US-NE/nemaha-12812" => array("type"=>"village", "pop"=>183, "name_ru"=>"Немаха", "name_en"=>"Nemaha", "lang"=>"en", "lat"=>40.337822, "lon"=>-95.675851),
"/US/US-NE/nenzel-12813" => array("type"=>"village", "pop"=>12, "name_en"=>"Nenzel", "lang"=>"en", "lat"=>42.926900, "lon"=>-101.101338),
"/US/US-NE/newcastle-12814" => array("type"=>"village", "pop"=>295, "name_en"=>"Newcastle", "lang"=>"en", "lat"=>42.652007, "lon"=>-96.874946),
"/US/US-NE/newman-grove-12815" => array("type"=>"city", "pop"=>734, "name_en"=>"Newman Grove", "lang"=>"en", "lat"=>41.747354, "lon"=>-97.777066),
"/US/US-NE/newport-12816" => array("type"=>"village", "pop"=>85, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>42.600774, "lon"=>-99.327591),
"/US/US-NE/nickerson-12817" => array("type"=>"village", "pop"=>422, "name_ru"=>"Никерсон", "name_en"=>"Nickerson", "lang"=>"en", "lat"=>41.535039, "lon"=>-96.470503),
"/US/US-NE/niobrara-12818" => array("type"=>"village", "pop"=>339, "name_en"=>"Niobrara", "lang"=>"en", "lat"=>42.750316, "lon"=>-98.030260),
"/US/US-NE/nora-12819" => array("type"=>"village", "pop"=>18, "name_ru"=>"Нора", "name_en"=>"Nora", "lang"=>"en", "lat"=>40.163780, "lon"=>-97.973629),
"/US/US-NE/norfolk-12820" => array("type"=>"city", "pop"=>22940, "name_en"=>"Norfolk", "lang"=>"en", "lat"=>42.034268, "lon"=>-97.418379),
"/US/US-NE/norman-12821" => array("type"=>"village", "pop"=>47, "name_ru"=>"Норман", "name_en"=>"Norman", "lang"=>"en", "lat"=>40.479254, "lon"=>-98.791013),
"/US/US-NE/north-bend-12822" => array("type"=>"city", "pop"=>1229, "name_en"=>"North Bend", "lang"=>"en", "lat"=>41.465243, "lon"=>-96.783136),
"/US/US-NE/north-loup-12823" => array("type"=>"village", "pop"=>298, "name_en"=>"North Loup", "lang"=>"en", "lat"=>41.494858, "lon"=>-98.771929),
"/US/US-NE/north-platte-12824" => array("type"=>"city", "pop"=>24107, "name_en"=>"North Platte", "lang"=>"en", "lat"=>41.132955, "lon"=>-100.771806),
"/US/US-NE/o'neill-12836" => array("type"=>"city", "pop"=>3309, "name_en"=>"O'Neill", "lang"=>"en", "lat"=>42.460750, "lon"=>-98.646651),
"/US/US-NE/oak-12825" => array("type"=>"village", "pop"=>54, "name_en"=>"Oak", "lang"=>"en", "lat"=>40.237257, "lon"=>-97.904044),
"/US/US-NE/oakdale-12826" => array("type"=>"village", "pop"=>301, "name_ru"=>"Оукдейл", "name_en"=>"Oakdale", "lang"=>"en", "lat"=>42.069994, "lon"=>-97.966104),
"/US/US-NE/oakland-12827" => array("type"=>"city", "pop"=>1231, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>41.835140, "lon"=>-96.465771),
"/US/US-NE/obert-12828" => array("type"=>"village", "pop"=>43, "name_en"=>"Obert", "lang"=>"en", "lat"=>42.688975, "lon"=>-97.027025),
"/US/US-NE/oconto-12829" => array("type"=>"village", "pop"=>131, "name_en"=>"Oconto", "lang"=>"en", "lat"=>41.141552, "lon"=>-99.761065),
"/US/US-NE/octavia-12830" => array("type"=>"village", "pop"=>135, "name_en"=>"Octavia", "lang"=>"en", "lat"=>41.347271, "lon"=>-97.059136),
"/US/US-NE/odell-12831" => array("type"=>"village", "pop"=>328, "name_ru"=>"Оделл", "name_en"=>"Odell", "lang"=>"en", "lat"=>40.050323, "lon"=>-96.800669),
"/US/US-NE/offutt-afb-west-12832" => array("type"=>"cdp", "pop"=>10883, "name_en"=>"Offutt AFB West", "lang"=>"en", "lat"=>41.111637, "lon"=>-95.918813),
"/US/US-NE/ogallala-12833" => array("type"=>"city", "pop"=>4429, "name_en"=>"Ogallala", "lang"=>"en", "lat"=>41.130495, "lon"=>-101.720547),
"/US/US-NE/ohiowa-12834" => array("type"=>"village", "pop"=>127, "name_en"=>"Ohiowa", "lang"=>"en", "lat"=>40.414054, "lon"=>-97.451867),
"/US/US-NE/omaha-12835" => array("type"=>"city", "pop"=>438646, "name_ru"=>"Омаха", "name_en"=>"Omaha", "lang"=>"en", "lat"=>41.263900, "lon"=>-96.011745),
"/US/US-NE/ong-12837" => array("type"=>"village", "pop"=>60, "name_en"=>"Ong", "lang"=>"en", "lat"=>40.398250, "lon"=>-97.839142),
"/US/US-NE/orchard-12838" => array("type"=>"village", "pop"=>340, "name_ru"=>"Орчард", "name_en"=>"Orchard", "lang"=>"en", "lat"=>42.336032, "lon"=>-98.241220),
"/US/US-NE/ord-12839" => array("type"=>"city", "pop"=>2028, "name_en"=>"Ord", "lang"=>"en", "lat"=>41.602544, "lon"=>-98.929597),
"/US/US-NE/orleans-12840" => array("type"=>"village", "pop"=>364, "name_ru"=>"Орлеанс", "name_en"=>"Orleans", "lang"=>"en", "lat"=>40.131610, "lon"=>-99.454134),
"/US/US-NE/osceola-12841" => array("type"=>"city", "pop"=>852, "name_ru"=>"Осеола", "name_en"=>"Osceola", "lang"=>"en", "lat"=>41.177943, "lon"=>-97.548405),
"/US/US-NE/oshkosh-12842" => array("type"=>"city", "pop"=>673, "name_en"=>"Oshkosh", "lang"=>"en", "lat"=>41.408578, "lon"=>-102.344367),
"/US/US-NE/osmond-12843" => array("type"=>"city", "pop"=>698, "name_en"=>"Osmond", "lang"=>"en", "lat"=>42.358211, "lon"=>-97.598909),
"/US/US-NE/otoe-12844" => array("type"=>"village", "pop"=>210, "name_en"=>"Otoe", "lang"=>"en", "lat"=>40.724462, "lon"=>-96.120262),
"/US/US-NE/overton-12845" => array("type"=>"village", "pop"=>655, "name_en"=>"Overton", "lang"=>"en", "lat"=>40.740394, "lon"=>-99.536975),
"/US/US-NE/oxford-12846" => array("type"=>"village", "pop"=>752, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>40.253088, "lon"=>-99.632497),
"/US/US-NE/page-12847" => array("type"=>"village", "pop"=>138, "name_ru"=>"Пейдж", "name_en"=>"Page", "lang"=>"en", "lat"=>42.399040, "lon"=>-98.417407),
"/US/US-NE/palisade-12848" => array("type"=>"village", "pop"=>360, "name_ru"=>"Палисейд", "name_en"=>"Palisade", "lang"=>"en", "lat"=>40.348410, "lon"=>-101.106688),
"/US/US-NE/palmer-12849" => array("type"=>"village", "pop"=>428, "name_ru"=>"Палмер", "name_en"=>"Palmer", "lang"=>"en", "lat"=>41.227496, "lon"=>-98.218879),
"/US/US-NE/palmyra-12850" => array("type"=>"village", "pop"=>533, "name_ru"=>"Палмира", "name_en"=>"Palmyra", "lang"=>"en", "lat"=>40.705740, "lon"=>-96.391618),
"/US/US-NE/panama-12851" => array("type"=>"village", "pop"=>259, "name_ru"=>"Панама", "name_en"=>"Panama", "lang"=>"en", "lat"=>40.599566, "lon"=>-96.511178),
"/US/US-NE/papillion-12852" => array("type"=>"city", "pop"=>23739, "name_en"=>"Papillion", "lang"=>"en", "lat"=>41.157250, "lon"=>-96.040468),
"/US/US-NE/pawnee-city-12853" => array("type"=>"city", "pop"=>864, "name_en"=>"Pawnee City", "lang"=>"en", "lat"=>40.110602, "lon"=>-96.152033),
"/US/US-NE/paxton-12854" => array("type"=>"village", "pop"=>512, "name_ru"=>"Пэкстон", "name_en"=>"Paxton", "lang"=>"en", "lat"=>41.124777, "lon"=>-101.355407),
"/US/US-NE/pender-12855" => array("type"=>"village", "pop"=>1117, "name_en"=>"Pender", "lang"=>"en", "lat"=>42.110010, "lon"=>-96.710643),
"/US/US-NE/peru-12856" => array("type"=>"city", "pop"=>801, "name_ru"=>"Перу", "name_en"=>"Peru", "lang"=>"en", "lat"=>40.478659, "lon"=>-95.731021),
"/US/US-NE/petersburg-12857" => array("type"=>"village", "pop"=>319, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>41.854705, "lon"=>-98.079907),
"/US/US-NE/phillips-12858" => array("type"=>"village", "pop"=>317, "name_en"=>"Phillips", "lang"=>"en", "lat"=>40.898034, "lon"=>-98.214885),
"/US/US-NE/pickrell-12859" => array("type"=>"village", "pop"=>180, "name_en"=>"Pickrell", "lang"=>"en", "lat"=>40.377846, "lon"=>-96.728504),
"/US/US-NE/pierce-12860" => array("type"=>"city", "pop"=>1640, "name_ru"=>"Пирс", "name_en"=>"Pierce", "lang"=>"en", "lat"=>42.199473, "lon"=>-97.527717),
"/US/US-NE/pilger-12861" => array("type"=>"village", "pop"=>350, "name_en"=>"Pilger", "lang"=>"en", "lat"=>42.007732, "lon"=>-97.053956),
"/US/US-NE/plainview-12862" => array("type"=>"city", "pop"=>1200, "name_ru"=>"Плейнвью", "name_en"=>"Plainview", "lang"=>"en", "lat"=>42.353937, "lon"=>-97.786919),
"/US/US-NE/platte-center-12863" => array("type"=>"village", "pop"=>355, "name_en"=>"Platte Center", "lang"=>"en", "lat"=>41.537974, "lon"=>-97.487169),
"/US/US-NE/plattsmouth-12864" => array("type"=>"city", "pop"=>6902, "name_en"=>"Plattsmouth", "lang"=>"en", "lat"=>41.007696, "lon"=>-95.891360),
"/US/US-NE/pleasant-dale-12865" => array("type"=>"village", "pop"=>245, "name_en"=>"Pleasant Dale", "lang"=>"en", "lat"=>40.791805, "lon"=>-96.932721),
"/US/US-NE/pleasanton-12866" => array("type"=>"village", "pop"=>351, "name_ru"=>"Плезантон", "name_en"=>"Pleasanton", "lang"=>"en", "lat"=>40.970132, "lon"=>-99.086708),
"/US/US-NE/plymouth-12867" => array("type"=>"village", "pop"=>409, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>40.303425, "lon"=>-96.988060),
"/US/US-NE/polk-12868" => array("type"=>"village", "pop"=>286, "name_en"=>"Polk", "lang"=>"en", "lat"=>41.075188, "lon"=>-97.782774),
"/US/US-NE/ponca-12869" => array("type"=>"city", "pop"=>1086, "name_en"=>"Ponca", "lang"=>"en", "lat"=>42.565374, "lon"=>-96.709110),
"/US/US-NE/potter-12870" => array("type"=>"village", "pop"=>397, "name_en"=>"Potter", "lang"=>"en", "lat"=>41.218912, "lon"=>-103.314287),
"/US/US-NE/prague-12871" => array("type"=>"village", "pop"=>329, "name_en"=>"Prague", "lang"=>"en", "lat"=>41.309712, "lon"=>-96.808592),
"/US/US-NE/preston-12872" => array("type"=>"village", "pop"=>45, "name_ru"=>"Престон", "name_en"=>"Preston", "lang"=>"en", "lat"=>40.034136, "lon"=>-95.517287),
"/US/US-NE/primrose-12873" => array("type"=>"village", "pop"=>60, "name_ru"=>"Примройз", "name_en"=>"Primrose", "lang"=>"en", "lat"=>41.623625, "lon"=>-98.237210),
"/US/US-NE/prosser-12874" => array("type"=>"village", "pop"=>99, "name_en"=>"Prosser", "lang"=>"en", "lat"=>40.688695, "lon"=>-98.577311),
"/US/US-NE/ragan-12875" => array("type"=>"village", "pop"=>41, "name_en"=>"Ragan", "lang"=>"en", "lat"=>40.310627, "lon"=>-99.290058),
"/US/US-NE/ralston-12876" => array("type"=>"city", "pop"=>6095, "name_ru"=>"Ралстон", "name_en"=>"Ralston", "lang"=>"en", "lat"=>41.201350, "lon"=>-96.036693),
"/US/US-NE/randolph-12877" => array("type"=>"city", "pop"=>822, "name_ru"=>"Рэндолф", "name_en"=>"Randolph", "lang"=>"en", "lat"=>42.377337, "lon"=>-97.357225),
"/US/US-NE/ravenna-12878" => array("type"=>"city", "pop"=>1316, "name_ru"=>"Равенна", "name_en"=>"Ravenna", "lang"=>"en", "lat"=>41.026786, "lon"=>-98.913004),
"/US/US-NE/raymond-12879" => array("type"=>"village", "pop"=>224, "name_ru"=>"Реймонд", "name_en"=>"Raymond", "lang"=>"en", "lat"=>40.956553, "lon"=>-96.781479),
"/US/US-NE/red-cloud-12880" => array("type"=>"city", "pop"=>960, "name_en"=>"Red Cloud", "lang"=>"en", "lat"=>40.086104, "lon"=>-98.522316),
"/US/US-NE/republican-city-12881" => array("type"=>"village", "pop"=>180, "name_en"=>"Republican City", "lang"=>"en", "lat"=>40.100260, "lon"=>-99.222439),
"/US/US-NE/reynolds-12882" => array("type"=>"village", "pop"=>79, "name_ru"=>"Рейнолдс", "name_en"=>"Reynolds", "lang"=>"en", "lat"=>40.060387, "lon"=>-97.335873),
"/US/US-NE/richland-12883" => array("type"=>"village", "pop"=>85, "name_ru"=>"Ричленд", "name_en"=>"Richland", "lang"=>"en", "lat"=>41.437286, "lon"=>-97.214125),
"/US/US-NE/rising-city-12884" => array("type"=>"village", "pop"=>362, "name_en"=>"Rising City", "lang"=>"en", "lat"=>41.198612, "lon"=>-97.296903),
"/US/US-NE/riverdale-12885" => array("type"=>"village", "pop"=>209, "name_ru"=>"Ривердейл", "name_en"=>"Riverdale", "lang"=>"en", "lat"=>40.783602, "lon"=>-99.159986),
"/US/US-NE/riverton-12886" => array("type"=>"village", "pop"=>124, "name_ru"=>"Ривертон", "name_en"=>"Riverton", "lang"=>"en", "lat"=>40.088856, "lon"=>-98.759234),
"/US/US-NE/roca-12887" => array("type"=>"village", "pop"=>218, "name_en"=>"Roca", "lang"=>"en", "lat"=>40.658100, "lon"=>-96.661655),
"/US/US-NE/rockville-12888" => array("type"=>"village", "pop"=>98, "name_ru"=>"Роквилл", "name_en"=>"Rockville", "lang"=>"en", "lat"=>41.119268, "lon"=>-98.831089),
"/US/US-NE/rogers-12889" => array("type"=>"village", "pop"=>89, "name_ru"=>"Роджерс", "name_en"=>"Rogers", "lang"=>"en", "lat"=>41.464628, "lon"=>-96.916155),
"/US/US-NE/rosalie-12890" => array("type"=>"village", "pop"=>184, "name_en"=>"Rosalie", "lang"=>"en", "lat"=>42.056868, "lon"=>-96.512713),
"/US/US-NE/roseland-12891" => array("type"=>"village", "pop"=>254, "name_ru"=>"Роузлэнд", "name_en"=>"Roseland", "lang"=>"en", "lat"=>40.470915, "lon"=>-98.557923),
"/US/US-NE/royal-12892" => array("type"=>"village", "pop"=>66, "name_ru"=>"Ройал", "name_en"=>"Royal", "lang"=>"en", "lat"=>42.332699, "lon"=>-98.125021),
"/US/US-NE/rulo-12893" => array("type"=>"village", "pop"=>200, "name_en"=>"Rulo", "lang"=>"en", "lat"=>40.052029, "lon"=>-95.429981),
"/US/US-NE/rushville-12894" => array("type"=>"city", "pop"=>849, "name_ru"=>"Рашвилл", "name_en"=>"Rushville", "lang"=>"en", "lat"=>42.712856, "lon"=>-102.465825),
"/US/US-NE/ruskin-12895" => array("type"=>"village", "pop"=>174, "name_ru"=>"Раскин", "name_en"=>"Ruskin", "lang"=>"en", "lat"=>40.144338, "lon"=>-97.866518),
"/US/US-NE/salem-12899" => array("type"=>"village", "pop"=>118, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>40.077099, "lon"=>-95.726818),
"/US/US-NE/santee-12900" => array("type"=>"village", "pop"=>282, "name_ru"=>"Санти", "name_en"=>"Santee", "lang"=>"en", "lat"=>42.838464, "lon"=>-97.849183),
"/US/US-NE/sargent-12901" => array("type"=>"city", "pop"=>580, "name_en"=>"Sargent", "lang"=>"en", "lat"=>41.640945, "lon"=>-99.369957),
"/US/US-NE/saronville-12902" => array("type"=>"village", "pop"=>55, "name_en"=>"Saronville", "lang"=>"en", "lat"=>40.602916, "lon"=>-97.938877),
"/US/US-NE/schuyler-12903" => array("type"=>"city", "pop"=>5163, "name_en"=>"Schuyler", "lang"=>"en", "lat"=>41.448050, "lon"=>-97.059885),
"/US/US-NE/scotia-12904" => array("type"=>"village", "pop"=>261, "name_en"=>"Scotia", "lang"=>"en", "lat"=>41.468284, "lon"=>-98.702457),
"/US/US-NE/scottsbluff-12905" => array("type"=>"city", "pop"=>14785, "name_en"=>"Scottsbluff", "lang"=>"en", "lat"=>41.867229, "lon"=>-103.661609),
"/US/US-NE/scribner-12906" => array("type"=>"city", "pop"=>954, "name_en"=>"Scribner", "lang"=>"en", "lat"=>41.664541, "lon"=>-96.664850),
"/US/US-NE/seneca-12907" => array("type"=>"village", "pop"=>41, "name_ru"=>"Сенека", "name_en"=>"Seneca", "lang"=>"en", "lat"=>42.042720, "lon"=>-100.831336),
"/US/US-NE/seward-12908" => array("type"=>"city", "pop"=>6841, "name_ru"=>"Сьюард", "name_en"=>"Seward", "lang"=>"en", "lat"=>40.912110, "lon"=>-97.096661),
"/US/US-NE/shelby-12909" => array("type"=>"village", "pop"=>614, "name_ru"=>"Шелби", "name_en"=>"Shelby", "lang"=>"en", "lat"=>41.194173, "lon"=>-97.426393),
"/US/US-NE/shelton-12910" => array("type"=>"village", "pop"=>1157, "name_ru"=>"Шелтон", "name_en"=>"Shelton", "lang"=>"en", "lat"=>40.778530, "lon"=>-98.730166),
"/US/US-NE/shickley-12911" => array("type"=>"village", "pop"=>337, "name_en"=>"Shickley", "lang"=>"en", "lat"=>40.416076, "lon"=>-97.723476),
"/US/US-NE/sholes-12912" => array("type"=>"village", "pop"=>23, "name_en"=>"Sholes", "lang"=>"en", "lat"=>42.335008, "lon"=>-97.293974),
"/US/US-NE/shubert-12913" => array("type"=>"village", "pop"=>224, "name_en"=>"Shubert", "lang"=>"en", "lat"=>40.235829, "lon"=>-95.683302),
"/US/US-NE/sidney-12914" => array("type"=>"city", "pop"=>6499, "name_ru"=>"Сидни", "name_en"=>"Sidney", "lang"=>"en", "lat"=>41.133742, "lon"=>-102.970047),
"/US/US-NE/silver-creek-12915" => array("type"=>"village", "pop"=>399, "name_ru"=>"Силвер Крик", "name_en"=>"Silver Creek", "lang"=>"en", "lat"=>41.332772, "lon"=>-97.665706),
"/US/US-NE/skyline-12916" => array("type"=>"cdp", "pop"=>2563, "name_ru"=>"Скайлайн", "name_en"=>"Skyline", "lang"=>"en", "lat"=>41.249700, "lon"=>-96.249294),
"/US/US-NE/smithfield-12917" => array("type"=>"village", "pop"=>60, "name_ru"=>"Смитфилд", "name_en"=>"Smithfield", "lang"=>"en", "lat"=>40.573156, "lon"=>-99.740948),
"/US/US-NE/snyder-12918" => array("type"=>"village", "pop"=>300, "name_en"=>"Snyder", "lang"=>"en", "lat"=>41.704482, "lon"=>-96.786240),
"/US/US-NE/south-bend-12919" => array("type"=>"village", "pop"=>84, "name_ru"=>"Саут Бенд", "name_en"=>"South Bend", "lang"=>"en", "lat"=>41.002051, "lon"=>-96.246174),
"/US/US-NE/south-sioux-city-12920" => array("type"=>"city", "pop"=>11936, "name_en"=>"South Sioux City", "lang"=>"en", "lat"=>42.465951, "lon"=>-96.412100),
"/US/US-NE/spalding-12921" => array("type"=>"village", "pop"=>479, "name_en"=>"Spalding", "lang"=>"en", "lat"=>41.688387, "lon"=>-98.362047),
"/US/US-NE/spencer-12922" => array("type"=>"village", "pop"=>466, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>42.874631, "lon"=>-98.701633),
"/US/US-NE/sprague-12923" => array("type"=>"village", "pop"=>147, "name_en"=>"Sprague", "lang"=>"en", "lat"=>40.626550, "lon"=>-96.745174),
"/US/US-NE/springfield-12924" => array("type"=>"city", "pop"=>1563, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>41.083100, "lon"=>-96.132174),
"/US/US-NE/springview-12925" => array("type"=>"village", "pop"=>239, "name_en"=>"Springview", "lang"=>"en", "lat"=>42.825286, "lon"=>-99.748971),
"/US/US-NE/st.-edward-12896" => array("type"=>"city", "pop"=>688, "name_en"=>"St. Edward", "lang"=>"en", "lat"=>41.571338, "lon"=>-97.860949),
"/US/US-NE/st.-helena-12897" => array("type"=>"village", "pop"=>75, "name_ru"=>"Сент-Хэлена", "name_en"=>"St. Helena", "lang"=>"en", "lat"=>42.809929, "lon"=>-97.248873),
"/US/US-NE/st.-paul-12898" => array("type"=>"city", "pop"=>2199, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>41.213713, "lon"=>-98.459534),
"/US/US-NE/stamford-12926" => array("type"=>"village", "pop"=>179, "name_ru"=>"Стэмфорд", "name_en"=>"Stamford", "lang"=>"en", "lat"=>40.130656, "lon"=>-99.594725),
"/US/US-NE/stanton-12927" => array("type"=>"city", "pop"=>1555, "name_ru"=>"Стэнтон", "name_en"=>"Stanton", "lang"=>"en", "lat"=>41.948856, "lon"=>-97.223971),
"/US/US-NE/staplehurst-12928" => array("type"=>"village", "pop"=>249, "name_en"=>"Staplehurst", "lang"=>"en", "lat"=>40.974802, "lon"=>-97.172031),
"/US/US-NE/stapleton-12929" => array("type"=>"village", "pop"=>286, "name_ru"=>"Стейпелтон", "name_en"=>"Stapleton", "lang"=>"en", "lat"=>41.479929, "lon"=>-100.512230),
"/US/US-NE/steele-city-12930" => array("type"=>"village", "pop"=>75, "name_en"=>"Steele City", "lang"=>"en", "lat"=>40.036463, "lon"=>-97.022511),
"/US/US-NE/steinauer-12931" => array("type"=>"village", "pop"=>63, "name_en"=>"Steinauer", "lang"=>"en", "lat"=>40.206994, "lon"=>-96.233556),
"/US/US-NE/stella-12932" => array("type"=>"village", "pop"=>195, "name_en"=>"Stella", "lang"=>"en", "lat"=>40.231637, "lon"=>-95.772558),
"/US/US-NE/sterling-12933" => array("type"=>"village", "pop"=>457, "name_ru"=>"Стерлинг", "name_en"=>"Sterling", "lang"=>"en", "lat"=>40.461894, "lon"=>-96.378064),
"/US/US-NE/stockham-12934" => array("type"=>"village", "pop"=>57, "name_en"=>"Stockham", "lang"=>"en", "lat"=>40.716803, "lon"=>-97.943377),
"/US/US-NE/stockville-12935" => array("type"=>"village", "pop"=>30, "name_en"=>"Stockville", "lang"=>"en", "lat"=>40.533243, "lon"=>-100.384125),
"/US/US-NE/strang-12936" => array("type"=>"village", "pop"=>30, "name_en"=>"Strang", "lang"=>"en", "lat"=>40.415092, "lon"=>-97.587112),
"/US/US-NE/stratton-12937" => array("type"=>"village", "pop"=>356, "name_ru"=>"Стрэттон", "name_en"=>"Stratton", "lang"=>"en", "lat"=>40.151665, "lon"=>-101.227678),
"/US/US-NE/stromsburg-12938" => array("type"=>"city", "pop"=>1110, "name_en"=>"Stromsburg", "lang"=>"en", "lat"=>41.116332, "lon"=>-97.590550),
"/US/US-NE/stuart-12939" => array("type"=>"village", "pop"=>554, "name_ru"=>"Стюарт", "name_en"=>"Stuart", "lang"=>"en", "lat"=>42.600033, "lon"=>-99.140167),
"/US/US-NE/sumner-12940" => array("type"=>"village", "pop"=>239, "name_ru"=>"Саммер", "name_en"=>"Sumner", "lang"=>"en", "lat"=>40.949471, "lon"=>-99.507184),
"/US/US-NE/superior-12941" => array("type"=>"city", "pop"=>1793, "name_ru"=>"Супериор", "name_en"=>"Superior", "lang"=>"en", "lat"=>40.022391, "lon"=>-98.067745),
"/US/US-NE/surprise-12942" => array("type"=>"village", "pop"=>41, "name_ru"=>"Серпрайз", "name_en"=>"Surprise", "lang"=>"en", "lat"=>41.104390, "lon"=>-97.308627),
"/US/US-NE/sutherland-12943" => array("type"=>"village", "pop"=>1235, "name_ru"=>"Сатерленд", "name_en"=>"Sutherland", "lang"=>"en", "lat"=>41.157529, "lon"=>-101.126536),
"/US/US-NE/sutton-12944" => array("type"=>"city", "pop"=>1317, "name_ru"=>"Саттон", "name_en"=>"Sutton", "lang"=>"en", "lat"=>40.607639, "lon"=>-97.858341),
"/US/US-NE/swanton-12945" => array("type"=>"village", "pop"=>100, "name_en"=>"Swanton", "lang"=>"en", "lat"=>40.379296, "lon"=>-97.079700),
"/US/US-NE/syracuse-12946" => array("type"=>"city", "pop"=>1845, "name_ru"=>"Сиракьюс", "name_en"=>"Syracuse", "lang"=>"en", "lat"=>40.661792, "lon"=>-96.182862),
"/US/US-NE/table-rock-12947" => array("type"=>"village", "pop"=>224, "name_en"=>"Table Rock", "lang"=>"en", "lat"=>40.178748, "lon"=>-96.089233),
"/US/US-NE/talmage-12948" => array("type"=>"village", "pop"=>260, "name_en"=>"Talmage", "lang"=>"en", "lat"=>40.531767, "lon"=>-96.023395),
"/US/US-NE/tamora-12949" => array("type"=>"village", "pop"=>51, "name_en"=>"Tamora", "lang"=>"en", "lat"=>40.895753, "lon"=>-97.224074),
"/US/US-NE/tarnov-12950" => array("type"=>"village", "pop"=>63, "name_en"=>"Tarnov", "lang"=>"en", "lat"=>41.614884, "lon"=>-97.502839),
"/US/US-NE/taylor-12951" => array("type"=>"village", "pop"=>176, "name_ru"=>"Тейлор", "name_en"=>"Taylor", "lang"=>"en", "lat"=>41.769722, "lon"=>-99.381271),
"/US/US-NE/tecumseh-12952" => array("type"=>"city", "pop"=>1577, "name_ru"=>"Текамсе", "name_en"=>"Tecumseh", "lang"=>"en", "lat"=>40.370347, "lon"=>-96.188761),
"/US/US-NE/tekamah-12953" => array("type"=>"city", "pop"=>1719, "name_en"=>"Tekamah", "lang"=>"en", "lat"=>41.777438, "lon"=>-96.222704),
"/US/US-NE/terrytown-12954" => array("type"=>"village", "pop"=>1190, "name_ru"=>"Территаун", "name_en"=>"Terrytown", "lang"=>"en", "lat"=>41.850007, "lon"=>-103.668450),
"/US/US-NE/thayer-12955" => array("type"=>"village", "pop"=>67, "name_ru"=>"Тайер", "name_en"=>"Thayer", "lang"=>"en", "lat"=>40.969660, "lon"=>-97.494509),
"/US/US-NE/thedford-12956" => array("type"=>"village", "pop"=>168, "name_en"=>"Thedford", "lang"=>"en", "lat"=>41.978848, "lon"=>-100.574541),
"/US/US-NE/thurston-12957" => array("type"=>"village", "pop"=>119, "name_en"=>"Thurston", "lang"=>"en", "lat"=>42.176541, "lon"=>-96.699858),
"/US/US-NE/tilden-12958" => array("type"=>"city", "pop"=>997, "name_ru"=>"Тилден", "name_en"=>"Tilden", "lang"=>"en", "lat"=>42.044022, "lon"=>-97.831993),
"/US/US-NE/tobias-12959" => array("type"=>"village", "pop"=>149, "name_en"=>"Tobias", "lang"=>"en", "lat"=>40.418423, "lon"=>-97.336303),
"/US/US-NE/trenton-12960" => array("type"=>"village", "pop"=>456, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>40.174285, "lon"=>-101.013317),
"/US/US-NE/trumbull-12961" => array("type"=>"village", "pop"=>185, "name_ru"=>"Трамбулл", "name_en"=>"Trumbull", "lang"=>"en", "lat"=>40.680073, "lon"=>-98.272754),
"/US/US-NE/uehling-12962" => array("type"=>"village", "pop"=>262, "name_en"=>"Uehling", "lang"=>"en", "lat"=>41.734286, "lon"=>-96.505400),
"/US/US-NE/ulysses-12963" => array("type"=>"village", "pop"=>250, "name_ru"=>"Улисс", "name_en"=>"Ulysses", "lang"=>"en", "lat"=>41.072364, "lon"=>-97.202445),
"/US/US-NE/unadilla-12964" => array("type"=>"village", "pop"=>334, "name_ru"=>"Унадилла", "name_en"=>"Unadilla", "lang"=>"en", "lat"=>40.682797, "lon"=>-96.270109),
"/US/US-NE/union-12965" => array("type"=>"village", "pop"=>256, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>40.814677, "lon"=>-95.920728),
"/US/US-NE/upland-12966" => array("type"=>"village", "pop"=>154, "name_ru"=>"Апленд", "name_en"=>"Upland", "lang"=>"en", "lat"=>40.319046, "lon"=>-98.901554),
"/US/US-NE/utica-12967" => array("type"=>"village", "pop"=>810, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>40.895569, "lon"=>-97.345035),
"/US/US-NE/valentine-12968" => array("type"=>"city", "pop"=>2598, "name_en"=>"Valentine", "lang"=>"en", "lat"=>42.873681, "lon"=>-100.549897),
"/US/US-NE/valley-12969" => array("type"=>"city", "pop"=>1922, "name_ru"=>"Валли", "name_en"=>"Valley", "lang"=>"en", "lat"=>41.313033, "lon"=>-96.346033),
"/US/US-NE/valparaiso-12970" => array("type"=>"village", "pop"=>592, "name_ru"=>"Валпарайсо", "name_en"=>"Valparaiso", "lang"=>"en", "lat"=>41.079432, "lon"=>-96.831969),
"/US/US-NE/venango-12971" => array("type"=>"village", "pop"=>152, "name_en"=>"Venango", "lang"=>"en", "lat"=>40.761744, "lon"=>-102.040769),
"/US/US-NE/verdel-12972" => array("type"=>"village", "pop"=>52, "name_en"=>"Verdel", "lang"=>"en", "lat"=>42.811476, "lon"=>-98.192935),
"/US/US-NE/verdigre-12973" => array("type"=>"village", "pop"=>469, "name_en"=>"Verdigre", "lang"=>"en", "lat"=>42.597521, "lon"=>-98.035178),
"/US/US-NE/verdon-12974" => array("type"=>"village", "pop"=>188, "name_en"=>"Verdon", "lang"=>"en", "lat"=>40.148864, "lon"=>-95.710757),
"/US/US-NE/virginia-12975" => array("type"=>"village", "pop"=>66, "name_ru"=>"Вирджиния", "name_en"=>"Virginia", "lang"=>"en", "lat"=>40.245196, "lon"=>-96.498857),
"/US/US-NE/waco-12976" => array("type"=>"village", "pop"=>256, "name_ru"=>"Вако", "name_en"=>"Waco", "lang"=>"en", "lat"=>40.896554, "lon"=>-97.461227),
"/US/US-NE/wahoo-12977" => array("type"=>"city", "pop"=>3966, "name_en"=>"Wahoo", "lang"=>"en", "lat"=>41.215165, "lon"=>-96.619956),
"/US/US-NE/wakefield-12978" => array("type"=>"city", "pop"=>1356, "name_ru"=>"Уэйкфилд", "name_en"=>"Wakefield", "lang"=>"en", "lat"=>42.267553, "lon"=>-96.867059),
"/US/US-NE/wallace-12979" => array("type"=>"village", "pop"=>314, "name_ru"=>"Уоллэс", "name_en"=>"Wallace", "lang"=>"en", "lat"=>40.837082, "lon"=>-101.163483),
"/US/US-NE/walthill-12980" => array("type"=>"village", "pop"=>859, "name_en"=>"Walthill", "lang"=>"en", "lat"=>42.148715, "lon"=>-96.492796),
"/US/US-NE/washington-12981" => array("type"=>"village", "pop"=>135, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>41.398269, "lon"=>-96.208711),
"/US/US-NE/waterbury-12982" => array("type"=>"village", "pop"=>88, "name_ru"=>"Уотербери", "name_en"=>"Waterbury", "lang"=>"en", "lat"=>42.457207, "lon"=>-96.735244),
"/US/US-NE/waterloo-12983" => array("type"=>"village", "pop"=>829, "name_ru"=>"Ватрелоо", "name_en"=>"Waterloo", "lang"=>"en", "lat"=>41.287750, "lon"=>-96.285664),
"/US/US-NE/wauneta-12984" => array("type"=>"village", "pop"=>546, "name_en"=>"Wauneta", "lang"=>"en", "lat"=>40.416608, "lon"=>-101.371560),
"/US/US-NE/wausa-12985" => array("type"=>"village", "pop"=>558, "name_en"=>"Wausa", "lang"=>"en", "lat"=>42.497803, "lon"=>-97.538174),
"/US/US-NE/waverly-12986" => array("type"=>"city", "pop"=>3199, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>40.914933, "lon"=>-96.529997),
"/US/US-NE/wayne-12987" => array("type"=>"city", "pop"=>5296, "name_ru"=>"Уэйн", "name_en"=>"Wayne", "lang"=>"en", "lat"=>42.237064, "lon"=>-97.016682),
"/US/US-NE/weeping-water-12988" => array("type"=>"city", "pop"=>1085, "name_en"=>"Weeping Water", "lang"=>"en", "lat"=>40.867772, "lon"=>-96.138789),
"/US/US-NE/wellfleet-12989" => array("type"=>"village", "pop"=>76, "name_en"=>"Wellfleet", "lang"=>"en", "lat"=>40.752710, "lon"=>-100.730407),
"/US/US-NE/west-point-12992" => array("type"=>"city", "pop"=>3354, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>41.837670, "lon"=>-96.707343),
"/US/US-NE/western-12990" => array("type"=>"village", "pop"=>271, "name_en"=>"Western", "lang"=>"en", "lat"=>40.393112, "lon"=>-97.199349),
"/US/US-NE/weston-12991" => array("type"=>"village", "pop"=>295, "name_ru"=>"Вэстон", "name_en"=>"Weston", "lang"=>"en", "lat"=>41.192034, "lon"=>-96.741220),
"/US/US-NE/whitney-12993" => array("type"=>"village", "pop"=>87, "name_en"=>"Whitney", "lang"=>"en", "lat"=>42.784054, "lon"=>-103.256809),
"/US/US-NE/wilber-12994" => array("type"=>"city", "pop"=>1729, "name_en"=>"Wilber", "lang"=>"en", "lat"=>40.480932, "lon"=>-96.963319),
"/US/US-NE/wilcox-12995" => array("type"=>"village", "pop"=>334, "name_en"=>"Wilcox", "lang"=>"en", "lat"=>40.364473, "lon"=>-99.168832),
"/US/US-NE/wilsonville-12996" => array("type"=>"village", "pop"=>105, "name_ru"=>"Уилсонвилл", "name_en"=>"Wilsonville", "lang"=>"en", "lat"=>40.111535, "lon"=>-100.105891),
"/US/US-NE/winnebago-12997" => array("type"=>"village", "pop"=>902, "name_ru"=>"Уиннебаго", "name_en"=>"Winnebago", "lang"=>"en", "lat"=>42.237174, "lon"=>-96.472948),
"/US/US-NE/winnetoon-12998" => array("type"=>"village", "pop"=>62, "name_en"=>"Winnetoon", "lang"=>"en", "lat"=>42.514291, "lon"=>-97.961389),
"/US/US-NE/winside-12999" => array("type"=>"village", "pop"=>425, "name_en"=>"Winside", "lang"=>"en", "lat"=>42.178025, "lon"=>-97.174110),
"/US/US-NE/winslow-13000" => array("type"=>"village", "pop"=>98, "name_ru"=>"Уинслоу", "name_en"=>"Winslow", "lang"=>"en", "lat"=>41.608795, "lon"=>-96.504739),
"/US/US-NE/wisner-13001" => array("type"=>"city", "pop"=>1156, "name_ru"=>"Уиснер", "name_en"=>"Wisner", "lang"=>"en", "lat"=>41.986699, "lon"=>-96.913512),
"/US/US-NE/wolbach-13002" => array("type"=>"village", "pop"=>243, "name_en"=>"Wolbach", "lang"=>"en", "lat"=>41.400787, "lon"=>-98.391697),
"/US/US-NE/wood-lake-13003" => array("type"=>"village", "pop"=>65, "name_ru"=>"Вуд Лейк", "name_en"=>"Wood Lake", "lang"=>"en", "lat"=>42.639056, "lon"=>-100.236774),
"/US/US-NE/wood-river-13004" => array("type"=>"city", "pop"=>1193, "name_ru"=>"Вуд Ривер", "name_en"=>"Wood River", "lang"=>"en", "lat"=>40.821790, "lon"=>-98.597867),
"/US/US-NE/wymore-13005" => array("type"=>"city", "pop"=>1587, "name_en"=>"Wymore", "lang"=>"en", "lat"=>40.122764, "lon"=>-96.666580),
"/US/US-NE/wynot-13006" => array("type"=>"village", "pop"=>161, "name_en"=>"Wynot", "lang"=>"en", "lat"=>42.739460, "lon"=>-97.169589),
"/US/US-NE/york-13007" => array("type"=>"city", "pop"=>7870, "name_ru"=>"Йорк", "name_en"=>"York", "lang"=>"en", "lat"=>40.871229, "lon"=>-97.595572),
"/US/US-NE/yutan-13008" => array("type"=>"village", "pop"=>1207, "name_en"=>"Yutan", "lang"=>"en", "lat"=>41.243234, "lon"=>-96.397276),
"/US/US-NH" => array("type"=>"state", "pop"=>1315809, "name_ru"=>"Нью-гемпшир", "name_en"=>"New Hampshire", "lang"=>"en"),
"/US/US-NH/antrim-13065" => array("type"=>"cdp", "pop"=>2621, "name_en"=>"Antrim", "lang"=>"en", "lat"=>43.031689, "lon"=>-71.945789),
"/US/US-NH/berlin-13066" => array("type"=>"city", "pop"=>9556, "name_ru"=>"Берлин", "name_en"=>"Berlin", "lang"=>"en", "lat"=>44.489702, "lon"=>-71.255125),
"/US/US-NH/bristol-13067" => array("type"=>"cdp", "pop"=>3123, "name_ru"=>"Бристол", "name_en"=>"Bristol", "lang"=>"en", "lat"=>43.600459, "lon"=>-71.745435),
"/US/US-NH/charlestown-13068" => array("type"=>"cdp", "pop"=>4840, "name_ru"=>"Чарльзтаун", "name_en"=>"Charlestown", "lang"=>"en", "lat"=>43.235888, "lon"=>-72.423359),
"/US/US-NH/claremont-13069" => array("type"=>"city", "pop"=>12968, "name_ru"=>"Клэрмонт", "name_en"=>"Claremont", "lang"=>"en", "lat"=>43.378082, "lon"=>-72.339142),
"/US/US-NH/concord-13070" => array("type"=>"city", "pop"=>42255, "name_ru"=>"Конкорд", "name_en"=>"Concord", "lang"=>"en", "lat"=>43.231594, "lon"=>-71.560077),
"/US/US-NH/contoocook-13071" => array("type"=>"cdp", "pop"=>1334, "name_en"=>"Contoocook", "lang"=>"en", "lat"=>43.223510, "lon"=>-71.715106),
"/US/US-NH/conway-13072" => array("type"=>"cdp", "pop"=>9178, "name_ru"=>"Конвэй", "name_en"=>"Conway", "lang"=>"en", "lat"=>43.979336, "lon"=>-71.127086),
"/US/US-NH/derry-13073" => array("type"=>"cdp", "pop"=>34242, "name_en"=>"Derry", "lang"=>"en", "lat"=>42.895173, "lon"=>-71.287830),
"/US/US-NH/dover-13074" => array("type"=>"city", "pop"=>28609, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>43.189750, "lon"=>-70.883373),
"/US/US-NH/durham-13075" => array("type"=>"cdp", "pop"=>13667, "name_ru"=>"Дарем", "name_en"=>"Durham", "lang"=>"en", "lat"=>43.139454, "lon"=>-70.923227),
"/US/US-NH/east-merrimack-13076" => array("type"=>"cdp", "pop"=>3656, "name_en"=>"East Merrimack", "lang"=>"en", "lat"=>42.867996, "lon"=>-71.483907),
"/US/US-NH/enfield-13077" => array("type"=>"cdp", "pop"=>4846, "name_ru"=>"Энфилд", "name_en"=>"Enfield", "lang"=>"en", "lat"=>43.643672, "lon"=>-72.146771),
"/US/US-NH/epping-13078" => array("type"=>"cdp", "pop"=>6256, "name_en"=>"Epping", "lang"=>"en", "lat"=>43.031415, "lon"=>-71.078933),
"/US/US-NH/exeter-13079" => array("type"=>"cdp", "pop"=>14762, "name_ru"=>"Эксетер", "name_en"=>"Exeter", "lang"=>"en", "lat"=>42.975900, "lon"=>-70.948843),
"/US/US-NH/farmington-13080" => array("type"=>"cdp", "pop"=>6712, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>43.401857, "lon"=>-71.071897),
"/US/US-NH/franklin-13081" => array("type"=>"city", "pop"=>8713, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>43.451264, "lon"=>-71.668484),
"/US/US-NH/gorham-13082" => array("type"=>"cdp", "pop"=>2798, "name_ru"=>"Горам", "name_en"=>"Gorham", "lang"=>"en", "lat"=>44.392678, "lon"=>-71.186646),
"/US/US-NH/greenville-13083" => array("type"=>"cdp", "pop"=>2181, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>42.769569, "lon"=>-71.799797),
"/US/US-NH/groveton-13084" => array("type"=>"cdp", "pop"=>1255, "name_en"=>"Groveton", "lang"=>"en", "lat"=>44.599205, "lon"=>-71.518221),
"/US/US-NH/hampton-13085" => array("type"=>"cdp", "pop"=>15367, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>42.942246, "lon"=>-70.825723),
"/US/US-NH/hanover-13086" => array("type"=>"cdp", "pop"=>11068, "name_ru"=>"Хановер", "name_en"=>"Hanover", "lang"=>"en", "lat"=>43.702449, "lon"=>-72.278927),
"/US/US-NH/henniker-13087" => array("type"=>"cdp", "pop"=>5051, "name_en"=>"Henniker", "lang"=>"en", "lat"=>43.179693, "lon"=>-71.818596),
"/US/US-NH/hillsborough-13088" => array("type"=>"cdp", "pop"=>402042, "name_ru"=>"Хилсборо", "name_en"=>"Hillsborough", "lang"=>"en", "lat"=>43.114290, "lon"=>-71.899837),
"/US/US-NH/hinsdale-13089" => array("type"=>"cdp", "pop"=>4170, "name_ru"=>"Хинсдейл", "name_en"=>"Hinsdale", "lang"=>"en", "lat"=>42.788187, "lon"=>-72.488330),
"/US/US-NH/hooksett-13090" => array("type"=>"cdp", "pop"=>13783, "name_en"=>"Hooksett", "lang"=>"en", "lat"=>43.093672, "lon"=>-71.458489),
"/US/US-NH/hudson-13091" => array("type"=>"cdp", "pop"=>24771, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>42.763800, "lon"=>-71.432114),
"/US/US-NH/jaffrey-13092" => array("type"=>"cdp", "pop"=>5666, "name_en"=>"Jaffrey", "lang"=>"en", "lat"=>42.812975, "lon"=>-72.024827),
"/US/US-NH/keene-13093" => array("type"=>"city", "pop"=>22407, "name_en"=>"Keene", "lang"=>"en", "lat"=>42.950006, "lon"=>-72.297427),
"/US/US-NH/laconia-13094" => array("type"=>"city", "pop"=>17065, "name_ru"=>"Лакония", "name_en"=>"Laconia", "lang"=>"en", "lat"=>43.569450, "lon"=>-71.480316),
"/US/US-NH/lancaster-13095" => array("type"=>"cdp", "pop"=>3250, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>44.488948, "lon"=>-71.575831),
"/US/US-NH/lebanon-13096" => array("type"=>"city", "pop"=>12806, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>43.635266, "lon"=>-72.254184),
"/US/US-NH/lisbon-13097" => array("type"=>"cdp", "pop"=>1660, "name_ru"=>"Лисбон", "name_en"=>"Lisbon", "lang"=>"en", "lat"=>44.212686, "lon"=>-71.909015),
"/US/US-NH/littleton-13098" => array("type"=>"cdp", "pop"=>6182, "name_ru"=>"Литтлтон", "name_en"=>"Littleton", "lang"=>"en", "lat"=>44.311402, "lon"=>-71.763536),
"/US/US-NH/londonderry-13099" => array("type"=>"cdp", "pop"=>25016, "name_en"=>"Londonderry", "lang"=>"en", "lat"=>42.854824, "lon"=>-71.362902),
"/US/US-NH/manchester-13100" => array("type"=>"city", "pop"=>108586, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>42.983600, "lon"=>-71.444899),
"/US/US-NH/marlborough-13101" => array("type"=>"cdp", "pop"=>2070, "name_ru"=>"Марлборо", "name_en"=>"Marlborough", "lang"=>"en", "lat"=>42.903879, "lon"=>-72.210190),
"/US/US-NH/meredith-13102" => array("type"=>"cdp", "pop"=>6623, "name_en"=>"Meredith", "lang"=>"en", "lat"=>43.658040, "lon"=>-71.504256),
"/US/US-NH/milford-13103" => array("type"=>"cdp", "pop"=>14972, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>42.833967, "lon"=>-71.661782),
"/US/US-NH/nashua-13104" => array("type"=>"city", "pop"=>86576, "name_ru"=>"Нашуа", "name_en"=>"Nashua", "lang"=>"en", "lat"=>42.748600, "lon"=>-71.491848),
"/US/US-NH/newmarket-13105" => array("type"=>"cdp", "pop"=>9436, "name_en"=>"Newmarket", "lang"=>"en", "lat"=>43.075383, "lon"=>-70.940611),
"/US/US-NH/newport-13106" => array("type"=>"cdp", "pop"=>6522, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>43.371327, "lon"=>-72.173193),
"/US/US-NH/north-conway-13107" => array("type"=>"cdp", "pop"=>2032, "name_en"=>"North Conway", "lang"=>"en", "lat"=>44.053752, "lon"=>-71.126764),
"/US/US-NH/peterborough-13108" => array("type"=>"cdp", "pop"=>6185, "name_en"=>"Peterborough", "lang"=>"en", "lat"=>42.876644, "lon"=>-71.962267),
"/US/US-NH/pinardville-13109" => array("type"=>"cdp", "pop"=>4654, "name_en"=>"Pinardville", "lang"=>"en", "lat"=>43.001200, "lon"=>-71.514043),
"/US/US-NH/pittsfield-13110" => array("type"=>"cdp", "pop"=>4361, "name_ru"=>"Питтсфилд", "name_en"=>"Pittsfield", "lang"=>"en", "lat"=>43.301012, "lon"=>-71.332956),
"/US/US-NH/plymouth-13111" => array("type"=>"cdp", "pop"=>6385, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>43.738647, "lon"=>-71.698338),
"/US/US-NH/portsmouth-13112" => array("type"=>"city", "pop"=>20443, "name_ru"=>"Портсмаут", "name_en"=>"Portsmouth", "lang"=>"en", "lat"=>43.056850, "lon"=>-70.782013),
"/US/US-NH/raymond-13113" => array("type"=>"cdp", "pop"=>10234, "name_ru"=>"Реймонд", "name_en"=>"Raymond", "lang"=>"en", "lat"=>43.032958, "lon"=>-71.174728),
"/US/US-NH/rochester-13114" => array("type"=>"city", "pop"=>30654, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>43.300841, "lon"=>-70.978964),
"/US/US-NH/somersworth-13115" => array("type"=>"city", "pop"=>11967, "name_en"=>"Somersworth", "lang"=>"en", "lat"=>43.255750, "lon"=>-70.884361),
"/US/US-NH/south-hooksett-13116" => array("type"=>"cdp", "pop"=>3638, "name_en"=>"South Hooksett", "lang"=>"en", "lat"=>43.034933, "lon"=>-71.422597),
"/US/US-NH/suncook-13117" => array("type"=>"cdp", "pop"=>5214, "name_en"=>"Suncook", "lang"=>"en", "lat"=>43.138729, "lon"=>-71.452055),
"/US/US-NH/tilton-northfield-13118" => array("type"=>"cdp", "pop"=>3081, "name_en"=>"Tilton-Northfield", "lang"=>"en", "lat"=>43.443415, "lon"=>-71.593090),
"/US/US-NH/west-swanzey-13119" => array("type"=>"cdp", "pop"=>1055, "name_en"=>"West Swanzey", "lang"=>"en", "lat"=>42.870754, "lon"=>-72.322270),
"/US/US-NH/whitefield-13120" => array("type"=>"cdp", "pop"=>1955, "name_en"=>"Whitefield", "lang"=>"en", "lat"=>44.372639, "lon"=>-71.611346),
"/US/US-NH/wilton-13121" => array("type"=>"cdp", "pop"=>3921, "name_ru"=>"Уилтон", "name_en"=>"Wilton", "lang"=>"en", "lat"=>42.846937, "lon"=>-71.736832),
"/US/US-NH/winchester-13122" => array("type"=>"cdp", "pop"=>4286, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>42.776196, "lon"=>-72.384667),
"/US/US-NH/wolfeboro-13123" => array("type"=>"cdp", "pop"=>6531, "name_en"=>"Wolfeboro", "lang"=>"en", "lat"=>43.589508, "lon"=>-71.208680),
"/US/US-NH/woodsville-13124" => array("type"=>"cdp", "pop"=>1122, "name_en"=>"Woodsville", "lang"=>"en", "lat"=>44.145368, "lon"=>-72.029038),
"/US/US-NJ" => array("type"=>"state", "pop"=>8682661, "name_ru"=>"Нью-джерси", "name_en"=>"New Jersey", "lang"=>"en"),
"/US/US-NJ/absecon-13125" => array("type"=>"city", "pop"=>8389, "name_en"=>"Absecon", "lang"=>"en", "lat"=>39.425200, "lon"=>-74.495569),
"/US/US-NJ/allamuchy-panther-valley-13126" => array("type"=>"cdp", "pop"=>2764, "name_en"=>"Allamuchy-Panther Valley", "lang"=>"en", "lat"=>40.912061, "lon"=>-74.831914),
"/US/US-NJ/allendale-13127" => array("type"=>"borough", "pop"=>6599, "name_ru"=>"Аллендейл", "name_en"=>"Allendale", "lang"=>"en", "lat"=>41.033186, "lon"=>-74.132829),
"/US/US-NJ/allenhurst-13128" => array("type"=>"borough", "pop"=>699, "name_ru"=>"Алленхерст", "name_en"=>"Allenhurst", "lang"=>"en", "lat"=>40.235950, "lon"=>-74.002482),
"/US/US-NJ/allentown-13129" => array("type"=>"borough", "pop"=>1845, "name_ru"=>"Аллентаун", "name_en"=>"Allentown", "lang"=>"en", "lat"=>40.175648, "lon"=>-74.586880),
"/US/US-NJ/alloway-13130" => array("type"=>"cdp", "pop"=>3072, "name_en"=>"Alloway", "lang"=>"en", "lat"=>39.563252, "lon"=>-75.346741),
"/US/US-NJ/alpha-13131" => array("type"=>"borough", "pop"=>2386, "name_ru"=>"Альфа", "name_en"=>"Alpha", "lang"=>"en", "lat"=>40.659752, "lon"=>-75.157457),
"/US/US-NJ/alpine-13132" => array("type"=>"borough", "pop"=>2480, "name_ru"=>"Алпайн", "name_en"=>"Alpine", "lang"=>"en", "lat"=>40.958040, "lon"=>-73.920466),
"/US/US-NJ/andover-13133" => array("type"=>"borough", "pop"=>633, "name_ru"=>"Андоувер", "name_en"=>"Andover", "lang"=>"en", "lat"=>40.985873, "lon"=>-74.742741),
"/US/US-NJ/annandale-13134" => array("type"=>"cdp", "pop"=>1074, "name_ru"=>"Аннандейл", "name_en"=>"Annandale", "lang"=>"en", "lat"=>40.645009, "lon"=>-74.891130),
"/US/US-NJ/asbury-park-13135" => array("type"=>"city", "pop"=>16546, "name_en"=>"Asbury Park", "lang"=>"en", "lat"=>40.223150, "lon"=>-74.011431),
"/US/US-NJ/atlantic-city-13136" => array("type"=>"city", "pop"=>39408, "name_en"=>"Atlantic City", "lang"=>"en", "lat"=>39.377600, "lon"=>-74.451443),
"/US/US-NJ/atlantic-highlands-13137" => array("type"=>"borough", "pop"=>4602, "name_en"=>"Atlantic Highlands", "lang"=>"en", "lat"=>40.420971, "lon"=>-74.016177),
"/US/US-NJ/audubon-13138" => array("type"=>"borough", "pop"=>8883, "name_ru"=>"Одубон", "name_en"=>"Audubon", "lang"=>"en", "lat"=>39.889700, "lon"=>-75.072756),
"/US/US-NJ/audubon-park-13139" => array("type"=>"borough", "pop"=>1058, "name_ru"=>"Одубон Парк", "name_en"=>"Audubon Park", "lang"=>"en", "lat"=>39.896000, "lon"=>-75.088712),
"/US/US-NJ/avalon-13140" => array("type"=>"borough", "pop"=>2093, "name_ru"=>"Авалон", "name_en"=>"Avalon", "lang"=>"en", "lat"=>39.091650, "lon"=>-74.737018),
"/US/US-NJ/avenel-13141" => array("type"=>"cdp", "pop"=>15504, "name_en"=>"Avenel", "lang"=>"en", "lat"=>40.582467, "lon"=>-74.273716),
"/US/US-NJ/avon-by-the-sea-13142" => array("type"=>"borough", "pop"=>2201, "name_en"=>"Avon-by-the-Sea", "lang"=>"en", "lat"=>40.191000, "lon"=>-74.015242),
"/US/US-NJ/barnegat-13143" => array("type"=>"cdp", "pop"=>22255, "name_en"=>"Barnegat", "lang"=>"en", "lat"=>39.754034, "lon"=>-74.221655),
"/US/US-NJ/barnegat-light-13144" => array("type"=>"borough", "pop"=>838, "name_en"=>"Barnegat Light", "lang"=>"en", "lat"=>39.751840, "lon"=>-74.111166),
"/US/US-NJ/barrington-13145" => array("type"=>"borough", "pop"=>6939, "name_ru"=>"Бэррингтон", "name_en"=>"Barrington", "lang"=>"en", "lat"=>39.870367, "lon"=>-75.052960),
"/US/US-NJ/bay-head-13146" => array("type"=>"borough", "pop"=>1265, "name_en"=>"Bay Head", "lang"=>"en", "lat"=>40.070200, "lon"=>-74.048657),
"/US/US-NJ/bayonne-13147" => array("type"=>"city", "pop"=>57448, "name_en"=>"Bayonne", "lang"=>"en", "lat"=>40.664980, "lon"=>-74.110464),
"/US/US-NJ/beach-haven-13148" => array("type"=>"borough", "pop"=>1391, "name_en"=>"Beach Haven", "lang"=>"en", "lat"=>39.564071, "lon"=>-74.245342),
"/US/US-NJ/beach-haven-west-13149" => array("type"=>"cdp", "pop"=>4237, "name_en"=>"Beach Haven West", "lang"=>"en", "lat"=>39.670188, "lon"=>-74.233389),
"/US/US-NJ/beachwood-13150" => array("type"=>"borough", "pop"=>10835, "name_en"=>"Beachwood", "lang"=>"en", "lat"=>39.927367, "lon"=>-74.202686),
"/US/US-NJ/beattyestown-13151" => array("type"=>"cdp", "pop"=>3966, "name_en"=>"Beattyestown", "lang"=>"en", "lat"=>40.819238, "lon"=>-74.855526),
"/US/US-NJ/beckett-13152" => array("type"=>"cdp", "pop"=>3815, "name_en"=>"Beckett", "lang"=>"en", "lat"=>39.753867, "lon"=>-75.357877),
"/US/US-NJ/belleville-13153" => array("type"=>"cdp", "pop"=>33831, "name_ru"=>"Белелвилл", "name_en"=>"Belleville", "lang"=>"en", "lat"=>40.794867, "lon"=>-74.162717),
"/US/US-NJ/bellmawr-13154" => array("type"=>"borough", "pop"=>11176, "name_en"=>"Bellmawr", "lang"=>"en", "lat"=>39.865900, "lon"=>-75.094892),
"/US/US-NJ/belmar-13155" => array("type"=>"borough", "pop"=>5908, "name_en"=>"Belmar", "lang"=>"en", "lat"=>40.180363, "lon"=>-74.025518),
"/US/US-NJ/belvidere-13156" => array("type"=>"town", "pop"=>2635, "name_ru"=>"Белвидер", "name_en"=>"Belvidere", "lang"=>"en", "lat"=>40.828053, "lon"=>-75.073223),
"/US/US-NJ/bergenfield-13157" => array("type"=>"borough", "pop"=>25611, "name_en"=>"Bergenfield", "lang"=>"en", "lat"=>40.923900, "lon"=>-73.998602),
"/US/US-NJ/berkeley-heights-13158" => array("type"=>"cdp", "pop"=>13344, "name_en"=>"Berkeley Heights", "lang"=>"en", "lat"=>40.675500, "lon"=>-74.425089),
"/US/US-NJ/berlin-13159" => array("type"=>"borough", "pop"=>7922, "name_ru"=>"Берлин", "name_en"=>"Berlin", "lang"=>"en", "lat"=>39.794035, "lon"=>-74.937872),
"/US/US-NJ/bernardsville-13160" => array("type"=>"borough", "pop"=>7750, "name_en"=>"Bernardsville", "lang"=>"en", "lat"=>40.730600, "lon"=>-74.592860),
"/US/US-NJ/beverly-13161" => array("type"=>"city", "pop"=>2563, "name_ru"=>"Беверли", "name_en"=>"Beverly", "lang"=>"en", "lat"=>40.064650, "lon"=>-74.921820),
"/US/US-NJ/blackwood-13162" => array("type"=>"cdp", "pop"=>5120, "name_en"=>"Blackwood", "lang"=>"en", "lat"=>39.798433, "lon"=>-75.062927),
"/US/US-NJ/bloomfield-13163" => array("type"=>"cdp", "pop"=>43885, "name_ru"=>"Блумфилд", "name_en"=>"Bloomfield", "lang"=>"en", "lat"=>40.810600, "lon"=>-74.186881),
"/US/US-NJ/bloomingdale-13164" => array("type"=>"borough", "pop"=>7454, "name_ru"=>"Блумингдэйл", "name_en"=>"Bloomingdale", "lang"=>"en", "lat"=>41.026678, "lon"=>-74.334515),
"/US/US-NJ/bloomsbury-13165" => array("type"=>"borough", "pop"=>857, "name_en"=>"Bloomsbury", "lang"=>"en", "lat"=>40.653021, "lon"=>-75.085615),
"/US/US-NJ/bogota-13166" => array("type"=>"borough", "pop"=>7917, "name_en"=>"Bogota", "lang"=>"en", "lat"=>40.875300, "lon"=>-74.029199),
"/US/US-NJ/boonton-13167" => array("type"=>"town", "pop"=>8469, "name_en"=>"Boonton", "lang"=>"en", "lat"=>40.904067, "lon"=>-74.405474),
"/US/US-NJ/bordentown-13168" => array("type"=>"city", "pop"=>3817, "name_en"=>"Bordentown", "lang"=>"en", "lat"=>40.148593, "lon"=>-74.709481),
"/US/US-NJ/bound-brook-13169" => array("type"=>"borough", "pop"=>10365, "name_en"=>"Bound Brook", "lang"=>"en", "lat"=>40.567550, "lon"=>-74.537736),
"/US/US-NJ/bradley-beach-13170" => array("type"=>"borough", "pop"=>4845, "name_en"=>"Bradley Beach", "lang"=>"en", "lat"=>40.201750, "lon"=>-74.012149),
"/US/US-NJ/branchville-13171" => array("type"=>"borough", "pop"=>814, "name_ru"=>"Бранчвилл", "name_en"=>"Branchville", "lang"=>"en", "lat"=>41.146652, "lon"=>-74.749634),
"/US/US-NJ/brass-castle-13172" => array("type"=>"cdp", "pop"=>1419, "name_en"=>"Brass Castle", "lang"=>"en", "lat"=>40.760831, "lon"=>-75.012190),
"/US/US-NJ/brick-township-13173" => array("type"=>"cdp", "pop"=>66473, "name_en"=>"Brick Township", "lang"=>"en", "lat"=>40.057516, "lon"=>-74.110006),
"/US/US-NJ/bridgeton-13174" => array("type"=>"city", "pop"=>24742, "name_ru"=>"Бриджтон", "name_en"=>"Bridgeton", "lang"=>"en", "lat"=>39.427533, "lon"=>-75.228117),
"/US/US-NJ/brielle-13175" => array("type"=>"borough", "pop"=>4882, "name_en"=>"Brielle", "lang"=>"en", "lat"=>40.104633, "lon"=>-74.062879),
"/US/US-NJ/brigantine-13176" => array("type"=>"city", "pop"=>12647, "name_en"=>"Brigantine", "lang"=>"en", "lat"=>39.413516, "lon"=>-74.379727),
"/US/US-NJ/brooklawn-13177" => array("type"=>"borough", "pop"=>2262, "name_en"=>"Brooklawn", "lang"=>"en", "lat"=>39.878133, "lon"=>-75.121535),
"/US/US-NJ/brown-mills-13178" => array("type"=>"cdp", "pop"=>11429, "name_en"=>"Brown Mills", "lang"=>"en", "lat"=>39.973342, "lon"=>-74.568044),
"/US/US-NJ/budd-lake-13179" => array("type"=>"cdp", "pop"=>7272, "name_en"=>"Budd Lake", "lang"=>"en", "lat"=>40.873256, "lon"=>-74.737053),
"/US/US-NJ/buena-13180" => array("type"=>"borough", "pop"=>3710, "name_en"=>"Buena", "lang"=>"en", "lat"=>39.528045, "lon"=>-74.944885),
"/US/US-NJ/burlington-13181" => array("type"=>"city", "pop"=>9396, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>40.078350, "lon"=>-74.852780),
"/US/US-NJ/butler-13182" => array("type"=>"borough", "pop"=>8118, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>40.999400, "lon"=>-74.346731),
"/US/US-NJ/caldwell-13183" => array("type"=>"cdp", "pop"=>7133, "name_ru"=>"Колдуэлл", "name_en"=>"Caldwell", "lang"=>"en", "lat"=>40.838367, "lon"=>-74.277860),
"/US/US-NJ/califon-13184" => array("type"=>"borough", "pop"=>1029, "name_en"=>"Califon", "lang"=>"en", "lat"=>40.719200, "lon"=>-74.836838),
"/US/US-NJ/camden-13185" => array("type"=>"city", "pop"=>79383, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>39.935950, "lon"=>-75.108416),
"/US/US-NJ/cape-may-13186" => array("type"=>"city", "pop"=>3686, "name_en"=>"Cape May", "lang"=>"en", "lat"=>38.940064, "lon"=>-74.905479),
"/US/US-NJ/cape-may-court-house-13187" => array("type"=>"cdp", "pop"=>4426, "name_en"=>"Cape May Court House", "lang"=>"en", "lat"=>39.079901, "lon"=>-74.823890),
"/US/US-NJ/cape-may-point-13188" => array("type"=>"borough", "pop"=>223, "name_en"=>"Cape May Point", "lang"=>"en", "lat"=>38.936784, "lon"=>-74.965645),
"/US/US-NJ/carlstadt-13189" => array("type"=>"borough", "pop"=>6019, "name_en"=>"Carlstadt", "lang"=>"en", "lat"=>40.824500, "lon"=>-74.063917),
"/US/US-NJ/carneys-point-13190" => array("type"=>"cdp", "pop"=>7967, "name_en"=>"Carneys Point", "lang"=>"en", "lat"=>39.707200, "lon"=>-75.467646),
"/US/US-NJ/carteret-13191" => array("type"=>"borough", "pop"=>23367, "name_en"=>"Carteret", "lang"=>"en", "lat"=>40.583550, "lon"=>-74.228770),
"/US/US-NJ/cedar-glen-lakes-13192" => array("type"=>"cdp", "pop"=>1611, "name_en"=>"Cedar Glen Lakes", "lang"=>"en", "lat"=>39.952118, "lon"=>-74.400199),
"/US/US-NJ/cedar-glen-west-13193" => array("type"=>"cdp", "pop"=>1396, "name_en"=>"Cedar Glen West", "lang"=>"en", "lat"=>40.043450, "lon"=>-74.287611),
"/US/US-NJ/cedar-grove-13194" => array("type"=>"cdp", "pop"=>12645, "name_ru"=>"Седар Гроув", "name_en"=>"Cedar Grove", "lang"=>"en", "lat"=>40.856100, "lon"=>-74.229480),
"/US/US-NJ/chatham-13195" => array("type"=>"borough", "pop"=>8212, "name_ru"=>"Чатем", "name_en"=>"Chatham", "lang"=>"en", "lat"=>40.740350, "lon"=>-74.383486),
"/US/US-NJ/cherry-hill-13196" => array("type"=>"cdp", "pop"=>70846, "name_en"=>"Cherry Hill", "lang"=>"en", "lat"=>39.902342, "lon"=>-74.996106),
"/US/US-NJ/chesilhurst-13197" => array("type"=>"borough", "pop"=>1909, "name_en"=>"Chesilhurst", "lang"=>"en", "lat"=>39.732208, "lon"=>-74.879731),
"/US/US-NJ/chester-13198" => array("type"=>"borough", "pop"=>1678, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>40.785900, "lon"=>-74.693298),
"/US/US-NJ/cinnaminson-13199" => array("type"=>"cdp", "pop"=>15360, "name_en"=>"Cinnaminson", "lang"=>"en", "lat"=>39.999900, "lon"=>-74.991489),
"/US/US-NJ/clark-13200" => array("type"=>"cdp", "pop"=>14357, "name_ru"=>"Кларк", "name_en"=>"Clark", "lang"=>"en", "lat"=>40.621050, "lon"=>-74.311536),
"/US/US-NJ/clayton-13201" => array("type"=>"borough", "pop"=>7553, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>39.659150, "lon"=>-75.082972),
"/US/US-NJ/clearbrook-park-13202" => array("type"=>"cdp", "pop"=>2853, "name_en"=>"Clearbrook Park", "lang"=>"en", "lat"=>40.309602, "lon"=>-74.465026),
"/US/US-NJ/clementon-13203" => array("type"=>"borough", "pop"=>4886, "name_en"=>"Clementon", "lang"=>"en", "lat"=>39.804233, "lon"=>-74.985795),
"/US/US-NJ/cliffside-park-13204" => array("type"=>"borough", "pop"=>22863, "name_en"=>"Cliffside Park", "lang"=>"en", "lat"=>40.821600, "lon"=>-73.988466),
"/US/US-NJ/cliffwood-beach-13205" => array("type"=>"cdp", "pop"=>3543, "name_en"=>"Cliffwood Beach", "lang"=>"en", "lat"=>40.442600, "lon"=>-74.217999),
"/US/US-NJ/clifton-13206" => array("type"=>"city", "pop"=>78219, "name_ru"=>"Слифтон", "name_en"=>"Clifton", "lang"=>"en", "lat"=>40.863950, "lon"=>-74.157645),
"/US/US-NJ/clinton-13207" => array("type"=>"town", "pop"=>2544, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>40.635346, "lon"=>-74.912318),
"/US/US-NJ/closter-13208" => array("type"=>"borough", "pop"=>8659, "name_en"=>"Closter", "lang"=>"en", "lat"=>40.972702, "lon"=>-73.960536),
"/US/US-NJ/collings-lakes-13209" => array("type"=>"cdp", "pop"=>2046, "name_en"=>"Collings Lakes", "lang"=>"en", "lat"=>39.594720, "lon"=>-74.883161),
"/US/US-NJ/collingswood-13210" => array("type"=>"borough", "pop"=>13817, "name_en"=>"Collingswood", "lang"=>"en", "lat"=>39.916150, "lon"=>-75.075041),
"/US/US-NJ/colonia-13211" => array("type"=>"cdp", "pop"=>18238, "name_en"=>"Colonia", "lang"=>"en", "lat"=>40.592570, "lon"=>-74.315752),
"/US/US-NJ/concordia-13212" => array("type"=>"cdp", "pop"=>2683, "name_ru"=>"Конкордия", "name_en"=>"Concordia", "lang"=>"en", "lat"=>40.310858, "lon"=>-74.448635),
"/US/US-NJ/corbin-city-13213" => array("type"=>"city", "pop"=>520, "name_en"=>"Corbin City", "lang"=>"en", "lat"=>39.301216, "lon"=>-74.710658),
"/US/US-NJ/country-lake-estates-13214" => array("type"=>"cdp", "pop"=>4492, "name_en"=>"Country Lake Estates", "lang"=>"en", "lat"=>39.948721, "lon"=>-74.541095),
"/US/US-NJ/crandon-lakes-13215" => array("type"=>"cdp", "pop"=>1177, "name_en"=>"Crandon Lakes", "lang"=>"en", "lat"=>41.124201, "lon"=>-74.840298),
"/US/US-NJ/cranford-13216" => array("type"=>"cdp", "pop"=>21895, "name_en"=>"Cranford", "lang"=>"en", "lat"=>40.656200, "lon"=>-74.304124),
"/US/US-NJ/cresskill-13217" => array("type"=>"borough", "pop"=>8615, "name_en"=>"Cresskill", "lang"=>"en", "lat"=>40.939933, "lon"=>-73.958918),
"/US/US-NJ/crestwood-village-13218" => array("type"=>"cdp", "pop"=>8030, "name_en"=>"Crestwood Village", "lang"=>"en", "lat"=>39.957203, "lon"=>-74.355096),
"/US/US-NJ/dayton-13219" => array("type"=>"cdp", "pop"=>4321, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>40.382600, "lon"=>-74.510253),
"/US/US-NJ/deal-13220" => array("type"=>"borough", "pop"=>1045, "name_en"=>"Deal", "lang"=>"en", "lat"=>40.249800, "lon"=>-73.997251),
"/US/US-NJ/delanco-13221" => array("type"=>"cdp", "pop"=>4654, "name_en"=>"Delanco", "lang"=>"en", "lat"=>40.049200, "lon"=>-74.949120),
"/US/US-NJ/demarest-13222" => array("type"=>"borough", "pop"=>5130, "name_en"=>"Demarest", "lang"=>"en", "lat"=>40.956150, "lon"=>-73.957710),
"/US/US-NJ/dover-13223" => array("type"=>"town", "pop"=>17860, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>40.885950, "lon"=>-74.559685),
"/US/US-NJ/dumont-13224" => array("type"=>"borough", "pop"=>16968, "name_ru"=>"Думонт", "name_en"=>"Dumont", "lang"=>"en", "lat"=>40.945500, "lon"=>-73.992640),
"/US/US-NJ/dunellen-13225" => array("type"=>"borough", "pop"=>6993, "name_en"=>"Dunellen", "lang"=>"en", "lat"=>40.589400, "lon"=>-74.465364),
"/US/US-NJ/east-brunswick-13226" => array("type"=>"cdp", "pop"=>47279, "name_en"=>"East Brunswick", "lang"=>"en", "lat"=>40.427733, "lon"=>-74.416477),
"/US/US-NJ/east-freehold-13227" => array("type"=>"cdp", "pop"=>3842, "name_en"=>"East Freehold", "lang"=>"en", "lat"=>40.268735, "lon"=>-74.240362),
"/US/US-NJ/east-hanover-13228" => array("type"=>"cdp", "pop"=>11396, "name_en"=>"East Hanover", "lang"=>"en", "lat"=>40.819913, "lon"=>-74.365151),
"/US/US-NJ/east-newark-13229" => array("type"=>"borough", "pop"=>2138, "name_en"=>"East Newark", "lang"=>"en", "lat"=>40.750350, "lon"=>-74.163260),
"/US/US-NJ/east-orange-13230" => array("type"=>"city", "pop"=>65390, "name_en"=>"East Orange", "lang"=>"en", "lat"=>40.765950, "lon"=>-74.212109),
"/US/US-NJ/east-rutherford-13231" => array("type"=>"borough", "pop"=>10074, "name_en"=>"East Rutherford", "lang"=>"en", "lat"=>40.817450, "lon"=>-74.084365),
"/US/US-NJ/eatontown-13232" => array("type"=>"borough", "pop"=>14195, "name_en"=>"Eatontown", "lang"=>"en", "lat"=>40.291500, "lon"=>-74.055145),
"/US/US-NJ/edgewater-13233" => array("type"=>"borough", "pop"=>9635, "name_ru"=>"Эджвотер", "name_en"=>"Edgewater", "lang"=>"en", "lat"=>40.821450, "lon"=>-73.978407),
"/US/US-NJ/edgewater-park-13234" => array("type"=>"cdp", "pop"=>7692, "name_en"=>"Edgewater Park", "lang"=>"en", "lat"=>40.054600, "lon"=>-74.912497),
"/US/US-NJ/edison-13235" => array("type"=>"cdp", "pop"=>99706, "name_ru"=>"Эдисон", "name_en"=>"Edison", "lang"=>"en", "lat"=>40.528767, "lon"=>-74.369286),
"/US/US-NJ/egg-harbor-city-13236" => array("type"=>"city", "pop"=>4379, "name_en"=>"Egg Harbor City", "lang"=>"en", "lat"=>39.564286, "lon"=>-74.596319),
"/US/US-NJ/elizabeth-13237" => array("type"=>"city", "pop"=>124755, "name_ru"=>"Элизабет", "name_en"=>"Elizabeth", "lang"=>"en", "lat"=>40.666367, "lon"=>-74.194157),
"/US/US-NJ/elmer-13238" => array("type"=>"borough", "pop"=>1345, "name_ru"=>"Элмер", "name_en"=>"Elmer", "lang"=>"en", "lat"=>39.591141, "lon"=>-75.173776),
"/US/US-NJ/elmwood-park-13239" => array("type"=>"borough", "pop"=>18634, "name_ru"=>"Элмвуд Парк", "name_en"=>"Elmwood Park", "lang"=>"en", "lat"=>40.904450, "lon"=>-74.119743),
"/US/US-NJ/elwood-magnolia-13240" => array("type"=>"cdp", "pop"=>1487, "name_en"=>"Elwood-Magnolia", "lang"=>"en", "lat"=>39.572812, "lon"=>-74.718519),
"/US/US-NJ/emerson-13241" => array("type"=>"borough", "pop"=>7353, "name_ru"=>"Эмерсон", "name_en"=>"Emerson", "lang"=>"en", "lat"=>40.974998, "lon"=>-74.023688),
"/US/US-NJ/englewood-13242" => array("type"=>"city", "pop"=>29112, "name_ru"=>"Энглвуд", "name_en"=>"Englewood", "lang"=>"en", "lat"=>40.892400, "lon"=>-73.974315),
"/US/US-NJ/englewood-cliffs-13243" => array("type"=>"borough", "pop"=>5804, "name_en"=>"Englewood Cliffs", "lang"=>"en", "lat"=>40.883650, "lon"=>-73.947579),
"/US/US-NJ/englishtown-13244" => array("type"=>"borough", "pop"=>1917, "name_en"=>"Englishtown", "lang"=>"en", "lat"=>40.296500, "lon"=>-74.360584),
"/US/US-NJ/erma-13245" => array("type"=>"cdp", "pop"=>2045, "name_en"=>"Erma", "lang"=>"en", "lat"=>38.998417, "lon"=>-74.894840),
"/US/US-NJ/estell-manor-13246" => array("type"=>"city", "pop"=>1716, "name_en"=>"Estell Manor", "lang"=>"en", "lat"=>39.354244, "lon"=>-74.773716),
"/US/US-NJ/ewing-13247" => array("type"=>"cdp", "pop"=>36062, "name_ru"=>"Юинг", "name_en"=>"Ewing", "lang"=>"en", "lat"=>40.265367, "lon"=>-74.801182),
"/US/US-NJ/fair-haven-13249" => array("type"=>"borough", "pop"=>5910, "name_ru"=>"Фэр Хейвен", "name_en"=>"Fair Haven", "lang"=>"en", "lat"=>40.360450, "lon"=>-74.037787),
"/US/US-NJ/fair-lawn-13250" => array("type"=>"borough", "pop"=>30521, "name_en"=>"Fair Lawn", "lang"=>"en", "lat"=>40.935967, "lon"=>-74.118043),
"/US/US-NJ/fairfield-13248" => array("type"=>"cdp", "pop"=>6742, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>40.883600, "lon"=>-74.306262),
"/US/US-NJ/fairton-13251" => array("type"=>"cdp", "pop"=>1359, "name_en"=>"Fairton", "lang"=>"en", "lat"=>39.377900, "lon"=>-75.214941),
"/US/US-NJ/fairview-13252" => array("type"=>"borough", "pop"=>13540, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>40.817500, "lon"=>-74.001610),
"/US/US-NJ/fairview-13253" => array("type"=>"cdp", "pop"=>3853, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>40.365444, "lon"=>-74.081164),
"/US/US-NJ/fanwood-13254" => array("type"=>"borough", "pop"=>7088, "name_en"=>"Fanwood", "lang"=>"en", "lat"=>40.641750, "lon"=>-74.385739),
"/US/US-NJ/far-hills-13255" => array("type"=>"borough", "pop"=>898, "name_en"=>"Far Hills", "lang"=>"en", "lat"=>40.690350, "lon"=>-74.622077),
"/US/US-NJ/farmingdale-13256" => array("type"=>"borough", "pop"=>1569, "name_ru"=>"Фармингдейл", "name_en"=>"Farmingdale", "lang"=>"en", "lat"=>40.198500, "lon"=>-74.171425),
"/US/US-NJ/fieldsboro-13257" => array("type"=>"borough", "pop"=>562, "name_en"=>"Fieldsboro", "lang"=>"en", "lat"=>40.136169, "lon"=>-74.729902),
"/US/US-NJ/flemington-13258" => array("type"=>"borough", "pop"=>4277, "name_ru"=>"Флемингтон", "name_en"=>"Flemington", "lang"=>"en", "lat"=>40.509100, "lon"=>-74.860322),
"/US/US-NJ/florence-roebling-13259" => array("type"=>"cdp", "pop"=>8564, "name_en"=>"Florence-Roebling", "lang"=>"en", "lat"=>40.118150, "lon"=>-74.794546),
"/US/US-NJ/florham-park-13260" => array("type"=>"borough", "pop"=>12389, "name_en"=>"Florham Park", "lang"=>"en", "lat"=>40.777650, "lon"=>-74.396802),
"/US/US-NJ/folsom-13261" => array("type"=>"borough", "pop"=>1907, "name_ru"=>"Фолсом", "name_en"=>"Folsom", "lang"=>"en", "lat"=>39.592084, "lon"=>-74.847528),
"/US/US-NJ/fords-13262" => array("type"=>"cdp", "pop"=>14392, "name_en"=>"Fords", "lang"=>"en", "lat"=>40.542133, "lon"=>-74.312240),
"/US/US-NJ/forked-river-13263" => array("type"=>"cdp", "pop"=>4243, "name_en"=>"Forked River", "lang"=>"en", "lat"=>39.813823, "lon"=>-74.147487),
"/US/US-NJ/fort-dix-13264" => array("type"=>"cdp", "pop"=>10205, "name_en"=>"Fort Dix", "lang"=>"en", "lat"=>40.004875, "lon"=>-74.613124),
"/US/US-NJ/fort-lee-13265" => array("type"=>"borough", "pop"=>36275, "name_en"=>"Fort Lee", "lang"=>"en", "lat"=>40.850433, "lon"=>-73.971325),
"/US/US-NJ/franklin-13266" => array("type"=>"borough", "pop"=>5077, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>41.110500, "lon"=>-74.589848),
"/US/US-NJ/franklin-lakes-13267" => array("type"=>"borough", "pop"=>11619, "name_en"=>"Franklin Lakes", "lang"=>"en", "lat"=>41.006899, "lon"=>-74.208013),
"/US/US-NJ/freehold-13268" => array("type"=>"borough", "pop"=>11425, "name_en"=>"Freehold", "lang"=>"en", "lat"=>40.258800, "lon"=>-74.275599),
"/US/US-NJ/frenchtown-13269" => array("type"=>"borough", "pop"=>1454, "name_en"=>"Frenchtown", "lang"=>"en", "lat"=>40.525778, "lon"=>-75.056277),
"/US/US-NJ/garfield-13270" => array("type"=>"city", "pop"=>28971, "name_ru"=>"Гарфилд", "name_en"=>"Garfield", "lang"=>"en", "lat"=>40.879800, "lon"=>-74.108485),
"/US/US-NJ/garwood-13271" => array("type"=>"borough", "pop"=>4408, "name_en"=>"Garwood", "lang"=>"en", "lat"=>40.651250, "lon"=>-74.324068),
"/US/US-NJ/gibbsboro-13272" => array("type"=>"borough", "pop"=>2434, "name_en"=>"Gibbsboro", "lang"=>"en", "lat"=>39.833500, "lon"=>-74.966396),
"/US/US-NJ/gibbstown-13273" => array("type"=>"cdp", "pop"=>3902, "name_en"=>"Gibbstown", "lang"=>"en", "lat"=>39.824333, "lon"=>-75.279836),
"/US/US-NJ/gilford-park-13274" => array("type"=>"cdp", "pop"=>8668, "name_en"=>"Gilford Park", "lang"=>"en", "lat"=>39.954100, "lon"=>-74.130174),
"/US/US-NJ/glassboro-13275" => array("type"=>"borough", "pop"=>19751, "name_en"=>"Glassboro", "lang"=>"en", "lat"=>39.701300, "lon"=>-75.113264),
"/US/US-NJ/glen-gardner-13277" => array("type"=>"borough", "pop"=>1945, "name_en"=>"Glen Gardner", "lang"=>"en", "lat"=>40.698957, "lon"=>-74.941242),
"/US/US-NJ/glen-ridge-13278" => array("type"=>"cdp", "pop"=>6662, "name_ru"=>"Глен Ридж", "name_en"=>"Glen Ridge", "lang"=>"en", "lat"=>40.803100, "lon"=>-74.205029),
"/US/US-NJ/glen-rock-13279" => array("type"=>"borough", "pop"=>11131, "name_en"=>"Glen Rock", "lang"=>"en", "lat"=>40.959150, "lon"=>-74.125395),
"/US/US-NJ/glendora-13276" => array("type"=>"cdp", "pop"=>5201, "name_ru"=>"Глендора", "name_en"=>"Glendora", "lang"=>"en", "lat"=>39.840350, "lon"=>-75.069090),
"/US/US-NJ/gloucester-city-13280" => array("type"=>"city", "pop"=>11503, "name_en"=>"Gloucester City", "lang"=>"en", "lat"=>39.892367, "lon"=>-75.117800),
"/US/US-NJ/great-meadows-vienna-13281" => array("type"=>"cdp", "pop"=>1108, "name_en"=>"Great Meadows-Vienna", "lang"=>"en", "lat"=>40.870754, "lon"=>-74.889811),
"/US/US-NJ/guttenberg-13282" => array("type"=>"town", "pop"=>10528, "name_ru"=>"Гаттенберг", "name_en"=>"Guttenberg", "lang"=>"en", "lat"=>40.792650, "lon"=>-74.004865),
"/US/US-NJ/hackensack-13283" => array("type"=>"city", "pop"=>42804, "name_ru"=>"Хэкенсэк", "name_en"=>"Hackensack", "lang"=>"en", "lat"=>40.889300, "lon"=>-74.046117),
"/US/US-NJ/hackettstown-13284" => array("type"=>"town", "pop"=>9511, "name_en"=>"Hackettstown", "lang"=>"en", "lat"=>40.853800, "lon"=>-74.825358),
"/US/US-NJ/haddon-heights-13286" => array("type"=>"borough", "pop"=>7589, "name_en"=>"Haddon Heights", "lang"=>"en", "lat"=>39.878750, "lon"=>-75.065954),
"/US/US-NJ/haddonfield-13285" => array("type"=>"borough", "pop"=>11442, "name_en"=>"Haddonfield", "lang"=>"en", "lat"=>39.894967, "lon"=>-75.034710),
"/US/US-NJ/haledon-13287" => array("type"=>"borough", "pop"=>8517, "name_en"=>"Haledon", "lang"=>"en", "lat"=>40.936800, "lon"=>-74.189019),
"/US/US-NJ/hamburg-13288" => array("type"=>"borough", "pop"=>3466, "name_ru"=>"Гамбург", "name_en"=>"Hamburg", "lang"=>"en", "lat"=>41.148797, "lon"=>-74.574739),
"/US/US-NJ/hammonton-13289" => array("type"=>"town", "pop"=>13424, "name_en"=>"Hammonton", "lang"=>"en", "lat"=>39.654922, "lon"=>-74.772383),
"/US/US-NJ/hampton-13290" => array("type"=>"borough", "pop"=>1628, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>40.708684, "lon"=>-74.967777),
"/US/US-NJ/hanover-township-13291" => array("type"=>"cdp", "pop"=>11538, "name_en"=>"Hanover Township", "lang"=>"en", "lat"=>40.818917, "lon"=>-74.430479),
"/US/US-NJ/harrington-park-13292" => array("type"=>"borough", "pop"=>4869, "name_en"=>"Harrington Park", "lang"=>"en", "lat"=>40.989262, "lon"=>-73.980007),
"/US/US-NJ/harrison-13293" => array("type"=>"town", "pop"=>15201, "name_ru"=>"Харрисон", "name_en"=>"Harrison", "lang"=>"en", "lat"=>40.743500, "lon"=>-74.153399),
"/US/US-NJ/harvey-cedars-13294" => array("type"=>"borough", "pop"=>393, "name_en"=>"Harvey Cedars", "lang"=>"en", "lat"=>39.699580, "lon"=>-74.142143),
"/US/US-NJ/hasbrouck-heights-13295" => array("type"=>"borough", "pop"=>11410, "name_en"=>"Hasbrouck Heights", "lang"=>"en", "lat"=>40.860800, "lon"=>-74.074184),
"/US/US-NJ/haworth-13296" => array("type"=>"borough", "pop"=>3413, "name_en"=>"Haworth", "lang"=>"en", "lat"=>40.961719, "lon"=>-73.998869),
"/US/US-NJ/hawthorne-13297" => array("type"=>"borough", "pop"=>17997, "name_ru"=>"Хоуторн", "name_en"=>"Hawthorne", "lang"=>"en", "lat"=>40.956750, "lon"=>-74.159045),
"/US/US-NJ/heathcote-13298" => array("type"=>"cdp", "pop"=>3112, "name_en"=>"Heathcote", "lang"=>"en", "lat"=>40.388563, "lon"=>-74.576162),
"/US/US-NJ/helmetta-13299" => array("type"=>"borough", "pop"=>2010, "name_en"=>"Helmetta", "lang"=>"en", "lat"=>40.376617, "lon"=>-74.424419),
"/US/US-NJ/hi-nella-13307" => array("type"=>"borough", "pop"=>998, "name_en"=>"Hi-Nella", "lang"=>"en", "lat"=>39.835200, "lon"=>-75.021993),
"/US/US-NJ/high-bridge-13300" => array("type"=>"borough", "pop"=>3674, "name_en"=>"High Bridge", "lang"=>"en", "lat"=>40.668368, "lon"=>-74.895721),
"/US/US-NJ/highland-lake-13301" => array("type"=>"cdp", "pop"=>4550, "name_ru"=>"Хайленд Лейк", "name_en"=>"Highland Lake", "lang"=>"en", "lat"=>41.176633, "lon"=>-74.456950),
"/US/US-NJ/highland-park-13302" => array("type"=>"borough", "pop"=>14245, "name_ru"=>"Хайленд Парк", "name_en"=>"Highland Park", "lang"=>"en", "lat"=>40.500350, "lon"=>-74.428449),
"/US/US-NJ/highlands-13303" => array("type"=>"borough", "pop"=>5272, "name_ru"=>"Хайлендс", "name_en"=>"Highlands", "lang"=>"en", "lat"=>40.404350, "lon"=>-73.990838),
"/US/US-NJ/hightstown-13304" => array("type"=>"borough", "pop"=>5295, "name_en"=>"Hightstown", "lang"=>"en", "lat"=>40.267833, "lon"=>-74.525423),
"/US/US-NJ/hillsdale-13305" => array("type"=>"borough", "pop"=>9848, "name_ru"=>"Хиллсдейл", "name_en"=>"Hillsdale", "lang"=>"en", "lat"=>41.008218, "lon"=>-74.043657),
"/US/US-NJ/hillside-13306" => array("type"=>"cdp", "pop"=>21171, "name_ru"=>"Хиллсайд", "name_en"=>"Hillside", "lang"=>"en", "lat"=>40.696150, "lon"=>-74.230260),
"/US/US-NJ/ho-ho-kus-13309" => array("type"=>"borough", "pop"=>4002, "name_en"=>"Ho-Ho-Kus", "lang"=>"en", "lat"=>40.999615, "lon"=>-74.097171),
"/US/US-NJ/hoboken-13308" => array("type"=>"city", "pop"=>40577, "name_ru"=>"Хобокен", "name_en"=>"Hoboken", "lang"=>"en", "lat"=>40.742550, "lon"=>-74.028832),
"/US/US-NJ/holiday-city-berkeley-13310" => array("type"=>"cdp", "pop"=>14293, "name_en"=>"Holiday City-Berkeley", "lang"=>"en", "lat"=>39.959592, "lon"=>-74.277712),
"/US/US-NJ/holiday-city-dover-13311" => array("type"=>"cdp", "pop"=>2391, "name_en"=>"Holiday City-Dover", "lang"=>"en", "lat"=>40.021200, "lon"=>-74.167162),
"/US/US-NJ/holiday-city-south-13312" => array("type"=>"cdp", "pop"=>5452, "name_en"=>"Holiday City South", "lang"=>"en", "lat"=>39.948683, "lon"=>-74.236633),
"/US/US-NJ/holiday-heights-13313" => array("type"=>"cdp", "pop"=>703, "name_en"=>"Holiday Heights", "lang"=>"en", "lat"=>39.945828, "lon"=>-74.254348),
"/US/US-NJ/hopatcong-13314" => array("type"=>"borough", "pop"=>15491, "name_en"=>"Hopatcong", "lang"=>"en", "lat"=>40.956090, "lon"=>-74.657718),
"/US/US-NJ/hopewell-13315" => array("type"=>"borough", "pop"=>1988, "name_ru"=>"Хоупуэлл", "name_en"=>"Hopewell", "lang"=>"en", "lat"=>40.388900, "lon"=>-74.764404),
"/US/US-NJ/interlaken-13316" => array("type"=>"borough", "pop"=>877, "name_ru"=>"Интерлакен", "name_en"=>"Interlaken", "lang"=>"en", "lat"=>40.234350, "lon"=>-74.015936),
"/US/US-NJ/irvington-13317" => array("type"=>"cdp", "pop"=>56299, "name_ru"=>"Ирвингтон", "name_en"=>"Irvington", "lang"=>"en", "lat"=>40.724517, "lon"=>-74.232526),
"/US/US-NJ/iselin-13318" => array("type"=>"cdp", "pop"=>16141, "name_en"=>"Iselin", "lang"=>"en", "lat"=>40.570208, "lon"=>-74.316676),
"/US/US-NJ/island-heights-13319" => array("type"=>"borough", "pop"=>1871, "name_en"=>"Island Heights", "lang"=>"en", "lat"=>39.942300, "lon"=>-74.149579),
"/US/US-NJ/jamesburg-13320" => array("type"=>"borough", "pop"=>6367, "name_en"=>"Jamesburg", "lang"=>"en", "lat"=>40.349000, "lon"=>-74.440169),
"/US/US-NJ/jersey-city-13321" => array("type"=>"city", "pop"=>241114, "name_en"=>"Jersey City", "lang"=>"en", "lat"=>40.711300, "lon"=>-74.064962),
"/US/US-NJ/keansburg-13322" => array("type"=>"borough", "pop"=>10550, "name_en"=>"Keansburg", "lang"=>"en", "lat"=>40.467125, "lon"=>-74.161872),
"/US/US-NJ/kearny-13323" => array("type"=>"town", "pop"=>36746, "name_ru"=>"Кирни", "name_en"=>"Kearny", "lang"=>"en", "lat"=>40.751333, "lon"=>-74.119713),
"/US/US-NJ/kendall-park-13324" => array("type"=>"cdp", "pop"=>7127, "name_en"=>"Kendall Park", "lang"=>"en", "lat"=>40.413400, "lon"=>-74.562599),
"/US/US-NJ/kenilworth-13325" => array("type"=>"borough", "pop"=>7619, "name_ru"=>"Кенилворт", "name_en"=>"Kenilworth", "lang"=>"en", "lat"=>40.678933, "lon"=>-74.288949),
"/US/US-NJ/keyport-13326" => array("type"=>"borough", "pop"=>7476, "name_en"=>"Keyport", "lang"=>"en", "lat"=>40.431950, "lon"=>-74.201533),
"/US/US-NJ/kingston-13327" => array("type"=>"cdp", "pop"=>1047, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>40.375272, "lon"=>-74.608164),
"/US/US-NJ/kinnelon-13328" => array("type"=>"borough", "pop"=>9582, "name_en"=>"Kinnelon", "lang"=>"en", "lat"=>40.984851, "lon"=>-74.386708),
"/US/US-NJ/lake-mohawk-13330" => array("type"=>"cdp", "pop"=>8930, "name_en"=>"Lake Mohawk", "lang"=>"en", "lat"=>41.018394, "lon"=>-74.660428),
"/US/US-NJ/lake-telemark-13331" => array("type"=>"cdp", "pop"=>1121, "name_en"=>"Lake Telemark", "lang"=>"en", "lat"=>40.961333, "lon"=>-74.496811),
"/US/US-NJ/lakehurst-13329" => array("type"=>"borough", "pop"=>2707, "name_en"=>"Lakehurst", "lang"=>"en", "lat"=>40.012950, "lon"=>-74.320799),
"/US/US-NJ/lakewood-13332" => array("type"=>"cdp", "pop"=>70983, "name_ru"=>"Лейквуд", "name_en"=>"Lakewood", "lang"=>"en", "lat"=>40.094800, "lon"=>-74.211912),
"/US/US-NJ/lambertville-13333" => array("type"=>"city", "pop"=>3741, "name_ru"=>"Ламбертвилл", "name_en"=>"Lambertville", "lang"=>"en", "lat"=>40.368022, "lon"=>-74.943322),
"/US/US-NJ/laurel-springs-13334" => array("type"=>"borough", "pop"=>1903, "name_en"=>"Laurel Springs", "lang"=>"en", "lat"=>39.820150, "lon"=>-75.005914),
"/US/US-NJ/laurence-harbor-13335" => array("type"=>"cdp", "pop"=>6361, "name_en"=>"Laurence Harbor", "lang"=>"en", "lat"=>40.447850, "lon"=>-74.249533),
"/US/US-NJ/lavallette-13336" => array("type"=>"borough", "pop"=>2754, "name_en"=>"Lavallette", "lang"=>"en", "lat"=>39.969100, "lon"=>-74.072423),
"/US/US-NJ/lawnside-13337" => array("type"=>"borough", "pop"=>2842, "name_en"=>"Lawnside", "lang"=>"en", "lat"=>39.865600, "lon"=>-75.030631),
"/US/US-NJ/lawrenceville-13338" => array("type"=>"cdp", "pop"=>6446, "name_ru"=>"Лоренсвилл", "name_en"=>"Lawrenceville", "lang"=>"en", "lat"=>40.301400, "lon"=>-74.730154),
"/US/US-NJ/lebanon-13339" => array("type"=>"borough", "pop"=>1955, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>40.643943, "lon"=>-74.835120),
"/US/US-NJ/leisure-knoll-13340" => array("type"=>"cdp", "pop"=>2707, "name_en"=>"Leisure Knoll", "lang"=>"en", "lat"=>40.018950, "lon"=>-74.292631),
"/US/US-NJ/leisure-village-13342" => array("type"=>"cdp", "pop"=>4295, "name_en"=>"Leisure Village", "lang"=>"en", "lat"=>40.044933, "lon"=>-74.182491),
"/US/US-NJ/leisure-village-east-13343" => array("type"=>"cdp", "pop"=>1989, "name_en"=>"Leisure Village East", "lang"=>"en", "lat"=>40.041476, "lon"=>-74.167337),
"/US/US-NJ/leisure-village-west-pine-lake-park-13344" => array("type"=>"cdp", "pop"=>10139, "name_en"=>"Leisure Village West-Pine Lake Park", "lang"=>"en", "lat"=>40.003850, "lon"=>-74.266462),
"/US/US-NJ/leisuretowne-13341" => array("type"=>"cdp", "pop"=>2552, "name_en"=>"Leisuretowne", "lang"=>"en", "lat"=>39.900955, "lon"=>-74.704462),
"/US/US-NJ/leonardo-13345" => array("type"=>"cdp", "pop"=>3788, "name_en"=>"Leonardo", "lang"=>"en", "lat"=>40.419200, "lon"=>-74.060645),
"/US/US-NJ/leonia-13346" => array("type"=>"borough", "pop"=>8597, "name_en"=>"Leonia", "lang"=>"en", "lat"=>40.863107, "lon"=>-73.991552),
"/US/US-NJ/lincoln-park-13347" => array("type"=>"borough", "pop"=>10635, "name_ru"=>"Линкольн Парк", "name_en"=>"Lincoln Park", "lang"=>"en", "lat"=>40.923950, "lon"=>-74.303940),
"/US/US-NJ/lincroft-13348" => array("type"=>"cdp", "pop"=>6193, "name_en"=>"Lincroft", "lang"=>"en", "lat"=>40.341400, "lon"=>-74.125237),
"/US/US-NJ/linden-13349" => array("type"=>"city", "pop"=>39162, "name_ru"=>"Линден", "name_en"=>"Linden", "lang"=>"en", "lat"=>40.625050, "lon"=>-74.238449),
"/US/US-NJ/lindenwold-13350" => array("type"=>"borough", "pop"=>17485, "name_en"=>"Lindenwold", "lang"=>"en", "lat"=>39.815850, "lon"=>-74.990157),
"/US/US-NJ/linwood-13351" => array("type"=>"city", "pop"=>7216, "name_ru"=>"Линвуд", "name_en"=>"Linwood", "lang"=>"en", "lat"=>39.342200, "lon"=>-74.570328),
"/US/US-NJ/little-falls-13352" => array("type"=>"cdp", "pop"=>11643, "name_ru"=>"Литтл Фоллс", "name_en"=>"Little Falls", "lang"=>"en", "lat"=>40.875491, "lon"=>-74.218520),
"/US/US-NJ/little-ferry-13353" => array("type"=>"borough", "pop"=>10477, "name_en"=>"Little Ferry", "lang"=>"en", "lat"=>40.845700, "lon"=>-74.039549),
"/US/US-NJ/little-silver-13354" => array("type"=>"borough", "pop"=>6119, "name_en"=>"Little Silver", "lang"=>"en", "lat"=>40.334967, "lon"=>-74.034951),
"/US/US-NJ/livingston-13355" => array("type"=>"cdp", "pop"=>27691, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>40.785567, "lon"=>-74.329513),
"/US/US-NJ/loch-arbour-13356" => array("type"=>"village", "pop"=>274, "name_en"=>"Loch Arbour", "lang"=>"en", "lat"=>40.231800, "lon"=>-74.001384),
"/US/US-NJ/lodi-13357" => array("type"=>"borough", "pop"=>23776, "name_ru"=>"Лоди", "name_en"=>"Lodi", "lang"=>"en", "lat"=>40.879016, "lon"=>-74.081196),
"/US/US-NJ/long-branch-13358" => array("type"=>"city", "pop"=>32622, "name_en"=>"Long Branch", "lang"=>"en", "lat"=>40.296700, "lon"=>-73.989199),
"/US/US-NJ/long-valley-13360" => array("type"=>"cdp", "pop"=>1744, "name_en"=>"Long Valley", "lang"=>"en", "lat"=>40.785386, "lon"=>-74.768369),
"/US/US-NJ/longport-13359" => array("type"=>"borough", "pop"=>1088, "name_en"=>"Longport", "lang"=>"en", "lat"=>39.311700, "lon"=>-74.525464),
"/US/US-NJ/lyndhurst-13361" => array("type"=>"cdp", "pop"=>19295, "name_en"=>"Lyndhurst", "lang"=>"en", "lat"=>40.796833, "lon"=>-74.113398),
"/US/US-NJ/madison-13363" => array("type"=>"borough", "pop"=>16009, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>40.758450, "lon"=>-74.417318),
"/US/US-NJ/madison-park-13364" => array("type"=>"cdp", "pop"=>7490, "name_en"=>"Madison Park", "lang"=>"en", "lat"=>40.446240, "lon"=>-74.295259),
"/US/US-NJ/magnolia-13365" => array("type"=>"borough", "pop"=>4328, "name_ru"=>"Магнолия", "name_en"=>"Magnolia", "lang"=>"en", "lat"=>39.855550, "lon"=>-75.036241),
"/US/US-NJ/manahawkin-13366" => array("type"=>"cdp", "pop"=>1594, "name_en"=>"Manahawkin", "lang"=>"en", "lat"=>39.695747, "lon"=>-74.253919),
"/US/US-NJ/manasquan-13367" => array("type"=>"borough", "pop"=>6251, "name_en"=>"Manasquan", "lang"=>"en", "lat"=>40.112800, "lon"=>-74.036956),
"/US/US-NJ/mantoloking-13368" => array("type"=>"borough", "pop"=>452, "name_en"=>"Mantoloking", "lang"=>"en", "lat"=>40.045250, "lon"=>-74.049649),
"/US/US-NJ/manville-13369" => array("type"=>"borough", "pop"=>10800, "name_en"=>"Manville", "lang"=>"en", "lat"=>40.541600, "lon"=>-74.589344),
"/US/US-NJ/maple-shade-13370" => array("type"=>"cdp", "pop"=>19154, "name_en"=>"Maple Shade", "lang"=>"en", "lat"=>39.951250, "lon"=>-74.995699),
"/US/US-NJ/maplewood-13371" => array("type"=>"cdp", "pop"=>21992, "name_ru"=>"Мейплвуд", "name_en"=>"Maplewood", "lang"=>"en", "lat"=>40.733133, "lon"=>-74.271895),
"/US/US-NJ/margate-city-13372" => array("type"=>"city", "pop"=>8501, "name_en"=>"Margate City", "lang"=>"en", "lat"=>39.330750, "lon"=>-74.507418),
"/US/US-NJ/marlton-13373" => array("type"=>"cdp", "pop"=>10228, "name_ru"=>"Марлтон", "name_en"=>"Marlton", "lang"=>"en", "lat"=>39.901400, "lon"=>-74.929314),
"/US/US-NJ/matawan-13374" => array("type"=>"borough", "pop"=>8754, "name_en"=>"Matawan", "lang"=>"en", "lat"=>40.412246, "lon"=>-74.237931),
"/US/US-NJ/mays-landing-13375" => array("type"=>"cdp", "pop"=>2090, "name_en"=>"Mays Landing", "lang"=>"en", "lat"=>39.452823, "lon"=>-74.724240),
"/US/US-NJ/maywood-13376" => array("type"=>"borough", "pop"=>9145, "name_ru"=>"Мэйвуд", "name_en"=>"Maywood", "lang"=>"en", "lat"=>40.902800, "lon"=>-74.063775),
"/US/US-NJ/mcguire-afb-13362" => array("type"=>"cdp", "pop"=>7580, "name_en"=>"McGuire AFB", "lang"=>"en", "lat"=>40.041267, "lon"=>-74.584368),
"/US/US-NJ/medford-lakes-13377" => array("type"=>"borough", "pop"=>4104, "name_en"=>"Medford Lakes", "lang"=>"en", "lat"=>39.859397, "lon"=>-74.805752),
"/US/US-NJ/mendham-13378" => array("type"=>"borough", "pop"=>5053, "name_en"=>"Mendham", "lang"=>"en", "lat"=>40.767144, "lon"=>-74.598255),
"/US/US-NJ/mercerville-hamilton-square-13379" => array("type"=>"cdp", "pop"=>26873, "name_en"=>"Mercerville-Hamilton Square", "lang"=>"en", "lat"=>40.230699, "lon"=>-74.671705),
"/US/US-NJ/merchantville-13380" => array("type"=>"borough", "pop"=>3764, "name_en"=>"Merchantville", "lang"=>"en", "lat"=>39.949867, "lon"=>-75.050632),
"/US/US-NJ/metuchen-13381" => array("type"=>"borough", "pop"=>13098, "name_en"=>"Metuchen", "lang"=>"en", "lat"=>40.541900, "lon"=>-74.363424),
"/US/US-NJ/middlesex-13382" => array("type"=>"borough", "pop"=>13644, "name_en"=>"Middlesex", "lang"=>"en", "lat"=>40.573700, "lon"=>-74.502142),
"/US/US-NJ/midland-park-13383" => array("type"=>"borough", "pop"=>6763, "name_en"=>"Midland Park", "lang"=>"en", "lat"=>40.994833, "lon"=>-74.142158),
"/US/US-NJ/milford-13384" => array("type"=>"borough", "pop"=>1186, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>40.568153, "lon"=>-75.090136),
"/US/US-NJ/millburn-13385" => array("type"=>"cdp", "pop"=>18546, "name_en"=>"Millburn", "lang"=>"en", "lat"=>40.740333, "lon"=>-74.324452),
"/US/US-NJ/millstone-13386" => array("type"=>"borough", "pop"=>410, "name_en"=>"Millstone", "lang"=>"en", "lat"=>40.499350, "lon"=>-74.591274),
"/US/US-NJ/milltown-13387" => array("type"=>"borough", "pop"=>6965, "name_ru"=>"Миллтаун", "name_en"=>"Milltown", "lang"=>"en", "lat"=>40.450100, "lon"=>-74.434720),
"/US/US-NJ/millville-13388" => array("type"=>"city", "pop"=>28930, "name_ru"=>"Миллвилл", "name_en"=>"Millville", "lang"=>"en", "lat"=>39.390028, "lon"=>-75.054689),
"/US/US-NJ/monmouth-beach-13389" => array("type"=>"borough", "pop"=>3570, "name_en"=>"Monmouth Beach", "lang"=>"en", "lat"=>40.335950, "lon"=>-73.986369),
"/US/US-NJ/monmouth-junction-13390" => array("type"=>"cdp", "pop"=>1570, "name_en"=>"Monmouth Junction", "lang"=>"en", "lat"=>40.380028, "lon"=>-74.545229),
"/US/US-NJ/montclair-13391" => array("type"=>"cdp", "pop"=>36966, "name_ru"=>"Монтклэр", "name_en"=>"Montclair", "lang"=>"en", "lat"=>40.824533, "lon"=>-74.211600),
"/US/US-NJ/montvale-13392" => array("type"=>"borough", "pop"=>7566, "name_en"=>"Montvale", "lang"=>"en", "lat"=>41.052192, "lon"=>-74.049068),
"/US/US-NJ/moonachie-13393" => array("type"=>"borough", "pop"=>2725, "name_en"=>"Moonachie", "lang"=>"en", "lat"=>40.842414, "lon"=>-74.059245),
"/US/US-NJ/moorestown-lenola-13394" => array("type"=>"cdp", "pop"=>13242, "name_en"=>"Moorestown-Lenola", "lang"=>"en", "lat"=>39.964317, "lon"=>-74.965970),
"/US/US-NJ/morris-plains-13395" => array("type"=>"borough", "pop"=>6005, "name_en"=>"Morris Plains", "lang"=>"en", "lat"=>40.836067, "lon"=>-74.481575),
"/US/US-NJ/morristown-13396" => array("type"=>"town", "pop"=>19268, "name_ru"=>"Морристаун", "name_en"=>"Morristown", "lang"=>"en", "lat"=>40.796733, "lon"=>-74.477556),
"/US/US-NJ/mount-arlington-13399" => array("type"=>"borough", "pop"=>5827, "name_en"=>"Mount Arlington", "lang"=>"en", "lat"=>40.919533, "lon"=>-74.640461),
"/US/US-NJ/mount-ephraim-13400" => array("type"=>"borough", "pop"=>4392, "name_en"=>"Mount Ephraim", "lang"=>"en", "lat"=>39.881000, "lon"=>-75.091727),
"/US/US-NJ/mount-holly-13401" => array("type"=>"cdp", "pop"=>10245, "name_en"=>"Mount Holly", "lang"=>"en", "lat"=>39.995033, "lon"=>-74.786824),
"/US/US-NJ/mountain-lakes-13397" => array("type"=>"borough", "pop"=>4274, "name_en"=>"Mountain Lakes", "lang"=>"en", "lat"=>40.891700, "lon"=>-74.440558),
"/US/US-NJ/mountainside-13398" => array("type"=>"borough", "pop"=>6551, "name_en"=>"Mountainside", "lang"=>"en", "lat"=>40.680600, "lon"=>-74.361049),
"/US/US-NJ/mullica-hill-13402" => array("type"=>"cdp", "pop"=>1117, "name_en"=>"Mullica Hill", "lang"=>"en", "lat"=>39.736174, "lon"=>-75.224564),
"/US/US-NJ/mystic-island-13403" => array("type"=>"cdp", "pop"=>7400, "name_en"=>"Mystic Island", "lang"=>"en", "lat"=>39.565559, "lon"=>-74.383286),
"/US/US-NJ/national-park-13404" => array("type"=>"borough", "pop"=>3227, "name_en"=>"National Park", "lang"=>"en", "lat"=>39.868000, "lon"=>-75.186149),
"/US/US-NJ/neptune-city-13405" => array("type"=>"borough", "pop"=>5114, "name_en"=>"Neptune City", "lang"=>"en", "lat"=>40.200600, "lon"=>-74.033109),
"/US/US-NJ/netcong-13406" => array("type"=>"borough", "pop"=>3222, "name_en"=>"Netcong", "lang"=>"en", "lat"=>40.899050, "lon"=>-74.701332),
"/US/US-NJ/new-brunswick-13408" => array("type"=>"city", "pop"=>51149, "name_en"=>"New Brunswick", "lang"=>"en", "lat"=>40.486400, "lon"=>-74.445133),
"/US/US-NJ/new-egypt-13409" => array("type"=>"cdp", "pop"=>2327, "name_en"=>"New Egypt", "lang"=>"en", "lat"=>40.063915, "lon"=>-74.531902),
"/US/US-NJ/new-milford-13411" => array("type"=>"borough", "pop"=>15947, "name_ru"=>"Нью Милфорд", "name_en"=>"New Milford", "lang"=>"en", "lat"=>40.934100, "lon"=>-74.019830),
"/US/US-NJ/new-providence-13412" => array("type"=>"borough", "pop"=>11905, "name_ru"=>"Нью Провиденс", "name_en"=>"New Providence", "lang"=>"en", "lat"=>40.700400, "lon"=>-74.403500),
"/US/US-NJ/newark-13407" => array("type"=>"city", "pop"=>278980, "name_ru"=>"Ньюарк", "name_en"=>"Newark", "lang"=>"en", "lat"=>40.724100, "lon"=>-74.173245),
"/US/US-NJ/newfield-13410" => array("type"=>"borough", "pop"=>1669, "name_en"=>"Newfield", "lang"=>"en", "lat"=>39.549362, "lon"=>-75.018724),
"/US/US-NJ/newton-13413" => array("type"=>"town", "pop"=>8096, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>41.054848, "lon"=>-74.753817),
"/US/US-NJ/north-arlington-13414" => array("type"=>"borough", "pop"=>14701, "name_en"=>"North Arlington", "lang"=>"en", "lat"=>40.790009, "lon"=>-74.126981),
"/US/US-NJ/north-beach-haven-13415" => array("type"=>"cdp", "pop"=>2413, "name_en"=>"North Beach Haven", "lang"=>"en", "lat"=>39.600173, "lon"=>-74.211890),
"/US/US-NJ/north-bergen-13416" => array("type"=>"cdp", "pop"=>55270, "name_en"=>"North Bergen", "lang"=>"en", "lat"=>40.792567, "lon"=>-74.025770),
"/US/US-NJ/north-brunswick-township-13417" => array("type"=>"cdp", "pop"=>31287, "name_en"=>"North Brunswick Township", "lang"=>"en", "lat"=>40.448450, "lon"=>-74.477578),
"/US/US-NJ/north-caldwell-13418" => array("type"=>"cdp", "pop"=>7051, "name_en"=>"North Caldwell", "lang"=>"en", "lat"=>40.863433, "lon"=>-74.259109),
"/US/US-NJ/north-cape-may-13419" => array("type"=>"cdp", "pop"=>3574, "name_en"=>"North Cape May", "lang"=>"en", "lat"=>38.976294, "lon"=>-74.951705),
"/US/US-NJ/north-haledon-13421" => array("type"=>"borough", "pop"=>8904, "name_en"=>"North Haledon", "lang"=>"en", "lat"=>40.961873, "lon"=>-74.186334),
"/US/US-NJ/north-middletown-13422" => array("type"=>"cdp", "pop"=>3160, "name_ru"=>"Норт Миддлтаун", "name_en"=>"North Middletown", "lang"=>"en", "lat"=>40.439572, "lon"=>-74.119307),
"/US/US-NJ/north-plainfield-13423" => array("type"=>"borough", "pop"=>21143, "name_en"=>"North Plainfield", "lang"=>"en", "lat"=>40.621150, "lon"=>-74.440097),
"/US/US-NJ/north-wildwood-13425" => array("type"=>"city", "pop"=>4800, "name_en"=>"North Wildwood", "lang"=>"en", "lat"=>39.004274, "lon"=>-74.799831),
"/US/US-NJ/northfield-13420" => array("type"=>"city", "pop"=>7885, "name_ru"=>"Нортфилд", "name_en"=>"Northfield", "lang"=>"en", "lat"=>39.373200, "lon"=>-74.554915),
"/US/US-NJ/northvale-13424" => array("type"=>"borough", "pop"=>4719, "name_en"=>"Northvale", "lang"=>"en", "lat"=>41.009754, "lon"=>-73.950062),
"/US/US-NJ/norwood-13426" => array("type"=>"borough", "pop"=>6211, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>40.994093, "lon"=>-73.951731),
"/US/US-NJ/nutley-13427" => array("type"=>"cdp", "pop"=>26171, "name_en"=>"Nutley", "lang"=>"en", "lat"=>40.818750, "lon"=>-74.157177),
"/US/US-NJ/oak-valley-13431" => array("type"=>"cdp", "pop"=>4055, "name_en"=>"Oak Valley", "lang"=>"en", "lat"=>39.805900, "lon"=>-75.159113),
"/US/US-NJ/oakhurst-13428" => array("type"=>"cdp", "pop"=>4130, "name_ru"=>"Оукхерст", "name_en"=>"Oakhurst", "lang"=>"en", "lat"=>40.261700, "lon"=>-74.027210),
"/US/US-NJ/oakland-13429" => array("type"=>"borough", "pop"=>13312, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>41.031383, "lon"=>-74.240368),
"/US/US-NJ/oaklyn-13430" => array("type"=>"borough", "pop"=>4030, "name_en"=>"Oaklyn", "lang"=>"en", "lat"=>39.901950, "lon"=>-75.080772),
"/US/US-NJ/ocean-acres-13432" => array("type"=>"cdp", "pop"=>5587, "name_en"=>"Ocean Acres", "lang"=>"en", "lat"=>39.743319, "lon"=>-74.281284),
"/US/US-NJ/ocean-city-13433" => array("type"=>"city", "pop"=>14756, "name_ru"=>"Оушен Сити", "name_en"=>"Ocean City", "lang"=>"en", "lat"=>39.269600, "lon"=>-74.603409),
"/US/US-NJ/ocean-gate-13434" => array("type"=>"borough", "pop"=>2136, "name_en"=>"Ocean Gate", "lang"=>"en", "lat"=>39.926150, "lon"=>-74.135071),
"/US/US-NJ/ocean-grove-13435" => array("type"=>"cdp", "pop"=>4818, "name_ru"=>"Оушен Гров", "name_en"=>"Ocean Grove", "lang"=>"en", "lat"=>40.211850, "lon"=>-74.009616),
"/US/US-NJ/oceanport-13436" => array("type"=>"borough", "pop"=>5733, "name_en"=>"Oceanport", "lang"=>"en", "lat"=>40.315000, "lon"=>-74.020379),
"/US/US-NJ/ogdensburg-13437" => array("type"=>"borough", "pop"=>2545, "name_en"=>"Ogdensburg", "lang"=>"en", "lat"=>41.075000, "lon"=>-74.597924),
"/US/US-NJ/old-bridge-13438" => array("type"=>"cdp", "pop"=>65859, "name_en"=>"Old Bridge", "lang"=>"en", "lat"=>40.392950, "lon"=>-74.330035),
"/US/US-NJ/old-tappan-13439" => array("type"=>"borough", "pop"=>6029, "name_en"=>"Old Tappan", "lang"=>"en", "lat"=>41.016203, "lon"=>-73.983031),
"/US/US-NJ/olivet-13440" => array("type"=>"cdp", "pop"=>1315, "name_ru"=>"Оливет", "name_en"=>"Olivet", "lang"=>"en", "lat"=>39.539652, "lon"=>-75.175011),
"/US/US-NJ/oradell-13441" => array("type"=>"borough", "pop"=>7780, "name_en"=>"Oradell", "lang"=>"en", "lat"=>40.955050, "lon"=>-74.031026),
"/US/US-NJ/orange-13442" => array("type"=>"cdp", "pop"=>29925, "name_ru"=>"Оранж", "name_en"=>"Orange", "lang"=>"en", "lat"=>40.768200, "lon"=>-74.236441),
"/US/US-NJ/oxford-13443" => array("type"=>"cdp", "pop"=>2575, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>40.812118, "lon"=>-74.995797),
"/US/US-NJ/palisades-park-13444" => array("type"=>"borough", "pop"=>19375, "name_en"=>"Palisades Park", "lang"=>"en", "lat"=>40.846950, "lon"=>-73.997278),
"/US/US-NJ/palmyra-13445" => array("type"=>"borough", "pop"=>7350, "name_ru"=>"Палмира", "name_en"=>"Palmyra", "lang"=>"en", "lat"=>40.002000, "lon"=>-75.034861),
"/US/US-NJ/paramus-13446" => array("type"=>"borough", "pop"=>26128, "name_en"=>"Paramus", "lang"=>"en", "lat"=>40.945117, "lon"=>-74.071524),
"/US/US-NJ/park-ridge-13447" => array("type"=>"borough", "pop"=>8917, "name_ru"=>"Парк Ридж", "name_en"=>"Park Ridge", "lang"=>"en", "lat"=>41.035790, "lon"=>-74.041511),
"/US/US-NJ/parsippany-troy-hills-township-13448" => array("type"=>"cdp", "pop"=>48478, "name_en"=>"Parsippany-Troy Hills Township", "lang"=>"en", "lat"=>40.859400, "lon"=>-74.423884),
"/US/US-NJ/passaic-13449" => array("type"=>"city", "pop"=>66884, "name_ru"=>"Пассаик", "name_en"=>"Passaic", "lang"=>"en", "lat"=>40.857300, "lon"=>-74.129404),
"/US/US-NJ/paterson-13450" => array("type"=>"city", "pop"=>145643, "name_en"=>"Paterson", "lang"=>"en", "lat"=>40.914450, "lon"=>-74.163413),
"/US/US-NJ/paulsboro-13451" => array("type"=>"borough", "pop"=>6075, "name_en"=>"Paulsboro", "lang"=>"en", "lat"=>39.839750, "lon"=>-75.240379),
"/US/US-NJ/peapack-and-gladstone-13452" => array("type"=>"borough", "pop"=>2556, "name_en"=>"Peapack and Gladstone", "lang"=>"en", "lat"=>40.715450, "lon"=>-74.657455),
"/US/US-NJ/pemberton-13453" => array("type"=>"borough", "pop"=>1540, "name_ru"=>"Пембертон", "name_en"=>"Pemberton", "lang"=>"en", "lat"=>39.971415, "lon"=>-74.686170),
"/US/US-NJ/pemberton-heights-13454" => array("type"=>"cdp", "pop"=>2941, "name_en"=>"Pemberton Heights", "lang"=>"en", "lat"=>39.957687, "lon"=>-74.678220),
"/US/US-NJ/pennington-13455" => array("type"=>"borough", "pop"=>2656, "name_ru"=>"Пеннингтон", "name_en"=>"Pennington", "lang"=>"en", "lat"=>40.326650, "lon"=>-74.792153),
"/US/US-NJ/penns-grove-13457" => array("type"=>"borough", "pop"=>4688, "name_en"=>"Penns Grove", "lang"=>"en", "lat"=>39.726496, "lon"=>-75.469436),
"/US/US-NJ/pennsauken-13456" => array("type"=>"cdp", "pop"=>34983, "name_en"=>"Pennsauken", "lang"=>"en", "lat"=>39.963650, "lon"=>-75.056171),
"/US/US-NJ/pennsville-13458" => array("type"=>"cdp", "pop"=>13345, "name_en"=>"Pennsville", "lang"=>"en", "lat"=>39.655424, "lon"=>-75.510651),
"/US/US-NJ/pequannock-township-13459" => array("type"=>"cdp", "pop"=>12844, "name_en"=>"Pequannock Township", "lang"=>"en", "lat"=>40.963333, "lon"=>-74.303741),
"/US/US-NJ/perth-amboy-13460" => array("type"=>"city", "pop"=>48742, "name_en"=>"Perth Amboy", "lang"=>"en", "lat"=>40.520200, "lon"=>-74.272098),
"/US/US-NJ/phillipsburg-13461" => array("type"=>"town", "pop"=>14528, "name_ru"=>"Филлипсбург", "name_en"=>"Phillipsburg", "lang"=>"en", "lat"=>40.688872, "lon"=>-75.182799),
"/US/US-NJ/pine-beach-13462" => array("type"=>"borough", "pop"=>2078, "name_en"=>"Pine Beach", "lang"=>"en", "lat"=>39.935500, "lon"=>-74.170099),
"/US/US-NJ/pine-hill-13463" => array("type"=>"borough", "pop"=>11308, "name_ru"=>"Пайн Хилл", "name_en"=>"Pine Hill", "lang"=>"en", "lat"=>39.786617, "lon"=>-74.985846),
"/US/US-NJ/pine-ridge-at-crestwood-13464" => array("type"=>"cdp", "pop"=>2372, "name_en"=>"Pine Ridge at Crestwood", "lang"=>"en", "lat"=>39.954572, "lon"=>-74.315608),
"/US/US-NJ/pine-valley-13465" => array("type"=>"borough", "pop"=>24, "name_ru"=>"Пайн Валли", "name_en"=>"Pine Valley", "lang"=>"en", "lat"=>39.788350, "lon"=>-74.972364),
"/US/US-NJ/pitman-13466" => array("type"=>"borough", "pop"=>9206, "name_en"=>"Pitman", "lang"=>"en", "lat"=>39.733250, "lon"=>-75.131035),
"/US/US-NJ/plainfield-13467" => array("type"=>"city", "pop"=>46126, "name_ru"=>"Плейнфилд", "name_en"=>"Plainfield", "lang"=>"en", "lat"=>40.615250, "lon"=>-74.416474),
"/US/US-NJ/pleasant-plains-13468" => array("type"=>"cdp", "pop"=>2577, "name_ru"=>"Плезант Плейнс", "name_en"=>"Pleasant Plains", "lang"=>"en", "lat"=>39.996963, "lon"=>-74.218958),
"/US/US-NJ/pleasantville-13469" => array("type"=>"city", "pop"=>18853, "name_ru"=>"Плезантвилл", "name_en"=>"Pleasantville", "lang"=>"en", "lat"=>39.390450, "lon"=>-74.518814),
"/US/US-NJ/point-pleasant-13470" => array("type"=>"borough", "pop"=>20139, "name_en"=>"Point Pleasant", "lang"=>"en", "lat"=>40.078400, "lon"=>-74.070916),
"/US/US-NJ/point-pleasant-beach-13471" => array("type"=>"borough", "pop"=>5418, "name_en"=>"Point Pleasant Beach", "lang"=>"en", "lat"=>40.092500, "lon"=>-74.045579),
"/US/US-NJ/pomona-13472" => array("type"=>"cdp", "pop"=>2624, "name_ru"=>"Помона", "name_en"=>"Pomona", "lang"=>"en", "lat"=>39.469624, "lon"=>-74.551921),
"/US/US-NJ/pompton-lakes-13473" => array("type"=>"borough", "pop"=>11031, "name_en"=>"Pompton Lakes", "lang"=>"en", "lat"=>41.003000, "lon"=>-74.285860),
"/US/US-NJ/port-monmouth-13474" => array("type"=>"cdp", "pop"=>3558, "name_en"=>"Port Monmouth", "lang"=>"en", "lat"=>40.432150, "lon"=>-74.101369),
"/US/US-NJ/port-norris-13475" => array("type"=>"cdp", "pop"=>1701, "name_en"=>"Port Norris", "lang"=>"en", "lat"=>39.247963, "lon"=>-75.041113),
"/US/US-NJ/port-reading-13476" => array("type"=>"cdp", "pop"=>3977, "name_en"=>"Port Reading", "lang"=>"en", "lat"=>40.564600, "lon"=>-74.247002),
"/US/US-NJ/port-republic-13477" => array("type"=>"city", "pop"=>1215, "name_en"=>"Port Republic", "lang"=>"en", "lat"=>39.538008, "lon"=>-74.487157),
"/US/US-NJ/presidential-lakes-estates-13478" => array("type"=>"cdp", "pop"=>2450, "name_en"=>"Presidential Lakes Estates", "lang"=>"en", "lat"=>39.913745, "lon"=>-74.564925),
"/US/US-NJ/princeton-13479" => array("type"=>"borough", "pop"=>13391, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>40.350250, "lon"=>-74.659447),
"/US/US-NJ/princeton-junction-13480" => array("type"=>"cdp", "pop"=>2362, "name_en"=>"Princeton Junction", "lang"=>"en", "lat"=>40.320300, "lon"=>-74.623683),
"/US/US-NJ/princeton-north-13481" => array("type"=>"cdp", "pop"=>4386, "name_en"=>"Princeton North", "lang"=>"en", "lat"=>40.361350, "lon"=>-74.648122),
"/US/US-NJ/prospect-park-13482" => array("type"=>"borough", "pop"=>5594, "name_en"=>"Prospect Park", "lang"=>"en", "lat"=>40.937100, "lon"=>-74.172554),
"/US/US-NJ/rahway-13483" => array("type"=>"city", "pop"=>28624, "name_en"=>"Rahway", "lang"=>"en", "lat"=>40.607000, "lon"=>-74.281407),
"/US/US-NJ/ramblewood-13484" => array("type"=>"cdp", "pop"=>6181, "name_en"=>"Ramblewood", "lang"=>"en", "lat"=>39.931600, "lon"=>-74.953248),
"/US/US-NJ/ramsey-13485" => array("type"=>"borough", "pop"=>14595, "name_ru"=>"Рэмси", "name_en"=>"Ramsey", "lang"=>"en", "lat"=>41.059223, "lon"=>-74.146495),
"/US/US-NJ/raritan-13486" => array("type"=>"borough", "pop"=>7289, "name_ru"=>"Раритан", "name_en"=>"Raritan", "lang"=>"en", "lat"=>40.571950, "lon"=>-74.641751),
"/US/US-NJ/red-bank-13487" => array("type"=>"borough", "pop"=>11866, "name_en"=>"Red Bank", "lang"=>"en", "lat"=>40.347350, "lon"=>-74.067457),
"/US/US-NJ/ridgefield-13488" => array("type"=>"borough", "pop"=>10851, "name_ru"=>"Риджфилд", "name_en"=>"Ridgefield", "lang"=>"en", "lat"=>40.831400, "lon"=>-74.015674),
"/US/US-NJ/ridgefield-park-13489" => array("type"=>"village", "pop"=>12370, "name_en"=>"Ridgefield Park", "lang"=>"en", "lat"=>40.854600, "lon"=>-74.020467),
"/US/US-NJ/ridgewood-13490" => array("type"=>"village", "pop"=>24163, "name_en"=>"Ridgewood", "lang"=>"en", "lat"=>40.981067, "lon"=>-74.113891),
"/US/US-NJ/ringwood-13491" => array("type"=>"borough", "pop"=>12681, "name_en"=>"Ringwood", "lang"=>"en", "lat"=>41.111504, "lon"=>-74.274753),
"/US/US-NJ/rio-grande-13492" => array("type"=>"cdp", "pop"=>2505, "name_en"=>"Rio Grande", "lang"=>"en", "lat"=>39.021968, "lon"=>-74.877718),
"/US/US-NJ/river-edge-13494" => array("type"=>"borough", "pop"=>10674, "name_en"=>"River Edge", "lang"=>"en", "lat"=>40.927200, "lon"=>-74.040033),
"/US/US-NJ/river-vale-13497" => array("type"=>"cdp", "pop"=>9620, "name_en"=>"River Vale", "lang"=>"en", "lat"=>41.013617, "lon"=>-74.006970),
"/US/US-NJ/riverdale-13493" => array("type"=>"borough", "pop"=>2858, "name_ru"=>"Ривердейл", "name_en"=>"Riverdale", "lang"=>"en", "lat"=>40.995233, "lon"=>-74.313797),
"/US/US-NJ/riverside-13495" => array("type"=>"cdp", "pop"=>7693, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>40.035500, "lon"=>-74.955915),
"/US/US-NJ/riverton-13496" => array("type"=>"borough", "pop"=>2627, "name_ru"=>"Ривертон", "name_en"=>"Riverton", "lang"=>"en", "lat"=>40.011750, "lon"=>-75.014909),
"/US/US-NJ/robertsville-13498" => array("type"=>"cdp", "pop"=>9841, "name_en"=>"Robertsville", "lang"=>"en", "lat"=>40.340288, "lon"=>-74.297415),
"/US/US-NJ/rochelle-park-13499" => array("type"=>"cdp", "pop"=>6079, "name_en"=>"Rochelle Park", "lang"=>"en", "lat"=>40.906450, "lon"=>-74.078386),
"/US/US-NJ/rockaway-13500" => array("type"=>"borough", "pop"=>6263, "name_en"=>"Rockaway", "lang"=>"en", "lat"=>40.896867, "lon"=>-74.514819),
"/US/US-NJ/rockleigh-13501" => array("type"=>"borough", "pop"=>388, "name_en"=>"Rockleigh", "lang"=>"en", "lat"=>41.002049, "lon"=>-73.935480),
"/US/US-NJ/rocky-hill-13502" => array("type"=>"borough", "pop"=>675, "name_en"=>"Rocky Hill", "lang"=>"en", "lat"=>40.400300, "lon"=>-74.639353),
"/US/US-NJ/roosevelt-13503" => array("type"=>"borough", "pop"=>907, "name_ru"=>"Рузвельт", "name_en"=>"Roosevelt", "lang"=>"en", "lat"=>40.221290, "lon"=>-74.474204),
"/US/US-NJ/roseland-13504" => array("type"=>"borough", "pop"=>5352, "name_ru"=>"Роузлэнд", "name_en"=>"Roseland", "lang"=>"en", "lat"=>40.821950, "lon"=>-74.308373),
"/US/US-NJ/roselle-13505" => array("type"=>"borough", "pop"=>20599, "name_ru"=>"Розл", "name_en"=>"Roselle", "lang"=>"en", "lat"=>40.652150, "lon"=>-74.260814),
"/US/US-NJ/roselle-park-13506" => array("type"=>"borough", "pop"=>12802, "name_en"=>"Roselle Park", "lang"=>"en", "lat"=>40.665650, "lon"=>-74.267149),
"/US/US-NJ/rosenhayn-13507" => array("type"=>"cdp", "pop"=>1053, "name_en"=>"Rosenhayn", "lang"=>"en", "lat"=>39.478850, "lon"=>-75.136508),
"/US/US-NJ/rossmoor-13508" => array("type"=>"cdp", "pop"=>3231, "name_ru"=>"Россмур", "name_en"=>"Rossmoor", "lang"=>"en", "lat"=>40.336485, "lon"=>-74.473887),
"/US/US-NJ/rumson-13509" => array("type"=>"borough", "pop"=>7264, "name_en"=>"Rumson", "lang"=>"en", "lat"=>40.362700, "lon"=>-74.003858),
"/US/US-NJ/runnemede-13510" => array("type"=>"borough", "pop"=>8410, "name_en"=>"Runnemede", "lang"=>"en", "lat"=>39.851450, "lon"=>-75.075115),
"/US/US-NJ/rutherford-13511" => array("type"=>"borough", "pop"=>17454, "name_en"=>"Rutherford", "lang"=>"en", "lat"=>40.819850, "lon"=>-74.107052),
"/US/US-NJ/saddle-brook-13512" => array("type"=>"cdp", "pop"=>13929, "name_en"=>"Saddle Brook", "lang"=>"en", "lat"=>40.904200, "lon"=>-74.095502),
"/US/US-NJ/saddle-river-13513" => array("type"=>"borough", "pop"=>3793, "name_en"=>"Saddle River", "lang"=>"en", "lat"=>41.026272, "lon"=>-74.098923),
"/US/US-NJ/salem-13514" => array("type"=>"city", "pop"=>5661, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>39.568040, "lon"=>-75.473168),
"/US/US-NJ/sayreville-13515" => array("type"=>"borough", "pop"=>42208, "name_en"=>"Sayreville", "lang"=>"en", "lat"=>40.465628, "lon"=>-74.322107),
"/US/US-NJ/scotch-plains-13516" => array("type"=>"cdp", "pop"=>22891, "name_en"=>"Scotch Plains", "lang"=>"en", "lat"=>40.633100, "lon"=>-74.373473),
"/US/US-NJ/sea-bright-13517" => array("type"=>"borough", "pop"=>1804, "name_en"=>"Sea Bright", "lang"=>"en", "lat"=>40.362093, "lon"=>-73.975554),
"/US/US-NJ/sea-girt-13519" => array("type"=>"borough", "pop"=>2070, "name_en"=>"Sea Girt", "lang"=>"en", "lat"=>40.129200, "lon"=>-74.034489),
"/US/US-NJ/sea-isle-city-13520" => array("type"=>"city", "pop"=>2909, "name_en"=>"Sea Isle City", "lang"=>"en", "lat"=>39.152406, "lon"=>-74.697843),
"/US/US-NJ/seabrook-farms-13518" => array("type"=>"cdp", "pop"=>1457, "name_en"=>"Seabrook Farms", "lang"=>"en", "lat"=>39.500965, "lon"=>-75.218295),
"/US/US-NJ/seaside-heights-13521" => array("type"=>"borough", "pop"=>3342, "name_en"=>"Seaside Heights", "lang"=>"en", "lat"=>39.943100, "lon"=>-74.076480),
"/US/US-NJ/seaside-park-13522" => array("type"=>"borough", "pop"=>2311, "name_en"=>"Seaside Park", "lang"=>"en", "lat"=>39.925750, "lon"=>-74.079067),
"/US/US-NJ/secaucus-13523" => array("type"=>"town", "pop"=>15372, "name_en"=>"Secaucus", "lang"=>"en", "lat"=>40.781114, "lon"=>-74.064778),
"/US/US-NJ/sewaren-13524" => array("type"=>"cdp", "pop"=>2569, "name_en"=>"Sewaren", "lang"=>"en", "lat"=>40.550850, "lon"=>-74.260931),
"/US/US-NJ/shark-river-hills-13525" => array("type"=>"cdp", "pop"=>4228, "name_en"=>"Shark River Hills", "lang"=>"en", "lat"=>40.193650, "lon"=>-74.048249),
"/US/US-NJ/shiloh-13526" => array("type"=>"borough", "pop"=>663, "name_ru"=>"Шилох", "name_en"=>"Shiloh", "lang"=>"en", "lat"=>39.458000, "lon"=>-75.297975),
"/US/US-NJ/ship-bottom-13527" => array("type"=>"borough", "pop"=>1449, "name_en"=>"Ship Bottom", "lang"=>"en", "lat"=>39.645099, "lon"=>-74.183269),
"/US/US-NJ/shrewsbury-13528" => array("type"=>"borough", "pop"=>3778, "name_ru"=>"Шрюсберри", "name_en"=>"Shrewsbury", "lang"=>"en", "lat"=>40.326600, "lon"=>-74.059730),
"/US/US-NJ/silver-ridge-13529" => array("type"=>"cdp", "pop"=>1138, "name_en"=>"Silver Ridge", "lang"=>"en", "lat"=>39.961035, "lon"=>-74.236217),
"/US/US-NJ/silverton-13530" => array("type"=>"cdp", "pop"=>9175, "name_ru"=>"Сильвертон", "name_en"=>"Silverton", "lang"=>"en", "lat"=>40.007867, "lon"=>-74.121824),
"/US/US-NJ/society-hill-13531" => array("type"=>"cdp", "pop"=>3577, "name_en"=>"Society Hill", "lang"=>"en", "lat"=>40.533963, "lon"=>-74.458291),
"/US/US-NJ/somerdale-13532" => array("type"=>"borough", "pop"=>5085, "name_en"=>"Somerdale", "lang"=>"en", "lat"=>39.845300, "lon"=>-75.023029),
"/US/US-NJ/somers-point-13534" => array("type"=>"city", "pop"=>11343, "name_en"=>"Somers Point", "lang"=>"en", "lat"=>39.317267, "lon"=>-74.606831),
"/US/US-NJ/somerset-13533" => array("type"=>"cdp", "pop"=>324563, "name_ru"=>"Сомерсет", "name_en"=>"Somerset", "lang"=>"en", "lat"=>40.497567, "lon"=>-74.488970),
"/US/US-NJ/somerville-13535" => array("type"=>"borough", "pop"=>12637, "name_ru"=>"Сомервилл", "name_en"=>"Somerville", "lang"=>"en", "lat"=>40.570200, "lon"=>-74.610449),
"/US/US-NJ/south-amboy-13536" => array("type"=>"city", "pop"=>7777, "name_en"=>"South Amboy", "lang"=>"en", "lat"=>40.485500, "lon"=>-74.277249),
"/US/US-NJ/south-belmar-13537" => array("type"=>"borough", "pop"=>1482, "name_en"=>"South Belmar", "lang"=>"en", "lat"=>40.170100, "lon"=>-74.026224),
"/US/US-NJ/south-bound-brook-13538" => array("type"=>"borough", "pop"=>5110, "name_en"=>"South Bound Brook", "lang"=>"en", "lat"=>40.553950, "lon"=>-74.528099),
"/US/US-NJ/south-orange-13539" => array("type"=>"cdp", "pop"=>16390, "name_en"=>"South Orange", "lang"=>"en", "lat"=>40.749000, "lon"=>-74.261622),
"/US/US-NJ/south-plainfield-13540" => array("type"=>"borough", "pop"=>22623, "name_en"=>"South Plainfield", "lang"=>"en", "lat"=>40.575350, "lon"=>-74.415275),
"/US/US-NJ/south-river-13541" => array("type"=>"borough", "pop"=>15682, "name_en"=>"South River", "lang"=>"en", "lat"=>40.445250, "lon"=>-74.378397),
"/US/US-NJ/south-toms-river-13542" => array("type"=>"borough", "pop"=>3709, "name_en"=>"South Toms River", "lang"=>"en", "lat"=>39.940759, "lon"=>-74.209697),
"/US/US-NJ/spotswood-13543" => array("type"=>"borough", "pop"=>8145, "name_en"=>"Spotswood", "lang"=>"en", "lat"=>40.393916, "lon"=>-74.390517),
"/US/US-NJ/spring-lake-13545" => array("type"=>"borough", "pop"=>3520, "name_ru"=>"Спринг Лейк", "name_en"=>"Spring Lake", "lang"=>"en", "lat"=>40.152750, "lon"=>-74.027385),
"/US/US-NJ/spring-lake-heights-13546" => array("type"=>"borough", "pop"=>5134, "name_en"=>"Spring Lake Heights", "lang"=>"en", "lat"=>40.152167, "lon"=>-74.042349),
"/US/US-NJ/springfield-13544" => array("type"=>"cdp", "pop"=>3466, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>40.699467, "lon"=>-74.325029),
"/US/US-NJ/stanhope-13547" => array("type"=>"borough", "pop"=>3562, "name_ru"=>"Стэнохоуп", "name_en"=>"Stanhope", "lang"=>"en", "lat"=>40.912900, "lon"=>-74.702866),
"/US/US-NJ/stockton-13548" => array("type"=>"borough", "pop"=>546, "name_ru"=>"Стоктон", "name_en"=>"Stockton", "lang"=>"en", "lat"=>40.405372, "lon"=>-74.977481),
"/US/US-NJ/stone-harbor-13549" => array("type"=>"borough", "pop"=>1012, "name_en"=>"Stone Harbor", "lang"=>"en", "lat"=>39.044025, "lon"=>-74.767783),
"/US/US-NJ/stratford-13550" => array("type"=>"borough", "pop"=>7049, "name_ru"=>"Стрэтфорд", "name_en"=>"Stratford", "lang"=>"en", "lat"=>39.828250, "lon"=>-75.015051),
"/US/US-NJ/strathmore-13551" => array("type"=>"cdp", "pop"=>7060, "name_ru"=>"Стратмор", "name_en"=>"Strathmore", "lang"=>"en", "lat"=>40.402767, "lon"=>-74.218882),
"/US/US-NJ/succasunna-kenvil-13552" => array("type"=>"cdp", "pop"=>11781, "name_en"=>"Succasunna-Kenvil", "lang"=>"en", "lat"=>40.855450, "lon"=>-74.652908),
"/US/US-NJ/summit-13553" => array("type"=>"city", "pop"=>20557, "name_ru"=>"Саммит", "name_en"=>"Summit", "lang"=>"en", "lat"=>40.714933, "lon"=>-74.365405),
"/US/US-NJ/surf-city-13554" => array("type"=>"borough", "pop"=>1558, "name_en"=>"Surf City", "lang"=>"en", "lat"=>39.662318, "lon"=>-74.169159),
"/US/US-NJ/sussex-13555" => array("type"=>"borough", "pop"=>2124, "name_en"=>"Sussex", "lang"=>"en", "lat"=>41.209793, "lon"=>-74.608468),
"/US/US-NJ/swedesboro-13556" => array("type"=>"borough", "pop"=>2250, "name_en"=>"Swedesboro", "lang"=>"en", "lat"=>39.746083, "lon"=>-75.312784),
"/US/US-NJ/tavistock-13557" => array("type"=>"borough", "pop"=>30, "name_en"=>"Tavistock", "lang"=>"en", "lat"=>39.875067, "lon"=>-75.028374),
"/US/US-NJ/teaneck-13558" => array("type"=>"cdp", "pop"=>38662, "name_en"=>"Teaneck", "lang"=>"en", "lat"=>40.889650, "lon"=>-74.012115),
"/US/US-NJ/tenafly-13559" => array("type"=>"borough", "pop"=>14662, "name_en"=>"Tenafly", "lang"=>"en", "lat"=>40.917150, "lon"=>-73.955174),
"/US/US-NJ/teterboro-13560" => array("type"=>"borough", "pop"=>17, "name_en"=>"Teterboro", "lang"=>"en", "lat"=>40.853550, "lon"=>-74.060789),
"/US/US-NJ/tinton-falls-13561" => array("type"=>"borough", "pop"=>19703, "name_en"=>"Tinton Falls", "lang"=>"en", "lat"=>40.272433, "lon"=>-74.089176),
"/US/US-NJ/toms-river-13562" => array("type"=>"cdp", "pop"=>95933, "name_en"=>"Toms River", "lang"=>"en", "lat"=>39.954600, "lon"=>-74.184382),
"/US/US-NJ/totowa-13563" => array("type"=>"borough", "pop"=>10621, "name_en"=>"Totowa", "lang"=>"en", "lat"=>40.904800, "lon"=>-74.222116),
"/US/US-NJ/trenton-13564" => array("type"=>"city", "pop"=>82883, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>40.223400, "lon"=>-74.764224),
"/US/US-NJ/tuckerton-13565" => array("type"=>"borough", "pop"=>3876, "name_en"=>"Tuckerton", "lang"=>"en", "lat"=>39.597829, "lon"=>-74.329450),
"/US/US-NJ/turnersville-13566" => array("type"=>"cdp", "pop"=>3843, "name_en"=>"Turnersville", "lang"=>"en", "lat"=>39.765387, "lon"=>-75.062074),
"/US/US-NJ/twin-rivers-13567" => array("type"=>"cdp", "pop"=>7715, "name_en"=>"Twin Rivers", "lang"=>"en", "lat"=>40.264132, "lon"=>-74.492554),
"/US/US-NJ/union-13568" => array("type"=>"cdp", "pop"=>6255, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>40.695300, "lon"=>-74.269740),
"/US/US-NJ/union-beach-13569" => array("type"=>"borough", "pop"=>6620, "name_en"=>"Union Beach", "lang"=>"en", "lat"=>40.447250, "lon"=>-74.169099),
"/US/US-NJ/union-city-13570" => array("type"=>"city", "pop"=>62183, "name_ru"=>"Юнион Сити", "name_en"=>"Union City", "lang"=>"en", "lat"=>40.767113, "lon"=>-74.032493),
"/US/US-NJ/upper-saddle-river-13571" => array("type"=>"borough", "pop"=>8454, "name_en"=>"Upper Saddle River", "lang"=>"en", "lat"=>41.063400, "lon"=>-74.100485),
"/US/US-NJ/ventnor-city-13572" => array("type"=>"city", "pop"=>12187, "name_en"=>"Ventnor City", "lang"=>"en", "lat"=>39.341850, "lon"=>-74.483231),
"/US/US-NJ/vernon-valley-13573" => array("type"=>"cdp", "pop"=>1696, "name_en"=>"Vernon Valley", "lang"=>"en", "lat"=>41.236796, "lon"=>-74.487518),
"/US/US-NJ/verona-13574" => array("type"=>"cdp", "pop"=>12493, "name_ru"=>"Верона", "name_en"=>"Verona", "lang"=>"en", "lat"=>40.832350, "lon"=>-74.242685),
"/US/US-NJ/victory-gardens-13575" => array("type"=>"borough", "pop"=>1486, "name_en"=>"Victory Gardens", "lang"=>"en", "lat"=>40.876166, "lon"=>-74.544338),
"/US/US-NJ/victory-lakes-13576" => array("type"=>"cdp", "pop"=>2160, "name_en"=>"Victory Lakes", "lang"=>"en", "lat"=>39.632949, "lon"=>-74.966505),
"/US/US-NJ/villas-13577" => array("type"=>"cdp", "pop"=>8136, "name_ru"=>"Виллас", "name_en"=>"Villas", "lang"=>"en", "lat"=>39.015582, "lon"=>-74.936681),
"/US/US-NJ/vineland-13578" => array("type"=>"city", "pop"=>58780, "name_ru"=>"Вайнленд", "name_en"=>"Vineland", "lang"=>"en", "lat"=>39.464800, "lon"=>-74.997395),
"/US/US-NJ/waldwick-13579" => array("type"=>"borough", "pop"=>9522, "name_en"=>"Waldwick", "lang"=>"en", "lat"=>41.012932, "lon"=>-74.125125),
"/US/US-NJ/wallington-13580" => array("type"=>"borough", "pop"=>11296, "name_en"=>"Wallington", "lang"=>"en", "lat"=>40.853100, "lon"=>-74.106782),
"/US/US-NJ/wanamassa-13581" => array("type"=>"cdp", "pop"=>4530, "name_en"=>"Wanamassa", "lang"=>"en", "lat"=>40.236700, "lon"=>-74.029214),
"/US/US-NJ/wanaque-13582" => array("type"=>"borough", "pop"=>12195, "name_en"=>"Wanaque", "lang"=>"en", "lat"=>41.043200, "lon"=>-74.290596),
"/US/US-NJ/waretown-13583" => array("type"=>"cdp", "pop"=>1283, "name_en"=>"Waretown", "lang"=>"en", "lat"=>39.789650, "lon"=>-74.193797),
"/US/US-NJ/washington-13584" => array("type"=>"borough", "pop"=>6693, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>40.759307, "lon"=>-74.982889),
"/US/US-NJ/washington-township-13585" => array("type"=>"cdp", "pop"=>9245, "name_en"=>"Washington Township", "lang"=>"en", "lat"=>40.989044, "lon"=>-74.065029),
"/US/US-NJ/watchung-13586" => array("type"=>"borough", "pop"=>6502, "name_en"=>"Watchung", "lang"=>"en", "lat"=>40.643850, "lon"=>-74.437734),
"/US/US-NJ/wayne-13587" => array("type"=>"cdp", "pop"=>53918, "name_ru"=>"Уэйн", "name_en"=>"Wayne", "lang"=>"en", "lat"=>40.945900, "lon"=>-74.245502),
"/US/US-NJ/weehawken-13588" => array("type"=>"cdp", "pop"=>12370, "name_en"=>"Weehawken", "lang"=>"en", "lat"=>40.768681, "lon"=>-74.018858),
"/US/US-NJ/wenonah-13589" => array("type"=>"borough", "pop"=>2355, "name_ru"=>"Уэнона", "name_en"=>"Wenonah", "lang"=>"en", "lat"=>39.791700, "lon"=>-75.148527),
"/US/US-NJ/west-belmar-13590" => array("type"=>"cdp", "pop"=>2498, "name_en"=>"West Belmar", "lang"=>"en", "lat"=>40.170200, "lon"=>-74.037224),
"/US/US-NJ/west-caldwell-13591" => array("type"=>"cdp", "pop"=>10441, "name_en"=>"West Caldwell", "lang"=>"en", "lat"=>40.849867, "lon"=>-74.298336),
"/US/US-NJ/west-cape-may-13592" => array("type"=>"borough", "pop"=>980, "name_en"=>"West Cape May", "lang"=>"en", "lat"=>38.941395, "lon"=>-74.937866),
"/US/US-NJ/west-freehold-13594" => array("type"=>"cdp", "pop"=>11166, "name_en"=>"West Freehold", "lang"=>"en", "lat"=>40.233033, "lon"=>-74.295504),
"/US/US-NJ/west-long-branch-13595" => array("type"=>"borough", "pop"=>8339, "name_en"=>"West Long Branch", "lang"=>"en", "lat"=>40.289996, "lon"=>-74.018325),
"/US/US-NJ/west-milford-13596" => array("type"=>"cdp", "pop"=>27739, "name_en"=>"West Milford", "lang"=>"en", "lat"=>41.106133, "lon"=>-74.396277),
"/US/US-NJ/west-new-york-13597" => array("type"=>"town", "pop"=>46472, "name_en"=>"West New York", "lang"=>"en", "lat"=>40.785950, "lon"=>-74.009845),
"/US/US-NJ/west-orange-13598" => array("type"=>"cdp", "pop"=>42617, "name_en"=>"West Orange", "lang"=>"en", "lat"=>40.792850, "lon"=>-74.263603),
"/US/US-NJ/west-paterson-13599" => array("type"=>"borough", "pop"=>11908, "name_en"=>"West Paterson", "lang"=>"en", "lat"=>40.889150, "lon"=>-74.195564),
"/US/US-NJ/west-wildwood-13601" => array("type"=>"borough", "pop"=>400, "name_en"=>"West Wildwood", "lang"=>"en", "lat"=>39.000097, "lon"=>-74.824334),
"/US/US-NJ/westfield-13593" => array("type"=>"town", "pop"=>29450, "name_ru"=>"Уэстфилд", "name_en"=>"Westfield", "lang"=>"en", "lat"=>40.651650, "lon"=>-74.344212),
"/US/US-NJ/westville-13600" => array("type"=>"borough", "pop"=>4467, "name_ru"=>"Уэствилл", "name_en"=>"Westville", "lang"=>"en", "lat"=>39.869733, "lon"=>-75.129960),
"/US/US-NJ/westwood-13602" => array("type"=>"borough", "pop"=>10699, "name_ru"=>"Уэствуд", "name_en"=>"Westwood", "lang"=>"en", "lat"=>40.988385, "lon"=>-74.031118),
"/US/US-NJ/wharton-13603" => array("type"=>"borough", "pop"=>6083, "name_en"=>"Wharton", "lang"=>"en", "lat"=>40.899550, "lon"=>-74.582304),
"/US/US-NJ/white-horse-13604" => array("type"=>"cdp", "pop"=>9397, "name_en"=>"White Horse", "lang"=>"en", "lat"=>40.190850, "lon"=>-74.704479),
"/US/US-NJ/white-house-station-13605" => array("type"=>"cdp", "pop"=>1287, "name_en"=>"White House Station", "lang"=>"en", "lat"=>40.616200, "lon"=>-74.772421),
"/US/US-NJ/white-meadow-lake-13606" => array("type"=>"cdp", "pop"=>8002, "name_en"=>"White Meadow Lake", "lang"=>"en", "lat"=>40.923500, "lon"=>-74.511055),
"/US/US-NJ/whitesboro-burleigh-13607" => array("type"=>"cdp", "pop"=>2080, "name_en"=>"Whitesboro-Burleigh", "lang"=>"en", "lat"=>39.045842, "lon"=>-74.865788),
"/US/US-NJ/wildwood-13608" => array("type"=>"city", "pop"=>5259, "name_ru"=>"Уайлдвуд", "name_en"=>"Wildwood", "lang"=>"en", "lat"=>38.988857, "lon"=>-74.820252),
"/US/US-NJ/wildwood-crest-13609" => array("type"=>"borough", "pop"=>3985, "name_en"=>"Wildwood Crest", "lang"=>"en", "lat"=>38.972352, "lon"=>-74.838753),
"/US/US-NJ/williamstown-13610" => array("type"=>"cdp", "pop"=>10891, "name_ru"=>"Уильямстаун", "name_en"=>"Williamstown", "lang"=>"en", "lat"=>39.685738, "lon"=>-74.983865),
"/US/US-NJ/willingboro-13611" => array("type"=>"cdp", "pop"=>36530, "name_en"=>"Willingboro", "lang"=>"en", "lat"=>40.027150, "lon"=>-74.887418),
"/US/US-NJ/wood-ridge-13618" => array("type"=>"borough", "pop"=>7453, "name_en"=>"Wood-Ridge", "lang"=>"en", "lat"=>40.850400, "lon"=>-74.087688),
"/US/US-NJ/woodbine-13612" => array("type"=>"borough", "pop"=>2463, "name_ru"=>"Вудбайн", "name_en"=>"Woodbine", "lang"=>"en", "lat"=>39.227949, "lon"=>-74.808490),
"/US/US-NJ/woodbridge-13613" => array("type"=>"cdp", "pop"=>97963, "name_ru"=>"Вудбридж", "name_en"=>"Woodbridge", "lang"=>"en", "lat"=>40.554715, "lon"=>-74.285797),
"/US/US-NJ/woodbury-13614" => array("type"=>"city", "pop"=>10450, "name_ru"=>"Вудбери", "name_en"=>"Woodbury", "lang"=>"en", "lat"=>39.837250, "lon"=>-75.153435),
"/US/US-NJ/woodbury-heights-13615" => array("type"=>"borough", "pop"=>3052, "name_en"=>"Woodbury Heights", "lang"=>"en", "lat"=>39.816300, "lon"=>-75.151042),
"/US/US-NJ/woodcliff-lake-13616" => array("type"=>"borough", "pop"=>5941, "name_en"=>"Woodcliff Lake", "lang"=>"en", "lat"=>41.025615, "lon"=>-74.061199),
"/US/US-NJ/woodlynne-13617" => array("type"=>"borough", "pop"=>2687, "name_en"=>"Woodlynne", "lang"=>"en", "lat"=>39.916333, "lon"=>-75.095786),
"/US/US-NJ/woodstown-13619" => array("type"=>"borough", "pop"=>3360, "name_en"=>"Woodstown", "lang"=>"en", "lat"=>39.650880, "lon"=>-75.325075),
"/US/US-NJ/wrightstown-13620" => array("type"=>"borough", "pop"=>736, "name_en"=>"Wrightstown", "lang"=>"en", "lat"=>40.027498, "lon"=>-74.631041),
"/US/US-NJ/wyckoff-13621" => array("type"=>"cdp", "pop"=>16898, "name_en"=>"Wyckoff", "lang"=>"en", "lat"=>40.997781, "lon"=>-74.166414),
"/US/US-NJ/yardville-groveville-13622" => array("type"=>"cdp", "pop"=>9248, "name_en"=>"Yardville-Groveville", "lang"=>"en", "lat"=>40.181200, "lon"=>-74.667722),
"/US/US-NJ/yorketown-13623" => array("type"=>"cdp", "pop"=>6313, "name_en"=>"Yorketown", "lang"=>"en", "lat"=>40.307767, "lon"=>-74.337980),
"/US/US-NM" => array("type"=>"state", "pop"=>1984356, "name_ru"=>"Нью-мексико", "name_en"=>"New Mexico", "lang"=>"en"),
"/US/US-NM/acomita-lake-13624" => array("type"=>"cdp", "pop"=>273, "name_en"=>"Acomita Lake", "lang"=>"en", "lat"=>35.070173, "lon"=>-107.613031),
"/US/US-NM/agua-fria-13625" => array("type"=>"cdp", "pop"=>3717, "name_en"=>"Agua Fria", "lang"=>"en", "lat"=>35.647689, "lon"=>-106.021531),
"/US/US-NM/alamogordo-13626" => array("type"=>"city", "pop"=>35757, "name_en"=>"Alamogordo", "lang"=>"en", "lat"=>32.884659, "lon"=>-105.957564),
"/US/US-NM/albuquerque-13627" => array("type"=>"city", "pop"=>521999, "name_en"=>"Albuquerque", "lang"=>"en", "lat"=>35.117218, "lon"=>-106.624636),
"/US/US-NM/alcalde-13628" => array("type"=>"cdp", "pop"=>308, "name_en"=>"Alcalde", "lang"=>"en", "lat"=>36.084742, "lon"=>-106.058445),
"/US/US-NM/angel-fire-13629" => array("type"=>"village", "pop"=>982, "name_en"=>"Angel Fire", "lang"=>"en", "lat"=>36.390614, "lon"=>-105.278655),
"/US/US-NM/anthony-13630" => array("type"=>"cdp", "pop"=>5160, "name_ru"=>"Энтони", "name_en"=>"Anthony", "lang"=>"en", "lat"=>32.005419, "lon"=>-106.595633),
"/US/US-NM/artesia-13631" => array("type"=>"city", "pop"=>10994, "name_ru"=>"Артижа", "name_en"=>"Artesia", "lang"=>"en", "lat"=>32.847450, "lon"=>-104.429710),
"/US/US-NM/aztec-13632" => array("type"=>"city", "pop"=>6927, "name_en"=>"Aztec", "lang"=>"en", "lat"=>36.828402, "lon"=>-108.004633),
"/US/US-NM/bayard-13633" => array("type"=>"city", "pop"=>2399, "name_ru"=>"Байяр", "name_en"=>"Bayard", "lang"=>"en", "lat"=>32.759494, "lon"=>-108.133437),
"/US/US-NM/belen-13634" => array("type"=>"city", "pop"=>7325, "name_en"=>"Belen", "lang"=>"en", "lat"=>34.658660, "lon"=>-106.779016),
"/US/US-NM/bernalillo-13635" => array("type"=>"town", "pop"=>9237, "name_en"=>"Bernalillo", "lang"=>"en", "lat"=>35.316012, "lon"=>-106.555718),
"/US/US-NM/black-rock-13636" => array("type"=>"cdp", "pop"=>858, "name_ru"=>"Блэк Рок", "name_en"=>"Black Rock", "lang"=>"en", "lat"=>35.086013, "lon"=>-108.790015),
"/US/US-NM/bloomfield-13637" => array("type"=>"city", "pop"=>7234, "name_ru"=>"Блумфилд", "name_en"=>"Bloomfield", "lang"=>"en", "lat"=>36.711732, "lon"=>-107.977446),
"/US/US-NM/boles-acres-13638" => array("type"=>"cdp", "pop"=>1409, "name_en"=>"Boles Acres", "lang"=>"en", "lat"=>32.808826, "lon"=>-105.986672),
"/US/US-NM/bosque-farms-13639" => array("type"=>"village", "pop"=>4060, "name_en"=>"Bosque Farms", "lang"=>"en", "lat"=>34.854765, "lon"=>-106.700879),
"/US/US-NM/cannon-afb-13640" => array("type"=>"cdp", "pop"=>3312, "name_en"=>"Cannon AFB", "lang"=>"en", "lat"=>34.380375, "lon"=>-103.314115),
"/US/US-NM/capitan-13641" => array("type"=>"village", "pop"=>1510, "name_en"=>"Capitan", "lang"=>"en", "lat"=>33.540354, "lon"=>-105.593930),
"/US/US-NM/carlsbad-13642" => array("type"=>"city", "pop"=>25629, "name_ru"=>"Карлсбад", "name_en"=>"Carlsbad", "lang"=>"en", "lat"=>32.404163, "lon"=>-104.236613),
"/US/US-NM/carlsbad-north-13643" => array("type"=>"cdp", "pop"=>1167, "name_en"=>"Carlsbad North", "lang"=>"en", "lat"=>32.446306, "lon"=>-104.222621),
"/US/US-NM/carrizozo-13644" => array("type"=>"town", "pop"=>1047, "name_en"=>"Carrizozo", "lang"=>"en", "lat"=>33.643821, "lon"=>-105.883302),
"/US/US-NM/causey-13645" => array("type"=>"village", "pop"=>58, "name_en"=>"Causey", "lang"=>"en", "lat"=>33.865230, "lon"=>-103.116827),
"/US/US-NM/central-13646" => array("type"=>"village", "pop"=>1835, "name_ru"=>"Сентрал", "name_en"=>"Central", "lang"=>"en", "lat"=>32.775297, "lon"=>-108.154242),
"/US/US-NM/chama-13647" => array("type"=>"village", "pop"=>1347, "name_en"=>"Chama", "lang"=>"en", "lat"=>36.893504, "lon"=>-106.582592),
"/US/US-NM/chamisal-13648" => array("type"=>"cdp", "pop"=>272, "name_en"=>"Chamisal", "lang"=>"en", "lat"=>36.176666, "lon"=>-105.749036),
"/US/US-NM/chaparral-13649" => array("type"=>"cdp", "pop"=>2962, "name_en"=>"Chaparral", "lang"=>"en", "lat"=>32.025929, "lon"=>-106.399456),
"/US/US-NM/chimayo-13650" => array("type"=>"cdp", "pop"=>2789, "name_en"=>"Chimayo", "lang"=>"en", "lat"=>35.998760, "lon"=>-105.928546),
"/US/US-NM/cimarron-13651" => array("type"=>"village", "pop"=>816, "name_ru"=>"Симаррон", "name_en"=>"Cimarron", "lang"=>"en", "lat"=>36.509860, "lon"=>-104.912588),
"/US/US-NM/clayton-13652" => array("type"=>"town", "pop"=>2303, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>36.450747, "lon"=>-103.175095),
"/US/US-NM/cloudcroft-13653" => array("type"=>"village", "pop"=>883, "name_en"=>"Cloudcroft", "lang"=>"en", "lat"=>32.952980, "lon"=>-105.739426),
"/US/US-NM/clovis-13654" => array("type"=>"city", "pop"=>32352, "name_ru"=>"Кловис", "name_en"=>"Clovis", "lang"=>"en", "lat"=>34.410220, "lon"=>-103.202710),
"/US/US-NM/cochiti-13655" => array("type"=>"cdp", "pop"=>434, "name_en"=>"Cochiti", "lang"=>"en", "lat"=>35.608845, "lon"=>-106.349644),
"/US/US-NM/columbus-13656" => array("type"=>"village", "pop"=>1841, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>31.828873, "lon"=>-107.640756),
"/US/US-NM/corona-13657" => array("type"=>"village", "pop"=>185, "name_ru"=>"Корона", "name_en"=>"Corona", "lang"=>"en", "lat"=>34.249696, "lon"=>-105.596709),
"/US/US-NM/corrales-13658" => array("type"=>"village", "pop"=>7787, "name_en"=>"Corrales", "lang"=>"en", "lat"=>35.234953, "lon"=>-106.619968),
"/US/US-NM/crownpoint-13659" => array("type"=>"cdp", "pop"=>2108, "name_en"=>"Crownpoint", "lang"=>"en", "lat"=>35.688067, "lon"=>-108.146061),
"/US/US-NM/cuba-13660" => array("type"=>"village", "pop"=>1358, "name_ru"=>"Куба", "name_en"=>"Cuba", "lang"=>"en", "lat"=>36.022477, "lon"=>-106.954246),
"/US/US-NM/cuyamungue-13661" => array("type"=>"cdp", "pop"=>329, "name_en"=>"Cuyamungue", "lang"=>"en", "lat"=>35.869799, "lon"=>-106.008537),
"/US/US-NM/deming-13662" => array("type"=>"city", "pop"=>15480, "name_en"=>"Deming", "lang"=>"en", "lat"=>32.261728, "lon"=>-107.750419),
"/US/US-NM/des-moines-13663" => array("type"=>"village", "pop"=>162, "name_ru"=>"Дес Мойн", "name_en"=>"Des Moines", "lang"=>"en", "lat"=>36.761335, "lon"=>-103.833291),
"/US/US-NM/dexter-13664" => array("type"=>"town", "pop"=>1231, "name_ru"=>"Декстер", "name_en"=>"Dexter", "lang"=>"en", "lat"=>33.194721, "lon"=>-104.368364),
"/US/US-NM/dona-ana-13665" => array("type"=>"cdp", "pop"=>201603, "name_en"=>"Dona Ana", "lang"=>"en", "lat"=>32.391265, "lon"=>-106.815301),
"/US/US-NM/dora-13666" => array("type"=>"village", "pop"=>125, "name_ru"=>"Дора", "name_en"=>"Dora", "lang"=>"en", "lat"=>33.933174, "lon"=>-103.335314),
"/US/US-NM/dulce-13667" => array("type"=>"cdp", "pop"=>2438, "name_en"=>"Dulce", "lang"=>"en", "lat"=>36.933420, "lon"=>-106.996048),
"/US/US-NM/eagle-nest-13668" => array("type"=>"village", "pop"=>299, "name_en"=>"Eagle Nest", "lang"=>"en", "lat"=>36.552968, "lon"=>-105.260783),
"/US/US-NM/edgewood-13669" => array("type"=>"cdp", "pop"=>2742, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>35.083944, "lon"=>-106.174906),
"/US/US-NM/eldorado-at-santa-fe-13670" => array("type"=>"cdp", "pop"=>2260, "name_en"=>"Eldorado at Santa Fe", "lang"=>"en", "lat"=>35.526365, "lon"=>-105.934240),
"/US/US-NM/elida-13671" => array("type"=>"town", "pop"=>182, "name_en"=>"Elida", "lang"=>"en", "lat"=>33.944177, "lon"=>-103.654646),
"/US/US-NM/encino-13672" => array("type"=>"village", "pop"=>89, "name_en"=>"Encino", "lang"=>"en", "lat"=>34.650957, "lon"=>-105.457958),
"/US/US-NM/espanola-13673" => array("type"=>"city", "pop"=>9691, "name_en"=>"Espanola", "lang"=>"en", "lat"=>36.003406, "lon"=>-106.069302),
"/US/US-NM/estancia-13674" => array("type"=>"town", "pop"=>1572, "name_en"=>"Estancia", "lang"=>"en", "lat"=>34.760320, "lon"=>-106.060175),
"/US/US-NM/eunice-13675" => array("type"=>"city", "pop"=>2771, "name_ru"=>"Юнис", "name_en"=>"Eunice", "lang"=>"en", "lat"=>32.442515, "lon"=>-103.193237),
"/US/US-NM/farmington-13676" => array("type"=>"city", "pop"=>42637, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>36.754663, "lon"=>-108.185560),
"/US/US-NM/flora-vista-13677" => array("type"=>"cdp", "pop"=>1021, "name_en"=>"Flora Vista", "lang"=>"en", "lat"=>36.802149, "lon"=>-108.081977),
"/US/US-NM/floyd-13678" => array("type"=>"village", "pop"=>74, "name_ru"=>"Флойд", "name_en"=>"Floyd", "lang"=>"en", "lat"=>34.214697, "lon"=>-103.548904),
"/US/US-NM/folsom-13679" => array("type"=>"village", "pop"=>66, "name_ru"=>"Фолсом", "name_en"=>"Folsom", "lang"=>"en", "lat"=>36.848128, "lon"=>-103.917654),
"/US/US-NM/fort-sumner-13680" => array("type"=>"village", "pop"=>972, "name_en"=>"Fort Sumner", "lang"=>"en", "lat"=>34.478060, "lon"=>-104.237169),
"/US/US-NM/gallup-13681" => array("type"=>"city", "pop"=>19979, "name_en"=>"Gallup", "lang"=>"en", "lat"=>35.520427, "lon"=>-108.735359),
"/US/US-NM/grady-13682" => array("type"=>"village", "pop"=>91, "name_ru"=>"Грейди", "name_en"=>"Grady", "lang"=>"en", "lat"=>34.821962, "lon"=>-103.315414),
"/US/US-NM/grants-13683" => array("type"=>"city", "pop"=>8871, "name_en"=>"Grants", "lang"=>"en", "lat"=>35.153390, "lon"=>-107.838278),
"/US/US-NM/grenville-13684" => array("type"=>"village", "pop"=>23, "name_en"=>"Grenville", "lang"=>"en", "lat"=>36.592867, "lon"=>-103.612739),
"/US/US-NM/hagerman-13685" => array("type"=>"town", "pop"=>1160, "name_ru"=>"Хаджерман", "name_en"=>"Hagerman", "lang"=>"en", "lat"=>33.113478, "lon"=>-104.328282),
"/US/US-NM/hatch-13686" => array("type"=>"village", "pop"=>1647, "name_en"=>"Hatch", "lang"=>"en", "lat"=>32.667588, "lon"=>-107.155794),
"/US/US-NM/hobbs-13687" => array("type"=>"city", "pop"=>30476, "name_en"=>"Hobbs", "lang"=>"en", "lat"=>32.734264, "lon"=>-103.162760),
"/US/US-NM/holloman-afb-13688" => array("type"=>"cdp", "pop"=>5891, "name_en"=>"Holloman AFB", "lang"=>"en", "lat"=>32.851211, "lon"=>-106.106210),
"/US/US-NM/hope-13689" => array("type"=>"village", "pop"=>107, "name_ru"=>"Хоуп", "name_en"=>"Hope", "lang"=>"en", "lat"=>32.817263, "lon"=>-104.736711),
"/US/US-NM/house-13690" => array("type"=>"village", "pop"=>63, "name_en"=>"House", "lang"=>"en", "lat"=>34.647968, "lon"=>-103.903281),
"/US/US-NM/hurley-13691" => array("type"=>"town", "pop"=>1410, "name_ru"=>"Херли", "name_en"=>"Hurley", "lang"=>"en", "lat"=>32.697587, "lon"=>-108.131385),
"/US/US-NM/isleta-pueblo-13692" => array("type"=>"cdp", "pop"=>1703, "name_en"=>"Isleta Pueblo", "lang"=>"en", "lat"=>34.882215, "lon"=>-106.683394),
"/US/US-NM/jaconita-13693" => array("type"=>"cdp", "pop"=>375, "name_en"=>"Jaconita", "lang"=>"en", "lat"=>35.892549, "lon"=>-106.034819),
"/US/US-NM/jal-13694" => array("type"=>"city", "pop"=>2045, "name_en"=>"Jal", "lang"=>"en", "lat"=>32.113506, "lon"=>-103.189361),
"/US/US-NM/jemez-pueblo-13695" => array("type"=>"cdp", "pop"=>1301, "name_en"=>"Jemez Pueblo", "lang"=>"en", "lat"=>35.616274, "lon"=>-106.726249),
"/US/US-NM/jemez-springs-13696" => array("type"=>"village", "pop"=>439, "name_en"=>"Jemez Springs", "lang"=>"en", "lat"=>35.773654, "lon"=>-106.688359),
"/US/US-NM/kirtland-13697" => array("type"=>"cdp", "pop"=>3552, "name_en"=>"Kirtland", "lang"=>"en", "lat"=>36.739503, "lon"=>-108.343095),
"/US/US-NM/la-cienega-13698" => array("type"=>"cdp", "pop"=>1066, "name_en"=>"La Cienega", "lang"=>"en", "lat"=>35.576487, "lon"=>-106.110432),
"/US/US-NM/la-luz-13701" => array("type"=>"cdp", "pop"=>1625, "name_en"=>"La Luz", "lang"=>"en", "lat"=>32.970503, "lon"=>-105.937409),
"/US/US-NM/laguna-13699" => array("type"=>"cdp", "pop"=>434, "name_ru"=>"Лагуна", "name_en"=>"Laguna", "lang"=>"en", "lat"=>35.038613, "lon"=>-107.392558),
"/US/US-NM/lake-arthur-13700" => array("type"=>"town", "pop"=>436, "name_ru"=>"Лейк Артур", "name_en"=>"Lake Arthur", "lang"=>"en", "lat"=>32.999746, "lon"=>-104.363400),
"/US/US-NM/las-cruces-13702" => array("type"=>"city", "pop"=>91865, "name_en"=>"Las Cruces", "lang"=>"en", "lat"=>32.336067, "lon"=>-106.755756),
"/US/US-NM/las-vegas-13703" => array("type"=>"city", "pop"=>13818, "name_en"=>"Las Vegas", "lang"=>"en", "lat"=>35.599236, "lon"=>-105.219582),
"/US/US-NM/logan-13704" => array("type"=>"village", "pop"=>1002, "name_ru"=>"Логан", "name_en"=>"Logan", "lang"=>"en", "lat"=>35.360998, "lon"=>-103.443416),
"/US/US-NM/lordsburg-13705" => array("type"=>"city", "pop"=>2812, "name_en"=>"Lordsburg", "lang"=>"en", "lat"=>32.343284, "lon"=>-108.702018),
"/US/US-NM/los-alamos-13706" => array("type"=>"cdp", "pop"=>18150, "name_en"=>"Los Alamos", "lang"=>"en", "lat"=>35.893777, "lon"=>-106.283852),
"/US/US-NM/los-chaves-13707" => array("type"=>"cdp", "pop"=>3872, "name_en"=>"Los Chaves", "lang"=>"en", "lat"=>34.729969, "lon"=>-106.760838),
"/US/US-NM/los-lunas-13708" => array("type"=>"village", "pop"=>14153, "name_en"=>"Los Lunas", "lang"=>"en", "lat"=>34.813253, "lon"=>-106.738098),
"/US/US-NM/los-ranchos-de-albuquerque-13709" => array("type"=>"village", "pop"=>5622, "name_en"=>"Los Ranchos de Albuquerque", "lang"=>"en", "lat"=>35.165133, "lon"=>-106.648987),
"/US/US-NM/los-trujillos-gabaldon-13710" => array("type"=>"cdp", "pop"=>1841, "name_en"=>"Los Trujillos-Gabaldon", "lang"=>"en", "lat"=>34.661273, "lon"=>-106.755154),
"/US/US-NM/loving-13711" => array("type"=>"village", "pop"=>1340, "name_en"=>"Loving", "lang"=>"en", "lat"=>32.286487, "lon"=>-104.095952),
"/US/US-NM/lovington-13712" => array("type"=>"city", "pop"=>9975, "name_ru"=>"Лавингтон", "name_en"=>"Lovington", "lang"=>"en", "lat"=>32.945299, "lon"=>-103.350299),
"/US/US-NM/magdalena-13713" => array("type"=>"village", "pop"=>974, "name_en"=>"Magdalena", "lang"=>"en", "lat"=>34.110386, "lon"=>-107.234547),
"/US/US-NM/maxwell-13714" => array("type"=>"village", "pop"=>249, "name_ru"=>"Максуэлл", "name_en"=>"Maxwell", "lang"=>"en", "lat"=>36.540475, "lon"=>-104.542111),
"/US/US-NM/meadow-lake-13715" => array("type"=>"cdp", "pop"=>1590, "name_en"=>"Meadow Lake", "lang"=>"en", "lat"=>34.802024, "lon"=>-106.590724),
"/US/US-NM/melrose-13716" => array("type"=>"village", "pop"=>678, "name_ru"=>"Мелроуз", "name_en"=>"Melrose", "lang"=>"en", "lat"=>34.429209, "lon"=>-103.628847),
"/US/US-NM/mescalero-13717" => array("type"=>"cdp", "pop"=>1159, "name_en"=>"Mescalero", "lang"=>"en", "lat"=>33.150813, "lon"=>-105.795992),
"/US/US-NM/mesilla-13718" => array("type"=>"town", "pop"=>2200, "name_en"=>"Mesilla", "lang"=>"en", "lat"=>32.270151, "lon"=>-106.806704),
"/US/US-NM/mesita-13719" => array("type"=>"cdp", "pop"=>627, "name_en"=>"Mesita", "lang"=>"en", "lat"=>35.013102, "lon"=>-107.330675),
"/US/US-NM/mexican-springs-13720" => array("type"=>"cdp", "pop"=>242, "name_en"=>"Mexican Springs", "lang"=>"en", "lat"=>35.787549, "lon"=>-108.807269),
"/US/US-NM/milan-13721" => array("type"=>"village", "pop"=>2484, "name_ru"=>"Милан", "name_en"=>"Milan", "lang"=>"en", "lat"=>35.188963, "lon"=>-107.891958),
"/US/US-NM/moriarty-13722" => array("type"=>"city", "pop"=>1952, "name_en"=>"Moriarty", "lang"=>"en", "lat"=>35.001355, "lon"=>-106.043733),
"/US/US-NM/mosquero-13723" => array("type"=>"village", "pop"=>81, "name_en"=>"Mosquero", "lang"=>"en", "lat"=>35.774547, "lon"=>-103.954276),
"/US/US-NM/mountainair-13724" => array("type"=>"town", "pop"=>1113, "name_en"=>"Mountainair", "lang"=>"en", "lat"=>34.520773, "lon"=>-106.242935),
"/US/US-NM/nambe-13725" => array("type"=>"cdp", "pop"=>1246, "name_en"=>"Nambe", "lang"=>"en", "lat"=>35.900235, "lon"=>-105.971791),
"/US/US-NM/naschitti-13726" => array("type"=>"cdp", "pop"=>323, "name_en"=>"Naschitti", "lang"=>"en", "lat"=>36.062221, "lon"=>-108.680973),
"/US/US-NM/navajo-13727" => array("type"=>"cdp", "pop"=>1985, "name_en"=>"Navajo", "lang"=>"en", "lat"=>35.896328, "lon"=>-109.031396),
"/US/US-NM/newcomb-13728" => array("type"=>"cdp", "pop"=>388, "name_en"=>"Newcomb", "lang"=>"en", "lat"=>36.283287, "lon"=>-108.706832),
"/US/US-NM/north-acomita-village-13729" => array("type"=>"cdp", "pop"=>314, "name_en"=>"North Acomita Village", "lang"=>"en", "lat"=>35.063244, "lon"=>-107.565410),
"/US/US-NM/north-valley-13730" => array("type"=>"cdp", "pop"=>12507, "name_en"=>"North Valley", "lang"=>"en", "lat"=>35.173229, "lon"=>-106.622655),
"/US/US-NM/ojo-amarillo-13731" => array("type"=>"cdp", "pop"=>955, "name_en"=>"Ojo Amarillo", "lang"=>"en", "lat"=>36.693286, "lon"=>-108.368536),
"/US/US-NM/paguate-13732" => array("type"=>"cdp", "pop"=>492, "name_en"=>"Paguate", "lang"=>"en", "lat"=>35.134594, "lon"=>-107.363806),
"/US/US-NM/paradise-hills-13733" => array("type"=>"cdp", "pop"=>5513, "name_en"=>"Paradise Hills", "lang"=>"en", "lat"=>35.196510, "lon"=>-106.694292),
"/US/US-NM/paraje-13734" => array("type"=>"cdp", "pop"=>622, "name_en"=>"Paraje", "lang"=>"en", "lat"=>35.046056, "lon"=>-107.467315),
"/US/US-NM/pecos-13735" => array("type"=>"village", "pop"=>1489, "name_en"=>"Pecos", "lang"=>"en", "lat"=>35.575031, "lon"=>-105.678731),
"/US/US-NM/pena-blanca-13736" => array("type"=>"cdp", "pop"=>300, "name_en"=>"Pena Blanca", "lang"=>"en", "lat"=>35.571414, "lon"=>-106.336214),
"/US/US-NM/penasco-13737" => array("type"=>"cdp", "pop"=>648, "name_en"=>"Penasco", "lang"=>"en", "lat"=>36.171204, "lon"=>-105.691913),
"/US/US-NM/peralta-13738" => array("type"=>"cdp", "pop"=>3597, "name_en"=>"Peralta", "lang"=>"en", "lat"=>34.829541, "lon"=>-106.687428),
"/US/US-NM/placitas-13739" => array("type"=>"cdp", "pop"=>1611, "name_en"=>"Placitas", "lang"=>"en", "lat"=>35.320189, "lon"=>-106.465837),
"/US/US-NM/pojoaque-13740" => array("type"=>"cdp", "pop"=>1037, "name_en"=>"Pojoaque", "lang"=>"en", "lat"=>35.893572, "lon"=>-106.009868),
"/US/US-NM/portales-13741" => array("type"=>"city", "pop"=>12215, "name_en"=>"Portales", "lang"=>"en", "lat"=>34.181093, "lon"=>-103.345307),
"/US/US-NM/questa-13742" => array("type"=>"village", "pop"=>1907, "name_en"=>"Questa", "lang"=>"en", "lat"=>36.710922, "lon"=>-105.592658),
"/US/US-NM/ranchos-de-taos-13743" => array("type"=>"cdp", "pop"=>1779, "name_en"=>"Ranchos De Taos", "lang"=>"en", "lat"=>36.362902, "lon"=>-105.603286),
"/US/US-NM/raton-13744" => array("type"=>"city", "pop"=>6465, "name_en"=>"Raton", "lang"=>"en", "lat"=>36.885738, "lon"=>-104.441509),
"/US/US-NM/red-river-13745" => array("type"=>"town", "pop"=>515, "name_en"=>"Red River", "lang"=>"en", "lat"=>36.707504, "lon"=>-105.406660),
"/US/US-NM/reserve-13746" => array("type"=>"village", "pop"=>389, "name_ru"=>"Резерв", "name_en"=>"Reserve", "lang"=>"en", "lat"=>33.708444, "lon"=>-108.761165),
"/US/US-NM/rio-communities-13747" => array("type"=>"cdp", "pop"=>3233, "name_en"=>"Rio Communities", "lang"=>"en", "lat"=>34.644073, "lon"=>-106.718038),
"/US/US-NM/rio-rancho-13748" => array("type"=>"city", "pop"=>79655, "name_en"=>"Rio Rancho", "lang"=>"en", "lat"=>35.273422, "lon"=>-106.660442),
"/US/US-NM/roswell-13749" => array("type"=>"city", "pop"=>46198, "name_ru"=>"Росуэл", "name_en"=>"Roswell", "lang"=>"en", "lat"=>33.371094, "lon"=>-104.529386),
"/US/US-NM/roy-13750" => array("type"=>"village", "pop"=>204, "name_en"=>"Roy", "lang"=>"en", "lat"=>35.945371, "lon"=>-104.196277),
"/US/US-NM/ruidoso-13751" => array("type"=>"village", "pop"=>9096, "name_en"=>"Ruidoso", "lang"=>"en", "lat"=>33.366049, "lon"=>-105.642487),
"/US/US-NM/ruidoso-downs-13752" => array("type"=>"village", "pop"=>2637, "name_en"=>"Ruidoso Downs", "lang"=>"en", "lat"=>33.331585, "lon"=>-105.603679),
"/US/US-NM/san-felipe-pueblo-13755" => array("type"=>"cdp", "pop"=>1557, "name_en"=>"San Felipe Pueblo", "lang"=>"en", "lat"=>35.434310, "lon"=>-106.428403),
"/US/US-NM/san-ildefonso-pueblo-13756" => array("type"=>"cdp", "pop"=>447, "name_en"=>"San Ildefonso Pueblo", "lang"=>"en", "lat"=>35.898052, "lon"=>-106.127502),
"/US/US-NM/san-jon-13757" => array("type"=>"village", "pop"=>249, "name_en"=>"San Jon", "lang"=>"en", "lat"=>35.112968, "lon"=>-103.327467),
"/US/US-NM/san-juan-13758" => array("type"=>"cdp", "pop"=>122500, "name_en"=>"San Juan", "lang"=>"en", "lat"=>36.051926, "lon"=>-106.068351),
"/US/US-NM/san-ysidro-13766" => array("type"=>"village", "pop"=>229, "name_en"=>"San Ysidro", "lang"=>"en", "lat"=>35.556545, "lon"=>-106.771924),
"/US/US-NM/sandia-13753" => array("type"=>"cdp", "pop"=>6742, "name_en"=>"Sandia", "lang"=>"en", "lat"=>35.053500, "lon"=>-106.565248),
"/US/US-NM/sandia-heights-13754" => array("type"=>"cdp", "pop"=>3519, "name_en"=>"Sandia Heights", "lang"=>"en", "lat"=>35.176931, "lon"=>-106.490724),
"/US/US-NM/sanostee-13759" => array("type"=>"cdp", "pop"=>626, "name_en"=>"Sanostee", "lang"=>"en", "lat"=>36.434753, "lon"=>-108.872714),
"/US/US-NM/santa-ana-pueblo-13760" => array("type"=>"cdp", "pop"=>476, "name_en"=>"Santa Ana Pueblo", "lang"=>"en", "lat"=>35.352536, "lon"=>-106.518479),
"/US/US-NM/santa-clara-pueblo-13761" => array("type"=>"cdp", "pop"=>1156, "name_en"=>"Santa Clara Pueblo", "lang"=>"en", "lat"=>35.971822, "lon"=>-106.092641),
"/US/US-NM/santa-cruz-13762" => array("type"=>"cdp", "pop"=>2504, "name_ru"=>"Санта Круз", "name_en"=>"Santa Cruz", "lang"=>"en", "lat"=>35.985497, "lon"=>-106.016094),
"/US/US-NM/santa-fe-13763" => array("type"=>"city", "pop"=>71831, "name_en"=>"Santa Fe", "lang"=>"en", "lat"=>35.678502, "lon"=>-105.954149),
"/US/US-NM/santa-rosa-13764" => array("type"=>"city", "pop"=>2642, "name_ru"=>"Санта Роза", "name_en"=>"Santa Rosa", "lang"=>"en", "lat"=>34.943381, "lon"=>-104.676698),
"/US/US-NM/santo-domingo-pueblo-13765" => array("type"=>"cdp", "pop"=>2866, "name_en"=>"Santo Domingo Pueblo", "lang"=>"en", "lat"=>35.515909, "lon"=>-106.365767),
"/US/US-NM/seama-13767" => array("type"=>"cdp", "pop"=>403, "name_en"=>"Seama", "lang"=>"en", "lat"=>35.047433, "lon"=>-107.527463),
"/US/US-NM/shiprock-13768" => array("type"=>"cdp", "pop"=>7687, "name_en"=>"Shiprock", "lang"=>"en", "lat"=>36.791955, "lon"=>-108.695996),
"/US/US-NM/silver-city-13769" => array("type"=>"town", "pop"=>10329, "name_ru"=>"Сильвер Сити", "name_en"=>"Silver City", "lang"=>"en", "lat"=>32.782853, "lon"=>-108.266825),
"/US/US-NM/skyline-ganipa-13770" => array("type"=>"cdp", "pop"=>946, "name_en"=>"Skyline-Ganipa", "lang"=>"en", "lat"=>35.034567, "lon"=>-107.615253),
"/US/US-NM/socorro-13771" => array("type"=>"city", "pop"=>8989, "name_en"=>"Socorro", "lang"=>"en", "lat"=>34.055150, "lon"=>-106.904225),
"/US/US-NM/south-valley-13772" => array("type"=>"cdp", "pop"=>35701, "name_en"=>"South Valley", "lang"=>"en", "lat"=>35.010014, "lon"=>-106.677532),
"/US/US-NM/springer-13773" => array("type"=>"town", "pop"=>1146, "name_en"=>"Springer", "lang"=>"en", "lat"=>36.366246, "lon"=>-104.593128),
"/US/US-NM/sunland-park-13774" => array("type"=>"city", "pop"=>14453, "name_en"=>"Sunland Park", "lang"=>"en", "lat"=>31.799146, "lon"=>-106.577166),
"/US/US-NM/taos-13775" => array("type"=>"town", "pop"=>5551, "name_en"=>"Taos", "lang"=>"en", "lat"=>36.386227, "lon"=>-105.577150),
"/US/US-NM/taos-pueblo-13776" => array("type"=>"cdp", "pop"=>1187, "name_en"=>"Taos Pueblo", "lang"=>"en", "lat"=>36.469277, "lon"=>-105.559506),
"/US/US-NM/tatum-13777" => array("type"=>"town", "pop"=>757, "name_en"=>"Tatum", "lang"=>"en", "lat"=>33.255289, "lon"=>-103.315301),
"/US/US-NM/tesuque-13778" => array("type"=>"cdp", "pop"=>1490, "name_en"=>"Tesuque", "lang"=>"en", "lat"=>35.766550, "lon"=>-105.923330),
"/US/US-NM/texico-13779" => array("type"=>"city", "pop"=>987, "name_en"=>"Texico", "lang"=>"en", "lat"=>34.389376, "lon"=>-103.051721),
"/US/US-NM/tijeras-13780" => array("type"=>"village", "pop"=>520, "name_en"=>"Tijeras", "lang"=>"en", "lat"=>35.088367, "lon"=>-106.374876),
"/US/US-NM/tohatchi-13781" => array("type"=>"cdp", "pop"=>661, "name_en"=>"Tohatchi", "lang"=>"en", "lat"=>35.850247, "lon"=>-108.750068),
"/US/US-NM/tome-adelino-13782" => array("type"=>"cdp", "pop"=>1695, "name_en"=>"Tome-Adelino", "lang"=>"en", "lat"=>34.731455, "lon"=>-106.720662),
"/US/US-NM/truth-or-consequences-13783" => array("type"=>"city", "pop"=>6784, "name_en"=>"Truth or Consequences", "lang"=>"en", "lat"=>33.135586, "lon"=>-107.248069),
"/US/US-NM/tucumcari-13784" => array("type"=>"city", "pop"=>5268, "name_en"=>"Tucumcari", "lang"=>"en", "lat"=>35.167449, "lon"=>-103.727482),
"/US/US-NM/tularosa-13785" => array("type"=>"village", "pop"=>3015, "name_en"=>"Tularosa", "lang"=>"en", "lat"=>33.075854, "lon"=>-106.019743),
"/US/US-NM/university-park-13786" => array("type"=>"cdp", "pop"=>4520, "name_ru"=>"Юниверсити Парк", "name_en"=>"University Park", "lang"=>"en", "lat"=>32.278150, "lon"=>-106.747336),
"/US/US-NM/vadito-13787" => array("type"=>"cdp", "pop"=>283, "name_en"=>"Vadito", "lang"=>"en", "lat"=>36.191077, "lon"=>-105.677104),
"/US/US-NM/valencia-13788" => array("type"=>"cdp", "pop"=>72207, "name_en"=>"Valencia", "lang"=>"en", "lat"=>34.796031, "lon"=>-106.690997),
"/US/US-NM/vaughn-13789" => array("type"=>"town", "pop"=>484, "name_en"=>"Vaughn", "lang"=>"en", "lat"=>34.606558, "lon"=>-105.212285),
"/US/US-NM/virden-13790" => array("type"=>"village", "pop"=>119, "name_ru"=>"Вирден", "name_en"=>"Virden", "lang"=>"en", "lat"=>32.688553, "lon"=>-109.001969),
"/US/US-NM/wagon-mound-13791" => array("type"=>"village", "pop"=>356, "name_en"=>"Wagon Mound", "lang"=>"en", "lat"=>36.002875, "lon"=>-104.714306),
"/US/US-NM/white-rock-13792" => array("type"=>"cdp", "pop"=>6192, "name_en"=>"White Rock", "lang"=>"en", "lat"=>35.803800, "lon"=>-106.210422),
"/US/US-NM/white-sands-13793" => array("type"=>"cdp", "pop"=>2616, "name_en"=>"White Sands", "lang"=>"en", "lat"=>32.382669, "lon"=>-106.492360),
"/US/US-NM/willard-13794" => array("type"=>"village", "pop"=>246, "name_ru"=>"Уиллард", "name_en"=>"Willard", "lang"=>"en", "lat"=>34.594396, "lon"=>-106.033745),
"/US/US-NM/williamsburg-13795" => array("type"=>"village", "pop"=>496, "name_ru"=>"Уилльямсбург", "name_en"=>"Williamsburg", "lang"=>"en", "lat"=>33.116065, "lon"=>-107.294044),
"/US/US-NM/zia-pueblo-13796" => array("type"=>"cdp", "pop"=>637, "name_en"=>"Zia Pueblo", "lang"=>"en", "lat"=>35.513898, "lon"=>-106.725384),
"/US/US-NM/zuni-pueblo-13797" => array("type"=>"cdp", "pop"=>5857, "name_en"=>"Zuni Pueblo", "lang"=>"en", "lat"=>35.072489, "lon"=>-108.849977),
"/US/US-NV" => array("type"=>"state", "pop"=>2600167, "name_ru"=>"Невада", "name_en"=>"Nevada", "lang"=>"en"),
"/US/US-NV/battle-mountain-13009" => array("type"=>"cdp", "pop"=>3542, "name_en"=>"Battle Mountain", "lang"=>"en", "lat"=>40.641522, "lon"=>-116.937721),
"/US/US-NV/beatty-13010" => array("type"=>"cdp", "pop"=>1623, "name_en"=>"Beatty", "lang"=>"en", "lat"=>36.940200, "lon"=>-116.707074),
"/US/US-NV/boulder-city-13011" => array("type"=>"city", "pop"=>14720, "name_en"=>"Boulder City", "lang"=>"en", "lat"=>35.965967, "lon"=>-114.838066),
"/US/US-NV/caliente-13012" => array("type"=>"city", "pop"=>1252, "name_en"=>"Caliente", "lang"=>"en", "lat"=>37.615647, "lon"=>-114.514457),
"/US/US-NV/carlin-13013" => array("type"=>"city", "pop"=>2083, "name_en"=>"Carlin", "lang"=>"en", "lat"=>40.717171, "lon"=>-116.112686),
"/US/US-NV/carson-city-13014" => array("type"=>"city", "pop"=>40443, "name_ru"=>"Карсон Сити", "name_en"=>"Carson City", "lang"=>"en", "lat"=>39.148328, "lon"=>-119.743243),
"/US/US-NV/dayton-13015" => array("type"=>"cdp", "pop"=>2217, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>39.256256, "lon"=>-119.569690),
"/US/US-NV/east-las-vegas-13016" => array("type"=>"cdp", "pop"=>11087, "name_en"=>"East Las Vegas", "lang"=>"en", "lat"=>36.096500, "lon"=>-115.043458),
"/US/US-NV/elko-13017" => array("type"=>"city", "pop"=>17156, "name_ru"=>"Элко", "name_en"=>"Elko", "lang"=>"en", "lat"=>40.838712, "lon"=>-115.760661),
"/US/US-NV/ely-13018" => array("type"=>"city", "pop"=>4049, "name_ru"=>"Или", "name_en"=>"Ely", "lang"=>"en", "lat"=>39.249128, "lon"=>-114.877926),
"/US/US-NV/enterprise-13019" => array("type"=>"cdp", "pop"=>6412, "name_ru"=>"Энтерпрайз", "name_en"=>"Enterprise", "lang"=>"en", "lat"=>36.025400, "lon"=>-115.241191),
"/US/US-NV/fallon-13020" => array("type"=>"city", "pop"=>8525, "name_en"=>"Fallon", "lang"=>"en", "lat"=>39.475792, "lon"=>-118.777827),
"/US/US-NV/fallon-station-13021" => array("type"=>"cdp", "pop"=>1092, "name_en"=>"Fallon Station", "lang"=>"en", "lat"=>39.418564, "lon"=>-118.720141),
"/US/US-NV/fernley-13022" => array("type"=>"cdp", "pop"=>12942, "name_en"=>"Fernley", "lang"=>"en", "lat"=>39.583799, "lon"=>-119.200860),
"/US/US-NV/gabbs-13023" => array("type"=>"city", "pop"=>667, "name_en"=>"Gabbs", "lang"=>"en", "lat"=>38.867700, "lon"=>-117.921526),
"/US/US-NV/gardnerville-13024" => array("type"=>"cdp", "pop"=>2177, "name_en"=>"Gardnerville", "lang"=>"en", "lat"=>38.939294, "lon"=>-119.735853),
"/US/US-NV/gardnerville-ranchos-13025" => array("type"=>"cdp", "pop"=>7455, "name_en"=>"Gardnerville Ranchos", "lang"=>"en", "lat"=>38.888222, "lon"=>-119.740180),
"/US/US-NV/hawthorne-13026" => array("type"=>"cdp", "pop"=>4162, "name_ru"=>"Хоуторн", "name_en"=>"Hawthorne", "lang"=>"en", "lat"=>38.525165, "lon"=>-118.627007),
"/US/US-NV/henderson-13027" => array("type"=>"city", "pop"=>252064, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>36.033669, "lon"=>-115.002364),
"/US/US-NV/incline-village-crystal-bay-13028" => array("type"=>"cdp", "pop"=>7119, "name_en"=>"Incline Village-Crystal Bay", "lang"=>"en", "lat"=>39.267210, "lon"=>-119.963988),
"/US/US-NV/indian-hills-13029" => array("type"=>"cdp", "pop"=>2544, "name_ru"=>"Индиан Хиллс", "name_en"=>"Indian Hills", "lang"=>"en", "lat"=>39.086015, "lon"=>-119.782968),
"/US/US-NV/indian-springs-13030" => array("type"=>"cdp", "pop"=>1164, "name_ru"=>"Индиан Спрингс", "name_en"=>"Indian Springs", "lang"=>"en", "lat"=>36.570757, "lon"=>-115.666550),
"/US/US-NV/johnson-lane-13031" => array("type"=>"cdp", "pop"=>2551, "name_en"=>"Johnson Lane", "lang"=>"en", "lat"=>39.047972, "lon"=>-119.721240),
"/US/US-NV/kingsbury-13032" => array("type"=>"cdp", "pop"=>2238, "name_ru"=>"Кингсбери", "name_en"=>"Kingsbury", "lang"=>"en", "lat"=>38.975872, "lon"=>-119.882069),
"/US/US-NV/las-vegas-13033" => array("type"=>"city", "pop"=>558383, "name_en"=>"Las Vegas", "lang"=>"en", "lat"=>36.205750, "lon"=>-115.222799),
"/US/US-NV/laughlin-13034" => array("type"=>"cdp", "pop"=>4791, "name_en"=>"Laughlin", "lang"=>"en", "lat"=>35.150710, "lon"=>-114.626991),
"/US/US-NV/lovelock-13035" => array("type"=>"city", "pop"=>1854, "name_en"=>"Lovelock", "lang"=>"en", "lat"=>40.179970, "lon"=>-118.476872),
"/US/US-NV/mcdermitt-13036" => array("type"=>"cdp", "pop"=>373, "name_en"=>"McDermitt", "lang"=>"en", "lat"=>41.972877, "lon"=>-117.604585),
"/US/US-NV/mcgill-13037" => array("type"=>"cdp", "pop"=>1258, "name_en"=>"McGill", "lang"=>"en", "lat"=>39.399691, "lon"=>-114.778889),
"/US/US-NV/mesquite-13038" => array("type"=>"city", "pop"=>16444, "name_en"=>"Mesquite", "lang"=>"en", "lat"=>36.802319, "lon"=>-114.107218),
"/US/US-NV/minden-13039" => array("type"=>"cdp", "pop"=>1441, "name_ru"=>"Минден", "name_en"=>"Minden", "lang"=>"en", "lat"=>38.960522, "lon"=>-119.772194),
"/US/US-NV/moapa-valley-13040" => array("type"=>"cdp", "pop"=>3444, "name_en"=>"Moapa Valley", "lang"=>"en", "lat"=>36.580500, "lon"=>-114.469576),
"/US/US-NV/nellis-afb-13041" => array("type"=>"cdp", "pop"=>8377, "name_en"=>"Nellis AFB", "lang"=>"en", "lat"=>36.244400, "lon"=>-115.053819),
"/US/US-NV/new-washoe-city-13042" => array("type"=>"cdp", "pop"=>2875, "name_en"=>"New Washoe City", "lang"=>"en", "lat"=>39.301333, "lon"=>-119.771787),
"/US/US-NV/north-las-vegas-13043" => array("type"=>"city", "pop"=>217253, "name_en"=>"North Las Vegas", "lang"=>"en", "lat"=>36.267831, "lon"=>-115.144219),
"/US/US-NV/owyhee-13044" => array("type"=>"cdp", "pop"=>908, "name_en"=>"Owyhee", "lang"=>"en", "lat"=>41.914438, "lon"=>-116.187879),
"/US/US-NV/pahrump-13045" => array("type"=>"cdp", "pop"=>7424, "name_en"=>"Pahrump", "lang"=>"en", "lat"=>36.252813, "lon"=>-116.018069),
"/US/US-NV/paradise-13046" => array("type"=>"cdp", "pop"=>124682, "name_ru"=>"Парадайс", "name_en"=>"Paradise", "lang"=>"en", "lat"=>36.080825, "lon"=>-115.133602),
"/US/US-NV/reno-13047" => array("type"=>"city", "pop"=>217016, "name_en"=>"Reno", "lang"=>"en", "lat"=>39.538700, "lon"=>-119.822499),
"/US/US-NV/schurz-13048" => array("type"=>"cdp", "pop"=>617, "name_en"=>"Schurz", "lang"=>"en", "lat"=>38.992860, "lon"=>-118.832820),
"/US/US-NV/silver-springs-13049" => array("type"=>"cdp", "pop"=>2253, "name_en"=>"Silver Springs", "lang"=>"en", "lat"=>39.380718, "lon"=>-119.214014),
"/US/US-NV/smith-valley-13050" => array("type"=>"cdp", "pop"=>1033, "name_en"=>"Smith Valley", "lang"=>"en", "lat"=>38.786118, "lon"=>-119.340789),
"/US/US-NV/sparks-13051" => array("type"=>"city", "pop"=>88602, "name_ru"=>"Спаркс", "name_en"=>"Sparks", "lang"=>"en", "lat"=>39.544244, "lon"=>-119.736931),
"/US/US-NV/spring-creek-13052" => array("type"=>"cdp", "pop"=>5866, "name_en"=>"Spring Creek", "lang"=>"en", "lat"=>40.738498, "lon"=>-115.595872),
"/US/US-NV/spring-valley-13053" => array("type"=>"cdp", "pop"=>51726, "name_ru"=>"Спринг Валли", "name_en"=>"Spring Valley", "lang"=>"en", "lat"=>36.108000, "lon"=>-115.244219),
"/US/US-NV/stateline-13054" => array("type"=>"cdp", "pop"=>1379, "name_en"=>"Stateline", "lang"=>"en", "lat"=>38.968655, "lon"=>-119.943085),
"/US/US-NV/sun-valley-13056" => array("type"=>"cdp", "pop"=>11391, "name_ru"=>"Сан Вэлли", "name_en"=>"Sun Valley", "lang"=>"en", "lat"=>39.598174, "lon"=>-119.778596),
"/US/US-NV/sunrise-manor-13055" => array("type"=>"cdp", "pop"=>95362, "name_en"=>"Sunrise Manor", "lang"=>"en", "lat"=>36.187400, "lon"=>-115.049985),
"/US/US-NV/tonopah-13057" => array("type"=>"cdp", "pop"=>3616, "name_en"=>"Tonopah", "lang"=>"en", "lat"=>38.097760, "lon"=>-117.247735),
"/US/US-NV/wadsworth-13058" => array("type"=>"cdp", "pop"=>640, "name_ru"=>"Уодсворт", "name_en"=>"Wadsworth", "lang"=>"en", "lat"=>39.632628, "lon"=>-119.290037),
"/US/US-NV/wells-13059" => array("type"=>"city", "pop"=>1297, "name_ru"=>"Уэллс", "name_en"=>"Wells", "lang"=>"en", "lat"=>41.114407, "lon"=>-114.948113),
"/US/US-NV/west-wendover-13060" => array("type"=>"cdp", "pop"=>4997, "name_en"=>"West Wendover", "lang"=>"en", "lat"=>40.767708, "lon"=>-114.115543),
"/US/US-NV/winchester-13061" => array("type"=>"cdp", "pop"=>23365, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>36.136600, "lon"=>-115.129789),
"/US/US-NV/winnemucca-13062" => array("type"=>"city", "pop"=>8085, "name_en"=>"Winnemucca", "lang"=>"en", "lat"=>40.968285, "lon"=>-117.725732),
"/US/US-NV/yerington-13063" => array("type"=>"city", "pop"=>3856, "name_en"=>"Yerington", "lang"=>"en", "lat"=>38.988090, "lon"=>-119.163066),
"/US/US-NV/zephyr-cove-round-hill-village-13064" => array("type"=>"cdp", "pop"=>1434, "name_en"=>"Zephyr Cove-Round Hill Village", "lang"=>"en", "lat"=>38.994484, "lon"=>-119.916294),
"/US/US-NY" => array("type"=>"state", "pop"=>19490297, "name_ru"=>"Нью-йорк", "name_en"=>"New York", "lang"=>"en"),
"/US/US-NY/adams-13798" => array("type"=>"village", "pop"=>1660, "name_ru"=>"Адамс", "name_en"=>"Adams", "lang"=>"en", "lat"=>43.809801, "lon"=>-76.022546),
"/US/US-NY/adams-center-13799" => array("type"=>"cdp", "pop"=>1675, "name_en"=>"Adams Center", "lang"=>"en", "lat"=>43.859625, "lon"=>-75.997624),
"/US/US-NY/addison-13800" => array("type"=>"village", "pop"=>1713, "name_ru"=>"Эддиссон", "name_en"=>"Addison", "lang"=>"en", "lat"=>42.105829, "lon"=>-77.233593),
"/US/US-NY/afton-13801" => array("type"=>"village", "pop"=>811, "name_ru"=>"Афтон", "name_en"=>"Afton", "lang"=>"en", "lat"=>42.229100, "lon"=>-75.525154),
"/US/US-NY/airmont-13802" => array("type"=>"cdp", "pop"=>8339, "name_en"=>"Airmont", "lang"=>"en", "lat"=>41.099067, "lon"=>-74.100424),
"/US/US-NY/akron-13803" => array("type"=>"village", "pop"=>2968, "name_ru"=>"Акрон", "name_en"=>"Akron", "lang"=>"en", "lat"=>43.019300, "lon"=>-78.495624),
"/US/US-NY/albany-13804" => array("type"=>"city", "pop"=>93539, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>42.665750, "lon"=>-73.799017),
"/US/US-NY/albertson-13805" => array("type"=>"cdp", "pop"=>5166, "name_en"=>"Albertson", "lang"=>"en", "lat"=>40.771183, "lon"=>-73.649253),
"/US/US-NY/albion-13806" => array("type"=>"village", "pop"=>5548, "name_ru"=>"Албион", "name_en"=>"Albion", "lang"=>"en", "lat"=>43.245884, "lon"=>-78.189347),
"/US/US-NY/alden-13807" => array("type"=>"village", "pop"=>2529, "name_ru"=>"Олден", "name_en"=>"Alden", "lang"=>"en", "lat"=>42.898350, "lon"=>-78.494524),
"/US/US-NY/alexander-13808" => array("type"=>"village", "pop"=>481, "name_ru"=>"Александр", "name_en"=>"Alexander", "lang"=>"en", "lat"=>42.901716, "lon"=>-78.260664),
"/US/US-NY/alexandria-bay-13809" => array("type"=>"village", "pop"=>1115, "name_en"=>"Alexandria Bay", "lang"=>"en", "lat"=>44.337417, "lon"=>-75.917670),
"/US/US-NY/alfred-13810" => array("type"=>"village", "pop"=>4843, "name_en"=>"Alfred", "lang"=>"en", "lat"=>42.254101, "lon"=>-77.789789),
"/US/US-NY/allegany-13811" => array("type"=>"village", "pop"=>1760, "name_en"=>"Allegany", "lang"=>"en", "lat"=>42.090058, "lon"=>-78.491513),
"/US/US-NY/almond-13812" => array("type"=>"village", "pop"=>435, "name_en"=>"Almond", "lang"=>"en", "lat"=>42.318985, "lon"=>-77.739544),
"/US/US-NY/altamont-13813" => array("type"=>"village", "pop"=>1701, "name_ru"=>"Альтамонт", "name_en"=>"Altamont", "lang"=>"en", "lat"=>42.705083, "lon"=>-74.036297),
"/US/US-NY/altmar-13814" => array("type"=>"village", "pop"=>348, "name_en"=>"Altmar", "lang"=>"en", "lat"=>43.510560, "lon"=>-76.004989),
"/US/US-NY/altona-13815" => array("type"=>"cdp", "pop"=>2918, "name_ru"=>"Альтона", "name_en"=>"Altona", "lang"=>"en", "lat"=>44.892200, "lon"=>-73.659270),
"/US/US-NY/amenia-13816" => array("type"=>"cdp", "pop"=>4113, "name_en"=>"Amenia", "lang"=>"en", "lat"=>41.846165, "lon"=>-73.552401),
"/US/US-NY/ames-13817" => array("type"=>"village", "pop"=>165, "name_ru"=>"Эймс", "name_en"=>"Ames", "lang"=>"en", "lat"=>42.835651, "lon"=>-74.602996),
"/US/US-NY/amityville-13818" => array("type"=>"village", "pop"=>9920, "name_en"=>"Amityville", "lang"=>"en", "lat"=>40.669850, "lon"=>-73.416115),
"/US/US-NY/amsterdam-13819" => array("type"=>"city", "pop"=>17533, "name_ru"=>"Амсердам", "name_en"=>"Amsterdam", "lang"=>"en", "lat"=>42.941772, "lon"=>-74.190854),
"/US/US-NY/andes-13820" => array("type"=>"village", "pop"=>1291, "name_en"=>"Andes", "lang"=>"en", "lat"=>42.190021, "lon"=>-74.784169),
"/US/US-NY/andover-13821" => array("type"=>"village", "pop"=>1011, "name_ru"=>"Андоувер", "name_en"=>"Andover", "lang"=>"en", "lat"=>42.157042, "lon"=>-77.796513),
"/US/US-NY/angelica-13822" => array("type"=>"village", "pop"=>856, "name_en"=>"Angelica", "lang"=>"en", "lat"=>42.304270, "lon"=>-78.021096),
"/US/US-NY/angola-13823" => array("type"=>"village", "pop"=>2134, "name_ru"=>"Ангола", "name_en"=>"Angola", "lang"=>"en", "lat"=>42.637650, "lon"=>-79.029797),
"/US/US-NY/angola-on-the-lake-13824" => array("type"=>"cdp", "pop"=>1719, "name_en"=>"Angola on the Lake", "lang"=>"en", "lat"=>42.654750, "lon"=>-79.049141),
"/US/US-NY/antwerp-13825" => array("type"=>"village", "pop"=>732, "name_en"=>"Antwerp", "lang"=>"en", "lat"=>44.199742, "lon"=>-75.609000),
"/US/US-NY/apalachin-13826" => array("type"=>"cdp", "pop"=>1208, "name_en"=>"Apalachin", "lang"=>"en", "lat"=>42.072330, "lon"=>-76.165887),
"/US/US-NY/aquebogue-13827" => array("type"=>"cdp", "pop"=>2060, "name_en"=>"Aquebogue", "lang"=>"en", "lat"=>40.938297, "lon"=>-72.614096),
"/US/US-NY/arcade-13828" => array("type"=>"village", "pop"=>1888, "name_ru"=>"Аркада", "name_en"=>"Arcade", "lang"=>"en", "lat"=>42.533429, "lon"=>-78.433156),
"/US/US-NY/ardsley-13829" => array("type"=>"village", "pop"=>4840, "name_en"=>"Ardsley", "lang"=>"en", "lat"=>41.014600, "lon"=>-73.841236),
"/US/US-NY/argyle-13830" => array("type"=>"village", "pop"=>278, "name_ru"=>"Аргайл", "name_en"=>"Argyle", "lang"=>"en", "lat"=>43.237015, "lon"=>-73.490371),
"/US/US-NY/arkport-13831" => array("type"=>"village", "pop"=>808, "name_en"=>"Arkport", "lang"=>"en", "lat"=>42.391940, "lon"=>-77.696489),
"/US/US-NY/arlington-13832" => array("type"=>"cdp", "pop"=>11948, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>41.695650, "lon"=>-73.884934),
"/US/US-NY/armonk-13833" => array("type"=>"cdp", "pop"=>2745, "name_en"=>"Armonk", "lang"=>"en", "lat"=>41.132292, "lon"=>-73.712815),
"/US/US-NY/asharoken-13834" => array("type"=>"village", "pop"=>660, "name_en"=>"Asharoken", "lang"=>"en", "lat"=>40.944750, "lon"=>-73.391185),
"/US/US-NY/athens-13835" => array("type"=>"village", "pop"=>1694, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>42.267091, "lon"=>-73.812900),
"/US/US-NY/atlantic-beach-13836" => array("type"=>"village", "pop"=>1998, "name_ru"=>"Атлантик Бич", "name_en"=>"Atlantic Beach", "lang"=>"en", "lat"=>40.588000, "lon"=>-73.729699),
"/US/US-NY/attica-13837" => array("type"=>"village", "pop"=>2412, "name_ru"=>"Аттика", "name_en"=>"Attica", "lang"=>"en", "lat"=>42.865478, "lon"=>-78.280111),
"/US/US-NY/auburn-13838" => array("type"=>"city", "pop"=>27138, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>42.933246, "lon"=>-76.568946),
"/US/US-NY/aurora-13839" => array("type"=>"village", "pop"=>872, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>42.752184, "lon"=>-76.698872),
"/US/US-NY/averill-park-13840" => array("type"=>"cdp", "pop"=>1656, "name_en"=>"Averill Park", "lang"=>"en", "lat"=>42.635004, "lon"=>-73.553357),
"/US/US-NY/avoca-13841" => array("type"=>"village", "pop"=>960, "name_ru"=>"Авока", "name_en"=>"Avoca", "lang"=>"en", "lat"=>42.408342, "lon"=>-77.423174),
"/US/US-NY/avon-13842" => array("type"=>"village", "pop"=>2911, "name_ru"=>"Эйвон", "name_en"=>"Avon", "lang"=>"en", "lat"=>42.912305, "lon"=>-77.746138),
"/US/US-NY/babylon-13843" => array("type"=>"village", "pop"=>12760, "name_en"=>"Babylon", "lang"=>"en", "lat"=>40.695950, "lon"=>-73.327149),
"/US/US-NY/bainbridge-13844" => array("type"=>"village", "pop"=>1322, "name_ru"=>"Бэйнбридж", "name_en"=>"Bainbridge", "lang"=>"en", "lat"=>42.296018, "lon"=>-75.480401),
"/US/US-NY/baldwin-13845" => array("type"=>"cdp", "pop"=>839, "name_ru"=>"Болдуин", "name_en"=>"Baldwin", "lang"=>"en", "lat"=>40.662983, "lon"=>-73.611216),
"/US/US-NY/baldwin-harbor-13846" => array("type"=>"cdp", "pop"=>7899, "name_en"=>"Baldwin Harbor", "lang"=>"en", "lat"=>40.629950, "lon"=>-73.602256),
"/US/US-NY/baldwinsville-13847" => array("type"=>"village", "pop"=>7247, "name_en"=>"Baldwinsville", "lang"=>"en", "lat"=>43.156400, "lon"=>-76.333813),
"/US/US-NY/ballston-spa-13848" => array("type"=>"village", "pop"=>5464, "name_en"=>"Ballston Spa", "lang"=>"en", "lat"=>43.005974, "lon"=>-73.851482),
"/US/US-NY/balmville-13849" => array("type"=>"cdp", "pop"=>2963, "name_en"=>"Balmville", "lang"=>"en", "lat"=>41.526800, "lon"=>-74.023452),
"/US/US-NY/bardonia-13850" => array("type"=>"cdp", "pop"=>4487, "name_en"=>"Bardonia", "lang"=>"en", "lat"=>41.111237, "lon"=>-73.981370),
"/US/US-NY/barker-13851" => array("type"=>"village", "pop"=>540, "name_en"=>"Barker", "lang"=>"en", "lat"=>43.327410, "lon"=>-78.553841),
"/US/US-NY/barneveld-13852" => array("type"=>"village", "pop"=>323, "name_en"=>"Barneveld", "lang"=>"en", "lat"=>43.274018, "lon"=>-75.189101),
"/US/US-NY/barnum-island-13853" => array("type"=>"cdp", "pop"=>2624, "name_en"=>"Barnum Island", "lang"=>"en", "lat"=>40.604506, "lon"=>-73.644353),
"/US/US-NY/batavia-13854" => array("type"=>"city", "pop"=>15202, "name_ru"=>"Батавия", "name_en"=>"Batavia", "lang"=>"en", "lat"=>42.998613, "lon"=>-78.181153),
"/US/US-NY/bath-13855" => array("type"=>"village", "pop"=>5432, "name_ru"=>"Бат", "name_en"=>"Bath", "lang"=>"en", "lat"=>42.336996, "lon"=>-77.318430),
"/US/US-NY/baxter-estates-13856" => array("type"=>"village", "pop"=>1000, "name_en"=>"Baxter Estates", "lang"=>"en", "lat"=>40.834133, "lon"=>-73.696178),
"/US/US-NY/bay-shore-13858" => array("type"=>"cdp", "pop"=>21279, "name_en"=>"Bay Shore", "lang"=>"en", "lat"=>40.726366, "lon"=>-73.250503),
"/US/US-NY/bayport-13857" => array("type"=>"cdp", "pop"=>7702, "name_ru"=>"Бэйпорт", "name_en"=>"Bayport", "lang"=>"en", "lat"=>40.747760, "lon"=>-73.054662),
"/US/US-NY/bayville-13859" => array("type"=>"village", "pop"=>7202, "name_en"=>"Bayville", "lang"=>"en", "lat"=>40.906950, "lon"=>-73.563149),
"/US/US-NY/baywood-13860" => array("type"=>"cdp", "pop"=>7351, "name_en"=>"Baywood", "lang"=>"en", "lat"=>40.751211, "lon"=>-73.291071),
"/US/US-NY/beacon-13861" => array("type"=>"city", "pop"=>14566, "name_ru"=>"Бикон", "name_en"=>"Beacon", "lang"=>"en", "lat"=>41.504150, "lon"=>-73.965993),
"/US/US-NY/beaverdam-lake-salisbury-mills-13862" => array("type"=>"cdp", "pop"=>2354, "name_en"=>"Beaverdam Lake-Salisbury Mills", "lang"=>"en", "lat"=>41.444464, "lon"=>-74.117974),
"/US/US-NY/bedford-13863" => array("type"=>"cdp", "pop"=>18585, "name_ru"=>"Бедфорд", "name_en"=>"Bedford", "lang"=>"en", "lat"=>41.196769, "lon"=>-73.647188),
"/US/US-NY/belle-terre-13865" => array("type"=>"village", "pop"=>844, "name_en"=>"Belle Terre", "lang"=>"en", "lat"=>40.962256, "lon"=>-73.067200),
"/US/US-NY/bellerose-13864" => array("type"=>"village", "pop"=>1159, "name_en"=>"Bellerose", "lang"=>"en", "lat"=>40.724300, "lon"=>-73.716728),
"/US/US-NY/bellmore-13866" => array("type"=>"cdp", "pop"=>16438, "name_en"=>"Bellmore", "lang"=>"en", "lat"=>40.657067, "lon"=>-73.529172),
"/US/US-NY/bellport-13867" => array("type"=>"village", "pop"=>2382, "name_en"=>"Bellport", "lang"=>"en", "lat"=>40.756936, "lon"=>-72.942053),
"/US/US-NY/belmont-13868" => array("type"=>"village", "pop"=>857, "name_ru"=>"Белмонт", "name_en"=>"Belmont", "lang"=>"en", "lat"=>42.222447, "lon"=>-78.033588),
"/US/US-NY/bemus-point-13869" => array("type"=>"village", "pop"=>334, "name_en"=>"Bemus Point", "lang"=>"en", "lat"=>42.163403, "lon"=>-79.391156),
"/US/US-NY/bergen-13870" => array("type"=>"village", "pop"=>1153, "name_en"=>"Bergen", "lang"=>"en", "lat"=>43.083328, "lon"=>-77.942021),
"/US/US-NY/bethpage-13871" => array("type"=>"cdp", "pop"=>15761, "name_en"=>"Bethpage", "lang"=>"en", "lat"=>40.749450, "lon"=>-73.485799),
"/US/US-NY/big-flats-13872" => array("type"=>"cdp", "pop"=>7520, "name_en"=>"Big Flats", "lang"=>"en", "lat"=>42.142334, "lon"=>-76.930449),
"/US/US-NY/big-flats-airport-13873" => array("type"=>"cdp", "pop"=>2248, "name_en"=>"Big Flats Airport", "lang"=>"en", "lat"=>42.171012, "lon"=>-76.889834),
"/US/US-NY/billington-heights-13874" => array("type"=>"cdp", "pop"=>1729, "name_en"=>"Billington Heights", "lang"=>"en", "lat"=>42.784300, "lon"=>-78.626652),
"/US/US-NY/binghamton-13875" => array("type"=>"city", "pop"=>44729, "name_en"=>"Binghamton", "lang"=>"en", "lat"=>42.101550, "lon"=>-75.909518),
"/US/US-NY/black-river-13876" => array("type"=>"village", "pop"=>1363, "name_en"=>"Black River", "lang"=>"en", "lat"=>44.009554, "lon"=>-75.796756),
"/US/US-NY/blasdell-13877" => array("type"=>"village", "pop"=>2504, "name_en"=>"Blasdell", "lang"=>"en", "lat"=>42.796233, "lon"=>-78.833360),
"/US/US-NY/blauvelt-13878" => array("type"=>"cdp", "pop"=>4838, "name_en"=>"Blauvelt", "lang"=>"en", "lat"=>41.068050, "lon"=>-73.955251),
"/US/US-NY/bloomingburg-13879" => array("type"=>"village", "pop"=>329, "name_en"=>"Bloomingburg", "lang"=>"en", "lat"=>41.556066, "lon"=>-74.441459),
"/US/US-NY/blue-point-13880" => array("type"=>"cdp", "pop"=>4230, "name_en"=>"Blue Point", "lang"=>"en", "lat"=>40.750276, "lon"=>-73.035027),
"/US/US-NY/bohemia-13881" => array("type"=>"cdp", "pop"=>9556, "name_en"=>"Bohemia", "lang"=>"en", "lat"=>40.774000, "lon"=>-73.121315),
"/US/US-NY/bolivar-13882" => array("type"=>"village", "pop"=>1113, "name_ru"=>"Боливар", "name_en"=>"Bolivar", "lang"=>"en", "lat"=>42.070257, "lon"=>-78.166710),
"/US/US-NY/boonville-13883" => array("type"=>"village", "pop"=>2042, "name_ru"=>"Бунвилл", "name_en"=>"Boonville", "lang"=>"en", "lat"=>43.480846, "lon"=>-75.330103),
"/US/US-NY/brasher-falls-winthrop-13884" => array("type"=>"cdp", "pop"=>1271, "name_en"=>"Brasher Falls-Winthrop", "lang"=>"en", "lat"=>44.804918, "lon"=>-74.793710),
"/US/US-NY/brentwood-13885" => array("type"=>"cdp", "pop"=>45218, "name_ru"=>"Брентвуд", "name_en"=>"Brentwood", "lang"=>"en", "lat"=>40.781352, "lon"=>-73.248451),
"/US/US-NY/brewerton-13886" => array("type"=>"cdp", "pop"=>2954, "name_en"=>"Brewerton", "lang"=>"en", "lat"=>43.235536, "lon"=>-76.141750),
"/US/US-NY/brewster-13887" => array("type"=>"village", "pop"=>2112, "name_ru"=>"Брюстер", "name_en"=>"Brewster", "lang"=>"en", "lat"=>41.396820, "lon"=>-73.615104),
"/US/US-NY/brewster-hill-13888" => array("type"=>"cdp", "pop"=>2226, "name_en"=>"Brewster Hill", "lang"=>"en", "lat"=>41.423782, "lon"=>-73.604836),
"/US/US-NY/briarcliff-manor-13889" => array("type"=>"village", "pop"=>7986, "name_en"=>"Briarcliff Manor", "lang"=>"en", "lat"=>41.138791, "lon"=>-73.842153),
"/US/US-NY/bridgehampton-13890" => array("type"=>"cdp", "pop"=>1997, "name_en"=>"Bridgehampton", "lang"=>"en", "lat"=>40.943392, "lon"=>-72.299085),
"/US/US-NY/bridgeport-13891" => array("type"=>"cdp", "pop"=>2107, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>43.154405, "lon"=>-75.975248),
"/US/US-NY/bridgewater-13892" => array("type"=>"village", "pop"=>561, "name_ru"=>"Бриджуотер", "name_en"=>"Bridgewater", "lang"=>"en", "lat"=>42.877663, "lon"=>-75.250227),
"/US/US-NY/brighton-13893" => array("type"=>"cdp", "pop"=>1539, "name_ru"=>"Брайтон", "name_en"=>"Brighton", "lang"=>"en", "lat"=>43.118590, "lon"=>-77.584391),
"/US/US-NY/brightwaters-13894" => array("type"=>"village", "pop"=>3278, "name_en"=>"Brightwaters", "lang"=>"en", "lat"=>40.718008, "lon"=>-73.265895),
"/US/US-NY/brinckerhoff-13895" => array("type"=>"cdp", "pop"=>2756, "name_en"=>"Brinckerhoff", "lang"=>"en", "lat"=>41.550300, "lon"=>-73.869922),
"/US/US-NY/broadalbin-13896" => array("type"=>"village", "pop"=>1397, "name_en"=>"Broadalbin", "lang"=>"en", "lat"=>43.059544, "lon"=>-74.197743),
"/US/US-NY/brockport-13897" => array("type"=>"village", "pop"=>8342, "name_en"=>"Brockport", "lang"=>"en", "lat"=>43.213351, "lon"=>-77.940987),
"/US/US-NY/brocton-13898" => array("type"=>"village", "pop"=>1442, "name_ru"=>"Броктон", "name_en"=>"Brocton", "lang"=>"en", "lat"=>42.389191, "lon"=>-79.442274),
"/US/US-NY/bronxville-13899" => array("type"=>"village", "pop"=>6516, "name_en"=>"Bronxville", "lang"=>"en", "lat"=>40.938850, "lon"=>-73.827313),
"/US/US-NY/brookhaven-13900" => array("type"=>"cdp", "pop"=>488800, "name_ru"=>"Брукхейвен", "name_en"=>"Brookhaven", "lang"=>"en", "lat"=>40.780248, "lon"=>-72.913011),
"/US/US-NY/brookville-13901" => array("type"=>"village", "pop"=>3874, "name_ru"=>"Бруквилл", "name_en"=>"Brookville", "lang"=>"en", "lat"=>40.807500, "lon"=>-73.570266),
"/US/US-NY/brownville-13902" => array("type"=>"village", "pop"=>1051, "name_en"=>"Brownville", "lang"=>"en", "lat"=>44.005431, "lon"=>-75.983552),
"/US/US-NY/brushton-13903" => array("type"=>"village", "pop"=>461, "name_en"=>"Brushton", "lang"=>"en", "lat"=>44.830867, "lon"=>-74.512937),
"/US/US-NY/buchanan-13904" => array("type"=>"village", "pop"=>2245, "name_ru"=>"Бачанан", "name_en"=>"Buchanan", "lang"=>"en", "lat"=>41.263225, "lon"=>-73.944853),
"/US/US-NY/buffalo-13905" => array("type"=>"city", "pop"=>270919, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>42.889800, "lon"=>-78.859684),
"/US/US-NY/burdett-13906" => array("type"=>"village", "pop"=>335, "name_ru"=>"Бердетт", "name_en"=>"Burdett", "lang"=>"en", "lat"=>42.417144, "lon"=>-76.847297),
"/US/US-NY/burke-13907" => array("type"=>"village", "pop"=>205, "name_en"=>"Burke", "lang"=>"en", "lat"=>44.903356, "lon"=>-74.170885),
"/US/US-NY/cairo-13908" => array("type"=>"cdp", "pop"=>6508, "name_ru"=>"Каир", "name_en"=>"Cairo", "lang"=>"en", "lat"=>42.305436, "lon"=>-74.009903),
"/US/US-NY/calcium-13909" => array("type"=>"cdp", "pop"=>2465, "name_en"=>"Calcium", "lang"=>"en", "lat"=>44.034697, "lon"=>-75.847975),
"/US/US-NY/caledonia-13910" => array("type"=>"village", "pop"=>2143, "name_ru"=>"Каледония", "name_en"=>"Caledonia", "lang"=>"en", "lat"=>42.975648, "lon"=>-77.857415),
"/US/US-NY/calverton-13911" => array("type"=>"cdp", "pop"=>4759, "name_ru"=>"Калвертон", "name_en"=>"Calverton", "lang"=>"en", "lat"=>40.923912, "lon"=>-72.761895),
"/US/US-NY/cambridge-13912" => array("type"=>"village", "pop"=>1819, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>43.026639, "lon"=>-73.380853),
"/US/US-NY/camden-13913" => array("type"=>"village", "pop"=>2252, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>43.336287, "lon"=>-75.747636),
"/US/US-NY/camillus-13914" => array("type"=>"village", "pop"=>1175, "name_en"=>"Camillus", "lang"=>"en", "lat"=>43.039300, "lon"=>-76.308164),
"/US/US-NY/canajoharie-13915" => array("type"=>"village", "pop"=>2170, "name_en"=>"Canajoharie", "lang"=>"en", "lat"=>42.903013, "lon"=>-74.571518),
"/US/US-NY/canandaigua-13916" => array("type"=>"city", "pop"=>11168, "name_en"=>"Canandaigua", "lang"=>"en", "lat"=>42.888996, "lon"=>-77.280679),
"/US/US-NY/canaseraga-13917" => array("type"=>"village", "pop"=>560, "name_en"=>"Canaseraga", "lang"=>"en", "lat"=>42.463974, "lon"=>-77.776864),
"/US/US-NY/canastota-13918" => array("type"=>"village", "pop"=>4308, "name_en"=>"Canastota", "lang"=>"en", "lat"=>43.084334, "lon"=>-75.755689),
"/US/US-NY/candor-13919" => array("type"=>"village", "pop"=>794, "name_en"=>"Candor", "lang"=>"en", "lat"=>42.227256, "lon"=>-76.337623),
"/US/US-NY/canisteo-13920" => array("type"=>"village", "pop"=>2221, "name_en"=>"Canisteo", "lang"=>"en", "lat"=>42.270303, "lon"=>-77.605064),
"/US/US-NY/canton-13921" => array("type"=>"village", "pop"=>6077, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>44.596588, "lon"=>-75.172537),
"/US/US-NY/cape-vincent-13922" => array("type"=>"village", "pop"=>782, "name_en"=>"Cape Vincent", "lang"=>"en", "lat"=>44.126258, "lon"=>-76.330972),
"/US/US-NY/carle-place-13923" => array("type"=>"cdp", "pop"=>5107, "name_en"=>"Carle Place", "lang"=>"en", "lat"=>40.750967, "lon"=>-73.611532),
"/US/US-NY/carmel-hamlet-13924" => array("type"=>"cdp", "pop"=>4800, "name_en"=>"Carmel Hamlet", "lang"=>"en", "lat"=>41.414547, "lon"=>-73.683758),
"/US/US-NY/carthage-13925" => array("type"=>"village", "pop"=>3744, "name_ru"=>"Картаж", "name_en"=>"Carthage", "lang"=>"en", "lat"=>43.984780, "lon"=>-75.604637),
"/US/US-NY/cassadaga-13926" => array("type"=>"village", "pop"=>630, "name_en"=>"Cassadaga", "lang"=>"en", "lat"=>42.341022, "lon"=>-79.317121),
"/US/US-NY/castile-13927" => array("type"=>"village", "pop"=>977, "name_en"=>"Castile", "lang"=>"en", "lat"=>42.630996, "lon"=>-78.052569),
"/US/US-NY/castleton-on-hudson-13928" => array("type"=>"village", "pop"=>1547, "name_en"=>"Castleton-on-Hudson", "lang"=>"en", "lat"=>42.532297, "lon"=>-73.750775),
"/US/US-NY/castorland-13929" => array("type"=>"village", "pop"=>285, "name_en"=>"Castorland", "lang"=>"en", "lat"=>43.884249, "lon"=>-75.517648),
"/US/US-NY/cato-13930" => array("type"=>"village", "pop"=>574, "name_en"=>"Cato", "lang"=>"en", "lat"=>43.168033, "lon"=>-76.572447),
"/US/US-NY/catskill-13931" => array("type"=>"village", "pop"=>4212, "name_en"=>"Catskill", "lang"=>"en", "lat"=>42.214771, "lon"=>-73.865492),
"/US/US-NY/cattaraugus-13932" => array("type"=>"village", "pop"=>984, "name_en"=>"Cattaraugus", "lang"=>"en", "lat"=>42.329686, "lon"=>-78.867854),
"/US/US-NY/cayuga-13933" => array("type"=>"village", "pop"=>488, "name_ru"=>"Каюга", "name_en"=>"Cayuga", "lang"=>"en", "lat"=>42.919449, "lon"=>-76.728349),
"/US/US-NY/cayuga-heights-13934" => array("type"=>"village", "pop"=>3670, "name_en"=>"Cayuga Heights", "lang"=>"en", "lat"=>42.469072, "lon"=>-76.486854),
"/US/US-NY/cazenovia-13935" => array("type"=>"village", "pop"=>2913, "name_en"=>"Cazenovia", "lang"=>"en", "lat"=>42.929383, "lon"=>-75.852193),
"/US/US-NY/cedarhurst-13936" => array("type"=>"village", "pop"=>6142, "name_en"=>"Cedarhurst", "lang"=>"en", "lat"=>40.625500, "lon"=>-73.727889),
"/US/US-NY/celoron-13937" => array("type"=>"village", "pop"=>1193, "name_en"=>"Celoron", "lang"=>"en", "lat"=>42.104833, "lon"=>-79.276057),
"/US/US-NY/center-moriches-13939" => array("type"=>"cdp", "pop"=>5987, "name_en"=>"Center Moriches", "lang"=>"en", "lat"=>40.799152, "lon"=>-72.796802),
"/US/US-NY/centereach-13938" => array("type"=>"cdp", "pop"=>26720, "name_en"=>"Centereach", "lang"=>"en", "lat"=>40.870776, "lon"=>-73.083275),
"/US/US-NY/centerport-13940" => array("type"=>"cdp", "pop"=>5333, "name_en"=>"Centerport", "lang"=>"en", "lat"=>40.900818, "lon"=>-73.372195),
"/US/US-NY/central-islip-13941" => array("type"=>"cdp", "pop"=>26028, "name_en"=>"Central Islip", "lang"=>"en", "lat"=>40.785416, "lon"=>-73.196934),
"/US/US-NY/central-square-13942" => array("type"=>"village", "pop"=>1727, "name_en"=>"Central Square", "lang"=>"en", "lat"=>43.286720, "lon"=>-76.140773),
"/US/US-NY/central-valley-13943" => array("type"=>"cdp", "pop"=>1929, "name_ru"=>"Сентрал Валли", "name_en"=>"Central Valley", "lang"=>"en", "lat"=>41.328919, "lon"=>-74.127261),
"/US/US-NY/centre-island-13944" => array("type"=>"village", "pop"=>445, "name_en"=>"Centre Island", "lang"=>"en", "lat"=>40.900167, "lon"=>-73.522092),
"/US/US-NY/champlain-13945" => array("type"=>"village", "pop"=>1132, "name_en"=>"Champlain", "lang"=>"en", "lat"=>44.987308, "lon"=>-73.444577),
"/US/US-NY/chateaugay-13946" => array("type"=>"village", "pop"=>772, "name_en"=>"Chateaugay", "lang"=>"en", "lat"=>44.927307, "lon"=>-74.080788),
"/US/US-NY/chatham-13947" => array("type"=>"village", "pop"=>1682, "name_ru"=>"Чатем", "name_en"=>"Chatham", "lang"=>"en", "lat"=>42.361324, "lon"=>-73.597928),
"/US/US-NY/chaumont-13948" => array("type"=>"village", "pop"=>612, "name_en"=>"Chaumont", "lang"=>"en", "lat"=>44.065295, "lon"=>-76.133356),
"/US/US-NY/cheektowaga-13949" => array("type"=>"cdp", "pop"=>87788, "name_en"=>"Cheektowaga", "lang"=>"en", "lat"=>42.908950, "lon"=>-78.750219),
"/US/US-NY/cherry-creek-13950" => array("type"=>"village", "pop"=>513, "name_en"=>"Cherry Creek", "lang"=>"en", "lat"=>42.295099, "lon"=>-79.101663),
"/US/US-NY/cherry-valley-13951" => array("type"=>"village", "pop"=>548, "name_ru"=>"Черри Валли", "name_en"=>"Cherry Valley", "lang"=>"en", "lat"=>42.797916, "lon"=>-74.752513),
"/US/US-NY/chester-13952" => array("type"=>"village", "pop"=>3576, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>41.357171, "lon"=>-74.276229),
"/US/US-NY/chestnut-ridge-13953" => array("type"=>"village", "pop"=>7871, "name_en"=>"Chestnut Ridge", "lang"=>"en", "lat"=>41.081867, "lon"=>-74.051924),
"/US/US-NY/chittenango-13954" => array("type"=>"village", "pop"=>4882, "name_en"=>"Chittenango", "lang"=>"en", "lat"=>43.045567, "lon"=>-75.873343),
"/US/US-NY/churchville-13955" => array("type"=>"village", "pop"=>1979, "name_en"=>"Churchville", "lang"=>"en", "lat"=>43.101484, "lon"=>-77.884129),
"/US/US-NY/clarence-center-13956" => array("type"=>"cdp", "pop"=>1376, "name_en"=>"Clarence Center", "lang"=>"en", "lat"=>43.008699, "lon"=>-78.631776),
"/US/US-NY/clark-mills-13957" => array("type"=>"cdp", "pop"=>1303, "name_en"=>"Clark Mills", "lang"=>"en", "lat"=>43.084188, "lon"=>-75.376598),
"/US/US-NY/claverack-red-mills-13958" => array("type"=>"cdp", "pop"=>1110, "name_en"=>"Claverack-Red Mills", "lang"=>"en", "lat"=>42.223227, "lon"=>-73.724855),
"/US/US-NY/clayton-13959" => array("type"=>"village", "pop"=>1873, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>44.235671, "lon"=>-76.086925),
"/US/US-NY/clayville-13960" => array("type"=>"village", "pop"=>422, "name_en"=>"Clayville", "lang"=>"en", "lat"=>42.974211, "lon"=>-75.249753),
"/US/US-NY/cleveland-13961" => array("type"=>"village", "pop"=>731, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>43.239249, "lon"=>-75.885033),
"/US/US-NY/clifton-springs-13962" => array("type"=>"village", "pop"=>2141, "name_en"=>"Clifton Springs", "lang"=>"en", "lat"=>42.959730, "lon"=>-77.133428),
"/US/US-NY/clinton-13963" => array("type"=>"village", "pop"=>1886, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>43.048950, "lon"=>-75.378728),
"/US/US-NY/clintondale-13964" => array("type"=>"cdp", "pop"=>1394, "name_en"=>"Clintondale", "lang"=>"en", "lat"=>41.691782, "lon"=>-74.047328),
"/US/US-NY/clyde-13965" => array("type"=>"village", "pop"=>2099, "name_ru"=>"Клайд", "name_en"=>"Clyde", "lang"=>"en", "lat"=>43.083373, "lon"=>-76.870995),
"/US/US-NY/cobleskill-13966" => array("type"=>"village", "pop"=>4593, "name_en"=>"Cobleskill", "lang"=>"en", "lat"=>42.679089, "lon"=>-74.486071),
"/US/US-NY/cohocton-13967" => array("type"=>"village", "pop"=>811, "name_en"=>"Cohocton", "lang"=>"en", "lat"=>42.500016, "lon"=>-77.499327),
"/US/US-NY/cohoes-13968" => array("type"=>"city", "pop"=>15025, "name_en"=>"Cohoes", "lang"=>"en", "lat"=>42.772650, "lon"=>-73.708422),
"/US/US-NY/cold-brook-13969" => array("type"=>"village", "pop"=>314, "name_en"=>"Cold Brook", "lang"=>"en", "lat"=>43.240890, "lon"=>-75.037617),
"/US/US-NY/cold-spring-13970" => array("type"=>"village", "pop"=>1996, "name_ru"=>"Коулд Спринг", "name_en"=>"Cold Spring", "lang"=>"en", "lat"=>41.418814, "lon"=>-73.954939),
"/US/US-NY/cold-spring-harbor-13971" => array("type"=>"cdp", "pop"=>4789, "name_en"=>"Cold Spring Harbor", "lang"=>"en", "lat"=>40.863050, "lon"=>-73.447898),
"/US/US-NY/colonie-13972" => array("type"=>"village", "pop"=>8591, "name_en"=>"Colonie", "lang"=>"en", "lat"=>42.720800, "lon"=>-73.834580),
"/US/US-NY/commack-13973" => array("type"=>"cdp", "pop"=>36124, "name_en"=>"Commack", "lang"=>"en", "lat"=>40.844939, "lon"=>-73.283636),
"/US/US-NY/congers-13974" => array("type"=>"cdp", "pop"=>8003, "name_en"=>"Congers", "lang"=>"en", "lat"=>41.147937, "lon"=>-73.946691),
"/US/US-NY/constableville-13975" => array("type"=>"village", "pop"=>278, "name_en"=>"Constableville", "lang"=>"en", "lat"=>43.564436, "lon"=>-75.428533),
"/US/US-NY/constantia-13976" => array("type"=>"cdp", "pop"=>5073, "name_en"=>"Constantia", "lang"=>"en", "lat"=>43.253941, "lon"=>-76.001944),
"/US/US-NY/cooperstown-13977" => array("type"=>"village", "pop"=>1898, "name_en"=>"Cooperstown", "lang"=>"en", "lat"=>42.699738, "lon"=>-74.930747),
"/US/US-NY/copenhagen-13978" => array("type"=>"village", "pop"=>792, "name_en"=>"Copenhagen", "lang"=>"en", "lat"=>43.893837, "lon"=>-75.674023),
"/US/US-NY/copiague-13979" => array("type"=>"cdp", "pop"=>20769, "name_en"=>"Copiague", "lang"=>"en", "lat"=>40.671350, "lon"=>-73.392019),
"/US/US-NY/coram-13980" => array("type"=>"cdp", "pop"=>30111, "name_en"=>"Coram", "lang"=>"en", "lat"=>40.879490, "lon"=>-73.004477),
"/US/US-NY/corfu-13981" => array("type"=>"village", "pop"=>735, "name_en"=>"Corfu", "lang"=>"en", "lat"=>42.960363, "lon"=>-78.400335),
"/US/US-NY/corinth-13982" => array("type"=>"village", "pop"=>2406, "name_ru"=>"Коринт", "name_en"=>"Corinth", "lang"=>"en", "lat"=>43.245479, "lon"=>-73.830661),
"/US/US-NY/corning-13983" => array("type"=>"city", "pop"=>10281, "name_ru"=>"Корнинг", "name_en"=>"Corning", "lang"=>"en", "lat"=>42.147224, "lon"=>-77.056709),
"/US/US-NY/cornwall-on-hudson-13984" => array("type"=>"village", "pop"=>3093, "name_en"=>"Cornwall on Hudson", "lang"=>"en", "lat"=>41.438652, "lon"=>-74.017251),
"/US/US-NY/cortland-13985" => array("type"=>"city", "pop"=>18434, "name_ru"=>"Кортленд", "name_en"=>"Cortland", "lang"=>"en", "lat"=>42.600908, "lon"=>-76.179404),
"/US/US-NY/cortland-west-13986" => array("type"=>"cdp", "pop"=>1270, "name_en"=>"Cortland West", "lang"=>"en", "lat"=>42.589042, "lon"=>-76.227897),
"/US/US-NY/country-knolls-13987" => array("type"=>"cdp", "pop"=>2287, "name_en"=>"Country Knolls", "lang"=>"en", "lat"=>42.914900, "lon"=>-73.805666),
"/US/US-NY/cove-neck-13988" => array("type"=>"village", "pop"=>304, "name_en"=>"Cove Neck", "lang"=>"en", "lat"=>40.884381, "lon"=>-73.496630),
"/US/US-NY/coxsackie-13989" => array("type"=>"village", "pop"=>2759, "name_en"=>"Coxsackie", "lang"=>"en", "lat"=>42.358195, "lon"=>-73.809250),
"/US/US-NY/croghan-13990" => array("type"=>"village", "pop"=>607, "name_en"=>"Croghan", "lang"=>"en", "lat"=>43.895277, "lon"=>-75.391269),
"/US/US-NY/croton-on-hudson-13991" => array("type"=>"village", "pop"=>7943, "name_en"=>"Croton-on-Hudson", "lang"=>"en", "lat"=>41.203203, "lon"=>-73.890711),
"/US/US-NY/crown-heights-13992" => array("type"=>"cdp", "pop"=>3200, "name_en"=>"Crown Heights", "lang"=>"en", "lat"=>41.637092, "lon"=>-73.938443),
"/US/US-NY/cuba-13993" => array("type"=>"village", "pop"=>1552, "name_ru"=>"Куба", "name_en"=>"Cuba", "lang"=>"en", "lat"=>42.218558, "lon"=>-78.276341),
"/US/US-NY/cumberland-head-13994" => array("type"=>"cdp", "pop"=>1698, "name_en"=>"Cumberland Head", "lang"=>"en", "lat"=>44.716303, "lon"=>-73.403025),
"/US/US-NY/cutchogue-13995" => array("type"=>"cdp", "pop"=>2627, "name_en"=>"Cutchogue", "lang"=>"en", "lat"=>41.012912, "lon"=>-72.486239),
"/US/US-NY/dannemora-13996" => array("type"=>"village", "pop"=>4055, "name_en"=>"Dannemora", "lang"=>"en", "lat"=>44.720837, "lon"=>-73.719318),
"/US/US-NY/dansville-13997" => array("type"=>"village", "pop"=>4484, "name_ru"=>"Дэнсвилл", "name_en"=>"Dansville", "lang"=>"en", "lat"=>42.560916, "lon"=>-77.695286),
"/US/US-NY/de-witt-14008" => array("type"=>"cdp", "pop"=>24424, "name_ru"=>"Де Витт", "name_en"=>"De Witt", "lang"=>"en", "lat"=>43.038500, "lon"=>-76.073379),
"/US/US-NY/deer-park-13998" => array("type"=>"cdp", "pop"=>28840, "name_ru"=>"Дир Парк", "name_en"=>"Deer Park", "lang"=>"en", "lat"=>40.762300, "lon"=>-73.322624),
"/US/US-NY/deferiet-13999" => array("type"=>"village", "pop"=>317, "name_en"=>"Deferiet", "lang"=>"en", "lat"=>44.034614, "lon"=>-75.679728),
"/US/US-NY/delanson-14000" => array("type"=>"village", "pop"=>414, "name_en"=>"Delanson", "lang"=>"en", "lat"=>42.747733, "lon"=>-74.184995),
"/US/US-NY/delevan-14001" => array("type"=>"village", "pop"=>999, "name_en"=>"Delevan", "lang"=>"en", "lat"=>42.491034, "lon"=>-78.479353),
"/US/US-NY/delhi-14002" => array("type"=>"village", "pop"=>2817, "name_ru"=>"Дели", "name_en"=>"Delhi", "lang"=>"en", "lat"=>42.279951, "lon"=>-74.914508),
"/US/US-NY/delmar-14003" => array("type"=>"cdp", "pop"=>8360, "name_ru"=>"Делмар", "name_en"=>"Delmar", "lang"=>"en", "lat"=>42.624333, "lon"=>-73.833529),
"/US/US-NY/depew-14004" => array("type"=>"village", "pop"=>15307, "name_en"=>"Depew", "lang"=>"en", "lat"=>42.912033, "lon"=>-78.704713),
"/US/US-NY/deposit-14005" => array("type"=>"village", "pop"=>1592, "name_en"=>"Deposit", "lang"=>"en", "lat"=>42.063972, "lon"=>-75.421876),
"/US/US-NY/dering-harbor-14006" => array("type"=>"village", "pop"=>13, "name_en"=>"Dering Harbor", "lang"=>"en", "lat"=>41.090555, "lon"=>-72.340694),
"/US/US-NY/deruyter-14007" => array("type"=>"village", "pop"=>505, "name_en"=>"Deruyter", "lang"=>"en", "lat"=>42.758923, "lon"=>-75.887302),
"/US/US-NY/dexter-14009" => array("type"=>"village", "pop"=>1174, "name_ru"=>"Декстер", "name_en"=>"Dexter", "lang"=>"en", "lat"=>44.008207, "lon"=>-76.045715),
"/US/US-NY/dix-hills-14010" => array("type"=>"cdp", "pop"=>25849, "name_en"=>"Dix Hills", "lang"=>"en", "lat"=>40.804816, "lon"=>-73.336575),
"/US/US-NY/dobbs-ferry-14011" => array("type"=>"village", "pop"=>11134, "name_en"=>"Dobbs Ferry", "lang"=>"en", "lat"=>41.014533, "lon"=>-73.869729),
"/US/US-NY/dolgeville-14012" => array("type"=>"village", "pop"=>2016, "name_en"=>"Dolgeville", "lang"=>"en", "lat"=>43.101223, "lon"=>-74.774547),
"/US/US-NY/dover-plains-14013" => array("type"=>"cdp", "pop"=>1847, "name_en"=>"Dover Plains", "lang"=>"en", "lat"=>41.745200, "lon"=>-73.580261),
"/US/US-NY/dresden-14014" => array("type"=>"village", "pop"=>288, "name_ru"=>"Дрезден", "name_en"=>"Dresden", "lang"=>"en", "lat"=>42.683067, "lon"=>-76.956931),
"/US/US-NY/dryden-14015" => array("type"=>"village", "pop"=>1820, "name_ru"=>"Драйден", "name_en"=>"Dryden", "lang"=>"en", "lat"=>42.490248, "lon"=>-76.298901),
"/US/US-NY/dundee-14016" => array("type"=>"village", "pop"=>1595, "name_ru"=>"Данди", "name_en"=>"Dundee", "lang"=>"en", "lat"=>42.523536, "lon"=>-76.978220),
"/US/US-NY/dunkirk-14017" => array("type"=>"city", "pop"=>12085, "name_ru"=>"Дюнкерк", "name_en"=>"Dunkirk", "lang"=>"en", "lat"=>42.565406, "lon"=>-79.324134),
"/US/US-NY/earlville-14018" => array("type"=>"village", "pop"=>754, "name_ru"=>"Эрлвилл", "name_en"=>"Earlville", "lang"=>"en", "lat"=>42.740275, "lon"=>-75.543571),
"/US/US-NY/east-aurora-14019" => array("type"=>"village", "pop"=>6242, "name_en"=>"East Aurora", "lang"=>"en", "lat"=>42.766750, "lon"=>-78.617374),
"/US/US-NY/east-bloomfield-14020" => array("type"=>"village", "pop"=>3578, "name_en"=>"East Bloomfield", "lang"=>"en", "lat"=>42.897077, "lon"=>-77.433772),
"/US/US-NY/east-farmingdale-14022" => array("type"=>"cdp", "pop"=>4510, "name_en"=>"East Farmingdale", "lang"=>"en", "lat"=>40.729147, "lon"=>-73.417317),
"/US/US-NY/east-glenville-14023" => array("type"=>"cdp", "pop"=>6518, "name_en"=>"East Glenville", "lang"=>"en", "lat"=>42.860050, "lon"=>-73.915914),
"/US/US-NY/east-greenbush-14024" => array("type"=>"cdp", "pop"=>16891, "name_en"=>"East Greenbush", "lang"=>"en", "lat"=>42.594400, "lon"=>-73.703207),
"/US/US-NY/east-hampton-14025" => array("type"=>"village", "pop"=>1394, "name_ru"=>"Ист Хэмптон", "name_en"=>"East Hampton", "lang"=>"en", "lat"=>40.950688, "lon"=>-72.196930),
"/US/US-NY/east-hampton-north-14026" => array("type"=>"cdp", "pop"=>2780, "name_en"=>"East Hampton North", "lang"=>"en", "lat"=>40.973488, "lon"=>-72.188872),
"/US/US-NY/east-hills-14027" => array("type"=>"village", "pop"=>6809, "name_en"=>"East Hills", "lang"=>"en", "lat"=>40.796450, "lon"=>-73.629166),
"/US/US-NY/east-islip-14028" => array("type"=>"cdp", "pop"=>14325, "name_en"=>"East Islip", "lang"=>"en", "lat"=>40.725232, "lon"=>-73.187002),
"/US/US-NY/east-ithaca-14029" => array("type"=>"cdp", "pop"=>2152, "name_en"=>"East Ithaca", "lang"=>"en", "lat"=>42.427020, "lon"=>-76.465817),
"/US/US-NY/east-massapequa-14030" => array("type"=>"cdp", "pop"=>19550, "name_en"=>"East Massapequa", "lang"=>"en", "lat"=>40.673400, "lon"=>-73.436899),
"/US/US-NY/east-meadow-14031" => array("type"=>"cdp", "pop"=>36909, "name_en"=>"East Meadow", "lang"=>"en", "lat"=>40.719850, "lon"=>-73.559275),
"/US/US-NY/east-middletown-14032" => array("type"=>"cdp", "pop"=>4974, "name_en"=>"East Middletown", "lang"=>"en", "lat"=>41.447564, "lon"=>-74.393528),
"/US/US-NY/east-moriches-14033" => array("type"=>"cdp", "pop"=>4021, "name_en"=>"East Moriches", "lang"=>"en", "lat"=>40.810091, "lon"=>-72.760535),
"/US/US-NY/east-northport-14034" => array("type"=>"cdp", "pop"=>20411, "name_en"=>"East Northport", "lang"=>"en", "lat"=>40.879150, "lon"=>-73.324572),
"/US/US-NY/east-norwich-14035" => array("type"=>"cdp", "pop"=>2698, "name_en"=>"East Norwich", "lang"=>"en", "lat"=>40.847033, "lon"=>-73.529247),
"/US/US-NY/east-patchogue-14036" => array("type"=>"cdp", "pop"=>20195, "name_en"=>"East Patchogue", "lang"=>"en", "lat"=>40.769896, "lon"=>-72.982351),
"/US/US-NY/east-quogue-14037" => array("type"=>"cdp", "pop"=>4372, "name_en"=>"East Quogue", "lang"=>"en", "lat"=>40.853112, "lon"=>-72.576748),
"/US/US-NY/east-randolph-14038" => array("type"=>"village", "pop"=>633, "name_en"=>"East Randolph", "lang"=>"en", "lat"=>42.170487, "lon"=>-78.950772),
"/US/US-NY/east-rochester-14039" => array("type"=>"village", "pop"=>6269, "name_en"=>"East Rochester", "lang"=>"en", "lat"=>43.111601, "lon"=>-77.487379),
"/US/US-NY/east-rockaway-14040" => array("type"=>"village", "pop"=>10438, "name_en"=>"East Rockaway", "lang"=>"en", "lat"=>40.643700, "lon"=>-73.667338),
"/US/US-NY/east-shoreham-14041" => array("type"=>"cdp", "pop"=>5461, "name_en"=>"East Shoreham", "lang"=>"en", "lat"=>40.944645, "lon"=>-72.880136),
"/US/US-NY/east-syracuse-14042" => array("type"=>"village", "pop"=>2972, "name_en"=>"East Syracuse", "lang"=>"en", "lat"=>43.063750, "lon"=>-76.069824),
"/US/US-NY/east-williston-14043" => array("type"=>"village", "pop"=>2544, "name_en"=>"East Williston", "lang"=>"en", "lat"=>40.760650, "lon"=>-73.633855),
"/US/US-NY/eastchester-14021" => array("type"=>"cdp", "pop"=>31450, "name_en"=>"Eastchester", "lang"=>"en", "lat"=>40.958161, "lon"=>-73.807821),
"/US/US-NY/eatons-neck-14044" => array("type"=>"cdp", "pop"=>1499, "name_en"=>"Eatons Neck", "lang"=>"en", "lat"=>40.930600, "lon"=>-73.401906),
"/US/US-NY/eden-14045" => array("type"=>"cdp", "pop"=>7736, "name_ru"=>"Иден", "name_en"=>"Eden", "lang"=>"en", "lat"=>42.651968, "lon"=>-78.900737),
"/US/US-NY/edwards-14046" => array("type"=>"village", "pop"=>440, "name_ru"=>"Эдвардс", "name_en"=>"Edwards", "lang"=>"en", "lat"=>44.323598, "lon"=>-75.253275),
"/US/US-NY/elba-14047" => array("type"=>"village", "pop"=>646, "name_ru"=>"Эльба", "name_en"=>"Elba", "lang"=>"en", "lat"=>43.075150, "lon"=>-78.188853),
"/US/US-NY/elbridge-14048" => array("type"=>"village", "pop"=>1059, "name_en"=>"Elbridge", "lang"=>"en", "lat"=>43.034200, "lon"=>-76.443625),
"/US/US-NY/ellenville-14049" => array("type"=>"village", "pop"=>3901, "name_en"=>"Ellenville", "lang"=>"en", "lat"=>41.702861, "lon"=>-74.364109),
"/US/US-NY/ellicottville-14050" => array("type"=>"village", "pop"=>522, "name_en"=>"Ellicottville", "lang"=>"en", "lat"=>42.274838, "lon"=>-78.672590),
"/US/US-NY/ellisburg-14051" => array("type"=>"village", "pop"=>277, "name_en"=>"Ellisburg", "lang"=>"en", "lat"=>43.735969, "lon"=>-76.133768),
"/US/US-NY/elma-center-14052" => array("type"=>"cdp", "pop"=>2354, "name_en"=>"Elma Center", "lang"=>"en", "lat"=>42.826200, "lon"=>-78.632680),
"/US/US-NY/elmira-14053" => array("type"=>"city", "pop"=>29271, "name_ru"=>"Элмира", "name_en"=>"Elmira", "lang"=>"en", "lat"=>42.093250, "lon"=>-76.810117),
"/US/US-NY/elmira-heights-14054" => array("type"=>"village", "pop"=>3897, "name_en"=>"Elmira Heights", "lang"=>"en", "lat"=>42.126350, "lon"=>-76.826641),
"/US/US-NY/elmont-14055" => array("type"=>"cdp", "pop"=>28612, "name_en"=>"Elmont", "lang"=>"en", "lat"=>40.703950, "lon"=>-73.707545),
"/US/US-NY/elmsford-14056" => array("type"=>"village", "pop"=>4757, "name_en"=>"Elmsford", "lang"=>"en", "lat"=>41.053150, "lon"=>-73.814559),
"/US/US-NY/elwood-14057" => array("type"=>"cdp", "pop"=>10916, "name_ru"=>"Элвуд", "name_en"=>"Elwood", "lang"=>"en", "lat"=>40.846233, "lon"=>-73.342124),
"/US/US-NY/endicott-14058" => array("type"=>"village", "pop"=>12411, "name_en"=>"Endicott", "lang"=>"en", "lat"=>42.097850, "lon"=>-76.063649),
"/US/US-NY/endwell-14059" => array("type"=>"cdp", "pop"=>12602, "name_en"=>"Endwell", "lang"=>"en", "lat"=>42.117250, "lon"=>-76.023277),
"/US/US-NY/esperance-14060" => array("type"=>"village", "pop"=>374, "name_en"=>"Esperance", "lang"=>"en", "lat"=>42.764401, "lon"=>-74.257690),
"/US/US-NY/evans-mills-14061" => array("type"=>"village", "pop"=>623, "name_en"=>"Evans Mills", "lang"=>"en", "lat"=>44.087835, "lon"=>-75.808288),
"/US/US-NY/fabius-14062" => array("type"=>"village", "pop"=>334, "name_en"=>"Fabius", "lang"=>"en", "lat"=>42.835485, "lon"=>-75.985507),
"/US/US-NY/fair-haven-14063" => array("type"=>"village", "pop"=>847, "name_ru"=>"Фэр Хейвен", "name_en"=>"Fair Haven", "lang"=>"en", "lat"=>43.325732, "lon"=>-76.702435),
"/US/US-NY/fairmount-14064" => array("type"=>"cdp", "pop"=>12266, "name_ru"=>"Фэрмаунт", "name_en"=>"Fairmount", "lang"=>"en", "lat"=>43.040550, "lon"=>-76.247231),
"/US/US-NY/fairport-14065" => array("type"=>"village", "pop"=>5472, "name_en"=>"Fairport", "lang"=>"en", "lat"=>43.100151, "lon"=>-77.443300),
"/US/US-NY/fairview-14066" => array("type"=>"cdp", "pop"=>4811, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>41.731854, "lon"=>-73.913381),
"/US/US-NY/falconer-14067" => array("type"=>"village", "pop"=>2342, "name_en"=>"Falconer", "lang"=>"en", "lat"=>42.118900, "lon"=>-79.197348),
"/US/US-NY/farmingdale-14068" => array("type"=>"village", "pop"=>8865, "name_ru"=>"Фармингдейл", "name_en"=>"Farmingdale", "lang"=>"en", "lat"=>40.733100, "lon"=>-73.447636),
"/US/US-NY/farmingville-14069" => array("type"=>"cdp", "pop"=>14842, "name_en"=>"Farmingville", "lang"=>"en", "lat"=>40.842504, "lon"=>-73.044193),
"/US/US-NY/farnham-14070" => array("type"=>"village", "pop"=>312, "name_en"=>"Farnham", "lang"=>"en", "lat"=>42.596944, "lon"=>-79.078861),
"/US/US-NY/fayetteville-14071" => array("type"=>"village", "pop"=>4110, "name_ru"=>"Фейетвилл", "name_en"=>"Fayetteville", "lang"=>"en", "lat"=>43.030100, "lon"=>-75.997830),
"/US/US-NY/fillmore-14072" => array("type"=>"village", "pop"=>455, "name_ru"=>"Фильмор", "name_en"=>"Fillmore", "lang"=>"en", "lat"=>42.467957, "lon"=>-78.113134),
"/US/US-NY/firthcliffe-14073" => array("type"=>"cdp", "pop"=>4427, "name_en"=>"Firthcliffe", "lang"=>"en", "lat"=>41.439991, "lon"=>-74.034198),
"/US/US-NY/fishkill-14074" => array("type"=>"village", "pop"=>1692, "name_en"=>"Fishkill", "lang"=>"en", "lat"=>41.534545, "lon"=>-73.894161),
"/US/US-NY/flanders-14075" => array("type"=>"cdp", "pop"=>3231, "name_en"=>"Flanders", "lang"=>"en", "lat"=>40.890944, "lon"=>-72.604819),
"/US/US-NY/fleischmanns-14076" => array("type"=>"village", "pop"=>325, "name_en"=>"Fleischmanns", "lang"=>"en", "lat"=>42.155578, "lon"=>-74.533501),
"/US/US-NY/floral-park-14077" => array("type"=>"village", "pop"=>15846, "name_en"=>"Floral Park", "lang"=>"en", "lat"=>40.722315, "lon"=>-73.703634),
"/US/US-NY/florida-14078" => array("type"=>"village", "pop"=>2805, "name_ru"=>"Флорида", "name_en"=>"Florida", "lang"=>"en", "lat"=>41.330275, "lon"=>-74.354294),
"/US/US-NY/flower-hill-14079" => array("type"=>"village", "pop"=>4595, "name_en"=>"Flower Hill", "lang"=>"en", "lat"=>40.808533, "lon"=>-73.675957),
"/US/US-NY/fonda-14080" => array("type"=>"village", "pop"=>775, "name_ru"=>"Фонда", "name_en"=>"Fonda", "lang"=>"en", "lat"=>42.953272, "lon"=>-74.373111),
"/US/US-NY/forest-home-14081" => array("type"=>"cdp", "pop"=>1125, "name_en"=>"Forest Home", "lang"=>"en", "lat"=>42.453086, "lon"=>-76.471812),
"/US/US-NY/forestville-14082" => array("type"=>"village", "pop"=>714, "name_ru"=>"Форествилл", "name_en"=>"Forestville", "lang"=>"en", "lat"=>42.468621, "lon"=>-79.174970),
"/US/US-NY/fort-ann-14083" => array("type"=>"village", "pop"=>454, "name_en"=>"Fort Ann", "lang"=>"en", "lat"=>43.414951, "lon"=>-73.490619),
"/US/US-NY/fort-drum-14084" => array("type"=>"cdp", "pop"=>11578, "name_en"=>"Fort Drum", "lang"=>"en", "lat"=>44.053410, "lon"=>-75.773786),
"/US/US-NY/fort-edward-14085" => array("type"=>"village", "pop"=>3025, "name_en"=>"Fort Edward", "lang"=>"en", "lat"=>43.268950, "lon"=>-73.582964),
"/US/US-NY/fort-johnson-14086" => array("type"=>"village", "pop"=>470, "name_en"=>"Fort Johnson", "lang"=>"en", "lat"=>42.957303, "lon"=>-74.239269),
"/US/US-NY/fort-montgomery-14087" => array("type"=>"cdp", "pop"=>1450, "name_en"=>"Fort Montgomery", "lang"=>"en", "lat"=>41.337577, "lon"=>-73.987738),
"/US/US-NY/fort-plain-14088" => array("type"=>"village", "pop"=>2185, "name_en"=>"Fort Plain", "lang"=>"en", "lat"=>42.931684, "lon"=>-74.628298),
"/US/US-NY/fort-salonga-14089" => array("type"=>"cdp", "pop"=>9176, "name_en"=>"Fort Salonga", "lang"=>"en", "lat"=>40.911195, "lon"=>-73.294402),
"/US/US-NY/frankfort-14090" => array("type"=>"village", "pop"=>2361, "name_ru"=>"Франкфорт", "name_en"=>"Frankfort", "lang"=>"en", "lat"=>43.038088, "lon"=>-75.071937),
"/US/US-NY/franklin-14091" => array("type"=>"village", "pop"=>366, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>42.342609, "lon"=>-75.166411),
"/US/US-NY/franklin-square-14092" => array("type"=>"cdp", "pop"=>28205, "name_en"=>"Franklin Square", "lang"=>"en", "lat"=>40.700000, "lon"=>-73.677024),
"/US/US-NY/franklinville-14093" => array("type"=>"village", "pop"=>1704, "name_en"=>"Franklinville", "lang"=>"en", "lat"=>42.335277, "lon"=>-78.456510),
"/US/US-NY/fredonia-14094" => array("type"=>"village", "pop"=>11064, "name_ru"=>"Фредониа", "name_en"=>"Fredonia", "lang"=>"en", "lat"=>42.440446, "lon"=>-79.331792),
"/US/US-NY/freeport-14095" => array("type"=>"village", "pop"=>43881, "name_ru"=>"Фрипорт", "name_en"=>"Freeport", "lang"=>"en", "lat"=>40.650800, "lon"=>-73.584975),
"/US/US-NY/freeville-14096" => array("type"=>"village", "pop"=>504, "name_en"=>"Freeville", "lang"=>"en", "lat"=>42.511943, "lon"=>-76.346001),
"/US/US-NY/frewsburg-14097" => array("type"=>"cdp", "pop"=>1817, "name_en"=>"Frewsburg", "lang"=>"en", "lat"=>42.057162, "lon"=>-79.161130),
"/US/US-NY/friendship-14098" => array("type"=>"cdp", "pop"=>1842, "name_ru"=>"Френдшип", "name_en"=>"Friendship", "lang"=>"en", "lat"=>42.205441, "lon"=>-78.142404),
"/US/US-NY/fulton-14099" => array("type"=>"city", "pop"=>11219, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>43.316792, "lon"=>-76.415605),
"/US/US-NY/fultonville-14100" => array("type"=>"village", "pop"=>680, "name_en"=>"Fultonville", "lang"=>"en", "lat"=>42.945099, "lon"=>-74.370526),
"/US/US-NY/gainesville-14101" => array("type"=>"village", "pop"=>282, "name_ru"=>"Гейнсвилл", "name_en"=>"Gainesville", "lang"=>"en", "lat"=>42.640833, "lon"=>-78.135168),
"/US/US-NY/galway-14102" => array("type"=>"village", "pop"=>207, "name_en"=>"Galway", "lang"=>"en", "lat"=>43.018782, "lon"=>-74.031818),
"/US/US-NY/gang-mills-14103" => array("type"=>"cdp", "pop"=>2738, "name_en"=>"Gang Mills", "lang"=>"en", "lat"=>42.153567, "lon"=>-77.134574),
"/US/US-NY/garden-city-14104" => array("type"=>"village", "pop"=>22275, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>40.725900, "lon"=>-73.644548),
"/US/US-NY/garden-city-park-14105" => array("type"=>"cdp", "pop"=>7437, "name_en"=>"Garden City Park", "lang"=>"en", "lat"=>40.744433, "lon"=>-73.663523),
"/US/US-NY/garden-city-south-14106" => array("type"=>"cdp", "pop"=>4073, "name_en"=>"Garden City South", "lang"=>"en", "lat"=>40.711200, "lon"=>-73.660632),
"/US/US-NY/gardnertown-14107" => array("type"=>"cdp", "pop"=>4209, "name_en"=>"Gardnertown", "lang"=>"en", "lat"=>41.534050, "lon"=>-74.059948),
"/US/US-NY/gasport-14108" => array("type"=>"cdp", "pop"=>1336, "name_en"=>"Gasport", "lang"=>"en", "lat"=>43.195424, "lon"=>-78.577189),
"/US/US-NY/gates-north-gates-14109" => array("type"=>"cdp", "pop"=>14995, "name_en"=>"Gates-North Gates", "lang"=>"en", "lat"=>43.163968, "lon"=>-77.700700),
"/US/US-NY/geneseo-14110" => array("type"=>"village", "pop"=>7702, "name_ru"=>"Дженесео", "name_en"=>"Geneseo", "lang"=>"en", "lat"=>42.798333, "lon"=>-77.810566),
"/US/US-NY/geneva-14111" => array("type"=>"city", "pop"=>13216, "name_ru"=>"Дженива", "name_en"=>"Geneva", "lang"=>"en", "lat"=>42.865294, "lon"=>-76.983474),
"/US/US-NY/gilbertsville-14112" => array("type"=>"village", "pop"=>343, "name_en"=>"Gilbertsville", "lang"=>"en", "lat"=>42.469416, "lon"=>-75.321363),
"/US/US-NY/glasco-14113" => array("type"=>"cdp", "pop"=>1538, "name_ru"=>"Гласко", "name_en"=>"Glasco", "lang"=>"en", "lat"=>42.045611, "lon"=>-73.950626),
"/US/US-NY/glen-cove-14114" => array("type"=>"city", "pop"=>26895, "name_en"=>"Glen Cove", "lang"=>"en", "lat"=>40.891529, "lon"=>-73.638799),
"/US/US-NY/glen-head-14115" => array("type"=>"cdp", "pop"=>4488, "name_en"=>"Glen Head", "lang"=>"en", "lat"=>40.843264, "lon"=>-73.619338),
"/US/US-NY/glen-park-14116" => array("type"=>"village", "pop"=>499, "name_en"=>"Glen Park", "lang"=>"en", "lat"=>44.000794, "lon"=>-75.955145),
"/US/US-NY/glens-falls-14117" => array("type"=>"city", "pop"=>13877, "name_en"=>"Glens Falls", "lang"=>"en", "lat"=>43.311067, "lon"=>-73.645491),
"/US/US-NY/glens-falls-north-14118" => array("type"=>"cdp", "pop"=>7978, "name_en"=>"Glens Falls North", "lang"=>"en", "lat"=>43.334680, "lon"=>-73.684335),
"/US/US-NY/glenwood-landing-14119" => array("type"=>"cdp", "pop"=>3407, "name_en"=>"Glenwood Landing", "lang"=>"en", "lat"=>40.829800, "lon"=>-73.637657),
"/US/US-NY/gloversville-14120" => array("type"=>"city", "pop"=>14990, "name_en"=>"Gloversville", "lang"=>"en", "lat"=>43.050217, "lon"=>-74.347004),
"/US/US-NY/golden's-bridge-14121" => array("type"=>"cdp", "pop"=>1589, "name_en"=>"Golden's Bridge", "lang"=>"en", "lat"=>41.285907, "lon"=>-73.670467),
"/US/US-NY/goshen-14122" => array("type"=>"village", "pop"=>5585, "name_ru"=>"Гошен", "name_en"=>"Goshen", "lang"=>"en", "lat"=>41.401131, "lon"=>-74.327528),
"/US/US-NY/gouverneur-14123" => array("type"=>"village", "pop"=>4012, "name_en"=>"Gouverneur", "lang"=>"en", "lat"=>44.336604, "lon"=>-75.466158),
"/US/US-NY/gowanda-14124" => array("type"=>"village", "pop"=>2603, "name_en"=>"Gowanda", "lang"=>"en", "lat"=>42.461646, "lon"=>-78.935319),
"/US/US-NY/grand-view-on-hudson-14125" => array("type"=>"village", "pop"=>285, "name_en"=>"Grand View-on-Hudson", "lang"=>"en", "lat"=>41.058400, "lon"=>-73.907376),
"/US/US-NY/granville-14126" => array("type"=>"village", "pop"=>2541, "name_ru"=>"Гранвилл", "name_en"=>"Granville", "lang"=>"en", "lat"=>43.408336, "lon"=>-73.260986),
"/US/US-NY/great-neck-14127" => array("type"=>"village", "pop"=>9923, "name_en"=>"Great Neck", "lang"=>"en", "lat"=>40.802350, "lon"=>-73.733370),
"/US/US-NY/great-neck-estates-14128" => array("type"=>"village", "pop"=>2769, "name_en"=>"Great Neck Estates", "lang"=>"en", "lat"=>40.784950, "lon"=>-73.739124),
"/US/US-NY/great-neck-plaza-14129" => array("type"=>"village", "pop"=>7001, "name_en"=>"Great Neck Plaza", "lang"=>"en", "lat"=>40.786867, "lon"=>-73.726466),
"/US/US-NY/greece-14130" => array("type"=>"cdp", "pop"=>93362, "name_en"=>"Greece", "lang"=>"en", "lat"=>43.211943, "lon"=>-77.702341),
"/US/US-NY/green-island-14132" => array("type"=>"village", "pop"=>2532, "name_ru"=>"Грин Айленд", "name_en"=>"Green Island", "lang"=>"en", "lat"=>42.747767, "lon"=>-73.692599),
"/US/US-NY/greene-14131" => array("type"=>"village", "pop"=>1646, "name_ru"=>"Грин", "name_en"=>"Greene", "lang"=>"en", "lat"=>42.329999, "lon"=>-75.770501),
"/US/US-NY/greenlawn-14133" => array("type"=>"cdp", "pop"=>13208, "name_en"=>"Greenlawn", "lang"=>"en", "lat"=>40.863150, "lon"=>-73.366568),
"/US/US-NY/greenport-14134" => array("type"=>"village", "pop"=>2144, "name_en"=>"Greenport", "lang"=>"en", "lat"=>41.102661, "lon"=>-72.367765),
"/US/US-NY/greenport-west-14135" => array("type"=>"cdp", "pop"=>1614, "name_en"=>"Greenport West", "lang"=>"en", "lat"=>41.099483, "lon"=>-72.374925),
"/US/US-NY/greenville-14136" => array("type"=>"cdp", "pop"=>3507, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>41.000917, "lon"=>-73.818706),
"/US/US-NY/greenwich-14137" => array("type"=>"village", "pop"=>1827, "name_en"=>"Greenwich", "lang"=>"en", "lat"=>43.086612, "lon"=>-73.496717),
"/US/US-NY/greenwood-lake-14138" => array("type"=>"village", "pop"=>3417, "name_en"=>"Greenwood Lake", "lang"=>"en", "lat"=>41.221977, "lon"=>-74.289011),
"/US/US-NY/groton-14139" => array("type"=>"village", "pop"=>2401, "name_ru"=>"Гротон", "name_en"=>"Groton", "lang"=>"en", "lat"=>42.586851, "lon"=>-76.365251),
"/US/US-NY/hagaman-14140" => array("type"=>"village", "pop"=>1304, "name_en"=>"Hagaman", "lang"=>"en", "lat"=>42.974847, "lon"=>-74.153773),
"/US/US-NY/halesite-14141" => array("type"=>"cdp", "pop"=>2687, "name_en"=>"Halesite", "lang"=>"en", "lat"=>40.887000, "lon"=>-73.414076),
"/US/US-NY/hamburg-14142" => array("type"=>"village", "pop"=>9400, "name_ru"=>"Гамбург", "name_en"=>"Hamburg", "lang"=>"en", "lat"=>42.724050, "lon"=>-78.834899),
"/US/US-NY/hamilton-14143" => array("type"=>"village", "pop"=>3784, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>42.824537, "lon"=>-75.547296),
"/US/US-NY/hammond-14144" => array("type"=>"village", "pop"=>285, "name_ru"=>"Хаммонд", "name_en"=>"Hammond", "lang"=>"en", "lat"=>44.447189, "lon"=>-75.694200),
"/US/US-NY/hammondsport-14145" => array("type"=>"village", "pop"=>731, "name_en"=>"Hammondsport", "lang"=>"en", "lat"=>42.408802, "lon"=>-77.223081),
"/US/US-NY/hampton-bays-14146" => array("type"=>"cdp", "pop"=>7893, "name_en"=>"Hampton Bays", "lang"=>"en", "lat"=>40.862043, "lon"=>-72.524405),
"/US/US-NY/hampton-manor-14147" => array("type"=>"cdp", "pop"=>2600, "name_en"=>"Hampton Manor", "lang"=>"en", "lat"=>42.620750, "lon"=>-73.728917),
"/US/US-NY/hancock-14148" => array("type"=>"village", "pop"=>1088, "name_ru"=>"Хэнкок", "name_en"=>"Hancock", "lang"=>"en", "lat"=>41.953439, "lon"=>-75.283343),
"/US/US-NY/hannibal-14149" => array("type"=>"village", "pop"=>513, "name_ru"=>"Ханнибал", "name_en"=>"Hannibal", "lang"=>"en", "lat"=>43.317757, "lon"=>-76.577572),
"/US/US-NY/harriman-14150" => array("type"=>"village", "pop"=>2249, "name_en"=>"Harriman", "lang"=>"en", "lat"=>41.308926, "lon"=>-74.144987),
"/US/US-NY/harris-hill-14151" => array("type"=>"cdp", "pop"=>4577, "name_en"=>"Harris Hill", "lang"=>"en", "lat"=>42.976600, "lon"=>-78.682699),
"/US/US-NY/harrison-14152" => array("type"=>"village", "pop"=>26665, "name_ru"=>"Харрисон", "name_en"=>"Harrison", "lang"=>"en", "lat"=>41.022578, "lon"=>-73.720202),
"/US/US-NY/harrisville-14153" => array("type"=>"village", "pop"=>596, "name_ru"=>"Харрисвилл", "name_en"=>"Harrisville", "lang"=>"en", "lat"=>44.153431, "lon"=>-75.321652),
"/US/US-NY/hartsdale-14154" => array("type"=>"cdp", "pop"=>9587, "name_en"=>"Hartsdale", "lang"=>"en", "lat"=>41.025654, "lon"=>-73.805227),
"/US/US-NY/hastings-on-hudson-14155" => array("type"=>"village", "pop"=>7919, "name_en"=>"Hastings-on-Hudson", "lang"=>"en", "lat"=>40.987867, "lon"=>-73.879584),
"/US/US-NY/hauppauge-14156" => array("type"=>"cdp", "pop"=>19750, "name_en"=>"Hauppauge", "lang"=>"en", "lat"=>40.819984, "lon"=>-73.212709),
"/US/US-NY/haverstraw-14157" => array("type"=>"village", "pop"=>11029, "name_en"=>"Haverstraw", "lang"=>"en", "lat"=>41.185291, "lon"=>-73.955013),
"/US/US-NY/haviland-14158" => array("type"=>"cdp", "pop"=>3605, "name_ru"=>"Хавиленд", "name_en"=>"Haviland", "lang"=>"en", "lat"=>41.766528, "lon"=>-73.901923),
"/US/US-NY/hawthorne-14159" => array("type"=>"cdp", "pop"=>4764, "name_ru"=>"Хоуторн", "name_en"=>"Hawthorne", "lang"=>"en", "lat"=>41.104275, "lon"=>-73.793845),
"/US/US-NY/head-of-the-harbor-14160" => array("type"=>"village", "pop"=>1531, "name_en"=>"Head of the Harbor", "lang"=>"en", "lat"=>40.898187, "lon"=>-73.162764),
"/US/US-NY/hempstead-14161" => array("type"=>"village", "pop"=>53915, "name_en"=>"Hempstead", "lang"=>"en", "lat"=>40.703400, "lon"=>-73.618972),
"/US/US-NY/herkimer-14162" => array("type"=>"village", "pop"=>6970, "name_en"=>"Herkimer", "lang"=>"en", "lat"=>43.028731, "lon"=>-74.992758),
"/US/US-NY/hermon-14163" => array("type"=>"village", "pop"=>378, "name_en"=>"Hermon", "lang"=>"en", "lat"=>44.468382, "lon"=>-75.232698),
"/US/US-NY/herricks-14164" => array("type"=>"cdp", "pop"=>4097, "name_en"=>"Herricks", "lang"=>"en", "lat"=>40.756850, "lon"=>-73.664343),
"/US/US-NY/herrings-14165" => array("type"=>"village", "pop"=>132, "name_en"=>"Herrings", "lang"=>"en", "lat"=>44.023535, "lon"=>-75.657740),
"/US/US-NY/heuvelton-14166" => array("type"=>"village", "pop"=>756, "name_en"=>"Heuvelton", "lang"=>"en", "lat"=>44.617123, "lon"=>-75.404428),
"/US/US-NY/hewlett-14167" => array("type"=>"cdp", "pop"=>6620, "name_en"=>"Hewlett", "lang"=>"en", "lat"=>40.641733, "lon"=>-73.694710),
"/US/US-NY/hewlett-bay-park-14168" => array("type"=>"village", "pop"=>491, "name_en"=>"Hewlett Bay Park", "lang"=>"en", "lat"=>40.635000, "lon"=>-73.696152),
"/US/US-NY/hewlett-harbor-14169" => array("type"=>"village", "pop"=>1292, "name_en"=>"Hewlett Harbor", "lang"=>"en", "lat"=>40.632000, "lon"=>-73.684575),
"/US/US-NY/hewlett-neck-14170" => array("type"=>"village", "pop"=>533, "name_en"=>"Hewlett Neck", "lang"=>"en", "lat"=>40.624712, "lon"=>-73.697236),
"/US/US-NY/hicksville-14171" => array("type"=>"cdp", "pop"=>40174, "name_en"=>"Hicksville", "lang"=>"en", "lat"=>40.763850, "lon"=>-73.525231),
"/US/US-NY/highland-14172" => array("type"=>"cdp", "pop"=>2495, "name_ru"=>"Хайленд", "name_en"=>"Highland", "lang"=>"en", "lat"=>41.717256, "lon"=>-73.965690),
"/US/US-NY/highland-falls-14173" => array("type"=>"village", "pop"=>3705, "name_en"=>"Highland Falls", "lang"=>"en", "lat"=>41.362724, "lon"=>-73.969200),
"/US/US-NY/highland-mills-14174" => array("type"=>"cdp", "pop"=>2576, "name_en"=>"Highland Mills", "lang"=>"en", "lat"=>41.351874, "lon"=>-74.126823),
"/US/US-NY/hillburn-14175" => array("type"=>"village", "pop"=>870, "name_en"=>"Hillburn", "lang"=>"en", "lat"=>41.127233, "lon"=>-74.170416),
"/US/US-NY/hillcrest-14176" => array("type"=>"cdp", "pop"=>6447, "name_ru"=>"Хиллкрест", "name_en"=>"Hillcrest", "lang"=>"en", "lat"=>41.129550, "lon"=>-74.035942),
"/US/US-NY/hillside-lake-14177" => array("type"=>"cdp", "pop"=>1692, "name_en"=>"Hillside Lake", "lang"=>"en", "lat"=>41.614832, "lon"=>-73.798617),
"/US/US-NY/hilton-14178" => array("type"=>"village", "pop"=>6319, "name_en"=>"Hilton", "lang"=>"en", "lat"=>43.289901, "lon"=>-77.792780),
"/US/US-NY/hobart-14179" => array("type"=>"village", "pop"=>359, "name_ru"=>"Хобарт", "name_en"=>"Hobart", "lang"=>"en", "lat"=>42.371170, "lon"=>-74.668884),
"/US/US-NY/holbrook-14180" => array("type"=>"cdp", "pop"=>25273, "name_ru"=>"Холбрук", "name_en"=>"Holbrook", "lang"=>"en", "lat"=>40.792584, "lon"=>-73.070765),
"/US/US-NY/holcomb-14181" => array("type"=>"village", "pop"=>790, "name_ru"=>"Холкомб", "name_en"=>"Holcomb", "lang"=>"en", "lat"=>42.900229, "lon"=>-77.418295),
"/US/US-NY/holland-14182" => array("type"=>"cdp", "pop"=>3475, "name_ru"=>"Холланд", "name_en"=>"Holland", "lang"=>"en", "lat"=>42.636417, "lon"=>-78.547937),
"/US/US-NY/holland-patent-14183" => array("type"=>"village", "pop"=>450, "name_en"=>"Holland Patent", "lang"=>"en", "lat"=>43.241750, "lon"=>-75.257446),
"/US/US-NY/holley-14184" => array("type"=>"village", "pop"=>1686, "name_en"=>"Holley", "lang"=>"en", "lat"=>43.224218, "lon"=>-78.028105),
"/US/US-NY/holtsville-14185" => array("type"=>"cdp", "pop"=>14972, "name_en"=>"Holtsville", "lang"=>"en", "lat"=>40.813296, "lon"=>-73.047623),
"/US/US-NY/homer-14186" => array("type"=>"village", "pop"=>3248, "name_ru"=>"Гомер", "name_en"=>"Homer", "lang"=>"en", "lat"=>42.637812, "lon"=>-76.183564),
"/US/US-NY/honeoye-falls-14187" => array("type"=>"village", "pop"=>2647, "name_en"=>"Honeoye Falls", "lang"=>"en", "lat"=>42.954401, "lon"=>-77.591845),
"/US/US-NY/hoosick-falls-14188" => array("type"=>"village", "pop"=>3269, "name_en"=>"Hoosick Falls", "lang"=>"en", "lat"=>42.900499, "lon"=>-73.350350),
"/US/US-NY/hopewell-junction-14189" => array("type"=>"cdp", "pop"=>1786, "name_en"=>"Hopewell Junction", "lang"=>"en", "lat"=>41.585696, "lon"=>-73.801293),
"/US/US-NY/hornell-14190" => array("type"=>"city", "pop"=>8508, "name_en"=>"Hornell", "lang"=>"en", "lat"=>42.324363, "lon"=>-77.660539),
"/US/US-NY/horseheads-14191" => array("type"=>"village", "pop"=>6254, "name_en"=>"Horseheads", "lang"=>"en", "lat"=>42.169700, "lon"=>-76.829877),
"/US/US-NY/horseheads-north-14192" => array("type"=>"cdp", "pop"=>3003, "name_en"=>"Horseheads North", "lang"=>"en", "lat"=>42.192000, "lon"=>-76.809149),
"/US/US-NY/houghton-14193" => array("type"=>"cdp", "pop"=>1740, "name_ru"=>"Хоутон", "name_en"=>"Houghton", "lang"=>"en", "lat"=>42.426536, "lon"=>-78.161369),
"/US/US-NY/hudson-14194" => array("type"=>"city", "pop"=>6925, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>42.251289, "lon"=>-73.786533),
"/US/US-NY/hudson-falls-14195" => array("type"=>"village", "pop"=>6644, "name_en"=>"Hudson Falls", "lang"=>"en", "lat"=>43.304650, "lon"=>-73.582008),
"/US/US-NY/hunter-14196" => array("type"=>"village", "pop"=>471, "name_ru"=>"Хантер", "name_en"=>"Hunter", "lang"=>"en", "lat"=>42.211243, "lon"=>-74.215672),
"/US/US-NY/huntington-14197" => array("type"=>"cdp", "pop"=>201594, "name_ru"=>"Хантингтон", "name_en"=>"Huntington", "lang"=>"en", "lat"=>40.877700, "lon"=>-73.407856),
"/US/US-NY/huntington-bay-14198" => array("type"=>"village", "pop"=>1513, "name_en"=>"Huntington Bay", "lang"=>"en", "lat"=>40.910800, "lon"=>-73.416381),
"/US/US-NY/huntington-station-14199" => array("type"=>"cdp", "pop"=>28247, "name_en"=>"Huntington Station", "lang"=>"en", "lat"=>40.844900, "lon"=>-73.405205),
"/US/US-NY/hurley-14200" => array("type"=>"cdp", "pop"=>6512, "name_ru"=>"Херли", "name_en"=>"Hurley", "lang"=>"en", "lat"=>41.913815, "lon"=>-74.058198),
"/US/US-NY/ilion-14201" => array("type"=>"village", "pop"=>8015, "name_en"=>"Ilion", "lang"=>"en", "lat"=>43.011197, "lon"=>-75.040707),
"/US/US-NY/interlaken-14202" => array("type"=>"village", "pop"=>650, "name_ru"=>"Интерлакен", "name_en"=>"Interlaken", "lang"=>"en", "lat"=>42.618740, "lon"=>-76.724737),
"/US/US-NY/inwood-14203" => array("type"=>"cdp", "pop"=>7767, "name_ru"=>"Инвуд", "name_en"=>"Inwood", "lang"=>"en", "lat"=>40.622800, "lon"=>-73.750047),
"/US/US-NY/irondequoit-14204" => array("type"=>"cdp", "pop"=>50007, "name_en"=>"Irondequoit", "lang"=>"en", "lat"=>43.210268, "lon"=>-77.572101),
"/US/US-NY/irvington-14205" => array("type"=>"village", "pop"=>6666, "name_ru"=>"Ирвингтон", "name_en"=>"Irvington", "lang"=>"en", "lat"=>41.038600, "lon"=>-73.867341),
"/US/US-NY/island-park-14207" => array("type"=>"village", "pop"=>4806, "name_ru"=>"Айленд Парк", "name_en"=>"Island Park", "lang"=>"en", "lat"=>40.603367, "lon"=>-73.655531),
"/US/US-NY/islandia-14206" => array("type"=>"village", "pop"=>3163, "name_ru"=>"Исландия", "name_en"=>"Islandia", "lang"=>"en", "lat"=>40.806688, "lon"=>-73.171915),
"/US/US-NY/islip-14208" => array("type"=>"cdp", "pop"=>336292, "name_en"=>"Islip", "lang"=>"en", "lat"=>40.732767, "lon"=>-73.217654),
"/US/US-NY/islip-terrace-14209" => array("type"=>"cdp", "pop"=>5530, "name_en"=>"Islip Terrace", "lang"=>"en", "lat"=>40.748640, "lon"=>-73.186858),
"/US/US-NY/ithaca-14210" => array("type"=>"city", "pop"=>29763, "name_ru"=>"Итака", "name_en"=>"Ithaca", "lang"=>"en", "lat"=>42.443873, "lon"=>-76.503359),
"/US/US-NY/jamesport-14211" => array("type"=>"cdp", "pop"=>1532, "name_ru"=>"Джеймспорт", "name_en"=>"Jamesport", "lang"=>"en", "lat"=>40.945440, "lon"=>-72.575030),
"/US/US-NY/jamestown-14212" => array("type"=>"city", "pop"=>29463, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>42.097400, "lon"=>-79.236341),
"/US/US-NY/jamestown-west-14213" => array("type"=>"cdp", "pop"=>2633, "name_en"=>"Jamestown West", "lang"=>"en", "lat"=>42.089433, "lon"=>-79.278782),
"/US/US-NY/jefferson-heights-14214" => array("type"=>"cdp", "pop"=>1512, "name_en"=>"Jefferson Heights", "lang"=>"en", "lat"=>42.236425, "lon"=>-73.872762),
"/US/US-NY/jefferson-valley-yorktown-14215" => array("type"=>"cdp", "pop"=>14118, "name_en"=>"Jefferson Valley-Yorktown", "lang"=>"en", "lat"=>41.318588, "lon"=>-73.801866),
"/US/US-NY/jeffersonville-14216" => array("type"=>"village", "pop"=>393, "name_ru"=>"Джефферсонвилл", "name_en"=>"Jeffersonville", "lang"=>"en", "lat"=>41.779429, "lon"=>-74.932599),
"/US/US-NY/jericho-14217" => array("type"=>"cdp", "pop"=>13141, "name_ru"=>"Джерико", "name_en"=>"Jericho", "lang"=>"en", "lat"=>40.788500, "lon"=>-73.541863),
"/US/US-NY/johnson-city-14218" => array("type"=>"village", "pop"=>14727, "name_ru"=>"Джонсон Сити", "name_en"=>"Johnson City", "lang"=>"en", "lat"=>42.123600, "lon"=>-75.961820),
"/US/US-NY/johnstown-14219" => array("type"=>"city", "pop"=>8438, "name_ru"=>"Джонстаун", "name_en"=>"Johnstown", "lang"=>"en", "lat"=>43.008488, "lon"=>-74.374385),
"/US/US-NY/jordan-14220" => array("type"=>"village", "pop"=>1307, "name_ru"=>"Джордан", "name_en"=>"Jordan", "lang"=>"en", "lat"=>43.065713, "lon"=>-76.473256),
"/US/US-NY/keeseville-14221" => array("type"=>"village", "pop"=>1742, "name_en"=>"Keeseville", "lang"=>"en", "lat"=>44.503652, "lon"=>-73.482687),
"/US/US-NY/kenmore-14222" => array("type"=>"village", "pop"=>15002, "name_en"=>"Kenmore", "lang"=>"en", "lat"=>42.964850, "lon"=>-78.871921),
"/US/US-NY/kensington-14223" => array("type"=>"village", "pop"=>1204, "name_ru"=>"Кенсингтон", "name_en"=>"Kensington", "lang"=>"en", "lat"=>40.794331, "lon"=>-73.723478),
"/US/US-NY/kerhonkson-14224" => array("type"=>"cdp", "pop"=>1629, "name_en"=>"Kerhonkson", "lang"=>"en", "lat"=>41.778995, "lon"=>-74.295947),
"/US/US-NY/kinderhook-14225" => array("type"=>"village", "pop"=>1292, "name_ru"=>"Киндерхук", "name_en"=>"Kinderhook", "lang"=>"en", "lat"=>42.394562, "lon"=>-73.704074),
"/US/US-NY/kings-park-14226" => array("type"=>"cdp", "pop"=>17773, "name_en"=>"Kings Park", "lang"=>"en", "lat"=>40.890720, "lon"=>-73.245534),
"/US/US-NY/kings-point-14227" => array("type"=>"village", "pop"=>5394, "name_ru"=>"Кингс Поинт", "name_en"=>"Kings Point", "lang"=>"en", "lat"=>40.817167, "lon"=>-73.743903),
"/US/US-NY/kingston-14228" => array("type"=>"city", "pop"=>22441, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>41.930112, "lon"=>-73.997269),
"/US/US-NY/kiryas-joel-14229" => array("type"=>"village", "pop"=>21793, "name_en"=>"Kiryas Joel", "lang"=>"en", "lat"=>41.339927, "lon"=>-74.167638),
"/US/US-NY/lackawanna-14230" => array("type"=>"city", "pop"=>17588, "name_en"=>"Lackawanna", "lang"=>"en", "lat"=>42.816773, "lon"=>-78.829252),
"/US/US-NY/lacona-14231" => array("type"=>"village", "pop"=>559, "name_ru"=>"Лакона", "name_en"=>"Lacona", "lang"=>"en", "lat"=>43.643723, "lon"=>-76.068917),
"/US/US-NY/lake-carmel-14232" => array("type"=>"cdp", "pop"=>8489, "name_en"=>"Lake Carmel", "lang"=>"en", "lat"=>41.461201, "lon"=>-73.670270),
"/US/US-NY/lake-erie-beach-14233" => array("type"=>"cdp", "pop"=>4509, "name_en"=>"Lake Erie Beach", "lang"=>"en", "lat"=>42.623650, "lon"=>-79.077724),
"/US/US-NY/lake-george-14234" => array("type"=>"village", "pop"=>976, "name_en"=>"Lake George", "lang"=>"en", "lat"=>43.425650, "lon"=>-73.715095),
"/US/US-NY/lake-grove-14235" => array("type"=>"village", "pop"=>10948, "name_en"=>"Lake Grove", "lang"=>"en", "lat"=>40.857987, "lon"=>-73.116936),
"/US/US-NY/lake-katrine-14236" => array("type"=>"cdp", "pop"=>1998, "name_en"=>"Lake Katrine", "lang"=>"en", "lat"=>41.983256, "lon"=>-73.990197),
"/US/US-NY/lake-luzerne-hadley-14237" => array("type"=>"cdp", "pop"=>2042, "name_en"=>"Lake Luzerne-Hadley", "lang"=>"en", "lat"=>43.318907, "lon"=>-73.837981),
"/US/US-NY/lake-placid-14238" => array("type"=>"village", "pop"=>2750, "name_ru"=>"Лэйк Плэйсид", "name_en"=>"Lake Placid", "lang"=>"en", "lat"=>44.283376, "lon"=>-73.985012),
"/US/US-NY/lake-ronkonkoma-14239" => array("type"=>"cdp", "pop"=>18997, "name_en"=>"Lake Ronkonkoma", "lang"=>"en", "lat"=>40.831389, "lon"=>-73.110888),
"/US/US-NY/lake-success-14240" => array("type"=>"village", "pop"=>2888, "name_en"=>"Lake Success", "lang"=>"en", "lat"=>40.768667, "lon"=>-73.710154),
"/US/US-NY/lakeview-14241" => array("type"=>"cdp", "pop"=>5476, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>40.675200, "lon"=>-73.651434),
"/US/US-NY/lakewood-14242" => array("type"=>"village", "pop"=>3041, "name_ru"=>"Лейквуд", "name_en"=>"Lakewood", "lang"=>"en", "lat"=>42.100225, "lon"=>-79.319983),
"/US/US-NY/lancaster-14243" => array("type"=>"village", "pop"=>11108, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>42.901600, "lon"=>-78.669837),
"/US/US-NY/lansing-14244" => array("type"=>"village", "pop"=>3428, "name_ru"=>"Ленсинг", "name_en"=>"Lansing", "lang"=>"en", "lat"=>42.490762, "lon"=>-76.486396),
"/US/US-NY/larchmont-14245" => array("type"=>"village", "pop"=>6567, "name_en"=>"Larchmont", "lang"=>"en", "lat"=>40.926100, "lon"=>-73.753542),
"/US/US-NY/latham-14246" => array("type"=>"cdp", "pop"=>10131, "name_ru"=>"Латам", "name_en"=>"Latham", "lang"=>"en", "lat"=>42.745248, "lon"=>-73.746927),
"/US/US-NY/lattingtown-14247" => array("type"=>"village", "pop"=>1884, "name_en"=>"Lattingtown", "lang"=>"en", "lat"=>40.894750, "lon"=>-73.594893),
"/US/US-NY/laurel-14248" => array("type"=>"cdp", "pop"=>1094, "name_ru"=>"Лорел", "name_en"=>"Laurel", "lang"=>"en", "lat"=>40.975288, "lon"=>-72.553616),
"/US/US-NY/laurel-hollow-14249" => array("type"=>"village", "pop"=>2042, "name_en"=>"Laurel Hollow", "lang"=>"en", "lat"=>40.857300, "lon"=>-73.477822),
"/US/US-NY/laurens-14250" => array("type"=>"village", "pop"=>256, "name_ru"=>"Лоренс", "name_en"=>"Laurens", "lang"=>"en", "lat"=>42.531357, "lon"=>-75.088566),
"/US/US-NY/lawrence-14251" => array("type"=>"village", "pop"=>6573, "name_ru"=>"Лоренс", "name_en"=>"Lawrence", "lang"=>"en", "lat"=>40.604433, "lon"=>-73.712821),
"/US/US-NY/le-roy-14253" => array("type"=>"village", "pop"=>4142, "name_ru"=>"Ле Рой", "name_en"=>"Le Roy", "lang"=>"en", "lat"=>42.977929, "lon"=>-77.991027),
"/US/US-NY/leicester-14252" => array("type"=>"village", "pop"=>436, "name_en"=>"Leicester", "lang"=>"en", "lat"=>42.771440, "lon"=>-77.898039),
"/US/US-NY/levittown-14254" => array("type"=>"cdp", "pop"=>53286, "name_en"=>"Levittown", "lang"=>"en", "lat"=>40.723750, "lon"=>-73.513108),
"/US/US-NY/lewiston-14255" => array("type"=>"village", "pop"=>2628, "name_ru"=>"Льюистон", "name_en"=>"Lewiston", "lang"=>"en", "lat"=>43.171650, "lon"=>-79.040752),
"/US/US-NY/liberty-14256" => array("type"=>"village", "pop"=>3860, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>41.797460, "lon"=>-74.746360),
"/US/US-NY/lido-beach-14257" => array("type"=>"cdp", "pop"=>2786, "name_en"=>"Lido Beach", "lang"=>"en", "lat"=>40.592150, "lon"=>-73.607755),
"/US/US-NY/lima-14258" => array("type"=>"village", "pop"=>2385, "name_ru"=>"Лима", "name_en"=>"Lima", "lang"=>"en", "lat"=>42.906448, "lon"=>-77.613088),
"/US/US-NY/lime-lake-machias-14259" => array("type"=>"cdp", "pop"=>1269, "name_en"=>"Lime Lake-Machias", "lang"=>"en", "lat"=>42.428364, "lon"=>-78.482323),
"/US/US-NY/limestone-14260" => array("type"=>"village", "pop"=>385, "name_ru"=>"Лаймстоун", "name_en"=>"Limestone", "lang"=>"en", "lat"=>42.024008, "lon"=>-78.630977),
"/US/US-NY/lincoln-park-14261" => array("type"=>"cdp", "pop"=>2457, "name_ru"=>"Линкольн Парк", "name_en"=>"Lincoln Park", "lang"=>"en", "lat"=>41.957609, "lon"=>-74.002665),
"/US/US-NY/lindenhurst-14262" => array("type"=>"village", "pop"=>28620, "name_ru"=>"Линденхерст", "name_en"=>"Lindenhurst", "lang"=>"en", "lat"=>40.685650, "lon"=>-73.371820),
"/US/US-NY/lisle-14263" => array("type"=>"village", "pop"=>284, "name_ru"=>"Лайл", "name_en"=>"Lisle", "lang"=>"en", "lat"=>42.350695, "lon"=>-76.003739),
"/US/US-NY/little-falls-14264" => array("type"=>"city", "pop"=>4867, "name_ru"=>"Литтл Фоллс", "name_en"=>"Little Falls", "lang"=>"en", "lat"=>43.045431, "lon"=>-74.855872),
"/US/US-NY/little-valley-14265" => array("type"=>"village", "pop"=>1060, "name_en"=>"Little Valley", "lang"=>"en", "lat"=>42.249276, "lon"=>-78.799334),
"/US/US-NY/liverpool-14266" => array("type"=>"village", "pop"=>2343, "name_ru"=>"Ливерпул", "name_en"=>"Liverpool", "lang"=>"en", "lat"=>43.105900, "lon"=>-76.209916),
"/US/US-NY/livingston-manor-14267" => array("type"=>"cdp", "pop"=>1482, "name_en"=>"Livingston Manor", "lang"=>"en", "lat"=>41.895064, "lon"=>-74.826549),
"/US/US-NY/livonia-14268" => array("type"=>"village", "pop"=>1611, "name_ru"=>"Ливония", "name_en"=>"Livonia", "lang"=>"en", "lat"=>42.820929, "lon"=>-77.668763),
"/US/US-NY/lloyd-harbor-14269" => array("type"=>"village", "pop"=>3739, "name_en"=>"Lloyd Harbor", "lang"=>"en", "lat"=>40.917533, "lon"=>-73.464050),
"/US/US-NY/lockport-14270" => array("type"=>"city", "pop"=>20630, "name_ru"=>"Локпорт", "name_en"=>"Lockport", "lang"=>"en", "lat"=>43.169700, "lon"=>-78.696324),
"/US/US-NY/locust-valley-14271" => array("type"=>"cdp", "pop"=>3963, "name_en"=>"Locust Valley", "lang"=>"en", "lat"=>40.880650, "lon"=>-73.587638),
"/US/US-NY/lodi-14272" => array("type"=>"village", "pop"=>325, "name_ru"=>"Лоди", "name_en"=>"Lodi", "lang"=>"en", "lat"=>42.613296, "lon"=>-76.822494),
"/US/US-NY/long-beach-14273" => array("type"=>"city", "pop"=>35698, "name_ru"=>"Лонг Бич", "name_en"=>"Long Beach", "lang"=>"en", "lat"=>40.585150, "lon"=>-73.665599),
"/US/US-NY/lorenz-park-14274" => array("type"=>"cdp", "pop"=>1811, "name_en"=>"Lorenz Park", "lang"=>"en", "lat"=>42.263674, "lon"=>-73.768755),
"/US/US-NY/loudonville-14275" => array("type"=>"cdp", "pop"=>10822, "name_en"=>"Loudonville", "lang"=>"en", "lat"=>42.707300, "lon"=>-73.766129),
"/US/US-NY/lowville-14276" => array("type"=>"village", "pop"=>3165, "name_en"=>"Lowville", "lang"=>"en", "lat"=>43.785778, "lon"=>-75.489731),
"/US/US-NY/lynbrook-14277" => array("type"=>"village", "pop"=>19852, "name_en"=>"Lynbrook", "lang"=>"en", "lat"=>40.658667, "lon"=>-73.674396),
"/US/US-NY/lyncourt-14278" => array("type"=>"cdp", "pop"=>4516, "name_en"=>"Lyncourt", "lang"=>"en", "lat"=>43.081350, "lon"=>-76.126249),
"/US/US-NY/lyndon-14279" => array("type"=>"cdp", "pop"=>637, "name_ru"=>"Линдон", "name_en"=>"Lyndon", "lang"=>"en", "lat"=>43.027817, "lon"=>-76.040643),
"/US/US-NY/lyndonville-14280" => array("type"=>"village", "pop"=>816, "name_en"=>"Lyndonville", "lang"=>"en", "lat"=>43.321871, "lon"=>-78.387043),
"/US/US-NY/lyons-14281" => array("type"=>"village", "pop"=>3419, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>43.063429, "lon"=>-76.993618),
"/US/US-NY/lyons-falls-14282" => array("type"=>"village", "pop"=>537, "name_en"=>"Lyons Falls", "lang"=>"en", "lat"=>43.616915, "lon"=>-75.362129),
"/US/US-NY/macedon-14283" => array("type"=>"village", "pop"=>1485, "name_en"=>"Macedon", "lang"=>"en", "lat"=>43.068690, "lon"=>-77.302701),
"/US/US-NY/madison-14285" => array("type"=>"village", "pop"=>301, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>42.897863, "lon"=>-75.510998),
"/US/US-NY/mahopac-14286" => array("type"=>"cdp", "pop"=>7755, "name_en"=>"Mahopac", "lang"=>"en", "lat"=>41.369534, "lon"=>-73.741587),
"/US/US-NY/malone-14287" => array("type"=>"village", "pop"=>5793, "name_ru"=>"Малон", "name_en"=>"Malone", "lang"=>"en", "lat"=>44.848219, "lon"=>-74.291410),
"/US/US-NY/malverne-14288" => array("type"=>"village", "pop"=>8880, "name_en"=>"Malverne", "lang"=>"en", "lat"=>40.674500, "lon"=>-73.672574),
"/US/US-NY/mamaroneck-14289" => array("type"=>"village", "pop"=>18413, "name_en"=>"Mamaroneck", "lang"=>"en", "lat"=>40.934544, "lon"=>-73.727366),
"/US/US-NY/manchester-14290" => array("type"=>"village", "pop"=>1416, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>42.968614, "lon"=>-77.231629),
"/US/US-NY/manhasset-14291" => array("type"=>"cdp", "pop"=>7718, "name_en"=>"Manhasset", "lang"=>"en", "lat"=>40.791527, "lon"=>-73.692813),
"/US/US-NY/manhasset-hills-14292" => array("type"=>"cdp", "pop"=>3722, "name_en"=>"Manhasset Hills", "lang"=>"en", "lat"=>40.759150, "lon"=>-73.680254),
"/US/US-NY/manlius-14293" => array("type"=>"village", "pop"=>4579, "name_ru"=>"Манлиус", "name_en"=>"Manlius", "lang"=>"en", "lat"=>43.000867, "lon"=>-75.982479),
"/US/US-NY/mannsville-14294" => array("type"=>"village", "pop"=>411, "name_en"=>"Mannsville", "lang"=>"en", "lat"=>43.715220, "lon"=>-76.066636),
"/US/US-NY/manorhaven-14295" => array("type"=>"village", "pop"=>6478, "name_en"=>"Manorhaven", "lang"=>"en", "lat"=>40.841850, "lon"=>-73.714185),
"/US/US-NY/manorville-14296" => array("type"=>"cdp", "pop"=>6198, "name_en"=>"Manorville", "lang"=>"en", "lat"=>40.858451, "lon"=>-72.794035),
"/US/US-NY/marathon-14297" => array("type"=>"village", "pop"=>1019, "name_ru"=>"Марафон", "name_en"=>"Marathon", "lang"=>"en", "lat"=>42.442815, "lon"=>-76.035070),
"/US/US-NY/marcellus-14298" => array("type"=>"village", "pop"=>1814, "name_ru"=>"Марселлус", "name_en"=>"Marcellus", "lang"=>"en", "lat"=>42.984113, "lon"=>-76.339986),
"/US/US-NY/margaretville-14299" => array("type"=>"village", "pop"=>626, "name_en"=>"Margaretville", "lang"=>"en", "lat"=>42.144399, "lon"=>-74.652098),
"/US/US-NY/marlboro-14300" => array("type"=>"cdp", "pop"=>2200, "name_en"=>"Marlboro", "lang"=>"en", "lat"=>41.603184, "lon"=>-73.973285),
"/US/US-NY/massapequa-14301" => array("type"=>"cdp", "pop"=>22018, "name_en"=>"Massapequa", "lang"=>"en", "lat"=>40.668250, "lon"=>-73.471803),
"/US/US-NY/massapequa-park-14302" => array("type"=>"village", "pop"=>17385, "name_en"=>"Massapequa Park", "lang"=>"en", "lat"=>40.680800, "lon"=>-73.449827),
"/US/US-NY/massena-14303" => array("type"=>"village", "pop"=>10539, "name_ru"=>"Массена", "name_en"=>"Massena", "lang"=>"en", "lat"=>44.928500, "lon"=>-74.893440),
"/US/US-NY/mastic-14304" => array("type"=>"cdp", "pop"=>13778, "name_en"=>"Mastic", "lang"=>"en", "lat"=>40.801992, "lon"=>-72.844406),
"/US/US-NY/mastic-beach-14305" => array("type"=>"cdp", "pop"=>10293, "name_en"=>"Mastic Beach", "lang"=>"en", "lat"=>40.761464, "lon"=>-72.840250),
"/US/US-NY/matinecock-14306" => array("type"=>"village", "pop"=>863, "name_en"=>"Matinecock", "lang"=>"en", "lat"=>40.864367, "lon"=>-73.584038),
"/US/US-NY/mattituck-14307" => array("type"=>"cdp", "pop"=>3902, "name_en"=>"Mattituck", "lang"=>"en", "lat"=>41.002738, "lon"=>-72.546440),
"/US/US-NY/mattydale-14308" => array("type"=>"cdp", "pop"=>6418, "name_en"=>"Mattydale", "lang"=>"en", "lat"=>43.099381, "lon"=>-76.145950),
"/US/US-NY/maybrook-14309" => array("type"=>"village", "pop"=>4006, "name_en"=>"Maybrook", "lang"=>"en", "lat"=>41.485696, "lon"=>-74.214349),
"/US/US-NY/mayfield-14310" => array("type"=>"village", "pop"=>790, "name_ru"=>"Мейфилд", "name_en"=>"Mayfield", "lang"=>"en", "lat"=>43.101973, "lon"=>-74.266496),
"/US/US-NY/mayville-14311" => array("type"=>"village", "pop"=>1685, "name_ru"=>"Мейвилл", "name_en"=>"Mayville", "lang"=>"en", "lat"=>42.252908, "lon"=>-79.502318),
"/US/US-NY/mcgraw-14284" => array("type"=>"village", "pop"=>954, "name_en"=>"McGraw", "lang"=>"en", "lat"=>42.592038, "lon"=>-76.098190),
"/US/US-NY/mechanicville-14312" => array("type"=>"city", "pop"=>4857, "name_en"=>"Mechanicville", "lang"=>"en", "lat"=>42.904046, "lon"=>-73.690112),
"/US/US-NY/medford-14313" => array("type"=>"cdp", "pop"=>21274, "name_ru"=>"Медфорд", "name_en"=>"Medford", "lang"=>"en", "lat"=>40.822448, "lon"=>-72.984717),
"/US/US-NY/medina-14314" => array("type"=>"village", "pop"=>6046, "name_ru"=>"Медина", "name_en"=>"Medina", "lang"=>"en", "lat"=>43.219806, "lon"=>-78.387980),
"/US/US-NY/melrose-park-14315" => array("type"=>"cdp", "pop"=>2091, "name_ru"=>"Мелроуз Парк", "name_en"=>"Melrose Park", "lang"=>"en", "lat"=>42.902313, "lon"=>-76.528294),
"/US/US-NY/melville-14316" => array("type"=>"cdp", "pop"=>12586, "name_ru"=>"Меовилл", "name_en"=>"Melville", "lang"=>"en", "lat"=>40.782050, "lon"=>-73.411966),
"/US/US-NY/menands-14317" => array("type"=>"village", "pop"=>3795, "name_en"=>"Menands", "lang"=>"en", "lat"=>42.691189, "lon"=>-73.728647),
"/US/US-NY/meridian-14318" => array("type"=>"village", "pop"=>335, "name_ru"=>"Меридиан", "name_en"=>"Meridian", "lang"=>"en", "lat"=>43.163464, "lon"=>-76.535382),
"/US/US-NY/merrick-14319" => array("type"=>"cdp", "pop"=>23042, "name_en"=>"Merrick", "lang"=>"en", "lat"=>40.650650, "lon"=>-73.554638),
"/US/US-NY/mexico-14320" => array("type"=>"village", "pop"=>1531, "name_ru"=>"Мексико", "name_en"=>"Mexico", "lang"=>"en", "lat"=>43.464104, "lon"=>-76.234992),
"/US/US-NY/middle-island-14322" => array("type"=>"cdp", "pop"=>7848, "name_en"=>"Middle Island", "lang"=>"en", "lat"=>40.884997, "lon"=>-72.943904),
"/US/US-NY/middleburgh-14321" => array("type"=>"village", "pop"=>1605, "name_en"=>"Middleburgh", "lang"=>"en", "lat"=>42.598920, "lon"=>-74.329899),
"/US/US-NY/middleport-14323" => array("type"=>"village", "pop"=>1779, "name_en"=>"Middleport", "lang"=>"en", "lat"=>43.211698, "lon"=>-78.475987),
"/US/US-NY/middletown-14324" => array("type"=>"city", "pop"=>25865, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>41.445661, "lon"=>-74.422466),
"/US/US-NY/middleville-14325" => array("type"=>"village", "pop"=>535, "name_ru"=>"Миддлвилл", "name_en"=>"Middleville", "lang"=>"en", "lat"=>43.138320, "lon"=>-74.971056),
"/US/US-NY/milford-14326" => array("type"=>"village", "pop"=>479, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>42.590024, "lon"=>-74.947320),
"/US/US-NY/mill-neck-14330" => array("type"=>"village", "pop"=>879, "name_en"=>"Mill Neck", "lang"=>"en", "lat"=>40.880100, "lon"=>-73.556200),
"/US/US-NY/millbrook-14327" => array("type"=>"village", "pop"=>1509, "name_ru"=>"Милбрук", "name_en"=>"Millbrook", "lang"=>"en", "lat"=>41.784312, "lon"=>-73.694731),
"/US/US-NY/miller-place-14328" => array("type"=>"cdp", "pop"=>9315, "name_en"=>"Miller Place", "lang"=>"en", "lat"=>40.932507, "lon"=>-72.987353),
"/US/US-NY/millerton-14329" => array("type"=>"village", "pop"=>902, "name_ru"=>"Миллертон", "name_en"=>"Millerton", "lang"=>"en", "lat"=>41.953643, "lon"=>-73.511225),
"/US/US-NY/millport-14331" => array("type"=>"village", "pop"=>283, "name_ru"=>"Милпорт", "name_en"=>"Millport", "lang"=>"en", "lat"=>42.267560, "lon"=>-76.836443),
"/US/US-NY/milton-14332" => array("type"=>"cdp", "pop"=>16925, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>43.033537, "lon"=>-73.852987),
"/US/US-NY/milton-14333" => array("type"=>"cdp", "pop"=>1140, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>41.657584, "lon"=>-73.961268),
"/US/US-NY/mineola-14334" => array("type"=>"village", "pop"=>19108, "name_en"=>"Mineola", "lang"=>"en", "lat"=>40.746967, "lon"=>-73.638649),
"/US/US-NY/minetto-14335" => array("type"=>"cdp", "pop"=>1651, "name_en"=>"Minetto", "lang"=>"en", "lat"=>43.399029, "lon"=>-76.481472),
"/US/US-NY/mineville-witherbee-14336" => array("type"=>"cdp", "pop"=>1740, "name_en"=>"Mineville-Witherbee", "lang"=>"en", "lat"=>44.089070, "lon"=>-73.527536),
"/US/US-NY/minoa-14337" => array("type"=>"village", "pop"=>3281, "name_en"=>"Minoa", "lang"=>"en", "lat"=>43.075450, "lon"=>-76.010426),
"/US/US-NY/mohawk-14338" => array("type"=>"village", "pop"=>2471, "name_en"=>"Mohawk", "lang"=>"en", "lat"=>43.009439, "lon"=>-75.007218),
"/US/US-NY/monroe-14339" => array("type"=>"village", "pop"=>8168, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>41.320631, "lon"=>-74.186293),
"/US/US-NY/monsey-14340" => array("type"=>"cdp", "pop"=>13986, "name_en"=>"Monsey", "lang"=>"en", "lat"=>41.118067, "lon"=>-74.067735),
"/US/US-NY/montauk-14341" => array("type"=>"cdp", "pop"=>3001, "name_en"=>"Montauk", "lang"=>"en", "lat"=>41.047464, "lon"=>-71.945247),
"/US/US-NY/montebello-14342" => array("type"=>"village", "pop"=>4007, "name_ru"=>"Монтебелло", "name_en"=>"Montebello", "lang"=>"en", "lat"=>41.128800, "lon"=>-74.118100),
"/US/US-NY/montgomery-14343" => array("type"=>"village", "pop"=>4741, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>41.522023, "lon"=>-74.238214),
"/US/US-NY/monticello-14344" => array("type"=>"village", "pop"=>6490, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>41.652490, "lon"=>-74.688017),
"/US/US-NY/montour-falls-14345" => array("type"=>"village", "pop"=>1769, "name_en"=>"Montour Falls", "lang"=>"en", "lat"=>42.350235, "lon"=>-76.849367),
"/US/US-NY/mooers-14346" => array("type"=>"village", "pop"=>3420, "name_en"=>"Mooers", "lang"=>"en", "lat"=>44.965269, "lon"=>-73.585930),
"/US/US-NY/moravia-14347" => array("type"=>"village", "pop"=>1293, "name_ru"=>"Моравия", "name_en"=>"Moravia", "lang"=>"en", "lat"=>42.712061, "lon"=>-76.422538),
"/US/US-NY/morris-14348" => array("type"=>"village", "pop"=>548, "name_ru"=>"Моррис", "name_en"=>"Morris", "lang"=>"en", "lat"=>42.547988, "lon"=>-75.245519),
"/US/US-NY/morrisonville-14349" => array("type"=>"cdp", "pop"=>1742, "name_ru"=>"Моррисонвилл", "name_en"=>"Morrisonville", "lang"=>"en", "lat"=>44.690679, "lon"=>-73.551189),
"/US/US-NY/morristown-14350" => array("type"=>"village", "pop"=>432, "name_ru"=>"Морристаун", "name_en"=>"Morristown", "lang"=>"en", "lat"=>44.583214, "lon"=>-75.645084),
"/US/US-NY/morrisville-14351" => array("type"=>"village", "pop"=>2172, "name_ru"=>"Моррисвилл", "name_en"=>"Morrisville", "lang"=>"en", "lat"=>42.898001, "lon"=>-75.644947),
"/US/US-NY/mount-ivy-14352" => array("type"=>"cdp", "pop"=>6013, "name_en"=>"Mount Ivy", "lang"=>"en", "lat"=>41.192893, "lon"=>-74.028681),
"/US/US-NY/mount-kisco-14353" => array("type"=>"village", "pop"=>10393, "name_en"=>"Mount Kisco", "lang"=>"en", "lat"=>41.201832, "lon"=>-73.730469),
"/US/US-NY/mount-morris-14354" => array("type"=>"village", "pop"=>2876, "name_ru"=>"Маунт Моррис", "name_en"=>"Mount Morris", "lang"=>"en", "lat"=>42.723720, "lon"=>-77.874984),
"/US/US-NY/mount-sinai-14355" => array("type"=>"cdp", "pop"=>8023, "name_en"=>"Mount Sinai", "lang"=>"en", "lat"=>40.938434, "lon"=>-73.019119),
"/US/US-NY/mount-vernon-14356" => array("type"=>"city", "pop"=>68653, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>40.912950, "lon"=>-73.829799),
"/US/US-NY/munnsville-14357" => array("type"=>"village", "pop"=>412, "name_en"=>"Munnsville", "lang"=>"en", "lat"=>42.977121, "lon"=>-75.587560),
"/US/US-NY/munsey-park-14358" => array("type"=>"village", "pop"=>2604, "name_en"=>"Munsey Park", "lang"=>"en", "lat"=>40.799371, "lon"=>-73.680133),
"/US/US-NY/munsons-corners-14359" => array("type"=>"cdp", "pop"=>2436, "name_en"=>"Munsons Corners", "lang"=>"en", "lat"=>42.580057, "lon"=>-76.205256),
"/US/US-NY/muttontown-14360" => array("type"=>"village", "pop"=>3696, "name_en"=>"Muttontown", "lang"=>"en", "lat"=>40.820600, "lon"=>-73.541138),
"/US/US-NY/myers-corner-14361" => array("type"=>"cdp", "pop"=>5599, "name_en"=>"Myers Corner", "lang"=>"en", "lat"=>41.597131, "lon"=>-73.865835),
"/US/US-NY/nanuet-14362" => array("type"=>"cdp", "pop"=>14065, "name_en"=>"Nanuet", "lang"=>"en", "lat"=>41.095200, "lon"=>-74.016038),
"/US/US-NY/napanoch-14363" => array("type"=>"cdp", "pop"=>1068, "name_en"=>"Napanoch", "lang"=>"en", "lat"=>41.750848, "lon"=>-74.374471),
"/US/US-NY/naples-14364" => array("type"=>"village", "pop"=>1019, "name_ru"=>"Нэплс", "name_en"=>"Naples", "lang"=>"en", "lat"=>42.616883, "lon"=>-77.401905),
"/US/US-NY/nassau-14365" => array("type"=>"village", "pop"=>1116, "name_ru"=>"Нассо", "name_en"=>"Nassau", "lang"=>"en", "lat"=>42.513717, "lon"=>-73.611581),
"/US/US-NY/nelliston-14366" => array("type"=>"village", "pop"=>601, "name_en"=>"Nelliston", "lang"=>"en", "lat"=>42.931979, "lon"=>-74.608152),
"/US/US-NY/nelsonville-14367" => array("type"=>"village", "pop"=>566, "name_en"=>"Nelsonville", "lang"=>"en", "lat"=>41.429492, "lon"=>-73.948521),
"/US/US-NY/nesconset-14368" => array("type"=>"cdp", "pop"=>10712, "name_en"=>"Nesconset", "lang"=>"en", "lat"=>40.846800, "lon"=>-73.152359),
"/US/US-NY/new-berlin-14371" => array("type"=>"village", "pop"=>1098, "name_ru"=>"Нью Берлин", "name_en"=>"New Berlin", "lang"=>"en", "lat"=>42.624062, "lon"=>-75.335721),
"/US/US-NY/new-cassel-14373" => array("type"=>"cdp", "pop"=>10257, "name_en"=>"New Cassel", "lang"=>"en", "lat"=>40.759000, "lon"=>-73.565594),
"/US/US-NY/new-city-14374" => array("type"=>"cdp", "pop"=>33673, "name_en"=>"New City", "lang"=>"en", "lat"=>41.153800, "lon"=>-73.991824),
"/US/US-NY/new-hartford-14377" => array("type"=>"village", "pop"=>1802, "name_ru"=>"Нью Хартфорд", "name_en"=>"New Hartford", "lang"=>"en", "lat"=>43.071100, "lon"=>-75.289635),
"/US/US-NY/new-hempstead-14378" => array("type"=>"village", "pop"=>4838, "name_en"=>"New Hempstead", "lang"=>"en", "lat"=>41.149567, "lon"=>-74.049673),
"/US/US-NY/new-hyde-park-14379" => array("type"=>"village", "pop"=>9596, "name_en"=>"New Hyde Park", "lang"=>"en", "lat"=>40.732000, "lon"=>-73.685918),
"/US/US-NY/new-paltz-14380" => array("type"=>"village", "pop"=>6553, "name_en"=>"New Paltz", "lang"=>"en", "lat"=>41.748197, "lon"=>-74.081855),
"/US/US-NY/new-rochelle-14382" => array("type"=>"city", "pop"=>74115, "name_en"=>"New Rochelle", "lang"=>"en", "lat"=>40.923200, "lon"=>-73.779902),
"/US/US-NY/new-square-14383" => array("type"=>"village", "pop"=>6461, "name_en"=>"New Square", "lang"=>"en", "lat"=>41.139650, "lon"=>-74.028612),
"/US/US-NY/new-windsor-14384" => array("type"=>"cdp", "pop"=>25149, "name_ru"=>"Нью Виндзор", "name_en"=>"New Windsor", "lang"=>"en", "lat"=>41.470150, "lon"=>-74.027150),
"/US/US-NY/new-york-14385" => array("type"=>"city", "pop"=>8363710, "name_en"=>"New York", "lang"=>"en", "lat"=>40.669800, "lon"=>-73.943849),
"/US/US-NY/new-york-mills-14386" => array("type"=>"village", "pop"=>3315, "name_ru"=>"Нью Йорк Миллс", "name_en"=>"New York Mills", "lang"=>"en", "lat"=>43.100333, "lon"=>-75.293245),
"/US/US-NY/newark-14369" => array("type"=>"village", "pop"=>9096, "name_ru"=>"Ньюарк", "name_en"=>"Newark", "lang"=>"en", "lat"=>43.041900, "lon"=>-77.094294),
"/US/US-NY/newark-valley-14370" => array("type"=>"village", "pop"=>1002, "name_en"=>"Newark Valley", "lang"=>"en", "lat"=>42.222984, "lon"=>-76.185908),
"/US/US-NY/newburgh-14372" => array("type"=>"city", "pop"=>28101, "name_ru"=>"Ньюберг", "name_en"=>"Newburgh", "lang"=>"en", "lat"=>41.503100, "lon"=>-74.020052),
"/US/US-NY/newfane-14375" => array("type"=>"cdp", "pop"=>9334, "name_en"=>"Newfane", "lang"=>"en", "lat"=>43.287950, "lon"=>-78.699299),
"/US/US-NY/newfield-hamlet-14376" => array("type"=>"cdp", "pop"=>692, "name_en"=>"Newfield Hamlet", "lang"=>"en", "lat"=>42.357930, "lon"=>-76.594058),
"/US/US-NY/newport-14381" => array("type"=>"village", "pop"=>614, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>43.187324, "lon"=>-75.014979),
"/US/US-NY/niagara-falls-14387" => array("type"=>"city", "pop"=>51345, "name_en"=>"Niagara Falls", "lang"=>"en", "lat"=>43.091150, "lon"=>-79.016248),
"/US/US-NY/nichols-14388" => array("type"=>"village", "pop"=>534, "name_ru"=>"Николс", "name_en"=>"Nichols", "lang"=>"en", "lat"=>42.019912, "lon"=>-76.368846),
"/US/US-NY/niskayuna-14389" => array("type"=>"cdp", "pop"=>21753, "name_en"=>"Niskayuna", "lang"=>"en", "lat"=>42.817150, "lon"=>-73.898008),
"/US/US-NY/nissequogue-14390" => array("type"=>"village", "pop"=>1604, "name_en"=>"Nissequogue", "lang"=>"en", "lat"=>40.905680, "lon"=>-73.188425),
"/US/US-NY/niverville-14391" => array("type"=>"cdp", "pop"=>1809, "name_en"=>"Niverville", "lang"=>"en", "lat"=>42.445487, "lon"=>-73.648810),
"/US/US-NY/norfolk-14392" => array("type"=>"cdp", "pop"=>4585, "name_en"=>"Norfolk", "lang"=>"en", "lat"=>44.800124, "lon"=>-74.985957),
"/US/US-NY/north-amityville-14393" => array("type"=>"cdp", "pop"=>13849, "name_en"=>"North Amityville", "lang"=>"en", "lat"=>40.700043, "lon"=>-73.413237),
"/US/US-NY/north-babylon-14394" => array("type"=>"cdp", "pop"=>18081, "name_en"=>"North Babylon", "lang"=>"en", "lat"=>40.731400, "lon"=>-73.325351),
"/US/US-NY/north-ballston-spa-14395" => array("type"=>"cdp", "pop"=>1362, "name_en"=>"North Ballston Spa", "lang"=>"en", "lat"=>43.019216, "lon"=>-73.852787),
"/US/US-NY/north-bay-shore-14396" => array("type"=>"cdp", "pop"=>12799, "name_en"=>"North Bay Shore", "lang"=>"en", "lat"=>40.759024, "lon"=>-73.260978),
"/US/US-NY/north-bellmore-14397" => array("type"=>"cdp", "pop"=>19707, "name_en"=>"North Bellmore", "lang"=>"en", "lat"=>40.689900, "lon"=>-73.538498),
"/US/US-NY/north-bellport-14398" => array("type"=>"cdp", "pop"=>8182, "name_en"=>"North Bellport", "lang"=>"en", "lat"=>40.788448, "lon"=>-72.946383),
"/US/US-NY/north-boston-14399" => array("type"=>"cdp", "pop"=>2581, "name_en"=>"North Boston", "lang"=>"en", "lat"=>42.672165, "lon"=>-78.784022),
"/US/US-NY/north-collins-14400" => array("type"=>"village", "pop"=>1002, "name_en"=>"North Collins", "lang"=>"en", "lat"=>42.594584, "lon"=>-78.936725),
"/US/US-NY/north-great-river-14402" => array("type"=>"cdp", "pop"=>3964, "name_en"=>"North Great River", "lang"=>"en", "lat"=>40.756272, "lon"=>-73.170087),
"/US/US-NY/north-haven-14403" => array("type"=>"village", "pop"=>914, "name_ru"=>"Норт Хейвен", "name_en"=>"North Haven", "lang"=>"en", "lat"=>41.023623, "lon"=>-72.315745),
"/US/US-NY/north-hills-14404" => array("type"=>"village", "pop"=>4589, "name_en"=>"North Hills", "lang"=>"en", "lat"=>40.776560, "lon"=>-73.680857),
"/US/US-NY/north-hornell-14405" => array("type"=>"village", "pop"=>817, "name_en"=>"North Hornell", "lang"=>"en", "lat"=>42.345665, "lon"=>-77.662538),
"/US/US-NY/north-lindenhurst-14406" => array("type"=>"cdp", "pop"=>10563, "name_en"=>"North Lindenhurst", "lang"=>"en", "lat"=>40.706867, "lon"=>-73.385446),
"/US/US-NY/north-massapequa-14407" => array("type"=>"cdp", "pop"=>19365, "name_en"=>"North Massapequa", "lang"=>"en", "lat"=>40.702800, "lon"=>-73.466620),
"/US/US-NY/north-merrick-14408" => array("type"=>"cdp", "pop"=>12113, "name_en"=>"North Merrick", "lang"=>"en", "lat"=>40.687150, "lon"=>-73.560821),
"/US/US-NY/north-new-hyde-park-14409" => array("type"=>"cdp", "pop"=>14359, "name_en"=>"North New Hyde Park", "lang"=>"en", "lat"=>40.745800, "lon"=>-73.687823),
"/US/US-NY/north-patchogue-14410" => array("type"=>"cdp", "pop"=>7374, "name_en"=>"North Patchogue", "lang"=>"en", "lat"=>40.784384, "lon"=>-73.024449),
"/US/US-NY/north-sea-14412" => array("type"=>"cdp", "pop"=>2530, "name_en"=>"North Sea", "lang"=>"en", "lat"=>40.937272, "lon"=>-72.405994),
"/US/US-NY/north-syracuse-14413" => array("type"=>"village", "pop"=>6570, "name_en"=>"North Syracuse", "lang"=>"en", "lat"=>43.132333, "lon"=>-76.130224),
"/US/US-NY/north-tarrytown-14414" => array("type"=>"village", "pop"=>8152, "name_en"=>"North Tarrytown", "lang"=>"en", "lat"=>41.091763, "lon"=>-73.871382),
"/US/US-NY/north-tonawanda-14415" => array("type"=>"city", "pop"=>31222, "name_en"=>"North Tonawanda", "lang"=>"en", "lat"=>43.043750, "lon"=>-78.865949),
"/US/US-NY/north-valley-stream-14416" => array("type"=>"cdp", "pop"=>14574, "name_en"=>"North Valley Stream", "lang"=>"en", "lat"=>40.684433, "lon"=>-73.707625),
"/US/US-NY/north-wantagh-14418" => array("type"=>"cdp", "pop"=>12276, "name_en"=>"North Wantagh", "lang"=>"en", "lat"=>40.699133, "lon"=>-73.508682),
"/US/US-NY/northeast-ithaca-14401" => array("type"=>"cdp", "pop"=>2533, "name_en"=>"Northeast Ithaca", "lang"=>"en", "lat"=>42.471214, "lon"=>-76.468031),
"/US/US-NY/northport-14411" => array("type"=>"village", "pop"=>7676, "name_ru"=>"Нортпорт", "name_en"=>"Northport", "lang"=>"en", "lat"=>40.903700, "lon"=>-73.345551),
"/US/US-NY/northville-14417" => array("type"=>"village", "pop"=>1153, "name_ru"=>"Нортвилл", "name_en"=>"Northville", "lang"=>"en", "lat"=>43.221972, "lon"=>-74.171023),
"/US/US-NY/northwest-harbor-14419" => array("type"=>"cdp", "pop"=>2167, "name_en"=>"Northwest Harbor", "lang"=>"en", "lat"=>41.009664, "lon"=>-72.221652),
"/US/US-NY/northwest-ithaca-14420" => array("type"=>"cdp", "pop"=>1144, "name_en"=>"Northwest Ithaca", "lang"=>"en", "lat"=>42.470765, "lon"=>-76.543187),
"/US/US-NY/norwich-14421" => array("type"=>"city", "pop"=>6992, "name_ru"=>"Норвич", "name_en"=>"Norwich", "lang"=>"en", "lat"=>42.534970, "lon"=>-75.523685),
"/US/US-NY/norwood-14422" => array("type"=>"village", "pop"=>1584, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>44.748360, "lon"=>-74.997628),
"/US/US-NY/noyack-14423" => array("type"=>"cdp", "pop"=>2059, "name_en"=>"Noyack", "lang"=>"en", "lat"=>40.976916, "lon"=>-72.333207),
"/US/US-NY/nunda-14424" => array("type"=>"village", "pop"=>1287, "name_en"=>"Nunda", "lang"=>"en", "lat"=>42.580564, "lon"=>-77.938667),
"/US/US-NY/nyack-14425" => array("type"=>"village", "pop"=>6833, "name_en"=>"Nyack", "lang"=>"en", "lat"=>41.090077, "lon"=>-73.915029),
"/US/US-NY/oakdale-14426" => array("type"=>"cdp", "pop"=>7875, "name_ru"=>"Оукдейл", "name_en"=>"Oakdale", "lang"=>"en", "lat"=>40.737739, "lon"=>-73.134201),
"/US/US-NY/oakfield-14427" => array("type"=>"village", "pop"=>1654, "name_en"=>"Oakfield", "lang"=>"en", "lat"=>43.064394, "lon"=>-78.270702),
"/US/US-NY/ocean-beach-14428" => array("type"=>"village", "pop"=>147, "name_en"=>"Ocean Beach", "lang"=>"en", "lat"=>40.645472, "lon"=>-73.156420),
"/US/US-NY/oceanside-14429" => array("type"=>"cdp", "pop"=>32423, "name_ru"=>"Ошенсайд", "name_en"=>"Oceanside", "lang"=>"en", "lat"=>40.632450, "lon"=>-73.635796),
"/US/US-NY/odessa-14430" => array("type"=>"village", "pop"=>583, "name_ru"=>"Одесса", "name_en"=>"Odessa", "lang"=>"en", "lat"=>42.311077, "lon"=>-76.817478),
"/US/US-NY/ogdensburg-14431" => array("type"=>"city", "pop"=>11119, "name_en"=>"Ogdensburg", "lang"=>"en", "lat"=>44.707011, "lon"=>-75.470434),
"/US/US-NY/olcott-14432" => array("type"=>"cdp", "pop"=>1432, "name_en"=>"Olcott", "lang"=>"en", "lat"=>43.335350, "lon"=>-78.712312),
"/US/US-NY/old-bethpage-14433" => array("type"=>"cdp", "pop"=>5610, "name_en"=>"Old Bethpage", "lang"=>"en", "lat"=>40.754267, "lon"=>-73.455504),
"/US/US-NY/old-brookville-14434" => array("type"=>"village", "pop"=>2314, "name_en"=>"Old Brookville", "lang"=>"en", "lat"=>40.827518, "lon"=>-73.604907),
"/US/US-NY/old-field-14435" => array("type"=>"village", "pop"=>1010, "name_en"=>"Old Field", "lang"=>"en", "lat"=>40.962272, "lon"=>-73.130766),
"/US/US-NY/old-westbury-14436" => array("type"=>"village", "pop"=>5283, "name_en"=>"Old Westbury", "lang"=>"en", "lat"=>40.786250, "lon"=>-73.596226),
"/US/US-NY/olean-14437" => array("type"=>"city", "pop"=>14149, "name_ru"=>"Олин", "name_en"=>"Olean", "lang"=>"en", "lat"=>42.082089, "lon"=>-78.432313),
"/US/US-NY/oneida-14438" => array("type"=>"city", "pop"=>10723, "name_ru"=>"Онида", "name_en"=>"Oneida", "lang"=>"en", "lat"=>43.075759, "lon"=>-75.663819),
"/US/US-NY/oneida-castle-14439" => array("type"=>"village", "pop"=>600, "name_en"=>"Oneida Castle", "lang"=>"en", "lat"=>43.083856, "lon"=>-75.632335),
"/US/US-NY/oneonta-14440" => array("type"=>"city", "pop"=>13208, "name_ru"=>"Онеонта", "name_en"=>"Oneonta", "lang"=>"en", "lat"=>42.454719, "lon"=>-75.067292),
"/US/US-NY/orange-lake-14442" => array("type"=>"cdp", "pop"=>5196, "name_en"=>"Orange Lake", "lang"=>"en", "lat"=>41.531833, "lon"=>-74.094158),
"/US/US-NY/orangeburg-14441" => array("type"=>"cdp", "pop"=>3583, "name_en"=>"Orangeburg", "lang"=>"en", "lat"=>41.050000, "lon"=>-73.945646),
"/US/US-NY/orchard-park-14443" => array("type"=>"village", "pop"=>3050, "name_en"=>"Orchard Park", "lang"=>"en", "lat"=>42.763350, "lon"=>-78.745824),
"/US/US-NY/oriskany-14444" => array("type"=>"village", "pop"=>1404, "name_en"=>"Oriskany", "lang"=>"en", "lat"=>43.155950, "lon"=>-75.332831),
"/US/US-NY/oriskany-falls-14445" => array("type"=>"village", "pop"=>665, "name_en"=>"Oriskany Falls", "lang"=>"en", "lat"=>42.938019, "lon"=>-75.463275),
"/US/US-NY/ossining-14446" => array("type"=>"village", "pop"=>23798, "name_en"=>"Ossining", "lang"=>"en", "lat"=>41.158484, "lon"=>-73.871459),
"/US/US-NY/oswego-14447" => array("type"=>"city", "pop"=>17351, "name_ru"=>"Освего", "name_en"=>"Oswego", "lang"=>"en", "lat"=>43.458392, "lon"=>-76.502945),
"/US/US-NY/otego-14448" => array("type"=>"village", "pop"=>984, "name_en"=>"Otego", "lang"=>"en", "lat"=>42.391604, "lon"=>-75.178407),
"/US/US-NY/otisville-14449" => array("type"=>"village", "pop"=>1085, "name_ru"=>"Отисвилл", "name_en"=>"Otisville", "lang"=>"en", "lat"=>41.470621, "lon"=>-74.539859),
"/US/US-NY/ovid-14450" => array("type"=>"village", "pop"=>590, "name_ru"=>"Овид", "name_en"=>"Ovid", "lang"=>"en", "lat"=>42.675339, "lon"=>-76.821565),
"/US/US-NY/owego-14451" => array("type"=>"village", "pop"=>3708, "name_en"=>"Owego", "lang"=>"en", "lat"=>42.104800, "lon"=>-76.262662),
"/US/US-NY/oxford-14452" => array("type"=>"village", "pop"=>1534, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>42.440577, "lon"=>-75.596171),
"/US/US-NY/oyster-bay-14453" => array("type"=>"cdp", "pop"=>301474, "name_en"=>"Oyster Bay", "lang"=>"en", "lat"=>40.868600, "lon"=>-73.533091),
"/US/US-NY/oyster-bay-cove-14454" => array("type"=>"village", "pop"=>2299, "name_en"=>"Oyster Bay Cove", "lang"=>"en", "lat"=>40.859200, "lon"=>-73.504391),
"/US/US-NY/painted-post-14455" => array("type"=>"village", "pop"=>1779, "name_en"=>"Painted Post", "lang"=>"en", "lat"=>42.162898, "lon"=>-77.091868),
"/US/US-NY/palatine-bridge-14456" => array("type"=>"village", "pop"=>696, "name_en"=>"Palatine Bridge", "lang"=>"en", "lat"=>42.910817, "lon"=>-74.575242),
"/US/US-NY/palmyra-14457" => array("type"=>"village", "pop"=>3411, "name_ru"=>"Палмира", "name_en"=>"Palmyra", "lang"=>"en", "lat"=>43.061600, "lon"=>-77.230691),
"/US/US-NY/panama-14458" => array("type"=>"village", "pop"=>460, "name_ru"=>"Панама", "name_en"=>"Panama", "lang"=>"en", "lat"=>42.075006, "lon"=>-79.486130),
"/US/US-NY/parish-14459" => array("type"=>"village", "pop"=>484, "name_en"=>"Parish", "lang"=>"en", "lat"=>43.404312, "lon"=>-76.128940),
"/US/US-NY/patchogue-14460" => array("type"=>"village", "pop"=>12194, "name_en"=>"Patchogue", "lang"=>"en", "lat"=>40.761307, "lon"=>-73.019370),
"/US/US-NY/pawling-14461" => array("type"=>"village", "pop"=>2245, "name_en"=>"Pawling", "lang"=>"en", "lat"=>41.563344, "lon"=>-73.599892),
"/US/US-NY/peach-lake-14462" => array("type"=>"cdp", "pop"=>1499, "name_en"=>"Peach Lake", "lang"=>"en", "lat"=>41.367364, "lon"=>-73.578367),
"/US/US-NY/pearl-river-14463" => array("type"=>"cdp", "pop"=>15314, "name_ru"=>"Перл Ривер", "name_en"=>"Pearl River", "lang"=>"en", "lat"=>41.060550, "lon"=>-74.003225),
"/US/US-NY/peconic-14464" => array("type"=>"cdp", "pop"=>1100, "name_en"=>"Peconic", "lang"=>"en", "lat"=>41.038984, "lon"=>-72.464930),
"/US/US-NY/peekskill-14465" => array("type"=>"city", "pop"=>24484, "name_en"=>"Peekskill", "lang"=>"en", "lat"=>41.289415, "lon"=>-73.923879),
"/US/US-NY/pelham-14466" => array("type"=>"village", "pop"=>6435, "name_ru"=>"Пелхэм", "name_en"=>"Pelham", "lang"=>"en", "lat"=>40.911000, "lon"=>-73.808315),
"/US/US-NY/pelham-manor-14467" => array("type"=>"village", "pop"=>5451, "name_en"=>"Pelham Manor", "lang"=>"en", "lat"=>40.893650, "lon"=>-73.805207),
"/US/US-NY/penn-yan-14468" => array("type"=>"village", "pop"=>5162, "name_en"=>"Penn Yan", "lang"=>"en", "lat"=>42.661175, "lon"=>-77.054873),
"/US/US-NY/perry-14469" => array("type"=>"village", "pop"=>3667, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>42.717324, "lon"=>-78.006733),
"/US/US-NY/perrysburg-14470" => array("type"=>"village", "pop"=>380, "name_en"=>"Perrysburg", "lang"=>"en", "lat"=>42.459485, "lon"=>-79.001499),
"/US/US-NY/peru-14471" => array("type"=>"cdp", "pop"=>6882, "name_ru"=>"Перу", "name_en"=>"Peru", "lang"=>"en", "lat"=>44.580015, "lon"=>-73.534465),
"/US/US-NY/phelps-14472" => array("type"=>"village", "pop"=>1903, "name_ru"=>"Фэлпс", "name_en"=>"Phelps", "lang"=>"en", "lat"=>42.958333, "lon"=>-77.062604),
"/US/US-NY/philadelphia-14473" => array("type"=>"village", "pop"=>1547, "name_ru"=>"Филадельфия", "name_en"=>"Philadelphia", "lang"=>"en", "lat"=>44.154763, "lon"=>-75.709423),
"/US/US-NY/philmont-14474" => array("type"=>"village", "pop"=>1366, "name_en"=>"Philmont", "lang"=>"en", "lat"=>42.248338, "lon"=>-73.646665),
"/US/US-NY/phoenix-14475" => array("type"=>"village", "pop"=>2138, "name_ru"=>"Феникс", "name_en"=>"Phoenix", "lang"=>"en", "lat"=>43.231808, "lon"=>-76.297265),
"/US/US-NY/piermont-14476" => array("type"=>"village", "pop"=>2566, "name_en"=>"Piermont", "lang"=>"en", "lat"=>41.039600, "lon"=>-73.915544),
"/US/US-NY/pike-14477" => array("type"=>"village", "pop"=>354, "name_en"=>"Pike", "lang"=>"en", "lat"=>42.555871, "lon"=>-78.154225),
"/US/US-NY/pine-bush-14478" => array("type"=>"cdp", "pop"=>1445, "name_en"=>"Pine Bush", "lang"=>"en", "lat"=>41.608930, "lon"=>-74.297395),
"/US/US-NY/pine-plains-14479" => array("type"=>"cdp", "pop"=>2689, "name_en"=>"Pine Plains", "lang"=>"en", "lat"=>41.977664, "lon"=>-73.658416),
"/US/US-NY/pine-valley-14480" => array("type"=>"village", "pop"=>1486, "name_ru"=>"Пайн Валли", "name_en"=>"Pine Valley", "lang"=>"en", "lat"=>40.904507, "lon"=>-72.679573),
"/US/US-NY/pittsford-14481" => array("type"=>"village", "pop"=>1320, "name_en"=>"Pittsford", "lang"=>"en", "lat"=>43.090351, "lon"=>-77.517120),
"/US/US-NY/plainedge-14482" => array("type"=>"cdp", "pop"=>8739, "name_en"=>"Plainedge", "lang"=>"en", "lat"=>40.724417, "lon"=>-73.477821),
"/US/US-NY/plainview-14483" => array("type"=>"cdp", "pop"=>26207, "name_ru"=>"Плейнвью", "name_en"=>"Plainview", "lang"=>"en", "lat"=>40.783300, "lon"=>-73.472649),
"/US/US-NY/plandome-14484" => array("type"=>"village", "pop"=>1273, "name_en"=>"Plandome", "lang"=>"en", "lat"=>40.807550, "lon"=>-73.700427),
"/US/US-NY/plandome-heights-14485" => array("type"=>"village", "pop"=>961, "name_en"=>"Plandome Heights", "lang"=>"en", "lat"=>40.801133, "lon"=>-73.705260),
"/US/US-NY/plandome-manor-14486" => array("type"=>"village", "pop"=>870, "name_en"=>"Plandome Manor", "lang"=>"en", "lat"=>40.813500, "lon"=>-73.697598),
"/US/US-NY/plattsburgh-14487" => array("type"=>"city", "pop"=>19393, "name_en"=>"Plattsburgh", "lang"=>"en", "lat"=>44.695461, "lon"=>-73.456825),
"/US/US-NY/plattsburgh-afb-14488" => array("type"=>"cdp", "pop"=>5483, "name_en"=>"Plattsburgh AFB", "lang"=>"en", "lat"=>44.663475, "lon"=>-73.453587),
"/US/US-NY/plattsburgh-west-14489" => array("type"=>"cdp", "pop"=>1274, "name_en"=>"Plattsburgh West", "lang"=>"en", "lat"=>44.682560, "lon"=>-73.503173),
"/US/US-NY/pleasant-valley-14490" => array("type"=>"cdp", "pop"=>9763, "name_ru"=>"Плезант Валли", "name_en"=>"Pleasant Valley", "lang"=>"en", "lat"=>41.748677, "lon"=>-73.820545),
"/US/US-NY/pleasantville-14491" => array("type"=>"village", "pop"=>7146, "name_ru"=>"Плезантвилл", "name_en"=>"Pleasantville", "lang"=>"en", "lat"=>41.138069, "lon"=>-73.784709),
"/US/US-NY/poestenkill-14492" => array("type"=>"cdp", "pop"=>4279, "name_en"=>"Poestenkill", "lang"=>"en", "lat"=>42.693550, "lon"=>-73.564502),
"/US/US-NY/poland-14493" => array("type"=>"village", "pop"=>432, "name_en"=>"Poland", "lang"=>"en", "lat"=>43.226120, "lon"=>-75.061475),
"/US/US-NY/pomona-14494" => array("type"=>"village", "pop"=>3001, "name_ru"=>"Помона", "name_en"=>"Pomona", "lang"=>"en", "lat"=>41.189850, "lon"=>-74.053709),
"/US/US-NY/poquott-14495" => array("type"=>"village", "pop"=>992, "name_en"=>"Poquott", "lang"=>"en", "lat"=>40.952218, "lon"=>-73.089180),
"/US/US-NY/port-byron-14496" => array("type"=>"village", "pop"=>1233, "name_ru"=>"Порт Бирон", "name_en"=>"Port Byron", "lang"=>"en", "lat"=>43.036729, "lon"=>-76.625773),
"/US/US-NY/port-chester-14497" => array("type"=>"village", "pop"=>28171, "name_en"=>"Port Chester", "lang"=>"en", "lat"=>41.004500, "lon"=>-73.667700),
"/US/US-NY/port-dickinson-14498" => array("type"=>"village", "pop"=>1595, "name_en"=>"Port Dickinson", "lang"=>"en", "lat"=>42.136583, "lon"=>-75.892040),
"/US/US-NY/port-ewen-14499" => array("type"=>"cdp", "pop"=>3444, "name_en"=>"Port Ewen", "lang"=>"en", "lat"=>41.906240, "lon"=>-73.978898),
"/US/US-NY/port-henry-14500" => array("type"=>"village", "pop"=>1047, "name_en"=>"Port Henry", "lang"=>"en", "lat"=>44.047013, "lon"=>-73.460685),
"/US/US-NY/port-jefferson-14501" => array("type"=>"village", "pop"=>8012, "name_en"=>"Port Jefferson", "lang"=>"en", "lat"=>40.947613, "lon"=>-73.059002),
"/US/US-NY/port-jefferson-station-14502" => array("type"=>"cdp", "pop"=>7232, "name_en"=>"Port Jefferson Station", "lang"=>"en", "lat"=>40.923477, "lon"=>-73.067740),
"/US/US-NY/port-jervis-14503" => array("type"=>"city", "pop"=>9111, "name_en"=>"Port Jervis", "lang"=>"en", "lat"=>41.377825, "lon"=>-74.691289),
"/US/US-NY/port-leyden-14504" => array("type"=>"village", "pop"=>605, "name_en"=>"Port Leyden", "lang"=>"en", "lat"=>43.583973, "lon"=>-75.341049),
"/US/US-NY/port-washington-14506" => array("type"=>"cdp", "pop"=>15387, "name_en"=>"Port Washington", "lang"=>"en", "lat"=>40.827230, "lon"=>-73.680312),
"/US/US-NY/port-washington-north-14507" => array("type"=>"village", "pop"=>3327, "name_en"=>"Port Washington North", "lang"=>"en", "lat"=>40.843304, "lon"=>-73.702512),
"/US/US-NY/portville-14505" => array("type"=>"village", "pop"=>951, "name_en"=>"Portville", "lang"=>"en", "lat"=>42.035604, "lon"=>-78.339126),
"/US/US-NY/potsdam-14508" => array("type"=>"village", "pop"=>9841, "name_en"=>"Potsdam", "lang"=>"en", "lat"=>44.668071, "lon"=>-74.985564),
"/US/US-NY/poughkeepsie-14509" => array("type"=>"city", "pop"=>29654, "name_en"=>"Poughkeepsie", "lang"=>"en", "lat"=>41.695800, "lon"=>-73.922473),
"/US/US-NY/prospect-14510" => array("type"=>"village", "pop"=>322, "name_ru"=>"Проспект", "name_en"=>"Prospect", "lang"=>"en", "lat"=>43.304240, "lon"=>-75.152845),
"/US/US-NY/pulaski-14511" => array("type"=>"village", "pop"=>2282, "name_ru"=>"Пуласки", "name_en"=>"Pulaski", "lang"=>"en", "lat"=>43.565309, "lon"=>-76.128185),
"/US/US-NY/putnam-lake-14512" => array("type"=>"cdp", "pop"=>3459, "name_en"=>"Putnam Lake", "lang"=>"en", "lat"=>41.477661, "lon"=>-73.549975),
"/US/US-NY/quogue-14513" => array("type"=>"village", "pop"=>1157, "name_en"=>"Quogue", "lang"=>"en", "lat"=>40.817552, "lon"=>-72.597497),
"/US/US-NY/randolph-14514" => array("type"=>"village", "pop"=>1211, "name_ru"=>"Рэндолф", "name_en"=>"Randolph", "lang"=>"en", "lat"=>42.162263, "lon"=>-78.979856),
"/US/US-NY/ransomville-14515" => array("type"=>"cdp", "pop"=>1542, "name_en"=>"Ransomville", "lang"=>"en", "lat"=>43.237800, "lon"=>-78.909079),
"/US/US-NY/rapids-14516" => array("type"=>"cdp", "pop"=>1152, "name_en"=>"Rapids", "lang"=>"en", "lat"=>43.101267, "lon"=>-78.641383),
"/US/US-NY/ravena-14517" => array("type"=>"village", "pop"=>3333, "name_en"=>"Ravena", "lang"=>"en", "lat"=>42.476576, "lon"=>-73.812293),
"/US/US-NY/red-creek-14518" => array("type"=>"village", "pop"=>485, "name_en"=>"Red Creek", "lang"=>"en", "lat"=>43.248017, "lon"=>-76.723071),
"/US/US-NY/red-hook-14519" => array("type"=>"village", "pop"=>1881, "name_en"=>"Red Hook", "lang"=>"en", "lat"=>41.995648, "lon"=>-73.876630),
"/US/US-NY/red-oaks-mill-14520" => array("type"=>"cdp", "pop"=>4906, "name_en"=>"Red Oaks Mill", "lang"=>"en", "lat"=>41.655700, "lon"=>-73.873552),
"/US/US-NY/remsen-14521" => array("type"=>"village", "pop"=>510, "name_ru"=>"Ремсен", "name_en"=>"Remsen", "lang"=>"en", "lat"=>43.327520, "lon"=>-75.187125),
"/US/US-NY/remsenburg-speonk-14522" => array("type"=>"cdp", "pop"=>1851, "name_en"=>"Remsenburg-Speonk", "lang"=>"en", "lat"=>40.813360, "lon"=>-72.705819),
"/US/US-NY/rensselaer-14523" => array("type"=>"city", "pop"=>7907, "name_ru"=>"Ренсселаер", "name_en"=>"Rensselaer", "lang"=>"en", "lat"=>42.644433, "lon"=>-73.735313),
"/US/US-NY/rensselaer-falls-14524" => array("type"=>"village", "pop"=>322, "name_en"=>"Rensselaer Falls", "lang"=>"en", "lat"=>44.590815, "lon"=>-75.320226),
"/US/US-NY/rhinebeck-14525" => array("type"=>"village", "pop"=>3049, "name_en"=>"Rhinebeck", "lang"=>"en", "lat"=>41.928688, "lon"=>-73.907533),
"/US/US-NY/richburg-14526" => array("type"=>"village", "pop"=>428, "name_en"=>"Richburg", "lang"=>"en", "lat"=>42.089867, "lon"=>-78.155982),
"/US/US-NY/richfield-springs-14527" => array("type"=>"village", "pop"=>1172, "name_en"=>"Richfield Springs", "lang"=>"en", "lat"=>42.854355, "lon"=>-74.986215),
"/US/US-NY/richmondville-14528" => array("type"=>"village", "pop"=>797, "name_en"=>"Richmondville", "lang"=>"en", "lat"=>42.633173, "lon"=>-74.564539),
"/US/US-NY/richville-14529" => array("type"=>"village", "pop"=>260, "name_ru"=>"Ричвилл", "name_en"=>"Richville", "lang"=>"en", "lat"=>44.415332, "lon"=>-75.390891),
"/US/US-NY/ridge-14530" => array("type"=>"cdp", "pop"=>11734, "name_en"=>"Ridge", "lang"=>"en", "lat"=>40.910184, "lon"=>-72.882051),
"/US/US-NY/ripley-14531" => array("type"=>"cdp", "pop"=>2653, "name_ru"=>"Рипли", "name_en"=>"Ripley", "lang"=>"en", "lat"=>42.265625, "lon"=>-79.711436),
"/US/US-NY/riverhead-14532" => array("type"=>"cdp", "pop"=>35383, "name_en"=>"Riverhead", "lang"=>"en", "lat"=>40.948720, "lon"=>-72.675109),
"/US/US-NY/riverside-14533" => array("type"=>"village", "pop"=>560, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>42.156022, "lon"=>-77.081108),
"/US/US-NY/riverside-14534" => array("type"=>"cdp", "pop"=>1300, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>40.881200, "lon"=>-72.678327),
"/US/US-NY/rochester-14535" => array("type"=>"city", "pop"=>206886, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>43.168651, "lon"=>-77.615838),
"/US/US-NY/rockville-centre-14536" => array("type"=>"village", "pop"=>24414, "name_en"=>"Rockville Centre", "lang"=>"en", "lat"=>40.663850, "lon"=>-73.638314),
"/US/US-NY/rocky-point-14537" => array("type"=>"cdp", "pop"=>8596, "name_en"=>"Rocky Point", "lang"=>"en", "lat"=>40.927056, "lon"=>-72.923370),
"/US/US-NY/roessleville-14538" => array("type"=>"cdp", "pop"=>10753, "name_en"=>"Roessleville", "lang"=>"en", "lat"=>42.699550, "lon"=>-73.803709),
"/US/US-NY/rome-14539" => array("type"=>"city", "pop"=>33673, "name_ru"=>"Роум", "name_en"=>"Rome", "lang"=>"en", "lat"=>43.225517, "lon"=>-75.489256),
"/US/US-NY/ronkonkoma-14540" => array("type"=>"cdp", "pop"=>20391, "name_en"=>"Ronkonkoma", "lang"=>"en", "lat"=>40.803456, "lon"=>-73.124873),
"/US/US-NY/roosevelt-14541" => array("type"=>"cdp", "pop"=>15030, "name_ru"=>"Рузвельт", "name_en"=>"Roosevelt", "lang"=>"en", "lat"=>40.679600, "lon"=>-73.583917),
"/US/US-NY/rosendale-village-14542" => array("type"=>"cdp", "pop"=>1284, "name_en"=>"Rosendale Village", "lang"=>"en", "lat"=>41.849303, "lon"=>-74.076198),
"/US/US-NY/roslyn-14543" => array("type"=>"village", "pop"=>3092, "name_en"=>"Roslyn", "lang"=>"en", "lat"=>40.799650, "lon"=>-73.648774),
"/US/US-NY/roslyn-estates-14544" => array("type"=>"village", "pop"=>1220, "name_en"=>"Roslyn Estates", "lang"=>"en", "lat"=>40.794050, "lon"=>-73.662323),
"/US/US-NY/roslyn-harbor-14545" => array("type"=>"village", "pop"=>1065, "name_en"=>"Roslyn Harbor", "lang"=>"en", "lat"=>40.818350, "lon"=>-73.640637),
"/US/US-NY/roslyn-heights-14546" => array("type"=>"cdp", "pop"=>6405, "name_en"=>"Roslyn Heights", "lang"=>"en", "lat"=>40.778750, "lon"=>-73.640619),
"/US/US-NY/rotterdam-14547" => array("type"=>"cdp", "pop"=>30024, "name_en"=>"Rotterdam", "lang"=>"en", "lat"=>42.778350, "lon"=>-73.954531),
"/US/US-NY/round-lake-14548" => array("type"=>"village", "pop"=>600, "name_ru"=>"Раунд Лейк", "name_en"=>"Round Lake", "lang"=>"en", "lat"=>42.937200, "lon"=>-73.796100),
"/US/US-NY/rouses-point-14549" => array("type"=>"village", "pop"=>2320, "name_en"=>"Rouses Point", "lang"=>"en", "lat"=>44.992028, "lon"=>-73.365974),
"/US/US-NY/rushville-14550" => array("type"=>"village", "pop"=>642, "name_ru"=>"Рашвилл", "name_en"=>"Rushville", "lang"=>"en", "lat"=>42.760632, "lon"=>-77.227482),
"/US/US-NY/russell-gardens-14551" => array("type"=>"village", "pop"=>1063, "name_en"=>"Russell Gardens", "lang"=>"en", "lat"=>40.781267, "lon"=>-73.725351),
"/US/US-NY/rye-14552" => array("type"=>"city", "pop"=>15060, "name_ru"=>"Рай", "name_en"=>"Rye", "lang"=>"en", "lat"=>40.952867, "lon"=>-73.683696),
"/US/US-NY/rye-brook-14553" => array("type"=>"village", "pop"=>9568, "name_en"=>"Rye Brook", "lang"=>"en", "lat"=>41.030410, "lon"=>-73.686595),
"/US/US-NY/sackets-harbor-14554" => array("type"=>"village", "pop"=>1424, "name_en"=>"Sackets Harbor", "lang"=>"en", "lat"=>43.941563, "lon"=>-76.122220),
"/US/US-NY/saddle-rock-14555" => array("type"=>"village", "pop"=>845, "name_en"=>"Saddle Rock", "lang"=>"en", "lat"=>40.794350, "lon"=>-73.749285),
"/US/US-NY/sag-harbor-14556" => array("type"=>"village", "pop"=>2428, "name_en"=>"Sag Harbor", "lang"=>"en", "lat"=>40.996368, "lon"=>-72.290207),
"/US/US-NY/salamanca-14560" => array("type"=>"city", "pop"=>5592, "name_en"=>"Salamanca", "lang"=>"en", "lat"=>42.164256, "lon"=>-78.722745),
"/US/US-NY/salem-14561" => array("type"=>"village", "pop"=>884, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>43.174079, "lon"=>-73.328293),
"/US/US-NY/salisbury-14562" => array("type"=>"cdp", "pop"=>1894, "name_ru"=>"Солсбери", "name_en"=>"Salisbury", "lang"=>"en", "lat"=>40.745650, "lon"=>-73.560581),
"/US/US-NY/saltaire-14563" => array("type"=>"village", "pop"=>45, "name_en"=>"Saltaire", "lang"=>"en", "lat"=>40.637985, "lon"=>-73.195065),
"/US/US-NY/sand-ridge-14564" => array("type"=>"cdp", "pop"=>1312, "name_en"=>"Sand Ridge", "lang"=>"en", "lat"=>43.255896, "lon"=>-76.230448),
"/US/US-NY/sands-point-14565" => array("type"=>"village", "pop"=>2915, "name_en"=>"Sands Point", "lang"=>"en", "lat"=>40.853050, "lon"=>-73.702982),
"/US/US-NY/sandy-creek-14566" => array("type"=>"village", "pop"=>748, "name_en"=>"Sandy Creek", "lang"=>"en", "lat"=>43.642854, "lon"=>-76.086081),
"/US/US-NY/saranac-lake-14567" => array("type"=>"village", "pop"=>4809, "name_en"=>"Saranac Lake", "lang"=>"en", "lat"=>44.325014, "lon"=>-74.133070),
"/US/US-NY/saratoga-springs-14568" => array("type"=>"city", "pop"=>28844, "name_en"=>"Saratoga Springs", "lang"=>"en", "lat"=>43.067555, "lon"=>-73.778837),
"/US/US-NY/saugerties-14569" => array("type"=>"village", "pop"=>3857, "name_en"=>"Saugerties", "lang"=>"en", "lat"=>42.074776, "lon"=>-73.948322),
"/US/US-NY/saugerties-south-14570" => array("type"=>"cdp", "pop"=>2346, "name_en"=>"Saugerties South", "lang"=>"en", "lat"=>42.058856, "lon"=>-73.952315),
"/US/US-NY/savona-14571" => array("type"=>"village", "pop"=>774, "name_en"=>"Savona", "lang"=>"en", "lat"=>42.284089, "lon"=>-77.223217),
"/US/US-NY/sayville-14572" => array("type"=>"cdp", "pop"=>16550, "name_en"=>"Sayville", "lang"=>"en", "lat"=>40.746784, "lon"=>-73.085823),
"/US/US-NY/scarsdale-14573" => array("type"=>"village", "pop"=>17695, "name_en"=>"Scarsdale", "lang"=>"en", "lat"=>40.988433, "lon"=>-73.776402),
"/US/US-NY/schaghticoke-14574" => array("type"=>"village", "pop"=>667, "name_en"=>"Schaghticoke", "lang"=>"en", "lat"=>42.899474, "lon"=>-73.586568),
"/US/US-NY/schenectady-14575" => array("type"=>"city", "pop"=>61152, "name_en"=>"Schenectady", "lang"=>"en", "lat"=>42.801800, "lon"=>-73.928102),
"/US/US-NY/schenevus-14576" => array("type"=>"village", "pop"=>513, "name_en"=>"Schenevus", "lang"=>"en", "lat"=>42.547316, "lon"=>-74.824612),
"/US/US-NY/schoharie-14577" => array("type"=>"village", "pop"=>1002, "name_en"=>"Schoharie", "lang"=>"en", "lat"=>42.666748, "lon"=>-74.312867),
"/US/US-NY/schuylerville-14578" => array("type"=>"village", "pop"=>1349, "name_en"=>"Schuylerville", "lang"=>"en", "lat"=>43.102002, "lon"=>-73.580554),
"/US/US-NY/scotchtown-14579" => array("type"=>"cdp", "pop"=>8765, "name_en"=>"Scotchtown", "lang"=>"en", "lat"=>41.476176, "lon"=>-74.368622),
"/US/US-NY/scotia-14580" => array("type"=>"village", "pop"=>8033, "name_en"=>"Scotia", "lang"=>"en", "lat"=>42.832000, "lon"=>-73.960845),
"/US/US-NY/scottsville-14581" => array("type"=>"village", "pop"=>2040, "name_ru"=>"Скоттсвилл", "name_en"=>"Scottsville", "lang"=>"en", "lat"=>43.021751, "lon"=>-77.753818),
"/US/US-NY/sea-cliff-14582" => array("type"=>"village", "pop"=>5033, "name_en"=>"Sea Cliff", "lang"=>"en", "lat"=>40.848500, "lon"=>-73.651329),
"/US/US-NY/seaford-14583" => array("type"=>"cdp", "pop"=>15597, "name_ru"=>"Сифорд", "name_en"=>"Seaford", "lang"=>"en", "lat"=>40.667843, "lon"=>-73.492986),
"/US/US-NY/searingtown-14584" => array("type"=>"cdp", "pop"=>5020, "name_en"=>"Searingtown", "lang"=>"en", "lat"=>40.770122, "lon"=>-73.660058),
"/US/US-NY/selden-14585" => array("type"=>"cdp", "pop"=>20608, "name_ru"=>"Селден", "name_en"=>"Selden", "lang"=>"en", "lat"=>40.869448, "lon"=>-73.047928),
"/US/US-NY/seneca-falls-14586" => array("type"=>"village", "pop"=>6629, "name_en"=>"Seneca Falls", "lang"=>"en", "lat"=>42.909427, "lon"=>-76.795771),
"/US/US-NY/setauket-east-setauket-14587" => array("type"=>"cdp", "pop"=>13634, "name_en"=>"Setauket-East Setauket", "lang"=>"en", "lat"=>40.929909, "lon"=>-73.102529),
"/US/US-NY/sharon-springs-14588" => array("type"=>"village", "pop"=>525, "name_ru"=>"Шэрон Спрингс", "name_en"=>"Sharon Springs", "lang"=>"en", "lat"=>42.794949, "lon"=>-74.617538),
"/US/US-NY/shelter-island-14589" => array("type"=>"cdp", "pop"=>2534, "name_en"=>"Shelter Island", "lang"=>"en", "lat"=>41.059984, "lon"=>-72.320218),
"/US/US-NY/shelter-island-heights-14590" => array("type"=>"cdp", "pop"=>1042, "name_en"=>"Shelter Island Heights", "lang"=>"en", "lat"=>41.075704, "lon"=>-72.342771),
"/US/US-NY/sherburne-14591" => array("type"=>"village", "pop"=>1412, "name_en"=>"Sherburne", "lang"=>"en", "lat"=>42.680331, "lon"=>-75.497536),
"/US/US-NY/sherman-14592" => array("type"=>"village", "pop"=>658, "name_ru"=>"Шерман", "name_en"=>"Sherman", "lang"=>"en", "lat"=>42.160066, "lon"=>-79.594915),
"/US/US-NY/sherrill-14593" => array("type"=>"city", "pop"=>3115, "name_ru"=>"Шеррилл", "name_en"=>"Sherrill", "lang"=>"en", "lat"=>43.069960, "lon"=>-75.600000),
"/US/US-NY/shinnecock-hills-14594" => array("type"=>"cdp", "pop"=>2847, "name_en"=>"Shinnecock Hills", "lang"=>"en", "lat"=>40.895501, "lon"=>-72.456422),
"/US/US-NY/shirley-14595" => array("type"=>"cdp", "pop"=>22936, "name_ru"=>"Ширли", "name_en"=>"Shirley", "lang"=>"en", "lat"=>40.795192, "lon"=>-72.874061),
"/US/US-NY/shoreham-14596" => array("type"=>"village", "pop"=>431, "name_ru"=>"Шорхэм", "name_en"=>"Shoreham", "lang"=>"en", "lat"=>40.955776, "lon"=>-72.906466),
"/US/US-NY/shortsville-14597" => array("type"=>"village", "pop"=>1313, "name_en"=>"Shortsville", "lang"=>"en", "lat"=>42.956100, "lon"=>-77.222255),
"/US/US-NY/sidney-14598" => array("type"=>"village", "pop"=>3692, "name_ru"=>"Сидни", "name_en"=>"Sidney", "lang"=>"en", "lat"=>42.305889, "lon"=>-75.397744),
"/US/US-NY/silver-creek-14599" => array("type"=>"village", "pop"=>2799, "name_ru"=>"Силвер Крик", "name_en"=>"Silver Creek", "lang"=>"en", "lat"=>42.543544, "lon"=>-79.168872),
"/US/US-NY/silver-springs-14600" => array("type"=>"village", "pop"=>785, "name_en"=>"Silver Springs", "lang"=>"en", "lat"=>42.660062, "lon"=>-78.084724),
"/US/US-NY/sinclairville-14601" => array("type"=>"village", "pop"=>626, "name_en"=>"Sinclairville", "lang"=>"en", "lat"=>42.264390, "lon"=>-79.258194),
"/US/US-NY/skaneateles-14602" => array("type"=>"village", "pop"=>2539, "name_en"=>"Skaneateles", "lang"=>"en", "lat"=>42.944967, "lon"=>-76.425463),
"/US/US-NY/sloan-14603" => array("type"=>"village", "pop"=>3450, "name_ru"=>"Слоан", "name_en"=>"Sloan", "lang"=>"en", "lat"=>42.892200, "lon"=>-78.791652),
"/US/US-NY/sloatsburg-14604" => array("type"=>"village", "pop"=>3142, "name_en"=>"Sloatsburg", "lang"=>"en", "lat"=>41.160629, "lon"=>-74.191248),
"/US/US-NY/smithtown-14605" => array("type"=>"cdp", "pop"=>121162, "name_en"=>"Smithtown", "lang"=>"en", "lat"=>40.857112, "lon"=>-73.214899),
"/US/US-NY/smyrna-14606" => array("type"=>"village", "pop"=>233, "name_ru"=>"Смирна", "name_en"=>"Smyrna", "lang"=>"en", "lat"=>42.685847, "lon"=>-75.567501),
"/US/US-NY/sodus-14607" => array("type"=>"village", "pop"=>1611, "name_en"=>"Sodus", "lang"=>"en", "lat"=>43.236710, "lon"=>-77.064061),
"/US/US-NY/sodus-point-14608" => array("type"=>"village", "pop"=>1108, "name_en"=>"Sodus Point", "lang"=>"en", "lat"=>43.262642, "lon"=>-76.994244),
"/US/US-NY/solvay-14609" => array("type"=>"village", "pop"=>6417, "name_en"=>"Solvay", "lang"=>"en", "lat"=>43.058283, "lon"=>-76.211989),
"/US/US-NY/sound-beach-14610" => array("type"=>"cdp", "pop"=>9102, "name_en"=>"Sound Beach", "lang"=>"en", "lat"=>40.955320, "lon"=>-72.966935),
"/US/US-NY/south-corning-14613" => array("type"=>"village", "pop"=>1181, "name_en"=>"South Corning", "lang"=>"en", "lat"=>42.124012, "lon"=>-77.036869),
"/US/US-NY/south-dayton-14614" => array("type"=>"village", "pop"=>609, "name_en"=>"South Dayton", "lang"=>"en", "lat"=>42.362736, "lon"=>-79.051954),
"/US/US-NY/south-fallsburg-14615" => array("type"=>"cdp", "pop"=>2115, "name_en"=>"South Fallsburg", "lang"=>"en", "lat"=>41.720550, "lon"=>-74.634830),
"/US/US-NY/south-farmingdale-14616" => array("type"=>"cdp", "pop"=>15377, "name_en"=>"South Farmingdale", "lang"=>"en", "lat"=>40.717800, "lon"=>-73.448375),
"/US/US-NY/south-floral-park-14617" => array("type"=>"village", "pop"=>1599, "name_en"=>"South Floral Park", "lang"=>"en", "lat"=>40.713050, "lon"=>-73.700935),
"/US/US-NY/south-glens-falls-14618" => array("type"=>"village", "pop"=>3382, "name_en"=>"South Glens Falls", "lang"=>"en", "lat"=>43.293380, "lon"=>-73.635268),
"/US/US-NY/south-hempstead-14619" => array("type"=>"cdp", "pop"=>3014, "name_en"=>"South Hempstead", "lang"=>"en", "lat"=>40.682750, "lon"=>-73.623730),
"/US/US-NY/south-hill-14620" => array("type"=>"cdp", "pop"=>5423, "name_en"=>"South Hill", "lang"=>"en", "lat"=>42.411526, "lon"=>-76.490986),
"/US/US-NY/south-huntington-14621" => array("type"=>"cdp", "pop"=>9624, "name_en"=>"South Huntington", "lang"=>"en", "lat"=>40.820833, "lon"=>-73.392164),
"/US/US-NY/south-lockport-14622" => array("type"=>"cdp", "pop"=>7112, "name_en"=>"South Lockport", "lang"=>"en", "lat"=>43.134000, "lon"=>-78.686379),
"/US/US-NY/south-nyack-14623" => array("type"=>"village", "pop"=>3378, "name_en"=>"South Nyack", "lang"=>"en", "lat"=>41.077733, "lon"=>-73.917053),
"/US/US-NY/south-valley-stream-14626" => array("type"=>"cdp", "pop"=>5328, "name_en"=>"South Valley Stream", "lang"=>"en", "lat"=>40.655900, "lon"=>-73.718122),
"/US/US-NY/southampton-14611" => array("type"=>"village", "pop"=>4323, "name_en"=>"Southampton", "lang"=>"en", "lat"=>40.880328, "lon"=>-72.395515),
"/US/US-NY/southampton-14612" => array("type"=>"cdp", "pop"=>1302, "name_en"=>"Southampton", "lang"=>"en", "lat"=>40.904248, "lon"=>-72.407813),
"/US/US-NY/southold-14624" => array("type"=>"cdp", "pop"=>22495, "name_en"=>"Southold", "lang"=>"en", "lat"=>41.059784, "lon"=>-72.427797),
"/US/US-NY/southport-14625" => array("type"=>"cdp", "pop"=>10586, "name_ru"=>"Саутпорт", "name_en"=>"Southport", "lang"=>"en", "lat"=>42.063800, "lon"=>-76.811600),
"/US/US-NY/spackenkill-14627" => array("type"=>"cdp", "pop"=>4660, "name_en"=>"Spackenkill", "lang"=>"en", "lat"=>41.655900, "lon"=>-73.914016),
"/US/US-NY/speculator-14628" => array("type"=>"village", "pop"=>327, "name_en"=>"Speculator", "lang"=>"en", "lat"=>43.580586, "lon"=>-74.380585),
"/US/US-NY/spencer-14629" => array("type"=>"village", "pop"=>683, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>42.215078, "lon"=>-76.494756),
"/US/US-NY/spencerport-14630" => array("type"=>"village", "pop"=>3464, "name_en"=>"Spencerport", "lang"=>"en", "lat"=>43.188051, "lon"=>-77.808621),
"/US/US-NY/spring-valley-14632" => array("type"=>"village", "pop"=>26291, "name_ru"=>"Спринг Валли", "name_en"=>"Spring Valley", "lang"=>"en", "lat"=>41.114350, "lon"=>-74.049239),
"/US/US-NY/springs-14631" => array("type"=>"cdp", "pop"=>4355, "name_en"=>"Springs", "lang"=>"en", "lat"=>41.020880, "lon"=>-72.157144),
"/US/US-NY/springville-14633" => array("type"=>"village", "pop"=>4181, "name_ru"=>"Спрингвилл", "name_en"=>"Springville", "lang"=>"en", "lat"=>42.509389, "lon"=>-78.670058),
"/US/US-NY/st.-bonaventure-14557" => array("type"=>"cdp", "pop"=>2397, "name_en"=>"St. Bonaventure", "lang"=>"en", "lat"=>42.080233, "lon"=>-78.475248),
"/US/US-NY/st.-james-14558" => array("type"=>"cdp", "pop"=>12703, "name_ru"=>"Сент Джеймс", "name_en"=>"St. James", "lang"=>"en", "lat"=>40.876392, "lon"=>-73.152578),
"/US/US-NY/st.-johnsville-14559" => array("type"=>"village", "pop"=>1607, "name_en"=>"St. Johnsville", "lang"=>"en", "lat"=>43.000483, "lon"=>-74.676808),
"/US/US-NY/stamford-14634" => array("type"=>"village", "pop"=>1221, "name_ru"=>"Стэмфорд", "name_en"=>"Stamford", "lang"=>"en", "lat"=>42.408876, "lon"=>-74.617604),
"/US/US-NY/stannards-14635" => array("type"=>"cdp", "pop"=>1028, "name_en"=>"Stannards", "lang"=>"en", "lat"=>42.078821, "lon"=>-77.912947),
"/US/US-NY/star-lake-14636" => array("type"=>"cdp", "pop"=>1092, "name_en"=>"Star Lake", "lang"=>"en", "lat"=>44.160169, "lon"=>-75.037541),
"/US/US-NY/stewart-manor-14637" => array("type"=>"village", "pop"=>1907, "name_en"=>"Stewart Manor", "lang"=>"en", "lat"=>40.720500, "lon"=>-73.685598),
"/US/US-NY/stillwater-14638" => array("type"=>"village", "pop"=>1681, "name_ru"=>"Стиллуотер", "name_en"=>"Stillwater", "lang"=>"en", "lat"=>42.950043, "lon"=>-73.644948),
"/US/US-NY/stony-brook-14639" => array("type"=>"cdp", "pop"=>13726, "name_en"=>"Stony Brook", "lang"=>"en", "lat"=>40.907872, "lon"=>-73.128344),
"/US/US-NY/stony-point-14640" => array("type"=>"cdp", "pop"=>15132, "name_ru"=>"Стоуни Пойнт", "name_en"=>"Stony Point", "lang"=>"en", "lat"=>41.229350, "lon"=>-73.997471),
"/US/US-NY/stottville-14641" => array("type"=>"cdp", "pop"=>1369, "name_en"=>"Stottville", "lang"=>"en", "lat"=>42.289210, "lon"=>-73.752280),
"/US/US-NY/suffern-14642" => array("type"=>"village", "pop"=>11088, "name_en"=>"Suffern", "lang"=>"en", "lat"=>41.113400, "lon"=>-74.144049),
"/US/US-NY/sylvan-beach-14643" => array("type"=>"village", "pop"=>1071, "name_en"=>"Sylvan Beach", "lang"=>"en", "lat"=>43.206620, "lon"=>-75.725525),
"/US/US-NY/syosset-14644" => array("type"=>"cdp", "pop"=>18967, "name_en"=>"Syosset", "lang"=>"en", "lat"=>40.815450, "lon"=>-73.502821),
"/US/US-NY/syracuse-14645" => array("type"=>"city", "pop"=>138068, "name_ru"=>"Сиракьюс", "name_en"=>"Syracuse", "lang"=>"en", "lat"=>43.041059, "lon"=>-76.144067),
"/US/US-NY/tannersville-14646" => array("type"=>"village", "pop"=>431, "name_en"=>"Tannersville", "lang"=>"en", "lat"=>42.193646, "lon"=>-74.134912),
"/US/US-NY/tappan-14647" => array("type"=>"cdp", "pop"=>6867, "name_en"=>"Tappan", "lang"=>"en", "lat"=>41.025800, "lon"=>-73.951599),
"/US/US-NY/tarrytown-14648" => array("type"=>"village", "pop"=>11031, "name_ru"=>"Тэрритаун", "name_en"=>"Tarrytown", "lang"=>"en", "lat"=>41.064833, "lon"=>-73.868665),
"/US/US-NY/terryville-14649" => array("type"=>"cdp", "pop"=>10275, "name_ru"=>"Терривилл", "name_en"=>"Terryville", "lang"=>"en", "lat"=>40.908456, "lon"=>-73.049775),
"/US/US-NY/theresa-14650" => array("type"=>"village", "pop"=>815, "name_en"=>"Theresa", "lang"=>"en", "lat"=>44.214808, "lon"=>-75.797168),
"/US/US-NY/thiells-14651" => array("type"=>"cdp", "pop"=>5204, "name_en"=>"Thiells", "lang"=>"en", "lat"=>41.207600, "lon"=>-74.010546),
"/US/US-NY/thomaston-14652" => array("type"=>"village", "pop"=>2600, "name_ru"=>"Томастон", "name_en"=>"Thomaston", "lang"=>"en", "lat"=>40.788000, "lon"=>-73.714535),
"/US/US-NY/thornwood-14653" => array("type"=>"cdp", "pop"=>7025, "name_en"=>"Thornwood", "lang"=>"en", "lat"=>41.113834, "lon"=>-73.764916),
"/US/US-NY/ticonderoga-14654" => array("type"=>"village", "pop"=>4963, "name_en"=>"Ticonderoga", "lang"=>"en", "lat"=>43.847009, "lon"=>-73.425646),
"/US/US-NY/tillson-14655" => array("type"=>"cdp", "pop"=>1688, "name_en"=>"Tillson", "lang"=>"en", "lat"=>41.831630, "lon"=>-74.069183),
"/US/US-NY/tivoli-14656" => array("type"=>"village", "pop"=>1135, "name_en"=>"Tivoli", "lang"=>"en", "lat"=>42.059280, "lon"=>-73.911085),
"/US/US-NY/tonawanda-14657" => array("type"=>"city", "pop"=>14819, "name_en"=>"Tonawanda", "lang"=>"en", "lat"=>43.010800, "lon"=>-78.881906),
"/US/US-NY/tonawanda-14658" => array("type"=>"cdp", "pop"=>65284, "name_en"=>"Tonawanda", "lang"=>"en", "lat"=>42.984150, "lon"=>-78.875671),
"/US/US-NY/town-line-14659" => array("type"=>"cdp", "pop"=>2721, "name_en"=>"Town Line", "lang"=>"en", "lat"=>42.884915, "lon"=>-78.558682),
"/US/US-NY/tribes-hill-14660" => array("type"=>"cdp", "pop"=>1060, "name_en"=>"Tribes Hill", "lang"=>"en", "lat"=>42.947669, "lon"=>-74.301773),
"/US/US-NY/troy-14661" => array("type"=>"city", "pop"=>47459, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>42.734350, "lon"=>-73.675469),
"/US/US-NY/trumansburg-14662" => array("type"=>"village", "pop"=>1591, "name_en"=>"Trumansburg", "lang"=>"en", "lat"=>42.540376, "lon"=>-76.660452),
"/US/US-NY/tuckahoe-14663" => array("type"=>"village", "pop"=>6244, "name_en"=>"Tuckahoe", "lang"=>"en", "lat"=>40.952700, "lon"=>-73.823196),
"/US/US-NY/tully-14664" => array("type"=>"village", "pop"=>867, "name_en"=>"Tully", "lang"=>"en", "lat"=>42.797872, "lon"=>-76.108015),
"/US/US-NY/tupper-lake-14665" => array("type"=>"village", "pop"=>3828, "name_en"=>"Tupper Lake", "lang"=>"en", "lat"=>44.234270, "lon"=>-74.468891),
"/US/US-NY/turin-14666" => array("type"=>"village", "pop"=>240, "name_ru"=>"Турин", "name_en"=>"Turin", "lang"=>"en", "lat"=>43.628851, "lon"=>-75.409734),
"/US/US-NY/tuxedo-park-14667" => array("type"=>"village", "pop"=>719, "name_en"=>"Tuxedo Park", "lang"=>"en", "lat"=>41.204134, "lon"=>-74.207219),
"/US/US-NY/unadilla-14668" => array("type"=>"village", "pop"=>1043, "name_ru"=>"Унадилла", "name_en"=>"Unadilla", "lang"=>"en", "lat"=>42.327031, "lon"=>-75.316658),
"/US/US-NY/union-springs-14670" => array("type"=>"village", "pop"=>1045, "name_ru"=>"Юнион Спрингс", "name_en"=>"Union Springs", "lang"=>"en", "lat"=>42.845036, "lon"=>-76.693333),
"/US/US-NY/uniondale-14669" => array("type"=>"cdp", "pop"=>20328, "name_ru"=>"Юниондейл", "name_en"=>"Uniondale", "lang"=>"en", "lat"=>40.702250, "lon"=>-73.591489),
"/US/US-NY/unionville-14671" => array("type"=>"village", "pop"=>561, "name_ru"=>"Юнионвилл", "name_en"=>"Unionville", "lang"=>"en", "lat"=>41.302028, "lon"=>-74.562557),
"/US/US-NY/university-gardens-14672" => array("type"=>"cdp", "pop"=>4419, "name_en"=>"University Gardens", "lang"=>"en", "lat"=>40.774750, "lon"=>-73.727909),
"/US/US-NY/upper-brookville-14673" => array("type"=>"village", "pop"=>1914, "name_en"=>"Upper Brookville", "lang"=>"en", "lat"=>40.844128, "lon"=>-73.564926),
"/US/US-NY/upper-nyack-14674" => array("type"=>"village", "pop"=>1891, "name_en"=>"Upper Nyack", "lang"=>"en", "lat"=>41.114481, "lon"=>-73.914544),
"/US/US-NY/utica-14675" => array("type"=>"city", "pop"=>58082, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>43.096850, "lon"=>-75.228380),
"/US/US-NY/vails-gate-14676" => array("type"=>"cdp", "pop"=>3014, "name_en"=>"Vails Gate", "lang"=>"en", "lat"=>41.457850, "lon"=>-74.053898),
"/US/US-NY/valatie-14677" => array("type"=>"village", "pop"=>1882, "name_en"=>"Valatie", "lang"=>"en", "lat"=>42.414759, "lon"=>-73.676664),
"/US/US-NY/valley-cottage-14678" => array("type"=>"cdp", "pop"=>9007, "name_en"=>"Valley Cottage", "lang"=>"en", "lat"=>41.115500, "lon"=>-73.943751),
"/US/US-NY/valley-falls-14679" => array("type"=>"village", "pop"=>476, "name_ru"=>"Валли Фоллс", "name_en"=>"Valley Falls", "lang"=>"en", "lat"=>42.900769, "lon"=>-73.562935),
"/US/US-NY/valley-stream-14680" => array("type"=>"village", "pop"=>36007, "name_en"=>"Valley Stream", "lang"=>"en", "lat"=>40.664450, "lon"=>-73.705235),
"/US/US-NY/van-etten-14681" => array("type"=>"village", "pop"=>550, "name_en"=>"Van Etten", "lang"=>"en", "lat"=>42.198236, "lon"=>-76.554670),
"/US/US-NY/vernon-14682" => array("type"=>"village", "pop"=>1147, "name_ru"=>"Вернон", "name_en"=>"Vernon", "lang"=>"en", "lat"=>43.080339, "lon"=>-75.538431),
"/US/US-NY/victor-14683" => array("type"=>"village", "pop"=>2560, "name_ru"=>"Виктор", "name_en"=>"Victor", "lang"=>"en", "lat"=>42.982628, "lon"=>-77.411707),
"/US/US-NY/victory-14684" => array("type"=>"village", "pop"=>535, "name_en"=>"Victory", "lang"=>"en", "lat"=>43.093552, "lon"=>-73.592368),
"/US/US-NY/village-green-14685" => array("type"=>"cdp", "pop"=>4198, "name_en"=>"Village Green", "lang"=>"en", "lat"=>43.133300, "lon"=>-76.313439),
"/US/US-NY/village-of-the-branch-14686" => array("type"=>"village", "pop"=>1968, "name_en"=>"Village of the Branch", "lang"=>"en", "lat"=>40.851259, "lon"=>-73.184635),
"/US/US-NY/viola-14687" => array("type"=>"cdp", "pop"=>4504, "name_ru"=>"Виола", "name_en"=>"Viola", "lang"=>"en", "lat"=>41.129383, "lon"=>-74.082468),
"/US/US-NY/voorheesville-14688" => array("type"=>"village", "pop"=>2767, "name_en"=>"Voorheesville", "lang"=>"en", "lat"=>42.646430, "lon"=>-73.933237),
"/US/US-NY/waddington-14689" => array("type"=>"village", "pop"=>926, "name_en"=>"Waddington", "lang"=>"en", "lat"=>44.858888, "lon"=>-75.196549),
"/US/US-NY/wading-river-14690" => array("type"=>"cdp", "pop"=>5317, "name_en"=>"Wading River", "lang"=>"en", "lat"=>40.947064, "lon"=>-72.819904),
"/US/US-NY/walden-14691" => array("type"=>"village", "pop"=>6977, "name_ru"=>"Уолден", "name_en"=>"Walden", "lang"=>"en", "lat"=>41.559749, "lon"=>-74.188625),
"/US/US-NY/wallkill-14692" => array("type"=>"cdp", "pop"=>27450, "name_en"=>"Wallkill", "lang"=>"en", "lat"=>41.608659, "lon"=>-74.164742),
"/US/US-NY/walton-14693" => array("type"=>"village", "pop"=>2846, "name_ru"=>"Уолтон", "name_en"=>"Walton", "lang"=>"en", "lat"=>42.169717, "lon"=>-75.130717),
"/US/US-NY/wampsville-14694" => array("type"=>"village", "pop"=>550, "name_en"=>"Wampsville", "lang"=>"en", "lat"=>43.077715, "lon"=>-75.709562),
"/US/US-NY/wantagh-14695" => array("type"=>"cdp", "pop"=>18567, "name_en"=>"Wantagh", "lang"=>"en", "lat"=>40.667700, "lon"=>-73.510815),
"/US/US-NY/wappingers-falls-14696" => array("type"=>"village", "pop"=>5659, "name_en"=>"Wappingers Falls", "lang"=>"en", "lat"=>41.599000, "lon"=>-73.918496),
"/US/US-NY/warrensburg-14697" => array("type"=>"cdp", "pop"=>4321, "name_ru"=>"Уорренсбург", "name_en"=>"Warrensburg", "lang"=>"en", "lat"=>43.501788, "lon"=>-73.775443),
"/US/US-NY/warsaw-14698" => array("type"=>"village", "pop"=>3622, "name_ru"=>"Варшава", "name_en"=>"Warsaw", "lang"=>"en", "lat"=>42.741397, "lon"=>-78.140964),
"/US/US-NY/warwick-14699" => array("type"=>"village", "pop"=>6724, "name_ru"=>"Уорвик", "name_en"=>"Warwick", "lang"=>"en", "lat"=>41.256173, "lon"=>-74.355507),
"/US/US-NY/washington-heights-14700" => array("type"=>"cdp", "pop"=>1159, "name_en"=>"Washington Heights", "lang"=>"en", "lat"=>41.469508, "lon"=>-74.417617),
"/US/US-NY/washingtonville-14701" => array("type"=>"village", "pop"=>6139, "name_en"=>"Washingtonville", "lang"=>"en", "lat"=>41.429674, "lon"=>-74.157565),
"/US/US-NY/waterford-14702" => array("type"=>"village", "pop"=>2119, "name_ru"=>"Уотерфорд", "name_en"=>"Waterford", "lang"=>"en", "lat"=>42.791776, "lon"=>-73.679353),
"/US/US-NY/waterloo-14703" => array("type"=>"village", "pop"=>4994, "name_ru"=>"Ватрелоо", "name_en"=>"Waterloo", "lang"=>"en", "lat"=>42.904498, "lon"=>-76.859356),
"/US/US-NY/watermill-14704" => array("type"=>"cdp", "pop"=>1893, "name_en"=>"Watermill", "lang"=>"en", "lat"=>40.929024, "lon"=>-72.348187),
"/US/US-NY/watertown-14705" => array("type"=>"city", "pop"=>27310, "name_ru"=>"Уотертаун", "name_en"=>"Watertown", "lang"=>"en", "lat"=>43.973611, "lon"=>-75.910189),
"/US/US-NY/waterville-14706" => array("type"=>"village", "pop"=>1819, "name_ru"=>"Уотервилл", "name_en"=>"Waterville", "lang"=>"en", "lat"=>42.931013, "lon"=>-75.381267),
"/US/US-NY/watervliet-14707" => array("type"=>"city", "pop"=>9765, "name_ru"=>"Уотервлит", "name_en"=>"Watervliet", "lang"=>"en", "lat"=>42.724750, "lon"=>-73.707540),
"/US/US-NY/watkins-glen-14708" => array("type"=>"village", "pop"=>2029, "name_en"=>"Watkins Glen", "lang"=>"en", "lat"=>42.380308, "lon"=>-76.867803),
"/US/US-NY/waverly-14709" => array("type"=>"village", "pop"=>4303, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>42.012199, "lon"=>-76.541131),
"/US/US-NY/wayland-14710" => array("type"=>"village", "pop"=>1789, "name_ru"=>"Уэйленд", "name_en"=>"Wayland", "lang"=>"en", "lat"=>42.568064, "lon"=>-77.592141),
"/US/US-NY/webster-14711" => array("type"=>"village", "pop"=>5185, "name_ru"=>"Вебстер", "name_en"=>"Webster", "lang"=>"en", "lat"=>43.214134, "lon"=>-77.422406),
"/US/US-NY/weedsport-14712" => array("type"=>"village", "pop"=>1915, "name_en"=>"Weedsport", "lang"=>"en", "lat"=>43.047035, "lon"=>-76.564688),
"/US/US-NY/wellsburg-14713" => array("type"=>"village", "pop"=>601, "name_ru"=>"Уэллсбург", "name_en"=>"Wellsburg", "lang"=>"en", "lat"=>42.012816, "lon"=>-76.729605),
"/US/US-NY/wellsville-14714" => array("type"=>"village", "pop"=>4896, "name_ru"=>"Уэллсвилл", "name_en"=>"Wellsville", "lang"=>"en", "lat"=>42.120956, "lon"=>-77.947470),
"/US/US-NY/wesley-hills-14715" => array("type"=>"village", "pop"=>5043, "name_en"=>"Wesley Hills", "lang"=>"en", "lat"=>41.157633, "lon"=>-74.079260),
"/US/US-NY/west-babylon-14716" => array("type"=>"cdp", "pop"=>42410, "name_en"=>"West Babylon", "lang"=>"en", "lat"=>40.710150, "lon"=>-73.358564),
"/US/US-NY/west-bay-shore-14717" => array("type"=>"cdp", "pop"=>4907, "name_en"=>"West Bay Shore", "lang"=>"en", "lat"=>40.707516, "lon"=>-73.271350),
"/US/US-NY/west-carthage-14719" => array("type"=>"village", "pop"=>2167, "name_en"=>"West Carthage", "lang"=>"en", "lat"=>43.973611, "lon"=>-75.621718),
"/US/US-NY/west-elmira-14720" => array("type"=>"cdp", "pop"=>5218, "name_en"=>"West Elmira", "lang"=>"en", "lat"=>42.089500, "lon"=>-76.850716),
"/US/US-NY/west-end-14721" => array("type"=>"cdp", "pop"=>1825, "name_en"=>"West End", "lang"=>"en", "lat"=>42.468569, "lon"=>-75.094146),
"/US/US-NY/west-glens-falls-14723" => array("type"=>"cdp", "pop"=>5964, "name_en"=>"West Glens Falls", "lang"=>"en", "lat"=>43.302567, "lon"=>-73.687960),
"/US/US-NY/west-haverstraw-14726" => array("type"=>"village", "pop"=>10252, "name_en"=>"West Haverstraw", "lang"=>"en", "lat"=>41.206400, "lon"=>-73.987507),
"/US/US-NY/west-hempstead-14727" => array("type"=>"cdp", "pop"=>17689, "name_en"=>"West Hempstead", "lang"=>"en", "lat"=>40.695333, "lon"=>-73.651073),
"/US/US-NY/west-hills-14728" => array("type"=>"cdp", "pop"=>5849, "name_en"=>"West Hills", "lang"=>"en", "lat"=>40.819250, "lon"=>-73.431735),
"/US/US-NY/west-hurley-14729" => array("type"=>"cdp", "pop"=>2252, "name_en"=>"West Hurley", "lang"=>"en", "lat"=>42.006531, "lon"=>-74.111294),
"/US/US-NY/west-islip-14730" => array("type"=>"cdp", "pop"=>28419, "name_en"=>"West Islip", "lang"=>"en", "lat"=>40.708880, "lon"=>-73.296649),
"/US/US-NY/west-nyack-14732" => array("type"=>"cdp", "pop"=>3437, "name_en"=>"West Nyack", "lang"=>"en", "lat"=>41.091000, "lon"=>-73.969203),
"/US/US-NY/west-point-14734" => array("type"=>"cdp", "pop"=>8024, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>41.362343, "lon"=>-74.027317),
"/US/US-NY/west-sand-lake-14736" => array("type"=>"cdp", "pop"=>2251, "name_en"=>"West Sand Lake", "lang"=>"en", "lat"=>42.638700, "lon"=>-73.606583),
"/US/US-NY/west-sayville-14737" => array("type"=>"cdp", "pop"=>4680, "name_en"=>"West Sayville", "lang"=>"en", "lat"=>40.728216, "lon"=>-73.105707),
"/US/US-NY/west-seneca-14738" => array("type"=>"cdp", "pop"=>43795, "name_en"=>"West Seneca", "lang"=>"en", "lat"=>42.837750, "lon"=>-78.751390),
"/US/US-NY/west-winfield-14740" => array("type"=>"village", "pop"=>817, "name_en"=>"West Winfield", "lang"=>"en", "lat"=>42.883638, "lon"=>-75.191747),
"/US/US-NY/westbury-14718" => array("type"=>"village", "pop"=>15065, "name_en"=>"Westbury", "lang"=>"en", "lat"=>40.759650, "lon"=>-73.588905),
"/US/US-NY/westfield-14722" => array("type"=>"village", "pop"=>3361, "name_ru"=>"Уэстфилд", "name_en"=>"Westfield", "lang"=>"en", "lat"=>42.322020, "lon"=>-79.575874),
"/US/US-NY/westhampton-14724" => array("type"=>"cdp", "pop"=>2129, "name_en"=>"Westhampton", "lang"=>"en", "lat"=>40.838704, "lon"=>-72.664914),
"/US/US-NY/westhampton-beach-14725" => array("type"=>"village", "pop"=>2011, "name_en"=>"Westhampton Beach", "lang"=>"en", "lat"=>40.808896, "lon"=>-72.646469),
"/US/US-NY/westmere-14731" => array("type"=>"cdp", "pop"=>6750, "name_en"=>"Westmere", "lang"=>"en", "lat"=>42.687550, "lon"=>-73.874805),
"/US/US-NY/weston-mills-14733" => array("type"=>"cdp", "pop"=>1750, "name_en"=>"Weston Mills", "lang"=>"en", "lat"=>42.075859, "lon"=>-78.372879),
"/US/US-NY/westport-14735" => array("type"=>"village", "pop"=>1276, "name_ru"=>"Уэстпорт", "name_en"=>"Westport", "lang"=>"en", "lat"=>44.178786, "lon"=>-73.436051),
"/US/US-NY/westvale-14739" => array("type"=>"cdp", "pop"=>5952, "name_en"=>"Westvale", "lang"=>"en", "lat"=>43.041850, "lon"=>-76.212305),
"/US/US-NY/wheatley-heights-14741" => array("type"=>"cdp", "pop"=>5027, "name_en"=>"Wheatley Heights", "lang"=>"en", "lat"=>40.763677, "lon"=>-73.370353),
"/US/US-NY/white-plains-14743" => array("type"=>"city", "pop"=>57342, "name_ru"=>"Уайт Плейнс", "name_en"=>"White Plains", "lang"=>"en", "lat"=>41.020850, "lon"=>-73.757422),
"/US/US-NY/whitehall-14742" => array("type"=>"village", "pop"=>2567, "name_ru"=>"Уайтхолл", "name_en"=>"Whitehall", "lang"=>"en", "lat"=>43.560167, "lon"=>-73.418420),
"/US/US-NY/whitesboro-14744" => array("type"=>"village", "pop"=>3761, "name_en"=>"Whitesboro", "lang"=>"en", "lat"=>43.123850, "lon"=>-75.297047),
"/US/US-NY/whitney-point-14745" => array("type"=>"village", "pop"=>940, "name_en"=>"Whitney Point", "lang"=>"en", "lat"=>42.331442, "lon"=>-75.967987),
"/US/US-NY/williamsville-14746" => array("type"=>"village", "pop"=>5129, "name_ru"=>"Уильямсвилл", "name_en"=>"Williamsville", "lang"=>"en", "lat"=>42.962950, "lon"=>-78.740889),
"/US/US-NY/williston-park-14747" => array("type"=>"village", "pop"=>7162, "name_en"=>"Williston Park", "lang"=>"en", "lat"=>40.758850, "lon"=>-73.646863),
"/US/US-NY/wilson-14748" => array("type"=>"village", "pop"=>1141, "name_ru"=>"Уилсон", "name_en"=>"Wilson", "lang"=>"en", "lat"=>43.310650, "lon"=>-78.826949),
"/US/US-NY/windsor-14749" => array("type"=>"village", "pop"=>866, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>42.077160, "lon"=>-75.640996),
"/US/US-NY/wolcott-14750" => array("type"=>"village", "pop"=>1597, "name_ru"=>"Уолкотт", "name_en"=>"Wolcott", "lang"=>"en", "lat"=>43.222543, "lon"=>-76.812745),
"/US/US-NY/woodbury-14751" => array("type"=>"cdp", "pop"=>8904, "name_ru"=>"Вудбери", "name_en"=>"Woodbury", "lang"=>"en", "lat"=>40.814900, "lon"=>-73.470342),
"/US/US-NY/woodmere-14752" => array("type"=>"cdp", "pop"=>15578, "name_en"=>"Woodmere", "lang"=>"en", "lat"=>40.637900, "lon"=>-73.721820),
"/US/US-NY/woodridge-14753" => array("type"=>"village", "pop"=>1065, "name_ru"=>"Вудридж", "name_en"=>"Woodridge", "lang"=>"en", "lat"=>41.711218, "lon"=>-74.573375),
"/US/US-NY/woodsburgh-14754" => array("type"=>"village", "pop"=>829, "name_en"=>"Woodsburgh", "lang"=>"en", "lat"=>40.620900, "lon"=>-73.706657),
"/US/US-NY/woodstock-14755" => array("type"=>"cdp", "pop"=>6149, "name_ru"=>"Вудсток", "name_en"=>"Woodstock", "lang"=>"en", "lat"=>42.038828, "lon"=>-74.110538),
"/US/US-NY/wurtsboro-14756" => array("type"=>"village", "pop"=>1306, "name_en"=>"Wurtsboro", "lang"=>"en", "lat"=>41.576383, "lon"=>-74.485425),
"/US/US-NY/wyandanch-14757" => array("type"=>"cdp", "pop"=>8950, "name_en"=>"Wyandanch", "lang"=>"en", "lat"=>40.746600, "lon"=>-73.377195),
"/US/US-NY/wyantskill-14758" => array("type"=>"cdp", "pop"=>3329, "name_en"=>"Wyantskill", "lang"=>"en", "lat"=>42.689143, "lon"=>-73.649279),
"/US/US-NY/wyoming-14759" => array("type"=>"village", "pop"=>476, "name_ru"=>"Вайоминг", "name_en"=>"Wyoming", "lang"=>"en", "lat"=>42.825144, "lon"=>-78.085495),
"/US/US-NY/yaphank-14760" => array("type"=>"cdp", "pop"=>4637, "name_en"=>"Yaphank", "lang"=>"en", "lat"=>40.831584, "lon"=>-72.925776),
"/US/US-NY/yonkers-14761" => array("type"=>"city", "pop"=>201588, "name_en"=>"Yonkers", "lang"=>"en", "lat"=>40.947033, "lon"=>-73.867514),
"/US/US-NY/yorkshire-14762" => array("type"=>"cdp", "pop"=>3942, "name_en"=>"Yorkshire", "lang"=>"en", "lat"=>42.521897, "lon"=>-78.478439),
"/US/US-NY/yorktown-heights-14763" => array("type"=>"cdp", "pop"=>7690, "name_en"=>"Yorktown Heights", "lang"=>"en", "lat"=>41.271264, "lon"=>-73.782582),
"/US/US-NY/yorkville-14764" => array("type"=>"village", "pop"=>2560, "name_ru"=>"Йорквилл", "name_en"=>"Yorkville", "lang"=>"en", "lat"=>43.111700, "lon"=>-75.273915),
"/US/US-NY/youngstown-14765" => array("type"=>"village", "pop"=>1865, "name_en"=>"Youngstown", "lang"=>"en", "lat"=>43.248700, "lon"=>-79.042753),
"/US/US-NY/zena-14766" => array("type"=>"cdp", "pop"=>1177, "name_en"=>"Zena", "lang"=>"en", "lat"=>42.022219, "lon"=>-74.086323),
"/US/US-OH" => array("type"=>"state", "pop"=>11485910, "name_ru"=>"Огайо", "name_en"=>"Ohio", "lang"=>"en"),
"/US/US-OH/aberdeen-15754" => array("type"=>"village", "pop"=>1509, "name_ru"=>"Абердин", "name_en"=>"Aberdeen", "lang"=>"en", "lat"=>38.666688, "lon"=>-83.769939),
"/US/US-OH/ada-15755" => array("type"=>"village", "pop"=>6052, "name_ru"=>"Ада", "name_en"=>"Ada", "lang"=>"en", "lat"=>40.768841, "lon"=>-83.823865),
"/US/US-OH/adamsville-15756" => array("type"=>"village", "pop"=>131, "name_ru"=>"Адамсвилл", "name_en"=>"Adamsville", "lang"=>"en", "lat"=>40.068775, "lon"=>-81.883080),
"/US/US-OH/addyston-15757" => array("type"=>"village", "pop"=>993, "name_en"=>"Addyston", "lang"=>"en", "lat"=>39.138124, "lon"=>-84.710619),
"/US/US-OH/adelphi-15758" => array("type"=>"village", "pop"=>383, "name_ru"=>"Аделфи", "name_en"=>"Adelphi", "lang"=>"en", "lat"=>39.464627, "lon"=>-82.746311),
"/US/US-OH/adena-15759" => array("type"=>"village", "pop"=>748, "name_en"=>"Adena", "lang"=>"en", "lat"=>40.216354, "lon"=>-80.876339),
"/US/US-OH/akron-15760" => array("type"=>"city", "pop"=>207510, "name_ru"=>"Акрон", "name_en"=>"Akron", "lang"=>"en", "lat"=>41.080400, "lon"=>-81.521499),
"/US/US-OH/albany-15761" => array("type"=>"village", "pop"=>832, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>39.223660, "lon"=>-82.198882),
"/US/US-OH/alexandria-15762" => array("type"=>"village", "pop"=>252, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>40.089749, "lon"=>-82.613056),
"/US/US-OH/alger-15763" => array("type"=>"village", "pop"=>870, "name_en"=>"Alger", "lang"=>"en", "lat"=>40.709506, "lon"=>-83.844575),
"/US/US-OH/alliance-15764" => array("type"=>"city", "pop"=>22476, "name_en"=>"Alliance", "lang"=>"en", "lat"=>40.911100, "lon"=>-81.117149),
"/US/US-OH/alvordton-15765" => array("type"=>"village", "pop"=>298, "name_en"=>"Alvordton", "lang"=>"en", "lat"=>41.665325, "lon"=>-84.434352),
"/US/US-OH/amanda-15766" => array("type"=>"village", "pop"=>710, "name_en"=>"Amanda", "lang"=>"en", "lat"=>39.650381, "lon"=>-82.743264),
"/US/US-OH/amberley-15767" => array("type"=>"village", "pop"=>3558, "name_en"=>"Amberley", "lang"=>"en", "lat"=>39.202368, "lon"=>-84.433224),
"/US/US-OH/amelia-15768" => array("type"=>"village", "pop"=>3560, "name_ru"=>"Амелия", "name_en"=>"Amelia", "lang"=>"en", "lat"=>39.025777, "lon"=>-84.216891),
"/US/US-OH/amesville-15769" => array("type"=>"village", "pop"=>188, "name_en"=>"Amesville", "lang"=>"en", "lat"=>39.401425, "lon"=>-81.955272),
"/US/US-OH/amherst-15770" => array("type"=>"city", "pop"=>11724, "name_ru"=>"Амхерст", "name_en"=>"Amherst", "lang"=>"en", "lat"=>41.400050, "lon"=>-82.230731),
"/US/US-OH/amsterdam-15771" => array("type"=>"village", "pop"=>537, "name_ru"=>"Амсердам", "name_en"=>"Amsterdam", "lang"=>"en", "lat"=>40.471350, "lon"=>-80.921352),
"/US/US-OH/andover-15772" => array("type"=>"village", "pop"=>1215, "name_ru"=>"Андоувер", "name_en"=>"Andover", "lang"=>"en", "lat"=>41.607246, "lon"=>-80.569565),
"/US/US-OH/anna-15773" => array("type"=>"village", "pop"=>1562, "name_ru"=>"Анна", "name_en"=>"Anna", "lang"=>"en", "lat"=>40.395634, "lon"=>-84.172914),
"/US/US-OH/ansonia-15774" => array("type"=>"village", "pop"=>1078, "name_ru"=>"Ансония", "name_en"=>"Ansonia", "lang"=>"en", "lat"=>40.214643, "lon"=>-84.635324),
"/US/US-OH/antioch-15775" => array("type"=>"village", "pop"=>85, "name_ru"=>"Антиок", "name_en"=>"Antioch", "lang"=>"en", "lat"=>39.661587, "lon"=>-81.067690),
"/US/US-OH/antwerp-15776" => array("type"=>"village", "pop"=>1636, "name_en"=>"Antwerp", "lang"=>"en", "lat"=>41.180744, "lon"=>-84.739109),
"/US/US-OH/apple-creek-15777" => array("type"=>"village", "pop"=>972, "name_en"=>"Apple Creek", "lang"=>"en", "lat"=>40.748322, "lon"=>-81.833506),
"/US/US-OH/aquilla-15778" => array("type"=>"village", "pop"=>365, "name_en"=>"Aquilla", "lang"=>"en", "lat"=>41.545958, "lon"=>-81.171243),
"/US/US-OH/arcadia-15779" => array("type"=>"village", "pop"=>564, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>41.111256, "lon"=>-83.511259),
"/US/US-OH/arcanum-15780" => array("type"=>"village", "pop"=>1965, "name_en"=>"Arcanum", "lang"=>"en", "lat"=>39.991552, "lon"=>-84.553823),
"/US/US-OH/archbold-15781" => array("type"=>"village", "pop"=>4438, "name_en"=>"Archbold", "lang"=>"en", "lat"=>41.515483, "lon"=>-84.304357),
"/US/US-OH/arlington-15782" => array("type"=>"village", "pop"=>1197, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>40.892692, "lon"=>-83.653678),
"/US/US-OH/arlington-heights-15783" => array("type"=>"village", "pop"=>836, "name_ru"=>"Арлингтон Хайтс", "name_en"=>"Arlington Heights", "lang"=>"en", "lat"=>39.215101, "lon"=>-84.455825),
"/US/US-OH/ashland-15784" => array("type"=>"city", "pop"=>21850, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>40.866945, "lon"=>-82.317606),
"/US/US-OH/ashley-15785" => array("type"=>"village", "pop"=>1295, "name_ru"=>"Эшли", "name_en"=>"Ashley", "lang"=>"en", "lat"=>40.409121, "lon"=>-82.954474),
"/US/US-OH/ashtabula-15786" => array("type"=>"city", "pop"=>19689, "name_en"=>"Ashtabula", "lang"=>"en", "lat"=>41.879487, "lon"=>-80.798217),
"/US/US-OH/ashville-15787" => array("type"=>"village", "pop"=>3289, "name_ru"=>"Ашвилл", "name_en"=>"Ashville", "lang"=>"en", "lat"=>39.717648, "lon"=>-82.955238),
"/US/US-OH/athalia-15788" => array("type"=>"village", "pop"=>336, "name_en"=>"Athalia", "lang"=>"en", "lat"=>38.512184, "lon"=>-82.307567),
"/US/US-OH/athens-15789" => array("type"=>"city", "pop"=>22088, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>39.324177, "lon"=>-82.096051),
"/US/US-OH/attica-15790" => array("type"=>"village", "pop"=>906, "name_ru"=>"Аттика", "name_en"=>"Attica", "lang"=>"en", "lat"=>41.065171, "lon"=>-82.886319),
"/US/US-OH/aurora-15791" => array("type"=>"city", "pop"=>14548, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>41.312167, "lon"=>-81.344577),
"/US/US-OH/austintown-15792" => array("type"=>"cdp", "pop"=>35119, "name_en"=>"Austintown", "lang"=>"en", "lat"=>41.091600, "lon"=>-80.738299),
"/US/US-OH/avon-15793" => array("type"=>"city", "pop"=>17086, "name_ru"=>"Эйвон", "name_en"=>"Avon", "lang"=>"en", "lat"=>41.451250, "lon"=>-82.023470),
"/US/US-OH/avon-lake-15794" => array("type"=>"city", "pop"=>23996, "name_en"=>"Avon Lake", "lang"=>"en", "lat"=>41.493743, "lon"=>-82.015895),
"/US/US-OH/bailey-lakes-15795" => array("type"=>"village", "pop"=>399, "name_en"=>"Bailey Lakes", "lang"=>"en", "lat"=>40.947434, "lon"=>-82.355481),
"/US/US-OH/bainbridge-15796" => array("type"=>"cdp", "pop"=>1052, "name_ru"=>"Бэйнбридж", "name_en"=>"Bainbridge", "lang"=>"en", "lat"=>41.393381, "lon"=>-81.337162),
"/US/US-OH/bainbridge-15797" => array("type"=>"village", "pop"=>968, "name_ru"=>"Бэйнбридж", "name_en"=>"Bainbridge", "lang"=>"en", "lat"=>39.226517, "lon"=>-83.269553),
"/US/US-OH/bairdstown-15798" => array("type"=>"village", "pop"=>125, "name_en"=>"Bairdstown", "lang"=>"en", "lat"=>41.171036, "lon"=>-83.607005),
"/US/US-OH/ballville-15799" => array("type"=>"cdp", "pop"=>6362, "name_en"=>"Ballville", "lang"=>"en", "lat"=>41.327671, "lon"=>-83.137318),
"/US/US-OH/baltic-15800" => array("type"=>"village", "pop"=>736, "name_en"=>"Baltic", "lang"=>"en", "lat"=>40.441331, "lon"=>-81.701841),
"/US/US-OH/baltimore-15801" => array("type"=>"village", "pop"=>2905, "name_ru"=>"Балтимор", "name_en"=>"Baltimore", "lang"=>"en", "lat"=>39.846256, "lon"=>-82.607618),
"/US/US-OH/barberton-15802" => array("type"=>"city", "pop"=>26581, "name_en"=>"Barberton", "lang"=>"en", "lat"=>41.011950, "lon"=>-81.605694),
"/US/US-OH/barnesville-15803" => array("type"=>"village", "pop"=>4042, "name_ru"=>"Барнсвилл", "name_en"=>"Barnesville", "lang"=>"en", "lat"=>39.989037, "lon"=>-81.173496),
"/US/US-OH/barnhill-15804" => array("type"=>"village", "pop"=>370, "name_en"=>"Barnhill", "lang"=>"en", "lat"=>40.448190, "lon"=>-81.367631),
"/US/US-OH/batavia-15805" => array("type"=>"village", "pop"=>1749, "name_ru"=>"Батавия", "name_en"=>"Batavia", "lang"=>"en", "lat"=>39.078434, "lon"=>-84.181245),
"/US/US-OH/batesville-15806" => array("type"=>"village", "pop"=>92, "name_ru"=>"Батсвилл", "name_en"=>"Batesville", "lang"=>"en", "lat"=>39.914231, "lon"=>-81.281211),
"/US/US-OH/bay-view-15807" => array("type"=>"village", "pop"=>639, "name_en"=>"Bay View", "lang"=>"en", "lat"=>41.468853, "lon"=>-82.823633),
"/US/US-OH/bay-village-15808" => array("type"=>"city", "pop"=>14606, "name_en"=>"Bay Village", "lang"=>"en", "lat"=>41.489750, "lon"=>-81.929299),
"/US/US-OH/beach-city-15809" => array("type"=>"village", "pop"=>1096, "name_en"=>"Beach City", "lang"=>"en", "lat"=>40.652663, "lon"=>-81.580120),
"/US/US-OH/beachwood-15810" => array("type"=>"city", "pop"=>11108, "name_en"=>"Beachwood", "lang"=>"en", "lat"=>41.478650, "lon"=>-81.502499),
"/US/US-OH/beallsville-15811" => array("type"=>"village", "pop"=>408, "name_en"=>"Beallsville", "lang"=>"en", "lat"=>39.848330, "lon"=>-81.035351),
"/US/US-OH/beaver-15812" => array("type"=>"village", "pop"=>455, "name_ru"=>"Бивер", "name_en"=>"Beaver", "lang"=>"en", "lat"=>39.032359, "lon"=>-82.825232),
"/US/US-OH/beavercreek-15813" => array("type"=>"city", "pop"=>39909, "name_en"=>"Beavercreek", "lang"=>"en", "lat"=>39.729300, "lon"=>-84.062375),
"/US/US-OH/beaverdam-15814" => array("type"=>"village", "pop"=>341, "name_en"=>"Beaverdam", "lang"=>"en", "lat"=>40.832915, "lon"=>-83.976370),
"/US/US-OH/beckett-ridge-15815" => array("type"=>"cdp", "pop"=>4505, "name_en"=>"Beckett Ridge", "lang"=>"en", "lat"=>39.347000, "lon"=>-84.435403),
"/US/US-OH/bedford-15816" => array("type"=>"city", "pop"=>12965, "name_ru"=>"Бедфорд", "name_en"=>"Bedford", "lang"=>"en", "lat"=>41.391271, "lon"=>-81.537270),
"/US/US-OH/bedford-heights-15817" => array("type"=>"city", "pop"=>10485, "name_en"=>"Bedford Heights", "lang"=>"en", "lat"=>41.405201, "lon"=>-81.506174),
"/US/US-OH/beechwood-trails-15818" => array("type"=>"cdp", "pop"=>1875, "name_en"=>"Beechwood Trails", "lang"=>"en", "lat"=>40.023662, "lon"=>-82.650935),
"/US/US-OH/bellaire-15819" => array("type"=>"city", "pop"=>4557, "name_ru"=>"Беллэр", "name_en"=>"Bellaire", "lang"=>"en", "lat"=>40.016933, "lon"=>-80.746475),
"/US/US-OH/bellbrook-15820" => array("type"=>"city", "pop"=>7000, "name_en"=>"Bellbrook", "lang"=>"en", "lat"=>39.640000, "lon"=>-84.085569),
"/US/US-OH/belle-center-15821" => array("type"=>"village", "pop"=>773, "name_en"=>"Belle Center", "lang"=>"en", "lat"=>40.509104, "lon"=>-83.744871),
"/US/US-OH/belle-valley-15823" => array("type"=>"village", "pop"=>254, "name_en"=>"Belle Valley", "lang"=>"en", "lat"=>39.788677, "lon"=>-81.555205),
"/US/US-OH/bellefontaine-15822" => array("type"=>"city", "pop"=>12665, "name_en"=>"Bellefontaine", "lang"=>"en", "lat"=>40.359974, "lon"=>-83.756994),
"/US/US-OH/bellevue-15824" => array("type"=>"city", "pop"=>7831, "name_ru"=>"Белвью", "name_en"=>"Bellevue", "lang"=>"en", "lat"=>41.275755, "lon"=>-82.842190),
"/US/US-OH/bellville-15825" => array("type"=>"village", "pop"=>1702, "name_ru"=>"Беллвилл", "name_en"=>"Bellville", "lang"=>"en", "lat"=>40.620812, "lon"=>-82.511028),
"/US/US-OH/belmont-15826" => array("type"=>"village", "pop"=>497, "name_ru"=>"Белмонт", "name_en"=>"Belmont", "lang"=>"en", "lat"=>40.027951, "lon"=>-81.041183),
"/US/US-OH/belmore-15827" => array("type"=>"village", "pop"=>165, "name_en"=>"Belmore", "lang"=>"en", "lat"=>41.154402, "lon"=>-83.941361),
"/US/US-OH/beloit-15828" => array("type"=>"village", "pop"=>968, "name_ru"=>"Белойт", "name_en"=>"Beloit", "lang"=>"en", "lat"=>40.920490, "lon"=>-81.000454),
"/US/US-OH/belpre-15829" => array("type"=>"city", "pop"=>6528, "name_ru"=>"Белпре", "name_en"=>"Belpre", "lang"=>"en", "lat"=>39.278253, "lon"=>-81.588620),
"/US/US-OH/bentleyville-15830" => array("type"=>"village", "pop"=>896, "name_en"=>"Bentleyville", "lang"=>"en", "lat"=>41.413467, "lon"=>-81.413161),
"/US/US-OH/benton-ridge-15831" => array("type"=>"village", "pop"=>290, "name_en"=>"Benton Ridge", "lang"=>"en", "lat"=>41.004826, "lon"=>-83.791962),
"/US/US-OH/berea-15832" => array("type"=>"city", "pop"=>17987, "name_ru"=>"Берея", "name_en"=>"Berea", "lang"=>"en", "lat"=>41.369900, "lon"=>-81.862733),
"/US/US-OH/bergholz-15833" => array("type"=>"village", "pop"=>727, "name_en"=>"Bergholz", "lang"=>"en", "lat"=>40.521086, "lon"=>-80.884741),
"/US/US-OH/berkey-15834" => array("type"=>"village", "pop"=>310, "name_en"=>"Berkey", "lang"=>"en", "lat"=>41.714350, "lon"=>-83.833452),
"/US/US-OH/berlin-heights-15835" => array("type"=>"village", "pop"=>635, "name_en"=>"Berlin Heights", "lang"=>"en", "lat"=>41.320900, "lon"=>-82.492690),
"/US/US-OH/bethel-15836" => array("type"=>"village", "pop"=>2644, "name_ru"=>"Бетел", "name_en"=>"Bethel", "lang"=>"en", "lat"=>38.962448, "lon"=>-84.084688),
"/US/US-OH/bethesda-15837" => array("type"=>"village", "pop"=>1357, "name_ru"=>"Бетесда", "name_en"=>"Bethesda", "lang"=>"en", "lat"=>40.016130, "lon"=>-81.072930),
"/US/US-OH/bettsville-15838" => array("type"=>"village", "pop"=>737, "name_en"=>"Bettsville", "lang"=>"en", "lat"=>41.244042, "lon"=>-83.233566),
"/US/US-OH/beverly-15839" => array("type"=>"village", "pop"=>1326, "name_ru"=>"Беверли", "name_en"=>"Beverly", "lang"=>"en", "lat"=>39.549667, "lon"=>-81.636291),
"/US/US-OH/bexley-15840" => array("type"=>"city", "pop"=>12507, "name_en"=>"Bexley", "lang"=>"en", "lat"=>39.964550, "lon"=>-82.934549),
"/US/US-OH/blacklick-estates-15841" => array("type"=>"cdp", "pop"=>10080, "name_en"=>"Blacklick Estates", "lang"=>"en", "lat"=>39.904933, "lon"=>-82.864518),
"/US/US-OH/blakeslee-15842" => array("type"=>"village", "pop"=>122, "name_en"=>"Blakeslee", "lang"=>"en", "lat"=>41.523875, "lon"=>-84.731070),
"/US/US-OH/blanchester-15843" => array("type"=>"village", "pop"=>4303, "name_en"=>"Blanchester", "lang"=>"en", "lat"=>39.289534, "lon"=>-83.979567),
"/US/US-OH/bloomdale-15844" => array("type"=>"village", "pop"=>709, "name_en"=>"Bloomdale", "lang"=>"en", "lat"=>41.170938, "lon"=>-83.553586),
"/US/US-OH/bloomingburg-15845" => array("type"=>"village", "pop"=>834, "name_en"=>"Bloomingburg", "lang"=>"en", "lat"=>39.607884, "lon"=>-83.395479),
"/US/US-OH/bloomingdale-15846" => array("type"=>"village", "pop"=>206, "name_ru"=>"Блумингдэйл", "name_en"=>"Bloomingdale", "lang"=>"en", "lat"=>40.341832, "lon"=>-80.817467),
"/US/US-OH/bloomville-15847" => array("type"=>"village", "pop"=>983, "name_en"=>"Bloomville", "lang"=>"en", "lat"=>41.051025, "lon"=>-83.013278),
"/US/US-OH/blue-ash-15848" => array("type"=>"city", "pop"=>12783, "name_en"=>"Blue Ash", "lang"=>"en", "lat"=>39.246718, "lon"=>-84.380773),
"/US/US-OH/bluffton-15849" => array("type"=>"village", "pop"=>3944, "name_ru"=>"Блаффтон", "name_en"=>"Bluffton", "lang"=>"en", "lat"=>40.890119, "lon"=>-83.885937),
"/US/US-OH/boardman-15850" => array("type"=>"cdp", "pop"=>39204, "name_en"=>"Boardman", "lang"=>"en", "lat"=>41.032150, "lon"=>-80.668496),
"/US/US-OH/bolindale-15851" => array("type"=>"cdp", "pop"=>2827, "name_en"=>"Bolindale", "lang"=>"en", "lat"=>41.207300, "lon"=>-80.777752),
"/US/US-OH/bolivar-15852" => array("type"=>"village", "pop"=>885, "name_ru"=>"Боливар", "name_en"=>"Bolivar", "lang"=>"en", "lat"=>40.649955, "lon"=>-81.454769),
"/US/US-OH/boston-heights-15853" => array("type"=>"village", "pop"=>1220, "name_en"=>"Boston Heights", "lang"=>"en", "lat"=>41.255659, "lon"=>-81.504562),
"/US/US-OH/botkins-15854" => array("type"=>"village", "pop"=>1174, "name_en"=>"Botkins", "lang"=>"en", "lat"=>40.467233, "lon"=>-84.182618),
"/US/US-OH/bowerston-15855" => array("type"=>"village", "pop"=>411, "name_en"=>"Bowerston", "lang"=>"en", "lat"=>40.426741, "lon"=>-81.187697),
"/US/US-OH/bowersville-15856" => array("type"=>"village", "pop"=>322, "name_ru"=>"Бауэрсвилл", "name_en"=>"Bowersville", "lang"=>"en", "lat"=>39.580137, "lon"=>-83.723712),
"/US/US-OH/bowling-green-15857" => array("type"=>"city", "pop"=>29542, "name_ru"=>"Боулинг Грин", "name_en"=>"Bowling Green", "lang"=>"en", "lat"=>41.376450, "lon"=>-83.649462),
"/US/US-OH/bradford-15858" => array("type"=>"village", "pop"=>1845, "name_ru"=>"Брадфорд", "name_en"=>"Bradford", "lang"=>"en", "lat"=>40.130904, "lon"=>-84.429307),
"/US/US-OH/bradner-15859" => array("type"=>"village", "pop"=>1158, "name_en"=>"Bradner", "lang"=>"en", "lat"=>41.323643, "lon"=>-83.436544),
"/US/US-OH/brady-lake-15860" => array("type"=>"village", "pop"=>497, "name_en"=>"Brady Lake", "lang"=>"en", "lat"=>41.162748, "lon"=>-81.313901),
"/US/US-OH/bratenahl-15861" => array("type"=>"village", "pop"=>1263, "name_en"=>"Bratenahl", "lang"=>"en", "lat"=>41.557762, "lon"=>-81.596999),
"/US/US-OH/brecksville-15862" => array("type"=>"city", "pop"=>12851, "name_en"=>"Brecksville", "lang"=>"en", "lat"=>41.307634, "lon"=>-81.618509),
"/US/US-OH/bremen-15863" => array("type"=>"village", "pop"=>1250, "name_ru"=>"Бремен", "name_en"=>"Bremen", "lang"=>"en", "lat"=>39.705937, "lon"=>-82.431369),
"/US/US-OH/brentwood-15864" => array("type"=>"cdp", "pop"=>3568, "name_ru"=>"Брентвуд", "name_en"=>"Brentwood", "lang"=>"en", "lat"=>40.353050, "lon"=>-80.730680),
"/US/US-OH/brewster-15865" => array("type"=>"village", "pop"=>2313, "name_ru"=>"Брюстер", "name_en"=>"Brewster", "lang"=>"en", "lat"=>40.712661, "lon"=>-81.600699),
"/US/US-OH/briarwood-beach-15866" => array("type"=>"village", "pop"=>682, "name_en"=>"Briarwood Beach", "lang"=>"en", "lat"=>41.074865, "lon"=>-81.904927),
"/US/US-OH/brice-15867" => array("type"=>"village", "pop"=>82, "name_en"=>"Brice", "lang"=>"en", "lat"=>39.916600, "lon"=>-82.832024),
"/US/US-OH/bridgeport-15868" => array("type"=>"village", "pop"=>2045, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>40.066484, "lon"=>-80.746841),
"/US/US-OH/bridgetown-north-15869" => array("type"=>"cdp", "pop"=>11748, "name_en"=>"Bridgetown North", "lang"=>"en", "lat"=>39.158101, "lon"=>-84.635560),
"/US/US-OH/brilliant-15870" => array("type"=>"village", "pop"=>1672, "name_ru"=>"Бриллиант", "name_en"=>"Brilliant", "lang"=>"en", "lat"=>40.268500, "lon"=>-80.624900),
"/US/US-OH/brimfield-15871" => array("type"=>"cdp", "pop"=>7910, "name_ru"=>"Бримфилд", "name_en"=>"Brimfield", "lang"=>"en", "lat"=>41.090250, "lon"=>-81.350149),
"/US/US-OH/broadview-heights-15872" => array("type"=>"city", "pop"=>17274, "name_en"=>"Broadview Heights", "lang"=>"en", "lat"=>41.319250, "lon"=>-81.679882),
"/US/US-OH/brook-park-15876" => array("type"=>"city", "pop"=>19194, "name_ru"=>"Брук Парк", "name_en"=>"Brook Park", "lang"=>"en", "lat"=>41.400067, "lon"=>-81.826923),
"/US/US-OH/brookfield-center-15873" => array("type"=>"cdp", "pop"=>1396, "name_en"=>"Brookfield Center", "lang"=>"en", "lat"=>41.240504, "lon"=>-80.558190),
"/US/US-OH/brooklyn-15874" => array("type"=>"city", "pop"=>10410, "name_ru"=>"Бруклин", "name_en"=>"Brooklyn", "lang"=>"en", "lat"=>41.434600, "lon"=>-81.749549),
"/US/US-OH/brooklyn-heights-15875" => array("type"=>"village", "pop"=>1454, "name_ru"=>"Бруклин Хайтс", "name_en"=>"Brooklyn Heights", "lang"=>"en", "lat"=>41.415601, "lon"=>-81.665391),
"/US/US-OH/brookside-15877" => array("type"=>"village", "pop"=>604, "name_ru"=>"Бруксайд", "name_en"=>"Brookside", "lang"=>"en", "lat"=>40.071150, "lon"=>-80.761790),
"/US/US-OH/brookville-15878" => array("type"=>"village", "pop"=>5423, "name_ru"=>"Бруквилл", "name_en"=>"Brookville", "lang"=>"en", "lat"=>39.835667, "lon"=>-84.415002),
"/US/US-OH/broughton-15879" => array("type"=>"village", "pop"=>153, "name_ru"=>"Бротон", "name_en"=>"Broughton", "lang"=>"en", "lat"=>41.087923, "lon"=>-84.535946),
"/US/US-OH/brunswick-15880" => array("type"=>"city", "pop"=>34711, "name_ru"=>"Брунсвик", "name_en"=>"Brunswick", "lang"=>"en", "lat"=>41.245667, "lon"=>-81.821167),
"/US/US-OH/bryan-15881" => array("type"=>"city", "pop"=>8314, "name_en"=>"Bryan", "lang"=>"en", "lat"=>41.470228, "lon"=>-84.548003),
"/US/US-OH/buchtel-15882" => array("type"=>"village", "pop"=>596, "name_en"=>"Buchtel", "lang"=>"en", "lat"=>39.462769, "lon"=>-82.182488),
"/US/US-OH/buckeye-lake-15883" => array("type"=>"village", "pop"=>3043, "name_en"=>"Buckeye Lake", "lang"=>"en", "lat"=>39.936971, "lon"=>-82.479204),
"/US/US-OH/buckland-15884" => array("type"=>"village", "pop"=>244, "name_ru"=>"Бакленд", "name_en"=>"Buckland", "lang"=>"en", "lat"=>40.624263, "lon"=>-84.260693),
"/US/US-OH/bucyrus-15885" => array("type"=>"city", "pop"=>12252, "name_en"=>"Bucyrus", "lang"=>"en", "lat"=>40.802800, "lon"=>-82.973469),
"/US/US-OH/burbank-15886" => array("type"=>"village", "pop"=>276, "name_ru"=>"Бурбанк", "name_en"=>"Burbank", "lang"=>"en", "lat"=>40.987597, "lon"=>-81.994556),
"/US/US-OH/burgoon-15887" => array("type"=>"village", "pop"=>193, "name_en"=>"Burgoon", "lang"=>"en", "lat"=>41.267322, "lon"=>-83.250709),
"/US/US-OH/burkettsville-15888" => array("type"=>"village", "pop"=>246, "name_en"=>"Burkettsville", "lang"=>"en", "lat"=>40.352912, "lon"=>-84.642489),
"/US/US-OH/burlington-15889" => array("type"=>"cdp", "pop"=>1169, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>38.409344, "lon"=>-82.529037),
"/US/US-OH/burton-15890" => array("type"=>"village", "pop"=>1426, "name_ru"=>"Бертон", "name_en"=>"Burton", "lang"=>"en", "lat"=>41.470547, "lon"=>-81.146401),
"/US/US-OH/butler-15891" => array("type"=>"village", "pop"=>880, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>40.586409, "lon"=>-82.420063),
"/US/US-OH/butlerville-15892" => array("type"=>"village", "pop"=>247, "name_en"=>"Butlerville", "lang"=>"en", "lat"=>39.301952, "lon"=>-84.088960),
"/US/US-OH/byesville-15893" => array("type"=>"village", "pop"=>2515, "name_en"=>"Byesville", "lang"=>"en", "lat"=>39.968009, "lon"=>-81.542633),
"/US/US-OH/cadiz-15894" => array("type"=>"village", "pop"=>3284, "name_ru"=>"Кадиз", "name_en"=>"Cadiz", "lang"=>"en", "lat"=>40.266724, "lon"=>-80.992336),
"/US/US-OH/cairo-15895" => array("type"=>"village", "pop"=>505, "name_ru"=>"Каир", "name_en"=>"Cairo", "lang"=>"en", "lat"=>40.831304, "lon"=>-84.084555),
"/US/US-OH/calcutta-15896" => array("type"=>"cdp", "pop"=>1212, "name_en"=>"Calcutta", "lang"=>"en", "lat"=>40.676036, "lon"=>-80.578345),
"/US/US-OH/caldwell-15897" => array("type"=>"village", "pop"=>2246, "name_ru"=>"Колдуэлл", "name_en"=>"Caldwell", "lang"=>"en", "lat"=>39.745763, "lon"=>-81.513145),
"/US/US-OH/caledonia-15898" => array("type"=>"village", "pop"=>531, "name_ru"=>"Каледония", "name_en"=>"Caledonia", "lang"=>"en", "lat"=>40.636009, "lon"=>-82.969394),
"/US/US-OH/cambridge-15899" => array("type"=>"city", "pop"=>11192, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>40.025047, "lon"=>-81.586931),
"/US/US-OH/camden-15900" => array("type"=>"village", "pop"=>2182, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>39.637128, "lon"=>-84.646455),
"/US/US-OH/campbell-15901" => array("type"=>"city", "pop"=>8436, "name_ru"=>"Кэмпбелл", "name_en"=>"Campbell", "lang"=>"en", "lat"=>41.077433, "lon"=>-80.590565),
"/US/US-OH/canal-fulton-15902" => array("type"=>"village", "pop"=>5023, "name_en"=>"Canal Fulton", "lang"=>"en", "lat"=>40.890005, "lon"=>-81.584535),
"/US/US-OH/canal-winchester-15903" => array("type"=>"village", "pop"=>6248, "name_en"=>"Canal Winchester", "lang"=>"en", "lat"=>39.846852, "lon"=>-82.821357),
"/US/US-OH/canfield-15904" => array("type"=>"city", "pop"=>6862, "name_en"=>"Canfield", "lang"=>"en", "lat"=>41.029712, "lon"=>-80.767387),
"/US/US-OH/canton-15905" => array("type"=>"city", "pop"=>78362, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>40.812200, "lon"=>-81.373265),
"/US/US-OH/cardington-15906" => array("type"=>"village", "pop"=>1992, "name_en"=>"Cardington", "lang"=>"en", "lat"=>40.499057, "lon"=>-82.893790),
"/US/US-OH/carey-15907" => array("type"=>"village", "pop"=>3773, "name_en"=>"Carey", "lang"=>"en", "lat"=>40.951470, "lon"=>-83.381861),
"/US/US-OH/carlisle-15908" => array("type"=>"village", "pop"=>6051, "name_ru"=>"Карлайл", "name_en"=>"Carlisle", "lang"=>"en", "lat"=>39.578335, "lon"=>-84.320930),
"/US/US-OH/carroll-15909" => array("type"=>"village", "pop"=>467, "name_ru"=>"Кэррол", "name_en"=>"Carroll", "lang"=>"en", "lat"=>39.799947, "lon"=>-82.703363),
"/US/US-OH/carrollton-15910" => array("type"=>"village", "pop"=>3208, "name_ru"=>"Карролтон", "name_en"=>"Carrollton", "lang"=>"en", "lat"=>40.579814, "lon"=>-81.091262),
"/US/US-OH/casstown-15911" => array("type"=>"village", "pop"=>320, "name_en"=>"Casstown", "lang"=>"en", "lat"=>40.052616, "lon"=>-84.128778),
"/US/US-OH/castalia-15912" => array("type"=>"village", "pop"=>873, "name_ru"=>"Касталия", "name_en"=>"Castalia", "lang"=>"en", "lat"=>41.401320, "lon"=>-82.809251),
"/US/US-OH/castine-15913" => array("type"=>"village", "pop"=>123, "name_en"=>"Castine", "lang"=>"en", "lat"=>39.930824, "lon"=>-84.624721),
"/US/US-OH/catawba-15914" => array("type"=>"village", "pop"=>313, "name_en"=>"Catawba", "lang"=>"en", "lat"=>39.999840, "lon"=>-83.622261),
"/US/US-OH/cecil-15915" => array("type"=>"village", "pop"=>198, "name_ru"=>"Сесил", "name_en"=>"Cecil", "lang"=>"en", "lat"=>41.219140, "lon"=>-84.601506),
"/US/US-OH/cedarville-15916" => array("type"=>"village", "pop"=>4033, "name_ru"=>"Сидарвилл", "name_en"=>"Cedarville", "lang"=>"en", "lat"=>39.744936, "lon"=>-83.810539),
"/US/US-OH/celina-15917" => array("type"=>"city", "pop"=>10255, "name_en"=>"Celina", "lang"=>"en", "lat"=>40.553345, "lon"=>-84.570101),
"/US/US-OH/centerburg-15918" => array("type"=>"village", "pop"=>1497, "name_en"=>"Centerburg", "lang"=>"en", "lat"=>40.303712, "lon"=>-82.696612),
"/US/US-OH/centerville-15919" => array("type"=>"city", "pop"=>142, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>39.641067, "lon"=>-84.140041),
"/US/US-OH/centerville-15920" => array("type"=>"village", "pop"=>128, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>38.898370, "lon"=>-82.445168),
"/US/US-OH/chagrin-falls-15921" => array("type"=>"village", "pop"=>3641, "name_en"=>"Chagrin Falls", "lang"=>"en", "lat"=>41.431150, "lon"=>-81.389561),
"/US/US-OH/champion-heights-15922" => array("type"=>"cdp", "pop"=>4665, "name_en"=>"Champion Heights", "lang"=>"en", "lat"=>41.288128, "lon"=>-80.848759),
"/US/US-OH/chardon-15923" => array("type"=>"village", "pop"=>5210, "name_en"=>"Chardon", "lang"=>"en", "lat"=>41.579074, "lon"=>-81.208531),
"/US/US-OH/chatfield-15924" => array("type"=>"village", "pop"=>195, "name_ru"=>"Чэтфилд", "name_en"=>"Chatfield", "lang"=>"en", "lat"=>40.952501, "lon"=>-82.941631),
"/US/US-OH/chauncey-15925" => array("type"=>"village", "pop"=>1101, "name_ru"=>"Чаунси", "name_en"=>"Chauncey", "lang"=>"en", "lat"=>39.399536, "lon"=>-82.128045),
"/US/US-OH/cherry-fork-15926" => array("type"=>"village", "pop"=>133, "name_en"=>"Cherry Fork", "lang"=>"en", "lat"=>38.888361, "lon"=>-83.613505),
"/US/US-OH/cherry-grove-15927" => array("type"=>"cdp", "pop"=>4972, "name_en"=>"Cherry Grove", "lang"=>"en", "lat"=>39.079701, "lon"=>-84.322811),
"/US/US-OH/chesapeake-15928" => array("type"=>"village", "pop"=>868, "name_en"=>"Chesapeake", "lang"=>"en", "lat"=>38.430520, "lon"=>-82.450537),
"/US/US-OH/cheshire-15929" => array("type"=>"village", "pop"=>85, "name_en"=>"Cheshire", "lang"=>"en", "lat"=>38.944029, "lon"=>-82.112240),
"/US/US-OH/chesterhill-15930" => array("type"=>"village", "pop"=>295, "name_en"=>"Chesterhill", "lang"=>"en", "lat"=>39.489879, "lon"=>-81.865945),
"/US/US-OH/chesterland-15931" => array("type"=>"cdp", "pop"=>2078, "name_en"=>"Chesterland", "lang"=>"en", "lat"=>41.524187, "lon"=>-81.342961),
"/US/US-OH/chesterville-15932" => array("type"=>"village", "pop"=>203, "name_en"=>"Chesterville", "lang"=>"en", "lat"=>40.480104, "lon"=>-82.682836),
"/US/US-OH/cheviot-15933" => array("type"=>"city", "pop"=>8385, "name_en"=>"Cheviot", "lang"=>"en", "lat"=>39.157701, "lon"=>-84.613825),
"/US/US-OH/chickasaw-15934" => array("type"=>"village", "pop"=>355, "name_ru"=>"Чикасо", "name_en"=>"Chickasaw", "lang"=>"en", "lat"=>40.436235, "lon"=>-84.494005),
"/US/US-OH/chillicothe-15935" => array("type"=>"city", "pop"=>22296, "name_ru"=>"Чилликот", "name_en"=>"Chillicothe", "lang"=>"en", "lat"=>39.335575, "lon"=>-82.988068),
"/US/US-OH/chilo-15936" => array("type"=>"village", "pop"=>104, "name_en"=>"Chilo", "lang"=>"en", "lat"=>38.793749, "lon"=>-84.138152),
"/US/US-OH/chippewa-on-the-lake-15937" => array("type"=>"village", "pop"=>271, "name_en"=>"Chippewa-on-the-Lake", "lang"=>"en", "lat"=>41.070078, "lon"=>-81.903688),
"/US/US-OH/chocktou-lake-15938" => array("type"=>"cdp", "pop"=>1234, "name_en"=>"Chocktou Lake", "lang"=>"en", "lat"=>39.960101, "lon"=>-83.485106),
"/US/US-OH/christiansburg-15939" => array("type"=>"village", "pop"=>537, "name_en"=>"Christiansburg", "lang"=>"en", "lat"=>40.056752, "lon"=>-84.025215),
"/US/US-OH/churchill-15940" => array("type"=>"cdp", "pop"=>2691, "name_en"=>"Churchill", "lang"=>"en", "lat"=>41.173608, "lon"=>-80.667301),
"/US/US-OH/cincinnati-15941" => array("type"=>"city", "pop"=>333336, "name_ru"=>"Цинциннати", "name_en"=>"Cincinnati", "lang"=>"en", "lat"=>39.139801, "lon"=>-84.505957),
"/US/US-OH/circleville-15942" => array("type"=>"city", "pop"=>13681, "name_ru"=>"Серклвилл", "name_en"=>"Circleville", "lang"=>"en", "lat"=>39.601816, "lon"=>-82.938959),
"/US/US-OH/clarington-15943" => array("type"=>"village", "pop"=>428, "name_en"=>"Clarington", "lang"=>"en", "lat"=>39.776780, "lon"=>-80.865084),
"/US/US-OH/clarksburg-15944" => array("type"=>"village", "pop"=>528, "name_ru"=>"Кларксбург", "name_en"=>"Clarksburg", "lang"=>"en", "lat"=>39.506326, "lon"=>-83.154614),
"/US/US-OH/clarksville-15945" => array("type"=>"village", "pop"=>489, "name_ru"=>"Кларксвилл", "name_en"=>"Clarksville", "lang"=>"en", "lat"=>39.402471, "lon"=>-83.982921),
"/US/US-OH/clay-center-15946" => array("type"=>"village", "pop"=>302, "name_ru"=>"Клэй Сентер", "name_en"=>"Clay Center", "lang"=>"en", "lat"=>41.568810, "lon"=>-83.360995),
"/US/US-OH/clayton-15947" => array("type"=>"village", "pop"=>12906, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>39.860667, "lon"=>-84.355085),
"/US/US-OH/cleveland-15948" => array("type"=>"city", "pop"=>433748, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>41.479700, "lon"=>-81.678511),
"/US/US-OH/cleveland-heights-15949" => array("type"=>"city", "pop"=>45827, "name_en"=>"Cleveland Heights", "lang"=>"en", "lat"=>41.509600, "lon"=>-81.563455),
"/US/US-OH/cleves-15950" => array("type"=>"village", "pop"=>2688, "name_en"=>"Cleves", "lang"=>"en", "lat"=>39.162301, "lon"=>-84.749278),
"/US/US-OH/clifton-15951" => array("type"=>"village", "pop"=>173, "name_ru"=>"Слифтон", "name_en"=>"Clifton", "lang"=>"en", "lat"=>39.797325, "lon"=>-83.825711),
"/US/US-OH/clinton-15952" => array("type"=>"village", "pop"=>1402, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>40.929100, "lon"=>-81.631118),
"/US/US-OH/cloverdale-15953" => array("type"=>"village", "pop"=>194, "name_ru"=>"Кловердейл", "name_en"=>"Cloverdale", "lang"=>"en", "lat"=>41.019621, "lon"=>-84.304180),
"/US/US-OH/clyde-15954" => array("type"=>"city", "pop"=>6121, "name_ru"=>"Клайд", "name_en"=>"Clyde", "lang"=>"en", "lat"=>41.305121, "lon"=>-82.977520),
"/US/US-OH/coal-grove-15955" => array("type"=>"village", "pop"=>2066, "name_en"=>"Coal Grove", "lang"=>"en", "lat"=>38.497989, "lon"=>-82.642672),
"/US/US-OH/coalton-15956" => array("type"=>"village", "pop"=>556, "name_ru"=>"Коалтон", "name_en"=>"Coalton", "lang"=>"en", "lat"=>39.112500, "lon"=>-82.611242),
"/US/US-OH/coldwater-15957" => array("type"=>"village", "pop"=>4396, "name_ru"=>"Коулдуотер", "name_en"=>"Coldwater", "lang"=>"en", "lat"=>40.482731, "lon"=>-84.631544),
"/US/US-OH/college-corner-15958" => array("type"=>"village", "pop"=>421, "name_en"=>"College Corner", "lang"=>"en", "lat"=>39.567800, "lon"=>-84.812582),
"/US/US-OH/columbiana-15959" => array("type"=>"village", "pop"=>5981, "name_ru"=>"Коламбиана", "name_en"=>"Columbiana", "lang"=>"en", "lat"=>40.885888, "lon"=>-80.688400),
"/US/US-OH/columbus-15960" => array("type"=>"city", "pop"=>754885, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>39.988933, "lon"=>-82.987381),
"/US/US-OH/columbus-grove-15961" => array("type"=>"village", "pop"=>2118, "name_en"=>"Columbus Grove", "lang"=>"en", "lat"=>40.919385, "lon"=>-84.058232),
"/US/US-OH/commercial-point-15962" => array("type"=>"village", "pop"=>830, "name_en"=>"Commercial Point", "lang"=>"en", "lat"=>39.769415, "lon"=>-83.057860),
"/US/US-OH/conesville-15963" => array("type"=>"village", "pop"=>365, "name_ru"=>"Коунсвилл", "name_en"=>"Conesville", "lang"=>"en", "lat"=>40.184978, "lon"=>-81.892296),
"/US/US-OH/congress-15964" => array("type"=>"village", "pop"=>188, "name_en"=>"Congress", "lang"=>"en", "lat"=>40.926295, "lon"=>-82.055290),
"/US/US-OH/conneaut-15965" => array("type"=>"city", "pop"=>12355, "name_en"=>"Conneaut", "lang"=>"en", "lat"=>41.926302, "lon"=>-80.570207),
"/US/US-OH/continental-15966" => array("type"=>"village", "pop"=>1162, "name_en"=>"Continental", "lang"=>"en", "lat"=>41.098888, "lon"=>-84.266822),
"/US/US-OH/convoy-15967" => array("type"=>"village", "pop"=>1050, "name_en"=>"Convoy", "lang"=>"en", "lat"=>40.917168, "lon"=>-84.706153),
"/US/US-OH/coolville-15968" => array("type"=>"village", "pop"=>544, "name_en"=>"Coolville", "lang"=>"en", "lat"=>39.220325, "lon"=>-81.798325),
"/US/US-OH/corning-15969" => array("type"=>"village", "pop"=>605, "name_ru"=>"Корнинг", "name_en"=>"Corning", "lang"=>"en", "lat"=>39.602316, "lon"=>-82.088657),
"/US/US-OH/cortland-15970" => array("type"=>"city", "pop"=>6364, "name_ru"=>"Кортленд", "name_en"=>"Cortland", "lang"=>"en", "lat"=>41.331463, "lon"=>-80.723417),
"/US/US-OH/corwin-15971" => array("type"=>"village", "pop"=>323, "name_en"=>"Corwin", "lang"=>"en", "lat"=>39.525045, "lon"=>-84.073204),
"/US/US-OH/coshocton-15972" => array("type"=>"city", "pop"=>11353, "name_en"=>"Coshocton", "lang"=>"en", "lat"=>40.263670, "lon"=>-81.849644),
"/US/US-OH/covedale-15973" => array("type"=>"cdp", "pop"=>6669, "name_en"=>"Covedale", "lang"=>"en", "lat"=>39.123801, "lon"=>-84.640608),
"/US/US-OH/covington-15974" => array("type"=>"village", "pop"=>2626, "name_ru"=>"Ковингтон", "name_en"=>"Covington", "lang"=>"en", "lat"=>40.119120, "lon"=>-84.350891),
"/US/US-OH/craig-beach-15975" => array("type"=>"village", "pop"=>1159, "name_en"=>"Craig Beach", "lang"=>"en", "lat"=>41.118504, "lon"=>-80.982391),
"/US/US-OH/crestline-15976" => array("type"=>"city", "pop"=>4882, "name_ru"=>"Крестлайн", "name_en"=>"Crestline", "lang"=>"en", "lat"=>40.785704, "lon"=>-82.738855),
"/US/US-OH/creston-15977" => array("type"=>"village", "pop"=>2114, "name_ru"=>"Крестон", "name_en"=>"Creston", "lang"=>"en", "lat"=>40.979390, "lon"=>-81.900409),
"/US/US-OH/cridersville-15978" => array("type"=>"village", "pop"=>1707, "name_en"=>"Cridersville", "lang"=>"en", "lat"=>40.652624, "lon"=>-84.145972),
"/US/US-OH/crooksville-15979" => array("type"=>"village", "pop"=>2443, "name_en"=>"Crooksville", "lang"=>"en", "lat"=>39.768120, "lon"=>-82.094912),
"/US/US-OH/crown-city-15980" => array("type"=>"village", "pop"=>430, "name_en"=>"Crown City", "lang"=>"en", "lat"=>38.588301, "lon"=>-82.289670),
"/US/US-OH/crystal-lakes-15981" => array("type"=>"cdp", "pop"=>1613, "name_ru"=>"Кристал Лейкс", "name_en"=>"Crystal Lakes", "lang"=>"en", "lat"=>39.884893, "lon"=>-84.025422),
"/US/US-OH/cumberland-15982" => array("type"=>"village", "pop"=>389, "name_ru"=>"Камберленд", "name_en"=>"Cumberland", "lang"=>"en", "lat"=>39.852299, "lon"=>-81.658632),
"/US/US-OH/custar-15983" => array("type"=>"village", "pop"=>207, "name_en"=>"Custar", "lang"=>"en", "lat"=>41.284246, "lon"=>-83.843459),
"/US/US-OH/cuyahoga-falls-15984" => array("type"=>"city", "pop"=>51090, "name_en"=>"Cuyahoga Falls", "lang"=>"en", "lat"=>41.168981, "lon"=>-81.522573),
"/US/US-OH/cuyahoga-heights-15985" => array("type"=>"village", "pop"=>533, "name_en"=>"Cuyahoga Heights", "lang"=>"en", "lat"=>41.436150, "lon"=>-81.653296),
"/US/US-OH/cygnet-15986" => array("type"=>"village", "pop"=>557, "name_en"=>"Cygnet", "lang"=>"en", "lat"=>41.240126, "lon"=>-83.643798),
"/US/US-OH/dalton-15987" => array("type"=>"village", "pop"=>1713, "name_ru"=>"Далтон", "name_en"=>"Dalton", "lang"=>"en", "lat"=>40.799240, "lon"=>-81.697179),
"/US/US-OH/danville-15988" => array("type"=>"village", "pop"=>1073, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>40.447939, "lon"=>-82.260114),
"/US/US-OH/darbyville-15989" => array("type"=>"village", "pop"=>298, "name_en"=>"Darbyville", "lang"=>"en", "lat"=>39.695744, "lon"=>-83.114188),
"/US/US-OH/day-heights-15990" => array("type"=>"cdp", "pop"=>2812, "name_en"=>"Day Heights", "lang"=>"en", "lat"=>39.173926, "lon"=>-84.226349),
"/US/US-OH/dayton-15991" => array("type"=>"city", "pop"=>154200, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>39.779067, "lon"=>-84.197357),
"/US/US-OH/de-graff-15995" => array("type"=>"village", "pop"=>1148, "name_ru"=>"Де Графф", "name_en"=>"De Graff", "lang"=>"en", "lat"=>40.310263, "lon"=>-83.914710),
"/US/US-OH/deer-park-15992" => array("type"=>"city", "pop"=>5751, "name_ru"=>"Дир Парк", "name_en"=>"Deer Park", "lang"=>"en", "lat"=>39.203401, "lon"=>-84.396967),
"/US/US-OH/deersville-15993" => array("type"=>"village", "pop"=>82, "name_en"=>"Deersville", "lang"=>"en", "lat"=>40.308146, "lon"=>-81.187696),
"/US/US-OH/defiance-15994" => array("type"=>"city", "pop"=>16063, "name_ru"=>"Дефианс", "name_en"=>"Defiance", "lang"=>"en", "lat"=>41.281169, "lon"=>-84.361900),
"/US/US-OH/delaware-15996" => array("type"=>"city", "pop"=>33719, "name_ru"=>"Делавэр", "name_en"=>"Delaware", "lang"=>"en", "lat"=>40.298664, "lon"=>-83.081343),
"/US/US-OH/dellroy-15997" => array("type"=>"village", "pop"=>285, "name_en"=>"Dellroy", "lang"=>"en", "lat"=>40.555743, "lon"=>-81.199414),
"/US/US-OH/delphos-15998" => array("type"=>"city", "pop"=>6730, "name_ru"=>"Делфос", "name_en"=>"Delphos", "lang"=>"en", "lat"=>40.844392, "lon"=>-84.339409),
"/US/US-OH/delta-15999" => array("type"=>"village", "pop"=>2900, "name_ru"=>"Дельта", "name_en"=>"Delta", "lang"=>"en", "lat"=>41.574124, "lon"=>-84.002525),
"/US/US-OH/dennison-16000" => array("type"=>"village", "pop"=>2878, "name_ru"=>"Деннисон", "name_en"=>"Dennison", "lang"=>"en", "lat"=>40.397146, "lon"=>-81.327882),
"/US/US-OH/dent-16001" => array("type"=>"cdp", "pop"=>6416, "name_ru"=>"Дент", "name_en"=>"Dent", "lang"=>"en", "lat"=>39.193751, "lon"=>-84.660907),
"/US/US-OH/deshler-16002" => array("type"=>"village", "pop"=>1797, "name_en"=>"Deshler", "lang"=>"en", "lat"=>41.208022, "lon"=>-83.905679),
"/US/US-OH/devola-16003" => array("type"=>"cdp", "pop"=>2736, "name_en"=>"Devola", "lang"=>"en", "lat"=>39.473256, "lon"=>-81.469574),
"/US/US-OH/dexter-city-16004" => array("type"=>"village", "pop"=>160, "name_en"=>"Dexter City", "lang"=>"en", "lat"=>39.659193, "lon"=>-81.474871),
"/US/US-OH/dillonvale-16005" => array("type"=>"cdp", "pop"=>721, "name_en"=>"Dillonvale", "lang"=>"en", "lat"=>39.218018, "lon"=>-84.402132),
"/US/US-OH/dillonvale-16006" => array("type"=>"village", "pop"=>857, "name_en"=>"Dillonvale", "lang"=>"en", "lat"=>40.198592, "lon"=>-80.775345),
"/US/US-OH/donnelsville-16007" => array("type"=>"village", "pop"=>282, "name_en"=>"Donnelsville", "lang"=>"en", "lat"=>39.918300, "lon"=>-83.946533),
"/US/US-OH/dover-16008" => array("type"=>"city", "pop"=>12466, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>40.529757, "lon"=>-81.479796),
"/US/US-OH/doylestown-16009" => array("type"=>"village", "pop"=>2898, "name_en"=>"Doylestown", "lang"=>"en", "lat"=>40.970328, "lon"=>-81.696698),
"/US/US-OH/dresden-16010" => array("type"=>"village", "pop"=>1415, "name_ru"=>"Дрезден", "name_en"=>"Dresden", "lang"=>"en", "lat"=>40.121830, "lon"=>-82.011141),
"/US/US-OH/drexel-16011" => array("type"=>"cdp", "pop"=>5143, "name_ru"=>"Дрексел", "name_en"=>"Drexel", "lang"=>"en", "lat"=>39.744050, "lon"=>-84.291676),
"/US/US-OH/dry-run-16012" => array("type"=>"cdp", "pop"=>5389, "name_en"=>"Dry Run", "lang"=>"en", "lat"=>39.104301, "lon"=>-84.330657),
"/US/US-OH/dublin-16013" => array("type"=>"city", "pop"=>38536, "name_ru"=>"Дублин", "name_en"=>"Dublin", "lang"=>"en", "lat"=>40.118433, "lon"=>-83.132926),
"/US/US-OH/dunkirk-16014" => array("type"=>"village", "pop"=>940, "name_ru"=>"Дюнкерк", "name_en"=>"Dunkirk", "lang"=>"en", "lat"=>40.787640, "lon"=>-83.642538),
"/US/US-OH/dupont-16015" => array("type"=>"village", "pop"=>258, "name_ru"=>"Дюпон", "name_en"=>"Dupont", "lang"=>"en", "lat"=>41.054879, "lon"=>-84.302109),
"/US/US-OH/east-canton-16016" => array("type"=>"village", "pop"=>1602, "name_en"=>"East Canton", "lang"=>"en", "lat"=>40.788950, "lon"=>-81.283374),
"/US/US-OH/east-cleveland-16017" => array("type"=>"city", "pop"=>24507, "name_en"=>"East Cleveland", "lang"=>"en", "lat"=>41.531233, "lon"=>-81.580165),
"/US/US-OH/east-liverpool-16019" => array("type"=>"city", "pop"=>11996, "name_en"=>"East Liverpool", "lang"=>"en", "lat"=>40.632939, "lon"=>-80.569448),
"/US/US-OH/east-palestine-16020" => array("type"=>"city", "pop"=>4690, "name_en"=>"East Palestine", "lang"=>"en", "lat"=>40.837388, "lon"=>-80.544884),
"/US/US-OH/east-sparta-16021" => array("type"=>"village", "pop"=>775, "name_en"=>"East Sparta", "lang"=>"en", "lat"=>40.669833, "lon"=>-81.353846),
"/US/US-OH/eastlake-16018" => array("type"=>"city", "pop"=>19520, "name_ru"=>"Истлейк", "name_en"=>"Eastlake", "lang"=>"en", "lat"=>41.659633, "lon"=>-81.430555),
"/US/US-OH/eaton-16022" => array("type"=>"city", "pop"=>8005, "name_ru"=>"Итон", "name_en"=>"Eaton", "lang"=>"en", "lat"=>39.748536, "lon"=>-84.633854),
"/US/US-OH/eaton-estates-16023" => array("type"=>"cdp", "pop"=>1586, "name_en"=>"Eaton Estates", "lang"=>"en", "lat"=>41.305550, "lon"=>-82.009646),
"/US/US-OH/edgerton-16024" => array("type"=>"village", "pop"=>1961, "name_ru"=>"Эджертон", "name_en"=>"Edgerton", "lang"=>"en", "lat"=>41.446717, "lon"=>-84.749838),
"/US/US-OH/edgewood-16025" => array("type"=>"cdp", "pop"=>5189, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>41.878725, "lon"=>-80.740774),
"/US/US-OH/edison-16026" => array("type"=>"village", "pop"=>451, "name_ru"=>"Эдисон", "name_en"=>"Edison", "lang"=>"en", "lat"=>40.558055, "lon"=>-82.863552),
"/US/US-OH/edon-16027" => array("type"=>"village", "pop"=>840, "name_en"=>"Edon", "lang"=>"en", "lat"=>41.556607, "lon"=>-84.768301),
"/US/US-OH/eldorado-16028" => array("type"=>"village", "pop"=>502, "name_ru"=>"Эльдорадо", "name_en"=>"Eldorado", "lang"=>"en", "lat"=>39.903840, "lon"=>-84.675826),
"/US/US-OH/elgin-16029" => array("type"=>"village", "pop"=>48, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>40.741658, "lon"=>-84.476097),
"/US/US-OH/elida-16030" => array("type"=>"village", "pop"=>1899, "name_en"=>"Elida", "lang"=>"en", "lat"=>40.786000, "lon"=>-84.199783),
"/US/US-OH/elmore-16031" => array("type"=>"village", "pop"=>1365, "name_ru"=>"Элмор", "name_en"=>"Elmore", "lang"=>"en", "lat"=>41.472952, "lon"=>-83.291919),
"/US/US-OH/elmwood-place-16032" => array("type"=>"village", "pop"=>2481, "name_en"=>"Elmwood Place", "lang"=>"en", "lat"=>39.184851, "lon"=>-84.488518),
"/US/US-OH/elyria-16033" => array("type"=>"city", "pop"=>54979, "name_en"=>"Elyria", "lang"=>"en", "lat"=>41.379867, "lon"=>-82.105581),
"/US/US-OH/empire-16034" => array("type"=>"village", "pop"=>276, "name_ru"=>"Эмпаер", "name_en"=>"Empire", "lang"=>"en", "lat"=>40.510960, "lon"=>-80.625019),
"/US/US-OH/englewood-16035" => array("type"=>"city", "pop"=>12686, "name_ru"=>"Энглвуд", "name_en"=>"Englewood", "lang"=>"en", "lat"=>39.860750, "lon"=>-84.306899),
"/US/US-OH/enon-16036" => array("type"=>"village", "pop"=>2542, "name_en"=>"Enon", "lang"=>"en", "lat"=>39.865519, "lon"=>-83.935519),
"/US/US-OH/euclid-16037" => array("type"=>"city", "pop"=>47415, "name_en"=>"Euclid", "lang"=>"en", "lat"=>41.592267, "lon"=>-81.519442),
"/US/US-OH/evendale-16038" => array("type"=>"village", "pop"=>2955, "name_en"=>"Evendale", "lang"=>"en", "lat"=>39.253801, "lon"=>-84.425704),
"/US/US-OH/fairborn-16039" => array("type"=>"city", "pop"=>32366, "name_en"=>"Fairborn", "lang"=>"en", "lat"=>39.804450, "lon"=>-84.017497),
"/US/US-OH/fairfax-16040" => array("type"=>"village", "pop"=>1855, "name_ru"=>"Фэрфакс", "name_en"=>"Fairfax", "lang"=>"en", "lat"=>39.142701, "lon"=>-84.396248),
"/US/US-OH/fairfield-16041" => array("type"=>"city", "pop"=>42384, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>39.331100, "lon"=>-84.542897),
"/US/US-OH/fairfield-beach-16042" => array("type"=>"cdp", "pop"=>1084, "name_en"=>"Fairfield Beach", "lang"=>"en", "lat"=>39.917842, "lon"=>-82.481125),
"/US/US-OH/fairlawn-16043" => array("type"=>"city", "pop"=>7030, "name_en"=>"Fairlawn", "lang"=>"en", "lat"=>41.124950, "lon"=>-81.621260),
"/US/US-OH/fairport-harbor-16044" => array("type"=>"village", "pop"=>3211, "name_en"=>"Fairport Harbor", "lang"=>"en", "lat"=>41.747500, "lon"=>-81.273299),
"/US/US-OH/fairview-16045" => array("type"=>"village", "pop"=>80, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>40.058889, "lon"=>-81.234082),
"/US/US-OH/fairview-lanes-16046" => array("type"=>"cdp", "pop"=>1120, "name_en"=>"Fairview Lanes", "lang"=>"en", "lat"=>41.427072, "lon"=>-82.656683),
"/US/US-OH/fairview-park-16047" => array("type"=>"city", "pop"=>15762, "name_ru"=>"Фэрвью Парк", "name_en"=>"Fairview Park", "lang"=>"en", "lat"=>41.440650, "lon"=>-81.854055),
"/US/US-OH/farmersville-16048" => array("type"=>"village", "pop"=>937, "name_ru"=>"Фармерсвилл", "name_en"=>"Farmersville", "lang"=>"en", "lat"=>39.678700, "lon"=>-84.427052),
"/US/US-OH/fayette-16049" => array("type"=>"village", "pop"=>1287, "name_ru"=>"Файет", "name_en"=>"Fayette", "lang"=>"en", "lat"=>41.672734, "lon"=>-84.329074),
"/US/US-OH/fayetteville-16050" => array("type"=>"village", "pop"=>367, "name_ru"=>"Фейетвилл", "name_en"=>"Fayetteville", "lang"=>"en", "lat"=>39.186366, "lon"=>-83.932082),
"/US/US-OH/felicity-16051" => array("type"=>"village", "pop"=>930, "name_en"=>"Felicity", "lang"=>"en", "lat"=>38.839867, "lon"=>-84.097928),
"/US/US-OH/findlay-16052" => array("type"=>"city", "pop"=>36987, "name_ru"=>"Файндлей", "name_en"=>"Findlay", "lang"=>"en", "lat"=>41.043011, "lon"=>-83.642183),
"/US/US-OH/finneytown-16053" => array("type"=>"cdp", "pop"=>13096, "name_en"=>"Finneytown", "lang"=>"en", "lat"=>39.216701, "lon"=>-84.514218),
"/US/US-OH/five-points-16054" => array("type"=>"cdp", "pop"=>1554, "name_ru"=>"Файв Поинтс", "name_en"=>"Five Points", "lang"=>"en", "lat"=>39.561567, "lon"=>-84.197147),
"/US/US-OH/fletcher-16055" => array("type"=>"village", "pop"=>504, "name_en"=>"Fletcher", "lang"=>"en", "lat"=>40.144432, "lon"=>-84.111766),
"/US/US-OH/florida-16056" => array("type"=>"village", "pop"=>236, "name_ru"=>"Флорида", "name_en"=>"Florida", "lang"=>"en", "lat"=>41.323000, "lon"=>-84.200584),
"/US/US-OH/flushing-16057" => array("type"=>"village", "pop"=>864, "name_ru"=>"Флашинг", "name_en"=>"Flushing", "lang"=>"en", "lat"=>40.148475, "lon"=>-81.065498),
"/US/US-OH/forest-16058" => array("type"=>"village", "pop"=>1449, "name_ru"=>"Форест", "name_en"=>"Forest", "lang"=>"en", "lat"=>40.803098, "lon"=>-83.511891),
"/US/US-OH/forest-park-16059" => array("type"=>"city", "pop"=>18514, "name_ru"=>"Форест Парк", "name_en"=>"Forest Park", "lang"=>"en", "lat"=>39.287216, "lon"=>-84.526936),
"/US/US-OH/forestville-16060" => array("type"=>"cdp", "pop"=>9185, "name_ru"=>"Форествилл", "name_en"=>"Forestville", "lang"=>"en", "lat"=>39.070068, "lon"=>-84.338840),
"/US/US-OH/fort-jennings-16061" => array("type"=>"village", "pop"=>419, "name_en"=>"Fort Jennings", "lang"=>"en", "lat"=>40.906142, "lon"=>-84.299298),
"/US/US-OH/fort-loramie-16062" => array("type"=>"village", "pop"=>1463, "name_en"=>"Fort Loramie", "lang"=>"en", "lat"=>40.346310, "lon"=>-84.370333),
"/US/US-OH/fort-mckinley-16063" => array("type"=>"cdp", "pop"=>9740, "name_en"=>"Fort McKinley", "lang"=>"en", "lat"=>39.806000, "lon"=>-84.256655),
"/US/US-OH/fort-recovery-16064" => array("type"=>"village", "pop"=>1332, "name_en"=>"Fort Recovery", "lang"=>"en", "lat"=>40.410897, "lon"=>-84.775104),
"/US/US-OH/fort-shawnee-16065" => array("type"=>"village", "pop"=>3704, "name_en"=>"Fort Shawnee", "lang"=>"en", "lat"=>40.681150, "lon"=>-84.131500),
"/US/US-OH/fostoria-16066" => array("type"=>"city", "pop"=>13000, "name_ru"=>"Фостория", "name_en"=>"Fostoria", "lang"=>"en", "lat"=>41.159165, "lon"=>-83.413578),
"/US/US-OH/frankfort-16067" => array("type"=>"village", "pop"=>1043, "name_ru"=>"Франкфорт", "name_en"=>"Frankfort", "lang"=>"en", "lat"=>39.407999, "lon"=>-83.182840),
"/US/US-OH/franklin-16068" => array("type"=>"city", "pop"=>12941, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>39.553918, "lon"=>-84.301167),
"/US/US-OH/franklin-furnace-16069" => array("type"=>"cdp", "pop"=>1212, "name_en"=>"Franklin Furnace", "lang"=>"en", "lat"=>38.607952, "lon"=>-82.846495),
"/US/US-OH/frazeysburg-16070" => array("type"=>"village", "pop"=>1291, "name_en"=>"Frazeysburg", "lang"=>"en", "lat"=>40.116253, "lon"=>-82.118654),
"/US/US-OH/fredericksburg-16071" => array("type"=>"village", "pop"=>475, "name_ru"=>"Фредериксбург", "name_en"=>"Fredericksburg", "lang"=>"en", "lat"=>40.677378, "lon"=>-81.869728),
"/US/US-OH/fredericktown-16072" => array("type"=>"village", "pop"=>2484, "name_ru"=>"Фредериктаун", "name_en"=>"Fredericktown", "lang"=>"en", "lat"=>40.478880, "lon"=>-82.545311),
"/US/US-OH/freeport-16073" => array("type"=>"village", "pop"=>394, "name_ru"=>"Фрипорт", "name_en"=>"Freeport", "lang"=>"en", "lat"=>40.210805, "lon"=>-81.268271),
"/US/US-OH/fremont-16074" => array("type"=>"city", "pop"=>16655, "name_ru"=>"Фремонт", "name_en"=>"Fremont", "lang"=>"en", "lat"=>41.350313, "lon"=>-83.113050),
"/US/US-OH/fruit-hill-16075" => array("type"=>"cdp", "pop"=>4101, "name_en"=>"Fruit Hill", "lang"=>"en", "lat"=>39.068051, "lon"=>-84.366198),
"/US/US-OH/fulton-16076" => array("type"=>"village", "pop"=>270, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>40.462360, "lon"=>-82.828411),
"/US/US-OH/fultonham-16077" => array("type"=>"village", "pop"=>156, "name_en"=>"Fultonham", "lang"=>"en", "lat"=>39.856082, "lon"=>-82.142884),
"/US/US-OH/gahanna-16078" => array("type"=>"city", "pop"=>33892, "name_en"=>"Gahanna", "lang"=>"en", "lat"=>40.025033, "lon"=>-82.875001),
"/US/US-OH/galena-16079" => array("type"=>"village", "pop"=>548, "name_ru"=>"Галена", "name_en"=>"Galena", "lang"=>"en", "lat"=>40.220501, "lon"=>-82.882075),
"/US/US-OH/galion-16080" => array("type"=>"city", "pop"=>10737, "name_en"=>"Galion", "lang"=>"en", "lat"=>40.731724, "lon"=>-82.790474),
"/US/US-OH/gallipolis-16081" => array("type"=>"city", "pop"=>4190, "name_en"=>"Gallipolis", "lang"=>"en", "lat"=>38.819962, "lon"=>-82.191459),
"/US/US-OH/gambier-16082" => array("type"=>"village", "pop"=>2061, "name_en"=>"Gambier", "lang"=>"en", "lat"=>40.376285, "lon"=>-82.394883),
"/US/US-OH/gann-16083" => array("type"=>"village", "pop"=>143, "name_en"=>"Gann", "lang"=>"en", "lat"=>40.469111, "lon"=>-82.190306),
"/US/US-OH/garfield-heights-16084" => array("type"=>"city", "pop"=>27734, "name_en"=>"Garfield Heights", "lang"=>"en", "lat"=>41.419686, "lon"=>-81.604008),
"/US/US-OH/garrettsville-16085" => array("type"=>"village", "pop"=>2206, "name_en"=>"Garrettsville", "lang"=>"en", "lat"=>41.283900, "lon"=>-81.095361),
"/US/US-OH/gates-mills-16086" => array("type"=>"village", "pop"=>2276, "name_en"=>"Gates Mills", "lang"=>"en", "lat"=>41.532533, "lon"=>-81.410691),
"/US/US-OH/geneva-16087" => array("type"=>"city", "pop"=>6291, "name_ru"=>"Дженива", "name_en"=>"Geneva", "lang"=>"en", "lat"=>41.799813, "lon"=>-80.948467),
"/US/US-OH/geneva-on-the-lake-16088" => array("type"=>"village", "pop"=>1486, "name_en"=>"Geneva-on-the-Lake", "lang"=>"en", "lat"=>41.856577, "lon"=>-80.945952),
"/US/US-OH/genoa-16089" => array("type"=>"village", "pop"=>2279, "name_ru"=>"Дженоа", "name_en"=>"Genoa", "lang"=>"en", "lat"=>41.518417, "lon"=>-83.361010),
"/US/US-OH/georgetown-16090" => array("type"=>"village", "pop"=>3430, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>38.866857, "lon"=>-83.902663),
"/US/US-OH/germantown-16091" => array("type"=>"city", "pop"=>5018, "name_ru"=>"Джермэнтаун", "name_en"=>"Germantown", "lang"=>"en", "lat"=>39.634955, "lon"=>-84.362458),
"/US/US-OH/gettysburg-16092" => array("type"=>"village", "pop"=>525, "name_en"=>"Gettysburg", "lang"=>"en", "lat"=>40.115728, "lon"=>-84.495955),
"/US/US-OH/gibsonburg-16093" => array("type"=>"village", "pop"=>2435, "name_en"=>"Gibsonburg", "lang"=>"en", "lat"=>41.386531, "lon"=>-83.322847),
"/US/US-OH/gilboa-16094" => array("type"=>"village", "pop"=>168, "name_en"=>"Gilboa", "lang"=>"en", "lat"=>41.018914, "lon"=>-83.921358),
"/US/US-OH/girard-16095" => array("type"=>"city", "pop"=>10069, "name_ru"=>"Жирар", "name_en"=>"Girard", "lang"=>"en", "lat"=>41.154550, "lon"=>-80.697549),
"/US/US-OH/glandorf-16096" => array("type"=>"village", "pop"=>955, "name_en"=>"Glandorf", "lang"=>"en", "lat"=>41.030309, "lon"=>-84.079173),
"/US/US-OH/glendale-16097" => array("type"=>"village", "pop"=>2214, "name_ru"=>"Глендейл", "name_en"=>"Glendale", "lang"=>"en", "lat"=>39.270749, "lon"=>-84.458365),
"/US/US-OH/glenford-16098" => array("type"=>"village", "pop"=>195, "name_en"=>"Glenford", "lang"=>"en", "lat"=>39.886152, "lon"=>-82.319844),
"/US/US-OH/glenmont-16099" => array("type"=>"village", "pop"=>294, "name_en"=>"Glenmont", "lang"=>"en", "lat"=>40.517047, "lon"=>-82.092736),
"/US/US-OH/glenmoor-16100" => array("type"=>"cdp", "pop"=>2307, "name_en"=>"Glenmoor", "lang"=>"en", "lat"=>40.665135, "lon"=>-80.613250),
"/US/US-OH/glenwillow-16101" => array("type"=>"village", "pop"=>617, "name_en"=>"Glenwillow", "lang"=>"en", "lat"=>41.360767, "lon"=>-81.472396),
"/US/US-OH/gloria-glens-park-16102" => array("type"=>"village", "pop"=>578, "name_en"=>"Gloria Glens Park", "lang"=>"en", "lat"=>41.058290, "lon"=>-81.900894),
"/US/US-OH/glouster-16103" => array("type"=>"village", "pop"=>2023, "name_en"=>"Glouster", "lang"=>"en", "lat"=>39.502801, "lon"=>-82.084757),
"/US/US-OH/gnadenhutten-16104" => array("type"=>"village", "pop"=>1283, "name_en"=>"Gnadenhutten", "lang"=>"en", "lat"=>40.359178, "lon"=>-81.431250),
"/US/US-OH/golf-manor-16105" => array("type"=>"village", "pop"=>3765, "name_en"=>"Golf Manor", "lang"=>"en", "lat"=>39.187051, "lon"=>-84.446424),
"/US/US-OH/gordon-16106" => array("type"=>"village", "pop"=>183, "name_ru"=>"Гордон", "name_en"=>"Gordon", "lang"=>"en", "lat"=>39.929912, "lon"=>-84.508577),
"/US/US-OH/grafton-16107" => array("type"=>"village", "pop"=>6716, "name_ru"=>"Графтон", "name_en"=>"Grafton", "lang"=>"en", "lat"=>41.280497, "lon"=>-82.033152),
"/US/US-OH/grand-rapids-16108" => array("type"=>"village", "pop"=>1015, "name_ru"=>"Гранд Рэпидс", "name_en"=>"Grand Rapids", "lang"=>"en", "lat"=>41.410400, "lon"=>-83.865748),
"/US/US-OH/grand-river-16109" => array("type"=>"village", "pop"=>366, "name_ru"=>"Гранд Ривер", "name_en"=>"Grand River", "lang"=>"en", "lat"=>41.742250, "lon"=>-81.286033),
"/US/US-OH/grandview-16110" => array("type"=>"cdp", "pop"=>1757, "name_ru"=>"Грандвью", "name_en"=>"Grandview", "lang"=>"en", "lat"=>39.194251, "lon"=>-84.724390),
"/US/US-OH/grandview-heights-16111" => array("type"=>"city", "pop"=>6306, "name_en"=>"Grandview Heights", "lang"=>"en", "lat"=>39.979500, "lon"=>-83.039933),
"/US/US-OH/granville-16112" => array("type"=>"village", "pop"=>5382, "name_ru"=>"Гранвилл", "name_en"=>"Granville", "lang"=>"en", "lat"=>40.071008, "lon"=>-82.501580),
"/US/US-OH/granville-south-16113" => array("type"=>"cdp", "pop"=>1124, "name_en"=>"Granville South", "lang"=>"en", "lat"=>40.052933, "lon"=>-82.540243),
"/US/US-OH/gratiot-16114" => array("type"=>"village", "pop"=>190, "name_en"=>"Gratiot", "lang"=>"en", "lat"=>39.951448, "lon"=>-82.217319),
"/US/US-OH/gratis-16115" => array("type"=>"village", "pop"=>878, "name_en"=>"Gratis", "lang"=>"en", "lat"=>39.648352, "lon"=>-84.528767),
"/US/US-OH/graysville-16116" => array("type"=>"village", "pop"=>109, "name_ru"=>"Грейсвилл", "name_en"=>"Graysville", "lang"=>"en", "lat"=>39.661068, "lon"=>-81.171060),
"/US/US-OH/green-16117" => array("type"=>"village", "pop"=>23404, "name_ru"=>"Грин", "name_en"=>"Green", "lang"=>"en", "lat"=>40.947650, "lon"=>-81.486482),
"/US/US-OH/green-camp-16118" => array("type"=>"village", "pop"=>313, "name_en"=>"Green Camp", "lang"=>"en", "lat"=>40.532825, "lon"=>-83.207970),
"/US/US-OH/green-meadows-16121" => array("type"=>"cdp", "pop"=>2526, "name_en"=>"Green Meadows", "lang"=>"en", "lat"=>39.868750, "lon"=>-83.944330),
"/US/US-OH/green-springs-16123" => array("type"=>"village", "pop"=>1200, "name_en"=>"Green Springs", "lang"=>"en", "lat"=>41.257098, "lon"=>-83.053700),
"/US/US-OH/greenfield-16119" => array("type"=>"city", "pop"=>5063, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>39.354188, "lon"=>-83.389060),
"/US/US-OH/greenhills-16120" => array("type"=>"village", "pop"=>3837, "name_en"=>"Greenhills", "lang"=>"en", "lat"=>39.264734, "lon"=>-84.517557),
"/US/US-OH/greensburg-16122" => array("type"=>"cdp", "pop"=>1519, "name_ru"=>"Гринсбург", "name_en"=>"Greensburg", "lang"=>"en", "lat"=>40.936350, "lon"=>-81.436598),
"/US/US-OH/greentown-16124" => array("type"=>"cdp", "pop"=>1856, "name_ru"=>"Гринтаун", "name_en"=>"Greentown", "lang"=>"en", "lat"=>40.927518, "lon"=>-81.401875),
"/US/US-OH/greenville-16125" => array("type"=>"city", "pop"=>12983, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>40.102814, "lon"=>-84.624765),
"/US/US-OH/greenwich-16126" => array("type"=>"village", "pop"=>1503, "name_en"=>"Greenwich", "lang"=>"en", "lat"=>41.031127, "lon"=>-82.517921),
"/US/US-OH/groesbeck-16127" => array("type"=>"cdp", "pop"=>6684, "name_en"=>"Groesbeck", "lang"=>"en", "lat"=>39.228801, "lon"=>-84.594284),
"/US/US-OH/grove-city-16128" => array("type"=>"city", "pop"=>33829, "name_ru"=>"Гроув Сити", "name_en"=>"Grove City", "lang"=>"en", "lat"=>39.873128, "lon"=>-83.074693),
"/US/US-OH/groveport-16129" => array("type"=>"village", "pop"=>5240, "name_en"=>"Groveport", "lang"=>"en", "lat"=>39.861617, "lon"=>-82.897324),
"/US/US-OH/grover-hill-16130" => array("type"=>"village", "pop"=>368, "name_en"=>"Grover Hill", "lang"=>"en", "lat"=>41.018672, "lon"=>-84.477091),
"/US/US-OH/hamden-16131" => array("type"=>"village", "pop"=>929, "name_en"=>"Hamden", "lang"=>"en", "lat"=>39.160005, "lon"=>-82.525357),
"/US/US-OH/hamersville-16132" => array("type"=>"village", "pop"=>488, "name_en"=>"Hamersville", "lang"=>"en", "lat"=>38.918276, "lon"=>-83.984959),
"/US/US-OH/hamilton-16133" => array("type"=>"city", "pop"=>62477, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>39.390750, "lon"=>-84.564099),
"/US/US-OH/hamler-16134" => array("type"=>"village", "pop"=>645, "name_en"=>"Hamler", "lang"=>"en", "lat"=>41.230192, "lon"=>-84.033254),
"/US/US-OH/hanging-rock-16135" => array("type"=>"village", "pop"=>289, "name_en"=>"Hanging Rock", "lang"=>"en", "lat"=>38.558941, "lon"=>-82.726184),
"/US/US-OH/hanover-16136" => array("type"=>"village", "pop"=>986, "name_ru"=>"Хановер", "name_en"=>"Hanover", "lang"=>"en", "lat"=>40.079167, "lon"=>-82.270528),
"/US/US-OH/hanoverton-16137" => array("type"=>"village", "pop"=>377, "name_en"=>"Hanoverton", "lang"=>"en", "lat"=>40.755025, "lon"=>-80.935628),
"/US/US-OH/harbor-hills-16138" => array("type"=>"cdp", "pop"=>1372, "name_en"=>"Harbor Hills", "lang"=>"en", "lat"=>39.936745, "lon"=>-82.435245),
"/US/US-OH/harbor-view-16139" => array("type"=>"village", "pop"=>99, "name_en"=>"Harbor View", "lang"=>"en", "lat"=>41.692633, "lon"=>-83.444599),
"/US/US-OH/harpster-16140" => array("type"=>"village", "pop"=>191, "name_en"=>"Harpster", "lang"=>"en", "lat"=>40.738597, "lon"=>-83.250586),
"/US/US-OH/harrisburg-16141" => array("type"=>"village", "pop"=>316, "name_ru"=>"Харрисбург", "name_en"=>"Harrisburg", "lang"=>"en", "lat"=>39.810840, "lon"=>-83.171470),
"/US/US-OH/harrison-16142" => array("type"=>"city", "pop"=>9387, "name_ru"=>"Харрисон", "name_en"=>"Harrison", "lang"=>"en", "lat"=>39.249868, "lon"=>-84.793025),
"/US/US-OH/harrisville-16143" => array("type"=>"village", "pop"=>256, "name_ru"=>"Харрисвилл", "name_en"=>"Harrisville", "lang"=>"en", "lat"=>40.181751, "lon"=>-80.887144),
"/US/US-OH/harrod-16144" => array("type"=>"village", "pop"=>470, "name_en"=>"Harrod", "lang"=>"en", "lat"=>40.708376, "lon"=>-83.920357),
"/US/US-OH/hartford-16145" => array("type"=>"village", "pop"=>398, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>40.239367, "lon"=>-82.687665),
"/US/US-OH/hartville-16146" => array("type"=>"village", "pop"=>2553, "name_ru"=>"Хартвилл", "name_en"=>"Hartville", "lang"=>"en", "lat"=>40.961000, "lon"=>-81.334004),
"/US/US-OH/harveysburg-16147" => array("type"=>"village", "pop"=>631, "name_en"=>"Harveysburg", "lang"=>"en", "lat"=>39.500332, "lon"=>-84.002259),
"/US/US-OH/haskins-16148" => array("type"=>"village", "pop"=>641, "name_en"=>"Haskins", "lang"=>"en", "lat"=>41.463846, "lon"=>-83.704065),
"/US/US-OH/haviland-16149" => array("type"=>"village", "pop"=>165, "name_ru"=>"Хавиленд", "name_en"=>"Haviland", "lang"=>"en", "lat"=>41.018378, "lon"=>-84.581138),
"/US/US-OH/hayesville-16150" => array("type"=>"village", "pop"=>348, "name_ru"=>"Хайесвилл", "name_en"=>"Hayesville", "lang"=>"en", "lat"=>40.774733, "lon"=>-82.260661),
"/US/US-OH/heath-16151" => array("type"=>"city", "pop"=>8903, "name_ru"=>"Хит", "name_en"=>"Heath", "lang"=>"en", "lat"=>40.026867, "lon"=>-82.437355),
"/US/US-OH/hebron-16152" => array("type"=>"village", "pop"=>2147, "name_ru"=>"Хеброн", "name_en"=>"Hebron", "lang"=>"en", "lat"=>39.962637, "lon"=>-82.491554),
"/US/US-OH/helena-16153" => array("type"=>"village", "pop"=>230, "name_ru"=>"Хелена", "name_en"=>"Helena", "lang"=>"en", "lat"=>41.339811, "lon"=>-83.291825),
"/US/US-OH/hemlock-16154" => array("type"=>"village", "pop"=>145, "name_ru"=>"Хемлок", "name_en"=>"Hemlock", "lang"=>"en", "lat"=>39.588936, "lon"=>-82.155343),
"/US/US-OH/hicksville-16155" => array("type"=>"village", "pop"=>3408, "name_en"=>"Hicksville", "lang"=>"en", "lat"=>41.294497, "lon"=>-84.766615),
"/US/US-OH/higginsport-16156" => array("type"=>"village", "pop"=>276, "name_en"=>"Higginsport", "lang"=>"en", "lat"=>38.790669, "lon"=>-83.968437),
"/US/US-OH/highland-16157" => array("type"=>"village", "pop"=>290, "name_ru"=>"Хайленд", "name_en"=>"Highland", "lang"=>"en", "lat"=>39.343579, "lon"=>-83.599629),
"/US/US-OH/highland-heights-16158" => array("type"=>"city", "pop"=>8533, "name_ru"=>"Хайленд Хайтс", "name_en"=>"Highland Heights", "lang"=>"en", "lat"=>41.548981, "lon"=>-81.471580),
"/US/US-OH/hilliard-16159" => array("type"=>"city", "pop"=>27901, "name_ru"=>"Хиллиард", "name_en"=>"Hilliard", "lang"=>"en", "lat"=>40.037975, "lon"=>-83.142580),
"/US/US-OH/hills-and-dales-16160" => array("type"=>"village", "pop"=>263, "name_ru"=>"Хиллс энд Дейлс", "name_en"=>"Hills and Dales", "lang"=>"en", "lat"=>40.828133, "lon"=>-81.444931),
"/US/US-OH/hillsboro-16161" => array("type"=>"city", "pop"=>6661, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>39.209246, "lon"=>-83.612526),
"/US/US-OH/hiram-16162" => array("type"=>"village", "pop"=>1192, "name_ru"=>"Хирам", "name_en"=>"Hiram", "lang"=>"en", "lat"=>41.310150, "lon"=>-81.144474),
"/US/US-OH/holgate-16163" => array("type"=>"village", "pop"=>1113, "name_en"=>"Holgate", "lang"=>"en", "lat"=>41.248847, "lon"=>-84.128595),
"/US/US-OH/holiday-valley-16164" => array("type"=>"cdp", "pop"=>1243, "name_en"=>"Holiday Valley", "lang"=>"en", "lat"=>39.847709, "lon"=>-83.966523),
"/US/US-OH/holland-16165" => array("type"=>"village", "pop"=>1327, "name_ru"=>"Холланд", "name_en"=>"Holland", "lang"=>"en", "lat"=>41.620314, "lon"=>-83.709512),
"/US/US-OH/hollansburg-16166" => array("type"=>"village", "pop"=>206, "name_en"=>"Hollansburg", "lang"=>"en", "lat"=>39.998467, "lon"=>-84.793030),
"/US/US-OH/holloway-16167" => array("type"=>"village", "pop"=>326, "name_ru"=>"Холлоуэй", "name_en"=>"Holloway", "lang"=>"en", "lat"=>40.160486, "lon"=>-81.126401),
"/US/US-OH/holmesville-16168" => array("type"=>"village", "pop"=>409, "name_en"=>"Holmesville", "lang"=>"en", "lat"=>40.628427, "lon"=>-81.923442),
"/US/US-OH/hopedale-16169" => array("type"=>"village", "pop"=>962, "name_ru"=>"Хоупдейл", "name_en"=>"Hopedale", "lang"=>"en", "lat"=>40.323323, "lon"=>-80.899020),
"/US/US-OH/howland-center-16170" => array("type"=>"cdp", "pop"=>6732, "name_en"=>"Howland Center", "lang"=>"en", "lat"=>41.251248, "lon"=>-80.745693),
"/US/US-OH/hoytville-16171" => array("type"=>"village", "pop"=>296, "name_en"=>"Hoytville", "lang"=>"en", "lat"=>41.190338, "lon"=>-83.783699),
"/US/US-OH/hubbard-16172" => array("type"=>"city", "pop"=>7688, "name_ru"=>"Хаббард", "name_en"=>"Hubbard", "lang"=>"en", "lat"=>41.159952, "lon"=>-80.569607),
"/US/US-OH/huber-heights-16173" => array("type"=>"city", "pop"=>37148, "name_en"=>"Huber Heights", "lang"=>"en", "lat"=>39.855850, "lon"=>-84.112254),
"/US/US-OH/huber-ridge-16174" => array("type"=>"cdp", "pop"=>5255, "name_en"=>"Huber Ridge", "lang"=>"en", "lat"=>40.090150, "lon"=>-82.918394),
"/US/US-OH/hudson-16175" => array("type"=>"village", "pop"=>23037, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>41.244658, "lon"=>-81.451167),
"/US/US-OH/hunting-valley-16176" => array("type"=>"village", "pop"=>701, "name_en"=>"Hunting Valley", "lang"=>"en", "lat"=>41.473000, "lon"=>-81.409069),
"/US/US-OH/huntsville-16177" => array("type"=>"village", "pop"=>421, "name_ru"=>"Хантсвилл", "name_en"=>"Huntsville", "lang"=>"en", "lat"=>40.442422, "lon"=>-83.803878),
"/US/US-OH/huron-16178" => array("type"=>"city", "pop"=>7348, "name_ru"=>"Гурон", "name_en"=>"Huron", "lang"=>"en", "lat"=>41.396472, "lon"=>-82.562139),
"/US/US-OH/independence-16179" => array("type"=>"city", "pop"=>6770, "name_ru"=>"Индепенденс", "name_en"=>"Independence", "lang"=>"en", "lat"=>41.380317, "lon"=>-81.641507),
"/US/US-OH/irondale-16180" => array("type"=>"village", "pop"=>396, "name_ru"=>"Айрондейл", "name_en"=>"Irondale", "lang"=>"en", "lat"=>40.572683, "lon"=>-80.726006),
"/US/US-OH/ironton-16181" => array("type"=>"city", "pop"=>11307, "name_ru"=>"Айронтон", "name_en"=>"Ironton", "lang"=>"en", "lat"=>38.532480, "lon"=>-82.678456),
"/US/US-OH/ithaca-16182" => array("type"=>"village", "pop"=>98, "name_ru"=>"Итака", "name_en"=>"Ithaca", "lang"=>"en", "lat"=>39.938224, "lon"=>-84.553443),
"/US/US-OH/jackson-16183" => array("type"=>"city", "pop"=>6138, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>39.045718, "lon"=>-82.629211),
"/US/US-OH/jackson-center-16185" => array("type"=>"village", "pop"=>1443, "name_en"=>"Jackson Center", "lang"=>"en", "lat"=>40.439137, "lon"=>-84.039465),
"/US/US-OH/jacksonburg-16184" => array("type"=>"village", "pop"=>69, "name_en"=>"Jacksonburg", "lang"=>"en", "lat"=>39.537929, "lon"=>-84.503880),
"/US/US-OH/jacksonville-16186" => array("type"=>"village", "pop"=>560, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>39.475756, "lon"=>-82.080053),
"/US/US-OH/jamestown-16187" => array("type"=>"village", "pop"=>1861, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>39.658616, "lon"=>-83.739976),
"/US/US-OH/jefferson-16188" => array("type"=>"village", "pop"=>3416, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>41.738615, "lon"=>-80.768732),
"/US/US-OH/jefferson-16189" => array("type"=>"village", "pop"=>4505, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>39.941177, "lon"=>-83.294390),
"/US/US-OH/jeffersonville-16190" => array("type"=>"village", "pop"=>1229, "name_ru"=>"Джефферсонвилл", "name_en"=>"Jeffersonville", "lang"=>"en", "lat"=>39.652006, "lon"=>-83.556885),
"/US/US-OH/jenera-16191" => array("type"=>"village", "pop"=>214, "name_en"=>"Jenera", "lang"=>"en", "lat"=>40.899829, "lon"=>-83.727404),
"/US/US-OH/jeromesville-16192" => array("type"=>"village", "pop"=>483, "name_en"=>"Jeromesville", "lang"=>"en", "lat"=>40.803880, "lon"=>-82.196237),
"/US/US-OH/jerry-city-16193" => array("type"=>"village", "pop"=>469, "name_en"=>"Jerry City", "lang"=>"en", "lat"=>41.253112, "lon"=>-83.602721),
"/US/US-OH/jerusalem-16194" => array("type"=>"village", "pop"=>146, "name_en"=>"Jerusalem", "lang"=>"en", "lat"=>39.852189, "lon"=>-81.095320),
"/US/US-OH/jewett-16195" => array("type"=>"village", "pop"=>767, "name_ru"=>"Джеветт", "name_en"=>"Jewett", "lang"=>"en", "lat"=>40.368075, "lon"=>-81.002712),
"/US/US-OH/johnstown-16196" => array("type"=>"village", "pop"=>4016, "name_ru"=>"Джонстаун", "name_en"=>"Johnstown", "lang"=>"en", "lat"=>40.151128, "lon"=>-82.687659),
"/US/US-OH/junction-city-16197" => array("type"=>"village", "pop"=>841, "name_ru"=>"Джанкшен Сити", "name_en"=>"Junction City", "lang"=>"en", "lat"=>39.721509, "lon"=>-82.300218),
"/US/US-OH/kalida-16198" => array("type"=>"village", "pop"=>1254, "name_en"=>"Kalida", "lang"=>"en", "lat"=>40.985866, "lon"=>-84.195477),
"/US/US-OH/kelleys-island-16199" => array("type"=>"village", "pop"=>386, "name_en"=>"Kelleys Island", "lang"=>"en", "lat"=>41.602757, "lon"=>-82.705971),
"/US/US-OH/kent-16200" => array("type"=>"city", "pop"=>27983, "name_ru"=>"Кент", "name_en"=>"Kent", "lang"=>"en", "lat"=>41.147067, "lon"=>-81.362487),
"/US/US-OH/kenton-16201" => array("type"=>"city", "pop"=>8052, "name_ru"=>"Кентон", "name_en"=>"Kenton", "lang"=>"en", "lat"=>40.644493, "lon"=>-83.610448),
"/US/US-OH/kenwood-16202" => array("type"=>"cdp", "pop"=>7469, "name_en"=>"Kenwood", "lang"=>"en", "lat"=>39.205599, "lon"=>-84.374132),
"/US/US-OH/kettering-16203" => array("type"=>"city", "pop"=>53708, "name_ru"=>"Кеттеринг", "name_en"=>"Kettering", "lang"=>"en", "lat"=>39.695852, "lon"=>-84.150076),
"/US/US-OH/kettlersville-16204" => array("type"=>"village", "pop"=>166, "name_en"=>"Kettlersville", "lang"=>"en", "lat"=>40.438764, "lon"=>-84.261008),
"/US/US-OH/killbuck-16205" => array("type"=>"village", "pop"=>887, "name_en"=>"Killbuck", "lang"=>"en", "lat"=>40.497171, "lon"=>-81.983613),
"/US/US-OH/kimbolton-16206" => array("type"=>"village", "pop"=>134, "name_en"=>"Kimbolton", "lang"=>"en", "lat"=>40.154021, "lon"=>-81.574315),
"/US/US-OH/kingston-16207" => array("type"=>"village", "pop"=>1037, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>39.472180, "lon"=>-82.911816),
"/US/US-OH/kipton-16208" => array("type"=>"village", "pop"=>245, "name_en"=>"Kipton", "lang"=>"en", "lat"=>41.267000, "lon"=>-82.304824),
"/US/US-OH/kirby-16209" => array("type"=>"village", "pop"=>123, "name_en"=>"Kirby", "lang"=>"en", "lat"=>40.813398, "lon"=>-83.418702),
"/US/US-OH/kirkersville-16210" => array("type"=>"village", "pop"=>544, "name_en"=>"Kirkersville", "lang"=>"en", "lat"=>39.950267, "lon"=>-82.597965),
"/US/US-OH/kirtland-16211" => array("type"=>"city", "pop"=>7362, "name_en"=>"Kirtland", "lang"=>"en", "lat"=>41.598450, "lon"=>-81.337990),
"/US/US-OH/kirtland-hills-16212" => array("type"=>"village", "pop"=>787, "name_en"=>"Kirtland Hills", "lang"=>"en", "lat"=>41.642484, "lon"=>-81.318051),
"/US/US-OH/la-croft-16213" => array("type"=>"cdp", "pop"=>1427, "name_en"=>"La Croft", "lang"=>"en", "lat"=>40.647162, "lon"=>-80.601278),
"/US/US-OH/la-rue-16223" => array("type"=>"village", "pop"=>703, "name_en"=>"La Rue", "lang"=>"en", "lat"=>40.577879, "lon"=>-83.382481),
"/US/US-OH/lafayette-16214" => array("type"=>"village", "pop"=>292, "name_ru"=>"Ла Файет", "name_en"=>"Lafayette", "lang"=>"en", "lat"=>40.758592, "lon"=>-83.949546),
"/US/US-OH/lagrange-16215" => array("type"=>"village", "pop"=>2012, "name_ru"=>"Лаграндж", "name_en"=>"Lagrange", "lang"=>"en", "lat"=>41.238850, "lon"=>-82.120499),
"/US/US-OH/lake-darby-16216" => array("type"=>"cdp", "pop"=>2798, "name_en"=>"Lake Darby", "lang"=>"en", "lat"=>39.957250, "lon"=>-83.228908),
"/US/US-OH/lakeline-16217" => array("type"=>"village", "pop"=>162, "name_en"=>"Lakeline", "lang"=>"en", "lat"=>41.658600, "lon"=>-81.453982),
"/US/US-OH/lakemore-16218" => array("type"=>"village", "pop"=>2857, "name_en"=>"Lakemore", "lang"=>"en", "lat"=>41.021346, "lon"=>-81.427680),
"/US/US-OH/lakeview-16219" => array("type"=>"village", "pop"=>1065, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>40.487782, "lon"=>-83.925722),
"/US/US-OH/lakewood-16220" => array("type"=>"city", "pop"=>50704, "name_ru"=>"Лейквуд", "name_en"=>"Lakewood", "lang"=>"en", "lat"=>41.483769, "lon"=>-81.801538),
"/US/US-OH/lancaster-16221" => array("type"=>"city", "pop"=>36976, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>39.722323, "lon"=>-82.599186),
"/US/US-OH/landen-16222" => array("type"=>"cdp", "pop"=>9263, "name_en"=>"Landen", "lang"=>"en", "lat"=>39.307909, "lon"=>-84.279697),
"/US/US-OH/latty-16224" => array("type"=>"village", "pop"=>184, "name_en"=>"Latty", "lang"=>"en", "lat"=>41.088095, "lon"=>-84.583369),
"/US/US-OH/laura-16225" => array("type"=>"village", "pop"=>497, "name_en"=>"Laura", "lang"=>"en", "lat"=>39.993608, "lon"=>-84.409103),
"/US/US-OH/laurelville-16226" => array("type"=>"village", "pop"=>547, "name_en"=>"Laurelville", "lang"=>"en", "lat"=>39.471460, "lon"=>-82.737485),
"/US/US-OH/lawrenceville-16227" => array("type"=>"village", "pop"=>304, "name_ru"=>"Лоренсвилл", "name_en"=>"Lawrenceville", "lang"=>"en", "lat"=>39.985033, "lon"=>-83.874807),
"/US/US-OH/lebanon-16228" => array("type"=>"city", "pop"=>20535, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>39.426665, "lon"=>-84.212695),
"/US/US-OH/leesburg-16229" => array("type"=>"village", "pop"=>1304, "name_ru"=>"Лисберг", "name_en"=>"Leesburg", "lang"=>"en", "lat"=>39.345799, "lon"=>-83.554621),
"/US/US-OH/leesville-16230" => array("type"=>"village", "pop"=>178, "name_ru"=>"Лисвилл", "name_en"=>"Leesville", "lang"=>"en", "lat"=>40.450948, "lon"=>-81.209563),
"/US/US-OH/leetonia-16231" => array("type"=>"village", "pop"=>1996, "name_en"=>"Leetonia", "lang"=>"en", "lat"=>40.878392, "lon"=>-80.760972),
"/US/US-OH/leipsic-16232" => array("type"=>"village", "pop"=>2173, "name_ru"=>"Лейпсик", "name_en"=>"Leipsic", "lang"=>"en", "lat"=>41.099327, "lon"=>-83.984523),
"/US/US-OH/lewisburg-16233" => array("type"=>"village", "pop"=>1726, "name_ru"=>"Льюисбург", "name_en"=>"Lewisburg", "lang"=>"en", "lat"=>39.849048, "lon"=>-84.540607),
"/US/US-OH/lewisville-16234" => array("type"=>"village", "pop"=>224, "name_ru"=>"Льюисвилл", "name_en"=>"Lewisville", "lang"=>"en", "lat"=>39.765880, "lon"=>-81.218847),
"/US/US-OH/lexington-16235" => array("type"=>"village", "pop"=>4092, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>40.678217, "lon"=>-82.586895),
"/US/US-OH/liberty-center-16236" => array("type"=>"village", "pop"=>1080, "name_en"=>"Liberty Center", "lang"=>"en", "lat"=>41.443439, "lon"=>-84.008106),
"/US/US-OH/lima-16237" => array("type"=>"city", "pop"=>37829, "name_ru"=>"Лима", "name_en"=>"Lima", "lang"=>"en", "lat"=>40.743250, "lon"=>-84.111399),
"/US/US-OH/limaville-16238" => array("type"=>"village", "pop"=>183, "name_en"=>"Limaville", "lang"=>"en", "lat"=>40.984150, "lon"=>-81.148099),
"/US/US-OH/lincoln-heights-16239" => array("type"=>"city", "pop"=>3911, "name_en"=>"Lincoln Heights", "lang"=>"en", "lat"=>39.243901, "lon"=>-84.458065),
"/US/US-OH/lincoln-village-16240" => array("type"=>"cdp", "pop"=>9958, "name_ru"=>"Линкольн Виллидж", "name_en"=>"Lincoln Village", "lang"=>"en", "lat"=>39.954700, "lon"=>-83.130920),
"/US/US-OH/lindsey-16241" => array("type"=>"village", "pop"=>473, "name_en"=>"Lindsey", "lang"=>"en", "lat"=>41.420576, "lon"=>-83.220744),
"/US/US-OH/linndale-16242" => array("type"=>"village", "pop"=>89, "name_en"=>"Linndale", "lang"=>"en", "lat"=>41.445100, "lon"=>-81.766523),
"/US/US-OH/lisbon-16243" => array("type"=>"village", "pop"=>2985, "name_ru"=>"Лисбон", "name_en"=>"Lisbon", "lang"=>"en", "lat"=>40.776085, "lon"=>-80.769334),
"/US/US-OH/lithopolis-16244" => array("type"=>"village", "pop"=>1070, "name_en"=>"Lithopolis", "lang"=>"en", "lat"=>39.804968, "lon"=>-82.811229),
"/US/US-OH/lockbourne-16245" => array("type"=>"village", "pop"=>268, "name_en"=>"Lockbourne", "lang"=>"en", "lat"=>39.808952, "lon"=>-82.970495),
"/US/US-OH/lockington-16246" => array("type"=>"village", "pop"=>198, "name_en"=>"Lockington", "lang"=>"en", "lat"=>40.207752, "lon"=>-84.235502),
"/US/US-OH/lockland-16247" => array("type"=>"village", "pop"=>3473, "name_en"=>"Lockland", "lang"=>"en", "lat"=>39.227551, "lon"=>-84.457400),
"/US/US-OH/lodi-16248" => array("type"=>"village", "pop"=>3375, "name_ru"=>"Лоди", "name_en"=>"Lodi", "lang"=>"en", "lat"=>41.033656, "lon"=>-82.013125),
"/US/US-OH/logan-16249" => array("type"=>"city", "pop"=>7436, "name_ru"=>"Логан", "name_en"=>"Logan", "lang"=>"en", "lat"=>39.536895, "lon"=>-82.401745),
"/US/US-OH/logan-elm-village-16250" => array("type"=>"cdp", "pop"=>1287, "name_en"=>"Logan Elm Village", "lang"=>"en", "lat"=>39.571691, "lon"=>-82.947444),
"/US/US-OH/london-16251" => array("type"=>"city", "pop"=>9587, "name_ru"=>"Лондон", "name_en"=>"London", "lang"=>"en", "lat"=>39.889197, "lon"=>-83.440889),
"/US/US-OH/lorain-16252" => array("type"=>"city", "pop"=>70239, "name_en"=>"Lorain", "lang"=>"en", "lat"=>41.441550, "lon"=>-82.183399),
"/US/US-OH/lordstown-16253" => array("type"=>"village", "pop"=>3522, "name_en"=>"Lordstown", "lang"=>"en", "lat"=>41.165360, "lon"=>-80.854397),
"/US/US-OH/lore-city-16254" => array("type"=>"village", "pop"=>295, "name_en"=>"Lore City", "lang"=>"en", "lat"=>39.983960, "lon"=>-81.459595),
"/US/US-OH/loudonville-16255" => array("type"=>"village", "pop"=>3069, "name_en"=>"Loudonville", "lang"=>"en", "lat"=>40.634679, "lon"=>-82.233978),
"/US/US-OH/louisville-16256" => array("type"=>"city", "pop"=>9459, "name_ru"=>"Луисвилл", "name_en"=>"Louisville", "lang"=>"en", "lat"=>40.837250, "lon"=>-81.260899),
"/US/US-OH/loveland-16257" => array("type"=>"city", "pop"=>11734, "name_ru"=>"Лавленд", "name_en"=>"Loveland", "lang"=>"en", "lat"=>39.267251, "lon"=>-84.274100),
"/US/US-OH/loveland-park-16258" => array("type"=>"cdp", "pop"=>1357, "name_en"=>"Loveland Park", "lang"=>"en", "lat"=>39.298423, "lon"=>-84.262478),
"/US/US-OH/lowell-16259" => array("type"=>"village", "pop"=>593, "name_ru"=>"Лоуэлл", "name_en"=>"Lowell", "lang"=>"en", "lat"=>39.529101, "lon"=>-81.507524),
"/US/US-OH/lowellville-16260" => array("type"=>"village", "pop"=>1146, "name_en"=>"Lowellville", "lang"=>"en", "lat"=>41.040150, "lon"=>-80.548140),
"/US/US-OH/lower-salem-16261" => array("type"=>"village", "pop"=>102, "name_en"=>"Lower Salem", "lang"=>"en", "lat"=>39.563541, "lon"=>-81.394592),
"/US/US-OH/lucas-16262" => array("type"=>"village", "pop"=>589, "name_ru"=>"Лукас", "name_en"=>"Lucas", "lang"=>"en", "lat"=>40.702880, "lon"=>-82.421237),
"/US/US-OH/lucasville-16263" => array("type"=>"cdp", "pop"=>1575, "name_en"=>"Lucasville", "lang"=>"en", "lat"=>38.878979, "lon"=>-82.995277),
"/US/US-OH/luckey-16264" => array("type"=>"village", "pop"=>989, "name_en"=>"Luckey", "lang"=>"en", "lat"=>41.452337, "lon"=>-83.483685),
"/US/US-OH/ludlow-falls-16265" => array("type"=>"village", "pop"=>210, "name_en"=>"Ludlow Falls", "lang"=>"en", "lat"=>39.998440, "lon"=>-84.339065),
"/US/US-OH/lynchburg-16266" => array("type"=>"village", "pop"=>1393, "name_ru"=>"Линчбург", "name_en"=>"Lynchburg", "lang"=>"en", "lat"=>39.244272, "lon"=>-83.788849),
"/US/US-OH/lyndhurst-16267" => array("type"=>"city", "pop"=>13848, "name_en"=>"Lyndhurst", "lang"=>"en", "lat"=>41.515500, "lon"=>-81.492637),
"/US/US-OH/lyons-16268" => array("type"=>"village", "pop"=>586, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>41.699633, "lon"=>-84.073025),
"/US/US-OH/macedonia-16274" => array("type"=>"city", "pop"=>10435, "name_ru"=>"Македония", "name_en"=>"Macedonia", "lang"=>"en", "lat"=>41.310983, "lon"=>-81.497334),
"/US/US-OH/mack-north-16276" => array("type"=>"cdp", "pop"=>2816, "name_en"=>"Mack North", "lang"=>"en", "lat"=>39.166828, "lon"=>-84.673156),
"/US/US-OH/mack-south-16278" => array("type"=>"cdp", "pop"=>5767, "name_en"=>"Mack South", "lang"=>"en", "lat"=>39.142051, "lon"=>-84.673211),
"/US/US-OH/macksburg-16277" => array("type"=>"village", "pop"=>200, "name_ru"=>"Максбург", "name_en"=>"Macksburg", "lang"=>"en", "lat"=>39.631169, "lon"=>-81.456932),
"/US/US-OH/madeira-16279" => array("type"=>"city", "pop"=>8558, "name_en"=>"Madeira", "lang"=>"en", "lat"=>39.184201, "lon"=>-84.374822),
"/US/US-OH/madison-16280" => array("type"=>"village", "pop"=>3104, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>41.771691, "lon"=>-81.053170),
"/US/US-OH/magnetic-springs-16281" => array("type"=>"village", "pop"=>314, "name_en"=>"Magnetic Springs", "lang"=>"en", "lat"=>40.353693, "lon"=>-83.262815),
"/US/US-OH/magnolia-16282" => array("type"=>"village", "pop"=>924, "name_ru"=>"Магнолия", "name_en"=>"Magnolia", "lang"=>"en", "lat"=>40.651510, "lon"=>-81.292493),
"/US/US-OH/maineville-16283" => array("type"=>"village", "pop"=>1033, "name_en"=>"Maineville", "lang"=>"en", "lat"=>39.315779, "lon"=>-84.223091),
"/US/US-OH/malinta-16284" => array("type"=>"village", "pop"=>271, "name_en"=>"Malinta", "lang"=>"en", "lat"=>41.319612, "lon"=>-84.037086),
"/US/US-OH/malta-16285" => array("type"=>"village", "pop"=>651, "name_ru"=>"Малта", "name_en"=>"Malta", "lang"=>"en", "lat"=>39.650439, "lon"=>-81.864563),
"/US/US-OH/malvern-16286" => array("type"=>"village", "pop"=>1193, "name_ru"=>"Малверн", "name_en"=>"Malvern", "lang"=>"en", "lat"=>40.688987, "lon"=>-81.181836),
"/US/US-OH/manchester-16287" => array("type"=>"village", "pop"=>2096, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>38.691397, "lon"=>-83.605214),
"/US/US-OH/mansfield-16288" => array("type"=>"city", "pop"=>49579, "name_ru"=>"Мансфилд", "name_en"=>"Mansfield", "lang"=>"en", "lat"=>40.762838, "lon"=>-82.526894),
"/US/US-OH/mantua-16289" => array("type"=>"village", "pop"=>1017, "name_en"=>"Mantua", "lang"=>"en", "lat"=>41.281450, "lon"=>-81.222699),
"/US/US-OH/maple-heights-16290" => array("type"=>"city", "pop"=>23667, "name_en"=>"Maple Heights", "lang"=>"en", "lat"=>41.408850, "lon"=>-81.562835),
"/US/US-OH/maple-ridge-16291" => array("type"=>"cdp", "pop"=>1018, "name_en"=>"Maple Ridge", "lang"=>"en", "lat"=>40.916345, "lon"=>-81.049969),
"/US/US-OH/marble-cliff-16292" => array("type"=>"village", "pop"=>650, "name_en"=>"Marble Cliff", "lang"=>"en", "lat"=>39.985150, "lon"=>-83.060496),
"/US/US-OH/marblehead-16293" => array("type"=>"village", "pop"=>831, "name_ru"=>"Марблхед", "name_en"=>"Marblehead", "lang"=>"en", "lat"=>41.531770, "lon"=>-82.721536),
"/US/US-OH/marengo-16294" => array("type"=>"village", "pop"=>304, "name_ru"=>"Маренго", "name_en"=>"Marengo", "lang"=>"en", "lat"=>40.401280, "lon"=>-82.810141),
"/US/US-OH/mariemont-16295" => array("type"=>"village", "pop"=>3183, "name_en"=>"Mariemont", "lang"=>"en", "lat"=>39.142051, "lon"=>-84.379041),
"/US/US-OH/marietta-16296" => array("type"=>"city", "pop"=>14307, "name_ru"=>"Мариетта", "name_en"=>"Marietta", "lang"=>"en", "lat"=>39.421388, "lon"=>-81.446872),
"/US/US-OH/marion-16297" => array("type"=>"city", "pop"=>35841, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>40.584931, "lon"=>-83.127921),
"/US/US-OH/marseilles-16298" => array("type"=>"village", "pop"=>117, "name_ru"=>"Марсейллес", "name_en"=>"Marseilles", "lang"=>"en", "lat"=>40.701152, "lon"=>-83.392750),
"/US/US-OH/marshallville-16299" => array("type"=>"village", "pop"=>810, "name_ru"=>"Маршалвилл", "name_en"=>"Marshallville", "lang"=>"en", "lat"=>40.900352, "lon"=>-81.733294),
"/US/US-OH/martins-ferry-16301" => array("type"=>"city", "pop"=>6654, "name_en"=>"Martins Ferry", "lang"=>"en", "lat"=>40.101333, "lon"=>-80.723865),
"/US/US-OH/martinsburg-16300" => array("type"=>"village", "pop"=>186, "name_ru"=>"Мартинсбург", "name_en"=>"Martinsburg", "lang"=>"en", "lat"=>40.269298, "lon"=>-82.354934),
"/US/US-OH/martinsville-16302" => array("type"=>"village", "pop"=>434, "name_ru"=>"Мартинсвилл", "name_en"=>"Martinsville", "lang"=>"en", "lat"=>39.322981, "lon"=>-83.811078),
"/US/US-OH/marysville-16303" => array("type"=>"city", "pop"=>18212, "name_ru"=>"Мэрисвилл", "name_en"=>"Marysville", "lang"=>"en", "lat"=>40.237650, "lon"=>-83.374685),
"/US/US-OH/mason-16304" => array("type"=>"city", "pop"=>29682, "name_ru"=>"Мейсон", "name_en"=>"Mason", "lang"=>"en", "lat"=>39.360001, "lon"=>-84.307037),
"/US/US-OH/massillon-16305" => array("type"=>"city", "pop"=>32613, "name_en"=>"Massillon", "lang"=>"en", "lat"=>40.786650, "lon"=>-81.520073),
"/US/US-OH/matamoras-16306" => array("type"=>"village", "pop"=>907, "name_en"=>"Matamoras", "lang"=>"en", "lat"=>39.518631, "lon"=>-81.070467),
"/US/US-OH/maumee-16307" => array("type"=>"city", "pop"=>13856, "name_en"=>"Maumee", "lang"=>"en", "lat"=>41.574233, "lon"=>-83.653438),
"/US/US-OH/mayfield-16308" => array("type"=>"village", "pop"=>3121, "name_ru"=>"Мейфилд", "name_en"=>"Mayfield", "lang"=>"en", "lat"=>41.553500, "lon"=>-81.435306),
"/US/US-OH/mayfield-heights-16309" => array("type"=>"city", "pop"=>17737, "name_en"=>"Mayfield Heights", "lang"=>"en", "lat"=>41.517350, "lon"=>-81.453999),
"/US/US-OH/mcarthur-16269" => array("type"=>"village", "pop"=>2022, "name_en"=>"McArthur", "lang"=>"en", "lat"=>39.247007, "lon"=>-82.478422),
"/US/US-OH/mcclure-16270" => array("type"=>"village", "pop"=>713, "name_en"=>"McClure", "lang"=>"en", "lat"=>41.370543, "lon"=>-83.941989),
"/US/US-OH/mccomb-16271" => array("type"=>"village", "pop"=>1599, "name_ru"=>"Мак-Комб", "name_en"=>"McComb", "lang"=>"en", "lat"=>41.106657, "lon"=>-83.789576),
"/US/US-OH/mcconnelsville-16272" => array("type"=>"village", "pop"=>1562, "name_en"=>"McConnelsville", "lang"=>"en", "lat"=>39.654097, "lon"=>-81.844417),
"/US/US-OH/mcdonald-16273" => array("type"=>"village", "pop"=>3214, "name_ru"=>"Макдональд", "name_en"=>"McDonald", "lang"=>"en", "lat"=>41.164224, "lon"=>-80.723807),
"/US/US-OH/mcguffey-16275" => array("type"=>"village", "pop"=>517, "name_en"=>"McGuffey", "lang"=>"en", "lat"=>40.693300, "lon"=>-83.785946),
"/US/US-OH/mechanicsburg-16310" => array("type"=>"village", "pop"=>1699, "name_ru"=>"Меканиксбург", "name_en"=>"Mechanicsburg", "lang"=>"en", "lat"=>40.074424, "lon"=>-83.557835),
"/US/US-OH/medina-16311" => array("type"=>"city", "pop"=>26011, "name_ru"=>"Медина", "name_en"=>"Medina", "lang"=>"en", "lat"=>41.136192, "lon"=>-81.867374),
"/US/US-OH/melrose-16312" => array("type"=>"village", "pop"=>298, "name_ru"=>"Мелроуз", "name_en"=>"Melrose", "lang"=>"en", "lat"=>41.089625, "lon"=>-84.419523),
"/US/US-OH/mendon-16313" => array("type"=>"village", "pop"=>683, "name_ru"=>"Мендон", "name_en"=>"Mendon", "lang"=>"en", "lat"=>40.671258, "lon"=>-84.517987),
"/US/US-OH/mentor-16314" => array("type"=>"city", "pop"=>51825, "name_ru"=>"Ментор", "name_en"=>"Mentor", "lang"=>"en", "lat"=>41.691526, "lon"=>-81.334452),
"/US/US-OH/mentor-on-the-lake-16315" => array("type"=>"city", "pop"=>8288, "name_en"=>"Mentor-on-the-Lake", "lang"=>"en", "lat"=>41.713702, "lon"=>-81.364872),
"/US/US-OH/metamora-16316" => array("type"=>"village", "pop"=>591, "name_ru"=>"Метамора", "name_en"=>"Metamora", "lang"=>"en", "lat"=>41.711096, "lon"=>-83.908036),
"/US/US-OH/meyers-lake-16317" => array("type"=>"village", "pop"=>556, "name_en"=>"Meyers Lake", "lang"=>"en", "lat"=>40.813850, "lon"=>-81.418162),
"/US/US-OH/miamisburg-16318" => array("type"=>"city", "pop"=>19793, "name_en"=>"Miamisburg", "lang"=>"en", "lat"=>39.635550, "lon"=>-84.274545),
"/US/US-OH/middle-point-16321" => array("type"=>"village", "pop"=>564, "name_en"=>"Middle Point", "lang"=>"en", "lat"=>40.855956, "lon"=>-84.446814),
"/US/US-OH/middleburg-heights-16319" => array("type"=>"city", "pop"=>14928, "name_en"=>"Middleburg Heights", "lang"=>"en", "lat"=>41.367650, "lon"=>-81.814884),
"/US/US-OH/middlefield-16320" => array("type"=>"village", "pop"=>2378, "name_en"=>"Middlefield", "lang"=>"en", "lat"=>41.459120, "lon"=>-81.075088),
"/US/US-OH/middleport-16322" => array("type"=>"village", "pop"=>2469, "name_en"=>"Middleport", "lang"=>"en", "lat"=>38.999829, "lon"=>-82.056858),
"/US/US-OH/middletown-16323" => array("type"=>"city", "pop"=>51422, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>39.505700, "lon"=>-84.373579),
"/US/US-OH/midland-16324" => array("type"=>"village", "pop"=>262, "name_ru"=>"Мидленд", "name_en"=>"Midland", "lang"=>"en", "lat"=>39.305853, "lon"=>-83.910388),
"/US/US-OH/midvale-16325" => array("type"=>"village", "pop"=>583, "name_ru"=>"Мидвейл", "name_en"=>"Midvale", "lang"=>"en", "lat"=>40.436983, "lon"=>-81.372140),
"/US/US-OH/midway-16326" => array("type"=>"village", "pop"=>262, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>39.732720, "lon"=>-83.476264),
"/US/US-OH/mifflin-16327" => array("type"=>"village", "pop"=>144, "name_en"=>"Mifflin", "lang"=>"en", "lat"=>40.774386, "lon"=>-82.364278),
"/US/US-OH/milan-16328" => array("type"=>"village", "pop"=>1345, "name_ru"=>"Милан", "name_en"=>"Milan", "lang"=>"en", "lat"=>41.288845, "lon"=>-82.601149),
"/US/US-OH/milford-16329" => array("type"=>"city", "pop"=>6332, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>39.170850, "lon"=>-84.281632),
"/US/US-OH/milford-center-16330" => array("type"=>"village", "pop"=>696, "name_en"=>"Milford Center", "lang"=>"en", "lat"=>40.179112, "lon"=>-83.437082),
"/US/US-OH/millbury-16331" => array("type"=>"village", "pop"=>1161, "name_en"=>"Millbury", "lang"=>"en", "lat"=>41.564050, "lon"=>-83.425684),
"/US/US-OH/milledgeville-16332" => array("type"=>"village", "pop"=>115, "name_ru"=>"Миледжвилл", "name_en"=>"Milledgeville", "lang"=>"en", "lat"=>39.593564, "lon"=>-83.587795),
"/US/US-OH/miller-city-16333" => array("type"=>"village", "pop"=>126, "name_en"=>"Miller City", "lang"=>"en", "lat"=>41.101936, "lon"=>-84.131969),
"/US/US-OH/millersburg-16334" => array("type"=>"village", "pop"=>3611, "name_ru"=>"Миллерсберг", "name_en"=>"Millersburg", "lang"=>"en", "lat"=>40.553410, "lon"=>-81.915643),
"/US/US-OH/millersport-16335" => array("type"=>"village", "pop"=>927, "name_en"=>"Millersport", "lang"=>"en", "lat"=>39.896272, "lon"=>-82.542068),
"/US/US-OH/millville-16336" => array("type"=>"village", "pop"=>891, "name_ru"=>"Миллвилл", "name_en"=>"Millville", "lang"=>"en", "lat"=>39.391850, "lon"=>-84.653536),
"/US/US-OH/milton-center-16337" => array("type"=>"village", "pop"=>195, "name_en"=>"Milton Center", "lang"=>"en", "lat"=>41.301408, "lon"=>-83.829531),
"/US/US-OH/miltonsburg-16338" => array("type"=>"village", "pop"=>28, "name_en"=>"Miltonsburg", "lang"=>"en", "lat"=>39.831816, "lon"=>-81.164348),
"/US/US-OH/mineral-city-16339" => array("type"=>"village", "pop"=>841, "name_en"=>"Mineral City", "lang"=>"en", "lat"=>40.602361, "lon"=>-81.361489),
"/US/US-OH/mineral-ridge-16340" => array("type"=>"cdp", "pop"=>3928, "name_en"=>"Mineral Ridge", "lang"=>"en", "lat"=>41.139080, "lon"=>-80.767259),
"/US/US-OH/minerva-16341" => array("type"=>"village", "pop"=>3906, "name_en"=>"Minerva", "lang"=>"en", "lat"=>40.729850, "lon"=>-81.102415),
"/US/US-OH/minerva-park-16342" => array("type"=>"village", "pop"=>1205, "name_en"=>"Minerva Park", "lang"=>"en", "lat"=>40.077624, "lon"=>-82.943558),
"/US/US-OH/mingo-junction-16343" => array("type"=>"city", "pop"=>3320, "name_en"=>"Mingo Junction", "lang"=>"en", "lat"=>40.320600, "lon"=>-80.613573),
"/US/US-OH/minster-16344" => array("type"=>"village", "pop"=>2781, "name_en"=>"Minster", "lang"=>"en", "lat"=>40.394323, "lon"=>-84.379498),
"/US/US-OH/mogadore-16345" => array("type"=>"village", "pop"=>3912, "name_en"=>"Mogadore", "lang"=>"en", "lat"=>41.050350, "lon"=>-81.400626),
"/US/US-OH/monfort-heights-east-16346" => array("type"=>"cdp", "pop"=>3661, "name_en"=>"Monfort Heights East", "lang"=>"en", "lat"=>39.182151, "lon"=>-84.583915),
"/US/US-OH/monfort-heights-south-16347" => array("type"=>"cdp", "pop"=>4587, "name_en"=>"Monfort Heights South", "lang"=>"en", "lat"=>39.173251, "lon"=>-84.606362),
"/US/US-OH/monroe-16348" => array("type"=>"village", "pop"=>12178, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>39.446100, "lon"=>-84.348112),
"/US/US-OH/monroeville-16349" => array("type"=>"village", "pop"=>1345, "name_ru"=>"Монровилл", "name_en"=>"Monroeville", "lang"=>"en", "lat"=>41.244005, "lon"=>-82.701200),
"/US/US-OH/montezuma-16350" => array("type"=>"village", "pop"=>185, "name_ru"=>"Мотезума", "name_en"=>"Montezuma", "lang"=>"en", "lat"=>40.489342, "lon"=>-84.548949),
"/US/US-OH/montgomery-16351" => array("type"=>"city", "pop"=>10458, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>39.247301, "lon"=>-84.347618),
"/US/US-OH/montpelier-16352" => array("type"=>"village", "pop"=>4015, "name_ru"=>"Монтпильер", "name_en"=>"Montpelier", "lang"=>"en", "lat"=>41.581619, "lon"=>-84.600539),
"/US/US-OH/montrose-ghent-16353" => array("type"=>"cdp", "pop"=>4906, "name_en"=>"Montrose-Ghent", "lang"=>"en", "lat"=>41.157301, "lon"=>-81.642462),
"/US/US-OH/moraine-16354" => array("type"=>"city", "pop"=>6467, "name_en"=>"Moraine", "lang"=>"en", "lat"=>39.703900, "lon"=>-84.228390),
"/US/US-OH/moreland-hills-16355" => array("type"=>"village", "pop"=>3064, "name_en"=>"Moreland Hills", "lang"=>"en", "lat"=>41.444500, "lon"=>-81.431068),
"/US/US-OH/morral-16356" => array("type"=>"village", "pop"=>368, "name_en"=>"Morral", "lang"=>"en", "lat"=>40.689927, "lon"=>-83.213062),
"/US/US-OH/morristown-16357" => array("type"=>"village", "pop"=>292, "name_ru"=>"Морристаун", "name_en"=>"Morristown", "lang"=>"en", "lat"=>40.062802, "lon"=>-81.071457),
"/US/US-OH/morrow-16358" => array("type"=>"village", "pop"=>1544, "name_ru"=>"Морроу", "name_en"=>"Morrow", "lang"=>"en", "lat"=>39.349141, "lon"=>-84.126618),
"/US/US-OH/moscow-16359" => array("type"=>"village", "pop"=>244, "name_ru"=>"Москва", "name_en"=>"Moscow", "lang"=>"en", "lat"=>38.860347, "lon"=>-84.228487),
"/US/US-OH/mount-blanchard-16360" => array("type"=>"village", "pop"=>425, "name_en"=>"Mount Blanchard", "lang"=>"en", "lat"=>40.898526, "lon"=>-83.557223),
"/US/US-OH/mount-carmel-16361" => array("type"=>"cdp", "pop"=>4462, "name_ru"=>"Маунт Кармел", "name_en"=>"Mount Carmel", "lang"=>"en", "lat"=>39.096651, "lon"=>-84.298427),
"/US/US-OH/mount-cory-16362" => array("type"=>"village", "pop"=>187, "name_en"=>"Mount Cory", "lang"=>"en", "lat"=>40.934815, "lon"=>-83.824313),
"/US/US-OH/mount-eaton-16363" => array("type"=>"village", "pop"=>240, "name_en"=>"Mount Eaton", "lang"=>"en", "lat"=>40.694982, "lon"=>-81.702825),
"/US/US-OH/mount-gilead-16364" => array("type"=>"village", "pop"=>3581, "name_en"=>"Mount Gilead", "lang"=>"en", "lat"=>40.550847, "lon"=>-82.835976),
"/US/US-OH/mount-healthy-16365" => array("type"=>"city", "pop"=>6070, "name_en"=>"Mount Healthy", "lang"=>"en", "lat"=>39.232751, "lon"=>-84.547217),
"/US/US-OH/mount-healthy-heights-16366" => array("type"=>"cdp", "pop"=>3863, "name_en"=>"Mount Healthy Heights", "lang"=>"en", "lat"=>39.270251, "lon"=>-84.568125),
"/US/US-OH/mount-orab-16367" => array("type"=>"village", "pop"=>2742, "name_en"=>"Mount Orab", "lang"=>"en", "lat"=>39.029277, "lon"=>-83.923739),
"/US/US-OH/mount-pleasant-16368" => array("type"=>"village", "pop"=>496, "name_ru"=>"Маунт Плезант", "name_en"=>"Mount Pleasant", "lang"=>"en", "lat"=>40.175880, "lon"=>-80.799952),
"/US/US-OH/mount-repose-16369" => array("type"=>"cdp", "pop"=>3093, "name_en"=>"Mount Repose", "lang"=>"en", "lat"=>39.185449, "lon"=>-84.224595),
"/US/US-OH/mount-sterling-16370" => array("type"=>"village", "pop"=>1835, "name_ru"=>"Маунт Стерлинг", "name_en"=>"Mount Sterling", "lang"=>"en", "lat"=>39.719467, "lon"=>-83.268085),
"/US/US-OH/mount-vernon-16371" => array("type"=>"city", "pop"=>16037, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>40.390482, "lon"=>-82.475153),
"/US/US-OH/mount-victory-16372" => array("type"=>"village", "pop"=>592, "name_en"=>"Mount Victory", "lang"=>"en", "lat"=>40.534109, "lon"=>-83.520230),
"/US/US-OH/mowrystown-16373" => array("type"=>"village", "pop"=>380, "name_en"=>"Mowrystown", "lang"=>"en", "lat"=>39.038761, "lon"=>-83.752533),
"/US/US-OH/mulberry-16374" => array("type"=>"cdp", "pop"=>2856, "name_ru"=>"Малберри", "name_en"=>"Mulberry", "lang"=>"en", "lat"=>39.197239, "lon"=>-84.248932),
"/US/US-OH/munroe-falls-16375" => array("type"=>"village", "pop"=>5165, "name_en"=>"Munroe Falls", "lang"=>"en", "lat"=>41.139700, "lon"=>-81.436259),
"/US/US-OH/murray-city-16376" => array("type"=>"village", "pop"=>462, "name_en"=>"Murray City", "lang"=>"en", "lat"=>39.510762, "lon"=>-82.167071),
"/US/US-OH/mutual-16377" => array("type"=>"village", "pop"=>129, "name_en"=>"Mutual", "lang"=>"en", "lat"=>40.079064, "lon"=>-83.637349),
"/US/US-OH/napoleon-16378" => array("type"=>"city", "pop"=>8817, "name_ru"=>"Наполеон", "name_en"=>"Napoleon", "lang"=>"en", "lat"=>41.394762, "lon"=>-84.126592),
"/US/US-OH/nashville-16379" => array("type"=>"village", "pop"=>183, "name_ru"=>"Нэшвилл", "name_en"=>"Nashville", "lang"=>"en", "lat"=>40.595957, "lon"=>-82.113300),
"/US/US-OH/navarre-16380" => array("type"=>"village", "pop"=>1896, "name_en"=>"Navarre", "lang"=>"en", "lat"=>40.722634, "lon"=>-81.522710),
"/US/US-OH/neffs-16381" => array("type"=>"cdp", "pop"=>1213, "name_en"=>"Neffs", "lang"=>"en", "lat"=>40.029151, "lon"=>-80.815553),
"/US/US-OH/nellie-16382" => array("type"=>"village", "pop"=>131, "name_en"=>"Nellie", "lang"=>"en", "lat"=>40.338297, "lon"=>-82.068432),
"/US/US-OH/nelsonville-16383" => array("type"=>"city", "pop"=>5375, "name_en"=>"Nelsonville", "lang"=>"en", "lat"=>39.458758, "lon"=>-82.226023),
"/US/US-OH/nevada-16384" => array("type"=>"village", "pop"=>763, "name_ru"=>"Невада", "name_en"=>"Nevada", "lang"=>"en", "lat"=>40.818596, "lon"=>-83.131370),
"/US/US-OH/neville-16385" => array("type"=>"village", "pop"=>140, "name_en"=>"Neville", "lang"=>"en", "lat"=>38.813213, "lon"=>-84.210790),
"/US/US-OH/new-albany-16386" => array("type"=>"village", "pop"=>7088, "name_ru"=>"Нью Олбани", "name_en"=>"New Albany", "lang"=>"en", "lat"=>40.079560, "lon"=>-82.808723),
"/US/US-OH/new-alexandria-16387" => array("type"=>"village", "pop"=>210, "name_en"=>"New Alexandria", "lang"=>"en", "lat"=>40.291227, "lon"=>-80.675310),
"/US/US-OH/new-athens-16389" => array("type"=>"village", "pop"=>339, "name_ru"=>"Нью Афины", "name_en"=>"New Athens", "lang"=>"en", "lat"=>40.184333, "lon"=>-80.995378),
"/US/US-OH/new-bavaria-16390" => array("type"=>"village", "pop"=>76, "name_en"=>"New Bavaria", "lang"=>"en", "lat"=>41.203554, "lon"=>-84.166980),
"/US/US-OH/new-bloomington-16391" => array("type"=>"village", "pop"=>531, "name_en"=>"New Bloomington", "lang"=>"en", "lat"=>40.585453, "lon"=>-83.315321),
"/US/US-OH/new-boston-16392" => array("type"=>"village", "pop"=>2169, "name_ru"=>"Нью Бостон", "name_en"=>"New Boston", "lang"=>"en", "lat"=>38.752050, "lon"=>-82.934958),
"/US/US-OH/new-bremen-16393" => array("type"=>"village", "pop"=>3071, "name_en"=>"New Bremen", "lang"=>"en", "lat"=>40.437379, "lon"=>-84.380598),
"/US/US-OH/new-carlisle-16395" => array("type"=>"city", "pop"=>5543, "name_ru"=>"Нью Карлайл", "name_en"=>"New Carlisle", "lang"=>"en", "lat"=>39.943472, "lon"=>-84.028086),
"/US/US-OH/new-concord-16397" => array("type"=>"village", "pop"=>2589, "name_en"=>"New Concord", "lang"=>"en", "lat"=>39.997372, "lon"=>-81.735788),
"/US/US-OH/new-holland-16398" => array("type"=>"village", "pop"=>786, "name_ru"=>"Нью Холланд", "name_en"=>"New Holland", "lang"=>"en", "lat"=>39.553244, "lon"=>-83.257408),
"/US/US-OH/new-knoxville-16399" => array("type"=>"village", "pop"=>890, "name_en"=>"New Knoxville", "lang"=>"en", "lat"=>40.494175, "lon"=>-84.318509),
"/US/US-OH/new-lebanon-16400" => array("type"=>"village", "pop"=>4100, "name_en"=>"New Lebanon", "lang"=>"en", "lat"=>39.745027, "lon"=>-84.400349),
"/US/US-OH/new-lexington-16401" => array("type"=>"city", "pop"=>4576, "name_en"=>"New Lexington", "lang"=>"en", "lat"=>39.713293, "lon"=>-82.208137),
"/US/US-OH/new-london-16402" => array("type"=>"village", "pop"=>2597, "name_ru"=>"Нью Лондон", "name_en"=>"New London", "lang"=>"en", "lat"=>41.079443, "lon"=>-82.406178),
"/US/US-OH/new-madison-16403" => array("type"=>"village", "pop"=>755, "name_en"=>"New Madison", "lang"=>"en", "lat"=>39.968520, "lon"=>-84.708282),
"/US/US-OH/new-miami-16404" => array("type"=>"village", "pop"=>2651, "name_en"=>"New Miami", "lang"=>"en", "lat"=>39.431500, "lon"=>-84.540983),
"/US/US-OH/new-middletown-16405" => array("type"=>"village", "pop"=>1553, "name_ru"=>"Нью Миддлтаун", "name_en"=>"New Middletown", "lang"=>"en", "lat"=>40.963664, "lon"=>-80.559988),
"/US/US-OH/new-paris-16406" => array("type"=>"village", "pop"=>1490, "name_ru"=>"Нью Парис", "name_en"=>"New Paris", "lang"=>"en", "lat"=>39.856974, "lon"=>-84.793531),
"/US/US-OH/new-philadelphia-16407" => array("type"=>"city", "pop"=>17319, "name_en"=>"New Philadelphia", "lang"=>"en", "lat"=>40.486658, "lon"=>-81.441259),
"/US/US-OH/new-richmond-16408" => array("type"=>"village", "pop"=>2510, "name_ru"=>"Нью Ричмонд", "name_en"=>"New Richmond", "lang"=>"en", "lat"=>38.962380, "lon"=>-84.279776),
"/US/US-OH/new-riegel-16409" => array("type"=>"village", "pop"=>212, "name_en"=>"New Riegel", "lang"=>"en", "lat"=>41.051482, "lon"=>-83.319024),
"/US/US-OH/new-rome-16410" => array("type"=>"village", "pop"=>111, "name_en"=>"New Rome", "lang"=>"en", "lat"=>39.951620, "lon"=>-83.141341),
"/US/US-OH/new-straitsville-16411" => array("type"=>"village", "pop"=>793, "name_en"=>"New Straitsville", "lang"=>"en", "lat"=>39.577850, "lon"=>-82.236049),
"/US/US-OH/new-vienna-16415" => array("type"=>"village", "pop"=>1373, "name_ru"=>"Нью Вьенна", "name_en"=>"New Vienna", "lang"=>"en", "lat"=>39.324698, "lon"=>-83.692416),
"/US/US-OH/new-washington-16416" => array("type"=>"village", "pop"=>911, "name_en"=>"New Washington", "lang"=>"en", "lat"=>40.961413, "lon"=>-82.854294),
"/US/US-OH/new-waterford-16417" => array("type"=>"village", "pop"=>1345, "name_en"=>"New Waterford", "lang"=>"en", "lat"=>40.848488, "lon"=>-80.620373),
"/US/US-OH/new-weston-16418" => array("type"=>"village", "pop"=>131, "name_en"=>"New Weston", "lang"=>"en", "lat"=>40.337042, "lon"=>-84.643978),
"/US/US-OH/newark-16388" => array("type"=>"city", "pop"=>47236, "name_ru"=>"Ньюарк", "name_en"=>"Newark", "lang"=>"en", "lat"=>40.067664, "lon"=>-82.423369),
"/US/US-OH/newburgh-heights-16394" => array("type"=>"village", "pop"=>2131, "name_en"=>"Newburgh Heights", "lang"=>"en", "lat"=>41.452550, "lon"=>-81.662349),
"/US/US-OH/newcomerstown-16396" => array("type"=>"village", "pop"=>3895, "name_en"=>"Newcomerstown", "lang"=>"en", "lat"=>40.274500, "lon"=>-81.598769),
"/US/US-OH/newton-falls-16412" => array("type"=>"city", "pop"=>4641, "name_en"=>"Newton Falls", "lang"=>"en", "lat"=>41.188925, "lon"=>-80.969832),
"/US/US-OH/newtonsville-16413" => array("type"=>"village", "pop"=>537, "name_en"=>"Newtonsville", "lang"=>"en", "lat"=>39.182886, "lon"=>-84.087434),
"/US/US-OH/newtown-16414" => array("type"=>"village", "pop"=>3889, "name_ru"=>"Ньютаун", "name_en"=>"Newtown", "lang"=>"en", "lat"=>39.122234, "lon"=>-84.350540),
"/US/US-OH/ney-16419" => array("type"=>"village", "pop"=>336, "name_en"=>"Ney", "lang"=>"en", "lat"=>41.380682, "lon"=>-84.520970),
"/US/US-OH/niles-16420" => array("type"=>"city", "pop"=>19268, "name_ru"=>"Найлс", "name_en"=>"Niles", "lang"=>"en", "lat"=>41.187800, "lon"=>-80.753099),
"/US/US-OH/north-baltimore-16421" => array("type"=>"village", "pop"=>3341, "name_en"=>"North Baltimore", "lang"=>"en", "lat"=>41.180307, "lon"=>-83.675592),
"/US/US-OH/north-bend-16422" => array("type"=>"village", "pop"=>638, "name_en"=>"North Bend", "lang"=>"en", "lat"=>39.148468, "lon"=>-84.741538),
"/US/US-OH/north-canton-16424" => array("type"=>"city", "pop"=>16925, "name_en"=>"North Canton", "lang"=>"en", "lat"=>40.873100, "lon"=>-81.398429),
"/US/US-OH/north-college-hill-16425" => array("type"=>"city", "pop"=>9549, "name_en"=>"North College Hill", "lang"=>"en", "lat"=>39.217151, "lon"=>-84.551647),
"/US/US-OH/north-fairfield-16426" => array("type"=>"village", "pop"=>559, "name_en"=>"North Fairfield", "lang"=>"en", "lat"=>41.104779, "lon"=>-82.612984),
"/US/US-OH/north-folk-village-16428" => array("type"=>"cdp", "pop"=>1247, "name_en"=>"North Folk Village", "lang"=>"en", "lat"=>39.336353, "lon"=>-83.027765),
"/US/US-OH/north-hampton-16430" => array("type"=>"village", "pop"=>352, "name_en"=>"North Hampton", "lang"=>"en", "lat"=>39.990533, "lon"=>-83.941743),
"/US/US-OH/north-kingsville-16431" => array("type"=>"village", "pop"=>2565, "name_en"=>"North Kingsville", "lang"=>"en", "lat"=>41.917174, "lon"=>-80.670211),
"/US/US-OH/north-lewisburg-16432" => array("type"=>"village", "pop"=>1585, "name_en"=>"North Lewisburg", "lang"=>"en", "lat"=>40.223811, "lon"=>-83.557825),
"/US/US-OH/north-madison-16433" => array("type"=>"cdp", "pop"=>8699, "name_en"=>"North Madison", "lang"=>"en", "lat"=>41.829724, "lon"=>-81.054767),
"/US/US-OH/north-olmsted-16434" => array("type"=>"city", "pop"=>31319, "name_en"=>"North Olmsted", "lang"=>"en", "lat"=>41.414700, "lon"=>-81.920474),
"/US/US-OH/north-perry-16435" => array("type"=>"village", "pop"=>964, "name_en"=>"North Perry", "lang"=>"en", "lat"=>41.806850, "lon"=>-81.125482),
"/US/US-OH/north-randall-16436" => array("type"=>"village", "pop"=>830, "name_en"=>"North Randall", "lang"=>"en", "lat"=>41.431150, "lon"=>-81.530739),
"/US/US-OH/north-ridgeville-16439" => array("type"=>"city", "pop"=>28153, "name_en"=>"North Ridgeville", "lang"=>"en", "lat"=>41.388050, "lon"=>-82.015156),
"/US/US-OH/north-robinson-16440" => array("type"=>"village", "pop"=>192, "name_en"=>"North Robinson", "lang"=>"en", "lat"=>40.792159, "lon"=>-82.856385),
"/US/US-OH/north-royalton-16441" => array("type"=>"city", "pop"=>29331, "name_en"=>"North Royalton", "lang"=>"en", "lat"=>41.312417, "lon"=>-81.745142),
"/US/US-OH/north-star-16442" => array("type"=>"village", "pop"=>203, "name_en"=>"North Star", "lang"=>"en", "lat"=>40.323940, "lon"=>-84.567989),
"/US/US-OH/north-zanesville-16445" => array("type"=>"cdp", "pop"=>2121, "name_en"=>"North Zanesville", "lang"=>"en", "lat"=>39.982204, "lon"=>-82.000071),
"/US/US-OH/northbrook-16423" => array("type"=>"cdp", "pop"=>11471, "name_ru"=>"Нортбрук", "name_en"=>"Northbrook", "lang"=>"en", "lat"=>39.247451, "lon"=>-84.579040),
"/US/US-OH/northfield-16427" => array("type"=>"village", "pop"=>3654, "name_ru"=>"Нортфилд", "name_en"=>"Northfield", "lang"=>"en", "lat"=>41.343650, "lon"=>-81.529740),
"/US/US-OH/northgate-16429" => array("type"=>"cdp", "pop"=>7864, "name_en"=>"Northgate", "lang"=>"en", "lat"=>39.252651, "lon"=>-84.592435),
"/US/US-OH/northridge-16437" => array("type"=>"cdp", "pop"=>5939, "name_en"=>"Northridge", "lang"=>"en", "lat"=>39.997100, "lon"=>-83.775649),
"/US/US-OH/northridge-16438" => array("type"=>"cdp", "pop"=>9448, "name_en"=>"Northridge", "lang"=>"en", "lat"=>39.810350, "lon"=>-84.191749),
"/US/US-OH/northview-16443" => array("type"=>"cdp", "pop"=>10337, "name_ru"=>"Нортвью", "name_en"=>"Northview", "lang"=>"en", "lat"=>39.843100, "lon"=>-84.279605),
"/US/US-OH/northwood-16444" => array("type"=>"city", "pop"=>5520, "name_ru"=>"Нортвуд", "name_en"=>"Northwood", "lang"=>"en", "lat"=>41.610350, "lon"=>-83.480278),
"/US/US-OH/norton-16446" => array("type"=>"city", "pop"=>11472, "name_ru"=>"Нортон", "name_en"=>"Norton", "lang"=>"en", "lat"=>41.028850, "lon"=>-81.645674),
"/US/US-OH/norwalk-16447" => array("type"=>"city", "pop"=>16637, "name_ru"=>"Норуолк", "name_en"=>"Norwalk", "lang"=>"en", "lat"=>41.242062, "lon"=>-82.610536),
"/US/US-OH/norwich-16448" => array("type"=>"village", "pop"=>117, "name_ru"=>"Норвич", "name_en"=>"Norwich", "lang"=>"en", "lat"=>39.984701, "lon"=>-81.792595),
"/US/US-OH/norwood-16449" => array("type"=>"city", "pop"=>20341, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>39.159601, "lon"=>-84.453717),
"/US/US-OH/oak-harbor-16450" => array("type"=>"village", "pop"=>2796, "name_en"=>"Oak Harbor", "lang"=>"en", "lat"=>41.511567, "lon"=>-83.146652),
"/US/US-OH/oak-hill-16451" => array("type"=>"village", "pop"=>1613, "name_ru"=>"Оук Хилл", "name_en"=>"Oak Hill", "lang"=>"en", "lat"=>38.895988, "lon"=>-82.569420),
"/US/US-OH/oakwood-16452" => array("type"=>"city", "pop"=>3719, "name_ru"=>"Оуквуд", "name_en"=>"Oakwood", "lang"=>"en", "lat"=>39.719933, "lon"=>-84.173576),
"/US/US-OH/oakwood-16453" => array("type"=>"village", "pop"=>3392, "name_ru"=>"Оуквуд", "name_en"=>"Oakwood", "lang"=>"en", "lat"=>41.367000, "lon"=>-81.503943),
"/US/US-OH/oakwood-16454" => array("type"=>"village", "pop"=>709, "name_ru"=>"Оуквуд", "name_en"=>"Oakwood", "lang"=>"en", "lat"=>41.092764, "lon"=>-84.376296),
"/US/US-OH/oberlin-16455" => array("type"=>"city", "pop"=>8381, "name_ru"=>"Оберлин", "name_en"=>"Oberlin", "lang"=>"en", "lat"=>41.290400, "lon"=>-82.216685),
"/US/US-OH/obetz-16456" => array("type"=>"village", "pop"=>4219, "name_en"=>"Obetz", "lang"=>"en", "lat"=>39.876700, "lon"=>-82.940649),
"/US/US-OH/octa-16457" => array("type"=>"village", "pop"=>79, "name_en"=>"Octa", "lang"=>"en", "lat"=>39.615256, "lon"=>-83.608577),
"/US/US-OH/ohio-city-16458" => array("type"=>"village", "pop"=>759, "name_en"=>"Ohio City", "lang"=>"en", "lat"=>40.770131, "lon"=>-84.616091),
"/US/US-OH/old-washington-16459" => array("type"=>"village", "pop"=>258, "name_en"=>"Old Washington", "lang"=>"en", "lat"=>40.037400, "lon"=>-81.444400),
"/US/US-OH/olmsted-falls-16460" => array("type"=>"city", "pop"=>8168, "name_en"=>"Olmsted Falls", "lang"=>"en", "lat"=>41.367511, "lon"=>-81.904957),
"/US/US-OH/ontario-16461" => array("type"=>"village", "pop"=>5204, "name_ru"=>"Онтарио", "name_en"=>"Ontario", "lang"=>"en", "lat"=>40.766050, "lon"=>-82.607610),
"/US/US-OH/orange-16462" => array("type"=>"village", "pop"=>3258, "name_ru"=>"Оранж", "name_en"=>"Orange", "lang"=>"en", "lat"=>41.438500, "lon"=>-81.472038),
"/US/US-OH/orangeville-16463" => array("type"=>"village", "pop"=>178, "name_ru"=>"Оранжвилл", "name_en"=>"Orangeville", "lang"=>"en", "lat"=>41.343288, "lon"=>-80.529230),
"/US/US-OH/oregon-16464" => array("type"=>"city", "pop"=>18921, "name_ru"=>"Орегон", "name_en"=>"Oregon", "lang"=>"en", "lat"=>41.666950, "lon"=>-83.419425),
"/US/US-OH/orient-16465" => array("type"=>"village", "pop"=>269, "name_ru"=>"Ориент", "name_en"=>"Orient", "lang"=>"en", "lat"=>39.805616, "lon"=>-83.151489),
"/US/US-OH/orrville-16466" => array("type"=>"city", "pop"=>8387, "name_ru"=>"Орвилл", "name_en"=>"Orrville", "lang"=>"en", "lat"=>40.846815, "lon"=>-81.771166),
"/US/US-OH/orwell-16467" => array("type"=>"village", "pop"=>1473, "name_en"=>"Orwell", "lang"=>"en", "lat"=>41.536082, "lon"=>-80.858505),
"/US/US-OH/osgood-16468" => array("type"=>"village", "pop"=>248, "name_ru"=>"Осгуд", "name_en"=>"Osgood", "lang"=>"en", "lat"=>40.340132, "lon"=>-84.494198),
"/US/US-OH/ostrander-16469" => array("type"=>"village", "pop"=>562, "name_ru"=>"Острэндер", "name_en"=>"Ostrander", "lang"=>"en", "lat"=>40.265072, "lon"=>-83.212129),
"/US/US-OH/ottawa-16470" => array("type"=>"village", "pop"=>4403, "name_ru"=>"Оттава", "name_en"=>"Ottawa", "lang"=>"en", "lat"=>41.020834, "lon"=>-84.041370),
"/US/US-OH/ottawa-hills-16471" => array("type"=>"village", "pop"=>4594, "name_en"=>"Ottawa Hills", "lang"=>"en", "lat"=>41.667500, "lon"=>-83.645007),
"/US/US-OH/ottoville-16472" => array("type"=>"village", "pop"=>860, "name_en"=>"Ottoville", "lang"=>"en", "lat"=>40.931638, "lon"=>-84.338822),
"/US/US-OH/otway-16473" => array("type"=>"village", "pop"=>83, "name_en"=>"Otway", "lang"=>"en", "lat"=>38.865001, "lon"=>-83.188262),
"/US/US-OH/overlook-page-manor-16474" => array("type"=>"cdp", "pop"=>13242, "name_en"=>"Overlook-Page Manor", "lang"=>"en", "lat"=>39.754100, "lon"=>-84.115371),
"/US/US-OH/owensville-16475" => array("type"=>"village", "pop"=>850, "name_ru"=>"Оуэнсвилл", "name_en"=>"Owensville", "lang"=>"en", "lat"=>39.123741, "lon"=>-84.135379),
"/US/US-OH/oxford-16476" => array("type"=>"city", "pop"=>22887, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>39.505667, "lon"=>-84.747241),
"/US/US-OH/painesville-16477" => array("type"=>"city", "pop"=>18474, "name_en"=>"Painesville", "lang"=>"en", "lat"=>41.727100, "lon"=>-81.248301),
"/US/US-OH/palestine-16478" => array("type"=>"village", "pop"=>163, "name_ru"=>"Палестин", "name_en"=>"Palestine", "lang"=>"en", "lat"=>40.050255, "lon"=>-84.744460),
"/US/US-OH/pandora-16479" => array("type"=>"village", "pop"=>1184, "name_en"=>"Pandora", "lang"=>"en", "lat"=>40.947677, "lon"=>-83.961230),
"/US/US-OH/park-layne-16480" => array("type"=>"cdp", "pop"=>4795, "name_en"=>"Park Layne", "lang"=>"en", "lat"=>39.886472, "lon"=>-84.039614),
"/US/US-OH/parma-16481" => array("type"=>"city", "pop"=>77947, "name_ru"=>"Парма", "name_en"=>"Parma", "lang"=>"en", "lat"=>41.383550, "lon"=>-81.729443),
"/US/US-OH/parma-heights-16482" => array("type"=>"city", "pop"=>19757, "name_en"=>"Parma Heights", "lang"=>"en", "lat"=>41.386000, "lon"=>-81.763605),
"/US/US-OH/parral-16483" => array("type"=>"village", "pop"=>233, "name_en"=>"Parral", "lang"=>"en", "lat"=>40.561221, "lon"=>-81.495005),
"/US/US-OH/pataskala-16484" => array("type"=>"village", "pop"=>12683, "name_en"=>"Pataskala", "lang"=>"en", "lat"=>40.000706, "lon"=>-82.673525),
"/US/US-OH/patterson-16485" => array("type"=>"village", "pop"=>136, "name_ru"=>"Паттерсон", "name_en"=>"Patterson", "lang"=>"en", "lat"=>40.781798, "lon"=>-83.526102),
"/US/US-OH/paulding-16486" => array("type"=>"village", "pop"=>3355, "name_en"=>"Paulding", "lang"=>"en", "lat"=>41.136803, "lon"=>-84.582661),
"/US/US-OH/payne-16487" => array("type"=>"village", "pop"=>1152, "name_ru"=>"Пейн", "name_en"=>"Payne", "lang"=>"en", "lat"=>41.079379, "lon"=>-84.727235),
"/US/US-OH/peebles-16488" => array("type"=>"village", "pop"=>1828, "name_en"=>"Peebles", "lang"=>"en", "lat"=>38.947215, "lon"=>-83.408270),
"/US/US-OH/pemberville-16489" => array("type"=>"village", "pop"=>1358, "name_en"=>"Pemberville", "lang"=>"en", "lat"=>41.409779, "lon"=>-83.457942),
"/US/US-OH/peninsula-16490" => array("type"=>"village", "pop"=>689, "name_en"=>"Peninsula", "lang"=>"en", "lat"=>41.236767, "lon"=>-81.553033),
"/US/US-OH/pepper-pike-16491" => array("type"=>"city", "pop"=>5715, "name_en"=>"Pepper Pike", "lang"=>"en", "lat"=>41.479550, "lon"=>-81.460199),
"/US/US-OH/perry-16492" => array("type"=>"village", "pop"=>1282, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>41.767347, "lon"=>-81.143049),
"/US/US-OH/perry-heights-16493" => array("type"=>"cdp", "pop"=>9055, "name_en"=>"Perry Heights", "lang"=>"en", "lat"=>40.796750, "lon"=>-81.469070),
"/US/US-OH/perrysburg-16494" => array("type"=>"city", "pop"=>17049, "name_en"=>"Perrysburg", "lang"=>"en", "lat"=>41.553050, "lon"=>-83.623834),
"/US/US-OH/perrysville-16495" => array("type"=>"village", "pop"=>816, "name_ru"=>"Перрисвилл", "name_en"=>"Perrysville", "lang"=>"en", "lat"=>40.656714, "lon"=>-82.312414),
"/US/US-OH/phillipsburg-16496" => array("type"=>"village", "pop"=>588, "name_ru"=>"Филлипсбург", "name_en"=>"Phillipsburg", "lang"=>"en", "lat"=>39.904350, "lon"=>-84.400672),
"/US/US-OH/philo-16497" => array("type"=>"village", "pop"=>761, "name_ru"=>"Фило", "name_en"=>"Philo", "lang"=>"en", "lat"=>39.861211, "lon"=>-81.909334),
"/US/US-OH/pickerington-16498" => array("type"=>"village", "pop"=>17531, "name_en"=>"Pickerington", "lang"=>"en", "lat"=>39.890538, "lon"=>-82.768367),
"/US/US-OH/pigeon-creek-16499" => array("type"=>"cdp", "pop"=>1008, "name_en"=>"Pigeon Creek", "lang"=>"en", "lat"=>41.110950, "lon"=>-81.673156),
"/US/US-OH/piketon-16500" => array("type"=>"village", "pop"=>1971, "name_en"=>"Piketon", "lang"=>"en", "lat"=>39.068233, "lon"=>-82.999948),
"/US/US-OH/pioneer-16501" => array("type"=>"village", "pop"=>1383, "name_ru"=>"Пайониер", "name_en"=>"Pioneer", "lang"=>"en", "lat"=>41.678034, "lon"=>-84.552081),
"/US/US-OH/piqua-16502" => array("type"=>"city", "pop"=>20582, "name_en"=>"Piqua", "lang"=>"en", "lat"=>40.146571, "lon"=>-84.250960),
"/US/US-OH/pitsburg-16503" => array("type"=>"village", "pop"=>380, "name_en"=>"Pitsburg", "lang"=>"en", "lat"=>39.986744, "lon"=>-84.487679),
"/US/US-OH/plain-city-16504" => array("type"=>"village", "pop"=>3543, "name_en"=>"Plain City", "lang"=>"en", "lat"=>40.106802, "lon"=>-83.269985),
"/US/US-OH/plainfield-16505" => array("type"=>"village", "pop"=>156, "name_ru"=>"Плейнфилд", "name_en"=>"Plainfield", "lang"=>"en", "lat"=>40.206278, "lon"=>-81.719033),
"/US/US-OH/pleasant-city-16506" => array("type"=>"village", "pop"=>425, "name_en"=>"Pleasant City", "lang"=>"en", "lat"=>39.903364, "lon"=>-81.544667),
"/US/US-OH/pleasant-grove-16507" => array("type"=>"cdp", "pop"=>2001, "name_ru"=>"Плезант Гров", "name_en"=>"Pleasant Grove", "lang"=>"en", "lat"=>39.949636, "lon"=>-81.959593),
"/US/US-OH/pleasant-hill-16508" => array("type"=>"village", "pop"=>1134, "name_ru"=>"Плезант Хилл", "name_en"=>"Pleasant Hill", "lang"=>"en", "lat"=>40.051328, "lon"=>-84.344923),
"/US/US-OH/pleasant-plain-16511" => array("type"=>"village", "pop"=>169, "name_ru"=>"Плезант Плейн", "name_en"=>"Pleasant Plain", "lang"=>"en", "lat"=>39.279723, "lon"=>-84.109316),
"/US/US-OH/pleasant-run-16509" => array("type"=>"cdp", "pop"=>4964, "name_en"=>"Pleasant Run", "lang"=>"en", "lat"=>39.290851, "lon"=>-84.572411),
"/US/US-OH/pleasant-run-farm-16510" => array("type"=>"cdp", "pop"=>4545, "name_en"=>"Pleasant Run Farm", "lang"=>"en", "lat"=>39.302968, "lon"=>-84.548036),
"/US/US-OH/pleasantville-16512" => array("type"=>"village", "pop"=>853, "name_ru"=>"Плезантвилл", "name_en"=>"Pleasantville", "lang"=>"en", "lat"=>39.808985, "lon"=>-82.522756),
"/US/US-OH/plymouth-16513" => array("type"=>"village", "pop"=>1803, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>40.994957, "lon"=>-82.667973),
"/US/US-OH/poland-16514" => array("type"=>"village", "pop"=>2662, "name_en"=>"Poland", "lang"=>"en", "lat"=>41.024050, "lon"=>-80.614116),
"/US/US-OH/polk-16515" => array("type"=>"village", "pop"=>350, "name_en"=>"Polk", "lang"=>"en", "lat"=>40.945219, "lon"=>-82.214349),
"/US/US-OH/pomeroy-16516" => array("type"=>"village", "pop"=>1952, "name_ru"=>"Померой", "name_en"=>"Pomeroy", "lang"=>"en", "lat"=>39.027622, "lon"=>-82.033225),
"/US/US-OH/port-clinton-16519" => array("type"=>"city", "pop"=>6135, "name_en"=>"Port Clinton", "lang"=>"en", "lat"=>41.509675, "lon"=>-82.938360),
"/US/US-OH/port-jefferson-16520" => array("type"=>"village", "pop"=>317, "name_en"=>"Port Jefferson", "lang"=>"en", "lat"=>40.329996, "lon"=>-84.092132),
"/US/US-OH/port-washington-16522" => array("type"=>"village", "pop"=>565, "name_en"=>"Port Washington", "lang"=>"en", "lat"=>40.326885, "lon"=>-81.519561),
"/US/US-OH/port-william-16523" => array("type"=>"village", "pop"=>253, "name_en"=>"Port William", "lang"=>"en", "lat"=>39.551712, "lon"=>-83.785735),
"/US/US-OH/portage-16517" => array("type"=>"village", "pop"=>430, "name_ru"=>"Портейдж", "name_en"=>"Portage", "lang"=>"en", "lat"=>41.323776, "lon"=>-83.647920),
"/US/US-OH/portage-lakes-16518" => array("type"=>"cdp", "pop"=>13373, "name_en"=>"Portage Lakes", "lang"=>"en", "lat"=>40.987132, "lon"=>-81.536614),
"/US/US-OH/portsmouth-16521" => array("type"=>"city", "pop"=>20297, "name_ru"=>"Портсмаут", "name_en"=>"Portsmouth", "lang"=>"en", "lat"=>38.755139, "lon"=>-82.950670),
"/US/US-OH/potsdam-16524" => array("type"=>"village", "pop"=>207, "name_en"=>"Potsdam", "lang"=>"en", "lat"=>39.965463, "lon"=>-84.412065),
"/US/US-OH/powell-16525" => array("type"=>"village", "pop"=>12814, "name_ru"=>"Пауэлл", "name_en"=>"Powell", "lang"=>"en", "lat"=>40.160506, "lon"=>-83.065633),
"/US/US-OH/powhatan-point-16526" => array("type"=>"village", "pop"=>1655, "name_en"=>"Powhatan Point", "lang"=>"en", "lat"=>39.861148, "lon"=>-80.809424),
"/US/US-OH/proctorville-16527" => array("type"=>"village", "pop"=>622, "name_en"=>"Proctorville", "lang"=>"en", "lat"=>38.437616, "lon"=>-82.380920),
"/US/US-OH/prospect-16528" => array("type"=>"village", "pop"=>1092, "name_ru"=>"Проспект", "name_en"=>"Prospect", "lang"=>"en", "lat"=>40.452419, "lon"=>-83.185706),
"/US/US-OH/put-in-bay-16529" => array("type"=>"village", "pop"=>131, "name_en"=>"Put-in-Bay", "lang"=>"en", "lat"=>41.653072, "lon"=>-82.815097),
"/US/US-OH/quaker-city-16530" => array("type"=>"village", "pop"=>546, "name_en"=>"Quaker City", "lang"=>"en", "lat"=>39.969136, "lon"=>-81.296425),
"/US/US-OH/quincy-16531" => array("type"=>"village", "pop"=>692, "name_ru"=>"Квинси", "name_en"=>"Quincy", "lang"=>"en", "lat"=>40.295888, "lon"=>-83.969484),
"/US/US-OH/racine-16532" => array("type"=>"village", "pop"=>741, "name_ru"=>"Расин", "name_en"=>"Racine", "lang"=>"en", "lat"=>38.968182, "lon"=>-81.912548),
"/US/US-OH/rarden-16533" => array("type"=>"village", "pop"=>171, "name_en"=>"Rarden", "lang"=>"en", "lat"=>38.922702, "lon"=>-83.242788),
"/US/US-OH/ravenna-16534" => array("type"=>"city", "pop"=>11381, "name_ru"=>"Равенна", "name_en"=>"Ravenna", "lang"=>"en", "lat"=>41.159800, "lon"=>-81.242574),
"/US/US-OH/rawson-16535" => array("type"=>"village", "pop"=>433, "name_en"=>"Rawson", "lang"=>"en", "lat"=>40.956702, "lon"=>-83.785733),
"/US/US-OH/rayland-16536" => array("type"=>"village", "pop"=>398, "name_en"=>"Rayland", "lang"=>"en", "lat"=>40.183367, "lon"=>-80.691447),
"/US/US-OH/reading-16537" => array("type"=>"city", "pop"=>10523, "name_ru"=>"Ридинг", "name_en"=>"Reading", "lang"=>"en", "lat"=>39.224392, "lon"=>-84.431409),
"/US/US-OH/reminderville-16538" => array("type"=>"village", "pop"=>2741, "name_en"=>"Reminderville", "lang"=>"en", "lat"=>41.327400, "lon"=>-81.397273),
"/US/US-OH/rendville-16539" => array("type"=>"village", "pop"=>47, "name_en"=>"Rendville", "lang"=>"en", "lat"=>39.619261, "lon"=>-82.090177),
"/US/US-OH/republic-16540" => array("type"=>"village", "pop"=>577, "name_ru"=>"Репаблик", "name_en"=>"Republic", "lang"=>"en", "lat"=>41.124764, "lon"=>-83.016078),
"/US/US-OH/reynoldsburg-16541" => array("type"=>"city", "pop"=>33657, "name_en"=>"Reynoldsburg", "lang"=>"en", "lat"=>39.955483, "lon"=>-82.799160),
"/US/US-OH/richfield-16542" => array("type"=>"village", "pop"=>3576, "name_ru"=>"Ричфилд", "name_en"=>"Richfield", "lang"=>"en", "lat"=>41.231650, "lon"=>-81.637322),
"/US/US-OH/richmond-16543" => array("type"=>"village", "pop"=>450, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>40.432808, "lon"=>-80.772090),
"/US/US-OH/richmond-heights-16544" => array("type"=>"city", "pop"=>10174, "name_ru"=>"Ричмонд Хайтс", "name_en"=>"Richmond Heights", "lang"=>"en", "lat"=>41.558133, "lon"=>-81.503809),
"/US/US-OH/richwood-16545" => array("type"=>"village", "pop"=>2175, "name_ru"=>"Ричвуд", "name_en"=>"Richwood", "lang"=>"en", "lat"=>40.427132, "lon"=>-83.296172),
"/US/US-OH/ridgeway-16546" => array("type"=>"village", "pop"=>344, "name_ru"=>"Риджуэй", "name_en"=>"Ridgeway", "lang"=>"en", "lat"=>40.512457, "lon"=>-83.568482),
"/US/US-OH/rio-grande-16547" => array("type"=>"village", "pop"=>873, "name_en"=>"Rio Grande", "lang"=>"en", "lat"=>38.881662, "lon"=>-82.378619),
"/US/US-OH/ripley-16548" => array("type"=>"village", "pop"=>1668, "name_ru"=>"Рипли", "name_en"=>"Ripley", "lang"=>"en", "lat"=>38.739630, "lon"=>-83.840248),
"/US/US-OH/risingsun-16549" => array("type"=>"village", "pop"=>607, "name_en"=>"Risingsun", "lang"=>"en", "lat"=>41.268758, "lon"=>-83.426416),
"/US/US-OH/rittman-16550" => array("type"=>"city", "pop"=>6274, "name_en"=>"Rittman", "lang"=>"en", "lat"=>40.971046, "lon"=>-81.783261),
"/US/US-OH/riverlea-16551" => array("type"=>"village", "pop"=>477, "name_en"=>"Riverlea", "lang"=>"en", "lat"=>40.080450, "lon"=>-83.025790),
"/US/US-OH/riverside-16552" => array("type"=>"village", "pop"=>25327, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>39.780633, "lon"=>-84.123936),
"/US/US-OH/roaming-shores-16553" => array("type"=>"village", "pop"=>1195, "name_en"=>"Roaming Shores", "lang"=>"en", "lat"=>41.636749, "lon"=>-80.828477),
"/US/US-OH/rochester-16554" => array("type"=>"village", "pop"=>181, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>41.124835, "lon"=>-82.306275),
"/US/US-OH/rock-creek-16555" => array("type"=>"village", "pop"=>554, "name_ru"=>"Рок Крик", "name_en"=>"Rock Creek", "lang"=>"en", "lat"=>41.660731, "lon"=>-80.853836),
"/US/US-OH/rockford-16556" => array("type"=>"village", "pop"=>1107, "name_ru"=>"Рокфорд", "name_en"=>"Rockford", "lang"=>"en", "lat"=>40.692015, "lon"=>-84.648694),
"/US/US-OH/rocky-ridge-16557" => array("type"=>"village", "pop"=>384, "name_ru"=>"Роки Ридж", "name_en"=>"Rocky Ridge", "lang"=>"en", "lat"=>41.530474, "lon"=>-83.213526),
"/US/US-OH/rocky-river-16558" => array("type"=>"city", "pop"=>18900, "name_en"=>"Rocky River", "lang"=>"en", "lat"=>41.473050, "lon"=>-81.854261),
"/US/US-OH/rogers-16559" => array("type"=>"village", "pop"=>251, "name_ru"=>"Роджерс", "name_en"=>"Rogers", "lang"=>"en", "lat"=>40.790208, "lon"=>-80.627057),
"/US/US-OH/rome-16560" => array("type"=>"village", "pop"=>123, "name_ru"=>"Роум", "name_en"=>"Rome", "lang"=>"en", "lat"=>38.664815, "lon"=>-83.379073),
"/US/US-OH/rosemount-16561" => array("type"=>"cdp", "pop"=>1926, "name_ru"=>"Роузмаунт", "name_en"=>"Rosemount", "lang"=>"en", "lat"=>38.787822, "lon"=>-82.971695),
"/US/US-OH/roseville-16562" => array("type"=>"village", "pop"=>1897, "name_ru"=>"Роузвилл", "name_en"=>"Roseville", "lang"=>"en", "lat"=>39.807148, "lon"=>-82.073646),
"/US/US-OH/ross-16563" => array("type"=>"cdp", "pop"=>7854, "name_ru"=>"Росс", "name_en"=>"Ross", "lang"=>"en", "lat"=>39.311750, "lon"=>-84.652895),
"/US/US-OH/rossburg-16564" => array("type"=>"village", "pop"=>216, "name_en"=>"Rossburg", "lang"=>"en", "lat"=>40.279809, "lon"=>-84.638159),
"/US/US-OH/rossford-16565" => array("type"=>"city", "pop"=>6404, "name_en"=>"Rossford", "lang"=>"en", "lat"=>41.603747, "lon"=>-83.561967),
"/US/US-OH/roswell-16566" => array("type"=>"village", "pop"=>279, "name_ru"=>"Росуэл", "name_en"=>"Roswell", "lang"=>"en", "lat"=>40.475422, "lon"=>-81.346430),
"/US/US-OH/rushsylvania-16567" => array("type"=>"village", "pop"=>516, "name_en"=>"Rushsylvania", "lang"=>"en", "lat"=>40.461171, "lon"=>-83.671065),
"/US/US-OH/rushville-16568" => array("type"=>"village", "pop"=>262, "name_ru"=>"Рашвилл", "name_en"=>"Rushville", "lang"=>"en", "lat"=>39.762676, "lon"=>-82.431326),
"/US/US-OH/russells-point-16569" => array("type"=>"village", "pop"=>1519, "name_en"=>"Russells Point", "lang"=>"en", "lat"=>40.467988, "lon"=>-83.893254),
"/US/US-OH/russellville-16570" => array("type"=>"village", "pop"=>425, "name_ru"=>"Расселлвилл", "name_en"=>"Russellville", "lang"=>"en", "lat"=>38.867398, "lon"=>-83.787532),
"/US/US-OH/russia-16571" => array("type"=>"village", "pop"=>615, "name_en"=>"Russia", "lang"=>"en", "lat"=>40.232400, "lon"=>-84.410870),
"/US/US-OH/rutland-16572" => array("type"=>"village", "pop"=>414, "name_ru"=>"Рутленд", "name_en"=>"Rutland", "lang"=>"en", "lat"=>39.042214, "lon"=>-82.129101),
"/US/US-OH/sabina-16573" => array("type"=>"village", "pop"=>2783, "name_en"=>"Sabina", "lang"=>"en", "lat"=>39.490274, "lon"=>-83.633758),
"/US/US-OH/salem-16581" => array("type"=>"city", "pop"=>11781, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>40.902624, "lon"=>-80.853727),
"/US/US-OH/salesville-16582" => array("type"=>"village", "pop"=>149, "name_ru"=>"Сейлсвилл", "name_en"=>"Salesville", "lang"=>"en", "lat"=>39.973440, "lon"=>-81.336975),
"/US/US-OH/salineville-16583" => array("type"=>"village", "pop"=>1318, "name_en"=>"Salineville", "lang"=>"en", "lat"=>40.621016, "lon"=>-80.832818),
"/US/US-OH/sandusky-16584" => array("type"=>"city", "pop"=>25688, "name_ru"=>"Сэндаски", "name_en"=>"Sandusky", "lang"=>"en", "lat"=>41.455881, "lon"=>-82.714369),
"/US/US-OH/sandusky-south-16585" => array("type"=>"cdp", "pop"=>6336, "name_en"=>"Sandusky South", "lang"=>"en", "lat"=>41.415970, "lon"=>-82.686992),
"/US/US-OH/sarahsville-16586" => array("type"=>"village", "pop"=>192, "name_en"=>"Sarahsville", "lang"=>"en", "lat"=>39.807713, "lon"=>-81.469503),
"/US/US-OH/sardinia-16587" => array("type"=>"village", "pop"=>828, "name_en"=>"Sardinia", "lang"=>"en", "lat"=>39.006941, "lon"=>-83.807016),
"/US/US-OH/savannah-16588" => array("type"=>"village", "pop"=>369, "name_ru"=>"Саванна", "name_en"=>"Savannah", "lang"=>"en", "lat"=>40.967372, "lon"=>-82.364290),
"/US/US-OH/scio-16589" => array("type"=>"village", "pop"=>759, "name_en"=>"Scio", "lang"=>"en", "lat"=>40.398669, "lon"=>-81.087568),
"/US/US-OH/sciotodale-16590" => array("type"=>"cdp", "pop"=>1128, "name_en"=>"Sciotodale", "lang"=>"en", "lat"=>38.755344, "lon"=>-82.861063),
"/US/US-OH/scott-16591" => array("type"=>"village", "pop"=>310, "name_ru"=>"Скотт", "name_en"=>"Scott", "lang"=>"en", "lat"=>40.988607, "lon"=>-84.584105),
"/US/US-OH/seaman-16592" => array("type"=>"village", "pop"=>1074, "name_en"=>"Seaman", "lang"=>"en", "lat"=>38.938820, "lon"=>-83.573109),
"/US/US-OH/sebring-16593" => array("type"=>"city", "pop"=>4534, "name_ru"=>"Себринг", "name_en"=>"Sebring", "lang"=>"en", "lat"=>40.923464, "lon"=>-81.022321),
"/US/US-OH/senecaville-16594" => array("type"=>"village", "pop"=>439, "name_en"=>"Senecaville", "lang"=>"en", "lat"=>39.934540, "lon"=>-81.459902),
"/US/US-OH/seven-hills-16595" => array("type"=>"city", "pop"=>11616, "name_en"=>"Seven Hills", "lang"=>"en", "lat"=>41.379467, "lon"=>-81.675069),
"/US/US-OH/seven-mile-16596" => array("type"=>"village", "pop"=>770, "name_en"=>"Seven Mile", "lang"=>"en", "lat"=>39.479317, "lon"=>-84.552491),
"/US/US-OH/seville-16597" => array("type"=>"village", "pop"=>2409, "name_en"=>"Seville", "lang"=>"en", "lat"=>41.022398, "lon"=>-81.867675),
"/US/US-OH/shadyside-16598" => array("type"=>"village", "pop"=>3530, "name_en"=>"Shadyside", "lang"=>"en", "lat"=>39.971312, "lon"=>-80.750946),
"/US/US-OH/shaker-heights-16599" => array("type"=>"city", "pop"=>26460, "name_en"=>"Shaker Heights", "lang"=>"en", "lat"=>41.475600, "lon"=>-81.547799),
"/US/US-OH/sharonville-16600" => array("type"=>"city", "pop"=>13377, "name_en"=>"Sharonville", "lang"=>"en", "lat"=>39.281184, "lon"=>-84.406481),
"/US/US-OH/shawnee-16601" => array("type"=>"village", "pop"=>631, "name_ru"=>"Шони", "name_en"=>"Shawnee", "lang"=>"en", "lat"=>39.610499, "lon"=>-82.207245),
"/US/US-OH/shawnee-hills-16602" => array("type"=>"village", "pop"=>577, "name_en"=>"Shawnee Hills", "lang"=>"en", "lat"=>40.159360, "lon"=>-83.135465),
"/US/US-OH/shawnee-hills-16603" => array("type"=>"cdp", "pop"=>2199, "name_en"=>"Shawnee Hills", "lang"=>"en", "lat"=>39.652456, "lon"=>-83.782622),
"/US/US-OH/sheffield-16604" => array("type"=>"village", "pop"=>3515, "name_ru"=>"Шеффилд", "name_en"=>"Sheffield", "lang"=>"en", "lat"=>41.456850, "lon"=>-82.094641),
"/US/US-OH/sheffield-lake-16605" => array("type"=>"city", "pop"=>8905, "name_en"=>"Sheffield Lake", "lang"=>"en", "lat"=>41.488850, "lon"=>-82.097924),
"/US/US-OH/shelby-16606" => array("type"=>"city", "pop"=>9311, "name_ru"=>"Шелби", "name_en"=>"Shelby", "lang"=>"en", "lat"=>40.882061, "lon"=>-82.661487),
"/US/US-OH/sherrodsville-16607" => array("type"=>"village", "pop"=>306, "name_en"=>"Sherrodsville", "lang"=>"en", "lat"=>40.494664, "lon"=>-81.244651),
"/US/US-OH/sherwood-16608" => array("type"=>"village", "pop"=>806, "name_ru"=>"Шервуд", "name_en"=>"Sherwood", "lang"=>"en", "lat"=>41.290106, "lon"=>-84.552968),
"/US/US-OH/sherwood-16609" => array("type"=>"cdp", "pop"=>3709, "name_ru"=>"Шервуд", "name_en"=>"Sherwood", "lang"=>"en", "lat"=>39.084734, "lon"=>-84.360697),
"/US/US-OH/shiloh-16610" => array("type"=>"village", "pop"=>687, "name_ru"=>"Шилох", "name_en"=>"Shiloh", "lang"=>"en", "lat"=>40.968713, "lon"=>-82.602199),
"/US/US-OH/shiloh-16611" => array("type"=>"cdp", "pop"=>11607, "name_ru"=>"Шилох", "name_en"=>"Shiloh", "lang"=>"en", "lat"=>39.813550, "lon"=>-84.231755),
"/US/US-OH/shreve-16612" => array("type"=>"village", "pop"=>1496, "name_en"=>"Shreve", "lang"=>"en", "lat"=>40.682190, "lon"=>-82.022347),
"/US/US-OH/sidney-16613" => array("type"=>"city", "pop"=>19944, "name_ru"=>"Сидни", "name_en"=>"Sidney", "lang"=>"en", "lat"=>40.288836, "lon"=>-84.163238),
"/US/US-OH/silver-lake-16614" => array("type"=>"village", "pop"=>3100, "name_ru"=>"Сильвер Лэйк", "name_en"=>"Silver Lake", "lang"=>"en", "lat"=>41.156600, "lon"=>-81.459521),
"/US/US-OH/silverton-16615" => array("type"=>"city", "pop"=>5155, "name_ru"=>"Сильвертон", "name_en"=>"Silverton", "lang"=>"en", "lat"=>39.189251, "lon"=>-84.401322),
"/US/US-OH/sinking-spring-16616" => array("type"=>"village", "pop"=>162, "name_en"=>"Sinking Spring", "lang"=>"en", "lat"=>39.075147, "lon"=>-83.387219),
"/US/US-OH/smithfield-16617" => array("type"=>"village", "pop"=>798, "name_ru"=>"Смитфилд", "name_en"=>"Smithfield", "lang"=>"en", "lat"=>40.272560, "lon"=>-80.776820),
"/US/US-OH/smithville-16618" => array("type"=>"village", "pop"=>1298, "name_ru"=>"Смитвилл", "name_en"=>"Smithville", "lang"=>"en", "lat"=>40.862778, "lon"=>-81.859833),
"/US/US-OH/solon-16619" => array("type"=>"city", "pop"=>21866, "name_ru"=>"Солон", "name_en"=>"Solon", "lang"=>"en", "lat"=>41.386233, "lon"=>-81.440874),
"/US/US-OH/somerset-16620" => array("type"=>"village", "pop"=>1556, "name_ru"=>"Сомерсет", "name_en"=>"Somerset", "lang"=>"en", "lat"=>39.805831, "lon"=>-82.300040),
"/US/US-OH/somerville-16621" => array("type"=>"village", "pop"=>321, "name_ru"=>"Сомервилл", "name_en"=>"Somerville", "lang"=>"en", "lat"=>39.563508, "lon"=>-84.638568),
"/US/US-OH/south-amherst-16622" => array("type"=>"village", "pop"=>1743, "name_ru"=>"Саут Амхерст", "name_en"=>"South Amherst", "lang"=>"en", "lat"=>41.353933, "lon"=>-82.238217),
"/US/US-OH/south-bloomfield-16623" => array("type"=>"village", "pop"=>1642, "name_en"=>"South Bloomfield", "lang"=>"en", "lat"=>39.715939, "lon"=>-82.990036),
"/US/US-OH/south-canal-16624" => array("type"=>"cdp", "pop"=>1319, "name_en"=>"South Canal", "lang"=>"en", "lat"=>41.177312, "lon"=>-80.986821),
"/US/US-OH/south-charleston-16625" => array("type"=>"village", "pop"=>1774, "name_en"=>"South Charleston", "lang"=>"en", "lat"=>39.824699, "lon"=>-83.644556),
"/US/US-OH/south-euclid-16626" => array("type"=>"city", "pop"=>21209, "name_en"=>"South Euclid", "lang"=>"en", "lat"=>41.524181, "lon"=>-81.525238),
"/US/US-OH/south-lebanon-16627" => array("type"=>"village", "pop"=>3691, "name_en"=>"South Lebanon", "lang"=>"en", "lat"=>39.371183, "lon"=>-84.211388),
"/US/US-OH/south-middletown-16628" => array("type"=>"cdp", "pop"=>3491, "name_en"=>"South Middletown", "lang"=>"en", "lat"=>39.481150, "lon"=>-84.400701),
"/US/US-OH/south-point-16629" => array("type"=>"village", "pop"=>3986, "name_en"=>"South Point", "lang"=>"en", "lat"=>38.419208, "lon"=>-82.574227),
"/US/US-OH/south-russell-16630" => array("type"=>"village", "pop"=>3913, "name_en"=>"South Russell", "lang"=>"en", "lat"=>41.434650, "lon"=>-81.332782),
"/US/US-OH/south-salem-16631" => array("type"=>"village", "pop"=>219, "name_en"=>"South Salem", "lang"=>"en", "lat"=>39.336354, "lon"=>-83.307480),
"/US/US-OH/south-solon-16632" => array("type"=>"village", "pop"=>387, "name_en"=>"South Solon", "lang"=>"en", "lat"=>39.737080, "lon"=>-83.613458),
"/US/US-OH/south-vienna-16633" => array("type"=>"village", "pop"=>450, "name_en"=>"South Vienna", "lang"=>"en", "lat"=>39.928680, "lon"=>-83.611751),
"/US/US-OH/south-webster-16634" => array("type"=>"village", "pop"=>742, "name_en"=>"South Webster", "lang"=>"en", "lat"=>38.815575, "lon"=>-82.727870),
"/US/US-OH/south-zanesville-16635" => array("type"=>"village", "pop"=>2021, "name_en"=>"South Zanesville", "lang"=>"en", "lat"=>39.903398, "lon"=>-82.019690),
"/US/US-OH/sparta-16636" => array("type"=>"village", "pop"=>195, "name_ru"=>"Спарта", "name_en"=>"Sparta", "lang"=>"en", "lat"=>40.394712, "lon"=>-82.699633),
"/US/US-OH/spencer-16637" => array("type"=>"village", "pop"=>820, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>41.097505, "lon"=>-82.122772),
"/US/US-OH/spencerville-16638" => array("type"=>"village", "pop"=>2160, "name_en"=>"Spencerville", "lang"=>"en", "lat"=>40.708013, "lon"=>-84.353214),
"/US/US-OH/spring-valley-16642" => array("type"=>"village", "pop"=>495, "name_ru"=>"Спринг Валли", "name_en"=>"Spring Valley", "lang"=>"en", "lat"=>39.610154, "lon"=>-84.006054),
"/US/US-OH/springboro-16639" => array("type"=>"city", "pop"=>17551, "name_en"=>"Springboro", "lang"=>"en", "lat"=>39.563368, "lon"=>-84.237705),
"/US/US-OH/springdale-16640" => array("type"=>"city", "pop"=>10384, "name_ru"=>"Спрингдейл", "name_en"=>"Springdale", "lang"=>"en", "lat"=>39.291601, "lon"=>-84.475525),
"/US/US-OH/springfield-16641" => array("type"=>"city", "pop"=>62269, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>39.926800, "lon"=>-83.796586),
"/US/US-OH/st.-bernard-16574" => array("type"=>"city", "pop"=>4626, "name_en"=>"St. Bernard", "lang"=>"en", "lat"=>39.170968, "lon"=>-84.495336),
"/US/US-OH/st.-clairsville-16575" => array("type"=>"city", "pop"=>5053, "name_en"=>"St. Clairsville", "lang"=>"en", "lat"=>40.078141, "lon"=>-80.899001),
"/US/US-OH/st.-henry-16576" => array("type"=>"village", "pop"=>2408, "name_en"=>"St. Henry", "lang"=>"en", "lat"=>40.420855, "lon"=>-84.633612),
"/US/US-OH/st.-louisville-16577" => array("type"=>"village", "pop"=>336, "name_en"=>"St. Louisville", "lang"=>"en", "lat"=>40.171521, "lon"=>-82.418986),
"/US/US-OH/st.-martin-16578" => array("type"=>"village", "pop"=>109, "name_ru"=>"Сент Мартин", "name_en"=>"St. Martin", "lang"=>"en", "lat"=>39.214666, "lon"=>-83.887865),
"/US/US-OH/st.-marys-16579" => array("type"=>"city", "pop"=>8138, "name_ru"=>"Сент Мэрис", "name_en"=>"St. Marys", "lang"=>"en", "lat"=>40.543929, "lon"=>-84.390866),
"/US/US-OH/st.-paris-16580" => array("type"=>"village", "pop"=>1965, "name_en"=>"St. Paris", "lang"=>"en", "lat"=>40.127912, "lon"=>-83.957964),
"/US/US-OH/stafford-16643" => array("type"=>"village", "pop"=>82, "name_ru"=>"Стэффорд", "name_en"=>"Stafford", "lang"=>"en", "lat"=>39.713154, "lon"=>-81.277599),
"/US/US-OH/steubenville-16644" => array("type"=>"city", "pop"=>18820, "name_en"=>"Steubenville", "lang"=>"en", "lat"=>40.367000, "lon"=>-80.646559),
"/US/US-OH/stockport-16645" => array("type"=>"village", "pop"=>524, "name_ru"=>"Стокпорт", "name_en"=>"Stockport", "lang"=>"en", "lat"=>39.549276, "lon"=>-81.794977),
"/US/US-OH/stone-creek-16646" => array("type"=>"village", "pop"=>180, "name_en"=>"Stone Creek", "lang"=>"en", "lat"=>40.398707, "lon"=>-81.558687),
"/US/US-OH/stony-prairie-16647" => array("type"=>"cdp", "pop"=>1536, "name_en"=>"Stony Prairie", "lang"=>"en", "lat"=>41.351301, "lon"=>-83.155202),
"/US/US-OH/stoutsville-16648" => array("type"=>"village", "pop"=>572, "name_en"=>"Stoutsville", "lang"=>"en", "lat"=>39.606577, "lon"=>-82.824553),
"/US/US-OH/stow-16649" => array("type"=>"city", "pop"=>33899, "name_en"=>"Stow", "lang"=>"en", "lat"=>41.176567, "lon"=>-81.436398),
"/US/US-OH/strasburg-16650" => array("type"=>"village", "pop"=>2676, "name_ru"=>"Страсбург", "name_en"=>"Strasburg", "lang"=>"en", "lat"=>40.602340, "lon"=>-81.529424),
"/US/US-OH/stratton-16651" => array("type"=>"village", "pop"=>273, "name_ru"=>"Стрэттон", "name_en"=>"Stratton", "lang"=>"en", "lat"=>40.524944, "lon"=>-80.630421),
"/US/US-OH/streetsboro-16652" => array("type"=>"city", "pop"=>14488, "name_en"=>"Streetsboro", "lang"=>"en", "lat"=>41.238850, "lon"=>-81.345799),
"/US/US-OH/strongsville-16653" => array("type"=>"city", "pop"=>42617, "name_en"=>"Strongsville", "lang"=>"en", "lat"=>41.312700, "lon"=>-81.832120),
"/US/US-OH/struthers-16654" => array("type"=>"city", "pop"=>10734, "name_en"=>"Struthers", "lang"=>"en", "lat"=>41.051200, "lon"=>-80.593124),
"/US/US-OH/stryker-16655" => array("type"=>"village", "pop"=>1354, "name_en"=>"Stryker", "lang"=>"en", "lat"=>41.503200, "lon"=>-84.417605),
"/US/US-OH/sugar-bush-knolls-16656" => array("type"=>"village", "pop"=>223, "name_en"=>"Sugar Bush Knolls", "lang"=>"en", "lat"=>41.204850, "lon"=>-81.346733),
"/US/US-OH/sugar-grove-16658" => array("type"=>"village", "pop"=>444, "name_ru"=>"Шугар Гров", "name_en"=>"Sugar Grove", "lang"=>"en", "lat"=>39.626682, "lon"=>-82.547145),
"/US/US-OH/sugarcreek-16657" => array("type"=>"village", "pop"=>2193, "name_en"=>"Sugarcreek", "lang"=>"en", "lat"=>40.506572, "lon"=>-81.641881),
"/US/US-OH/summerfield-16659" => array("type"=>"village", "pop"=>287, "name_ru"=>"Саммерфилд", "name_en"=>"Summerfield", "lang"=>"en", "lat"=>39.797696, "lon"=>-81.335910),
"/US/US-OH/summerside-16660" => array("type"=>"cdp", "pop"=>4573, "name_en"=>"Summerside", "lang"=>"en", "lat"=>39.122834, "lon"=>-84.286445),
"/US/US-OH/summit-station-16661" => array("type"=>"cdp", "pop"=>1380, "name_en"=>"Summit Station", "lang"=>"en", "lat"=>40.004843, "lon"=>-82.750799),
"/US/US-OH/summitville-16662" => array("type"=>"village", "pop"=>103, "name_ru"=>"Саммитвилл", "name_en"=>"Summitville", "lang"=>"en", "lat"=>40.677201, "lon"=>-80.886078),
"/US/US-OH/sunbury-16663" => array("type"=>"village", "pop"=>3375, "name_en"=>"Sunbury", "lang"=>"en", "lat"=>40.245072, "lon"=>-82.864783),
"/US/US-OH/swanton-16664" => array("type"=>"village", "pop"=>3651, "name_en"=>"Swanton", "lang"=>"en", "lat"=>41.583843, "lon"=>-83.891658),
"/US/US-OH/sycamore-16665" => array("type"=>"village", "pop"=>867, "name_ru"=>"Сикамор", "name_en"=>"Sycamore", "lang"=>"en", "lat"=>40.951008, "lon"=>-83.170898),
"/US/US-OH/sylvania-16666" => array("type"=>"city", "pop"=>19185, "name_ru"=>"Сильвания", "name_en"=>"Sylvania", "lang"=>"en", "lat"=>41.714650, "lon"=>-83.706026),
"/US/US-OH/syracuse-16667" => array("type"=>"village", "pop"=>856, "name_ru"=>"Сиракьюс", "name_en"=>"Syracuse", "lang"=>"en", "lat"=>38.999679, "lon"=>-81.972969),
"/US/US-OH/tallmadge-16668" => array("type"=>"city", "pop"=>17293, "name_en"=>"Tallmadge", "lang"=>"en", "lat"=>41.103350, "lon"=>-81.423222),
"/US/US-OH/tarlton-16669" => array("type"=>"village", "pop"=>308, "name_en"=>"Tarlton", "lang"=>"en", "lat"=>39.554275, "lon"=>-82.777372),
"/US/US-OH/terrace-park-16670" => array("type"=>"village", "pop"=>2220, "name_en"=>"Terrace Park", "lang"=>"en", "lat"=>39.157951, "lon"=>-84.312117),
"/US/US-OH/the-plains-16671" => array("type"=>"cdp", "pop"=>2644, "name_en"=>"The Plains", "lang"=>"en", "lat"=>39.365569, "lon"=>-82.133985),
"/US/US-OH/the-village-of-indian-hill-16672" => array("type"=>"city", "pop"=>6030, "name_en"=>"The Village of Indian Hill", "lang"=>"en", "lat"=>39.189451, "lon"=>-84.334831),
"/US/US-OH/thornville-16673" => array("type"=>"village", "pop"=>1138, "name_en"=>"Thornville", "lang"=>"en", "lat"=>39.895844, "lon"=>-82.419204),
"/US/US-OH/thurston-16674" => array("type"=>"village", "pop"=>602, "name_en"=>"Thurston", "lang"=>"en", "lat"=>39.841275, "lon"=>-82.544982),
"/US/US-OH/tiffin-16675" => array("type"=>"city", "pop"=>17347, "name_ru"=>"Тиффин", "name_en"=>"Tiffin", "lang"=>"en", "lat"=>41.116728, "lon"=>-83.177502),
"/US/US-OH/tiltonsville-16676" => array("type"=>"village", "pop"=>1201, "name_en"=>"Tiltonsville", "lang"=>"en", "lat"=>40.172537, "lon"=>-80.697682),
"/US/US-OH/timberlake-16677" => array("type"=>"village", "pop"=>735, "name_en"=>"Timberlake", "lang"=>"en", "lat"=>41.664600, "lon"=>-81.442857),
"/US/US-OH/tipp-city-16678" => array("type"=>"city", "pop"=>9254, "name_en"=>"Tipp City", "lang"=>"en", "lat"=>39.963821, "lon"=>-84.183762),
"/US/US-OH/tiro-16679" => array("type"=>"village", "pop"=>251, "name_en"=>"Tiro", "lang"=>"en", "lat"=>40.906031, "lon"=>-82.768307),
"/US/US-OH/toledo-16680" => array("type"=>"city", "pop"=>293201, "name_ru"=>"Толедо", "name_en"=>"Toledo", "lang"=>"en", "lat"=>41.663950, "lon"=>-83.581649),
"/US/US-OH/tontogany-16681" => array("type"=>"village", "pop"=>371, "name_en"=>"Tontogany", "lang"=>"en", "lat"=>41.420100, "lon"=>-83.740114),
"/US/US-OH/toronto-16682" => array("type"=>"city", "pop"=>5221, "name_ru"=>"Торонто", "name_en"=>"Toronto", "lang"=>"en", "lat"=>40.458900, "lon"=>-80.606277),
"/US/US-OH/tremont-city-16683" => array("type"=>"village", "pop"=>342, "name_en"=>"Tremont City", "lang"=>"en", "lat"=>40.014654, "lon"=>-83.836072),
"/US/US-OH/trenton-16684" => array("type"=>"city", "pop"=>10866, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>39.477050, "lon"=>-84.462334),
"/US/US-OH/trimble-16685" => array("type"=>"village", "pop"=>480, "name_en"=>"Trimble", "lang"=>"en", "lat"=>39.486392, "lon"=>-82.079916),
"/US/US-OH/trotwood-16686" => array("type"=>"city", "pop"=>26039, "name_en"=>"Trotwood", "lang"=>"en", "lat"=>39.802800, "lon"=>-84.304540),
"/US/US-OH/troy-16687" => array("type"=>"city", "pop"=>21975, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>40.044407, "lon"=>-84.219211),
"/US/US-OH/turnpike-interchange-16688" => array("type"=>"cdp", "pop"=>1188, "name_en"=>"Turnpike Interchange", "lang"=>"en", "lat"=>41.216373, "lon"=>-80.934352),
"/US/US-OH/turpin-hills-16689" => array("type"=>"cdp", "pop"=>4927, "name_en"=>"Turpin Hills", "lang"=>"en", "lat"=>39.107101, "lon"=>-84.373199),
"/US/US-OH/tuscarawas-16690" => array("type"=>"village", "pop"=>981, "name_en"=>"Tuscarawas", "lang"=>"en", "lat"=>40.397456, "lon"=>-81.404215),
"/US/US-OH/twinsburg-16691" => array("type"=>"city", "pop"=>17394, "name_en"=>"Twinsburg", "lang"=>"en", "lat"=>41.324200, "lon"=>-81.452886),
"/US/US-OH/uhrichsville-16692" => array("type"=>"city", "pop"=>5516, "name_en"=>"Uhrichsville", "lang"=>"en", "lat"=>40.400056, "lon"=>-81.350849),
"/US/US-OH/union-16693" => array("type"=>"city", "pop"=>6293, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>39.902698, "lon"=>-84.309920),
"/US/US-OH/union-city-16694" => array("type"=>"village", "pop"=>1649, "name_ru"=>"Юнион Сити", "name_en"=>"Union City", "lang"=>"en", "lat"=>40.198151, "lon"=>-84.794789),
"/US/US-OH/uniontown-16695" => array("type"=>"cdp", "pop"=>3074, "name_ru"=>"Юнионтаун", "name_en"=>"Uniontown", "lang"=>"en", "lat"=>40.973501, "lon"=>-81.403452),
"/US/US-OH/unionville-center-16696" => array("type"=>"village", "pop"=>286, "name_en"=>"Unionville Center", "lang"=>"en", "lat"=>40.136276, "lon"=>-83.341582),
"/US/US-OH/uniopolis-16697" => array("type"=>"village", "pop"=>247, "name_en"=>"Uniopolis", "lang"=>"en", "lat"=>40.601936, "lon"=>-84.086592),
"/US/US-OH/university-heights-16698" => array("type"=>"city", "pop"=>12519, "name_ru"=>"Юниверсити Хайтс", "name_en"=>"University Heights", "lang"=>"en", "lat"=>41.494700, "lon"=>-81.535521),
"/US/US-OH/upper-arlington-16699" => array("type"=>"city", "pop"=>31878, "name_en"=>"Upper Arlington", "lang"=>"en", "lat"=>40.026550, "lon"=>-83.070511),
"/US/US-OH/upper-sandusky-16700" => array("type"=>"city", "pop"=>6364, "name_en"=>"Upper Sandusky", "lang"=>"en", "lat"=>40.830959, "lon"=>-83.280756),
"/US/US-OH/urbana-16701" => array("type"=>"city", "pop"=>11442, "name_ru"=>"Урбана", "name_en"=>"Urbana", "lang"=>"en", "lat"=>40.108768, "lon"=>-83.751693),
"/US/US-OH/urbancrest-16702" => array("type"=>"village", "pop"=>879, "name_en"=>"Urbancrest", "lang"=>"en", "lat"=>39.899400, "lon"=>-83.088977),
"/US/US-OH/utica-16703" => array("type"=>"village", "pop"=>2104, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>40.233450, "lon"=>-82.443585),
"/US/US-OH/valley-hi-16704" => array("type"=>"village", "pop"=>231, "name_en"=>"Valley Hi", "lang"=>"en", "lat"=>40.316146, "lon"=>-83.673041),
"/US/US-OH/valley-view-16705" => array("type"=>"village", "pop"=>2012, "name_en"=>"Valley View", "lang"=>"en", "lat"=>41.386333, "lon"=>-81.606931),
"/US/US-OH/valleyview-16706" => array("type"=>"village", "pop"=>569, "name_en"=>"Valleyview", "lang"=>"en", "lat"=>39.964800, "lon"=>-83.072853),
"/US/US-OH/van-buren-16707" => array("type"=>"village", "pop"=>316, "name_ru"=>"Ван Бурен", "name_en"=>"Van Buren", "lang"=>"en", "lat"=>41.138961, "lon"=>-83.649183),
"/US/US-OH/van-wert-16710" => array("type"=>"city", "pop"=>10211, "name_ru"=>"Ван Верт", "name_en"=>"Van Wert", "lang"=>"en", "lat"=>40.864614, "lon"=>-84.581675),
"/US/US-OH/vandalia-16708" => array("type"=>"city", "pop"=>14124, "name_ru"=>"Вандалия", "name_en"=>"Vandalia", "lang"=>"en", "lat"=>39.878131, "lon"=>-84.194022),
"/US/US-OH/vanlue-16709" => array("type"=>"village", "pop"=>332, "name_en"=>"Vanlue", "lang"=>"en", "lat"=>40.976425, "lon"=>-83.481792),
"/US/US-OH/venedocia-16711" => array("type"=>"village", "pop"=>149, "name_en"=>"Venedocia", "lang"=>"en", "lat"=>40.785427, "lon"=>-84.455455),
"/US/US-OH/vermilion-16712" => array("type"=>"city", "pop"=>10754, "name_ru"=>"Вермилион", "name_en"=>"Vermilion", "lang"=>"en", "lat"=>41.411967, "lon"=>-82.318228),
"/US/US-OH/verona-16713" => array("type"=>"village", "pop"=>443, "name_ru"=>"Верона", "name_en"=>"Verona", "lang"=>"en", "lat"=>39.902932, "lon"=>-84.486531),
"/US/US-OH/versailles-16714" => array("type"=>"village", "pop"=>2521, "name_ru"=>"Версайллес", "name_en"=>"Versailles", "lang"=>"en", "lat"=>40.222960, "lon"=>-84.484027),
"/US/US-OH/vienna-center-16715" => array("type"=>"cdp", "pop"=>1067, "name_en"=>"Vienna Center", "lang"=>"en", "lat"=>41.255064, "lon"=>-80.677092),
"/US/US-OH/vinton-16716" => array("type"=>"village", "pop"=>338, "name_ru"=>"Винтон", "name_en"=>"Vinton", "lang"=>"en", "lat"=>38.977414, "lon"=>-82.340295),
"/US/US-OH/wadsworth-16717" => array("type"=>"city", "pop"=>20696, "name_ru"=>"Уодсворт", "name_en"=>"Wadsworth", "lang"=>"en", "lat"=>41.023776, "lon"=>-81.731929),
"/US/US-OH/waite-hill-16718" => array("type"=>"village", "pop"=>553, "name_en"=>"Waite Hill", "lang"=>"en", "lat"=>41.609800, "lon"=>-81.387968),
"/US/US-OH/wakeman-16719" => array("type"=>"village", "pop"=>929, "name_en"=>"Wakeman", "lang"=>"en", "lat"=>41.255900, "lon"=>-82.403974),
"/US/US-OH/walbridge-16720" => array("type"=>"village", "pop"=>3089, "name_en"=>"Walbridge", "lang"=>"en", "lat"=>41.585600, "lon"=>-83.489224),
"/US/US-OH/waldo-16721" => array("type"=>"village", "pop"=>311, "name_ru"=>"Уолдо", "name_en"=>"Waldo", "lang"=>"en", "lat"=>40.461590, "lon"=>-83.085573),
"/US/US-OH/walton-hills-16722" => array("type"=>"village", "pop"=>2278, "name_en"=>"Walton Hills", "lang"=>"en", "lat"=>41.364250, "lon"=>-81.554535),
"/US/US-OH/wapakoneta-16723" => array("type"=>"city", "pop"=>9438, "name_en"=>"Wapakoneta", "lang"=>"en", "lat"=>40.570200, "lon"=>-84.192224),
"/US/US-OH/warren-16724" => array("type"=>"city", "pop"=>43789, "name_ru"=>"Уоррен", "name_en"=>"Warren", "lang"=>"en", "lat"=>41.238050, "lon"=>-80.816899),
"/US/US-OH/warrensville-heights-16725" => array("type"=>"city", "pop"=>13618, "name_en"=>"Warrensville Heights", "lang"=>"en", "lat"=>41.438600, "lon"=>-81.523419),
"/US/US-OH/warsaw-16726" => array("type"=>"village", "pop"=>769, "name_ru"=>"Варшава", "name_en"=>"Warsaw", "lang"=>"en", "lat"=>40.335332, "lon"=>-82.001435),
"/US/US-OH/washington-16727" => array("type"=>"city", "pop"=>2313, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>39.537705, "lon"=>-83.431130),
"/US/US-OH/washingtonville-16728" => array("type"=>"village", "pop"=>743, "name_en"=>"Washingtonville", "lang"=>"en", "lat"=>40.900152, "lon"=>-80.767916),
"/US/US-OH/waterville-16729" => array("type"=>"village", "pop"=>5230, "name_ru"=>"Уотервилл", "name_en"=>"Waterville", "lang"=>"en", "lat"=>41.497983, "lon"=>-83.733292),
"/US/US-OH/wauseon-16730" => array("type"=>"city", "pop"=>7246, "name_en"=>"Wauseon", "lang"=>"en", "lat"=>41.553521, "lon"=>-84.141939),
"/US/US-OH/waverly-city-16731" => array("type"=>"city", "pop"=>4437, "name_en"=>"Waverly City", "lang"=>"en", "lat"=>39.125336, "lon"=>-82.982462),
"/US/US-OH/wayne-16732" => array("type"=>"village", "pop"=>855, "name_ru"=>"Уэйн", "name_en"=>"Wayne", "lang"=>"en", "lat"=>41.301237, "lon"=>-83.472352),
"/US/US-OH/wayne-lakes-16733" => array("type"=>"village", "pop"=>646, "name_en"=>"Wayne Lakes", "lang"=>"en", "lat"=>40.021135, "lon"=>-84.662380),
"/US/US-OH/waynesburg-16734" => array("type"=>"village", "pop"=>969, "name_en"=>"Waynesburg", "lang"=>"en", "lat"=>40.668100, "lon"=>-81.259209),
"/US/US-OH/waynesfield-16735" => array("type"=>"village", "pop"=>785, "name_en"=>"Waynesfield", "lang"=>"en", "lat"=>40.601826, "lon"=>-83.973769),
"/US/US-OH/waynesville-16736" => array("type"=>"village", "pop"=>3055, "name_ru"=>"Уэйнсвилл", "name_en"=>"Waynesville", "lang"=>"en", "lat"=>39.534107, "lon"=>-84.086603),
"/US/US-OH/wellington-16737" => array("type"=>"village", "pop"=>4699, "name_ru"=>"Веллингтон", "name_en"=>"Wellington", "lang"=>"en", "lat"=>41.164600, "lon"=>-82.223610),
"/US/US-OH/wellston-16738" => array("type"=>"city", "pop"=>5936, "name_en"=>"Wellston", "lang"=>"en", "lat"=>39.117127, "lon"=>-82.538339),
"/US/US-OH/wellsville-16739" => array("type"=>"city", "pop"=>3893, "name_ru"=>"Уэллсвилл", "name_en"=>"Wellsville", "lang"=>"en", "lat"=>40.602720, "lon"=>-80.656329),
"/US/US-OH/west-alexandria-16740" => array("type"=>"village", "pop"=>1298, "name_en"=>"West Alexandria", "lang"=>"en", "lat"=>39.742864, "lon"=>-84.533555),
"/US/US-OH/west-carrollton-city-16741" => array("type"=>"city", "pop"=>12761, "name_en"=>"West Carrollton City", "lang"=>"en", "lat"=>39.669500, "lon"=>-84.252524),
"/US/US-OH/west-elkton-16742" => array("type"=>"village", "pop"=>173, "name_en"=>"West Elkton", "lang"=>"en", "lat"=>39.588800, "lon"=>-84.554327),
"/US/US-OH/west-farmington-16744" => array("type"=>"village", "pop"=>482, "name_en"=>"West Farmington", "lang"=>"en", "lat"=>41.391024, "lon"=>-80.973455),
"/US/US-OH/west-hill-16746" => array("type"=>"cdp", "pop"=>2954, "name_en"=>"West Hill", "lang"=>"en", "lat"=>41.230933, "lon"=>-80.527402),
"/US/US-OH/west-lafayette-16747" => array("type"=>"village", "pop"=>2469, "name_ru"=>"Уэст Лафайет", "name_en"=>"West Lafayette", "lang"=>"en", "lat"=>40.273886, "lon"=>-81.751009),
"/US/US-OH/west-leipsic-16749" => array("type"=>"village", "pop"=>260, "name_en"=>"West Leipsic", "lang"=>"en", "lat"=>41.105394, "lon"=>-84.001831),
"/US/US-OH/west-liberty-16750" => array("type"=>"village", "pop"=>1729, "name_ru"=>"Вэст Либерти", "name_en"=>"West Liberty", "lang"=>"en", "lat"=>40.253837, "lon"=>-83.757961),
"/US/US-OH/west-manchester-16751" => array("type"=>"village", "pop"=>398, "name_en"=>"West Manchester", "lang"=>"en", "lat"=>39.902304, "lon"=>-84.626247),
"/US/US-OH/west-mansfield-16752" => array("type"=>"village", "pop"=>696, "name_en"=>"West Mansfield", "lang"=>"en", "lat"=>40.399631, "lon"=>-83.545012),
"/US/US-OH/west-millgrove-16753" => array("type"=>"village", "pop"=>77, "name_en"=>"West Millgrove", "lang"=>"en", "lat"=>41.243777, "lon"=>-83.492149),
"/US/US-OH/west-milton-16754" => array("type"=>"village", "pop"=>4613, "name_en"=>"West Milton", "lang"=>"en", "lat"=>39.952808, "lon"=>-84.325088),
"/US/US-OH/west-portsmouth-16756" => array("type"=>"cdp", "pop"=>3551, "name_en"=>"West Portsmouth", "lang"=>"en", "lat"=>38.762847, "lon"=>-83.038294),
"/US/US-OH/west-rushville-16757" => array("type"=>"village", "pop"=>137, "name_en"=>"West Rushville", "lang"=>"en", "lat"=>39.762697, "lon"=>-82.448558),
"/US/US-OH/west-salem-16758" => array("type"=>"village", "pop"=>1464, "name_ru"=>"Уэст Салем", "name_en"=>"West Salem", "lang"=>"en", "lat"=>40.969611, "lon"=>-82.107790),
"/US/US-OH/west-union-16759" => array("type"=>"village", "pop"=>3085, "name_ru"=>"Вэст Юнион", "name_en"=>"West Union", "lang"=>"en", "lat"=>38.793151, "lon"=>-83.543658),
"/US/US-OH/west-unity-16760" => array("type"=>"village", "pop"=>1745, "name_en"=>"West Unity", "lang"=>"en", "lat"=>41.585967, "lon"=>-84.434216),
"/US/US-OH/westerville-16743" => array("type"=>"city", "pop"=>36231, "name_en"=>"Westerville", "lang"=>"en", "lat"=>40.121200, "lon"=>-82.917604),
"/US/US-OH/westfield-center-16745" => array("type"=>"village", "pop"=>1170, "name_en"=>"Westfield Center", "lang"=>"en", "lat"=>41.026629, "lon"=>-81.933441),
"/US/US-OH/westlake-16748" => array("type"=>"city", "pop"=>30546, "name_ru"=>"Вэстлейк", "name_en"=>"Westlake", "lang"=>"en", "lat"=>41.454395, "lon"=>-81.928795),
"/US/US-OH/weston-16755" => array("type"=>"village", "pop"=>1650, "name_ru"=>"Вэстон", "name_en"=>"Weston", "lang"=>"en", "lat"=>41.345755, "lon"=>-83.794605),
"/US/US-OH/wharton-16761" => array("type"=>"village", "pop"=>384, "name_en"=>"Wharton", "lang"=>"en", "lat"=>40.862026, "lon"=>-83.464728),
"/US/US-OH/wheelersburg-16762" => array("type"=>"cdp", "pop"=>5113, "name_en"=>"Wheelersburg", "lang"=>"en", "lat"=>38.734079, "lon"=>-82.845343),
"/US/US-OH/white-oak-16765" => array("type"=>"cdp", "pop"=>12430, "name_ru"=>"Уайт Оук", "name_en"=>"White Oak", "lang"=>"en", "lat"=>39.220168, "lon"=>-84.605414),
"/US/US-OH/white-oak-east-16766" => array("type"=>"cdp", "pop"=>3544, "name_en"=>"White Oak East", "lang"=>"en", "lat"=>39.202034, "lon"=>-84.590169),
"/US/US-OH/white-oak-west-16767" => array("type"=>"cdp", "pop"=>2879, "name_en"=>"White Oak West", "lang"=>"en", "lat"=>39.199298, "lon"=>-84.614260),
"/US/US-OH/whitehall-16763" => array("type"=>"city", "pop"=>18180, "name_ru"=>"Уайтхолл", "name_en"=>"Whitehall", "lang"=>"en", "lat"=>39.968605, "lon"=>-82.883716),
"/US/US-OH/whitehouse-16764" => array("type"=>"village", "pop"=>3830, "name_en"=>"Whitehouse", "lang"=>"en", "lat"=>41.518850, "lon"=>-83.802621),
"/US/US-OH/wickliffe-16768" => array("type"=>"city", "pop"=>12983, "name_ru"=>"Уиклайфф", "name_en"=>"Wickliffe", "lang"=>"en", "lat"=>41.608100, "lon"=>-81.467529),
"/US/US-OH/wilberforce-16769" => array("type"=>"cdp", "pop"=>2639, "name_en"=>"Wilberforce", "lang"=>"en", "lat"=>39.716544, "lon"=>-83.883607),
"/US/US-OH/wilkesville-16770" => array("type"=>"village", "pop"=>161, "name_en"=>"Wilkesville", "lang"=>"en", "lat"=>39.076522, "lon"=>-82.326632),
"/US/US-OH/willard-16771" => array("type"=>"city", "pop"=>6664, "name_ru"=>"Уиллард", "name_en"=>"Willard", "lang"=>"en", "lat"=>41.050833, "lon"=>-82.722557),
"/US/US-OH/williamsburg-16772" => array("type"=>"village", "pop"=>2382, "name_ru"=>"Уилльямсбург", "name_en"=>"Williamsburg", "lang"=>"en", "lat"=>39.053123, "lon"=>-84.050654),
"/US/US-OH/williamsport-16773" => array("type"=>"village", "pop"=>1025, "name_ru"=>"Уильямспорт", "name_en"=>"Williamsport", "lang"=>"en", "lat"=>39.583413, "lon"=>-83.116700),
"/US/US-OH/willoughby-16774" => array("type"=>"city", "pop"=>22561, "name_en"=>"Willoughby", "lang"=>"en", "lat"=>41.646267, "lon"=>-81.409412),
"/US/US-OH/willoughby-hills-16775" => array("type"=>"city", "pop"=>8540, "name_en"=>"Willoughby Hills", "lang"=>"en", "lat"=>41.584800, "lon"=>-81.435155),
"/US/US-OH/willowick-16776" => array("type"=>"city", "pop"=>13727, "name_en"=>"Willowick", "lang"=>"en", "lat"=>41.635033, "lon"=>-81.468452),
"/US/US-OH/willshire-16777" => array("type"=>"village", "pop"=>438, "name_en"=>"Willshire", "lang"=>"en", "lat"=>40.746606, "lon"=>-84.792050),
"/US/US-OH/wilmington-16778" => array("type"=>"city", "pop"=>12603, "name_ru"=>"Уилмингтон", "name_en"=>"Wilmington", "lang"=>"en", "lat"=>39.443622, "lon"=>-83.830897),
"/US/US-OH/wilmot-16779" => array("type"=>"village", "pop"=>338, "name_ru"=>"Уилмот", "name_en"=>"Wilmot", "lang"=>"en", "lat"=>40.656051, "lon"=>-81.635032),
"/US/US-OH/wilson-16780" => array("type"=>"village", "pop"=>115, "name_ru"=>"Уилсон", "name_en"=>"Wilson", "lang"=>"en", "lat"=>39.858220, "lon"=>-81.067777),
"/US/US-OH/winchester-16781" => array("type"=>"village", "pop"=>1077, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>38.942638, "lon"=>-83.653555),
"/US/US-OH/windham-16782" => array("type"=>"village", "pop"=>2724, "name_en"=>"Windham", "lang"=>"en", "lat"=>41.236200, "lon"=>-81.036360),
"/US/US-OH/wintersville-16783" => array("type"=>"village", "pop"=>3877, "name_en"=>"Wintersville", "lang"=>"en", "lat"=>40.378983, "lon"=>-80.706582),
"/US/US-OH/withamsville-16784" => array("type"=>"cdp", "pop"=>2834, "name_en"=>"Withamsville", "lang"=>"en", "lat"=>39.058601, "lon"=>-84.269965),
"/US/US-OH/woodbourne-hyde-park-16785" => array("type"=>"cdp", "pop"=>7837, "name_en"=>"Woodbourne-Hyde Park", "lang"=>"en", "lat"=>39.660717, "lon"=>-84.173625),
"/US/US-OH/woodlawn-16786" => array("type"=>"village", "pop"=>2645, "name_ru"=>"Вудлон", "name_en"=>"Woodlawn", "lang"=>"en", "lat"=>39.256668, "lon"=>-84.467508),
"/US/US-OH/woodmere-16787" => array("type"=>"village", "pop"=>755, "name_en"=>"Woodmere", "lang"=>"en", "lat"=>41.459200, "lon"=>-81.479369),
"/US/US-OH/woodsfield-16788" => array("type"=>"village", "pop"=>2424, "name_en"=>"Woodsfield", "lang"=>"en", "lat"=>39.762954, "lon"=>-81.116734),
"/US/US-OH/woodstock-16789" => array("type"=>"village", "pop"=>309, "name_ru"=>"Вудсток", "name_en"=>"Woodstock", "lang"=>"en", "lat"=>40.173453, "lon"=>-83.527946),
"/US/US-OH/woodville-16790" => array("type"=>"village", "pop"=>1959, "name_ru"=>"Вудвилл", "name_en"=>"Woodville", "lang"=>"en", "lat"=>41.451294, "lon"=>-83.363014),
"/US/US-OH/wooster-16791" => array("type"=>"city", "pop"=>26212, "name_ru"=>"Вустер", "name_en"=>"Wooster", "lang"=>"en", "lat"=>40.818469, "lon"=>-81.932870),
"/US/US-OH/worthington-16792" => array("type"=>"city", "pop"=>13314, "name_ru"=>"Уортингтон", "name_en"=>"Worthington", "lang"=>"en", "lat"=>40.097733, "lon"=>-83.020854),
"/US/US-OH/wren-16793" => array("type"=>"village", "pop"=>197, "name_en"=>"Wren", "lang"=>"en", "lat"=>40.800262, "lon"=>-84.774134),
"/US/US-OH/wright-patterson-afb-16794" => array("type"=>"cdp", "pop"=>8579, "name_en"=>"Wright-Patterson AFB", "lang"=>"en", "lat"=>39.801584, "lon"=>-84.070798),
"/US/US-OH/wyoming-16795" => array("type"=>"city", "pop"=>8355, "name_ru"=>"Вайоминг", "name_en"=>"Wyoming", "lang"=>"en", "lat"=>39.229451, "lon"=>-84.481598),
"/US/US-OH/xenia-16796" => array("type"=>"city", "pop"=>27357, "name_ru"=>"Ксениа", "name_en"=>"Xenia", "lang"=>"en", "lat"=>39.685027, "lon"=>-83.940808),
"/US/US-OH/yankee-lake-16797" => array("type"=>"village", "pop"=>95, "name_en"=>"Yankee Lake", "lang"=>"en", "lat"=>41.271407, "lon"=>-80.570659),
"/US/US-OH/yellow-springs-16798" => array("type"=>"village", "pop"=>3427, "name_en"=>"Yellow Springs", "lang"=>"en", "lat"=>39.799621, "lon"=>-83.894596),
"/US/US-OH/yorkshire-16799" => array("type"=>"village", "pop"=>106, "name_en"=>"Yorkshire", "lang"=>"en", "lat"=>40.325495, "lon"=>-84.495990),
"/US/US-OH/yorkville-16800" => array("type"=>"village", "pop"=>1165, "name_ru"=>"Йорквилл", "name_en"=>"Yorkville", "lang"=>"en", "lat"=>40.152544, "lon"=>-80.708095),
"/US/US-OH/youngstown-16801" => array("type"=>"city", "pop"=>72925, "name_en"=>"Youngstown", "lang"=>"en", "lat"=>41.099100, "lon"=>-80.646389),
"/US/US-OH/zaleski-16802" => array("type"=>"village", "pop"=>401, "name_en"=>"Zaleski", "lang"=>"en", "lat"=>39.280639, "lon"=>-82.396276),
"/US/US-OH/zanesfield-16803" => array("type"=>"village", "pop"=>203, "name_en"=>"Zanesfield", "lang"=>"en", "lat"=>40.338188, "lon"=>-83.678059),
"/US/US-OH/zanesville-16804" => array("type"=>"city", "pop"=>25129, "name_en"=>"Zanesville", "lang"=>"en", "lat"=>39.952438, "lon"=>-82.011983),
"/US/US-OH/zoar-16805" => array("type"=>"village", "pop"=>191, "name_en"=>"Zoar", "lang"=>"en", "lat"=>40.613968, "lon"=>-81.424321),
"/US/US-OK" => array("type"=>"state", "pop"=>3642361, "name_ru"=>"Оклахома", "name_en"=>"Oklahoma", "lang"=>"en"),
"/US/US-OK/achille-16806" => array("type"=>"town", "pop"=>534, "name_en"=>"Achille", "lang"=>"en", "lat"=>33.834539, "lon"=>-96.389919),
"/US/US-OK/ada-16807" => array("type"=>"city", "pop"=>16729, "name_ru"=>"Ада", "name_en"=>"Ada", "lang"=>"en", "lat"=>34.777005, "lon"=>-96.660408),
"/US/US-OK/adair-16808" => array("type"=>"town", "pop"=>716, "name_ru"=>"Адэр", "name_en"=>"Adair", "lang"=>"en", "lat"=>36.437829, "lon"=>-95.265756),
"/US/US-OK/addington-16809" => array("type"=>"town", "pop"=>107, "name_en"=>"Addington", "lang"=>"en", "lat"=>34.243116, "lon"=>-97.966279),
"/US/US-OK/afton-16810" => array("type"=>"town", "pop"=>1081, "name_ru"=>"Афтон", "name_en"=>"Afton", "lang"=>"en", "lat"=>36.692265, "lon"=>-94.967726),
"/US/US-OK/agra-16811" => array("type"=>"town", "pop"=>356, "name_ru"=>"Агра", "name_en"=>"Agra", "lang"=>"en", "lat"=>35.895110, "lon"=>-96.870108),
"/US/US-OK/albion-16812" => array("type"=>"town", "pop"=>143, "name_ru"=>"Албион", "name_en"=>"Albion", "lang"=>"en", "lat"=>34.661965, "lon"=>-95.099201),
"/US/US-OK/alderson-16813" => array("type"=>"town", "pop"=>267, "name_en"=>"Alderson", "lang"=>"en", "lat"=>34.900747, "lon"=>-95.690963),
"/US/US-OK/alex-16814" => array("type"=>"town", "pop"=>662, "name_en"=>"Alex", "lang"=>"en", "lat"=>34.920165, "lon"=>-97.779397),
"/US/US-OK/aline-16815" => array("type"=>"town", "pop"=>190, "name_en"=>"Aline", "lang"=>"en", "lat"=>36.509002, "lon"=>-98.448306),
"/US/US-OK/allen-16816" => array("type"=>"town", "pop"=>993, "name_ru"=>"Аллен", "name_en"=>"Allen", "lang"=>"en", "lat"=>34.878321, "lon"=>-96.414040),
"/US/US-OK/altus-16817" => array("type"=>"city", "pop"=>18912, "name_ru"=>"Алтус", "name_en"=>"Altus", "lang"=>"en", "lat"=>34.649190, "lon"=>-99.312970),
"/US/US-OK/alva-16818" => array("type"=>"city", "pop"=>4848, "name_ru"=>"Алва", "name_en"=>"Alva", "lang"=>"en", "lat"=>36.801924, "lon"=>-98.667226),
"/US/US-OK/amber-16819" => array("type"=>"town", "pop"=>536, "name_en"=>"Amber", "lang"=>"en", "lat"=>35.161417, "lon"=>-97.880507),
"/US/US-OK/ames-16820" => array("type"=>"town", "pop"=>189, "name_ru"=>"Эймс", "name_en"=>"Ames", "lang"=>"en", "lat"=>36.247752, "lon"=>-98.186381),
"/US/US-OK/amorita-16821" => array("type"=>"town", "pop"=>40, "name_en"=>"Amorita", "lang"=>"en", "lat"=>36.923891, "lon"=>-98.293486),
"/US/US-OK/anadarko-16822" => array("type"=>"city", "pop"=>6323, "name_en"=>"Anadarko", "lang"=>"en", "lat"=>35.065974, "lon"=>-98.244101),
"/US/US-OK/antlers-16823" => array("type"=>"town", "pop"=>2485, "name_en"=>"Antlers", "lang"=>"en", "lat"=>34.231774, "lon"=>-95.618370),
"/US/US-OK/apache-16824" => array("type"=>"town", "pop"=>1536, "name_en"=>"Apache", "lang"=>"en", "lat"=>34.893864, "lon"=>-98.358159),
"/US/US-OK/arapaho-16825" => array("type"=>"town", "pop"=>734, "name_en"=>"Arapaho", "lang"=>"en", "lat"=>35.578111, "lon"=>-98.959590),
"/US/US-OK/arcadia-16826" => array("type"=>"town", "pop"=>281, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>35.665400, "lon"=>-97.326074),
"/US/US-OK/ardmore-16827" => array("type"=>"city", "pop"=>24810, "name_ru"=>"Ардмор", "name_en"=>"Ardmore", "lang"=>"en", "lat"=>34.190700, "lon"=>-97.119667),
"/US/US-OK/arkoma-16828" => array("type"=>"town", "pop"=>2175, "name_en"=>"Arkoma", "lang"=>"en", "lat"=>35.332003, "lon"=>-94.451409),
"/US/US-OK/armstrong-16829" => array("type"=>"town", "pop"=>149, "name_ru"=>"Армстронг", "name_en"=>"Armstrong", "lang"=>"en", "lat"=>34.052407, "lon"=>-96.343924),
"/US/US-OK/arnett-16830" => array("type"=>"town", "pop"=>505, "name_en"=>"Arnett", "lang"=>"en", "lat"=>36.135346, "lon"=>-99.770404),
"/US/US-OK/asher-16831" => array("type"=>"town", "pop"=>431, "name_en"=>"Asher", "lang"=>"en", "lat"=>34.988239, "lon"=>-96.924397),
"/US/US-OK/ashland-16832" => array("type"=>"town", "pop"=>53, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>34.766836, "lon"=>-96.069472),
"/US/US-OK/atoka-16833" => array("type"=>"city", "pop"=>3101, "name_en"=>"Atoka", "lang"=>"en", "lat"=>34.382684, "lon"=>-96.134309),
"/US/US-OK/avant-16834" => array("type"=>"town", "pop"=>355, "name_en"=>"Avant", "lang"=>"en", "lat"=>36.488584, "lon"=>-96.060016),
"/US/US-OK/avard-16835" => array("type"=>"town", "pop"=>19, "name_en"=>"Avard", "lang"=>"en", "lat"=>36.699101, "lon"=>-98.789701),
"/US/US-OK/barnsdall-16836" => array("type"=>"city", "pop"=>1237, "name_en"=>"Barnsdall", "lang"=>"en", "lat"=>36.560458, "lon"=>-96.162038),
"/US/US-OK/bartlesville-16837" => array("type"=>"city", "pop"=>35914, "name_en"=>"Bartlesville", "lang"=>"en", "lat"=>36.737821, "lon"=>-95.948237),
"/US/US-OK/bearden-16838" => array("type"=>"town", "pop"=>131, "name_ru"=>"Бирден", "name_en"=>"Bearden", "lang"=>"en", "lat"=>35.357913, "lon"=>-96.388004),
"/US/US-OK/beaver-16839" => array("type"=>"city", "pop"=>1371, "name_ru"=>"Бивер", "name_en"=>"Beaver", "lang"=>"en", "lat"=>36.814503, "lon"=>-100.523878),
"/US/US-OK/beggs-16840" => array("type"=>"city", "pop"=>1355, "name_en"=>"Beggs", "lang"=>"en", "lat"=>35.781296, "lon"=>-96.023046),
"/US/US-OK/bennington-16841" => array("type"=>"town", "pop"=>304, "name_ru"=>"Беннингтон", "name_en"=>"Bennington", "lang"=>"en", "lat"=>34.003944, "lon"=>-96.037329),
"/US/US-OK/bernice-16842" => array("type"=>"town", "pop"=>534, "name_ru"=>"Бернис", "name_en"=>"Bernice", "lang"=>"en", "lat"=>36.622396, "lon"=>-94.912040),
"/US/US-OK/bessie-16843" => array("type"=>"town", "pop"=>197, "name_en"=>"Bessie", "lang"=>"en", "lat"=>35.384850, "lon"=>-98.988348),
"/US/US-OK/bethany-16844" => array("type"=>"city", "pop"=>19729, "name_ru"=>"Бетани", "name_en"=>"Bethany", "lang"=>"en", "lat"=>35.506000, "lon"=>-97.643374),
"/US/US-OK/bethel-acres-16845" => array("type"=>"town", "pop"=>2871, "name_en"=>"Bethel Acres", "lang"=>"en", "lat"=>35.311328, "lon"=>-97.044872),
"/US/US-OK/big-cabin-16846" => array("type"=>"town", "pop"=>296, "name_en"=>"Big Cabin", "lang"=>"en", "lat"=>36.537391, "lon"=>-95.221994),
"/US/US-OK/billings-16847" => array("type"=>"town", "pop"=>562, "name_ru"=>"Биллингс", "name_en"=>"Billings", "lang"=>"en", "lat"=>36.531358, "lon"=>-97.443173),
"/US/US-OK/binger-16848" => array("type"=>"town", "pop"=>670, "name_en"=>"Binger", "lang"=>"en", "lat"=>35.309249, "lon"=>-98.343665),
"/US/US-OK/bixby-16849" => array("type"=>"city", "pop"=>20678, "name_en"=>"Bixby", "lang"=>"en", "lat"=>35.944950, "lon"=>-95.877949),
"/US/US-OK/blackburn-16850" => array("type"=>"town", "pop"=>99, "name_ru"=>"Блэкберн", "name_en"=>"Blackburn", "lang"=>"en", "lat"=>36.373346, "lon"=>-96.596189),
"/US/US-OK/blackwell-16851" => array("type"=>"city", "pop"=>7144, "name_en"=>"Blackwell", "lang"=>"en", "lat"=>36.801643, "lon"=>-97.298531),
"/US/US-OK/blair-16852" => array("type"=>"town", "pop"=>774, "name_en"=>"Blair", "lang"=>"en", "lat"=>34.778992, "lon"=>-99.332965),
"/US/US-OK/blanchard-16853" => array("type"=>"city", "pop"=>6599, "name_ru"=>"Бланчард", "name_en"=>"Blanchard", "lang"=>"en", "lat"=>35.145215, "lon"=>-97.658596),
"/US/US-OK/bluejacket-16854" => array("type"=>"town", "pop"=>288, "name_en"=>"Bluejacket", "lang"=>"en", "lat"=>36.800511, "lon"=>-95.073329),
"/US/US-OK/boise-city-16855" => array("type"=>"city", "pop"=>1187, "name_ru"=>"Бойсе", "name_en"=>"Boise City", "lang"=>"en", "lat"=>36.730091, "lon"=>-102.510003),
"/US/US-OK/bokchito-16856" => array("type"=>"town", "pop"=>585, "name_en"=>"Bokchito", "lang"=>"en", "lat"=>34.018261, "lon"=>-96.141371),
"/US/US-OK/bokoshe-16857" => array("type"=>"town", "pop"=>463, "name_en"=>"Bokoshe", "lang"=>"en", "lat"=>35.190686, "lon"=>-94.790346),
"/US/US-OK/boley-16858" => array("type"=>"town", "pop"=>1087, "name_en"=>"Boley", "lang"=>"en", "lat"=>35.490097, "lon"=>-96.478293),
"/US/US-OK/boswell-16859" => array("type"=>"town", "pop"=>680, "name_ru"=>"Босуэлл", "name_en"=>"Boswell", "lang"=>"en", "lat"=>34.026691, "lon"=>-95.869661),
"/US/US-OK/bowlegs-16860" => array("type"=>"town", "pop"=>362, "name_en"=>"Bowlegs", "lang"=>"en", "lat"=>35.147489, "lon"=>-96.669309),
"/US/US-OK/boynton-16861" => array("type"=>"town", "pop"=>317, "name_en"=>"Boynton", "lang"=>"en", "lat"=>35.649304, "lon"=>-95.653417),
"/US/US-OK/bradley-16862" => array("type"=>"town", "pop"=>199, "name_ru"=>"Брадли", "name_en"=>"Bradley", "lang"=>"en", "lat"=>34.877273, "lon"=>-97.708505),
"/US/US-OK/braggs-16863" => array("type"=>"town", "pop"=>308, "name_en"=>"Braggs", "lang"=>"en", "lat"=>35.662668, "lon"=>-95.197849),
"/US/US-OK/braman-16864" => array("type"=>"town", "pop"=>235, "name_en"=>"Braman", "lang"=>"en", "lat"=>36.922944, "lon"=>-97.333651),
"/US/US-OK/bray-16865" => array("type"=>"town", "pop"=>1060, "name_en"=>"Bray", "lang"=>"en", "lat"=>34.608827, "lon"=>-97.817497),
"/US/US-OK/breckenridge-16866" => array("type"=>"town", "pop"=>232, "name_ru"=>"Брекенридж", "name_en"=>"Breckenridge", "lang"=>"en", "lat"=>36.460679, "lon"=>-97.723700),
"/US/US-OK/bridgeport-16867" => array("type"=>"city", "pop"=>106, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>35.548727, "lon"=>-98.377423),
"/US/US-OK/bristow-16868" => array("type"=>"city", "pop"=>4383, "name_ru"=>"Бристоу", "name_en"=>"Bristow", "lang"=>"en", "lat"=>35.833276, "lon"=>-96.393973),
"/US/US-OK/broken-arrow-16869" => array("type"=>"city", "pop"=>92931, "name_en"=>"Broken Arrow", "lang"=>"en", "lat"=>36.037520, "lon"=>-95.787180),
"/US/US-OK/broken-bow-16870" => array("type"=>"city", "pop"=>4146, "name_en"=>"Broken Bow", "lang"=>"en", "lat"=>34.025723, "lon"=>-94.736324),
"/US/US-OK/bromide-16871" => array("type"=>"town", "pop"=>158, "name_en"=>"Bromide", "lang"=>"en", "lat"=>34.417817, "lon"=>-96.494291),
"/US/US-OK/brooksville-16872" => array("type"=>"town", "pop"=>93, "name_ru"=>"Бруксвилл", "name_en"=>"Brooksville", "lang"=>"en", "lat"=>35.214080, "lon"=>-96.955687),
"/US/US-OK/buffalo-16873" => array("type"=>"town", "pop"=>1087, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>36.834906, "lon"=>-99.627648),
"/US/US-OK/burbank-16874" => array("type"=>"town", "pop"=>152, "name_ru"=>"Бурбанк", "name_en"=>"Burbank", "lang"=>"en", "lat"=>36.695163, "lon"=>-96.728570),
"/US/US-OK/burlington-16875" => array("type"=>"town", "pop"=>143, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>36.900826, "lon"=>-98.423293),
"/US/US-OK/burns-flat-16876" => array("type"=>"town", "pop"=>1749, "name_en"=>"Burns Flat", "lang"=>"en", "lat"=>35.355082, "lon"=>-99.175109),
"/US/US-OK/butler-16877" => array("type"=>"town", "pop"=>357, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>35.636055, "lon"=>-99.185409),
"/US/US-OK/byars-16878" => array("type"=>"town", "pop"=>292, "name_en"=>"Byars", "lang"=>"en", "lat"=>34.871830, "lon"=>-97.053620),
"/US/US-OK/byng-16879" => array("type"=>"town", "pop"=>1161, "name_en"=>"Byng", "lang"=>"en", "lat"=>34.869420, "lon"=>-96.663582),
"/US/US-OK/byron-16880" => array("type"=>"town", "pop"=>41, "name_ru"=>"Байрон", "name_en"=>"Byron", "lang"=>"en", "lat"=>36.901178, "lon"=>-98.294668),
"/US/US-OK/cache-16881" => array("type"=>"town", "pop"=>2360, "name_en"=>"Cache", "lang"=>"en", "lat"=>34.628617, "lon"=>-98.615657),
"/US/US-OK/caddo-16882" => array("type"=>"town", "pop"=>989, "name_en"=>"Caddo", "lang"=>"en", "lat"=>34.131318, "lon"=>-96.269068),
"/US/US-OK/calera-16883" => array("type"=>"town", "pop"=>1831, "name_ru"=>"Калера", "name_en"=>"Calera", "lang"=>"en", "lat"=>33.934574, "lon"=>-96.430744),
"/US/US-OK/calumet-16884" => array("type"=>"town", "pop"=>532, "name_ru"=>"Калумет", "name_en"=>"Calumet", "lang"=>"en", "lat"=>35.601435, "lon"=>-98.123748),
"/US/US-OK/calvin-16885" => array("type"=>"town", "pop"=>258, "name_ru"=>"Калвин", "name_en"=>"Calvin", "lang"=>"en", "lat"=>34.967368, "lon"=>-96.253234),
"/US/US-OK/camargo-16886" => array("type"=>"town", "pop"=>106, "name_ru"=>"Камарго", "name_en"=>"Camargo", "lang"=>"en", "lat"=>36.017846, "lon"=>-99.288250),
"/US/US-OK/cameron-16887" => array("type"=>"town", "pop"=>320, "name_ru"=>"Кэмерон", "name_en"=>"Cameron", "lang"=>"en", "lat"=>35.134935, "lon"=>-94.535870),
"/US/US-OK/canadian-16888" => array("type"=>"town", "pop"=>245, "name_en"=>"Canadian", "lang"=>"en", "lat"=>35.172627, "lon"=>-95.648870),
"/US/US-OK/caney-16889" => array("type"=>"town", "pop"=>212, "name_ru"=>"Кэни", "name_en"=>"Caney", "lang"=>"en", "lat"=>34.232892, "lon"=>-96.213380),
"/US/US-OK/canton-16890" => array("type"=>"town", "pop"=>570, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>36.056047, "lon"=>-98.588622),
"/US/US-OK/canute-16891" => array("type"=>"town", "pop"=>543, "name_en"=>"Canute", "lang"=>"en", "lat"=>35.425123, "lon"=>-99.276685),
"/US/US-OK/capron-16892" => array("type"=>"town", "pop"=>39, "name_ru"=>"Капрон", "name_en"=>"Capron", "lang"=>"en", "lat"=>36.897389, "lon"=>-98.576732),
"/US/US-OK/cardin-16893" => array("type"=>"town", "pop"=>144, "name_en"=>"Cardin", "lang"=>"en", "lat"=>36.975256, "lon"=>-94.851633),
"/US/US-OK/carmen-16894" => array("type"=>"town", "pop"=>359, "name_en"=>"Carmen", "lang"=>"en", "lat"=>36.578383, "lon"=>-98.457734),
"/US/US-OK/carnegie-16895" => array("type"=>"town", "pop"=>1543, "name_en"=>"Carnegie", "lang"=>"en", "lat"=>35.104277, "lon"=>-98.599280),
"/US/US-OK/carney-16896" => array("type"=>"town", "pop"=>641, "name_ru"=>"Карни", "name_en"=>"Carney", "lang"=>"en", "lat"=>35.805845, "lon"=>-97.015372),
"/US/US-OK/carrier-16897" => array("type"=>"town", "pop"=>77, "name_en"=>"Carrier", "lang"=>"en", "lat"=>36.476753, "lon"=>-98.014384),
"/US/US-OK/carter-16898" => array("type"=>"town", "pop"=>269, "name_en"=>"Carter", "lang"=>"en", "lat"=>35.217243, "lon"=>-99.503418),
"/US/US-OK/cashion-16899" => array("type"=>"town", "pop"=>866, "name_en"=>"Cashion", "lang"=>"en", "lat"=>35.803311, "lon"=>-97.676310),
"/US/US-OK/castle-16900" => array("type"=>"town", "pop"=>115, "name_en"=>"Castle", "lang"=>"en", "lat"=>35.475095, "lon"=>-96.383786),
"/US/US-OK/catoosa-16901" => array("type"=>"city", "pop"=>6607, "name_en"=>"Catoosa", "lang"=>"en", "lat"=>36.181244, "lon"=>-95.765728),
"/US/US-OK/cedar-valley-16902" => array("type"=>"city", "pop"=>222, "name_en"=>"Cedar Valley", "lang"=>"en", "lat"=>35.864036, "lon"=>-97.561583),
"/US/US-OK/cement-16903" => array("type"=>"town", "pop"=>507, "name_en"=>"Cement", "lang"=>"en", "lat"=>34.936314, "lon"=>-98.136236),
"/US/US-OK/centrahoma-16904" => array("type"=>"city", "pop"=>105, "name_en"=>"Centrahoma", "lang"=>"en", "lat"=>34.609197, "lon"=>-96.343494),
"/US/US-OK/chandler-16905" => array("type"=>"city", "pop"=>2822, "name_ru"=>"Чандлер", "name_en"=>"Chandler", "lang"=>"en", "lat"=>35.708019, "lon"=>-96.876495),
"/US/US-OK/chattanooga-16906" => array("type"=>"town", "pop"=>412, "name_en"=>"Chattanooga", "lang"=>"en", "lat"=>34.423974, "lon"=>-98.654695),
"/US/US-OK/checotah-16907" => array("type"=>"city", "pop"=>3469, "name_en"=>"Checotah", "lang"=>"en", "lat"=>35.477878, "lon"=>-95.521703),
"/US/US-OK/chelsea-16908" => array("type"=>"city", "pop"=>2228, "name_ru"=>"Челси", "name_en"=>"Chelsea", "lang"=>"en", "lat"=>36.534734, "lon"=>-95.431859),
"/US/US-OK/cherokee-16909" => array("type"=>"city", "pop"=>1437, "name_ru"=>"Чероки", "name_en"=>"Cherokee", "lang"=>"en", "lat"=>36.754133, "lon"=>-98.354932),
"/US/US-OK/cheyenne-16910" => array("type"=>"town", "pop"=>744, "name_en"=>"Cheyenne", "lang"=>"en", "lat"=>35.611365, "lon"=>-99.677046),
"/US/US-OK/chickasha-16911" => array("type"=>"city", "pop"=>17062, "name_en"=>"Chickasha", "lang"=>"en", "lat"=>35.040449, "lon"=>-97.952271),
"/US/US-OK/choctaw-16912" => array("type"=>"city", "pop"=>11354, "name_en"=>"Choctaw", "lang"=>"en", "lat"=>35.482370, "lon"=>-97.265371),
"/US/US-OK/chouteau-16913" => array("type"=>"town", "pop"=>2012, "name_en"=>"Chouteau", "lang"=>"en", "lat"=>36.189607, "lon"=>-95.338605),
"/US/US-OK/cimarron-city-16914" => array("type"=>"town", "pop"=>125, "name_en"=>"Cimarron City", "lang"=>"en", "lat"=>35.889366, "lon"=>-97.601281),
"/US/US-OK/claremore-16915" => array("type"=>"city", "pop"=>17458, "name_en"=>"Claremore", "lang"=>"en", "lat"=>36.313402, "lon"=>-95.619780),
"/US/US-OK/clayton-16916" => array("type"=>"town", "pop"=>727, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>34.586030, "lon"=>-95.355445),
"/US/US-OK/clearview-16917" => array("type"=>"town", "pop"=>52, "name_en"=>"Clearview", "lang"=>"en", "lat"=>35.397548, "lon"=>-96.190407),
"/US/US-OK/cleo-springs-16918" => array("type"=>"town", "pop"=>299, "name_en"=>"Cleo Springs", "lang"=>"en", "lat"=>36.406890, "lon"=>-98.439538),
"/US/US-OK/cleveland-16919" => array("type"=>"city", "pop"=>3140, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>36.299228, "lon"=>-96.461190),
"/US/US-OK/clinton-16920" => array("type"=>"city", "pop"=>8768, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>35.506376, "lon"=>-98.972980),
"/US/US-OK/coalgate-16921" => array("type"=>"city", "pop"=>1874, "name_en"=>"Coalgate", "lang"=>"en", "lat"=>34.533570, "lon"=>-96.218452),
"/US/US-OK/colbert-16922" => array("type"=>"town", "pop"=>1123, "name_ru"=>"Колберт", "name_en"=>"Colbert", "lang"=>"en", "lat"=>33.857433, "lon"=>-96.504214),
"/US/US-OK/colcord-16923" => array("type"=>"town", "pop"=>867, "name_en"=>"Colcord", "lang"=>"en", "lat"=>36.265096, "lon"=>-94.692013),
"/US/US-OK/cole-16924" => array("type"=>"town", "pop"=>496, "name_en"=>"Cole", "lang"=>"en", "lat"=>35.102820, "lon"=>-97.572486),
"/US/US-OK/collinsville-16925" => array("type"=>"city", "pop"=>5039, "name_ru"=>"Коллинсвилл", "name_en"=>"Collinsville", "lang"=>"en", "lat"=>36.378270, "lon"=>-95.855018),
"/US/US-OK/colony-16926" => array("type"=>"town", "pop"=>152, "name_ru"=>"Колони", "name_en"=>"Colony", "lang"=>"en", "lat"=>35.347651, "lon"=>-98.670110),
"/US/US-OK/comanche-16927" => array("type"=>"city", "pop"=>1523, "name_en"=>"Comanche", "lang"=>"en", "lat"=>34.368697, "lon"=>-97.967375),
"/US/US-OK/commerce-16928" => array("type"=>"city", "pop"=>2506, "name_ru"=>"Коммерс", "name_en"=>"Commerce", "lang"=>"en", "lat"=>36.931882, "lon"=>-94.869942),
"/US/US-OK/cooperton-16929" => array("type"=>"town", "pop"=>19, "name_en"=>"Cooperton", "lang"=>"en", "lat"=>34.863298, "lon"=>-98.877817),
"/US/US-OK/copan-16930" => array("type"=>"town", "pop"=>820, "name_en"=>"Copan", "lang"=>"en", "lat"=>36.899847, "lon"=>-95.924970),
"/US/US-OK/corn-16931" => array("type"=>"town", "pop"=>584, "name_en"=>"Corn", "lang"=>"en", "lat"=>35.378869, "lon"=>-98.781310),
"/US/US-OK/cornish-16932" => array("type"=>"town", "pop"=>160, "name_en"=>"Cornish", "lang"=>"en", "lat"=>34.162548, "lon"=>-97.595168),
"/US/US-OK/council-hill-16933" => array("type"=>"town", "pop"=>131, "name_en"=>"Council Hill", "lang"=>"en", "lat"=>35.556036, "lon"=>-95.652373),
"/US/US-OK/covington-16934" => array("type"=>"town", "pop"=>554, "name_ru"=>"Ковингтон", "name_en"=>"Covington", "lang"=>"en", "lat"=>36.307478, "lon"=>-97.585623),
"/US/US-OK/coweta-16935" => array("type"=>"city", "pop"=>9096, "name_en"=>"Coweta", "lang"=>"en", "lat"=>35.963933, "lon"=>-95.660789),
"/US/US-OK/cowlington-16936" => array("type"=>"town", "pop"=>137, "name_en"=>"Cowlington", "lang"=>"en", "lat"=>35.276446, "lon"=>-94.728071),
"/US/US-OK/coyle-16937" => array("type"=>"town", "pop"=>378, "name_en"=>"Coyle", "lang"=>"en", "lat"=>35.954037, "lon"=>-97.237357),
"/US/US-OK/crescent-16938" => array("type"=>"city", "pop"=>1401, "name_ru"=>"Кресент", "name_en"=>"Crescent", "lang"=>"en", "lat"=>35.949861, "lon"=>-97.589612),
"/US/US-OK/cromwell-16939" => array("type"=>"town", "pop"=>257, "name_ru"=>"Кромуэлл", "name_en"=>"Cromwell", "lang"=>"en", "lat"=>35.341664, "lon"=>-96.454843),
"/US/US-OK/crowder-16940" => array("type"=>"town", "pop"=>448, "name_ru"=>"Краудер", "name_en"=>"Crowder", "lang"=>"en", "lat"=>35.122859, "lon"=>-95.666945),
"/US/US-OK/cushing-16941" => array("type"=>"city", "pop"=>8767, "name_ru"=>"Кушинг", "name_en"=>"Cushing", "lang"=>"en", "lat"=>35.979449, "lon"=>-96.761484),
"/US/US-OK/custer-city-16942" => array("type"=>"town", "pop"=>398, "name_en"=>"Custer City", "lang"=>"en", "lat"=>35.664272, "lon"=>-98.886794),
"/US/US-OK/cyril-16943" => array("type"=>"town", "pop"=>1126, "name_en"=>"Cyril", "lang"=>"en", "lat"=>34.899369, "lon"=>-98.202181),
"/US/US-OK/dacoma-16944" => array("type"=>"town", "pop"=>139, "name_en"=>"Dacoma", "lang"=>"en", "lat"=>36.660027, "lon"=>-98.563693),
"/US/US-OK/davenport-16945" => array("type"=>"town", "pop"=>874, "name_ru"=>"Дэйвенпорт", "name_en"=>"Davenport", "lang"=>"en", "lat"=>35.709321, "lon"=>-96.765567),
"/US/US-OK/davidson-16946" => array("type"=>"town", "pop"=>319, "name_en"=>"Davidson", "lang"=>"en", "lat"=>34.242105, "lon"=>-99.078169),
"/US/US-OK/davis-16947" => array("type"=>"city", "pop"=>2636, "name_ru"=>"Давис", "name_en"=>"Davis", "lang"=>"en", "lat"=>34.480930, "lon"=>-97.130455),
"/US/US-OK/deer-creek-16948" => array("type"=>"town", "pop"=>128, "name_ru"=>"Дир Крик", "name_en"=>"Deer Creek", "lang"=>"en", "lat"=>36.806304, "lon"=>-97.518772),
"/US/US-OK/del-city-16950" => array("type"=>"city", "pop"=>22067, "name_en"=>"Del City", "lang"=>"en", "lat"=>35.447450, "lon"=>-97.439749),
"/US/US-OK/delaware-16949" => array("type"=>"town", "pop"=>465, "name_ru"=>"Делавэр", "name_en"=>"Delaware", "lang"=>"en", "lat"=>36.778973, "lon"=>-95.642872),
"/US/US-OK/depew-16951" => array("type"=>"town", "pop"=>569, "name_en"=>"Depew", "lang"=>"en", "lat"=>35.800349, "lon"=>-96.507612),
"/US/US-OK/devol-16952" => array("type"=>"town", "pop"=>143, "name_en"=>"Devol", "lang"=>"en", "lat"=>34.194695, "lon"=>-98.587668),
"/US/US-OK/dewar-16953" => array("type"=>"town", "pop"=>898, "name_en"=>"Dewar", "lang"=>"en", "lat"=>35.458183, "lon"=>-95.946689),
"/US/US-OK/dewey-16954" => array("type"=>"city", "pop"=>3324, "name_en"=>"Dewey", "lang"=>"en", "lat"=>36.787808, "lon"=>-95.933315),
"/US/US-OK/dibble-16955" => array("type"=>"town", "pop"=>644, "name_en"=>"Dibble", "lang"=>"en", "lat"=>35.034024, "lon"=>-97.628039),
"/US/US-OK/dickson-16956" => array("type"=>"town", "pop"=>1262, "name_en"=>"Dickson", "lang"=>"en", "lat"=>34.188670, "lon"=>-96.994107),
"/US/US-OK/dill-city-16957" => array("type"=>"town", "pop"=>543, "name_en"=>"Dill City", "lang"=>"en", "lat"=>35.282524, "lon"=>-99.133789),
"/US/US-OK/disney-16958" => array("type"=>"town", "pop"=>229, "name_en"=>"Disney", "lang"=>"en", "lat"=>36.477637, "lon"=>-95.020566),
"/US/US-OK/dougherty-16959" => array("type"=>"town", "pop"=>227, "name_ru"=>"Доэрти", "name_en"=>"Dougherty", "lang"=>"en", "lat"=>34.399768, "lon"=>-97.051083),
"/US/US-OK/douglas-16960" => array("type"=>"town", "pop"=>32, "name_ru"=>"Дуглас", "name_en"=>"Douglas", "lang"=>"en", "lat"=>36.259968, "lon"=>-97.667309),
"/US/US-OK/dover-16961" => array("type"=>"town", "pop"=>360, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>35.981558, "lon"=>-97.910714),
"/US/US-OK/drummond-16962" => array("type"=>"town", "pop"=>393, "name_ru"=>"Драммонд", "name_en"=>"Drummond", "lang"=>"en", "lat"=>36.300950, "lon"=>-98.035391),
"/US/US-OK/drumright-16963" => array("type"=>"city", "pop"=>2885, "name_en"=>"Drumright", "lang"=>"en", "lat"=>35.990700, "lon"=>-96.597444),
"/US/US-OK/duncan-16964" => array("type"=>"city", "pop"=>22617, "name_ru"=>"Дункан", "name_en"=>"Duncan", "lang"=>"en", "lat"=>34.520072, "lon"=>-97.968173),
"/US/US-OK/durant-16965" => array("type"=>"city", "pop"=>16450, "name_ru"=>"Дурант", "name_en"=>"Durant", "lang"=>"en", "lat"=>33.984070, "lon"=>-96.392329),
"/US/US-OK/dustin-16966" => array("type"=>"town", "pop"=>441, "name_en"=>"Dustin", "lang"=>"en", "lat"=>35.271184, "lon"=>-96.031370),
"/US/US-OK/eakly-16967" => array("type"=>"town", "pop"=>265, "name_en"=>"Eakly", "lang"=>"en", "lat"=>35.306279, "lon"=>-98.555734),
"/US/US-OK/earlsboro-16968" => array("type"=>"town", "pop"=>660, "name_en"=>"Earlsboro", "lang"=>"en", "lat"=>35.325758, "lon"=>-96.798016),
"/US/US-OK/east-duke-16969" => array("type"=>"town", "pop"=>400, "name_en"=>"East Duke", "lang"=>"en", "lat"=>34.662893, "lon"=>-99.569262),
"/US/US-OK/east-ninnekah-16970" => array("type"=>"town", "pop"=>1016, "name_en"=>"East Ninnekah", "lang"=>"en", "lat"=>34.964060, "lon"=>-97.943250),
"/US/US-OK/edmond-16971" => array("type"=>"city", "pop"=>79559, "name_ru"=>"Эдмонд", "name_en"=>"Edmond", "lang"=>"en", "lat"=>35.672000, "lon"=>-97.413170),
"/US/US-OK/el-reno-16977" => array("type"=>"city", "pop"=>16545, "name_en"=>"El Reno", "lang"=>"en", "lat"=>35.544548, "lon"=>-97.965111),
"/US/US-OK/eldorado-16972" => array("type"=>"town", "pop"=>473, "name_ru"=>"Эльдорадо", "name_en"=>"Eldorado", "lang"=>"en", "lat"=>34.472534, "lon"=>-99.650170),
"/US/US-OK/elgin-16973" => array("type"=>"town", "pop"=>1524, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>34.782554, "lon"=>-98.291235),
"/US/US-OK/elk-city-16974" => array("type"=>"city", "pop"=>11311, "name_ru"=>"Элк Сити", "name_en"=>"Elk City", "lang"=>"en", "lat"=>35.402930, "lon"=>-99.411334),
"/US/US-OK/elmer-16975" => array("type"=>"town", "pop"=>87, "name_ru"=>"Элмер", "name_en"=>"Elmer", "lang"=>"en", "lat"=>34.481422, "lon"=>-99.352333),
"/US/US-OK/elmore-city-16976" => array("type"=>"town", "pop"=>787, "name_en"=>"Elmore City", "lang"=>"en", "lat"=>34.623343, "lon"=>-97.396458),
"/US/US-OK/empire-city-16978" => array("type"=>"town", "pop"=>744, "name_en"=>"Empire City", "lang"=>"en", "lat"=>34.464831, "lon"=>-98.043033),
"/US/US-OK/enid-16979" => array("type"=>"city", "pop"=>47396, "name_en"=>"Enid", "lang"=>"en", "lat"=>36.408600, "lon"=>-97.868149),
"/US/US-OK/erick-16980" => array("type"=>"city", "pop"=>1076, "name_en"=>"Erick", "lang"=>"en", "lat"=>35.213486, "lon"=>-99.866900),
"/US/US-OK/erin-springs-16981" => array("type"=>"town", "pop"=>139, "name_en"=>"Erin Springs", "lang"=>"en", "lat"=>34.810654, "lon"=>-97.605138),
"/US/US-OK/etowah-16982" => array("type"=>"town", "pop"=>127, "name_en"=>"Etowah", "lang"=>"en", "lat"=>35.131204, "lon"=>-97.168661),
"/US/US-OK/eufaula-16983" => array("type"=>"city", "pop"=>2756, "name_ru"=>"Юфола", "name_en"=>"Eufaula", "lang"=>"en", "lat"=>35.292856, "lon"=>-95.586684),
"/US/US-OK/fair-oaks-16987" => array("type"=>"town", "pop"=>75, "name_ru"=>"Фэр Оукс", "name_en"=>"Fair Oaks", "lang"=>"en", "lat"=>36.147842, "lon"=>-95.708069),
"/US/US-OK/fairfax-16984" => array("type"=>"town", "pop"=>1454, "name_ru"=>"Фэрфакс", "name_en"=>"Fairfax", "lang"=>"en", "lat"=>36.570011, "lon"=>-96.707289),
"/US/US-OK/fairland-16985" => array("type"=>"town", "pop"=>986, "name_ru"=>"Фэрленд", "name_en"=>"Fairland", "lang"=>"en", "lat"=>36.751003, "lon"=>-94.848611),
"/US/US-OK/fairmont-16986" => array("type"=>"town", "pop"=>152, "name_ru"=>"Фэрмонт", "name_en"=>"Fairmont", "lang"=>"en", "lat"=>36.354900, "lon"=>-97.703268),
"/US/US-OK/fairview-16988" => array("type"=>"city", "pop"=>2524, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>36.269391, "lon"=>-98.476332),
"/US/US-OK/fallis-16989" => array("type"=>"town", "pop"=>28, "name_en"=>"Fallis", "lang"=>"en", "lat"=>35.749195, "lon"=>-97.118134),
"/US/US-OK/fanshawe-16990" => array("type"=>"town", "pop"=>395, "name_en"=>"Fanshawe", "lang"=>"en", "lat"=>34.981689, "lon"=>-94.869271),
"/US/US-OK/fargo-16991" => array("type"=>"town", "pop"=>321, "name_ru"=>"Фарго", "name_en"=>"Fargo", "lang"=>"en", "lat"=>36.373411, "lon"=>-99.622357),
"/US/US-OK/faxon-16992" => array("type"=>"town", "pop"=>129, "name_en"=>"Faxon", "lang"=>"en", "lat"=>34.460374, "lon"=>-98.578253),
"/US/US-OK/fitzhugh-16993" => array("type"=>"town", "pop"=>215, "name_en"=>"Fitzhugh", "lang"=>"en", "lat"=>34.662259, "lon"=>-96.775564),
"/US/US-OK/fletcher-16994" => array("type"=>"town", "pop"=>1077, "name_en"=>"Fletcher", "lang"=>"en", "lat"=>34.824461, "lon"=>-98.240324),
"/US/US-OK/foraker-16995" => array("type"=>"town", "pop"=>22, "name_en"=>"Foraker", "lang"=>"en", "lat"=>36.872588, "lon"=>-96.568961),
"/US/US-OK/forest-park-16996" => array("type"=>"town", "pop"=>1178, "name_ru"=>"Форест Парк", "name_en"=>"Forest Park", "lang"=>"en", "lat"=>35.514833, "lon"=>-97.446463),
"/US/US-OK/forgan-16997" => array("type"=>"town", "pop"=>482, "name_en"=>"Forgan", "lang"=>"en", "lat"=>36.906358, "lon"=>-100.537570),
"/US/US-OK/fort-cobb-16998" => array("type"=>"town", "pop"=>637, "name_en"=>"Fort Cobb", "lang"=>"en", "lat"=>35.100015, "lon"=>-98.441417),
"/US/US-OK/fort-gibson-16999" => array("type"=>"town", "pop"=>4368, "name_en"=>"Fort Gibson", "lang"=>"en", "lat"=>35.785718, "lon"=>-95.248845),
"/US/US-OK/fort-sill-17000" => array("type"=>"cdp", "pop"=>12107, "name_en"=>"Fort Sill", "lang"=>"en", "lat"=>34.658558, "lon"=>-98.406598),
"/US/US-OK/fort-supply-17001" => array("type"=>"town", "pop"=>336, "name_en"=>"Fort Supply", "lang"=>"en", "lat"=>36.572237, "lon"=>-99.573464),
"/US/US-OK/fort-towson-17002" => array("type"=>"town", "pop"=>591, "name_en"=>"Fort Towson", "lang"=>"en", "lat"=>34.026493, "lon"=>-95.295365),
"/US/US-OK/foss-17003" => array("type"=>"town", "pop"=>132, "name_en"=>"Foss", "lang"=>"en", "lat"=>35.450639, "lon"=>-99.171566),
"/US/US-OK/foyil-17004" => array("type"=>"town", "pop"=>268, "name_en"=>"Foyil", "lang"=>"en", "lat"=>36.433698, "lon"=>-95.520454),
"/US/US-OK/francis-17005" => array("type"=>"town", "pop"=>352, "name_en"=>"Francis", "lang"=>"en", "lat"=>34.874905, "lon"=>-96.592640),
"/US/US-OK/frederick-17006" => array("type"=>"city", "pop"=>3869, "name_ru"=>"Фредерик", "name_en"=>"Frederick", "lang"=>"en", "lat"=>34.388943, "lon"=>-99.012107),
"/US/US-OK/freedom-17007" => array("type"=>"town", "pop"=>257, "name_ru"=>"Фридом", "name_en"=>"Freedom", "lang"=>"en", "lat"=>36.766524, "lon"=>-99.112557),
"/US/US-OK/gage-17008" => array("type"=>"town", "pop"=>405, "name_en"=>"Gage", "lang"=>"en", "lat"=>36.318080, "lon"=>-99.756654),
"/US/US-OK/gans-17009" => array("type"=>"town", "pop"=>213, "name_en"=>"Gans", "lang"=>"en", "lat"=>35.386995, "lon"=>-94.694793),
"/US/US-OK/garber-17010" => array("type"=>"city", "pop"=>807, "name_ru"=>"Гарбер", "name_en"=>"Garber", "lang"=>"en", "lat"=>36.436542, "lon"=>-97.580399),
"/US/US-OK/garvin-17011" => array("type"=>"town", "pop"=>222, "name_ru"=>"Гарвин", "name_en"=>"Garvin", "lang"=>"en", "lat"=>33.954780, "lon"=>-94.941291),
"/US/US-OK/gate-17012" => array("type"=>"town", "pop"=>100, "name_en"=>"Gate", "lang"=>"en", "lat"=>36.851873, "lon"=>-100.055392),
"/US/US-OK/geary-17013" => array("type"=>"city", "pop"=>1241, "name_en"=>"Geary", "lang"=>"en", "lat"=>35.629438, "lon"=>-98.317836),
"/US/US-OK/gene-autry-17014" => array("type"=>"town", "pop"=>103, "name_en"=>"Gene Autry", "lang"=>"en", "lat"=>34.284810, "lon"=>-97.037671),
"/US/US-OK/geronimo-17015" => array("type"=>"town", "pop"=>943, "name_en"=>"Geronimo", "lang"=>"en", "lat"=>34.482468, "lon"=>-98.383270),
"/US/US-OK/gerty-17016" => array("type"=>"town", "pop"=>98, "name_en"=>"Gerty", "lang"=>"en", "lat"=>34.835974, "lon"=>-96.289618),
"/US/US-OK/glencoe-17017" => array("type"=>"town", "pop"=>628, "name_ru"=>"Гленко", "name_en"=>"Glencoe", "lang"=>"en", "lat"=>36.226086, "lon"=>-96.927370),
"/US/US-OK/glenpool-17018" => array("type"=>"city", "pop"=>9914, "name_en"=>"Glenpool", "lang"=>"en", "lat"=>35.957700, "lon"=>-96.001049),
"/US/US-OK/goldsby-17019" => array("type"=>"town", "pop"=>1597, "name_en"=>"Goldsby", "lang"=>"en", "lat"=>35.128046, "lon"=>-97.471357),
"/US/US-OK/goltry-17020" => array("type"=>"town", "pop"=>236, "name_en"=>"Goltry", "lang"=>"en", "lat"=>36.531990, "lon"=>-98.150737),
"/US/US-OK/goodwell-17021" => array("type"=>"town", "pop"=>1183, "name_en"=>"Goodwell", "lang"=>"en", "lat"=>36.594496, "lon"=>-101.632145),
"/US/US-OK/gore-17022" => array("type"=>"town", "pop"=>909, "name_en"=>"Gore", "lang"=>"en", "lat"=>35.543341, "lon"=>-95.111380),
"/US/US-OK/gotebo-17023" => array("type"=>"town", "pop"=>255, "name_en"=>"Gotebo", "lang"=>"en", "lat"=>35.071090, "lon"=>-98.873802),
"/US/US-OK/gould-17024" => array("type"=>"town", "pop"=>178, "name_ru"=>"Гулд", "name_en"=>"Gould", "lang"=>"en", "lat"=>34.669095, "lon"=>-99.773785),
"/US/US-OK/gracemont-17025" => array("type"=>"town", "pop"=>323, "name_en"=>"Gracemont", "lang"=>"en", "lat"=>35.187844, "lon"=>-98.256469),
"/US/US-OK/grainola-17026" => array("type"=>"town", "pop"=>30, "name_en"=>"Grainola", "lang"=>"en", "lat"=>36.938379, "lon"=>-96.648906),
"/US/US-OK/grand-lake-towne-17028" => array("type"=>"town", "pop"=>72, "name_en"=>"Grand Lake Towne", "lang"=>"en", "lat"=>36.505912, "lon"=>-95.027364),
"/US/US-OK/grandfield-17027" => array("type"=>"city", "pop"=>942, "name_en"=>"Grandfield", "lang"=>"en", "lat"=>34.231224, "lon"=>-98.687305),
"/US/US-OK/granite-17029" => array("type"=>"town", "pop"=>1788, "name_en"=>"Granite", "lang"=>"en", "lat"=>34.962938, "lon"=>-99.378301),
"/US/US-OK/grayson-17030" => array("type"=>"town", "pop"=>135, "name_ru"=>"Грейсон", "name_en"=>"Grayson", "lang"=>"en", "lat"=>35.504578, "lon"=>-95.869686),
"/US/US-OK/greenfield-17031" => array("type"=>"town", "pop"=>119, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>35.728700, "lon"=>-98.377103),
"/US/US-OK/grove-17032" => array("type"=>"city", "pop"=>6338, "name_en"=>"Grove", "lang"=>"en", "lat"=>36.595249, "lon"=>-94.787429),
"/US/US-OK/guthrie-17033" => array("type"=>"city", "pop"=>11043, "name_ru"=>"Гатри", "name_en"=>"Guthrie", "lang"=>"en", "lat"=>35.840228, "lon"=>-97.428076),
"/US/US-OK/guymon-17034" => array("type"=>"city", "pop"=>10702, "name_en"=>"Guymon", "lang"=>"en", "lat"=>36.686282, "lon"=>-101.483009),
"/US/US-OK/haileyville-17035" => array("type"=>"city", "pop"=>904, "name_en"=>"Haileyville", "lang"=>"en", "lat"=>34.853831, "lon"=>-95.578519),
"/US/US-OK/hall-park-17037" => array("type"=>"town", "pop"=>1090, "name_en"=>"Hall Park", "lang"=>"en", "lat"=>35.240068, "lon"=>-97.406798),
"/US/US-OK/hallett-17036" => array("type"=>"town", "pop"=>166, "name_en"=>"Hallett", "lang"=>"en", "lat"=>36.232363, "lon"=>-96.566897),
"/US/US-OK/hammon-17038" => array("type"=>"town", "pop"=>468, "name_en"=>"Hammon", "lang"=>"en", "lat"=>35.632322, "lon"=>-99.383437),
"/US/US-OK/hanna-17039" => array("type"=>"town", "pop"=>133, "name_en"=>"Hanna", "lang"=>"en", "lat"=>35.204089, "lon"=>-95.889149),
"/US/US-OK/hardesty-17040" => array("type"=>"town", "pop"=>281, "name_en"=>"Hardesty", "lang"=>"en", "lat"=>36.614517, "lon"=>-101.193347),
"/US/US-OK/harrah-17041" => array("type"=>"town", "pop"=>5208, "name_en"=>"Harrah", "lang"=>"en", "lat"=>35.485704, "lon"=>-97.184129),
"/US/US-OK/hartshorne-17042" => array("type"=>"city", "pop"=>2079, "name_en"=>"Hartshorne", "lang"=>"en", "lat"=>34.839458, "lon"=>-95.557363),
"/US/US-OK/haskell-17043" => array("type"=>"town", "pop"=>1987, "name_ru"=>"Хаскел", "name_en"=>"Haskell", "lang"=>"en", "lat"=>35.823888, "lon"=>-95.678525),
"/US/US-OK/hastings-17044" => array("type"=>"town", "pop"=>143, "name_ru"=>"Гастингс", "name_en"=>"Hastings", "lang"=>"en", "lat"=>34.225143, "lon"=>-98.108369),
"/US/US-OK/haworth-17045" => array("type"=>"town", "pop"=>360, "name_en"=>"Haworth", "lang"=>"en", "lat"=>33.844235, "lon"=>-94.654847),
"/US/US-OK/headrick-17046" => array("type"=>"town", "pop"=>119, "name_en"=>"Headrick", "lang"=>"en", "lat"=>34.626736, "lon"=>-99.137137),
"/US/US-OK/healdton-17047" => array("type"=>"city", "pop"=>2782, "name_en"=>"Healdton", "lang"=>"en", "lat"=>34.233530, "lon"=>-97.486833),
"/US/US-OK/heavener-17048" => array("type"=>"city", "pop"=>3233, "name_en"=>"Heavener", "lang"=>"en", "lat"=>34.891892, "lon"=>-94.607477),
"/US/US-OK/helena-17049" => array("type"=>"town", "pop"=>1424, "name_ru"=>"Хелена", "name_en"=>"Helena", "lang"=>"en", "lat"=>36.546446, "lon"=>-98.270210),
"/US/US-OK/hendrix-17050" => array("type"=>"town", "pop"=>80, "name_en"=>"Hendrix", "lang"=>"en", "lat"=>33.774658, "lon"=>-96.407987),
"/US/US-OK/hennessey-17051" => array("type"=>"town", "pop"=>2031, "name_en"=>"Hennessey", "lang"=>"en", "lat"=>36.105935, "lon"=>-97.898017),
"/US/US-OK/henryetta-17052" => array("type"=>"city", "pop"=>6031, "name_en"=>"Henryetta", "lang"=>"en", "lat"=>35.441247, "lon"=>-95.984731),
"/US/US-OK/hickory-17053" => array("type"=>"town", "pop"=>87, "name_ru"=>"Хикори", "name_en"=>"Hickory", "lang"=>"en", "lat"=>34.555966, "lon"=>-96.855403),
"/US/US-OK/hillsdale-17054" => array("type"=>"town", "pop"=>97, "name_ru"=>"Хиллсдейл", "name_en"=>"Hillsdale", "lang"=>"en", "lat"=>36.564003, "lon"=>-97.992465),
"/US/US-OK/hinton-17055" => array("type"=>"town", "pop"=>2133, "name_ru"=>"Хинтон", "name_en"=>"Hinton", "lang"=>"en", "lat"=>35.495631, "lon"=>-98.359648),
"/US/US-OK/hitchcock-17056" => array("type"=>"town", "pop"=>137, "name_en"=>"Hitchcock", "lang"=>"en", "lat"=>35.968450, "lon"=>-98.348914),
"/US/US-OK/hitchita-17057" => array("type"=>"town", "pop"=>110, "name_en"=>"Hitchita", "lang"=>"en", "lat"=>35.519399, "lon"=>-95.752142),
"/US/US-OK/hobart-17058" => array("type"=>"city", "pop"=>3633, "name_ru"=>"Хобарт", "name_en"=>"Hobart", "lang"=>"en", "lat"=>35.026232, "lon"=>-99.091922),
"/US/US-OK/hoffman-17059" => array("type"=>"town", "pop"=>146, "name_ru"=>"Хоффман", "name_en"=>"Hoffman", "lang"=>"en", "lat"=>35.488271, "lon"=>-95.844450),
"/US/US-OK/holdenville-17060" => array("type"=>"city", "pop"=>5417, "name_en"=>"Holdenville", "lang"=>"en", "lat"=>35.084757, "lon"=>-96.399021),
"/US/US-OK/hollis-17061" => array("type"=>"city", "pop"=>1961, "name_ru"=>"Холлис", "name_en"=>"Hollis", "lang"=>"en", "lat"=>34.689656, "lon"=>-99.915372),
"/US/US-OK/hollister-17062" => array("type"=>"town", "pop"=>52, "name_ru"=>"Холлистер", "name_en"=>"Hollister", "lang"=>"en", "lat"=>34.341630, "lon"=>-98.870367),
"/US/US-OK/hominy-17063" => array("type"=>"city", "pop"=>3631, "name_en"=>"Hominy", "lang"=>"en", "lat"=>36.421069, "lon"=>-96.393265),
"/US/US-OK/hooker-17064" => array("type"=>"city", "pop"=>1722, "name_en"=>"Hooker", "lang"=>"en", "lat"=>36.859913, "lon"=>-101.211964),
"/US/US-OK/hoot-owl-17065" => array("type"=>"town", "pop"=>1, "name_en"=>"Hoot Owl", "lang"=>"en", "lat"=>36.359945, "lon"=>-95.121370),
"/US/US-OK/howe-17066" => array("type"=>"town", "pop"=>716, "name_en"=>"Howe", "lang"=>"en", "lat"=>34.949265, "lon"=>-94.637401),
"/US/US-OK/hugo-17067" => array("type"=>"city", "pop"=>5390, "name_ru"=>"Хьюго", "name_en"=>"Hugo", "lang"=>"en", "lat"=>34.013689, "lon"=>-95.515797),
"/US/US-OK/hulbert-17068" => array("type"=>"town", "pop"=>540, "name_en"=>"Hulbert", "lang"=>"en", "lat"=>35.931461, "lon"=>-95.142978),
"/US/US-OK/hunter-17069" => array("type"=>"town", "pop"=>173, "name_ru"=>"Хантер", "name_en"=>"Hunter", "lang"=>"en", "lat"=>36.564021, "lon"=>-97.662437),
"/US/US-OK/hydro-17070" => array("type"=>"town", "pop"=>1009, "name_en"=>"Hydro", "lang"=>"en", "lat"=>35.548820, "lon"=>-98.578619),
"/US/US-OK/idabel-17071" => array("type"=>"city", "pop"=>6851, "name_en"=>"Idabel", "lang"=>"en", "lat"=>33.901973, "lon"=>-94.818784),
"/US/US-OK/indiahoma-17072" => array("type"=>"town", "pop"=>339, "name_en"=>"Indiahoma", "lang"=>"en", "lat"=>34.620283, "lon"=>-98.751751),
"/US/US-OK/indianola-17073" => array("type"=>"town", "pop"=>194, "name_ru"=>"Индианола", "name_en"=>"Indianola", "lang"=>"en", "lat"=>35.166205, "lon"=>-95.775544),
"/US/US-OK/inola-17074" => array("type"=>"town", "pop"=>1745, "name_en"=>"Inola", "lang"=>"en", "lat"=>36.133454, "lon"=>-95.532123),
"/US/US-OK/jamestown-17075" => array("type"=>"town", "pop"=>4, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>36.569971, "lon"=>-95.629062),
"/US/US-OK/jay-17076" => array("type"=>"city", "pop"=>2994, "name_ru"=>"Джэй", "name_en"=>"Jay", "lang"=>"en", "lat"=>36.424536, "lon"=>-94.797538),
"/US/US-OK/jefferson-17077" => array("type"=>"town", "pop"=>32, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>36.719098, "lon"=>-97.790166),
"/US/US-OK/jenks-17078" => array("type"=>"city", "pop"=>15590, "name_en"=>"Jenks", "lang"=>"en", "lat"=>36.002150, "lon"=>-95.974647),
"/US/US-OK/jennings-17079" => array("type"=>"town", "pop"=>368, "name_ru"=>"Дженнингс", "name_en"=>"Jennings", "lang"=>"en", "lat"=>36.181680, "lon"=>-96.568592),
"/US/US-OK/jet-17080" => array("type"=>"town", "pop"=>212, "name_en"=>"Jet", "lang"=>"en", "lat"=>36.666760, "lon"=>-98.180708),
"/US/US-OK/johnson-17081" => array("type"=>"town", "pop"=>233, "name_ru"=>"Джонсон", "name_en"=>"Johnson", "lang"=>"en", "lat"=>35.405123, "lon"=>-96.842564),
"/US/US-OK/jones-17082" => array("type"=>"town", "pop"=>2702, "name_en"=>"Jones", "lang"=>"en", "lat"=>35.566600, "lon"=>-97.290574),
"/US/US-OK/kansas-17083" => array("type"=>"town", "pop"=>723, "name_ru"=>"Канзас", "name_en"=>"Kansas", "lang"=>"en", "lat"=>36.204482, "lon"=>-94.790888),
"/US/US-OK/kaw-city-17084" => array("type"=>"city", "pop"=>367, "name_en"=>"Kaw City", "lang"=>"en", "lat"=>36.764923, "lon"=>-96.863016),
"/US/US-OK/kellyville-17085" => array("type"=>"town", "pop"=>996, "name_en"=>"Kellyville", "lang"=>"en", "lat"=>35.940546, "lon"=>-96.216185),
"/US/US-OK/kemp-17086" => array("type"=>"town", "pop"=>148, "name_en"=>"Kemp", "lang"=>"en", "lat"=>33.768991, "lon"=>-96.354637),
"/US/US-OK/kendrick-17087" => array("type"=>"town", "pop"=>142, "name_ru"=>"Кендрик", "name_en"=>"Kendrick", "lang"=>"en", "lat"=>35.785281, "lon"=>-96.774871),
"/US/US-OK/kenefic-17088" => array("type"=>"town", "pop"=>203, "name_en"=>"Kenefic", "lang"=>"en", "lat"=>34.148559, "lon"=>-96.360303),
"/US/US-OK/keota-17089" => array("type"=>"town", "pop"=>527, "name_ru"=>"Кеота", "name_en"=>"Keota", "lang"=>"en", "lat"=>35.257914, "lon"=>-94.922403),
"/US/US-OK/ketchum-17090" => array("type"=>"town", "pop"=>295, "name_ru"=>"Кетчум", "name_en"=>"Ketchum", "lang"=>"en", "lat"=>36.525545, "lon"=>-95.025141),
"/US/US-OK/keyes-17091" => array("type"=>"town", "pop"=>331, "name_ru"=>"Кейес", "name_en"=>"Keyes", "lang"=>"en", "lat"=>36.807602, "lon"=>-102.251480),
"/US/US-OK/kiefer-17092" => array("type"=>"town", "pop"=>1583, "name_en"=>"Kiefer", "lang"=>"en", "lat"=>35.944125, "lon"=>-96.052617),
"/US/US-OK/kildare-17093" => array("type"=>"town", "pop"=>89, "name_en"=>"Kildare", "lang"=>"en", "lat"=>36.809013, "lon"=>-97.049619),
"/US/US-OK/kingfisher-17094" => array("type"=>"city", "pop"=>4343, "name_en"=>"Kingfisher", "lang"=>"en", "lat"=>35.844564, "lon"=>-97.938940),
"/US/US-OK/kingston-17095" => array("type"=>"town", "pop"=>1576, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>34.002250, "lon"=>-96.721705),
"/US/US-OK/kinta-17096" => array("type"=>"town", "pop"=>248, "name_en"=>"Kinta", "lang"=>"en", "lat"=>35.119485, "lon"=>-95.237903),
"/US/US-OK/kiowa-17097" => array("type"=>"town", "pop"=>705, "name_ru"=>"Кайова", "name_en"=>"Kiowa", "lang"=>"en", "lat"=>34.722088, "lon"=>-95.901830),
"/US/US-OK/knowles-17098" => array("type"=>"town", "pop"=>29, "name_en"=>"Knowles", "lang"=>"en", "lat"=>36.872615, "lon"=>-100.192474),
"/US/US-OK/konawa-17099" => array("type"=>"city", "pop"=>1400, "name_en"=>"Konawa", "lang"=>"en", "lat"=>34.957526, "lon"=>-96.753752),
"/US/US-OK/krebs-17100" => array("type"=>"city", "pop"=>2134, "name_en"=>"Krebs", "lang"=>"en", "lat"=>34.929558, "lon"=>-95.718456),
"/US/US-OK/kremlin-17101" => array("type"=>"town", "pop"=>234, "name_en"=>"Kremlin", "lang"=>"en", "lat"=>36.546580, "lon"=>-97.830865),
"/US/US-OK/lahoma-17102" => array("type"=>"town", "pop"=>578, "name_en"=>"Lahoma", "lang"=>"en", "lat"=>36.387894, "lon"=>-98.090273),
"/US/US-OK/lake-aluma-17103" => array("type"=>"town", "pop"=>93, "name_en"=>"Lake Aluma", "lang"=>"en", "lat"=>35.535583, "lon"=>-97.448299),
"/US/US-OK/lamar-17104" => array("type"=>"town", "pop"=>168, "name_ru"=>"Ламар", "name_en"=>"Lamar", "lang"=>"en", "lat"=>35.097891, "lon"=>-96.126811),
"/US/US-OK/lambert-17105" => array("type"=>"town", "pop"=>8, "name_ru"=>"Ламберт", "name_en"=>"Lambert", "lang"=>"en", "lat"=>36.683715, "lon"=>-98.423918),
"/US/US-OK/lamont-17106" => array("type"=>"town", "pop"=>408, "name_ru"=>"Ламонт", "name_en"=>"Lamont", "lang"=>"en", "lat"=>36.691133, "lon"=>-97.556061),
"/US/US-OK/langley-17107" => array("type"=>"town", "pop"=>684, "name_en"=>"Langley", "lang"=>"en", "lat"=>36.464930, "lon"=>-95.051395),
"/US/US-OK/langston-17108" => array("type"=>"town", "pop"=>1712, "name_ru"=>"Лангстон", "name_en"=>"Langston", "lang"=>"en", "lat"=>35.946038, "lon"=>-97.263581),
"/US/US-OK/laverne-17109" => array("type"=>"town", "pop"=>1018, "name_en"=>"Laverne", "lang"=>"en", "lat"=>36.706136, "lon"=>-99.896680),
"/US/US-OK/lawrence-creek-17110" => array("type"=>"town", "pop"=>126, "name_en"=>"Lawrence Creek", "lang"=>"en", "lat"=>36.085065, "lon"=>-96.424326),
"/US/US-OK/lawton-17111" => array("type"=>"city", "pop"=>90091, "name_ru"=>"Лотон", "name_en"=>"Lawton", "lang"=>"en", "lat"=>34.600400, "lon"=>-98.422821),
"/US/US-OK/le-flore-17113" => array("type"=>"town", "pop"=>172, "name_en"=>"Le Flore", "lang"=>"en", "lat"=>34.896849, "lon"=>-94.981140),
"/US/US-OK/leedey-17112" => array("type"=>"town", "pop"=>320, "name_en"=>"Leedey", "lang"=>"en", "lat"=>35.869480, "lon"=>-99.344995),
"/US/US-OK/lehigh-17114" => array("type"=>"city", "pop"=>301, "name_ru"=>"Лихай", "name_en"=>"Lehigh", "lang"=>"en", "lat"=>34.469763, "lon"=>-96.220458),
"/US/US-OK/lenapah-17115" => array("type"=>"town", "pop"=>302, "name_en"=>"Lenapah", "lang"=>"en", "lat"=>36.851409, "lon"=>-95.634929),
"/US/US-OK/leon-17116" => array("type"=>"town", "pop"=>98, "name_ru"=>"Леон", "name_en"=>"Leon", "lang"=>"en", "lat"=>33.877006, "lon"=>-97.429524),
"/US/US-OK/lexington-17117" => array("type"=>"town", "pop"=>2114, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>35.018178, "lon"=>-97.337394),
"/US/US-OK/liberty-17118" => array("type"=>"town", "pop"=>178, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>35.858724, "lon"=>-95.977197),
"/US/US-OK/lima-17119" => array("type"=>"town", "pop"=>71, "name_ru"=>"Лима", "name_en"=>"Lima", "lang"=>"en", "lat"=>35.173285, "lon"=>-96.598045),
"/US/US-OK/lindsay-17120" => array("type"=>"city", "pop"=>2911, "name_ru"=>"Линдсей", "name_en"=>"Lindsay", "lang"=>"en", "lat"=>34.839646, "lon"=>-97.609701),
"/US/US-OK/loco-17121" => array("type"=>"town", "pop"=>153, "name_en"=>"Loco", "lang"=>"en", "lat"=>34.329533, "lon"=>-97.680225),
"/US/US-OK/locust-grove-17122" => array("type"=>"town", "pop"=>1593, "name_ru"=>"Локуст Гров", "name_en"=>"Locust Grove", "lang"=>"en", "lat"=>36.199417, "lon"=>-95.165641),
"/US/US-OK/lone-grove-17123" => array("type"=>"city", "pop"=>5276, "name_en"=>"Lone Grove", "lang"=>"en", "lat"=>34.189753, "lon"=>-97.273841),
"/US/US-OK/lone-wolf-17124" => array("type"=>"town", "pop"=>451, "name_en"=>"Lone Wolf", "lang"=>"en", "lat"=>34.990148, "lon"=>-99.245675),
"/US/US-OK/longdale-17125" => array("type"=>"town", "pop"=>306, "name_en"=>"Longdale", "lang"=>"en", "lat"=>36.134266, "lon"=>-98.550212),
"/US/US-OK/longtown-17126" => array("type"=>"cdp", "pop"=>1641, "name_ru"=>"Лонгтаун", "name_en"=>"Longtown", "lang"=>"en", "lat"=>35.242185, "lon"=>-95.514885),
"/US/US-OK/lookeba-17127" => array("type"=>"town", "pop"=>127, "name_en"=>"Lookeba", "lang"=>"en", "lat"=>35.364760, "lon"=>-98.365400),
"/US/US-OK/lotsee-17128" => array("type"=>"town", "pop"=>11, "name_en"=>"Lotsee", "lang"=>"en", "lat"=>36.133351, "lon"=>-96.209170),
"/US/US-OK/loveland-17129" => array("type"=>"town", "pop"=>12, "name_ru"=>"Лавленд", "name_en"=>"Loveland", "lang"=>"en", "lat"=>34.304226, "lon"=>-98.770812),
"/US/US-OK/loyal-17130" => array("type"=>"town", "pop"=>83, "name_en"=>"Loyal", "lang"=>"en", "lat"=>35.972798, "lon"=>-98.117340),
"/US/US-OK/luther-17131" => array("type"=>"town", "pop"=>1140, "name_ru"=>"Лютер", "name_en"=>"Luther", "lang"=>"en", "lat"=>35.664264, "lon"=>-97.180413),
"/US/US-OK/macomb-17137" => array("type"=>"town", "pop"=>63, "name_ru"=>"Макомб", "name_en"=>"Macomb", "lang"=>"en", "lat"=>35.147768, "lon"=>-97.008091),
"/US/US-OK/madill-17138" => array("type"=>"city", "pop"=>3770, "name_en"=>"Madill", "lang"=>"en", "lat"=>34.091534, "lon"=>-96.772323),
"/US/US-OK/manchester-17139" => array("type"=>"town", "pop"=>91, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>36.994431, "lon"=>-98.035039),
"/US/US-OK/mangum-17140" => array("type"=>"city", "pop"=>2711, "name_en"=>"Mangum", "lang"=>"en", "lat"=>34.878785, "lon"=>-99.503161),
"/US/US-OK/manitou-17141" => array("type"=>"town", "pop"=>267, "name_en"=>"Manitou", "lang"=>"en", "lat"=>34.506799, "lon"=>-98.980491),
"/US/US-OK/mannford-17142" => array("type"=>"town", "pop"=>2861, "name_en"=>"Mannford", "lang"=>"en", "lat"=>36.121683, "lon"=>-96.329452),
"/US/US-OK/mannsville-17143" => array("type"=>"town", "pop"=>555, "name_en"=>"Mannsville", "lang"=>"en", "lat"=>34.187433, "lon"=>-96.879962),
"/US/US-OK/maramec-17144" => array("type"=>"town", "pop"=>102, "name_en"=>"Maramec", "lang"=>"en", "lat"=>36.242459, "lon"=>-96.680694),
"/US/US-OK/marble-city-17145" => array("type"=>"town", "pop"=>247, "name_en"=>"Marble City", "lang"=>"en", "lat"=>35.583099, "lon"=>-94.817158),
"/US/US-OK/marietta-17146" => array("type"=>"city", "pop"=>2530, "name_ru"=>"Мариетта", "name_en"=>"Marietta", "lang"=>"en", "lat"=>33.933919, "lon"=>-97.123253),
"/US/US-OK/marland-17147" => array("type"=>"town", "pop"=>279, "name_en"=>"Marland", "lang"=>"en", "lat"=>36.561263, "lon"=>-97.152726),
"/US/US-OK/marlow-17148" => array("type"=>"city", "pop"=>4599, "name_en"=>"Marlow", "lang"=>"en", "lat"=>34.639760, "lon"=>-97.961414),
"/US/US-OK/marshall-17149" => array("type"=>"town", "pop"=>288, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>36.155412, "lon"=>-97.624788),
"/US/US-OK/martha-17150" => array("type"=>"town", "pop"=>186, "name_en"=>"Martha", "lang"=>"en", "lat"=>34.726503, "lon"=>-99.386031),
"/US/US-OK/maud-17151" => array("type"=>"city", "pop"=>1150, "name_en"=>"Maud", "lang"=>"en", "lat"=>35.132728, "lon"=>-96.778456),
"/US/US-OK/may-17152" => array("type"=>"town", "pop"=>31, "name_en"=>"May", "lang"=>"en", "lat"=>36.616502, "lon"=>-99.748965),
"/US/US-OK/maysville-17153" => array("type"=>"town", "pop"=>1301, "name_ru"=>"Мейсвилл", "name_en"=>"Maysville", "lang"=>"en", "lat"=>34.816732, "lon"=>-97.410168),
"/US/US-OK/mcalester-17132" => array("type"=>"city", "pop"=>18388, "name_en"=>"McAlester", "lang"=>"en", "lat"=>34.924255, "lon"=>-95.760270),
"/US/US-OK/mcbride-17133" => array("type"=>"town", "pop"=>80, "name_ru"=>"МакБрайд", "name_en"=>"McBride", "lang"=>"en", "lat"=>33.935903, "lon"=>-96.635980),
"/US/US-OK/mccord-17134" => array("type"=>"cdp", "pop"=>2170, "name_en"=>"McCord", "lang"=>"en", "lat"=>36.678685, "lon"=>-97.036898),
"/US/US-OK/mccurtain-17135" => array("type"=>"town", "pop"=>476, "name_en"=>"McCurtain", "lang"=>"en", "lat"=>35.150627, "lon"=>-94.967295),
"/US/US-OK/mcloud-17136" => array("type"=>"town", "pop"=>4347, "name_en"=>"McLoud", "lang"=>"en", "lat"=>35.412134, "lon"=>-97.098095),
"/US/US-OK/mead-17154" => array("type"=>"town", "pop"=>129, "name_ru"=>"Мид", "name_en"=>"Mead", "lang"=>"en", "lat"=>34.000654, "lon"=>-96.511274),
"/US/US-OK/medford-17155" => array("type"=>"city", "pop"=>993, "name_ru"=>"Медфорд", "name_en"=>"Medford", "lang"=>"en", "lat"=>36.804334, "lon"=>-97.737882),
"/US/US-OK/medicine-park-17156" => array("type"=>"town", "pop"=>378, "name_en"=>"Medicine Park", "lang"=>"en", "lat"=>34.728570, "lon"=>-98.464015),
"/US/US-OK/meeker-17157" => array("type"=>"town", "pop"=>981, "name_ru"=>"Микер", "name_en"=>"Meeker", "lang"=>"en", "lat"=>35.485000, "lon"=>-96.893642),
"/US/US-OK/meno-17158" => array("type"=>"town", "pop"=>185, "name_en"=>"Meno", "lang"=>"en", "lat"=>36.387770, "lon"=>-98.177251),
"/US/US-OK/meridian-17159" => array("type"=>"town", "pop"=>61, "name_ru"=>"Меридиан", "name_en"=>"Meridian", "lang"=>"en", "lat"=>35.843492, "lon"=>-97.246312),
"/US/US-OK/meridian-17160" => array("type"=>"cdp", "pop"=>1471, "name_ru"=>"Меридиан", "name_en"=>"Meridian", "lang"=>"en", "lat"=>34.427101, "lon"=>-97.977792),
"/US/US-OK/miami-17161" => array("type"=>"city", "pop"=>13027, "name_ru"=>"Майами", "name_en"=>"Miami", "lang"=>"en", "lat"=>36.886846, "lon"=>-94.876025),
"/US/US-OK/midwest-city-17162" => array("type"=>"city", "pop"=>56394, "name_en"=>"Midwest City", "lang"=>"en", "lat"=>35.462600, "lon"=>-97.369385),
"/US/US-OK/milburn-17163" => array("type"=>"town", "pop"=>301, "name_en"=>"Milburn", "lang"=>"en", "lat"=>34.240915, "lon"=>-96.551649),
"/US/US-OK/mill-creek-17164" => array("type"=>"town", "pop"=>328, "name_ru"=>"Милл Крик", "name_en"=>"Mill Creek", "lang"=>"en", "lat"=>34.402599, "lon"=>-96.824882),
"/US/US-OK/millerton-17165" => array("type"=>"town", "pop"=>351, "name_ru"=>"Миллертон", "name_en"=>"Millerton", "lang"=>"en", "lat"=>33.984706, "lon"=>-95.015063),
"/US/US-OK/minco-17166" => array("type"=>"city", "pop"=>1802, "name_en"=>"Minco", "lang"=>"en", "lat"=>35.315504, "lon"=>-97.951222),
"/US/US-OK/moffett-17167" => array("type"=>"town", "pop"=>174, "name_en"=>"Moffett", "lang"=>"en", "lat"=>35.389700, "lon"=>-94.447246),
"/US/US-OK/moore-17168" => array("type"=>"city", "pop"=>52615, "name_ru"=>"Мур", "name_en"=>"Moore", "lang"=>"en", "lat"=>35.329496, "lon"=>-97.474699),
"/US/US-OK/mooreland-17169" => array("type"=>"town", "pop"=>1254, "name_ru"=>"Мурленд", "name_en"=>"Mooreland", "lang"=>"en", "lat"=>36.438488, "lon"=>-99.205034),
"/US/US-OK/morris-17170" => array("type"=>"city", "pop"=>1309, "name_ru"=>"Моррис", "name_en"=>"Morris", "lang"=>"en", "lat"=>35.613292, "lon"=>-95.860780),
"/US/US-OK/morrison-17171" => array("type"=>"town", "pop"=>649, "name_ru"=>"Моррисон", "name_en"=>"Morrison", "lang"=>"en", "lat"=>36.294650, "lon"=>-97.011591),
"/US/US-OK/mounds-17172" => array("type"=>"town", "pop"=>1269, "name_ru"=>"Маундс", "name_en"=>"Mounds", "lang"=>"en", "lat"=>35.874660, "lon"=>-96.059731),
"/US/US-OK/mountain-park-17173" => array("type"=>"town", "pop"=>367, "name_ru"=>"Маунтин Парк", "name_en"=>"Mountain Park", "lang"=>"en", "lat"=>34.698453, "lon"=>-98.951609),
"/US/US-OK/mountain-view-17174" => array("type"=>"town", "pop"=>786, "name_ru"=>"Маунтин Вью", "name_en"=>"Mountain View", "lang"=>"en", "lat"=>35.098656, "lon"=>-98.750406),
"/US/US-OK/muldrow-17175" => array("type"=>"town", "pop"=>3153, "name_en"=>"Muldrow", "lang"=>"en", "lat"=>35.404070, "lon"=>-94.600845),
"/US/US-OK/mule-barn-17176" => array("type"=>"town", "pop"=>0, "name_en"=>"Mule Barn", "lang"=>"en", "lat"=>36.217367, "lon"=>-96.311416),
"/US/US-OK/mulhall-17177" => array("type"=>"town", "pop"=>276, "name_en"=>"Mulhall", "lang"=>"en", "lat"=>36.064252, "lon"=>-97.404762),
"/US/US-OK/muskogee-17178" => array("type"=>"city", "pop"=>40099, "name_en"=>"Muskogee", "lang"=>"en", "lat"=>35.748509, "lon"=>-95.351878),
"/US/US-OK/mustang-17179" => array("type"=>"city", "pop"=>17727, "name_en"=>"Mustang", "lang"=>"en", "lat"=>35.392590, "lon"=>-97.722737),
"/US/US-OK/mutual-17180" => array("type"=>"town", "pop"=>80, "name_en"=>"Mutual", "lang"=>"en", "lat"=>36.229777, "lon"=>-99.167325),
"/US/US-OK/nardin-17181" => array("type"=>"town", "pop"=>75, "name_en"=>"Nardin", "lang"=>"en", "lat"=>36.806289, "lon"=>-97.448048),
"/US/US-OK/nash-17182" => array("type"=>"town", "pop"=>197, "name_en"=>"Nash", "lang"=>"en", "lat"=>36.665366, "lon"=>-98.052095),
"/US/US-OK/new-alluwe-17183" => array("type"=>"town", "pop"=>96, "name_en"=>"New Alluwe", "lang"=>"en", "lat"=>36.609420, "lon"=>-95.488945),
"/US/US-OK/new-cordell-17185" => array("type"=>"city", "pop"=>2932, "name_en"=>"New Cordell", "lang"=>"en", "lat"=>35.297523, "lon"=>-98.981045),
"/US/US-OK/new-tulsa-17187" => array("type"=>"town", "pop"=>272, "name_en"=>"New Tulsa", "lang"=>"en", "lat"=>36.099500, "lon"=>-95.735054),
"/US/US-OK/newcastle-17184" => array("type"=>"city", "pop"=>7250, "name_en"=>"Newcastle", "lang"=>"en", "lat"=>35.247720, "lon"=>-97.599404),
"/US/US-OK/newkirk-17186" => array("type"=>"city", "pop"=>2124, "name_en"=>"Newkirk", "lang"=>"en", "lat"=>36.881712, "lon"=>-97.055420),
"/US/US-OK/nichols-hills-17188" => array("type"=>"city", "pop"=>4033, "name_en"=>"Nichols Hills", "lang"=>"en", "lat"=>35.546150, "lon"=>-97.544394),
"/US/US-OK/nicoma-park-17189" => array("type"=>"city", "pop"=>2401, "name_en"=>"Nicoma Park", "lang"=>"en", "lat"=>35.491139, "lon"=>-97.324018),
"/US/US-OK/noble-17190" => array("type"=>"town", "pop"=>5812, "name_ru"=>"Ноубл", "name_en"=>"Noble", "lang"=>"en", "lat"=>35.140935, "lon"=>-97.371847),
"/US/US-OK/norge-17191" => array("type"=>"town", "pop"=>92, "name_en"=>"Norge", "lang"=>"en", "lat"=>34.987861, "lon"=>-97.995676),
"/US/US-OK/norman-17192" => array("type"=>"city", "pop"=>106957, "name_ru"=>"Норман", "name_en"=>"Norman", "lang"=>"en", "lat"=>35.225450, "lon"=>-97.343987),
"/US/US-OK/north-enid-17193" => array("type"=>"town", "pop"=>862, "name_en"=>"North Enid", "lang"=>"en", "lat"=>36.446650, "lon"=>-97.861524),
"/US/US-OK/north-miami-17194" => array("type"=>"town", "pop"=>424, "name_ru"=>"Норт Майами", "name_en"=>"North Miami", "lang"=>"en", "lat"=>36.919495, "lon"=>-94.879071),
"/US/US-OK/nowata-17195" => array("type"=>"city", "pop"=>3996, "name_en"=>"Nowata", "lang"=>"en", "lat"=>36.697429, "lon"=>-95.638708),
"/US/US-OK/oak-grove-17196" => array("type"=>"town", "pop"=>20, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>36.205550, "lon"=>-96.337999),
"/US/US-OK/oakhurst-17197" => array("type"=>"cdp", "pop"=>3030, "name_ru"=>"Оукхерст", "name_en"=>"Oakhurst", "lang"=>"en", "lat"=>36.081700, "lon"=>-96.065549),
"/US/US-OK/oakland-17198" => array("type"=>"town", "pop"=>770, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>34.100008, "lon"=>-96.793118),
"/US/US-OK/oaks-17199" => array("type"=>"town", "pop"=>367, "name_ru"=>"Оукс", "name_en"=>"Oaks", "lang"=>"en", "lat"=>36.168157, "lon"=>-94.852976),
"/US/US-OK/oakwood-17200" => array("type"=>"town", "pop"=>67, "name_ru"=>"Оуквуд", "name_en"=>"Oakwood", "lang"=>"en", "lat"=>35.930827, "lon"=>-98.703180),
"/US/US-OK/ochelata-17201" => array("type"=>"town", "pop"=>509, "name_en"=>"Ochelata", "lang"=>"en", "lat"=>36.601196, "lon"=>-95.982013),
"/US/US-OK/oilton-17202" => array("type"=>"city", "pop"=>1124, "name_en"=>"Oilton", "lang"=>"en", "lat"=>36.084979, "lon"=>-96.584765),
"/US/US-OK/okarche-17203" => array("type"=>"town", "pop"=>1145, "name_en"=>"Okarche", "lang"=>"en", "lat"=>35.727539, "lon"=>-97.977115),
"/US/US-OK/okay-17204" => array("type"=>"town", "pop"=>589, "name_en"=>"Okay", "lang"=>"en", "lat"=>35.852218, "lon"=>-95.314296),
"/US/US-OK/okeene-17205" => array("type"=>"town", "pop"=>1159, "name_en"=>"Okeene", "lang"=>"en", "lat"=>36.116480, "lon"=>-98.318075),
"/US/US-OK/okemah-17206" => array("type"=>"city", "pop"=>2911, "name_en"=>"Okemah", "lang"=>"en", "lat"=>35.429660, "lon"=>-96.298696),
"/US/US-OK/oklahoma-city-17207" => array("type"=>"city", "pop"=>551789, "name_en"=>"Oklahoma City", "lang"=>"en", "lat"=>35.467050, "lon"=>-97.513491),
"/US/US-OK/okmulgee-17208" => array("type"=>"city", "pop"=>12634, "name_en"=>"Okmulgee", "lang"=>"en", "lat"=>35.629954, "lon"=>-95.956240),
"/US/US-OK/oktaha-17209" => array("type"=>"town", "pop"=>336, "name_en"=>"Oktaha", "lang"=>"en", "lat"=>35.577673, "lon"=>-95.476508),
"/US/US-OK/olustee-17210" => array("type"=>"town", "pop"=>617, "name_en"=>"Olustee", "lang"=>"en", "lat"=>34.547901, "lon"=>-99.423941),
"/US/US-OK/oolagah-(oologah)-17211" => array("type"=>"town", "pop"=>828, "name_en"=>"Oolagah (Oologah)", "lang"=>"en", "lat"=>36.446541, "lon"=>-95.710085),
"/US/US-OK/optima-17212" => array("type"=>"town", "pop"=>273, "name_en"=>"Optima", "lang"=>"en", "lat"=>36.758337, "lon"=>-101.354022),
"/US/US-OK/orlando-17213" => array("type"=>"town", "pop"=>224, "name_ru"=>"Орландо", "name_en"=>"Orlando", "lang"=>"en", "lat"=>36.148282, "lon"=>-97.375839),
"/US/US-OK/osage-17214" => array("type"=>"town", "pop"=>189, "name_ru"=>"Осейдж", "name_en"=>"Osage", "lang"=>"en", "lat"=>36.293749, "lon"=>-96.417480),
"/US/US-OK/owasso-17215" => array("type"=>"city", "pop"=>27467, "name_en"=>"Owasso", "lang"=>"en", "lat"=>36.280107, "lon"=>-95.835050),
"/US/US-OK/paden-17216" => array("type"=>"town", "pop"=>419, "name_ru"=>"Пейден", "name_en"=>"Paden", "lang"=>"en", "lat"=>35.508270, "lon"=>-96.566895),
"/US/US-OK/panama-17217" => array("type"=>"town", "pop"=>1399, "name_ru"=>"Панама", "name_en"=>"Panama", "lang"=>"en", "lat"=>35.170880, "lon"=>-94.669875),
"/US/US-OK/paoli-17218" => array("type"=>"town", "pop"=>660, "name_ru"=>"Паоли", "name_en"=>"Paoli", "lang"=>"en", "lat"=>34.826919, "lon"=>-97.262128),
"/US/US-OK/paradise-hill-17219" => array("type"=>"town", "pop"=>100, "name_en"=>"Paradise Hill", "lang"=>"en", "lat"=>35.610296, "lon"=>-95.071359),
"/US/US-OK/pauls-valley-17220" => array("type"=>"city", "pop"=>6121, "name_en"=>"Pauls Valley", "lang"=>"en", "lat"=>34.725340, "lon"=>-97.224444),
"/US/US-OK/pawhuska-17221" => array("type"=>"city", "pop"=>3426, "name_en"=>"Pawhuska", "lang"=>"en", "lat"=>36.669436, "lon"=>-96.331254),
"/US/US-OK/pawnee-17222" => array("type"=>"city", "pop"=>2149, "name_ru"=>"Пони", "name_en"=>"Pawnee", "lang"=>"en", "lat"=>36.335072, "lon"=>-96.801155),
"/US/US-OK/pensacola-17223" => array("type"=>"town", "pop"=>73, "name_ru"=>"Пенсакола", "name_en"=>"Pensacola", "lang"=>"en", "lat"=>36.455369, "lon"=>-95.128582),
"/US/US-OK/peoria-17224" => array("type"=>"town", "pop"=>137, "name_ru"=>"Пеория", "name_en"=>"Peoria", "lang"=>"en", "lat"=>36.915964, "lon"=>-94.669218),
"/US/US-OK/perkins-17225" => array("type"=>"town", "pop"=>2396, "name_en"=>"Perkins", "lang"=>"en", "lat"=>35.976582, "lon"=>-97.028127),
"/US/US-OK/perry-17226" => array("type"=>"city", "pop"=>5060, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>36.287665, "lon"=>-97.304224),
"/US/US-OK/phillips-17227" => array("type"=>"town", "pop"=>143, "name_en"=>"Phillips", "lang"=>"en", "lat"=>34.502301, "lon"=>-96.223892),
"/US/US-OK/picher-17228" => array("type"=>"city", "pop"=>1555, "name_en"=>"Picher", "lang"=>"en", "lat"=>36.986110, "lon"=>-94.831803),
"/US/US-OK/piedmont-17229" => array("type"=>"city", "pop"=>5504, "name_ru"=>"Пидмонт", "name_en"=>"Piedmont", "lang"=>"en", "lat"=>35.674544, "lon"=>-97.753697),
"/US/US-OK/pink-17230" => array("type"=>"town", "pop"=>1226, "name_en"=>"Pink", "lang"=>"en", "lat"=>35.232075, "lon"=>-97.106770),
"/US/US-OK/pittsburg-17231" => array("type"=>"town", "pop"=>285, "name_ru"=>"Питсбург", "name_en"=>"Pittsburg", "lang"=>"en", "lat"=>34.710876, "lon"=>-95.850732),
"/US/US-OK/pocola-17232" => array("type"=>"town", "pop"=>4464, "name_en"=>"Pocola", "lang"=>"en", "lat"=>35.241722, "lon"=>-94.478844),
"/US/US-OK/ponca-city-17233" => array("type"=>"city", "pop"=>24507, "name_en"=>"Ponca City", "lang"=>"en", "lat"=>36.723637, "lon"=>-97.068813),
"/US/US-OK/pond-creek-17234" => array("type"=>"city", "pop"=>759, "name_en"=>"Pond Creek", "lang"=>"en", "lat"=>36.667483, "lon"=>-97.801733),
"/US/US-OK/porter-17235" => array("type"=>"town", "pop"=>595, "name_ru"=>"Портер", "name_en"=>"Porter", "lang"=>"en", "lat"=>35.869214, "lon"=>-95.522857),
"/US/US-OK/porum-17236" => array("type"=>"town", "pop"=>735, "name_en"=>"Porum", "lang"=>"en", "lat"=>35.356977, "lon"=>-95.262761),
"/US/US-OK/poteau-17237" => array("type"=>"city", "pop"=>8258, "name_en"=>"Poteau", "lang"=>"en", "lat"=>35.041028, "lon"=>-94.634835),
"/US/US-OK/prague-17238" => array("type"=>"city", "pop"=>2127, "name_en"=>"Prague", "lang"=>"en", "lat"=>35.488220, "lon"=>-96.694267),
"/US/US-OK/prue-17239" => array("type"=>"town", "pop"=>432, "name_en"=>"Prue", "lang"=>"en", "lat"=>36.249017, "lon"=>-96.265640),
"/US/US-OK/pryor-creek-17240" => array("type"=>"city", "pop"=>9293, "name_en"=>"Pryor Creek", "lang"=>"en", "lat"=>36.294779, "lon"=>-95.319130),
"/US/US-OK/purcell-17241" => array("type"=>"city", "pop"=>6129, "name_ru"=>"Перселл", "name_en"=>"Purcell", "lang"=>"en", "lat"=>35.014542, "lon"=>-97.371522),
"/US/US-OK/putnam-17242" => array("type"=>"town", "pop"=>43, "name_en"=>"Putnam", "lang"=>"en", "lat"=>35.999611, "lon"=>-99.234146),
"/US/US-OK/quapaw-17243" => array("type"=>"town", "pop"=>951, "name_en"=>"Quapaw", "lang"=>"en", "lat"=>36.952130, "lon"=>-94.789389),
"/US/US-OK/quay-17244" => array("type"=>"town", "pop"=>59, "name_en"=>"Quay", "lang"=>"en", "lat"=>36.160070, "lon"=>-96.709668),
"/US/US-OK/quinlan-17245" => array("type"=>"town", "pop"=>23, "name_en"=>"Quinlan", "lang"=>"en", "lat"=>36.454793, "lon"=>-99.045467),
"/US/US-OK/quinton-17246" => array("type"=>"town", "pop"=>1087, "name_en"=>"Quinton", "lang"=>"en", "lat"=>35.126875, "lon"=>-95.366319),
"/US/US-OK/ralston-17247" => array("type"=>"town", "pop"=>346, "name_ru"=>"Ралстон", "name_en"=>"Ralston", "lang"=>"en", "lat"=>36.503263, "lon"=>-96.736296),
"/US/US-OK/ramona-17248" => array("type"=>"town", "pop"=>581, "name_ru"=>"Рамона", "name_en"=>"Ramona", "lang"=>"en", "lat"=>36.531389, "lon"=>-95.925286),
"/US/US-OK/randlett-17249" => array("type"=>"town", "pop"=>489, "name_en"=>"Randlett", "lang"=>"en", "lat"=>34.177743, "lon"=>-98.462724),
"/US/US-OK/ratliff-city-17250" => array("type"=>"town", "pop"=>135, "name_en"=>"Ratliff City", "lang"=>"en", "lat"=>34.448628, "lon"=>-97.511957),
"/US/US-OK/rattan-17251" => array("type"=>"town", "pop"=>243, "name_en"=>"Rattan", "lang"=>"en", "lat"=>34.197950, "lon"=>-95.414778),
"/US/US-OK/ravia-17252" => array("type"=>"town", "pop"=>439, "name_en"=>"Ravia", "lang"=>"en", "lat"=>34.240551, "lon"=>-96.755707),
"/US/US-OK/red-oak-17254" => array("type"=>"town", "pop"=>563, "name_ru"=>"Ред Оук", "name_en"=>"Red Oak", "lang"=>"en", "lat"=>34.952407, "lon"=>-95.080657),
"/US/US-OK/red-rock-17255" => array("type"=>"town", "pop"=>291, "name_en"=>"Red Rock", "lang"=>"en", "lat"=>36.459893, "lon"=>-97.178905),
"/US/US-OK/redbird-17253" => array("type"=>"town", "pop"=>149, "name_en"=>"Redbird", "lang"=>"en", "lat"=>35.886717, "lon"=>-95.587934),
"/US/US-OK/renfrow-17256" => array("type"=>"town", "pop"=>14, "name_en"=>"Renfrow", "lang"=>"en", "lat"=>36.924188, "lon"=>-97.656426),
"/US/US-OK/rentiesville-17257" => array("type"=>"town", "pop"=>102, "name_en"=>"Rentiesville", "lang"=>"en", "lat"=>35.520376, "lon"=>-95.492957),
"/US/US-OK/reydon-17258" => array("type"=>"town", "pop"=>176, "name_en"=>"Reydon", "lang"=>"en", "lat"=>35.650062, "lon"=>-99.923673),
"/US/US-OK/ringling-17259" => array("type"=>"town", "pop"=>1044, "name_en"=>"Ringling", "lang"=>"en", "lat"=>34.176703, "lon"=>-97.590420),
"/US/US-OK/ringwood-17260" => array("type"=>"city", "pop"=>406, "name_en"=>"Ringwood", "lang"=>"en", "lat"=>36.382174, "lon"=>-98.240821),
"/US/US-OK/ripley-17261" => array("type"=>"town", "pop"=>466, "name_ru"=>"Рипли", "name_en"=>"Ripley", "lang"=>"en", "lat"=>36.017485, "lon"=>-96.904873),
"/US/US-OK/rock-island-17262" => array("type"=>"town", "pop"=>735, "name_ru"=>"Рок Айленд", "name_en"=>"Rock Island", "lang"=>"en", "lat"=>35.183817, "lon"=>-94.482339),
"/US/US-OK/rocky-17263" => array("type"=>"town", "pop"=>179, "name_en"=>"Rocky", "lang"=>"en", "lat"=>35.156355, "lon"=>-99.058518),
"/US/US-OK/roff-17264" => array("type"=>"town", "pop"=>759, "name_en"=>"Roff", "lang"=>"en", "lat"=>34.628891, "lon"=>-96.841118),
"/US/US-OK/roland-17265" => array("type"=>"town", "pop"=>3398, "name_ru"=>"Роланд", "name_en"=>"Roland", "lang"=>"en", "lat"=>35.415800, "lon"=>-94.514526),
"/US/US-OK/roosevelt-17266" => array("type"=>"town", "pop"=>263, "name_ru"=>"Рузвельт", "name_en"=>"Roosevelt", "lang"=>"en", "lat"=>34.848048, "lon"=>-99.021957),
"/US/US-OK/rosedale-17267" => array("type"=>"town", "pop"=>69, "name_ru"=>"Роуздейл", "name_en"=>"Rosedale", "lang"=>"en", "lat"=>34.918717, "lon"=>-97.184567),
"/US/US-OK/rosston-17268" => array("type"=>"town", "pop"=>63, "name_ru"=>"Росстон", "name_en"=>"Rosston", "lang"=>"en", "lat"=>36.812219, "lon"=>-99.929918),
"/US/US-OK/rush-springs-17269" => array("type"=>"town", "pop"=>1344, "name_en"=>"Rush Springs", "lang"=>"en", "lat"=>34.778355, "lon"=>-97.956784),
"/US/US-OK/ryan-17270" => array("type"=>"town", "pop"=>825, "name_ru"=>"Райан", "name_en"=>"Ryan", "lang"=>"en", "lat"=>34.021583, "lon"=>-97.953987),
"/US/US-OK/salina-17272" => array("type"=>"town", "pop"=>1471, "name_ru"=>"Салина", "name_en"=>"Salina", "lang"=>"en", "lat"=>36.290866, "lon"=>-95.150692),
"/US/US-OK/sallisaw-17273" => array("type"=>"city", "pop"=>8742, "name_en"=>"Sallisaw", "lang"=>"en", "lat"=>35.460707, "lon"=>-94.806726),
"/US/US-OK/sand-springs-17274" => array("type"=>"city", "pop"=>18489, "name_en"=>"Sand Springs", "lang"=>"en", "lat"=>36.139517, "lon"=>-96.130042),
"/US/US-OK/sapulpa-17275" => array("type"=>"city", "pop"=>21173, "name_en"=>"Sapulpa", "lang"=>"en", "lat"=>35.999700, "lon"=>-96.095079),
"/US/US-OK/sasakwa-17276" => array("type"=>"town", "pop"=>145, "name_en"=>"Sasakwa", "lang"=>"en", "lat"=>34.946585, "lon"=>-96.524968),
"/US/US-OK/savanna-17277" => array("type"=>"town", "pop"=>750, "name_ru"=>"Саванна", "name_en"=>"Savanna", "lang"=>"en", "lat"=>34.829424, "lon"=>-95.838356),
"/US/US-OK/sayre-17278" => array("type"=>"city", "pop"=>4336, "name_en"=>"Sayre", "lang"=>"en", "lat"=>35.297298, "lon"=>-99.634821),
"/US/US-OK/seiling-17279" => array("type"=>"city", "pop"=>782, "name_en"=>"Seiling", "lang"=>"en", "lat"=>36.149570, "lon"=>-98.924744),
"/US/US-OK/seminole-17280" => array("type"=>"city", "pop"=>6804, "name_ru"=>"Семиноул", "name_en"=>"Seminole", "lang"=>"en", "lat"=>35.235838, "lon"=>-96.649696),
"/US/US-OK/sentinel-17281" => array("type"=>"city", "pop"=>881, "name_en"=>"Sentinel", "lang"=>"en", "lat"=>35.156606, "lon"=>-99.173472),
"/US/US-OK/shady-grove-17282" => array("type"=>"town", "pop"=>44, "name_en"=>"Shady Grove", "lang"=>"en", "lat"=>36.189190, "lon"=>-96.287638),
"/US/US-OK/shady-point-17283" => array("type"=>"town", "pop"=>873, "name_en"=>"Shady Point", "lang"=>"en", "lat"=>35.135205, "lon"=>-94.661654),
"/US/US-OK/shamrock-17284" => array("type"=>"town", "pop"=>123, "name_en"=>"Shamrock", "lang"=>"en", "lat"=>35.910339, "lon"=>-96.577394),
"/US/US-OK/sharon-17285" => array("type"=>"town", "pop"=>129, "name_ru"=>"Шэрон", "name_en"=>"Sharon", "lang"=>"en", "lat"=>36.275439, "lon"=>-99.338113),
"/US/US-OK/shattuck-17286" => array("type"=>"town", "pop"=>1227, "name_en"=>"Shattuck", "lang"=>"en", "lat"=>36.272567, "lon"=>-99.877078),
"/US/US-OK/shawnee-17287" => array("type"=>"city", "pop"=>30562, "name_ru"=>"Шони", "name_en"=>"Shawnee", "lang"=>"en", "lat"=>35.365332, "lon"=>-96.968860),
"/US/US-OK/shidler-17288" => array("type"=>"town", "pop"=>510, "name_en"=>"Shidler", "lang"=>"en", "lat"=>36.782910, "lon"=>-96.659654),
"/US/US-OK/silo-17289" => array("type"=>"town", "pop"=>296, "name_en"=>"Silo", "lang"=>"en", "lat"=>34.035051, "lon"=>-96.475364),
"/US/US-OK/skedee-17290" => array("type"=>"town", "pop"=>100, "name_en"=>"Skedee", "lang"=>"en", "lat"=>36.381161, "lon"=>-96.703632),
"/US/US-OK/skiatook-17291" => array("type"=>"town", "pop"=>6820, "name_en"=>"Skiatook", "lang"=>"en", "lat"=>36.369430, "lon"=>-96.039559),
"/US/US-OK/slaughterville-17292" => array("type"=>"town", "pop"=>3949, "name_en"=>"Slaughterville", "lang"=>"en", "lat"=>35.096968, "lon"=>-97.282443),
"/US/US-OK/slick-17293" => array("type"=>"town", "pop"=>150, "name_en"=>"Slick", "lang"=>"en", "lat"=>35.776440, "lon"=>-96.266566),
"/US/US-OK/smith-village-17294" => array("type"=>"town", "pop"=>40, "name_en"=>"Smith Village", "lang"=>"en", "lat"=>35.451010, "lon"=>-97.456901),
"/US/US-OK/smithville-17295" => array("type"=>"town", "pop"=>111, "name_ru"=>"Смитвилл", "name_en"=>"Smithville", "lang"=>"en", "lat"=>34.468113, "lon"=>-94.646109),
"/US/US-OK/snyder-17296" => array("type"=>"city", "pop"=>1378, "name_en"=>"Snyder", "lang"=>"en", "lat"=>34.656180, "lon"=>-98.952197),
"/US/US-OK/soper-17297" => array("type"=>"town", "pop"=>291, "name_en"=>"Soper", "lang"=>"en", "lat"=>34.032322, "lon"=>-95.696642),
"/US/US-OK/south-coffeyville-17298" => array("type"=>"town", "pop"=>787, "name_en"=>"South Coffeyville", "lang"=>"en", "lat"=>36.993984, "lon"=>-95.622056),
"/US/US-OK/sparks-17299" => array("type"=>"town", "pop"=>136, "name_ru"=>"Спаркс", "name_en"=>"Sparks", "lang"=>"en", "lat"=>35.610852, "lon"=>-96.818250),
"/US/US-OK/spavinaw-17300" => array("type"=>"town", "pop"=>580, "name_en"=>"Spavinaw", "lang"=>"en", "lat"=>36.393264, "lon"=>-95.051131),
"/US/US-OK/spencer-17301" => array("type"=>"city", "pop"=>4024, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>35.507700, "lon"=>-97.370349),
"/US/US-OK/sperry-17302" => array("type"=>"town", "pop"=>1032, "name_en"=>"Sperry", "lang"=>"en", "lat"=>36.296669, "lon"=>-95.986749),
"/US/US-OK/spiro-17303" => array("type"=>"city", "pop"=>2323, "name_en"=>"Spiro", "lang"=>"en", "lat"=>35.242355, "lon"=>-94.618007),
"/US/US-OK/sportsmen-acres-17304" => array("type"=>"town", "pop"=>210, "name_en"=>"Sportsmen Acres", "lang"=>"en", "lat"=>36.245731, "lon"=>-95.251181),
"/US/US-OK/springer-17305" => array("type"=>"town", "pop"=>600, "name_en"=>"Springer", "lang"=>"en", "lat"=>34.286497, "lon"=>-97.120370),
"/US/US-OK/st.-louis-17271" => array("type"=>"town", "pop"=>214, "name_ru"=>"Сент Луис", "name_en"=>"St. Louis", "lang"=>"en", "lat"=>35.079667, "lon"=>-96.852753),
"/US/US-OK/sterling-17306" => array("type"=>"town", "pop"=>716, "name_ru"=>"Стерлинг", "name_en"=>"Sterling", "lang"=>"en", "lat"=>34.749515, "lon"=>-98.172439),
"/US/US-OK/stidham-17307" => array("type"=>"town", "pop"=>23, "name_en"=>"Stidham", "lang"=>"en", "lat"=>35.368805, "lon"=>-95.700453),
"/US/US-OK/stigler-17308" => array("type"=>"city", "pop"=>2866, "name_en"=>"Stigler", "lang"=>"en", "lat"=>35.256790, "lon"=>-95.122422),
"/US/US-OK/stillwater-17309" => array("type"=>"city", "pop"=>47653, "name_ru"=>"Стиллуотер", "name_en"=>"Stillwater", "lang"=>"en", "lat"=>36.131018, "lon"=>-97.071690),
"/US/US-OK/stilwell-17310" => array("type"=>"city", "pop"=>3501, "name_en"=>"Stilwell", "lang"=>"en", "lat"=>35.814153, "lon"=>-94.631356),
"/US/US-OK/stonewall-17311" => array("type"=>"town", "pop"=>493, "name_ru"=>"Стоунуолл", "name_en"=>"Stonewall", "lang"=>"en", "lat"=>34.652028, "lon"=>-96.525018),
"/US/US-OK/strang-17312" => array("type"=>"town", "pop"=>103, "name_en"=>"Strang", "lang"=>"en", "lat"=>36.411161, "lon"=>-95.134475),
"/US/US-OK/stratford-17313" => array("type"=>"town", "pop"=>1490, "name_ru"=>"Стрэтфорд", "name_en"=>"Stratford", "lang"=>"en", "lat"=>34.795055, "lon"=>-96.960033),
"/US/US-OK/stringtown-17314" => array("type"=>"town", "pop"=>421, "name_en"=>"Stringtown", "lang"=>"en", "lat"=>34.465284, "lon"=>-96.053197),
"/US/US-OK/strong-city-17315" => array("type"=>"town", "pop"=>42, "name_ru"=>"Стронг Сити", "name_en"=>"Strong City", "lang"=>"en", "lat"=>35.670052, "lon"=>-99.600065),
"/US/US-OK/stroud-17316" => array("type"=>"city", "pop"=>2726, "name_en"=>"Stroud", "lang"=>"en", "lat"=>35.769145, "lon"=>-96.649505),
"/US/US-OK/stuart-17317" => array("type"=>"town", "pop"=>209, "name_ru"=>"Стюарт", "name_en"=>"Stuart", "lang"=>"en", "lat"=>34.900747, "lon"=>-96.099756),
"/US/US-OK/sugden-17318" => array("type"=>"town", "pop"=>55, "name_en"=>"Sugden", "lang"=>"en", "lat"=>34.081194, "lon"=>-97.978352),
"/US/US-OK/sulphur-17319" => array("type"=>"city", "pop"=>4838, "name_ru"=>"Салфер", "name_en"=>"Sulphur", "lang"=>"en", "lat"=>34.506726, "lon"=>-96.976976),
"/US/US-OK/summit-17320" => array("type"=>"town", "pop"=>232, "name_ru"=>"Саммит", "name_en"=>"Summit", "lang"=>"en", "lat"=>35.668901, "lon"=>-95.422353),
"/US/US-OK/taft-17321" => array("type"=>"town", "pop"=>609, "name_ru"=>"Тафт", "name_en"=>"Taft", "lang"=>"en", "lat"=>35.759813, "lon"=>-95.545808),
"/US/US-OK/tahlequah-17322" => array("type"=>"city", "pop"=>16623, "name_en"=>"Tahlequah", "lang"=>"en", "lat"=>35.911884, "lon"=>-94.972439),
"/US/US-OK/talala-17323" => array("type"=>"town", "pop"=>300, "name_en"=>"Talala", "lang"=>"en", "lat"=>36.530210, "lon"=>-95.699990),
"/US/US-OK/talihina-17324" => array("type"=>"town", "pop"=>1234, "name_en"=>"Talihina", "lang"=>"en", "lat"=>34.751508, "lon"=>-95.040708),
"/US/US-OK/taloga-17325" => array("type"=>"town", "pop"=>348, "name_en"=>"Taloga", "lang"=>"en", "lat"=>36.040231, "lon"=>-98.964271),
"/US/US-OK/tamaha-17326" => array("type"=>"town", "pop"=>205, "name_en"=>"Tamaha", "lang"=>"en", "lat"=>35.409190, "lon"=>-95.014154),
"/US/US-OK/tatums-17327" => array("type"=>"town", "pop"=>177, "name_en"=>"Tatums", "lang"=>"en", "lat"=>34.479404, "lon"=>-97.465260),
"/US/US-OK/tecumseh-17328" => array("type"=>"city", "pop"=>6667, "name_ru"=>"Текамсе", "name_en"=>"Tecumseh", "lang"=>"en", "lat"=>35.263260, "lon"=>-96.933198),
"/US/US-OK/temple-17329" => array("type"=>"town", "pop"=>1084, "name_ru"=>"Тэмпл", "name_en"=>"Temple", "lang"=>"en", "lat"=>34.273510, "lon"=>-98.234722),
"/US/US-OK/terlton-17330" => array("type"=>"town", "pop"=>83, "name_en"=>"Terlton", "lang"=>"en", "lat"=>36.187678, "lon"=>-96.490811),
"/US/US-OK/terral-17331" => array("type"=>"town", "pop"=>346, "name_en"=>"Terral", "lang"=>"en", "lat"=>33.896124, "lon"=>-97.938667),
"/US/US-OK/texhoma-17332" => array("type"=>"town", "pop"=>938, "name_en"=>"Texhoma", "lang"=>"en", "lat"=>36.505402, "lon"=>-101.787315),
"/US/US-OK/texola-17333" => array("type"=>"town", "pop"=>49, "name_en"=>"Texola", "lang"=>"en", "lat"=>35.218605, "lon"=>-99.989398),
"/US/US-OK/thackerville-17334" => array("type"=>"town", "pop"=>418, "name_en"=>"Thackerville", "lang"=>"en", "lat"=>33.793873, "lon"=>-97.143402),
"/US/US-OK/the-village-17335" => array("type"=>"city", "pop"=>9769, "name_en"=>"The Village", "lang"=>"en", "lat"=>35.570150, "lon"=>-97.555611),
"/US/US-OK/thomas-17336" => array("type"=>"town", "pop"=>1209, "name_en"=>"Thomas", "lang"=>"en", "lat"=>35.747310, "lon"=>-98.749104),
"/US/US-OK/tipton-17337" => array("type"=>"town", "pop"=>778, "name_ru"=>"Типтон", "name_en"=>"Tipton", "lang"=>"en", "lat"=>34.501264, "lon"=>-99.137503),
"/US/US-OK/tishomingo-17338" => array("type"=>"city", "pop"=>3150, "name_ru"=>"Тишоминго", "name_en"=>"Tishomingo", "lang"=>"en", "lat"=>34.239121, "lon"=>-96.678913),
"/US/US-OK/tonkawa-17339" => array("type"=>"city", "pop"=>3136, "name_en"=>"Tonkawa", "lang"=>"en", "lat"=>36.682835, "lon"=>-97.307810),
"/US/US-OK/tribbey-17340" => array("type"=>"town", "pop"=>288, "name_en"=>"Tribbey", "lang"=>"en", "lat"=>35.092666, "lon"=>-97.091486),
"/US/US-OK/tryon-17341" => array("type"=>"town", "pop"=>448, "name_en"=>"Tryon", "lang"=>"en", "lat"=>35.879460, "lon"=>-96.964144),
"/US/US-OK/tullahassee-17342" => array("type"=>"town", "pop"=>104, "name_en"=>"Tullahassee", "lang"=>"en", "lat"=>35.841562, "lon"=>-95.440345),
"/US/US-OK/tulsa-17343" => array("type"=>"city", "pop"=>385635, "name_en"=>"Tulsa", "lang"=>"en", "lat"=>36.127750, "lon"=>-95.916407),
"/US/US-OK/tupelo-17344" => array("type"=>"city", "pop"=>359, "name_ru"=>"Тупело", "name_en"=>"Tupelo", "lang"=>"en", "lat"=>34.602092, "lon"=>-96.420245),
"/US/US-OK/turley-17345" => array("type"=>"cdp", "pop"=>2930, "name_en"=>"Turley", "lang"=>"en", "lat"=>36.246700, "lon"=>-95.965773),
"/US/US-OK/tushka-17346" => array("type"=>"town", "pop"=>370, "name_en"=>"Tushka", "lang"=>"en", "lat"=>34.319461, "lon"=>-96.165997),
"/US/US-OK/tuttle-17347" => array("type"=>"town", "pop"=>6036, "name_en"=>"Tuttle", "lang"=>"en", "lat"=>35.296064, "lon"=>-97.751923),
"/US/US-OK/tyrone-17348" => array("type"=>"town", "pop"=>857, "name_ru"=>"Тайрон", "name_en"=>"Tyrone", "lang"=>"en", "lat"=>36.955605, "lon"=>-101.066444),
"/US/US-OK/union-city-17349" => array("type"=>"town", "pop"=>1453, "name_ru"=>"Юнион Сити", "name_en"=>"Union City", "lang"=>"en", "lat"=>35.401387, "lon"=>-97.898490),
"/US/US-OK/valley-brook-17350" => array("type"=>"town", "pop"=>783, "name_en"=>"Valley Brook", "lang"=>"en", "lat"=>35.402450, "lon"=>-97.481967),
"/US/US-OK/valley-park-17351" => array("type"=>"town", "pop"=>34, "name_en"=>"Valley Park", "lang"=>"en", "lat"=>36.283571, "lon"=>-95.733391),
"/US/US-OK/valliant-17352" => array("type"=>"town", "pop"=>743, "name_en"=>"Valliant", "lang"=>"en", "lat"=>34.004924, "lon"=>-95.089709),
"/US/US-OK/velma-17353" => array("type"=>"town", "pop"=>703, "name_en"=>"Velma", "lang"=>"en", "lat"=>34.456130, "lon"=>-97.665924),
"/US/US-OK/vera-17354" => array("type"=>"town", "pop"=>205, "name_en"=>"Vera", "lang"=>"en", "lat"=>36.448854, "lon"=>-95.881086),
"/US/US-OK/verden-17355" => array("type"=>"town", "pop"=>698, "name_en"=>"Verden", "lang"=>"en", "lat"=>35.083745, "lon"=>-98.087134),
"/US/US-OK/vian-17356" => array("type"=>"town", "pop"=>1438, "name_en"=>"Vian", "lang"=>"en", "lat"=>35.497555, "lon"=>-94.971180),
"/US/US-OK/vici-17357" => array("type"=>"town", "pop"=>624, "name_en"=>"Vici", "lang"=>"en", "lat"=>36.149146, "lon"=>-99.298882),
"/US/US-OK/vinita-17358" => array("type"=>"city", "pop"=>6059, "name_en"=>"Vinita", "lang"=>"en", "lat"=>36.641141, "lon"=>-95.153570),
"/US/US-OK/wagoner-17359" => array("type"=>"city", "pop"=>8081, "name_en"=>"Wagoner", "lang"=>"en", "lat"=>35.958491, "lon"=>-95.381839),
"/US/US-OK/wainwright-17360" => array("type"=>"town", "pop"=>201, "name_ru"=>"Уэйнрайт", "name_en"=>"Wainwright", "lang"=>"en", "lat"=>35.613326, "lon"=>-95.565858),
"/US/US-OK/wakita-17361" => array("type"=>"town", "pop"=>373, "name_en"=>"Wakita", "lang"=>"en", "lat"=>36.881755, "lon"=>-97.923421),
"/US/US-OK/walters-17362" => array("type"=>"city", "pop"=>2432, "name_ru"=>"Уолтерс", "name_en"=>"Walters", "lang"=>"en", "lat"=>34.357286, "lon"=>-98.307914),
"/US/US-OK/wanette-17363" => array("type"=>"town", "pop"=>417, "name_en"=>"Wanette", "lang"=>"en", "lat"=>34.962629, "lon"=>-97.030870),
"/US/US-OK/wann-17364" => array("type"=>"town", "pop"=>136, "name_en"=>"Wann", "lang"=>"en", "lat"=>36.914934, "lon"=>-95.804755),
"/US/US-OK/wapanucka-17365" => array("type"=>"town", "pop"=>431, "name_en"=>"Wapanucka", "lang"=>"en", "lat"=>34.373406, "lon"=>-96.424395),
"/US/US-OK/wardville-17366" => array("type"=>"town", "pop"=>52, "name_en"=>"Wardville", "lang"=>"en", "lat"=>34.657024, "lon"=>-96.032491),
"/US/US-OK/warner-17367" => array("type"=>"town", "pop"=>1447, "name_en"=>"Warner", "lang"=>"en", "lat"=>35.493484, "lon"=>-95.308894),
"/US/US-OK/warr-acres-17368" => array("type"=>"city", "pop"=>9427, "name_en"=>"Warr Acres", "lang"=>"en", "lat"=>35.526767, "lon"=>-97.617659),
"/US/US-OK/warwick-17369" => array("type"=>"town", "pop"=>237, "name_ru"=>"Уорвик", "name_en"=>"Warwick", "lang"=>"en", "lat"=>35.687484, "lon"=>-96.999971),
"/US/US-OK/washington-17370" => array("type"=>"town", "pop"=>552, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>35.058329, "lon"=>-97.483510),
"/US/US-OK/watonga-17371" => array("type"=>"city", "pop"=>5657, "name_en"=>"Watonga", "lang"=>"en", "lat"=>35.849212, "lon"=>-98.412177),
"/US/US-OK/watts-17372" => array("type"=>"town", "pop"=>325, "name_en"=>"Watts", "lang"=>"en", "lat"=>36.107188, "lon"=>-94.571281),
"/US/US-OK/waukomis-17373" => array("type"=>"town", "pop"=>1214, "name_en"=>"Waukomis", "lang"=>"en", "lat"=>36.281572, "lon"=>-97.904607),
"/US/US-OK/waurika-17374" => array("type"=>"city", "pop"=>1823, "name_en"=>"Waurika", "lang"=>"en", "lat"=>34.186536, "lon"=>-98.022267),
"/US/US-OK/wayne-17375" => array("type"=>"town", "pop"=>744, "name_ru"=>"Уэйн", "name_en"=>"Wayne", "lang"=>"en", "lat"=>34.917190, "lon"=>-97.315771),
"/US/US-OK/waynoka-17376" => array("type"=>"city", "pop"=>901, "name_en"=>"Waynoka", "lang"=>"en", "lat"=>36.585012, "lon"=>-98.879413),
"/US/US-OK/weatherford-17377" => array("type"=>"city", "pop"=>10185, "name_en"=>"Weatherford", "lang"=>"en", "lat"=>35.537140, "lon"=>-98.689944),
"/US/US-OK/webb-city-17378" => array("type"=>"town", "pop"=>91, "name_en"=>"Webb City", "lang"=>"en", "lat"=>36.807188, "lon"=>-96.712757),
"/US/US-OK/webbers-falls-17379" => array("type"=>"town", "pop"=>716, "name_en"=>"Webbers Falls", "lang"=>"en", "lat"=>35.508384, "lon"=>-95.160916),
"/US/US-OK/welch-17380" => array("type"=>"town", "pop"=>604, "name_en"=>"Welch", "lang"=>"en", "lat"=>36.874534, "lon"=>-95.095242),
"/US/US-OK/weleetka-17381" => array("type"=>"town", "pop"=>926, "name_en"=>"Weleetka", "lang"=>"en", "lat"=>35.338089, "lon"=>-96.135945),
"/US/US-OK/wellston-17382" => array("type"=>"town", "pop"=>843, "name_en"=>"Wellston", "lang"=>"en", "lat"=>35.687938, "lon"=>-97.061762),
"/US/US-OK/west-siloam-springs-17384" => array("type"=>"town", "pop"=>925, "name_en"=>"West Siloam Springs", "lang"=>"en", "lat"=>36.174587, "lon"=>-94.584178),
"/US/US-OK/westport-17383" => array("type"=>"town", "pop"=>260, "name_ru"=>"Уэстпорт", "name_en"=>"Westport", "lang"=>"en", "lat"=>36.194763, "lon"=>-96.336771),
"/US/US-OK/westville-17385" => array("type"=>"town", "pop"=>1638, "name_ru"=>"Уэствилл", "name_en"=>"Westville", "lang"=>"en", "lat"=>35.989168, "lon"=>-94.574183),
"/US/US-OK/wetumka-17386" => array("type"=>"city", "pop"=>1405, "name_en"=>"Wetumka", "lang"=>"en", "lat"=>35.241569, "lon"=>-96.238343),
"/US/US-OK/wewoka-17387" => array("type"=>"city", "pop"=>3315, "name_en"=>"Wewoka", "lang"=>"en", "lat"=>35.144174, "lon"=>-96.495767),
"/US/US-OK/whitefield-17388" => array("type"=>"town", "pop"=>221, "name_en"=>"Whitefield", "lang"=>"en", "lat"=>35.252092, "lon"=>-95.237627),
"/US/US-OK/wilburton-17389" => array("type"=>"city", "pop"=>2893, "name_en"=>"Wilburton", "lang"=>"en", "lat"=>34.917835, "lon"=>-95.303717),
"/US/US-OK/willow-17390" => array("type"=>"town", "pop"=>107, "name_ru"=>"Уиллоу", "name_en"=>"Willow", "lang"=>"en", "lat"=>35.052473, "lon"=>-99.509164),
"/US/US-OK/wilson-17391" => array("type"=>"city", "pop"=>1634, "name_ru"=>"Уилсон", "name_en"=>"Wilson", "lang"=>"en", "lat"=>34.158950, "lon"=>-97.432493),
"/US/US-OK/winchester-17392" => array("type"=>"town", "pop"=>449, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>35.785991, "lon"=>-95.991995),
"/US/US-OK/wister-17393" => array("type"=>"town", "pop"=>1029, "name_en"=>"Wister", "lang"=>"en", "lat"=>34.970153, "lon"=>-94.723181),
"/US/US-OK/woodlawn-park-17394" => array("type"=>"town", "pop"=>155, "name_ru"=>"Вудлон Парк", "name_en"=>"Woodlawn Park", "lang"=>"en", "lat"=>35.509100, "lon"=>-97.647399),
"/US/US-OK/woodville-17395" => array("type"=>"town", "pop"=>31, "name_ru"=>"Вудвилл", "name_en"=>"Woodville", "lang"=>"en", "lat"=>33.968626, "lon"=>-96.654881),
"/US/US-OK/woodward-17396" => array("type"=>"city", "pop"=>12299, "name_ru"=>"Вудуорд", "name_en"=>"Woodward", "lang"=>"en", "lat"=>36.424907, "lon"=>-99.404457),
"/US/US-OK/wright-city-17397" => array("type"=>"town", "pop"=>791, "name_en"=>"Wright City", "lang"=>"en", "lat"=>34.063685, "lon"=>-95.007381),
"/US/US-OK/wyandotte-17398" => array("type"=>"town", "pop"=>350, "name_ru"=>"Уайандотт", "name_en"=>"Wyandotte", "lang"=>"en", "lat"=>36.792858, "lon"=>-94.723227),
"/US/US-OK/wynnewood-17399" => array("type"=>"city", "pop"=>2289, "name_en"=>"Wynnewood", "lang"=>"en", "lat"=>34.644316, "lon"=>-97.161697),
"/US/US-OK/wynona-17400" => array("type"=>"town", "pop"=>526, "name_en"=>"Wynona", "lang"=>"en", "lat"=>36.545514, "lon"=>-96.325673),
"/US/US-OK/yale-17401" => array("type"=>"city", "pop"=>1356, "name_ru"=>"Йель", "name_en"=>"Yale", "lang"=>"en", "lat"=>36.114884, "lon"=>-96.700620),
"/US/US-OK/yeager-17402" => array("type"=>"town", "pop"=>65, "name_en"=>"Yeager", "lang"=>"en", "lat"=>35.156788, "lon"=>-96.339559),
"/US/US-OK/yukon-17403" => array("type"=>"city", "pop"=>22849, "name_en"=>"Yukon", "lang"=>"en", "lat"=>35.516800, "lon"=>-97.768471),
"/US/US-OR" => array("type"=>"state", "pop"=>3790060, "name_ru"=>"Орегон", "name_en"=>"Oregon", "lang"=>"en"),
"/US/US-OR/adair-village-17404" => array("type"=>"city", "pop"=>783, "name_en"=>"Adair Village", "lang"=>"en", "lat"=>44.674320, "lon"=>-123.217669),
"/US/US-OR/adams-17405" => array("type"=>"city", "pop"=>289, "name_ru"=>"Адамс", "name_en"=>"Adams", "lang"=>"en", "lat"=>45.767259, "lon"=>-118.562654),
"/US/US-OR/adrian-17406" => array("type"=>"city", "pop"=>140, "name_ru"=>"Эдриан", "name_en"=>"Adrian", "lang"=>"en", "lat"=>43.740859, "lon"=>-117.070063),
"/US/US-OR/albany-17407" => array("type"=>"city", "pop"=>48081, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>44.617017, "lon"=>-123.091263),
"/US/US-OR/aloha-17408" => array("type"=>"cdp", "pop"=>34284, "name_en"=>"Aloha", "lang"=>"en", "lat"=>45.492056, "lon"=>-122.871765),
"/US/US-OR/altamont-17409" => array("type"=>"cdp", "pop"=>18591, "name_ru"=>"Альтамонт", "name_en"=>"Altamont", "lang"=>"en", "lat"=>42.198176, "lon"=>-121.719721),
"/US/US-OR/amity-17410" => array("type"=>"city", "pop"=>1467, "name_ru"=>"Амити", "name_en"=>"Amity", "lang"=>"en", "lat"=>45.115699, "lon"=>-123.203128),
"/US/US-OR/antelope-17411" => array("type"=>"city", "pop"=>58, "name_en"=>"Antelope", "lang"=>"en", "lat"=>44.910885, "lon"=>-120.722531),
"/US/US-OR/arlington-17412" => array("type"=>"city", "pop"=>475, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>45.717847, "lon"=>-120.192607),
"/US/US-OR/ashland-17413" => array("type"=>"city", "pop"=>21390, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>42.190205, "lon"=>-122.699569),
"/US/US-OR/astoria-17414" => array("type"=>"city", "pop"=>9851, "name_ru"=>"Астория", "name_en"=>"Astoria", "lang"=>"en", "lat"=>46.188026, "lon"=>-123.820481),
"/US/US-OR/athena-17415" => array("type"=>"city", "pop"=>1184, "name_en"=>"Athena", "lang"=>"en", "lat"=>45.812778, "lon"=>-118.491449),
"/US/US-OR/aumsville-17416" => array("type"=>"city", "pop"=>3603, "name_en"=>"Aumsville", "lang"=>"en", "lat"=>44.845659, "lon"=>-122.868096),
"/US/US-OR/aurora-17417" => array("type"=>"city", "pop"=>1003, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>45.228297, "lon"=>-122.755574),
"/US/US-OR/baker-17418" => array("type"=>"city", "pop"=>15983, "name_ru"=>"Бейкер", "name_en"=>"Baker", "lang"=>"en", "lat"=>44.774194, "lon"=>-117.831637),
"/US/US-OR/bandon-17419" => array("type"=>"city", "pop"=>3248, "name_en"=>"Bandon", "lang"=>"en", "lat"=>43.115869, "lon"=>-124.414636),
"/US/US-OR/banks-17420" => array("type"=>"city", "pop"=>1606, "name_ru"=>"Бэнкс", "name_en"=>"Banks", "lang"=>"en", "lat"=>45.615432, "lon"=>-123.110009),
"/US/US-OR/barlow-17421" => array("type"=>"city", "pop"=>144, "name_ru"=>"Барлоу", "name_en"=>"Barlow", "lang"=>"en", "lat"=>45.252400, "lon"=>-122.720111),
"/US/US-OR/barview-17422" => array("type"=>"cdp", "pop"=>1402, "name_en"=>"Barview", "lang"=>"en", "lat"=>43.347613, "lon"=>-124.307957),
"/US/US-OR/bay-city-17423" => array("type"=>"city", "pop"=>1145, "name_ru"=>"Бэй Сити", "name_en"=>"Bay City", "lang"=>"en", "lat"=>45.521241, "lon"=>-123.885005),
"/US/US-OR/beaverton-17424" => array("type"=>"city", "pop"=>91757, "name_ru"=>"Бивертон", "name_en"=>"Beaverton", "lang"=>"en", "lat"=>45.475500, "lon"=>-122.816599),
"/US/US-OR/bend-17425" => array("type"=>"city", "pop"=>77181, "name_en"=>"Bend", "lang"=>"en", "lat"=>44.066047, "lon"=>-121.311048),
"/US/US-OR/boardman-17426" => array("type"=>"city", "pop"=>2915, "name_en"=>"Boardman", "lang"=>"en", "lat"=>45.837593, "lon"=>-119.702477),
"/US/US-OR/bonanza-17427" => array("type"=>"town", "pop"=>404, "name_ru"=>"Бонанза", "name_en"=>"Bonanza", "lang"=>"en", "lat"=>42.199554, "lon"=>-121.405948),
"/US/US-OR/brookings-17428" => array("type"=>"city", "pop"=>6213, "name_en"=>"Brookings", "lang"=>"en", "lat"=>42.059763, "lon"=>-124.290797),
"/US/US-OR/brownsville-17429" => array("type"=>"city", "pop"=>1638, "name_ru"=>"Браунсвилл", "name_en"=>"Brownsville", "lang"=>"en", "lat"=>44.393031, "lon"=>-122.981930),
"/US/US-OR/bunker-hill-17430" => array("type"=>"cdp", "pop"=>1242, "name_ru"=>"Банкер Хилл", "name_en"=>"Bunker Hill", "lang"=>"en", "lat"=>43.350208, "lon"=>-124.209085),
"/US/US-OR/burns-17431" => array("type"=>"city", "pop"=>2664, "name_ru"=>"Бернс", "name_en"=>"Burns", "lang"=>"en", "lat"=>43.587735, "lon"=>-119.061178),
"/US/US-OR/butte-falls-17432" => array("type"=>"town", "pop"=>431, "name_en"=>"Butte Falls", "lang"=>"en", "lat"=>42.542032, "lon"=>-122.567818),
"/US/US-OR/canby-17433" => array("type"=>"city", "pop"=>15637, "name_ru"=>"Кэнби", "name_en"=>"Canby", "lang"=>"en", "lat"=>45.266514, "lon"=>-122.691730),
"/US/US-OR/cannon-beach-17434" => array("type"=>"city", "pop"=>1726, "name_en"=>"Cannon Beach", "lang"=>"en", "lat"=>45.888792, "lon"=>-123.959786),
"/US/US-OR/canyon-city-17435" => array("type"=>"town", "pop"=>556, "name_en"=>"Canyon City", "lang"=>"en", "lat"=>44.392201, "lon"=>-118.948401),
"/US/US-OR/canyonville-17436" => array("type"=>"city", "pop"=>1607, "name_en"=>"Canyonville", "lang"=>"en", "lat"=>42.928746, "lon"=>-123.278339),
"/US/US-OR/carlton-17437" => array("type"=>"city", "pop"=>1679, "name_ru"=>"Карлтон", "name_en"=>"Carlton", "lang"=>"en", "lat"=>45.294980, "lon"=>-123.174977),
"/US/US-OR/cascade-locks-17438" => array("type"=>"city", "pop"=>1087, "name_en"=>"Cascade Locks", "lang"=>"en", "lat"=>45.677848, "lon"=>-121.872559),
"/US/US-OR/cave-junction-17439" => array("type"=>"city", "pop"=>1425, "name_en"=>"Cave Junction", "lang"=>"en", "lat"=>42.167345, "lon"=>-123.644674),
"/US/US-OR/cedar-hills-17440" => array("type"=>"cdp", "pop"=>9294, "name_en"=>"Cedar Hills", "lang"=>"en", "lat"=>45.503650, "lon"=>-122.805562),
"/US/US-OR/cedar-mill-17441" => array("type"=>"cdp", "pop"=>9697, "name_en"=>"Cedar Mill", "lang"=>"en", "lat"=>45.536568, "lon"=>-122.799523),
"/US/US-OR/central-point-17442" => array("type"=>"city", "pop"=>16503, "name_en"=>"Central Point", "lang"=>"en", "lat"=>42.373700, "lon"=>-122.910969),
"/US/US-OR/chenoweth-17443" => array("type"=>"cdp", "pop"=>3246, "name_en"=>"Chenoweth", "lang"=>"en", "lat"=>45.627339, "lon"=>-121.230134),
"/US/US-OR/chiloquin-17444" => array("type"=>"city", "pop"=>697, "name_en"=>"Chiloquin", "lang"=>"en", "lat"=>42.576491, "lon"=>-121.866890),
"/US/US-OR/city-of-the-dalles-17445" => array("type"=>"city", "pop"=>11060, "name_en"=>"City of the Dalles", "lang"=>"en", "lat"=>45.599156, "lon"=>-121.174997),
"/US/US-OR/clackamas-17446" => array("type"=>"cdp", "pop"=>380576, "name_en"=>"Clackamas", "lang"=>"en", "lat"=>45.408553, "lon"=>-122.552520),
"/US/US-OR/clatskanie-17447" => array("type"=>"city", "pop"=>1637, "name_en"=>"Clatskanie", "lang"=>"en", "lat"=>46.103745, "lon"=>-123.205016),
"/US/US-OR/coburg-17448" => array("type"=>"city", "pop"=>1074, "name_ru"=>"Коберг", "name_en"=>"Coburg", "lang"=>"en", "lat"=>44.138093, "lon"=>-123.059816),
"/US/US-OR/columbia-city-17449" => array("type"=>"city", "pop"=>1964, "name_ru"=>"Колумбия Сити", "name_en"=>"Columbia City", "lang"=>"en", "lat"=>45.897355, "lon"=>-122.810825),
"/US/US-OR/condon-17450" => array("type"=>"city", "pop"=>668, "name_en"=>"Condon", "lang"=>"en", "lat"=>45.236760, "lon"=>-120.183947),
"/US/US-OR/coos-bay-17451" => array("type"=>"city", "pop"=>15665, "name_en"=>"Coos Bay", "lang"=>"en", "lat"=>43.378446, "lon"=>-124.231014),
"/US/US-OR/coquille-17452" => array("type"=>"city", "pop"=>4117, "name_en"=>"Coquille", "lang"=>"en", "lat"=>43.181296, "lon"=>-124.181717),
"/US/US-OR/cornelius-17453" => array("type"=>"city", "pop"=>11464, "name_en"=>"Cornelius", "lang"=>"en", "lat"=>45.518880, "lon"=>-123.054091),
"/US/US-OR/corvallis-17454" => array("type"=>"city", "pop"=>51110, "name_en"=>"Corvallis", "lang"=>"en", "lat"=>44.571006, "lon"=>-123.276033),
"/US/US-OR/cottage-grove-17455" => array("type"=>"city", "pop"=>9124, "name_ru"=>"Коттэдж Гров", "name_en"=>"Cottage Grove", "lang"=>"en", "lat"=>43.797661, "lon"=>-123.055651),
"/US/US-OR/cove-17456" => array("type"=>"city", "pop"=>610, "name_ru"=>"Ков", "name_en"=>"Cove", "lang"=>"en", "lat"=>45.296700, "lon"=>-117.809490),
"/US/US-OR/creswell-17457" => array("type"=>"city", "pop"=>5058, "name_en"=>"Creswell", "lang"=>"en", "lat"=>43.918312, "lon"=>-123.019259),
"/US/US-OR/culver-17458" => array("type"=>"city", "pop"=>1116, "name_ru"=>"Калвер", "name_en"=>"Culver", "lang"=>"en", "lat"=>44.524943, "lon"=>-121.209278),
"/US/US-OR/dallas-17459" => array("type"=>"city", "pop"=>15892, "name_ru"=>"Даллас", "name_en"=>"Dallas", "lang"=>"en", "lat"=>44.921340, "lon"=>-123.312298),
"/US/US-OR/dayton-17460" => array("type"=>"city", "pop"=>2368, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>45.220496, "lon"=>-123.076641),
"/US/US-OR/dayville-17461" => array("type"=>"town", "pop"=>113, "name_en"=>"Dayville", "lang"=>"en", "lat"=>44.466795, "lon"=>-119.532268),
"/US/US-OR/depoe-bay-17462" => array("type"=>"city", "pop"=>1384, "name_en"=>"Depoe Bay", "lang"=>"en", "lat"=>44.810147, "lon"=>-124.058252),
"/US/US-OR/deschutes-river-woods-17463" => array("type"=>"cdp", "pop"=>2373, "name_en"=>"Deschutes River Woods", "lang"=>"en", "lat"=>43.991741, "lon"=>-121.357135),
"/US/US-OR/detroit-17464" => array("type"=>"city", "pop"=>271, "name_ru"=>"Детройт", "name_en"=>"Detroit", "lang"=>"en", "lat"=>44.733858, "lon"=>-122.150841),
"/US/US-OR/donald-17465" => array("type"=>"city", "pop"=>967, "name_en"=>"Donald", "lang"=>"en", "lat"=>45.223025, "lon"=>-122.837652),
"/US/US-OR/drain-17466" => array("type"=>"city", "pop"=>1038, "name_en"=>"Drain", "lang"=>"en", "lat"=>43.662107, "lon"=>-123.314066),
"/US/US-OR/dufur-17467" => array("type"=>"city", "pop"=>581, "name_en"=>"Dufur", "lang"=>"en", "lat"=>45.453374, "lon"=>-121.127072),
"/US/US-OR/dundee-17468" => array("type"=>"city", "pop"=>3076, "name_ru"=>"Данди", "name_en"=>"Dundee", "lang"=>"en", "lat"=>45.276485, "lon"=>-123.005903),
"/US/US-OR/dunes-city-17469" => array("type"=>"city", "pop"=>1467, "name_en"=>"Dunes City", "lang"=>"en", "lat"=>43.908104, "lon"=>-124.095060),
"/US/US-OR/durham-17470" => array("type"=>"city", "pop"=>1444, "name_ru"=>"Дарем", "name_en"=>"Durham", "lang"=>"en", "lat"=>45.394667, "lon"=>-122.757482),
"/US/US-OR/eagle-point-17471" => array("type"=>"city", "pop"=>8302, "name_en"=>"Eagle Point", "lang"=>"en", "lat"=>42.466066, "lon"=>-122.799527),
"/US/US-OR/echo-17472" => array("type"=>"city", "pop"=>666, "name_ru"=>"Эхо", "name_en"=>"Echo", "lang"=>"en", "lat"=>45.743955, "lon"=>-119.192030),
"/US/US-OR/elgin-17473" => array("type"=>"city", "pop"=>1652, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>45.564482, "lon"=>-117.920215),
"/US/US-OR/elkton-17474" => array("type"=>"city", "pop"=>163, "name_ru"=>"Элктон", "name_en"=>"Elkton", "lang"=>"en", "lat"=>43.637340, "lon"=>-123.565979),
"/US/US-OR/enterprise-17475" => array("type"=>"city", "pop"=>1710, "name_ru"=>"Энтерпрайз", "name_en"=>"Enterprise", "lang"=>"en", "lat"=>45.425268, "lon"=>-117.276071),
"/US/US-OR/estacada-17476" => array("type"=>"city", "pop"=>2558, "name_en"=>"Estacada", "lang"=>"en", "lat"=>45.292190, "lon"=>-122.333286),
"/US/US-OR/eugene-17477" => array("type"=>"city", "pop"=>150104, "name_ru"=>"Юджин", "name_en"=>"Eugene", "lang"=>"en", "lat"=>44.053000, "lon"=>-123.112172),
"/US/US-OR/fairview-17478" => array("type"=>"city", "pop"=>9805, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>45.546367, "lon"=>-122.434725),
"/US/US-OR/falls-city-17479" => array("type"=>"city", "pop"=>1070, "name_en"=>"Falls City", "lang"=>"en", "lat"=>44.865618, "lon"=>-123.436743),
"/US/US-OR/florence-17480" => array("type"=>"city", "pop"=>8694, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>43.986388, "lon"=>-124.103108),
"/US/US-OR/forest-grove-17481" => array("type"=>"city", "pop"=>20985, "name_en"=>"Forest Grove", "lang"=>"en", "lat"=>45.522264, "lon"=>-123.104431),
"/US/US-OR/fossil-17482" => array("type"=>"city", "pop"=>379, "name_en"=>"Fossil", "lang"=>"en", "lat"=>44.998411, "lon"=>-120.213191),
"/US/US-OR/four-corners-17483" => array("type"=>"cdp", "pop"=>12156, "name_en"=>"Four Corners", "lang"=>"en", "lat"=>44.929000, "lon"=>-122.969874),
"/US/US-OR/garden-home-whitford-17484" => array("type"=>"cdp", "pop"=>6652, "name_en"=>"Garden Home-Whitford", "lang"=>"en", "lat"=>45.465125, "lon"=>-122.756835),
"/US/US-OR/garibaldi-17485" => array("type"=>"city", "pop"=>881, "name_en"=>"Garibaldi", "lang"=>"en", "lat"=>45.560455, "lon"=>-123.910223),
"/US/US-OR/gaston-17486" => array("type"=>"city", "pop"=>783, "name_ru"=>"Гэстон", "name_en"=>"Gaston", "lang"=>"en", "lat"=>45.435622, "lon"=>-123.140774),
"/US/US-OR/gates-17487" => array("type"=>"city", "pop"=>485, "name_en"=>"Gates", "lang"=>"en", "lat"=>44.755720, "lon"=>-122.419482),
"/US/US-OR/gearhart-17488" => array("type"=>"city", "pop"=>1189, "name_en"=>"Gearhart", "lang"=>"en", "lat"=>46.024986, "lon"=>-123.917472),
"/US/US-OR/gervais-17489" => array("type"=>"city", "pop"=>2406, "name_en"=>"Gervais", "lang"=>"en", "lat"=>45.108456, "lon"=>-122.894791),
"/US/US-OR/gladstone-17490" => array("type"=>"city", "pop"=>12079, "name_ru"=>"Глэдстоун", "name_en"=>"Gladstone", "lang"=>"en", "lat"=>45.385800, "lon"=>-122.591523),
"/US/US-OR/glendale-17491" => array("type"=>"city", "pop"=>872, "name_ru"=>"Глендейл", "name_en"=>"Glendale", "lang"=>"en", "lat"=>42.737645, "lon"=>-123.428387),
"/US/US-OR/gold-beach-17492" => array("type"=>"city", "pop"=>1822, "name_en"=>"Gold Beach", "lang"=>"en", "lat"=>42.412905, "lon"=>-124.418517),
"/US/US-OR/gold-hill-17493" => array("type"=>"city", "pop"=>1070, "name_en"=>"Gold Hill", "lang"=>"en", "lat"=>42.434592, "lon"=>-123.050327),
"/US/US-OR/granite-17494" => array("type"=>"city", "pop"=>20, "name_en"=>"Granite", "lang"=>"en", "lat"=>44.810560, "lon"=>-118.419414),
"/US/US-OR/grants-pass-17495" => array("type"=>"city", "pop"=>33217, "name_en"=>"Grants Pass", "lang"=>"en", "lat"=>42.441234, "lon"=>-123.324286),
"/US/US-OR/grass-valley-17496" => array("type"=>"city", "pop"=>136, "name_ru"=>"Грас Валли", "name_en"=>"Grass Valley", "lang"=>"en", "lat"=>45.359892, "lon"=>-120.783179),
"/US/US-OR/green-17497" => array("type"=>"cdp", "pop"=>5076, "name_ru"=>"Грин", "name_en"=>"Green", "lang"=>"en", "lat"=>43.149212, "lon"=>-123.382222),
"/US/US-OR/greenhorn-17498" => array("type"=>"city", "pop"=>0, "name_en"=>"Greenhorn", "lang"=>"en", "lat"=>44.708863, "lon"=>-118.495815),
"/US/US-OR/gresham-17499" => array("type"=>"city", "pop"=>101221, "name_en"=>"Gresham", "lang"=>"en", "lat"=>45.503850, "lon"=>-122.438224),
"/US/US-OR/haines-17500" => array("type"=>"city", "pop"=>391, "name_ru"=>"Хайнс", "name_en"=>"Haines", "lang"=>"en", "lat"=>44.911856, "lon"=>-117.938958),
"/US/US-OR/halfway-17501" => array("type"=>"town", "pop"=>308, "name_ru"=>"Хафуэй", "name_en"=>"Halfway", "lang"=>"en", "lat"=>44.876766, "lon"=>-117.108973),
"/US/US-OR/halsey-17502" => array("type"=>"city", "pop"=>872, "name_en"=>"Halsey", "lang"=>"en", "lat"=>44.383117, "lon"=>-123.108086),
"/US/US-OR/hammond-17503" => array("type"=>"town", "pop"=>589, "name_ru"=>"Хаммонд", "name_en"=>"Hammond", "lang"=>"en", "lat"=>46.198146, "lon"=>-123.946757),
"/US/US-OR/happy-valley-17504" => array("type"=>"city", "pop"=>12643, "name_ru"=>"Хэппи Валли", "name_en"=>"Happy Valley", "lang"=>"en", "lat"=>45.443816, "lon"=>-122.535751),
"/US/US-OR/harbeck-fruitdale-17505" => array("type"=>"cdp", "pop"=>3982, "name_en"=>"Harbeck-Fruitdale", "lang"=>"en", "lat"=>42.415839, "lon"=>-123.320460),
"/US/US-OR/harbor-17506" => array("type"=>"cdp", "pop"=>2143, "name_en"=>"Harbor", "lang"=>"en", "lat"=>42.039141, "lon"=>-124.254667),
"/US/US-OR/harrisburg-17507" => array("type"=>"city", "pop"=>3494, "name_ru"=>"Харрисбург", "name_en"=>"Harrisburg", "lang"=>"en", "lat"=>44.268490, "lon"=>-123.164986),
"/US/US-OR/hayesville-17508" => array("type"=>"cdp", "pop"=>14318, "name_ru"=>"Хайесвилл", "name_en"=>"Hayesville", "lang"=>"en", "lat"=>44.984500, "lon"=>-122.970473),
"/US/US-OR/hazelwood-17509" => array("type"=>"cdp", "pop"=>11480, "name_ru"=>"Хейзелвуд", "name_en"=>"Hazelwood", "lang"=>"en", "lat"=>45.516600, "lon"=>-122.522718),
"/US/US-OR/helix-17510" => array("type"=>"city", "pop"=>178, "name_en"=>"Helix", "lang"=>"en", "lat"=>45.850250, "lon"=>-118.657387),
"/US/US-OR/heppner-17511" => array("type"=>"city", "pop"=>1361, "name_en"=>"Heppner", "lang"=>"en", "lat"=>45.354166, "lon"=>-119.552977),
"/US/US-OR/hermiston-17512" => array("type"=>"city", "pop"=>15297, "name_en"=>"Hermiston", "lang"=>"en", "lat"=>45.832813, "lon"=>-119.283515),
"/US/US-OR/hillsboro-17513" => array("type"=>"city", "pop"=>93638, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>45.527679, "lon"=>-122.939055),
"/US/US-OR/hines-17514" => array("type"=>"city", "pop"=>1406, "name_en"=>"Hines", "lang"=>"en", "lat"=>43.567237, "lon"=>-119.078776),
"/US/US-OR/hood-river-17515" => array("type"=>"city", "pop"=>6877, "name_en"=>"Hood River", "lang"=>"en", "lat"=>45.709842, "lon"=>-121.521935),
"/US/US-OR/hubbard-17516" => array("type"=>"city", "pop"=>2812, "name_ru"=>"Хаббард", "name_en"=>"Hubbard", "lang"=>"en", "lat"=>45.181298, "lon"=>-122.805846),
"/US/US-OR/huntington-17517" => array("type"=>"city", "pop"=>470, "name_ru"=>"Хантингтон", "name_en"=>"Huntington", "lang"=>"en", "lat"=>44.350308, "lon"=>-117.266012),
"/US/US-OR/idanha-17518" => array("type"=>"city", "pop"=>227, "name_en"=>"Idanha", "lang"=>"en", "lat"=>44.703083, "lon"=>-122.083429),
"/US/US-OR/imbler-17519" => array("type"=>"city", "pop"=>283, "name_en"=>"Imbler", "lang"=>"en", "lat"=>45.462445, "lon"=>-117.962490),
"/US/US-OR/independence-17520" => array("type"=>"city", "pop"=>9375, "name_ru"=>"Индепенденс", "name_en"=>"Independence", "lang"=>"en", "lat"=>44.855342, "lon"=>-123.192489),
"/US/US-OR/ione-17521" => array("type"=>"city", "pop"=>314, "name_ru"=>"Ен", "name_en"=>"Ione", "lang"=>"en", "lat"=>45.501018, "lon"=>-119.822252),
"/US/US-OR/irrigon-17522" => array("type"=>"city", "pop"=>1755, "name_en"=>"Irrigon", "lang"=>"en", "lat"=>45.896859, "lon"=>-119.488053),
"/US/US-OR/island-city-17523" => array("type"=>"city", "pop"=>916, "name_en"=>"Island City", "lang"=>"en", "lat"=>45.338842, "lon"=>-118.046060),
"/US/US-OR/jacksonville-17524" => array("type"=>"city", "pop"=>2181, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>42.313092, "lon"=>-122.967559),
"/US/US-OR/jefferson-17525" => array("type"=>"city", "pop"=>3085, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>44.717986, "lon"=>-123.005271),
"/US/US-OR/jennings-lodge-17526" => array("type"=>"cdp", "pop"=>6530, "name_en"=>"Jennings Lodge", "lang"=>"en", "lat"=>45.392900, "lon"=>-122.614019),
"/US/US-OR/john-day-17527" => array("type"=>"city", "pop"=>1512, "name_en"=>"John Day", "lang"=>"en", "lat"=>44.418427, "lon"=>-118.949399),
"/US/US-OR/johnson-city-17528" => array("type"=>"city", "pop"=>661, "name_ru"=>"Джонсон Сити", "name_en"=>"Johnson City", "lang"=>"en", "lat"=>45.404450, "lon"=>-122.577907),
"/US/US-OR/jordan-valley-17529" => array("type"=>"city", "pop"=>228, "name_en"=>"Jordan Valley", "lang"=>"en", "lat"=>42.976562, "lon"=>-117.054307),
"/US/US-OR/joseph-17530" => array("type"=>"city", "pop"=>955, "name_en"=>"Joseph", "lang"=>"en", "lat"=>45.352019, "lon"=>-117.227876),
"/US/US-OR/junction-city-17531" => array("type"=>"city", "pop"=>5618, "name_ru"=>"Джанкшен Сити", "name_en"=>"Junction City", "lang"=>"en", "lat"=>44.217944, "lon"=>-123.203864),
"/US/US-OR/keizer-17532" => array("type"=>"city", "pop"=>35864, "name_en"=>"Keizer", "lang"=>"en", "lat"=>45.003983, "lon"=>-123.021202),
"/US/US-OR/king-city-17533" => array("type"=>"city", "pop"=>2931, "name_ru"=>"Кинг Сити", "name_en"=>"King City", "lang"=>"en", "lat"=>45.405431, "lon"=>-122.801580),
"/US/US-OR/klamath-falls-17534" => array("type"=>"city", "pop"=>20276, "name_en"=>"Klamath Falls", "lang"=>"en", "lat"=>42.221019, "lon"=>-121.772380),
"/US/US-OR/la-grande-17536" => array("type"=>"city", "pop"=>12682, "name_en"=>"La Grande", "lang"=>"en", "lat"=>45.325345, "lon"=>-118.087300),
"/US/US-OR/lafayette-17535" => array("type"=>"city", "pop"=>3605, "name_ru"=>"Ла Файет", "name_en"=>"Lafayette", "lang"=>"en", "lat"=>45.246131, "lon"=>-123.110417),
"/US/US-OR/lake-oswego-17537" => array("type"=>"city", "pop"=>36926, "name_en"=>"Lake Oswego", "lang"=>"en", "lat"=>45.412200, "lon"=>-122.698118),
"/US/US-OR/lakeside-17538" => array("type"=>"city", "pop"=>1463, "name_ru"=>"Лэйксайд", "name_en"=>"Lakeside", "lang"=>"en", "lat"=>43.579044, "lon"=>-124.173236),
"/US/US-OR/lakeview-17539" => array("type"=>"town", "pop"=>2427, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>42.189171, "lon"=>-120.345439),
"/US/US-OR/lebanon-17540" => array("type"=>"city", "pop"=>15397, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>44.534854, "lon"=>-122.904354),
"/US/US-OR/lexington-17541" => array("type"=>"town", "pop"=>258, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>45.445657, "lon"=>-119.686600),
"/US/US-OR/lincoln-beach-17542" => array("type"=>"cdp", "pop"=>1507, "name_en"=>"Lincoln Beach", "lang"=>"en", "lat"=>44.873899, "lon"=>-124.028465),
"/US/US-OR/lincoln-city-17543" => array("type"=>"city", "pop"=>8066, "name_en"=>"Lincoln City", "lang"=>"en", "lat"=>44.972555, "lon"=>-124.007155),
"/US/US-OR/lonerock-17544" => array("type"=>"city", "pop"=>21, "name_en"=>"Lonerock", "lang"=>"en", "lat"=>45.089229, "lon"=>-119.883109),
"/US/US-OR/long-creek-17545" => array("type"=>"city", "pop"=>187, "name_ru"=>"Лонг Крик", "name_en"=>"Long Creek", "lang"=>"en", "lat"=>44.713757, "lon"=>-119.101935),
"/US/US-OR/lostine-17546" => array("type"=>"city", "pop"=>236, "name_en"=>"Lostine", "lang"=>"en", "lat"=>45.487050, "lon"=>-117.429049),
"/US/US-OR/lowell-17547" => array("type"=>"city", "pop"=>1022, "name_ru"=>"Лоуэлл", "name_en"=>"Lowell", "lang"=>"en", "lat"=>43.920678, "lon"=>-122.779613),
"/US/US-OR/lyons-17548" => array("type"=>"city", "pop"=>1150, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>44.777326, "lon"=>-122.606918),
"/US/US-OR/madras-17550" => array("type"=>"city", "pop"=>6011, "name_en"=>"Madras", "lang"=>"en", "lat"=>44.629871, "lon"=>-121.129015),
"/US/US-OR/malin-17551" => array("type"=>"city", "pop"=>618, "name_en"=>"Malin", "lang"=>"en", "lat"=>42.013966, "lon"=>-121.407766),
"/US/US-OR/manzanita-17552" => array("type"=>"city", "pop"=>618, "name_en"=>"Manzanita", "lang"=>"en", "lat"=>45.717293, "lon"=>-123.933576),
"/US/US-OR/maupin-17553" => array("type"=>"city", "pop"=>406, "name_en"=>"Maupin", "lang"=>"en", "lat"=>45.175165, "lon"=>-121.082861),
"/US/US-OR/maywood-park-17554" => array("type"=>"city", "pop"=>764, "name_en"=>"Maywood Park", "lang"=>"en", "lat"=>45.552733, "lon"=>-122.560846),
"/US/US-OR/mcminnville-17549" => array("type"=>"city", "pop"=>31185, "name_en"=>"McMinnville", "lang"=>"en", "lat"=>45.211210, "lon"=>-123.190248),
"/US/US-OR/medford-17555" => array("type"=>"city", "pop"=>73212, "name_ru"=>"Медфорд", "name_en"=>"Medford", "lang"=>"en", "lat"=>42.339800, "lon"=>-122.853099),
"/US/US-OR/merrill-17556" => array("type"=>"city", "pop"=>869, "name_ru"=>"Меррил", "name_en"=>"Merrill", "lang"=>"en", "lat"=>42.026601, "lon"=>-121.599711),
"/US/US-OR/metolius-17557" => array("type"=>"city", "pop"=>779, "name_en"=>"Metolius", "lang"=>"en", "lat"=>44.587595, "lon"=>-121.176556),
"/US/US-OR/metzger-17558" => array("type"=>"cdp", "pop"=>3149, "name_en"=>"Metzger", "lang"=>"en", "lat"=>45.449933, "lon"=>-122.761532),
"/US/US-OR/mill-city-17559" => array("type"=>"city", "pop"=>1685, "name_en"=>"Mill City", "lang"=>"en", "lat"=>44.751313, "lon"=>-122.477136),
"/US/US-OR/millersburg-17560" => array("type"=>"city", "pop"=>1112, "name_ru"=>"Миллерсберг", "name_en"=>"Millersburg", "lang"=>"en", "lat"=>44.680695, "lon"=>-123.070589),
"/US/US-OR/milton-freewater-17561" => array("type"=>"city", "pop"=>6421, "name_en"=>"Milton-Freewater", "lang"=>"en", "lat"=>45.934926, "lon"=>-118.390371),
"/US/US-OR/milwaukie-17562" => array("type"=>"city", "pop"=>20707, "name_en"=>"Milwaukie", "lang"=>"en", "lat"=>45.444200, "lon"=>-122.619749),
"/US/US-OR/mission-17563" => array("type"=>"cdp", "pop"=>664, "name_ru"=>"Мишен", "name_en"=>"Mission", "lang"=>"en", "lat"=>45.667093, "lon"=>-118.671809),
"/US/US-OR/mitchell-17564" => array("type"=>"city", "pop"=>137, "name_ru"=>"Митчелл", "name_en"=>"Mitchell", "lang"=>"en", "lat"=>44.567284, "lon"=>-120.152460),
"/US/US-OR/molalla-17565" => array("type"=>"city", "pop"=>7263, "name_en"=>"Molalla", "lang"=>"en", "lat"=>45.149733, "lon"=>-122.578007),
"/US/US-OR/monmouth-17566" => array("type"=>"city", "pop"=>9607, "name_ru"=>"Монмаут", "name_en"=>"Monmouth", "lang"=>"en", "lat"=>44.850375, "lon"=>-123.228810),
"/US/US-OR/monroe-17567" => array("type"=>"city", "pop"=>594, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>44.316818, "lon"=>-123.298499),
"/US/US-OR/monument-17568" => array("type"=>"city", "pop"=>124, "name_ru"=>"Моньюмент", "name_en"=>"Monument", "lang"=>"en", "lat"=>44.819978, "lon"=>-119.418802),
"/US/US-OR/moro-17569" => array("type"=>"city", "pop"=>315, "name_ru"=>"Моро", "name_en"=>"Moro", "lang"=>"en", "lat"=>45.485675, "lon"=>-120.731730),
"/US/US-OR/mosier-17570" => array("type"=>"city", "pop"=>479, "name_en"=>"Mosier", "lang"=>"en", "lat"=>45.683741, "lon"=>-121.394857),
"/US/US-OR/mount-angel-17571" => array("type"=>"city", "pop"=>3470, "name_en"=>"Mount Angel", "lang"=>"en", "lat"=>45.069341, "lon"=>-122.794343),
"/US/US-OR/mount-hood-village-17572" => array("type"=>"cdp", "pop"=>2234, "name_en"=>"Mount Hood Village", "lang"=>"en", "lat"=>45.355544, "lon"=>-121.979476),
"/US/US-OR/mount-vernon-17573" => array("type"=>"city", "pop"=>493, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>44.417778, "lon"=>-119.112196),
"/US/US-OR/myrtle-creek-17574" => array("type"=>"city", "pop"=>3487, "name_en"=>"Myrtle Creek", "lang"=>"en", "lat"=>43.025785, "lon"=>-123.282562),
"/US/US-OR/myrtle-point-17575" => array("type"=>"city", "pop"=>2442, "name_en"=>"Myrtle Point", "lang"=>"en", "lat"=>43.062219, "lon"=>-124.131882),
"/US/US-OR/nehalem-17576" => array("type"=>"city", "pop"=>212, "name_en"=>"Nehalem", "lang"=>"en", "lat"=>45.721084, "lon"=>-123.891543),
"/US/US-OR/newberg-17577" => array("type"=>"city", "pop"=>22953, "name_en"=>"Newberg", "lang"=>"en", "lat"=>45.306728, "lon"=>-122.959752),
"/US/US-OR/newport-17578" => array("type"=>"city", "pop"=>9943, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>44.616705, "lon"=>-124.053893),
"/US/US-OR/north-albany-17579" => array("type"=>"cdp", "pop"=>4325, "name_en"=>"North Albany", "lang"=>"en", "lat"=>44.661688, "lon"=>-123.113573),
"/US/US-OR/north-bend-17580" => array("type"=>"city", "pop"=>9636, "name_en"=>"North Bend", "lang"=>"en", "lat"=>43.406108, "lon"=>-124.236162),
"/US/US-OR/north-plains-17581" => array("type"=>"city", "pop"=>1897, "name_en"=>"North Plains", "lang"=>"en", "lat"=>45.596736, "lon"=>-122.998804),
"/US/US-OR/north-powder-17582" => array("type"=>"city", "pop"=>487, "name_en"=>"North Powder", "lang"=>"en", "lat"=>45.030644, "lon"=>-117.918617),
"/US/US-OR/north-springfield-17583" => array("type"=>"cdp", "pop"=>5451, "name_en"=>"North Springfield", "lang"=>"en", "lat"=>44.075000, "lon"=>-123.001870),
"/US/US-OR/nyssa-17584" => array("type"=>"city", "pop"=>2997, "name_en"=>"Nyssa", "lang"=>"en", "lat"=>43.878459, "lon"=>-116.997316),
"/US/US-OR/oak-grove-17585" => array("type"=>"cdp", "pop"=>12576, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>45.413500, "lon"=>-122.637654),
"/US/US-OR/oak-hills-17586" => array("type"=>"cdp", "pop"=>6450, "name_en"=>"Oak Hills", "lang"=>"en", "lat"=>45.541366, "lon"=>-122.839984),
"/US/US-OR/oakland-17587" => array("type"=>"city", "pop"=>950, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>43.422982, "lon"=>-123.295812),
"/US/US-OR/oakridge-17588" => array("type"=>"city", "pop"=>3193, "name_en"=>"Oakridge", "lang"=>"en", "lat"=>43.747429, "lon"=>-122.471958),
"/US/US-OR/oatfield-17589" => array("type"=>"cdp", "pop"=>15348, "name_en"=>"Oatfield", "lang"=>"en", "lat"=>45.413687, "lon"=>-122.596148),
"/US/US-OR/ontario-17590" => array("type"=>"city", "pop"=>10991, "name_ru"=>"Онтарио", "name_en"=>"Ontario", "lang"=>"en", "lat"=>44.024699, "lon"=>-116.974311),
"/US/US-OR/oregon-city-17591" => array("type"=>"city", "pop"=>31404, "name_en"=>"Oregon City", "lang"=>"en", "lat"=>45.344905, "lon"=>-122.596815),
"/US/US-OR/paisley-17592" => array("type"=>"city", "pop"=>230, "name_en"=>"Paisley", "lang"=>"en", "lat"=>42.692981, "lon"=>-120.544385),
"/US/US-OR/pendleton-17593" => array("type"=>"city", "pop"=>16380, "name_ru"=>"Пендлтон", "name_en"=>"Pendleton", "lang"=>"en", "lat"=>45.673791, "lon"=>-118.817280),
"/US/US-OR/philomath-17594" => array("type"=>"city", "pop"=>4512, "name_en"=>"Philomath", "lang"=>"en", "lat"=>44.540499, "lon"=>-123.357076),
"/US/US-OR/phoenix-17595" => array("type"=>"city", "pop"=>4402, "name_ru"=>"Феникс", "name_en"=>"Phoenix", "lang"=>"en", "lat"=>42.274487, "lon"=>-122.815974),
"/US/US-OR/pilot-rock-17596" => array("type"=>"city", "pop"=>1514, "name_en"=>"Pilot Rock", "lang"=>"en", "lat"=>45.481503, "lon"=>-118.832204),
"/US/US-OR/port-orford-17598" => array("type"=>"city", "pop"=>1110, "name_en"=>"Port Orford", "lang"=>"en", "lat"=>42.749889, "lon"=>-124.495545),
"/US/US-OR/portland-17597" => array("type"=>"city", "pop"=>557706, "name_ru"=>"Портленд", "name_en"=>"Portland", "lang"=>"en", "lat"=>45.538250, "lon"=>-122.656496),
"/US/US-OR/powellhurst-centennial-17599" => array("type"=>"cdp", "pop"=>28756, "name_en"=>"Powellhurst-Centennial", "lang"=>"en", "lat"=>45.496112, "lon"=>-122.515126),
"/US/US-OR/powers-17600" => array("type"=>"city", "pop"=>724, "name_ru"=>"Пауэрс", "name_en"=>"Powers", "lang"=>"en", "lat"=>42.885354, "lon"=>-124.072152),
"/US/US-OR/prairie-city-17601" => array("type"=>"city", "pop"=>899, "name_ru"=>"Прейри Сити", "name_en"=>"Prairie City", "lang"=>"en", "lat"=>44.461577, "lon"=>-118.709536),
"/US/US-OR/prescott-17602" => array("type"=>"city", "pop"=>73, "name_ru"=>"Прескотт", "name_en"=>"Prescott", "lang"=>"en", "lat"=>46.047325, "lon"=>-122.885825),
"/US/US-OR/prineville-17603" => array("type"=>"city", "pop"=>10085, "name_en"=>"Prineville", "lang"=>"en", "lat"=>44.305462, "lon"=>-120.841544),
"/US/US-OR/rainier-17604" => array("type"=>"city", "pop"=>1806, "name_en"=>"Rainier", "lang"=>"en", "lat"=>46.092331, "lon"=>-122.946431),
"/US/US-OR/raleigh-hills-17605" => array("type"=>"cdp", "pop"=>6066, "name_en"=>"Raleigh Hills", "lang"=>"en", "lat"=>45.484950, "lon"=>-122.754290),
"/US/US-OR/redmond-17606" => array("type"=>"city", "pop"=>24551, "name_en"=>"Redmond", "lang"=>"en", "lat"=>44.259475, "lon"=>-121.171228),
"/US/US-OR/redwood-17607" => array("type"=>"cdp", "pop"=>3702, "name_en"=>"Redwood", "lang"=>"en", "lat"=>42.422199, "lon"=>-123.386141),
"/US/US-OR/reedsport-17608" => array("type"=>"city", "pop"=>4212, "name_en"=>"Reedsport", "lang"=>"en", "lat"=>43.698951, "lon"=>-124.110918),
"/US/US-OR/richland-17609" => array("type"=>"city", "pop"=>134, "name_ru"=>"Ричленд", "name_en"=>"Richland", "lang"=>"en", "lat"=>44.767664, "lon"=>-117.167559),
"/US/US-OR/riddle-17610" => array("type"=>"city", "pop"=>998, "name_en"=>"Riddle", "lang"=>"en", "lat"=>42.953697, "lon"=>-123.366122),
"/US/US-OR/river-road-17612" => array("type"=>"cdp", "pop"=>9443, "name_en"=>"River Road", "lang"=>"en", "lat"=>44.084200, "lon"=>-123.132549),
"/US/US-OR/rivergrove-17611" => array("type"=>"city", "pop"=>345, "name_en"=>"Rivergrove", "lang"=>"en", "lat"=>45.385400, "lon"=>-122.733827),
"/US/US-OR/rockaway-beach-17613" => array("type"=>"city", "pop"=>1366, "name_ru"=>"Рокуэй Бич", "name_en"=>"Rockaway Beach", "lang"=>"en", "lat"=>45.614971, "lon"=>-123.938319),
"/US/US-OR/rockcreek-17614" => array("type"=>"cdp", "pop"=>8282, "name_en"=>"Rockcreek", "lang"=>"en", "lat"=>45.550317, "lon"=>-122.875924),
"/US/US-OR/rogue-river-17615" => array("type"=>"city", "pop"=>1932, "name_en"=>"Rogue River", "lang"=>"en", "lat"=>42.435788, "lon"=>-123.167148),
"/US/US-OR/rose-lodge-17618" => array("type"=>"cdp", "pop"=>1257, "name_en"=>"Rose Lodge", "lang"=>"en", "lat"=>45.022052, "lon"=>-123.879746),
"/US/US-OR/roseburg-17616" => array("type"=>"city", "pop"=>20681, "name_en"=>"Roseburg", "lang"=>"en", "lat"=>43.219690, "lon"=>-123.357624),
"/US/US-OR/roseburg-north-17617" => array("type"=>"cdp", "pop"=>6831, "name_en"=>"Roseburg North", "lang"=>"en", "lat"=>43.253623, "lon"=>-123.321107),
"/US/US-OR/rufus-17619" => array("type"=>"city", "pop"=>214, "name_en"=>"Rufus", "lang"=>"en", "lat"=>45.693715, "lon"=>-120.739707),
"/US/US-OR/salem-17622" => array("type"=>"city", "pop"=>153435, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>44.924500, "lon"=>-123.022057),
"/US/US-OR/sandy-17623" => array("type"=>"city", "pop"=>8823, "name_en"=>"Sandy", "lang"=>"en", "lat"=>45.399646, "lon"=>-122.266655),
"/US/US-OR/santa-clara-17624" => array("type"=>"cdp", "pop"=>12834, "name_ru"=>"Санта Клара", "name_en"=>"Santa Clara", "lang"=>"en", "lat"=>44.114500, "lon"=>-123.131874),
"/US/US-OR/scappoose-17625" => array("type"=>"city", "pop"=>6377, "name_en"=>"Scappoose", "lang"=>"en", "lat"=>45.751644, "lon"=>-122.880510),
"/US/US-OR/scio-17626" => array("type"=>"city", "pop"=>783, "name_en"=>"Scio", "lang"=>"en", "lat"=>44.705450, "lon"=>-122.848545),
"/US/US-OR/scotts-mills-17627" => array("type"=>"city", "pop"=>343, "name_en"=>"Scotts Mills", "lang"=>"en", "lat"=>45.040919, "lon"=>-122.668086),
"/US/US-OR/seaside-17628" => array("type"=>"city", "pop"=>6258, "name_ru"=>"Сисайд", "name_en"=>"Seaside", "lang"=>"en", "lat"=>45.990247, "lon"=>-123.920054),
"/US/US-OR/seneca-17629" => array("type"=>"city", "pop"=>183, "name_ru"=>"Сенека", "name_en"=>"Seneca", "lang"=>"en", "lat"=>44.134704, "lon"=>-118.975641),
"/US/US-OR/shady-cove-17630" => array("type"=>"city", "pop"=>2612, "name_en"=>"Shady Cove", "lang"=>"en", "lat"=>42.611790, "lon"=>-122.818703),
"/US/US-OR/shaniko-17631" => array("type"=>"city", "pop"=>26, "name_en"=>"Shaniko", "lang"=>"en", "lat"=>45.004186, "lon"=>-120.750522),
"/US/US-OR/sheridan-17632" => array("type"=>"city", "pop"=>5747, "name_ru"=>"Шеридан", "name_en"=>"Sheridan", "lang"=>"en", "lat"=>45.095412, "lon"=>-123.395100),
"/US/US-OR/sherwood-17633" => array("type"=>"city", "pop"=>17373, "name_ru"=>"Шервуд", "name_en"=>"Sherwood", "lang"=>"en", "lat"=>45.359863, "lon"=>-122.841929),
"/US/US-OR/siletz-17634" => array("type"=>"city", "pop"=>1083, "name_en"=>"Siletz", "lang"=>"en", "lat"=>44.722122, "lon"=>-123.916500),
"/US/US-OR/silverton-17635" => array("type"=>"city", "pop"=>9649, "name_ru"=>"Сильвертон", "name_en"=>"Silverton", "lang"=>"en", "lat"=>45.007728, "lon"=>-122.780701),
"/US/US-OR/sisters-17636" => array("type"=>"city", "pop"=>1642, "name_en"=>"Sisters", "lang"=>"en", "lat"=>44.290990, "lon"=>-121.548184),
"/US/US-OR/sodaville-17637" => array("type"=>"city", "pop"=>293, "name_en"=>"Sodaville", "lang"=>"en", "lat"=>44.483925, "lon"=>-122.867382),
"/US/US-OR/south-lebanon-17638" => array("type"=>"cdp", "pop"=>1203, "name_en"=>"South Lebanon", "lang"=>"en", "lat"=>44.506436, "lon"=>-122.902004),
"/US/US-OR/spray-17639" => array("type"=>"town", "pop"=>113, "name_en"=>"Spray", "lang"=>"en", "lat"=>44.832021, "lon"=>-119.793315),
"/US/US-OR/springfield-17640" => array("type"=>"city", "pop"=>57224, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>44.052917, "lon"=>-122.977072),
"/US/US-OR/st.-helens-17620" => array("type"=>"city", "pop"=>12510, "name_en"=>"St. Helens", "lang"=>"en", "lat"=>45.860600, "lon"=>-122.810970),
"/US/US-OR/st.-paul-17621" => array("type"=>"city", "pop"=>447, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>45.212952, "lon"=>-122.975518),
"/US/US-OR/stanfield-17641" => array("type"=>"city", "pop"=>1926, "name_en"=>"Stanfield", "lang"=>"en", "lat"=>45.786312, "lon"=>-119.219535),
"/US/US-OR/stayton-17642" => array("type"=>"city", "pop"=>7319, "name_en"=>"Stayton", "lang"=>"en", "lat"=>44.801804, "lon"=>-122.796224),
"/US/US-OR/sublimity-17643" => array("type"=>"city", "pop"=>2591, "name_en"=>"Sublimity", "lang"=>"en", "lat"=>44.829208, "lon"=>-122.791670),
"/US/US-OR/summerville-17644" => array("type"=>"town", "pop"=>116, "name_ru"=>"Саммервилл", "name_en"=>"Summerville", "lang"=>"en", "lat"=>45.489843, "lon"=>-118.002523),
"/US/US-OR/sumpter-17645" => array("type"=>"city", "pop"=>157, "name_en"=>"Sumpter", "lang"=>"en", "lat"=>44.743847, "lon"=>-118.195859),
"/US/US-OR/sunnyside-17646" => array("type"=>"cdp", "pop"=>4423, "name_ru"=>"Саннисайд", "name_en"=>"Sunnyside", "lang"=>"en", "lat"=>45.432180, "lon"=>-122.555784),
"/US/US-OR/sutherlin-17647" => array("type"=>"city", "pop"=>7108, "name_en"=>"Sutherlin", "lang"=>"en", "lat"=>43.388596, "lon"=>-123.314988),
"/US/US-OR/sweet-home-17648" => array("type"=>"city", "pop"=>8930, "name_en"=>"Sweet Home", "lang"=>"en", "lat"=>44.401687, "lon"=>-122.702121),
"/US/US-OR/talent-17649" => array("type"=>"city", "pop"=>6215, "name_en"=>"Talent", "lang"=>"en", "lat"=>42.240134, "lon"=>-122.780955),
"/US/US-OR/tangent-17650" => array("type"=>"city", "pop"=>994, "name_en"=>"Tangent", "lang"=>"en", "lat"=>44.551198, "lon"=>-123.107901),
"/US/US-OR/terrebonne-17651" => array("type"=>"cdp", "pop"=>1143, "name_en"=>"Terrebonne", "lang"=>"en", "lat"=>44.342008, "lon"=>-121.177369),
"/US/US-OR/three-rivers-17652" => array("type"=>"cdp", "pop"=>1268, "name_ru"=>"Три Риверс", "name_en"=>"Three Rivers", "lang"=>"en", "lat"=>43.820139, "lon"=>-121.468007),
"/US/US-OR/tigard-17653" => array("type"=>"city", "pop"=>48713, "name_en"=>"Tigard", "lang"=>"en", "lat"=>45.425000, "lon"=>-122.773936),
"/US/US-OR/tillamook-17654" => array("type"=>"city", "pop"=>4430, "name_en"=>"Tillamook", "lang"=>"en", "lat"=>45.456673, "lon"=>-123.837558),
"/US/US-OR/toledo-17655" => array("type"=>"city", "pop"=>3296, "name_ru"=>"Толедо", "name_en"=>"Toledo", "lang"=>"en", "lat"=>44.620949, "lon"=>-123.932521),
"/US/US-OR/tri-city-17656" => array("type"=>"cdp", "pop"=>3585, "name_en"=>"Tri-City", "lang"=>"en", "lat"=>42.984657, "lon"=>-123.310455),
"/US/US-OR/troutdale-17657" => array("type"=>"city", "pop"=>15438, "name_en"=>"Troutdale", "lang"=>"en", "lat"=>45.536909, "lon"=>-122.390751),
"/US/US-OR/tualatin-17658" => array("type"=>"city", "pop"=>26476, "name_en"=>"Tualatin", "lang"=>"en", "lat"=>45.376331, "lon"=>-122.771680),
"/US/US-OR/turner-17659" => array("type"=>"city", "pop"=>1717, "name_ru"=>"Тернер", "name_en"=>"Turner", "lang"=>"en", "lat"=>44.846088, "lon"=>-122.951512),
"/US/US-OR/ukiah-17660" => array("type"=>"city", "pop"=>249, "name_ru"=>"Юкайа", "name_en"=>"Ukiah", "lang"=>"en", "lat"=>45.133869, "lon"=>-118.931718),
"/US/US-OR/umatilla-17661" => array("type"=>"city", "pop"=>6280, "name_ru"=>"Уматилла", "name_en"=>"Umatilla", "lang"=>"en", "lat"=>45.913294, "lon"=>-119.327815),
"/US/US-OR/union-17662" => array("type"=>"city", "pop"=>1954, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>45.209007, "lon"=>-117.867126),
"/US/US-OR/unity-17663" => array("type"=>"city", "pop"=>120, "name_en"=>"Unity", "lang"=>"en", "lat"=>44.436799, "lon"=>-118.187300),
"/US/US-OR/vale-17664" => array("type"=>"city", "pop"=>1883, "name_en"=>"Vale", "lang"=>"en", "lat"=>43.983114, "lon"=>-117.240526),
"/US/US-OR/veneta-17665" => array("type"=>"city", "pop"=>4275, "name_en"=>"Veneta", "lang"=>"en", "lat"=>44.048344, "lon"=>-123.352329),
"/US/US-OR/vernonia-17666" => array("type"=>"city", "pop"=>2275, "name_en"=>"Vernonia", "lang"=>"en", "lat"=>45.860538, "lon"=>-123.185208),
"/US/US-OR/waldport-17667" => array("type"=>"city", "pop"=>2025, "name_en"=>"Waldport", "lang"=>"en", "lat"=>44.421435, "lon"=>-124.064146),
"/US/US-OR/wallowa-17668" => array("type"=>"city", "pop"=>782, "name_en"=>"Wallowa", "lang"=>"en", "lat"=>45.570674, "lon"=>-117.527703),
"/US/US-OR/warm-springs-17669" => array("type"=>"cdp", "pop"=>2287, "name_ru"=>"Уорм Спрингс", "name_en"=>"Warm Springs", "lang"=>"en", "lat"=>44.769178, "lon"=>-121.285738),
"/US/US-OR/warrenton-17670" => array("type"=>"city", "pop"=>4448, "name_ru"=>"Уоррентон", "name_en"=>"Warrenton", "lang"=>"en", "lat"=>46.170252, "lon"=>-123.920887),
"/US/US-OR/wasco-17671" => array("type"=>"city", "pop"=>304, "name_ru"=>"Васко", "name_en"=>"Wasco", "lang"=>"en", "lat"=>45.591587, "lon"=>-120.695741),
"/US/US-OR/waterloo-17672" => array("type"=>"town", "pop"=>241, "name_ru"=>"Ватрелоо", "name_en"=>"Waterloo", "lang"=>"en", "lat"=>44.495275, "lon"=>-122.821634),
"/US/US-OR/west-haven-sylvan-17674" => array("type"=>"cdp", "pop"=>6009, "name_en"=>"West Haven-Sylvan", "lang"=>"en", "lat"=>45.517800, "lon"=>-122.767668),
"/US/US-OR/west-linn-17675" => array("type"=>"city", "pop"=>25236, "name_en"=>"West Linn", "lang"=>"en", "lat"=>45.368536, "lon"=>-122.638220),
"/US/US-OR/west-slope-17677" => array("type"=>"cdp", "pop"=>7959, "name_en"=>"West Slope", "lang"=>"en", "lat"=>45.496985, "lon"=>-122.769380),
"/US/US-OR/westfir-17673" => array("type"=>"city", "pop"=>278, "name_en"=>"Westfir", "lang"=>"en", "lat"=>43.758237, "lon"=>-122.506442),
"/US/US-OR/weston-17676" => array("type"=>"city", "pop"=>693, "name_ru"=>"Вэстон", "name_en"=>"Weston", "lang"=>"en", "lat"=>45.815803, "lon"=>-118.424724),
"/US/US-OR/wheeler-17678" => array("type"=>"city", "pop"=>389, "name_ru"=>"Уилер", "name_en"=>"Wheeler", "lang"=>"en", "lat"=>45.686971, "lon"=>-123.884066),
"/US/US-OR/white-city-17679" => array("type"=>"cdp", "pop"=>5891, "name_ru"=>"Уайт Сити", "name_en"=>"White City", "lang"=>"en", "lat"=>42.431987, "lon"=>-122.830070),
"/US/US-OR/willamina-17680" => array("type"=>"city", "pop"=>1942, "name_en"=>"Willamina", "lang"=>"en", "lat"=>45.079129, "lon"=>-123.483486),
"/US/US-OR/wilsonville-17681" => array("type"=>"city", "pop"=>19055, "name_ru"=>"Уилсонвилл", "name_en"=>"Wilsonville", "lang"=>"en", "lat"=>45.309088, "lon"=>-122.766301),
"/US/US-OR/winston-17682" => array("type"=>"city", "pop"=>5528, "name_ru"=>"Уинстон", "name_en"=>"Winston", "lang"=>"en", "lat"=>43.120428, "lon"=>-123.411357),
"/US/US-OR/wood-village-17684" => array("type"=>"city", "pop"=>3166, "name_en"=>"Wood Village", "lang"=>"en", "lat"=>45.533464, "lon"=>-122.420279),
"/US/US-OR/woodburn-17683" => array("type"=>"city", "pop"=>22728, "name_ru"=>"Вудберн", "name_en"=>"Woodburn", "lang"=>"en", "lat"=>45.148423, "lon"=>-122.855794),
"/US/US-OR/yachats-17685" => array("type"=>"city", "pop"=>675, "name_en"=>"Yachats", "lang"=>"en", "lat"=>44.312680, "lon"=>-124.100460),
"/US/US-OR/yamhill-17686" => array("type"=>"city", "pop"=>915, "name_en"=>"Yamhill", "lang"=>"en", "lat"=>45.341337, "lon"=>-123.184363),
"/US/US-OR/yoncalla-17687" => array("type"=>"city", "pop"=>1036, "name_en"=>"Yoncalla", "lang"=>"en", "lat"=>43.600182, "lon"=>-123.285396),
"/US/US-PA" => array("type"=>"state", "pop"=>12448279, "name_ru"=>"Пенсильвания", "name_en"=>"Pennsylvania", "lang"=>"en"),
"/US/US-PA/abbottstown-17688" => array("type"=>"borough", "pop"=>934, "name_en"=>"Abbottstown", "lang"=>"en", "lat"=>39.884543, "lon"=>-76.989694),
"/US/US-PA/adamsburg-17689" => array("type"=>"borough", "pop"=>204, "name_en"=>"Adamsburg", "lang"=>"en", "lat"=>40.310432, "lon"=>-79.654659),
"/US/US-PA/adamstown-17690" => array("type"=>"borough", "pop"=>1530, "name_en"=>"Adamstown", "lang"=>"en", "lat"=>40.240932, "lon"=>-76.062595),
"/US/US-PA/addison-17691" => array("type"=>"borough", "pop"=>197, "name_ru"=>"Эддиссон", "name_en"=>"Addison", "lang"=>"en", "lat"=>39.743502, "lon"=>-79.335254),
"/US/US-PA/akron-17692" => array("type"=>"borough", "pop"=>4017, "name_ru"=>"Акрон", "name_en"=>"Akron", "lang"=>"en", "lat"=>40.157550, "lon"=>-76.203663),
"/US/US-PA/alba-17693" => array("type"=>"borough", "pop"=>174, "name_ru"=>"Альба", "name_en"=>"Alba", "lang"=>"en", "lat"=>41.704269, "lon"=>-76.827839),
"/US/US-PA/albion-17694" => array("type"=>"borough", "pop"=>1501, "name_ru"=>"Албион", "name_en"=>"Albion", "lang"=>"en", "lat"=>41.889602, "lon"=>-80.363845),
"/US/US-PA/alburtis-17695" => array("type"=>"borough", "pop"=>2392, "name_en"=>"Alburtis", "lang"=>"en", "lat"=>40.509456, "lon"=>-75.600708),
"/US/US-PA/aldan-17696" => array("type"=>"borough", "pop"=>4238, "name_en"=>"Aldan", "lang"=>"en", "lat"=>39.922861, "lon"=>-75.288801),
"/US/US-PA/alexandria-17697" => array("type"=>"borough", "pop"=>373, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>40.558863, "lon"=>-78.100087),
"/US/US-PA/aliquippa-17698" => array("type"=>"city", "pop"=>10631, "name_en"=>"Aliquippa", "lang"=>"en", "lat"=>40.617400, "lon"=>-80.255042),
"/US/US-PA/allenport-17699" => array("type"=>"borough", "pop"=>513, "name_en"=>"Allenport", "lang"=>"en", "lat"=>40.090872, "lon"=>-79.856921),
"/US/US-PA/allentown-17700" => array("type"=>"city", "pop"=>107250, "name_ru"=>"Аллентаун", "name_en"=>"Allentown", "lang"=>"en", "lat"=>40.596333, "lon"=>-75.477665),
"/US/US-PA/almedia-17701" => array("type"=>"cdp", "pop"=>1116, "name_en"=>"Almedia", "lang"=>"en", "lat"=>41.014407, "lon"=>-76.387979),
"/US/US-PA/altoona-17702" => array("type"=>"city", "pop"=>46144, "name_ru"=>"Альтоона", "name_en"=>"Altoona", "lang"=>"en", "lat"=>40.507850, "lon"=>-78.401285),
"/US/US-PA/ambler-17703" => array("type"=>"borough", "pop"=>6196, "name_ru"=>"Амблер", "name_en"=>"Ambler", "lang"=>"en", "lat"=>40.156050, "lon"=>-75.221949),
"/US/US-PA/ambridge-17704" => array("type"=>"borough", "pop"=>7024, "name_en"=>"Ambridge", "lang"=>"en", "lat"=>40.593100, "lon"=>-80.225418),
"/US/US-PA/amity-gardens-17705" => array("type"=>"cdp", "pop"=>2714, "name_en"=>"Amity Gardens", "lang"=>"en", "lat"=>40.274135, "lon"=>-75.735622),
"/US/US-PA/ancient-oaks-17706" => array("type"=>"cdp", "pop"=>2663, "name_en"=>"Ancient Oaks", "lang"=>"en", "lat"=>40.532536, "lon"=>-75.585733),
"/US/US-PA/annville-17707" => array("type"=>"cdp", "pop"=>4758, "name_ru"=>"Эннвилл", "name_en"=>"Annville", "lang"=>"en", "lat"=>40.332501, "lon"=>-76.506550),
"/US/US-PA/apollo-17708" => array("type"=>"borough", "pop"=>1633, "name_en"=>"Apollo", "lang"=>"en", "lat"=>40.584560, "lon"=>-79.564622),
"/US/US-PA/applewold-17709" => array("type"=>"borough", "pop"=>333, "name_en"=>"Applewold", "lang"=>"en", "lat"=>40.807462, "lon"=>-79.522338),
"/US/US-PA/archbald-17710" => array("type"=>"borough", "pop"=>6488, "name_en"=>"Archbald", "lang"=>"en", "lat"=>41.507183, "lon"=>-75.542220),
"/US/US-PA/ardmore-17711" => array("type"=>"cdp", "pop"=>12646, "name_ru"=>"Ардмор", "name_en"=>"Ardmore", "lang"=>"en", "lat"=>40.002850, "lon"=>-75.294795),
"/US/US-PA/arendtsville-17712" => array("type"=>"borough", "pop"=>891, "name_en"=>"Arendtsville", "lang"=>"en", "lat"=>39.924974, "lon"=>-77.301203),
"/US/US-PA/arlington-heights-17713" => array("type"=>"cdp", "pop"=>4768, "name_ru"=>"Арлингтон Хайтс", "name_en"=>"Arlington Heights", "lang"=>"en", "lat"=>41.004392, "lon"=>-75.213083),
"/US/US-PA/armagh-17714" => array("type"=>"borough", "pop"=>122, "name_en"=>"Armagh", "lang"=>"en", "lat"=>40.453216, "lon"=>-79.033097),
"/US/US-PA/arnold-17715" => array("type"=>"city", "pop"=>5212, "name_ru"=>"Арнолд", "name_en"=>"Arnold", "lang"=>"en", "lat"=>40.578950, "lon"=>-79.764876),
"/US/US-PA/arona-17716" => array("type"=>"borough", "pop"=>376, "name_en"=>"Arona", "lang"=>"en", "lat"=>40.268240, "lon"=>-79.657479),
"/US/US-PA/ashland-17717" => array("type"=>"borough", "pop"=>3099, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>40.780951, "lon"=>-76.345661),
"/US/US-PA/ashley-17718" => array("type"=>"borough", "pop"=>2663, "name_ru"=>"Эшли", "name_en"=>"Ashley", "lang"=>"en", "lat"=>41.213667, "lon"=>-75.898964),
"/US/US-PA/ashville-17719" => array("type"=>"borough", "pop"=>258, "name_ru"=>"Ашвилл", "name_en"=>"Ashville", "lang"=>"en", "lat"=>40.559145, "lon"=>-78.547421),
"/US/US-PA/aspinwall-17720" => array("type"=>"borough", "pop"=>2704, "name_ru"=>"Аспинуолл", "name_en"=>"Aspinwall", "lang"=>"en", "lat"=>40.492300, "lon"=>-79.903699),
"/US/US-PA/atglen-17721" => array("type"=>"borough", "pop"=>1365, "name_en"=>"Atglen", "lang"=>"en", "lat"=>39.947059, "lon"=>-75.975266),
"/US/US-PA/athens-17722" => array("type"=>"borough", "pop"=>3209, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>41.950376, "lon"=>-76.517075),
"/US/US-PA/atwood-17723" => array("type"=>"borough", "pop"=>105, "name_ru"=>"Атвуд", "name_en"=>"Atwood", "lang"=>"en", "lat"=>40.748013, "lon"=>-79.261866),
"/US/US-PA/auburn-17724" => array("type"=>"borough", "pop"=>799, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>40.596259, "lon"=>-76.096179),
"/US/US-PA/audubon-17725" => array("type"=>"cdp", "pop"=>6328, "name_ru"=>"Одубон", "name_en"=>"Audubon", "lang"=>"en", "lat"=>40.129850, "lon"=>-75.428362),
"/US/US-PA/austin-17726" => array("type"=>"borough", "pop"=>562, "name_ru"=>"Остин", "name_en"=>"Austin", "lang"=>"en", "lat"=>41.636047, "lon"=>-78.089008),
"/US/US-PA/avalon-17727" => array("type"=>"borough", "pop"=>4812, "name_ru"=>"Авалон", "name_en"=>"Avalon", "lang"=>"en", "lat"=>40.501000, "lon"=>-80.068699),
"/US/US-PA/avis-17728" => array("type"=>"borough", "pop"=>1461, "name_en"=>"Avis", "lang"=>"en", "lat"=>41.186214, "lon"=>-77.316417),
"/US/US-PA/avoca-17729" => array("type"=>"borough", "pop"=>2657, "name_ru"=>"Авока", "name_en"=>"Avoca", "lang"=>"en", "lat"=>41.337683, "lon"=>-75.743097),
"/US/US-PA/avon-heights-17731" => array("type"=>"cdp", "pop"=>2714, "name_en"=>"Avon Heights", "lang"=>"en", "lat"=>40.345832, "lon"=>-76.387075),
"/US/US-PA/avondale-17730" => array("type"=>"borough", "pop"=>1086, "name_ru"=>"Авондейл", "name_en"=>"Avondale", "lang"=>"en", "lat"=>39.823916, "lon"=>-75.782395),
"/US/US-PA/avonia-17732" => array("type"=>"cdp", "pop"=>1336, "name_en"=>"Avonia", "lang"=>"en", "lat"=>42.044072, "lon"=>-80.273465),
"/US/US-PA/avonmore-17733" => array("type"=>"borough", "pop"=>760, "name_en"=>"Avonmore", "lang"=>"en", "lat"=>40.526822, "lon"=>-79.469978),
"/US/US-PA/baden-17734" => array("type"=>"borough", "pop"=>4015, "name_en"=>"Baden", "lang"=>"en", "lat"=>40.638550, "lon"=>-80.221900),
"/US/US-PA/baidland-17735" => array("type"=>"cdp", "pop"=>1620, "name_en"=>"Baidland", "lang"=>"en", "lat"=>40.189080, "lon"=>-79.955074),
"/US/US-PA/baldwin-17736" => array("type"=>"borough", "pop"=>18562, "name_ru"=>"Болдуин", "name_en"=>"Baldwin", "lang"=>"en", "lat"=>40.363208, "lon"=>-79.966150),
"/US/US-PA/bally-17737" => array("type"=>"borough", "pop"=>1094, "name_en"=>"Bally", "lang"=>"en", "lat"=>40.400947, "lon"=>-75.588736),
"/US/US-PA/bangor-17738" => array("type"=>"borough", "pop"=>5256, "name_ru"=>"Бангор", "name_en"=>"Bangor", "lang"=>"en", "lat"=>40.868496, "lon"=>-75.208162),
"/US/US-PA/barkeyville-17739" => array("type"=>"borough", "pop"=>218, "name_en"=>"Barkeyville", "lang"=>"en", "lat"=>41.200192, "lon"=>-79.983112),
"/US/US-PA/barnesboro-17740" => array("type"=>"borough", "pop"=>2530, "name_en"=>"Barnesboro", "lang"=>"en", "lat"=>40.666257, "lon"=>-78.784963),
"/US/US-PA/bath-17741" => array("type"=>"borough", "pop"=>2731, "name_ru"=>"Бат", "name_en"=>"Bath", "lang"=>"en", "lat"=>40.727584, "lon"=>-75.391790),
"/US/US-PA/beallsville-17742" => array("type"=>"borough", "pop"=>483, "name_en"=>"Beallsville", "lang"=>"en", "lat"=>40.058560, "lon"=>-80.033114),
"/US/US-PA/bear-lake-17743" => array("type"=>"borough", "pop"=>175, "name_ru"=>"Бэр Лейк", "name_en"=>"Bear Lake", "lang"=>"en", "lat"=>41.993051, "lon"=>-79.500957),
"/US/US-PA/beaver-17744" => array("type"=>"borough", "pop"=>4371, "name_ru"=>"Бивер", "name_en"=>"Beaver", "lang"=>"en", "lat"=>40.692900, "lon"=>-80.307510),
"/US/US-PA/beaver-falls-17746" => array("type"=>"city", "pop"=>9055, "name_en"=>"Beaver Falls", "lang"=>"en", "lat"=>40.760933, "lon"=>-80.322395),
"/US/US-PA/beaver-meadows-17747" => array("type"=>"borough", "pop"=>946, "name_en"=>"Beaver Meadows", "lang"=>"en", "lat"=>40.929648, "lon"=>-75.912746),
"/US/US-PA/beaverdale-lloydell-17745" => array("type"=>"cdp", "pop"=>1278, "name_en"=>"Beaverdale-Lloydell", "lang"=>"en", "lat"=>40.322909, "lon"=>-78.698849),
"/US/US-PA/beavertown-17748" => array("type"=>"borough", "pop"=>871, "name_en"=>"Beavertown", "lang"=>"en", "lat"=>40.752631, "lon"=>-77.169744),
"/US/US-PA/bechtelsville-17749" => array("type"=>"borough", "pop"=>968, "name_en"=>"Bechtelsville", "lang"=>"en", "lat"=>40.370863, "lon"=>-75.630715),
"/US/US-PA/bedford-17750" => array("type"=>"borough", "pop"=>2999, "name_ru"=>"Бедфорд", "name_en"=>"Bedford", "lang"=>"en", "lat"=>40.014903, "lon"=>-78.504360),
"/US/US-PA/beech-creek-17751" => array("type"=>"borough", "pop"=>691, "name_en"=>"Beech Creek", "lang"=>"en", "lat"=>41.074672, "lon"=>-77.586045),
"/US/US-PA/belfast-17752" => array("type"=>"cdp", "pop"=>1428, "name_ru"=>"Бэлфаст", "name_en"=>"Belfast", "lang"=>"en", "lat"=>40.782522, "lon"=>-75.273145),
"/US/US-PA/bell-acres-17753" => array("type"=>"borough", "pop"=>1384, "name_en"=>"Bell Acres", "lang"=>"en", "lat"=>40.592112, "lon"=>-80.176285),
"/US/US-PA/belle-vernon-17755" => array("type"=>"borough", "pop"=>1137, "name_en"=>"Belle Vernon", "lang"=>"en", "lat"=>40.124603, "lon"=>-79.865785),
"/US/US-PA/bellefonte-17754" => array("type"=>"borough", "pop"=>6151, "name_ru"=>"Беллефонт", "name_en"=>"Bellefonte", "lang"=>"en", "lat"=>40.913767, "lon"=>-77.768577),
"/US/US-PA/belleville-17756" => array("type"=>"cdp", "pop"=>1589, "name_ru"=>"Белелвилл", "name_en"=>"Belleville", "lang"=>"en", "lat"=>40.602827, "lon"=>-77.725677),
"/US/US-PA/bellevue-17757" => array("type"=>"borough", "pop"=>7980, "name_ru"=>"Белвью", "name_en"=>"Bellevue", "lang"=>"en", "lat"=>40.493900, "lon"=>-80.055926),
"/US/US-PA/bellwood-17758" => array("type"=>"borough", "pop"=>1869, "name_ru"=>"Беллвуд", "name_en"=>"Bellwood", "lang"=>"en", "lat"=>40.601117, "lon"=>-78.334207),
"/US/US-PA/belmont-17759" => array("type"=>"cdp", "pop"=>3184, "name_ru"=>"Белмонт", "name_en"=>"Belmont", "lang"=>"en", "lat"=>40.282433, "lon"=>-78.891738),
"/US/US-PA/ben-avon-17760" => array("type"=>"borough", "pop"=>1747, "name_en"=>"Ben Avon", "lang"=>"en", "lat"=>40.506550, "lon"=>-80.082447),
"/US/US-PA/ben-avon-heights-17761" => array("type"=>"borough", "pop"=>359, "name_en"=>"Ben Avon Heights", "lang"=>"en", "lat"=>40.511500, "lon"=>-80.072923),
"/US/US-PA/bendersville-17762" => array("type"=>"borough", "pop"=>601, "name_en"=>"Bendersville", "lang"=>"en", "lat"=>39.982374, "lon"=>-77.249903),
"/US/US-PA/benson-17763" => array("type"=>"borough", "pop"=>177, "name_ru"=>"Бенсон", "name_en"=>"Benson", "lang"=>"en", "lat"=>40.202569, "lon"=>-78.929436),
"/US/US-PA/bentleyville-17764" => array("type"=>"borough", "pop"=>2380, "name_en"=>"Bentleyville", "lang"=>"en", "lat"=>40.117796, "lon"=>-80.003648),
"/US/US-PA/benton-17765" => array("type"=>"borough", "pop"=>914, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>41.194944, "lon"=>-76.385261),
"/US/US-PA/berlin-17766" => array("type"=>"borough", "pop"=>2072, "name_ru"=>"Берлин", "name_en"=>"Berlin", "lang"=>"en", "lat"=>39.922372, "lon"=>-78.952502),
"/US/US-PA/bernville-17767" => array("type"=>"borough", "pop"=>877, "name_en"=>"Bernville", "lang"=>"en", "lat"=>40.434150, "lon"=>-76.110747),
"/US/US-PA/berrysburg-17768" => array("type"=>"borough", "pop"=>339, "name_en"=>"Berrysburg", "lang"=>"en", "lat"=>40.602837, "lon"=>-76.809063),
"/US/US-PA/berwick-17769" => array("type"=>"borough", "pop"=>10216, "name_ru"=>"Бервик", "name_en"=>"Berwick", "lang"=>"en", "lat"=>41.055616, "lon"=>-76.248947),
"/US/US-PA/bessemer-17770" => array("type"=>"borough", "pop"=>1096, "name_ru"=>"Бессемер", "name_en"=>"Bessemer", "lang"=>"en", "lat"=>40.976775, "lon"=>-80.487457),
"/US/US-PA/bethany-17771" => array("type"=>"borough", "pop"=>282, "name_ru"=>"Бетани", "name_en"=>"Bethany", "lang"=>"en", "lat"=>41.614234, "lon"=>-75.288916),
"/US/US-PA/bethel-park-17772" => array("type"=>"borough", "pop"=>31562, "name_en"=>"Bethel Park", "lang"=>"en", "lat"=>40.323600, "lon"=>-80.036062),
"/US/US-PA/bethlehem-17773" => array("type"=>"city", "pop"=>72241, "name_ru"=>"Вифлеем", "name_en"=>"Bethlehem", "lang"=>"en", "lat"=>40.626700, "lon"=>-75.368049),
"/US/US-PA/big-beaver-17774" => array("type"=>"borough", "pop"=>2106, "name_en"=>"Big Beaver", "lang"=>"en", "lat"=>40.818339, "lon"=>-80.363576),
"/US/US-PA/big-run-17776" => array("type"=>"borough", "pop"=>646, "name_en"=>"Big Run", "lang"=>"en", "lat"=>40.969882, "lon"=>-78.878093),
"/US/US-PA/biglerville-17775" => array("type"=>"borough", "pop"=>1146, "name_en"=>"Biglerville", "lang"=>"en", "lat"=>39.930145, "lon"=>-77.247227),
"/US/US-PA/birdsboro-17777" => array("type"=>"borough", "pop"=>5168, "name_en"=>"Birdsboro", "lang"=>"en", "lat"=>40.260700, "lon"=>-75.810523),
"/US/US-PA/birmingham-17778" => array("type"=>"borough", "pop"=>85, "name_ru"=>"Бирмингем", "name_en"=>"Birmingham", "lang"=>"en", "lat"=>40.647211, "lon"=>-78.195448),
"/US/US-PA/black-lick-17779" => array("type"=>"cdp", "pop"=>1283, "name_en"=>"Black Lick", "lang"=>"en", "lat"=>40.464973, "lon"=>-79.187951),
"/US/US-PA/blain-17780" => array("type"=>"borough", "pop"=>245, "name_en"=>"Blain", "lang"=>"en", "lat"=>40.337221, "lon"=>-77.512224),
"/US/US-PA/blairsville-17781" => array("type"=>"borough", "pop"=>3370, "name_ru"=>"Блэйрсвилл", "name_en"=>"Blairsville", "lang"=>"en", "lat"=>40.431513, "lon"=>-79.260105),
"/US/US-PA/blakely-17782" => array("type"=>"borough", "pop"=>6747, "name_ru"=>"Блэйкли", "name_en"=>"Blakely", "lang"=>"en", "lat"=>41.487836, "lon"=>-75.597830),
"/US/US-PA/blawnox-17783" => array("type"=>"borough", "pop"=>1436, "name_en"=>"Blawnox", "lang"=>"en", "lat"=>40.491050, "lon"=>-79.862248),
"/US/US-PA/bloomfield-17784" => array("type"=>"borough", "pop"=>1090, "name_ru"=>"Блумфилд", "name_en"=>"Bloomfield", "lang"=>"en", "lat"=>40.419687, "lon"=>-77.188757),
"/US/US-PA/blooming-valley-17785" => array("type"=>"borough", "pop"=>467, "name_en"=>"Blooming Valley", "lang"=>"en", "lat"=>41.680255, "lon"=>-80.041589),
"/US/US-PA/bloomsburg-17786" => array("type"=>"town", "pop"=>12772, "name_en"=>"Bloomsburg", "lang"=>"en", "lat"=>41.002345, "lon"=>-76.456893),
"/US/US-PA/blossburg-17787" => array("type"=>"borough", "pop"=>1447, "name_en"=>"Blossburg", "lang"=>"en", "lat"=>41.682851, "lon"=>-77.066625),
"/US/US-PA/blue-bell-17788" => array("type"=>"cdp", "pop"=>6091, "name_en"=>"Blue Bell", "lang"=>"en", "lat"=>40.145731, "lon"=>-75.267656),
"/US/US-PA/boalsburg-17789" => array("type"=>"cdp", "pop"=>2206, "name_en"=>"Boalsburg", "lang"=>"en", "lat"=>40.773513, "lon"=>-77.791646),
"/US/US-PA/boiling-springs-17790" => array("type"=>"cdp", "pop"=>1978, "name_en"=>"Boiling Springs", "lang"=>"en", "lat"=>40.157758, "lon"=>-77.137276),
"/US/US-PA/bolivar-17791" => array("type"=>"borough", "pop"=>464, "name_ru"=>"Боливар", "name_en"=>"Bolivar", "lang"=>"en", "lat"=>40.394823, "lon"=>-79.152143),
"/US/US-PA/bonneauville-17792" => array("type"=>"borough", "pop"=>1457, "name_en"=>"Bonneauville", "lang"=>"en", "lat"=>39.811416, "lon"=>-77.136832),
"/US/US-PA/boothwyn-17793" => array("type"=>"cdp", "pop"=>5069, "name_en"=>"Boothwyn", "lang"=>"en", "lat"=>39.835550, "lon"=>-75.445226),
"/US/US-PA/boswell-17794" => array("type"=>"borough", "pop"=>1253, "name_ru"=>"Босуэлл", "name_en"=>"Boswell", "lang"=>"en", "lat"=>40.160814, "lon"=>-79.027107),
"/US/US-PA/bowmanstown-17795" => array("type"=>"borough", "pop"=>887, "name_en"=>"Bowmanstown", "lang"=>"en", "lat"=>40.801474, "lon"=>-75.661475),
"/US/US-PA/boyertown-17796" => array("type"=>"borough", "pop"=>3919, "name_en"=>"Boyertown", "lang"=>"en", "lat"=>40.331767, "lon"=>-75.637805),
"/US/US-PA/brackenridge-17797" => array("type"=>"borough", "pop"=>3227, "name_en"=>"Brackenridge", "lang"=>"en", "lat"=>40.607416, "lon"=>-79.742115),
"/US/US-PA/braddock-17798" => array("type"=>"borough", "pop"=>2667, "name_en"=>"Braddock", "lang"=>"en", "lat"=>40.402000, "lon"=>-79.869349),
"/US/US-PA/braddock-hills-17799" => array("type"=>"borough", "pop"=>1827, "name_en"=>"Braddock Hills", "lang"=>"en", "lat"=>40.416750, "lon"=>-79.863415),
"/US/US-PA/bradford-17800" => array("type"=>"city", "pop"=>8391, "name_ru"=>"Брадфорд", "name_en"=>"Bradford", "lang"=>"en", "lat"=>41.960457, "lon"=>-78.640229),
"/US/US-PA/bradfordwoods-17801" => array("type"=>"borough", "pop"=>1077, "name_en"=>"Bradfordwoods", "lang"=>"en", "lat"=>40.635771, "lon"=>-80.080307),
"/US/US-PA/brentwood-17802" => array("type"=>"borough", "pop"=>9522, "name_ru"=>"Брентвуд", "name_en"=>"Brentwood", "lang"=>"en", "lat"=>40.373267, "lon"=>-79.975966),
"/US/US-PA/bressler-enhaut-oberlin-17803" => array("type"=>"cdp", "pop"=>2660, "name_en"=>"Bressler-Enhaut-Oberlin", "lang"=>"en", "lat"=>40.232500, "lon"=>-76.819610),
"/US/US-PA/briar-creek-17804" => array("type"=>"borough", "pop"=>636, "name_en"=>"Briar Creek", "lang"=>"en", "lat"=>41.047002, "lon"=>-76.286829),
"/US/US-PA/brickerville-17805" => array("type"=>"cdp", "pop"=>1268, "name_en"=>"Brickerville", "lang"=>"en", "lat"=>40.226804, "lon"=>-76.286541),
"/US/US-PA/bridgeport-17806" => array("type"=>"borough", "pop"=>4350, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>40.103933, "lon"=>-75.343446),
"/US/US-PA/bridgeville-17807" => array("type"=>"borough", "pop"=>4876, "name_ru"=>"Бриджвилл", "name_en"=>"Bridgeville", "lang"=>"en", "lat"=>40.356567, "lon"=>-80.106070),
"/US/US-PA/bridgewater-17808" => array("type"=>"borough", "pop"=>863, "name_ru"=>"Бриджуотер", "name_en"=>"Bridgewater", "lang"=>"en", "lat"=>40.709000, "lon"=>-80.300505),
"/US/US-PA/brisbin-17809" => array("type"=>"borough", "pop"=>387, "name_en"=>"Brisbin", "lang"=>"en", "lat"=>40.839206, "lon"=>-78.352206),
"/US/US-PA/bristol-17810" => array("type"=>"borough", "pop"=>9635, "name_ru"=>"Бристол", "name_en"=>"Bristol", "lang"=>"en", "lat"=>40.101650, "lon"=>-74.853154),
"/US/US-PA/brittany-farms-highlands-17811" => array("type"=>"cdp", "pop"=>2747, "name_en"=>"Brittany Farms-Highlands", "lang"=>"en", "lat"=>40.268823, "lon"=>-75.214490),
"/US/US-PA/broad-top-city-17812" => array("type"=>"borough", "pop"=>385, "name_en"=>"Broad Top City", "lang"=>"en", "lat"=>40.201536, "lon"=>-78.140663),
"/US/US-PA/brockway-17813" => array("type"=>"borough", "pop"=>2057, "name_en"=>"Brockway", "lang"=>"en", "lat"=>41.247054, "lon"=>-78.793330),
"/US/US-PA/brodheadsville-17814" => array("type"=>"cdp", "pop"=>1389, "name_en"=>"Brodheadsville", "lang"=>"en", "lat"=>40.924059, "lon"=>-75.403306),
"/US/US-PA/brookhaven-17815" => array("type"=>"borough", "pop"=>7886, "name_ru"=>"Брукхейвен", "name_en"=>"Brookhaven", "lang"=>"en", "lat"=>39.870550, "lon"=>-75.391278),
"/US/US-PA/brookville-17816" => array("type"=>"borough", "pop"=>3993, "name_ru"=>"Бруквилл", "name_en"=>"Brookville", "lang"=>"en", "lat"=>41.161603, "lon"=>-79.082830),
"/US/US-PA/broomall-17817" => array("type"=>"cdp", "pop"=>10930, "name_en"=>"Broomall", "lang"=>"en", "lat"=>39.969950, "lon"=>-75.353023),
"/US/US-PA/brownstown-17818" => array("type"=>"borough", "pop"=>799, "name_ru"=>"Браунстаун", "name_en"=>"Brownstown", "lang"=>"en", "lat"=>40.332850, "lon"=>-78.937000),
"/US/US-PA/brownsville-17819" => array("type"=>"borough", "pop"=>2631, "name_ru"=>"Браунсвилл", "name_en"=>"Brownsville", "lang"=>"en", "lat"=>40.018983, "lon"=>-79.891420),
"/US/US-PA/bruin-17820" => array("type"=>"borough", "pop"=>492, "name_en"=>"Bruin", "lang"=>"en", "lat"=>41.055099, "lon"=>-79.727327),
"/US/US-PA/bryn-athyn-17821" => array("type"=>"borough", "pop"=>1327, "name_en"=>"Bryn Athyn", "lang"=>"en", "lat"=>40.138450, "lon"=>-75.065990),
"/US/US-PA/bryn-mawr-17822" => array("type"=>"cdp", "pop"=>3271, "name_en"=>"Bryn Mawr", "lang"=>"en", "lat"=>40.021833, "lon"=>-75.315854),
"/US/US-PA/burgettstown-17823" => array("type"=>"borough", "pop"=>1478, "name_en"=>"Burgettstown", "lang"=>"en", "lat"=>40.381190, "lon"=>-80.392242),
"/US/US-PA/burlington-17824" => array("type"=>"borough", "pop"=>172, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>41.782832, "lon"=>-76.608879),
"/US/US-PA/burnham-17825" => array("type"=>"borough", "pop"=>2039, "name_ru"=>"Бернхэм", "name_en"=>"Burnham", "lang"=>"en", "lat"=>40.636554, "lon"=>-77.563236),
"/US/US-PA/burnside-17826" => array("type"=>"borough", "pop"=>276, "name_ru"=>"Бернсайд", "name_en"=>"Burnside", "lang"=>"en", "lat"=>40.812846, "lon"=>-78.789800),
"/US/US-PA/butler-17827" => array("type"=>"city", "pop"=>13933, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>40.861872, "lon"=>-79.896744),
"/US/US-PA/california-17828" => array("type"=>"borough", "pop"=>6252, "name_ru"=>"Калифорния", "name_en"=>"California", "lang"=>"en", "lat"=>40.066248, "lon"=>-79.910783),
"/US/US-PA/callensburg-17829" => array("type"=>"borough", "pop"=>202, "name_en"=>"Callensburg", "lang"=>"en", "lat"=>41.125648, "lon"=>-79.558525),
"/US/US-PA/callery-17830" => array("type"=>"borough", "pop"=>419, "name_en"=>"Callery", "lang"=>"en", "lat"=>40.739456, "lon"=>-80.037617),
"/US/US-PA/callimont-17831" => array("type"=>"borough", "pop"=>43, "name_en"=>"Callimont", "lang"=>"en", "lat"=>39.796585, "lon"=>-78.920466),
"/US/US-PA/calumet-norvelt-17832" => array("type"=>"cdp", "pop"=>1790, "name_en"=>"Calumet-Norvelt", "lang"=>"en", "lat"=>40.207318, "lon"=>-79.493128),
"/US/US-PA/cambridge-springs-17833" => array("type"=>"borough", "pop"=>2632, "name_en"=>"Cambridge Springs", "lang"=>"en", "lat"=>41.801667, "lon"=>-80.060130),
"/US/US-PA/camp-hill-17835" => array("type"=>"borough", "pop"=>7378, "name_ru"=>"Кэмп Хилл", "name_en"=>"Camp Hill", "lang"=>"en", "lat"=>40.242500, "lon"=>-76.926781),
"/US/US-PA/campbelltown-17834" => array("type"=>"cdp", "pop"=>1609, "name_en"=>"Campbelltown", "lang"=>"en", "lat"=>40.276380, "lon"=>-76.584485),
"/US/US-PA/canonsburg-17836" => array("type"=>"borough", "pop"=>8717, "name_en"=>"Canonsburg", "lang"=>"en", "lat"=>40.263829, "lon"=>-80.186814),
"/US/US-PA/canton-17837" => array("type"=>"borough", "pop"=>1695, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>41.656986, "lon"=>-76.853026),
"/US/US-PA/carbondale-17838" => array("type"=>"city", "pop"=>9208, "name_ru"=>"Карбондейл", "name_en"=>"Carbondale", "lang"=>"en", "lat"=>41.571100, "lon"=>-75.505501),
"/US/US-PA/carlisle-17839" => array("type"=>"borough", "pop"=>18379, "name_ru"=>"Карлайл", "name_en"=>"Carlisle", "lang"=>"en", "lat"=>40.200942, "lon"=>-77.204185),
"/US/US-PA/carmichaels-17840" => array("type"=>"borough", "pop"=>521, "name_en"=>"Carmichaels", "lang"=>"en", "lat"=>39.897680, "lon"=>-79.975246),
"/US/US-PA/carnegie-17841" => array("type"=>"borough", "pop"=>7920, "name_en"=>"Carnegie", "lang"=>"en", "lat"=>40.407600, "lon"=>-80.086649),
"/US/US-PA/carnot-moon-17842" => array("type"=>"cdp", "pop"=>10187, "name_en"=>"Carnot-Moon", "lang"=>"en", "lat"=>40.518567, "lon"=>-80.214162),
"/US/US-PA/carroll-valley-17844" => array("type"=>"borough", "pop"=>3532, "name_en"=>"Carroll Valley", "lang"=>"en", "lat"=>39.750623, "lon"=>-77.381065),
"/US/US-PA/carrolltown-17843" => array("type"=>"borough", "pop"=>959, "name_en"=>"Carrolltown", "lang"=>"en", "lat"=>40.604247, "lon"=>-78.707116),
"/US/US-PA/cassandra-17845" => array("type"=>"borough", "pop"=>123, "name_en"=>"Cassandra", "lang"=>"en", "lat"=>40.408070, "lon"=>-78.641601),
"/US/US-PA/casselman-17846" => array("type"=>"borough", "pop"=>90, "name_en"=>"Casselman", "lang"=>"en", "lat"=>39.885437, "lon"=>-79.211012),
"/US/US-PA/cassville-17847" => array("type"=>"borough", "pop"=>142, "name_ru"=>"Кассвилл", "name_en"=>"Cassville", "lang"=>"en", "lat"=>40.293668, "lon"=>-78.027757),
"/US/US-PA/castanea-17848" => array("type"=>"cdp", "pop"=>1176, "name_en"=>"Castanea", "lang"=>"en", "lat"=>41.122892, "lon"=>-77.428718),
"/US/US-PA/castle-shannon-17849" => array("type"=>"borough", "pop"=>8028, "name_en"=>"Castle Shannon", "lang"=>"en", "lat"=>40.366300, "lon"=>-80.020199),
"/US/US-PA/catasauqua-17850" => array("type"=>"borough", "pop"=>6558, "name_en"=>"Catasauqua", "lang"=>"en", "lat"=>40.652133, "lon"=>-75.464704),
"/US/US-PA/catawissa-17851" => array("type"=>"borough", "pop"=>1538, "name_en"=>"Catawissa", "lang"=>"en", "lat"=>40.953037, "lon"=>-76.460424),
"/US/US-PA/cecil-bishop-17852" => array("type"=>"cdp", "pop"=>2701, "name_en"=>"Cecil-Bishop", "lang"=>"en", "lat"=>40.316442, "lon"=>-80.191008),
"/US/US-PA/centerport-17853" => array("type"=>"borough", "pop"=>328, "name_en"=>"Centerport", "lang"=>"en", "lat"=>40.486650, "lon"=>-76.006289),
"/US/US-PA/centerville-17854" => array("type"=>"borough", "pop"=>236, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>41.736318, "lon"=>-79.761191),
"/US/US-PA/centerville-17855" => array("type"=>"borough", "pop"=>3842, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>40.030693, "lon"=>-79.959335),
"/US/US-PA/central-city-17856" => array("type"=>"borough", "pop"=>1160, "name_ru"=>"Сентрал Сити", "name_en"=>"Central City", "lang"=>"en", "lat"=>40.109742, "lon"=>-78.805333),
"/US/US-PA/centralia-17857" => array("type"=>"borough", "pop"=>20, "name_ru"=>"Сентралия", "name_en"=>"Centralia", "lang"=>"en", "lat"=>40.804240, "lon"=>-76.344237),
"/US/US-PA/centre-hall-17858" => array("type"=>"borough", "pop"=>1072, "name_en"=>"Centre Hall", "lang"=>"en", "lat"=>40.844422, "lon"=>-77.684621),
"/US/US-PA/chalfant-17859" => array("type"=>"borough", "pop"=>789, "name_en"=>"Chalfant", "lang"=>"en", "lat"=>40.409900, "lon"=>-79.838515),
"/US/US-PA/chalfont-17860" => array("type"=>"borough", "pop"=>4161, "name_en"=>"Chalfont", "lang"=>"en", "lat"=>40.288895, "lon"=>-75.209665),
"/US/US-PA/chambersburg-17861" => array("type"=>"borough", "pop"=>18302, "name_en"=>"Chambersburg", "lang"=>"en", "lat"=>39.932245, "lon"=>-77.656432),
"/US/US-PA/chapman-17862" => array("type"=>"borough", "pop"=>230, "name_ru"=>"Чэпмен", "name_en"=>"Chapman", "lang"=>"en", "lat"=>40.759176, "lon"=>-75.401479),
"/US/US-PA/charleroi-17863" => array("type"=>"borough", "pop"=>4549, "name_en"=>"Charleroi", "lang"=>"en", "lat"=>40.139152, "lon"=>-79.900476),
"/US/US-PA/cherry-tree-17864" => array("type"=>"borough", "pop"=>414, "name_en"=>"Cherry Tree", "lang"=>"en", "lat"=>40.725677, "lon"=>-78.807827),
"/US/US-PA/cherry-valley-17865" => array("type"=>"borough", "pop"=>73, "name_ru"=>"Черри Валли", "name_en"=>"Cherry Valley", "lang"=>"en", "lat"=>41.161930, "lon"=>-79.799065),
"/US/US-PA/chest-springs-17871" => array("type"=>"borough", "pop"=>101, "name_en"=>"Chest Springs", "lang"=>"en", "lat"=>40.578736, "lon"=>-78.610065),
"/US/US-PA/chester-17866" => array("type"=>"city", "pop"=>36564, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>39.845450, "lon"=>-75.372202),
"/US/US-PA/chester-heights-17868" => array("type"=>"borough", "pop"=>2461, "name_en"=>"Chester Heights", "lang"=>"en", "lat"=>39.894383, "lon"=>-75.467669),
"/US/US-PA/chester-hill-17869" => array("type"=>"borough", "pop"=>853, "name_en"=>"Chester Hill", "lang"=>"en", "lat"=>40.889834, "lon"=>-78.229207),
"/US/US-PA/chester-township-17870" => array("type"=>"cdp", "pop"=>5399, "name_en"=>"Chester Township", "lang"=>"en", "lat"=>39.850015, "lon"=>-75.397379),
"/US/US-PA/chesterbrook-17867" => array("type"=>"cdp", "pop"=>4561, "name_en"=>"Chesterbrook", "lang"=>"en", "lat"=>40.075681, "lon"=>-75.459309),
"/US/US-PA/cheswick-17872" => array("type"=>"borough", "pop"=>1741, "name_en"=>"Cheswick", "lang"=>"en", "lat"=>40.542864, "lon"=>-79.801324),
"/US/US-PA/chevy-chase-heights-17873" => array("type"=>"cdp", "pop"=>1535, "name_en"=>"Chevy Chase Heights", "lang"=>"en", "lat"=>40.638972, "lon"=>-79.143705),
"/US/US-PA/chicora-17874" => array("type"=>"borough", "pop"=>969, "name_en"=>"Chicora", "lang"=>"en", "lat"=>40.948899, "lon"=>-79.741878),
"/US/US-PA/christiana-17875" => array("type"=>"borough", "pop"=>1110, "name_en"=>"Christiana", "lang"=>"en", "lat"=>39.954080, "lon"=>-75.997651),
"/US/US-PA/churchill-17876" => array("type"=>"borough", "pop"=>3254, "name_en"=>"Churchill", "lang"=>"en", "lat"=>40.438300, "lon"=>-79.840750),
"/US/US-PA/churchville-17877" => array("type"=>"cdp", "pop"=>4255, "name_en"=>"Churchville", "lang"=>"en", "lat"=>40.200732, "lon"=>-74.997436),
"/US/US-PA/clairton-17878" => array("type"=>"city", "pop"=>7832, "name_en"=>"Clairton", "lang"=>"en", "lat"=>40.296450, "lon"=>-79.885474),
"/US/US-PA/clarendon-17879" => array("type"=>"borough", "pop"=>510, "name_ru"=>"Кларендон", "name_en"=>"Clarendon", "lang"=>"en", "lat"=>41.780710, "lon"=>-79.096395),
"/US/US-PA/clarion-17880" => array("type"=>"borough", "pop"=>5258, "name_ru"=>"Клэрион", "name_en"=>"Clarion", "lang"=>"en", "lat"=>41.210149, "lon"=>-79.380448),
"/US/US-PA/clark-17881" => array("type"=>"borough", "pop"=>620, "name_ru"=>"Кларк", "name_en"=>"Clark", "lang"=>"en", "lat"=>41.285294, "lon"=>-80.404164),
"/US/US-PA/clarks-green-17882" => array("type"=>"borough", "pop"=>1551, "name_en"=>"Clarks Green", "lang"=>"en", "lat"=>41.499667, "lon"=>-75.695619),
"/US/US-PA/clarks-summit-17883" => array("type"=>"borough", "pop"=>4964, "name_en"=>"Clarks Summit", "lang"=>"en", "lat"=>41.490167, "lon"=>-75.707457),
"/US/US-PA/clarksville-17884" => array("type"=>"borough", "pop"=>218, "name_ru"=>"Кларксвилл", "name_en"=>"Clarksville", "lang"=>"en", "lat"=>39.974448, "lon"=>-80.044285),
"/US/US-PA/claysburg-17885" => array("type"=>"cdp", "pop"=>1399, "name_en"=>"Claysburg", "lang"=>"en", "lat"=>40.292127, "lon"=>-78.449612),
"/US/US-PA/claysville-17886" => array("type"=>"borough", "pop"=>675, "name_en"=>"Claysville", "lang"=>"en", "lat"=>40.120606, "lon"=>-80.413176),
"/US/US-PA/clearfield-17887" => array("type"=>"borough", "pop"=>6173, "name_ru"=>"Клирфилд", "name_en"=>"Clearfield", "lang"=>"en", "lat"=>41.021378, "lon"=>-78.439184),
"/US/US-PA/cleona-17888" => array("type"=>"borough", "pop"=>2121, "name_en"=>"Cleona", "lang"=>"en", "lat"=>40.338798, "lon"=>-76.477308),
"/US/US-PA/clifton-heights-17889" => array("type"=>"borough", "pop"=>6541, "name_en"=>"Clifton Heights", "lang"=>"en", "lat"=>39.928950, "lon"=>-75.296130),
"/US/US-PA/clintonville-17890" => array("type"=>"borough", "pop"=>488, "name_en"=>"Clintonville", "lang"=>"en", "lat"=>41.201799, "lon"=>-79.876131),
"/US/US-PA/clymer-17891" => array("type"=>"borough", "pop"=>1430, "name_en"=>"Clymer", "lang"=>"en", "lat"=>40.668289, "lon"=>-79.014021),
"/US/US-PA/coal-center-17892" => array("type"=>"borough", "pop"=>126, "name_en"=>"Coal Center", "lang"=>"en", "lat"=>40.069861, "lon"=>-79.901248),
"/US/US-PA/coaldale-17893" => array("type"=>"borough", "pop"=>138, "name_en"=>"Coaldale", "lang"=>"en", "lat"=>40.167967, "lon"=>-78.217102),
"/US/US-PA/coaldale-17894" => array("type"=>"borough", "pop"=>2531, "name_en"=>"Coaldale", "lang"=>"en", "lat"=>40.820803, "lon"=>-75.916221),
"/US/US-PA/coalmont-17895" => array("type"=>"borough", "pop"=>119, "name_en"=>"Coalmont", "lang"=>"en", "lat"=>40.211016, "lon"=>-78.200670),
"/US/US-PA/coalport-17896" => array("type"=>"borough", "pop"=>453, "name_en"=>"Coalport", "lang"=>"en", "lat"=>40.748461, "lon"=>-78.532127),
"/US/US-PA/coatesville-17897" => array("type"=>"city", "pop"=>11624, "name_ru"=>"Коатсвилл", "name_en"=>"Coatesville", "lang"=>"en", "lat"=>39.984800, "lon"=>-75.819289),
"/US/US-PA/cochranton-17898" => array("type"=>"borough", "pop"=>1065, "name_en"=>"Cochranton", "lang"=>"en", "lat"=>41.518545, "lon"=>-80.046939),
"/US/US-PA/cokeburg-17899" => array("type"=>"borough", "pop"=>660, "name_en"=>"Cokeburg", "lang"=>"en", "lat"=>40.098384, "lon"=>-80.066279),
"/US/US-PA/collegeville-17900" => array("type"=>"borough", "pop"=>5058, "name_ru"=>"Коллиджвилл", "name_en"=>"Collegeville", "lang"=>"en", "lat"=>40.187100, "lon"=>-75.457399),
"/US/US-PA/collingdale-17901" => array("type"=>"borough", "pop"=>8365, "name_en"=>"Collingdale", "lang"=>"en", "lat"=>39.914500, "lon"=>-75.277413),
"/US/US-PA/colonial-park-17902" => array("type"=>"cdp", "pop"=>13777, "name_en"=>"Colonial Park", "lang"=>"en", "lat"=>40.299883, "lon"=>-76.806400),
"/US/US-PA/columbia-17903" => array("type"=>"borough", "pop"=>10033, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>40.033383, "lon"=>-76.495049),
"/US/US-PA/colver-17904" => array("type"=>"cdp", "pop"=>1024, "name_en"=>"Colver", "lang"=>"en", "lat"=>40.542912, "lon"=>-78.787559),
"/US/US-PA/colwyn-17905" => array("type"=>"borough", "pop"=>2394, "name_en"=>"Colwyn", "lang"=>"en", "lat"=>39.911450, "lon"=>-75.253299),
"/US/US-PA/confluence-17906" => array("type"=>"borough", "pop"=>777, "name_en"=>"Confluence", "lang"=>"en", "lat"=>39.809481, "lon"=>-79.354340),
"/US/US-PA/conneaut-lake-17907" => array("type"=>"borough", "pop"=>660, "name_en"=>"Conneaut Lake", "lang"=>"en", "lat"=>41.602314, "lon"=>-80.309224),
"/US/US-PA/conneaut-lakeshore-17908" => array("type"=>"cdp", "pop"=>1852, "name_en"=>"Conneaut Lakeshore", "lang"=>"en", "lat"=>41.624110, "lon"=>-80.307157),
"/US/US-PA/conneautville-17909" => array("type"=>"borough", "pop"=>788, "name_en"=>"Conneautville", "lang"=>"en", "lat"=>41.757662, "lon"=>-80.368089),
"/US/US-PA/connellsville-17910" => array("type"=>"city", "pop"=>8468, "name_en"=>"Connellsville", "lang"=>"en", "lat"=>40.016283, "lon"=>-79.589096),
"/US/US-PA/connoquenessing-17911" => array("type"=>"borough", "pop"=>546, "name_en"=>"Connoquenessing", "lang"=>"en", "lat"=>40.818032, "lon"=>-80.013927),
"/US/US-PA/conshohocken-17912" => array("type"=>"borough", "pop"=>8409, "name_en"=>"Conshohocken", "lang"=>"en", "lat"=>40.077521, "lon"=>-75.303198),
"/US/US-PA/conway-17913" => array("type"=>"borough", "pop"=>2132, "name_ru"=>"Конвэй", "name_en"=>"Conway", "lang"=>"en", "lat"=>40.664500, "lon"=>-80.240904),
"/US/US-PA/conyngham-17914" => array("type"=>"borough", "pop"=>1835, "name_en"=>"Conyngham", "lang"=>"en", "lat"=>40.991505, "lon"=>-76.060225),
"/US/US-PA/coopersburg-17915" => array("type"=>"borough", "pop"=>2562, "name_en"=>"Coopersburg", "lang"=>"en", "lat"=>40.510368, "lon"=>-75.391382),
"/US/US-PA/cooperstown-17916" => array("type"=>"borough", "pop"=>421, "name_en"=>"Cooperstown", "lang"=>"en", "lat"=>41.499585, "lon"=>-79.874044),
"/US/US-PA/coplay-17917" => array("type"=>"borough", "pop"=>3367, "name_en"=>"Coplay", "lang"=>"en", "lat"=>40.671100, "lon"=>-75.496502),
"/US/US-PA/coraopolis-17918" => array("type"=>"borough", "pop"=>5606, "name_en"=>"Coraopolis", "lang"=>"en", "lat"=>40.514100, "lon"=>-80.162100),
"/US/US-PA/cornwall-17919" => array("type"=>"borough", "pop"=>3499, "name_en"=>"Cornwall", "lang"=>"en", "lat"=>40.269514, "lon"=>-76.408112),
"/US/US-PA/cornwells-heights-eddington-17920" => array("type"=>"cdp", "pop"=>3621, "name_en"=>"Cornwells Heights-Eddington", "lang"=>"en", "lat"=>40.081700, "lon"=>-74.947066),
"/US/US-PA/corry-17921" => array("type"=>"city", "pop"=>6307, "name_en"=>"Corry", "lang"=>"en", "lat"=>41.925897, "lon"=>-79.637337),
"/US/US-PA/corsica-17922" => array("type"=>"borough", "pop"=>333, "name_en"=>"Corsica", "lang"=>"en", "lat"=>41.181668, "lon"=>-79.202943),
"/US/US-PA/coudersport-17923" => array("type"=>"borough", "pop"=>2375, "name_en"=>"Coudersport", "lang"=>"en", "lat"=>41.774201, "lon"=>-78.014344),
"/US/US-PA/courtdale-17924" => array("type"=>"borough", "pop"=>741, "name_en"=>"Courtdale", "lang"=>"en", "lat"=>41.285455, "lon"=>-75.915246),
"/US/US-PA/crafton-17925" => array("type"=>"borough", "pop"=>6109, "name_en"=>"Crafton", "lang"=>"en", "lat"=>40.433250, "lon"=>-80.070934),
"/US/US-PA/cranesville-17926" => array("type"=>"borough", "pop"=>574, "name_en"=>"Cranesville", "lang"=>"en", "lat"=>41.903781, "lon"=>-80.344732),
"/US/US-PA/creekside-17927" => array("type"=>"borough", "pop"=>298, "name_ru"=>"Криксайд", "name_en"=>"Creekside", "lang"=>"en", "lat"=>40.679951, "lon"=>-79.193505),
"/US/US-PA/cresson-17928" => array("type"=>"borough", "pop"=>1481, "name_en"=>"Cresson", "lang"=>"en", "lat"=>40.462560, "lon"=>-78.586580),
"/US/US-PA/cressona-17929" => array("type"=>"borough", "pop"=>1554, "name_en"=>"Cressona", "lang"=>"en", "lat"=>40.630467, "lon"=>-76.194819),
"/US/US-PA/cross-roads-17930" => array("type"=>"borough", "pop"=>550, "name_en"=>"Cross Roads", "lang"=>"en", "lat"=>39.820469, "lon"=>-76.570304),
"/US/US-PA/croydon-17931" => array("type"=>"cdp", "pop"=>9967, "name_en"=>"Croydon", "lang"=>"en", "lat"=>40.090250, "lon"=>-74.895528),
"/US/US-PA/curtisville-17932" => array("type"=>"cdp", "pop"=>1285, "name_en"=>"Curtisville", "lang"=>"en", "lat"=>40.648720, "lon"=>-79.849569),
"/US/US-PA/curwensville-17933" => array("type"=>"borough", "pop"=>2463, "name_en"=>"Curwensville", "lang"=>"en", "lat"=>40.972715, "lon"=>-78.517998),
"/US/US-PA/daisytown-17934" => array("type"=>"borough", "pop"=>337, "name_en"=>"Daisytown", "lang"=>"en", "lat"=>40.320350, "lon"=>-78.903480),
"/US/US-PA/dale-17935" => array("type"=>"borough", "pop"=>1358, "name_ru"=>"Дейл", "name_en"=>"Dale", "lang"=>"en", "lat"=>40.311850, "lon"=>-78.904092),
"/US/US-PA/dallas-17936" => array("type"=>"borough", "pop"=>2475, "name_ru"=>"Даллас", "name_en"=>"Dallas", "lang"=>"en", "lat"=>41.331800, "lon"=>-75.972292),
"/US/US-PA/dallastown-17937" => array("type"=>"borough", "pop"=>4062, "name_en"=>"Dallastown", "lang"=>"en", "lat"=>39.899750, "lon"=>-76.641074),
"/US/US-PA/dalton-17938" => array("type"=>"borough", "pop"=>1224, "name_ru"=>"Далтон", "name_en"=>"Dalton", "lang"=>"en", "lat"=>41.537856, "lon"=>-75.738631),
"/US/US-PA/danville-17939" => array("type"=>"borough", "pop"=>4450, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>40.961482, "lon"=>-76.612311),
"/US/US-PA/darby-17940" => array("type"=>"borough", "pop"=>9903, "name_en"=>"Darby", "lang"=>"en", "lat"=>39.920450, "lon"=>-75.261029),
"/US/US-PA/darby-township-17941" => array("type"=>"cdp", "pop"=>10955, "name_en"=>"Darby Township", "lang"=>"en", "lat"=>39.899700, "lon"=>-75.273231),
"/US/US-PA/darlington-17942" => array("type"=>"borough", "pop"=>270, "name_ru"=>"Дарлингтон", "name_en"=>"Darlington", "lang"=>"en", "lat"=>40.810037, "lon"=>-80.423942),
"/US/US-PA/dauphin-17943" => array("type"=>"borough", "pop"=>719, "name_en"=>"Dauphin", "lang"=>"en", "lat"=>40.368767, "lon"=>-76.930394),
"/US/US-PA/davidsville-17944" => array("type"=>"cdp", "pop"=>1167, "name_en"=>"Davidsville", "lang"=>"en", "lat"=>40.231864, "lon"=>-78.936928),
"/US/US-PA/dawson-17945" => array("type"=>"borough", "pop"=>422, "name_ru"=>"Доусон", "name_en"=>"Dawson", "lang"=>"en", "lat"=>40.047818, "lon"=>-79.658898),
"/US/US-PA/dayton-17946" => array("type"=>"borough", "pop"=>498, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>40.881030, "lon"=>-79.241766),
"/US/US-PA/deemston-17947" => array("type"=>"borough", "pop"=>776, "name_en"=>"Deemston", "lang"=>"en", "lat"=>40.032096, "lon"=>-80.026728),
"/US/US-PA/deer-lake-17948" => array("type"=>"borough", "pop"=>622, "name_en"=>"Deer Lake", "lang"=>"en", "lat"=>40.620641, "lon"=>-76.056627),
"/US/US-PA/delaware-water-gap-17949" => array("type"=>"borough", "pop"=>791, "name_en"=>"Delaware Water Gap", "lang"=>"en", "lat"=>40.978611, "lon"=>-75.137590),
"/US/US-PA/delmont-17950" => array("type"=>"borough", "pop"=>2418, "name_en"=>"Delmont", "lang"=>"en", "lat"=>40.413944, "lon"=>-79.573388),
"/US/US-PA/delta-17951" => array("type"=>"borough", "pop"=>719, "name_ru"=>"Дельта", "name_en"=>"Delta", "lang"=>"en", "lat"=>39.726050, "lon"=>-76.327738),
"/US/US-PA/denver-17952" => array("type"=>"borough", "pop"=>3650, "name_ru"=>"Денвер", "name_en"=>"Denver", "lang"=>"en", "lat"=>40.232500, "lon"=>-76.138731),
"/US/US-PA/derry-17953" => array("type"=>"borough", "pop"=>2775, "name_en"=>"Derry", "lang"=>"en", "lat"=>40.333317, "lon"=>-79.301731),
"/US/US-PA/devon-berwyn-17954" => array("type"=>"cdp", "pop"=>5019, "name_en"=>"Devon-Berwyn", "lang"=>"en", "lat"=>40.043450, "lon"=>-75.438658),
"/US/US-PA/dickson-city-17955" => array("type"=>"borough", "pop"=>5903, "name_en"=>"Dickson City", "lang"=>"en", "lat"=>41.468767, "lon"=>-75.634775),
"/US/US-PA/dillsburg-17956" => array("type"=>"borough", "pop"=>2377, "name_en"=>"Dillsburg", "lang"=>"en", "lat"=>40.110143, "lon"=>-77.034808),
"/US/US-PA/donegal-17957" => array("type"=>"borough", "pop"=>155, "name_en"=>"Donegal", "lang"=>"en", "lat"=>40.112838, "lon"=>-79.382532),
"/US/US-PA/donora-17958" => array("type"=>"borough", "pop"=>5261, "name_en"=>"Donora", "lang"=>"en", "lat"=>40.178272, "lon"=>-79.863402),
"/US/US-PA/dormont-17959" => array("type"=>"borough", "pop"=>8420, "name_en"=>"Dormont", "lang"=>"en", "lat"=>40.393033, "lon"=>-80.037547),
"/US/US-PA/dover-17960" => array("type"=>"borough", "pop"=>1923, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>40.003750, "lon"=>-76.849704),
"/US/US-PA/downingtown-17961" => array("type"=>"borough", "pop"=>7937, "name_en"=>"Downingtown", "lang"=>"en", "lat"=>40.007150, "lon"=>-75.702449),
"/US/US-PA/doylestown-17962" => array("type"=>"borough", "pop"=>8113, "name_en"=>"Doylestown", "lang"=>"en", "lat"=>40.314000, "lon"=>-75.127863),
"/US/US-PA/dravosburg-17963" => array("type"=>"borough", "pop"=>1828, "name_en"=>"Dravosburg", "lang"=>"en", "lat"=>40.352100, "lon"=>-79.888654),
"/US/US-PA/drexel-hill-17964" => array("type"=>"cdp", "pop"=>29744, "name_en"=>"Drexel Hill", "lang"=>"en", "lat"=>39.949233, "lon"=>-75.303850),
"/US/US-PA/driftwood-17965" => array("type"=>"borough", "pop"=>90, "name_en"=>"Driftwood", "lang"=>"en", "lat"=>41.343227, "lon"=>-78.135387),
"/US/US-PA/dublin-17966" => array("type"=>"borough", "pop"=>2141, "name_ru"=>"Дублин", "name_en"=>"Dublin", "lang"=>"en", "lat"=>40.370582, "lon"=>-75.204685),
"/US/US-PA/dubois-17967" => array("type"=>"city", "pop"=>7648, "name_ru"=>"Дюбуа", "name_en"=>"DuBois", "lang"=>"en", "lat"=>41.122304, "lon"=>-78.755786),
"/US/US-PA/duboistown-17968" => array("type"=>"borough", "pop"=>1194, "name_en"=>"Duboistown", "lang"=>"en", "lat"=>41.223050, "lon"=>-77.038325),
"/US/US-PA/dudley-17969" => array("type"=>"borough", "pop"=>202, "name_ru"=>"Дадли", "name_en"=>"Dudley", "lang"=>"en", "lat"=>40.201500, "lon"=>-78.176518),
"/US/US-PA/dunbar-17970" => array("type"=>"borough", "pop"=>1149, "name_en"=>"Dunbar", "lang"=>"en", "lat"=>39.979572, "lon"=>-79.614148),
"/US/US-PA/duncannon-17971" => array("type"=>"borough", "pop"=>1493, "name_en"=>"Duncannon", "lang"=>"en", "lat"=>40.395633, "lon"=>-77.027080),
"/US/US-PA/duncansville-17972" => array("type"=>"borough", "pop"=>1167, "name_en"=>"Duncansville", "lang"=>"en", "lat"=>40.426327, "lon"=>-78.429596),
"/US/US-PA/dunlevy-17973" => array("type"=>"borough", "pop"=>393, "name_en"=>"Dunlevy", "lang"=>"en", "lat"=>40.112408, "lon"=>-79.858675),
"/US/US-PA/dunmore-17974" => array("type"=>"borough", "pop"=>13978, "name_en"=>"Dunmore", "lang"=>"en", "lat"=>41.415600, "lon"=>-75.607086),
"/US/US-PA/dunnstown-17975" => array("type"=>"cdp", "pop"=>1486, "name_en"=>"Dunnstown", "lang"=>"en", "lat"=>41.146855, "lon"=>-77.421146),
"/US/US-PA/dupont-17976" => array("type"=>"borough", "pop"=>2577, "name_ru"=>"Дюпон", "name_en"=>"Dupont", "lang"=>"en", "lat"=>41.324020, "lon"=>-75.742426),
"/US/US-PA/duquesne-17977" => array("type"=>"city", "pop"=>6682, "name_ru"=>"Даквесн", "name_en"=>"Duquesne", "lang"=>"en", "lat"=>40.373200, "lon"=>-79.850785),
"/US/US-PA/duryea-17978" => array("type"=>"borough", "pop"=>4335, "name_en"=>"Duryea", "lang"=>"en", "lat"=>41.351800, "lon"=>-75.776290),
"/US/US-PA/dushore-17979" => array("type"=>"borough", "pop"=>575, "name_en"=>"Dushore", "lang"=>"en", "lat"=>41.525289, "lon"=>-76.399092),
"/US/US-PA/eagles-mere-17980" => array("type"=>"borough", "pop"=>142, "name_en"=>"Eagles Mere", "lang"=>"en", "lat"=>41.409722, "lon"=>-76.583323),
"/US/US-PA/eagleville-17981" => array("type"=>"cdp", "pop"=>3637, "name_ru"=>"Иглвилл", "name_en"=>"Eagleville", "lang"=>"en", "lat"=>40.157864, "lon"=>-75.407753),
"/US/US-PA/east-bangor-17982" => array("type"=>"borough", "pop"=>1113, "name_en"=>"East Bangor", "lang"=>"en", "lat"=>40.880565, "lon"=>-75.186621),
"/US/US-PA/east-berlin-17983" => array("type"=>"borough", "pop"=>1425, "name_en"=>"East Berlin", "lang"=>"en", "lat"=>39.936822, "lon"=>-76.980376),
"/US/US-PA/east-berwick-17984" => array("type"=>"cdp", "pop"=>2128, "name_en"=>"East Berwick", "lang"=>"en", "lat"=>41.064674, "lon"=>-76.220696),
"/US/US-PA/east-brady-17985" => array("type"=>"borough", "pop"=>993, "name_en"=>"East Brady", "lang"=>"en", "lat"=>40.984760, "lon"=>-79.614068),
"/US/US-PA/east-butler-17986" => array("type"=>"borough", "pop"=>635, "name_en"=>"East Butler", "lang"=>"en", "lat"=>40.878551, "lon"=>-79.847564),
"/US/US-PA/east-conemaugh-17987" => array("type"=>"borough", "pop"=>1163, "name_en"=>"East Conemaugh", "lang"=>"en", "lat"=>40.347450, "lon"=>-78.886053),
"/US/US-PA/east-greenville-17988" => array("type"=>"borough", "pop"=>3017, "name_en"=>"East Greenville", "lang"=>"en", "lat"=>40.406039, "lon"=>-75.506522),
"/US/US-PA/east-lansdowne-17989" => array("type"=>"borough", "pop"=>2473, "name_en"=>"East Lansdowne", "lang"=>"en", "lat"=>39.943850, "lon"=>-75.261670),
"/US/US-PA/east-mckeesport-17991" => array("type"=>"borough", "pop"=>2135, "name_en"=>"East McKeesport", "lang"=>"en", "lat"=>40.384250, "lon"=>-79.806848),
"/US/US-PA/east-norriton-17992" => array("type"=>"cdp", "pop"=>13587, "name_en"=>"East Norriton", "lang"=>"en", "lat"=>40.151800, "lon"=>-75.336825),
"/US/US-PA/east-petersburg-17994" => array("type"=>"borough", "pop"=>4347, "name_en"=>"East Petersburg", "lang"=>"en", "lat"=>40.100750, "lon"=>-76.353495),
"/US/US-PA/east-pittsburgh-17995" => array("type"=>"borough", "pop"=>1839, "name_en"=>"East Pittsburgh", "lang"=>"en", "lat"=>40.397167, "lon"=>-79.836885),
"/US/US-PA/east-prospect-17996" => array("type"=>"borough", "pop"=>758, "name_en"=>"East Prospect", "lang"=>"en", "lat"=>39.971408, "lon"=>-76.520914),
"/US/US-PA/east-rochester-17997" => array("type"=>"borough", "pop"=>564, "name_en"=>"East Rochester", "lang"=>"en", "lat"=>40.698100, "lon"=>-80.268249),
"/US/US-PA/east-side-17998" => array("type"=>"borough", "pop"=>286, "name_en"=>"East Side", "lang"=>"en", "lat"=>41.062720, "lon"=>-75.762003),
"/US/US-PA/east-stroudsburg-17999" => array("type"=>"borough", "pop"=>10390, "name_en"=>"East Stroudsburg", "lang"=>"en", "lat"=>41.001539, "lon"=>-75.178397),
"/US/US-PA/east-uniontown-18000" => array("type"=>"cdp", "pop"=>2822, "name_en"=>"East Uniontown", "lang"=>"en", "lat"=>39.899693, "lon"=>-79.698111),
"/US/US-PA/east-vandergrift-18002" => array("type"=>"borough", "pop"=>679, "name_en"=>"East Vandergrift", "lang"=>"en", "lat"=>40.597424, "lon"=>-79.562976),
"/US/US-PA/east-washington-18003" => array("type"=>"borough", "pop"=>1861, "name_en"=>"East Washington", "lang"=>"en", "lat"=>40.173656, "lon"=>-80.232862),
"/US/US-PA/east-york-18004" => array("type"=>"cdp", "pop"=>8487, "name_en"=>"East York", "lang"=>"en", "lat"=>39.968650, "lon"=>-76.674972),
"/US/US-PA/eastlawn-gardens-17990" => array("type"=>"cdp", "pop"=>1794, "name_en"=>"Eastlawn Gardens", "lang"=>"en", "lat"=>40.750653, "lon"=>-75.296125),
"/US/US-PA/easton-17993" => array("type"=>"city", "pop"=>26080, "name_ru"=>"Истон", "name_en"=>"Easton", "lang"=>"en", "lat"=>40.686300, "lon"=>-75.220578),
"/US/US-PA/eastvale-18001" => array("type"=>"borough", "pop"=>266, "name_en"=>"Eastvale", "lang"=>"en", "lat"=>40.767700, "lon"=>-80.315499),
"/US/US-PA/eau-claire-18005" => array("type"=>"borough", "pop"=>331, "name_ru"=>"О Клэр", "name_en"=>"Eau Claire", "lang"=>"en", "lat"=>41.136860, "lon"=>-79.798182),
"/US/US-PA/ebensburg-18006" => array("type"=>"borough", "pop"=>2924, "name_en"=>"Ebensburg", "lang"=>"en", "lat"=>40.488626, "lon"=>-78.726987),
"/US/US-PA/economy-18007" => array("type"=>"borough", "pop"=>9136, "name_ru"=>"Экономи", "name_en"=>"Economy", "lang"=>"en", "lat"=>40.638400, "lon"=>-80.185109),
"/US/US-PA/eddystone-18008" => array("type"=>"borough", "pop"=>2341, "name_en"=>"Eddystone", "lang"=>"en", "lat"=>39.856100, "lon"=>-75.332821),
"/US/US-PA/edgewood-18009" => array("type"=>"borough", "pop"=>3017, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>40.431390, "lon"=>-79.883659),
"/US/US-PA/edgewood-18010" => array("type"=>"cdp", "pop"=>2719, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>40.788003, "lon"=>-76.578740),
"/US/US-PA/edgeworth-18011" => array("type"=>"borough", "pop"=>1589, "name_en"=>"Edgeworth", "lang"=>"en", "lat"=>40.554333, "lon"=>-80.192199),
"/US/US-PA/edinboro-18012" => array("type"=>"borough", "pop"=>6655, "name_en"=>"Edinboro", "lang"=>"en", "lat"=>41.874662, "lon"=>-80.124663),
"/US/US-PA/edwardsville-18013" => array("type"=>"borough", "pop"=>4658, "name_ru"=>"Эдвардсвилл", "name_en"=>"Edwardsville", "lang"=>"en", "lat"=>41.260418, "lon"=>-75.907012),
"/US/US-PA/ehrenfeld-18014" => array("type"=>"borough", "pop"=>213, "name_en"=>"Ehrenfeld", "lang"=>"en", "lat"=>40.372718, "lon"=>-78.776205),
"/US/US-PA/elco-18015" => array("type"=>"borough", "pop"=>339, "name_en"=>"Elco", "lang"=>"en", "lat"=>40.085029, "lon"=>-79.883145),
"/US/US-PA/elderton-18016" => array("type"=>"borough", "pop"=>336, "name_en"=>"Elderton", "lang"=>"en", "lat"=>40.694064, "lon"=>-79.341909),
"/US/US-PA/eldred-18017" => array("type"=>"borough", "pop"=>795, "name_ru"=>"Эльдред", "name_en"=>"Eldred", "lang"=>"en", "lat"=>41.956675, "lon"=>-78.381007),
"/US/US-PA/elgin-18018" => array("type"=>"borough", "pop"=>225, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>41.910831, "lon"=>-79.743434),
"/US/US-PA/elim-18019" => array("type"=>"cdp", "pop"=>3861, "name_ru"=>"Элим", "name_en"=>"Elim", "lang"=>"en", "lat"=>40.297800, "lon"=>-78.942869),
"/US/US-PA/elizabeth-18020" => array("type"=>"borough", "pop"=>1460, "name_ru"=>"Элизабет", "name_en"=>"Elizabeth", "lang"=>"en", "lat"=>40.271520, "lon"=>-79.886464),
"/US/US-PA/elizabethtown-18021" => array("type"=>"borough", "pop"=>12069, "name_ru"=>"Элизабеттаун", "name_en"=>"Elizabethtown", "lang"=>"en", "lat"=>40.153550, "lon"=>-76.598625),
"/US/US-PA/elizabethville-18022" => array("type"=>"borough", "pop"=>1281, "name_en"=>"Elizabethville", "lang"=>"en", "lat"=>40.547474, "lon"=>-76.816002),
"/US/US-PA/elkland-18023" => array("type"=>"borough", "pop"=>1663, "name_en"=>"Elkland", "lang"=>"en", "lat"=>41.989235, "lon"=>-77.314818),
"/US/US-PA/ellport-18024" => array("type"=>"borough", "pop"=>1096, "name_en"=>"Ellport", "lang"=>"en", "lat"=>40.861368, "lon"=>-80.262295),
"/US/US-PA/ellsworth-18025" => array("type"=>"borough", "pop"=>1007, "name_ru"=>"Эллсворт", "name_en"=>"Ellsworth", "lang"=>"en", "lat"=>40.108027, "lon"=>-80.022064),
"/US/US-PA/ellwood-city-18026" => array("type"=>"borough", "pop"=>7966, "name_en"=>"Ellwood City", "lang"=>"en", "lat"=>40.860928, "lon"=>-80.283535),
"/US/US-PA/elverson-18027" => array("type"=>"borough", "pop"=>1198, "name_en"=>"Elverson", "lang"=>"en", "lat"=>40.157793, "lon"=>-75.831504),
"/US/US-PA/elysburg-18028" => array("type"=>"cdp", "pop"=>1890, "name_en"=>"Elysburg", "lang"=>"en", "lat"=>40.868159, "lon"=>-76.549278),
"/US/US-PA/emigsville-18029" => array("type"=>"cdp", "pop"=>2580, "name_en"=>"Emigsville", "lang"=>"en", "lat"=>40.007800, "lon"=>-76.730216),
"/US/US-PA/emlenton-18030" => array("type"=>"borough", "pop"=>733, "name_en"=>"Emlenton", "lang"=>"en", "lat"=>41.180676, "lon"=>-79.708350),
"/US/US-PA/emmaus-18031" => array("type"=>"borough", "pop"=>11349, "name_en"=>"Emmaus", "lang"=>"en", "lat"=>40.536100, "lon"=>-75.498023),
"/US/US-PA/emporium-18032" => array("type"=>"borough", "pop"=>2207, "name_en"=>"Emporium", "lang"=>"en", "lat"=>41.510866, "lon"=>-78.237122),
"/US/US-PA/emsworth-18033" => array("type"=>"borough", "pop"=>2377, "name_en"=>"Emsworth", "lang"=>"en", "lat"=>40.510600, "lon"=>-80.096449),
"/US/US-PA/enola-18034" => array("type"=>"cdp", "pop"=>5961, "name_ru"=>"Энола", "name_en"=>"Enola", "lang"=>"en", "lat"=>40.288533, "lon"=>-76.935974),
"/US/US-PA/enon-valley-18035" => array("type"=>"borough", "pop"=>362, "name_en"=>"Enon Valley", "lang"=>"en", "lat"=>40.855800, "lon"=>-80.457409),
"/US/US-PA/ephrata-18036" => array("type"=>"borough", "pop"=>13074, "name_en"=>"Ephrata", "lang"=>"en", "lat"=>40.180650, "lon"=>-76.182755),
"/US/US-PA/erie-18037" => array("type"=>"city", "pop"=>103817, "name_ru"=>"Эри", "name_en"=>"Erie", "lang"=>"en", "lat"=>42.125850, "lon"=>-80.086596),
"/US/US-PA/ernest-18038" => array("type"=>"borough", "pop"=>464, "name_en"=>"Ernest", "lang"=>"en", "lat"=>40.678471, "lon"=>-79.165001),
"/US/US-PA/espy-18039" => array("type"=>"cdp", "pop"=>1430, "name_en"=>"Espy", "lang"=>"en", "lat"=>41.004996, "lon"=>-76.417694),
"/US/US-PA/etna-18040" => array("type"=>"borough", "pop"=>3560, "name_ru"=>"Этна", "name_en"=>"Etna", "lang"=>"en", "lat"=>40.498550, "lon"=>-79.947257),
"/US/US-PA/evans-city-18042" => array("type"=>"borough", "pop"=>1883, "name_en"=>"Evans City", "lang"=>"en", "lat"=>40.768461, "lon"=>-80.058744),
"/US/US-PA/evansburg-18041" => array("type"=>"cdp", "pop"=>1047, "name_en"=>"Evansburg", "lang"=>"en", "lat"=>40.190200, "lon"=>-75.435988),
"/US/US-PA/everett-18043" => array("type"=>"borough", "pop"=>1850, "name_ru"=>"Эверетт", "name_en"=>"Everett", "lang"=>"en", "lat"=>40.012707, "lon"=>-78.366317),
"/US/US-PA/everson-18044" => array("type"=>"borough", "pop"=>787, "name_en"=>"Everson", "lang"=>"en", "lat"=>40.090509, "lon"=>-79.587282),
"/US/US-PA/exeter-18045" => array("type"=>"borough", "pop"=>5905, "name_ru"=>"Эксетер", "name_en"=>"Exeter", "lang"=>"en", "lat"=>41.330250, "lon"=>-75.822500),
"/US/US-PA/export-18046" => array("type"=>"borough", "pop"=>830, "name_en"=>"Export", "lang"=>"en", "lat"=>40.418688, "lon"=>-79.624954),
"/US/US-PA/exton-18047" => array("type"=>"cdp", "pop"=>2550, "name_en"=>"Exton", "lang"=>"en", "lat"=>40.031578, "lon"=>-75.630019),
"/US/US-PA/factoryville-18048" => array("type"=>"borough", "pop"=>1169, "name_en"=>"Factoryville", "lang"=>"en", "lat"=>41.560040, "lon"=>-75.783460),
"/US/US-PA/fairchance-18049" => array("type"=>"borough", "pop"=>2050, "name_en"=>"Fairchance", "lang"=>"en", "lat"=>39.825854, "lon"=>-79.754593),
"/US/US-PA/fairdale-18050" => array("type"=>"cdp", "pop"=>2049, "name_ru"=>"Фэрдейл", "name_en"=>"Fairdale", "lang"=>"en", "lat"=>39.888384, "lon"=>-79.969065),
"/US/US-PA/fairfield-18051" => array("type"=>"borough", "pop"=>509, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>39.787707, "lon"=>-77.369207),
"/US/US-PA/fairless-hills-18052" => array("type"=>"cdp", "pop"=>9026, "name_en"=>"Fairless Hills", "lang"=>"en", "lat"=>40.178800, "lon"=>-74.853436),
"/US/US-PA/fairview-18053" => array("type"=>"borough", "pop"=>206, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>41.015226, "lon"=>-79.743532),
"/US/US-PA/fairview-18054" => array("type"=>"borough", "pop"=>1988, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>42.029347, "lon"=>-80.255361),
"/US/US-PA/fairview-ferndale-18055" => array("type"=>"cdp", "pop"=>2895, "name_en"=>"Fairview-Ferndale", "lang"=>"en", "lat"=>40.779962, "lon"=>-76.575605),
"/US/US-PA/falls-creek-18056" => array("type"=>"borough", "pop"=>926, "name_en"=>"Falls Creek", "lang"=>"en", "lat"=>41.142117, "lon"=>-78.807234),
"/US/US-PA/fallston-18057" => array("type"=>"borough", "pop"=>288, "name_ru"=>"Фоллстон", "name_en"=>"Fallston", "lang"=>"en", "lat"=>40.721467, "lon"=>-80.314472),
"/US/US-PA/farrell-18058" => array("type"=>"city", "pop"=>5751, "name_en"=>"Farrell", "lang"=>"en", "lat"=>41.209550, "lon"=>-80.497449),
"/US/US-PA/fawn-grove-18059" => array("type"=>"borough", "pop"=>469, "name_en"=>"Fawn Grove", "lang"=>"en", "lat"=>39.729900, "lon"=>-76.451076),
"/US/US-PA/fayette-city-18060" => array("type"=>"borough", "pop"=>667, "name_en"=>"Fayette City", "lang"=>"en", "lat"=>40.100176, "lon"=>-79.839086),
"/US/US-PA/fayetteville-18061" => array("type"=>"cdp", "pop"=>3033, "name_ru"=>"Фейетвилл", "name_en"=>"Fayetteville", "lang"=>"en", "lat"=>39.912370, "lon"=>-77.565708),
"/US/US-PA/feasterville-trevose-18062" => array("type"=>"cdp", "pop"=>6696, "name_en"=>"Feasterville-Trevose", "lang"=>"en", "lat"=>40.152650, "lon"=>-74.986586),
"/US/US-PA/felton-18063" => array("type"=>"borough", "pop"=>468, "name_ru"=>"Фелтон", "name_en"=>"Felton", "lang"=>"en", "lat"=>39.856267, "lon"=>-76.561536),
"/US/US-PA/ferndale-18064" => array("type"=>"borough", "pop"=>1653, "name_ru"=>"Ферндейл", "name_en"=>"Ferndale", "lang"=>"en", "lat"=>40.288350, "lon"=>-78.917214),
"/US/US-PA/fernway-18065" => array("type"=>"cdp", "pop"=>9072, "name_en"=>"Fernway", "lang"=>"en", "lat"=>40.694664, "lon"=>-80.130980),
"/US/US-PA/finleyville-18066" => array("type"=>"borough", "pop"=>439, "name_en"=>"Finleyville", "lang"=>"en", "lat"=>40.252912, "lon"=>-80.002184),
"/US/US-PA/fleetwood-18067" => array("type"=>"borough", "pop"=>3998, "name_en"=>"Fleetwood", "lang"=>"en", "lat"=>40.455578, "lon"=>-75.821405),
"/US/US-PA/flemington-18068" => array("type"=>"borough", "pop"=>1253, "name_ru"=>"Флемингтон", "name_en"=>"Flemington", "lang"=>"en", "lat"=>41.127104, "lon"=>-77.470502),
"/US/US-PA/flourtown-18069" => array("type"=>"cdp", "pop"=>4754, "name_en"=>"Flourtown", "lang"=>"en", "lat"=>40.102800, "lon"=>-75.208649),
"/US/US-PA/flying-hills-18070" => array("type"=>"cdp", "pop"=>1526, "name_en"=>"Flying Hills", "lang"=>"en", "lat"=>40.276667, "lon"=>-75.914517),
"/US/US-PA/folcroft-18071" => array("type"=>"borough", "pop"=>6823, "name_en"=>"Folcroft", "lang"=>"en", "lat"=>39.888571, "lon"=>-75.276589),
"/US/US-PA/folsom-18072" => array("type"=>"cdp", "pop"=>8173, "name_ru"=>"Фолсом", "name_en"=>"Folsom", "lang"=>"en", "lat"=>39.891388, "lon"=>-75.329105),
"/US/US-PA/ford-city-18073" => array("type"=>"borough", "pop"=>3179, "name_ru"=>"Форд Сити", "name_en"=>"Ford City", "lang"=>"en", "lat"=>40.768187, "lon"=>-79.534515),
"/US/US-PA/ford-cliff-18074" => array("type"=>"borough", "pop"=>385, "name_en"=>"Ford Cliff", "lang"=>"en", "lat"=>40.759872, "lon"=>-79.536612),
"/US/US-PA/forest-city-18075" => array("type"=>"borough", "pop"=>1725, "name_ru"=>"Форест Сити", "name_en"=>"Forest City", "lang"=>"en", "lat"=>41.651616, "lon"=>-75.469697),
"/US/US-PA/forest-hills-18076" => array("type"=>"borough", "pop"=>6240, "name_ru"=>"Форест Хиллс", "name_en"=>"Forest Hills", "lang"=>"en", "lat"=>40.424750, "lon"=>-79.852549),
"/US/US-PA/forksville-18077" => array("type"=>"borough", "pop"=>131, "name_en"=>"Forksville", "lang"=>"en", "lat"=>41.490904, "lon"=>-76.603883),
"/US/US-PA/fort-washington-18078" => array("type"=>"cdp", "pop"=>3699, "name_ru"=>"Форт Вашингтон", "name_en"=>"Fort Washington", "lang"=>"en", "lat"=>40.138450, "lon"=>-75.191799),
"/US/US-PA/forty-fort-18079" => array("type"=>"borough", "pop"=>4243, "name_en"=>"Forty Fort", "lang"=>"en", "lat"=>41.289022, "lon"=>-75.870534),
"/US/US-PA/fountain-hill-18080" => array("type"=>"borough", "pop"=>4575, "name_ru"=>"Фонтейн Хилл", "name_en"=>"Fountain Hill", "lang"=>"en", "lat"=>40.602550, "lon"=>-75.396666),
"/US/US-PA/fox-chapel-18082" => array("type"=>"borough", "pop"=>5138, "name_en"=>"Fox Chapel", "lang"=>"en", "lat"=>40.525433, "lon"=>-79.889628),
"/US/US-PA/fox-run-18083" => array("type"=>"cdp", "pop"=>2384, "name_en"=>"Fox Run", "lang"=>"en", "lat"=>40.702296, "lon"=>-80.083015),
"/US/US-PA/foxburg-18081" => array("type"=>"borough", "pop"=>261, "name_en"=>"Foxburg", "lang"=>"en", "lat"=>41.144989, "lon"=>-79.680628),
"/US/US-PA/frackville-18084" => array("type"=>"borough", "pop"=>4123, "name_en"=>"Frackville", "lang"=>"en", "lat"=>40.783703, "lon"=>-76.233334),
"/US/US-PA/frankfort-springs-18085" => array("type"=>"borough", "pop"=>119, "name_en"=>"Frankfort Springs", "lang"=>"en", "lat"=>40.481320, "lon"=>-80.445564),
"/US/US-PA/franklin-18086" => array("type"=>"borough", "pop"=>398, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>40.341950, "lon"=>-78.883499),
"/US/US-PA/franklin-18087" => array("type"=>"city", "pop"=>7329, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>41.391951, "lon"=>-79.838795),
"/US/US-PA/franklin-park-18088" => array("type"=>"borough", "pop"=>12172, "name_ru"=>"Франклин Парк", "name_en"=>"Franklin Park", "lang"=>"en", "lat"=>40.591728, "lon"=>-80.098691),
"/US/US-PA/franklintown-18089" => array("type"=>"borough", "pop"=>553, "name_en"=>"Franklintown", "lang"=>"en", "lat"=>40.075175, "lon"=>-77.029144),
"/US/US-PA/fredericksburg-18090" => array("type"=>"cdp", "pop"=>1269, "name_ru"=>"Фредериксбург", "name_en"=>"Fredericksburg", "lang"=>"en", "lat"=>41.647601, "lon"=>-80.179503),
"/US/US-PA/fredericksburg-18091" => array("type"=>"cdp", "pop"=>2338, "name_ru"=>"Фредериксбург", "name_en"=>"Fredericksburg", "lang"=>"en", "lat"=>40.456319, "lon"=>-76.425571),
"/US/US-PA/fredericktown-millsboro-18092" => array("type"=>"cdp", "pop"=>1237, "name_en"=>"Fredericktown-Millsboro", "lang"=>"en", "lat"=>39.997432, "lon"=>-80.005112),
"/US/US-PA/fredonia-18093" => array("type"=>"borough", "pop"=>597, "name_ru"=>"Фредониа", "name_en"=>"Fredonia", "lang"=>"en", "lat"=>41.322216, "lon"=>-80.258822),
"/US/US-PA/freeburg-18094" => array("type"=>"borough", "pop"=>577, "name_ru"=>"Фрибург", "name_en"=>"Freeburg", "lang"=>"en", "lat"=>40.762810, "lon"=>-76.941223),
"/US/US-PA/freedom-18095" => array("type"=>"borough", "pop"=>1595, "name_ru"=>"Фридом", "name_en"=>"Freedom", "lang"=>"en", "lat"=>40.684650, "lon"=>-80.253360),
"/US/US-PA/freeland-18096" => array("type"=>"borough", "pop"=>3383, "name_ru"=>"Фриленд", "name_en"=>"Freeland", "lang"=>"en", "lat"=>41.020750, "lon"=>-75.896349),
"/US/US-PA/freemansburg-18097" => array("type"=>"borough", "pop"=>2022, "name_en"=>"Freemansburg", "lang"=>"en", "lat"=>40.627250, "lon"=>-75.340199),
"/US/US-PA/freeport-18098" => array("type"=>"borough", "pop"=>1810, "name_ru"=>"Фрипорт", "name_en"=>"Freeport", "lang"=>"en", "lat"=>40.682635, "lon"=>-79.682812),
"/US/US-PA/friedens-18099" => array("type"=>"cdp", "pop"=>1576, "name_en"=>"Friedens", "lang"=>"en", "lat"=>40.042595, "lon"=>-79.005266),
"/US/US-PA/friendsville-18100" => array("type"=>"borough", "pop"=>85, "name_ru"=>"Френдсвилл", "name_en"=>"Friendsville", "lang"=>"en", "lat"=>41.917912, "lon"=>-76.047306),
"/US/US-PA/fullerton-18101" => array("type"=>"cdp", "pop"=>13127, "name_ru"=>"Фуллертон", "name_en"=>"Fullerton", "lang"=>"en", "lat"=>40.631967, "lon"=>-75.485875),
"/US/US-PA/galeton-18102" => array("type"=>"borough", "pop"=>1213, "name_en"=>"Galeton", "lang"=>"en", "lat"=>41.733245, "lon"=>-77.644033),
"/US/US-PA/gallitzin-18103" => array("type"=>"borough", "pop"=>1604, "name_en"=>"Gallitzin", "lang"=>"en", "lat"=>40.482219, "lon"=>-78.555953),
"/US/US-PA/gap-18104" => array("type"=>"cdp", "pop"=>1226, "name_en"=>"Gap", "lang"=>"en", "lat"=>39.988937, "lon"=>-76.024234),
"/US/US-PA/garden-view-18105" => array("type"=>"cdp", "pop"=>2687, "name_en"=>"Garden View", "lang"=>"en", "lat"=>41.256500, "lon"=>-77.048697),
"/US/US-PA/garrett-18106" => array("type"=>"borough", "pop"=>420, "name_ru"=>"Гарретт", "name_en"=>"Garrett", "lang"=>"en", "lat"=>39.864167, "lon"=>-79.061653),
"/US/US-PA/gastonville-18107" => array("type"=>"cdp", "pop"=>3090, "name_en"=>"Gastonville", "lang"=>"en", "lat"=>40.264432, "lon"=>-80.007270),
"/US/US-PA/geistown-18108" => array("type"=>"borough", "pop"=>2352, "name_en"=>"Geistown", "lang"=>"en", "lat"=>40.293850, "lon"=>-78.871749),
"/US/US-PA/georgetown-18109" => array("type"=>"borough", "pop"=>155, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>40.641710, "lon"=>-80.499817),
"/US/US-PA/gettysburg-18110" => array("type"=>"borough", "pop"=>8040, "name_en"=>"Gettysburg", "lang"=>"en", "lat"=>39.830132, "lon"=>-77.234547),
"/US/US-PA/gilberton-18111" => array("type"=>"borough", "pop"=>822, "name_en"=>"Gilberton", "lang"=>"en", "lat"=>40.796556, "lon"=>-76.223282),
"/US/US-PA/gilbertsville-18112" => array("type"=>"cdp", "pop"=>3994, "name_en"=>"Gilbertsville", "lang"=>"en", "lat"=>40.322830, "lon"=>-75.609042),
"/US/US-PA/girard-18113" => array("type"=>"borough", "pop"=>2928, "name_ru"=>"Жирар", "name_en"=>"Girard", "lang"=>"en", "lat"=>42.003991, "lon"=>-80.319401),
"/US/US-PA/girardville-18114" => array("type"=>"borough", "pop"=>1641, "name_en"=>"Girardville", "lang"=>"en", "lat"=>40.792257, "lon"=>-76.284228),
"/US/US-PA/glasgow-18115" => array("type"=>"borough", "pop"=>58, "name_ru"=>"Гласго", "name_en"=>"Glasgow", "lang"=>"en", "lat"=>40.644577, "lon"=>-80.508559),
"/US/US-PA/glassport-18116" => array("type"=>"borough", "pop"=>4548, "name_en"=>"Glassport", "lang"=>"en", "lat"=>40.325350, "lon"=>-79.886633),
"/US/US-PA/glen-campbell-18118" => array("type"=>"borough", "pop"=>285, "name_en"=>"Glen Campbell", "lang"=>"en", "lat"=>40.818309, "lon"=>-78.830517),
"/US/US-PA/glen-hope-18121" => array("type"=>"borough", "pop"=>139, "name_en"=>"Glen Hope", "lang"=>"en", "lat"=>40.798952, "lon"=>-78.498088),
"/US/US-PA/glen-lyon-18122" => array("type"=>"cdp", "pop"=>2082, "name_en"=>"Glen Lyon", "lang"=>"en", "lat"=>41.184050, "lon"=>-76.071433),
"/US/US-PA/glen-rock-18124" => array("type"=>"borough", "pop"=>1837, "name_en"=>"Glen Rock", "lang"=>"en", "lat"=>39.793365, "lon"=>-76.730461),
"/US/US-PA/glenburn-18117" => array("type"=>"cdp", "pop"=>1279, "name_en"=>"Glenburn", "lang"=>"en", "lat"=>41.509173, "lon"=>-75.735244),
"/US/US-PA/glendon-18119" => array("type"=>"borough", "pop"=>358, "name_en"=>"Glendon", "lang"=>"en", "lat"=>40.661308, "lon"=>-75.235775),
"/US/US-PA/glenfield-18120" => array("type"=>"borough", "pop"=>218, "name_en"=>"Glenfield", "lang"=>"en", "lat"=>40.521733, "lon"=>-80.132144),
"/US/US-PA/glenolden-18123" => array("type"=>"borough", "pop"=>7203, "name_en"=>"Glenolden", "lang"=>"en", "lat"=>39.898700, "lon"=>-75.292826),
"/US/US-PA/glenside-18125" => array("type"=>"cdp", "pop"=>8704, "name_en"=>"Glenside", "lang"=>"en", "lat"=>40.102850, "lon"=>-75.152574),
"/US/US-PA/goldsboro-18126" => array("type"=>"borough", "pop"=>938, "name_ru"=>"Гоулсборо", "name_en"=>"Goldsboro", "lang"=>"en", "lat"=>40.154512, "lon"=>-76.750458),
"/US/US-PA/gordon-18127" => array("type"=>"borough", "pop"=>740, "name_ru"=>"Гордон", "name_en"=>"Gordon", "lang"=>"en", "lat"=>40.750355, "lon"=>-76.340568),
"/US/US-PA/grampian-18128" => array("type"=>"borough", "pop"=>408, "name_en"=>"Grampian", "lang"=>"en", "lat"=>40.964026, "lon"=>-78.611012),
"/US/US-PA/grandview-park-18129" => array("type"=>"cdp", "pop"=>2170, "name_en"=>"Grandview Park", "lang"=>"en", "lat"=>41.416978, "lon"=>-78.542451),
"/US/US-PA/grantley-18130" => array("type"=>"cdp", "pop"=>3069, "name_en"=>"Grantley", "lang"=>"en", "lat"=>39.940217, "lon"=>-76.729451),
"/US/US-PA/gratz-18131" => array("type"=>"borough", "pop"=>658, "name_ru"=>"Гратц", "name_en"=>"Gratz", "lang"=>"en", "lat"=>40.605602, "lon"=>-76.717348),
"/US/US-PA/great-bend-18132" => array("type"=>"borough", "pop"=>658, "name_ru"=>"Грейт Бенд", "name_en"=>"Great Bend", "lang"=>"en", "lat"=>41.972732, "lon"=>-75.745583),
"/US/US-PA/green-hills-18134" => array("type"=>"borough", "pop"=>17, "name_en"=>"Green Hills", "lang"=>"en", "lat"=>40.115639, "lon"=>-80.308780),
"/US/US-PA/green-lane-18135" => array("type"=>"borough", "pop"=>570, "name_en"=>"Green Lane", "lang"=>"en", "lat"=>40.335780, "lon"=>-75.470625),
"/US/US-PA/green-tree-18138" => array("type"=>"borough", "pop"=>4335, "name_en"=>"Green Tree", "lang"=>"en", "lat"=>40.418467, "lon"=>-80.055262),
"/US/US-PA/greencastle-18133" => array("type"=>"borough", "pop"=>4053, "name_ru"=>"Гринкасл", "name_en"=>"Greencastle", "lang"=>"en", "lat"=>39.790519, "lon"=>-77.726920),
"/US/US-PA/greensboro-18136" => array("type"=>"borough", "pop"=>273, "name_ru"=>"Гринсборо", "name_en"=>"Greensboro", "lang"=>"en", "lat"=>39.792067, "lon"=>-79.913037),
"/US/US-PA/greensburg-18137" => array("type"=>"city", "pop"=>15249, "name_ru"=>"Гринсбург", "name_en"=>"Greensburg", "lang"=>"en", "lat"=>40.310936, "lon"=>-79.544396),
"/US/US-PA/greenville-18139" => array("type"=>"borough", "pop"=>6103, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>41.406035, "lon"=>-80.384282),
"/US/US-PA/greenville-east-18140" => array("type"=>"cdp", "pop"=>1419, "name_en"=>"Greenville East", "lang"=>"en", "lat"=>41.402718, "lon"=>-80.359031),
"/US/US-PA/grindstone-rowes-run-18141" => array("type"=>"cdp", "pop"=>1041, "name_en"=>"Grindstone-Rowes Run", "lang"=>"en", "lat"=>40.015192, "lon"=>-79.815216),
"/US/US-PA/grove-city-18142" => array("type"=>"borough", "pop"=>7715, "name_ru"=>"Гроув Сити", "name_en"=>"Grove City", "lang"=>"en", "lat"=>41.156925, "lon"=>-80.088753),
"/US/US-PA/guilford-18143" => array("type"=>"cdp", "pop"=>14473, "name_ru"=>"Гилфорд", "name_en"=>"Guilford", "lang"=>"en", "lat"=>39.915515, "lon"=>-77.599350),
"/US/US-PA/halfway-house-18144" => array("type"=>"cdp", "pop"=>1415, "name_en"=>"Halfway House", "lang"=>"en", "lat"=>40.281864, "lon"=>-75.643645),
"/US/US-PA/halifax-18145" => array("type"=>"borough", "pop"=>833, "name_en"=>"Halifax", "lang"=>"en", "lat"=>40.464999, "lon"=>-76.933438),
"/US/US-PA/hallam-18146" => array("type"=>"borough", "pop"=>2867, "name_en"=>"Hallam", "lang"=>"en", "lat"=>40.002500, "lon"=>-76.604449),
"/US/US-PA/hallstead-18147" => array("type"=>"borough", "pop"=>1128, "name_en"=>"Hallstead", "lang"=>"en", "lat"=>41.961909, "lon"=>-75.747335),
"/US/US-PA/hamburg-18148" => array("type"=>"borough", "pop"=>4211, "name_ru"=>"Гамбург", "name_en"=>"Hamburg", "lang"=>"en", "lat"=>40.556179, "lon"=>-75.983013),
"/US/US-PA/hampton-township-18149" => array("type"=>"cdp", "pop"=>15568, "name_en"=>"Hampton Township", "lang"=>"en", "lat"=>40.585840, "lon"=>-79.954985),
"/US/US-PA/hanover-18150" => array("type"=>"borough", "pop"=>15025, "name_ru"=>"Хановер", "name_en"=>"Hanover", "lang"=>"en", "lat"=>39.812118, "lon"=>-76.983720),
"/US/US-PA/harleysville-18151" => array("type"=>"cdp", "pop"=>7405, "name_en"=>"Harleysville", "lang"=>"en", "lat"=>40.278373, "lon"=>-75.388020),
"/US/US-PA/harmony-18152" => array("type"=>"borough", "pop"=>868, "name_ru"=>"Хармони", "name_en"=>"Harmony", "lang"=>"en", "lat"=>40.801264, "lon"=>-80.125165),
"/US/US-PA/harmony-township-18153" => array("type"=>"cdp", "pop"=>3694, "name_en"=>"Harmony Township", "lang"=>"en", "lat"=>40.607150, "lon"=>-80.220539),
"/US/US-PA/harrisburg-18154" => array("type"=>"city", "pop"=>47148, "name_ru"=>"Харрисбург", "name_en"=>"Harrisburg", "lang"=>"en", "lat"=>40.276050, "lon"=>-76.884503),
"/US/US-PA/harrison-township-18155" => array("type"=>"cdp", "pop"=>11763, "name_en"=>"Harrison Township", "lang"=>"en", "lat"=>40.636640, "lon"=>-79.716690),
"/US/US-PA/harrisville-18156" => array("type"=>"borough", "pop"=>876, "name_ru"=>"Харрисвилл", "name_en"=>"Harrisville", "lang"=>"en", "lat"=>41.135959, "lon"=>-80.011028),
"/US/US-PA/hartleton-18157" => array("type"=>"borough", "pop"=>263, "name_en"=>"Hartleton", "lang"=>"en", "lat"=>40.899449, "lon"=>-77.156750),
"/US/US-PA/harveys-lake-18158" => array("type"=>"borough", "pop"=>2930, "name_en"=>"Harveys Lake", "lang"=>"en", "lat"=>41.362120, "lon"=>-76.032207),
"/US/US-PA/hasson-heights-18159" => array("type"=>"cdp", "pop"=>1610, "name_en"=>"Hasson Heights", "lang"=>"en", "lat"=>41.448830, "lon"=>-79.677134),
"/US/US-PA/hastings-18160" => array("type"=>"borough", "pop"=>1296, "name_ru"=>"Гастингс", "name_en"=>"Hastings", "lang"=>"en", "lat"=>40.664906, "lon"=>-78.708969),
"/US/US-PA/hatboro-18161" => array("type"=>"borough", "pop"=>7125, "name_en"=>"Hatboro", "lang"=>"en", "lat"=>40.177550, "lon"=>-75.104795),
"/US/US-PA/hatfield-18162" => array("type"=>"borough", "pop"=>2836, "name_ru"=>"Хэтфилд", "name_en"=>"Hatfield", "lang"=>"en", "lat"=>40.277981, "lon"=>-75.299175),
"/US/US-PA/hawley-18163" => array("type"=>"borough", "pop"=>1299, "name_ru"=>"Холи", "name_en"=>"Hawley", "lang"=>"en", "lat"=>41.476574, "lon"=>-75.177972),
"/US/US-PA/hawthorn-18164" => array("type"=>"borough", "pop"=>546, "name_en"=>"Hawthorn", "lang"=>"en", "lat"=>41.020786, "lon"=>-79.276185),
"/US/US-PA/haysville-18165" => array("type"=>"borough", "pop"=>72, "name_en"=>"Haysville", "lang"=>"en", "lat"=>40.526100, "lon"=>-80.154377),
"/US/US-PA/hazleton-18166" => array("type"=>"city", "pop"=>21732, "name_ru"=>"Хейзлтон", "name_en"=>"Hazleton", "lang"=>"en", "lat"=>40.950100, "lon"=>-75.972312),
"/US/US-PA/heidelberg-18167" => array("type"=>"borough", "pop"=>1143, "name_ru"=>"Хайделберг", "name_en"=>"Heidelberg", "lang"=>"en", "lat"=>40.391400, "lon"=>-80.092032),
"/US/US-PA/hellertown-18168" => array("type"=>"borough", "pop"=>5659, "name_en"=>"Hellertown", "lang"=>"en", "lat"=>40.581233, "lon"=>-75.338315),
"/US/US-PA/hermitage-18169" => array("type"=>"city", "pop"=>16327, "name_ru"=>"Хермитедж", "name_en"=>"Hermitage", "lang"=>"en", "lat"=>41.229850, "lon"=>-80.442922),
"/US/US-PA/herndon-18170" => array("type"=>"borough", "pop"=>354, "name_en"=>"Herndon", "lang"=>"en", "lat"=>40.710555, "lon"=>-76.854564),
"/US/US-PA/hershey-18171" => array("type"=>"cdp", "pop"=>11860, "name_en"=>"Hershey", "lang"=>"en", "lat"=>40.280085, "lon"=>-76.647699),
"/US/US-PA/highland-park-18172" => array("type"=>"cdp", "pop"=>1583, "name_ru"=>"Хайленд Парк", "name_en"=>"Highland Park", "lang"=>"en", "lat"=>40.621703, "lon"=>-77.570770),
"/US/US-PA/highspire-18173" => array("type"=>"borough", "pop"=>2606, "name_en"=>"Highspire", "lang"=>"en", "lat"=>40.208500, "lon"=>-76.785938),
"/US/US-PA/hiller-18174" => array("type"=>"cdp", "pop"=>1401, "name_en"=>"Hiller", "lang"=>"en", "lat"=>40.007224, "lon"=>-79.903850),
"/US/US-PA/hokendauqua-18175" => array("type"=>"cdp", "pop"=>3413, "name_en"=>"Hokendauqua", "lang"=>"en", "lat"=>40.658667, "lon"=>-75.495817),
"/US/US-PA/hollidaysburg-18176" => array("type"=>"borough", "pop"=>5466, "name_en"=>"Hollidaysburg", "lang"=>"en", "lat"=>40.430150, "lon"=>-78.393061),
"/US/US-PA/homeacre-lyndora-18177" => array("type"=>"cdp", "pop"=>7511, "name_en"=>"Homeacre-Lyndora", "lang"=>"en", "lat"=>40.871010, "lon"=>-79.923020),
"/US/US-PA/homer-city-18178" => array("type"=>"borough", "pop"=>1711, "name_en"=>"Homer City", "lang"=>"en", "lat"=>40.539605, "lon"=>-79.159488),
"/US/US-PA/homestead-18179" => array("type"=>"borough", "pop"=>3490, "name_ru"=>"Хоумстид", "name_en"=>"Homestead", "lang"=>"en", "lat"=>40.406750, "lon"=>-79.909798),
"/US/US-PA/hometown-18180" => array("type"=>"cdp", "pop"=>1545, "name_ru"=>"Хоумтаун", "name_en"=>"Hometown", "lang"=>"en", "lat"=>40.822032, "lon"=>-75.985998),
"/US/US-PA/homewood-18181" => array("type"=>"borough", "pop"=>138, "name_ru"=>"Хомвуд", "name_en"=>"Homewood", "lang"=>"en", "lat"=>40.813376, "lon"=>-80.329216),
"/US/US-PA/honesdale-18182" => array("type"=>"borough", "pop"=>4720, "name_en"=>"Honesdale", "lang"=>"en", "lat"=>41.577216, "lon"=>-75.253557),
"/US/US-PA/honey-brook-18183" => array("type"=>"borough", "pop"=>1587, "name_en"=>"Honey Brook", "lang"=>"en", "lat"=>40.093375, "lon"=>-75.910982),
"/US/US-PA/hookstown-18184" => array("type"=>"borough", "pop"=>138, "name_en"=>"Hookstown", "lang"=>"en", "lat"=>40.598804, "lon"=>-80.474362),
"/US/US-PA/hooversville-18185" => array("type"=>"borough", "pop"=>708, "name_en"=>"Hooversville", "lang"=>"en", "lat"=>40.150421, "lon"=>-78.914563),
"/US/US-PA/hop-bottom-18186" => array("type"=>"borough", "pop"=>304, "name_en"=>"Hop Bottom", "lang"=>"en", "lat"=>41.706435, "lon"=>-75.767224),
"/US/US-PA/hopewell-18187" => array("type"=>"borough", "pop"=>209, "name_ru"=>"Хоупуэлл", "name_en"=>"Hopewell", "lang"=>"en", "lat"=>40.134606, "lon"=>-78.266694),
"/US/US-PA/hopwood-18188" => array("type"=>"cdp", "pop"=>2021, "name_en"=>"Hopwood", "lang"=>"en", "lat"=>39.877475, "lon"=>-79.701815),
"/US/US-PA/horsham-18189" => array("type"=>"cdp", "pop"=>24720, "name_en"=>"Horsham", "lang"=>"en", "lat"=>40.182500, "lon"=>-75.139229),
"/US/US-PA/houston-18190" => array("type"=>"borough", "pop"=>1245, "name_ru"=>"Хьюстон", "name_en"=>"Houston", "lang"=>"en", "lat"=>40.249992, "lon"=>-80.211816),
"/US/US-PA/houtzdale-18191" => array("type"=>"borough", "pop"=>879, "name_en"=>"Houtzdale", "lang"=>"en", "lat"=>40.824969, "lon"=>-78.351017),
"/US/US-PA/howard-18192" => array("type"=>"borough", "pop"=>668, "name_ru"=>"Ховарт", "name_en"=>"Howard", "lang"=>"en", "lat"=>41.011965, "lon"=>-77.657029),
"/US/US-PA/hughestown-18193" => array("type"=>"borough", "pop"=>1443, "name_en"=>"Hughestown", "lang"=>"en", "lat"=>41.330100, "lon"=>-75.771934),
"/US/US-PA/hughesville-18194" => array("type"=>"borough", "pop"=>2045, "name_ru"=>"Хьюсвилл", "name_en"=>"Hughesville", "lang"=>"en", "lat"=>41.238463, "lon"=>-76.726351),
"/US/US-PA/hulmeville-18195" => array("type"=>"borough", "pop"=>873, "name_en"=>"Hulmeville", "lang"=>"en", "lat"=>40.143150, "lon"=>-74.907508),
"/US/US-PA/hummels-wharf-18197" => array("type"=>"cdp", "pop"=>1069, "name_en"=>"Hummels Wharf", "lang"=>"en", "lat"=>40.833609, "lon"=>-76.839612),
"/US/US-PA/hummelstown-18196" => array("type"=>"borough", "pop"=>4424, "name_en"=>"Hummelstown", "lang"=>"en", "lat"=>40.266450, "lon"=>-76.712932),
"/US/US-PA/hunker-18198" => array("type"=>"borough", "pop"=>312, "name_en"=>"Hunker", "lang"=>"en", "lat"=>40.204032, "lon"=>-79.615296),
"/US/US-PA/huntingdon-18199" => array("type"=>"borough", "pop"=>6759, "name_en"=>"Huntingdon", "lang"=>"en", "lat"=>40.496520, "lon"=>-78.010305),
"/US/US-PA/hyde-18200" => array("type"=>"cdp", "pop"=>1643, "name_en"=>"Hyde", "lang"=>"en", "lat"=>41.007416, "lon"=>-78.468328),
"/US/US-PA/hyde-park-18201" => array("type"=>"borough", "pop"=>478, "name_en"=>"Hyde Park", "lang"=>"en", "lat"=>40.632216, "lon"=>-79.589027),
"/US/US-PA/hydetown-18202" => array("type"=>"borough", "pop"=>567, "name_en"=>"Hydetown", "lang"=>"en", "lat"=>41.651621, "lon"=>-79.724280),
"/US/US-PA/hyndman-18203" => array("type"=>"borough", "pop"=>957, "name_en"=>"Hyndman", "lang"=>"en", "lat"=>39.820954, "lon"=>-78.721113),
"/US/US-PA/imperial-enlow-18204" => array("type"=>"cdp", "pop"=>3449, "name_en"=>"Imperial-Enlow", "lang"=>"en", "lat"=>40.453413, "lon"=>-80.248530),
"/US/US-PA/indian-lake-18206" => array("type"=>"borough", "pop"=>458, "name_en"=>"Indian Lake", "lang"=>"en", "lat"=>40.047537, "lon"=>-78.859454),
"/US/US-PA/indiana-18205" => array("type"=>"borough", "pop"=>14754, "name_en"=>"Indiana", "lang"=>"en", "lat"=>40.621931, "lon"=>-79.155832),
"/US/US-PA/industry-18207" => array("type"=>"borough", "pop"=>1789, "name_ru"=>"Индастри", "name_en"=>"Industry", "lang"=>"en", "lat"=>40.655952, "lon"=>-80.409751),
"/US/US-PA/ingram-18208" => array("type"=>"borough", "pop"=>3371, "name_en"=>"Ingram", "lang"=>"en", "lat"=>40.445700, "lon"=>-80.068387),
"/US/US-PA/irvona-18209" => array("type"=>"borough", "pop"=>634, "name_en"=>"Irvona", "lang"=>"en", "lat"=>40.773454, "lon"=>-78.552807),
"/US/US-PA/irwin-18210" => array("type"=>"borough", "pop"=>4050, "name_ru"=>"Ирвин", "name_en"=>"Irwin", "lang"=>"en", "lat"=>40.325850, "lon"=>-79.699550),
"/US/US-PA/ivyland-18211" => array("type"=>"borough", "pop"=>805, "name_en"=>"Ivyland", "lang"=>"en", "lat"=>40.208100, "lon"=>-75.072920),
"/US/US-PA/jackson-center-18212" => array("type"=>"borough", "pop"=>207, "name_en"=>"Jackson Center", "lang"=>"en", "lat"=>41.273212, "lon"=>-80.138831),
"/US/US-PA/jacksonville-18213" => array("type"=>"borough", "pop"=>89, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>40.541523, "lon"=>-79.283395),
"/US/US-PA/jacobus-18214" => array("type"=>"borough", "pop"=>1191, "name_en"=>"Jacobus", "lang"=>"en", "lat"=>39.882416, "lon"=>-76.712377),
"/US/US-PA/jamestown-18215" => array("type"=>"borough", "pop"=>577, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>41.484885, "lon"=>-80.437030),
"/US/US-PA/jeannette-18216" => array("type"=>"city", "pop"=>9859, "name_en"=>"Jeannette", "lang"=>"en", "lat"=>40.327232, "lon"=>-79.614085),
"/US/US-PA/jeddo-18217" => array("type"=>"borough", "pop"=>134, "name_en"=>"Jeddo", "lang"=>"en", "lat"=>40.992450, "lon"=>-75.897621),
"/US/US-PA/jefferson-18218" => array("type"=>"borough", "pop"=>313, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>40.293733, "lon"=>-79.932190),
"/US/US-PA/jefferson-18219" => array("type"=>"borough", "pop"=>355, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>39.930144, "lon"=>-80.057825),
"/US/US-PA/jefferson-18220" => array("type"=>"borough", "pop"=>675, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>39.815958, "lon"=>-76.841380),
"/US/US-PA/jenkintown-18221" => array("type"=>"borough", "pop"=>4299, "name_en"=>"Jenkintown", "lang"=>"en", "lat"=>40.095950, "lon"=>-75.129797),
"/US/US-PA/jennerstown-18222" => array("type"=>"borough", "pop"=>692, "name_en"=>"Jennerstown", "lang"=>"en", "lat"=>40.164431, "lon"=>-79.062145),
"/US/US-PA/jermyn-18223" => array("type"=>"borough", "pop"=>2231, "name_en"=>"Jermyn", "lang"=>"en", "lat"=>41.526667, "lon"=>-75.546566),
"/US/US-PA/jerome-18224" => array("type"=>"cdp", "pop"=>1074, "name_ru"=>"Джером", "name_en"=>"Jerome", "lang"=>"en", "lat"=>40.213361, "lon"=>-78.983937),
"/US/US-PA/jersey-shore-18225" => array("type"=>"borough", "pop"=>4302, "name_en"=>"Jersey Shore", "lang"=>"en", "lat"=>41.200739, "lon"=>-77.266963),
"/US/US-PA/jessup-18226" => array("type"=>"borough", "pop"=>4550, "name_ru"=>"Джессап", "name_en"=>"Jessup", "lang"=>"en", "lat"=>41.458210, "lon"=>-75.547659),
"/US/US-PA/jim-thorpe-18227" => array("type"=>"borough", "pop"=>4863, "name_en"=>"Jim Thorpe", "lang"=>"en", "lat"=>40.871367, "lon"=>-75.739817),
"/US/US-PA/johnsonburg-18228" => array("type"=>"borough", "pop"=>2678, "name_en"=>"Johnsonburg", "lang"=>"en", "lat"=>41.491935, "lon"=>-78.678863),
"/US/US-PA/johnstown-18229" => array("type"=>"city", "pop"=>21641, "name_ru"=>"Джонстаун", "name_en"=>"Johnstown", "lang"=>"en", "lat"=>40.325350, "lon"=>-78.919548),
"/US/US-PA/jonestown-18230" => array("type"=>"borough", "pop"=>1016, "name_ru"=>"Джоунстаун", "name_en"=>"Jonestown", "lang"=>"en", "lat"=>40.413196, "lon"=>-76.480712),
"/US/US-PA/juniata-terrace-18231" => array("type"=>"borough", "pop"=>476, "name_en"=>"Juniata Terrace", "lang"=>"en", "lat"=>40.584171, "lon"=>-77.574566),
"/US/US-PA/kane-18232" => array("type"=>"borough", "pop"=>3765, "name_ru"=>"Кейн", "name_en"=>"Kane", "lang"=>"en", "lat"=>41.663046, "lon"=>-78.809278),
"/US/US-PA/karns-city-18233" => array("type"=>"borough", "pop"=>225, "name_en"=>"Karns City", "lang"=>"en", "lat"=>40.997101, "lon"=>-79.726015),
"/US/US-PA/kenhorst-18234" => array("type"=>"borough", "pop"=>2647, "name_en"=>"Kenhorst", "lang"=>"en", "lat"=>40.306733, "lon"=>-75.944118),
"/US/US-PA/kenilworth-18235" => array("type"=>"cdp", "pop"=>1890, "name_ru"=>"Кенилворт", "name_en"=>"Kenilworth", "lang"=>"en", "lat"=>40.225445, "lon"=>-75.642615),
"/US/US-PA/kennedy-township-18236" => array("type"=>"cdp", "pop"=>7152, "name_en"=>"Kennedy Township", "lang"=>"en", "lat"=>40.476181, "lon"=>-80.101888),
"/US/US-PA/kennett-square-18237" => array("type"=>"borough", "pop"=>5270, "name_en"=>"Kennett Square", "lang"=>"en", "lat"=>39.843658, "lon"=>-75.711670),
"/US/US-PA/king-of-prussia-18238" => array("type"=>"cdp", "pop"=>18406, "name_en"=>"King of Prussia", "lang"=>"en", "lat"=>40.095522, "lon"=>-75.383417),
"/US/US-PA/kingston-18239" => array("type"=>"borough", "pop"=>12943, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>41.264100, "lon"=>-75.887418),
"/US/US-PA/kistler-18240" => array("type"=>"borough", "pop"=>329, "name_en"=>"Kistler", "lang"=>"en", "lat"=>40.377338, "lon"=>-77.867844),
"/US/US-PA/kittanning-18241" => array("type"=>"borough", "pop"=>4356, "name_en"=>"Kittanning", "lang"=>"en", "lat"=>40.827100, "lon"=>-79.523450),
"/US/US-PA/knox-18242" => array("type"=>"borough", "pop"=>1097, "name_ru"=>"Нокс", "name_en"=>"Knox", "lang"=>"en", "lat"=>41.234774, "lon"=>-79.535702),
"/US/US-PA/knoxville-18243" => array("type"=>"borough", "pop"=>581, "name_ru"=>"Ноксвилл", "name_en"=>"Knoxville", "lang"=>"en", "lat"=>41.956842, "lon"=>-77.435830),
"/US/US-PA/koppel-18244" => array("type"=>"borough", "pop"=>774, "name_en"=>"Koppel", "lang"=>"en", "lat"=>40.834808, "lon"=>-80.322471),
"/US/US-PA/kulpmont-18245" => array("type"=>"borough", "pop"=>2763, "name_en"=>"Kulpmont", "lang"=>"en", "lat"=>40.795352, "lon"=>-76.472207),
"/US/US-PA/kulpsville-18246" => array("type"=>"cdp", "pop"=>5183, "name_en"=>"Kulpsville", "lang"=>"en", "lat"=>40.243700, "lon"=>-75.340852),
"/US/US-PA/kutztown-18247" => array("type"=>"borough", "pop"=>5074, "name_en"=>"Kutztown", "lang"=>"en", "lat"=>40.520914, "lon"=>-75.778278),
"/US/US-PA/laceyville-18248" => array("type"=>"borough", "pop"=>368, "name_en"=>"Laceyville", "lang"=>"en", "lat"=>41.645034, "lon"=>-76.159484),
"/US/US-PA/laflin-18249" => array("type"=>"borough", "pop"=>1487, "name_en"=>"Laflin", "lang"=>"en", "lat"=>41.289550, "lon"=>-75.796212),
"/US/US-PA/lake-city-18250" => array("type"=>"borough", "pop"=>2873, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>42.017293, "lon"=>-80.346853),
"/US/US-PA/lake-wynonah-18251" => array("type"=>"cdp", "pop"=>1055, "name_en"=>"Lake Wynonah", "lang"=>"en", "lat"=>40.587091, "lon"=>-76.180245),
"/US/US-PA/lancaster-18252" => array("type"=>"city", "pop"=>54626, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>40.041850, "lon"=>-76.301247),
"/US/US-PA/landingville-18253" => array("type"=>"borough", "pop"=>169, "name_en"=>"Landingville", "lang"=>"en", "lat"=>40.625492, "lon"=>-76.123063),
"/US/US-PA/landisburg-18254" => array("type"=>"borough", "pop"=>190, "name_en"=>"Landisburg", "lang"=>"en", "lat"=>40.343167, "lon"=>-77.305884),
"/US/US-PA/lanesboro-18255" => array("type"=>"borough", "pop"=>551, "name_ru"=>"Лейнсборо", "name_en"=>"Lanesboro", "lang"=>"en", "lat"=>41.962291, "lon"=>-75.582393),
"/US/US-PA/langhorne-18256" => array("type"=>"borough", "pop"=>1948, "name_en"=>"Langhorne", "lang"=>"en", "lat"=>40.177300, "lon"=>-74.919270),
"/US/US-PA/langhorne-manor-18257" => array("type"=>"borough", "pop"=>1063, "name_en"=>"Langhorne Manor", "lang"=>"en", "lat"=>40.165750, "lon"=>-74.918185),
"/US/US-PA/lansdale-18258" => array("type"=>"borough", "pop"=>15526, "name_en"=>"Lansdale", "lang"=>"en", "lat"=>40.241800, "lon"=>-75.282145),
"/US/US-PA/lansdowne-18259" => array("type"=>"borough", "pop"=>10638, "name_en"=>"Lansdowne", "lang"=>"en", "lat"=>39.940500, "lon"=>-75.275265),
"/US/US-PA/lansford-18260" => array("type"=>"borough", "pop"=>4132, "name_en"=>"Lansford", "lang"=>"en", "lat"=>40.832057, "lon"=>-75.885992),
"/US/US-PA/laporte-18261" => array("type"=>"borough", "pop"=>264, "name_ru"=>"Лапорт", "name_en"=>"Laporte", "lang"=>"en", "lat"=>41.418563, "lon"=>-76.491721),
"/US/US-PA/larksville-18262" => array("type"=>"borough", "pop"=>4435, "name_en"=>"Larksville", "lang"=>"en", "lat"=>41.259374, "lon"=>-75.932321),
"/US/US-PA/latrobe-18263" => array("type"=>"borough", "pop"=>8347, "name_en"=>"Latrobe", "lang"=>"en", "lat"=>40.312102, "lon"=>-79.382684),
"/US/US-PA/laurel-mountain-18265" => array("type"=>"borough", "pop"=>170, "name_en"=>"Laurel Mountain", "lang"=>"en", "lat"=>40.211345, "lon"=>-79.184815),
"/US/US-PA/laurel-run-18266" => array("type"=>"borough", "pop"=>688, "name_en"=>"Laurel Run", "lang"=>"en", "lat"=>41.219296, "lon"=>-75.847586),
"/US/US-PA/laureldale-18264" => array("type"=>"borough", "pop"=>3735, "name_en"=>"Laureldale", "lang"=>"en", "lat"=>40.389200, "lon"=>-75.914500),
"/US/US-PA/lawnton-18267" => array("type"=>"cdp", "pop"=>3221, "name_en"=>"Lawnton", "lang"=>"en", "lat"=>40.265830, "lon"=>-76.798914),
"/US/US-PA/lawrence-park-18268" => array("type"=>"cdp", "pop"=>3687, "name_en"=>"Lawrence Park", "lang"=>"en", "lat"=>42.151867, "lon"=>-80.021702),
"/US/US-PA/lawrenceville-18269" => array("type"=>"borough", "pop"=>606, "name_ru"=>"Лоренсвилл", "name_en"=>"Lawrenceville", "lang"=>"en", "lat"=>41.999076, "lon"=>-77.129502),
"/US/US-PA/lawson-heights-18270" => array("type"=>"cdp", "pop"=>2464, "name_en"=>"Lawson Heights", "lang"=>"en", "lat"=>40.292014, "lon"=>-79.388313),
"/US/US-PA/le-raysville-18282" => array("type"=>"borough", "pop"=>298, "name_en"=>"Le Raysville", "lang"=>"en", "lat"=>41.837623, "lon"=>-76.180741),
"/US/US-PA/leacock-leola-bareville-18271" => array("type"=>"cdp", "pop"=>5685, "name_en"=>"Leacock-Leola-Bareville", "lang"=>"en", "lat"=>40.091400, "lon"=>-76.189147),
"/US/US-PA/lebanon-18272" => array("type"=>"city", "pop"=>24097, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>40.341308, "lon"=>-76.423257),
"/US/US-PA/lebanon-south-18273" => array("type"=>"cdp", "pop"=>1764, "name_en"=>"Lebanon South", "lang"=>"en", "lat"=>40.328560, "lon"=>-76.408298),
"/US/US-PA/leechburg-18274" => array("type"=>"borough", "pop"=>2212, "name_en"=>"Leechburg", "lang"=>"en", "lat"=>40.630432, "lon"=>-79.602814),
"/US/US-PA/leesport-18275" => array("type"=>"borough", "pop"=>1908, "name_en"=>"Leesport", "lang"=>"en", "lat"=>40.444900, "lon"=>-75.969899),
"/US/US-PA/leetsdale-18276" => array("type"=>"borough", "pop"=>1117, "name_en"=>"Leetsdale", "lang"=>"en", "lat"=>40.568300, "lon"=>-80.214781),
"/US/US-PA/lehighton-18277" => array("type"=>"borough", "pop"=>5428, "name_en"=>"Lehighton", "lang"=>"en", "lat"=>40.831213, "lon"=>-75.716595),
"/US/US-PA/leith-hatfield-18278" => array("type"=>"cdp", "pop"=>2437, "name_en"=>"Leith-Hatfield", "lang"=>"en", "lat"=>39.879761, "lon"=>-79.734664),
"/US/US-PA/lemoyne-18279" => array("type"=>"borough", "pop"=>3949, "name_en"=>"Lemoyne", "lang"=>"en", "lat"=>40.244500, "lon"=>-76.899399),
"/US/US-PA/lenape-heights-18280" => array("type"=>"cdp", "pop"=>1355, "name_en"=>"Lenape Heights", "lang"=>"en", "lat"=>40.764294, "lon"=>-79.520810),
"/US/US-PA/lenhartsville-18281" => array("type"=>"borough", "pop"=>169, "name_en"=>"Lenhartsville", "lang"=>"en", "lat"=>40.572655, "lon"=>-75.891176),
"/US/US-PA/levittown-18283" => array("type"=>"cdp", "pop"=>55362, "name_en"=>"Levittown", "lang"=>"en", "lat"=>40.153500, "lon"=>-74.853135),
"/US/US-PA/lewis-run-18286" => array("type"=>"borough", "pop"=>566, "name_en"=>"Lewis Run", "lang"=>"en", "lat"=>41.870781, "lon"=>-78.660712),
"/US/US-PA/lewisberry-18284" => array("type"=>"borough", "pop"=>385, "name_en"=>"Lewisberry", "lang"=>"en", "lat"=>40.135208, "lon"=>-76.860810),
"/US/US-PA/lewisburg-18285" => array("type"=>"borough", "pop"=>5454, "name_ru"=>"Льюисбург", "name_en"=>"Lewisburg", "lang"=>"en", "lat"=>40.964098, "lon"=>-76.890480),
"/US/US-PA/lewistown-18287" => array("type"=>"borough", "pop"=>8528, "name_ru"=>"Льюистаун", "name_en"=>"Lewistown", "lang"=>"en", "lat"=>40.597312, "lon"=>-77.573514),
"/US/US-PA/liberty-18288" => array("type"=>"borough", "pop"=>2436, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>40.323800, "lon"=>-79.857978),
"/US/US-PA/liberty-18289" => array("type"=>"borough", "pop"=>199, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>41.559871, "lon"=>-77.105109),
"/US/US-PA/ligonier-18290" => array("type"=>"borough", "pop"=>1599, "name_ru"=>"Лигоньер", "name_en"=>"Ligonier", "lang"=>"en", "lat"=>40.244446, "lon"=>-79.237313),
"/US/US-PA/lilly-18291" => array("type"=>"borough", "pop"=>861, "name_ru"=>"Лилли", "name_en"=>"Lilly", "lang"=>"en", "lat"=>40.423931, "lon"=>-78.619985),
"/US/US-PA/lima-18292" => array("type"=>"cdp", "pop"=>2670, "name_ru"=>"Лима", "name_en"=>"Lima", "lang"=>"en", "lat"=>39.916710, "lon"=>-75.442319),
"/US/US-PA/lime-ridge-18293" => array("type"=>"cdp", "pop"=>1051, "name_en"=>"Lime Ridge", "lang"=>"en", "lat"=>41.023618, "lon"=>-76.355271),
"/US/US-PA/lincoln-18294" => array("type"=>"borough", "pop"=>1120, "name_ru"=>"Линкольн", "name_en"=>"Lincoln", "lang"=>"en", "lat"=>40.293215, "lon"=>-79.852462),
"/US/US-PA/linesville-18295" => array("type"=>"borough", "pop"=>1091, "name_en"=>"Linesville", "lang"=>"en", "lat"=>41.656825, "lon"=>-80.421767),
"/US/US-PA/linglestown-18296" => array("type"=>"cdp", "pop"=>5862, "name_en"=>"Linglestown", "lang"=>"en", "lat"=>40.343150, "lon"=>-76.793999),
"/US/US-PA/linntown-18297" => array("type"=>"cdp", "pop"=>1640, "name_en"=>"Linntown", "lang"=>"en", "lat"=>40.956853, "lon"=>-76.900886),
"/US/US-PA/linwood-18298" => array("type"=>"cdp", "pop"=>3425, "name_ru"=>"Линвуд", "name_en"=>"Linwood", "lang"=>"en", "lat"=>39.823417, "lon"=>-75.424507),
"/US/US-PA/lionville-marchwood-18299" => array("type"=>"cdp", "pop"=>6468, "name_en"=>"Lionville-Marchwood", "lang"=>"en", "lat"=>40.052728, "lon"=>-75.645585),
"/US/US-PA/lititz-18300" => array("type"=>"borough", "pop"=>9034, "name_en"=>"Lititz", "lang"=>"en", "lat"=>40.153738, "lon"=>-76.305327),
"/US/US-PA/little-meadows-18301" => array("type"=>"borough", "pop"=>282, "name_en"=>"Little Meadows", "lang"=>"en", "lat"=>41.993125, "lon"=>-76.130860),
"/US/US-PA/littlestown-18302" => array("type"=>"borough", "pop"=>4115, "name_en"=>"Littlestown", "lang"=>"en", "lat"=>39.744100, "lon"=>-77.088719),
"/US/US-PA/liverpool-18303" => array("type"=>"borough", "pop"=>884, "name_ru"=>"Ливерпул", "name_en"=>"Liverpool", "lang"=>"en", "lat"=>40.572375, "lon"=>-76.992773),
"/US/US-PA/lock-haven-18304" => array("type"=>"city", "pop"=>8541, "name_en"=>"Lock Haven", "lang"=>"en", "lat"=>41.135939, "lon"=>-77.453217),
"/US/US-PA/loganton-18305" => array("type"=>"borough", "pop"=>462, "name_en"=>"Loganton", "lang"=>"en", "lat"=>41.033738, "lon"=>-77.302431),
"/US/US-PA/loganville-18306" => array("type"=>"borough", "pop"=>1023, "name_ru"=>"Логанвилл", "name_en"=>"Loganville", "lang"=>"en", "lat"=>39.855056, "lon"=>-76.708768),
"/US/US-PA/long-branch-18307" => array("type"=>"borough", "pop"=>505, "name_en"=>"Long Branch", "lang"=>"en", "lat"=>40.102659, "lon"=>-79.878919),
"/US/US-PA/lorain-18308" => array("type"=>"borough", "pop"=>677, "name_en"=>"Lorain", "lang"=>"en", "lat"=>40.295767, "lon"=>-78.895259),
"/US/US-PA/lorane-18309" => array("type"=>"cdp", "pop"=>2580, "name_en"=>"Lorane", "lang"=>"en", "lat"=>40.293633, "lon"=>-75.846287),
"/US/US-PA/loretto-18310" => array("type"=>"borough", "pop"=>1338, "name_ru"=>"Лоретто", "name_en"=>"Loretto", "lang"=>"en", "lat"=>40.508675, "lon"=>-78.636257),
"/US/US-PA/lower-allen-18311" => array("type"=>"cdp", "pop"=>17466, "name_en"=>"Lower Allen", "lang"=>"en", "lat"=>40.226333, "lon"=>-76.900722),
"/US/US-PA/lower-burrell-18312" => array("type"=>"city", "pop"=>12087, "name_en"=>"Lower Burrell", "lang"=>"en", "lat"=>40.582400, "lon"=>-79.713050),
"/US/US-PA/lucerne-mines-18313" => array("type"=>"cdp", "pop"=>1074, "name_en"=>"Lucerne Mines", "lang"=>"en", "lat"=>40.554380, "lon"=>-79.155828),
"/US/US-PA/lumber-city-18314" => array("type"=>"borough", "pop"=>81, "name_ru"=>"Ламбер Сити", "name_en"=>"Lumber City", "lang"=>"en", "lat"=>40.928869, "lon"=>-78.572110),
"/US/US-PA/luzerne-18315" => array("type"=>"borough", "pop"=>2751, "name_ru"=>"Лузерн", "name_en"=>"Luzerne", "lang"=>"en", "lat"=>41.285967, "lon"=>-75.897238),
"/US/US-PA/lykens-18316" => array("type"=>"borough", "pop"=>1848, "name_en"=>"Lykens", "lang"=>"en", "lat"=>40.563263, "lon"=>-76.697908),
"/US/US-PA/lynnwood-pricedale-18317" => array("type"=>"cdp", "pop"=>2664, "name_en"=>"Lynnwood-Pricedale", "lang"=>"en", "lat"=>40.129784, "lon"=>-79.852260),
"/US/US-PA/lyons-18318" => array("type"=>"borough", "pop"=>510, "name_ru"=>"Лайонс", "name_en"=>"Lyons", "lang"=>"en", "lat"=>40.480950, "lon"=>-75.760214),
"/US/US-PA/macungie-18332" => array("type"=>"borough", "pop"=>3123, "name_en"=>"Macungie", "lang"=>"en", "lat"=>40.515816, "lon"=>-75.554326),
"/US/US-PA/madison-18334" => array("type"=>"borough", "pop"=>478, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>40.246576, "lon"=>-79.672926),
"/US/US-PA/mahaffey-18335" => array("type"=>"borough", "pop"=>372, "name_en"=>"Mahaffey", "lang"=>"en", "lat"=>40.875222, "lon"=>-78.728120),
"/US/US-PA/mahanoy-city-18336" => array("type"=>"borough", "pop"=>4375, "name_en"=>"Mahanoy City", "lang"=>"en", "lat"=>40.812736, "lon"=>-76.138467),
"/US/US-PA/malvern-18337" => array("type"=>"borough", "pop"=>3098, "name_ru"=>"Малверн", "name_en"=>"Malvern", "lang"=>"en", "lat"=>40.033450, "lon"=>-75.514023),
"/US/US-PA/manchester-18338" => array("type"=>"borough", "pop"=>2456, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>40.061272, "lon"=>-76.720325),
"/US/US-PA/manheim-18339" => array("type"=>"borough", "pop"=>4646, "name_en"=>"Manheim", "lang"=>"en", "lat"=>40.163267, "lon"=>-76.396491),
"/US/US-PA/manns-choice-18340" => array("type"=>"borough", "pop"=>277, "name_en"=>"Manns Choice", "lang"=>"en", "lat"=>40.002756, "lon"=>-78.591699),
"/US/US-PA/manor-18341" => array("type"=>"borough", "pop"=>2905, "name_en"=>"Manor", "lang"=>"en", "lat"=>40.346544, "lon"=>-79.670762),
"/US/US-PA/manorville-18342" => array("type"=>"borough", "pop"=>369, "name_en"=>"Manorville", "lang"=>"en", "lat"=>40.787693, "lon"=>-79.521070),
"/US/US-PA/mansfield-18343" => array("type"=>"borough", "pop"=>3177, "name_ru"=>"Мансфилд", "name_en"=>"Mansfield", "lang"=>"en", "lat"=>41.805690, "lon"=>-77.078716),
"/US/US-PA/maple-glen-18344" => array("type"=>"cdp", "pop"=>5881, "name_en"=>"Maple Glen", "lang"=>"en", "lat"=>40.175250, "lon"=>-75.180513),
"/US/US-PA/mapleton-18345" => array("type"=>"borough", "pop"=>437, "name_ru"=>"Мейплтон", "name_en"=>"Mapleton", "lang"=>"en", "lat"=>40.392928, "lon"=>-77.939453),
"/US/US-PA/marcus-hook-18346" => array("type"=>"borough", "pop"=>2228, "name_en"=>"Marcus Hook", "lang"=>"en", "lat"=>39.812750, "lon"=>-75.416373),
"/US/US-PA/marianna-18347" => array("type"=>"borough", "pop"=>579, "name_ru"=>"Марианна", "name_en"=>"Marianna", "lang"=>"en", "lat"=>40.011250, "lon"=>-80.114433),
"/US/US-PA/marietta-18348" => array("type"=>"borough", "pop"=>2593, "name_ru"=>"Мариетта", "name_en"=>"Marietta", "lang"=>"en", "lat"=>40.057000, "lon"=>-76.552949),
"/US/US-PA/marion-center-18349" => array("type"=>"borough", "pop"=>416, "name_ru"=>"Мэрион Сентер", "name_en"=>"Marion Center", "lang"=>"en", "lat"=>40.770233, "lon"=>-79.047988),
"/US/US-PA/marion-heights-18350" => array("type"=>"borough", "pop"=>684, "name_en"=>"Marion Heights", "lang"=>"en", "lat"=>40.803199, "lon"=>-76.465217),
"/US/US-PA/marklesburg-18351" => array("type"=>"borough", "pop"=>211, "name_en"=>"Marklesburg", "lang"=>"en", "lat"=>40.383977, "lon"=>-78.169279),
"/US/US-PA/markleysburg-18352" => array("type"=>"borough", "pop"=>268, "name_en"=>"Markleysburg", "lang"=>"en", "lat"=>39.737101, "lon"=>-79.452886),
"/US/US-PA/mars-18353" => array("type"=>"borough", "pop"=>1653, "name_en"=>"Mars", "lang"=>"en", "lat"=>40.696632, "lon"=>-80.014090),
"/US/US-PA/marshallton-18354" => array("type"=>"cdp", "pop"=>1482, "name_en"=>"Marshallton", "lang"=>"en", "lat"=>40.785542, "lon"=>-76.534646),
"/US/US-PA/martinsburg-18355" => array("type"=>"borough", "pop"=>2121, "name_ru"=>"Мартинсбург", "name_en"=>"Martinsburg", "lang"=>"en", "lat"=>40.310264, "lon"=>-78.324461),
"/US/US-PA/marysville-18356" => array("type"=>"borough", "pop"=>2425, "name_ru"=>"Мэрисвилл", "name_en"=>"Marysville", "lang"=>"en", "lat"=>40.337600, "lon"=>-76.932583),
"/US/US-PA/masontown-18357" => array("type"=>"borough", "pop"=>3394, "name_en"=>"Masontown", "lang"=>"en", "lat"=>39.848252, "lon"=>-79.908650),
"/US/US-PA/matamoras-18358" => array("type"=>"borough", "pop"=>2612, "name_en"=>"Matamoras", "lang"=>"en", "lat"=>41.366421, "lon"=>-74.700054),
"/US/US-PA/mayfield-18359" => array("type"=>"borough", "pop"=>1703, "name_ru"=>"Мейфилд", "name_en"=>"Mayfield", "lang"=>"en", "lat"=>41.537000, "lon"=>-75.533729),
"/US/US-PA/maytown-18360" => array("type"=>"cdp", "pop"=>1720, "name_ru"=>"Мейтаун", "name_en"=>"Maytown", "lang"=>"en", "lat"=>40.076972, "lon"=>-76.580417),
"/US/US-PA/mcadoo-18319" => array("type"=>"borough", "pop"=>2088, "name_en"=>"McAdoo", "lang"=>"en", "lat"=>40.900840, "lon"=>-75.992380),
"/US/US-PA/mccandless-township-18320" => array("type"=>"cdp", "pop"=>28781, "name_en"=>"McCandless Township", "lang"=>"en", "lat"=>40.583480, "lon"=>-80.029965),
"/US/US-PA/mcchesneytown-loyalhanna-18321" => array("type"=>"cdp", "pop"=>3708, "name_en"=>"McChesneytown-Loyalhanna", "lang"=>"en", "lat"=>40.312800, "lon"=>-79.356007),
"/US/US-PA/mcclure-18322" => array("type"=>"borough", "pop"=>925, "name_en"=>"McClure", "lang"=>"en", "lat"=>40.707637, "lon"=>-77.311326),
"/US/US-PA/mcconnellsburg-18323" => array("type"=>"borough", "pop"=>1040, "name_en"=>"McConnellsburg", "lang"=>"en", "lat"=>39.931525, "lon"=>-77.995280),
"/US/US-PA/mcdonald-18324" => array("type"=>"borough", "pop"=>2120, "name_ru"=>"Макдональд", "name_en"=>"McDonald", "lang"=>"en", "lat"=>40.370160, "lon"=>-80.232501),
"/US/US-PA/mcewensville-18325" => array("type"=>"borough", "pop"=>298, "name_en"=>"McEwensville", "lang"=>"en", "lat"=>41.072055, "lon"=>-76.819035),
"/US/US-PA/mcgovern-18326" => array("type"=>"cdp", "pop"=>2504, "name_en"=>"McGovern", "lang"=>"en", "lat"=>40.240312, "lon"=>-80.229079),
"/US/US-PA/mckean-18327" => array("type"=>"borough", "pop"=>359, "name_en"=>"McKean", "lang"=>"en", "lat"=>41.999056, "lon"=>-80.138294),
"/US/US-PA/mckees-rocks-18329" => array("type"=>"borough", "pop"=>6018, "name_en"=>"McKees Rocks", "lang"=>"en", "lat"=>40.469300, "lon"=>-80.062132),
"/US/US-PA/mckeesport-18328" => array("type"=>"city", "pop"=>22130, "name_en"=>"McKeesport", "lang"=>"en", "lat"=>40.341454, "lon"=>-79.844339),
"/US/US-PA/mcmurray-18330" => array("type"=>"cdp", "pop"=>4082, "name_en"=>"McMurray", "lang"=>"en", "lat"=>40.280795, "lon"=>-80.088328),
"/US/US-PA/mcsherrystown-18331" => array("type"=>"borough", "pop"=>2791, "name_en"=>"McSherrystown", "lang"=>"en", "lat"=>39.804103, "lon"=>-77.019599),
"/US/US-PA/mcveytown-18333" => array("type"=>"borough", "pop"=>390, "name_en"=>"McVeytown", "lang"=>"en", "lat"=>40.498133, "lon"=>-77.741342),
"/US/US-PA/meadowood-18361" => array("type"=>"cdp", "pop"=>3011, "name_en"=>"Meadowood", "lang"=>"en", "lat"=>40.841977, "lon"=>-79.894030),
"/US/US-PA/meadville-18362" => array("type"=>"city", "pop"=>13233, "name_ru"=>"Мидвилл", "name_en"=>"Meadville", "lang"=>"en", "lat"=>41.646181, "lon"=>-80.146545),
"/US/US-PA/mechanicsburg-18363" => array("type"=>"borough", "pop"=>8721, "name_ru"=>"Меканиксбург", "name_en"=>"Mechanicsburg", "lang"=>"en", "lat"=>40.210100, "lon"=>-77.005488),
"/US/US-PA/mechanicsville-18364" => array("type"=>"borough", "pop"=>508, "name_ru"=>"Меканиксвилл", "name_en"=>"Mechanicsville", "lang"=>"en", "lat"=>40.689588, "lon"=>-76.181544),
"/US/US-PA/mechanicsville-18365" => array("type"=>"cdp", "pop"=>2803, "name_ru"=>"Меканиксвилл", "name_en"=>"Mechanicsville", "lang"=>"en", "lat"=>40.966442, "lon"=>-76.585477),
"/US/US-PA/media-18366" => array("type"=>"borough", "pop"=>5396, "name_ru"=>"Медиа", "name_en"=>"Media", "lang"=>"en", "lat"=>39.919533, "lon"=>-75.388837),
"/US/US-PA/mercer-18367" => array("type"=>"borough", "pop"=>2207, "name_ru"=>"Мерсер", "name_en"=>"Mercer", "lang"=>"en", "lat"=>41.226278, "lon"=>-80.236626),
"/US/US-PA/mercersburg-18368" => array("type"=>"borough", "pop"=>1584, "name_en"=>"Mercersburg", "lang"=>"en", "lat"=>39.831364, "lon"=>-77.901811),
"/US/US-PA/meridian-18369" => array("type"=>"cdp", "pop"=>3473, "name_ru"=>"Меридиан", "name_en"=>"Meridian", "lang"=>"en", "lat"=>40.854351, "lon"=>-79.956580),
"/US/US-PA/meshoppen-18370" => array("type"=>"borough", "pop"=>427, "name_en"=>"Meshoppen", "lang"=>"en", "lat"=>41.613585, "lon"=>-76.045862),
"/US/US-PA/meyersdale-18371" => array("type"=>"borough", "pop"=>2273, "name_en"=>"Meyersdale", "lang"=>"en", "lat"=>39.813077, "lon"=>-79.026759),
"/US/US-PA/middleburg-18372" => array("type"=>"borough", "pop"=>1327, "name_ru"=>"Миддлберг", "name_en"=>"Middleburg", "lang"=>"en", "lat"=>40.789034, "lon"=>-77.045896),
"/US/US-PA/middleport-18373" => array("type"=>"borough", "pop"=>433, "name_en"=>"Middleport", "lang"=>"en", "lat"=>40.727264, "lon"=>-76.087151),
"/US/US-PA/middletown-18374" => array("type"=>"borough", "pop"=>8841, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>40.200900, "lon"=>-76.729349),
"/US/US-PA/middletown-18375" => array("type"=>"cdp", "pop"=>6866, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>40.643700, "lon"=>-75.325772),
"/US/US-PA/midland-18376" => array("type"=>"borough", "pop"=>2846, "name_ru"=>"Мидленд", "name_en"=>"Midland", "lang"=>"en", "lat"=>40.639856, "lon"=>-80.457914),
"/US/US-PA/midway-18377" => array("type"=>"borough", "pop"=>924, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>40.368553, "lon"=>-80.291736),
"/US/US-PA/midway-18378" => array("type"=>"cdp", "pop"=>2254, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>39.803914, "lon"=>-77.005268),
"/US/US-PA/mifflin-18379" => array("type"=>"borough", "pop"=>605, "name_en"=>"Mifflin", "lang"=>"en", "lat"=>40.567803, "lon"=>-77.404773),
"/US/US-PA/mifflinburg-18380" => array("type"=>"borough", "pop"=>3517, "name_en"=>"Mifflinburg", "lang"=>"en", "lat"=>40.920084, "lon"=>-77.047111),
"/US/US-PA/mifflintown-18381" => array("type"=>"borough", "pop"=>822, "name_en"=>"Mifflintown", "lang"=>"en", "lat"=>40.571107, "lon"=>-77.395090),
"/US/US-PA/mifflinville-18382" => array("type"=>"cdp", "pop"=>1329, "name_en"=>"Mifflinville", "lang"=>"en", "lat"=>41.029184, "lon"=>-76.300573),
"/US/US-PA/milesburg-18383" => array("type"=>"borough", "pop"=>1144, "name_en"=>"Milesburg", "lang"=>"en", "lat"=>40.942628, "lon"=>-77.791901),
"/US/US-PA/milford-18384" => array("type"=>"borough", "pop"=>1204, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>41.323990, "lon"=>-74.801894),
"/US/US-PA/mill-creek-18386" => array("type"=>"borough", "pop"=>327, "name_ru"=>"Милл Крик", "name_en"=>"Mill Creek", "lang"=>"en", "lat"=>40.437380, "lon"=>-77.929416),
"/US/US-PA/mill-hall-18390" => array("type"=>"borough", "pop"=>1478, "name_en"=>"Mill Hall", "lang"=>"en", "lat"=>41.104439, "lon"=>-77.490368),
"/US/US-PA/mill-village-18393" => array("type"=>"borough", "pop"=>386, "name_en"=>"Mill Village", "lang"=>"en", "lat"=>41.877662, "lon"=>-79.972604),
"/US/US-PA/millbourne-18385" => array("type"=>"borough", "pop"=>905, "name_en"=>"Millbourne", "lang"=>"en", "lat"=>39.963150, "lon"=>-75.252773),
"/US/US-PA/millersburg-18387" => array("type"=>"borough", "pop"=>2460, "name_ru"=>"Миллерсберг", "name_en"=>"Millersburg", "lang"=>"en", "lat"=>40.542423, "lon"=>-76.955987),
"/US/US-PA/millerstown-18388" => array("type"=>"borough", "pop"=>683, "name_en"=>"Millerstown", "lang"=>"en", "lat"=>40.552783, "lon"=>-77.152811),
"/US/US-PA/millersville-18389" => array("type"=>"borough", "pop"=>7341, "name_en"=>"Millersville", "lang"=>"en", "lat"=>40.003584, "lon"=>-76.353857),
"/US/US-PA/millheim-18391" => array("type"=>"borough", "pop"=>747, "name_en"=>"Millheim", "lang"=>"en", "lat"=>40.892514, "lon"=>-77.476162),
"/US/US-PA/millvale-18392" => array("type"=>"borough", "pop"=>3656, "name_en"=>"Millvale", "lang"=>"en", "lat"=>40.482054, "lon"=>-79.974449),
"/US/US-PA/millville-18394" => array("type"=>"borough", "pop"=>944, "name_ru"=>"Миллвилл", "name_en"=>"Millville", "lang"=>"en", "lat"=>41.121236, "lon"=>-76.527435),
"/US/US-PA/milroy-18395" => array("type"=>"cdp", "pop"=>1456, "name_ru"=>"Милрой", "name_en"=>"Milroy", "lang"=>"en", "lat"=>40.714835, "lon"=>-77.586895),
"/US/US-PA/milton-18396" => array("type"=>"borough", "pop"=>6359, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>41.008014, "lon"=>-76.850574),
"/US/US-PA/minersville-18397" => array("type"=>"borough", "pop"=>4221, "name_en"=>"Minersville", "lang"=>"en", "lat"=>40.690765, "lon"=>-76.259336),
"/US/US-PA/modena-18398" => array("type"=>"borough", "pop"=>598, "name_en"=>"Modena", "lang"=>"en", "lat"=>39.962150, "lon"=>-75.802489),
"/US/US-PA/mohnton-18399" => array("type"=>"borough", "pop"=>3070, "name_en"=>"Mohnton", "lang"=>"en", "lat"=>40.286533, "lon"=>-75.987668),
"/US/US-PA/monaca-18400" => array("type"=>"borough", "pop"=>5749, "name_en"=>"Monaca", "lang"=>"en", "lat"=>40.683350, "lon"=>-80.273240),
"/US/US-PA/monessen-18401" => array("type"=>"city", "pop"=>8024, "name_en"=>"Monessen", "lang"=>"en", "lat"=>40.152224, "lon"=>-79.882204),
"/US/US-PA/monongahela-18402" => array("type"=>"city", "pop"=>4422, "name_en"=>"Monongahela", "lang"=>"en", "lat"=>40.198344, "lon"=>-79.922622),
"/US/US-PA/monroe-18403" => array("type"=>"borough", "pop"=>480, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>41.713597, "lon"=>-76.475631),
"/US/US-PA/mont-alto-18404" => array("type"=>"borough", "pop"=>1815, "name_en"=>"Mont Alto", "lang"=>"en", "lat"=>39.839640, "lon"=>-77.554706),
"/US/US-PA/montgomery-18405" => array("type"=>"borough", "pop"=>1570, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>41.170928, "lon"=>-76.875195),
"/US/US-PA/montgomeryville-18406" => array("type"=>"cdp", "pop"=>9114, "name_en"=>"Montgomeryville", "lang"=>"en", "lat"=>40.248055, "lon"=>-75.241685),
"/US/US-PA/montoursville-18407" => array("type"=>"borough", "pop"=>4558, "name_en"=>"Montoursville", "lang"=>"en", "lat"=>41.247050, "lon"=>-76.919968),
"/US/US-PA/montrose-18408" => array("type"=>"borough", "pop"=>1532, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>41.834048, "lon"=>-75.876909),
"/US/US-PA/moosic-18409" => array("type"=>"borough", "pop"=>5758, "name_en"=>"Moosic", "lang"=>"en", "lat"=>41.357000, "lon"=>-75.702329),
"/US/US-PA/morrisville-18410" => array("type"=>"borough", "pop"=>9594, "name_ru"=>"Моррисвилл", "name_en"=>"Morrisville", "lang"=>"en", "lat"=>40.207350, "lon"=>-74.780312),
"/US/US-PA/morrisville-18411" => array("type"=>"cdp", "pop"=>1365, "name_ru"=>"Моррисвилл", "name_en"=>"Morrisville", "lang"=>"en", "lat"=>39.901916, "lon"=>-80.166670),
"/US/US-PA/morton-18412" => array("type"=>"borough", "pop"=>2634, "name_ru"=>"Мортон", "name_en"=>"Morton", "lang"=>"en", "lat"=>39.910900, "lon"=>-75.326958),
"/US/US-PA/moscow-18413" => array("type"=>"borough", "pop"=>1938, "name_ru"=>"Москва", "name_en"=>"Moscow", "lang"=>"en", "lat"=>41.340320, "lon"=>-75.529113),
"/US/US-PA/mount-carbon-18415" => array("type"=>"borough", "pop"=>82, "name_en"=>"Mount Carbon", "lang"=>"en", "lat"=>40.674162, "lon"=>-76.187970),
"/US/US-PA/mount-carmel-18416" => array("type"=>"borough", "pop"=>5888, "name_ru"=>"Маунт Кармел", "name_en"=>"Mount Carmel", "lang"=>"en", "lat"=>40.795287, "lon"=>-76.412264),
"/US/US-PA/mount-cobb-18417" => array("type"=>"cdp", "pop"=>2043, "name_en"=>"Mount Cobb", "lang"=>"en", "lat"=>41.428558, "lon"=>-75.497237),
"/US/US-PA/mount-gretna-18418" => array("type"=>"borough", "pop"=>234, "name_en"=>"Mount Gretna", "lang"=>"en", "lat"=>40.245367, "lon"=>-76.471556),
"/US/US-PA/mount-holly-springs-18419" => array("type"=>"borough", "pop"=>1909, "name_en"=>"Mount Holly Springs", "lang"=>"en", "lat"=>40.112662, "lon"=>-77.184822),
"/US/US-PA/mount-jewett-18420" => array("type"=>"borough", "pop"=>995, "name_en"=>"Mount Jewett", "lang"=>"en", "lat"=>41.724726, "lon"=>-78.643954),
"/US/US-PA/mount-joy-18421" => array("type"=>"borough", "pop"=>7249, "name_en"=>"Mount Joy", "lang"=>"en", "lat"=>40.110450, "lon"=>-76.507858),
"/US/US-PA/mount-lebanon-18422" => array("type"=>"cdp", "pop"=>30381, "name_ru"=>"Маунт Лебанон", "name_en"=>"Mount Lebanon", "lang"=>"en", "lat"=>40.374000, "lon"=>-80.048859),
"/US/US-PA/mount-oliver-18423" => array("type"=>"borough", "pop"=>3653, "name_en"=>"Mount Oliver", "lang"=>"en", "lat"=>40.410950, "lon"=>-79.986427),
"/US/US-PA/mount-penn-18424" => array("type"=>"borough", "pop"=>3063, "name_en"=>"Mount Penn", "lang"=>"en", "lat"=>40.328550, "lon"=>-75.889178),
"/US/US-PA/mount-pleasant-18425" => array("type"=>"borough", "pop"=>4380, "name_ru"=>"Маунт Плезант", "name_en"=>"Mount Pleasant", "lang"=>"en", "lat"=>40.150368, "lon"=>-79.544090),
"/US/US-PA/mount-pocono-18426" => array("type"=>"borough", "pop"=>2934, "name_en"=>"Mount Pocono", "lang"=>"en", "lat"=>41.122451, "lon"=>-75.357942),
"/US/US-PA/mount-union-18427" => array("type"=>"borough", "pop"=>2344, "name_ru"=>"Маунт Юнион", "name_en"=>"Mount Union", "lang"=>"en", "lat"=>40.384935, "lon"=>-77.881973),
"/US/US-PA/mount-wolf-18429" => array("type"=>"borough", "pop"=>1331, "name_en"=>"Mount Wolf", "lang"=>"en", "lat"=>40.060216, "lon"=>-76.705331),
"/US/US-PA/mountainhome-18414" => array("type"=>"cdp", "pop"=>1042, "name_en"=>"Mountainhome", "lang"=>"en", "lat"=>41.176582, "lon"=>-75.263245),
"/US/US-PA/mountville-18428" => array("type"=>"borough", "pop"=>2820, "name_en"=>"Mountville", "lang"=>"en", "lat"=>40.040650, "lon"=>-76.434428),
"/US/US-PA/muncy-18430" => array("type"=>"borough", "pop"=>2459, "name_en"=>"Muncy", "lang"=>"en", "lat"=>41.201928, "lon"=>-76.785867),
"/US/US-PA/munhall-18431" => array("type"=>"borough", "pop"=>11199, "name_en"=>"Munhall", "lang"=>"en", "lat"=>40.393350, "lon"=>-79.901145),
"/US/US-PA/municipality-of-monroeville-18432" => array("type"=>"borough", "pop"=>29169, "name_en"=>"Municipality of Monroeville", "lang"=>"en", "lat"=>40.426004, "lon"=>-79.760948),
"/US/US-PA/municipality-of-murrysville-18433" => array("type"=>"borough", "pop"=>17240, "name_en"=>"Municipality of Murrysville", "lang"=>"en", "lat"=>40.441525, "lon"=>-79.655872),
"/US/US-PA/myerstown-18434" => array("type"=>"borough", "pop"=>3120, "name_en"=>"Myerstown", "lang"=>"en", "lat"=>40.371653, "lon"=>-76.305438),
"/US/US-PA/nanticoke-18435" => array("type"=>"city", "pop"=>10199, "name_en"=>"Nanticoke", "lang"=>"en", "lat"=>41.200272, "lon"=>-76.000011),
"/US/US-PA/nanty-glo-18436" => array("type"=>"borough", "pop"=>2793, "name_en"=>"Nanty-Glo", "lang"=>"en", "lat"=>40.470290, "lon"=>-78.834942),
"/US/US-PA/narberth-18437" => array("type"=>"borough", "pop"=>4040, "name_en"=>"Narberth", "lang"=>"en", "lat"=>40.007850, "lon"=>-75.263839),
"/US/US-PA/nazareth-18438" => array("type"=>"borough", "pop"=>6020, "name_en"=>"Nazareth", "lang"=>"en", "lat"=>40.739888, "lon"=>-75.313015),
"/US/US-PA/nemacolin-18439" => array("type"=>"cdp", "pop"=>1097, "name_en"=>"Nemacolin", "lang"=>"en", "lat"=>39.877608, "lon"=>-79.924159),
"/US/US-PA/nescopeck-18440" => array("type"=>"borough", "pop"=>1427, "name_en"=>"Nescopeck", "lang"=>"en", "lat"=>41.052699, "lon"=>-76.212537),
"/US/US-PA/nesquehoning-18441" => array("type"=>"borough", "pop"=>3303, "name_en"=>"Nesquehoning", "lang"=>"en", "lat"=>40.862028, "lon"=>-75.829536),
"/US/US-PA/nether-providence-township-18442" => array("type"=>"cdp", "pop"=>13229, "name_en"=>"Nether Providence Township", "lang"=>"en", "lat"=>39.897933, "lon"=>-75.368182),
"/US/US-PA/new-albany-18443" => array("type"=>"borough", "pop"=>286, "name_ru"=>"Нью Олбани", "name_en"=>"New Albany", "lang"=>"en", "lat"=>41.600810, "lon"=>-76.443287),
"/US/US-PA/new-alexandria-18444" => array("type"=>"borough", "pop"=>567, "name_en"=>"New Alexandria", "lang"=>"en", "lat"=>40.393486, "lon"=>-79.418362),
"/US/US-PA/new-baltimore-18445" => array("type"=>"borough", "pop"=>153, "name_ru"=>"Нью Балтимор", "name_en"=>"New Baltimore", "lang"=>"en", "lat"=>39.985753, "lon"=>-78.771996),
"/US/US-PA/new-beaver-18446" => array("type"=>"borough", "pop"=>1579, "name_en"=>"New Beaver", "lang"=>"en", "lat"=>40.882415, "lon"=>-80.364612),
"/US/US-PA/new-berlin-18447" => array("type"=>"borough", "pop"=>818, "name_ru"=>"Нью Берлин", "name_en"=>"New Berlin", "lang"=>"en", "lat"=>40.880425, "lon"=>-76.985306),
"/US/US-PA/new-bethlehem-18448" => array("type"=>"borough", "pop"=>981, "name_en"=>"New Bethlehem", "lang"=>"en", "lat"=>41.002992, "lon"=>-79.328833),
"/US/US-PA/new-brighton-18449" => array("type"=>"borough", "pop"=>6068, "name_ru"=>"Нью Брайтон", "name_en"=>"New Brighton", "lang"=>"en", "lat"=>40.735600, "lon"=>-80.308883),
"/US/US-PA/new-britain-18450" => array("type"=>"borough", "pop"=>2265, "name_ru"=>"Нью Бритн", "name_en"=>"New Britain", "lang"=>"en", "lat"=>40.299150, "lon"=>-75.178607),
"/US/US-PA/new-buffalo-18451" => array("type"=>"borough", "pop"=>120, "name_ru"=>"Нью Буффало", "name_en"=>"New Buffalo", "lang"=>"en", "lat"=>40.454400, "lon"=>-76.970944),
"/US/US-PA/new-castle-18454" => array("type"=>"city", "pop"=>24212, "name_ru"=>"Нью Касл", "name_en"=>"New Castle", "lang"=>"en", "lat"=>40.995677, "lon"=>-80.346278),
"/US/US-PA/new-castle-northwest-18455" => array("type"=>"cdp", "pop"=>1515, "name_en"=>"New Castle Northwest", "lang"=>"en", "lat"=>41.021925, "lon"=>-80.356211),
"/US/US-PA/new-centerville-18456" => array("type"=>"borough", "pop"=>178, "name_en"=>"New Centerville", "lang"=>"en", "lat"=>39.941063, "lon"=>-79.191993),
"/US/US-PA/new-columbus-18457" => array("type"=>"borough", "pop"=>204, "name_en"=>"New Columbus", "lang"=>"en", "lat"=>41.168868, "lon"=>-76.286565),
"/US/US-PA/new-cumberland-18458" => array("type"=>"borough", "pop"=>7048, "name_en"=>"New Cumberland", "lang"=>"en", "lat"=>40.230500, "lon"=>-76.875187),
"/US/US-PA/new-eagle-18459" => array("type"=>"borough", "pop"=>2201, "name_en"=>"New Eagle", "lang"=>"en", "lat"=>40.206176, "lon"=>-79.954532),
"/US/US-PA/new-florence-18461" => array("type"=>"borough", "pop"=>726, "name_ru"=>"Нью Флоренс", "name_en"=>"New Florence", "lang"=>"en", "lat"=>40.379070, "lon"=>-79.074996),
"/US/US-PA/new-freedom-18462" => array("type"=>"borough", "pop"=>4205, "name_en"=>"New Freedom", "lang"=>"en", "lat"=>39.736300, "lon"=>-76.696807),
"/US/US-PA/new-galilee-18463" => array("type"=>"borough", "pop"=>385, "name_en"=>"New Galilee", "lang"=>"en", "lat"=>40.833104, "lon"=>-80.400868),
"/US/US-PA/new-holland-18464" => array("type"=>"borough", "pop"=>5152, "name_ru"=>"Нью Холланд", "name_en"=>"New Holland", "lang"=>"en", "lat"=>40.101013, "lon"=>-76.090395),
"/US/US-PA/new-hope-18465" => array("type"=>"borough", "pop"=>2278, "name_ru"=>"Нью Хоуп", "name_en"=>"New Hope", "lang"=>"en", "lat"=>40.361419, "lon"=>-74.959310),
"/US/US-PA/new-kensington-18466" => array("type"=>"city", "pop"=>13667, "name_en"=>"New Kensington", "lang"=>"en", "lat"=>40.569892, "lon"=>-79.750816),
"/US/US-PA/new-lebanon-18467" => array("type"=>"borough", "pop"=>195, "name_en"=>"New Lebanon", "lang"=>"en", "lat"=>41.416743, "lon"=>-80.075887),
"/US/US-PA/new-milford-18469" => array("type"=>"borough", "pop"=>806, "name_ru"=>"Нью Милфорд", "name_en"=>"New Milford", "lang"=>"en", "lat"=>41.875883, "lon"=>-75.726893),
"/US/US-PA/new-oxford-18470" => array("type"=>"borough", "pop"=>1764, "name_en"=>"New Oxford", "lang"=>"en", "lat"=>39.862216, "lon"=>-77.055706),
"/US/US-PA/new-paris-18471" => array("type"=>"borough", "pop"=>201, "name_ru"=>"Нью Парис", "name_en"=>"New Paris", "lang"=>"en", "lat"=>40.107728, "lon"=>-78.643170),
"/US/US-PA/new-philadelphia-18472" => array("type"=>"borough", "pop"=>1089, "name_en"=>"New Philadelphia", "lang"=>"en", "lat"=>40.717767, "lon"=>-76.118046),
"/US/US-PA/new-ringgold-18474" => array("type"=>"borough", "pop"=>278, "name_en"=>"New Ringgold", "lang"=>"en", "lat"=>40.688484, "lon"=>-75.993835),
"/US/US-PA/new-salem-18476" => array("type"=>"borough", "pop"=>641, "name_ru"=>"Нью Салем", "name_en"=>"New Salem", "lang"=>"en", "lat"=>39.902427, "lon"=>-76.793495),
"/US/US-PA/new-salem-buffington-18477" => array("type"=>"cdp", "pop"=>1169, "name_en"=>"New Salem-Buffington", "lang"=>"en", "lat"=>39.931888, "lon"=>-79.820631),
"/US/US-PA/new-stanton-18478" => array("type"=>"borough", "pop"=>2099, "name_en"=>"New Stanton", "lang"=>"en", "lat"=>40.221653, "lon"=>-79.609013),
"/US/US-PA/new-washington-18483" => array("type"=>"borough", "pop"=>82, "name_en"=>"New Washington", "lang"=>"en", "lat"=>40.822676, "lon"=>-78.701409),
"/US/US-PA/new-wilmington-18484" => array("type"=>"borough", "pop"=>2253, "name_en"=>"New Wilmington", "lang"=>"en", "lat"=>41.117565, "lon"=>-80.332178),
"/US/US-PA/newburg-18452" => array("type"=>"borough", "pop"=>76, "name_ru"=>"Ньюбург", "name_en"=>"Newburg", "lang"=>"en", "lat"=>40.839163, "lon"=>-78.687111),
"/US/US-PA/newburg-18453" => array("type"=>"borough", "pop"=>312, "name_ru"=>"Ньюбург", "name_en"=>"Newburg", "lang"=>"en", "lat"=>40.137726, "lon"=>-77.554500),
"/US/US-PA/newell-18460" => array("type"=>"borough", "pop"=>517, "name_ru"=>"Ньюуэлл", "name_en"=>"Newell", "lang"=>"en", "lat"=>40.073544, "lon"=>-79.888620),
"/US/US-PA/newmanstown-18468" => array("type"=>"cdp", "pop"=>1410, "name_en"=>"Newmanstown", "lang"=>"en", "lat"=>40.350794, "lon"=>-76.212894),
"/US/US-PA/newport-18473" => array("type"=>"borough", "pop"=>1463, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>40.478277, "lon"=>-77.134115),
"/US/US-PA/newry-18475" => array("type"=>"borough", "pop"=>231, "name_en"=>"Newry", "lang"=>"en", "lat"=>40.393150, "lon"=>-78.435349),
"/US/US-PA/newton-hamilton-18479" => array("type"=>"borough", "pop"=>257, "name_en"=>"Newton Hamilton", "lang"=>"en", "lat"=>40.392754, "lon"=>-77.835363),
"/US/US-PA/newtown-18480" => array("type"=>"borough", "pop"=>2376, "name_ru"=>"Ньютаун", "name_en"=>"Newtown", "lang"=>"en", "lat"=>40.228100, "lon"=>-74.932728),
"/US/US-PA/newtown-grant-18481" => array("type"=>"cdp", "pop"=>2141, "name_en"=>"Newtown Grant", "lang"=>"en", "lat"=>40.259997, "lon"=>-74.955265),
"/US/US-PA/newville-18482" => array("type"=>"borough", "pop"=>1303, "name_ru"=>"Ньювилл", "name_en"=>"Newville", "lang"=>"en", "lat"=>40.170872, "lon"=>-77.401923),
"/US/US-PA/nicholson-18485" => array("type"=>"borough", "pop"=>658, "name_ru"=>"Николсон", "name_en"=>"Nicholson", "lang"=>"en", "lat"=>41.627403, "lon"=>-75.785747),
"/US/US-PA/nixon-18486" => array("type"=>"cdp", "pop"=>1342, "name_en"=>"Nixon", "lang"=>"en", "lat"=>40.783230, "lon"=>-79.929732),
"/US/US-PA/norristown-18487" => array("type"=>"borough", "pop"=>31843, "name_en"=>"Norristown", "lang"=>"en", "lat"=>40.122150, "lon"=>-75.339727),
"/US/US-PA/north-apollo-18489" => array("type"=>"borough", "pop"=>1324, "name_en"=>"North Apollo", "lang"=>"en", "lat"=>40.593208, "lon"=>-79.557105),
"/US/US-PA/north-belle-vernon-18490" => array("type"=>"borough", "pop"=>1949, "name_en"=>"North Belle Vernon", "lang"=>"en", "lat"=>40.131976, "lon"=>-79.865091),
"/US/US-PA/north-braddock-18491" => array("type"=>"borough", "pop"=>5821, "name_en"=>"North Braddock", "lang"=>"en", "lat"=>40.400950, "lon"=>-79.852054),
"/US/US-PA/north-catasauqua-18492" => array("type"=>"borough", "pop"=>2834, "name_en"=>"North Catasauqua", "lang"=>"en", "lat"=>40.664350, "lon"=>-75.474305),
"/US/US-PA/north-charleroi-18493" => array("type"=>"borough", "pop"=>1314, "name_en"=>"North Charleroi", "lang"=>"en", "lat"=>40.150368, "lon"=>-79.908372),
"/US/US-PA/north-east-18494" => array("type"=>"borough", "pop"=>4184, "name_ru"=>"Норт Ист", "name_en"=>"North East", "lang"=>"en", "lat"=>42.213328, "lon"=>-79.833940),
"/US/US-PA/north-irwin-18495" => array("type"=>"borough", "pop"=>838, "name_en"=>"North Irwin", "lang"=>"en", "lat"=>40.338850, "lon"=>-79.711199),
"/US/US-PA/north-vandergrift-pleasant-view-18497" => array("type"=>"cdp", "pop"=>1431, "name_en"=>"North Vandergrift-Pleasant View", "lang"=>"en", "lat"=>40.608508, "lon"=>-79.564379),
"/US/US-PA/north-versailles-18498" => array("type"=>"cdp", "pop"=>10270, "name_en"=>"North Versailles", "lang"=>"en", "lat"=>40.379800, "lon"=>-79.809699),
"/US/US-PA/north-wales-18499" => array("type"=>"borough", "pop"=>3226, "name_en"=>"North Wales", "lang"=>"en", "lat"=>40.210600, "lon"=>-75.275377),
"/US/US-PA/north-york-18501" => array("type"=>"borough", "pop"=>1742, "name_en"=>"North York", "lang"=>"en", "lat"=>39.977716, "lon"=>-76.731793),
"/US/US-PA/northampton-18488" => array("type"=>"borough", "pop"=>9841, "name_ru"=>"Нортгемптон", "name_en"=>"Northampton", "lang"=>"en", "lat"=>40.688000, "lon"=>-75.488694),
"/US/US-PA/northumberland-18496" => array("type"=>"borough", "pop"=>3512, "name_en"=>"Northumberland", "lang"=>"en", "lat"=>40.897054, "lon"=>-76.793672),
"/US/US-PA/northwest-harborcreek-18500" => array("type"=>"cdp", "pop"=>6662, "name_en"=>"Northwest Harborcreek", "lang"=>"en", "lat"=>42.146203, "lon"=>-79.994120),
"/US/US-PA/norwood-18502" => array("type"=>"borough", "pop"=>5774, "name_ru"=>"Норвуд", "name_en"=>"Norwood", "lang"=>"en", "lat"=>39.885900, "lon"=>-75.295994),
"/US/US-PA/nuangola-18503" => array("type"=>"borough", "pop"=>650, "name_en"=>"Nuangola", "lang"=>"en", "lat"=>41.158250, "lon"=>-75.977982),
"/US/US-PA/o'hara-township-18510" => array("type"=>"cdp", "pop"=>9096, "name_en"=>"O'Hara Township", "lang"=>"en", "lat"=>40.499350, "lon"=>-79.901730),
"/US/US-PA/oak-hills-18505" => array("type"=>"cdp", "pop"=>2245, "name_en"=>"Oak Hills", "lang"=>"en", "lat"=>40.824744, "lon"=>-79.913429),
"/US/US-PA/oakdale-18504" => array("type"=>"borough", "pop"=>1435, "name_ru"=>"Оукдейл", "name_en"=>"Oakdale", "lang"=>"en", "lat"=>40.399680, "lon"=>-80.187732),
"/US/US-PA/oakland-18506" => array("type"=>"borough", "pop"=>572, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>41.948910, "lon"=>-75.611131),
"/US/US-PA/oakland-18507" => array("type"=>"cdp", "pop"=>1766, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>40.992235, "lon"=>-80.365958),
"/US/US-PA/oakmont-18508" => array("type"=>"borough", "pop"=>6423, "name_en"=>"Oakmont", "lang"=>"en", "lat"=>40.519400, "lon"=>-79.836497),
"/US/US-PA/oakwood-18509" => array("type"=>"cdp", "pop"=>2541, "name_ru"=>"Оуквуд", "name_en"=>"Oakwood", "lang"=>"en", "lat"=>41.010610, "lon"=>-80.379618),
"/US/US-PA/ohiopyle-18511" => array("type"=>"borough", "pop"=>74, "name_en"=>"Ohiopyle", "lang"=>"en", "lat"=>39.867511, "lon"=>-79.494173),
"/US/US-PA/ohioville-18512" => array("type"=>"borough", "pop"=>3614, "name_en"=>"Ohioville", "lang"=>"en", "lat"=>40.682885, "lon"=>-80.477275),
"/US/US-PA/oil-city-18513" => array("type"=>"city", "pop"=>10598, "name_ru"=>"Ойл Сити", "name_en"=>"Oil City", "lang"=>"en", "lat"=>41.426642, "lon"=>-79.704258),
"/US/US-PA/oklahoma-18514" => array("type"=>"borough", "pop"=>841, "name_en"=>"Oklahoma", "lang"=>"en", "lat"=>40.580923, "lon"=>-79.574650),
"/US/US-PA/old-forge-18515" => array("type"=>"borough", "pop"=>8524, "name_en"=>"Old Forge", "lang"=>"en", "lat"=>41.368950, "lon"=>-75.742349),
"/US/US-PA/old-orchard-18516" => array("type"=>"cdp", "pop"=>2598, "name_en"=>"Old Orchard", "lang"=>"en", "lat"=>40.657786, "lon"=>-75.262598),
"/US/US-PA/oliver-18517" => array("type"=>"cdp", "pop"=>1150, "name_ru"=>"Оливер", "name_en"=>"Oliver", "lang"=>"en", "lat"=>39.915182, "lon"=>-79.722057),
"/US/US-PA/olyphant-18518" => array("type"=>"borough", "pop"=>4956, "name_en"=>"Olyphant", "lang"=>"en", "lat"=>41.449250, "lon"=>-75.580174),
"/US/US-PA/orangeville-18519" => array("type"=>"borough", "pop"=>471, "name_ru"=>"Оранжвилл", "name_en"=>"Orangeville", "lang"=>"en", "lat"=>41.076630, "lon"=>-76.413090),
"/US/US-PA/orbisonia-18520" => array("type"=>"borough", "pop"=>402, "name_en"=>"Orbisonia", "lang"=>"en", "lat"=>40.243264, "lon"=>-77.893643),
"/US/US-PA/orchard-hills-18521" => array("type"=>"cdp", "pop"=>2019, "name_en"=>"Orchard Hills", "lang"=>"en", "lat"=>40.593034, "lon"=>-79.541440),
"/US/US-PA/oreland-18522" => array("type"=>"cdp", "pop"=>5695, "name_en"=>"Oreland", "lang"=>"en", "lat"=>40.114400, "lon"=>-75.180260),
"/US/US-PA/orrstown-18523" => array("type"=>"borough", "pop"=>230, "name_en"=>"Orrstown", "lang"=>"en", "lat"=>40.058291, "lon"=>-77.609462),
"/US/US-PA/orwigsburg-18524" => array("type"=>"borough", "pop"=>2979, "name_en"=>"Orwigsburg", "lang"=>"en", "lat"=>40.653700, "lon"=>-76.103670),
"/US/US-PA/osborne-18525" => array("type"=>"borough", "pop"=>522, "name_ru"=>"Осборн", "name_en"=>"Osborne", "lang"=>"en", "lat"=>40.530825, "lon"=>-80.169468),
"/US/US-PA/osceola-mills-18526" => array("type"=>"borough", "pop"=>1152, "name_en"=>"Osceola Mills", "lang"=>"en", "lat"=>40.852529, "lon"=>-78.269875),
"/US/US-PA/oswayo-18527" => array("type"=>"borough", "pop"=>140, "name_en"=>"Oswayo", "lang"=>"en", "lat"=>41.920065, "lon"=>-78.020510),
"/US/US-PA/oxford-18528" => array("type"=>"borough", "pop"=>4658, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>39.785678, "lon"=>-75.980376),
"/US/US-PA/paint-18529" => array("type"=>"borough", "pop"=>1034, "name_en"=>"Paint", "lang"=>"en", "lat"=>40.242800, "lon"=>-78.849003),
"/US/US-PA/palmer-heights-18530" => array("type"=>"cdp", "pop"=>3960, "name_en"=>"Palmer Heights", "lang"=>"en", "lat"=>40.689633, "lon"=>-75.268502),
"/US/US-PA/palmerton-18531" => array("type"=>"borough", "pop"=>5209, "name_en"=>"Palmerton", "lang"=>"en", "lat"=>40.801166, "lon"=>-75.617040),
"/US/US-PA/palmyra-18532" => array("type"=>"borough", "pop"=>6997, "name_ru"=>"Палмира", "name_en"=>"Palmyra", "lang"=>"en", "lat"=>40.309538, "lon"=>-76.595011),
"/US/US-PA/palo-alto-18533" => array("type"=>"borough", "pop"=>1106, "name_ru"=>"Пало-Альто", "name_en"=>"Palo Alto", "lang"=>"en", "lat"=>40.682832, "lon"=>-76.176755),
"/US/US-PA/paoli-18534" => array("type"=>"cdp", "pop"=>5603, "name_ru"=>"Паоли", "name_en"=>"Paoli", "lang"=>"en", "lat"=>40.042350, "lon"=>-75.492751),
"/US/US-PA/paradise-18535" => array("type"=>"cdp", "pop"=>4858, "name_ru"=>"Парадайс", "name_en"=>"Paradise", "lang"=>"en", "lat"=>40.006769, "lon"=>-76.122302),
"/US/US-PA/park-forest-village-18538" => array("type"=>"cdp", "pop"=>6703, "name_en"=>"Park Forest Village", "lang"=>"en", "lat"=>40.801000, "lon"=>-77.909699),
"/US/US-PA/parker-18536" => array("type"=>"city", "pop"=>747, "name_ru"=>"Паркер", "name_en"=>"Parker", "lang"=>"en", "lat"=>41.092037, "lon"=>-79.684268),
"/US/US-PA/parkesburg-18537" => array("type"=>"borough", "pop"=>3430, "name_en"=>"Parkesburg", "lang"=>"en", "lat"=>39.958682, "lon"=>-75.916711),
"/US/US-PA/parkside-18539" => array("type"=>"borough", "pop"=>2176, "name_en"=>"Parkside", "lang"=>"en", "lat"=>39.866450, "lon"=>-75.378671),
"/US/US-PA/parkville-18540" => array("type"=>"cdp", "pop"=>6014, "name_ru"=>"Парквилл", "name_en"=>"Parkville", "lang"=>"en", "lat"=>39.782908, "lon"=>-76.967487),
"/US/US-PA/parryville-18541" => array("type"=>"borough", "pop"=>504, "name_en"=>"Parryville", "lang"=>"en", "lat"=>40.824311, "lon"=>-75.669614),
"/US/US-PA/patterson-heights-18542" => array("type"=>"borough", "pop"=>614, "name_en"=>"Patterson Heights", "lang"=>"en", "lat"=>40.739050, "lon"=>-80.326999),
"/US/US-PA/patterson-township-18543" => array("type"=>"cdp", "pop"=>3074, "name_en"=>"Patterson Township", "lang"=>"en", "lat"=>40.742400, "lon"=>-80.332442),
"/US/US-PA/patton-18544" => array("type"=>"borough", "pop"=>1848, "name_en"=>"Patton", "lang"=>"en", "lat"=>40.632728, "lon"=>-78.649253),
"/US/US-PA/paxtang-18545" => array("type"=>"borough", "pop"=>1486, "name_en"=>"Paxtang", "lang"=>"en", "lat"=>40.261900, "lon"=>-76.834264),
"/US/US-PA/paxtonia-18546" => array("type"=>"cdp", "pop"=>4862, "name_en"=>"Paxtonia", "lang"=>"en", "lat"=>40.316200, "lon"=>-76.790837),
"/US/US-PA/pen-argyl-18547" => array("type"=>"borough", "pop"=>3627, "name_en"=>"Pen Argyl", "lang"=>"en", "lat"=>40.867946, "lon"=>-75.254190),
"/US/US-PA/penbrook-18548" => array("type"=>"borough", "pop"=>2919, "name_en"=>"Penbrook", "lang"=>"en", "lat"=>40.278267, "lon"=>-76.848744),
"/US/US-PA/penn-18549" => array("type"=>"borough", "pop"=>422, "name_en"=>"Penn", "lang"=>"en", "lat"=>40.328760, "lon"=>-79.641673),
"/US/US-PA/penn-hills-18551" => array("type"=>"cdp", "pop"=>44061, "name_en"=>"Penn Hills", "lang"=>"en", "lat"=>40.475933, "lon"=>-79.827015),
"/US/US-PA/penn-lake-park-18552" => array("type"=>"borough", "pop"=>283, "name_en"=>"Penn Lake Park", "lang"=>"en", "lat"=>41.114864, "lon"=>-75.774581),
"/US/US-PA/penn-wynne-18556" => array("type"=>"cdp", "pop"=>5807, "name_en"=>"Penn Wynne", "lang"=>"en", "lat"=>39.986250, "lon"=>-75.272705),
"/US/US-PA/penndel-18550" => array("type"=>"borough", "pop"=>2470, "name_en"=>"Penndel", "lang"=>"en", "lat"=>40.154900, "lon"=>-74.913869),
"/US/US-PA/pennsburg-18553" => array("type"=>"borough", "pop"=>3302, "name_en"=>"Pennsburg", "lang"=>"en", "lat"=>40.393586, "lon"=>-75.495895),
"/US/US-PA/pennsbury-village-18554" => array("type"=>"borough", "pop"=>685, "name_en"=>"Pennsbury Village", "lang"=>"en", "lat"=>40.427744, "lon"=>-80.100933),
"/US/US-PA/pennville-18555" => array("type"=>"cdp", "pop"=>1559, "name_ru"=>"Пеннвилл", "name_en"=>"Pennville", "lang"=>"en", "lat"=>39.785633, "lon"=>-76.990354),
"/US/US-PA/perkasie-18557" => array("type"=>"borough", "pop"=>8636, "name_en"=>"Perkasie", "lang"=>"en", "lat"=>40.370998, "lon"=>-75.292425),
"/US/US-PA/perryopolis-18558" => array("type"=>"borough", "pop"=>1712, "name_en"=>"Perryopolis", "lang"=>"en", "lat"=>40.086536, "lon"=>-79.753197),
"/US/US-PA/petersburg-18559" => array("type"=>"borough", "pop"=>423, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>40.573599, "lon"=>-78.049592),
"/US/US-PA/petrolia-18560" => array("type"=>"borough", "pop"=>201, "name_en"=>"Petrolia", "lang"=>"en", "lat"=>41.021803, "lon"=>-79.717200),
"/US/US-PA/philadelphia-18561" => array("type"=>"city", "pop"=>1447395, "name_ru"=>"Филадельфия", "name_en"=>"Philadelphia", "lang"=>"en", "lat"=>40.006817, "lon"=>-75.134678),
"/US/US-PA/philipsburg-18562" => array("type"=>"borough", "pop"=>2915, "name_en"=>"Philipsburg", "lang"=>"en", "lat"=>40.892584, "lon"=>-78.211715),
"/US/US-PA/phoenixville-18563" => array("type"=>"borough", "pop"=>16382, "name_en"=>"Phoenixville", "lang"=>"en", "lat"=>40.135933, "lon"=>-75.519211),
"/US/US-PA/picture-rocks-18564" => array("type"=>"borough", "pop"=>670, "name_ru"=>"Пикчер Рокс", "name_en"=>"Picture Rocks", "lang"=>"en", "lat"=>41.280022, "lon"=>-76.709221),
"/US/US-PA/pillow-18565" => array("type"=>"borough", "pop"=>288, "name_en"=>"Pillow", "lang"=>"en", "lat"=>40.641192, "lon"=>-76.802428),
"/US/US-PA/pine-grove-18566" => array("type"=>"borough", "pop"=>2043, "name_en"=>"Pine Grove", "lang"=>"en", "lat"=>40.553344, "lon"=>-76.387021),
"/US/US-PA/pine-grove-mills-18567" => array("type"=>"cdp", "pop"=>1129, "name_en"=>"Pine Grove Mills", "lang"=>"en", "lat"=>40.738843, "lon"=>-77.873916),
"/US/US-PA/pitcairn-18568" => array("type"=>"borough", "pop"=>3343, "name_en"=>"Pitcairn", "lang"=>"en", "lat"=>40.407650, "lon"=>-79.776049),
"/US/US-PA/pittsburgh-18569" => array("type"=>"city", "pop"=>310037, "name_en"=>"Pittsburgh", "lang"=>"en", "lat"=>40.439207, "lon"=>-79.976702),
"/US/US-PA/pittston-18570" => array("type"=>"city", "pop"=>7547, "name_en"=>"Pittston", "lang"=>"en", "lat"=>41.326133, "lon"=>-75.787884),
"/US/US-PA/plains-18571" => array("type"=>"cdp", "pop"=>10414, "name_ru"=>"Плейнс", "name_en"=>"Plains", "lang"=>"en", "lat"=>41.273500, "lon"=>-75.853893),
"/US/US-PA/platea-18572" => array("type"=>"borough", "pop"=>438, "name_en"=>"Platea", "lang"=>"en", "lat"=>41.951413, "lon"=>-80.330292),
"/US/US-PA/pleasant-gap-18573" => array("type"=>"cdp", "pop"=>1699, "name_en"=>"Pleasant Gap", "lang"=>"en", "lat"=>40.864573, "lon"=>-77.737954),
"/US/US-PA/pleasant-hill-18574" => array("type"=>"cdp", "pop"=>1659, "name_ru"=>"Плезант Хилл", "name_en"=>"Pleasant Hill", "lang"=>"en", "lat"=>40.336731, "lon"=>-76.447245),
"/US/US-PA/pleasant-hills-18575" => array("type"=>"borough", "pop"=>7751, "name_ru"=>"Плезант Хилс", "name_en"=>"Pleasant Hills", "lang"=>"en", "lat"=>40.328650, "lon"=>-79.961096),
"/US/US-PA/pleasantville-18576" => array("type"=>"borough", "pop"=>201, "name_ru"=>"Плезантвилл", "name_en"=>"Pleasantville", "lang"=>"en", "lat"=>40.180337, "lon"=>-78.613331),
"/US/US-PA/pleasantville-18577" => array("type"=>"borough", "pop"=>991, "name_ru"=>"Плезантвилл", "name_en"=>"Pleasantville", "lang"=>"en", "lat"=>41.593558, "lon"=>-79.579326),
"/US/US-PA/plum-18578" => array("type"=>"borough", "pop"=>26198, "name_en"=>"Plum", "lang"=>"en", "lat"=>40.502392, "lon"=>-79.753953),
"/US/US-PA/plumville-18579" => array("type"=>"borough", "pop"=>317, "name_en"=>"Plumville", "lang"=>"en", "lat"=>40.793759, "lon"=>-79.180476),
"/US/US-PA/plymouth-18580" => array("type"=>"borough", "pop"=>6041, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>41.241100, "lon"=>-75.950187),
"/US/US-PA/plymouth-meeting-18581" => array("type"=>"cdp", "pop"=>6241, "name_en"=>"Plymouth Meeting", "lang"=>"en", "lat"=>40.108748, "lon"=>-75.280820),
"/US/US-PA/plymptonville-18582" => array("type"=>"cdp", "pop"=>1074, "name_en"=>"Plymptonville", "lang"=>"en", "lat"=>41.045919, "lon"=>-78.445943),
"/US/US-PA/pocono-pines-18583" => array("type"=>"cdp", "pop"=>824, "name_en"=>"Pocono Pines", "lang"=>"en", "lat"=>41.116765, "lon"=>-75.457901),
"/US/US-PA/point-marion-18584" => array("type"=>"borough", "pop"=>1247, "name_en"=>"Point Marion", "lang"=>"en", "lat"=>39.734453, "lon"=>-79.900681),
"/US/US-PA/polk-18585" => array("type"=>"borough", "pop"=>995, "name_en"=>"Polk", "lang"=>"en", "lat"=>41.369219, "lon"=>-79.930131),
"/US/US-PA/port-allegany-18587" => array("type"=>"borough", "pop"=>2190, "name_en"=>"Port Allegany", "lang"=>"en", "lat"=>41.813563, "lon"=>-78.277664),
"/US/US-PA/port-carbon-18588" => array("type"=>"borough", "pop"=>1747, "name_en"=>"Port Carbon", "lang"=>"en", "lat"=>40.697121, "lon"=>-76.167071),
"/US/US-PA/port-clinton-18589" => array("type"=>"borough", "pop"=>277, "name_en"=>"Port Clinton", "lang"=>"en", "lat"=>40.582375, "lon"=>-76.026724),
"/US/US-PA/port-matilda-18592" => array("type"=>"borough", "pop"=>638, "name_en"=>"Port Matilda", "lang"=>"en", "lat"=>40.799049, "lon"=>-78.052643),
"/US/US-PA/port-royal-18593" => array("type"=>"borough", "pop"=>950, "name_en"=>"Port Royal", "lang"=>"en", "lat"=>40.533855, "lon"=>-77.390889),
"/US/US-PA/port-vue-18594" => array("type"=>"borough", "pop"=>3843, "name_en"=>"Port Vue", "lang"=>"en", "lat"=>40.336900, "lon"=>-79.870535),
"/US/US-PA/portage-18586" => array("type"=>"borough", "pop"=>2581, "name_ru"=>"Портейдж", "name_en"=>"Portage", "lang"=>"en", "lat"=>40.386761, "lon"=>-78.675499),
"/US/US-PA/portersville-18590" => array("type"=>"borough", "pop"=>250, "name_en"=>"Portersville", "lang"=>"en", "lat"=>40.925222, "lon"=>-80.144447),
"/US/US-PA/portland-18591" => array("type"=>"borough", "pop"=>565, "name_ru"=>"Портленд", "name_en"=>"Portland", "lang"=>"en", "lat"=>40.920526, "lon"=>-75.098118),
"/US/US-PA/pottsgrove-18595" => array("type"=>"cdp", "pop"=>3122, "name_en"=>"Pottsgrove", "lang"=>"en", "lat"=>40.264629, "lon"=>-75.612224),
"/US/US-PA/pottstown-18596" => array("type"=>"borough", "pop"=>21288, "name_en"=>"Pottstown", "lang"=>"en", "lat"=>40.250250, "lon"=>-75.644049),
"/US/US-PA/pottsville-18597" => array("type"=>"city", "pop"=>14405, "name_ru"=>"Поттсвилл", "name_en"=>"Pottsville", "lang"=>"en", "lat"=>40.679260, "lon"=>-76.209752),
"/US/US-PA/pringle-18598" => array("type"=>"borough", "pop"=>937, "name_en"=>"Pringle", "lang"=>"en", "lat"=>41.276600, "lon"=>-75.903087),
"/US/US-PA/progress-18599" => array("type"=>"cdp", "pop"=>9654, "name_en"=>"Progress", "lang"=>"en", "lat"=>40.290467, "lon"=>-76.836574),
"/US/US-PA/prompton-18600" => array("type"=>"borough", "pop"=>237, "name_en"=>"Prompton", "lang"=>"en", "lat"=>41.589776, "lon"=>-75.329549),
"/US/US-PA/prospect-18601" => array("type"=>"borough", "pop"=>1221, "name_ru"=>"Проспект", "name_en"=>"Prospect", "lang"=>"en", "lat"=>40.900737, "lon"=>-80.049585),
"/US/US-PA/prospect-park-18602" => array("type"=>"borough", "pop"=>6374, "name_en"=>"Prospect Park", "lang"=>"en", "lat"=>39.885600, "lon"=>-75.307535),
"/US/US-PA/punxsutawney-18603" => array("type"=>"borough", "pop"=>5944, "name_en"=>"Punxsutawney", "lang"=>"en", "lat"=>40.944068, "lon"=>-78.977606),
"/US/US-PA/quakertown-18604" => array("type"=>"borough", "pop"=>8643, "name_en"=>"Quakertown", "lang"=>"en", "lat"=>40.439575, "lon"=>-75.346012),
"/US/US-PA/quarryville-18605" => array("type"=>"borough", "pop"=>2146, "name_en"=>"Quarryville", "lang"=>"en", "lat"=>39.895750, "lon"=>-76.161228),
"/US/US-PA/radnor-township-18606" => array("type"=>"cdp", "pop"=>28705, "name_en"=>"Radnor Township", "lang"=>"en", "lat"=>40.028344, "lon"=>-75.368307),
"/US/US-PA/railroad-18607" => array("type"=>"borough", "pop"=>302, "name_en"=>"Railroad", "lang"=>"en", "lat"=>39.760074, "lon"=>-76.696940),
"/US/US-PA/rainsburg-18608" => array("type"=>"borough", "pop"=>139, "name_en"=>"Rainsburg", "lang"=>"en", "lat"=>39.895175, "lon"=>-78.516209),
"/US/US-PA/ramblewood-18609" => array("type"=>"cdp", "pop"=>1104, "name_en"=>"Ramblewood", "lang"=>"en", "lat"=>40.719460, "lon"=>-77.956650),
"/US/US-PA/ramey-18610" => array("type"=>"borough", "pop"=>500, "name_en"=>"Ramey", "lang"=>"en", "lat"=>40.800716, "lon"=>-78.398846),
"/US/US-PA/rankin-18611" => array("type"=>"borough", "pop"=>2108, "name_ru"=>"Ранкин", "name_en"=>"Rankin", "lang"=>"en", "lat"=>40.411367, "lon"=>-79.882138),
"/US/US-PA/reading-18612" => array("type"=>"city", "pop"=>80506, "name_ru"=>"Ридинг", "name_en"=>"Reading", "lang"=>"en", "lat"=>40.339000, "lon"=>-75.926423),
"/US/US-PA/reamstown-18613" => array("type"=>"cdp", "pop"=>2649, "name_en"=>"Reamstown", "lang"=>"en", "lat"=>40.211953, "lon"=>-76.117919),
"/US/US-PA/red-hill-18614" => array("type"=>"borough", "pop"=>2302, "name_en"=>"Red Hill", "lang"=>"en", "lat"=>40.376761, "lon"=>-75.483726),
"/US/US-PA/red-lion-18615" => array("type"=>"borough", "pop"=>6041, "name_en"=>"Red Lion", "lang"=>"en", "lat"=>39.898700, "lon"=>-76.607737),
"/US/US-PA/reedsville-18616" => array("type"=>"cdp", "pop"=>1030, "name_en"=>"Reedsville", "lang"=>"en", "lat"=>40.659277, "lon"=>-77.602020),
"/US/US-PA/reiffton-18617" => array("type"=>"cdp", "pop"=>2522, "name_en"=>"Reiffton", "lang"=>"en", "lat"=>40.315950, "lon"=>-75.867572),
"/US/US-PA/reinerton-orwin-muir-18618" => array("type"=>"cdp", "pop"=>1255, "name_en"=>"Reinerton-Orwin-Muir", "lang"=>"en", "lat"=>40.585515, "lon"=>-76.528551),
"/US/US-PA/renovo-18619" => array("type"=>"borough", "pop"=>1217, "name_en"=>"Renovo", "lang"=>"en", "lat"=>41.329136, "lon"=>-77.747798),
"/US/US-PA/republic-18620" => array("type"=>"cdp", "pop"=>1603, "name_ru"=>"Репаблик", "name_en"=>"Republic", "lang"=>"en", "lat"=>39.964832, "lon"=>-79.876624),
"/US/US-PA/reserve-township-18621" => array("type"=>"cdp", "pop"=>3866, "name_en"=>"Reserve Township", "lang"=>"en", "lat"=>40.482749, "lon"=>-79.987147),
"/US/US-PA/reynoldsville-18622" => array("type"=>"borough", "pop"=>2551, "name_en"=>"Reynoldsville", "lang"=>"en", "lat"=>41.094297, "lon"=>-78.888046),
"/US/US-PA/rheems-18623" => array("type"=>"cdp", "pop"=>1044, "name_en"=>"Rheems", "lang"=>"en", "lat"=>40.128950, "lon"=>-76.570679),
"/US/US-PA/rices-landing-18624" => array("type"=>"borough", "pop"=>408, "name_en"=>"Rices Landing", "lang"=>"en", "lat"=>39.946104, "lon"=>-79.994770),
"/US/US-PA/richboro-18625" => array("type"=>"cdp", "pop"=>5332, "name_en"=>"Richboro", "lang"=>"en", "lat"=>40.227300, "lon"=>-75.000837),
"/US/US-PA/richland-18626" => array("type"=>"borough", "pop"=>1492, "name_ru"=>"Ричленд", "name_en"=>"Richland", "lang"=>"en", "lat"=>40.357030, "lon"=>-76.257514),
"/US/US-PA/richlandtown-18627" => array("type"=>"borough", "pop"=>1327, "name_en"=>"Richlandtown", "lang"=>"en", "lat"=>40.472282, "lon"=>-75.321327),
"/US/US-PA/ridgway-18628" => array("type"=>"borough", "pop"=>4096, "name_ru"=>"Риджуэй", "name_en"=>"Ridgway", "lang"=>"en", "lat"=>41.425560, "lon"=>-78.728795),
"/US/US-PA/ridley-park-18629" => array("type"=>"borough", "pop"=>6979, "name_en"=>"Ridley Park", "lang"=>"en", "lat"=>39.877650, "lon"=>-75.325644),
"/US/US-PA/riegelsville-18630" => array("type"=>"borough", "pop"=>833, "name_en"=>"Riegelsville", "lang"=>"en", "lat"=>40.596924, "lon"=>-75.197797),
"/US/US-PA/rimersburg-18631" => array("type"=>"borough", "pop"=>1013, "name_en"=>"Rimersburg", "lang"=>"en", "lat"=>41.041226, "lon"=>-79.502141),
"/US/US-PA/ringtown-18632" => array("type"=>"borough", "pop"=>762, "name_en"=>"Ringtown", "lang"=>"en", "lat"=>40.856892, "lon"=>-76.235370),
"/US/US-PA/riverside-18633" => array("type"=>"borough", "pop"=>1853, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>40.942614, "lon"=>-76.645675),
"/US/US-PA/roaring-spring-18634" => array("type"=>"borough", "pop"=>2252, "name_en"=>"Roaring Spring", "lang"=>"en", "lat"=>40.333652, "lon"=>-78.396750),
"/US/US-PA/robesonia-18635" => array("type"=>"borough", "pop"=>2049, "name_en"=>"Robesonia", "lang"=>"en", "lat"=>40.348560, "lon"=>-76.141138),
"/US/US-PA/robinson-township-18636" => array("type"=>"cdp", "pop"=>10830, "name_en"=>"Robinson Township", "lang"=>"en", "lat"=>40.462112, "lon"=>-80.126835),
"/US/US-PA/rochester-18637" => array("type"=>"borough", "pop"=>3662, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>40.702000, "lon"=>-80.283299),
"/US/US-PA/rockhill-furnace-18638" => array("type"=>"borough", "pop"=>421, "name_en"=>"Rockhill Furnace", "lang"=>"en", "lat"=>40.242008, "lon"=>-77.900681),
"/US/US-PA/rockledge-18639" => array("type"=>"borough", "pop"=>2478, "name_ru"=>"Рокледж", "name_en"=>"Rockledge", "lang"=>"en", "lat"=>40.082200, "lon"=>-75.089983),
"/US/US-PA/rockwood-18640" => array("type"=>"borough", "pop"=>905, "name_ru"=>"Роквуд", "name_en"=>"Rockwood", "lang"=>"en", "lat"=>39.916747, "lon"=>-79.157137),
"/US/US-PA/rome-18641" => array("type"=>"borough", "pop"=>361, "name_ru"=>"Роум", "name_en"=>"Rome", "lang"=>"en", "lat"=>41.857915, "lon"=>-76.341635),
"/US/US-PA/roscoe-18642" => array("type"=>"borough", "pop"=>788, "name_ru"=>"Роско", "name_en"=>"Roscoe", "lang"=>"en", "lat"=>40.077600, "lon"=>-79.864908),
"/US/US-PA/rose-valley-18644" => array("type"=>"borough", "pop"=>924, "name_en"=>"Rose Valley", "lang"=>"en", "lat"=>39.894833, "lon"=>-75.386113),
"/US/US-PA/roseto-18643" => array("type"=>"borough", "pop"=>1644, "name_en"=>"Roseto", "lang"=>"en", "lat"=>40.878182, "lon"=>-75.220769),
"/US/US-PA/roseville-18645" => array("type"=>"borough", "pop"=>195, "name_ru"=>"Роузвилл", "name_en"=>"Roseville", "lang"=>"en", "lat"=>41.865429, "lon"=>-76.958906),
"/US/US-PA/ross-township-18647" => array("type"=>"cdp", "pop"=>33482, "name_en"=>"Ross Township", "lang"=>"en", "lat"=>40.526600, "lon"=>-80.024934),
"/US/US-PA/rosslyn-farms-18646" => array("type"=>"borough", "pop"=>423, "name_en"=>"Rosslyn Farms", "lang"=>"en", "lat"=>40.421600, "lon"=>-80.089839),
"/US/US-PA/rothsville-18648" => array("type"=>"cdp", "pop"=>2097, "name_en"=>"Rothsville", "lang"=>"en", "lat"=>40.152667, "lon"=>-76.245736),
"/US/US-PA/rouseville-18649" => array("type"=>"borough", "pop"=>432, "name_en"=>"Rouseville", "lang"=>"en", "lat"=>41.471042, "lon"=>-79.686090),
"/US/US-PA/rouzerville-18650" => array("type"=>"cdp", "pop"=>1188, "name_en"=>"Rouzerville", "lang"=>"en", "lat"=>39.733998, "lon"=>-77.524836),
"/US/US-PA/royalton-18651" => array("type"=>"borough", "pop"=>946, "name_ru"=>"Ройалтон", "name_en"=>"Royalton", "lang"=>"en", "lat"=>40.187150, "lon"=>-76.726149),
"/US/US-PA/royersford-18652" => array("type"=>"borough", "pop"=>4356, "name_en"=>"Royersford", "lang"=>"en", "lat"=>40.185200, "lon"=>-75.537967),
"/US/US-PA/rural-valley-18653" => array("type"=>"borough", "pop"=>852, "name_en"=>"Rural Valley", "lang"=>"en", "lat"=>40.799108, "lon"=>-79.314669),
"/US/US-PA/russellton-18654" => array("type"=>"cdp", "pop"=>1691, "name_en"=>"Russellton", "lang"=>"en", "lat"=>40.610736, "lon"=>-79.837716),
"/US/US-PA/rutherford-18655" => array("type"=>"cdp", "pop"=>3481, "name_en"=>"Rutherford", "lang"=>"en", "lat"=>40.268850, "lon"=>-76.768200),
"/US/US-PA/rutledge-18656" => array("type"=>"borough", "pop"=>823, "name_ru"=>"Рутледж", "name_en"=>"Rutledge", "lang"=>"en", "lat"=>39.900700, "lon"=>-75.327772),
"/US/US-PA/s.n.p.j.-18735" => array("type"=>"borough", "pop"=>0, "name_en"=>"S.N.P.J.", "lang"=>"en", "lat"=>40.928855, "lon"=>-80.499038),
"/US/US-PA/saegertown-18657" => array("type"=>"borough", "pop"=>1036, "name_en"=>"Saegertown", "lang"=>"en", "lat"=>41.713267, "lon"=>-80.138338),
"/US/US-PA/salisbury-18664" => array("type"=>"borough", "pop"=>810, "name_ru"=>"Солсбери", "name_en"=>"Salisbury", "lang"=>"en", "lat"=>39.754175, "lon"=>-79.084649),
"/US/US-PA/salix-beauty-line-park-18665" => array("type"=>"cdp", "pop"=>1257, "name_en"=>"Salix-Beauty Line Park", "lang"=>"en", "lat"=>40.298485, "lon"=>-78.759430),
"/US/US-PA/salladasburg-18666" => array("type"=>"borough", "pop"=>248, "name_en"=>"Salladasburg", "lang"=>"en", "lat"=>41.276520, "lon"=>-77.227613),
"/US/US-PA/saltillo-18667" => array("type"=>"borough", "pop"=>325, "name_ru"=>"Салтилло", "name_en"=>"Saltillo", "lang"=>"en", "lat"=>40.211794, "lon"=>-78.007843),
"/US/US-PA/saltsburg-18668" => array("type"=>"borough", "pop"=>883, "name_en"=>"Saltsburg", "lang"=>"en", "lat"=>40.485196, "lon"=>-79.446512),
"/US/US-PA/salunga-landisville-18669" => array("type"=>"cdp", "pop"=>4239, "name_en"=>"Salunga-Landisville", "lang"=>"en", "lat"=>40.091550, "lon"=>-76.418625),
"/US/US-PA/sanatoga-18670" => array("type"=>"cdp", "pop"=>5534, "name_en"=>"Sanatoga", "lang"=>"en", "lat"=>40.249521, "lon"=>-75.589241),
"/US/US-PA/sand-hill-18671" => array("type"=>"cdp", "pop"=>2307, "name_en"=>"Sand Hill", "lang"=>"en", "lat"=>40.360861, "lon"=>-76.421487),
"/US/US-PA/sandy-18672" => array("type"=>"cdp", "pop"=>11575, "name_en"=>"Sandy", "lang"=>"en", "lat"=>41.107223, "lon"=>-78.773805),
"/US/US-PA/sandy-lake-18673" => array("type"=>"borough", "pop"=>691, "name_en"=>"Sandy Lake", "lang"=>"en", "lat"=>41.347601, "lon"=>-80.084599),
"/US/US-PA/sankertown-18674" => array("type"=>"borough", "pop"=>655, "name_en"=>"Sankertown", "lang"=>"en", "lat"=>40.470620, "lon"=>-78.592917),
"/US/US-PA/saxonburg-18675" => array("type"=>"borough", "pop"=>1593, "name_en"=>"Saxonburg", "lang"=>"en", "lat"=>40.752015, "lon"=>-79.815542),
"/US/US-PA/saxton-18676" => array("type"=>"borough", "pop"=>762, "name_en"=>"Saxton", "lang"=>"en", "lat"=>40.212957, "lon"=>-78.247236),
"/US/US-PA/sayre-18677" => array("type"=>"borough", "pop"=>5440, "name_en"=>"Sayre", "lang"=>"en", "lat"=>41.985864, "lon"=>-76.520986),
"/US/US-PA/scalp-level-18678" => array("type"=>"borough", "pop"=>789, "name_en"=>"Scalp Level", "lang"=>"en", "lat"=>40.249283, "lon"=>-78.843945),
"/US/US-PA/schellsburg-18679" => array("type"=>"borough", "pop"=>306, "name_en"=>"Schellsburg", "lang"=>"en", "lat"=>40.048611, "lon"=>-78.643014),
"/US/US-PA/schlusser-18680" => array("type"=>"cdp", "pop"=>4728, "name_en"=>"Schlusser", "lang"=>"en", "lat"=>40.241586, "lon"=>-77.177289),
"/US/US-PA/schnecksville-18681" => array("type"=>"cdp", "pop"=>1780, "name_en"=>"Schnecksville", "lang"=>"en", "lat"=>40.678232, "lon"=>-75.619601),
"/US/US-PA/schuylkill-haven-18682" => array("type"=>"borough", "pop"=>5161, "name_en"=>"Schuylkill Haven", "lang"=>"en", "lat"=>40.628084, "lon"=>-76.172922),
"/US/US-PA/schwenksville-18683" => array("type"=>"borough", "pop"=>1325, "name_en"=>"Schwenksville", "lang"=>"en", "lat"=>40.257900, "lon"=>-75.466517),
"/US/US-PA/scott-township-18685" => array("type"=>"cdp", "pop"=>17118, "name_en"=>"Scott Township", "lang"=>"en", "lat"=>40.386200, "lon"=>-80.077234),
"/US/US-PA/scottdale-18684" => array("type"=>"borough", "pop"=>4409, "name_ru"=>"Скотдейл", "name_en"=>"Scottdale", "lang"=>"en", "lat"=>40.102819, "lon"=>-79.589917),
"/US/US-PA/scranton-18686" => array("type"=>"city", "pop"=>72233, "name_ru"=>"Скрантон", "name_en"=>"Scranton", "lang"=>"en", "lat"=>41.404050, "lon"=>-75.666103),
"/US/US-PA/selinsgrove-18687" => array("type"=>"borough", "pop"=>5344, "name_en"=>"Selinsgrove", "lang"=>"en", "lat"=>40.801059, "lon"=>-76.865548),
"/US/US-PA/sellersville-18688" => array("type"=>"borough", "pop"=>4476, "name_en"=>"Sellersville", "lang"=>"en", "lat"=>40.360309, "lon"=>-75.308265),
"/US/US-PA/seneca-18689" => array("type"=>"cdp", "pop"=>1029, "name_ru"=>"Сенека", "name_en"=>"Seneca", "lang"=>"en", "lat"=>41.375744, "lon"=>-79.705532),
"/US/US-PA/seven-fields-18690" => array("type"=>"borough", "pop"=>2550, "name_en"=>"Seven Fields", "lang"=>"en", "lat"=>40.688416, "lon"=>-80.062010),
"/US/US-PA/seven-springs-18691" => array("type"=>"borough", "pop"=>117, "name_en"=>"Seven Springs", "lang"=>"en", "lat"=>40.025095, "lon"=>-79.291537),
"/US/US-PA/seven-valleys-18692" => array("type"=>"borough", "pop"=>507, "name_en"=>"Seven Valleys", "lang"=>"en", "lat"=>39.852617, "lon"=>-76.766474),
"/US/US-PA/seward-18693" => array("type"=>"borough", "pop"=>451, "name_ru"=>"Сьюард", "name_en"=>"Seward", "lang"=>"en", "lat"=>40.412525, "lon"=>-79.021730),
"/US/US-PA/sewickley-18694" => array("type"=>"borough", "pop"=>3562, "name_en"=>"Sewickley", "lang"=>"en", "lat"=>40.539667, "lon"=>-80.180499),
"/US/US-PA/sewickley-heights-18695" => array("type"=>"borough", "pop"=>918, "name_en"=>"Sewickley Heights", "lang"=>"en", "lat"=>40.561024, "lon"=>-80.155761),
"/US/US-PA/sewickley-hills-18696" => array("type"=>"borough", "pop"=>700, "name_en"=>"Sewickley Hills", "lang"=>"en", "lat"=>40.560256, "lon"=>-80.120195),
"/US/US-PA/shade-gap-18697" => array("type"=>"borough", "pop"=>91, "name_en"=>"Shade Gap", "lang"=>"en", "lat"=>40.180024, "lon"=>-77.865716),
"/US/US-PA/shaler-township-18698" => array("type"=>"cdp", "pop"=>30533, "name_en"=>"Shaler Township", "lang"=>"en", "lat"=>40.520858, "lon"=>-79.964009),
"/US/US-PA/shamokin-18699" => array("type"=>"city", "pop"=>7367, "name_en"=>"Shamokin", "lang"=>"en", "lat"=>40.788072, "lon"=>-76.554906),
"/US/US-PA/shamokin-dam-18700" => array("type"=>"borough", "pop"=>1438, "name_en"=>"Shamokin Dam", "lang"=>"en", "lat"=>40.853128, "lon"=>-76.822455),
"/US/US-PA/shanksville-18701" => array("type"=>"borough", "pop"=>222, "name_en"=>"Shanksville", "lang"=>"en", "lat"=>40.017412, "lon"=>-78.906850),
"/US/US-PA/shanor-northvue-18702" => array("type"=>"cdp", "pop"=>3517, "name_en"=>"Shanor-Northvue", "lang"=>"en", "lat"=>40.909283, "lon"=>-79.917191),
"/US/US-PA/sharon-18703" => array("type"=>"city", "pop"=>14869, "name_ru"=>"Шэрон", "name_en"=>"Sharon", "lang"=>"en", "lat"=>41.234200, "lon"=>-80.499827),
"/US/US-PA/sharon-hill-18704" => array("type"=>"borough", "pop"=>5308, "name_en"=>"Sharon Hill", "lang"=>"en", "lat"=>39.907750, "lon"=>-75.269184),
"/US/US-PA/sharpsburg-18705" => array("type"=>"borough", "pop"=>3263, "name_ru"=>"Шарпсбург", "name_en"=>"Sharpsburg", "lang"=>"en", "lat"=>40.493167, "lon"=>-79.926589),
"/US/US-PA/sharpsville-18706" => array("type"=>"borough", "pop"=>4100, "name_ru"=>"Шарпсвилл", "name_en"=>"Sharpsville", "lang"=>"en", "lat"=>41.258950, "lon"=>-80.481999),
"/US/US-PA/sheakleyville-18707" => array("type"=>"borough", "pop"=>151, "name_en"=>"Sheakleyville", "lang"=>"en", "lat"=>41.444653, "lon"=>-80.207069),
"/US/US-PA/sheffield-18708" => array("type"=>"cdp", "pop"=>2200, "name_ru"=>"Шеффилд", "name_en"=>"Sheffield", "lang"=>"en", "lat"=>41.696283, "lon"=>-79.028330),
"/US/US-PA/shelocta-18709" => array("type"=>"borough", "pop"=>119, "name_en"=>"Shelocta", "lang"=>"en", "lat"=>40.656512, "lon"=>-79.302386),
"/US/US-PA/shenandoah-18710" => array("type"=>"borough", "pop"=>5155, "name_ru"=>"Шенандоа", "name_en"=>"Shenandoah", "lang"=>"en", "lat"=>40.816551, "lon"=>-76.200838),
"/US/US-PA/shenandoah-heights-18711" => array("type"=>"cdp", "pop"=>1386, "name_en"=>"Shenandoah Heights", "lang"=>"en", "lat"=>40.833424, "lon"=>-76.208029),
"/US/US-PA/shickshinny-18712" => array("type"=>"borough", "pop"=>892, "name_en"=>"Shickshinny", "lang"=>"en", "lat"=>41.153212, "lon"=>-76.151621),
"/US/US-PA/shillington-18713" => array("type"=>"borough", "pop"=>4997, "name_en"=>"Shillington", "lang"=>"en", "lat"=>40.302550, "lon"=>-75.967221),
"/US/US-PA/shiloh-18714" => array("type"=>"cdp", "pop"=>8245, "name_ru"=>"Шилох", "name_en"=>"Shiloh", "lang"=>"en", "lat"=>39.973789, "lon"=>-76.791767),
"/US/US-PA/shinglehouse-18715" => array("type"=>"borough", "pop"=>1105, "name_en"=>"Shinglehouse", "lang"=>"en", "lat"=>41.965318, "lon"=>-78.191499),
"/US/US-PA/shippensburg-18716" => array("type"=>"borough", "pop"=>5578, "name_en"=>"Shippensburg", "lang"=>"en", "lat"=>40.048128, "lon"=>-77.523030),
"/US/US-PA/shippenville-18717" => array("type"=>"borough", "pop"=>470, "name_en"=>"Shippenville", "lang"=>"en", "lat"=>41.250928, "lon"=>-79.463027),
"/US/US-PA/shippingport-18718" => array("type"=>"borough", "pop"=>219, "name_en"=>"Shippingport", "lang"=>"en", "lat"=>40.625421, "lon"=>-80.424213),
"/US/US-PA/shiremanstown-18719" => array("type"=>"borough", "pop"=>1464, "name_en"=>"Shiremanstown", "lang"=>"en", "lat"=>40.222350, "lon"=>-76.956291),
"/US/US-PA/shirleysburg-18720" => array("type"=>"borough", "pop"=>130, "name_en"=>"Shirleysburg", "lang"=>"en", "lat"=>40.296935, "lon"=>-77.877415),
"/US/US-PA/shoemakersville-18721" => array("type"=>"borough", "pop"=>1393, "name_en"=>"Shoemakersville", "lang"=>"en", "lat"=>40.499949, "lon"=>-75.969392),
"/US/US-PA/shrewsbury-18722" => array("type"=>"borough", "pop"=>3573, "name_ru"=>"Шрюсберри", "name_en"=>"Shrewsbury", "lang"=>"en", "lat"=>39.770954, "lon"=>-76.680304),
"/US/US-PA/silverdale-18723" => array("type"=>"borough", "pop"=>965, "name_en"=>"Silverdale", "lang"=>"en", "lat"=>40.346355, "lon"=>-75.271336),
"/US/US-PA/sinking-spring-18724" => array("type"=>"borough", "pop"=>3605, "name_en"=>"Sinking Spring", "lang"=>"en", "lat"=>40.324633, "lon"=>-76.022726),
"/US/US-PA/skippack-18725" => array("type"=>"cdp", "pop"=>13453, "name_en"=>"Skippack", "lang"=>"en", "lat"=>40.221039, "lon"=>-75.401572),
"/US/US-PA/skyline-view-18726" => array("type"=>"cdp", "pop"=>2370, "name_en"=>"Skyline View", "lang"=>"en", "lat"=>40.338250, "lon"=>-76.726296),
"/US/US-PA/slatington-18727" => array("type"=>"borough", "pop"=>4404, "name_en"=>"Slatington", "lang"=>"en", "lat"=>40.754329, "lon"=>-75.612291),
"/US/US-PA/sligo-18728" => array("type"=>"borough", "pop"=>690, "name_en"=>"Sligo", "lang"=>"en", "lat"=>41.108769, "lon"=>-79.495900),
"/US/US-PA/slippery-rock-18729" => array("type"=>"borough", "pop"=>3068, "name_en"=>"Slippery Rock", "lang"=>"en", "lat"=>41.070011, "lon"=>-80.060278),
"/US/US-PA/smethport-18730" => array("type"=>"borough", "pop"=>1563, "name_en"=>"Smethport", "lang"=>"en", "lat"=>41.807430, "lon"=>-78.444288),
"/US/US-PA/smicksburg-18731" => array("type"=>"borough", "pop"=>45, "name_en"=>"Smicksburg", "lang"=>"en", "lat"=>40.869978, "lon"=>-79.171797),
"/US/US-PA/smithfield-18732" => array("type"=>"borough", "pop"=>803, "name_ru"=>"Смитфилд", "name_en"=>"Smithfield", "lang"=>"en", "lat"=>39.801989, "lon"=>-79.809507),
"/US/US-PA/smithton-18733" => array("type"=>"borough", "pop"=>407, "name_ru"=>"Смиттон", "name_en"=>"Smithton", "lang"=>"en", "lat"=>40.153288, "lon"=>-79.741660),
"/US/US-PA/snow-shoe-18734" => array("type"=>"borough", "pop"=>774, "name_en"=>"Snow Shoe", "lang"=>"en", "lat"=>41.027214, "lon"=>-77.950308),
"/US/US-PA/snydertown-18736" => array("type"=>"borough", "pop"=>331, "name_en"=>"Snydertown", "lang"=>"en", "lat"=>40.872391, "lon"=>-76.674494),
"/US/US-PA/somerset-18737" => array("type"=>"borough", "pop"=>6359, "name_ru"=>"Сомерсет", "name_en"=>"Somerset", "lang"=>"en", "lat"=>40.005002, "lon"=>-79.078267),
"/US/US-PA/souderton-18738" => array("type"=>"borough", "pop"=>6553, "name_en"=>"Souderton", "lang"=>"en", "lat"=>40.310132, "lon"=>-75.322095),
"/US/US-PA/south-bethlehem-18739" => array("type"=>"borough", "pop"=>412, "name_en"=>"South Bethlehem", "lang"=>"en", "lat"=>40.999632, "lon"=>-79.339546),
"/US/US-PA/south-coatesville-18740" => array("type"=>"borough", "pop"=>1060, "name_en"=>"South Coatesville", "lang"=>"en", "lat"=>39.968125, "lon"=>-75.815285),
"/US/US-PA/south-connellsville-18741" => array("type"=>"borough", "pop"=>2151, "name_en"=>"South Connellsville", "lang"=>"en", "lat"=>39.992561, "lon"=>-79.582453),
"/US/US-PA/south-fork-18742" => array("type"=>"borough", "pop"=>1028, "name_en"=>"South Fork", "lang"=>"en", "lat"=>40.363800, "lon"=>-78.791589),
"/US/US-PA/south-greensburg-18743" => array("type"=>"borough", "pop"=>2217, "name_en"=>"South Greensburg", "lang"=>"en", "lat"=>40.276936, "lon"=>-79.547941),
"/US/US-PA/south-heights-18744" => array("type"=>"borough", "pop"=>492, "name_en"=>"South Heights", "lang"=>"en", "lat"=>40.574567, "lon"=>-80.236313),
"/US/US-PA/south-new-castle-18746" => array("type"=>"borough", "pop"=>737, "name_en"=>"South New Castle", "lang"=>"en", "lat"=>40.974179, "lon"=>-80.346022),
"/US/US-PA/south-park-township-18747" => array("type"=>"cdp", "pop"=>14292, "name_en"=>"South Park Township", "lang"=>"en", "lat"=>40.296950, "lon"=>-79.993864),
"/US/US-PA/south-philipsburg-18748" => array("type"=>"borough", "pop"=>438, "name_en"=>"South Philipsburg", "lang"=>"en", "lat"=>40.887216, "lon"=>-78.218524),
"/US/US-PA/south-pottstown-18749" => array("type"=>"cdp", "pop"=>1966, "name_en"=>"South Pottstown", "lang"=>"en", "lat"=>40.237022, "lon"=>-75.664593),
"/US/US-PA/south-renovo-18750" => array("type"=>"borough", "pop"=>517, "name_en"=>"South Renovo", "lang"=>"en", "lat"=>41.324438, "lon"=>-77.742528),
"/US/US-PA/south-waverly-18751" => array("type"=>"borough", "pop"=>964, "name_en"=>"South Waverly", "lang"=>"en", "lat"=>41.995556, "lon"=>-76.544344),
"/US/US-PA/south-williamsport-18753" => array("type"=>"borough", "pop"=>6017, "name_en"=>"South Williamsport", "lang"=>"en", "lat"=>41.229900, "lon"=>-77.001267),
"/US/US-PA/southmont-18745" => array("type"=>"borough", "pop"=>2063, "name_en"=>"Southmont", "lang"=>"en", "lat"=>40.310400, "lon"=>-78.930937),
"/US/US-PA/southwest-greensburg-18752" => array("type"=>"borough", "pop"=>2201, "name_en"=>"Southwest Greensburg", "lang"=>"en", "lat"=>40.291976, "lon"=>-79.547753),
"/US/US-PA/spangler-18754" => array("type"=>"borough", "pop"=>2068, "name_en"=>"Spangler", "lang"=>"en", "lat"=>40.645164, "lon"=>-78.773276),
"/US/US-PA/spartansburg-18755" => array("type"=>"borough", "pop"=>314, "name_en"=>"Spartansburg", "lang"=>"en", "lat"=>41.823826, "lon"=>-79.682489),
"/US/US-PA/speers-18756" => array("type"=>"borough", "pop"=>1165, "name_en"=>"Speers", "lang"=>"en", "lat"=>40.122616, "lon"=>-79.880056),
"/US/US-PA/spring-city-18758" => array("type"=>"borough", "pop"=>3375, "name_en"=>"Spring City", "lang"=>"en", "lat"=>40.176700, "lon"=>-75.546870),
"/US/US-PA/spring-grove-18762" => array("type"=>"borough", "pop"=>2206, "name_ru"=>"Спринг Гров", "name_en"=>"Spring Grove", "lang"=>"en", "lat"=>39.881360, "lon"=>-76.864351),
"/US/US-PA/spring-hill-18763" => array("type"=>"cdp", "pop"=>1014, "name_ru"=>"Спринг Хилл", "name_en"=>"Spring Hill", "lang"=>"en", "lat"=>40.371143, "lon"=>-78.666471),
"/US/US-PA/spring-house-18764" => array("type"=>"cdp", "pop"=>2782, "name_en"=>"Spring House", "lang"=>"en", "lat"=>40.183626, "lon"=>-75.226415),
"/US/US-PA/spring-mount-18765" => array("type"=>"cdp", "pop"=>1365, "name_en"=>"Spring Mount", "lang"=>"en", "lat"=>40.273587, "lon"=>-75.465396),
"/US/US-PA/springboro-18757" => array("type"=>"borough", "pop"=>459, "name_en"=>"Springboro", "lang"=>"en", "lat"=>41.800000, "lon"=>-80.371244),
"/US/US-PA/springdale-18759" => array("type"=>"borough", "pop"=>3487, "name_ru"=>"Спрингдейл", "name_en"=>"Springdale", "lang"=>"en", "lat"=>40.539560, "lon"=>-79.781952),
"/US/US-PA/springetts-manor-yorklyn-18760" => array("type"=>"cdp", "pop"=>3433, "name_en"=>"Springetts Manor-Yorklyn", "lang"=>"en", "lat"=>39.989866, "lon"=>-76.647403),
"/US/US-PA/springfield-18761" => array("type"=>"cdp", "pop"=>1147, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>39.927733, "lon"=>-75.337558),
"/US/US-PA/spry-18766" => array("type"=>"cdp", "pop"=>4271, "name_en"=>"Spry", "lang"=>"en", "lat"=>39.912500, "lon"=>-76.687529),
"/US/US-PA/st.-clair-18658" => array("type"=>"borough", "pop"=>2987, "name_ru"=>"Сент Клэр", "name_en"=>"St. Clair", "lang"=>"en", "lat"=>40.720408, "lon"=>-76.190812),
"/US/US-PA/st.-clairsville-18659" => array("type"=>"borough", "pop"=>81, "name_en"=>"St. Clairsville", "lang"=>"en", "lat"=>40.155907, "lon"=>-78.510777),
"/US/US-PA/st.-lawrence-18660" => array("type"=>"borough", "pop"=>1861, "name_en"=>"St. Lawrence", "lang"=>"en", "lat"=>40.326000, "lon"=>-75.864952),
"/US/US-PA/st.-marys-18661" => array("type"=>"borough", "pop"=>13409, "name_ru"=>"Сент Мэрис", "name_en"=>"St. Marys", "lang"=>"en", "lat"=>41.428564, "lon"=>-78.557665),
"/US/US-PA/st.-michael-sidman-18662" => array("type"=>"cdp", "pop"=>1189, "name_en"=>"St. Michael-Sidman", "lang"=>"en", "lat"=>40.323287, "lon"=>-78.765473),
"/US/US-PA/st.-petersburg-18663" => array("type"=>"borough", "pop"=>393, "name_ru"=>"Сент Петерсберг", "name_en"=>"St. Petersburg", "lang"=>"en", "lat"=>41.160883, "lon"=>-79.654361),
"/US/US-PA/starrucca-18767" => array("type"=>"borough", "pop"=>224, "name_en"=>"Starrucca", "lang"=>"en", "lat"=>41.902758, "lon"=>-75.454367),
"/US/US-PA/state-college-18768" => array("type"=>"borough", "pop"=>39419, "name_en"=>"State College", "lang"=>"en", "lat"=>40.790550, "lon"=>-77.857919),
"/US/US-PA/steelton-18769" => array("type"=>"borough", "pop"=>5598, "name_en"=>"Steelton", "lang"=>"en", "lat"=>40.226150, "lon"=>-76.825887),
"/US/US-PA/stewartstown-18770" => array("type"=>"borough", "pop"=>1997, "name_en"=>"Stewartstown", "lang"=>"en", "lat"=>39.752404, "lon"=>-76.593042),
"/US/US-PA/stillwater-18771" => array("type"=>"borough", "pop"=>203, "name_ru"=>"Стиллуотер", "name_en"=>"Stillwater", "lang"=>"en", "lat"=>41.150160, "lon"=>-76.367088),
"/US/US-PA/stockdale-18772" => array("type"=>"borough", "pop"=>520, "name_en"=>"Stockdale", "lang"=>"en", "lat"=>40.083275, "lon"=>-79.850689),
"/US/US-PA/stockertown-18773" => array("type"=>"borough", "pop"=>753, "name_en"=>"Stockertown", "lang"=>"en", "lat"=>40.755334, "lon"=>-75.263766),
"/US/US-PA/stoneboro-18774" => array("type"=>"borough", "pop"=>1015, "name_en"=>"Stoneboro", "lang"=>"en", "lat"=>41.336637, "lon"=>-80.109726),
"/US/US-PA/stonybrook-wilshire-18775" => array("type"=>"cdp", "pop"=>4887, "name_en"=>"Stonybrook-Wilshire", "lang"=>"en", "lat"=>39.974750, "lon"=>-76.642953),
"/US/US-PA/stowe-18776" => array("type"=>"cdp", "pop"=>6094, "name_en"=>"Stowe", "lang"=>"en", "lat"=>40.251600, "lon"=>-75.681599),
"/US/US-PA/stowe-township-18777" => array("type"=>"cdp", "pop"=>7681, "name_en"=>"Stowe Township", "lang"=>"en", "lat"=>40.483133, "lon"=>-80.073321),
"/US/US-PA/stoystown-18778" => array("type"=>"borough", "pop"=>391, "name_en"=>"Stoystown", "lang"=>"en", "lat"=>40.103351, "lon"=>-78.955265),
"/US/US-PA/strasburg-18779" => array("type"=>"borough", "pop"=>2777, "name_ru"=>"Страсбург", "name_en"=>"Strasburg", "lang"=>"en", "lat"=>39.983600, "lon"=>-76.185643),
"/US/US-PA/strattanville-18780" => array("type"=>"borough", "pop"=>531, "name_en"=>"Strattanville", "lang"=>"en", "lat"=>41.202945, "lon"=>-79.327612),
"/US/US-PA/strausstown-18781" => array("type"=>"borough", "pop"=>334, "name_en"=>"Strausstown", "lang"=>"en", "lat"=>40.491961, "lon"=>-76.184434),
"/US/US-PA/stroudsburg-18782" => array("type"=>"borough", "pop"=>6294, "name_en"=>"Stroudsburg", "lang"=>"en", "lat"=>40.983634, "lon"=>-75.197570),
"/US/US-PA/sturgeon-noblestown-18783" => array("type"=>"cdp", "pop"=>1350, "name_en"=>"Sturgeon-Noblestown", "lang"=>"en", "lat"=>40.388803, "lon"=>-80.203141),
"/US/US-PA/sugar-grove-18785" => array("type"=>"borough", "pop"=>555, "name_ru"=>"Шугар Гров", "name_en"=>"Sugar Grove", "lang"=>"en", "lat"=>41.983449, "lon"=>-79.339964),
"/US/US-PA/sugar-notch-18786" => array("type"=>"borough", "pop"=>956, "name_en"=>"Sugar Notch", "lang"=>"en", "lat"=>41.192225, "lon"=>-75.932212),
"/US/US-PA/sugarcreek-18784" => array("type"=>"borough", "pop"=>4969, "name_en"=>"Sugarcreek", "lang"=>"en", "lat"=>41.439450, "lon"=>-79.811494),
"/US/US-PA/summerhill-18787" => array("type"=>"borough", "pop"=>477, "name_en"=>"Summerhill", "lang"=>"en", "lat"=>40.375700, "lon"=>-78.761330),
"/US/US-PA/summerville-18788" => array("type"=>"borough", "pop"=>493, "name_ru"=>"Саммервилл", "name_en"=>"Summerville", "lang"=>"en", "lat"=>41.116364, "lon"=>-79.188631),
"/US/US-PA/summit-hill-18789" => array("type"=>"borough", "pop"=>2956, "name_en"=>"Summit Hill", "lang"=>"en", "lat"=>40.824920, "lon"=>-75.846660),
"/US/US-PA/sunbury-18790" => array("type"=>"city", "pop"=>9819, "name_en"=>"Sunbury", "lang"=>"en", "lat"=>40.862100, "lon"=>-76.788074),
"/US/US-PA/susquehanna-depot-18791" => array("type"=>"borough", "pop"=>1560, "name_en"=>"Susquehanna Depot", "lang"=>"en", "lat"=>41.944521, "lon"=>-75.604398),
"/US/US-PA/susquehanna-trails-18792" => array("type"=>"cdp", "pop"=>1419, "name_en"=>"Susquehanna Trails", "lang"=>"en", "lat"=>39.758501, "lon"=>-76.368433),
"/US/US-PA/sutersville-18793" => array("type"=>"borough", "pop"=>587, "name_en"=>"Sutersville", "lang"=>"en", "lat"=>40.235704, "lon"=>-79.802591),
"/US/US-PA/swarthmore-18794" => array("type"=>"borough", "pop"=>6090, "name_en"=>"Swarthmore", "lang"=>"en", "lat"=>39.902150, "lon"=>-75.348302),
"/US/US-PA/swissvale-18795" => array("type"=>"borough", "pop"=>8765, "name_en"=>"Swissvale", "lang"=>"en", "lat"=>40.420457, "lon"=>-79.886465),
"/US/US-PA/swoyersville-18796" => array("type"=>"borough", "pop"=>4836, "name_en"=>"Swoyersville", "lang"=>"en", "lat"=>41.298550, "lon"=>-75.879024),
"/US/US-PA/sykesville-18797" => array("type"=>"borough", "pop"=>1170, "name_ru"=>"Сайксвилл", "name_en"=>"Sykesville", "lang"=>"en", "lat"=>41.049291, "lon"=>-78.818927),
"/US/US-PA/sylvania-18798" => array("type"=>"borough", "pop"=>186, "name_ru"=>"Сильвания", "name_en"=>"Sylvania", "lang"=>"en", "lat"=>41.805038, "lon"=>-76.857523),
"/US/US-PA/tamaqua-18799" => array("type"=>"borough", "pop"=>6581, "name_en"=>"Tamaqua", "lang"=>"en", "lat"=>40.805242, "lon"=>-75.934855),
"/US/US-PA/tarentum-18800" => array("type"=>"borough", "pop"=>4534, "name_en"=>"Tarentum", "lang"=>"en", "lat"=>40.605158, "lon"=>-79.760224),
"/US/US-PA/tatamy-18801" => array("type"=>"borough", "pop"=>1079, "name_en"=>"Tatamy", "lang"=>"en", "lat"=>40.741216, "lon"=>-75.255269),
"/US/US-PA/taylor-18802" => array("type"=>"borough", "pop"=>6151, "name_ru"=>"Тейлор", "name_en"=>"Taylor", "lang"=>"en", "lat"=>41.398719, "lon"=>-75.713727),
"/US/US-PA/telford-18803" => array("type"=>"borough", "pop"=>4599, "name_en"=>"Telford", "lang"=>"en", "lat"=>40.325508, "lon"=>-75.327324),
"/US/US-PA/temple-18804" => array("type"=>"borough", "pop"=>1491, "name_ru"=>"Тэмпл", "name_en"=>"Temple", "lang"=>"en", "lat"=>40.408300, "lon"=>-75.921035),
"/US/US-PA/terre-hill-18805" => array("type"=>"borough", "pop"=>1261, "name_en"=>"Terre Hill", "lang"=>"en", "lat"=>40.157470, "lon"=>-76.050743),
"/US/US-PA/thompson-18806" => array("type"=>"borough", "pop"=>275, "name_ru"=>"Томпсон", "name_en"=>"Thompson", "lang"=>"en", "lat"=>41.862321, "lon"=>-75.513923),
"/US/US-PA/thompsontown-18807" => array("type"=>"borough", "pop"=>688, "name_en"=>"Thompsontown", "lang"=>"en", "lat"=>40.566164, "lon"=>-77.235309),
"/US/US-PA/thompsonville-18808" => array("type"=>"cdp", "pop"=>3560, "name_ru"=>"Томпсонвилл", "name_en"=>"Thompsonville", "lang"=>"en", "lat"=>40.281304, "lon"=>-80.119888),
"/US/US-PA/thornburg-18809" => array("type"=>"borough", "pop"=>429, "name_ru"=>"Торнбург", "name_en"=>"Thornburg", "lang"=>"en", "lat"=>40.433500, "lon"=>-80.083388),
"/US/US-PA/thorndale-18810" => array("type"=>"cdp", "pop"=>3518, "name_en"=>"Thorndale", "lang"=>"en", "lat"=>39.998500, "lon"=>-75.752159),
"/US/US-PA/three-springs-18811" => array("type"=>"borough", "pop"=>424, "name_en"=>"Three Springs", "lang"=>"en", "lat"=>40.195174, "lon"=>-77.983929),
"/US/US-PA/throop-18812" => array("type"=>"borough", "pop"=>4051, "name_en"=>"Throop", "lang"=>"en", "lat"=>41.438500, "lon"=>-75.595972),
"/US/US-PA/tidioute-18813" => array("type"=>"borough", "pop"=>718, "name_en"=>"Tidioute", "lang"=>"en", "lat"=>41.683554, "lon"=>-79.402123),
"/US/US-PA/timblin-18814" => array("type"=>"borough", "pop"=>142, "name_en"=>"Timblin", "lang"=>"en", "lat"=>40.966117, "lon"=>-79.199348),
"/US/US-PA/tinicum-township-18815" => array("type"=>"cdp", "pop"=>4440, "name_en"=>"Tinicum Township", "lang"=>"en", "lat"=>39.864998, "lon"=>-75.275196),
"/US/US-PA/tioga-18816" => array("type"=>"borough", "pop"=>582, "name_en"=>"Tioga", "lang"=>"en", "lat"=>41.904943, "lon"=>-77.135389),
"/US/US-PA/tionesta-18817" => array("type"=>"borough", "pop"=>567, "name_en"=>"Tionesta", "lang"=>"en", "lat"=>41.495991, "lon"=>-79.446813),
"/US/US-PA/tipton-18818" => array("type"=>"cdp", "pop"=>1194, "name_ru"=>"Типтон", "name_en"=>"Tipton", "lang"=>"en", "lat"=>40.636326, "lon"=>-78.298575),
"/US/US-PA/titusville-18819" => array("type"=>"city", "pop"=>5774, "name_ru"=>"Титусвилл", "name_en"=>"Titusville", "lang"=>"en", "lat"=>41.626721, "lon"=>-79.670937),
"/US/US-PA/topton-18820" => array("type"=>"borough", "pop"=>1965, "name_en"=>"Topton", "lang"=>"en", "lat"=>40.503499, "lon"=>-75.702633),
"/US/US-PA/toughkenamon-18821" => array("type"=>"cdp", "pop"=>1273, "name_en"=>"Toughkenamon", "lang"=>"en", "lat"=>39.830883, "lon"=>-75.755640),
"/US/US-PA/towanda-18822" => array("type"=>"borough", "pop"=>2826, "name_ru"=>"Тованда", "name_en"=>"Towanda", "lang"=>"en", "lat"=>41.770886, "lon"=>-76.446946),
"/US/US-PA/tower-city-18823" => array("type"=>"borough", "pop"=>1318, "name_en"=>"Tower City", "lang"=>"en", "lat"=>40.589547, "lon"=>-76.553774),
"/US/US-PA/townville-18824" => array("type"=>"borough", "pop"=>286, "name_en"=>"Townville", "lang"=>"en", "lat"=>41.679622, "lon"=>-79.881955),
"/US/US-PA/trafford-18825" => array("type"=>"borough", "pop"=>3011, "name_ru"=>"Траффорд", "name_en"=>"Trafford", "lang"=>"en", "lat"=>40.384200, "lon"=>-79.757126),
"/US/US-PA/trainer-18826" => array("type"=>"borough", "pop"=>1831, "name_en"=>"Trainer", "lang"=>"en", "lat"=>39.823595, "lon"=>-75.405027),
"/US/US-PA/trappe-18827" => array("type"=>"borough", "pop"=>3347, "name_ru"=>"Трапп", "name_en"=>"Trappe", "lang"=>"en", "lat"=>40.194100, "lon"=>-75.477414),
"/US/US-PA/treasure-lake-18828" => array("type"=>"cdp", "pop"=>2185, "name_en"=>"Treasure Lake", "lang"=>"en", "lat"=>41.173384, "lon"=>-78.716118),
"/US/US-PA/tremont-18829" => array("type"=>"borough", "pop"=>1696, "name_ru"=>"Тремонт", "name_en"=>"Tremont", "lang"=>"en", "lat"=>40.629745, "lon"=>-76.391947),
"/US/US-PA/tresckow-18830" => array("type"=>"cdp", "pop"=>1033, "name_en"=>"Tresckow", "lang"=>"en", "lat"=>40.915424, "lon"=>-75.965231),
"/US/US-PA/trevorton-18831" => array("type"=>"cdp", "pop"=>2058, "name_en"=>"Trevorton", "lang"=>"en", "lat"=>40.782948, "lon"=>-76.673505),
"/US/US-PA/trooper-18832" => array("type"=>"cdp", "pop"=>5137, "name_en"=>"Trooper", "lang"=>"en", "lat"=>40.149600, "lon"=>-75.399354),
"/US/US-PA/troutville-18833" => array("type"=>"borough", "pop"=>239, "name_en"=>"Troutville", "lang"=>"en", "lat"=>41.024852, "lon"=>-78.785519),
"/US/US-PA/troy-18834" => array("type"=>"borough", "pop"=>1457, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>41.784655, "lon"=>-76.789565),
"/US/US-PA/trumbauersville-18835" => array("type"=>"borough", "pop"=>1071, "name_en"=>"Trumbauersville", "lang"=>"en", "lat"=>40.414100, "lon"=>-75.379421),
"/US/US-PA/tullytown-18836" => array("type"=>"borough", "pop"=>1956, "name_en"=>"Tullytown", "lang"=>"en", "lat"=>40.139295, "lon"=>-74.812780),
"/US/US-PA/tunkhannock-18837" => array("type"=>"borough", "pop"=>1773, "name_en"=>"Tunkhannock", "lang"=>"en", "lat"=>41.541046, "lon"=>-75.949334),
"/US/US-PA/tunnelhill-18838" => array("type"=>"borough", "pop"=>378, "name_en"=>"Tunnelhill", "lang"=>"en", "lat"=>40.473896, "lon"=>-78.543371),
"/US/US-PA/turbotville-18839" => array("type"=>"borough", "pop"=>645, "name_en"=>"Turbotville", "lang"=>"en", "lat"=>41.100974, "lon"=>-76.769408),
"/US/US-PA/turtle-creek-18840" => array("type"=>"borough", "pop"=>5547, "name_en"=>"Turtle Creek", "lang"=>"en", "lat"=>40.408050, "lon"=>-79.820949),
"/US/US-PA/tuscarora-18841" => array("type"=>"cdp", "pop"=>1063, "name_en"=>"Tuscarora", "lang"=>"en", "lat"=>40.760215, "lon"=>-76.057901),
"/US/US-PA/twilight-18842" => array("type"=>"borough", "pop"=>228, "name_en"=>"Twilight", "lang"=>"en", "lat"=>40.114320, "lon"=>-79.890520),
"/US/US-PA/tyler-run-queens-gate-18843" => array("type"=>"cdp", "pop"=>2739, "name_en"=>"Tyler Run-Queens Gate", "lang"=>"en", "lat"=>39.933950, "lon"=>-76.696901),
"/US/US-PA/tyrone-18844" => array("type"=>"borough", "pop"=>5197, "name_ru"=>"Тайрон", "name_en"=>"Tyrone", "lang"=>"en", "lat"=>40.676802, "lon"=>-78.246187),
"/US/US-PA/ulysses-18845" => array("type"=>"borough", "pop"=>631, "name_ru"=>"Улисс", "name_en"=>"Ulysses", "lang"=>"en", "lat"=>41.904135, "lon"=>-77.755209),
"/US/US-PA/union-city-18846" => array("type"=>"borough", "pop"=>3276, "name_ru"=>"Юнион Сити", "name_en"=>"Union City", "lang"=>"en", "lat"=>41.896066, "lon"=>-79.844666),
"/US/US-PA/union-dale-18847" => array("type"=>"borough", "pop"=>341, "name_en"=>"Union Dale", "lang"=>"en", "lat"=>41.714320, "lon"=>-75.483885),
"/US/US-PA/uniontown-18848" => array("type"=>"city", "pop"=>11682, "name_ru"=>"Юнионтаун", "name_en"=>"Uniontown", "lang"=>"en", "lat"=>39.899155, "lon"=>-79.724684),
"/US/US-PA/unionville-18849" => array("type"=>"borough", "pop"=>287, "name_ru"=>"Юнионвилл", "name_en"=>"Unionville", "lang"=>"en", "lat"=>40.905918, "lon"=>-77.876609),
"/US/US-PA/upland-18850" => array("type"=>"borough", "pop"=>2864, "name_ru"=>"Апленд", "name_en"=>"Upland", "lang"=>"en", "lat"=>39.856650, "lon"=>-75.379793),
"/US/US-PA/upper-providence-township-18851" => array("type"=>"cdp", "pop"=>9727, "name_en"=>"Upper Providence Township", "lang"=>"en", "lat"=>39.937655, "lon"=>-75.396488),
"/US/US-PA/upper-st.-clair-18852" => array("type"=>"cdp", "pop"=>18849, "name_en"=>"Upper St. Clair", "lang"=>"en", "lat"=>40.335733, "lon"=>-80.083735),
"/US/US-PA/ursina-18853" => array("type"=>"borough", "pop"=>233, "name_en"=>"Ursina", "lang"=>"en", "lat"=>39.815627, "lon"=>-79.332892),
"/US/US-PA/utica-18854" => array("type"=>"borough", "pop"=>193, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>41.437035, "lon"=>-79.957601),
"/US/US-PA/valencia-18855" => array("type"=>"borough", "pop"=>537, "name_en"=>"Valencia", "lang"=>"en", "lat"=>40.675696, "lon"=>-79.988181),
"/US/US-PA/valley-green-18857" => array("type"=>"cdp", "pop"=>3017, "name_en"=>"Valley Green", "lang"=>"en", "lat"=>40.157189, "lon"=>-76.792970),
"/US/US-PA/valley-hi-18856" => array("type"=>"borough", "pop"=>19, "name_en"=>"Valley-Hi", "lang"=>"en", "lat"=>40.027018, "lon"=>-78.190823),
"/US/US-PA/valley-view-18858" => array("type"=>"cdp", "pop"=>1749, "name_en"=>"Valley View", "lang"=>"en", "lat"=>40.646083, "lon"=>-76.540119),
"/US/US-PA/valley-view-18859" => array("type"=>"cdp", "pop"=>2911, "name_en"=>"Valley View", "lang"=>"en", "lat"=>39.949900, "lon"=>-76.701291),
"/US/US-PA/vanderbilt-18860" => array("type"=>"borough", "pop"=>517, "name_ru"=>"Вандербилт", "name_en"=>"Vanderbilt", "lang"=>"en", "lat"=>40.033149, "lon"=>-79.663003),
"/US/US-PA/vandergrift-18861" => array("type"=>"borough", "pop"=>4997, "name_en"=>"Vandergrift", "lang"=>"en", "lat"=>40.599520, "lon"=>-79.574353),
"/US/US-PA/vandling-18862" => array("type"=>"borough", "pop"=>700, "name_en"=>"Vandling", "lang"=>"en", "lat"=>41.631135, "lon"=>-75.470592),
"/US/US-PA/venango-18863" => array("type"=>"borough", "pop"=>267, "name_en"=>"Venango", "lang"=>"en", "lat"=>41.772758, "lon"=>-80.112004),
"/US/US-PA/verona-18864" => array("type"=>"borough", "pop"=>2842, "name_ru"=>"Верона", "name_en"=>"Verona", "lang"=>"en", "lat"=>40.505050, "lon"=>-79.841149),
"/US/US-PA/versailles-18865" => array("type"=>"borough", "pop"=>1566, "name_ru"=>"Версайллес", "name_en"=>"Versailles", "lang"=>"en", "lat"=>40.317200, "lon"=>-79.830507),
"/US/US-PA/village-green-green-ridge-18866" => array("type"=>"cdp", "pop"=>9026, "name_en"=>"Village Green-Green Ridge", "lang"=>"en", "lat"=>39.863900, "lon"=>-75.424860),
"/US/US-PA/village-shires-18867" => array("type"=>"cdp", "pop"=>4364, "name_en"=>"Village Shires", "lang"=>"en", "lat"=>40.203000, "lon"=>-74.970850),
"/US/US-PA/vinco-18868" => array("type"=>"cdp", "pop"=>1586, "name_en"=>"Vinco", "lang"=>"en", "lat"=>40.409550, "lon"=>-78.842233),
"/US/US-PA/vintondale-18869" => array("type"=>"borough", "pop"=>477, "name_en"=>"Vintondale", "lang"=>"en", "lat"=>40.478563, "lon"=>-78.912528),
"/US/US-PA/volant-18870" => array("type"=>"borough", "pop"=>110, "name_en"=>"Volant", "lang"=>"en", "lat"=>41.113642, "lon"=>-80.259063),
"/US/US-PA/wall-18871" => array("type"=>"borough", "pop"=>664, "name_en"=>"Wall", "lang"=>"en", "lat"=>40.393167, "lon"=>-79.785086),
"/US/US-PA/wallaceton-18872" => array("type"=>"borough", "pop"=>327, "name_en"=>"Wallaceton", "lang"=>"en", "lat"=>40.965484, "lon"=>-78.292414),
"/US/US-PA/walnutport-18873" => array("type"=>"borough", "pop"=>2185, "name_en"=>"Walnutport", "lang"=>"en", "lat"=>40.751533, "lon"=>-75.595614),
"/US/US-PA/wampum-18874" => array("type"=>"borough", "pop"=>623, "name_en"=>"Wampum", "lang"=>"en", "lat"=>40.886467, "lon"=>-80.338797),
"/US/US-PA/warminster-heights-18875" => array("type"=>"cdp", "pop"=>4310, "name_en"=>"Warminster Heights", "lang"=>"en", "lat"=>40.186700, "lon"=>-75.084832),
"/US/US-PA/warren-18876" => array("type"=>"city", "pop"=>9403, "name_ru"=>"Уоррен", "name_en"=>"Warren", "lang"=>"en", "lat"=>41.843727, "lon"=>-79.144050),
"/US/US-PA/warren-south-18877" => array("type"=>"cdp", "pop"=>1780, "name_en"=>"Warren South", "lang"=>"en", "lat"=>41.831530, "lon"=>-79.157466),
"/US/US-PA/warrior-run-18878" => array("type"=>"borough", "pop"=>581, "name_en"=>"Warrior Run", "lang"=>"en", "lat"=>41.187381, "lon"=>-75.952941),
"/US/US-PA/washington-18879" => array("type"=>"city", "pop"=>14754, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>40.174336, "lon"=>-80.247353),
"/US/US-PA/washingtonville-18880" => array("type"=>"borough", "pop"=>187, "name_en"=>"Washingtonville", "lang"=>"en", "lat"=>41.051878, "lon"=>-76.675128),
"/US/US-PA/waterford-18881" => array("type"=>"borough", "pop"=>1413, "name_ru"=>"Уотерфорд", "name_en"=>"Waterford", "lang"=>"en", "lat"=>41.945343, "lon"=>-79.985395),
"/US/US-PA/watsontown-18882" => array("type"=>"borough", "pop"=>2102, "name_en"=>"Watsontown", "lang"=>"en", "lat"=>41.084680, "lon"=>-76.864628),
"/US/US-PA/wattsburg-18883" => array("type"=>"borough", "pop"=>344, "name_en"=>"Wattsburg", "lang"=>"en", "lat"=>42.002816, "lon"=>-79.807069),
"/US/US-PA/waymart-18884" => array("type"=>"borough", "pop"=>3135, "name_en"=>"Waymart", "lang"=>"en", "lat"=>41.587163, "lon"=>-75.409297),
"/US/US-PA/wayne-heights-18885" => array("type"=>"cdp", "pop"=>1683, "name_en"=>"Wayne Heights", "lang"=>"en", "lat"=>39.745171, "lon"=>-77.547608),
"/US/US-PA/waynesboro-18886" => array("type"=>"borough", "pop"=>9947, "name_ru"=>"Уэйнсборо", "name_en"=>"Waynesboro", "lang"=>"en", "lat"=>39.751982, "lon"=>-77.582434),
"/US/US-PA/waynesburg-18887" => array("type"=>"borough", "pop"=>4147, "name_en"=>"Waynesburg", "lang"=>"en", "lat"=>39.898637, "lon"=>-80.185769),
"/US/US-PA/weatherly-18888" => array("type"=>"borough", "pop"=>2588, "name_en"=>"Weatherly", "lang"=>"en", "lat"=>40.943016, "lon"=>-75.821712),
"/US/US-PA/weigelstown-18889" => array("type"=>"cdp", "pop"=>8665, "name_en"=>"Weigelstown", "lang"=>"en", "lat"=>39.987400, "lon"=>-76.829444),
"/US/US-PA/weissport-18890" => array("type"=>"borough", "pop"=>424, "name_en"=>"Weissport", "lang"=>"en", "lat"=>40.829013, "lon"=>-75.701184),
"/US/US-PA/weissport-east-18891" => array("type"=>"cdp", "pop"=>1843, "name_en"=>"Weissport East", "lang"=>"en", "lat"=>40.835535, "lon"=>-75.691054),
"/US/US-PA/wellersburg-18892" => array("type"=>"borough", "pop"=>162, "name_en"=>"Wellersburg", "lang"=>"en", "lat"=>39.727782, "lon"=>-78.849847),
"/US/US-PA/wellsboro-18893" => array("type"=>"borough", "pop"=>3242, "name_en"=>"Wellsboro", "lang"=>"en", "lat"=>41.746128, "lon"=>-77.302933),
"/US/US-PA/wellsville-18894" => array("type"=>"borough", "pop"=>274, "name_ru"=>"Уэллсвилл", "name_en"=>"Wellsville", "lang"=>"en", "lat"=>40.050241, "lon"=>-76.940836),
"/US/US-PA/wernersville-18895" => array("type"=>"borough", "pop"=>2444, "name_en"=>"Wernersville", "lang"=>"en", "lat"=>40.330350, "lon"=>-76.083136),
"/US/US-PA/wesleyville-18896" => array("type"=>"borough", "pop"=>3299, "name_en"=>"Wesleyville", "lang"=>"en", "lat"=>42.137000, "lon"=>-80.012615),
"/US/US-PA/west-alexander-18897" => array("type"=>"borough", "pop"=>301, "name_en"=>"West Alexander", "lang"=>"en", "lat"=>40.104117, "lon"=>-80.508050),
"/US/US-PA/west-brownsville-18898" => array("type"=>"borough", "pop"=>1021, "name_en"=>"West Brownsville", "lang"=>"en", "lat"=>40.031257, "lon"=>-79.892009),
"/US/US-PA/west-chester-18899" => array("type"=>"borough", "pop"=>18315, "name_ru"=>"Вэст Честер", "name_en"=>"West Chester", "lang"=>"en", "lat"=>39.959800, "lon"=>-75.606076),
"/US/US-PA/west-conshohocken-18900" => array("type"=>"borough", "pop"=>1495, "name_en"=>"West Conshohocken", "lang"=>"en", "lat"=>40.069950, "lon"=>-75.318991),
"/US/US-PA/west-easton-18901" => array("type"=>"borough", "pop"=>1160, "name_en"=>"West Easton", "lang"=>"en", "lat"=>40.678200, "lon"=>-75.236523),
"/US/US-PA/west-elizabeth-18902" => array("type"=>"borough", "pop"=>516, "name_en"=>"West Elizabeth", "lang"=>"en", "lat"=>40.273104, "lon"=>-79.893895),
"/US/US-PA/west-fairview-18903" => array("type"=>"borough", "pop"=>1403, "name_en"=>"West Fairview", "lang"=>"en", "lat"=>40.277400, "lon"=>-76.918643),
"/US/US-PA/west-goshen-18905" => array("type"=>"cdp", "pop"=>21192, "name_en"=>"West Goshen", "lang"=>"en", "lat"=>39.968185, "lon"=>-75.572977),
"/US/US-PA/west-grove-18906" => array("type"=>"borough", "pop"=>2761, "name_en"=>"West Grove", "lang"=>"en", "lat"=>39.821391, "lon"=>-75.828911),
"/US/US-PA/west-hazleton-18907" => array("type"=>"borough", "pop"=>3302, "name_en"=>"West Hazleton", "lang"=>"en", "lat"=>40.970733, "lon"=>-76.013167),
"/US/US-PA/west-hills-18908" => array("type"=>"cdp", "pop"=>1240, "name_en"=>"West Hills", "lang"=>"en", "lat"=>40.827928, "lon"=>-79.546979),
"/US/US-PA/west-homestead-18909" => array("type"=>"borough", "pop"=>1997, "name_en"=>"West Homestead", "lang"=>"en", "lat"=>40.399038, "lon"=>-79.920412),
"/US/US-PA/west-kittanning-18910" => array("type"=>"borough", "pop"=>1138, "name_en"=>"West Kittanning", "lang"=>"en", "lat"=>40.811895, "lon"=>-79.531686),
"/US/US-PA/west-lawn-18911" => array("type"=>"borough", "pop"=>1606, "name_en"=>"West Lawn", "lang"=>"en", "lat"=>40.328550, "lon"=>-75.993349),
"/US/US-PA/west-leechburg-18912" => array("type"=>"borough", "pop"=>1256, "name_en"=>"West Leechburg", "lang"=>"en", "lat"=>40.632064, "lon"=>-79.617369),
"/US/US-PA/west-liberty-18913" => array("type"=>"borough", "pop"=>334, "name_ru"=>"Вэст Либерти", "name_en"=>"West Liberty", "lang"=>"en", "lat"=>41.007301, "lon"=>-80.055014),
"/US/US-PA/west-mayfield-18914" => array("type"=>"borough", "pop"=>1079, "name_en"=>"West Mayfield", "lang"=>"en", "lat"=>40.778950, "lon"=>-80.338068),
"/US/US-PA/west-middlesex-18915" => array("type"=>"borough", "pop"=>848, "name_en"=>"West Middlesex", "lang"=>"en", "lat"=>41.174114, "lon"=>-80.456142),
"/US/US-PA/west-middletown-18916" => array("type"=>"borough", "pop"=>133, "name_en"=>"West Middletown", "lang"=>"en", "lat"=>40.245082, "lon"=>-80.425457),
"/US/US-PA/west-mifflin-18917" => array("type"=>"borough", "pop"=>20669, "name_en"=>"West Mifflin", "lang"=>"en", "lat"=>40.357050, "lon"=>-79.907835),
"/US/US-PA/west-newton-18919" => array("type"=>"borough", "pop"=>2848, "name_en"=>"West Newton", "lang"=>"en", "lat"=>40.208370, "lon"=>-79.769579),
"/US/US-PA/west-norriton-18920" => array("type"=>"cdp", "pop"=>14530, "name_en"=>"West Norriton", "lang"=>"en", "lat"=>40.129400, "lon"=>-75.378972),
"/US/US-PA/west-pittston-18922" => array("type"=>"borough", "pop"=>4916, "name_en"=>"West Pittston", "lang"=>"en", "lat"=>41.329400, "lon"=>-75.800033),
"/US/US-PA/west-reading-18923" => array("type"=>"borough", "pop"=>4032, "name_en"=>"West Reading", "lang"=>"en", "lat"=>40.333450, "lon"=>-75.946899),
"/US/US-PA/west-sunbury-18924" => array("type"=>"borough", "pop"=>96, "name_en"=>"West Sunbury", "lang"=>"en", "lat"=>41.006090, "lon"=>-79.896332),
"/US/US-PA/west-view-18925" => array("type"=>"borough", "pop"=>6677, "name_en"=>"West View", "lang"=>"en", "lat"=>40.518300, "lon"=>-80.033868),
"/US/US-PA/west-wyoming-18926" => array("type"=>"borough", "pop"=>2683, "name_en"=>"West Wyoming", "lang"=>"en", "lat"=>41.319350, "lon"=>-75.852444),
"/US/US-PA/west-wyomissing-18927" => array("type"=>"cdp", "pop"=>3097, "name_en"=>"West Wyomissing", "lang"=>"en", "lat"=>40.322700, "lon"=>-75.994927),
"/US/US-PA/west-york-18928" => array("type"=>"borough", "pop"=>4196, "name_en"=>"West York", "lang"=>"en", "lat"=>39.953050, "lon"=>-76.760231),
"/US/US-PA/westfield-18904" => array("type"=>"borough", "pop"=>1116, "name_ru"=>"Уэстфилд", "name_en"=>"Westfield", "lang"=>"en", "lat"=>41.917840, "lon"=>-77.540608),
"/US/US-PA/westmont-18918" => array("type"=>"borough", "pop"=>5033, "name_ru"=>"Уэстмонт", "name_en"=>"Westmont", "lang"=>"en", "lat"=>40.319133, "lon"=>-78.952277),
"/US/US-PA/westover-18921" => array("type"=>"borough", "pop"=>435, "name_en"=>"Westover", "lang"=>"en", "lat"=>40.749588, "lon"=>-78.681816),
"/US/US-PA/wheatland-18929" => array("type"=>"borough", "pop"=>695, "name_ru"=>"Уитленд", "name_en"=>"Wheatland", "lang"=>"en", "lat"=>41.196867, "lon"=>-80.495997),
"/US/US-PA/whitaker-18930" => array("type"=>"borough", "pop"=>1216, "name_en"=>"Whitaker", "lang"=>"en", "lat"=>40.399650, "lon"=>-79.887399),
"/US/US-PA/white-haven-18932" => array("type"=>"borough", "pop"=>1139, "name_en"=>"White Haven", "lang"=>"en", "lat"=>41.055264, "lon"=>-75.777223),
"/US/US-PA/white-oak-18933" => array("type"=>"borough", "pop"=>8003, "name_ru"=>"Уайт Оук", "name_en"=>"White Oak", "lang"=>"en", "lat"=>40.339200, "lon"=>-79.804287),
"/US/US-PA/whitehall-18931" => array("type"=>"borough", "pop"=>13400, "name_ru"=>"Уайтхолл", "name_en"=>"Whitehall", "lang"=>"en", "lat"=>40.359100, "lon"=>-79.991762),
"/US/US-PA/whitfield-18934" => array("type"=>"cdp", "pop"=>2585, "name_ru"=>"Уайтфилд", "name_en"=>"Whitfield", "lang"=>"en", "lat"=>40.335950, "lon"=>-76.006329),
"/US/US-PA/wickerham-manor-fisher-18935" => array("type"=>"cdp", "pop"=>1931, "name_en"=>"Wickerham Manor-Fisher", "lang"=>"en", "lat"=>40.175424, "lon"=>-79.904518),
"/US/US-PA/wilkes-barre-18936" => array("type"=>"city", "pop"=>40932, "name_en"=>"Wilkes-Barre", "lang"=>"en", "lat"=>41.246400, "lon"=>-75.875802),
"/US/US-PA/wilkes-barre-township-18937" => array("type"=>"cdp", "pop"=>3572, "name_en"=>"Wilkes-Barre Township", "lang"=>"en", "lat"=>41.237493, "lon"=>-75.863053),
"/US/US-PA/wilkins-township-18939" => array("type"=>"cdp", "pop"=>7487, "name_en"=>"Wilkins Township", "lang"=>"en", "lat"=>40.426350, "lon"=>-79.824295),
"/US/US-PA/wilkinsburg-18938" => array("type"=>"borough", "pop"=>17522, "name_en"=>"Wilkinsburg", "lang"=>"en", "lat"=>40.444767, "lon"=>-79.873298),
"/US/US-PA/williamsburg-18940" => array("type"=>"borough", "pop"=>1248, "name_ru"=>"Уилльямсбург", "name_en"=>"Williamsburg", "lang"=>"en", "lat"=>40.463871, "lon"=>-78.205725),
"/US/US-PA/williamsport-18941" => array("type"=>"city", "pop"=>29456, "name_ru"=>"Уильямспорт", "name_en"=>"Williamsport", "lang"=>"en", "lat"=>41.238900, "lon"=>-77.037743),
"/US/US-PA/williamstown-18942" => array("type"=>"borough", "pop"=>1350, "name_ru"=>"Уильямстаун", "name_en"=>"Williamstown", "lang"=>"en", "lat"=>40.580685, "lon"=>-76.617885),
"/US/US-PA/willow-grove-18943" => array("type"=>"cdp", "pop"=>16325, "name_en"=>"Willow Grove", "lang"=>"en", "lat"=>40.149000, "lon"=>-75.118473),
"/US/US-PA/willow-street-18944" => array("type"=>"cdp", "pop"=>5817, "name_en"=>"Willow Street", "lang"=>"en", "lat"=>39.984700, "lon"=>-76.270384),
"/US/US-PA/wilmerding-18945" => array("type"=>"borough", "pop"=>1951, "name_en"=>"Wilmerding", "lang"=>"en", "lat"=>40.393583, "lon"=>-79.810920),
"/US/US-PA/wilmore-18946" => array("type"=>"borough", "pop"=>229, "name_ru"=>"Уилмор", "name_en"=>"Wilmore", "lang"=>"en", "lat"=>40.386127, "lon"=>-78.716898),
"/US/US-PA/wilson-18947" => array("type"=>"borough", "pop"=>7625, "name_ru"=>"Уилсон", "name_en"=>"Wilson", "lang"=>"en", "lat"=>40.685250, "lon"=>-75.240071),
"/US/US-PA/wind-gap-18949" => array("type"=>"borough", "pop"=>2785, "name_en"=>"Wind Gap", "lang"=>"en", "lat"=>40.847548, "lon"=>-75.291680),
"/US/US-PA/windber-18948" => array("type"=>"borough", "pop"=>4013, "name_en"=>"Windber", "lang"=>"en", "lat"=>40.233850, "lon"=>-78.825902),
"/US/US-PA/windsor-18950" => array("type"=>"borough", "pop"=>1306, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>39.915478, "lon"=>-76.583285),
"/US/US-PA/winterstown-18951" => array("type"=>"borough", "pop"=>538, "name_en"=>"Winterstown", "lang"=>"en", "lat"=>39.840944, "lon"=>-76.613247),
"/US/US-PA/wolfdale-18952" => array("type"=>"cdp", "pop"=>2906, "name_en"=>"Wolfdale", "lang"=>"en", "lat"=>40.198686, "lon"=>-80.299704),
"/US/US-PA/womelsdorf-18953" => array("type"=>"borough", "pop"=>2789, "name_en"=>"Womelsdorf", "lang"=>"en", "lat"=>40.365636, "lon"=>-76.186729),
"/US/US-PA/woodbourne-18954" => array("type"=>"cdp", "pop"=>2953, "name_en"=>"Woodbourne", "lang"=>"en", "lat"=>40.202108, "lon"=>-74.886428),
"/US/US-PA/woodbury-18955" => array("type"=>"borough", "pop"=>260, "name_ru"=>"Вудбери", "name_en"=>"Woodbury", "lang"=>"en", "lat"=>40.225722, "lon"=>-78.366719),
"/US/US-PA/woodcock-18956" => array("type"=>"borough", "pop"=>136, "name_en"=>"Woodcock", "lang"=>"en", "lat"=>41.753603, "lon"=>-80.084310),
"/US/US-PA/woodland-heights-18957" => array("type"=>"cdp", "pop"=>1471, "name_en"=>"Woodland Heights", "lang"=>"en", "lat"=>41.409275, "lon"=>-79.701786),
"/US/US-PA/woodlyn-18958" => array("type"=>"cdp", "pop"=>10151, "name_en"=>"Woodlyn", "lang"=>"en", "lat"=>39.877400, "lon"=>-75.344139),
"/US/US-PA/woodside-18959" => array("type"=>"cdp", "pop"=>2947, "name_ru"=>"Вудсайд", "name_en"=>"Woodside", "lang"=>"en", "lat"=>40.230287, "lon"=>-74.858685),
"/US/US-PA/woodside-drifton-18960" => array("type"=>"cdp", "pop"=>1360, "name_en"=>"Woodside-Drifton", "lang"=>"en", "lat"=>41.011850, "lon"=>-75.898076),
"/US/US-PA/wormleysburg-18961" => array("type"=>"borough", "pop"=>2622, "name_en"=>"Wormleysburg", "lang"=>"en", "lat"=>40.261800, "lon"=>-76.908270),
"/US/US-PA/worthington-18962" => array("type"=>"borough", "pop"=>721, "name_ru"=>"Уортингтон", "name_en"=>"Worthington", "lang"=>"en", "lat"=>40.837940, "lon"=>-79.634363),
"/US/US-PA/worthville-18963" => array("type"=>"borough", "pop"=>80, "name_ru"=>"Уортвилл", "name_en"=>"Worthville", "lang"=>"en", "lat"=>41.024413, "lon"=>-79.141102),
"/US/US-PA/wrightsville-18964" => array("type"=>"borough", "pop"=>2233, "name_ru"=>"Райтсвилл", "name_en"=>"Wrightsville", "lang"=>"en", "lat"=>40.024384, "lon"=>-76.531534),
"/US/US-PA/wyalusing-18965" => array("type"=>"borough", "pop"=>535, "name_en"=>"Wyalusing", "lang"=>"en", "lat"=>41.669115, "lon"=>-76.262673),
"/US/US-PA/wyncote-18966" => array("type"=>"cdp", "pop"=>2960, "name_en"=>"Wyncote", "lang"=>"en", "lat"=>40.092200, "lon"=>-75.147930),
"/US/US-PA/wyndmoor-18967" => array("type"=>"cdp", "pop"=>5682, "name_en"=>"Wyndmoor", "lang"=>"en", "lat"=>40.083250, "lon"=>-75.193002),
"/US/US-PA/wyoming-18968" => array("type"=>"borough", "pop"=>2995, "name_ru"=>"Вайоминг", "name_en"=>"Wyoming", "lang"=>"en", "lat"=>41.305833, "lon"=>-75.841806),
"/US/US-PA/wyomissing-18969" => array("type"=>"borough", "pop"=>10399, "name_en"=>"Wyomissing", "lang"=>"en", "lat"=>40.330500, "lon"=>-75.967749),
"/US/US-PA/wyomissing-hills-18970" => array("type"=>"borough", "pop"=>2469, "name_en"=>"Wyomissing Hills", "lang"=>"en", "lat"=>40.334750, "lon"=>-75.982006),
"/US/US-PA/yardley-18971" => array("type"=>"borough", "pop"=>2492, "name_en"=>"Yardley", "lang"=>"en", "lat"=>40.239750, "lon"=>-74.838409),
"/US/US-PA/yatesville-18972" => array("type"=>"borough", "pop"=>611, "name_ru"=>"Йетсвилл", "name_en"=>"Yatesville", "lang"=>"en", "lat"=>41.302867, "lon"=>-75.781979),
"/US/US-PA/yeadon-18973" => array("type"=>"borough", "pop"=>11367, "name_en"=>"Yeadon", "lang"=>"en", "lat"=>39.932750, "lon"=>-75.251913),
"/US/US-PA/yeagertown-18974" => array("type"=>"cdp", "pop"=>1150, "name_en"=>"Yeagertown", "lang"=>"en", "lat"=>40.640869, "lon"=>-77.580430),
"/US/US-PA/yoe-18975" => array("type"=>"borough", "pop"=>1013, "name_en"=>"Yoe", "lang"=>"en", "lat"=>39.909717, "lon"=>-76.636627),
"/US/US-PA/york-18976" => array("type"=>"city", "pop"=>40097, "name_ru"=>"Йорк", "name_en"=>"York", "lang"=>"en", "lat"=>39.964500, "lon"=>-76.731805),
"/US/US-PA/york-haven-18978" => array("type"=>"borough", "pop"=>793, "name_en"=>"York Haven", "lang"=>"en", "lat"=>40.110117, "lon"=>-76.715787),
"/US/US-PA/york-springs-18979" => array("type"=>"borough", "pop"=>650, "name_en"=>"York Springs", "lang"=>"en", "lat"=>40.008859, "lon"=>-77.116641),
"/US/US-PA/yorkana-18977" => array("type"=>"borough", "pop"=>274, "name_en"=>"Yorkana", "lang"=>"en", "lat"=>39.974235, "lon"=>-76.584604),
"/US/US-PA/youngstown-18980" => array("type"=>"borough", "pop"=>369, "name_en"=>"Youngstown", "lang"=>"en", "lat"=>40.280886, "lon"=>-79.365771),
"/US/US-PA/youngsville-18981" => array("type"=>"borough", "pop"=>1666, "name_ru"=>"Янгсвилл", "name_en"=>"Youngsville", "lang"=>"en", "lat"=>41.851858, "lon"=>-79.315730),
"/US/US-PA/youngwood-18982" => array("type"=>"borough", "pop"=>3069, "name_en"=>"Youngwood", "lang"=>"en", "lat"=>40.243744, "lon"=>-79.581084),
"/US/US-PA/zelienople-18983" => array("type"=>"borough", "pop"=>3923, "name_en"=>"Zelienople", "lang"=>"en", "lat"=>40.789435, "lon"=>-80.143505),
"/US/US-PA/zion-18984" => array("type"=>"cdp", "pop"=>1573, "name_ru"=>"Зион", "name_en"=>"Zion", "lang"=>"en", "lat"=>40.921406, "lon"=>-77.667880),
"/US/US-PR" => array("type"=>"outlying area", "pop"=>3954037, "name_ru"=>"Пуэрто-рико", "name_en"=>"Puerto Rico", "lang"=>"en"),
"/US/US-PR/aceitunas-23557" => array("type"=>"comunidad", "pop"=>1834, "name_en"=>"Aceitunas", "lang"=>"en", "lat"=>18.447998, "lon"=>-67.067822),
"/US/US-PR/adjuntas-23558" => array("type"=>"zona urbana", "pop"=>5081, "name_en"=>"Adjuntas", "lang"=>"en", "lat"=>18.165742, "lon"=>-66.724091),
"/US/US-PR/aguada-23559" => array("type"=>"zona urbana", "pop"=>4608, "name_en"=>"Aguada", "lang"=>"en", "lat"=>18.382156, "lon"=>-67.187814),
"/US/US-PR/aguadilla-23560" => array("type"=>"zona urbana", "pop"=>18347, "name_en"=>"Aguadilla", "lang"=>"en", "lat"=>18.439712, "lon"=>-67.154909),
"/US/US-PR/aguas-buenas-23561" => array("type"=>"zona urbana", "pop"=>4452, "name_en"=>"Aguas Buenas", "lang"=>"en", "lat"=>18.259409, "lon"=>-66.105429),
"/US/US-PR/aguas-claras-23562" => array("type"=>"comunidad", "pop"=>3091, "name_en"=>"Aguas Claras", "lang"=>"en", "lat"=>18.249486, "lon"=>-65.666332),
"/US/US-PR/aguilita-23563" => array("type"=>"comunidad", "pop"=>4091, "name_en"=>"Aguilita", "lang"=>"en", "lat"=>18.030021, "lon"=>-66.533550),
"/US/US-PR/aibonito-23564" => array("type"=>"zona urbana", "pop"=>9978, "name_en"=>"Aibonito", "lang"=>"en", "lat"=>18.141786, "lon"=>-66.266096),
"/US/US-PR/anasco-23565" => array("type"=>"zona urbana", "pop"=>5826, "name_en"=>"Anasco", "lang"=>"en", "lat"=>18.288466, "lon"=>-67.141345),
"/US/US-PR/animas-23566" => array("type"=>"comunidad", "pop"=>1397, "name_en"=>"Animas", "lang"=>"en", "lat"=>18.447630, "lon"=>-66.635173),
"/US/US-PR/anton-ruiz-23567" => array("type"=>"comunidad", "pop"=>1859, "name_en"=>"Anton Ruiz", "lang"=>"en", "lat"=>18.190113, "lon"=>-65.807633),
"/US/US-PR/arecibo-23568" => array("type"=>"zona urbana", "pop"=>49545, "name_en"=>"Arecibo", "lang"=>"en", "lat"=>18.456723, "lon"=>-66.735899),
"/US/US-PR/arroyo-23569" => array("type"=>"zona urbana", "pop"=>8763, "name_en"=>"Arroyo", "lang"=>"en", "lat"=>17.970815, "lon"=>-66.062111),
"/US/US-PR/bairoa-23570" => array("type"=>"comunidad", "pop"=>2324, "name_en"=>"Bairoa", "lang"=>"en", "lat"=>18.256699, "lon"=>-66.065274),
"/US/US-PR/bajadero-23571" => array("type"=>"comunidad", "pop"=>4112, "name_en"=>"Bajadero", "lang"=>"en", "lat"=>18.425285, "lon"=>-66.676542),
"/US/US-PR/bajandas-23572" => array("type"=>"comunidad", "pop"=>1550, "name_en"=>"Bajandas", "lang"=>"en", "lat"=>18.162302, "lon"=>-65.782826),
"/US/US-PR/barahona-23573" => array("type"=>"comunidad", "pop"=>2246, "name_en"=>"Barahona", "lang"=>"en", "lat"=>18.353443, "lon"=>-66.445945),
"/US/US-PR/barceloneta-23574" => array("type"=>"zona urbana", "pop"=>4873, "name_en"=>"Barceloneta", "lang"=>"en", "lat"=>18.454802, "lon"=>-66.538961),
"/US/US-PR/barranquitas-23575" => array("type"=>"zona urbana", "pop"=>2786, "name_en"=>"Barranquitas", "lang"=>"en", "lat"=>18.187659, "lon"=>-66.309272),
"/US/US-PR/bartolo-23576" => array("type"=>"comunidad", "pop"=>1170, "name_en"=>"Bartolo", "lang"=>"en", "lat"=>18.363466, "lon"=>-65.841402),
"/US/US-PR/bayamon-23577" => array("type"=>"comunidad", "pop"=>1053, "name_en"=>"Bayamon", "lang"=>"en", "lat"=>18.179769, "lon"=>-66.113556),
"/US/US-PR/bayamon-23578" => array("type"=>"zona urbana", "pop"=>202103, "name_en"=>"Bayamon", "lang"=>"en", "lat"=>18.381271, "lon"=>-66.163871),
"/US/US-PR/benitez-23579" => array("type"=>"comunidad", "pop"=>1240, "name_en"=>"Benitez", "lang"=>"en", "lat"=>18.275438, "lon"=>-65.879339),
"/US/US-PR/betances-23580" => array("type"=>"comunidad", "pop"=>1031, "name_en"=>"Betances", "lang"=>"en", "lat"=>18.031497, "lon"=>-67.133448),
"/US/US-PR/boqueron-23581" => array("type"=>"comunidad", "pop"=>1122, "name_en"=>"Boqueron", "lang"=>"en", "lat"=>18.209544, "lon"=>-65.848864),
"/US/US-PR/brenas-23582" => array("type"=>"comunidad", "pop"=>2060, "name_en"=>"Brenas", "lang"=>"en", "lat"=>18.473912, "lon"=>-66.334098),
"/US/US-PR/buena-vista-23583" => array("type"=>"comunidad", "pop"=>2321, "name_ru"=>"Буена Виста", "name_en"=>"Buena Vista", "lang"=>"en", "lat"=>17.998372, "lon"=>-66.052319),
"/US/US-PR/bufalo-23584" => array("type"=>"comunidad", "pop"=>1121, "name_en"=>"Bufalo", "lang"=>"en", "lat"=>18.420353, "lon"=>-66.573686),
"/US/US-PR/caban-23585" => array("type"=>"comunidad", "pop"=>4327, "name_en"=>"Caban", "lang"=>"en", "lat"=>18.445390, "lon"=>-67.136285),
"/US/US-PR/cabo-rojo-23586" => array("type"=>"zona urbana", "pop"=>10131, "name_en"=>"Cabo Rojo", "lang"=>"en", "lat"=>18.088289, "lon"=>-67.148564),
"/US/US-PR/cacao-23587" => array("type"=>"comunidad", "pop"=>1690, "name_en"=>"Cacao", "lang"=>"en", "lat"=>18.438549, "lon"=>-66.937463),
"/US/US-PR/caguas-23588" => array("type"=>"zona urbana", "pop"=>92429, "name_en"=>"Caguas", "lang"=>"en", "lat"=>18.234271, "lon"=>-66.039357),
"/US/US-PR/campanilla-23589" => array("type"=>"comunidad", "pop"=>6681, "name_en"=>"Campanilla", "lang"=>"en", "lat"=>18.424613, "lon"=>-66.238734),
"/US/US-PR/campo-rico-23590" => array("type"=>"comunidad", "pop"=>2952, "name_en"=>"Campo Rico", "lang"=>"en", "lat"=>18.339059, "lon"=>-65.898245),
"/US/US-PR/camuy-23591" => array("type"=>"zona urbana", "pop"=>3890, "name_en"=>"Camuy", "lang"=>"en", "lat"=>18.485664, "lon"=>-66.849072),
"/US/US-PR/candelaria-23592" => array("type"=>"comunidad", "pop"=>16603, "name_en"=>"Candelaria", "lang"=>"en", "lat"=>18.406738, "lon"=>-66.218288),
"/US/US-PR/candelaria-arenas-23593" => array("type"=>"comunidad", "pop"=>5712, "name_en"=>"Candelaria Arenas", "lang"=>"en", "lat"=>18.419187, "lon"=>-66.217841),
"/US/US-PR/candelero-arriba-23594" => array("type"=>"comunidad", "pop"=>1491, "name_en"=>"Candelero Arriba", "lang"=>"en", "lat"=>18.104025, "lon"=>-65.837086),
"/US/US-PR/canovanas-23595" => array("type"=>"zona urbana", "pop"=>7697, "name_en"=>"Canovanas", "lang"=>"en", "lat"=>18.379678, "lon"=>-65.906789),
"/US/US-PR/capitanejo-23596" => array("type"=>"comunidad", "pop"=>2281, "name_en"=>"Capitanejo", "lang"=>"en", "lat"=>18.015284, "lon"=>-66.536067),
"/US/US-PR/carolina-23597" => array("type"=>"zona urbana", "pop"=>162404, "name_ru"=>"Каролина", "name_en"=>"Carolina", "lang"=>"en", "lat"=>18.408384, "lon"=>-65.981091),
"/US/US-PR/carrizales-23598" => array("type"=>"comunidad", "pop"=>1672, "name_en"=>"Carrizales", "lang"=>"en", "lat"=>18.483365, "lon"=>-66.788111),
"/US/US-PR/catano-23599" => array("type"=>"zona urbana", "pop"=>34587, "name_en"=>"Catano", "lang"=>"en", "lat"=>18.445414, "lon"=>-66.139425),
"/US/US-PR/cayey-23600" => array("type"=>"zona urbana", "pop"=>23332, "name_en"=>"Cayey", "lang"=>"en", "lat"=>18.116389, "lon"=>-66.162702),
"/US/US-PR/cayuco-23601" => array("type"=>"comunidad", "pop"=>1163, "name_en"=>"Cayuco", "lang"=>"en", "lat"=>18.295254, "lon"=>-66.735432),
"/US/US-PR/ceiba-23602" => array("type"=>"comunidad", "pop"=>3454, "name_en"=>"Ceiba", "lang"=>"en", "lat"=>18.444920, "lon"=>-66.346609),
"/US/US-PR/ceiba-23603" => array("type"=>"zona urbana", "pop"=>6289, "name_en"=>"Ceiba", "lang"=>"en", "lat"=>18.266585, "lon"=>-65.649010),
"/US/US-PR/celada-23604" => array("type"=>"comunidad", "pop"=>4158, "name_en"=>"Celada", "lang"=>"en", "lat"=>18.271965, "lon"=>-65.962566),
"/US/US-PR/central-aguirre-23605" => array("type"=>"comunidad", "pop"=>1601, "name_en"=>"Central Aguirre", "lang"=>"en", "lat"=>17.956956, "lon"=>-66.226613),
"/US/US-PR/ciales-23606" => array("type"=>"zona urbana", "pop"=>3370, "name_en"=>"Ciales", "lang"=>"en", "lat"=>18.336028, "lon"=>-66.470989),
"/US/US-PR/cidra-23607" => array("type"=>"zona urbana", "pop"=>5580, "name_en"=>"Cidra", "lang"=>"en", "lat"=>18.179226, "lon"=>-66.159174),
"/US/US-PR/coamo-23608" => array("type"=>"zona urbana", "pop"=>13266, "name_en"=>"Coamo", "lang"=>"en", "lat"=>18.081052, "lon"=>-66.360581),
"/US/US-PR/coco-23609" => array("type"=>"comunidad", "pop"=>3478, "name_en"=>"Coco", "lang"=>"en", "lat"=>18.004544, "lon"=>-66.260035),
"/US/US-PR/comerio-23610" => array("type"=>"zona urbana", "pop"=>4978, "name_en"=>"Comerio", "lang"=>"en", "lat"=>18.221736, "lon"=>-66.225024),
"/US/US-PR/comunas-23611" => array("type"=>"comunidad", "pop"=>1808, "name_en"=>"Comunas", "lang"=>"en", "lat"=>18.089511, "lon"=>-65.840404),
"/US/US-PR/coqui-23612" => array("type"=>"comunidad", "pop"=>3051, "name_en"=>"Coqui", "lang"=>"en", "lat"=>17.980552, "lon"=>-66.226435),
"/US/US-PR/corazon-23613" => array("type"=>"comunidad", "pop"=>2747, "name_en"=>"Corazon", "lang"=>"en", "lat"=>17.995494, "lon"=>-66.083589),
"/US/US-PR/corcovado-23614" => array("type"=>"comunidad", "pop"=>1674, "name_en"=>"Corcovado", "lang"=>"en", "lat"=>18.463186, "lon"=>-66.777169),
"/US/US-PR/corozal-23615" => array("type"=>"zona urbana", "pop"=>11746, "name_en"=>"Corozal", "lang"=>"en", "lat"=>18.343668, "lon"=>-66.312669),
"/US/US-PR/coto-laurel-23616" => array("type"=>"comunidad", "pop"=>4044, "name_en"=>"Coto Laurel", "lang"=>"en", "lat"=>18.049855, "lon"=>-66.551938),
"/US/US-PR/coto-norte-23617" => array("type"=>"comunidad", "pop"=>1945, "name_en"=>"Coto Norte", "lang"=>"en", "lat"=>18.435244, "lon"=>-66.441115),
"/US/US-PR/culebra-23618" => array("type"=>"zona urbana", "pop"=>1244, "name_en"=>"Culebra", "lang"=>"en", "lat"=>18.313222, "lon"=>-65.305675),
"/US/US-PR/daguao-23619" => array("type"=>"comunidad", "pop"=>1628, "name_en"=>"Daguao", "lang"=>"en", "lat"=>18.223489, "lon"=>-65.680678),
"/US/US-PR/dorado-23620" => array("type"=>"zona urbana", "pop"=>12003, "name_en"=>"Dorado", "lang"=>"en", "lat"=>18.470150, "lon"=>-66.271509),
"/US/US-PR/duque-23621" => array("type"=>"comunidad", "pop"=>1445, "name_en"=>"Duque", "lang"=>"en", "lat"=>18.240067, "lon"=>-65.744589),
"/US/US-PR/el-mango-23622" => array("type"=>"comunidad", "pop"=>1424, "name_en"=>"El Mango", "lang"=>"en", "lat"=>18.236896, "lon"=>-65.875563),
"/US/US-PR/el-negro-23623" => array("type"=>"comunidad", "pop"=>1387, "name_en"=>"El Negro", "lang"=>"en", "lat"=>18.039345, "lon"=>-65.851721),
"/US/US-PR/el-ojo-23624" => array("type"=>"comunidad", "pop"=>1674, "name_en"=>"El Ojo", "lang"=>"en", "lat"=>18.005861, "lon"=>-66.392068),
"/US/US-PR/emajagua-23625" => array("type"=>"comunidad", "pop"=>2457, "name_en"=>"Emajagua", "lang"=>"en", "lat"=>18.002532, "lon"=>-65.883038),
"/US/US-PR/esperanza-23626" => array("type"=>"comunidad", "pop"=>1183, "name_en"=>"Esperanza", "lang"=>"en", "lat"=>18.103179, "lon"=>-65.480016),
"/US/US-PR/espino-23627" => array("type"=>"comunidad", "pop"=>1327, "name_en"=>"Espino", "lang"=>"en", "lat"=>18.278291, "lon"=>-67.119753),
"/US/US-PR/fajardo-23628" => array("type"=>"zona urbana", "pop"=>31659, "name_en"=>"Fajardo", "lang"=>"en", "lat"=>18.335017, "lon"=>-65.659686),
"/US/US-PR/florida-23629" => array("type"=>"zona urbana", "pop"=>5748, "name_ru"=>"Флорида", "name_en"=>"Florida", "lang"=>"en", "lat"=>18.364391, "lon"=>-66.561529),
"/US/US-PR/franquez-23630" => array("type"=>"comunidad", "pop"=>2052, "name_en"=>"Franquez", "lang"=>"en", "lat"=>18.342136, "lon"=>-66.428001),
"/US/US-PR/fuig-23631" => array("type"=>"comunidad", "pop"=>1172, "name_en"=>"Fuig", "lang"=>"en", "lat"=>17.988326, "lon"=>-66.917689),
"/US/US-PR/g.-l.-garcia-23634" => array("type"=>"comunidad", "pop"=>1703, "name_en"=>"G. L. Garcia", "lang"=>"en", "lat"=>18.129539, "lon"=>-66.104317),
"/US/US-PR/galateo-23632" => array("type"=>"comunidad", "pop"=>2475, "name_en"=>"Galateo", "lang"=>"en", "lat"=>18.364362, "lon"=>-66.259284),
"/US/US-PR/garrochales-23633" => array("type"=>"comunidad", "pop"=>1316, "name_en"=>"Garrochales", "lang"=>"en", "lat"=>18.464356, "lon"=>-66.570133),
"/US/US-PR/guanica-23635" => array("type"=>"zona urbana", "pop"=>9263, "name_en"=>"Guanica", "lang"=>"en", "lat"=>17.970369, "lon"=>-66.930667),
"/US/US-PR/guayabal-23636" => array("type"=>"comunidad", "pop"=>2379, "name_en"=>"Guayabal", "lang"=>"en", "lat"=>18.078176, "lon"=>-66.502110),
"/US/US-PR/guayama-23637" => array("type"=>"zona urbana", "pop"=>21692, "name_en"=>"Guayama", "lang"=>"en", "lat"=>17.978166, "lon"=>-66.111004),
"/US/US-PR/guayanilla-23638" => array("type"=>"zona urbana", "pop"=>5469, "name_en"=>"Guayanilla", "lang"=>"en", "lat"=>18.024881, "lon"=>-66.789968),
"/US/US-PR/guaynabo-23639" => array("type"=>"zona urbana", "pop"=>73385, "name_en"=>"Guaynabo", "lang"=>"en", "lat"=>18.386438, "lon"=>-66.114343),
"/US/US-PR/gurabo-23640" => array("type"=>"zona urbana", "pop"=>9199, "name_en"=>"Gurabo", "lang"=>"en", "lat"=>18.256130, "lon"=>-65.978042),
"/US/US-PR/h.-rivera-colon-23644" => array("type"=>"comunidad", "pop"=>2024, "name_en"=>"H. Rivera Colon", "lang"=>"en", "lat"=>18.349664, "lon"=>-66.274094),
"/US/US-PR/hatillo-23641" => array("type"=>"zona urbana", "pop"=>5242, "name_en"=>"Hatillo", "lang"=>"en", "lat"=>18.487172, "lon"=>-66.823500),
"/US/US-PR/hato-candal-23642" => array("type"=>"comunidad", "pop"=>2011, "name_en"=>"Hato Candal", "lang"=>"en", "lat"=>18.376556, "lon"=>-65.789880),
"/US/US-PR/hormigueros-23643" => array("type"=>"zona urbana", "pop"=>13585, "name_en"=>"Hormigueros", "lang"=>"en", "lat"=>18.144351, "lon"=>-67.120759),
"/US/US-PR/humacao-23645" => array("type"=>"zona urbana", "pop"=>21306, "name_en"=>"Humacao", "lang"=>"en", "lat"=>18.154510, "lon"=>-65.819726),
"/US/US-PR/imbery-23646" => array("type"=>"comunidad", "pop"=>2554, "name_en"=>"Imbery", "lang"=>"en", "lat"=>18.439228, "lon"=>-66.557262),
"/US/US-PR/indios-23647" => array("type"=>"comunidad", "pop"=>1595, "name_en"=>"Indios", "lang"=>"en", "lat"=>17.996138, "lon"=>-66.819755),
"/US/US-PR/ingenio-23648" => array("type"=>"comunidad", "pop"=>5111, "name_en"=>"Ingenio", "lang"=>"en", "lat"=>18.447279, "lon"=>-66.225843),
"/US/US-PR/isabela-23649" => array("type"=>"zona urbana", "pop"=>13515, "name_en"=>"Isabela", "lang"=>"en", "lat"=>18.505278, "lon"=>-67.021040),
"/US/US-PR/jagual-23650" => array("type"=>"comunidad", "pop"=>1061, "name_en"=>"Jagual", "lang"=>"en", "lat"=>18.162779, "lon"=>-65.997726),
"/US/US-PR/jauca-23651" => array("type"=>"comunidad", "pop"=>1871, "name_en"=>"Jauca", "lang"=>"en", "lat"=>17.965842, "lon"=>-66.365961),
"/US/US-PR/jayuya-23652" => array("type"=>"zona urbana", "pop"=>3702, "name_en"=>"Jayuya", "lang"=>"en", "lat"=>18.220457, "lon"=>-66.596969),
"/US/US-PR/jobos-23653" => array("type"=>"comunidad", "pop"=>3772, "name_en"=>"Jobos", "lang"=>"en", "lat"=>17.962530, "lon"=>-66.165697),
"/US/US-PR/juana-diaz-23654" => array("type"=>"zona urbana", "pop"=>9660, "name_en"=>"Juana Diaz", "lang"=>"en", "lat"=>18.054809, "lon"=>-66.505314),
"/US/US-PR/juncal-23655" => array("type"=>"comunidad", "pop"=>1122, "name_en"=>"Juncal", "lang"=>"en", "lat"=>18.315695, "lon"=>-66.919381),
"/US/US-PR/juncos-23656" => array("type"=>"zona urbana", "pop"=>9132, "name_en"=>"Juncos", "lang"=>"en", "lat"=>18.228513, "lon"=>-65.916762),
"/US/US-PR/la-alianza-23657" => array("type"=>"comunidad", "pop"=>1901, "name_en"=>"La Alianza", "lang"=>"en", "lat"=>18.402399, "lon"=>-66.601796),
"/US/US-PR/la-dolores-23658" => array("type"=>"comunidad", "pop"=>3255, "name_en"=>"La Dolores", "lang"=>"en", "lat"=>18.377331, "lon"=>-65.855155),
"/US/US-PR/la-fermina-23659" => array("type"=>"comunidad", "pop"=>2525, "name_en"=>"La Fermina", "lang"=>"en", "lat"=>18.177386, "lon"=>-65.853766),
"/US/US-PR/la-luisa-23661" => array("type"=>"comunidad", "pop"=>2448, "name_en"=>"La Luisa", "lang"=>"en", "lat"=>18.450832, "lon"=>-66.510163),
"/US/US-PR/la-parguera-23663" => array("type"=>"comunidad", "pop"=>1192, "name_en"=>"La Parguera", "lang"=>"en", "lat"=>17.976823, "lon"=>-67.046864),
"/US/US-PR/la-playa-23664" => array("type"=>"comunidad", "pop"=>2165, "name_en"=>"La Playa", "lang"=>"en", "lat"=>18.289446, "lon"=>-67.187353),
"/US/US-PR/la-plena-23665" => array("type"=>"comunidad", "pop"=>1320, "name_en"=>"La Plena", "lang"=>"en", "lat"=>18.044329, "lon"=>-66.206562),
"/US/US-PR/lajas-23660" => array("type"=>"zona urbana", "pop"=>4495, "name_en"=>"Lajas", "lang"=>"en", "lat"=>18.046367, "lon"=>-67.060264),
"/US/US-PR/lamboglia-23662" => array("type"=>"comunidad", "pop"=>1149, "name_en"=>"Lamboglia", "lang"=>"en", "lat"=>17.986346, "lon"=>-65.984942),
"/US/US-PR/lares-23666" => array("type"=>"zona urbana", "pop"=>5627, "name_en"=>"Lares", "lang"=>"en", "lat"=>18.296274, "lon"=>-66.882222),
"/US/US-PR/las-marias-23667" => array("type"=>"comunidad", "pop"=>1785, "name_en"=>"Las Marias", "lang"=>"en", "lat"=>18.294321, "lon"=>-67.144848),
"/US/US-PR/las-marias-23668" => array("type"=>"zona urbana", "pop"=>901, "name_en"=>"Las Marias", "lang"=>"en", "lat"=>18.255761, "lon"=>-66.986245),
"/US/US-PR/las-ochenta-23669" => array("type"=>"comunidad", "pop"=>1275, "name_en"=>"Las Ochenta", "lang"=>"en", "lat"=>17.985779, "lon"=>-66.318086),
"/US/US-PR/las-ollas-23670" => array("type"=>"comunidad", "pop"=>1893, "name_en"=>"Las Ollas", "lang"=>"en", "lat"=>18.016356, "lon"=>-66.421694),
"/US/US-PR/las-piedras-23671" => array("type"=>"zona urbana", "pop"=>6195, "name_en"=>"Las Piedras", "lang"=>"en", "lat"=>18.181269, "lon"=>-65.869846),
"/US/US-PR/levittown-23672" => array("type"=>"comunidad", "pop"=>30807, "name_en"=>"Levittown", "lang"=>"en", "lat"=>18.447607, "lon"=>-66.176331),
"/US/US-PR/liborio-negron-torres-23673" => array("type"=>"comunidad", "pop"=>1246, "name_en"=>"Liborio Negron Torres", "lang"=>"en", "lat"=>18.045121, "lon"=>-66.942814),
"/US/US-PR/lluveras-23674" => array("type"=>"comunidad", "pop"=>1727, "name_en"=>"Lluveras", "lang"=>"en", "lat"=>18.040368, "lon"=>-66.905082),
"/US/US-PR/loiza-23675" => array("type"=>"zona urbana", "pop"=>4300, "name_en"=>"Loiza", "lang"=>"en", "lat"=>18.432818, "lon"=>-65.880628),
"/US/US-PR/lomas-23676" => array("type"=>"comunidad", "pop"=>1077, "name_en"=>"Lomas", "lang"=>"en", "lat"=>18.270543, "lon"=>-65.909164),
"/US/US-PR/los-llanos-23677" => array("type"=>"comunidad", "pop"=>2216, "name_en"=>"Los Llanos", "lang"=>"en", "lat"=>18.058239, "lon"=>-66.409626),
"/US/US-PR/luis-llorens-torres-23678" => array("type"=>"comunidad", "pop"=>2990, "name_en"=>"Luis Llorens Torres", "lang"=>"en", "lat"=>18.058793, "lon"=>-66.527348),
"/US/US-PR/luis-m.-cintron-23679" => array("type"=>"comunidad", "pop"=>1926, "name_en"=>"Luis M. Cintron", "lang"=>"en", "lat"=>18.301600, "lon"=>-65.638938),
"/US/US-PR/luquillo-23680" => array("type"=>"zona urbana", "pop"=>8672, "name_en"=>"Luquillo", "lang"=>"en", "lat"=>18.377486, "lon"=>-65.721034),
"/US/US-PR/luyando-23681" => array("type"=>"comunidad", "pop"=>3121, "name_en"=>"Luyando", "lang"=>"en", "lat"=>18.358423, "lon"=>-67.156291),
"/US/US-PR/magas-arriba-23682" => array("type"=>"comunidad", "pop"=>1352, "name_en"=>"Magas Arriba", "lang"=>"en", "lat"=>18.024870, "lon"=>-66.765829),
"/US/US-PR/manati-23683" => array("type"=>"zona urbana", "pop"=>16352, "name_en"=>"Manati", "lang"=>"en", "lat"=>18.430958, "lon"=>-66.484242),
"/US/US-PR/maria-antonia-23684" => array("type"=>"comunidad", "pop"=>1265, "name_en"=>"Maria Antonia", "lang"=>"en", "lat"=>17.980259, "lon"=>-66.889706),
"/US/US-PR/mariano-colon-23685" => array("type"=>"comunidad", "pop"=>1543, "name_en"=>"Mariano Colon", "lang"=>"en", "lat"=>18.025370, "lon"=>-66.332785),
"/US/US-PR/maricao-23686" => array("type"=>"zona urbana", "pop"=>1058, "name_en"=>"Maricao", "lang"=>"en", "lat"=>18.186137, "lon"=>-66.980430),
"/US/US-PR/martorell-23687" => array("type"=>"comunidad", "pop"=>2889, "name_en"=>"Martorell", "lang"=>"en", "lat"=>18.075391, "lon"=>-65.898167),
"/US/US-PR/marueno-23688" => array("type"=>"comunidad", "pop"=>1202, "name_en"=>"Marueno", "lang"=>"en", "lat"=>18.058626, "lon"=>-66.656349),
"/US/US-PR/maunabo-23689" => array("type"=>"zona urbana", "pop"=>2516, "name_en"=>"Maunabo", "lang"=>"en", "lat"=>18.009833, "lon"=>-65.901437),
"/US/US-PR/mayaguez-23690" => array("type"=>"zona urbana", "pop"=>83010, "name_en"=>"Mayaguez", "lang"=>"en", "lat"=>18.203568, "lon"=>-67.143361),
"/US/US-PR/miranda-23691" => array("type"=>"comunidad", "pop"=>1878, "name_en"=>"Miranda", "lang"=>"en", "lat"=>18.388669, "lon"=>-66.380249),
"/US/US-PR/moca-23692" => array("type"=>"zona urbana", "pop"=>4231, "name_en"=>"Moca", "lang"=>"en", "lat"=>18.396600, "lon"=>-67.114157),
"/US/US-PR/monserrate-23693" => array("type"=>"comunidad", "pop"=>2647, "name_en"=>"Monserrate", "lang"=>"en", "lat"=>18.439155, "lon"=>-66.356775),
"/US/US-PR/mora-23694" => array("type"=>"comunidad", "pop"=>1706, "name_ru"=>"Мора", "name_en"=>"Mora", "lang"=>"en", "lat"=>18.465017, "lon"=>-67.032970),
"/US/US-PR/morovis-23695" => array("type"=>"zona urbana", "pop"=>2355, "name_en"=>"Morovis", "lang"=>"en", "lat"=>18.326822, "lon"=>-66.405195),
"/US/US-PR/mucarabones-23696" => array("type"=>"comunidad", "pop"=>1670, "name_en"=>"Mucarabones", "lang"=>"en", "lat"=>18.392735, "lon"=>-66.216521),
"/US/US-PR/naguabo-23697" => array("type"=>"zona urbana", "pop"=>4855, "name_en"=>"Naguabo", "lang"=>"en", "lat"=>18.213511, "lon"=>-65.737612),
"/US/US-PR/naranjito-23698" => array("type"=>"zona urbana", "pop"=>2368, "name_en"=>"Naranjito", "lang"=>"en", "lat"=>18.305126, "lon"=>-66.245028),
"/US/US-PR/olimpo-23699" => array("type"=>"comunidad", "pop"=>2460, "name_en"=>"Olimpo", "lang"=>"en", "lat"=>18.004666, "lon"=>-66.109476),
"/US/US-PR/orocovis-23700" => array("type"=>"zona urbana", "pop"=>1032, "name_en"=>"Orocovis", "lang"=>"en", "lat"=>18.226943, "lon"=>-66.391814),
"/US/US-PR/pajaros-23701" => array("type"=>"comunidad", "pop"=>6877, "name_en"=>"Pajaros", "lang"=>"en", "lat"=>18.362071, "lon"=>-66.219074),
"/US/US-PR/palmarejo-23702" => array("type"=>"comunidad", "pop"=>1088, "name_en"=>"Palmarejo", "lang"=>"en", "lat"=>18.070041, "lon"=>-66.325960),
"/US/US-PR/palmarejo-23703" => array("type"=>"comunidad", "pop"=>1150, "name_en"=>"Palmarejo", "lang"=>"en", "lat"=>18.043455, "lon"=>-67.077541),
"/US/US-PR/palmas-23704" => array("type"=>"comunidad", "pop"=>2254, "name_en"=>"Palmas", "lang"=>"en", "lat"=>17.989038, "lon"=>-66.025737),
"/US/US-PR/palmer-23705" => array("type"=>"comunidad", "pop"=>1472, "name_ru"=>"Палмер", "name_en"=>"Palmer", "lang"=>"en", "lat"=>18.372368, "lon"=>-65.774493),
"/US/US-PR/palo-seco-23708" => array("type"=>"comunidad", "pop"=>1221, "name_en"=>"Palo Seco", "lang"=>"en", "lat"=>18.009563, "lon"=>-65.937330),
"/US/US-PR/palomas-23706" => array("type"=>"comunidad", "pop"=>1833, "name_en"=>"Palomas", "lang"=>"en", "lat"=>18.238030, "lon"=>-66.250620),
"/US/US-PR/palomas-23707" => array("type"=>"comunidad", "pop"=>2393, "name_en"=>"Palomas", "lang"=>"en", "lat"=>18.015422, "lon"=>-66.868731),
"/US/US-PR/parcelas-la-milagrosa-23709" => array("type"=>"comunidad", "pop"=>1161, "name_en"=>"Parcelas La Milagrosa", "lang"=>"en", "lat"=>18.172741, "lon"=>-66.188474),
"/US/US-PR/parcelas-nuevas-23710" => array("type"=>"comunidad", "pop"=>1279, "name_en"=>"Parcelas Nuevas", "lang"=>"en", "lat"=>18.142231, "lon"=>-66.172726),
"/US/US-PR/parcelas-penuelas-23711" => array("type"=>"comunidad", "pop"=>1099, "name_en"=>"Parcelas Penuelas", "lang"=>"en", "lat"=>18.004020, "lon"=>-66.340102),
"/US/US-PR/pastos-23712" => array("type"=>"comunidad", "pop"=>1538, "name_en"=>"Pastos", "lang"=>"en", "lat"=>18.119712, "lon"=>-66.259050),
"/US/US-PR/patillas-23713" => array("type"=>"zona urbana", "pop"=>4410, "name_en"=>"Patillas", "lang"=>"en", "lat"=>18.009035, "lon"=>-66.013899),
"/US/US-PR/pena-pobre-23714" => array("type"=>"comunidad", "pop"=>1127, "name_en"=>"Pena Pobre", "lang"=>"en", "lat"=>18.218734, "lon"=>-65.822882),
"/US/US-PR/penuelas-23715" => array("type"=>"zona urbana", "pop"=>5918, "name_en"=>"Penuelas", "lang"=>"en", "lat"=>18.061508, "lon"=>-66.721403),
"/US/US-PR/piedra-gorda-23716" => array("type"=>"comunidad", "pop"=>1771, "name_en"=>"Piedra Gorda", "lang"=>"en", "lat"=>18.436049, "lon"=>-66.888006),
"/US/US-PR/playa-fortuna-23717" => array("type"=>"comunidad", "pop"=>2232, "name_en"=>"Playa Fortuna", "lang"=>"en", "lat"=>18.381757, "lon"=>-65.745449),
"/US/US-PR/playita-23718" => array("type"=>"comunidad", "pop"=>2399, "name_en"=>"Playita", "lang"=>"en", "lat"=>18.041658, "lon"=>-65.908335),
"/US/US-PR/playita-cortada-23719" => array("type"=>"comunidad", "pop"=>1521, "name_en"=>"Playita Cortada", "lang"=>"en", "lat"=>17.978772, "lon"=>-66.444717),
"/US/US-PR/pole-ojea-23720" => array("type"=>"comunidad", "pop"=>1539, "name_en"=>"Pole Ojea", "lang"=>"en", "lat"=>17.978896, "lon"=>-67.184735),
"/US/US-PR/ponce-23721" => array("type"=>"zona urbana", "pop"=>159151, "name_en"=>"Ponce", "lang"=>"en", "lat"=>17.987145, "lon"=>-66.609061),
"/US/US-PR/potala-pastillo-23722" => array("type"=>"comunidad", "pop"=>4129, "name_en"=>"Potala Pastillo", "lang"=>"en", "lat"=>17.993464, "lon"=>-66.496941),
"/US/US-PR/pueblito-del-rio-23723" => array("type"=>"comunidad", "pop"=>1361, "name_en"=>"Pueblito del Rio", "lang"=>"en", "lat"=>18.230027, "lon"=>-65.863291),
"/US/US-PR/puerto-real-23724" => array("type"=>"comunidad", "pop"=>3429, "name_en"=>"Puerto Real", "lang"=>"en", "lat"=>18.072680, "lon"=>-67.191123),
"/US/US-PR/punta-santiago-23725" => array("type"=>"comunidad", "pop"=>6112, "name_en"=>"Punta Santiago", "lang"=>"en", "lat"=>18.164672, "lon"=>-65.758425),
"/US/US-PR/quebrada-23726" => array("type"=>"comunidad", "pop"=>1191, "name_en"=>"Quebrada", "lang"=>"en", "lat"=>18.362096, "lon"=>-66.836317),
"/US/US-PR/quebradillas-23727" => array("type"=>"zona urbana", "pop"=>4574, "name_en"=>"Quebradillas", "lang"=>"en", "lat"=>18.475841, "lon"=>-66.936284),
"/US/US-PR/rafael-capo-23728" => array("type"=>"comunidad", "pop"=>1933, "name_en"=>"Rafael Capo", "lang"=>"en", "lat"=>18.405606, "lon"=>-66.783562),
"/US/US-PR/rafael-gonzalez-23729" => array("type"=>"comunidad", "pop"=>2446, "name_en"=>"Rafael Gonzalez", "lang"=>"en", "lat"=>18.429463, "lon"=>-66.787279),
"/US/US-PR/rafael-hernandez-23730" => array("type"=>"comunidad", "pop"=>1374, "name_en"=>"Rafael Hernandez", "lang"=>"en", "lat"=>18.473321, "lon"=>-67.079489),
"/US/US-PR/ramos-23731" => array("type"=>"comunidad", "pop"=>1977, "name_en"=>"Ramos", "lang"=>"en", "lat"=>18.341576, "lon"=>-65.712981),
"/US/US-PR/rincon-23732" => array("type"=>"zona urbana", "pop"=>1319, "name_ru"=>"Ринкон", "name_en"=>"Rincon", "lang"=>"en", "lat"=>18.340953, "lon"=>-67.252636),
"/US/US-PR/rio-blanco-23733" => array("type"=>"comunidad", "pop"=>1341, "name_en"=>"Rio Blanco", "lang"=>"en", "lat"=>18.214409, "lon"=>-65.793186),
"/US/US-PR/rio-canas-abajo-23734" => array("type"=>"comunidad", "pop"=>1245, "name_en"=>"Rio Canas Abajo", "lang"=>"en", "lat"=>18.041075, "lon"=>-66.468680),
"/US/US-PR/rio-grande-23735" => array("type"=>"zona urbana", "pop"=>15428, "name_en"=>"Rio Grande", "lang"=>"en", "lat"=>18.380687, "lon"=>-65.839327),
"/US/US-PR/rio-lajas-23736" => array("type"=>"comunidad", "pop"=>2268, "name_en"=>"Rio Lajas", "lang"=>"en", "lat"=>18.400419, "lon"=>-66.262796),
"/US/US-PR/roosevelt-roads-23737" => array("type"=>"comunidad", "pop"=>4158, "name_en"=>"Roosevelt Roads", "lang"=>"en", "lat"=>18.238193, "lon"=>-65.627778),
"/US/US-PR/rosa-sanchez-23738" => array("type"=>"comunidad", "pop"=>1402, "name_en"=>"Rosa Sanchez", "lang"=>"en", "lat"=>18.063526, "lon"=>-65.913767),
"/US/US-PR/sabana-23739" => array("type"=>"comunidad", "pop"=>1207, "name_en"=>"Sabana", "lang"=>"en", "lat"=>18.464048, "lon"=>-66.352810),
"/US/US-PR/sabana-eneas-23740" => array("type"=>"comunidad", "pop"=>1817, "name_en"=>"Sabana Eneas", "lang"=>"en", "lat"=>18.084143, "lon"=>-67.085376),
"/US/US-PR/sabana-grande-23741" => array("type"=>"zona urbana", "pop"=>8577, "name_en"=>"Sabana Grande", "lang"=>"en", "lat"=>18.084023, "lon"=>-66.967050),
"/US/US-PR/sabana-hoyos-23742" => array("type"=>"comunidad", "pop"=>1543, "name_en"=>"Sabana Hoyos", "lang"=>"en", "lat"=>18.428979, "lon"=>-66.612076),
"/US/US-PR/sabana-seca-23743" => array("type"=>"comunidad", "pop"=>6755, "name_en"=>"Sabana Seca", "lang"=>"en", "lat"=>18.425638, "lon"=>-66.185361),
"/US/US-PR/salinas-23744" => array("type"=>"zona urbana", "pop"=>6786, "name_ru"=>"Салинас", "name_en"=>"Salinas", "lang"=>"en", "lat"=>17.979402, "lon"=>-66.296646),
"/US/US-PR/san-antonio-23745" => array("type"=>"comunidad", "pop"=>2458, "name_ru"=>"Сан Антонио", "name_en"=>"San Antonio", "lang"=>"en", "lat"=>18.497310, "lon"=>-67.097023),
"/US/US-PR/san-antonio-23746" => array("type"=>"comunidad", "pop"=>5875, "name_ru"=>"Сан Антонио", "name_en"=>"San Antonio", "lang"=>"en", "lat"=>18.449517, "lon"=>-66.298609),
"/US/US-PR/san-antonio-23747" => array("type"=>"comunidad", "pop"=>1006, "name_ru"=>"Сан Антонио", "name_en"=>"San Antonio", "lang"=>"en", "lat"=>18.453583, "lon"=>-66.950134),
"/US/US-PR/san-german-23748" => array("type"=>"zona urbana", "pop"=>11977, "name_en"=>"San German", "lang"=>"en", "lat"=>18.084687, "lon"=>-67.045500),
"/US/US-PR/san-isidro-23749" => array("type"=>"comunidad", "pop"=>7378, "name_en"=>"San Isidro", "lang"=>"en", "lat"=>18.394137, "lon"=>-65.885699),
"/US/US-PR/san-jose-23750" => array("type"=>"comunidad", "pop"=>3076, "name_ru"=>"Сан-Хосе", "name_en"=>"San Jose", "lang"=>"en", "lat"=>18.405960, "lon"=>-66.248479),
"/US/US-PR/san-juan-23751" => array("type"=>"zona urbana", "pop"=>426832, "name_en"=>"San Juan", "lang"=>"en", "lat"=>18.408386, "lon"=>-66.064425),
"/US/US-PR/san-lorenzo-23752" => array("type"=>"zona urbana", "pop"=>8481, "name_ru"=>"Сан Лоренсо", "name_en"=>"San Lorenzo", "lang"=>"en", "lat"=>18.190836, "lon"=>-65.966394),
"/US/US-PR/san-sebastian-23753" => array("type"=>"zona urbana", "pop"=>11605, "name_en"=>"San Sebastian", "lang"=>"en", "lat"=>18.337985, "lon"=>-66.995937),
"/US/US-PR/santa-barbara-23754" => array("type"=>"comunidad", "pop"=>5252, "name_ru"=>"Санта Брабара", "name_en"=>"Santa Barbara", "lang"=>"en", "lat"=>18.395604, "lon"=>-65.919048),
"/US/US-PR/santa-isabel-23755" => array("type"=>"zona urbana", "pop"=>7512, "name_en"=>"Santa Isabel", "lang"=>"en", "lat"=>17.970853, "lon"=>-66.405138),
"/US/US-PR/santo-domingo-23756" => array("type"=>"comunidad", "pop"=>2691, "name_en"=>"Santo Domingo", "lang"=>"en", "lat"=>18.071710, "lon"=>-66.753726),
"/US/US-PR/stella-23757" => array("type"=>"comunidad", "pop"=>1286, "name_en"=>"Stella", "lang"=>"en", "lat"=>18.323955, "lon"=>-67.247226),
"/US/US-PR/suarez-23758" => array("type"=>"comunidad", "pop"=>2438, "name_en"=>"Suarez", "lang"=>"en", "lat"=>18.432911, "lon"=>-65.851033),
"/US/US-PR/sumidero-23759" => array("type"=>"comunidad", "pop"=>1145, "name_en"=>"Sumidero", "lang"=>"en", "lat"=>18.212758, "lon"=>-66.134268),
"/US/US-PR/tallaboa-23760" => array("type"=>"comunidad", "pop"=>1021, "name_en"=>"Tallaboa", "lang"=>"en", "lat"=>17.998655, "lon"=>-66.710855),
"/US/US-PR/tallaboa-alta-23761" => array("type"=>"comunidad", "pop"=>2399, "name_en"=>"Tallaboa Alta", "lang"=>"en", "lat"=>18.052081, "lon"=>-66.702881),
"/US/US-PR/tiburones-23762" => array("type"=>"comunidad", "pop"=>1119, "name_en"=>"Tiburones", "lang"=>"en", "lat"=>18.438721, "lon"=>-66.579994),
"/US/US-PR/tierras-nuevas-poniente-23763" => array("type"=>"comunidad", "pop"=>1688, "name_en"=>"Tierras Nuevas Poniente", "lang"=>"en", "lat"=>18.463861, "lon"=>-66.489007),
"/US/US-PR/toa-alta-23764" => array("type"=>"zona urbana", "pop"=>4941, "name_en"=>"Toa Alta", "lang"=>"en", "lat"=>18.390063, "lon"=>-66.250136),
"/US/US-PR/toa-baja-23765" => array("type"=>"zona urbana", "pop"=>1960, "name_en"=>"Toa Baja", "lang"=>"en", "lat"=>18.444343, "lon"=>-66.254811),
"/US/US-PR/trujillo-alto-23766" => array("type"=>"zona urbana", "pop"=>44336, "name_en"=>"Trujillo Alto", "lang"=>"en", "lat"=>18.362973, "lon"=>-66.017614),
"/US/US-PR/utuado-23767" => array("type"=>"zona urbana", "pop"=>9943, "name_en"=>"Utuado", "lang"=>"en", "lat"=>18.271407, "lon"=>-66.705606),
"/US/US-PR/vazquez-23768" => array("type"=>"comunidad", "pop"=>2182, "name_en"=>"Vazquez", "lang"=>"en", "lat"=>18.067686, "lon"=>-66.239099),
"/US/US-PR/vega-alta-23769" => array("type"=>"zona urbana", "pop"=>11187, "name_en"=>"Vega Alta", "lang"=>"en", "lat"=>18.416617, "lon"=>-66.321965),
"/US/US-PR/vega-baja-23770" => array("type"=>"zona urbana", "pop"=>27753, "name_en"=>"Vega Baja", "lang"=>"en", "lat"=>18.442827, "lon"=>-66.399706),
"/US/US-PR/vieques-23771" => array("type"=>"comunidad", "pop"=>3971, "name_en"=>"Vieques", "lang"=>"en", "lat"=>18.428396, "lon"=>-65.835388),
"/US/US-PR/vieques-23772" => array("type"=>"zona urbana", "pop"=>2359, "name_en"=>"Vieques", "lang"=>"en", "lat"=>18.149552, "lon"=>-65.445223),
"/US/US-PR/villalba-23773" => array("type"=>"zona urbana", "pop"=>4159, "name_en"=>"Villalba", "lang"=>"en", "lat"=>18.129985, "lon"=>-66.482150),
"/US/US-PR/yabucoa-23774" => array("type"=>"zona urbana", "pop"=>7262, "name_en"=>"Yabucoa", "lang"=>"en", "lat"=>18.049602, "lon"=>-65.879650),
"/US/US-PR/yauco-23775" => array("type"=>"zona urbana", "pop"=>18158, "name_en"=>"Yauco", "lang"=>"en", "lat"=>18.036563, "lon"=>-66.861964),
"/US/US-PR/yaurel-23776" => array("type"=>"comunidad", "pop"=>1382, "name_en"=>"Yaurel", "lang"=>"en", "lat"=>18.028876, "lon"=>-66.057506),
"/US/US-RI" => array("type"=>"state", "pop"=>1050788, "name_ru"=>"Род-айленд", "name_en"=>"Rhode Island", "lang"=>"en"),
"/US/US-RI/ashaway-18985" => array("type"=>"cdp", "pop"=>1584, "name_en"=>"Ashaway", "lang"=>"en", "lat"=>41.426462, "lon"=>-71.787698),
"/US/US-RI/barrington-18986" => array("type"=>"cdp", "pop"=>16368, "name_ru"=>"Бэррингтон", "name_en"=>"Barrington", "lang"=>"en", "lat"=>41.735167, "lon"=>-71.318633),
"/US/US-RI/bradford-18987" => array("type"=>"cdp", "pop"=>1604, "name_ru"=>"Брадфорд", "name_en"=>"Bradford", "lang"=>"en", "lat"=>41.393722, "lon"=>-71.750341),
"/US/US-RI/bristol-18988" => array("type"=>"cdp", "pop"=>49838, "name_ru"=>"Бристол", "name_en"=>"Bristol", "lang"=>"en", "lat"=>41.677750, "lon"=>-71.274320),
"/US/US-RI/central-falls-18989" => array("type"=>"city", "pop"=>18683, "name_en"=>"Central Falls", "lang"=>"en", "lat"=>41.889371, "lon"=>-71.393870),
"/US/US-RI/cranston-18990" => array("type"=>"city", "pop"=>79980, "name_en"=>"Cranston", "lang"=>"en", "lat"=>41.766650, "lon"=>-71.481108),
"/US/US-RI/cumberland-hill-18991" => array("type"=>"cdp", "pop"=>6379, "name_en"=>"Cumberland Hill", "lang"=>"en", "lat"=>41.972049, "lon"=>-71.460079),
"/US/US-RI/east-providence-18992" => array("type"=>"city", "pop"=>48480, "name_en"=>"East Providence", "lang"=>"en", "lat"=>41.800200, "lon"=>-71.359488),
"/US/US-RI/greenville-18993" => array("type"=>"cdp", "pop"=>8303, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>41.879000, "lon"=>-71.556662),
"/US/US-RI/harrisville-18994" => array("type"=>"cdp", "pop"=>1654, "name_ru"=>"Харрисвилл", "name_en"=>"Harrisville", "lang"=>"en", "lat"=>41.968154, "lon"=>-71.677368),
"/US/US-RI/hope-valley-18995" => array("type"=>"cdp", "pop"=>1446, "name_en"=>"Hope Valley", "lang"=>"en", "lat"=>41.514801, "lon"=>-71.722316),
"/US/US-RI/kingston-18996" => array("type"=>"cdp", "pop"=>6504, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>41.472225, "lon"=>-71.523439),
"/US/US-RI/melville-18997" => array("type"=>"cdp", "pop"=>4426, "name_ru"=>"Меовилл", "name_en"=>"Melville", "lang"=>"en", "lat"=>41.563882, "lon"=>-71.303805),
"/US/US-RI/narragansett-pier-18998" => array("type"=>"cdp", "pop"=>3721, "name_en"=>"Narragansett Pier", "lang"=>"en", "lat"=>41.427100, "lon"=>-71.466910),
"/US/US-RI/newport-18999" => array("type"=>"city", "pop"=>23523, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>41.483146, "lon"=>-71.320265),
"/US/US-RI/newport-east-19000" => array("type"=>"cdp", "pop"=>11080, "name_en"=>"Newport East", "lang"=>"en", "lat"=>41.516018, "lon"=>-71.288338),
"/US/US-RI/north-providence-19001" => array("type"=>"cdp", "pop"=>32695, "name_en"=>"North Providence", "lang"=>"en", "lat"=>41.861200, "lon"=>-71.457489),
"/US/US-RI/pascoag-19002" => array("type"=>"cdp", "pop"=>5011, "name_en"=>"Pascoag", "lang"=>"en", "lat"=>41.953657, "lon"=>-71.704868),
"/US/US-RI/pawtucket-19003" => array("type"=>"city", "pop"=>71765, "name_en"=>"Pawtucket", "lang"=>"en", "lat"=>41.873300, "lon"=>-71.374322),
"/US/US-RI/providence-19004" => array("type"=>"city", "pop"=>171557, "name_ru"=>"Провиденс", "name_en"=>"Providence", "lang"=>"en", "lat"=>41.821950, "lon"=>-71.419732),
"/US/US-RI/tiverton-19005" => array("type"=>"cdp", "pop"=>14951, "name_en"=>"Tiverton", "lang"=>"en", "lat"=>41.654190, "lon"=>-71.200913),
"/US/US-RI/valley-falls-19006" => array("type"=>"cdp", "pop"=>11175, "name_ru"=>"Валли Фоллс", "name_en"=>"Valley Falls", "lang"=>"en", "lat"=>41.923850, "lon"=>-71.393453),
"/US/US-RI/wakefield-peacedale-19007" => array("type"=>"cdp", "pop"=>7134, "name_en"=>"Wakefield-Peacedale", "lang"=>"en", "lat"=>41.445767, "lon"=>-71.501078),
"/US/US-RI/warwick-19008" => array("type"=>"city", "pop"=>84483, "name_ru"=>"Уорвик", "name_en"=>"Warwick", "lang"=>"en", "lat"=>41.702200, "lon"=>-71.421854),
"/US/US-RI/west-warwick-19010" => array("type"=>"cdp", "pop"=>29255, "name_en"=>"West Warwick", "lang"=>"en", "lat"=>41.697550, "lon"=>-71.518328),
"/US/US-RI/westerly-19009" => array("type"=>"cdp", "pop"=>23377, "name_en"=>"Westerly", "lang"=>"en", "lat"=>41.370551, "lon"=>-71.810337),
"/US/US-RI/woonsocket-19011" => array("type"=>"city", "pop"=>43268, "name_en"=>"Woonsocket", "lang"=>"en", "lat"=>41.999883, "lon"=>-71.500378),
"/US/US-SC" => array("type"=>"state", "pop"=>4479800, "name_ru"=>"Южная Каролина", "name_en"=>"South Carolina", "lang"=>"en"),
"/US/US-SC/abbeville-19012" => array("type"=>"city", "pop"=>5547, "name_ru"=>"Аббевилл", "name_en"=>"Abbeville", "lang"=>"en", "lat"=>34.179567, "lon"=>-82.378232),
"/US/US-SC/aiken-19013" => array("type"=>"city", "pop"=>29434, "name_en"=>"Aiken", "lang"=>"en", "lat"=>33.540832, "lon"=>-81.728365),
"/US/US-SC/allendale-19014" => array("type"=>"town", "pop"=>3659, "name_ru"=>"Аллендейл", "name_en"=>"Allendale", "lang"=>"en", "lat"=>33.008088, "lon"=>-81.309739),
"/US/US-SC/anderson-19015" => array("type"=>"city", "pop"=>27027, "name_ru"=>"Андерсон", "name_en"=>"Anderson", "lang"=>"en", "lat"=>34.518567, "lon"=>-82.645779),
"/US/US-SC/andrews-19016" => array("type"=>"town", "pop"=>2978, "name_ru"=>"Эндрюс", "name_en"=>"Andrews", "lang"=>"en", "lat"=>33.448591, "lon"=>-79.566198),
"/US/US-SC/arcadia-lakes-19017" => array("type"=>"town", "pop"=>836, "name_en"=>"Arcadia Lakes", "lang"=>"en", "lat"=>34.053478, "lon"=>-80.962136),
"/US/US-SC/arial-19018" => array("type"=>"cdp", "pop"=>2604, "name_en"=>"Arial", "lang"=>"en", "lat"=>34.846295, "lon"=>-82.639506),
"/US/US-SC/atlantic-beach-19019" => array("type"=>"town", "pop"=>397, "name_ru"=>"Атлантик Бич", "name_en"=>"Atlantic Beach", "lang"=>"en", "lat"=>33.803019, "lon"=>-78.717496),
"/US/US-SC/aynor-19020" => array("type"=>"town", "pop"=>640, "name_en"=>"Aynor", "lang"=>"en", "lat"=>33.998409, "lon"=>-79.201083),
"/US/US-SC/bamberg-19021" => array("type"=>"town", "pop"=>3432, "name_en"=>"Bamberg", "lang"=>"en", "lat"=>33.298284, "lon"=>-81.032083),
"/US/US-SC/barnwell-19022" => array("type"=>"city", "pop"=>4783, "name_en"=>"Barnwell", "lang"=>"en", "lat"=>33.241358, "lon"=>-81.365102),
"/US/US-SC/batesburg-19023" => array("type"=>"town", "pop"=>4082, "name_en"=>"Batesburg", "lang"=>"en", "lat"=>33.907391, "lon"=>-81.546893),
"/US/US-SC/beaufort-19024" => array("type"=>"city", "pop"=>11755, "name_en"=>"Beaufort", "lang"=>"en", "lat"=>32.425910, "lon"=>-80.689056),
"/US/US-SC/belton-19025" => array("type"=>"city", "pop"=>4670, "name_ru"=>"Белтон", "name_en"=>"Belton", "lang"=>"en", "lat"=>34.523900, "lon"=>-82.495086),
"/US/US-SC/belvedere-19026" => array("type"=>"cdp", "pop"=>6133, "name_ru"=>"Бельведер", "name_en"=>"Belvedere", "lang"=>"en", "lat"=>33.535650, "lon"=>-81.941938),
"/US/US-SC/bennettsville-19027" => array("type"=>"city", "pop"=>9012, "name_en"=>"Bennettsville", "lang"=>"en", "lat"=>34.631569, "lon"=>-79.686083),
"/US/US-SC/berea-19028" => array("type"=>"cdp", "pop"=>13535, "name_ru"=>"Берея", "name_en"=>"Berea", "lang"=>"en", "lat"=>34.881536, "lon"=>-82.465060),
"/US/US-SC/bethune-19029" => array("type"=>"town", "pop"=>370, "name_ru"=>"Бетун", "name_en"=>"Bethune", "lang"=>"en", "lat"=>34.413947, "lon"=>-80.349072),
"/US/US-SC/bishopville-19030" => array("type"=>"town", "pop"=>3859, "name_en"=>"Bishopville", "lang"=>"en", "lat"=>34.219858, "lon"=>-80.247914),
"/US/US-SC/blacksburg-19031" => array("type"=>"town", "pop"=>1903, "name_en"=>"Blacksburg", "lang"=>"en", "lat"=>35.121143, "lon"=>-81.518727),
"/US/US-SC/blackville-19032" => array("type"=>"town", "pop"=>2837, "name_en"=>"Blackville", "lang"=>"en", "lat"=>33.360242, "lon"=>-81.280960),
"/US/US-SC/blenheim-19033" => array("type"=>"town", "pop"=>128, "name_en"=>"Blenheim", "lang"=>"en", "lat"=>34.510104, "lon"=>-79.652631),
"/US/US-SC/bluffton-19034" => array("type"=>"town", "pop"=>4312, "name_ru"=>"Блаффтон", "name_en"=>"Bluffton", "lang"=>"en", "lat"=>32.235777, "lon"=>-80.863054),
"/US/US-SC/blythewood-19035" => array("type"=>"town", "pop"=>1407, "name_en"=>"Blythewood", "lang"=>"en", "lat"=>34.217023, "lon"=>-80.975821),
"/US/US-SC/boiling-springs-19036" => array("type"=>"cdp", "pop"=>3522, "name_en"=>"Boiling Springs", "lang"=>"en", "lat"=>35.044981, "lon"=>-81.976728),
"/US/US-SC/bonneau-19037" => array("type"=>"town", "pop"=>352, "name_en"=>"Bonneau", "lang"=>"en", "lat"=>33.306552, "lon"=>-79.958036),
"/US/US-SC/bowman-19038" => array("type"=>"town", "pop"=>1149, "name_ru"=>"Баумэн", "name_en"=>"Bowman", "lang"=>"en", "lat"=>33.348897, "lon"=>-80.683798),
"/US/US-SC/branchville-19039" => array("type"=>"town", "pop"=>1035, "name_ru"=>"Бранчвилл", "name_en"=>"Branchville", "lang"=>"en", "lat"=>33.251101, "lon"=>-80.815774),
"/US/US-SC/briarcliffe-acres-19040" => array("type"=>"town", "pop"=>548, "name_en"=>"Briarcliffe Acres", "lang"=>"en", "lat"=>33.789095, "lon"=>-78.747559),
"/US/US-SC/brookdale-19041" => array("type"=>"cdp", "pop"=>5339, "name_en"=>"Brookdale", "lang"=>"en", "lat"=>33.515426, "lon"=>-80.835031),
"/US/US-SC/brunson-19042" => array("type"=>"town", "pop"=>571, "name_en"=>"Brunson", "lang"=>"en", "lat"=>32.924032, "lon"=>-81.188926),
"/US/US-SC/bucksport-19043" => array("type"=>"cdp", "pop"=>1022, "name_ru"=>"Бакспорт", "name_en"=>"Bucksport", "lang"=>"en", "lat"=>33.672037, "lon"=>-79.112744),
"/US/US-SC/buffalo-19044" => array("type"=>"cdp", "pop"=>1569, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>34.724263, "lon"=>-81.683481),
"/US/US-SC/burnettown-19045" => array("type"=>"town", "pop"=>2665, "name_en"=>"Burnettown", "lang"=>"en", "lat"=>33.510467, "lon"=>-81.854222),
"/US/US-SC/burton-19046" => array("type"=>"cdp", "pop"=>6917, "name_ru"=>"Бертон", "name_en"=>"Burton", "lang"=>"en", "lat"=>32.420430, "lon"=>-80.744105),
"/US/US-SC/calhoun-falls-19047" => array("type"=>"town", "pop"=>2183, "name_en"=>"Calhoun Falls", "lang"=>"en", "lat"=>34.091722, "lon"=>-82.596544),
"/US/US-SC/camden-19048" => array("type"=>"city", "pop"=>7033, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>34.261640, "lon"=>-80.611653),
"/US/US-SC/cameron-19049" => array("type"=>"town", "pop"=>407, "name_ru"=>"Кэмерон", "name_en"=>"Cameron", "lang"=>"en", "lat"=>33.558400, "lon"=>-80.714882),
"/US/US-SC/campobello-19050" => array("type"=>"town", "pop"=>586, "name_en"=>"Campobello", "lang"=>"en", "lat"=>35.116596, "lon"=>-82.149440),
"/US/US-SC/carlisle-19051" => array("type"=>"town", "pop"=>458, "name_ru"=>"Карлайл", "name_en"=>"Carlisle", "lang"=>"en", "lat"=>34.591535, "lon"=>-81.464697),
"/US/US-SC/cayce-19052" => array("type"=>"city", "pop"=>12646, "name_en"=>"Cayce", "lang"=>"en", "lat"=>33.955800, "lon"=>-81.060901),
"/US/US-SC/centerville-19053" => array("type"=>"cdp", "pop"=>4866, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>34.524633, "lon"=>-82.718903),
"/US/US-SC/central-19054" => array("type"=>"town", "pop"=>4104, "name_ru"=>"Сентрал", "name_en"=>"Central", "lang"=>"en", "lat"=>34.723349, "lon"=>-82.782966),
"/US/US-SC/central-pacolet-19055" => array("type"=>"town", "pop"=>280, "name_en"=>"Central Pacolet", "lang"=>"en", "lat"=>34.909400, "lon"=>-81.751722),
"/US/US-SC/chapin-19056" => array("type"=>"town", "pop"=>709, "name_ru"=>"Чапин", "name_en"=>"Chapin", "lang"=>"en", "lat"=>34.165201, "lon"=>-81.347988),
"/US/US-SC/chappells-19057" => array("type"=>"town", "pop"=>45, "name_en"=>"Chappells", "lang"=>"en", "lat"=>34.182466, "lon"=>-81.869989),
"/US/US-SC/charleston-19058" => array("type"=>"city", "pop"=>111978, "name_ru"=>"Чарльстон", "name_en"=>"Charleston", "lang"=>"en", "lat"=>32.785250, "lon"=>-79.994024),
"/US/US-SC/cheraw-19059" => array("type"=>"town", "pop"=>5427, "name_ru"=>"Черо", "name_en"=>"Cheraw", "lang"=>"en", "lat"=>34.696417, "lon"=>-79.895301),
"/US/US-SC/cherryvale-19060" => array("type"=>"cdp", "pop"=>3061, "name_ru"=>"Черривейл", "name_en"=>"Cherryvale", "lang"=>"en", "lat"=>33.955257, "lon"=>-80.459496),
"/US/US-SC/chesnee-19061" => array("type"=>"city", "pop"=>1078, "name_en"=>"Chesnee", "lang"=>"en", "lat"=>35.147872, "lon"=>-81.862838),
"/US/US-SC/chester-19062" => array("type"=>"city", "pop"=>6045, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>34.705655, "lon"=>-81.214156),
"/US/US-SC/chesterfield-19063" => array("type"=>"town", "pop"=>1318, "name_ru"=>"Честерфилд", "name_en"=>"Chesterfield", "lang"=>"en", "lat"=>34.734731, "lon"=>-80.084652),
"/US/US-SC/city-view-19064" => array("type"=>"town", "pop"=>1490, "name_en"=>"City View", "lang"=>"en", "lat"=>34.861983, "lon"=>-82.426843),
"/US/US-SC/clearwater-19065" => array("type"=>"cdp", "pop"=>4731, "name_ru"=>"Клирвотер", "name_en"=>"Clearwater", "lang"=>"en", "lat"=>33.505450, "lon"=>-81.907741),
"/US/US-SC/clemson-19066" => array("type"=>"city", "pop"=>13012, "name_en"=>"Clemson", "lang"=>"en", "lat"=>34.684019, "lon"=>-82.812853),
"/US/US-SC/clinton-19067" => array("type"=>"city", "pop"=>8892, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>34.477768, "lon"=>-81.866059),
"/US/US-SC/clio-19068" => array("type"=>"town", "pop"=>721, "name_ru"=>"Клио", "name_en"=>"Clio", "lang"=>"en", "lat"=>34.578535, "lon"=>-79.546425),
"/US/US-SC/clover-19069" => array("type"=>"town", "pop"=>4916, "name_en"=>"Clover", "lang"=>"en", "lat"=>35.112058, "lon"=>-81.227051),
"/US/US-SC/columbia-19070" => array("type"=>"city", "pop"=>127029, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>34.039236, "lon"=>-80.886341),
"/US/US-SC/conway-19071" => array("type"=>"city", "pop"=>15894, "name_ru"=>"Конвэй", "name_en"=>"Conway", "lang"=>"en", "lat"=>33.838576, "lon"=>-79.060910),
"/US/US-SC/cope-19072" => array("type"=>"town", "pop"=>102, "name_en"=>"Cope", "lang"=>"en", "lat"=>33.378198, "lon"=>-81.006840),
"/US/US-SC/cordova-19073" => array("type"=>"town", "pop"=>151, "name_ru"=>"Кордова", "name_en"=>"Cordova", "lang"=>"en", "lat"=>33.435605, "lon"=>-80.920036),
"/US/US-SC/cottageville-19074" => array("type"=>"town", "pop"=>677, "name_en"=>"Cottageville", "lang"=>"en", "lat"=>32.936730, "lon"=>-80.479751),
"/US/US-SC/coward-19075" => array("type"=>"town", "pop"=>676, "name_en"=>"Coward", "lang"=>"en", "lat"=>33.972149, "lon"=>-79.748296),
"/US/US-SC/cowpens-19076" => array("type"=>"town", "pop"=>2413, "name_en"=>"Cowpens", "lang"=>"en", "lat"=>35.018848, "lon"=>-81.803597),
"/US/US-SC/cross-hill-19077" => array("type"=>"town", "pop"=>583, "name_en"=>"Cross Hill", "lang"=>"en", "lat"=>34.303975, "lon"=>-81.983239),
"/US/US-SC/darlington-19078" => array("type"=>"city", "pop"=>6666, "name_ru"=>"Дарлингтон", "name_en"=>"Darlington", "lang"=>"en", "lat"=>34.301811, "lon"=>-79.866101),
"/US/US-SC/denmark-19079" => array("type"=>"city", "pop"=>3005, "name_en"=>"Denmark", "lang"=>"en", "lat"=>33.318739, "lon"=>-81.138515),
"/US/US-SC/dentsville-19080" => array("type"=>"cdp", "pop"=>11839, "name_en"=>"Dentsville", "lang"=>"en", "lat"=>34.076500, "lon"=>-80.956149),
"/US/US-SC/dillon-19081" => array("type"=>"city", "pop"=>6353, "name_ru"=>"Диллон", "name_en"=>"Dillon", "lang"=>"en", "lat"=>34.421989, "lon"=>-79.369812),
"/US/US-SC/donalds-19082" => array("type"=>"town", "pop"=>335, "name_en"=>"Donalds", "lang"=>"en", "lat"=>34.377601, "lon"=>-82.347024),
"/US/US-SC/due-west-19083" => array("type"=>"town", "pop"=>1268, "name_en"=>"Due West", "lang"=>"en", "lat"=>34.332250, "lon"=>-82.388607),
"/US/US-SC/duncan-19084" => array("type"=>"town", "pop"=>3118, "name_ru"=>"Дункан", "name_en"=>"Duncan", "lang"=>"en", "lat"=>34.935584, "lon"=>-82.137048),
"/US/US-SC/dunean-19085" => array("type"=>"cdp", "pop"=>4637, "name_en"=>"Dunean", "lang"=>"en", "lat"=>34.820050, "lon"=>-82.422823),
"/US/US-SC/easley-19086" => array("type"=>"city", "pop"=>20325, "name_en"=>"Easley", "lang"=>"en", "lat"=>34.821968, "lon"=>-82.588018),
"/US/US-SC/east-gaffney-19087" => array("type"=>"cdp", "pop"=>3278, "name_en"=>"East Gaffney", "lang"=>"en", "lat"=>35.084420, "lon"=>-81.619929),
"/US/US-SC/east-sumter-19089" => array("type"=>"cdp", "pop"=>1590, "name_en"=>"East Sumter", "lang"=>"en", "lat"=>33.925401, "lon"=>-80.296340),
"/US/US-SC/eastover-19088" => array("type"=>"town", "pop"=>779, "name_en"=>"Eastover", "lang"=>"en", "lat"=>33.878411, "lon"=>-80.695694),
"/US/US-SC/edgefield-19090" => array("type"=>"town", "pop"=>4418, "name_ru"=>"Эджфилд", "name_en"=>"Edgefield", "lang"=>"en", "lat"=>33.789412, "lon"=>-81.930385),
"/US/US-SC/edisto-19091" => array("type"=>"cdp", "pop"=>2815, "name_en"=>"Edisto", "lang"=>"en", "lat"=>33.476493, "lon"=>-80.898851),
"/US/US-SC/edisto-beach-19092" => array("type"=>"town", "pop"=>718, "name_en"=>"Edisto Beach", "lang"=>"en", "lat"=>32.489357, "lon"=>-80.319095),
"/US/US-SC/ehrhardt-19093" => array("type"=>"town", "pop"=>552, "name_en"=>"Ehrhardt", "lang"=>"en", "lat"=>33.098353, "lon"=>-81.013422),
"/US/US-SC/elgin-19094" => array("type"=>"town", "pop"=>1256, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>34.168375, "lon"=>-80.794888),
"/US/US-SC/elgin-19095" => array("type"=>"cdp", "pop"=>2196, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>34.673730, "lon"=>-80.721651),
"/US/US-SC/elko-19096" => array("type"=>"town", "pop"=>206, "name_ru"=>"Элко", "name_en"=>"Elko", "lang"=>"en", "lat"=>33.379073, "lon"=>-81.379513),
"/US/US-SC/elloree-19097" => array("type"=>"town", "pop"=>699, "name_en"=>"Elloree", "lang"=>"en", "lat"=>33.529789, "lon"=>-80.572372),
"/US/US-SC/estill-19098" => array("type"=>"town", "pop"=>2322, "name_en"=>"Estill", "lang"=>"en", "lat"=>32.754233, "lon"=>-81.240110),
"/US/US-SC/eureka-mill-19099" => array("type"=>"cdp", "pop"=>1738, "name_en"=>"Eureka Mill", "lang"=>"en", "lat"=>34.717626, "lon"=>-81.193841),
"/US/US-SC/eutawville-19100" => array("type"=>"town", "pop"=>328, "name_en"=>"Eutawville", "lang"=>"en", "lat"=>33.397855, "lon"=>-80.343648),
"/US/US-SC/fairfax-19101" => array("type"=>"town", "pop"=>3158, "name_ru"=>"Фэрфакс", "name_en"=>"Fairfax", "lang"=>"en", "lat"=>32.959100, "lon"=>-81.236348),
"/US/US-SC/florence-19102" => array("type"=>"city", "pop"=>31570, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>34.181000, "lon"=>-79.780399),
"/US/US-SC/folly-beach-19103" => array("type"=>"city", "pop"=>2412, "name_en"=>"Folly Beach", "lang"=>"en", "lat"=>32.674520, "lon"=>-79.927423),
"/US/US-SC/forest-acres-19104" => array("type"=>"city", "pop"=>9925, "name_en"=>"Forest Acres", "lang"=>"en", "lat"=>34.026300, "lon"=>-80.976763),
"/US/US-SC/forestbrook-19105" => array("type"=>"cdp", "pop"=>2502, "name_en"=>"Forestbrook", "lang"=>"en", "lat"=>33.723033, "lon"=>-78.966143),
"/US/US-SC/fort-lawn-19106" => array("type"=>"town", "pop"=>810, "name_en"=>"Fort Lawn", "lang"=>"en", "lat"=>34.701142, "lon"=>-80.898492),
"/US/US-SC/fort-mill-19107" => array("type"=>"town", "pop"=>10032, "name_en"=>"Fort Mill", "lang"=>"en", "lat"=>35.006138, "lon"=>-80.943455),
"/US/US-SC/fountain-inn-19108" => array("type"=>"town", "pop"=>7760, "name_en"=>"Fountain Inn", "lang"=>"en", "lat"=>34.693520, "lon"=>-82.200608),
"/US/US-SC/furman-19109" => array("type"=>"town", "pop"=>278, "name_en"=>"Furman", "lang"=>"en", "lat"=>32.681410, "lon"=>-81.188022),
"/US/US-SC/gaffney-19110" => array("type"=>"city", "pop"=>13200, "name_en"=>"Gaffney", "lang"=>"en", "lat"=>35.072316, "lon"=>-81.656123),
"/US/US-SC/gantt-19111" => array("type"=>"cdp", "pop"=>13891, "name_ru"=>"Гантт", "name_en"=>"Gantt", "lang"=>"en", "lat"=>34.785760, "lon"=>-82.403463),
"/US/US-SC/garden-city-19112" => array("type"=>"cdp", "pop"=>6305, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>33.592650, "lon"=>-79.008971),
"/US/US-SC/gaston-19113" => array("type"=>"town", "pop"=>1442, "name_ru"=>"Гэстон", "name_en"=>"Gaston", "lang"=>"en", "lat"=>33.815774, "lon"=>-81.101120),
"/US/US-SC/gayle-mill-19114" => array("type"=>"cdp", "pop"=>1037, "name_en"=>"Gayle Mill", "lang"=>"en", "lat"=>34.701813, "lon"=>-81.240290),
"/US/US-SC/georgetown-19115" => array("type"=>"city", "pop"=>8501, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>33.362995, "lon"=>-79.296135),
"/US/US-SC/gifford-19116" => array("type"=>"town", "pop"=>355, "name_ru"=>"Гиффорд", "name_en"=>"Gifford", "lang"=>"en", "lat"=>32.860668, "lon"=>-81.239132),
"/US/US-SC/gilbert-19117" => array("type"=>"town", "pop"=>594, "name_ru"=>"Гилберт", "name_en"=>"Gilbert", "lang"=>"en", "lat"=>33.923558, "lon"=>-81.394238),
"/US/US-SC/gloverville-19118" => array("type"=>"cdp", "pop"=>2753, "name_en"=>"Gloverville", "lang"=>"en", "lat"=>33.528080, "lon"=>-81.813951),
"/US/US-SC/golden-grove-19119" => array("type"=>"cdp", "pop"=>2055, "name_en"=>"Golden Grove", "lang"=>"en", "lat"=>34.730688, "lon"=>-82.439123),
"/US/US-SC/goose-creek-19120" => array("type"=>"city", "pop"=>37900, "name_ru"=>"Гуз Крик", "name_en"=>"Goose Creek", "lang"=>"en", "lat"=>32.987477, "lon"=>-79.998680),
"/US/US-SC/govan-19121" => array("type"=>"town", "pop"=>62, "name_en"=>"Govan", "lang"=>"en", "lat"=>33.223234, "lon"=>-81.175129),
"/US/US-SC/gray-court-19122" => array("type"=>"town", "pop"=>987, "name_en"=>"Gray Court", "lang"=>"en", "lat"=>34.607331, "lon"=>-82.112938),
"/US/US-SC/great-falls-19123" => array("type"=>"town", "pop"=>2043, "name_en"=>"Great Falls", "lang"=>"en", "lat"=>34.574582, "lon"=>-80.903241),
"/US/US-SC/greeleyville-19124" => array("type"=>"town", "pop"=>397, "name_en"=>"Greeleyville", "lang"=>"en", "lat"=>33.577946, "lon"=>-79.989453),
"/US/US-SC/greenville-19125" => array("type"=>"city", "pop"=>59988, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>34.838650, "lon"=>-82.371555),
"/US/US-SC/greenwood-19126" => array("type"=>"city", "pop"=>22533, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>34.190161, "lon"=>-82.152820),
"/US/US-SC/greer-19127" => array("type"=>"city", "pop"=>24557, "name_en"=>"Greer", "lang"=>"en", "lat"=>34.939040, "lon"=>-82.231847),
"/US/US-SC/hampton-19128" => array("type"=>"town", "pop"=>2754, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>32.866972, "lon"=>-81.108903),
"/US/US-SC/hanahan-19129" => array("type"=>"city", "pop"=>16065, "name_en"=>"Hanahan", "lang"=>"en", "lat"=>32.930618, "lon"=>-80.008716),
"/US/US-SC/hardeeville-19130" => array("type"=>"town", "pop"=>2763, "name_en"=>"Hardeeville", "lang"=>"en", "lat"=>32.275747, "lon"=>-81.076615),
"/US/US-SC/harleyville-19131" => array("type"=>"town", "pop"=>689, "name_en"=>"Harleyville", "lang"=>"en", "lat"=>33.212305, "lon"=>-80.447888),
"/US/US-SC/hartsville-19132" => array("type"=>"city", "pop"=>7465, "name_ru"=>"Хартсвилл", "name_en"=>"Hartsville", "lang"=>"en", "lat"=>34.369515, "lon"=>-80.083805),
"/US/US-SC/heath-springs-19133" => array("type"=>"town", "pop"=>1044, "name_en"=>"Heath Springs", "lang"=>"en", "lat"=>34.590164, "lon"=>-80.674505),
"/US/US-SC/hemingway-19134" => array("type"=>"town", "pop"=>500, "name_en"=>"Hemingway", "lang"=>"en", "lat"=>33.753985, "lon"=>-79.444030),
"/US/US-SC/hickory-grove-19135" => array("type"=>"town", "pop"=>422, "name_en"=>"Hickory Grove", "lang"=>"en", "lat"=>34.981995, "lon"=>-81.415450),
"/US/US-SC/hilda-19136" => array("type"=>"town", "pop"=>420, "name_en"=>"Hilda", "lang"=>"en", "lat"=>33.273822, "lon"=>-81.246566),
"/US/US-SC/hilton-head-island-19137" => array("type"=>"town", "pop"=>33913, "name_en"=>"Hilton Head Island", "lang"=>"en", "lat"=>32.195672, "lon"=>-80.745626),
"/US/US-SC/hodges-19138" => array("type"=>"town", "pop"=>167, "name_ru"=>"Ходжес", "name_en"=>"Hodges", "lang"=>"en", "lat"=>34.287638, "lon"=>-82.247970),
"/US/US-SC/holly-hill-19139" => array("type"=>"town", "pop"=>1344, "name_ru"=>"Холли Хилл", "name_en"=>"Holly Hill", "lang"=>"en", "lat"=>33.324965, "lon"=>-80.413737),
"/US/US-SC/hollywood-19140" => array("type"=>"town", "pop"=>4398, "name_ru"=>"Холливуд", "name_en"=>"Hollywood", "lang"=>"en", "lat"=>32.751642, "lon"=>-80.201212),
"/US/US-SC/homeland-park-19141" => array("type"=>"cdp", "pop"=>6569, "name_en"=>"Homeland Park", "lang"=>"en", "lat"=>34.463850, "lon"=>-82.659054),
"/US/US-SC/honea-path-19142" => array("type"=>"town", "pop"=>3674, "name_en"=>"Honea Path", "lang"=>"en", "lat"=>34.447025, "lon"=>-82.394588),
"/US/US-SC/india-hook-19143" => array("type"=>"cdp", "pop"=>1506, "name_en"=>"India Hook", "lang"=>"en", "lat"=>35.011820, "lon"=>-81.039670),
"/US/US-SC/inman-19144" => array("type"=>"city", "pop"=>2096, "name_ru"=>"Инман", "name_en"=>"Inman", "lang"=>"en", "lat"=>35.051976, "lon"=>-82.093299),
"/US/US-SC/inman-mills-19145" => array("type"=>"cdp", "pop"=>1571, "name_en"=>"Inman Mills", "lang"=>"en", "lat"=>35.041093, "lon"=>-82.099563),
"/US/US-SC/irmo-19146" => array("type"=>"town", "pop"=>11684, "name_en"=>"Irmo", "lang"=>"en", "lat"=>34.093631, "lon"=>-81.187958),
"/US/US-SC/irwin-19147" => array("type"=>"cdp", "pop"=>1296, "name_ru"=>"Ирвин", "name_en"=>"Irwin", "lang"=>"en", "lat"=>34.693712, "lon"=>-80.822611),
"/US/US-SC/isle-of-palms-19148" => array("type"=>"city", "pop"=>4678, "name_en"=>"Isle of Palms", "lang"=>"en", "lat"=>32.801352, "lon"=>-79.757297),
"/US/US-SC/iva-19149" => array("type"=>"town", "pop"=>1200, "name_en"=>"Iva", "lang"=>"en", "lat"=>34.306753, "lon"=>-82.663477),
"/US/US-SC/jackson-19150" => array("type"=>"town", "pop"=>1647, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>33.328448, "lon"=>-81.792198),
"/US/US-SC/jamestown-19151" => array("type"=>"town", "pop"=>102, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>33.285943, "lon"=>-79.696480),
"/US/US-SC/jefferson-19152" => array("type"=>"town", "pop"=>693, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>34.651350, "lon"=>-80.388099),
"/US/US-SC/joanna-19153" => array("type"=>"cdp", "pop"=>1735, "name_en"=>"Joanna", "lang"=>"en", "lat"=>34.415687, "lon"=>-81.810776),
"/US/US-SC/johnsonville-19154" => array("type"=>"city", "pop"=>1491, "name_ru"=>"Джонсонвилл", "name_en"=>"Johnsonville", "lang"=>"en", "lat"=>33.816795, "lon"=>-79.449002),
"/US/US-SC/johnston-19155" => array("type"=>"town", "pop"=>2336, "name_ru"=>"Джонстон", "name_en"=>"Johnston", "lang"=>"en", "lat"=>33.834382, "lon"=>-81.805148),
"/US/US-SC/jonesville-19156" => array("type"=>"town", "pop"=>901, "name_ru"=>"Джонсвилл", "name_en"=>"Jonesville", "lang"=>"en", "lat"=>34.835558, "lon"=>-81.679869),
"/US/US-SC/judson-19157" => array("type"=>"cdp", "pop"=>2859, "name_ru"=>"Джадсон", "name_en"=>"Judson", "lang"=>"en", "lat"=>34.832950, "lon"=>-82.427778),
"/US/US-SC/kershaw-19158" => array("type"=>"town", "pop"=>1947, "name_en"=>"Kershaw", "lang"=>"en", "lat"=>34.545070, "lon"=>-80.587681),
"/US/US-SC/kiawah-island-19159" => array("type"=>"town", "pop"=>1118, "name_en"=>"Kiawah Island", "lang"=>"en", "lat"=>32.616025, "lon"=>-80.062006),
"/US/US-SC/kingstree-19160" => array("type"=>"town", "pop"=>3248, "name_en"=>"Kingstree", "lang"=>"en", "lat"=>33.665132, "lon"=>-79.829230),
"/US/US-SC/kline-19161" => array("type"=>"town", "pop"=>231, "name_en"=>"Kline", "lang"=>"en", "lat"=>33.124208, "lon"=>-81.343227),
"/US/US-SC/ladson-19162" => array("type"=>"cdp", "pop"=>13540, "name_en"=>"Ladson", "lang"=>"en", "lat"=>33.009392, "lon"=>-80.105741),
"/US/US-SC/lake-city-19163" => array("type"=>"city", "pop"=>6666, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>33.867750, "lon"=>-79.753217),
"/US/US-SC/lake-view-19164" => array("type"=>"town", "pop"=>782, "name_ru"=>"Лейк Вью", "name_en"=>"Lake View", "lang"=>"en", "lat"=>34.340224, "lon"=>-79.166883),
"/US/US-SC/lake-wylie-19165" => array("type"=>"cdp", "pop"=>2599, "name_en"=>"Lake Wylie", "lang"=>"en", "lat"=>35.110768, "lon"=>-81.053911),
"/US/US-SC/lamar-19166" => array("type"=>"town", "pop"=>991, "name_ru"=>"Ламар", "name_en"=>"Lamar", "lang"=>"en", "lat"=>34.169203, "lon"=>-80.065690),
"/US/US-SC/lancaster-19167" => array("type"=>"city", "pop"=>9981, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>34.723875, "lon"=>-80.779580),
"/US/US-SC/lancaster-mill-19168" => array("type"=>"cdp", "pop"=>2373, "name_en"=>"Lancaster Mill", "lang"=>"en", "lat"=>34.709148, "lon"=>-80.794941),
"/US/US-SC/landrum-19169" => array("type"=>"city", "pop"=>2615, "name_en"=>"Landrum", "lang"=>"en", "lat"=>35.174180, "lon"=>-82.186951),
"/US/US-SC/lane-19170" => array("type"=>"town", "pop"=>516, "name_ru"=>"Лейн", "name_en"=>"Lane", "lang"=>"en", "lat"=>33.524813, "lon"=>-79.880474),
"/US/US-SC/latta-19171" => array("type"=>"town", "pop"=>1478, "name_en"=>"Latta", "lang"=>"en", "lat"=>34.338914, "lon"=>-79.434001),
"/US/US-SC/laurel-bay-19172" => array("type"=>"cdp", "pop"=>4972, "name_en"=>"Laurel Bay", "lang"=>"en", "lat"=>32.460425, "lon"=>-80.788846),
"/US/US-SC/laurens-19173" => array("type"=>"city", "pop"=>9599, "name_ru"=>"Лоренс", "name_en"=>"Laurens", "lang"=>"en", "lat"=>34.501551, "lon"=>-82.025129),
"/US/US-SC/leesville-19174" => array("type"=>"town", "pop"=>2025, "name_ru"=>"Лисвилл", "name_en"=>"Leesville", "lang"=>"en", "lat"=>33.918109, "lon"=>-81.512292),
"/US/US-SC/lexington-19175" => array("type"=>"town", "pop"=>15901, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>33.986334, "lon"=>-81.229005),
"/US/US-SC/liberty-19176" => array("type"=>"town", "pop"=>3064, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>34.792472, "lon"=>-82.695364),
"/US/US-SC/lincolnville-19177" => array("type"=>"town", "pop"=>842, "name_ru"=>"Линколнвилл", "name_en"=>"Lincolnville", "lang"=>"en", "lat"=>33.007728, "lon"=>-80.156545),
"/US/US-SC/little-mountain-19178" => array("type"=>"town", "pop"=>260, "name_en"=>"Little Mountain", "lang"=>"en", "lat"=>34.195017, "lon"=>-81.414123),
"/US/US-SC/little-river-19179" => array("type"=>"cdp", "pop"=>3470, "name_ru"=>"Литтл Ривер", "name_en"=>"Little River", "lang"=>"en", "lat"=>33.873021, "lon"=>-78.631524),
"/US/US-SC/livingston-19180" => array("type"=>"town", "pop"=>142, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>33.553423, "lon"=>-81.120148),
"/US/US-SC/lockhart-19181" => array("type"=>"town", "pop"=>490, "name_ru"=>"Локхарт", "name_en"=>"Lockhart", "lang"=>"en", "lat"=>34.794693, "lon"=>-81.461037),
"/US/US-SC/lodge-19182" => array("type"=>"town", "pop"=>112, "name_en"=>"Lodge", "lang"=>"en", "lat"=>33.067520, "lon"=>-80.954563),
"/US/US-SC/loris-19183" => array("type"=>"city", "pop"=>2351, "name_en"=>"Loris", "lang"=>"en", "lat"=>34.056563, "lon"=>-78.890339),
"/US/US-SC/lowndesville-19184" => array("type"=>"town", "pop"=>159, "name_en"=>"Lowndesville", "lang"=>"en", "lat"=>34.210275, "lon"=>-82.649483),
"/US/US-SC/lowrys-19185" => array("type"=>"town", "pop"=>198, "name_en"=>"Lowrys", "lang"=>"en", "lat"=>34.804901, "lon"=>-81.238203),
"/US/US-SC/lugoff-19186" => array("type"=>"cdp", "pop"=>3211, "name_en"=>"Lugoff", "lang"=>"en", "lat"=>34.225737, "lon"=>-80.680605),
"/US/US-SC/luray-19187" => array("type"=>"town", "pop"=>112, "name_ru"=>"Лурей", "name_en"=>"Luray", "lang"=>"en", "lat"=>32.813391, "lon"=>-81.240170),
"/US/US-SC/lyman-19188" => array("type"=>"town", "pop"=>2870, "name_ru"=>"Лиман", "name_en"=>"Lyman", "lang"=>"en", "lat"=>34.951368, "lon"=>-82.123332),
"/US/US-SC/lynchburg-19189" => array("type"=>"town", "pop"=>553, "name_ru"=>"Линчбург", "name_en"=>"Lynchburg", "lang"=>"en", "lat"=>34.059169, "lon"=>-80.076994),
"/US/US-SC/manning-19195" => array("type"=>"city", "pop"=>3943, "name_ru"=>"Мэннинг", "name_en"=>"Manning", "lang"=>"en", "lat"=>33.693836, "lon"=>-80.215606),
"/US/US-SC/marion-19196" => array("type"=>"city", "pop"=>6801, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>34.175878, "lon"=>-79.399992),
"/US/US-SC/mauldin-19197" => array("type"=>"city", "pop"=>21784, "name_en"=>"Mauldin", "lang"=>"en", "lat"=>34.787115, "lon"=>-82.301465),
"/US/US-SC/mayesville-19198" => array("type"=>"town", "pop"=>1007, "name_en"=>"Mayesville", "lang"=>"en", "lat"=>33.987078, "lon"=>-80.207445),
"/US/US-SC/mayo-19199" => array("type"=>"cdp", "pop"=>1569, "name_ru"=>"Майо", "name_en"=>"Mayo", "lang"=>"en", "lat"=>35.084530, "lon"=>-81.854376),
"/US/US-SC/mcbee-19190" => array("type"=>"town", "pop"=>706, "name_en"=>"McBee", "lang"=>"en", "lat"=>34.467651, "lon"=>-80.258283),
"/US/US-SC/mcclellanville-19191" => array("type"=>"town", "pop"=>496, "name_en"=>"McClellanville", "lang"=>"en", "lat"=>33.085746, "lon"=>-79.467288),
"/US/US-SC/mccoll-19192" => array("type"=>"town", "pop"=>2308, "name_en"=>"McColl", "lang"=>"en", "lat"=>34.667648, "lon"=>-79.545343),
"/US/US-SC/mcconnells-19193" => array("type"=>"town", "pop"=>367, "name_en"=>"McConnells", "lang"=>"en", "lat"=>34.868729, "lon"=>-81.227577),
"/US/US-SC/mccormick-19194" => array("type"=>"town", "pop"=>2660, "name_en"=>"McCormick", "lang"=>"en", "lat"=>33.909960, "lon"=>-82.292941),
"/US/US-SC/meggett-19200" => array("type"=>"town", "pop"=>1363, "name_en"=>"Meggett", "lang"=>"en", "lat"=>32.705337, "lon"=>-80.263286),
"/US/US-SC/millwood-19201" => array("type"=>"cdp", "pop"=>1070, "name_en"=>"Millwood", "lang"=>"en", "lat"=>33.909712, "lon"=>-80.394234),
"/US/US-SC/monarch-mill-19202" => array("type"=>"cdp", "pop"=>2214, "name_en"=>"Monarch Mill", "lang"=>"en", "lat"=>34.718956, "lon"=>-81.581023),
"/US/US-SC/moncks-corner-19203" => array("type"=>"town", "pop"=>7044, "name_en"=>"Moncks Corner", "lang"=>"en", "lat"=>33.195359, "lon"=>-79.999044),
"/US/US-SC/monetta-19204" => array("type"=>"town", "pop"=>217, "name_en"=>"Monetta", "lang"=>"en", "lat"=>33.849826, "lon"=>-81.610507),
"/US/US-SC/mount-carmel-19205" => array("type"=>"town", "pop"=>117, "name_ru"=>"Маунт Кармел", "name_en"=>"Mount Carmel", "lang"=>"en", "lat"=>34.006095, "lon"=>-82.507942),
"/US/US-SC/mount-croghan-19206" => array("type"=>"town", "pop"=>152, "name_en"=>"Mount Croghan", "lang"=>"en", "lat"=>34.769314, "lon"=>-80.225138),
"/US/US-SC/mount-pleasant-19207" => array("type"=>"town", "pop"=>65472, "name_ru"=>"Маунт Плезант", "name_en"=>"Mount Pleasant", "lang"=>"en", "lat"=>32.823120, "lon"=>-79.863830),
"/US/US-SC/mulberry-19208" => array("type"=>"cdp", "pop"=>1097, "name_ru"=>"Малберри", "name_en"=>"Mulberry", "lang"=>"en", "lat"=>33.951859, "lon"=>-80.333115),
"/US/US-SC/mullins-19209" => array("type"=>"city", "pop"=>4716, "name_en"=>"Mullins", "lang"=>"en", "lat"=>34.204617, "lon"=>-79.253631),
"/US/US-SC/murrells-inlet-19210" => array("type"=>"cdp", "pop"=>3334, "name_en"=>"Murrells Inlet", "lang"=>"en", "lat"=>33.554889, "lon"=>-79.050009),
"/US/US-SC/myrtle-beach-19211" => array("type"=>"city", "pop"=>30596, "name_en"=>"Myrtle Beach", "lang"=>"en", "lat"=>33.698789, "lon"=>-78.892156),
"/US/US-SC/neeses-19212" => array("type"=>"town", "pop"=>395, "name_en"=>"Neeses", "lang"=>"en", "lat"=>33.536234, "lon"=>-81.126144),
"/US/US-SC/new-ellenton-19214" => array("type"=>"town", "pop"=>2227, "name_en"=>"New Ellenton", "lang"=>"en", "lat"=>33.418764, "lon"=>-81.686635),
"/US/US-SC/newberry-19213" => array("type"=>"town", "pop"=>10907, "name_ru"=>"Ньюберри", "name_en"=>"Newberry", "lang"=>"en", "lat"=>34.278901, "lon"=>-81.611869),
"/US/US-SC/nichols-19215" => array("type"=>"town", "pop"=>396, "name_ru"=>"Николс", "name_en"=>"Nichols", "lang"=>"en", "lat"=>34.234746, "lon"=>-79.149252),
"/US/US-SC/ninety-six-19216" => array("type"=>"town", "pop"=>1921, "name_en"=>"Ninety Six", "lang"=>"en", "lat"=>34.172016, "lon"=>-82.023543),
"/US/US-SC/norris-19217" => array("type"=>"town", "pop"=>852, "name_ru"=>"Норрис", "name_en"=>"Norris", "lang"=>"en", "lat"=>34.764794, "lon"=>-82.752330),
"/US/US-SC/north-19218" => array("type"=>"town", "pop"=>776, "name_en"=>"North", "lang"=>"en", "lat"=>33.617059, "lon"=>-81.103134),
"/US/US-SC/north-augusta-19219" => array("type"=>"city", "pop"=>20712, "name_en"=>"North Augusta", "lang"=>"en", "lat"=>33.517733, "lon"=>-81.956220),
"/US/US-SC/north-charleston-19220" => array("type"=>"city", "pop"=>94407, "name_en"=>"North Charleston", "lang"=>"en", "lat"=>32.905317, "lon"=>-80.041414),
"/US/US-SC/north-hartsville-19221" => array("type"=>"cdp", "pop"=>2906, "name_en"=>"North Hartsville", "lang"=>"en", "lat"=>34.399503, "lon"=>-80.071201),
"/US/US-SC/north-myrtle-beach-19223" => array("type"=>"city", "pop"=>15816, "name_en"=>"North Myrtle Beach", "lang"=>"en", "lat"=>33.827621, "lon"=>-78.671138),
"/US/US-SC/northlake-19222" => array("type"=>"cdp", "pop"=>3162, "name_ru"=>"Нортлейк", "name_en"=>"Northlake", "lang"=>"en", "lat"=>34.566238, "lon"=>-82.684200),
"/US/US-SC/norway-19224" => array("type"=>"town", "pop"=>367, "name_ru"=>"Норуэй", "name_en"=>"Norway", "lang"=>"en", "lat"=>33.449592, "lon"=>-81.126832),
"/US/US-SC/oak-grove-19225" => array("type"=>"cdp", "pop"=>7173, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>33.980650, "lon"=>-81.143433),
"/US/US-SC/oakland-19226" => array("type"=>"cdp", "pop"=>1298, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>33.987899, "lon"=>-80.499467),
"/US/US-SC/olanta-19227" => array("type"=>"town", "pop"=>627, "name_en"=>"Olanta", "lang"=>"en", "lat"=>33.935420, "lon"=>-79.932808),
"/US/US-SC/olar-19228" => array("type"=>"town", "pop"=>212, "name_en"=>"Olar", "lang"=>"en", "lat"=>33.180291, "lon"=>-81.185194),
"/US/US-SC/orangeburg-19229" => array("type"=>"city", "pop"=>13224, "name_en"=>"Orangeburg", "lang"=>"en", "lat"=>33.493934, "lon"=>-80.865401),
"/US/US-SC/pacolet-19230" => array("type"=>"town", "pop"=>2825, "name_en"=>"Pacolet", "lang"=>"en", "lat"=>34.899179, "lon"=>-81.766717),
"/US/US-SC/pacolet-mills-19231" => array("type"=>"town", "pop"=>696, "name_en"=>"Pacolet Mills", "lang"=>"en", "lat"=>34.918576, "lon"=>-81.747179),
"/US/US-SC/pageland-19232" => array("type"=>"town", "pop"=>2513, "name_en"=>"Pageland", "lang"=>"en", "lat"=>34.772973, "lon"=>-80.389729),
"/US/US-SC/pamplico-19233" => array("type"=>"town", "pop"=>1154, "name_en"=>"Pamplico", "lang"=>"en", "lat"=>33.995544, "lon"=>-79.569045),
"/US/US-SC/parker-19234" => array("type"=>"cdp", "pop"=>11072, "name_ru"=>"Паркер", "name_en"=>"Parker", "lang"=>"en", "lat"=>34.850485, "lon"=>-82.453592),
"/US/US-SC/parksville-19235" => array("type"=>"town", "pop"=>115, "name_en"=>"Parksville", "lang"=>"en", "lat"=>33.786159, "lon"=>-82.217925),
"/US/US-SC/parris-island-19236" => array("type"=>"cdp", "pop"=>7172, "name_en"=>"Parris Island", "lang"=>"en", "lat"=>32.328746, "lon"=>-80.694790),
"/US/US-SC/patrick-19237" => array("type"=>"town", "pop"=>347, "name_en"=>"Patrick", "lang"=>"en", "lat"=>34.575703, "lon"=>-80.044653),
"/US/US-SC/pawley's-island-19238" => array("type"=>"town", "pop"=>176, "name_en"=>"Pawley's Island", "lang"=>"en", "lat"=>33.422007, "lon"=>-79.126103),
"/US/US-SC/paxville-19239" => array("type"=>"town", "pop"=>244, "name_en"=>"Paxville", "lang"=>"en", "lat"=>33.739762, "lon"=>-80.360437),
"/US/US-SC/peak-19240" => array("type"=>"town", "pop"=>96, "name_en"=>"Peak", "lang"=>"en", "lat"=>34.238106, "lon"=>-81.329058),
"/US/US-SC/pelion-19241" => array("type"=>"town", "pop"=>593, "name_en"=>"Pelion", "lang"=>"en", "lat"=>33.780230, "lon"=>-81.247904),
"/US/US-SC/pelzer-19242" => array("type"=>"town", "pop"=>103, "name_en"=>"Pelzer", "lang"=>"en", "lat"=>34.643808, "lon"=>-82.461888),
"/US/US-SC/pendleton-19243" => array("type"=>"town", "pop"=>3119, "name_ru"=>"Пендлтон", "name_en"=>"Pendleton", "lang"=>"en", "lat"=>34.650818, "lon"=>-82.782765),
"/US/US-SC/perry-19244" => array("type"=>"town", "pop"=>239, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>33.627000, "lon"=>-81.309210),
"/US/US-SC/pickens-19245" => array("type"=>"town", "pop"=>3026, "name_ru"=>"Пикенс", "name_en"=>"Pickens", "lang"=>"en", "lat"=>34.882775, "lon"=>-82.708106),
"/US/US-SC/piedmont-19246" => array("type"=>"cdp", "pop"=>4143, "name_ru"=>"Пидмонт", "name_en"=>"Piedmont", "lang"=>"en", "lat"=>34.710147, "lon"=>-82.464316),
"/US/US-SC/pineridge-19247" => array("type"=>"town", "pop"=>1731, "name_en"=>"Pineridge", "lang"=>"en", "lat"=>33.910928, "lon"=>-81.095031),
"/US/US-SC/pinewood-19248" => array("type"=>"town", "pop"=>482, "name_ru"=>"Пайнвуд", "name_en"=>"Pinewood", "lang"=>"en", "lat"=>33.739300, "lon"=>-80.462716),
"/US/US-SC/plum-branch-19249" => array("type"=>"town", "pop"=>94, "name_en"=>"Plum Branch", "lang"=>"en", "lat"=>33.849429, "lon"=>-82.259244),
"/US/US-SC/pomaria-19250" => array("type"=>"town", "pop"=>181, "name_en"=>"Pomaria", "lang"=>"en", "lat"=>34.266391, "lon"=>-81.419758),
"/US/US-SC/port-royal-19251" => array("type"=>"town", "pop"=>11064, "name_en"=>"Port Royal", "lang"=>"en", "lat"=>32.382896, "lon"=>-80.693277),
"/US/US-SC/prosperity-19252" => array("type"=>"town", "pop"=>1059, "name_en"=>"Prosperity", "lang"=>"en", "lat"=>34.210330, "lon"=>-81.534304),
"/US/US-SC/quinby-19253" => array("type"=>"town", "pop"=>867, "name_en"=>"Quinby", "lang"=>"en", "lat"=>34.232250, "lon"=>-79.732553),
"/US/US-SC/ravenel-19254" => array("type"=>"town", "pop"=>2288, "name_en"=>"Ravenel", "lang"=>"en", "lat"=>32.777100, "lon"=>-80.224764),
"/US/US-SC/red-bank-19255" => array("type"=>"cdp", "pop"=>5950, "name_en"=>"Red Bank", "lang"=>"en", "lat"=>33.928900, "lon"=>-81.233280),
"/US/US-SC/red-hill-19256" => array("type"=>"cdp", "pop"=>6112, "name_en"=>"Red Hill", "lang"=>"en", "lat"=>33.787073, "lon"=>-79.009746),
"/US/US-SC/reevesville-19257" => array("type"=>"town", "pop"=>211, "name_en"=>"Reevesville", "lang"=>"en", "lat"=>33.202435, "lon"=>-80.646004),
"/US/US-SC/richburg-19258" => array("type"=>"town", "pop"=>312, "name_en"=>"Richburg", "lang"=>"en", "lat"=>34.715547, "lon"=>-81.020287),
"/US/US-SC/ridge-spring-19260" => array("type"=>"town", "pop"=>781, "name_en"=>"Ridge Spring", "lang"=>"en", "lat"=>33.845151, "lon"=>-81.662693),
"/US/US-SC/ridgeland-19259" => array("type"=>"town", "pop"=>2589, "name_ru"=>"Риджленд", "name_en"=>"Ridgeland", "lang"=>"en", "lat"=>32.483704, "lon"=>-80.981430),
"/US/US-SC/ridgeville-19261" => array("type"=>"town", "pop"=>1984, "name_ru"=>"Риджвилл", "name_en"=>"Ridgeville", "lang"=>"en", "lat"=>33.089914, "lon"=>-80.306682),
"/US/US-SC/ridgeway-19262" => array("type"=>"town", "pop"=>328, "name_ru"=>"Риджуэй", "name_en"=>"Ridgeway", "lang"=>"en", "lat"=>34.306594, "lon"=>-80.960658),
"/US/US-SC/rock-hill-19263" => array("type"=>"city", "pop"=>67339, "name_ru"=>"Рок Хилл", "name_en"=>"Rock Hill", "lang"=>"en", "lat"=>34.939300, "lon"=>-81.028449),
"/US/US-SC/roebuck-19264" => array("type"=>"cdp", "pop"=>1966, "name_en"=>"Roebuck", "lang"=>"en", "lat"=>34.876064, "lon"=>-81.964189),
"/US/US-SC/rowesville-19265" => array("type"=>"town", "pop"=>360, "name_en"=>"Rowesville", "lang"=>"en", "lat"=>33.372212, "lon"=>-80.836336),
"/US/US-SC/ruby-19266" => array("type"=>"town", "pop"=>341, "name_ru"=>"Раби", "name_en"=>"Ruby", "lang"=>"en", "lat"=>34.744139, "lon"=>-80.179736),
"/US/US-SC/salem-19271" => array("type"=>"town", "pop"=>132, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>34.891460, "lon"=>-82.977854),
"/US/US-SC/salley-19272" => array("type"=>"town", "pop"=>414, "name_en"=>"Salley", "lang"=>"en", "lat"=>33.566617, "lon"=>-81.304344),
"/US/US-SC/saluda-19273" => array("type"=>"town", "pop"=>2910, "name_en"=>"Saluda", "lang"=>"en", "lat"=>34.000734, "lon"=>-81.771468),
"/US/US-SC/sans-souci-19274" => array("type"=>"cdp", "pop"=>7612, "name_en"=>"Sans Souci", "lang"=>"en", "lat"=>34.891404, "lon"=>-82.425410),
"/US/US-SC/santee-19275" => array("type"=>"town", "pop"=>718, "name_ru"=>"Санти", "name_en"=>"Santee", "lang"=>"en", "lat"=>33.480065, "lon"=>-80.487149),
"/US/US-SC/saxon-19276" => array("type"=>"cdp", "pop"=>4002, "name_en"=>"Saxon", "lang"=>"en", "lat"=>34.962067, "lon"=>-81.971312),
"/US/US-SC/scotia-19277" => array("type"=>"town", "pop"=>221, "name_en"=>"Scotia", "lang"=>"en", "lat"=>32.681936, "lon"=>-81.243069),
"/US/US-SC/scranton-19278" => array("type"=>"town", "pop"=>1032, "name_ru"=>"Скрантон", "name_en"=>"Scranton", "lang"=>"en", "lat"=>33.917252, "lon"=>-79.743589),
"/US/US-SC/seabrook-island-19279" => array("type"=>"town", "pop"=>1202, "name_en"=>"Seabrook Island", "lang"=>"en", "lat"=>32.580235, "lon"=>-80.175511),
"/US/US-SC/sellers-19280" => array("type"=>"town", "pop"=>261, "name_en"=>"Sellers", "lang"=>"en", "lat"=>34.282539, "lon"=>-79.472197),
"/US/US-SC/seneca-19281" => array("type"=>"town", "pop"=>8038, "name_ru"=>"Сенека", "name_en"=>"Seneca", "lang"=>"en", "lat"=>34.686129, "lon"=>-82.957825),
"/US/US-SC/seven-oaks-19282" => array("type"=>"cdp", "pop"=>15722, "name_en"=>"Seven Oaks", "lang"=>"en", "lat"=>34.048650, "lon"=>-81.146542),
"/US/US-SC/sharon-19283" => array("type"=>"town", "pop"=>504, "name_ru"=>"Шэрон", "name_en"=>"Sharon", "lang"=>"en", "lat"=>34.953963, "lon"=>-81.344308),
"/US/US-SC/shell-point-19284" => array("type"=>"cdp", "pop"=>2885, "name_en"=>"Shell Point", "lang"=>"en", "lat"=>32.378714, "lon"=>-80.749364),
"/US/US-SC/silverstreet-19285" => array("type"=>"town", "pop"=>225, "name_en"=>"Silverstreet", "lang"=>"en", "lat"=>34.218059, "lon"=>-81.714828),
"/US/US-SC/simpsonville-19286" => array("type"=>"town", "pop"=>17144, "name_ru"=>"Симпсонвилл", "name_en"=>"Simpsonville", "lang"=>"en", "lat"=>34.731360, "lon"=>-82.258991),
"/US/US-SC/six-mile-19287" => array("type"=>"town", "pop"=>582, "name_en"=>"Six Mile", "lang"=>"en", "lat"=>34.807788, "lon"=>-82.819085),
"/US/US-SC/slater-marietta-19288" => array("type"=>"cdp", "pop"=>2245, "name_en"=>"Slater-Marietta", "lang"=>"en", "lat"=>35.033468, "lon"=>-82.494199),
"/US/US-SC/smoaks-19289" => array("type"=>"town", "pop"=>134, "name_en"=>"Smoaks", "lang"=>"en", "lat"=>33.089459, "lon"=>-80.816338),
"/US/US-SC/smyrna-19290" => array("type"=>"town", "pop"=>80, "name_ru"=>"Смирна", "name_en"=>"Smyrna", "lang"=>"en", "lat"=>35.041990, "lon"=>-81.411300),
"/US/US-SC/snelling-19291" => array("type"=>"town", "pop"=>239, "name_en"=>"Snelling", "lang"=>"en", "lat"=>33.240325, "lon"=>-81.455936),
"/US/US-SC/socastee-19292" => array("type"=>"cdp", "pop"=>10426, "name_en"=>"Socastee", "lang"=>"en", "lat"=>33.687082, "lon"=>-79.008934),
"/US/US-SC/society-hill-19293" => array("type"=>"town", "pop"=>685, "name_en"=>"Society Hill", "lang"=>"en", "lat"=>34.509152, "lon"=>-79.852279),
"/US/US-SC/south-congaree-19294" => array("type"=>"town", "pop"=>2370, "name_en"=>"South Congaree", "lang"=>"en", "lat"=>33.909515, "lon"=>-81.137060),
"/US/US-SC/south-sumter-19296" => array("type"=>"cdp", "pop"=>4371, "name_en"=>"South Sumter", "lang"=>"en", "lat"=>33.888394, "lon"=>-80.340408),
"/US/US-SC/southern-shops-19295" => array("type"=>"cdp", "pop"=>3378, "name_en"=>"Southern Shops", "lang"=>"en", "lat"=>34.985750, "lon"=>-81.994799),
"/US/US-SC/spartanburg-19297" => array("type"=>"city", "pop"=>39584, "name_en"=>"Spartanburg", "lang"=>"en", "lat"=>34.944888, "lon"=>-81.928319),
"/US/US-SC/springdale-19298" => array("type"=>"town", "pop"=>2939, "name_ru"=>"Спрингдейл", "name_en"=>"Springdale", "lang"=>"en", "lat"=>33.959800, "lon"=>-81.112375),
"/US/US-SC/springdale-19299" => array("type"=>"cdp", "pop"=>2643, "name_ru"=>"Спрингдейл", "name_en"=>"Springdale", "lang"=>"en", "lat"=>34.689117, "lon"=>-80.786280),
"/US/US-SC/springfield-19300" => array("type"=>"town", "pop"=>482, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>33.495669, "lon"=>-81.279468),
"/US/US-SC/st.-andrews-19267" => array("type"=>"cdp", "pop"=>25692, "name_en"=>"St. Andrews", "lang"=>"en", "lat"=>34.048500, "lon"=>-81.103379),
"/US/US-SC/st.-george-19268" => array("type"=>"town", "pop"=>2121, "name_ru"=>"Сент-Джрдж", "name_en"=>"St. George", "lang"=>"en", "lat"=>33.185475, "lon"=>-80.581548),
"/US/US-SC/st.-matthews-19269" => array("type"=>"town", "pop"=>1948, "name_ru"=>"Сент Мэтьюс", "name_en"=>"St. Matthews", "lang"=>"en", "lat"=>33.663678, "lon"=>-80.777914),
"/US/US-SC/st.-stephen-19270" => array("type"=>"town", "pop"=>1778, "name_ru"=>"Сент Стифен", "name_en"=>"St. Stephen", "lang"=>"en", "lat"=>33.404535, "lon"=>-79.924656),
"/US/US-SC/starr-19301" => array("type"=>"town", "pop"=>199, "name_en"=>"Starr", "lang"=>"en", "lat"=>34.377132, "lon"=>-82.695755),
"/US/US-SC/startex-19302" => array("type"=>"cdp", "pop"=>1162, "name_en"=>"Startex", "lang"=>"en", "lat"=>34.930376, "lon"=>-82.098861),
"/US/US-SC/stuckey-19303" => array("type"=>"town", "pop"=>233, "name_en"=>"Stuckey", "lang"=>"en", "lat"=>33.730921, "lon"=>-79.512889),
"/US/US-SC/sullivan's-island-19304" => array("type"=>"town", "pop"=>1870, "name_en"=>"Sullivan's Island", "lang"=>"en", "lat"=>32.765848, "lon"=>-79.835854),
"/US/US-SC/summerton-19305" => array("type"=>"town", "pop"=>1028, "name_en"=>"Summerton", "lang"=>"en", "lat"=>33.602177, "lon"=>-80.351809),
"/US/US-SC/summerville-19306" => array("type"=>"town", "pop"=>45193, "name_ru"=>"Саммервилл", "name_en"=>"Summerville", "lang"=>"en", "lat"=>33.004240, "lon"=>-80.178861),
"/US/US-SC/summit-19307" => array("type"=>"town", "pop"=>263, "name_ru"=>"Саммит", "name_en"=>"Summit", "lang"=>"en", "lat"=>33.924331, "lon"=>-81.423261),
"/US/US-SC/sumter-19308" => array("type"=>"city", "pop"=>38625, "name_en"=>"Sumter", "lang"=>"en", "lat"=>33.944365, "lon"=>-80.392472),
"/US/US-SC/surfside-beach-19309" => array("type"=>"town", "pop"=>4783, "name_en"=>"Surfside Beach", "lang"=>"en", "lat"=>33.609594, "lon"=>-78.977737),
"/US/US-SC/swansea-19310" => array("type"=>"town", "pop"=>799, "name_ru"=>"Свонси", "name_en"=>"Swansea", "lang"=>"en", "lat"=>33.736055, "lon"=>-81.106402),
"/US/US-SC/sycamore-19311" => array("type"=>"town", "pop"=>171, "name_ru"=>"Сикамор", "name_en"=>"Sycamore", "lang"=>"en", "lat"=>33.036504, "lon"=>-81.222643),
"/US/US-SC/tatum-19312" => array("type"=>"town", "pop"=>62, "name_en"=>"Tatum", "lang"=>"en", "lat"=>34.643798, "lon"=>-79.586735),
"/US/US-SC/taylors-19313" => array("type"=>"cdp", "pop"=>19619, "name_en"=>"Taylors", "lang"=>"en", "lat"=>34.915184, "lon"=>-82.313545),
"/US/US-SC/tega-cay-19314" => array("type"=>"city", "pop"=>5014, "name_en"=>"Tega Cay", "lang"=>"en", "lat"=>35.034653, "lon"=>-81.018782),
"/US/US-SC/timmonsville-19315" => array("type"=>"town", "pop"=>2373, "name_en"=>"Timmonsville", "lang"=>"en", "lat"=>34.134152, "lon"=>-79.942403),
"/US/US-SC/travelers-rest-19316" => array("type"=>"city", "pop"=>4523, "name_en"=>"Travelers Rest", "lang"=>"en", "lat"=>34.967216, "lon"=>-82.437847),
"/US/US-SC/trenton-19317" => array("type"=>"town", "pop"=>270, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>33.740900, "lon"=>-81.840159),
"/US/US-SC/troy-19318" => array("type"=>"town", "pop"=>107, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>33.987357, "lon"=>-82.296811),
"/US/US-SC/turbeville-19319" => array("type"=>"town", "pop"=>707, "name_en"=>"Turbeville", "lang"=>"en", "lat"=>33.890330, "lon"=>-80.015362),
"/US/US-SC/ulmer-19320" => array("type"=>"town", "pop"=>94, "name_en"=>"Ulmer", "lang"=>"en", "lat"=>33.095363, "lon"=>-81.207666),
"/US/US-SC/union-19321" => array("type"=>"city", "pop"=>8039, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>34.722796, "lon"=>-81.623809),
"/US/US-SC/utica-19322" => array("type"=>"cdp", "pop"=>1478, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>34.679617, "lon"=>-82.926199),
"/US/US-SC/valencia-heights-19323" => array("type"=>"cdp", "pop"=>4122, "name_en"=>"Valencia Heights", "lang"=>"en", "lat"=>33.976433, "lon"=>-80.985344),
"/US/US-SC/valley-falls-19324" => array("type"=>"cdp", "pop"=>3504, "name_ru"=>"Валли Фоллс", "name_en"=>"Valley Falls", "lang"=>"en", "lat"=>35.008728, "lon"=>-81.974364),
"/US/US-SC/vance-19325" => array("type"=>"town", "pop"=>199, "name_ru"=>"Ванс", "name_en"=>"Vance", "lang"=>"en", "lat"=>33.435758, "lon"=>-80.421474),
"/US/US-SC/varnville-19326" => array("type"=>"town", "pop"=>2042, "name_en"=>"Varnville", "lang"=>"en", "lat"=>32.851570, "lon"=>-81.080393),
"/US/US-SC/wade-hampton-19327" => array("type"=>"cdp", "pop"=>20014, "name_en"=>"Wade Hampton", "lang"=>"en", "lat"=>34.880256, "lon"=>-82.333899),
"/US/US-SC/wagener-19328" => array("type"=>"town", "pop"=>875, "name_en"=>"Wagener", "lang"=>"en", "lat"=>33.650908, "lon"=>-81.362964),
"/US/US-SC/walhalla-19329" => array("type"=>"town", "pop"=>3576, "name_en"=>"Walhalla", "lang"=>"en", "lat"=>34.772014, "lon"=>-83.061775),
"/US/US-SC/walterboro-19330" => array("type"=>"city", "pop"=>5780, "name_en"=>"Walterboro", "lang"=>"en", "lat"=>32.900757, "lon"=>-80.670388),
"/US/US-SC/ward-19331" => array("type"=>"town", "pop"=>105, "name_ru"=>"Уорд", "name_en"=>"Ward", "lang"=>"en", "lat"=>33.857193, "lon"=>-81.732324),
"/US/US-SC/ware-shoals-19332" => array("type"=>"town", "pop"=>2348, "name_en"=>"Ware Shoals", "lang"=>"en", "lat"=>34.390109, "lon"=>-82.244227),
"/US/US-SC/waterloo-19333" => array("type"=>"town", "pop"=>201, "name_ru"=>"Ватрелоо", "name_en"=>"Waterloo", "lang"=>"en", "lat"=>34.352845, "lon"=>-82.058236),
"/US/US-SC/watts-mills-19334" => array("type"=>"cdp", "pop"=>1535, "name_en"=>"Watts Mills", "lang"=>"en", "lat"=>34.516025, "lon"=>-81.985160),
"/US/US-SC/welcome-19335" => array("type"=>"cdp", "pop"=>6560, "name_ru"=>"Вэлкам", "name_en"=>"Welcome", "lang"=>"en", "lat"=>34.820944, "lon"=>-82.458027),
"/US/US-SC/wellford-19336" => array("type"=>"city", "pop"=>2465, "name_en"=>"Wellford", "lang"=>"en", "lat"=>34.956789, "lon"=>-82.097187),
"/US/US-SC/west-columbia-19337" => array("type"=>"city", "pop"=>13914, "name_en"=>"West Columbia", "lang"=>"en", "lat"=>33.991500, "lon"=>-81.081430),
"/US/US-SC/west-pelzer-19339" => array("type"=>"town", "pop"=>915, "name_en"=>"West Pelzer", "lang"=>"en", "lat"=>34.644662, "lon"=>-82.475022),
"/US/US-SC/west-union-19340" => array("type"=>"town", "pop"=>301, "name_ru"=>"Вэст Юнион", "name_en"=>"West Union", "lang"=>"en", "lat"=>34.760334, "lon"=>-83.043064),
"/US/US-SC/westminster-19338" => array("type"=>"town", "pop"=>2663, "name_ru"=>"Уэстминстер", "name_en"=>"Westminster", "lang"=>"en", "lat"=>34.666214, "lon"=>-83.091631),
"/US/US-SC/whitmire-19341" => array("type"=>"town", "pop"=>1528, "name_en"=>"Whitmire", "lang"=>"en", "lat"=>34.503399, "lon"=>-81.615565),
"/US/US-SC/wilkinson-heights-19342" => array("type"=>"cdp", "pop"=>3394, "name_en"=>"Wilkinson Heights", "lang"=>"en", "lat"=>33.491646, "lon"=>-80.829944),
"/US/US-SC/williams-19343" => array("type"=>"town", "pop"=>116, "name_ru"=>"Уилльямс", "name_en"=>"Williams", "lang"=>"en", "lat"=>33.033914, "lon"=>-80.843106),
"/US/US-SC/williamston-19344" => array("type"=>"town", "pop"=>3964, "name_ru"=>"Уилльямстон", "name_en"=>"Williamston", "lang"=>"en", "lat"=>34.617586, "lon"=>-82.479407),
"/US/US-SC/williston-19345" => array("type"=>"town", "pop"=>3190, "name_ru"=>"Уиллистон", "name_en"=>"Williston", "lang"=>"en", "lat"=>33.401742, "lon"=>-81.422668),
"/US/US-SC/windsor-19346" => array("type"=>"town", "pop"=>131, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>33.480796, "lon"=>-81.515103),
"/US/US-SC/winnsboro-19347" => array("type"=>"town", "pop"=>3577, "name_ru"=>"Уиннсборо", "name_en"=>"Winnsboro", "lang"=>"en", "lat"=>34.376603, "lon"=>-81.089709),
"/US/US-SC/winnsboro-mills-19348" => array("type"=>"cdp", "pop"=>2275, "name_en"=>"Winnsboro Mills", "lang"=>"en", "lat"=>34.354056, "lon"=>-81.069481),
"/US/US-SC/woodfield-19349" => array("type"=>"cdp", "pop"=>8862, "name_en"=>"Woodfield", "lang"=>"en", "lat"=>34.059100, "lon"=>-80.930994),
"/US/US-SC/woodford-19350" => array("type"=>"town", "pop"=>188, "name_en"=>"Woodford", "lang"=>"en", "lat"=>33.667769, "lon"=>-81.111708),
"/US/US-SC/woodruff-19351" => array("type"=>"town", "pop"=>4103, "name_en"=>"Woodruff", "lang"=>"en", "lat"=>34.740405, "lon"=>-82.032756),
"/US/US-SC/yemassee-19352" => array("type"=>"town", "pop"=>860, "name_en"=>"Yemassee", "lang"=>"en", "lat"=>32.694594, "lon"=>-80.853262),
"/US/US-SC/york-19353" => array("type"=>"city", "pop"=>7967, "name_ru"=>"Йорк", "name_en"=>"York", "lang"=>"en", "lat"=>34.995158, "lon"=>-81.236113),
"/US/US-SD" => array("type"=>"state", "pop"=>804194, "name_ru"=>"Южная Дакота", "name_en"=>"South Dakota", "lang"=>"en"),
"/US/US-SD/aberdeen-19354" => array("type"=>"city", "pop"=>24460, "name_ru"=>"Абердин", "name_en"=>"Aberdeen", "lang"=>"en", "lat"=>45.467504, "lon"=>-98.482401),
"/US/US-SD/agar-19355" => array("type"=>"town", "pop"=>70, "name_en"=>"Agar", "lang"=>"en", "lat"=>44.838696, "lon"=>-100.072469),
"/US/US-SD/akaska-19356" => array("type"=>"town", "pop"=>28, "name_en"=>"Akaska", "lang"=>"en", "lat"=>45.331854, "lon"=>-100.120742),
"/US/US-SD/albee-19357" => array("type"=>"town", "pop"=>15, "name_en"=>"Albee", "lang"=>"en", "lat"=>45.051331, "lon"=>-96.554316),
"/US/US-SD/alcester-19358" => array("type"=>"city", "pop"=>892, "name_en"=>"Alcester", "lang"=>"en", "lat"=>43.023063, "lon"=>-96.628696),
"/US/US-SD/alexandria-19359" => array("type"=>"city", "pop"=>650, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>43.653919, "lon"=>-97.779022),
"/US/US-SD/alpena-19360" => array("type"=>"town", "pop"=>225, "name_ru"=>"Алпена", "name_en"=>"Alpena", "lang"=>"en", "lat"=>44.183388, "lon"=>-98.367887),
"/US/US-SD/altamont-19361" => array("type"=>"town", "pop"=>35, "name_ru"=>"Альтамонт", "name_en"=>"Altamont", "lang"=>"en", "lat"=>44.840379, "lon"=>-96.690138),
"/US/US-SD/andover-19362" => array("type"=>"town", "pop"=>86, "name_ru"=>"Андоувер", "name_en"=>"Andover", "lang"=>"en", "lat"=>45.410569, "lon"=>-97.903324),
"/US/US-SD/antelope-19363" => array("type"=>"cdp", "pop"=>20, "name_en"=>"Antelope", "lang"=>"en", "lat"=>43.305617, "lon"=>-100.628328),
"/US/US-SD/arlington-19364" => array("type"=>"city", "pop"=>935, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>44.362479, "lon"=>-97.133911),
"/US/US-SD/armour-19365" => array("type"=>"city", "pop"=>654, "name_en"=>"Armour", "lang"=>"en", "lat"=>43.319386, "lon"=>-98.343870),
"/US/US-SD/artas-19366" => array("type"=>"town", "pop"=>10, "name_en"=>"Artas", "lang"=>"en", "lat"=>45.887739, "lon"=>-99.805968),
"/US/US-SD/artesian-19367" => array("type"=>"town", "pop"=>138, "name_en"=>"Artesian", "lang"=>"en", "lat"=>44.008527, "lon"=>-97.923014),
"/US/US-SD/ashton-19368" => array("type"=>"city", "pop"=>132, "name_ru"=>"Аштон", "name_en"=>"Ashton", "lang"=>"en", "lat"=>44.993521, "lon"=>-98.498205),
"/US/US-SD/astoria-19369" => array("type"=>"town", "pop"=>133, "name_ru"=>"Астория", "name_en"=>"Astoria", "lang"=>"en", "lat"=>44.557739, "lon"=>-96.545998),
"/US/US-SD/aurora-19370" => array("type"=>"town", "pop"=>471, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>44.282719, "lon"=>-96.686602),
"/US/US-SD/avon-19371" => array("type"=>"city", "pop"=>516, "name_ru"=>"Эйвон", "name_en"=>"Avon", "lang"=>"en", "lat"=>43.005140, "lon"=>-98.059146),
"/US/US-SD/badger-19372" => array("type"=>"town", "pop"=>132, "name_ru"=>"Баджер", "name_en"=>"Badger", "lang"=>"en", "lat"=>44.485708, "lon"=>-97.209396),
"/US/US-SD/baltic-19373" => array("type"=>"city", "pop"=>1004, "name_en"=>"Baltic", "lang"=>"en", "lat"=>43.759630, "lon"=>-96.738375),
"/US/US-SD/bancroft-19374" => array("type"=>"town", "pop"=>34, "name_ru"=>"Бэнкрофт", "name_en"=>"Bancroft", "lang"=>"en", "lat"=>44.488728, "lon"=>-97.750435),
"/US/US-SD/batesland-19375" => array("type"=>"town", "pop"=>97, "name_en"=>"Batesland", "lang"=>"en", "lat"=>43.127661, "lon"=>-102.101304),
"/US/US-SD/belle-fourche-19376" => array("type"=>"city", "pop"=>4979, "name_en"=>"Belle Fourche", "lang"=>"en", "lat"=>44.664318, "lon"=>-103.852044),
"/US/US-SD/belvidere-19377" => array("type"=>"town", "pop"=>53, "name_ru"=>"Белвидер", "name_en"=>"Belvidere", "lang"=>"en", "lat"=>43.831211, "lon"=>-101.270358),
"/US/US-SD/beresford-19378" => array("type"=>"city", "pop"=>2150, "name_en"=>"Beresford", "lang"=>"en", "lat"=>43.079705, "lon"=>-96.780092),
"/US/US-SD/big-stone-city-19379" => array("type"=>"city", "pop"=>542, "name_en"=>"Big Stone City", "lang"=>"en", "lat"=>45.295158, "lon"=>-96.464373),
"/US/US-SD/bison-19380" => array("type"=>"town", "pop"=>326, "name_ru"=>"Бизон", "name_en"=>"Bison", "lang"=>"en", "lat"=>45.521735, "lon"=>-102.466871),
"/US/US-SD/blackhawk-19381" => array("type"=>"cdp", "pop"=>1995, "name_ru"=>"Блэкхок", "name_en"=>"Blackhawk", "lang"=>"en", "lat"=>44.151682, "lon"=>-103.315265),
"/US/US-SD/blunt-19382" => array("type"=>"city", "pop"=>349, "name_en"=>"Blunt", "lang"=>"en", "lat"=>44.515472, "lon"=>-99.987965),
"/US/US-SD/bonesteel-19383" => array("type"=>"city", "pop"=>255, "name_en"=>"Bonesteel", "lang"=>"en", "lat"=>43.078104, "lon"=>-98.946828),
"/US/US-SD/bowdle-19384" => array("type"=>"city", "pop"=>501, "name_en"=>"Bowdle", "lang"=>"en", "lat"=>45.451665, "lon"=>-99.655894),
"/US/US-SD/box-elder-19385" => array("type"=>"city", "pop"=>3517, "name_en"=>"Box Elder", "lang"=>"en", "lat"=>44.119567, "lon"=>-103.072069),
"/US/US-SD/bradley-19386" => array("type"=>"town", "pop"=>91, "name_ru"=>"Брадли", "name_en"=>"Bradley", "lang"=>"en", "lat"=>45.090113, "lon"=>-97.641389),
"/US/US-SD/brandon-19387" => array("type"=>"city", "pop"=>8757, "name_ru"=>"Брэндон", "name_en"=>"Brandon", "lang"=>"en", "lat"=>43.591298, "lon"=>-96.577956),
"/US/US-SD/brandt-19388" => array("type"=>"town", "pop"=>103, "name_en"=>"Brandt", "lang"=>"en", "lat"=>44.665535, "lon"=>-96.625013),
"/US/US-SD/brentford-19389" => array("type"=>"town", "pop"=>56, "name_en"=>"Brentford", "lang"=>"en", "lat"=>45.159235, "lon"=>-98.321547),
"/US/US-SD/bridgewater-19390" => array("type"=>"city", "pop"=>557, "name_ru"=>"Бриджуотер", "name_en"=>"Bridgewater", "lang"=>"en", "lat"=>43.550200, "lon"=>-97.498576),
"/US/US-SD/bristol-19391" => array("type"=>"city", "pop"=>323, "name_ru"=>"Бристол", "name_en"=>"Bristol", "lang"=>"en", "lat"=>45.346645, "lon"=>-97.748959),
"/US/US-SD/britton-19392" => array("type"=>"city", "pop"=>1266, "name_ru"=>"Бриттон", "name_en"=>"Britton", "lang"=>"en", "lat"=>45.789550, "lon"=>-97.749489),
"/US/US-SD/broadland-19393" => array("type"=>"town", "pop"=>35, "name_en"=>"Broadland", "lang"=>"en", "lat"=>44.493398, "lon"=>-98.348122),
"/US/US-SD/brookings-19394" => array("type"=>"city", "pop"=>19865, "name_en"=>"Brookings", "lang"=>"en", "lat"=>44.306284, "lon"=>-96.786151),
"/US/US-SD/bruce-19395" => array("type"=>"city", "pop"=>260, "name_ru"=>"Брюс", "name_en"=>"Bruce", "lang"=>"en", "lat"=>44.438086, "lon"=>-96.889435),
"/US/US-SD/bryant-19396" => array("type"=>"city", "pop"=>373, "name_ru"=>"Брайант", "name_en"=>"Bryant", "lang"=>"en", "lat"=>44.589949, "lon"=>-97.467418),
"/US/US-SD/buffalo-19397" => array("type"=>"town", "pop"=>320, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>45.586094, "lon"=>-103.543006),
"/US/US-SD/buffalo-gap-19398" => array("type"=>"town", "pop"=>153, "name_en"=>"Buffalo Gap", "lang"=>"en", "lat"=>43.492579, "lon"=>-103.313792),
"/US/US-SD/bullhead-19399" => array("type"=>"cdp", "pop"=>179, "name_en"=>"Bullhead", "lang"=>"en", "lat"=>45.767938, "lon"=>-101.080067),
"/US/US-SD/burke-19400" => array("type"=>"city", "pop"=>571, "name_en"=>"Burke", "lang"=>"en", "lat"=>43.183021, "lon"=>-99.292511),
"/US/US-SD/bushnell-19401" => array("type"=>"town", "pop"=>68, "name_ru"=>"Бушнелл", "name_en"=>"Bushnell", "lang"=>"en", "lat"=>44.328371, "lon"=>-96.643181),
"/US/US-SD/butler-19402" => array("type"=>"town", "pop"=>12, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>45.258185, "lon"=>-97.711220),
"/US/US-SD/camp-crook-19403" => array("type"=>"town", "pop"=>47, "name_en"=>"Camp Crook", "lang"=>"en", "lat"=>45.549732, "lon"=>-103.974819),
"/US/US-SD/canistota-19404" => array("type"=>"city", "pop"=>656, "name_en"=>"Canistota", "lang"=>"en", "lat"=>43.597511, "lon"=>-97.291560),
"/US/US-SD/canova-19405" => array("type"=>"town", "pop"=>118, "name_en"=>"Canova", "lang"=>"en", "lat"=>43.880780, "lon"=>-97.502966),
"/US/US-SD/canton-19406" => array("type"=>"city", "pop"=>3633, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>43.303166, "lon"=>-96.580949),
"/US/US-SD/carthage-19407" => array("type"=>"city", "pop"=>155, "name_ru"=>"Картаж", "name_en"=>"Carthage", "lang"=>"en", "lat"=>44.168856, "lon"=>-97.713985),
"/US/US-SD/castlewood-19408" => array("type"=>"city", "pop"=>677, "name_ru"=>"Каслвуд", "name_en"=>"Castlewood", "lang"=>"en", "lat"=>44.724328, "lon"=>-97.030722),
"/US/US-SD/cavour-19409" => array("type"=>"town", "pop"=>132, "name_en"=>"Cavour", "lang"=>"en", "lat"=>44.372799, "lon"=>-98.041040),
"/US/US-SD/centerville-19410" => array("type"=>"city", "pop"=>844, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>43.117655, "lon"=>-96.958852),
"/US/US-SD/central-city-19411" => array("type"=>"city", "pop"=>138, "name_ru"=>"Сентрал Сити", "name_en"=>"Central City", "lang"=>"en", "lat"=>44.369167, "lon"=>-103.769185),
"/US/US-SD/chamberlain-19412" => array("type"=>"city", "pop"=>2264, "name_en"=>"Chamberlain", "lang"=>"en", "lat"=>43.807630, "lon"=>-99.320240),
"/US/US-SD/chancellor-19413" => array("type"=>"town", "pop"=>302, "name_en"=>"Chancellor", "lang"=>"en", "lat"=>43.372401, "lon"=>-96.987431),
"/US/US-SD/chelsea-19414" => array("type"=>"town", "pop"=>28, "name_ru"=>"Челси", "name_en"=>"Chelsea", "lang"=>"en", "lat"=>45.167178, "lon"=>-98.742209),
"/US/US-SD/claire-city-19415" => array("type"=>"town", "pop"=>81, "name_en"=>"Claire City", "lang"=>"en", "lat"=>45.856428, "lon"=>-97.102361),
"/US/US-SD/claremont-19416" => array("type"=>"town", "pop"=>124, "name_ru"=>"Клэрмонт", "name_en"=>"Claremont", "lang"=>"en", "lat"=>45.672007, "lon"=>-98.014482),
"/US/US-SD/clark-19417" => array("type"=>"city", "pop"=>1062, "name_ru"=>"Кларк", "name_en"=>"Clark", "lang"=>"en", "lat"=>44.880740, "lon"=>-97.733812),
"/US/US-SD/clear-lake-19418" => array("type"=>"city", "pop"=>1225, "name_ru"=>"Клир Лейк", "name_en"=>"Clear Lake", "lang"=>"en", "lat"=>44.766025, "lon"=>-96.681344),
"/US/US-SD/colman-19419" => array("type"=>"city", "pop"=>548, "name_en"=>"Colman", "lang"=>"en", "lat"=>43.984691, "lon"=>-96.815925),
"/US/US-SD/colome-19420" => array("type"=>"city", "pop"=>291, "name_en"=>"Colome", "lang"=>"en", "lat"=>43.259955, "lon"=>-99.715716),
"/US/US-SD/colonial-pine-hills-19421" => array("type"=>"cdp", "pop"=>1553, "name_en"=>"Colonial Pine Hills", "lang"=>"en", "lat"=>44.007751, "lon"=>-103.314972),
"/US/US-SD/colton-19422" => array("type"=>"city", "pop"=>676, "name_ru"=>"Колтон", "name_en"=>"Colton", "lang"=>"en", "lat"=>43.786385, "lon"=>-96.927008),
"/US/US-SD/columbia-19423" => array("type"=>"city", "pop"=>134, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>45.614905, "lon"=>-98.311577),
"/US/US-SD/conde-19424" => array("type"=>"city", "pop"=>162, "name_en"=>"Conde", "lang"=>"en", "lat"=>45.157836, "lon"=>-98.096671),
"/US/US-SD/corona-19425" => array("type"=>"town", "pop"=>105, "name_ru"=>"Корона", "name_en"=>"Corona", "lang"=>"en", "lat"=>45.334825, "lon"=>-96.765156),
"/US/US-SD/corsica-19426" => array("type"=>"city", "pop"=>570, "name_en"=>"Corsica", "lang"=>"en", "lat"=>43.424339, "lon"=>-98.405675),
"/US/US-SD/cottonwood-19427" => array("type"=>"town", "pop"=>6, "name_ru"=>"Коттонвуд", "name_en"=>"Cottonwood", "lang"=>"en", "lat"=>43.965522, "lon"=>-101.901833),
"/US/US-SD/cresbard-19428" => array("type"=>"town", "pop"=>120, "name_en"=>"Cresbard", "lang"=>"en", "lat"=>45.169626, "lon"=>-98.947567),
"/US/US-SD/crooks-19429" => array("type"=>"city", "pop"=>1218, "name_en"=>"Crooks", "lang"=>"en", "lat"=>43.660169, "lon"=>-96.807263),
"/US/US-SD/custer-19430" => array("type"=>"city", "pop"=>1940, "name_ru"=>"Кастер", "name_en"=>"Custer", "lang"=>"en", "lat"=>43.766481, "lon"=>-103.600733),
"/US/US-SD/dallas-19431" => array("type"=>"town", "pop"=>120, "name_ru"=>"Даллас", "name_en"=>"Dallas", "lang"=>"en", "lat"=>43.237896, "lon"=>-99.517712),
"/US/US-SD/dante-19432" => array("type"=>"town", "pop"=>78, "name_en"=>"Dante", "lang"=>"en", "lat"=>43.039903, "lon"=>-98.186406),
"/US/US-SD/davis-19433" => array("type"=>"town", "pop"=>94, "name_ru"=>"Давис", "name_en"=>"Davis", "lang"=>"en", "lat"=>43.258288, "lon"=>-96.995762),
"/US/US-SD/de-smet-19437" => array("type"=>"city", "pop"=>1051, "name_en"=>"De Smet", "lang"=>"en", "lat"=>44.386639, "lon"=>-97.547767),
"/US/US-SD/deadwood-19434" => array("type"=>"city", "pop"=>1283, "name_en"=>"Deadwood", "lang"=>"en", "lat"=>44.381965, "lon"=>-103.723322),
"/US/US-SD/dell-rapids-19435" => array("type"=>"city", "pop"=>3431, "name_en"=>"Dell Rapids", "lang"=>"en", "lat"=>43.824940, "lon"=>-96.712474),
"/US/US-SD/delmont-19436" => array("type"=>"city", "pop"=>215, "name_en"=>"Delmont", "lang"=>"en", "lat"=>43.265971, "lon"=>-98.159858),
"/US/US-SD/dimock-19438" => array("type"=>"town", "pop"=>133, "name_en"=>"Dimock", "lang"=>"en", "lat"=>43.475992, "lon"=>-97.987705),
"/US/US-SD/doland-19439" => array("type"=>"city", "pop"=>256, "name_en"=>"Doland", "lang"=>"en", "lat"=>44.894432, "lon"=>-98.099154),
"/US/US-SD/dolton-19440" => array("type"=>"town", "pop"=>38, "name_ru"=>"Долтон", "name_en"=>"Dolton", "lang"=>"en", "lat"=>43.491503, "lon"=>-97.384765),
"/US/US-SD/draper-19441" => array("type"=>"town", "pop"=>80, "name_en"=>"Draper", "lang"=>"en", "lat"=>43.925853, "lon"=>-100.537696),
"/US/US-SD/dupree-19442" => array("type"=>"city", "pop"=>434, "name_en"=>"Dupree", "lang"=>"en", "lat"=>45.049095, "lon"=>-101.600634),
"/US/US-SD/eagle-butte-19443" => array("type"=>"city", "pop"=>925, "name_en"=>"Eagle Butte", "lang"=>"en", "lat"=>44.990363, "lon"=>-101.229143),
"/US/US-SD/eden-19444" => array("type"=>"town", "pop"=>92, "name_ru"=>"Иден", "name_en"=>"Eden", "lang"=>"en", "lat"=>45.616244, "lon"=>-97.420055),
"/US/US-SD/edgemont-19445" => array("type"=>"city", "pop"=>788, "name_en"=>"Edgemont", "lang"=>"en", "lat"=>43.298486, "lon"=>-103.827075),
"/US/US-SD/egan-19446" => array("type"=>"city", "pop"=>245, "name_en"=>"Egan", "lang"=>"en", "lat"=>43.999676, "lon"=>-96.650493),
"/US/US-SD/elk-point-19447" => array("type"=>"city", "pop"=>1939, "name_en"=>"Elk Point", "lang"=>"en", "lat"=>42.682669, "lon"=>-96.680750),
"/US/US-SD/elkton-19448" => array("type"=>"city", "pop"=>644, "name_ru"=>"Элктон", "name_en"=>"Elkton", "lang"=>"en", "lat"=>44.234550, "lon"=>-96.479833),
"/US/US-SD/ellsworth-afb-19449" => array("type"=>"cdp", "pop"=>7017, "name_en"=>"Ellsworth AFB", "lang"=>"en", "lat"=>44.143736, "lon"=>-103.075564),
"/US/US-SD/emery-19450" => array("type"=>"city", "pop"=>506, "name_en"=>"Emery", "lang"=>"en", "lat"=>43.602659, "lon"=>-97.620206),
"/US/US-SD/erwin-19451" => array("type"=>"town", "pop"=>53, "name_en"=>"Erwin", "lang"=>"en", "lat"=>44.488237, "lon"=>-97.441030),
"/US/US-SD/estelline-19452" => array("type"=>"city", "pop"=>666, "name_en"=>"Estelline", "lang"=>"en", "lat"=>44.576248, "lon"=>-96.900130),
"/US/US-SD/ethan-19453" => array("type"=>"town", "pop"=>312, "name_en"=>"Ethan", "lang"=>"en", "lat"=>43.545894, "lon"=>-97.982933),
"/US/US-SD/eureka-19454" => array("type"=>"city", "pop"=>937, "name_ru"=>"Юрика", "name_en"=>"Eureka", "lang"=>"en", "lat"=>45.768709, "lon"=>-99.622962),
"/US/US-SD/fairburn-19455" => array("type"=>"town", "pop"=>75, "name_ru"=>"Фэрберн", "name_en"=>"Fairburn", "lang"=>"en", "lat"=>43.686749, "lon"=>-103.208206),
"/US/US-SD/fairfax-19456" => array("type"=>"town", "pop"=>105, "name_ru"=>"Фэрфакс", "name_en"=>"Fairfax", "lang"=>"en", "lat"=>43.028072, "lon"=>-98.888951),
"/US/US-SD/fairview-19457" => array("type"=>"town", "pop"=>109, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>43.222141, "lon"=>-96.489011),
"/US/US-SD/faith-19458" => array("type"=>"city", "pop"=>441, "name_en"=>"Faith", "lang"=>"en", "lat"=>45.025880, "lon"=>-102.036429),
"/US/US-SD/farmer-19459" => array("type"=>"town", "pop"=>21, "name_en"=>"Farmer", "lang"=>"en", "lat"=>43.724768, "lon"=>-97.688145),
"/US/US-SD/faulkton-19460" => array("type"=>"city", "pop"=>666, "name_en"=>"Faulkton", "lang"=>"en", "lat"=>45.034275, "lon"=>-99.127699),
"/US/US-SD/flandreau-19461" => array("type"=>"city", "pop"=>2222, "name_en"=>"Flandreau", "lang"=>"en", "lat"=>44.046912, "lon"=>-96.597747),
"/US/US-SD/florence-19462" => array("type"=>"town", "pop"=>290, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>45.055568, "lon"=>-97.325907),
"/US/US-SD/fort-pierre-19463" => array("type"=>"city", "pop"=>2010, "name_en"=>"Fort Pierre", "lang"=>"en", "lat"=>44.362744, "lon"=>-100.379346),
"/US/US-SD/fort-thompson-19464" => array("type"=>"cdp", "pop"=>1088, "name_en"=>"Fort Thompson", "lang"=>"en", "lat"=>44.049775, "lon"=>-99.411982),
"/US/US-SD/frankfort-19465" => array("type"=>"city", "pop"=>144, "name_ru"=>"Франкфорт", "name_en"=>"Frankfort", "lang"=>"en", "lat"=>44.877406, "lon"=>-98.308568),
"/US/US-SD/frederick-19466" => array("type"=>"town", "pop"=>242, "name_ru"=>"Фредерик", "name_en"=>"Frederick", "lang"=>"en", "lat"=>45.832838, "lon"=>-98.506650),
"/US/US-SD/freeman-19467" => array("type"=>"city", "pop"=>1194, "name_ru"=>"Фримэн", "name_en"=>"Freeman", "lang"=>"en", "lat"=>43.350877, "lon"=>-97.432359),
"/US/US-SD/fruitdale-19468" => array("type"=>"town", "pop"=>62, "name_en"=>"Fruitdale", "lang"=>"en", "lat"=>44.668229, "lon"=>-103.695512),
"/US/US-SD/fulton-19469" => array("type"=>"town", "pop"=>99, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>43.728233, "lon"=>-97.822290),
"/US/US-SD/garden-city-19470" => array("type"=>"town", "pop"=>58, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>44.957908, "lon"=>-97.580714),
"/US/US-SD/garretson-19471" => array("type"=>"city", "pop"=>1198, "name_en"=>"Garretson", "lang"=>"en", "lat"=>43.716805, "lon"=>-96.500280),
"/US/US-SD/gary-19472" => array("type"=>"city", "pop"=>206, "name_ru"=>"Гэри", "name_en"=>"Gary", "lang"=>"en", "lat"=>44.795186, "lon"=>-96.457969),
"/US/US-SD/gayville-19473" => array("type"=>"town", "pop"=>401, "name_en"=>"Gayville", "lang"=>"en", "lat"=>42.888221, "lon"=>-97.172762),
"/US/US-SD/geddes-19474" => array("type"=>"city", "pop"=>238, "name_en"=>"Geddes", "lang"=>"en", "lat"=>43.253987, "lon"=>-98.697332),
"/US/US-SD/gettysburg-19475" => array("type"=>"city", "pop"=>1058, "name_en"=>"Gettysburg", "lang"=>"en", "lat"=>45.006535, "lon"=>-99.953490),
"/US/US-SD/glenham-19476" => array("type"=>"town", "pop"=>118, "name_en"=>"Glenham", "lang"=>"en", "lat"=>45.533317, "lon"=>-100.270328),
"/US/US-SD/goodwin-19477" => array("type"=>"town", "pop"=>148, "name_en"=>"Goodwin", "lang"=>"en", "lat"=>44.877191, "lon"=>-96.848520),
"/US/US-SD/gregory-19478" => array("type"=>"city", "pop"=>1182, "name_en"=>"Gregory", "lang"=>"en", "lat"=>43.231762, "lon"=>-99.425120),
"/US/US-SD/grenville-19479" => array("type"=>"town", "pop"=>54, "name_en"=>"Grenville", "lang"=>"en", "lat"=>45.466435, "lon"=>-97.389671),
"/US/US-SD/groton-19480" => array("type"=>"city", "pop"=>1408, "name_ru"=>"Гротон", "name_en"=>"Groton", "lang"=>"en", "lat"=>45.452063, "lon"=>-98.100114),
"/US/US-SD/harrisburg-19481" => array("type"=>"city", "pop"=>4270, "name_ru"=>"Харрисбург", "name_en"=>"Harrisburg", "lang"=>"en", "lat"=>43.430292, "lon"=>-96.695797),
"/US/US-SD/harrold-19482" => array("type"=>"town", "pop"=>200, "name_en"=>"Harrold", "lang"=>"en", "lat"=>44.523279, "lon"=>-99.738436),
"/US/US-SD/hartford-19483" => array("type"=>"city", "pop"=>2436, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>43.623516, "lon"=>-96.944758),
"/US/US-SD/hayti-19484" => array("type"=>"town", "pop"=>355, "name_ru"=>"Хейти", "name_en"=>"Hayti", "lang"=>"en", "lat"=>44.656562, "lon"=>-97.204041),
"/US/US-SD/hazel-19485" => array("type"=>"town", "pop"=>106, "name_ru"=>"Хейзел", "name_en"=>"Hazel", "lang"=>"en", "lat"=>44.757951, "lon"=>-97.380146),
"/US/US-SD/hecla-19486" => array("type"=>"city", "pop"=>297, "name_en"=>"Hecla", "lang"=>"en", "lat"=>45.882052, "lon"=>-98.151430),
"/US/US-SD/henry-19487" => array("type"=>"town", "pop"=>252, "name_ru"=>"Хенри", "name_en"=>"Henry", "lang"=>"en", "lat"=>44.880954, "lon"=>-97.462596),
"/US/US-SD/hermosa-19488" => array("type"=>"town", "pop"=>350, "name_en"=>"Hermosa", "lang"=>"en", "lat"=>43.840023, "lon"=>-103.191538),
"/US/US-SD/herreid-19489" => array("type"=>"city", "pop"=>360, "name_en"=>"Herreid", "lang"=>"en", "lat"=>45.836942, "lon"=>-100.075190),
"/US/US-SD/herrick-19490" => array("type"=>"town", "pop"=>90, "name_ru"=>"Херрик", "name_en"=>"Herrick", "lang"=>"en", "lat"=>43.115376, "lon"=>-99.187186),
"/US/US-SD/hetland-19491" => array("type"=>"town", "pop"=>39, "name_en"=>"Hetland", "lang"=>"en", "lat"=>44.376487, "lon"=>-97.234497),
"/US/US-SD/highmore-19492" => array("type"=>"city", "pop"=>714, "name_en"=>"Highmore", "lang"=>"en", "lat"=>44.520740, "lon"=>-99.439033),
"/US/US-SD/hill-city-19493" => array("type"=>"city", "pop"=>908, "name_ru"=>"Хилл Сити", "name_en"=>"Hill City", "lang"=>"en", "lat"=>43.932068, "lon"=>-103.571438),
"/US/US-SD/hillsview-19494" => array("type"=>"town", "pop"=>3, "name_en"=>"Hillsview", "lang"=>"en", "lat"=>45.665158, "lon"=>-99.559499),
"/US/US-SD/hitchcock-19495" => array("type"=>"town", "pop"=>99, "name_en"=>"Hitchcock", "lang"=>"en", "lat"=>44.629308, "lon"=>-98.407801),
"/US/US-SD/hosmer-19496" => array("type"=>"city", "pop"=>251, "name_en"=>"Hosmer", "lang"=>"en", "lat"=>45.578866, "lon"=>-99.473739),
"/US/US-SD/hot-springs-19497" => array("type"=>"city", "pop"=>4028, "name_ru"=>"Хот Спрингс", "name_en"=>"Hot Springs", "lang"=>"en", "lat"=>43.433035, "lon"=>-103.480282),
"/US/US-SD/hoven-19498" => array("type"=>"town", "pop"=>395, "name_en"=>"Hoven", "lang"=>"en", "lat"=>45.241824, "lon"=>-99.776353),
"/US/US-SD/howard-19499" => array("type"=>"city", "pop"=>891, "name_ru"=>"Ховарт", "name_en"=>"Howard", "lang"=>"en", "lat"=>44.011863, "lon"=>-97.523272),
"/US/US-SD/hudson-19500" => array("type"=>"town", "pop"=>445, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>43.129828, "lon"=>-96.455541),
"/US/US-SD/humboldt-19501" => array("type"=>"town", "pop"=>589, "name_ru"=>"Гумбольдт", "name_en"=>"Humboldt", "lang"=>"en", "lat"=>43.644956, "lon"=>-97.073985),
"/US/US-SD/hurley-19502" => array("type"=>"city", "pop"=>385, "name_ru"=>"Херли", "name_en"=>"Hurley", "lang"=>"en", "lat"=>43.279829, "lon"=>-97.090266),
"/US/US-SD/huron-19503" => array("type"=>"city", "pop"=>11033, "name_ru"=>"Гурон", "name_en"=>"Huron", "lang"=>"en", "lat"=>44.363895, "lon"=>-98.217835),
"/US/US-SD/interior-19504" => array("type"=>"town", "pop"=>73, "name_en"=>"Interior", "lang"=>"en", "lat"=>43.727445, "lon"=>-101.983426),
"/US/US-SD/ipswich-19505" => array("type"=>"city", "pop"=>866, "name_ru"=>"Ипсвич", "name_en"=>"Ipswich", "lang"=>"en", "lat"=>45.444051, "lon"=>-99.029867),
"/US/US-SD/irene-19506" => array("type"=>"city", "pop"=>404, "name_en"=>"Irene", "lang"=>"en", "lat"=>43.083664, "lon"=>-97.158082),
"/US/US-SD/iroquois-19507" => array("type"=>"city", "pop"=>268, "name_ru"=>"Ирокез", "name_en"=>"Iroquois", "lang"=>"en", "lat"=>44.366462, "lon"=>-97.849734),
"/US/US-SD/isabel-19508" => array("type"=>"city", "pop"=>231, "name_ru"=>"Изабел", "name_en"=>"Isabel", "lang"=>"en", "lat"=>45.395764, "lon"=>-101.428529),
"/US/US-SD/java-19509" => array("type"=>"city", "pop"=>169, "name_en"=>"Java", "lang"=>"en", "lat"=>45.503576, "lon"=>-99.884234),
"/US/US-SD/jefferson-19510" => array("type"=>"city", "pop"=>599, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>42.603542, "lon"=>-96.561400),
"/US/US-SD/kadoka-19511" => array("type"=>"city", "pop"=>635, "name_en"=>"Kadoka", "lang"=>"en", "lat"=>43.831989, "lon"=>-101.508757),
"/US/US-SD/kennebec-19512" => array("type"=>"town", "pop"=>277, "name_en"=>"Kennebec", "lang"=>"en", "lat"=>43.904294, "lon"=>-99.862407),
"/US/US-SD/keystone-19513" => array("type"=>"town", "pop"=>316, "name_ru"=>"Кистоун", "name_en"=>"Keystone", "lang"=>"en", "lat"=>43.894349, "lon"=>-103.428874),
"/US/US-SD/kimball-19514" => array("type"=>"city", "pop"=>692, "name_en"=>"Kimball", "lang"=>"en", "lat"=>43.746155, "lon"=>-98.957128),
"/US/US-SD/kranzburg-19515" => array("type"=>"town", "pop"=>180, "name_en"=>"Kranzburg", "lang"=>"en", "lat"=>44.886523, "lon"=>-96.909777),
"/US/US-SD/kyle-19516" => array("type"=>"cdp", "pop"=>914, "name_en"=>"Kyle", "lang"=>"en", "lat"=>43.427439, "lon"=>-102.164547),
"/US/US-SD/la-bolt-19517" => array("type"=>"town", "pop"=>76, "name_en"=>"La Bolt", "lang"=>"en", "lat"=>45.050025, "lon"=>-96.674215),
"/US/US-SD/lake-andes-19518" => array("type"=>"city", "pop"=>756, "name_en"=>"Lake Andes", "lang"=>"en", "lat"=>43.155675, "lon"=>-98.535024),
"/US/US-SD/lake-city-19519" => array("type"=>"town", "pop"=>44, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>45.723836, "lon"=>-97.413563),
"/US/US-SD/lake-norden-19520" => array("type"=>"city", "pop"=>422, "name_en"=>"Lake Norden", "lang"=>"en", "lat"=>44.579127, "lon"=>-97.207996),
"/US/US-SD/lake-preston-19521" => array("type"=>"city", "pop"=>653, "name_en"=>"Lake Preston", "lang"=>"en", "lat"=>44.361770, "lon"=>-97.376204),
"/US/US-SD/lane-19522" => array("type"=>"town", "pop"=>50, "name_ru"=>"Лейн", "name_en"=>"Lane", "lang"=>"en", "lat"=>44.069159, "lon"=>-98.424099),
"/US/US-SD/langford-19523" => array("type"=>"town", "pop"=>284, "name_en"=>"Langford", "lang"=>"en", "lat"=>45.602316, "lon"=>-97.829489),
"/US/US-SD/lead-19524" => array("type"=>"city", "pop"=>2892, "name_en"=>"Lead", "lang"=>"en", "lat"=>44.352131, "lon"=>-103.766931),
"/US/US-SD/lebanon-19525" => array("type"=>"town", "pop"=>68, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>45.068911, "lon"=>-99.765884),
"/US/US-SD/lemmon-19526" => array("type"=>"city", "pop"=>1185, "name_en"=>"Lemmon", "lang"=>"en", "lat"=>45.938392, "lon"=>-102.158176),
"/US/US-SD/lennox-19527" => array("type"=>"city", "pop"=>2417, "name_ru"=>"Леннокс", "name_en"=>"Lennox", "lang"=>"en", "lat"=>43.352829, "lon"=>-96.896400),
"/US/US-SD/leola-19528" => array("type"=>"city", "pop"=>390, "name_ru"=>"Леола", "name_en"=>"Leola", "lang"=>"en", "lat"=>45.721262, "lon"=>-98.938119),
"/US/US-SD/lesterville-19529" => array("type"=>"town", "pop"=>151, "name_en"=>"Lesterville", "lang"=>"en", "lat"=>43.038684, "lon"=>-97.590512),
"/US/US-SD/letcher-19530" => array("type"=>"town", "pop"=>158, "name_en"=>"Letcher", "lang"=>"en", "lat"=>43.897763, "lon"=>-98.143455),
"/US/US-SD/lily-19531" => array("type"=>"town", "pop"=>18, "name_en"=>"Lily", "lang"=>"en", "lat"=>45.181375, "lon"=>-97.681178),
"/US/US-SD/little-eagle-19532" => array("type"=>"cdp", "pop"=>294, "name_en"=>"Little Eagle", "lang"=>"en", "lat"=>45.681588, "lon"=>-100.796252),
"/US/US-SD/long-lake-19533" => array("type"=>"town", "pop"=>49, "name_ru"=>"Лонг Лейк", "name_en"=>"Long Lake", "lang"=>"en", "lat"=>45.856284, "lon"=>-99.206181),
"/US/US-SD/lower-brule-19534" => array("type"=>"cdp", "pop"=>1268, "name_en"=>"Lower Brule", "lang"=>"en", "lat"=>44.073987, "lon"=>-99.582489),
"/US/US-SD/lowry-19535" => array("type"=>"town", "pop"=>9, "name_ru"=>"Лоури", "name_en"=>"Lowry", "lang"=>"en", "lat"=>45.315435, "lon"=>-99.982006),
"/US/US-SD/madison-19538" => array("type"=>"city", "pop"=>6482, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>44.008417, "lon"=>-97.107377),
"/US/US-SD/manderson-white-horse-creek-19539" => array("type"=>"cdp", "pop"=>243, "name_en"=>"Manderson-White Horse Creek", "lang"=>"en", "lat"=>43.229524, "lon"=>-102.470606),
"/US/US-SD/marion-19540" => array("type"=>"city", "pop"=>825, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>43.423990, "lon"=>-97.260627),
"/US/US-SD/martin-19541" => array("type"=>"city", "pop"=>995, "name_ru"=>"Мартин", "name_en"=>"Martin", "lang"=>"en", "lat"=>43.174870, "lon"=>-101.723883),
"/US/US-SD/marty-19542" => array("type"=>"cdp", "pop"=>436, "name_en"=>"Marty", "lang"=>"en", "lat"=>42.992904, "lon"=>-98.429263),
"/US/US-SD/marvin-19543" => array("type"=>"town", "pop"=>61, "name_en"=>"Marvin", "lang"=>"en", "lat"=>45.260877, "lon"=>-96.912531),
"/US/US-SD/mcintosh-19536" => array("type"=>"city", "pop"=>205, "name_ru"=>"Макинтош", "name_en"=>"McIntosh", "lang"=>"en", "lat"=>45.920623, "lon"=>-101.349650),
"/US/US-SD/mclaughlin-19537" => array("type"=>"city", "pop"=>730, "name_en"=>"McLaughlin", "lang"=>"en", "lat"=>45.813031, "lon"=>-100.811036),
"/US/US-SD/mellette-19544" => array("type"=>"city", "pop"=>215, "name_en"=>"Mellette", "lang"=>"en", "lat"=>45.153995, "lon"=>-98.497965),
"/US/US-SD/menno-19545" => array("type"=>"city", "pop"=>655, "name_en"=>"Menno", "lang"=>"en", "lat"=>43.237850, "lon"=>-97.577096),
"/US/US-SD/midland-19546" => array("type"=>"town", "pop"=>148, "name_ru"=>"Мидленд", "name_en"=>"Midland", "lang"=>"en", "lat"=>44.070960, "lon"=>-101.154516),
"/US/US-SD/milbank-19547" => array("type"=>"city", "pop"=>3203, "name_en"=>"Milbank", "lang"=>"en", "lat"=>45.219580, "lon"=>-96.634845),
"/US/US-SD/miller-19548" => array("type"=>"city", "pop"=>1341, "name_ru"=>"Миллер", "name_en"=>"Miller", "lang"=>"en", "lat"=>44.519951, "lon"=>-98.987662),
"/US/US-SD/mission-19549" => array("type"=>"city", "pop"=>1223, "name_ru"=>"Мишен", "name_en"=>"Mission", "lang"=>"en", "lat"=>43.306466, "lon"=>-100.660453),
"/US/US-SD/mission-hill-19550" => array("type"=>"town", "pop"=>168, "name_en"=>"Mission Hill", "lang"=>"en", "lat"=>42.921160, "lon"=>-97.279118),
"/US/US-SD/mitchell-19551" => array("type"=>"city", "pop"=>14752, "name_ru"=>"Митчелл", "name_en"=>"Mitchell", "lang"=>"en", "lat"=>43.732610, "lon"=>-98.033846),
"/US/US-SD/mobridge-19552" => array("type"=>"city", "pop"=>3154, "name_en"=>"Mobridge", "lang"=>"en", "lat"=>45.540687, "lon"=>-100.435766),
"/US/US-SD/monroe-19553" => array("type"=>"town", "pop"=>152, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>43.487052, "lon"=>-97.216030),
"/US/US-SD/montrose-19554" => array("type"=>"city", "pop"=>450, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>43.701479, "lon"=>-97.183043),
"/US/US-SD/morristown-19555" => array("type"=>"town", "pop"=>82, "name_ru"=>"Морристаун", "name_en"=>"Morristown", "lang"=>"en", "lat"=>45.938997, "lon"=>-101.718266),
"/US/US-SD/mound-city-19556" => array("type"=>"town", "pop"=>64, "name_ru"=>"Маунд Сити", "name_en"=>"Mound City", "lang"=>"en", "lat"=>45.726803, "lon"=>-100.067665),
"/US/US-SD/mount-vernon-19557" => array("type"=>"city", "pop"=>464, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>43.712170, "lon"=>-98.260905),
"/US/US-SD/murdo-19558" => array("type"=>"city", "pop"=>526, "name_en"=>"Murdo", "lang"=>"en", "lat"=>43.889201, "lon"=>-100.713587),
"/US/US-SD/naples-19559" => array("type"=>"town", "pop"=>20, "name_ru"=>"Нэплс", "name_en"=>"Naples", "lang"=>"en", "lat"=>44.771331, "lon"=>-97.512606),
"/US/US-SD/new-effington-19560" => array("type"=>"town", "pop"=>225, "name_en"=>"New Effington", "lang"=>"en", "lat"=>45.855208, "lon"=>-96.918054),
"/US/US-SD/new-underwood-19562" => array("type"=>"city", "pop"=>675, "name_en"=>"New Underwood", "lang"=>"en", "lat"=>44.093857, "lon"=>-102.833607),
"/US/US-SD/new-witten-19563" => array("type"=>"town", "pop"=>45, "name_en"=>"New Witten", "lang"=>"en", "lat"=>43.440316, "lon"=>-100.082788),
"/US/US-SD/newell-19561" => array("type"=>"city", "pop"=>627, "name_ru"=>"Ньюуэлл", "name_en"=>"Newell", "lang"=>"en", "lat"=>44.715272, "lon"=>-103.417755),
"/US/US-SD/nisland-19564" => array("type"=>"town", "pop"=>206, "name_en"=>"Nisland", "lang"=>"en", "lat"=>44.673502, "lon"=>-103.552480),
"/US/US-SD/north-eagle-butte-19565" => array("type"=>"cdp", "pop"=>1423, "name_en"=>"North Eagle Butte", "lang"=>"en", "lat"=>45.001324, "lon"=>-101.227081),
"/US/US-SD/north-sioux-city-19566" => array("type"=>"city", "pop"=>2520, "name_en"=>"North Sioux City", "lang"=>"en", "lat"=>42.536579, "lon"=>-96.502134),
"/US/US-SD/north-spearfish-19567" => array("type"=>"cdp", "pop"=>2274, "name_en"=>"North Spearfish", "lang"=>"en", "lat"=>44.506706, "lon"=>-103.891778),
"/US/US-SD/northville-19568" => array("type"=>"town", "pop"=>108, "name_ru"=>"Нортвилл", "name_en"=>"Northville", "lang"=>"en", "lat"=>45.154854, "lon"=>-98.578535),
"/US/US-SD/nunda-19569" => array("type"=>"town", "pop"=>47, "name_en"=>"Nunda", "lang"=>"en", "lat"=>44.160985, "lon"=>-97.017882),
"/US/US-SD/oacoma-19570" => array("type"=>"town", "pop"=>406, "name_en"=>"Oacoma", "lang"=>"en", "lat"=>43.799225, "lon"=>-99.384408),
"/US/US-SD/oelrichs-19571" => array("type"=>"town", "pop"=>139, "name_en"=>"Oelrichs", "lang"=>"en", "lat"=>43.181483, "lon"=>-103.233126),
"/US/US-SD/oglala-19572" => array("type"=>"cdp", "pop"=>422, "name_en"=>"Oglala", "lang"=>"en", "lat"=>43.182821, "lon"=>-102.729636),
"/US/US-SD/oldham-19573" => array("type"=>"city", "pop"=>181, "name_en"=>"Oldham", "lang"=>"en", "lat"=>44.228248, "lon"=>-97.308634),
"/US/US-SD/olivet-19574" => array("type"=>"town", "pop"=>63, "name_ru"=>"Оливет", "name_en"=>"Olivet", "lang"=>"en", "lat"=>43.241670, "lon"=>-97.673500),
"/US/US-SD/onaka-19575" => array("type"=>"town", "pop"=>25, "name_en"=>"Onaka", "lang"=>"en", "lat"=>45.191229, "lon"=>-99.464579),
"/US/US-SD/onida-19576" => array("type"=>"city", "pop"=>619, "name_en"=>"Onida", "lang"=>"en", "lat"=>44.704699, "lon"=>-100.066491),
"/US/US-SD/orient-19577" => array("type"=>"town", "pop"=>47, "name_ru"=>"Ориент", "name_en"=>"Orient", "lang"=>"en", "lat"=>44.902862, "lon"=>-99.088332),
"/US/US-SD/ortley-19578" => array("type"=>"town", "pop"=>53, "name_en"=>"Ortley", "lang"=>"en", "lat"=>45.335211, "lon"=>-97.204020),
"/US/US-SD/parker-19579" => array("type"=>"city", "pop"=>983, "name_ru"=>"Паркер", "name_en"=>"Parker", "lang"=>"en", "lat"=>43.396229, "lon"=>-97.138437),
"/US/US-SD/parkston-19580" => array("type"=>"city", "pop"=>1486, "name_en"=>"Parkston", "lang"=>"en", "lat"=>43.393300, "lon"=>-97.985551),
"/US/US-SD/parmelee-19581" => array("type"=>"cdp", "pop"=>618, "name_en"=>"Parmelee", "lang"=>"en", "lat"=>43.324427, "lon"=>-101.036403),
"/US/US-SD/peever-19582" => array("type"=>"town", "pop"=>198, "name_en"=>"Peever", "lang"=>"en", "lat"=>45.542494, "lon"=>-96.955743),
"/US/US-SD/philip-19583" => array("type"=>"city", "pop"=>731, "name_en"=>"Philip", "lang"=>"en", "lat"=>44.040540, "lon"=>-101.663837),
"/US/US-SD/pickstown-19584" => array("type"=>"town", "pop"=>194, "name_en"=>"Pickstown", "lang"=>"en", "lat"=>43.066935, "lon"=>-98.530488),
"/US/US-SD/pierpont-19585" => array("type"=>"town", "pop"=>102, "name_en"=>"Pierpont", "lang"=>"en", "lat"=>45.496931, "lon"=>-97.831130),
"/US/US-SD/pierre-19586" => array("type"=>"city", "pop"=>13899, "name_en"=>"Pierre", "lang"=>"en", "lat"=>44.372982, "lon"=>-100.322483),
"/US/US-SD/pine-ridge-19587" => array("type"=>"cdp", "pop"=>2596, "name_ru"=>"Пайн Ридж", "name_en"=>"Pine Ridge", "lang"=>"en", "lat"=>43.027776, "lon"=>-102.558412),
"/US/US-SD/plankinton-19588" => array("type"=>"city", "pop"=>569, "name_en"=>"Plankinton", "lang"=>"en", "lat"=>43.715672, "lon"=>-98.484139),
"/US/US-SD/platte-19589" => array("type"=>"city", "pop"=>1284, "name_en"=>"Platte", "lang"=>"en", "lat"=>43.387382, "lon"=>-98.843884),
"/US/US-SD/pollock-19590" => array("type"=>"town", "pop"=>249, "name_ru"=>"Поллок", "name_en"=>"Pollock", "lang"=>"en", "lat"=>45.901491, "lon"=>-100.290169),
"/US/US-SD/porcupine-19591" => array("type"=>"cdp", "pop"=>783, "name_en"=>"Porcupine", "lang"=>"en", "lat"=>43.271967, "lon"=>-102.332889),
"/US/US-SD/presho-19592" => array("type"=>"city", "pop"=>625, "name_en"=>"Presho", "lang"=>"en", "lat"=>43.907126, "lon"=>-100.057683),
"/US/US-SD/pringle-19593" => array("type"=>"town", "pop"=>112, "name_en"=>"Pringle", "lang"=>"en", "lat"=>43.609151, "lon"=>-103.593711),
"/US/US-SD/pukwana-19594" => array("type"=>"town", "pop"=>280, "name_en"=>"Pukwana", "lang"=>"en", "lat"=>43.778944, "lon"=>-99.182369),
"/US/US-SD/quinn-19595" => array("type"=>"town", "pop"=>45, "name_en"=>"Quinn", "lang"=>"en", "lat"=>43.988591, "lon"=>-102.128055),
"/US/US-SD/ramona-19596" => array("type"=>"town", "pop"=>191, "name_ru"=>"Рамона", "name_en"=>"Ramona", "lang"=>"en", "lat"=>44.120084, "lon"=>-97.215041),
"/US/US-SD/rapid-city-19597" => array("type"=>"city", "pop"=>65491, "name_en"=>"Rapid City", "lang"=>"en", "lat"=>44.075201, "lon"=>-103.232802),
"/US/US-SD/rapid-valley-19598" => array("type"=>"cdp", "pop"=>5968, "name_en"=>"Rapid Valley", "lang"=>"en", "lat"=>44.072547, "lon"=>-103.129606),
"/US/US-SD/ravinia-19599" => array("type"=>"town", "pop"=>76, "name_en"=>"Ravinia", "lang"=>"en", "lat"=>43.136009, "lon"=>-98.427131),
"/US/US-SD/raymond-19600" => array("type"=>"town", "pop"=>69, "name_ru"=>"Реймонд", "name_en"=>"Raymond", "lang"=>"en", "lat"=>44.911144, "lon"=>-97.936557),
"/US/US-SD/redfield-19601" => array("type"=>"city", "pop"=>2273, "name_ru"=>"Редфилд", "name_en"=>"Redfield", "lang"=>"en", "lat"=>44.873023, "lon"=>-98.520371),
"/US/US-SD/ree-heights-19602" => array("type"=>"town", "pop"=>74, "name_en"=>"Ree Heights", "lang"=>"en", "lat"=>44.516096, "lon"=>-99.200358),
"/US/US-SD/reliance-19603" => array("type"=>"town", "pop"=>206, "name_en"=>"Reliance", "lang"=>"en", "lat"=>43.880361, "lon"=>-99.602271),
"/US/US-SD/revillo-19604" => array("type"=>"town", "pop"=>132, "name_en"=>"Revillo", "lang"=>"en", "lat"=>45.014568, "lon"=>-96.570317),
"/US/US-SD/rockham-19605" => array("type"=>"town", "pop"=>45, "name_en"=>"Rockham", "lang"=>"en", "lat"=>44.905040, "lon"=>-98.822867),
"/US/US-SD/roscoe-19606" => array("type"=>"city", "pop"=>287, "name_ru"=>"Роско", "name_en"=>"Roscoe", "lang"=>"en", "lat"=>45.450318, "lon"=>-99.334903),
"/US/US-SD/rosebud-19607" => array("type"=>"cdp", "pop"=>42, "name_ru"=>"Роузбад", "name_en"=>"Rosebud", "lang"=>"en", "lat"=>43.235388, "lon"=>-100.838837),
"/US/US-SD/rosholt-19608" => array("type"=>"town", "pop"=>426, "name_en"=>"Rosholt", "lang"=>"en", "lat"=>45.866145, "lon"=>-96.731704),
"/US/US-SD/roslyn-19609" => array("type"=>"town", "pop"=>193, "name_en"=>"Roslyn", "lang"=>"en", "lat"=>45.496643, "lon"=>-97.493396),
"/US/US-SD/roswell-19610" => array("type"=>"town", "pop"=>18, "name_ru"=>"Росуэл", "name_en"=>"Roswell", "lang"=>"en", "lat"=>43.999069, "lon"=>-97.701001),
"/US/US-SD/salem-19613" => array("type"=>"city", "pop"=>1321, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>43.724264, "lon"=>-97.388630),
"/US/US-SD/scotland-19614" => array("type"=>"city", "pop"=>790, "name_ru"=>"Скотленд", "name_en"=>"Scotland", "lang"=>"en", "lat"=>43.148109, "lon"=>-97.719498),
"/US/US-SD/selby-19615" => array("type"=>"city", "pop"=>647, "name_en"=>"Selby", "lang"=>"en", "lat"=>45.504743, "lon"=>-100.032402),
"/US/US-SD/seneca-19616" => array("type"=>"town", "pop"=>49, "name_ru"=>"Сенека", "name_en"=>"Seneca", "lang"=>"en", "lat"=>45.060973, "lon"=>-99.508076),
"/US/US-SD/sherman-19617" => array("type"=>"town", "pop"=>93, "name_ru"=>"Шерман", "name_en"=>"Sherman", "lang"=>"en", "lat"=>43.757697, "lon"=>-96.475376),
"/US/US-SD/sinai-19618" => array("type"=>"town", "pop"=>120, "name_en"=>"Sinai", "lang"=>"en", "lat"=>44.245069, "lon"=>-97.043041),
"/US/US-SD/sioux-falls-19619" => array("type"=>"city", "pop"=>154997, "name_en"=>"Sioux Falls", "lang"=>"en", "lat"=>43.544201, "lon"=>-96.730100),
"/US/US-SD/sisseton-19620" => array("type"=>"city", "pop"=>2447, "name_en"=>"Sisseton", "lang"=>"en", "lat"=>45.662408, "lon"=>-97.046073),
"/US/US-SD/south-shore-19621" => array("type"=>"town", "pop"=>254, "name_ru"=>"Саут Шор", "name_en"=>"South Shore", "lang"=>"en", "lat"=>45.101700, "lon"=>-96.930027),
"/US/US-SD/spearfish-19622" => array("type"=>"city", "pop"=>10010, "name_en"=>"Spearfish", "lang"=>"en", "lat"=>44.488761, "lon"=>-103.847107),
"/US/US-SD/spencer-19623" => array("type"=>"city", "pop"=>151, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>43.727874, "lon"=>-97.591688),
"/US/US-SD/spring-creek-19624" => array("type"=>"cdp", "pop"=>144, "name_en"=>"Spring Creek", "lang"=>"en", "lat"=>43.120531, "lon"=>-101.027389),
"/US/US-SD/springfield-19625" => array("type"=>"city", "pop"=>1487, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>42.854729, "lon"=>-97.892961),
"/US/US-SD/st.-francis-19611" => array("type"=>"town", "pop"=>767, "name_ru"=>"Сент-Фрэнсис", "name_en"=>"St. Francis", "lang"=>"en", "lat"=>43.142344, "lon"=>-100.902041),
"/US/US-SD/st.-lawrence-19612" => array("type"=>"town", "pop"=>177, "name_en"=>"St. Lawrence", "lang"=>"en", "lat"=>44.516966, "lon"=>-98.939968),
"/US/US-SD/stickney-19626" => array("type"=>"town", "pop"=>298, "name_ru"=>"Стикни", "name_en"=>"Stickney", "lang"=>"en", "lat"=>43.589561, "lon"=>-98.437546),
"/US/US-SD/stockholm-19627" => array("type"=>"town", "pop"=>93, "name_en"=>"Stockholm", "lang"=>"en", "lat"=>45.102452, "lon"=>-96.798415),
"/US/US-SD/strandburg-19628" => array("type"=>"town", "pop"=>60, "name_en"=>"Strandburg", "lang"=>"en", "lat"=>45.043347, "lon"=>-96.761024),
"/US/US-SD/stratford-19629" => array("type"=>"town", "pop"=>92, "name_ru"=>"Стрэтфорд", "name_en"=>"Stratford", "lang"=>"en", "lat"=>45.317135, "lon"=>-98.303814),
"/US/US-SD/sturgis-19630" => array("type"=>"city", "pop"=>5950, "name_ru"=>"Стерджис", "name_en"=>"Sturgis", "lang"=>"en", "lat"=>44.411613, "lon"=>-103.514593),
"/US/US-SD/summit-19631" => array("type"=>"town", "pop"=>267, "name_ru"=>"Саммит", "name_en"=>"Summit", "lang"=>"en", "lat"=>45.303782, "lon"=>-97.036751),
"/US/US-SD/tabor-19632" => array("type"=>"town", "pop"=>365, "name_ru"=>"Табор", "name_en"=>"Tabor", "lang"=>"en", "lat"=>42.947657, "lon"=>-97.659684),
"/US/US-SD/tea-19633" => array("type"=>"city", "pop"=>3761, "name_en"=>"Tea", "lang"=>"en", "lat"=>43.448078, "lon"=>-96.837258),
"/US/US-SD/timber-lake-19634" => array("type"=>"city", "pop"=>421, "name_en"=>"Timber Lake", "lang"=>"en", "lat"=>45.427619, "lon"=>-101.074175),
"/US/US-SD/tolstoy-19635" => array("type"=>"town", "pop"=>50, "name_en"=>"Tolstoy", "lang"=>"en", "lat"=>45.206756, "lon"=>-99.614739),
"/US/US-SD/toronto-19636" => array("type"=>"town", "pop"=>175, "name_ru"=>"Торонто", "name_en"=>"Toronto", "lang"=>"en", "lat"=>44.572370, "lon"=>-96.641566),
"/US/US-SD/trent-19637" => array("type"=>"town", "pop"=>237, "name_en"=>"Trent", "lang"=>"en", "lat"=>43.906858, "lon"=>-96.657282),
"/US/US-SD/tripp-19638" => array("type"=>"city", "pop"=>631, "name_en"=>"Tripp", "lang"=>"en", "lat"=>43.224964, "lon"=>-97.966426),
"/US/US-SD/tulare-19639" => array("type"=>"town", "pop"=>192, "name_ru"=>"Тулар", "name_en"=>"Tulare", "lang"=>"en", "lat"=>44.738815, "lon"=>-98.508692),
"/US/US-SD/turton-19640" => array("type"=>"town", "pop"=>53, "name_en"=>"Turton", "lang"=>"en", "lat"=>45.048966, "lon"=>-98.096469),
"/US/US-SD/twin-brooks-19641" => array("type"=>"town", "pop"=>49, "name_en"=>"Twin Brooks", "lang"=>"en", "lat"=>45.207726, "lon"=>-96.786913),
"/US/US-SD/two-strike-19642" => array("type"=>"cdp", "pop"=>112, "name_en"=>"Two Strike", "lang"=>"en", "lat"=>43.213041, "lon"=>-100.875213),
"/US/US-SD/tyndall-19643" => array("type"=>"city", "pop"=>1102, "name_en"=>"Tyndall", "lang"=>"en", "lat"=>42.989732, "lon"=>-97.864601),
"/US/US-SD/utica-19644" => array("type"=>"town", "pop"=>82, "name_ru"=>"Ютика", "name_en"=>"Utica", "lang"=>"en", "lat"=>42.981304, "lon"=>-97.496233),
"/US/US-SD/valley-springs-19645" => array("type"=>"city", "pop"=>857, "name_ru"=>"Валли Спрингс", "name_en"=>"Valley Springs", "lang"=>"en", "lat"=>43.583188, "lon"=>-96.464665),
"/US/US-SD/veblen-19646" => array("type"=>"city", "pop"=>254, "name_en"=>"Veblen", "lang"=>"en", "lat"=>45.861862, "lon"=>-97.286825),
"/US/US-SD/verdon-19647" => array("type"=>"town", "pop"=>6, "name_en"=>"Verdon", "lang"=>"en", "lat"=>45.244691, "lon"=>-98.098057),
"/US/US-SD/vermillion-19648" => array("type"=>"city", "pop"=>10495, "name_ru"=>"Вермиллион", "name_en"=>"Vermillion", "lang"=>"en", "lat"=>42.782001, "lon"=>-96.926574),
"/US/US-SD/viborg-19649" => array("type"=>"city", "pop"=>774, "name_en"=>"Viborg", "lang"=>"en", "lat"=>43.171160, "lon"=>-97.080469),
"/US/US-SD/vienna-19650" => array("type"=>"town", "pop"=>63, "name_ru"=>"Вьенна", "name_en"=>"Vienna", "lang"=>"en", "lat"=>44.703114, "lon"=>-97.499115),
"/US/US-SD/vilas-19651" => array("type"=>"town", "pop"=>16, "name_ru"=>"Вилас", "name_en"=>"Vilas", "lang"=>"en", "lat"=>44.008511, "lon"=>-97.595689),
"/US/US-SD/virgil-19652" => array("type"=>"town", "pop"=>23, "name_ru"=>"Верджил", "name_en"=>"Virgil", "lang"=>"en", "lat"=>44.290251, "lon"=>-98.426738),
"/US/US-SD/volga-19653" => array("type"=>"city", "pop"=>1583, "name_ru"=>"Волга", "name_en"=>"Volga", "lang"=>"en", "lat"=>44.323272, "lon"=>-96.922588),
"/US/US-SD/volin-19654" => array("type"=>"town", "pop"=>197, "name_en"=>"Volin", "lang"=>"en", "lat"=>42.959000, "lon"=>-97.180712),
"/US/US-SD/wagner-19655" => array("type"=>"city", "pop"=>1551, "name_en"=>"Wagner", "lang"=>"en", "lat"=>43.078071, "lon"=>-98.301552),
"/US/US-SD/wakonda-19656" => array("type"=>"town", "pop"=>350, "name_en"=>"Wakonda", "lang"=>"en", "lat"=>43.008123, "lon"=>-97.105351),
"/US/US-SD/wall-19657" => array("type"=>"town", "pop"=>802, "name_en"=>"Wall", "lang"=>"en", "lat"=>43.991571, "lon"=>-102.239127),
"/US/US-SD/wallace-19658" => array("type"=>"town", "pop"=>83, "name_ru"=>"Уоллэс", "name_en"=>"Wallace", "lang"=>"en", "lat"=>45.085108, "lon"=>-97.477255),
"/US/US-SD/wanblee-19659" => array("type"=>"cdp", "pop"=>654, "name_en"=>"Wanblee", "lang"=>"en", "lat"=>43.574200, "lon"=>-101.661708),
"/US/US-SD/ward-19660" => array("type"=>"town", "pop"=>49, "name_ru"=>"Уорд", "name_en"=>"Ward", "lang"=>"en", "lat"=>44.155356, "lon"=>-96.460786),
"/US/US-SD/warner-19661" => array("type"=>"town", "pop"=>444, "name_en"=>"Warner", "lang"=>"en", "lat"=>45.324342, "lon"=>-98.495204),
"/US/US-SD/wasta-19662" => array("type"=>"town", "pop"=>75, "name_en"=>"Wasta", "lang"=>"en", "lat"=>44.069306, "lon"=>-102.445869),
"/US/US-SD/watertown-19663" => array("type"=>"city", "pop"=>20488, "name_ru"=>"Уотертаун", "name_en"=>"Watertown", "lang"=>"en", "lat"=>44.912596, "lon"=>-97.166597),
"/US/US-SD/waubay-19664" => array("type"=>"city", "pop"=>585, "name_en"=>"Waubay", "lang"=>"en", "lat"=>45.333227, "lon"=>-97.306665),
"/US/US-SD/webster-19665" => array("type"=>"city", "pop"=>1669, "name_ru"=>"Вебстер", "name_en"=>"Webster", "lang"=>"en", "lat"=>45.335724, "lon"=>-97.521937),
"/US/US-SD/wentworth-19666" => array("type"=>"village", "pop"=>186, "name_en"=>"Wentworth", "lang"=>"en", "lat"=>43.996841, "lon"=>-96.964311),
"/US/US-SD/wessington-19667" => array("type"=>"city", "pop"=>226, "name_en"=>"Wessington", "lang"=>"en", "lat"=>44.454856, "lon"=>-98.696157),
"/US/US-SD/wessington-springs-19668" => array("type"=>"city", "pop"=>846, "name_en"=>"Wessington Springs", "lang"=>"en", "lat"=>44.080789, "lon"=>-98.571254),
"/US/US-SD/westport-19669" => array("type"=>"town", "pop"=>120, "name_ru"=>"Уэстпорт", "name_en"=>"Westport", "lang"=>"en", "lat"=>45.647895, "lon"=>-98.497260),
"/US/US-SD/wetonka-19670" => array("type"=>"town", "pop"=>10, "name_en"=>"Wetonka", "lang"=>"en", "lat"=>45.624567, "lon"=>-98.769978),
"/US/US-SD/white-19671" => array("type"=>"city", "pop"=>512, "name_ru"=>"Уайт", "name_en"=>"White", "lang"=>"en", "lat"=>44.433383, "lon"=>-96.645424),
"/US/US-SD/white-horse-19672" => array("type"=>"cdp", "pop"=>152, "name_en"=>"White Horse", "lang"=>"en", "lat"=>43.308286, "lon"=>-100.595983),
"/US/US-SD/white-lake-19673" => array("type"=>"city", "pop"=>378, "name_en"=>"White Lake", "lang"=>"en", "lat"=>43.728434, "lon"=>-98.712949),
"/US/US-SD/white-river-19674" => array("type"=>"city", "pop"=>553, "name_en"=>"White River", "lang"=>"en", "lat"=>43.567327, "lon"=>-100.744306),
"/US/US-SD/white-rock-19675" => array("type"=>"town", "pop"=>18, "name_en"=>"White Rock", "lang"=>"en", "lat"=>45.925258, "lon"=>-96.572543),
"/US/US-SD/whitewood-19676" => array("type"=>"city", "pop"=>849, "name_en"=>"Whitewood", "lang"=>"en", "lat"=>44.461571, "lon"=>-103.640080),
"/US/US-SD/willow-lake-19677" => array("type"=>"city", "pop"=>237, "name_en"=>"Willow Lake", "lang"=>"en", "lat"=>44.627924, "lon"=>-97.637844),
"/US/US-SD/wilmot-19678" => array("type"=>"city", "pop"=>514, "name_ru"=>"Уилмот", "name_en"=>"Wilmot", "lang"=>"en", "lat"=>45.409295, "lon"=>-96.856904),
"/US/US-SD/winfred-19679" => array("type"=>"town", "pop"=>203, "name_en"=>"Winfred", "lang"=>"en", "lat"=>43.997617, "lon"=>-97.362586),
"/US/US-SD/winner-19680" => array("type"=>"city", "pop"=>2744, "name_en"=>"Winner", "lang"=>"en", "lat"=>43.373858, "lon"=>-99.858695),
"/US/US-SD/wolsey-19681" => array("type"=>"town", "pop"=>386, "name_en"=>"Wolsey", "lang"=>"en", "lat"=>44.410871, "lon"=>-98.473769),
"/US/US-SD/wood-19682" => array("type"=>"town", "pop"=>63, "name_en"=>"Wood", "lang"=>"en", "lat"=>43.496760, "lon"=>-100.479821),
"/US/US-SD/woonsocket-19683" => array("type"=>"city", "pop"=>641, "name_en"=>"Woonsocket", "lang"=>"en", "lat"=>44.054712, "lon"=>-98.273739),
"/US/US-SD/worthing-19684" => array("type"=>"town", "pop"=>993, "name_en"=>"Worthing", "lang"=>"en", "lat"=>43.328876, "lon"=>-96.764944),
"/US/US-SD/wounded-knee-19685" => array("type"=>"cdp", "pop"=>18, "name_en"=>"Wounded Knee", "lang"=>"en", "lat"=>43.142480, "lon"=>-102.364634),
"/US/US-SD/yale-19686" => array("type"=>"town", "pop"=>109, "name_ru"=>"Йель", "name_en"=>"Yale", "lang"=>"en", "lat"=>44.433436, "lon"=>-97.989051),
"/US/US-SD/yankton-19687" => array("type"=>"city", "pop"=>13798, "name_en"=>"Yankton", "lang"=>"en", "lat"=>42.889849, "lon"=>-97.392079),
"/US/US-TN" => array("type"=>"state", "pop"=>6214888, "name_ru"=>"Теннесси", "name_en"=>"Tennessee", "lang"=>"en"),
"/US/US-TN/adams-19688" => array("type"=>"city", "pop"=>611, "name_ru"=>"Адамс", "name_en"=>"Adams", "lang"=>"en", "lat"=>36.581645, "lon"=>-87.065116),
"/US/US-TN/adamsville-19689" => array("type"=>"town", "pop"=>2122, "name_ru"=>"Адамсвилл", "name_en"=>"Adamsville", "lang"=>"en", "lat"=>35.234715, "lon"=>-88.387599),
"/US/US-TN/alamo-19690" => array("type"=>"town", "pop"=>2316, "name_ru"=>"Аламо", "name_en"=>"Alamo", "lang"=>"en", "lat"=>35.782630, "lon"=>-89.115726),
"/US/US-TN/alcoa-19691" => array("type"=>"city", "pop"=>8606, "name_en"=>"Alcoa", "lang"=>"en", "lat"=>35.792955, "lon"=>-83.979232),
"/US/US-TN/alexandria-19692" => array("type"=>"town", "pop"=>873, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>36.076677, "lon"=>-86.034596),
"/US/US-TN/algood-19693" => array("type"=>"town", "pop"=>3379, "name_en"=>"Algood", "lang"=>"en", "lat"=>36.197716, "lon"=>-85.447234),
"/US/US-TN/allardt-19694" => array("type"=>"city", "pop"=>677, "name_en"=>"Allardt", "lang"=>"en", "lat"=>36.381022, "lon"=>-84.881398),
"/US/US-TN/altamont-19695" => array("type"=>"town", "pop"=>1113, "name_ru"=>"Альтамонт", "name_en"=>"Altamont", "lang"=>"en", "lat"=>35.429735, "lon"=>-85.728441),
"/US/US-TN/ardmore-19696" => array("type"=>"city", "pop"=>1163, "name_ru"=>"Ардмор", "name_en"=>"Ardmore", "lang"=>"en", "lat"=>34.999893, "lon"=>-86.856565),
"/US/US-TN/arlington-19697" => array("type"=>"town", "pop"=>3918, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>35.282200, "lon"=>-89.664214),
"/US/US-TN/ashland-city-19698" => array("type"=>"town", "pop"=>4582, "name_en"=>"Ashland City", "lang"=>"en", "lat"=>36.276877, "lon"=>-87.059785),
"/US/US-TN/athens-19699" => array("type"=>"city", "pop"=>14275, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>35.451497, "lon"=>-84.601956),
"/US/US-TN/atoka-19700" => array("type"=>"town", "pop"=>7334, "name_en"=>"Atoka", "lang"=>"en", "lat"=>35.445166, "lon"=>-89.772666),
"/US/US-TN/atwood-19701" => array("type"=>"town", "pop"=>968, "name_ru"=>"Атвуд", "name_en"=>"Atwood", "lang"=>"en", "lat"=>35.976963, "lon"=>-88.670968),
"/US/US-TN/auburntown-19702" => array("type"=>"town", "pop"=>266, "name_en"=>"Auburntown", "lang"=>"en", "lat"=>35.948943, "lon"=>-86.094264),
"/US/US-TN/baileyton-19703" => array("type"=>"town", "pop"=>492, "name_ru"=>"Бейлейтон", "name_en"=>"Baileyton", "lang"=>"en", "lat"=>36.327995, "lon"=>-82.830621),
"/US/US-TN/baneberry-19704" => array("type"=>"city", "pop"=>475, "name_en"=>"Baneberry", "lang"=>"en", "lat"=>36.044844, "lon"=>-83.283454),
"/US/US-TN/banner-hill-19705" => array("type"=>"cdp", "pop"=>1717, "name_en"=>"Banner Hill", "lang"=>"en", "lat"=>36.127645, "lon"=>-82.405573),
"/US/US-TN/bartlett-19706" => array("type"=>"town", "pop"=>47501, "name_ru"=>"Барлет", "name_en"=>"Bartlett", "lang"=>"en", "lat"=>35.215512, "lon"=>-89.835329),
"/US/US-TN/baxter-19707" => array("type"=>"town", "pop"=>1412, "name_ru"=>"Бэкстер", "name_en"=>"Baxter", "lang"=>"en", "lat"=>36.154261, "lon"=>-85.638311),
"/US/US-TN/beersheba-springs-19708" => array("type"=>"town", "pop"=>546, "name_en"=>"Beersheba Springs", "lang"=>"en", "lat"=>35.457383, "lon"=>-85.671314),
"/US/US-TN/belinda-city-19709" => array("type"=>"cdp", "pop"=>2098, "name_en"=>"Belinda City", "lang"=>"en", "lat"=>36.169998, "lon"=>-86.477938),
"/US/US-TN/bell-buckle-19710" => array("type"=>"town", "pop"=>411, "name_en"=>"Bell Buckle", "lang"=>"en", "lat"=>35.590678, "lon"=>-86.354054),
"/US/US-TN/belle-meade-19711" => array("type"=>"city", "pop"=>3452, "name_en"=>"Belle Meade", "lang"=>"en", "lat"=>36.098850, "lon"=>-86.855034),
"/US/US-TN/bells-19712" => array("type"=>"town", "pop"=>2254, "name_en"=>"Bells", "lang"=>"en", "lat"=>35.718350, "lon"=>-89.085321),
"/US/US-TN/benton-19713" => array("type"=>"town", "pop"=>1202, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>35.175299, "lon"=>-84.652816),
"/US/US-TN/berry-hill-19714" => array("type"=>"city", "pop"=>806, "name_en"=>"Berry Hill", "lang"=>"en", "lat"=>36.119300, "lon"=>-86.767783),
"/US/US-TN/bethel-springs-19715" => array("type"=>"town", "pop"=>776, "name_en"=>"Bethel Springs", "lang"=>"en", "lat"=>35.235455, "lon"=>-88.609911),
"/US/US-TN/big-sandy-19716" => array("type"=>"town", "pop"=>509, "name_en"=>"Big Sandy", "lang"=>"en", "lat"=>36.232670, "lon"=>-88.085749),
"/US/US-TN/blaine-19717" => array("type"=>"city", "pop"=>1751, "name_ru"=>"Блейн", "name_en"=>"Blaine", "lang"=>"en", "lat"=>36.144357, "lon"=>-83.694859),
"/US/US-TN/bloomingdale-19718" => array("type"=>"cdp", "pop"=>10953, "name_ru"=>"Блумингдэйл", "name_en"=>"Bloomingdale", "lang"=>"en", "lat"=>36.578347, "lon"=>-82.508441),
"/US/US-TN/blountville-19719" => array("type"=>"cdp", "pop"=>2605, "name_en"=>"Blountville", "lang"=>"en", "lat"=>36.533110, "lon"=>-82.327327),
"/US/US-TN/bluff-city-19720" => array("type"=>"city", "pop"=>1653, "name_ru"=>"Блафф Сити", "name_en"=>"Bluff City", "lang"=>"en", "lat"=>36.463735, "lon"=>-82.274178),
"/US/US-TN/bolivar-19721" => array("type"=>"city", "pop"=>5647, "name_ru"=>"Боливар", "name_en"=>"Bolivar", "lang"=>"en", "lat"=>35.264019, "lon"=>-88.999108),
"/US/US-TN/braden-19722" => array("type"=>"town", "pop"=>281, "name_en"=>"Braden", "lang"=>"en", "lat"=>35.369682, "lon"=>-89.575896),
"/US/US-TN/bradford-19723" => array("type"=>"town", "pop"=>1068, "name_ru"=>"Брадфорд", "name_en"=>"Bradford", "lang"=>"en", "lat"=>36.073949, "lon"=>-88.817684),
"/US/US-TN/brentwood-19724" => array("type"=>"city", "pop"=>36176, "name_ru"=>"Брентвуд", "name_en"=>"Brentwood", "lang"=>"en", "lat"=>35.996727, "lon"=>-86.787725),
"/US/US-TN/brighton-19725" => array("type"=>"town", "pop"=>2617, "name_ru"=>"Брайтон", "name_en"=>"Brighton", "lang"=>"en", "lat"=>35.479803, "lon"=>-89.739033),
"/US/US-TN/bristol-19726" => array("type"=>"city", "pop"=>25817, "name_ru"=>"Бристол", "name_en"=>"Bristol", "lang"=>"en", "lat"=>36.567793, "lon"=>-82.197363),
"/US/US-TN/brownsville-19727" => array("type"=>"city", "pop"=>10316, "name_ru"=>"Браунсвилл", "name_en"=>"Brownsville", "lang"=>"en", "lat"=>35.594484, "lon"=>-89.255004),
"/US/US-TN/bruceton-19728" => array("type"=>"town", "pop"=>1450, "name_en"=>"Bruceton", "lang"=>"en", "lat"=>36.034406, "lon"=>-88.247100),
"/US/US-TN/bulls-gap-19729" => array("type"=>"town", "pop"=>749, "name_en"=>"Bulls Gap", "lang"=>"en", "lat"=>36.264228, "lon"=>-83.079026),
"/US/US-TN/burlison-19730" => array("type"=>"town", "pop"=>441, "name_en"=>"Burlison", "lang"=>"en", "lat"=>35.558713, "lon"=>-89.785435),
"/US/US-TN/burns-19731" => array("type"=>"town", "pop"=>1398, "name_ru"=>"Бернс", "name_en"=>"Burns", "lang"=>"en", "lat"=>36.053207, "lon"=>-87.312043),
"/US/US-TN/byrdstown-19732" => array("type"=>"town", "pop"=>871, "name_en"=>"Byrdstown", "lang"=>"en", "lat"=>36.573994, "lon"=>-85.133315),
"/US/US-TN/calhoun-19733" => array("type"=>"town", "pop"=>523, "name_ru"=>"Калхун", "name_en"=>"Calhoun", "lang"=>"en", "lat"=>35.296993, "lon"=>-84.747333),
"/US/US-TN/camden-19734" => array("type"=>"town", "pop"=>3658, "name_ru"=>"Камден", "name_en"=>"Camden", "lang"=>"en", "lat"=>36.064445, "lon"=>-88.104877),
"/US/US-TN/carthage-19735" => array("type"=>"town", "pop"=>2236, "name_ru"=>"Картаж", "name_en"=>"Carthage", "lang"=>"en", "lat"=>36.256777, "lon"=>-85.943720),
"/US/US-TN/caryville-19736" => array("type"=>"town", "pop"=>2385, "name_ru"=>"Кэривилл", "name_en"=>"Caryville", "lang"=>"en", "lat"=>36.324434, "lon"=>-84.219327),
"/US/US-TN/cedar-hill-19737" => array("type"=>"town", "pop"=>321, "name_ru"=>"Сидар Хилл", "name_en"=>"Cedar Hill", "lang"=>"en", "lat"=>36.550613, "lon"=>-87.000100),
"/US/US-TN/celina-19738" => array("type"=>"town", "pop"=>1337, "name_en"=>"Celina", "lang"=>"en", "lat"=>36.548318, "lon"=>-85.502011),
"/US/US-TN/centertown-19739" => array("type"=>"town", "pop"=>266, "name_ru"=>"Сентертаун", "name_en"=>"Centertown", "lang"=>"en", "lat"=>35.724837, "lon"=>-85.919666),
"/US/US-TN/centerville-19740" => array("type"=>"town", "pop"=>3977, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>35.794097, "lon"=>-87.449832),
"/US/US-TN/central-19741" => array("type"=>"cdp", "pop"=>2635, "name_ru"=>"Сентрал", "name_en"=>"Central", "lang"=>"en", "lat"=>36.326633, "lon"=>-82.295446),
"/US/US-TN/chapel-hill-19742" => array("type"=>"town", "pop"=>1341, "name_en"=>"Chapel Hill", "lang"=>"en", "lat"=>35.628085, "lon"=>-86.696206),
"/US/US-TN/charleston-19743" => array("type"=>"city", "pop"=>664, "name_ru"=>"Чарльстон", "name_en"=>"Charleston", "lang"=>"en", "lat"=>35.287185, "lon"=>-84.760394),
"/US/US-TN/charlotte-19744" => array("type"=>"town", "pop"=>1166, "name_ru"=>"Шарлотт", "name_en"=>"Charlotte", "lang"=>"en", "lat"=>36.184682, "lon"=>-87.347546),
"/US/US-TN/chattanooga-19745" => array("type"=>"city", "pop"=>170880, "name_en"=>"Chattanooga", "lang"=>"en", "lat"=>35.066209, "lon"=>-85.256956),
"/US/US-TN/church-hill-19746" => array("type"=>"town", "pop"=>6765, "name_ru"=>"Черч Хилл", "name_en"=>"Church Hill", "lang"=>"en", "lat"=>36.525264, "lon"=>-82.711965),
"/US/US-TN/clarksburg-19747" => array("type"=>"town", "pop"=>370, "name_ru"=>"Кларксбург", "name_en"=>"Clarksburg", "lang"=>"en", "lat"=>35.871628, "lon"=>-88.389805),
"/US/US-TN/clarksville-19748" => array("type"=>"city", "pop"=>119735, "name_ru"=>"Кларксвилл", "name_en"=>"Clarksville", "lang"=>"en", "lat"=>36.561139, "lon"=>-87.347305),
"/US/US-TN/cleveland-19749" => array("type"=>"city", "pop"=>39753, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>35.177558, "lon"=>-84.867544),
"/US/US-TN/clifton-19750" => array("type"=>"city", "pop"=>2694, "name_ru"=>"Слифтон", "name_en"=>"Clifton", "lang"=>"en", "lat"=>35.381441, "lon"=>-87.992599),
"/US/US-TN/clinton-19751" => array("type"=>"town", "pop"=>9625, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>36.090773, "lon"=>-84.130548),
"/US/US-TN/coalmont-19752" => array("type"=>"city", "pop"=>945, "name_en"=>"Coalmont", "lang"=>"en", "lat"=>35.339032, "lon"=>-85.713428),
"/US/US-TN/collegedale-19753" => array("type"=>"city", "pop"=>7947, "name_en"=>"Collegedale", "lang"=>"en", "lat"=>35.044683, "lon"=>-85.050323),
"/US/US-TN/collierville-19754" => array("type"=>"town", "pop"=>39211, "name_en"=>"Collierville", "lang"=>"en", "lat"=>35.055257, "lon"=>-89.678514),
"/US/US-TN/collinwood-19755" => array("type"=>"city", "pop"=>1012, "name_en"=>"Collinwood", "lang"=>"en", "lat"=>35.176364, "lon"=>-87.744303),
"/US/US-TN/colonial-heights-19756" => array("type"=>"cdp", "pop"=>6716, "name_en"=>"Colonial Heights", "lang"=>"en", "lat"=>36.479567, "lon"=>-82.511830),
"/US/US-TN/columbia-19757" => array("type"=>"city", "pop"=>34402, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>35.622451, "lon"=>-87.048952),
"/US/US-TN/cookeville-19758" => array("type"=>"city", "pop"=>29234, "name_en"=>"Cookeville", "lang"=>"en", "lat"=>36.156856, "lon"=>-85.505411),
"/US/US-TN/copperhill-19759" => array("type"=>"city", "pop"=>442, "name_en"=>"Copperhill", "lang"=>"en", "lat"=>34.991481, "lon"=>-84.364889),
"/US/US-TN/cornersville-19760" => array("type"=>"town", "pop"=>965, "name_en"=>"Cornersville", "lang"=>"en", "lat"=>35.358744, "lon"=>-86.841030),
"/US/US-TN/cottage-grove-19761" => array("type"=>"town", "pop"=>98, "name_ru"=>"Коттэдж Гров", "name_en"=>"Cottage Grove", "lang"=>"en", "lat"=>36.379793, "lon"=>-88.477861),
"/US/US-TN/covington-19762" => array("type"=>"city", "pop"=>9253, "name_ru"=>"Ковингтон", "name_en"=>"Covington", "lang"=>"en", "lat"=>35.567136, "lon"=>-89.649434),
"/US/US-TN/cowan-19763" => array("type"=>"city", "pop"=>1716, "name_en"=>"Cowan", "lang"=>"en", "lat"=>35.165223, "lon"=>-86.016129),
"/US/US-TN/crab-orchard-19764" => array("type"=>"city", "pop"=>930, "name_ru"=>"Крэб Орчард", "name_en"=>"Crab Orchard", "lang"=>"en", "lat"=>35.909598, "lon"=>-84.882079),
"/US/US-TN/cross-plains-19765" => array("type"=>"city", "pop"=>1640, "name_en"=>"Cross Plains", "lang"=>"en", "lat"=>36.540137, "lon"=>-86.681152),
"/US/US-TN/crossville-19766" => array("type"=>"city", "pop"=>11599, "name_ru"=>"Кроссвилл", "name_en"=>"Crossville", "lang"=>"en", "lat"=>35.951912, "lon"=>-85.036814),
"/US/US-TN/crump-19767" => array("type"=>"city", "pop"=>1459, "name_en"=>"Crump", "lang"=>"en", "lat"=>35.239762, "lon"=>-88.321832),
"/US/US-TN/cumberland-city-19768" => array("type"=>"town", "pop"=>327, "name_en"=>"Cumberland City", "lang"=>"en", "lat"=>36.382187, "lon"=>-87.644017),
"/US/US-TN/cumberland-gap-19769" => array("type"=>"town", "pop"=>208, "name_en"=>"Cumberland Gap", "lang"=>"en", "lat"=>36.598030, "lon"=>-83.665208),
"/US/US-TN/dandridge-19770" => array("type"=>"town", "pop"=>2687, "name_en"=>"Dandridge", "lang"=>"en", "lat"=>36.023105, "lon"=>-83.418719),
"/US/US-TN/dayton-19771" => array("type"=>"city", "pop"=>6771, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>35.492771, "lon"=>-85.013047),
"/US/US-TN/decatur-19772" => array("type"=>"town", "pop"=>1476, "name_ru"=>"Декатур", "name_en"=>"Decatur", "lang"=>"en", "lat"=>35.525048, "lon"=>-84.791236),
"/US/US-TN/decaturville-19773" => array("type"=>"town", "pop"=>823, "name_en"=>"Decaturville", "lang"=>"en", "lat"=>35.582333, "lon"=>-88.119246),
"/US/US-TN/decherd-19774" => array("type"=>"town", "pop"=>2125, "name_en"=>"Decherd", "lang"=>"en", "lat"=>35.209270, "lon"=>-86.087919),
"/US/US-TN/dickson-19775" => array("type"=>"city", "pop"=>13972, "name_en"=>"Dickson", "lang"=>"en", "lat"=>36.076701, "lon"=>-87.377092),
"/US/US-TN/dover-19776" => array("type"=>"city", "pop"=>1613, "name_ru"=>"Довер", "name_en"=>"Dover", "lang"=>"en", "lat"=>36.481390, "lon"=>-87.843664),
"/US/US-TN/dowelltown-19777" => array("type"=>"town", "pop"=>319, "name_en"=>"Dowelltown", "lang"=>"en", "lat"=>36.013067, "lon"=>-85.943067),
"/US/US-TN/doyle-19778" => array("type"=>"town", "pop"=>572, "name_en"=>"Doyle", "lang"=>"en", "lat"=>35.852109, "lon"=>-85.512763),
"/US/US-TN/dresden-19779" => array("type"=>"town", "pop"=>2841, "name_ru"=>"Дрезден", "name_en"=>"Dresden", "lang"=>"en", "lat"=>36.277430, "lon"=>-88.690449),
"/US/US-TN/ducktown-19780" => array("type"=>"city", "pop"=>381, "name_en"=>"Ducktown", "lang"=>"en", "lat"=>35.035832, "lon"=>-84.389774),
"/US/US-TN/dunlap-19781" => array("type"=>"city", "pop"=>5412, "name_ru"=>"Данлап", "name_en"=>"Dunlap", "lang"=>"en", "lat"=>35.372703, "lon"=>-85.386532),
"/US/US-TN/dyer-19782" => array("type"=>"city", "pop"=>2414, "name_ru"=>"Дайер", "name_en"=>"Dyer", "lang"=>"en", "lat"=>36.070861, "lon"=>-88.992048),
"/US/US-TN/dyersburg-19783" => array("type"=>"city", "pop"=>17126, "name_en"=>"Dyersburg", "lang"=>"en", "lat"=>36.039227, "lon"=>-89.380162),
"/US/US-TN/eagleton-village-19784" => array("type"=>"cdp", "pop"=>5169, "name_en"=>"Eagleton Village", "lang"=>"en", "lat"=>35.794016, "lon"=>-83.936703),
"/US/US-TN/eagleville-19785" => array("type"=>"city", "pop"=>523, "name_ru"=>"Иглвилл", "name_en"=>"Eagleville", "lang"=>"en", "lat"=>35.740144, "lon"=>-86.653603),
"/US/US-TN/east-brainerd-19786" => array("type"=>"cdp", "pop"=>11594, "name_en"=>"East Brainerd", "lang"=>"en", "lat"=>35.013808, "lon"=>-85.108850),
"/US/US-TN/east-cleveland-19787" => array("type"=>"cdp", "pop"=>1249, "name_en"=>"East Cleveland", "lang"=>"en", "lat"=>35.149953, "lon"=>-84.853596),
"/US/US-TN/east-ridge-19788" => array("type"=>"city", "pop"=>19642, "name_en"=>"East Ridge", "lang"=>"en", "lat"=>34.997024, "lon"=>-85.229021),
"/US/US-TN/eastview-19789" => array("type"=>"town", "pop"=>640, "name_en"=>"Eastview", "lang"=>"en", "lat"=>35.088977, "lon"=>-88.553822),
"/US/US-TN/elizabethton-19790" => array("type"=>"city", "pop"=>13955, "name_en"=>"Elizabethton", "lang"=>"en", "lat"=>36.338759, "lon"=>-82.231996),
"/US/US-TN/elkton-19791" => array("type"=>"town", "pop"=>594, "name_ru"=>"Элктон", "name_en"=>"Elkton", "lang"=>"en", "lat"=>35.060192, "lon"=>-86.896431),
"/US/US-TN/englewood-19792" => array("type"=>"town", "pop"=>1750, "name_ru"=>"Энглвуд", "name_en"=>"Englewood", "lang"=>"en", "lat"=>35.422641, "lon"=>-84.489746),
"/US/US-TN/enville-19793" => array("type"=>"town", "pop"=>233, "name_en"=>"Enville", "lang"=>"en", "lat"=>35.387884, "lon"=>-88.424722),
"/US/US-TN/erin-19794" => array("type"=>"city", "pop"=>1460, "name_en"=>"Erin", "lang"=>"en", "lat"=>36.315882, "lon"=>-87.702595),
"/US/US-TN/erwin-19795" => array("type"=>"city", "pop"=>5802, "name_en"=>"Erwin", "lang"=>"en", "lat"=>36.144928, "lon"=>-82.410936),
"/US/US-TN/estill-springs-19796" => array("type"=>"town", "pop"=>2261, "name_en"=>"Estill Springs", "lang"=>"en", "lat"=>35.264463, "lon"=>-86.131019),
"/US/US-TN/ethridge-19797" => array("type"=>"town", "pop"=>553, "name_en"=>"Ethridge", "lang"=>"en", "lat"=>35.321151, "lon"=>-87.302289),
"/US/US-TN/etowah-19798" => array("type"=>"city", "pop"=>3775, "name_en"=>"Etowah", "lang"=>"en", "lat"=>35.335011, "lon"=>-84.528912),
"/US/US-TN/fairfield-glade-19799" => array("type"=>"cdp", "pop"=>2209, "name_en"=>"Fairfield Glade", "lang"=>"en", "lat"=>36.000115, "lon"=>-84.899399),
"/US/US-TN/fairmount-19800" => array("type"=>"cdp", "pop"=>1578, "name_ru"=>"Фэрмаунт", "name_en"=>"Fairmount", "lang"=>"en", "lat"=>35.177077, "lon"=>-85.335390),
"/US/US-TN/fairview-19801" => array("type"=>"city", "pop"=>7870, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>35.979743, "lon"=>-87.124151),
"/US/US-TN/fall-branch-19802" => array("type"=>"cdp", "pop"=>1203, "name_en"=>"Fall Branch", "lang"=>"en", "lat"=>36.417236, "lon"=>-82.623363),
"/US/US-TN/farragut-19803" => array("type"=>"town", "pop"=>20350, "name_ru"=>"Фаррагут", "name_en"=>"Farragut", "lang"=>"en", "lat"=>35.874388, "lon"=>-84.182215),
"/US/US-TN/fayetteville-19804" => array("type"=>"city", "pop"=>7166, "name_ru"=>"Фейетвилл", "name_en"=>"Fayetteville", "lang"=>"en", "lat"=>35.149899, "lon"=>-86.564436),
"/US/US-TN/finger-19805" => array("type"=>"city", "pop"=>363, "name_en"=>"Finger", "lang"=>"en", "lat"=>35.354736, "lon"=>-88.615478),
"/US/US-TN/forest-hills-19806" => array("type"=>"city", "pop"=>6018, "name_ru"=>"Форест Хиллс", "name_en"=>"Forest Hills", "lang"=>"en", "lat"=>36.063400, "lon"=>-86.836850),
"/US/US-TN/franklin-19807" => array("type"=>"city", "pop"=>58481, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>35.929410, "lon"=>-86.841982),
"/US/US-TN/friendship-19808" => array("type"=>"city", "pop"=>594, "name_ru"=>"Френдшип", "name_en"=>"Friendship", "lang"=>"en", "lat"=>35.910298, "lon"=>-89.241758),
"/US/US-TN/friendsville-19809" => array("type"=>"city", "pop"=>921, "name_ru"=>"Френдсвилл", "name_en"=>"Friendsville", "lang"=>"en", "lat"=>35.757227, "lon"=>-84.130438),
"/US/US-TN/gadsden-19810" => array("type"=>"town", "pop"=>533, "name_ru"=>"Гадсден", "name_en"=>"Gadsden", "lang"=>"en", "lat"=>35.778372, "lon"=>-88.983557),
"/US/US-TN/gainesboro-19811" => array("type"=>"town", "pop"=>845, "name_en"=>"Gainesboro", "lang"=>"en", "lat"=>36.360478, "lon"=>-85.653795),
"/US/US-TN/gallatin-19812" => array("type"=>"city", "pop"=>29343, "name_ru"=>"Галлатин", "name_en"=>"Gallatin", "lang"=>"en", "lat"=>36.378650, "lon"=>-86.453132),
"/US/US-TN/gallaway-19813" => array("type"=>"city", "pop"=>692, "name_en"=>"Gallaway", "lang"=>"en", "lat"=>35.338264, "lon"=>-89.603463),
"/US/US-TN/garland-19814" => array("type"=>"town", "pop"=>327, "name_ru"=>"Гарленд", "name_en"=>"Garland", "lang"=>"en", "lat"=>35.586476, "lon"=>-89.752037),
"/US/US-TN/gates-19815" => array("type"=>"town", "pop"=>851, "name_en"=>"Gates", "lang"=>"en", "lat"=>35.839408, "lon"=>-89.407813),
"/US/US-TN/gatlinburg-19816" => array("type"=>"city", "pop"=>5555, "name_en"=>"Gatlinburg", "lang"=>"en", "lat"=>35.723395, "lon"=>-83.493743),
"/US/US-TN/germantown-19817" => array("type"=>"city", "pop"=>37251, "name_ru"=>"Джермэнтаун", "name_en"=>"Germantown", "lang"=>"en", "lat"=>35.088867, "lon"=>-89.791699),
"/US/US-TN/gibson-19818" => array("type"=>"town", "pop"=>401, "name_ru"=>"Гибсон", "name_en"=>"Gibson", "lang"=>"en", "lat"=>35.873981, "lon"=>-88.845906),
"/US/US-TN/gilt-edge-19819" => array("type"=>"city", "pop"=>481, "name_en"=>"Gilt Edge", "lang"=>"en", "lat"=>35.532819, "lon"=>-89.830685),
"/US/US-TN/gleason-19820" => array("type"=>"town", "pop"=>1401, "name_en"=>"Gleason", "lang"=>"en", "lat"=>36.215621, "lon"=>-88.610182),
"/US/US-TN/goodlettsville-19821" => array("type"=>"city", "pop"=>17075, "name_en"=>"Goodlettsville", "lang"=>"en", "lat"=>36.333725, "lon"=>-86.703831),
"/US/US-TN/gordonsville-19822" => array("type"=>"town", "pop"=>1291, "name_en"=>"Gordonsville", "lang"=>"en", "lat"=>36.183057, "lon"=>-85.932634),
"/US/US-TN/grand-junction-19823" => array("type"=>"city", "pop"=>306, "name_ru"=>"Гранд Джанкшен", "name_en"=>"Grand Junction", "lang"=>"en", "lat"=>35.049158, "lon"=>-89.188626),
"/US/US-TN/gray-19824" => array("type"=>"cdp", "pop"=>1071, "name_ru"=>"Грей", "name_en"=>"Gray", "lang"=>"en", "lat"=>36.411167, "lon"=>-82.479739),
"/US/US-TN/graysville-19825" => array("type"=>"town", "pop"=>1429, "name_ru"=>"Грейсвилл", "name_en"=>"Graysville", "lang"=>"en", "lat"=>35.447795, "lon"=>-85.082495),
"/US/US-TN/green-hill-19830" => array("type"=>"cdp", "pop"=>6763, "name_en"=>"Green Hill", "lang"=>"en", "lat"=>36.234798, "lon"=>-86.570917),
"/US/US-TN/greenback-19826" => array("type"=>"city", "pop"=>1110, "name_en"=>"Greenback", "lang"=>"en", "lat"=>35.651763, "lon"=>-84.175165),
"/US/US-TN/greenbrier-19827" => array("type"=>"town", "pop"=>6553, "name_ru"=>"Гринбрир", "name_en"=>"Greenbrier", "lang"=>"en", "lat"=>36.422950, "lon"=>-86.800874),
"/US/US-TN/greeneville-19828" => array("type"=>"town", "pop"=>15440, "name_en"=>"Greeneville", "lang"=>"en", "lat"=>36.165310, "lon"=>-82.818975),
"/US/US-TN/greenfield-19829" => array("type"=>"town", "pop"=>2039, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>36.155445, "lon"=>-88.801405),
"/US/US-TN/gruetli-laager-19831" => array("type"=>"city", "pop"=>1855, "name_en"=>"Gruetli-Laager", "lang"=>"en", "lat"=>35.371005, "lon"=>-85.637079),
"/US/US-TN/guys-19832" => array("type"=>"town", "pop"=>503, "name_en"=>"Guys", "lang"=>"en", "lat"=>35.015938, "lon"=>-88.544170),
"/US/US-TN/halls-19833" => array("type"=>"town", "pop"=>2191, "name_en"=>"Halls", "lang"=>"en", "lat"=>35.880612, "lon"=>-89.400836),
"/US/US-TN/halls-19834" => array("type"=>"cdp", "pop"=>6450, "name_en"=>"Halls", "lang"=>"en", "lat"=>36.081847, "lon"=>-83.932906),
"/US/US-TN/harriman-19835" => array("type"=>"city", "pop"=>6658, "name_en"=>"Harriman", "lang"=>"en", "lat"=>35.933368, "lon"=>-84.561681),
"/US/US-TN/harrison-19836" => array("type"=>"cdp", "pop"=>7191, "name_ru"=>"Харрисон", "name_en"=>"Harrison", "lang"=>"en", "lat"=>35.127902, "lon"=>-85.146017),
"/US/US-TN/harrogate-shawanee-19837" => array("type"=>"cdp", "pop"=>2657, "name_en"=>"Harrogate-Shawanee", "lang"=>"en", "lat"=>36.581368, "lon"=>-83.643671),
"/US/US-TN/hartsville-19838" => array("type"=>"town", "pop"=>2188, "name_ru"=>"Хартсвилл", "name_en"=>"Hartsville", "lang"=>"en", "lat"=>36.390019, "lon"=>-86.159599),
"/US/US-TN/henderson-19839" => array("type"=>"city", "pop"=>6418, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>35.444913, "lon"=>-88.647713),
"/US/US-TN/hendersonville-19840" => array("type"=>"city", "pop"=>47725, "name_en"=>"Hendersonville", "lang"=>"en", "lat"=>36.293967, "lon"=>-86.603513),
"/US/US-TN/henning-19841" => array("type"=>"town", "pop"=>1279, "name_ru"=>"Хеннинг", "name_en"=>"Henning", "lang"=>"en", "lat"=>35.673337, "lon"=>-89.575403),
"/US/US-TN/henry-19842" => array("type"=>"town", "pop"=>546, "name_ru"=>"Хенри", "name_en"=>"Henry", "lang"=>"en", "lat"=>36.200497, "lon"=>-88.410562),
"/US/US-TN/hickory-valley-19843" => array("type"=>"town", "pop"=>129, "name_en"=>"Hickory Valley", "lang"=>"en", "lat"=>35.154787, "lon"=>-89.126512),
"/US/US-TN/hohenwald-19844" => array("type"=>"city", "pop"=>3805, "name_en"=>"Hohenwald", "lang"=>"en", "lat"=>35.550855, "lon"=>-87.559673),
"/US/US-TN/hollow-rock-19845" => array("type"=>"town", "pop"=>933, "name_en"=>"Hollow Rock", "lang"=>"en", "lat"=>36.035183, "lon"=>-88.272630),
"/US/US-TN/hopewell-19846" => array("type"=>"cdp", "pop"=>2569, "name_ru"=>"Хоупуэлл", "name_en"=>"Hopewell", "lang"=>"en", "lat"=>35.234851, "lon"=>-84.905281),
"/US/US-TN/hornbeak-19847" => array("type"=>"town", "pop"=>416, "name_en"=>"Hornbeak", "lang"=>"en", "lat"=>36.331413, "lon"=>-89.296949),
"/US/US-TN/hornsby-19848" => array("type"=>"town", "pop"=>292, "name_en"=>"Hornsby", "lang"=>"en", "lat"=>35.229672, "lon"=>-88.830385),
"/US/US-TN/humboldt-19849" => array("type"=>"city", "pop"=>9159, "name_ru"=>"Гумбольдт", "name_en"=>"Humboldt", "lang"=>"en", "lat"=>35.821770, "lon"=>-88.906490),
"/US/US-TN/hunter-19850" => array("type"=>"cdp", "pop"=>1250, "name_ru"=>"Хантер", "name_en"=>"Hunter", "lang"=>"en", "lat"=>36.378782, "lon"=>-82.163344),
"/US/US-TN/huntingdon-19851" => array("type"=>"town", "pop"=>4125, "name_en"=>"Huntingdon", "lang"=>"en", "lat"=>36.008539, "lon"=>-88.412001),
"/US/US-TN/huntland-19852" => array("type"=>"town", "pop"=>866, "name_en"=>"Huntland", "lang"=>"en", "lat"=>35.055843, "lon"=>-86.268569),
"/US/US-TN/huntsville-19853" => array("type"=>"town", "pop"=>1035, "name_ru"=>"Хантсвилл", "name_en"=>"Huntsville", "lang"=>"en", "lat"=>36.413622, "lon"=>-84.487537),
"/US/US-TN/iron-city-19854" => array("type"=>"city", "pop"=>378, "name_ru"=>"Айрон Сити", "name_en"=>"Iron City", "lang"=>"en", "lat"=>35.025604, "lon"=>-87.583680),
"/US/US-TN/jacksboro-19855" => array("type"=>"town", "pop"=>2102, "name_en"=>"Jacksboro", "lang"=>"en", "lat"=>36.331768, "lon"=>-84.191040),
"/US/US-TN/jackson-19856" => array("type"=>"city", "pop"=>63158, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>35.633701, "lon"=>-88.828528),
"/US/US-TN/jamestown-19857" => array("type"=>"city", "pop"=>1918, "name_ru"=>"Джеймстаун", "name_en"=>"Jamestown", "lang"=>"en", "lat"=>36.432137, "lon"=>-84.935306),
"/US/US-TN/jasper-19858" => array("type"=>"town", "pop"=>3122, "name_ru"=>"Джаспер", "name_en"=>"Jasper", "lang"=>"en", "lat"=>35.059855, "lon"=>-85.628680),
"/US/US-TN/jefferson-city-19859" => array("type"=>"city", "pop"=>8156, "name_ru"=>"Джефферсон Сити", "name_en"=>"Jefferson City", "lang"=>"en", "lat"=>36.119006, "lon"=>-83.487570),
"/US/US-TN/jellico-19860" => array("type"=>"city", "pop"=>2524, "name_en"=>"Jellico", "lang"=>"en", "lat"=>36.577076, "lon"=>-84.128683),
"/US/US-TN/johnson-city-19861" => array("type"=>"city", "pop"=>61990, "name_ru"=>"Джонсон Сити", "name_en"=>"Johnson City", "lang"=>"en", "lat"=>36.331400, "lon"=>-82.368724),
"/US/US-TN/jonesborough-19862" => array("type"=>"town", "pop"=>5221, "name_en"=>"Jonesborough", "lang"=>"en", "lat"=>36.293600, "lon"=>-82.477727),
"/US/US-TN/karns-19863" => array("type"=>"cdp", "pop"=>1458, "name_en"=>"Karns", "lang"=>"en", "lat"=>35.975984, "lon"=>-84.118229),
"/US/US-TN/kenton-19864" => array("type"=>"town", "pop"=>1291, "name_ru"=>"Кентон", "name_en"=>"Kenton", "lang"=>"en", "lat"=>36.199686, "lon"=>-89.014170),
"/US/US-TN/kimball-19865" => array("type"=>"town", "pop"=>1401, "name_en"=>"Kimball", "lang"=>"en", "lat"=>35.045165, "lon"=>-85.672972),
"/US/US-TN/kingsport-19866" => array("type"=>"city", "pop"=>44473, "name_en"=>"Kingsport", "lang"=>"en", "lat"=>36.529550, "lon"=>-82.556859),
"/US/US-TN/kingston-19867" => array("type"=>"city", "pop"=>5607, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>35.870064, "lon"=>-84.510060),
"/US/US-TN/kingston-springs-19868" => array("type"=>"town", "pop"=>2953, "name_en"=>"Kingston Springs", "lang"=>"en", "lat"=>36.096518, "lon"=>-87.100449),
"/US/US-TN/knoxville-19869" => array("type"=>"city", "pop"=>184802, "name_ru"=>"Ноксвилл", "name_en"=>"Knoxville", "lang"=>"en", "lat"=>35.974550, "lon"=>-83.946288),
"/US/US-TN/la-follette-19871" => array("type"=>"city", "pop"=>8194, "name_en"=>"La Follette", "lang"=>"en", "lat"=>36.372954, "lon"=>-84.127057),
"/US/US-TN/la-grange-19872" => array("type"=>"town", "pop"=>142, "name_ru"=>"Ла Гранж", "name_en"=>"La Grange", "lang"=>"en", "lat"=>35.049250, "lon"=>-89.234694),
"/US/US-TN/la-vergne-19877" => array("type"=>"city", "pop"=>30034, "name_en"=>"La Vergne", "lang"=>"en", "lat"=>35.999300, "lon"=>-86.573195),
"/US/US-TN/lafayette-19870" => array("type"=>"city", "pop"=>4426, "name_ru"=>"Ла Файет", "name_en"=>"Lafayette", "lang"=>"en", "lat"=>36.524480, "lon"=>-86.030933),
"/US/US-TN/lake-city-19873" => array("type"=>"town", "pop"=>1847, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>36.224156, "lon"=>-84.155979),
"/US/US-TN/lakeland-19874" => array("type"=>"city", "pop"=>8093, "name_ru"=>"Лэйклэнд", "name_en"=>"Lakeland", "lang"=>"en", "lat"=>35.240151, "lon"=>-89.729794),
"/US/US-TN/lakesite-19875" => array("type"=>"city", "pop"=>1902, "name_en"=>"Lakesite", "lang"=>"en", "lat"=>35.205193, "lon"=>-85.130481),
"/US/US-TN/lakewood-19876" => array("type"=>"city", "pop"=>2595, "name_ru"=>"Лейквуд", "name_en"=>"Lakewood", "lang"=>"en", "lat"=>36.239533, "lon"=>-86.634328),
"/US/US-TN/lawrenceburg-19878" => array("type"=>"city", "pop"=>10778, "name_ru"=>"Лоренсбург", "name_en"=>"Lawrenceburg", "lang"=>"en", "lat"=>35.245735, "lon"=>-87.337975),
"/US/US-TN/lebanon-19879" => array("type"=>"city", "pop"=>24648, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>36.209780, "lon"=>-86.322200),
"/US/US-TN/lenoir-city-19880" => array("type"=>"city", "pop"=>8010, "name_en"=>"Lenoir City", "lang"=>"en", "lat"=>35.800905, "lon"=>-84.270967),
"/US/US-TN/lewisburg-19881" => array("type"=>"city", "pop"=>11003, "name_ru"=>"Льюисбург", "name_en"=>"Lewisburg", "lang"=>"en", "lat"=>35.446972, "lon"=>-86.788130),
"/US/US-TN/lexington-19882" => array("type"=>"city", "pop"=>7860, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>35.653860, "lon"=>-88.392959),
"/US/US-TN/liberty-19883" => array("type"=>"town", "pop"=>389, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>36.003982, "lon"=>-85.977824),
"/US/US-TN/linden-19884" => array("type"=>"town", "pop"=>999, "name_ru"=>"Линден", "name_en"=>"Linden", "lang"=>"en", "lat"=>35.611237, "lon"=>-87.843912),
"/US/US-TN/livingston-19885" => array("type"=>"town", "pop"=>3567, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>36.388742, "lon"=>-85.327502),
"/US/US-TN/lobelville-19886" => array("type"=>"city", "pop"=>931, "name_en"=>"Lobelville", "lang"=>"en", "lat"=>35.750977, "lon"=>-87.797094),
"/US/US-TN/lookout-mountain-19887" => array("type"=>"town", "pop"=>1866, "name_ru"=>"Лукаут Маунтин", "name_en"=>"Lookout Mountain", "lang"=>"en", "lat"=>34.993950, "lon"=>-85.352249),
"/US/US-TN/loretto-19888" => array("type"=>"city", "pop"=>1705, "name_ru"=>"Лоретто", "name_en"=>"Loretto", "lang"=>"en", "lat"=>35.080318, "lon"=>-87.436083),
"/US/US-TN/loudon-19889" => array("type"=>"town", "pop"=>4921, "name_en"=>"Loudon", "lang"=>"en", "lat"=>35.735251, "lon"=>-84.357805),
"/US/US-TN/luttrell-19890" => array("type"=>"town", "pop"=>933, "name_en"=>"Luttrell", "lang"=>"en", "lat"=>36.208768, "lon"=>-83.744617),
"/US/US-TN/lynchburg,-moore-county-19891" => array("type"=>"county", "pop"=>4721, "name_en"=>"Lynchburg, Moore County", "lang"=>"en", "lat"=>35.284220, "lon"=>-86.357414),
"/US/US-TN/lynnville-19892" => array("type"=>"town", "pop"=>338, "name_ru"=>"Линнвилл", "name_en"=>"Lynnville", "lang"=>"en", "lat"=>35.377871, "lon"=>-87.005551),
"/US/US-TN/madisonville-19897" => array("type"=>"town", "pop"=>4714, "name_ru"=>"Мэдисонвилл", "name_en"=>"Madisonville", "lang"=>"en", "lat"=>35.522831, "lon"=>-84.359892),
"/US/US-TN/manchester-19898" => array("type"=>"city", "pop"=>10012, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>35.468593, "lon"=>-86.080778),
"/US/US-TN/martin-19899" => array("type"=>"city", "pop"=>10208, "name_ru"=>"Мартин", "name_en"=>"Martin", "lang"=>"en", "lat"=>36.342316, "lon"=>-88.858538),
"/US/US-TN/maryville-19900" => array("type"=>"city", "pop"=>27156, "name_ru"=>"Мэривилл", "name_en"=>"Maryville", "lang"=>"en", "lat"=>35.747454, "lon"=>-83.973049),
"/US/US-TN/mascot-19901" => array("type"=>"cdp", "pop"=>2138, "name_en"=>"Mascot", "lang"=>"en", "lat"=>36.068208, "lon"=>-83.766139),
"/US/US-TN/mason-19902" => array("type"=>"town", "pop"=>1231, "name_ru"=>"Мейсон", "name_en"=>"Mason", "lang"=>"en", "lat"=>35.413230, "lon"=>-89.536625),
"/US/US-TN/maury-city-19903" => array("type"=>"town", "pop"=>691, "name_en"=>"Maury City", "lang"=>"en", "lat"=>35.815462, "lon"=>-89.223798),
"/US/US-TN/maynardville-19904" => array("type"=>"city", "pop"=>1909, "name_en"=>"Maynardville", "lang"=>"en", "lat"=>36.246902, "lon"=>-83.806154),
"/US/US-TN/mcewen-19893" => array("type"=>"city", "pop"=>1675, "name_en"=>"McEwen", "lang"=>"en", "lat"=>36.106789, "lon"=>-87.635684),
"/US/US-TN/mckenzie-19894" => array("type"=>"city", "pop"=>5357, "name_ru"=>"Маккензи", "name_en"=>"McKenzie", "lang"=>"en", "lat"=>36.139571, "lon"=>-88.511334),
"/US/US-TN/mclemoresville-19895" => array("type"=>"town", "pop"=>298, "name_en"=>"McLemoresville", "lang"=>"en", "lat"=>35.986019, "lon"=>-88.579416),
"/US/US-TN/mcminnville-19896" => array("type"=>"city", "pop"=>13270, "name_en"=>"McMinnville", "lang"=>"en", "lat"=>35.687502, "lon"=>-85.776279),
"/US/US-TN/medina-19905" => array("type"=>"town", "pop"=>2141, "name_ru"=>"Медина", "name_en"=>"Medina", "lang"=>"en", "lat"=>35.803422, "lon"=>-88.774285),
"/US/US-TN/medon-19906" => array("type"=>"town", "pop"=>191, "name_en"=>"Medon", "lang"=>"en", "lat"=>35.458461, "lon"=>-88.865692),
"/US/US-TN/memphis-19907" => array("type"=>"city", "pop"=>669651, "name_ru"=>"Мемфис", "name_en"=>"Memphis", "lang"=>"en", "lat"=>35.105600, "lon"=>-90.006991),
"/US/US-TN/michie-19908" => array("type"=>"town", "pop"=>673, "name_en"=>"Michie", "lang"=>"en", "lat"=>35.062869, "lon"=>-88.427486),
"/US/US-TN/middle-valley-19910" => array("type"=>"cdp", "pop"=>12255, "name_en"=>"Middle Valley", "lang"=>"en", "lat"=>35.187718, "lon"=>-85.189402),
"/US/US-TN/middleton-19909" => array("type"=>"town", "pop"=>609, "name_ru"=>"Мидлтон", "name_en"=>"Middleton", "lang"=>"en", "lat"=>35.060305, "lon"=>-88.891434),
"/US/US-TN/midway-19911" => array("type"=>"cdp", "pop"=>2953, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>36.289500, "lon"=>-82.423706),
"/US/US-TN/milan-19912" => array("type"=>"city", "pop"=>7934, "name_ru"=>"Милан", "name_en"=>"Milan", "lang"=>"en", "lat"=>35.918914, "lon"=>-88.754564),
"/US/US-TN/milledgeville-19913" => array("type"=>"town", "pop"=>294, "name_ru"=>"Миледжвилл", "name_en"=>"Milledgeville", "lang"=>"en", "lat"=>35.376009, "lon"=>-88.366642),
"/US/US-TN/millersville-19914" => array("type"=>"city", "pop"=>6279, "name_en"=>"Millersville", "lang"=>"en", "lat"=>36.381767, "lon"=>-86.701287),
"/US/US-TN/millington-19915" => array("type"=>"city", "pop"=>10165, "name_ru"=>"Миллингтон", "name_en"=>"Millington", "lang"=>"en", "lat"=>35.334398, "lon"=>-89.888290),
"/US/US-TN/minor-hill-19916" => array("type"=>"city", "pop"=>451, "name_en"=>"Minor Hill", "lang"=>"en", "lat"=>35.038384, "lon"=>-87.171122),
"/US/US-TN/mitchellville-19917" => array("type"=>"town", "pop"=>205, "name_ru"=>"Митчелвилл", "name_en"=>"Mitchellville", "lang"=>"en", "lat"=>36.632742, "lon"=>-86.539204),
"/US/US-TN/monteagle-19918" => array("type"=>"town", "pop"=>1203, "name_en"=>"Monteagle", "lang"=>"en", "lat"=>35.238226, "lon"=>-85.840666),
"/US/US-TN/monterey-19919" => array("type"=>"town", "pop"=>2924, "name_ru"=>"Монтерей", "name_en"=>"Monterey", "lang"=>"en", "lat"=>36.133025, "lon"=>-85.244283),
"/US/US-TN/morrison-19920" => array("type"=>"town", "pop"=>702, "name_ru"=>"Моррисон", "name_en"=>"Morrison", "lang"=>"en", "lat"=>35.602383, "lon"=>-85.918550),
"/US/US-TN/morristown-19921" => array("type"=>"city", "pop"=>27636, "name_ru"=>"Морристаун", "name_en"=>"Morristown", "lang"=>"en", "lat"=>36.214650, "lon"=>-83.295982),
"/US/US-TN/moscow-19922" => array("type"=>"town", "pop"=>548, "name_ru"=>"Москва", "name_en"=>"Moscow", "lang"=>"en", "lat"=>35.061349, "lon"=>-89.396847),
"/US/US-TN/mosheim-19923" => array("type"=>"town", "pop"=>2090, "name_en"=>"Mosheim", "lang"=>"en", "lat"=>36.192241, "lon"=>-82.956825),
"/US/US-TN/mount-carmel-19925" => array("type"=>"town", "pop"=>5462, "name_ru"=>"Маунт Кармел", "name_en"=>"Mount Carmel", "lang"=>"en", "lat"=>36.560867, "lon"=>-82.658681),
"/US/US-TN/mount-juliet-19926" => array("type"=>"city", "pop"=>21718, "name_en"=>"Mount Juliet", "lang"=>"en", "lat"=>36.206162, "lon"=>-86.522468),
"/US/US-TN/mount-pleasant-19927" => array("type"=>"city", "pop"=>4443, "name_ru"=>"Маунт Плезант", "name_en"=>"Mount Pleasant", "lang"=>"en", "lat"=>35.551312, "lon"=>-87.191198),
"/US/US-TN/mountain-city-19924" => array("type"=>"town", "pop"=>2407, "name_ru"=>"Маунтин Сити", "name_en"=>"Mountain City", "lang"=>"en", "lat"=>36.471818, "lon"=>-81.805303),
"/US/US-TN/munford-19928" => array("type"=>"town", "pop"=>6634, "name_en"=>"Munford", "lang"=>"en", "lat"=>35.449803, "lon"=>-89.804208),
"/US/US-TN/murfreesboro-19929" => array("type"=>"city", "pop"=>101753, "name_ru"=>"Мерфрисборо", "name_en"=>"Murfreesboro", "lang"=>"en", "lat"=>35.853391, "lon"=>-86.394592),
"/US/US-TN/nashville-davidson-(remainder)-19930" => array("type"=>"(remainder)", "pop"=>488374, "name_en"=>"Nashville-Davidson (remainder)", "lang"=>"en", "lat"=>36.171550, "lon"=>-86.784829),
"/US/US-TN/new-hope-19932" => array("type"=>"city", "pop"=>1048, "name_ru"=>"Нью Хоуп", "name_en"=>"New Hope", "lang"=>"en", "lat"=>35.005064, "lon"=>-85.658566),
"/US/US-TN/new-johnsonville-19933" => array("type"=>"city", "pop"=>1953, "name_en"=>"New Johnsonville", "lang"=>"en", "lat"=>36.016867, "lon"=>-87.957386),
"/US/US-TN/new-market-19934" => array("type"=>"town", "pop"=>1326, "name_ru"=>"Нью Маркет", "name_en"=>"New Market", "lang"=>"en", "lat"=>36.098879, "lon"=>-83.552433),
"/US/US-TN/new-tazewell-19936" => array("type"=>"town", "pop"=>2882, "name_en"=>"New Tazewell", "lang"=>"en", "lat"=>36.436009, "lon"=>-83.605851),
"/US/US-TN/newbern-19931" => array("type"=>"town", "pop"=>3141, "name_ru"=>"Ньюберн", "name_en"=>"Newbern", "lang"=>"en", "lat"=>36.116390, "lon"=>-89.268029),
"/US/US-TN/newport-19935" => array("type"=>"city", "pop"=>7515, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>35.960189, "lon"=>-83.191236),
"/US/US-TN/niota-19937" => array("type"=>"city", "pop"=>800, "name_en"=>"Niota", "lang"=>"en", "lat"=>35.514585, "lon"=>-84.549316),
"/US/US-TN/nolensville-19938" => array("type"=>"cdp", "pop"=>2680, "name_en"=>"Nolensville", "lang"=>"en", "lat"=>35.947904, "lon"=>-86.647984),
"/US/US-TN/normandy-19939" => array("type"=>"town", "pop"=>152, "name_ru"=>"Норманди", "name_en"=>"Normandy", "lang"=>"en", "lat"=>35.451356, "lon"=>-86.258079),
"/US/US-TN/norris-19940" => array("type"=>"city", "pop"=>1477, "name_ru"=>"Норрис", "name_en"=>"Norris", "lang"=>"en", "lat"=>36.213097, "lon"=>-84.061032),
"/US/US-TN/oak-grove-19942" => array("type"=>"cdp", "pop"=>3498, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>36.421350, "lon"=>-82.427982),
"/US/US-TN/oak-hill-19943" => array("type"=>"city", "pop"=>5389, "name_ru"=>"Оук Хилл", "name_en"=>"Oak Hill", "lang"=>"en", "lat"=>36.076900, "lon"=>-86.786877),
"/US/US-TN/oak-ridge-19945" => array("type"=>"city", "pop"=>27677, "name_ru"=>"Оук Ридж", "name_en"=>"Oak Ridge", "lang"=>"en", "lat"=>35.962739, "lon"=>-84.296156),
"/US/US-TN/oakdale-19941" => array("type"=>"town", "pop"=>235, "name_ru"=>"Оукдейл", "name_en"=>"Oakdale", "lang"=>"en", "lat"=>35.988836, "lon"=>-84.557940),
"/US/US-TN/oakland-19944" => array("type"=>"town", "pop"=>5038, "name_ru"=>"Оукленд", "name_en"=>"Oakland", "lang"=>"en", "lat"=>35.224222, "lon"=>-89.513410),
"/US/US-TN/obion-19946" => array("type"=>"town", "pop"=>1070, "name_en"=>"Obion", "lang"=>"en", "lat"=>36.260798, "lon"=>-89.191955),
"/US/US-TN/oliver-springs-19947" => array("type"=>"town", "pop"=>3326, "name_en"=>"Oliver Springs", "lang"=>"en", "lat"=>36.036449, "lon"=>-84.329521),
"/US/US-TN/oneida-19948" => array("type"=>"town", "pop"=>3857, "name_ru"=>"Онида", "name_en"=>"Oneida", "lang"=>"en", "lat"=>36.503923, "lon"=>-84.515026),
"/US/US-TN/ooltewah-19949" => array("type"=>"cdp", "pop"=>4903, "name_en"=>"Ooltewah", "lang"=>"en", "lat"=>35.062517, "lon"=>-85.091117),
"/US/US-TN/orlinda-19950" => array("type"=>"town", "pop"=>670, "name_en"=>"Orlinda", "lang"=>"en", "lat"=>36.600739, "lon"=>-86.671861),
"/US/US-TN/orme-19951" => array("type"=>"town", "pop"=>119, "name_en"=>"Orme", "lang"=>"en", "lat"=>35.011857, "lon"=>-85.806449),
"/US/US-TN/palmer-19952" => array("type"=>"town", "pop"=>705, "name_ru"=>"Палмер", "name_en"=>"Palmer", "lang"=>"en", "lat"=>35.357943, "lon"=>-85.565150),
"/US/US-TN/paris-19953" => array("type"=>"city", "pop"=>9966, "name_ru"=>"Парис", "name_en"=>"Paris", "lang"=>"en", "lat"=>36.300229, "lon"=>-88.307321),
"/US/US-TN/parker's-cross-roads-19954" => array("type"=>"city", "pop"=>161, "name_en"=>"Parker's Cross Roads", "lang"=>"en", "lat"=>35.788844, "lon"=>-88.392712),
"/US/US-TN/parrottsville-19955" => array("type"=>"town", "pop"=>221, "name_en"=>"Parrottsville", "lang"=>"en", "lat"=>36.009293, "lon"=>-83.091309),
"/US/US-TN/parsons-19956" => array("type"=>"town", "pop"=>2348, "name_ru"=>"Парсонс", "name_en"=>"Parsons", "lang"=>"en", "lat"=>35.653260, "lon"=>-88.124987),
"/US/US-TN/pegram-19957" => array("type"=>"town", "pop"=>2127, "name_en"=>"Pegram", "lang"=>"en", "lat"=>36.104300, "lon"=>-87.056655),
"/US/US-TN/petersburg-19958" => array("type"=>"town", "pop"=>603, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>35.317126, "lon"=>-86.640119),
"/US/US-TN/philadelphia-19959" => array("type"=>"city", "pop"=>603, "name_ru"=>"Филадельфия", "name_en"=>"Philadelphia", "lang"=>"en", "lat"=>35.679138, "lon"=>-84.400421),
"/US/US-TN/pigeon-forge-19960" => array("type"=>"city", "pop"=>6204, "name_en"=>"Pigeon Forge", "lang"=>"en", "lat"=>35.792663, "lon"=>-83.558136),
"/US/US-TN/pikeville-19961" => array("type"=>"town", "pop"=>1896, "name_ru"=>"Пайквилл", "name_en"=>"Pikeville", "lang"=>"en", "lat"=>35.606704, "lon"=>-85.195791),
"/US/US-TN/pine-crest-19962" => array("type"=>"cdp", "pop"=>3821, "name_en"=>"Pine Crest", "lang"=>"en", "lat"=>36.297300, "lon"=>-82.305536),
"/US/US-TN/piperton-19963" => array("type"=>"city", "pop"=>1246, "name_en"=>"Piperton", "lang"=>"en", "lat"=>35.026461, "lon"=>-89.630216),
"/US/US-TN/pittman-center-19964" => array("type"=>"town", "pop"=>672, "name_en"=>"Pittman Center", "lang"=>"en", "lat"=>35.758781, "lon"=>-83.388287),
"/US/US-TN/pleasant-hill-19965" => array("type"=>"town", "pop"=>633, "name_ru"=>"Плезант Хилл", "name_en"=>"Pleasant Hill", "lang"=>"en", "lat"=>35.974978, "lon"=>-85.197191),
"/US/US-TN/portland-19966" => array("type"=>"town", "pop"=>11154, "name_ru"=>"Портленд", "name_en"=>"Portland", "lang"=>"en", "lat"=>36.585930, "lon"=>-86.516582),
"/US/US-TN/powell-19967" => array("type"=>"cdp", "pop"=>7534, "name_ru"=>"Пауэлл", "name_en"=>"Powell", "lang"=>"en", "lat"=>36.033680, "lon"=>-84.028276),
"/US/US-TN/powells-crossroads-19968" => array("type"=>"town", "pop"=>1273, "name_en"=>"Powells Crossroads", "lang"=>"en", "lat"=>35.183464, "lon"=>-85.482934),
"/US/US-TN/pulaski-19969" => array("type"=>"city", "pop"=>7847, "name_ru"=>"Пуласки", "name_en"=>"Pulaski", "lang"=>"en", "lat"=>35.195690, "lon"=>-87.034323),
"/US/US-TN/puryear-19970" => array("type"=>"city", "pop"=>675, "name_en"=>"Puryear", "lang"=>"en", "lat"=>36.442477, "lon"=>-88.334807),
"/US/US-TN/ramer-19971" => array("type"=>"city", "pop"=>367, "name_en"=>"Ramer", "lang"=>"en", "lat"=>35.070293, "lon"=>-88.617749),
"/US/US-TN/red-bank-19972" => array("type"=>"city", "pop"=>11573, "name_en"=>"Red Bank", "lang"=>"en", "lat"=>35.111400, "lon"=>-85.294995),
"/US/US-TN/red-boiling-springs-19973" => array("type"=>"city", "pop"=>1111, "name_en"=>"Red Boiling Springs", "lang"=>"en", "lat"=>36.531716, "lon"=>-85.850686),
"/US/US-TN/ridgely-19974" => array("type"=>"town", "pop"=>1520, "name_ru"=>"Риджли", "name_en"=>"Ridgely", "lang"=>"en", "lat"=>36.264492, "lon"=>-89.482594),
"/US/US-TN/ridgeside-19975" => array("type"=>"city", "pop"=>394, "name_en"=>"Ridgeside", "lang"=>"en", "lat"=>35.035850, "lon"=>-85.246945),
"/US/US-TN/ridgetop-19976" => array("type"=>"town", "pop"=>73, "name_en"=>"Ridgetop", "lang"=>"en", "lat"=>36.396750, "lon"=>-86.773352),
"/US/US-TN/ripley-19977" => array("type"=>"city", "pop"=>7636, "name_ru"=>"Рипли", "name_en"=>"Ripley", "lang"=>"en", "lat"=>35.745923, "lon"=>-89.534065),
"/US/US-TN/rives-19978" => array("type"=>"town", "pop"=>315, "name_en"=>"Rives", "lang"=>"en", "lat"=>36.357060, "lon"=>-89.049679),
"/US/US-TN/roan-mountain-19979" => array("type"=>"cdp", "pop"=>1220, "name_en"=>"Roan Mountain", "lang"=>"en", "lat"=>36.183689, "lon"=>-82.073854),
"/US/US-TN/rockford-19980" => array("type"=>"city", "pop"=>814, "name_ru"=>"Рокфорд", "name_en"=>"Rockford", "lang"=>"en", "lat"=>35.829785, "lon"=>-83.935738),
"/US/US-TN/rockwood-19981" => array("type"=>"city", "pop"=>5541, "name_ru"=>"Роквуд", "name_en"=>"Rockwood", "lang"=>"en", "lat"=>35.874075, "lon"=>-84.675192),
"/US/US-TN/rogersville-19982" => array("type"=>"town", "pop"=>4346, "name_ru"=>"Роджерсвилл", "name_en"=>"Rogersville", "lang"=>"en", "lat"=>36.406739, "lon"=>-83.005180),
"/US/US-TN/rossville-19983" => array("type"=>"town", "pop"=>561, "name_ru"=>"Россвилл", "name_en"=>"Rossville", "lang"=>"en", "lat"=>35.043845, "lon"=>-89.542852),
"/US/US-TN/rural-hill-19984" => array("type"=>"cdp", "pop"=>1329, "name_en"=>"Rural Hill", "lang"=>"en", "lat"=>36.111400, "lon"=>-86.511292),
"/US/US-TN/rutherford-19985" => array("type"=>"town", "pop"=>1255, "name_en"=>"Rutherford", "lang"=>"en", "lat"=>36.125416, "lon"=>-88.992789),
"/US/US-TN/rutledge-19986" => array("type"=>"town", "pop"=>1284, "name_ru"=>"Рутледж", "name_en"=>"Rutledge", "lang"=>"en", "lat"=>36.274210, "lon"=>-83.524830),
"/US/US-TN/saltillo-19988" => array("type"=>"town", "pop"=>349, "name_ru"=>"Салтилло", "name_en"=>"Saltillo", "lang"=>"en", "lat"=>35.377833, "lon"=>-88.210500),
"/US/US-TN/samburg-19989" => array("type"=>"town", "pop"=>249, "name_en"=>"Samburg", "lang"=>"en", "lat"=>36.382283, "lon"=>-89.351868),
"/US/US-TN/sardis-19990" => array("type"=>"town", "pop"=>461, "name_ru"=>"Сардис", "name_en"=>"Sardis", "lang"=>"en", "lat"=>35.443072, "lon"=>-88.294295),
"/US/US-TN/saulsbury-19991" => array("type"=>"town", "pop"=>95, "name_en"=>"Saulsbury", "lang"=>"en", "lat"=>35.049188, "lon"=>-89.088794),
"/US/US-TN/savannah-19992" => array("type"=>"city", "pop"=>7320, "name_ru"=>"Саванна", "name_en"=>"Savannah", "lang"=>"en", "lat"=>35.222924, "lon"=>-88.237181),
"/US/US-TN/scotts-hill-19993" => array("type"=>"town", "pop"=>924, "name_en"=>"Scotts Hill", "lang"=>"en", "lat"=>35.515361, "lon"=>-88.250581),
"/US/US-TN/selmer-19994" => array("type"=>"town", "pop"=>4697, "name_en"=>"Selmer", "lang"=>"en", "lat"=>35.171808, "lon"=>-88.589954),
"/US/US-TN/sevierville-19995" => array("type"=>"town", "pop"=>16846, "name_en"=>"Sevierville", "lang"=>"en", "lat"=>35.892033, "lon"=>-83.578232),
"/US/US-TN/sewanee-19996" => array("type"=>"cdp", "pop"=>2128, "name_en"=>"Sewanee", "lang"=>"en", "lat"=>35.201474, "lon"=>-85.920515),
"/US/US-TN/seymour-19997" => array("type"=>"cdp", "pop"=>7026, "name_ru"=>"Сеймур", "name_en"=>"Seymour", "lang"=>"en", "lat"=>35.874675, "lon"=>-83.773896),
"/US/US-TN/sharon-19998" => array("type"=>"town", "pop"=>915, "name_ru"=>"Шэрон", "name_en"=>"Sharon", "lang"=>"en", "lat"=>36.233204, "lon"=>-88.824940),
"/US/US-TN/shelbyville-19999" => array("type"=>"city", "pop"=>19669, "name_ru"=>"Шелбивилл", "name_en"=>"Shelbyville", "lang"=>"en", "lat"=>35.490410, "lon"=>-86.446708),
"/US/US-TN/signal-mountain-20000" => array("type"=>"town", "pop"=>7109, "name_en"=>"Signal Mountain", "lang"=>"en", "lat"=>35.130433, "lon"=>-85.342100),
"/US/US-TN/silerton-20001" => array("type"=>"town", "pop"=>57, "name_en"=>"Silerton", "lang"=>"en", "lat"=>35.340495, "lon"=>-88.804887),
"/US/US-TN/slayden-20002" => array("type"=>"town", "pop"=>190, "name_en"=>"Slayden", "lang"=>"en", "lat"=>36.293219, "lon"=>-87.467819),
"/US/US-TN/smithville-20003" => array("type"=>"town", "pop"=>4389, "name_ru"=>"Смитвилл", "name_en"=>"Smithville", "lang"=>"en", "lat"=>35.957128, "lon"=>-85.820771),
"/US/US-TN/smyrna-20004" => array("type"=>"town", "pop"=>37867, "name_ru"=>"Смирна", "name_en"=>"Smyrna", "lang"=>"en", "lat"=>35.982617, "lon"=>-86.517241),
"/US/US-TN/sneedville-20005" => array("type"=>"town", "pop"=>1312, "name_en"=>"Sneedville", "lang"=>"en", "lat"=>36.537597, "lon"=>-83.211087),
"/US/US-TN/soddy-daisy-20006" => array("type"=>"city", "pop"=>12511, "name_en"=>"Soddy-Daisy", "lang"=>"en", "lat"=>35.262590, "lon"=>-85.172192),
"/US/US-TN/somerville-20007" => array("type"=>"town", "pop"=>2935, "name_ru"=>"Сомервилл", "name_en"=>"Somerville", "lang"=>"en", "lat"=>35.237527, "lon"=>-89.350683),
"/US/US-TN/south-carthage-20008" => array("type"=>"town", "pop"=>1357, "name_en"=>"South Carthage", "lang"=>"en", "lat"=>36.240704, "lon"=>-85.958026),
"/US/US-TN/south-cleveland-20009" => array("type"=>"cdp", "pop"=>5372, "name_en"=>"South Cleveland", "lang"=>"en", "lat"=>35.113212, "lon"=>-84.913232),
"/US/US-TN/south-fulton-20010" => array("type"=>"city", "pop"=>2401, "name_en"=>"South Fulton", "lang"=>"en", "lat"=>36.492021, "lon"=>-88.884944),
"/US/US-TN/south-pittsburg-20011" => array("type"=>"city", "pop"=>3140, "name_en"=>"South Pittsburg", "lang"=>"en", "lat"=>35.015842, "lon"=>-85.711423),
"/US/US-TN/sparta-20012" => array("type"=>"city", "pop"=>4938, "name_ru"=>"Спарта", "name_en"=>"Sparta", "lang"=>"en", "lat"=>35.934471, "lon"=>-85.473149),
"/US/US-TN/spencer-20013" => array("type"=>"town", "pop"=>1698, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>35.742330, "lon"=>-85.456689),
"/US/US-TN/spring-city-20014" => array("type"=>"town", "pop"=>2021, "name_en"=>"Spring City", "lang"=>"en", "lat"=>35.687389, "lon"=>-84.864133),
"/US/US-TN/spring-hill-20016" => array("type"=>"town", "pop"=>26230, "name_ru"=>"Спринг Хилл", "name_en"=>"Spring Hill", "lang"=>"en", "lat"=>35.741966, "lon"=>-86.923732),
"/US/US-TN/springfield-20015" => array("type"=>"city", "pop"=>17261, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>36.499463, "lon"=>-86.878719),
"/US/US-TN/spurgeon-20017" => array("type"=>"cdp", "pop"=>3149, "name_ru"=>"Сперджен", "name_en"=>"Spurgeon", "lang"=>"en", "lat"=>36.443133, "lon"=>-82.461448),
"/US/US-TN/st.-joseph-19987" => array("type"=>"city", "pop"=>856, "name_ru"=>"Сент Джозеф", "name_en"=>"St. Joseph", "lang"=>"en", "lat"=>35.029541, "lon"=>-87.506149),
"/US/US-TN/stanton-20018" => array("type"=>"town", "pop"=>581, "name_ru"=>"Стэнтон", "name_en"=>"Stanton", "lang"=>"en", "lat"=>35.462018, "lon"=>-89.400847),
"/US/US-TN/stantonville-20019" => array("type"=>"town", "pop"=>325, "name_en"=>"Stantonville", "lang"=>"en", "lat"=>35.158060, "lon"=>-88.428312),
"/US/US-TN/surgoinsville-20020" => array("type"=>"town", "pop"=>1811, "name_en"=>"Surgoinsville", "lang"=>"en", "lat"=>36.473323, "lon"=>-82.857916),
"/US/US-TN/sweetwater-20021" => array("type"=>"city", "pop"=>6645, "name_ru"=>"Свитвотер", "name_en"=>"Sweetwater", "lang"=>"en", "lat"=>35.600324, "lon"=>-84.468307),
"/US/US-TN/tazewell-20022" => array("type"=>"town", "pop"=>2223, "name_en"=>"Tazewell", "lang"=>"en", "lat"=>36.458168, "lon"=>-83.575960),
"/US/US-TN/tellico-plains-20023" => array("type"=>"town", "pop"=>964, "name_en"=>"Tellico Plains", "lang"=>"en", "lat"=>35.366498, "lon"=>-84.299035),
"/US/US-TN/tennessee-ridge-20024" => array("type"=>"city", "pop"=>1338, "name_en"=>"Tennessee Ridge", "lang"=>"en", "lat"=>36.319835, "lon"=>-87.763805),
"/US/US-TN/tiptonville-20025" => array("type"=>"town", "pop"=>4008, "name_en"=>"Tiptonville", "lang"=>"en", "lat"=>36.377264, "lon"=>-89.472306),
"/US/US-TN/toone-20026" => array("type"=>"town", "pop"=>350, "name_en"=>"Toone", "lang"=>"en", "lat"=>35.353046, "lon"=>-88.953521),
"/US/US-TN/townsend-20027" => array("type"=>"city", "pop"=>272, "name_ru"=>"Таунсенд", "name_en"=>"Townsend", "lang"=>"en", "lat"=>35.676150, "lon"=>-83.751886),
"/US/US-TN/tracy-city-20028" => array("type"=>"town", "pop"=>1652, "name_en"=>"Tracy City", "lang"=>"en", "lat"=>35.259016, "lon"=>-85.751856),
"/US/US-TN/trenton-20029" => array("type"=>"city", "pop"=>4518, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>35.972118, "lon"=>-88.939558),
"/US/US-TN/trezevant-20030" => array("type"=>"town", "pop"=>882, "name_en"=>"Trezevant", "lang"=>"en", "lat"=>36.013485, "lon"=>-88.619983),
"/US/US-TN/trimble-20031" => array("type"=>"town", "pop"=>718, "name_en"=>"Trimble", "lang"=>"en", "lat"=>36.202159, "lon"=>-89.188944),
"/US/US-TN/troy-20032" => array("type"=>"town", "pop"=>1208, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>36.339535, "lon"=>-89.161603),
"/US/US-TN/tullahoma-20033" => array("type"=>"city", "pop"=>18579, "name_en"=>"Tullahoma", "lang"=>"en", "lat"=>35.371234, "lon"=>-86.215536),
"/US/US-TN/tusculum-20034" => array("type"=>"city", "pop"=>2277, "name_en"=>"Tusculum", "lang"=>"en", "lat"=>36.175561, "lon"=>-82.750884),
"/US/US-TN/union-city-20035" => array("type"=>"city", "pop"=>10569, "name_ru"=>"Юнион Сити", "name_en"=>"Union City", "lang"=>"en", "lat"=>36.422524, "lon"=>-89.050961),
"/US/US-TN/vanleer-20036" => array("type"=>"town", "pop"=>325, "name_en"=>"Vanleer", "lang"=>"en", "lat"=>36.237388, "lon"=>-87.444796),
"/US/US-TN/viola-20037" => array("type"=>"town", "pop"=>133, "name_ru"=>"Виола", "name_en"=>"Viola", "lang"=>"en", "lat"=>35.538027, "lon"=>-85.860631),
"/US/US-TN/vonore-20038" => array("type"=>"town", "pop"=>1492, "name_en"=>"Vonore", "lang"=>"en", "lat"=>35.596776, "lon"=>-84.231890),
"/US/US-TN/walden-20039" => array("type"=>"town", "pop"=>2078, "name_ru"=>"Уолден", "name_en"=>"Walden", "lang"=>"en", "lat"=>35.164000, "lon"=>-85.304785),
"/US/US-TN/walnut-hill-20040" => array("type"=>"cdp", "pop"=>3332, "name_ru"=>"Уолнат Хилл", "name_en"=>"Walnut Hill", "lang"=>"en", "lat"=>36.572557, "lon"=>-82.259212),
"/US/US-TN/walterhill-20041" => array("type"=>"cdp", "pop"=>1043, "name_en"=>"Walterhill", "lang"=>"en", "lat"=>35.962785, "lon"=>-86.379219),
"/US/US-TN/wartburg-20042" => array("type"=>"city", "pop"=>928, "name_en"=>"Wartburg", "lang"=>"en", "lat"=>36.103385, "lon"=>-84.588208),
"/US/US-TN/wartrace-20043" => array("type"=>"town", "pop"=>589, "name_en"=>"Wartrace", "lang"=>"en", "lat"=>35.527758, "lon"=>-86.332884),
"/US/US-TN/watauga-20044" => array("type"=>"city", "pop"=>425, "name_en"=>"Watauga", "lang"=>"en", "lat"=>36.368050, "lon"=>-82.292305),
"/US/US-TN/watertown-20045" => array("type"=>"city", "pop"=>1419, "name_ru"=>"Уотертаун", "name_en"=>"Watertown", "lang"=>"en", "lat"=>36.099637, "lon"=>-86.139526),
"/US/US-TN/waverly-20046" => array("type"=>"city", "pop"=>4186, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>36.092365, "lon"=>-87.787785),
"/US/US-TN/waynesboro-20047" => array("type"=>"city", "pop"=>2139, "name_ru"=>"Уэйнсборо", "name_en"=>"Waynesboro", "lang"=>"en", "lat"=>35.322759, "lon"=>-87.757787),
"/US/US-TN/westmoreland-20048" => array("type"=>"town", "pop"=>2189, "name_ru"=>"Вэстморленд", "name_en"=>"Westmoreland", "lang"=>"en", "lat"=>36.563768, "lon"=>-86.246881),
"/US/US-TN/white-bluff-20049" => array("type"=>"town", "pop"=>2513, "name_en"=>"White Bluff", "lang"=>"en", "lat"=>36.105939, "lon"=>-87.213469),
"/US/US-TN/white-house-20050" => array("type"=>"city", "pop"=>10048, "name_en"=>"White House", "lang"=>"en", "lat"=>36.463240, "lon"=>-86.667812),
"/US/US-TN/white-pine-20051" => array("type"=>"town", "pop"=>2143, "name_en"=>"White Pine", "lang"=>"en", "lat"=>36.099336, "lon"=>-83.297730),
"/US/US-TN/whiteville-20052" => array("type"=>"town", "pop"=>4464, "name_en"=>"Whiteville", "lang"=>"en", "lat"=>35.327586, "lon"=>-89.147662),
"/US/US-TN/whitwell-20053" => array("type"=>"city", "pop"=>1603, "name_en"=>"Whitwell", "lang"=>"en", "lat"=>35.201464, "lon"=>-85.518689),
"/US/US-TN/wildwood-lake-20054" => array("type"=>"cdp", "pop"=>2680, "name_en"=>"Wildwood Lake", "lang"=>"en", "lat"=>35.091902, "lon"=>-84.854564),
"/US/US-TN/williston-20055" => array("type"=>"city", "pop"=>350, "name_ru"=>"Уиллистон", "name_en"=>"Williston", "lang"=>"en", "lat"=>35.159357, "lon"=>-89.376007),
"/US/US-TN/winchester-20056" => array("type"=>"city", "pop"=>7862, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>35.188400, "lon"=>-86.108524),
"/US/US-TN/winfield-20057" => array("type"=>"town", "pop"=>993, "name_ru"=>"Винфилд", "name_en"=>"Winfield", "lang"=>"en", "lat"=>36.561810, "lon"=>-84.450488),
"/US/US-TN/woodbury-20058" => array("type"=>"town", "pop"=>2610, "name_ru"=>"Вудбери", "name_en"=>"Woodbury", "lang"=>"en", "lat"=>35.824836, "lon"=>-86.070516),
"/US/US-TN/woodland-mills-20059" => array("type"=>"city", "pop"=>279, "name_en"=>"Woodland Mills", "lang"=>"en", "lat"=>36.476314, "lon"=>-89.112644),
"/US/US-TN/yorkville-20060" => array("type"=>"city", "pop"=>288, "name_ru"=>"Йорквилл", "name_en"=>"Yorkville", "lang"=>"en", "lat"=>36.098669, "lon"=>-89.118954),
"/US/US-TX" => array("type"=>"state", "pop"=>24326974, "name_ru"=>"Техас", "name_en"=>"Texas", "lang"=>"en"),
"/US/US-TX/abbott-20061" => array("type"=>"city", "pop"=>329, "name_en"=>"Abbott", "lang"=>"en", "lat"=>31.883698, "lon"=>-97.077409),
"/US/US-TX/abernathy-20062" => array("type"=>"city", "pop"=>2732, "name_en"=>"Abernathy", "lang"=>"en", "lat"=>33.831761, "lon"=>-101.844630),
"/US/US-TX/abilene-20063" => array("type"=>"city", "pop"=>116484, "name_ru"=>"Абилин", "name_en"=>"Abilene", "lang"=>"en", "lat"=>32.454500, "lon"=>-99.738973),
"/US/US-TX/abram-perezville-20064" => array("type"=>"cdp", "pop"=>3999, "name_en"=>"Abram-Perezville", "lang"=>"en", "lat"=>26.222798, "lon"=>-98.401174),
"/US/US-TX/ackerly-20065" => array("type"=>"city", "pop"=>225, "name_en"=>"Ackerly", "lang"=>"en", "lat"=>32.524613, "lon"=>-101.715762),
"/US/US-TX/addison-20066" => array("type"=>"city", "pop"=>14964, "name_ru"=>"Эддиссон", "name_en"=>"Addison", "lang"=>"en", "lat"=>32.961167, "lon"=>-96.837234),
"/US/US-TX/adrian-20067" => array("type"=>"city", "pop"=>149, "name_ru"=>"Эдриан", "name_en"=>"Adrian", "lang"=>"en", "lat"=>35.275933, "lon"=>-102.663907),
"/US/US-TX/agua-dulce-20068" => array("type"=>"city", "pop"=>713, "name_en"=>"Agua Dulce", "lang"=>"en", "lat"=>27.782774, "lon"=>-97.909756),
"/US/US-TX/alamo-20069" => array("type"=>"city", "pop"=>16608, "name_ru"=>"Аламо", "name_en"=>"Alamo", "lang"=>"en", "lat"=>26.188600, "lon"=>-98.118520),
"/US/US-TX/alamo-heights-20070" => array("type"=>"city", "pop"=>7353, "name_en"=>"Alamo Heights", "lang"=>"en", "lat"=>29.482650, "lon"=>-98.466974),
"/US/US-TX/alba-20071" => array("type"=>"town", "pop"=>497, "name_ru"=>"Альба", "name_en"=>"Alba", "lang"=>"en", "lat"=>32.791375, "lon"=>-95.635154),
"/US/US-TX/albany-20072" => array("type"=>"city", "pop"=>1795, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>32.726731, "lon"=>-99.295636),
"/US/US-TX/aldine-20073" => array("type"=>"cdp", "pop"=>11133, "name_en"=>"Aldine", "lang"=>"en", "lat"=>29.913584, "lon"=>-95.375987),
"/US/US-TX/aledo-20074" => array("type"=>"city", "pop"=>2858, "name_ru"=>"Аледо", "name_en"=>"Aledo", "lang"=>"en", "lat"=>32.696046, "lon"=>-97.603840),
"/US/US-TX/alice-20075" => array("type"=>"city", "pop"=>19797, "name_en"=>"Alice", "lang"=>"en", "lat"=>27.755238, "lon"=>-98.065489),
"/US/US-TX/allen-20076" => array("type"=>"city", "pop"=>81268, "name_ru"=>"Аллен", "name_en"=>"Allen", "lang"=>"en", "lat"=>33.109683, "lon"=>-96.669479),
"/US/US-TX/alma-20077" => array("type"=>"town", "pop"=>321, "name_ru"=>"Алма", "name_en"=>"Alma", "lang"=>"en", "lat"=>32.285352, "lon"=>-96.544710),
"/US/US-TX/alpine-20078" => array("type"=>"city", "pop"=>6280, "name_ru"=>"Алпайн", "name_en"=>"Alpine", "lang"=>"en", "lat"=>30.363206, "lon"=>-103.664953),
"/US/US-TX/alto-20079" => array("type"=>"town", "pop"=>1177, "name_ru"=>"Алто", "name_en"=>"Alto", "lang"=>"en", "lat"=>31.650441, "lon"=>-95.070858),
"/US/US-TX/alton-20080" => array("type"=>"city", "pop"=>11523, "name_ru"=>"Альтон", "name_en"=>"Alton", "lang"=>"en", "lat"=>26.283959, "lon"=>-98.305648),
"/US/US-TX/alvarado-20081" => array("type"=>"city", "pop"=>4188, "name_ru"=>"Альварадо", "name_en"=>"Alvarado", "lang"=>"en", "lat"=>32.410013, "lon"=>-97.211449),
"/US/US-TX/alvin-20082" => array("type"=>"city", "pop"=>22853, "name_ru"=>"Алвин", "name_en"=>"Alvin", "lang"=>"en", "lat"=>29.432076, "lon"=>-95.246020),
"/US/US-TX/alvord-20083" => array("type"=>"town", "pop"=>1448, "name_ru"=>"Алворд", "name_en"=>"Alvord", "lang"=>"en", "lat"=>33.358926, "lon"=>-97.694997),
"/US/US-TX/amarillo-20084" => array("type"=>"city", "pop"=>187236, "name_en"=>"Amarillo", "lang"=>"en", "lat"=>35.202550, "lon"=>-101.818682),
"/US/US-TX/ames-20085" => array("type"=>"city", "pop"=>1137, "name_ru"=>"Эймс", "name_en"=>"Ames", "lang"=>"en", "lat"=>30.046933, "lon"=>-94.737727),
"/US/US-TX/amherst-20086" => array("type"=>"city", "pop"=>717, "name_ru"=>"Амхерст", "name_en"=>"Amherst", "lang"=>"en", "lat"=>34.012425, "lon"=>-102.413960),
"/US/US-TX/anahuac-20087" => array("type"=>"city", "pop"=>1970, "name_en"=>"Anahuac", "lang"=>"en", "lat"=>29.764848, "lon"=>-94.678329),
"/US/US-TX/anderson-mill-20088" => array("type"=>"cdp", "pop"=>9468, "name_en"=>"Anderson Mill", "lang"=>"en", "lat"=>30.454800, "lon"=>-97.805457),
"/US/US-TX/andrews-20089" => array("type"=>"city", "pop"=>10142, "name_ru"=>"Эндрюс", "name_en"=>"Andrews", "lang"=>"en", "lat"=>32.321674, "lon"=>-102.549526),
"/US/US-TX/angleton-20090" => array("type"=>"city", "pop"=>18613, "name_en"=>"Angleton", "lang"=>"en", "lat"=>29.166915, "lon"=>-95.427630),
"/US/US-TX/angus-20091" => array("type"=>"city", "pop"=>411, "name_en"=>"Angus", "lang"=>"en", "lat"=>31.997723, "lon"=>-96.425713),
"/US/US-TX/anna-20092" => array("type"=>"city", "pop"=>1889, "name_ru"=>"Анна", "name_en"=>"Anna", "lang"=>"en", "lat"=>33.352836, "lon"=>-96.549811),
"/US/US-TX/annetta-20093" => array("type"=>"town", "pop"=>1284, "name_en"=>"Annetta", "lang"=>"en", "lat"=>32.695539, "lon"=>-97.657700),
"/US/US-TX/annetta-north-20094" => array("type"=>"town", "pop"=>544, "name_en"=>"Annetta North", "lang"=>"en", "lat"=>32.719869, "lon"=>-97.673552),
"/US/US-TX/annetta-south-20095" => array("type"=>"town", "pop"=>647, "name_en"=>"Annetta South", "lang"=>"en", "lat"=>32.669273, "lon"=>-97.647031),
"/US/US-TX/annona-20096" => array("type"=>"town", "pop"=>249, "name_en"=>"Annona", "lang"=>"en", "lat"=>33.580955, "lon"=>-94.913123),
"/US/US-TX/anson-20097" => array("type"=>"city", "pop"=>2288, "name_en"=>"Anson", "lang"=>"en", "lat"=>32.756434, "lon"=>-99.896476),
"/US/US-TX/anthony-20098" => array("type"=>"town", "pop"=>4344, "name_ru"=>"Энтони", "name_en"=>"Anthony", "lang"=>"en", "lat"=>31.993519, "lon"=>-106.591580),
"/US/US-TX/anton-20099" => array("type"=>"city", "pop"=>1128, "name_en"=>"Anton", "lang"=>"en", "lat"=>33.811253, "lon"=>-102.162045),
"/US/US-TX/appleby-20100" => array("type"=>"city", "pop"=>461, "name_en"=>"Appleby", "lang"=>"en", "lat"=>31.717791, "lon"=>-94.607522),
"/US/US-TX/aquilla-20101" => array("type"=>"city", "pop"=>150, "name_en"=>"Aquilla", "lang"=>"en", "lat"=>31.855170, "lon"=>-97.221597),
"/US/US-TX/aransas-pass-20102" => array("type"=>"city", "pop"=>8810, "name_en"=>"Aransas Pass", "lang"=>"en", "lat"=>27.888140, "lon"=>-97.108667),
"/US/US-TX/archer-city-20103" => array("type"=>"city", "pop"=>1832, "name_en"=>"Archer City", "lang"=>"en", "lat"=>33.593475, "lon"=>-98.625167),
"/US/US-TX/arcola-20104" => array("type"=>"village", "pop"=>1230, "name_ru"=>"Аркола", "name_en"=>"Arcola", "lang"=>"en", "lat"=>29.501103, "lon"=>-95.463541),
"/US/US-TX/argyle-20105" => array("type"=>"city", "pop"=>3525, "name_ru"=>"Аргайл", "name_en"=>"Argyle", "lang"=>"en", "lat"=>33.110487, "lon"=>-97.179876),
"/US/US-TX/arlington-20106" => array("type"=>"city", "pop"=>374417, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>32.694501, "lon"=>-97.127510),
"/US/US-TX/arp-20107" => array("type"=>"city", "pop"=>962, "name_en"=>"Arp", "lang"=>"en", "lat"=>32.227692, "lon"=>-95.056322),
"/US/US-TX/asherton-20108" => array("type"=>"city", "pop"=>1253, "name_en"=>"Asherton", "lang"=>"en", "lat"=>28.446035, "lon"=>-99.760471),
"/US/US-TX/aspermont-20109" => array("type"=>"town", "pop"=>871, "name_en"=>"Aspermont", "lang"=>"en", "lat"=>33.139858, "lon"=>-100.224854),
"/US/US-TX/athens-20110" => array("type"=>"city", "pop"=>12320, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>32.204815, "lon"=>-95.855367),
"/US/US-TX/atlanta-20111" => array("type"=>"city", "pop"=>5475, "name_ru"=>"Атланта", "name_en"=>"Atlanta", "lang"=>"en", "lat"=>33.116436, "lon"=>-94.164287),
"/US/US-TX/aubrey-20112" => array("type"=>"city", "pop"=>2778, "name_ru"=>"Обри", "name_en"=>"Aubrey", "lang"=>"en", "lat"=>33.303895, "lon"=>-96.983109),
"/US/US-TX/aurora-20113" => array("type"=>"town", "pop"=>1023, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>33.058173, "lon"=>-97.515919),
"/US/US-TX/austin-20114" => array("type"=>"city", "pop"=>757688, "name_ru"=>"Остин", "name_en"=>"Austin", "lang"=>"en", "lat"=>30.305880, "lon"=>-97.750522),
"/US/US-TX/austwell-20115" => array("type"=>"city", "pop"=>183, "name_en"=>"Austwell", "lang"=>"en", "lat"=>28.390132, "lon"=>-96.843292),
"/US/US-TX/avery-20116" => array("type"=>"town", "pop"=>421, "name_en"=>"Avery", "lang"=>"en", "lat"=>33.551016, "lon"=>-94.779834),
"/US/US-TX/avinger-20117" => array("type"=>"town", "pop"=>442, "name_en"=>"Avinger", "lang"=>"en", "lat"=>32.898155, "lon"=>-94.554264),
"/US/US-TX/azle-20118" => array("type"=>"city", "pop"=>11347, "name_en"=>"Azle", "lang"=>"en", "lat"=>32.895203, "lon"=>-97.531059),
"/US/US-TX/bacliff-20119" => array("type"=>"cdp", "pop"=>5549, "name_en"=>"Bacliff", "lang"=>"en", "lat"=>29.508160, "lon"=>-94.991899),
"/US/US-TX/bailey's-prairie-20121" => array("type"=>"village", "pop"=>705, "name_en"=>"Bailey's Prairie", "lang"=>"en", "lat"=>29.153444, "lon"=>-95.494268),
"/US/US-TX/bailey-20120" => array("type"=>"city", "pop"=>223, "name_en"=>"Bailey", "lang"=>"en", "lat"=>33.432591, "lon"=>-96.164536),
"/US/US-TX/baird-20122" => array("type"=>"city", "pop"=>1671, "name_en"=>"Baird", "lang"=>"en", "lat"=>32.395816, "lon"=>-99.395537),
"/US/US-TX/balch-springs-20123" => array("type"=>"city", "pop"=>20003, "name_en"=>"Balch Springs", "lang"=>"en", "lat"=>32.718801, "lon"=>-96.616210),
"/US/US-TX/balcones-heights-20124" => array("type"=>"city", "pop"=>2984, "name_en"=>"Balcones Heights", "lang"=>"en", "lat"=>29.489988, "lon"=>-98.551911),
"/US/US-TX/ballinger-20125" => array("type"=>"city", "pop"=>3724, "name_en"=>"Ballinger", "lang"=>"en", "lat"=>31.740665, "lon"=>-99.955502),
"/US/US-TX/balmorhea-20126" => array("type"=>"city", "pop"=>446, "name_en"=>"Balmorhea", "lang"=>"en", "lat"=>30.984093, "lon"=>-103.743460),
"/US/US-TX/bandera-20127" => array("type"=>"city", "pop"=>1217, "name_en"=>"Bandera", "lang"=>"en", "lat"=>29.724813, "lon"=>-99.073844),
"/US/US-TX/bangs-20128" => array("type"=>"city", "pop"=>1567, "name_en"=>"Bangs", "lang"=>"en", "lat"=>31.714522, "lon"=>-99.130484),
"/US/US-TX/bardwell-20129" => array("type"=>"city", "pop"=>634, "name_ru"=>"Бардуэлл", "name_en"=>"Bardwell", "lang"=>"en", "lat"=>32.267210, "lon"=>-96.695200),
"/US/US-TX/barrett-20130" => array("type"=>"cdp", "pop"=>3052, "name_ru"=>"Барретт", "name_en"=>"Barrett", "lang"=>"en", "lat"=>29.873001, "lon"=>-95.061960),
"/US/US-TX/barry-20131" => array("type"=>"city", "pop"=>227, "name_ru"=>"Бэрри", "name_en"=>"Barry", "lang"=>"en", "lat"=>32.098677, "lon"=>-96.637881),
"/US/US-TX/barstow-20132" => array("type"=>"town", "pop"=>399, "name_ru"=>"Барстоу", "name_en"=>"Barstow", "lang"=>"en", "lat"=>31.461358, "lon"=>-103.394964),
"/US/US-TX/bartlett-20133" => array("type"=>"city", "pop"=>1731, "name_ru"=>"Барлет", "name_en"=>"Bartlett", "lang"=>"en", "lat"=>30.794722, "lon"=>-97.431888),
"/US/US-TX/bartonville-20134" => array("type"=>"town", "pop"=>1601, "name_ru"=>"Бартонвилл", "name_en"=>"Bartonville", "lang"=>"en", "lat"=>33.088703, "lon"=>-97.156376),
"/US/US-TX/bastrop-20135" => array("type"=>"city", "pop"=>8378, "name_ru"=>"Бастроп", "name_en"=>"Bastrop", "lang"=>"en", "lat"=>30.113029, "lon"=>-97.305426),
"/US/US-TX/batesville-20136" => array("type"=>"cdp", "pop"=>1313, "name_ru"=>"Батсвилл", "name_en"=>"Batesville", "lang"=>"en", "lat"=>28.947308, "lon"=>-99.614338),
"/US/US-TX/bay-city-20137" => array("type"=>"city", "pop"=>17891, "name_ru"=>"Бэй Сити", "name_en"=>"Bay City", "lang"=>"en", "lat"=>28.978616, "lon"=>-95.963361),
"/US/US-TX/bayou-vista-20138" => array("type"=>"village", "pop"=>1686, "name_ru"=>"Байу Виста", "name_en"=>"Bayou Vista", "lang"=>"en", "lat"=>29.325342, "lon"=>-94.937751),
"/US/US-TX/bayside-20139" => array("type"=>"town", "pop"=>354, "name_en"=>"Bayside", "lang"=>"en", "lat"=>28.095830, "lon"=>-97.210551),
"/US/US-TX/baytown-20140" => array("type"=>"city", "pop"=>70330, "name_en"=>"Baytown", "lang"=>"en", "lat"=>29.748483, "lon"=>-94.968959),
"/US/US-TX/bayview-20141" => array("type"=>"town", "pop"=>535, "name_ru"=>"Бэйвью", "name_en"=>"Bayview", "lang"=>"en", "lat"=>26.124330, "lon"=>-97.398990),
"/US/US-TX/beach-city-20142" => array("type"=>"town", "pop"=>1888, "name_en"=>"Beach City", "lang"=>"en", "lat"=>29.727425, "lon"=>-94.848796),
"/US/US-TX/beasley-20143" => array("type"=>"town", "pop"=>670, "name_en"=>"Beasley", "lang"=>"en", "lat"=>29.497425, "lon"=>-95.917946),
"/US/US-TX/beaumont-20144" => array("type"=>"city", "pop"=>110553, "name_ru"=>"Бомонт", "name_en"=>"Beaumont", "lang"=>"en", "lat"=>30.087767, "lon"=>-94.144463),
"/US/US-TX/beckville-20145" => array("type"=>"city", "pop"=>763, "name_en"=>"Beckville", "lang"=>"en", "lat"=>32.242850, "lon"=>-94.455865),
"/US/US-TX/bedford-20146" => array("type"=>"city", "pop"=>49155, "name_ru"=>"Бедфорд", "name_en"=>"Bedford", "lang"=>"en", "lat"=>32.846083, "lon"=>-97.134420),
"/US/US-TX/bee-cave-20147" => array("type"=>"village", "pop"=>2680, "name_en"=>"Bee Cave", "lang"=>"en", "lat"=>30.307971, "lon"=>-97.955248),
"/US/US-TX/beeville-20148" => array("type"=>"city", "pop"=>12682, "name_en"=>"Beeville", "lang"=>"en", "lat"=>28.405558, "lon"=>-97.749129),
"/US/US-TX/bellaire-20149" => array("type"=>"city", "pop"=>18197, "name_ru"=>"Беллэр", "name_en"=>"Bellaire", "lang"=>"en", "lat"=>29.703800, "lon"=>-95.460774),
"/US/US-TX/bellevue-20150" => array("type"=>"city", "pop"=>377, "name_ru"=>"Белвью", "name_en"=>"Bellevue", "lang"=>"en", "lat"=>33.634212, "lon"=>-98.015930),
"/US/US-TX/bellmead-20151" => array("type"=>"city", "pop"=>9579, "name_en"=>"Bellmead", "lang"=>"en", "lat"=>31.600467, "lon"=>-97.091435),
"/US/US-TX/bells-20152" => array("type"=>"town", "pop"=>1312, "name_en"=>"Bells", "lang"=>"en", "lat"=>33.620022, "lon"=>-96.412946),
"/US/US-TX/bellville-20153" => array("type"=>"city", "pop"=>4415, "name_ru"=>"Беллвилл", "name_en"=>"Bellville", "lang"=>"en", "lat"=>29.946423, "lon"=>-96.258671),
"/US/US-TX/belton-20154" => array("type"=>"city", "pop"=>17876, "name_ru"=>"Белтон", "name_en"=>"Belton", "lang"=>"en", "lat"=>31.068090, "lon"=>-97.463717),
"/US/US-TX/benavides-20155" => array("type"=>"city", "pop"=>1517, "name_en"=>"Benavides", "lang"=>"en", "lat"=>27.597733, "lon"=>-98.409161),
"/US/US-TX/benbrook-20156" => array("type"=>"city", "pop"=>22957, "name_en"=>"Benbrook", "lang"=>"en", "lat"=>32.678401, "lon"=>-97.461200),
"/US/US-TX/benjamin-20157" => array("type"=>"city", "pop"=>213, "name_en"=>"Benjamin", "lang"=>"en", "lat"=>33.583328, "lon"=>-99.793026),
"/US/US-TX/berryville-20158" => array("type"=>"town", "pop"=>984, "name_ru"=>"Берривилл", "name_en"=>"Berryville", "lang"=>"en", "lat"=>32.087412, "lon"=>-95.465000),
"/US/US-TX/bertram-20159" => array("type"=>"city", "pop"=>1408, "name_ru"=>"Бертрам", "name_en"=>"Bertram", "lang"=>"en", "lat"=>30.744215, "lon"=>-98.057429),
"/US/US-TX/beverly-hills-20160" => array("type"=>"city", "pop"=>2047, "name_ru"=>"Беверли Хиллс", "name_en"=>"Beverly Hills", "lang"=>"en", "lat"=>31.521750, "lon"=>-97.156003),
"/US/US-TX/bevil-oaks-20161" => array("type"=>"town", "pop"=>1213, "name_en"=>"Bevil Oaks", "lang"=>"en", "lat"=>30.155448, "lon"=>-94.271560),
"/US/US-TX/big-lake-20162" => array("type"=>"city", "pop"=>2666, "name_ru"=>"Биг Лейк", "name_en"=>"Big Lake", "lang"=>"en", "lat"=>31.193754, "lon"=>-101.458440),
"/US/US-TX/big-sandy-20163" => array("type"=>"town", "pop"=>1360, "name_en"=>"Big Sandy", "lang"=>"en", "lat"=>32.585689, "lon"=>-95.112575),
"/US/US-TX/big-spring-20164" => array("type"=>"city", "pop"=>24270, "name_en"=>"Big Spring", "lang"=>"en", "lat"=>32.239977, "lon"=>-101.478904),
"/US/US-TX/big-wells-20165" => array("type"=>"city", "pop"=>699, "name_en"=>"Big Wells", "lang"=>"en", "lat"=>28.569519, "lon"=>-99.570043),
"/US/US-TX/bishop-20166" => array("type"=>"city", "pop"=>3139, "name_ru"=>"Бишоп", "name_en"=>"Bishop", "lang"=>"en", "lat"=>27.584765, "lon"=>-97.797594),
"/US/US-TX/blackwell-20167" => array("type"=>"town", "pop"=>353, "name_en"=>"Blackwell", "lang"=>"en", "lat"=>32.084533, "lon"=>-100.318617),
"/US/US-TX/blanco-20168" => array("type"=>"city", "pop"=>1557, "name_en"=>"Blanco", "lang"=>"en", "lat"=>30.098444, "lon"=>-98.422318),
"/US/US-TX/blanket-20169" => array("type"=>"town", "pop"=>400, "name_en"=>"Blanket", "lang"=>"en", "lat"=>31.824266, "lon"=>-98.789926),
"/US/US-TX/bloomburg-20170" => array("type"=>"town", "pop"=>361, "name_en"=>"Bloomburg", "lang"=>"en", "lat"=>33.137528, "lon"=>-94.058711),
"/US/US-TX/blooming-grove-20171" => array("type"=>"town", "pop"=>927, "name_en"=>"Blooming Grove", "lang"=>"en", "lat"=>32.091932, "lon"=>-96.716819),
"/US/US-TX/bloomington-20172" => array("type"=>"cdp", "pop"=>1888, "name_ru"=>"Блумингтон", "name_en"=>"Bloomington", "lang"=>"en", "lat"=>28.650543, "lon"=>-96.901995),
"/US/US-TX/blossom-20173" => array("type"=>"city", "pop"=>1465, "name_en"=>"Blossom", "lang"=>"en", "lat"=>33.661988, "lon"=>-95.382226),
"/US/US-TX/blue-mound-20174" => array("type"=>"city", "pop"=>2365, "name_ru"=>"Блю Маунд", "name_en"=>"Blue Mound", "lang"=>"en", "lat"=>32.853764, "lon"=>-97.337217),
"/US/US-TX/blue-ridge-20175" => array("type"=>"town", "pop"=>970, "name_ru"=>"Блу Ридж", "name_en"=>"Blue Ridge", "lang"=>"en", "lat"=>33.298370, "lon"=>-96.399939),
"/US/US-TX/blum-20176" => array("type"=>"town", "pop"=>445, "name_en"=>"Blum", "lang"=>"en", "lat"=>32.141816, "lon"=>-97.396782),
"/US/US-TX/boerne-20177" => array("type"=>"city", "pop"=>10283, "name_en"=>"Boerne", "lang"=>"en", "lat"=>29.801532, "lon"=>-98.738549),
"/US/US-TX/bogata-20178" => array("type"=>"town", "pop"=>1235, "name_en"=>"Bogata", "lang"=>"en", "lat"=>33.469185, "lon"=>-95.213386),
"/US/US-TX/boling-iago-20179" => array("type"=>"cdp", "pop"=>1119, "name_en"=>"Boling-Iago", "lang"=>"en", "lat"=>29.260502, "lon"=>-95.948582),
"/US/US-TX/bonham-20180" => array("type"=>"city", "pop"=>10660, "name_en"=>"Bonham", "lang"=>"en", "lat"=>33.589343, "lon"=>-96.180007),
"/US/US-TX/bonney-20181" => array("type"=>"village", "pop"=>441, "name_en"=>"Bonney", "lang"=>"en", "lat"=>29.301706, "lon"=>-95.458473),
"/US/US-TX/booker-20182" => array("type"=>"town", "pop"=>1280, "name_en"=>"Booker", "lang"=>"en", "lat"=>36.455947, "lon"=>-100.539559),
"/US/US-TX/borger-20183" => array("type"=>"city", "pop"=>12677, "name_en"=>"Borger", "lang"=>"en", "lat"=>35.659266, "lon"=>-101.400233),
"/US/US-TX/bovina-20184" => array("type"=>"city", "pop"=>1697, "name_en"=>"Bovina", "lang"=>"en", "lat"=>34.516209, "lon"=>-102.888317),
"/US/US-TX/bowie-20185" => array("type"=>"city", "pop"=>5580, "name_ru"=>"Боуи", "name_en"=>"Bowie", "lang"=>"en", "lat"=>33.559110, "lon"=>-97.844750),
"/US/US-TX/boyd-20186" => array("type"=>"town", "pop"=>1378, "name_ru"=>"Бойд", "name_en"=>"Boyd", "lang"=>"en", "lat"=>33.080521, "lon"=>-97.561579),
"/US/US-TX/brackettville-20187" => array("type"=>"city", "pop"=>1767, "name_en"=>"Brackettville", "lang"=>"en", "lat"=>29.317773, "lon"=>-100.410362),
"/US/US-TX/brady-20188" => array("type"=>"city", "pop"=>5324, "name_en"=>"Brady", "lang"=>"en", "lat"=>31.133522, "lon"=>-99.375695),
"/US/US-TX/brazoria-20189" => array("type"=>"city", "pop"=>2959, "name_en"=>"Brazoria", "lang"=>"en", "lat"=>29.047096, "lon"=>-95.568238),
"/US/US-TX/breckenridge-20190" => array("type"=>"city", "pop"=>5632, "name_ru"=>"Брекенридж", "name_en"=>"Breckenridge", "lang"=>"en", "lat"=>32.756651, "lon"=>-98.912492),
"/US/US-TX/bremond-20191" => array("type"=>"city", "pop"=>855, "name_en"=>"Bremond", "lang"=>"en", "lat"=>31.165123, "lon"=>-96.676263),
"/US/US-TX/brenham-20192" => array("type"=>"city", "pop"=>15270, "name_en"=>"Brenham", "lang"=>"en", "lat"=>30.159690, "lon"=>-96.397348),
"/US/US-TX/briar-20193" => array("type"=>"cdp", "pop"=>3899, "name_en"=>"Briar", "lang"=>"en", "lat"=>32.992142, "lon"=>-97.554639),
"/US/US-TX/briarcliff-20194" => array("type"=>"village", "pop"=>890, "name_en"=>"Briarcliff", "lang"=>"en", "lat"=>30.405912, "lon"=>-98.046000),
"/US/US-TX/briaroaks-20195" => array("type"=>"city", "pop"=>533, "name_en"=>"Briaroaks", "lang"=>"en", "lat"=>32.492329, "lon"=>-97.301934),
"/US/US-TX/bridge-city-20196" => array("type"=>"city", "pop"=>8593, "name_ru"=>"Бридж Сити", "name_en"=>"Bridge City", "lang"=>"en", "lat"=>30.029888, "lon"=>-93.850903),
"/US/US-TX/bridgeport-20197" => array("type"=>"city", "pop"=>6099, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>33.207584, "lon"=>-97.757206),
"/US/US-TX/broaddus-20198" => array("type"=>"town", "pop"=>180, "name_en"=>"Broaddus", "lang"=>"en", "lat"=>31.305394, "lon"=>-94.269283),
"/US/US-TX/bronson-20199" => array("type"=>"city", "pop"=>259, "name_ru"=>"Бронсон", "name_en"=>"Bronson", "lang"=>"en", "lat"=>31.345493, "lon"=>-94.008949),
"/US/US-TX/bronte-20200" => array("type"=>"town", "pop"=>977, "name_en"=>"Bronte", "lang"=>"en", "lat"=>31.885455, "lon"=>-100.294773),
"/US/US-TX/brookshire-20201" => array("type"=>"city", "pop"=>3949, "name_en"=>"Brookshire", "lang"=>"en", "lat"=>29.782440, "lon"=>-95.953460),
"/US/US-TX/brookside-village-20202" => array("type"=>"city", "pop"=>1984, "name_en"=>"Brookside Village", "lang"=>"en", "lat"=>29.589443, "lon"=>-95.316281),
"/US/US-TX/browndell-20203" => array("type"=>"city", "pop"=>215, "name_en"=>"Browndell", "lang"=>"en", "lat"=>31.124400, "lon"=>-93.981974),
"/US/US-TX/brownfield-20204" => array("type"=>"city", "pop"=>8947, "name_en"=>"Brownfield", "lang"=>"en", "lat"=>33.179734, "lon"=>-102.270467),
"/US/US-TX/brownsboro-20205" => array("type"=>"city", "pop"=>837, "name_en"=>"Brownsboro", "lang"=>"en", "lat"=>32.297306, "lon"=>-95.614703),
"/US/US-TX/brownsville-20206" => array("type"=>"city", "pop"=>175494, "name_ru"=>"Браунсвилл", "name_en"=>"Brownsville", "lang"=>"en", "lat"=>25.925100, "lon"=>-97.482049),
"/US/US-TX/brownwood-20207" => array("type"=>"city", "pop"=>19140, "name_en"=>"Brownwood", "lang"=>"en", "lat"=>31.720637, "lon"=>-98.975574),
"/US/US-TX/bruceville-eddy-20208" => array("type"=>"city", "pop"=>1539, "name_en"=>"Bruceville-Eddy", "lang"=>"en", "lat"=>31.305774, "lon"=>-97.248499),
"/US/US-TX/brushy-creek-20209" => array("type"=>"cdp", "pop"=>5833, "name_en"=>"Brushy Creek", "lang"=>"en", "lat"=>30.513295, "lon"=>-97.739389),
"/US/US-TX/bryan-20210" => array("type"=>"city", "pop"=>72357, "name_en"=>"Bryan", "lang"=>"en", "lat"=>30.668550, "lon"=>-96.365206),
"/US/US-TX/bryson-20211" => array("type"=>"city", "pop"=>513, "name_en"=>"Bryson", "lang"=>"en", "lat"=>33.159703, "lon"=>-98.386612),
"/US/US-TX/buchanan-dam-20212" => array("type"=>"cdp", "pop"=>1099, "name_en"=>"Buchanan Dam", "lang"=>"en", "lat"=>30.799139, "lon"=>-98.438326),
"/US/US-TX/buckholts-20213" => array("type"=>"town", "pop"=>399, "name_en"=>"Buckholts", "lang"=>"en", "lat"=>30.872585, "lon"=>-97.127488),
"/US/US-TX/buckingham-20214" => array("type"=>"town", "pop"=>102, "name_ru"=>"Бакингем", "name_en"=>"Buckingham", "lang"=>"en", "lat"=>32.934101, "lon"=>-96.724952),
"/US/US-TX/buda-20215" => array("type"=>"city", "pop"=>6788, "name_ru"=>"Буда", "name_en"=>"Buda", "lang"=>"en", "lat"=>30.085800, "lon"=>-97.839732),
"/US/US-TX/buffalo-20216" => array("type"=>"city", "pop"=>1973, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>31.461227, "lon"=>-96.062783),
"/US/US-TX/buffalo-gap-20217" => array("type"=>"town", "pop"=>448, "name_en"=>"Buffalo Gap", "lang"=>"en", "lat"=>32.282186, "lon"=>-99.831262),
"/US/US-TX/bullard-20218" => array("type"=>"town", "pop"=>1817, "name_en"=>"Bullard", "lang"=>"en", "lat"=>32.141113, "lon"=>-95.319143),
"/US/US-TX/buna-20219" => array("type"=>"cdp", "pop"=>2127, "name_en"=>"Buna", "lang"=>"en", "lat"=>30.444311, "lon"=>-93.961713),
"/US/US-TX/bunker-hill-village-20220" => array("type"=>"city", "pop"=>3852, "name_en"=>"Bunker Hill Village", "lang"=>"en", "lat"=>29.764665, "lon"=>-95.531636),
"/US/US-TX/burkburnett-20221" => array("type"=>"city", "pop"=>10418, "name_en"=>"Burkburnett", "lang"=>"en", "lat"=>34.079159, "lon"=>-98.564806),
"/US/US-TX/burke-20222" => array("type"=>"city", "pop"=>325, "name_en"=>"Burke", "lang"=>"en", "lat"=>31.228314, "lon"=>-94.769242),
"/US/US-TX/burleson-20223" => array("type"=>"city", "pop"=>34908, "name_en"=>"Burleson", "lang"=>"en", "lat"=>32.533087, "lon"=>-97.328966),
"/US/US-TX/burnet-20224" => array("type"=>"city", "pop"=>5974, "name_en"=>"Burnet", "lang"=>"en", "lat"=>30.761079, "lon"=>-98.229787),
"/US/US-TX/burton-20225" => array("type"=>"town", "pop"=>369, "name_ru"=>"Бертон", "name_en"=>"Burton", "lang"=>"en", "lat"=>30.179343, "lon"=>-96.597266),
"/US/US-TX/byers-20226" => array("type"=>"city", "pop"=>510, "name_ru"=>"Байерс", "name_en"=>"Byers", "lang"=>"en", "lat"=>34.071152, "lon"=>-98.190204),
"/US/US-TX/bynum-20227" => array("type"=>"town", "pop"=>246, "name_ru"=>"Бинум", "name_en"=>"Bynum", "lang"=>"en", "lat"=>31.968858, "lon"=>-97.003030),
"/US/US-TX/cactus-20228" => array("type"=>"city", "pop"=>2605, "name_en"=>"Cactus", "lang"=>"en", "lat"=>36.044734, "lon"=>-102.011103),
"/US/US-TX/caddo-mills-20229" => array("type"=>"city", "pop"=>1204, "name_en"=>"Caddo Mills", "lang"=>"en", "lat"=>33.065705, "lon"=>-96.226940),
"/US/US-TX/caldwell-20230" => array("type"=>"city", "pop"=>3736, "name_ru"=>"Колдуэлл", "name_en"=>"Caldwell", "lang"=>"en", "lat"=>30.529139, "lon"=>-96.702988),
"/US/US-TX/callisburg-20231" => array("type"=>"town", "pop"=>347, "name_en"=>"Callisburg", "lang"=>"en", "lat"=>33.699771, "lon"=>-97.016423),
"/US/US-TX/calvert-20232" => array("type"=>"city", "pop"=>1358, "name_en"=>"Calvert", "lang"=>"en", "lat"=>30.979581, "lon"=>-96.671715),
"/US/US-TX/cameron-20233" => array("type"=>"city", "pop"=>5740, "name_ru"=>"Кэмерон", "name_en"=>"Cameron", "lang"=>"en", "lat"=>30.858112, "lon"=>-96.976981),
"/US/US-TX/cameron-park-20234" => array("type"=>"cdp", "pop"=>3802, "name_ru"=>"Камерон Парк", "name_en"=>"Cameron Park", "lang"=>"en", "lat"=>25.964250, "lon"=>-97.476402),
"/US/US-TX/camp-swift-20236" => array("type"=>"cdp", "pop"=>2681, "name_en"=>"Camp Swift", "lang"=>"en", "lat"=>30.190652, "lon"=>-97.292053),
"/US/US-TX/camp-wood-20237" => array("type"=>"city", "pop"=>784, "name_en"=>"Camp Wood", "lang"=>"en", "lat"=>29.668859, "lon"=>-100.011295),
"/US/US-TX/campbell-20235" => array("type"=>"town", "pop"=>742, "name_ru"=>"Кэмпбелл", "name_en"=>"Campbell", "lang"=>"en", "lat"=>33.147915, "lon"=>-95.954972),
"/US/US-TX/canadian-20238" => array("type"=>"town", "pop"=>2290, "name_en"=>"Canadian", "lang"=>"en", "lat"=>35.910659, "lon"=>-100.383350),
"/US/US-TX/caney-city-20239" => array("type"=>"town", "pop"=>256, "name_en"=>"Caney City", "lang"=>"en", "lat"=>32.208672, "lon"=>-96.039364),
"/US/US-TX/canton-20240" => array("type"=>"city", "pop"=>3665, "name_ru"=>"Кэнтон", "name_en"=>"Canton", "lang"=>"en", "lat"=>32.547542, "lon"=>-95.863854),
"/US/US-TX/canutillo-20241" => array("type"=>"cdp", "pop"=>4442, "name_en"=>"Canutillo", "lang"=>"en", "lat"=>31.918678, "lon"=>-106.599427),
"/US/US-TX/canyon-20242" => array("type"=>"city", "pop"=>14602, "name_en"=>"Canyon", "lang"=>"en", "lat"=>34.981284, "lon"=>-101.921605),
"/US/US-TX/canyon-lake-20243" => array("type"=>"cdp", "pop"=>9975, "name_ru"=>"Каньон Лейк", "name_en"=>"Canyon Lake", "lang"=>"en", "lat"=>29.874958, "lon"=>-98.262257),
"/US/US-TX/carbon-20244" => array("type"=>"town", "pop"=>223, "name_ru"=>"Карбон", "name_en"=>"Carbon", "lang"=>"en", "lat"=>32.268576, "lon"=>-98.826712),
"/US/US-TX/carl's-corner-20245" => array("type"=>"town", "pop"=>158, "name_en"=>"Carl's Corner", "lang"=>"en", "lat"=>32.085251, "lon"=>-97.044878),
"/US/US-TX/carmine-20246" => array("type"=>"city", "pop"=>230, "name_en"=>"Carmine", "lang"=>"en", "lat"=>30.149596, "lon"=>-96.688831),
"/US/US-TX/carrizo-springs-20247" => array("type"=>"city", "pop"=>5332, "name_en"=>"Carrizo Springs", "lang"=>"en", "lat"=>28.525501, "lon"=>-99.858986),
"/US/US-TX/carrollton-20248" => array("type"=>"city", "pop"=>125595, "name_ru"=>"Карролтон", "name_en"=>"Carrollton", "lang"=>"en", "lat"=>32.988141, "lon"=>-96.896710),
"/US/US-TX/carthage-20249" => array("type"=>"city", "pop"=>6633, "name_ru"=>"Картаж", "name_en"=>"Carthage", "lang"=>"en", "lat"=>32.151779, "lon"=>-94.336925),
"/US/US-TX/castle-hills-20250" => array("type"=>"city", "pop"=>4205, "name_en"=>"Castle Hills", "lang"=>"en", "lat"=>29.523833, "lon"=>-98.519181),
"/US/US-TX/castroville-20251" => array("type"=>"city", "pop"=>3059, "name_ru"=>"Кастровилл", "name_en"=>"Castroville", "lang"=>"en", "lat"=>29.348029, "lon"=>-98.873204),
"/US/US-TX/cedar-hill-20252" => array("type"=>"city", "pop"=>44786, "name_ru"=>"Сидар Хилл", "name_en"=>"Cedar Hill", "lang"=>"en", "lat"=>32.580276, "lon"=>-96.957623),
"/US/US-TX/cedar-park-20253" => array("type"=>"city", "pop"=>62308, "name_en"=>"Cedar Park", "lang"=>"en", "lat"=>30.514570, "lon"=>-97.830813),
"/US/US-TX/celeste-20254" => array("type"=>"town", "pop"=>864, "name_en"=>"Celeste", "lang"=>"en", "lat"=>33.294161, "lon"=>-96.194131),
"/US/US-TX/celina-20255" => array("type"=>"town", "pop"=>5557, "name_en"=>"Celina", "lang"=>"en", "lat"=>33.319980, "lon"=>-96.783775),
"/US/US-TX/center-20256" => array("type"=>"city", "pop"=>5746, "name_ru"=>"Сентер", "name_en"=>"Center", "lang"=>"en", "lat"=>31.793532, "lon"=>-94.178266),
"/US/US-TX/centerville-20257" => array("type"=>"city", "pop"=>971, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>31.258303, "lon"=>-95.976799),
"/US/US-TX/central-gardens-20258" => array("type"=>"cdp", "pop"=>4026, "name_en"=>"Central Gardens", "lang"=>"en", "lat"=>29.991847, "lon"=>-94.020698),
"/US/US-TX/chandler-20259" => array("type"=>"town", "pop"=>2631, "name_ru"=>"Чандлер", "name_en"=>"Chandler", "lang"=>"en", "lat"=>32.305813, "lon"=>-95.478911),
"/US/US-TX/channelview-20260" => array("type"=>"cdp", "pop"=>25564, "name_en"=>"Channelview", "lang"=>"en", "lat"=>29.788583, "lon"=>-95.115084),
"/US/US-TX/channing-20261" => array("type"=>"city", "pop"=>327, "name_en"=>"Channing", "lang"=>"en", "lat"=>35.682122, "lon"=>-102.331738),
"/US/US-TX/charlotte-20262" => array("type"=>"city", "pop"=>1788, "name_ru"=>"Шарлотт", "name_en"=>"Charlotte", "lang"=>"en", "lat"=>28.859190, "lon"=>-98.700011),
"/US/US-TX/chateau-woods-20263" => array("type"=>"city", "pop"=>641, "name_en"=>"Chateau Woods", "lang"=>"en", "lat"=>30.165426, "lon"=>-95.427390),
"/US/US-TX/chester-20264" => array("type"=>"town", "pop"=>259, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>30.922645, "lon"=>-94.602640),
"/US/US-TX/chico-20265" => array("type"=>"city", "pop"=>1079, "name_ru"=>"Чико", "name_en"=>"Chico", "lang"=>"en", "lat"=>33.294200, "lon"=>-97.799235),
"/US/US-TX/childress-20266" => array("type"=>"city", "pop"=>6513, "name_en"=>"Childress", "lang"=>"en", "lat"=>34.424505, "lon"=>-100.248319),
"/US/US-TX/chillicothe-20267" => array("type"=>"city", "pop"=>687, "name_ru"=>"Чилликот", "name_en"=>"Chillicothe", "lang"=>"en", "lat"=>34.255582, "lon"=>-99.513851),
"/US/US-TX/china-20268" => array("type"=>"city", "pop"=>1032, "name_en"=>"China", "lang"=>"en", "lat"=>30.052240, "lon"=>-94.335639),
"/US/US-TX/china-grove-20269" => array("type"=>"town", "pop"=>1303, "name_en"=>"China Grove", "lang"=>"en", "lat"=>29.395650, "lon"=>-98.348142),
"/US/US-TX/chireno-20270" => array("type"=>"city", "pop"=>421, "name_en"=>"Chireno", "lang"=>"en", "lat"=>31.498751, "lon"=>-94.345418),
"/US/US-TX/christine-20271" => array("type"=>"town", "pop"=>489, "name_en"=>"Christine", "lang"=>"en", "lat"=>28.785150, "lon"=>-98.498644),
"/US/US-TX/cibolo-20272" => array("type"=>"city", "pop"=>14555, "name_en"=>"Cibolo", "lang"=>"en", "lat"=>29.573899, "lon"=>-98.232796),
"/US/US-TX/circle-d-kc-estates-20273" => array("type"=>"cdp", "pop"=>1247, "name_en"=>"Circle D-KC Estates", "lang"=>"en", "lat"=>30.160805, "lon"=>-97.231350),
"/US/US-TX/cisco-20274" => array("type"=>"city", "pop"=>3724, "name_ru"=>"Сиско", "name_en"=>"Cisco", "lang"=>"en", "lat"=>32.385034, "lon"=>-98.979959),
"/US/US-TX/clarendon-20275" => array("type"=>"city", "pop"=>2000, "name_ru"=>"Кларендон", "name_en"=>"Clarendon", "lang"=>"en", "lat"=>34.936741, "lon"=>-100.891755),
"/US/US-TX/clarksville-20276" => array("type"=>"city", "pop"=>3478, "name_ru"=>"Кларксвилл", "name_en"=>"Clarksville", "lang"=>"en", "lat"=>33.610537, "lon"=>-95.055463),
"/US/US-TX/clarksville-city-20277" => array("type"=>"city", "pop"=>844, "name_en"=>"Clarksville City", "lang"=>"en", "lat"=>32.535099, "lon"=>-94.895525),
"/US/US-TX/claude-20278" => array("type"=>"city", "pop"=>1293, "name_en"=>"Claude", "lang"=>"en", "lat"=>35.107313, "lon"=>-101.362282),
"/US/US-TX/clear-lake-shores-20279" => array("type"=>"city", "pop"=>1294, "name_en"=>"Clear Lake Shores", "lang"=>"en", "lat"=>29.548192, "lon"=>-95.033935),
"/US/US-TX/cleburne-20280" => array("type"=>"city", "pop"=>29889, "name_en"=>"Cleburne", "lang"=>"en", "lat"=>32.349253, "lon"=>-97.406338),
"/US/US-TX/cleveland-20281" => array("type"=>"city", "pop"=>7960, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>30.344246, "lon"=>-95.078720),
"/US/US-TX/clifton-20282" => array("type"=>"city", "pop"=>3565, "name_ru"=>"Слифтон", "name_en"=>"Clifton", "lang"=>"en", "lat"=>31.781309, "lon"=>-97.580117),
"/US/US-TX/clint-20283" => array("type"=>"town", "pop"=>972, "name_en"=>"Clint", "lang"=>"en", "lat"=>31.590131, "lon"=>-106.228359),
"/US/US-TX/cloverleaf-20284" => array("type"=>"cdp", "pop"=>18230, "name_en"=>"Cloverleaf", "lang"=>"en", "lat"=>29.787350, "lon"=>-95.172799),
"/US/US-TX/clute-20285" => array("type"=>"city", "pop"=>10765, "name_en"=>"Clute", "lang"=>"en", "lat"=>29.024573, "lon"=>-95.397484),
"/US/US-TX/clyde-20286" => array("type"=>"town", "pop"=>3809, "name_ru"=>"Клайд", "name_en"=>"Clyde", "lang"=>"en", "lat"=>32.402362, "lon"=>-99.498186),
"/US/US-TX/coahoma-20287" => array("type"=>"town", "pop"=>914, "name_ru"=>"Коахома", "name_en"=>"Coahoma", "lang"=>"en", "lat"=>32.295019, "lon"=>-101.306337),
"/US/US-TX/cockrell-hill-20288" => array("type"=>"city", "pop"=>4272, "name_en"=>"Cockrell Hill", "lang"=>"en", "lat"=>32.738601, "lon"=>-96.888900),
"/US/US-TX/coffee-city-20289" => array("type"=>"town", "pop"=>207, "name_en"=>"Coffee City", "lang"=>"en", "lat"=>32.129943, "lon"=>-95.478209),
"/US/US-TX/coldspring-20290" => array("type"=>"city", "pop"=>767, "name_en"=>"Coldspring", "lang"=>"en", "lat"=>30.590947, "lon"=>-95.129163),
"/US/US-TX/coleman-20291" => array("type"=>"city", "pop"=>4691, "name_ru"=>"Коулман", "name_en"=>"Coleman", "lang"=>"en", "lat"=>31.831046, "lon"=>-99.421806),
"/US/US-TX/college-station-20292" => array("type"=>"city", "pop"=>84128, "name_en"=>"College Station", "lang"=>"en", "lat"=>30.600500, "lon"=>-96.312351),
"/US/US-TX/colleyville-20293" => array("type"=>"city", "pop"=>24428, "name_en"=>"Colleyville", "lang"=>"en", "lat"=>32.892487, "lon"=>-97.147135),
"/US/US-TX/collinsville-20294" => array("type"=>"town", "pop"=>1494, "name_ru"=>"Коллинсвилл", "name_en"=>"Collinsville", "lang"=>"en", "lat"=>33.559661, "lon"=>-96.910005),
"/US/US-TX/colmesneil-20295" => array("type"=>"city", "pop"=>626, "name_en"=>"Colmesneil", "lang"=>"en", "lat"=>30.909009, "lon"=>-94.423069),
"/US/US-TX/colorado-city-20296" => array("type"=>"city", "pop"=>3888, "name_ru"=>"Колорадо Сити", "name_en"=>"Colorado City", "lang"=>"en", "lat"=>32.399746, "lon"=>-100.857555),
"/US/US-TX/columbus-20297" => array("type"=>"city", "pop"=>3886, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>29.704445, "lon"=>-96.551530),
"/US/US-TX/comanche-20298" => array("type"=>"city", "pop"=>4201, "name_en"=>"Comanche", "lang"=>"en", "lat"=>31.899743, "lon"=>-98.604161),
"/US/US-TX/combes-20299" => array("type"=>"town", "pop"=>2845, "name_en"=>"Combes", "lang"=>"en", "lat"=>26.242400, "lon"=>-97.724210),
"/US/US-TX/combine-20300" => array("type"=>"city", "pop"=>2128, "name_en"=>"Combine", "lang"=>"en", "lat"=>32.588213, "lon"=>-96.514879),
"/US/US-TX/comfort-20301" => array("type"=>"cdp", "pop"=>1477, "name_en"=>"Comfort", "lang"=>"en", "lat"=>29.972332, "lon"=>-98.902806),
"/US/US-TX/commerce-20302" => array("type"=>"city", "pop"=>9238, "name_ru"=>"Коммерс", "name_en"=>"Commerce", "lang"=>"en", "lat"=>33.238158, "lon"=>-95.900770),
"/US/US-TX/como-20303" => array("type"=>"town", "pop"=>655, "name_ru"=>"Комо", "name_en"=>"Como", "lang"=>"en", "lat"=>33.060504, "lon"=>-95.475083),
"/US/US-TX/conroe-20304" => array("type"=>"city", "pop"=>55429, "name_en"=>"Conroe", "lang"=>"en", "lat"=>30.323320, "lon"=>-95.471373),
"/US/US-TX/converse-20305" => array("type"=>"city", "pop"=>17510, "name_ru"=>"Конверс", "name_en"=>"Converse", "lang"=>"en", "lat"=>29.513750, "lon"=>-98.311852),
"/US/US-TX/cool-20306" => array("type"=>"city", "pop"=>190, "name_en"=>"Cool", "lang"=>"en", "lat"=>32.798344, "lon"=>-98.012461),
"/US/US-TX/coolidge-20307" => array("type"=>"town", "pop"=>859, "name_ru"=>"Кулидж", "name_en"=>"Coolidge", "lang"=>"en", "lat"=>31.751187, "lon"=>-96.651961),
"/US/US-TX/cooper-20308" => array("type"=>"city", "pop"=>2175, "name_en"=>"Cooper", "lang"=>"en", "lat"=>33.371844, "lon"=>-95.690482),
"/US/US-TX/coppell-20309" => array("type"=>"city", "pop"=>39154, "name_en"=>"Coppell", "lang"=>"en", "lat"=>32.962949, "lon"=>-96.990456),
"/US/US-TX/copper-canyon-20311" => array("type"=>"town", "pop"=>1467, "name_en"=>"Copper Canyon", "lang"=>"en", "lat"=>33.097079, "lon"=>-97.096656),
"/US/US-TX/copperas-cove-20310" => array("type"=>"city", "pop"=>30710, "name_en"=>"Copperas Cove", "lang"=>"en", "lat"=>31.118750, "lon"=>-97.901783),
"/US/US-TX/corinth-20312" => array("type"=>"town", "pop"=>20773, "name_ru"=>"Коринт", "name_en"=>"Corinth", "lang"=>"en", "lat"=>33.143793, "lon"=>-97.064707),
"/US/US-TX/corpus-christi-20313" => array("type"=>"city", "pop"=>286462, "name_en"=>"Corpus Christi", "lang"=>"en", "lat"=>27.705728, "lon"=>-97.292827),
"/US/US-TX/corral-city-20314" => array("type"=>"town", "pop"=>106, "name_en"=>"Corral City", "lang"=>"en", "lat"=>33.099343, "lon"=>-97.226190),
"/US/US-TX/corrigan-20315" => array("type"=>"town", "pop"=>1872, "name_en"=>"Corrigan", "lang"=>"en", "lat"=>30.999452, "lon"=>-94.824444),
"/US/US-TX/corsicana-20316" => array("type"=>"city", "pop"=>26459, "name_en"=>"Corsicana", "lang"=>"en", "lat"=>32.084214, "lon"=>-96.464953),
"/US/US-TX/cottonwood-20317" => array("type"=>"city", "pop"=>214, "name_ru"=>"Коттонвуд", "name_en"=>"Cottonwood", "lang"=>"en", "lat"=>32.456896, "lon"=>-96.391479),
"/US/US-TX/cottonwood-shores-20318" => array("type"=>"city", "pop"=>1185, "name_en"=>"Cottonwood Shores", "lang"=>"en", "lat"=>30.555833, "lon"=>-98.325489),
"/US/US-TX/cotulla-20319" => array("type"=>"city", "pop"=>3529, "name_en"=>"Cotulla", "lang"=>"en", "lat"=>28.436084, "lon"=>-99.236400),
"/US/US-TX/cove-20320" => array("type"=>"town", "pop"=>289, "name_ru"=>"Ков", "name_en"=>"Cove", "lang"=>"en", "lat"=>29.813224, "lon"=>-94.824905),
"/US/US-TX/covington-20321" => array("type"=>"city", "pop"=>302, "name_ru"=>"Ковингтон", "name_en"=>"Covington", "lang"=>"en", "lat"=>32.177402, "lon"=>-97.258883),
"/US/US-TX/crandall-20322" => array("type"=>"city", "pop"=>3850, "name_ru"=>"Крэндолл", "name_en"=>"Crandall", "lang"=>"en", "lat"=>32.630703, "lon"=>-96.453726),
"/US/US-TX/crane-20323" => array("type"=>"city", "pop"=>3194, "name_ru"=>"Крейн", "name_en"=>"Crane", "lang"=>"en", "lat"=>31.391836, "lon"=>-102.350041),
"/US/US-TX/cranfills-gap-20324" => array("type"=>"city", "pop"=>348, "name_en"=>"Cranfills Gap", "lang"=>"en", "lat"=>31.776151, "lon"=>-97.829849),
"/US/US-TX/crawford-20325" => array("type"=>"town", "pop"=>790, "name_ru"=>"Крофорд", "name_en"=>"Crawford", "lang"=>"en", "lat"=>31.533794, "lon"=>-97.440089),
"/US/US-TX/creedmoor-20326" => array("type"=>"city", "pop"=>210, "name_en"=>"Creedmoor", "lang"=>"en", "lat"=>30.091880, "lon"=>-97.741675),
"/US/US-TX/crockett-20327" => array("type"=>"city", "pop"=>6815, "name_ru"=>"Крокетт", "name_en"=>"Crockett", "lang"=>"en", "lat"=>31.317918, "lon"=>-95.456320),
"/US/US-TX/crosby-20328" => array("type"=>"cdp", "pop"=>6192, "name_ru"=>"Кросби", "name_en"=>"Crosby", "lang"=>"en", "lat"=>29.914076, "lon"=>-95.058179),
"/US/US-TX/crosbyton-20329" => array("type"=>"city", "pop"=>1614, "name_en"=>"Crosbyton", "lang"=>"en", "lat"=>33.641948, "lon"=>-101.237360),
"/US/US-TX/cross-mountain-20330" => array("type"=>"cdp", "pop"=>1112, "name_en"=>"Cross Mountain", "lang"=>"en", "lat"=>29.645241, "lon"=>-98.659060),
"/US/US-TX/cross-plains-20331" => array("type"=>"town", "pop"=>1078, "name_en"=>"Cross Plains", "lang"=>"en", "lat"=>32.126593, "lon"=>-99.165276),
"/US/US-TX/cross-roads-20332" => array("type"=>"town", "pop"=>689, "name_en"=>"Cross Roads", "lang"=>"en", "lat"=>33.230632, "lon"=>-96.994061),
"/US/US-TX/crowell-20333" => array("type"=>"city", "pop"=>953, "name_en"=>"Crowell", "lang"=>"en", "lat"=>33.983578, "lon"=>-99.724078),
"/US/US-TX/crowley-20334" => array("type"=>"city", "pop"=>12368, "name_ru"=>"Кроули", "name_en"=>"Crowley", "lang"=>"en", "lat"=>32.577329, "lon"=>-97.358070),
"/US/US-TX/crystal-city-20335" => array("type"=>"city", "pop"=>7161, "name_ru"=>"Кристал Сити", "name_en"=>"Crystal City", "lang"=>"en", "lat"=>28.689910, "lon"=>-99.825484),
"/US/US-TX/cuero-20336" => array("type"=>"city", "pop"=>6417, "name_en"=>"Cuero", "lang"=>"en", "lat"=>29.094082, "lon"=>-97.287401),
"/US/US-TX/cumby-20337" => array("type"=>"city", "pop"=>674, "name_en"=>"Cumby", "lang"=>"en", "lat"=>33.133017, "lon"=>-95.838777),
"/US/US-TX/cuney-20338" => array("type"=>"town", "pop"=>147, "name_en"=>"Cuney", "lang"=>"en", "lat"=>32.038239, "lon"=>-95.415425),
"/US/US-TX/cushing-20339" => array("type"=>"city", "pop"=>659, "name_ru"=>"Кушинг", "name_en"=>"Cushing", "lang"=>"en", "lat"=>31.813319, "lon"=>-94.840074),
"/US/US-TX/cut-and-shoot-20340" => array("type"=>"town", "pop"=>1235, "name_en"=>"Cut and Shoot", "lang"=>"en", "lat"=>30.338601, "lon"=>-95.352899),
"/US/US-TX/daingerfield-20341" => array("type"=>"town", "pop"=>2452, "name_en"=>"Daingerfield", "lang"=>"en", "lat"=>33.030595, "lon"=>-94.724245),
"/US/US-TX/daisetta-20342" => array("type"=>"city", "pop"=>1079, "name_en"=>"Daisetta", "lang"=>"en", "lat"=>30.114112, "lon"=>-94.642715),
"/US/US-TX/dalhart-20343" => array("type"=>"city", "pop"=>7032, "name_en"=>"Dalhart", "lang"=>"en", "lat"=>36.061819, "lon"=>-102.514594),
"/US/US-TX/dallas-20344" => array("type"=>"city", "pop"=>1279910, "name_ru"=>"Даллас", "name_en"=>"Dallas", "lang"=>"en", "lat"=>32.794151, "lon"=>-96.765249),
"/US/US-TX/dalworthington-gardens-20345" => array("type"=>"city", "pop"=>2412, "name_en"=>"Dalworthington Gardens", "lang"=>"en", "lat"=>32.693651, "lon"=>-97.158653),
"/US/US-TX/danbury-20346" => array("type"=>"city", "pop"=>1684, "name_ru"=>"Данбери", "name_en"=>"Danbury", "lang"=>"en", "lat"=>29.226579, "lon"=>-95.345655),
"/US/US-TX/darrouzett-20347" => array("type"=>"town", "pop"=>287, "name_en"=>"Darrouzett", "lang"=>"en", "lat"=>36.445095, "lon"=>-100.325419),
"/US/US-TX/dawson-20348" => array("type"=>"town", "pop"=>961, "name_ru"=>"Доусон", "name_en"=>"Dawson", "lang"=>"en", "lat"=>31.893750, "lon"=>-96.713989),
"/US/US-TX/dayton-20349" => array("type"=>"city", "pop"=>7371, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>30.050409, "lon"=>-94.895321),
"/US/US-TX/dayton-lakes-20350" => array("type"=>"city", "pop"=>105, "name_en"=>"Dayton Lakes", "lang"=>"en", "lat"=>30.145215, "lon"=>-94.821484),
"/US/US-TX/de-kalb-20354" => array("type"=>"town", "pop"=>1805, "name_ru"=>"Де Калб", "name_en"=>"De Kalb", "lang"=>"en", "lat"=>33.507224, "lon"=>-94.616509),
"/US/US-TX/de-leon-20355" => array("type"=>"city", "pop"=>2335, "name_en"=>"De Leon", "lang"=>"en", "lat"=>32.111689, "lon"=>-98.536666),
"/US/US-TX/dean-20351" => array("type"=>"city", "pop"=>340, "name_en"=>"Dean", "lang"=>"en", "lat"=>33.933316, "lon"=>-98.368895),
"/US/US-TX/decatur-20352" => array("type"=>"city", "pop"=>6432, "name_ru"=>"Декатур", "name_en"=>"Decatur", "lang"=>"en", "lat"=>33.233327, "lon"=>-97.591361),
"/US/US-TX/deer-park-20353" => array("type"=>"city", "pop"=>30890, "name_ru"=>"Дир Парк", "name_en"=>"Deer Park", "lang"=>"en", "lat"=>29.690389, "lon"=>-95.115313),
"/US/US-TX/del-rio-20357" => array("type"=>"city", "pop"=>36682, "name_ru"=>"Дель Рио", "name_en"=>"Del Rio", "lang"=>"en", "lat"=>29.376347, "lon"=>-100.895659),
"/US/US-TX/dell-city-20356" => array("type"=>"city", "pop"=>387, "name_en"=>"Dell City", "lang"=>"en", "lat"=>31.935212, "lon"=>-105.199787),
"/US/US-TX/denison-20358" => array("type"=>"city", "pop"=>24001, "name_ru"=>"Денисон", "name_en"=>"Denison", "lang"=>"en", "lat"=>33.755894, "lon"=>-96.564105),
"/US/US-TX/denton-20359" => array("type"=>"city", "pop"=>119454, "name_ru"=>"Дентон", "name_en"=>"Denton", "lang"=>"en", "lat"=>33.214204, "lon"=>-97.130942),
"/US/US-TX/denver-city-20360" => array("type"=>"town", "pop"=>4074, "name_en"=>"Denver City", "lang"=>"en", "lat"=>32.969611, "lon"=>-102.830733),
"/US/US-TX/deport-20361" => array("type"=>"city", "pop"=>661, "name_en"=>"Deport", "lang"=>"en", "lat"=>33.529069, "lon"=>-95.317455),
"/US/US-TX/desoto-20362" => array("type"=>"city", "pop"=>47568, "name_en"=>"DeSoto", "lang"=>"en", "lat"=>32.599200, "lon"=>-96.862550),
"/US/US-TX/detroit-20363" => array("type"=>"town", "pop"=>708, "name_ru"=>"Детройт", "name_en"=>"Detroit", "lang"=>"en", "lat"=>33.660759, "lon"=>-95.265976),
"/US/US-TX/devers-20364" => array("type"=>"city", "pop"=>432, "name_en"=>"Devers", "lang"=>"en", "lat"=>30.027586, "lon"=>-94.585077),
"/US/US-TX/devine-20365" => array("type"=>"city", "pop"=>4545, "name_en"=>"Devine", "lang"=>"en", "lat"=>29.145407, "lon"=>-98.904698),
"/US/US-TX/deweyville-20366" => array("type"=>"cdp", "pop"=>1218, "name_en"=>"Deweyville", "lang"=>"en", "lat"=>30.306912, "lon"=>-93.757945),
"/US/US-TX/diboll-20367" => array("type"=>"city", "pop"=>5537, "name_en"=>"Diboll", "lang"=>"en", "lat"=>31.188612, "lon"=>-94.783465),
"/US/US-TX/dickens-20368" => array("type"=>"city", "pop"=>288, "name_ru"=>"Дикенс", "name_en"=>"Dickens", "lang"=>"en", "lat"=>33.621257, "lon"=>-100.835830),
"/US/US-TX/dickinson-20369" => array("type"=>"city", "pop"=>17886, "name_en"=>"Dickinson", "lang"=>"en", "lat"=>29.453712, "lon"=>-95.066321),
"/US/US-TX/dilley-20370" => array("type"=>"city", "pop"=>3604, "name_en"=>"Dilley", "lang"=>"en", "lat"=>28.668816, "lon"=>-99.171002),
"/US/US-TX/dimmitt-20371" => array("type"=>"city", "pop"=>3701, "name_en"=>"Dimmitt", "lang"=>"en", "lat"=>34.547924, "lon"=>-102.315205),
"/US/US-TX/dodd-city-20372" => array("type"=>"town", "pop"=>444, "name_en"=>"Dodd City", "lang"=>"en", "lat"=>33.575756, "lon"=>-96.074656),
"/US/US-TX/dodson-20373" => array("type"=>"town", "pop"=>110, "name_ru"=>"Додсон", "name_en"=>"Dodson", "lang"=>"en", "lat"=>34.764985, "lon"=>-100.019537),
"/US/US-TX/dominion-20374" => array("type"=>"cdp", "pop"=>1196, "name_en"=>"Dominion", "lang"=>"en", "lat"=>29.656570, "lon"=>-98.616224),
"/US/US-TX/domino-20375" => array("type"=>"town", "pop"=>50, "name_en"=>"Domino", "lang"=>"en", "lat"=>33.251441, "lon"=>-94.114745),
"/US/US-TX/donna-20376" => array("type"=>"city", "pop"=>17094, "name_en"=>"Donna", "lang"=>"en", "lat"=>26.168936, "lon"=>-98.050001),
"/US/US-TX/dorchester-20377" => array("type"=>"town", "pop"=>109, "name_ru"=>"Дорчестер", "name_en"=>"Dorchester", "lang"=>"en", "lat"=>33.530087, "lon"=>-96.691944),
"/US/US-TX/double-oak-20378" => array("type"=>"town", "pop"=>3314, "name_en"=>"Double Oak", "lang"=>"en", "lat"=>33.064072, "lon"=>-97.112611),
"/US/US-TX/douglassville-20379" => array("type"=>"town", "pop"=>163, "name_en"=>"Douglassville", "lang"=>"en", "lat"=>33.191801, "lon"=>-94.351423),
"/US/US-TX/dripping-springs-20380" => array("type"=>"city", "pop"=>1828, "name_en"=>"Dripping Springs", "lang"=>"en", "lat"=>30.188484, "lon"=>-98.088253),
"/US/US-TX/driscoll-20381" => array("type"=>"city", "pop"=>802, "name_en"=>"Driscoll", "lang"=>"en", "lat"=>27.672242, "lon"=>-97.750786),
"/US/US-TX/dublin-20382" => array("type"=>"city", "pop"=>3803, "name_ru"=>"Дублин", "name_en"=>"Dublin", "lang"=>"en", "lat"=>32.086574, "lon"=>-98.338915),
"/US/US-TX/dumas-20383" => array("type"=>"city", "pop"=>13916, "name_ru"=>"Дюма", "name_en"=>"Dumas", "lang"=>"en", "lat"=>35.863230, "lon"=>-101.964500),
"/US/US-TX/duncanville-20384" => array("type"=>"city", "pop"=>36109, "name_en"=>"Duncanville", "lang"=>"en", "lat"=>32.645867, "lon"=>-96.913467),
"/US/US-TX/eagle-lake-20385" => array("type"=>"city", "pop"=>3679, "name_ru"=>"Игл Лэйк", "name_en"=>"Eagle Lake", "lang"=>"en", "lat"=>29.587127, "lon"=>-96.328286),
"/US/US-TX/eagle-mountain-20386" => array("type"=>"cdp", "pop"=>5847, "name_en"=>"Eagle Mountain", "lang"=>"en", "lat"=>32.893442, "lon"=>-97.444273),
"/US/US-TX/eagle-pass-20387" => array("type"=>"city", "pop"=>26668, "name_en"=>"Eagle Pass", "lang"=>"en", "lat"=>28.709759, "lon"=>-100.491409),
"/US/US-TX/early-20388" => array("type"=>"city", "pop"=>2760, "name_ru"=>"Эрли", "name_en"=>"Early", "lang"=>"en", "lat"=>31.744518, "lon"=>-98.940725),
"/US/US-TX/earth-20389" => array("type"=>"city", "pop"=>1001, "name_en"=>"Earth", "lang"=>"en", "lat"=>34.233490, "lon"=>-102.408433),
"/US/US-TX/east-bernard-20390" => array("type"=>"cdp", "pop"=>2215, "name_en"=>"East Bernard", "lang"=>"en", "lat"=>29.527967, "lon"=>-96.062953),
"/US/US-TX/east-mountain-20392" => array("type"=>"city", "pop"=>628, "name_en"=>"East Mountain", "lang"=>"en", "lat"=>32.593910, "lon"=>-94.855222),
"/US/US-TX/east-tawakoni-20394" => array("type"=>"city", "pop"=>966, "name_en"=>"East Tawakoni", "lang"=>"en", "lat"=>32.897713, "lon"=>-95.933610),
"/US/US-TX/eastland-20391" => array("type"=>"city", "pop"=>3886, "name_en"=>"Eastland", "lang"=>"en", "lat"=>32.401162, "lon"=>-98.819103),
"/US/US-TX/easton-20393" => array("type"=>"city", "pop"=>555, "name_ru"=>"Истон", "name_en"=>"Easton", "lang"=>"en", "lat"=>32.382948, "lon"=>-94.593399),
"/US/US-TX/ector-20395" => array("type"=>"town", "pop"=>630, "name_en"=>"Ector", "lang"=>"en", "lat"=>33.578952, "lon"=>-96.272882),
"/US/US-TX/edcouch-20396" => array("type"=>"city", "pop"=>4613, "name_en"=>"Edcouch", "lang"=>"en", "lat"=>26.293994, "lon"=>-97.964268),
"/US/US-TX/eden-20397" => array("type"=>"city", "pop"=>2381, "name_ru"=>"Иден", "name_en"=>"Eden", "lang"=>"en", "lat"=>31.215665, "lon"=>-99.843556),
"/US/US-TX/edgecliff-20398" => array("type"=>"village", "pop"=>2715, "name_en"=>"Edgecliff", "lang"=>"en", "lat"=>32.655184, "lon"=>-97.341322),
"/US/US-TX/edgewood-20399" => array("type"=>"town", "pop"=>1442, "name_ru"=>"Эджвуд", "name_en"=>"Edgewood", "lang"=>"en", "lat"=>32.694818, "lon"=>-95.882062),
"/US/US-TX/edinburg-20400" => array("type"=>"city", "pop"=>71520, "name_ru"=>"Эдинбург", "name_en"=>"Edinburg", "lang"=>"en", "lat"=>26.295800, "lon"=>-98.164874),
"/US/US-TX/edmonson-20401" => array("type"=>"town", "pop"=>120, "name_en"=>"Edmonson", "lang"=>"en", "lat"=>34.281067, "lon"=>-101.894838),
"/US/US-TX/edna-20402" => array("type"=>"city", "pop"=>5807, "name_ru"=>"Эдна", "name_en"=>"Edna", "lang"=>"en", "lat"=>28.973984, "lon"=>-96.647314),
"/US/US-TX/edom-20403" => array("type"=>"city", "pop"=>348, "name_en"=>"Edom", "lang"=>"en", "lat"=>32.373018, "lon"=>-95.611124),
"/US/US-TX/el-campo-20404" => array("type"=>"city", "pop"=>10773, "name_en"=>"El Campo", "lang"=>"en", "lat"=>29.196487, "lon"=>-96.276290),
"/US/US-TX/el-cenizo-20405" => array("type"=>"city", "pop"=>3570, "name_en"=>"El Cenizo", "lang"=>"en", "lat"=>27.332211, "lon"=>-99.503091),
"/US/US-TX/el-lago-20410" => array("type"=>"city", "pop"=>2952, "name_en"=>"El Lago", "lang"=>"en", "lat"=>29.572808, "lon"=>-95.044481),
"/US/US-TX/el-paso-20412" => array("type"=>"city", "pop"=>613190, "name_ru"=>"Эль Пасо", "name_en"=>"El Paso", "lang"=>"en", "lat"=>31.849250, "lon"=>-106.437549),
"/US/US-TX/eldorado-20406" => array("type"=>"town", "pop"=>1853, "name_ru"=>"Эльдорадо", "name_en"=>"Eldorado", "lang"=>"en", "lat"=>30.861992, "lon"=>-100.597699),
"/US/US-TX/electra-20407" => array("type"=>"city", "pop"=>2896, "name_en"=>"Electra", "lang"=>"en", "lat"=>34.030282, "lon"=>-98.918473),
"/US/US-TX/elgin-20408" => array("type"=>"city", "pop"=>9909, "name_ru"=>"Элгин", "name_en"=>"Elgin", "lang"=>"en", "lat"=>30.349190, "lon"=>-97.372512),
"/US/US-TX/elkhart-20409" => array("type"=>"town", "pop"=>1272, "name_ru"=>"Элкхарт", "name_en"=>"Elkhart", "lang"=>"en", "lat"=>31.628255, "lon"=>-95.578357),
"/US/US-TX/elmendorf-20411" => array("type"=>"city", "pop"=>1365, "name_en"=>"Elmendorf", "lang"=>"en", "lat"=>29.258533, "lon"=>-98.334604),
"/US/US-TX/elsa-20413" => array("type"=>"city", "pop"=>6624, "name_en"=>"Elsa", "lang"=>"en", "lat"=>26.298274, "lon"=>-97.993419),
"/US/US-TX/emhouse-20414" => array("type"=>"town", "pop"=>176, "name_en"=>"Emhouse", "lang"=>"en", "lat"=>32.160630, "lon"=>-96.577629),
"/US/US-TX/emory-20415" => array("type"=>"city", "pop"=>1490, "name_en"=>"Emory", "lang"=>"en", "lat"=>32.875705, "lon"=>-95.768797),
"/US/US-TX/encantada-ranchito-el-calaboz-20416" => array("type"=>"cdp", "pop"=>1143, "name_en"=>"Encantada-Ranchito El Calaboz", "lang"=>"en", "lat"=>26.030150, "lon"=>-97.629121),
"/US/US-TX/enchanted-oaks-20417" => array("type"=>"town", "pop"=>430, "name_en"=>"Enchanted Oaks", "lang"=>"en", "lat"=>32.266332, "lon"=>-96.107678),
"/US/US-TX/encinal-20418" => array("type"=>"city", "pop"=>685, "name_en"=>"Encinal", "lang"=>"en", "lat"=>28.041252, "lon"=>-99.354729),
"/US/US-TX/ennis-20419" => array("type"=>"city", "pop"=>19506, "name_en"=>"Ennis", "lang"=>"en", "lat"=>32.338031, "lon"=>-96.629090),
"/US/US-TX/escobares-20420" => array("type"=>"cdp", "pop"=>1330, "name_en"=>"Escobares", "lang"=>"en", "lat"=>26.404346, "lon"=>-98.968631),
"/US/US-TX/estelline-20421" => array("type"=>"town", "pop"=>155, "name_en"=>"Estelline", "lang"=>"en", "lat"=>34.547126, "lon"=>-100.439381),
"/US/US-TX/euless-20422" => array("type"=>"city", "pop"=>52579, "name_en"=>"Euless", "lang"=>"en", "lat"=>32.849108, "lon"=>-97.077296),
"/US/US-TX/eureka-20423" => array("type"=>"city", "pop"=>376, "name_ru"=>"Юрика", "name_en"=>"Eureka", "lang"=>"en", "lat"=>32.009718, "lon"=>-96.285939),
"/US/US-TX/eustace-20424" => array("type"=>"city", "pop"=>925, "name_en"=>"Eustace", "lang"=>"en", "lat"=>32.305902, "lon"=>-96.012142),
"/US/US-TX/evadale-20425" => array("type"=>"cdp", "pop"=>1422, "name_en"=>"Evadale", "lang"=>"en", "lat"=>30.338643, "lon"=>-94.061173),
"/US/US-TX/evant-20426" => array("type"=>"town", "pop"=>371, "name_en"=>"Evant", "lang"=>"en", "lat"=>31.475768, "lon"=>-98.149634),
"/US/US-TX/everman-20427" => array("type"=>"city", "pop"=>5771, "name_en"=>"Everman", "lang"=>"en", "lat"=>32.630150, "lon"=>-97.282600),
"/US/US-TX/fabens-20428" => array("type"=>"cdp", "pop"=>5599, "name_en"=>"Fabens", "lang"=>"en", "lat"=>31.507999, "lon"=>-106.153786),
"/US/US-TX/fair-oaks-ranch-20430" => array("type"=>"city", "pop"=>6367, "name_en"=>"Fair Oaks Ranch", "lang"=>"en", "lat"=>29.736070, "lon"=>-98.635637),
"/US/US-TX/fairfield-20429" => array("type"=>"city", "pop"=>3618, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>31.718787, "lon"=>-96.165519),
"/US/US-TX/fairview-20431" => array("type"=>"city", "pop"=>8093, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>33.113851, "lon"=>-97.444747),
"/US/US-TX/fairview-20432" => array("type"=>"town", "pop"=>1554, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>33.142201, "lon"=>-96.624883),
"/US/US-TX/falfurrias-20433" => array("type"=>"city", "pop"=>4941, "name_en"=>"Falfurrias", "lang"=>"en", "lat"=>27.223062, "lon"=>-98.145021),
"/US/US-TX/falls-city-20434" => array("type"=>"city", "pop"=>598, "name_en"=>"Falls City", "lang"=>"en", "lat"=>28.981316, "lon"=>-98.021324),
"/US/US-TX/farmers-branch-20435" => array("type"=>"city", "pop"=>26455, "name_en"=>"Farmers Branch", "lang"=>"en", "lat"=>32.926801, "lon"=>-96.878919),
"/US/US-TX/farmersville-20436" => array("type"=>"city", "pop"=>3476, "name_ru"=>"Фармерсвилл", "name_en"=>"Farmersville", "lang"=>"en", "lat"=>33.161586, "lon"=>-96.357622),
"/US/US-TX/farwell-20437" => array("type"=>"city", "pop"=>1251, "name_ru"=>"Фаруэлл", "name_en"=>"Farwell", "lang"=>"en", "lat"=>34.385408, "lon"=>-103.036498),
"/US/US-TX/fate-20438" => array("type"=>"city", "pop"=>5679, "name_en"=>"Fate", "lang"=>"en", "lat"=>32.940793, "lon"=>-96.386293),
"/US/US-TX/fayetteville-20439" => array("type"=>"town", "pop"=>279, "name_ru"=>"Фейетвилл", "name_en"=>"Fayetteville", "lang"=>"en", "lat"=>29.906426, "lon"=>-96.674428),
"/US/US-TX/ferris-20440" => array("type"=>"city", "pop"=>2566, "name_ru"=>"Феррис", "name_en"=>"Ferris", "lang"=>"en", "lat"=>32.535121, "lon"=>-96.663852),
"/US/US-TX/first-colony-20441" => array("type"=>"cdp", "pop"=>18327, "name_en"=>"First Colony", "lang"=>"en", "lat"=>29.574536, "lon"=>-95.608309),
"/US/US-TX/flatonia-20442" => array("type"=>"town", "pop"=>1431, "name_en"=>"Flatonia", "lang"=>"en", "lat"=>29.687299, "lon"=>-97.107465),
"/US/US-TX/florence-20443" => array("type"=>"town", "pop"=>1129, "name_ru"=>"Флоренс", "name_en"=>"Florence", "lang"=>"en", "lat"=>30.839670, "lon"=>-97.792648),
"/US/US-TX/floresville-20444" => array("type"=>"city", "pop"=>7639, "name_en"=>"Floresville", "lang"=>"en", "lat"=>29.136825, "lon"=>-98.156525),
"/US/US-TX/flower-mound-20445" => array("type"=>"town", "pop"=>69307, "name_en"=>"Flower Mound", "lang"=>"en", "lat"=>33.028333, "lon"=>-97.092668),
"/US/US-TX/floydada-20446" => array("type"=>"city", "pop"=>3059, "name_en"=>"Floydada", "lang"=>"en", "lat"=>33.983693, "lon"=>-101.336824),
"/US/US-TX/follett-20447" => array("type"=>"city", "pop"=>396, "name_en"=>"Follett", "lang"=>"en", "lat"=>36.434176, "lon"=>-100.140677),
"/US/US-TX/forest-hill-20448" => array("type"=>"city", "pop"=>13858, "name_ru"=>"Форест Хилл", "name_en"=>"Forest Hill", "lang"=>"en", "lat"=>32.661450, "lon"=>-97.265350),
"/US/US-TX/forney-20449" => array("type"=>"town", "pop"=>15223, "name_en"=>"Forney", "lang"=>"en", "lat"=>32.752687, "lon"=>-96.466119),
"/US/US-TX/forsan-20450" => array("type"=>"city", "pop"=>221, "name_en"=>"Forsan", "lang"=>"en", "lat"=>32.110218, "lon"=>-101.366063),
"/US/US-TX/fort-bliss-20451" => array("type"=>"cdp", "pop"=>13915, "name_en"=>"Fort Bliss", "lang"=>"en", "lat"=>31.807550, "lon"=>-106.421663),
"/US/US-TX/fort-gates-20452" => array("type"=>"city", "pop"=>818, "name_en"=>"Fort Gates", "lang"=>"en", "lat"=>31.403321, "lon"=>-97.709568),
"/US/US-TX/fort-hood-20453" => array("type"=>"cdp", "pop"=>35580, "name_en"=>"Fort Hood", "lang"=>"en", "lat"=>31.133792, "lon"=>-97.774126),
"/US/US-TX/fort-stockton-20454" => array("type"=>"city", "pop"=>7526, "name_en"=>"Fort Stockton", "lang"=>"en", "lat"=>30.891239, "lon"=>-102.885710),
"/US/US-TX/fort-worth-20455" => array("type"=>"city", "pop"=>703073, "name_en"=>"Fort Worth", "lang"=>"en", "lat"=>32.753901, "lon"=>-97.336249),
"/US/US-TX/franklin-20456" => array("type"=>"city", "pop"=>1443, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>31.024934, "lon"=>-96.485662),
"/US/US-TX/frankston-20457" => array("type"=>"town", "pop"=>1234, "name_en"=>"Frankston", "lang"=>"en", "lat"=>32.055096, "lon"=>-95.503634),
"/US/US-TX/fredericksburg-20458" => array("type"=>"city", "pop"=>11098, "name_ru"=>"Фредериксбург", "name_en"=>"Fredericksburg", "lang"=>"en", "lat"=>30.270482, "lon"=>-98.872241),
"/US/US-TX/freeport-20459" => array("type"=>"city", "pop"=>12494, "name_ru"=>"Фрипорт", "name_en"=>"Freeport", "lang"=>"en", "lat"=>28.951365, "lon"=>-95.364284),
"/US/US-TX/freer-20460" => array("type"=>"city", "pop"=>2936, "name_en"=>"Freer", "lang"=>"en", "lat"=>27.881416, "lon"=>-98.618224),
"/US/US-TX/fresno-20461" => array("type"=>"cdp", "pop"=>3182, "name_ru"=>"Фресно", "name_en"=>"Fresno", "lang"=>"en", "lat"=>29.537388, "lon"=>-95.456933),
"/US/US-TX/friendswood-20462" => array("type"=>"city", "pop"=>33979, "name_en"=>"Friendswood", "lang"=>"en", "lat"=>29.512299, "lon"=>-95.197722),
"/US/US-TX/friona-20463" => array("type"=>"city", "pop"=>3522, "name_en"=>"Friona", "lang"=>"en", "lat"=>34.638920, "lon"=>-102.722332),
"/US/US-TX/frisco-20464" => array("type"=>"city", "pop"=>96676, "name_ru"=>"Фриско", "name_en"=>"Frisco", "lang"=>"en", "lat"=>33.144503, "lon"=>-96.812957),
"/US/US-TX/fritch-20465" => array("type"=>"city", "pop"=>2006, "name_en"=>"Fritch", "lang"=>"en", "lat"=>35.640640, "lon"=>-101.596846),
"/US/US-TX/frost-20466" => array("type"=>"town", "pop"=>725, "name_ru"=>"Фрост", "name_en"=>"Frost", "lang"=>"en", "lat"=>32.078147, "lon"=>-96.807965),
"/US/US-TX/fruitvale-20467" => array("type"=>"city", "pop"=>450, "name_ru"=>"Фрутвейл", "name_en"=>"Fruitvale", "lang"=>"en", "lat"=>32.683933, "lon"=>-95.803619),
"/US/US-TX/fulshear-20468" => array("type"=>"city", "pop"=>954, "name_en"=>"Fulshear", "lang"=>"en", "lat"=>29.691135, "lon"=>-95.884217),
"/US/US-TX/fulton-20469" => array("type"=>"town", "pop"=>1653, "name_ru"=>"Фултон", "name_en"=>"Fulton", "lang"=>"en", "lat"=>28.071287, "lon"=>-97.035753),
"/US/US-TX/gainesville-20470" => array("type"=>"city", "pop"=>16452, "name_ru"=>"Гейнсвилл", "name_en"=>"Gainesville", "lang"=>"en", "lat"=>33.637250, "lon"=>-97.150579),
"/US/US-TX/galena-park-20471" => array("type"=>"city", "pop"=>10207, "name_en"=>"Galena Park", "lang"=>"en", "lat"=>29.744900, "lon"=>-95.234221),
"/US/US-TX/gallatin-20472" => array("type"=>"city", "pop"=>392, "name_ru"=>"Галлатин", "name_en"=>"Gallatin", "lang"=>"en", "lat"=>31.895107, "lon"=>-95.152518),
"/US/US-TX/galveston-20473" => array("type"=>"city", "pop"=>57086, "name_ru"=>"Галвестон", "name_en"=>"Galveston", "lang"=>"en", "lat"=>29.229337, "lon"=>-94.892064),
"/US/US-TX/ganado-20474" => array("type"=>"town", "pop"=>1841, "name_ru"=>"Ганадо", "name_en"=>"Ganado", "lang"=>"en", "lat"=>29.042067, "lon"=>-96.511402),
"/US/US-TX/garden-ridge-20476" => array("type"=>"city", "pop"=>3205, "name_en"=>"Garden Ridge", "lang"=>"en", "lat"=>29.634700, "lon"=>-98.293628),
"/US/US-TX/gardendale-20475" => array("type"=>"cdp", "pop"=>1103, "name_ru"=>"Гардендейл", "name_en"=>"Gardendale", "lang"=>"en", "lat"=>32.020401, "lon"=>-102.361341),
"/US/US-TX/garfield-20477" => array("type"=>"cdp", "pop"=>1336, "name_ru"=>"Гарфилд", "name_en"=>"Garfield", "lang"=>"en", "lat"=>30.189728, "lon"=>-97.560607),
"/US/US-TX/garland-20478" => array("type"=>"city", "pop"=>218577, "name_ru"=>"Гарленд", "name_en"=>"Garland", "lang"=>"en", "lat"=>32.910800, "lon"=>-96.629300),
"/US/US-TX/garrett-20479" => array("type"=>"town", "pop"=>740, "name_ru"=>"Гарретт", "name_en"=>"Garrett", "lang"=>"en", "lat"=>32.360840, "lon"=>-96.651755),
"/US/US-TX/garrison-20480" => array("type"=>"town", "pop"=>858, "name_ru"=>"Гаррисон", "name_en"=>"Garrison", "lang"=>"en", "lat"=>31.824233, "lon"=>-94.492457),
"/US/US-TX/gary-city-20481" => array("type"=>"town", "pop"=>308, "name_en"=>"Gary City", "lang"=>"en", "lat"=>32.027922, "lon"=>-94.368252),
"/US/US-TX/gatesville-20482" => array("type"=>"city", "pop"=>15252, "name_en"=>"Gatesville", "lang"=>"en", "lat"=>31.440313, "lon"=>-97.733976),
"/US/US-TX/george-west-20484" => array("type"=>"city", "pop"=>2240, "name_en"=>"George West", "lang"=>"en", "lat"=>28.330273, "lon"=>-98.117694),
"/US/US-TX/georgetown-20483" => array("type"=>"city", "pop"=>49618, "name_ru"=>"Джорджтаун", "name_en"=>"Georgetown", "lang"=>"en", "lat"=>30.647568, "lon"=>-97.686263),
"/US/US-TX/gholson-20485" => array("type"=>"city", "pop"=>977, "name_en"=>"Gholson", "lang"=>"en", "lat"=>31.711741, "lon"=>-97.236792),
"/US/US-TX/giddings-20486" => array("type"=>"city", "pop"=>5423, "name_en"=>"Giddings", "lang"=>"en", "lat"=>30.183344, "lon"=>-96.927851),
"/US/US-TX/gilmer-20487" => array("type"=>"city", "pop"=>5249, "name_en"=>"Gilmer", "lang"=>"en", "lat"=>32.730232, "lon"=>-94.945664),
"/US/US-TX/gladewater-20488" => array("type"=>"city", "pop"=>6275, "name_en"=>"Gladewater", "lang"=>"en", "lat"=>32.543049, "lon"=>-94.945071),
"/US/US-TX/glen-rose-20490" => array("type"=>"city", "pop"=>2906, "name_en"=>"Glen Rose", "lang"=>"en", "lat"=>32.236280, "lon"=>-97.754637),
"/US/US-TX/glenn-heights-20489" => array("type"=>"city", "pop"=>11078, "name_en"=>"Glenn Heights", "lang"=>"en", "lat"=>32.550974, "lon"=>-96.853872),
"/US/US-TX/godley-20491" => array("type"=>"town", "pop"=>1009, "name_ru"=>"Годли", "name_en"=>"Godley", "lang"=>"en", "lat"=>32.449473, "lon"=>-97.526579),
"/US/US-TX/goldsmith-20492" => array("type"=>"city", "pop"=>274, "name_en"=>"Goldsmith", "lang"=>"en", "lat"=>31.983080, "lon"=>-102.616513),
"/US/US-TX/goldthwaite-20493" => array("type"=>"city", "pop"=>1736, "name_en"=>"Goldthwaite", "lang"=>"en", "lat"=>31.453867, "lon"=>-98.571022),
"/US/US-TX/goliad-20494" => array("type"=>"city", "pop"=>2017, "name_en"=>"Goliad", "lang"=>"en", "lat"=>28.669730, "lon"=>-97.391579),
"/US/US-TX/golinda-20495" => array("type"=>"city", "pop"=>399, "name_en"=>"Golinda", "lang"=>"en", "lat"=>31.370353, "lon"=>-97.081078),
"/US/US-TX/gonzales-20496" => array("type"=>"city", "pop"=>7298, "name_ru"=>"Гонзалес", "name_en"=>"Gonzales", "lang"=>"en", "lat"=>29.508600, "lon"=>-97.447535),
"/US/US-TX/goodlow-20497" => array("type"=>"city", "pop"=>287, "name_en"=>"Goodlow", "lang"=>"en", "lat"=>32.112633, "lon"=>-96.216354),
"/US/US-TX/goodrich-20498" => array("type"=>"city", "pop"=>266, "name_ru"=>"Гудрич", "name_en"=>"Goodrich", "lang"=>"en", "lat"=>30.609599, "lon"=>-94.945119),
"/US/US-TX/gordon-20499" => array("type"=>"city", "pop"=>467, "name_ru"=>"Гордон", "name_en"=>"Gordon", "lang"=>"en", "lat"=>32.545847, "lon"=>-98.365803),
"/US/US-TX/goree-20500" => array("type"=>"city", "pop"=>257, "name_en"=>"Goree", "lang"=>"en", "lat"=>33.467614, "lon"=>-99.523474),
"/US/US-TX/gorman-20501" => array("type"=>"city", "pop"=>1231, "name_en"=>"Gorman", "lang"=>"en", "lat"=>32.213970, "lon"=>-98.671815),
"/US/US-TX/graford-20502" => array("type"=>"city", "pop"=>605, "name_en"=>"Graford", "lang"=>"en", "lat"=>32.936677, "lon"=>-98.247548),
"/US/US-TX/graham-20503" => array("type"=>"city", "pop"=>8479, "name_ru"=>"Грэхэм", "name_en"=>"Graham", "lang"=>"en", "lat"=>33.100674, "lon"=>-98.577449),
"/US/US-TX/granbury-20504" => array("type"=>"city", "pop"=>8620, "name_en"=>"Granbury", "lang"=>"en", "lat"=>32.440192, "lon"=>-97.787680),
"/US/US-TX/grand-prairie-20506" => array("type"=>"city", "pop"=>160641, "name_en"=>"Grand Prairie", "lang"=>"en", "lat"=>32.692239, "lon"=>-97.016653),
"/US/US-TX/grand-saline-20507" => array("type"=>"city", "pop"=>3171, "name_en"=>"Grand Saline", "lang"=>"en", "lat"=>32.677532, "lon"=>-95.711231),
"/US/US-TX/grandfalls-20505" => array("type"=>"town", "pop"=>386, "name_en"=>"Grandfalls", "lang"=>"en", "lat"=>31.340639, "lon"=>-102.854354),
"/US/US-TX/grandview-20508" => array("type"=>"city", "pop"=>1583, "name_ru"=>"Грандвью", "name_en"=>"Grandview", "lang"=>"en", "lat"=>32.267981, "lon"=>-97.177240),
"/US/US-TX/granger-20509" => array("type"=>"city", "pop"=>1363, "name_ru"=>"Грейнджер", "name_en"=>"Granger", "lang"=>"en", "lat"=>30.717889, "lon"=>-97.440716),
"/US/US-TX/granite-shoals-20510" => array("type"=>"city", "pop"=>2828, "name_en"=>"Granite Shoals", "lang"=>"en", "lat"=>30.588670, "lon"=>-98.383752),
"/US/US-TX/grapeland-20511" => array("type"=>"city", "pop"=>1380, "name_en"=>"Grapeland", "lang"=>"en", "lat"=>31.491258, "lon"=>-95.481772),
"/US/US-TX/grapevine-20512" => array("type"=>"city", "pop"=>50471, "name_en"=>"Grapevine", "lang"=>"en", "lat"=>32.933498, "lon"=>-97.071677),
"/US/US-TX/grayburg-20513" => array("type"=>"city", "pop"=>257, "name_en"=>"Grayburg", "lang"=>"en", "lat"=>30.114407, "lon"=>-94.411783),
"/US/US-TX/grays-prairie-20514" => array("type"=>"village", "pop"=>354, "name_en"=>"Grays Prairie", "lang"=>"en", "lat"=>32.472479, "lon"=>-96.349999),
"/US/US-TX/greenville-20515" => array("type"=>"city", "pop"=>25676, "name_ru"=>"Гринвилл", "name_en"=>"Greenville", "lang"=>"en", "lat"=>33.116861, "lon"=>-96.101188),
"/US/US-TX/gregory-20516" => array("type"=>"city", "pop"=>2200, "name_en"=>"Gregory", "lang"=>"en", "lat"=>27.922104, "lon"=>-97.290755),
"/US/US-TX/grey-forest-20517" => array("type"=>"city", "pop"=>442, "name_en"=>"Grey Forest", "lang"=>"en", "lat"=>29.614150, "lon"=>-98.682410),
"/US/US-TX/groesbeck-20518" => array("type"=>"city", "pop"=>4280, "name_en"=>"Groesbeck", "lang"=>"en", "lat"=>31.524508, "lon"=>-96.531059),
"/US/US-TX/groom-20519" => array("type"=>"city", "pop"=>558, "name_en"=>"Groom", "lang"=>"en", "lat"=>35.204082, "lon"=>-101.106005),
"/US/US-TX/groves-20520" => array("type"=>"city", "pop"=>14393, "name_en"=>"Groves", "lang"=>"en", "lat"=>29.945850, "lon"=>-93.916668),
"/US/US-TX/groveton-20521" => array("type"=>"city", "pop"=>1116, "name_en"=>"Groveton", "lang"=>"en", "lat"=>31.057280, "lon"=>-95.127098),
"/US/US-TX/gruver-20522" => array("type"=>"city", "pop"=>1127, "name_ru"=>"Грувер", "name_en"=>"Gruver", "lang"=>"en", "lat"=>36.256543, "lon"=>-101.411142),
"/US/US-TX/gun-barrel-city-20523" => array("type"=>"town", "pop"=>5975, "name_en"=>"Gun Barrel City", "lang"=>"en", "lat"=>32.327383, "lon"=>-96.136445),
"/US/US-TX/gunter-20524" => array("type"=>"town", "pop"=>1332, "name_en"=>"Gunter", "lang"=>"en", "lat"=>33.450764, "lon"=>-96.741856),
"/US/US-TX/gustine-20525" => array("type"=>"town", "pop"=>442, "name_ru"=>"Гастин", "name_en"=>"Gustine", "lang"=>"en", "lat"=>31.845337, "lon"=>-98.402127),
"/US/US-TX/hackberry-20526" => array("type"=>"town", "pop"=>644, "name_ru"=>"Хэкберри", "name_en"=>"Hackberry", "lang"=>"en", "lat"=>33.150014, "lon"=>-96.918427),
"/US/US-TX/hale-center-20527" => array("type"=>"city", "pop"=>2117, "name_en"=>"Hale Center", "lang"=>"en", "lat"=>34.064719, "lon"=>-101.844498),
"/US/US-TX/hallettsville-20528" => array("type"=>"city", "pop"=>2490, "name_en"=>"Hallettsville", "lang"=>"en", "lat"=>29.444950, "lon"=>-96.941751),
"/US/US-TX/hallsburg-20529" => array("type"=>"city", "pop"=>492, "name_en"=>"Hallsburg", "lang"=>"en", "lat"=>31.545794, "lon"=>-96.946773),
"/US/US-TX/hallsville-20530" => array("type"=>"city", "pop"=>3010, "name_ru"=>"Холлсвилл", "name_en"=>"Hallsville", "lang"=>"en", "lat"=>32.503404, "lon"=>-94.576118),
"/US/US-TX/haltom-city-20531" => array("type"=>"city", "pop"=>40181, "name_en"=>"Haltom City", "lang"=>"en", "lat"=>32.817600, "lon"=>-97.270650),
"/US/US-TX/hamilton-20532" => array("type"=>"city", "pop"=>2922, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>31.705027, "lon"=>-98.121184),
"/US/US-TX/hamlin-20533" => array("type"=>"city", "pop"=>1904, "name_ru"=>"Хэмлин", "name_en"=>"Hamlin", "lang"=>"en", "lat"=>32.888913, "lon"=>-100.131773),
"/US/US-TX/happy-20534" => array("type"=>"town", "pop"=>606, "name_en"=>"Happy", "lang"=>"en", "lat"=>34.741345, "lon"=>-101.855872),
"/US/US-TX/hardin-20535" => array("type"=>"town", "pop"=>791, "name_ru"=>"Хардин", "name_en"=>"Hardin", "lang"=>"en", "lat"=>30.149958, "lon"=>-94.736431),
"/US/US-TX/harker-heights-20536" => array("type"=>"city", "pop"=>25417, "name_en"=>"Harker Heights", "lang"=>"en", "lat"=>31.061200, "lon"=>-97.654101),
"/US/US-TX/harlingen-20537" => array("type"=>"city", "pop"=>64843, "name_en"=>"Harlingen", "lang"=>"en", "lat"=>26.195300, "lon"=>-97.686274),
"/US/US-TX/hart-20538" => array("type"=>"city", "pop"=>1029, "name_ru"=>"Харт", "name_en"=>"Hart", "lang"=>"en", "lat"=>34.385644, "lon"=>-102.114592),
"/US/US-TX/haskell-20539" => array("type"=>"city", "pop"=>2623, "name_ru"=>"Хаскел", "name_en"=>"Haskell", "lang"=>"en", "lat"=>33.158984, "lon"=>-99.731813),
"/US/US-TX/haslet-20540" => array("type"=>"city", "pop"=>1721, "name_en"=>"Haslet", "lang"=>"en", "lat"=>32.958235, "lon"=>-97.339744),
"/US/US-TX/hawkins-20541" => array("type"=>"city", "pop"=>1524, "name_en"=>"Hawkins", "lang"=>"en", "lat"=>32.590844, "lon"=>-95.202452),
"/US/US-TX/hawley-20542" => array("type"=>"city", "pop"=>576, "name_ru"=>"Холи", "name_en"=>"Hawley", "lang"=>"en", "lat"=>32.601488, "lon"=>-99.810131),
"/US/US-TX/hays-20543" => array("type"=>"city", "pop"=>254, "name_en"=>"Hays", "lang"=>"en", "lat"=>30.122773, "lon"=>-97.870983),
"/US/US-TX/hearne-20544" => array("type"=>"city", "pop"=>4578, "name_en"=>"Hearne", "lang"=>"en", "lat"=>30.876997, "lon"=>-96.594842),
"/US/US-TX/heath-20545" => array("type"=>"city", "pop"=>7793, "name_ru"=>"Хит", "name_en"=>"Heath", "lang"=>"en", "lat"=>32.847765, "lon"=>-96.478009),
"/US/US-TX/hebbronville-20546" => array("type"=>"cdp", "pop"=>4465, "name_en"=>"Hebbronville", "lang"=>"en", "lat"=>27.317440, "lon"=>-98.686020),
"/US/US-TX/hebron-20547" => array("type"=>"town", "pop"=>261, "name_ru"=>"Хеброн", "name_en"=>"Hebron", "lang"=>"en", "lat"=>33.042290, "lon"=>-96.899256),
"/US/US-TX/hedley-20548" => array("type"=>"town", "pop"=>373, "name_en"=>"Hedley", "lang"=>"en", "lat"=>34.867047, "lon"=>-100.659350),
"/US/US-TX/hedwig-village-20549" => array("type"=>"city", "pop"=>2344, "name_en"=>"Hedwig Village", "lang"=>"en", "lat"=>29.779762, "lon"=>-95.519195),
"/US/US-TX/helotes-20550" => array("type"=>"city", "pop"=>7831, "name_en"=>"Helotes", "lang"=>"en", "lat"=>29.568090, "lon"=>-98.693528),
"/US/US-TX/hemphill-20551" => array("type"=>"city", "pop"=>1026, "name_en"=>"Hemphill", "lang"=>"en", "lat"=>31.341751, "lon"=>-93.851309),
"/US/US-TX/hempstead-20552" => array("type"=>"city", "pop"=>7401, "name_en"=>"Hempstead", "lang"=>"en", "lat"=>30.093076, "lon"=>-96.080365),
"/US/US-TX/henderson-20553" => array("type"=>"city", "pop"=>11631, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>32.157005, "lon"=>-94.797235),
"/US/US-TX/henrietta-20554" => array("type"=>"city", "pop"=>3213, "name_ru"=>"Хенриетта", "name_en"=>"Henrietta", "lang"=>"en", "lat"=>33.816777, "lon"=>-98.194477),
"/US/US-TX/hereford-20555" => array("type"=>"city", "pop"=>14499, "name_en"=>"Hereford", "lang"=>"en", "lat"=>34.823370, "lon"=>-102.398399),
"/US/US-TX/hewitt-20556" => array("type"=>"city", "pop"=>13561, "name_ru"=>"Хьюитт", "name_en"=>"Hewitt", "lang"=>"en", "lat"=>31.451103, "lon"=>-97.193625),
"/US/US-TX/hickory-creek-20557" => array("type"=>"town", "pop"=>3779, "name_en"=>"Hickory Creek", "lang"=>"en", "lat"=>33.109334, "lon"=>-97.028239),
"/US/US-TX/hico-20558" => array("type"=>"city", "pop"=>1342, "name_en"=>"Hico", "lang"=>"en", "lat"=>31.983407, "lon"=>-98.028475),
"/US/US-TX/hidalgo-20559" => array("type"=>"city", "pop"=>11984, "name_ru"=>"Идальго", "name_en"=>"Hidalgo", "lang"=>"en", "lat"=>26.101848, "lon"=>-98.258372),
"/US/US-TX/higgins-20560" => array("type"=>"city", "pop"=>422, "name_en"=>"Higgins", "lang"=>"en", "lat"=>36.120341, "lon"=>-100.026945),
"/US/US-TX/highland-park-20561" => array("type"=>"town", "pop"=>9184, "name_ru"=>"Хайленд Парк", "name_en"=>"Highland Park", "lang"=>"en", "lat"=>32.831451, "lon"=>-96.800823),
"/US/US-TX/highland-village-20563" => array("type"=>"city", "pop"=>16737, "name_en"=>"Highland Village", "lang"=>"en", "lat"=>33.088141, "lon"=>-97.055625),
"/US/US-TX/highlands-20562" => array("type"=>"cdp", "pop"=>6632, "name_ru"=>"Хайлендс", "name_en"=>"Highlands", "lang"=>"en", "lat"=>29.812664, "lon"=>-95.057453),
"/US/US-TX/hill-country-village-20564" => array("type"=>"city", "pop"=>1114, "name_en"=>"Hill Country Village", "lang"=>"en", "lat"=>29.583850, "lon"=>-98.491758),
"/US/US-TX/hillcrest-20565" => array("type"=>"village", "pop"=>723, "name_ru"=>"Хиллкрест", "name_en"=>"Hillcrest", "lang"=>"en", "lat"=>29.392087, "lon"=>-95.222649),
"/US/US-TX/hillsboro-20566" => array("type"=>"city", "pop"=>8941, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>32.007368, "lon"=>-97.115579),
"/US/US-TX/hilshire-village-20567" => array("type"=>"city", "pop"=>736, "name_en"=>"Hilshire Village", "lang"=>"en", "lat"=>29.790167, "lon"=>-95.487959),
"/US/US-TX/hitchcock-20568" => array("type"=>"city", "pop"=>7201, "name_en"=>"Hitchcock", "lang"=>"en", "lat"=>29.315712, "lon"=>-95.017286),
"/US/US-TX/holiday-lakes-20569" => array("type"=>"town", "pop"=>1094, "name_en"=>"Holiday Lakes", "lang"=>"en", "lat"=>29.208806, "lon"=>-95.514208),
"/US/US-TX/holland-20570" => array("type"=>"town", "pop"=>1140, "name_ru"=>"Холланд", "name_en"=>"Holland", "lang"=>"en", "lat"=>30.883835, "lon"=>-97.403547),
"/US/US-TX/holliday-20571" => array("type"=>"city", "pop"=>1801, "name_ru"=>"Холлидей", "name_en"=>"Holliday", "lang"=>"en", "lat"=>33.812930, "lon"=>-98.689732),
"/US/US-TX/hollywood-park-20572" => array("type"=>"town", "pop"=>3307, "name_en"=>"Hollywood Park", "lang"=>"en", "lat"=>29.598900, "lon"=>-98.484902),
"/US/US-TX/homestead-meadows-20573" => array("type"=>"cdp", "pop"=>4978, "name_en"=>"Homestead Meadows", "lang"=>"en", "lat"=>31.841576, "lon"=>-106.175322),
"/US/US-TX/hondo-20574" => array("type"=>"city", "pop"=>8992, "name_en"=>"Hondo", "lang"=>"en", "lat"=>29.356323, "lon"=>-99.158302),
"/US/US-TX/honey-grove-20575" => array("type"=>"city", "pop"=>1839, "name_en"=>"Honey Grove", "lang"=>"en", "lat"=>33.583424, "lon"=>-95.909922),
"/US/US-TX/hooks-20576" => array("type"=>"city", "pop"=>2954, "name_en"=>"Hooks", "lang"=>"en", "lat"=>33.469285, "lon"=>-94.284958),
"/US/US-TX/horizon-city-20577" => array("type"=>"town", "pop"=>13043, "name_en"=>"Horizon City", "lang"=>"en", "lat"=>31.670082, "lon"=>-106.191297),
"/US/US-TX/horseshoe-bay-20578" => array("type"=>"cdp", "pop"=>2491, "name_en"=>"Horseshoe Bay", "lang"=>"en", "lat"=>30.537108, "lon"=>-98.354801),
"/US/US-TX/houston-20579" => array("type"=>"city", "pop"=>2242193, "name_ru"=>"Хьюстон", "name_en"=>"Houston", "lang"=>"en", "lat"=>29.768700, "lon"=>-95.386728),
"/US/US-TX/howardwick-20580" => array("type"=>"city", "pop"=>440, "name_en"=>"Howardwick", "lang"=>"en", "lat"=>35.035233, "lon"=>-100.908736),
"/US/US-TX/howe-20581" => array("type"=>"town", "pop"=>2694, "name_en"=>"Howe", "lang"=>"en", "lat"=>33.512783, "lon"=>-96.613138),
"/US/US-TX/hubbard-20582" => array("type"=>"city", "pop"=>1752, "name_ru"=>"Хаббард", "name_en"=>"Hubbard", "lang"=>"en", "lat"=>31.845864, "lon"=>-96.800536),
"/US/US-TX/hudson-20583" => array("type"=>"city", "pop"=>4310, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>31.330081, "lon"=>-94.797150),
"/US/US-TX/hudson-oaks-20584" => array("type"=>"town", "pop"=>2070, "name_en"=>"Hudson Oaks", "lang"=>"en", "lat"=>32.747254, "lon"=>-97.701061),
"/US/US-TX/hughes-springs-20585" => array("type"=>"city", "pop"=>1827, "name_en"=>"Hughes Springs", "lang"=>"en", "lat"=>32.998755, "lon"=>-94.630224),
"/US/US-TX/humble-20586" => array("type"=>"city", "pop"=>14899, "name_en"=>"Humble", "lang"=>"en", "lat"=>29.994700, "lon"=>-95.264662),
"/US/US-TX/hunters-creek-village-20587" => array("type"=>"city", "pop"=>4577, "name_en"=>"Hunters Creek Village", "lang"=>"en", "lat"=>29.773800, "lon"=>-95.497688),
"/US/US-TX/huntington-20588" => array("type"=>"city", "pop"=>2109, "name_ru"=>"Хантингтон", "name_en"=>"Huntington", "lang"=>"en", "lat"=>31.279818, "lon"=>-94.577527),
"/US/US-TX/huntsville-20589" => array("type"=>"city", "pop"=>38480, "name_ru"=>"Хантсвилл", "name_en"=>"Huntsville", "lang"=>"en", "lat"=>30.711896, "lon"=>-95.549817),
"/US/US-TX/hurst-20590" => array("type"=>"city", "pop"=>38550, "name_ru"=>"Херст", "name_en"=>"Hurst", "lang"=>"en", "lat"=>32.835501, "lon"=>-97.179882),
"/US/US-TX/hutchins-20591" => array("type"=>"city", "pop"=>3095, "name_en"=>"Hutchins", "lang"=>"en", "lat"=>32.643651, "lon"=>-96.707264),
"/US/US-TX/hutto-20592" => array("type"=>"town", "pop"=>13599, "name_en"=>"Hutto", "lang"=>"en", "lat"=>30.545808, "lon"=>-97.542102),
"/US/US-TX/huxley-20593" => array("type"=>"city", "pop"=>320, "name_ru"=>"Хаксли", "name_en"=>"Huxley", "lang"=>"en", "lat"=>31.762237, "lon"=>-93.887379),
"/US/US-TX/idalou-20594" => array("type"=>"town", "pop"=>2118, "name_en"=>"Idalou", "lang"=>"en", "lat"=>33.662941, "lon"=>-101.681941),
"/US/US-TX/impact-20595" => array("type"=>"town", "pop"=>39, "name_en"=>"Impact", "lang"=>"en", "lat"=>32.498750, "lon"=>-99.746274),
"/US/US-TX/indian-lake-20596" => array("type"=>"town", "pop"=>559, "name_en"=>"Indian Lake", "lang"=>"en", "lat"=>26.087300, "lon"=>-97.502210),
"/US/US-TX/inez-20597" => array("type"=>"cdp", "pop"=>1371, "name_ru"=>"Инез", "name_en"=>"Inez", "lang"=>"en", "lat"=>28.875051, "lon"=>-96.793861),
"/US/US-TX/ingleside-20598" => array("type"=>"city", "pop"=>9039, "name_en"=>"Ingleside", "lang"=>"en", "lat"=>27.870048, "lon"=>-97.200097),
"/US/US-TX/ingram-20599" => array("type"=>"city", "pop"=>1909, "name_en"=>"Ingram", "lang"=>"en", "lat"=>30.078495, "lon"=>-99.236667),
"/US/US-TX/iowa-colony-20600" => array("type"=>"village", "pop"=>944, "name_en"=>"Iowa Colony", "lang"=>"en", "lat"=>29.442726, "lon"=>-95.415389),
"/US/US-TX/iowa-park-20601" => array("type"=>"city", "pop"=>6276, "name_en"=>"Iowa Park", "lang"=>"en", "lat"=>33.959690, "lon"=>-98.680631),
"/US/US-TX/iraan-20602" => array("type"=>"city", "pop"=>1228, "name_en"=>"Iraan", "lang"=>"en", "lat"=>30.912312, "lon"=>-101.899436),
"/US/US-TX/iredell-20603" => array("type"=>"city", "pop"=>380, "name_en"=>"Iredell", "lang"=>"en", "lat"=>31.986215, "lon"=>-97.871077),
"/US/US-TX/irving-20604" => array("type"=>"city", "pop"=>201358, "name_ru"=>"Ирвинг", "name_en"=>"Irving", "lang"=>"en", "lat"=>32.857685, "lon"=>-96.969338),
"/US/US-TX/italy-20605" => array("type"=>"town", "pop"=>2144, "name_en"=>"Italy", "lang"=>"en", "lat"=>32.181389, "lon"=>-96.883881),
"/US/US-TX/itasca-20606" => array("type"=>"city", "pop"=>1696, "name_ru"=>"Итаска", "name_en"=>"Itasca", "lang"=>"en", "lat"=>32.158359, "lon"=>-97.147567),
"/US/US-TX/jacinto-city-20607" => array("type"=>"city", "pop"=>9923, "name_en"=>"Jacinto City", "lang"=>"en", "lat"=>29.766050, "lon"=>-95.240874),
"/US/US-TX/jacksboro-20608" => array("type"=>"city", "pop"=>4493, "name_en"=>"Jacksboro", "lang"=>"en", "lat"=>33.223079, "lon"=>-98.154062),
"/US/US-TX/jacksonville-20609" => array("type"=>"city", "pop"=>14367, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>31.965273, "lon"=>-95.259998),
"/US/US-TX/jamaica-beach-20610" => array("type"=>"village", "pop"=>1100, "name_en"=>"Jamaica Beach", "lang"=>"en", "lat"=>29.191160, "lon"=>-94.979770),
"/US/US-TX/jasper-20611" => array("type"=>"city", "pop"=>7353, "name_ru"=>"Джаспер", "name_en"=>"Jasper", "lang"=>"en", "lat"=>30.920447, "lon"=>-93.995064),
"/US/US-TX/jayton-20612" => array("type"=>"town", "pop"=>426, "name_en"=>"Jayton", "lang"=>"en", "lat"=>33.250119, "lon"=>-100.574379),
"/US/US-TX/jefferson-20613" => array("type"=>"city", "pop"=>1919, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>32.760873, "lon"=>-94.349139),
"/US/US-TX/jersey-village-20614" => array("type"=>"city", "pop"=>7279, "name_en"=>"Jersey Village", "lang"=>"en", "lat"=>29.888329, "lon"=>-95.568433),
"/US/US-TX/jewett-20615" => array("type"=>"city", "pop"=>951, "name_ru"=>"Джеветт", "name_en"=>"Jewett", "lang"=>"en", "lat"=>31.361577, "lon"=>-96.145817),
"/US/US-TX/joaquin-20616" => array("type"=>"city", "pop"=>946, "name_en"=>"Joaquin", "lang"=>"en", "lat"=>31.966782, "lon"=>-94.050547),
"/US/US-TX/johnson-city-20617" => array("type"=>"city", "pop"=>1570, "name_ru"=>"Джонсон Сити", "name_en"=>"Johnson City", "lang"=>"en", "lat"=>30.274335, "lon"=>-98.406492),
"/US/US-TX/jolly-20618" => array("type"=>"city", "pop"=>186, "name_en"=>"Jolly", "lang"=>"en", "lat"=>33.876813, "lon"=>-98.345925),
"/US/US-TX/jollyville-20619" => array("type"=>"cdp", "pop"=>15206, "name_en"=>"Jollyville", "lang"=>"en", "lat"=>30.449550, "lon"=>-97.750997),
"/US/US-TX/jones-creek-20620" => array("type"=>"village", "pop"=>2103, "name_en"=>"Jones Creek", "lang"=>"en", "lat"=>28.974629, "lon"=>-95.467789),
"/US/US-TX/jonestown-20621" => array("type"=>"city", "pop"=>2432, "name_ru"=>"Джоунстаун", "name_en"=>"Jonestown", "lang"=>"en", "lat"=>30.489160, "lon"=>-97.928080),
"/US/US-TX/josephine-20622" => array("type"=>"town", "pop"=>944, "name_en"=>"Josephine", "lang"=>"en", "lat"=>33.061595, "lon"=>-96.316347),
"/US/US-TX/joshua-20623" => array("type"=>"city", "pop"=>5812, "name_en"=>"Joshua", "lang"=>"en", "lat"=>32.462172, "lon"=>-97.380368),
"/US/US-TX/jourdanton-20624" => array("type"=>"city", "pop"=>4315, "name_en"=>"Jourdanton", "lang"=>"en", "lat"=>28.912741, "lon"=>-98.541356),
"/US/US-TX/junction-20625" => array("type"=>"city", "pop"=>2554, "name_ru"=>"Джанкшен", "name_en"=>"Junction", "lang"=>"en", "lat"=>30.489584, "lon"=>-99.770991),
"/US/US-TX/justin-20626" => array("type"=>"city", "pop"=>3473, "name_en"=>"Justin", "lang"=>"en", "lat"=>33.086614, "lon"=>-97.299856),
"/US/US-TX/karnes-city-20627" => array("type"=>"town", "pop"=>3338, "name_en"=>"Karnes City", "lang"=>"en", "lat"=>28.886957, "lon"=>-97.901055),
"/US/US-TX/katy-20628" => array("type"=>"city", "pop"=>13913, "name_en"=>"Katy", "lang"=>"en", "lat"=>29.798252, "lon"=>-95.823131),
"/US/US-TX/kaufman-20629" => array("type"=>"city", "pop"=>8794, "name_en"=>"Kaufman", "lang"=>"en", "lat"=>32.581817, "lon"=>-96.308231),
"/US/US-TX/keene-20630" => array("type"=>"city", "pop"=>6350, "name_en"=>"Keene", "lang"=>"en", "lat"=>32.393348, "lon"=>-97.330234),
"/US/US-TX/keller-20631" => array("type"=>"city", "pop"=>38951, "name_en"=>"Keller", "lang"=>"en", "lat"=>32.930150, "lon"=>-97.226081),
"/US/US-TX/kemah-20632" => array("type"=>"city", "pop"=>2498, "name_en"=>"Kemah", "lang"=>"en", "lat"=>29.528768, "lon"=>-95.020530),
"/US/US-TX/kemp-20633" => array("type"=>"town", "pop"=>1334, "name_en"=>"Kemp", "lang"=>"en", "lat"=>32.436648, "lon"=>-96.223677),
"/US/US-TX/kendleton-20634" => array("type"=>"city", "pop"=>525, "name_en"=>"Kendleton", "lang"=>"en", "lat"=>29.445897, "lon"=>-96.000134),
"/US/US-TX/kenedy-20635" => array("type"=>"city", "pop"=>3307, "name_en"=>"Kenedy", "lang"=>"en", "lat"=>28.815192, "lon"=>-97.850542),
"/US/US-TX/kenefick-20636" => array("type"=>"town", "pop"=>706, "name_en"=>"Kenefick", "lang"=>"en", "lat"=>30.108752, "lon"=>-94.854134),
"/US/US-TX/kennard-20637" => array("type"=>"city", "pop"=>312, "name_ru"=>"Кеннард", "name_en"=>"Kennard", "lang"=>"en", "lat"=>31.358236, "lon"=>-95.184891),
"/US/US-TX/kennedale-20638" => array("type"=>"city", "pop"=>7115, "name_en"=>"Kennedale", "lang"=>"en", "lat"=>32.647533, "lon"=>-97.219095),
"/US/US-TX/kerens-20639" => array("type"=>"city", "pop"=>1810, "name_en"=>"Kerens", "lang"=>"en", "lat"=>32.131038, "lon"=>-96.225224),
"/US/US-TX/kermit-20640" => array("type"=>"city", "pop"=>5231, "name_en"=>"Kermit", "lang"=>"en", "lat"=>31.853930, "lon"=>-103.091967),
"/US/US-TX/kerrville-20641" => array("type"=>"city", "pop"=>22821, "name_en"=>"Kerrville", "lang"=>"en", "lat"=>30.041863, "lon"=>-99.136119),
"/US/US-TX/kilgore-20642" => array("type"=>"city", "pop"=>12010, "name_en"=>"Kilgore", "lang"=>"en", "lat"=>32.387234, "lon"=>-94.868306),
"/US/US-TX/killeen-20643" => array("type"=>"city", "pop"=>116934, "name_en"=>"Killeen", "lang"=>"en", "lat"=>31.098300, "lon"=>-97.717219),
"/US/US-TX/kingsland-20644" => array("type"=>"cdp", "pop"=>2725, "name_ru"=>"Кингсленд", "name_en"=>"Kingsland", "lang"=>"en", "lat"=>30.664701, "lon"=>-98.453520),
"/US/US-TX/kingsville-20645" => array("type"=>"city", "pop"=>24702, "name_ru"=>"Кингсвилл", "name_en"=>"Kingsville", "lang"=>"en", "lat"=>27.508923, "lon"=>-97.860738),
"/US/US-TX/kingwood-20646" => array("type"=>"cdp", "pop"=>37397, "name_en"=>"Kingwood", "lang"=>"en", "lat"=>30.057856, "lon"=>-95.186241),
"/US/US-TX/kirby-20647" => array("type"=>"city", "pop"=>8606, "name_en"=>"Kirby", "lang"=>"en", "lat"=>29.460000, "lon"=>-98.385649),
"/US/US-TX/kirbyville-20648" => array("type"=>"city", "pop"=>1945, "name_en"=>"Kirbyville", "lang"=>"en", "lat"=>30.657900, "lon"=>-93.899906),
"/US/US-TX/kirvin-20649" => array("type"=>"town", "pop"=>128, "name_en"=>"Kirvin", "lang"=>"en", "lat"=>31.766831, "lon"=>-96.329820),
"/US/US-TX/knollwood-20650" => array("type"=>"village", "pop"=>404, "name_en"=>"Knollwood", "lang"=>"en", "lat"=>33.688977, "lon"=>-96.618530),
"/US/US-TX/knox-city-20651" => array("type"=>"city", "pop"=>977, "name_ru"=>"Нокс Сити", "name_en"=>"Knox City", "lang"=>"en", "lat"=>33.417458, "lon"=>-99.815804),
"/US/US-TX/kosse-20652" => array("type"=>"town", "pop"=>504, "name_en"=>"Kosse", "lang"=>"en", "lat"=>31.306905, "lon"=>-96.628731),
"/US/US-TX/kountze-20653" => array("type"=>"city", "pop"=>2171, "name_en"=>"Kountze", "lang"=>"en", "lat"=>30.372410, "lon"=>-94.315514),
"/US/US-TX/kress-20654" => array("type"=>"city", "pop"=>767, "name_en"=>"Kress", "lang"=>"en", "lat"=>34.366160, "lon"=>-101.747490),
"/US/US-TX/krugerville-20655" => array("type"=>"city", "pop"=>1767, "name_en"=>"Krugerville", "lang"=>"en", "lat"=>33.280731, "lon"=>-96.990555),
"/US/US-TX/krum-20656" => array("type"=>"city", "pop"=>4309, "name_en"=>"Krum", "lang"=>"en", "lat"=>33.265576, "lon"=>-97.229583),
"/US/US-TX/kyle-20657" => array("type"=>"town", "pop"=>26103, "name_en"=>"Kyle", "lang"=>"en", "lat"=>29.994031, "lon"=>-97.870639),
"/US/US-TX/la-casita-garciasville-20658" => array("type"=>"cdp", "pop"=>1186, "name_en"=>"La Casita-Garciasville", "lang"=>"en", "lat"=>26.332297, "lon"=>-98.717076),
"/US/US-TX/la-feria-20663" => array("type"=>"city", "pop"=>7023, "name_en"=>"La Feria", "lang"=>"en", "lat"=>26.157000, "lon"=>-97.822649),
"/US/US-TX/la-grange-20665" => array("type"=>"city", "pop"=>4709, "name_ru"=>"Ла Гранж", "name_en"=>"La Grange", "lang"=>"en", "lat"=>29.908700, "lon"=>-96.878116),
"/US/US-TX/la-grulla-20666" => array("type"=>"city", "pop"=>1838, "name_en"=>"La Grulla", "lang"=>"en", "lat"=>26.268930, "lon"=>-98.646766),
"/US/US-TX/la-homa-20669" => array("type"=>"cdp", "pop"=>1403, "name_en"=>"La Homa", "lang"=>"en", "lat"=>26.249681, "lon"=>-98.363200),
"/US/US-TX/la-joya-20670" => array("type"=>"city", "pop"=>4795, "name_en"=>"La Joya", "lang"=>"en", "lat"=>26.248320, "lon"=>-98.478165),
"/US/US-TX/la-marque-20686" => array("type"=>"city", "pop"=>14236, "name_en"=>"La Marque", "lang"=>"en", "lat"=>29.367050, "lon"=>-94.993305),
"/US/US-TX/la-porte-20690" => array("type"=>"city", "pop"=>34274, "name_ru"=>"Ла Порт", "name_en"=>"La Porte", "lang"=>"en", "lat"=>29.666553, "lon"=>-95.044664),
"/US/US-TX/la-pryor-20691" => array("type"=>"cdp", "pop"=>1343, "name_en"=>"La Pryor", "lang"=>"en", "lat"=>28.948640, "lon"=>-99.848787),
"/US/US-TX/la-vernia-20695" => array("type"=>"city", "pop"=>1218, "name_en"=>"La Vernia", "lang"=>"en", "lat"=>29.354051, "lon"=>-98.114338),
"/US/US-TX/la-villa-20696" => array("type"=>"city", "pop"=>1434, "name_en"=>"La Villa", "lang"=>"en", "lat"=>26.297279, "lon"=>-97.926660),
"/US/US-TX/la-ward-20698" => array("type"=>"city", "pop"=>193, "name_en"=>"La Ward", "lang"=>"en", "lat"=>28.847611, "lon"=>-96.465377),
"/US/US-TX/lackland-afb-20659" => array("type"=>"cdp", "pop"=>9352, "name_en"=>"Lackland AFB", "lang"=>"en", "lat"=>29.382779, "lon"=>-98.626545),
"/US/US-TX/lacoste-20660" => array("type"=>"city", "pop"=>1407, "name_en"=>"LaCoste", "lang"=>"en", "lat"=>29.309305, "lon"=>-98.810897),
"/US/US-TX/lacy-lakeview-20661" => array("type"=>"city", "pop"=>5883, "name_en"=>"Lacy-Lakeview", "lang"=>"en", "lat"=>31.619867, "lon"=>-97.107599),
"/US/US-TX/ladonia-20662" => array("type"=>"town", "pop"=>695, "name_ru"=>"Ладония", "name_en"=>"Ladonia", "lang"=>"en", "lat"=>33.427799, "lon"=>-95.945297),
"/US/US-TX/lago-vista-20664" => array("type"=>"city", "pop"=>6293, "name_en"=>"Lago Vista", "lang"=>"en", "lat"=>30.451888, "lon"=>-97.996723),
"/US/US-TX/laguna-heights-20667" => array("type"=>"cdp", "pop"=>1671, "name_en"=>"Laguna Heights", "lang"=>"en", "lat"=>26.080064, "lon"=>-97.256886),
"/US/US-TX/laguna-vista-20668" => array("type"=>"village", "pop"=>4017, "name_en"=>"Laguna Vista", "lang"=>"en", "lat"=>26.100487, "lon"=>-97.293272),
"/US/US-TX/lake-bridgeport-20671" => array("type"=>"city", "pop"=>455, "name_en"=>"Lake Bridgeport", "lang"=>"en", "lat"=>33.207101, "lon"=>-97.830570),
"/US/US-TX/lake-brownwood-20672" => array("type"=>"cdp", "pop"=>1221, "name_en"=>"Lake Brownwood", "lang"=>"en", "lat"=>31.816030, "lon"=>-99.102900),
"/US/US-TX/lake-city-20673" => array("type"=>"town", "pop"=>512, "name_ru"=>"Лэйк Сити", "name_en"=>"Lake City", "lang"=>"en", "lat"=>28.082983, "lon"=>-97.882447),
"/US/US-TX/lake-dallas-20674" => array("type"=>"city", "pop"=>7811, "name_en"=>"Lake Dallas", "lang"=>"en", "lat"=>33.126463, "lon"=>-97.024693),
"/US/US-TX/lake-jackson-20676" => array("type"=>"city", "pop"=>27417, "name_en"=>"Lake Jackson", "lang"=>"en", "lat"=>29.041858, "lon"=>-95.446326),
"/US/US-TX/lake-tanglewood-20681" => array("type"=>"village", "pop"=>886, "name_en"=>"Lake Tanglewood", "lang"=>"en", "lat"=>35.058966, "lon"=>-101.782862),
"/US/US-TX/lake-worth-20685" => array("type"=>"city", "pop"=>4810, "name_ru"=>"Лэйк Ворт", "name_en"=>"Lake Worth", "lang"=>"en", "lat"=>32.813743, "lon"=>-97.432900),
"/US/US-TX/lakehills-20675" => array("type"=>"cdp", "pop"=>2147, "name_en"=>"Lakehills", "lang"=>"en", "lat"=>29.604374, "lon"=>-98.942888),
"/US/US-TX/lakeport-20677" => array("type"=>"city", "pop"=>984, "name_ru"=>"Лейкпорт", "name_en"=>"Lakeport", "lang"=>"en", "lat"=>32.401956, "lon"=>-94.711492),
"/US/US-TX/lakeside-20678" => array("type"=>"town", "pop"=>322, "name_ru"=>"Лэйксайд", "name_en"=>"Lakeside", "lang"=>"en", "lat"=>28.104519, "lon"=>-97.861105),
"/US/US-TX/lakeside-20679" => array("type"=>"town", "pop"=>816, "name_ru"=>"Лэйксайд", "name_en"=>"Lakeside", "lang"=>"en", "lat"=>32.821550, "lon"=>-97.489255),
"/US/US-TX/lakeside-city-20680" => array("type"=>"town", "pop"=>1048, "name_en"=>"Lakeside City", "lang"=>"en", "lat"=>33.828227, "lon"=>-98.541178),
"/US/US-TX/lakeview-20682" => array("type"=>"town", "pop"=>140, "name_ru"=>"Лейквью", "name_en"=>"Lakeview", "lang"=>"en", "lat"=>34.672825, "lon"=>-100.697227),
"/US/US-TX/lakeway-20683" => array("type"=>"city", "pop"=>11193, "name_en"=>"Lakeway", "lang"=>"en", "lat"=>30.366587, "lon"=>-97.979403),
"/US/US-TX/lakewood-village-20684" => array("type"=>"city", "pop"=>393, "name_en"=>"Lakewood Village", "lang"=>"en", "lat"=>33.140384, "lon"=>-96.976460),
"/US/US-TX/lamesa-20687" => array("type"=>"city", "pop"=>8880, "name_en"=>"Lamesa", "lang"=>"en", "lat"=>32.732599, "lon"=>-101.955788),
"/US/US-TX/lampasas-20688" => array("type"=>"city", "pop"=>7975, "name_en"=>"Lampasas", "lang"=>"en", "lat"=>31.063397, "lon"=>-98.180937),
"/US/US-TX/lancaster-20689" => array("type"=>"city", "pop"=>35761, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>32.595619, "lon"=>-96.781045),
"/US/US-TX/laredo-20692" => array("type"=>"city", "pop"=>221659, "name_ru"=>"Ларедо", "name_en"=>"Laredo", "lang"=>"en", "lat"=>27.534300, "lon"=>-99.486599),
"/US/US-TX/latexo-20693" => array("type"=>"city", "pop"=>268, "name_en"=>"Latexo", "lang"=>"en", "lat"=>31.389106, "lon"=>-95.474473),
"/US/US-TX/laughlin-afb-20694" => array("type"=>"cdp", "pop"=>2556, "name_en"=>"Laughlin AFB", "lang"=>"en", "lat"=>29.355661, "lon"=>-100.783841),
"/US/US-TX/lavon-20697" => array("type"=>"town", "pop"=>423, "name_en"=>"Lavon", "lang"=>"en", "lat"=>33.026676, "lon"=>-96.438204),
"/US/US-TX/lawn-20699" => array("type"=>"town", "pop"=>353, "name_en"=>"Lawn", "lang"=>"en", "lat"=>32.136230, "lon"=>-99.749972),
"/US/US-TX/lawrence-20700" => array("type"=>"village", "pop"=>231, "name_ru"=>"Лоренс", "name_en"=>"Lawrence", "lang"=>"en", "lat"=>32.741177, "lon"=>-96.346892),
"/US/US-TX/league-city-20701" => array("type"=>"city", "pop"=>71222, "name_en"=>"League City", "lang"=>"en", "lat"=>29.485611, "lon"=>-95.107462),
"/US/US-TX/leakey-20702" => array("type"=>"city", "pop"=>359, "name_en"=>"Leakey", "lang"=>"en", "lat"=>29.724377, "lon"=>-99.762250),
"/US/US-TX/leander-20703" => array("type"=>"city", "pop"=>25424, "name_en"=>"Leander", "lang"=>"en", "lat"=>30.562367, "lon"=>-97.860520),
"/US/US-TX/leary-20704" => array("type"=>"city", "pop"=>537, "name_ru"=>"Лири", "name_en"=>"Leary", "lang"=>"en", "lat"=>33.470505, "lon"=>-94.212145),
"/US/US-TX/lefors-20705" => array("type"=>"town", "pop"=>563, "name_en"=>"Lefors", "lang"=>"en", "lat"=>35.439663, "lon"=>-100.803324),
"/US/US-TX/leon-valley-20708" => array("type"=>"city", "pop"=>10269, "name_en"=>"Leon Valley", "lang"=>"en", "lat"=>29.495156, "lon"=>-98.613420),
"/US/US-TX/leona-20706" => array("type"=>"town", "pop"=>201, "name_ru"=>"Леона", "name_en"=>"Leona", "lang"=>"en", "lat"=>31.154580, "lon"=>-95.975469),
"/US/US-TX/leonard-20707" => array("type"=>"city", "pop"=>2095, "name_ru"=>"Леонард", "name_en"=>"Leonard", "lang"=>"en", "lat"=>33.381052, "lon"=>-96.246451),
"/US/US-TX/leroy-20709" => array("type"=>"city", "pop"=>342, "name_en"=>"Leroy", "lang"=>"en", "lat"=>31.731632, "lon"=>-97.019964),
"/US/US-TX/levelland-20710" => array("type"=>"city", "pop"=>12450, "name_en"=>"Levelland", "lang"=>"en", "lat"=>33.579906, "lon"=>-102.363944),
"/US/US-TX/lewisville-20711" => array("type"=>"city", "pop"=>101624, "name_ru"=>"Льюисвилл", "name_en"=>"Lewisville", "lang"=>"en", "lat"=>33.041663, "lon"=>-96.982918),
"/US/US-TX/lexington-20712" => array("type"=>"town", "pop"=>1236, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>30.414692, "lon"=>-97.009382),
"/US/US-TX/liberty-20713" => array("type"=>"city", "pop"=>8347, "name_ru"=>"Либерти", "name_en"=>"Liberty", "lang"=>"en", "lat"=>30.046017, "lon"=>-94.797686),
"/US/US-TX/liberty-city-20714" => array("type"=>"cdp", "pop"=>1607, "name_en"=>"Liberty City", "lang"=>"en", "lat"=>32.450058, "lon"=>-94.944611),
"/US/US-TX/lincoln-park-20715" => array("type"=>"town", "pop"=>616, "name_ru"=>"Линкольн Парк", "name_en"=>"Lincoln Park", "lang"=>"en", "lat"=>33.222583, "lon"=>-96.972356),
"/US/US-TX/lindale-20716" => array("type"=>"town", "pop"=>4687, "name_ru"=>"Линдейл", "name_en"=>"Lindale", "lang"=>"en", "lat"=>32.505441, "lon"=>-95.405976),
"/US/US-TX/linden-20717" => array("type"=>"city", "pop"=>2129, "name_ru"=>"Линден", "name_en"=>"Linden", "lang"=>"en", "lat"=>33.011009, "lon"=>-94.362067),
"/US/US-TX/lindsay-20718" => array("type"=>"town", "pop"=>1007, "name_ru"=>"Линдсей", "name_en"=>"Lindsay", "lang"=>"en", "lat"=>33.636189, "lon"=>-97.219786),
"/US/US-TX/lipan-20719" => array("type"=>"city", "pop"=>507, "name_en"=>"Lipan", "lang"=>"en", "lat"=>32.518913, "lon"=>-98.046675),
"/US/US-TX/little-elm-20720" => array("type"=>"town", "pop"=>24947, "name_en"=>"Little Elm", "lang"=>"en", "lat"=>33.163836, "lon"=>-96.929999),
"/US/US-TX/little-river-academy-20722" => array("type"=>"city", "pop"=>1648, "name_en"=>"Little River-Academy", "lang"=>"en", "lat"=>30.991304, "lon"=>-97.351868),
"/US/US-TX/littlefield-20721" => array("type"=>"city", "pop"=>5929, "name_en"=>"Littlefield", "lang"=>"en", "lat"=>33.919919, "lon"=>-102.334593),
"/US/US-TX/live-oak-20723" => array("type"=>"city", "pop"=>13142, "name_ru"=>"Лайв Оук", "name_en"=>"Live Oak", "lang"=>"en", "lat"=>29.554150, "lon"=>-98.339831),
"/US/US-TX/liverpool-20724" => array("type"=>"village", "pop"=>422, "name_ru"=>"Ливерпул", "name_en"=>"Liverpool", "lang"=>"en", "lat"=>29.299034, "lon"=>-95.277258),
"/US/US-TX/livingston-20725" => array("type"=>"town", "pop"=>6203, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>30.709442, "lon"=>-94.936069),
"/US/US-TX/llano-20726" => array("type"=>"city", "pop"=>3237, "name_en"=>"Llano", "lang"=>"en", "lat"=>30.751341, "lon"=>-98.673896),
"/US/US-TX/lockhart-20727" => array("type"=>"city", "pop"=>13661, "name_ru"=>"Локхарт", "name_en"=>"Lockhart", "lang"=>"en", "lat"=>29.874201, "lon"=>-97.677601),
"/US/US-TX/lockney-20728" => array("type"=>"town", "pop"=>1668, "name_en"=>"Lockney", "lang"=>"en", "lat"=>34.122899, "lon"=>-101.441845),
"/US/US-TX/log-cabin-20729" => array("type"=>"city", "pop"=>792, "name_en"=>"Log Cabin", "lang"=>"en", "lat"=>32.223278, "lon"=>-96.022169),
"/US/US-TX/lometa-20730" => array("type"=>"city", "pop"=>897, "name_en"=>"Lometa", "lang"=>"en", "lat"=>31.216109, "lon"=>-98.392842),
"/US/US-TX/lone-oak-20731" => array("type"=>"town", "pop"=>564, "name_ru"=>"Лоун Оук", "name_en"=>"Lone Oak", "lang"=>"en", "lat"=>32.996281, "lon"=>-95.941542),
"/US/US-TX/lone-star-20732" => array("type"=>"city", "pop"=>1588, "name_ru"=>"Лоун Стар", "name_en"=>"Lone Star", "lang"=>"en", "lat"=>32.937399, "lon"=>-94.708507),
"/US/US-TX/longview-20733" => array("type"=>"city", "pop"=>77211, "name_ru"=>"Лонгвью", "name_en"=>"Longview", "lang"=>"en", "lat"=>32.517742, "lon"=>-94.762280),
"/US/US-TX/lopezville-20734" => array("type"=>"cdp", "pop"=>2827, "name_en"=>"Lopezville", "lang"=>"en", "lat"=>26.247536, "lon"=>-98.158878),
"/US/US-TX/loraine-20735" => array("type"=>"town", "pop"=>604, "name_ru"=>"Лорейн", "name_en"=>"Loraine", "lang"=>"en", "lat"=>32.408317, "lon"=>-100.712173),
"/US/US-TX/lorena-20736" => array("type"=>"town", "pop"=>1670, "name_en"=>"Lorena", "lang"=>"en", "lat"=>31.382312, "lon"=>-97.212717),
"/US/US-TX/lorenzo-20737" => array("type"=>"town", "pop"=>1185, "name_en"=>"Lorenzo", "lang"=>"en", "lat"=>33.670069, "lon"=>-101.534467),
"/US/US-TX/los-fresnos-20738" => array("type"=>"city", "pop"=>5538, "name_en"=>"Los Fresnos", "lang"=>"en", "lat"=>26.073885, "lon"=>-97.479899),
"/US/US-TX/los-ybanez-20740" => array("type"=>"city", "pop"=>29, "name_en"=>"Los Ybanez", "lang"=>"en", "lat"=>32.718900, "lon"=>-101.917592),
"/US/US-TX/lost-creek-20739" => array("type"=>"cdp", "pop"=>4095, "name_en"=>"Lost Creek", "lang"=>"en", "lat"=>30.295375, "lon"=>-97.844043),
"/US/US-TX/lott-20741" => array("type"=>"city", "pop"=>675, "name_en"=>"Lott", "lang"=>"en", "lat"=>31.206022, "lon"=>-97.033351),
"/US/US-TX/lovelady-20742" => array("type"=>"city", "pop"=>598, "name_en"=>"Lovelady", "lang"=>"en", "lat"=>31.125963, "lon"=>-95.445345),
"/US/US-TX/lowry-crossing-20743" => array("type"=>"city", "pop"=>2100, "name_en"=>"Lowry Crossing", "lang"=>"en", "lat"=>33.167923, "lon"=>-96.544536),
"/US/US-TX/lubbock-20744" => array("type"=>"city", "pop"=>220483, "name_en"=>"Lubbock", "lang"=>"en", "lat"=>33.575850, "lon"=>-101.875374),
"/US/US-TX/lucas-20745" => array("type"=>"town", "pop"=>4914, "name_ru"=>"Лукас", "name_en"=>"Lucas", "lang"=>"en", "lat"=>33.101408, "lon"=>-96.577719),
"/US/US-TX/lueders-20746" => array("type"=>"city", "pop"=>264, "name_en"=>"Lueders", "lang"=>"en", "lat"=>32.799463, "lon"=>-99.623259),
"/US/US-TX/luella-20747" => array("type"=>"town", "pop"=>559, "name_en"=>"Luella", "lang"=>"en", "lat"=>33.569263, "lon"=>-96.542509),
"/US/US-TX/lufkin-20748" => array("type"=>"city", "pop"=>34530, "name_en"=>"Lufkin", "lang"=>"en", "lat"=>31.333518, "lon"=>-94.726135),
"/US/US-TX/luling-20749" => array("type"=>"city", "pop"=>5458, "name_ru"=>"Лулинг", "name_en"=>"Luling", "lang"=>"en", "lat"=>29.682925, "lon"=>-97.648978),
"/US/US-TX/lumberton-20750" => array("type"=>"city", "pop"=>10354, "name_ru"=>"Ламбертон", "name_en"=>"Lumberton", "lang"=>"en", "lat"=>30.264530, "lon"=>-94.201791),
"/US/US-TX/lyford-20751" => array("type"=>"city", "pop"=>2545, "name_en"=>"Lyford", "lang"=>"en", "lat"=>26.410899, "lon"=>-97.789628),
"/US/US-TX/lytle-20752" => array("type"=>"city", "pop"=>2818, "name_en"=>"Lytle", "lang"=>"en", "lat"=>29.234113, "lon"=>-98.797070),
"/US/US-TX/mabank-20753" => array("type"=>"town", "pop"=>3090, "name_en"=>"Mabank", "lang"=>"en", "lat"=>32.365471, "lon"=>-96.103976),
"/US/US-TX/madisonville-20761" => array("type"=>"city", "pop"=>4372, "name_ru"=>"Мэдисонвилл", "name_en"=>"Madisonville", "lang"=>"en", "lat"=>30.951996, "lon"=>-95.908353),
"/US/US-TX/magnolia-20762" => array("type"=>"town", "pop"=>1249, "name_ru"=>"Магнолия", "name_en"=>"Magnolia", "lang"=>"en", "lat"=>30.210989, "lon"=>-95.748742),
"/US/US-TX/malakoff-20763" => array("type"=>"city", "pop"=>2326, "name_en"=>"Malakoff", "lang"=>"en", "lat"=>32.173050, "lon"=>-96.012360),
"/US/US-TX/malone-20764" => array("type"=>"town", "pop"=>302, "name_ru"=>"Малон", "name_en"=>"Malone", "lang"=>"en", "lat"=>31.916561, "lon"=>-96.894863),
"/US/US-TX/manor-20765" => array("type"=>"city", "pop"=>3471, "name_en"=>"Manor", "lang"=>"en", "lat"=>30.345968, "lon"=>-97.555713),
"/US/US-TX/mansfield-20766" => array("type"=>"city", "pop"=>46142, "name_ru"=>"Мансфилд", "name_en"=>"Mansfield", "lang"=>"en", "lat"=>32.567113, "lon"=>-97.117367),
"/US/US-TX/manvel-20767" => array("type"=>"city", "pop"=>6051, "name_en"=>"Manvel", "lang"=>"en", "lat"=>29.483773, "lon"=>-95.359608),
"/US/US-TX/marble-falls-20768" => array("type"=>"city", "pop"=>7451, "name_en"=>"Marble Falls", "lang"=>"en", "lat"=>30.579731, "lon"=>-98.272085),
"/US/US-TX/marfa-20769" => array("type"=>"city", "pop"=>1862, "name_en"=>"Marfa", "lang"=>"en", "lat"=>30.310389, "lon"=>-104.024906),
"/US/US-TX/marietta-20770" => array("type"=>"town", "pop"=>107, "name_ru"=>"Мариетта", "name_en"=>"Marietta", "lang"=>"en", "lat"=>33.173794, "lon"=>-94.541761),
"/US/US-TX/marion-20771" => array("type"=>"city", "pop"=>1152, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>29.573600, "lon"=>-98.143049),
"/US/US-TX/markham-20772" => array("type"=>"cdp", "pop"=>1206, "name_ru"=>"Маркхэм", "name_en"=>"Markham", "lang"=>"en", "lat"=>28.963808, "lon"=>-96.063445),
"/US/US-TX/marlin-20773" => array("type"=>"city", "pop"=>5817, "name_en"=>"Marlin", "lang"=>"en", "lat"=>31.308956, "lon"=>-96.893005),
"/US/US-TX/marquez-20774" => array("type"=>"city", "pop"=>238, "name_en"=>"Marquez", "lang"=>"en", "lat"=>31.236918, "lon"=>-96.256811),
"/US/US-TX/marshall-20775" => array("type"=>"city", "pop"=>23798, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>32.536002, "lon"=>-94.350756),
"/US/US-TX/marshall-creek-20776" => array("type"=>"town", "pop"=>508, "name_en"=>"Marshall Creek", "lang"=>"en", "lat"=>33.017604, "lon"=>-97.208333),
"/US/US-TX/mart-20777" => array("type"=>"city", "pop"=>2415, "name_en"=>"Mart", "lang"=>"en", "lat"=>31.541351, "lon"=>-96.830043),
"/US/US-TX/martindale-20778" => array("type"=>"town", "pop"=>1148, "name_en"=>"Martindale", "lang"=>"en", "lat"=>29.842309, "lon"=>-97.843302),
"/US/US-TX/mason-20779" => array("type"=>"city", "pop"=>2222, "name_ru"=>"Мейсон", "name_en"=>"Mason", "lang"=>"en", "lat"=>30.749254, "lon"=>-99.232094),
"/US/US-TX/matador-20780" => array("type"=>"town", "pop"=>652, "name_en"=>"Matador", "lang"=>"en", "lat"=>34.015072, "lon"=>-100.821243),
"/US/US-TX/mathis-20781" => array("type"=>"city", "pop"=>5295, "name_en"=>"Mathis", "lang"=>"en", "lat"=>28.093373, "lon"=>-97.824516),
"/US/US-TX/maud-20782" => array("type"=>"city", "pop"=>1022, "name_en"=>"Maud", "lang"=>"en", "lat"=>33.330379, "lon"=>-94.344719),
"/US/US-TX/mauriceville-20783" => array("type"=>"cdp", "pop"=>2046, "name_en"=>"Mauriceville", "lang"=>"en", "lat"=>30.218288, "lon"=>-93.872919),
"/US/US-TX/maypearl-20784" => array("type"=>"city", "pop"=>936, "name_en"=>"Maypearl", "lang"=>"en", "lat"=>32.315557, "lon"=>-97.006559),
"/US/US-TX/mcallen-20754" => array("type"=>"city", "pop"=>129776, "name_en"=>"McAllen", "lang"=>"en", "lat"=>26.215500, "lon"=>-98.239316),
"/US/US-TX/mccamey-20755" => array("type"=>"city", "pop"=>1643, "name_en"=>"McCamey", "lang"=>"en", "lat"=>31.132600, "lon"=>-102.219459),
"/US/US-TX/mcgregor-20756" => array("type"=>"city", "pop"=>4891, "name_ru"=>"Мак-Грегор", "name_en"=>"McGregor", "lang"=>"en", "lat"=>31.453998, "lon"=>-97.377330),
"/US/US-TX/mckinney-20757" => array("type"=>"city", "pop"=>121211, "name_en"=>"McKinney", "lang"=>"en", "lat"=>33.200942, "lon"=>-96.654871),
"/US/US-TX/mclean-20758" => array("type"=>"town", "pop"=>828, "name_ru"=>"Мак-Лин", "name_en"=>"McLean", "lang"=>"en", "lat"=>35.232871, "lon"=>-100.599958),
"/US/US-TX/mclendon-chisholm-20759" => array("type"=>"city", "pop"=>1136, "name_en"=>"McLendon-Chisholm", "lang"=>"en", "lat"=>32.845910, "lon"=>-96.389862),
"/US/US-TX/mcqueeney-20760" => array("type"=>"cdp", "pop"=>2063, "name_en"=>"McQueeney", "lang"=>"en", "lat"=>29.597875, "lon"=>-98.043184),
"/US/US-TX/meadow-20785" => array("type"=>"town", "pop"=>618, "name_en"=>"Meadow", "lang"=>"en", "lat"=>33.338235, "lon"=>-102.206199),
"/US/US-TX/meadowlakes-20786" => array("type"=>"city", "pop"=>1892, "name_en"=>"Meadowlakes", "lang"=>"en", "lat"=>30.563245, "lon"=>-98.295207),
"/US/US-TX/meadows-20787" => array("type"=>"city", "pop"=>4606, "name_en"=>"Meadows", "lang"=>"en", "lat"=>29.651460, "lon"=>-95.586290),
"/US/US-TX/megargel-20788" => array("type"=>"town", "pop"=>253, "name_en"=>"Megargel", "lang"=>"en", "lat"=>33.453255, "lon"=>-98.929186),
"/US/US-TX/melissa-20789" => array("type"=>"city", "pop"=>4276, "name_en"=>"Melissa", "lang"=>"en", "lat"=>33.282609, "lon"=>-96.577192),
"/US/US-TX/melvin-20790" => array("type"=>"town", "pop"=>149, "name_ru"=>"Мэлвин", "name_en"=>"Melvin", "lang"=>"en", "lat"=>31.198322, "lon"=>-99.580935),
"/US/US-TX/memphis-20791" => array("type"=>"city", "pop"=>2197, "name_ru"=>"Мемфис", "name_en"=>"Memphis", "lang"=>"en", "lat"=>34.726663, "lon"=>-100.541157),
"/US/US-TX/menard-20792" => array("type"=>"city", "pop"=>1495, "name_en"=>"Menard", "lang"=>"en", "lat"=>30.919346, "lon"=>-99.784099),
"/US/US-TX/mercedes-20793" => array("type"=>"city", "pop"=>15131, "name_en"=>"Mercedes", "lang"=>"en", "lat"=>26.147999, "lon"=>-97.920756),
"/US/US-TX/meridian-20794" => array("type"=>"city", "pop"=>1495, "name_ru"=>"Меридиан", "name_en"=>"Meridian", "lang"=>"en", "lat"=>31.923955, "lon"=>-97.647233),
"/US/US-TX/merkel-20795" => array("type"=>"town", "pop"=>2614, "name_en"=>"Merkel", "lang"=>"en", "lat"=>32.468335, "lon"=>-100.012831),
"/US/US-TX/mertens-20796" => array("type"=>"town", "pop"=>160, "name_en"=>"Mertens", "lang"=>"en", "lat"=>32.058949, "lon"=>-96.893351),
"/US/US-TX/mertzon-20797" => array("type"=>"city", "pop"=>806, "name_en"=>"Mertzon", "lang"=>"en", "lat"=>31.262318, "lon"=>-100.820358),
"/US/US-TX/mesquite-20798" => array("type"=>"city", "pop"=>132123, "name_en"=>"Mesquite", "lang"=>"en", "lat"=>32.769001, "lon"=>-96.599800),
"/US/US-TX/mexia-20799" => array("type"=>"city", "pop"=>6552, "name_en"=>"Mexia", "lang"=>"en", "lat"=>31.679723, "lon"=>-96.480733),
"/US/US-TX/miami-20800" => array("type"=>"city", "pop"=>552, "name_ru"=>"Майами", "name_en"=>"Miami", "lang"=>"en", "lat"=>35.691795, "lon"=>-100.640440),
"/US/US-TX/midland-20801" => array("type"=>"city", "pop"=>106561, "name_ru"=>"Мидленд", "name_en"=>"Midland", "lang"=>"en", "lat"=>32.026491, "lon"=>-102.100151),
"/US/US-TX/midlothian-20802" => array("type"=>"city", "pop"=>16478, "name_ru"=>"Мидлотиан", "name_en"=>"Midlothian", "lang"=>"en", "lat"=>32.493967, "lon"=>-96.997712),
"/US/US-TX/midway-20803" => array("type"=>"city", "pop"=>300, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>31.026499, "lon"=>-95.752755),
"/US/US-TX/mila-doce-20804" => array("type"=>"cdp", "pop"=>2089, "name_en"=>"Mila Doce", "lang"=>"en", "lat"=>26.225770, "lon"=>-97.958652),
"/US/US-TX/milano-20805" => array("type"=>"city", "pop"=>413, "name_en"=>"Milano", "lang"=>"en", "lat"=>30.706781, "lon"=>-96.859895),
"/US/US-TX/mildred-20806" => array("type"=>"town", "pop"=>447, "name_ru"=>"Милдред", "name_en"=>"Mildred", "lang"=>"en", "lat"=>32.036891, "lon"=>-96.344762),
"/US/US-TX/miles-20807" => array("type"=>"city", "pop"=>762, "name_ru"=>"Майлс", "name_en"=>"Miles", "lang"=>"en", "lat"=>31.600129, "lon"=>-100.183859),
"/US/US-TX/milford-20808" => array("type"=>"town", "pop"=>749, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>32.122229, "lon"=>-96.949401),
"/US/US-TX/miller's-cove-20809" => array("type"=>"town", "pop"=>129, "name_en"=>"Miller's Cove", "lang"=>"en", "lat"=>33.155739, "lon"=>-95.115576),
"/US/US-TX/millsap-20810" => array("type"=>"town", "pop"=>410, "name_en"=>"Millsap", "lang"=>"en", "lat"=>32.750016, "lon"=>-98.008611),
"/US/US-TX/mineola-20811" => array("type"=>"city", "pop"=>5179, "name_en"=>"Mineola", "lang"=>"en", "lat"=>32.665876, "lon"=>-95.488914),
"/US/US-TX/mineral-wells-20812" => array("type"=>"city", "pop"=>16880, "name_en"=>"Mineral Wells", "lang"=>"en", "lat"=>32.820327, "lon"=>-98.078314),
"/US/US-TX/mingus-20813" => array("type"=>"city", "pop"=>256, "name_en"=>"Mingus", "lang"=>"en", "lat"=>32.536829, "lon"=>-98.423928),
"/US/US-TX/mission-20814" => array("type"=>"city", "pop"=>67119, "name_ru"=>"Мишен", "name_en"=>"Mission", "lang"=>"en", "lat"=>26.211450, "lon"=>-98.316887),
"/US/US-TX/mission-bend-20815" => array("type"=>"cdp", "pop"=>24945, "name_en"=>"Mission Bend", "lang"=>"en", "lat"=>29.693706, "lon"=>-95.664620),
"/US/US-TX/missouri-city-20816" => array("type"=>"city", "pop"=>74723, "name_ru"=>"Миссури Сити", "name_en"=>"Missouri City", "lang"=>"en", "lat"=>29.572302, "lon"=>-95.536728),
"/US/US-TX/mobeetie-20817" => array("type"=>"city", "pop"=>100, "name_en"=>"Mobeetie", "lang"=>"en", "lat"=>35.530471, "lon"=>-100.442142),
"/US/US-TX/monahans-20818" => array("type"=>"city", "pop"=>6516, "name_en"=>"Monahans", "lang"=>"en", "lat"=>31.639894, "lon"=>-103.061328),
"/US/US-TX/mont-belvieu-20819" => array("type"=>"town", "pop"=>2694, "name_en"=>"Mont Belvieu", "lang"=>"en", "lat"=>29.851566, "lon"=>-94.881140),
"/US/US-TX/montgomery-20820" => array("type"=>"city", "pop"=>596, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>30.391193, "lon"=>-95.695772),
"/US/US-TX/moody-20821" => array("type"=>"town", "pop"=>1421, "name_ru"=>"Муди", "name_en"=>"Moody", "lang"=>"en", "lat"=>31.308439, "lon"=>-97.360245),
"/US/US-TX/moore-station-20822" => array("type"=>"city", "pop"=>195, "name_en"=>"Moore Station", "lang"=>"en", "lat"=>32.190825, "lon"=>-95.572385),
"/US/US-TX/moran-20823" => array("type"=>"city", "pop"=>218, "name_ru"=>"Моран", "name_en"=>"Moran", "lang"=>"en", "lat"=>32.546784, "lon"=>-99.165596),
"/US/US-TX/morgan's-point-20825" => array("type"=>"city", "pop"=>347, "name_en"=>"Morgan's Point", "lang"=>"en", "lat"=>29.676071, "lon"=>-95.005238),
"/US/US-TX/morgan's-point-resort-20826" => array("type"=>"city", "pop"=>4414, "name_en"=>"Morgan's Point Resort", "lang"=>"en", "lat"=>31.155566, "lon"=>-97.454165),
"/US/US-TX/morgan-20824" => array("type"=>"city", "pop"=>510, "name_ru"=>"Морган", "name_en"=>"Morgan", "lang"=>"en", "lat"=>32.015724, "lon"=>-97.605734),
"/US/US-TX/morton-20827" => array("type"=>"city", "pop"=>1772, "name_ru"=>"Мортон", "name_en"=>"Morton", "lang"=>"en", "lat"=>33.724608, "lon"=>-102.758214),
"/US/US-TX/moulton-20828" => array("type"=>"town", "pop"=>959, "name_ru"=>"Молтон", "name_en"=>"Moulton", "lang"=>"en", "lat"=>29.571609, "lon"=>-97.146014),
"/US/US-TX/mount-calm-20830" => array("type"=>"city", "pop"=>343, "name_en"=>"Mount Calm", "lang"=>"en", "lat"=>31.756866, "lon"=>-96.880788),
"/US/US-TX/mount-enterprise-20831" => array("type"=>"city", "pop"=>543, "name_en"=>"Mount Enterprise", "lang"=>"en", "lat"=>31.910065, "lon"=>-94.681875),
"/US/US-TX/mount-pleasant-20832" => array("type"=>"city", "pop"=>15011, "name_ru"=>"Маунт Плезант", "name_en"=>"Mount Pleasant", "lang"=>"en", "lat"=>33.165545, "lon"=>-94.975836),
"/US/US-TX/mount-vernon-20833" => array("type"=>"town", "pop"=>2781, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>33.172717, "lon"=>-95.221809),
"/US/US-TX/mountain-city-20829" => array("type"=>"city", "pop"=>737, "name_ru"=>"Маунтин Сити", "name_en"=>"Mountain City", "lang"=>"en", "lat"=>30.038462, "lon"=>-97.893041),
"/US/US-TX/muenster-20834" => array("type"=>"city", "pop"=>1647, "name_en"=>"Muenster", "lang"=>"en", "lat"=>33.651884, "lon"=>-97.376684),
"/US/US-TX/muleshoe-20835" => array("type"=>"city", "pop"=>4262, "name_en"=>"Muleshoe", "lang"=>"en", "lat"=>34.228562, "lon"=>-102.728128),
"/US/US-TX/mullin-20836" => array("type"=>"town", "pop"=>169, "name_en"=>"Mullin", "lang"=>"en", "lat"=>31.554698, "lon"=>-98.664981),
"/US/US-TX/munday-20837" => array("type"=>"city", "pop"=>1209, "name_en"=>"Munday", "lang"=>"en", "lat"=>33.446861, "lon"=>-99.623252),
"/US/US-TX/murchison-20838" => array("type"=>"city", "pop"=>622, "name_en"=>"Murchison", "lang"=>"en", "lat"=>32.274885, "lon"=>-95.758099),
"/US/US-TX/murphy-20839" => array("type"=>"city", "pop"=>14779, "name_ru"=>"Мерфи", "name_en"=>"Murphy", "lang"=>"en", "lat"=>33.012804, "lon"=>-96.607415),
"/US/US-TX/mustang-20840" => array("type"=>"town", "pop"=>54, "name_en"=>"Mustang", "lang"=>"en", "lat"=>32.012474, "lon"=>-96.430919),
"/US/US-TX/mustang-ridge-20841" => array("type"=>"city", "pop"=>933, "name_en"=>"Mustang Ridge", "lang"=>"en", "lat"=>30.056555, "lon"=>-97.690857),
"/US/US-TX/nacogdoches-20842" => array("type"=>"city", "pop"=>32205, "name_en"=>"Nacogdoches", "lang"=>"en", "lat"=>31.614977, "lon"=>-94.651909),
"/US/US-TX/naples-20843" => array("type"=>"town", "pop"=>1402, "name_ru"=>"Нэплс", "name_en"=>"Naples", "lang"=>"en", "lat"=>33.203611, "lon"=>-94.677696),
"/US/US-TX/nash-20844" => array("type"=>"city", "pop"=>2402, "name_en"=>"Nash", "lang"=>"en", "lat"=>33.443467, "lon"=>-94.129373),
"/US/US-TX/nassau-bay-20845" => array("type"=>"city", "pop"=>4040, "name_en"=>"Nassau Bay", "lang"=>"en", "lat"=>29.545277, "lon"=>-95.087660),
"/US/US-TX/natalia-20846" => array("type"=>"city", "pop"=>1870, "name_en"=>"Natalia", "lang"=>"en", "lat"=>29.190523, "lon"=>-98.855664),
"/US/US-TX/navarro-20847" => array("type"=>"town", "pop"=>210, "name_en"=>"Navarro", "lang"=>"en", "lat"=>32.000313, "lon"=>-96.381875),
"/US/US-TX/navasota-20848" => array("type"=>"city", "pop"=>7558, "name_en"=>"Navasota", "lang"=>"en", "lat"=>30.389149, "lon"=>-96.087980),
"/US/US-TX/nazareth-20849" => array("type"=>"city", "pop"=>316, "name_en"=>"Nazareth", "lang"=>"en", "lat"=>34.541432, "lon"=>-102.103657),
"/US/US-TX/nederland-20850" => array("type"=>"city", "pop"=>16096, "name_ru"=>"Недерленд", "name_en"=>"Nederland", "lang"=>"en", "lat"=>29.970700, "lon"=>-93.995939),
"/US/US-TX/needville-20851" => array("type"=>"city", "pop"=>3457, "name_en"=>"Needville", "lang"=>"en", "lat"=>29.395721, "lon"=>-95.839229),
"/US/US-TX/nesbitt-20852" => array("type"=>"town", "pop"=>327, "name_en"=>"Nesbitt", "lang"=>"en", "lat"=>32.589280, "lon"=>-94.445290),
"/US/US-TX/nevada-20853" => array("type"=>"city", "pop"=>748, "name_ru"=>"Невада", "name_en"=>"Nevada", "lang"=>"en", "lat"=>33.039071, "lon"=>-96.371701),
"/US/US-TX/new-berlin-20855" => array("type"=>"city", "pop"=>563, "name_ru"=>"Нью Берлин", "name_en"=>"New Berlin", "lang"=>"en", "lat"=>29.463700, "lon"=>-98.101016),
"/US/US-TX/new-boston-20856" => array("type"=>"town", "pop"=>4650, "name_ru"=>"Нью Бостон", "name_en"=>"New Boston", "lang"=>"en", "lat"=>33.460284, "lon"=>-94.418492),
"/US/US-TX/new-braunfels-20857" => array("type"=>"city", "pop"=>53547, "name_en"=>"New Braunfels", "lang"=>"en", "lat"=>29.704347, "lon"=>-98.117429),
"/US/US-TX/new-chapel-hill-20859" => array("type"=>"city", "pop"=>612, "name_en"=>"New Chapel Hill", "lang"=>"en", "lat"=>32.303405, "lon"=>-95.169005),
"/US/US-TX/new-deal-20860" => array("type"=>"town", "pop"=>749, "name_en"=>"New Deal", "lang"=>"en", "lat"=>33.731336, "lon"=>-101.838147),
"/US/US-TX/new-home-20861" => array("type"=>"city", "pop"=>300, "name_en"=>"New Home", "lang"=>"en", "lat"=>33.327312, "lon"=>-101.910758),
"/US/US-TX/new-hope-20862" => array("type"=>"town", "pop"=>717, "name_ru"=>"Нью Хоуп", "name_en"=>"New Hope", "lang"=>"en", "lat"=>33.212440, "lon"=>-96.564649),
"/US/US-TX/new-london-20863" => array("type"=>"city", "pop"=>1028, "name_ru"=>"Нью Лондон", "name_en"=>"New London", "lang"=>"en", "lat"=>32.269212, "lon"=>-94.931308),
"/US/US-TX/new-summerfield-20864" => array("type"=>"city", "pop"=>1022, "name_en"=>"New Summerfield", "lang"=>"en", "lat"=>31.985167, "lon"=>-95.104775),
"/US/US-TX/new-waverly-20866" => array("type"=>"town", "pop"=>911, "name_en"=>"New Waverly", "lang"=>"en", "lat"=>30.536509, "lon"=>-95.481193),
"/US/US-TX/newark-20854" => array("type"=>"city", "pop"=>1218, "name_ru"=>"Ньюарк", "name_en"=>"Newark", "lang"=>"en", "lat"=>32.994894, "lon"=>-97.470790),
"/US/US-TX/newcastle-20858" => array("type"=>"city", "pop"=>553, "name_en"=>"Newcastle", "lang"=>"en", "lat"=>33.193978, "lon"=>-98.743998),
"/US/US-TX/newton-20865" => array("type"=>"city", "pop"=>2263, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>30.850503, "lon"=>-93.752774),
"/US/US-TX/neylandville-20867" => array("type"=>"town", "pop"=>59, "name_en"=>"Neylandville", "lang"=>"en", "lat"=>33.199632, "lon"=>-96.004023),
"/US/US-TX/niederwald-20868" => array("type"=>"town", "pop"=>501, "name_en"=>"Niederwald", "lang"=>"en", "lat"=>30.013283, "lon"=>-97.736851),
"/US/US-TX/nixon-20869" => array("type"=>"city", "pop"=>2188, "name_en"=>"Nixon", "lang"=>"en", "lat"=>29.269260, "lon"=>-97.761863),
"/US/US-TX/nocona-20870" => array("type"=>"city", "pop"=>3236, "name_en"=>"Nocona", "lang"=>"en", "lat"=>33.783522, "lon"=>-97.729983),
"/US/US-TX/nolanville-20871" => array("type"=>"city", "pop"=>2914, "name_en"=>"Nolanville", "lang"=>"en", "lat"=>31.078201, "lon"=>-97.610088),
"/US/US-TX/nome-20872" => array("type"=>"city", "pop"=>481, "name_ru"=>"Ном", "name_en"=>"Nome", "lang"=>"en", "lat"=>30.034428, "lon"=>-94.418119),
"/US/US-TX/noonday-20873" => array("type"=>"city", "pop"=>559, "name_en"=>"Noonday", "lang"=>"en", "lat"=>32.243855, "lon"=>-95.397642),
"/US/US-TX/nordheim-20874" => array("type"=>"city", "pop"=>318, "name_en"=>"Nordheim", "lang"=>"en", "lat"=>28.923340, "lon"=>-97.613641),
"/US/US-TX/normangee-20875" => array("type"=>"town", "pop"=>785, "name_en"=>"Normangee", "lang"=>"en", "lat"=>31.030026, "lon"=>-96.115521),
"/US/US-TX/north-cleveland-20876" => array("type"=>"city", "pop"=>273, "name_en"=>"North Cleveland", "lang"=>"en", "lat"=>30.362664, "lon"=>-95.101600),
"/US/US-TX/north-richland-hills-20879" => array("type"=>"city", "pop"=>65078, "name_en"=>"North Richland Hills", "lang"=>"en", "lat"=>32.860144, "lon"=>-97.219073),
"/US/US-TX/north-san-pedro-20880" => array("type"=>"cdp", "pop"=>953, "name_en"=>"North San Pedro", "lang"=>"en", "lat"=>27.802368, "lon"=>-97.682697),
"/US/US-TX/northcrest-20877" => array("type"=>"town", "pop"=>1725, "name_en"=>"Northcrest", "lang"=>"en", "lat"=>31.639419, "lon"=>-97.097394),
"/US/US-TX/northlake-20878" => array("type"=>"town", "pop"=>2036, "name_ru"=>"Нортлейк", "name_en"=>"Northlake", "lang"=>"en", "lat"=>33.073993, "lon"=>-97.257302),
"/US/US-TX/novice-20881" => array("type"=>"city", "pop"=>134, "name_en"=>"Novice", "lang"=>"en", "lat"=>31.987463, "lon"=>-99.624941),
"/US/US-TX/o'brien-20892" => array("type"=>"city", "pop"=>115, "name_en"=>"O'Brien", "lang"=>"en", "lat"=>33.380617, "lon"=>-99.843791),
"/US/US-TX/o'donnell-20895" => array("type"=>"city", "pop"=>911, "name_en"=>"O'Donnell", "lang"=>"en", "lat"=>32.965191, "lon"=>-101.830111),
"/US/US-TX/oak-grove-20882" => array("type"=>"town", "pop"=>1008, "name_ru"=>"Оук Гров", "name_en"=>"Oak Grove", "lang"=>"en", "lat"=>32.528802, "lon"=>-96.315615),
"/US/US-TX/oak-leaf-20884" => array("type"=>"town", "pop"=>1495, "name_en"=>"Oak Leaf", "lang"=>"en", "lat"=>32.520886, "lon"=>-96.849871),
"/US/US-TX/oak-point-20885" => array("type"=>"town", "pop"=>2963, "name_en"=>"Oak Point", "lang"=>"en", "lat"=>33.182234, "lon"=>-96.994907),
"/US/US-TX/oak-ridge-20886" => array("type"=>"town", "pop"=>248, "name_ru"=>"Оук Ридж", "name_en"=>"Oak Ridge", "lang"=>"en", "lat"=>33.647905, "lon"=>-97.038521),
"/US/US-TX/oak-ridge-20887" => array("type"=>"town", "pop"=>268, "name_ru"=>"Оук Ридж", "name_en"=>"Oak Ridge", "lang"=>"en", "lat"=>32.657576, "lon"=>-96.263995),
"/US/US-TX/oak-ridge-north-20888" => array("type"=>"town", "pop"=>3410, "name_en"=>"Oak Ridge North", "lang"=>"en", "lat"=>30.159009, "lon"=>-95.444971),
"/US/US-TX/oak-trail-shores-20889" => array("type"=>"cdp", "pop"=>1750, "name_en"=>"Oak Trail Shores", "lang"=>"en", "lat"=>32.488606, "lon"=>-97.833889),
"/US/US-TX/oak-valley-20890" => array("type"=>"town", "pop"=>459, "name_en"=>"Oak Valley", "lang"=>"en", "lat"=>32.030353, "lon"=>-96.511005),
"/US/US-TX/oakhurst-20883" => array("type"=>"city", "pop"=>219, "name_ru"=>"Оукхерст", "name_en"=>"Oakhurst", "lang"=>"en", "lat"=>30.744709, "lon"=>-95.309479),
"/US/US-TX/oakwood-20891" => array("type"=>"town", "pop"=>512, "name_ru"=>"Оуквуд", "name_en"=>"Oakwood", "lang"=>"en", "lat"=>31.584353, "lon"=>-95.848925),
"/US/US-TX/odem-20893" => array("type"=>"city", "pop"=>2525, "name_en"=>"Odem", "lang"=>"en", "lat"=>27.945496, "lon"=>-97.586800),
"/US/US-TX/odessa-20894" => array("type"=>"city", "pop"=>98801, "name_ru"=>"Одесса", "name_en"=>"Odessa", "lang"=>"en", "lat"=>31.874433, "lon"=>-102.348341),
"/US/US-TX/oglesby-20896" => array("type"=>"city", "pop"=>415, "name_ru"=>"Оглсби", "name_en"=>"Oglesby", "lang"=>"en", "lat"=>31.418920, "lon"=>-97.509852),
"/US/US-TX/old-river-winfree-20897" => array("type"=>"town", "pop"=>1650, "name_en"=>"Old River-Winfree", "lang"=>"en", "lat"=>29.878258, "lon"=>-94.824275),
"/US/US-TX/olmos-park-20898" => array("type"=>"city", "pop"=>2312, "name_en"=>"Olmos Park", "lang"=>"en", "lat"=>29.474500, "lon"=>-98.485886),
"/US/US-TX/olney-20899" => array("type"=>"city", "pop"=>3236, "name_ru"=>"Олни", "name_en"=>"Olney", "lang"=>"en", "lat"=>33.364398, "lon"=>-98.758094),
"/US/US-TX/olton-20900" => array("type"=>"city", "pop"=>2151, "name_en"=>"Olton", "lang"=>"en", "lat"=>34.179516, "lon"=>-102.136532),
"/US/US-TX/omaha-20901" => array("type"=>"city", "pop"=>988, "name_ru"=>"Омаха", "name_en"=>"Omaha", "lang"=>"en", "lat"=>33.179637, "lon"=>-94.742151),
"/US/US-TX/onalaska-20902" => array("type"=>"city", "pop"=>1435, "name_en"=>"Onalaska", "lang"=>"en", "lat"=>30.808842, "lon"=>-95.105853),
"/US/US-TX/onion-creek-20903" => array("type"=>"cdp", "pop"=>1544, "name_en"=>"Onion Creek", "lang"=>"en", "lat"=>30.136335, "lon"=>-97.783920),
"/US/US-TX/opdyke-west-20904" => array("type"=>"town", "pop"=>193, "name_en"=>"Opdyke West", "lang"=>"en", "lat"=>33.591154, "lon"=>-102.299477),
"/US/US-TX/orange-20905" => array("type"=>"city", "pop"=>19494, "name_ru"=>"Оранж", "name_en"=>"Orange", "lang"=>"en", "lat"=>30.103923, "lon"=>-93.758177),
"/US/US-TX/orange-grove-20906" => array("type"=>"city", "pop"=>1404, "name_ru"=>"Оранж Гров", "name_en"=>"Orange Grove", "lang"=>"en", "lat"=>27.955833, "lon"=>-97.938855),
"/US/US-TX/orchard-20907" => array("type"=>"town", "pop"=>477, "name_ru"=>"Орчард", "name_en"=>"Orchard", "lang"=>"en", "lat"=>29.603041, "lon"=>-95.968741),
"/US/US-TX/ore-city-20908" => array("type"=>"city", "pop"=>1168, "name_en"=>"Ore City", "lang"=>"en", "lat"=>32.801463, "lon"=>-94.716888),
"/US/US-TX/overton-20909" => array("type"=>"city", "pop"=>2377, "name_en"=>"Overton", "lang"=>"en", "lat"=>32.275564, "lon"=>-94.972286),
"/US/US-TX/ovilla-20910" => array("type"=>"city", "pop"=>3961, "name_en"=>"Ovilla", "lang"=>"en", "lat"=>32.539673, "lon"=>-96.892673),
"/US/US-TX/oyster-creek-20911" => array("type"=>"village", "pop"=>1238, "name_en"=>"Oyster Creek", "lang"=>"en", "lat"=>28.997449, "lon"=>-95.330526),
"/US/US-TX/ozona-20912" => array("type"=>"cdp", "pop"=>3181, "name_en"=>"Ozona", "lang"=>"en", "lat"=>30.706204, "lon"=>-101.206308),
"/US/US-TX/paducah-20913" => array("type"=>"town", "pop"=>1257, "name_ru"=>"Падака", "name_en"=>"Paducah", "lang"=>"en", "lat"=>34.014829, "lon"=>-100.303395),
"/US/US-TX/paint-rock-20914" => array("type"=>"town", "pop"=>280, "name_ru"=>"Пейнт Рок", "name_en"=>"Paint Rock", "lang"=>"en", "lat"=>31.509511, "lon"=>-99.925320),
"/US/US-TX/palacios-20915" => array("type"=>"town", "pop"=>5069, "name_en"=>"Palacios", "lang"=>"en", "lat"=>28.713625, "lon"=>-96.235816),
"/US/US-TX/palestine-20916" => array("type"=>"city", "pop"=>18390, "name_ru"=>"Палестин", "name_en"=>"Palestine", "lang"=>"en", "lat"=>31.756549, "lon"=>-95.646495),
"/US/US-TX/palm-valley-20919" => array("type"=>"town", "pop"=>1262, "name_ru"=>"Палм Вэлли", "name_en"=>"Palm Valley", "lang"=>"en", "lat"=>26.203233, "lon"=>-97.753742),
"/US/US-TX/palmer-20917" => array("type"=>"town", "pop"=>2258, "name_ru"=>"Палмер", "name_en"=>"Palmer", "lang"=>"en", "lat"=>32.427293, "lon"=>-96.672421),
"/US/US-TX/palmhurst-20918" => array("type"=>"city", "pop"=>4988, "name_en"=>"Palmhurst", "lang"=>"en", "lat"=>26.253800, "lon"=>-98.300872),
"/US/US-TX/palmview-20920" => array("type"=>"city", "pop"=>5502, "name_en"=>"Palmview", "lang"=>"en", "lat"=>26.229856, "lon"=>-98.372511),
"/US/US-TX/pampa-20921" => array("type"=>"city", "pop"=>17382, "name_en"=>"Pampa", "lang"=>"en", "lat"=>35.546968, "lon"=>-100.964347),
"/US/US-TX/panhandle-20922" => array("type"=>"town", "pop"=>2484, "name_en"=>"Panhandle", "lang"=>"en", "lat"=>35.346245, "lon"=>-101.380067),
"/US/US-TX/panorama-village-20923" => array("type"=>"city", "pop"=>2376, "name_en"=>"Panorama Village", "lang"=>"en", "lat"=>30.382124, "lon"=>-95.492677),
"/US/US-TX/pantego-20924" => array("type"=>"town", "pop"=>2381, "name_en"=>"Pantego", "lang"=>"en", "lat"=>32.715051, "lon"=>-97.154556),
"/US/US-TX/paris-20925" => array("type"=>"city", "pop"=>26050, "name_ru"=>"Парис", "name_en"=>"Paris", "lang"=>"en", "lat"=>33.666636, "lon"=>-95.547623),
"/US/US-TX/parker-20926" => array("type"=>"city", "pop"=>2911, "name_ru"=>"Паркер", "name_en"=>"Parker", "lang"=>"en", "lat"=>33.054018, "lon"=>-96.636408),
"/US/US-TX/pasadena-20927" => array("type"=>"city", "pop"=>146439, "name_ru"=>"Пасадена", "name_en"=>"Pasadena", "lang"=>"en", "lat"=>29.660600, "lon"=>-95.152021),
"/US/US-TX/pattison-20928" => array("type"=>"city", "pop"=>481, "name_en"=>"Pattison", "lang"=>"en", "lat"=>29.813984, "lon"=>-95.977477),
"/US/US-TX/patton-village-20929" => array("type"=>"city", "pop"=>1483, "name_en"=>"Patton Village", "lang"=>"en", "lat"=>30.194749, "lon"=>-95.166903),
"/US/US-TX/payne-springs-20930" => array("type"=>"town", "pop"=>722, "name_en"=>"Payne Springs", "lang"=>"en", "lat"=>32.280878, "lon"=>-96.089495),
"/US/US-TX/pearland-20931" => array("type"=>"city", "pop"=>82903, "name_en"=>"Pearland", "lang"=>"en", "lat"=>29.555595, "lon"=>-95.275493),
"/US/US-TX/pearsall-20932" => array("type"=>"city", "pop"=>7663, "name_en"=>"Pearsall", "lang"=>"en", "lat"=>28.887748, "lon"=>-99.092107),
"/US/US-TX/pecan-acres-20933" => array("type"=>"cdp", "pop"=>1587, "name_en"=>"Pecan Acres", "lang"=>"en", "lat"=>32.970017, "lon"=>-97.474435),
"/US/US-TX/pecan-gap-20934" => array("type"=>"city", "pop"=>219, "name_en"=>"Pecan Gap", "lang"=>"en", "lat"=>33.436642, "lon"=>-95.851552),
"/US/US-TX/pecan-grove-20935" => array("type"=>"cdp", "pop"=>9502, "name_en"=>"Pecan Grove", "lang"=>"en", "lat"=>29.625826, "lon"=>-95.731365),
"/US/US-TX/pecan-hill-20936" => array("type"=>"city", "pop"=>663, "name_en"=>"Pecan Hill", "lang"=>"en", "lat"=>32.496860, "lon"=>-96.775080),
"/US/US-TX/pecos-20937" => array("type"=>"city", "pop"=>7804, "name_en"=>"Pecos", "lang"=>"en", "lat"=>31.402486, "lon"=>-103.505575),
"/US/US-TX/pelican-bay-20938" => array("type"=>"town", "pop"=>1615, "name_en"=>"Pelican Bay", "lang"=>"en", "lat"=>32.923334, "lon"=>-97.519218),
"/US/US-TX/penelope-20939" => array("type"=>"town", "pop"=>232, "name_en"=>"Penelope", "lang"=>"en", "lat"=>31.858389, "lon"=>-96.929083),
"/US/US-TX/penitas-20940" => array("type"=>"cdp", "pop"=>1181, "name_en"=>"Penitas", "lang"=>"en", "lat"=>26.233381, "lon"=>-98.448493),
"/US/US-TX/pernitas-point-20941" => array("type"=>"village", "pop"=>253, "name_en"=>"Pernitas Point", "lang"=>"en", "lat"=>28.058292, "lon"=>-97.906102),
"/US/US-TX/perryton-20942" => array("type"=>"city", "pop"=>8304, "name_en"=>"Perryton", "lang"=>"en", "lat"=>36.391732, "lon"=>-100.802970),
"/US/US-TX/petersburg-20943" => array("type"=>"city", "pop"=>1221, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>33.870888, "lon"=>-101.596753),
"/US/US-TX/petrolia-20944" => array("type"=>"city", "pop"=>776, "name_en"=>"Petrolia", "lang"=>"en", "lat"=>34.013012, "lon"=>-98.231333),
"/US/US-TX/petronila-20945" => array("type"=>"city", "pop"=>79, "name_en"=>"Petronila", "lang"=>"en", "lat"=>27.669409, "lon"=>-97.633614),
"/US/US-TX/pflugerville-20946" => array("type"=>"city", "pop"=>39653, "name_en"=>"Pflugerville", "lang"=>"en", "lat"=>30.440656, "lon"=>-97.620202),
"/US/US-TX/pharr-20947" => array("type"=>"city", "pop"=>65258, "name_en"=>"Pharr", "lang"=>"en", "lat"=>26.181950, "lon"=>-98.188480),
"/US/US-TX/pilot-point-20948" => array("type"=>"city", "pop"=>4409, "name_ru"=>"Пайлот Пойнт", "name_en"=>"Pilot Point", "lang"=>"en", "lat"=>33.398483, "lon"=>-96.955107),
"/US/US-TX/pine-forest-20949" => array("type"=>"city", "pop"=>613, "name_en"=>"Pine Forest", "lang"=>"en", "lat"=>30.179864, "lon"=>-94.029774),
"/US/US-TX/pine-island-20952" => array("type"=>"town", "pop"=>833, "name_ru"=>"Пайн Айленд", "name_en"=>"Pine Island", "lang"=>"en", "lat"=>30.053371, "lon"=>-96.028634),
"/US/US-TX/pinehurst-20950" => array("type"=>"cdp", "pop"=>2163, "name_ru"=>"Пайнхерст", "name_en"=>"Pinehurst", "lang"=>"en", "lat"=>30.188854, "lon"=>-95.705123),
"/US/US-TX/pinehurst-20951" => array("type"=>"city", "pop"=>2682, "name_ru"=>"Пайнхерст", "name_en"=>"Pinehurst", "lang"=>"en", "lat"=>30.107219, "lon"=>-93.770018),
"/US/US-TX/pineland-20953" => array("type"=>"city", "pop"=>874, "name_en"=>"Pineland", "lang"=>"en", "lat"=>31.246399, "lon"=>-93.971663),
"/US/US-TX/pinewood-estates-20954" => array("type"=>"cdp", "pop"=>1174, "name_en"=>"Pinewood Estates", "lang"=>"en", "lat"=>30.164091, "lon"=>-94.321330),
"/US/US-TX/piney-point-village-20955" => array("type"=>"city", "pop"=>3534, "name_en"=>"Piney Point Village", "lang"=>"en", "lat"=>29.759900, "lon"=>-95.515510),
"/US/US-TX/pittsburg-20956" => array("type"=>"city", "pop"=>4667, "name_ru"=>"Питсбург", "name_en"=>"Pittsburg", "lang"=>"en", "lat"=>32.996912, "lon"=>-94.965647),
"/US/US-TX/plains-20957" => array("type"=>"town", "pop"=>1491, "name_ru"=>"Плейнс", "name_en"=>"Plains", "lang"=>"en", "lat"=>33.190149, "lon"=>-102.827090),
"/US/US-TX/plainview-20958" => array("type"=>"city", "pop"=>21334, "name_ru"=>"Плейнвью", "name_en"=>"Plainview", "lang"=>"en", "lat"=>34.190931, "lon"=>-101.724614),
"/US/US-TX/plano-20959" => array("type"=>"city", "pop"=>267480, "name_ru"=>"Плано", "name_en"=>"Plano", "lang"=>"en", "lat"=>33.046201, "lon"=>-96.746664),
"/US/US-TX/pleak-20960" => array("type"=>"village", "pop"=>1338, "name_en"=>"Pleak", "lang"=>"en", "lat"=>29.483903, "lon"=>-95.809866),
"/US/US-TX/pleasant-valley-20962" => array("type"=>"town", "pop"=>370, "name_ru"=>"Плезант Валли", "name_en"=>"Pleasant Valley", "lang"=>"en", "lat"=>33.937488, "lon"=>-98.600962),
"/US/US-TX/pleasanton-20961" => array("type"=>"city", "pop"=>9681, "name_ru"=>"Плезантон", "name_en"=>"Pleasanton", "lang"=>"en", "lat"=>28.964601, "lon"=>-98.491455),
"/US/US-TX/plum-grove-20963" => array("type"=>"city", "pop"=>991, "name_en"=>"Plum Grove", "lang"=>"en", "lat"=>30.198531, "lon"=>-95.087373),
"/US/US-TX/point-20964" => array("type"=>"city", "pop"=>958, "name_en"=>"Point", "lang"=>"en", "lat"=>32.930129, "lon"=>-95.870137),
"/US/US-TX/point-blank-20965" => array("type"=>"city", "pop"=>626, "name_en"=>"Point Blank", "lang"=>"en", "lat"=>30.745919, "lon"=>-95.216782),
"/US/US-TX/point-comfort-20966" => array("type"=>"city", "pop"=>713, "name_en"=>"Point Comfort", "lang"=>"en", "lat"=>28.672127, "lon"=>-96.557001),
"/US/US-TX/ponder-20967" => array("type"=>"town", "pop"=>1288, "name_en"=>"Ponder", "lang"=>"en", "lat"=>33.179134, "lon"=>-97.287601),
"/US/US-TX/port-aransas-20968" => array("type"=>"city", "pop"=>3842, "name_en"=>"Port Aransas", "lang"=>"en", "lat"=>27.830724, "lon"=>-97.082557),
"/US/US-TX/port-arthur-20969" => array("type"=>"city", "pop"=>55910, "name_en"=>"Port Arthur", "lang"=>"en", "lat"=>29.826667, "lon"=>-93.934630),
"/US/US-TX/port-isabel-20971" => array("type"=>"city", "pop"=>5290, "name_en"=>"Port Isabel", "lang"=>"en", "lat"=>26.074762, "lon"=>-97.220145),
"/US/US-TX/port-lavaca-20973" => array("type"=>"city", "pop"=>11432, "name_en"=>"Port Lavaca", "lang"=>"en", "lat"=>28.614677, "lon"=>-96.621178),
"/US/US-TX/port-neches-20974" => array("type"=>"city", "pop"=>12604, "name_en"=>"Port Neches", "lang"=>"en", "lat"=>29.974150, "lon"=>-93.941201),
"/US/US-TX/porter-heights-20970" => array("type"=>"cdp", "pop"=>1448, "name_en"=>"Porter Heights", "lang"=>"en", "lat"=>30.151705, "lon"=>-95.321555),
"/US/US-TX/portland-20972" => array("type"=>"city", "pop"=>16490, "name_ru"=>"Портленд", "name_en"=>"Portland", "lang"=>"en", "lat"=>27.878872, "lon"=>-97.326929),
"/US/US-TX/post-20975" => array("type"=>"city", "pop"=>3483, "name_en"=>"Post", "lang"=>"en", "lat"=>33.190871, "lon"=>-101.381306),
"/US/US-TX/post-oak-bend-city-20976" => array("type"=>"town", "pop"=>487, "name_en"=>"Post Oak Bend City", "lang"=>"en", "lat"=>32.632069, "lon"=>-96.313260),
"/US/US-TX/poteet-20977" => array("type"=>"city", "pop"=>3640, "name_en"=>"Poteet", "lang"=>"en", "lat"=>29.037398, "lon"=>-98.575107),
"/US/US-TX/poth-20978" => array("type"=>"town", "pop"=>2356, "name_en"=>"Poth", "lang"=>"en", "lat"=>29.072405, "lon"=>-98.080912),
"/US/US-TX/potosi-20979" => array("type"=>"cdp", "pop"=>1441, "name_ru"=>"Потоси", "name_en"=>"Potosi", "lang"=>"en", "lat"=>32.335647, "lon"=>-99.684177),
"/US/US-TX/pottsboro-20980" => array("type"=>"town", "pop"=>2108, "name_en"=>"Pottsboro", "lang"=>"en", "lat"=>33.772243, "lon"=>-96.671606),
"/US/US-TX/powell-20981" => array("type"=>"town", "pop"=>114, "name_ru"=>"Пауэлл", "name_en"=>"Powell", "lang"=>"en", "lat"=>32.116476, "lon"=>-96.328716),
"/US/US-TX/poynor-20982" => array("type"=>"town", "pop"=>330, "name_en"=>"Poynor", "lang"=>"en", "lat"=>32.076006, "lon"=>-95.594950),
"/US/US-TX/prairie-view-20983" => array("type"=>"city", "pop"=>4492, "name_ru"=>"Прейри Вью", "name_en"=>"Prairie View", "lang"=>"en", "lat"=>30.083187, "lon"=>-95.989120),
"/US/US-TX/premont-20984" => array("type"=>"city", "pop"=>2789, "name_en"=>"Premont", "lang"=>"en", "lat"=>27.357778, "lon"=>-98.124090),
"/US/US-TX/presidio-20985" => array("type"=>"city", "pop"=>4695, "name_en"=>"Presidio", "lang"=>"en", "lat"=>29.561107, "lon"=>-104.363687),
"/US/US-TX/primera-20986" => array("type"=>"town", "pop"=>4230, "name_en"=>"Primera", "lang"=>"en", "lat"=>26.222800, "lon"=>-97.751199),
"/US/US-TX/princeton-20987" => array("type"=>"city", "pop"=>6117, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>33.185135, "lon"=>-96.504126),
"/US/US-TX/progreso-20988" => array("type"=>"cdp", "pop"=>5511, "name_en"=>"Progreso", "lang"=>"en", "lat"=>26.097475, "lon"=>-97.960458),
"/US/US-TX/progreso-lakes-20989" => array("type"=>"city", "pop"=>266, "name_en"=>"Progreso Lakes", "lang"=>"en", "lat"=>26.073029, "lon"=>-97.961091),
"/US/US-TX/prosper-20990" => array("type"=>"town", "pop"=>7053, "name_en"=>"Prosper", "lang"=>"en", "lat"=>33.240229, "lon"=>-96.785894),
"/US/US-TX/putnam-20991" => array("type"=>"town", "pop"=>92, "name_en"=>"Putnam", "lang"=>"en", "lat"=>32.369619, "lon"=>-99.195083),
"/US/US-TX/pyote-20992" => array("type"=>"town", "pop"=>131, "name_en"=>"Pyote", "lang"=>"en", "lat"=>31.537215, "lon"=>-103.122519),
"/US/US-TX/quanah-20993" => array("type"=>"city", "pop"=>2511, "name_en"=>"Quanah", "lang"=>"en", "lat"=>34.295598, "lon"=>-99.742507),
"/US/US-TX/queen-city-20994" => array("type"=>"city", "pop"=>1548, "name_ru"=>"Квин Сити", "name_en"=>"Queen City", "lang"=>"en", "lat"=>33.149723, "lon"=>-94.151290),
"/US/US-TX/quinlan-20995" => array("type"=>"city", "pop"=>1432, "name_en"=>"Quinlan", "lang"=>"en", "lat"=>32.908737, "lon"=>-96.132491),
"/US/US-TX/quintana-20996" => array("type"=>"town", "pop"=>40, "name_en"=>"Quintana", "lang"=>"en", "lat"=>28.930505, "lon"=>-95.315834),
"/US/US-TX/quitaque-20997" => array("type"=>"city", "pop"=>352, "name_en"=>"Quitaque", "lang"=>"en", "lat"=>34.367435, "lon"=>-101.054867),
"/US/US-TX/quitman-20998" => array("type"=>"city", "pop"=>2231, "name_ru"=>"Квитман", "name_en"=>"Quitman", "lang"=>"en", "lat"=>32.795899, "lon"=>-95.444271),
"/US/US-TX/ralls-20999" => array("type"=>"city", "pop"=>1974, "name_en"=>"Ralls", "lang"=>"en", "lat"=>33.678314, "lon"=>-101.384448),
"/US/US-TX/rancho-viejo-21000" => array("type"=>"town", "pop"=>1847, "name_en"=>"Rancho Viejo", "lang"=>"en", "lat"=>26.034700, "lon"=>-97.557595),
"/US/US-TX/ranger-21001" => array("type"=>"city", "pop"=>2559, "name_ru"=>"Рейнджер", "name_en"=>"Ranger", "lang"=>"en", "lat"=>32.470093, "lon"=>-98.674344),
"/US/US-TX/rangerville-21002" => array("type"=>"village", "pop"=>280, "name_en"=>"Rangerville", "lang"=>"en", "lat"=>26.102750, "lon"=>-97.738712),
"/US/US-TX/rankin-21003" => array("type"=>"city", "pop"=>749, "name_ru"=>"Ранкин", "name_en"=>"Rankin", "lang"=>"en", "lat"=>31.225249, "lon"=>-101.939353),
"/US/US-TX/ransom-canyon-21004" => array("type"=>"town", "pop"=>1114, "name_en"=>"Ransom Canyon", "lang"=>"en", "lat"=>33.529577, "lon"=>-101.682541),
"/US/US-TX/raymondville-21005" => array("type"=>"city", "pop"=>9522, "name_ru"=>"Реймондвилл", "name_en"=>"Raymondville", "lang"=>"en", "lat"=>26.479107, "lon"=>-97.780822),
"/US/US-TX/red-oak-21006" => array("type"=>"city", "pop"=>9255, "name_ru"=>"Ред Оук", "name_en"=>"Red Oak", "lang"=>"en", "lat"=>32.517902, "lon"=>-96.808833),
"/US/US-TX/redwater-21007" => array("type"=>"city", "pop"=>896, "name_ru"=>"Редуотер", "name_en"=>"Redwater", "lang"=>"en", "lat"=>33.358222, "lon"=>-94.254439),
"/US/US-TX/reese-afb-21008" => array("type"=>"cdp", "pop"=>1263, "name_en"=>"Reese AFB", "lang"=>"en", "lat"=>33.592541, "lon"=>-102.026875),
"/US/US-TX/refugio-21009" => array("type"=>"town", "pop"=>2717, "name_en"=>"Refugio", "lang"=>"en", "lat"=>28.307087, "lon"=>-97.275036),
"/US/US-TX/reklaw-21010" => array("type"=>"town", "pop"=>336, "name_en"=>"Reklaw", "lang"=>"en", "lat"=>31.864561, "lon"=>-94.984056),
"/US/US-TX/rendon-21011" => array("type"=>"cdp", "pop"=>7658, "name_en"=>"Rendon", "lang"=>"en", "lat"=>32.579979, "lon"=>-97.236542),
"/US/US-TX/reno-21012" => array("type"=>"city", "pop"=>3086, "name_en"=>"Reno", "lang"=>"en", "lat"=>33.666791, "lon"=>-95.480170),
"/US/US-TX/reno-21013" => array("type"=>"city", "pop"=>2322, "name_en"=>"Reno", "lang"=>"en", "lat"=>32.940865, "lon"=>-97.583041),
"/US/US-TX/retreat-21014" => array("type"=>"town", "pop"=>396, "name_en"=>"Retreat", "lang"=>"en", "lat"=>32.051301, "lon"=>-96.477089),
"/US/US-TX/rhome-21015" => array("type"=>"city", "pop"=>1051, "name_en"=>"Rhome", "lang"=>"en", "lat"=>33.050964, "lon"=>-97.467278),
"/US/US-TX/rice-21016" => array("type"=>"city", "pop"=>980, "name_ru"=>"Райс", "name_en"=>"Rice", "lang"=>"en", "lat"=>32.244688, "lon"=>-96.501328),
"/US/US-TX/richardson-21017" => array("type"=>"city", "pop"=>101589, "name_en"=>"Richardson", "lang"=>"en", "lat"=>32.972751, "lon"=>-96.709846),
"/US/US-TX/richland-21018" => array("type"=>"town", "pop"=>320, "name_ru"=>"Ричленд", "name_en"=>"Richland", "lang"=>"en", "lat"=>31.925045, "lon"=>-96.425828),
"/US/US-TX/richland-hills-21019" => array("type"=>"city", "pop"=>8094, "name_en"=>"Richland Hills", "lang"=>"en", "lat"=>32.809951, "lon"=>-97.226078),
"/US/US-TX/richland-springs-21020" => array("type"=>"town", "pop"=>330, "name_en"=>"Richland Springs", "lang"=>"en", "lat"=>31.270161, "lon"=>-98.945331),
"/US/US-TX/richmond-21021" => array("type"=>"city", "pop"=>13493, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>29.581411, "lon"=>-95.765049),
"/US/US-TX/richwood-21022" => array("type"=>"city", "pop"=>3451, "name_ru"=>"Ричвуд", "name_en"=>"Richwood", "lang"=>"en", "lat"=>29.057996, "lon"=>-95.407499),
"/US/US-TX/riesel-21023" => array("type"=>"city", "pop"=>1013, "name_en"=>"Riesel", "lang"=>"en", "lat"=>31.473905, "lon"=>-96.930777),
"/US/US-TX/rio-grande-city-21024" => array("type"=>"cdp", "pop"=>13950, "name_en"=>"Rio Grande City", "lang"=>"en", "lat"=>26.378455, "lon"=>-98.826250),
"/US/US-TX/rio-hondo-21025" => array("type"=>"city", "pop"=>2128, "name_en"=>"Rio Hondo", "lang"=>"en", "lat"=>26.234300, "lon"=>-97.582621),
"/US/US-TX/rio-vista-21026" => array("type"=>"city", "pop"=>818, "name_ru"=>"Рио Виста", "name_en"=>"Rio Vista", "lang"=>"en", "lat"=>32.235688, "lon"=>-97.375463),
"/US/US-TX/rising-star-21027" => array("type"=>"town", "pop"=>827, "name_en"=>"Rising Star", "lang"=>"en", "lat"=>32.097322, "lon"=>-98.965730),
"/US/US-TX/river-oaks-21028" => array("type"=>"city", "pop"=>6965, "name_en"=>"River Oaks", "lang"=>"en", "lat"=>32.775701, "lon"=>-97.397950),
"/US/US-TX/riverside-21029" => array("type"=>"city", "pop"=>406, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>30.847476, "lon"=>-95.398840),
"/US/US-TX/roanoke-21030" => array("type"=>"city", "pop"=>4235, "name_ru"=>"Роанок", "name_en"=>"Roanoke", "lang"=>"en", "lat"=>33.002128, "lon"=>-97.232264),
"/US/US-TX/roaring-springs-21031" => array("type"=>"town", "pop"=>234, "name_en"=>"Roaring Springs", "lang"=>"en", "lat"=>33.899362, "lon"=>-100.855799),
"/US/US-TX/robert-lee-21032" => array("type"=>"city", "pop"=>1031, "name_en"=>"Robert Lee", "lang"=>"en", "lat"=>31.895544, "lon"=>-100.484536),
"/US/US-TX/robinson-21033" => array("type"=>"city", "pop"=>10327, "name_ru"=>"Робинсон", "name_en"=>"Robinson", "lang"=>"en", "lat"=>31.449588, "lon"=>-97.124363),
"/US/US-TX/robstown-21034" => array("type"=>"city", "pop"=>12179, "name_en"=>"Robstown", "lang"=>"en", "lat"=>27.799176, "lon"=>-97.660723),
"/US/US-TX/roby-21035" => array("type"=>"city", "pop"=>600, "name_en"=>"Roby", "lang"=>"en", "lat"=>32.746310, "lon"=>-100.378482),
"/US/US-TX/rochester-21036" => array("type"=>"town", "pop"=>326, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>33.314168, "lon"=>-99.856141),
"/US/US-TX/rockdale-21037" => array("type"=>"city", "pop"=>5966, "name_ru"=>"Рокдейл", "name_en"=>"Rockdale", "lang"=>"en", "lat"=>30.655313, "lon"=>-97.007690),
"/US/US-TX/rockport-21038" => array("type"=>"city", "pop"=>9810, "name_ru"=>"Рокпорт", "name_en"=>"Rockport", "lang"=>"en", "lat"=>28.040731, "lon"=>-97.051534),
"/US/US-TX/rocksprings-21039" => array("type"=>"town", "pop"=>1159, "name_en"=>"Rocksprings", "lang"=>"en", "lat"=>30.016845, "lon"=>-100.211847),
"/US/US-TX/rockwall-21040" => array("type"=>"city", "pop"=>35247, "name_en"=>"Rockwall", "lang"=>"en", "lat"=>32.921953, "lon"=>-96.457498),
"/US/US-TX/rocky-mound-21041" => array("type"=>"town", "pop"=>104, "name_en"=>"Rocky Mound", "lang"=>"en", "lat"=>33.028465, "lon"=>-95.031688),
"/US/US-TX/rogers-21042" => array("type"=>"town", "pop"=>1138, "name_ru"=>"Роджерс", "name_en"=>"Rogers", "lang"=>"en", "lat"=>30.932206, "lon"=>-97.228455),
"/US/US-TX/rolling-meadows-21043" => array("type"=>"city", "pop"=>291, "name_ru"=>"Роллинг Медоус", "name_en"=>"Rolling Meadows", "lang"=>"en", "lat"=>32.412568, "lon"=>-94.807591),
"/US/US-TX/rollingwood-21044" => array("type"=>"city", "pop"=>1435, "name_en"=>"Rollingwood", "lang"=>"en", "lat"=>30.273833, "lon"=>-97.786833),
"/US/US-TX/roma-21045" => array("type"=>"city", "pop"=>11257, "name_en"=>"Roma", "lang"=>"en", "lat"=>26.403399, "lon"=>-99.002415),
"/US/US-TX/roman-forest-21046" => array("type"=>"town", "pop"=>3762, "name_en"=>"Roman Forest", "lang"=>"en", "lat"=>30.177749, "lon"=>-95.159278),
"/US/US-TX/ropesville-21047" => array("type"=>"city", "pop"=>509, "name_en"=>"Ropesville", "lang"=>"en", "lat"=>33.413658, "lon"=>-102.154330),
"/US/US-TX/roscoe-21048" => array("type"=>"city", "pop"=>1271, "name_ru"=>"Роско", "name_en"=>"Roscoe", "lang"=>"en", "lat"=>32.442429, "lon"=>-100.542031),
"/US/US-TX/rose-city-21050" => array("type"=>"city", "pop"=>504, "name_ru"=>"Роуз Сити", "name_en"=>"Rose City", "lang"=>"en", "lat"=>30.104096, "lon"=>-94.050021),
"/US/US-TX/rose-hill-acres-21051" => array("type"=>"city", "pop"=>501, "name_en"=>"Rose Hill Acres", "lang"=>"en", "lat"=>30.195600, "lon"=>-94.193911),
"/US/US-TX/rosebud-21049" => array("type"=>"city", "pop"=>1336, "name_ru"=>"Роузбад", "name_en"=>"Rosebud", "lang"=>"en", "lat"=>31.074951, "lon"=>-96.975258),
"/US/US-TX/rosenberg-21052" => array("type"=>"city", "pop"=>33595, "name_en"=>"Rosenberg", "lang"=>"en", "lat"=>29.548565, "lon"=>-95.794566),
"/US/US-TX/ross-21053" => array("type"=>"city", "pop"=>234, "name_ru"=>"Росс", "name_en"=>"Ross", "lang"=>"en", "lat"=>31.728441, "lon"=>-97.111209),
"/US/US-TX/rosser-21054" => array("type"=>"village", "pop"=>462, "name_en"=>"Rosser", "lang"=>"en", "lat"=>32.462727, "lon"=>-96.442211),
"/US/US-TX/rotan-21055" => array("type"=>"city", "pop"=>1429, "name_en"=>"Rotan", "lang"=>"en", "lat"=>32.853620, "lon"=>-100.465417),
"/US/US-TX/round-mountain-21056" => array("type"=>"town", "pop"=>117, "name_en"=>"Round Mountain", "lang"=>"en", "lat"=>30.436114, "lon"=>-98.353830),
"/US/US-TX/round-rock-21057" => array("type"=>"city", "pop"=>104446, "name_en"=>"Round Rock", "lang"=>"en", "lat"=>30.517477, "lon"=>-97.672068),
"/US/US-TX/round-top-21058" => array("type"=>"town", "pop"=>84, "name_en"=>"Round Top", "lang"=>"en", "lat"=>30.064406, "lon"=>-96.695673),
"/US/US-TX/rowlett-21059" => array("type"=>"city", "pop"=>55780, "name_en"=>"Rowlett", "lang"=>"en", "lat"=>32.908103, "lon"=>-96.549956),
"/US/US-TX/roxton-21060" => array("type"=>"city", "pop"=>695, "name_en"=>"Roxton", "lang"=>"en", "lat"=>33.545490, "lon"=>-95.724690),
"/US/US-TX/royse-city-21061" => array("type"=>"city", "pop"=>9101, "name_en"=>"Royse City", "lang"=>"en", "lat"=>32.975289, "lon"=>-96.332926),
"/US/US-TX/rule-21062" => array("type"=>"town", "pop"=>599, "name_en"=>"Rule", "lang"=>"en", "lat"=>33.181998, "lon"=>-99.892942),
"/US/US-TX/runaway-bay-21063" => array("type"=>"city", "pop"=>1424, "name_en"=>"Runaway Bay", "lang"=>"en", "lat"=>33.174889, "lon"=>-97.866542),
"/US/US-TX/runge-21064" => array("type"=>"town", "pop"=>1052, "name_en"=>"Runge", "lang"=>"en", "lat"=>28.885891, "lon"=>-97.712820),
"/US/US-TX/rusk-21065" => array("type"=>"city", "pop"=>5312, "name_en"=>"Rusk", "lang"=>"en", "lat"=>31.797577, "lon"=>-95.148535),
"/US/US-TX/sabinal-21066" => array("type"=>"city", "pop"=>1621, "name_en"=>"Sabinal", "lang"=>"en", "lat"=>29.321406, "lon"=>-99.468721),
"/US/US-TX/sachse-21067" => array("type"=>"city", "pop"=>18688, "name_en"=>"Sachse", "lang"=>"en", "lat"=>32.974577, "lon"=>-96.580842),
"/US/US-TX/sadler-21068" => array("type"=>"city", "pop"=>437, "name_en"=>"Sadler", "lang"=>"en", "lat"=>33.683126, "lon"=>-96.845692),
"/US/US-TX/saginaw-21069" => array("type"=>"city", "pop"=>20295, "name_ru"=>"Сагино", "name_en"=>"Saginaw", "lang"=>"en", "lat"=>32.865228, "lon"=>-97.365342),
"/US/US-TX/salado-21073" => array("type"=>"cdp", "pop"=>2061, "name_en"=>"Salado", "lang"=>"en", "lat"=>30.950771, "lon"=>-97.524690),
"/US/US-TX/san-angelo-21074" => array("type"=>"city", "pop"=>91880, "name_en"=>"San Angelo", "lang"=>"en", "lat"=>31.446050, "lon"=>-100.452424),
"/US/US-TX/san-antonio-21075" => array("type"=>"city", "pop"=>1351305, "name_ru"=>"Сан Антонио", "name_en"=>"San Antonio", "lang"=>"en", "lat"=>29.457650, "lon"=>-98.505355),
"/US/US-TX/san-augustine-21076" => array("type"=>"town", "pop"=>2346, "name_en"=>"San Augustine", "lang"=>"en", "lat"=>31.529852, "lon"=>-94.110991),
"/US/US-TX/san-benito-21077" => array("type"=>"city", "pop"=>25072, "name_en"=>"San Benito", "lang"=>"en", "lat"=>26.137100, "lon"=>-97.635334),
"/US/US-TX/san-diego-21080" => array("type"=>"city", "pop"=>4433, "name_ru"=>"Сан Диего", "name_en"=>"San Diego", "lang"=>"en", "lat"=>27.759161, "lon"=>-98.238258),
"/US/US-TX/san-elizario-21081" => array("type"=>"cdp", "pop"=>4385, "name_en"=>"San Elizario", "lang"=>"en", "lat"=>31.578428, "lon"=>-106.262146),
"/US/US-TX/san-felipe-21082" => array("type"=>"town", "pop"=>973, "name_en"=>"San Felipe", "lang"=>"en", "lat"=>29.798555, "lon"=>-96.105602),
"/US/US-TX/san-juan-21085" => array("type"=>"city", "pop"=>33970, "name_en"=>"San Juan", "lang"=>"en", "lat"=>26.191300, "lon"=>-98.154119),
"/US/US-TX/san-leanna-21086" => array("type"=>"village", "pop"=>503, "name_en"=>"San Leanna", "lang"=>"en", "lat"=>30.144900, "lon"=>-97.818660),
"/US/US-TX/san-leon-21087" => array("type"=>"cdp", "pop"=>3328, "name_en"=>"San Leon", "lang"=>"en", "lat"=>29.488552, "lon"=>-94.938261),
"/US/US-TX/san-marcos-21088" => array("type"=>"city", "pop"=>52927, "name_ru"=>"Сан Маркос", "name_en"=>"San Marcos", "lang"=>"en", "lat"=>29.880178, "lon"=>-97.929043),
"/US/US-TX/san-patricio-21089" => array("type"=>"city", "pop"=>303, "name_en"=>"San Patricio", "lang"=>"en", "lat"=>27.978500, "lon"=>-97.772590),
"/US/US-TX/san-perlita-21090" => array("type"=>"city", "pop"=>695, "name_en"=>"San Perlita", "lang"=>"en", "lat"=>26.500418, "lon"=>-97.639819),
"/US/US-TX/san-saba-21091" => array("type"=>"town", "pop"=>2499, "name_en"=>"San Saba", "lang"=>"en", "lat"=>31.195142, "lon"=>-98.725557),
"/US/US-TX/sanctuary-21078" => array("type"=>"town", "pop"=>438, "name_en"=>"Sanctuary", "lang"=>"en", "lat"=>32.908697, "lon"=>-97.583907),
"/US/US-TX/sanderson-21079" => array("type"=>"cdp", "pop"=>1128, "name_en"=>"Sanderson", "lang"=>"en", "lat"=>30.152639, "lon"=>-102.407720),
"/US/US-TX/sanford-21083" => array("type"=>"town", "pop"=>189, "name_ru"=>"Санфорд", "name_en"=>"Sanford", "lang"=>"en", "lat"=>35.702585, "lon"=>-101.531326),
"/US/US-TX/sanger-21084" => array("type"=>"city", "pop"=>7766, "name_ru"=>"Сангер", "name_en"=>"Sanger", "lang"=>"en", "lat"=>33.360724, "lon"=>-97.176801),
"/US/US-TX/sansom-park-21092" => array("type"=>"city", "pop"=>4196, "name_en"=>"Sansom Park", "lang"=>"en", "lat"=>32.802600, "lon"=>-97.401725),
"/US/US-TX/santa-anna-21093" => array("type"=>"town", "pop"=>1016, "name_en"=>"Santa Anna", "lang"=>"en", "lat"=>31.737508, "lon"=>-99.324867),
"/US/US-TX/santa-fe-21094" => array("type"=>"city", "pop"=>10479, "name_en"=>"Santa Fe", "lang"=>"en", "lat"=>29.381971, "lon"=>-95.098268),
"/US/US-TX/santa-rosa-21095" => array("type"=>"town", "pop"=>3152, "name_ru"=>"Санта Роза", "name_en"=>"Santa Rosa", "lang"=>"en", "lat"=>26.254250, "lon"=>-97.825399),
"/US/US-TX/savoy-21096" => array("type"=>"city", "pop"=>895, "name_ru"=>"Савой", "name_en"=>"Savoy", "lang"=>"en", "lat"=>33.599494, "lon"=>-96.369193),
"/US/US-TX/scenic-oaks-21097" => array("type"=>"cdp", "pop"=>2352, "name_en"=>"Scenic Oaks", "lang"=>"en", "lat"=>29.710644, "lon"=>-98.675655),
"/US/US-TX/schertz-21098" => array("type"=>"city", "pop"=>30669, "name_en"=>"Schertz", "lang"=>"en", "lat"=>29.554800, "lon"=>-98.261997),
"/US/US-TX/schulenburg-21099" => array("type"=>"city", "pop"=>2692, "name_en"=>"Schulenburg", "lang"=>"en", "lat"=>29.681198, "lon"=>-96.906899),
"/US/US-TX/scissors-21100" => array("type"=>"cdp", "pop"=>1513, "name_en"=>"Scissors", "lang"=>"en", "lat"=>26.130009, "lon"=>-98.051569),
"/US/US-TX/scotland-21101" => array("type"=>"city", "pop"=>427, "name_ru"=>"Скотленд", "name_en"=>"Scotland", "lang"=>"en", "lat"=>33.648681, "lon"=>-98.455644),
"/US/US-TX/scottsville-21102" => array("type"=>"city", "pop"=>272, "name_ru"=>"Скоттсвилл", "name_en"=>"Scottsville", "lang"=>"en", "lat"=>32.535948, "lon"=>-94.241420),
"/US/US-TX/seabrook-21103" => array("type"=>"city", "pop"=>11525, "name_en"=>"Seabrook", "lang"=>"en", "lat"=>29.579013, "lon"=>-94.992965),
"/US/US-TX/seadrift-21104" => array("type"=>"city", "pop"=>1433, "name_en"=>"Seadrift", "lang"=>"en", "lat"=>28.413945, "lon"=>-96.715177),
"/US/US-TX/seagoville-21105" => array("type"=>"city", "pop"=>11998, "name_en"=>"Seagoville", "lang"=>"en", "lat"=>32.656675, "lon"=>-96.546473),
"/US/US-TX/seagraves-21106" => array("type"=>"city", "pop"=>2348, "name_en"=>"Seagraves", "lang"=>"en", "lat"=>32.941523, "lon"=>-102.564893),
"/US/US-TX/sealy-21107" => array("type"=>"city", "pop"=>6275, "name_en"=>"Sealy", "lang"=>"en", "lat"=>29.773668, "lon"=>-96.152938),
"/US/US-TX/sebastian-21108" => array("type"=>"cdp", "pop"=>1598, "name_ru"=>"Себастьян", "name_en"=>"Sebastian", "lang"=>"en", "lat"=>26.343900, "lon"=>-97.797299),
"/US/US-TX/seguin-21109" => array("type"=>"city", "pop"=>26394, "name_en"=>"Seguin", "lang"=>"en", "lat"=>29.579500, "lon"=>-97.968774),
"/US/US-TX/selma-21110" => array("type"=>"city", "pop"=>4632, "name_ru"=>"Сельма", "name_en"=>"Selma", "lang"=>"en", "lat"=>29.586300, "lon"=>-98.316831),
"/US/US-TX/seminole-21111" => array("type"=>"city", "pop"=>6122, "name_ru"=>"Семиноул", "name_en"=>"Seminole", "lang"=>"en", "lat"=>32.718322, "lon"=>-102.652553),
"/US/US-TX/serenada-21112" => array("type"=>"cdp", "pop"=>3242, "name_en"=>"Serenada", "lang"=>"en", "lat"=>30.699133, "lon"=>-97.691539),
"/US/US-TX/seth-ward-21113" => array("type"=>"cdp", "pop"=>1402, "name_en"=>"Seth Ward", "lang"=>"en", "lat"=>34.213657, "lon"=>-101.690400),
"/US/US-TX/seven-oaks-21114" => array("type"=>"city", "pop"=>144, "name_en"=>"Seven Oaks", "lang"=>"en", "lat"=>30.853374, "lon"=>-94.856218),
"/US/US-TX/seven-points-21115" => array("type"=>"city", "pop"=>1323, "name_en"=>"Seven Points", "lang"=>"en", "lat"=>32.325590, "lon"=>-96.214833),
"/US/US-TX/seymour-21116" => array("type"=>"city", "pop"=>2619, "name_ru"=>"Сеймур", "name_en"=>"Seymour", "lang"=>"en", "lat"=>33.596927, "lon"=>-99.259269),
"/US/US-TX/shady-shores-21117" => array("type"=>"town", "pop"=>2604, "name_en"=>"Shady Shores", "lang"=>"en", "lat"=>33.162728, "lon"=>-97.044302),
"/US/US-TX/shallowater-21118" => array("type"=>"city", "pop"=>2289, "name_en"=>"Shallowater", "lang"=>"en", "lat"=>33.690065, "lon"=>-101.992340),
"/US/US-TX/shamrock-21119" => array("type"=>"city", "pop"=>1807, "name_en"=>"Shamrock", "lang"=>"en", "lat"=>35.215990, "lon"=>-100.245546),
"/US/US-TX/shavano-park-21120" => array("type"=>"city", "pop"=>3225, "name_en"=>"Shavano Park", "lang"=>"en", "lat"=>29.584100, "lon"=>-98.555013),
"/US/US-TX/sheldon-21121" => array("type"=>"cdp", "pop"=>1653, "name_ru"=>"Шелдон", "name_en"=>"Sheldon", "lang"=>"en", "lat"=>29.860346, "lon"=>-95.133593),
"/US/US-TX/shenandoah-21122" => array("type"=>"city", "pop"=>2002, "name_ru"=>"Шенандоа", "name_en"=>"Shenandoah", "lang"=>"en", "lat"=>30.181001, "lon"=>-95.454815),
"/US/US-TX/shepherd-21123" => array("type"=>"city", "pop"=>2259, "name_ru"=>"Шеперд", "name_en"=>"Shepherd", "lang"=>"en", "lat"=>30.490451, "lon"=>-95.002991),
"/US/US-TX/sherman-21124" => array("type"=>"city", "pop"=>38077, "name_ru"=>"Шерман", "name_en"=>"Sherman", "lang"=>"en", "lat"=>33.618335, "lon"=>-96.609456),
"/US/US-TX/shiner-21125" => array("type"=>"city", "pop"=>1984, "name_en"=>"Shiner", "lang"=>"en", "lat"=>29.431991, "lon"=>-97.171790),
"/US/US-TX/shoreacres-21126" => array("type"=>"city", "pop"=>1656, "name_en"=>"Shoreacres", "lang"=>"en", "lat"=>29.617918, "lon"=>-95.001556),
"/US/US-TX/silsbee-21127" => array("type"=>"city", "pop"=>6870, "name_en"=>"Silsbee", "lang"=>"en", "lat"=>30.344730, "lon"=>-94.180376),
"/US/US-TX/silverton-21128" => array("type"=>"city", "pop"=>628, "name_ru"=>"Сильвертон", "name_en"=>"Silverton", "lang"=>"en", "lat"=>34.471193, "lon"=>-101.304311),
"/US/US-TX/simonton-21129" => array("type"=>"city", "pop"=>873, "name_en"=>"Simonton", "lang"=>"en", "lat"=>29.681443, "lon"=>-95.995240),
"/US/US-TX/sinton-21130" => array("type"=>"city", "pop"=>5354, "name_en"=>"Sinton", "lang"=>"en", "lat"=>28.033482, "lon"=>-97.509572),
"/US/US-TX/skellytown-21131" => array("type"=>"town", "pop"=>572, "name_en"=>"Skellytown", "lang"=>"en", "lat"=>35.573088, "lon"=>-101.173043),
"/US/US-TX/slaton-21132" => array("type"=>"city", "pop"=>5803, "name_en"=>"Slaton", "lang"=>"en", "lat"=>33.439066, "lon"=>-101.642275),
"/US/US-TX/smiley-21133" => array("type"=>"city", "pop"=>474, "name_en"=>"Smiley", "lang"=>"en", "lat"=>29.271476, "lon"=>-97.637547),
"/US/US-TX/smithville-21134" => array("type"=>"city", "pop"=>4458, "name_ru"=>"Смитвилл", "name_en"=>"Smithville", "lang"=>"en", "lat"=>30.007734, "lon"=>-97.156554),
"/US/US-TX/smyer-21135" => array("type"=>"town", "pop"=>479, "name_en"=>"Smyer", "lang"=>"en", "lat"=>33.586788, "lon"=>-102.163239),
"/US/US-TX/snook-21136" => array("type"=>"city", "pop"=>569, "name_en"=>"Snook", "lang"=>"en", "lat"=>30.490552, "lon"=>-96.465735),
"/US/US-TX/snyder-21137" => array("type"=>"city", "pop"=>10406, "name_en"=>"Snyder", "lang"=>"en", "lat"=>32.713495, "lon"=>-100.915115),
"/US/US-TX/socorro-21138" => array("type"=>"town", "pop"=>32007, "name_en"=>"Socorro", "lang"=>"en", "lat"=>31.638744, "lon"=>-106.269691),
"/US/US-TX/somerset-21139" => array("type"=>"city", "pop"=>1840, "name_ru"=>"Сомерсет", "name_en"=>"Somerset", "lang"=>"en", "lat"=>29.227589, "lon"=>-98.656690),
"/US/US-TX/somerville-21140" => array("type"=>"city", "pop"=>1686, "name_ru"=>"Сомервилл", "name_en"=>"Somerville", "lang"=>"en", "lat"=>30.345220, "lon"=>-96.530537),
"/US/US-TX/sonora-21141" => array("type"=>"city", "pop"=>3037, "name_ru"=>"Сонора", "name_en"=>"Sonora", "lang"=>"en", "lat"=>30.572168, "lon"=>-100.643744),
"/US/US-TX/sour-lake-21142" => array("type"=>"city", "pop"=>1734, "name_en"=>"Sour Lake", "lang"=>"en", "lat"=>30.136005, "lon"=>-94.403929),
"/US/US-TX/south-houston-21143" => array("type"=>"city", "pop"=>16351, "name_en"=>"South Houston", "lang"=>"en", "lat"=>29.660750, "lon"=>-95.228049),
"/US/US-TX/south-mountain-21146" => array("type"=>"town", "pop"=>372, "name_en"=>"South Mountain", "lang"=>"en", "lat"=>31.438283, "lon"=>-97.676298),
"/US/US-TX/south-padre-island-21147" => array("type"=>"town", "pop"=>2837, "name_en"=>"South Padre Island", "lang"=>"en", "lat"=>26.110888, "lon"=>-97.167160),
"/US/US-TX/southlake-21144" => array("type"=>"city", "pop"=>26595, "name_en"=>"Southlake", "lang"=>"en", "lat"=>32.953176, "lon"=>-97.148341),
"/US/US-TX/southmayd-21145" => array("type"=>"town", "pop"=>1078, "name_en"=>"Southmayd", "lang"=>"en", "lat"=>33.621406, "lon"=>-96.720111),
"/US/US-TX/southside-place-21148" => array("type"=>"city", "pop"=>1667, "name_en"=>"Southside Place", "lang"=>"en", "lat"=>29.708450, "lon"=>-95.435199),
"/US/US-TX/sparks-21149" => array("type"=>"cdp", "pop"=>1276, "name_ru"=>"Спаркс", "name_en"=>"Sparks", "lang"=>"en", "lat"=>31.672565, "lon"=>-106.239305),
"/US/US-TX/spearman-21150" => array("type"=>"city", "pop"=>2944, "name_en"=>"Spearman", "lang"=>"en", "lat"=>36.197484, "lon"=>-101.193613),
"/US/US-TX/splendora-21151" => array("type"=>"city", "pop"=>1604, "name_en"=>"Splendora", "lang"=>"en", "lat"=>30.230093, "lon"=>-95.163991),
"/US/US-TX/spofford-21152" => array("type"=>"city", "pop"=>72, "name_en"=>"Spofford", "lang"=>"en", "lat"=>29.172442, "lon"=>-100.411034),
"/US/US-TX/spring-21153" => array("type"=>"cdp", "pop"=>33111, "name_en"=>"Spring", "lang"=>"en", "lat"=>30.061941, "lon"=>-95.383807),
"/US/US-TX/spring-valley-21156" => array("type"=>"city", "pop"=>3879, "name_ru"=>"Спринг Валли", "name_en"=>"Spring Valley", "lang"=>"en", "lat"=>29.788067, "lon"=>-95.503946),
"/US/US-TX/springlake-21154" => array("type"=>"town", "pop"=>128, "name_en"=>"Springlake", "lang"=>"en", "lat"=>34.232198, "lon"=>-102.305071),
"/US/US-TX/springtown-21155" => array("type"=>"city", "pop"=>3130, "name_en"=>"Springtown", "lang"=>"en", "lat"=>32.969289, "lon"=>-97.680985),
"/US/US-TX/spur-21157" => array("type"=>"city", "pop"=>937, "name_en"=>"Spur", "lang"=>"en", "lat"=>33.478130, "lon"=>-100.854866),
"/US/US-TX/st.-hedwig-21070" => array("type"=>"town", "pop"=>2107, "name_en"=>"St. Hedwig", "lang"=>"en", "lat"=>29.424250, "lon"=>-98.214056),
"/US/US-TX/st.-jo-21071" => array("type"=>"city", "pop"=>974, "name_en"=>"St. Jo", "lang"=>"en", "lat"=>33.695239, "lon"=>-97.522388),
"/US/US-TX/st.-paul-21072" => array("type"=>"town", "pop"=>980, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>33.046173, "lon"=>-96.548590),
"/US/US-TX/stafford-21158" => array("type"=>"town", "pop"=>19541, "name_ru"=>"Стэффорд", "name_en"=>"Stafford", "lang"=>"en", "lat"=>29.623953, "lon"=>-95.563137),
"/US/US-TX/stagecoach-21159" => array("type"=>"town", "pop"=>508, "name_en"=>"Stagecoach", "lang"=>"en", "lat"=>30.141242, "lon"=>-95.710062),
"/US/US-TX/stamford-21160" => array("type"=>"city", "pop"=>3111, "name_ru"=>"Стэмфорд", "name_en"=>"Stamford", "lang"=>"en", "lat"=>33.021530, "lon"=>-99.642237),
"/US/US-TX/stanton-21161" => array("type"=>"city", "pop"=>2225, "name_ru"=>"Стэнтон", "name_en"=>"Stanton", "lang"=>"en", "lat"=>32.129301, "lon"=>-101.789854),
"/US/US-TX/star-harbor-21162" => array("type"=>"town", "pop"=>467, "name_en"=>"Star Harbor", "lang"=>"en", "lat"=>32.193753, "lon"=>-96.055744),
"/US/US-TX/stephenville-21163" => array("type"=>"city", "pop"=>16921, "name_en"=>"Stephenville", "lang"=>"en", "lat"=>32.219100, "lon"=>-98.215284),
"/US/US-TX/sterling-city-21164" => array("type"=>"city", "pop"=>978, "name_en"=>"Sterling City", "lang"=>"en", "lat"=>31.838951, "lon"=>-100.985480),
"/US/US-TX/stinnett-21165" => array("type"=>"city", "pop"=>1796, "name_en"=>"Stinnett", "lang"=>"en", "lat"=>35.822221, "lon"=>-101.443155),
"/US/US-TX/stockdale-21166" => array("type"=>"city", "pop"=>1670, "name_en"=>"Stockdale", "lang"=>"en", "lat"=>29.233829, "lon"=>-97.962191),
"/US/US-TX/stowell-21167" => array("type"=>"cdp", "pop"=>1419, "name_en"=>"Stowell", "lang"=>"en", "lat"=>29.785538, "lon"=>-94.379907),
"/US/US-TX/stratford-21168" => array("type"=>"city", "pop"=>1913, "name_ru"=>"Стрэтфорд", "name_en"=>"Stratford", "lang"=>"en", "lat"=>36.337227, "lon"=>-102.073573),
"/US/US-TX/strawn-21169" => array("type"=>"city", "pop"=>762, "name_ru"=>"Строн", "name_en"=>"Strawn", "lang"=>"en", "lat"=>32.550783, "lon"=>-98.497775),
"/US/US-TX/streetman-21170" => array("type"=>"town", "pop"=>210, "name_en"=>"Streetman", "lang"=>"en", "lat"=>31.875883, "lon"=>-96.323853),
"/US/US-TX/sudan-21171" => array("type"=>"city", "pop"=>979, "name_en"=>"Sudan", "lang"=>"en", "lat"=>34.067523, "lon"=>-102.524864),
"/US/US-TX/sugar-land-21172" => array("type"=>"city", "pop"=>80704, "name_en"=>"Sugar Land", "lang"=>"en", "lat"=>29.618777, "lon"=>-95.616183),
"/US/US-TX/sullivan-city-21173" => array("type"=>"cdp", "pop"=>4435, "name_en"=>"Sullivan City", "lang"=>"en", "lat"=>26.281707, "lon"=>-98.560751),
"/US/US-TX/sulphur-springs-21174" => array("type"=>"city", "pop"=>15473, "name_ru"=>"Салфер Спрингс", "name_en"=>"Sulphur Springs", "lang"=>"en", "lat"=>33.143873, "lon"=>-95.610110),
"/US/US-TX/sun-valley-21180" => array("type"=>"city", "pop"=>53, "name_ru"=>"Сан Вэлли", "name_en"=>"Sun Valley", "lang"=>"en", "lat"=>33.672497, "lon"=>-95.428797),
"/US/US-TX/sundown-21175" => array("type"=>"city", "pop"=>1496, "name_en"=>"Sundown", "lang"=>"en", "lat"=>33.456836, "lon"=>-102.490736),
"/US/US-TX/sunnyvale-21176" => array("type"=>"town", "pop"=>4518, "name_ru"=>"Саннивейл", "name_en"=>"Sunnyvale", "lang"=>"en", "lat"=>32.799324, "lon"=>-96.558257),
"/US/US-TX/sunray-21177" => array("type"=>"city", "pop"=>1911, "name_en"=>"Sunray", "lang"=>"en", "lat"=>36.018214, "lon"=>-101.823426),
"/US/US-TX/sunrise-beach-village-21178" => array("type"=>"city", "pop"=>753, "name_en"=>"Sunrise Beach Village", "lang"=>"en", "lat"=>30.588298, "lon"=>-98.418784),
"/US/US-TX/sunset-valley-21179" => array("type"=>"city", "pop"=>884, "name_en"=>"Sunset Valley", "lang"=>"en", "lat"=>30.226100, "lon"=>-97.816415),
"/US/US-TX/surfside-beach-21181" => array("type"=>"city", "pop"=>882, "name_en"=>"Surfside Beach", "lang"=>"en", "lat"=>28.953802, "lon"=>-95.282245),
"/US/US-TX/sweeny-21182" => array("type"=>"town", "pop"=>3617, "name_en"=>"Sweeny", "lang"=>"en", "lat"=>29.045347, "lon"=>-95.699329),
"/US/US-TX/sweetwater-21183" => array("type"=>"city", "pop"=>10657, "name_ru"=>"Свитвотер", "name_en"=>"Sweetwater", "lang"=>"en", "lat"=>32.469047, "lon"=>-100.404629),
"/US/US-TX/taft-21184" => array("type"=>"city", "pop"=>3339, "name_ru"=>"Тафт", "name_en"=>"Taft", "lang"=>"en", "lat"=>27.981080, "lon"=>-97.390838),
"/US/US-TX/taft-southwest-21185" => array("type"=>"cdp", "pop"=>2012, "name_en"=>"Taft Southwest", "lang"=>"en", "lat"=>27.972272, "lon"=>-97.405278),
"/US/US-TX/tahoka-21186" => array("type"=>"city", "pop"=>2516, "name_en"=>"Tahoka", "lang"=>"en", "lat"=>33.164519, "lon"=>-101.794425),
"/US/US-TX/talco-21187" => array("type"=>"city", "pop"=>579, "name_en"=>"Talco", "lang"=>"en", "lat"=>33.362019, "lon"=>-95.103871),
"/US/US-TX/tanglewood-forest-21188" => array("type"=>"cdp", "pop"=>2941, "name_en"=>"Tanglewood Forest", "lang"=>"en", "lat"=>30.174347, "lon"=>-97.838912),
"/US/US-TX/tatum-21189" => array("type"=>"city", "pop"=>1211, "name_en"=>"Tatum", "lang"=>"en", "lat"=>32.314697, "lon"=>-94.518673),
"/US/US-TX/taylor-21190" => array("type"=>"city", "pop"=>16090, "name_ru"=>"Тейлор", "name_en"=>"Taylor", "lang"=>"en", "lat"=>30.571781, "lon"=>-97.410046),
"/US/US-TX/taylor-lake-village-21191" => array("type"=>"city", "pop"=>3553, "name_en"=>"Taylor Lake Village", "lang"=>"en", "lat"=>29.573673, "lon"=>-95.054784),
"/US/US-TX/teague-21192" => array("type"=>"city", "pop"=>4749, "name_en"=>"Teague", "lang"=>"en", "lat"=>31.630413, "lon"=>-96.281576),
"/US/US-TX/tehuacana-21193" => array("type"=>"town", "pop"=>311, "name_en"=>"Tehuacana", "lang"=>"en", "lat"=>31.744241, "lon"=>-96.543978),
"/US/US-TX/temple-21194" => array("type"=>"city", "pop"=>59654, "name_ru"=>"Тэмпл", "name_en"=>"Temple", "lang"=>"en", "lat"=>31.097658, "lon"=>-97.363458),
"/US/US-TX/tenaha-21195" => array("type"=>"town", "pop"=>1112, "name_en"=>"Tenaha", "lang"=>"en", "lat"=>31.943722, "lon"=>-94.244635),
"/US/US-TX/terrell-21196" => array("type"=>"city", "pop"=>19527, "name_en"=>"Terrell", "lang"=>"en", "lat"=>32.732834, "lon"=>-96.291858),
"/US/US-TX/terrell-hills-21197" => array("type"=>"city", "pop"=>5241, "name_en"=>"Terrell Hills", "lang"=>"en", "lat"=>29.477335, "lon"=>-98.446626),
"/US/US-TX/texarkana-21198" => array("type"=>"city", "pop"=>36611, "name_ru"=>"Тексаркана", "name_en"=>"Texarkana", "lang"=>"en", "lat"=>33.441050, "lon"=>-94.072430),
"/US/US-TX/texas-city-21199" => array("type"=>"city", "pop"=>44491, "name_en"=>"Texas City", "lang"=>"en", "lat"=>29.409367, "lon"=>-94.958020),
"/US/US-TX/texhoma-21200" => array("type"=>"city", "pop"=>314, "name_en"=>"Texhoma", "lang"=>"en", "lat"=>36.497733, "lon"=>-101.790013),
"/US/US-TX/texline-21201" => array("type"=>"town", "pop"=>520, "name_en"=>"Texline", "lang"=>"en", "lat"=>36.376744, "lon"=>-103.018746),
"/US/US-TX/the-colony-21202" => array("type"=>"city", "pop"=>42281, "name_en"=>"The Colony", "lang"=>"en", "lat"=>33.090942, "lon"=>-96.887118),
"/US/US-TX/the-woodlands-21203" => array("type"=>"cdp", "pop"=>29205, "name_en"=>"The Woodlands", "lang"=>"en", "lat"=>30.157674, "lon"=>-95.489289),
"/US/US-TX/thompsons-21204" => array("type"=>"town", "pop"=>439, "name_en"=>"Thompsons", "lang"=>"en", "lat"=>29.491019, "lon"=>-95.619634),
"/US/US-TX/thorndale-21205" => array("type"=>"city", "pop"=>1309, "name_en"=>"Thorndale", "lang"=>"en", "lat"=>30.614252, "lon"=>-97.204951),
"/US/US-TX/thornton-21206" => array("type"=>"town", "pop"=>531, "name_ru"=>"Торнтон", "name_en"=>"Thornton", "lang"=>"en", "lat"=>31.411738, "lon"=>-96.574614),
"/US/US-TX/thorntonville-21207" => array("type"=>"town", "pop"=>439, "name_en"=>"Thorntonville", "lang"=>"en", "lat"=>31.577987, "lon"=>-102.921647),
"/US/US-TX/thrall-21208" => array("type"=>"town", "pop"=>914, "name_en"=>"Thrall", "lang"=>"en", "lat"=>30.588441, "lon"=>-97.298434),
"/US/US-TX/three-rivers-21209" => array("type"=>"city", "pop"=>1675, "name_ru"=>"Три Риверс", "name_en"=>"Three Rivers", "lang"=>"en", "lat"=>28.465627, "lon"=>-98.177917),
"/US/US-TX/throckmorton-21210" => array("type"=>"town", "pop"=>813, "name_en"=>"Throckmorton", "lang"=>"en", "lat"=>33.181696, "lon"=>-99.179944),
"/US/US-TX/tiki-island-21211" => array("type"=>"village", "pop"=>1313, "name_en"=>"Tiki Island", "lang"=>"en", "lat"=>29.298700, "lon"=>-94.914177),
"/US/US-TX/timbercreek-canyon-21212" => array("type"=>"village", "pop"=>486, "name_en"=>"Timbercreek Canyon", "lang"=>"en", "lat"=>35.054844, "lon"=>-101.816002),
"/US/US-TX/timberwood-park-21213" => array("type"=>"cdp", "pop"=>2578, "name_en"=>"Timberwood Park", "lang"=>"en", "lat"=>29.705450, "lon"=>-98.478073),
"/US/US-TX/timpson-21214" => array("type"=>"city", "pop"=>1156, "name_en"=>"Timpson", "lang"=>"en", "lat"=>31.906213, "lon"=>-94.396354),
"/US/US-TX/tioga-21215" => array("type"=>"town", "pop"=>927, "name_en"=>"Tioga", "lang"=>"en", "lat"=>33.472230, "lon"=>-96.917192),
"/US/US-TX/tira-21216" => array("type"=>"town", "pop"=>259, "name_en"=>"Tira", "lang"=>"en", "lat"=>33.326182, "lon"=>-95.562321),
"/US/US-TX/toco-21217" => array("type"=>"city", "pop"=>89, "name_en"=>"Toco", "lang"=>"en", "lat"=>33.653355, "lon"=>-95.649105),
"/US/US-TX/todd-mission-21218" => array("type"=>"city", "pop"=>83, "name_en"=>"Todd Mission", "lang"=>"en", "lat"=>30.260366, "lon"=>-95.829231),
"/US/US-TX/tolar-21219" => array("type"=>"city", "pop"=>681, "name_en"=>"Tolar", "lang"=>"en", "lat"=>32.390006, "lon"=>-97.916838),
"/US/US-TX/tom-bean-21221" => array("type"=>"town", "pop"=>1030, "name_en"=>"Tom Bean", "lang"=>"en", "lat"=>33.520679, "lon"=>-96.483720),
"/US/US-TX/tomball-21220" => array("type"=>"city", "pop"=>10234, "name_en"=>"Tomball", "lang"=>"en", "lat"=>30.096947, "lon"=>-95.616405),
"/US/US-TX/tool-21222" => array("type"=>"city", "pop"=>2439, "name_en"=>"Tool", "lang"=>"en", "lat"=>32.280252, "lon"=>-96.172424),
"/US/US-TX/town-west-21223" => array("type"=>"cdp", "pop"=>6166, "name_en"=>"Town West", "lang"=>"en", "lat"=>29.663503, "lon"=>-95.630924),
"/US/US-TX/toyah-21224" => array("type"=>"town", "pop"=>85, "name_en"=>"Toyah", "lang"=>"en", "lat"=>31.312514, "lon"=>-103.793470),
"/US/US-TX/trent-21225" => array("type"=>"town", "pop"=>316, "name_en"=>"Trent", "lang"=>"en", "lat"=>32.487886, "lon"=>-100.122952),
"/US/US-TX/trenton-21226" => array("type"=>"city", "pop"=>719, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>33.429457, "lon"=>-96.339884),
"/US/US-TX/trinidad-21227" => array("type"=>"city", "pop"=>1109, "name_ru"=>"Тринидад", "name_en"=>"Trinidad", "lang"=>"en", "lat"=>32.148463, "lon"=>-96.099690),
"/US/US-TX/trinity-21228" => array("type"=>"city", "pop"=>2725, "name_ru"=>"Тринити", "name_en"=>"Trinity", "lang"=>"en", "lat"=>30.944203, "lon"=>-95.373335),
"/US/US-TX/trophy-club-21229" => array("type"=>"town", "pop"=>8069, "name_en"=>"Trophy Club", "lang"=>"en", "lat"=>33.000384, "lon"=>-97.192890),
"/US/US-TX/troup-21230" => array("type"=>"city", "pop"=>2119, "name_en"=>"Troup", "lang"=>"en", "lat"=>32.144615, "lon"=>-95.123150),
"/US/US-TX/troy-21231" => array("type"=>"city", "pop"=>1425, "name_ru"=>"Трой", "name_en"=>"Troy", "lang"=>"en", "lat"=>31.195613, "lon"=>-97.306816),
"/US/US-TX/tulia-21232" => array("type"=>"city", "pop"=>4589, "name_en"=>"Tulia", "lang"=>"en", "lat"=>34.537643, "lon"=>-101.771474),
"/US/US-TX/turkey-21233" => array("type"=>"city", "pop"=>456, "name_en"=>"Turkey", "lang"=>"en", "lat"=>34.394186, "lon"=>-100.895464),
"/US/US-TX/tuscola-21234" => array("type"=>"town", "pop"=>726, "name_ru"=>"Таскола", "name_en"=>"Tuscola", "lang"=>"en", "lat"=>32.210742, "lon"=>-99.798884),
"/US/US-TX/tye-21235" => array("type"=>"town", "pop"=>1140, "name_en"=>"Tye", "lang"=>"en", "lat"=>32.453200, "lon"=>-99.869355),
"/US/US-TX/tyler-21236" => array("type"=>"city", "pop"=>97705, "name_ru"=>"Тайлер", "name_en"=>"Tyler", "lang"=>"en", "lat"=>32.328400, "lon"=>-95.303599),
"/US/US-TX/uhland-21237" => array("type"=>"city", "pop"=>436, "name_en"=>"Uhland", "lang"=>"en", "lat"=>29.963652, "lon"=>-97.794599),
"/US/US-TX/uncertain-21238" => array("type"=>"city", "pop"=>155, "name_en"=>"Uncertain", "lang"=>"en", "lat"=>32.708559, "lon"=>-94.122488),
"/US/US-TX/union-grove-21239" => array("type"=>"city", "pop"=>375, "name_ru"=>"Юнион Гров", "name_en"=>"Union Grove", "lang"=>"en", "lat"=>32.583100, "lon"=>-94.908464),
"/US/US-TX/universal-city-21240" => array("type"=>"city", "pop"=>18376, "name_en"=>"Universal City", "lang"=>"en", "lat"=>29.551650, "lon"=>-98.306239),
"/US/US-TX/university-park-21241" => array("type"=>"city", "pop"=>24853, "name_ru"=>"Юниверсити Парк", "name_en"=>"University Park", "lang"=>"en", "lat"=>32.850084, "lon"=>-96.793108),
"/US/US-TX/uvalde-21242" => array("type"=>"city", "pop"=>16142, "name_en"=>"Uvalde", "lang"=>"en", "lat"=>29.215459, "lon"=>-99.784859),
"/US/US-TX/valentine-21243" => array("type"=>"town", "pop"=>190, "name_en"=>"Valentine", "lang"=>"en", "lat"=>30.589134, "lon"=>-104.495031),
"/US/US-TX/valley-mills-21244" => array("type"=>"city", "pop"=>1121, "name_en"=>"Valley Mills", "lang"=>"en", "lat"=>31.658399, "lon"=>-97.471797),
"/US/US-TX/valley-view-21245" => array("type"=>"town", "pop"=>788, "name_en"=>"Valley View", "lang"=>"en", "lat"=>33.485382, "lon"=>-97.160808),
"/US/US-TX/van-21246" => array("type"=>"city", "pop"=>2579, "name_en"=>"Van", "lang"=>"en", "lat"=>32.523675, "lon"=>-95.637308),
"/US/US-TX/van-alstyne-21247" => array("type"=>"town", "pop"=>2957, "name_en"=>"Van Alstyne", "lang"=>"en", "lat"=>33.419762, "lon"=>-96.573492),
"/US/US-TX/van-horn-21248" => array("type"=>"town", "pop"=>2013, "name_en"=>"Van Horn", "lang"=>"en", "lat"=>31.040005, "lon"=>-104.834570),
"/US/US-TX/van-vleck-21249" => array("type"=>"cdp", "pop"=>1534, "name_en"=>"Van Vleck", "lang"=>"en", "lat"=>29.036101, "lon"=>-95.881806),
"/US/US-TX/vega-21250" => array("type"=>"town", "pop"=>871, "name_en"=>"Vega", "lang"=>"en", "lat"=>35.246579, "lon"=>-102.425931),
"/US/US-TX/venus-21251" => array("type"=>"town", "pop"=>2818, "name_en"=>"Venus", "lang"=>"en", "lat"=>32.427450, "lon"=>-97.104594),
"/US/US-TX/vernon-21252" => array("type"=>"city", "pop"=>10924, "name_ru"=>"Вернон", "name_en"=>"Vernon", "lang"=>"en", "lat"=>34.147438, "lon"=>-99.300090),
"/US/US-TX/victoria-21253" => array("type"=>"city", "pop"=>62558, "name_ru"=>"Виктория", "name_en"=>"Victoria", "lang"=>"en", "lat"=>28.824200, "lon"=>-96.982949),
"/US/US-TX/vidor-21254" => array("type"=>"city", "pop"=>11063, "name_en"=>"Vidor", "lang"=>"en", "lat"=>30.129281, "lon"=>-94.001571),
"/US/US-TX/vinton-21255" => array("type"=>"village", "pop"=>1939, "name_ru"=>"Винтон", "name_en"=>"Vinton", "lang"=>"en", "lat"=>31.961668, "lon"=>-106.595668),
"/US/US-TX/waco-21256" => array("type"=>"city", "pop"=>124009, "name_ru"=>"Вако", "name_en"=>"Waco", "lang"=>"en", "lat"=>31.568950, "lon"=>-97.183198),
"/US/US-TX/waelder-21257" => array("type"=>"city", "pop"=>992, "name_en"=>"Waelder", "lang"=>"en", "lat"=>29.693781, "lon"=>-97.296345),
"/US/US-TX/wake-village-21258" => array("type"=>"city", "pop"=>5628, "name_en"=>"Wake Village", "lang"=>"en", "lat"=>33.423328, "lon"=>-94.112406),
"/US/US-TX/waller-21259" => array("type"=>"city", "pop"=>2058, "name_en"=>"Waller", "lang"=>"en", "lat"=>30.057811, "lon"=>-95.923331),
"/US/US-TX/wallis-21260" => array("type"=>"city", "pop"=>1325, "name_en"=>"Wallis", "lang"=>"en", "lat"=>29.631304, "lon"=>-96.064214),
"/US/US-TX/walnut-springs-21261" => array("type"=>"city", "pop"=>793, "name_en"=>"Walnut Springs", "lang"=>"en", "lat"=>32.057036, "lon"=>-97.751478),
"/US/US-TX/warren-city-21262" => array("type"=>"city", "pop"=>346, "name_en"=>"Warren City", "lang"=>"en", "lat"=>32.553372, "lon"=>-94.902174),
"/US/US-TX/waskom-21263" => array("type"=>"city", "pop"=>2137, "name_en"=>"Waskom", "lang"=>"en", "lat"=>32.477932, "lon"=>-94.064814),
"/US/US-TX/watauga-21264" => array("type"=>"city", "pop"=>24005, "name_en"=>"Watauga", "lang"=>"en", "lat"=>32.871526, "lon"=>-97.250920),
"/US/US-TX/waxahachie-21265" => array("type"=>"city", "pop"=>28668, "name_en"=>"Waxahachie", "lang"=>"en", "lat"=>32.407186, "lon"=>-96.851995),
"/US/US-TX/weatherford-21266" => array("type"=>"city", "pop"=>26686, "name_en"=>"Weatherford", "lang"=>"en", "lat"=>32.755670, "lon"=>-97.771306),
"/US/US-TX/webster-21267" => array("type"=>"city", "pop"=>10665, "name_ru"=>"Вебстер", "name_en"=>"Webster", "lang"=>"en", "lat"=>29.532835, "lon"=>-95.116396),
"/US/US-TX/weimar-21268" => array("type"=>"city", "pop"=>2016, "name_en"=>"Weimar", "lang"=>"en", "lat"=>29.700787, "lon"=>-96.778324),
"/US/US-TX/weinert-21269" => array("type"=>"city", "pop"=>153, "name_en"=>"Weinert", "lang"=>"en", "lat"=>33.324326, "lon"=>-99.672898),
"/US/US-TX/weir-21270" => array("type"=>"town", "pop"=>648, "name_ru"=>"Вейр", "name_en"=>"Weir", "lang"=>"en", "lat"=>30.673273, "lon"=>-97.591163),
"/US/US-TX/wellington-21271" => array("type"=>"city", "pop"=>2090, "name_ru"=>"Веллингтон", "name_en"=>"Wellington", "lang"=>"en", "lat"=>34.853542, "lon"=>-100.213225),
"/US/US-TX/wellman-21272" => array("type"=>"town", "pop"=>198, "name_ru"=>"Уэллман", "name_en"=>"Wellman", "lang"=>"en", "lat"=>33.047014, "lon"=>-102.427557),
"/US/US-TX/wells-21273" => array("type"=>"town", "pop"=>798, "name_ru"=>"Уэллс", "name_en"=>"Wells", "lang"=>"en", "lat"=>31.491789, "lon"=>-94.946720),
"/US/US-TX/wells-branch-21274" => array("type"=>"cdp", "pop"=>7094, "name_en"=>"Wells Branch", "lang"=>"en", "lat"=>30.445753, "lon"=>-97.679056),
"/US/US-TX/weslaco-21275" => array("type"=>"city", "pop"=>33354, "name_en"=>"Weslaco", "lang"=>"en", "lat"=>26.161581, "lon"=>-97.984351),
"/US/US-TX/west-21276" => array("type"=>"city", "pop"=>2690, "name_ru"=>"Вэст", "name_en"=>"West", "lang"=>"en", "lat"=>31.803209, "lon"=>-97.092826),
"/US/US-TX/west-columbia-21278" => array("type"=>"city", "pop"=>4183, "name_en"=>"West Columbia", "lang"=>"en", "lat"=>29.141498, "lon"=>-95.648053),
"/US/US-TX/west-lake-hills-21280" => array("type"=>"city", "pop"=>3156, "name_en"=>"West Lake Hills", "lang"=>"en", "lat"=>30.292133, "lon"=>-97.809612),
"/US/US-TX/west-odessa-21282" => array("type"=>"cdp", "pop"=>16568, "name_en"=>"West Odessa", "lang"=>"en", "lat"=>31.842245, "lon"=>-102.498260),
"/US/US-TX/west-orange-21284" => array("type"=>"city", "pop"=>3832, "name_en"=>"West Orange", "lang"=>"en", "lat"=>30.078227, "lon"=>-93.759898),
"/US/US-TX/west-tawakoni-21286" => array("type"=>"town", "pop"=>1741, "name_en"=>"West Tawakoni", "lang"=>"en", "lat"=>32.901820, "lon"=>-96.020946),
"/US/US-TX/west-university-place-21287" => array("type"=>"city", "pop"=>15583, "name_en"=>"West University Place", "lang"=>"en", "lat"=>29.715200, "lon"=>-95.431349),
"/US/US-TX/westbrook-21277" => array("type"=>"city", "pop"=>193, "name_ru"=>"Вэстбрук", "name_en"=>"Westbrook", "lang"=>"en", "lat"=>32.356908, "lon"=>-101.012161),
"/US/US-TX/westlake-21279" => array("type"=>"town", "pop"=>211, "name_ru"=>"Вэстлейк", "name_en"=>"Westlake", "lang"=>"en", "lat"=>32.985048, "lon"=>-97.198039),
"/US/US-TX/westminster-21281" => array("type"=>"town", "pop"=>388, "name_ru"=>"Уэстминстер", "name_en"=>"Westminster", "lang"=>"en", "lat"=>33.358865, "lon"=>-96.457880),
"/US/US-TX/weston-21283" => array("type"=>"town", "pop"=>636, "name_ru"=>"Вэстон", "name_en"=>"Weston", "lang"=>"en", "lat"=>33.328153, "lon"=>-96.644084),
"/US/US-TX/westover-hills-21285" => array("type"=>"town", "pop"=>725, "name_en"=>"Westover Hills", "lang"=>"en", "lat"=>32.745290, "lon"=>-97.411515),
"/US/US-TX/westway-21288" => array("type"=>"cdp", "pop"=>2381, "name_en"=>"Westway", "lang"=>"en", "lat"=>31.958667, "lon"=>-106.577574),
"/US/US-TX/westworth-21289" => array("type"=>"village", "pop"=>2350, "name_en"=>"Westworth", "lang"=>"en", "lat"=>32.759967, "lon"=>-97.422225),
"/US/US-TX/wharton-21290" => array("type"=>"city", "pop"=>9150, "name_en"=>"Wharton", "lang"=>"en", "lat"=>29.316371, "lon"=>-96.099046),
"/US/US-TX/wheeler-21291" => array("type"=>"town", "pop"=>1218, "name_ru"=>"Уилер", "name_en"=>"Wheeler", "lang"=>"en", "lat"=>35.440259, "lon"=>-100.274984),
"/US/US-TX/white-deer-21292" => array("type"=>"town", "pop"=>1007, "name_en"=>"White Deer", "lang"=>"en", "lat"=>35.432446, "lon"=>-101.174836),
"/US/US-TX/white-oak-21295" => array("type"=>"city", "pop"=>6298, "name_ru"=>"Уайт Оук", "name_en"=>"White Oak", "lang"=>"en", "lat"=>32.532085, "lon"=>-94.855037),
"/US/US-TX/white-settlement-21297" => array("type"=>"city", "pop"=>16301, "name_en"=>"White Settlement", "lang"=>"en", "lat"=>32.754751, "lon"=>-97.460233),
"/US/US-TX/whiteface-21293" => array("type"=>"town", "pop"=>366, "name_en"=>"Whiteface", "lang"=>"en", "lat"=>33.599153, "lon"=>-102.612499),
"/US/US-TX/whitehouse-21294" => array("type"=>"city", "pop"=>7724, "name_en"=>"Whitehouse", "lang"=>"en", "lat"=>32.226026, "lon"=>-95.219575),
"/US/US-TX/whitesboro-21296" => array("type"=>"city", "pop"=>4029, "name_en"=>"Whitesboro", "lang"=>"en", "lat"=>33.659459, "lon"=>-96.903367),
"/US/US-TX/whitewright-21298" => array("type"=>"town", "pop"=>1697, "name_en"=>"Whitewright", "lang"=>"en", "lat"=>33.511053, "lon"=>-96.394697),
"/US/US-TX/whitney-21299" => array("type"=>"town", "pop"=>2052, "name_en"=>"Whitney", "lang"=>"en", "lat"=>31.949806, "lon"=>-97.320382),
"/US/US-TX/wichita-falls-21300" => array("type"=>"city", "pop"=>101202, "name_en"=>"Wichita Falls", "lang"=>"en", "lat"=>33.897933, "lon"=>-98.516280),
"/US/US-TX/wickett-21301" => array("type"=>"town", "pop"=>438, "name_en"=>"Wickett", "lang"=>"en", "lat"=>31.570101, "lon"=>-103.009425),
"/US/US-TX/wild-peach-village-21302" => array("type"=>"cdp", "pop"=>2440, "name_en"=>"Wild Peach Village", "lang"=>"en", "lat"=>29.083404, "lon"=>-95.633721),
"/US/US-TX/willis-21303" => array("type"=>"city", "pop"=>4290, "name_ru"=>"Уиллс", "name_en"=>"Willis", "lang"=>"en", "lat"=>30.424839, "lon"=>-95.478088),
"/US/US-TX/willow-park-21304" => array("type"=>"city", "pop"=>4523, "name_en"=>"Willow Park", "lang"=>"en", "lat"=>32.755806, "lon"=>-97.651277),
"/US/US-TX/wills-point-21305" => array("type"=>"city", "pop"=>3804, "name_en"=>"Wills Point", "lang"=>"en", "lat"=>32.709419, "lon"=>-96.006183),
"/US/US-TX/wilmer-21306" => array("type"=>"city", "pop"=>3576, "name_ru"=>"Уилмер", "name_en"=>"Wilmer", "lang"=>"en", "lat"=>32.598801, "lon"=>-96.684159),
"/US/US-TX/wilson-21307" => array("type"=>"city", "pop"=>460, "name_ru"=>"Уилсон", "name_en"=>"Wilson", "lang"=>"en", "lat"=>33.319777, "lon"=>-101.726685),
"/US/US-TX/wimberley-21308" => array("type"=>"cdp", "pop"=>2851, "name_en"=>"Wimberley", "lang"=>"en", "lat"=>29.992865, "lon"=>-98.101523),
"/US/US-TX/windcrest-21309" => array("type"=>"city", "pop"=>5357, "name_en"=>"Windcrest", "lang"=>"en", "lat"=>29.517667, "lon"=>-98.378899),
"/US/US-TX/windemere-21310" => array("type"=>"cdp", "pop"=>3207, "name_en"=>"Windemere", "lang"=>"en", "lat"=>30.458874, "lon"=>-97.648788),
"/US/US-TX/windom-21311" => array("type"=>"town", "pop"=>253, "name_ru"=>"Уиндом", "name_en"=>"Windom", "lang"=>"en", "lat"=>33.564398, "lon"=>-95.998548),
"/US/US-TX/windthorst-21312" => array("type"=>"town", "pop"=>448, "name_en"=>"Windthorst", "lang"=>"en", "lat"=>33.575533, "lon"=>-98.436739),
"/US/US-TX/winfield-21313" => array("type"=>"city", "pop"=>532, "name_ru"=>"Винфилд", "name_en"=>"Winfield", "lang"=>"en", "lat"=>33.166507, "lon"=>-95.109329),
"/US/US-TX/wink-21314" => array("type"=>"city", "pop"=>907, "name_en"=>"Wink", "lang"=>"en", "lat"=>31.755228, "lon"=>-103.154371),
"/US/US-TX/winnie-21315" => array("type"=>"cdp", "pop"=>2238, "name_en"=>"Winnie", "lang"=>"en", "lat"=>29.816663, "lon"=>-94.380172),
"/US/US-TX/winnsboro-21316" => array("type"=>"city", "pop"=>3929, "name_ru"=>"Уиннсборо", "name_en"=>"Winnsboro", "lang"=>"en", "lat"=>32.956223, "lon"=>-95.289836),
"/US/US-TX/winona-21317" => array("type"=>"town", "pop"=>620, "name_ru"=>"Уайнона", "name_en"=>"Winona", "lang"=>"en", "lat"=>32.490692, "lon"=>-95.170876),
"/US/US-TX/winters-21318" => array("type"=>"city", "pop"=>2569, "name_ru"=>"Уинтерс", "name_en"=>"Winters", "lang"=>"en", "lat"=>31.965540, "lon"=>-99.934343),
"/US/US-TX/wixon-valley-21319" => array("type"=>"city", "pop"=>235, "name_en"=>"Wixon Valley", "lang"=>"en", "lat"=>30.762621, "lon"=>-96.317202),
"/US/US-TX/wolfe-city-21320" => array("type"=>"city", "pop"=>1633, "name_en"=>"Wolfe City", "lang"=>"en", "lat"=>33.369764, "lon"=>-96.067685),
"/US/US-TX/wolfforth-21321" => array("type"=>"town", "pop"=>3499, "name_en"=>"Wolfforth", "lang"=>"en", "lat"=>33.508000, "lon"=>-102.010315),
"/US/US-TX/woodbranch-21322" => array("type"=>"village", "pop"=>1366, "name_en"=>"Woodbranch", "lang"=>"en", "lat"=>30.183979, "lon"=>-95.184258),
"/US/US-TX/woodcreek-21323" => array("type"=>"town", "pop"=>1604, "name_en"=>"Woodcreek", "lang"=>"en", "lat"=>30.031080, "lon"=>-98.112219),
"/US/US-TX/woodloch-21324" => array("type"=>"town", "pop"=>248, "name_en"=>"Woodloch", "lang"=>"en", "lat"=>30.217687, "lon"=>-95.413981),
"/US/US-TX/woodsboro-21325" => array("type"=>"town", "pop"=>1574, "name_ru"=>"Вудсборо", "name_en"=>"Woodsboro", "lang"=>"en", "lat"=>28.237541, "lon"=>-97.324798),
"/US/US-TX/woodson-21326" => array("type"=>"town", "pop"=>266, "name_ru"=>"Вудсон", "name_en"=>"Woodson", "lang"=>"en", "lat"=>33.014365, "lon"=>-99.053400),
"/US/US-TX/woodville-21327" => array("type"=>"town", "pop"=>2280, "name_ru"=>"Вудвилл", "name_en"=>"Woodville", "lang"=>"en", "lat"=>30.774143, "lon"=>-94.422405),
"/US/US-TX/woodway-21328" => array("type"=>"city", "pop"=>8814, "name_en"=>"Woodway", "lang"=>"en", "lat"=>31.496600, "lon"=>-97.232111),
"/US/US-TX/wortham-21329" => array("type"=>"town", "pop"=>1089, "name_en"=>"Wortham", "lang"=>"en", "lat"=>31.789955, "lon"=>-96.460840),
"/US/US-TX/wyldwood-21330" => array("type"=>"cdp", "pop"=>1764, "name_en"=>"Wyldwood", "lang"=>"en", "lat"=>30.129182, "lon"=>-97.472509),
"/US/US-TX/wylie-21331" => array("type"=>"city", "pop"=>37861, "name_en"=>"Wylie", "lang"=>"en", "lat"=>33.042559, "lon"=>-96.513878),
"/US/US-TX/yantis-21332" => array("type"=>"town", "pop"=>374, "name_en"=>"Yantis", "lang"=>"en", "lat"=>32.931898, "lon"=>-95.575028),
"/US/US-TX/yoakum-21333" => array("type"=>"city", "pop"=>5469, "name_en"=>"Yoakum", "lang"=>"en", "lat"=>29.292753, "lon"=>-97.146527),
"/US/US-TX/yorktown-21334" => array("type"=>"city", "pop"=>2157, "name_ru"=>"Йорктаун", "name_en"=>"Yorktown", "lang"=>"en", "lat"=>28.982905, "lon"=>-97.504145),
"/US/US-TX/zapata-21335" => array("type"=>"cdp", "pop"=>13847, "name_en"=>"Zapata", "lang"=>"en", "lat"=>26.894322, "lon"=>-99.262145),
"/US/US-TX/zavalla-21336" => array("type"=>"city", "pop"=>665, "name_en"=>"Zavalla", "lang"=>"en", "lat"=>31.156947, "lon"=>-94.417439),
"/US/US-UM" => array("type"=>"outlying area", "name_ru"=>"Внешние Малые Ислэндс", "name_en"=>"United States Minor Outlying Islands", "lang"=>"en"),
"/US/US-UT" => array("type"=>"state", "pop"=>2736424, "name_ru"=>"Юта", "name_en"=>"Utah", "lang"=>"en"),
"/US/US-UT/alpine-21337" => array("type"=>"city", "pop"=>9885, "name_ru"=>"Алпайн", "name_en"=>"Alpine", "lang"=>"en", "lat"=>40.460261, "lon"=>-111.774515),
"/US/US-UT/alta-21338" => array("type"=>"town", "pop"=>374, "name_ru"=>"Альта", "name_en"=>"Alta", "lang"=>"en", "lat"=>40.579124, "lon"=>-111.619140),
"/US/US-UT/altamont-21339" => array("type"=>"town", "pop"=>194, "name_ru"=>"Альтамонт", "name_en"=>"Altamont", "lang"=>"en", "lat"=>40.358588, "lon"=>-110.286274),
"/US/US-UT/alton-21340" => array("type"=>"town", "pop"=>140, "name_ru"=>"Альтон", "name_en"=>"Alton", "lang"=>"en", "lat"=>37.438789, "lon"=>-112.483199),
"/US/US-UT/amalga-21341" => array("type"=>"town", "pop"=>480, "name_en"=>"Amalga", "lang"=>"en", "lat"=>41.857513, "lon"=>-111.896782),
"/US/US-UT/american-fork-21342" => array("type"=>"city", "pop"=>27064, "name_en"=>"American Fork", "lang"=>"en", "lat"=>40.384250, "lon"=>-111.793873),
"/US/US-UT/annabella-21343" => array("type"=>"town", "pop"=>647, "name_en"=>"Annabella", "lang"=>"en", "lat"=>38.706197, "lon"=>-112.057502),
"/US/US-UT/antimony-21344" => array("type"=>"town", "pop"=>114, "name_en"=>"Antimony", "lang"=>"en", "lat"=>38.100891, "lon"=>-111.983579),
"/US/US-UT/aurora-21345" => array("type"=>"city", "pop"=>952, "name_ru"=>"Аврора", "name_en"=>"Aurora", "lang"=>"en", "lat"=>38.920364, "lon"=>-111.932426),
"/US/US-UT/ballard-21346" => array("type"=>"town", "pop"=>689, "name_en"=>"Ballard", "lang"=>"en", "lat"=>40.284328, "lon"=>-109.946958),
"/US/US-UT/bear-river-city-21347" => array("type"=>"town", "pop"=>833, "name_en"=>"Bear River City", "lang"=>"en", "lat"=>41.614423, "lon"=>-112.123150),
"/US/US-UT/beaver-21348" => array("type"=>"city", "pop"=>2597, "name_ru"=>"Бивер", "name_en"=>"Beaver", "lang"=>"en", "lat"=>38.279002, "lon"=>-112.639769),
"/US/US-UT/bicknell-21349" => array("type"=>"town", "pop"=>347, "name_ru"=>"Бикнелл", "name_en"=>"Bicknell", "lang"=>"en", "lat"=>38.341404, "lon"=>-111.544727),
"/US/US-UT/big-water-21350" => array("type"=>"town", "pop"=>406, "name_en"=>"Big Water", "lang"=>"en", "lat"=>37.068512, "lon"=>-111.661588),
"/US/US-UT/blanding-21351" => array("type"=>"city", "pop"=>3290, "name_en"=>"Blanding", "lang"=>"en", "lat"=>37.624671, "lon"=>-109.480093),
"/US/US-UT/bluffdale-21352" => array("type"=>"city", "pop"=>8016, "name_en"=>"Bluffdale", "lang"=>"en", "lat"=>40.476033, "lon"=>-111.955468),
"/US/US-UT/boulder-21353" => array("type"=>"town", "pop"=>186, "name_ru"=>"Боулдер", "name_en"=>"Boulder", "lang"=>"en", "lat"=>37.929041, "lon"=>-111.427776),
"/US/US-UT/bountiful-21354" => array("type"=>"city", "pop"=>44473, "name_en"=>"Bountiful", "lang"=>"en", "lat"=>40.876069, "lon"=>-111.865214),
"/US/US-UT/brian-head-21355" => array("type"=>"town", "pop"=>127, "name_en"=>"Brian Head", "lang"=>"en", "lat"=>37.698353, "lon"=>-112.841743),
"/US/US-UT/brigham-city-21356" => array("type"=>"city", "pop"=>18709, "name_en"=>"Brigham City", "lang"=>"en", "lat"=>41.509064, "lon"=>-112.009589),
"/US/US-UT/cannonville-21357" => array("type"=>"town", "pop"=>138, "name_en"=>"Cannonville", "lang"=>"en", "lat"=>37.566099, "lon"=>-112.054132),
"/US/US-UT/canyon-rim-21358" => array("type"=>"cdp", "pop"=>10527, "name_en"=>"Canyon Rim", "lang"=>"en", "lat"=>40.706650, "lon"=>-111.821151),
"/US/US-UT/castle-dale-21359" => array("type"=>"city", "pop"=>1582, "name_en"=>"Castle Dale", "lang"=>"en", "lat"=>39.221363, "lon"=>-111.019762),
"/US/US-UT/castle-valley-21360" => array("type"=>"town", "pop"=>386, "name_en"=>"Castle Valley", "lang"=>"en", "lat"=>38.632597, "lon"=>-109.399413),
"/US/US-UT/cedar-city-21361" => array("type"=>"city", "pop"=>28667, "name_en"=>"Cedar City", "lang"=>"en", "lat"=>37.689755, "lon"=>-113.073541),
"/US/US-UT/cedar-fort-21362" => array("type"=>"town", "pop"=>410, "name_en"=>"Cedar Fort", "lang"=>"en", "lat"=>40.331039, "lon"=>-112.096217),
"/US/US-UT/cedar-hills-21363" => array("type"=>"town", "pop"=>9551, "name_en"=>"Cedar Hills", "lang"=>"en", "lat"=>40.404661, "lon"=>-111.750786),
"/US/US-UT/centerfield-21364" => array("type"=>"town", "pop"=>1096, "name_en"=>"Centerfield", "lang"=>"en", "lat"=>39.127632, "lon"=>-111.817177),
"/US/US-UT/centerville-21365" => array("type"=>"city", "pop"=>15720, "name_ru"=>"Сентервилл", "name_en"=>"Centerville", "lang"=>"en", "lat"=>40.926600, "lon"=>-111.886080),
"/US/US-UT/charleston-21366" => array("type"=>"town", "pop"=>457, "name_ru"=>"Чарльстон", "name_en"=>"Charleston", "lang"=>"en", "lat"=>40.465487, "lon"=>-111.467931),
"/US/US-UT/circleville-21367" => array("type"=>"town", "pop"=>485, "name_ru"=>"Серклвилл", "name_en"=>"Circleville", "lang"=>"en", "lat"=>38.163443, "lon"=>-112.261849),
"/US/US-UT/clarkston-21368" => array("type"=>"town", "pop"=>754, "name_ru"=>"Кларкстон", "name_en"=>"Clarkston", "lang"=>"en", "lat"=>41.920978, "lon"=>-112.048417),
"/US/US-UT/clawson-21369" => array("type"=>"town", "pop"=>173, "name_ru"=>"Клосон", "name_en"=>"Clawson", "lang"=>"en", "lat"=>39.132854, "lon"=>-111.097161),
"/US/US-UT/clearfield-21370" => array("type"=>"city", "pop"=>27851, "name_ru"=>"Клирфилд", "name_en"=>"Clearfield", "lang"=>"en", "lat"=>41.104462, "lon"=>-112.022612),
"/US/US-UT/cleveland-21371" => array("type"=>"town", "pop"=>508, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>39.348958, "lon"=>-110.856522),
"/US/US-UT/clinton-21372" => array("type"=>"city", "pop"=>19855, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>41.141000, "lon"=>-112.062618),
"/US/US-UT/coalville-21373" => array("type"=>"city", "pop"=>1327, "name_en"=>"Coalville", "lang"=>"en", "lat"=>40.917648, "lon"=>-111.394416),
"/US/US-UT/corinne-21374" => array("type"=>"city", "pop"=>677, "name_en"=>"Corinne", "lang"=>"en", "lat"=>41.547736, "lon"=>-112.116566),
"/US/US-UT/cornish-21375" => array("type"=>"town", "pop"=>285, "name_en"=>"Cornish", "lang"=>"en", "lat"=>41.973122, "lon"=>-111.950985),
"/US/US-UT/cottonwood-heights-21376" => array("type"=>"cdp", "pop"=>35418, "name_en"=>"Cottonwood Heights", "lang"=>"en", "lat"=>40.610150, "lon"=>-111.809586),
"/US/US-UT/cottonwood-west-21377" => array("type"=>"cdp", "pop"=>17476, "name_en"=>"Cottonwood West", "lang"=>"en", "lat"=>40.646500, "lon"=>-111.850309),
"/US/US-UT/delta-21378" => array("type"=>"city", "pop"=>3172, "name_ru"=>"Дельта", "name_en"=>"Delta", "lang"=>"en", "lat"=>39.353294, "lon"=>-112.565664),
"/US/US-UT/deweyville-21379" => array("type"=>"town", "pop"=>334, "name_en"=>"Deweyville", "lang"=>"en", "lat"=>41.689992, "lon"=>-112.089793),
"/US/US-UT/draper-21380" => array("type"=>"city", "pop"=>42317, "name_en"=>"Draper", "lang"=>"en", "lat"=>40.499383, "lon"=>-111.863729),
"/US/US-UT/duchesne-21381" => array("type"=>"city", "pop"=>1612, "name_en"=>"Duchesne", "lang"=>"en", "lat"=>40.175429, "lon"=>-110.392467),
"/US/US-UT/dugway-21382" => array("type"=>"cdp", "pop"=>1761, "name_en"=>"Dugway", "lang"=>"en", "lat"=>40.231234, "lon"=>-112.750511),
"/US/US-UT/east-carbon-21383" => array("type"=>"city", "pop"=>1258, "name_en"=>"East Carbon", "lang"=>"en", "lat"=>39.535671, "lon"=>-110.406638),
"/US/US-UT/east-millcreek-21384" => array("type"=>"cdp", "pop"=>21184, "name_en"=>"East Millcreek", "lang"=>"en", "lat"=>40.688800, "lon"=>-111.820499),
"/US/US-UT/elk-ridge-21385" => array("type"=>"town", "pop"=>2476, "name_en"=>"Elk Ridge", "lang"=>"en", "lat"=>40.006620, "lon"=>-111.678870),
"/US/US-UT/elmo-21386" => array("type"=>"town", "pop"=>363, "name_ru"=>"Элмо", "name_en"=>"Elmo", "lang"=>"en", "lat"=>39.388789, "lon"=>-110.815485),
"/US/US-UT/elsinore-21387" => array("type"=>"town", "pop"=>741, "name_en"=>"Elsinore", "lang"=>"en", "lat"=>38.683872, "lon"=>-112.148861),
"/US/US-UT/elwood-21388" => array("type"=>"town", "pop"=>877, "name_ru"=>"Элвуд", "name_en"=>"Elwood", "lang"=>"en", "lat"=>41.678795, "lon"=>-112.139260),
"/US/US-UT/emery-21389" => array("type"=>"town", "pop"=>295, "name_en"=>"Emery", "lang"=>"en", "lat"=>38.925266, "lon"=>-111.250565),
"/US/US-UT/enoch-21390" => array("type"=>"city", "pop"=>5085, "name_en"=>"Enoch", "lang"=>"en", "lat"=>37.762160, "lon"=>-113.041991),
"/US/US-UT/enterprise-21391" => array("type"=>"city", "pop"=>1636, "name_ru"=>"Энтерпрайз", "name_en"=>"Enterprise", "lang"=>"en", "lat"=>37.570361, "lon"=>-113.714635),
"/US/US-UT/ephraim-21392" => array("type"=>"city", "pop"=>5284, "name_en"=>"Ephraim", "lang"=>"en", "lat"=>39.359716, "lon"=>-111.584357),
"/US/US-UT/erda-21393" => array("type"=>"cdp", "pop"=>1113, "name_en"=>"Erda", "lang"=>"en", "lat"=>40.590727, "lon"=>-112.321811),
"/US/US-UT/escalante-21394" => array("type"=>"town", "pop"=>763, "name_en"=>"Escalante", "lang"=>"en", "lat"=>37.764882, "lon"=>-111.600773),
"/US/US-UT/eureka-21395" => array("type"=>"city", "pop"=>796, "name_ru"=>"Юрика", "name_en"=>"Eureka", "lang"=>"en", "lat"=>39.955025, "lon"=>-112.115585),
"/US/US-UT/fairview-21396" => array("type"=>"city", "pop"=>1210, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>39.630472, "lon"=>-111.435442),
"/US/US-UT/farmington-21397" => array("type"=>"city", "pop"=>17217, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>40.988700, "lon"=>-111.894890),
"/US/US-UT/farr-west-21398" => array("type"=>"city", "pop"=>5335, "name_en"=>"Farr West", "lang"=>"en", "lat"=>41.298000, "lon"=>-112.030274),
"/US/US-UT/fayette-21399" => array("type"=>"town", "pop"=>211, "name_ru"=>"Файет", "name_en"=>"Fayette", "lang"=>"en", "lat"=>39.226109, "lon"=>-111.852796),
"/US/US-UT/ferron-21400" => array("type"=>"city", "pop"=>1544, "name_en"=>"Ferron", "lang"=>"en", "lat"=>39.090749, "lon"=>-111.132277),
"/US/US-UT/fielding-21401" => array("type"=>"town", "pop"=>422, "name_en"=>"Fielding", "lang"=>"en", "lat"=>41.812012, "lon"=>-112.116360),
"/US/US-UT/fillmore-21402" => array("type"=>"city", "pop"=>2136, "name_ru"=>"Фильмор", "name_en"=>"Fillmore", "lang"=>"en", "lat"=>38.967699, "lon"=>-112.336817),
"/US/US-UT/fort-duchesne-21403" => array("type"=>"cdp", "pop"=>655, "name_en"=>"Fort Duchesne", "lang"=>"en", "lat"=>40.278232, "lon"=>-109.870986),
"/US/US-UT/fountain-green-21404" => array("type"=>"city", "pop"=>978, "name_en"=>"Fountain Green", "lang"=>"en", "lat"=>39.627694, "lon"=>-111.638556),
"/US/US-UT/francis-21405" => array("type"=>"town", "pop"=>894, "name_en"=>"Francis", "lang"=>"en", "lat"=>40.612209, "lon"=>-111.275295),
"/US/US-UT/fruit-heights-21406" => array("type"=>"city", "pop"=>5312, "name_en"=>"Fruit Heights", "lang"=>"en", "lat"=>41.028300, "lon"=>-111.905664),
"/US/US-UT/garden-city-21407" => array("type"=>"town", "pop"=>460, "name_ru"=>"Гарден Сити", "name_en"=>"Garden City", "lang"=>"en", "lat"=>41.934112, "lon"=>-111.409692),
"/US/US-UT/garland-21408" => array("type"=>"city", "pop"=>2059, "name_ru"=>"Гарленд", "name_en"=>"Garland", "lang"=>"en", "lat"=>41.736804, "lon"=>-112.161178),
"/US/US-UT/genola-21409" => array("type"=>"town", "pop"=>1159, "name_ru"=>"Дженола", "name_en"=>"Genola", "lang"=>"en", "lat"=>40.017356, "lon"=>-111.852969),
"/US/US-UT/glendale-21410" => array("type"=>"town", "pop"=>339, "name_ru"=>"Глендейл", "name_en"=>"Glendale", "lang"=>"en", "lat"=>37.332064, "lon"=>-112.600653),
"/US/US-UT/glenwood-21411" => array("type"=>"town", "pop"=>438, "name_ru"=>"Гленвуд", "name_en"=>"Glenwood", "lang"=>"en", "lat"=>38.762373, "lon"=>-111.988597),
"/US/US-UT/goshen-21412" => array("type"=>"town", "pop"=>934, "name_ru"=>"Гошен", "name_en"=>"Goshen", "lang"=>"en", "lat"=>39.951803, "lon"=>-111.900884),
"/US/US-UT/granite-21413" => array("type"=>"cdp", "pop"=>3300, "name_en"=>"Granite", "lang"=>"en", "lat"=>40.567433, "lon"=>-111.805599),
"/US/US-UT/grantsville-21414" => array("type"=>"city", "pop"=>9049, "name_ru"=>"Грантсвилл", "name_en"=>"Grantsville", "lang"=>"en", "lat"=>40.596957, "lon"=>-112.467468),
"/US/US-UT/green-river-21415" => array("type"=>"city", "pop"=>923, "name_en"=>"Green River", "lang"=>"en", "lat"=>38.981554, "lon"=>-110.089879),
"/US/US-UT/gunnison-21416" => array("type"=>"city", "pop"=>3016, "name_ru"=>"Ганнисон", "name_en"=>"Gunnison", "lang"=>"en", "lat"=>39.158290, "lon"=>-111.812638),
"/US/US-UT/harrisville-21417" => array("type"=>"city", "pop"=>6060, "name_ru"=>"Харрисвилл", "name_en"=>"Harrisville", "lang"=>"en", "lat"=>41.280950, "lon"=>-111.984322),
"/US/US-UT/hatch-21418" => array("type"=>"town", "pop"=>118, "name_en"=>"Hatch", "lang"=>"en", "lat"=>37.652357, "lon"=>-112.432597),
"/US/US-UT/heber-21419" => array("type"=>"city", "pop"=>9830, "name_ru"=>"Хебер", "name_en"=>"Heber", "lang"=>"en", "lat"=>40.506840, "lon"=>-111.411523),
"/US/US-UT/helper-21420" => array("type"=>"city", "pop"=>1876, "name_en"=>"Helper", "lang"=>"en", "lat"=>39.689682, "lon"=>-110.859102),
"/US/US-UT/henefer-21421" => array("type"=>"town", "pop"=>680, "name_en"=>"Henefer", "lang"=>"en", "lat"=>41.019153, "lon"=>-111.492766),
"/US/US-UT/henrieville-21422" => array("type"=>"town", "pop"=>148, "name_en"=>"Henrieville", "lang"=>"en", "lat"=>37.564631, "lon"=>-111.994561),
"/US/US-UT/hiawatha-21423" => array("type"=>"town", "pop"=>43, "name_ru"=>"Хайавата", "name_en"=>"Hiawatha", "lang"=>"en", "lat"=>39.496208, "lon"=>-111.028974),
"/US/US-UT/highland-21424" => array("type"=>"city", "pop"=>16189, "name_ru"=>"Хайленд", "name_en"=>"Highland", "lang"=>"en", "lat"=>40.421560, "lon"=>-111.788755),
"/US/US-UT/hildale-21425" => array("type"=>"town", "pop"=>1970, "name_en"=>"Hildale", "lang"=>"en", "lat"=>37.008208, "lon"=>-112.968104),
"/US/US-UT/hinckley-21426" => array("type"=>"town", "pop"=>708, "name_ru"=>"Хинкли", "name_en"=>"Hinckley", "lang"=>"en", "lat"=>39.332390, "lon"=>-112.672772),
"/US/US-UT/holden-21427" => array("type"=>"town", "pop"=>372, "name_ru"=>"Холден", "name_en"=>"Holden", "lang"=>"en", "lat"=>39.099928, "lon"=>-112.268925),
"/US/US-UT/holladay-cottonwood-21428" => array("type"=>"cdp", "pop"=>14095, "name_en"=>"Holladay-Cottonwood", "lang"=>"en", "lat"=>40.648511, "lon"=>-111.809416),
"/US/US-UT/honeyville-21429" => array("type"=>"city", "pop"=>1354, "name_en"=>"Honeyville", "lang"=>"en", "lat"=>41.634784, "lon"=>-112.082790),
"/US/US-UT/hooper-21430" => array("type"=>"cdp", "pop"=>5655, "name_ru"=>"Хупер", "name_en"=>"Hooper", "lang"=>"en", "lat"=>41.172974, "lon"=>-112.124475),
"/US/US-UT/howell-21431" => array("type"=>"town", "pop"=>245, "name_ru"=>"Хауэлл", "name_en"=>"Howell", "lang"=>"en", "lat"=>41.771228, "lon"=>-112.444882),
"/US/US-UT/huntington-21432" => array("type"=>"city", "pop"=>2033, "name_ru"=>"Хантингтон", "name_en"=>"Huntington", "lang"=>"en", "lat"=>39.331065, "lon"=>-110.962909),
"/US/US-UT/huntsville-21433" => array("type"=>"town", "pop"=>653, "name_ru"=>"Хантсвилл", "name_en"=>"Huntsville", "lang"=>"en", "lat"=>41.259923, "lon"=>-111.772675),
"/US/US-UT/hurricane-21434" => array("type"=>"city", "pop"=>13321, "name_en"=>"Hurricane", "lang"=>"en", "lat"=>37.159901, "lon"=>-113.339189),
"/US/US-UT/hyde-park-21435" => array("type"=>"city", "pop"=>3927, "name_en"=>"Hyde Park", "lang"=>"en", "lat"=>41.797991, "lon"=>-111.814661),
"/US/US-UT/hyrum-21436" => array("type"=>"city", "pop"=>7636, "name_en"=>"Hyrum", "lang"=>"en", "lat"=>41.632927, "lon"=>-111.844662),
"/US/US-UT/ivins-21437" => array("type"=>"town", "pop"=>7870, "name_en"=>"Ivins", "lang"=>"en", "lat"=>37.174872, "lon"=>-113.684537),
"/US/US-UT/joseph-21438" => array("type"=>"town", "pop"=>273, "name_en"=>"Joseph", "lang"=>"en", "lat"=>38.625070, "lon"=>-112.218726),
"/US/US-UT/junction-21439" => array("type"=>"town", "pop"=>171, "name_ru"=>"Джанкшен", "name_en"=>"Junction", "lang"=>"en", "lat"=>38.238371, "lon"=>-112.224057),
"/US/US-UT/kamas-21440" => array("type"=>"city", "pop"=>1492, "name_en"=>"Kamas", "lang"=>"en", "lat"=>40.643309, "lon"=>-111.273746),
"/US/US-UT/kanab-21441" => array("type"=>"city", "pop"=>3782, "name_en"=>"Kanab", "lang"=>"en", "lat"=>37.020033, "lon"=>-112.516996),
"/US/US-UT/kanarraville-21442" => array("type"=>"town", "pop"=>315, "name_en"=>"Kanarraville", "lang"=>"en", "lat"=>37.537458, "lon"=>-113.179447),
"/US/US-UT/kanosh-21443" => array("type"=>"town", "pop"=>470, "name_en"=>"Kanosh", "lang"=>"en", "lat"=>38.804633, "lon"=>-112.438252),
"/US/US-UT/kaysville-21444" => array("type"=>"city", "pop"=>25820, "name_en"=>"Kaysville", "lang"=>"en", "lat"=>41.030650, "lon"=>-111.944323),
"/US/US-UT/kearns-21445" => array("type"=>"cdp", "pop"=>28374, "name_en"=>"Kearns", "lang"=>"en", "lat"=>40.652500, "lon"=>-112.008324),
"/US/US-UT/kingston-21446" => array("type"=>"town", "pop"=>136, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>38.204985, "lon"=>-112.179229),
"/US/US-UT/koosharem-21447" => array("type"=>"town", "pop"=>292, "name_en"=>"Koosharem", "lang"=>"en", "lat"=>38.511475, "lon"=>-111.880305),
"/US/US-UT/la-verkin-21449" => array("type"=>"city", "pop"=>4487, "name_en"=>"La Verkin", "lang"=>"en", "lat"=>37.206414, "lon"=>-113.272475),
"/US/US-UT/laketown-21448" => array("type"=>"town", "pop"=>191, "name_en"=>"Laketown", "lang"=>"en", "lat"=>41.822251, "lon"=>-111.317675),
"/US/US-UT/layton-21450" => array("type"=>"city", "pop"=>65514, "name_ru"=>"Лэйтон", "name_en"=>"Layton", "lang"=>"en", "lat"=>41.077976, "lon"=>-111.954554),
"/US/US-UT/leamington-21451" => array("type"=>"town", "pop"=>206, "name_en"=>"Leamington", "lang"=>"en", "lat"=>39.531038, "lon"=>-112.289288),
"/US/US-UT/leeds-21452" => array("type"=>"town", "pop"=>764, "name_ru"=>"Лидс", "name_en"=>"Leeds", "lang"=>"en", "lat"=>37.236962, "lon"=>-113.358045),
"/US/US-UT/lehi-21453" => array("type"=>"city", "pop"=>46802, "name_en"=>"Lehi", "lang"=>"en", "lat"=>40.402323, "lon"=>-111.854323),
"/US/US-UT/levan-21454" => array("type"=>"town", "pop"=>864, "name_en"=>"Levan", "lang"=>"en", "lat"=>39.556775, "lon"=>-111.860244),
"/US/US-UT/lewiston-21455" => array("type"=>"city", "pop"=>2030, "name_ru"=>"Льюистон", "name_en"=>"Lewiston", "lang"=>"en", "lat"=>41.960922, "lon"=>-111.874943),
"/US/US-UT/lindon-21456" => array("type"=>"city", "pop"=>10466, "name_en"=>"Lindon", "lang"=>"en", "lat"=>40.341267, "lon"=>-111.716597),
"/US/US-UT/little-cottonwood-creek-valley-21457" => array("type"=>"cdp", "pop"=>5042, "name_en"=>"Little Cottonwood Creek Valley", "lang"=>"en", "lat"=>40.604400, "lon"=>-111.828716),
"/US/US-UT/loa-21458" => array("type"=>"town", "pop"=>516, "name_en"=>"Loa", "lang"=>"en", "lat"=>38.402875, "lon"=>-111.643161),
"/US/US-UT/logan-21459" => array("type"=>"city", "pop"=>48657, "name_ru"=>"Логан", "name_en"=>"Logan", "lang"=>"en", "lat"=>41.740040, "lon"=>-111.835125),
"/US/US-UT/lyman-21460" => array("type"=>"town", "pop"=>230, "name_ru"=>"Лиман", "name_en"=>"Lyman", "lang"=>"en", "lat"=>38.396162, "lon"=>-111.588180),
"/US/US-UT/lynndyl-21461" => array("type"=>"town", "pop"=>120, "name_en"=>"Lynndyl", "lang"=>"en", "lat"=>39.508061, "lon"=>-112.389002),
"/US/US-UT/maeser-21462" => array("type"=>"cdp", "pop"=>2598, "name_en"=>"Maeser", "lang"=>"en", "lat"=>40.473552, "lon"=>-109.580908),
"/US/US-UT/magna-21463" => array("type"=>"cdp", "pop"=>17829, "name_en"=>"Magna", "lang"=>"en", "lat"=>40.704000, "lon"=>-112.086224),
"/US/US-UT/manila-21464" => array("type"=>"town", "pop"=>324, "name_ru"=>"Манила", "name_en"=>"Manila", "lang"=>"en", "lat"=>40.992753, "lon"=>-109.720433),
"/US/US-UT/manti-21465" => array("type"=>"city", "pop"=>3312, "name_en"=>"Manti", "lang"=>"en", "lat"=>39.266883, "lon"=>-111.636357),
"/US/US-UT/mantua-21466" => array("type"=>"town", "pop"=>756, "name_en"=>"Mantua", "lang"=>"en", "lat"=>41.502755, "lon"=>-111.932353),
"/US/US-UT/mapleton-21467" => array("type"=>"city", "pop"=>7954, "name_ru"=>"Мейплтон", "name_en"=>"Mapleton", "lang"=>"en", "lat"=>40.121711, "lon"=>-111.572975),
"/US/US-UT/marysvale-21468" => array("type"=>"town", "pop"=>352, "name_en"=>"Marysvale", "lang"=>"en", "lat"=>38.437652, "lon"=>-112.258099),
"/US/US-UT/mayfield-21469" => array("type"=>"town", "pop"=>440, "name_ru"=>"Мейфилд", "name_en"=>"Mayfield", "lang"=>"en", "lat"=>39.116164, "lon"=>-111.707339),
"/US/US-UT/meadow-21470" => array("type"=>"town", "pop"=>237, "name_en"=>"Meadow", "lang"=>"en", "lat"=>38.886799, "lon"=>-112.405926),
"/US/US-UT/mendon-21471" => array("type"=>"city", "pop"=>1190, "name_ru"=>"Мендон", "name_en"=>"Mendon", "lang"=>"en", "lat"=>41.711015, "lon"=>-111.979690),
"/US/US-UT/mexican-hat-21472" => array("type"=>"cdp", "pop"=>259, "name_en"=>"Mexican Hat", "lang"=>"en", "lat"=>37.128586, "lon"=>-109.915498),
"/US/US-UT/midvale-21473" => array("type"=>"city", "pop"=>28129, "name_ru"=>"Мидвейл", "name_en"=>"Midvale", "lang"=>"en", "lat"=>40.611799, "lon"=>-111.902252),
"/US/US-UT/midway-21474" => array("type"=>"city", "pop"=>3701, "name_ru"=>"Мидуэй", "name_en"=>"Midway", "lang"=>"en", "lat"=>40.514469, "lon"=>-111.474466),
"/US/US-UT/milford-21475" => array("type"=>"city", "pop"=>1399, "name_ru"=>"Милфорд", "name_en"=>"Milford", "lang"=>"en", "lat"=>38.394532, "lon"=>-113.011068),
"/US/US-UT/millcreek-21476" => array("type"=>"cdp", "pop"=>32230, "name_en"=>"Millcreek", "lang"=>"en", "lat"=>40.686883, "lon"=>-111.874722),
"/US/US-UT/millville-21477" => array("type"=>"city", "pop"=>1825, "name_ru"=>"Миллвилл", "name_en"=>"Millville", "lang"=>"en", "lat"=>41.680631, "lon"=>-111.818869),
"/US/US-UT/minersville-21478" => array("type"=>"town", "pop"=>817, "name_en"=>"Minersville", "lang"=>"en", "lat"=>38.213273, "lon"=>-112.924144),
"/US/US-UT/moab-21479" => array("type"=>"city", "pop"=>5121, "name_en"=>"Moab", "lang"=>"en", "lat"=>38.572528, "lon"=>-109.547100),
"/US/US-UT/mona-21480" => array("type"=>"town", "pop"=>1402, "name_en"=>"Mona", "lang"=>"en", "lat"=>39.815442, "lon"=>-111.857524),
"/US/US-UT/monroe-21481" => array("type"=>"city", "pop"=>1853, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>38.624256, "lon"=>-112.119542),
"/US/US-UT/montezuma-creek-21482" => array("type"=>"cdp", "pop"=>345, "name_en"=>"Montezuma Creek", "lang"=>"en", "lat"=>37.260110, "lon"=>-109.313310),
"/US/US-UT/monticello-21483" => array("type"=>"city", "pop"=>2018, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>37.871502, "lon"=>-109.335535),
"/US/US-UT/morgan-21484" => array("type"=>"city", "pop"=>3321, "name_ru"=>"Морган", "name_en"=>"Morgan", "lang"=>"en", "lat"=>41.042588, "lon"=>-111.682660),
"/US/US-UT/moroni-21485" => array("type"=>"city", "pop"=>1327, "name_en"=>"Moroni", "lang"=>"en", "lat"=>39.528105, "lon"=>-111.582268),
"/US/US-UT/mount-olympus-21486" => array("type"=>"cdp", "pop"=>7413, "name_en"=>"Mount Olympus", "lang"=>"en", "lat"=>40.685467, "lon"=>-111.787862),
"/US/US-UT/mount-pleasant-21487" => array("type"=>"city", "pop"=>2813, "name_ru"=>"Маунт Плезант", "name_en"=>"Mount Pleasant", "lang"=>"en", "lat"=>39.541723, "lon"=>-111.455603),
"/US/US-UT/murray-21488" => array("type"=>"city", "pop"=>46201, "name_ru"=>"Мюррей", "name_en"=>"Murray", "lang"=>"en", "lat"=>40.652450, "lon"=>-111.892461),
"/US/US-UT/myton-21489" => array("type"=>"city", "pop"=>597, "name_en"=>"Myton", "lang"=>"en", "lat"=>40.193489, "lon"=>-110.061794),
"/US/US-UT/naples-21490" => array("type"=>"city", "pop"=>1694, "name_ru"=>"Нэплс", "name_en"=>"Naples", "lang"=>"en", "lat"=>40.429846, "lon"=>-109.491003),
"/US/US-UT/neola-21491" => array("type"=>"cdp", "pop"=>511, "name_ru"=>"Неола", "name_en"=>"Neola", "lang"=>"en", "lat"=>40.440259, "lon"=>-110.036668),
"/US/US-UT/nephi-21492" => array("type"=>"city", "pop"=>5408, "name_en"=>"Nephi", "lang"=>"en", "lat"=>39.710801, "lon"=>-111.828788),
"/US/US-UT/new-harmony-21493" => array("type"=>"town", "pop"=>194, "name_ru"=>"Нью Хармони", "name_en"=>"New Harmony", "lang"=>"en", "lat"=>37.478429, "lon"=>-113.308078),
"/US/US-UT/newton-21494" => array("type"=>"town", "pop"=>803, "name_ru"=>"Ньютон", "name_en"=>"Newton", "lang"=>"en", "lat"=>41.861243, "lon"=>-111.989153),
"/US/US-UT/nibley-21495" => array("type"=>"city", "pop"=>4410, "name_en"=>"Nibley", "lang"=>"en", "lat"=>41.672514, "lon"=>-111.841887),
"/US/US-UT/north-logan-21496" => array("type"=>"city", "pop"=>8466, "name_en"=>"North Logan", "lang"=>"en", "lat"=>41.771546, "lon"=>-111.814127),
"/US/US-UT/north-ogden-21497" => array("type"=>"city", "pop"=>17682, "name_en"=>"North Ogden", "lang"=>"en", "lat"=>41.311400, "lon"=>-111.958949),
"/US/US-UT/north-salt-lake-21498" => array("type"=>"city", "pop"=>13446, "name_en"=>"North Salt Lake", "lang"=>"en", "lat"=>40.844350, "lon"=>-111.922779),
"/US/US-UT/oak-city-21499" => array("type"=>"town", "pop"=>606, "name_en"=>"Oak City", "lang"=>"en", "lat"=>39.376055, "lon"=>-112.336029),
"/US/US-UT/oakley-21500" => array("type"=>"town", "pop"=>1327, "name_ru"=>"Оукли", "name_en"=>"Oakley", "lang"=>"en", "lat"=>40.722052, "lon"=>-111.287583),
"/US/US-UT/ogden-21501" => array("type"=>"city", "pop"=>82865, "name_ru"=>"Огден", "name_en"=>"Ogden", "lang"=>"en", "lat"=>41.227900, "lon"=>-111.965818),
"/US/US-UT/ophir-21502" => array("type"=>"town", "pop"=>28, "name_ru"=>"Офир", "name_en"=>"Ophir", "lang"=>"en", "lat"=>40.369675, "lon"=>-112.254292),
"/US/US-UT/oquirrh-21503" => array("type"=>"cdp", "pop"=>7593, "name_en"=>"Oquirrh", "lang"=>"en", "lat"=>40.630676, "lon"=>-112.033162),
"/US/US-UT/orangeville-21504" => array("type"=>"city", "pop"=>1352, "name_ru"=>"Оранжвилл", "name_en"=>"Orangeville", "lang"=>"en", "lat"=>39.231325, "lon"=>-111.058922),
"/US/US-UT/orderville-21505" => array("type"=>"town", "pop"=>603, "name_en"=>"Orderville", "lang"=>"en", "lat"=>37.276154, "lon"=>-112.632558),
"/US/US-UT/orem-21506" => array("type"=>"city", "pop"=>93250, "name_en"=>"Orem", "lang"=>"en", "lat"=>40.298233, "lon"=>-111.698090),
"/US/US-UT/panguitch-21507" => array("type"=>"city", "pop"=>1520, "name_en"=>"Panguitch", "lang"=>"en", "lat"=>37.823675, "lon"=>-112.437664),
"/US/US-UT/paradise-21508" => array("type"=>"town", "pop"=>900, "name_ru"=>"Парадайс", "name_en"=>"Paradise", "lang"=>"en", "lat"=>41.567606, "lon"=>-111.833205),
"/US/US-UT/paragonah-21509" => array("type"=>"town", "pop"=>478, "name_en"=>"Paragonah", "lang"=>"en", "lat"=>37.886919, "lon"=>-112.771623),
"/US/US-UT/park-city-21510" => array("type"=>"city", "pop"=>7980, "name_ru"=>"Парк Сити", "name_en"=>"Park City", "lang"=>"en", "lat"=>40.656433, "lon"=>-111.492979),
"/US/US-UT/parowan-21511" => array("type"=>"city", "pop"=>2624, "name_en"=>"Parowan", "lang"=>"en", "lat"=>37.832232, "lon"=>-112.820145),
"/US/US-UT/payson-21512" => array("type"=>"city", "pop"=>17429, "name_ru"=>"Пейсон", "name_en"=>"Payson", "lang"=>"en", "lat"=>40.033904, "lon"=>-111.728057),
"/US/US-UT/perry-21513" => array("type"=>"city", "pop"=>3889, "name_ru"=>"Перри", "name_en"=>"Perry", "lang"=>"en", "lat"=>41.464853, "lon"=>-112.034340),
"/US/US-UT/plain-city-21514" => array("type"=>"city", "pop"=>5288, "name_en"=>"Plain City", "lang"=>"en", "lat"=>41.298883, "lon"=>-112.075132),
"/US/US-UT/pleasant-grove-21515" => array("type"=>"city", "pop"=>33798, "name_ru"=>"Плезант Гров", "name_en"=>"Pleasant Grove", "lang"=>"en", "lat"=>40.370400, "lon"=>-111.733324),
"/US/US-UT/pleasant-view-21516" => array("type"=>"city", "pop"=>7052, "name_en"=>"Pleasant View", "lang"=>"en", "lat"=>41.323700, "lon"=>-111.998685),
"/US/US-UT/plymouth-21517" => array("type"=>"town", "pop"=>364, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>41.876376, "lon"=>-112.144772),
"/US/US-UT/portage-21518" => array("type"=>"town", "pop"=>276, "name_ru"=>"Портейдж", "name_en"=>"Portage", "lang"=>"en", "lat"=>41.976564, "lon"=>-112.237788),
"/US/US-UT/price-21519" => array("type"=>"city", "pop"=>8039, "name_en"=>"Price", "lang"=>"en", "lat"=>39.601665, "lon"=>-110.801286),
"/US/US-UT/providence-21520" => array("type"=>"city", "pop"=>6538, "name_ru"=>"Провиденс", "name_en"=>"Providence", "lang"=>"en", "lat"=>41.705156, "lon"=>-111.813437),
"/US/US-UT/provo-21521" => array("type"=>"city", "pop"=>118581, "name_en"=>"Provo", "lang"=>"en", "lat"=>40.247150, "lon"=>-111.642674),
"/US/US-UT/randlett-21522" => array("type"=>"cdp", "pop"=>283, "name_en"=>"Randlett", "lang"=>"en", "lat"=>40.230578, "lon"=>-109.827660),
"/US/US-UT/randolph-21523" => array("type"=>"city", "pop"=>489, "name_ru"=>"Рэндолф", "name_en"=>"Randolph", "lang"=>"en", "lat"=>41.664455, "lon"=>-111.182709),
"/US/US-UT/redmond-21524" => array("type"=>"town", "pop"=>813, "name_en"=>"Redmond", "lang"=>"en", "lat"=>39.005321, "lon"=>-111.864494),
"/US/US-UT/richfield-21525" => array("type"=>"city", "pop"=>7217, "name_ru"=>"Ричфилд", "name_en"=>"Richfield", "lang"=>"en", "lat"=>38.765971, "lon"=>-112.090342),
"/US/US-UT/richmond-21526" => array("type"=>"city", "pop"=>2364, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>41.919410, "lon"=>-111.809390),
"/US/US-UT/river-heights-21528" => array("type"=>"city", "pop"=>1707, "name_en"=>"River Heights", "lang"=>"en", "lat"=>41.722542, "lon"=>-111.820373),
"/US/US-UT/riverdale-21527" => array("type"=>"city", "pop"=>8126, "name_ru"=>"Ривердейл", "name_en"=>"Riverdale", "lang"=>"en", "lat"=>41.171150, "lon"=>-112.002227),
"/US/US-UT/riverton-21529" => array("type"=>"city", "pop"=>39751, "name_ru"=>"Ривертон", "name_en"=>"Riverton", "lang"=>"en", "lat"=>40.519733, "lon"=>-111.945197),
"/US/US-UT/rockville-21530" => array("type"=>"town", "pop"=>264, "name_ru"=>"Роквилл", "name_en"=>"Rockville", "lang"=>"en", "lat"=>37.153598, "lon"=>-113.059380),
"/US/US-UT/roosevelt-21531" => array("type"=>"city", "pop"=>5025, "name_ru"=>"Рузвельт", "name_en"=>"Roosevelt", "lang"=>"en", "lat"=>40.295539, "lon"=>-110.000003),
"/US/US-UT/roy-21532" => array("type"=>"city", "pop"=>35672, "name_en"=>"Roy", "lang"=>"en", "lat"=>41.172600, "lon"=>-112.044974),
"/US/US-UT/rush-valley-21533" => array("type"=>"town", "pop"=>602, "name_en"=>"Rush Valley", "lang"=>"en", "lat"=>40.362978, "lon"=>-112.449586),
"/US/US-UT/salem-21535" => array("type"=>"city", "pop"=>6435, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>40.052275, "lon"=>-111.671643),
"/US/US-UT/salina-21536" => array("type"=>"city", "pop"=>2414, "name_ru"=>"Салина", "name_en"=>"Salina", "lang"=>"en", "lat"=>38.949103, "lon"=>-111.859246),
"/US/US-UT/salt-lake-city-21537" => array("type"=>"city", "pop"=>181698, "name_en"=>"Salt Lake City", "lang"=>"en", "lat"=>40.777267, "lon"=>-111.929921),
"/US/US-UT/sandy-21538" => array("type"=>"city", "pop"=>96660, "name_en"=>"Sandy", "lang"=>"en", "lat"=>40.569684, "lon"=>-111.854340),
"/US/US-UT/santa-clara-21539" => array("type"=>"city", "pop"=>6866, "name_ru"=>"Санта Клара", "name_en"=>"Santa Clara", "lang"=>"en", "lat"=>37.133901, "lon"=>-113.651302),
"/US/US-UT/santaquin-21540" => array("type"=>"city", "pop"=>8400, "name_en"=>"Santaquin", "lang"=>"en", "lat"=>39.974804, "lon"=>-111.783596),
"/US/US-UT/scipio-21541" => array("type"=>"town", "pop"=>298, "name_en"=>"Scipio", "lang"=>"en", "lat"=>39.247573, "lon"=>-112.101984),
"/US/US-UT/scofield-21542" => array("type"=>"town", "pop"=>26, "name_en"=>"Scofield", "lang"=>"en", "lat"=>39.725048, "lon"=>-111.161248),
"/US/US-UT/sigurd-21543" => array("type"=>"town", "pop"=>432, "name_en"=>"Sigurd", "lang"=>"en", "lat"=>38.850368, "lon"=>-111.965419),
"/US/US-UT/smithfield-21544" => array("type"=>"city", "pop"=>9535, "name_ru"=>"Смитфилд", "name_en"=>"Smithfield", "lang"=>"en", "lat"=>41.835364, "lon"=>-111.829672),
"/US/US-UT/snowville-21545" => array("type"=>"town", "pop"=>164, "name_en"=>"Snowville", "lang"=>"en", "lat"=>41.969764, "lon"=>-112.714743),
"/US/US-UT/south-jordan-21546" => array("type"=>"city", "pop"=>51131, "name_en"=>"South Jordan", "lang"=>"en", "lat"=>40.558767, "lon"=>-111.973208),
"/US/US-UT/south-ogden-21547" => array("type"=>"city", "pop"=>15891, "name_en"=>"South Ogden", "lang"=>"en", "lat"=>41.174284, "lon"=>-111.957976),
"/US/US-UT/south-salt-lake-21548" => array("type"=>"city", "pop"=>21607, "name_en"=>"South Salt Lake", "lang"=>"en", "lat"=>40.710200, "lon"=>-111.896799),
"/US/US-UT/south-weber-21549" => array("type"=>"city", "pop"=>6167, "name_en"=>"South Weber", "lang"=>"en", "lat"=>41.133940, "lon"=>-111.934921),
"/US/US-UT/spanish-fork-21550" => array("type"=>"city", "pop"=>31538, "name_en"=>"Spanish Fork", "lang"=>"en", "lat"=>40.114260, "lon"=>-111.639172),
"/US/US-UT/spring-city-21551" => array("type"=>"city", "pop"=>1044, "name_en"=>"Spring City", "lang"=>"en", "lat"=>39.480246, "lon"=>-111.490533),
"/US/US-UT/springdale-21552" => array("type"=>"town", "pop"=>579, "name_ru"=>"Спрингдейл", "name_en"=>"Springdale", "lang"=>"en", "lat"=>37.184032, "lon"=>-112.999144),
"/US/US-UT/springville-21553" => array("type"=>"city", "pop"=>28520, "name_ru"=>"Спрингвилл", "name_en"=>"Springville", "lang"=>"en", "lat"=>40.164850, "lon"=>-111.614493),
"/US/US-UT/st.-george-21534" => array("type"=>"city", "pop"=>72718, "name_ru"=>"Сент-Джрдж", "name_en"=>"St. George", "lang"=>"en", "lat"=>37.078472, "lon"=>-113.578586),
"/US/US-UT/stansbury-park-21554" => array("type"=>"cdp", "pop"=>1049, "name_en"=>"Stansbury Park", "lang"=>"en", "lat"=>40.637983, "lon"=>-112.303383),
"/US/US-UT/sterling-21555" => array("type"=>"town", "pop"=>274, "name_ru"=>"Стерлинг", "name_en"=>"Sterling", "lang"=>"en", "lat"=>39.194302, "lon"=>-111.690874),
"/US/US-UT/stockton-21556" => array("type"=>"town", "pop"=>580, "name_ru"=>"Стоктон", "name_en"=>"Stockton", "lang"=>"en", "lat"=>40.452721, "lon"=>-112.361955),
"/US/US-UT/sunnyside-21557" => array("type"=>"city", "pop"=>377, "name_ru"=>"Саннисайд", "name_en"=>"Sunnyside", "lang"=>"en", "lat"=>39.552160, "lon"=>-110.400125),
"/US/US-UT/sunset-21558" => array("type"=>"city", "pop"=>4945, "name_ru"=>"Сансет", "name_en"=>"Sunset", "lang"=>"en", "lat"=>41.138667, "lon"=>-112.027199),
"/US/US-UT/syracuse-21559" => array("type"=>"city", "pop"=>22195, "name_ru"=>"Сиракьюс", "name_en"=>"Syracuse", "lang"=>"en", "lat"=>41.096150, "lon"=>-112.057442),
"/US/US-UT/tabiona-21560" => array("type"=>"town", "pop"=>163, "name_en"=>"Tabiona", "lang"=>"en", "lat"=>40.353912, "lon"=>-110.708578),
"/US/US-UT/taylorsville-bennion-21561" => array("type"=>"cdp", "pop"=>52351, "name_en"=>"Taylorsville-Bennion", "lang"=>"en", "lat"=>40.655821, "lon"=>-111.944786),
"/US/US-UT/tooele-21562" => array("type"=>"city", "pop"=>30120, "name_en"=>"Tooele", "lang"=>"en", "lat"=>40.536457, "lon"=>-112.301388),
"/US/US-UT/toquerville-21563" => array("type"=>"town", "pop"=>1373, "name_en"=>"Toquerville", "lang"=>"en", "lat"=>37.236655, "lon"=>-113.285535),
"/US/US-UT/torrey-21564" => array("type"=>"town", "pop"=>191, "name_en"=>"Torrey", "lang"=>"en", "lat"=>38.301619, "lon"=>-111.419184),
"/US/US-UT/tremonton-21565" => array("type"=>"city", "pop"=>6789, "name_en"=>"Tremonton", "lang"=>"en", "lat"=>41.715270, "lon"=>-112.177617),
"/US/US-UT/trenton-21566" => array("type"=>"town", "pop"=>510, "name_ru"=>"Трентон", "name_en"=>"Trenton", "lang"=>"en", "lat"=>41.913932, "lon"=>-111.935168),
"/US/US-UT/tropic-21567" => array("type"=>"town", "pop"=>476, "name_en"=>"Tropic", "lang"=>"en", "lat"=>37.624642, "lon"=>-112.087698),
"/US/US-UT/uintah-21568" => array("type"=>"town", "pop"=>1258, "name_en"=>"Uintah", "lang"=>"en", "lat"=>41.142824, "lon"=>-111.933779),
"/US/US-UT/union-21569" => array("type"=>"cdp", "pop"=>13684, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>40.619103, "lon"=>-111.872570),
"/US/US-UT/val-verda-21570" => array("type"=>"cdp", "pop"=>3712, "name_en"=>"Val Verda", "lang"=>"en", "lat"=>40.853750, "lon"=>-111.888774),
"/US/US-UT/vernal-21571" => array("type"=>"city", "pop"=>8696, "name_en"=>"Vernal", "lang"=>"en", "lat"=>40.451984, "lon"=>-109.535538),
"/US/US-UT/vernon-21572" => array("type"=>"town", "pop"=>313, "name_ru"=>"Вернон", "name_en"=>"Vernon", "lang"=>"en", "lat"=>40.091255, "lon"=>-112.449783),
"/US/US-UT/vineyard-21573" => array("type"=>"town", "pop"=>148, "name_en"=>"Vineyard", "lang"=>"en", "lat"=>40.307450, "lon"=>-111.752077),
"/US/US-UT/virgin-21574" => array("type"=>"town", "pop"=>555, "name_en"=>"Virgin", "lang"=>"en", "lat"=>37.200830, "lon"=>-113.198524),
"/US/US-UT/wales-21575" => array("type"=>"town", "pop"=>234, "name_ru"=>"Уэльс", "name_en"=>"Wales", "lang"=>"en", "lat"=>39.486209, "lon"=>-111.635298),
"/US/US-UT/wallsburg-21576" => array("type"=>"town", "pop"=>305, "name_en"=>"Wallsburg", "lang"=>"en", "lat"=>40.386887, "lon"=>-111.420856),
"/US/US-UT/washington-21577" => array("type"=>"city", "pop"=>17716, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>37.134665, "lon"=>-113.493474),
"/US/US-UT/washington-terrace-21578" => array("type"=>"city", "pop"=>8515, "name_en"=>"Washington Terrace", "lang"=>"en", "lat"=>41.170050, "lon"=>-111.979507),
"/US/US-UT/wellington-21579" => array("type"=>"city", "pop"=>1571, "name_ru"=>"Веллингтон", "name_en"=>"Wellington", "lang"=>"en", "lat"=>39.537905, "lon"=>-110.737389),
"/US/US-UT/wellsville-21580" => array("type"=>"city", "pop"=>3259, "name_ru"=>"Уэллсвилл", "name_en"=>"Wellsville", "lang"=>"en", "lat"=>41.632277, "lon"=>-111.931789),
"/US/US-UT/wendover-21581" => array("type"=>"city", "pop"=>1632, "name_en"=>"Wendover", "lang"=>"en", "lat"=>40.735597, "lon"=>-114.035849),
"/US/US-UT/west-bountiful-21582" => array("type"=>"city", "pop"=>5337, "name_en"=>"West Bountiful", "lang"=>"en", "lat"=>40.900200, "lon"=>-111.902188),
"/US/US-UT/west-jordan-21583" => array("type"=>"city", "pop"=>104447, "name_en"=>"West Jordan", "lang"=>"en", "lat"=>40.599891, "lon"=>-111.994574),
"/US/US-UT/west-point-21584" => array("type"=>"city", "pop"=>9001, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>41.121524, "lon"=>-112.097071),
"/US/US-UT/west-valley-city-21585" => array("type"=>"city", "pop"=>123447, "name_en"=>"West Valley City", "lang"=>"en", "lat"=>40.690906, "lon"=>-112.008570),
"/US/US-UT/white-city-21586" => array("type"=>"cdp", "pop"=>6506, "name_ru"=>"Уайт Сити", "name_en"=>"White City", "lang"=>"en", "lat"=>40.567250, "lon"=>-111.860981),
"/US/US-UT/whiterocks-21587" => array("type"=>"cdp", "pop"=>312, "name_en"=>"Whiterocks", "lang"=>"en", "lat"=>40.472874, "lon"=>-109.943019),
"/US/US-UT/willard-21588" => array("type"=>"city", "pop"=>1747, "name_ru"=>"Уиллард", "name_en"=>"Willard", "lang"=>"en", "lat"=>41.411729, "lon"=>-112.044000),
"/US/US-UT/woodland-hills-21589" => array("type"=>"town", "pop"=>1325, "name_ru"=>"Вудленд Хилс", "name_en"=>"Woodland Hills", "lang"=>"en", "lat"=>40.014977, "lon"=>-111.649827),
"/US/US-UT/woodruff-21590" => array("type"=>"town", "pop"=>198, "name_en"=>"Woodruff", "lang"=>"en", "lat"=>41.522876, "lon"=>-111.163566),
"/US/US-UT/woods-cross-21591" => array("type"=>"city", "pop"=>8705, "name_en"=>"Woods Cross", "lang"=>"en", "lat"=>40.873603, "lon"=>-111.911864),
"/US/US-VA" => array("type"=>"state", "pop"=>7769089, "name_ru"=>"Вирджиния", "name_en"=>"Virginia", "lang"=>"en"),
"/US/US-VA/abingdon-21661" => array("type"=>"town", "pop"=>8009, "name_ru"=>"Абингдон", "name_en"=>"Abingdon", "lang"=>"en", "lat"=>36.708874, "lon"=>-81.969282),
"/US/US-VA/accomac-21662" => array("type"=>"town", "pop"=>524, "name_en"=>"Accomac", "lang"=>"en", "lat"=>37.718549, "lon"=>-75.667672),
"/US/US-VA/adwolf-21663" => array("type"=>"cdp", "pop"=>1292, "name_en"=>"Adwolf", "lang"=>"en", "lat"=>36.789098, "lon"=>-81.588451),
"/US/US-VA/alberta-21664" => array("type"=>"town", "pop"=>290, "name_ru"=>"Алберта", "name_en"=>"Alberta", "lang"=>"en", "lat"=>36.864633, "lon"=>-77.882993),
"/US/US-VA/alexandria-21665" => array("type"=>"city", "pop"=>143885, "name_ru"=>"Александрия", "name_en"=>"Alexandria", "lang"=>"en", "lat"=>38.821000, "lon"=>-77.086178),
"/US/US-VA/altavista-21666" => array("type"=>"town", "pop"=>3368, "name_en"=>"Altavista", "lang"=>"en", "lat"=>37.123547, "lon"=>-79.284480),
"/US/US-VA/amherst-21667" => array("type"=>"town", "pop"=>2219, "name_ru"=>"Амхерст", "name_en"=>"Amherst", "lang"=>"en", "lat"=>37.586089, "lon"=>-79.050434),
"/US/US-VA/annandale-21668" => array("type"=>"cdp", "pop"=>50975, "name_ru"=>"Аннандейл", "name_en"=>"Annandale", "lang"=>"en", "lat"=>38.833176, "lon"=>-77.212052),
"/US/US-VA/appalachia-21669" => array("type"=>"town", "pop"=>1739, "name_en"=>"Appalachia", "lang"=>"en", "lat"=>36.908841, "lon"=>-82.786954),
"/US/US-VA/appomattox-21670" => array("type"=>"town", "pop"=>1748, "name_en"=>"Appomattox", "lang"=>"en", "lat"=>37.358828, "lon"=>-78.826703),
"/US/US-VA/aquia-harbour-21671" => array("type"=>"cdp", "pop"=>6308, "name_en"=>"Aquia Harbour", "lang"=>"en", "lat"=>38.458898, "lon"=>-77.379886),
"/US/US-VA/arlington-21672" => array("type"=>"cdp", "pop"=>209969, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>38.878726, "lon"=>-77.101909),
"/US/US-VA/ashburn-21673" => array("type"=>"cdp", "pop"=>3393, "name_ru"=>"Эшберн", "name_en"=>"Ashburn", "lang"=>"en", "lat"=>39.041408, "lon"=>-77.481018),
"/US/US-VA/ashland-21674" => array("type"=>"town", "pop"=>7100, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>37.761600, "lon"=>-77.476378),
"/US/US-VA/atkins-21675" => array("type"=>"cdp", "pop"=>1130, "name_ru"=>"Аткинс", "name_en"=>"Atkins", "lang"=>"en", "lat"=>36.867490, "lon"=>-81.398629),
"/US/US-VA/bailey's-crossroads-21676" => array("type"=>"cdp", "pop"=>19507, "name_en"=>"Bailey's Crossroads", "lang"=>"en", "lat"=>38.848033, "lon"=>-77.129160),
"/US/US-VA/barracks-21677" => array("type"=>"cdp", "pop"=>4710, "name_en"=>"Barracks", "lang"=>"en", "lat"=>38.057404, "lon"=>-78.513523),
"/US/US-VA/bassett-21678" => array("type"=>"cdp", "pop"=>1579, "name_ru"=>"Бассет", "name_en"=>"Bassett", "lang"=>"en", "lat"=>36.763553, "lon"=>-79.985835),
"/US/US-VA/bedford-21679" => array("type"=>"city", "pop"=>6312, "name_ru"=>"Бедфорд", "name_en"=>"Bedford", "lang"=>"en", "lat"=>37.337818, "lon"=>-79.521114),
"/US/US-VA/belle-haven-21680" => array("type"=>"town", "pop"=>458, "name_en"=>"Belle Haven", "lang"=>"en", "lat"=>37.553924, "lon"=>-75.829145),
"/US/US-VA/belle-haven-21681" => array("type"=>"cdp", "pop"=>6427, "name_en"=>"Belle Haven", "lang"=>"en", "lat"=>38.777300, "lon"=>-77.057449),
"/US/US-VA/bellwood-21682" => array("type"=>"cdp", "pop"=>6178, "name_ru"=>"Беллвуд", "name_en"=>"Bellwood", "lang"=>"en", "lat"=>37.407167, "lon"=>-77.436824),
"/US/US-VA/bensley-21683" => array("type"=>"cdp", "pop"=>5093, "name_en"=>"Bensley", "lang"=>"en", "lat"=>37.446650, "lon"=>-77.443694),
"/US/US-VA/berryville-21684" => array("type"=>"town", "pop"=>3168, "name_ru"=>"Берривилл", "name_en"=>"Berryville", "lang"=>"en", "lat"=>39.149596, "lon"=>-77.981092),
"/US/US-VA/big-stone-gap-21685" => array("type"=>"town", "pop"=>5662, "name_en"=>"Big Stone Gap", "lang"=>"en", "lat"=>36.860026, "lon"=>-82.777917),
"/US/US-VA/blacksburg-21686" => array("type"=>"town", "pop"=>41796, "name_en"=>"Blacksburg", "lang"=>"en", "lat"=>37.232748, "lon"=>-80.428414),
"/US/US-VA/blackstone-21687" => array("type"=>"town", "pop"=>3582, "name_en"=>"Blackstone", "lang"=>"en", "lat"=>37.078747, "lon"=>-78.001704),
"/US/US-VA/bloxom-21688" => array("type"=>"town", "pop"=>384, "name_en"=>"Bloxom", "lang"=>"en", "lat"=>37.830831, "lon"=>-75.620159),
"/US/US-VA/blue-ridge-21690" => array("type"=>"cdp", "pop"=>2840, "name_ru"=>"Блу Ридж", "name_en"=>"Blue Ridge", "lang"=>"en", "lat"=>37.383472, "lon"=>-79.820148),
"/US/US-VA/bluefield-21689" => array("type"=>"town", "pop"=>5143, "name_en"=>"Bluefield", "lang"=>"en", "lat"=>37.236911, "lon"=>-81.266075),
"/US/US-VA/bon-air-21691" => array("type"=>"cdp", "pop"=>16413, "name_ru"=>"Бон Эр", "name_en"=>"Bon Air", "lang"=>"en", "lat"=>37.519800, "lon"=>-77.569058),
"/US/US-VA/boones-mill-21692" => array("type"=>"town", "pop"=>291, "name_en"=>"Boones Mill", "lang"=>"en", "lat"=>37.115206, "lon"=>-79.950709),
"/US/US-VA/bowling-green-21693" => array("type"=>"town", "pop"=>1021, "name_ru"=>"Боулинг Грин", "name_en"=>"Bowling Green", "lang"=>"en", "lat"=>38.049779, "lon"=>-77.351012),
"/US/US-VA/boyce-21694" => array("type"=>"town", "pop"=>451, "name_ru"=>"Бойс", "name_en"=>"Boyce", "lang"=>"en", "lat"=>39.093012, "lon"=>-78.059470),
"/US/US-VA/boydton-21695" => array("type"=>"town", "pop"=>458, "name_en"=>"Boydton", "lang"=>"en", "lat"=>36.665972, "lon"=>-78.389776),
"/US/US-VA/boykins-21696" => array("type"=>"town", "pop"=>601, "name_en"=>"Boykins", "lang"=>"en", "lat"=>36.577248, "lon"=>-77.198892),
"/US/US-VA/branchville-21697" => array("type"=>"town", "pop"=>120, "name_ru"=>"Бранчвилл", "name_en"=>"Branchville", "lang"=>"en", "lat"=>36.569115, "lon"=>-77.249338),
"/US/US-VA/bridgewater-21698" => array("type"=>"town", "pop"=>5430, "name_ru"=>"Бриджуотер", "name_en"=>"Bridgewater", "lang"=>"en", "lat"=>38.384917, "lon"=>-78.969048),
"/US/US-VA/bristol-21699" => array("type"=>"city", "pop"=>17424, "name_ru"=>"Бристол", "name_en"=>"Bristol", "lang"=>"en", "lat"=>36.613216, "lon"=>-82.168103),
"/US/US-VA/broadway-21700" => array("type"=>"town", "pop"=>3202, "name_en"=>"Broadway", "lang"=>"en", "lat"=>38.608243, "lon"=>-78.795936),
"/US/US-VA/brodnax-21701" => array("type"=>"town", "pop"=>290, "name_en"=>"Brodnax", "lang"=>"en", "lat"=>36.703481, "lon"=>-78.032939),
"/US/US-VA/brookneal-21702" => array("type"=>"town", "pop"=>1249, "name_en"=>"Brookneal", "lang"=>"en", "lat"=>37.053756, "lon"=>-78.948404),
"/US/US-VA/buchanan-21703" => array("type"=>"town", "pop"=>1219, "name_ru"=>"Бачанан", "name_en"=>"Buchanan", "lang"=>"en", "lat"=>37.520708, "lon"=>-79.689391),
"/US/US-VA/buena-vista-21704" => array("type"=>"city", "pop"=>6471, "name_ru"=>"Буена Виста", "name_en"=>"Buena Vista", "lang"=>"en", "lat"=>37.731693, "lon"=>-79.357099),
"/US/US-VA/bull-run-21705" => array("type"=>"cdp", "pop"=>5525, "name_en"=>"Bull Run", "lang"=>"en", "lat"=>38.783628, "lon"=>-77.520704),
"/US/US-VA/burke-21706" => array("type"=>"cdp", "pop"=>57734, "name_en"=>"Burke", "lang"=>"en", "lat"=>38.780239, "lon"=>-77.272058),
"/US/US-VA/burkeville-21707" => array("type"=>"town", "pop"=>472, "name_en"=>"Burkeville", "lang"=>"en", "lat"=>37.186979, "lon"=>-78.201983),
"/US/US-VA/cape-charles-21708" => array("type"=>"town", "pop"=>1464, "name_en"=>"Cape Charles", "lang"=>"en", "lat"=>37.270406, "lon"=>-76.014604),
"/US/US-VA/capron-21709" => array("type"=>"town", "pop"=>382, "name_ru"=>"Капрон", "name_en"=>"Capron", "lang"=>"en", "lat"=>36.710638, "lon"=>-77.202050),
"/US/US-VA/castlewood-21710" => array("type"=>"cdp", "pop"=>2110, "name_ru"=>"Каслвуд", "name_en"=>"Castlewood", "lang"=>"en", "lat"=>36.878315, "lon"=>-82.287169),
"/US/US-VA/cave-spring-21711" => array("type"=>"cdp", "pop"=>24053, "name_ru"=>"Кейв Спринг", "name_en"=>"Cave Spring", "lang"=>"en", "lat"=>37.223833, "lon"=>-80.007599),
"/US/US-VA/cedar-bluff-21712" => array("type"=>"town", "pop"=>1043, "name_ru"=>"Сидар Блафф", "name_en"=>"Cedar Bluff", "lang"=>"en", "lat"=>37.086820, "lon"=>-81.763858),
"/US/US-VA/centreville-21713" => array("type"=>"cdp", "pop"=>26585, "name_ru"=>"Сентревилл", "name_en"=>"Centreville", "lang"=>"en", "lat"=>38.846933, "lon"=>-77.443809),
"/US/US-VA/chamberlayne-21714" => array("type"=>"cdp", "pop"=>4577, "name_en"=>"Chamberlayne", "lang"=>"en", "lat"=>37.626400, "lon"=>-77.428752),
"/US/US-VA/chantilly-21715" => array("type"=>"cdp", "pop"=>29337, "name_en"=>"Chantilly", "lang"=>"en", "lat"=>38.875633, "lon"=>-77.408698),
"/US/US-VA/charlotte-court-house-21716" => array("type"=>"town", "pop"=>436, "name_en"=>"Charlotte Court House", "lang"=>"en", "lat"=>37.054237, "lon"=>-78.638455),
"/US/US-VA/charlottesville-21717" => array("type"=>"city", "pop"=>41487, "name_en"=>"Charlottesville", "lang"=>"en", "lat"=>38.037450, "lon"=>-78.485745),
"/US/US-VA/chase-city-21718" => array("type"=>"town", "pop"=>2321, "name_en"=>"Chase City", "lang"=>"en", "lat"=>36.799171, "lon"=>-78.461300),
"/US/US-VA/chatham-21719" => array("type"=>"town", "pop"=>1518, "name_ru"=>"Чатем", "name_en"=>"Chatham", "lang"=>"en", "lat"=>36.819346, "lon"=>-79.398323),
"/US/US-VA/chatmoss-laurel-park-21720" => array("type"=>"cdp", "pop"=>2194, "name_en"=>"Chatmoss-Laurel Park", "lang"=>"en", "lat"=>36.679526, "lon"=>-79.805886),
"/US/US-VA/cheriton-21721" => array("type"=>"town", "pop"=>469, "name_en"=>"Cheriton", "lang"=>"en", "lat"=>37.287977, "lon"=>-75.967622),
"/US/US-VA/chesapeake-21722" => array("type"=>"city", "pop"=>220111, "name_en"=>"Chesapeake", "lang"=>"en", "lat"=>36.678812, "lon"=>-76.308771),
"/US/US-VA/chester-21723" => array("type"=>"cdp", "pop"=>14986, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>37.353300, "lon"=>-77.436062),
"/US/US-VA/chilhowie-21724" => array("type"=>"town", "pop"=>1746, "name_en"=>"Chilhowie", "lang"=>"en", "lat"=>36.800099, "lon"=>-81.684008),
"/US/US-VA/chincoteague-21725" => array("type"=>"town", "pop"=>4296, "name_en"=>"Chincoteague", "lang"=>"en", "lat"=>37.949348, "lon"=>-75.355320),
"/US/US-VA/christiansburg-21726" => array("type"=>"town", "pop"=>19477, "name_en"=>"Christiansburg", "lang"=>"en", "lat"=>37.141655, "lon"=>-80.402697),
"/US/US-VA/claremont-21727" => array("type"=>"town", "pop"=>335, "name_ru"=>"Клэрмонт", "name_en"=>"Claremont", "lang"=>"en", "lat"=>37.227143, "lon"=>-76.965771),
"/US/US-VA/clarksville-21728" => array("type"=>"town", "pop"=>1258, "name_ru"=>"Кларксвилл", "name_en"=>"Clarksville", "lang"=>"en", "lat"=>36.620543, "lon"=>-78.565643),
"/US/US-VA/claypool-hill-21729" => array("type"=>"cdp", "pop"=>1468, "name_en"=>"Claypool Hill", "lang"=>"en", "lat"=>37.065069, "lon"=>-81.756028),
"/US/US-VA/cleveland-21730" => array("type"=>"town", "pop"=>141, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>36.943567, "lon"=>-82.152478),
"/US/US-VA/clifton-21731" => array("type"=>"town", "pop"=>210, "name_ru"=>"Слифтон", "name_en"=>"Clifton", "lang"=>"en", "lat"=>38.779888, "lon"=>-77.387682),
"/US/US-VA/clifton-forge-21732" => array("type"=>"city", "pop"=>3932, "name_en"=>"Clifton Forge", "lang"=>"en", "lat"=>37.823468, "lon"=>-79.825532),
"/US/US-VA/clinchport-21733" => array("type"=>"town", "pop"=>73, "name_en"=>"Clinchport", "lang"=>"en", "lat"=>36.679254, "lon"=>-82.744970),
"/US/US-VA/clintwood-21734" => array("type"=>"town", "pop"=>1520, "name_en"=>"Clintwood", "lang"=>"en", "lat"=>37.150624, "lon"=>-82.456239),
"/US/US-VA/clover-21735" => array("type"=>"town", "pop"=>198, "name_en"=>"Clover", "lang"=>"en", "lat"=>36.833886, "lon"=>-78.733887),
"/US/US-VA/cloverdale-21736" => array("type"=>"cdp", "pop"=>1689, "name_ru"=>"Кловердейл", "name_en"=>"Cloverdale", "lang"=>"en", "lat"=>37.356960, "lon"=>-79.906850),
"/US/US-VA/coeburn-21737" => array("type"=>"town", "pop"=>1978, "name_en"=>"Coeburn", "lang"=>"en", "lat"=>36.945722, "lon"=>-82.469331),
"/US/US-VA/collinsville-21738" => array("type"=>"cdp", "pop"=>7280, "name_ru"=>"Коллинсвилл", "name_en"=>"Collinsville", "lang"=>"en", "lat"=>36.724787, "lon"=>-79.911953),
"/US/US-VA/colonial-beach-21739" => array("type"=>"town", "pop"=>3789, "name_en"=>"Colonial Beach", "lang"=>"en", "lat"=>38.254801, "lon"=>-76.970693),
"/US/US-VA/colonial-heights-21740" => array("type"=>"city", "pop"=>17768, "name_en"=>"Colonial Heights", "lang"=>"en", "lat"=>37.264000, "lon"=>-77.398786),
"/US/US-VA/columbia-21741" => array("type"=>"town", "pop"=>56, "name_ru"=>"Колумбия", "name_en"=>"Columbia", "lang"=>"en", "lat"=>37.752988, "lon"=>-78.165091),
"/US/US-VA/commonwealth-21742" => array("type"=>"cdp", "pop"=>5538, "name_en"=>"Commonwealth", "lang"=>"en", "lat"=>38.074625, "lon"=>-78.487461),
"/US/US-VA/countryside-21743" => array("type"=>"cdp", "pop"=>8349, "name_ru"=>"Кантрисайд", "name_en"=>"Countryside", "lang"=>"en", "lat"=>39.052361, "lon"=>-77.417628),
"/US/US-VA/courtland-21744" => array("type"=>"town", "pop"=>1242, "name_ru"=>"Кортленд", "name_en"=>"Courtland", "lang"=>"en", "lat"=>36.711626, "lon"=>-77.062921),
"/US/US-VA/covington-21745" => array("type"=>"city", "pop"=>6120, "name_ru"=>"Ковингтон", "name_en"=>"Covington", "lang"=>"en", "lat"=>37.777966, "lon"=>-79.990271),
"/US/US-VA/craigsville-21746" => array("type"=>"town", "pop"=>1019, "name_en"=>"Craigsville", "lang"=>"en", "lat"=>38.081030, "lon"=>-79.384365),
"/US/US-VA/crewe-21747" => array("type"=>"town", "pop"=>2284, "name_en"=>"Crewe", "lang"=>"en", "lat"=>37.180717, "lon"=>-78.130927),
"/US/US-VA/crimora-21748" => array("type"=>"cdp", "pop"=>1752, "name_en"=>"Crimora", "lang"=>"en", "lat"=>38.160712, "lon"=>-78.839342),
"/US/US-VA/crozet-21749" => array("type"=>"cdp", "pop"=>2256, "name_en"=>"Crozet", "lang"=>"en", "lat"=>38.069309, "lon"=>-78.699036),
"/US/US-VA/culpeper-21750" => array("type"=>"town", "pop"=>13977, "name_en"=>"Culpeper", "lang"=>"en", "lat"=>38.471018, "lon"=>-78.001075),
"/US/US-VA/dale-city-21751" => array("type"=>"cdp", "pop"=>47170, "name_en"=>"Dale City", "lang"=>"en", "lat"=>38.650500, "lon"=>-77.344317),
"/US/US-VA/daleville-21752" => array("type"=>"cdp", "pop"=>1163, "name_ru"=>"Дейлвилл", "name_en"=>"Daleville", "lang"=>"en", "lat"=>37.415232, "lon"=>-79.921522),
"/US/US-VA/damascus-21753" => array("type"=>"town", "pop"=>1070, "name_ru"=>"Дамаскус", "name_en"=>"Damascus", "lang"=>"en", "lat"=>36.632779, "lon"=>-81.789349),
"/US/US-VA/danville-21754" => array("type"=>"city", "pop"=>44660, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>36.583187, "lon"=>-79.408300),
"/US/US-VA/dayton-21755" => array("type"=>"town", "pop"=>1353, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>38.415177, "lon"=>-78.942163),
"/US/US-VA/dendron-21756" => array("type"=>"town", "pop"=>293, "name_en"=>"Dendron", "lang"=>"en", "lat"=>37.034599, "lon"=>-76.924216),
"/US/US-VA/dillwyn-21757" => array("type"=>"town", "pop"=>433, "name_en"=>"Dillwyn", "lang"=>"en", "lat"=>37.540218, "lon"=>-78.461611),
"/US/US-VA/dooms-21758" => array("type"=>"cdp", "pop"=>1307, "name_en"=>"Dooms", "lang"=>"en", "lat"=>38.103474, "lon"=>-78.848351),
"/US/US-VA/drakes-branch-21759" => array("type"=>"town", "pop"=>472, "name_en"=>"Drakes Branch", "lang"=>"en", "lat"=>36.992752, "lon"=>-78.601036),
"/US/US-VA/dublin-21760" => array("type"=>"town", "pop"=>2190, "name_ru"=>"Дублин", "name_en"=>"Dublin", "lang"=>"en", "lat"=>37.105950, "lon"=>-80.685077),
"/US/US-VA/duffield-21761" => array("type"=>"town", "pop"=>82, "name_en"=>"Duffield", "lang"=>"en", "lat"=>36.719856, "lon"=>-82.797781),
"/US/US-VA/dumbarton-21762" => array("type"=>"cdp", "pop"=>8526, "name_en"=>"Dumbarton", "lang"=>"en", "lat"=>37.608745, "lon"=>-77.500499),
"/US/US-VA/dumfries-21763" => array("type"=>"town", "pop"=>4805, "name_en"=>"Dumfries", "lang"=>"en", "lat"=>38.567682, "lon"=>-77.325091),
"/US/US-VA/dungannon-21764" => array("type"=>"town", "pop"=>298, "name_en"=>"Dungannon", "lang"=>"en", "lat"=>36.828400, "lon"=>-82.467569),
"/US/US-VA/dunn-loring-21765" => array("type"=>"cdp", "pop"=>6509, "name_en"=>"Dunn Loring", "lang"=>"en", "lat"=>38.893250, "lon"=>-77.230290),
"/US/US-VA/east-highland-park-21766" => array("type"=>"cdp", "pop"=>11850, "name_en"=>"East Highland Park", "lang"=>"en", "lat"=>37.574800, "lon"=>-77.384484),
"/US/US-VA/eastville-21767" => array("type"=>"town", "pop"=>193, "name_en"=>"Eastville", "lang"=>"en", "lat"=>37.350824, "lon"=>-75.940403),
"/US/US-VA/edinburg-21768" => array("type"=>"town", "pop"=>897, "name_ru"=>"Эдинбург", "name_en"=>"Edinburg", "lang"=>"en", "lat"=>38.823375, "lon"=>-78.564881),
"/US/US-VA/elkton-21769" => array("type"=>"town", "pop"=>2626, "name_ru"=>"Элктон", "name_en"=>"Elkton", "lang"=>"en", "lat"=>38.413093, "lon"=>-78.617102),
"/US/US-VA/elliston-lafayette-21770" => array("type"=>"cdp", "pop"=>1243, "name_en"=>"Elliston-Lafayette", "lang"=>"en", "lat"=>37.223039, "lon"=>-80.215710),
"/US/US-VA/emory-meadow-view-21771" => array("type"=>"cdp", "pop"=>2248, "name_en"=>"Emory-Meadow View", "lang"=>"en", "lat"=>36.773158, "lon"=>-81.846443),
"/US/US-VA/emporia-21772" => array("type"=>"city", "pop"=>5643, "name_ru"=>"Эмпория", "name_en"=>"Emporia", "lang"=>"en", "lat"=>36.694793, "lon"=>-77.536704),
"/US/US-VA/ettrick-21773" => array("type"=>"cdp", "pop"=>5290, "name_en"=>"Ettrick", "lang"=>"en", "lat"=>37.240850, "lon"=>-77.428258),
"/US/US-VA/exmore-21774" => array("type"=>"town", "pop"=>1349, "name_en"=>"Exmore", "lang"=>"en", "lat"=>37.533704, "lon"=>-75.824106),
"/US/US-VA/fairfax-21775" => array("type"=>"city", "pop"=>23844, "name_ru"=>"Фэрфакс", "name_en"=>"Fairfax", "lang"=>"en", "lat"=>38.853650, "lon"=>-77.298227),
"/US/US-VA/fairlawn-21776" => array("type"=>"cdp", "pop"=>2399, "name_en"=>"Fairlawn", "lang"=>"en", "lat"=>37.151257, "lon"=>-80.559535),
"/US/US-VA/falls-church-21777" => array("type"=>"city", "pop"=>11169, "name_en"=>"Falls Church", "lang"=>"en", "lat"=>38.885050, "lon"=>-77.174558),
"/US/US-VA/falmouth-21778" => array("type"=>"cdp", "pop"=>3541, "name_ru"=>"Фолмут", "name_en"=>"Falmouth", "lang"=>"en", "lat"=>38.335287, "lon"=>-77.465503),
"/US/US-VA/farmville-21779" => array("type"=>"town", "pop"=>7396, "name_en"=>"Farmville", "lang"=>"en", "lat"=>37.297486, "lon"=>-78.395980),
"/US/US-VA/ferrum-21780" => array("type"=>"cdp", "pop"=>1514, "name_en"=>"Ferrum", "lang"=>"en", "lat"=>36.929326, "lon"=>-80.021849),
"/US/US-VA/fieldale-21781" => array("type"=>"cdp", "pop"=>1018, "name_en"=>"Fieldale", "lang"=>"en", "lat"=>36.703097, "lon"=>-79.943184),
"/US/US-VA/fincastle-21782" => array("type"=>"town", "pop"=>353, "name_ru"=>"Финкасл", "name_en"=>"Fincastle", "lang"=>"en", "lat"=>37.498567, "lon"=>-79.875462),
"/US/US-VA/fishersville-21783" => array("type"=>"cdp", "pop"=>3230, "name_en"=>"Fishersville", "lang"=>"en", "lat"=>38.098282, "lon"=>-78.969333),
"/US/US-VA/floyd-21784" => array("type"=>"town", "pop"=>436, "name_ru"=>"Флойд", "name_en"=>"Floyd", "lang"=>"en", "lat"=>36.911987, "lon"=>-80.318630),
"/US/US-VA/forest-21785" => array("type"=>"cdp", "pop"=>5624, "name_ru"=>"Форест", "name_en"=>"Forest", "lang"=>"en", "lat"=>37.373544, "lon"=>-79.278341),
"/US/US-VA/fort-belvoir-21786" => array("type"=>"cdp", "pop"=>8590, "name_en"=>"Fort Belvoir", "lang"=>"en", "lat"=>38.709710, "lon"=>-77.146988),
"/US/US-VA/fort-hunt-21787" => array("type"=>"cdp", "pop"=>12989, "name_en"=>"Fort Hunt", "lang"=>"en", "lat"=>38.732700, "lon"=>-77.058367),
"/US/US-VA/fort-lee-21788" => array("type"=>"cdp", "pop"=>6895, "name_en"=>"Fort Lee", "lang"=>"en", "lat"=>37.233424, "lon"=>-77.330278),
"/US/US-VA/franconia-21789" => array("type"=>"cdp", "pop"=>19882, "name_en"=>"Franconia", "lang"=>"en", "lat"=>38.763090, "lon"=>-77.149055),
"/US/US-VA/franklin-21790" => array("type"=>"city", "pop"=>8877, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>36.680804, "lon"=>-76.935967),
"/US/US-VA/fredericksburg-21791" => array("type"=>"city", "pop"=>22818, "name_ru"=>"Фредериксбург", "name_en"=>"Fredericksburg", "lang"=>"en", "lat"=>38.299252, "lon"=>-77.487129),
"/US/US-VA/fries-21792" => array("type"=>"town", "pop"=>552, "name_en"=>"Fries", "lang"=>"en", "lat"=>36.715412, "lon"=>-80.976046),
"/US/US-VA/front-royal-21793" => array("type"=>"town", "pop"=>14616, "name_en"=>"Front Royal", "lang"=>"en", "lat"=>38.924845, "lon"=>-78.184697),
"/US/US-VA/galax-21794" => array("type"=>"city", "pop"=>6808, "name_en"=>"Galax", "lang"=>"en", "lat"=>36.664502, "lon"=>-80.917721),
"/US/US-VA/gate-city-21795" => array("type"=>"town", "pop"=>2052, "name_en"=>"Gate City", "lang"=>"en", "lat"=>36.638570, "lon"=>-82.581238),
"/US/US-VA/glade-spring-21796" => array("type"=>"town", "pop"=>1520, "name_en"=>"Glade Spring", "lang"=>"en", "lat"=>36.790447, "lon"=>-81.772130),
"/US/US-VA/glasgow-21797" => array("type"=>"town", "pop"=>1007, "name_ru"=>"Гласго", "name_en"=>"Glasgow", "lang"=>"en", "lat"=>37.634628, "lon"=>-79.452922),
"/US/US-VA/glen-allen-21798" => array("type"=>"cdp", "pop"=>9010, "name_ru"=>"Глен Аллен", "name_en"=>"Glen Allen", "lang"=>"en", "lat"=>37.663984, "lon"=>-77.484485),
"/US/US-VA/glen-lyn-21799" => array("type"=>"town", "pop"=>163, "name_en"=>"Glen Lyn", "lang"=>"en", "lat"=>37.371633, "lon"=>-80.859185),
"/US/US-VA/gloucester-courthouse-21800" => array("type"=>"cdp", "pop"=>2118, "name_en"=>"Gloucester Courthouse", "lang"=>"en", "lat"=>37.409758, "lon"=>-76.526961),
"/US/US-VA/gloucester-point-21801" => array("type"=>"cdp", "pop"=>8509, "name_en"=>"Gloucester Point", "lang"=>"en", "lat"=>37.267683, "lon"=>-76.495535),
"/US/US-VA/gordonsville-21802" => array("type"=>"town", "pop"=>1697, "name_en"=>"Gordonsville", "lang"=>"en", "lat"=>38.136619, "lon"=>-78.188115),
"/US/US-VA/goshen-21803" => array("type"=>"town", "pop"=>394, "name_ru"=>"Гошен", "name_en"=>"Goshen", "lang"=>"en", "lat"=>37.990790, "lon"=>-79.506266),
"/US/US-VA/great-falls-21804" => array("type"=>"cdp", "pop"=>6945, "name_en"=>"Great Falls", "lang"=>"en", "lat"=>39.010160, "lon"=>-77.288543),
"/US/US-VA/gretna-21805" => array("type"=>"town", "pop"=>1192, "name_ru"=>"Гретна", "name_en"=>"Gretna", "lang"=>"en", "lat"=>36.953050, "lon"=>-79.363046),
"/US/US-VA/grottoes-21806" => array("type"=>"town", "pop"=>2190, "name_en"=>"Grottoes", "lang"=>"en", "lat"=>38.268091, "lon"=>-78.824610),
"/US/US-VA/groveton-21807" => array("type"=>"cdp", "pop"=>19997, "name_en"=>"Groveton", "lang"=>"en", "lat"=>38.764835, "lon"=>-77.095892),
"/US/US-VA/grundy-21808" => array("type"=>"town", "pop"=>961, "name_en"=>"Grundy", "lang"=>"en", "lat"=>37.276808, "lon"=>-82.100011),
"/US/US-VA/halifax-21809" => array("type"=>"town", "pop"=>1290, "name_en"=>"Halifax", "lang"=>"en", "lat"=>36.765412, "lon"=>-78.932368),
"/US/US-VA/hallwood-21810" => array("type"=>"town", "pop"=>275, "name_en"=>"Hallwood", "lang"=>"en", "lat"=>37.877797, "lon"=>-75.589465),
"/US/US-VA/hamilton-21811" => array("type"=>"town", "pop"=>770, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>39.135450, "lon"=>-77.665442),
"/US/US-VA/hampden-sydney-21812" => array("type"=>"cdp", "pop"=>1240, "name_en"=>"Hampden Sydney", "lang"=>"en", "lat"=>37.240808, "lon"=>-78.457870),
"/US/US-VA/hampton-21813" => array("type"=>"city", "pop"=>145494, "name_ru"=>"Хэмптон", "name_en"=>"Hampton", "lang"=>"en", "lat"=>37.049850, "lon"=>-76.294973),
"/US/US-VA/harrisonburg-21814" => array("type"=>"city", "pop"=>44015, "name_ru"=>"Харрисонбург", "name_en"=>"Harrisonburg", "lang"=>"en", "lat"=>38.436724, "lon"=>-78.873976),
"/US/US-VA/haymarket-21815" => array("type"=>"town", "pop"=>1226, "name_en"=>"Haymarket", "lang"=>"en", "lat"=>38.812100, "lon"=>-77.637524),
"/US/US-VA/haysi-21816" => array("type"=>"town", "pop"=>309, "name_en"=>"Haysi", "lang"=>"en", "lat"=>37.205123, "lon"=>-82.293088),
"/US/US-VA/herndon-21817" => array("type"=>"town", "pop"=>22045, "name_en"=>"Herndon", "lang"=>"en", "lat"=>38.969200, "lon"=>-77.387321),
"/US/US-VA/highland-springs-21818" => array("type"=>"cdp", "pop"=>13823, "name_en"=>"Highland Springs", "lang"=>"en", "lat"=>37.550800, "lon"=>-77.327248),
"/US/US-VA/hillsboro-21819" => array("type"=>"town", "pop"=>116, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>39.197949, "lon"=>-77.724547),
"/US/US-VA/hillsville-21820" => array("type"=>"town", "pop"=>2650, "name_en"=>"Hillsville", "lang"=>"en", "lat"=>36.766129, "lon"=>-80.736271),
"/US/US-VA/hollins-21821" => array("type"=>"cdp", "pop"=>13305, "name_en"=>"Hollins", "lang"=>"en", "lat"=>37.342900, "lon"=>-79.953454),
"/US/US-VA/hollymead-21822" => array("type"=>"cdp", "pop"=>2628, "name_en"=>"Hollymead", "lang"=>"en", "lat"=>38.121909, "lon"=>-78.439639),
"/US/US-VA/honaker-21823" => array("type"=>"town", "pop"=>1445, "name_en"=>"Honaker", "lang"=>"en", "lat"=>37.016237, "lon"=>-81.977211),
"/US/US-VA/hopewell-21824" => array("type"=>"city", "pop"=>23142, "name_ru"=>"Хоупуэлл", "name_en"=>"Hopewell", "lang"=>"en", "lat"=>37.290565, "lon"=>-77.298425),
"/US/US-VA/horse-pasture-21825" => array("type"=>"cdp", "pop"=>2224, "name_en"=>"Horse Pasture", "lang"=>"en", "lat"=>36.629589, "lon"=>-79.948663),
"/US/US-VA/huntington-21826" => array("type"=>"cdp", "pop"=>7489, "name_ru"=>"Хантингтон", "name_en"=>"Huntington", "lang"=>"en", "lat"=>38.793000, "lon"=>-77.071572),
"/US/US-VA/hurt-21827" => array("type"=>"town", "pop"=>1214, "name_en"=>"Hurt", "lang"=>"en", "lat"=>37.094799, "lon"=>-79.298698),
"/US/US-VA/hybla-valley-21828" => array("type"=>"cdp", "pop"=>15491, "name_en"=>"Hybla Valley", "lang"=>"en", "lat"=>38.748750, "lon"=>-77.076614),
"/US/US-VA/idylwood-21829" => array("type"=>"cdp", "pop"=>14710, "name_en"=>"Idylwood", "lang"=>"en", "lat"=>38.890950, "lon"=>-77.206874),
"/US/US-VA/independence-21830" => array("type"=>"town", "pop"=>895, "name_ru"=>"Индепенденс", "name_en"=>"Independence", "lang"=>"en", "lat"=>36.622837, "lon"=>-81.151010),
"/US/US-VA/iron-gate-21831" => array("type"=>"town", "pop"=>371, "name_en"=>"Iron Gate", "lang"=>"en", "lat"=>37.798803, "lon"=>-79.790585),
"/US/US-VA/irvington-21832" => array("type"=>"town", "pop"=>642, "name_ru"=>"Ирвингтон", "name_en"=>"Irvington", "lang"=>"en", "lat"=>37.662233, "lon"=>-76.420176),
"/US/US-VA/ivor-21833" => array("type"=>"town", "pop"=>310, "name_en"=>"Ivor", "lang"=>"en", "lat"=>36.904859, "lon"=>-76.899410),
"/US/US-VA/jarratt-21834" => array("type"=>"town", "pop"=>560, "name_en"=>"Jarratt", "lang"=>"en", "lat"=>36.815803, "lon"=>-77.470108),
"/US/US-VA/jefferson-21835" => array("type"=>"cdp", "pop"=>25782, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>38.864500, "lon"=>-77.188032),
"/US/US-VA/jolivue-21836" => array("type"=>"cdp", "pop"=>1092, "name_en"=>"Jolivue", "lang"=>"en", "lat"=>38.118950, "lon"=>-79.066587),
"/US/US-VA/jonesville-21837" => array("type"=>"town", "pop"=>967, "name_ru"=>"Джонсвилл", "name_en"=>"Jonesville", "lang"=>"en", "lat"=>36.688179, "lon"=>-83.115505),
"/US/US-VA/keller-21838" => array("type"=>"town", "pop"=>164, "name_en"=>"Keller", "lang"=>"en", "lat"=>37.620394, "lon"=>-75.764517),
"/US/US-VA/kenbridge-21839" => array("type"=>"town", "pop"=>1279, "name_en"=>"Kenbridge", "lang"=>"en", "lat"=>36.964103, "lon"=>-78.130616),
"/US/US-VA/keysville-21840" => array("type"=>"town", "pop"=>764, "name_ru"=>"Кисвилл", "name_en"=>"Keysville", "lang"=>"en", "lat"=>37.038871, "lon"=>-78.481520),
"/US/US-VA/kilmarnock-21841" => array("type"=>"town", "pop"=>1295, "name_en"=>"Kilmarnock", "lang"=>"en", "lat"=>37.714462, "lon"=>-76.382654),
"/US/US-VA/la-crosse-21842" => array("type"=>"town", "pop"=>589, "name_ru"=>"Ла Кросс", "name_en"=>"La Crosse", "lang"=>"en", "lat"=>36.696782, "lon"=>-78.093725),
"/US/US-VA/lake-barcroft-21843" => array("type"=>"cdp", "pop"=>8686, "name_en"=>"Lake Barcroft", "lang"=>"en", "lat"=>38.851250, "lon"=>-77.158246),
"/US/US-VA/lake-monticello-21844" => array("type"=>"cdp", "pop"=>2331, "name_en"=>"Lake Monticello", "lang"=>"en", "lat"=>37.923067, "lon"=>-78.335092),
"/US/US-VA/lake-ridge-21845" => array("type"=>"cdp", "pop"=>23862, "name_en"=>"Lake Ridge", "lang"=>"en", "lat"=>38.687650, "lon"=>-77.297974),
"/US/US-VA/lakeside-21846" => array("type"=>"cdp", "pop"=>12081, "name_ru"=>"Лэйксайд", "name_en"=>"Lakeside", "lang"=>"en", "lat"=>37.617483, "lon"=>-77.467930),
"/US/US-VA/laurel-21847" => array("type"=>"cdp", "pop"=>13011, "name_ru"=>"Лорел", "name_en"=>"Laurel", "lang"=>"en", "lat"=>37.633867, "lon"=>-77.506954),
"/US/US-VA/lawrenceville-21848" => array("type"=>"town", "pop"=>1333, "name_ru"=>"Лоренсвилл", "name_en"=>"Lawrenceville", "lang"=>"en", "lat"=>36.757586, "lon"=>-77.851964),
"/US/US-VA/laymantown-21849" => array("type"=>"cdp", "pop"=>1942, "name_en"=>"Laymantown", "lang"=>"en", "lat"=>37.365676, "lon"=>-79.857868),
"/US/US-VA/lebanon-21850" => array("type"=>"town", "pop"=>3196, "name_ru"=>"Лебанон", "name_en"=>"Lebanon", "lang"=>"en", "lat"=>36.899397, "lon"=>-82.078532),
"/US/US-VA/leesburg-21851" => array("type"=>"town", "pop"=>39641, "name_ru"=>"Лисберг", "name_en"=>"Leesburg", "lang"=>"en", "lat"=>39.104690, "lon"=>-77.556079),
"/US/US-VA/lexington-21852" => array("type"=>"city", "pop"=>7003, "name_ru"=>"Лексингтон", "name_en"=>"Lexington", "lang"=>"en", "lat"=>37.782292, "lon"=>-79.444621),
"/US/US-VA/lincolnia-21853" => array("type"=>"cdp", "pop"=>13041, "name_en"=>"Lincolnia", "lang"=>"en", "lat"=>38.824400, "lon"=>-77.154220),
"/US/US-VA/loch-lomond-21854" => array("type"=>"cdp", "pop"=>3292, "name_en"=>"Loch Lomond", "lang"=>"en", "lat"=>38.781450, "lon"=>-77.482750),
"/US/US-VA/lorton-21855" => array("type"=>"cdp", "pop"=>15385, "name_en"=>"Lorton", "lang"=>"en", "lat"=>38.700690, "lon"=>-77.240786),
"/US/US-VA/louisa-21856" => array("type"=>"town", "pop"=>1564, "name_ru"=>"Луиза", "name_en"=>"Louisa", "lang"=>"en", "lat"=>38.019244, "lon"=>-77.999457),
"/US/US-VA/lovettsville-21857" => array("type"=>"town", "pop"=>1274, "name_en"=>"Lovettsville", "lang"=>"en", "lat"=>39.274088, "lon"=>-77.639050),
"/US/US-VA/luray-21858" => array("type"=>"town", "pop"=>4843, "name_ru"=>"Лурей", "name_en"=>"Luray", "lang"=>"en", "lat"=>38.664713, "lon"=>-78.455277),
"/US/US-VA/lynchburg-21859" => array("type"=>"city", "pop"=>72596, "name_ru"=>"Линчбург", "name_en"=>"Lynchburg", "lang"=>"en", "lat"=>37.401217, "lon"=>-79.191434),
"/US/US-VA/madison-21862" => array("type"=>"town", "pop"=>212, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>38.377546, "lon"=>-78.259292),
"/US/US-VA/madison-heights-21863" => array("type"=>"cdp", "pop"=>11700, "name_ru"=>"Мэдисон Хайтс", "name_en"=>"Madison Heights", "lang"=>"en", "lat"=>37.439904, "lon"=>-79.104148),
"/US/US-VA/manassas-21864" => array("type"=>"city", "pop"=>35205, "name_ru"=>"Манассас", "name_en"=>"Manassas", "lang"=>"en", "lat"=>38.747400, "lon"=>-77.485372),
"/US/US-VA/manassas-park-21865" => array("type"=>"city", "pop"=>11319, "name_en"=>"Manassas Park", "lang"=>"en", "lat"=>38.773267, "lon"=>-77.452060),
"/US/US-VA/mantua-21866" => array("type"=>"cdp", "pop"=>6804, "name_en"=>"Mantua", "lang"=>"en", "lat"=>38.852600, "lon"=>-77.257995),
"/US/US-VA/marion-21867" => array("type"=>"town", "pop"=>6001, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>36.838728, "lon"=>-81.514047),
"/US/US-VA/martinsville-21868" => array("type"=>"city", "pop"=>14543, "name_ru"=>"Мартинсвилл", "name_en"=>"Martinsville", "lang"=>"en", "lat"=>36.683477, "lon"=>-79.863824),
"/US/US-VA/massanutten-21869" => array("type"=>"cdp", "pop"=>990, "name_en"=>"Massanutten", "lang"=>"en", "lat"=>38.409581, "lon"=>-78.737971),
"/US/US-VA/mckenney-21860" => array("type"=>"town", "pop"=>484, "name_en"=>"McKenney", "lang"=>"en", "lat"=>36.985640, "lon"=>-77.721735),
"/US/US-VA/mclean-21861" => array("type"=>"cdp", "pop"=>38168, "name_ru"=>"Мак-Лин", "name_en"=>"McLean", "lang"=>"en", "lat"=>38.938472, "lon"=>-77.177012),
"/US/US-VA/mechanicsville-21870" => array("type"=>"cdp", "pop"=>22027, "name_ru"=>"Меканиксвилл", "name_en"=>"Mechanicsville", "lang"=>"en", "lat"=>37.627304, "lon"=>-77.355958),
"/US/US-VA/melfa-21871" => array("type"=>"town", "pop"=>425, "name_en"=>"Melfa", "lang"=>"en", "lat"=>37.649192, "lon"=>-75.740768),
"/US/US-VA/merrifield-21872" => array("type"=>"cdp", "pop"=>8399, "name_en"=>"Merrifield", "lang"=>"en", "lat"=>38.873033, "lon"=>-77.243290),
"/US/US-VA/merrimac-21873" => array("type"=>"cdp", "pop"=>1713, "name_en"=>"Merrimac", "lang"=>"en", "lat"=>37.190618, "lon"=>-80.425842),
"/US/US-VA/middleburg-21874" => array("type"=>"town", "pop"=>943, "name_ru"=>"Миддлберг", "name_en"=>"Middleburg", "lang"=>"en", "lat"=>38.970980, "lon"=>-77.739754),
"/US/US-VA/middletown-21875" => array("type"=>"town", "pop"=>1141, "name_ru"=>"Мидлтаун", "name_en"=>"Middletown", "lang"=>"en", "lat"=>39.028738, "lon"=>-78.277868),
"/US/US-VA/mineral-21876" => array("type"=>"town", "pop"=>476, "name_ru"=>"Минерал", "name_en"=>"Mineral", "lang"=>"en", "lat"=>38.007272, "lon"=>-77.906585),
"/US/US-VA/montclair-21877" => array("type"=>"cdp", "pop"=>11399, "name_ru"=>"Монтклэр", "name_en"=>"Montclair", "lang"=>"en", "lat"=>38.610867, "lon"=>-77.339868),
"/US/US-VA/monterey-21878" => array("type"=>"town", "pop"=>144, "name_ru"=>"Монтерей", "name_en"=>"Monterey", "lang"=>"en", "lat"=>38.411768, "lon"=>-79.580698),
"/US/US-VA/montrose-21879" => array("type"=>"cdp", "pop"=>6405, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>37.520500, "lon"=>-77.378509),
"/US/US-VA/montross-21880" => array("type"=>"town", "pop"=>299, "name_en"=>"Montross", "lang"=>"en", "lat"=>38.093084, "lon"=>-76.822813),
"/US/US-VA/mount-crawford-21881" => array("type"=>"town", "pop"=>286, "name_en"=>"Mount Crawford", "lang"=>"en", "lat"=>38.356878, "lon"=>-78.940953),
"/US/US-VA/mount-jackson-21882" => array("type"=>"town", "pop"=>2001, "name_en"=>"Mount Jackson", "lang"=>"en", "lat"=>38.749836, "lon"=>-78.637901),
"/US/US-VA/mount-vernon-21883" => array("type"=>"cdp", "pop"=>27485, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>38.718617, "lon"=>-77.109239),
"/US/US-VA/narrows-21884" => array("type"=>"town", "pop"=>2148, "name_en"=>"Narrows", "lang"=>"en", "lat"=>37.331382, "lon"=>-80.809095),
"/US/US-VA/nassawadox-21885" => array("type"=>"town", "pop"=>592, "name_en"=>"Nassawadox", "lang"=>"en", "lat"=>37.476470, "lon"=>-75.862845),
"/US/US-VA/new-castle-21886" => array("type"=>"town", "pop"=>169, "name_ru"=>"Нью Касл", "name_en"=>"New Castle", "lang"=>"en", "lat"=>37.499973, "lon"=>-80.112438),
"/US/US-VA/new-market-21888" => array("type"=>"town", "pop"=>1852, "name_ru"=>"Нью Маркет", "name_en"=>"New Market", "lang"=>"en", "lat"=>38.647626, "lon"=>-78.674851),
"/US/US-VA/newington-21887" => array("type"=>"cdp", "pop"=>17965, "name_ru"=>"Ньюингтон", "name_en"=>"Newington", "lang"=>"en", "lat"=>38.736019, "lon"=>-77.204165),
"/US/US-VA/newport-news-21889" => array("type"=>"city", "pop"=>179614, "name_en"=>"Newport News", "lang"=>"en", "lat"=>37.075850, "lon"=>-76.514160),
"/US/US-VA/newsoms-21890" => array("type"=>"town", "pop"=>280, "name_en"=>"Newsoms", "lang"=>"en", "lat"=>36.627564, "lon"=>-77.124866),
"/US/US-VA/nickelsville-21891" => array("type"=>"town", "pop"=>424, "name_en"=>"Nickelsville", "lang"=>"en", "lat"=>36.751317, "lon"=>-82.416271),
"/US/US-VA/norfolk-21892" => array("type"=>"city", "pop"=>234220, "name_en"=>"Norfolk", "lang"=>"en", "lat"=>36.923200, "lon"=>-76.244943),
"/US/US-VA/north-springfield-21893" => array("type"=>"cdp", "pop"=>8996, "name_en"=>"North Springfield", "lang"=>"en", "lat"=>38.801067, "lon"=>-77.211451),
"/US/US-VA/norton-21894" => array("type"=>"city", "pop"=>3702, "name_ru"=>"Нортон", "name_en"=>"Norton", "lang"=>"en", "lat"=>36.931339, "lon"=>-82.626090),
"/US/US-VA/oakton-21895" => array("type"=>"cdp", "pop"=>24610, "name_en"=>"Oakton", "lang"=>"en", "lat"=>38.892181, "lon"=>-77.301173),
"/US/US-VA/occoquan-21896" => array("type"=>"town", "pop"=>816, "name_en"=>"Occoquan", "lang"=>"en", "lat"=>38.682433, "lon"=>-77.260499),
"/US/US-VA/onancock-21897" => array("type"=>"town", "pop"=>1389, "name_en"=>"Onancock", "lang"=>"en", "lat"=>37.709892, "lon"=>-75.744314),
"/US/US-VA/onley-21898" => array("type"=>"town", "pop"=>471, "name_en"=>"Onley", "lang"=>"en", "lat"=>37.690637, "lon"=>-75.719343),
"/US/US-VA/orange-21899" => array("type"=>"town", "pop"=>4629, "name_ru"=>"Оранж", "name_en"=>"Orange", "lang"=>"en", "lat"=>38.246982, "lon"=>-78.111393),
"/US/US-VA/painter-21900" => array("type"=>"town", "pop"=>234, "name_en"=>"Painter", "lang"=>"en", "lat"=>37.585499, "lon"=>-75.783793),
"/US/US-VA/pamplin-city-21901" => array("type"=>"town", "pop"=>197, "name_en"=>"Pamplin City", "lang"=>"en", "lat"=>37.263553, "lon"=>-78.682134),
"/US/US-VA/parksley-21902" => array("type"=>"town", "pop"=>790, "name_en"=>"Parksley", "lang"=>"en", "lat"=>37.785796, "lon"=>-75.653974),
"/US/US-VA/pearisburg-21903" => array("type"=>"town", "pop"=>2764, "name_en"=>"Pearisburg", "lang"=>"en", "lat"=>37.326580, "lon"=>-80.727509),
"/US/US-VA/pembroke-21904" => array("type"=>"town", "pop"=>1163, "name_ru"=>"Пемброук", "name_en"=>"Pembroke", "lang"=>"en", "lat"=>37.321778, "lon"=>-80.636616),
"/US/US-VA/pennington-gap-21905" => array("type"=>"town", "pop"=>1726, "name_en"=>"Pennington Gap", "lang"=>"en", "lat"=>36.759576, "lon"=>-83.028856),
"/US/US-VA/petersburg-21906" => array("type"=>"city", "pop"=>32916, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>37.204350, "lon"=>-77.392576),
"/US/US-VA/phenix-21907" => array("type"=>"town", "pop"=>186, "name_en"=>"Phenix", "lang"=>"en", "lat"=>37.080685, "lon"=>-78.748915),
"/US/US-VA/pimmit-hills-21908" => array("type"=>"cdp", "pop"=>6019, "name_en"=>"Pimmit Hills", "lang"=>"en", "lat"=>38.910000, "lon"=>-77.199038),
"/US/US-VA/pocahontas-21909" => array("type"=>"town", "pop"=>419, "name_ru"=>"Покахонтас", "name_en"=>"Pocahontas", "lang"=>"en", "lat"=>37.305788, "lon"=>-81.343199),
"/US/US-VA/poquoson-21910" => array("type"=>"city", "pop"=>11829, "name_en"=>"Poquoson", "lang"=>"en", "lat"=>37.153396, "lon"=>-76.290436),
"/US/US-VA/port-royal-21911" => array("type"=>"town", "pop"=>166, "name_en"=>"Port Royal", "lang"=>"en", "lat"=>38.169669, "lon"=>-77.191064),
"/US/US-VA/portsmouth-21912" => array("type"=>"city", "pop"=>100577, "name_ru"=>"Портсмаут", "name_en"=>"Portsmouth", "lang"=>"en", "lat"=>36.855534, "lon"=>-76.356410),
"/US/US-VA/pound-21913" => array("type"=>"town", "pop"=>1072, "name_en"=>"Pound", "lang"=>"en", "lat"=>37.124719, "lon"=>-82.607476),
"/US/US-VA/pulaski-21914" => array("type"=>"town", "pop"=>8983, "name_ru"=>"Пуласки", "name_en"=>"Pulaski", "lang"=>"en", "lat"=>37.052568, "lon"=>-80.762216),
"/US/US-VA/purcellville-21915" => array("type"=>"town", "pop"=>5137, "name_en"=>"Purcellville", "lang"=>"en", "lat"=>39.138003, "lon"=>-77.711459),
"/US/US-VA/quantico-21916" => array("type"=>"town", "pop"=>607, "name_en"=>"Quantico", "lang"=>"en", "lat"=>38.522450, "lon"=>-77.289687),
"/US/US-VA/quantico-station-21917" => array("type"=>"cdp", "pop"=>7425, "name_en"=>"Quantico Station", "lang"=>"en", "lat"=>38.518731, "lon"=>-77.314440),
"/US/US-VA/radford-21918" => array("type"=>"city", "pop"=>16125, "name_en"=>"Radford", "lang"=>"en", "lat"=>37.124075, "lon"=>-80.558897),
"/US/US-VA/raven-21919" => array("type"=>"cdp", "pop"=>2640, "name_en"=>"Raven", "lang"=>"en", "lat"=>37.094158, "lon"=>-81.862026),
"/US/US-VA/remington-21920" => array("type"=>"town", "pop"=>673, "name_ru"=>"Ремингтон", "name_en"=>"Remington", "lang"=>"en", "lat"=>38.534141, "lon"=>-77.809040),
"/US/US-VA/reston-21921" => array("type"=>"cdp", "pop"=>48556, "name_en"=>"Reston", "lang"=>"en", "lat"=>38.952067, "lon"=>-77.345174),
"/US/US-VA/rich-creek-21922" => array("type"=>"town", "pop"=>681, "name_en"=>"Rich Creek", "lang"=>"en", "lat"=>37.383965, "lon"=>-80.821886),
"/US/US-VA/richlands-21923" => array("type"=>"town", "pop"=>3997, "name_en"=>"Richlands", "lang"=>"en", "lat"=>37.095160, "lon"=>-81.802310),
"/US/US-VA/richmond-21924" => array("type"=>"city", "pop"=>202002, "name_ru"=>"Ричмонд", "name_en"=>"Richmond", "lang"=>"en", "lat"=>37.531050, "lon"=>-77.474584),
"/US/US-VA/ridgeway-21925" => array("type"=>"town", "pop"=>779, "name_ru"=>"Риджуэй", "name_en"=>"Ridgeway", "lang"=>"en", "lat"=>36.579150, "lon"=>-79.858243),
"/US/US-VA/rio-21926" => array("type"=>"cdp", "pop"=>5133, "name_ru"=>"Рио", "name_en"=>"Rio", "lang"=>"en", "lat"=>38.076150, "lon"=>-78.461290),
"/US/US-VA/roanoke-21927" => array("type"=>"city", "pop"=>92967, "name_ru"=>"Роанок", "name_en"=>"Roanoke", "lang"=>"en", "lat"=>37.278400, "lon"=>-79.958399),
"/US/US-VA/rocky-mount-21928" => array("type"=>"town", "pop"=>4535, "name_en"=>"Rocky Mount", "lang"=>"en", "lat"=>36.996575, "lon"=>-79.890619),
"/US/US-VA/rose-hill-21929" => array("type"=>"cdp", "pop"=>12675, "name_ru"=>"Роуз Хилл", "name_en"=>"Rose Hill", "lang"=>"en", "lat"=>38.788600, "lon"=>-77.112954),
"/US/US-VA/round-hill-21930" => array("type"=>"town", "pop"=>746, "name_en"=>"Round Hill", "lang"=>"en", "lat"=>39.133520, "lon"=>-77.770525),
"/US/US-VA/rural-retreat-21931" => array("type"=>"town", "pop"=>1346, "name_en"=>"Rural Retreat", "lang"=>"en", "lat"=>36.894268, "lon"=>-81.276430),
"/US/US-VA/rushmere-21932" => array("type"=>"cdp", "pop"=>1064, "name_en"=>"Rushmere", "lang"=>"en", "lat"=>37.083680, "lon"=>-76.673212),
"/US/US-VA/salem-21935" => array("type"=>"city", "pop"=>25449, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>37.286650, "lon"=>-80.056418),
"/US/US-VA/saltville-21936" => array("type"=>"town", "pop"=>2206, "name_en"=>"Saltville", "lang"=>"en", "lat"=>36.875019, "lon"=>-81.762838),
"/US/US-VA/saxis-21937" => array("type"=>"town", "pop"=>320, "name_en"=>"Saxis", "lang"=>"en", "lat"=>37.925378, "lon"=>-75.723636),
"/US/US-VA/scottsburg-21938" => array("type"=>"town", "pop"=>143, "name_ru"=>"Скоттсбург", "name_en"=>"Scottsburg", "lang"=>"en", "lat"=>36.759752, "lon"=>-78.790786),
"/US/US-VA/scottsville-21939" => array("type"=>"town", "pop"=>575, "name_ru"=>"Скоттсвилл", "name_en"=>"Scottsville", "lang"=>"en", "lat"=>37.798939, "lon"=>-78.493364),
"/US/US-VA/seven-corners-21940" => array("type"=>"cdp", "pop"=>7280, "name_en"=>"Seven Corners", "lang"=>"en", "lat"=>38.865250, "lon"=>-77.144697),
"/US/US-VA/shawsville-21941" => array("type"=>"cdp", "pop"=>1260, "name_en"=>"Shawsville", "lang"=>"en", "lat"=>37.167925, "lon"=>-80.241496),
"/US/US-VA/shenandoah-21942" => array("type"=>"town", "pop"=>1858, "name_ru"=>"Шенандоа", "name_en"=>"Shenandoah", "lang"=>"en", "lat"=>38.487426, "lon"=>-78.622536),
"/US/US-VA/smithfield-21943" => array("type"=>"town", "pop"=>7034, "name_ru"=>"Смитфилд", "name_en"=>"Smithfield", "lang"=>"en", "lat"=>36.982679, "lon"=>-76.617760),
"/US/US-VA/south-boston-21944" => array("type"=>"city", "pop"=>7884, "name_en"=>"South Boston", "lang"=>"en", "lat"=>36.707791, "lon"=>-78.906246),
"/US/US-VA/south-hill-21945" => array("type"=>"town", "pop"=>4556, "name_en"=>"South Hill", "lang"=>"en", "lat"=>36.728981, "lon"=>-78.128291),
"/US/US-VA/spotsylvania-courthouse-21946" => array("type"=>"cdp", "pop"=>2694, "name_en"=>"Spotsylvania Courthouse", "lang"=>"en", "lat"=>38.197743, "lon"=>-77.588167),
"/US/US-VA/springfield-21947" => array("type"=>"cdp", "pop"=>23706, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>38.786033, "lon"=>-77.179620),
"/US/US-VA/st.-charles-21933" => array("type"=>"town", "pop"=>153, "name_ru"=>"Сент-Чарльз", "name_en"=>"St. Charles", "lang"=>"en", "lat"=>36.803918, "lon"=>-83.057892),
"/US/US-VA/st.-paul-21934" => array("type"=>"town", "pop"=>968, "name_ru"=>"Сент-Пол", "name_en"=>"St. Paul", "lang"=>"en", "lat"=>36.905500, "lon"=>-82.310277),
"/US/US-VA/stanardsville-21948" => array("type"=>"town", "pop"=>500, "name_en"=>"Stanardsville", "lang"=>"en", "lat"=>38.300362, "lon"=>-78.438267),
"/US/US-VA/stanley-21949" => array("type"=>"town", "pop"=>1564, "name_ru"=>"Стэнли", "name_en"=>"Stanley", "lang"=>"en", "lat"=>38.579123, "lon"=>-78.504731),
"/US/US-VA/stanleytown-21950" => array("type"=>"cdp", "pop"=>1563, "name_en"=>"Stanleytown", "lang"=>"en", "lat"=>36.751668, "lon"=>-79.951277),
"/US/US-VA/staunton-21951" => array("type"=>"city", "pop"=>23967, "name_ru"=>"Стонтон", "name_en"=>"Staunton", "lang"=>"en", "lat"=>38.158808, "lon"=>-79.062277),
"/US/US-VA/stephens-city-21952" => array("type"=>"town", "pop"=>1488, "name_en"=>"Stephens City", "lang"=>"en", "lat"=>39.096950, "lon"=>-78.221407),
"/US/US-VA/sterling-21953" => array("type"=>"cdp", "pop"=>20512, "name_ru"=>"Стерлинг", "name_en"=>"Sterling", "lang"=>"en", "lat"=>39.001333, "lon"=>-77.408297),
"/US/US-VA/stony-creek-21954" => array("type"=>"town", "pop"=>185, "name_en"=>"Stony Creek", "lang"=>"en", "lat"=>36.947353, "lon"=>-77.400169),
"/US/US-VA/strasburg-21955" => array("type"=>"town", "pop"=>4373, "name_ru"=>"Страсбург", "name_en"=>"Strasburg", "lang"=>"en", "lat"=>38.988473, "lon"=>-78.357669),
"/US/US-VA/stuart-21956" => array("type"=>"town", "pop"=>1436, "name_ru"=>"Стюарт", "name_en"=>"Stuart", "lang"=>"en", "lat"=>36.640061, "lon"=>-80.274017),
"/US/US-VA/stuarts-draft-21957" => array("type"=>"cdp", "pop"=>5087, "name_en"=>"Stuarts Draft", "lang"=>"en", "lat"=>38.025597, "lon"=>-79.024856),
"/US/US-VA/sudley-21958" => array("type"=>"cdp", "pop"=>7321, "name_en"=>"Sudley", "lang"=>"en", "lat"=>38.792650, "lon"=>-77.495235),
"/US/US-VA/suffolk-21959" => array("type"=>"city", "pop"=>82302, "name_en"=>"Suffolk", "lang"=>"en", "lat"=>36.703293, "lon"=>-76.638147),
"/US/US-VA/sugarland-run-21960" => array("type"=>"cdp", "pop"=>9357, "name_en"=>"Sugarland Run", "lang"=>"en", "lat"=>39.037454, "lon"=>-77.374327),
"/US/US-VA/surry-21961" => array("type"=>"town", "pop"=>253, "name_en"=>"Surry", "lang"=>"en", "lat"=>37.137444, "lon"=>-76.835025),
"/US/US-VA/tangier-21962" => array("type"=>"town", "pop"=>652, "name_en"=>"Tangier", "lang"=>"en", "lat"=>37.824961, "lon"=>-75.993362),
"/US/US-VA/tappahannock-21963" => array("type"=>"town", "pop"=>2195, "name_en"=>"Tappahannock", "lang"=>"en", "lat"=>37.919965, "lon"=>-76.867831),
"/US/US-VA/tazewell-21964" => array("type"=>"town", "pop"=>4267, "name_en"=>"Tazewell", "lang"=>"en", "lat"=>37.122838, "lon"=>-81.516731),
"/US/US-VA/the-plains-21965" => array("type"=>"town", "pop"=>286, "name_en"=>"The Plains", "lang"=>"en", "lat"=>38.862770, "lon"=>-77.773596),
"/US/US-VA/timberlake-21966" => array("type"=>"cdp", "pop"=>10314, "name_en"=>"Timberlake", "lang"=>"en", "lat"=>37.323160, "lon"=>-79.249886),
"/US/US-VA/timberville-21967" => array("type"=>"town", "pop"=>1713, "name_en"=>"Timberville", "lang"=>"en", "lat"=>38.635952, "lon"=>-78.772397),
"/US/US-VA/toms-brook-21968" => array("type"=>"town", "pop"=>262, "name_en"=>"Toms Brook", "lang"=>"en", "lat"=>38.947105, "lon"=>-78.439928),
"/US/US-VA/triangle-21969" => array("type"=>"cdp", "pop"=>4740, "name_en"=>"Triangle", "lang"=>"en", "lat"=>38.545700, "lon"=>-77.314625),
"/US/US-VA/troutdale-21970" => array("type"=>"town", "pop"=>181, "name_en"=>"Troutdale", "lang"=>"en", "lat"=>36.702732, "lon"=>-81.443907),
"/US/US-VA/troutville-21971" => array("type"=>"town", "pop"=>421, "name_en"=>"Troutville", "lang"=>"en", "lat"=>37.413595, "lon"=>-79.878012),
"/US/US-VA/tuckahoe-21972" => array("type"=>"cdp", "pop"=>42629, "name_en"=>"Tuckahoe", "lang"=>"en", "lat"=>37.587032, "lon"=>-77.587022),
"/US/US-VA/tysons-corner-21973" => array("type"=>"cdp", "pop"=>13124, "name_en"=>"Tysons Corner", "lang"=>"en", "lat"=>38.918650, "lon"=>-77.231749),
"/US/US-VA/university-heights-21974" => array("type"=>"cdp", "pop"=>6900, "name_ru"=>"Юниверсити Хайтс", "name_en"=>"University Heights", "lang"=>"en", "lat"=>38.037083, "lon"=>-78.533086),
"/US/US-VA/urbanna-21975" => array("type"=>"town", "pop"=>538, "name_en"=>"Urbanna", "lang"=>"en", "lat"=>37.639730, "lon"=>-76.575437),
"/US/US-VA/vansant-21976" => array("type"=>"cdp", "pop"=>1187, "name_en"=>"Vansant", "lang"=>"en", "lat"=>37.230613, "lon"=>-82.089649),
"/US/US-VA/verona-21977" => array("type"=>"cdp", "pop"=>3479, "name_ru"=>"Верона", "name_en"=>"Verona", "lang"=>"en", "lat"=>38.199848, "lon"=>-79.006161),
"/US/US-VA/victoria-21978" => array("type"=>"town", "pop"=>1731, "name_ru"=>"Виктория", "name_en"=>"Victoria", "lang"=>"en", "lat"=>36.994598, "lon"=>-78.224550),
"/US/US-VA/vienna-21979" => array("type"=>"town", "pop"=>14903, "name_ru"=>"Вьенна", "name_en"=>"Vienna", "lang"=>"en", "lat"=>38.900400, "lon"=>-77.260159),
"/US/US-VA/villa-heights-21980" => array("type"=>"cdp", "pop"=>1021, "name_en"=>"Villa Heights", "lang"=>"en", "lat"=>36.697314, "lon"=>-79.904049),
"/US/US-VA/vint-hill-farms-station-21981" => array("type"=>"cdp", "pop"=>1332, "name_en"=>"Vint Hill Farms Station", "lang"=>"en", "lat"=>38.745231, "lon"=>-77.676545),
"/US/US-VA/vinton-21982" => array("type"=>"town", "pop"=>7868, "name_ru"=>"Винтон", "name_en"=>"Vinton", "lang"=>"en", "lat"=>37.273067, "lon"=>-79.888389),
"/US/US-VA/virgilina-21983" => array("type"=>"town", "pop"=>147, "name_en"=>"Virgilina", "lang"=>"en", "lat"=>36.545480, "lon"=>-78.775949),
"/US/US-VA/virginia-beach-21984" => array("type"=>"city", "pop"=>433746, "name_en"=>"Virginia Beach", "lang"=>"en", "lat"=>36.739356, "lon"=>-76.043668),
"/US/US-VA/wachapreague-21985" => array("type"=>"town", "pop"=>222, "name_en"=>"Wachapreague", "lang"=>"en", "lat"=>37.606233, "lon"=>-75.689797),
"/US/US-VA/wakefield-21986" => array("type"=>"town", "pop"=>965, "name_ru"=>"Уэйкфилд", "name_en"=>"Wakefield", "lang"=>"en", "lat"=>36.968648, "lon"=>-76.988236),
"/US/US-VA/warrenton-21987" => array("type"=>"town", "pop"=>8947, "name_ru"=>"Уоррентон", "name_en"=>"Warrenton", "lang"=>"en", "lat"=>38.718228, "lon"=>-77.797133),
"/US/US-VA/warsaw-21988" => array("type"=>"town", "pop"=>1352, "name_ru"=>"Варшава", "name_en"=>"Warsaw", "lang"=>"en", "lat"=>37.961101, "lon"=>-76.761283),
"/US/US-VA/washington-21989" => array("type"=>"town", "pop"=>172, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>38.712742, "lon"=>-78.159773),
"/US/US-VA/waverly-21990" => array("type"=>"town", "pop"=>2157, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>37.033764, "lon"=>-77.095668),
"/US/US-VA/waynesboro-21991" => array("type"=>"city", "pop"=>21953, "name_ru"=>"Уэйнсборо", "name_en"=>"Waynesboro", "lang"=>"en", "lat"=>38.069502, "lon"=>-78.902634),
"/US/US-VA/weber-city-21992" => array("type"=>"town", "pop"=>1331, "name_en"=>"Weber City", "lang"=>"en", "lat"=>36.622433, "lon"=>-82.560336),
"/US/US-VA/west-gate-21993" => array("type"=>"cdp", "pop"=>6565, "name_en"=>"West Gate", "lang"=>"en", "lat"=>38.782650, "lon"=>-77.497887),
"/US/US-VA/west-point-21994" => array("type"=>"town", "pop"=>3145, "name_ru"=>"Вэст Поинт", "name_en"=>"West Point", "lang"=>"en", "lat"=>37.552308, "lon"=>-76.800270),
"/US/US-VA/west-springfield-21995" => array("type"=>"cdp", "pop"=>28126, "name_ru"=>"Вэст Спрингфилд", "name_en"=>"West Springfield", "lang"=>"en", "lat"=>38.785867, "lon"=>-77.234041),
"/US/US-VA/white-stone-21996" => array("type"=>"town", "pop"=>338, "name_en"=>"White Stone", "lang"=>"en", "lat"=>37.645046, "lon"=>-76.391574),
"/US/US-VA/williamsburg-21997" => array("type"=>"city", "pop"=>12481, "name_ru"=>"Уилльямсбург", "name_en"=>"Williamsburg", "lang"=>"en", "lat"=>37.270008, "lon"=>-76.706855),
"/US/US-VA/winchester-21998" => array("type"=>"city", "pop"=>25897, "name_ru"=>"Винчестер", "name_en"=>"Winchester", "lang"=>"en", "lat"=>39.174479, "lon"=>-78.175001),
"/US/US-VA/windsor-21999" => array("type"=>"town", "pop"=>2493, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>36.808172, "lon"=>-76.739457),
"/US/US-VA/wise-22000" => array("type"=>"town", "pop"=>3231, "name_en"=>"Wise", "lang"=>"en", "lat"=>36.977255, "lon"=>-82.580440),
"/US/US-VA/wolf-trap-22001" => array("type"=>"cdp", "pop"=>13133, "name_en"=>"Wolf Trap", "lang"=>"en", "lat"=>38.939827, "lon"=>-77.286278),
"/US/US-VA/woodbridge-22002" => array("type"=>"cdp", "pop"=>26401, "name_ru"=>"Вудбридж", "name_en"=>"Woodbridge", "lang"=>"en", "lat"=>38.635650, "lon"=>-77.262095),
"/US/US-VA/woodstock-22003" => array("type"=>"town", "pop"=>4284, "name_ru"=>"Вудсток", "name_en"=>"Woodstock", "lang"=>"en", "lat"=>38.876341, "lon"=>-78.519036),
"/US/US-VA/wytheville-22004" => array("type"=>"town", "pop"=>8282, "name_en"=>"Wytheville", "lang"=>"en", "lat"=>36.953680, "lon"=>-81.087020),
"/US/US-VA/yorkshire-22005" => array("type"=>"cdp", "pop"=>5699, "name_en"=>"Yorkshire", "lang"=>"en", "lat"=>38.788350, "lon"=>-77.450378),
"/US/US-VI" => array("type"=>"outlying area", "pop"=>108448, "name_ru"=>"Виргинские Ислэндс", "name_en"=>"Virgin Islands", "lang"=>"en"),
"/US/US-VI/anna's-retreat-23777" => array("type"=>"cdp", "pop"=>9084, "name_en"=>"Anna's Retreat", "lang"=>"en", "lat"=>18.342079, "lon"=>-64.887906),
"/US/US-VI/charlotte-amalie-23778" => array("type"=>"town", "pop"=>12331, "name_en"=>"Charlotte Amalie", "lang"=>"en", "lat"=>18.344032, "lon"=>-64.933536),
"/US/US-VI/charlotte-amalie-east-23779" => array("type"=>"cdp", "pop"=>2836, "name_en"=>"Charlotte Amalie East", "lang"=>"en", "lat"=>18.337018, "lon"=>-64.912684),
"/US/US-VI/charlotte-amalie-west-23780" => array("type"=>"cdp", "pop"=>5422, "name_en"=>"Charlotte Amalie West", "lang"=>"en", "lat"=>18.339165, "lon"=>-64.960538),
"/US/US-VI/christiansted-23781" => array("type"=>"town", "pop"=>2555, "name_en"=>"Christiansted", "lang"=>"en", "lat"=>17.746289, "lon"=>-64.707256),
"/US/US-VI/cruz-bay-23782" => array("type"=>"cdp", "pop"=>2466, "name_en"=>"Cruz Bay", "lang"=>"en", "lat"=>18.322496, "lon"=>-64.782444),
"/US/US-VI/frederiksted-23783" => array("type"=>"town", "pop"=>1064, "name_en"=>"Frederiksted", "lang"=>"en", "lat"=>17.713676, "lon"=>-64.881673),
"/US/US-VI/frederiksted-southeast-23784" => array("type"=>"cdp", "pop"=>3002, "name_en"=>"Frederiksted Southeast", "lang"=>"en", "lat"=>17.707358, "lon"=>-64.880139),
"/US/US-VI/grove-place-23785" => array("type"=>"cdp", "pop"=>2655, "name_en"=>"Grove Place", "lang"=>"en", "lat"=>17.726297, "lon"=>-64.827477),
"/US/US-VI/grove-place-23786" => array("type"=>"cdp", "pop"=>2655, "name_en"=>"Grove Place", "lang"=>"en", "lat"=>17.726297, "lon"=>-64.827477),
"/US/US-VT" => array("type"=>"state", "pop"=>621270, "name_ru"=>"Вермонт", "name_en"=>"Vermont", "lang"=>"en"),
"/US/US-VT/albany-21592" => array("type"=>"village", "pop"=>158, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>44.730535, "lon"=>-72.382024),
"/US/US-VT/alburg-21593" => array("type"=>"village", "pop"=>511, "name_en"=>"Alburg", "lang"=>"en", "lat"=>44.977392, "lon"=>-73.301325),
"/US/US-VT/arlington-21594" => array("type"=>"cdp", "pop"=>2425, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>43.067278, "lon"=>-73.151863),
"/US/US-VT/barre-21595" => array("type"=>"city", "pop"=>8837, "name_ru"=>"Барр", "name_en"=>"Barre", "lang"=>"en", "lat"=>44.200061, "lon"=>-72.507914),
"/US/US-VT/barton-21596" => array("type"=>"village", "pop"=>719, "name_ru"=>"Бартон", "name_en"=>"Barton", "lang"=>"en", "lat"=>44.749561, "lon"=>-72.177389),
"/US/US-VT/bellows-falls-21597" => array("type"=>"village", "pop"=>2900, "name_en"=>"Bellows Falls", "lang"=>"en", "lat"=>43.136137, "lon"=>-72.453207),
"/US/US-VT/bennington-21598" => array("type"=>"cdp", "pop"=>36382, "name_ru"=>"Беннингтон", "name_en"=>"Bennington", "lang"=>"en", "lat"=>42.872842, "lon"=>-73.185061),
"/US/US-VT/bradford-21599" => array("type"=>"village", "pop"=>2655, "name_ru"=>"Брадфорд", "name_en"=>"Bradford", "lang"=>"en", "lat"=>43.994080, "lon"=>-72.128164),
"/US/US-VT/brandon-21600" => array("type"=>"cdp", "pop"=>3866, "name_ru"=>"Брэндон", "name_en"=>"Brandon", "lang"=>"en", "lat"=>43.801198, "lon"=>-73.086484),
"/US/US-VT/brattleboro-21601" => array("type"=>"cdp", "pop"=>11491, "name_en"=>"Brattleboro", "lang"=>"en", "lat"=>42.861283, "lon"=>-72.574198),
"/US/US-VT/bristol-21602" => array("type"=>"village", "pop"=>3741, "name_ru"=>"Бристол", "name_en"=>"Bristol", "lang"=>"en", "lat"=>44.136545, "lon"=>-73.080934),
"/US/US-VT/burlington-21603" => array("type"=>"city", "pop"=>38897, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>44.488093, "lon"=>-73.226177),
"/US/US-VT/cabot-21604" => array("type"=>"village", "pop"=>247, "name_ru"=>"Кабот", "name_en"=>"Cabot", "lang"=>"en", "lat"=>44.408596, "lon"=>-72.311019),
"/US/US-VT/cambridge-21605" => array("type"=>"village", "pop"=>225, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>44.635386, "lon"=>-72.882468),
"/US/US-VT/chester-chester-depot-21606" => array("type"=>"cdp", "pop"=>1057, "name_en"=>"Chester-Chester Depot", "lang"=>"en", "lat"=>43.262786, "lon"=>-72.593868),
"/US/US-VT/derby-center-21607" => array("type"=>"village", "pop"=>668, "name_en"=>"Derby Center", "lang"=>"en", "lat"=>44.955829, "lon"=>-72.132261),
"/US/US-VT/derby-line-21608" => array("type"=>"village", "pop"=>770, "name_en"=>"Derby Line", "lang"=>"en", "lat"=>45.002354, "lon"=>-72.104048),
"/US/US-VT/enosburg-falls-21609" => array("type"=>"village", "pop"=>1435, "name_en"=>"Enosburg Falls", "lang"=>"en", "lat"=>44.908272, "lon"=>-72.804373),
"/US/US-VT/essex-junction-21610" => array("type"=>"village", "pop"=>9056, "name_en"=>"Essex Junction", "lang"=>"en", "lat"=>44.490350, "lon"=>-73.112399),
"/US/US-VT/fair-haven-21611" => array("type"=>"cdp", "pop"=>2924, "name_ru"=>"Фэр Хейвен", "name_en"=>"Fair Haven", "lang"=>"en", "lat"=>43.595755, "lon"=>-73.271064),
"/US/US-VT/graniteville-east-barre-21612" => array("type"=>"cdp", "pop"=>2189, "name_en"=>"Graniteville-East Barre", "lang"=>"en", "lat"=>44.157865, "lon"=>-72.470371),
"/US/US-VT/hyde-park-21613" => array("type"=>"village", "pop"=>527, "name_en"=>"Hyde Park", "lang"=>"en", "lat"=>44.594278, "lon"=>-72.612900),
"/US/US-VT/island-pond-21614" => array("type"=>"cdp", "pop"=>1222, "name_en"=>"Island Pond", "lang"=>"en", "lat"=>44.812943, "lon"=>-71.885456),
"/US/US-VT/jacksonville-21615" => array("type"=>"village", "pop"=>218, "name_ru"=>"Джексонвилл", "name_en"=>"Jacksonville", "lang"=>"en", "lat"=>42.797273, "lon"=>-72.820761),
"/US/US-VT/jeffersonville-21616" => array("type"=>"village", "pop"=>557, "name_ru"=>"Джефферсонвилл", "name_en"=>"Jeffersonville", "lang"=>"en", "lat"=>44.644417, "lon"=>-72.828124),
"/US/US-VT/jericho-21617" => array("type"=>"village", "pop"=>5190, "name_ru"=>"Джерико", "name_en"=>"Jericho", "lang"=>"en", "lat"=>44.501777, "lon"=>-72.986961),
"/US/US-VT/johnson-21618" => array("type"=>"village", "pop"=>1388, "name_ru"=>"Джонсон", "name_en"=>"Johnson", "lang"=>"en", "lat"=>44.637231, "lon"=>-72.677949),
"/US/US-VT/ludlow-21619" => array("type"=>"village", "pop"=>1026, "name_ru"=>"Ладлоу", "name_en"=>"Ludlow", "lang"=>"en", "lat"=>43.396301, "lon"=>-72.697469),
"/US/US-VT/lyndonville-21620" => array("type"=>"village", "pop"=>1215, "name_en"=>"Lyndonville", "lang"=>"en", "lat"=>44.534889, "lon"=>-72.002975),
"/US/US-VT/manchester-21621" => array("type"=>"village", "pop"=>695, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>43.161998, "lon"=>-73.072004),
"/US/US-VT/manchester-center-21622" => array("type"=>"cdp", "pop"=>1574, "name_en"=>"Manchester Center", "lang"=>"en", "lat"=>43.177813, "lon"=>-73.043518),
"/US/US-VT/marshfield-21623" => array("type"=>"village", "pop"=>263, "name_ru"=>"Маршфилд", "name_en"=>"Marshfield", "lang"=>"en", "lat"=>44.350827, "lon"=>-72.352138),
"/US/US-VT/middlebury-21624" => array("type"=>"cdp", "pop"=>8271, "name_ru"=>"Миддлбери", "name_en"=>"Middlebury", "lang"=>"en", "lat"=>44.007721, "lon"=>-73.157176),
"/US/US-VT/milton-21625" => array("type"=>"village", "pop"=>10714, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>44.638392, "lon"=>-73.112471),
"/US/US-VT/montpelier-21626" => array("type"=>"city", "pop"=>7760, "name_ru"=>"Монтпильер", "name_en"=>"Montpelier", "lang"=>"en", "lat"=>44.266482, "lon"=>-72.571854),
"/US/US-VT/morrisville-21627" => array("type"=>"village", "pop"=>2040, "name_ru"=>"Моррисвилл", "name_en"=>"Morrisville", "lang"=>"en", "lat"=>44.558560, "lon"=>-72.596356),
"/US/US-VT/newbury-21628" => array("type"=>"village", "pop"=>432, "name_en"=>"Newbury", "lang"=>"en", "lat"=>44.080359, "lon"=>-72.060550),
"/US/US-VT/newfane-21629" => array("type"=>"village", "pop"=>110, "name_en"=>"Newfane", "lang"=>"en", "lat"=>42.987643, "lon"=>-72.656380),
"/US/US-VT/newport-21630" => array("type"=>"city", "pop"=>5148, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>44.936960, "lon"=>-72.208938),
"/US/US-VT/north-bennington-21631" => array("type"=>"village", "pop"=>1326, "name_en"=>"North Bennington", "lang"=>"en", "lat"=>42.925322, "lon"=>-73.244312),
"/US/US-VT/north-troy-21633" => array("type"=>"village", "pop"=>605, "name_en"=>"North Troy", "lang"=>"en", "lat"=>44.995412, "lon"=>-72.404851),
"/US/US-VT/north-westminster-21634" => array("type"=>"village", "pop"=>255, "name_en"=>"North Westminster", "lang"=>"en", "lat"=>43.119988, "lon"=>-72.455181),
"/US/US-VT/northfield-21632" => array("type"=>"village", "pop"=>3124, "name_ru"=>"Нортфилд", "name_en"=>"Northfield", "lang"=>"en", "lat"=>44.152291, "lon"=>-72.656534),
"/US/US-VT/old-bennington-21635" => array("type"=>"village", "pop"=>196, "name_en"=>"Old Bennington", "lang"=>"en", "lat"=>42.883326, "lon"=>-73.213759),
"/US/US-VT/orleans-21636" => array("type"=>"village", "pop"=>803, "name_ru"=>"Орлеанс", "name_en"=>"Orleans", "lang"=>"en", "lat"=>44.810642, "lon"=>-72.201816),
"/US/US-VT/perkinsville-21637" => array("type"=>"village", "pop"=>136, "name_en"=>"Perkinsville", "lang"=>"en", "lat"=>43.369718, "lon"=>-72.516181),
"/US/US-VT/poultney-21638" => array("type"=>"village", "pop"=>1532, "name_en"=>"Poultney", "lang"=>"en", "lat"=>43.517755, "lon"=>-73.236105),
"/US/US-VT/richford-21639" => array("type"=>"village", "pop"=>2293, "name_en"=>"Richford", "lang"=>"en", "lat"=>44.995685, "lon"=>-72.673417),
"/US/US-VT/rutland-21640" => array("type"=>"city", "pop"=>16742, "name_ru"=>"Рутленд", "name_en"=>"Rutland", "lang"=>"en", "lat"=>43.608900, "lon"=>-72.979619),
"/US/US-VT/saxtons-river-21643" => array("type"=>"village", "pop"=>477, "name_en"=>"Saxtons River", "lang"=>"en", "lat"=>43.138925, "lon"=>-72.511284),
"/US/US-VT/south-barre-21644" => array("type"=>"cdp", "pop"=>1314, "name_en"=>"South Barre", "lang"=>"en", "lat"=>44.163339, "lon"=>-72.506628),
"/US/US-VT/south-burlington-21645" => array("type"=>"city", "pop"=>17574, "name_en"=>"South Burlington", "lang"=>"en", "lat"=>44.445633, "lon"=>-73.216149),
"/US/US-VT/springfield-21646" => array("type"=>"cdp", "pop"=>8602, "name_ru"=>"Спрингфилд", "name_en"=>"Springfield", "lang"=>"en", "lat"=>43.288371, "lon"=>-72.476168),
"/US/US-VT/st.-albans-21641" => array("type"=>"city", "pop"=>7250, "name_en"=>"St. Albans", "lang"=>"en", "lat"=>44.811327, "lon"=>-73.084901),
"/US/US-VT/st.-johnsbury-21642" => array("type"=>"cdp", "pop"=>7421, "name_en"=>"St. Johnsbury", "lang"=>"en", "lat"=>44.430421, "lon"=>-72.013238),
"/US/US-VT/stowe-21647" => array("type"=>"village", "pop"=>4919, "name_en"=>"Stowe", "lang"=>"en", "lat"=>44.465934, "lon"=>-72.685681),
"/US/US-VT/swanton-21648" => array("type"=>"village", "pop"=>2526, "name_en"=>"Swanton", "lang"=>"en", "lat"=>44.921513, "lon"=>-73.120990),
"/US/US-VT/vergennes-21649" => array("type"=>"city", "pop"=>2666, "name_ru"=>"Вердженес", "name_en"=>"Vergennes", "lang"=>"en", "lat"=>44.167360, "lon"=>-73.255749),
"/US/US-VT/wallingford-21650" => array("type"=>"cdp", "pop"=>2324, "name_ru"=>"Уоллингтон", "name_en"=>"Wallingford", "lang"=>"en", "lat"=>43.480520, "lon"=>-72.975652),
"/US/US-VT/waterbury-21651" => array("type"=>"village", "pop"=>1756, "name_ru"=>"Уотербери", "name_en"=>"Waterbury", "lang"=>"en", "lat"=>44.339378, "lon"=>-72.752694),
"/US/US-VT/wells-river-21652" => array("type"=>"village", "pop"=>340, "name_en"=>"Wells River", "lang"=>"en", "lat"=>44.151355, "lon"=>-72.065855),
"/US/US-VT/west-brattleboro-21653" => array("type"=>"cdp", "pop"=>3135, "name_en"=>"West Brattleboro", "lang"=>"en", "lat"=>42.851508, "lon"=>-72.610981),
"/US/US-VT/west-burke-21654" => array("type"=>"village", "pop"=>367, "name_en"=>"West Burke", "lang"=>"en", "lat"=>44.643795, "lon"=>-71.980339),
"/US/US-VT/west-rutland-21656" => array("type"=>"cdp", "pop"=>2513, "name_en"=>"West Rutland", "lang"=>"en", "lat"=>43.594914, "lon"=>-73.045463),
"/US/US-VT/westminster-21655" => array("type"=>"village", "pop"=>262, "name_ru"=>"Уэстминстер", "name_en"=>"Westminster", "lang"=>"en", "lat"=>43.076255, "lon"=>-72.456016),
"/US/US-VT/white-river-junction-21657" => array("type"=>"cdp", "pop"=>2521, "name_en"=>"White River Junction", "lang"=>"en", "lat"=>43.650107, "lon"=>-72.323385),
"/US/US-VT/wilder-21658" => array("type"=>"cdp", "pop"=>1576, "name_ru"=>"Уайлдер", "name_en"=>"Wilder", "lang"=>"en", "lat"=>43.672329, "lon"=>-72.311299),
"/US/US-VT/winooski-21659" => array("type"=>"city", "pop"=>6429, "name_en"=>"Winooski", "lang"=>"en", "lat"=>44.495450, "lon"=>-73.185099),
"/US/US-VT/woodstock-21660" => array("type"=>"village", "pop"=>927, "name_ru"=>"Вудсток", "name_en"=>"Woodstock", "lang"=>"en", "lat"=>43.626199, "lon"=>-72.517062),
"/US/US-WA" => array("type"=>"state", "pop"=>6549224, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en"),
"/US/US-WA/aberdeen-22006" => array("type"=>"city", "pop"=>16042, "name_ru"=>"Абердин", "name_en"=>"Aberdeen", "lang"=>"en", "lat"=>46.977024, "lon"=>-123.808369),
"/US/US-WA/airway-heights-22007" => array("type"=>"city", "pop"=>5289, "name_en"=>"Airway Heights", "lang"=>"en", "lat"=>47.644400, "lon"=>-117.582899),
"/US/US-WA/albion-22008" => array("type"=>"town", "pop"=>566, "name_ru"=>"Албион", "name_en"=>"Albion", "lang"=>"en", "lat"=>46.792526, "lon"=>-117.249505),
"/US/US-WA/alderwood-manor-bothell-north-22009" => array("type"=>"cdp", "pop"=>22945, "name_en"=>"Alderwood Manor-Bothell North", "lang"=>"en", "lat"=>47.799591, "lon"=>-122.242109),
"/US/US-WA/algona-22010" => array("type"=>"city", "pop"=>2731, "name_ru"=>"Алгона", "name_en"=>"Algona", "lang"=>"en", "lat"=>47.283750, "lon"=>-122.250141),
"/US/US-WA/allyn-grapeview-22011" => array("type"=>"cdp", "pop"=>1526, "name_en"=>"Allyn-Grapeview", "lang"=>"en", "lat"=>47.344359, "lon"=>-122.831153),
"/US/US-WA/almira-22012" => array("type"=>"town", "pop"=>292, "name_en"=>"Almira", "lang"=>"en", "lat"=>47.710560, "lon"=>-118.939293),
"/US/US-WA/anacortes-22013" => array("type"=>"city", "pop"=>16798, "name_en"=>"Anacortes", "lang"=>"en", "lat"=>48.492392, "lon"=>-122.629404),
"/US/US-WA/arlington-22014" => array("type"=>"city", "pop"=>16810, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>48.175782, "lon"=>-122.138322),
"/US/US-WA/artondale-22015" => array("type"=>"cdp", "pop"=>7141, "name_en"=>"Artondale", "lang"=>"en", "lat"=>47.313769, "lon"=>-122.634453),
"/US/US-WA/asotin-22016" => array("type"=>"city", "pop"=>1129, "name_en"=>"Asotin", "lang"=>"en", "lat"=>46.338827, "lon"=>-117.043912),
"/US/US-WA/auburn-22017" => array("type"=>"city", "pop"=>55426, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>47.299000, "lon"=>-122.211800),
"/US/US-WA/ault-field-22018" => array("type"=>"cdp", "pop"=>3795, "name_en"=>"Ault Field", "lang"=>"en", "lat"=>48.337105, "lon"=>-122.676162),
"/US/US-WA/bangor-trident-base-22019" => array("type"=>"cdp", "pop"=>3702, "name_en"=>"Bangor Trident Base", "lang"=>"en", "lat"=>47.692017, "lon"=>-122.716106),
"/US/US-WA/battle-ground-22020" => array("type"=>"city", "pop"=>16812, "name_ru"=>"Бэттл Граунд", "name_en"=>"Battle Ground", "lang"=>"en", "lat"=>45.782030, "lon"=>-122.541061),
"/US/US-WA/beaux-arts-village-22021" => array("type"=>"town", "pop"=>298, "name_en"=>"Beaux Arts Village", "lang"=>"en", "lat"=>47.586058, "lon"=>-122.203203),
"/US/US-WA/bellevue-22022" => array("type"=>"city", "pop"=>123771, "name_ru"=>"Белвью", "name_en"=>"Bellevue", "lang"=>"en", "lat"=>47.603393, "lon"=>-122.155018),
"/US/US-WA/bellingham-22023" => array("type"=>"city", "pop"=>78905, "name_ru"=>"Беллингхэм", "name_en"=>"Bellingham", "lang"=>"en", "lat"=>48.744048, "lon"=>-122.466870),
"/US/US-WA/benton-city-22024" => array("type"=>"city", "pop"=>2964, "name_ru"=>"Бентон Сити", "name_en"=>"Benton City", "lang"=>"en", "lat"=>46.268016, "lon"=>-119.487095),
"/US/US-WA/bingen-22025" => array("type"=>"city", "pop"=>680, "name_en"=>"Bingen", "lang"=>"en", "lat"=>45.714201, "lon"=>-121.468146),
"/US/US-WA/birch-bay-22026" => array("type"=>"cdp", "pop"=>2656, "name_en"=>"Birch Bay", "lang"=>"en", "lat"=>48.923536, "lon"=>-122.753207),
"/US/US-WA/black-diamond-22027" => array("type"=>"city", "pop"=>4003, "name_en"=>"Black Diamond", "lang"=>"en", "lat"=>47.311695, "lon"=>-122.009899),
"/US/US-WA/blaine-22028" => array("type"=>"city", "pop"=>4975, "name_ru"=>"Блейн", "name_en"=>"Blaine", "lang"=>"en", "lat"=>48.989123, "lon"=>-122.758148),
"/US/US-WA/bonney-lake-22029" => array("type"=>"city", "pop"=>16809, "name_en"=>"Bonney Lake", "lang"=>"en", "lat"=>47.187270, "lon"=>-122.168810),
"/US/US-WA/bothell-22030" => array("type"=>"city", "pop"=>32307, "name_en"=>"Bothell", "lang"=>"en", "lat"=>47.760033, "lon"=>-122.196884),
"/US/US-WA/bremerton-22031" => array("type"=>"city", "pop"=>36006, "name_en"=>"Bremerton", "lang"=>"en", "lat"=>47.548900, "lon"=>-122.701999),
"/US/US-WA/brewster-22032" => array("type"=>"city", "pop"=>2090, "name_ru"=>"Брюстер", "name_en"=>"Brewster", "lang"=>"en", "lat"=>48.102804, "lon"=>-119.779569),
"/US/US-WA/bridgeport-22033" => array("type"=>"town", "pop"=>1978, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>48.005675, "lon"=>-119.671225),
"/US/US-WA/brier-22034" => array("type"=>"city", "pop"=>6344, "name_en"=>"Brier", "lang"=>"en", "lat"=>47.792500, "lon"=>-122.271695),
"/US/US-WA/brush-prairie-22035" => array("type"=>"cdp", "pop"=>2650, "name_en"=>"Brush Prairie", "lang"=>"en", "lat"=>45.731400, "lon"=>-122.551383),
"/US/US-WA/bryn-mawr-skyway-22036" => array("type"=>"cdp", "pop"=>12514, "name_en"=>"Bryn Mawr-Skyway", "lang"=>"en", "lat"=>47.495250, "lon"=>-122.238199),
"/US/US-WA/buckley-22037" => array("type"=>"city", "pop"=>4462, "name_ru"=>"Бакли", "name_en"=>"Buckley", "lang"=>"en", "lat"=>47.162550, "lon"=>-122.018702),
"/US/US-WA/bucoda-22038" => array("type"=>"town", "pop"=>633, "name_en"=>"Bucoda", "lang"=>"en", "lat"=>46.798195, "lon"=>-122.867946),
"/US/US-WA/burbank-22039" => array("type"=>"cdp", "pop"=>1745, "name_ru"=>"Бурбанк", "name_en"=>"Burbank", "lang"=>"en", "lat"=>46.191643, "lon"=>-118.981596),
"/US/US-WA/burien-22040" => array("type"=>"cdp", "pop"=>31316, "name_en"=>"Burien", "lang"=>"en", "lat"=>47.464933, "lon"=>-122.346408),
"/US/US-WA/burlington-22041" => array("type"=>"city", "pop"=>8697, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>48.469163, "lon"=>-122.331833),
"/US/US-WA/camas-22042" => array("type"=>"city", "pop"=>17950, "name_en"=>"Camas", "lang"=>"en", "lat"=>45.587856, "lon"=>-122.423505),
"/US/US-WA/carbonado-22043" => array("type"=>"town", "pop"=>599, "name_en"=>"Carbonado", "lang"=>"en", "lat"=>47.080910, "lon"=>-122.053401),
"/US/US-WA/carnation-22044" => array("type"=>"city", "pop"=>1814, "name_en"=>"Carnation", "lang"=>"en", "lat"=>47.646144, "lon"=>-121.909250),
"/US/US-WA/carson-river-valley-22045" => array("type"=>"cdp", "pop"=>1678, "name_en"=>"Carson River Valley", "lang"=>"en", "lat"=>45.731911, "lon"=>-121.816811),
"/US/US-WA/cascade-fairwood-22046" => array("type"=>"cdp", "pop"=>30107, "name_en"=>"Cascade-Fairwood", "lang"=>"en", "lat"=>47.449227, "lon"=>-122.162672),
"/US/US-WA/cascade-park-east-22047" => array("type"=>"cdp", "pop"=>6996, "name_en"=>"Cascade Park East", "lang"=>"en", "lat"=>45.605504, "lon"=>-122.513472),
"/US/US-WA/cascade-park-west-22048" => array("type"=>"cdp", "pop"=>6656, "name_en"=>"Cascade Park West", "lang"=>"en", "lat"=>45.611472, "lon"=>-122.537995),
"/US/US-WA/cascade-valley-22049" => array("type"=>"cdp", "pop"=>1288, "name_en"=>"Cascade Valley", "lang"=>"en", "lat"=>47.134782, "lon"=>-119.327067),
"/US/US-WA/cashmere-22050" => array("type"=>"city", "pop"=>2843, "name_en"=>"Cashmere", "lang"=>"en", "lat"=>47.518370, "lon"=>-120.467607),
"/US/US-WA/castle-rock-22051" => array("type"=>"city", "pop"=>2120, "name_ru"=>"Касл Рок", "name_en"=>"Castle Rock", "lang"=>"en", "lat"=>46.272875, "lon"=>-122.901458),
"/US/US-WA/cathan-22052" => array("type"=>"cdp", "pop"=>428, "name_en"=>"Cathan", "lang"=>"en", "lat"=>48.114338, "lon"=>-122.276184),
"/US/US-WA/cathlamet-22053" => array("type"=>"town", "pop"=>603, "name_en"=>"Cathlamet", "lang"=>"en", "lat"=>46.200159, "lon"=>-123.379596),
"/US/US-WA/central-park-22055" => array("type"=>"cdp", "pop"=>2669, "name_en"=>"Central Park", "lang"=>"en", "lat"=>46.970578, "lon"=>-123.701510),
"/US/US-WA/centralia-22054" => array("type"=>"city", "pop"=>15710, "name_ru"=>"Сентралия", "name_en"=>"Centralia", "lang"=>"en", "lat"=>46.721220, "lon"=>-122.962046),
"/US/US-WA/chehalis-22056" => array("type"=>"city", "pop"=>7361, "name_en"=>"Chehalis", "lang"=>"en", "lat"=>46.664851, "lon"=>-122.965045),
"/US/US-WA/chehalis-village-22057" => array("type"=>"cdp", "pop"=>282, "name_en"=>"Chehalis Village", "lang"=>"en", "lat"=>46.803634, "lon"=>-123.167885),
"/US/US-WA/chelan-22058" => array("type"=>"city", "pop"=>3993, "name_en"=>"Chelan", "lang"=>"en", "lat"=>47.846314, "lon"=>-120.027723),
"/US/US-WA/cheney-22059" => array("type"=>"city", "pop"=>10385, "name_ru"=>"Чини", "name_en"=>"Cheney", "lang"=>"en", "lat"=>47.490607, "lon"=>-117.579545),
"/US/US-WA/chewelah-22060" => array("type"=>"city", "pop"=>2300, "name_en"=>"Chewelah", "lang"=>"en", "lat"=>48.285980, "lon"=>-117.729702),
"/US/US-WA/clarkston-22061" => array("type"=>"city", "pop"=>7207, "name_ru"=>"Кларкстон", "name_en"=>"Clarkston", "lang"=>"en", "lat"=>46.415491, "lon"=>-117.049047),
"/US/US-WA/clarkston-heights-vineland-22062" => array("type"=>"cdp", "pop"=>2832, "name_en"=>"Clarkston Heights-Vineland", "lang"=>"en", "lat"=>46.386261, "lon"=>-117.076601),
"/US/US-WA/cle-elum-22063" => array("type"=>"city", "pop"=>1918, "name_en"=>"Cle Elum", "lang"=>"en", "lat"=>47.193994, "lon"=>-120.933059),
"/US/US-WA/clinton-22064" => array("type"=>"cdp", "pop"=>1564, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>47.978904, "lon"=>-122.358111),
"/US/US-WA/clyde-hill-22065" => array("type"=>"town", "pop"=>2725, "name_en"=>"Clyde Hill", "lang"=>"en", "lat"=>47.629195, "lon"=>-122.217769),
"/US/US-WA/colfax-22066" => array("type"=>"city", "pop"=>2754, "name_ru"=>"Колфакс", "name_en"=>"Colfax", "lang"=>"en", "lat"=>46.884146, "lon"=>-117.362652),
"/US/US-WA/college-place-22067" => array("type"=>"city", "pop"=>8978, "name_en"=>"College Place", "lang"=>"en", "lat"=>46.044155, "lon"=>-118.383622),
"/US/US-WA/colton-22068" => array("type"=>"town", "pop"=>394, "name_ru"=>"Колтон", "name_en"=>"Colton", "lang"=>"en", "lat"=>46.568433, "lon"=>-117.126980),
"/US/US-WA/colville-22069" => array("type"=>"city", "pop"=>4929, "name_en"=>"Colville", "lang"=>"en", "lat"=>48.543270, "lon"=>-117.895114),
"/US/US-WA/conconully-22070" => array("type"=>"town", "pop"=>182, "name_en"=>"Conconully", "lang"=>"en", "lat"=>48.558848, "lon"=>-119.750956),
"/US/US-WA/concrete-22071" => array("type"=>"town", "pop"=>864, "name_en"=>"Concrete", "lang"=>"en", "lat"=>48.537130, "lon"=>-121.748882),
"/US/US-WA/connell-22072" => array("type"=>"city", "pop"=>3190, "name_en"=>"Connell", "lang"=>"en", "lat"=>46.659967, "lon"=>-118.861026),
"/US/US-WA/cosmopolis-22073" => array("type"=>"city", "pop"=>1661, "name_en"=>"Cosmopolis", "lang"=>"en", "lat"=>46.955544, "lon"=>-123.769849),
"/US/US-WA/coulee-city-22074" => array("type"=>"town", "pop"=>632, "name_en"=>"Coulee City", "lang"=>"en", "lat"=>47.611852, "lon"=>-119.289765),
"/US/US-WA/coulee-dam-22075" => array("type"=>"town", "pop"=>1050, "name_en"=>"Coulee Dam", "lang"=>"en", "lat"=>47.968743, "lon"=>-118.974446),
"/US/US-WA/country-homes-22076" => array("type"=>"cdp", "pop"=>5126, "name_en"=>"Country Homes", "lang"=>"en", "lat"=>47.749000, "lon"=>-117.422880),
"/US/US-WA/coupeville-22077" => array("type"=>"town", "pop"=>1888, "name_en"=>"Coupeville", "lang"=>"en", "lat"=>48.217279, "lon"=>-122.675759),
"/US/US-WA/covington-sawyer-wilderness-22078" => array("type"=>"cdp", "pop"=>24321, "name_en"=>"Covington-Sawyer-Wilderness", "lang"=>"en", "lat"=>47.352810, "lon"=>-122.073107),
"/US/US-WA/creston-22079" => array("type"=>"town", "pop"=>224, "name_ru"=>"Крестон", "name_en"=>"Creston", "lang"=>"en", "lat"=>47.760000, "lon"=>-118.519400),
"/US/US-WA/cusick-22080" => array("type"=>"town", "pop"=>219, "name_en"=>"Cusick", "lang"=>"en", "lat"=>48.334464, "lon"=>-117.294061),
"/US/US-WA/darrington-22081" => array("type"=>"town", "pop"=>1334, "name_en"=>"Darrington", "lang"=>"en", "lat"=>48.252429, "lon"=>-121.603906),
"/US/US-WA/davenport-22082" => array("type"=>"city", "pop"=>1703, "name_ru"=>"Дэйвенпорт", "name_en"=>"Davenport", "lang"=>"en", "lat"=>47.651280, "lon"=>-118.153173),
"/US/US-WA/dayton-22083" => array("type"=>"city", "pop"=>2617, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>46.317688, "lon"=>-117.976751),
"/US/US-WA/deer-park-22084" => array("type"=>"city", "pop"=>3500, "name_ru"=>"Дир Парк", "name_en"=>"Deer Park", "lang"=>"en", "lat"=>47.963335, "lon"=>-117.433963),
"/US/US-WA/des-moines-22085" => array("type"=>"city", "pop"=>28755, "name_ru"=>"Дес Мойн", "name_en"=>"Des Moines", "lang"=>"en", "lat"=>47.395500, "lon"=>-122.309469),
"/US/US-WA/dishman-22086" => array("type"=>"cdp", "pop"=>9671, "name_en"=>"Dishman", "lang"=>"en", "lat"=>47.658950, "lon"=>-117.275453),
"/US/US-WA/dupont-22087" => array("type"=>"city", "pop"=>6944, "name_ru"=>"Дюпон", "name_en"=>"DuPont", "lang"=>"en", "lat"=>47.106767, "lon"=>-122.653083),
"/US/US-WA/duvall-22088" => array("type"=>"city", "pop"=>5982, "name_en"=>"Duvall", "lang"=>"en", "lat"=>47.733711, "lon"=>-121.969195),
"/US/US-WA/east-hill-meridian-22090" => array("type"=>"cdp", "pop"=>42696, "name_en"=>"East Hill-Meridian", "lang"=>"en", "lat"=>47.389806, "lon"=>-122.172020),
"/US/US-WA/east-port-orchard-22091" => array("type"=>"cdp", "pop"=>5409, "name_en"=>"East Port Orchard", "lang"=>"en", "lat"=>47.521186, "lon"=>-122.623837),
"/US/US-WA/east-renton-highlands-22092" => array("type"=>"cdp", "pop"=>13218, "name_en"=>"East Renton Highlands", "lang"=>"en", "lat"=>47.485101, "lon"=>-122.111240),
"/US/US-WA/east-wenatchee-22093" => array("type"=>"city", "pop"=>12286, "name_en"=>"East Wenatchee", "lang"=>"en", "lat"=>47.417148, "lon"=>-120.288304),
"/US/US-WA/east-wenatchee-bench-22094" => array("type"=>"cdp", "pop"=>12539, "name_en"=>"East Wenatchee Bench", "lang"=>"en", "lat"=>47.425919, "lon"=>-120.280071),
"/US/US-WA/eastgate-22089" => array("type"=>"cdp", "pop"=>4434, "name_en"=>"Eastgate", "lang"=>"en", "lat"=>47.572300, "lon"=>-122.134912),
"/US/US-WA/eatonville-22095" => array("type"=>"town", "pop"=>2459, "name_ru"=>"Истонвилл", "name_en"=>"Eatonville", "lang"=>"en", "lat"=>46.868427, "lon"=>-122.268841),
"/US/US-WA/edgewood-north-hill-22096" => array("type"=>"cdp", "pop"=>9120, "name_en"=>"Edgewood-North Hill", "lang"=>"en", "lat"=>47.232150, "lon"=>-122.280493),
"/US/US-WA/edmonds-22097" => array("type"=>"city", "pop"=>40158, "name_en"=>"Edmonds", "lang"=>"en", "lat"=>47.825900, "lon"=>-122.367931),
"/US/US-WA/electric-city-22098" => array("type"=>"town", "pop"=>988, "name_en"=>"Electric City", "lang"=>"en", "lat"=>47.930158, "lon"=>-119.036197),
"/US/US-WA/elk-plain-22099" => array("type"=>"cdp", "pop"=>12197, "name_en"=>"Elk Plain", "lang"=>"en", "lat"=>47.053348, "lon"=>-122.378041),
"/US/US-WA/ellensburg-22100" => array("type"=>"city", "pop"=>17141, "name_en"=>"Ellensburg", "lang"=>"en", "lat"=>46.997147, "lon"=>-120.548004),
"/US/US-WA/ellsworth-north-22101" => array("type"=>"cdp", "pop"=>5796, "name_en"=>"Ellsworth North", "lang"=>"en", "lat"=>45.633035, "lon"=>-122.571238),
"/US/US-WA/ellsworth-south-22102" => array("type"=>"cdp", "pop"=>4423, "name_en"=>"Ellsworth South", "lang"=>"en", "lat"=>45.609906, "lon"=>-122.567047),
"/US/US-WA/elma-22103" => array("type"=>"city", "pop"=>3113, "name_ru"=>"Элма", "name_en"=>"Elma", "lang"=>"en", "lat"=>47.005061, "lon"=>-123.400855),
"/US/US-WA/elmer-city-22104" => array("type"=>"town", "pop"=>251, "name_en"=>"Elmer City", "lang"=>"en", "lat"=>47.999026, "lon"=>-118.952392),
"/US/US-WA/endicott-22105" => array("type"=>"town", "pop"=>314, "name_en"=>"Endicott", "lang"=>"en", "lat"=>46.928877, "lon"=>-117.684994),
"/US/US-WA/entiat-22106" => array("type"=>"city", "pop"=>1043, "name_en"=>"Entiat", "lang"=>"en", "lat"=>47.675427, "lon"=>-120.211609),
"/US/US-WA/enumclaw-22107" => array("type"=>"city", "pop"=>10684, "name_en"=>"Enumclaw", "lang"=>"en", "lat"=>47.200875, "lon"=>-121.989087),
"/US/US-WA/ephrata-22108" => array("type"=>"city", "pop"=>7344, "name_en"=>"Ephrata", "lang"=>"en", "lat"=>47.312337, "lon"=>-119.536912),
"/US/US-WA/erlands-point-kitsap-lake-22109" => array("type"=>"cdp", "pop"=>2764, "name_en"=>"Erlands Point-Kitsap Lake", "lang"=>"en", "lat"=>47.596301, "lon"=>-122.702957),
"/US/US-WA/esperance-22110" => array("type"=>"cdp", "pop"=>11236, "name_en"=>"Esperance", "lang"=>"en", "lat"=>47.789052, "lon"=>-122.354302),
"/US/US-WA/everett-22111" => array("type"=>"city", "pop"=>98212, "name_ru"=>"Эверетт", "name_en"=>"Everett", "lang"=>"en", "lat"=>47.963257, "lon"=>-122.198590),
"/US/US-WA/evergreen-22112" => array("type"=>"cdp", "pop"=>11249, "name_ru"=>"Эвергрин", "name_en"=>"Evergreen", "lang"=>"en", "lat"=>45.627288, "lon"=>-122.527347),
"/US/US-WA/everson-22113" => array("type"=>"city", "pop"=>2192, "name_en"=>"Everson", "lang"=>"en", "lat"=>48.914812, "lon"=>-122.350878),
"/US/US-WA/fairchild-afb-22114" => array("type"=>"cdp", "pop"=>4854, "name_en"=>"Fairchild AFB", "lang"=>"en", "lat"=>47.613068, "lon"=>-117.643746),
"/US/US-WA/fairfield-22115" => array("type"=>"town", "pop"=>582, "name_ru"=>"Фэрфилд", "name_en"=>"Fairfield", "lang"=>"en", "lat"=>47.385201, "lon"=>-117.174045),
"/US/US-WA/fairview-sumach-22116" => array("type"=>"cdp", "pop"=>2749, "name_en"=>"Fairview-Sumach", "lang"=>"en", "lat"=>46.587753, "lon"=>-120.474570),
"/US/US-WA/fairwood-22117" => array("type"=>"cdp", "pop"=>5807, "name_en"=>"Fairwood", "lang"=>"en", "lat"=>47.767050, "lon"=>-117.416348),
"/US/US-WA/fall-city-22118" => array("type"=>"cdp", "pop"=>1582, "name_en"=>"Fall City", "lang"=>"en", "lat"=>47.566303, "lon"=>-121.902141),
"/US/US-WA/farmington-22119" => array("type"=>"town", "pop"=>140, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>47.088949, "lon"=>-117.045089),
"/US/US-WA/federal-way-22120" => array("type"=>"cdp", "pop"=>84309, "name_en"=>"Federal Way", "lang"=>"en", "lat"=>47.311870, "lon"=>-122.340805),
"/US/US-WA/felida-22121" => array("type"=>"cdp", "pop"=>3109, "name_en"=>"Felida", "lang"=>"en", "lat"=>45.716808, "lon"=>-122.708186),
"/US/US-WA/ferndale-22122" => array("type"=>"city", "pop"=>11359, "name_ru"=>"Ферндейл", "name_en"=>"Ferndale", "lang"=>"en", "lat"=>48.850936, "lon"=>-122.587218),
"/US/US-WA/fife-22123" => array("type"=>"city", "pop"=>8114, "name_en"=>"Fife", "lang"=>"en", "lat"=>47.239954, "lon"=>-122.361355),
"/US/US-WA/finley-22124" => array("type"=>"cdp", "pop"=>4897, "name_en"=>"Finley", "lang"=>"en", "lat"=>46.168971, "lon"=>-119.042191),
"/US/US-WA/fircrest-22125" => array("type"=>"town", "pop"=>6213, "name_en"=>"Fircrest", "lang"=>"en", "lat"=>47.231150, "lon"=>-122.512974),
"/US/US-WA/five-corners-22126" => array("type"=>"cdp", "pop"=>6776, "name_en"=>"Five Corners", "lang"=>"en", "lat"=>45.684750, "lon"=>-122.573913),
"/US/US-WA/fords-prairie-22127" => array("type"=>"cdp", "pop"=>2480, "name_en"=>"Fords Prairie", "lang"=>"en", "lat"=>46.740084, "lon"=>-123.001495),
"/US/US-WA/forks-22128" => array("type"=>"city", "pop"=>3275, "name_en"=>"Forks", "lang"=>"en", "lat"=>47.947054, "lon"=>-124.389933),
"/US/US-WA/fort-lewis-22129" => array("type"=>"cdp", "pop"=>22224, "name_en"=>"Fort Lewis", "lang"=>"en", "lat"=>47.106133, "lon"=>-122.581853),
"/US/US-WA/fox-island-22130" => array("type"=>"cdp", "pop"=>2017, "name_en"=>"Fox Island", "lang"=>"en", "lat"=>47.251800, "lon"=>-122.627835),
"/US/US-WA/frederickson-22131" => array("type"=>"cdp", "pop"=>3502, "name_en"=>"Frederickson", "lang"=>"en", "lat"=>47.086482, "lon"=>-122.361868),
"/US/US-WA/freeland-22132" => array("type"=>"cdp", "pop"=>1278, "name_ru"=>"Фриленд", "name_en"=>"Freeland", "lang"=>"en", "lat"=>48.013051, "lon"=>-122.537737),
"/US/US-WA/friday-harbor-22133" => array("type"=>"town", "pop"=>2130, "name_en"=>"Friday Harbor", "lang"=>"en", "lat"=>48.536920, "lon"=>-123.031320),
"/US/US-WA/fruitvale-22134" => array("type"=>"cdp", "pop"=>4125, "name_ru"=>"Фрутвейл", "name_en"=>"Fruitvale", "lang"=>"en", "lat"=>46.617181, "lon"=>-120.541693),
"/US/US-WA/garfield-22135" => array("type"=>"town", "pop"=>616, "name_ru"=>"Гарфилд", "name_en"=>"Garfield", "lang"=>"en", "lat"=>47.008383, "lon"=>-117.140731),
"/US/US-WA/garrett-22136" => array("type"=>"cdp", "pop"=>1004, "name_ru"=>"Гарретт", "name_en"=>"Garrett", "lang"=>"en", "lat"=>46.061372, "lon"=>-118.383311),
"/US/US-WA/george-22137" => array("type"=>"town", "pop"=>571, "name_ru"=>"Джордж", "name_en"=>"George", "lang"=>"en", "lat"=>47.078590, "lon"=>-119.855968),
"/US/US-WA/gig-harbor-22138" => array("type"=>"city", "pop"=>7008, "name_en"=>"Gig Harbor", "lang"=>"en", "lat"=>47.328888, "lon"=>-122.587872),
"/US/US-WA/gold-bar-22139" => array("type"=>"town", "pop"=>2373, "name_en"=>"Gold Bar", "lang"=>"en", "lat"=>47.857745, "lon"=>-121.695701),
"/US/US-WA/goldendale-22140" => array("type"=>"city", "pop"=>3742, "name_en"=>"Goldendale", "lang"=>"en", "lat"=>45.822024, "lon"=>-120.818317),
"/US/US-WA/grand-coulee-22141" => array("type"=>"city", "pop"=>928, "name_en"=>"Grand Coulee", "lang"=>"en", "lat"=>47.939565, "lon"=>-119.003947),
"/US/US-WA/grand-mound-22142" => array("type"=>"cdp", "pop"=>1394, "name_ru"=>"Гранд Маунд", "name_en"=>"Grand Mound", "lang"=>"en", "lat"=>46.809985, "lon"=>-123.012697),
"/US/US-WA/grandview-22143" => array("type"=>"city", "pop"=>9392, "name_ru"=>"Грандвью", "name_en"=>"Grandview", "lang"=>"en", "lat"=>46.244358, "lon"=>-119.910846),
"/US/US-WA/granger-22144" => array("type"=>"town", "pop"=>2979, "name_ru"=>"Грейнджер", "name_en"=>"Granger", "lang"=>"en", "lat"=>46.342805, "lon"=>-120.189622),
"/US/US-WA/granite-falls-22145" => array("type"=>"town", "pop"=>3088, "name_ru"=>"Гранти Фоллс", "name_en"=>"Granite Falls", "lang"=>"en", "lat"=>48.082862, "lon"=>-121.967090),
"/US/US-WA/green-acres-22146" => array("type"=>"cdp", "pop"=>4626, "name_en"=>"Green Acres", "lang"=>"en", "lat"=>47.663550, "lon"=>-117.160949),
"/US/US-WA/hadlock-irondale-22147" => array("type"=>"cdp", "pop"=>2742, "name_en"=>"Hadlock-Irondale", "lang"=>"en", "lat"=>48.049035, "lon"=>-122.771445),
"/US/US-WA/hamilton-22148" => array("type"=>"town", "pop"=>354, "name_ru"=>"Гамильтон", "name_en"=>"Hamilton", "lang"=>"en", "lat"=>48.523328, "lon"=>-121.988211),
"/US/US-WA/harbour-pointe-22149" => array("type"=>"cdp", "pop"=>9107, "name_en"=>"Harbour Pointe", "lang"=>"en", "lat"=>47.885972, "lon"=>-122.303982),
"/US/US-WA/harrah-22150" => array("type"=>"town", "pop"=>503, "name_en"=>"Harrah", "lang"=>"en", "lat"=>46.404333, "lon"=>-120.541310),
"/US/US-WA/harrington-22151" => array("type"=>"city", "pop"=>408, "name_ru"=>"Харрингтон", "name_en"=>"Harrington", "lang"=>"en", "lat"=>47.480374, "lon"=>-118.254353),
"/US/US-WA/hartline-22152" => array("type"=>"town", "pop"=>143, "name_en"=>"Hartline", "lang"=>"en", "lat"=>47.689432, "lon"=>-119.105930),
"/US/US-WA/hatton-22153" => array("type"=>"town", "pop"=>98, "name_en"=>"Hatton", "lang"=>"en", "lat"=>46.775518, "lon"=>-118.827880),
"/US/US-WA/hazel-dell-north-22154" => array("type"=>"cdp", "pop"=>6924, "name_en"=>"Hazel Dell North", "lang"=>"en", "lat"=>45.687050, "lon"=>-122.647999),
"/US/US-WA/hazel-dell-south-22155" => array("type"=>"cdp", "pop"=>5796, "name_en"=>"Hazel Dell South", "lang"=>"en", "lat"=>45.670000, "lon"=>-122.659497),
"/US/US-WA/highland-22156" => array("type"=>"cdp", "pop"=>3656, "name_ru"=>"Хайленд", "name_en"=>"Highland", "lang"=>"en", "lat"=>46.131633, "lon"=>-119.113005),
"/US/US-WA/hoquiam-22157" => array("type"=>"city", "pop"=>8848, "name_en"=>"Hoquiam", "lang"=>"en", "lat"=>46.975413, "lon"=>-123.886641),
"/US/US-WA/hunts-point-22158" => array("type"=>"town", "pop"=>472, "name_en"=>"Hunts Point", "lang"=>"en", "lat"=>47.638850, "lon"=>-122.227481),
"/US/US-WA/ilwaco-22159" => array("type"=>"city", "pop"=>986, "name_en"=>"Ilwaco", "lang"=>"en", "lat"=>46.310898, "lon"=>-124.033053),
"/US/US-WA/inchelium-22160" => array("type"=>"cdp", "pop"=>393, "name_en"=>"Inchelium", "lang"=>"en", "lat"=>48.341189, "lon"=>-118.247355),
"/US/US-WA/index-22161" => array("type"=>"town", "pop"=>166, "name_en"=>"Index", "lang"=>"en", "lat"=>47.820708, "lon"=>-121.552392),
"/US/US-WA/indianola-22162" => array("type"=>"cdp", "pop"=>1729, "name_ru"=>"Индианола", "name_en"=>"Indianola", "lang"=>"en", "lat"=>47.754500, "lon"=>-122.513271),
"/US/US-WA/inglewood-finn-hill-22163" => array("type"=>"cdp", "pop"=>29132, "name_en"=>"Inglewood-Finn Hill", "lang"=>"en", "lat"=>47.725997, "lon"=>-122.234982),
"/US/US-WA/ione-22164" => array("type"=>"town", "pop"=>496, "name_ru"=>"Ен", "name_en"=>"Ione", "lang"=>"en", "lat"=>48.740607, "lon"=>-117.420904),
"/US/US-WA/issaquah-22165" => array("type"=>"city", "pop"=>24057, "name_en"=>"Issaquah", "lang"=>"en", "lat"=>47.532300, "lon"=>-122.038487),
"/US/US-WA/john-sam-lake-22166" => array("type"=>"cdp", "pop"=>432, "name_en"=>"John Sam Lake", "lang"=>"en", "lat"=>48.109432, "lon"=>-122.244624),
"/US/US-WA/kahlotus-22167" => array("type"=>"city", "pop"=>236, "name_en"=>"Kahlotus", "lang"=>"en", "lat"=>46.643623, "lon"=>-118.552398),
"/US/US-WA/kalama-22168" => array("type"=>"city", "pop"=>2249, "name_en"=>"Kalama", "lang"=>"en", "lat"=>46.010286, "lon"=>-122.840868),
"/US/US-WA/kelso-22169" => array("type"=>"city", "pop"=>12188, "name_ru"=>"Келсо", "name_en"=>"Kelso", "lang"=>"en", "lat"=>46.126449, "lon"=>-122.891370),
"/US/US-WA/kenmore-22170" => array("type"=>"cdp", "pop"=>20374, "name_en"=>"Kenmore", "lang"=>"en", "lat"=>47.764631, "lon"=>-122.242617),
"/US/US-WA/kennewick-22171" => array("type"=>"city", "pop"=>63216, "name_en"=>"Kennewick", "lang"=>"en", "lat"=>46.203051, "lon"=>-119.170380),
"/US/US-WA/kent-22172" => array("type"=>"city", "pop"=>83978, "name_ru"=>"Кент", "name_en"=>"Kent", "lang"=>"en", "lat"=>47.394300, "lon"=>-122.238533),
"/US/US-WA/kettle-falls-22173" => array("type"=>"city", "pop"=>1425, "name_en"=>"Kettle Falls", "lang"=>"en", "lat"=>48.605998, "lon"=>-118.058909),
"/US/US-WA/kingsgate-22174" => array("type"=>"cdp", "pop"=>14259, "name_en"=>"Kingsgate", "lang"=>"en", "lat"=>47.730599, "lon"=>-122.167148),
"/US/US-WA/kingston-22175" => array("type"=>"cdp", "pop"=>1270, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>47.797810, "lon"=>-122.496656),
"/US/US-WA/kirkland-22176" => array("type"=>"city", "pop"=>47303, "name_ru"=>"Киркленд", "name_en"=>"Kirkland", "lang"=>"en", "lat"=>47.684600, "lon"=>-122.193939),
"/US/US-WA/kittitas-22177" => array("type"=>"city", "pop"=>1231, "name_en"=>"Kittitas", "lang"=>"en", "lat"=>46.985001, "lon"=>-120.417041),
"/US/US-WA/krupp-22178" => array("type"=>"town", "pop"=>62, "name_en"=>"Krupp", "lang"=>"en", "lat"=>47.409892, "lon"=>-118.987818),
"/US/US-WA/la-center-22179" => array("type"=>"town", "pop"=>2545, "name_ru"=>"Ла Сентер", "name_en"=>"La Center", "lang"=>"en", "lat"=>45.864160, "lon"=>-122.666474),
"/US/US-WA/la-conner-22181" => array("type"=>"town", "pop"=>788, "name_en"=>"La Conner", "lang"=>"en", "lat"=>48.392750, "lon"=>-122.493176),
"/US/US-WA/la-crosse-22182" => array("type"=>"town", "pop"=>348, "name_ru"=>"Ла Кросс", "name_en"=>"La Crosse", "lang"=>"en", "lat"=>46.814411, "lon"=>-117.879048),
"/US/US-WA/lacey-22180" => array("type"=>"city", "pop"=>40702, "name_en"=>"Lacey", "lang"=>"en", "lat"=>47.033050, "lon"=>-122.805796),
"/US/US-WA/lake-forest-north-22183" => array("type"=>"cdp", "pop"=>8002, "name_en"=>"Lake Forest North", "lang"=>"en", "lat"=>47.767750, "lon"=>-122.285552),
"/US/US-WA/lake-forest-park-22184" => array("type"=>"city", "pop"=>12438, "name_en"=>"Lake Forest Park", "lang"=>"en", "lat"=>47.762700, "lon"=>-122.286208),
"/US/US-WA/lake-goodwin-22185" => array("type"=>"cdp", "pop"=>2437, "name_en"=>"Lake Goodwin", "lang"=>"en", "lat"=>48.139103, "lon"=>-122.278448),
"/US/US-WA/lake-serene-north-lynnwood-22188" => array("type"=>"cdp", "pop"=>14290, "name_en"=>"Lake Serene-North Lynnwood", "lang"=>"en", "lat"=>47.855033, "lon"=>-122.284073),
"/US/US-WA/lake-shore-22189" => array("type"=>"cdp", "pop"=>6268, "name_ru"=>"Лейк Шор", "name_en"=>"Lake Shore", "lang"=>"en", "lat"=>45.690800, "lon"=>-122.689816),
"/US/US-WA/lake-stevens-22190" => array("type"=>"city", "pop"=>13599, "name_en"=>"Lake Stevens", "lang"=>"en", "lat"=>48.017059, "lon"=>-122.065293),
"/US/US-WA/lakeland-north-22186" => array("type"=>"cdp", "pop"=>14402, "name_en"=>"Lakeland North", "lang"=>"en", "lat"=>47.329200, "lon"=>-122.277980),
"/US/US-WA/lakeland-south-22187" => array("type"=>"cdp", "pop"=>9027, "name_en"=>"Lakeland South", "lang"=>"en", "lat"=>47.277440, "lon"=>-122.282757),
"/US/US-WA/lakewood-22191" => array("type"=>"cdp", "pop"=>56983, "name_ru"=>"Лейквуд", "name_en"=>"Lakewood", "lang"=>"en", "lat"=>47.161850, "lon"=>-122.530180),
"/US/US-WA/lamont-22192" => array("type"=>"town", "pop"=>97, "name_ru"=>"Ламонт", "name_en"=>"Lamont", "lang"=>"en", "lat"=>47.200574, "lon"=>-117.903459),
"/US/US-WA/langley-22193" => array("type"=>"city", "pop"=>1072, "name_en"=>"Langley", "lang"=>"en", "lat"=>48.037106, "lon"=>-122.407244),
"/US/US-WA/latah-22194" => array("type"=>"town", "pop"=>150, "name_en"=>"Latah", "lang"=>"en", "lat"=>47.281828, "lon"=>-117.154307),
"/US/US-WA/lea-hill-22195" => array("type"=>"cdp", "pop"=>6876, "name_en"=>"Lea Hill", "lang"=>"en", "lat"=>47.326250, "lon"=>-122.180319),
"/US/US-WA/leavenworth-22196" => array("type"=>"city", "pop"=>2324, "name_ru"=>"Ливенуорт", "name_en"=>"Leavenworth", "lang"=>"en", "lat"=>47.591317, "lon"=>-120.664248),
"/US/US-WA/liberty-lake-22197" => array("type"=>"cdp", "pop"=>7018, "name_en"=>"Liberty Lake", "lang"=>"en", "lat"=>47.654412, "lon"=>-117.082479),
"/US/US-WA/lind-22198" => array("type"=>"town", "pop"=>581, "name_en"=>"Lind", "lang"=>"en", "lat"=>46.971648, "lon"=>-118.613570),
"/US/US-WA/long-beach-22199" => array("type"=>"city", "pop"=>1360, "name_ru"=>"Лонг Бич", "name_en"=>"Long Beach", "lang"=>"en", "lat"=>46.355466, "lon"=>-124.053176),
"/US/US-WA/longview-22200" => array("type"=>"city", "pop"=>36562, "name_ru"=>"Лонгвью", "name_en"=>"Longview", "lang"=>"en", "lat"=>46.146499, "lon"=>-122.954952),
"/US/US-WA/longview-heights-22201" => array("type"=>"cdp", "pop"=>3310, "name_en"=>"Longview Heights", "lang"=>"en", "lat"=>46.180499, "lon"=>-122.955866),
"/US/US-WA/lyman-22202" => array("type"=>"town", "pop"=>421, "name_ru"=>"Лиман", "name_en"=>"Lyman", "lang"=>"en", "lat"=>48.523456, "lon"=>-122.063621),
"/US/US-WA/lynden-22203" => array("type"=>"city", "pop"=>11737, "name_en"=>"Lynden", "lang"=>"en", "lat"=>48.948440, "lon"=>-122.456110),
"/US/US-WA/lynnwood-22204" => array("type"=>"city", "pop"=>33572, "name_en"=>"Lynnwood", "lang"=>"en", "lat"=>47.827250, "lon"=>-122.304263),
"/US/US-WA/mabton-22205" => array("type"=>"town", "pop"=>2067, "name_en"=>"Mabton", "lang"=>"en", "lat"=>46.212893, "lon"=>-119.995625),
"/US/US-WA/malden-22208" => array("type"=>"town", "pop"=>196, "name_ru"=>"Малден", "name_en"=>"Malden", "lang"=>"en", "lat"=>47.231296, "lon"=>-117.471884),
"/US/US-WA/manchester-22209" => array("type"=>"cdp", "pop"=>4031, "name_ru"=>"Манчестер", "name_en"=>"Manchester", "lang"=>"en", "lat"=>47.552126, "lon"=>-122.541727),
"/US/US-WA/mansfield-22210" => array("type"=>"town", "pop"=>341, "name_ru"=>"Мансфилд", "name_en"=>"Mansfield", "lang"=>"en", "lat"=>47.811513, "lon"=>-119.636993),
"/US/US-WA/maple-valley-22211" => array("type"=>"cdp", "pop"=>19929, "name_en"=>"Maple Valley", "lang"=>"en", "lat"=>47.401306, "lon"=>-122.025698),
"/US/US-WA/marcus-22212" => array("type"=>"town", "pop"=>162, "name_ru"=>"Маркус", "name_en"=>"Marcus", "lang"=>"en", "lat"=>48.663203, "lon"=>-118.057927),
"/US/US-WA/marietta-alderwood-22213" => array("type"=>"cdp", "pop"=>2766, "name_en"=>"Marietta-Alderwood", "lang"=>"en", "lat"=>48.788978, "lon"=>-122.554795),
"/US/US-WA/martha-lake-22214" => array("type"=>"cdp", "pop"=>10155, "name_en"=>"Martha Lake", "lang"=>"en", "lat"=>47.850983, "lon"=>-122.238054),
"/US/US-WA/marysville-22215" => array("type"=>"city", "pop"=>34482, "name_ru"=>"Мэрисвилл", "name_en"=>"Marysville", "lang"=>"en", "lat"=>48.059513, "lon"=>-122.154710),
"/US/US-WA/mattawa-22216" => array("type"=>"town", "pop"=>3143, "name_en"=>"Mattawa", "lang"=>"en", "lat"=>46.736456, "lon"=>-119.900829),
"/US/US-WA/mcchord-afb-22206" => array("type"=>"cdp", "pop"=>4538, "name_en"=>"McChord AFB", "lang"=>"en", "lat"=>47.133767, "lon"=>-122.496525),
"/US/US-WA/mccleary-22207" => array("type"=>"town", "pop"=>1604, "name_en"=>"McCleary", "lang"=>"en", "lat"=>47.055251, "lon"=>-123.267610),
"/US/US-WA/meadow-glade-22217" => array("type"=>"cdp", "pop"=>1584, "name_en"=>"Meadow Glade", "lang"=>"en", "lat"=>45.754578, "lon"=>-122.554009),
"/US/US-WA/medical-lake-22218" => array("type"=>"city", "pop"=>4710, "name_en"=>"Medical Lake", "lang"=>"en", "lat"=>47.570043, "lon"=>-117.692149),
"/US/US-WA/medina-22219" => array("type"=>"city", "pop"=>3144, "name_ru"=>"Медина", "name_en"=>"Medina", "lang"=>"en", "lat"=>47.616309, "lon"=>-122.236524),
"/US/US-WA/mercer-island-22220" => array("type"=>"city", "pop"=>24163, "name_en"=>"Mercer Island", "lang"=>"en", "lat"=>47.569267, "lon"=>-122.230917),
"/US/US-WA/mesa-22221" => array("type"=>"town", "pop"=>443, "name_ru"=>"Меса", "name_en"=>"Mesa", "lang"=>"en", "lat"=>46.573195, "lon"=>-118.999560),
"/US/US-WA/metaline-22222" => array("type"=>"town", "pop"=>168, "name_en"=>"Metaline", "lang"=>"en", "lat"=>48.854567, "lon"=>-117.388110),
"/US/US-WA/metaline-falls-22223" => array("type"=>"town", "pop"=>230, "name_en"=>"Metaline Falls", "lang"=>"en", "lat"=>48.862383, "lon"=>-117.369394),
"/US/US-WA/midland-22224" => array("type"=>"cdp", "pop"=>5587, "name_ru"=>"Мидленд", "name_en"=>"Midland", "lang"=>"en", "lat"=>47.172300, "lon"=>-122.408899),
"/US/US-WA/mill-creek-22225" => array("type"=>"city", "pop"=>16907, "name_ru"=>"Милл Крик", "name_en"=>"Mill Creek", "lang"=>"en", "lat"=>47.857715, "lon"=>-122.205846),
"/US/US-WA/millwood-22226" => array("type"=>"town", "pop"=>1613, "name_en"=>"Millwood", "lang"=>"en", "lat"=>47.685700, "lon"=>-117.279336),
"/US/US-WA/milton-22227" => array("type"=>"city", "pop"=>6786, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>47.251900, "lon"=>-122.315931),
"/US/US-WA/minnehaha-22228" => array("type"=>"cdp", "pop"=>9661, "name_en"=>"Minnehaha", "lang"=>"en", "lat"=>45.656950, "lon"=>-122.626348),
"/US/US-WA/mirrormont-22229" => array("type"=>"cdp", "pop"=>2360, "name_en"=>"Mirrormont", "lang"=>"en", "lat"=>47.462620, "lon"=>-121.994330),
"/US/US-WA/monroe-22230" => array("type"=>"city", "pop"=>17059, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>47.858084, "lon"=>-121.981776),
"/US/US-WA/montesano-22231" => array("type"=>"city", "pop"=>3599, "name_en"=>"Montesano", "lang"=>"en", "lat"=>47.014631, "lon"=>-123.583917),
"/US/US-WA/morton-22232" => array("type"=>"city", "pop"=>1086, "name_ru"=>"Мортон", "name_en"=>"Morton", "lang"=>"en", "lat"=>46.558033, "lon"=>-122.280075),
"/US/US-WA/moses-lake-22233" => array("type"=>"city", "pop"=>18800, "name_en"=>"Moses Lake", "lang"=>"en", "lat"=>47.115666, "lon"=>-119.286163),
"/US/US-WA/moses-lake-north-22234" => array("type"=>"cdp", "pop"=>3677, "name_en"=>"Moses Lake North", "lang"=>"en", "lat"=>47.195201, "lon"=>-119.322686),
"/US/US-WA/mossyrock-22235" => array("type"=>"city", "pop"=>503, "name_en"=>"Mossyrock", "lang"=>"en", "lat"=>46.530096, "lon"=>-122.481307),
"/US/US-WA/mount-vernon-22237" => array("type"=>"city", "pop"=>31513, "name_ru"=>"Маунт Вернон", "name_en"=>"Mount Vernon", "lang"=>"en", "lat"=>48.421735, "lon"=>-122.314498),
"/US/US-WA/mountlake-terrace-22236" => array("type"=>"city", "pop"=>19912, "name_en"=>"Mountlake Terrace", "lang"=>"en", "lat"=>47.791200, "lon"=>-122.306649),
"/US/US-WA/moxee-22238" => array("type"=>"city", "pop"=>2330, "name_en"=>"Moxee", "lang"=>"en", "lat"=>46.563403, "lon"=>-120.398767),
"/US/US-WA/mukilteo-22239" => array("type"=>"city", "pop"=>20742, "name_en"=>"Mukilteo", "lang"=>"en", "lat"=>47.926831, "lon"=>-122.308686),
"/US/US-WA/naches-22240" => array("type"=>"town", "pop"=>718, "name_en"=>"Naches", "lang"=>"en", "lat"=>46.729495, "lon"=>-120.697333),
"/US/US-WA/napavine-22241" => array("type"=>"city", "pop"=>1611, "name_en"=>"Napavine", "lang"=>"en", "lat"=>46.577064, "lon"=>-122.908503),
"/US/US-WA/navy-yard-city-22242" => array("type"=>"cdp", "pop"=>2905, "name_en"=>"Navy Yard City", "lang"=>"en", "lat"=>47.551600, "lon"=>-122.666954),
"/US/US-WA/neah-bay-22243" => array("type"=>"cdp", "pop"=>916, "name_en"=>"Neah Bay", "lang"=>"en", "lat"=>48.364637, "lon"=>-124.610967),
"/US/US-WA/nespelem-22244" => array("type"=>"town", "pop"=>201, "name_en"=>"Nespelem", "lang"=>"en", "lat"=>48.167659, "lon"=>-118.971419),
"/US/US-WA/nespelem-community-22245" => array("type"=>"cdp", "pop"=>291, "name_en"=>"Nespelem Community", "lang"=>"en", "lat"=>48.167123, "lon"=>-119.019094),
"/US/US-WA/newport-22246" => array("type"=>"city", "pop"=>2126, "name_ru"=>"Ньюпорт", "name_en"=>"Newport", "lang"=>"en", "lat"=>48.180752, "lon"=>-117.050762),
"/US/US-WA/newport-hills-22247" => array("type"=>"cdp", "pop"=>14736, "name_en"=>"Newport Hills", "lang"=>"en", "lat"=>47.547527, "lon"=>-122.173972),
"/US/US-WA/nisqually-indian-community-22248" => array("type"=>"cdp", "pop"=>558, "name_en"=>"Nisqually Indian Community", "lang"=>"en", "lat"=>47.005854, "lon"=>-122.669829),
"/US/US-WA/nooksack-22249" => array("type"=>"city", "pop"=>1049, "name_en"=>"Nooksack", "lang"=>"en", "lat"=>48.928389, "lon"=>-122.321246),
"/US/US-WA/normandy-park-22250" => array("type"=>"city", "pop"=>6240, "name_en"=>"Normandy Park", "lang"=>"en", "lat"=>47.444750, "lon"=>-122.355792),
"/US/US-WA/north-bend-22251" => array("type"=>"city", "pop"=>4635, "name_en"=>"North Bend", "lang"=>"en", "lat"=>47.495719, "lon"=>-121.788722),
"/US/US-WA/north-bonneville-22252" => array("type"=>"city", "pop"=>955, "name_en"=>"North Bonneville", "lang"=>"en", "lat"=>45.642921, "lon"=>-121.969488),
"/US/US-WA/north-city-ridgecrest-22253" => array("type"=>"cdp", "pop"=>13832, "name_en"=>"North City-Ridgecrest", "lang"=>"en", "lat"=>47.750900, "lon"=>-122.313074),
"/US/US-WA/north-creek-canyon-park-22254" => array("type"=>"cdp", "pop"=>23236, "name_en"=>"North Creek-Canyon Park", "lang"=>"en", "lat"=>47.810567, "lon"=>-122.186380),
"/US/US-WA/north-hill-22255" => array("type"=>"cdp", "pop"=>5706, "name_en"=>"North Hill", "lang"=>"en", "lat"=>47.419350, "lon"=>-122.322328),
"/US/US-WA/north-marysville-22256" => array("type"=>"cdp", "pop"=>18711, "name_en"=>"North Marysville", "lang"=>"en", "lat"=>48.099435, "lon"=>-122.147598),
"/US/US-WA/north-omak-22257" => array("type"=>"cdp", "pop"=>515, "name_en"=>"North Omak", "lang"=>"en", "lat"=>48.445380, "lon"=>-119.443360),
"/US/US-WA/north-puyallup-22259" => array("type"=>"cdp", "pop"=>2886, "name_en"=>"North Puyallup", "lang"=>"en", "lat"=>47.211877, "lon"=>-122.311525),
"/US/US-WA/north-yelm-22260" => array("type"=>"cdp", "pop"=>2075, "name_en"=>"North Yelm", "lang"=>"en", "lat"=>46.963244, "lon"=>-122.601745),
"/US/US-WA/northport-22258" => array("type"=>"town", "pop"=>333, "name_ru"=>"Нортпорт", "name_en"=>"Northport", "lang"=>"en", "lat"=>48.916744, "lon"=>-117.778268),
"/US/US-WA/oak-harbor-22262" => array("type"=>"city", "pop"=>22744, "name_en"=>"Oak Harbor", "lang"=>"en", "lat"=>48.292926, "lon"=>-122.622554),
"/US/US-WA/oakesdale-22261" => array("type"=>"town", "pop"=>391, "name_en"=>"Oakesdale", "lang"=>"en", "lat"=>47.130865, "lon"=>-117.245522),
"/US/US-WA/oakville-22263" => array("type"=>"city", "pop"=>721, "name_ru"=>"Оуквилл", "name_en"=>"Oakville", "lang"=>"en", "lat"=>46.840428, "lon"=>-123.233567),
"/US/US-WA/ocean-park-22264" => array("type"=>"cdp", "pop"=>1409, "name_en"=>"Ocean Park", "lang"=>"en", "lat"=>46.495745, "lon"=>-124.043180),
"/US/US-WA/ocean-shores-22265" => array("type"=>"city", "pop"=>5029, "name_en"=>"Ocean Shores", "lang"=>"en", "lat"=>46.971075, "lon"=>-124.152344),
"/US/US-WA/odessa-22266" => array("type"=>"town", "pop"=>917, "name_ru"=>"Одесса", "name_en"=>"Odessa", "lang"=>"en", "lat"=>47.333311, "lon"=>-118.687119),
"/US/US-WA/okanogan-22267" => array("type"=>"city", "pop"=>2351, "name_en"=>"Okanogan", "lang"=>"en", "lat"=>48.368785, "lon"=>-119.576129),
"/US/US-WA/olympia-22268" => array("type"=>"city", "pop"=>45322, "name_en"=>"Olympia", "lang"=>"en", "lat"=>47.041917, "lon"=>-122.893766),
"/US/US-WA/omak-22269" => array("type"=>"city", "pop"=>4706, "name_en"=>"Omak", "lang"=>"en", "lat"=>48.409604, "lon"=>-119.525503),
"/US/US-WA/opportunity-22270" => array("type"=>"cdp", "pop"=>22326, "name_en"=>"Opportunity", "lang"=>"en", "lat"=>47.649450, "lon"=>-117.240452),
"/US/US-WA/orchards-north-22271" => array("type"=>"cdp", "pop"=>6479, "name_en"=>"Orchards North", "lang"=>"en", "lat"=>45.683389, "lon"=>-122.525753),
"/US/US-WA/orchards-south-22272" => array("type"=>"cdp", "pop"=>12956, "name_en"=>"Orchards South", "lang"=>"en", "lat"=>45.655900, "lon"=>-122.538144),
"/US/US-WA/oroville-22273" => array("type"=>"town", "pop"=>1658, "name_ru"=>"Оровилл", "name_en"=>"Oroville", "lang"=>"en", "lat"=>48.942365, "lon"=>-119.431617),
"/US/US-WA/orting-22274" => array("type"=>"city", "pop"=>6188, "name_en"=>"Orting", "lang"=>"en", "lat"=>47.088800, "lon"=>-122.205395),
"/US/US-WA/othello-22275" => array("type"=>"city", "pop"=>6577, "name_en"=>"Othello", "lang"=>"en", "lat"=>46.820445, "lon"=>-119.167086),
"/US/US-WA/otis-orchards-east-farms-22276" => array("type"=>"cdp", "pop"=>5811, "name_en"=>"Otis Orchards-East Farms", "lang"=>"en", "lat"=>47.700170, "lon"=>-117.090130),
"/US/US-WA/pacific-22277" => array("type"=>"city", "pop"=>6040, "name_ru"=>"Пасифик", "name_en"=>"Pacific", "lang"=>"en", "lat"=>47.265750, "lon"=>-122.246724),
"/US/US-WA/paine-field-lake-stickney-22278" => array("type"=>"cdp", "pop"=>18670, "name_en"=>"Paine Field-Lake Stickney", "lang"=>"en", "lat"=>47.887050, "lon"=>-122.262371),
"/US/US-WA/palouse-22279" => array("type"=>"city", "pop"=>934, "name_en"=>"Palouse", "lang"=>"en", "lat"=>46.911105, "lon"=>-117.073817),
"/US/US-WA/parkland-22280" => array("type"=>"cdp", "pop"=>20882, "name_ru"=>"Паркланд", "name_en"=>"Parkland", "lang"=>"en", "lat"=>47.137850, "lon"=>-122.429324),
"/US/US-WA/parkwood-22281" => array("type"=>"cdp", "pop"=>6853, "name_ru"=>"Парквуд", "name_en"=>"Parkwood", "lang"=>"en", "lat"=>47.525978, "lon"=>-122.596670),
"/US/US-WA/pasco-22282" => array("type"=>"city", "pop"=>55246, "name_en"=>"Pasco", "lang"=>"en", "lat"=>46.252067, "lon"=>-119.127763),
"/US/US-WA/pateros-22283" => array("type"=>"city", "pop"=>616, "name_en"=>"Pateros", "lang"=>"en", "lat"=>48.054811, "lon"=>-119.899852),
"/US/US-WA/pe-ell-22284" => array("type"=>"town", "pop"=>679, "name_en"=>"Pe Ell", "lang"=>"en", "lat"=>46.571534, "lon"=>-123.296921),
"/US/US-WA/pine-lake-22285" => array("type"=>"cdp", "pop"=>13940, "name_ru"=>"Пайн Лейк", "name_en"=>"Pine Lake", "lang"=>"en", "lat"=>47.578950, "lon"=>-122.044887),
"/US/US-WA/pomeroy-22286" => array("type"=>"city", "pop"=>1281, "name_ru"=>"Померой", "name_en"=>"Pomeroy", "lang"=>"en", "lat"=>46.473961, "lon"=>-117.595545),
"/US/US-WA/port-angeles-22287" => array("type"=>"city", "pop"=>18982, "name_en"=>"Port Angeles", "lang"=>"en", "lat"=>48.190986, "lon"=>-123.459641),
"/US/US-WA/port-angeles-east-22288" => array("type"=>"cdp", "pop"=>2672, "name_en"=>"Port Angeles East", "lang"=>"en", "lat"=>48.105908, "lon"=>-123.376086),
"/US/US-WA/port-orchard-22289" => array("type"=>"city", "pop"=>7953, "name_en"=>"Port Orchard", "lang"=>"en", "lat"=>47.529049, "lon"=>-122.641158),
"/US/US-WA/port-townsend-22290" => array("type"=>"city", "pop"=>9133, "name_en"=>"Port Townsend", "lang"=>"en", "lat"=>48.122031, "lon"=>-122.779180),
"/US/US-WA/poulsbo-22291" => array("type"=>"city", "pop"=>8149, "name_en"=>"Poulsbo", "lang"=>"en", "lat"=>47.737134, "lon"=>-122.640039),
"/US/US-WA/prairie-ridge-22292" => array("type"=>"cdp", "pop"=>8278, "name_en"=>"Prairie Ridge", "lang"=>"en", "lat"=>47.137855, "lon"=>-122.147412),
"/US/US-WA/prescott-22293" => array("type"=>"town", "pop"=>317, "name_ru"=>"Прескотт", "name_en"=>"Prescott", "lang"=>"en", "lat"=>46.299054, "lon"=>-118.311468),
"/US/US-WA/priest-point-22294" => array("type"=>"cdp", "pop"=>703, "name_en"=>"Priest Point", "lang"=>"en", "lat"=>48.036906, "lon"=>-122.249727),
"/US/US-WA/prosser-22295" => array("type"=>"city", "pop"=>5124, "name_en"=>"Prosser", "lang"=>"en", "lat"=>46.206210, "lon"=>-119.766319),
"/US/US-WA/pullman-22296" => array("type"=>"city", "pop"=>26920, "name_en"=>"Pullman", "lang"=>"en", "lat"=>46.733253, "lon"=>-117.161959),
"/US/US-WA/puyallup-22297" => array("type"=>"city", "pop"=>36330, "name_en"=>"Puyallup", "lang"=>"en", "lat"=>47.181946, "lon"=>-122.284160),
"/US/US-WA/quincy-22298" => array("type"=>"city", "pop"=>5882, "name_ru"=>"Квинси", "name_en"=>"Quincy", "lang"=>"en", "lat"=>47.233836, "lon"=>-119.851135),
"/US/US-WA/rainier-22299" => array("type"=>"town", "pop"=>1645, "name_en"=>"Rainier", "lang"=>"en", "lat"=>46.890523, "lon"=>-122.684604),
"/US/US-WA/raymond-22300" => array("type"=>"city", "pop"=>2865, "name_ru"=>"Реймонд", "name_en"=>"Raymond", "lang"=>"en", "lat"=>46.683475, "lon"=>-123.736939),
"/US/US-WA/reardan-22301" => array("type"=>"town", "pop"=>597, "name_en"=>"Reardan", "lang"=>"en", "lat"=>47.669174, "lon"=>-117.877405),
"/US/US-WA/redmond-22302" => array("type"=>"city", "pop"=>49548, "name_en"=>"Redmond", "lang"=>"en", "lat"=>47.674200, "lon"=>-122.114799),
"/US/US-WA/renton-22303" => array("type"=>"city", "pop"=>62266, "name_en"=>"Renton", "lang"=>"en", "lat"=>47.481883, "lon"=>-122.196616),
"/US/US-WA/republic-22304" => array("type"=>"town", "pop"=>941, "name_ru"=>"Репаблик", "name_en"=>"Republic", "lang"=>"en", "lat"=>48.649060, "lon"=>-118.731849),
"/US/US-WA/richland-22305" => array("type"=>"city", "pop"=>46155, "name_ru"=>"Ричленд", "name_en"=>"Richland", "lang"=>"en", "lat"=>46.294000, "lon"=>-119.290720),
"/US/US-WA/richmond-beach-innis-arden-22306" => array("type"=>"cdp", "pop"=>7242, "name_en"=>"Richmond Beach-Innis Arden", "lang"=>"en", "lat"=>47.755250, "lon"=>-122.373393),
"/US/US-WA/richmond-highlands-22307" => array("type"=>"cdp", "pop"=>26037, "name_en"=>"Richmond Highlands", "lang"=>"en", "lat"=>47.759133, "lon"=>-122.343042),
"/US/US-WA/ridgefield-22308" => array("type"=>"city", "pop"=>4314, "name_ru"=>"Риджфилд", "name_en"=>"Ridgefield", "lang"=>"en", "lat"=>45.817306, "lon"=>-122.740232),
"/US/US-WA/ritzville-22309" => array("type"=>"city", "pop"=>1736, "name_en"=>"Ritzville", "lang"=>"en", "lat"=>47.126563, "lon"=>-118.375997),
"/US/US-WA/riverside-22310" => array("type"=>"town", "pop"=>327, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>48.502741, "lon"=>-119.507531),
"/US/US-WA/riverton-boulevard-park-22311" => array("type"=>"cdp", "pop"=>15337, "name_en"=>"Riverton-Boulevard Park", "lang"=>"en", "lat"=>47.497623, "lon"=>-122.309231),
"/US/US-WA/rochester-22312" => array("type"=>"cdp", "pop"=>1250, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>46.829113, "lon"=>-123.070734),
"/US/US-WA/rock-island-22314" => array("type"=>"town", "pop"=>827, "name_ru"=>"Рок Айленд", "name_en"=>"Rock Island", "lang"=>"en", "lat"=>47.371833, "lon"=>-120.137093),
"/US/US-WA/rockford-22313" => array("type"=>"town", "pop"=>512, "name_ru"=>"Рокфорд", "name_en"=>"Rockford", "lang"=>"en", "lat"=>47.451022, "lon"=>-117.129200),
"/US/US-WA/rosalia-22315" => array("type"=>"town", "pop"=>587, "name_en"=>"Rosalia", "lang"=>"en", "lat"=>47.236975, "lon"=>-117.367869),
"/US/US-WA/roslyn-22316" => array("type"=>"city", "pop"=>992, "name_en"=>"Roslyn", "lang"=>"en", "lat"=>47.244848, "lon"=>-121.102623),
"/US/US-WA/roy-22317" => array("type"=>"city", "pop"=>689, "name_en"=>"Roy", "lang"=>"en", "lat"=>47.001164, "lon"=>-122.544299),
"/US/US-WA/royal-city-22318" => array("type"=>"city", "pop"=>1966, "name_en"=>"Royal City", "lang"=>"en", "lat"=>46.898890, "lon"=>-119.614081),
"/US/US-WA/ruston-22319" => array("type"=>"town", "pop"=>734, "name_ru"=>"Растон", "name_en"=>"Ruston", "lang"=>"en", "lat"=>47.297817, "lon"=>-122.509261),
"/US/US-WA/sahalee-22320" => array("type"=>"cdp", "pop"=>13951, "name_en"=>"Sahalee", "lang"=>"en", "lat"=>47.634900, "lon"=>-122.054114),
"/US/US-WA/salmon-creek-22322" => array("type"=>"cdp", "pop"=>11989, "name_en"=>"Salmon Creek", "lang"=>"en", "lat"=>45.713400, "lon"=>-122.661051),
"/US/US-WA/satus-22323" => array("type"=>"cdp", "pop"=>1343, "name_en"=>"Satus", "lang"=>"en", "lat"=>46.231965, "lon"=>-120.118374),
"/US/US-WA/sea-tac-22324" => array("type"=>"cdp", "pop"=>22694, "name_en"=>"Sea-Tac", "lang"=>"en", "lat"=>47.443483, "lon"=>-122.297767),
"/US/US-WA/seattle-22325" => array("type"=>"city", "pop"=>598541, "name_en"=>"Seattle", "lang"=>"en", "lat"=>47.621800, "lon"=>-122.350326),
"/US/US-WA/sedro-woolley-22326" => array("type"=>"city", "pop"=>10871, "name_en"=>"Sedro-Woolley", "lang"=>"en", "lat"=>48.508170, "lon"=>-122.235005),
"/US/US-WA/selah-22327" => array("type"=>"city", "pop"=>7126, "name_en"=>"Selah", "lang"=>"en", "lat"=>46.656160, "lon"=>-120.533925),
"/US/US-WA/sequim-22328" => array("type"=>"city", "pop"=>5809, "name_en"=>"Sequim", "lang"=>"en", "lat"=>48.077490, "lon"=>-123.105254),
"/US/US-WA/shaker-church-22329" => array("type"=>"cdp", "pop"=>670, "name_en"=>"Shaker Church", "lang"=>"en", "lat"=>48.052780, "lon"=>-122.227115),
"/US/US-WA/shelter-bay-22330" => array("type"=>"cdp", "pop"=>1069, "name_en"=>"Shelter Bay", "lang"=>"en", "lat"=>48.382606, "lon"=>-122.513953),
"/US/US-WA/shelton-22331" => array("type"=>"city", "pop"=>9276, "name_ru"=>"Шелтон", "name_en"=>"Shelton", "lang"=>"en", "lat"=>47.214754, "lon"=>-123.105275),
"/US/US-WA/sheridan-beach-22332" => array("type"=>"cdp", "pop"=>6518, "name_en"=>"Sheridan Beach", "lang"=>"en", "lat"=>47.744917, "lon"=>-122.286815),
"/US/US-WA/silver-lake-fircrest-22334" => array("type"=>"cdp", "pop"=>24474, "name_en"=>"Silver Lake-Fircrest", "lang"=>"en", "lat"=>47.873650, "lon"=>-122.169807),
"/US/US-WA/silverdale-22333" => array("type"=>"cdp", "pop"=>7660, "name_en"=>"Silverdale", "lang"=>"en", "lat"=>47.652150, "lon"=>-122.683083),
"/US/US-WA/skokomish-22335" => array("type"=>"cdp", "pop"=>532, "name_en"=>"Skokomish", "lang"=>"en", "lat"=>47.333558, "lon"=>-123.156230),
"/US/US-WA/skykomish-22336" => array("type"=>"town", "pop"=>205, "name_en"=>"Skykomish", "lang"=>"en", "lat"=>47.711198, "lon"=>-121.357381),
"/US/US-WA/smokey-point-22337" => array("type"=>"cdp", "pop"=>2620, "name_en"=>"Smokey Point", "lang"=>"en", "lat"=>48.150412, "lon"=>-122.193337),
"/US/US-WA/snee-oosh-22338" => array("type"=>"cdp", "pop"=>302, "name_en"=>"Snee Oosh", "lang"=>"en", "lat"=>48.422535, "lon"=>-122.553110),
"/US/US-WA/snohomish-22339" => array("type"=>"city", "pop"=>8801, "name_en"=>"Snohomish", "lang"=>"en", "lat"=>47.919525, "lon"=>-122.093441),
"/US/US-WA/snoqualmie-22340" => array("type"=>"city", "pop"=>8161, "name_en"=>"Snoqualmie", "lang"=>"en", "lat"=>47.524265, "lon"=>-121.817629),
"/US/US-WA/soap-lake-22341" => array("type"=>"city", "pop"=>1826, "name_en"=>"Soap Lake", "lang"=>"en", "lat"=>47.388318, "lon"=>-119.489713),
"/US/US-WA/south-bend-22342" => array("type"=>"city", "pop"=>1785, "name_ru"=>"Саут Бенд", "name_en"=>"South Bend", "lang"=>"en", "lat"=>46.670652, "lon"=>-123.803043),
"/US/US-WA/south-broadway-22343" => array("type"=>"cdp", "pop"=>2735, "name_en"=>"South Broadway", "lang"=>"en", "lat"=>46.566400, "lon"=>-120.515649),
"/US/US-WA/south-cle-elum-22344" => array("type"=>"town", "pop"=>573, "name_en"=>"South Cle Elum", "lang"=>"en", "lat"=>47.186751, "lon"=>-120.951803),
"/US/US-WA/south-hill-22345" => array("type"=>"cdp", "pop"=>12963, "name_en"=>"South Hill", "lang"=>"en", "lat"=>47.141250, "lon"=>-122.268977),
"/US/US-WA/south-prairie-22346" => array("type"=>"town", "pop"=>427, "name_en"=>"South Prairie", "lang"=>"en", "lat"=>47.138500, "lon"=>-122.091842),
"/US/US-WA/south-wenatchee-22347" => array("type"=>"cdp", "pop"=>1207, "name_en"=>"South Wenatchee", "lang"=>"en", "lat"=>47.391173, "lon"=>-120.292642),
"/US/US-WA/spanaway-22348" => array("type"=>"cdp", "pop"=>15001, "name_en"=>"Spanaway", "lang"=>"en", "lat"=>47.106209, "lon"=>-122.426474),
"/US/US-WA/spangle-22349" => array("type"=>"city", "pop"=>236, "name_en"=>"Spangle", "lang"=>"en", "lat"=>47.430535, "lon"=>-117.378813),
"/US/US-WA/spokane-22350" => array("type"=>"city", "pop"=>202319, "name_en"=>"Spokane", "lang"=>"en", "lat"=>47.672300, "lon"=>-117.414024),
"/US/US-WA/sprague-22351" => array("type"=>"city", "pop"=>478, "name_en"=>"Sprague", "lang"=>"en", "lat"=>47.300021, "lon"=>-117.974129),
"/US/US-WA/springdale-22352" => array("type"=>"town", "pop"=>277, "name_ru"=>"Спрингдейл", "name_en"=>"Springdale", "lang"=>"en", "lat"=>48.056930, "lon"=>-117.746314),
"/US/US-WA/st.-john-22321" => array("type"=>"town", "pop"=>514, "name_ru"=>"Сент Джон", "name_en"=>"St. John", "lang"=>"en", "lat"=>47.088754, "lon"=>-117.581063),
"/US/US-WA/stanwood-22353" => array("type"=>"city", "pop"=>6089, "name_ru"=>"Стэнвуд", "name_en"=>"Stanwood", "lang"=>"en", "lat"=>48.243202, "lon"=>-122.348156),
"/US/US-WA/starbuck-22354" => array("type"=>"town", "pop"=>124, "name_ru"=>"Старбак", "name_en"=>"Starbuck", "lang"=>"en", "lat"=>46.519107, "lon"=>-118.124966),
"/US/US-WA/steilacoom-22355" => array("type"=>"town", "pop"=>6061, "name_en"=>"Steilacoom", "lang"=>"en", "lat"=>47.169600, "lon"=>-122.592422),
"/US/US-WA/stevenson-22356" => array("type"=>"city", "pop"=>1392, "name_ru"=>"Стивенсон", "name_en"=>"Stevenson", "lang"=>"en", "lat"=>45.693621, "lon"=>-121.893160),
"/US/US-WA/stimson-crossing-22357" => array("type"=>"cdp", "pop"=>591, "name_en"=>"Stimson Crossing", "lang"=>"en", "lat"=>48.107932, "lon"=>-122.207035),
"/US/US-WA/sudden-valley-22358" => array("type"=>"cdp", "pop"=>2615, "name_en"=>"Sudden Valley", "lang"=>"en", "lat"=>48.720835, "lon"=>-122.344740),
"/US/US-WA/sultan-22359" => array("type"=>"town", "pop"=>4228, "name_en"=>"Sultan", "lang"=>"en", "lat"=>47.864602, "lon"=>-121.804300),
"/US/US-WA/sumas-22360" => array("type"=>"city", "pop"=>1232, "name_en"=>"Sumas", "lang"=>"en", "lat"=>48.996136, "lon"=>-122.264743),
"/US/US-WA/summit-22361" => array("type"=>"cdp", "pop"=>6312, "name_ru"=>"Саммит", "name_en"=>"Summit", "lang"=>"en", "lat"=>47.172127, "lon"=>-122.358003),
"/US/US-WA/sumner-22362" => array("type"=>"city", "pop"=>9704, "name_ru"=>"Саммер", "name_en"=>"Sumner", "lang"=>"en", "lat"=>47.210550, "lon"=>-122.235998),
"/US/US-WA/sunnyside-22363" => array("type"=>"city", "pop"=>14949, "name_ru"=>"Саннисайд", "name_en"=>"Sunnyside", "lang"=>"en", "lat"=>46.320969, "lon"=>-120.012530),
"/US/US-WA/sunnyslope-22364" => array("type"=>"cdp", "pop"=>1907, "name_ru"=>"Саннислоуп", "name_en"=>"Sunnyslope", "lang"=>"en", "lat"=>47.491842, "lon"=>-120.341281),
"/US/US-WA/suquamish-22365" => array("type"=>"cdp", "pop"=>3105, "name_en"=>"Suquamish", "lang"=>"en", "lat"=>47.729094, "lon"=>-122.584289),
"/US/US-WA/swinomish-village-22366" => array("type"=>"cdp", "pop"=>563, "name_en"=>"Swinomish Village", "lang"=>"en", "lat"=>48.408404, "lon"=>-122.509923),
"/US/US-WA/tacoma-22367" => array("type"=>"city", "pop"=>197181, "name_en"=>"Tacoma", "lang"=>"en", "lat"=>47.252000, "lon"=>-122.459766),
"/US/US-WA/taholah-22368" => array("type"=>"cdp", "pop"=>788, "name_en"=>"Taholah", "lang"=>"en", "lat"=>47.340047, "lon"=>-124.280145),
"/US/US-WA/tanglewilde-thompson-place-22369" => array("type"=>"cdp", "pop"=>6061, "name_en"=>"Tanglewilde-Thompson Place", "lang"=>"en", "lat"=>47.052500, "lon"=>-122.778061),
"/US/US-WA/tekoa-22370" => array("type"=>"city", "pop"=>760, "name_en"=>"Tekoa", "lang"=>"en", "lat"=>47.225645, "lon"=>-117.073299),
"/US/US-WA/tenino-22371" => array("type"=>"town", "pop"=>1590, "name_en"=>"Tenino", "lang"=>"en", "lat"=>46.857381, "lon"=>-122.847366),
"/US/US-WA/terrace-heights-22372" => array("type"=>"cdp", "pop"=>4223, "name_en"=>"Terrace Heights", "lang"=>"en", "lat"=>46.605967, "lon"=>-120.437680),
"/US/US-WA/tieton-22373" => array("type"=>"town", "pop"=>1162, "name_en"=>"Tieton", "lang"=>"en", "lat"=>46.701752, "lon"=>-120.752863),
"/US/US-WA/toledo-22374" => array("type"=>"city", "pop"=>678, "name_ru"=>"Толедо", "name_en"=>"Toledo", "lang"=>"en", "lat"=>46.440470, "lon"=>-122.847658),
"/US/US-WA/tonasket-22375" => array("type"=>"town", "pop"=>947, "name_en"=>"Tonasket", "lang"=>"en", "lat"=>48.704785, "lon"=>-119.437837),
"/US/US-WA/toppenish-22376" => array("type"=>"city", "pop"=>9166, "name_en"=>"Toppenish", "lang"=>"en", "lat"=>46.381118, "lon"=>-120.311549),
"/US/US-WA/town-and-country-22377" => array("type"=>"cdp", "pop"=>4921, "name_en"=>"Town and Country", "lang"=>"en", "lat"=>47.727450, "lon"=>-117.420524),
"/US/US-WA/tracyton-22378" => array("type"=>"cdp", "pop"=>2621, "name_en"=>"Tracyton", "lang"=>"en", "lat"=>47.609600, "lon"=>-122.654499),
"/US/US-WA/trentwood-22379" => array("type"=>"cdp", "pop"=>4060, "name_en"=>"Trentwood", "lang"=>"en", "lat"=>47.697950, "lon"=>-117.213374),
"/US/US-WA/tukwila-22380" => array("type"=>"city", "pop"=>17092, "name_en"=>"Tukwila", "lang"=>"en", "lat"=>47.477700, "lon"=>-122.268299),
"/US/US-WA/tulalip-bay-22381" => array("type"=>"cdp", "pop"=>1395, "name_en"=>"Tulalip Bay", "lang"=>"en", "lat"=>48.037406, "lon"=>-122.308984),
"/US/US-WA/tumwater-22382" => array("type"=>"city", "pop"=>14100, "name_en"=>"Tumwater", "lang"=>"en", "lat"=>46.998107, "lon"=>-122.917865),
"/US/US-WA/twisp-22383" => array("type"=>"town", "pop"=>900, "name_en"=>"Twisp", "lang"=>"en", "lat"=>48.363275, "lon"=>-120.119515),
"/US/US-WA/union-gap-22384" => array("type"=>"city", "pop"=>5699, "name_en"=>"Union Gap", "lang"=>"en", "lat"=>46.555133, "lon"=>-120.487382),
"/US/US-WA/uniontown-22385" => array("type"=>"town", "pop"=>316, "name_ru"=>"Юнионтаун", "name_en"=>"Uniontown", "lang"=>"en", "lat"=>46.537392, "lon"=>-117.084603),
"/US/US-WA/university-place-22386" => array("type"=>"cdp", "pop"=>30345, "name_en"=>"University Place", "lang"=>"en", "lat"=>47.219098, "lon"=>-122.542440),
"/US/US-WA/vader-22387" => array("type"=>"city", "pop"=>612, "name_en"=>"Vader", "lang"=>"en", "lat"=>46.402781, "lon"=>-122.955647),
"/US/US-WA/vancouver-22388" => array("type"=>"city", "pop"=>163186, "name_en"=>"Vancouver", "lang"=>"en", "lat"=>45.634200, "lon"=>-122.643272),
"/US/US-WA/vancouver-mall-22389" => array("type"=>"cdp", "pop"=>6938, "name_en"=>"Vancouver Mall", "lang"=>"en", "lat"=>45.653250, "lon"=>-122.583710),
"/US/US-WA/veradale-22390" => array("type"=>"cdp", "pop"=>7836, "name_en"=>"Veradale", "lang"=>"en", "lat"=>47.649450, "lon"=>-117.207524),
"/US/US-WA/waitsburg-22391" => array("type"=>"city", "pop"=>1230, "name_en"=>"Waitsburg", "lang"=>"en", "lat"=>46.269110, "lon"=>-118.150395),
"/US/US-WA/walla-walla-22392" => array("type"=>"city", "pop"=>30643, "name_en"=>"Walla Walla", "lang"=>"en", "lat"=>46.066902, "lon"=>-118.341229),
"/US/US-WA/walla-walla-east-22393" => array("type"=>"cdp", "pop"=>2959, "name_en"=>"Walla Walla East", "lang"=>"en", "lat"=>46.054538, "lon"=>-118.302608),
"/US/US-WA/waller-22394" => array("type"=>"cdp", "pop"=>6415, "name_en"=>"Waller", "lang"=>"en", "lat"=>47.200870, "lon"=>-122.368190),
"/US/US-WA/walnut-grove-22395" => array("type"=>"cdp", "pop"=>3906, "name_ru"=>"Уолнат Гров", "name_en"=>"Walnut Grove", "lang"=>"en", "lat"=>45.676750, "lon"=>-122.607566),
"/US/US-WA/wapato-22396" => array("type"=>"city", "pop"=>4543, "name_en"=>"Wapato", "lang"=>"en", "lat"=>46.445672, "lon"=>-120.420651),
"/US/US-WA/warden-22397" => array("type"=>"town", "pop"=>2582, "name_en"=>"Warden", "lang"=>"en", "lat"=>46.967887, "lon"=>-119.046990),
"/US/US-WA/washougal-22398" => array("type"=>"city", "pop"=>13509, "name_en"=>"Washougal", "lang"=>"en", "lat"=>45.579032, "lon"=>-122.343937),
"/US/US-WA/washtucna-22399" => array("type"=>"town", "pop"=>262, "name_en"=>"Washtucna", "lang"=>"en", "lat"=>46.753507, "lon"=>-118.311604),
"/US/US-WA/waterville-22400" => array("type"=>"town", "pop"=>1129, "name_ru"=>"Уотервилл", "name_en"=>"Waterville", "lang"=>"en", "lat"=>47.648048, "lon"=>-120.069540),
"/US/US-WA/waverly-22401" => array("type"=>"town", "pop"=>116, "name_ru"=>"Уэверли", "name_en"=>"Waverly", "lang"=>"en", "lat"=>47.339548, "lon"=>-117.228749),
"/US/US-WA/weallup-lake-22402" => array("type"=>"cdp", "pop"=>681, "name_en"=>"Weallup Lake", "lang"=>"en", "lat"=>48.110389, "lon"=>-122.304884),
"/US/US-WA/wenatchee-22403" => array("type"=>"city", "pop"=>29898, "name_en"=>"Wenatchee", "lang"=>"en", "lat"=>47.425661, "lon"=>-120.324928),
"/US/US-WA/west-clarkston-highland-22404" => array("type"=>"cdp", "pop"=>3913, "name_en"=>"West Clarkston-Highland", "lang"=>"en", "lat"=>46.398611, "lon"=>-117.057239),
"/US/US-WA/west-lake-sammamish-22405" => array("type"=>"cdp", "pop"=>6087, "name_en"=>"West Lake Sammamish", "lang"=>"en", "lat"=>47.577867, "lon"=>-122.099887),
"/US/US-WA/west-lake-stevens-22406" => array("type"=>"cdp", "pop"=>12453, "name_en"=>"West Lake Stevens", "lang"=>"en", "lat"=>47.993526, "lon"=>-122.100681),
"/US/US-WA/west-longview-22407" => array("type"=>"cdp", "pop"=>3163, "name_en"=>"West Longview", "lang"=>"en", "lat"=>46.168032, "lon"=>-122.997773),
"/US/US-WA/west-pasco-22408" => array("type"=>"cdp", "pop"=>7312, "name_en"=>"West Pasco", "lang"=>"en", "lat"=>46.245600, "lon"=>-119.181699),
"/US/US-WA/west-richland-22410" => array("type"=>"city", "pop"=>10857, "name_en"=>"West Richland", "lang"=>"en", "lat"=>46.324505, "lon"=>-119.397289),
"/US/US-WA/west-side-highway-22411" => array("type"=>"cdp", "pop"=>3641, "name_en"=>"West Side Highway", "lang"=>"en", "lat"=>46.184749, "lon"=>-122.910898),
"/US/US-WA/west-valley-22412" => array("type"=>"cdp", "pop"=>6594, "name_en"=>"West Valley", "lang"=>"en", "lat"=>46.592280, "lon"=>-120.603977),
"/US/US-WA/west-wenatchee-22413" => array("type"=>"cdp", "pop"=>2220, "name_en"=>"West Wenatchee", "lang"=>"en", "lat"=>47.443980, "lon"=>-120.352339),
"/US/US-WA/westport-22409" => array("type"=>"city", "pop"=>2577, "name_ru"=>"Уэстпорт", "name_en"=>"Westport", "lang"=>"en", "lat"=>46.891965, "lon"=>-124.110985),
"/US/US-WA/white-center-shorewood-22414" => array("type"=>"cdp", "pop"=>20531, "name_en"=>"White Center-Shorewood", "lang"=>"en", "lat"=>47.502000, "lon"=>-122.349774),
"/US/US-WA/white-salmon-22415" => array("type"=>"city", "pop"=>2411, "name_en"=>"White Salmon", "lang"=>"en", "lat"=>45.727694, "lon"=>-121.483668),
"/US/US-WA/white-swan-22416" => array("type"=>"cdp", "pop"=>2669, "name_en"=>"White Swan", "lang"=>"en", "lat"=>46.388236, "lon"=>-120.720833),
"/US/US-WA/wilbur-22417" => array("type"=>"town", "pop"=>878, "name_en"=>"Wilbur", "lang"=>"en", "lat"=>47.758134, "lon"=>-118.704531),
"/US/US-WA/wilkeson-22418" => array("type"=>"town", "pop"=>421, "name_en"=>"Wilkeson", "lang"=>"en", "lat"=>47.107150, "lon"=>-122.046956),
"/US/US-WA/wilson-creek-22419" => array("type"=>"town", "pop"=>243, "name_en"=>"Wilson Creek", "lang"=>"en", "lat"=>47.422607, "lon"=>-119.115715),
"/US/US-WA/winlock-22420" => array("type"=>"city", "pop"=>1249, "name_en"=>"Winlock", "lang"=>"en", "lat"=>46.491494, "lon"=>-122.935892),
"/US/US-WA/winslow-22421" => array("type"=>"city", "pop"=>3081, "name_ru"=>"Уинслоу", "name_en"=>"Winslow", "lang"=>"en", "lat"=>47.629654, "lon"=>-122.517780),
"/US/US-WA/winthrop-22422" => array("type"=>"town", "pop"=>387, "name_ru"=>"Уинтроп", "name_en"=>"Winthrop", "lang"=>"en", "lat"=>48.472877, "lon"=>-120.178262),
"/US/US-WA/woodinville-22423" => array("type"=>"cdp", "pop"=>11240, "name_en"=>"Woodinville", "lang"=>"en", "lat"=>47.746158, "lon"=>-122.111867),
"/US/US-WA/woodland-22424" => array("type"=>"city", "pop"=>4843, "name_ru"=>"Вудленд", "name_en"=>"Woodland", "lang"=>"en", "lat"=>45.915003, "lon"=>-122.743044),
"/US/US-WA/woodmont-beach-22425" => array("type"=>"cdp", "pop"=>7493, "name_en"=>"Woodmont Beach", "lang"=>"en", "lat"=>47.361833, "lon"=>-122.316298),
"/US/US-WA/woodway-22426" => array("type"=>"city", "pop"=>1133, "name_en"=>"Woodway", "lang"=>"en", "lat"=>47.794550, "lon"=>-122.415361),
"/US/US-WA/yacolt-22427" => array("type"=>"town", "pop"=>1213, "name_en"=>"Yacolt", "lang"=>"en", "lat"=>45.865670, "lon"=>-122.405853),
"/US/US-WA/yakima-22428" => array("type"=>"city", "pop"=>84074, "name_en"=>"Yakima", "lang"=>"en", "lat"=>46.592633, "lon"=>-120.528908),
"/US/US-WA/yarrow-point-22429" => array("type"=>"town", "pop"=>1073, "name_en"=>"Yarrow Point", "lang"=>"en", "lat"=>47.645750, "lon"=>-122.215332),
"/US/US-WA/yelm-22430" => array("type"=>"town", "pop"=>5800, "name_en"=>"Yelm", "lang"=>"en", "lat"=>46.945668, "lon"=>-122.603518),
"/US/US-WA/zillah-22431" => array("type"=>"city", "pop"=>2674, "name_en"=>"Zillah", "lang"=>"en", "lat"=>46.403710, "lon"=>-120.259531),
"/US/US-WI" => array("type"=>"state", "pop"=>5627967, "name_ru"=>"Висконсин", "name_en"=>"Wisconsin", "lang"=>"en"),
"/US/US-WI/abbotsford-22709" => array("type"=>"city", "pop"=>1862, "name_en"=>"Abbotsford", "lang"=>"en", "lat"=>44.948293, "lon"=>-90.318747),
"/US/US-WI/adams-22710" => array("type"=>"city", "pop"=>1744, "name_ru"=>"Адамс", "name_en"=>"Adams", "lang"=>"en", "lat"=>43.955287, "lon"=>-89.816929),
"/US/US-WI/adell-22711" => array("type"=>"village", "pop"=>501, "name_en"=>"Adell", "lang"=>"en", "lat"=>43.620788, "lon"=>-87.946174),
"/US/US-WI/albany-22712" => array("type"=>"village", "pop"=>1116, "name_ru"=>"Олбани", "name_en"=>"Albany", "lang"=>"en", "lat"=>42.707031, "lon"=>-89.436841),
"/US/US-WI/algoma-22713" => array("type"=>"city", "pop"=>3116, "name_ru"=>"Алгома", "name_en"=>"Algoma", "lang"=>"en", "lat"=>44.605929, "lon"=>-87.445442),
"/US/US-WI/allouez-22714" => array("type"=>"village", "pop"=>14718, "name_en"=>"Allouez", "lang"=>"en", "lat"=>44.472250, "lon"=>-88.026046),
"/US/US-WI/alma-22715" => array("type"=>"city", "pop"=>869, "name_ru"=>"Алма", "name_en"=>"Alma", "lang"=>"en", "lat"=>44.330990, "lon"=>-91.922157),
"/US/US-WI/alma-center-22716" => array("type"=>"village", "pop"=>435, "name_en"=>"Alma Center", "lang"=>"en", "lat"=>44.436635, "lon"=>-90.913015),
"/US/US-WI/almena-22717" => array("type"=>"village", "pop"=>736, "name_ru"=>"Алмена", "name_en"=>"Almena", "lang"=>"en", "lat"=>45.414511, "lon"=>-92.036471),
"/US/US-WI/almond-22718" => array("type"=>"village", "pop"=>415, "name_en"=>"Almond", "lang"=>"en", "lat"=>44.261306, "lon"=>-89.408736),
"/US/US-WI/altoona-22719" => array("type"=>"city", "pop"=>6753, "name_ru"=>"Альтоона", "name_en"=>"Altoona", "lang"=>"en", "lat"=>44.805050, "lon"=>-91.441999),
"/US/US-WI/amery-22720" => array("type"=>"city", "pop"=>2777, "name_en"=>"Amery", "lang"=>"en", "lat"=>45.304607, "lon"=>-92.363079),
"/US/US-WI/amherst-22721" => array("type"=>"village", "pop"=>995, "name_ru"=>"Амхерст", "name_en"=>"Amherst", "lang"=>"en", "lat"=>44.448247, "lon"=>-89.284798),
"/US/US-WI/amherst-junction-22722" => array("type"=>"village", "pop"=>351, "name_en"=>"Amherst Junction", "lang"=>"en", "lat"=>44.469537, "lon"=>-89.316672),
"/US/US-WI/aniwa-22723" => array("type"=>"village", "pop"=>256, "name_en"=>"Aniwa", "lang"=>"en", "lat"=>45.009968, "lon"=>-89.208214),
"/US/US-WI/antigo-22724" => array("type"=>"city", "pop"=>7917, "name_en"=>"Antigo", "lang"=>"en", "lat"=>45.144098, "lon"=>-89.155080),
"/US/US-WI/appleton-22725" => array("type"=>"city", "pop"=>70305, "name_ru"=>"Эпплтон", "name_en"=>"Appleton", "lang"=>"en", "lat"=>44.266333, "lon"=>-88.396278),
"/US/US-WI/arcadia-22726" => array("type"=>"city", "pop"=>2321, "name_ru"=>"Аркадия", "name_en"=>"Arcadia", "lang"=>"en", "lat"=>44.252401, "lon"=>-91.493360),
"/US/US-WI/arena-22727" => array("type"=>"village", "pop"=>806, "name_en"=>"Arena", "lang"=>"en", "lat"=>43.163333, "lon"=>-89.913206),
"/US/US-WI/argyle-22728" => array("type"=>"village", "pop"=>753, "name_ru"=>"Аргайл", "name_en"=>"Argyle", "lang"=>"en", "lat"=>42.700635, "lon"=>-89.866245),
"/US/US-WI/arlington-22729" => array("type"=>"village", "pop"=>680, "name_ru"=>"Арлингтон", "name_en"=>"Arlington", "lang"=>"en", "lat"=>43.338518, "lon"=>-89.376578),
"/US/US-WI/arpin-22730" => array("type"=>"village", "pop"=>307, "name_en"=>"Arpin", "lang"=>"en", "lat"=>44.541272, "lon"=>-90.032516),
"/US/US-WI/ashland-22731" => array("type"=>"city", "pop"=>8118, "name_ru"=>"Ашленд", "name_en"=>"Ashland", "lang"=>"en", "lat"=>46.580154, "lon"=>-90.874147),
"/US/US-WI/ashwaubenon-22732" => array("type"=>"village", "pop"=>17235, "name_en"=>"Ashwaubenon", "lang"=>"en", "lat"=>44.485767, "lon"=>-88.083555),
"/US/US-WI/athens-22733" => array("type"=>"village", "pop"=>1035, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>45.033634, "lon"=>-90.076309),
"/US/US-WI/auburndale-22734" => array("type"=>"village", "pop"=>702, "name_ru"=>"Ауберндэйл", "name_en"=>"Auburndale", "lang"=>"en", "lat"=>44.628279, "lon"=>-90.014882),
"/US/US-WI/augusta-22735" => array("type"=>"city", "pop"=>1407, "name_ru"=>"Огаста", "name_en"=>"Augusta", "lang"=>"en", "lat"=>44.678668, "lon"=>-91.121217),
"/US/US-WI/avoca-22736" => array("type"=>"village", "pop"=>567, "name_ru"=>"Авока", "name_en"=>"Avoca", "lang"=>"en", "lat"=>43.188247, "lon"=>-90.325186),
"/US/US-WI/bagley-22737" => array("type"=>"village", "pop"=>309, "name_ru"=>"Багли", "name_en"=>"Bagley", "lang"=>"en", "lat"=>42.903572, "lon"=>-91.098197),
"/US/US-WI/baldwin-22738" => array("type"=>"village", "pop"=>3709, "name_ru"=>"Болдуин", "name_en"=>"Baldwin", "lang"=>"en", "lat"=>44.962857, "lon"=>-92.371057),
"/US/US-WI/balsam-lake-22739" => array("type"=>"village", "pop"=>1031, "name_en"=>"Balsam Lake", "lang"=>"en", "lat"=>45.458092, "lon"=>-92.454674),
"/US/US-WI/bangor-22740" => array("type"=>"village", "pop"=>1405, "name_ru"=>"Бангор", "name_en"=>"Bangor", "lang"=>"en", "lat"=>43.891965, "lon"=>-90.991115),
"/US/US-WI/baraboo-22741" => array("type"=>"city", "pop"=>11244, "name_en"=>"Baraboo", "lang"=>"en", "lat"=>43.468532, "lon"=>-89.741506),
"/US/US-WI/barneveld-22742" => array("type"=>"village", "pop"=>1171, "name_en"=>"Barneveld", "lang"=>"en", "lat"=>43.014537, "lon"=>-89.895910),
"/US/US-WI/barron-22743" => array("type"=>"city", "pop"=>3122, "name_en"=>"Barron", "lang"=>"en", "lat"=>45.400580, "lon"=>-91.846491),
"/US/US-WI/bay-city-22744" => array("type"=>"village", "pop"=>489, "name_ru"=>"Бэй Сити", "name_en"=>"Bay City", "lang"=>"en", "lat"=>44.585336, "lon"=>-92.452634),
"/US/US-WI/bayfield-22745" => array("type"=>"city", "pop"=>570, "name_ru"=>"Бэйфилд", "name_en"=>"Bayfield", "lang"=>"en", "lat"=>46.811614, "lon"=>-90.821435),
"/US/US-WI/bayside-22746" => array("type"=>"village", "pop"=>4232, "name_en"=>"Bayside", "lang"=>"en", "lat"=>43.182950, "lon"=>-87.901848),
"/US/US-WI/bear-creek-22747" => array("type"=>"village", "pop"=>388, "name_ru"=>"Бэар Крик", "name_en"=>"Bear Creek", "lang"=>"en", "lat"=>44.530835, "lon"=>-88.726896),
"/US/US-WI/beaver-dam-22748" => array("type"=>"city", "pop"=>15133, "name_ru"=>"Бивер Дэм", "name_en"=>"Beaver Dam", "lang"=>"en", "lat"=>43.464003, "lon"=>-88.837164),
"/US/US-WI/belgium-22749" => array("type"=>"village", "pop"=>2272, "name_ru"=>"Белджиум", "name_en"=>"Belgium", "lang"=>"en", "lat"=>43.500584, "lon"=>-87.850286),
"/US/US-WI/bell-center-22750" => array("type"=>"village", "pop"=>108, "name_en"=>"Bell Center", "lang"=>"en", "lat"=>43.291963, "lon"=>-90.825367),
"/US/US-WI/belleville-22751" => array("type"=>"village", "pop"=>2265, "name_ru"=>"Белелвилл", "name_en"=>"Belleville", "lang"=>"en", "lat"=>42.861713, "lon"=>-89.536511),
"/US/US-WI/bellevue-town-22752" => array("type"=>"cdp", "pop"=>7541, "name_en"=>"Bellevue Town", "lang"=>"en", "lat"=>44.459667, "lon"=>-87.955862),
"/US/US-WI/belmont-22753" => array("type"=>"village", "pop"=>877, "name_ru"=>"Белмонт", "name_en"=>"Belmont", "lang"=>"en", "lat"=>42.737963, "lon"=>-90.332583),
"/US/US-WI/beloit-22754" => array("type"=>"city", "pop"=>36160, "name_ru"=>"Белойт", "name_en"=>"Beloit", "lang"=>"en", "lat"=>42.523750, "lon"=>-89.020098),
"/US/US-WI/benton-22755" => array("type"=>"village", "pop"=>945, "name_ru"=>"Бентон", "name_en"=>"Benton", "lang"=>"en", "lat"=>42.570923, "lon"=>-90.382164),
"/US/US-WI/berlin-22756" => array("type"=>"city", "pop"=>4997, "name_ru"=>"Берлин", "name_en"=>"Berlin", "lang"=>"en", "lat"=>43.969972, "lon"=>-88.950195),
"/US/US-WI/big-bend-22757" => array("type"=>"village", "pop"=>1332, "name_en"=>"Big Bend", "lang"=>"en", "lat"=>42.882304, "lon"=>-88.204082),
"/US/US-WI/big-falls-22758" => array("type"=>"village", "pop"=>74, "name_ru"=>"Биг Фолс", "name_en"=>"Big Falls", "lang"=>"en", "lat"=>44.619498, "lon"=>-89.016855),
"/US/US-WI/birchwood-22759" => array("type"=>"village", "pop"=>548, "name_en"=>"Birchwood", "lang"=>"en", "lat"=>45.657100, "lon"=>-91.552157),
"/US/US-WI/birnamwood-22760" => array("type"=>"village", "pop"=>750, "name_en"=>"Birnamwood", "lang"=>"en", "lat"=>44.931229, "lon"=>-89.209530),
"/US/US-WI/biron-22761" => array("type"=>"village", "pop"=>810, "name_en"=>"Biron", "lang"=>"en", "lat"=>44.428389, "lon"=>-89.764381),
"/US/US-WI/black-creek-22762" => array("type"=>"village", "pop"=>1245, "name_en"=>"Black Creek", "lang"=>"en", "lat"=>44.474532, "lon"=>-88.450024),
"/US/US-WI/black-earth-22763" => array("type"=>"village", "pop"=>1322, "name_en"=>"Black Earth", "lang"=>"en", "lat"=>43.136617, "lon"=>-89.746051),
"/US/US-WI/black-river-falls-22764" => array("type"=>"city", "pop"=>3406, "name_en"=>"Black River Falls", "lang"=>"en", "lat"=>44.298712, "lon"=>-90.846998),
"/US/US-WI/blair-22765" => array("type"=>"city", "pop"=>1246, "name_en"=>"Blair", "lang"=>"en", "lat"=>44.295938, "lon"=>-91.229946),
"/US/US-WI/blanchardville-22766" => array("type"=>"village", "pop"=>736, "name_en"=>"Blanchardville", "lang"=>"en", "lat"=>42.809324, "lon"=>-89.860633),
"/US/US-WI/bloomer-22767" => array("type"=>"city", "pop"=>3304, "name_en"=>"Bloomer", "lang"=>"en", "lat"=>45.102493, "lon"=>-91.490800),
"/US/US-WI/bloomington-22768" => array("type"=>"village", "pop"=>668, "name_ru"=>"Блумингтон", "name_en"=>"Bloomington", "lang"=>"en", "lat"=>42.892143, "lon"=>-90.927330),
"/US/US-WI/blue-mounds-22769" => array("type"=>"village", "pop"=>773, "name_en"=>"Blue Mounds", "lang"=>"en", "lat"=>43.020719, "lon"=>-89.826184),
"/US/US-WI/blue-river-22770" => array("type"=>"village", "pop"=>417, "name_ru"=>"Блу Ривер", "name_en"=>"Blue River", "lang"=>"en", "lat"=>43.186191, "lon"=>-90.572449),
"/US/US-WI/boaz-22771" => array("type"=>"village", "pop"=>126, "name_ru"=>"Бос", "name_en"=>"Boaz", "lang"=>"en", "lat"=>43.330812, "lon"=>-90.527154),
"/US/US-WI/bohners-lake-22772" => array("type"=>"cdp", "pop"=>1553, "name_en"=>"Bohners Lake", "lang"=>"en", "lat"=>42.622986, "lon"=>-88.280252),
"/US/US-WI/bonduel-22773" => array("type"=>"village", "pop"=>1354, "name_en"=>"Bonduel", "lang"=>"en", "lat"=>44.741289, "lon"=>-88.448352),
"/US/US-WI/boscobel-22774" => array("type"=>"city", "pop"=>3183, "name_en"=>"Boscobel", "lang"=>"en", "lat"=>43.139606, "lon"=>-90.704580),
"/US/US-WI/bowler-22775" => array("type"=>"village", "pop"=>322, "name_en"=>"Bowler", "lang"=>"en", "lat"=>44.863494, "lon"=>-88.982002),
"/US/US-WI/boyceville-22776" => array("type"=>"village", "pop"=>1039, "name_en"=>"Boyceville", "lang"=>"en", "lat"=>45.042612, "lon"=>-92.039915),
"/US/US-WI/boyd-22777" => array("type"=>"village", "pop"=>628, "name_ru"=>"Бойд", "name_en"=>"Boyd", "lang"=>"en", "lat"=>44.953511, "lon"=>-91.040150),
"/US/US-WI/brandon-22778" => array("type"=>"village", "pop"=>870, "name_ru"=>"Брэндон", "name_en"=>"Brandon", "lang"=>"en", "lat"=>43.737380, "lon"=>-88.783843),
"/US/US-WI/brice-prairie-22779" => array("type"=>"cdp", "pop"=>1031, "name_en"=>"Brice Prairie", "lang"=>"en", "lat"=>43.938700, "lon"=>-91.299699),
"/US/US-WI/brillion-22780" => array("type"=>"city", "pop"=>2865, "name_en"=>"Brillion", "lang"=>"en", "lat"=>44.176919, "lon"=>-88.069443),
"/US/US-WI/brodhead-22781" => array("type"=>"city", "pop"=>3124, "name_ru"=>"Бродхэд", "name_en"=>"Brodhead", "lang"=>"en", "lat"=>42.616690, "lon"=>-89.376173),
"/US/US-WI/brokaw-22782" => array("type"=>"village", "pop"=>303, "name_en"=>"Brokaw", "lang"=>"en", "lat"=>45.027370, "lon"=>-89.652652),
"/US/US-WI/brookfield-22783" => array("type"=>"city", "pop"=>39020, "name_ru"=>"Брукфилд", "name_en"=>"Brookfield", "lang"=>"en", "lat"=>43.062800, "lon"=>-88.119761),
"/US/US-WI/brooklyn-22784" => array("type"=>"village", "pop"=>1245, "name_ru"=>"Бруклин", "name_en"=>"Brooklyn", "lang"=>"en", "lat"=>42.852892, "lon"=>-89.372197),
"/US/US-WI/brown-deer-22785" => array("type"=>"village", "pop"=>11716, "name_en"=>"Brown Deer", "lang"=>"en", "lat"=>43.175000, "lon"=>-87.974774),
"/US/US-WI/browns-lake-22786" => array("type"=>"cdp", "pop"=>1725, "name_en"=>"Browns Lake", "lang"=>"en", "lat"=>42.692416, "lon"=>-88.231105),
"/US/US-WI/brownsville-22787" => array("type"=>"village", "pop"=>539, "name_ru"=>"Браунсвилл", "name_en"=>"Brownsville", "lang"=>"en", "lat"=>43.615696, "lon"=>-88.491344),
"/US/US-WI/browntown-22788" => array("type"=>"village", "pop"=>264, "name_en"=>"Browntown", "lang"=>"en", "lat"=>42.577900, "lon"=>-89.791365),
"/US/US-WI/bruce-22789" => array("type"=>"village", "pop"=>682, "name_ru"=>"Брюс", "name_en"=>"Bruce", "lang"=>"en", "lat"=>45.458574, "lon"=>-91.272789),
"/US/US-WI/buffalo-22790" => array("type"=>"city", "pop"=>13741, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>44.223518, "lon"=>-91.869362),
"/US/US-WI/burlington-22791" => array("type"=>"city", "pop"=>10930, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>42.677929, "lon"=>-88.278188),
"/US/US-WI/butler-22792" => array("type"=>"village", "pop"=>1771, "name_ru"=>"Батлер", "name_en"=>"Butler", "lang"=>"en", "lat"=>43.108350, "lon"=>-88.072249),
"/US/US-WI/butternut-22793" => array("type"=>"village", "pop"=>386, "name_en"=>"Butternut", "lang"=>"en", "lat"=>46.013660, "lon"=>-90.497689),
"/US/US-WI/cadott-22794" => array("type"=>"village", "pop"=>1321, "name_en"=>"Cadott", "lang"=>"en", "lat"=>44.951169, "lon"=>-91.151195),
"/US/US-WI/cambria-22795" => array("type"=>"village", "pop"=>766, "name_ru"=>"Камбрия", "name_en"=>"Cambria", "lang"=>"en", "lat"=>43.540853, "lon"=>-89.113826),
"/US/US-WI/cambridge-22796" => array("type"=>"village", "pop"=>1367, "name_ru"=>"Кэмбридж", "name_en"=>"Cambridge", "lang"=>"en", "lat"=>43.007184, "lon"=>-89.018703),
"/US/US-WI/cameron-22797" => array("type"=>"village", "pop"=>1707, "name_ru"=>"Кэмерон", "name_en"=>"Cameron", "lang"=>"en", "lat"=>45.403341, "lon"=>-91.743408),
"/US/US-WI/camp-douglas-22799" => array("type"=>"village", "pop"=>565, "name_en"=>"Camp Douglas", "lang"=>"en", "lat"=>43.918914, "lon"=>-90.268913),
"/US/US-WI/camp-lake-22800" => array("type"=>"cdp", "pop"=>2291, "name_en"=>"Camp Lake", "lang"=>"en", "lat"=>42.530750, "lon"=>-88.145914),
"/US/US-WI/campbellsport-22798" => array("type"=>"village", "pop"=>1940, "name_en"=>"Campbellsport", "lang"=>"en", "lat"=>43.599368, "lon"=>-88.281335),
"/US/US-WI/cascade-22801" => array("type"=>"village", "pop"=>691, "name_ru"=>"Каскейд", "name_en"=>"Cascade", "lang"=>"en", "lat"=>43.659415, "lon"=>-88.008511),
"/US/US-WI/casco-22802" => array("type"=>"village", "pop"=>539, "name_en"=>"Casco", "lang"=>"en", "lat"=>44.553803, "lon"=>-87.618487),
"/US/US-WI/cashton-22803" => array("type"=>"village", "pop"=>1055, "name_en"=>"Cashton", "lang"=>"en", "lat"=>43.743088, "lon"=>-90.781447),
"/US/US-WI/cassville-22804" => array("type"=>"village", "pop"=>1008, "name_ru"=>"Кассвилл", "name_en"=>"Cassville", "lang"=>"en", "lat"=>42.715577, "lon"=>-90.990387),
"/US/US-WI/catawba-22805" => array("type"=>"village", "pop"=>121, "name_en"=>"Catawba", "lang"=>"en", "lat"=>45.535845, "lon"=>-90.532908),
"/US/US-WI/cazenovia-22806" => array("type"=>"village", "pop"=>335, "name_en"=>"Cazenovia", "lang"=>"en", "lat"=>43.525320, "lon"=>-90.198799),
"/US/US-WI/cecil-22807" => array("type"=>"village", "pop"=>556, "name_ru"=>"Сесил", "name_en"=>"Cecil", "lang"=>"en", "lat"=>44.812334, "lon"=>-88.449067),
"/US/US-WI/cedar-grove-22809" => array("type"=>"village", "pop"=>2060, "name_ru"=>"Седар Гроув", "name_en"=>"Cedar Grove", "lang"=>"en", "lat"=>43.567072, "lon"=>-87.823619),
"/US/US-WI/cedarburg-22808" => array("type"=>"city", "pop"=>11101, "name_en"=>"Cedarburg", "lang"=>"en", "lat"=>43.296645, "lon"=>-87.988248),
"/US/US-WI/centuria-22810" => array("type"=>"village", "pop"=>924, "name_en"=>"Centuria", "lang"=>"en", "lat"=>45.449612, "lon"=>-92.555776),
"/US/US-WI/chain-o'-lakes-king-22811" => array("type"=>"cdp", "pop"=>1667, "name_en"=>"Chain O' Lakes-King", "lang"=>"en", "lat"=>44.330802, "lon"=>-89.167803),
"/US/US-WI/chaseburg-22812" => array("type"=>"village", "pop"=>304, "name_en"=>"Chaseburg", "lang"=>"en", "lat"=>43.654770, "lon"=>-91.099945),
"/US/US-WI/chenequa-22813" => array("type"=>"village", "pop"=>590, "name_en"=>"Chenequa", "lang"=>"en", "lat"=>43.123392, "lon"=>-88.381077),
"/US/US-WI/chetek-22814" => array("type"=>"city", "pop"=>2140, "name_en"=>"Chetek", "lang"=>"en", "lat"=>45.315963, "lon"=>-91.653058),
"/US/US-WI/chief-lake-22815" => array("type"=>"cdp", "pop"=>570, "name_en"=>"Chief Lake", "lang"=>"en", "lat"=>45.926636, "lon"=>-91.332767),
"/US/US-WI/chilton-22816" => array("type"=>"city", "pop"=>3576, "name_en"=>"Chilton", "lang"=>"en", "lat"=>44.029557, "lon"=>-88.164517),
"/US/US-WI/chippewa-falls-22817" => array("type"=>"city", "pop"=>12929, "name_en"=>"Chippewa Falls", "lang"=>"en", "lat"=>44.932550, "lon"=>-91.391211),
"/US/US-WI/clayton-22818" => array("type"=>"village", "pop"=>588, "name_ru"=>"Клэйтон", "name_en"=>"Clayton", "lang"=>"en", "lat"=>45.323673, "lon"=>-92.170004),
"/US/US-WI/clear-lake-22819" => array("type"=>"village", "pop"=>1098, "name_ru"=>"Клир Лейк", "name_en"=>"Clear Lake", "lang"=>"en", "lat"=>45.250146, "lon"=>-92.268618),
"/US/US-WI/cleveland-22820" => array("type"=>"village", "pop"=>1403, "name_ru"=>"Кливленд", "name_en"=>"Cleveland", "lang"=>"en", "lat"=>43.919173, "lon"=>-87.747790),
"/US/US-WI/clinton-22821" => array("type"=>"village", "pop"=>3166, "name_ru"=>"Клинтон", "name_en"=>"Clinton", "lang"=>"en", "lat"=>42.556501, "lon"=>-88.866121),
"/US/US-WI/clintonville-22822" => array("type"=>"city", "pop"=>4283, "name_en"=>"Clintonville", "lang"=>"en", "lat"=>44.621542, "lon"=>-88.750988),
"/US/US-WI/clyman-22823" => array("type"=>"village", "pop"=>384, "name_en"=>"Clyman", "lang"=>"en", "lat"=>43.312056, "lon"=>-88.719302),
"/US/US-WI/cobb-22824" => array("type"=>"village", "pop"=>417, "name_ru"=>"Кобб", "name_en"=>"Cobb", "lang"=>"en", "lat"=>42.965273, "lon"=>-90.329348),
"/US/US-WI/cochrane-22825" => array("type"=>"village", "pop"=>390, "name_en"=>"Cochrane", "lang"=>"en", "lat"=>44.228777, "lon"=>-91.836599),
"/US/US-WI/colby-22826" => array("type"=>"city", "pop"=>1681, "name_ru"=>"Колби", "name_en"=>"Colby", "lang"=>"en", "lat"=>44.908939, "lon"=>-90.316923),
"/US/US-WI/coleman-22827" => array("type"=>"village", "pop"=>699, "name_ru"=>"Коулман", "name_en"=>"Coleman", "lang"=>"en", "lat"=>45.070389, "lon"=>-88.033409),
"/US/US-WI/colfax-22828" => array("type"=>"village", "pop"=>1062, "name_ru"=>"Колфакс", "name_en"=>"Colfax", "lang"=>"en", "lat"=>44.996943, "lon"=>-91.725908),
"/US/US-WI/coloma-22829" => array("type"=>"village", "pop"=>465, "name_ru"=>"Колома", "name_en"=>"Coloma", "lang"=>"en", "lat"=>44.033448, "lon"=>-89.520955),
"/US/US-WI/columbus-22830" => array("type"=>"city", "pop"=>5093, "name_ru"=>"Колумбус", "name_en"=>"Columbus", "lang"=>"en", "lat"=>43.335697, "lon"=>-89.028387),
"/US/US-WI/combined-locks-22831" => array("type"=>"village", "pop"=>3247, "name_en"=>"Combined Locks", "lang"=>"en", "lat"=>44.268700, "lon"=>-88.310867),
"/US/US-WI/como-22832" => array("type"=>"cdp", "pop"=>1353, "name_ru"=>"Комо", "name_en"=>"Como", "lang"=>"en", "lat"=>42.609444, "lon"=>-88.495188),
"/US/US-WI/conrath-22833" => array("type"=>"village", "pop"=>85, "name_en"=>"Conrath", "lang"=>"en", "lat"=>45.384702, "lon"=>-91.035821),
"/US/US-WI/coon-valley-22834" => array("type"=>"village", "pop"=>744, "name_en"=>"Coon Valley", "lang"=>"en", "lat"=>43.701805, "lon"=>-91.010211),
"/US/US-WI/cornell-22835" => array("type"=>"city", "pop"=>1385, "name_ru"=>"Корнелл", "name_en"=>"Cornell", "lang"=>"en", "lat"=>45.162463, "lon"=>-91.152300),
"/US/US-WI/cottage-grove-22836" => array("type"=>"village", "pop"=>6166, "name_ru"=>"Коттэдж Гров", "name_en"=>"Cottage Grove", "lang"=>"en", "lat"=>43.085040, "lon"=>-89.202328),
"/US/US-WI/couderay-22837" => array("type"=>"village", "pop"=>98, "name_en"=>"Couderay", "lang"=>"en", "lat"=>45.796173, "lon"=>-91.299163),
"/US/US-WI/crandon-22838" => array("type"=>"city", "pop"=>1816, "name_en"=>"Crandon", "lang"=>"en", "lat"=>45.569905, "lon"=>-88.897260),
"/US/US-WI/crivitz-22839" => array("type"=>"village", "pop"=>1013, "name_en"=>"Crivitz", "lang"=>"en", "lat"=>45.234875, "lon"=>-88.006822),
"/US/US-WI/cross-plains-22840" => array("type"=>"village", "pop"=>3552, "name_en"=>"Cross Plains", "lang"=>"en", "lat"=>43.113922, "lon"=>-89.646178),
"/US/US-WI/cuba-city-22841" => array("type"=>"city", "pop"=>2046, "name_en"=>"Cuba City", "lang"=>"en", "lat"=>42.603086, "lon"=>-90.431332),
"/US/US-WI/cudahy-22842" => array("type"=>"city", "pop"=>18809, "name_ru"=>"Кудаи", "name_en"=>"Cudahy", "lang"=>"en", "lat"=>42.946583, "lon"=>-87.864799),
"/US/US-WI/cumberland-22843" => array("type"=>"city", "pop"=>2240, "name_ru"=>"Камберленд", "name_en"=>"Cumberland", "lang"=>"en", "lat"=>45.535938, "lon"=>-92.023210),
"/US/US-WI/curtiss-22844" => array("type"=>"village", "pop"=>192, "name_en"=>"Curtiss", "lang"=>"en", "lat"=>44.953855, "lon"=>-90.435226),
"/US/US-WI/dallas-22845" => array("type"=>"village", "pop"=>346, "name_ru"=>"Даллас", "name_en"=>"Dallas", "lang"=>"en", "lat"=>45.258061, "lon"=>-91.814664),
"/US/US-WI/dane-22846" => array("type"=>"village", "pop"=>1001, "name_en"=>"Dane", "lang"=>"en", "lat"=>43.250372, "lon"=>-89.500408),
"/US/US-WI/darien-22847" => array("type"=>"village", "pop"=>1696, "name_ru"=>"Дариен", "name_en"=>"Darien", "lang"=>"en", "lat"=>42.600923, "lon"=>-88.710146),
"/US/US-WI/darlington-22848" => array("type"=>"city", "pop"=>2212, "name_ru"=>"Дарлингтон", "name_en"=>"Darlington", "lang"=>"en", "lat"=>42.680454, "lon"=>-90.117002),
"/US/US-WI/de-forest-22851" => array("type"=>"village", "pop"=>4882, "name_en"=>"De Forest", "lang"=>"en", "lat"=>43.245760, "lon"=>-89.345753),
"/US/US-WI/de-pere-22856" => array("type"=>"city", "pop"=>24779, "name_en"=>"De Pere", "lang"=>"en", "lat"=>44.438483, "lon"=>-88.071688),
"/US/US-WI/de-soto-22857" => array("type"=>"village", "pop"=>381, "name_ru"=>"Де Сото", "name_en"=>"De Soto", "lang"=>"en", "lat"=>43.426931, "lon"=>-91.196710),
"/US/US-WI/deer-park-22850" => array("type"=>"village", "pop"=>223, "name_ru"=>"Дир Парк", "name_en"=>"Deer Park", "lang"=>"en", "lat"=>45.188614, "lon"=>-92.386600),
"/US/US-WI/deerfield-22849" => array("type"=>"village", "pop"=>2373, "name_ru"=>"Дирфилд", "name_en"=>"Deerfield", "lang"=>"en", "lat"=>43.051429, "lon"=>-89.074276),
"/US/US-WI/delafield-22852" => array("type"=>"city", "pop"=>6787, "name_en"=>"Delafield", "lang"=>"en", "lat"=>43.069595, "lon"=>-88.391884),
"/US/US-WI/delavan-22853" => array("type"=>"city", "pop"=>8505, "name_ru"=>"Делаван", "name_en"=>"Delavan", "lang"=>"en", "lat"=>42.626683, "lon"=>-88.632569),
"/US/US-WI/delavan-lake-22854" => array("type"=>"cdp", "pop"=>2177, "name_en"=>"Delavan Lake", "lang"=>"en", "lat"=>42.602650, "lon"=>-88.618966),
"/US/US-WI/denmark-22855" => array("type"=>"village", "pop"=>2126, "name_en"=>"Denmark", "lang"=>"en", "lat"=>44.348306, "lon"=>-87.829605),
"/US/US-WI/dickeyville-22858" => array("type"=>"village", "pop"=>1055, "name_en"=>"Dickeyville", "lang"=>"en", "lat"=>42.625758, "lon"=>-90.592082),
"/US/US-WI/dodgeville-22859" => array("type"=>"city", "pop"=>4975, "name_en"=>"Dodgeville", "lang"=>"en", "lat"=>42.963811, "lon"=>-90.130018),
"/US/US-WI/dorchester-22860" => array("type"=>"village", "pop"=>847, "name_ru"=>"Дорчестер", "name_en"=>"Dorchester", "lang"=>"en", "lat"=>45.002299, "lon"=>-90.333341),
"/US/US-WI/dousman-22861" => array("type"=>"village", "pop"=>2052, "name_en"=>"Dousman", "lang"=>"en", "lat"=>43.014840, "lon"=>-88.471363),
"/US/US-WI/downing-22862" => array("type"=>"village", "pop"=>240, "name_ru"=>"Даунинг", "name_en"=>"Downing", "lang"=>"en", "lat"=>45.048539, "lon"=>-92.124101),
"/US/US-WI/doylestown-22863" => array("type"=>"village", "pop"=>340, "name_en"=>"Doylestown", "lang"=>"en", "lat"=>43.427602, "lon"=>-89.146185),
"/US/US-WI/dresser-22864" => array("type"=>"village", "pop"=>852, "name_en"=>"Dresser", "lang"=>"en", "lat"=>45.361513, "lon"=>-92.633756),
"/US/US-WI/durand-22865" => array("type"=>"city", "pop"=>1859, "name_ru"=>"Дуранд", "name_en"=>"Durand", "lang"=>"en", "lat"=>44.629085, "lon"=>-91.960634),
"/US/US-WI/eagle-22866" => array("type"=>"village", "pop"=>1823, "name_ru"=>"Игл", "name_en"=>"Eagle", "lang"=>"en", "lat"=>42.879627, "lon"=>-88.471266),
"/US/US-WI/eagle-lake-22867" => array("type"=>"cdp", "pop"=>1196, "name_ru"=>"Игл Лэйк", "name_en"=>"Eagle Lake", "lang"=>"en", "lat"=>42.707067, "lon"=>-88.128067),
"/US/US-WI/eagle-river-22868" => array("type"=>"city", "pop"=>1724, "name_en"=>"Eagle River", "lang"=>"en", "lat"=>45.924631, "lon"=>-89.257902),
"/US/US-WI/east-troy-22870" => array("type"=>"village", "pop"=>4483, "name_en"=>"East Troy", "lang"=>"en", "lat"=>42.789100, "lon"=>-88.399437),
"/US/US-WI/eastman-22869" => array("type"=>"village", "pop"=>424, "name_ru"=>"Истман", "name_en"=>"Eastman", "lang"=>"en", "lat"=>43.162284, "lon"=>-91.020336),
"/US/US-WI/eau-claire-22871" => array("type"=>"city", "pop"=>65426, "name_ru"=>"О Клэр", "name_en"=>"Eau Claire", "lang"=>"en", "lat"=>44.821150, "lon"=>-91.492274),
"/US/US-WI/eden-22872" => array("type"=>"village", "pop"=>740, "name_ru"=>"Иден", "name_en"=>"Eden", "lang"=>"en", "lat"=>43.692424, "lon"=>-88.362846),
"/US/US-WI/edgar-22873" => array("type"=>"village", "pop"=>1344, "name_en"=>"Edgar", "lang"=>"en", "lat"=>44.922798, "lon"=>-89.963041),
"/US/US-WI/edgerton-22874" => array("type"=>"city", "pop"=>5329, "name_ru"=>"Эджертон", "name_en"=>"Edgerton", "lang"=>"en", "lat"=>42.837840, "lon"=>-89.072558),
"/US/US-WI/egg-harbor-22875" => array("type"=>"village", "pop"=>260, "name_en"=>"Egg Harbor", "lang"=>"en", "lat"=>45.045820, "lon"=>-87.291210),
"/US/US-WI/eland-22876" => array("type"=>"village", "pop"=>228, "name_en"=>"Eland", "lang"=>"en", "lat"=>44.865841, "lon"=>-89.211355),
"/US/US-WI/elderon-22877" => array("type"=>"village", "pop"=>170, "name_en"=>"Elderon", "lang"=>"en", "lat"=>44.783278, "lon"=>-89.250378),
"/US/US-WI/eleva-22878" => array("type"=>"village", "pop"=>637, "name_en"=>"Eleva", "lang"=>"en", "lat"=>44.576391, "lon"=>-91.470307),
"/US/US-WI/elk-mound-22881" => array("type"=>"village", "pop"=>811, "name_en"=>"Elk Mound", "lang"=>"en", "lat"=>44.875054, "lon"=>-91.687709),
"/US/US-WI/elkhart-lake-22879" => array("type"=>"village", "pop"=>1169, "name_en"=>"Elkhart Lake", "lang"=>"en", "lat"=>43.832814, "lon"=>-88.014601),
"/US/US-WI/elkhorn-22880" => array("type"=>"city", "pop"=>9244, "name_ru"=>"Элкхом", "name_en"=>"Elkhorn", "lang"=>"en", "lat"=>42.671037, "lon"=>-88.540251),
"/US/US-WI/ellsworth-22882" => array("type"=>"village", "pop"=>3098, "name_ru"=>"Эллсворт", "name_en"=>"Ellsworth", "lang"=>"en", "lat"=>44.737442, "lon"=>-92.479831),
"/US/US-WI/elm-grove-22883" => array("type"=>"village", "pop"=>6133, "name_en"=>"Elm Grove", "lang"=>"en", "lat"=>43.047650, "lon"=>-88.086661),
"/US/US-WI/elmwood-22884" => array("type"=>"village", "pop"=>799, "name_ru"=>"Элмвуд", "name_en"=>"Elmwood", "lang"=>"en", "lat"=>44.780111, "lon"=>-92.149322),
"/US/US-WI/elmwood-park-22885" => array("type"=>"village", "pop"=>451, "name_ru"=>"Элмвуд Парк", "name_en"=>"Elmwood Park", "lang"=>"en", "lat"=>42.692861, "lon"=>-87.822232),
"/US/US-WI/elroy-22886" => array("type"=>"city", "pop"=>1481, "name_en"=>"Elroy", "lang"=>"en", "lat"=>43.742743, "lon"=>-90.272358),
"/US/US-WI/embarrass-22887" => array("type"=>"village", "pop"=>448, "name_en"=>"Embarrass", "lang"=>"en", "lat"=>44.670735, "lon"=>-88.703342),
"/US/US-WI/endeavor-22888" => array("type"=>"village", "pop"=>461, "name_en"=>"Endeavor", "lang"=>"en", "lat"=>43.714681, "lon"=>-89.468905),
"/US/US-WI/ephraim-22889" => array("type"=>"village", "pop"=>311, "name_en"=>"Ephraim", "lang"=>"en", "lat"=>45.157654, "lon"=>-87.166260),
"/US/US-WI/ettrick-22890" => array("type"=>"village", "pop"=>501, "name_en"=>"Ettrick", "lang"=>"en", "lat"=>44.169834, "lon"=>-91.266124),
"/US/US-WI/evansville-22891" => array("type"=>"city", "pop"=>4957, "name_ru"=>"Эвансвилл", "name_en"=>"Evansville", "lang"=>"en", "lat"=>42.779912, "lon"=>-89.298686),
"/US/US-WI/evergreen-22892" => array("type"=>"cdp", "pop"=>478, "name_ru"=>"Эвергрин", "name_en"=>"Evergreen", "lang"=>"en", "lat"=>44.842464, "lon"=>-89.637385),
"/US/US-WI/exeland-22893" => array("type"=>"village", "pop"=>203, "name_en"=>"Exeland", "lang"=>"en", "lat"=>45.667961, "lon"=>-91.240195),
"/US/US-WI/fairchild-22894" => array("type"=>"village", "pop"=>525, "name_en"=>"Fairchild", "lang"=>"en", "lat"=>44.605457, "lon"=>-90.956416),
"/US/US-WI/fairwater-22895" => array("type"=>"village", "pop"=>356, "name_en"=>"Fairwater", "lang"=>"en", "lat"=>43.741064, "lon"=>-88.868776),
"/US/US-WI/fall-creek-22896" => array("type"=>"village", "pop"=>1276, "name_en"=>"Fall Creek", "lang"=>"en", "lat"=>44.762102, "lon"=>-91.274013),
"/US/US-WI/fall-river-22897" => array("type"=>"village", "pop"=>1515, "name_ru"=>"Фолл Ривер", "name_en"=>"Fall River", "lang"=>"en", "lat"=>43.385146, "lon"=>-89.047729),
"/US/US-WI/fennimore-22898" => array("type"=>"city", "pop"=>2314, "name_en"=>"Fennimore", "lang"=>"en", "lat"=>42.980618, "lon"=>-90.650638),
"/US/US-WI/fenwood-22899" => array("type"=>"village", "pop"=>159, "name_en"=>"Fenwood", "lang"=>"en", "lat"=>44.865518, "lon"=>-90.013627),
"/US/US-WI/ferryville-22900" => array("type"=>"village", "pop"=>177, "name_en"=>"Ferryville", "lang"=>"en", "lat"=>43.343233, "lon"=>-91.083548),
"/US/US-WI/fitchburg-22901" => array("type"=>"city", "pop"=>23344, "name_ru"=>"Фитчбург", "name_en"=>"Fitchburg", "lang"=>"en", "lat"=>42.986000, "lon"=>-89.425142),
"/US/US-WI/fond-du-lac-22902" => array("type"=>"city", "pop"=>42025, "name_en"=>"Fond du Lac", "lang"=>"en", "lat"=>43.773300, "lon"=>-88.445511),
"/US/US-WI/fontana-on-geneva-lake-22903" => array("type"=>"village", "pop"=>1938, "name_en"=>"Fontana-on-Geneva Lake", "lang"=>"en", "lat"=>42.545343, "lon"=>-88.566428),
"/US/US-WI/footville-22904" => array("type"=>"village", "pop"=>745, "name_en"=>"Footville", "lang"=>"en", "lat"=>42.671452, "lon"=>-89.208908),
"/US/US-WI/forestville-22905" => array("type"=>"village", "pop"=>372, "name_ru"=>"Форествилл", "name_en"=>"Forestville", "lang"=>"en", "lat"=>44.690929, "lon"=>-87.478337),
"/US/US-WI/fort-atkinson-22906" => array("type"=>"city", "pop"=>11895, "name_ru"=>"Форт Аткинсон", "name_en"=>"Fort Atkinson", "lang"=>"en", "lat"=>42.926675, "lon"=>-88.841527),
"/US/US-WI/fountain-city-22907" => array("type"=>"city", "pop"=>992, "name_ru"=>"Фаунтейн Сити", "name_en"=>"Fountain City", "lang"=>"en", "lat"=>44.119750, "lon"=>-91.697773),
"/US/US-WI/fox-lake-22908" => array("type"=>"city", "pop"=>1462, "name_ru"=>"Фокс Лейк", "name_en"=>"Fox Lake", "lang"=>"en", "lat"=>43.563384, "lon"=>-88.912607),
"/US/US-WI/fox-point-22909" => array("type"=>"village", "pop"=>6816, "name_en"=>"Fox Point", "lang"=>"en", "lat"=>43.158450, "lon"=>-87.901709),
"/US/US-WI/francis-creek-22910" => array("type"=>"village", "pop"=>673, "name_en"=>"Francis Creek", "lang"=>"en", "lat"=>44.201508, "lon"=>-87.722215),
"/US/US-WI/franklin-22911" => array("type"=>"city", "pop"=>35500, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>42.886300, "lon"=>-88.009824),
"/US/US-WI/frederic-22912" => array("type"=>"village", "pop"=>1197, "name_en"=>"Frederic", "lang"=>"en", "lat"=>45.656422, "lon"=>-92.463968),
"/US/US-WI/fredonia-22913" => array("type"=>"village", "pop"=>2184, "name_ru"=>"Фредониа", "name_en"=>"Fredonia", "lang"=>"en", "lat"=>43.468472, "lon"=>-87.951740),
"/US/US-WI/fremont-22914" => array("type"=>"village", "pop"=>681, "name_ru"=>"Фремонт", "name_en"=>"Fremont", "lang"=>"en", "lat"=>44.259677, "lon"=>-88.870726),
"/US/US-WI/french-island-22915" => array("type"=>"cdp", "pop"=>4478, "name_en"=>"French Island", "lang"=>"en", "lat"=>43.858300, "lon"=>-91.260304),
"/US/US-WI/friendship-22916" => array("type"=>"village", "pop"=>752, "name_ru"=>"Френдшип", "name_en"=>"Friendship", "lang"=>"en", "lat"=>43.971788, "lon"=>-89.819993),
"/US/US-WI/friesland-22917" => array("type"=>"village", "pop"=>296, "name_en"=>"Friesland", "lang"=>"en", "lat"=>43.589002, "lon"=>-89.067759),
"/US/US-WI/galesville-22918" => array("type"=>"city", "pop"=>1447, "name_en"=>"Galesville", "lang"=>"en", "lat"=>44.084182, "lon"=>-91.356227),
"/US/US-WI/gays-mills-22919" => array("type"=>"village", "pop"=>600, "name_en"=>"Gays Mills", "lang"=>"en", "lat"=>43.320091, "lon"=>-90.850354),
"/US/US-WI/genoa-22920" => array("type"=>"village", "pop"=>256, "name_ru"=>"Дженоа", "name_en"=>"Genoa", "lang"=>"en", "lat"=>43.573395, "lon"=>-91.225795),
"/US/US-WI/genoa-city-22921" => array("type"=>"village", "pop"=>2874, "name_en"=>"Genoa City", "lang"=>"en", "lat"=>42.501522, "lon"=>-88.323443),
"/US/US-WI/germantown-22922" => array("type"=>"village", "pop"=>19594, "name_ru"=>"Джермэнтаун", "name_en"=>"Germantown", "lang"=>"en", "lat"=>43.236051, "lon"=>-88.120309),
"/US/US-WI/gillett-22923" => array("type"=>"city", "pop"=>1105, "name_ru"=>"Гиллетт", "name_en"=>"Gillett", "lang"=>"en", "lat"=>44.889757, "lon"=>-88.306321),
"/US/US-WI/gilman-22924" => array("type"=>"village", "pop"=>446, "name_ru"=>"Гилман", "name_en"=>"Gilman", "lang"=>"en", "lat"=>45.166924, "lon"=>-90.806878),
"/US/US-WI/glen-flora-22927" => array("type"=>"village", "pop"=>81, "name_en"=>"Glen Flora", "lang"=>"en", "lat"=>45.496761, "lon"=>-90.893051),
"/US/US-WI/glenbeulah-22925" => array("type"=>"village", "pop"=>449, "name_en"=>"Glenbeulah", "lang"=>"en", "lat"=>43.798635, "lon"=>-88.046650),
"/US/US-WI/glendale-22926" => array("type"=>"city", "pop"=>12992, "name_ru"=>"Глендейл", "name_en"=>"Glendale", "lang"=>"en", "lat"=>43.130050, "lon"=>-87.927633),
"/US/US-WI/glenwood-city-22928" => array("type"=>"city", "pop"=>1238, "name_en"=>"Glenwood City", "lang"=>"en", "lat"=>45.057303, "lon"=>-92.171573),
"/US/US-WI/grafton-22929" => array("type"=>"village", "pop"=>11613, "name_ru"=>"Графтон", "name_en"=>"Grafton", "lang"=>"en", "lat"=>43.318633, "lon"=>-87.954632),
"/US/US-WI/granton-22930" => array("type"=>"village", "pop"=>390, "name_en"=>"Granton", "lang"=>"en", "lat"=>44.588541, "lon"=>-90.462003),
"/US/US-WI/grantsburg-22931" => array("type"=>"village", "pop"=>1356, "name_en"=>"Grantsburg", "lang"=>"en", "lat"=>45.780589, "lon"=>-92.684514),
"/US/US-WI/gratiot-22932" => array("type"=>"village", "pop"=>226, "name_en"=>"Gratiot", "lang"=>"en", "lat"=>42.579493, "lon"=>-90.022887),
"/US/US-WI/green-bay-22933" => array("type"=>"city", "pop"=>101025, "name_en"=>"Green Bay", "lang"=>"en", "lat"=>44.521600, "lon"=>-87.989750),
"/US/US-WI/green-lake-22936" => array("type"=>"city", "pop"=>1095, "name_en"=>"Green Lake", "lang"=>"en", "lat"=>43.842683, "lon"=>-88.955695),
"/US/US-WI/greendale-22934" => array("type"=>"village", "pop"=>13978, "name_ru"=>"Гриндейл", "name_en"=>"Greendale", "lang"=>"en", "lat"=>42.935700, "lon"=>-88.001549),
"/US/US-WI/greenfield-22935" => array("type"=>"city", "pop"=>36075, "name_ru"=>"Гринфилд", "name_en"=>"Greenfield", "lang"=>"en", "lat"=>42.962767, "lon"=>-88.005603),
"/US/US-WI/greenwood-22937" => array("type"=>"city", "pop"=>1052, "name_ru"=>"Гринвуд", "name_en"=>"Greenwood", "lang"=>"en", "lat"=>44.765429, "lon"=>-90.598809),
"/US/US-WI/gresham-22938" => array("type"=>"village", "pop"=>563, "name_en"=>"Gresham", "lang"=>"en", "lat"=>44.848576, "lon"=>-88.785948),
"/US/US-WI/hales-corners-22939" => array("type"=>"village", "pop"=>7574, "name_en"=>"Hales Corners", "lang"=>"en", "lat"=>42.940500, "lon"=>-88.049324),
"/US/US-WI/hammond-22940" => array("type"=>"village", "pop"=>1723, "name_ru"=>"Хаммонд", "name_en"=>"Hammond", "lang"=>"en", "lat"=>44.972970, "lon"=>-92.437092),
"/US/US-WI/hancock-22941" => array("type"=>"village", "pop"=>452, "name_ru"=>"Хэнкок", "name_en"=>"Hancock", "lang"=>"en", "lat"=>44.134516, "lon"=>-89.520244),
"/US/US-WI/hartford-22942" => array("type"=>"city", "pop"=>13853, "name_ru"=>"Хартфорд", "name_en"=>"Hartford", "lang"=>"en", "lat"=>43.323656, "lon"=>-88.388777),
"/US/US-WI/hartland-22943" => array("type"=>"village", "pop"=>8692, "name_ru"=>"Хартленд", "name_en"=>"Hartland", "lang"=>"en", "lat"=>43.098216, "lon"=>-88.344619),
"/US/US-WI/hatley-22944" => array("type"=>"village", "pop"=>509, "name_ru"=>"Хэтли", "name_en"=>"Hatley", "lang"=>"en", "lat"=>44.887066, "lon"=>-89.337014),
"/US/US-WI/haugen-22945" => array("type"=>"village", "pop"=>270, "name_en"=>"Haugen", "lang"=>"en", "lat"=>45.607775, "lon"=>-91.778431),
"/US/US-WI/hawkins-22946" => array("type"=>"village", "pop"=>284, "name_en"=>"Hawkins", "lang"=>"en", "lat"=>45.511756, "lon"=>-90.713325),
"/US/US-WI/hayward-22947" => array("type"=>"city", "pop"=>2342, "name_ru"=>"Хейворд", "name_en"=>"Hayward", "lang"=>"en", "lat"=>46.010032, "lon"=>-91.480295),
"/US/US-WI/hazel-green-22948" => array("type"=>"village", "pop"=>1177, "name_ru"=>"Хазел Грин", "name_en"=>"Hazel Green", "lang"=>"en", "lat"=>42.534316, "lon"=>-90.435613),
"/US/US-WI/hewitt-22949" => array("type"=>"village", "pop"=>712, "name_ru"=>"Хьюитт", "name_en"=>"Hewitt", "lang"=>"en", "lat"=>44.645195, "lon"=>-90.105108),
"/US/US-WI/highland-22950" => array("type"=>"village", "pop"=>801, "name_ru"=>"Хайленд", "name_en"=>"Highland", "lang"=>"en", "lat"=>43.046356, "lon"=>-90.378959),
"/US/US-WI/hilbert-22951" => array("type"=>"village", "pop"=>1085, "name_en"=>"Hilbert", "lang"=>"en", "lat"=>44.140050, "lon"=>-88.159149),
"/US/US-WI/hillsboro-22952" => array("type"=>"city", "pop"=>1345, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>43.654063, "lon"=>-90.337701),
"/US/US-WI/hixton-22953" => array("type"=>"village", "pop"=>437, "name_en"=>"Hixton", "lang"=>"en", "lat"=>44.384030, "lon"=>-91.011053),
"/US/US-WI/hollandale-22954" => array("type"=>"village", "pop"=>261, "name_ru"=>"Холландейл", "name_en"=>"Hollandale", "lang"=>"en", "lat"=>42.876122, "lon"=>-89.933288),
"/US/US-WI/holmen-22955" => array("type"=>"village", "pop"=>8487, "name_en"=>"Holmen", "lang"=>"en", "lat"=>43.956921, "lon"=>-91.258994),
"/US/US-WI/horicon-22956" => array("type"=>"city", "pop"=>3495, "name_en"=>"Horicon", "lang"=>"en", "lat"=>43.444904, "lon"=>-88.637058),
"/US/US-WI/hortonville-22957" => array("type"=>"village", "pop"=>2736, "name_en"=>"Hortonville", "lang"=>"en", "lat"=>44.338787, "lon"=>-88.632860),
"/US/US-WI/howard-22958" => array("type"=>"village", "pop"=>16753, "name_ru"=>"Ховарт", "name_en"=>"Howard", "lang"=>"en", "lat"=>44.573350, "lon"=>-88.076757),
"/US/US-WI/howards-grove-22959" => array("type"=>"village", "pop"=>3178, "name_en"=>"Howards Grove", "lang"=>"en", "lat"=>43.829600, "lon"=>-87.825919),
"/US/US-WI/hudson-22960" => array("type"=>"city", "pop"=>12196, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>44.966950, "lon"=>-92.742099),
"/US/US-WI/hurley-22961" => array("type"=>"city", "pop"=>1543, "name_ru"=>"Херли", "name_en"=>"Hurley", "lang"=>"en", "lat"=>46.444265, "lon"=>-90.196466),
"/US/US-WI/hustisford-22962" => array("type"=>"village", "pop"=>1097, "name_en"=>"Hustisford", "lang"=>"en", "lat"=>43.345541, "lon"=>-88.603622),
"/US/US-WI/hustler-22963" => array("type"=>"village", "pop"=>117, "name_en"=>"Hustler", "lang"=>"en", "lat"=>43.879339, "lon"=>-90.268242),
"/US/US-WI/independence-22964" => array("type"=>"city", "pop"=>1212, "name_ru"=>"Индепенденс", "name_en"=>"Independence", "lang"=>"en", "lat"=>44.362159, "lon"=>-91.419308),
"/US/US-WI/ingram-22965" => array("type"=>"village", "pop"=>68, "name_en"=>"Ingram", "lang"=>"en", "lat"=>45.505506, "lon"=>-90.813220),
"/US/US-WI/iola-22966" => array("type"=>"village", "pop"=>1243, "name_ru"=>"Айола", "name_en"=>"Iola", "lang"=>"en", "lat"=>44.509410, "lon"=>-89.126331),
"/US/US-WI/iron-ridge-22967" => array("type"=>"village", "pop"=>977, "name_en"=>"Iron Ridge", "lang"=>"en", "lat"=>43.397558, "lon"=>-88.532063),
"/US/US-WI/ironton-22968" => array("type"=>"village", "pop"=>260, "name_ru"=>"Айронтон", "name_en"=>"Ironton", "lang"=>"en", "lat"=>43.544725, "lon"=>-90.141651),
"/US/US-WI/jackson-22969" => array("type"=>"village", "pop"=>6755, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>43.320912, "lon"=>-88.163873),
"/US/US-WI/janesville-22970" => array("type"=>"city", "pop"=>62516, "name_ru"=>"Джейнсвилл", "name_en"=>"Janesville", "lang"=>"en", "lat"=>42.684650, "lon"=>-89.015552),
"/US/US-WI/jefferson-22971" => array("type"=>"city", "pop"=>7822, "name_ru"=>"Джефферсон", "name_en"=>"Jefferson", "lang"=>"en", "lat"=>43.006436, "lon"=>-88.808069),
"/US/US-WI/johnson-creek-22972" => array("type"=>"village", "pop"=>2261, "name_en"=>"Johnson Creek", "lang"=>"en", "lat"=>43.079808, "lon"=>-88.773916),
"/US/US-WI/junction-city-22973" => array("type"=>"village", "pop"=>410, "name_ru"=>"Джанкшен Сити", "name_en"=>"Junction City", "lang"=>"en", "lat"=>44.591362, "lon"=>-89.766017),
"/US/US-WI/juneau-22974" => array("type"=>"city", "pop"=>2601, "name_ru"=>"Джуно", "name_en"=>"Juneau", "lang"=>"en", "lat"=>43.406257, "lon"=>-88.703296),
"/US/US-WI/kaukauna-22975" => array("type"=>"city", "pop"=>15414, "name_en"=>"Kaukauna", "lang"=>"en", "lat"=>44.278967, "lon"=>-88.268384),
"/US/US-WI/kekoskee-22976" => array("type"=>"village", "pop"=>159, "name_en"=>"Kekoskee", "lang"=>"en", "lat"=>43.527633, "lon"=>-88.561710),
"/US/US-WI/kellnersville-22977" => array("type"=>"village", "pop"=>356, "name_en"=>"Kellnersville", "lang"=>"en", "lat"=>44.224237, "lon"=>-87.803128),
"/US/US-WI/kendall-22978" => array("type"=>"village", "pop"=>463, "name_ru"=>"Кендалл", "name_en"=>"Kendall", "lang"=>"en", "lat"=>43.793155, "lon"=>-90.367616),
"/US/US-WI/kennan-22979" => array("type"=>"village", "pop"=>148, "name_en"=>"Kennan", "lang"=>"en", "lat"=>45.530899, "lon"=>-90.586087),
"/US/US-WI/kenosha-22980" => array("type"=>"city", "pop"=>96950, "name_en"=>"Kenosha", "lang"=>"en", "lat"=>42.584100, "lon"=>-87.854549),
"/US/US-WI/keshena-22981" => array("type"=>"cdp", "pop"=>685, "name_en"=>"Keshena", "lang"=>"en", "lat"=>44.883844, "lon"=>-88.636192),
"/US/US-WI/kewaskum-22982" => array("type"=>"village", "pop"=>3927, "name_en"=>"Kewaskum", "lang"=>"en", "lat"=>43.521877, "lon"=>-88.230414),
"/US/US-WI/kewaunee-22983" => array("type"=>"city", "pop"=>2753, "name_en"=>"Kewaunee", "lang"=>"en", "lat"=>44.460324, "lon"=>-87.511201),
"/US/US-WI/kiel-22984" => array("type"=>"city", "pop"=>3587, "name_en"=>"Kiel", "lang"=>"en", "lat"=>43.916521, "lon"=>-88.030462),
"/US/US-WI/kimberly-22985" => array("type"=>"village", "pop"=>6244, "name_ru"=>"Кимберли", "name_en"=>"Kimberly", "lang"=>"en", "lat"=>44.270032, "lon"=>-88.340681),
"/US/US-WI/kingston-22986" => array("type"=>"village", "pop"=>279, "name_ru"=>"Кингстон", "name_en"=>"Kingston", "lang"=>"en", "lat"=>43.693582, "lon"=>-89.131265),
"/US/US-WI/knapp-22987" => array("type"=>"village", "pop"=>415, "name_en"=>"Knapp", "lang"=>"en", "lat"=>44.954100, "lon"=>-92.076950),
"/US/US-WI/kohler-22988" => array("type"=>"village", "pop"=>1949, "name_en"=>"Kohler", "lang"=>"en", "lat"=>43.735450, "lon"=>-87.779039),
"/US/US-WI/la-crosse-22991" => array("type"=>"city", "pop"=>50902, "name_ru"=>"Ла Кросс", "name_en"=>"La Crosse", "lang"=>"en", "lat"=>43.826950, "lon"=>-91.233899),
"/US/US-WI/la-farge-22993" => array("type"=>"village", "pop"=>788, "name_en"=>"La Farge", "lang"=>"en", "lat"=>43.577502, "lon"=>-90.638132),
"/US/US-WI/la-valle-23004" => array("type"=>"village", "pop"=>317, "name_en"=>"La Valle", "lang"=>"en", "lat"=>43.582718, "lon"=>-90.130622),
"/US/US-WI/lac-du-flambeau-22989" => array("type"=>"cdp", "pop"=>3081, "name_en"=>"Lac du Flambeau", "lang"=>"en", "lat"=>45.970538, "lon"=>-89.902728),
"/US/US-WI/lac-la-belle-22990" => array("type"=>"village", "pop"=>433, "name_en"=>"Lac La Belle", "lang"=>"en", "lat"=>43.145456, "lon"=>-88.525658),
"/US/US-WI/ladysmith-22992" => array("type"=>"city", "pop"=>3339, "name_en"=>"Ladysmith", "lang"=>"en", "lat"=>45.462751, "lon"=>-91.095931),
"/US/US-WI/lake-delton-22994" => array("type"=>"village", "pop"=>3155, "name_en"=>"Lake Delton", "lang"=>"en", "lat"=>43.593308, "lon"=>-89.780369),
"/US/US-WI/lake-geneva-22995" => array("type"=>"city", "pop"=>8371, "name_en"=>"Lake Geneva", "lang"=>"en", "lat"=>42.588647, "lon"=>-88.429638),
"/US/US-WI/lake-mills-22996" => array("type"=>"city", "pop"=>5557, "name_ru"=>"Лейк Миллс", "name_en"=>"Lake Mills", "lang"=>"en", "lat"=>43.074024, "lon"=>-88.907965),
"/US/US-WI/lake-nebagamon-22997" => array("type"=>"village", "pop"=>1089, "name_en"=>"Lake Nebagamon", "lang"=>"en", "lat"=>46.507797, "lon"=>-91.694504),
"/US/US-WI/lake-ripley-22998" => array("type"=>"cdp", "pop"=>1218, "name_en"=>"Lake Ripley", "lang"=>"en", "lat"=>43.005707, "lon"=>-88.986063),
"/US/US-WI/lake-wazeecha-22999" => array("type"=>"cdp", "pop"=>2278, "name_en"=>"Lake Wazeecha", "lang"=>"en", "lat"=>44.371240, "lon"=>-89.756259),
"/US/US-WI/lake-wisconsin-23000" => array("type"=>"cdp", "pop"=>2268, "name_en"=>"Lake Wisconsin", "lang"=>"en", "lat"=>43.373475, "lon"=>-89.575555),
"/US/US-WI/lake-wissota-23001" => array("type"=>"cdp", "pop"=>2175, "name_en"=>"Lake Wissota", "lang"=>"en", "lat"=>44.920204, "lon"=>-91.302541),
"/US/US-WI/lancaster-23002" => array("type"=>"city", "pop"=>3920, "name_ru"=>"Ланкастер", "name_en"=>"Lancaster", "lang"=>"en", "lat"=>42.848538, "lon"=>-90.707989),
"/US/US-WI/lannon-23003" => array("type"=>"village", "pop"=>1130, "name_en"=>"Lannon", "lang"=>"en", "lat"=>43.151373, "lon"=>-88.162408),
"/US/US-WI/lena-23005" => array("type"=>"village", "pop"=>464, "name_ru"=>"Лена", "name_en"=>"Lena", "lang"=>"en", "lat"=>44.952656, "lon"=>-88.050115),
"/US/US-WI/lime-ridge-23006" => array("type"=>"village", "pop"=>153, "name_en"=>"Lime Ridge", "lang"=>"en", "lat"=>43.468230, "lon"=>-90.156421),
"/US/US-WI/linden-23007" => array("type"=>"village", "pop"=>557, "name_ru"=>"Линден", "name_en"=>"Linden", "lang"=>"en", "lat"=>42.919199, "lon"=>-90.273959),
"/US/US-WI/little-chute-23008" => array("type"=>"village", "pop"=>11121, "name_en"=>"Little Chute", "lang"=>"en", "lat"=>44.285750, "lon"=>-88.313499),
"/US/US-WI/little-round-lake-23009" => array("type"=>"cdp", "pop"=>871, "name_en"=>"Little Round Lake", "lang"=>"en", "lat"=>45.964992, "lon"=>-91.367782),
"/US/US-WI/livingston-23010" => array("type"=>"village", "pop"=>573, "name_ru"=>"Ливингстон", "name_en"=>"Livingston", "lang"=>"en", "lat"=>42.900722, "lon"=>-90.433738),
"/US/US-WI/lodi-23011" => array("type"=>"city", "pop"=>3008, "name_ru"=>"Лоди", "name_en"=>"Lodi", "lang"=>"en", "lat"=>43.316400, "lon"=>-89.532069),
"/US/US-WI/loganville-23012" => array("type"=>"village", "pop"=>280, "name_ru"=>"Логанвилл", "name_en"=>"Loganville", "lang"=>"en", "lat"=>43.439833, "lon"=>-90.036969),
"/US/US-WI/lohrville-23013" => array("type"=>"village", "pop"=>405, "name_ru"=>"Лорвилл", "name_en"=>"Lohrville", "lang"=>"en", "lat"=>44.037960, "lon"=>-89.121107),
"/US/US-WI/lomira-23014" => array("type"=>"village", "pop"=>2333, "name_en"=>"Lomira", "lang"=>"en", "lat"=>43.588856, "lon"=>-88.444716),
"/US/US-WI/lone-rock-23015" => array("type"=>"village", "pop"=>877, "name_ru"=>"Лоун Рок", "name_en"=>"Lone Rock", "lang"=>"en", "lat"=>43.184465, "lon"=>-90.201710),
"/US/US-WI/lowell-23016" => array("type"=>"village", "pop"=>353, "name_ru"=>"Лоуэлл", "name_en"=>"Lowell", "lang"=>"en", "lat"=>43.337812, "lon"=>-88.820933),
"/US/US-WI/loyal-23017" => array("type"=>"city", "pop"=>1244, "name_en"=>"Loyal", "lang"=>"en", "lat"=>44.736116, "lon"=>-90.496413),
"/US/US-WI/lublin-23018" => array("type"=>"village", "pop"=>101, "name_en"=>"Lublin", "lang"=>"en", "lat"=>45.075175, "lon"=>-90.724040),
"/US/US-WI/luck-23019" => array("type"=>"village", "pop"=>1181, "name_en"=>"Luck", "lang"=>"en", "lat"=>45.575168, "lon"=>-92.467295),
"/US/US-WI/luxemburg-23020" => array("type"=>"village", "pop"=>2237, "name_ru"=>"Люксембург", "name_en"=>"Luxemburg", "lang"=>"en", "lat"=>44.542622, "lon"=>-87.705267),
"/US/US-WI/lyndon-station-23021" => array("type"=>"village", "pop"=>455, "name_en"=>"Lyndon Station", "lang"=>"en", "lat"=>43.710098, "lon"=>-89.894433),
"/US/US-WI/lynxville-23022" => array("type"=>"village", "pop"=>162, "name_en"=>"Lynxville", "lang"=>"en", "lat"=>43.247984, "lon"=>-91.047858),
"/US/US-WI/madison-23024" => array("type"=>"city", "pop"=>231916, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>43.079800, "lon"=>-89.387519),
"/US/US-WI/maiden-rock-23025" => array("type"=>"village", "pop"=>146, "name_en"=>"Maiden Rock", "lang"=>"en", "lat"=>44.567845, "lon"=>-92.308361),
"/US/US-WI/manawa-23026" => array("type"=>"city", "pop"=>1264, "name_en"=>"Manawa", "lang"=>"en", "lat"=>44.461215, "lon"=>-88.920381),
"/US/US-WI/manitowoc-23027" => array("type"=>"city", "pop"=>33169, "name_en"=>"Manitowoc", "lang"=>"en", "lat"=>44.098633, "lon"=>-87.677420),
"/US/US-WI/maple-bluff-23028" => array("type"=>"village", "pop"=>1360, "name_en"=>"Maple Bluff", "lang"=>"en", "lat"=>43.109807, "lon"=>-89.369435),
"/US/US-WI/marathon-city-23029" => array("type"=>"village", "pop"=>1540, "name_en"=>"Marathon City", "lang"=>"en", "lat"=>44.932497, "lon"=>-89.841754),
"/US/US-WI/maribel-23030" => array("type"=>"village", "pop"=>276, "name_en"=>"Maribel", "lang"=>"en", "lat"=>44.279400, "lon"=>-87.802052),
"/US/US-WI/marinette-23031" => array("type"=>"city", "pop"=>10800, "name_en"=>"Marinette", "lang"=>"en", "lat"=>45.088543, "lon"=>-87.626904),
"/US/US-WI/marion-23032" => array("type"=>"city", "pop"=>1177, "name_ru"=>"Марион", "name_en"=>"Marion", "lang"=>"en", "lat"=>44.671223, "lon"=>-88.887394),
"/US/US-WI/markesan-23033" => array("type"=>"city", "pop"=>1282, "name_en"=>"Markesan", "lang"=>"en", "lat"=>43.709224, "lon"=>-88.989691),
"/US/US-WI/marquette-23034" => array("type"=>"village", "pop"=>162, "name_ru"=>"Маркетт", "name_en"=>"Marquette", "lang"=>"en", "lat"=>43.742211, "lon"=>-89.138791),
"/US/US-WI/marshall-23035" => array("type"=>"village", "pop"=>3744, "name_ru"=>"Маршалл", "name_en"=>"Marshall", "lang"=>"en", "lat"=>43.172540, "lon"=>-89.064060),
"/US/US-WI/marshfield-23036" => array("type"=>"city", "pop"=>18267, "name_ru"=>"Маршфилд", "name_en"=>"Marshfield", "lang"=>"en", "lat"=>44.663487, "lon"=>-90.174946),
"/US/US-WI/mason-23037" => array("type"=>"village", "pop"=>80, "name_ru"=>"Мейсон", "name_en"=>"Mason", "lang"=>"en", "lat"=>46.435505, "lon"=>-91.062269),
"/US/US-WI/mattoon-23038" => array("type"=>"village", "pop"=>424, "name_ru"=>"Мэттун", "name_en"=>"Mattoon", "lang"=>"en", "lat"=>45.004731, "lon"=>-89.042641),
"/US/US-WI/mauston-23039" => array("type"=>"city", "pop"=>4413, "name_en"=>"Mauston", "lang"=>"en", "lat"=>43.800563, "lon"=>-90.081053),
"/US/US-WI/mayville-23040" => array("type"=>"city", "pop"=>5090, "name_ru"=>"Мейвилл", "name_en"=>"Mayville", "lang"=>"en", "lat"=>43.498216, "lon"=>-88.546220),
"/US/US-WI/mazomanie-23041" => array("type"=>"village", "pop"=>1680, "name_en"=>"Mazomanie", "lang"=>"en", "lat"=>43.174396, "lon"=>-89.794837),
"/US/US-WI/mcfarland-23023" => array("type"=>"village", "pop"=>7767, "name_ru"=>"Мак-Фарленд", "name_en"=>"McFarland", "lang"=>"en", "lat"=>43.018850, "lon"=>-89.293863),
"/US/US-WI/medford-23042" => array("type"=>"city", "pop"=>4115, "name_ru"=>"Медфорд", "name_en"=>"Medford", "lang"=>"en", "lat"=>45.138647, "lon"=>-90.347897),
"/US/US-WI/mellen-23043" => array("type"=>"city", "pop"=>775, "name_en"=>"Mellen", "lang"=>"en", "lat"=>46.323264, "lon"=>-90.661106),
"/US/US-WI/melrose-23044" => array("type"=>"village", "pop"=>500, "name_ru"=>"Мелроуз", "name_en"=>"Melrose", "lang"=>"en", "lat"=>44.132030, "lon"=>-90.996847),
"/US/US-WI/melvina-23045" => array("type"=>"village", "pop"=>88, "name_en"=>"Melvina", "lang"=>"en", "lat"=>43.804288, "lon"=>-90.780265),
"/US/US-WI/menasha-23046" => array("type"=>"city", "pop"=>16627, "name_en"=>"Menasha", "lang"=>"en", "lat"=>44.211650, "lon"=>-88.436732),
"/US/US-WI/menomonee-falls-23047" => array("type"=>"village", "pop"=>34607, "name_en"=>"Menomonee Falls", "lang"=>"en", "lat"=>43.147168, "lon"=>-88.125927),
"/US/US-WI/menomonie-23048" => array("type"=>"city", "pop"=>15581, "name_en"=>"Menomonie", "lang"=>"en", "lat"=>44.887891, "lon"=>-91.913986),
"/US/US-WI/mequon-23049" => array("type"=>"city", "pop"=>23584, "name_en"=>"Mequon", "lang"=>"en", "lat"=>43.235883, "lon"=>-87.989257),
"/US/US-WI/merrill-23050" => array("type"=>"city", "pop"=>9670, "name_ru"=>"Меррил", "name_en"=>"Merrill", "lang"=>"en", "lat"=>45.182023, "lon"=>-89.703459),
"/US/US-WI/merrillan-23051" => array("type"=>"village", "pop"=>565, "name_en"=>"Merrillan", "lang"=>"en", "lat"=>44.450990, "lon"=>-90.836624),
"/US/US-WI/merrimac-23052" => array("type"=>"village", "pop"=>464, "name_en"=>"Merrimac", "lang"=>"en", "lat"=>43.373719, "lon"=>-89.625363),
"/US/US-WI/merton-23053" => array("type"=>"village", "pop"=>2973, "name_en"=>"Merton", "lang"=>"en", "lat"=>43.140976, "lon"=>-88.313537),
"/US/US-WI/middleton-23054" => array("type"=>"city", "pop"=>16244, "name_ru"=>"Мидлтон", "name_en"=>"Middleton", "lang"=>"en", "lat"=>43.102867, "lon"=>-89.506333),
"/US/US-WI/milladore-23055" => array("type"=>"village", "pop"=>252, "name_en"=>"Milladore", "lang"=>"en", "lat"=>44.604636, "lon"=>-89.854138),
"/US/US-WI/milltown-23056" => array("type"=>"village", "pop"=>880, "name_ru"=>"Миллтаун", "name_en"=>"Milltown", "lang"=>"en", "lat"=>45.526850, "lon"=>-92.499656),
"/US/US-WI/milton-23057" => array("type"=>"city", "pop"=>5579, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>42.778943, "lon"=>-88.949419),
"/US/US-WI/milwaukee-23058" => array("type"=>"city", "pop"=>604477, "name_en"=>"Milwaukee", "lang"=>"en", "lat"=>43.063350, "lon"=>-87.966623),
"/US/US-WI/mineral-point-23059" => array("type"=>"city", "pop"=>2462, "name_ru"=>"Минерал Пойнт", "name_en"=>"Mineral Point", "lang"=>"en", "lat"=>42.861927, "lon"=>-90.183084),
"/US/US-WI/minong-23060" => array("type"=>"village", "pop"=>565, "name_en"=>"Minong", "lang"=>"en", "lat"=>46.098378, "lon"=>-91.823932),
"/US/US-WI/mishicot-23061" => array("type"=>"village", "pop"=>1384, "name_en"=>"Mishicot", "lang"=>"en", "lat"=>44.230808, "lon"=>-87.643415),
"/US/US-WI/mondovi-23062" => array("type"=>"city", "pop"=>2561, "name_en"=>"Mondovi", "lang"=>"en", "lat"=>44.568901, "lon"=>-91.669475),
"/US/US-WI/monona-23063" => array("type"=>"city", "pop"=>7837, "name_ru"=>"Монона", "name_en"=>"Monona", "lang"=>"en", "lat"=>43.054400, "lon"=>-89.333201),
"/US/US-WI/monroe-23064" => array("type"=>"city", "pop"=>10481, "name_ru"=>"Монро", "name_en"=>"Monroe", "lang"=>"en", "lat"=>42.599781, "lon"=>-89.643843),
"/US/US-WI/montello-23065" => array("type"=>"city", "pop"=>1459, "name_en"=>"Montello", "lang"=>"en", "lat"=>43.794295, "lon"=>-89.334360),
"/US/US-WI/montfort-23066" => array("type"=>"village", "pop"=>668, "name_en"=>"Montfort", "lang"=>"en", "lat"=>42.970174, "lon"=>-90.432851),
"/US/US-WI/monticello-23067" => array("type"=>"village", "pop"=>1135, "name_ru"=>"Монтиселло", "name_en"=>"Monticello", "lang"=>"en", "lat"=>42.745803, "lon"=>-89.589668),
"/US/US-WI/montreal-23068" => array("type"=>"city", "pop"=>713, "name_en"=>"Montreal", "lang"=>"en", "lat"=>46.430482, "lon"=>-90.238795),
"/US/US-WI/mosinee-23069" => array("type"=>"city", "pop"=>3973, "name_en"=>"Mosinee", "lang"=>"en", "lat"=>44.787213, "lon"=>-89.683445),
"/US/US-WI/mount-calvary-23070" => array("type"=>"village", "pop"=>910, "name_en"=>"Mount Calvary", "lang"=>"en", "lat"=>43.826503, "lon"=>-88.246755),
"/US/US-WI/mount-hope-23071" => array("type"=>"village", "pop"=>176, "name_ru"=>"Маунт Хоуп", "name_en"=>"Mount Hope", "lang"=>"en", "lat"=>42.969668, "lon"=>-90.858593),
"/US/US-WI/mount-horeb-23072" => array("type"=>"village", "pop"=>6703, "name_en"=>"Mount Horeb", "lang"=>"en", "lat"=>43.009062, "lon"=>-89.733630),
"/US/US-WI/mount-sterling-23073" => array("type"=>"village", "pop"=>196, "name_ru"=>"Маунт Стерлинг", "name_en"=>"Mount Sterling", "lang"=>"en", "lat"=>43.314323, "lon"=>-90.930626),
"/US/US-WI/mukwonago-23074" => array("type"=>"village", "pop"=>7232, "name_en"=>"Mukwonago", "lang"=>"en", "lat"=>42.866336, "lon"=>-88.331215),
"/US/US-WI/muscoda-23075" => array("type"=>"village", "pop"=>1352, "name_en"=>"Muscoda", "lang"=>"en", "lat"=>43.187281, "lon"=>-90.434221),
"/US/US-WI/muskego-23076" => array("type"=>"city", "pop"=>23234, "name_en"=>"Muskego", "lang"=>"en", "lat"=>42.891104, "lon"=>-88.122910),
"/US/US-WI/nashotah-23077" => array("type"=>"village", "pop"=>1361, "name_en"=>"Nashotah", "lang"=>"en", "lat"=>43.093016, "lon"=>-88.401344),
"/US/US-WI/necedah-23078" => array("type"=>"village", "pop"=>853, "name_en"=>"Necedah", "lang"=>"en", "lat"=>44.024568, "lon"=>-90.071711),
"/US/US-WI/neenah-23079" => array("type"=>"city", "pop"=>25060, "name_en"=>"Neenah", "lang"=>"en", "lat"=>44.169533, "lon"=>-88.472189),
"/US/US-WI/neillsville-23080" => array("type"=>"city", "pop"=>2591, "name_en"=>"Neillsville", "lang"=>"en", "lat"=>44.560423, "lon"=>-90.590888),
"/US/US-WI/nekoosa-23081" => array("type"=>"city", "pop"=>2439, "name_en"=>"Nekoosa", "lang"=>"en", "lat"=>44.313159, "lon"=>-89.907954),
"/US/US-WI/nelson-23082" => array("type"=>"village", "pop"=>372, "name_ru"=>"Нельсон", "name_en"=>"Nelson", "lang"=>"en", "lat"=>44.423555, "lon"=>-92.004870),
"/US/US-WI/nelsonville-23083" => array("type"=>"village", "pop"=>172, "name_en"=>"Nelsonville", "lang"=>"en", "lat"=>44.494729, "lon"=>-89.310204),
"/US/US-WI/neopit-23084" => array("type"=>"cdp", "pop"=>615, "name_en"=>"Neopit", "lang"=>"en", "lat"=>44.981173, "lon"=>-88.826442),
"/US/US-WI/neosho-23085" => array("type"=>"village", "pop"=>558, "name_ru"=>"Неошо", "name_en"=>"Neosho", "lang"=>"en", "lat"=>43.309733, "lon"=>-88.517841),
"/US/US-WI/neshkoro-23086" => array("type"=>"village", "pop"=>440, "name_en"=>"Neshkoro", "lang"=>"en", "lat"=>43.965186, "lon"=>-89.213888),
"/US/US-WI/new-auburn-23087" => array("type"=>"village", "pop"=>558, "name_ru"=>"Нью Обурн", "name_en"=>"New Auburn", "lang"=>"en", "lat"=>45.198078, "lon"=>-91.565015),
"/US/US-WI/new-berlin-23088" => array("type"=>"city", "pop"=>38649, "name_ru"=>"Нью Берлин", "name_en"=>"New Berlin", "lang"=>"en", "lat"=>42.972783, "lon"=>-88.129221),
"/US/US-WI/new-glarus-23090" => array("type"=>"village", "pop"=>2071, "name_en"=>"New Glarus", "lang"=>"en", "lat"=>42.813192, "lon"=>-89.634590),
"/US/US-WI/new-holstein-23091" => array("type"=>"city", "pop"=>3164, "name_en"=>"New Holstein", "lang"=>"en", "lat"=>43.949283, "lon"=>-88.094767),
"/US/US-WI/new-lisbon-23092" => array("type"=>"city", "pop"=>2372, "name_en"=>"New Lisbon", "lang"=>"en", "lat"=>43.878560, "lon"=>-90.164688),
"/US/US-WI/new-london-23093" => array("type"=>"city", "pop"=>6745, "name_ru"=>"Нью Лондон", "name_en"=>"New London", "lang"=>"en", "lat"=>44.394481, "lon"=>-88.739556),
"/US/US-WI/new-post-23094" => array("type"=>"cdp", "pop"=>243, "name_en"=>"New Post", "lang"=>"en", "lat"=>45.897165, "lon"=>-91.191951),
"/US/US-WI/new-richmond-23095" => array("type"=>"city", "pop"=>8259, "name_ru"=>"Нью Ричмонд", "name_en"=>"New Richmond", "lang"=>"en", "lat"=>45.122087, "lon"=>-92.536719),
"/US/US-WI/newburg-23089" => array("type"=>"village", "pop"=>1248, "name_ru"=>"Ньюбург", "name_en"=>"Newburg", "lang"=>"en", "lat"=>43.431440, "lon"=>-88.047539),
"/US/US-WI/niagara-23096" => array("type"=>"village", "pop"=>1735, "name_en"=>"Niagara", "lang"=>"en", "lat"=>45.780103, "lon"=>-88.002215),
"/US/US-WI/nichols-23097" => array("type"=>"village", "pop"=>301, "name_ru"=>"Николс", "name_en"=>"Nichols", "lang"=>"en", "lat"=>44.565451, "lon"=>-88.467079),
"/US/US-WI/north-bay-23098" => array("type"=>"village", "pop"=>242, "name_en"=>"North Bay", "lang"=>"en", "lat"=>42.764850, "lon"=>-87.780774),
"/US/US-WI/north-fond-du-lac-23099" => array("type"=>"village", "pop"=>5198, "name_en"=>"North Fond du Lac", "lang"=>"en", "lat"=>43.809017, "lon"=>-88.483308),
"/US/US-WI/north-freedom-23100" => array("type"=>"village", "pop"=>608, "name_en"=>"North Freedom", "lang"=>"en", "lat"=>43.458736, "lon"=>-89.863104),
"/US/US-WI/north-hudson-23101" => array("type"=>"village", "pop"=>3831, "name_en"=>"North Hudson", "lang"=>"en", "lat"=>44.995761, "lon"=>-92.759011),
"/US/US-WI/north-prairie-23102" => array("type"=>"village", "pop"=>2061, "name_en"=>"North Prairie", "lang"=>"en", "lat"=>42.937176, "lon"=>-88.395622),
"/US/US-WI/norwalk-23103" => array("type"=>"village", "pop"=>616, "name_ru"=>"Норуолк", "name_en"=>"Norwalk", "lang"=>"en", "lat"=>43.833326, "lon"=>-90.625416),
"/US/US-WI/oak-creek-23104" => array("type"=>"city", "pop"=>33482, "name_ru"=>"Оук Крик", "name_en"=>"Oak Creek", "lang"=>"en", "lat"=>42.882300, "lon"=>-87.897301),
"/US/US-WI/oakdale-23105" => array("type"=>"village", "pop"=>312, "name_ru"=>"Оукдейл", "name_en"=>"Oakdale", "lang"=>"en", "lat"=>43.962659, "lon"=>-90.378597),
"/US/US-WI/oakfield-23106" => array("type"=>"village", "pop"=>1010, "name_en"=>"Oakfield", "lang"=>"en", "lat"=>43.684578, "lon"=>-88.546870),
"/US/US-WI/oconomowoc-23107" => array("type"=>"city", "pop"=>14172, "name_en"=>"Oconomowoc", "lang"=>"en", "lat"=>43.102760, "lon"=>-88.496752),
"/US/US-WI/oconomowoc-lake-23108" => array("type"=>"village", "pop"=>630, "name_en"=>"Oconomowoc Lake", "lang"=>"en", "lat"=>43.098680, "lon"=>-88.455038),
"/US/US-WI/oconto-23109" => array("type"=>"city", "pop"=>4320, "name_en"=>"Oconto", "lang"=>"en", "lat"=>44.891306, "lon"=>-87.868238),
"/US/US-WI/oconto-falls-23110" => array("type"=>"city", "pop"=>2607, "name_en"=>"Oconto Falls", "lang"=>"en", "lat"=>44.872577, "lon"=>-88.145190),
"/US/US-WI/odanah-23111" => array("type"=>"cdp", "pop"=>190, "name_en"=>"Odanah", "lang"=>"en", "lat"=>46.604537, "lon"=>-90.683520),
"/US/US-WI/ogdensburg-23112" => array("type"=>"village", "pop"=>204, "name_en"=>"Ogdensburg", "lang"=>"en", "lat"=>44.453916, "lon"=>-89.032197),
"/US/US-WI/okauchee-lake-23113" => array("type"=>"cdp", "pop"=>3819, "name_en"=>"Okauchee Lake", "lang"=>"en", "lat"=>43.123392, "lon"=>-88.440562),
"/US/US-WI/oliver-23114" => array("type"=>"village", "pop"=>394, "name_ru"=>"Оливер", "name_en"=>"Oliver", "lang"=>"en", "lat"=>46.649900, "lon"=>-92.189949),
"/US/US-WI/omro-23115" => array("type"=>"city", "pop"=>3419, "name_en"=>"Omro", "lang"=>"en", "lat"=>44.038978, "lon"=>-88.741535),
"/US/US-WI/onalaska-23116" => array("type"=>"city", "pop"=>16778, "name_en"=>"Onalaska", "lang"=>"en", "lat"=>43.887050, "lon"=>-91.218849),
"/US/US-WI/oneida-23117" => array("type"=>"cdp", "pop"=>36031, "name_ru"=>"Онида", "name_en"=>"Oneida", "lang"=>"en", "lat"=>44.492683, "lon"=>-88.216127),
"/US/US-WI/ontario-23118" => array("type"=>"village", "pop"=>475, "name_ru"=>"Онтарио", "name_en"=>"Ontario", "lang"=>"en", "lat"=>43.722063, "lon"=>-90.592899),
"/US/US-WI/oostburg-23119" => array("type"=>"village", "pop"=>2838, "name_en"=>"Oostburg", "lang"=>"en", "lat"=>43.624260, "lon"=>-87.797579),
"/US/US-WI/oregon-23120" => array("type"=>"village", "pop"=>9369, "name_ru"=>"Орегон", "name_en"=>"Oregon", "lang"=>"en", "lat"=>42.926280, "lon"=>-89.382149),
"/US/US-WI/orfordville-23121" => array("type"=>"village", "pop"=>1386, "name_en"=>"Orfordville", "lang"=>"en", "lat"=>42.629156, "lon"=>-89.257268),
"/US/US-WI/osceola-23122" => array("type"=>"village", "pop"=>2722, "name_ru"=>"Осеола", "name_en"=>"Osceola", "lang"=>"en", "lat"=>45.319220, "lon"=>-92.695853),
"/US/US-WI/oshkosh-23123" => array("type"=>"city", "pop"=>63679, "name_en"=>"Oshkosh", "lang"=>"en", "lat"=>44.017767, "lon"=>-88.553991),
"/US/US-WI/osseo-23124" => array("type"=>"city", "pop"=>1629, "name_ru"=>"Оссео", "name_en"=>"Osseo", "lang"=>"en", "lat"=>44.578345, "lon"=>-91.218116),
"/US/US-WI/owen-23125" => array("type"=>"city", "pop"=>890, "name_en"=>"Owen", "lang"=>"en", "lat"=>44.949364, "lon"=>-90.567823),
"/US/US-WI/oxford-23126" => array("type"=>"village", "pop"=>533, "name_ru"=>"Оксфорд", "name_en"=>"Oxford", "lang"=>"en", "lat"=>43.780163, "lon"=>-89.562673),
"/US/US-WI/paddock-lake-23127" => array("type"=>"village", "pop"=>3138, "name_en"=>"Paddock Lake", "lang"=>"en", "lat"=>42.571541, "lon"=>-88.102371),
"/US/US-WI/palmyra-23128" => array("type"=>"village", "pop"=>1747, "name_ru"=>"Палмира", "name_en"=>"Palmyra", "lang"=>"en", "lat"=>42.878989, "lon"=>-88.588110),
"/US/US-WI/pardeeville-23129" => array("type"=>"village", "pop"=>2106, "name_en"=>"Pardeeville", "lang"=>"en", "lat"=>43.534070, "lon"=>-89.300324),
"/US/US-WI/park-falls-23130" => array("type"=>"city", "pop"=>2262, "name_en"=>"Park Falls", "lang"=>"en", "lat"=>45.933413, "lon"=>-90.445918),
"/US/US-WI/park-ridge-23131" => array("type"=>"village", "pop"=>462, "name_ru"=>"Парк Ридж", "name_en"=>"Park Ridge", "lang"=>"en", "lat"=>44.520028, "lon"=>-89.546221),
"/US/US-WI/patch-grove-23132" => array("type"=>"village", "pop"=>158, "name_en"=>"Patch Grove", "lang"=>"en", "lat"=>42.939341, "lon"=>-90.972654),
"/US/US-WI/pell-lake-23133" => array("type"=>"cdp", "pop"=>2018, "name_en"=>"Pell Lake", "lang"=>"en", "lat"=>42.540883, "lon"=>-88.357975),
"/US/US-WI/pepin-23134" => array("type"=>"village", "pop"=>908, "name_en"=>"Pepin", "lang"=>"en", "lat"=>44.442403, "lon"=>-92.146274),
"/US/US-WI/peshtigo-23135" => array("type"=>"city", "pop"=>3216, "name_en"=>"Peshtigo", "lang"=>"en", "lat"=>45.055494, "lon"=>-87.745838),
"/US/US-WI/pewaukee-23136" => array("type"=>"village", "pop"=>12494, "name_en"=>"Pewaukee", "lang"=>"en", "lat"=>43.085712, "lon"=>-88.251992),
"/US/US-WI/phillips-23137" => array("type"=>"city", "pop"=>1407, "name_en"=>"Phillips", "lang"=>"en", "lat"=>45.695609, "lon"=>-90.397702),
"/US/US-WI/pigeon-falls-23138" => array("type"=>"village", "pop"=>387, "name_en"=>"Pigeon Falls", "lang"=>"en", "lat"=>44.424701, "lon"=>-91.208212),
"/US/US-WI/pittsville-23139" => array("type"=>"city", "pop"=>840, "name_ru"=>"Питтсвилл", "name_en"=>"Pittsville", "lang"=>"en", "lat"=>44.440448, "lon"=>-90.129870),
"/US/US-WI/plain-23140" => array("type"=>"village", "pop"=>765, "name_en"=>"Plain", "lang"=>"en", "lat"=>43.278286, "lon"=>-90.042183),
"/US/US-WI/plainfield-23141" => array("type"=>"village", "pop"=>885, "name_ru"=>"Плейнфилд", "name_en"=>"Plainfield", "lang"=>"en", "lat"=>44.214096, "lon"=>-89.494282),
"/US/US-WI/platteville-23142" => array("type"=>"city", "pop"=>10297, "name_ru"=>"Платтевилл", "name_en"=>"Platteville", "lang"=>"en", "lat"=>42.735180, "lon"=>-90.476866),
"/US/US-WI/pleasant-prairie-23143" => array("type"=>"village", "pop"=>19847, "name_en"=>"Pleasant Prairie", "lang"=>"en", "lat"=>42.524050, "lon"=>-87.881824),
"/US/US-WI/plover-23144" => array("type"=>"village", "pop"=>11720, "name_ru"=>"Пловер", "name_en"=>"Plover", "lang"=>"en", "lat"=>44.462596, "lon"=>-89.542950),
"/US/US-WI/plum-city-23145" => array("type"=>"village", "pop"=>570, "name_en"=>"Plum City", "lang"=>"en", "lat"=>44.632972, "lon"=>-92.191789),
"/US/US-WI/plymouth-23146" => array("type"=>"city", "pop"=>8241, "name_ru"=>"Плимут", "name_en"=>"Plymouth", "lang"=>"en", "lat"=>43.747202, "lon"=>-87.971950),
"/US/US-WI/poplar-23147" => array("type"=>"village", "pop"=>620, "name_en"=>"Poplar", "lang"=>"en", "lat"=>46.581927, "lon"=>-91.787804),
"/US/US-WI/port-edwards-23149" => array("type"=>"village", "pop"=>1742, "name_en"=>"Port Edwards", "lang"=>"en", "lat"=>44.346539, "lon"=>-89.846379),
"/US/US-WI/port-washington-23150" => array("type"=>"city", "pop"=>11174, "name_en"=>"Port Washington", "lang"=>"en", "lat"=>43.393685, "lon"=>-87.879925),
"/US/US-WI/portage-23148" => array("type"=>"city", "pop"=>9919, "name_ru"=>"Портейдж", "name_en"=>"Portage", "lang"=>"en", "lat"=>43.547941, "lon"=>-89.465903),
"/US/US-WI/potosi-23151" => array("type"=>"village", "pop"=>715, "name_ru"=>"Потоси", "name_en"=>"Potosi", "lang"=>"en", "lat"=>42.689385, "lon"=>-90.711029),
"/US/US-WI/potter-23152" => array("type"=>"village", "pop"=>237, "name_en"=>"Potter", "lang"=>"en", "lat"=>44.119609, "lon"=>-88.097741),
"/US/US-WI/potter-lake-23153" => array("type"=>"cdp", "pop"=>1096, "name_en"=>"Potter Lake", "lang"=>"en", "lat"=>42.821584, "lon"=>-88.348764),
"/US/US-WI/pound-23154" => array("type"=>"village", "pop"=>331, "name_en"=>"Pound", "lang"=>"en", "lat"=>45.095879, "lon"=>-88.032846),
"/US/US-WI/powers-lake-23155" => array("type"=>"cdp", "pop"=>1044, "name_en"=>"Powers Lake", "lang"=>"en", "lat"=>42.542580, "lon"=>-88.301429),
"/US/US-WI/poynette-23156" => array("type"=>"village", "pop"=>2549, "name_en"=>"Poynette", "lang"=>"en", "lat"=>43.390836, "lon"=>-89.408232),
"/US/US-WI/prairie-du-chien-23157" => array("type"=>"city", "pop"=>5839, "name_en"=>"Prairie du Chien", "lang"=>"en", "lat"=>43.043219, "lon"=>-91.139955),
"/US/US-WI/prairie-du-sac-23158" => array("type"=>"village", "pop"=>3717, "name_en"=>"Prairie du Sac", "lang"=>"en", "lat"=>43.291068, "lon"=>-89.728504),
"/US/US-WI/prairie-farm-23159" => array("type"=>"village", "pop"=>498, "name_en"=>"Prairie Farm", "lang"=>"en", "lat"=>45.237601, "lon"=>-91.980758),
"/US/US-WI/prentice-23160" => array("type"=>"village", "pop"=>512, "name_en"=>"Prentice", "lang"=>"en", "lat"=>45.541748, "lon"=>-90.288076),
"/US/US-WI/prescott-23161" => array("type"=>"city", "pop"=>4020, "name_ru"=>"Прескотт", "name_en"=>"Prescott", "lang"=>"en", "lat"=>44.750650, "lon"=>-92.790078),
"/US/US-WI/princeton-23162" => array("type"=>"city", "pop"=>1399, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>43.852249, "lon"=>-89.129465),
"/US/US-WI/pulaski-23163" => array("type"=>"village", "pop"=>3552, "name_ru"=>"Пуласки", "name_en"=>"Pulaski", "lang"=>"en", "lat"=>44.669227, "lon"=>-88.236892),
"/US/US-WI/racine-23164" => array("type"=>"city", "pop"=>82196, "name_ru"=>"Расин", "name_en"=>"Racine", "lang"=>"en", "lat"=>42.728200, "lon"=>-87.807099),
"/US/US-WI/radisson-23165" => array("type"=>"village", "pop"=>221, "name_en"=>"Radisson", "lang"=>"en", "lat"=>45.768525, "lon"=>-91.218728),
"/US/US-WI/randolph-23166" => array("type"=>"village", "pop"=>1774, "name_ru"=>"Рэндолф", "name_en"=>"Randolph", "lang"=>"en", "lat"=>43.537839, "lon"=>-89.002187),
"/US/US-WI/random-lake-23167" => array("type"=>"village", "pop"=>1606, "name_en"=>"Random Lake", "lang"=>"en", "lat"=>43.554963, "lon"=>-87.956022),
"/US/US-WI/readstown-23168" => array("type"=>"village", "pop"=>386, "name_en"=>"Readstown", "lang"=>"en", "lat"=>43.448786, "lon"=>-90.758902),
"/US/US-WI/redgranite-23169" => array("type"=>"village", "pop"=>2060, "name_en"=>"Redgranite", "lang"=>"en", "lat"=>44.050279, "lon"=>-89.102759),
"/US/US-WI/reedsburg-23170" => array("type"=>"city", "pop"=>8666, "name_en"=>"Reedsburg", "lang"=>"en", "lat"=>43.534913, "lon"=>-89.999022),
"/US/US-WI/reedsville-23171" => array("type"=>"village", "pop"=>1133, "name_en"=>"Reedsville", "lang"=>"en", "lat"=>44.153597, "lon"=>-87.953123),
"/US/US-WI/reeseville-23172" => array("type"=>"village", "pop"=>690, "name_en"=>"Reeseville", "lang"=>"en", "lat"=>43.305904, "lon"=>-88.845104),
"/US/US-WI/reserve-23173" => array("type"=>"cdp", "pop"=>371, "name_ru"=>"Резерв", "name_en"=>"Reserve", "lang"=>"en", "lat"=>45.828338, "lon"=>-91.360116),
"/US/US-WI/rewey-23174" => array("type"=>"village", "pop"=>285, "name_en"=>"Rewey", "lang"=>"en", "lat"=>42.841896, "lon"=>-90.396337),
"/US/US-WI/rhinelander-23175" => array("type"=>"city", "pop"=>7649, "name_en"=>"Rhinelander", "lang"=>"en", "lat"=>45.639559, "lon"=>-89.411956),
"/US/US-WI/rib-lake-23176" => array("type"=>"village", "pop"=>855, "name_en"=>"Rib Lake", "lang"=>"en", "lat"=>45.318869, "lon"=>-90.203603),
"/US/US-WI/rib-mountain-23177" => array("type"=>"cdp", "pop"=>7547, "name_en"=>"Rib Mountain", "lang"=>"en", "lat"=>44.912879, "lon"=>-89.675363),
"/US/US-WI/rice-lake-23178" => array("type"=>"city", "pop"=>8257, "name_en"=>"Rice Lake", "lang"=>"en", "lat"=>45.496924, "lon"=>-91.739189),
"/US/US-WI/richland-center-23179" => array("type"=>"city", "pop"=>5050, "name_en"=>"Richland Center", "lang"=>"en", "lat"=>43.337819, "lon"=>-90.383226),
"/US/US-WI/ridgeland-23180" => array("type"=>"village", "pop"=>258, "name_ru"=>"Риджленд", "name_en"=>"Ridgeland", "lang"=>"en", "lat"=>45.202217, "lon"=>-91.897373),
"/US/US-WI/ridgeway-23181" => array("type"=>"village", "pop"=>642, "name_ru"=>"Риджуэй", "name_en"=>"Ridgeway", "lang"=>"en", "lat"=>42.999153, "lon"=>-89.992203),
"/US/US-WI/rio-23182" => array("type"=>"village", "pop"=>989, "name_ru"=>"Рио", "name_en"=>"Rio", "lang"=>"en", "lat"=>43.450274, "lon"=>-89.237652),
"/US/US-WI/ripon-23183" => array("type"=>"city", "pop"=>7396, "name_ru"=>"Рипон", "name_en"=>"Ripon", "lang"=>"en", "lat"=>43.845821, "lon"=>-88.841576),
"/US/US-WI/river-falls-23184" => array("type"=>"city", "pop"=>14352, "name_ru"=>"Ривер Фолс", "name_en"=>"River Falls", "lang"=>"en", "lat"=>44.855741, "lon"=>-92.621609),
"/US/US-WI/river-hills-23185" => array("type"=>"village", "pop"=>1657, "name_en"=>"River Hills", "lang"=>"en", "lat"=>43.173600, "lon"=>-87.935061),
"/US/US-WI/roberts-23186" => array("type"=>"village", "pop"=>1651, "name_ru"=>"Робертс", "name_en"=>"Roberts", "lang"=>"en", "lat"=>44.985008, "lon"=>-92.552030),
"/US/US-WI/rochester-23187" => array("type"=>"village", "pop"=>1149, "name_ru"=>"Рочестер", "name_en"=>"Rochester", "lang"=>"en", "lat"=>42.740569, "lon"=>-88.224019),
"/US/US-WI/rock-springs-23190" => array("type"=>"village", "pop"=>400, "name_en"=>"Rock Springs", "lang"=>"en", "lat"=>43.479879, "lon"=>-89.917964),
"/US/US-WI/rockdale-23188" => array("type"=>"village", "pop"=>207, "name_ru"=>"Рокдейл", "name_en"=>"Rockdale", "lang"=>"en", "lat"=>42.972464, "lon"=>-89.032627),
"/US/US-WI/rockland-23189" => array("type"=>"village", "pop"=>647, "name_ru"=>"Рокленд", "name_en"=>"Rockland", "lang"=>"en", "lat"=>43.906149, "lon"=>-90.918236),
"/US/US-WI/rosendale-23191" => array("type"=>"village", "pop"=>1024, "name_ru"=>"Розендейл", "name_en"=>"Rosendale", "lang"=>"en", "lat"=>43.808276, "lon"=>-88.675794),
"/US/US-WI/rosholt-23192" => array("type"=>"village", "pop"=>473, "name_en"=>"Rosholt", "lang"=>"en", "lat"=>44.630157, "lon"=>-89.304939),
"/US/US-WI/rothschild-23193" => array("type"=>"village", "pop"=>5042, "name_en"=>"Rothschild", "lang"=>"en", "lat"=>44.886850, "lon"=>-89.620924),
"/US/US-WI/rudolph-23194" => array("type"=>"village", "pop"=>412, "name_en"=>"Rudolph", "lang"=>"en", "lat"=>44.497238, "lon"=>-89.801773),
"/US/US-WI/sauk-city-23199" => array("type"=>"village", "pop"=>2945, "name_en"=>"Sauk City", "lang"=>"en", "lat"=>43.272192, "lon"=>-89.728513),
"/US/US-WI/saukville-23200" => array("type"=>"village", "pop"=>4283, "name_en"=>"Saukville", "lang"=>"en", "lat"=>43.383808, "lon"=>-87.941427),
"/US/US-WI/scandinavia-23201" => array("type"=>"village", "pop"=>335, "name_en"=>"Scandinavia", "lang"=>"en", "lat"=>44.460758, "lon"=>-89.146541),
"/US/US-WI/schofield-23202" => array("type"=>"city", "pop"=>2114, "name_en"=>"Schofield", "lang"=>"en", "lat"=>44.915233, "lon"=>-89.611883),
"/US/US-WI/seymour-23203" => array("type"=>"cdp", "pop"=>3474, "name_ru"=>"Сеймур", "name_en"=>"Seymour", "lang"=>"en", "lat"=>44.826567, "lon"=>-91.431095),
"/US/US-WI/seymour-23204" => array("type"=>"city", "pop"=>2782, "name_ru"=>"Сеймур", "name_en"=>"Seymour", "lang"=>"en", "lat"=>44.514479, "lon"=>-88.327370),
"/US/US-WI/sharon-23205" => array("type"=>"village", "pop"=>1553, "name_ru"=>"Шэрон", "name_en"=>"Sharon", "lang"=>"en", "lat"=>42.501564, "lon"=>-88.730118),
"/US/US-WI/shawano-23206" => array("type"=>"city", "pop"=>8666, "name_en"=>"Shawano", "lang"=>"en", "lat"=>44.776548, "lon"=>-88.588685),
"/US/US-WI/sheboygan-23207" => array("type"=>"city", "pop"=>47895, "name_en"=>"Sheboygan", "lang"=>"en", "lat"=>43.745700, "lon"=>-87.730071),
"/US/US-WI/sheboygan-falls-23208" => array("type"=>"city", "pop"=>7895, "name_en"=>"Sheboygan Falls", "lang"=>"en", "lat"=>43.729750, "lon"=>-87.822624),
"/US/US-WI/sheldon-23209" => array("type"=>"village", "pop"=>218, "name_ru"=>"Шелдон", "name_en"=>"Sheldon", "lang"=>"en", "lat"=>45.311080, "lon"=>-90.957374),
"/US/US-WI/shell-lake-23210" => array("type"=>"city", "pop"=>1393, "name_en"=>"Shell Lake", "lang"=>"en", "lat"=>45.738590, "lon"=>-91.899086),
"/US/US-WI/sherwood-23211" => array("type"=>"village", "pop"=>2388, "name_ru"=>"Шервуд", "name_en"=>"Sherwood", "lang"=>"en", "lat"=>44.173550, "lon"=>-88.275151),
"/US/US-WI/shiocton-23212" => array("type"=>"village", "pop"=>917, "name_en"=>"Shiocton", "lang"=>"en", "lat"=>44.445232, "lon"=>-88.576502),
"/US/US-WI/shorewood-23213" => array("type"=>"village", "pop"=>13226, "name_ru"=>"Шорвуд", "name_en"=>"Shorewood", "lang"=>"en", "lat"=>43.091250, "lon"=>-87.886249),
"/US/US-WI/shorewood-hills-23214" => array("type"=>"village", "pop"=>1681, "name_en"=>"Shorewood Hills", "lang"=>"en", "lat"=>43.077733, "lon"=>-89.446482),
"/US/US-WI/shullsburg-23215" => array("type"=>"city", "pop"=>1136, "name_en"=>"Shullsburg", "lang"=>"en", "lat"=>42.573082, "lon"=>-90.231521),
"/US/US-WI/silver-lake-23216" => array("type"=>"village", "pop"=>2522, "name_ru"=>"Сильвер Лэйк", "name_en"=>"Silver Lake", "lang"=>"en", "lat"=>42.551700, "lon"=>-88.167611),
"/US/US-WI/siren-23217" => array("type"=>"village", "pop"=>999, "name_en"=>"Siren", "lang"=>"en", "lat"=>45.783685, "lon"=>-92.379966),
"/US/US-WI/sister-bay-23218" => array("type"=>"village", "pop"=>867, "name_en"=>"Sister Bay", "lang"=>"en", "lat"=>45.186888, "lon"=>-87.127451),
"/US/US-WI/slinger-23219" => array("type"=>"village", "pop"=>4657, "name_en"=>"Slinger", "lang"=>"en", "lat"=>43.332064, "lon"=>-88.285675),
"/US/US-WI/soldiers-grove-23220" => array("type"=>"village", "pop"=>611, "name_en"=>"Soldiers Grove", "lang"=>"en", "lat"=>43.392433, "lon"=>-90.773557),
"/US/US-WI/solon-springs-23221" => array("type"=>"village", "pop"=>562, "name_en"=>"Solon Springs", "lang"=>"en", "lat"=>46.349447, "lon"=>-91.819997),
"/US/US-WI/somerset-23222" => array("type"=>"village", "pop"=>2681, "name_ru"=>"Сомерсет", "name_en"=>"Somerset", "lang"=>"en", "lat"=>45.127253, "lon"=>-92.676047),
"/US/US-WI/south-milwaukee-23223" => array("type"=>"city", "pop"=>21129, "name_en"=>"South Milwaukee", "lang"=>"en", "lat"=>42.912900, "lon"=>-87.862160),
"/US/US-WI/south-wayne-23224" => array("type"=>"village", "pop"=>449, "name_en"=>"South Wayne", "lang"=>"en", "lat"=>42.566653, "lon"=>-89.880439),
"/US/US-WI/sparta-23225" => array("type"=>"city", "pop"=>8823, "name_ru"=>"Спарта", "name_en"=>"Sparta", "lang"=>"en", "lat"=>43.942756, "lon"=>-90.809753),
"/US/US-WI/spencer-23226" => array("type"=>"village", "pop"=>1805, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>44.755394, "lon"=>-90.298213),
"/US/US-WI/spooner-23227" => array("type"=>"city", "pop"=>2655, "name_en"=>"Spooner", "lang"=>"en", "lat"=>45.824533, "lon"=>-91.893502),
"/US/US-WI/spring-green-23228" => array("type"=>"village", "pop"=>1497, "name_en"=>"Spring Green", "lang"=>"en", "lat"=>43.175635, "lon"=>-90.067143),
"/US/US-WI/spring-valley-23229" => array("type"=>"village", "pop"=>1310, "name_ru"=>"Спринг Валли", "name_en"=>"Spring Valley", "lang"=>"en", "lat"=>44.849970, "lon"=>-92.239847),
"/US/US-WI/st.-cloud-23195" => array("type"=>"village", "pop"=>505, "name_ru"=>"Сент Клауд", "name_en"=>"St. Cloud", "lang"=>"en", "lat"=>43.822611, "lon"=>-88.167303),
"/US/US-WI/st.-croix-falls-23196" => array("type"=>"city", "pop"=>2152, "name_en"=>"St. Croix Falls", "lang"=>"en", "lat"=>45.409413, "lon"=>-92.633817),
"/US/US-WI/st.-francis-23197" => array("type"=>"city", "pop"=>9733, "name_ru"=>"Сент-Фрэнсис", "name_en"=>"St. Francis", "lang"=>"en", "lat"=>42.970700, "lon"=>-87.871729),
"/US/US-WI/st.-nazianz-23198" => array("type"=>"village", "pop"=>793, "name_en"=>"St. Nazianz", "lang"=>"en", "lat"=>44.007074, "lon"=>-87.924176),
"/US/US-WI/stanley-23230" => array("type"=>"city", "pop"=>3319, "name_ru"=>"Стэнли", "name_en"=>"Stanley", "lang"=>"en", "lat"=>44.962548, "lon"=>-90.942761),
"/US/US-WI/star-prairie-23231" => array("type"=>"village", "pop"=>652, "name_en"=>"Star Prairie", "lang"=>"en", "lat"=>45.198310, "lon"=>-92.531789),
"/US/US-WI/stetsonville-23232" => array("type"=>"village", "pop"=>525, "name_en"=>"Stetsonville", "lang"=>"en", "lat"=>45.076445, "lon"=>-90.312515),
"/US/US-WI/steuben-23233" => array("type"=>"village", "pop"=>153, "name_en"=>"Steuben", "lang"=>"en", "lat"=>43.181236, "lon"=>-90.857343),
"/US/US-WI/stevens-point-23234" => array("type"=>"city", "pop"=>25327, "name_en"=>"Stevens Point", "lang"=>"en", "lat"=>44.527685, "lon"=>-89.560507),
"/US/US-WI/stockbridge-23235" => array("type"=>"village", "pop"=>670, "name_ru"=>"Стокбридж", "name_en"=>"Stockbridge", "lang"=>"en", "lat"=>44.074038, "lon"=>-88.315999),
"/US/US-WI/stockholm-23236" => array("type"=>"village", "pop"=>95, "name_en"=>"Stockholm", "lang"=>"en", "lat"=>44.483592, "lon"=>-92.260278),
"/US/US-WI/stoddard-23237" => array("type"=>"village", "pop"=>832, "name_en"=>"Stoddard", "lang"=>"en", "lat"=>43.661184, "lon"=>-91.218444),
"/US/US-WI/stoughton-23238" => array("type"=>"city", "pop"=>12966, "name_en"=>"Stoughton", "lang"=>"en", "lat"=>42.921056, "lon"=>-89.222764),
"/US/US-WI/stratford-23239" => array("type"=>"village", "pop"=>1515, "name_ru"=>"Стрэтфорд", "name_en"=>"Stratford", "lang"=>"en", "lat"=>44.801137, "lon"=>-90.072101),
"/US/US-WI/strum-23240" => array("type"=>"village", "pop"=>1024, "name_en"=>"Strum", "lang"=>"en", "lat"=>44.552899, "lon"=>-91.386181),
"/US/US-WI/sturgeon-bay-23241" => array("type"=>"city", "pop"=>8792, "name_en"=>"Sturgeon Bay", "lang"=>"en", "lat"=>44.821635, "lon"=>-87.368287),
"/US/US-WI/sturtevant-23242" => array("type"=>"village", "pop"=>6913, "name_en"=>"Sturtevant", "lang"=>"en", "lat"=>42.700300, "lon"=>-87.899124),
"/US/US-WI/sullivan-23243" => array("type"=>"village", "pop"=>657, "name_ru"=>"Салливан", "name_en"=>"Sullivan", "lang"=>"en", "lat"=>43.012549, "lon"=>-88.589536),
"/US/US-WI/sun-prairie-23244" => array("type"=>"city", "pop"=>28322, "name_en"=>"Sun Prairie", "lang"=>"en", "lat"=>43.183680, "lon"=>-89.232338),
"/US/US-WI/superior-23245" => array("type"=>"city", "pop"=>26223, "name_ru"=>"Супериор", "name_en"=>"Superior", "lang"=>"en", "lat"=>46.696870, "lon"=>-92.059603),
"/US/US-WI/superior-23246" => array("type"=>"village", "pop"=>481, "name_ru"=>"Супериор", "name_en"=>"Superior", "lang"=>"en", "lat"=>46.653660, "lon"=>-92.105499),
"/US/US-WI/suring-23247" => array("type"=>"village", "pop"=>526, "name_en"=>"Suring", "lang"=>"en", "lat"=>45.000419, "lon"=>-88.370289),
"/US/US-WI/sussex-23248" => array("type"=>"village", "pop"=>10048, "name_en"=>"Sussex", "lang"=>"en", "lat"=>43.135949, "lon"=>-88.217945),
"/US/US-WI/tainter-lake-23249" => array("type"=>"cdp", "pop"=>1716, "name_en"=>"Tainter Lake", "lang"=>"en", "lat"=>44.989290, "lon"=>-91.847602),
"/US/US-WI/taylor-23250" => array("type"=>"village", "pop"=>486, "name_ru"=>"Тейлор", "name_en"=>"Taylor", "lang"=>"en", "lat"=>44.322059, "lon"=>-91.121228),
"/US/US-WI/tennyson-23251" => array("type"=>"village", "pop"=>351, "name_ru"=>"Теннисон", "name_en"=>"Tennyson", "lang"=>"en", "lat"=>42.689652, "lon"=>-90.686628),
"/US/US-WI/theresa-23252" => array("type"=>"village", "pop"=>1249, "name_en"=>"Theresa", "lang"=>"en", "lat"=>43.517824, "lon"=>-88.452732),
"/US/US-WI/thiensville-23253" => array("type"=>"village", "pop"=>3036, "name_en"=>"Thiensville", "lang"=>"en", "lat"=>43.236800, "lon"=>-87.979863),
"/US/US-WI/thorp-23254" => array("type"=>"city", "pop"=>1509, "name_en"=>"Thorp", "lang"=>"en", "lat"=>44.959825, "lon"=>-90.800925),
"/US/US-WI/tigerton-23255" => array("type"=>"village", "pop"=>720, "name_en"=>"Tigerton", "lang"=>"en", "lat"=>44.740496, "lon"=>-89.060904),
"/US/US-WI/tomah-23256" => array("type"=>"city", "pop"=>8664, "name_en"=>"Tomah", "lang"=>"en", "lat"=>43.989609, "lon"=>-90.503327),
"/US/US-WI/tomahawk-23257" => array("type"=>"city", "pop"=>3695, "name_en"=>"Tomahawk", "lang"=>"en", "lat"=>45.473409, "lon"=>-89.724151),
"/US/US-WI/tony-23258" => array("type"=>"village", "pop"=>89, "name_en"=>"Tony", "lang"=>"en", "lat"=>45.481125, "lon"=>-90.995816),
"/US/US-WI/trempealeau-23259" => array("type"=>"village", "pop"=>1518, "name_en"=>"Trempealeau", "lang"=>"en", "lat"=>44.007025, "lon"=>-91.435957),
"/US/US-WI/turtle-lake-23260" => array("type"=>"village", "pop"=>1029, "name_en"=>"Turtle Lake", "lang"=>"en", "lat"=>45.392384, "lon"=>-92.141289),
"/US/US-WI/twin-lakes-23261" => array("type"=>"village", "pop"=>5752, "name_ru"=>"Твин Лейкс", "name_en"=>"Twin Lakes", "lang"=>"en", "lat"=>42.518829, "lon"=>-88.258815),
"/US/US-WI/two-rivers-23262" => array("type"=>"city", "pop"=>11807, "name_ru"=>"Ту Риверс", "name_en"=>"Two Rivers", "lang"=>"en", "lat"=>44.156283, "lon"=>-87.581188),
"/US/US-WI/union-center-23263" => array("type"=>"village", "pop"=>217, "name_en"=>"Union Center", "lang"=>"en", "lat"=>43.684069, "lon"=>-90.263171),
"/US/US-WI/union-grove-23264" => array("type"=>"village", "pop"=>4662, "name_ru"=>"Юнион Гров", "name_en"=>"Union Grove", "lang"=>"en", "lat"=>42.684192, "lon"=>-88.050991),
"/US/US-WI/unity-23265" => array("type"=>"village", "pop"=>341, "name_en"=>"Unity", "lang"=>"en", "lat"=>44.850568, "lon"=>-90.313887),
"/US/US-WI/valders-23266" => array("type"=>"village", "pop"=>971, "name_en"=>"Valders", "lang"=>"en", "lat"=>44.066036, "lon"=>-87.885619),
"/US/US-WI/verona-23267" => array("type"=>"city", "pop"=>11486, "name_ru"=>"Верона", "name_en"=>"Verona", "lang"=>"en", "lat"=>42.988848, "lon"=>-89.535427),
"/US/US-WI/vesper-23268" => array("type"=>"village", "pop"=>511, "name_en"=>"Vesper", "lang"=>"en", "lat"=>44.481056, "lon"=>-89.967295),
"/US/US-WI/viola-23269" => array("type"=>"village", "pop"=>626, "name_ru"=>"Виола", "name_en"=>"Viola", "lang"=>"en", "lat"=>43.507565, "lon"=>-90.673106),
"/US/US-WI/viroqua-23270" => array("type"=>"city", "pop"=>4435, "name_en"=>"Viroqua", "lang"=>"en", "lat"=>43.558310, "lon"=>-90.886328),
"/US/US-WI/waldo-23271" => array("type"=>"village", "pop"=>466, "name_ru"=>"Уолдо", "name_en"=>"Waldo", "lang"=>"en", "lat"=>43.675590, "lon"=>-87.946544),
"/US/US-WI/wales-23272" => array("type"=>"village", "pop"=>2692, "name_ru"=>"Уэльс", "name_en"=>"Wales", "lang"=>"en", "lat"=>43.004896, "lon"=>-88.377743),
"/US/US-WI/walworth-23273" => array("type"=>"village", "pop"=>2693, "name_en"=>"Walworth", "lang"=>"en", "lat"=>42.530515, "lon"=>-88.597423),
"/US/US-WI/warrens-23274" => array("type"=>"village", "pop"=>347, "name_en"=>"Warrens", "lang"=>"en", "lat"=>44.130385, "lon"=>-90.500567),
"/US/US-WI/washburn-23275" => array("type"=>"city", "pop"=>2167, "name_ru"=>"Уошберн", "name_en"=>"Washburn", "lang"=>"en", "lat"=>46.673877, "lon"=>-90.893344),
"/US/US-WI/waterford-23276" => array("type"=>"village", "pop"=>4973, "name_ru"=>"Уотерфорд", "name_en"=>"Waterford", "lang"=>"en", "lat"=>42.762840, "lon"=>-88.215540),
"/US/US-WI/waterford-north-23277" => array("type"=>"cdp", "pop"=>1604, "name_en"=>"Waterford North", "lang"=>"en", "lat"=>42.793213, "lon"=>-88.222041),
"/US/US-WI/waterloo-23278" => array("type"=>"city", "pop"=>3272, "name_ru"=>"Ватрелоо", "name_en"=>"Waterloo", "lang"=>"en", "lat"=>43.183034, "lon"=>-88.989856),
"/US/US-WI/watertown-23279" => array("type"=>"city", "pop"=>22930, "name_ru"=>"Уотертаун", "name_en"=>"Watertown", "lang"=>"en", "lat"=>43.191625, "lon"=>-88.724858),
"/US/US-WI/waukesha-23280" => array("type"=>"city", "pop"=>68008, "name_en"=>"Waukesha", "lang"=>"en", "lat"=>43.012467, "lon"=>-88.238208),
"/US/US-WI/waunakee-23281" => array("type"=>"village", "pop"=>11290, "name_en"=>"Waunakee", "lang"=>"en", "lat"=>43.186616, "lon"=>-89.452039),
"/US/US-WI/waupaca-23282" => array("type"=>"city", "pop"=>5826, "name_en"=>"Waupaca", "lang"=>"en", "lat"=>44.354647, "lon"=>-89.075461),
"/US/US-WI/waupun-23283" => array("type"=>"city", "pop"=>10469, "name_en"=>"Waupun", "lang"=>"en", "lat"=>43.631589, "lon"=>-88.735728),
"/US/US-WI/wausau-23284" => array("type"=>"city", "pop"=>37576, "name_ru"=>"Ваусау", "name_en"=>"Wausau", "lang"=>"en", "lat"=>44.961700, "lon"=>-89.644504),
"/US/US-WI/wausaukee-23285" => array("type"=>"village", "pop"=>530, "name_en"=>"Wausaukee", "lang"=>"en", "lat"=>45.378552, "lon"=>-87.954978),
"/US/US-WI/wautoma-23286" => array("type"=>"city", "pop"=>2111, "name_en"=>"Wautoma", "lang"=>"en", "lat"=>44.069728, "lon"=>-89.292715),
"/US/US-WI/wauwatosa-23287" => array("type"=>"city", "pop"=>45004, "name_en"=>"Wauwatosa", "lang"=>"en", "lat"=>43.062100, "lon"=>-88.033597),
"/US/US-WI/wauzeka-23288" => array("type"=>"village", "pop"=>778, "name_en"=>"Wauzeka", "lang"=>"en", "lat"=>43.084371, "lon"=>-90.895462),
"/US/US-WI/webster-23289" => array("type"=>"village", "pop"=>656, "name_ru"=>"Вебстер", "name_en"=>"Webster", "lang"=>"en", "lat"=>45.878571, "lon"=>-92.361985),
"/US/US-WI/west-allis-23290" => array("type"=>"city", "pop"=>59416, "name_en"=>"West Allis", "lang"=>"en", "lat"=>43.007250, "lon"=>-88.030023),
"/US/US-WI/west-baraboo-23291" => array("type"=>"village", "pop"=>1539, "name_en"=>"West Baraboo", "lang"=>"en", "lat"=>43.475933, "lon"=>-89.767934),
"/US/US-WI/west-bend-23292" => array("type"=>"city", "pop"=>29894, "name_ru"=>"Вэст Бенд", "name_en"=>"West Bend", "lang"=>"en", "lat"=>43.422784, "lon"=>-88.184227),
"/US/US-WI/west-milwaukee-23295" => array("type"=>"village", "pop"=>4025, "name_en"=>"West Milwaukee", "lang"=>"en", "lat"=>43.012850, "lon"=>-87.970056),
"/US/US-WI/west-salem-23297" => array("type"=>"village", "pop"=>4823, "name_ru"=>"Уэст Салем", "name_en"=>"West Salem", "lang"=>"en", "lat"=>43.898592, "lon"=>-91.081669),
"/US/US-WI/westby-23293" => array("type"=>"city", "pop"=>2173, "name_en"=>"Westby", "lang"=>"en", "lat"=>43.653010, "lon"=>-90.857120),
"/US/US-WI/westfield-23294" => array("type"=>"village", "pop"=>1178, "name_ru"=>"Уэстфилд", "name_en"=>"Westfield", "lang"=>"en", "lat"=>43.884610, "lon"=>-89.494072),
"/US/US-WI/weston-23296" => array("type"=>"cdp", "pop"=>13447, "name_ru"=>"Вэстон", "name_en"=>"Weston", "lang"=>"en", "lat"=>44.899150, "lon"=>-89.569473),
"/US/US-WI/weyauwega-23298" => array("type"=>"city", "pop"=>1720, "name_en"=>"Weyauwega", "lang"=>"en", "lat"=>44.322656, "lon"=>-88.928670),
"/US/US-WI/weyerhaeuser-23299" => array("type"=>"village", "pop"=>309, "name_en"=>"Weyerhaeuser", "lang"=>"en", "lat"=>45.423799, "lon"=>-91.414398),
"/US/US-WI/wheeler-23300" => array("type"=>"village", "pop"=>309, "name_ru"=>"Уилер", "name_en"=>"Wheeler", "lang"=>"en", "lat"=>45.043673, "lon"=>-91.905463),
"/US/US-WI/white-lake-23303" => array("type"=>"village", "pop"=>323, "name_en"=>"White Lake", "lang"=>"en", "lat"=>45.163424, "lon"=>-88.766576),
"/US/US-WI/whitefish-bay-23301" => array("type"=>"village", "pop"=>13584, "name_en"=>"Whitefish Bay", "lang"=>"en", "lat"=>43.112567, "lon"=>-87.900916),
"/US/US-WI/whitehall-23302" => array("type"=>"city", "pop"=>1600, "name_ru"=>"Уайтхолл", "name_en"=>"Whitehall", "lang"=>"en", "lat"=>44.365814, "lon"=>-91.313694),
"/US/US-WI/whitelaw-23304" => array("type"=>"village", "pop"=>721, "name_en"=>"Whitelaw", "lang"=>"en", "lat"=>44.145117, "lon"=>-87.827554),
"/US/US-WI/whitewater-23305" => array("type"=>"city", "pop"=>14291, "name_ru"=>"Уайтуотер", "name_en"=>"Whitewater", "lang"=>"en", "lat"=>42.834941, "lon"=>-88.736014),
"/US/US-WI/whiting-23306" => array("type"=>"village", "pop"=>1634, "name_ru"=>"Уайтинг", "name_en"=>"Whiting", "lang"=>"en", "lat"=>44.489023, "lon"=>-89.561908),
"/US/US-WI/wild-rose-23307" => array("type"=>"village", "pop"=>742, "name_en"=>"Wild Rose", "lang"=>"en", "lat"=>44.176288, "lon"=>-89.248045),
"/US/US-WI/williams-bay-23308" => array("type"=>"village", "pop"=>2703, "name_en"=>"Williams Bay", "lang"=>"en", "lat"=>42.576410, "lon"=>-88.542789),
"/US/US-WI/wilson-23309" => array("type"=>"village", "pop"=>204, "name_ru"=>"Уилсон", "name_en"=>"Wilson", "lang"=>"en", "lat"=>44.955631, "lon"=>-92.172365),
"/US/US-WI/wilton-23310" => array("type"=>"village", "pop"=>544, "name_ru"=>"Уилтон", "name_en"=>"Wilton", "lang"=>"en", "lat"=>43.814523, "lon"=>-90.527397),
"/US/US-WI/wind-lake-23311" => array("type"=>"cdp", "pop"=>3748, "name_en"=>"Wind Lake", "lang"=>"en", "lat"=>42.822428, "lon"=>-88.156804),
"/US/US-WI/wind-point-23312" => array("type"=>"village", "pop"=>1792, "name_en"=>"Wind Point", "lang"=>"en", "lat"=>42.781664, "lon"=>-87.772929),
"/US/US-WI/windsor-23313" => array("type"=>"cdp", "pop"=>5586, "name_ru"=>"Виндзор", "name_en"=>"Windsor", "lang"=>"en", "lat"=>43.216736, "lon"=>-89.342119),
"/US/US-WI/winneconne-23314" => array("type"=>"village", "pop"=>2462, "name_en"=>"Winneconne", "lang"=>"en", "lat"=>44.112030, "lon"=>-88.711882),
"/US/US-WI/winter-23315" => array("type"=>"village", "pop"=>339, "name_en"=>"Winter", "lang"=>"en", "lat"=>45.821053, "lon"=>-91.012013),
"/US/US-WI/wisconsin-dells-23316" => array("type"=>"city", "pop"=>2476, "name_en"=>"Wisconsin Dells", "lang"=>"en", "lat"=>43.626393, "lon"=>-89.777922),
"/US/US-WI/wisconsin-rapids-23317" => array("type"=>"city", "pop"=>17164, "name_en"=>"Wisconsin Rapids", "lang"=>"en", "lat"=>44.394494, "lon"=>-89.832872),
"/US/US-WI/withee-23318" => array("type"=>"village", "pop"=>486, "name_en"=>"Withee", "lang"=>"en", "lat"=>44.952780, "lon"=>-90.597010),
"/US/US-WI/wittenberg-23319" => array("type"=>"village", "pop"=>1061, "name_en"=>"Wittenberg", "lang"=>"en", "lat"=>44.826576, "lon"=>-89.170729),
"/US/US-WI/wonewoc-23320" => array("type"=>"village", "pop"=>769, "name_en"=>"Wonewoc", "lang"=>"en", "lat"=>43.653066, "lon"=>-90.223959),
"/US/US-WI/woodman-23321" => array("type"=>"village", "pop"=>92, "name_en"=>"Woodman", "lang"=>"en", "lat"=>43.090822, "lon"=>-90.797919),
"/US/US-WI/woodville-23322" => array("type"=>"village", "pop"=>1349, "name_ru"=>"Вудвилл", "name_en"=>"Woodville", "lang"=>"en", "lat"=>44.952513, "lon"=>-92.286355),
"/US/US-WI/wrightstown-23323" => array("type"=>"village", "pop"=>2395, "name_en"=>"Wrightstown", "lang"=>"en", "lat"=>44.325000, "lon"=>-88.174359),
"/US/US-WI/wyeville-23324" => array("type"=>"village", "pop"=>140, "name_en"=>"Wyeville", "lang"=>"en", "lat"=>44.027693, "lon"=>-90.385801),
"/US/US-WI/wyocena-23325" => array("type"=>"village", "pop"=>760, "name_en"=>"Wyocena", "lang"=>"en", "lat"=>43.493132, "lon"=>-89.310048),
"/US/US-WI/yuba-23326" => array("type"=>"village", "pop"=>84, "name_en"=>"Yuba", "lang"=>"en", "lat"=>43.536252, "lon"=>-90.427939),
"/US/US-WV" => array("type"=>"state", "pop"=>1814468, "name_ru"=>"Западная Вирджиния", "name_en"=>"West Virginia", "lang"=>"en"),
"/US/US-WV/addison-(webster-springs)-22432" => array("type"=>"town", "pop"=>747, "name_en"=>"Addison (Webster Springs)", "lang"=>"en", "lat"=>38.478385, "lon"=>-80.410358),
"/US/US-WV/albright-22433" => array("type"=>"town", "pop"=>251, "name_en"=>"Albright", "lang"=>"en", "lat"=>39.493415, "lon"=>-79.638511),
"/US/US-WV/alderson-22434" => array("type"=>"town", "pop"=>1070, "name_en"=>"Alderson", "lang"=>"en", "lat"=>37.726795, "lon"=>-80.644194),
"/US/US-WV/alum-creek-22435" => array("type"=>"cdp", "pop"=>1602, "name_en"=>"Alum Creek", "lang"=>"en", "lat"=>38.281925, "lon"=>-81.841734),
"/US/US-WV/amherstdale-robinette-22436" => array("type"=>"cdp", "pop"=>2435, "name_en"=>"Amherstdale-Robinette", "lang"=>"en", "lat"=>37.779776, "lon"=>-81.773896),
"/US/US-WV/anawalt-22437" => array("type"=>"town", "pop"=>231, "name_en"=>"Anawalt", "lang"=>"en", "lat"=>37.333475, "lon"=>-81.441509),
"/US/US-WV/anmoore-22438" => array("type"=>"town", "pop"=>767, "name_en"=>"Anmoore", "lang"=>"en", "lat"=>39.258872, "lon"=>-80.291196),
"/US/US-WV/ansted-22439" => array("type"=>"town", "pop"=>1577, "name_en"=>"Ansted", "lang"=>"en", "lat"=>38.135921, "lon"=>-81.102126),
"/US/US-WV/athens-22440" => array("type"=>"town", "pop"=>1182, "name_ru"=>"Афины", "name_en"=>"Athens", "lang"=>"en", "lat"=>37.422810, "lon"=>-81.016063),
"/US/US-WV/auburn-22441" => array("type"=>"town", "pop"=>102, "name_ru"=>"Обурн", "name_en"=>"Auburn", "lang"=>"en", "lat"=>39.095344, "lon"=>-80.856711),
"/US/US-WV/bancroft-22442" => array("type"=>"town", "pop"=>376, "name_ru"=>"Бэнкрофт", "name_en"=>"Bancroft", "lang"=>"en", "lat"=>38.510797, "lon"=>-81.841843),
"/US/US-WV/barboursville-22443" => array("type"=>"village", "pop"=>3410, "name_en"=>"Barboursville", "lang"=>"en", "lat"=>38.404800, "lon"=>-82.294464),
"/US/US-WV/barrackville-22444" => array("type"=>"town", "pop"=>1289, "name_en"=>"Barrackville", "lang"=>"en", "lat"=>39.501525, "lon"=>-80.170225),
"/US/US-WV/bath-(berkeley-springs)-22445" => array("type"=>"town", "pop"=>682, "name_en"=>"Bath (Berkeley Springs)", "lang"=>"en", "lat"=>39.625469, "lon"=>-78.227133),
"/US/US-WV/bayard-22446" => array("type"=>"town", "pop"=>288, "name_ru"=>"Байяр", "name_en"=>"Bayard", "lang"=>"en", "lat"=>39.271795, "lon"=>-79.366903),
"/US/US-WV/beaver-22447" => array("type"=>"cdp", "pop"=>1244, "name_ru"=>"Бивер", "name_en"=>"Beaver", "lang"=>"en", "lat"=>37.740503, "lon"=>-81.150905),
"/US/US-WV/beckley-22448" => array("type"=>"city", "pop"=>16832, "name_en"=>"Beckley", "lang"=>"en", "lat"=>37.787480, "lon"=>-81.185849),
"/US/US-WV/beech-bottom-22449" => array("type"=>"village", "pop"=>555, "name_en"=>"Beech Bottom", "lang"=>"en", "lat"=>40.226296, "lon"=>-80.652045),
"/US/US-WV/belington-22450" => array("type"=>"town", "pop"=>1797, "name_en"=>"Belington", "lang"=>"en", "lat"=>39.022321, "lon"=>-79.938721),
"/US/US-WV/belle-22451" => array("type"=>"town", "pop"=>1158, "name_ru"=>"Бель", "name_en"=>"Belle", "lang"=>"en", "lat"=>38.232504, "lon"=>-81.539857),
"/US/US-WV/belmont-22452" => array("type"=>"city", "pop"=>975, "name_ru"=>"Белмонт", "name_en"=>"Belmont", "lang"=>"en", "lat"=>39.377413, "lon"=>-81.262256),
"/US/US-WV/benwood-22453" => array("type"=>"city", "pop"=>1431, "name_en"=>"Benwood", "lang"=>"en", "lat"=>40.009960, "lon"=>-80.733872),
"/US/US-WV/bethany-22454" => array("type"=>"town", "pop"=>970, "name_ru"=>"Бетани", "name_en"=>"Bethany", "lang"=>"en", "lat"=>40.205464, "lon"=>-80.563642),
"/US/US-WV/bethlehem-22455" => array("type"=>"village", "pop"=>2487, "name_ru"=>"Вифлеем", "name_en"=>"Bethlehem", "lang"=>"en", "lat"=>40.044712, "lon"=>-80.691976),
"/US/US-WV/beverly-22456" => array("type"=>"town", "pop"=>643, "name_ru"=>"Беверли", "name_en"=>"Beverly", "lang"=>"en", "lat"=>38.842415, "lon"=>-79.874802),
"/US/US-WV/blacksville-22457" => array("type"=>"town", "pop"=>176, "name_ru"=>"Блэксвилл", "name_en"=>"Blacksville", "lang"=>"en", "lat"=>39.718672, "lon"=>-80.215630),
"/US/US-WV/blennerhassett-22458" => array("type"=>"cdp", "pop"=>2924, "name_en"=>"Blennerhassett", "lang"=>"en", "lat"=>39.258536, "lon"=>-81.633564),
"/US/US-WV/bluefield-22459" => array("type"=>"city", "pop"=>11093, "name_en"=>"Bluefield", "lang"=>"en", "lat"=>37.260462, "lon"=>-81.218682),
"/US/US-WV/boaz-22460" => array("type"=>"cdp", "pop"=>1137, "name_ru"=>"Бос", "name_en"=>"Boaz", "lang"=>"en", "lat"=>39.371048, "lon"=>-81.486216),
"/US/US-WV/bolivar-22461" => array("type"=>"town", "pop"=>1059, "name_ru"=>"Боливар", "name_en"=>"Bolivar", "lang"=>"en", "lat"=>39.323448, "lon"=>-77.752924),
"/US/US-WV/bradley-22462" => array("type"=>"cdp", "pop"=>2144, "name_ru"=>"Брадли", "name_en"=>"Bradley", "lang"=>"en", "lat"=>37.861365, "lon"=>-81.196308),
"/US/US-WV/bradshaw-22463" => array("type"=>"town", "pop"=>247, "name_en"=>"Bradshaw", "lang"=>"en", "lat"=>37.352882, "lon"=>-81.800869),
"/US/US-WV/bramwell-22464" => array("type"=>"town", "pop"=>412, "name_en"=>"Bramwell", "lang"=>"en", "lat"=>37.326711, "lon"=>-81.312746),
"/US/US-WV/brandonville-22465" => array("type"=>"town", "pop"=>107, "name_en"=>"Brandonville", "lang"=>"en", "lat"=>39.665426, "lon"=>-79.625167),
"/US/US-WV/bridgeport-22466" => array("type"=>"city", "pop"=>7928, "name_ru"=>"Бриджпорт", "name_en"=>"Bridgeport", "lang"=>"en", "lat"=>39.291526, "lon"=>-80.256945),
"/US/US-WV/brookhaven-22467" => array("type"=>"cdp", "pop"=>3836, "name_ru"=>"Брукхейвен", "name_en"=>"Brookhaven", "lang"=>"en", "lat"=>39.610352, "lon"=>-79.884410),
"/US/US-WV/bruceton-mills-22468" => array("type"=>"town", "pop"=>74, "name_en"=>"Bruceton Mills", "lang"=>"en", "lat"=>39.659514, "lon"=>-79.641014),
"/US/US-WV/buckhannon-22469" => array("type"=>"city", "pop"=>5505, "name_en"=>"Buckhannon", "lang"=>"en", "lat"=>38.989513, "lon"=>-80.225433),
"/US/US-WV/buffalo-22470" => array("type"=>"town", "pop"=>1206, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>38.613165, "lon"=>-81.981302),
"/US/US-WV/burnsville-22471" => array("type"=>"town", "pop"=>467, "name_ru"=>"Бернсвилл", "name_en"=>"Burnsville", "lang"=>"en", "lat"=>38.855632, "lon"=>-80.656015),
"/US/US-WV/cairo-22472" => array("type"=>"town", "pop"=>262, "name_ru"=>"Каир", "name_en"=>"Cairo", "lang"=>"en", "lat"=>39.208136, "lon"=>-81.155713),
"/US/US-WV/camden-on-gauley-22473" => array("type"=>"town", "pop"=>148, "name_en"=>"Camden-on-Gauley", "lang"=>"en", "lat"=>38.368870, "lon"=>-80.597825),
"/US/US-WV/cameron-22474" => array("type"=>"city", "pop"=>1077, "name_ru"=>"Кэмерон", "name_en"=>"Cameron", "lang"=>"en", "lat"=>39.826385, "lon"=>-80.570491),
"/US/US-WV/capon-bridge-22475" => array("type"=>"town", "pop"=>272, "name_en"=>"Capon Bridge", "lang"=>"en", "lat"=>39.297753, "lon"=>-78.435489),
"/US/US-WV/cassville-22476" => array("type"=>"cdp", "pop"=>1458, "name_ru"=>"Кассвилл", "name_en"=>"Cassville", "lang"=>"en", "lat"=>39.663887, "lon"=>-80.058930),
"/US/US-WV/cedar-grove-22477" => array("type"=>"town", "pop"=>802, "name_ru"=>"Седар Гроув", "name_en"=>"Cedar Grove", "lang"=>"en", "lat"=>38.220744, "lon"=>-81.428794),
"/US/US-WV/ceredo-22478" => array("type"=>"city", "pop"=>1586, "name_en"=>"Ceredo", "lang"=>"en", "lat"=>38.396624, "lon"=>-82.554135),
"/US/US-WV/chapmanville-22479" => array("type"=>"town", "pop"=>1117, "name_en"=>"Chapmanville", "lang"=>"en", "lat"=>37.973379, "lon"=>-82.018164),
"/US/US-WV/charles-town-22481" => array("type"=>"city", "pop"=>4765, "name_en"=>"Charles Town", "lang"=>"en", "lat"=>39.285057, "lon"=>-77.857534),
"/US/US-WV/charleston-22480" => array("type"=>"city", "pop"=>50302, "name_ru"=>"Чарльстон", "name_en"=>"Charleston", "lang"=>"en", "lat"=>38.350550, "lon"=>-81.630439),
"/US/US-WV/chattaroy-22482" => array("type"=>"cdp", "pop"=>1182, "name_en"=>"Chattaroy", "lang"=>"en", "lat"=>37.705043, "lon"=>-82.268417),
"/US/US-WV/cheat-lake-22483" => array("type"=>"cdp", "pop"=>3992, "name_en"=>"Cheat Lake", "lang"=>"en", "lat"=>39.672046, "lon"=>-79.853631),
"/US/US-WV/chesapeake-22484" => array("type"=>"town", "pop"=>1542, "name_en"=>"Chesapeake", "lang"=>"en", "lat"=>38.222440, "lon"=>-81.536202),
"/US/US-WV/chester-22485" => array("type"=>"city", "pop"=>2332, "name_ru"=>"Честер", "name_en"=>"Chester", "lang"=>"en", "lat"=>40.612192, "lon"=>-80.563782),
"/US/US-WV/clarksburg-22486" => array("type"=>"city", "pop"=>16441, "name_ru"=>"Кларксбург", "name_en"=>"Clarksburg", "lang"=>"en", "lat"=>39.280394, "lon"=>-80.331821),
"/US/US-WV/clay-22487" => array("type"=>"town", "pop"=>561, "name_ru"=>"Клей", "name_en"=>"Clay", "lang"=>"en", "lat"=>38.462754, "lon"=>-81.080194),
"/US/US-WV/clearview-22488" => array("type"=>"village", "pop"=>548, "name_en"=>"Clearview", "lang"=>"en", "lat"=>40.142368, "lon"=>-80.691277),
"/US/US-WV/clendenin-22489" => array("type"=>"town", "pop"=>1042, "name_en"=>"Clendenin", "lang"=>"en", "lat"=>38.487448, "lon"=>-81.349314),
"/US/US-WV/coal-city-22490" => array("type"=>"cdp", "pop"=>1876, "name_ru"=>"Коал Сити", "name_en"=>"Coal City", "lang"=>"en", "lat"=>37.676182, "lon"=>-81.214491),
"/US/US-WV/coal-fork-22491" => array("type"=>"cdp", "pop"=>2100, "name_en"=>"Coal Fork", "lang"=>"en", "lat"=>38.316523, "lon"=>-81.520927),
"/US/US-WV/corporation-of-ranson-22492" => array("type"=>"town", "pop"=>2890, "name_en"=>"Corporation of Ranson", "lang"=>"en", "lat"=>39.299480, "lon"=>-77.859830),
"/US/US-WV/cowen-22493" => array("type"=>"town", "pop"=>486, "name_en"=>"Cowen", "lang"=>"en", "lat"=>38.410231, "lon"=>-80.555115),
"/US/US-WV/crab-orchard-22494" => array("type"=>"cdp", "pop"=>2919, "name_ru"=>"Крэб Орчард", "name_en"=>"Crab Orchard", "lang"=>"en", "lat"=>37.740589, "lon"=>-81.231398),
"/US/US-WV/craigsville-22495" => array("type"=>"cdp", "pop"=>1955, "name_en"=>"Craigsville", "lang"=>"en", "lat"=>38.320505, "lon"=>-80.645839),
"/US/US-WV/cross-lanes-22496" => array("type"=>"cdp", "pop"=>10878, "name_en"=>"Cross Lanes", "lang"=>"en", "lat"=>38.435081, "lon"=>-81.769903),
"/US/US-WV/culloden-22497" => array("type"=>"cdp", "pop"=>2907, "name_ru"=>"Каллоден", "name_en"=>"Culloden", "lang"=>"en", "lat"=>38.417062, "lon"=>-82.067823),
"/US/US-WV/daniels-22498" => array("type"=>"cdp", "pop"=>1714, "name_en"=>"Daniels", "lang"=>"en", "lat"=>37.722597, "lon"=>-81.125993),
"/US/US-WV/danville-22499" => array("type"=>"town", "pop"=>532, "name_ru"=>"Данвилл", "name_en"=>"Danville", "lang"=>"en", "lat"=>38.081323, "lon"=>-81.834752),
"/US/US-WV/davis-22500" => array("type"=>"town", "pop"=>578, "name_ru"=>"Давис", "name_en"=>"Davis", "lang"=>"en", "lat"=>39.131003, "lon"=>-79.466545),
"/US/US-WV/davy-22501" => array("type"=>"town", "pop"=>338, "name_en"=>"Davy", "lang"=>"en", "lat"=>37.474109, "lon"=>-81.650026),
"/US/US-WV/delbarton-22502" => array("type"=>"town", "pop"=>432, "name_en"=>"Delbarton", "lang"=>"en", "lat"=>37.704886, "lon"=>-82.185793),
"/US/US-WV/despard-22503" => array("type"=>"cdp", "pop"=>1018, "name_en"=>"Despard", "lang"=>"en", "lat"=>39.292169, "lon"=>-80.317636),
"/US/US-WV/dunbar-22504" => array("type"=>"city", "pop"=>7664, "name_en"=>"Dunbar", "lang"=>"en", "lat"=>38.368187, "lon"=>-81.736728),
"/US/US-WV/durbin-22505" => array("type"=>"town", "pop"=>238, "name_en"=>"Durbin", "lang"=>"en", "lat"=>38.547146, "lon"=>-79.828150),
"/US/US-WV/east-bank-22506" => array("type"=>"town", "pop"=>888, "name_en"=>"East Bank", "lang"=>"en", "lat"=>38.215665, "lon"=>-81.445197),
"/US/US-WV/eleanor-22507" => array("type"=>"town", "pop"=>1555, "name_en"=>"Eleanor", "lang"=>"en", "lat"=>38.538949, "lon"=>-81.931189),
"/US/US-WV/elizabeth-22508" => array("type"=>"town", "pop"=>927, "name_ru"=>"Элизабет", "name_en"=>"Elizabeth", "lang"=>"en", "lat"=>39.062418, "lon"=>-81.397372),
"/US/US-WV/elk-garden-22509" => array("type"=>"town", "pop"=>205, "name_en"=>"Elk Garden", "lang"=>"en", "lat"=>39.385962, "lon"=>-79.154680),
"/US/US-WV/elkins-22510" => array("type"=>"city", "pop"=>6976, "name_ru"=>"Элкинс", "name_en"=>"Elkins", "lang"=>"en", "lat"=>38.925060, "lon"=>-79.853565),
"/US/US-WV/elkview-22511" => array("type"=>"cdp", "pop"=>1047, "name_en"=>"Elkview", "lang"=>"en", "lat"=>38.434911, "lon"=>-81.474873),
"/US/US-WV/ellenboro-22512" => array("type"=>"town", "pop"=>361, "name_en"=>"Ellenboro", "lang"=>"en", "lat"=>39.265459, "lon"=>-81.054999),
"/US/US-WV/enterprise-22513" => array("type"=>"cdp", "pop"=>1058, "name_ru"=>"Энтерпрайз", "name_en"=>"Enterprise", "lang"=>"en", "lat"=>39.418710, "lon"=>-80.277728),
"/US/US-WV/fairlea-22514" => array("type"=>"cdp", "pop"=>1743, "name_en"=>"Fairlea", "lang"=>"en", "lat"=>37.775093, "lon"=>-80.459823),
"/US/US-WV/fairmont-22515" => array("type"=>"city", "pop"=>19024, "name_ru"=>"Фэрмонт", "name_en"=>"Fairmont", "lang"=>"en", "lat"=>39.480798, "lon"=>-80.146087),
"/US/US-WV/fairview-22516" => array("type"=>"town", "pop"=>441, "name_ru"=>"Фэрвью", "name_en"=>"Fairview", "lang"=>"en", "lat"=>39.592617, "lon"=>-80.247322),
"/US/US-WV/falling-spring-22517" => array("type"=>"town", "pop"=>199, "name_en"=>"Falling Spring", "lang"=>"en", "lat"=>37.991995, "lon"=>-80.357286),
"/US/US-WV/farmington-22518" => array("type"=>"town", "pop"=>387, "name_ru"=>"Фармингтон", "name_en"=>"Farmington", "lang"=>"en", "lat"=>39.511503, "lon"=>-80.252533),
"/US/US-WV/fayetteville-22519" => array("type"=>"town", "pop"=>2663, "name_ru"=>"Фейетвилл", "name_en"=>"Fayetteville", "lang"=>"en", "lat"=>38.052093, "lon"=>-81.105466),
"/US/US-WV/flatwoods-22520" => array("type"=>"town", "pop"=>349, "name_ru"=>"Флэтвудс", "name_en"=>"Flatwoods", "lang"=>"en", "lat"=>38.721275, "lon"=>-80.653141),
"/US/US-WV/flemington-22521" => array("type"=>"town", "pop"=>287, "name_ru"=>"Флемингтон", "name_en"=>"Flemington", "lang"=>"en", "lat"=>39.267312, "lon"=>-80.128153),
"/US/US-WV/follansbee-22522" => array("type"=>"city", "pop"=>2855, "name_en"=>"Follansbee", "lang"=>"en", "lat"=>40.337200, "lon"=>-80.597179),
"/US/US-WV/fort-ashby-22523" => array("type"=>"cdp", "pop"=>1288, "name_en"=>"Fort Ashby", "lang"=>"en", "lat"=>39.498175, "lon"=>-78.765394),
"/US/US-WV/fort-gay-22524" => array("type"=>"town", "pop"=>796, "name_en"=>"Fort Gay", "lang"=>"en", "lat"=>38.118483, "lon"=>-82.594539),
"/US/US-WV/franklin-22525" => array("type"=>"town", "pop"=>789, "name_ru"=>"Франклин", "name_en"=>"Franklin", "lang"=>"en", "lat"=>38.646030, "lon"=>-79.333105),
"/US/US-WV/friendly-22526" => array("type"=>"town", "pop"=>143, "name_ru"=>"Френдли", "name_en"=>"Friendly", "lang"=>"en", "lat"=>39.512904, "lon"=>-81.062068),
"/US/US-WV/gary-22527" => array("type"=>"city", "pop"=>748, "name_ru"=>"Гэри", "name_en"=>"Gary", "lang"=>"en", "lat"=>37.361874, "lon"=>-81.538127),
"/US/US-WV/gassaway-22528" => array("type"=>"town", "pop"=>875, "name_en"=>"Gassaway", "lang"=>"en", "lat"=>38.670614, "lon"=>-80.770592),
"/US/US-WV/gauley-bridge-22529" => array("type"=>"town", "pop"=>690, "name_en"=>"Gauley Bridge", "lang"=>"en", "lat"=>38.168294, "lon"=>-81.201704),
"/US/US-WV/gilbert-22530" => array("type"=>"town", "pop"=>387, "name_ru"=>"Гилберт", "name_en"=>"Gilbert", "lang"=>"en", "lat"=>37.614388, "lon"=>-81.867372),
"/US/US-WV/gilbert-creek-22531" => array("type"=>"cdp", "pop"=>1784, "name_en"=>"Gilbert Creek", "lang"=>"en", "lat"=>37.575737, "lon"=>-81.895026),
"/US/US-WV/glasgow-22532" => array("type"=>"town", "pop"=>732, "name_ru"=>"Гласго", "name_en"=>"Glasgow", "lang"=>"en", "lat"=>38.209776, "lon"=>-81.422563),
"/US/US-WV/glen-dale-22533" => array("type"=>"city", "pop"=>1396, "name_en"=>"Glen Dale", "lang"=>"en", "lat"=>39.949024, "lon"=>-80.756208),
"/US/US-WV/glenville-22534" => array("type"=>"town", "pop"=>1456, "name_ru"=>"Гленвилл", "name_en"=>"Glenville", "lang"=>"en", "lat"=>38.938407, "lon"=>-80.833441),
"/US/US-WV/grafton-22535" => array("type"=>"city", "pop"=>5315, "name_ru"=>"Графтон", "name_en"=>"Grafton", "lang"=>"en", "lat"=>39.340566, "lon"=>-80.017153),
"/US/US-WV/grant-town-22537" => array("type"=>"town", "pop"=>637, "name_en"=>"Grant Town", "lang"=>"en", "lat"=>39.557276, "lon"=>-80.178686),
"/US/US-WV/grantsville-22536" => array("type"=>"town", "pop"=>533, "name_ru"=>"Грантсвилл", "name_en"=>"Grantsville", "lang"=>"en", "lat"=>38.920459, "lon"=>-81.093354),
"/US/US-WV/granville-22538" => array("type"=>"town", "pop"=>812, "name_ru"=>"Гранвилл", "name_en"=>"Granville", "lang"=>"en", "lat"=>39.646667, "lon"=>-79.990553),
"/US/US-WV/hambleton-22539" => array("type"=>"town", "pop"=>249, "name_en"=>"Hambleton", "lang"=>"en", "lat"=>39.081241, "lon"=>-79.646074),
"/US/US-WV/hamlin-22540" => array("type"=>"town", "pop"=>1102, "name_ru"=>"Хэмлин", "name_en"=>"Hamlin", "lang"=>"en", "lat"=>38.279125, "lon"=>-82.103841),
"/US/US-WV/handley-22541" => array("type"=>"town", "pop"=>340, "name_en"=>"Handley", "lang"=>"en", "lat"=>38.186558, "lon"=>-81.366377),
"/US/US-WV/harman-22542" => array("type"=>"town", "pop"=>123, "name_en"=>"Harman", "lang"=>"en", "lat"=>38.921445, "lon"=>-79.524805),
"/US/US-WV/harpers-ferry-22543" => array("type"=>"town", "pop"=>309, "name_ru"=>"Харперс Ферри", "name_en"=>"Harpers Ferry", "lang"=>"en", "lat"=>39.325128, "lon"=>-77.742235),
"/US/US-WV/harrisville-22544" => array("type"=>"town", "pop"=>1817, "name_ru"=>"Харрисвилл", "name_en"=>"Harrisville", "lang"=>"en", "lat"=>39.210718, "lon"=>-81.049657),
"/US/US-WV/hartford-city-22545" => array("type"=>"town", "pop"=>518, "name_ru"=>"Хартфорд Сити", "name_en"=>"Hartford City", "lang"=>"en", "lat"=>38.998554, "lon"=>-81.988680),
"/US/US-WV/harts-22546" => array("type"=>"cdp", "pop"=>2332, "name_en"=>"Harts", "lang"=>"en", "lat"=>38.030591, "lon"=>-82.130810),
"/US/US-WV/hedgesville-22547" => array("type"=>"town", "pop"=>242, "name_en"=>"Hedgesville", "lang"=>"en", "lat"=>39.554668, "lon"=>-77.994391),
"/US/US-WV/henderson-22548" => array("type"=>"town", "pop"=>309, "name_ru"=>"Хендерсон", "name_en"=>"Henderson", "lang"=>"en", "lat"=>38.831459, "lon"=>-82.135693),
"/US/US-WV/hendricks-22549" => array("type"=>"town", "pop"=>316, "name_ru"=>"Хендрикс", "name_en"=>"Hendricks", "lang"=>"en", "lat"=>39.075235, "lon"=>-79.630257),
"/US/US-WV/hillsboro-22550" => array("type"=>"town", "pop"=>223, "name_ru"=>"Хилсборо", "name_en"=>"Hillsboro", "lang"=>"en", "lat"=>38.135002, "lon"=>-80.212466),
"/US/US-WV/hinton-22551" => array("type"=>"city", "pop"=>2532, "name_ru"=>"Хинтон", "name_en"=>"Hinton", "lang"=>"en", "lat"=>37.662947, "lon"=>-80.883134),
"/US/US-WV/holden-22552" => array("type"=>"cdp", "pop"=>1246, "name_ru"=>"Холден", "name_en"=>"Holden", "lang"=>"en", "lat"=>37.814388, "lon"=>-82.066661),
"/US/US-WV/hooverson-heights-22553" => array("type"=>"cdp", "pop"=>3056, "name_en"=>"Hooverson Heights", "lang"=>"en", "lat"=>40.318400, "lon"=>-80.579825),
"/US/US-WV/hundred-22554" => array("type"=>"town", "pop"=>322, "name_en"=>"Hundred", "lang"=>"en", "lat"=>39.684105, "lon"=>-80.457516),
"/US/US-WV/huntington-22555" => array("type"=>"city", "pop"=>49185, "name_ru"=>"Хантингтон", "name_en"=>"Huntington", "lang"=>"en", "lat"=>38.412950, "lon"=>-82.433767),
"/US/US-WV/hurricane-22556" => array("type"=>"city", "pop"=>6335, "name_en"=>"Hurricane", "lang"=>"en", "lat"=>38.436812, "lon"=>-82.016939),
"/US/US-WV/huttonsville-22557" => array("type"=>"town", "pop"=>212, "name_en"=>"Huttonsville", "lang"=>"en", "lat"=>38.714737, "lon"=>-79.976269),
"/US/US-WV/iaeger-22558" => array("type"=>"town", "pop"=>294, "name_en"=>"Iaeger", "lang"=>"en", "lat"=>37.463752, "lon"=>-81.813397),
"/US/US-WV/inwood-22559" => array("type"=>"cdp", "pop"=>1360, "name_ru"=>"Инвуд", "name_en"=>"Inwood", "lang"=>"en", "lat"=>39.352573, "lon"=>-78.051238),
"/US/US-WV/jane-lew-22560" => array("type"=>"town", "pop"=>411, "name_en"=>"Jane Lew", "lang"=>"en", "lat"=>39.109594, "lon"=>-80.408978),
"/US/US-WV/junior-22561" => array("type"=>"town", "pop"=>444, "name_en"=>"Junior", "lang"=>"en", "lat"=>38.977276, "lon"=>-79.950504),
"/US/US-WV/kenova-22562" => array("type"=>"city", "pop"=>3272, "name_en"=>"Kenova", "lang"=>"en", "lat"=>38.403200, "lon"=>-82.580149),
"/US/US-WV/kermit-22563" => array("type"=>"town", "pop"=>220, "name_en"=>"Kermit", "lang"=>"en", "lat"=>37.839190, "lon"=>-82.407819),
"/US/US-WV/keyser-22564" => array("type"=>"city", "pop"=>5230, "name_en"=>"Keyser", "lang"=>"en", "lat"=>39.439500, "lon"=>-78.984014),
"/US/US-WV/keystone-22565" => array("type"=>"city", "pop"=>366, "name_ru"=>"Кистоун", "name_en"=>"Keystone", "lang"=>"en", "lat"=>37.415005, "lon"=>-81.446257),
"/US/US-WV/kimball-22566" => array("type"=>"town", "pop"=>337, "name_en"=>"Kimball", "lang"=>"en", "lat"=>37.427382, "lon"=>-81.506924),
"/US/US-WV/kingwood-22567" => array("type"=>"city", "pop"=>2947, "name_en"=>"Kingwood", "lang"=>"en", "lat"=>39.471053, "lon"=>-79.683065),
"/US/US-WV/leon-22568" => array("type"=>"town", "pop"=>126, "name_ru"=>"Леон", "name_en"=>"Leon", "lang"=>"en", "lat"=>38.747346, "lon"=>-81.956128),
"/US/US-WV/lester-22569" => array("type"=>"town", "pop"=>314, "name_ru"=>"Лестер", "name_en"=>"Lester", "lang"=>"en", "lat"=>37.735106, "lon"=>-81.300931),
"/US/US-WV/lewisburg-22570" => array("type"=>"city", "pop"=>3533, "name_ru"=>"Льюисбург", "name_en"=>"Lewisburg", "lang"=>"en", "lat"=>37.808955, "lon"=>-80.431900),
"/US/US-WV/littleton-22571" => array("type"=>"town", "pop"=>198, "name_ru"=>"Литтлтон", "name_en"=>"Littleton", "lang"=>"en", "lat"=>39.699557, "lon"=>-80.514774),
"/US/US-WV/logan-22572" => array("type"=>"city", "pop"=>1501, "name_ru"=>"Логан", "name_en"=>"Logan", "lang"=>"en", "lat"=>37.849110, "lon"=>-81.986858),
"/US/US-WV/lost-creek-22573" => array("type"=>"town", "pop"=>501, "name_en"=>"Lost Creek", "lang"=>"en", "lat"=>39.159127, "lon"=>-80.350531),
"/US/US-WV/lubeck-22574" => array("type"=>"cdp", "pop"=>1579, "name_en"=>"Lubeck", "lang"=>"en", "lat"=>39.236363, "lon"=>-81.629485),
"/US/US-WV/lumberport-22575" => array("type"=>"town", "pop"=>967, "name_en"=>"Lumberport", "lang"=>"en", "lat"=>39.372998, "lon"=>-80.348542),
"/US/US-WV/mabscott-22576" => array("type"=>"town", "pop"=>1350, "name_en"=>"Mabscott", "lang"=>"en", "lat"=>37.768350, "lon"=>-81.212306),
"/US/US-WV/macarthur-22577" => array("type"=>"cdp", "pop"=>1595, "name_en"=>"MacArthur", "lang"=>"en", "lat"=>37.757038, "lon"=>-81.209491),
"/US/US-WV/madison-22579" => array("type"=>"city", "pop"=>2576, "name_ru"=>"Мадисон", "name_en"=>"Madison", "lang"=>"en", "lat"=>38.059315, "lon"=>-81.800820),
"/US/US-WV/mallory-22580" => array("type"=>"cdp", "pop"=>1126, "name_en"=>"Mallory", "lang"=>"en", "lat"=>37.734795, "lon"=>-81.818858),
"/US/US-WV/man-22581" => array("type"=>"town", "pop"=>692, "name_en"=>"Man", "lang"=>"en", "lat"=>37.742592, "lon"=>-81.874343),
"/US/US-WV/mannington-22582" => array("type"=>"city", "pop"=>2079, "name_en"=>"Mannington", "lang"=>"en", "lat"=>39.526502, "lon"=>-80.341243),
"/US/US-WV/marlinton-22583" => array("type"=>"town", "pop"=>1190, "name_en"=>"Marlinton", "lang"=>"en", "lat"=>38.220057, "lon"=>-80.087076),
"/US/US-WV/marmet-22584" => array("type"=>"city", "pop"=>1605, "name_en"=>"Marmet", "lang"=>"en", "lat"=>38.246257, "lon"=>-81.569822),
"/US/US-WV/martinsburg-22585" => array("type"=>"city", "pop"=>17020, "name_ru"=>"Мартинсбург", "name_en"=>"Martinsburg", "lang"=>"en", "lat"=>39.460281, "lon"=>-77.969283),
"/US/US-WV/mason-22586" => array("type"=>"town", "pop"=>1040, "name_ru"=>"Мейсон", "name_en"=>"Mason", "lang"=>"en", "lat"=>39.018851, "lon"=>-82.032449),
"/US/US-WV/masontown-22587" => array("type"=>"town", "pop"=>653, "name_en"=>"Masontown", "lang"=>"en", "lat"=>39.551738, "lon"=>-79.800376),
"/US/US-WV/matewan-22588" => array("type"=>"town", "pop"=>479, "name_en"=>"Matewan", "lang"=>"en", "lat"=>37.623716, "lon"=>-82.167719),
"/US/US-WV/matoaka-22589" => array("type"=>"town", "pop"=>305, "name_en"=>"Matoaka", "lang"=>"en", "lat"=>37.418625, "lon"=>-81.242207),
"/US/US-WV/mcmechen-22578" => array("type"=>"city", "pop"=>1724, "name_en"=>"McMechen", "lang"=>"en", "lat"=>39.986528, "lon"=>-80.733167),
"/US/US-WV/meadow-bridge-22590" => array("type"=>"town", "pop"=>306, "name_en"=>"Meadow Bridge", "lang"=>"en", "lat"=>37.862282, "lon"=>-80.855357),
"/US/US-WV/middlebourne-22591" => array("type"=>"town", "pop"=>810, "name_en"=>"Middlebourne", "lang"=>"en", "lat"=>39.493593, "lon"=>-80.909398),
"/US/US-WV/mill-creek-22592" => array("type"=>"town", "pop"=>641, "name_ru"=>"Милл Крик", "name_en"=>"Mill Creek", "lang"=>"en", "lat"=>38.731640, "lon"=>-79.972487),
"/US/US-WV/milton-22593" => array("type"=>"town", "pop"=>2399, "name_ru"=>"Мильтон", "name_en"=>"Milton", "lang"=>"en", "lat"=>38.433872, "lon"=>-82.137709),
"/US/US-WV/mineralwells-22594" => array("type"=>"cdp", "pop"=>1698, "name_en"=>"Mineralwells", "lang"=>"en", "lat"=>39.179096, "lon"=>-81.511578),
"/US/US-WV/mitchell-heights-22595" => array("type"=>"town", "pop"=>274, "name_en"=>"Mitchell Heights", "lang"=>"en", "lat"=>37.909230, "lon"=>-81.988666),
"/US/US-WV/monongah-22596" => array("type"=>"town", "pop"=>911, "name_en"=>"Monongah", "lang"=>"en", "lat"=>39.459622, "lon"=>-80.217663),
"/US/US-WV/montcalm-22597" => array("type"=>"cdp", "pop"=>1023, "name_en"=>"Montcalm", "lang"=>"en", "lat"=>37.350261, "lon"=>-81.251838),
"/US/US-WV/montgomery-22598" => array("type"=>"city", "pop"=>1921, "name_ru"=>"Монтгомери", "name_en"=>"Montgomery", "lang"=>"en", "lat"=>38.173849, "lon"=>-81.318465),
"/US/US-WV/montrose-22599" => array("type"=>"town", "pop"=>156, "name_ru"=>"Монтроуз", "name_en"=>"Montrose", "lang"=>"en", "lat"=>39.069636, "lon"=>-79.810663),
"/US/US-WV/moorefield-22600" => array("type"=>"town", "pop"=>2438, "name_ru"=>"Мурфилд", "name_en"=>"Moorefield", "lang"=>"en", "lat"=>39.062892, "lon"=>-78.966173),
"/US/US-WV/morgantown-22601" => array("type"=>"city", "pop"=>29642, "name_ru"=>"Моргантаун", "name_en"=>"Morgantown", "lang"=>"en", "lat"=>39.635649, "lon"=>-79.949772),
"/US/US-WV/moundsville-22602" => array("type"=>"city", "pop"=>9109, "name_en"=>"Moundsville", "lang"=>"en", "lat"=>39.922776, "lon"=>-80.741782),
"/US/US-WV/mount-gay-shamrock-22603" => array("type"=>"cdp", "pop"=>3377, "name_en"=>"Mount Gay-Shamrock", "lang"=>"en", "lat"=>37.854184, "lon"=>-82.042691),
"/US/US-WV/mount-hope-22604" => array("type"=>"city", "pop"=>1381, "name_ru"=>"Маунт Хоуп", "name_en"=>"Mount Hope", "lang"=>"en", "lat"=>37.895280, "lon"=>-81.171014),
"/US/US-WV/mullens-22605" => array("type"=>"city", "pop"=>1584, "name_en"=>"Mullens", "lang"=>"en", "lat"=>37.579704, "lon"=>-81.385198),
"/US/US-WV/new-cumberland-22607" => array("type"=>"city", "pop"=>990, "name_en"=>"New Cumberland", "lang"=>"en", "lat"=>40.495984, "lon"=>-80.607149),
"/US/US-WV/new-haven-22609" => array("type"=>"town", "pop"=>1528, "name_ru"=>"Нью Хейвен", "name_en"=>"New Haven", "lang"=>"en", "lat"=>38.987980, "lon"=>-81.967458),
"/US/US-WV/new-martinsville-22610" => array("type"=>"city", "pop"=>5564, "name_en"=>"New Martinsville", "lang"=>"en", "lat"=>39.664088, "lon"=>-80.859718),
"/US/US-WV/newburg-22606" => array("type"=>"town", "pop"=>360, "name_ru"=>"Ньюбург", "name_en"=>"Newburg", "lang"=>"en", "lat"=>39.388362, "lon"=>-79.854266),
"/US/US-WV/newell-22608" => array("type"=>"cdp", "pop"=>1724, "name_ru"=>"Ньюуэлл", "name_en"=>"Newell", "lang"=>"en", "lat"=>40.617480, "lon"=>-80.601063),
"/US/US-WV/nitro-22611" => array("type"=>"city", "pop"=>6784, "name_en"=>"Nitro", "lang"=>"en", "lat"=>38.417632, "lon"=>-81.831112),
"/US/US-WV/north-hills-22613" => array("type"=>"town", "pop"=>857, "name_en"=>"North Hills", "lang"=>"en", "lat"=>39.314720, "lon"=>-81.510468),
"/US/US-WV/northfork-22612" => array("type"=>"town", "pop"=>420, "name_en"=>"Northfork", "lang"=>"en", "lat"=>37.419052, "lon"=>-81.426184),
"/US/US-WV/nutter-fort-22614" => array("type"=>"town", "pop"=>1633, "name_en"=>"Nutter Fort", "lang"=>"en", "lat"=>39.259540, "lon"=>-80.324611),
"/US/US-WV/oak-hill-22615" => array("type"=>"city", "pop"=>7211, "name_ru"=>"Оук Хилл", "name_en"=>"Oak Hill", "lang"=>"en", "lat"=>37.981958, "lon"=>-81.147711),
"/US/US-WV/oakvale-22616" => array("type"=>"town", "pop"=>140, "name_en"=>"Oakvale", "lang"=>"en", "lat"=>37.332912, "lon"=>-80.965975),
"/US/US-WV/oceana-22617" => array("type"=>"town", "pop"=>1411, "name_en"=>"Oceana", "lang"=>"en", "lat"=>37.691832, "lon"=>-81.632742),
"/US/US-WV/osage-22618" => array("type"=>"town", "pop"=>183, "name_ru"=>"Осейдж", "name_en"=>"Osage", "lang"=>"en", "lat"=>39.658266, "lon"=>-80.009445),
"/US/US-WV/paden-city-22619" => array("type"=>"city", "pop"=>2599, "name_en"=>"Paden City", "lang"=>"en", "lat"=>39.602924, "lon"=>-80.935360),
"/US/US-WV/parkersburg-22620" => array("type"=>"city", "pop"=>31611, "name_ru"=>"Паркерсбург", "name_en"=>"Parkersburg", "lang"=>"en", "lat"=>39.263967, "lon"=>-81.542969),
"/US/US-WV/parsons-22621" => array("type"=>"city", "pop"=>1376, "name_ru"=>"Парсонс", "name_en"=>"Parsons", "lang"=>"en", "lat"=>39.093346, "lon"=>-79.679265),
"/US/US-WV/paw-paw-22622" => array("type"=>"town", "pop"=>489, "name_ru"=>"По По", "name_en"=>"Paw Paw", "lang"=>"en", "lat"=>39.530893, "lon"=>-78.456068),
"/US/US-WV/pax-22623" => array("type"=>"town", "pop"=>164, "name_en"=>"Pax", "lang"=>"en", "lat"=>37.909611, "lon"=>-81.263884),
"/US/US-WV/pea-ridge-22624" => array("type"=>"cdp", "pop"=>6535, "name_ru"=>"Пи Ридж", "name_en"=>"Pea Ridge", "lang"=>"en", "lat"=>38.413967, "lon"=>-82.319961),
"/US/US-WV/pennsboro-22625" => array("type"=>"city", "pop"=>1155, "name_en"=>"Pennsboro", "lang"=>"en", "lat"=>39.281632, "lon"=>-80.974969),
"/US/US-WV/petersburg-22626" => array("type"=>"city", "pop"=>2774, "name_ru"=>"Петребург", "name_en"=>"Petersburg", "lang"=>"en", "lat"=>38.996330, "lon"=>-79.128218),
"/US/US-WV/peterstown-22627" => array("type"=>"town", "pop"=>499, "name_en"=>"Peterstown", "lang"=>"en", "lat"=>37.398517, "lon"=>-80.795304),
"/US/US-WV/philippi-22628" => array("type"=>"city", "pop"=>2783, "name_en"=>"Philippi", "lang"=>"en", "lat"=>39.152154, "lon"=>-80.044245),
"/US/US-WV/piedmont-22629" => array("type"=>"town", "pop"=>919, "name_ru"=>"Пидмонт", "name_en"=>"Piedmont", "lang"=>"en", "lat"=>39.479317, "lon"=>-79.047839),
"/US/US-WV/pinch-22630" => array("type"=>"cdp", "pop"=>2695, "name_en"=>"Pinch", "lang"=>"en", "lat"=>38.407365, "lon"=>-81.479829),
"/US/US-WV/pine-grove-22631" => array("type"=>"town", "pop"=>520, "name_en"=>"Pine Grove", "lang"=>"en", "lat"=>39.564744, "lon"=>-80.682586),
"/US/US-WV/pineville-22632" => array("type"=>"town", "pop"=>655, "name_ru"=>"Пайнвилл", "name_en"=>"Pineville", "lang"=>"en", "lat"=>37.585602, "lon"=>-81.534995),
"/US/US-WV/piney-view-22633" => array("type"=>"cdp", "pop"=>1085, "name_en"=>"Piney View", "lang"=>"en", "lat"=>37.847108, "lon"=>-81.138875),
"/US/US-WV/poca-22634" => array("type"=>"town", "pop"=>1032, "name_en"=>"Poca", "lang"=>"en", "lat"=>38.474395, "lon"=>-81.812843),
"/US/US-WV/point-pleasant-22635" => array("type"=>"city", "pop"=>4451, "name_en"=>"Point Pleasant", "lang"=>"en", "lat"=>38.853880, "lon"=>-82.130662),
"/US/US-WV/powellton-22636" => array("type"=>"cdp", "pop"=>1905, "name_en"=>"Powellton", "lang"=>"en", "lat"=>38.090488, "lon"=>-81.316226),
"/US/US-WV/pratt-22637" => array("type"=>"town", "pop"=>530, "name_ru"=>"Прэтт", "name_en"=>"Pratt", "lang"=>"en", "lat"=>38.207045, "lon"=>-81.387518),
"/US/US-WV/princeton-22638" => array("type"=>"city", "pop"=>6261, "name_ru"=>"Принстон", "name_en"=>"Princeton", "lang"=>"en", "lat"=>37.368641, "lon"=>-81.096226),
"/US/US-WV/prosperity-22639" => array("type"=>"cdp", "pop"=>1322, "name_en"=>"Prosperity", "lang"=>"en", "lat"=>37.835255, "lon"=>-81.203176),
"/US/US-WV/pullman-22640" => array("type"=>"town", "pop"=>170, "name_en"=>"Pullman", "lang"=>"en", "lat"=>39.187404, "lon"=>-80.949564),
"/US/US-WV/quinwood-22641" => array("type"=>"town", "pop"=>410, "name_en"=>"Quinwood", "lang"=>"en", "lat"=>38.057777, "lon"=>-80.703562),
"/US/US-WV/rainelle-22642" => array("type"=>"town", "pop"=>1470, "name_en"=>"Rainelle", "lang"=>"en", "lat"=>37.968611, "lon"=>-80.771041),
"/US/US-WV/ravenswood-22643" => array("type"=>"city", "pop"=>3935, "name_en"=>"Ravenswood", "lang"=>"en", "lat"=>38.957206, "lon"=>-81.763114),
"/US/US-WV/red-jacket-22644" => array("type"=>"cdp", "pop"=>760, "name_en"=>"Red Jacket", "lang"=>"en", "lat"=>37.649150, "lon"=>-82.132301),
"/US/US-WV/reedsville-22645" => array("type"=>"town", "pop"=>536, "name_en"=>"Reedsville", "lang"=>"en", "lat"=>39.510397, "lon"=>-79.800107),
"/US/US-WV/reedy-22646" => array("type"=>"town", "pop"=>184, "name_en"=>"Reedy", "lang"=>"en", "lat"=>38.898811, "lon"=>-81.427194),
"/US/US-WV/rhodell-22647" => array("type"=>"town", "pop"=>230, "name_en"=>"Rhodell", "lang"=>"en", "lat"=>37.609783, "lon"=>-81.304957),
"/US/US-WV/richwood-22648" => array("type"=>"city", "pop"=>2309, "name_ru"=>"Ричвуд", "name_en"=>"Richwood", "lang"=>"en", "lat"=>38.221549, "lon"=>-80.537204),
"/US/US-WV/ridgeley-22649" => array("type"=>"town", "pop"=>681, "name_en"=>"Ridgeley", "lang"=>"en", "lat"=>39.642700, "lon"=>-78.770349),
"/US/US-WV/ripley-22650" => array("type"=>"city", "pop"=>3241, "name_ru"=>"Рипли", "name_en"=>"Ripley", "lang"=>"en", "lat"=>38.820102, "lon"=>-81.709128),
"/US/US-WV/rivesville-22651" => array("type"=>"town", "pop"=>905, "name_en"=>"Rivesville", "lang"=>"en", "lat"=>39.531592, "lon"=>-80.120749),
"/US/US-WV/romney-22652" => array("type"=>"city", "pop"=>1912, "name_en"=>"Romney", "lang"=>"en", "lat"=>39.346052, "lon"=>-78.755464),
"/US/US-WV/ronceverte-22653" => array("type"=>"city", "pop"=>1603, "name_en"=>"Ronceverte", "lang"=>"en", "lat"=>37.748864, "lon"=>-80.472595),
"/US/US-WV/rowlesburg-22654" => array("type"=>"town", "pop"=>613, "name_en"=>"Rowlesburg", "lang"=>"en", "lat"=>39.348491, "lon"=>-79.677437),
"/US/US-WV/rupert-22655" => array("type"=>"town", "pop"=>920, "name_ru"=>"Руперт", "name_en"=>"Rupert", "lang"=>"en", "lat"=>37.965836, "lon"=>-80.687310),
"/US/US-WV/salem-22658" => array("type"=>"city", "pop"=>2050, "name_ru"=>"Салем", "name_en"=>"Salem", "lang"=>"en", "lat"=>39.282921, "lon"=>-80.565108),
"/US/US-WV/sand-fork-22659" => array("type"=>"town", "pop"=>169, "name_en"=>"Sand Fork", "lang"=>"en", "lat"=>38.915243, "lon"=>-80.748763),
"/US/US-WV/shady-spring-22660" => array("type"=>"cdp", "pop"=>1929, "name_en"=>"Shady Spring", "lang"=>"en", "lat"=>37.702981, "lon"=>-81.095183),
"/US/US-WV/shepherdstown-22661" => array("type"=>"town", "pop"=>1146, "name_en"=>"Shepherdstown", "lang"=>"en", "lat"=>39.431125, "lon"=>-77.805071),
"/US/US-WV/shinnston-22662" => array("type"=>"city", "pop"=>2243, "name_en"=>"Shinnston", "lang"=>"en", "lat"=>39.393215, "lon"=>-80.299478),
"/US/US-WV/sissonville-22663" => array("type"=>"cdp", "pop"=>4290, "name_en"=>"Sissonville", "lang"=>"en", "lat"=>38.501901, "lon"=>-81.644680),
"/US/US-WV/sistersville-22664" => array("type"=>"city", "pop"=>1413, "name_en"=>"Sistersville", "lang"=>"en", "lat"=>39.559548, "lon"=>-80.998502),
"/US/US-WV/smithers-22665" => array("type"=>"city", "pop"=>844, "name_en"=>"Smithers", "lang"=>"en", "lat"=>38.178508, "lon"=>-81.306500),
"/US/US-WV/smithfield-22666" => array("type"=>"town", "pop"=>161, "name_ru"=>"Смитфилд", "name_en"=>"Smithfield", "lang"=>"en", "lat"=>39.498854, "lon"=>-80.560648),
"/US/US-WV/sophia-22667" => array("type"=>"town", "pop"=>1234, "name_en"=>"Sophia", "lang"=>"en", "lat"=>37.707782, "lon"=>-81.256023),
"/US/US-WV/south-charleston-22668" => array("type"=>"city", "pop"=>12427, "name_en"=>"South Charleston", "lang"=>"en", "lat"=>38.356390, "lon"=>-81.707210),
"/US/US-WV/spencer-22669" => array("type"=>"city", "pop"=>2156, "name_ru"=>"Спенсер", "name_en"=>"Spencer", "lang"=>"en", "lat"=>38.804200, "lon"=>-81.356269),
"/US/US-WV/st.-albans-22656" => array("type"=>"city", "pop"=>10996, "name_en"=>"St. Albans", "lang"=>"en", "lat"=>38.378576, "lon"=>-81.816972),
"/US/US-WV/st.-marys-22657" => array("type"=>"city", "pop"=>1926, "name_ru"=>"Сент Мэрис", "name_en"=>"St. Marys", "lang"=>"en", "lat"=>39.402498, "lon"=>-81.195471),
"/US/US-WV/stanaford-22670" => array("type"=>"cdp", "pop"=>1706, "name_en"=>"Stanaford", "lang"=>"en", "lat"=>37.815010, "lon"=>-81.152028),
"/US/US-WV/star-city-22671" => array("type"=>"town", "pop"=>1654, "name_ru"=>"Стар Сити", "name_en"=>"Star City", "lang"=>"en", "lat"=>39.659359, "lon"=>-79.986593),
"/US/US-WV/stonewood-22672" => array("type"=>"city", "pop"=>1840, "name_en"=>"Stonewood", "lang"=>"en", "lat"=>39.248947, "lon"=>-80.305189),
"/US/US-WV/summersville-22673" => array("type"=>"town", "pop"=>3303, "name_en"=>"Summersville", "lang"=>"en", "lat"=>38.284424, "lon"=>-80.844129),
"/US/US-WV/sutton-22674" => array("type"=>"town", "pop"=>983, "name_ru"=>"Саттон", "name_en"=>"Sutton", "lang"=>"en", "lat"=>38.666742, "lon"=>-80.710719),
"/US/US-WV/switzer-22675" => array("type"=>"cdp", "pop"=>1004, "name_en"=>"Switzer", "lang"=>"en", "lat"=>37.786327, "lon"=>-81.973907),
"/US/US-WV/sylvester-22676" => array("type"=>"town", "pop"=>458, "name_ru"=>"Силвестер", "name_en"=>"Sylvester", "lang"=>"en", "lat"=>38.008109, "lon"=>-81.561010),
"/US/US-WV/teays-valley-22677" => array("type"=>"cdp", "pop"=>8436, "name_en"=>"Teays Valley", "lang"=>"en", "lat"=>38.448031, "lon"=>-81.927506),
"/US/US-WV/terra-alta-22678" => array("type"=>"town", "pop"=>1503, "name_en"=>"Terra Alta", "lang"=>"en", "lat"=>39.444250, "lon"=>-79.544230),
"/US/US-WV/thomas-22679" => array("type"=>"city", "pop"=>552, "name_en"=>"Thomas", "lang"=>"en", "lat"=>39.147375, "lon"=>-79.499097),
"/US/US-WV/thurmond-22680" => array("type"=>"town", "pop"=>7, "name_en"=>"Thurmond", "lang"=>"en", "lat"=>37.963759, "lon"=>-81.080294),
"/US/US-WV/tornado-22681" => array("type"=>"cdp", "pop"=>1006, "name_en"=>"Tornado", "lang"=>"en", "lat"=>38.332848, "lon"=>-81.855676),
"/US/US-WV/triadelphia-22682" => array("type"=>"town", "pop"=>800, "name_en"=>"Triadelphia", "lang"=>"en", "lat"=>40.050632, "lon"=>-80.627217),
"/US/US-WV/tunnelton-22683" => array("type"=>"town", "pop"=>347, "name_en"=>"Tunnelton", "lang"=>"en", "lat"=>39.393389, "lon"=>-79.746469),
"/US/US-WV/union-22684" => array("type"=>"town", "pop"=>561, "name_ru"=>"Юнион", "name_en"=>"Union", "lang"=>"en", "lat"=>37.590630, "lon"=>-80.543192),
"/US/US-WV/valley-grove-22685" => array("type"=>"village", "pop"=>394, "name_en"=>"Valley Grove", "lang"=>"en", "lat"=>40.091232, "lon"=>-80.570710),
"/US/US-WV/vienna-22686" => array("type"=>"city", "pop"=>10536, "name_ru"=>"Вьенна", "name_en"=>"Vienna", "lang"=>"en", "lat"=>39.322674, "lon"=>-81.538177),
"/US/US-WV/war-22687" => array("type"=>"city", "pop"=>646, "name_en"=>"War", "lang"=>"en", "lat"=>37.302773, "lon"=>-81.680082),
"/US/US-WV/wardensville-22688" => array("type"=>"town", "pop"=>258, "name_en"=>"Wardensville", "lang"=>"en", "lat"=>39.076465, "lon"=>-78.593692),
"/US/US-WV/washington-22689" => array("type"=>"cdp", "pop"=>1030, "name_ru"=>"Вашингтон", "name_en"=>"Washington", "lang"=>"en", "lat"=>39.240904, "lon"=>-81.673359),
"/US/US-WV/wayne-22690" => array("type"=>"town", "pop"=>1121, "name_ru"=>"Уэйн", "name_en"=>"Wayne", "lang"=>"en", "lat"=>38.220891, "lon"=>-82.442710),
"/US/US-WV/weirton-22691" => array("type"=>"city", "pop"=>18748, "name_en"=>"Weirton", "lang"=>"en", "lat"=>40.400942, "lon"=>-80.564122),
"/US/US-WV/welch-22692" => array("type"=>"city", "pop"=>2220, "name_en"=>"Welch", "lang"=>"en", "lat"=>37.432818, "lon"=>-81.581503),
"/US/US-WV/wellsburg-22693" => array("type"=>"city", "pop"=>2591, "name_ru"=>"Уэллсбург", "name_en"=>"Wellsburg", "lang"=>"en", "lat"=>40.281700, "lon"=>-80.609834),
"/US/US-WV/west-hamlin-22694" => array("type"=>"town", "pop"=>686, "name_en"=>"West Hamlin", "lang"=>"en", "lat"=>38.283405, "lon"=>-82.195339),
"/US/US-WV/west-liberty-22695" => array("type"=>"town", "pop"=>1233, "name_ru"=>"Вэст Либерти", "name_en"=>"West Liberty", "lang"=>"en", "lat"=>40.167264, "lon"=>-80.594829),
"/US/US-WV/west-logan-22696" => array("type"=>"town", "pop"=>389, "name_en"=>"West Logan", "lang"=>"en", "lat"=>37.867936, "lon"=>-81.993041),
"/US/US-WV/west-milford-22697" => array("type"=>"town", "pop"=>649, "name_en"=>"West Milford", "lang"=>"en", "lat"=>39.205674, "lon"=>-80.404862),
"/US/US-WV/west-union-22700" => array("type"=>"town", "pop"=>777, "name_ru"=>"Вэст Юнион", "name_en"=>"West Union", "lang"=>"en", "lat"=>39.296344, "lon"=>-80.775631),
"/US/US-WV/weston-22698" => array("type"=>"city", "pop"=>4262, "name_ru"=>"Вэстон", "name_en"=>"Weston", "lang"=>"en", "lat"=>39.041631, "lon"=>-80.468826),
"/US/US-WV/westover-22699" => array("type"=>"city", "pop"=>4088, "name_en"=>"Westover", "lang"=>"en", "lat"=>39.635128, "lon"=>-79.976359),
"/US/US-WV/wheeling-22701" => array("type"=>"city", "pop"=>28913, "name_ru"=>"Уилинг", "name_en"=>"Wheeling", "lang"=>"en", "lat"=>40.075067, "lon"=>-80.696995),
"/US/US-WV/white-sulphur-springs-22702" => array("type"=>"city", "pop"=>2279, "name_en"=>"White Sulphur Springs", "lang"=>"en", "lat"=>37.798833, "lon"=>-80.297345),
"/US/US-WV/whitesville-22703" => array("type"=>"town", "pop"=>504, "name_ru"=>"Уайтсвилл", "name_en"=>"Whitesville", "lang"=>"en", "lat"=>37.979473, "lon"=>-81.536045),
"/US/US-WV/williamson-22704" => array("type"=>"city", "pop"=>3053, "name_ru"=>"Уильямсон", "name_en"=>"Williamson", "lang"=>"en", "lat"=>37.673662, "lon"=>-82.272740),
"/US/US-WV/williamstown-22705" => array("type"=>"city", "pop"=>2982, "name_ru"=>"Уильямстаун", "name_en"=>"Williamstown", "lang"=>"en", "lat"=>39.401212, "lon"=>-81.450901),
"/US/US-WV/winfield-22706" => array("type"=>"town", "pop"=>2055, "name_ru"=>"Винфилд", "name_en"=>"Winfield", "lang"=>"en", "lat"=>38.525797, "lon"=>-81.883104),
"/US/US-WV/womelsdorf-(coalton)-22707" => array("type"=>"town", "pop"=>247, "name_en"=>"Womelsdorf (Coalton)", "lang"=>"en", "lat"=>38.897662, "lon"=>-79.964452),
"/US/US-WV/worthington-22708" => array("type"=>"town", "pop"=>163, "name_ru"=>"Уортингтон", "name_en"=>"Worthington", "lang"=>"en", "lat"=>39.458041, "lon"=>-80.261911),
"/US/US-WY" => array("type"=>"state", "pop"=>532668, "name_ru"=>"Вайоминг", "name_en"=>"Wyoming", "lang"=>"en"),
"/US/US-WY/afton-23327" => array("type"=>"town", "pop"=>1864, "name_ru"=>"Афтон", "name_en"=>"Afton", "lang"=>"en", "lat"=>42.726335, "lon"=>-110.928370),
"/US/US-WY/albin-23328" => array("type"=>"town", "pop"=>116, "name_en"=>"Albin", "lang"=>"en", "lat"=>41.417104, "lon"=>-104.099658),
"/US/US-WY/alpine-23329" => array("type"=>"town", "pop"=>810, "name_ru"=>"Алпайн", "name_en"=>"Alpine", "lang"=>"en", "lat"=>43.160846, "lon"=>-111.013629),
"/US/US-WY/antelope-valley-crestview-23330" => array("type"=>"cdp", "pop"=>1099, "name_en"=>"Antelope Valley-Crestview", "lang"=>"en", "lat"=>44.225363, "lon"=>-105.459523),
"/US/US-WY/arapahoe-23331" => array("type"=>"cdp", "pop"=>393, "name_en"=>"Arapahoe", "lang"=>"en", "lat"=>42.970496, "lon"=>-108.475063),
"/US/US-WY/baggs-23332" => array("type"=>"town", "pop"=>400, "name_en"=>"Baggs", "lang"=>"en", "lat"=>41.035599, "lon"=>-107.655888),
"/US/US-WY/bairoil-23333" => array("type"=>"town", "pop"=>96, "name_en"=>"Bairoil", "lang"=>"en", "lat"=>42.237978, "lon"=>-107.558892),
"/US/US-WY/bar-nunn-23334" => array("type"=>"town", "pop"=>1828, "name_en"=>"Bar Nunn", "lang"=>"en", "lat"=>42.926212, "lon"=>-106.346333),
"/US/US-WY/basin-23335" => array("type"=>"town", "pop"=>1243, "name_en"=>"Basin", "lang"=>"en", "lat"=>44.379585, "lon"=>-108.042468),
"/US/US-WY/big-piney-23336" => array("type"=>"town", "pop"=>495, "name_en"=>"Big Piney", "lang"=>"en", "lat"=>42.541026, "lon"=>-110.117122),
"/US/US-WY/buffalo-23337" => array("type"=>"city", "pop"=>4832, "name_ru"=>"Буффало", "name_en"=>"Buffalo", "lang"=>"en", "lat"=>44.342282, "lon"=>-106.717437),
"/US/US-WY/burlington-23338" => array("type"=>"town", "pop"=>253, "name_ru"=>"Бёрлингтон", "name_en"=>"Burlington", "lang"=>"en", "lat"=>44.447436, "lon"=>-108.431733),
"/US/US-WY/burns-23339" => array("type"=>"town", "pop"=>314, "name_ru"=>"Бернс", "name_en"=>"Burns", "lang"=>"en", "lat"=>41.190139, "lon"=>-104.358664),
"/US/US-WY/byron-23340" => array("type"=>"town", "pop"=>557, "name_ru"=>"Байрон", "name_en"=>"Byron", "lang"=>"en", "lat"=>44.795395, "lon"=>-108.507590),
"/US/US-WY/casper-23341" => array("type"=>"city", "pop"=>54047, "name_en"=>"Casper", "lang"=>"en", "lat"=>42.832600, "lon"=>-106.329340),
"/US/US-WY/cheyenne-23342" => array("type"=>"city", "pop"=>56915, "name_en"=>"Cheyenne", "lang"=>"en", "lat"=>41.145450, "lon"=>-104.792349),
"/US/US-WY/chugwater-23343" => array("type"=>"town", "pop"=>221, "name_en"=>"Chugwater", "lang"=>"en", "lat"=>41.755692, "lon"=>-104.821137),
"/US/US-WY/clearmont-23344" => array("type"=>"town", "pop"=>114, "name_ru"=>"Клирмонт", "name_en"=>"Clearmont", "lang"=>"en", "lat"=>44.640288, "lon"=>-106.380465),
"/US/US-WY/cody-23345" => array("type"=>"city", "pop"=>9309, "name_en"=>"Cody", "lang"=>"en", "lat"=>44.519488, "lon"=>-109.054149),
"/US/US-WY/cokeville-23346" => array("type"=>"town", "pop"=>488, "name_en"=>"Cokeville", "lang"=>"en", "lat"=>42.083208, "lon"=>-110.955024),
"/US/US-WY/cowley-23347" => array("type"=>"town", "pop"=>614, "name_en"=>"Cowley", "lang"=>"en", "lat"=>44.885840, "lon"=>-108.469776),
"/US/US-WY/dayton-23348" => array("type"=>"town", "pop"=>757, "name_ru"=>"Дейтон", "name_en"=>"Dayton", "lang"=>"en", "lat"=>44.872872, "lon"=>-107.263368),
"/US/US-WY/deaver-23349" => array("type"=>"town", "pop"=>177, "name_en"=>"Deaver", "lang"=>"en", "lat"=>44.889190, "lon"=>-108.594843),
"/US/US-WY/diamondville-23350" => array("type"=>"town", "pop"=>663, "name_en"=>"Diamondville", "lang"=>"en", "lat"=>41.777736, "lon"=>-110.536440),
"/US/US-WY/dixon-23351" => array("type"=>"town", "pop"=>81, "name_ru"=>"Диксон", "name_en"=>"Dixon", "lang"=>"en", "lat"=>41.034112, "lon"=>-107.534284),
"/US/US-WY/douglas-23352" => array("type"=>"city", "pop"=>5971, "name_ru"=>"Дуглас", "name_en"=>"Douglas", "lang"=>"en", "lat"=>42.755807, "lon"=>-105.392726),
"/US/US-WY/dubois-23353" => array("type"=>"town", "pop"=>1053, "name_ru"=>"Дюбуа", "name_en"=>"Dubois", "lang"=>"en", "lat"=>43.536278, "lon"=>-109.642095),
"/US/US-WY/east-thermopolis-23354" => array("type"=>"town", "pop"=>264, "name_en"=>"East Thermopolis", "lang"=>"en", "lat"=>43.645409, "lon"=>-108.198667),
"/US/US-WY/edgerton-23355" => array("type"=>"town", "pop"=>176, "name_ru"=>"Эджертон", "name_en"=>"Edgerton", "lang"=>"en", "lat"=>43.414476, "lon"=>-106.247241),
"/US/US-WY/elk-mountain-23356" => array("type"=>"town", "pop"=>200, "name_en"=>"Elk Mountain", "lang"=>"en", "lat"=>41.687833, "lon"=>-106.412854),
"/US/US-WY/ethete-23357" => array("type"=>"cdp", "pop"=>1059, "name_en"=>"Ethete", "lang"=>"en", "lat"=>43.021747, "lon"=>-108.757036),
"/US/US-WY/evanston-23358" => array("type"=>"city", "pop"=>11781, "name_ru"=>"Эванстон", "name_en"=>"Evanston", "lang"=>"en", "lat"=>41.258855, "lon"=>-110.963440),
"/US/US-WY/evansville-23359" => array("type"=>"town", "pop"=>2393, "name_ru"=>"Эвансвилл", "name_en"=>"Evansville", "lang"=>"en", "lat"=>42.872464, "lon"=>-106.258717),
"/US/US-WY/fort-laramie-23360" => array("type"=>"town", "pop"=>228, "name_en"=>"Fort Laramie", "lang"=>"en", "lat"=>42.212385, "lon"=>-104.516596),
"/US/US-WY/fort-washakie-23361" => array("type"=>"cdp", "pop"=>1334, "name_en"=>"Fort Washakie", "lang"=>"en", "lat"=>43.007936, "lon"=>-108.925873),
"/US/US-WY/fox-farm-college-23362" => array("type"=>"cdp", "pop"=>2965, "name_en"=>"Fox Farm-College", "lang"=>"en", "lat"=>41.110092, "lon"=>-104.787429),
"/US/US-WY/frannie-23363" => array("type"=>"town", "pop"=>208, "name_en"=>"Frannie", "lang"=>"en", "lat"=>44.970852, "lon"=>-108.619685),
"/US/US-WY/gillette-23364" => array("type"=>"city", "pop"=>26871, "name_en"=>"Gillette", "lang"=>"en", "lat"=>44.281783, "lon"=>-105.517908),
"/US/US-WY/glendo-23365" => array("type"=>"town", "pop"=>215, "name_en"=>"Glendo", "lang"=>"en", "lat"=>42.504019, "lon"=>-105.024897),
"/US/US-WY/glenrock-23366" => array("type"=>"town", "pop"=>2423, "name_en"=>"Glenrock", "lang"=>"en", "lat"=>42.857122, "lon"=>-105.857091),
"/US/US-WY/grand-encampment-23367" => array("type"=>"town", "pop"=>490, "name_en"=>"Grand Encampment", "lang"=>"en", "lat"=>41.208778, "lon"=>-106.794638),
"/US/US-WY/granger-23368" => array("type"=>"town", "pop"=>145, "name_ru"=>"Грейнджер", "name_en"=>"Granger", "lang"=>"en", "lat"=>41.594081, "lon"=>-109.965904),
"/US/US-WY/green-river-23369" => array("type"=>"city", "pop"=>12149, "name_en"=>"Green River", "lang"=>"en", "lat"=>41.515850, "lon"=>-109.467866),
"/US/US-WY/greybull-23370" => array("type"=>"town", "pop"=>1739, "name_en"=>"Greybull", "lang"=>"en", "lat"=>44.491035, "lon"=>-108.057227),
"/US/US-WY/guernsey-23371" => array("type"=>"town", "pop"=>1072, "name_ru"=>"Гернси", "name_en"=>"Guernsey", "lang"=>"en", "lat"=>42.266306, "lon"=>-104.744330),
"/US/US-WY/hanna-23372" => array("type"=>"town", "pop"=>866, "name_en"=>"Hanna", "lang"=>"en", "lat"=>41.869922, "lon"=>-106.558690),
"/US/US-WY/hartville-23373" => array("type"=>"town", "pop"=>70, "name_ru"=>"Хартвилл", "name_en"=>"Hartville", "lang"=>"en", "lat"=>42.327350, "lon"=>-104.724230),
"/US/US-WY/hudson-23374" => array("type"=>"town", "pop"=>429, "name_ru"=>"Хадсон", "name_en"=>"Hudson", "lang"=>"en", "lat"=>42.901963, "lon"=>-108.581643),
"/US/US-WY/hulett-23375" => array("type"=>"town", "pop"=>471, "name_en"=>"Hulett", "lang"=>"en", "lat"=>44.686102, "lon"=>-104.599986),
"/US/US-WY/jackson-23376" => array("type"=>"town", "pop"=>9806, "name_ru"=>"Джексон", "name_en"=>"Jackson", "lang"=>"en", "lat"=>43.473998, "lon"=>-110.763878),
"/US/US-WY/kaycee-23377" => array("type"=>"town", "pop"=>290, "name_en"=>"Kaycee", "lang"=>"en", "lat"=>43.709835, "lon"=>-106.637739),
"/US/US-WY/kemmerer-23378" => array("type"=>"city", "pop"=>2468, "name_en"=>"Kemmerer", "lang"=>"en", "lat"=>41.782150, "lon"=>-110.551428),
"/US/US-WY/kirby-23379" => array("type"=>"town", "pop"=>55, "name_en"=>"Kirby", "lang"=>"en", "lat"=>43.804107, "lon"=>-108.180170),
"/US/US-WY/la-barge-23380" => array("type"=>"town", "pop"=>466, "name_en"=>"La Barge", "lang"=>"en", "lat"=>42.261197, "lon"=>-110.196259),
"/US/US-WY/la-grange-23381" => array("type"=>"town", "pop"=>326, "name_ru"=>"Ла Гранж", "name_en"=>"La Grange", "lang"=>"en", "lat"=>41.639392, "lon"=>-104.163790),
"/US/US-WY/lander-23382" => array("type"=>"city", "pop"=>7264, "name_en"=>"Lander", "lang"=>"en", "lat"=>42.830022, "lon"=>-108.728566),
"/US/US-WY/laramie-23383" => array("type"=>"city", "pop"=>27664, "name_en"=>"Laramie", "lang"=>"en", "lat"=>41.310880, "lon"=>-105.583037),
"/US/US-WY/lingle-23384" => array("type"=>"town", "pop"=>484, "name_en"=>"Lingle", "lang"=>"en", "lat"=>42.138787, "lon"=>-104.345970),
"/US/US-WY/lost-springs-23385" => array("type"=>"town", "pop"=>1, "name_ru"=>"Лост Спрингс", "name_en"=>"Lost Springs", "lang"=>"en", "lat"=>42.765286, "lon"=>-104.923796),
"/US/US-WY/lovell-23386" => array("type"=>"town", "pop"=>2276, "name_en"=>"Lovell", "lang"=>"en", "lat"=>44.836834, "lon"=>-108.391508),
"/US/US-WY/lusk-23387" => array("type"=>"town", "pop"=>1479, "name_en"=>"Lusk", "lang"=>"en", "lat"=>42.760662, "lon"=>-104.453570),
"/US/US-WY/lyman-23388" => array("type"=>"town", "pop"=>2009, "name_ru"=>"Лиман", "name_en"=>"Lyman", "lang"=>"en", "lat"=>41.327776, "lon"=>-110.300280),
"/US/US-WY/manderson-23389" => array("type"=>"town", "pop"=>100, "name_en"=>"Manderson", "lang"=>"en", "lat"=>44.269939, "lon"=>-107.963440),
"/US/US-WY/manville-23390" => array("type"=>"town", "pop"=>100, "name_en"=>"Manville", "lang"=>"en", "lat"=>42.779486, "lon"=>-104.616861),
"/US/US-WY/marbleton-23391" => array("type"=>"town", "pop"=>1084, "name_en"=>"Marbleton", "lang"=>"en", "lat"=>42.557464, "lon"=>-110.099385),
"/US/US-WY/medicine-bow-23392" => array("type"=>"town", "pop"=>267, "name_en"=>"Medicine Bow", "lang"=>"en", "lat"=>41.899762, "lon"=>-106.201031),
"/US/US-WY/meeteetse-23393" => array("type"=>"town", "pop"=>342, "name_en"=>"Meeteetse", "lang"=>"en", "lat"=>44.155540, "lon"=>-108.871270),
"/US/US-WY/midwest-23394" => array("type"=>"town", "pop"=>435, "name_en"=>"Midwest", "lang"=>"en", "lat"=>43.411316, "lon"=>-106.277531),
"/US/US-WY/mills-23395" => array("type"=>"town", "pop"=>3143, "name_en"=>"Mills", "lang"=>"en", "lat"=>42.845800, "lon"=>-106.375919),
"/US/US-WY/moorcroft-23396" => array("type"=>"town", "pop"=>892, "name_en"=>"Moorcroft", "lang"=>"en", "lat"=>44.264894, "lon"=>-104.951628),
"/US/US-WY/mountain-view-23397" => array("type"=>"cdp", "pop"=>1202, "name_ru"=>"Маунтин Вью", "name_en"=>"Mountain View", "lang"=>"en", "lat"=>42.884067, "lon"=>-106.444889),
"/US/US-WY/mountain-view-23398" => array("type"=>"town", "pop"=>1189, "name_ru"=>"Маунтин Вью", "name_en"=>"Mountain View", "lang"=>"en", "lat"=>41.271465, "lon"=>-110.335603),
"/US/US-WY/newcastle-23399" => array("type"=>"city", "pop"=>3390, "name_en"=>"Newcastle", "lang"=>"en", "lat"=>43.849002, "lon"=>-104.210143),
"/US/US-WY/north-rock-springs-23400" => array("type"=>"cdp", "pop"=>2471, "name_en"=>"North Rock Springs", "lang"=>"en", "lat"=>41.642290, "lon"=>-109.252488),
"/US/US-WY/opal-23401" => array("type"=>"town", "pop"=>95, "name_en"=>"Opal", "lang"=>"en", "lat"=>41.769960, "lon"=>-110.320816),
"/US/US-WY/pavillion-23402" => array("type"=>"town", "pop"=>169, "name_en"=>"Pavillion", "lang"=>"en", "lat"=>43.244010, "lon"=>-108.690756),
"/US/US-WY/pine-bluffs-23403" => array("type"=>"town", "pop"=>1155, "name_en"=>"Pine Bluffs", "lang"=>"en", "lat"=>41.180390, "lon"=>-104.068199),
"/US/US-WY/pine-haven-23405" => array("type"=>"town", "pop"=>380, "name_en"=>"Pine Haven", "lang"=>"en", "lat"=>44.348828, "lon"=>-104.810080),
"/US/US-WY/pinedale-23404" => array("type"=>"town", "pop"=>2162, "name_en"=>"Pinedale", "lang"=>"en", "lat"=>42.866408, "lon"=>-109.864702),
"/US/US-WY/powell-23406" => array("type"=>"city", "pop"=>5524, "name_ru"=>"Пауэлл", "name_en"=>"Powell", "lang"=>"en", "lat"=>44.791832, "lon"=>-108.735895),
"/US/US-WY/rafter-j-ranch-23407" => array("type"=>"cdp", "pop"=>1092, "name_en"=>"Rafter J Ranch", "lang"=>"en", "lat"=>43.426248, "lon"=>-110.798440),
"/US/US-WY/ranchester-23408" => array("type"=>"town", "pop"=>798, "name_en"=>"Ranchester", "lang"=>"en", "lat"=>44.907321, "lon"=>-107.164075),
"/US/US-WY/ranchettes-23409" => array("type"=>"cdp", "pop"=>4038, "name_en"=>"Ranchettes", "lang"=>"en", "lat"=>41.218517, "lon"=>-104.789690),
"/US/US-WY/rawlins-23410" => array("type"=>"city", "pop"=>8740, "name_en"=>"Rawlins", "lang"=>"en", "lat"=>41.784507, "lon"=>-107.227891),
"/US/US-WY/riverside-23411" => array("type"=>"town", "pop"=>63, "name_ru"=>"Риверсайд", "name_en"=>"Riverside", "lang"=>"en", "lat"=>41.214401, "lon"=>-106.780674),
"/US/US-WY/riverton-23412" => array("type"=>"city", "pop"=>10032, "name_ru"=>"Ривертон", "name_en"=>"Riverton", "lang"=>"en", "lat"=>43.035566, "lon"=>-108.414785),
"/US/US-WY/rock-river-23413" => array("type"=>"town", "pop"=>213, "name_en"=>"Rock River", "lang"=>"en", "lat"=>41.731202, "lon"=>-105.972728),
"/US/US-WY/rock-springs-23414" => array("type"=>"city", "pop"=>20200, "name_en"=>"Rock Springs", "lang"=>"en", "lat"=>41.598138, "lon"=>-109.221458),
"/US/US-WY/rolling-hills-23415" => array("type"=>"town", "pop"=>505, "name_ru"=>"Роллинг Хилс", "name_en"=>"Rolling Hills", "lang"=>"en", "lat"=>42.902385, "lon"=>-105.845998),
"/US/US-WY/saratoga-23416" => array("type"=>"town", "pop"=>1759, "name_ru"=>"Саратога", "name_en"=>"Saratoga", "lang"=>"en", "lat"=>41.451481, "lon"=>-106.812226),
"/US/US-WY/sheridan-23417" => array("type"=>"city", "pop"=>17197, "name_ru"=>"Шеридан", "name_en"=>"Sheridan", "lang"=>"en", "lat"=>44.792090, "lon"=>-106.957096),
"/US/US-WY/shoshoni-23418" => array("type"=>"town", "pop"=>689, "name_en"=>"Shoshoni", "lang"=>"en", "lat"=>43.237920, "lon"=>-108.102874),
"/US/US-WY/sinclair-23419" => array("type"=>"town", "pop"=>405, "name_en"=>"Sinclair", "lang"=>"en", "lat"=>41.777477, "lon"=>-107.116742),
"/US/US-WY/sleepy-hollow-23420" => array("type"=>"cdp", "pop"=>1194, "name_ru"=>"Слипи Холлоу", "name_en"=>"Sleepy Hollow", "lang"=>"en", "lat"=>44.241020, "lon"=>-105.421557),
"/US/US-WY/south-greeley-23421" => array("type"=>"cdp", "pop"=>3723, "name_en"=>"South Greeley", "lang"=>"en", "lat"=>41.097052, "lon"=>-104.805803),
"/US/US-WY/sundance-23422" => array("type"=>"town", "pop"=>1253, "name_en"=>"Sundance", "lang"=>"en", "lat"=>44.404702, "lon"=>-104.374235),
"/US/US-WY/superior-23423" => array("type"=>"town", "pop"=>237, "name_ru"=>"Супериор", "name_en"=>"Superior", "lang"=>"en", "lat"=>41.760645, "lon"=>-108.963580),
"/US/US-WY/ten-sleep-23424" => array("type"=>"town", "pop"=>315, "name_en"=>"Ten Sleep", "lang"=>"en", "lat"=>44.034733, "lon"=>-107.447351),
"/US/US-WY/thayne-23425" => array("type"=>"town", "pop"=>370, "name_en"=>"Thayne", "lang"=>"en", "lat"=>42.919834, "lon"=>-111.001058),
"/US/US-WY/thermopolis-23426" => array("type"=>"town", "pop"=>2971, "name_en"=>"Thermopolis", "lang"=>"en", "lat"=>43.647962, "lon"=>-108.213877),
"/US/US-WY/torrington-23427" => array("type"=>"town", "pop"=>5514, "name_ru"=>"Торрингтон", "name_en"=>"Torrington", "lang"=>"en", "lat"=>42.067539, "lon"=>-104.181973),
"/US/US-WY/upton-23428" => array("type"=>"town", "pop"=>907, "name_ru"=>"Аптон", "name_en"=>"Upton", "lang"=>"en", "lat"=>44.101435, "lon"=>-104.623495),
"/US/US-WY/van-tassell-23429" => array("type"=>"town", "pop"=>18, "name_en"=>"Van Tassell", "lang"=>"en", "lat"=>42.663646, "lon"=>-104.090655),
"/US/US-WY/wamsutter-23430" => array("type"=>"town", "pop"=>269, "name_en"=>"Wamsutter", "lang"=>"en", "lat"=>41.668870, "lon"=>-107.975896),
"/US/US-WY/warren-afb-23431" => array("type"=>"cdp", "pop"=>3832, "name_en"=>"Warren AFB", "lang"=>"en", "lat"=>41.153059, "lon"=>-104.860563),
"/US/US-WY/wheatland-23432" => array("type"=>"town", "pop"=>3298, "name_ru"=>"Уитленд", "name_en"=>"Wheatland", "lang"=>"en", "lat"=>42.051225, "lon"=>-104.958892),
"/US/US-WY/worland-23433" => array("type"=>"city", "pop"=>4958, "name_en"=>"Worland", "lang"=>"en", "lat"=>44.005402, "lon"=>-107.958034),
"/US/US-WY/wright-23434" => array("type"=>"town", "pop"=>1462, "name_ru"=>"Райт", "name_en"=>"Wright", "lang"=>"en", "lat"=>43.750697, "lon"=>-105.498077),
"/US/US-WY/yoder-23435" => array("type"=>"town", "pop"=>159, "name_en"=>"Yoder", "lang"=>"en", "lat"=>41.917590, "lon"=>-104.294551),
"/UY" => array("type"=>"country", "pop"=>3323906, "name_ru"=>"Уругвай", "name_en"=>"Uruguay", "lang"=>"es", "lat"=>-32.73, "lon"=>-56.65),
"/UZ" => array("type"=>"country", "pop"=>27555300, "name_ru"=>"Узбекистан", "name_uz"=>"O‘zbekiston", "name_en"=>"Uzbekistan", "lang"=>"uz", "lat"=>41.66, "lon"=>64.06),
"/VA" => array("type"=>"country", "pop"=>826, "name_la"=>"Vaticanæ", "name_ru"=>"Ватикан", "name_en"=>"Vatican", "lang"=>"la", "lat"=>41.90, "lon"=>12.45),
"/VC" => array("type"=>"country", "pop"=>117534, "name_ru"=>"Сент-Винсент и Гренадины", "name_en"=>"Saint Vincent and the Grenadines", "lang"=>"en", "lat"=>13.25, "lon"=>-61.2),
"/VE" => array("type"=>"country", "pop"=>27730469, "name_ru"=>"Венесуэла", "name_en"=>"Venezuela", "lang"=>"es", "lat"=>7.66, "lon"=>-65.21),
"/VN" => array("type"=>"country", "pop"=>83535576, "name_ru"=>"Вьетнам", "name_en"=>"Viet Nam", "lang"=>"vi", "lat"=>14.03, "lon"=>107.9),
"/WS" => array("type"=>"country", "pop"=>188540, "name_ru"=>"Самоа", "name_en"=>"Samoa", "lang"=>"sm", "lat"=>-14.2, "lon"=>-170.75),
"/YE" => array("type"=>"country", "pop"=>20727063, "name_ru"=>"Йемен", "name_en"=>"Yemen", "lang"=>"ar", "lat"=>15.00, "lon"=>48.00),
"/ZA" => array("type"=>"country", "pop"=>48700000, "name_ru"=>"ЮАР", "name_en"=>"South Africa", "lang"=>"af", "lat"=>-28.61, "lon"=>24.33),
"/ZM" => array("type"=>"country", "pop"=>11261795, "name_ru"=>"Замбия", "name_en"=>"Zambia", "lang"=>"en", "lat"=>-14.11, "lon"=>27.63),
"/ZW" => array("type"=>"country", "pop"=>12746990, "name_ru"=>"Зимбабве", "name_en"=>"Zimbabwe", "lang"=>"en", "lat"=>-19.01, "lon"=>30.01),
"_"=>"");
?>
